/* Array.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_array_
#define _H_qu_array_ 1

#include "Container.h"


/** Array
	The Array class.
*/
extern qu_class QuArrayClass;


/** qu_array
	qu_array represents an instance of an Array class.
	It is a pointer to the instance struct qu_array_s.

	See also ={QU_CONTAINER_HEAD}.

	All fields are read-only.

	See ={qu_array_set_type} to set the type for newly created Arrays.
*/
struct qu_array_s
{
	QU_CONTAINER_HEAD;
	long    cap;   /* allocated capacity <= QU_GC_MAX_ITEMS */
	qu_any *objs;  /* the items, array is qu_mem_alloc'ed */
	qu_any  type;  /* item type/validator, maybe 0 */
};


/** QuArrayEmpty
	This is a predefined frozen empty Array.

	Expression C{const []} in programs simply points to this object.

	@warn
	You may not modify this Array.
*/
extern qu_array QuArrayEmpty;


/** qu_array_is
	Whether or not an object is an Array.
*/
#define qu_array_is(x) qu_object_instof (x, QuArrayClass)


/** QU_ARRAY_FLAG
	Flags used by Array.

	Do not access directly. Use the accessors instead.
*/
#define QU_ARRAY_FLAG_PROTECT  QU_CONTAINER_FLAG_PROTECT /* protected? */
#define QU_ARRAY_FLAG_FROZEN   QU_CONTAINER_FLAG_FROZEN  /* immutable? */
#define QU_ARRAY_FLAG_FIXED    QU_CONTAINER_FLAG_FIXED   /* resizable? */
#define QU_ARRAY_FLAG_FORLAST  QU_FLAG_10                /* indicate *c in for a, b, *c */


/** qu_array_frozen qu_array_fixed
	qu_array_frozen checks whether or not the Array is mutable.

	qu_array_freeze must only be used after you have created a new Array and
	don't want it to be modified by others.

	qu_array_unfreeze must only be used on an Array you have created.
	Normally, to modify the Array then freeze it again.

	qu_array_fixed checks whether or not the Array is resizable.

	You should have no reason to set/unset the QU_ARRAY_FLAG_FIXED in C level.
*/
#define qu_array_frozen(self)   qu_container_frozen   (self)
#define qu_array_freeze(self)   qu_container_freeze   (self)
#define qu_array_unfreeze(self) qu_container_unfreeze (self)
#define qu_array_fixed(self)    qu_container_fixed    (self)


/** qu_array_forlast
	This is strictly for the interpreter/compiler.

	The QU_ARRAY_FLAG_FORLAST flag is used to indicate that the Array is a collection
	of the last items v{z} in a C{for x, y, *z} syntax.

	Once set by the compiler, the flag must never be unset.
*/
#define qu_array_forlast(self)     qu_flag_test (self, QU_ARRAY_FLAG_FORLAST)
#define qu_array_forlast_set(self) qu_flag_set  (self, QU_ARRAY_FLAG_FORLAST)


/** qu_array_new
	Various constructors.

	In any case v{len} must be within 0..={QU_GC_MAX_ITEMS}.

	qu_array_new creates a new regular GC Array with a capacity enough to hold
	a specified number of items.

	qu_array_new_root creates a root GC Array object.
	It is always marked by the GC.

	qu_array_new_static creates a static Array object.
	It is never marked by the GC, thus you must make sure that all items are
	={qu_gc_is_static} objects. No checks are made.

	qu_array_new_argv and qu_array_new_ex create a new Array with the specified items.
*/
extern qu_array qu_array_new        (long len);
extern qu_array qu_array_new_root   (long len);
extern qu_array qu_array_new_static (long len);
extern qu_array qu_array_new_argv   (int argc, qu_any *argv);
extern qu_array qu_array_new_ex     (long len, ...);


/** qu_array_to_static
	Create an Array which is never garbage collected nor marked from an
	existing one. All items must satisfy ={qu_gc_is_static} otherwise
	EValue is thrown.
*/
extern qu_array qu_array_to_static (qu_array self);


/** qu_array_copy
	Return a shallow copy.
*/
#define qu_array_copy(self) qu__array__copy(self)


/** qu_array_clear qu_array_CLEAR
	Clears the content.

	qu_array_clear throws if the Array is immutable while qu_array_CLEAR does not
	care and does not resize the Array.
*/
#define qu_array_clear(self) qu__array__clear (self)
#define qu_array_CLEAR(self) QU_BEGIN (self)->len = 0; QU_END


/** qu_array_flatten
	Return a new Array by flattening the arguments.

	This is used to call functions/methods with flattened arguments by the
	interpreter. You should have no reason to call this function directly.
*/
extern qu_array qu_array_flatten (int argc, qu_any *argv);


/** qu_array_make_space
	Make sure the Array has enough space for v{more} items, grow if necessary.

	Return 1 on success, return 0 and throw EMemory on failure.
*/
extern int qu_array_make_space (qu_array self, long more);


/** qu_array_contains
	Whether the Array contains a specified object.

	This never fails and return 1 or 0.
*/
#define qu_array_contains(self, x) (qu__array__in (self, x) == QuTrue)


/** qu_array_like
	Whether two Arrays are of the same length and contain the same items in the
	same order.

	This never fails and return 1 or 0.
*/
#define qu_array_like(self, x) (qu__array__like (self, x) == QuTrue)


/** qu_array_push qu_array_push_quick
	Push an item.

	qu_array_push returns self on success, throws on error.

	qu_array_push_quick does not care whether the Array is mutable and assumes
	that the Array can grow if necessary, that is, the current number of items
	is less than ={QU_GC_MAX_ITEMS}.
*/
#define qu_array_push(self, x) qu__array__push (self, x)
extern void qu_array_push_quick (qu_array self, qu_any x);


/** qu_array_pop qu_array_pop_quick
	Pops an item from the end.

	qu_array_pop throws if the Array is empty or is immutable.

	qu_array_pop_quick simply remove the last item and does nothing if the
	Array is empty. It does not care whether the Array is immutable and does
	not resize the Array. Use with care.
*/
#define qu_array_pop(self)       qu__array__pop (self)
#define qu_array_pop_quick(self) QU_BEGIN if ((self)->len) --(self)->len; QU_END


/** qu_array_of qu_array_set_type
	Set the type of the items.

	qu_array_of returns self on success, returns 0 and throws on error.

	Use qu_array_set_type to set the type if you have just created the Array.
	Notice the assertions in the macros.
*/
#define qu_array_of(self, type) qu__array__of (self, type)

#define qu_array_set_type(self, type) \
QU_BEGIN \
	assert (qu_object_valid_type (type)); \
	assert ((self)->type == 0); \
	(self)->type = type; \
QU_END


/** qu_array_typeof
	Whether all items are of a specified type.
*/
#define qu_array_typeof(self, type) (qu__array_typeof (self, type) == QuTrue)


/** qu_array_index qu_array_rindex
	qu_array_index returns the index of v{x} optionally starting from offset v{i} going
	forwards while qu_array_rindex does the same but backwards.

	Thess functions take care of negative index in the usual manner.

	Both return -1 if the Array does not contain the object.
*/
extern long qu_array_index  (qu_array self, qu_any x, long i);
extern long qu_array_rindex (qu_array self, qu_any x, long i);


/** qu_array_GETITEM qu_array_SETITEM
	Get or set an item.

	@warn
	Make sure C{0 <= i && i < self->len}.
*/
#define qu_array_GETITEM(self, i)    ((self)->objs[i])
#define qu_array_SETITEM(self, i, x) QU_BEGIN (self)->objs[i] = x; QU_END


/** qu_array_del_quick
	qu_array_del_quick removes v{x} by comparing the pointers (not by calling
	the __eq method). It does not care whether the Array is mutable.

	qu_array_DELFIRST removes the first item if the Array is not empty, does
	nothing otherwise. It does not care whether the Array is mutable.
*/
extern void qu_array_del_quick (qu_array self, qu_any x);

#define qu_array_DELFIRST(self) \
QU_BEGIN \
	if ((self)->len) \
		QU_OBJECT_LIST_COPY ((self)->objs, (self)->objs + 1, --((self)->len)); \
QU_END


/** qu_array_getslice
	Get a slice from the Array.

	Return 0 without throwing if the index is out of bound.

	This function takes care of negative index in the usual manner.
*/
extern qu_array qu_array_getslice (qu_array self, long index, long len);


/** qu_array_as
	Various representations.
*/
extern qu_set  qu_array_as_set  (qu_array self);
extern qu_dict qu_array_as_dict (qu_array self);
#define qu_array_as_string(self)  qu__array__string (self)
#define qu_array_as_boolean(self) qu_container_as_boolean (self)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_array__setup ();
extern void qu_array__init  ();

/* for efficiency - may only be called from this file */
extern qu_array   qu__array__clear  (qu_array self);
extern qu_array   qu__array__copy   (qu_array self);
extern qu_boolean qu__array__in     (qu_array self, qu_any x);
extern qu_boolean qu__array__like   (qu_array self, qu_any x);
extern qu_array   qu__array__of     (qu_array self, qu_any type);
extern qu_array   qu__array__of     (qu_array self, qu_any type);
extern qu_string  qu__array__string (qu_array self);
extern qu_array   qu__array__push   (qu_array self, qu_any x);
extern qu_boolean qu__array_typeof  (qu_array self, qu_any type);


#endif /*_H_qu_array_*/
