/* Big.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_big_
#define _H_qu_big_ 1


/** Big
	This is the final Big class.
*/
extern qu_class QuBigClass;


/** QU_BIG_XXX
	Constants used by the Big class.

	qu_big_digit_t is the C type for the digits.

	QU_BIG_BASE is the digit base.

	QU_BIG_FIG is the number of ASCII characters in one digit.
	The relation between qu_big_digit_t and QU_BIG_BASE is C{QU_BIG_BASE = 10 ^ QU_BIG_FIG}.

	QU_BIG_MAXEXPO is the maximum absolute value of the exponent.
	It is such that we can always do +/- without overflow.

	QU_BIG_MINPREC and QU_BIG_MAXPREC are the minimum and maximum precision.
	Multiply by QU_BIG_FIG to get the number of significant digits.
	The value must fit in an ={Int}.

	QU_BIG_MAXLEN is the practical maximum allocatable number of digits.
	It must be such that it is always possible to convert a ={Long} to a Big
	and vice versa. It must also fit in an ={Int}.
*/
#define qu_big_digit_t  unsigned long
#define QU_BIG_BASE     10000
#define QU_BIG_FIG      4
#define QU_BIG_FORMAT   "%04lu"
#define QU_BIG_FIG_ZERO  "0000"
#define QU_BIG_MAXEXPO  (INT_MAX / SIZEOF_LONG / QU_BIG_FIG / 16)
#define QU_BIG_MINPREC  2
#define QU_BIG_MAXPREC  (INT_MAX / SIZEOF_LONG / 8)
#define QU_BIG_MAXLEN   (INT_MAX / SIZEOF_LONG / 4)


/** qu_big
	qu_big represents an instance of a Big class.
	It is a pointer to the instance struct qu_big_s.

	@code
	value = 0.xxxxxxxxxxxxxxxx * BASE ^ expo
	          <-----len------>
	          MSB          LSB digits
*/
struct qu_big_s
{
	QU_OBJECT_HEAD;
	qu_big_digit_t *digits; /* the digits, MSB first */
	long            len;    /* actual number of digits */
	long            expo;   /* the exponent */
	long            cap;    /* allocated capacity of digits */
};


/** QU_BIG_FLAG
	Flags used by Big objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_BIG_FLAG_NEG  QU_FLAG_0  /* negative? same as QU_LONG_FLAG_NEG */
#define QU_BIG_FLAG_TMP  QU_FLAG_9  /* on local C stack? same as QU_LONG_FLAG_TMP */


/** QuBigN
	Predefined system Big objects.

	See ={qu_big_new_static} to create your own.
*/
extern qu_big QuBig0;    /*  0.0  */
extern qu_big QuBig1;    /*  1.0  */
extern qu_big QuBig2;    /*  2.0  */
extern qu_big QuBig3;    /*  3.0  */
extern qu_big QuBig4;    /*  4.0  */
extern qu_big QuBig5;    /*  5.0  */
extern qu_big QuBig6;    /*  6.0  */
extern qu_big QuBig7;    /*  7.0  */
extern qu_big QuBig8;    /*  8.0  */
extern qu_big QuBig9;    /*  9.0  */
extern qu_big QuBig10;   /* 10.0  */
extern qu_big QuBig16;   /* 16.0  */
extern qu_big QuBig_1;   /* -1.0  */
extern qu_big QuBig_2;   /* -2.0  */
extern qu_big QuBig_4;   /* -4.0  */
extern qu_big QuBig_8;   /* -8.0  */
extern qu_big QuBig1_2;  /*  0.5  */
extern qu_big QuBig_1_2; /* -0.5  */
extern qu_big QuBig1_4;  /*  0.25 */
extern qu_big QuBig_1_4; /* -0.25 */


/** __qu_big_precision
	Set or get the current precision.

	@warn
	Its value must be between ={QU_BIG_MINPREC} and ={QU_BIG_MAXPREC}.
	So be careful.

	@xmp
		long old;
		__qu_big_precision_push (old, 16);
		{
			... no throwing here ...
		}
		__qu_big_precision_pop (old);
*/
#define __qu_big_precision() __QuBigPrecision

#define __qu_big_precision_push(variable, newvalue) \
QU_BEGIN \
	variable = __QuBigPrecision; \
	__QuBigPrecision = (value); \
	assert (QU_BIG_MINPREC <= __QuBigPrecision && __QuBigPrecision<= QU_BIG_MAX_PREC); \
QU_END

#define __qu_big_precision_pop(variable) \
QU_BEGIN \
	__QuBigPrecision = (variable); \
	assert (QU_BIG_MINPREC <= __QuBigPrecision && __QuBigPrecision<= QU_BIG_MAX_PREC); \
QU_END

/***/
extern long __QuBigPrecision;


/** qu_big_new
	Create a new Big object from various C types and other objects.

	Note that except for qu_big_new and qu_big_new_string, all always succeed.

	@warn
	You may not pass a C double NaN nor infinity to qu_big_newd and qu_big_newld.
	Remember that NaN is an error in Qu and infinities are represented with
	={QuInfPos} and ={QuInfNeg} which are ={Inf} objects. Not Float nor Big objects.
*/
#define       qu_big_newh(v)       qu_big_newl  ((long) v)
#define       qu_big_newuh(v)      qu_big_newul ((unsigned long) v)
#define       qu_big_newi(v)       qu_big_newl  ((long) v)
#define       qu_big_newui(v)      qu_big_newul ((unsigned long) v)
extern qu_big qu_big_newl          (long v);
extern qu_big qu_big_newul         (unsigned long v);
extern qu_big qu_big_newd          (double v);
#ifdef HAVE_LONG_LONG
extern qu_big qu_big_newll         (long long v);
extern qu_big qu_big_newull        (unsigned long long v);
#endif
#ifdef HAVE_LONG_DOUBLE
extern qu_big qu_big_newld         (long double v);
#endif

#define       qu_big_new_int(x)     qu_big_newl (qu_int_tol (x))
extern qu_big qu_big_new_long       (qu_long x);
extern qu_big qu_big_new_rational   (qu_rational x);
#define       qu_big_new_float(x)   qu_big_newd (qu_float_tod (x))
#define       qu_big_new_big(x)     (x)
#define       qu_big_new_complex(x) qu_complex_as_big (x)
#define       qu_big_new_integer(x) qu_integer_as_big (x)
#define       qu_big_new_decimal(x) qu_decimal_as_big (x)
#define       qu_big_new_real(x)    qu_real_as_big (x)
#define       qu_big_new_number(x)  qu_number_as_big (x)
#define       qu_big_new_boolean(x) (((x) == QuTrue) ? QuBig1 : QuBig0)
extern qu_big qu_big_new_string     (qu_string x);
extern qu_big qu_big_new            (qu_any x);


/** qu_big_newdx
	Create a new Big/Real object from a C double by checking whether or not the
	value is NaN or infinity.

	If the value is NaN then QuEValue is thrown and NULL is returned.

	If the value is infinity then QuEInf is thrown and NULL is returned unless the
	={qu_number_context} allows infinities to be returned, in which case one of
	={QuInfPos} or ={QuInfNeg} is returned based on the sign.
*/
extern qu_real qu_big_newdx  (double v);
#ifdef HAVE_LONG_DOUBLE
extern qu_real qu_big_newldx (long double v);
#endif


/** qu_big_new_static
	Create a new system Big object not marked by GC.

	qu_big_to_static creates a new one from an existing object.
*/
extern qu_big qu_big_new_static (const char *s);
extern qu_big qu_big_to_static  (qu_big x);


/** qu_big_make
	Make a Big value.

	qu_big_make creates a Big by specifying the digits and exponent.
	v{digits} must be MSB (most significant digit) first.
	v{len} is the number of digits.
	If it is zero then QuBig0 is returned.

	qu_big_make1 creates a Big with one digit and a specified exponent.
	This is useful for things such as 1E-40.
	If v{digit} is zero then QuBig0 is returned.
*/
extern qu_big qu_big_make  (qu_big_digit_t *digits, long len, int expo);
extern qu_big qu_big_make1 (qu_big_digit_t  digit, int expo);


/** qu_big_to
	Convert a Big to various C types.

	v{errno} is set to 0 on success, to ERANGE on failure and return a maximum
	value (e.g. LONG_MAX, ULONG_MAX) for a positive number and a minimum value
	(e.g. LONG_MIN, ULONG_MIN) for a negative number.
*/
#define                   qu_big_toh(x)  qu_number__ltoh   (qu_big_tol  (x))
#define                   qu_big_touh(x) qu_number__ultouh (qu_big_toul (x))
#define                   qu_big_toi(x)  qu_number__ltoi   (qu_big_tol  (x))
#define                   qu_big_toui(x) qu_number__ultoui (qu_big_toul (x))
#define                   qu_big_toli(x) qu_number__ltoli  (qu_big_tol  (x))
extern long               qu_big_tol     (qu_big x);
extern unsigned long      qu_big_toul    (qu_big x);
extern float              qu_big_tof     (qu_big x);
extern double             qu_big_tod     (qu_big x);
#ifdef HAVE_LONG_LONG
extern long long          qu_big_toll    (qu_big x);
extern unsigned long long qu_big_toull   (qu_big x);
#endif
#ifdef HAVE_LONG_DOUBLE
extern long double        qu_big_told    (qu_big x);
#endif


/** qu_big_as
	Convert a Big to various objects.

	NULL is returned and an exception thrown on failure.
	Only qu_big_as_int and qu_big_as_float can fail.
	The others always succeed.
*/
extern qu_int      qu_big_as_int        (qu_big x);
extern qu_long     qu_big_as_long       (qu_big x);
extern qu_rational qu_big_as_rational   (qu_big x);
extern qu_float    qu_big_as_float      (qu_big x);
#define            qu_big_as_big(x)     (x)
#define            qu_big_as_complex(x) qu_complex_new1 ((qu_real) (x))
extern qu_integer  qu_big_as_integer    (qu_big x);
#define            qu_big_as_decimal(x) ((qu_decimal) (x))
#define            qu_big_as_real(x)    ((qu_real) (x))
#define            qu_big_as_number(x)  ((qu_number) (x))
extern qu_array    qu_big_as_array      (qu_big x);
#define            qu_big_as_string(x)  qu__big__string (x)
#define            qu_big_as_boolean(x) (qu_big_iszero (x) ? QuFalse : QuTrue)


/** qu_big_sign
	qu_big_sign returns -1, 0, or 1 if the number is negative, zero, or positive
	respectively.
*/
extern int qu_big_sign (qu_big x);


/** qu_big_is
	Identity functions.

	qu_big_is tests whether or not an object is a Big.

	qu_big_isneg returns 1 if the number is negative.
	Returns 0 otherwise.

	qu_big_iszero returns 1 if the number is zero.
	Returns 0 otherwise.

	qu_big_isone returns 1 if the number is -1 or 1.
	Returns 0 otherwise.

	qu_big_is1 returns 1 if the number is 1.
	Returns 0 otherwise.

	qu_big_iswhole returns 1 if there is no fractional part.
	Returns 0 otherwise.

	qu_big_isodd returns 1 if the number is whole and odd.
	Returns 0 otherwise.

	qu_big_isfrac returns 1 if C{0 < |x| < 1}.
	Returns 0 otherwise.
*/
#define qu_big_is(x)      (qu_classof (x) == QuBigClass)
#define qu_big_isneg(x)   qu_flag_test ((x), QU_BIG_FLAG_NEG)
#define qu_big_iszero(x)  ((x)->len == 0)
extern int qu_big_isone   (qu_big x);
extern int qu_big_is1     (qu_big x);
extern int qu_big_iswhole (qu_big x);
extern int qu_big_isfrac  (qu_big x);
extern int qu_big_isodd   (qu_big x);


/** qu_big_unary
	Unary functions.

	qu_big_pos returns the number itself, obviously.

	qu_big_abs returns the absolute value.
	The result is a Big and always succeed.

	qu_big_neg returns a negated value.
	The result is a Big and always succeed.

	qu_big_incr returns the qu_big incremented by 1.
	On success, the result is a Big or ={QuInfPos}.
	QuEBig is thrown if the ={qu_number_context} does not allow infinities.

	qu_big_decr returns the qu_big decremented by 1.
	On success, the result is a Big or ={QuInfPos}.
	QuEBig is thrown if the ={qu_number_context} does not allow infinities.
*/
#define qu_big_pos(x)   (x)
#define qu_big_abs(x)   qu__big_abs (x)
#define qu_big_neg(x)   qu__big__neg (x)
#define qu_big_incr(x)  qu_big_plus_big (x, QuBig1)
#define qu_big_decr(x)  qu_big_plus_big (x, QuBig_1)


/** qu_big_eq
	Whether the value is the same as another.
	All return 0 or 1.
*/
extern int qu_big_eq_int           (qu_big x, qu_int y);
extern int qu_big_eq_long          (qu_big x, qu_long y);
extern int qu_big_eq_rational      (qu_big x, qu_rational y);
extern int qu_big_eq_float         (qu_big x, qu_float y);
extern int qu_big_eq_big           (qu_big x, qu_big y);
#define    qu_big_eq_complex(x, y) qu_complex_eq_big (y, x)
#define    qu_big_eq_inf(x, y)     (0)
extern int qu_big_eq_integer       (qu_big x, qu_integer y);
extern int qu_big_eq_decimal       (qu_big x, qu_decimal y);
extern int qu_big_eq_real          (qu_big x, qu_real y);
extern int qu_big_eq_number        (qu_big x, qu_number y);


/** qu_big_like
	Whether the absolute value is the same as the absolute value of another.
	All return 0 or 1.
*/
#define    qu_big_like_int(x, y)      (qu_big_cmpabs_int (x, y) == 0)
#define    qu_big_like_long(x, y)     (qu_big_cmpabs_long (x, y) == 0)
#define    qu_big_like_rational(x, y) (qu_big_cmpabs_rational (x, y) == 0)
#define    qu_big_like_float(x, y)    (qu_big_cmpabs_float (x, y) == 0)
#define    qu_big_like_big(x, y)      (qu_big_cmpabs_big (x, y) == 0)
#define    qu_big_like_complex(x, y)  qu_complex_like_big (y, x)
#define    qu_big_like_inf(x, y)      (0)
extern int qu_big_like_integer        (qu_big x, qu_integer y);
extern int qu_big_like_decimal        (qu_big x, qu_decimal y);
extern int qu_big_like_real           (qu_big x, qu_real y);
extern int qu_big_like_number         (qu_big x, qu_number y);


/** qu_big_cmp
	Compare with another.
	All return -1, 0, or 1 in the usual manner.
*/
extern int qu_big_cmp_int           (qu_big x, qu_int y);
extern int qu_big_cmp_long          (qu_big x, qu_long y);
extern int qu_big_cmp_rational      (qu_big x, qu_rational y);
extern int qu_big_cmp_float         (qu_big x, qu_float y);
extern int qu_big_cmp_big           (qu_big x, qu_big y);
#define    qu_big_cmp_complex(x, y) (-qu_complex_cmp_big (y, x))
#define    qu_big_cmp_inf(x, y)     (-qu_inf_cmp_big (y, x))
extern int qu_big_cmp_integer       (qu_big x, qu_integer y);
extern int qu_big_cmp_decimal       (qu_big x, qu_decimal y);
extern int qu_big_cmp_real          (qu_big x, qu_real y);
extern int qu_big_cmp_number        (qu_big x, qu_number y);



/** qu_big_cmpabs
	Compare the absolute value with the absolute value of another.
	All return -1, 0, or 1 in the usual manner.
*/
extern int qu_big_cmpabs_int           (qu_big x, qu_int y);
extern int qu_big_cmpabs_long          (qu_big x, qu_long y);
extern int qu_big_cmpabs_rational      (qu_big x, qu_rational y);
extern int qu_big_cmpabs_float         (qu_big x, qu_float y);
extern int qu_big_cmpabs_big           (qu_big x, qu_big y);
#define    qu_big_cmpabs_complex(x, y) (-qu_complex_cmpabs_big (y, x))
#define    qu_big_cmpabs_inf(x, y)     (-1)
extern int qu_big_cmpabs_integer       (qu_big x, qu_integer y);
extern int qu_big_cmpabs_decimal       (qu_big x, qu_decimal y);
extern int qu_big_cmpabs_real          (qu_big x, qu_real y);
extern int qu_big_cmpabs_number        (qu_big x, qu_number y);


/** qu_big_plus
	Add with another.
*/
extern qu_real   qu_big_plus_int           (qu_big x, qu_int y);
extern qu_real   qu_big_plus_long          (qu_big x, qu_long y);
extern qu_real   qu_big_plus_rational      (qu_big x, qu_rational y);
extern qu_real   qu_big_plus_float         (qu_big x, qu_float y);
extern qu_real   qu_big_plus_big           (qu_big x, qu_big y);
#define          qu_big_plus_complex(x, y) qu_complex_plus_big (y, x)
#define          qu_big_plus_inf(x, y)     qu_inf_plus_big (y, x)
extern qu_real   qu_big_plus_integer       (qu_big x, qu_integer y);
extern qu_real   qu_big_plus_decimal       (qu_big x, qu_decimal y);
extern qu_real   qu_big_plus_real          (qu_big x, qu_real y);
extern qu_number qu_big_plus_number        (qu_big x, qu_number y);


/** qu_big_minus
	Subtract with another.
*/
extern qu_real   qu_big_minus_int           (qu_big x, qu_int y);
extern qu_real   qu_big_minus_long          (qu_big x, qu_long y);
extern qu_real   qu_big_minus_rational      (qu_big x, qu_rational y);
extern qu_real   qu_big_minus_float         (qu_big x, qu_float y);
extern qu_real   qu_big_minus_big           (qu_big x, qu_big y);
#define          qu_big_minus_complex(x, y) qu_complex_rminus_big (y, x)
#define          qu_big_minus_inf(x, y)     qu_inf_plus_big (qu_inf_neg (y), x)
extern qu_real   qu_big_minus_integer       (qu_big x, qu_integer y);
extern qu_real   qu_big_minus_decimal       (qu_big x, qu_decimal y);
extern qu_real   qu_big_minus_real          (qu_big x, qu_real y);
extern qu_number qu_big_minus_number        (qu_big x, qu_number y);


/** qu_big_rminus
	Subtract another with this one.
	Provided for efficiency.
*/
extern qu_real qu_big_rminus_int      (qu_big x, qu_int y);
extern qu_real qu_big_rminus_long     (qu_big x, qu_long y);
extern qu_real qu_big_rminus_rational (qu_big x, qu_rational y);
extern qu_real qu_big_rminus_float    (qu_big x, qu_float y);


/** qu_big_mul
	Multiply with another.
*/
extern qu_real   qu_big_mul_int           (qu_big x, qu_int y);
extern qu_real   qu_big_mul_long          (qu_big x, qu_long y);
extern qu_real   qu_big_mul_rational      (qu_big x, qu_rational y);
extern qu_real   qu_big_mul_float         (qu_big x, qu_float y);
extern qu_real   qu_big_mul_big           (qu_big x, qu_big y);
#define          qu_big_mul_complex(x, y) qu_complex_mul_big (y, x)
#define          qu_big_mul_inf(x, y)     qu_inf_mul_big (y, x)
extern qu_real   qu_big_mul_integer       (qu_big x, qu_integer y);
extern qu_real   qu_big_mul_decimal       (qu_big x, qu_decimal y);
extern qu_real   qu_big_mul_real          (qu_big x, qu_real y);
extern qu_number qu_big_mul_number        (qu_big x, qu_number y);


/** qu_big_div
	Divide with another.
*/
extern qu_real   qu_big_div_int           (qu_big x, qu_int y);
extern qu_real   qu_big_div_long          (qu_big x, qu_long y);
extern qu_real   qu_big_div_rational      (qu_big x, qu_rational y);
extern qu_real   qu_big_div_float         (qu_big x, qu_float y);
extern qu_real   qu_big_div_big           (qu_big x, qu_big y);
#define          qu_big_div_complex(x, y) qu_complex_rdiv_big (y, x)
extern qu_big    qu_big_div_inf           (qu_big x, qu_inf y);
extern qu_real   qu_big_div_integer       (qu_big x, qu_integer y);
extern qu_real   qu_big_div_decimal       (qu_big x, qu_decimal y);
extern qu_real   qu_big_div_real          (qu_big x, qu_real y);
extern qu_number qu_big_div_number        (qu_big x, qu_number y);


/** qu_big_rdiv
	Divide another with this one.
	Provided for efficiency.
*/
extern qu_real qu_big_rdiv_int      (qu_big x, qu_int y);
extern qu_real qu_big_rdiv_long     (qu_big x, qu_long y);
extern qu_real qu_big_rdiv_rational (qu_big x, qu_rational y);
extern qu_real qu_big_rdiv_float    (qu_big x, qu_float y);


/** qu_big_idiv
	Integer division with another.
*/
extern qu_integer qu_big_idiv_int           (qu_big x, qu_int y);
extern qu_integer qu_big_idiv_long          (qu_big x, qu_long y);
extern qu_integer qu_big_idiv_rational      (qu_big x, qu_rational y);
extern qu_integer qu_big_idiv_float         (qu_big x, qu_float y);
extern qu_integer qu_big_idiv_big           (qu_big x, qu_big y);
#define           qu_big_idiv_complex(x, y) qu_complex_ridiv_big (y, x)
extern qu_integer qu_big_idiv_inf           (qu_big x, qu_inf y);
extern qu_integer qu_big_idiv_integer       (qu_big x, qu_integer y);
extern qu_integer qu_big_idiv_decimal       (qu_big x, qu_decimal y);
extern qu_integer qu_big_idiv_real          (qu_big x, qu_real y);
extern qu_integer qu_big_idiv_number        (qu_big x, qu_number y);


/** qu_big_ridiv
	Integer division of another with this one.
	Provided for efficiency.
*/
extern qu_integer qu_big_ridiv_int      (qu_big x, qu_int y);
extern qu_integer qu_big_ridiv_long     (qu_big x, qu_long y);
extern qu_integer qu_big_ridiv_rational (qu_big x, qu_rational y);
extern qu_integer qu_big_ridiv_float    (qu_big x, qu_float y);


/** qu_big_rem
	Remainder of dividing with another.
*/
extern qu_real   qu_big_rem_int           (qu_big x, qu_int y);
extern qu_real   qu_big_rem_long          (qu_big x, qu_long y);
extern qu_real   qu_big_rem_rational      (qu_big x, qu_rational y);
extern qu_real   qu_big_rem_float         (qu_big x, qu_float y);
extern qu_real   qu_big_rem_big           (qu_big x, qu_big y);
#define          qu_big_rem_complex(x, y) qu_complex_rrem_big (y, x)
extern qu_big    qu_big_rem_inf           (qu_big x, qu_inf y);
extern qu_real   qu_big_rem_integer       (qu_big x, qu_integer y);
extern qu_real   qu_big_rem_decimal       (qu_big x, qu_decimal y);
extern qu_real   qu_big_rem_real          (qu_big x, qu_real y);
extern qu_number qu_big_rem_number        (qu_big x, qu_number y);


/** qu_big_rrem
	Remainder of dividing another with this one.
	Provided for efficiency.
*/
extern qu_real qu_big_rrem_int      (qu_big x, qu_int y);
extern qu_real qu_big_rrem_long     (qu_big x, qu_long y);
extern qu_real qu_big_rrem_rational (qu_big x, qu_rational y);
extern qu_real qu_big_rrem_float    (qu_big x, qu_float y);


/** qu_big_pow
	Raise to the power of another.
*/
extern qu_real   qu_big_pow_int           (qu_big x, qu_int y);
extern qu_real   qu_big_pow_long          (qu_big x, qu_long y);
extern qu_number qu_big_pow_rational      (qu_big x, qu_rational y);
extern qu_number qu_big_pow_float         (qu_big x, qu_float y);
extern qu_number qu_big_pow_big           (qu_big x, qu_big y);
#define          qu_big_pow_complex(x, y) qu_complex_rpow_big (y, x)
extern qu_real   qu_big_pow_inf           (qu_big x, qu_inf y);
extern qu_real   qu_big_pow_integer       (qu_big x, qu_integer y);
extern qu_number qu_big_pow_decimal       (qu_big x, qu_decimal y);
extern qu_number qu_big_pow_real          (qu_big x, qu_real y);
extern qu_number qu_big_pow_number        (qu_big x, qu_number y);


/** qu_big_rpow
	Raise another to the power of this one.
	Provided for efficiency.
*/
extern qu_number qu_big_rpow_int      (qu_big x, qu_int y);
extern qu_number qu_big_rpow_long     (qu_big x, qu_long y);
extern qu_number qu_big_rpow_rational (qu_big x, qu_rational y);
extern qu_number qu_big_rpow_float    (qu_big x, qu_float y);


/** qu_big_round
	Rounding functions.

	qu_big_ceil rounds the number to the nearest integer towards positive infinity.
	EBig is thrown if the number is too big.

	qu_big_floor rounds the number to the nearest integer towards negative infinity.
	EBig is thrown if the number is too big.

	qu_big_trunc rounds the number towards zero, that is, it removes the
	fraction part. This function always succeed and returns an Int or Long.

	qu_big_ceil_ex, qu_big_floor_ex, and qu_big_trunc_ex do the same but return a
	Big instead of an Integer.

	qu_big_round rounds a Big to the specified number of fraction digits.
*/
#define qu_big_ceil(x)  qu__big_ceil  (x)
#define qu_big_floor(x) qu__big_floor (x)
#define qu_big_trunc(x) qu__big_trunc (x)
extern qu_big qu_big_round (qu_big x, long prec);


/** qu_big_frac
	Return the fractional part.

	The precision used is the precision of v{x} thus there is no precision
	lost here.
*/
#define qu_big_frac(x) qu__big_frac (x)


/** qu_big_format
	Format a Big.

	See ={Format} for details.

	This function is provided for efficiency between core modules.
	It is not intended to be called by users directly.
*/
extern qu_string qu_big_format (qu_big x, qu_format_s *f);


/** qu_big_scan
	Scan an Big number.

	v{s} is the start of the string.

	On success, a Big object is returned and v{s} is updated to the next cursor
	position after the number.

	On failure, NULL is returned and errno is set.

	qu_big_scan_string does the same but throws ESyntax on failure.
*/
extern qu_big qu_big_scan        (char **s, int neg);
extern qu_big qu_big_scan_string (qu_string s);


/** qu_big_integer_factor
	Return the Integer which must be multiplied to the Big to get an Integer.

	@xmp
		qu_big_integer_factor (1.2 ) = 10
		qu_big_integer_factor (1.23) = 100
*/
extern qu_integer qu_big_integer_factor (qu_big x);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize the class */
extern void qu_big__setup ();
extern void qu_big__init  ();

/* for efficiency - must only be called from within this file */
extern qu_big    qu__big__neg    (qu_big self);
extern qu_string qu__big__string (qu_big self);
extern qu_big    qu__big_abs     (qu_big self);
extern qu_real   qu__big_ceil    (qu_big self);
extern qu_real   qu__big_floor   (qu_big self);
extern qu_big    qu__big_frac    (qu_big self);
extern qu_big    qu__big_trunc   (qu_big self);

/* set by the Math module when loaded */
typedef qu_number (*qu_bigpowbig_f) (qu_big self, qu_big x);
extern qu_bigpowbig_f __QuBigPowBigMath;

#endif /*_H_qu_big_*/
