/* Buffer.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_buffer_
#define _H_qu_buffer_ 1

#include "String.h"


/** Buffer
	The final Buffer class.
*/
extern qu_class QuBufferClass;


/** qu_buffer
	qu_buffer represents an instance of a Buffer class.
	It is a pointer to the instance struct qu_buffer_s.

	See ={QU_STRING_HEAD}
*/
struct qu_buffer_s
{
	QU_STRING_HEAD;
	long    cap;    /* current allocated space */
	long    limit;  /* maximum size allowed */
	qu_any  scope;  /* protector scope */
};


/** qu_buffer_is
	Whether or not an object is a qu_buffer.
*/
#define qu_buffer_is(x) (qu_classof (x) == QuBufferClass)


/** qu_stringy
	qu_stringy is used to denote that a function/method can accept a Buffer and String
	argument. It is typed as qu_string for convenience.

	@warn
	Be careful. If you are blocking then you must protect the Buffer using
	={qu_buffer_protect}.

*/
typedef qu_string qu_stringy;
#define qu_stringy_is(x) (qu_string_is (x) || qu_buffer_is (x))


/** QU_BUFFER_FLAG
	Flags used by Buffer objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_BUFFER_FLAG_PROTECT QU_CONTAINER_FLAG_PROTECT
#define QU_BUFFER_FLAG_FROZEN  QU_CONTAINER_FLAG_FROZEN


/** qu_buffer_protect
	Temporarily make the Buffer immutable.
*/
#define QU_BUFFER_PROTECT_BEGIN(self, saved_scope) \
QU_BEGIN \
	saved_scope = (self)->scope; \
	(self)->scope = QuSysModule; \
	qu_flag_set (self, QU_BUFFER_FLAG_PROTECT); \
QU_END

#define QU_BUFFER_PROTECT_END(self, saved_scope) \
QU_BEGIN \
	(self)->scope = saved_scope; \
	if (!saved_scope) \
		qu_flag_unset (self, QU_BUFFER_FLAG_PROTECT); \
QU_END

#define __qu_buffer_protect_begin(self) \
QU_BEGIN \
	qu_buffer tHiS = self; \
	qu_any    sCoPe = tHiS->scope; \
	tHiS->scope = QuSysModule; \
	qu_flag_set (tHiS, QU_BUFFER_FLAG_PROTECT)

#define __qu_buffer_protect_end \
	tHiS->scope = sCoPe; \
	if (!sCoPe) \
		qu_flag_unset (tHiS, QU_BUFFER_FLAG_PROTECT); \
QU_END


/** qu_buffer_frozen
	Whether the Buffer is mutable.
*/
#define qu_buffer_frozen(self) \
	qu_flag_test (self, QU_BUFFER_FLAG_FROZEN | QU_BUFFER_FLAG_PROTECT)


/** QU_BUFFER_ENSURE_NOT_FROZEN
	Throw EModify if the Buffer if frozen.
*/
#define QU_BUFFER_ENSURE_NOT_FROZEN(self) \
	QU_REQUIRE (QuEModify, !qu_buffer_frozen (self))


/** qu_buffer_new
	Various Buffer constructors.

	qu_buffer_alloc create a new empty Buffer with v{cap} capacity.

	qu_buffer_new create a new Buffer from a C string.

	qu_buffer_newz simply calls C{qu_buffer_new (z, strlen (z))}.
*/
extern qu_buffer qu_buffer_alloc      (long cap);
extern qu_buffer qu_buffer_new        (const char *z, long len);
extern qu_buffer qu_buffer_newz       (const char *z);
extern qu_buffer qu_buffer_new_static (long cap);


/** qu_buffer_clear qu_buffer_CLEAR
	Clear content.

	qu_buffer_clear throws if the Buffer is immutable while qu_buffer_CLEAR does
	not care and does not resize the Buffer.
*/
#define qu_buffer_clear(self) qu__buffer__clear(self)
#define qu_buffer_CLEAR(self) QU_BEGIN (self)->len = 0; QU_END


/** qu_buffer_exchange_tmp
	Exchange the field v{str} and adjust other fields.

	Assumes a mutable Buffer.

	@warn
	This is a delicate function provided only for efficiency.
	Use with care.
*/
extern void qu_buffer_exchange_tmp        (qu_buffer self, qu_buffer tmp);
extern void qu_buffer_exchange_tmp_string (qu_buffer self, qu_string tmp);


/** qu_buffer_tmp_to_string
	Magically turn the Buffer into a String, that is, the v{klass} field is set
	to QuStringClass.

	After a call to this function you must no longer use the Buffer methods.
	You must use the String methods instead.

	@warn
	This is a delicate function provided only for efficiency.
	Use with care.
*/
extern qu_string qu_buffer_tmp_to_string (qu_buffer self);


/** qu_buffer_make_space
	Ensure there is enough room for v{more} bytes.

	Return 1 on success, 0 and throw EOverflow otherwise.
*/
extern int qu_buffer_make_space (qu_buffer self, long more);


/** qu_buffer_avail
	The current space available.
*/
#define qu_buffer_avail(self) ((self)->cap - (self)->len)


/** qu_buffer_put
	Put string into the Buffer.

	All functions return 1 on success, throw EMemory and return 0 on failure.

	Assumes a mutable Buffer.
*/
extern int qu_buffer_put  (qu_buffer self, const char *s, long len);
extern int qu_buffer_putz (qu_buffer self, const char *s);
extern int qu_buffer_puts (qu_buffer self, qu_stringy s);
extern int qu_buffer_putc (qu_buffer self, char c);


/** qu_buffer_insert
	Insert string to Buffer.

	Assumes a mutable Buffer.

	@warn
	Make sure C{i <= 0 && i <= self->len}.
	No checks are made.
*/
extern int qu_buffer_insert (qu_buffer self, long i, const char *s, long n);


/** qu_buffer_set_len
	Set the length of the content.

	Use this function if you have previously entered new content directly.
*/
#define qu_buffer_set_len(self, length) \
QU_BEGIN \
	(self)->len = length; \
	(self)->str [length] = '\0'; \
QU_END


/** qu_buffer_del
	Delete v{n} bytes starting from v{i}.

	qu_buffer_consume is C{qu_buffer_del (self, 0, n)}.

	Both return the actual number of bytes deleted.

	Assumes a mutable Buffer.

	@warn
	Make sure C{0 <= i < self->len}.
	No checks are made.
*/
extern long qu_buffer_del     (qu_buffer self, long i, long n);
extern long qu_buffer_consume (qu_buffer self, long n);


/** qu_buffer_chop
	Delete v{n} bytes from the head or tail.

	If v{n} is negative then delete from head.
	Delete from tail otherwise.

	Assumes a mutable Buffer.
*/
extern void qu_buffer_chop (qu_buffer self, long n);


/** qu_buffer_as
	Various representations.
*/
#define qu_buffer_as_string(self) qu_string_new ((self)->str, (self)->len)
#define qu_buffer_as_boolean(self) ((self)->len ? QuTrue : QuFalse)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_buffer__setup ();
extern void qu_buffer__init  ();

/* for efficiency - may only be called from within this file */
extern qu_buffer qu__buffer__clear (qu_buffer self);

#endif /*_H_qu_buffer_*/
