/* Code.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_code_
#define _H_qu_code_ 1


/** Code
	This is the final Code class.

	Only ={Module} and ={Sub} objects have Code objects.
	={Class} objects share it with the module where they are declared.
	This means that the initialization code within the class block is executed
	when the module code is initialized.

	A Code object can not live alone because the objects contained in
	the bytecode array (C{code->bytes}) are stored in C{module->literals}.

	Sb objects will eventually refer to a Module object so C{module->literals}
	is always automatically referred by a Code object.

	@xmp
		Sub -> Module
		|
		Code

		Sub -> ... -> Class -> Module
		|
		Code
*/
extern qu_class QuCodeClass;


/** qu_code
	qu_code represents an instance of a Code class.
	It is a pointer to the instance struct qu_code_s.

	All fields are read-only. They are generated by ={qu_code_generate}.
*/
struct qu_code_s
{
	QU_OBJECT_HEAD;
	unsigned char *bytes;     /* Bytecode */
	qu_symbol      parent;    /* The associated Module or Sub object */
	int            len;       /* Bytecode length */
	int            stacksize; /* Stack size required (including for locals) */
};


/** qu_code_is
	Whether or not an object is a Code object.
*/
#define qu_code_is(o) (qu_classof(o) == QuCodeClass)


/** QU_CODE_FLAG
	Flags used by Code objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_CODE_FLAG_JIT QU_FLAG_0  /* 'bytes' point to JIT code */


/** qu_code_parent
	Get the symbol object (Module or Sub) where this code belong.
*/
#define qu_code_parent(self) ((self)->parent)


/** qu_code_is_jit
	Whether this is JIT-ed code.
*/
#define qu_code_is_jit(self) qu_flag_test (self, QU_CODE_FLAG_JIT)


/** qu_code_generate
	Generate the Code object for a Module and everything inside it.

	This function is called by the compiler.
	You must not call this function directly.
*/
extern int qu_code_generate (qu_module m);


/** qu_code_pcname qu_code_pctype
	Get the name or type of an instruction.

	If it does not exist then ENone is thrown.

	See #{QU_PC_XXX} for the list.
*/
extern qu_string qu_code_pcname (int type);
extern qu_int    qu_code_pctype (int type);


/** qu_code_get_module
	Return the ={Module} that own this Code.

	If this is module level code then C{self->parent} is returned, otherwise
	returns C{qu_sub_get_module (self->parent)}.
*/
extern qu_module qu_code_get_module (qu_code self);


/** qu_code_print_literal
	Print any literal to stdout.

	It is intended for Ast so we get a uniform print out.
*/
extern int qu_code_print_literal (qu_any x);


/***/
/*-----------------------------------------------------------------------
	instruction list
-----------------------------------------------------------------------*/

/** QU_PC
	List of instructions.
	An instruction may have as much as 3 arguments.

	See ={QU_PC_XXX} for details.
*/
#define QU_CODE_IN_HEADER
#define p_(name, arg) QU_PC_##name,
enum
{
#include "Codes.h"
};
#undef p_
#undef QU_CODE_IN_HEADER


/** QU_CODE_BLOCK
	Type of blocks.
*/
#define QU_CODE_BLOCK_NONE     0   /* not a block */
#define QU_CODE_BLOCK_SWITCH   1
#define QU_CODE_BLOCK_DO       2
#define QU_CODE_BLOCK_LOOP     3
#define QU_CODE_BLOCK_WHILE    4
#define QU_CODE_BLOCK_FOR      5
#define QU_CODE_BLOCK_RANGE    6
#define QU_CODE_BLOCK_ATOMIC   7
#define QU_CODE_BLOCK_TRY      8
#define QU_CODE_BLOCK_CATCH    9
#define QU_CODE_BLOCK_ELSETRY  10
#define QU_CODE_BLOCK_FINALLY  11

#define QU_CODE_NUMBLOCKS 12


/***/
/*----------------------------------------------------------------------------
	JIT
----------------------------------------------------------------------------*/

/** qu_code_jit
	These are the interfaces to JIT code.

	A JIT engine must set all of these variables.

	qu_code_jit_generate is called by the compiler to generate JIT code.
	It must return 1 on success, 0 on failure. If it is generating ={qu_sub_is_jit}
	then it must free C{code->bytes} by using ={qu_mem_free} and replace it
	with native code then call ={qu_sub_jit_set} to give the Sub the appropriate flags
	and callers. If it produces ={qu_sub_is_extern}, then it must set ={sub}->value
	to the native code then call ={qu_sub_jit_set_extern}.

	__QuCodeJitPrint is called by the __print method for JIT code.
	It must return 1 on success, 0 on failure.
	If not set then a default string is printed.
*/
typedef int (*qu_codejitgen_f  ) (qu_sub  u);
typedef int (*qu_codejitprint_f) (qu_code c);

extern qu_codejitgen_f   qu_code_jit_generate;
extern qu_codejitprint_f __QuCodeJitPrint;

/* call it */
#define qu_code_jit_call(f) (((qu_subc1_f) (f)->code->bytes) (f))


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize the class */
extern void qu_code__setup ();
extern void qu_code__init  ();

/* for the interpreter - list of binop functions */
extern qu_binop_f QU_CODE_BINOPS [QU_PC_LAST_INSTRUCTION + 1];

/* for JIT */
extern qu_string __QuCodePcName [QU_PC_LAST_INSTRUCTION + 1];
extern int       __QuCodePcType [QU_PC_LAST_INSTRUCTION + 1];

#endif /*_H_qu_code_*/
