/* Container.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_container_
#define _H_qu_container_ 1


/** Container
	This is the parent class of ={Array}, ={Dict}, ={Set}, ={Queue} and ={Record}.
*/
extern qu_class QuContainerClass;


/** qu_container QU_CONTAINER_HEAD
	qu_container represents an instance of a Container class.
	It is a pointer to the instance struct qu_container_s.

	Container instances must use QU_CONTAINER_HEAD instead of ={QU_OBJECT_HEAD}.

	See also ={qu_interp_scope}.

	If defined, QU_CONTAINER_WITH_PROTECT allows for program level protection.
	It takes an extra pointer overhead but it is a nice feature.
*/
#define QU_CONTAINER_WITH_PROTECT /* always define! */

#ifdef QU_CONTAINER_WITH_PROTECT

#define QU_CONTAINER_HEAD \
	QU_OBJECT_HEAD; \
	qu_any scope;  /* protector scope - see qu_interp_scope */ \
	long   len     /* number of items <= QU_GC_MAX_ITEMS */

#else

#define QU_CONTAINER_HEAD \
	QU_OBJECT_HEAD; \
	long   len

#endif

struct qu_container_s
{
	QU_CONTAINER_HEAD;
};


/** qu_container_is
	Whether or not an object is a Container.
*/
#define qu_container_is(o) (qu_object_instof(o) == QuContainerClass)


/** QU_CONTAINER_FLAG
	Flags used by Container objects.

	Do not access directly.
	Use the accessor macros provided instead.
*/
#define QU_CONTAINER_FLAG_PROTECT QU_FLAG_0   /* protected by program level */
#define QU_CONTAINER_FLAG_FROZEN  QU_FLAG_1   /* immutable? */
#define QU_CONTAINER_FLAG_FIXED   QU_FLAG_2   /* resizable? */


/** qu_container_frozen qu_container_fixed
	qu_container_frozen checks whether or not the Container is immutable.

	qu_container_freeze must only be used after you have created a new Container and
	don't want the Container to be modified by others.

	qu_container_unfreeze must only be used on a Container you have created.
	Normally, to modify the Container then freeze it again.

	qu_container_fixed checks whether or not the Container is resizable.
*/
#define qu_container_frozen(self)   qu_flag_test  (self, QU_CONTAINER_FLAG_FROZEN | QU_CONTAINER_FLAG_PROTECT)
#define qu_container_freeze(self)   qu_flag_set   (self, QU_CONTAINER_FLAG_FROZEN)
#define qu_container_unfreeze(self) qu_flag_unset (self, QU_CONTAINER_FLAG_FROZEN)
#define qu_container_fixed(self)    qu_flag_test  (self, QU_CONTAINER_FLAG_FIXED)
#define qu_container_fix(self)      qu_flag_set   (self, QU_CONTAINER_FLAG_FIXED)
#define qu_container_unfix(self)    qu_flag_unset (self, QU_CONTAINER_FLAG_FIXED)


/** __qu_container_protect
	If you are not the creator of the Container then use these macros to
	temporarily make the Container immutable.

	The correct idiom is:
	@code
		...
		...
		__qu_container_protect_begin (self)
			...
			result = ... do not QU_THROW here -----------+
			...                                          |
		__qu_container_protect_end;                      |
		...                     QU_THROW here instead <--+
		...
*/
#ifdef QU_CONTAINER_WITH_PROTECT

#define __qu_container_protect_begin(self) \
QU_BEGIN \
	qu_container tHiS = (qu_container) self; \
	qu_any       pRoT = tHiS->scope; \
	tHiS->scope = QuSysModule; \
	qu_flag_set (tHiS, QU_CONTAINER_FLAG_PROTECT)

#define __qu_container_protect_end \
	tHiS->scope = pRoT; \
	if (!tHiS->scope) \
		qu_flag_unset (tHiS, QU_CONTAINER_FLAG_PROTECT); \
QU_END

#else /*QU_CONTAINER_WITH_PROTECT*/

#define __qu_container_protect_begin(self) \
QU_BEGIN \
	qu_container tHiS = (qu_container) self; \
	assert (!qu_container_frozen (tHiS)); \
	qu_container_freeze (tHiS)

#define __qu_container_protect_end \
	qu_container_unfreeze (tHiS); \
QU_END

#endif /*QU_CONTAINER_WITH_PROTECT*/


/** QU_CONTAINER_INIT QU_CONTAINER_MARK
	Container classes must call QU_CONTAINER_INIT from their allocation routines
	and QU_CONTAINER_MARK from within their f{mark} method.
*/
#ifdef QU_CONTAINER_WITH_PROTECT
#define QU_CONTAINER_INIT(self)  (self)->scope = 0
#define QU_CONTAINER_MARK(self)  qu_gc_xmark ((self)->scope)
#else
#define QU_CONTAINER_INIT(self)  QU_BEGIN QU_END
#define QU_CONTAINER_MARK(self)  QU_BEGIN QU_END
#endif


/** QU_CONTAINER_ENSURE
	Throw EModify if the Container is frozen, fixed or both.
*/
#define QU_CONTAINER_ENSURE_NOT_FROZEN(self) \
	QU_REQUIRE (QuEModify, !qu_container_frozen (self))

#define QU_CONTAINER_ENSURE_NOT_FIXED(self) \
	QU_REQUIRE (QuEModify, !qu_container_fixed (self))

#define QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED(self) \
	QU_REQUIRE (QuEModify, !qu_container_frozen (self) && !qu_container_fixed (self))


/** QU_CONTAINER_QUOTE
	If it is a String then quote it.
	Otherwise make it a String.
*/
#define QU_CONTAINER_QUOTE(x) (qu_string_is (x) ? qu_string_quote ((qu_string) x) : qu_sys_string (x))


/** QU_CONTAINER_PRINT_QUOTED
	If it is a String then print it quoted.
	Otherwise print as is.
*/
#define QU_CONTAINER_PRINT_QUOTED(x) \
	(qu_string_is (x) ? qu_string_print (x, '\'') : qu_sys_print (x))


/** QU_CONTAINER_TRUE
	The default truth value of a container.
*/
#define QU_CONTAINER_TRUE(self) ((self)->len != 0)


/** qu_container_sizer
	Get a nice size.
*/
extern long qu_container_sizer (long cap);

/* not smaller than this */
#define QU_CONTAINER_SIZE_MIN 8


/** qu_container_mask
	Various functions for Dict/Set like objects.

	qu_container_mask_sizer get a nice mask value.

	qu_container_mask_grow returns 0 if growing is not required, the new mask value
	otherwise.

	qu_container_mask_shrink returns 0 if shrinking is not required, the new mask value
	otherwise.
*/
extern long qu_container_mask_sizer (long cap);

/* small mask must be n^2 - 1 */
#define QU_CONTAINER_MASK_MIN 15

/* limit mask must be n^2 - 1 */
#define QU_CONTAINER_MASK_MAX 0X3FFFFFFF

inline static long qu_container_mask_grow (long mask, long len)
{
	if (mask < QU_CONTAINER_MASK_MAX && len > (mask << 1) / 3)
		return ((mask + 1) << 1) - 1;
	return 0;
}

inline static long qu_container_mask_shrink (long mask, long len)
{
	if (mask > QU_CONTAINER_MASK_MIN && len < (mask / 3))
		return ((mask + 1) >> 1) - 1;
	return 0;
}


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_container__setup ();
extern void qu_container__init  ();

/* for efficiency - may only be called by Container objects */
#define qu_container_true(self)       ((self)->len != 0)
#define qu_container_as_boolean(self) ((self)->len ? QuTrue : QuFalse)

#endif /*_H_qu_container_*/
