/* Dict.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_dict_
#define _H_qu_dict_ 1

#include "Container.h"


/** Dict
	The Dict class.

	@warn
	Arg is a final subclass of Dict.
	Do not use it.
	It is an experimental special class used for passing named arguments to
	functions/methods.

	Currently program level functions/methods do not support named arguments
	for efficiency (otherwise every single Sub will need one more Arg field).
	Thus C functions/method should not support it as well to avoid confusion.

	In the future, if we ever decide to support named arguments then this is
	the class to use.

	Note that the compiler/interpreter already supports this feature (the
	QU_PC_ARG instruction). At the moment it is completely useless.
*/
extern qu_class QuDictClass;
extern qu_class QuArgClass;


/** qu_dict
	qu_dict represents an instance of the Dict class.
	It is a pointer to the instance struct qu_dict_s.

	See also ={QU_CONTAINER_HEAD}.
*/
typedef struct qu_dictentry_s qu_dictentry_s;
struct qu_dictentry_s
{
	qu_dictentry_s  *prev;    /* previous item */
	qu_dictentry_s  *next;    /* next item */
	qu_dictentry_s  *right;   /* next item in bucket */
	qu_any           key;     /* the key */
	qu_any           value;   /* the value */
	long             hash;    /* key hash value */
};

struct qu_dict_s
{
	QU_CONTAINER_HEAD;
	long             mask;     /* current mask */
	qu_any           ktype;    /* key type/validator, maybe 0 */
	qu_any           vtype;    /* value type/validator, maybe 0 */
	qu_dictentry_s **buckets;  /* buckets */
	qu_dictentry_s   sentinel; /* sentinel */
};


/** qu_dict_is
	Whether or not an object is a Dict.
*/
#define qu_dict_is(o) qu_object_instof (o, QuDictClass)
#define qu_arg_is(o)  (qu_classof (o) == QuArgClass)


/** QU_DICT_FLAG
	Flags used by Dict objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_DICT_FLAG_PROTECT QU_CONTAINER_FLAG_PROTECT /* program level protected? */
#define QU_DICT_FLAG_FROZEN  QU_CONTAINER_FLAG_FROZEN  /* mutable? implies QU_DICT_FLAG_FIXED */
#define QU_DICT_FLAG_FIXED   QU_CONTAINER_FLAG_FIXED   /* resizable? */
#define QU_DICT_FLAG_ENUM    QU_FLAG_10                /* built from enum statement? */


/** qu_dict_frozen qu_dict_fixed
	qu_dict_frozen checks whether or not the Dict is mutable.

	qu_dict_freeze must only be used after you have created a new Dict and
	don't want it to be modified by others.

	qu_dict_unfreeze must only be used on an Dict you have created.
	Normally, to modify the Dict then freeze it again.

	qu_dict_fixed checks whether or not the Dict is resizable.

	You should have no reason to set/unset the QU_DICT_FLAG_FIXED in C level.
*/
#define qu_dict_frozen(self)    qu_container_frozen   (self)
#define qu_dict_freeze(self)    qu_container_freeze   (self)
#define qu_dict_unfreeze(self)  qu_container_unfreeze (self)
#define qu_dict_fixed(self)     qu_container_fixed    (self)


/** qu_dict_is_enum
	Whether the Dict came from an C{enum} statement.

	Only the compiler may set this flag when building a qu_dict from
	an enum statement. Others may only query this flag.
*/
#define qu_dict_is_enum(self)  qu_flag_test ((self), QU_DICT_FLAG_ENUM)
#define qu_dict_set_enum(self) qu_flag_set  ((self), QU_DICT_FLAG_ENUM)


/** QuDictEMPTY
	This is a predefined system frozen empty Dict object.

	Expression C{const [:]} in Qu programs simply point to this object.

	@warn
	You may not modify this object.
*/
extern qu_dict QuDictEMPTY;


/** qu_dict_new
	Create a new Dict with a specified initial capacity.

	qu_dict_new_root creates a root Dict object.
	It is always marked by the GC and may contain anything.

	qu_dict_new_static creates a system static Dict object.
	It is never marked by the GC and may only contain other static objects
	={qu_gc_isstatic}, such as ={Int}, ={Float}, ={String}, etc.

	Make sure v{len} is smaller than ={QU_GC_MAX_ITEMS}.

	qu_dict_new_argv is called by the interpreter.
	You should have no reason to call this function directly.
*/
extern qu_dict qu_dict_new        (long len);
extern qu_dict qu_dict_new_root   (long len);
extern qu_dict qu_dict_new_static (long len);
extern qu_dict qu_dict_new_argv   (int argc, qu_any *argv);


/** qu_dict_to_static
	Create a Dict which is never garbage collected nor marked from an
	existing one. All items must satisfy ={qu_gc_is_static} otherwise
	EValue is thrown.
*/
extern qu_dict qu_dict_to_static (qu_dict self);


/** qu_dict_copy
	Get a shallow copy of the qu_dict.
*/
#define qu_dict_copy(self) qu__dict__copy (self)


/** qu_dict_clear
	Empty the Dict.

	An exception is thrown if the Dict is immutable.
*/
#define qu_dict_clear(self) qu__dict__clear (self)


/** qu_dict_getitem
	qu_dict_getitem find the associated value of a key without throwing an exception.
	Return NULL if the key is not found.

	qu_dict_setitem set the value of an existing item or create a new entry.
	Return 0 and throw EModify, EArgValue, or EMemory on failure.
	Return 1 on success.

	qu_dict_del remove a key.
	Return 0 and throw EModify on failure.
	Return 1 on success.

	qu_dict_del_any remove any arbitrary key without resizing the Dict.
	If the Dict is empty the nothing happens.
*/
extern qu_any qu_dict_getitem  (qu_dict self, qu_any key);
extern qu_any qu_dict_getitemz (qu_dict self, const char *key, long len);
#define       qu_dict_setitem(self, key, val) qu__dict__setitem (self, key, val)
extern int    qu_dict_del      (qu_dict self, qu_any key);
extern void   qu_dict_del_any  (qu_dict self);


/** qu_dict_like
	Return 1 if all keys in self are found in v{other}, 0 otherwise.
*/
extern int qu_dict_like (qu_dict self, qu_dict other);


/** qu_dict_of
	Set key/value type. To set key then type may be C{[type, nil] or C{type}.
	To set the value, it must be C{[nil, type]}. To set both it must be
	C{[keytype, valuetype]}.
*/
#define qu_dict_of(self, type) qu__dict__of (self, type)


/** qu_dict_walk
	Walk through the items.

	Generally you should not call this function unless you are the creator
	of the Dict.

	Note that the Dict is not frozen while walking thus may be modified at any
	time. The algorithm used, however, makes it safe from crashing.

	v{b} and v{i} are index used to keep a safe iteration.
	You don't need to know what they are except that they must be set
	to zero on the first call.

	@xmp
		long b = 0, i = 0;
		qu_any key, val;

		while ((val = qu_dict_walk (MyFooDict, &key, &b, &i)) != 0)
		{
			... do something with key/val ...
		}
*/
extern qu_any qu_dict_walk (qu_dict self, qu_any *key, long *b, long *i);


/** qu_dict_as
	Various representations.
*/
#define         qu_dict_as_string(self)  qu__dict__string (self)
#define         qu_dict_as_boolean(self) qu_container_as_boolean (self)
extern qu_array qu_dict_as_array         (qu_dict self);
extern qu_set   qu_dict_as_set           (qu_dict self);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_dict__setup ();
extern void qu_dict__init  ();

/* for efficiency - may only be called from within this file */
extern qu_dict   qu__dict_new      (qu_class self, int argc, qu_any *argv);
extern qu_string qu__dict__string  (qu_dict self);
extern qu_dict   qu__dict__copy    (qu_dict self);
extern qu_dict   qu__dict__clear   (qu_dict self);
extern qu_any    qu__dict__setitem (qu_dict self, qu_any key, qu_any type);
extern qu_dict   qu__dict__of      (qu_dict self, qu_any type);

/*
	qu_arg_new is experimental.
	It is used when creating a Dict when calling a function/method by argument
	name. Do not use it.
*/
extern qu_dict qu_arg_new (int argc, qu_any *argv);

#endif /*_H_qu_dict_*/
