/* Exception.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_exception_
#define _H_qu_exception_ 1


/** Exception
	This is the Exception class.

	It is the base class of all exception classes.
*/
extern qu_class QuExceptionClass;


/** qu_exception
	qu_exception represents an instance of the Exception class.
	It is a pointer to the instance struct qu_exception.

	Note that the message is an instance property named v{msg}.
*/
struct qu_exception_s
{
	QU_OBJECT_HEAD;
};


/** qu_exception_class_new
	Create a new exception class.

	Note that you can also use ={qu_class_new} and ={qu_class_set_super} and
	={qu_attr_set} but this is much more convenient.
*/
extern qu_class qu_exception_class_new (qu_module m, qu_class sup, const char *name, const char *doc);


/** qu_exception_errno
	Find the exception class associated with v{errno}.
*/
extern qu_class qu_exception_errno (int eno);


/** qu_exception_class_is
	Whether or not the object is an exception class.
*/
#define qu_exception_class_is(o) (qu_class_is(o) && qu_class_isa ((qu_class)(o), QuExceptionClass))


/** qu_exception_is
	Whether or not the object is an instance of an exception class.
*/
#define qu_exception_is(o) qu_object_instof (o, QuExceptionClass)


/** QU_EXCEPTION_LAST_IS
	Check whether the last exception thrown by this thread is a specific class.

	@xmp
		if (QU_EXCEPTION_LAST_IS (QuEStop))
			...
*/
#define QU_EXCEPTION_LAST_IS(kls) \
	(__QuThreadCurrent->eclass == kls##Class)


/** __QuExceptionExitCode
	The exit code set by the builtin -{__exit__} function is the error code
	number passed to the system f{exit} function.

	Do not access this variable directly.
*/
extern volatile int __QuExceptionExitCode;


/** QU_THROW
	Use these macros to throw exceptions from C code.

	For the method or function name to be appropriately included in the
	message, you must name it using this convention:

	@ul
	. C{Xxx__xxx} : for standard methods.
	. C{Xxx_xxx}  : for non-standard methods or functions.

	@xmp
		static qu_int Foo__cmp (qu_foo self, qu_int a)
		{
			...
		}

		static qu_int Foo_bar (qu_int a, qu_string b)
		{
			...
		}

	@warn
	v{strz} must be a static string, e.g. C{"foo bar"}. It may not be a
	dynamic pointer.

	The arguments to QU_THROWF must be C values, i.e. ala f{fprint}.

	@xmp
		static void Foo_bar (qu_string s)
		{
			...
			QU_THROWF (QuEValue, "%s is fooed", s);        # wrong!
			QU_THROWF (QuEValue, "%s is fooed", s->str);   # correct
			...
		}
*/
#define QU_RETHROW \
QU_BEGIN \
	qu_thread_rethrow (); \
	return 0; \
QU_END

#define QU_THROW(kls) \
QU_BEGIN \
	qu_thread_throw (kls##Class, 0); \
	return 0; \
QU_END

#define QU_THROWV(kls, v) \
QU_BEGIN \
	qu_thread_throw (kls##Class, 0); \
	return v; \
QU_END

#define QU_THROWS(kls, s) \
QU_BEGIN \
	qu_thread_throw (kls##Class, s); \
	return 0; \
QU_END

#define QU_THROWZ(kls, strz) \
QU_BEGIN \
	qu_thread_throw (kls##Class, qu_string_newz (strz)); \
	return 0; \
QU_END

#define QU_THROWF(kls, ...) \
QU_BEGIN \
	qu_thread_throw (kls##Class, qu_string_newf (__VA_ARGS__)); \
	return 0; \
QU_END

#define QU_THROWERRNO \
QU_BEGIN \
	qu_thread_throw (qu_exception_errno (errno), 0); \
	return 0; \
QU_END

#define QU_THROWERRNOF(fmt, ...) \
QU_BEGIN \
	qu_thread_throw (qu_exception_errno (errno), qu_string_newf (fmt, __VA_ARGS__)); \
	return 0; \
QU_END

#define QU_THROWERRNOV(val) \
QU_BEGIN \
	qu_thread_throw (qu_exception_errno (errno), 0); \
	return (val); \
QU_END


/** QU_ENSURE QU_REQUIRE
	Use these macros to throw an exception by testing an expression.

	Use QU_ENSURE to override ={Thread}->ewhere set by a previous function
	call.

	@xmp
		static qu_int _foo_do_something (qu_int x)
		{
			assert (qu_int_is (x));
			...
			QU_THROW (QuEValue);
				--> sets __QuThreadCurrent->ewhere to "_foo_do_something"
			...
			return QuInt0;
		}

		static sub Foo_bar (qu_int x)
		{
			long n;

			QU_ARG_TOLI (x, n);
			QU_ARG_CHECK (n > 0);
			QU_ENSURE (_do_something (x));
				--> if _foo_do_something throws the this will override
				    __QuThreadCurrent->ewhere to "Foo_bar" then return 0
			...
		}
*/
#define QU_ENSURE(expr) \
QU_BEGIN \
	if ((expr) == 0) \
		QU_RETHROW; \
QU_END


#define QU_ENSUREZ(expr, msgz) \
QU_BEGIN \
	if ((expr) == 0) \
	{ \
		qu_thread_set_errmsg (qu_string_newz (msgz)); \
		QU_RETHROW; \
	} \
QU_END


#define QU_ENSUREV(expr, ret) \
QU_BEGIN \
	if ((expr) == 0) \
	{ \
		qu_thread_rethrow (); \
		return ret; \
	} \
QU_END


/* note that this takes care of negative index */
#define QU_ENSURE_INDEX(idx, len) \
QU_BEGIN \
	if ((idx) < 0) \
		idx += len; \
	QU_REQUIRE (QuEIndex, 0 <= (idx) && (idx) < (len)); \
QU_END


#define QU_REQUIRE(kls, expr) \
QU_BEGIN \
	if (!(expr)) \
	{ \
		qu_thread_throw (kls##Class, 0); \
		return 0; \
	} \
QU_END


#define QU_REQUIREZ(kls, expr, strz) \
QU_BEGIN \
	if (!(expr)) \
	{ \
		qu_thread_throw (kls##Class, qu_string_newz (strz)); \
		return 0; \
	} \
QU_END


#define QU_REQUIREF(kls, expr, fmt, ...) \
QU_BEGIN \
	if (!(expr)) \
	{ \
		qu_thread_throw (kls##Class, qu_string_newf (fmt, __VA_ARGS__)); \
		return 0; \
	} \
QU_END


/** qu_exception_defmsg
	Get the default message of an exception class as String.
*/
qu_string qu_exception_defmsg (qu_class self);


/** qu_exception_getmsg
	Get the message associated with an Exception object.
	The returned value may be any object, not just String.
*/
#define qu_exception_getmsg(self) qu__exception_getmsg (self)


/** qu_exception_as
	Various representations.
*/
#define qu_exception_as_string(self)  qu__exception__string (self)
#define qu_exception_as_boolean(self) QuTrue


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize exception classes */
extern void qu_exception__setup ();
extern void qu_exception__init  ();

/* for efficiency - may only be called from within this file */
extern qu_any    qu__exception_defmsg  (qu_class self);
extern qu_any    qu__exception_getmsg  (qu_exception self);
extern qu_string qu__exception__string (qu_exception self);

/* the list */
#define QU_EXCEPTION_IN_HEADER
#include "Exceptions.h"

/* for Sys */
extern qu_exception qu__exception_new (qu_class self, int argc, qu_any *argv);

#endif /*_H_qu_exception_*/
