/* File.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_file_
#define _H_qu_file_ 1


/** File
	This is the final File class.

	Most methods will throw an ESystem exception class according to v{errno}.
	See ={QU_THROW} and ={qu_exception_errno}.
*/
extern qu_class QuFileClass;


/** qu_file
	qu_file represents an instance of the File class.
	It is a pointer to the instance struct qu_file_s.
*/
/* qu_file_s is intentionally hidden */


/** QuFilePathSepStr
	The file path separator as a ={String} object, e.g. C{"/"}.
*/
extern qu_string QuFilePathSepStr;


/** qu_file_is
	Whether or not the object is a File object.
*/
#define qu_file_is(o) (qu_classof(o) == QuFileClass)


/** __qu_file_debug
	If __QuFileDebug is 1, then some File operations such as chmod, copy, rename, dump,
	etc. are not actually performed.

	@xmp
		__qu_file_debug_push;
		{
			... do not throw here ...
		}
		__qu_file_debug_pop;
*/
#define __qu_file_debug_push QU_BEGIN __QuFileDebug++; QU_END
#define __qu_file_debug_pop  QU_BEGIN __QuFileDebug--; QU_END
/***/
extern volatile int __QuFileDebug;


/** qu_file_functions
	Convenience functions needed by some other modules.

	Will probably document these later ...
*/
extern qu_string qu_file_dir       (qu_string path);
extern qu_string qu_file_ext       (qu_string path);
extern int       qu_file_exist     (qu_string path, qu_string type);
extern int       qu_file_is_abs    (qu_string path);
extern int       qu_file_is_dir    (qu_string path);
extern int       qu_file_is_root   (qu_string path);
extern int       qu_file_is_type   (qu_string path, qu_string filtertype);
extern qu_set    qu_file_list      (qu_string path, qu_string filtertype, qu_string filtername, int strip);
extern int       qu_file_info      (qu_string path, long *size, double *mtime);
extern qu_string qu_file_load_ex   (qu_string path, long timeout, long max, long *size, double *mtime);
extern qu_string qu_file_load      (qu_string path);
extern qu_string qu_file_make      (qu_string path, qu_string defdir, qu_string defext);
extern qu_string qu_file_name      (qu_string path);
extern qu_string qu_file_name_ext  (qu_string path);
extern qu_string qu_file_repext    (qu_string path, qu_string ext);
extern qu_string qu_file_strip     (qu_string path);
extern qu_string qu_file_toabs     (qu_string path, qu_string defdir);
extern int       qu_file_valid     (qu_any path, int allowspace);


/** qu_file_match
	Match name with pattern.

	Wildcards recognized are: C{"*"} C{"?"} and C{"."}

	@xmp
		"f?o.*"
*/
extern int qu_file_match (const char *pat, long lp, const char *name, long ln);


/** qu_file_get_permission_mode
	Get permision mode from a string specification.

	On success 1 is returned and v{p} will contain the flags (a combination
	of S_IRUSR, S_IWUSR, S_IXUSR, etc.).

	On failure, 0 is returned.

	The string must be a combination of:
	@args
	.C{u}  user
	.C{g}  group
	.C{o}  others

	@p
	each followed by a combination of:
	@args
	.C{r}  has read permission
	.C{w}  has write permission
	.C{x}  has execute permission

	@xmp
		"urwx grw orw"
*/
extern int qu_file_get_permission_mode (const char *s, int *p);


/** qu_file_get_open_mode
	Get file open mode from a string specification.

	On success 1 is returned and v{mode} and v{perm} will contain the open
	and permission flags.

	On failure, 0 is returned.

	The mode is a combination of:
	@args
		.C{r}  for read           (O_RDONLY)
		.C{w}  for write          (O_WRONLY)
		.C{rw} for read and write (O_RDWR)
		.C{a}  for append only    (O_APPEND)
	@p
	followed by a combination of:
	@args
		.C{t}  truncate after opened    (O_TRUNC)
		.C{c}  create if does not exist (O_CREAT)
		.C{l}  dont follow link         (O_NOFOLLOW)
		.C{n}  non-blocking             (O_NONBLOCK)
		.C{x}  exclusive                (O_EXCL)
		.C{s}  auto synchronize         (O_SYNC)

	@p
	and possibly followed by C{":"} then a permission (see ={qu_file_get_permission_mode}).

	@xmp
		"rwtc:urwxgrworw"
*/
extern int qu_file_get_open_mode (const char *s, int *mode, int *perm);


/** qu_file_get_filter_mode
	Get filter mode from a string specification.

	On success 1 is returned and v{filter} will contain the flags.

	On failure, 0 is returned.

	@args
	.C{s}   socket           (S_ISSOCK)
	.C{l}   symbolic link    (S_IFLNK)
	.C{f}   regular file     (S_IFREG)
	.C{b}   block device     (S_IFBLK)
	.C{d}   directory        (S_IFDIR)
	.C{c}   character device (S_IFCHR)
	.C{i}   fifo             (S_ISFIFO)
	.C{h}   hidden file      (S_IFHIDDEN)
	.C{v}   sticky bit       (S_ISVTX)
	.C{u}   set UID bit      (S_ISUID) or user  permission if followed by C{rwx} combination
	.C{g}   set GID bit      (S_ISGID) or group permission if followed by C{rwx} combination
	.C{o}                    others permission followed by C{rwx} combination
*/
extern int qu_file_get_filter_mode (const char *s, int *filter);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_file__setup ();
extern void qu_file__init  ();

/* for Sub __interface */
extern int qu_file_modified (qu_string path, double lasttime);

#endif /*_H_qu_file_*/
