/* Frame.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_frame_
#define _H_qu_frame_ 1

#include "Compile.h"


/** Frame
	The final Frame class.

	Frames are created to execute program level code, i.e. initializing modules
	and calling functions/methods.
*/
extern qu_class QuFrameClass;


/** qu_frame
	qu_frame represents an instance of the Frame class.
	It is a pointer to the instance struct qu_frame_s.
*/
typedef struct qu_frblk_s qu_frblk_s;
struct qu_frblk_s
{
	qu_any obj;     /* switch value or for-each iterator object or 0 */
	long   ex1;     /* \__ for builtin class iterators */
	long   ex2;     /* /   used internally by interp.c */
	long   cont;    /* where to 'continue' */
	long   nextip;  /* where to jump when leaving the block */

	/*
		Note that obj is not necessarily an Iterator. It may be an Array, Dict,
		Range, or Ret. The interpreter does not create an Iterator object for
		those. This is much more efficient and faster.

		For bytecode, using pointers for the ip and jump targets is not a win.
		I have tried it and lost about 10% speed. And that includes a number
		of delay jump strategies. If you can do better, let me know. -marc-

		JIT uses 'nextip' as the address of block info embedded in the JIT
		code and 'cont' as *jit_insn. That is why we use longs.
	*/
};

/* be careful not to exceed QU_GC_SIZE_LIMIT */
struct qu_frame_s
{
	QU_OBJECT_HEAD;
	qu_code   code;    /* the associated Code object */
	qu_any    self;    /* the 'self' object */
	qu_any    result;  /* result */
	qu_class  eclass;  /* last exception thrown */
	long      ip;      /* current instruction pointer */
	qu_any   *ss;      /* stack base */
	qu_any   *bp;      /* start of stack after locals */
	qu_any   *sp;      /* current stack pointer */
	int       argc;    /* number of arguments passed */
	long      line;    /* current line */
	int       atomic;  /* depth of atomic block */
	int       level;   /* block level */

	qu_frblk_s blocks [QU_COMPILE_MAXLEVEL]; /* for/switch/loop/etc. blocks */
};


/** qu_frame_is
	Whether or not an object is a Frame.
*/
#define qu_frame_is(o) (qu_classof(o) == QuFrameClass)


/** QU_FRAME_CSTACK_MAX
	Non-generator frames generally use the C stack provided that they don't
	require a stack bigger than this.

	Setting this to a big value is not something you want to do.
*/
#define QU_FRAME_CSTACK_MAX 32


/** QU_FRAME_FLAG
	Flags used by Frame objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_FRAME_FLAG_SUB       QU_FLAG_0  /* Sub or Module code? */
#define QU_FRAME_FLAG_GENERATOR QU_FLAG_1  /* is a generator */
#define QU_FRAME_FLAG_RUNNING   QU_FLAG_2  /* currently running? */
#define QU_FRAME_FLAG_DONE      QU_FLAG_3  /* completed? */
#define QU_FRAME_FLAG_THROWN    QU_FLAG_4  /* exception thrown? */
#define QU_FRAME_FLAG_RETURNING QU_FLAG_5  /* return issued? */
#define QU_FRAME_FLAG_JIT       QU_FLAG_6  /* a JIT frame */
#define QU_FRAME_FLAG_SWAPPING  QU_FLAG_7  /* yielded from __swap__ */
#define QU_FRAME_FLAG_CFRAME    QU_FLAG_8  /* frame is on the C stack (not allocated) */


/** qu_frame_get_symbol
	Get the symbol object (Module or Sub) where this frame originated.
*/
#define qu_frame_get_symbol(self) qu_code_parent ((self)->code)


/** qu_frame_is_sub
	Whether this is a Sub or Module code Frame.
*/
#define qu_frame_is_sub(self) qu_flag_test (self, QU_FRAME_FLAG_SUB)


/** qu_frame_is_generator
	Whether this comes from a generator Sub.
*/
#define qu_frame_is_generator(self) qu_flag_test (self, QU_FRAME_FLAG_GENERATOR)


/** qu_frame_running
	Whether the Frame is currently running.

	QU_FRAME_FLAG_RUNNING may only be set/unset by the interpreter.
*/
#define qu_frame_running(self)        qu_flag_test  (self, QU_FRAME_FLAG_RUNNING)
#define qu_frame_set_running(self)    qu_flag_set   (self, QU_FRAME_FLAG_RUNNING)
#define qu_frame_unset_running(self)  qu_flag_unset (self, QU_FRAME_FLAG_RUNNING)


/** qu_frame_done
	Whether the Frame has completed running.

	qu_frame_set_done should only be set by the interpreter.
	Once set, it must never be unset.
*/
#define qu_frame_done(self)     qu_flag_test (self, QU_FRAME_FLAG_DONE)
#define qu_frame_set_done(self) qu_flag_set  (self, QU_FRAME_FLAG_DONE)


/** qu_frame_thrown
	Whether an exception is thrown.

	The QU_FRAME_FLAG_THROWN should only be set by the interpreter.
*/
#define qu_frame_thrown(self)       qu_flag_test  (self, QU_FRAME_FLAG_THROWN)
#define qu_frame_set_thrown(self)   qu_flag_set   (self, QU_FRAME_FLAG_THROWN)
#define qu_frame_unset_thrown(self) qu_flag_unset (self, QU_FRAME_FLAG_THROWN)


/** qu_frame_returning
	Whether k{return} has been issued.

	qu_frame_set_returning should only be used by the interpreter.
	Once set, it must never be unset.
*/
#define qu_frame_returning(self)     qu_flag_test (self, QU_FRAME_FLAG_RETURNING)
#define qu_frame_set_returning(self) qu_flag_set  (self, QU_FRAME_FLAG_RETURNING)


/** qu_frame_is_jit
	Whether this is a JIT frame.

	QU_FRAME_FLAG_JIT may only be set by the interpreter.
	Once set, it must never be unset.
*/
#define qu_frame_is_jit(self)  qu_flag_test (self, QU_FRAME_FLAG_JIT)
#define qu_frame_set_jit(self) qu_flag_set  (self, QU_FRAME_FLAG_JIT)


/** qu_frame_swapping
	Whether __swap__ has been issued.

	QU_FRAME_FLAG_SWAPPING should only be set/unset by the interpreter.
*/
#define qu_frame_swapping(self)       qu_flag_test  (self, QU_FRAME_FLAG_SWAPPING)
#define qu_frame_set_swapping(self)   qu_flag_set   (self, QU_FRAME_FLAG_SWAPPING)
#define qu_frame_unset_swapping(self) qu_flag_unset (self, QU_FRAME_FLAG_SWAPPING)


/** qu_frame_is_cframe
	Whether the frame is on the C stack (not allocated).
*/
#define qu_frame_is_cframe(self) qu_flag_test (self, QU_FRAME_FLAG_CFRAME)


/** qu_frame_get_module
	Get the ={Module} that owns the Frame.
*/
extern qu_module qu_frame_get_module (qu_frame self);


/** qu_frame_next
	Get the next yielded value.
*/
#define qu_frame_next(self) qu__frame__next (self)


/** qu_frame_where
	Get a String that tells you the whereabouts of this Frame.
*/
#define qu_frame_where(self) qu__frame_where (self)


/** qu_frame_as
	Various representations.
*/
extern qu_array qu_frame_as_array (qu_frame self);
extern qu_dict  qu_frame_as_dict  (qu_frame self);
extern qu_set   qu_frame_as_set   (qu_frame self);
#define qu_frame_as_string(self)  qu__frame__string (self)
#define qu_frame_as_boolean(self) (qu_frame_done (self) ? QuTrue : QuFalse)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_frame__setup ();
extern void qu_frame__init  ();

/* for efficiency - may only be called from within this file */
extern qu_string  qu__frame__string (qu_frame self);
extern qu_boolean qu__frame__true   (qu_frame self);
extern qu_any     qu__frame__next   (qu_frame self);
extern qu_string  qu__frame_where   (qu_frame self);

/* create frame to run qu_module code */
#define qu_frame_new_module(mod) qu__frame_new (((qu_module) mod)->code, mod, 0)
extern qu_frame qu__frame_new (qu_code c, qu_any theself, int nlocal);

/* for interp */
extern qu_frame qu_frame_new_c (qu_frame f, qu_any *stack, qu_sub u, qu_any theself, int argc, qu_any *argv);
extern qu_frame qu_frame_new_a (qu_sub u, qu_any theself, int argc, qu_any *argv);
extern qu_frame qu_frame_new_g (qu_sub u, qu_any theself, int argc, qu_any *argv);

/* throw/rethrow */
extern void qu_frame_rethrow  (qu_frame f);
extern void qu_frame_throw    (qu_frame f, qu_class c);
#define qu_frame_clear_err(f) QU_BEGIN (f)->eclass = 0; QU_END

/* dont touch yet */
#define qu_frame_stop(self, argc, argv) qu__frame__stop (self, argc, argv)
#define qu_frame_call(self, argc, argv) qu__frame__call (self, argc, argv)
extern qu_null qu__frame__stop (qu_frame self, int argc, qu_any *argv);
extern qu_any  qu__frame__call (qu_frame self, int argc, qu_any *argv);

#endif /*_H_qu_frame_*/
