/* Int.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_int_
#define _H_qu_int_ 1


/** Int qu_int
	This is the final Int class.

	qu_int represents an instance of the Int class.

	An Int object is a C long minus the first bit disguised as a pointer.
	The range is #{QU_INT_MIN} .. #{QU_INT_MAX}, where QU_INT_MIN is intentionally
	-QU_INT_MAX.

	We dont allocate memory for Int objects. We rely on word aligned memory.
	If an object address has the first bit set then it is an Int, otherwise it
	is another object.
*/
extern qu_class QuIntClass;


/** QU_INT_MIN QU_INT_MAX
	The maximum C long value an Int object can hold.

	Note that the minimum value is intentionally -QU_INT_MAX.
	Note also that the result of this macro is a long.

	@warn
	The __len method must never return an integer bigger than QU_INT_MAX.
	Meaning, in Qu there must be no object which length is bigger then QU_INT_MAX.
*/
#define QU_INT_MAX (LONG_MAX >> 1)
#define QU_INT_MIN (-QU_INT_MAX)


/** QuIntN
	Some predefined Int objects.

	See ={qu_int_new} to create your own.
*/
#define QuInt16  qu_int_new (16)
#define QuInt10  qu_int_new (10)
#define QuInt9   qu_int_new (9)
#define QuInt8   qu_int_new (8)
#define QuInt7   qu_int_new (7)
#define QuInt6   qu_int_new (6)
#define QuInt5   qu_int_new (5)
#define QuInt4   qu_int_new (4)
#define QuInt3   qu_int_new (3)
#define QuInt2   qu_int_new (2)
#define QuInt1   qu_int_new (1)
#define QuInt0   qu_int_new (0)
#define QuInt_1  qu_int_new (-1)
#define QuInt_2  qu_int_new (-2)
#define QuInt_3  qu_int_new (-3)
#define QuInt_4  qu_int_new (-4)
#define QuInt_5  qu_int_new (-5)
#define QuInt_6  qu_int_new (-6)
#define QuInt_7  qu_int_new (-7)
#define QuInt_8  qu_int_new (-8)
#define QuInt_9  qu_int_new (-9)
#define QuInt_10 qu_int_new (-10)
#define QuInt_16 qu_int_new (-16)


/** qu_int_new
	Create a new Int object from various C types or other objects.

	When using qu_int_new, make sure the C long value is in the range
	#{QU_INT_MIN} .. #{QU_INT_MAX} otherwise you loose the value.
	No checks are made!

	qu_int_newli does the exact same thing as qu_int_new. It is defined for the sake
	of consistent naming convention.

	qu_int_newh and qu_int_newuh never fail because the size of a short is always
	smaller than the size of a I{long minus one bit}.

	The others return NULL and throw EValue if the value fall out of the Int range.

	@warn
	qu_int_new is a macro with no side effect by itself.
	You must never change this behavior.
*/
#define        qu_int_new(l)   ((qu_int) (((long) (l) << 1) | 1))

#define       qu_int_newh(v)  qu_int_newl  (v)
#define       qu_int_newuh(v) qu_int_newul (v)
#define       qu_int_newi(v)  qu_int_newl  (v)
#define       qu_int_newui(v) qu_int_newul (v)
#define       qu_int_newl(v)  qu_int_new   (v)
extern qu_int qu_int_newul    (unsigned long v);
#define       qu_int_newli(v) qu_int_new (v)
extern qu_int qu_int_newd     (double v);
#ifdef HAVE_LONG_LONG
extern qu_int qu_int_newll    (long long v);
extern qu_int qu_int_newull   (unsigned long long v);
#endif
#ifdef HAVE_LONG_DOUBLE
extern qu_int qu_int_newld    (long double v);
#endif

#define       qu_int_new_int(x)      (x)
#define       qu_int_new_float(x)    qu_float_as_int (x)
#define       qu_int_new_long(x)     qu_long_as_int (x)
#define       qu_int_new_rational(x) qu_rational_as_int (x)
#define       qu_int_new_big(x)      qu_big_as_int (x)
#define       qu_int_new_complex(x)  qu_complex_as_int (x)
#define       qu_int_new_integer(x)  qu_integer_toInt (x)
#define       qu_int_new_decimal(x)  qu_decimal_as_int (x)
#define       qu_int_new_real(x)     qu_real_as_int (x)
#define       qu_int_new_number(x)   qu_number_as_int (x)
#define       qu_int_new_boolean(x)  (((x) == QuTrue) ? QuInt1 : QuInt0)
extern qu_int qu_int_new_string      (qu_string x);


/** qu_int_to
	Conversion to C types and other objects.

	The functions set v{errno} to zero on success, to EDOM if conversion is not
	possible, to ERANGE if the value does not fit in the type range.

	The I{unsigned} versions fail with ERANGE if the value is negative.

	Like C, if ERANGE is set then a minimum or maximum possible value for the
	corresponding type is returned, e.g. USHORT_MAX, INT_MIN, etc.

	@warn
	qu_int_tol is a function or a macro with no side effect by itself.
	You must never change this behavior.
*/
#ifdef QU_RSHIFT_PRESERVES_SIGNBIT
#  define qu_int_tol(x) (((long) (x)) >> 1)
#else
   extern long qu__int_tol (long x);
#  define qu_int_tol(x) qu__int_tol ((long) (x))
#endif

#define              qu_int_toh(x)   qu_number__ltoh   (qu_int_tol  (x))
#define              qu_int_touh(x)  qu_number__ultouh (qu_int_toul (x))
#define              qu_int_toi(x)   qu_number__ltoi   (qu_int_tol  (x))
#define              qu_int_toui(x)  qu_number__ultouh (qu_int_toul (x))
#define              qu_int_toli(x)  (qu_int_tol (x))
extern unsigned long qu_int_toul     (qu_int x);
#define              qu_int_tof(x)   ((float) qu_int_tol (x))
#define              qu_int_tod(x)   ((double) qu_int_tol (x))
#ifdef HAVE_LONG_LONG
#define              qu_int_toll(x)  ((long long) qu_int_tol (x))
#define              qu_int_toull(x) ((unsigned long long) qu_int_tol (x))
#endif
#ifdef HAVE_LONG_DOUBLE
#define              qu_int_told(x)  ((long double) qu_int_tol (x))
#endif


/** qu_int_as
	Convert an Int to other objects.

	All functions never fail.
*/
#define          qu_int_as_int(x)      (x)
#define          qu_int_as_long(x)     qu_long_new_int (x)
#define          qu_int_as_rational(x) qu_rational_new_int (x)
#define          qu_int_as_float(x)    qu_float_new_int (x)
#define          qu_int_as_big(x)      qu_big_new_int (x)
#define          qu_int_as_complex(x)  qu_complex_new_int (x)
#define          qu_int_as_integer(x)  ((qu_integer) (x))
#define          qu_int_as_real(x)     ((qu_real) (x))
#define          qu_int_as_decimal(x)  ((qu_decimal) qu_int_as_float (x))
#define          qu_int_as_number(x)   ((qu_number) (x))
#define          qu_int_as_string(x)   qu__int__string (x)
#define          qu_int_as_boolean(x)  (qu_int_tol (x) ? QuTrue : QuFalse)
extern qu_range  qu_int_as_range       (qu_int x);


/** qu_int_sign
	Returns -1, 0, 1 if the value is negative, 0, positive respectively.
*/
extern int qu_int_sign (qu_int x);


/** qu_int_is
	Identity functions.

	qu_int_is tests whether or not an object is an Int.

	qu_int_isneg returns 1 if the value is negative.
	Returns 0 otherwise.

	qu_int_iszero returns 1 if the value is 0.
	Returns 0 otherwise.

	qu_int_isone returns 1 if the value is 1 or -1.
	Returns 0 otherwise.

	qu_int_is1 returns 1 if the value is 1.
	Returns 0 otherwise.

	qu_int_iswhole returns 1, obviously.

	qu_inf_isfrac returns 0, obviously.

	qu_int_isodd returns 1 if the number is odd.
	Returns 0 otherwise.
*/
#define    qu_int_is(x)      ((unsigned long) (x) & 1)
#define    qu_int_iszero(x)  ((x) == QuInt0)
extern int qu_int_isone      (qu_int x);
#define    qu_int_is1(x)     ((x) == QuInt1)
#define    qu_int_isneg(x)   (qu_int_tol (x) < 0)
#define    qu_int_iswhole(x) (1)
#define    qu_int_isfrac(x)  (0)
#define    qu_int_isodd(x)   (qu_int_tol (x) & 1)


/** qu_int_unary
	Unary functions.

	qu_int_pos returns the number itself, obviously.

	qu_int_neg returns a negated value.
	The result is an qu_int.

	qu_int_incr returns the qu_int incremented by 1.
	The result is either an qu_int or a qu_long.

	qu_int_decr returns the qu_int decremented by 1.
	The result is either an qu_int or a qu_long.

	qu_int_inot returns a bit-not (inverted) value.
	The result is an qu_int.
*/
#define qu_int_pos(x)  (x)
#define qu_int_neg(x)  qu__int__neg  (x)
#define qu_int_incr(x) qu__int__incr (x)
#define qu_int_decr(x) qu__int__decr (x)
#define qu_int_inot(x) qu__int__inot (x)


/** qu_int_eq
	Whether the value is the same as another.
	All return 0 or 1.
*/
#define    qu_int_eq_int(x, y)      ((x) == (y))
#define    qu_int_eq_long(x, y)     qu_long_eq_int (y, x)
#define    qu_int_eq_rational(x, y) qu_rational_eq_int (y, x)
#define    qu_int_eq_float(x, y)    qu_float_eq_int (y, x)
#define    qu_int_eq_big(x, y)      qu_big_eq_int (y, x)
#define    qu_int_eq_complex(x, y)  qu_complex_eq_int (y, x)
#define    qu_int_eq_inf(x, y)      (0)
extern int qu_int_eq_integer        (qu_int x, qu_integer y);
extern int qu_int_eq_decimal        (qu_int x, qu_decimal y);
extern int qu_int_eq_real           (qu_int x, qu_real y);
extern int qu_int_eq_number         (qu_int x, qu_number y);


/** qu_int_like
	Whether the absolute value is the same as another.
	All return 0 or 1.
*/
#define    qu_int_like_int(x, y)      (qu_int_cmpabs_int (x, y) == 0)
#define    qu_int_like_long(x, y)     qu_long_like_int (y, x)
#define    qu_int_like_rational(x, y) qu_rational_like_int (y, x)
#define    qu_int_like_float(x, y)    qu_float_like_int (y, x)
#define    qu_int_like_big(x, y)      qu_big_like_int (y, x)
#define    qu_int_like_complex(x, y)  qu_complex_like_int (y, x)
#define    qu_int_like_inf(x, y)      (0)
extern int qu_int_like_integer        (qu_int x, qu_integer y);
extern int qu_int_like_decimal        (qu_int x, qu_decimal y);
extern int qu_int_like_real           (qu_int x, qu_real y);
extern int qu_int_like_number         (qu_int x, qu_number y);


/** qu_int_cmp
	Compare with another.
	All return -1, 0, or 1 in the usual manner.
*/
extern int qu_int_cmp_int            (qu_int x, qu_int y);
#define    qu_int_cmp_long(x, y)     (-qu_long_cmp_int (y, x))
#define    qu_int_cmp_rational(x, y) (-qu_rational_cmp_int (y, x))
#define    qu_int_cmp_float(x, y)    (-qu_float_cmp_int (y, x))
#define    qu_int_cmp_big(x, y)      (-qu_big_cmp_int (y, x))
#define    qu_int_cmp_complex(x, y)  (-qu_complex_cmp_int (y, x))
#define    qu_int_cmp_inf(x, y)      (-qu_inf_cmp_int (y, x))
extern int qu_int_cmp_integer        (qu_int x, qu_integer y);
extern int qu_int_cmp_decimal        (qu_int x, qu_decimal y);
extern int qu_int_cmp_real           (qu_int x, qu_real y);
extern int qu_int_cmp_number         (qu_int x, qu_number y);


/** qu_int_cmpabs
	Compare the absolute value with the absolute value of another.
	All return -1, 0, or 1 in the usual manner.
*/
extern int qu_int_cmpabs_int            (qu_int x, qu_int y);
#define    qu_int_cmpabs_long(x, y)     (-qu_long_cmpabs_int (y, x))
#define    qu_int_cmpabs_rational(x, y) (-qu_rational_cmpabs_int (y, x))
#define    qu_int_cmpabs_float(x, y)    (-qu_float_cmpabs_int (y, x))
#define    qu_int_cmpabs_big(x, y)      (-qu_big_cmpabs_int (y, x))
#define    qu_int_cmpabs_complex(x, y)  (-qu_complex_cmpabs_int (y, x))
#define    qu_int_cmpabs_inf(x, y)      (-1)
extern int qu_int_cmpabs_integer        (qu_int x, qu_integer y);
extern int qu_int_cmpabs_decimal        (qu_int x, qu_decimal y);
extern int qu_int_cmpabs_real           (qu_int x, qu_real y);
extern int qu_int_cmpabs_number         (qu_int x, qu_number y);


/** qu_int_plus
	Add with another.
*/
extern qu_integer qu_int_plus_int            (qu_int x, qu_int y);
#define           qu_int_plus_long(x, y)     qu_long_plus_int (y, x)
#define           qu_int_plus_rational(x, y) qu_rational_plus_int (y, x)
#define           qu_int_plus_float(x, y)    qu_float_plus_int (y, x)
#define           qu_int_plus_big(x, y)      qu_big_plus_int (y, x)
#define           qu_int_plus_complex(x, y)  qu_complex_plus_int (y, x)
#define           qu_int_plus_inf(x, y)      qu_inf_plus_int (y, x)
extern qu_real    qu_int_plus_integer        (qu_int x, qu_integer y);
extern qu_real    qu_int_plus_decimal        (qu_int x, qu_decimal y);
extern qu_real    qu_int_plus_real           (qu_int x, qu_real y);
extern qu_number  qu_int_plus_number         (qu_int x, qu_number y);


/** qu_int_minus
	Subtract with another.
*/
extern qu_integer qu_int_minus_int            (qu_int x, qu_int y);
#define           qu_int_minus_long(x, y)     qu_long_rminus_int (y, x)
#define           qu_int_minus_rational(x, y) qu_rational_rminus_int (y, x)
#define           qu_int_minus_float(x, y)    qu_float_rminus_int (y, x)
#define           qu_int_minus_big(x, y)      qu_big_rminus_int (y, x)
#define           qu_int_minus_complex(x, y)  qu_complex_rminus_int (y, x)
#define           qu_int_minus_inf(x, y)      qu_inf_plus_int (qu_inf_neg (y), x)
extern qu_real    qu_int_minus_integer        (qu_int x, qu_integer y);
extern qu_real    qu_int_minus_decimal        (qu_int x, qu_decimal y);
extern qu_real    qu_int_minus_real           (qu_int x, qu_real y);
extern qu_number  qu_int_minus_number         (qu_int x, qu_number y);


/** qu_int_mul
	Multiply with another.
*/
extern qu_integer qu_int_mul_int            (qu_int x, qu_int y);
#define           qu_int_mul_long(x, y)     qu_long_mul_int (y, x)
#define           qu_int_mul_rational(x, y) qu_rational_mul_int (y, x)
#define           qu_int_mul_float(x, y)    qu_float_mul_int (y, x)
#define           qu_int_mul_big(x, y)      qu_big_mul_int (y, x)
#define           qu_int_mul_complex(x, y)  qu_complex_mul_int (y, x)
#define           qu_int_mul_inf(x, y)      qu_inf_mul_int (y, x)
extern qu_real    qu_int_mul_integer        (qu_int x, qu_integer y);
extern qu_real    qu_int_mul_decimal        (qu_int x, qu_decimal y);
extern qu_real    qu_int_mul_real           (qu_int x, qu_real y);
extern qu_number  qu_int_mul_number         (qu_int x, qu_number y);


/** qu_int_div
	Divide with another.
*/
extern qu_real    qu_int_div_int            (qu_int x, qu_int y);
#define           qu_int_div_long(x, y)     qu_long_rdiv_int (y, x)
#define           qu_int_div_rational(x, y) qu_rational_rdiv_int (y, x)
#define           qu_int_div_float(x, y)    qu_float_rdiv_int (y, x)
#define           qu_int_div_big(x, y)      qu_big_rdiv_int (y, x)
#define           qu_int_div_complex(x, y)  qu_complex_rdiv_int (y, x)
extern qu_int     qu_int_div_inf            (qu_int x, qu_inf y);
extern qu_real    qu_int_div_integer        (qu_int x, qu_integer y);
extern qu_real    qu_int_div_decimal        (qu_int x, qu_decimal y);
extern qu_real    qu_int_div_real           (qu_int x, qu_real y);
extern qu_number  qu_int_div_number         (qu_int x, qu_number y);


/** qu_int_idiv
	Integer division with another.
*/
extern qu_int     qu_int_idiv_int            (qu_int x, qu_int y);
#define           qu_int_idiv_long(x, y)     qu_long_ridiv_int (y, x)
#define           qu_int_idiv_rational(x, y) qu_rational_ridiv_int (y, x)
#define           qu_int_idiv_float(x, y)    qu_float_ridiv_int (y, x)
#define           qu_int_idiv_big(x, y)      qu_big_ridiv_int (y, x)
#define           qu_int_idiv_complex(x, y)  qu_complex_ridiv_int (y, x)
extern qu_int     qu_int_idiv_inf            (qu_int x, qu_inf y);
extern qu_integer qu_int_idiv_integer        (qu_int x, qu_integer y);
extern qu_integer qu_int_idiv_decimal        (qu_int x, qu_decimal y);
extern qu_integer qu_int_idiv_real           (qu_int x, qu_real y);
extern qu_integer qu_int_idiv_number         (qu_int x, qu_number y);


/** qu_int_rem
	Remainder of dividing with another.
*/
extern qu_int     qu_int_rem_int            (qu_int x, qu_int y);
#define           qu_int_rem_long(x, y)     qu_long_rrem_int (y, x)
#define           qu_int_rem_rational(x, y) qu_rational_rrem_int (y, x)
#define           qu_int_rem_float(x, y)    qu_float_rrem_int (y, x)
#define           qu_int_rem_big(x, y)      qu_big_rrem_int (y, x)
#define           qu_int_rem_complex(x, y)  qu_complex_rrem_int (y, x)
extern qu_int     qu_int_rem_inf            (qu_int x, qu_inf y);
extern qu_integer qu_int_rem_integer        (qu_int x, qu_integer y);
extern qu_real    qu_int_rem_decimal        (qu_int x, qu_decimal y);
extern qu_real    qu_int_rem_real           (qu_int x, qu_real y);
extern qu_number  qu_int_rem_number         (qu_int x, qu_number y);


/** qu_int_pow
	Raise to the power of another.
*/
extern qu_real   qu_int_pow_int            (qu_int x, qu_int y);
#define          qu_int_pow_long(x, y)     qu_long_rpow_int (y, x)
#define          qu_int_pow_rational(x, y) qu_rational_rpow_int (y, x)
#define          qu_int_pow_float(x, y)    qu_float_rpow_int (y, x)
#define          qu_int_pow_big(x, y)      qu_big_rpow_int (y, x)
#define          qu_int_pow_complex(x, y)  qu_complex_rpow_int (y, x)
extern qu_real   qu_int_pow_inf            (qu_int x, qu_inf y);
extern qu_number qu_int_pow_integer        (qu_int x, qu_integer y);
extern qu_number qu_int_pow_decimal        (qu_int x, qu_decimal y);
extern qu_number qu_int_pow_real           (qu_int x, qu_real y);
extern qu_number qu_int_pow_number         (qu_int x, qu_number y);


/** qu_int_iand
	Bit-and with another.
*/
extern qu_int qu_int_iand_int        (qu_int x, qu_int y);
#define       qu_int_iand_long(x, y) qu_long_iand_int (y, x)


/** qu_int_ior
	Bit-or with another.
*/
extern qu_int qu_int_ior_int        (qu_int x, qu_int y);
#define      qu_int_ior_long(x, y) qu_long_ior_int (y, x)


/** qu_int_ixor
	Bit-xor with another.
*/
extern qu_int qu_int_ixor_int        (qu_int x, qu_int y);
#define       qu_int_ixor_long(x, y) qu_long_ixor_int (y, x)


/** qu_int_lshift
	Left-shift by the amount of another.
*/
extern qu_real qu_int_lshift_int        (qu_int x, qu_int y);
#define        qu_int_lshift_long(x, y) qu_long_rlshift_int (y, x)


/** qu_int_rshift
	Right-shift by the amount of another.
*/
extern qu_int qu_int_rshift_int  (qu_int x, qu_int y);
extern qu_int qu_int_rshift_long (qu_int x, qu_long y);


/** qu_int_format
	Format an Int.

	See ={Format} for details.

	This function is provided for efficiency between core modules.
	It is not intended to be called by users directly.
*/
extern qu_string qu_int_format (qu_int x, qu_format_s *f);


/** qu_int_scan
	Scan an qu_int number.

	v{s} is the start of a nil terminated string.

	On success an Int object is returned and v{s} is updated to the next cursor
	position after the number.

	On failure NULL is returned.
	v{errno} is set to ERANGE if the number is an integer but does not fit in the Int
	range, to EDOM if the number is not an integer number.
*/
extern qu_int qu_int_scan        (char **s, int neg);
extern qu_int qu_int_scan_hex    (char **s, int neg);
extern qu_int qu_int_scan_oct    (char **s, int neg);
extern qu_int qu_int_scan_binary (char **s, int neg);


/** qu_int_round
	Rounding functions.

	These are only required by some macros. You should have no reason to call
	them since the result is always the same as the argument.
*/
#define qu_int_ceil(x)        (x)
#define qu_int_floor(x)       (x)
#define qu_int_trunc(x)       (x)
#define qu_int_round(x, prec) (x)


/** qu_int_frac
	Return the fraction part, which is always Int (0).
*/
#define qu_int_frac(x) QuInt0


/** qu_int_bin qu_int_oct qu_int_hex
	Return binary, octal and hexidecimal String representation of the Int.
*/
#define qu_int_bin(x) qu__int_bin (x)
#define qu_int_oct(x) qu__int_oct (x)
#define qu_int_hex(x) qu__int_hex (x)


/** qu_int_hash
	Return the Int hash value of an Int which happens to be itself.
*/
#define qu_int_hash(x) (x)


/** qu_int_digit_sum
	Return the sum of the digits.
*/
extern qu_int qu_int_digit_sum (qu_int x);


/** QU_INT_CMP
	QU_INT_CMP compare two C values, return Int(1), Int(0) or Int(-1).
	This is useful for __cmp methods.

	QU_INT_CMPVAL does the same for an int value.
*/
#define QU_INT_CMP(u, v) (((u) > (v)) ? QuInt1  : (((u) < (v)) ? QuInt_1 : QuInt0))

inline static qu_int QU_INT_CMPVAL (int x)
{
	return (x > 0) ? QuInt1 : ((x < 0) ? QuInt_1 : QuInt0);
}


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_int__setup ();
extern void qu_int__init  ();

/* for efficiency - may only be called from within this file */
extern qu_string  qu__int__string (qu_int self);
extern qu_int     qu__int__neg    (qu_int self);
extern qu_integer qu__int__incr   (qu_int self);
extern qu_integer qu__int__decr   (qu_int self);
extern qu_int     qu__int__inot   (qu_int self);
extern qu_string  qu__int_bin     (qu_int x);
extern qu_string  qu__int_oct     (qu_int x);
extern qu_string  qu__int_hex     (qu_int x);

#endif /*_H_qu_int_ */
