/* Iterator.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_iterator_
#define _H_qu_iterator_ 1


/** Iterator
	This is the final Iterator class.
*/
extern qu_class QuIteratorClass;


/** qu_iterator
	qu_iterator represents an instance of the Iterator class.
	It is a pointer to the instance struct qu_iterator_s.

	v{ex1} and v{ex2} may be used by C code. They are not internally used
	by the Iterator class.

	@warn
	v{pos} may not be NULL because it is used by scripts to create custom
	iterators. If not used then leave it to QuInt0.
*/
struct qu_iterator_s
{
	QU_OBJECT_HEAD;
	qu_any  obj;  /* the object to iterate through */
	void   *met;  /* Sub/Method or C method/function to call for the next value */
	qu_any  pos;  /* current position, initially QuInt0 and may not be 0 */
	long    ex1;  /* optional value, initially 0 - may be used as pointer */
	long    ex2;  /* -ditto- */

	/*
		'met' is a qu_iterator_f C method/function address for QU_ITERATOR_EXTERN
		a Sub/Method object otherwise (marked by Iterator)
	*/
};


/** qu_iterator_f
	qu_iterator_f is the C function prototype to be called when the flag
	v{QU_ITERATOR_FLAG_EXTERN} is set (thus the field v{met} is a pointer to
	a C function instead of a Sub).
*/
typedef qu_any (*qu_iterator_f) (qu_any self, qu_iterator i);


/** qu_iterator_is
	Whether or not an object is an Iterator.
*/
#define qu_iterator_is(o) (qu_classof(o) == QuIteratorClass)


/** QU_ITERATOR_FLAG
	Flags used by Iterator objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_ITERATOR_FLAG_EXTERN    QU_FLAG_0  /* field "met" is a pointer to a C function? */
#define QU_ITERATOR_FLAG_KEYS      QU_FLAG_1  /* iterating through keys? */
#define QU_ITERATOR_FLAG_VALUES    QU_FLAG_2  /* iterating through values? */
#define QU_ITERATOR_FLAG_MUTABLES  QU_FLAG_3  /* yield frozen or mutable object? */


/** qu_iterator_is_extern
	Whether the field v{met} is a pointer to a C function.

	See ={QU_ITERATOR_NEW} to setup the QU_ITERATOR_FLAG_EXTERN flag.

	The QU_ITERATOR_FLAG_EXTERN flag must only be set by the creator of the Iterator
	object. Once set, it must never be unset.
*/
#define qu_iterator_is_extern(self)  qu_flag_test (self, QU_ITERATOR_FLAG_EXTERN)
#define qu_iterator_set_extern(self) qu_flag_set  (self, QU_ITERATOR_FLAG_EXTERN)


/** qu_iterator_yield_mutables
	Whether the Iterator requires to return mutable objects.

	The QU_ITERATOR_FLAG_MUTABLES flag must only be set by the creator of the Iterator
	object. Once set, it must never be unset.
*/
#define qu_iterator_yield_mutables(self)     qu_flag_test (self, QU_ITERATOR_FLAG_MUTABLES)
#define qu_iterator_set_yield_mutables(self) qu_flag_set  (self, QU_ITERATOR_FLAG_MUTABLES)


/** qu_iterator_yield_keys
	Whether the Iterator object iterates through keys.

	The QU_ITERATOR_FLAG_KEYS flag must only be set by the creator of the Iterator object.
	Once set, it must never be unset.
*/
#define qu_iterator_yield_keys(self)     qu_flag_test (self, QU_ITERATOR_FLAG_KEYS)
#define qu_iterator_set_yield_keys(self) qu_flag_set  (self, QU_ITERATOR_FLAG_KEYS)


/** qu_iterator_yield_values
	Whether the Iterator object iterates through keys.

	The QU_ITERATOR_FLAG_VALUES flag must only be set by the creator of the Iterator
	object. Once set, it must never be unset.
*/
#define qu_iterator_yield_values(self)    qu_flag_test (self, QU_ITERATOR_FLAG_VALUES)
#define qu_iterator_set_yield_values(self) qu_flag_set  (self, QU_ITERATOR_FLAG_VALUES)


/** QU_ITERATOR_SETUP QU_ITERATOR_NEW
	Initialize or create a new Iterator object.

	v{i} is a qu_iterator type variable, e.g. C{Iterator i;}

	v{o} is the object to iterate through.

	v{u} is the qu_sub to call for QU_ITERATOR_NEW and a ={qu_iterator_f} C function
	for QU_ITERATOR_NEW_EXTERN.

	See Array for an example.

	@warn
	i->pos may not be NULL. Set to QuInt0 if not used.
*/
#define QU_ITERATOR_SETUP(i, o, u) \
QU_BEGIN \
	i->obj = (qu_obj) o; \
	i->pos = QuInt0; \
	i->met = u; \
	i->ex1 = i->ex2 = 0; \
QU_END

#define QU_ITERATOR_SETUP_EXTERN(i, o, u) \
QU_BEGIN \
	QU_ITERATOR_SETUP (i, o, u); \
	qu_iterator_set_extern (i); \
QU_END

#define QU_ITERATOR_NEW(i, o, u) \
QU_BEGIN \
	i = qu_gc_new (QuIteratorClass); \
	QU_ITERATOR_SETUP (i, o, u); \
QU_END

#define QU_ITERATOR_NEW_EXTERN(i, o, u) \
QU_BEGIN \
	i = qu_gc_new (QuIteratorClass); \
	QU_ITERATOR_SETUP_EXTERN (i, o, u); \
QU_END


/** QU_ITERATOR_STOP
	Throw EStop if v{expr} is false.
*/
#define QU_ITERATOR_STOP(expr) QU_REQUIRE (QuEStop, (expr))


/** qu_iterator_do
	Iterate through an object at C level.

	qu_iterator_do takes an object v{x} and a callback function.
	Optionally pass v{data} to the callback.

	qu_iterator_do knows how to iterate through some builtin classes thus is more
	efficient than creating an Iterator object or performed at program level.

	qu_itercallback_f is a callback function that will be called for each item.
	It must return one of:
	@args
	.QU_ITERATOR_CALLBACK_ERROR     an error occured
	.QU_ITERATOR_CALLBACK_END       no more items or QuEStop is thrown
	.QU_ITERATOR_CALLBACK_CONTINUE  continue with the next item
	.QU_ITERATOR_CALLBACK_RESULT    stop with a result

	@p
	v{k} and v{v} are set as follow:
	@rcc
	.| B{object} | B{item}       | B{idx}
	.| String    | character     | index
	.| Array     | item          | index
	.| Dict      | key           | value
	.| Set       | item          | index
	.| Frame     | yielded value | index
	.| Range     | item          | index
	.| Iterator  | next item     | index
	.| Record    | field value   | index
	.| Struct    | field name    | field type
	.| Buffer    | character     | index

	@p
	qu_iterator_do returns one of:
	@args
	.QU_ITERATOR_CALLBACK_ERROR   an error occured
	.QU_ITERATOR_CALLBACK_END     no more items or EStop is thrown
	.QU_ITERATOR_CALLBACK_RESULT  stop with a result
*/
typedef int (*qu_itercallback_f) (qu_any item, qu_any idx, void *data);
extern int qu_iterator_do (qu_any x, qu_itercallback_f f, void *data);

#define QU_ITERATOR_CALLBACK_ERROR    0
#define QU_ITERATOR_CALLBACK_END      1
#define QU_ITERATOR_CALLBACK_CONTINUE 2
#define QU_ITERATOR_CALLBACK_RESULT   3


/** qu_iterator_from
	Make sure we ge an Iterator, not an Array, Dict, etc.

	The interpreter knows how to iterate through some builtin classes
	and the program level __each method is allowed to return those classes
	(see qu_sys_each).

	If v{x} does not support __each then EImplement is thrown.
*/
extern qu_iterator qu_iterator_from (qu_any x);


/** qu_iterator_next
	Get the next value.

	Throws EStop if there is no more value.
*/
#define qu_iterator_next(self) qu__iterator__next (self)


/** qu_iterator_as
	Various representations.
*/
#define qu_iterator_as_string(self)  qu_object_default_string (self)
#define qu_iterator_as_boolean(self) QuTrue


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_iterator__setup ();
extern void qu_iterator__init  ();

/* for efficiency - may only be called from within this file */
extern qu_any qu__iterator__next (qu_iterator self);

#endif /*_H_qu_iterator_*/
