/* Long.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_long_
#define _H_qu_long_ 1


/** Long
	This is the final Long class.
*/
extern qu_class QuLongClass;


/** qu_long_digit_t qu_long_digit2_t QU_LONG_SDIGIT2 QU_LONG_SIZE QU_LONG_FORMAT
	Constants used by the Long class.

	qu_long_digit_t is the C type used for the digits.

	qu_long_digit2_t is twice qu_long_digit_t unsigned while QU_LONG_SDIGIT2 is signed.

	QU_LONG_SIZE is the size of the digit.

	QU_LONG_FORMAT is the format of a digit. This must match QU_LONG_SIZE.

	QU_LONG_MAXLEN is the practical maximum allocatable number of digits.
	It must be such that it is always possible to convert a Long to a ={Big}
	and vice versa. It must also fit in an ={Int}.
	It should be be QU_BIG_MAXLEN / 4.
*/
#ifdef HAVE_LONG_LONG
#  if SIZEOF_LONG_LONG == (SIZEOF_LONG + SIZEOF_LONG)
#    define QU_LONG_FORMAT     2
#    define QU_LONG_SIZE       SIZEOF_LONG
#    define QU_LONG_SDIGIT     long
#    define QU_LONG_SDIGIT2    long long
#    define QU_LONG_2ISLL      1
#  elif SIZEOF_LONG_LONG == (SIZEOF_INT + SIZEOF_INT)
#    define QU_LONG_FORMAT     1
#    define QU_LONG_SIZE       SIZEOF_INT
#    define QU_LONG_SDIGIT     int
#    define QU_LONG_SDIGIT2    long long
#    define QU_LONG_2ISLL      1
#  else
#    define QU_LONG_FORMAT     0
#    define QU_LONG_SIZE       SIZEOF_SHORT
#    define QU_LONG_SDIGIT     short
#    define QU_LONG_SDIGIT2    long
#    undef QU_LONG_2ISLL
#  endif
#else
#    define QU_LONG_FORMAT     0
#    define QU_LONG_SIZE       SIZEOF_SHORT
#    define QU_LONG_SDIGIT     short
#    define QU_LONG_SDIGIT2    long
#    undef QU_LONG_2ISLL
#endif

#define qu_long_digit_t  unsigned QU_LONG_SDIGIT
#define qu_long_digit2_t unsigned QU_LONG_SDIGIT2

#define QU_LONG_MAXLEN (INT_MAX / QU_LONG_SIZE / 16)


/** qu_long
	qu_long represents an instance of the Long class.
	It is a pointer to the instance struct qu_long_s.
*/
struct qu_long_s
{
	QU_OBJECT_HEAD;
	qu_long_digit_t *digits; /* the digits, LSB first */
	long             len;    /* the number of digits */
};


/** QuLongN
	Predefined system objects.

	See ={qu_long_new_static} to create your own.
*/
extern qu_long QuLong0;   /* 0 */
extern qu_long QuLong1;   /* 1 */
extern qu_long QuLong2;   /* 2 */
extern qu_long QuLong_1;  /* -1 */
extern qu_long QuLong_2;  /* -2 */
extern qu_long QuLong10;  /* 10 */


/** QU_LONG_FLAG
	Flags used by Long objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_LONG_FLAG_NEG QU_FLAG_0  /* negative? = QU_BIG_FLAG_NEG */
#define QU_LONG_FLAG_TMP QU_FLAG_9  /* on local C stack? = QU_BIG_FLAG_TMP */


/** qu_long_new
	Create a new Long object from various C types or other objects.

	@warn
	You may not pass a NaN or infinity value to qu_long_newd and qu_long_newld.
*/
#define        qu_long_newh(v)  qu_long_newl  (v)
#define        qu_long_newuh(v) qu_long_newul (v)
#define        qu_long_newi(v)  qu_long_newl  (v)
#define        qu_long_newui(v) qu_long_newul (v)
extern qu_long qu_long_newl     (long v);
extern qu_long qu_long_newul    (unsigned long v);
#define        qu_long_newli(v) qu_long_newl (v)
extern qu_long qu_long_newd     (double v);
#ifdef HAVE_LONG_LONG
extern qu_long qu_long_newll    (long long v);
extern qu_long qu_long_newull   (unsigned long long v);
#endif
#ifdef HAVE_LONG_DOUBLE
extern qu_long qu_long_newld    (long double v);
#endif

#define        qu_long_new_int(x)      qu_long_newl (qu_int_tol (x))
#define        qu_long_new_long(x)     (x)
#define        qu_long_new_float(x)    qu_long_newd ((x)->value)
#define        qu_long_new_rational(x) qu_rational_as_long (x)
#define        qu_long_new_big(x)      qu_big_as_long (x)
#define        qu_long_new_complex(x)  qu_complex_as_long (x)
#define        qu_long_new_integer(x)  qu_integer_as_long (x)
#define        qu_long_new_decimal(x)  qu_decimal_as_long (x)
#define        qu_long_new_real(x)     qu_real_as_long (x)
#define        qu_long_new_number(x)   qu_number_as_long (x)
#define        qu_long_new_boolean(x)  ((x) == QuTrue ? QuLong1 : QuLong0)
extern qu_long qu_long_new_string      (qu_string s);


/** qu_long_new_static
	Create a static Long object never marked by the GC.
*/
extern qu_long qu_long_new_static   (const char *s);
extern qu_long qu_long_new_statici  (int v);
extern qu_long qu_long_new_staticui (unsigned int v);
extern qu_long qu_long_new_staticl  (long v);
extern qu_long qu_long_new_staticul (unsigned long v);
extern qu_long qu_long_to_static    (qu_long x);



/** qu_long_to
	Convert a Long to various C types.

	qu_long_tod returns infinity in case it does not fit in a double so you must
	always check the result before you use it to create ={Float} objects.

	The others C type conversions set v{errno} to 0 on success, to ERANGE on failure
	and return a minimum or maximum value e.g. ULONG_MAX, LONG_LONG_MIN, etc.
*/
#define                   qu_long_toh(x)  qu_number__ltoh   (qu_long_tol (x))
#define                   qu_long_touh(x) qu_number__ultouh (qu_long_toul (x))
#define                   qu_long_toi(x)  qu_number__ltoi   (qu_long_tol (x))
#define                   qu_long_toui(x) qu_number__ultouh (qu_long_toul (x))
#define                   qu_long_toli(x) qu_number__ltoli  (qu_long_tol (x))
extern long               qu_long_tol     (qu_long x);
extern unsigned long      qu_long_toul    (qu_long x);
#ifdef HAVE_LONG_LONG
extern long long          qu_long_toll    (qu_long x);
extern unsigned long long qu_long_toull   (qu_long x);
#endif
extern float              qu_long_tof     (qu_long x);
extern double             qu_long_tod     (qu_long x);
#ifdef HAVE_LONG_DOUBLE
extern long double        qu_long_told    (qu_long x);
#endif


/** qu_long_as
	Convert a Long to other objects.

	An exception is thrown and NULL is returned on failure.
	Only qu_long_as_int and qu_long_as_float can fail. The others always succeed.
*/
extern qu_int     qu_long_as_int         (qu_long x);
extern qu_decimal qu_long_as_decimal     (qu_long x);
#define           qu_long_as_long(x)     (x)
#define           qu_long_as_float(x)    qu_float_new_long (x)
#define           qu_long_as_big(x)      qu_big_new_long (x)
#define           qu_long_as_rational(x) qu_rational_new_integer ((qu_integer) x)
#define           qu_long_as_complex(x)  qu_complex_new1 ((qu_real) x)
#define           qu_long_as_integer(x)  ((qu_integer) (x))
#define           qu_long_as_real(x)     ((qu_real) x)
#define           qu_long_as_number(x)   ((qu_number) x)
#define           qu_long_as_boolean(x)  (qu_long_iszero (x) ? QuFalse : QuTrue)
#define           qu_long_as_string(x)   qu__long__string (x)
extern qu_array   qu_long_as_array       (qu_long x);


/** qu_long_sign
	Returns -1, 0, or 1 if the number is negative, 0, or positive respectively.
*/
extern int qu_long_sign (qu_long x);


/** qu_long_is
	Identity functions.

	qu_long_is tests Whether or not an object is a Long.

	qu_long_isneg return 1 if the number is negative, 0 otherwise.

	qu_long_iszero return 1 if the number is zero, 0 otherwise.

	qu_long_isone return 1 if the number is -1 or 1, 0 otherwise.

	qu_long_is1 return 1 if the number is 1, 0 otherwise.

	qu_long_iswhole always return 1, obviously.

	qu_long_isfrac always return 0, obviously.

	qu_long_isodd return 1 if the number is odd, 0 otherwise.
*/
#define    qu_long_is(x)      (qu_classof(x) == QuLongClass)
#define    qu_long_isneg(x)   qu_flag_test (x, QU_LONG_FLAG_NEG)
#define    qu_long_iszero(x)  ((x)->len == 0)
#define    qu_long_iswhole(x) (1)
#define    qu_long_isfrac(x)  (0)
extern int qu_long_isone      (qu_long x);
extern int qu_long_is1        (qu_long x);
extern int qu_long_isodd      (qu_long x);


/** qu_long_unary
	Unary functions.

	qu_long_pos, obviously, return the Long itself.

	qu_long_neg returns a negated value.
	The result is a qu_long and always succeed.

	qu_long_incr returns the qu_long incremented by 1.
	On success, the result is an qu_int, a qu_long or ={QuInfPos}.
	QuEBig is thrown if the ={qu_number_context} does not allow infinities.

	qu_long_decr returns the qu_long decremented by 1.
	On success, the result is an qu_int, a qu_long or ={QuInfNeg}.
	QuEBig is thrown if the ={qu_number_context} does not allow infinities.

	qu_long_inot returns a bit-not (inverted) value.
	The result is either an qu_int or a qu_long and always succeed.
*/
#define qu_long_pos(x)  (x)
#define qu_long_neg(x)  qu__long__neg  (x)
#define qu_long_incr(x) qu__long__incr (x)
#define qu_long_decr(x) qu__long__decr (x)
#define qu_long_inot(x) qu__long__inot (x)


/** qu_long_eq
	Whether the value is the same as another.
	All return 0 or 1.
*/
extern int qu_long_eq_int            (qu_long x, qu_int y);
extern int qu_long_eq_long           (qu_long x, qu_long y);
#define    qu_long_eq_rational(x, y) qu_rational_eq_long (y, x)
#define    qu_long_eq_float(x, y)    qu_float_eq_long (y, x)
#define    qu_long_eq_big(x, y)      qu_big_eq_long (y, x)
#define    qu_long_eq_complex(x, y)  qu_complex_eq_long (y, x)
#define    qu_long_eq_inf(x, y)      (0)
extern int qu_long_eq_integer        (qu_long x, qu_integer y);
extern int qu_long_eq_decimal        (qu_long x, qu_decimal y);
extern int qu_long_eq_real           (qu_long x, qu_real y);
extern int qu_long_eq_number         (qu_long x, qu_number y);


/** qu_long_like
	Whether the absolute value is the same as the absolute value of another.
	All return 0 or 1.
*/
#define    qu_long_like_int(x, y)      (qu_long_cmpabs_int (x, y) == 0)
#define    qu_long_like_long(x, y)     (qu_long_cmpabs_long (x, y) == 0)
#define    qu_long_like_rational(x, y) qu_rational_like_long (y, x)
#define    qu_long_like_float(x, y)    qu_float_like_long (y, x)
#define    qu_long_like_big(x, y)      qu_big_like_long (y, x)
#define    qu_long_like_complex(x, y)  qu_complex_like_long (y, x)
#define    qu_long_like_inf(x, y)      (0)
extern int qu_long_like_integer        (qu_long x, qu_integer y);
extern int qu_long_like_decimal        (qu_long x, qu_decimal y);
extern int qu_long_like_real           (qu_long x, qu_real y);
extern int qu_long_like_number         (qu_long x, qu_number y);


/** qu_long_cmp
	Compare with another.
	All return -1, 0 or 1 in the usual manner.
*/
extern int qu_long_cmp_int            (qu_long x, qu_int y);
extern int qu_long_cmp_long           (qu_long x, qu_long y);
#define    qu_long_cmp_rational(x, y) (-qu_rational_cmp_long (y, x))
#define    qu_long_cmp_float(x, y)    (-qu_float_cmp_long (y, x))
#define    qu_long_cmp_big(x, y)      (-qu_big_cmp_long (y, x))
#define    qu_long_cmp_complex(x, y)  (-qu_complex_cmp_long (y, x))
#define    qu_long_cmp_inf(x, y)      (-qu_inf_cmp_long (y, x))
extern int qu_long_cmp_integer        (qu_long x, qu_integer y);
extern int qu_long_cmp_decimal        (qu_long x, qu_decimal y);
extern int qu_long_cmp_real           (qu_long x, qu_real y);
extern int qu_long_cmp_number         (qu_long x, qu_number y);


/** qu_long_cmpabs
	Compare the absolute value with the absolute value of another.
	All return -1, 0 or 1 in the usual manner.
*/
extern int qu_long_cmpabs_int            (qu_long x, qu_int y);
extern int qu_long_cmpabs_long           (qu_long x, qu_long y);
#define    qu_long_cmpabs_rational(x, y) (-qu_rational_cmpabs_long (y, x))
#define    qu_long_cmpabs_float(x, y)    (-qu_float_cmpabs_long (y, x))
#define    qu_long_cmpabs_big(x, y)      (-qu_big_cmpabs_long (y, x))
#define    qu_long_cmpabs_complex(x, y)  (-qu_complex_cmpabs_long (y, x))
#define    qu_long_cmpabs_inf(x, y)      (-1)
extern int qu_long_cmpabs_integer        (qu_long x, qu_integer y);
extern int qu_long_cmpabs_decimal        (qu_long x, qu_decimal y);
extern int qu_long_cmpabs_real           (qu_long x, qu_real y);
extern int qu_long_cmpabs_number         (qu_long x, qu_number y);


/** qu_long_plus
	Add with another.
*/
extern qu_real   qu_long_plus_int            (qu_long x, qu_int y);
extern qu_real   qu_long_plus_long           (qu_long x, qu_long y);
#define          qu_long_plus_rational(x, y) qu_rational_plus_long (y, x)
#define          qu_long_plus_float(x, y)    qu_float_plus_long (y, x)
#define          qu_long_plus_big(x, y)      qu_big_plus_long (y, x)
#define          qu_long_plus_complex(x, y)  qu_complex_plus_long (y, x)
#define          qu_long_plus_inf(x, y)      qu_inf_plus_long (y, x)
extern qu_real   qu_long_plus_integer        (qu_long x, qu_integer y);
extern qu_real   qu_long_plus_decimal        (qu_long x, qu_decimal y);
extern qu_real   qu_long_plus_real           (qu_long x, qu_real y);
extern qu_number qu_long_plus_number         (qu_long x, qu_number y);


/** qu_long_minus
	Subtract with another.
*/
extern qu_real   qu_long_minus_int            (qu_long x, qu_int y);
extern qu_real   qu_long_minus_long           (qu_long x, qu_long y);
#define          qu_long_minus_rational(x, y) qu_rational_rminus_long (y, x)
#define          qu_long_minus_float(x, y)    qu_float_rminus_long (y, x)
#define          qu_long_minus_big(x, y)      qu_big_rminus_long (y, x)
#define          qu_long_minus_complex(x, y)  qu_complex_rminus_long (y, x)
#define          qu_long_minus_inf(x, y)      qu_inf_plus_long (qu_inf_neg (y), x)
extern qu_real   qu_long_minus_integer        (qu_long x, qu_integer y);
extern qu_real   qu_long_minus_decimal        (qu_long x, qu_decimal y);
extern qu_real   qu_long_minus_real           (qu_long x, qu_real y);
extern qu_number qu_long_minus_number         (qu_long x, qu_number y);


/** qu_long_rminus
	Subtract another with this one.
	Provided for efficiency.
*/
extern qu_real qu_long_rminus_int (qu_long x, qu_int y);


/** qu_long_mul
	Multiply with another.
*/
extern qu_real   qu_long_mul_int            (qu_long x, qu_int y);
extern qu_real   qu_long_mul_long           (qu_long x, qu_long y);
#define          qu_long_mul_rational(x, y) qu_rational_mul_long (y, x)
#define          qu_long_mul_float(x, y)    qu_float_mul_long (y, x)
#define          qu_long_mul_big(x, y)      qu_big_mul_long (y, x)
#define          qu_long_mul_complex(x, y)  qu_complex_mul_long (y, x)
#define          qu_long_mul_inf(x, y)      qu_inf_mul_long (y, x)
extern qu_real   qu_long_mul_integer        (qu_long x, qu_integer y);
extern qu_real   qu_long_mul_decimal        (qu_long x, qu_decimal y);
extern qu_real   qu_long_mul_real           (qu_long x, qu_real y);
extern qu_number qu_long_mul_number         (qu_long x, qu_number y);


/** qu_long_div
	Divide with another.
*/
extern qu_real   qu_long_div_int            (qu_long x, qu_int y);
extern qu_real   qu_long_div_long           (qu_long x, qu_long y);
#define          qu_long_div_rational(x, y) qu_rational_rdiv_long (y, x)
#define          qu_long_div_float(x, y)    qu_float_rdiv_long (y, x)
#define          qu_long_div_big(x, y)      qu_big_rdiv_long (y, x)
#define          qu_long_div_complex(x, y)  qu_complex_rdiv_long (y, x)
extern qu_int    qu_long_div_inf            (qu_long x, qu_inf y);
extern qu_real   qu_long_div_integer        (qu_long x, qu_integer y);
extern qu_real   qu_long_div_decimal        (qu_long x, qu_decimal y);
extern qu_real   qu_long_div_real           (qu_long x, qu_real y);
extern qu_number qu_long_div_number         (qu_long x, qu_number y);


/** qu_long_rdiv
	Divide another with this one.
	Provided for efficiency.
*/
extern qu_real qu_long_rdiv_int (qu_long x, qu_int y);


/** qu_long_idiv
	Integer divison with another.
*/
extern qu_integer qu_long_idiv_int            (qu_long x, qu_int y);
extern qu_integer qu_long_idiv_long           (qu_long x, qu_long y);
#define           qu_long_idiv_rational(x, y) qu_rational_ridiv_long (y, x)
#define           qu_long_idiv_float(x, y)    qu_float_ridiv_long (y, x)
#define           qu_long_idiv_big(x, y)      qu_big_ridiv_long (y, x)
#define           qu_long_idiv_complex(x, y)  qu_complex_ridiv_long (y, x)
extern qu_int     qu_long_idiv_inf            (qu_long x, qu_inf y);
extern qu_integer qu_long_idiv_integer        (qu_long x, qu_integer y);
extern qu_integer qu_long_idiv_decimal        (qu_long x, qu_decimal y);
extern qu_integer qu_long_idiv_real           (qu_long x, qu_real y);
extern qu_integer qu_long_idiv_number         (qu_long x, qu_number y);


/** qu_long_ridiv
	Integer divison of another with this one.
	Provided for efficiency.
*/
extern qu_integer qu_long_ridiv_int (qu_long x, qu_int y);


/** qu_long_rem
	Remainder of dividing with another.
*/
extern qu_real    qu_long_rem_int            (qu_long x, qu_int y);
extern qu_real    qu_long_rem_long           (qu_long x, qu_long y);
#define           qu_long_rem_rational(x, y) qu_rational_rrem_long (y, x)
#define           qu_long_rem_float(x, y)    qu_float_rrem_long (y, x)
#define           qu_long_rem_big(x, y)      qu_big_rrem_long (y, x)
#define           qu_long_rem_complex(x, y)  qu_complex_rrem_long (y, x)
extern qu_integer qu_long_rem_inf            (qu_long x, qu_inf y);
extern qu_real    qu_long_rem_integer        (qu_long x, qu_integer y);
extern qu_real    qu_long_rem_decimal        (qu_long x, qu_decimal y);
extern qu_real    qu_long_rem_real           (qu_long x, qu_real y);
extern qu_number  qu_long_rem_number         (qu_long x, qu_number y);


/** qu_long_rrem
	Remainder of dividing another with this one.
	Provided for efficiency.
*/
extern qu_real qu_long_rrem_int (qu_long x, qu_int y);


/** qu_long_pow
	Raise to the power of another.
*/
extern qu_real   qu_long_pow_int            (qu_long x, qu_int y);
extern qu_real   qu_long_pow_long           (qu_long x, qu_long y);
#define          qu_long_pow_rational(x, y) qu_rational_rpow_long (y, x)
#define          qu_long_pow_float(x, y)    qu_float_rpow_long (y, x)
#define          qu_long_pow_big(x, y)      qu_big_rpow_long (y, x)
#define          qu_long_pow_complex(x, y)  qu_complex_rpow_long (y, x)
extern qu_real   qu_long_pow_inf            (qu_long x, qu_inf y);
extern qu_real   qu_long_pow_integer        (qu_long x, qu_integer y);
extern qu_real   qu_long_pow_decimal        (qu_long x, qu_decimal y);
extern qu_real   qu_long_pow_real           (qu_long x, qu_real y);
extern qu_number qu_long_pow_number         (qu_long x, qu_number y);


/** qu_long_rpow
	Raise another to the power of this one.
	Provided for efficiency.
*/
extern qu_real qu_long_rpow_int (qu_long x, qu_int y);


/** qu_long_pow_10
	Return a Long which is the result of raising 10 to the specified power.

	@warn
	v{p} must be a non-negative number.
	If the value is huge then NULL might be returned so be careful.
*/
extern qu_long qu_long_pow_10 (int p);


/** qu_long_iand
	Bit-and with another.
*/
extern qu_int     qu_long_iand_int     (qu_long x, qu_int y);
extern qu_integer qu_long_iand_long    (qu_long x, qu_long y);
extern qu_integer qu_long_iand_integer (qu_long x, qu_integer y);


/** qu_long_riand
	Bit-and another with this one.
	Provided for efficiency.
*/
extern qu_integer qu_long_riand_int (qu_long x, qu_int y);


/** qu_long_ior
	Bit-or with another.
*/
extern qu_integer qu_long_ior_int     (qu_long x, qu_int y);
extern qu_integer qu_long_ior_long    (qu_long x, qu_long y);
extern qu_integer qu_long_ior_integer (qu_long x, qu_integer y);


/** qu_long_rior
	Bit-or another with this one.
	Provided for efficiency.
*/
extern qu_integer qu_long_rior_int (qu_long x, qu_int y);


/** qu_long_ixor
	Bit-xor with another.
*/
extern qu_integer qu_long_ixor_int     (qu_long x, qu_int y);
extern qu_integer qu_long_ixor_long    (qu_long x, qu_long y);
extern qu_integer qu_long_ixor_integer (qu_long x, qu_integer y);


/** qu_long_rixor
	Bit-xor another with this one.
	Provided for efficiency.
*/
extern qu_integer qu_long_rixor_int (qu_long x, qu_int y);


/** qu_long_rshift
	Right shift with another.
*/
extern qu_integer qu_long_rshift_int     (qu_long x, qu_int y);
extern qu_integer qu_long_rshift_long    (qu_long x, qu_long y);
extern qu_integer qu_long_rshift_integer (qu_long x, qu_integer y);


/** qu_long_rrshift
	Right shift another with this one.
*/
/* no such thing because we don't need it */


/** qu_long_lshift
	Left shift with another.
*/
extern qu_real qu_long_lshift_int     (qu_long x, qu_int y);
extern qu_real qu_long_lshift_long    (qu_long x, qu_long y);
extern qu_real qu_long_lshift_integer (qu_long x, qu_integer y);


/** qu_long_rlshift
	Left shift another with this one.
	Provided for efficiency.
*/
extern qu_real qu_long_rlshift_int (qu_long x, qu_int y);


/** qu_long_round
	Rounding functions.

	These are only required by some macros. You should have no reason to call
	them since the result is always the same as the argument.
*/
#define qu_long_ceil(x)        (x)
#define qu_long_floor(x)       (x)
#define qu_long_trunc(x)       (x)
#define qu_long_round(x, prec) (x)


/** qu_long_frac
	Return the fraction part, which is always Long (0).
*/
#define qu_long_frac(x) QuLong0


/** qu_long_format
	Format a Long.

	See ={Format} for details.

	This function is provided for efficiency between core modules.
	It is not intended to be called by users directly.
*/
extern qu_string qu_long_format (qu_long self, qu_format_s *f);


/** qu_long_scan
	Scan an Long number.

	v{s} is the start of a nil terminated string.

	On success, a Long or ={Int} object is returned and v{s} is updated to the next
	cursor position after the number.

	On failure NULL is returned.

	qu_long_scan_string scans a base-10 integer and throws ESyntax on failure.
*/
extern qu_long qu_long_scan        (char **s, int neg);
extern qu_long qu_long_scan_hex    (char **s, int neg);
extern qu_long qu_long_scan_oct    (char **s, int neg);
extern qu_long qu_long_scan_binary (char **s, int neg);
extern qu_long qu_long_scan_string (qu_string s);


/** qu_long_bin qu_long_oct qu_long_hex
	Return binary, octal and hexidecimal String representation of the Long.
*/
#define qu_long_bin(x) qu__long_bin (x)
#define qu_long_oct(x) qu__long_oct (x)
#define qu_long_hex(x) qu__long_hex (x)


/** qu_long_hash
	Return the Int hash value of a Long.
*/
#define qu_long_hash(x) qu__long__hash(x)


/** qu_long_digit_sum
	Return the sum of the digits.

	The result is a Qu Integer object, not a C type.
*/
#define qu_long_digit_sum(x) qu__long_digitsum (qu_long x);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_long__setup ();
extern void qu_long__init  ();

/* for efficiency - may only be called from within this file */
extern qu_long    qu__long__neg     (qu_long self);
extern qu_real    qu__long__incr    (qu_long self);
extern qu_real    qu__long__decr    (qu_long self);
extern qu_integer qu__long__inot    (qu_long self);
extern qu_int     qu__long__hash    (qu_long self);
extern qu_string  qu__long__string  (qu_long self);
extern qu_string  qu__long_bin      (qu_long self);
extern qu_string  qu__long_oct      (qu_long self);
extern qu_string  qu__long_hex      (qu_long self);
extern qu_integer qu__long_digitsum (qu_long self);

#endif /*_H_qu_long_*/
