/* Module.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_module_
#define _H_qu_module_ 1


/** Module
	This is the final Module class.

	The main program, modules, and k{eval} are all Module objects.
*/
extern qu_class QuModuleClass;


/** qu_module
	qu_module represents an instance of the Module class.
	It is a pointer to the instance struct qu_module_s.

	All fields are read-only.
*/
typedef qu_any (*qu_module_getcfunc_f) (qu_module self, qu_string symbol, qu_string spec);
typedef qu_any (*qu_module_callfunc_f) (qu_module self, qu_string symbol, qu_string spec);

struct qu_module_s
{
	QU_SYMBOL_HEAD;
	qu_set      literals; /* Literals, may be shared */
	qu_set      use;      /* Used module (of Module objects) */
	void       *dlhandle; /* External library handle */
	qu_code     code;     /* qu_code object initially 0, see qu_code.h */
	qu_string   source;   /* Source string, during compilation only */
	qu_ast      tree;     /* AST root node, during compilation only */
	qu_dict     doc;      /* Documentation, may be 0 */

	qu_module_getcfunc_f getcfunc; /* to get a static Cfunc object */

#ifdef QU_HAS_UNIVERSE
	qu_array    universes; /* Defined universe names */
#endif

};


/** qu_module_is
	Whether or not the object is a Module object.
*/
#define qu_module_is(o) (qu_classof(o) == QuModuleClass)


/** QU_MODULE_FLAG
	Flags used by Module objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_MODULE_FLAG_DOC        QU_SYMBOL_FLAG_DOC /* documentation */
#define QU_MODULE_FLAG_RESIDENT   QU_FLAG_11         /* module */
#define QU_MODULE_FLAG_EVAL       QU_FLAG_12         /* eval */
#define QU_MODULE_FLAG_TRUSTED    QU_FLAG_13         /* trusted module? */

/* for interp */
#define QU_MODULE_FLAG_FINALIZED  QU_FLAG_14         /* __destroy called? */

/* Status */
#define QU_MODULE_FLAG_PARSED     QU_FLAG_18         /* parsed */
#define QU_MODULE_FLAG_COMPILED   QU_FLAG_19         /* compiled */
#define QU_MODULE_FLAG_READY      QU_FLAG_20         /* ready to run */
#define QU_MODULE_FLAG_INIT       QU_FLAG_21         /* initialized */
#define QU_MODULE_FLAG_ERROR      QU_FLAG_22         /* failed */


/** qu_module_is_trusted
	Whether this is "trusted" module.

	If not, then it is restricted to certain things such as reading files, etc.

	The flag QU_MODULE_FLAG_TRUSTED is set by the interpreter.
	Once set it must never be unset.
*/
#define qu_module_is_trusted(self)  qu_flag_test (self, QU_MODULE_FLAG_TRUSTED)
#define qu_module_set_trusted(self) qu_flag_set  (self, QU_MODULE_FLAG_TRUSTED)


/** qu_module_doc
	The QU_MODULE_FLAG_DOC is set by the compiler.
	It indicates whether the module was compiled as a documentation module.

	Once set it may never be unset.
*/
#define qu_module_is_doc(self)  qu_flag_test (self, QU_MODULE_FLAG_DOC)
#define qu_module_set_doc(self) qu_flag_set  (self, QU_MODULE_FLAG_DOC)


/** qu_module_is_eval
	Whether or not this is an eval module.

	The QU_MODULE_FLAG_EVAL is set by qu_module_new_eval and may only be queried.
*/
#define qu_module_is_eval(self) qu_flag_test (self, QU_MODULE_FLAG_EVAL)


/** qu_module_is_resident
	Whether or not this module is kept in memory all time.

	The QU_MODULE_FLAG_RESIDENT is set by Module and may only be queried.
*/
#define qu_module_is_resident(self) qu_flag_test (self, QU_MODULE_FLAG_RESIDENT)


/** qu_module_parse
	The QU_MODULE_FLAG_PARSED is set/unset by the compiler to keep track of whether
	or not the module has been parsed.
*/
#define qu_module_parsed(self)      qu_flag_test  (self, QU_MODULE_FLAG_PARSED)
#define qu_module_set_parsed(self)   qu_flag_set   (self, QU_MODULE_FLAG_PARSED)
#define qu_module_unset_parsed(self) qu_flag_unset (self, QU_MODULE_FLAG_PARSED)


/** qu_module_compiled
	The QU_MODULE_FLAG_COMPILED is set/unset by the compiler to keep track of whether
	or not the module has been compiled.

	This flag may only be set after a successful parsing
	(the flag QU_MODULE_FLAG_PARSED must be set).
*/
#define qu_module_compiled(self)       qu_flag_test  (self, QU_MODULE_FLAG_COMPILED)
#define qu_module_set_compiled(self)   qu_flag_set   (self, QU_MODULE_FLAG_COMPILED)
#define qu_module_unset_compiled(self) qu_flag_unset (self, QU_MODULE_FLAG_COMPILED)


/** qu_module_ready
	The QU_MODULE_FLAG_READY is set by the compiler to indicate whether or not the module
	is ready to run by the interpreter.

	This flag must only be set after a successful compilation
	(the flag QU_MODULE_FLAG_COMPILED is set).
*/
#define qu_module_is_ready(self)  qu_flag_test (self, QU_MODULE_FLAG_READY)
#define qu_module_set_ready(self) qu_flag_set  (self, QU_MODULE_FLAG_READY)


/** qu_module_error
	The QU_MODULE_FLAG_ERROR is set and used by the compiler to indicate that
	a module has an error (can not be properly compiled).

	Once set this flag must never be unset.
*/
#define qu_module_haserror(self)     qu_flag_test (self, QU_MODULE_FLAG_ERROR)
#define qu_module_set_haserror(self) qu_flag_set  (self, QU_MODULE_FLAG_ERROR)


/** qu_module_initialized
	The QU_MODULE_FLAG_INIT flag is set by the interpreter after the module has been
	initialized. Once set, this flag may never be unset.
*/
#define qu_module_initialized(self)     qu_flag_test (self, QU_MODULE_FLAG_INIT)
#define qu_module_set_initialized(self) qu_flag_set  (self, QU_MODULE_FLAG_INIT)


/** QuModuleDir QuModuleExt QuModuleDlExt
	QuModuleDir is the installed library directory. Generally C{/opt/qu/lib/}
	(with the trailing slash).

	QuModuleExt is our module file extension, e.g. C{".qm"} (with the dot).

	QuModuleDlExt is the shared library file extension, e.g. C{".so"} (with the dot).
*/
extern qu_string QuModuleDir;
extern qu_string QuModuleExt;
extern qu_string QuModuleDlExt;


/** QuModuleMain qu_module_new_main qu_module_new_eval
	qu_module_new_main creates the main program module QuModuleMain.
	Only the main program may call this function.

	qu_module_new_eval creates an k{eval} module. Only the interpreter should call
	this function. See C{interp.c}.
*/
extern qu_module QuModuleMain;
extern qu_module qu_module_new_main (qu_string s, int ispath);
extern qu_module qu_module_new_eval (qu_module self, qu_string src, qu_string path);


/** qu_module_use
	For the compiler, called for each k{use} statements.

	On success, if v{path} is a directory then an ={Array} of Module objects is
	returned, otherwise a Module object is returned. In case of a directory
	then v{as} must be 0 otherwise an exception is thrown.
*/
extern qu_obj qu_module_use (qu_module self, qu_string use, qu_string as, int option);
#define QU_MODULE_USE_DEFAULT  0
#define QU_MODULE_USE_PACKAGE  1
#define QU_MODULE_USE_STANDARD 2


/** qu_module_failed
	For the compiler, when compilation failed. This ensures that failed
	modules are not kept in memory.
*/
extern void qu_module_failed (qu_module self);


/** qu_module_getattr qu_module_setattr
	For the compiler, to get/set module attributes.
*/
extern qu_symbol qu_module_getattr (qu_module self, qu_string name);
extern void      qu_module_setattr (qu_module self, qu_string name, qu_symbol thing);


/** qu_module_getattr_callable_name
	Get class or function which is named the same as the module.
	Return NULL if none found.

	For example, class Foo or function Foo in module C{/opt/qu/lib/Foo.qm}.

	If the symbol is a variable and contains a callable then return that
	callable iff v{runtime}, or not v{runtime} and variable is frozen (constant).
*/
extern qu_symbol qu_module_getattr_callable_name (qu_module self, int runtime);


/** qu_module_import qu_module_import_all
	For the compiler, to import symbols from another Module.
	See C{qu_code.c}.
*/
extern int qu_module_import     (qu_module self, qu_module u, qu_string name);
extern int qu_module_import_all (qu_module self, qu_module u);


/** qu_module_loaded
	If the module is loaded then return the Module object, NULL otherwise.
	v{path} must be an absolute path.
*/
extern qu_module qu_module_loaded (qu_string path);


/** qu_module_load qu_module_math
	qu_module_load loads a Module.

	qu_module_math loads the Math module. Reload it if necessary.

	qu_module_html loads the html module. Reload it if necessary.
*/
extern qu_module qu_module_load (qu_string path);
extern qu_module qu_module_math ();
extern qu_module qu_module_html ();
/***/


/** qu_module_add_sub
	Add a new Sub.

	qu_module_add_sub creates a module level function.
*/
extern qu_sub qu_module_add_sub (qu_module self, const char *name, void *func, int argc, long flags);

#define qu_module_add_function(self, namez, faddr, argc) \
	qu_module_add_sub (self, namez, faddr, argc, QU_SUB_FLAG_CFUNC)


/** qu_module_add_var
	Add a new Var.
*/
extern qu_var qu_module_add_var (qu_module self, const char *name, qu_any value);


/** qu_module_add_const qu_module_add_float qu_module_add_int qu_module_add_string
	Add a new constant.
*/
extern qu_var qu_module_add_const  (qu_module self, const char *name, qu_any value);
extern qu_var qu_module_add_float  (qu_module self, const char *name, double val);
extern qu_var qu_module_add_int    (qu_module self, const char *name, long val);
extern qu_var qu_module_add_string (qu_module self, const char *name, const char *val);


/** qu_module_add_alias
	Add an alias of an existing symbol.
*/
extern qu_symbol qu_module_add_alias (qu_module self, const char *name, const char *alias);


/** qu_dl_xxx
	Interface to dlfcn. Use these instead of the dl functions directly.

	qu_dl_open_ex opens the library defined in the qu.conf file.
*/
#define QU_DL_LAZY    RTLD_LAZY
#define QU_DL_NOW     RTLD_NOW
#define QU_DL_GLOBAL  RTLD_GLOBAL
#define QU_DL_DEFAULT RTLD_DEFAULT
#define QU_DL_NEXT    RTLD_NEXT

extern void      *qu_dl_open    (const char *filename, int flag);
extern void      *qu_dl_open_ex (const char *name, int flag);
extern int        qu_dl_close   (void *handle);
extern void      *qu_dl_error   ();
extern void      *qu_dl_sym     (void *handle, char *symbol);
extern qu_string  qu_dl_config  (const char *name);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class. */
extern void qu_module__setup ();
extern void qu_module__init  ();

/* for Compile */
extern void qu_module_add_use (qu_module self, qu_module m, qu_string as);

/* these are set by the html module */
typedef qu_null (*htopen_f)   (qu_string s, long op);
typedef qu_null (*htclose_f)  (qu_string s, int linebreak);
typedef qu_null (*htentity_f) (qu_string s);

extern htopen_f   __QuModuleHtml_open;
extern htclose_f  __QuModuleHtml_close;
extern htentity_f __QuModuleHtml_entity;

#ifdef QU_HAS_UNIVERSE
extern qu_dict QuUniverses;
extern qu_symbol qu_universe_find_symbol (qu_string uniname, qu_string symname);
#endif

/* for C modules with __extern support */
extern qu_any qu_module_get_cfunc (qu_module self, void *handle, qu_string name, qu_string spec, qu_dict resolved);

#endif /*_H_qu_module_*/
