/* Number.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_number_
#define _H_qu_number_ 1


/** Number qu_number
	This is the abstract parent class of all number objects.

	qu_number represents a Number object.
*/
extern qu_class QuNumberClass;


/** QuNumberN
	Predefined system objects.
*/
#define QuNumber0  ((qu_number) QuInt0)   /*  0 */
#define QuNumber1  ((qu_number) QuInt1)   /*  1 */
#define QuNumber2  ((qu_number) QuInt2)   /*  2 */
#define QuNumber_1 ((qu_number) QuInt_1)  /* -1 */
#define QuNumber_2 ((qu_number) QuInt_2)  /* -2 */


/** qu_number_context
	Qu provides a global number context that can be set by an application.

	__QuNumberContext is a combination of the QU_NUMBER_CONTEXT flags.

	@warn
	All number classes must respect the setting.
*/
#define qu_number_context_get()      (__QuNumberContext)
#define qu_number_context_set(value) QU_BEGIN __QuNumberContext = value; QU_END

#define QU_NUMBER_CONTEXT_ALLOW_INF       0x01 /* throw EInf or EBig or return an infinity */
#define QU_NUMBER_CONTEXT_ALLOW_TINY      0x02 /* throw ETiny or return 0 */
#define QU_NUMBER_CONTEXT_ALLOW_HUGE      0x04 /* throw EHuge or return an infinity */
#define QU_NUMBER_CONTEXT_ALLOW_ZERODIV   0x08 /* throw EDiv or return an infinity */
#define QU_NUMBER_CONTEXT_ALLOW_COMPLEX   0x10 /* throw EComplex if result is Complex */
#define QU_NUMBER_CONTEXT_REALIFY_COMPLEX 0x20 /* return Real if imaginary part is 0 */
#define QU_NUMBER_CONTEXT_EXACT           0x40 /* try to return "exact" values */

/* this is the default setting */
#define QU_NUMBER_CONTEXT_DEFAULT \
	(QU_NUMBER_CONTEXT_ALLOW_TINY \
	| QU_NUMBER_CONTEXT_ALLOW_HUGE \
	| QU_NUMBER_CONTEXT_ALLOW_ZERODIV \
	| QU_NUMBER_CONTEXT_ALLOW_COMPLEX)

/* test macros */
#define QU_NUMBER_ALLOW_INF       (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_INF)
#define QU_NUMBER_ALLOW_HUGE      (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_HUGE)
#define QU_NUMBER_ALLOW_TINY      (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_TINY)
#define QU_NUMBER_ALLOW_ZERODIV   (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_ZERODIV)
#define QU_NUMBER_ALLOW_COMPLEX   (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_COMPLEX)
#define QU_NUMBER_REALIFY_COMPLEX (__QuNumberContext & QU_NUMBER_CONTEXT_REALIFY_COMPLEX)
#define QU_NUMBER_EXACT           (__QuNumberContext & QU_NUMBER_CONTEXT_EXACT)

/* ensure macros */
#define QU_NUMBER_ENSURE_ALLOW_INF      QU_REQUIRE (QuEInf    , QU_NUMBER_ALLOW_INF)
#define QU_NUMBER_ENSURE_ALLOW_HUGE     QU_REQUIRE (QuEHuge   , QU_NUMBER_ALLOW_HUGE)
#define QU_NUMBER_ENSURE_ALLOW_TINY     QU_REQUIRE (QuETiny   , QU_NUMBER_ALLOW_TINY)
#define QU_NUMBER_ENSURE_ALLOW_ZERODIV  QU_REQUIRE (QuEDiv    , QU_NUMBER_ALLOW_ZERODIV)
#define QU_NUMBER_ENSURE_ALLOW_COMPLEX  QU_REQUIRE (QuEComplex, QU_NUMBER_ALLOW_COMPLEX)
/***/
extern volatile long __QuNumberContext;


/** __qu_number_context
	Sometimes it is necessary to temporarily change the context to perform some
	calculations. Use these macros for that.

	@xmp
		int saved_context;
		__qu_number_context_begin (saved_context, __QuNumberContext | QU_NUMBER_ALLOW_INF)
		  ...
		__qu_number_context_end (saved_context);

	@p
	Since any thread can change the context at any time then it might be necessary
	to use ={__atomic_begin}/end.

	@xmp
		int saved_context;
		__qu_atomic_begin;
		__qu_number_context_begin (saved_context, QU_NUMBER_CONTEXT_DEFAULT)
		  ...
		__qu_number_context_end (saved_context);
		__qu_atomic_end;
*/
#define __qu_number_context_begin(var, new) \
QU_BEGIN \
	var = __QuNumberContext; \
	__QuNumberContext = new; \
QU_END

#define __qu_number_context_end(var) \
QU_BEGIN \
	__QuNumberContext = var; \
QU_END


/** qu_number_new
	Create a Number object from C types or other objects.

	These are just convenient wrappers around ={qu_real_new} and ={qu_complex_new}.
*/
#define qu_number_newh(v)   ((qu_number) qu_real_newh   (v))
#define qu_number_newuh(v)  ((qu_number) qu_real_newuh  (v))
#define qu_number_newi(v)   ((qu_number) qu_real_newi   (v))
#define qu_number_newui(v)  ((qu_number) qu_real_newui  (v))
#define qu_number_newl(v)   ((qu_number) qu_real_newl   (v))
#define qu_number_newul(v)  ((qu_number) qu_real_newul  (v))
#define qu_number_newli(v)  ((qu_number) qu_real_newli  (v))
#define qu_number_newd(v)   ((qu_number) qu_real_newd   (v))
#ifdef HAVE_LONG_LONG
#define qu_number_newll(v)  ((qu_number) qu_real_newll  (v))
#define qu_number_newull(v) ((qu_number) qu_real_newull (v))
#endif
#ifdef HAVE_LONG_DOUBLE
#define qu_number_newld(v)  ((qu_number) qu_real_newui  (v))
#endif

#define qu_number_new_int(x)      ((qu_number) (x))
#define qu_number_new_float(x)    ((qu_number) (x))
#define qu_number_new_rational(x) ((qu_number) (x))
#define qu_number_new_big(x)      ((qu_number) (x))
#define qu_number_new_complex(x)  ((qu_number) (x))
#define qu_number_new_integer(x)  ((qu_number) (x))
#define qu_number_new_decimal(x)  ((qu_number) (x))
#define qu_number_new_real(x)     ((qu_number) (x))
#define qu_number_new_number(x)   (x)


/** qu_number_to
	Functions to convert any object into C types or other qu_number objects.

	The functions set v{errno} to zero on success. On failure, v{errno} is set to EDOM
	if v{x} is conversion is not possible, to ERANGE if the value falls out of the C
	type range. The I{unsigned} versions fail with ERANGE if the value is negative.

	@note
	Qu uses a global abbreviation convention for C types.
	@code
		h     short
		uh    unsigned short
		i     int
		ui    unsigned int
		i16   int 2 bytes
		ui16  unsigned int 2 bytes
		i32   int 4 bytes
		ui32  unsigned int 4 bytes
		l     long
		ul    unsigned long
		li    long in the qu_int range
		d     double
		ll    long long
		ull   unsigned long long
		ld    long double

	@p
	These are just convenient wrappers around ={qu_real_to} and ={qu_complex_to}.
*/
#define                   qu_number_tob(x)    qu_number__ltob     (qu_number_tol  (x))
#define                   qu_number_toub(x)   qu_number__ultoub    (qu_number_toul (x))
#define                   qu_number_toh(x)    qu_number__ltoh     (qu_number_tol  (x))
#define                   qu_number_touh(x)   qu_number__ultouh   (qu_number_toul (x))
#define                   qu_number_toi(x)    qu_number__ltoi     (qu_number_tol  (x))
#define                   qu_number_toui(x)   qu_number__ultoui   (qu_number_toul (x))
#define                   qu_number_toi16(x)  qu_number__ltoi16   (qu_number_tol  (x))
#define                   qu_number_toui16(x) qu_number__ultoui16 (qu_number_toul (x))
#define                   qu_number_toi32(x)  qu_number__ltoi32   (qu_number_tol  (x))
#define                   qu_number_toui32(x) qu_number__ultoui32 (qu_number_toul (x))
extern long               qu_number_tol       (qu_any x);
extern unsigned long      qu_number_toul      (qu_any x);
extern long               qu_number_toli      (qu_any x);
#define                   qu_number_touli(x)  qu_number__ltouli   (qu_number_tol (x))
#define                   qu_number_tof(x)    qu_number__dtof     (qu_number_tod  (x))
extern double             qu_number_tod       (qu_any x);
#ifdef HAVE_LONG_LONG
extern long long          qu_number_toll      (qu_any x);
extern unsigned long long qu_number_toull     (qu_any x);
#endif
#ifdef HAVE_LONG_DOUBLE
extern long double        qu_number_told      (qu_any x);
#endif


/** qu_number_as
	Convert any object into a Number objects.

	An exception is thrown and NULL is returned on failure.

	@p
	These are just convenient wrappers around ={qu_real_as} and ={qu_complex_as}.
*/
extern qu_int      qu_number_as_int      (qu_any x);
extern qu_float    qu_number_as_float    (qu_any x);
extern qu_long     qu_number_as_long     (qu_any x);
extern qu_rational qu_number_as_rational (qu_any x);
extern qu_big      qu_number_as_big      (qu_any x);
extern qu_complex  qu_number_as_complex  (qu_any x);
extern qu_integer  qu_number_as_integer  (qu_any x);
extern qu_decimal  qu_number_as_decimal  (qu_any x);
extern qu_real     qu_number_as_real     (qu_any x);
extern qu_number   qu_number_as_number   (qu_any x);


/** qu_number_sign
	Returns -1, 0, 1 if the value is negative, 0, positive respectively.

	This is just a convenient wrapper around ={qu_real_sign} and ={qu_complex_sign}.
*/
extern int qu_number_sign (qu_number x);


/** qu_number_is
	Identity functions.

	qu_number_is tests whether or not an object is a Number.

	qu_number_isneg returns 1 if the value is negative.
	Returns 0 otherwise.

	qu_number_iszero returns 1 if the value is 0.
	Returns 0 otherwise.

	qu_number_isone returns 1 if the value is 1 or -1.
	Returns 0 otherwise.

	qu_number_is1 return 1 if the value is 1.
	Returns 0 otherwise.

	qu_number_iswhole return 1 if the value is a whole number.
	Returns 0 otherwise.

	qu_number_isfrac return 1 if 0 < |x| < 1.
	Returns 0 otherwise.

	qu_number_isodd return 1 if the value is a whole and odd number.
	Returns 0 otherwise.

	These are just convenient wrappers around ={qu_real_is} and ={qu_complex_is}.
*/
extern int qu_number_is      (qu_any x);
extern int qu_number_iszero  (qu_number x);
extern int qu_number_isneg   (qu_number x);
extern int qu_number_isone   (qu_number x);
extern int qu_number_is1     (qu_number x);
extern int qu_number_iswhole (qu_number x);
extern int qu_number_isfrac  (qu_number x);
extern int qu_number_isodd   (qu_number x);


/** qu_number_unary
	Unary functions.

	qu_number_pos returns the number itself, obviously.

	qu_number_neg returns a negated value.

	qu_number_incr returns the number incremented by 1.

	qu_number_decr returns the number decremented by 1.

	These are just convenient wrappers around
	={qu_real_pos} ={qu_real_neg} ={qu_real_incr} ={qu_real_decr}
	={qu_complex_pos} ={qu_complex_neg} ={qu_complex_incr} ={qu_complex_decr}.
*/
extern qu_number qu_number_pos  (qu_number x);
extern qu_number qu_number_neg  (qu_number x);
extern qu_number qu_number_incr (qu_number x);
extern qu_number qu_number_decr (qu_number x);


/** qu_number_eq
	Whether the value is the same as the value of another.
	All return 0 or 1.

	These are just convenient wrappers around ={qu_real_eq} and ={qu_complex_eq}.
*/
extern int qu_number_eq_int      (qu_number x, qu_int y);
extern int qu_number_eq_float    (qu_number x, qu_float y);
extern int qu_number_eq_long     (qu_number x, qu_long y);
extern int qu_number_eq_rational (qu_number x, qu_rational y);
extern int qu_number_eq_big      (qu_number x, qu_big y);
extern int qu_number_eq_complex  (qu_number x, qu_complex y);
extern int qu_number_eq_inf      (qu_number x, qu_inf y);
extern int qu_number_eq_integer  (qu_number x, qu_integer y);
extern int qu_number_eq_decimal  (qu_number x, qu_decimal y);
extern int qu_number_eq_real     (qu_number x, qu_real y);
extern int qu_number_eq_number   (qu_number x, qu_number y);


/** qu_number_like
	Whether the absolute value is the same as the absolute value of another.
	All return 0 or 1.

	These are just convenient wrappers around ={qu_real_like} and ={qu_complex_like}.
*/
extern int qu_number_like_int      (qu_number x, qu_int y);
extern int qu_number_like_float    (qu_number x, qu_float y);
extern int qu_number_like_long     (qu_number x, qu_long y);
extern int qu_number_like_rational (qu_number x, qu_rational y);
extern int qu_number_like_big      (qu_number x, qu_big y);
extern int qu_number_like_complex  (qu_number x, qu_complex y);
extern int qu_number_like_inf      (qu_number x, qu_inf y);
extern int qu_number_like_integer  (qu_number x, qu_integer y);
extern int qu_number_like_decimal  (qu_number x, qu_decimal y);
extern int qu_number_like_real     (qu_number x, qu_real y);
extern int qu_number_like_number   (qu_number x, qu_number y);


/** qu_number_cmp
	Compare with another.
	All return -1, 0, or -1 in the usual manner.

	These are just convenient wrappers around ={qu_real_cmp} and ={qu_complex_cmp}.
*/
extern int qu_number_cmp_int      (qu_number x, qu_int y);
extern int qu_number_cmp_float    (qu_number x, qu_float y);
extern int qu_number_cmp_long     (qu_number x, qu_long y);
extern int qu_number_cmp_rational (qu_number x, qu_rational y);
extern int qu_number_cmp_big      (qu_number x, qu_big y);
extern int qu_number_cmp_complex  (qu_number x, qu_complex y);
extern int qu_number_cmp_inf      (qu_number x, qu_inf y);
extern int qu_number_cmp_integer  (qu_number x, qu_integer y);
extern int qu_number_cmp_decimal  (qu_number x, qu_decimal y);
extern int qu_number_cmp_real     (qu_number x, qu_real y);
extern int qu_number_cmp_number   (qu_number x, qu_number y);


/** qu_number_cmpabs
	Compare the absolute value with the absolute value of another.
	All return -1, 0, or -1 in the usual manner.

	These are just convenient wrappers around ={qu_real_cmpabs} and ={qu_complex_cmpabs}.
*/
extern int qu_number_cmpabs_int      (qu_number x, qu_int y);
extern int qu_number_cmpabs_float    (qu_number x, qu_float y);
extern int qu_number_cmpabs_long     (qu_number x, qu_long y);
extern int qu_number_cmpabs_rational (qu_number x, qu_rational y);
extern int qu_number_cmpabs_big      (qu_number x, qu_big y);
extern int qu_number_cmpabs_complex  (qu_number x, qu_complex y);
extern int qu_number_cmpabs_inf      (qu_number x, qu_inf y);
extern int qu_number_cmpabs_integer  (qu_number x, qu_integer y);
extern int qu_number_cmpabs_decimal  (qu_number x, qu_decimal y);
extern int qu_number_cmpabs_real     (qu_number x, qu_real y);
extern int qu_number_cmpabs_number   (qu_number x, qu_number y);


/** qu_number_plus
	Add with another.

	These are just convenient wrappers around ={qu_real_plus} and ={qu_complex_plus}.
*/
extern qu_number  qu_number_plus_int      (qu_number x, qu_int y);
extern qu_number  qu_number_plus_float    (qu_number x, qu_float y);
extern qu_number  qu_number_plus_long     (qu_number x, qu_long y);
extern qu_number  qu_number_plus_rational (qu_number x, qu_rational y);
extern qu_number  qu_number_plus_big      (qu_number x, qu_big y);
extern qu_complex qu_number_plus_complex  (qu_number x, qu_complex y);
extern qu_number  qu_number_plus_inf      (qu_number x, qu_inf y);
extern qu_number  qu_number_plus_integer  (qu_number x, qu_integer y);
extern qu_number  qu_number_plus_decimal  (qu_number x, qu_decimal y);
extern qu_number  qu_number_plus_real     (qu_number x, qu_real y);
extern qu_number  qu_number_plus_number   (qu_number x, qu_number y);


/** qu_number_minus
	Subtract with another.

	These are just convenient wrappers around ={qu_real_minus} and ={qu_complex_minus}.
*/
extern qu_number  qu_number_minus_int      (qu_number x, qu_int y);
extern qu_number  qu_number_minus_float    (qu_number x, qu_float y);
extern qu_number  qu_number_minus_long     (qu_number x, qu_long y);
extern qu_number  qu_number_minus_rational (qu_number x, qu_rational y);
extern qu_number  qu_number_minus_big      (qu_number x, qu_big y);
extern qu_complex qu_number_minus_complex  (qu_number x, qu_complex y);
extern qu_number  qu_number_minus_inf      (qu_number x, qu_inf y);
extern qu_number  qu_number_minus_integer  (qu_number x, qu_integer y);
extern qu_number  qu_number_minus_decimal  (qu_number x, qu_decimal y);
extern qu_number  qu_number_minus_real     (qu_number x, qu_real y);
extern qu_number  qu_number_minus_number   (qu_number x, qu_number y);


/** qu_number_mul
	Multiply with another.

	These are just convenient wrappers around ={qu_real_mul} and ={qu_complex_mul}.
*/
extern qu_number  qu_number_mul_int      (qu_number x, qu_int y);
extern qu_number  qu_number_mul_float    (qu_number x, qu_float y);
extern qu_number  qu_number_mul_long     (qu_number x, qu_long y);
extern qu_number  qu_number_mul_rational (qu_number x, qu_rational y);
extern qu_number  qu_number_mul_big      (qu_number x, qu_big y);
extern qu_complex qu_number_mul_complex  (qu_number x, qu_complex y);
extern qu_number  qu_number_mul_inf      (qu_number x, qu_inf y);
extern qu_number  qu_number_mul_integer  (qu_number x, qu_integer y);
extern qu_number  qu_number_mul_decimal  (qu_number x, qu_decimal y);
extern qu_number  qu_number_mul_real     (qu_number x, qu_real y);
extern qu_number  qu_number_mul_number   (qu_number x, qu_number y);


/** qu_number_div
	Divide with another.

	These are just convenient wrappers around ={qu_real_div} and ={qu_complex_div}.
*/
extern qu_number  qu_number_div_int      (qu_number x, qu_int y);
extern qu_number  qu_number_div_float    (qu_number x, qu_float y);
extern qu_number  qu_number_div_long     (qu_number x, qu_long y);
extern qu_number  qu_number_div_rational (qu_number x, qu_rational y);
extern qu_number  qu_number_div_big      (qu_number x, qu_big y);
extern qu_complex qu_number_div_complex  (qu_number x, qu_complex y);
extern qu_number  qu_number_div_inf      (qu_number x, qu_inf y);
extern qu_number  qu_number_div_integer  (qu_number x, qu_integer y);
extern qu_number  qu_number_div_decimal  (qu_number x, qu_decimal y);
extern qu_number  qu_number_div_real     (qu_number x, qu_real y);
extern qu_number  qu_number_div_number   (qu_number x, qu_number y);


/** qu_number_idiv
	Integer division with another.

	These are just convenient wrappers around ={qu_real_idiv} and ={qu_complex_idiv}.
*/
extern qu_integer qu_number_idiv_int      (qu_number x, qu_int y);
extern qu_integer qu_number_idiv_float    (qu_number x, qu_float y);
extern qu_integer qu_number_idiv_long     (qu_number x, qu_long y);
extern qu_integer qu_number_idiv_rational (qu_number x, qu_rational y);
extern qu_integer qu_number_idiv_big      (qu_number x, qu_big y);
extern qu_integer qu_number_idiv_complex  (qu_number x, qu_complex y);
extern qu_integer qu_number_idiv_inf      (qu_number x, qu_inf y);
extern qu_integer qu_number_idiv_integer  (qu_number x, qu_integer y);
extern qu_integer qu_number_idiv_decimal  (qu_number x, qu_decimal y);
extern qu_integer qu_number_idiv_real     (qu_number x, qu_real y);
extern qu_integer qu_number_idiv_number   (qu_number x, qu_number y);


/** qu_number_rem
	Remainder of dividing with another.

	These are just convenient wrappers around ={qu_real_rem} and ={qu_complex_rem}.
*/
extern qu_number  qu_number_rem_int      (qu_number x, qu_int y);
extern qu_number  qu_number_rem_float    (qu_number x, qu_float y);
extern qu_number  qu_number_rem_long     (qu_number x, qu_long y);
extern qu_number  qu_number_rem_rational (qu_number x, qu_rational y);
extern qu_number  qu_number_rem_big      (qu_number x, qu_big y);
extern qu_complex qu_number_rem_complex  (qu_number x, qu_complex y);
extern qu_number  qu_number_rem_inf      (qu_number x, qu_inf y);
extern qu_number  qu_number_rem_integer  (qu_number x, qu_integer y);
extern qu_number  qu_number_rem_decimal  (qu_number x, qu_decimal y);
extern qu_number  qu_number_rem_real     (qu_number x, qu_real y);
extern qu_number  qu_number_rem_number   (qu_number x, qu_number y);


/** qu_number_pow
	Raise to the power of another.

	These are just convenient wrappers around ={qu_real_pow} and ={qu_complex_pow}.
*/
extern qu_number  qu_number_pow_int      (qu_number x, qu_int y);
extern qu_number  qu_number_pow_float    (qu_number x, qu_float y);
extern qu_number  qu_number_pow_long     (qu_number x, qu_long y);
extern qu_number  qu_number_pow_rational (qu_number x, qu_rational y);
extern qu_number  qu_number_pow_big      (qu_number x, qu_big y);
extern qu_complex qu_number_pow_complex  (qu_number x, qu_complex y);
extern qu_number  qu_number_pow_inf      (qu_number x, qu_inf y);
extern qu_number  qu_number_pow_integer  (qu_number x, qu_integer y);
extern qu_number  qu_number_pow_decimal  (qu_number x, qu_decimal y);
extern qu_number  qu_number_pow_real     (qu_number x, qu_real y);
extern qu_number  qu_number_pow_number   (qu_number x, qu_number y);


/** qu_number_round
	Rounding functions.

	qu_number_ceil rounds the number to the nearest integer towards positive infinity.
	EBig is thrown on failure.

	qu_number_floor rounds the number to the nearest integer towards negative infinity.
	EBig is thrown on failure.

	qu_number_trunc round the number towards zero, that is, it removes the
	fraction part. This function throws EMath if the number is an infinity.

	These are just convenient wrappers around
	={qu_real_floor} ={qu_real_ceil} ={qu_real_trunc}
	={qu_complex_floor} ={qu_complex_ceil} ={qu_complex_trunc}
*/
extern qu_number qu_number_floor (qu_number x);
extern qu_number qu_number_ceil  (qu_number x);
extern qu_number qu_number_trunc (qu_number x);
extern qu_number qu_number_frac  (qu_number x);
extern qu_number qu_number_round (qu_number x, int prec);


/** qu_number_format
	Format a Number.

	See ={Format} for details.

	This is just a convenient wrapper around ={qu_real_format} and ={qu_complex_format}.

	This function is provided for efficiency between core modules.
	It is not intended to be called by users directly.
*/
extern qu_string qu_number_format (qu_number x, qu_format_s *f);


/** qu_number_scan
	Scan number.

	v{s} is the start of the string to scan.

	If v{options} does not contain the QU_NUMBER_SCAN_SKIPSPACE flag then the scanner
	will not skip leading blanks and will not read the sign, that is, the string may
	not contain the sign nor leading spaces.

	If it contains the QU_NUMBER_SCAN_COMPLEX flag then the scanner will treat
	things like C{-123j} and {-123 + 456j} as a complex number.

	On success a qu_number object is returned, NULL otherwise.

	qu_number_scan_string does the same but throws ESyntax on failure.
*/
extern qu_number qu_number_scan        (char **s, int options);
extern qu_number qu_number_scan_string (qu_string s, int options);

#define QU_NUMBER_SCAN_SKIPSPACE 0x1
#define QU_NUMBER_SCAN_COMPLEX   0x2


/** QU_NUMBER_COMPARE_VALUE
	Compare two C values, yielding -1, 0 or 1 if v{x} is smaller, the same,
	and greater than v{y} respectively.
*/
#define QU_NUMBER_COMPARE_VALUE(x, y)  (((x) > (y)) ? 1 : ((x) < (y)) ? -1 : 0)


/** QU_NUMBER_COMPARE_SIGN
	Compare two signs and issues the appropriate return statements.
	This is generally used on top of cmp functions.
*/
#define QU_NUMBER_COMPARE_SIGN(sx, sy) \
QU_BEGIN \
	if ((sx) > 0) \
	{ \
		if ((sy) <= 0) \
			return 1; \
	} \
	else if ((sx) < 0) \
	{ \
		if ((sy) >= 0) \
			return -1; \
	} \
	else \
	{ \
		return -(sy); \
	} \
QU_END
/* ... here you must compare the values with the same sign */



/** QU_NUMBER_BINOP_integer
	Make a binop function with an Integer, e.g:
	@code
		qu_bar qu_foo_plus_integer (qu_foo x, qu_integer y)

	@p
	Foo must provide the functions:
	@code
		qu_bar qu_foo_plus_int  (qu_foo x, qu_int  y) ...
		qu_bar qu_foo_plus_long (qu_foo x, qu_long y) ...
*/
#define QU_NUMBER_BINOP_integer(xtype, name, rtype) \
rtype xtype##_##name##_integer (xtype x, qu_integer y) \
{ \
	return qu_int_is (y) \
		? (rtype) xtype##_##name##_int  (x, (qu_int ) y) \
		: (rtype) xtype##_##name##_long (x, (qu_long) y); \
}


/** QU_NUMBER_BINOP_decimal
	Make a binop function with a Decimal, e.g:
	@code
		qu_bar qu_foo_plus_decimal (qu_foo x, qu_decimal y)

	@p
	Foo must provide the functions:
	@code
		qu_bar qu_foo_plus_float (qu_foo x, qu_float y) ...
		qu_bar qu_foo_plus_big   (qu_foo x, qu_big   y) ...
*/
#define QU_NUMBER_BINOP_decimal(xtype, name, rtype) \
rtype xtype##_##name##_decimal (xtype x, qu_decimal y) \
{ \
	return qu_float_is (y) \
		? (rtype) xtype##_##name##_float (x, (qu_float) y) \
		: (rtype) xtype##_##name##_big   (x, (qu_big  ) y); \
}


/** QU_NUMBER_BINOP_real
	Make a binop function with a Real, e.g:
	@code
		qu_bar qu_foo_plus_real (qu_foo x, qu_real y)

	@p
	Foo must provide the functions:
	@code
		qu_bar qu_foo_plus_int      (qu_foo x, qu_int     y) ...
		qu_bar qu_foo_plus_long     (qu_foo x, qu_long    y) ...
		qu_bar qu_foo_plus_rational (qu_foo x, qu_rational y) ...
		qu_bar qu_foo_plus_float    (qu_foo x, qu_float    y) ...
		qu_bar qu_foo_plus_big      (qu_foo x, qu_big      y) ...
		qu_bar qu_foo_plus_inf      (qu_foo x, qu_inf      y) ...
*/
#define QU_NUMBER_BINOP_real(xtype, name, rtype) \
rtype xtype##_##name##_real (xtype x, qu_real y) \
{ \
	switch (qu_classid (y)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) xtype##_##name##_int      (x, (qu_int     ) y); \
		case QU_CLASS_ID_Long:     return (rtype) xtype##_##name##_long     (x, (qu_long    ) y); \
		case QU_CLASS_ID_Rational: return (rtype) xtype##_##name##_rational (x, (qu_rational) y); \
		case QU_CLASS_ID_Float:    return (rtype) xtype##_##name##_float    (x, (qu_float   ) y); \
		case QU_CLASS_ID_Big:      return (rtype) xtype##_##name##_big      (x, (qu_big     ) y); \
		case QU_CLASS_ID_Inf:      return (rtype) xtype##_##name##_inf      (x, (qu_inf     ) y); \
	} \
	QU_FATAL ("y is not a Real object"); \
	return 0; /*never*/ \
}


/** QU_NUMBER_BINOP_number
	Make a binop function with a Number, e.g:
	@code
		qu_bar qu_foo_plus_number (qu_foo x, qu_number y)

	@p
	Foo must provide the functions:
	@code
		qu_bar qu_foo_plus_int      (qu_foo x, qu_int     y) ...
		qu_bar qu_foo_plus_long     (qu_foo x, qu_long    y) ...
		qu_bar qu_foo_plus_rational (qu_foo x, qu_rational y) ...
		qu_bar qu_foo_plus_float    (qu_foo x, qu_float    y) ...
		qu_bar qu_foo_plus_big      (qu_foo x, qu_big      y) ...
		qu_bar qu_foo_plus_inf      (qu_foo x, qu_inf      y) ...
		qu_bar qu_foo_plus_complex  (qu_foo x, qu_complex  y) ...
*/
#define QU_NUMBER_BINOP_number(xtype, name, rtype) \
rtype xtype##_##name##_number (xtype x, qu_number y) \
{ \
	switch (qu_classid (y)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) xtype##_##name##_int      (x, (qu_int     ) y); \
		case QU_CLASS_ID_Long:     return (rtype) xtype##_##name##_long     (x, (qu_long    ) y); \
		case QU_CLASS_ID_Rational: return (rtype) xtype##_##name##_rational (x, (qu_rational) y); \
		case QU_CLASS_ID_Float:    return (rtype) xtype##_##name##_float    (x, (qu_float   ) y); \
		case QU_CLASS_ID_Big:      return (rtype) xtype##_##name##_big      (x, (qu_big     ) y); \
		case QU_CLASS_ID_Inf:      return (rtype) xtype##_##name##_inf      (x, (qu_inf     ) y); \
		case QU_CLASS_ID_Complex:  return (rtype) xtype##_##name##_complex  (x, (qu_complex ) y); \
	} \
	QU_FATAL ("y is not a Number object"); \
	return 0; /*never*/ \
}


/** qu_number_helpers
	C type conversion helpers.

	All must be entered with errno = 0 or ERANGE.
	These function do not change errno.

	If entered with errno == ERANGE then return a minimum/maximum value.
*/
extern char           qu_number__ltob     (long v);
extern unsigned char  qu_number__ultoub   (unsigned long v);
extern short          qu_number__ltoh     (long v);
extern unsigned short qu_number__ultouh   (unsigned long v);
extern int            qu_number__ltoi16   (long v);
extern unsigned int   qu_number__ultoui16 (unsigned long v);
extern long           qu_number__ltoi32   (long v);
extern unsigned long  qu_number__ultoui32 (unsigned long v);
extern long           qu_number__ltoli    (long v);
extern long           qu_number__ltouli   (long v);
#if SIZEOF_INT != SIZEOF_LONG
extern int            qu_number__ltoi     (long v);
extern unsigned int   qu_number__ultoui   (unsigned long v);
#else
#define               qu_number__ltoi(v)   ((int) (v))
#define               qu_number__ultoui(v) ((unsigned int) (v))
#endif
extern float          qu_number__dtof     (double v);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_number__setup ();
extern void qu_number__init  ();

/* for Sys */
extern qu_string qu__number_context (qu_class self, int argc, qu_any *argv);

#endif /*_H_qu_number_*/
