/* Object.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_object_
#define _H_qu_object_ 1

#include "base.h"


#if (SIZEOF_LONG != SIZEOF_VOID_P)
#error *** We need sizeof (long) = sizeof (void *)
#endif


/** QU_OBJECT_HEAD
	All object structs must use QU_OBJECT_HEAD as the first entry.

	v{klass} must be the first field. You must never change this because
	it is used by the GC to mark an object or memory chunk as being freed.
	It is also the most frequent used field, so this helps performance too.

	v{attrs} is used to store the attributes and is initially set to NULL
	by the GC. It must only be allocated as necessary since many objects
	may never need it. It is not efficient for objects that don't use it
	(e.g. String, Float, etc.) but lets be practical. Most user objects
	will use it anyway.

	v{flags} are used by most builtin objects. Program level objects usually
	don't use it but Qu provides program level methods to access it so this
	is also not entirely wasted. See ={QU_FLAG}.
*/
#define QU_OBJECT_HEAD \
	qu_class   klass; /* class object, never 0 */ \
	long       flags; /* see QU_FLAG */ \
	qu_set     weak;  /* weak references, may be 0 */ \
	qu_attr_s *attrs  /* attributes, may be 0 */

#ifndef QU_ATTR_S_DECLARED
#define QU_ATTR_S_DECLARED
typedef struct qu_attr_s qu_attr_s;  /* see attr.h */
#endif


/** qu_object_s qu_obj qu_object qu_any
	Predeclare all builtin object structs here.

	qu_any is used for any object including ={Int} while v{qu_obj} is used for
	any object excluding Int.
*/
typedef struct qu_object_s    qu_object_s;
typedef struct qu_array_s     qu_array_s;
typedef struct qu_ast_s       qu_ast_s;
typedef struct qu_big_s       qu_big_s;
typedef struct qu_buffer_s    qu_buffer_s;
typedef struct qu_class_s     qu_class_s;
typedef struct qu_code_s      qu_code_s;
typedef struct qu_complex_s   qu_complex_s;
typedef struct qu_container_s qu_container_s;
typedef struct qu_dict_s      qu_dict_s;
typedef struct qu_dlib_s      qu_dlib_s;
typedef struct qu_exception_s qu_exception_s;
typedef struct qu_float_s     qu_float_s;
typedef struct qu_frame_s     qu_frame_s;
typedef struct qu_file_s      qu_file_s;
typedef struct qu_iterator_s  qu_iterator_s;
typedef struct qu_long_s      qu_long_s;
typedef struct qu_method_s    qu_method_s;
typedef struct qu_module_s    qu_module_s;
typedef struct qu_opaque_s    qu_opaque_s;
typedef struct qu_proc_s      qu_proc_s;
typedef struct qu_pointer_s   qu_pointer_s;
typedef struct qu_queue_s     qu_queue_s;
typedef struct qu_range_s     qu_range_s;
typedef struct qu_rational_s  qu_rational_s;
typedef struct qu_record_s    qu_record_s;
typedef struct qu_set_s       qu_set_s;
typedef struct qu_stream_s    qu_stream_s;
typedef struct qu_string_s    qu_string_s;
typedef struct qu_struct_s    qu_struct_s;
typedef struct qu_sub_s       qu_sub_s;
typedef struct qu_thread_s    qu_thread_s;
typedef struct qu_tokenizer_s qu_tokenizer_s;
typedef struct qu_var_s       qu_var_s;
typedef struct qu_weak_s      qu_weak_s;

/* abstracts - to enforce unique pointer types */
typedef struct qu_decimal_s  qu_decimal_s;
typedef struct qu_function_s qu_function_s;
typedef struct qu_int_s      qu_int_s;
typedef struct qu_integer_s  qu_integer_s;
typedef struct qu_inf_s      qu_inf_s;
typedef struct qu_null_s     qu_null_s;
typedef struct qu_number_s   qu_number_s;
typedef struct qu_real_s     qu_real_s;

typedef void            *qu_any;       /* any object including qu_int */
typedef qu_dict_s       *qu_arg;
typedef qu_array_s      *qu_array;
typedef qu_ast_s        *qu_ast;
typedef qu_big_s        *qu_big;
typedef qu_object_s     *qu_boolean;
typedef qu_buffer_s     *qu_buffer;
typedef qu_class_s      *qu_class;
typedef qu_code_s       *qu_code;
typedef qu_complex_s    *qu_complex;
typedef qu_container_s  *qu_container; /* mixin */
typedef qu_object_s     *qu_decimal;
typedef qu_dict_s       *qu_dict;
typedef qu_dlib_s       *qu_dlib;
typedef qu_exception_s  *qu_exception;
typedef qu_file_s       *qu_file;
typedef qu_float_s      *qu_float;
typedef qu_frame_s      *qu_frame;
typedef qu_function_s   *qu_function;  /* abstract */
typedef qu_int_s        *qu_int;
typedef qu_integer_s    *qu_integer;   /* abstract */
typedef qu_inf_s        *qu_inf;
typedef qu_iterator_s   *qu_iterator;
typedef qu_long_s       *qu_long;
typedef qu_method_s     *qu_method;
typedef qu_module_s     *qu_module;
typedef qu_null_s       *qu_null;
typedef qu_number_s     *qu_number;    /* abstract */
typedef qu_object_s     *qu_obj;       /* any object excluding qu_int */
typedef qu_object_s     *qu_object;    /* instance of Object */
typedef qu_opaque_s     *qu_opaque;
typedef qu_proc_s       *qu_proc;
typedef qu_pointer_s    *qu_pointer;
typedef qu_queue_s      *qu_queue;
typedef qu_range_s      *qu_range;
typedef qu_rational_s   *qu_rational;
typedef qu_real_s       *qu_real;      /* abstract */
typedef qu_record_s     *qu_record;
typedef qu_set_s        *qu_set;
typedef qu_stream_s     *qu_stream;
typedef qu_string_s     *qu_string;
typedef qu_struct_s     *qu_struct;
typedef qu_sub_s        *qu_sub;
typedef qu_thread_s     *qu_thread;
typedef qu_tokenizer_s  *qu_tokenizer;
typedef qu_var_s        *qu_var;
typedef qu_weak_s       *qu_weak;

struct qu_object_s
{
	QU_OBJECT_HEAD;
};



/** Object
	This is the U{not} base ={Class} of all classes.

	The Object class simply provides default methods such as __call, __inst,
	__cmp, etc. See the program documentation for details.
*/
extern qu_class QuObjectClass;


/** QU_FLAG
	List of flags. The first four are strictly for use by the GC.
	The others may be used by classes.

	Observe that some classes share the same of the flags.

	@warn
	Flags may not be used with ={Int} objects, of course.
*/

/* used by the GC */
#define QU_FLAG_MARK    0x1000000  /* GC: marked? */
#define QU_FLAG_FIN     0x2000000  /* GC: finalizer already called? */
#define QU_FLAG_DONTGC  0x4000000  /* GC: dont GC? */
#define QU_FLAG_STATIC  0x8000000  /* GC: never GC'ed */

/* usable by classes */
#define QU_FLAG_0   0x1
#define QU_FLAG_1   0x2
#define QU_FLAG_2   0x4
#define QU_FLAG_3   0x8
#define QU_FLAG_4   0x10
#define QU_FLAG_5   0x20
#define QU_FLAG_6   0x40
#define QU_FLAG_7   0x80
#define QU_FLAG_8   0x100
#define QU_FLAG_9   0x200
#define QU_FLAG_10  0x400
#define QU_FLAG_11  0x800
#define QU_FLAG_12  0x1000
#define QU_FLAG_13  0x2000
#define QU_FLAG_14  0x4000
#define QU_FLAG_15  0x8000
#define QU_FLAG_16  0x10000
#define QU_FLAG_17  0x20000
#define QU_FLAG_18  0x40000
#define QU_FLAG_19  0x80000
#define QU_FLAG_20  0x100000
#define QU_FLAG_21  0x200000
#define QU_FLAG_22  0x400000
#define QU_FLAG_23  0x800000


/** qu_flag_test qu_flag_set qu_flag_unset
	These are the 3 macros to test, set and unset a flag.
	See ={QU_FLAG}.

	You should generally not use these macros and use the flag accessors
	provided by the corresponding classes. That is, ideally these macros should
	only be used within classes.
*/
#define qu_flag_test(o, f)   ((((qu_obj)(o))->flags & (f)) != 0)
#define qu_flag_set(o, f)    QU_BEGIN ((qu_obj)(o))->flags |= (f); QU_END
#define qu_flag_unset(o, f)  QU_BEGIN ((qu_obj)(o))->flags &= ~(f); QU_END


/** QU_OBJECT_MAXIDENTLEN
	The maximum number of characters in any identifiers.

	This limit is used because in many cases we can use the C stack
	to build strings and save memory allocation.
*/
#define QU_OBJECT_MAXIDENTLEN 64


/** qu_destroy_f qu_mark_f
	Low-level instance routine prototypes.

	See ={Class}.

	Notice the argument type is qu_obj because it must not be an ={Int}.

	qu_destroy_f is the prototype to free memory allocated by an object.
	It must not free the v{attrs} field nor the object itself since that
	is taken care by the GC.

	qu_mark_f is the prototype to mark objects referenced by an object.
*/
typedef void (*qu_destroy_f) (qu_obj self);
typedef void (*qu_mark_f   ) (qu_obj self);


/** qu_unop_f qu_binop_f qu_triop_f quartop_f
	C method operator prototypes.
*/
typedef qu_any (*qu_unop_f   ) (qu_any self);
typedef qu_any (*qu_binop_f  ) (qu_any self, qu_any x);
typedef qu_any (*qu_triop_f  ) (qu_any self, qu_any x, qu_any y);
typedef qu_any (*qu_quartop_f) (qu_any self, qu_any x, qu_any y, qu_any z);


/** qu_subn_f qu_sub3_f qu_sub2_f qu_sub1_f qu_sub0_f qu_subcn_f qu_subc3_f qu_subc2_f qu_subc1_f qu_subc0_f
	C method and function prototypes.
*/
typedef qu_any (*qu_subn_f) (qu_any self, int argc, qu_any *argv);
typedef qu_any (*qu_sub4_f) (qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
typedef qu_any (*qu_sub3_f) (qu_any self, qu_any a1, qu_any a2, qu_any a3);
typedef qu_any (*qu_sub2_f) (qu_any self, qu_any a1, qu_any a2);
typedef qu_any (*qu_sub1_f) (qu_any self, qu_any a1);
typedef qu_any (*qu_sub0_f) (qu_any self);

typedef qu_any (*qu_subcn_f) (int argc, qu_any *argv);
typedef qu_any (*qu_subc4_f) (qu_any a1, qu_any a2, qu_any a3, qu_any a4);
typedef qu_any (*qu_subc3_f) (qu_any a1, qu_any a2, qu_any a3);
typedef qu_any (*qu_subc2_f) (qu_any a1, qu_any a2);
typedef qu_any (*qu_subc1_f) (qu_any a1);
typedef qu_any (*qu_subc0_f) ();


/***/
/*----------------------------------------------------------------------------
	Symbol = Module, Class, Sub, Var
----------------------------------------------------------------------------*/

/** Symbol QU_SYMBOL_HEAD qu_symbol_s qu_symbol
	={Module}, ={Class}, ={Sub}, and ={Var} must use QU_SYMBOL_HEAD instead
	of ={QU_OBJECT_HEAD}.
*/
#define QU_SYMBOL_HEAD \
	QU_OBJECT_HEAD; \
	qu_string name;   /* symbol name */ \
	qu_symbol parent  /* parent symbol */


typedef struct qu_symbol_s  qu_symbol_s;
typedef struct qu_symbol_s *qu_symbol;

struct qu_symbol_s
{
	QU_SYMBOL_HEAD;
};


/** QU_SYMBOL_FLAG
	Flags shared by Module, Class, Sub, and Var.
*/
#define QU_SYMBOL_FLAG_EXTERN    QU_FLAG_0  /* Class/Sub: C (not program level) */
#define QU_SYMBOL_FLAG_ABSTRACT  QU_FLAG_1  /* Class/Sub: abstract */
#define QU_SYMBOL_FLAG_PRIVATE   QU_FLAG_2  /* Class/Sub/Var : visible to block & parent */
#define QU_SYMBOL_FLAG_FINAL     QU_FLAG_3  /* Class/Sub/Var: cant override */
#define QU_SYMBOL_FLAG_LOCAL     QU_FLAG_4  /* Sub/Var: visible to block only */
#define QU_SYMBOL_FLAG_STATIC    QU_FLAG_5  /* Sub: is static (self can be anything) */
#define QU_SYMBOL_FLAG_CLASS     QU_FLAG_6  /* Sub/Var: is class attribute */
#define QU_SYMBOL_FLAG_INST      QU_FLAG_7  /* Sub/Var: is instance attribute */
#define QU_SYMBOL_FLAG_DOC       QU_FLAG_8  /* Module/Class: compiled for documentation? */

#define qu_symbol_is_extern(x)   qu_flag_test (x, QU_SYMBOL_FLAG_EXTERN)
#define qu_symbol_is_abstract(x) qu_flag_test (x, QU_SYMBOL_FLAG_ABSTRACT)
#define qu_symbol_is_private(x)  qu_flag_test (x, QU_SYMBOL_FLAG_PRIVATE)
#define qu_symbol_is_final(x)    qu_flag_test (x, QU_SYMBOL_FLAG_FINAL)
#define qu_symbol_is_local(x)    qu_flag_test (x, QU_SYMBOL_FLAG_LOCAL)
#define qu_symbol_is_static(x)   qu_flag_test (x, QU_SYMBOL_FLAG_STATIC)
#define qu_symbol_is_class(x)    qu_flag_test (x, QU_SYMBOL_FLAG_CLASS)
#define qu_symbol_is_inst(x)     qu_flag_test (x, QU_SYMBOL_FLAG_INST)
#define qu_symbol_is_doc(x)      qu_flag_test (x, QU_SYMBOL_FLAG_DOC)


/** qu_symbol_is
	Whether an object is a symbol.
*/
#define qu_symbol_is(self) qu_class_is_symbol (qu_classof (self))


/** qu_symbol_get_module qu_symbol_get_class
	Return the ={Module} or ={Class} object where a symbol is declared.
	Return 0 if none.
*/
extern qu_module qu_symbol_get_module (qu_symbol self);
extern qu_class  qu_symbol_get_class  (qu_symbol self);


/** qu_symbol_find
	Find symbol in block, then parent blocks, then the builtin Sys module.
*/
extern qu_symbol qu_symbol_find (qu_symbol from, qu_string name);


/***/
/*----------------------------------------------------------------------------
	Object
----------------------------------------------------------------------------*/

/** qu_object_as
	Convenient function to be called from f{new} and __as methods.
*/
extern qu_any qu_object_as (qu_any self, qu_any x);


/** qu_object_hash
	Return the hash value of an object.
*/
extern long qu_object_hash (qu_any self);


/** qu_object_method
	Methods that always return self, false, true, etc. can just point to these methods
	to save code.

	Note that methods that throw an exception must always be provided since Qu
	uses the method name to report exception messages.
*/
extern qu_any     qu_object_method_self  (qu_any self);
extern qu_boolean qu_object_method_true  (qu_any self);
extern qu_boolean qu_object_method_false (qu_any self);
extern qu_null    qu_object_method_nil   (qu_any self);
extern qu_int     qu_object_method_int0  (qu_any self);


/** qu_object_defines
	Whether or not an object defines a specified attribute.

	This function is provided for efficiency over C{Object.__defines}.
*/
extern int qu_object_defines (qu_any self, qu_string name);


/** qu_object_iterable
	Whether an object is iterable.
*/
extern int qu_object_iterable (qu_any self);


/** qu_object_callable
	Whether the object is a callable: Sub, Method or Class.
*/
extern int qu_object_callable (qu_any self);


/** qu_object_call
	qu_object_call call a method by name where v{name} must be a ={String}.

	qu_object_call_x automatically make the call base on what v{self} is.
	v{self} may be a Sub, Method or Class, and not be an instance method.
	Otherwise QuECall is thrown.

	qu_object_call_ex is a flexible call.
	v{how} may be a String, Sub, Method or Class.
	If v{how} is as String then call C{self.(how)(*arg)}
	otherwise call C{how (self, *arg)} setting self as ={qu_interp_self}.
*/
extern qu_any qu_object_calln   (qu_any self, qu_string name, int argc, qu_any *argv);
extern qu_any qu_object_call4   (qu_any self, qu_string name, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_object_call3   (qu_any self, qu_string name, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_object_call2   (qu_any self, qu_string name, qu_any a1, qu_any a2);
extern qu_any qu_object_call1   (qu_any self, qu_string name, qu_any a1);
extern qu_any qu_object_call0   (qu_any self, qu_string name);

extern qu_any qu_object_call_xn  (qu_any self, int argc, qu_any *argv);
extern qu_any qu_object_call_x4  (qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_object_call_x3  (qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_object_call_x2  (qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_object_call_x1  (qu_any self, qu_any a1);
extern qu_any qu_object_call_x0  (qu_any self);

extern qu_any qu_object_call_exn (qu_any self, qu_any how, int argc, qu_any *argv);
extern qu_any qu_object_call_ex3 (qu_any self, qu_any how, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_object_call_ex2 (qu_any self, qu_any how, qu_any a1, qu_any a2);
extern qu_any qu_object_call_ex1 (qu_any self, qu_any how, qu_any a1);
extern qu_any qu_object_call_ex0 (qu_any self, qu_any how);


/** qu_object_rank
	Return the rank of an object.

	The value must not be taken literally as it may change in the future.
	The only purpose of this function is for sorting.
*/
extern int qu_object_rank (qu_any self);


/** qu_object_isa qu_object_instof
	qu_object_isa is the "is" operator. It works for any object including classes.
	Both v{self} and v{x} may be anything. If they are not classes then ={qu_classof}
	is used. Be careful, this is a very flexible operator.

	@xmp
		"foo"     is String -> true
		qu_int    is Number -> true
		qu_string is Class  -> true

	@p
	qu_object_instof checks whether self is an instance of Class v{x} or a class
	derived from Class v{x}.

	@xmp
		"foo"    instof  String -> true
		Int      instof  Number -> false
		String   instof  ClassClass  -> true
*/
extern int qu_object_isa (qu_any self, qu_any x);
#define qu_object_instof(o, klass) qu_class_isa (qu_classof(o), klass)


/** qu_object_specific
	For C modules, to create or delete instance specific attribute.

	We do not directly support specific attributes in program level for a number
	of reasons (security, documentability, etc.). C classes will have to implement
	it themselves as necessary using these functions.

	Instance specifics must be avoided if possible. Generally you would only use
	them if you know there are not a lot of instances that will be created.

	@warn
	The class must have the QU_CLASS_FLAG_DYNAMIC flag set otherwise a ={QU_FATAL} error
	is issued.

	Class objects may not have specific attributes.
	v{self} may not be a class otherwise a ={QU_FATAL} error is issued.

	Never assign methods from foreign classes. You risk the danger of calling a
	C method with the wrong instance/class.

	v{func} must be a C function address.
*/
extern qu_any  qu_object_specific_prop   (qu_obj self, qu_string name, qu_any value);
extern qu_sub  qu_object_specific_method (qu_obj self, qu_string name, void *func, int argc, long flags);
extern qu_null qu_object_specific_delete (qu_obj self, qu_string name);


/** qu_object_where_string
	Use to print symbols (={Module}, ={Class}, ={Sub} and ={Var}) so we have
	a consistent form.
*/
extern qu_string qu_object_where_string (qu_any sym, const char *prefix, const char *suffix);


/** qu_object_validator
	These are I{validator} routines.

	qu_object_valid_type return 1 if v{x} is a valid validator, 1 otherwise.
	Valid validators: Class, Struct, Range, Array, Set, Dict, static Sub,
	and the special ={qu_var_is_validator} variable.

	qu_object_fix_type fixes a compile-time generated type which is still in the form
	of a Var into one of qu_var_is_validator, Struct or Dict (enum).

	qu_object_compat_type return 1 if v{othertype} is a compatible validator for
	v{type}. For example, if both are frozen Arrays containing the same objects
	then they are compatible. If v{othertype} is a subclass of v{type} then
	return 1 but not the other way around.

	qu_object_check_value is designed for compile-time. It validates the value but does
	not call the Sub if k{type} is a Sub.

	qu_object_validate is designed for run-time. It validates the value and call Sub
	validators as necessary.
*/
extern int     qu_object_valid_type  (qu_any x);
extern qu_any  qu_object_fix_type    (qu_any x);
extern int     qu_object_compat_type (qu_any type, qu_any othertype);
extern int     qu_object_check_value (qu_any type, qu_any value);
extern qu_any  qu_object_validate    (qu_any self, qu_any type, qu_any value);


/** qu_object_default_string
	Return a default representation String of an object.

	@warn
	This does not call the __string method.
	It returns a default value.
*/
#define qu_object_default_string(self) qu__object__string (self)


/** qu_object_default_true
	This method always return QuTrue.
*/
#define qu_object_default_true(self) QuTrue


/** qu_object_default_cmp
	Return default comparisson value of two objects.

	qu_object_default_cmpi return -1, 0, and 1.Not just arbitrary integer.

	qu_object_default_cmp does the same but returns an ={Int}.

	@warn
	This does not call the __cmp method.
	It returns a default value.
*/
extern int qu_object_default_cmpi (qu_any self, qu_any x);
#define qu_object_default_cmp(self, x) qu_int_new (qu_object_default_cmpi (self, x))


/** qu_object_default_eq
	Whether self and x are the very same object.

	@warn
	This does not call the __eq method.
	It returns a default value which is either QuTrue or QuFalse.
*/
#define qu_object_default_eq(self, x) qu__object__eq (self, x)


/** qu_object_default_print
	Default __print method.
*/
#define qu_object_default_print(self) qu__object__print (self)


/** qu_object_get_module
	Get the ={Module} where the object came from.
*/
#define qu_object_get_module(self) qu__object__module (self)


/** qu_object_class_name
	Get the class name of an object.
*/
#define qu_object_class_name(self) qu__object__name (self)


/** qu_object_getattr qu_object_setattr
	Get or set attribute at run-time.

	These look for attributes in all superclasses and mixin classes.

	Prefer these over using the v{get} and v{set} of the qu_class_s struct
	directly.
*/
#define qu_object_getattr(self, name, restrict) \
	qu_classof (self)->get (self, name, restrict)

#define qu_object_setattr(self, name, value, restrict) \
	qu_classof (self)->set (self, name, value, restrict)


/** qu_object_to_static
	Convert an object into a static (one which is never garbage collected
	and never marked).

	v{x} must be an instance of the following classes:
	@ul
	.String
	.Null
	.Int
	.Boolean
	.Inf
	.Long
	.Rational
	.Float
	.Big
	.Range
	.Array
	.Dict
	.Set
*/
#define qu_object_to_static(self) qu__object__static (self)


/** QU_OBJECT_LIST_COPY QU_OBJECT_LIST_RCOPY QU_OBJECT_LIST_SETVAL
	Some handy tools.

	QU_OBJECT_LIST_COPY should be faster than f{memcpy}. If not then it should be
	translated into a macro.
*/
inline static void QU_OBJECT_LIST_COPY (register qu_any *dst, register qu_any *src, register long n)
{
	while (n--)
		*dst++ = *src++;
}

inline static void QU_OBJECT_LIST_RCOPY (register qu_any *dst, register qu_any *src, register long n)
{
	dst += n;
	src += n;
	while (n--)
		*--dst = *--src;
}

inline static void QU_OBJECT_LIST_SETVAL (register qu_any *dst, register qu_any v, register long n)
{
	while (n--)
		*dst++ = v;
}


/***/
/*--------------------------------------------------------------------------*/

/* internally called by class to initialize this class */
extern void qu_object__setup ();

/* for efficiency - may only be called from within this file */
extern qu_module  qu__object__module (qu_any self);
extern qu_string  qu__object__name   (qu_any self);
extern qu_boolean qu__object__eq     (qu_any self, qu_any x);
extern qu_null    qu__object__print  (qu_any self);
extern qu_string  qu__object__string (qu_any self);
extern qu_any     qu__object__static (qu_any self);

/* for Emit - this is the Object.new method */
extern qu_sub __QuObjectNew;

#endif /*_H_qu_object_*/
