/* Queue.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_queue_
#define _H_qu_queue_ 1

#include "Container.h"


/** Queue
	The Queue class.
*/
extern qu_class QuQueueClass;


/** qu_queue
	qu_queue represents an instance of the Queue class.
	It is a pointer to the instance struct qu_queue_s.

	See also ={QU_CONTAINER_HEAD}.
*/
typedef struct qu_queueitem_s qu_queueitem_s;
struct qu_queueitem_s
{
	qu_queueitem_s *next;   /* next item */
	qu_any          obj;    /* the object */
};

struct qu_queue_s
{
	QU_CONTAINER_HEAD;
	long            size;   /* max items */
	qu_queueitem_s *first;  /* first item */
	qu_queueitem_s *last;   /* last item */
};


/** qu_queue_is
	Whether the object is a Queue.
*/
#define qu_queue_is(o) qu_object_instof (o, QuQueueClass)


/** QU_QUEUE_FLAG
	Flags used by Queue objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_QUEUE_FLAG_PROTECT QU_CONTAINER_FLAG_PROTECT /* protected by program level */
#define QU_QUEUE_FLAG_FROZEN  QU_CONTAINER_FLAG_FROZEN  /* immutable? */
#define QU_QUEUE_FLAG_FIFO    QU_FLAG_10                /* first in first out */
#define QU_QUEUE_FLAG_LIFO    QU_FLAG_11                /* last in last out */


/** qu_queue_frozen
	qu_queue_frozen checks whether or not the Queue is mutable.

	qu_queue_freeze must only be used after you have created a new Queue and
	don't want the Queue to be modified by others.

	qu_queue_unfreeze must only be used on an Queue you have created.
	Normally, to modify the Queue then freeze it again.
*/
#define qu_queue_frozen(self)    qu_container_frozen   (self)
#define qu_queue_freeze(self)    qu_container_freeze   (self)
#define qu_queue_unfreeze(self)  qu_container_unfreeze (self)


/** qu_queue_is_fifo
	Whether this is a FIFO queue (LIFO otherwise).

	You may only query the QU_QUEUE_FLAG_LIFO and QU_QUEUE_FLAG_FIFO flags.
	Setting/unsetting may only be done by Queue.
*/
#define qu_queue_is_fifo(self) qu_flag_test ((self), QU_QUEUE_FLAG_FIFO)


/** qu_queue_new
	Create a new Queue.

	Use the macros instead of calling qu_queue_new directly.
*/
extern qu_queue qu_queue_new (long flags, long size);
#define qu_queue_new_fifo(size) qu_queue_new (QU_QUEUE_FLAG_FIFO, size)
#define qu_queue_new_lifo(size) qu_queue_new (QU_QUEUE_FLAG_LIFO, size)


/** qu_queue_clear
	Remove everything from the Queue.

	Throws if the Queue is immutable.
*/
#define qu_queue_clear(self) qu__queue__clear (self)


/** qu_queue_copy
	Make a shallow copy.

	This function always succeed.
*/
#define qu_queue_copy(self) qu__queue__copy (self)


/** qu_queue_pop
	Pop item from the Queue.

	Throws if the Queue is empty or immutable.
*/
#define qu_queue_pop(self) qu__queue__pop (self)


/** qu_queue_push
	Append an object into the Queue.

	Throws if the hard limit is reached or the Queue is immutable.
*/
#define qu_queue_push(self, x) qu__queue__push (self, x)


/** qu_queue_putback
	Put back an object into the Queue, no questions asked.

	Always succeed except if the hard limit is reached.
*/
#define qu_queue_putback(self, x) qu__queue_putback (self, x)


/** qu_queue_rotate
	Rotate the items.

	This function always succeed and return self.
*/
#define qu_queue_rotate(self) qu__queue_rotate(self)


/** qu_queue_true
	The truth value.
*/
#define qu_queue_true(self) (qu_queue_as_boolean (self) = QuTrue)


/** qu_queue_as
	Various representations.
*/
extern qu_array qu_queue_as_array (qu_queue self);
#define qu_queue_as_string(self)  qu__queue__string (self)
#define qu_queue_as_boolean(self) ((self)->len ? QuTrue : QuFalse)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_queue__setup ();
extern void qu_queue__init  ();

/* for efficiency - may only be called from within this file */
extern qu_queue   qu__queue__clear  (qu_queue self);
extern qu_queue   qu__queue__copy   (qu_queue self);
extern qu_any     qu__queue__pop    (qu_queue self);
extern qu_queue   qu__queue__push   (qu_queue self, qu_any x);
extern qu_string  qu__queue__string (qu_queue self);
extern qu_queue   qu__queue_rotate  (qu_queue self);
extern qu_queue   qu__queue_putback (qu_queue self, qu_any x);

#endif /*_H_qu_queue_*/
