/* Rational.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_rational_
#define _H_qu_rational_ 1


/** Rational
	This is the final Rational class.
*/
extern qu_class QuRationalClass;


/** qu_rational
	qu_rational represents an instance of the Rational class.
	It is a pointer to the instance struct qu_rational_s.

	The denominator is never 0 and the sign is carried by the numerator.
*/
struct qu_rational_s
{
	QU_OBJECT_HEAD;
	qu_integer num;  /* numerator - carries the sign */
	qu_integer den;  /* denominator */
};


/** QuRationalN
	Predefined system Rational objects.
*/
extern qu_rational QuRational0;  /*  0 / 1 */
extern qu_rational QuRational1;  /*  1 / 1 */
extern qu_rational QuRational_1; /* -1 / 1 */


/** qu_rational_new
	Create a new Rational object.

	You may not pass 0 to v{den} otherwise EValue is thrown and NULL is returned.

	qu_rational_new_ex returns an ={Integer} if the numerator is an exact multiplication
	of the denominator, a ={Rational} otherwise.
*/
extern qu_rational qu_rational_newl            (long num, long den);
extern qu_rational qu_rational_newr            (qu_integer num, qu_integer den);
extern qu_rational qu_rational_new_num         (qu_integer num);
extern qu_rational qu_rational_new_den         (qu_integer den);
extern qu_real     qu_rational_new_ex          (qu_integer num, qu_integer den);
extern qu_rational qu_rational_newd            (double v);
#define            qu_rational_new_int(x)      qu_rational_new_num ((qu_integer) x)
#define            qu_rational_new_long(x)     qu_rational_new_num ((qu_integer) x)
#define            qu_rational_new_rational(x) (x)
#define            qu_rational_new_float(x)    qu_rational_newd ((x)->value)
#define            qu_rational_new_big(x)      qu_big_as_rational (x)
#define            qu_rational_new_complex     qu_complex_as_rational (x)
#define            qu_rational_new_integer(x)  qu_rational_new_num (x)
#define            qu_rational_new_decimal(x)  qu_decimal_as_rational (x)
#define            qu_rational_new_real(x)     qu_real_as_rational (x)
#define            qu_rational_new_number(x)   qu_number_as_rational (x)
#define            qu_rational_new_boolean(x)  (((x) == QuTrue) ? QuRational1 : QuRational0)
extern qu_rational qu_rational_new_string      (qu_string x);


/** qu_rational_new_static
	Create a system static Rational object.

	@warn
	The denominator may not be zero. No checks are made!

	@p
	qu_rational_to_static creates one from an existing object.
*/
extern qu_rational qu_rational_new_static  (qu_integer n, qu_integer d);
extern qu_rational qu_rational_new_staticl (long n, long d);
extern qu_rational qu_rational_to_static   (qu_rational x);


/** qu_rational_to
	Conversions.

	On success errno is set to 0, non-zero otherwise.
*/
#define                   qu_rational_toh(x)  qu_number__ltoh   (qu_rational_tol (x))
#define                   qu_rational_touh(x) qu_number__ultouh (qu_rational_toul (x))
#define                   qu_rational_toi(x)  qu_number__ltoi   (qu_rational_tol (x))
#define                   qu_rational_toui(x) qu_number__ultoui (qu_rational_toul (x))
#define                   qu_rational_toli(x) qu_number__ltoli  (qu_rational_tol (x))
extern long               qu_rational_tol     (qu_rational x);
extern unsigned long      qu_rational_toul    (qu_rational x);
extern float              qu_rational_tof     (qu_rational x);
extern double             qu_rational_tod     (qu_rational x);
#ifdef HAVE_LONG_LONG
extern long long          qu_rational_toll    (qu_rational x);
extern unsigned long long qu_rational_toull   (qu_rational x);
#endif
#ifdef HAVE_LONG_DOUBLE
extern long double        qu_rational_told    (qu_rational x);
#endif


/** qu_rational_as
	Conversion to another object.

	NULL is returned and an excpetion thrown on failure.
	Only qu_rational_as_int and qu_rational_as_float can fail.
	The others always succeed.
*/
extern qu_int     qu_rational_as_int          (qu_rational x);
extern qu_long    qu_rational_as_long         (qu_rational x);
#define           qu_rational_as_rational(x)  (x)
#define           qu_rational_as_float(x)     qu_float_new_rational (x)
#define           qu_rational_as_big(x)       qu_big_new_rational(x)
extern qu_integer qu_rational_as_integer      (qu_rational x);
extern qu_decimal qu_rational_as_decimal      (qu_rational x);
#define           qu_rational_as_real(x)      ((qu_real) (x))
#define           qu_rational_as_number(x)    ((qu_number) (x))
#define           qu_rational_as_complex(x)   qu_complex_new1 ((qu_real) (x))
#define           qu_rational_as_string(x)    qu__rational__string (x)
#define           qu_rational_as_boolean(x)   (qu_rational_iszero (x) ? QuFalse : QuTrue)


/** qu_rational_sign
	Returns -1, 0, 1 if the value is negative, 0, positive respectively.
*/
#define qu_rational_sign(x) qu_integer_sign ((x)->num)


/** qu_rational_is
	Identity functions.

	qu_rational_is tests whether or not an object is a Rational.

	qu_rational_isneg returnss 1 if the value is negative.
	Returns 0 otherwise.

	qu_rational_iszero return 1 if the value is zero.
	Returns 0 otherwise.

	qu_rational_isone returns 1 if the value is 1 or -1.
	Returns 0 otherwise.

	qu_rational_iswhole returns 1 if there is no fraction part.
	Returns 0 otherwise.

	qu_rational_isfrac returns 1 if 0 < |x| < 1.
	Returns 0 otherwise.

	qu_rational_isodd returns 1 if the number is whole and odd.
	Returns 0 otherwise.
*/
#define    qu_rational_is(x)      (qu_classof (x) == QuRationalClass)
#define    qu_rational_isneg(x)   qu_integer_isneg ((x)->num)
#define    qu_rational_iszero(x)  qu_integer_iszero ((x)->num)
#define    qu_rational_iswhole(x) ((x)->den == (qu_integer) QuInt1)
extern int qu_rational_isfrac     (qu_rational x);
extern int qu_rational_isone      (qu_rational x);
extern int qu_rational_is1        (qu_rational x);
extern int qu_rational_isodd      (qu_rational x);


/** qu_rational_unary
	Unary functions.

	qu_rational_pos returns the number itself, obviously.

	qu_rational_neg returns a negated value.
	The result is a qu_rational and always succeed.

	qu_rational_incr returns the qu_rational incremented by 1.
	The result is a qu_rational or NULL on failure.

	qu_rational_decr returns the qu_rational decremented by 1.
	The result is a qu_rational or NULL on failure.
*/
#define qu_rational_pos(x)  (x)
#define qu_rational_neg(x)  qu__rational__neg  (x)
#define qu_rational_incr(x) qu__rational__incr (x)
#define qu_rational_decr(x) qu__rational__decr (x)


/** qu_rational_eq
	Whether the value is the same as the value of another.
	All return 1 or 0.
*/
extern int qu_rational_eq_int           (qu_rational x, qu_int y);
extern int qu_rational_eq_long          (qu_rational x, qu_long y);
extern int qu_rational_eq_rational      (qu_rational x, qu_rational y);
#define    qu_rational_eq_float(x, y)   qu_float_eq_rational (y, x)
#define    qu_rational_eq_big(x, y)     qu_big_eq_rational (y, x)
#define    qu_rational_eq_complex(x, y) qu_complex_eq_rational (y, x)
#define    qu_rational_eq_inf(x, y)     (0)
extern int qu_rational_eq_integer       (qu_rational x, qu_integer y);
extern int qu_rational_eq_decimal       (qu_rational x, qu_decimal y);
extern int qu_rational_eq_real          (qu_rational x, qu_real y);
extern int qu_rational_eq_number        (qu_rational x, qu_number y);



/** qu_rational_like
	Whether the absolute value is the same as the absolute value of another.
	All return int 1 or 0.
*/
extern int qu_rational_like_int           (qu_rational x, qu_int y);
extern int qu_rational_like_long          (qu_rational x, qu_long y);
extern int qu_rational_like_rational      (qu_rational x, qu_rational y);
#define    qu_rational_like_float(x, y)   qu_float_like_rational (y, x)
#define    qu_rational_like_big(x, y)     qu_big_like_rational (y, x)
#define    qu_rational_like_complex(x, y) qu_complex_like_rational (y, x)
#define    qu_rational_like_inf(x, y)     (0)
extern int qu_rational_like_integer       (qu_rational x, qu_integer y);
extern int qu_rational_like_decimal       (qu_rational x, qu_decimal y);
extern int qu_rational_like_real          (qu_rational x, qu_real y);
extern int qu_rational_like_number        (qu_rational x, qu_number y);


/** qu_rational_cmp
	Compare with another.
	All return -1, 0, 1 in the usual manner.
*/
extern int qu_rational_cmp_int           (qu_rational x, qu_int y);
extern int qu_rational_cmp_long          (qu_rational x, qu_long y);
extern int qu_rational_cmp_rational      (qu_rational x, qu_rational y);
#define    qu_rational_cmp_float(x, y)   (-qu_float_cmp_rational (y, x))
#define    qu_rational_cmp_big(x, y)     (-qu_big_cmp_rational (y, x))
#define    qu_rational_cmp_complex(x, y) (-qu_complex_cmp_rational (y, x))
#define    qu_rational_cmp_inf(x, y)     (-qu_inf_cmp_rational (y, x))
extern int qu_rational_cmp_integer       (qu_rational x, qu_integer y);
extern int qu_rational_cmp_decimal       (qu_rational x, qu_decimal y);
extern int qu_rational_cmp_real          (qu_rational x, qu_real y);
extern int qu_rational_cmp_number        (qu_rational x, qu_number y);


/** qu_rational_cmpabs
	Compare the absolute value with the absolute value of another.
	All return -1, 0, 1 in the usual maner.
*/
extern int qu_rational_cmpabs_int           (qu_rational x, qu_int y);
extern int qu_rational_cmpabs_long          (qu_rational x, qu_long y);
extern int qu_rational_cmpabs_rational      (qu_rational x, qu_rational y);
#define    qu_rational_cmpabs_float(x, y)   (-qu_float_cmpabs_rational (y, x))
#define    qu_rational_cmpabs_big(x, y)     (-qu_big_cmpabs_rational (y, x))
#define    qu_rational_cmpabs_complex(x, y) (-qu_complex_cmpabs_rational (y, x))
#define    qu_rational_cmpabs_inf(x, y)     (-qu_inf_cmpabs_rational (y, x))
extern int qu_rational_cmpabs_integer       (qu_rational x, qu_integer y);
extern int qu_rational_cmpabs_decimal       (qu_rational x, qu_decimal y);
extern int qu_rational_cmpabs_real          (qu_rational x, qu_real y);
extern int qu_rational_cmpabs_number        (qu_rational x, qu_number y);


/** qu_rational_plus
	Add with another.
*/
extern qu_real   qu_rational_plus_int           (qu_rational x, qu_int y);
extern qu_real   qu_rational_plus_long          (qu_rational x, qu_long  y);
extern qu_real   qu_rational_plus_rational      (qu_rational x, qu_rational y);
#define          qu_rational_plus_float(x, y)   qu_float_plus_rational (y, x)
#define          qu_rational_plus_big(x, y)     qu_big_plus_rational (y, x)
#define          qu_rational_plus_complex(x, y) qu_complex_plus_rational (y, x)
#define          qu_rational_plus_inf(x, y)     qu_inf_plus_rational (y, x)
extern qu_real   qu_rational_plus_integer       (qu_rational x, qu_integer y);
extern qu_real   qu_rational_plus_decimal       (qu_rational x, qu_decimal y);
extern qu_real   qu_rational_plus_real          (qu_rational x, qu_real y);
extern qu_number qu_rational_plus_number        (qu_rational x, qu_number y);


/** qu_rational_minus
	Subtract with another.
*/
extern qu_real   qu_rational_minus_int           (qu_rational x, qu_int y);
extern qu_real   qu_rational_minus_long          (qu_rational x, qu_long y);
extern qu_real   qu_rational_minus_rational      (qu_rational x, qu_rational y);
#define          qu_rational_minus_float(x, y)   qu_float_rminus_rational (y, x)
#define          qu_rational_minus_big(x, y)     qu_big_rminus_rational (y, x)
#define          qu_rational_minus_complex(x, y) qu_complex_rminus_rational (y, x)
#define          qu_rational_minus_inf(x, y)     qu_inf_plus_rational (qu_inf_neg (y), x)
extern qu_real   qu_rational_minus_integer       (qu_rational x, qu_integer y);
extern qu_real   qu_rational_minus_decimal       (qu_rational x, qu_decimal y);
extern qu_real   qu_rational_minus_real          (qu_rational x, qu_real y);
extern qu_number qu_rational_minus_number        (qu_rational x, qu_number y);


/** qu_rational_rminus
	Subtract another with this one.
	Provided for efficiency.
*/
extern qu_real qu_rational_rminus_int   (qu_rational x, qu_int y);
extern qu_real qu_rational_rminus_long  (qu_rational x, qu_long y);


/** qu_rational_mul
	Multiply with another.
*/
extern qu_real   qu_rational_mul_int           (qu_rational x, qu_int y);
extern qu_real   qu_rational_mul_long          (qu_rational x, qu_long y);
extern qu_real   qu_rational_mul_rational      (qu_rational x, qu_rational y);
#define          qu_rational_mul_float(x, y)   qu_float_mul_rational (y, x)
#define          qu_rational_mul_big(x, y)     qu_big_mul_rational (y, x)
#define          qu_rational_mul_complex(x, y) qu_complex_mul_rational (y, x)
#define          qu_rational_mul_inf(x, y)     qu_inf_mul_rational (y, x)
extern qu_real   qu_rational_mul_integer       (qu_rational x, qu_integer y);
extern qu_real   qu_rational_mul_decimal       (qu_rational x, qu_decimal y);
extern qu_real   qu_rational_mul_real          (qu_rational x, qu_real y);
extern qu_number qu_rational_mul_number        (qu_rational x, qu_number y);


/** qu_rational_div
	Divide with another.
*/
extern qu_real   qu_rational_div_int           (qu_rational x, qu_int y);
extern qu_real   qu_rational_div_long          (qu_rational x, qu_long y);
extern qu_real   qu_rational_div_rational      (qu_rational x, qu_rational y);
#define          qu_rational_div_float(x, y)   qu_float_rdiv_rational (y, x)
#define          qu_rational_div_big(x, y)     qu_big_rdiv_rational (y, x)
#define          qu_rational_div_complex(x, y) qu_complex_rdiv_rational (y, x)
extern qu_int    qu_rational_div_inf           (qu_rational x, qu_inf y);
extern qu_real   qu_rational_div_integer       (qu_rational x, qu_integer y);
extern qu_real   qu_rational_div_decimal       (qu_rational x, qu_decimal y);
extern qu_real   qu_rational_div_real          (qu_rational x, qu_real y);
extern qu_number qu_rational_div_number        (qu_rational x, qu_number y);


/** qu_rational_rdiv
	Divide another with this one.
	Provided for efficiency.
*/
extern qu_real qu_rational_rdiv_int  (qu_rational x, qu_int y);
extern qu_real qu_rational_rdiv_long (qu_rational x, qu_long y);


/** qu_rational_idiv
	Integer division with another.
*/
extern qu_integer qu_rational_idiv_int           (qu_rational x, qu_int y);
extern qu_integer qu_rational_idiv_long          (qu_rational x, qu_long y);
extern qu_integer qu_rational_idiv_rational      (qu_rational x, qu_rational y);
#define           qu_rational_idiv_float(x, y)   qu_float_ridiv_rational (y, x)
#define           qu_rational_idiv_big(x, y)     qu_big_ridiv_rational (y, x)
#define           qu_rational_idiv_complex(x, y) qu_complex_ridiv_rational (y, x)
extern qu_integer qu_rational_idiv_inf           (qu_rational x, qu_inf y);
extern qu_integer qu_rational_idiv_integer       (qu_rational x, qu_integer y);
extern qu_integer qu_rational_idiv_decimal       (qu_rational x, qu_decimal y);
extern qu_integer qu_rational_idiv_real          (qu_rational x, qu_real y);
extern qu_integer qu_rational_idiv_number        (qu_rational x, qu_number y);


/** qu_rational_ridiv
	Integer division of another with this one.
	Provided for efficiency.
*/
extern qu_integer qu_rational_ridiv_int  (qu_rational x, qu_int y);
extern qu_integer qu_rational_ridiv_long (qu_rational x, qu_long y);


/** qu_rational_rem
	Remainder of dividing with another.
*/
extern qu_real     qu_rational_rem_int           (qu_rational x, qu_int y);
extern qu_real     qu_rational_rem_long          (qu_rational x, qu_long y);
extern qu_real     qu_rational_rem_rational      (qu_rational x, qu_rational y);
#define            qu_rational_rem_float(x, y)   qu_float_rrem_rational (y, x)
#define            qu_rational_rem_big(x, y)     qu_big_rrem_rational (y, x)
#define            qu_rational_rem_complex(x, y) qu_complex_rrem_rational (y, x)
extern qu_rational qu_rational_rem_inf           (qu_rational x, qu_inf y);
extern qu_real     qu_rational_rem_integer       (qu_rational x, qu_integer y);
extern qu_real     qu_rational_rem_decimal       (qu_rational x, qu_decimal y);
extern qu_real     qu_rational_rem_real          (qu_rational x, qu_real y);
extern qu_number   qu_rational_rem_number        (qu_rational x, qu_number y);


/** qu_rational_rrem
	Remainder of dividing another with this one.
	Provided for efficiency.
*/
extern qu_real qu_rational_rrem_int  (qu_rational x, qu_int y);
extern qu_real qu_rational_rrem_long (qu_rational x, qu_long y);


/** qu_rational_pow
	Raise to the power of another.
*/
extern qu_real   qu_rational_pow_int           (qu_rational x, qu_int y);
extern qu_real   qu_rational_pow_long          (qu_rational x, qu_long y);
extern qu_number qu_rational_pow_rational      (qu_rational x, qu_rational y);
#define          qu_rational_pow_float(x, y)   qu_float_rpow_rational (y, x)
#define          qu_rational_pow_big(x, y)     qu_big_rpow_rational (y, x)
#define          qu_rational_pow_complex(x, y) qu_complex_rpow_rational (y, x)
extern qu_real   qu_rational_pow_inf           (qu_rational x, qu_inf y);
extern qu_real   qu_rational_pow_integer       (qu_rational x, qu_integer y);
extern qu_real   qu_rational_pow_decimal       (qu_rational x, qu_decimal y);
extern qu_real   qu_rational_pow_real          (qu_rational x, qu_real y);
extern qu_number qu_rational_pow_number        (qu_rational x, qu_number y);


/** qu_rational_rpow
	Raise another to the power of this one.
	Provided for efficiency.
*/
extern qu_number qu_rational_rpow_int  (qu_rational x, qu_int y);
extern qu_number qu_rational_rpow_long (qu_rational x, qu_long y);


/** qu_rational_round
	Rounding routines.

	qu_rational_ceil rounds to the nearest integer towards positive infinity.
	Thus it is possible to get an EInf.

	qu_rational_floor rounds to the nearest integer towards negative infinity.
	Thus it is possible to get an EInf.

	qu_rational_trunc round the number towards zero, that is, it removes the
	fraction part. This function always succeed.

	qu_rational_ceil_ex, qu_rational_floor_ex, and qu_rational_trunc_ex do the
	same but return a Rational instead of an Integer.
*/
#define qu_rational_ceil(x)  qu__rational_ceil  (x)
#define qu_rational_floor(x) qu__rational_floor (x)
#define qu_rational_trunc(x) qu__rational_trunc (x);
extern qu_rational qu_rational_round (qu_rational x, int prec);


/** qu_rational_frac
	Return the fraction part.
	The result is exact and always succeed.
*/
#define qu_rational_frac(x) qu__rational_frac (x)


/** qu_rational_format
	Format a Rational.

	See ={Format} for details.

	This function is provided for efficiency between core modules.
	It is not intended to be called by users directly.
*/
extern qu_string qu_rational_format (qu_rational x, qu_format_s *f);


/** qu_rational_inv
	Return a new Rational by inverting the numerator and denominator.

	If the numerator is zero then EArgValue is thrown.
*/
#define qu_rational_inv(self) qu__rational_inv (self)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_rational__setup ();
extern void qu_rational__init  ();

/* for efficiency - may only be called from within this file */
extern qu_rational qu__rational__neg    (qu_rational self);
extern qu_real     qu__rational__incr   (qu_rational self);
extern qu_real     qu__rational__decr   (qu_rational self);
extern qu_string   qu__rational__string (qu_rational self);
extern qu_rational qu__rational_ceil    (qu_rational self);
extern qu_rational qu__rational_floor   (qu_rational self);
extern qu_rational qu__rational_trunc   (qu_rational self);
extern qu_rational qu__rational_frac    (qu_rational self);
extern qu_rational qu__rational_inv     (qu_rational self);

/* for interp and Sys */
extern qu_rational qu_rational_new_ratio (qu_any x, qu_any y);

#endif /*_H_qu_rational_*/
