/* Set.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_set_
#define _H_qu_set_ 1

#include "Container.h"


/** Set
	The Set class.
*/
extern qu_class QuSetClass;


/** qu_set
	qu_set represents an instance of the Set class.
	It is a pointer to the instance struct qu_set_s.

	The v{qu_setentry_s} struct is intentionally hidden.

	See also ={QU_CONTAINER_HEAD}.
*/
typedef struct qu_setentry_s qu_setentry_s;
struct qu_set_s
{
	QU_CONTAINER_HEAD;
	long            mask;    /* current mask */
	qu_any          type;    /* type/validator, maybe 0 */
	qu_setentry_s **buckets; /* the buckets */
};


/** QuSetEMPTY
	This is a predefined system object frozen empty Set.

	Expressions C{const {}} in programs simply return this object.
*/
extern qu_set QuSetEMPTY;


/** qu_set_is
	Whether or not the objet is a Set.
*/
#define qu_set_is(o) qu_object_instof (o, QuSetClass)


/** QU_SET_FLAG
	Flags used by Set objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_SET_FLAG_PROTECT QU_CONTAINER_FLAG_PROTECT /* program level protected? */
#define QU_SET_FLAG_FROZEN  QU_CONTAINER_FLAG_FROZEN  /* mutable? */
#define QU_SET_FLAG_ENUM    QU_FLAG_10                /* built from enum statement? */


/** qu_set_frozen
	Whether or not the Set is mutable.

	qu_set_freeze must only be used after you have created a new Set and
	don't want the Set to be modified by others.

	qu_set_unfreeze must only be used on an Set you have created.
	Normally, to modify the Set then freeze it again.
*/
#define qu_set_frozen(self)    qu_container_frozen (self)
#define qu_set_freeze(self)    qu_container_freeze   (self)
#define qu_set_unfreeze(self)  qu_container_unfreeze (self)


/** qu_set_new
	Create a new Set with an initial capacity.

	qu_set_new creates a regular GC Set object.

	qu_set_new_root creates a root Set object always marked by the GC.

	qu_set_new_static creates a static Set object never marked by the GC,
	thus you must make sure that all of the items are also static objects.
	No checks are made.

	@warn
	Make sure v{len} <= #{QU_GC_MAX_ITEMS}.
*/
extern qu_set qu_set_new        (long len);
extern qu_set qu_set_new_root   (long len);
extern qu_set qu_set_new_static (long len);


/** qu_set_to_static
	Create a Set which is never garbage collected nor marked from an
	existing one. All items must satisfy ={qu_gc_is_static} otherwise
	EValue is thrown.
*/
extern qu_set qu_set_to_static (qu_set self);


/** qu_set_add qu_set_addif qu_set_addz qu_set_add_float
	Include an object in the Set.

	qu_set_add obediently add the object. It is the responsibility of the caller
	to make sure that the object is not in the Set. Be careful. You must have
	good knowledge of the object's __eq method behavior.

	qu_set_addif returns the same object in the Set if one exist otherwise add
	the object and return that object.

	qu_set_addif_ex does the same as qu_set_addif but the object's class is compared
	for equality. This allows you to store, e.g. a Big and a Float with the same
	value as two different items.

	qu_set_addz is for efficiency. If the same string is already in the Set
	then it is returned, created otherwise. This may save creating a ={String}
	object.

	qu_set_add_float does the same for a ={Float} object.

	@warn
	All of these functions return 0 if the set becomes too big.
*/
extern qu_any    qu_set_add       (qu_set self, qu_any o);
extern qu_any    qu_set_addif     (qu_set self, qu_any o);
extern qu_any    qu_set_addif_ex  (qu_set self, qu_any o);
extern qu_string qu_set_addz      (qu_set self, const char *z, long len);
extern qu_float  qu_set_add_float (qu_set self, double v);


/** qu_set_walk
	Walk through the objects.

	On first call v{b} and v{i} must be set to 0.

	@xmp
		qu_any x;
		long   b, i;

		b = i = 0;
		while ((x = qu_set_walk (MyFooSet, &b, &i)) != 0)
		{
			... do something with x ...
		}
*/
extern qu_any qu_set_walk (qu_set self, long *b, long *i);


/** qu_set_contains
	Whether the qu_set contains a specified object.
*/
extern int qu_set_contains (qu_set self, qu_any x);


/** qu_set_of
	qu_set type of items.
*/
#define qu_set_of(self, type) qu__set__of (self, type)


/** qu_set_push
	Push an item.

	Returns self on success, returns 0 and throws on error.
*/
#define qu_set_push(self, type) qu__set__push (self, type)


/** qu_set_del qu_set_del_any
	Remove an object from the Set.

	qu_set_del throws EModify if the Set is immutable.

	qu_set_del_any remove any arbitrary item. If the Set is empty then
	nothing happens. This function does not care whether the Set is immutable.

	Returns self on success, returns 0 and throws on error.
*/
#define qu_set_del(self, x) qu__set__del (self, x)
extern void qu_set_del_any (qu_set self);


/** qu_set_clear
	Clear the qu_set.

	Returns self on success, returns 0 and throws on error.
*/
#define qu_set_clear(self) qu__set__clear (self)


/** qu_set_exchange
	Exchange the items in the two Sets.

	@warn
	This function does not care whether the Sets are mutable.
	Use with care.
*/
extern void qu_set_exchange (qu_set self, qu_set x);


/** qu_set_like
	Whether two sets are alike.
*/
#define qu_set_like(self, x) (qu__set__like (self, x) == QuTrue)


/** qu_set_as
	Various representation.
*/
extern qu_array qu_set_as_array (qu_set self);
extern qu_dict  qu_set_as_dict  (qu_set self);
#define qu_set_as_string(self)  qu__set__string (self)
#define qu_set_as_boolean(self) qu_container_as_boolean (self)


/***/
/*--------------------------------------------------------------------------*/

/* internally by Class to initialize this class */
extern void qu_set__setup ();
extern void qu_set__init  ();

/* for efficiency - may only be called from within this file */
extern qu_set     qu__set_new     (qu_class self, int argc, qu_any *argv);
extern qu_boolean qu__set__like   (qu_set self, qu_any x);
extern qu_set     qu__set__of     (qu_set self, qu_any type);
extern qu_set     qu__set__push   (qu_set self, qu_any x);
extern qu_set     qu__set__del    (qu_set self, qu_any x);
extern qu_set     qu__set__clear  (qu_set self);
extern qu_string  qu__set__string (qu_set self);
extern qu_boolean qu__set_typeof  (qu_set self, qu_any type);

#endif /*_H_qu_set_*/
