/* Stream.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_stream_
#define _H_qu_stream_ 1


/** Stream
	This the base class of all Stream classes (File, Socket, etc.).
*/
extern qu_class QuStreamClass;


/** QuStreamStd
	These are the Stream objects of v{stderr}, v{stdin}, v{stdout}, and k{null}.
*/
extern qu_stream QuStreamStderr;
extern qu_stream QuStreamStdin;
extern qu_stream QuStreamStdout;
extern qu_stream QuStreamNull;


/** qu_fstream_s
	Stream functions.

	See ={qu_stream_s} and ={QuStreamDefaultFuncs}.
*/
typedef	ssize_t   (*qu_stream_rdr_f  ) (int fd, char *buf, size_t len);
typedef	ssize_t   (*qu_stream_wrt_f  ) (int fd, const char *buf, size_t len);
typedef	ssize_t   (*qu_stream_ordr_f ) (qu_any obj, char *buf, size_t len);
typedef	ssize_t   (*qu_stream_owrt_f ) (qu_any obj, const char *buf, size_t len);
typedef	qu_stream (*qu_stream_acc_f  ) (qu_any obj);
typedef	qu_null   (*qu_stream_con_f  ) (qu_any obj, qu_any addr);
typedef	ssize_t   (*qu_stream_seek_f ) (int fd, off_t pos, int whence);
typedef	ssize_t   (*qu_stream_oseek_f) (qu_any obj, off_t pos, int whence);

typedef struct qu_fstream_s qu_fstream_s;
struct qu_fstream_s
{
	qu_stream_rdr_f   rdr;   /* read bytes from descriptor */
	qu_stream_wrt_f   wrt;   /* write bytes to descriptor */
	qu_stream_ordr_f  ordr;  /* read bytes from object */
	qu_stream_owrt_f  owrt;  /* write bytes to object */

	/* for the benefit of Socket/SSL */
	qu_stream_acc_f   acc;   /* accept a new client qu_stream */
	qu_stream_con_f   con;   /* connect to an address */

	qu_stream_seek_f  seek;  /* move file pointer of descriptor */
	qu_stream_oseek_f oseek; /* move file pointer of object */
};


/** QuStreamDefaultFuncs
	The default readers/writers.

	@args
	.rdr    read
	.wrt    write
	.ordr   internal, always return EBADF
	.owrt   -ditto-

	@warn
	Read-only! You must never change the content.
*/
extern qu_fstream_s QuStreamDefaultFuncs;


/** qu_stream QU_STREAM_HEAD
	qu_stream represents an instance of the Stream class.
	It is a pointer to the instance struct qu_stream_s.

	All Stream subclasses must use QU_STREAM_HEAD as the header instead of
	={QU_OBJECT_HEAD}.

	@warn
	The readers/writers must not take care of errors, including EINTR, EAGAIN, etc.
	Stream already takes care of those.

	Note that v{obj} must be a Qu object. It is marked by qu_stream__mark.

	If v{obj} is NULL then rdr/wrt/snd/rcv are used, otherwise ordr/owrt/osnd/orcv
	are used thus they may not be NULL.
*/
#define QU_STREAM_HEAD \
	QU_OBJECT_HEAD; \
	int           fd;       /* file descriptor  */ \
	qu_string     br;       /* linebreak string, may not be 0 */ \
	double        timeout;  /* when to timeout  */ \
	qu_fstream_s *funcs;    /* readers/writers */ \
	qu_any        obj;      /* object to pass to ordr/owrt/acc/con, may be NULL */ \
	long          sid       /* unique ID */

struct qu_stream_s
{
	QU_STREAM_HEAD;
};


/** QU_STREAM_FLAG
	Flags used by Stream objects.

	For C classes, these flags should be used directly when an instance is
	created, e.g. C{inst->flags = QU_STREAM_FLAG_CANEOF | QU_STREAM_FLAG_CANCLOSE}.
	Once created, the flags must never be modified.
*/
/* 0-9 reserved for Stream */
#define QU_STREAM_FLAG_CANEOF    QU_FLAG_0  /* EOF is relevant? */
#define QU_STREAM_FLAG_CANSELECT QU_FLAG_2  /* select is not relevant */
#define QU_STREAM_FLAG_CANBLOCK  QU_FLAG_3  /* can set blocking mode? */
#define QU_STREAM_FLAG_CANCLOSE  QU_FLAG_4  /* can be closed? */


/** qu_stream_can
	qu_stream_can_close checks whether the qu_stream can be closed.
	If not then the f{close} method should do nothing.

	qu_stream_can_block checks whether the qu_stream can be set/unset to blocking mode.
	If not then the f{blocking} method should do nothing.

	qu_stream_can_eof checks whether the Stream has EOF (end of file) notion.
	This is used by qu_stream_read et.al. to stop immediately when zero bytes
	are read from the Stream. A socket does not have EOF. A regular file does.

	qu_stream_can_select checks whether the Stream can be tested for readyness
	using the system f{select} call. If not then the Stream is assumed to be
	always ready.
*/
#define qu_stream_can_close(self)  qu_flag_test (self, QU_STREAM_FLAG_CANCLOSE)
#define qu_stream_can_block(self)  qu_flag_test (self, QU_STREAM_FLAG_CANBLOCK)
#define qu_stream_can_eof(self)    qu_flag_test (self, QU_STREAM_FLAG_CANEOF)
#define qu_stream_can_select(self) qu_flag_test (self, QU_STREAM_FLAG_CANSELECT)


/** qu_stream_is
	Whether or not an object is a Stream.
*/
#define qu_stream_is(o) qu_object_instof (o, QuStreamClass)


/** qu_stream_new
	Allocate a new Stream object.

	The following fields are set:
	@arg
	.fd      -1
	.br      qu_string_byte ('\n')
	.buf     NULL
	.timeout maximum value of time_t
	.funcs   ={QuStreamDefaultFuncs}
	.obj     NULL.

	@warn
	If you later set v{obj} then you must also set v{ordr}, v{owrt}, v{osnd} and v{orcv}.
*/
extern qu_stream qu_stream_new (qu_class self);


/** qu_stream_print
	These are called by the builtin -{print} functions.
	See ={qu_sys_print}.

	On success QuNil is returned, otherwise an exception is thrown and
	0 is returned.
*/
extern qu_null qu_stream_prints  (qu_stream self, qu_string s);
extern qu_null qu_stream_print   (qu_stream self, qu_any x);
extern qu_null qu_stream_println (qu_stream self, qu_any x);
extern qu_null qu_stream_printz  (qu_stream self, const char *s, long len);


/** qu_stream_select
	Get Stream status.

	Returns what the system f{select} function returns.

	As long as the Stream is valid, then the only possible v{errno} is
	ENOMEM. Note that EINTR is already handled by qu_stream_select.

	v{timeout} is in microseconds. A negative value means to use the
	default value (1000 microsecs).

	v{how} is one of:
	@args
	.r ready for reading
	.w ready for writing
	.e has error

	@p
	Calling this prior to do a read/write may improve performance a great
	deal.

	@warn
	={qu_stream_read} and ={qu_stream_write} already perform this so you must not
	do it yourself.

	f{qu_stream_select} does not wrap itself in ={__qu_blocking}. You must
	do it yourself if necessary.

	@p
	ENOMEM should have never been possible in the first place!
*/
extern int qu_stream_select (qu_stream self, char how, long timeout);


/** __qu_now__ qu_stream_now
	Return the current time since epoch as a double.
*/
#define __qu_now__ qu_stream_now ()
extern double qu_stream_now ();


/** qu_stream_read
	Read from Stream.

	In any case v{len} must be <= SSIZE_MAX.

	qu_stream_read is the read primitive. It return the length read on success,
	throw and return -1 otherwise.

	qu_stream_readln read until linebreak or v{max}. It return the length read on success,
	throw and return -1 otherwise.

	qu_stream_reads and qu_stream_readsln do the same but return a new String.

	qu_stream_recv and qu_stream_recvln do the same but read into a Buffer.
*/
extern long      qu_stream_read    (qu_stream self, char *buf, long len, int all);
extern long      qu_stream_readln  (qu_stream self, char *buf, long max);
extern qu_string qu_stream_reads   (qu_stream self, long n, int all);
extern qu_string qu_stream_readsln (qu_stream self, long max);
extern long      qu_stream_recv    (qu_stream self, qu_buffer b, long n, int all);
extern long      qu_stream_recvln  (qu_stream self, qu_buffer b, long max);


/** qu_stream_write
	Write to Stream.

	In any case v{len} must be <= SSIZE_MAX.

	All return the number of bytes written on success, -1 on error.

	qu_stream_write is the write primitive.
	qu_stream_writeln does the same then writes the linebreak.

	qu_stream_writes writes a String.
	qu_stream_writesln does the same then writes the linebreak.

	qu_stream_send does the same. It consumes the Buffer.
*/
extern long qu_stream_write    (qu_stream self, const char *buf, long len, int all);
extern long qu_stream_writeln  (qu_stream self, const char *buf, long len);
extern long qu_stream_writes   (qu_stream self, qu_string x, long ofs, long len, int all);
extern long qu_stream_writesln (qu_stream self, qu_string x, long ofs, long len);
extern long qu_stream_send     (qu_stream self, qu_buffer x, long len, int all);


/** qu_stream_transfer
	qu_stream_transfer read v{len} bytes from self and write it to v{dst}.

	If v{len} is negative then read until self reaches EOF. Return 1 on success
	0 otherwise.
*/
extern int qu_stream_transfer (qu_stream self, qu_stream dst, long len);


/** qu_stream_close
	Close the Stream.
*/
#define qu_stream_close(self) qu__stream_close (self)


/** qu_stream_contains
	Keep reading from Stream until the string is found or EOF is reached or
	an exception is thrown.

	Return QuTrue if found, QuFalse if not found, NULL if an error occurred,
*/
#define qu_stream_contains(self, str) qu__stream__in (self, str)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_stream__setup ();
extern void qu_stream__init  ();

/* published for the benefit of C subclasses */
extern void qu__stream__destroy (qu_stream self);
extern void qu__stream__mark    (qu_stream self);

/* for efficiency - may only be called from within this file */
extern qu_boolean qu__stream__in      (qu_stream self, qu_string x);
extern qu_null    qu__stream_close    (qu_stream self);
extern qu_boolean qu__stream_contains (qu_stream self, qu_string x);

#endif /*_H_qu_stream_*/
