/* String.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_string_
#define _H_qu_string_ 1


/** String
	This is the final String class.

	String objects are immutable byte based strings.
	The string may contain nils and is nil terminated just like C strings.

	Thus if you pass a string to a C function and the string happens to contain
	a nil somewhere in the middle then C will interpret it as the end of the
	string.
*/
extern qu_class QuStringClass;


/** qu_string QU_STRING_HEAD
	qu_string represents an instance of the String class.
	It is a pointer to the instance struct qu_string_s.
*/
#define QU_STRING_HEAD \
	QU_OBJECT_HEAD; \
	long   len;   /* content length */ \
	long   hash;  /* cached hash value, -1 = not cached */ \
	char  *str    /* content is always nil terminated */ \

struct qu_string_s
{
	QU_STRING_HEAD;
};


/** qu_string_is
	Whether or not the object is a String.
**/
#define qu_string_is(o) (qu_classof(o) == QuStringClass)


/** qu_string_byte
	__QuStringByteList is the list of system predefined single byte String objects.
	You access a particular object with the qu_string_byte macro.
*/
#define qu_string_byte(c) __QuStringByteList [(c) & 255]
/***/
extern qu_string __QuStringByteList [256];


/** QuStringN
	Predefined static String objects.
*/
extern qu_string QuStringEMPTY;  /* the empty string */
extern qu_string QuStringALL8;
extern qu_string QuStringALL7;
extern qu_string QuStringALNUM;
extern qu_string QuStringALPHA;
extern qu_string QuStringCNTRL;
extern qu_string QuStringBDIGIT;
extern qu_string QuStringBLANK;
extern qu_string QuStringDIGIT;
extern qu_string QuStringGRAPH;
extern qu_string QuStringIDENT;
extern qu_string QuStringIDNUM;
extern qu_string QuStringLOWER;
extern qu_string QuStringNAME;
extern qu_string QuStringODIGIT;
extern qu_string QuStringPATH;
extern qu_string QuStringPRINT;
extern qu_string QuStringPUNCT;
extern qu_string QuStringPWD;
extern qu_string QuStringSPACE;
extern qu_string QuStringUPPER;
extern qu_string QuStringXDIGIT;
extern qu_string QuStringWHITE;
extern qu_string QuStringWORD;


/** QU_STRING_FLAG
	Flags used by String objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_STRING_FLAG_PROTECT QU_FLAG_0  /* \__ match qu_buffer/qu_container */
#define QU_STRING_FLAG_FROZEN  QU_FLAG_1  /* /                          */
#define QU_STRING_FLAG_STATIC  QU_FLAG_7  /* static memory? */
#define QU_STRING_FLAG_UTF8    QU_FLAG_3  /* UTF-8? (currently not used) */

/*
	10..end are reserved for Qu modules
	These may only be touched by the corresonding modules.
*/
#define QU_STRING_FLAG_NETADDR   QU_FLAG_11  /* reserved for the Net module */
#define QU_STRING_FLAG_SOCKADDR  QU_FLAG_12  /* -ditto- */
#define QU_STRING_FLAG_REGEX     QU_FLAG_13  /* reserved for the Regex module */


/** qu_string_is_static
	Whether or not the String is a static String.

	A static String is never marked by the GC and it's v{str} field points to static
	memory or permanent ={qu_mem_alloc} memory.
*/
#define qu_string_is_static(self)  qu_flag_test (self, QU_STRING_FLAG_STATIC)


/** qu_string_hash
	Get the string hash value.

	@warn
	Prefer qu_string_HASH over qu_string_hash but notice the macro side effect.
*/
extern long qu_string_hash   (qu_string self);
#define qu_string_HASH(self) (((self)->hash == -1) ? qu_string_hash (self) : (self)->hash)


/** qu_string_eq
	Whether or not a String equals another.
*/
extern int qu_string_eq (qu_string self, qu_string x);

#define qu_string_EQ(self, x) \
	((qu_string) (self) == (qu_string) (x) || qu_string_eq (self, x))


/** qu_string_new
	Various string constructors.

	qu_string_new takes a memory location and a length.

	qu_string_newz takes a C nil terminated string.

	qu_string_newl creates an uninitialized string with the v{hash} field set to -1.
	Use this function to set the content yourself.

	qu_string_newf creates a string using arguments ala C f{printf}.

	qu_string_newc creates a string by repreating a character.

	qu_string_new_trim does the same as qu_string_new except that the leading and
	trailing spaces are removed.

	qu_string_source creates string from source code.
	This function is used by the compiler. It translates escape sequences.

	qu_string_dup creates a new string with the same content.
	The v{hash} field is set to -1 since you generally would call this
	function to modify the content. This is actually the same as calling
	qu_string_newl then copy the content.

	qu_string_cat concatenate all strings.
	All arguments must be String objects.
	The last argument must be NULL to indicate termination.

	qu_string_new_static creates a new string from a static C string.
	The string will not be included in the GC thus will never be freed.
	If one already exists in the static string table then no new string is created.

	qu_string_static_dup does the same but reallocate the C string.

	qu_string_static_dupz does the same and calculates the length.

	qu_string_new_staticz does the same as qu_string_new_static and computes f{strlen}
	all by itself.

	qu_string_get_static returns an existing string from the static string table
	if one exists, v{s} otherwise. This function is for efficiency to avoid
	creating duplicate strings, that is, if an existing string exist then
	forget v{s}.

	qu_string_static_findz is a version of qu_string_get_static.
	It returns NULL if no match is found.
	An existing string otherwise.

	@warn
	Make sure v{len} is within C{0..}={QU_MEM_MAX}.
	No checks are made.

	The argument v{z} to qu_string_new_static and qu_string_new_staticz must be a nil
	terminated C static string e.g. C{"foo bar"}. It may not be an allocated memory
	address unless it is permanent (never freed).
*/
extern qu_string qu_string_new          (const char *z, long len);
extern qu_string qu_string_newz         (const char *z);
extern qu_string qu_string_newl         (long len);
extern qu_string qu_string_newf         (const char *format, ...);
extern qu_string qu_string_newc         (char c, long n);
extern qu_string qu_string_new_trim     (const char *s, long len);
extern qu_string qu_string_dup          (qu_string s);
extern qu_string qu_string_source       (const char *z, long len);
extern qu_string qu_string_cat          (qu_string s, ...);
extern qu_string qu_string_new_static   (const char *z, long len);
extern qu_string qu_string_static_dup   (const char *z, long len);
extern qu_string qu_string_static_dupz  (const char *z);
extern qu_string qu_string_new_staticz  (const char *z);
extern qu_string qu_string_new_staticv  (const char *fmt, va_list va);
extern qu_string qu_string_new_staticf  (const char *fmt, ...);
extern qu_string qu_string_get_static   (qu_string s);
extern qu_string qu_string_static_findz (const char *z, long len);

extern qu_string qu_string_to_static    (qu_string self);


/** qu_string_print
	Print a String to v{stdout}.

	To print the String unquoted, set v{quot} to 0.
	Otherwise v{quot} must be one of C{'}, C{`}, or C{"}.

	Note that this might throw an exception because writing to v{stdout} may
	fail for various reasons.
*/
extern qu_null qu_string_print (qu_string self, char quot);


/** qu_string_format
	Format this string.

	See ={Format} for details.

	This function is provided for efficiency between core modules.
	It is not intended to be called by users directly.
*/
extern qu_string qu_string_format (qu_string self, qu_format_s *f);


/** qu_string_quote
	Return a quoted string.
*/
#define qu_string_quote(self) qu_strz_quote ((self)->str, (self)->len, '\'')


/***/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	C string
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/** qu_strz
	Functions dealing with C strings.

	qu_strz_hash computes the string hash value.
	Note that there is a cut-off implemented for reasonable performance.

	qu_strz_eq return 1 if the strings are equal, 0 otherwise.
	qu_strz_eqnc does the same in case insensitive way.
	The functions are intentionally taking only one length argument to force you to
	check for the same length before calling the functions.

	qu_strz_chr returns the position of a character in the string, NULL if not found.
	qu_strz_chrnc does the same in case insensitive way.

	qu_strz_str returns the position of a string in the string, NULL if not found.
	qu_strz_strnc does the same in case insensitive way.

	qu_strz_set_cset fills a 256 character buffer with characters from a string.
	v{cset} is not cleared and the values are set to 1 if the character exists,
	unchanged otherwise. v{cmin} and v{cmax} are set to the lowest and highest
	character value.

	qu_strz_set_csetnc does the same except that it also sets the lowercase and
	uppercase characters in the set for all alphabet characters.

	qu_strz_slice return a new String from a slice of the string.

	qu_strz_index find the index of a string in a string.
	Return -1 if not found.

	qu_strz_startwith test whether the string starts with v{what} while
	qu_strz_endwith test whether the string ends with v{what}.
	They return the matched item or NULL if none found.
	v{what} may be a String or an iterable that yields String objects.
	In case of a Dict then the keys are used and the corresponding value is stored
	in v{value} which may be anything.
	Non String keys are ignored.

	qu_strz_translate translates a character.
	v{what} may be a String or an iterable that yields String objects.
	In case of a Dict then the keys are used and the corresponding value is stored
	in v{value} which may be anything.
	Non String keys are ignored.
	Return NULL if no translation occured.
*/
extern long       qu_strz_hash      (const char *z, long l);
extern int        qu_strz_eq        (const char *z, const char *x, long l);
extern int        qu_strz_eqnc      (const char *z, const char *x, long l);
extern char      *qu_strz_chr       (const char *z, long lz, char c);
extern char      *qu_strz_chrnc     (const char *z, long lz, char c);
extern int        qu_strz_cmp       (const char *z, long lz, const char *x, long lx);
extern int        qu_strz_casecmp   (const char *z, long lz, const char *x, long lx);
extern char      *qu_strz_str       (const char *z, long lz, const char *x, long lx);
extern char      *qu_strz_strnc     (const char *z, long lz, const char *x, long lx);
extern long       qu_strz_copy      (const char *z, long lz, char *x, long len);
extern qu_string  qu_strz_slice     (const char *z, long lz, long i, long n);
extern long       qu_strz_index     (const char *z, long lz, const char *x, long lx, long ofs);
extern long       qu_strz_indexnc   (const char *z, long lz, const char *x, long lx, long ofs);
extern long       qu_strz_rindex    (const char *z, long lz, const char *x, long lx, long ofs);
extern long       qu_strz_rindexnc  (const char *z, long lz, const char *x, long lx, long ofs);
extern qu_string  qu_strz_startwith (const char *z, long lz, int nocase, qu_any what, qu_any *value);
extern qu_string  qu_strz_endwith   (const char *z, long lz, int nocase, qu_any what, qu_any *value);
extern void       qu_strz_reverse   (char *z, long lz);
extern qu_string  qu_strz_translate (unsigned char c, qu_any x);
extern qu_string  qu_strz_quote     (const char *z, long lz, char quot);

extern void qu_strz_set_cset (const unsigned char *z, long lz, unsigned char *cset256,
                              unsigned char *cmin, unsigned char *cmax);

extern void qu_strz_set_csetnc (const unsigned char *z, long lz, unsigned char *cset256,
                                unsigned char *cmin, unsigned char *cmax);

#define qu_strz_EQ(za, la, zb, lb)   (((la) == (lb)) && qu_strz_eq   (za, zb, la))
#define qu_strz_EQNC(za, la, zb, lb) (((la) == (lb)) && qu_strz_eqnc (za, zb, la))

#define qu_string_index(sa, sb, ofs)   qu_strz_index  ((sa)->str, (sa)->len, (sb)->str, (sb)->len, ofs)
#define qu_string_rindex(sa, sb, ofs)  qu_strz_rindex ((sa)->str, (sa)->len, (sb)->str, (sb)->len, ofs)
#define qu_string_getslice(sa, ofs, n) qu_strz_slice  ((sa)->str, (sa)->len, ofs, n)


/***/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	buffer
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/** qu_string_buffer
	String buffer used by the builtin print functions etc.

	@warn
	String buffers use the v{hash} field to store the current allocated capacity.
	You must never touch this field.

	Make sure to call qu_string_buffer_return when the String is ready to be returned
	to program level.

	@p
	qu_string_buffer_new creates a new string buffer.

	qu_string_buffer_put append a string to the buffer.

	qu_string_buffer_clear clears the buffer.

	qu_string_buffer_reverse reverses the content.

	qu_string_buffer_set_len sets the content length. Use this only if you are
	injecting the content directly. Be careful!

	qu_string_buffer_make_space makes sure there is enough room to put more bytes,
	growing the buffer if necessary.

	qu_string_buffer_return must be called prior to returning the string to program
	level. This puts the nil terminator and sets the v{hash} field to -1.
*/
extern qu_string qu_string_buffer_new        (long cap);
extern int       qu_string_buffer_puts       (qu_string self, qu_string s);
extern int       qu_string_buffer_put        (qu_string self, const char *z, long len);
extern int       qu_string_buffer_putz       (qu_string self, const char *z);
extern int       qu_string_buffer_putc       (qu_string self, char c);
extern void      qu_string_buffer_clear      (qu_string self);
extern void      qu_string_buffer_reverse    (qu_string self);
extern qu_string qu_string_buffer_return     (qu_string self);
extern void      qu_string_buffer_set_len    (qu_string self, long len);
extern int       qu_string_buffer_make_space (qu_string self, long more);
extern long      qu_string_buffer_avail      (qu_string self);
extern qu_string qu_string_buffer_return     (qu_string self);


/***/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/** qu_string_cmpint
	Compare two Strings and return an Int in the usual manner.
*/
#define qu_string_cmpint(self, x) qu__string__cmp (self, x)


/** qu_string_delete
	Return a new String by replacing occurences of v{what}.
*/
extern qu_string qu_string_delete (qu_string self, qu_string what);


/** qu_string_pad
	Return a new String by padding v{fill} to the left and right.
*/
extern qu_string qu_string_pad (qu_string self, long left, long right, char fill);


/** qu_string_trim
	Return a new String by trimming spaces.
*/
#define qu_string_ltrim(self) qu__string_ltrim (self)
#define qu_string_rtrim(self) qu__string_rtrim (self)
#define qu_string_trim(self)  qu__string_trim (self)


/** qu_string_interval
	Return a new String by inserting v{delim} at v{intvl} intervals.
*/
extern qu_string qu_string_interval (qu_string self, qu_string delim, long intvl);


/** qu_string_cfunc_name
	Convert C function name to something that makes sense.

	This is used by Qu to report error messages. It translates something like
	C{"qu__string__cmp"} or C{"_string__cmp"} into C{"String.__cmp"}.
*/
void qu_string_cfunc_name (const char *func, char *buf, size_t size);


/** qu_string_as
	Various representations.
*/
extern qu_array qu_string_as_array (qu_string self);
extern qu_set   qu_string_as_set   (qu_string self);
extern qu_dict  qu_string_as_dict  (qu_string self);
#define         qu_string_as_boolean(self) ((self)->len ? QuTrue : QuFalse)
#define         qu_string_as_buffer(self)  qu_buffer_new ((self)->str, (self)->len)


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_string__setup ();
extern void qu_string__init  ();

/* for efficiency - may only be called from within this file */
extern qu_int    qu__string__cmp  (qu_string self, qu_any x);
extern qu_string qu__string_ltrim (qu_string self);
extern qu_string qu__string_rtrim (qu_string self);
extern qu_string qu__string_trim  (qu_string self);

/* to avoid duplicate codes in Buffer - must only be called by Buffer */
extern qu_boolean qu__string_isalnum  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isalpha  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isbdigit (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isblank  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_iscntrl  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isdigit  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isgraph  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_ishi     (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isident  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isidnum  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_islo     (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_islower  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isname   (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isodigit (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_ispath   (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isprint  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_ispwd    (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_ispunct  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isspace  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_istext   (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isupper  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isxdigit (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_iswhite  (qu_string self, int argc, qu_any *argv);
extern qu_boolean qu__string_isword   (qu_string self, int argc, qu_any *argv);

extern qu_any qu__string__as (qu_string self, qu_class c);
extern qu_int qu__string_asc (qu_string self);
extern qu_int qu__string_ord (qu_string self);

#endif /*_H_qu_string_*/
