/* Struct.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_struct_
#define _H_qu_struct_ 1


/** qu_struct
	This is the final Struct class.
*/
extern qu_class QuStructClass;


/** qu_struct
	qu_sub represents an instance of the Struct class.
	It is a pointer to the instance struct qu_struct_s.
*/
typedef struct qu_structentry_s qu_structentry_s;
struct qu_structentry_s
{
	qu_string name; /* field name */
	qu_any    type; /* field type Class, Var, Sub, or Struct */
};

struct qu_struct_s
{
	QU_OBJECT_HEAD; /* attrs used for specialized methods, 0 if none */

	/* never access these fields
		they are implementation specific */

	qu_attr_s        *indexes;  /* list of indexes (key=name, value=index) */
	qu_structentry_s *items;    /* items are allocated */
};


/** QU_STRUCT_FLAG
	Flags used by Struct objects.
*/
/* none but reserve all */


/** qu_struct_is
	Whether or not an object is a Struct.
*/
#define qu_struct_is(o) (qu_classof(o) == QuStructClass)


/** qu_struct_len
	Get the number of items in the Struct.
*/
#define qu_struct_len(self) qu_attr_len ((self)->indexes)


/** qu_struct_new
	Create a new Struct object.

	qu_struct_new_static creates a Struct that is never garbage collected.

	@xmp
	qu_struct mystruct = qu_struct_new (3, QuIntClass, "foo", QuFloatClass, "bar", QuNil, "baz");
*/
extern qu_struct qu_struct_new        (int nelt, ...);
extern qu_struct qu_struct_new_static (int nelt, ...);


/** qu_struct_set_methodz
	A wrapper arount f{qu__struct_setmethod}.

	@warn
	v{namez} must be a static C string, e.g. C{"foo"}.
	It must not be a dynamic pointer.
*/
#define qu_struct_set_methodz(self, namez, f) qu__struct_setmethod (self, qu_string_new_staticz (namez), f)


/** qu_struct_index qu_struct_index_ex
	qu_struct_index finds the index of a field by name.

	qu_struct_index_ex takes any object. If it is a String then qu_struct_index is called.
	If it is an Integer then validates the index. Otherwise returns -1.

	If not found, -1 is returned.
*/
extern long qu_struct_index    (qu_struct self, qu_string x);
extern long qu_struct_index_ex (qu_struct self, qu_any x);


/** qu_struct_type_s
	Lists the Qu "type", C type size and C type alignment for each type characters.

	See ={qu_struct_type_to_char} for a list of type characters.

	Use qu_struct_type_info to get the qu_struct_type_s of a character.
*/
typedef struct qu_struct_type_s qu_struct_type_s;
struct qu_struct_type_s
{
	qu_any      type;  /* Class or validator Sub */
	int         align; /* alignment */
	int         size;  /* C size */
	const char *name;  /* C name */
};
#define qu_struct_type_info(ch) (&__QuStructTypeList[(unsigned char) (ch)])
/***/
extern qu_struct_type_s __QuStructTypeList [256];


/** QU_STRUCT_VALID_SPEC qu_struct_valid_spec
	QU_STRUCT_VALID_SPEC checks for a valid specification character.

	qu_struct_valid_spec checks for a valid string specification.
*/
#define QU_STRUCT_VALID_SPEC(c) (__QuStructTypeList[(unsigned char) (c)].type != NULL)
extern int qu_struct_valid_spec (char *s, long len);


/** qu_struct_type_to_char
	qu_struct_type_to_char translates a type object into the corresponding character
	specification.

	The following table shows the list of character types and the corresponding
	C type, Qu "type", and Qu object.

	@table
	.| a | char               | QuType_char               | String
	.| y | byte               | QuType_byte               | Int
	.| Y | unsigned byte      | QuType_ubyte              | Int
	.| h | short              | QuType_short              | Int
	.| H | unsigned short     | QuType_ushort             | Int
	.| i | int                | QuType_int                | Int or Long
	.| I | unsigned int       | QuType_uint               | Int or Long
	.| j | int16              | QuType_int16              | Int
	.| J | unsigned int16     | QuType_uint16             | Int
	.| k | int32              | QuType_int32              | Int or Long
	.| K | unsigned int32     | QuType_uint32             | Int or Long
	.| l | long               | QuType_long               | Int or Long
	.| L | unsigned long      | QuType_ulong              | Int or Long
	.| o | long long          | QuType_longlong           | Int or Long
	.| O | unsigned long long | QuType_ulonglong          | Int or Long
	.| g | float              | QuType_float              | Float
	.| f | double             | QuType_double or Float    | Float
	.| F | long double        | QuType_longdouble         | Float or Big
	.| z | ssize_t            | QuType_ssize_t            | Int or Long
	.| Z | size_t             | QuType_size_t             | Int or Long
	.| e | off_t              | QuType_off_t              | Int or Long
	.| s | char *             | QuType_char_p             | String
	.| X | void *             | QuType_void_p             | Pointer
	.| n | NULL               | Null                      | Null
	.| b | int                | Boolean                   | Boolean
	.| q | long               | Int                       | Int
	.| p | void *             | Pointer                   | Pointer
	.| S | String             | String                    | String
	.| A | Array              | Array                     | Array
	.| C | Class              | Class                     | Class
	.| D | Dict               | Dict                      | Dict
	.| E | Set                | Set                       | Set
	.| P | Proc               | Proc                      | Proc
	.| Q | Queue              | Queue                     | Queue
	.| R | Range              | Range                     | Range
	.| T | Stream             | Stream                    | Stream
	.| x | Any                | Any                       | any Qu object as is
*/
extern char qu_struct_type_to_char (qu_any type);

#define QU_STRUCT_ID_char       'a'
#define QU_STRUCT_ID_byte       'y'
#define QU_STRUCT_ID_ubyte      'Y'
#define QU_STRUCT_ID_short      'h'
#define QU_STRUCT_ID_ushort     'H'
#define QU_STRUCT_ID_int        'i'
#define QU_STRUCT_ID_uint       'I'
#define QU_STRUCT_ID_int16      'j'
#define QU_STRUCT_ID_uint16     'J'
#define QU_STRUCT_ID_int32      'k'
#define QU_STRUCT_ID_uint32     'K'
#define QU_STRUCT_ID_long       'l'
#define QU_STRUCT_ID_ulong      'L'
#define QU_STRUCT_ID_longlong   'o'
#define QU_STRUCT_ID_ulonglong  'O'
#define QU_STRUCT_ID_float      'g'
#define QU_STRUCT_ID_double     'f'
#define QU_STRUCT_ID_longdouble 'F'
#define QU_STRUCT_ID_ssize_t    'z'
#define QU_STRUCT_ID_size_t     'Z'
#define QU_STRUCT_ID_off_t      'e'
#define QU_STRUCT_ID_char_p     'S'
#define QU_STRUCT_ID_void_p     'X'
#define QU_STRUCT_ID_Null       'n'
#define QU_STRUCT_ID_Boolean    'b'
#define QU_STRUCT_ID_Int        'q'
#define QU_STRUCT_ID_Pointer    'p'
#define QU_STRUCT_ID_String     's'
#define QU_STRUCT_ID_Stringy    'N'
#define QU_STRUCT_ID_Any        'x'
#define QU_STRUCT_ID_Array      'A'
#define QU_STRUCT_ID_Buffer     'B'
#define QU_STRUCT_ID_Class      'C'
#define QU_STRUCT_ID_Dict       'D'
#define QU_STRUCT_ID_Function   'U'
#define QU_STRUCT_ID_Method     'M'
#define QU_STRUCT_ID_Proc       'P'
#define QU_STRUCT_ID_Queue      'Q'
#define QU_STRUCT_ID_Range      'R'
#define QU_STRUCT_ID_Set        'E'
#define QU_STRUCT_ID_Stream     'T'
#define QU_STRUCT_ID_Sub        'v'
#ifdef QU_SUPPORT_GLIB
#define QU_STRUCT_ID_GObject    'G'
#endif

#define QU_STRUCT_IDS_char       "a"
#define QU_STRUCT_IDS_byte       "y"
#define QU_STRUCT_IDS_ubyte      "Y"
#define QU_STRUCT_IDS_short      "h"
#define QU_STRUCT_IDS_ushort     "H"
#define QU_STRUCT_IDS_int        "i"
#define QU_STRUCT_IDS_uint       "I"
#define QU_STRUCT_IDS_int16      "j"
#define QU_STRUCT_IDS_uint16     "J"
#define QU_STRUCT_IDS_int32      "k"
#define QU_STRUCT_IDS_uint32     "K"
#define QU_STRUCT_IDS_long       "l"
#define QU_STRUCT_IDS_ulong      "L"
#define QU_STRUCT_IDS_longlong   "o"
#define QU_STRUCT_IDS_ulonglong  "O"
#define QU_STRUCT_IDS_float      "g"
#define QU_STRUCT_IDS_double     "f"
#define QU_STRUCT_IDS_longdouble "F"
#define QU_STRUCT_IDS_ssize_t    "z"
#define QU_STRUCT_IDS_size_t     "Z"
#define QU_STRUCT_IDS_off_t      "e"
#define QU_STRUCT_IDS_char_p     "S"
#define QU_STRUCT_IDS_void_p     "X"
#define QU_STRUCT_IDS_Null       "n"
#define QU_STRUCT_IDS_Boolean    "b"
#define QU_STRUCT_IDS_Int        "q"
#define QU_STRUCT_IDS_Pointer    "p"
#define QU_STRUCT_IDS_String     "s"
#define QU_STRUCT_IDS_Stringy    "N"
#define QU_STRUCT_IDS_Any        "x"
#define QU_STRUCT_IDS_Array      "A"
#define QU_STRUCT_IDS_Buffer     "B"
#define QU_STRUCT_IDS_Class      "C"
#define QU_STRUCT_IDS_Dict       "D"
#define QU_STRUCT_IDS_Function   "U"
#define QU_STRUCT_IDS_Method     "M"
#define QU_STRUCT_IDS_Proc       "P"
#define QU_STRUCT_IDS_Queue      "Q"
#define QU_STRUCT_IDS_Range      "R"
#define QU_STRUCT_IDS_Set        "E"
#define QU_STRUCT_IDS_Stream     "T"
#define QU_STRUCT_IDS_Sub        "v"
#ifdef QU_SUPPORT_GLIB
#define QU_STRUCT_IDS_GObject    "G"
#endif

/** qu_struct_offset
	Find the type and offset of a C value if the Struct were a C struct.

	qu_struct_offset_ex does the same but v{x} may be a Struct or a String
	specification e.g. C{"ssfi"}.
*/
extern long qu_struct_offset    (qu_struct self, long i, char *type);
extern long qu_struct_offset_ex (qu_any x, long i, char *type);


/** qu_struct_size
	Return the size of the Struct if it were a C struct.

	qu_struct_size_ex does the same but v{x} may be a Struct or a String
	specifiction e.g. C{"ssfi"}.
*/
extern long qu_struct_size    (qu_struct self);
extern long qu_struct_size_ex (qu_any x);


/** qu_struct_getitem
	Get an item from the Struct.
*/
#define qu_struct_getitem(self, x) qu__struct__getitem (self, x)


/** qu_struct_like
	Whether two Structs are alike.
*/
#define qu_struct_like(self, x) (qu__struct__like (self, x) == QuTrue)


/** qu_struct_as
	Various representations.
*/
extern qu_dict  qu_struct_as_dict  (qu_struct self);
extern qu_array qu_struct_as_array (qu_struct self);
#define qu_struct_as_string(self)  qu__struct__string (self)
#define qu_struct_as_boolean(self) QuTrue


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_struct__setup ();
extern void qu_struct__init  ();

/* for efficiency - may only be called from within this file */
extern qu_boolean qu__struct__like    (qu_struct self, qu_struct x);
extern qu_string  qu__struct__string  (qu_struct self);
extern qu_any     qu__struct__getitem (qu_struct self, qu_any x);

/* called by qu_sys__init */
extern void qu_struct_sys_init ();

/* for Interp and Code */
extern qu_sub QuStructMethodNew;

/* for Reduce */
extern int       qu_struct_compile_setup    (qu_var v, qu_ast n);
extern qu_struct qu_struct_compile_prealloc ();

#endif /*_H_qu_struct_*/
