/* Sub.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_sub_
#define _H_qu_sub_ 1


/** QU_SUB_ALLOW_ASSIGN
	Whether or not calling Sub can be done via the assignment operator.

	This must always be defined.
	The macro only serves as a placeholder so tracking is easier.
*/
#define QU_SUB_ALLOW_ASSIGN 1


/** Sub
	This is the final Sub class.
*/
extern qu_class QuSubClass;


/** qu_sub
	qu_sub represents an instance of the Sub class.
	It is a pointer to the instance struct qu_sub_s.

	Methods and functions are all Sub objects.

	We have three types:
	@args
	.external  C level function/method, QU_SUB_FLAG_EXTERN set
	.regular   script level function/method, QU_SUB_FLAG_EXTERN not set
	.closure   script level closures, QU_SUB_FLAG_CLOSURE set, QU_SUB_FLAG_EXTERN not set

	@warn
	Some fields are not allocated depending on the type.

	@table
	.| QU_SUB_FLAG_EXTERN                   | QU_SUB_HEAD only
	.| QU_SUB_FLAG_EXTERN  QU_SUB_FLAG_JIT  | QU_SUB_HEAD + QU_SUB_SCRIPT_FIELDS
	.| QU_SUB_FLAG_CLOSURE                  | all
	.| others                               | QU_SUB_HEAD + QU_SUB_SCRIPT_FIELDS
*/
#define QU_SUB_HEAD \
	QU_SYMBOL_HEAD; \
	qu_any        value;   /* Code object or C function address or "once" value or JIT code */ \
	int           argc;    /* Minimum number of arguments, -1 means any */ \
	qu_any        type;    /* Return type, maybe 0 */ \
	qu_any       *ltype;   /* List of lexical types, 0 or set by qu_sub_reindex */ \
	qu_dict       tag;     /* The tag entries, maybe 0 */ \
	qu_subcall_s *call     /* How to call, set by Sub based on flags */

#define QU_SUB_SCRIPT_FIELDS \
	qu_sub        root;    /* Root sub, self for root */ \
	int           narg;    /* Actual number of arguments */ \
	int           nlocal;  /* Number of lexicals */ \
	int           first    /* Index of first lexical */ \

#define QU_SUB_CLOSURE_FIELDS \
	qu_array      clex;   /* Closure lexicals */ \
	qu_sub        corg    /* The Sub object used to create the closure */


typedef qu_any (*qu_subcalln_f) (qu_sub self, qu_any theself, int argc, qu_any *argv);
typedef qu_any (*qu_subcall4_f) (qu_sub self, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
typedef qu_any (*qu_subcall3_f) (qu_sub self, qu_any theself, qu_any a1, qu_any a2, qu_any a3);
typedef qu_any (*qu_subcall2_f) (qu_sub self, qu_any theself, qu_any a1, qu_any a2);
typedef qu_any (*qu_subcall1_f) (qu_sub self, qu_any theself, qu_any a);
typedef qu_any (*qu_subcall0_f) (qu_sub self, qu_any theself);

typedef struct qu_subcall_s qu_subcall_s;
struct qu_subcall_s
{
	qu_subcalln_f fn;
	qu_subcall4_f f4;
	qu_subcall3_f f3;
	qu_subcall2_f f2;
	qu_subcall1_f f1;
	qu_subcall0_f f0;
};

struct qu_sub_s
{
	QU_SUB_HEAD;
	QU_SUB_SCRIPT_FIELDS;  /* allocated for script subs only */
	QU_SUB_CLOSURE_FIELDS; /* allocated for closures only */
};


/** qu_sub_is
	Whether or not the object is a Sub.
*/
#define qu_sub_is(o) (qu_classof(o) == QuSubClass)


/** QU_SUB_FLAG
	Flags use by Sub objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_SUB_FLAG_EXTERN     QU_SYMBOL_FLAG_EXTERN   /* C or script? */
#define QU_SUB_FLAG_PRIVATE    QU_SYMBOL_FLAG_PRIVATE  /* private? */
#define QU_SUB_FLAG_LOCAL      QU_SYMBOL_FLAG_LOCAL    /* local? */
#define QU_SUB_FLAG_STATIC     QU_SYMBOL_FLAG_STATIC   /* self may be anything? */
#define QU_SUB_FLAG_CLASS      QU_SYMBOL_FLAG_CLASS    /* class method? */
#define QU_SUB_FLAG_INST       QU_SYMBOL_FLAG_INST     /* instance method? */
#define QU_SUB_FLAG_FINAL      QU_SYMBOL_FLAG_FINAL    /* cant override? */
#define QU_SUB_FLAG_ABSTRACT   QU_SYMBOL_FLAG_ABSTRACT /* must override? */
#define QU_SUB_FLAG_NOJIT      QU_FLAG_9               /* do not JIT */
#define QU_SUB_FLAG_VARARG     QU_FLAG_10              /* last arg is vararg? e.g. foo (a, *b) */
#define QU_SUB_FLAG_ARGMIN     QU_FLAG_11              /* has default arguments? (argc is minimum) */
#define QU_SUB_FLAG_ARGCHECK   QU_FLAG_12              /* must check argument type? */
#define QU_SUB_FLAG_CFUNC      QU_FLAG_13              /* C function? (don't pass self) */
#define QU_SUB_FLAG_CLOSURE    QU_FLAG_14              /* is closure? */
#define QU_SUB_FLAG_GENERATOR  QU_FLAG_15              /* is generator? */
#define QU_SUB_FLAG_UNSAFE     QU_FLAG_16              /* __unsafe? */
#define QU_SUB_FLAG_ONCE       QU_FLAG_17              /* __once? */
#define QU_SUB_FLAG_ACCUM      QU_FLAG_18              /* __accumulator? */
#define QU_SUB_FLAG_ENSURE     QU_FLAG_19              /* property validator? (ensure) */
#define QU_SUB_FLAG_NOAUTO     QU_FLAG_20              /* call requires parentheses? */
#define QU_SUB_FLAG_JIT        QU_FLAG_21              /* jit-ed? */
#define QU_SUB_FLAG_INTERFACE  QU_FLAG_22              /* __interface? */
#define QU_SUB_FLAG_ARGS       QU_FLAG_23              /* __args? */


/** qu_sub_is_extern
	Whether this is an external function (C or whatever) or program level Sub.

	Note that JIT might turn a program level Sub to "external".
*/
#define qu_sub_is_extern(self) qu_flag_test (self, QU_SUB_FLAG_EXTERN)


/** qu_sub_is_class
	Whether this is a class method.

	@xmp
		class Foo
			sub class bar  <-
*/
#define qu_sub_is_class(self) qu_flag_test (self, QU_SUB_FLAG_CLASS)


/** qu_sub_is_inst
	Whether this is an instance method.

	@xmp
		class Foo
			sub bar  <-
*/
#define qu_sub_is_inst(self) qu_flag_test (self, QU_SUB_FLAG_INST)


/** qu_sub_is_static
	Whether self in this Sub may have any value (unbound).

	This has nothing to do with the word "static" as known in other language such
	as C, Java, etc.

	All module level functions and child Sub (Sub within Sub) are "static".

	@xmp
		sub foo             <-
			sub bar         <-

		class Foo
			sub static foo  <-
			sub bar
				sub baz     <-
*/
#define qu_sub_is_static(self) qu_flag_test (self, QU_SUB_FLAG_STATIC)


/** qu_sub_is_private
	Whether this is a private function known to block/childs, or method known
	to class/subclasses.

	@xmp
		sub foo
		sub _bar            <-
		sub baz __private   <-
*/
#define qu_sub_is_private(self) qu_flag_test (self, QU_SUB_FLAG_PRIVATE)


/** qu_sub_is_local
	Whether it is local to the class (unknown to subclasses).

	@xmp
		class Foo
			sub local bar   <-
*/
#define qu_sub_is_local(self) qu_flag_test (self, QU_SUB_FLAG_LOCAL)


/** qu_sub_is_final
	Whether this is a final method (can not override).

	@xmp
		class Foo
			sub bar __final  <-
*/
#define qu_sub_is_final(self) qu_flag_test (self, QU_SUB_FLAG_FINAL)


/** qu_sub_is_abstract
	Whether this is an abstract method (must be defined by real subclasses).

	@xmp
		class Foo __abstract
			sub bar __abstract  <-
*/
#define qu_sub_is_abstract(self) qu_flag_test (self, QU_SUB_FLAG_ABSTRACT)


/** qu_sub_is_vararg
	Whether the last argument of the Sub is a variable argument.

	@xmp
	sub foo (a, b, *c)
	               |
	               +--- makes foo qu_sub_is_vararg
*/
#define qu_sub_is_vararg(self) qu_flag_test (self, QU_SUB_FLAG_VARARG)


/** qu_sub_is_argmin
	Whether the Sub has default arguments, thus the field 'argc' is the minimum
	number of arguments that must be passed to the Sub.

	@xmp
	sub foo (a, b = 5, c = true)
	              |
	              +--- makes foo qu_sub_is_argmin (argc = 1)
*/
#define qu_sub_is_argmin(self) qu_flag_test (self, QU_SUB_FLAG_ARGMIN)


/** qu_sub_need_arg_check
	Whether the one or more arguments are typed, i.e. the Sub needs to validate
	arguments passed to it.

	@xmp
	sub foo (a: qu_boolean, b, c)
	          |
	          +--- makes foo qu_sub_need_arg_check
*/
#define qu_sub_need_arg_check(self) qu_flag_test (self, QU_SUB_FLAG_ARGCHECK)


/** qu_sub_is_cfunc
	Whether or not calling the C function Sub must not pass self as the first
	argument.

	@xmp
		static qu_int Foo_bar (qu_any self, qu_string a, qu_array b) <- no
		static qu_int Foo_bar (qu_string a, qu_array b)              <- yes
*/
#define qu_sub_is_cfunc(self) qu_flag_test (self, QU_SUB_FLAG_CFUNC)


/** qu_sub_is_closure
	Whether this is a run-time created closure.

	@xmp
		sub foo (x)
			return lambda (x) x * 2   <-
*/
#define qu_sub_is_closure(self) qu_flag_test (self, QU_SUB_FLAG_CLOSURE)


/** qu_sub_noauto
	Whether calling requires parentheses.
*/
#define qu_sub_noauto(self) qu_flag_test (self, QU_SUB_FLAG_NOAUTO)


/** qu_sub_is_generator
	Whether the Sub issues k{yield} or is tagged __generator.

	@xmp
		sub foo (x)
			yield x

		sub foo (x) __generator
*/
#define qu_sub_is_generator(self) qu_flag_test (self, QU_SUB_FLAG_GENERATOR)


/** qu_sub_is_accum
	Whether this is an accumulator.

	An accummulator Sub preserves the value of it's lexical over calls.

	@xmp
		sub make (x)
			return lambda (z) __accumulator x + z
		;;
*/
#define qu_sub_is_accum(self) qu_flag_test (self, QU_SUB_FLAG_ACCUM)


/** qu_sub_is_ensure
	Whether this is an k{ensure} method.

	@xmp
		var foo
			ensure
				...
			;;

	@warn
	The QU_SUB_FLAG_ENSURE flag must always be coupled with QU_SUB_FLAG_PRIVATE.
*/
#define qu_sub_is_ensure(self) qu_flag_test (self, QU_SUB_FLAG_ENSURE)


/** qu_sub_is_once
	Whether the Sub may only be called once.

	If it is then after the first call the field v{value} will have the result
	and subsequent calls will just return this result.

	@xmp
		sub foo (x) __once
			return x
		;;

	@p
	Note that C functions may also set this flag.
*/
#define qu_sub_is_once(self) qu_flag_test (self, QU_SUB_FLAG_ONCE)


/** qu_sub_is_unsafe
	Whether this is an __unsafe Sub.

	If it is, the compiler generates some extra prolog code to make sure that
	the caller does not come from an untrusted module.

	@xmp
		sub foo (x) __unsafe
*/
#define qu_sub_is_unsafe(self) qu_flag_test (self, QU_SUB_FLAG_UNSAFE)


/** qu_sub_is_jit
	Whether the Sub is JIT-ed, i.e. the code is compiled into native code
	thus the field v{value} is the native code and must be ={qu_mem_free} when
	the Sub is destroyed.
*/
#define qu_sub_is_jit(self) qu_flag_test (self, QU_SUB_FLAG_JIT)



/** qu_sub_new
	Create a new Sub.

	This is called by the compiler, ={qu_class_method} etc.

	@warn
	Do not call directly unless you know exactly what you are doing.
*/
extern qu_sub qu_sub_new (qu_symbol block, qu_string name, qu_any value, int argc, long flags);


/** qu_sub_reindex qu_sub_enddecl qu_sub_set_once
	For the compiler, to reindex locals and to finalize declaration.

	For qu_sub_reindex, the Sub must be the root Sub.

	qu_sub_enddecl frees all local Var objects and only keep the types,
	if any, in the v{args} field.

	qu_sub_set_once sets the necessary __once properties.
	It is called by the interpreter.
*/
extern void qu_sub_reindex  (qu_sub self);
extern void qu_sub_enddecl  (qu_sub self);
extern void qu_sub_set_once (qu_sub self, qu_any result);


/** qu_sub_jit
	These are for the JIT engine.

	See ={qu_code_jit} for more info.
*/
extern void qu_sub_jit_set        (qu_sub self);
extern void qu_sub_jit_set_extern (qu_sub self);


/** qu_sub_get_module
	Get the ={Module} object where the Sub is declared.
*/
extern qu_module qu_sub_get_module (qu_sub self);


/** qu_sub_get_class
	Get the ={Class} object where the Sub is declared.

	If the Sub is a function (not a method, i.e. not in a class block)
	then 0 is returned, no exception thrown.
*/
extern qu_class qu_sub_get_class (qu_sub self);


/** qu_sub_isroot qu_sub_isparent
	Whether a Sub is a nested Sub or the parent of a Sub.
*/
extern int qu_sub_isroot   (qu_sub self);
extern int qu_sub_isparent (qu_sub self, qu_sub p);


/** qu_sub_new_closure
	For the interpreter, to create a new closure.
*/
extern qu_sub qu_sub_new_closure (qu_sub u, qu_frame f);


/** qu_sub_getattr qu_sub_setattr
	qu_sub_getattr find a local symbol in the Sub.

	qu_sub_setattr sets a local symbol in the Sub where v{thing} must be a
	Sub or a ={Var} object.

	This is designed for the compiler.
	Do not call unless you know exactly what you are doing.
*/
extern qu_symbol qu_sub_getattr (qu_sub self, qu_string name);
extern void      qu_sub_setattr (qu_sub self, qu_string name, qu_symbol attr);


/** qu_sub_gettag qu_sub_settag
	qu_sub_gettag returns a value in the C{__tag} property.
	Returns 0 if none.

	qu_sub_settag set a value in the C{__tag} property.

	qu_sub_gettag_dict return all in a new Dict.
	Return 0 if none.
*/
extern qu_any  qu_sub_gettag      (qu_sub self, qu_string name);
extern void    qu_sub_settag      (qu_sub self, qu_string name, qu_any value);
extern qu_dict qu_sub_gettag_dict (qu_sub self);


/** qu_sub_arg_check
	Checks for valid arguments.

	If v{runtime} is true then validators are actually called. Otherwise
	only a type match is performed (e.g. used by the compiler).
*/
extern qu_sub qu_sub_arg_check (qu_sub self, int argc, qu_any *argv, int runtime);


/** qu_sub_as
	Various representation.
*/
#define qu_sub_as_string(self)  qu__sub__string (self)
#define qu_sub_as_boolean(self) (QuTrue)


/** QU_RETURN
	Use this instead of C{return x} to override the function name in the exception
	message to the name of this function.

	As a rule of thumb, always use QU_RETURN instead of k{return} unless you
	know exactly that the called function never fails.

	@xmp
		static qu_int _do_something ()
		{
			...
			QU_THROW (EFoo);
				--> sets thread->eblock to "_do_something"
			...
			return QuInt0;
		}

		static qu_int Foo_bar (qu_string baz)
		{
			...
			QU_RETURN (_do_something ());
				--> if _do_something throws the this will override
				    thread->eblock to "Foo_bar" then return 0
		}

	@p
	If you are certain that C{_do_something} never fails then you can simply
	use k{return} instead of QU_RETURN.
*/
#define QU_RETURN(x) \
QU_BEGIN \
	qu_any _rEtvAl_ = x; \
	QU_ENSURE (_rEtvAl_); \
	return _rEtvAl_; \
QU_END


/** qu_sub_call
	Convenient macros to call a qu_sub (function or method).
*/
#define qu_sub_calln(self, theself, argc, argv) \
	((qu_sub) self)->call->fn ((qu_sub) self, theself, argc, argv)

#define qu_sub_call4(self, theself, a1, a2, a3, a4) \
	((qu_sub) self)->call->f4 ((qu_sub) self, theself, a1, a2, a3, a4)

#define qu_sub_call3(self, theself, a1, a2, a3) \
	((qu_sub) self)->call->f3 ((qu_sub) self, theself, a1, a2, a3)

#define qu_sub_call2(self, theself, a1, a2) \
	((qu_sub) self)->call->f2 ((qu_sub) self, theself, a1, a2)

#define qu_sub_call1(self, theself, a1) \
	((qu_sub) self)->call->f1 ((qu_sub) self, theself, a1)

#define qu_sub_call0(self, theself) \
	((qu_sub) self)->call->f0 ((qu_sub) self, theself)


/***/
/*----------------------------------------------------------------------------
	dealing with arguments (for C functions/methods)
----------------------------------------------------------------------------*/

/** qu_arg
	Various macros to deal with function/method arguments.

	See a lot of examples in the source code.
*/
#define QU_ARG_CHECK(expr)            QU_REQUIRE  (QuEArgValue, (expr))
#define QU_ARG_CHECKZ(expr, msgz)     QU_REQUIREZ (QuEArgValue, (expr), msgz)
#define QU_ARG_CHECKF(expr, fmt, ...) QU_REQUIREF (QuEArgValue, (expr), fmt, __VA_ARGS__)

#define QU_ARG_THROW               QU_THROW  (QuEArgValue)
#define QU_ARG_THROWZ(msgz)        QU_THROWZ (QuEArgValue, msgz)
#define QU_ARG_THROWF(fmt, ...)    QU_THROWF (QuEArgValue, fmt, __VA_ARGS__)

#define QU_ARG_TYPE(x, type) QU_ARG_CHECK (type##_is (x))

#define QU_ARG_bool(x, v)       QU_BEGIN if (!qu_boolean_is (x)) { qu_arg__throw_boolean (__FUNCTION__); return 0; } v = (((qu_boolean) (x)) == QuTrue); QU_END;
#define QU_ARG_byte(x, v)       QU_BEGIN v = qu_integer_tob   (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_ubyte(x, v)      QU_BEGIN v = qu_integer_toub  (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_ilong(x, v)      QU_BEGIN v = qu_integer_toli  (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_uilong(x, v)     QU_BEGIN v = qu_integer_touli (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_short(x, v)      QU_BEGIN v = qu_integer_toh   (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_ushort(x, v)     QU_BEGIN v = qu_integer_touh  (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_int(x, v)        QU_BEGIN v = qu_integer_toi   (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_uint(x, v)       QU_BEGIN v = qu_integer_toui  (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_long(x, v)       QU_BEGIN v = qu_integer_tol   (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_ulong(x, v)      QU_BEGIN v = qu_integer_toul  (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_float(x, v)      QU_BEGIN v = qu_real_tof      (x); if (errno) { qu_arg__throw_float   (__FUNCTION__); return 0; } QU_END
#define QU_ARG_double(x, v)     QU_BEGIN v = qu_real_tod      (x); if (errno) { qu_arg__throw_float   (__FUNCTION__); return 0; } QU_END
#ifdef HAVE_LONG_LONG
#define QU_ARG_longlong(x, v)   QU_BEGIN v = qu_integer_toll  (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#define QU_ARG_ulonglong(x, v)  QU_BEGIN v = qu_integer_toull (x); if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } QU_END
#endif
#ifdef HAVE_LONG_DOUBLE
#define QU_ARG_longdouble(x, v) QU_BEGIN v = qu_real_told     (x); if (errno) { qu_arg__throw_float   (__FUNCTION__); return 0; } QU_END
#endif

#define QU_ARG_record(x, v, strct) \
QU_BEGIN \
	v = x; \
	QU_ARG_CHECK (qu_record_of (v, strct)); \
QU_END

/* this only works with integer values */
#define QU_ARG_enum(x, v, enu) \
QU_BEGIN \
	assert (qu_var_is (enu) && qu_var_is_enum (enu)); \
	assert (qu_dict_is ((enu)->value)); \
	v = qu_integer_toi (x); \
	if (errno) { qu_arg__throw_integer (__FUNCTION__); return 0; } \
	if (!qu_dict_getitem ((qu_dict) (enu)->value, x)) { qu_arg__throw_enum (__FUNCTION__, enu); return 0; } \
QU_END

#define QU_ARG_VCOUNT(min, max)   QU_BEGIN if ((argc) < (min) || (argc) > (max)) { qu_arg__throw_count (__FUNCTION__, argc, min, max); return 0; } QU_END
#define QU_ARG_VEXACT(count)      QU_BEGIN if ((argc) != (count)) { qu_arg__throw_exact (__FUNCTION__, argc, count); return 0; } QU_END
#define QU_ARG_VMIN(min)          QU_BEGIN if ((argc) < (min)) { qu_arg__throw_min (__FUNCTION__, argc, min); return 0; } QU_END
#define QU_ARG_VMAX(max)          QU_BEGIN if ((argc) > (max)) { qu_arg__throw_max (__FUNCTION__, argc, max); return 0; } QU_END

#define QU_ARG_VTYPE(i, v, type)   QU_BEGIN v = argv[i]; QU_ARG_TYPE (v, type); QU_END

#define QU_ARG_vbool(i, v)         QU_ARG_bool       (argv[i], v)
#define QU_ARG_vbyte(i, v)         QU_ARG_byte       (argv[i], v)
#define QU_ARG_vubyte(i, v)        QU_ARG_ubyte      (argv[i], v)
#define QU_ARG_vilong(i, v)        QU_ARG_ilong      (argv[i], v)
#define QU_ARG_vuilong(i, v)       QU_ARG_uilong     (argv[i], v)
#define QU_ARG_vshort(i, v)        QU_ARG_short      (argv[i], v)
#define QU_ARG_vushort(i, v)       QU_ARG_ushort     (argv[i], v)
#define QU_ARG_vint(i, v)          QU_ARG_int        (argv[i], v)
#define QU_ARG_vuint(i, v)         QU_ARG_uint       (argv[i], v)
#define QU_ARG_vlong(i, v)         QU_ARG_long       (argv[i], v)
#define QU_ARG_vulong(i, v)        QU_ARG_ulong      (argv[i], v)
#define QU_ARG_vfloat(i, v)        QU_ARG_float      (argv[i], v)
#define QU_ARG_vdouble(i, v)       QU_ARG_double     (argv[i], v)
#ifdef HAVE_LONG_LONG
#define QU_ARG_vlonglong(i, v)     QU_ARG_longlong   (argv[i], v)
#define QU_ARG_vulonglong(i, v)    QU_ARG_ulonglong  (argv[i], v)
#endif
#ifdef HAVE_LONG_DOUBLE
#define QU_ARG_vlongdouble(i, v)    QU_ARG_longdouble (argv[i], v)
#endif
#define QU_ARG_venum(i, v, enu)     QU_ARG_enum       (argv[i], v, enu)
#define QU_ARG_vrecord(i, v, strct) QU_ARG_record     (argv[i], v, strct)
/***/
extern void qu_arg__throw_boolean (const char *cfunc);
extern void qu_arg__throw_integer (const char *cfunc);
extern void qu_arg__throw_float   (const char *cfunc);
extern void qu_arg__throw_enum    (const char *cfunc, qu_var v);
extern void qu_arg__throw_count   (const char *cfunc, int argc, int min, int max);
extern void qu_arg__throw_min     (const char *cfunc, int argc, int min);
extern void qu_arg__throw_max     (const char *cfunc, int argc, int max);
extern void qu_arg__throw_exact   (const char *cfunc, int argc, int n);


/** qu_arg_validate
	Use these macros to validate all arguments passed in v{argv}.
*/
extern int qu_arg_validate (qu_any type, int argc, qu_any *argv);


/** qu_arg_PARSE
	Parse arguments.

	Throw EArgValue if one or more fields do not match, EArgCount
	if the number of fields is less than required.

	@warn
	You must name the arguments v{argc}, v{argv} and v{argx}.
	Do not invent new names or the macros will not work.

	@xmp
		class Foo
			sub bar (how: Dict , nocase = false) -> Boolean


		qu_boolean Foo_bar (Foo self, int argc, qu_any *argv)
		{
			static char *argx[] = {"D", "how", ":", "b", "nocase", NULL};

			qu_dict how;
			int     nocase = 0;

			QU_ARG_XPARSE (&how, &nocase);

			...
			...
		}

	@p
	Do not call the functions directly.
	Use the macro instead.

	Note that the list is similiar as ={qu_struct_type_to_char}.

	@table
	.| : | -        | start optional arguments
	.| a | String   | char
	.| y | Integer  | int (-127 .. 127)
	.| Y | Integer  | int (0 .. 256)
	.| b | Boolean  | int (0 or 1)
	.| h | Integer  | short
	.| H | Integer  | unsigned short
	.| q | Integer  | long (QU_INT_MIN .. QU_INT_MAX)
	.| i | Integer  | int
	.| I | Integer  | unsigned int
	.| j | Integer  | int16
	.| J | Integer  | unsigned int16
	.| k | Integer  | int32
	.| K | Integer  | unsigned int32
	.| m | Integer  | int64
	.| M | Integer  | unsigned int64
	.| l | Integer  | long
	.| L | Integer  | unsigned long
	.| o | Integer  | long long
	.| O | Integer  | unsigned long long
	.| g | Real     | float
	.| f | Real     | double
	.| F | Real     | long double
	.| z | Integer  | ssize_t
	.| Z | Integer  | size_t
	.| e | Integer  | off_t
	.| p | Pointer  | void *
	.| s | String   | String   (not converted)
	.| A | Array    | Array    (not converted)
	.| D | Dict     | Dict     (not converted)
	.| E | Set      | Set      (not converted)
	.| R | Range    | Range    (not converted)
	.| C | Class    | Class    (not converted)
	.| B | Buffer   | Buffer   (not converted)
	.| Q | Queue    | Queue    (not converted)
	.| T | Stream   | Stream   (not converted)
	.| P | Proc     | Proc     (not converted)
	.| U | Function | Function (not converted)
	.| v | Sub      | Sub      (not converted)
	.| w | Method   | Method   (not converted)
	.| x | Any      | Any      (not converted)
*/
extern int qu_arg_vparse (const char *funcname, const char *spec, int argc, qu_any *argv, ...);
#define QU_ARG_VPARSE(spec, ...) \
	QU_ENSURE (qu_arg_vparse (__FUNCTION__, spec, argc, argv, __VA_ARGS__))

extern int qu_arg_xparse (const char *funcname, char **argx, int argc, qu_any *argv, ...);
#define QU_ARG_XPARSE(...) \
	QU_ENSURE (qu_arg_xparse (__FUNCTION__, argx, argc, argv, __VA_ARGS__))

extern int qu_arg_dparse (const char *funcname, char **argx, qu_dict argd, ...);
#define QU_ARG_DPARSE(...) \
QU_BEGIN \
	QU_ARG_CHECK (qu_dict_is (argd)); \
	QU_ENSURE (qu_arg_dparse (__FUNCTION__, argx, argd, __VA_ARGS__)); \
QU_END


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_sub__setup (void);
extern void qu_sub__init  (void);

/* for efficiency - may only be called from within this file or Method.c */
extern qu_string qu__sub__string   (qu_sub self);
extern qu_any    qu__sub__getitem  (qu_sub self, qu_string name);
extern qu_any    qu__sub__getslice (qu_sub self, qu_string name, qu_any defval);
extern qu_any    qu__sub__setitem  (qu_sub self, qu_string name, qu_any val);
extern qu_code   qu__sub_code      (qu_sub self);
extern qu_int    qu__sub_argmatch  (qu_sub self, int argc, qu_any *argv);
extern qu_any    qu__sub_argnames  (qu_sub self);
extern qu_any    qu__sub_argtypes  (qu_sub self);

/* special Sub to simulate yield from C code - see Code */
extern qu_sub QuSubYield;

/* for Reduce/Emit */
extern int qu_sub_can_access (qu_sub self, qu_symbol from);

#ifdef QU_DEBUG_SEGV
extern void __qu_sub_trace      (qu_sub u);
extern void __qu_sub_trace_show ();
#endif

#endif /*_H_qu_sub_*/
