/* Sys.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_sys_
#define _H_qu_sys_ 1


/** QuSysModule
	This is the builtin Sys module.

	It contains all the builtin functions, operators and printing functions.
*/
extern qu_module QuSysModule;


/** qu_sys_warn
	Functions and macros to help keep warnings and error messages consistent
	across files.

	QU_SYS_WARN takes arguments as in the C f{printf} function.

	Warnings are printed to C{stderr} but only if ={QU_APP_NO_WARNING} is not
	set (Qu is invoked with the C{-W} option).
*/
#define QU_SYS_WARN(...) \
	QU_BEGIN if (!QU_APP_NO_WARNING) qu_sys__warn (__VA_ARGS__); QU_END

#define QU_SYS_WARN_CLASS(any, msgz) \
	QU_BEGIN if (!QU_APP_NO_WARNING) qu_sys__warn_class (any, msgz); QU_END

#define QU_SYS_WARN_RETVAL(any, namez, expectedz) \
	QU_BEGIN if (!QU_APP_NO_WARNING) qu_sys__warn_retval (any, namez, expectedz); QU_END

#define QU_SYS_EIMPLEMENT(any, namez) \
	QU_BEGIN qu_sys__eimplement (any, namez, __FUNCTION__); QU_RETHROW; QU_END


/** qu_sys_print qu_sys_printz qu_sys_printf qu_sys_sprintf qu_sys_printb qu_sys_printh
	These are the function to call from the __print methods.

	qu_sys_print prints any object. This is the function to call from C code
	to print a Qu object.

	qu_sys_printz is provided for efficiency. It is generally preferable over
	qu_sys_print if the argument is a immutable string (such as ={String}).

	qu_sys_printh does qu_sys_print but HTML escape the characters.

	qu_sys_printf and qu_sys_sprintf use our formatting standard, not of C.

	@warn
	An object must not call qu_sys_print to print itself otherwise we fall in
	a deadly infinite loop.

	qu_sys_printf and qu_sys_sprintf can only take QU_SYS_PRINT_MAXARG arguments max.

	qu_sys_printb prints with a buffer thus improve performance.
	See examples in the source code.

	@warn
	qu_sys_printb buffer size must be QU_SYS_PRINTB_BUFSIZ.
	It may not have any other value.
*/
#define QU_SYS_PRINT_MAXARG 256
extern qu_null   qu_sys_print   (qu_any x);
extern qu_null   qu_sys_printh  (qu_any x);
extern qu_null   qu_sys_printz  (const char *s, long len);
extern qu_null   qu_sys_printf  (qu_string format, int argc, ...);
extern qu_string qu_sys_sprintf (qu_string format, int argc, ...);

#define QU_SYS_PRINTB_BUFSIZ 1024

typedef struct qu_sys_printb_s qu_sys_printb_s;

struct qu_sys_printb_s
{
	char  buf [QU_SYS_PRINTB_BUFSIZ];
	char *ptr;
};
extern int qu_sys_print_buff (qu_sys_printb_s *b, const char *z, long n);

#define qu_sys_printb_init(b)      QU_BEGIN b.ptr = b.buf; QU_END
#define qu_sys_printb(b, z, n)     qu_sys_print_buff (&b, z, n)
#define qu_sys_printb_string(b, s) qu_sys_print_buff (&b, ((qu_string) s)->str, ((qu_string) s)->len)
#define qu_sys_printb_final(b)     ((b.ptr > b.buf) ? (qu_sys_printz (b.buf, b.ptr - b.buf) != 0) : 1)


/** qu_sys_binop
	Generic binary operators.

	qu_sys_binop is intended for the interpreter.
	You should have no reason to call it.
*/
extern qu_any     qu_sys_binop    (qu_any self, qu_any x, int op);
extern qu_boolean qu_sys_are      (qu_any self, qu_any x);


/** qu_sys_operator
	These are the builtin operators.

	Refer the I{operator} section of the manual for details.
*/
extern qu_any     qu_sys_as       (qu_any self, qu_any x);
extern qu_any     qu_sys_and      (qu_any self, qu_any x);
extern qu_boolean qu_sys_band     (qu_any self, qu_any x);
extern qu_boolean qu_sys_bor      (qu_any self, qu_any x);
extern qu_any     qu_sys_clear    (qu_any self);
extern qu_int     qu_sys_cmp      (qu_any self, qu_any x);
extern qu_string  qu_sys_concat   (qu_any self, qu_any x);
extern qu_any     qu_sys_copy     (qu_any self);
extern qu_any     qu_sys_decr     (qu_any self);
extern qu_int     qu_sys_del      (qu_any self, qu_any x);
extern qu_any     qu_sys_div      (qu_any self, qu_any x);
extern qu_boolean qu_sys_eq       (qu_any self, qu_any x);
extern qu_any     qu_sys_each     (qu_any self);
extern qu_any     qu_sys_format   (qu_any self, int argc, qu_any *argv);
extern qu_string  qu_sys_frmt     (qu_any self, qu_any x);
extern qu_boolean qu_sys_ge       (qu_any self, qu_any x);
extern qu_any     qu_sys_getitem  (qu_any self, qu_any x);
extern qu_any     qu_sys_getslice (qu_any self, qu_any i, qu_any len);
extern qu_boolean qu_sys_gt       (qu_any self, qu_any x);
extern qu_any     qu_sys_iand     (qu_any self, qu_any x);
extern qu_any     qu_sys_idiv     (qu_any self, qu_any x);
extern qu_boolean qu_sys_in       (qu_any self, qu_any x);
extern qu_any     qu_sys_incr     (qu_any self);
extern qu_any     qu_sys_inot     (qu_any self);
extern qu_any     qu_sys_ior      (qu_any self, qu_any x);
extern qu_boolean qu_sys_is       (qu_any self, qu_any x);
extern qu_any     qu_sys_ixor     (qu_any self, qu_any x);
extern qu_boolean qu_sys_le       (qu_any self, qu_any x);
extern qu_int     qu_sys_len      (qu_any self);
extern qu_boolean qu_sys_like     (qu_any self, qu_any x);
extern qu_any     qu_sys_lshift   (qu_any self, qu_any x);
extern qu_boolean qu_sys_lt       (qu_any self, qu_any x);
extern qu_any     qu_sys_minus    (qu_any self, qu_any x);
extern qu_any     qu_sys_mul      (qu_any self, qu_any x);
extern qu_boolean qu_sys_ne       (qu_any self, qu_any x);
extern qu_any     qu_sys_neg      (qu_any self);
extern qu_boolean qu_sys_not      (qu_any self);
extern qu_any     qu_sys_of       (qu_any self, qu_any x);
extern qu_any     qu_sys_or       (qu_any self, qu_any x);
extern qu_any     qu_sys_plus     (qu_any self, qu_any x);
extern qu_any     qu_sys_pop      (qu_any self);
extern qu_any     qu_sys_pos      (qu_any self);
extern qu_any     qu_sys_pow      (qu_any self, qu_any x);
extern qu_any     qu_sys_push     (qu_any self, qu_any x);
extern qu_any     qu_sys_range    (qu_any self, qu_any x);
extern qu_any     qu_sys_ratio    (qu_any self, qu_any x);
extern qu_any     qu_sys_rem      (qu_any self, qu_any x);
extern qu_any     qu_sys_rep      (qu_any self, qu_any x);
extern qu_boolean qu_sys_rin      (qu_any self, qu_any x);
extern qu_any     qu_sys_rshift   (qu_any self, qu_any x);
extern qu_any     qu_sys_setitem  (qu_any self, qu_any x, qu_any v);
extern qu_any     qu_sys_setslice (qu_any self, qu_any from, qu_any len, qu_any v);
extern qu_boolean qu_sys_true     (qu_any self);
extern qu_string  qu_sys_string   (qu_any self);
extern qu_any     qu_sys_xchg     (qu_any self, qu_any x);


/** qu_sys_binop_get
	Get the binop function for a specified operator name.

	This is used, for example, by the builtin sysmap function when passed
	an operator name.
*/
extern qu_binop_f qu_sys_binop_get (qu_string op);


/** Format
	qu_format_s is used for formatting objects.

	qu_format_parse parses a format. It returns non-zero on success, throws EFormat
	and return 0 on failure.

	qu_format_ptr handles the 'p' mode format.
	You must always use this function instead of providing your own.

	qu_format_yesno handles the 'y' mode format.
	You must always use this function instead of providing your own.

	qu_format_string handles the 's' mode format.
	You must always use this function instead of providing your own.

	qu_format_number does the final arrangements. The number formatter must provide
	the integral and fractional part, whether the number is negative, and
	the exponent value.

	QU_FORMAT_MAX_WIDTH is the maximum allowed width of a formatted string.
	We consider the maximum length of a string and a "reasonable" value.
	We also make it possible to check for the maximum value I{after} a calculation
	such as: C{width = width * 10 + (*i - '0')}

	qu_format_align does the final alignment.

	Note that for the exponent, 'E' and 'e' must always be used. All number
	parsers and formatters use this convention.
*/
typedef struct qu_format_s qu_format_s;
struct qu_format_s
{
	char mode;    /* mode: 'g', 'f', etc. */
	int  align;   /* 0: align right, otherwise: align left */
	int  sign;    /* 0:"+"/"-" 1:"-" 2:" "/"-" -1:dont print sign */
	int  zero;    /* pad 0 on front */
	int  group;   /* number of characters to group, -1 means none */
	int  width;   /* total available space */
	int  prec;    /* number of decimal digits */
	int  ucase;   /* use uppercase letters */
	char sep;     /* group separator character */
	char dec;     /* decimal point character */
};

#define QU_FORMAT_MAX_WIDTH  (INT_MAX / 256)

extern int       qu_format_parse  (const char *fmt, qu_format_s *f);
extern qu_string qu_format_ptr    (qu_format_s *f, qu_any x);
extern qu_string qu_format_string (qu_format_s *f, qu_any x);
extern qu_string qu_format_yesno  (qu_format_s *f, qu_any x);
extern qu_string qu_format_number (qu_format_s *f, qu_string s, int neg, int expo);
extern qu_string qu_format_align  (qu_format_s *f, qu_string s);


/** qu_sys__arg
	The global C{__arg__} variable is set by the main program from the command line.
	It contains an ={Array} of ={String} objects. You must never modify the
	content.

	For example, if Qu is invoked as: C{qu foo.qu -a bc 1 2} then C{__arg__} will
	be C{['-a', 'bc', '1', '2']}.
*/
#define qu_sys__arg() (QuSys__arg__)
/***/
extern qu_var QuSys__arg__;


/** qu_sys_debugging
	qu_sys_debugging checks whether the value of __debug__ is true.

	QuSys__debug__ is the variable that holds the value and must never be
	accessed directly.
*/
#define qu_sys_debugging() (QuSys__debug__ && QuSys__debug__->value == QuTrue)
/***/
extern qu_var QuSys__debug__;


/** QuSysTrace QuSys__tracefilter__ QuSys__tracehow__
	QuSysTrace determines whether or not exceptions should be displayed.

	QuSys__tracefilter__ holds the list of exception classes that should or should
	not be displayed when trace is on. The value is a ={Set}.

	If QuSys__tracehow__ is QuFalse, then display those exceptions.

	This is strictly used by the interpreter. You must not access any of
	these symbols.
*/
extern int volatile QuSysTrace;
extern qu_var QuSys__tracefilter__;
extern qu_var QuSys__tracehow__;


/** QuSys__ignore
	The global __ read/write variable.

	This is specially used by for-in loops and strictly used by the interpreter.
	You must never access this symbol.
*/
extern qu_var QuSys__ignore;


/** QuSys__keyboard__
	This variable holds the keyboard interrupt handler (a static Sub)
	or nil if none is set.

	This is strictly used by the interpreter.
	You must never access this symbol.
*/
extern qu_var QuSys__keyboard__;


/** QuSys__quit__
	This variable holds the QuEExit exception handler (a static Sub)
	or nil if none is set.

	This is strictly used by the interpreter.
	You must never access this symbol.
*/
extern qu_var QuSys__quit__;


/** qu_sys_builtins
	These are the script level builtin functions.

	They are meant to be called by core modules only.
*/
#define qu_sys_raise(argc, argv) qu__sys_raise (argc, argv)
#define qu_sys_filter(x, ap)     qu__sys_sysfilter (x, ap)
#define qu_sys_map(x, ap)        qu__sys_sysmap (x, ap)
#define qu_sys_cat(argc, argv)   qu__sys_syscat (argc, argv)


/** __qu_buffer__
	Interface to the builtin functions __buffer__, __clear__, __trash__, and
	__collect__.
*/
#define __qu_buffer__  qu__sys__buffer__()
#define __qu_clear__   qu__sys__clear__()
#define __qu_trash__   qu__sys__trash__()
#define __qu_collect__ qu__sys__collect__()


/** __qu_err__
	Interface to the builtin functions __err__ and __errmsg__.
*/
#define __qu_err__     qu__sys__err__ ()
#define __qu_errmsg__  qu__sys__errmsg__ ()


/** QuType
	These are the C type validators.
*/
extern qu_sub QuType_char;
extern qu_sub QuType_byte;
extern qu_sub QuType_ubyte;
extern qu_sub QuType_short;
extern qu_sub QuType_ushort;
extern qu_sub QuType_int;
extern qu_sub QuType_uint;
extern qu_sub QuType_int16;
extern qu_sub QuType_uint16;
extern qu_sub QuType_int32;
extern qu_sub QuType_uint32;
extern qu_sub QuType_long;
extern qu_sub QuType_ulong;
extern qu_sub QuType_longlong;
extern qu_sub QuType_ulonglong;
extern qu_sub QuType_float;
extern qu_sub QuType_double;
extern qu_sub QuType_longdouble;
extern qu_sub QuType_size_t;
extern qu_sub QuType_ssize_t;
extern qu_sub QuType_off_t;
extern qu_sub QuType_char_p;
extern qu_sub QuType_void_p;
#ifdef QU_SUPPORT_GLIB
extern qu_sub QuType_GObject;
#endif


/** qu_sys_sort
	Use this as a replacement of qsort.
*/
/* comparison function prototype */
typedef int (*qu_sys_osort_f) (const void *i, const void *j, qu_any obj, qu_any self);
typedef int (*qu_sys_sort_f ) (const void *i, const void *j);

extern int qu_sys_osort (qu_sys_osort_f f, void *p, long size, long count, qu_any obj, qu_any self);
extern int qu_sys_sort  (qu_sys_sort_f  f, void *p, long size, long count);

/* possible return value of qu_sys_sort_f */
#define QU_SYS_SORT_ERROR   -2
#define QU_SYS_SORT_SMALLER -1
#define QU_SYS_SORT_EQUAL    0
#define QU_SYS_SORT_GREATER  1

/* maximum size of item that can be sorted */
#define QU_SYS_SORT_MAXSIZE SIZEOF_DOUBLE

/* return a sort result returned by a call to a Qu function/method */
#define QU_SYS_SORT_RETURN_RESULT(res) \
QU_BEGIN \
	if ((res) == 0) \
		return QU_SYS_SORT_ERROR; \
	else \
	{ \
		long vAl = qu_integer_toi (res); \
		if (errno) \
		{ \
			qu_thread_throw (QuECmpClass, 0); \
			return QU_SYS_SORT_ERROR; \
		} \
		return (vAl < 0) ? QU_SYS_SORT_SMALLER : ((vAl > 0) ? QU_SYS_SORT_GREATER : QU_SYS_SORT_EQUAL); \
	} \
QU_END


/***/
/*--------------------------------------------------------------------------*/

/* called internally by the main program to initialize the Sys module */
extern void qu_sys__init ();

/* for efficiency - may only be called from within this file */
extern void       qu_sys__warn        (const char *fmt, ...);
extern void       qu_sys__warn_class  (qu_any self, const char *msg);
extern void       qu_sys__warn_retval (qu_any self, const char *name, const char *expected);
extern void       qu_sys__eimplement  (qu_any self, const char *name, const char *func);
extern qu_null    qu__sys__buffer__   ();
extern qu_null    qu__sys__clear__    ();
extern qu_string  qu__sys__collect__  ();
extern qu_null    qu__sys__trash__    ();
extern qu_any     qu__sys__err__      ();
extern qu_any     qu__sys__errmsg__   ();
extern qu_any     qu__sys_raise       (int argc, qu_any *argv);
extern qu_boolean qu__sys_sysfilter   (qu_any x, qu_any ap);
extern qu_any     qu__sys_sysmap      (qu_any x, qu_any ap);
extern qu_string  qu__sys_syscat      (int argc, qu_any *argv);

/*HACK - documentation */
extern qu_sub __QuSys__undocumented__;

/* for Compile/Emit/interp */
extern qu_sub __QuSys_syscat;
extern qu_sub __QuSys__err__;

#endif /*_H_qu_sys_*/
