/* Thread.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_thread_
#define _H_qu_thread_ 1


/** Thread Mutex Cond
	The final Thread, Mutex, and Cond classes.

	All program level methods are not published since you should never use them
	from C code.
*/
extern qu_class QuThreadClass;
extern qu_class QuMutexClass;
extern qu_class QuCondClass;


/** qu_thread
	qu_thread represents an instance of the Thread class.
	It is a pointer to the instance struct qu_thread_s.
*/
struct qu_thread_s
{
	QU_OBJECT_HEAD;
	qu_thread_t tid;      /* qu_thread ID */
	int         ticks;    /* For the scheduler, see interp.c */
	int         depth;    /* To check recursive function call limit */
	long        atomic;   /* __qu_atomic block counter, initially 0 */

	/* These are zeroed once the thread runs */
	qu_array    args;     /* Args to pass to the start function, initally 0 */
	qu_sub      sub;      /* Start function, initally 0 */

	/* Internally used by the builtin functions buffer/collect/clear */
	qu_array    buffers;  /* qu_array of qu_string objects, initally 0 */

	/* These are set everytime an exception is thrown.
		When a joinable thread dies with an exception, they are used to rethrow
		the exception when the method 'join' is called.

		!!! If ewhere is a C function name, it must be __FUNCTION__.
		!!! It may not be a dynamic/allocated string.
	*/
	qu_class    eclass; /* exception class */
	qu_any      emsg;   /* the message, may be 0 */
	void       *ewhere; /* the qu_module, qu_sub or qu_class, or (char *) C function name */
	long        eline;  /* the line number */

	/* This is the result set by qu_interp_run_thread. If an error occured then
		eclass etc. are set and this is 0. This is the result returned
		by the 'join' method. */
	qu_any      result;

	/* To mark the C stack */
	qu_any     *stacklo;
	qu_any     *stackhi;

	/* thread state - managed internally, don't touch! */
	qu_array    state;

	/* User data - a qu_dict or 0 */
	qu_dict     data;

	/* to support framework applications - don't touch! */
	qu_array    extra;

	/* to support the html module - don't touch! */
	qu_any      html;
};


/** QU_THREAD_FLAG
	Flags used by Thread objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_THREAD_FLAG_ALIVE     QU_FLAG_0  /* alive? */
#define QU_THREAD_FLAG_DETACHED  QU_FLAG_1  /* in detached mode? */
#define QU_THREAD_FLAG_EFUNCTION QU_FLAG_2  /* ewhere is a char * instead of an object */
#define QU_THREAD_FLAG_BLOCKING  QU_FLAG_3  /* thread is blocking? */
/***/
#define QU_THREAD_FLAG_EXTERROR  QU_FLAG_9  /* to support stupid globalish libraries */


/** qu_thread_blocking
	These are strictly for use by the GC and the interpreter.

	It is used to assert that the thread is not in blocking mode when
	allocating/freeing memory and used to verify the qu_gc_lock status.
	Thus they are only meaningful in DEBUG mode.
*/
#define qu_thread_blocking(self)       qu_flag_test  (self, QU_THREAD_FLAG_BLOCKING)
#define qu_thread_set_blocking(self)   qu_flag_set   (self, QU_THREAD_FLAG_BLOCKING)
#define qu_thread_unset_blocking(self) qu_flag_unset (self, QU_THREAD_FLAG_BLOCKING)


/** qu_thread_is_efunction qu_thread_set_efunction qu_thread_set_eobject
	These are strictly for use by the interpreter.

	If the QU_THREAD_FLAG_EFUNCTION flag is set then it means that the field
	thread->ewhere is a C function name thus a C{char *} that comes from
	__FUNCTION__. Otherwise it is a Module, Class or Sub object.
*/
#define qu_thread_is_efunction(self)  qu_flag_test  (self, QU_THREAD_FLAG_EFUNCTION)
#define qu_thread_set_efunction(self) qu_flag_set   (self, QU_THREAD_FLAG_EFUNCTION)
#define qu_thread_set_eobject(self)   qu_flag_unset (self, QU_THREAD_FLAG_EFUNCTION)


/** qu_thread_state
	Use qu_thread_state to create a new thread state.

	This is used for printing or constructing strings of containers that may
	be recursive such as Array, etc.

	Acceptable keys are one of QU_THREAD_STATE_XXX.
*/
extern void qu_thread_state_begin (qu_string key, qu_any value);
extern int  qu_thread_state_exist (qu_string key, qu_any value);
extern void qu_thread_state_end   (qu_string key, qu_any value);

#define __qu_thread_state_begin(k, v) \
QU_BEGIN \
	qu_any tHrEaDkEy   = k; \
	qu_any tHrEaDvAlUe = v; \
	qu_thread_state_begin (k, v)

#define __qu_thread_state_end \
	qu_thread_state_end (tHrEaDkEy, tHrEaDvAlUe); \
QU_END

#define QU_THREAD_STATE_PRINT      qu_string_byte ('p')  /* __print */
#define QU_THREAD_STATE_STRING     qu_string_byte ('s')  /* __string */
#define QU_THREAD_STATE_JOIN       qu_string_byte ('j')  /* join */
#define QU_THREAD_STATE_JOIN_KEY   qu_string_byte ('k')  /* join key (qu_dict only) */
#define QU_THREAD_STATE_JOIN_VALUE qu_string_byte ('v')  /* join value (qu_dict only) */


/** qu_thread_main
	This is the main thread.
*/
#define qu_thread_main() (__QuThreadMain)
/***/
extern qu_thread __QuThreadMain;


/** qu_thread_me
	This is the thread which currently holds the ={qu_gc_lock}.
	It is set by the interpreter.
*/
#define qu_thread_me() (__QuThreadCurrent)
/***/
extern qu_thread volatile __QuThreadCurrent;


/** qu_thread_self
	Get the Thread object of this thread.

	@warn
	If this thread is currently ={__qu_blocking} then qu_thread_self will
	return a different value than ={qu_thread_me}.

	At Qu initialization time, this returns 0 because the main thread
	={qu_thread_main} is not yet created.
*/
#define qu_thread_self() ((qu_thread) qu_thread_specific_get (__QuThreadKey))
/***/
extern qu_thread_key_t __QuThreadKey;



/** __QuThreadMulti
	Initially 0, set to 1 once another thread is created.

	This should help speed-up single threaded applications because
	they don't have to do a context switch.

	This is strictly used by the interpreter.
	You must not access this variable.
*/
extern int volatile __QuThreadMulti;


/** qu_thread_is qu_mutex_is qu_cond_is
	Whether or not the object is a Thread, Mutex, or Cond.
*/
#define qu_thread_is(o) (qu_classof(o) == QuThreadClass)
#define qu_mutex_is(o)  (qu_classof(o) == QuMutexClass)
#define qu_cond_is(o)   (qu_classof(o) == QuCondClass)


/** qu_thread_data
	Set or get user data.
*/
extern qu_any qu_thread_data_get (qu_thread self, qu_any key);
extern qu_any qu_thread_data_set (qu_thread self, qu_any key, qu_any val);


/** __qu_atomic
	Use __qu_atomic_begin and __qu_atomic_end to start and end an atomic state.

	@warn
	You must not use these within ={__qu_blocking}.
*/
#define __qu_atomic_begin QU_BEGIN __QuThreadCurrent->atomic++; QU_END
#define __qu_atomic_end   QU_BEGIN __QuThreadCurrent->atomic--; QU_END


/** qu_thread_throw
	Throwing exceptions from C code.

	v{kls} must be an exception Class or NULL. NULL means the Exception class.

	v{msg} may be 0 which means to use the class default message.

	v{cfunc} must be __FUNCTION__. Never call qu_thread__throw and
	qu_thread__rethrow directly. Use the macros instead.

	Make sure v{obj} is either a Module, Class or Sub.
*/
extern void qu_thread__throw   (qu_class kls, qu_any msg, const char *cfunc);
extern void qu_thread__rethrow (const char *cfunc);
extern void qu_thread_throwex  (qu_class kls, qu_any msg, qu_any obj, long line);

#define qu_thread_throw(kls,msg) qu_thread__throw (kls, msg, __FUNCTION__)
#define qu_thread_rethrow()      qu_thread__rethrow (__FUNCTION__)

#define qu_thread_set_errmsgz(msgz) \
QU_BEGIN \
	if (__QuThreadCurrent) \
		__QuThreadCurrent->emsg = qu_string_newz (msgz); \
QU_END

#define qu_thread_set_errmsg(msg) \
QU_BEGIN \
	if (__QuThreadCurrent) \
		__QuThreadCurrent->emsg = msg; \
QU_END


/** qu_thread_err_string
	Get the last exception as a String.
*/
extern qu_string qu_thread_err_string ();


/** qu_thread_print_err qu_thread_print_trace
	Print the last exception.

	qu_thread_print_trace takes into account ={QuSysTrace}.
*/
extern void qu_thread_print_err   ();
extern void qu_thread_print_trace ();


/** qu_thread_err_is_estop
	Whether or not the last exception was QuEStop.
	This is useful for iterators.

	@warn
	Assumes __QuThreadCurrent thus do not use at Qu initialization time.
*/
#define qu_thread_err_is_estop() (__QuThreadCurrent && __QuThreadCurrent->eclass == QuEStopClass)


/** qu_thread_err_is
	Whether the last exception is a specified class.
*/
#define qu_thread_err_is(kls) (__QuThreadCurrent && __QuThreadCurrent->eclass == kls##Class)


/** qu_thread_check_external
	Do the best we can to test whether this thread is our thread.
	Return 1 if it is, 0 if not.
*/
extern int qu_thread_check_external ();


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_thread__setup ();
extern void qu_thread__init  ();

#endif /*_H_qu_thread_*/
