/* Tokenizer.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_tokenizer_
#define _H_qu_tokenizer_ 1


/** QU_TOKENIZER_MAXPEEK
	The maximum look-ahead value.
*/
#define QU_TOKENIZER_MAXPEEK 7


/** Tokenizer
	This is the Tokenizer class.

	A tokenizer object interpret a source code into series of tokens.
	You must supply an ASCII string as the source code.
*/
extern qu_class QuTokenizerClass;


/** qu_token_s
	This represents a token in a Tokenizer buffer.
*/
typedef struct qu_token_s qu_token_s;
struct qu_token_s
{
	qu_token_s *next;      /* Next token */
	int         type;      /* One of QU_T_XXX */
	const char *cursor;    /* Position in source */
	long        len;       /* Length of token */
	long        line;      /* Line number in source */
	qu_any      thing;     /* Literal (String, Int, Float, Big, etc. or 0) */
	int         nobr;      /* same line with next? */
};


/** qu_tokenizer
	qu_tokenizer represents an instance of the Tokenizer class. This is
	a pointer to the qu_tokenizer_s struct.
*/
struct qu_tokenizer_s
{
	QU_OBJECT_HEAD;
	qu_token_s  *token;     /* current token */
	qu_token_s  *last;      /* last token read */
	long         line;      /* current line */
	char        *cursor;    /* current cursor position */
	qu_string    source;    /* ASCII source string */
	qu_set       literals;  /* where to store literals */
	int          indent;    /* current indentation level */

	/* token buffer */
	qu_token_s   tokbuf [QU_TOKENIZER_MAXPEEK + 1];
};


/** QU_TOKENIZER_FLAG
	Flags used by Tokenizer objects.

	Do not access directly. Use the accessors instead.
*/
/* 0..9 not used but reserved by Container */
#define QU_TOKENIZER_FLAG_ERROR    QU_FLAG_10 /* set on error */
#define QU_TOKENIZER_FLAG_DOC      QU_FLAG_11 /* include documentation strings? */
#define QU_TOKENIZER_FLAG_INDENT   QU_FLAG_12 /* indentations are tokens? */
#define QU_TOKENIZER_FLAG_CONFIG   QU_FLAG_13 /* for Config */
#define QU_TOKENIZER_FLAG_HTML     QU_FLAG_14 /* /abc -> "</abc>" */
#define QU_TOKENIZER_FLAG_RESIDENT QU_FLAG_15 /* compiling a resident module? */


/** qu_tokenizer_include_doc
	Whether or not to include documentation strings.
*/
#define qu_tokenizer_include_doc(self)       qu_flag_test  (self, QU_TOKENIZER_FLAG_DOC)
#define qu_tokenizer_set_include_doc(self)   qu_flag_set   (self, QU_TOKENIZER_FLAG_DOC)
#define qu_tokenizer_unset_include_doc(self) qu_flag_unset (self, QU_TOKENIZER_FLAG_DOC)


/** qu_tokenizer_has_error
	Set/get the QU_TOKENIZER_FLAG_ERROR flag which indicates that the Tokenizer
	has an error.

	Setting may only be done by Tokenizer.
*/
#define qu_tokenizer_has_error(self)   qu_flag_test  (self, QU_TOKENIZER_FLAG_ERROR)
#define qu_tokenizer_set_error(self)   qu_flag_set   (self, QU_TOKENIZER_FLAG_ERROR)
#define qu_tokenizer_unset_error(self) qu_flag_unset (self, QU_TOKENIZER_FLAG_ERROR)


/** qu_tokenizer_use_indent
	Necessary because the compiler needs to do it later after creating
	the Tokenizer.
*/
#define qu_tokenizer_use_indent(self)       qu_flag_test  (self, QU_TOKENIZER_FLAG_INDENT)
#define qu_tokenizer_set_use_indent(self)   qu_flag_set   (self, QU_TOKENIZER_FLAG_INDENT)
#define qu_tokenizer_unset_use_indent(self) qu_flag_unset (self, QU_TOKENIZER_FLAG_INDENT)


/** qu_tokenizer_use_html
	Necessary because the compiler needs to do it later after creating
	the Tokenizer.
*/
#define qu_tokenizer_use_html(self)       qu_flag_test  (self, QU_TOKENIZER_FLAG_HTML)
#define qu_tokenizer_set_use_html(self)   qu_flag_set   (self, QU_TOKENIZER_FLAG_HTML)
#define qu_tokenizer_unset_use_html(self) qu_flag_unset (self, QU_TOKENIZER_FLAG_HTML)


/** qu_tokenizer_is_resident
	Set/get the QU_TOKENIZER_FLAG_RESIDENT flag which indicates that the Tokenizer
	is being used to compile a resident module.

	If set then all literals such as String, Int, Float, etc. are created
	with qu_gc_new_static instead of qu_gc_new thus reducing the GC workload.
*/
#define qu_tokenizer_is_resident(self)    qu_flag_test  (self, QU_TOKENIZER_FLAG_RESIDENT)
#define qu_tokenizer_set_resident(self)   qu_flag_set   (self, QU_TOKENIZER_FLAG_RESIDENT)
#define qu_tokenizer_unset_resident(self) qu_flag_unset (self, QU_TOKENIZER_FLAG_RESIDENT)


/** qu_tokenizer_is
	Whether or not the object is a Tokenizer.
*/
#define qu_tokenizer_is(o) (qu_classof(o) == QuTokenizerClass)


/** qu_tokenizer_keyword
	Return the token ID, or 0 if the string is not a token.
*/
extern int qu_tokenizer_keyword (const char *s, size_t len);


/** qu_tokenizer_name
	Return the token as a ={String} object.

	v{type} is one of #{QU_T_XXX}.
*/
extern qu_string qu_tokenizer_name (int type);


/** QuTokenizerBy
	These frozen ={qu_dict} objects hold the list of all tokens except the first
	group C{NAME .. DEDENT}.

	@warn
	You must never modify these objects.
*/
extern qu_dict QuTokenizerByName;
extern qu_dict QuTokenizerById;


/** qu_tokenizer_new
	Create a new Tokenizer object.

	Make sure that v{literals} is not frozen.

	v{op} is a combination of:
	@ul
	.QU_TOKENIZER_FLAG_DOC = include documentation strings
*/
extern qu_tokenizer qu_tokenizer_new      (qu_set literals, qu_string source, long op);
extern qu_tokenizer qu_tokenizer_new_root (qu_set literals, qu_string source, long op);


/** qu_tokenizer_eos
	Return 1 if the end of the source has been reached, 0 otherwise.
*/
extern int qu_tokenizer_eos (qu_tokenizer self);


/** qu_tokenizer_reset
	Use this function to reset the Tokenizer.
*/
extern void qu_tokenizer_reset (qu_tokenizer self, qu_set literals, qu_string source, long op);


/** qu_tokenizer_peek
	Return the v{n}-th token type ahead. If the end of the source has been
	reached then 0 is returned.

	v{n} must be in 1 .. #{QU_TOKENIZER_MAXPEEK}.
*/
extern int qu_tokenizer_peek (qu_tokenizer self, int n);


/** qu_tokenizer_take
	Make the next token current.

	@warn
	Make sure to call #{qu_tokenizer_next} before you call this function.
*/
extern void qu_tokenizer_take (qu_tokenizer self);


/** qu_tokenizer_match
	Whether or not the next token is of a specified type. If the end of the
	source has been reached then 0 is returned.
*/
extern int qu_tokenizer_match (qu_tokenizer self, int type);


/** qu_tokenizer_next
	Return the next token type. If the end of the source has been reached then
	0 is returned.
*/
extern int qu_tokenizer_next (qu_tokenizer self);


/** qu_tokenizer_need
	Ensure that the next token is of a specified type. If it is then
	make it current and the type is returned. Otherwise 0 is returned.
*/
extern int qu_tokenizer_need (qu_tokenizer self, int type);


/** qu_tokenizer_nobr
	Whether or not the next token is on the same line as the current token.
	It is it then 1 is returned. If it is not or the end of the source has
	been reached then 0 is returned.
*/
extern int qu_tokenizer_nobr (qu_tokenizer self);


/** Tokens
	This is the list of all tokens.
	See ={QU_T_XXX} for details.
*/
#define QU_TOKENIZER_IN_HEADER
#define t_(id, strz) QU_T_##id,
enum
{
#include "Tokens.h"
};
#undef t_
#undef QU_TOKENIZER_IN_HEADER


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_tokenizer__setup ();
extern void qu_tokenizer__init  ();

#endif /*_H_qu_tokenizer_*/
