/* Var.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_var_
#define _H_qu_var_ 1


/** Var
	This is the final Var class.
*/
extern qu_class QuVarClass;


/** qu_var
	qu_var represents an instance of the Var class.
	It is a pointer to the instance struct qu_var_s.

	@warn
	Instance properties, once created are stored in C{object->attrs}
	as the values, not as Var objects.
*/
struct qu_var_s
{
	QU_SYMBOL_HEAD;
	qu_any  value;  /* Value, never 0, initially nil */
	qu_any  type;   /* Validator or 0 */
	int     index;  /* Local index, if QU_VAR_FLAG_LEXICAL set */
};


/** QU_VAR_FLAG
	Flags used by Var objects.

	Do not access directly. Use the accessors instead.
*/
#define QU_VAR_FLAG_FINAL     QU_SYMBOL_FLAG_FINAL   /* final property */
#define QU_VAR_FLAG_CLASS     QU_SYMBOL_FLAG_CLASS   /* class property */
#define QU_VAR_FLAG_INST      QU_SYMBOL_FLAG_INST    /* instance property */
#define QU_VAR_FLAG_PRIVATE   QU_SYMBOL_FLAG_PRIVATE /* class/derived property */
#define QU_VAR_FLAG_LOCAL     QU_SYMBOL_FLAG_LOCAL   /* local to block and child block */
#define QU_VAR_FLAG_LEXICAL   QU_FLAG_10             /* local sub variable (lexical) */
#define QU_VAR_FLAG_STATIC    QU_FLAG_11             /* local sub static */
#define QU_VAR_FLAG_ARG       QU_FLAG_12             /* function argument */
#define QU_VAR_FLAG_VARARG    QU_FLAG_13             /* function vararg */
#define QU_VAR_FLAG_FLATTENER QU_FLAG_14             /* argument flattener */
#define QU_VAR_FLAG_DOCENTRY  QU_FLAG_15             /* is __doc__ */
#define QU_VAR_FLAG_FROZEN    QU_FLAG_16             /* cant modify */
#define QU_VAR_FLAG_RDWR      QU_FLAG_17             /* read-only or read-write? */
#define QU_VAR_FLAG_TYPEOF    QU_FLAG_18             /* from Array|Dict|Set of x statement? */
#define QU_VAR_FLAG_STRUCT    QU_FLAG_19             /* is struct */
#define QU_VAR_FLAG_ENUM      QU_FLAG_20             /* is enum */
#define QU_VAR_FLAG_ENUMITEM  QU_FLAG_21             /* is an enum item */
#define QU_VAR_FLAG_ALIAS     QU_FLAG_22             /* documentation hack for alias */


/** qu_var_is
	Whether or not the object is a qu_var.
*/
#define qu_var_is(o) (qu_classof(o) == QuVarClass)


/** qu_var_is_alias
	Whether this is an alias documentation hack.
*/
#define qu_var_is_alias(self) qu_flag_test (self, QU_VAR_FLAG_ALIAS)


/** qu_var_is_struct qu_var_is_typeof qu_var_is_enum qu_var_is_enumitem
	Various validators.

	qu_var_is_struct checks whether it comes from a C{struct} statement.

	qu_var_is_enum checks whether the variable comes from an C{enum} statement.

	qu_var_is_enumitem checks whether the variable is an C{enum} item.

	qu_var_is_typeof checks whether it is a C{Array|Dict|Set of Foo} validator,
	i.e:
	@code
		var foo: Array of String
		var bar: Dict of @baz

	@p
	Thus instead of having to create validator methods, e.g. Array.String,
	Set.Int, Dict.Array, etc., we use a more flexible validator Var object
	which fields are:
	@ul
	.name   = '*' (not important)
	.value  = QuArrayClass | QuSetClass | QuDictClass
	.type   = any valid validator
	.flag  |= QU_VAR_FLAG_TYPEOF
*/
#define qu_var_is_struct(self)    qu_flag_test (self, QU_VAR_FLAG_STRUCT)
#define qu_var_is_enum(self)      qu_flag_test (self, QU_VAR_FLAG_ENUM)
#define qu_var_is_enumitem(self)  qu_flag_test (self, QU_VAR_FLAG_ENUMITEM)
#define qu_var_is_typeof(self)    qu_flag_test (self, QU_VAR_FLAG_TYPEOF)


/** qu_var_is_class
	Whether this is a class property.
*/
#define qu_var_is_class(self) qu_flag_test (self, QU_VAR_FLAG_CLASS)


/** qu_var_is_inst
	Whether this is a class property.
*/
#define qu_var_is_inst(self) qu_flag_test (self, QU_VAR_FLAG_INST)


/** qu_var_is_local
	Whether this is a local variable visible to block and child blocks.
*/
#define qu_var_is_local(self) qu_flag_test (self, QU_VAR_FLAG_LOCAL)


/** qu_var_is_lexical
	Whether this is a local Sub variable.

	Note that lexical variables only lives in compile-time.
	They are indexed and live in the Frame stack at run-time.
*/
#define qu_var_is_lexical(self) qu_flag_test (self, QU_VAR_FLAG_LEXICAL)


/** qu_var_is_static
	Whether this is a Sub static variable.
*/
#define qu_var_is_static(self) qu_flag_test (self, QU_VAR_FLAG_STATIC)


/** qu_var_is_arg qu_var_is_vararg
	Whether a variable is a Sub argument.

	qu_var_is_vararg checks whether it is a variable argument specifier,
	i.e. C{c} in C{sub foo (a, b, *c)}.
*/
#define qu_var_is_arg(self)    qu_flag_test (self, QU_VAR_FLAG_ARG | QU_VAR_FLAG_VARARG)
#define qu_var_is_vararg(self) qu_flag_test (self, QU_VAR_FLAG_VARARG)


/** qu_var_is_flattener
	Whether this is an argument flattener.

	This must only be created by the interpreter to flatten argument to functions,
	i.e. v{bar} in C{foo, *bar, baz}.

	The use of flattener variables allows for small and calculatable stack size.
*/
#define qu_var_is_flattener(self) qu_flag_test (self, QU_VAR_FLAG_FLATTENER)


/** qu_var_is_docentry
	Whether this variable comes from a __doc__ statement.
*/
#define qu_var_is_docentry(self) qu_flag_test (self, QU_VAR_FLAG_DOCENTRY)


/** qu_var_is_private
	Whether this is a private variable known to block/childs, or a property
	known to class/subclasses.
*/
#define qu_var_is_private(self) qu_flag_test (self, QU_VAR_FLAG_PRIVATE)


/** qu_var_frozen qu_var_freeze
	qu_var_frozen checks whether the value of the variable may be modified (constant).

	qu_var_freeze makes the variable a constant.
*/
#define qu_var_frozen(self) qu_flag_test (self, QU_VAR_FLAG_FROZEN)
#define qu_var_freeze(self)   qu_flag_set  (self, QU_VAR_FLAG_FROZEN)


/** qu_var_is_rdwr
	Whether the global variable or property is read/write (read-only otherwise).
*/
#define qu_var_is_rdwr(self) qu_flag_test (self, QU_VAR_FLAG_RDWR)


/** qu_var_new
	Create a new Var object. Make sure that v{block} is a ={Module}, ={Class} or ={Sub}.

	Use qu_var_new_ex to type the variable. v{type} must be ={qu_object_valid_type}
*/
extern qu_var qu_var_new    (qu_symbol block, qu_string name, long flags);
extern qu_var qu_var_new_ex (qu_symbol block, qu_string name, long flags, qu_any type);

#define qu_var_new_rdwr(b, name)  qu_var_new (b, name, QU_VAR_FLAG_RDWR)
#define qu_var_new_const(b, name) qu_var_new (b, name, QU_VAR_FLAG_FROZEN)


/** qu_var_new_flattener
	Create a new flattener variable, i.e. v{c} from C{a, b, *c}.

	This is strictl for the interpreter.
*/
#define qu_var_new_flattener() \
	qu_var_new ((qu_symbol) QuSysModule, qu_string_byte ('*'), \
		QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_FROZEN | QU_VAR_FLAG_FLATTENER)


/** qu_var_get_module qu_var_get_class
	qu_var_get_module returns the ={Module} object where the variable is declared.

	qu_var_get_class returns the ={Class} object where the property is declared. If
	the Var is not a property then 0 is returned, no exception thrown.
*/
extern qu_module qu_var_get_module (qu_var self);
extern qu_class  qu_var_get_class  (qu_var self);


/** qu_var_gettag qu_var_settag
	qu_var_gettag returns a value in the C{__tag} property.
	Returns 0 if none.

	qu_var_settag set a value in the C{__tag} property.

	qu_var_gettag_dict gets all in a qu_dict.
	Returns 0 if none.
*/
extern qu_any  qu_var_gettag      (qu_var self, qu_string name);
extern void    qu_var_settag      (qu_var self, qu_string name, qu_any value);
extern qu_dict qu_var_gettag_dict (qu_var self);


/** qu_var_enum
	Does what the compiler does with the k{enum} keyword. This creates a
	constant named v{name} in v{where} which contains a frozen ={Dict}
	of the enumerated values. All enumerated values will be qu_var objects
	in v{where} as well.

	All entries must be a pair of C{const char *} key and values.

	For qu_var_enum the values must be C{int} the ={Int} class range.
	For qu_var_enum_ul the values must be C{unsigned long} values.
	The last argument must be NULL to mark the end of the entries.

	For qu_var_enum_ex you must specify the types.
	See ={qu_struct_type_to_char} for details.

	@warn
	v{where} must be a ={Module} or ={Class}, otherwise QU_FATAL is issued.
	QU_VAR_FLAG_FROZEN is added to ={Var} flags and QU_VAR_FLAG_CLASS is automatically
	added if v{where} is a Class.

	The ={Dict} and Var objects are static objects by means of ={qu_gc_new_static}.
	You must never modify the content.

	qu_var_enum_ex only supports C{abqiIlLfsx}.
	Others will give you a ={QU_FATAL} error.
	If you use "x" make sure the object is a static object otherwise you get
	a ={QU_FATAL} error as well.
*/
extern qu_var qu_var_enum    (qu_symbol where, const char *name, const char *first, ...);
extern qu_var qu_var_enum_ul (qu_symbol where, const char *name, const char *first, ...);
extern qu_var qu_var_enum_ex (qu_symbol where, const char *name, const char *spec, ...);


/***/
/*--------------------------------------------------------------------------*/

/* called internally by Class to initialize this class */
extern void qu_var__setup ();
extern void qu_var__init  ();

/* for interp */
extern qu_any qu_var_set_value (qu_var self, qu_any val);

/* for Reduce */
extern int qu_var_can_access (qu_var self, qu_symbol setter);

#endif /*_H_qu_var_*/
