/* attr.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_attr_
#define _H_qu_attr_ 1


/** qu_attr_s
	This represents an attribute table.

	It is used to store object attributes.
	It is a special case of a Dict where all keys are (and must be) ={String} objects.

	The struct is intentionally hidden.
	See C{attr.c} if you are curious.

	@warn
	An attribute table is not an object.
*/
#ifndef QU_ATTR_S_DECLARED
#define QU_ATTR_S_DECLARED
typedef struct qu_attr_s qu_attr_s;
#endif


/** qu_attr_alloc
	Allocate a new attribute table.
*/
extern qu_attr_s *qu_attr_alloc ();


/** qu_attr_free qu_attr_xfree
	Free an attribute table.

	qu_attr_xfree simply calls qu_attr_free if its argument is not NULL.
*/
extern void qu_attr_free (qu_attr_s *t);

#define qu_attr_xfree(t) \
QU_BEGIN \
	if (t) \
		qu_attr_free (t);\
QU_END


/** qu_attr_mark qu_attr_xmark
	Mark the items in the attribute table.

	These are called by the GC. You must not call this function directly unless
	you are using the table for something else then the v{attrs} field in
	={QU_OBJECT_HEAD}.

	qu_attr_xmark simply calls qu_attr_mark if its argument is not NULL.
*/
extern void qu_attr_mark (qu_attr_s *t);

#define qu_attr_xmark(t) \
QU_BEGIN \
	if (t) \
		qu_attr_mark (t); \
QU_END


/** qu_attr_find
	Find a value in the attribute table.

	Return 0 if not found but does not throw.
*/
extern qu_any qu_attr_find  (qu_attr_s *t, qu_string key);
extern qu_any qu_attr_findz (qu_attr_s *t, const char *key, long len);


/** qu_attr_add qu_attr_set
	qu_attr_add adds an entry without checking for an existing key. If the key
	already exists then there will be two entries with the same key, so be
	careful. The only reason qu_attr_add is provided is for efficiency, i.e.
	when you know an entry with the same key does not exist (e.g. when you
	just created the table). Normally you would use qu_attr_set instead.

	qu_attr_set checks for an existing key. If the key already exists, its
	value is set and 1 is returned. If not, a new entry is added and 0
	is returned instead. If the table is fixed then ={QU_FATAL} is issued.

	qu_attr_setex is the same but does it even if the table is fixed.
	It always succeed.
*/
extern void qu_attr_add   (qu_attr_s *t, qu_string key, qu_any val);
extern int  qu_attr_set   (qu_attr_s *t, qu_string key, qu_any val);
extern void qu_attr_setex (qu_attr_s *t, qu_string key, qu_any val);


/** qu_attr_del
	Delete an entry, return 1 if the entry existed and 0 otherwise.
*/
extern int qu_attr_del (qu_attr_s *t, qu_string key);


/** qu_attr_len
	Return the number of items currently in the table.
*/
extern long qu_attr_len (qu_attr_s *t);


/** qu_attr_walk
	Walk through the items.

	The first call must set C{i = 0}.
	@xmp
		qu_string key;
		qu_any    val;
		long      i = 0;

		while ((val = qu_attr_walk (MyFooAttr, &key, &i)) != 0)
		{
			... do something with key, val ...
		}

	@warn
	It is not safe to delete nor add an entry inside the loop.
*/
extern qu_any qu_attr_walk (qu_attr_s *t, qu_string *key, long *i);


/** qu_attr_merge_to_dict
	Export entries to a Dict.

	@warn
	Entries in the Dict with the same name will be overidden.
*/
extern void qu_attr_merge_to_dict (qu_attr_s *a, qu_dict d);


/** qu_attr_end_declsub
	This is only to be called by ={qu_sub_enddecl}.
	It frees all lexical ={Var} objects to save memory at runtime.
*/
extern void qu_attr_end_declsub (qu_sub u);


/** qu_attr_inst_get qu_attr_inst_set
	Get/set instance attribute.

	qu_attr_inst_get finds an instance attribute value. It throws EImplement
	if the attribute could not be found.

	qu_attr_inst_set sets an instance attribute value. It throws EImplement
	if the attribute could not be found, EModify if modification is prohibited.

	v{restrict} means whether or not private attributes may be accessed.
	If 1, then only public attributes are looked up.

	Use qu_attr_inst_getnt instead of qu_attr_inst_get to avoid exception.

	@note
	These should really be in C{qu_object.c} but we need to do some internal
	v{attr} stuffs: the values are stored in a flat array which is much
	more efficient.

	If you are calling C{class->get} then you can use {__qu_attr_nothrow} to
	avoid exceptions.

	@warn
	self may not be an ={Int}.
*/
extern qu_any qu_attr_inst_get   (qu_obj self, qu_string name, int restrict);
extern qu_any qu_attr_inst_set   (qu_obj self, qu_string name, qu_any val, int restrict);
extern qu_any qu_attr_inst_getnt (qu_obj self, qu_string name, int restrict);


/** qu_attr_search
	Search declared attribute in class, including mixins and superclasses.

	Return 0 if not found and does not throw any exception.

	@xmp
		class3, mix31, mix32
		|
		class2, mix21, mix22
		|
		class1, mix11, mix12

		# the search order will be:

		class1 -> mix21 -> mix11 -> class2 -> mix22 -> mix21 -> class3 -> mix32 -> mix31
*/
extern qu_any qu_attr_search (qu_class c, qu_string name);


/** __qu_attr_nothrow
	Use these when you are calling C{class->get} and don't want to throw
	useless exceptions (you are just querying and it is okay if the attribute
	does not exist).
*/
#define __qu_attr_nothrow_begin QU_BEGIN __QuAttrNoThrow++; QU_END
#define __qu_attr_nothrow_end   QU_BEGIN __QuAttrNoThrow--; QU_END
/***/
extern volatile int __QuAttrNoThrow;

/***/
#endif /*_H_qu_attr_*/
