/* gc.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_gc_
#define _H_qu_gc_ 1


/** QU_GC_SIZE_LIMIT
	This is the biggest possible size of an object struct.

	All qu_xxx_s structs size must be smaller or equal this limit.

	If you need a bigger struct then use an extra field and use
	={gc_mem_alloc} instead.
*/
#define QU_GC_SIZE_LIMIT 1024


/** QU_GC_MAX_ITEMS
	The maximum number of items that can be allocated.

	Note that this value is guaranteed to be <= ={QU_INT_MAX}.

	All container objects should use this value instead of defining their own
	even if they can actually contain more than this value.
*/
#define QU_GC_MAX_ITEMS ((INT_MAX >> 1) / SIZEOF_VOID_P)


/** QU_MEM_MAX QU_MEM_CHECK
	QU_MEM_MAX is the maximum size of memory you are allowed to allocate.

	Use QU_MEM_CHECK to check. It is not perfect but at least it reminds
	you that you must check the size.
*/
#define QU_MEM_MAX (INT_MAX >> 1)
#define QU_MEM_CHECK(expr) \
QU_BEGIN \
	if (!(expr)) \
		QU_THROW (QuEMemory); \
QU_END


/** qu_gc_disable qu_gc_enable
	Enable/disable GC.
*/
#define qu_gc_disable() QU_BEGIN __QuGcDisable++; QU_END
#define qu_gc_enable()  QU_BEGIN __QuGcDisable--; QU_END
/***/
extern int volatile __QuGcDisable;


/** __qu_gc_critical
	Always enclose the creation of composite objects within these two macros.

	This ensures that the GC doesn't get triggered and accidentally call your
	mark function with an incomplete object.

	In the following example, GC can occur right before qu_dict_new. This
	will trigger a call to C{_foo_mark} with C{self->dict} and C{self->set}
	not yet initialized.

	@xmp
		void _foo_mark (Foo self)
		{
			qu_gc_mark (self->dict);
			qu_gc_mark (self->set);
		}

	  __qu_gc_critical_begin;
		a = qu_gc_new (Foo);
		a->dict = qu_dict_new (4);
		a->set = qu_set_new (8);
	  __qu_gc_critical_end;
		a->mem = qu_mem_alloc (10);

	@code
	That is generally simpler than writing code such as the following example.
	It also eliminates the qu_flag_test in C{_foo_mark} thus is faster.

	@xmp
		void _foo_mark (Foo self)
		{
			if (qu_flag_test (self, FOO_FLAG_INITIALIZED))
			{
				qu_gc_mark (self->dict);
				qu_gc_mark (self->set);
			}
		}

		a = qu_gc_new (FooClass);
		a->dict = qu_dict_new (4);
		a->set = qu_set_new (8);
		qu_flag_set (a, FOO_FLAG_INITIALIZED);
		a->mem = qu_mem_alloc (10);

	@note
	Note that qu_mem_alloc et.al. do not trigger GC thus is safe to do
	out of the enclosing macros.
*/
#define __qu_gc_critical_begin (__QuGcDisable++)
#define __qu_gc_critical_end   (__QuGcDisable--)


/** qu_gc_new
	Object allocation routines.

	qu_gc_new_root creates a root object. A root object is always marked and never
	freed. It may refer to other objects.

	qu_gc_new_static creates a static object. A static object is never marked and
	never freed. It must not reference other objects. You can use ={qu_gc_is_static}
	to test whether or not an object is a static object.

	={Int} objects are static objects.

	qu_gc_new create a normal (non-root non-static) object.
	This object is marked and freed automatically by the GC. It may refer
	to other objects, of course.

	You never need to free an object because that is taken care by the GC
	automatically when the object is no longer referenced. You must never
	call ={qu_gc_delete} unless you know exactly what you are doing.
	qu_gc_delete forces the object to be deleted and is intended to be used
	by the interpreter only.

	@note
	Allocation clears C{object->flags} except for qu_gc_new_static which sets
	QU_FLAG_STATIC (see ={Flag}).

	Always use the macros instead of the functions.

	@warn
	You might need to use ={__qu_gc_critical}.
*/
extern qu_any qu_gc__new_root   (qu_class c, size_t size);
extern qu_any qu_gc__new_static (qu_class c, size_t size);
extern qu_any qu_gc__new        (qu_class c, size_t size);
#define qu_gc_new_root(klass)    qu_gc__new_root   (klass, (klass)->size)
#define qu_gc_new_static(klass)  qu_gc__new_static (klass, (klass)->size)
#define qu_gc_new(klass)         qu_gc__new        (klass, (klass)->size)


/** qu_gc_is_static
	Whether or not an object is static.

	For ={Int} objects, this function returns true.
*/
extern int qu_gc_is_static (qu_any o);


/** qu_gc_delete
	Force an object to be deleted.

	@warn
	Use with care. Make sure there are no other object referencing this object.
	You would normally only use this with temporary objects.

	This function is designed to for the interpreter in case the -u option
	is set, i.e. to destroy ={Frame} objects as soon as they are no longer used.
*/
extern void qu_gc_delete (qu_obj o);


/** qu_mem_alloc qu_mem_calloc qu_mem_realloc qu_mem_free qu_mem_copy
	These are the replacement routines for malloc etc.

	Note that you may not allocate a memory bigger than ={QU_MEM_MAX}.
	You must check for this yourself before you call the routines.
*/
#ifdef QU_DEBUG_SEGV
extern void *qu_mem__alloc   (size_t size, const char *func);
extern void *qu_mem__calloc  (size_t nelts, size_t size, const char *func);
extern void *qu_mem__realloc (void *p, size_t size, const char *func);
extern void  qu_mem__free    (void *p, const char *func);
extern void *qu_mem__copy    (void *p, const char *func);
#define qu_mem_alloc(z)      qu_mem__alloc   (z, __FUNCTION__)
#define qu_mem_calloc(n,z)   qu_mem__calloc  (n, z, __FUNCTION__)
#define qu_mem_realloc(p,z)  qu_mem__realloc (p, z, __FUNCTION__)
#define qu_mem_free(p)       qu_mem__free    (p, __FUNCTION__)
#define qu_mem_copy(p)       qu_mem__copy    (p, __FUNCTION__)
#else
extern void *qu_mem_alloc   (size_t size);
extern void *qu_mem_calloc  (size_t nelts, size_t size);
extern void *qu_mem_realloc (void *p, size_t size);
extern void  qu_mem_free    (void *p);
extern void *qu_mem_copy    (void *p);
#endif /*QU_DEBUG_SEGV*/


/** qu_mem_static
	At initialization Qu allocates a malloc'ed space for used as static memory.

	Static memory can only be allocated. It can not be resized nor freed.
	You must never use qu_mem_realloc nor qu_mem_free with the pointer returned
	by qu_mem_static.
*/
extern void *qu_mem_static (size_t size);


/** qu_gc_need_reclaim
	Whether or not we need to start GC.

	Use qu_gc_need_reclaim_set to force GC as soon as possible.
	To immediately start GC call ={qu_gc_reclaim} instead.
*/
#define qu_gc_need_reclaim()     (__QuGcNeedReclaim != 0)
#define qu_gc_need_reclaim_set() (__QuGcNeedReclaim = 1)
/***/
extern int volatile __QuGcNeedReclaim;


/** qu_gc_reclaim
	Start collecting garbage.

	To simply have GC to run as soon as possible, use ={qu_gc_need_reclaim_set}
	instead.
*/
extern void qu_gc_reclaim ();


/** qu_gc_num_roots qu_gc_num_statics qu_gc_num_regulars
	qu_gc_num_roots returns the number of root objects, i.e. those allocated with
	={qu_gc_new_root}.

	qu_gc_num_statics returns the number of static objects, i.e. those allocated with
	={qu_gc_new_static}.

	qu_gc_num_regulars returns the number of non-static non-root objects, i.e. those
	allocated with ={qu_gc_new}.

	@note
	These functions may return 0 because the GC may not keep track of the number
	of objects for performance reasons.
*/
extern long qu_gc_num_regulars ();
extern long qu_gc_num_roots    ();
extern long qu_gc_num_statics  ();


/** qu_gc_limit
	Return the current limit (in bytes) when GC should occur.
*/
#define qu_gc_limit() (__QuGcLimit)
/***/
extern long __QuGcLimit ();


/** qu_gc_mark qu_gc_xmark
	Mark referenced objects.

	Call these functions from within the v{mark} methods. See plenty of
	examples in the source code.
*/
extern void qu_gc_mark  (qu_any o);
extern void qu_gc_xmark (qu_any o);


/** __QuGcMutex
	This is the one and only global lock.
	Only one thread hold this lock at any given time.

	@warn
	Never acquire the lock directly.
	Use the ={__qu_blocking} macros instead.
*/
extern qu_thread_mutex_t __QuGcMutex;
#define qu_gc_lock()   qu_thread_mutex_lock (&__QuGcMutex)
#define qu_gc_unlock() qu_thread_mutex_unlock (&__QuGcMutex)


/** qu_gc_mem_allocated
	Get the current amount of active memory.
	This may be used by the GC to determine whether or not GC must start.
*/
#define qu_gc_mem_allocated() __QuGcMemAllocated
/***/
extern unsigned long volatile __QuGcMemAllocated;


/** qu_gc_stat
	Prints some GC information on v{stderr}.
*/
extern void qu_gc_stat ();


/** __QuGcMemTrap
	Set function that must be called when memory allocation fails.

	@warn
	This is intended for the interpreter. You should have no reasons to
	change it.
*/
typedef void (*qu_gcmemtrap_f) ();
extern qu_gcmemtrap_f __QuGcMemTrap;


/***/
/*--------------------------------------------------------------------------*/

/* called internally by the main program to initialize GC */
extern void qu_gc__init ();

/* for Thread to mark the C stack */
extern void qu_gc_mark_cstack (qu_thread self, qu_any *lo, qu_any *hi);

/* for Weak to add a Weak to self */
extern int qu_gc_weak_set (qu_obj self, qu_weak w);

/* for Weak to remove a Weak from self - called by qu__weak_unref */
extern void qu_gc_weak_delete (qu_obj self, qu_weak w);

#endif /*_H_qu_gc_*/
