/* interp.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_interp_
#define _H_qu_interp_ 1


/** qu_interp_run_main
	Run the main thread. This is called by the main program. Do not call
	directly.
*/
extern qu_any qu_interp_run_main (qu_module m);


/** qu_interp_run_thread
	All new threads enter here. This function is called from ={Thread}.
	Do not call directly.
*/
extern void qu_interp_run_thread (qu_thread t, qu_any self);


/** qu_interp_frame_next
	This is called to get the next value from a generator ={Frame}.

	It is only meant to be called by ={qu_frame_next}.
*/
extern qu_any qu_interp_frame_next (qu_frame f);


/** qu_interp_exiting
	True when exiting the main program. Read-only.
*/
#define qu_interp_exiting() (__QuInterpExiting != 0)
/***/
extern volatile int __QuInterpExiting;


/** qu_interp_where qu_interp_self qu_interp_get_module qu_interp_where_string qu_interp_scope
	Get info from the running code.

	qu_interp_where returns the block where the code is currently running.
	It returns the associated ={Module}, ={Class} or ={Sub} object.

	qu_interp_where_string returns a nicely formatted string.

	qu_interp_self returns the current value of self.

	qu_interp_get_module returns the ={Module} object of the current running code.

	qu_interp_scope returns a Class if the code is currently running from within
	a Class, theModule otherwise.

	@warn
	It is an error to call these functions before the main program runs.
*/
extern qu_symbol qu_interp_where        ();
extern qu_string qu_interp_where_string ();
extern qu_any    qu_interp_self         ();
extern qu_module qu_interp_get_module   ();
extern qu_symbol qu_interp_scope        ();


/** qu_interp_errstr qu_interp_print_err
	Get the last exception as a string or print it.
*/
extern qu_string qu_interp_errstr    ();
extern void      qu_interp_print_err ();


/** qu_interp_self
	Return the associated self object of the current running ={Frame}.

	@warn
	It is an error to call this function before the main program runs.
*/
extern qu_any qu_interp_self ();


/** qu_interp_add_finalizer
	Add an object for finalization, i.e. have the interpreter call its
	__destroy method.

	You must not call this function directly. It is called by the GC.

	The function call can happen at any time. There is absolutely no guarantee
	when and whether or not it is actually called (e.g. abort may be issued
	before the function is called).
*/
extern void qu_interp_add_finalizer (qu_any o);


/** qu_interp_unsafe __qu_unsafe__
	Whether or not the call comes from an untrusted eval module.

	Put __qu_unsafe__ on top of a C function to throw ESecurity if the call came from
	an untrusted eval module.

	This has the same effect as k{__unsafe} at program level.

	@xmp
		static qu_int Foo_bar (qu_string baz)
		{
			int a, b;

			__qu_unsafe__;
			...
		}
*/
extern int qu_interp_unsafe ();
#define __qu_unsafe__ QU_REQUIRE (QuESecurity, !qu_interp_unsafe())



/** __qu_blocking
	Use the macros (not the functions) to do a blocking operation.

	Use __qu_blocking_check_signal to check for termination signals every now and
	then while blocking, i.e. __qu_blocking_check_signal so we are a bit more
	responsive to, say, C{Ctrl-c}.

	@warn
	Only one thread has the interpreter lock at any given time.

	@p
	__qu_blocking_begin releases the lock while __qu_blocking_end acquires the lock.

	A thread may allocate objects or allocate memory only if it owns the
	interpreter lock.

	={qu_gc_reclaim} may be run by any thread at any time. Thus in between
	__qu_blocking_begin and __qu_blocking_end, another thread may run GC. This
	means that you must never allocate objects or allocate memory in
	between __qu_blocking_begin and __qu_blocking_end. Even accessing an object
	is safe only if it is immutable such as String, Float, etc.

	The pairs must be visible in one file.

	If you have issued __qu_blocking_begin then you must issue __qu_blocking_end
	before calling a function unless you are certain that the function does
	not allocate objects/memory. If the function is not within the same C file,
	you must never assume anything.

	If you do not issue __qu_blocking_end before calling a function and the
	function happens to issue __qu_blocking_begin then you have a deadlock.

	As a rule of thumb, only perform system calls in between __qu_blocking_begin
	and __qu_blocking_end.

	@warn
	The C compiler might perform an over-optimization such that an object stored
	in a local variable is not actually stored in the C stack. This makes it impossible
	for the GC to reach and mark the object. Thus you may need to declared the variable
	holding your objects created before __qu_blocking_begin with the volatile
	keyword.

	@xmp
		static int Foo_bar (Foo self, qu_int x)
		{
			volatile qu_array a = qu_array_new (3);

			....
			__qu_blocking_begin
			...
			__qu_blocking_end
			...
		}
*/
extern void qu_interp_lock   ();
extern void qu_interp_unlock ();
#define __qu_blocking_begin qu_interp_unlock ()
#define __qu_blocking_end   qu_interp_lock ()

#define __qu_blocking_check_signal \
QU_BEGIN \
	if (__QuSigTERM || __QuSigINT) \
	{ \
		__qu_blocking_end; \
		if (__QuSigTERM) \
			QU_THROW (QuEQuit); \
		QU_THROW (QuEInterrupt); \
	} \
QU_END


#define __qu_blocking_check_signalv(retval) \
QU_BEGIN \
	if (__QuSigTERM || __QuSigINT) \
	{ \
		__qu_blocking_end; \
		if (__QuSigTERM) \
			QU_THROWV (QuEQuit, retval); \
		QU_THROWV (QuEInterrupt, retval); \
	} \
QU_END


/** qu_interp_eval
	Evaluate a code string.

	v{mod} is the Module object where eval occurs.

	v{src} is the code string.

	v{imports} is a list of symbol to export.

	v{options} is a combination of QU_INTERP_EVAL flags.

	If v{path} is not NULL then set the module's f{name} field to path.
*/
extern qu_any qu_interp_eval (qu_module mod, qu_string src, qu_dict imports, long options, qu_string path);

/* options is one of these */
#define QU_INTERP_EVAL_EXPR       0x0001 /* compile as expression */
#define QU_INTERP_EVAL_MODULE     0x0002 /* compile as qu_module */
#define QU_INTERP_EVAL_DOC        0x0008 /* compile as qu_module, do not initialize */

/* possibly combined with one of these */
#define QU_INTERP_EVAL_TRUSTED    0x1000 /* trusted source? */
#define QU_INTERP_EVAL_GETIMPORT  0x2000 /* get back imports result */


/***/
/*--------------------------------------------------------------------------*/

/* internally called by the main program to initialize the interpreter */
extern void qu_interp__init ();

/* inform the interpreter of a signal */
extern volatile int __QuInterpGotSignal;

/* the current running frame - highly toxic, use with care! */
extern volatile qu_frame __QuInterpFrame;

/* for the JIT */
extern qu_any qu_interp_run_frame (qu_frame f);

/* when a thread->ticks reaches this limit, it's time to switch to other threads */
#define QU_INTERP_THREAD_TICKS 10001

/* run finalizers - for the interpreter only, don't call otherwise. */
extern volatile int      __QuInterpInFinalize;
extern volatile qu_array __QuInterpFinalizeList;

extern void qu__interp_run_finalize ();
#define qu_interp_run_finalize() \
QU_BEGIN \
	if (!__QuInterpInFinalize && __QuInterpFinalizeList->len) \
	{ \
		__QuInterpInFinalize = 1; \
		qu__interp_run_finalize (); \
		__QuInterpInFinalize = 0; \
	} \
QU_END


/*
	Callers - see Sub.c

	c = frame and frame stack is on the C stack (non-generator)
	a = frame and frame stack are allocated (non-generator)
	g = frame and frame stack is allocated (generator)
*/
extern qu_any qu_interp_call_cn (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_any qu_interp_call_c4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_interp_call_c3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_interp_call_c2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_interp_call_c1 (qu_sub u, qu_any self, qu_any a1);
extern qu_any qu_interp_call_c0 (qu_sub u, qu_any self);

extern qu_any qu_interp_call_an (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_any qu_interp_call_a4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_interp_call_a3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_interp_call_a2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_interp_call_a1 (qu_sub u, qu_any self, qu_any a1);
extern qu_any qu_interp_call_a0 (qu_sub u, qu_any self);

extern qu_frame qu_interp_call_gn (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_frame qu_interp_call_g4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_frame qu_interp_call_g3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_frame qu_interp_call_g2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_frame qu_interp_call_g1 (qu_sub u, qu_any self, qu_any a1);
extern qu_frame qu_interp_call_g0 (qu_sub u, qu_any self);

extern qu_any qu_interp_call_once_cn (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_any qu_interp_call_once_c4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_interp_call_once_c3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_interp_call_once_c2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_interp_call_once_c1 (qu_sub u, qu_any self, qu_any a1);
extern qu_any qu_interp_call_once_c0 (qu_sub u, qu_any self);

extern qu_any qu_interp_call_once_an (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_any qu_interp_call_once_a4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_interp_call_once_a3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_interp_call_once_a2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_interp_call_once_a1 (qu_sub u, qu_any self, qu_any a1);
extern qu_any qu_interp_call_once_a0 (qu_sub u, qu_any self);

extern qu_any qu_interp_call_jit_cn (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_any qu_interp_call_jit_c4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_interp_call_jit_c3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_interp_call_jit_c2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_interp_call_jit_c1 (qu_sub u, qu_any self, qu_any a1);
extern qu_any qu_interp_call_jit_c0 (qu_sub u, qu_any self);

extern qu_any qu_interp_call_jit_an (qu_sub u, qu_any self, int argc, qu_any *argv);
extern qu_any qu_interp_call_jit_a4 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4);
extern qu_any qu_interp_call_jit_a3 (qu_sub u, qu_any self, qu_any a1, qu_any a2, qu_any a3);
extern qu_any qu_interp_call_jit_a2 (qu_sub u, qu_any self, qu_any a1, qu_any a2);
extern qu_any qu_interp_call_jit_a1 (qu_sub u, qu_any self, qu_any a1);
extern qu_any qu_interp_call_jit_a0 (qu_sub u, qu_any self);

#endif /*_H_qu_interp_*/
