/* threads.h - Copyright (c) Marc Krisnanto */

/*
	This is the low-level interface to the underlying system thread library.
	If you are developing thread modules and need more functionalities, then
	please add them here.

	Note that this header file must not include anything from a high-level
	interface. Only system libraries are allowed.
*/

#ifndef _H_qu_threads_
#define _H_qu_threads_ 1

#include <pthread.h>

/** qu_thread_yield
	Give other threads a chance.
*/
#ifdef HAVE_SCHED_YIELD
#  include <sched.h>
#  define qu_thread_yield() sched_yield()
#else
extern void qu_thread_yield ();
#endif


/** QU_THREAD_IDFORMAT
	The format of pthread_t.
*/
#define QU_THREAD_IDFORMAT "%lu"
#ifdef HAVE_LONG_LONG
#  if SIZEOF_PTHREAD_T == SIZEOF_LONG_LONG
#     undef QU_THREAD_IDFORMAT
#     define QU_THREAD_IDFORMAT "%llu"
#  endif
#endif


/** QU_THREAD_STACK_MIN
	The minimum size of a thread system stack.
*/
#ifdef PTHREAD_STACK_MIN
#  define QU_THREAD_STACK_MIN (PTHREAD_STACK_MIN + 32768)
#else
#  define QU_THREAD_STACK_MIN 49152
#endif


/** qu_thread_t qu_thread_key_t qu_thread_mutex_t
	Basic system thread types.

	See the pthread library for details.
*/
#define qu_thread_t       pthread_t
#define qu_thread_key_t   pthread_key_t
#define qu_thread_mutex_t pthread_mutex_t
#define qu_thread_cond_t  pthread_cond_t


/** qu_thread_create
	Create a new system thread.

	See the pthread library for details.
*/
extern int qu_thread_create (qu_thread_t *tid, void *(*start)(void *), void *arg,
                             int stacksize, int detach);

/** qu_thread_functions
	System thread functions.

	See the pthread library for details.
*/
#define qu_thread_sigmask(h,n,o)       pthread_sigmask(h,n,o)
#define qu_thread_id()                 pthread_self ()
#define qu_thread_exit(v)              pthread_exit (v)
#define qu_thread_kill(t,s)            pthread_kill (t, s)
#define qu_thread_detach(t)            pthread_detach (t)
#define qu_thread_join(t, p)           pthread_join (t, p)

#define qu_thread_specific_create(k,f) pthread_key_create (k, f)
#define qu_thread_specific_set(k,v)    pthread_setspecific (k, v)
#define qu_thread_specific_get(k)      pthread_getspecific (k)

#define QU_THREAD_MUTEX_INITIALIZER    PTHREAD_MUTEX_INITIALIZER
#define qu_thread_mutex_init(m)        pthread_mutex_init (m, NULL)
#define qu_thread_mutex_destroy(m)     pthread_mutex_destroy (m)
#define qu_thread_mutex_lock(m)        pthread_mutex_lock (m)
#define qu_thread_mutex_trylock(m)     pthread_mutex_trylock (m)
#define qu_thread_mutex_unlock(m)      pthread_mutex_unlock (m)

#define QU_THREAD_COND_INITIALIZER     PTHREAD_COND_INITIALIZER
#define qu_thread_cond_init(c)         pthread_cond_init (c, NULL)
#define qu_thread_cond_destroy(c)      pthread_cond_destroy (c)
#define qu_thread_cond_signal(c)       pthread_cond_signal (c)
#define qu_thread_cond_broadcast(c)    pthread_cond_broadcast (c)
#define qu_thread_cond_wait(c, m)      pthread_cond_wait (c, m)

/***/
#endif /*_H_qu_threads_*/
