/* Bits.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_NETINET_IN_H

#include <netinet/in.h>


/*) module Bits
	This module provides a number of functions to deal with bits.

	@xmp
		use Bits

		println (Bits.htonl ('foobar'))
		println (Bits.htonl (123))

	@out
		1718579042
		2063597568
*/



/*) sub htonl (x, ofs = 0) -> Integer
	Convert the unsigned integer from host byte order to network byte order.

	v{x} may be one of:
	@rc
	.| String Buffer | Read a 32-bits unsigned integer
	.| Int Long      | Convert byte order

	@p
	v{ofs} is only valid when v{x} is a String or Buffer.

	@exc
	.EArgValue
	.EIndex

	@xmp
		use Bits

		println (Bits.htonl ('foobar'))

		x = Bits.htonl (123)
		println (x)
		x = Bits.ntohl (x)
		println (x)

	@out
		1718579042
		2063597568
		123
*/
static qu_integer Bits_htonl (int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (qu_stringy_is (x))
	{
		qu_stringy s = x;
		long       ofs = 0;
		char      *z;
		uint32_t   v;

		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			QU_ENSURE_INDEX (ofs, s->len);
		}
		QU_REQUIRE (QuEIndex, s->len - ofs >= 4);
		z = s->str + ofs;
		v = htonl (*(uint32_t *) z);
#if SIZEOF_SHORT == 4
		return qu_integer_newuh (v);
#elif SIZEOF_INT == 4
		return qu_integer_newui (v);
#elif SIZEOF_LONG == 4
		return qu_integer_newui (v);
#else
		QU_THROW (QuEImplement);
#endif
	}
	else
	{
		uint32_t v;

#if SIZEOF_SHORT == 4
		QU_ARG_vushort (0, v);
		v = htonl (v);
		return qu_integer_newuh (v);
#elif SIZEOF_INT == 4
		QU_ARG_vuint (0, v);
		v = htonl (v);
		return qu_integer_newuh (v);
#elif SIZEOF_LONG == 4
		QU_ARG_vulong (0, v);
		v = htonl (v);
		return qu_integer_newuh (v);
#else
		v = 0;
		QU_THROW (QuEImplement);
#endif
	}
}



/*) sub htons (x, ofs = 0) -> Integer
	Convert the unsigned integer from host byte order to network byte order.

	v{x} may be one of:
	@rc
	.| String Buffer | Read a 16-bits unsigned integer
	.| Int Long      | Convert byte order

	@p
	v{ofs} is only valid when v{x} is a String or Buffer.

	@exc
	.EArgValue
	.EIndex

	@xmp
		use Bits

		println (Bits.htons ('foobar'))

		x = Bits.htons (123)
		println (x)
		x = Bits.ntohs (x)
		println (x)

	@out
		26223
		31488
		123
*/
static qu_integer Bits_htons (int argc, qu_any *argv)
{
#if SIZEOF_SHORT == 2
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (qu_stringy_is (x))
	{
		qu_stringy  s = x;
		long        ofs = 0;
		char       *z;
		uint16_t    v;

		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			QU_ENSURE_INDEX (ofs, s->len);
		}
		QU_REQUIRE (QuEIndex, s->len - ofs >= 2);
		z = s->str + ofs;
		v = htons (*(uint16_t *) z);
		return qu_integer_newuh (v);
	}
	else
	{
		uint16_t v;

		QU_ARG_vushort (0, v);
		v = htons (v);
		return qu_integer_newuh (v);
	}
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub ntohl (x, ofs = 0) -> Integer
	Does the reverse of #{htonl}.
*/
static qu_integer Bits_ntohl (int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (qu_stringy_is (x))
	{
		qu_stringy  s = x;
		long        ofs = 0;
		char       *z;
		uint32_t    v;

		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			QU_ENSURE_INDEX (ofs, s->len);
		}
		QU_REQUIRE (QuEIndex, s->len - ofs >= 4);
		z = s->str + ofs;
		v = ntohl (*(uint32_t *) z);
#if SIZEOF_SHORT == 4
		return qu_integer_newuh (v);
#elif SIZEOF_INT == 4
		return qu_integer_newui (v);
#elif SIZEOF_LONG == 4
		return qu_integer_newui (v);
#else
		QU_THROW (QuEImplement);
#endif
	}
	else
	{
		uint32_t v;

#if SIZEOF_SHORT == 4
		QU_ARG_vushort (0, v);
		v = ntohl (v);
		return qu_integer_newuh (v);
#elif SIZEOF_INT == 4
		QU_ARG_vuint (0, v);
		v = ntohl (v);
		return qu_integer_newuh (v);
#elif SIZEOF_LONG == 4
		QU_ARG_vulong (0, v);
		v = ntohl (v);
		return qu_integer_newuh (v);
#else
		v = 0;
		QU_THROW (QuEImplement);
#endif
	}
}



/*) sub ntohs (x, ofs = 0) -> Integer
	Does the reverse of #{htons}.
*/
static qu_integer Bits_ntohs (int argc, qu_any *argv)
{
#if SIZEOF_SHORT == 2
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (qu_stringy_is (x))
	{
		qu_stringy  s = x;
		long        ofs = 0;
		char       *z;
		uint16_t    v;

		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			QU_ENSURE_INDEX (ofs, s->len);
		}
		QU_REQUIRE (QuEIndex, s->len - ofs >= 2);
		z = s->str + ofs;
		v = ntohs (*(uint16_t *) z);
		return qu_integer_newuh (v);
	}
	else
	{
		uint16_t v;

		QU_ARG_vushort (0, v);
		v = ntohs (v);
		return qu_integer_newuh (v);
	}
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub get (x, type: String, ofs = 0) -> Any
	Read a value from the String or Buffer v{x} optionally start from v{ofs}.

	See -{Struct.spec} for v{type}.

	@exc
	.EArgValue
	.EIndex

	@warn
		Be careful when you read pointers. Your program may crash.
		Generally you would never read Qu object pointers directly.

	@xmp
		use Bits

		b = Buffer ('foobar')
		Bits.set (b, 'I', 123)
		println (b.quote)
		println (Bits.get (b, 'I'))

	@out
		'{\x00\x00\x00ar'
		123
*/
qu_any Bits_get (int argc, qu_any *argv)
{
	long              ofs = 0;
	unsigned char     c;
	long              m;
	qu_stringy        x, y;
	qu_struct_type_s *t;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, x, qu_string);
	QU_ARG_VTYPE  (1, y, qu_string);
	QU_ARG_CHECK  (y->len);
	if (argc > 2)
	{
		QU_ARG_vlong (2, ofs);
		QU_ENSURE_INDEX (ofs, x->len);
	}
	c = y->str[0];
	t = qu_struct_type_info (c);
	QU_ARG_CHECK (t->type);
	m = x->len - ofs;
	QU_REQUIRE (QuEIndex, m >= (long) t->size);
	return qu_pointer_get (x->str + ofs, c);
}



/*) sub set (x, v, type: String, ofs = 0) -> Any
	Set value v{v} in String or Buffer v{x} optionally start from v{ofs}.

	If v{x} is a String then a new String is returned. If v{x} is a Buffer
	then the Buffer is returned.

	See -{Struct.spec} for v{type}.

	@exc
	.EArgValue
	.EIndex

	@xmp
		use Bits

		b = Buffer ('foobar')
		Bits.set (b, 'I', 123)
		println (b.quote)
		println (Bits.get (b, 'I'))

	@out
		'{\x00\x00\x00ar'
		123
*/
qu_stringy Bits_set (int argc, qu_any *argv)
{
	long              ofs = 0;
	unsigned char     c;
	long              m;
	qu_stringy        x, y;
	qu_any            v;
	qu_struct_type_s *t;

	QU_ARG_VCOUNT (3, 4);
	QU_ARG_VTYPE  (0, x, qu_string);
	QU_ARG_VTYPE  (1, y, qu_string);
	QU_ARG_CHECK  (y->len);
	v = argv[2];
	if (argc > 3)
	{
		QU_ARG_vlong (3, ofs);
		QU_ENSURE_INDEX (ofs, x->len);
	}
	c = y->str[0];
	t = qu_struct_type_info (c);
	QU_ARG_CHECK (t->type);
	m = x->len - ofs;
	QU_REQUIRE (QuEIndex, m >= (long) t->size);
	if (qu_string_is (x))
	{
		x = qu_string_dup (x);
		/* dang! */
		x->hash = -1;
	}
	QU_ENSURE (qu_pointer_set (x->str + ofs, c, v));
	return x;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Bits_init (qu_module m)
{
	qu_module_add_function (m, "htonl", Bits_htonl, -1);
	qu_module_add_function (m, "htons", Bits_htons, -1);
	qu_module_add_function (m, "ntohl", Bits_ntohl, -1);
	qu_module_add_function (m, "ntohs", Bits_ntohs, -1);

	qu_module_add_function (m, "get"  , Bits_get  , -1);
	qu_module_add_function (m, "set"  , Bits_set  , -1);

	return 1;
}


#else /*HAVE_NETINET_IN_H*/

int qu_Bits_init (qu_module m)
{
	QU_THROWZ (QuEImplement, "Bits not available on your platform");
}

#endif /*HAVE_NETINET_IN_H*/
