/* Curses.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/*) module Curses
	This module provides an interface to the m{ncurses} library.

	Note that the documentations are intentionally made short because they
	should only serve quick references. You must consult the curses library
	documentation (C{man ncurses}) for details.
*/


/*----------------------------------------------------------------------------
	crappy stuffs
----------------------------------------------------------------------------*/
static int   *COLORS_p;
static int   *COLOR_PAIRS_p;
static int   *LINES_p;
static int   *COLS_p;
static int   *TABSIZE_p;
static int   *ESCDELAY_p;

#if 0
static void **stdscr_p;
static void **curscr_p;
static void **newscr_p;
#endif


static qu_integer qu_curses__int (int *p)
{
	int v;

	if (p == NULL)
		QU_THROW (QuEImplement);
	v = *p;
	return qu_integer_newi (v);
}



/*----------------------------------------------------------------------------
	DL stuffs
----------------------------------------------------------------------------*/
static void    *Handle;
static qu_dict  Resolved;


static qu_any qu_curses_getcfunc (qu_module self, qu_string name, qu_string spec)
{
	return qu_module_get_cfunc (self, Handle, name, spec, Resolved);
}


static void qu_curses_atexit ()
{
	if (Handle)
		qu_dl_close (Handle);
}



/*----------------------------------------------------------------------------
	functions
----------------------------------------------------------------------------*/

static qu_any qu_curses_call (int argc, qu_any *argv)
{
	return qu_dlib_call (Handle, argc, argv);
}


/*) sub COLORS -> Integer
	Return the number of colors.
*/
static qu_integer qu_curses_COLORS ()
{
	return qu_curses__int (COLORS_p);
}


/*) sub COLOR_PAIRS -> Integer
	Return the number of color-pairs.
*/
static qu_integer qu_curses_COLOR_PAIRS ()
{
	return qu_curses__int (COLOR_PAIRS_p);
}


/*) sub LINES -> Integer
	Return the number of lines.
*/
static qu_integer qu_curses_LINES ()
{
	return qu_curses__int (LINES_p);
}


/*) sub COLS -> Integer
	Return the number of columns.
*/
static qu_integer qu_curses_COLS ()
{
	return qu_curses__int (COLS_p);
}


/*) sub TABSIZE -> Integer
	Return the size of one tab.
*/
static qu_integer qu_curses_TABSIZE ()
{
	return qu_curses__int (TABSIZE_p);
}


/*) sub ESCDELAY -> Integer
	I dont know that this is but it looks appealing.
*/
static qu_integer qu_curses_ESCDELAY ()
{
	return qu_curses__int (ESCDELAY_p);
}


#if 0
static qu_pointer qu_curses_stdscr ()
{
	return qu_curses__ptr (stdscr_p);
}


static qu_pointer qu_curses_curscr ()
{
	return qu_curses__ptr (curscr_p);
}


static qu_pointer qu_curses_newscr ()
{
	return qu_curses__ptr (newscr_p);
}
#endif


/*) sub chtype (c, a: Integer, i = 0) -> Integer
	Compute a v{chtype} integer.

	v{c} is the character. This may be an integer or a character.
	If it is a character then -{String.asc} is called to get the
	integer value.

	v{a} is the attribute.

	v{i} is the color.

	@exc
	.EArgValue
*/
static qu_integer qu_curses_chtype (int argc, qu_any *argv)
{
	qu_any x;
	int    c, a, i = 0;

	QU_ARG_VCOUNT (2, 3);
	x = argv[0];
	if (qu_stringy_is (x))
		c = ((qu_string) x)->str[0];
	else
		QU_ARG_vint (0, c);
	QU_ARG_vint (1, a);
	if (argc > 2)
		QU_ARG_vint (2, i);
	a = a | c | (i << 8);
	return qu_integer_newi (a);
}


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Curses_init (qu_module m)
{
	Handle = qu_dl_open_ex ("curses", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (Handle);
	m->getcfunc = qu_curses_getcfunc;

	Resolved = qu_dict_new_static (512);

	COLORS_p      = qu_dl_sym (Handle, "COLORS");
	COLOR_PAIRS_p = qu_dl_sym (Handle, "COLOR_PAIRS");
	LINES_p       = qu_dl_sym (Handle, "LINES");
	COLS_p        = qu_dl_sym (Handle, "COLS");
	TABSIZE_p     = qu_dl_sym (Handle, "TABSIZE");
	ESCDELAY_p    = qu_dl_sym (Handle, "ESCDELAY");
#if 0
	stdscr_p      = qu_dl_sym (Handle, "stdscr");
	curscr_p      = qu_dl_sym (Handle, "curscr");
	newscr_p      = qu_dl_sym (Handle, "newscr");
#endif

	qu_module_add_function (m, "call"       , qu_curses_call       , -1);
	qu_module_add_function (m, "COLORS"     , qu_curses_COLORS     ,  0);
	qu_module_add_function (m, "COLOR_PAIRS", qu_curses_COLOR_PAIRS,  0);
	qu_module_add_function (m, "LINES"      , qu_curses_LINES      ,  0);
	qu_module_add_function (m, "COLS"       , qu_curses_COLS       ,  0);
	qu_module_add_function (m, "TABSIZE"    , qu_curses_TABSIZE    ,  0);
	qu_module_add_function (m, "ESCDELAY"   , qu_curses_ESCDELAY   ,  0);

#if 0
	qu_module_add_function (m, "__stdscr"   , qu_curses_stdscr     ,  0);
	qu_module_add_function (m, "__curscr"   , qu_curses_curscr     ,  0);
	qu_module_add_function (m, "__newscr"   , qu_curses_newscr     ,  0);
#endif

	qu_module_add_function (m, "chtype"     , qu_curses_chtype     , -1);


#ifdef DEBUG
#define NCURSES_ATTR_SHIFT       8
#define NCURSES_BITS(mask,shift) ((mask) << ((shift) + NCURSES_ATTR_SHIFT))
	printf ("A_ATTRIBUTES = %lx\n", NCURSES_BITS(~(1UL - 1UL),0));
	printf ("A_CHARTEXT   = %lx\n", (NCURSES_BITS(1UL,0) - 1UL));
	printf ("A_COLOR      = %lx\n", NCURSES_BITS(((1UL) << 8) - 1UL,0));
	printf ("A_STANDOUT   = %lx\n", NCURSES_BITS(1UL,8));
	printf ("A_UNDERLINE  = %lx\n", NCURSES_BITS(1UL,9));
	printf ("A_REVERSE    = %lx\n", NCURSES_BITS(1UL,10));
	printf ("A_BLINK      = %lx\n", NCURSES_BITS(1UL,11));
	printf ("A_DIM        = %lx\n", NCURSES_BITS(1UL,12));
	printf ("A_BOLD       = %lx\n", NCURSES_BITS(1UL,13));
	printf ("A_ALTCHARSET = %lx\n", NCURSES_BITS(1UL,14));
	printf ("A_INVIS      = %lx\n", NCURSES_BITS(1UL,15));
	printf ("A_PROTECT    = %lx\n", NCURSES_BITS(1UL,16));
	printf ("A_HORIZONTAL = %lx\n", NCURSES_BITS(1UL,17));
	printf ("A_LEFT       = %lx\n", NCURSES_BITS(1UL,18));
	printf ("A_LOW        = %lx\n", NCURSES_BITS(1UL,19));
	printf ("A_RIGHT      = %lx\n", NCURSES_BITS(1UL,20));
	printf ("A_TOP        = %lx\n", NCURSES_BITS(1UL,21));
	printf ("A_VERTICAL   = %lx\n", NCURSES_BITS(1UL,22));
#endif

	atexit (qu_curses_atexit);
	return 1;
}
