/* Date.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Date
	This module provides basic functions to deal with date/time.

	Like the %{Time} module, a time value is the number of seconds after
	the Epoch. It has no notion of timezone, daylight saving, etc.

	Unlike the %{Time} module, the Epoch is set to 01-Jan-2000 00:00:00.
	The lower limit is 00-Jan-0000 00:00:00 and the upper limit is
	31-Dec-9999 59:59:59. Thus dates before the year 2000 have negative
	values.

	The time value does not include microseconds for efficiency.
	The value is an -{Integer} instead of a -{Float}.

	The Date module assumes that dates around the year 2000 are more common
	thus in most cases the value should be an Int (which does not allocate
	memory).

	@xmp
		use Date, Time

		t = Date.now
		println (t)

		t = Time.now
		println (t)

	@out
		63311915467 <Class Sys.Long>
		1144782667.472353 <Class Sys.Float>
*/

#define AVAILABLE 1

#if SIZEOF_INT >= 8
#  define DATE                int
#  define QU_INTEGER_NEWDATE  qu_integer_newi
#  define QU_INTEGER_TODATE   qu_integer_toi
#  define QU_ARG_DATE         QU_ARG_int
#  define QU_ARG_VDATE        QU_ARG_vint
#  define OFFSET              63113904000
#  define VALUEMAX            (315537897600 - OFFSET)
#  define VALUEMIN            (0 - OFFSET)
#  define YEARMAX             9999
#  define MULV                1
#  define MUL10               10000000000
#  define MUL8                100000000
#elif SIZEOF_LONG >= 8
#  define DATE                long
#  define QU_INTEGER_NEWDATE  qu_integer_newl
#  define QU_INTEGER_TODATE   qu_integer_tol
#  define QU_ARG_DATE         QU_ARG_long
#  define QU_ARG_VDATE        QU_ARG_vlong
#  define OFFSET              63113904000L
#  define VALUEMAX            (315537897600L - OFFSET)
#  define VALUEMIN            (0L - OFFSET)
#  define YEARMAX             9999
#  define MULV                1L
#  define MUL10               10000000000L
#  define MUL8                100000000L
#elif defined (HAVE_LONG_LONG)
#  define DATE                long long
#  define QU_INTEGER_NEWDATE  qu_integer_newll
#  define QU_INTEGER_TODATE   qu_integer_toll
#  define QU_ARG_DATE         QU_ARG_longlong
#  define QU_ARG_VDATE        QU_ARG_vlonglong
#  define OFFSET              63113904000LL
#  define VALUEMAX            (315537897600LL - OFFSET)
#  define VALUEMIN            (0LL - OFFSET)
#  define YEARMAX             9999
#  define MULV                1LL
#  define MUL10               10000000000LL
#  define MUL8                100000000LL
#else
#  undef AVAILABLE
#endif


#ifndef AVAILABLE

int qu_Date_init (qu_module m)
{
	QU_THROWZ (QuESystem, "the Date module require 8-bytes integer support");
}


#else /*AVAILABLE*/


/*----------------------------------------------------------------------------
	internals
----------------------------------------------------------------------------*/

/* stuffs */
static int DaysInMonth [12]     = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
static int DaysBeforeMonth [12] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
static int DaysBeforeYear [YEARMAX + 1];


inline static int isleap (int y)
{
	assert (y >= 0);
	return (y % 4) == 0 && ((y % 100) || (y % 400) == 0);
}


inline static int monthdays (int y, int m)
{
	assert (0 <= m && m <= 11);
	if (m == 1 && isleap (y))
		return 29;
	return DaysInMonth [m];
}


inline static int daysbeforemonth (int y, int m)
{
	assert (0 <= m && m <= 11);
	if (m > 1 && isleap (y))
		return DaysBeforeMonth [m] + 1;
	return DaysBeforeMonth [m];
}


inline static int daysbeforeyear (int y)
{
	assert (0 <= y && y <= YEARMAX);
	return DaysBeforeYear [y];
}


inline static DATE ymd2v (int y, int m, int d)
{
	return (MULV * 86400 * (daysbeforeyear (y) + daysbeforemonth (y, m) + d - 1)) - OFFSET;
}


inline static DATE ymdhns2v (int y, int m, int d, int h, int n, int s)
{
	return (MULV * 86400 * (daysbeforeyear (y) + daysbeforemonth (y, m) + d - 1)
		+ h * 3600 + n * 60 + s) - OFFSET;
}


inline static int weekday (int y, int m, int d)
{
	return ((ymd2v (y, m, d) + OFFSET) / 86400 + 6) % 7;
}



#define v2minute(v) ((((v) + OFFSET) % 3600 ) / 60)
#define v2hour(v)   ((((v) + OFFSET) % 86400) / 3600)
#define v2second(v) (((v) + OFFSET) % 60)


static void v2ymd (DATE v, int *year, int *mon, int *day)
{
#define ndelta 5
	static int delta [ndelta] = {1000, 400, 100, 4, 1};
	int y = 0, m = 0, i, x;

	assert (VALUEMIN <= v && v <= VALUEMAX);
	v += OFFSET;
	v /= 86400;
	for (i = 0; i < ndelta; ++i)
	{
		int d = delta [i];
		while (v >= daysbeforeyear (y + d))
			y += d;
	}

	/* switch to int */
	x = v - daysbeforeyear (y);
	assert (0 <= x && x < (isleap (y) ? 366 : 365));
	while (m < 11 && x >= daysbeforemonth (y, m + 1))
		++m;

	*year = y;
	*mon  = m;
	*day  = x - daysbeforemonth (y, m) + 1;
}



static DATE now2v ()
{
	time_t    t = time (NULL);
	struct tm u;

	localtime_r (&t, &u);
	return ymdhns2v (u.tm_year + 1900, u.tm_mon, u.tm_mday, u.tm_hour, u.tm_min, u.tm_sec);
}



static void now2ymd (int *y, int *m, int *d)
{
	time_t    t = time (NULL);
	struct tm u;

	localtime_r (&t, &u);
	*y = u.tm_year + 1900;
	*m = u.tm_mon;
	*d = u.tm_mday;
}



static void now2ymdhns (int *y, int *m, int *d, int *h, int *n, int *s)
{
	time_t    t = time (NULL);
	struct tm u;

	localtime_r (&t, &u);
	*y = u.tm_year + 1900;
	*m = u.tm_mon;
	*d = u.tm_mday;
	*h = u.tm_hour;
	*n = u.tm_min;
	*s = u.tm_sec;
}



static int obj2ymd (qu_any val, int *y, int *m, int *d)
{
	DATE v = QU_INTEGER_TODATE (val);

	if (errno == 0 && VALUEMIN <= v && v <= VALUEMAX)
	{
		v2ymd (v, y, m, d);
		return 1;
	}
	return 0;
}



static int obj2ymdhns (qu_any val, int *y, int *m, int *d, int *h, int *t, int *s)
{
	DATE v = QU_INTEGER_TODATE (val);

	if (errno == 0 && VALUEMIN <= v && v <= VALUEMAX)
	{
		v2ymd (v, y, m, d);
		*h = v2hour   (v);
		*t = v2minute (v);
		*s = v2second (v);
		return 1;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	make
----------------------------------------------------------------------------*/

/*) sub now -> Integer
	Get the current local time measured in seconds since Epoch.

	@noexc

	@xmp
		use Date
		println (Date.now)
		println (Date.date)

	@out
		63311915984
		63311932800
*/
static qu_integer Date_now ()
{
	DATE v = now2v ();
	return QU_INTEGER_NEWDATE (v);
}



/*) sub date -> Integer
	This is the same as #{now} except that the hour, minute and second are
	ignored.

	@noexc

	@xmp
		use Date
		println (Date.now)
		println (Date.date)

	@out
		63311915984
		63311932800
*/
static qu_integer Date_date ()
{
	DATE v;
	int  y, m, d;

	now2ymd (&y, &m, &d);
	v = ymd2v (y, m, d);
	return QU_INTEGER_NEWDATE (v);
}



/*) sub make (y = year, m = mon, d = day, h = 0, n = 0, s = 0, u = 0) -> Integer
	Build a time value from the arguments.

	The result is a value measured from the Epoch.

	Missing argument means to use the value of the current local time.

	@xmp
		use Date
		println (Date.make (2004, 9, 1))
		println (Date.make (d:1 y:2004 m:9))

	@out
		63263808000
		63263808540
*/
static qu_integer Date_make (int argc, qu_any *argv)
{
	static char *argx[] =
	{
		":",
		"i", "y",
		"i", "m",
		"i", "d",
		"i", "h",
		"i", "n",
		"i", "s",
		"i", "u",
		NULL
	};

	DATE v;
	int  y, m, d, h, n, s;

	now2ymdhns (&y, &m, &d, &h, &n, &s);
	h = n = s = 0;

	QU_ARG_XPARSE (&y, &m, &d, &h, &n, &s);
	QU_ARG_CHECK (
		   0 <= y && y <= YEARMAX
		&& 0 <= m && m <= 11
		&& 1 <= d && d <= monthdays (y, m)
		&& 0 <= h && h <= 23
		&& 0 <= n && n <= 59
		&& 0 <= s && s <= 59);

	v = ymdhns2v (y, m, d, h, n, s);
	return QU_INTEGER_NEWDATE (v);
}



/* return 1 if pattern matches, 0 otherwise */
static int _ispattern (const char *s, const char *format)
{
	register char f, c;

	while ((f = *format++) != '\0')
	{
		if ((c = *s++) == '\0')
			return 0;

		switch (f)
		{
			case '%':
				if (qu_ascii_isdigit (c))
					continue;
				return 0;

			case '$':
				if (qu_ascii_isalpha (c))
					continue;
				return 0;

			case '!':
				if ((qu_ascii_isdigit (c)) || (c == ' '))
					continue;
				return 0;

			default:
				if (f != c)
					return 0;
		}
	}
	return 1;
}



/*) sub value (x: String, defval = 0) -> Integer
	Convert v{x} to a time value.

	If the format is not understood, then if v{defval} is supplied then it is
	returned, otherwise ESyntax is thrown.

	Formats understood:
	@ul
	. RFC-822 : Sun, 06 Nov 1994 08:49:37 GMT
	. RFC-850 : Sunday, 06-Nov-94 08:49:37 GMT
	. ANSI C asctime() : Sun Nov  6 08:49:37 1994
	. extended: basically d-m-y h:m:s am/pm GMT

	@exc
	.EArgValue
	.ESyntax format is not understood

	@xmp
		use Date

		v = Date.value ('9-Jun-2004')
		println (v)
		println (Date.year  (v))
		println (Date.month (v))
		println (Date.day   (v))

	@out
		63253958400
		2004
		5
		9
*/
static qu_integer Date_value (int argc, qu_any *argv)
{
	qu_any    x;
	struct tm e;
	DATE      v;
	int       g = 0;
	char     *h, *m, *s;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, x, qu_string);

	s = ((qu_string) x)->str;

#define K(c) ((s[c] - '0') * 10 + (s[c+1] - '0'))
#define Y  e.tm_year
#define M  e.tm_mon
#define D  e.tm_mday
#define H  e.tm_hour
#define N  e.tm_min
#define S  e.tm_sec

#define skipspace_(s)  QU_BEGIN while (qu_ascii_isspace (*(s))) (s)++; QU_END
#define skipto_(s, is) QU_BEGIN while (*(s)) { if (is (*(s))) break; (s)++; } QU_END

	skipspace_ (s);
	if (qu_ascii_isdigit (*s))
	{
		/*
			A bit more flexible format.
			Never allow the idiotic 10-11-03 notation!
			10...nov...[1995...[01...[19...[59...[GMT]]]]] */

		D = *s++ - '0';
		if (qu_ascii_isdigit (*s))
			D = D * 10 + *s++ - '0';

		skipto_ (s, qu_ascii_isalpha);
		m = (char *) s;
		while (qu_ascii_isalpha (*s))
			s++;
		if (m == s)
			goto nope;

		skipto_ (s, qu_ascii_isdigit);

		Y = 0;
		if (qu_ascii_isdigit (*s))
		{
			Y = *s++ - '0';
			if (qu_ascii_isdigit (*s))
			{
				Y = Y * 10 + *s++ - '0';
				if (qu_ascii_isdigit (*s))
				{
					Y = Y * 10 + *s++ - '0';
					if (!qu_ascii_isdigit (*s))
						goto nope;
					Y = Y * 10 + *s++ - '0';
				}
			}
		}
		if (Y == 0)
		{
			time_t    t = time (NULL);
			struct tm u;

			localtime_r (&t, &u);
			Y = u.tm_year + 1900;
		}

		H = N = S = 0;

		if (qu_ascii_isdigit (*s))
		{
			H = *s++ - '0';
			if (qu_ascii_isdigit (*s))
				H = H * 10 + *s++ - '0';
			skipto_ (s, qu_ascii_isdigit);
			if (qu_ascii_isdigit (*s))
			{
				N = *s++ - '0';
				if (qu_ascii_isdigit (*s))
					N = N * 10 + *s++ - '0';
				skipto_ (s, qu_ascii_isdigit);
				if (qu_ascii_isdigit (*s))
				{
					S = *s++ - '0';
					if (qu_ascii_isdigit (*s))
						S = S * 10 + *s++ - '0';
				}
			}
			if (strchr (s, 'P') || strchr (s, 'p'))
			{
				if (H < 12)
					H += 12;
			}
			else if (strchr (s, 'A') || strchr (s, 'a'))
			{
				if (H == 12)
					H -= 12;
			}
		}
		g = strchr (s, 'G') || strchr (s, 'g');
	}
	else
	{
		s = strchr (s, ' ');
		if (s == NULL)
			goto nope;

		++s;

		if (_ispattern (s, "%% $$$ %%%% %%:%%:%% GMT"))
		{
			/* RFC 822
			Sun, 06 Nov 1994 08:49:37 GMT */

			g = 1;
			h = (char *) s + 12;
			m = (char *) s + 3;
			D = K(0);
			Y = K(7) * 100 + K(9);
		}
		else if (_ispattern (s, "%%-$$$-%% %%:%%:%% GMT"))
		{
			/* RFC 850
			Sunday, 06-Nov-94 08:49:37 GMT */

			g = 1;
			h = (char *) s + 10;
			m = (char *) s + 3;
			D = K(0);
			if ((Y = K(7)) < 70)
				Y += 1900;
			else
				Y += 2000;
		}
		else if (_ispattern (s, "$$$ !% %%:%%:%% %%%%"))
		{
			/* ANSI C asctime()
			Sun Nov  6 08:49:37 1994 */

			g = 0;
			h = (char *) s + 7;
			m = (char *) s;
			D = (s[4] == ' ') ? (s[5] - '0') : K(4);
			Y = K(16) * 100 + K(18);
		}
		else
			goto nope;

		s = h;
		H = K(0);
		N = K(3);
		S = K(6);
	}

	switch (qu_ascii_tolower (m[0]))
	{
		case 'j': M = (qu_ascii_tolower (m[1]) == 'a') ? 0 : (qu_ascii_tolower (m[2]) == 'n') ? 5 : 6; break;
		case 'a': M = (qu_ascii_tolower (m[1]) == 'p') ? 3 : 7; break;
		case 'm': M = (qu_ascii_tolower (m[2]) == 'r') ? 2 : 4; break;
		case 's': M =  8; break;
		case 'f': M =  1; break;
		case 'o': M =  9; break;
		case 'n': M = 10; break;
		case 'd': M = 11; break;
		default : goto nope;
	}

	v = ymdhns2v (Y, M, D, H, N, S) - (g ? timezone : 0);
	return QU_INTEGER_NEWDATE (v);

  nope:
	if (argc > 1)
	{
		DATE v;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		return argv[0];
	}
	QU_THROW (QuESyntax);

#undef S
#undef T
#undef H
#undef D
#undef M
#undef Y
#undef I

}



/*----------------------------------------------------------------------------
	field value
----------------------------------------------------------------------------*/

/*) sub year (v = now) -> Int
	Return the year of the time value v{v}.

	If v{v} is not specified then return the current local time year.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.year)

	@out
		2006
*/
static qu_int Date_year (int argc, qu_any *argv)
{
	int y;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		y = u.tm_year + 1900;
	}
	else
	{
		DATE v;
		int  m, d;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		v2ymd (v, &y, &m, &d);
	}
	return qu_int_newi (y);
}



/*) sub month (v = now) -> Int
	Return the month of the time value v{v}.

	If v{v} is not specified then return the current local time month.

	@warn
	Month range is C{0 .. 11}.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.month)

	@out
		3
*/
static qu_int Date_month (int argc, qu_any *argv)
{
	int m;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		m = u.tm_mon;
	}
	else
	{
		DATE v;
		int  y, d;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		v2ymd (v, &y, &m, &d);
	}
	return qu_int_newi (m);
}



/*) sub day (v = now) -> Int
	Return the day of the time value v{v}.

	If v{v} is not specified then return the current local time day.

	@warn
	Day range is C{1 .. 31}.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.day)

	@out
		14
*/
static qu_int Date_day (int argc, qu_any *argv)
{
	int d;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		d = u.tm_mday;
	}
	else
	{
		DATE v;
		int  y, m;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		v2ymd (v, &y, &m, &d);
	}
	return qu_int_newi (d);
}



/*) sub hour (v = now) -> Int
	Return the hour of the time value v{v}.

	If v{v} is not specified then return the current local time hour.

	@warn
	Hour range is C{0 .. 23}.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.hour)

	@out
		22
*/
static qu_int Date_hour (int argc, qu_any *argv)
{
	int h;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		h = u.tm_hour;
	}
	else
	{
		DATE v;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		h = v2hour (v);
	}
	return qu_int_newi (h);
}



/*) sub minute (v = now) -> Int
	Return the minute of the time value v{v}.

	If v{v} is not specified then return the current local time minute.

	@warn
	Minute range is C{0 .. 59}.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.minute)

	@out
		57
*/
static qu_int Date_minute (int argc, qu_any *argv)
{
	int n;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		n = u.tm_min;
	}
	else
	{
		DATE v;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		n = v2minute (v);
	}
	return qu_int_newi (n);
}



/*) sub second (v = now) -> Int
	Return the second of the time value v{v}.

	If v{v} is not specified then return the current local time second.

	@warn
	Second range is C{0 .. 59}.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.second)

	@out
		49
*/
static qu_int Date_second (int argc, qu_any *argv)
{
	int s;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		s = u.tm_sec;
	}
	else
	{
		DATE v;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		s = v2second (v);
	}
	return qu_int_newi (s);
}



/*) sub weekday (v = now) -> Int
	Return the weekday of the time value v{v}.

	If v{v} is not specified then return the current local time weekday.

	@warn
	Weekday range is C{0 .. 6} where 0 is Sunday.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.weekday)

	@out
		4
*/
static qu_int Date_weekday (int argc, qu_any *argv)
{
	int w;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		w = u.tm_wday;
	}
	else
	{
		DATE v;
		int  y, m, d;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		v2ymd (v, &y, &m, &d);
		w = weekday (y, m, d);
	}
	return qu_int_newi (w);
}



/*) sub yday (v = now) -> Int
	Return the day number.

	If v{v} is not specified then use the current local time.

	@warn
	The range is C{1 .. 366}. Thus 1-Jan is 1.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.yday)

	@out
		211
*/
static qu_int Date_yday (int argc, qu_any *argv)
{
	int y, m, d;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		now2ymd (&y, &m, &d);
	}
	else
	{
		DATE v;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		v2ymd (v, &y, &m, &d);
	}

	d += daysbeforemonth (y, m);
	return qu_int_new (d);
}



/*) sub toa (v: Integer) -> Array
	Return an Array consisting of the year, month, day, hour, minute, second,
	and weekday of the time value v{v}.

	If v{v} is not specified then use the current local time.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.toa)

	@out
		[2006, 3, 12, 2, 5, 14, 3]
*/
static qu_array Date_toa (int argc, qu_any *argv)
{
	int y, m, d, h, n, s, w;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		now2ymdhns (&y, &m, &d, &h, &n, &s);
	}
	else
	{
		DATE v;

		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
		v2ymd (v, &y, &m, &d);
		h = v2hour   (v);
		n = v2minute (v);
		s = v2second (v);
	}

	w = weekday (y, m, d);

	return qu_array_new_ex (7,
		qu_int_newi (y), qu_int_newi (m), qu_int_newi (d),
		qu_int_newi (h), qu_int_newi (n), qu_int_newi (s), qu_int_new (w));
}



/*----------------------------------------------------------------------------
	format
----------------------------------------------------------------------------*/

/* the buffer size required to format a HTTP time */
#define DATE_WEBSIZE 64

static const char MonthLong[12][9] =
{
	"January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December"
};


static const char MonthShort[12][4] =
{
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};


static const char DayLong[7][9] =
{
	"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
};


static const char DayShort[7][4] =
{
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};



/*) sub format (fmt, v = Date.now, names: Array = const []) -> String
	Format the time v{v} using a specified format.

	If v{v} is not supplied then the current local time is used.

	@warn
	The format is different from C's strftime.

	@p
	Unless you supply v{names} then English names for weekdays and months are used.

	v{names} must be an -{Array} of four Array objects:
	@args
	.first  at least  7 Strings of the full name of weekdays
	.second at least  7 Strings of the abbreviated name of weekdays
	.third  at least 12 Strings of the full name of months
	.fourth at least 12 Strings of the abbreviated name of months

	@p
	v{fmt} is the format. Special characters known are:
	@rcc
	.C{%%} | A literal '%' character.

	.C{%w} | The abbreviated weekday name according to the current locale.
	.C{%W} | The full weekday name according to the current locale.

    .C{%d} | The day of the month (range 1 to 31).
    .C{%D} | The day of the month as two digits.

	.C{%m} | The abbreviated month name in English or as supplied.
	.C{%M} | The full month name in English or as supplied.

	.C{%y} | The year including.
	.C{%Y} | An alias of C{%y}.

	.C{%h} | The hour (range 00 to 12).
	.C{%H} | The hour (range 00 to 23).

	.C{%p} | 'am' or 'pm' based on the hour. Should not be used with C{%H}.
	.C{%P} | Same as C{%p} but uses 'AM' or 'PM'.

	.C{%n} | The minute (range 00 to 59).
	.C{%N} | An alias of C{%t}.

	.C{%s} | The second (range 00 to 59).
	.C{%S} | An alias of C{%s}.

	.C{%z} | The time zone name using #{tzname} C{(0)}.
	.C{%Z} | The time zone name using #{tzname} C{(1)}.

	@exc
	.EArgValue
	.EFormat

	@xmp
		use Date
		println (Date.format ("%w, %d-%m-%y %H:%n:%s%P"))

	@out
		Wed, 12-Apr-2006 03:15:11AM
*/
static qu_string Date_format (int argc, qu_any *argv)
{
	qu_string  fmt;
	qu_string  r, name;
	long       l;
	char      *i, *j, buf [64];
	int        y, m, d, h, n, s, w, ok;
	qu_array   wlong = 0, wshort = 0, mlong = 0, mshort = 0;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, fmt, qu_string);

	if (argc == 1)
	{
		now2ymdhns (&y, &m, &d, &h, &n, &s);
	}
	else
	{
		QU_ARG_CHECK (obj2ymdhns (argv[1], &y, &m, &d, &h, &n, &s));
		if (argc > 2)
		{
			qu_array names = (qu_array) argv [2];
			QU_ARG_CHECK (qu_array_is (names));
			if (names->len)
			{
				QU_ARG_CHECK (names->len >= 4);

				wlong  = qu_array_GETITEM (names, 0);
				wshort = qu_array_GETITEM (names, 1);
				mlong  = qu_array_GETITEM (names, 2);
				mshort = qu_array_GETITEM (names, 3);

				QU_ARG_CHECK (wlong->len >= 7 && wshort->len >= 7
					&& mlong->len >= 12 && mshort->len >= 12);
			}
		}
	}

	l = fmt->len;
	r = qu_string_buffer_new (l);
	i = j = fmt->str;
	while (l)
	{
		if (*i == '%')
		{
			++i;
			--l;
			switch (*i)
			{
				case '%':
					ok = qu_string_buffer_putc (r, '%');
					break;

				case 'w':
					w = weekday (y, m, d);
					if (wshort)
					{
						name = qu_array_GETITEM (wshort, w);
						QU_ARG_TYPE (name, qu_string);
						ok = qu_string_buffer_puts (r, name);
					}
					else
					{
						ok = qu_string_buffer_putz (r, DayShort [w]);
					}
					break;

				case 'W':
					w = weekday (y, m, d);
					if (wlong)
					{
						name = qu_array_GETITEM (wlong, w);
						QU_ARG_TYPE (name, qu_string);
						ok = qu_string_buffer_puts (r, name);
					}
					else
					{
						ok = qu_string_buffer_putz (r, DayLong [w]);
					}
					break;

				case 'd':
					sprintf (buf, "%d", d);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 'D':
					sprintf (buf, "%02d", d);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 'm':
					if (mshort)
					{
						name = qu_array_GETITEM (mshort, m);
						QU_ARG_TYPE (name, qu_string);
						ok = qu_string_buffer_puts (r, name);
					}
					else
					{
						ok = qu_string_buffer_putz (r, MonthShort [m]);
					}
					break;

				case 'M':
					if (mlong)
					{
						name = qu_array_GETITEM (mlong, m);
						QU_ARG_TYPE (name, qu_string);
						ok = qu_string_buffer_puts (r, name);
					}
					else
					{
						ok = qu_string_buffer_putz (r, MonthLong [m]);
					}
					break;

				case 'y':
				case 'Y':
					sprintf (buf, "%d", y);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 'h':
					sprintf (buf, "%02d", (h > 12) ? (h - 12) : h);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 'H':
					sprintf (buf, "%02d", h);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 'p':
					ok = qu_string_buffer_putz (r, (h > 12) ? "pm" : "am");
					break;

				case 'P':
					ok = qu_string_buffer_putz (r, (h > 12) ? "PM" : "AM");
					break;

				case 'n':
				case 'N':
					sprintf (buf, "%02d", n);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 's':
				case 'S':
					sprintf (buf, "%02d", s);
					ok = qu_string_buffer_putz (r, buf);
					break;

				case 'z':
					ok = 1;
					if (tzname [0])
						ok = qu_string_buffer_putz (r, tzname[0]);
					break;

				case 'Z':
					ok = 1;
					if (tzname [1])
						ok = qu_string_buffer_putz (r, tzname[1]);
					break;

				default:
					QU_THROW (QuEFormat);
			}
		}
		else
		{
			ok = qu_string_buffer_putc (r, *i);
		}
		QU_ENSURE (ok);
		++i;
		--l;
	}
	return qu_string_buffer_return (r);
}



/*) sub web (v = Date.now) -> String
	Return a formatted HTTP time.

	If v{v} is not supplied then the current local time is used.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.web)

	@out
		Tue, 11 Apr 2006 18:21:37 GMT
*/
static qu_string Date_web (int argc, qu_any *argv)
{
	char buf [DATE_WEBSIZE + 1];
	int  y, m, d, h, n, s;
	DATE v;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		v = now2v ();
	}
	else
	{
		QU_ARG_VDATE (0, v);
		QU_ARG_CHECK (VALUEMIN <= v && v <= VALUEMAX);
	}

	v += timezone;

	v2ymd (v, &y, &m, &d);
	h = v2hour   (v);
	n = v2minute (v);
	s = v2second (v);

	snprintf (buf, DATE_WEBSIZE, "%s, %.2d %s %d %.2d:%.2d:%.2d GMT",
	          DayShort [weekday (y, m, d)], d, MonthShort [m], y, h, n, s);
	buf [DATE_WEBSIZE] = '\0';
	return qu_string_newz (buf);
}



/*) sub smtp (v = Date.now) -> String
	Return a formatted SMTP time.

	If v{v} is not supplied then the current local time is used.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.smtp)

	@out
		Wed, 12 Apr 2006 01:21 +0700
*/
static qu_string Date_smtp (int argc, qu_any *argv)
{
	char buf [DATE_WEBSIZE + 1];
	int  y, m, d, h, n, s;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
		now2ymdhns (&y, &m, &d, &h, &n, &s);
	else
		QU_ARG_CHECK (obj2ymdhns (argv[0], &y, &m, &d, &h, &n, &s));

	snprintf (buf, DATE_WEBSIZE, "%s, %d %s %d %02d:%02d %+03d00",
			  DayShort [weekday (y, m, d)], d, MonthShort [m],
	          y, h, n, (int) -timezone / 3600);
	buf [DATE_WEBSIZE] = '\0';
	return qu_string_newz (buf);
}



/*) sub ymd (v = Date.now) -> Int
	Return a formatted time which the first 4 digits is the year, the next
	2 digits is the month and the last 2 digits is the day.

	If v{v} is not supplied then the current local time is used.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.ymd)

	@out
		20060312
*/
static qu_int Date_ymd (int argc, qu_any *argv)
{
	int y, m, d;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
		now2ymd (&y, &m, &d);
	else
		QU_ARG_CHECK (obj2ymd (argv[0], &y, &m, &d));

	d = y * 10000 + m * 100 + d;
	assert (d < QU_INT_MAX);
	return qu_int_newi (d);
}



/*) sub ymdhns (v = Date.now) -> Integer
	Return a formatted time which the first 4 digits is the year, the next
	2 digits is the month, the next 2 digits is the day, the next 2 digits
	is the hour, the next 2 digits is the minute and the last 2 digits is
	the second.

	If v{v} is not supplied then the current local time is used.

	@exc
	.EArgValue

	@xmp
		use Date
		println (Date.ymdhns)

	@out
		20060317002856
*/
static qu_integer Date_ymdhns (int argc, qu_any *argv)
{
	int       y, m, d, h, n, s;
	DATE v;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
		now2ymdhns (&y, &m, &d, &h, &n, &s);
	else
		QU_ARG_CHECK (obj2ymdhns (argv[0], &y, &m, &d, &h, &n, &s));

	v = y * MUL10 + m * MUL8 + d * 1000000 + h * 10000 + n * 100 + s;
	return qu_integer_newll (v);
}



/*----------------------------------------------------------------------------
	info
----------------------------------------------------------------------------*/

/*) sub isleap (y = year) -> Boolean
	Whether the year is a leap year.

	v{y} must be within C{0 .. 9999}.
	If not specified then use the current year.

	@exc
	.EArgValue not a valid year
*/
static qu_boolean Date_isleap (int argc, qu_any *argv)
{
	int y;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		y = u.tm_year + 1900;
	}
	else
	{
		QU_ARG_vint (0, y);
		QU_ARG_CHECK (y >= 0);
	}
	return isleap (y) ? QuTrue : QuFalse;
}



/*) sub ydays (y = year) -> Int
	Return the number of days in a specified year.

	v{y} must be within C{0 .. 9999}.
	If not specified then use the current year.

	Thus this function returns 366 if the year is a leap year, 365 otherwise.

	@exc
	.EArgValue not a valid year
*/
static qu_int Date_ydays (int argc, qu_any *argv)
{
	int y, d;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		y = u.tm_year + 1900;
	}
	else
	{
		QU_ARG_vint (0, y);
		QU_ARG_CHECK (y >= 0);
	}
	d = isleap (y) ? 366 : 365;
	return qu_int_newi (d);
}



/*) sub ybdays (year: Integer) -> Int
	Return the number of days since Epoch.

	v{y} must be within C{0 .. 9999}.
	If not specified then use the current year.

	@exc
	.EArgValue not a valid year
*/
static qu_int Date_ybdays (int argc, qu_any *argv)
{
	int y, d;

	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		y = u.tm_year + 1900;
	}
	else
	{
		QU_ARG_vint (0, y);
		QU_ARG_CHECK (y >= 0);
	}
	d = daysbeforeyear (y);
	return qu_int_newi (d);
}



/*) sub mdays (y = year, m = month) -> Int
	Return the number of days in a specified month.

	v{y} must be within C{0 .. 9999} and v{m} within C{0 .. 11}.
	If not specified then use the current year/month.

	@exc
	.EArgValue not a valid year or month
*/
static qu_int Date_mdays (int argc, qu_any *argv)
{
	int y, m, d;

	QU_ARG_VCOUNT (0, 2);
	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		y = u.tm_year + 1900;
		m = u.tm_mon;
	}
	else
	{
		QU_ARG_vint (0, y);
		QU_ARG_CHECK (y >= 0);
		if (argc > 1)
		{
			QU_ARG_vint (1, m);
			QU_ARG_CHECK (m && m <= 11);
		}
		else
		{
			time_t    t = time (NULL);
		 	struct tm u;

			localtime_r (&t, &u);
			m = u.tm_mon;
		}
	}
	d = monthdays (y, m);
	return qu_int_newi (d);
}



/*) sub mbdays (y = year, m = month) -> Int
	Return the number of days since 1-Jan.

	v{y} must be within C{0 .. 9999} and v{m} within C{0 .. 11}.
	If not specified then use the current year/month.

	@exc
	.EArgValue not a valid year or month
*/
static qu_int Date_mbdays (int argc, qu_any *argv)
{
	int y, m, d;

	QU_ARG_VCOUNT (0, 2);
	if (argc == 0)
	{
		time_t    t = time (NULL);
	 	struct tm u;

		localtime_r (&t, &u);
		y = u.tm_year + 1900;
		m = u.tm_mon;
	}
	else
	{
		QU_ARG_vint (0, y);
		QU_ARG_CHECK (y >= 0);
		if (argc > 1)
		{
			QU_ARG_vint (1, m);
			QU_ARG_CHECK (m && m <= 11);
		}
		else
		{
			time_t    t = time (NULL);
		 	struct tm u;

			localtime_r (&t, &u);
			m = u.tm_mon;
		}
	}
	d = daysbeforemonth (y, m);
	return qu_int_newi (d);
}



/*) sub timezone -> Int
	Return the timezone value.

	This is the difference between GMT and the local standard time in seconds.

	See C{man 3 tzset} for details.

	@noexc
*/
static qu_int Date_timezone ()
{
	return qu_int_new (timezone);
}



/*) sub tzname (which = 0) -> String
	Return the timezone name.

	v{which} is 0 or 1. If it is not 1, then it is assumed to be 0.

	See C{man 3 tzset} for details.

	@noexc

	@xmp
		use Time
		println (Time.tzname)

	@out
		WIB
*/
static qu_string Date_tzname (int argc, qu_any *argv)
{
	int   i = 0;
	char *z;

	if (argc)
	{
		i = qu_integer_toi (argv[0]);
		if (i < 0 || i > 1)
			i = 0;
	}
	z = tzname[i];
	if (z)
		return qu_string_newz (z);
	return QuStringEMPTY;
}



/*) sub daylight -> Int
	Return the daylight saving time.

	See C{man 3 tzset} for details.

	@xmp
		use Date
		println (Date.daylight)

	@out
		0
*/
static qu_int Date_daylight ()
{
	return qu_int_new (daylight);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Date_init (qu_module m)
{
	time_t t = time (NULL);
	mktime (localtime (&t));

	{
		int y;

		DaysBeforeYear [0] = 0;
		for (y = 1; y <= YEARMAX; ++y)
			DaysBeforeYear [y] = DaysBeforeYear [y - 1] + (isleap (y - 1) ? 366 : 365);
	}


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		make
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "now"  , Date_now  , -1);
	qu_module_add_function (m, "date" , Date_date ,  0);
	qu_module_add_function (m, "value", Date_value, -1);
	qu_module_add_function (m, "make" , Date_make , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		field value
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "year"   , Date_year   , -1);
	qu_module_add_function (m, "month"  , Date_month  , -1);
	qu_module_add_function (m, "day"    , Date_day    , -1);
	qu_module_add_function (m, "hour"   , Date_hour   , -1);
	qu_module_add_function (m, "minute" , Date_minute , -1);
	qu_module_add_function (m, "second" , Date_second , -1);
	qu_module_add_function (m, "weekday", Date_weekday, -1);
	qu_module_add_function (m, "yday"   , Date_yday   , -1);
	qu_module_add_function (m, "toa"    , Date_toa    , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		info
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "isleap"  , Date_isleap  , -1);
	qu_module_add_function (m, "mdays"   , Date_mdays   , -1);
	qu_module_add_function (m, "ydays"   , Date_ydays   , -1);
	qu_module_add_function (m, "mbdays"  , Date_mbdays  , -1);
	qu_module_add_function (m, "ybdays"  , Date_ybdays  , -1);
	qu_module_add_function (m, "timezone", Date_timezone,  0);
	qu_module_add_function (m, "tzname"  , Date_tzname  , -1);
	qu_module_add_function (m, "daylight", Date_daylight,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		format
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "format", Date_format, -1);
	qu_module_add_function (m, "web"   , Date_web   , -1);
	qu_module_add_function (m, "smtp"  , Date_smtp  , -1);
	qu_module_add_function (m, "ymd"   , Date_ymd   , -1);
	qu_module_add_function (m, "ymdhns", Date_ymdhns, -1);

	return 1;
}

#endif /*AVAILABLE*/
