/* Db.c - Copyright (c) Marc Krisnanto */

/* TODO

	Scheme.close --> on exit

	Scheme.gethandle returns an instance of Handle not the subclass of Handle.
	This may require reworking everything ...

*/

/*
	!!! THIS NOTICE SHOULD BE INCLUDED IN ALL DBD MODULES

	Thread synchronization is managed by the Db module. Functions that
	are calling the driver functions may be wrapped in __qu_blocking_begin/end.
	This means that driver functions may not allocate memory via qu_mem_alloc
	et.al., and may not create nor access objects.

	The only exception is 'connect'. When called, this function is not wrapped
	in __qu_blocking_begin/end. If the driver needs to allocate memory, create or
	access objects, the only place to do it is in the driver 'connect' function.

	The driver function 'free' is called by the GC. It may call qu_mem_free
	to free memory allocated with qu_mem_alloc et.al.

	Driver functions may, of course, use C malloc/free.
*/

#include "Db.h"


/*) module Db
	This module provides access to relational databases.

	Here are the steps:
	@ul
	.Create a ${Scheme} object using the ${Scheme.new} method.
	.Get a database ${Handle} with ${Scheme.gethandle} or ${Handle.new}.
	.Do something with the handle, e.g. ${Handle.lookup}.
	.Release the handle with ${Handle.release}.

	@xmp
		use Db

		FooScheme = Db.Scheme ("Foo", "Mysql", 10,
			[
				database : 'Bar'
				host     : 'localhost'
				user     : 'foo'
				password : 'bar'
				port     : 3306
				socket   : '/var/lib/mysql/mysql.sock'
				flags    : 0
				owner    : 'Me!'  # not required by the driver
				foo      : 'bar'  # not required by the driver
			])

		h = Handle (FooScheme) # or the slower h = FooScheme.gethandle
		a = h.selectall ("SELECT * FROM Something")
		println (a)
		h.release
*/

/*) enum TYPE
		TYPE_UNKNOWN    # unknown type gets a String
		TYPE_NULL       # nil
		TYPE_INTEGER    # Integer
		TYPE_FLOAT      # Float
		TYPE_STR        # String
		TYPE_TEXT       # String (TEXT)
		TYPE_BLOB       # String (BLOB)
	;;
	Possible database field value types.

	Use this with ${Handle.fieldtype}.

	@warn
	The enumerated values are set by the system. You must not assume that
	f{TYPE_INTEGER}, for example, has the value of 0.
*/

/*) enum CAPABILITY
		CAPABILITY_THREADSAFE     # is thread-safe
		CAPABILITY_TRANSACTION    # supports transaction
		CAPABILITY_AUTORECONNECT  # has auto-reconnect feature
	;;
	Driver capability.

	Use this with ${Driver.capability}.

	@warn
	The enumerated values are set by the system. You must not assume that
	f{CAPABILITY_THREADSAFE}, for example, has the value of 2.
*/



/* The drivers directory with a trail "/" , e.g. "/opt/qu/lib/Dbd/"  */
static qu_string DriverDir;

/* forward */
typedef struct TDriver TDriver;
typedef TDriver *Driver;

typedef struct TScheme TScheme;
typedef TScheme *Scheme;
struct TScheme
{
	QU_OBJECT_HEAD;
	qu_string  name;     /* scheme name */
	Driver     driver;   /* database driver */
	qu_array   handles;  /* handle pool */
	int        count;    /* current number of handles created */
	int        max;      /* max number of handles */
	qu_dict    data;     /* connection data */
};
static qu_class SchemeClass;

#define SchemeIs(o)  (qu_classof (o) == SchemeClass)



/*============================================================================

	Driver

============================================================================*/

/*) class Driver __final
	A Driver object represents a database driver.

	You can not create a Driver object directly. You use the ${Scheme.new}
	method to create a ${Scheme} object which will load the Driver you specify.

	@xmp
		use Db

		FooScheme = Db.Scheme ('Foo', 'Mysql', 10,
			[
				database : 'Bar'
				host     : 'localhost'
				user     : 'foo'
				password : 'bar'
				port     : 3306
				socket   : '/var/lib/mysql/mysql.sock'
				flags    : 0
				owner    : 'Me!'  # not required by the driver
				foo      : 'bar'  # not required by the driver
			])

		d = FooScheme.driver
		println (d)
		ptinltn (d.name)

	@out
		<Driver Mysql>
		Mysql

	@note
	See the C{lib/dbd} directory for the available drivers. A driver name is
	the name of the file without the extension, e.g. C{'Mysql'}.

	The C{"Generic"} driver might be useful for testing.

	You can use the same Driver for different Scheme objects. The Driver
	will only be loaded once.
*/
static qu_class DriverClass;

struct TDriver
{
	QU_OBJECT_HEAD;
	qu_string   name;     /* driver name */
	qu_dbd_s   *dbd;      /* database driver */
	void       *dlhandle; /* DL handle */
};

#define DriverIs(o)  (qu_classof (o) == DriverClass)


/* All drivers, key = name, value = Driver object */
static qu_dict Drivers;



/*----------------------------------------------------------------------------
	Driver --- low-level instance methods
----------------------------------------------------------------------------*/

void driver__free (Driver self)
{
	if (self->dlhandle)
		qu_dl_close (self->dlhandle);
}


void driver__mark (Driver self)
{
	qu_gc_mark (self->name);
}



/*----------------------------------------------------------------------------
	Driver --- standard instance methods
----------------------------------------------------------------------------*/

/*) sub __print */
static qu_null driver__print (Driver self)
{
	QU_ENSURE (qu_sys_printz ("<Driver ", 8));
	QU_ENSURE (qu_sys_printz (self->name->str, self->name->len));
	QU_ENSURE (qu_sys_printz (">", 1));
	return QuNil;
}



/*) sub __string */
static qu_string driver__string (Driver self)
{
	return qu_string_newf ("<Driver %s>", self->name->str);
}



/*----------------------------------------------------------------------------
	Driver --- specific instance methods
----------------------------------------------------------------------------*/

/*) sub name -> String
	Return the name of the driver.
*/
static qu_string driver_name (Driver self)
{
	return self->name;
}


/*) sub capability (what: Integer) -> Boolean
	Return k{true} if the driver has a specified capability, k{false} otherwise.

	v{what} is one of #{CAPABILITY}.

	@xmp
		d = FooScheme.driver
		println (d.capability (Db.CAPABILITY_TRANSACTION)

	@out
		true  # or perhaps false
*/
static qu_boolean driver_capability (Driver self, qu_integer what)
{
	int cap;

	QU_ARG_int (what, cap);
	return self->dbd->cap (cap) ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	Driver --- interface
----------------------------------------------------------------------------*/

static Driver DriverNew (qu_string name, qu_dbd_s *dbd, void *dlhandle)
{
	Driver d = qu_gc_new (DriverClass);

	d->name     = name;
	d->dbd      = dbd;
	d->dlhandle = dlhandle;
	return d;
}



static Driver DriverLoad (qu_string name)
{
	Driver       d;
	void        *h;
	qu_dbd_s    *dbd;
	qu_dbd_init  ini;
	char         buf [128];
	volatile qu_string  p;

	QU_ARG_CHECKZ (name->len < sizeof (buf) - 16, "driver name too long");

	d = qu_dict_getitem (Drivers, name);
	if (d)
		return d;

	p = qu_string_cat (DriverDir, name, QuModuleDlExt, 0);

  __qu_blocking_begin;
	h = qu_dl_open (p->str, RTLD_NOW | RTLD_GLOBAL);
  __qu_blocking_end;

	if (h == 0)
		QU_THROWF (QuESystem, "%s", qu_dl_error ());

	strcpy (buf, "qu_dbd_");
	strcpy (buf + 7, name->str);

	ini = qu_dl_sym (h, buf);
	if (ini == 0)
	{
		qu_dl_close (h);
		QU_THROWF (QuESystem, "%s missing initialization '%s'", p->str, buf);
	}

	dbd = ini ();
	if (dbd == 0)
	{
		qu_dl_close (h);
		QU_RETHROW;
	}

	d = DriverNew (name, dbd, h);
	qu_dict_setitem (Drivers, name, d);
	return d;
}



/*----------------------------------------------------------------------------
	Driver --- init
----------------------------------------------------------------------------*/

static void Driver_init (qu_module m)
{
	DriverClass = qu_class_new_finalz (m, "Driver", TDriver, driver__free, driver__mark);

	/* standard instance methods */
	qu_class_method_inst  (DriverClass, QU_NAME__print , driver__print , 0);
	qu_class_method_inst  (DriverClass, QU_NAME__string, driver__string, 0);

	/* specific instance methods */
	qu_class_method_instz (DriverClass, "name"      , driver_name      , 0);
	qu_class_method_instz (DriverClass, "capability", driver_capability, 1);

	/* all loaded drivers */
	Drivers = qu_dict_new_root (8);
}



/*============================================================================

	Handle

============================================================================*/

/*) class Handle
	A Handle object represents a database handle (a connection to the
	database server).

	You get a Handle from a handle pool maintained by a ${Scheme} object
	with the #{new} method or the less efficient ${Scheme.gethandle} method.

	You must never forget to #{release} a Handle once you are done with it,
	otherwise you will run out of handles since a ${Scheme} has a maximum number
	of handles it may create.
*/
static qu_class HandleClass;

typedef struct THandle THandle;
typedef THandle *Handle;

struct THandle
{
	QU_OBJECT_HEAD;
	Scheme     scheme;  /* scheme where it came from */
	qu_dbd_s  *dbd;     /* to save deferencing pointers, self->scheme->driver->dbd */
	void      *dbh;     /* handle returned by the driver */
};


/* flags */
#define HandleFlagACTIVE  QU_FLAG_0  /* not released */
#define HandleFlagFATAL   QU_FLAG_1  /* has fatal error */

/* flags accessors */
#define HandleIsActive(self)    qu_flag_test  (self, HandleFlagACTIVE)
#define HandleSetActive(self)   qu_flag_set   (self, HandleFlagACTIVE)
#define HandleUnsetActive(self) qu_flag_unset (self, HandleFlagACTIVE)

#define HandleIsFatal(self)     qu_flag_test  (self, HandleFlagFATAL)
#define HandleSetFatal(self)    qu_flag_set   (self, HandleFlagFATAL)
#define HandleUnseFatal(self)   qu_flag_unset (self, HandleFlagFATAL)

/* make sure the handle has not been released */
#define HANDLE_ENSURE_ACTIVE(self) \
	QU_REQUIREZ (QuECall, HandleIsActive (self), "handle already released")

/* make sure the handle has not been released and does not have a fatal error */
#define HANDLE_ENSURE_USABLE(self) \
QU_BEGIN \
	HANDLE_ENSURE_ACTIVE(self); \
	QU_REQUIREZ (QuECall, !HandleIsFatal (self), "handle has a fatal error"); \
QU_END


/*----------------------------------------------------------------------------
	Handle --- stuffs
----------------------------------------------------------------------------*/

static qu_array _handle_build_row (qu_dbd_s *d, void *h, int nfield, const char **r)
{
	qu_array a;
	int      i;

	if (r == 0)
		return qu_array_new (0);

	if (nfield > QU_INT_MAX)
		QU_ARG_THROWZ ("too many fields");

	a = qu_array_new (nfield);
	for (i = 0; i < nfield; i++)
	{
		if (r[i] == 0)
		{
			/*
				This shouldn't be necessary but just in case some drivers
				have a funny way of doing it ...
			*/
			qu_array_push_quick (a, QuNil);
		}
		else
		{
			switch (d->fieldtype (h, i))
			{
				case QU_DB_TYPE_NULL:
					qu_array_push_quick (a, QuNil);
					break;

				case QU_DB_TYPE_INT:
					qu_array_push_quick (a, qu_integer_newi (atoi (r[i])));
					break;

				case QU_DB_TYPE_LONG:
					qu_array_push_quick (a, qu_integer_newl (atol (r[i])));
					break;

#ifdef HAVE_LONG_LONG
				case QU_DB_TYPE_LONGLONG:
					qu_array_push_quick (a, qu_integer_newll (atoll (r[i])));
					break;
#endif

				case QU_DB_TYPE_FLOAT:
					qu_array_push_quick (a, qu_float_new (atof (r[i])));
					break;

				default:
				{
					long l = strlen (r[i]);
					QU_MEM_CHECK (l < QU_MEM_MAX);
					qu_array_push_quick (a, qu_string_new (r[i], l));
				}
			}
		}
	}
	return a;
}



/*----------------------------------------------------------------------------
	Handle --- standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (s: Scheme) -> Handle  __final
	Get a new database handle from the ${Scheme}.

	See #{new} for details.

	This is used when you are subclassing the Handle class.
*/
static Handle scheme_gethandle (Scheme self);
static qu_null handle_release  (Handle self);

static Handle Handle__inst (qu_class self, Scheme s)
{
	Handle h;

	QU_ARG_CHECK (SchemeIs (s));
	h = scheme_gethandle (s);
	QU_ENSURE (h);
	h->klass = self;

	if (qu_class_init_new (self, h, 1, (qu_any *) &s) == 0)
	{
		handle_release (h);
		QU_RETHROW;
	}
	return h;
}



/*) sub class new (s: Scheme) -> Handle
	Get a new database handle from the ${Scheme}.

	If no handles are available, the Scheme will create a new Handle
	provided that you have not used the maximum number of handles the Scheme
	is allowed to create. Otherwise the calling thread will be blocked until
	a handle is available.

	See ${Scheme.new}.

	@warn
	Never forget to call #{release} once you are done with the handle.

	A database handle should never be used by more than one thread unless
	you know exactly what you are doing. Otherwise weird things, including
	crashes, might happen.

	@exc
	.EArgValue  v{s} not a Scheme object
	.ESystem could not make a new connection

	@xmp
		use Db

		FooScheme = Db.Scheme ("Foo", "Mysql", 10,
			[
				database : 'Bar'
				host     : 'localhost'
				user     : 'foo'
				password : 'bar'
				port     : 3306
				socket   : '/var/lib/mysql/mysql.sock'
				flags    : 0
				owner    : 'Me!'  # not required by the driver
				foo      : 'bar'  # not required by the driver
			])

		h = Handle (FooScheme) # or the slower h = FooScheme.get
		a = h.selectall ("SELECT * FROM Something")
		println (a)
		h.release

	@note
	You can also use ${Scheme.gethandle} but this method is more efficient.
*/
static Handle Handle_new (qu_class self, qu_any s)
{
	return Handle__inst (self, s);
}


/*----------------------------------------------------------------------------
	Handle --- low-level instance methods
----------------------------------------------------------------------------*/

static void handle__free (Handle self)
{
	if (HandleIsActive (self))
		self->dbd->finish (self->dbh);
	self->dbd->disconnect (self->dbh);
	self->dbd->free (self->dbh);
}



static void handle__mark (Handle self)
{
	qu_gc_mark (self->scheme);
}



/*----------------------------------------------------------------------------
	Handle --- standard instance methods
----------------------------------------------------------------------------*/

/*) sub __print */
static qu_null handle__print (Handle self)
{
	QU_ENSURE (qu_sys_printz ("<Handle ", 8));
	QU_ENSURE (qu_sys_printz (self->scheme->name->str, self->scheme->name->len));
	QU_ENSURE (qu_sys_printz (">", 1));
	return QuNil;
}



/*) sub __string */
static qu_string handle__string (Handle self)
{
	return qu_string_newf ("<Handle %s>", self->scheme->name->str);
}



/*----------------------------------------------------------------------------
	Handle --- specific instance methods
----------------------------------------------------------------------------*/
static qu_string handle_sql (Handle self, int argc, qu_any *argv);


/*) sub action (sql: String, *params) -> Integer
	Perform an action query such as I{DELETE}, I{INSERT}, etc.
	On success the number of affected rows is returned.

	See #{sql} for the arguments.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		n = h.action ('DELETE from Foo WHERE id < 10')
*/
static qu_integer handle_action (Handle self, int argc, qu_any *argv)
{
	qu_dbd_s *d;
	void     *h;
	long      n;
	volatile qu_string sql;

	HANDLE_ENSURE_USABLE (self);
	sql = handle_sql (self, argc, argv);
	QU_ENSURE (sql);
	d = self->dbd;
	h = self->dbh;
  __qu_blocking_begin;
	n = d->action (h, sql->str);
  __qu_blocking_end;
	QU_REQUIRE (QuEDatabase, n >= 0);
	return qu_integer_newl (n);
}



/*) sub begin -> Null
	Start transaction.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		try
			h.begin
		catch
			println (__err__)
			h.release
		;;
*/
static qu_null handle_begin (Handle self)
{
	int ret;

	HANDLE_ENSURE_USABLE (self);
  __qu_blocking_begin;
	ret = self->dbd->begin (self->dbh);
  __qu_blocking_end;
	QU_REQUIRE (QuEDatabase, ret == 0);
	return QuNil;
}



/*) sub commit -> Null
	Commit the current transaction.

	@warn
	If this method fails then the Handle is marked as has a fatal error
	and can not be used further. The only thing you can do is to call the
	#{release} method which will destroy the handle.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase
*/
static qu_null handle_commit (Handle self)
{
	int ret;

	HANDLE_ENSURE_USABLE (self);
  __qu_blocking_begin;
	ret = self->dbd->commit (self->dbh);
  __qu_blocking_end;
	if (ret)
	{
		HandleSetFatal (self);
		QU_THROW (QuEDatabase);
	}
	return QuNil;
}



/*) sub errnum -> Integer
	Get the last error code number. Useful only for logging purposes since
	there is no standard code, i.e. it varies among database drivers.

	@exc
	.ECall      handle already released
	.EDatabase

	@xmp
		try
			h.select ('SELEC * FROM bar')
		catch
			println (h.errnum)
		;;

	@out
		1064
*/
static qu_integer handle_errnum (Handle self)
{
	int e;

	HANDLE_ENSURE_ACTIVE (self);
  __qu_blocking_begin;
	e = self->dbd->errnum (self->dbh);
  __qu_blocking_end;
	return qu_integer_newi (e);
}



/*) sub errstr -> String
	Get the last error message. Useful only for logging purposes since
	the string returned can be anything.

	@exc
	.ECall      handle already released
	.EDatabase

	@xmp
		try
			h.select ('SELEC * FROM bar')
		catch
			println (h.errstr)
		;;

	@out
		You have an error in your SQL syntax near 'SELEC * FROM bar' at line 1
*/
static qu_string handle_errstr (Handle self)
{
	const char *s;

	HANDLE_ENSURE_ACTIVE (self);
  __qu_blocking_begin;
	s = self->dbd->errstr (self->dbh);
  __qu_blocking_end;
	if (s)
		return qu_string_newz (s);
	return QuStringEMPTY;
}



/*) sub escape (s: String) -> String
	Escape the string v{s} suitable for the I{SQL} language.

	@exc
	.EArgValue
	.EMemory    string too big
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		s = 'foo says "hello"'
		println (h.escape (s))

	@out
		foo says \"hello\"
*/
static qu_string handle_escape (Handle self, qu_string s)
{
	qu_string e;
	int       l, r;

	HANDLE_ENSURE_USABLE (self);
	QU_ARG_TYPE (s, qu_string);
	if (s->len < 1)
		return QuStringEMPTY;
	QU_MEM_CHECK (s->len < QU_INT_MAX / 2);
	l = s->len * 2 + 1;
	for (;;)
	{
		QU_MEM_CHECK (0 <= l && l < QU_INT_MAX);
		e = qu_string_newl (l);
		r = self->dbd->escape (self->dbh, e->str, s->str, s->len);
		if (r >= 0 && r <= l)
		{
			if (r == 0)
				e = QuStringEMPTY;
			else if (r == 1)
				e = qu_string_byte (e->str[0]);
			else
			{
				e->len = r;
				e->str[r] = '\0';
			}
			break;
		}
		l += 4;
	}
	return e;
}



/*) sub fields -> Array
	Return fields data of a I{SELECT} query. On success the returned value
	is an Array containing Arrays of two elements: the field name and
	the field data type. See the #{fieldtype} method or #{TYPE} for the
	possible data type.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		h.select ("SELECT id, name, age FROM foos")
		println (h.fields)

	@out
		[['id', 2], ['name', 4], ['age', 2]]
*/
static qu_array handle_fields (Handle self)
{
	qu_array   a, b;
	int        i, n, t;
	char      *x;
	qu_string  s;

	HANDLE_ENSURE_USABLE (self);
	n = self->dbd->numfields (self->dbh);
	QU_REQUIRE (QuEDatabase, n > 0);
	a = qu_array_new (n);
	for (i = 0; i < n; i++)
	{
		x = (char *) self->dbd->fieldname (self->dbh, i);
		QU_REQUIRE (QuEDatabase, x);
		s = qu_string_newz (x);
		t = self->dbd->fieldtype (self->dbh, i);
		b = qu_array_new_ex (2, s, qu_int_new (t));
		qu_array_push (a, b);
	}
	return a;
}



/*) sub fieldname (i: Integer) -> String
	Return the name of the v{i}-th field.

	@exc
	.EArgValue
	.EIndex
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		h.select ('SELECT id, name FROM baz')
		println (h.fieldmame (1))

	@out
		name
*/
static qu_string handle_fieldname (Handle self, qu_integer idx)
{
	long        n, i;
	const char *s;

	HANDLE_ENSURE_USABLE (self);
	QU_ARG_long (idx, i);
	n = self->dbd->numfields (self->dbh);
	QU_REQUIRE (QuEDatabase, n > 0);
	QU_REQUIRE (QuEIndex, 0 <= i && i < n);
	s = self->dbd->fieldname (self->dbh, (int) i);
	QU_REQUIRE (QuEDatabase, s);
	return qu_string_newz (s);
}



/*) sub fieldtype (i: Integer) -> Integer
	Return the data type of the v{i}-th field.

	On success the returned value is one of #{TYPE}.

	@exc
	.EArgValue
	.EIndex
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		h.select ('SELECT id, name FROM baz')
		println (h.fieldtype (0))

	@out
		2
*/
static qu_int handle_fieldtype (Handle self, qu_integer idx)
{
	long n, i;
	int  t;

	HANDLE_ENSURE_USABLE (self);
	QU_ARG_long (idx, i);
	n = self->dbd->numfields (self->dbh);
	QU_REQUIRE (QuEDatabase, n > 0);
	QU_REQUIRE (QuEIndex, 0 <= i && i <= n);
	t = self->dbd->fieldtype (self->dbh, (int) i);
	return qu_int_new (t);
}



/*) sub fetchrow -> Array
	Get the next row. You use this with the #{select} method.
	The returned value is an Array. If there are no more rows then
	an empty Array is returned.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		h.select ('SELECT id, name FROM foo')
		while a = h.fetchrow
			println (a)
		;;
*/
static qu_array handle_fetchrow (Handle self)
{
	qu_dbd_s    *d;
	void        *h;
	const char **r;
	int          ret;

	HANDLE_ENSURE_USABLE (self);
	d = self->dbd;
	h = self->dbh;
  __qu_blocking_begin;
	ret = d->fetchrow (h, &r);
  __qu_blocking_end;
	QU_REQUIRE (QuEDatabase, ret == 0);
	return _handle_build_row (d, h, d->numfields (h), r);
}



/*) sub finish -> Null
	This method is generally not required but it may help free up some
	memory. You generally call this method after you have finished using
	a query result. If you perform another query with the same handle,
	this method is automatically called.

	@exc
	.ECall  handle already released or has a fatal error

	@xmp
		a = h.selecteach ('SELECT id, name FROM foo') as Array
		h.finish
*/
static qu_null handle_finish (Handle self)
{
	HANDLE_ENSURE_USABLE (self);
  __qu_blocking_begin;
	self->dbd->finish (self->dbh);
  __qu_blocking_end;
	return QuNil;
}



/*) sub lastid -> Integer
	Return the last record ID created. You would generally use this function
	after an INSERT query.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		h.action ('INSERT INTO foo VALUES (NULL, 1, 2)')
		id = h.lastid
*/
static qu_integer handle_lastid (Handle self)
{
	long i;

	HANDLE_ENSURE_USABLE (self);
  __qu_blocking_begin;
	i = self->dbd->lastid (self->dbh);
  __qu_blocking_end;
	QU_REQUIRE (QuEDatabase, i >= 0);
	return qu_integer_newl (i);
}



/*) sub lookup (sql: String, *params) -> Array
	Perform a I{SELECT} query and return only the first row. The result
	is an Array of column values.

	See #{sql} for the arguments.

	@warn
	You still need to append C{LIMIT 1} in the v{sql} statement yourself
	if required.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		a = h.lookup ('SELECT foo, bar FROM baz LIMIT 1')
		println (a)
*/
static qu_array handle_lookup (Handle self, int argc, qu_any *argv)
{
	qu_dbd_s    *d;
	void        *h;
	const char **r;
	int          ret;
	volatile qu_string sql;

	HANDLE_ENSURE_USABLE (self);
	sql = handle_sql (self, argc, argv);
	QU_ENSURE (sql);
	d = self->dbd;
	h = self->dbh;
  __qu_blocking_begin;
	ret = d->select (h, sql->str);
	if (ret)
	{
	  __qu_blocking_end;
		QU_THROW (QuEDatabase);
	}
	ret = d->fetchrow (h, &r);
  __qu_blocking_end;
	QU_REQUIRE (QuEDatabase, ret == 0);
	return _handle_build_row (d, h, d->numfields (h), r);
}



/*) sub numfields -> Integer
	Return the number of fields returned by a I{SELECT} query.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		h.select ('SELECT foo, bar FROM baz')
		println (h.numfields)

	@out
		2
*/
static qu_integer handle_numfields (Handle self)
{
	long n;

	HANDLE_ENSURE_USABLE (self);
	n = self->dbd->numfields (self->dbh);
	QU_REQUIRE (QuEDatabase, n >= 0);
	return qu_integer_newl (n);
}



/*) sub release -> Null
	Release the database Handle you get from the #{new} method.

	If the Handle has a fatal error then it will be destroyed. Otherwise
	this will put it back in the Scheme handle pool.

	@warn
	Once released, you must never use it again.

	@exc
	.ECall handle already released

	@xmp
		a = h.selectall ('SELECT foo, bar FROM baz')
		h.release
*/
static void _scheme_putback (Scheme self, Handle h);
static qu_null handle_release (Handle self)
{
	HANDLE_ENSURE_ACTIVE (self);
  __qu_blocking_begin;
	self->dbd->finish (self->dbh);
  __qu_blocking_end;
	HandleUnsetActive (self);
	_scheme_putback (self->scheme, self);
	return QuNil;
}



/*) sub rollback -> Null
	Cancel the current transaction.

	@warn
	If this method fails then the Handle is marked as has a fatal error
	and can not be used further. The only thing you can do is to call the
	#{release} method which will destroy the handle.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		try
			h.rollback
		catch
			println ('something is seriously wrong')
			h.release
			throw
		;;
*/
static qu_null handle_rollback (Handle self)
{
	int ret;

	HANDLE_ENSURE_USABLE (self);
  __qu_blocking_begin;
	ret = self->dbd->rollback (self->dbh);
  __qu_blocking_end;
	if (ret)
	{
		HandleSetFatal (self);
		QU_THROW (QuEDatabase);
	}
	return QuNil;
}



/*) sub select (sql: String, *params) -> Null
	Perform a I{SELECT} query. You must then use the #{fetchrow} method
	to get the result one row at a time.

	See #{sql} for the arguments.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		a = h.select ('SELECT foo, bar FROM baz')
		while r = h.fetchrow
			println (r)
		;;
*/
static qu_null handle_select (Handle self, int argc, qu_any *argv)
{
	qu_dbd_s  *d;
	void      *h;
	int        ret;
	volatile qu_string sql;

	HANDLE_ENSURE_USABLE (self);
	sql = handle_sql (self, argc, argv);
	QU_ENSURE (sql);
	d = self->dbd;
	h = self->dbh;
  __qu_blocking_begin;
	ret = d->select (h, sql->str);
  __qu_blocking_end;
	QU_REQUIRE (QuEDatabase, ret == 0);
	return QuNil;
}



/*) sub selectall (sql: String, *params) -> Array
	Perform a I{SELECT} query and return everything in an two dimensional Array.

	See #{sql} for the arguments.

	@warn
	You should normally use C{LIMIT} or you may get a huge result and crash
	on out of memory.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		f = ['foo', 'bar']
		t = 'baz'
		n = 10
		h.selectall ('SELECT 'f.cat' FROM 't' LIMIT 'n).every(@println)
*/
static qu_array handle_selectall (Handle self, int argc, qu_any *argv)
{
	qu_dbd_s    *d;
	void        *h;
	const char **r;
	int          n;
	int          ret;
	volatile qu_string sql;
	volatile qu_array  a, b;

	HANDLE_ENSURE_USABLE (self);
	sql = handle_sql (self, argc, argv);
	QU_ENSURE (sql);
	d = self->dbd;
	h = self->dbh;
  __qu_blocking_begin;
	ret = d->select (h, sql->str);
	if (ret)
	{
	  __qu_blocking_end;
		QU_THROW (QuEDatabase);
	}
  __qu_blocking_end;
	n = d->numfields (h);
	a = qu_array_new (8);
	for (;;)
	{
	  __qu_blocking_begin;
		ret = d->fetchrow (h, &r);
	  __qu_blocking_end;
		if (ret)
			QU_THROW (QuEDatabase);
		if (r == 0)
			break;
		b = _handle_build_row (d, h, n, r);
		QU_ENSURE (b);
		QU_ENSURE (qu_array_push (a, b));
	}
	return a;
}



/*) sub selecteach (sql: String, *params) -> Iterator
	This is the same as the #{select} method except that it returns an Iterator
	object you can use to iterate through the rows.

	@exc
	.ECall      handle already released or has a fatal error
	.EDatabase

	@xmp
		for each a in self.selecteach ('SELECT * FROM foo')
			println (a)
		;;

		a = self.selecteach ('SELECT * FROM foo') as Array
*/
static qu_array _handle_selecteach_next (Handle self, qu_iterator i)
{
	qu_array a = handle_fetchrow (self);
	QU_ENSURE (a);
	QU_REQUIRE (QuEStop, a->len);
	return a;
}


static qu_iterator handle_selecteach (Handle self, int argc, qu_any *argv)
{
	register qu_iterator i;

	QU_ENSURE (handle_select (self, argc, argv));
	QU_ITERATOR_NEW_EXTERN (i, self, _handle_selecteach_next);
	return i;
}



/*) sub sql (sql: String, *params) -> String
	Methods that take an SQL statement such as #{action}, #{select}, etc. take
	the this format.

	v{sql} is the SQL statement and v{params} are the parameters to substitute
	C{%} items in the SQL statement.

	The possible items are:
	@ul
	.C{%q} string escaped, argument is converted into a String
	.C{%o} any object will call its C{__string} method

	@p
	In fact, anything other then C{q} is treated as C{o} so feel free to use any
	character which makes sense to you such as those provided by the corresponding
	object class.

	Note that C{%} inside literal strings are ignored.

	@warn
	Literal strings must only use the single quote character (ANSI).

	@xmp
		dbh.select ("SELECT id FROM %s WHERE id < %n AND stuff=%q", "Foo", 10, 'photo')
		dbh.select ("SELECT id FROM Bar WHERE name LIKE '%foo%')
*/
static qu_string handle_sql (Handle self, int argc, qu_any *argv)
{
 	qu_string  sql, res, x;
	char      *i;

	QU_ARG_VMIN  (1);
	QU_ARG_VTYPE (0, sql, qu_stringy);

	if (qu_buffer_is (sql))
		sql = qu_sys_string (sql);
	if (argc == 1)
		return sql;

	res = qu_string_buffer_new (sql->len);
	i = sql->str;
	argc--;
	argv++;

	while (*i)
	{
		if (*i == '\'')
		{
			char *j = i++;
			while (*i && *i != '\'')
			{
				if (*i == '\\')
					++i;
				++i;
			}
			QU_ENSURE (qu_string_buffer_put (res, j, i - j));
		}
		else if (*i == '%')
		{
			QU_REQUIRE (QuEArgCount, argc);
			argc--;
			x = *argv++;
			if (*++i == 'q')
			{
				if (!qu_string_is (x))
					x = qu_sys_string (x);
				x = handle_escape (self, x);
				QU_ENSURE (x);
				QU_ENSURE (qu_string_buffer_putc (res, '\''));
				QU_ENSURE (qu_string_buffer_puts (res, x));
				QU_ENSURE (qu_string_buffer_putc (res, '\''));
			}
			else
			{
				if (! (qu_string_is (x) || qu_buffer_is (x)))
					x = qu_sys_string (x);
				x = qu_sys_string (x);
				QU_ENSURE (qu_string_buffer_puts (res, x));
			}
			++i;
		}
		else
		{
			qu_string_buffer_putc (res, *i);
			++i;
		}
	}
	return qu_string_buffer_return (res);
}



/*----------------------------------------------------------------------------
	Handle --- interface
----------------------------------------------------------------------------*/

static Handle HandleNew (Scheme s, void *dbh)
{
	Handle h = qu_gc_new (HandleClass);

	h->scheme = s;
	h->dbd    = s->driver->dbd;
	h->dbh    = dbh;
	return h;
}



/*----------------------------------------------------------------------------
	Handle --- init
----------------------------------------------------------------------------*/

static void Handle_init (qu_module m)
{
	HandleClass = qu_class_newz (m, "Handle", THandle, handle__free, handle__mark);

	/* standard class method */
	qu_class_method_class (HandleClass, QU_NAME__inst, Handle__inst, 1);
	qu_class_method_class (HandleClass, QU_NAME_new  , Handle_new  , 1);

	/* standard instance method */
	qu_class_method_inst  (HandleClass, QU_NAME__print , handle__print , 0);
	qu_class_method_inst  (HandleClass, QU_NAME__string, handle__string, 0);

	/* specific instance method */
	qu_class_method_instz (HandleClass, "action"    , handle_action    , -1);
	qu_class_method_instz (HandleClass, "begin"     , handle_begin     ,  0);
	qu_class_method_instz (HandleClass, "commit"    , handle_commit    ,  0);
	qu_class_method_instz (HandleClass, "errnum"    , handle_errnum    ,  0);
	qu_class_method_instz (HandleClass, "errstr"    , handle_errstr    ,  0);
	qu_class_method_instz (HandleClass, "escape"    , handle_escape    ,  1);
	qu_class_method_instz (HandleClass, "fields"    , handle_fields    ,  0);
	qu_class_method_instz (HandleClass, "fieldtype" , handle_fieldtype ,  1);
	qu_class_method_instz (HandleClass, "fieldname" , handle_fieldname ,  1);
	qu_class_method_instz (HandleClass, "fetchrow"  , handle_fetchrow  ,  0);
	qu_class_method_instz (HandleClass, "finish"    , handle_finish    ,  0);
	qu_class_method_instz (HandleClass, "lastid"    , handle_lastid    ,  0);
	qu_class_method_instz (HandleClass, "lookup"    , handle_lookup    , -1);
	qu_class_method_instz (HandleClass, "numfields" , handle_numfields ,  0);
	qu_class_method_instz (HandleClass, "release"   , handle_release   ,  0);
	qu_class_method_instz (HandleClass, "rollback"  , handle_rollback  ,  0);
	qu_class_method_instz (HandleClass, "select"    , handle_select    , -1);
	qu_class_method_instz (HandleClass, "selectall" , handle_selectall , -1);
	qu_class_method_instz (HandleClass, "selecteach", handle_selecteach, -1);
	qu_class_method_instz (HandleClass, "sql"       , handle_sql       , -1);
}



/*============================================================================

	Scheme

============================================================================*/

/*) class Scheme __final
	A Scheme object represents a database scheme.

	You create a Scheme object via the #{new} method. Once created you can
	acquire ${Handle} objects to access your database via ${Handle.new} or
	the #{gethandle} method.

	A Scheme object maintains a pool of reusable ${Handle} objects.
*/
/* static qu_class SchemeClass; */

/* All schemes: key = name, value = Scheme object */
static qu_dict Schemes;



/*----------------------------------------------------------------------------
	Scheme --- standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (name: String, driver: String, maxhandle: Integer, data: Dict) -> Scheme
	Create a new database scheme.

	@args
	.name       A unique name.
	.driver     The name of the driver to load, e.g. C{'Mysql'}.
	.maxhandle  The maximum number of ${Handle} objects that may be created.
	            This must be a positive value within the Int range.
	.data       Data to pass to the ${Driver} when creating a new ${Handle} object.

	@warn
	Make sure you pass the correct v{data} fields otherwise you will not be
	able to make ${Handle} objects later on.

	You must never pass the ${Scheme} object to untrusted source, otherwise
	the source will be able to create ${Handle} objects and access your
	database.

	@exc
	.EKey     a scheme with the same name already exists
	.ESystem  can not load the driver

	@xmp
		use Db

		FooScheme = Db.Scheme ("Foo", "Mysql", 10,
			[
				database : 'my_database_name'
				host     : 'localhost'
				user     : 'foo'
				password : 'bar'
				port     : 3306
				socket   : '/var/lib/mysql/mysql.sock'
				flags    : 0
				owner    : 'Me!'  # not required by the driver
				foo      : 'bar'  # not required by the driver
			])

		println (FooScheme)
		println (FooScheme.name)

	@out
		<Scheme Foo>
		Foo

	@note
	A v{driver} name is the file name in the C{lib/dbd} directory.
	The C{"Generic"} driver is a fake driver that may be useful for testing.
*/
static Scheme Scheme_new (qu_class self, qu_string name, qu_string driver, qu_integer maxhandle, qu_dict data)
{
	int    max;
	Driver d;
	Scheme s;

	__qu_unsafe__;

	QU_REQUIRE   (QuEArgValue, qu_string_is (name) && qu_string_is (driver) && qu_dict_is (data));
	QU_ARG_int   (maxhandle, max);
	QU_ARG_CHECK (max > 0);
	QU_REQUIREZ  (QuEKey, qu_dict_getitem (Schemes, name) == 0, "duplicate database scheme name");

	d = DriverLoad (driver);
	QU_ENSURE (d);

	s = qu_gc_new (SchemeClass);
	s->name    = name;
	s->driver  = d;
	s->handles = qu_array_new (max);
	s->count   = 0;
	s->max     = max;
	s->data    = data;

	qu_dict_setitem (Schemes, name, s);
	return s;
}




/*----------------------------------------------------------------------------
	Scheme --- low-level instance methods
----------------------------------------------------------------------------*/

void scheme__mark (Scheme s)
{
	qu_gc_mark (s->name);
	qu_gc_mark (s->handles);
	qu_gc_mark (s->data);
}



/*----------------------------------------------------------------------------
	Scheme --- standard instance methods
----------------------------------------------------------------------------*/

/*) sub __print */
static qu_null scheme__print (Scheme self)
{
	QU_ENSURE (qu_sys_printz ("<Scheme ", 8));
	QU_ENSURE (qu_sys_printz (self->name->str, self->name->len));
	QU_ENSURE (qu_sys_printz (">", 1));
	return QuNil;
}



/*) sub __string */
static qu_string scheme__string (Scheme self)
{
	return qu_string_newf ("<Scheme %s>", self->name->str);
}



/*----------------------------------------------------------------------------
	Scheme --- specific instance methods
----------------------------------------------------------------------------*/

/*) sub data -> Dict
	Return the data used to create the Scheme object.

	This is the v{data} argument you passed to the #{new} method.
*/
static qu_dict scheme_data (Scheme self)
{
	return self->data;
}



/*) sub driver -> Driver
	Return the associated ${Driver} object.
*/
static Driver scheme_driver (Scheme self)
{
	return self->driver;
}



/*) sub name -> String
	Return the name of this Scheme.
*/
static qu_string scheme_name (Scheme self)
{
 	return self->name;
}



/*) sub maxhandle -> Int
	Return the maximum number of ${Handle} objects that may be created by
	this Scheme.

	This is the v{maxhandle} argument you passed to the #{new} method.
*/
static qu_integer scheme_maxhandle (Scheme self)
{
	return qu_integer_newi (self->max);
}



/*----------------------------------------------------------------------------
	Scheme --- specific instance class methods
----------------------------------------------------------------------------*/

/*) sub gethandle -> Handle
	Get a ${Handle} from the handle pool.

	Note that using the f{Handle.new} method is more efficient.

	See ${Handle.new} for more information.
*/
static Handle scheme_gethandle (Scheme self)
{
	Handle    h;
	qu_dbd_s *dbd = self->driver->dbd;
	qu_array  a = self->handles;

	for (;;)
	{
		if (a->len)
		{
			h = a->objs[--(a->len)];
			break;
		}
		if (self->count >= self->max)
		{
			__qu_blocking_begin;
			__qu_blocking_check_signal;
			/* zzz ... */
			__qu_blocking_end;
		}
		else
		{
			void *d = dbd->connect (self->data);
			QU_ENSURE (d);
			self->count++;
			h = HandleNew (self, d);
			break;
		}
	}
	HandleSetActive (h);
	return h;
}



/*----------------------------------------------------------------------------
	Scheme --- interface
----------------------------------------------------------------------------*/

static void _scheme_putback (Scheme self, Handle h)
{
	if (!HandleIsFatal (h))
		qu_array_push (self->handles, h);
}



/*----------------------------------------------------------------------------
	Scheme --- init
----------------------------------------------------------------------------*/

static void Scheme_init (qu_module m)
{
	SchemeClass = qu_class_new_finalz (m, "Scheme", TScheme, 0, scheme__mark);

	/* standard class methods */
	qu_class_method_class (SchemeClass, QU_NAME_new, Scheme_new, 4);

	/* standard instance methods */
	qu_class_method_inst  (SchemeClass, QU_NAME__print , scheme__print , 0);
	qu_class_method_inst  (SchemeClass, QU_NAME__string, scheme__string, 0);

	/* specific instance methods */
	qu_class_method_instz (SchemeClass, "data"     , scheme_data      , 0);
	qu_class_method_instz (SchemeClass, "driver"   , scheme_driver    , 0);
	qu_class_method_instz (SchemeClass, "gethandle", scheme_gethandle , 0);
	qu_class_method_instz (SchemeClass, "name"     , scheme_name      , 0);
	qu_class_method_instz (SchemeClass, "maxhandle", scheme_maxhandle , 0);

	/* all schemes */
	Schemes = qu_dict_new_root (8);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Db_init (qu_module m)
{
	DriverDir = qu_string_new_staticf ("%s" QU_PATH_SEPSTR "lib/Dbd" QU_PATH_SEPSTR, QU_APP_HOMEDIR);

	qu_var_enum ((qu_symbol) m, "TYPE",
		"TYPE_UNKNOWN" , QU_DB_TYPE_UNKNOWN,
		"TYPE_NULL"    , QU_DB_TYPE_NULL,
		"TYPE_INTEGER" , QU_DB_TYPE_INT,
		"TYPE_FLOAT"   , QU_DB_TYPE_FLOAT,
		"TYPE_STR"     , QU_DB_TYPE_STR,
		"TYPE_TEXT"    , QU_DB_TYPE_TEXT,
		"TYPE_BLOB"    , QU_DB_TYPE_BLOB,
		0);

	qu_var_enum ((qu_symbol) m, "CAPABILITY",
		"CAPABILITY_THREADSAFE"    , QU_DB_CAPABILITY_THREADSAFE,
		"CAPABILITY_TRANSACTION"   , QU_DB_CAPABILITY_TRANSACTION,
		"CAPABILITY_AUTORECONNECT" , QU_DB_CAPABILITY_AUTORECONNECT,
		0);

	Driver_init (m);
	Scheme_init (m);
	Handle_init (m);

	return 1;
}
