/* Db.h - Copyright (c) Marc Krisnanto */

#ifndef _H_qu_db_
#define _H_qu_db_ 1

#include "qu.h"

/*
	This module provides standard interface to relational databases.

	See the dbd/Generic.c and dbd/MySQL.c source files for examples on how
	to write drivers.. They are simple and straightforward, no explanation
	required.
*/

/*
	These are the database function prototypes.
	See qu_dbd_s for what they mean.
*/
typedef int         (*qu_dbd_capability_f) (int what);
typedef void       *(*qu_dbd_connect_f   ) (qu_dict data);
typedef void        (*qu_dbd_disconnect_f) (void *h);
typedef void        (*qu_dbd_free_f      ) (void *h);
typedef int         (*qu_dbd_errnum_f    ) (void *h);
typedef const char *(*qu_dbd_errstr_f    ) (void *h);
typedef void        (*qu_dbd_finish_f    ) (void *h);
typedef int         (*qu_dbd_begin_f     ) (void *h);
typedef int         (*qu_dbd_rollback_f  ) (void *h);
typedef int         (*qu_dbd_commit_f    ) (void *h);
typedef long        (*qu_dbd_action_f    ) (void *h, const char *sql);
typedef long        (*qu_dbd_lastid_f    ) (void *h);
typedef int         (*qu_dbd_select_f    ) (void *h, const char *sql);
typedef int         (*qu_dbd_fetchrow_f  ) (void *h, const char ***row);
typedef int         (*qu_dbd_numfields_f ) (void *h);
typedef int         (*qu_dbd_fieldtype_f ) (void *h, int index);
typedef const char *(*qu_dbd_fieldname_f ) (void *h, int index);
typedef int         (*qu_dbd_escape_f    ) (void *h, char *buf, const char *s, int len);

/*
	This struct represents a database driver.
*/
typedef struct qu_dbd_s qu_dbd_s;
struct qu_dbd_s
{
	qu_dbd_capability_f  cap;        /* Get driver capability */
	qu_dbd_connect_f     connect;    /* Make a new connection */
	qu_dbd_disconnect_f  disconnect; /* Close an existing connection */
	qu_dbd_free_f        free;       /* Free an existing connection */
	qu_dbd_errnum_f      errnum;     /* Get current error code */
	qu_dbd_errstr_f      errstr;     /* Get current error message */
	qu_dbd_finish_f      finish;     /* Finish using query result */
	qu_dbd_begin_f       begin;      /* Begin transaction */
	qu_dbd_rollback_f    rollback;   /* Abort transaction */
	qu_dbd_commit_f      commit;     /* Commit transaction */
	qu_dbd_action_f      action;     /* Perform an action query */
	qu_dbd_lastid_f      lastid;     /* Get last record ID number */
	qu_dbd_select_f      select;     /* Perform a select query */
	qu_dbd_fetchrow_f    fetchrow;   /* Fetch one row of a query result */
	qu_dbd_numfields_f   numfields;  /* Get number of fields of a select query result */
	qu_dbd_fieldtype_f   fieldtype;  /* Get field type of a select query result */
	qu_dbd_fieldname_f   fieldname;  /* Get field name of a select query result */
	qu_dbd_escape_f      escape;     /* SQL escape a string */
};


/*
	This is the initialization function.

	It must be named qu_dbd_xxx where xxx is the name of the file
	e.g. C{int qu_dbd_Mysql ()}

	This function must throw an exception on failure (return NULL).
*/
typedef qu_dbd_s *(*qu_dbd_init) ();


/*
	For the field 'what' of qu_dbd_capability_f.
*/
#define QU_DB_CAPABILITY_THREADSAFE    0x000000
#define QU_DB_CAPABILITY_TRANSACTION   0x000001
#define QU_DB_CAPABILITY_AUTORECONNECT 0x000002


/*
	For qu_dbd_fieldtype_f, the supported value types.

	Note that in script level QU_DB_TYPE_INT is named QU_DB_TYPE_INTEGER and covers
	all qu_integer types (QU_DB_TYPE_LONG and QU_DB_TYPE_LONGLONG).
*/
#define QU_DB_TYPE_UNKNOWN   0   /* converted to Str */
#define QU_DB_TYPE_NULL      1   /* converted to nil */
#define QU_DB_TYPE_INT       2   /* converted to qu_integer */
#define QU_DB_TYPE_LONG      3   /* converted to qu_integer */
#define QU_DB_TYPE_LONGLONG  4   /* converted to qu_integer */
#define QU_DB_TYPE_FLOAT     8   /* converted to qu_float */
#define QU_DB_TYPE_STR       9   /* converted to Str */
#define QU_DB_TYPE_TEXT      10  /* converted to Str */
#define QU_DB_TYPE_BLOB      11  /* converted to Str */

/***/
#endif /*_H_qu_db_*/
