/* Mysql.c - Copyright (c) Marc Krisnanto */

/*
	!!! THIS NOTICE SHOULD BE INCLUDED IN ALL DBD MODULES

	Thread synchronization is managed by the Db module. Functions that
	are calling the driver functions may be wrapped in __qu_blocking_begin/end.
	This means that driver functions may not allocate memory via qu_mem_alloc
	et.al., and may not create nor access objects.

	The only exception is 'connect'. When called, this function is not wrapped
	in __qu_blocking_begin/end. If the driver needs to allocate memory, create or
	access objects, the only place to do it is in the driver 'connect' function.

	The driver function 'free' is called by the GC. It may call qu_mem_free
	to free memory allocated with qu_mem_alloc et.al.

	Driver functions may, of course, use C malloc/free.
*/

#include "Mysql.h"
#include "../Db.h"

void *QuMysqlHandle;
int   QuMysqlSizeLongLong = 0;
int   QuMysqlFieldNew = 0;


typedef struct conn_s conn_s;
struct conn_s
{
	MYSQL     *mysql;  /* mysql handle */
	MYSQL_RES *result; /* result of the last query */
	int        closed; /* connection closed */
	int        fatal;  /* has fatal error */
};



static void Mysql_free (conn_s *c)
{
	if (c->result)
		mysql_free_result (c->result);
	if (!c->closed)
		mysql_close (c->mysql);
	/* qu_mem_free (c->mysql); */
	qu_mem_free (c);
}



int Mysql_cap (int what)
{
	switch (what)
	{
		case QU_DB_CAPABILITY_TRANSACTION:
		case QU_DB_CAPABILITY_AUTORECONNECT:
			return 1;

		case QU_DB_CAPABILITY_THREADSAFE:
			return mysql_thread_safe ();
	}
	return 0;
}



static conn_s *Mysql_connect (qu_dict argd)
{
	MYSQL        *m;
	conn_s       *c;
	qu_string     host, database, user, password, socket = QuStringEMPTY;
	unsigned int  port = 3306, flags = 0;

	static char *argx[] =
	{
		"s", "host",
		"s", "database",
		"s", "user",
		"s", "password",
		":",
		"s", "socket",
		"I", "port",
		"I", "flags",
		NULL
	};

	QU_ARG_DPARSE (&host, &database, &user, &password, &socket, &port, &flags);
	m = mysql_init (NULL);
	QU_REQUIRE (QuEMemory, m);

	/*
		CHECKME
		We want UPDATE to always return the number of matching
		rows, not the number of updated rows!
	*/
	flags |= CLIENT_FOUND_ROWS;

	if (mysql_real_connect (m, host->str, user->str,
			password->str, database->str,
			port, socket->str, flags) == NULL)
	{
		const char *e = mysql_error (m);
		qu_string   s = qu_string_newz (e);

		/* qu_mem_free (m); */
		mysql_close (m);
		qu_thread_throw (QuEDatabaseClass, s);
		return 0;
	}

	c = qu_mem_alloc (sizeof (conn_s));
 	c->mysql  = m;
	c->result = NULL;
	c->fatal  = 0;
	return c;
}



static void Mysql_disconnect (conn_s *c)
{
	if (!c->closed)
	{
		if (c->result != 0)
			mysql_free_result (c->result);
		c->result = 0;
		mysql_close (c->mysql);
		c->closed = 1;
	}
}



static int Mysql_errnum (conn_s *c)
{
	return mysql_errno (c->mysql);
}



static const char *Mysql_errstr (conn_s *c)
{
	return mysql_error (c->mysql);
}



static void Mysql_finish (conn_s *c)
{
	if (c->result)
	{
		mysql_free_result (c->result);
		c->result = 0;
	}
}



static long Mysql_action (conn_s *c, const char *sql)
{
	Mysql_finish (c);
	if (mysql_query (c->mysql, sql))
		return -1;

	/*FIXME - this is bad */
	if (QuMysqlSizeLongLong)
		return (long) mysql_affected_rows (c->mysql);
	return mysql_affected_rows_l (c->mysql);
}



static int Mysql_begin (conn_s *c)
{
	return (Mysql_action (c, "BEGIN") < 0) ? -1 : 0;
}



static int Mysql_rollback (conn_s *c)
{
	return (Mysql_action (c, "ROLLBACK") < 0) ? -1 : 0;
}



static int Mysql_commit (conn_s *c)
{
	return (Mysql_action (c, "COMMIT") < 0) ? -1 : 0;
}



static long Mysql_lastid (conn_s *c)
{
	/*FIXME - this is bad */
	if (QuMysqlSizeLongLong)
		return (long) mysql_insert_id (c->mysql);
	return mysql_insert_id_l (c->mysql);
}



static int Mysql_select (conn_s *c, const char *sql)
{
	Mysql_finish (c);
	if (mysql_query (c->mysql, sql))
		return -1;
	c->result = mysql_use_result (c->mysql);
	if (c->result == 0)
		return -1;
	return 0;
}



static int Mysql_fetchrow (conn_s *c, const char ***row)
{
	assert (c->result != 0);
	*row = (const char **) mysql_fetch_row (c->result);
	return *row ? 0 : (Mysql_errnum (c) ? -1 : 0);
}



static int Mysql_numfields (conn_s *c)
{
	return c->result ? mysql_num_fields (c->result) : -1;
}



static int Mysql_fieldtype (conn_s *c, int index)
{
	int n, t;

	n = (int) mysql_num_fields (c->result);
	if (index < 0 || index >= n)
		return -1;

	if (QuMysqlFieldNew)
	{
		MYSQL_FIELD_NEW *f = (MYSQL_FIELD_NEW *) mysql_fetch_fields (c->result);
		t = MYSQL_FIELD_GET_TYPE (f, index);
	}
	else
	{
		MYSQL_FIELD *f = (MYSQL_FIELD *) mysql_fetch_fields (c->result);
		t = MYSQL_FIELD_GET_TYPE (f, index);
	}

	switch (t)
	{
		case MYSQL_TYPE_TINY:
		case MYSQL_TYPE_SHORT:
			return QU_DB_TYPE_INT;

		case MYSQL_TYPE_LONG:
		case MYSQL_TYPE_INT24:
		case MYSQL_TYPE_TIMESTAMP:
		case MYSQL_TYPE_DATE:
		case MYSQL_TYPE_TIME:
		case MYSQL_TYPE_DATETIME:
		case MYSQL_TYPE_YEAR:
		case MYSQL_TYPE_SET:
		case MYSQL_TYPE_ENUM:
		case MYSQL_TYPE_NEWDATE:
			return QU_DB_TYPE_LONG;

		case MYSQL_TYPE_DECIMAL:
		case MYSQL_TYPE_FLOAT:
		case MYSQL_TYPE_DOUBLE:
			return QU_DB_TYPE_FLOAT;

		case MYSQL_TYPE_LONGLONG:
#ifdef HAVE_LONG_LONG
			return QU_DB_TYPE_LONGLONG;
#else
			return QU_DB_TYPE_FLOAT;
#endif

		case MYSQL_TYPE_STRING:
		case MYSQL_TYPE_VAR_STRING:
			return QU_DB_TYPE_STR;

		case MYSQL_TYPE_BLOB:
		case MYSQL_TYPE_TINY_BLOB:
		case MYSQL_TYPE_MEDIUM_BLOB:
		case MYSQL_TYPE_LONG_BLOB:
			return QU_DB_TYPE_BLOB;

		case MYSQL_TYPE_NULL:
			return QU_DB_TYPE_NULL;

		default:
			break;
	}
	return QU_DB_TYPE_UNKNOWN;
}



static const char *Mysql_fieldname (conn_s *c, int index)
{
	if (c->result)
	{
		int n = (int) mysql_num_fields (c->result);

		if (QuMysqlFieldNew)
		{
			MYSQL_FIELD_NEW *f = (MYSQL_FIELD_NEW *) mysql_fetch_fields (c->result);
			if (0 <= index && index < n)
				return MYSQL_FIELD_GET_NAME (f, index);
		}
		else
		{
			MYSQL_FIELD *f = (MYSQL_FIELD *) mysql_fetch_fields (c->result);
			if (0 <= index && index < n)
				return MYSQL_FIELD_GET_NAME (f, index);
		}
	}
	return 0;
}



static int Mysql_escape (conn_s *c, char *buf, const char *s, int len)
{
	return mysql_real_escape_string (c->mysql, buf, s, len);
}



static qu_dbd_s MysqlDbd =
{
	(qu_dbd_capability_f) Mysql_cap,
	(qu_dbd_connect_f   ) Mysql_connect,
	(qu_dbd_disconnect_f) Mysql_disconnect,
	(qu_dbd_free_f      ) Mysql_free,
	(qu_dbd_errnum_f    ) Mysql_errnum,
	(qu_dbd_errstr_f    ) Mysql_errstr,
	(qu_dbd_finish_f    ) Mysql_finish,
	(qu_dbd_begin_f     ) Mysql_begin,
	(qu_dbd_rollback_f  ) Mysql_rollback,
	(qu_dbd_commit_f    ) Mysql_commit,
	(qu_dbd_action_f    ) Mysql_action,
	(qu_dbd_lastid_f    ) Mysql_lastid,
	(qu_dbd_select_f    ) Mysql_select,
	(qu_dbd_fetchrow_f  ) Mysql_fetchrow,
	(qu_dbd_numfields_f ) Mysql_numfields,
	(qu_dbd_fieldtype_f ) Mysql_fieldtype,
	(qu_dbd_fieldname_f ) Mysql_fieldname,
	(qu_dbd_escape_f    ) Mysql_escape,
};


void *QuMysqlHandle;

static void qu_mysql_at_exit ()
{
	if (QuMysqlHandle)
		qu_dl_close (QuMysqlHandle);
}


static void *qu_mysql_missing (const char *name)
{
	qu_dl_close (QuMysqlHandle);
	QU_THROWF (QuESystem, "MySql missing symbol %s", name);
}


qu_dbd_s *qu_dbd_Mysql ()
{
	qu_string s, v;

	QuMysqlHandle = qu_dl_open_ex ("mysql", QU_DL_GLOBAL | QU_DL_NOW);
	s = qu_dl_config ("mysqlsize");
	v = qu_dl_config ("mysqlversion");

	QU_ENSURE (QuMysqlHandle && s && v);
	QuMysqlSizeLongLong = (strcmp (s->str, "LL") == 0);
	QuMysqlFieldNew     = (strcmp (v->str, "4") == 0);

#define f_(name) \
	name = qu_dl_sym (QuMysqlHandle, #name); \
	if (name == 0) \
		return qu_mysql_missing ( #name)

#define g_(name, orig) \
	name = qu_dl_sym (QuMysqlHandle, orig); \
	if (name == 0) \
		return qu_mysql_missing (orig)

	f_(mysql_close);
	f_(mysql_free_result);
	f_(mysql_thread_safe);
	f_(mysql_init);
	f_(mysql_real_connect);
	f_(mysql_errno);
	f_(mysql_error);
	f_(mysql_query);
	g_(mysql_affected_rows_l, "mysql_affected_rows");
	g_(mysql_insert_id_l, "mysql_insert_id");
	f_(mysql_affected_rows);
	f_(mysql_insert_id);
	f_(mysql_use_result);
	f_(mysql_fetch_row);
	f_(mysql_num_fields);
	f_(mysql_fetch_fields);
	f_(mysql_real_escape_string);

	atexit (qu_mysql_at_exit);
	return &MysqlDbd;
}
