/* Sqlite.c - Copyright (c) Bruno Luca */

/*
	Revisions:
	- 16-May-2006 (Marc K.) dynamic loading
	- 03-Apr-2006 (Marc K.) global rename
	- 20-Mar-2006 (Marc K.) initial corrections
*/

/*
	!!! THIS NOTICE SHOULD BE INCLUDED IN ALL DBD MODULES

	Thread synchronization is managed by the Db module. Functions that
	are calling the driver functions may be wrapped in __qu_blocking_begin/end.
	This means that driver functions may not allocate memory via qu_mem_alloc
	et.al., and may not create nor access objects.

	The only exception is 'connect'. When called, this function is not wrapped
	in __qu_blocking_begin/end. If the driver needs to allocate memory, create or
	access objects, the only place to do it is in the driver 'connect' function.

	The driver function 'free' is called by the GC. It may call qu_mem_free
	to free memory allocated with qu_mem_alloc et.al.

	Driver functions may, of course, use C malloc/free.
*/

#include "../Db.h"
#include "Sqlite3.h"


typedef struct conn conn;
struct conn
{
	sqlite3       *db;    /* the database */
	sqlite3_stmt  *stmt;  /* the current SQL */
	char         **cols;  /* result of the current SQL --> malloc'ed */
};



static void Sqlite_free (conn *c)
{
	if (c->stmt)
	{
		sqlite3_finalize (c->stmt);
		c->stmt = 0;
	}

	if (c->cols)
	{
		free (c->cols);
		c->cols = 0;
	}
}



static int Sqlite_capability (int what)
{
	switch (what)
	{
		case QU_DB_CAPABILITY_TRANSACTION:
			return 1;

		case QU_DB_CAPABILITY_THREADSAFE:
			/* there is nothing in sqlite3.h to check this out
				we are assuming that sqlite was compiled with -DTHREADSAFE=1 */
			return 1;
	}
	return 0;
}



static conn *Sqlite_connect (qu_dict argd)
{
	sqlite3    *db;
	conn       *c;
	qu_string   database = QuStringEMPTY;
	int         rc;

	static char *argx[] =
	{
		"s", "database",
		NULL
	};

	QU_ARG_DPARSE (&database);

	rc = sqlite3_open (database->str, &db);
	if (rc)
	{
		const char *e = sqlite3_errmsg (db);
		qu_string   s = qu_string_newz (e);
		qu_thread_throw (QuEDatabaseClass, s);
		return 0;
	}

	c       = qu_mem_alloc (sizeof (conn));
	c->db   = db;
	c->stmt = 0;
	c->cols = 0;
	return c;
}



static void Sqlite_disconnect (conn *c)
{
	Sqlite_free (c);
	sqlite3_close (c->db);
	qu_mem_free (c);
}



static int Sqlite_errnum (conn *c)
{
	return sqlite3_errcode (c->db);
}



static const char *Sqlite_errstr (conn *c)
{
	return sqlite3_errmsg (c->db);
}



static void Sqlite_finish (void *c)
{
	Sqlite_free (c);
}



static long Sqlite_action (conn *c, const char *sql)
{
	int           rc, rows = 0;
	sqlite3_stmt *stmt;

	rc = sqlite3_prepare (c->db, sql, -1, &stmt, 0);
	if (rc != SQLITE_OK || !stmt)
		return -1;
	rc = sqlite3_step (stmt);
	while (rc == SQLITE_ROW)
	{
		rows++;
		rc = sqlite3_step (stmt);
	}
	sqlite3_finalize (stmt);
	return rc == SQLITE_DONE ? rows : -1;
}



static int Sqlite_begin (conn *c)
{
	return (Sqlite_action (c, "BEGIN TRANSACTION") < 0) ? -1 : 0;
}



static int Sqlite_rollback (conn *c)
{
	return (Sqlite_action (c, "ROLLBACK TRANSACTION") < 0) ? -1 : 0;
}


static int Sqlite_commit (conn *c)
{
	return (Sqlite_action (c, "COMMIT TRANSACTION") < 0) ? -1 : 0;
}



static long Sqlite_lastid (conn *c)
{
	return sqlite3_last_insert_rowid (c->db);
}



static int Sqlite_select (conn *c, const char *sql)
{
	int rc;

	Sqlite_free (c);
	rc = sqlite3_prepare (c->db, sql, -1, &(c->stmt), 0);
	return rc == SQLITE_OK || !c->stmt ? 0 : -1;
}



static int Sqlite_fetchrow (conn *c, const char ***row)
{
	assert (c->stmt);
	switch (sqlite3_step (c->stmt))
	{
		case SQLITE_DONE:
			sqlite3_finalize (c->stmt);
			*row = 0;
			return 0;

		case SQLITE_ROW:
		{
			int    i, n;
			char **r;

			if (c->cols)
			{
				free (c->cols);
				c->cols = 0;
			}

			n = sqlite3_column_count (c->stmt);
			r = c->cols = malloc (sizeof (char *) * n);
			for (i = 0; i < n; ++i)
				r [i] = (char *) sqlite3_column_text (c->stmt, i);
			*((char ***) row) =  r;
			return 0;
		}
	}
	return -1;
}



static int Sqlite_numfields (conn *c)
{
	assert (c->stmt);
	return sqlite3_column_count (c->stmt);
}



static int Sqlite_fieldtype (conn *c, int index)
{
	assert (c->stmt);
	switch (sqlite3_column_type (c->stmt, index))
	{
		case SQLITE_INTEGER: return QU_DB_TYPE_INT;
		case SQLITE_FLOAT:   return QU_DB_TYPE_FLOAT;
		case SQLITE_TEXT:    return QU_DB_TYPE_STR;
		case SQLITE_BLOB:    return QU_DB_TYPE_BLOB;
		case SQLITE_NULL:    return QU_DB_TYPE_NULL;
    }
	return -1;
}



static const char *Sqlite_fieldname (conn *c, int index)
{
	assert (c->stmt);
	return sqlite3_column_name (c->stmt, index);
}



static int Sqlite_escape (void *c, char *to, const char *from, int len)
{
	char *p = sqlite3_mprintf ("%q", from);
	int   l = strlen (p);

	if (l > len)
	{
		sqlite3_free (p);
	}
	else
	{
		strcpy (to, p);
		sqlite3_free (p);
	}
	return l;
}



static qu_dbd_s SqliteDbd =
{
	(qu_dbd_capability_f) Sqlite_capability,
	(qu_dbd_connect_f   ) Sqlite_connect,
	(qu_dbd_disconnect_f) Sqlite_disconnect,
	(qu_dbd_free_f      ) Sqlite_free,
	(qu_dbd_errnum_f    ) Sqlite_errnum,
	(qu_dbd_errstr_f    ) Sqlite_errstr,
	(qu_dbd_finish_f    ) Sqlite_finish,
	(qu_dbd_begin_f     ) Sqlite_begin,
	(qu_dbd_rollback_f  ) Sqlite_rollback,
	(qu_dbd_commit_f    ) Sqlite_commit,
	(qu_dbd_action_f    ) Sqlite_action,
	(qu_dbd_lastid_f    ) Sqlite_lastid,
	(qu_dbd_select_f    ) Sqlite_select,
	(qu_dbd_fetchrow_f  ) Sqlite_fetchrow,
	(qu_dbd_numfields_f ) Sqlite_numfields,
	(qu_dbd_fieldtype_f ) Sqlite_fieldtype,
	(qu_dbd_fieldname_f ) Sqlite_fieldname,
	(qu_dbd_escape_f    ) Sqlite_escape,
};



void *QuSqliteHandle;

static void qu_sqlite_at_exit ()
{
	if (QuSqliteHandle)
		qu_dl_close (QuSqliteHandle);
}


static void *qu_sqlite_missing (const char *name)
{
	qu_dl_close (QuSqliteHandle);
	QU_THROWF (QuESystem, "MySql missing symbol %s", name);
}


qu_dbd_s *qu_dbd_Sqlite ()
{
	QuSqliteHandle = qu_dl_open_ex ("sqlite3", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (QuSqliteHandle);

#define f_(name) \
	name = qu_dl_sym (QuSqliteHandle, #name); \
	if (name == 0) \
		return qu_sqlite_missing ( #name)

	f_(sqlite3_finalize         );
	f_(sqlite3_open             );
	f_(sqlite3_close            );
	f_(sqlite3_errcode          );
	f_(sqlite3_errmsg           );
	f_(sqlite3_prepare          );
	f_(sqlite3_step             );
	f_(sqlite3_finalize         );
	f_(sqlite3_last_insert_rowid);
	f_(sqlite3_column_count     );
	f_(sqlite3_column_text      );
	f_(sqlite3_column_count     );
	f_(sqlite3_column_type      );
	f_(sqlite3_column_name      );
	f_(sqlite3_mprintf          );
	f_(sqlite3_free             );

	atexit (qu_sqlite_at_exit);
	return &SqliteDbd;
}
