/* Dump.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) mod Dump
	This module prints out the internal data of an object to v{stderr}.

	@prog
		use Dump

		q = Queue ("F", 100)
		q <<< 1 <<< 2 <<< 3 <<< 4
		Dump (q)

	@out
		Dump 0x40078478
			class  : Queue (size 32)
			flags  : 400
			weakref: 0
			attrs  : (nil)
			size   : 100
			len    : 4
			type   : FIFO
			items
			1
			2
			3
			4

	@p
	If the class specifies an instance method named __dump then it will
	be called.

	@todo
	Code, Dlib, Exception, File, Frame, Iterator, Method, Node, Opaque, Proc,
	Pointer, Range, Record, Stream, Sub, Thread (+ Mutex, Cond), Tokenizer,
	Var, Weak.
*/


/*-----------------------------------------------------------------------------
	helpers
-----------------------------------------------------------------------------*/

#define P(...) fprintf (stderr, __VA_ARGS__)
#define S(x)   (x) ? qu_sys_string (x)->str : "(nil)"

static void Dump_default (qu_any self)
{
	qu_class c = qu_classof (self);

	P ("Dump %p\n", self);
	P ("\tclass  : %s (size %d)\n", c->name->str, c->size);
	if (!qu_int_is (self))
	{
		qu_obj o = (qu_obj) self;
		P ("\tflags: %lx\n", o->flags);
		P ("\tweak : %ld\n", o->weak ? o->weak->len : 0L);
		P ("\tattrs: %p\n" , o->attrs);
	}
}



/*-----------------------------------------------------------------------------
	qu_class
-----------------------------------------------------------------------------*/

static void Dump_class (qu_class self)
{
	P ("\tid   : %ld\n",   self->id     );
	P ("\tmod  : %s\n" , S(self->parent));
	P ("\tsuper: %s\n" , S(self->super) );
	P ("\tsize : %d\n" ,   self->size   );
	P ("\tfree : %p\n" ,   self->dtor   );
	P ("\tmark : %p\n" ,   self->mark   );
	P ("\taccel: %p\n" ,   self->accel  );
	P ("\tmixin: %s\n" , S(self->mixin) );
	P ("\tget  : %p\n" ,   self->get    );
	P ("\tset  : %p\n" ,   self->set    );
	P ("\tinit : %s\n" , S(self->init)  );
};



/*----------------------------------------------------------------------------
	qu_null
----------------------------------------------------------------------------*/

static void Dump_null (qu_null self)
{
	P ("\tvalue  : %s\n", S(self));
}



/*----------------------------------------------------------------------------
	qu_boolean
----------------------------------------------------------------------------*/

static void Dump_boolean (qu_boolean self)
{
	P ("\tvalue  : %s\n", S(self));
}



/*-----------------------------------------------------------------------------
	qu_int
-----------------------------------------------------------------------------*/

static void Dump_int (qu_int self)
{
	P ("value  : %ld\n", qu_int_tol (self));
}



/*----------------------------------------------------------------------------
	qu_long
----------------------------------------------------------------------------*/

static void Dump_long (qu_long self)
{
	long             n = self->len;
	qu_long_digit_t *b = self->digits + n;

	P ("\t(%ld) ", n);
	while (n--)
	{
#if QU_LONG_FORMAT == 2
		P ("%lx ", *--b);
#elif QU_LONG_FORMAT == 1
		P ("%x " , *--b);
#else
		P ("%hx ", *--b);
#endif
	}
	P ("\n");
}



/*----------------------------------------------------------------------------
	qu_rational
----------------------------------------------------------------------------*/

static void Dump_rational (qu_rational self)
{
	P ("\tvalue  : %s _/ %s\n", S(self->num), S(self->den));
}



/*-----------------------------------------------------------------------------
	qu_float
-----------------------------------------------------------------------------*/

static void Dump_float (qu_float self)
{
	P ("\tvalue  : %.16e\n", qu_float_tod (self));
}



/*----------------------------------------------------------------------------
	qu_big
----------------------------------------------------------------------------*/

static void Dump_big (qu_big self)
{
	long            l = self->len;
	qu_big_digit_t *d = self->digits;

	P ("\tvalue  :");
	if (qu_big_isneg (self))
		P ("- 0. ");
	else
		P ("  0. ");
	if (l == 0)
	{
		P ("zero ^ 0");
	}
	else
	{
		while (l--)
			P (QU_BIG_FORMAT " ", *d++);

		P ("^ %ld", self->expo * QU_BIG_FIG);
	}
	P ("\n");
}



/*----------------------------------------------------------------------------
	qu_complex
----------------------------------------------------------------------------*/

static void Dump_complex (qu_complex self)
{
	P ("\tvalue  : %s + %s j\n", S(self->real), S(self->imag));
}



/*----------------------------------------------------------------------------
	qu_inf
----------------------------------------------------------------------------*/

static void Dump_inf (qu_inf self)
{
	P ("\tvalue  : %s\n",
		(self == QuInfPos)
			? "__infp__"
			: ((self == QuInfNeg) ? "__infn__" : "*** unknown ***"));
}



/*----------------------------------------------------------------------------
	qu_string
----------------------------------------------------------------------------*/

static void Dump_string (qu_string self)
{
	long i;

	P ("\tlen    : %ld\n", self->len);
	P ("\tvalue  : ");
	for (i = 0; i < self->len; ++i)
	{
		char c = self->str[i];
		if (c < 32)
			P ("");
		else
			P ("%c", c);
	}
	if (self->str[self->len] != '\0')
		P (" *** no terminating nil ***");
	P ("\n");
}



/*----------------------------------------------------------------------------
	qu_string
----------------------------------------------------------------------------*/

static void Dump_buffer (qu_buffer self)
{
	long i;

	P ("\tlen    : %ld\n", self->len);
	P ("\tcap    : %ld\n", self->cap);
	P ("\tlimit  : %ld\n", self->limit);
	P ("\tvalue  : ");
	for (i = 0; i < self->len; ++i)
	{
		char c = self->str[i];
		if (c < 32)
			P ("");
		else
			P ("%c", c);
	}
	if (self->str[self->len] != '\0')
		P (" *** no terminating nil ***");
	P ("\n");
}



/*-----------------------------------------------------------------------------
	qu_array
-----------------------------------------------------------------------------*/

static void Dump_array (qu_array self)
{
	long i;

	P ("\tcap    : %ld\n",   self->cap  );
	P ("\tlen    : %ld\n",   self->len  );
	P ("\ttype   : %s\n" , S(self->type));
	P ("\titems\n");
	for (i = 0; i < self->len; ++i)
		P ("\t%s\n", S(self->objs[i]));
}



/*-----------------------------------------------------------------------------
	qu_queue
-----------------------------------------------------------------------------*/

static void Dump_queue (qu_queue self)
{
	qu_queueitem_s *i;

	P ("\tsize   : %ld\n", self->size);
	P ("\tlen    : %ld\n", self->len);
	P ("\ttype   : %s\n" , qu_queue_is_fifo (self) ? "FIFO" : "LIFO");
	P ("\titems\n");

	for (i = self->first; i; i = i->next)
		P ("\t%s\n", S(i->obj));
}



/*-----------------------------------------------------------------------------
	qu_dict
-----------------------------------------------------------------------------*/

static void Dump_dict (qu_dict self)
{
	qu_dictentry_s **b = self->buckets, *i;
	long m = self->mask;
	long j, k;

	P ("\tmask   : %ld\n" , self->mask);
	P ("\titems  : %ld\n", self->len);
	P ("\tbuckets\n");

	for (k = 0; k < m; ++k)
	{
		i = b[k];
		if (i == 0)
		{
			P ("\t%4ld", k);
			j = k;
			while (k + 1 < m && b[k + 1] == 0)
				++k;
			if (k > j)
				P (" - %4ld", k);
			else
				P ("       ");
			P (": -");
		}
		else
		{
			P ("\t%4ld       :", k);
			for (; i; i = i->right)
				P (" [%s, %s]", S(i->key), S(i->value));
		}
		P ("\n");
	}
}



/*-----------------------------------------------------------------------------
	qu_set
-----------------------------------------------------------------------------*/

struct qu_setentry_s
{
	qu_setentry_s *next;  /* Next entry in bucket */
	qu_any         obj;   /* qu_object */
	long           hash;  /* Cached hash value */
};


static void Dump_set (qu_set self)
{
	qu_setentry_s **b = self->buckets, *i;
	long m = self->mask;
	long j, k;

	P ("\tmask   : %ld\n", self->mask);
	P ("\titems  : %ld\n", self->len);
	P ("\tbuckets\n");

	for (k = 0; k < m; ++k)
	{
		i = b[k];
		if (i == 0)
		{
			P ("\t%4ld", k);
			j = k;
			while (k + 1 < m && b[k + 1] == 0)
				++k;
			if (k > j)
				P (" - %4ld", k);
			else
				P ("       ");
			P (": -");
		}
		else
		{
			P ("\t%4ld       :", k);
			for (; i; i = i->next)
				P (" [%s]", S(i->obj));
		}
		P ("\n");
	}
}



/*-----------------------------------------------------------------------------
	dump
-----------------------------------------------------------------------------*/

static qu_null Dump (qu_any self)
{
	static qu_string sdump = 0;

	__qu_unsafe__;

	Dump_default (self);

#define d_(klass) if (qu_##klass##_is (self)) Dump_##klass (self); else

	d_(class)
	d_(null)
	d_(boolean)
	d_(int)
	d_(long)
	d_(rational)
	d_(float)
	d_(big)
	d_(complex)
	d_(inf)
	d_(string)
	d_(buffer)
	d_(dict)
	d_(set)
	d_(array)
	d_(queue)
	{};

	if (sdump == 0)
		sdump = qu_string_new_staticz ("__dump");

	qu_object_call0 (self, sdump);
	return QuNil;
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

int qu_Dump_init (qu_module m)
{
	qu_module_add_function (m, "Dump", Dump, 1);

	return 1;
}
