/* Base64.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Base64
	This module deals with the Base64 encoding algorithm as per RFC-2045.

	@xmp
		use "Encode/Base64"

		s = Base64.encode_string ('fooba')
		println (s)

	@out
		Zm9vYmE=

	@p
	Following is an example on how you could do a buffered encoding.

	@xmp
		use "Encode/Base64"

		state = Base64.EncodeState (false, 0, 0)

		f = File.open ('/opt/qu/demo/sample.qu')
		buf = Buffer (100)
		while f.recv (buf, buf.avail)
			print (Base64.encode (buf, state))
			println
			println (' ... and continuing')
			buf.chop (-state.consumed)
		;;
		f.close

		state.flush = true
		print (Base64.encode (buf, state))
		println
		println ('done!')

	@out
		IyBBIHNhbXBsZSBwcm9ncmFtCgpoZWxsbwoKc3ViIGhlbGxvIChtc2cgPSBuaWwpCglgYGAKCVBy
		aW50cyBhIG1lc3NhZ2UgZm9yIHRoZSB3b3JsZC4KCWBgYAoJbG9jYWwgaGVhZCA9IGZhbHNlCglk
		byB0cnkKCQltc2cg
		... and continuing
		IyBBIHNhbXBsZSBwcm9ncmFtCgpoZWxsbwoKc3ViIGhlbGxvIChtc2cgPSBu
		aWwpCglgYGAKCVByaW50cyBhIG1lc3NhZ2UgZm9yIHRoZSB3b3JsZC4KCWBgYAoJbG9jYWwgaGVh
		ZCA9IGZhbHNlCglkbyB0cnkKCQltc2cgaXM=
		done!
*/


/*----------------------------------------------------------------------------
	encoding
----------------------------------------------------------------------------*/

/* the Base64 alphabet */
static char EncodeAlphabet [] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/*
	maximum number of encoded characters per-line, excluding the linebreak
	note that this is a multiple of 4
*/
#define ENCODE_MAXPERLINE 76

/*
	the required space for encoding n bytes
	this is an over-estimate close enough to the final result
	it works for all Int objects
*/
#define ENCODE_LEN(n) \
	((((n) + 4) / 3 * 4 + 1) + ((n) + ENCODE_MAXPERLINE) / ENCODE_MAXPERLINE)

/*
	3 bytes are encoded as 4 bytes
	!!! both i and o must be unsigned char pointers
*/
#define ENCODE_3_BYTES(i, o) \
QU_BEGIN \
	*o++ = EncodeAlphabet [(i[0] >> 2) & 0x3F]; \
	*o++ = EncodeAlphabet [((i[0] & 0x3) << 4) | ((i[1] & 0xF0) >> 4)]; \
	*o++ = EncodeAlphabet [((i[1] & 0xF) << 2) | ((i[2] & 0xC0) >> 6)]; \
	*o++ = EncodeAlphabet [i[2] & 0x3F]; \
	i += 3; \
QU_END

/*
	the last 1 or 2 bytes are also encoded as 4 bytes
	!!! both i and o must be unsigned char pointers
*/
#define ENCODE_LAST_1_OR_2_BYTES(i, o, n) \
QU_BEGIN \
	*o++ = EncodeAlphabet [(i[0] >> 2) & 0x3F]; \
	if (n == 1) \
	{ \
		*o++ = EncodeAlphabet [((i[0] & 0x3) << 4)]; \
		*o++ = '='; \
	} \
	else \
	{ \
		*o++ = EncodeAlphabet [((i[0] & 0x3) << 4) | ((i[1] & 0xF0) >> 4)]; \
		*o++ = EncodeAlphabet [((i[1] & 0xF) << 2)]; \
	} \
	*o++ = '='; \
	i += n; \
QU_END


/*) struct EncodeState
		flush    : Boolean # whether we're in a flush state
		counter  : Int  # the current line counter
		consumed : Int  # the length consumed
	;;
	This is the I{state-machine} used by #{encode}.
*/
static qu_struct EncodeStateStruct;



/*
	This is the encoder engine.
	Returns the bytes output, and update consumed/counter.
	Sets errno = EOVERFLOW if we need more space than avail, sets errno = 0 otherwise.
*/
static long _encode (int flush, unsigned char *i, long l, unsigned char *o, long avail,
                     long *consumed, int *counter)
{
	unsigned char *p = o;
	int  overflow = 0;
	long len      = l;
	long cnt      = *counter;

	while (l >= 3)
	{
		if (avail < 5)
		{
			overflow = EOVERFLOW;
			break;
		}
		ENCODE_3_BYTES (i, o);
		l -= 3;
		avail -= 4;
		cnt += 4;
		if (cnt == ENCODE_MAXPERLINE)
		{
			*o++ = '\n';
			--avail;
			cnt = 0;
		}
	}
	if (flush && l && !overflow && avail > 3)
	{
		assert (l == 1 || l == 2);
		ENCODE_LAST_1_OR_2_BYTES (i, o, l);
		l = 0;
	}
	errno = overflow;
	*counter  = cnt;
	*consumed = len - l;
	return o - p;
}



/*) sub encode_string (x) -> String
	Encode v{x} into a Base64 string.

	v{x} must be a String or Buffer.

	This is similar with #{encode} with the v{flush} field set to true.

	@exc
	.EArgValue
	.EMemory

	@xmp
		use "Encode/Base64"
		println (Base64.encode_string ('fooba'))

	@out
		Zm9vYmE=
*/
static qu_string Base64_encode_string (qu_string x)
{
	int       counter = 0;
	long      consumed, l, n;
	qu_string r;

	QU_ARG_TYPE (x, qu_string);
	l = x->len;
	if (l == 0)
		return QuStringEMPTY;

	/* EMemory should only happen if the string is too damn huge */
	n = ENCODE_LEN (l);
	QU_MEM_CHECK (n < QU_MEM_MAX);

	/* just encode everything and ignore the remainder */
	r = qu_string_buffer_new (n);
	r->len = _encode (1, (unsigned char *) x->str, l, (unsigned char *) r->str, n, &consumed, &counter);

	/* take care of the nil terminator, empty and single character string */
	return qu_string_buffer_return (r);
}



/*) sub encode (x, state) -> String
	Encode v{x} into a Base64 string.

	This does the same as #{encode_string} except that the encoding can be done in
	stages.

	v{x} must be a String or Buffer.

	v{state} must be an #{EncodeState} record.
	@ul
	.The v{flush} field must be set to true to encode everything, v{false}
	 otherwise.

	.The v{counter} field must be set to zero on first call. It will be set by this
	 function after each call, to zero if the v{flush} field is true.

	.The v{consumed} field is set by this function after each call.
	 It will contain the total number of bytes encoded from the string.

	@p
	Note that value of the v{counter} field must be within 0..72 at all time, otherwise
	you will get a screwed-up result. The best thing is do not set the value yourself
	except for the first time.

	@exc
	.EArgValue
	.EMemory

	@xmp
		use "Encode/Base64"

		s = 'fooba'

		state = Base64.EncodeState (false, 0, 0)
		result = Base64.encode (s, state)
		println (state.consumed)
		println (result)
		println ('--------------')

		state.flush = true
		result = Base64.encode (s, state)
		println (state.consumed)
		println (result)

	@out
		4
		Zm9v
		--------------
		7
		Zm9vYmE=

	@p
	See the %{Base64} module documentation for another example.
*/
static qu_string Base64_encode (qu_string x, qu_record state)
{
	long      consumed, l, n;
	int       counter = 0;
	int       flush = 0;
	qu_string r;

	QU_ARG_CHECK (qu_string_is (x) && qu_record_of (state, EncodeStateStruct));

	/* read from the state Record */
	flush = state->objs[0] == QuTrue;
	counter = qu_int_toi (state->objs[1]); /* I don't care for bad value */

	l = x->len;
	if (l == 0)
	{
		/* string is empty, just set 'consumed' to 0 */
		state->objs[2] = QuInt0;
		return QuStringEMPTY;
	}

	/* EMemory should only happen if the string is too damn huge */
	n = ENCODE_LEN (l);
	QU_MEM_CHECK (n < QU_MEM_MAX);

	r = qu_string_buffer_new (n);
	r->len = _encode (flush, (unsigned char *) x->str, l, (unsigned char *) r->str, n, &consumed, &counter);

	/* don't forget to update the state */
	state->objs[1] = flush ? QuInt0 : qu_int_new (counter);
	state->objs[2] = qu_int_new (consumed);

	/* take care of the nil terminator, empty and single character string */
	return qu_string_buffer_return (r);
}



/*) sub encode_buffer (input: Buffer, output: Buffer, counter = 0, flush = false) -> Int
	This does the same as #{encode} except that it is more efficient and takes an
	v{input} Buffer and v{output} Buffer avoiding the need for a double copy.

	Since the only I{state-machine} required is the characters-per-line v{counter}
	then it is passed as an argument. Just call this function first time with v{counter}
	set to 0 and this function will return the next counter.

	This function is aware of the v{output} Buffer capacity and will try to encode
	only such that does not cause EOverflow, except if v{flushing}.

	If v{flushing} and the output Buffer does not have enough room then EOverflow
	is thrown.

	@warn
	The consumed bytes from the v{input} Buffer are automatically removed.

	@exc
	.EArgValue
	.EOverflow
*/
static qu_int Base64_encode_buffer (int argc, qu_any *argv)
{
	qu_buffer in, out;
	long      lout, consumed;
	int       counter = 0;
	int       flush = 0;

	QU_ARG_VCOUNT (2, 4);
	QU_ARG_VTYPE (0, in, qu_buffer);
	QU_ARG_VTYPE (1, out, qu_buffer);
	QU_ARG_CHECK (in != out);
	if (argc > 2)
	{
		QU_ARG_vint (2, counter);  /* I don't care for bad value */
		if (argc > 3)
			QU_ARG_vbool (3, flush);
	}

	lout = _encode (flush, (unsigned char *) in->str, in->len, (unsigned char *) out->str + out->len,
					out->cap - out->len, &consumed, &counter);

	if (errno && flush)
	{
		/* we are here because we are flushing and there was not enough room */
		long remain = in->len - consumed;
		long more   = ENCODE_LEN (remain);
		long taken  = consumed;

		/* try make room - this throws EOverflow on failure */
		QU_ENSURE (qu_buffer_make_space (out, more));

		/* this always succeeds */
		lout += _encode (flush, (unsigned char *) in->str + consumed, remain, 
				(unsigned char *) out->str + out->len + lout, more, &consumed, &counter);
		assert (errno == 0);
		consumed += taken;
	}

	/* don't forget to adjust the output qu_buffer */
	out->len += lout;
	out->str [out->len] = '\0';

	/* and remove what we have consumed from the input qu_buffer */
	qu_buffer_consume (in, consumed);

	/* and finally return the new counter */
	return flush ? QuInt0 : qu_int_new (counter);
}



/*) sub encode_stream (input: Stream, output: Stream) -> Nil
	This does the same as #{encode_buffer} except that it reads from an input
	Stream and writes to an output Stream.

	@warn
	This function does not work with input Stream without EOF.

	@exc
	.EArgValue
	.Exception anything thrown by the Stream
*/
static qu_null Base64_encode_stream (qu_stream in, qu_stream out)
{
	char i [2048], o [4096];
	long len, consumed;
	int  counter = 0;
	long remain = 0;

	QU_ARG_CHECK (qu_stream_is (in) && qu_stream_is (out) && in != out);
	for (;;)
	{
		/* this should never happen */
		QU_REQUIRE (QuEOverflow, sizeof (i) > remain);

		len = qu_stream_read (in, i + remain, sizeof (i) - remain, 0);
		QU_ENSURE (len >= 0);
		if (len == 0)
			break;

		remain += len;
		len = _encode (0, (unsigned char *) i, remain, (unsigned char *) o, sizeof (o), &consumed, &counter);
		if (consumed < remain)
		{
			remain -= consumed;
			memcpy (i, i + consumed, remain);
		}
		QU_ENSURE (qu_stream_write (out, o, len, 1) >= 0);
	}

	if (remain)
	{
		len = _encode (1, (unsigned char *) i, remain, (unsigned char *) o, sizeof (o), &consumed, &counter);
		QU_ENSURE (qu_stream_write (out, o, len, 1) >= 0);
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	decoding
----------------------------------------------------------------------------*/

#define _o_ 64
static unsigned char DecodeValue [256] =
{
	/*   0 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/*   8 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/*  16 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/*  24 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/*  32 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/*  40 */ _o_, _o_, _o_,  62, _o_, _o_, _o_,  63,
	/*  48 */  52,  53,  54,  55,  56,  57,  58,  59,
	/*  56 */  60,  61, _o_, _o_, _o_, _o_, _o_, _o_,
	/*  64 */ _o_,   0,   1,   2,   3,   4,   5,   6,
	/*  72 */   7,   8,   9,  10,  11,  12,  13,  14,
	/*  80 */  15,  16,  17,  18,  19,  20,  21,  22,
	/*  88 */  23,  24,  25, _o_, _o_, _o_, _o_, _o_,
	/*  96 */ _o_,  26,  27,  28,  29,  30,  31,  32,
	/* 104 */  33,  34,  35,  36,  37,  38,  39,  40,
	/* 112 */  41,  42,  43,  44,  45,  46,  47,  48,
	/* 120 */  49,  50,  51, _o_, _o_, _o_, _o_, _o_,
	/* 128 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 136 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 144 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 152 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 160 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 168 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 176 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 184 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 192 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 200 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 208 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 216 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 224 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 232 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 240 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_,
	/* 248 */ _o_, _o_, _o_, _o_, _o_, _o_, _o_, _o_
};
#undef _o_


/* valid characters */
static unsigned char EncodeValid [256] =
{
	/*   0 */   0,   0,   0,   0,   0,   0,   0,   0,
	/*   8 */   0,   0,   0,   0,   0,   0,   0,   0,
	/*  16 */   0,   0,   0,   0,   0,   0,   0,   0,
	/*  24 */   0,   0,   0,   0,   0,   0,   0,   0,
	/*  32 */   0,   0,   0,   0,   0,   0,   0,   0,
	/*  40 */   0,   0,   0, '+',   0,   0,   0, '/',
	/*  48 */ '0', '1', '2', '3', '4', '5', '6', '7',
	/*  56 */ '8', '9',   0,   0,   0, '=',   0,   0,
	/*  64 */   0, 'A', 'B', 'C', 'D', 'E', 'F', 'G',
	/*  72 */ 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
	/*  80 */ 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
	/*  88 */ 'X', 'Y', 'Z',   0,   0,   0,   0,   0,
	/*  96 */   0, 'a', 'b', 'c', 'd', 'e', 'f', 'g',
	/* 104 */ 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
	/* 112 */ 'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
	/* 120 */ 'x', 'y', 'z',   0,   0,   0,   0,   0,
	/* 128 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 136 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 144 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 152 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 160 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 168 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 176 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 184 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 192 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 200 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 208 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 216 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 224 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 232 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 240 */   0,   0,   0,   0,   0,   0,   0,   0,
	/* 248 */   0,   0,   0,   0,   0,   0,   0,   0
};


/*
	the required space for decoding n bytes
	this is either accurate or an over-estimate very close to the final result
	it works for all Int objects
*/
#define DECODE_LEN(n) \
	(((n) + 3) / 4 * 3 + 1)

/*
	4 bytes are decoded as 3 bytes
	!!! both i and o must be unsigned char pointers
*/
#define DECODE_4_BYTES(i, o) \
QU_BEGIN \
	o[0] = (DecodeValue [i[0]] << 2) | (DecodeValue [i[1]] >> 4); \
	o[1] = (DecodeValue [i[1]] << 4) | (DecodeValue [i[2]] >> 2); \
	o[2] = (DecodeValue [i[2]] << 6) | (DecodeValue [i[3]]); \
QU_END


/*
	n will be the number of the resulting decoded bytes
	!!! both i and o must be unsigned char pointers
*/
#define DECODE_LAST_4_BYTES(i, o, n) \
QU_BEGIN \
	n = 1;\
	o[0] = (DecodeValue [i[0]] << 2) | (DecodeValue [i[1]] >> 4); \
	o[1]  = DecodeValue [i[1]] << 4; \
	if (i[2] != '=') \
	{ \
		n = 2; \
		o[1] |= DecodeValue [i[2]] >> 2; \
		o[2]  = DecodeValue [i[2]] << 6; \
		if (i[3] != '=') \
		{ \
			n = 3; \
			o[2] |= DecodeValue [i[3]]; \
		} \
	} \
QU_END


/*) struct DecodeState
		flush    : Boolean # whether we're in a flush state
		final    : Boolean # true if the last decoded octet contains '='
		consumed : Int  # the length consumed
	;;
	This is the I{state-machine} used by #{decode} and #{decode_buffer}.
*/
static qu_struct DecodeStateStruct;



/*
	This is the decoder engine.
	Returns the bytes output, and update consumed/final.
	Sets errno = EOVERFLOW if we need more space than avail, sets errno = 0 otherwise.
*/
static long _decode (unsigned char *i, long l, unsigned char *buf4, int *bytes,
                     unsigned char *o, long avail, long *consumed, int *final)
{
	unsigned char *p = i;
	unsigned char *b = o;
	unsigned char *start = i;
	unsigned char c;
	int           overflow = 0;
	long          n;

	*final = 0;
	while (l--)
	{
		c = *i++;

		if (!(EncodeValid [c] || (c == '=' && *bytes > 1)))
		{
			if (*bytes == 0)
				start = i;
			continue;
		}

		buf4[*bytes] = c;
		*bytes += 1;
		if (*bytes < 4)
			continue;

		start = i;

		if (avail < 3)
		{
			overflow = EOVERFLOW;
			break;
		}
		avail -= 3;

		if (buf4[2] == '=' || buf4[3] == '=')
		{
			DECODE_LAST_4_BYTES (buf4, o, n);
			o += n;
			*final = 1;
			break;
		}
		else
		{
			DECODE_4_BYTES (buf4, o);
			o += 3;
			*bytes = 0;
		}
	}
	errno = overflow;
	*consumed = start - p;
	return o - b;
}



/*) sub decode_string (x) -> String
	Decode an encoded Base64 string.

	v{x} must be a String or Buffer.

	This does the same as #{decode} with the v{flush} field set to true.

	Note that the remaining unencoded characters, if any, are not included.

	@xmp
		NTY3ODkt**3*A*
		        <----> these are ignored
	@exc
	.EArgValue
	.EMemory

	@xmp
		use "Encode/Base64"
		println (Base64.decode_string ('Zm9vZm8=....'))

	@out
		fooba
*/
static qu_string Base64_decode_string (qu_string x)
{
	unsigned char buf4[4];
	long          consumed, l, n;
	int           bytes = 0, final;
	qu_string     r;

	QU_ARG_TYPE (x, qu_string);
	l = x->len;
	if (l == 0)
		return QuStringEMPTY;

	/* EMemory should only happen if the string is too damn huge */
	n = DECODE_LEN (l);
	QU_MEM_CHECK (n < QU_MEM_MAX);

	/* encode everything and ignore the remainder */
	r = qu_string_buffer_new (n);
	r->len = _decode ((unsigned char *) x->str, l, buf4, &bytes, (unsigned char *) r->str, n, &consumed, &final);

	/* take care of empty and single character string */
	return qu_string_buffer_return (r);
}



/*) sub decode (x, state) -> String
	Decode a Base64 string v{x}.

	This does the reverse of #{decode}.

	v{x} must be a String or Buffer.

	Per RFC-2045, non-Base64 alphabet characters are ignored and the presence of
	'=' at the 3rd or 4th byte position of the octet is taken as evidence that
	the end of the data is reached. Note however that the silly algorithm does not
	ensure that this is always the case since the last octet may not contain any
	'=' at all.

	v{state} must be an #{DecodeState} record.
	@ul
	.The v{flush} field must be set to true to decode everything, v{false}
	 otherwise.

	.The v{final} field is set by this function to true if a final octet is
	 encountered, false otherwise.

	.The v{consumed} field is set by this function after each call.
	 It will contain the number of bytes decoded from the string.

	@exc
	.EArgValue
	.EMemory

	@xmp
		use "Encode/Base64"

		s = 'Zm9vYmFmb29iYWZvb2JhZm9vYmFmb29iYWZvb2JhZm.....'

		state = Base64.DecodeState (false, false, 0)
		result = Base64.decode (s, state)
		println (state.consumed ' ' state.final)
		println (result)
		println ('REMAINING: ' s[state.consumed,])
		println ('--------------')

		s = 'Zm9vYmFmb29iYWZvb2JhZm9vYmFmb29iYWZvb2JhZm9vYmE=.....'
		state.flush = true
		result = Base64.decode (s, state)
		println (state.consumed ' ' state.final)
		println (result)

	@out
		40 false
		foobafoobafoobafoobafoobafooba
		REMAINING: Zm.....
		--------------
		48 true
		foobafoobafoobafoobafoobafoobafooba
*/
static qu_string Base64_decode (qu_string x, qu_record state)
{
	long          consumed, l, n;
	int           flush = 0;
	unsigned char buf4 [4];
	int           bytes = 0, final;
	qu_string     r;

	QU_ARG_CHECK (qu_string_is (x) && qu_record_of (state, DecodeStateStruct));

	/* read the state qu_record - we only need the 'flush' field */
	flush = state->objs[0] == QuTrue;

	l = x->len;
	if (l == 0)
	{
		/* empty string, just set 'final' to true and 'consumed' to 0 */
		state->objs[1] = QuFalse;
		state->objs[2] = QuInt0;
		return QuStringEMPTY;
	}

	/* EMemory is only thrown if the string is too damn huge */
	n = DECODE_LEN (l);
	QU_MEM_CHECK (n < QU_MEM_MAX);

	r = qu_string_buffer_new (n);
	r->len = _decode ((unsigned char *) x->str, x->len, buf4, &bytes, (unsigned char *) r->str, n, &consumed, &final);

	/* don't forget the nil terminator */
	r->str[r->len] = '\0';

	/* and update the state */
	state->objs[1] = ((final || flush) ? QuTrue : QuFalse);
	state->objs[2] = qu_int_new (consumed);

	/* take care of empty and single character string */
	return qu_string_buffer_return (r);
}



/*) sub decode_buffer (input: Buffer, output: Buffer, flush = false) -> Boolean
	This does the same as #{decode} except that it is more efficient and takes an
	v{input} Buffer and v{output} Buffer avoiding the need for a double copy.

	Since the only I{state-machine} required is v{flush} then it is passed as an
	argument. This function will return whether or not the final octect is found.

	This function is aware of the v{output} Buffer capacity and will try to decode
	only such that does not cause EOverflow, except if v{flushing}.

	If v{flushing} and the output Buffer does not have enough room then EOverflow
	is thrown.

	@warn
	The consumed bytes from the v{input} Buffer are automatically removed.

	@exc
	.EArgValue
	.EOverflow
*/
static qu_boolean Base64_decode_buffer (int argc, qu_any *argv)
{
	qu_buffer     in, out;
	long          lout, consumed;
	int           flush = 0;
	unsigned char buf4 [4];
	int           bytes, final;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, in, qu_buffer);
	QU_ARG_VTYPE  (1, out, qu_buffer);
	QU_ARG_CHECK  (in != out);
	if (argc > 2)
		QU_ARG_vbool (2, flush);

	bytes = 0;
	lout = _decode ((unsigned char *) in->str, in->len, buf4, &bytes,
				(unsigned char *) out->str + out->len,
				out->cap - out->len, &consumed, &final);

	if (errno && flush)
	{
		/* we are here because we are flushing and there was not enough room */
		long remain = in->len - consumed;
		long more   = DECODE_LEN (remain);
		long taken  = consumed;

		/* try make room - this throws EOverflow on failure */
		QU_ENSURE (qu_buffer_make_space (out, more));

		/* this always succeeds */
		lout += _decode ((unsigned char *) in->str + consumed, remain, buf4, &bytes,
					(unsigned char *) out->str + out->len + lout,
					more, &consumed, &final);
		assert (errno == 0);
		consumed += taken;
	}

	/* don't forget to adjust the output qu_buffer */
	out->len += lout;
	out->str [out->len] = '\0';

	/* and remove what we have consumed from the input qu_buffer */
	qu_buffer_consume (in, consumed);

	/* and return whether we encountered the final octet */
	return final ? QuTrue : QuFalse;
}



/*) sub decode_stream (input: Stream, output: Stream) -> Nil
	This does the same as #{decode_buffer} except that it reads from an input
	Stream and writes to an output Stream.

	@warn
	This function does not work with input Stream without EOF.
	It also may read past a final octet.

	@exc
	.EArgValue
	.Exception anything thrown by the Stream
*/
static qu_null Base64_decode_stream (qu_stream in, qu_stream out)
{
	char          i [2048], o [2048];
	long          len, consumed, remain = 0;
	unsigned char buf4 [4];
	int           bytes = 0, final = 0;

	QU_ARG_CHECK (qu_stream_is (in) && qu_stream_is (out) && in != out);
	while (!final)
	{
		/* this should never happen */
		QU_REQUIRE (QuEOverflow, sizeof (i) > remain);

		len = qu_stream_read (in, i + remain, sizeof (i) - remain, 0);
		QU_ENSURE (len >= 0);
		if (len == 0)
			break;

		remain += len;
		len = _decode ((unsigned char *) i, remain, buf4, &bytes, (unsigned char *) o, sizeof (o),
				&consumed, &final);
		if (consumed < remain)
		{
			remain -= consumed;
			memcpy (i, i + consumed, remain);
		}
		QU_ENSURE (qu_stream_write (out, o, len, 1) >= 0);
	}

	if (remain && !final)
	{
		len = _decode ((unsigned char *) i, remain, buf4, &bytes, (unsigned char *) o, sizeof (o), &consumed, &final);
		assert (errno == 0);
		QU_ENSURE (qu_stream_write (out, o, len, 1) >= 0);
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Base64_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		encoding
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	EncodeStateStruct = qu_struct_new_static (3,
		"flush"    , QuBooleanClass,
		"counter"  , QuIntClass,
		"consumed" , QuIntClass);

	qu_module_add_const (m, "EncodeState", EncodeStateStruct);

	qu_module_add_function (m, "encode_string", Base64_encode_string,  1);
	qu_module_add_function (m, "encode_buffer", Base64_encode_buffer, -1);
	qu_module_add_function (m, "encode_stream", Base64_encode_stream,  2);
	qu_module_add_function (m, "encode"       , Base64_encode       ,  2);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		decoding
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	DecodeStateStruct = qu_struct_new_static (3,
		"flush"    , QuBooleanClass,
		"final"    , QuBooleanClass,
		"consumed" , QuIntClass);

	qu_module_add_const (m, "DecodeState", DecodeStateStruct);

	qu_module_add_function (m, "decode_string", Base64_decode_string,  1);
	qu_module_add_function (m, "decode_buffer", Base64_decode_buffer, -1);
	qu_module_add_function (m, "decode_stream", Base64_decode_stream,  2);
	qu_module_add_function (m, "decode"       , Base64_decode       ,  2);

	return 1;
}
