/* Chunk.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Chunk
	This module deals with the HTTP Chunk encoding algorithm as per RFC-2616.

	@xmp
		use "Encode/Chunk"

		s = Chunk.encode_string ('foobar')
		println (s)

	@out
		'0006;\r\nfoobar\r\n0;\r\n'
*/



/*----------------------------------------------------------------------------
	encoding
----------------------------------------------------------------------------*/

/* chunk size for encoding */
#define ENCODE_CHUNKSIZE 4080

/*
	the maximum size that can be chunked
	this works for all qu_int values
*/
#define ENCODE_MAX_SIZE(n) \
	(QU_MEM_MAX - 4 - ((n) + ENCODE_CHUNKSIZE) / ENCODE_CHUNKSIZE * 9)

/*
	the required size for encoding n bytes including the terminating chunk
	must only be called after n < ENCODE_MAX_SIZE (n)
*/
static long _encode_len (long n)
{
	long chunks = (n + ENCODE_CHUNKSIZE) / ENCODE_CHUNKSIZE;
	return n + chunks * 9 + 4;
}



/* !!! requires n + 9 bytes of space */
#define ENCODE_CHUNK(i, o, n) \
QU_BEGIN \
	sprintf (o, "%04lX;\r\n", n); \
	o += 7; \
	memcpy (o, i, n); \
	o += n; \
	*o++ = '\r'; \
	*o++ = '\n'; \
	i += n; \
QU_END


/* !!! requires 4 bytes of space */
#define ENCODE_LAST_CHUNK(i, o) \
QU_BEGIN \
	*o++ = '0'; \
	*o++ = ';'; \
	*o++ = '\r'; \
	*o++ = '\n'; \
QU_END



/*) sub encode_string (x) -> String
	Encode v{x} into a Chunk string.

	v{x} must be a String or Buffer.

	Note that you still get some result when v{x} is empty.

	@exc
	.EArgValue
	.EMemory

	@xmp
		use "Encode/Chunk"

		s = 'foo.bar'
		e = Chunk.encode_string (s)
		println (e.quote)

		e = Chunk.encode_string ('')
		println (e.quote)

	@out
		'0007;\r\nfoo.bar\r\n0;\r\n'
		'0;\r\n'
*/
static qu_string Chunk_encode_string (qu_string x)
{
	qu_string  r;
	long       l, n;
	char      *i, *o;

	QU_ARG_TYPE (x, qu_string);
	l = x->len;
	QU_MEM_CHECK (l < ENCODE_MAX_SIZE (l));
	n = _encode_len (l);
	r = qu_string_newl (n);
	i = x->str;
	o = r->str;
	while (l > ENCODE_CHUNKSIZE)
	{
		ENCODE_CHUNK (i, o, (long) ENCODE_CHUNKSIZE);
		l -= ENCODE_CHUNKSIZE;
	}
	if (l)
		ENCODE_CHUNK (i, o, l);
	ENCODE_LAST_CHUNK (i, o);
	*o = '\0';
	l = o - r->str;
	assert (l <= n);
	r->len = l;
	return r;
}



/*) sub encode_buffer (input: Buffer, output: Buffer, flush = false) -> Null
	Encode data in the v{input} Buffer to an v{output} Buffer.

	This function will only consume such that the result does not overflow the
	v{output} Buffer, except when v{flush} is true.

	If v{flush} is true then the v{output} Buffer must have enough space to
	receive all of the encoded data.

	This function figures out the size of the chunks based on the size of the
	buffers. The bigger the v{output} Buffer is, the bigger the chunk size is.
	An ideal v{output} Buffer will be around 4KBytes.

	@warn
	You must always consume the v{output} Buffer after calling this function
	otherwise you might loop forever.

	You must eventually call this function with k{flush} set to true (and
	exactly once only) to produce the last closing chunk, i.e. C{0;[CRLF]}.

	@exc
	.EArgValue
	.EOverflow

	@xmp
		use "Encode/Chunk"

		input  = Buffer ('foo.bar.' * 20)
		output = Buffer (60)
		counter = 0

		while input
			Chunk.encode_buffer (input, output)
			print (output)
			output.__clear
		;;

		Chunk.encode_buffer (input, output, true)
		println (output)

	@out
		0033;
		foo.bar.foo.bar.foo.bar.foo.bar.foo.bar.foo.bar.foo
		0033;
		.bar.foo.bar.foo.bar.foo.bar.foo.bar.foo.bar.foo.ba
		0033;
		r.foo.bar.foo.bar.foo.bar.foo.bar.foo.bar.foo.bar.f
		0007;
		oo.bar.
		0;
*/
static qu_null Chunk_encode_buffer (int argc, qu_any *argv)
{
	qu_buffer  in, out;
	int        flush = 0;
	char      *i, *o;
	long       l, avail;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, in, qu_buffer);
	QU_ARG_VTYPE  (1, out, qu_buffer);
	QU_ARG_CHECK  (in != out);
	if (argc > 2)
		QU_ARG_vbool (2, flush);

	o = out->str + out->len;
	i = in->str;
	l = in->len;
	avail = out->cap - out->len;

	if (flush)
	{
		QU_REQUIRE (QuEOverflow, avail >= (l ? (l + 13) : 4));
		if (l)
		{
			ENCODE_CHUNK (i, o, l);
			l = 0;
		}
		ENCODE_LAST_CHUNK (i, o);
	}
	else
	{
		long target, max, probe = ENCODE_CHUNKSIZE;

		if (probe > in->cap - 13)
			probe = in->cap - 13;
		if (probe > out->cap - 9)
			probe = out->cap - 9;

		QU_REQUIRE (QuEOverflow, probe > 0);

		while (l)
		{
			target = probe;
			if (target > l)
				target = l;
			max = avail - 13;
			if (target >= max)
				target = max;
			if (target < 1)
				break;
			ENCODE_CHUNK (i, o, target);
			l -= target;
			avail -= (target + 9);
		}
	}
	*o = '\0';
	out->len = (char *) o - out->str;
	qu_buffer_consume (in, in->len - l);
	return QuNil;
}



/*) sub encode_stream (input: Stream, output: Stream) -> Null
	Encode data from an v{input} Stream to an v{output} Stream.

	@warn
	This function does not work with Stream objects with no EOF, such as Socket.

	@exc
	.EArgValue
	.Exception anything thrown by v{input} or v{output}

	@xmp
		use "Encode/Chunk"

		f = File.open ('/opt/qu/demo/sample.qu')
		Chunk.encode_stream (f, Stream.stdout)
		f.close
*/
static qu_null Chunk_encode_stream (qu_stream in, qu_stream out)
{
	char i[ENCODE_CHUNKSIZE], b[8];
	long n;

	QU_ARG_CHECK (qu_stream_is (in) && qu_stream_is (out) && in != out);
	for (;;)
	{
		n = qu_stream_read (in, i, sizeof (i), 0);
		QU_ENSURE (n >= 0);
		if (n == 0)
			break;
		sprintf (b, "%04lX;\r\n", n);
		QU_ENSURE (qu_stream_write (out, b, 7, 1) > 0);
		QU_ENSURE (qu_stream_write (out, i, n, 1) > 0);
		QU_ENSURE (qu_stream_write (out, "\r\n", 2, 1) > 0);
	}
	QU_ENSURE (qu_stream_write (out, "0;\r\n", 4, 1) > 0);
	return QuNil;
}



/*----------------------------------------------------------------------------
	decode
----------------------------------------------------------------------------*/

#define STATE_ERROR       -1
#define STATE_GET_HEADER   0
#define STATE_HEADER_CRLF  2
#define STATE_READ_CHUNK   3
#define STATE_CHUNK_CRLF   4
#define STATE_FINISH       5


static long _decode (char *i, long l, char *o, long *consumed,
                     int *state, long *chunksize, long *avail)
{
	long r = l;

	switch (*state)
	{
		case STATE_ERROR:
			/* this always means a syntax error */
		  error:
			*consumed = 0;
			*state = STATE_ERROR;
			return -1;


		case STATE_FINISH:
			*consumed = 0;
			return 0;


		case STATE_GET_HEADER:
			if (l < 3)
			{
				*consumed = 0;
				return 0;
			}
			else
			{
				long size = 0;
				while (l && qu_ascii_isxdigit (*i))
				{
					/* limit to a sane size */
					if (size >= (QU_INT_MAX / 4))
						goto error;
					size = size * 16 + qu_ascii_hexval (i[0]);
					--l;
					++i;
				}

				/* must have a terminator other than a hex digit */
				if (l == 0)
				{
					*consumed = 0;
					return 0;
				}

				*chunksize = size;
				*state = STATE_HEADER_CRLF;
				*consumed = r - l;
				return 0;
			}


		case STATE_HEADER_CRLF:
		{
			long n = -1;
			while (l > 1)
			{
				if (i[0] == '\r' && i[1] == '\n')
				{
					if (*chunksize <= 0)
						*state = STATE_FINISH;
					else
						*state = STATE_READ_CHUNK;

					l -= 2;
					*consumed = r - l;
					return 0;
				}
				--l;
				++i;
				++n;
			}
			*consumed = (n > 0) ? n : 0;
			return 0;
		}


		case STATE_READ_CHUNK:
		{
			long n = *chunksize;
			long m;

			if (n > l)
				n = l;

			if (n > *avail)
				n = *avail;

			*avail -= n;
			*chunksize -= n;
			m = n;
			l -= n;
			while (n--)
				*o++ = *i++;
			if (*chunksize == 0)
			{
				if (l >= 2)
				{
					if (i[0] != '\r' || i[1] != '\n')
						goto error;
					l -= 2;
					*state = STATE_GET_HEADER;
				}
				else
				{
					*state = STATE_CHUNK_CRLF;
				}
			}
			*consumed = r - l;
			return m;
		}


		case STATE_CHUNK_CRLF:
			if (l < 2)
			{
				*consumed = 0;
				return 0;
			}
			if (i[0] != '\r' || i[1] != '\n')
				goto error;
			*state = STATE_GET_HEADER;
			*consumed = 2;
			return 0;
	}
	QU_FATAL ("never");
	return -1;
}



/*) sub decode_string (x: String, maxlen = 8388608) -> String
	Decode a Chunk string.

	All the chunked data information including the terminating chunk must be available
	otherwise you get a ESyntax.

	This function works based on the fact that it must consume something at each
	decoding step. If it can't then the data is incomplete and ESyntax is thrown.

	This is different than #{decode_buffer} since the data may come later after each
	call, thus requires a I{state-machine}.

	@exc
	.EArgValue
	.EOverflow decoded data is bigger than v{maxlen}
	.ESyntax   invalid syntax

	@xmp
		use "Encode/Chunk"

		x = '03;\r\nabc\r\n4;\r\ndefg\r\n00000;\r\n...ignored'
		s = Chunk.decode_string (x)
		println (s)

	@out
		abcdefg
*/
static qu_string Chunk_decode_string (int argc, qu_any *argv)
{
	qu_string  x;
	int        state = STATE_GET_HEADER;
	long       maxlen = 8L * 1024L * 1024L;
	long       l, n, consumed, avail, size, chunksize;
	char      *i;
	qu_string  r;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, x, qu_string);
	if (argc > 1)
	{
		QU_ARG_vlong (1, maxlen);
		QU_ARG_CHECK  (maxlen >= 0);
		QU_MEM_CHECK  (maxlen < QU_MEM_MAX);
	}

	l = x->len;
	if (l == 0)
		return QuStringEMPTY;

	avail = l;
	r = qu_string_buffer_new (l);
	i = x->str;
	size = 0;

	while (state != STATE_FINISH)
	{
		if (avail <= 0)
		{
			QU_ENSURE (qu_string_buffer_make_space (r, chunksize));
			avail = chunksize;
		}
		n = _decode (i, l, r->str + size, &consumed, &state, &chunksize, &avail);
		QU_REQUIRE (QuESyntax, consumed && n >= 0);
		i += consumed;
		l -= consumed;
		avail -= n;
		size += n;
		r->len = size;
		QU_REQUIRE (QuEOverflow, size <= maxlen);
	}
	return qu_string_buffer_return (r);
}



/*) sub decode_buffer (*arg) -> Any
	@code
	sub decode_buffer -> Any
	sub decode_buffer (input: Buffer, output: Buffer, state) -> Any
	sub decode_buffer (input: Buffer, output: Buffer, state) -> Boolean
	@p
	Decode an input Buffer into an output Buffer.

	Decoding a Chunk data requires a I{state-machine}.

	First you must call this function without any argument. The returned value is
	the I{state-machine} which you must supply for subsequent calls.

	If this function function returns false then it means the end of the data
	is reached. Otherwise you must consume the v{output} Buffer then call this
	function again passing the returned updated I{state-machine}.

	@warn
	If you don't consume the v{output} Buffer then EOverflow is thrown after
	four times trying.

	@exc
	.EArgValue
	.ESyntax
	.EOverflow

	@xmp
		use "Encode/Chunk"

		i = Buffer (100)
		o = Buffer (32)

		state = Chunk.decode_buffer

		i <<< '0024;\r\n123456789-123456789-123456789-123456\r\n0004;\r\nab'
		state = Chunk.decode_buffer (i, o, state)
		println ('out > ' o.quote)
		println ('in  > ' i.quote)
		o.__clear

		i <<< 'de\r\n0000;\r\n...ignored'
		println ('in  > ' i.quote)

		loop
			state = Chunk.decode_buffer (i, o, state)
			println ('out > ' o.quote)
			println ('in  > ' i.quote)
			o.__clear
		until not state

	@out
		out > '123456789-123456789-123456789-12'
		in  > '3456\r\n0004;\r\nab'
		in  > '3456\r\n0004;\r\nabde\r\n0000;\r\n...ignored'
		out > '3456abde'
		in  > '...ignored'

	@note
	To cope with abusive data, e.g. a never ending header, then you would
	generally limit the size of the v{input} Buffer combined with a timeout.
*/
static qu_any Chunk_decode_buffer (int argc, qu_any *argv)
{
	qu_array sta;

	if (argc == 0)
	{
		qu_any arg[3] = {qu_int_new (STATE_GET_HEADER), QuInt0, QuInt0};
		sta = qu_array_new_argv (3, arg);
		qu_array_freeze (sta);
		return sta;
	}
	else
	{
		qu_buffer in, out;
		qu_array  sta;
		int       state;
		long      n, chunksize, avail, consumed;
		int       overflow;

		QU_ARG_VEXACT (3);
		in  = argv[0];
		out = argv[1];
		sta = argv[2];

		if (sta == (qu_any) QuFalse)
			return QuFalse;

		QU_ARG_CHECK (qu_buffer_is (in)
			&& qu_buffer_is (out)
			&& in != out
			&& qu_array_is (sta)
			&& sta->len >= 3);

		QU_ARG_int   (sta->objs[0], state);
		QU_ARG_long (sta->objs[1], chunksize);
		QU_ARG_CHECK (chunksize > 0);
		QU_ARG_int   (sta->objs[2], overflow);

		QU_ARG_CHECK (STATE_ERROR <= state && state <= STATE_FINISH);

		avail = out->cap - out->len;
		while (state != STATE_FINISH && in->len)
		{
			n = _decode (in->str, in->len, out->str + out->len, &consumed, &state, &chunksize, &avail);
			QU_REQUIRE (QuESyntax, n >= 0);
			out->len += n;
			out->str[out->len] = '\0';
			if (consumed)
				qu_buffer_consume (in, consumed);
			else
			{
				if (state == STATE_READ_CHUNK && avail <= 0)
				{
					QU_REQUIRE (QuEOverflow, overflow < 4);
					++overflow;
				}
				break;
			}
		}
		sta->objs[0] = qu_int_newi (state);
		if (state == STATE_FINISH)
			return QuFalse;
		sta->objs[1] = qu_int_new (chunksize);
		sta->objs[2] = qu_int_newi (overflow);
		return sta;
	}
}



/*) sub decode_stream (input: Stream, output: Stream) -> Null
	Decode an input Stream into an output Stream.

	@warn
	This function does not work with Stream objects with no EOF, such as Socket.

	@exc
	.EArgValue
	.ESyntax
	.EOverflow

	@p
	The following example confirms the encoding and decoding.

	@xmp
		use "Encode/Chunk"

		src = '/opt/qu/demo/sample.qu'
		dst = '/tmp/foo.txt'

		i = File.open (src)
		o = File.open (dst, 'wct')
		Chunk.encode_stream (i, o)
		i.close
		o.close

		i = File.open (dst)
		o = Stream.stdout
		Chunk.decode_stream (i, o)
		i.close
*/
static qu_any Chunk_decode_stream (qu_stream in, qu_stream out)
{
	int  state = STATE_GET_HEADER;
	char i[1024], o[1024];
	long n, len, chunksize, avail, consumed;
	int  eof = 0;

	QU_ARG_CHECK (qu_stream_is (in) && qu_stream_is (out) && in != out);
	len = qu_stream_read (in, i, sizeof (i), 0);
	QU_ENSURE (len >= 0);
	if (len)
	{
		for (;;)
		{
			avail = sizeof (o);
			n = _decode (i, len, o, &consumed, &state, &chunksize, &avail);
			QU_REQUIRE (QuESyntax, n >= 0);
			if (n)
				QU_ENSURE (qu_stream_write (out, o, n, 1) > 0);
			if (state == STATE_FINISH)
				break;
			len -= consumed;
			if (consumed)
			{
				if (len)
				{
					memcpy (i, i + consumed, len);
					continue;
				}
			}
			QU_REQUIRE (QuESyntax, !eof);
			QU_REQUIRE (QuEOverflow, sizeof (i) > len);
			n = qu_stream_read (in, i + len, sizeof (i) - len, 0);
			QU_ENSURE (n >= 0);
			eof = (n == 0);
			len += n;
		}
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Chunk_init (qu_module m)
{
	qu_module_add_function (m, "encode_string" , Chunk_encode_string ,  1);
	qu_module_add_function (m, "encode_buffer" , Chunk_encode_buffer , -1);
	qu_module_add_function (m, "encode_stream" , Chunk_encode_stream ,  2);

	qu_module_add_function (m, "decode_string" , Chunk_decode_string , -1);
	qu_module_add_function (m, "decode_buffer" , Chunk_decode_buffer , -1);
	qu_module_add_function (m, "decode_stream" , Chunk_decode_stream ,  2);

	return 1;
}
