/* Md5.c - Copyright (c) Marc Krisnanto */

/*
	This work is derived from the RSA Data Security, Inc.
	MD5 Message-Digest Algorithm.
*/
/*
 * Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
 * rights reserved.
 *
 * License to copy and use this software is granted provided that it
 * is identified as the "RSA Data Security, Inc. MD5 Message-Digest
 * Algorithm" in all material mentioning or referencing this software
 * or this function.
 *
 * License is also granted to make and use derivative works provided
 * that such works are identified as "derived from the RSA Data
 * Security, Inc. MD5 Message-Digest Algorithm" in all material
 * mentioning or referencing the derived work.
 *
 * RSA Data Security, Inc. makes no representations concerning either
 * the merchantability of this software or the suitability of this
 * software for any particular purpose. It is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this
 * documentation and/or software.
 */


#include "qu.h"


/*) module Md5
	This module provides an interface to the MD5 message-digest algorithm.

	Note that MD5 is a one-way encoding thus there is not decoding function.

	@xmp
		use "Encode/Md5"

		s = Md5.encode ('foo', 'bar', 'baz')
		println (s)

	@out
		6df23dc03f9b54cc38a0fc1483df6e21
*/



/*----------------------------------------------------------------------------
	MD5 stuffs
----------------------------------------------------------------------------*/

#define MD5_DIGESTLEN 16
#define MD5_HEXLEN    33

#if SIZEOF_INT == 4
#  define MD5_UINT4 unsigned int
#elif SIZEOF_LONG == 4
#  define MD5_UINT4 unsigned long
#else
#  error cant find suitable size for MD5_UINT4
#endif

typedef struct md5_s md5_s;
struct md5_s
{
	MD5_UINT4 state[4];
	MD5_UINT4 count[2];
	unsigned char buffer[64];
};


#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21


static unsigned char MD5_PADDING [64] =
{
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

#define FF(a, b, c, d, x, s, ac) { \
	(a) += F ((b), (c), (d)) + (x) + (MD5_UINT4)(ac); \
	(a) = ROTATE_LEFT ((a), (s)); \
	(a) += (b); }

#define GG(a, b, c, d, x, s, ac) { \
	(a) += G ((b), (c), (d)) + (x) + (MD5_UINT4)(ac); \
	(a) = ROTATE_LEFT ((a), (s)); \
	(a) += (b); }

#define HH(a, b, c, d, x, s, ac) { \
	(a) += H ((b), (c), (d)) + (x) + (MD5_UINT4)(ac); \
	(a) = ROTATE_LEFT ((a), (s)); \
	(a) += (b); }

#define II(a, b, c, d, x, s, ac) { \
	(a) += I ((b), (c), (d)) + (x) + (MD5_UINT4)(ac); \
	(a) = ROTATE_LEFT ((a), (s)); \
	(a) += (b); }



static void _encode (unsigned char *output, MD5_UINT4 *input, unsigned int len)
{
	unsigned int i, j;

	for (i = 0, j = 0; j < len; i++, j += 4)
	{
		output [j]     = (unsigned char) (input[i] & 0xff);
		output [j + 1] = (unsigned char) ((input[i] >>  8) & 0xff);
		output [j + 2] = (unsigned char) ((input[i] >> 16) & 0xff);
		output [j + 3] = (unsigned char) ((input[i] >> 24) & 0xff);
	}
}




static void _decode (MD5_UINT4 *output, const unsigned char *input, unsigned int len)
{
	unsigned int i, j;

	for (i = 0, j = 0; j < len; i++, j += 4)
		output[i] = ((MD5_UINT4) input [j])
			| (((MD5_UINT4) input [j + 1]) << 8)
			| (((MD5_UINT4) input [j + 2]) << 16)
			| (((MD5_UINT4) input [j + 3]) << 24);
}



static void _transform (MD5_UINT4 state[4], const unsigned char block[64])
{
	MD5_UINT4 a, b, c, d, x[MD5_DIGESTLEN];

	a = state[0];
	b = state[1];
	c = state[2];
	d = state[3];

	_decode (x, block, 64);

	FF (a, b, c, d, x[ 0], S11, 0xd76aa478);
	FF (d, a, b, c, x[ 1], S12, 0xe8c7b756);
	FF (c, d, a, b, x[ 2], S13, 0x242070db);
	FF (b, c, d, a, x[ 3], S14, 0xc1bdceee);
	FF (a, b, c, d, x[ 4], S11, 0xf57c0faf);
	FF (d, a, b, c, x[ 5], S12, 0x4787c62a);
	FF (c, d, a, b, x[ 6], S13, 0xa8304613);
	FF (b, c, d, a, x[ 7], S14, 0xfd469501);
	FF (a, b, c, d, x[ 8], S11, 0x698098d8);
	FF (d, a, b, c, x[ 9], S12, 0x8b44f7af);
	FF (c, d, a, b, x[10], S13, 0xffff5bb1);
	FF (b, c, d, a, x[11], S14, 0x895cd7be);
	FF (a, b, c, d, x[12], S11, 0x6b901122);
	FF (d, a, b, c, x[13], S12, 0xfd987193);
	FF (c, d, a, b, x[14], S13, 0xa679438e);
	FF (b, c, d, a, x[15], S14, 0x49b40821);

	GG (a, b, c, d, x[ 1], S21, 0xf61e2562);
	GG (d, a, b, c, x[ 6], S22, 0xc040b340);
	GG (c, d, a, b, x[11], S23, 0x265e5a51);
	GG (b, c, d, a, x[ 0], S24, 0xe9b6c7aa);
	GG (a, b, c, d, x[ 5], S21, 0xd62f105d);
	GG (d, a, b, c, x[10], S22,  0x2441453);
	GG (c, d, a, b, x[15], S23, 0xd8a1e681);
	GG (b, c, d, a, x[ 4], S24, 0xe7d3fbc8);
	GG (a, b, c, d, x[ 9], S21, 0x21e1cde6);
	GG (d, a, b, c, x[14], S22, 0xc33707d6);
	GG (c, d, a, b, x[ 3], S23, 0xf4d50d87);
	GG (b, c, d, a, x[ 8], S24, 0x455a14ed);
	GG (a, b, c, d, x[13], S21, 0xa9e3e905);
	GG (d, a, b, c, x[ 2], S22, 0xfcefa3f8);
	GG (c, d, a, b, x[ 7], S23, 0x676f02d9);
	GG (b, c, d, a, x[12], S24, 0x8d2a4c8a);

	HH (a, b, c, d, x[ 5], S31, 0xfffa3942);
	HH (d, a, b, c, x[ 8], S32, 0x8771f681);
	HH (c, d, a, b, x[11], S33, 0x6d9d6122);
	HH (b, c, d, a, x[14], S34, 0xfde5380c);
	HH (a, b, c, d, x[ 1], S31, 0xa4beea44);
	HH (d, a, b, c, x[ 4], S32, 0x4bdecfa9);
	HH (c, d, a, b, x[ 7], S33, 0xf6bb4b60);
	HH (b, c, d, a, x[10], S34, 0xbebfbc70);
	HH (a, b, c, d, x[13], S31, 0x289b7ec6);
	HH (d, a, b, c, x[ 0], S32, 0xeaa127fa);
	HH (c, d, a, b, x[ 3], S33, 0xd4ef3085);
	HH (b, c, d, a, x[ 6], S34,  0x4881d05);
	HH (a, b, c, d, x[ 9], S31, 0xd9d4d039);
	HH (d, a, b, c, x[12], S32, 0xe6db99e5);
	HH (c, d, a, b, x[15], S33, 0x1fa27cf8);
	HH (b, c, d, a, x[ 2], S34, 0xc4ac5665);

	II (a, b, c, d, x[ 0], S41, 0xf4292244);
	II (d, a, b, c, x[ 7], S42, 0x432aff97);
	II (c, d, a, b, x[14], S43, 0xab9423a7);
	II (b, c, d, a, x[ 5], S44, 0xfc93a039);
	II (a, b, c, d, x[12], S41, 0x655b59c3);
	II (d, a, b, c, x[ 3], S42, 0x8f0ccc92);
	II (c, d, a, b, x[10], S43, 0xffeff47d);
	II (b, c, d, a, x[ 1], S44, 0x85845dd1);
	II (a, b, c, d, x[ 8], S41, 0x6fa87e4f);
	II (d, a, b, c, x[15], S42, 0xfe2ce6e0);
	II (c, d, a, b, x[ 6], S43, 0xa3014314);
	II (b, c, d, a, x[13], S44, 0x4e0811a1);
	II (a, b, c, d, x[ 4], S41, 0xf7537e82);
	II (d, a, b, c, x[11], S42, 0xbd3af235);
	II (c, d, a, b, x[ 2], S43, 0x2ad7d2bb);
	II (b, c, d, a, x[ 9], S44, 0xeb86d391);

	state[0] += a;
	state[1] += b;
	state[2] += c;
	state[3] += d;

	memset (x, 0, sizeof (x));
}



static void _tohex33 (unsigned char dest [MD5_HEXLEN], unsigned char *src)
{
	static unsigned char *hexs = "0123456789abcdef";

	int i;
	unsigned char c;

	for (i = 0; i < MD5_DIGESTLEN; i++)
	{
		c = src [i];
		dest [i * 2] = hexs [(c >> 4) & 0xf];
		dest [i * 2 + 1] = hexs [c & 0xf];
	}
	dest [32] = '\0';
}



/*----------------------------------------------------------------------------
	MD5 interface
----------------------------------------------------------------------------*/

static void md5_init (md5_s *m)
{
	m->count[0] = m->count[1] = 0;
	m->state[0] = 0x67452301;
	m->state[1] = 0xefcdab89;
	m->state[2] = 0x98badcfe;
	m->state[3] = 0x10325476;
}



static void md5_update (md5_s *m, const unsigned char *input, unsigned int inputLen)
{
	unsigned int i, index, partLen;

	index = (unsigned int) ((m->count[0] >> 3) & 0x3F);
	if ((m->count[0] += ((MD5_UINT4) inputLen << 3)) < ((MD5_UINT4) inputLen << 3))
		m->count[1]++;
	m->count[1] += ((MD5_UINT4) inputLen >> 29);
	partLen = 64 - index;
	if (inputLen >= partLen)
	{
		memcpy (&m->buffer[index], input, partLen);
		_transform (m->state, m->buffer);
		for (i = partLen; (i + 63) < inputLen; i += 64)
			_transform (m->state, &input[i]);
		index = 0;
	}
	else
		i = 0;
	memcpy (&m->buffer[index], &input[i], inputLen - i);
}



static void md5_final (md5_s *m, unsigned char digest [MD5_DIGESTLEN])
{
	unsigned char bits [8];
	unsigned int  index, padLen;

	_encode (bits, m->count, 8);
	index  = (unsigned int) ((m->count[0] >> 3) & 0x3f);
	padLen = (index < 56) ? (56 - index) : (120 - index);
	md5_update (m, MD5_PADDING, padLen);
	md5_update (m, bits, 8);
	_encode (digest, m->state, MD5_DIGESTLEN);
	memset (m, 0, sizeof (md5_s));
}



static void md5_finalhex (md5_s *m, unsigned char hexdigest [MD5_HEXLEN])
{
	unsigned char digest [MD5_DIGESTLEN];

	md5_final (m, digest);
	_tohex33 (hexdigest, digest);
}



/*----------------------------------------------------------------------------
	encode
----------------------------------------------------------------------------*/

/*) sub encode (*arg) -> String
	@code
	sub encode (x: String, y: String, z: String, ...) -> String
	sub encode (x: Array) -> String
	sub encode (i: Stream) -> String

	@p
	Encode arguments using the MD5 algorithm.

	All arguments must be String objects, or alternatively a single Array
	containing one or more String objects, or an input Stream.

	@exc
	.EValue

	@xmp
		use "Encode/Md5"

		s = Md5.encode ('foo', 'bar', 'baz')
		println (s)

		a = ['foo', 'bar', 'baz']
		s = Md5.encode (a)         # s = Md5.encode (*a) is not as efficient
		println (s)

	@out
		6df23dc03f9b54cc38a0fc1483df6e21
		6df23dc03f9b54cc38a0fc1483df6e21
*/
static int _md5_String (md5_s *m, int argc, qu_any *argv)
{
	while (argc--)
	{
		qu_stringy s = (qu_stringy) *argv++;
		QU_ARG_TYPE (s, qu_stringy);
		md5_update (m, (unsigned char *) s->str, (size_t) s->len);
	}
	return 1;
}



static int _md5_Array (md5_s *m, qu_array a)
{
	long    l = a->len;
	qu_any *p = a->objs;

	while (l--)
	{
		qu_stringy s = (qu_stringy) *p++;
		QU_ARG_TYPE (s, qu_stringy);
		md5_update (m, (unsigned char *) s->str, (size_t) s->len);
	}
	return 1;
}



static int _md5_Stream (md5_s *m, qu_stream in)
{
	char buf [1024];
	long r;

	for (;;)
	{
		r = qu_stream_read (in, buf, 1024, 0);
		if (r < 0)
			QU_THROWERRNO;
		if (r == 0)
			break;
		md5_update (m, (unsigned char *) buf, (size_t) r);
	}
	return 1;
}



static qu_string Md5_encode (int argc, qu_any *argv)
{
	qu_any    x;
	qu_string s;
	int       r = 0;
	md5_s     m;

	QU_ARG_VMIN (1);
	x = argv[0];

	md5_init (&m);
	if (qu_stringy_is (x))
	{
		r = _md5_String (&m, argc, argv);
	}
	else if (qu_array_is (x))
	{
		r = _md5_Array (&m, x);
	}
	else if (qu_stream_is (x))
	{
		r = _md5_Stream (&m, x);
	}
	else
	{
		QU_ARG_THROW;
	}
	QU_ENSURE (r);
	s = qu_string_newl (MD5_HEXLEN);
	md5_finalhex (&m, (unsigned char *) s->str);
	return s;
}



#undef S11
#undef S12
#undef S13
#undef S14
#undef S21
#undef S22
#undef S23
#undef S24
#undef S31
#undef S32
#undef S33
#undef S34
#undef S41
#undef S42
#undef S43
#undef S44
#undef F
#undef G
#undef H
#undef I
#undef ROTATE_LEFT
#undef FF
#undef GG
#undef HH
#undef II



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Md5_init (qu_module m)
{
	qu_module_add_function (m, "encode", Md5_encode, -1);
	return 1;
}
