/* Quopri.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Quopri
	This module deals with the Quoted-Printable encoding algorithm as per RFC-2045.

	@xmp
		use "Encode/Quopri"

		s = Quopri.encode_string ('foobar' * 20)
		println (s)

	@out
		foobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoo=
		barfoobarfoobarfoobarfoobarfoobarfoobarfoobar


	@h3
	Encoding
		@indent
		Encoding respects the maximum 76 characters-per-line as defined by RFC-2045.

		The I{soft-linebreak} character used is LF (C{'\n'}).

		To make things simpler, and faster, all non-printable characters including C{"."}
		are encoded as their hexadecimal values. This makes the result "not quite readable"
		as the linebreaks will not be presented as themselves. We feel that there is not much
		sense to trade speed for a "readable" output. Besides, you can always do an extra
		processing if you want to be able to read the output.
		@unindent


	@h3
	Decoding
		@indent
		Decoding ignores invalid escape octets and treat them as regular 8-bit characters.

		If v{flush} is true and the last character is C{'='}, for example, then it is
		decoded as C{'='}. It is not considered as a syntax error.

		In other words, decoding treats everything in the input Buffer as decodable
		octets.
		@unindent
*/



/*----------------------------------------------------------------------------
	encoding
----------------------------------------------------------------------------*/

/* list of characters needs to be escaped */
static unsigned char QuopriEscape [256] =
{
	/*   0 */    1,	  1,   1,   1,   1,   1,   1,   1,
	/*   8 */    1,'\t','\n','\v','\f','\r',   1,   1,
	/*  16 */    1,   1,   1,   1,   1,   1,   1,   1,
	/*  24 */    1,   1,   1,   1,   1,   1,   1,   1,
	/*  32 */  ' ',   0,   0,   0,   0,   0,   0,   0,
	/*  40 */    0,   0,   0,   0,   0,   0, '.',   0,
	/*  48 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  56 */    0,   0,   0,   0,   0, '=',   0,   0,
	/*  64 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  72 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  80 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  88 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  96 */    0,   0,   0,   0,   0,   0,   0,   0,
	/* 104 */    0,   0,   0,   0,   0,   0,   0,   0,
	/* 112 */    0,   0,   0,   0,   0,   0,   0,   0,
	/* 120 */    0,   0,   0,   0,   0,   0,   0,   1,
	/* 128 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 136 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 144 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 152 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 160 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 168 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 176 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 184 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 192 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 200 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 208 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 216 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 224 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 232 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 240 */    1,   1,   1,   1,   1,   1,   1,   1,
	/* 248 */    1,   1,   1,   1,   1,   1,   1,   1
};


/*
	maximum number of encoded characters per-line, excluding the linebreak
*/
#define ENCODE_MAX_CHARS_PER_LINE 76


/*
	the maximum size of a string that can be encoded
	an undervalued value but let's be practical
	this works for all Int values
*/
#define ENCODE_MAX_SIZE (QU_MEM_MAX / 4)


static long _encode (char *i, long len, char *o, long avail, long *consumed, long *counter)
{
	char *p = o;
	long  l = len;

	while (l && avail >= 5)
	{
		char c = *i++;
		--l;

		if (QuopriEscape [255 & c])
		{
			if (*counter > ENCODE_MAX_CHARS_PER_LINE - 5)
			{
				*o++ = '=';
				*o++ = '\n';
				avail -= 2;
				*counter = 0;
			}
			sprintf (o, "=%02X", c);
			o += 3;
			avail -= 3;
			*counter += 3;
		}
		else
		{
			if (*counter > ENCODE_MAX_CHARS_PER_LINE - 2)
			{
				*o++ = '=';
				*o++ = '\n';
				avail -= 2;
				*counter = 0;
			}
			*o++ = c;
			--avail;
			*counter += 1;
		}
	}
	*consumed = len - l;
	return o - p;
}



/*) sub encode_string (x) -> String
	Encode v{x} into a Quoted-Printable string.

	v{x} must be a String or Buffer.

	@exc
	.EMemory

	@xmp
		use "Encode/Quopri"

		s = Quopri.encode_string ('foobar' * 20)
		println (s)

	@out
		foobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoobarfoo=
		barfoobarfoobarfoobarfoobarfoobarfoobarfoobar
*/
static qu_string Quopri_encode_string (qu_string x)
{
	long       l, avail, counter, consumed, size;
	char      *i;
	qu_string  r;

	QU_ARG_TYPE (x, qu_string);
	l = x->len;
	if (l == 0)
		return QuStringEMPTY;

	QU_MEM_CHECK (l <= ENCODE_MAX_SIZE);

	avail = l * 3;
	r = qu_string_buffer_new (avail);
	counter = 0;
	size = 0;
	i = x->str;
	for (;;)
	{
		size += _encode (i, l, r->str + size, avail, &consumed, &counter);
		l -= consumed;
		if (l == 0)
			break;
		avail = l * 3 + 6;
		i += consumed;
		QU_ENSURE (qu_string_buffer_make_space (r, avail));
	}
	r->len = size;
	return qu_string_buffer_return (r);
}



/*) sub encode_buffer (input: Buffer, output: Buffer, counter: Integer, flush = false) -> Int
	Encode data in the v{input} Buffer to an v{output} Buffer.

	To produce the encoded data properly, you need to maintain a characters-per-line
	v{counter}. At the first call to this function, you must set it to 0.

	The returned value of this function is the new counter or 0 if v{flush} is true.

	Note that value of the v{counter} must be within 0..72 at all time, otherwise you
	get an EValue exception.

	This function will only consume such that the result does not overflow the
	v{output} Buffer, except when v{flush} is true.

	If v{flush} is true then the v{output} Buffer must have enough space to consume
	all of the encoded data.

	@exc
	.EValue
	.EOverflow

	@xmp
		use "Encode/Quopri"

		input  = Buffer ('foo+bar+' * 20)
		output = Buffer (40)
		counter = 0

		while input
			counter = Quopri.encode_buffer (input, output, counter)
			output or break
			print (output)
			output.__clear
		;;
		Quopri.encode_buffer (input, output, counter, true)
		println (output)

	@out
		foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo=
		+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+bar+foo+ba=
		r+foo+bar+
*/
static qu_int Quopri_encode_buffer (int argc, qu_any *argv)
{
	qu_buffer in, out;
	long      counter = 0;
	int       flush = 0;
	long      len, size, consumed;

	QU_ARG_VCOUNT (2, 4);
	QU_ARG_VTYPE  (0, in, qu_buffer);
	QU_ARG_VTYPE  (1, out, qu_buffer);
	QU_ARG_CHECK  (in != out);
	if (argc > 2)
	{
		QU_ARG_vlong (2, counter);
		QU_ARG_CHECK (0 <= counter && counter <= ENCODE_MAX_CHARS_PER_LINE - 5);
		if (argc > 3)
			QU_ARG_vbool (3, flush);
	}
	len = in->len;
	size = _encode (in->str, len, out->str + out->len, out->cap - out->len, &consumed, &counter);
	out->len += size;
	out->str[out->len] = '\0';
	qu_buffer_consume (in, consumed);
	if (flush && consumed != len)
		QU_THROW (QuEOverflow);
	if (flush)
		return QuInt0;
	return qu_int_new (counter);
}



/*) sub encode_stream (input: Stream, output: Stream) -> Null
	Encode an input Stream into an output Stream.

	@warn
	This function assumes that all bytes read from the input Stream belongs to
	the encoded data.

	It does not work with input Stream without EOF (such as Socket).

	@exc
	.EValue

	@xmp
		use "Encode/Quopri"

		f = File.open ('/opt/qu/demo/sample.qu')
		Quopri.encode_stream (f, Stream.stdout)
		f.close
*/
static qu_null Quopri_encode_stream (qu_stream in, qu_stream out)
{
	char i[1024], o[4096];
	long counter, consumed, l, n;

	QU_ARG_CHECK (qu_stream_is (in) && qu_stream_is (out) && in != out);
	counter = 0;
	for (;;)
	{
		n = qu_stream_read (in, i, sizeof (i), 0);
		if (n < 0)
			return 0;
		if (n == 0)
			break;
		l = _encode (i, n, o, sizeof (o), &consumed, &counter);
		QU_REQUIRE (QuEOverflow, consumed == n);
		if (qu_stream_write (out, o, l, 1) < 0)
			return 0;
	}
	return QuNil;
}




/*----------------------------------------------------------------------------
	decoding
----------------------------------------------------------------------------*/

static long _decode (int flush, char *i, long len, char *o, long avail, long *consumed)
{
	char *p = o;
	long  l = len;

	while (l && avail > 0)
	{
		char c = *i++;
		long m =  l--;

		if (c == '=')
		{
			if (l && qu_ascii_iswhite (*i))
			{
				while (--l && *++i != '\n');
				if (l == 0)
				{
					if (!flush)
						l = m;
					break;
				}
				--l;
				++i;
			}
			else if (l && *i == '\n')
			{
				--l;
				++i;
			}
			else if (l > 1 && qu_ascii_isxdigit (i[0]) && qu_ascii_isxdigit (i[1]))
			{
				*o++ = qu_ascii_hexval (i[0]) * 16 + qu_ascii_hexval (i[1]);
				i += 2;
				l -= 2;
				--avail;
			}
			else if (l < 2)
			{
				if (flush && avail >= l)
				{
					*o++ = '=';
					while (l--)
						*o++ = *i++;
				}
				else
				{
					l = m;
				}
				break;
			}
			else
			{
				*o++ = '=';
				--avail;
			}
		}
		else
		{
			*o++ = c;
			--avail;
		}
	}
	*consumed = len - l;
	return o - p;
}



/*) sub decode_string (x, option = '') -> String
	Decode an encoded Quoted-Printable string.

	v{x} must be a String or Buffer.

	The only v{option} recognized is C{'_'} (underscore) to translate underscore
	characters into spaces.

	@exc
	.EValue
	.EMemory
	.ESyntax

	@xmp
		use "Encode/Quopri"
		println (Quopri.decode_string ('foo_bar=  \n +=202', '_'))

	@out
		foo bar + 2
*/
static qu_string Quopri_decode_string (qu_string x)
{
	long      l, n;
	qu_string r;

	QU_ARG_TYPE (x, qu_string);
	l = x->len;
	if (l == 0)
		return QuStringEMPTY;
	QU_MEM_CHECK (l < QU_MEM_MAX);
	r = qu_string_buffer_new (l);
	r->len = _decode (1, x->str, l, r->str, l, &n);
	return qu_string_buffer_return (r);
}



/*) sub decode_buffer (input: Buffer, output: Buffer, flush = false) -> Null
	Decode data in the v{input} Buffer to an v{output} Buffer.

	This function will only consume the v{input} Buffer such that the result does
	not overflow the v{output} Buffer, except when v{flush} is true.

	If v{flush} is true then the v{output} Buffer must have enough space to consume
	all of the encoded data otherwise EOverflow is thrown.

	@exc
	.EValue
	.EOverflow

	@xmp
		input  = Buffer (100)
		output = Buffer (80)

		input <<< 'foo+=3Dbarfoo+=3Dbarfoo+=3Dbarfoo+=3Dbarfoo+=3'
		Quopri.decode_buffer (input, output)
		println ('out > ' output)
		println ('in  > ' input)
		output.__clear

		input <<< 'Dbarfoo+=3Dbarfoo+=3Db= \t  \r \narfoo+=3Dbarfoo+='
		Quopri.decode_buffer (input, output, true)
		println ('out > ' output)
		println ('in  > ' input)
		output.__clear

	@out
		out > foo+=barfoo+=barfoo+=barfoo+=barfoo+
		in  > =3
		out > =barfoo+=barfoo+=barfoo+=barfoo+=
		in  >
*/
static qu_null Quopri_decode_buffer (int argc, qu_any *argv)
{
	qu_buffer in, out;
	int       flush = 0;
	long      l, n;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, in, qu_buffer);
	QU_ARG_VTYPE  (1, out, qu_buffer);
	QU_ARG_CHECK  (in != out);
	if (argc > 2)
		QU_ARG_vbool (2, flush);
	l = _decode (flush, in->str, in->len, out->str + out->len, out->cap - out->len, &n);
	qu_buffer_consume (in, n);
	out->str[l] = '\0';
	out->len = l;
	if (flush && in->len)
		QU_THROW (QuEOverflow);
	return QuNil;
}




/*) sub decode_stream (input: Stream, output: Stream) -> Null
	Decode an input Stream into an output Stream.

	@warn
	This function assumes that all bytes read from the input Stream belongs to
	the encoded data.

	It does not work with input Stream without EOF (such as Socket).

	EOverflow is thrown in weird cases where the soft-line-break contains long
	whitespaces. If this happens, you should assume that the encoded data is abusive.

	@exc
	.EValue
	.EOverflow

	@xmp
		use "Encode/Quopri"

		f = File.open ('/foo/bar.txt')
		Quopri.decode_stream (f, Stream.stdout)
		f.close
*/
static qu_null Quopri_decode_stream (qu_stream in, qu_stream out)
{
	char i[2048], o[2048];
	long n, l, m;

	QU_ARG_CHECK (qu_stream_is (in) && qu_stream_is (out) && in != out);
	l = 0;
	for (;;)
	{
		QU_REQUIRE (QuEOverflow, sizeof (i) > l);
		n = qu_stream_read (in, i + l, sizeof (i) - l, 0);
		if (n < 0)
			return 0;
		if (n == 0)
			break;
		l += n;
		n = _decode (0, i, l, o, sizeof (o), &m);
		if (qu_stream_write (out, o, n, 1) < 0)
			return 0;
		l -= m;
		if (l)
			memcpy (i, i + m, l);
	}
	if (l)
	{
		n = _decode (1, i, l, o, sizeof (o), &m);
		if (n)
		{
			if (qu_stream_write (out, i, n, 1) < 0)
				return 0;
		}
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Quopri_init (qu_module m)
{
	qu_module_add_function (m, "encode_string" , Quopri_encode_string ,  1);
	qu_module_add_function (m, "encode_buffer" , Quopri_encode_buffer , -1);
	qu_module_add_function (m, "encode_stream" , Quopri_encode_stream ,  2);

	qu_module_add_function (m, "decode_string" , Quopri_decode_string ,  1);
	qu_module_add_function (m, "decode_buffer" , Quopri_decode_buffer , -1);
	qu_module_add_function (m, "decode_stream" , Quopri_decode_stream ,  2);

	return 1;
}
