/* Env.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Env
	This module provides an interface to the system environment.

	Refer to C{man environ} for details.

	@xmp
		use Env

		s = Env.get ('USER')
		println (s)

	@out
		foo
*/



/*) sub clear -> Null
	Clear the environment of all name-value pairs.

	See C{man clearenv} for more information.

	@exc
	.ESystem

	@xmp
		use Env
		Env.clear
*/
static qu_null Env_clear ()
{
	QU_REQUIRE (QuESystem, clearenv () == 0);
	return QuNil;
}



/*) sub env -> Array
	Get all environment variables into an Array.

	The Array will have items in the form C{name=value}.

	See C{man environ} for details.

	@noexc

	@xmp
		use Env

		for s in Env.environ
			println (s)
		;;

	@out
		PWD=/opt/qu/demo
		HOSTNAME=localhost.localdomain
		USER=foo
*/
static qu_array Env_environ ()
{
	extern char **environ;

	char     **s = environ;
	qu_array   a = qu_array_new (16);

	while (*s)
	{
		qu_array_push_quick (a, qu_string_newz (*s));
		++s;
	}
	return a;
}



/*) sub get (name: String) -> String
	Return the variable value specified by v{name}.

	See C{man getenv} for more information.

	@exc
	.ENone variable not found
	.EArgValue

	@xmp
		use Env
		s = Env.get ('USER')
		println (s)

	@out
		foo
*/
static qu_any Env_get (qu_string name)
{
	char *s;

	QU_ARG_TYPE (name, qu_string);
	s = getenv (name->str);
	QU_REQUIRE (QuENone, s);
	return qu_string_newz (s);
}



/*) sub set (name: String, value: String, overwrite = true) -> Null
	Change or add an environment variable.

	See C{man setenv} for more information.

	This function adds the variable name to the environment with the specified
	v{value}, if v{name} does not already exist.  If v{name} does exist in
	the environment, then its value is changed to v{value} if v{overwrite}
	is true; if overwrite is false then the value is not changed.

	@exc
	.EArgValue
	.ESystem

	@xmp
		use Env
		Env.set ('FOO', 'foo', false)
*/
static qu_null Env_set (int argc, qu_any *argv)
{
	qu_string name, value;
	int       overwrite = 1;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, name, qu_string);
	QU_ARG_VTYPE  (1, value, qu_string);
	if (argc > 2)
		QU_ARG_vbool (2, overwrite);
	QU_REQUIRE (QuESystem, setenv (name->str, value->str, overwrite) == 0);
	return QuNil;
}



/*) sub unset (name: String) -> Null
	Delete the variable name from the environment.

	If the variable does not exist then nothing happens.

	See C{man unsetenv} for more information.

	@noexc

	@xmp
		use Env
		Env.unset ('FOO')
*/
static qu_null Env_unset (qu_string name)
{
	QU_ARG_TYPE (name, qu_string);
	unsetenv (name->str);
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Env_init (qu_module m)
{
	qu_module_add_function (m, "environ", Env_environ,  0);
	qu_module_add_function (m, "get"    , Env_get    ,  1);
	qu_module_add_function (m, "set"    , Env_set    , -1);
	qu_module_add_function (m, "clear"  , Env_clear  ,  0);
	qu_module_add_function (m, "unset"  , Env_unset  ,  1);

	return 1;
}
