/* Gc.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/*) module Gc
	This module provides access to the Garbage Collector (GC).
*/


/*) sub __critical_begin
	Start critical section, disable GC.

	@warn
	You must eventually call #{__critical_end}.

	@exc
	.ESecurity
*/
static qu_null Gc___critical_begin ()
{
	__qu_unsafe__;
	__qu_gc_critical_begin;
	return QuNil;
}



/*) sub __critical_end
	End critical section, probably reenable GC.

	You must eventually call this method after #{__critical_begin}.

	Note that this does not necessarily means that GC is enabled because
	this part of code may be nested within another critical section.

	@exc
	.ESecurity
*/
static qu_null Gc___critical_end ()
{
	__qu_unsafe__;
	__qu_gc_critical_end;
	return QuNil;
}



/*) sub limit -> Integer
	Get the current GC limit in KB.

	When the size of memory reaches this limit then GC automatically starts,
	unless of course, it is #{disable}d.

	@noexc
*/
static qu_integer Gc_limit ()
{
	return qu_integer_newl (__QuGcLimit () / 1024);
}



/*) sub memory -> Integer
	Get the total size of memory currently in use in KB.

	Note that this is not the size of memory actually allocated by the
	interpreter, which is generally bigger.

	@noexc
*/
static qu_integer Gc_memory ()
{
	return qu_integer_newul (__QuGcMemAllocated / 1024);
}



/*) sub reclaim -> Null
	Start collecting garbage now.

	@exc
	.ESecurity
*/
static qu_null Gc_reclaim ()
{
	__qu_unsafe__;
	__QuGcNeedReclaim = 1;
	qu_gc_reclaim ();
	return QuNil;
}



/*) sub regulars -> Integer
	Return the total number of non-root non-static objects currently allocated.

	Note that this probably always return 0 since the GC may not record the
	number of objects for performance reasons.

	@noexc
*/
static qu_integer Gc_regulars ()
{
	return qu_integer_newl (qu_gc_num_regulars ());
}



/*) sub roots -> Integer
	Return the total number of root objects currently allocated.

	Note that this probably always return 0 since the GC may not record the
	number of objects for performance reasons.

	@noexc
*/
static qu_integer Gc_roots ()
{
	return qu_integer_newl (qu_gc_num_roots ());
}



/*) sub stat -> Null
	Print memory statistic on %{sys$File.stderr}.

	@noexc
*/
static qu_null Gc_stat ()
{
	qu_gc_stat ();
	return QuNil;
}



/*) sub statics -> Integer
	Return the total number of static objects currently allocated.

	Note that this probably always return 0 since the GC may not record the
	number of objects for performance reasons.

	@noexc
*/
static qu_integer Gc_statics ()
{
	return qu_integer_newl (qu_gc_num_statics ());
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Gc_init (qu_module m)
{

	qu_module_add_function (m, "__critical_begin", Gc___critical_begin, 0);
	qu_module_add_function (m, "__critical_end"  , Gc___critical_end  , 0);
	qu_module_add_function (m, "limit"           , Gc_limit           , 0);
	qu_module_add_function (m, "memory"          , Gc_memory          , 0);
	qu_module_add_function (m, "regulars"        , Gc_regulars        , 0);
	qu_module_add_function (m, "reclaim"         , Gc_reclaim         , 0);
	qu_module_add_function (m, "roots"           , Gc_roots           , 0);
	qu_module_add_function (m, "stat"            , Gc_stat            , 0);
	qu_module_add_function (m, "statics"         , Gc_statics         , 0);
	return 1;
}
