/* Gd.c - Copyright (c) Marc Krisnanto */

#include "qu.h"
#include "Gd.h"


/*) module Gd
	This module provides basic routines to manipulate simple images.
	It uses the gd library.

	@xmp
		use Gd

		im = Gd.Image (10, 20)
		background = im.color (0, 0, 0)
		white = im.color (255, 255, 255)
		im.line (im, 0, 0, 63, 63, white)
		im.png  ("/tmp/gdtest.png")
		im.jpeg ("/tmp/gdtest.jpeg")

		s = im.data ('jpeg', 50)
		File.dump ("/tmp/gdtest.jpeg", s)

	@warn
	You must have the gd library installed on your system.

	@p
	Note that you don't need to recompile Qu nor this module.
	Once you have installed the gd library then it will be automatically
	available.
*/

/*----------------------------------------------------------------------------
	stuffs
----------------------------------------------------------------------------*/

#define QU_GD_TYPE_UNKNOWN -1
#define QU_GD_TYPE_JPEG     0
#define QU_GD_TYPE_PNG      1
#define QU_GD_TYPE_GIF      2
#define QU_GD_TYPE_XPM      3
#define QU_GD_TYPE_XBM      4
#define QU_GD_TYPE_GD       5
#define QU_GD_TYPE_GD2      6

static int qu_gd_type_from_string (qu_string s)
{
	char  e [12];
	char *z = s->str;
	char *r = strrchr (z, QU_PATH_SEPCHR);
	char *i;

	if (r == 0)
		r = z;
	else
	{
		r = strchr (r, '.');
		if (r == 0)
			r = z;
		else
			++r;
	}

	if (strlen (r) > 4)
		return QU_GD_TYPE_UNKNOWN;

	i = e;
	while (*r)
	{
		*i++ = qu_ascii_tolower (*r);
		++r;
	}
	*i = '\0';

	if (strcmp (e, "jpeg") == 0 || strcmp (e, "jpg") == 0) return QU_GD_TYPE_JPEG;
	if (strcmp (e, "png" ) == 0) return QU_GD_TYPE_PNG;
	if (strcmp (e, "gif" ) == 0) return QU_GD_TYPE_GIF;
	if (strcmp (e, "xpm" ) == 0) return QU_GD_TYPE_XPM;
	if (strcmp (e, "xbm" ) == 0) return QU_GD_TYPE_XBM;
	if (strcmp (e, "gd"  ) == 0) return QU_GD_TYPE_GD;
	if (strcmp (e, "gd2" ) == 0) return QU_GD_TYPE_GD2;
	return QU_GD_TYPE_UNKNOWN;
}



/*----------------------------------------------------------------------------
	class Image
----------------------------------------------------------------------------*/

/*) class Image
	Use this class to manipulate images.
*/
static qu_class QuImageClass;

typedef struct qu_image_s qu_image_s;
struct qu_image_s
{
	QU_OBJECT_HEAD;
	gdImagePtr im;
};
typedef struct qu_image_s *qu_image;

#define qu_image_is(x) qu_object_instof (x, QuImageClass)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	standard class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub class __inst (width: Integer, height: Integer, truecolor = false) -> Image __final
	Create a new Image with a specified dimension.

	If v{truecolor} is false then the image will have at most 256 colors.

	@exc
	.ESecurity
	.EArgValue
	.EMemory
*/
static qu_image qu_image__inst (qu_class self, int argc, qu_any *argv)
{
	int        w, h, tc = 0;
	qu_image   o;
	gdImagePtr im;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_vint (0, w);
	QU_ARG_vint (1, h);
	QU_ARG_CHECK (0 < w && w < QU_INT_MAX && 0 < h && h < QU_INT_MAX);

	if (w > 64 || h > 64)
	{
		__qu_unsafe__;
	}

	if (argc > 2)
		QU_ARG_vbool (2, tc);

	if (tc && gdImageCreateTrueColor)
		im = gdImageCreateTrueColor (w, h);
	else
		im = gdImageCreate (w, h);

	QU_REQUIRE (QuEMemory, im);
	o = qu_gc_new (self);
	o->im = im;
	return o;
}



/*) sub class new (width: Integer, height: Integer, truecolor = false) -> Image
	This method simply calls #{__inst} then f{__init} if one is defined,
	passing all arguments.
*/
static qu_image qu_image_new (qu_class self, int argc, qu_any *argv)
{
	qu_image o = qu_image__inst (self, argc, argv);

	QU_ENSURE (o);
	return qu_class_init_new (self, o, argc, argv);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	specific class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub class open (path: String, type = nil) -> Image
	Create a new Image from a file.

	v{type} must be one of:
	@ul
	.C{'jpeg'}
	.C{'png'}
	.C{'gif'}
	.C{'xpm'}
	.C{'xbm'}
	.C{'gd'}
	.C{'gd2'}

	@p
	or nil to get it from the file extension.

	@exc
	.ESecurity
	.EArgValue
	.EImplement
	.EMemory
*/
static qu_image qu_image_open (qu_class self, int argc, qu_any *argv)
{
	qu_string    path;
	int          type;
	FILE        *f;
	gdImagePtr   im;
	qu_image     o;
	gdImagePtr (*func) (FILE *) = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
	{
		qu_string t;
		QU_ARG_VTYPE (1, t, qu_string);
		type = qu_gd_type_from_string (t);
	}
	else
	{
		type = qu_gd_type_from_string (path);
	}

	switch (type)
	{
		case QU_GD_TYPE_JPEG: func = gdImageCreateFromJpeg; break;
		case QU_GD_TYPE_PNG:  func = gdImageCreateFromPng;  break;
		case QU_GD_TYPE_GIF:  func = gdImageCreateFromGif;  break;
		case QU_GD_TYPE_XPM:  func = gdImageCreateFromXpm;  break;
		case QU_GD_TYPE_XBM:  func = gdImageCreateFromXbm;  break;
		case QU_GD_TYPE_GD:   func = gdImageCreateFromGd;   break;
		case QU_GD_TYPE_GD2:  func = gdImageCreateFromGd2;  break;
	}

	if (func == 0)
		QU_THROWZ (QuEImplement, "unknown type");

	f = fopen (path->str, "r");
	if (f == 0)
		QU_THROWERRNO;

	im = func (f);
	fclose (f);
	if (im == 0)
		QU_THROWZ (QuESystem, "image corrupted or too big");
	o = qu_gc_new (self);
	o->im = im;
	return o;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	low-level instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static void qu_image__destroy (qu_image self)
{
	gdImageDestroy (self->im);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	specific instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub save (path, type = '', quality = -1) -> self
	Save image to file.

	v{quality} must be -1 to use the default, or a number between C{0..95}.
	It is only applicable for C{'jpeg'}.

	v{type} must be one of:
	@ul
	.C{'jpeg'}
	.C{'png'}
	.C{'gif'}
	.C{'xpm'}
	.C{'xbm'}
	.C{'gd'}
	.C{'gd2'}

	@p
	or nil to get it from the file extension.

	@xmp
	.EArgValue
	.EImplement
*/
static qu_image qu_image_save (qu_image self, int argc, qu_any *argv)
{
	qu_string  path;
	int        type, q = -1;
	FILE      *f;
	void     (*func) (gdImagePtr, FILE *) = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
	{
		qu_string t;
		QU_ARG_VTYPE (1, t, qu_string);
		if (t->len)
			type = qu_gd_type_from_string (t);
		else
			type = qu_gd_type_from_string (path);
		if (argc > 2)
			QU_ARG_vint (2, q);
	}
	else
	{
		type = qu_gd_type_from_string (path);
	}

	switch (type)
	{
		case QU_GD_TYPE_JPEG: func = (void *) gdImageJpeg; break;
		case QU_GD_TYPE_PNG:  func = gdImagePng;  break;
		case QU_GD_TYPE_GIF:  func = gdImageGif;  break;
		case QU_GD_TYPE_XPM:  func = gdImageXpm;  break;
		case QU_GD_TYPE_XBM:  func = gdImageXbm;  break;
		case QU_GD_TYPE_GD:   func = gdImageGd;   break;
		case QU_GD_TYPE_GD2:  func = gdImageGd2;  break;
	}

	if (func == 0)
		QU_THROWZ (QuEImplement, "unknown type");

	f = fopen (path->str, "wb");
	if (f == 0)
		QU_THROWERRNO;

	if (type == QU_GD_TYPE_JPEG)
		gdImageJpeg (self->im, f, q);
	else
		func (self->im, f);

	if (ferror (f))
	{
		fclose (f);
		QU_THROWZ (QuESystem, "save image failed");
	}

	fclose (f);
	return self;
}


/*) sub data (type, quality = -1) -> String
	Get the image data as a String.

	v{quality} must be -1 to use the default, or a number between C{0..95}.
	It is only applicable for C{'jpeg'}.

	v{type} must be one of:
	@ul
	.C{'jpeg'}
	.C{'png'}
	.C{'gif'}
	.C{'xpm'}
	.C{'xbm'}
	.C{'gd'}
	.C{'gd2'}

	@exc
	.EArgValue
	.EImplement
*/
static qu_string qu_image_data (qu_image self, int argc, qu_any *argv)
{
	qu_string  spec;
	int        type, q = -1, size;
	void    *(*func) (gdImagePtr, int *) = 0;
	void      *data;
	qu_string  r;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, spec, qu_string);
	if (argc > 1)
		QU_ARG_vint (2, q);

	type = qu_gd_type_from_string (spec);
	switch (type)
	{
		case QU_GD_TYPE_JPEG: func = (void *) gdImageJpegPtr; break;
		case QU_GD_TYPE_PNG:  func = gdImagePngPtr;  break;
		case QU_GD_TYPE_GIF:  func = gdImageGifPtr;  break;
		case QU_GD_TYPE_XPM:  func = gdImageXpmPtr;  break;
		case QU_GD_TYPE_XBM:  func = gdImageXbmPtr;  break;
		case QU_GD_TYPE_GD:   func = gdImageGdPtr;   break;
		case QU_GD_TYPE_GD2:  func = gdImageGd2Ptr;  break;
	}

	if (func == 0)
		QU_THROWZ (QuEImplement, "unknown type");

	if (type == QU_GD_TYPE_JPEG)
		data = gdImageJpegPtr (self->im, &size, q);
	else
		data = func (self->im, &size);

	QU_REQUIRE (QuEMemory, data);
	r = qu_string_new (data, size);
	gdFree (data);
	return r;
}



/*) sub width -> Integer
	Get the width of the image.

	@noexc
*/
static qu_integer qu_image_width (qu_image self)
{
	return qu_integer_newi (self->im->sx);
}



/*) sub height -> Integer
	Get the height of the image.

	@noexc
*/
static qu_integer qu_image_height (qu_image self)
{
	return qu_integer_newi (self->im->sy);
}



/*) sub clip (x1: Integer, y1: Integer, x2: Integer y2: Integer) -> self
	Set clipping area.

	@exc
	.EArgValue
*/
static qu_image qu_image_clip (qu_image self, int argc, qu_any *argv)
{
	int x1, y1, x2, y2;

	QU_ARG_VEXACT (4);
	QU_ARG_vint (0, x1);
	QU_ARG_vint (1, y1);
	QU_ARG_vint (2, x2);
	QU_ARG_vint (3, y2);
	QU_REQUIRE (QuEImplement, gdImageSetClip);
	gdImageSetClip (self->im, x1, y1, x2, y2);
	return self;
}



/*) sub color (r: Integer, g: Integer, b: Integer) -> Integer
	Create a color.

	Return the index or value of the color.

	@exc
	.EArgValue
*/
static qu_integer qu_image_color (qu_image self, qu_any rr, qu_any gg, qu_any bb)
{
	int r, g, b, c;

	QU_ARG_int (rr, r);
	QU_ARG_int (gg, g);
	QU_ARG_int (bb, b);
	QU_REQUIRE (QuEImplement, gdImageColorAllocate);
	c = gdImageColorAllocate (self->im, r, g, b);
	return qu_integer_newi (c);
}



/*) sub pixel (*arg) -> self
	@def
	sub pixel (x: Integer, y: Integer, color = 0) -> self
	sub pixel (x: Integer, y: Integer) -> Integer
	@p
	Get or set pixel using the specified color index.

	@exc
	.EArgValue
*/
static qu_any qu_image_pixel (qu_image self, int argc, qu_any *argv)
{
	int x, y, c = 0;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_vint (0, x);
	QU_ARG_vint (1, y);
	if (argc > 2)
	{
		QU_ARG_vint (2, c);
		QU_REQUIRE (QuEImplement, gdImageSetPixel);
		gdImageSetPixel (self->im, x, y, c);
		return self;
	}
	else
	{
		QU_REQUIRE (QuEImplement, gdImageGetPixel);
		c = gdImageGetPixel (self->im, x, y);
		return qu_integer_newi (c);
	}
}



/*) sub line (x1: Integer, y1: Integer, x2: Integer y2: Integer, color = 0) -> self
	Draw a line using the specified color index.

	@exc
	.EArgValue
*/
static qu_image qu_image_line (qu_image self, int argc, qu_any *argv)
{
	int x1, y1, x2, y2, c = 0;

	QU_ARG_VCOUNT (4, 5);
	QU_ARG_vint (0, x1);
	QU_ARG_vint (1, y1);
	QU_ARG_vint (2, x2);
	QU_ARG_vint (3, y2);
	if (argc > 4)
		QU_ARG_vint (4, c);
	QU_REQUIRE (QuEImplement, gdImageLine);
	gdImageLine (self->im, x1, y1, x2, y2, c);
	return self;
}



/*) sub arc (x: Integer, y: Integer, w: Integer h: Integer, s: Integer, e: Integer, color = 0) -> self
	Draw arc using the specified color index.

	@exc
	.EArgValue
*/
static qu_image qu_image_arc (qu_image self, int argc, qu_any *argv)
{
	int x, y, h, w, s, e, c = 0;

	QU_ARG_VCOUNT (6, 7);
	QU_ARG_vint (0, x);
	QU_ARG_vint (1, y);
	QU_ARG_vint (2, w);
	QU_ARG_vint (3, h);
	QU_ARG_vint (4, s);
	QU_ARG_vint (5, e);
	if (argc > 6)
		QU_ARG_vint (6, c);
	QU_REQUIRE (QuEImplement, gdImageArc);
	gdImageArc (self->im, x, y, w, h, s, e, c);
	return self;
}



/*) sub fill (x: Integer, y: Integer, color = 0) -> self
	Fill flood using the specified color index.

	@exc
	.EArgValue
*/
static qu_image qu_image_fill (qu_image self, int argc, qu_any *argv)
{
	int x, y, c = 0;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_vint (0, x);
	QU_ARG_vint (1, y);
	if (argc > 2)
		QU_ARG_vint (2, c);
	QU_REQUIRE (QuEImplement, gdImageFill);
	gdImageFill (self->im, x, y,c);
	return self;
}



/*) sub fillborder (x: Integer, y: Integer, border: Integer, color = 0) -> self
	Flood a portion of the image with the specified color, beginning at the
	specified point and stopping at the specified border color.

	@exc
	.EArgValue
*/
static qu_image qu_image_fillborder (qu_image self, int argc, qu_any *argv)
{
	int x, y, b, c = 0;

	QU_ARG_VCOUNT (3, 4);
	QU_ARG_vint (0, x);
	QU_ARG_vint (1, y);
	QU_ARG_vint (2, y);
	if (argc > 3)
		QU_ARG_vint (3, c);
	QU_REQUIRE (QuEImplement, gdImageFillToBorder);
	gdImageFillToBorder (self->im, x, y, b, c);
	return self;
}



/*) sub string (s, size: String, x: Integer, y: Integer, color = 0) -> self
	Draw a string.

	v{size} must be one of:
	@ul
	.C{'t'} tiny
	.C{'s'} small
	.C{'m'} medium
	.C{'l'} large
	.C{'g'} giant

	@exc
	.EArgValue
*/
static qu_image qu_image_string (qu_image self, int argc, qu_any *argv)
{
	int       x, y, c = 0;
	qu_string s, z;
	gdFontPtr f = 0;

	QU_ARG_VCOUNT (4, 5);
	QU_ARG_VTYPE (0, s, qu_stringy);
	QU_ARG_VTYPE (1, z, qu_stringy);
	QU_ARG_vint  (2, x);
	QU_ARG_vint  (3, y);
	if (argc > 4)
		QU_ARG_vint (4, c);
	QU_REQUIRE (QuEImplement, gdImageString);

	switch (z->str[0])
	{
		default:             f = gdFontTiny;       break;
		case 's': case 'S':  f = gdFontSmall;      break;
		case 'm': case 'M':  f = gdFontMediumBold; break;
		case 'l': case 'L':  f = gdFontLarge;      break;
		case 'g': case 'G':  f = gdFontGiant;      break;
	}
	QU_REQUIRE (QuEImplement, f);
	gdImageString (self->im, f, x, y, s->str, c);
	return self;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/
static void *QuGdHandle;

void qu_gd_atexit ()
{
	if (QuGdHandle)
		qu_dl_close (QuGdHandle);
}


int qu_Gd_init (qu_module m)
{

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		gd
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuGdHandle = qu_dl_open_ex ("gd", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (QuGdHandle);

#define f_(name) name = qu_dl_sym (QuGdHandle, #name)

	f_(gdImageDestroy);
	f_(gdFree);
	f_(gdImageCreate);

	if (gdImageDestroy == 0
		|| gdFree == 0
		|| gdImageCreate == 0)
	{
		qu_dl_close (QuGdHandle);
		QU_THROWZ (QuESystem, "incompatible gd library");
	}

	f_(gdImageCreateTrueColor);
	f_(gdImageCreateFromJpeg);
	f_(gdImageCreateFromJPegPtr);
	f_(gdImageCreateFromPng);
	f_(gdImageCreateFromPngPtr);
	f_(gdImageCreateFromGif);
	f_(gdImageCreateFromGifPtr);
	f_(gdImageCreateFromXpm);
	f_(gdImageCreateFromXpmPtr);
	f_(gdImageCreateFromXbm);
	f_(gdImageCreateFromXbmPtr);
	f_(gdImageCreateFromGd);
	f_(gdImageCreateFromGdPtr);
	f_(gdImageCreateFromGd2);
	f_(gdImageCreateFromGd2Ptr);
	f_(gdImageJpeg);
	f_(gdImageJpegPtr);
	f_(gdImagePng);
	f_(gdImagePngPtr);
	f_(gdImageGif);
	f_(gdImageGifPtr);
	f_(gdImageXpm);
	f_(gdImageXpmPtr);
	f_(gdImageXbm);
	f_(gdImageXbmPtr);
	f_(gdImageGd);
	f_(gdImageGdPtr);
	f_(gdImageGd2);
	f_(gdImageGd2Ptr);
	f_(gdImageSetClip);
	f_(gdImageColorAllocate);
	f_(gdImageSetPixel);
	f_(gdImageGetPixel);
	f_(gdImageLine);
	f_(gdImageArc);
	f_(gdImageFill);
	f_(gdImageFillToBorder);
	f_(gdImageChar);
	f_(gdImageString);
	f_(gdFontTiny);
	f_(gdFontSmall);
	f_(gdFontMediumBold);
	f_(gdFontLarge);
	f_(gdFontGiant);

	atexit (qu_gd_atexit);

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Image
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuImageClass = qu_class_newz (m, "Image", qu_image_s, qu_image__destroy, 0);

	/* standard class methods */
	qu_class_method_class  (QuImageClass, QU_NAME__inst, qu_image__inst, -1);
	qu_class_method_class  (QuImageClass, QU_NAME_new  , qu_image_new  , -1);

	/* specific class methods */
	qu_class_method_classz (QuImageClass, "open", qu_image_open, -1);

	/* specific instance methods */
	qu_class_method_instz  (QuImageClass, "save"      , qu_image_save      , -1);
	qu_class_method_instz  (QuImageClass, "data"      , qu_image_data      , -1);
	qu_class_method_instz  (QuImageClass, "width"     , qu_image_width     ,  0);
	qu_class_method_instz  (QuImageClass, "height"    , qu_image_height    ,  0);
	qu_class_method_instz  (QuImageClass, "clip"      , qu_image_clip      , -1);
	qu_class_method_instz  (QuImageClass, "color"     , qu_image_color     ,  3);
	qu_class_method_instz  (QuImageClass, "pixel"     , qu_image_pixel     , -1);
	qu_class_method_instz  (QuImageClass, "line"      , qu_image_line      , -1);
	qu_class_method_instz  (QuImageClass, "arc"       , qu_image_arc       , -1);
	qu_class_method_instz  (QuImageClass, "fill"      , qu_image_fill      , -1);
	qu_class_method_instz  (QuImageClass, "fillborder", qu_image_fillborder, -1);
	qu_class_method_instz  (QuImageClass, "string"    , qu_image_string    , -1);

	return 1;
}
