/* Gdbm.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Gdbm
	This module provides an interface to the gdbm library.

	@xmp
		use Gdbm

		path = "/tmp/qutest.gdbm"
		g = Gdbm (path: path, readwrite:Gdbm.WRCREATE)

		# set new keys
		g['foo'] = 123
		g['bar'] = 456

		# retrieve
		println (g['foo'])
		println (g['bar'])

		# iterate
		for k in g
			println (k ' = ' g[k])
		;;

		# dont forget to close
		g.close

		File.delete (path)

	@p
	See C{man gdbm} for more information.
*/


/*----------------------------------------------------------------------------
	stuffs we need from gdbm
---------------------------------------------------------------------------*/
#define GDBM_READER  0
#define GDBM_WRITER  1
#define GDBM_WRCREAT 2
#define GDBM_NEWDB   3
#define GDBM_FAST    0x10
#define GDBM_SYNC    0x20
#define GDBM_NOLOCK  0x40

#define GDBM_INSERT  0
#define GDBM_REPLACE 1

#define GDBM_CACHESIZE    1
#define GDBM_FASTMODE     2
#define GDBM_SYNCMODE     3
#define GDBM_CENTFREE     4
#define GDBM_COALESCEBLKS 5

typedef struct {
	char *dptr;
	int   dsize;
} datum;


typedef struct {int dummy[10];} *GDBM_FILE;

static char *gdbm_version;
static GDBM_FILE (*gdbm_open)       (char *, int, int, int, void (*)());
static void      (*gdbm_close)      (GDBM_FILE);
static int       (*gdbm_store)      (GDBM_FILE, datum, datum, int);
static datum     (*gdbm_fetch)      (GDBM_FILE, datum);
static int       (*gdbm_delete)     (GDBM_FILE, datum);
static datum     (*gdbm_firstkey)   (GDBM_FILE);
static datum     (*gdbm_nextkey)    (GDBM_FILE, datum);
static int       (*gdbm_reorganize) (GDBM_FILE);
static void      (*gdbm_sync)       (GDBM_FILE);
static int       (*gdbm_exists)     (GDBM_FILE, datum);
static int       (*gdbm_setopt)     (GDBM_FILE, int, int *, int);
static int       (*gdbm_fdesc)      (GDBM_FILE);

#define	GDBM_NO_ERROR		0
#define	GDBM_MALLOC_ERROR	1
#define	GDBM_BLOCK_SIZE_ERROR	2
#define	GDBM_FILE_OPEN_ERROR	3
#define	GDBM_FILE_WRITE_ERROR	4
#define	GDBM_FILE_SEEK_ERROR	5
#define	GDBM_FILE_READ_ERROR	6
#define	GDBM_BAD_MAGIC_NUMBER	7
#define	GDBM_EMPTY_DATABASE	8
#define	GDBM_CANT_BE_READER	9
#define	GDBM_CANT_BE_WRITER	10
#define	GDBM_READER_CANT_DELETE	11
#define	GDBM_READER_CANT_STORE	12
#define	GDBM_READER_CANT_REORGANIZE	13
#define	GDBM_UNKNOWN_UPDATE	14
#define	GDBM_ITEM_NOT_FOUND	15
#define	GDBM_REORGANIZE_FAILED	16
#define	GDBM_CANNOT_REPLACE	17
#define	GDBM_ILLEGAL_DATA	18
#define	GDBM_OPT_ALREADY_SET	19
#define	GDBM_OPT_ILLEGAL	29

typedef int gdbm_error;
static gdbm_error *gdbm_errno;
static char *(*gdbm_strerror) (gdbm_error);

#define gdbm__errno *gdbm_errno


/*----------------------------------------------------------------------------
	class Gdbm
----------------------------------------------------------------------------*/

/*) class Gdbm
	Create an instance of this class to access a gdbm file.
*/
qu_class QuGdbmClass;

typedef struct qu_gdbm_s qu_gdbm_s;
struct qu_gdbm_s
{
	QU_OBJECT_HEAD;
	GDBM_FILE dbf;
	long      len;
};

typedef struct qu_gdbm_s *qu_gdbm;


#define QU_GDBM_FLAG_READER QU_FLAG_0

#define qu_gdbm_isreader(self)  qu_flag_test (self, QU_GDBM_FLAG_READER)
#define qu_gdbm_setreader(self) qu_flag_set  (self, QU_GDBM_FLAG_READER)

#define QU_GDBM_ENSURE_OPENED(self) \
	QU_REQUIREZ (QuECall, (self)->dbf, "Gdbm already closed")

#define QU_GDBM_CLEAR_ERROR(self) \
QU_BEGIN \
	qu_thread tHr = qu_thread_self (); \
	assert (tHr); \
	qu_flag_unset (tHr, QU_THREAD_FLAG_EXTERROR); \
QU_END


#define QU_GDBM_ENSURE_NONFATAL(self) \
QU_BEGIN \
	if (qu_gdbm__isfatal (self)) \
		QU_RETHROW; \
QU_END


static int qu_gdbm__isfatal (qu_gdbm self)
{
	qu_thread tHr = qu_thread_self ();
	assert (tHr);
	return qu_flag_test (tHr, QU_THREAD_FLAG_EXTERROR);
}


static void qu_gdbm__fatal (const char *msg)
{
	qu_thread t = qu_thread_self ();

	assert (t);
	qu_thread__throw (QuEDatabaseClass, qu_string_newz (msg), __FUNCTION__);
	qu_flag_set (t, QU_THREAD_FLAG_EXTERROR);
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (path: String, blocksize = 512, readwrite = READER, perm = 0666) -> Gdbm
	Create or open a gdbm file.

	@exc
	.EArgValue
	.ESystem
*/
static qu_gdbm qu_gdbm_new (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	int       block_size = 512;
	int       read_write = GDBM_READER;
	int       mode = 0666;
	GDBM_FILE f;
	qu_gdbm   g;

	static char *argx[] =
	{
		"s", "path",
		":",
		"i", "blocksize",
		"i", "readwrite",
		"i", "perm",
		NULL
	};

	__qu_unsafe__;

	QU_ARG_XPARSE (&path, &block_size, &read_write, &mode);
	f = gdbm_open (path->str, block_size, read_write, mode, qu_gdbm__fatal);
	if (f == 0)
		QU_THROWZ (QuESystem, gdbm_strerror (gdbm__errno));
	g = qu_gc_new (QuGdbmClass);
	g->dbf = f;
	g->len = -1;
	if (read_write & GDBM_READER)
		qu_gdbm_setreader (g);
	return g;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu_gdbm__destroy (qu_gdbm self)
{
	if (self->dbf)
		gdbm_close (self->dbf);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __del (key) -> self
	Delete a key.

	@exc
	.EModify    opened for read-only
	.ECall      already closed
	.EKey       key not found
	.EArgValue
*/
/*) sub delete (key)
	This is an alias of #{__del}.
*/
static qu_gdbm qu_gdbm__del (qu_gdbm self, qu_any akey)
{
	volatile qu_string key;
	datum k;
	int   r;

	QU_GDBM_ENSURE_OPENED (self);
	key = qu_sys_string (akey);
	QU_REQUIRE (QuEKey, key->len && key->len < INT_MAX);
	k.dptr = key->str;
	k.dsize = key->len;
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	r = gdbm_delete (self->dbf, k);
	__qu_blocking_end;
	if (r == -1)
	{
		if (qu_gdbm_isreader (self))
			QU_THROWZ (QuEModify, "reader cant write to database");
		QU_THROW (QuEKey);
	}
	else if (r)
	{
		QU_RETHROW;
	}
	self->len = -1;
	return self;
}



/*) sub __each -> Iterator
	Return a -{Iterator} object to iterate through the keys.

	@exc
	.ECall already closed
*/
static qu_any qu_gdbm__next (qu_gdbm self, qu_iterator i)
{
	qu_string key;
	datum     k;

	QU_GDBM_ENSURE_OPENED (self);
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	if (i->pos == QuStringEMPTY)
	{
		k = gdbm_firstkey (self->dbf);
	}
	else
	{
		datum p;

		key = i->pos;
		assert (qu_string_is (key));
		p.dptr = key->str;
		p.dsize = key->len;
		k = gdbm_nextkey (self->dbf, p);
	}
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	QU_ITERATOR_STOP (k.dptr);
	key = qu_string_new (k.dptr, k.dsize);
	free (k.dptr);
	i->pos = key;
	return key;
}



static qu_iterator qu_gdbm__each (qu_gdbm self)
{
	register qu_iterator i;

	QU_GDBM_ENSURE_OPENED (self);
	QU_ITERATOR_NEW_EXTERN (i, self, qu_gdbm__next);
	qu_iterator_set_yield_keys (i);
	i->pos = QuStringEMPTY;
	return i;
}



/*) sub __getitem (key) -> String
	Get the value associated with a key.

	@exc
	.ECall      already closed
	.EKey       key not found
	.EArgValue
*/
static qu_any qu_gdbm__getitem (qu_gdbm self, qu_any akey)
{
	volatile qu_string key, val;
	datum k, v;

	QU_GDBM_ENSURE_OPENED (self);
	key = qu_sys_string (akey);
	QU_REQUIRE (QuEKey, key->len && key->len < INT_MAX);
	k.dptr = key->str;
	k.dsize = key->len;
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	v = gdbm_fetch (self->dbf, k);
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	QU_REQUIRE (QuEKey, v.dptr);
	val = qu_string_new (v.dptr, v.dsize);
	free (v.dptr);
	return val;
}



/*) sub __getslice (key, defval) -> Any
	This does the same as #{__getitem} except that it does not throw EKey
	if the key is not found. It returns v{defval} instead.

	@exc
	.ECall already closed
*/
static qu_any qu_gdbm__getslice (qu_gdbm self, qu_any akey, qu_any defval)
{
	volatile qu_string key, val;
	datum k, v;

	QU_GDBM_ENSURE_OPENED (self);
	key = qu_sys_string (akey);
	if (key->len == 0 || key->len >= INT_MAX)
		return defval;
	k.dptr = key->str;
	k.dsize = key->len;
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	v = gdbm_fetch (self->dbf, k);
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	if (v.dptr == 0)
		return defval;
	val = qu_string_new (v.dptr, v.dsize);
	free (v.dptr);
	return val;
}



/*) sub __in (key) -> Boolean
	Whether or not the key is in the databse.

	v{key} may be anything. If it is not a String then it is converted into
	a String by calling its __string method.

	If the database is already closed then this method returns k{false}.

	@noexc
*/
/*) sub exists (key) -> Boolean
	This is an alias of #{__in}.
*/
/*) sub haskey (key) -> Boolean
	This is an alias of #{__in}.
*/
static qu_any qu_gdbm__in (qu_gdbm self, qu_any akey)
{
	volatile qu_string key;
	datum k;
	int   r;

	if (self->dbf == 0)
		return QuFalse;
	key = qu_sys_string (akey);
	if (key->len == 0 || key->len >= INT_MAX)
		return QuFalse;
	k.dptr = key->str;
	k.dsize = key->len;
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	r = gdbm_exists (self->dbf, k);
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	return r ? QuTrue : QuFalse;
}



/*) sub __len -> Integer
	Return the current number of records.

	@exc
	.ECall already closed
	.EDatabase
*/
/*) sub len
	This is an alias of #{__len}.
*/
static qu_integer qu_gdbm__len (qu_gdbm self)
{
	if (self->len < 0)
	{
		datum k, x;
		long  n = 0;

		QU_GDBM_ENSURE_OPENED (self);
		QU_GDBM_CLEAR_ERROR (self);
		__qu_blocking_begin;
		x = gdbm_firstkey (self->dbf);
		if (x.dptr)
		{
			++n;
			for (;;)
			{
				k = gdbm_nextkey (self->dbf, x);
				free (x.dptr);
				if (k.dptr)
				{
					++n;
					x = k;
				}
				else
				{
					if (qu_gdbm__isfatal (self))
					{
						__qu_blocking_end;
						QU_RETHROW;
					}
					break;
				}
			}
		}
		else
		{
			if (qu_gdbm__isfatal (self))
			{
				__qu_blocking_end;
				QU_RETHROW;
			}
		}
		__qu_blocking_end;
		self->len = n;
	}
	return qu_integer_newl (self->len);
}



/*) sub __setitem (key, value) -> value
	Set an existing value or create a new one if it doesn't exist.

	@exc
	.ECall      already closed
	.EModify    opened for read-only
	.EArgValue
*/
static qu_any qu_gdbm__setitem (qu_gdbm self, qu_any akey, qu_any aval)
{
	volatile qu_string key, val;
	datum k, v;
	int   r;

	QU_GDBM_ENSURE_OPENED (self);
	key = qu_sys_string (akey);
	val = qu_sys_string (aval);

	QU_ARG_CHECKZ (key->len, "key can not be an emptry string");
	QU_ARG_CHECKZ (key->len < INT_MAX && val->len < INT_MAX, "key or value too big");

	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	{
		k.dptr  = key->str;
		k.dsize = key->len;
		v.dptr  = val->str;
		v.dsize = val->len;
		r = gdbm_store (self->dbf, k, v, GDBM_REPLACE);
	}
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);

	if (r == -1)
		QU_THROWZ (QuEModify, "reader cant write to database");
	else if (r == 1)
		QU_THROWZ (QuEKey, "key already exist"); /* should never happen */

	return aval;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/


/*) sub close
	Close the file.

	If it is already closed then nothing happens.

	@noexc
*/
static qu_null qu_gdbm_close (qu_gdbm self)
{
	if (self->dbf)
	{
		gdbm_close (self->dbf);
		self->dbf = 0;
	}
	return QuNil;
}



/*) sub firstkey -> String
	Get the first key.

	@exc
	.ECall already closed
	.ENone
*/
static qu_string qu_gdbm_firstkey (qu_gdbm self)
{
	qu_string key;
	datum     k;

	QU_GDBM_ENSURE_OPENED (self);
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	k = gdbm_firstkey (self->dbf);
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	QU_REQUIRE (QuENone, k.dptr);
	key = qu_string_new (k.dptr, k.dsize);
	free (k.dptr);
	return key;
}



/*) sub nextkey (key) -> String
	Get the next key after a specified key.

	@exc
	.ECall already closed
	.ENone
*/
static qu_string qu_gdbm_nextkey (qu_gdbm self, qu_any akey)
{
	volatile qu_string key;
	datum k, x;

	QU_GDBM_ENSURE_OPENED (self);
	key = qu_sys_string (akey);
	QU_REQUIRE (QuENone, key->len && key->len < INT_MAX);
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	k.dptr = key->str;
	k.dsize = key->len;
	x = gdbm_nextkey (self->dbf, k);
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	QU_REQUIRE (QuENone, x.dptr);
	key = qu_string_new (x.dptr, x.dsize);
	free (x.dptr);
	return key;
}



/*) sub reorganize -> Null
	Reorganize the database file.

	Call only if you have deleted many items.

	@exc
	.ESystem  failed
*/
static qu_null qu_gdbm_reorganize (qu_gdbm self)
{
	int r;

	QU_GDBM_ENSURE_OPENED (self);
	QU_GDBM_CLEAR_ERROR (self);
	__qu_blocking_begin;
	r = gdbm_reorganize (self->dbf);
	__qu_blocking_end;
	QU_GDBM_ENSURE_NONFATAL (self);
	QU_REQUIRE (QuESystem, r == 0);
	return QuNil;
}



/*) sub sync -> Null
	Flush data in memory to disk.

	@noexc
*/
static qu_null qu_gdbm_sync (qu_gdbm self)
{
	if (self->dbf)
       gdbm_sync (self->dbf);
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/
static void *QuGdbmHandle;

static void qu_gdbm_atexit ()
{
	if (QuGdbmHandle)
		qu_dl_close (QuGdbmHandle);
}

static int qu_gdbm_failed (const char *z)
{
	qu_gdbm_atexit ();
	QU_THROWF (QuESystem, "cant initialize Gdbm missing symbol %s", z);
}


int qu_Gdbm_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		gdbm
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuGdbmHandle = qu_dl_open_ex ("gdbm", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (QuGdbmHandle);

#define f_(name) \
	name = qu_dl_sym (QuGdbmHandle, #name); \
	if (name == 0) \
		return qu_gdbm_failed ( #name);

	f_(gdbm_version)
	f_(gdbm_open)
	f_(gdbm_close)
	f_(gdbm_store)
	f_(gdbm_fetch)
	f_(gdbm_delete)
	f_(gdbm_firstkey)
	f_(gdbm_nextkey)
	f_(gdbm_reorganize)
	f_(gdbm_sync)
	f_(gdbm_exists)
	f_(gdbm_setopt)
	f_(gdbm_fdesc)
	f_(gdbm_errno)
	f_(gdbm_strerror)

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		module
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_const (m, "READER"  , qu_int_newi (GDBM_READER));
	qu_module_add_const (m, "WRITER"  , qu_int_newi (GDBM_WRITER));
	qu_module_add_const (m, "WRCREATE", qu_int_newi (GDBM_WRCREAT));
	qu_module_add_const (m, "NEWDB"   , qu_int_newi (GDBM_NEWDB));
	qu_module_add_const (m, "SYNC"    , qu_int_newi (GDBM_SYNC));
	qu_module_add_const (m, "NOLOCK"  , qu_int_newi (GDBM_NOLOCK));


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Gdbm
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuGdbmClass = qu_class_new_finalz (m, "Gdbm", qu_gdbm_s, qu_gdbm__destroy, 0);

	/* standard class methods */
	qu_class_method_class (QuGdbmClass, QU_NAME_new, qu_gdbm_new, -1);

	/* standard instance methods */
	qu_class_method_inst  (QuGdbmClass, QU_NAME__each    , qu_gdbm__each    , 0);
	qu_class_method_inst  (QuGdbmClass, QU_NAME__del     , qu_gdbm__del     , 1);
	qu_class_alias        (QuGdbmClass, QU_NAME__del     , QU_NAME_delete);
	qu_class_method_inst  (QuGdbmClass, QU_NAME__in      , qu_gdbm__in      , 1);
	qu_class_aliasz       (QuGdbmClass, "__in"           , "haskey");
	qu_class_aliasz       (QuGdbmClass, "__in"           , "exists");
	qu_class_method_inst  (QuGdbmClass, QU_NAME__len     , qu_gdbm__len     , 0);
	qu_class_alias        (QuGdbmClass, QU_NAME__len     , QU_NAME_len);
	qu_class_method_inst  (QuGdbmClass, QU_NAME__setitem , qu_gdbm__setitem , 2);
	qu_class_method_inst  (QuGdbmClass, QU_NAME__getitem , qu_gdbm__getitem , 1);
	qu_class_method_inst  (QuGdbmClass, QU_NAME__getslice, qu_gdbm__getslice, 2);

	/* specific instance methods */
	qu_class_method_instz (QuGdbmClass, "close"     , qu_gdbm_close     , 0);
	qu_class_method_instz (QuGdbmClass, "firstkey"  , qu_gdbm_firstkey  , 0);
	qu_class_method_instz (QuGdbmClass, "nextkey"   , qu_gdbm_nextkey   , 0);
	qu_class_method_instz (QuGdbmClass, "reorganize", qu_gdbm_reorganize, 0);
	qu_class_method_instz (QuGdbmClass, "sync"      , qu_gdbm_sync      , 0);

	/* finalize declaration */
	qu_class_enddecl (QuGdbmClass);

	atexit (qu_gdbm_atexit);
	return 1;
}
