/* Gtk.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Gtk
	This module provides the necessary basic routines for Gtk modules.

	Note that you don't need to recompile Qu nor this module.
	Once you have installed the GTK library then it will be automatically
	available.

	Check the C{qu.conf} in the directory where Qu is installed and adjust
	the GTK entry as necessary.

	@p
	@code
		gtk     = /usr/lib/libgtk-x11-2.0.so
		gthread = /usr/lib/libgthread-2.0.so

	@p
	All GTK classes can be found in the Gtk directory.

	@xmp
		use Gtk
		use Gtk.Window
		use Gtk.Button

		win = Window ()
		win.title = "Test"
		win.connect ("destroy", @quit)

		btn = Button.from_stock ('gtk-zoom-100')
		btn.connect ("clicked", @hello)
		win.add (btn)

		win.show_all

		dialog = ColorSelectionDialog ()
		r = dialog.run
		println ('Result = ' r)
		dialog.destroy

		Gtk.main

		sub quit (p: Pointer, x: Any) -> Cint __pure
			# here self == x == win, and x.ptr == p
			println ('goodbye ' self ' ' x ' ' p)
			return Gtk.quit
		;;

		sub hello (p: Pointer, x: Any) -> Any __pure
			# here self == x == btn, and x.ptr == p
			println ('hello ' self ' ' x ' ' p ' ' self.get_pointer)
		;;

	@p
	@warn
		There are a few things you need to be aware of.

	@p
	The Gtk module does not care about reference counting. If a Qu object owns
	a reference to a GObject then that GObject will be alive as long as the
	Qu Object. For GtkObject, the Gtk module adds a reference automatically.

	This means that when the Qu object is garbage collected then a call to
	f{g_object_unref} is performed.

	The only important thing you must remember is that when you pass a Qu
	object to a GTK signal handler or a callback, you are responsible to keep
	that object alive as long as the signal handler or callback lives.

	All symbols are named without the C{"Gtk"} prefix and
	all methods are named without the C{"gtk_xxx_"} prefix.

	All setter methods that do not return anything in GTK (void), return k{self}
	in Qu thus allowing you to do something like:

	@prog
		win.set_title ('Hello').set_resizable (false).set_position (Gtk.WIN_POS_CENTER)

	@warn
		Signal handlers must not include the first and last arguments specified
		by the GTK manual. They will all receive the correct k{self} object
		instead.

		Make sure they have the correct signature otherwise your program
		will crash.

		Note also that if you plan to run your program with the -O option,
		you must use C{__pure} for all of your handlers to prevent if from
		being optimized. It is not possible to call JIT-ed handlers.

	@xmp
		sub myhandler (x: _int_) __pure
			...
		;;

	@p
	Conversion is as follows:
	@rc
	.| C{gboolean} | C{Boolean} | or C{Cint}
	.| C{gint}     | C{Cint}    | including all enumerations
	.| C{guint}    | C{Cuint}   | -
	.| C{gdouble}  | C{Cdouble} | -
	.| C{GObject}  | C{Gobject} | for all GObject objects
	.| C{gpointer} | C{Pointer} | including any other pointers

	@p
	For example:
	@c
		void add (GtkContainer *container, GtkWidget *widget, gpointer user_data)
		gboolean frame-event (GtkWindow *window, GdkEvent *event, gpointer user_data)
		void move_focus (GtkWindow *window, GtkDirectionType arg1, gpointer user_data)
	@p
	become:
	@prog
		sub user_function (widget: GObject)
			# here self is a Gtk.Container object
		;;

		sub frame-event (event: Pointer) -> Boolean
			# here self is a Gtk.Window object
		;;

		void move_focus (arg1: Cint)
			# here self is a Gtk.Window object
		;;

	@note
	As of Qu 1.21.05, all signal handlers callbacks no longer requires Ligthning.
	Multithreaded GTK application is possible. See the C{demo/gtk.qu} example.
*/


#ifndef QU_SUPPORT_GLIB

int qu_Gtk_init (qu_module m)
{
	QU_THROWZ (QuEImplement, "Qu not compiled to support GTK");
}

#else /*QU_SUPPORT_GLIB*/


/*============================================================================

	stuffs we need from GTK

============================================================================*/

/*
	basic
*/
#if	SIZEOF_LONG == SIZEOF_SIZE_T
#define GType unsigned long
#else
#define GType size_t
#endif

typedef struct GTypeClass GTypeClass;
struct GTypeClass
{
	GType g_type;
	/* whatever */
};

typedef struct GTypeInstance GTypeInstance;
struct GTypeInstance
{
	GTypeClass *g_class;
	/* whatever */
};

#define G_TYPE_FROM_INSTANCE(instance) (G_TYPE_FROM_CLASS (((GTypeInstance*) (instance))->g_class))
#define G_TYPE_FROM_CLASS(g_class)     (((GTypeClass*) (g_class))->g_type)

#define gulong         unsigned long
#define gboolean       int
#define guint          unsigned int
#define gchar          char
#define GObject        void
#define gpointer       void *
#define GConnectFlags  int

#define G_CONNECT_SWAPPED (1 << 1)

/* supported types */
#define G_TYPE_CHAR    (3  << 2)
#define G_TYPE_UCHAR   (4  << 2)
#define G_TYPE_BOOLEAN (5  << 2)
#define G_TYPE_INT	   (6  << 2)
#define G_TYPE_UINT	   (7  << 2)
#define G_TYPE_LONG	   (8  << 2)
#define G_TYPE_ULONG   (9  << 2)
#define G_TYPE_FLOAT   (14 << 2)
#define G_TYPE_DOUBLE  (15 << 2)
#define G_TYPE_STRING  (16 << 2)
#define G_TYPE_POINTER (17 << 2)
#define G_TYPE_OBJECT  (20 << 2)


/*
	core
*/
static void     (*gtk_init                   ) (int arg);
static gpointer (*g_object_ref               ) (GObject *o);
static void     (*g_object_unref             ) (GObject *o);
static gpointer (*g_object_get_data          ) (GObject *o, const gchar *k);
static void     (*g_object_set_data          ) (GObject *o, const gchar *k, gpointer data);
static void     (*g_object_set_data_full     ) (GObject *o, const gchar *k, gpointer data, void *destroy);
static gpointer (*g_object_steal_data        ) (GObject *o, const gchar *k);
static gulong   (*g_signal_connect_data      ) (GObject *o, const gchar *s, void *h, gpointer data, void *dest, GConnectFlags flags);
static void     (*g_signal_handler_disconnect) (GObject *instance, gulong id);
static gchar   *(*g_strdup                   ) (const gchar *s);
static gchar   *(*g_type_name                ) (GType type);
static void     (*gtk_object_sink            ) ();
static void     (*gtk_main_quit              ) ();
static void     (*gtk_main                   ) ();


/*
	GValue
*/
typedef struct GValue GValue;
struct GValue
{
	GType g_type;
	union
	{
		int       v_int;
		long      v_long;
		float     v_float;
		double    v_double;
		void     *v_pointer;
#if defined (HAVE_LONG_LONG) && SIZEOF_LONG_LONG == 8
		long long v_int64;
#elif (SIZEOF_SHORT + SIZEOF_SHORT) == 8
        short     v_int64[2];
#elif (SIZEOF_INT + SIZEOF_INT) == 8
        int       v_int64[2];
#elif SIZEOF_LONG == 8
		long      v_int64;
#else
#error cant get a 64bits integer type
#endif
	} data[2];
};


static GValue       *(*g_value_init       ) (GValue *g, GType t);
static void          (*g_value_unset      ) (GValue *g);
static void          (*g_value_set_char   ) (GValue *g, char v);
static char          (*g_value_get_char   ) (GValue *g);
static void          (*g_value_set_uchar  ) (GValue *g, unsigned char v);
static unsigned      (*g_value_get_uchar  ) (GValue *g);
static void          (*g_value_set_boolean) (GValue *g, int v);
static int           (*g_value_get_boolean) (GValue *g);
static void          (*g_value_set_int    ) (GValue *g, int v);
static int           (*g_value_get_int    ) (GValue *g);
static void          (*g_value_set_uint   ) (GValue *g, unsigned int v);
static unsigned int  (*g_value_get_uint   ) (GValue *g);
static void          (*g_value_set_long   ) (GValue *g, long v);
static long          (*g_value_get_long   ) (GValue *g);
static void          (*g_value_set_ulong  ) (GValue *g, unsigned long v);
static unsigned long (*g_value_get_ulong  ) (GValue *g);
static void          (*g_value_set_float  ) (GValue *g, float v);
static float         (*g_value_get_float  ) (GValue *g);
static void          (*g_value_set_double ) (GValue *g, double v);
static double        (*g_value_get_double ) (GValue *g);
static void          (*g_value_set_string ) (GValue *g, const char *v);
static char *        (*g_value_get_string ) (GValue *g);
static void          (*g_value_set_pointer) (GValue *g, void *v);
static void *        (*g_value_get_pointer) (GValue *g);
static void          (*g_value_set_object ) (GValue *g, GObject *v);
static GObject *     (*g_value_get_object ) (GValue *g);


/*
	property
*/
static void (*g_object_get_property) (GObject *g, const char *name, GValue *v);
static void (*g_object_set_property) (GObject *g, const char *name, const GValue *v);


/*
	Closure
*/
typedef struct GClosure GClosure;

static GClosure *(*g_cclosure_new            ) (void *f, gpointer u, void *d);
static void      (*g_closure_set_meta_marshal) (GClosure *closure, gpointer data, void *marshal);


/*
	GError
*/
typedef struct GError GError;
struct GError
{
	/* GQuark = guint32 idiotic */
#if SIZEOF_SHORT == 4
	short domain;
#elif SIZEOF_INT == 4
	int   domain;
#else
#error cant find a 32bits integer
#endif
	int        code;
	char      *msg;
};

static void (*g_error_free) (void *e);


/*
	helpers
*/
/* mem */
static void (*g_free    ) (void *m);
static void (*g_strfreev) (char **s);

/* GSlist */
typedef struct GSList GSList;
struct GSList
{
	void   *data;
	GSList *next;
};

static void         (*g_slist_free  ) (void *list);
static void *       (*g_slist_append) (void *list, void *data);
static unsigned int (*g_slist_length) (void *list);

/* GList */
typedef struct GList GList;
struct GList
{
	void  *data;
	GList *next;
	GList *prev;
};

static void         (*g_list_free  ) (void *list);
static void *       (*g_list_append) (void *list, void *data);
static unsigned int (*g_list_length) (void *list);


/*
	Missing - PangoLogAttr
*/

typedef struct PangoLogAttr PangoLogAttr;
struct PangoLogAttr
{
	unsigned int is_line_break               : 1; /*  0 */
	unsigned int is_mandatory_break          : 1; /*  1 */
	unsigned int is_char_break               : 1; /*  2 */
	unsigned int is_white                    : 1; /*  3 */
	unsigned int is_cursor_position          : 1; /*  4 */
	unsigned int is_word_start               : 1; /*  5 */
	unsigned int is_word_end                 : 1; /*  6 */
	unsigned int is_sentence_boundary        : 1; /*  7 */
	unsigned int is_sentence_start           : 1; /*  8 */
	unsigned int is_sentence_end             : 1; /*  9 */
	unsigned int backspace_deletes_character : 1; /* 10 */
};


typedef struct PangoGlyphVisAttr PangoGlyphVisAttr;
struct PangoGlyphVisAttr
{
	unsigned int is_cluster_start : 1;
};



/*============================================================================

	our own stuffs

============================================================================*/

/* key name to use to store our corresponding object in a GObject */
#define QU_GTK_KEY "Qu"

/* this module */
static qu_module QuGtkModule;

/* the DL handle */
static void *QuGtkHandle     = NULL;
static void *QuGthreadHandle = NULL;


/*
	List of all Qu classes.
	This is a root Dict with key = Gtk name (String), value = Qu Class.
	Items are collected via the 'register' function.
*/
static qu_dict AllClasses;


/*
	List of all resolved Methods.
*/
static qu_dict AllMethods;


/*
	The GtkObject class.
*/
static qu_class QuGtkObjectClass;


/*
	This is called by GTK when the object is destroyed.
	Here we disassociate the GObject and the Qu object by setting the 'ptr'
	field to 0 and remove it and all of its handlers from root list.
*/
static void qu_gtk__destroy_notify (qu_gobject self)
{
	self->ptr = 0;
}


/*
	Wrap a GObject into our object.

	This is called to automatically create our object from a returned value
	as a result of calling a GTK function.
*/
static qu_gobject qu_gtk_from_gobject (GObject *g)
{
	qu_gobject o;

	/* see if we already have a Qu object associated with this GTK object*/
	o = g_object_get_data (g, QU_GTK_KEY);
	if (o == 0)
	{
		/* nope, so set up a new one */

		char     *z = (char *) g_type_name (G_TYPE_FROM_INSTANCE (g));
		qu_class  c = qu_dict_getitemz (AllClasses, z, strlen (z));

		QU_REQUIREF (QuENone, c
				&& qu_class_is (c)
				&& !qu_class_is_abstract (c)
				&& qu_class_isa (c, QuGObjectClass), "%s: no associated Qu class", z);

		o = qu_gc_new (c);
		o->ptr = g;
		o->callbacks = NULL;

		g_object_ref (g);
		g_object_set_data_full (g, QU_GTK_KEY, o, qu_gtk__destroy_notify);
	}
	return o;
}


/*
	Guess the GType of a Qu instance.
*/
static GType qu_gtk_gtype_guess (qu_any x)
{
	if (qu_gobject_is (x)) return G_TYPE_OBJECT;
	if (qu_pointer_is (x)) return G_TYPE_POINTER;
	if (qu_int_is     (x)) return G_TYPE_INT;
	if (qu_long_is    (x)) return G_TYPE_LONG;
	if (qu_stringy_is (x)) return G_TYPE_STRING;
	if (qu_decimal_is (x)) return G_TYPE_DOUBLE;
	if (qu_boolean_is (x)) return G_TYPE_BOOLEAN;
	return 0;
}



/*
	The GType of a Qu type.
*/
static GType qu_gtk_type_qu2g (qu_any t)
{
	if (t == QuType_byte   ) return G_TYPE_UCHAR;
	if (t == QuBooleanClass) return G_TYPE_BOOLEAN;
	if (t == QuType_int    ) return G_TYPE_INT;
	if (t == QuType_uint   ) return G_TYPE_UINT;
	if (t == QuType_long   ) return G_TYPE_LONG;
	if (t == QuType_ulong  ) return G_TYPE_ULONG;
	if (t == QuType_float  ) return G_TYPE_FLOAT;
	if (t == QuType_double ) return G_TYPE_DOUBLE;
	if (t == QuStringClass ) return G_TYPE_STRING;
	if (t == QuPointerClass) return G_TYPE_POINTER;
	if (t == QuGObjectClass) return G_TYPE_OBJECT;
	if (qu_class_is (t))
	{
		qu_class c = (qu_class) t;

		if (qu_class_isa (c, QuGObjectClass))
			return G_TYPE_OBJECT;

		if (qu_class_isa (c, QuPointerClass))
			return G_TYPE_POINTER;
	}
	return 0;
}


/*
	The Qu type of a GType.
*/
static qu_any qu_gtk_type_g2qu (GType t)
{
	switch (t)
	{
		case G_TYPE_UCHAR:   return QuType_byte;
		case G_TYPE_BOOLEAN: return QuBooleanClass;
		case G_TYPE_INT:     return QuType_int;
		case G_TYPE_UINT:    return QuType_uint;
		case G_TYPE_LONG:    return QuType_long;
		case G_TYPE_ULONG:   return QuType_ulong;
		case G_TYPE_FLOAT:   return QuType_float;
		case G_TYPE_DOUBLE:  return QuType_double;
		case G_TYPE_STRING:  return QuStringClass;
		case G_TYPE_POINTER: return QuPointerClass;
		case G_TYPE_OBJECT:  return QuGObjectClass;
	}
	return 0;
}


/*
	The argument character a Qu type.
*/
static char qu_gtk_type_qu2c (qu_any t)
{
	if (t == QuType_byte   ) return 'y';
	if (t == QuBooleanClass) return 'b';
	if (t == QuType_int    ) return 'i';
	if (t == QuType_uint   ) return 'I';
	if (t == QuType_long   ) return 'l';
	if (t == QuType_ulong  ) return 'L';
	if (t == QuType_float  ) return 'g';
	if (t == QuType_double ) return 'f';
	if (t == QuStringClass ) return 's';
	if (t == QuPointerClass) return 'p';
	if (t == QuGObjectClass) return 'W';
	if (qu_class_is (t))
	{
		qu_class c = (qu_class) t;

		if (qu_class_isa (c, QuGObjectClass))
			return 'W';

		if (qu_class_isa (c, QuPointerClass))
			return 'p';
	}
	return 0;
}


/*
	Set GValue from a Qu object.
*/
static int qu_gtk_q2v (GValue *g, qu_any x)
{
	switch (g->g_type)
	{
		case G_TYPE_CHAR:
		{
			char v;
			QU_ARG_byte (x, v);
			g_value_set_char (g, v);
			return 1;
		}

		case G_TYPE_UCHAR:
		{
			unsigned char v;
			QU_ARG_ubyte (x, v);
			g_value_set_uchar (g, v);
			return 1;
		}

		case G_TYPE_BOOLEAN:
		{
			int v;
			QU_ARG_bool (x, v);
			g_value_set_boolean (g, v);
			return 1;
		}

		case G_TYPE_INT:
		{
			int v;
			QU_ARG_int (x, v);
			g_value_set_int (g, v);
			return 1;
		}

		case G_TYPE_UINT:
		{
			unsigned int v;
			QU_ARG_uint (x, v);
			g_value_set_uint (g, v);
			return 1;
		}

		case G_TYPE_LONG:
		{
			long v;
			QU_ARG_long (x, v);
			g_value_set_long (g, v);
			return 1;
		}

		case G_TYPE_ULONG:
		{
			unsigned long v;
			QU_ARG_ulong (x, v);
			g_value_set_ulong (g, v);
			return 1;
		}

		case G_TYPE_FLOAT:
		{
			float v;
			QU_ARG_float (x, v);
			g_value_set_float (g, v);
			return 1;
		}

		case G_TYPE_DOUBLE:
		{
			double v;
			QU_ARG_double (x, v);
			g_value_set_double (g, v);
			return 1;
		}

		case G_TYPE_STRING:
		{
			QU_ARG_TYPE (x, qu_string);
			g_value_set_string (g, ((qu_string) x)->str);
			return 1;
		}

		case G_TYPE_POINTER:
		{
			QU_ARG_TYPE (x, qu_pointer);
			g_value_set_string (g, ((qu_pointer) x)->ptr);
			return 1;
		}

		case G_TYPE_OBJECT:
		{
			QU_ARG_TYPE (x, qu_gobject);
			g_value_set_object (g, ((qu_gobject) x)->ptr);
			return 1;
		}
	}
	QU_ARG_THROW;
}


/* get Qu value from GValue */
static qu_any qu_gtk_v2q (GValue *g)
{
	switch (g->g_type)
	{
		case G_TYPE_CHAR:
		case G_TYPE_UCHAR:
		{
			char v = g_value_get_char (g);
			return qu_string_byte (v);
		}

		case G_TYPE_BOOLEAN:
		{
			int v = g_value_get_boolean (g);
			return v ? QuTrue: QuFalse;
		}

		case G_TYPE_INT:
		{
			int v = g_value_get_int (g);
			return qu_integer_newi (v);
		}

		case G_TYPE_UINT:
		{
			unsigned int v = g_value_get_uint (g);
			return qu_integer_newui (v);
		}

		case G_TYPE_LONG:
		{
			long v = g_value_get_long (g);
			return qu_integer_newl (v);
		}

		case G_TYPE_ULONG:
		{
			unsigned long v = g_value_get_ulong (g);
			return qu_integer_newul (v);
		}

		case G_TYPE_FLOAT:
		{
			float v = g_value_get_float (g);
			return qu_float_newdx (v);
		}

		case G_TYPE_DOUBLE:
		{
			double v = g_value_get_double (g);
			return qu_float_newdx (v);
		}

		case G_TYPE_STRING:
		{
			char *v = g_value_get_string (g);
			return qu_string_newz (v);
		}

		case G_TYPE_POINTER:
		{
			void *v = g_value_get_pointer (g);
			return qu_pointer_new (v);
		}

		case G_TYPE_OBJECT:
		{
			GObject *v = g_value_get_object (g);
			if (v == NULL)
				return QuNil;
			return qu_gtk_from_gobject (v);
		}
	}
	return NULL;
}



/*============================================================================

	signal handler

============================================================================*/
#define MAXARGS 8

typedef struct qu_callback_s qu_callback_s;
struct qu_callback_s
{
	QU_OPAQUE_HEAD;
	qu_sub     sub;   /* the Sub object to call */
	qu_any     self;  /* the self object to pass to "sub" */
	gulong     id;    /* non-zero only for signal handlers */
};
typedef qu_callback_s *qu_callback ;


static void qu_callback__mark (qu_callback self)
{
	qu_gc_mark  (self->sub);
	qu_gc_xmark (self->self);
}



/* possible GTK arguments passed to handlers */
static int qu_callback_valid_arg (char type)
{
	switch (type)
	{
		case QU_STRUCT_ID_int:
		case QU_STRUCT_ID_uint:
		case QU_STRUCT_ID_double:
		case QU_STRUCT_ID_Boolean:
		case QU_STRUCT_ID_String:
		case QU_STRUCT_ID_Pointer:
		case QU_STRUCT_ID_GObject:
			return 1;
	}
	return 0;
}


/* validate callback object */
static int qu_callback_validate (qu_sub u)
{
	int i;

	QU_ARG_TYPE   (u, qu_sub);
	QU_ARG_CHECKF (qu_sub_is_static (u), "%s must be a static sub", u->name->str);

	if (qu_sub_is_closure (u) || qu_sub_is_extern (u) || qu_sub_is_jit (u))
		QU_ARG_THROWF ("%s cant use closure or extern or JIT-ed handler", u->name->str);

	if (qu_sub_is_argmin (u)
			|| qu_sub_is_vararg (u)
			|| u->argc < 0
			|| (u->argc && u->ltype == NULL))
		QU_ARG_THROWF ("%s must have fixed typed arguments", u->name->str);

	if (u->argc > MAXARGS)
		QU_ARG_THROWF ("%s has too many arguments", u->name->str);

	if (u->type)
	{
		switch (qu_struct_type_to_char (u->type))
		{
			case QU_STRUCT_ID_Boolean:
			case QU_STRUCT_ID_int:
			case QU_STRUCT_ID_float:
			case QU_STRUCT_ID_double:
			case QU_STRUCT_ID_String:
			case QU_STRUCT_ID_GObject:
				break;

			default:
				QU_ARG_THROWF ("%s has an unsupported return type", u->name->str);
		}
	}

	for (i = 0; i < u->argc; i++)
	{
		qu_any x = u->ltype [i];
		if (!qu_callback_valid_arg (qu_struct_type_to_char (x)))
			QU_ARG_THROWF ("%s has invalid argument", u->name->str);
	}

	return 1;
}



static qu_callback qu_gtk_callback_alloc (qu_any self, qu_sub u)
{
	qu_callback o = (qu_callback) qu_opaque__new ((qu_opaquemark_f) qu_callback__mark,
		(qu_opaquefree_f) NULL, sizeof (qu_callback_s));

	o->sub  = u;
	o->self = self;
	o->id   = 0;
	return o;
}


static void qu_gtk_callback_setarg (qu_sub u, void *p, int argc, qu_any *argv)
{
	int i;

	for (i = 0; i < argc; ++i)
	{
		switch (qu_struct_type_to_char (u->ltype [i]))
		{
			default:
			case QU_STRUCT_ID_int:
			{
				int v = *(int *) p;
				argv[i] = qu_integer_newi (v);
				p += sizeof (long);
				continue;
			}

			case QU_STRUCT_ID_uint:
			{
				unsigned int v = *(unsigned int *) p;
				argv[i] = qu_integer_newui (v);
				p += sizeof (long);
				continue;
			}

			case QU_STRUCT_ID_double:
			{
				double v = *(double *) p;
				qu_any o = qu_float_newdx (v);
				argv[i] = o ? o : QuFloat0;
				p += sizeof (double);
				continue;
			}

			case QU_STRUCT_ID_Null:
			{
				argv[i] = QuNil;
				p += sizeof (long);
				continue;
			}

			case QU_STRUCT_ID_Boolean:
			{
				int v = *(int *) p;
				argv[i] = v ? QuTrue : QuFalse;
				p += sizeof (long);
				continue;
			}

			case QU_STRUCT_ID_Pointer:
			{
				void *v = *(void **) p;
				argv[i] = v ? qu_pointer_new (v) : QuPointerNULL;
				p += sizeof (void *);
				continue;
			}

			case QU_STRUCT_ID_String:
			{
				char *v = *(char **) p;
				argv[i] = v ? qu_string_newz (v) : QuStringEMPTY;
				p += sizeof (void *);
				continue;
			}

			case QU_STRUCT_ID_GObject:
			{
				void *v = *(void **) p;
				if (v)
				{
					qu_any x = __QuGObjectNew (v);
					argv[i] = x ? x : (qu_any) QuNil;
				}
				else
				{
					argv[i] = QuNil;
				}
				p += sizeof (void *);
				continue;
			}
		}
	}
}


#define ENTER \
	qu_thread tHr = qu_thread_self (); \
	int       bLoCk = qu_thread_blocking (tHr); \
	if (bLoCk) \
		__qu_blocking_end


#define LEAVE \
	if (bLoCk) \
		__qu_blocking_begin


/*
	Generic marshaller.
*/
static qu_pointer QuGtkMarshallPtr;
static void qu_gtk_marshal (GClosure *dontcare, GValue *ret, guint argc, GValue *arg, gpointer hint, qu_callback b)
{

#define MAXCLOSUREARGS 64

	qu_any  argv [MAXCLOSUREARGS];
	guint   i;
	qu_any  r;

	ENTER;

	if (argc > MAXCLOSUREARGS)
	{
		QU_SYS_WARN ("closure has too many arguments");
		goto done;
	}

	for (i = 0; i < argc; ++i, arg++)
	{
		qu_any x = qu_gtk_v2q (arg);

		if (x == 0)
		{
			QU_SYS_WARN ("cant convert GValue to Qu");
			goto done;
		}
		argv[i] = x;
	}

	r = qu_sub_calln (b->sub, b->self, argc, argv);
	if (r == 0)
	{
		QU_SYS_WARN ("closure threw an exception");
	}
	else if (ret)
	{
		if (qu_gtk_q2v (ret, r) == 0)
			QU_SYS_WARN ("closure threw an exception");
	}

  done:
	LEAVE;
}



/*============================================================================

	functions

============================================================================*/
/*) sub main -> Null
	This is the main loop.

	@warn
	You must never call gtk_main directly. Always call this function
	instead. This ensure that your multithreaded Qu application run as
	expected.

	@p
	This function blocks the calling thread so that your other Qu threads
	can still run.
*/
static qu_null qu_gtk_main ()
{
	__qu_blocking_begin;
	gtk_main ();
	__qu_blocking_check_signal;
	__qu_blocking_end;
	return QuNil;
}



/*) sub quit -> _int_
	Abort the main loop then return 1.
*/
static qu_int qu_gtk_quit ()
{
	gtk_main_quit ();
	return QuInt1;
}



/*) sub register (klass: Class, name = klass.__name)
	You must register all real classes here so that #{get} works.

	v{name} is the GTK class name, e.g. C{'GtkWindow'} and v{klass} is
	the corresponding Qu class which must be a descendant of the GObject class.
*/
static qu_null qu_gtk_register (int argc, qu_any *argv)
{
	qu_class  c;
	qu_string name;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, c, qu_class);
	QU_ARG_CHECK (qu_class_isa (c, QuGObjectClass));

	if (argc > 1)
		QU_ARG_VTYPE (1, name, qu_string);
	else
		name = c->name;

	/* !!! AllClasses is a static Dict */
	qu_dict_setitem (AllClasses, qu_string_static_dup (name->str, name->len), c);

	return QuNil;
}


/*) sub call (*arg) -> Any
	Call a GTK function.

	This is just a wrapper around -{Dlib.call}.
	It should be a bit more efficient than doing it from Qu programs.
*/
static qu_any qu_gtk_call (int argc, qu_any *argv)
{
	return qu_dlib_call (QuGtkHandle, argc, argv);
}



/*) sub gobject (p, allownil = false) -> Any
	Get or create the Qu object from a GObject.

	If v{p} is nil or a NULL pointer and v{allownil} is true then return nil,
	otherwise v{ptr} must point to a valid GObject.

	If no Qu object is already associated with the GObject object then
	a new object is created. Otherwise, that object is simply returned.

	@exc
	.EArgValue
	.ENone     no associated registered Qu class
*/
static qu_any qu_gtk_gobject (int argc, qu_any *argv)
{
	int    n = 0;
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	if (argc > 1)
		QU_ARG_vbool (1, n);

	x = argv[0];
	if (x != (qu_any) QuNil)
	{
		QU_ARG_TYPE (x, qu_pointer);
		if (((qu_pointer) x)->ptr)
			return qu_gtk_from_gobject ((GObject *) ((qu_pointer) x)->ptr);
	}

	if (n)
		return QuNil;
	QU_ARG_THROW;
}



/*) sub gobjectptr (p, allownil = false) -> Any
	Does the same as #{gobject} except that the pointer is the pointer contained in v{p}.

	@exc
	.EArgValue
	.ENone     no associated registered Qu class
*/
static qu_any qu_gtk_gobjectptr (int argc, qu_any *argv)
{
	int    n = 0;
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	if (argc > 1)
		QU_ARG_vbool (1, n);

	x = argv[0];
	if (x != (qu_any) QuNil)
	{
		void *p;

		QU_ARG_TYPE (x, qu_pointer);
		p = ((qu_pointer) x)->ptr;
		if (p)
		{
			void *g = *((void **) p);
			if (g)
				return qu_gtk_from_gobject ((GObject *) g);
		}
	}

	if (n)
		return QuNil;
	QU_ARG_THROW;
}



/*) sub closure (f: Sub, theself = nil, retval = false) -> Array
	Create a Qu closure to be called with the GClosure mechanism.

	The returned value is an Array of 2 or 3 items. The thirs items is only present
	if v{retval} is true.
	@ul
	.Pointer to the GClosure object
	.Opaque object to keepalive as long as the closure is in use
	.Pointer to the GValue returned value (NULL if the Sub does not return any value)

	@p
	See #{closurearg} for an example.
*/
static int qu_gtk_fake_callback ()
{
	return 1;
}

static qu_array qu_gtk_closure (int argc, qu_any *argv)
{
	int          retval = 0;
	qu_any       self = QuNil;
	qu_sub       u;
	qu_callback  h;
	GClosure    *c;
	qu_pointer   ret = 0;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, u, qu_sub);
	if (argc > 1)
	{
		self = argv[1];
		if (argc > 2)
			QU_ARG_vbool (2, retval);
	}

	QU_ENSURE (qu_callback_validate (u));
	if (u->type && retval)
	{
		GType t = qu_gtk_type_qu2g (u->type);
		assert (t);
		ret = qu_pointer_new_alloc (sizeof (GValue));
		g_value_init (ret->ptr, t);
	}

	h = qu_gtk_callback_alloc (self, u);
	assert (h);

	c = g_cclosure_new (qu_gtk_fake_callback, NULL, NULL);
	QU_REQUIREZ (QuESystem, c, "g_cclosure_new failed");

	g_closure_set_meta_marshal (c, h, qu_gtk_marshal);

	if (retval)
		return qu_array_new_ex (3, qu_pointer_new (c), h, ret);
	return qu_array_new_ex (2, qu_pointer_new (c), h);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GValue
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*) sub gvalue_size -> Int
	Return the size of the GValue struct.

	@noexc
*/
static qu_int qu_gtk_gvalue_size ()
{
	return qu_int_new (sizeof (GValue));
}



/*) sub gvalue_new -> Pointer
	Allocate an uninitialized GValue to be filled by a call to a GTK function.
	You will then use #{gvalue_get} to get the value.

	@noexc
*/
static qu_pointer qu_gtk_gvalue_new ()
{
	return qu_pointer_new_calloc (sizeof (GValue));
}



/*) sub gvalue_init (g: Pointer, type) -> Pointer
	Initialize a GValue with a type.

	Following are the supported types:
	@table
	.| Cbyte    | G_TYPE_UCHAR
	.| Boolean  | G_TYPE_BOOLEAN
	.| Cint     | G_TYPE_INT
	.| Cuint    | G_TYPE_UINT
	.| Clong    | G_TYPE_LONG
	.| Culong   | G_TYPE_ULONG
	.| Cfloat   | G_TYPE_FLOAT
	.| Cdouble  | G_TYPE_DOUBLE
	.| String   | G_TYPE_STRING
	.| Pointer  | G_TYPE_POINTER
	.| GObject  | G_TYPE_OBJECT

	@exc
	.EArgValue invalid or unsupported type
*/
static qu_pointer qu_gtk_gvalue_init (qu_pointer p, qu_any type)
{
	GType t;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && p->size >= sizeof (GValue));
	t = qu_gtk_type_qu2g (type);
	if (t == 0)
		QU_THROWZ (QuEArgValue, "invalid or unsupported GValue type");
	g_value_init (p->ptr, t);
	return p;
}



/*) sub gvalue_unset (p: Pointer)
	Call this function when GTK tells you to call g_value_unset.
*/
static qu_any qu_gtk_gvalue_unset (qu_pointer p)
{
	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && p->size >= sizeof (GValue));
	g_value_unset (p->ptr);
	return QuNil;
}



/*) sub gvalue_newv (value: Any, type = nil) -> Pointer
	Allocate and initialize a GValue.

	See #{gvalue_init} for valid v{type} values.
	If v{type} is nil then it is guessed from value.

	@exc
	.EArgValue invalid or unsupported type
*/
static qu_pointer qu_gtk_gvalue_newv (int argc, qu_any *argv)
{
	qu_any x;
	GType  t;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (argc > 1)
		t = qu_gtk_type_qu2g (argv[1]);
	else
		t = qu_gtk_gtype_guess (x);
	if (t)
	{
		qu_pointer p = qu_pointer_new_calloc (sizeof (GValue));
		g_value_init (p->ptr, t);
		if (qu_gtk_q2v (p->ptr, x))
			return p;
	}
	QU_ARG_THROWZ ("invalid or unsupported GValue type");
}



/*) sub gvalue_get (p: Pointer) -> Any
	Get the value stored in a GValue.

	@exc
	.EArgValue invalid or unsupported type
*/
static qu_any qu_gtk_gvalue_get (qu_pointer p)
{
	qu_any x;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && p->size >= sizeof (GValue));
	x = qu_gtk_v2q (p->ptr);
	if (x)
		return x;
	QU_ARG_THROWZ ("invalid or unsupported GValue type");
}



/*) sub gvalue_set (gvalue: Pointer, v: Any) -> Any
	Set the value in a GValue.

	@warn
	The GValue is assumed to be already initialized with the proper type.
	You must pass v{v} as the proper value.

	@exc
	.EArgValue invalid or unsupported type
*/
static qu_any qu_gtk_gvalue_set (qu_pointer p, qu_any v)
{
	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && p->size >= sizeof (GValue));
	if (qu_gtk_q2v (p->ptr, v))
		return v;
	QU_ARG_THROWZ ("invalid or unsupported GValue type");
}



/*) sub gvalue_type (gvalue: Pointer) -> Any
	Get the Qu value type of an initialized GValue.

	@exc
	.EArgValue invalid or unsupported type
*/
static qu_any qu_gtk_gvalue_type (qu_pointer p)
{
	GValue *v;
	qu_any  t;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && p->size >= sizeof (GValue));
	v = p->ptr;
	t = qu_gtk_type_g2qu (v->g_type);
	if (t)
		return t;
	QU_ARG_THROWZ ("invalid or unsupported GValue type");
}



/*) sub gvalue_gtype (gvalue: Pointer) -> Integer
	Get the GType of an initialized GValue.

	@exc
	.EArgValue
*/
static qu_any qu_gtk_gvalue_gtype (qu_pointer p)
{
	GValue *v;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && p->size >= sizeof (GValue));
	v = p->ptr;
	return qu_integer_newul (v->g_type);
}



/*) sub gvalue_array (*arg) -> Pointer
	Create a Pointer of GValue array.

	If you pass no arguments then a Pointer.NULL is returned.

	@exc
	.EArgValue

	@xmp
		use Gtk

		sub foo (a: Cint, b: String) -> Cint
			println ("In foo...")
			println ('a = ' a)
			println ('b = ' b)
			return a * 3
		;;

		(closure, keepalive, retval) = Gtk.closure (@foo, nil, true)
		args = Gtk.gvalue_array (123, "Hello World!")
		__extern Gtk.g_closure_invoke "-:ppIpp" (closure, retval, 2, args, nil)
		println (Gtk.gvalue_get (retval))

	@out
		In foo ...
		a = 123
		b = Hello World!
		369
*/
static qu_pointer qu_gtk_gvalue_array (int argc, qu_any *argv)
{
	if (argc == 0)
		return QuPointerNULL;
	else
	{
		qu_pointer r = qu_pointer_new_calloc ((argc + 1) * sizeof (GValue));
		GValue    *g = (GValue *) r->ptr;
		int        i = 0;
		qu_any     x;
		GType      t;

		for (i = 0; i < argc; ++i, ++g)
		{
			x = argv [i];
			t = qu_gtk_gtype_guess (x);
			if (t == 0)
				goto nope;
			g_value_init (g, t);
			if (qu_gtk_q2v (g, x) == 0)
				goto nope;
		}
		return r;

	  nope:
		QU_THROWF (QuEArgValue, "cant convert argument %d to GValue", i);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	type
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*) sub gtypeof (type: Any) -> Integer
	Return the corresponding GType of a Qu type.

	v{type} must be one of C{Cbyte}, C{Boolean}, etc. as in the following table:
	@table
	Following are the supported types:
	@table
	.| Cbyte    | G_TYPE_UCHAR
	.| Boolean  | G_TYPE_BOOLEAN
	.| Cint     | G_TYPE_INT
	.| Cuint    | G_TYPE_UINT
	.| Clong    | G_TYPE_LONG
	.| Culong   | G_TYPE_ULONG
	.| Cfloat   | G_TYPE_FLOAT
	.| Cdouble  | G_TYPE_DOUBLE
	.| String   | G_TYPE_STRING
	.| Pointer  | G_TYPE_POINTER
	.| GObject  | G_TYPE_OBJECT

	@exc
	.ENone
*/
static qu_integer qu_gtk_gtypeof (qu_any qu)
{
	GType g;

	if (qu_module_is (qu))
	{
		qu = qu_module_getattr_callable_name ((qu_module) qu, 1);
		QU_REQUIRE (QuENone, qu);
	}
	g = qu_gtk_type_qu2g (qu);
	QU_REQUIRE (QuENone, g);
	return qu_integer_newul (g);
}



/*) sub typeof (gtype: Integer) -> Any
	Return the corresponding Qu type of a GType.

	See #{gtypeof} for supported v{gtype} values.

	@exc
	.ENone
*/
static qu_any qu_gtk_typeof (qu_integer gt)
{
	GType  g;
	qu_any qu;

	QU_ARG_ulong (gt, g);
	qu = qu_gtk_type_g2qu (g);
	QU_REQUIRE (QuENone, qu);
	return qu;
}



/*) sub args (*arg) -> String
	Return the character to pass to Dlib et.al. for one or more GType value.

	v{arg} is a list of Qu type as explained in #{gtypeof}.

	@exc
	.ENone

	@xmp
		use Gtk

		s = Gtk.args (Cint, Cfloat, String)
		println (s)

	@out
		igs
*/
static qu_string qu_gtk_args (int argc, qu_any *argv)
{
	char buf [256], *s = buf;
	int  n = argc;

	QU_ARG_VCOUNT (1, sizeof (buf));
	while (n--)
	{
		qu_any x = *argv++;
		char   c = qu_gtk_type_qu2c (x);

		QU_ARG_CHECKZ (c, "unsupported type");
		*s++ = c;
	}
	return qu_string_new (buf, argc);
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	callbacks
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*) sub hook (handler: Sub, theself = self) -> Array
	Create a callback which the last argument is user data.

	The returned value is an Array of two items:
	@ul
	.a Pointer which must be passed as the callback function to the GTK function.
	.an Opaque object which must be passed to the GTK function as user data

	@p
	The opaque object must be saved into a variable or instance property of self
	if necessary.

	@warn
		The Qu callback function must not include the last user data argument.
		In the function, self will be v{theself}.

	@xmp
		sub myhook (arg1: Cint, user_data: Any)  #-> WRONG!
		sub myhook (arg1: Cint)                  #-> CORRECT

	@exc
	.EArgValue
*/
#include "GtkHook.c"
static qu_array qu_gtk_hook (int argc, qu_any *argv)
{
	qu_any       theself = QuNil;
	qu_sub       u;
	void        *f;
	qu_callback  h;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, u, qu_sub);
	if (argc > 1)
		theself = argv [1];

	QU_ENSURE (qu_callback_validate (u));
	f = qu_gtk_hook_getfunc (u);
	assert (f);
	h = qu_gtk_callback_alloc (theself, u);
	assert (h);
	return qu_array_new_ex (2, qu_pointer_new (f), h);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GError
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
static qu_class QuEGErrorClass;


/*) sub gerror_throw (p: Pointer)
	Throw EGError with the error message reported by a GError.

	@warn
	v{p} must contain a pointer to a GError set by a GTK function.

	This function already calls f{g_error_free} so you must not call it yourself.

	@xmp
		use Qu
		use Gtk

		__extern Gtk.gtk_whatever "-:p" (Qu.PTR1)
		Gtk.gerror_throw (Qu.PTR1) unless Qu.PTR1.isnull
*/
static qu_null qu_gerror_throw (qu_pointer p)
{
	GError *e;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr);
	e = *((GError **) p->ptr);
	QU_ARG_CHECK (e && e->msg);
	qu_thread_throw (QuEGErrorClass, qu_string_newz (e->msg));
	g_error_free (e);
	return 0;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	helpers
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*) sub glist_get_gobjects (p: Pointer, free = true) -> Array
	Get an Array of GObjects contained in a C{GList}.

	If free is true then C{g_list_free} is called by this function, otherwise you
	must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_glist_get_gobjects (int argc, qu_any *argv)
{
	qu_pointer  p;
	int         f = 1;
	GList      *g, *i;
	qu_gobject  o;
	long        n;
	qu_array    a = NULL;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, p, qu_pointer);
	if (argc > 1)
		QU_ARG_vbool (1, f);
	g = (GList *) p->ptr;
	if (g)
	{
		n = g_list_length (g);
		assert (n <= QU_GC_MAX_ITEMS);
		a = qu_array_new (n);
		i = g;
		do
		{
			assert (i->data);
			o = qu_gtk_from_gobject (i->data);
			QU_ENSURE (o);
			qu_array_push_quick (a, o);
			i = i->next;

		} while (i);

		if (f)
			g_list_free (g);
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub glist_get_pointers (p: Pointer, free = true) -> Array
	Get an Array of Pointers contained in a C{GList}.

	If free is true then C{g_list_free} is called by this function, otherwise you
	must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_glist_get_pointers (int argc, qu_any *argv)
{
	qu_pointer  p;
	int         f = 1;
	GList      *g, *i;
	long        n;
	qu_array    a = NULL;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, p, qu_pointer);
	if (argc > 1)
		QU_ARG_vbool (1, f);
	g = (GList *) p->ptr;
	if (g)
	{
		n = g_list_length (g);
		assert (n <= QU_GC_MAX_ITEMS);
		a = qu_array_new (n);
		i = g;
		do
		{
			assert (i->data);
			qu_array_push_quick (a, qu_pointer_new (i->data));
			i = i->next;

		} while (i);

		if (f)
			g_list_free (g);
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub gslist_get_gobjects (p: Pointer, free = true) -> Array
	Get an Array of GObjects contained in a C{GSList}.

	If free is true then C{g_slist_free} is called by this function, otherwise you
	must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_gslist_get_gobjects (int argc, qu_any *argv)
{
	qu_pointer  p;
	int         f = 1;
	GSList     *g, *i;
	qu_gobject  o;
	long        n;
	qu_array    a = NULL;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, p, qu_pointer);
	if (argc > 1)
		QU_ARG_vbool (1, f);
	g = (GSList *) p->ptr;
	if (g)
	{
		n = g_slist_length (g);
		assert (n <= QU_GC_MAX_ITEMS);
		a = qu_array_new (n);
		i = g;
		do
		{
			assert (i->data);
			o = qu_gtk_from_gobject (i->data);
			QU_ENSURE (o);
			qu_array_push_quick (a, o);
			i = i->next;

		} while (i);

		if (f)
			g_list_free (g);
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub gslist_get_pointers (p: Pointer, free = true) -> Array
	Get an Array of Pointers contained in a C{GSList}.

	If free is true then C{g_slist_free} is called by this function, otherwise you
	must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_gslist_get_pointers (int argc, qu_any *argv)
{
	qu_pointer  p;
	int         f = 1;
	GSList     *g, *i;
	long        n;
	qu_array    a = NULL;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, p, qu_pointer);
	if (argc > 1)
		QU_ARG_vbool (1, f);
	g = (GSList *) p->ptr;
	if (g)
	{
		n = g_slist_length (g);
		assert (n <= QU_GC_MAX_ITEMS);
		a = qu_array_new (n);
		i = g;
		do
		{
			assert (i->data);
			qu_array_push_quick (a, qu_pointer_new (i->data));
			i = i->next;

		} while (i);

		if (f)
			g_list_free (g);
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub glist_make_gobjects (a: Array) -> Pointer
	Build a C{GList} from an Array of GObjects.

	@warn
		You must call f{g_list_free} after you have done with the result.

		Make sure you keep the Array alive as long as the C{GList} is alive.
		This function is intended to just pass the GList to a GTK function
		then free it.

	@exc
	.EArgValue
*/
static qu_pointer qu_gtk_glist_make_gobjects (qu_array a)
{
	GList      *g = NULL;
	long        n;
	qu_any     *o;
	qu_gobject  x;

	QU_ARG_TYPE (a, qu_array);
	n = a->len;
	o = a->objs;
	while (n--)
	{
		x = (qu_gobject) *o++;
		QU_ARG_TYPE (x, qu_gobject);
		g = g_list_append (g, x->ptr);
	}
	if (g)
		return qu_pointer_new (g);
	return QuPointerNULL;
}



/*) sub gslist_make_gobjects (a: Array) -> Pointer
	Build a C{GSList} from an Array of GObjects.

	Note that if the Array is empty then a NULL pointer is returned.

	@warn
		You must call f{g_slist_free} after you have done with the result.

		Make sure you keep the Array alive as long as the GSList is alive.
		This function is intended to just pass the GSList to a GTK function
		then free it.

	@exc
	.EArgValue
*/
static qu_pointer qu_gtk_gslist_make_gobjects (qu_array a)
{
	GSList     *g = NULL;
	long        n;
	qu_any     *o;
	qu_gobject  x;

	QU_ARG_TYPE (a, qu_array);
	n = a->len;
	o = a->objs;
	while (n--)
	{
		x = (qu_gobject) *o++;
		QU_ARG_TYPE (x, qu_gobject);
		g = g_slist_append (g, x->ptr);
	}
	if (g)
		return qu_pointer_new (g);
	return QuPointerNULL;
}



/*) sub glist_get_strings (p: Pointer, free = true) -> Array
	Get an Array of strings contained in a C{GList}.

	Thus v{p} must point to the C{GList} and the v{data} field is C{char *}.

	If v{free} is true then this function calls f{g_free} to free individual strings
	and f{g_list_free} to free the C{GList}, otherwise you must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_glist_get_strings (int argc, qu_any *argv)
{
	qu_pointer p;
	int        f = 1;
	qu_array   a = NULL;
	GList     *g, *i;
	long       n;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, p, qu_pointer);
	if (argc > 1)
		QU_ARG_vbool (1, f);

	g = (GList *) p->ptr;
	if (g)
	{
		n = g_list_length (i);
		if (n)
		{
			assert (n <= QU_GC_MAX_ITEMS);
			a = qu_array_new (n);
			i = g;
			while (i)
			{
				qu_array_push_quick (a, qu_string_newz ((char *) i->data));
				if (f)
					g_free (i->data);
				i = i->next;
			}
			if (f)
				g_list_free (g);
		}
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub garray_get_gobjects (p: Pointer, n = -1, free = true) -> Array
	Get an Array of GObjects contained in an C array.

	v{p} must be a Pointer to the array of pointers.

	v{n} is the number of items. It may be negative to indicate that the array
	is terminated by a NULL pointer.

	If v{free} is true then call C{g_free} otherwise you must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_garray_get_gobjects (int argc, qu_any *argv)
{
	qu_pointer   p;
	long         n = -1;
	int          f = 1;
	qu_array     a = NULL;
	GObject    **g, **z, *x;
	qu_gobject   o;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, p, qu_pointer);
	if (argc > 1)
	{
		QU_ARG_vlong (1, n);
		if (argc > 2)
			QU_ARG_vbool (2, f);
	}
	g = (GObject **) p->ptr;
	if (g)
	{
		if (n >= 0)
		{
			assert (n <= QU_GC_MAX_ITEMS);
			a = qu_array_new (n);
		}
		else
		{
			a = qu_array_new (4);
			n = QU_GC_MAX_ITEMS;
		}

		z = g;
		for (x = *z; x && n; x = *++z, n--)
		{
			o = qu_gtk_from_gobject (x);
			QU_ENSURE (o);
			qu_array_push_quick (a, o);
		}
		if (f)
			g_free (g);
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub garray_get_strings (p: Pointer, n = -1, free = true) -> Array
	Get an Array of string pointers contained in an C string array.

	v{p} must be a Pointer that contains a C pointer which points to the
	string array.

	v{n} is the number of items. It may be negative to indicate that the array is
	terminated by a NULL pointer.

	If v{free} is true then this function calls C{g_strfreev} to free the array
	and all strings, otherwise you must do it yourself.

	Note that v{p} may be a NULL pointer, in which case an empty Array is returned
	no matter what v{n} is.

	@exc
	.EArgValue
*/
static qu_array qu_gtk_garray_get_strings (int argc, qu_any *argv)
{
	qu_pointer p;
	long       n = -1;
	int        f = 1;
	char     **g, **z, *s;
	qu_array   a = NULL;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, p, qu_pointer);
	if (argc > 1)
	{
		QU_ARG_vlong  (1, n);
		if (argc > 2)
			QU_ARG_vbool (2, f);
	}

	g = *((char ***) p->ptr);
	if (g)
	{
		if (n < 0)
		{
			a = qu_array_new (4);
			z = g;
			for (s = *z; s; s = *++z)
				QU_ENSURE (qu_array_push (a, qu_string_newz (s)));
		}
		else
		{
			QU_MEM_CHECK (n <= QU_GC_MAX_ITEMS);
			a = qu_array_new (n);
			z = g;
			while (n--)
			{
				s = *z++;
				assert (s);
				qu_array_push_quick (a, qu_string_newz (s));
			}
		}
		if (f)
			g_strfreev (g);
	}
	if (a)
		return a;
	return qu_array_new (0);
}



/*) sub ptr_to_string (p: Pointer, defval = '', free = true) -> Any
	Get a String from a pointer returned by a GTK function.

	If the pointer is NULL then return defval.
	If v{free} is true then call C{g_free} to free the string.

	@exc
	.EArgValue
*/
static qu_any qu_gtk_ptr_to_string (int argc, qu_any *argv)
{
	qu_pointer p;
	qu_any     d = QuStringEMPTY;
	int        f = 0;
	char      *z;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, p, qu_pointer);
	if (argc > 1)
	{
		d = argv[1];
		if (argc > 2)
			QU_ARG_vbool (2, f);
	}
	z = (char *) p->ptr;
	if (z)
	{
		d = (qu_any) qu_string_newz (z);
		if (f)
			g_free (z);
	}
	return d;
}



/*============================================================================

	things not possible from Qu

============================================================================*/
#include "GtkMissing.c"
/*) include GtkMissing.c */



/*============================================================================

	class GObject

============================================================================*/
/*) class GObject __abstract
	This class wraps the GObject class.

	The correct way to implement a subclass is to define a f{new}
	method and create the instance via the #{__inst} method.
*/
/* see Dlib.h for definitions */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	standard class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub class __inst (p: Pointer) -> GObject __final
	Call this from your Qu class to create a new Qu GObject from a GObject
	returned by a GTK function.

	For GObject, this method assumes that you own the reference.
	For GtkObject, f{g_object_ref} is called.
*/
static qu_gobject qu_gobject__inst (qu_class self, qu_pointer p)
{
	qu_gobject  o;
	GObject    *g;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr);

	g = (GObject *) p->ptr;

	/* see if we already have a Qu object associated with this object*/
	o = g_object_get_data (g, QU_GTK_KEY);
	if (o)
		return o;

	/* nope, so set up a new one */
	o = qu_gc_new (self);
	o->ptr = p->ptr;
	o->callbacks = NULL;
	g_object_set_data_full (g, QU_GTK_KEY, o, qu_gtk__destroy_notify);
	g_object_ref (g);

	if (qu_class_isa (self, QuGtkObjectClass))
		gtk_object_sink (g);

	return o;
}



/*) sub class new (p: Pointer) -> GObject
	This default method simply calls #{__inst} then the f{__init}
	method if one is defined then return the new instance.
*/
static qu_gobject qu_gobject_new (qu_class self, qu_any p)
{
	qu_gobject o = qu_gobject__inst (self, p);
	QU_ENSURE (o);
	return qu_class_init_new (self, o, 1, &p);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	low-level instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static void qu__gobject__mark (qu_gobject self)
{
	qu_gc_xmark (self->callbacks);
}


static void qu__gobject__destroy (qu_gobject self)
{
	if (self->ptr)
	{
		if (QU_APP_DEBUG)
			fprintf (stderr, "Qu GObject destroyed : %s %p\n", self->klass->name->str, self->ptr);
		g_object_steal_data (self->ptr, QU_GTK_KEY);
		g_object_unref (self->ptr);
		self->ptr = 0;
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	specific instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub ptr -> Pointer
	Get the GTK object as a pointer.

	@warn
	Be careful when you use this method.
	Call only if you have to build some struct to pass to a GTK function.

	@exc
	.ENone the associated GObject has been destroyed
*/
static qu_pointer qu_gobject_ptr (qu_gobject self)
{
	QU_REQUIREZ (QuENone, self->ptr, "GObject is gone");
	return qu_pointer_new (self->ptr);
}



/*) sub gtype -> Integer
	Get the GType of the associated GObject class.

	@exc
	.ECall GObject already gone
*/
static qu_integer qu_gobject_gtype (qu_gobject self)
{
	GObject *o = self->ptr;
	GType    t;

	QU_REQUIREZ (QuECall, o, "GObject already gone");
	t = G_TYPE_FROM_INSTANCE (o);
	return qu_integer_newul (t);
}



/*) sub connect (sig: String, handler: Sub) -> self __final
	Connect a GTK object with a handler.

	See the %{Gtk} module documentation for an example.

	@exc
	.EArgValue
*/
#include "GtkSignal.c"
static qu_gobject qu_gobject_connect (qu_gobject self, qu_string sig, qu_sub u)
{
	GObject     *g = self->ptr;
	qu_callback  h;
	void        *func;
	qu_dict      callbacks = self->callbacks;

	QU_ARG_CHECK (qu_string_is (sig));
	QU_REQUIREZ (QuECall, g, "GObject already gone");

	QU_ENSURE (qu_callback_validate (u));
	func = qu_gtk_signal_getfunc (u);

	h = qu_gtk_callback_alloc (self, u);
	assert (h);

	if (callbacks)
	{
		qu_callback old = qu_dict_getitem (callbacks, sig);
		if (old)
			g_signal_handler_disconnect (g, old->id);
	}
	else
	{
		callbacks = self->callbacks = qu_dict_new (4);
	}

	qu_dict_setitem (callbacks, sig, h);
	h->id = g_signal_connect_data (g, sig->str, func, h, 0, G_CONNECT_SWAPPED);
	return self;
}



/*) sub disconnect (sig: String) -> self __final
	Disconnect handler previously installed with #{connect}.

	@exc
	.EArgValue
*/
static qu_gobject qu_gobject_disconnect (qu_gobject self, qu_string sig)
{
	qu_dict  callbacks = self->callbacks;

	QU_ARG_CHECK (qu_string_is (sig));

	if (callbacks)
	{
		qu_callback h = qu_dict_getitem (callbacks, sig);
		if (h)
		{
			GObject *g = self->ptr;

			if (g)
				g_signal_handler_disconnect (g, h->id);

			qu_dict_del (callbacks, sig);
			if (callbacks->len == 0)
				self->callbacks = 0;
		}
	}
	return self;
}



/*) sub hook (handler: Sub, name = handler.ident, theself = self) -> Array __final
	Create a callback which the last argument is user data.

	v{name} must be a unique name.

	The returned value is an Array of two items:
	@ul
	.a Pointer which must be passed as the callback function to the GTK function.
	.an Opaque object which must be passed to the GTK function as user data

	@warn
		The Qu callback function must not include the last user data argument.
		In the function, self will be v{theself}.

	@xmp
		sub myhook (arg1: Cint, user_data: Any)  #-> WRONG!
		sub myhook (arg1: Cint)                  #-> CORRECT

	@exc
	.EArgValue
*/
static qu_array qu_gobject_hook (qu_gobject self, int argc, qu_any *argv)
{
	qu_string    name;
	qu_any       theself = self;
	qu_sub       u;
	GObject     *g = self->ptr;
	void        *f;
	qu_callback  h;
	qu_dict      callbacks = self->callbacks;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, u, qu_sub);
	if (argc > 1)
	{
		QU_ARG_VTYPE (1, name, qu_string);
		if (argc > 2)
			theself = argv [2];
	}
	else
	{
		name = u->name;
	}

	QU_REQUIREZ (QuECall, g, "GObject already gone");
	QU_ENSURE (qu_callback_validate (u));
	f = qu_gtk_hook_getfunc (u);
	assert (f);
	h = qu_gtk_callback_alloc (theself, u);
	assert (h);

	if (callbacks == 0)
		self->callbacks = callbacks = qu_dict_new (4);

	qu_dict_setitem (callbacks, name, h);
	return qu_array_new_ex (2, qu_pointer_new (f), h);
}



/*) sub callback (handler: Sub, name = handler.ident, theself = self) -> Pointer __final
	Create a callback which the first argument is user data.

	v{name} must be a unique name.

	The returned value is an Array of two items:
	@ul
	.a Pointer which must be passed as the callback function to the GTK function.
	.an Opaque object which must be passed to the GTK function as user data

	@warn
		The Qu callback function must not include the first user data argument.
		In the function, self will be v{theself}.

	@xmp
		sub myhook (user_data: Any, arg1: Cint)  #-> WRONG!
		sub myhook (arg1: Cint)                  #-> CORRECT

	@exc
	.EArgValue
*/
#include "GtkCallback.c"
static qu_array qu_gobject_callback (qu_gobject self, int argc, qu_any *argv)
{
	qu_string    name;
	qu_any       theself = self;
	qu_sub       u;
	GObject     *g = self->ptr;
	void        *f;
	qu_callback  h;
	qu_dict      callbacks = self->callbacks;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, u, qu_sub);
	if (argc > 1)
	{
		QU_ARG_VTYPE (1, name, qu_string);
		if (argc > 2)
			theself = argv [2];
	}
	else
	{
		name = u->name;
	}

	QU_REQUIREZ (QuECall, g, "GObject already gone");
	QU_ENSURE (qu_callback_validate (u));
	f = qu_gtk_callback_getfunc (u);
	assert (f);
	h = qu_gtk_callback_alloc (theself, u);
	assert (h);

	if (callbacks == 0)
		self->callbacks = callbacks = qu_dict_new (4);

	qu_dict_setitem (callbacks, name, h);
	return qu_array_new_ex (2, qu_pointer_new (f), h);
}



/*) sub closure (f: Sub, name = f.ident) -> Any
	Create a Qu closure to be called by GTK.

	v{name} must be a unique indentifier.

	The returned value is a Pointer to the GClosure object that you must pass to the
	GTK function.
*/
static qu_pointer qu_gobject_closure (qu_gobject self, int argc, qu_any *argv)
{
	qu_sub       u;
	qu_string    name;
	qu_callback  h;
	GClosure    *c;
	qu_dict      callbacks = self->callbacks;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, u, qu_sub);
	if (argc > 1)
		QU_ARG_VTYPE (1, name, qu_string);
	else
		name = u->name;

	QU_ENSURE (qu_callback_validate (u));
	h = qu_gtk_callback_alloc (self, u);
	assert (h);
	c = g_cclosure_new (qu_gtk_fake_callback, NULL, NULL);
	QU_REQUIREZ (QuESystem, c, "g_cclosure_new failed");
	g_closure_set_meta_marshal (c, h, qu_gtk_marshal);
	if (callbacks == 0)
		callbacks = self->callbacks = qu_dict_new (4);
	qu_dict_setitem (callbacks, name, h);
	return qu_pointer_new (c);
}



/*) sub getprop (name: String, type) -> self
	Get property by name.

	This is a wrapper around f{g_object_get_property}.

	See #{gvalue_init} for valid v{type} value.

	@exc
	.ECall      GObject already gone
	.EArgValue
*/
static qu_any qu_gobject_getprop (qu_gobject self, qu_string name, qu_any type)
{
	GValue g;
	GType  t;

	QU_REQUIREZ (QuECall, self->ptr, "GObject already gone");
	QU_ARG_TYPE (name, qu_string);

	memset (&g, 0, sizeof (g));
	t = qu_gtk_type_qu2g (type);
	if (t)
	{
		qu_any x;

		g_value_init (&g, t);
		g_object_get_property (self->ptr, name->str, &g);
		x = qu_gtk_v2q (&g);
		if (x)
			return x;
	}
	QU_THROWF (QuEImplement, "property '%s' value type not supported", name->str);
}



/*) sub setprop (name: String, val: Any, type = nil) -> Any
	Set property by name.

	This is a wrapper around f{g_object_set_property}.

	See #{gvalue_init} for valid v{type} value.

	If v{type} is not specified then a guess is made so you better make sure v{val}
	is valid.

	@exc
	.ECall      GObject already gone
	.EArgValue
*/
static qu_any qu_gobject_setprop (qu_gobject self, int argc, qu_any *argv)
{
	qu_string name;
	qu_any    x;
	GValue    g;
	GType     t;

	QU_REQUIREZ (QuECall, self->ptr, "GObject already gone");
	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, name, qu_string);
	x = argv[1];

	memset (&g, 0, sizeof (g));

	if (argc > 2)
		t = qu_gtk_type_qu2g (x);
	else
		t = qu_gtk_gtype_guess (x);

	if (t)
	{
		g_value_init (&g, t);
		if (qu_gtk_q2v (&g, x))
		{
			g_object_set_property (self->ptr, name->str, &g);
			return self;
		}
	}
	QU_THROWF (QuEImplement, "property '%s' value type not supported", name->str);
}



/*) sub __ref -> self
 	Increment the reference count. This simply calls f{g_object_ref}.

	If the GObject is already gone then nothing happens.
*/
static qu_gobject qu_gobject__ref (qu_gobject self)
{
	GObject *g = self->ptr;

	if (g)
		g_object_ref (g);
	return self;
}



/*) sub __unref -> self
 	Decrement the reference count. This simply calls f{g_object_unref}.

	If the GObject is already gone then nothing happens.
*/
static qu_gobject qu_gobject__unref (qu_gobject self)
{
	GObject *g = self->ptr;

	if (g)
		g_object_unref (g);
	return self;
}



/*) sub __refcount -> Integer
	Returns the reference count on a GObject.

	@warn
	This is for debugging only.
*/
static qu_integer qu_gobject__refcount (qu_gobject self)
{
	struct _GObject
	{
		void     *g_type_instance;
		volatile  unsigned int ref_count;
		void     *qdata;
	};
	QU_REQUIRE (QuENone, self->ptr);
	return qu_integer_newui (((struct _GObject *) self->ptr)->ref_count);
}



/*============================================================================

	class Object

============================================================================*/
/*) class Object is GObject __abstract
	This is the GtkObject class.

	@warn
	Unlike GObject, you do not own a reference to the GtkObject. When the
	GtkObject is destroyed then you can still access your Qu Object but
	you can not access the GtkObject.

	@p
	The Gtk module makes sure that your Qu object lives as long as the
	GtkObject lives but no the other way around.
*/
/* static qu_class QuGtkObjectClass; */



/*============================================================================

	class EGError

============================================================================*/
/*) class EGError is ERuntime
	This is the exception that should be thrown by a GError.
*/
/* static qu_class QuEGErrorClass; */



/*============================================================================

	init

============================================================================*/
static qu_any qu_gtk_getcfunc (qu_module self, qu_string name, qu_string spec)
{
	qu_any m = qu_dict_getitem (AllMethods, name);

	if (m == 0)
	{
		void *h = qu_dl_sym (QuGtkHandle, name->str);

		if (h == 0)
			QU_THROWF (QuEImplement, "symbol %s not found", name->str);
		m = qu_dlib_cfunc_static (h, name, spec);
		if (m == 0)
			QU_THROWF (QuEValue, "%s: %s\n", name->str, qu_thread_err_string ()->str);
		qu_dict_setitem (AllMethods, qu_string_to_static (name), m);
	}
	return m;
}


static void qu_gtk_atexit ()
{
	if (QuGtkHandle)
		qu_dl_close (QuGtkHandle);

	if (QuGthreadHandle)
		qu_dl_close (QuGthreadHandle);
}


static int qu_gtk_failed (const char *z)
{
	qu_gtk_atexit ();
	QU_THROWF (QuESystem, "cant initialize Gtk missing symbol %s", z);
}



int qu_Gtk_init (qu_module m)
{
#ifdef DEBUG
	fprintf (stderr, "sizeof (GValue) = %d\n", sizeof (GValue));
#endif

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		module
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuGtkModule = m;

	QuGtkHandle     = qu_dl_open_ex ("gtk"    , QU_DL_GLOBAL | QU_DL_NOW);
	QuGthreadHandle = qu_dl_open_ex ("gthread", QU_DL_GLOBAL | QU_DL_NOW);

	if (! (QuGtkHandle && QuGthreadHandle))
	{
		qu_gtk_atexit ();
		QU_RETHROW;
	}

	/* initialize thread support */
	{
		static void  (*g_thread_init   ) (void *ignore);
		static void  (*gdk_threads_init) ();

		g_thread_init    = qu_dl_sym (QuGthreadHandle, "g_thread_init");
		gdk_threads_init = qu_dl_sym (QuGtkHandle, "gdk_threads_init");

		if (g_thread_init && gdk_threads_init)
		{
			g_thread_init (NULL);
			gdk_threads_init ();
		}
		else
		{
			QU_SYS_WARN ("cant initialize GTK thread support");
		}
	}

	m->getcfunc = qu_gtk_getcfunc;

#define f_(name) \
	name = qu_dl_sym (QuGtkHandle, #name); \
	if (name == 0) \
		return qu_gtk_failed ( #name);

	f_(gtk_init);
	f_(g_object_ref);
	f_(g_object_unref);
	f_(g_object_get_data);
	f_(g_object_set_data);
	f_(g_object_set_data_full);
	f_(g_object_steal_data);
	f_(g_signal_connect_data);
	f_(g_signal_handler_disconnect);
	f_(g_strdup);
	f_(g_type_name);
	f_(gtk_object_sink);
	f_(gtk_main_quit);
	f_(gtk_main);

	f_(g_value_init);
	f_(g_value_unset);
	f_(g_value_set_char);
	f_(g_value_get_char);
	f_(g_value_set_uchar);
	f_(g_value_get_uchar);
	f_(g_value_set_boolean);
	f_(g_value_get_boolean);
	f_(g_value_set_int);
	f_(g_value_get_int);
	f_(g_value_set_uint);
	f_(g_value_get_uint);
	f_(g_value_set_long);
	f_(g_value_get_long);
	f_(g_value_set_ulong);
	f_(g_value_get_ulong);
	f_(g_value_set_float);
	f_(g_value_get_float);
	f_(g_value_set_double);
	f_(g_value_get_double);
	f_(g_value_set_string);
	f_(g_value_get_string);
	f_(g_value_set_pointer);
	f_(g_value_get_pointer);
	f_(g_value_set_object);
	f_(g_value_get_object);
	f_(g_object_get_property);
	f_(g_object_set_property);

	f_(g_cclosure_new);
	f_(g_closure_set_meta_marshal);

	f_(g_free);
	f_(g_slist_free);
	f_(g_slist_append);
	f_(g_slist_length);
	f_(g_list_free);
	f_(g_list_append);
	f_(g_list_length);
	f_(g_strfreev);

	f_(g_error_free);

	gtk_init (0);

	/* see Dlib.h */
	__QuGObjectNew = qu_gtk_from_gobject;

	qu_gtk_missing_init (m);

	QuGtkMarshallPtr = qu_pointer_new_static (qu_gtk_marshal);

	/* strings are static, classes don't need marking */
	AllClasses = qu_dict_new_static (512);

	/* strings are static, values are static Methods */
	AllMethods = qu_dict_new_static (512);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Core
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "main"      , qu_gtk_main      ,  0);
	qu_module_add_function (m, "quit"      , qu_gtk_quit      ,  0);
	qu_module_add_function (m, "register"  , qu_gtk_register  , -1);
	qu_module_add_function (m, "call"      , qu_gtk_call      , -1);
	qu_module_add_function (m, "gobject"   , qu_gtk_gobject   , -1);
	qu_module_add_function (m, "gobjectptr", qu_gtk_gobjectptr, -1);
	qu_module_add_function (m, "closure"   , qu_gtk_closure   , -1);

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		GError
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuEGErrorClass = qu_exception_class_new (m, QuERuntimeClass, "EGError",
		"This is the exception that should be thrown by a GError.");

	qu_module_add_function (m, "gerror_throw", qu_gerror_throw, 1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		GValue
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "gvalue_size"  , qu_gtk_gvalue_size  ,  0);
	qu_module_add_function (m, "gvalue_new"   , qu_gtk_gvalue_new   ,  0);
	qu_module_add_function (m, "gvalue_newv"  , qu_gtk_gvalue_newv  , -1);
	qu_module_add_function (m, "gvalue_init"  , qu_gtk_gvalue_init  ,  2);
	qu_module_add_function (m, "gvalue_unset" , qu_gtk_gvalue_unset ,  1);
	qu_module_add_function (m, "gvalue_get"   , qu_gtk_gvalue_get   ,  1);
	qu_module_add_function (m, "gvalue_set"   , qu_gtk_gvalue_set   ,  2);
	qu_module_add_function (m, "gvalue_type"  , qu_gtk_gvalue_type  ,  1);
	qu_module_add_function (m, "gvalue_gtype" , qu_gtk_gvalue_gtype ,  1);
	qu_module_add_function (m, "gvalue_array" , qu_gtk_gvalue_array , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		type stuffs
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "gtypeof" , qu_gtk_gtypeof,  1);
	qu_module_add_function (m, "typeof"  , qu_gtk_typeof ,  1);
	qu_module_add_function (m, "args"    , qu_gtk_args   , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		callbacks
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "hook"    , qu_gtk_hook    , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		helpers
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "glist_get_gobjects"  , qu_gtk_glist_get_gobjects  , -1);
	qu_module_add_function (m, "glist_get_pointers"  , qu_gtk_glist_get_pointers  , -1);
	qu_module_add_function (m, "gslist_get_gobjects" , qu_gtk_gslist_get_gobjects , -1);
	qu_module_add_function (m, "gslist_get_pointers" , qu_gtk_gslist_get_pointers , -1);
	qu_module_add_function (m, "glist_make_gobjects" , qu_gtk_glist_make_gobjects ,  1);
	qu_module_add_function (m, "gslist_make_gobjects", qu_gtk_gslist_make_gobjects,  1);
	qu_module_add_function (m, "glist_get_strings"   , qu_gtk_glist_get_strings   , -1);
	qu_module_add_function (m, "garray_get_gobjects" , qu_gtk_garray_get_gobjects , -1);
	qu_module_add_function (m, "garray_get_strings"  , qu_gtk_garray_get_strings  , -1);
	qu_module_add_function (m, "ptr_to_string"       , qu_gtk_ptr_to_string       , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class GObject
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
#if 0
	QuGObjectClass = qu_class_new_abstractz (m, "GObject", qu_gobject_s,
		qu_gobject__mark, qu_gobject__destroy);
#else
	qu_class_setup (QuGObjectClass, qu_gobject_s, qu__gobject__destroy, qu__gobject__mark);
	qu_flag_set (QuGObjectClass, QU_CLASS_FLAG_EXTERN | QU_CLASS_FLAG_ABSTRACT);
	qu_module_setattr (m, QuGObjectClass->name, (qu_symbol) QuGObjectClass);
#endif

	qu_class_method_class  (QuGObjectClass, QU_NAME__inst , qu_gobject__inst      ,  1);
	qu_class_method_class  (QuGObjectClass, QU_NAME_new   , qu_gobject_new        ,  1);
	qu_class_method_instzf (QuGObjectClass, "gtype"       , qu_gobject_gtype      ,  0);
	qu_class_method_instzf (QuGObjectClass, "connect"     , qu_gobject_connect    ,  2);
	qu_class_method_instzf (QuGObjectClass, "disconnect"  , qu_gobject_disconnect ,  1);
	qu_class_method_instzf (QuGObjectClass, "hook"        , qu_gobject_hook       , -1);
	qu_class_method_instzf (QuGObjectClass, "callback"    , qu_gobject_callback   , -1);
	qu_class_method_instzf (QuGObjectClass, "closure"     , qu_gobject_closure    , -1);
	qu_class_method_instzf (QuGObjectClass, "ptr"         , qu_gobject_ptr        ,  0);
	qu_class_method_instzf (QuGObjectClass, "getprop"     , qu_gobject_getprop    ,  2);
	qu_class_method_instzf (QuGObjectClass, "setprop"     , qu_gobject_setprop    , -1);
	qu_class_method_instzf (QuGObjectClass, "__ref"       , qu_gobject__ref       ,  0);
	qu_class_method_instzf (QuGObjectClass, "__unref"     , qu_gobject__unref     ,  0);
	qu_class_method_instzf (QuGObjectClass, "__refcount"  , qu_gobject__refcount  ,  0);

	qu_class_enddecl (QuGObjectClass);
	qu_module_setattr (QuSysModule, QuGObjectClass->name, (qu_symbol) QuGObjectClass);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Object
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuGtkObjectClass = qu_class_new_abstractz (m, "Object", qu_gobject_s,
		qu__gobject__destroy, qu__gobject__mark);

	qu_class_set_super (QuGtkObjectClass, QuGObjectClass);
	qu_class_enddecl   (QuGtkObjectClass);

	atexit (qu_gtk_atexit);
	return 1;
}

#endif /*QU_SUPPORT_GLIB*/
