/* GtkMissing.c - Copyright (c) Marc Krisnanto */
/*
	For stuffs that are just impossible:
	- callbacks which are impossible without Proc-ing
	- other stuffs
*/


/*---------------------------------------------------------------------------
	ColorSelection
---------------------------------------------------------------------------*/

static qu_sub QuColorSelectionChangePaletteWithScreenFunc = NULL;


static void qu_color_selection_set_change_palette_with_screen_hook (GObject *screen, void *colors, int ncolors)
{
	if (QuColorSelectionChangePaletteWithScreenFunc)
	{
		ENTER;
		if (qu_sub_call3 (QuColorSelectionChangePaletteWithScreenFunc,
				QuNil,
				qu_gtk_from_gobject (screen),
				qu_pointer_new (colors),
				qu_integer_newi (ncolors)) == NULL)
		{
			QU_SYS_WARN ("ColorSelection screen hook threw an exception");
		}
		LEAVE;
	}
}


/*) sub class color_selection_set_change_palette_with_screen_hook (f)
	Called by Selection.color_selection_set_change_palette_with_screen_hook.

	The callback signature is:
	@code
		sub callback (s: Screen, colors: Pointer, n_colors: Cint)
			# self is nil
			# to get an Array of Color objects do this:
			colors = Color.array (colors, n_colors)
*/
static qu_pointer qu_gtk_color_selection_set_change_palette_with_screen_hook (qu_sub f)
{
	QuColorSelectionChangePaletteWithScreenFunc = f;
	return qu_pointer_new (qu_color_selection_set_change_palette_with_screen_hook);
}



/*---------------------------------------------------------------------------
	GtkClipboard
---------------------------------------------------------------------------*/

static void qu_gtk_clipboard_get_func (void *clipboard, void *selection_data, unsigned int info, qu_callback x)
{
	ENTER;
	if (qu_sub_call2 (x->sub,
			qu_gtk_from_gobject (clipboard),
			qu_pointer_new (selection_data),
			qu_integer_newui (info)) == NULL)
	{
		QU_SYS_WARN ("Clipboard get hook threw an exception");
	}
	LEAVE;
}


static void qu_gtk_clipboard_clear_func (void *clipboard, qu_callback x)
{
	ENTER;
	if (qu_sub_call0 ((qu_sub) x->self, qu_gtk_from_gobject (clipboard)))
		QU_SYS_WARN ("Clipboard get hook threw an exception");
	LEAVE;
}


/*) sub clipboard_set_with_data (o: GObject, f1: Sub, f2: Sub, name = f1.ident) -> Array
	Called by Clipboard.set_with_data.

	Return an Array of 3 items:
	@ul
	.Pointer to pass as get_func
	.Pointer to pass as clear_func
	.Opaque object to pass as user_data

	@p
	The get-function v{f1} signature:
	@code
		sub get_func (selection_data: Pointer, info: Cuint)
        	# self is the Clipboard object

			# to get the Qu SelectionData object:
			s = Selection.SelectionData.acquire (selection_data)
		;;


	@p
	The clear-function v{f1} signature:
	@code
		sub clear_func
        	# self is the Clipboard object
		;;
*/
static qu_array qu_gtk_clipboard_set_with_data (int argc, qu_any *argv)
{
	qu_gobject   self;
	qu_sub       u1, u2;
	qu_string    name;
	GObject     *g;
	qu_callback  h;

	QU_ARG_VCOUNT (3, 4);
	QU_ARG_VTYPE  (0, self, qu_gobject);

	g = self->ptr;
	QU_REQUIREZ (QuECall, g, "GObject already gone");

	QU_ARG_VTYPE (1, u1, qu_sub);
	QU_ARG_VTYPE (2, u2, qu_sub);
	if (argc > 3)
		QU_ARG_VTYPE (3, name, qu_string);
	else
		name = u1->name;

	QU_ENSURE (qu_callback_validate (u1));
	QU_ENSURE (qu_callback_validate (u2));

	h = qu_gtk_callback_alloc (u2, u1);
	if (self->callbacks == 0)
		self->callbacks = qu_dict_new (4);

	qu_dict_setitem (self->callbacks, name, h);
	return qu_array_new_ex (3,
		qu_pointer_new (qu_gtk_clipboard_get_func),
		qu_pointer_new (qu_gtk_clipboard_clear_func),
		h);

}


/*---------------------------------------------------------------------------
	GtkMenu
---------------------------------------------------------------------------*/

static void qu_gtk_menu_detach_func (void *widget, void *menu)
{
	qu_gobject w, m;

	ENTER;
	w = qu_gtk_from_gobject (widget);
	m = qu_gtk_from_gobject (menu);
	if (m && w)
	{
		qu_string s = qu_string_newz ("__detach_func");
		qu_sub    u = qu_object_getattr (m, s, 0);

		if (u && !qu_sub_is_inst (u))
		{
			if (qu_sub_call1 (u, m, w) == NULL)
				QU_SYS_WARN ("Menu detach hook threw an exception");
		}
	}
	LEAVE;
}


/*) sub menu_attach_to_widget -> Pointer
	Called by Menu.attach_to_widget.

	Return the pointer to pass as the callback.

	The caller must set the instance attribute C{__detach_func} to the callback Sub
	to call.

	@p
	The callback must have this signature:
	@code
		sub callback (w: Widget)
        	# self is the Menu object
		;;
*/
static qu_pointer qu_gtk_menu_attach_to_widget ()
{
	return qu_pointer_new (qu_gtk_menu_detach_func);
}



/*---------------------------------------------------------------------------
	PangoLogAttr
---------------------------------------------------------------------------*/

/*) sub pango_logattr_size -> Int
	Called by LogAttr.
*/
static qu_int qu_gtk_pango_logattr_size ()
{
	long l = sizeof (PangoLogAttr);
	return qu_int_new (l);
}



/*) sub pango_logattr_getitem (p: Pointer, idx: Integer) -> Boolean
	Called by LogAttr.
*/
static qu_boolean qu_gtk_pango_logattr_getitem (qu_pointer p, qu_int idx)
{
	PangoLogAttr *a;
	long          i;
	unsigned int  v;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr);
	QU_ARG_long (idx, i);
	a = p->ptr;
	switch (i)
	{
		case  0: v = a->is_line_break;               break;
		case  1: v = a->is_mandatory_break;          break;
		case  2: v = a->is_char_break;               break;
		case  3: v = a->is_white;                    break;
		case  4: v = a->is_cursor_position;          break;
		case  5: v = a->is_word_start;               break;
		case  6: v = a->is_word_end;                 break;
		case  7: v = a->is_sentence_boundary;        break;
		case  8: v = a->is_sentence_start;           break;
		case  9: v = a->is_sentence_end;             break;
		case 10: v = a->backspace_deletes_character; break;

		default: QU_THROW (QuEIndex);
	}
	return v ? QuTrue : QuFalse;
}



/*) sub pango_logattr_setitem (p: Pointer, idx: Integer, val: Boolean) -> Boolean
	Called by LogAttr.
*/
static qu_boolean qu_gtk_pango_logattr_setitem (qu_pointer p, qu_int idx, qu_boolean val)
{
	PangoLogAttr *a;
	long          i;
	unsigned int  v = (val == QuTrue);

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr);
	QU_ARG_long (idx, i);

	a = p->ptr;
	switch (i)
	{
		case  0: a->is_line_break               = v; break;
		case  1: a->is_mandatory_break          = v; break;
		case  2: a->is_char_break               = v; break;
		case  3: a->is_white                    = v; break;
		case  4: a->is_cursor_position          = v; break;
		case  5: a->is_word_start               = v; break;
		case  6: a->is_word_end                 = v; break;
		case  7: a->is_sentence_boundary        = v; break;
		case  8: a->is_sentence_start           = v; break;
		case  9: a->is_sentence_end             = v; break;
		case 10: a->backspace_deletes_character = v; break;

		default: QU_THROW (QuEIndex);
	}
	return val;
}



/*) sub pango_layout_get_log_attrs (p: Pointer, n: Integer) -> Array
	Called by Layout.

	@warn
	All items in the returned Array must be acquireed with C{own = false}.
*/
static qu_array qu_gtk_pango_layout_get_log_attrs (qu_pointer p, qu_integer l)
{
	PangoLogAttr *a;
	long          n;
	qu_array      r;

	QU_ARG_long  (l, n);
	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr && n > 0);
	assert (n <= QU_GC_MAX_ITEMS);
	r = qu_array_new (n);
	a = p->ptr;
	while (n--)
	{
		p = qu_pointer_new_alloc (sizeof (PangoLogAttr));
		memcpy (p->ptr, a++, sizeof (PangoLogAttr));
		qu_array_push_quick (r, p);
	}
	return r;
}



/*---------------------------------------------------------------------------
	PangoGlyphVisAttr
---------------------------------------------------------------------------*/

/*) sub pango_glyph_visattr_size -> Int
	Called by GlyphVisAttr.
*/
static qu_int qu_gtk_pango_glyph_visattr_size ()
{
	long l = sizeof (PangoGlyphVisAttr);
	return qu_int_new (l);
}



/*) sub pango_glyph_visattr_getitem (p: Pointer, idx: Integer) -> Boolean
	Called by GlyphVisAttr.
*/
static qu_boolean qu_gtk_pango_glyph_visattr_getitem (qu_pointer p, qu_int idx)
{
	PangoGlyphVisAttr *a;
	long               i;
	unsigned int       v;

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr);
	QU_ARG_long (idx, i);
	a = p->ptr;
	switch (i)
	{
		case  0: v = a->is_cluster_start; break;
		default: QU_THROW (QuEIndex);
	}
	return v ? QuTrue : QuFalse;
}



/*) sub pango_glyph_visattr_setitem (p: Pointer, idx: Integer, val: Boolean) -> Boolean
	Called by GlyphVisAttr.
*/
static qu_boolean qu_gtk_pango_glyph_visattr_setitem (qu_pointer p, qu_int idx, qu_boolean val)
{
	PangoGlyphVisAttr *a;
	long               i;
	unsigned int       v = (val == QuTrue);

	QU_ARG_CHECK (qu_pointer_is (p) && p->ptr);
	QU_ARG_long (idx, i);

	a = p->ptr;
	switch (i)
	{
		case  0: a->is_cluster_start = v; break;
		default: QU_THROW (QuEIndex);
	}
	return val;
}



/*---------------------------------------------------------------------------
	init
---------------------------------------------------------------------------*/

void qu_gtk_missing_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		ColorSelection
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
	qu_module_add_function (m, "color_selection_set_change_palette_with_screen_hook",
			qu_gtk_color_selection_set_change_palette_with_screen_hook, 1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		GtkClipboard
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
	qu_module_add_function (m, "clipboard_set_with_data",
		qu_gtk_clipboard_set_with_data, -1);

	qu_module_add_function (m, "menu_attach_to_widget",
			qu_gtk_menu_attach_to_widget, 0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		PangoLogAttr
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
	qu_module_add_function (m, "pango_logattr_size",
			qu_gtk_pango_logattr_size, 0);

	qu_module_add_function (m, "pango_logattr_getitem",
			qu_gtk_pango_logattr_getitem, 2);

	qu_module_add_function (m, "pango_logattr_setitem",
			qu_gtk_pango_logattr_setitem, 3);

	qu_module_add_function (m, "pango_layout_get_log_attrs",
			qu_gtk_pango_layout_get_log_attrs, 2);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		PangoGlyphVisAttr
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

	qu_module_add_function (m, "pango_glyph_visattr_size",
			qu_gtk_pango_glyph_visattr_size, 0);

	qu_module_add_function (m, "pango_glyph_visattr_getitem",
			qu_gtk_pango_glyph_visattr_getitem, 2);

	qu_module_add_function (m, "pango_glyph_visattr_setitem",
			qu_gtk_pango_glyph_visattr_setitem, 3);
}
