/* Iconv.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Iconv
	This module provides an interface to the system I{iconv} library.

	See #{convert} for an example.
*/

#ifndef HAVE_ICONV_H


int qu_Iconv_init (qu_module m)
{
	QU_THROWZ ("Iconv is not available on your system"):
}


#else /*HAVE_ICONV_H*/


#include <iconv.h>


/* forward */
static qu_class QuEIconvClass;


#define QU_THROW_ICONV QU_BEGIN qu_iconv__throw (); QU_RETHROW; QU_END
static int qu_iconv__throw ()
{
	if (errno == EINVAL)
		QU_THROWZ (QuEIconv, "invalid multibyte sequence");

	else if (errno == EILSEQ)
		QU_THROWZ (QuEIconv, "incomplete multibyte sequence");

	else
		QU_THROWZ (QuEIconv, "unknown Iconv error");
}


/* !!! don't call if s->len == 0 */
static qu_string qu_iconv___convert (iconv_t c, qu_stringy s)
{
	char     *ibuf, *obuf;
	size_t    n, ileft, oleft;
	long      more;
	volatile qu_string r;

	assert (s->len);

	r = qu_string_buffer_new (s->len);

	ibuf  = s->str;
	obuf  = r->str;
	ileft = oleft = more = s->len;

	iconv (c, 0, 0, 0, 0);

  again:
	n = iconv (c, &ibuf, &ileft, &obuf, &oleft);
	if (n == (size_t) -1)
	{
		if (errno == E2BIG)
		{
			qu_string_buffer_set_len (r, obuf - r->str);
			QU_REQUIRE (QuEMemory, qu_string_buffer_make_space (r, more));
			obuf = r->str + r->len;
			oleft = more;
			goto again;
		}
		else
		{
			iconv_close (c);
			QU_THROW_ICONV;
		}
	}
	qu_string_buffer_set_len (r, obuf - r->str);
	return qu_string_buffer_return (r);
}



#define GENERIC_SUPERSET      "UCS-4LE"
#define GENERIC_SUPERSET_WLEN 4

static long qu_iconv__width (qu_stringy s, qu_string enc)
{
	iconv_t  c;
	char    *ibuf, *obuf, buf [GENERIC_SUPERSET_WLEN * 2 + 1];
	size_t   ileft, oleft, prev;
	long     n;

	__qu_blocking_begin;
	c = iconv_open (GENERIC_SUPERSET, enc->str);
	__qu_blocking_end;
	if (c == (iconv_t) -1)
	{
		if (errno == EINVAL)
			QU_THROWV (QuEImplement, -1);
		QU_THROWERRNOV (-1);
	}

	ibuf  = s->str;
	ileft = s->len;
	oleft = 0;
	n     = 0;
	errno = 0;

	for (; ileft > 0; n += 2)
	{
		obuf = buf;
		oleft = sizeof (buf);
		prev = ileft;
		if (iconv (c, &ibuf, &ileft, &obuf, &oleft) == (size_t) -1)
		{
			if (prev == ileft)
				break;
		}
	}
	if (oleft > 0)
		n -= oleft / GENERIC_SUPERSET_WLEN;
	if (errno == 0 || errno == E2BIG)
		return n;
	return -1;
}



/*============================================================================
	functions
============================================================================*/

/*) sub convert (s: String, fromcode: String, tocode: String) -> String
	Perform character encoding conversion v{fromcode} to v{tocode}.

	The values permitted for v{fromcode} and v{tocode} and the supported
	combinations are system dependent.

	Try C{iconv --list} command for supported values.

	@exc
	.EArgValue
	.EMemory
	.EImplement conversion not supported

	@xmp
		use Iconv

		s = 'Copyright Foo'
		r = Iconv.convert (s, '8859_1', 'UTF-8')
		println (r.quote)
		s = Iconv.convert (r, 'UTF-8', '8859_1')
		println (s)

	@out
		'Copyright \xc2\xa9Foo'
		Copyright Foo
*/
static qu_string qu_iconv_convert (qu_stringy s, qu_string from, qu_string to)
{
	iconv_t c;

	QU_ARG_TYPE (s, qu_stringy);
	QU_ARG_TYPE (to, qu_string);
	QU_ARG_TYPE (from, qu_string);

	if (s->len)
	{
		__qu_blocking_begin;
		c = iconv_open (to->str, from->str);
		__qu_blocking_end;
		if (c == (iconv_t) -1)
		{
			if (errno == EINVAL)
				QU_THROW (QuEImplement);
			QU_THROWERRNO;
		}
		s = qu_iconv___convert (c, s);
		iconv_close (c);
		QU_ENSURE (s);
	}
	return s;
}



/*) sub width (s: String, encoding: String) -> Integer
	Count the number of wide characters in a String with a given encoding.

	@exc
	.EConv      invalid characters in input
	.EImplement conversion not supported

	@xmp
		use Iconv

		s = 'Copyright Foo'
		r = Iconv.convert (s, '8859_1', 'UTF-8')

		println (Iconv.width (s, '8859_1')) # \_ should print the
		println (Iconv.width (r, 'UTF-8'))  # /  same result

	@out
		14
		14
*/
static qu_integer qu_iconv_width (qu_stringy s, qu_string enc)
{
	long n;

	QU_ARG_TYPE (s, qu_stringy);
	QU_ARG_TYPE (enc, qu_stringy);
	if (s->len == 0)
		return QuInteger0;
	n = qu_iconv__width (s, enc);
	QU_ENSURE (n > 0);
	return qu_integer_newl (n);
}



/*============================================================================
	class EIconv
============================================================================*/

/*) class EIconv
	This is the standard exception thrown by the Iconv library.
*/



/*============================================================================
	class Iconv
============================================================================*/

/*) class Iconv
	Use this class to convert a number of strings.

	This is more efficient than using the !{convert} function.

	@xmp
		use Iconv

		i = Iconv ('8859_1', 'UTF-8')
		for __ in 1, 3
			s = i.convert ('Copyright Foo')
			println (s.quote)
		;;
		i.close

	@out
		'Copyright \xc2\xa9Foo'
		'Copyright \xc2\xa9Foo'
		'Copyright \xc2\xa9Foo'
*/

static qu_class QuIconvClass;

typedef struct qu_iconv_s qu_iconv_s;
struct qu_iconv_s
{
	QU_OBJECT_HEAD;
	iconv_t  cd;
};

typedef struct qu_iconv_s *qu_iconv;


#define QU_ICONV_FLAG_CLOSED QU_FLAG_0

#define qu_iconv_setclosed(self) qu_flag_set  (self, QU_ICONV_FLAG_CLOSED)
#define qu_iconv_closed(self)    qu_flag_test (self, QU_ICONV_FLAG_CLOSED)

#define ENSURE_NOT_CLOSED(self) QU_REQUIREZ (QuECall, !qu_iconv_closed (self), "Iconv already closed")



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (fromcode: String, tocode: String) -> Iconv __final
	Create a new Iconv object.

	@exc
	.EArgValue
	.EImplement conversion not supported
*/
static qu_iconv qu_iconv__inst (qu_class self, qu_string f, qu_string t)
{
	iconv_t  c;
	qu_iconv o;

	QU_ARG_TYPE (f, qu_string);
	QU_ARG_TYPE (t, qu_string);

	__qu_blocking_begin;
	c = iconv_open (t->str, f->str);
	__qu_blocking_end;
	if (c == (iconv_t) -1)
	{
		if (errno == EINVAL)
			QU_THROW (QuEImplement);
		QU_THROWERRNO;
	}
	o = qu_gc_new (self);
	o->cd = c;
	return o;
}



/*) sub class new (fromcode: String, tocode: String) -> Iconv __final
	Create a new Iconv object.

	This simply calls the #{__inst} method then f{__init} if one is defined.
*/
static qu_iconv qu_iconv_new (qu_class self, qu_string f, qu_string t)
{
	qu_any argv [2] = {f, t};

	qu_iconv o = qu_iconv__inst (self, f, t);
	QU_ENSURE (o);
	return qu_class_init_new (self, o, 2, argv);
}



/*) sub convert (s: String) -> String
	Perform character encoding.

	@exc
	.EArgValue
	.EMemory
*/
static qu_string qu_iconv__convert (qu_iconv self, qu_string s)
{
	QU_ARG_TYPE (s, qu_string);

	ENSURE_NOT_CLOSED (self);

	if (s->len)
	{
		s = qu_iconv___convert (self->cd, s);
		QU_ENSURE (s);
	}
	return s;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu_iconv__destroy (qu_iconv self)
{
	if (!qu_iconv_closed (self))
		iconv_close (self->cd);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub close
	Always call this method when you have finished converting.

	Note that you can not use the Iconv object after closing it.

	@noexc
*/
static void qu_iconv_close (qu_iconv self)
{
	if (!qu_iconv_closed (self))
	{
		iconv_close (self->cd);
		qu_iconv_setclosed (self);
	}
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Iconv_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		functions
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "convert", qu_iconv_convert, 3);
	qu_module_add_function (m, "width"  , qu_iconv_width  , 2);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class EIconv
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuEIconvClass = qu_exception_class_new (m, QuERuntimeClass, "EIconv", "IConv error");


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Iconv
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuIconvClass = qu_class_newz (m, "Iconv", qu_iconv_s, qu_iconv__destroy, 0);

	/* standard class methods */
	qu_class_method_class (QuIconvClass, QU_NAME__inst, qu_iconv__inst, 2);
	qu_class_method_class (QuIconvClass, QU_NAME_new  , qu_iconv_new  , 2);

	/* specific instance methods */
	qu_class_method_instz (QuIconvClass, "close"  , qu_iconv_close   , 0);
	qu_class_method_instz (QuIconvClass, "convert", qu_iconv__convert, 1);

	/* finalize declaration */
	qu_class_enddecl (QuIconvClass);
	return 1;
}


#endif /*HAVE_ICONV_H*/
