/* System.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module System
	This module provide some basic functionalities for Kata.

	Since this module imports symbols to the Sys module then it must be the
	first module to use.

	@xmp
		use System
		# ... then the others

	@p
	The System module sets a number of data in client threads.

	All are imported into the builtin %{Sys} module thus eliminating the need to
	use the System prefix.

	@xmp
		use System

		__protocol__ = 'HTTP/1.1'
		println (__protocol__)

	@out
		HTTP/1.1

	@h2
	__boundary__
		@indent
		The boundary -{String} in case of a multipart body, nil otherwise.

		This is set by %{Kata/Server} from the request headers.
		@unindent

	@h2
	__created__
		@indent
		The newly created object ID, nil if none.

		This is set by %{Kata/Classes/Base$Base.create} so that
		%{Kata/Admin!create} can do a redirection.
		@unindent

	@h2
	__files__
		@indent
		The uploaded files.

		This is set by %{Kata/Service} when reading the request body.
		It is nil if there are no files, a Dict otherwise in which case the
		keys are the name of the query fields and the values are the temporary
		absolute file paths.

		%{Kata/Server} deletes these files after completion of the request.
		@unindent

	@h2
	__message__
		@indent
		The message to display on top of the page by %{Kata/Service}.
		@unindent

	@h2
	__mime__
		@indent
		The content-type String of the body, or nil if there is none.

		This is set by %{Kata/Server} from the request headers
		e.g. C{"multipart/form-data"}.
		@unindent

	@h2
	__oid__
		@indent
		The requested Kata object ID.

		This is set by %{Kata/Server} from the request URI.

		A positive integer means request to an object e.g. C{http://foo.bar.baz/123}.
		Otherwise it was a request to a dynamic page to __path__.
		@unindent

	@h2
	__path__
		@indent
		The requested path translated to real path.

		This is set by %{Kata/Server} from the request URI.

		For a Kata object request this is always in the form of C{"1234"} (the
		object ID as a -{String}).

		For others this points to the actual file, e.g. C{"/MyStuffs/MyKata/foo/bar.qu"}.
		@unindent

	@h2
	__protocol__
		@indent
		The protocol String.

		This is set by %{Kata/Server}, e.g. C{"HTTP/1.1"}.
		@unindent

	@h2
	__query__
		@indent
		The request query.

		This is set by %{Kata/Server} from the request URI, and %{Kata/Service}
		from the request body. The value is a Dict or nil.
		@unindent

	@h2
	__request__
		@indent
		Request data headers.

		This is set by %{Kata/Server}. The value is a Dict.
		@unindent

	@h2
	__remote__
		@indent
		The remote client IP address.

		The value is a -{String} set by %{Kata/Server}, e.g. C{"123.456.789"}.
		@unindent

	@h2
	__response__
		@indent
		The response headers.

		The value is a -{Dict}. An initial value is set by %{Kata/Server} depending
		on the URI, generally C{["content-type": "text/html"]}.

		Service must set this as necessary.
		@unindent

	@h2
	__secure__
		@indent
		Whether the connection is a secured one.

		This is a -{Boolean} set by %{Kata/Server}.
		@unindent

	@h2
	__start__
		@indent
		The time when the request has been accepted.

		This is set by %{Kata/Server}.
		The value is a time value as per the %{Time} module (a -{Float} number).
		@unindent

	@h2
	__status__
		@indent
		The response status.

		An initial C{"OK"} value is set by %{Kata/Server}.

		Service must set this as necessary.
		The value must be one of:

		@rc
		.| C{"OK"}           | request okay
		.| C{"FILE: path"}   | send this file instead
		.| C{"REDIR: url"}   | redirect to this URL instead
		.| C{"ERROR: msg"}   | send error code 500 this message
		.| C{"errcode: msg"} | send error code with this message, e.g. C{"500: it failed"}
		.| C{code"}          | a positive integer code
		@unindent

	@h2
	__trailers__
		@indent
		The trailer request headers in case of HTTP chunk encoding body.

		This is set by %{Kata/Server} after Service has read the request body.

		The value is a Dict.
		@unindent

	@h2
	__trans__
		@indent
		The %{Kata/Transaction} object.

		This is set by %{Kata/Service} before performing any services.
		@unindent

	@h2
	__url__
		@indent
		The path part of the request URI.

		This is set by %{Kata/Server}.

		For a Kata object request this is always in the form of C{"1234"} (the
		object ID as a -{String}).

		For others this is something like C{"/foo/bar/baz.qu"}.
		@unindent

	@h2
	__user__
		@indent
		The %{Kata/Classes/User$User} object.

		This is set by %{Kata/Service} from a login form or auto-login via cookies.
		@unindent

	@h2
	read
		@indent
		@def
		C{sub read (len: Integer = Config.IoChunkSize) -> String}

		@p
		Function to read from the client. Throws on error.
		v{len} must be a positive value (it may not be zero).

		If this function throws, then it is useless to continue.
		@unindent

	@h2
	recv
		@indent
		@def
		C{sub recv (buf, len = Config.IoChunkSize) -> Integer}

		@p
		Function to read from the client into a -{Buffer}.
		Return the actual number of bytes read on success.

		v{len} must be a positive value (it may not be zero).

		If this function throws, then it is useless to continue.
		@unindent

	@h2
	send
		@indent
		@def
		C{sub send (buf: Buffer, len = buf.len) -> Integer}

		@p
		Function to write content of -{Buffer} to client.
		Return the actual number of bytes written on success.

		The Buffer is consumed.

		If this function throws, then it is useless to continue.
		@unindent

	@h2
	write
		@indent
		@def
		C{sub write (s) -> Null}

		@p
		Function to write to client. v{s} may be a -{String} or -{Buffer}.

		Everything is written. You have to manage your own partial writes if
		necessary.

		If this function throws, then it is useless to continue.
		@unindent
*/



/*----------------------------------------------------------------------------
	thread data
----------------------------------------------------------------------------*/

enum
{
	INDEX_boundary,
	INDEX_created,
	INDEX_files,
	INDEX_message,
	INDEX_mime,
	INDEX_oid,
	INDEX_path,
	INDEX_protocol,
	INDEX_query,
	INDEX_request,
	INDEX_remote,
	INDEX_response,
	INDEX_secure,
	INDEX_start,
	INDEX_status,
	INDEX_trailers,
	INDEX_trans,
	INDEX_url,
	INDEX_user,

	INDEX_read,
	INDEX_recv,
	INDEX_send,
	INDEX_write,
};

#define INDEX_LAST INDEX_write


static qu_any _System_get (int index)
{
	qu_array a;

	assert (__QuThreadCurrent);
	assert (0 <= index && index <= INDEX_LAST);

	if (__QuThreadCurrent->extra == 0)
		return QuNil;

	a = __QuThreadCurrent->extra;
	assert (a && qu_array_is (a));
	return a->objs [index];
}


static qu_any _System_set (int index, qu_any v)
{
	qu_array a;

	assert (__QuThreadCurrent);
	assert (0 <= index && index <= INDEX_LAST);

	a = __QuThreadCurrent->extra;
	if (a == 0)
	{
		a = __QuThreadCurrent->extra = qu_array_new (INDEX_LAST + 1);
		QU_OBJECT_LIST_SETVAL (a->objs, QuNil, INDEX_LAST + 1);
		a->len = INDEX_LAST + 1;
	}
	a->objs [index] = v;
	return v;
}


#define DATA(xname) \
static qu_any System_##xname (int argc, qu_any *argv) \
{ \
	if (argc) \
		return _System_set (INDEX_##xname, argv[0]); \
	return _System_get (INDEX_##xname); \
}


#define IO(xname) \
static qu_any System_##xname (int argc, qu_any *argv) \
{ \
	if (argc) \
		return _System_set (INDEX_##xname, argv[0]); \
	return _System_get (INDEX_##xname); \
} \
static qu_any System_io_##xname (int argc, qu_any *argv) \
{ \
	qu_any f = _System_get (INDEX_##xname); \
	assert (f); \
	if (qu_sub_is (f)) \
	{ \
		return qu_sub_calln (f, QuNil, argc, argv); \
	} \
	if (qu_method_is (f)) \
	{ \
		return qu_method_calln (f, argc, argv); \
	} \
	QU_THROWF (QuECall, "calling %s", qu_classof (f)->name->str); \
}


DATA (boundary)
DATA (created)
DATA (files)
DATA (message)
DATA (mime)
DATA (oid)
DATA (path)
DATA (protocol)
DATA (query)
DATA (request)
DATA (remote)
DATA (response)
DATA (secure)
DATA (start)
DATA (status)
DATA (trailers)
DATA (trans)
DATA (url)
DATA (user)

IO   (read)
IO   (recv)
IO   (send)
IO   (write)

#undef DATA
#undef IO



/*----------------------------------------------------------------------------
	functions
----------------------------------------------------------------------------*/

/*) sub __finish__ -> Nil
	Reset all thread data. Called by Server after completion of a request.

	This sets all thread data to Nil.

	@noexc

	@warn
		This method is imported to the builtin %{Sys} module thus you must not
		use the C{System} prefix.
*/
static qu_null System_finish ()
{
	qu_array a;
	int      i;

	assert (__QuThreadCurrent);
	a = __QuThreadCurrent->extra;
	if (a)
	{
		for (i = 0; i <= INDEX_LAST; ++i)
			a->objs[i] = QuNil;
	}
	return QuNil;
}



/*) sub gettag (*arg) -> Any
	@def
	sub gettag (v: Var, name: String) -> Any
	sub gettag (u: Sub, name: String) -> Any
	sub gettag (o: Any, methodname: String, name: String) -> Any
	@p
	Get a tag value of a method or property.

	If the method/property does not define it then search in superclass.

	Return nil if not found.

	@exc
	.EArgValue
*/
static qu_any System_gettag (int argc, qu_any *argv)
{
	qu_symbol y;
	qu_string name;
	qu_any    x;

	QU_ARG_VCOUNT (2, 3);
	if (argc == 3)
	{
		QU_ARG_VTYPE (1, name, qu_string);
		x = argv[0];
		y = qu_classof (x)->get (x, name, 0);
		QU_ENSURE (y);
		name = argv[2];
	}
	else
	{
		y    = argv[0];
		name = argv[1];
	}

	QU_ARG_CHECK ((qu_sub_is (y) || qu_var_is (y)) && qu_string_is (name) && name->len);
	for (;;)
	{
		qu_symbol c;

		if (qu_sub_is (y))
		{
			x = qu_sub_gettag ((qu_sub) y, name);
			c = ((qu_sub) y)->parent;
		}
		else
		{
			x = qu_var_gettag ((qu_var) y, name);
			c = ((qu_var) y)->parent;
		}

		if (x)
			return x;

		if (c == 0 || !qu_class_is (c))
			break;

		c = (qu_symbol) ((qu_class) c)->super;
		if (c == 0)
			break;

		y = qu_attr_find (c->attrs, y->name);
		if (y == 0 || !(qu_sub_is (y) || qu_var_is (y)))
			break;
	}
	return QuNil;
}



/*) sub dispatch (obj, name: String) -> Any
	Call a method v{name} of object v{obj} with C{__query__} as the arguments.

	Note that C{__query__} may not be empty and the method must define the
	C{__args} decorator.

	@warn
	This function does not check for authorization.

	@exc
	.ECall     you passed invalid arguments to this function
	.EValue    invalid __query__ arguments
	.ESystem   something wrong
*/
static qu_any System_dispatch (qu_any self, qu_string name)
{

#define MAXARGS 64

	qu_any    argv [MAXARGS], x;
	int       argc, i;
	qu_dict   query;
	qu_array  argnames = 0;
	qu_class  c;
	qu_sub    u;

	c = qu_classof (self);
	__qu_attr_nothrow_begin;
	u = (qu_sub) c->get (self, name, 0);
	__qu_attr_nothrow_end;

	if (u == 0 || !qu_sub_is (u))
		QU_THROWF (QuECall, "method %s.%s does not exist", c->name->str, name->str);

	if (qu_sub_is_inst (u))
	{
		if (qu_class_is (self))
			QU_THROWF (QuECall, "calling instance method %s.%s with class object", c->name->str, name->str);
	}
	else if (qu_sub_is_class (u))
	{
		if (!qu_class_is (self))
			QU_THROWF (QuECall, "calling class method %s.%s with instance object is not supported", c->name->str, name->str);
	}

	if (!qu_flag_test (u, QU_SUB_FLAG_ARGS))
		QU_THROWF (QuECall, "method %s.%s does not define __args", c->name->str, name->str);

	query = _System_get (INDEX_query);
	QU_REQUIREZ (QuEValue, query && qu_dict_is (query), "invalid query");

	argnames = qu__sub_argnames (u);
	assert (argnames && qu_array_is (argnames));

	argc = argnames->len;
	if (argc > MAXARGS)
		/* this really should never happen */
		QU_THROWF (QuECall, "method %s.%s has too many arguments", c->name->str, name->str);

	for (i = 0; i < argc; ++i)
	{
		name = argnames->objs[i];
		x = qu_dict_getitem (query, name);
		if (x == 0)
			QU_THROWF (QuEValue, "missing value: %s", name->str);
		argv [i] = x;
	}

	x = qu_sub_calln (u, self, argc, argv);
	if (x == 0)
	{
		if (__QuThreadCurrent->eclass == QuEArgValueClass)
			__QuThreadCurrent->eclass = QuEValueClass;
		else
			__QuThreadCurrent->eclass = QuESystemClass;
	}
	return x;
}



/*) sub getvalue (m: Module, name: String) -> Any
	If the module specifies a variable or constant with a specified v{name} then
	return its value, otherwise return nil.

	If v{m} is not a Module or name is not a String then nil is returned, thus
	no exception thrown.

	@noexc
*/
static qu_any System_getvalue (qu_module m, qu_stringy name)
{
	if (qu_module_is (m) && qu_stringy_is (name))
	{
		qu_var v = (qu_var) qu_module_getattr (m, name);
		if (v && qu_var_is (v))
			return v->value;
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_System_init (qu_module m)
{
#define DATA(name) \
	qu_module_add_function (QuSysModule, "__"#name"__", System_##name, -1);

#define IO(name) \
	qu_module_add_function (QuSysModule, "__"#name"__", System_##name, -1); \
	qu_module_add_function (QuSysModule, #name, System_io_##name, -1);

DATA (boundary)
DATA (created)
DATA (files)
DATA (message)
DATA (mime)
DATA (oid)
DATA (path)
DATA (protocol)
DATA (query)
DATA (request)
DATA (remote)
DATA (response)
DATA (secure)
DATA (start)
DATA (status)
DATA (trailers)
DATA (trans)
DATA (url)
DATA (user)

IO   (read)
IO   (recv)
IO   (send)
IO   (write)

#undef DATA
#undef IO

	qu_module_add_function (QuSysModule, "__finish__", System_finish, 0);

	qu_module_add_function (m, "gettag"  , System_gettag  , -1);
	qu_module_add_function (m, "dispatch", System_dispatch,  2);
	qu_module_add_function (m, "getvalue", System_getvalue,  2);

	return 1;
}
