/* Locale.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/*) module Locale
	This module provides access to the system locale information.
*/


/*) sub econv -> Dict
	This function returns a Dict containing all locale information.

	See C{/usr/include/locale.h} for more information.

	@xmp
		use Locale
		n = 3
		for k, v in Locale.econv
			n-- or break
			println (k ' ' v is String ? v.quote : v)
		;;

	@out
		negative_sign ''
		mon_thousands_sep ''
		decimal_point '.'

	@note
	Note that the result may change after a call to #{set}.

	If available, use the #{langinfo} function instead of this function.
*/
static qu_array _locale_grouping_to_array (char * s)
{
	if (*s == '\0')
		return QuArrayEmpty;
	else
	{
		qu_array a = qu_array_new (4);
		while (*s && *s!= CHAR_MAX)
			qu_array_push_quick (a, qu_int_new (*s));
		return a;
	}
}

static qu_dict Locale_econv ()
{
	struct lconv *l = localeconv ();
	qu_dict r = qu_dict_new (24);

#define c_(name)      qu_dict_setitem (r, qu_string_newz ( #name ), qu_string_byte (l->name))
#define s_(name)      qu_dict_setitem (r, qu_string_newz ( #name ), qu_string_newz (l->name))
#define a_(name, val) qu_dict_setitem (r, qu_string_newz ( #name ), val)

	/* Numeric (non-monetary) information. */
	s_(decimal_point);
	s_(thousands_sep);
	a_(grouping, _locale_grouping_to_array (l->grouping));

	/* Monetary information. */
	s_(int_curr_symbol);
	s_(currency_symbol);
	s_(mon_decimal_point);
	s_(mon_thousands_sep);
	a_(mon_grouping, _locale_grouping_to_array (l->mon_grouping));
	s_(positive_sign);
	s_(negative_sign);
	c_(int_frac_digits);
	c_(frac_digits);
	c_(int_frac_digits);
	c_(p_cs_precedes);
	c_(p_sep_by_space);
	c_(n_cs_precedes);
	c_(n_sep_by_space);
	c_(p_sign_posn);
	c_(n_sign_posn);
    return r;

#undef c_
#undef s_
#undef a_

}


#ifdef HAVE_LANGINFO_H

#define i_(name) {#name, name},
struct langinfo_s
{
	char *name;
	int   value;

} LangInfo [] =
{
    i_(ABDAY_1)
    i_(ABDAY_2)
    i_(ABDAY_3)
    i_(ABDAY_4)
    i_(ABDAY_5)
    i_(ABDAY_6)
    i_(ABDAY_7)

    i_(DAY_1)
    i_(DAY_2)
    i_(DAY_3)
    i_(DAY_4)
    i_(DAY_5)
    i_(DAY_6)
    i_(DAY_7)

    i_(ABMON_1)
    i_(ABMON_2)
    i_(ABMON_3)
    i_(ABMON_4)
    i_(ABMON_5)
    i_(ABMON_6)
    i_(ABMON_7)
    i_(ABMON_8)
    i_(ABMON_9)
    i_(ABMON_10)
    i_(ABMON_11)
    i_(ABMON_12)

    i_(MON_1)
    i_(MON_2)
    i_(MON_3)
    i_(MON_4)
    i_(MON_5)
    i_(MON_6)
    i_(MON_7)
    i_(MON_8)
    i_(MON_9)
    i_(MON_10)
    i_(MON_11)
    i_(MON_12)

#ifdef AM_STR
	i_(AM_STR)
#endif
#ifdef PM_STR
	i_(PM_STR)
#endif
#ifdef D_T_FMT
	i_(D_T_FMT)
#endif
#ifdef D_FMT
	i_(D_FMT)
#endif
#ifdef T_FMT
	i_(T_FMT)
#endif
#ifdef T_FMT_AMPM
	i_(T_FMT_AMPM)
#endif
#ifdef ERA
	i_(ERA)
#endif
#ifdef ERA_D_FMT
	i_(ERA_D_FMT)
#endif
#ifdef ALT_DIGITS
	i_(ALT_DIGITS)
#endif
#ifdef ERA_D_T_FMT
	i_(ERA_D_T_FMT)
#endif
#ifdef ERA_T_FMT
	i_(ERA_T_FMT)
#endif
#ifdef CODESET
    i_(CODESET)
#endif
#ifdef INT_CURR_SYMBOL
	i_(INT_CURR_SYMBOL)
#endif
#ifdef CURRENCY_SYMBOL
	i_(CURRENCY_SYMBOL)
#endif
#ifdef MON_DECIMAL_POINT
	i_(MON_DECIMAL_POINT)
#endif
#ifdef MON_THOUSANDS_SEP
	i_(MON_THOUSANDS_SEP)
#endif
#ifdef MON_GROUPING
	i_(MON_GROUPING)
#endif
#ifdef POSITIVE_SIGN
	i_(POSITIVE_SIGN)
#endif
#ifdef NEGATIVE_SIGN
	i_(NEGATIVE_SIGN)
#endif
#ifdef INT_FRAC_DIGITS
	i_(INT_FRAC_DIGITS)
#endif
#ifdef FRAC_DIGITS
	i_(FRAC_DIGITS)
#endif
#ifdef P_CS_PRECEDES
	i_(P_CS_PRECEDES)
#endif
#ifdef P_SEP_BY_SPACE
	i_(P_SEP_BY_SPACE)
#endif
#ifdef N_CS_PRECEDES
	i_(N_CS_PRECEDES)
#endif
#ifdef N_SEP_BY_SPACE
	i_(N_SEP_BY_SPACE)
#endif
#ifdef P_SIGN_POSN
	i_(P_SIGN_POSN)
#endif
#ifdef N_SIGN_POSN
	i_(N_SIGN_POSN)
#endif
#ifdef CRNCYSTR
	i_(CRNCYSTR)
#endif
#ifdef INT_P_CS_PRECEDES
	i_(INT_P_CS_PRECEDES)
#endif
#ifdef INT_P_SEP_BY_SPACE
	i_(INT_P_SEP_BY_SPACE)
#endif
#ifdef INT_N_CS_PRECEDES
	i_(INT_N_CS_PRECEDES)
#endif
#ifdef INT_N_SEP_BY_SPACE
	i_(INT_N_SEP_BY_SPACE)
#endif
#ifdef INT_P_SIGN_POSN
	i_(INT_P_SIGN_POSN)
#endif
#ifdef INT_N_SIGN_POSN
	i_(INT_N_SIGN_POSN)
#endif
#ifdef DECIMAL_POINT
	i_(INT_NDECIMAL_POINT_SIGN_POSN)
#endif
#ifdef RADIXCHAR
	i_(RADIXCHAR)
#endif
#ifdef THOUSANDS_SEP
    i_(THOUSANDS_SEP)
#endif
#ifdef THOUSEP
    i_(THOUSEP)
#endif
#ifdef GROUPING
    i_(GROUPING)
#endif
#ifdef YESEXPR
    i_(YESEXPR)
#endif
#ifdef NOEXPR
    i_(NOEXPR)
#endif

	{ 0, 0 }
};

#undef i_



/*) sub langinfo (item: Integer) -> String
	This function provides access to locale information in a more flexible way than
	#{econv}. Individual and additional elements of the locale categories can be queried.

	v{item} is a constant defined in C{/usr/include/langinfo.h}, e.g.
	CODESET, DAY_1, MON_10, ERA, etc.

	Refer to C{man nl_langinfo} for details.

	@xmp
		use Locale
		println (Locale.langinfo (Locale.DAY_1))

	@out
		Sunday
		ANSI_X3.4-1968

	@warn
	This function and the constants are only available if your system supports it.
	It would be strange if it doesn't, however.
*/
static qu_string Locale_langinfo (qu_integer x)
{
	int item, i;

	QU_ARG_int (x, item);
	for (i = 0; LangInfo[i].name != NULL; i++)
	{
		if (LangInfo[i].value == item)
		{
			const char *r = nl_langinfo (item);
			return r ? qu_string_newz (r) : QuStringEMPTY;
		}
	}
	QU_THROW (QuENone);
}
#endif /* HAVE_LANGINFO_H */



/*) sub get (category: Integer) -> String
	Query the program's current locale.

	The v{category} must be one of the LC constants.

	See C{man setlocale} for more information.

	@exc
	.EValue
	.ENone   no such category

	@xmp
		use Locale
		println (Locale.get (Locale.LC_NUMERIC))

	@out
		C
*/
static qu_any Locale_get (qu_integer x)
{
	int   category;
	char *res;

	QU_ARG_int (x, category);
	res = setlocale (category, NULL);
	QU_REQUIRE (QuENone, res);
	return qu_string_newz (res);
}



/*) sub set (category: Integer, locale: String) -> Null
	Set the program's current locale.

	v{category} must be one of the LC constants.

	v{locale} is a string containing the required setting of category.
	Such a string is either a well-known constant like "C" or "da_DK", or an
	opaque string that was returned by another call of #{get}.

	See C{man setlocale} for more information.

	@exc
	.EValue
	.ESystem  request can not be honored

	@xmp
		use Locale
		Locale.set (Locale.LC_NUMERIC, 'C')
*/
static qu_any Locale_set (qu_integer c, qu_string locale)
{
	int   category;
	char *res;

	QU_ARG_int (c, category);
	QU_ARG_TYPE (locale, qu_string);
	res = setlocale (category, locale->str);
	QU_REQUIRE (QuESystem, res);
	return qu_string_newz (res);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Locale_init (qu_module m)
{
	qu_module_add_function (m, "get"     , Locale_get     , 1);
	qu_module_add_function (m, "set"     , Locale_set     , 2);
	qu_module_add_function (m, "econv"   , Locale_econv   , 0);
	qu_module_add_function (m, "langinfo", Locale_langinfo, 1);

#define l_(name) qu_module_add_const (m, #name, qu_integer_newl (name))
	l_(LC_CTYPE);
	l_(LC_TIME);
	l_(LC_COLLATE);
	l_(LC_MONETARY);
	l_(LC_NUMERIC);
	l_(LC_ALL);
	l_(CHAR_MAX);
#ifdef LC_MESSAGES
	l_(LC_MESSAGES);
#endif

#ifdef HAVE_LANGINFO_H
	{
		int i;
		for (i = 0; LangInfo[i].name != NULL; i++)
		{
			qu_module_add_const (m, LangInfo[i].name, qu_integer_newl (LangInfo[i].value));
		}
	}
#endif

	return 1;
}
