/* Lock.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#define MAXIDi32  (QU_I32_MAX / 2)
#ifdef qu_i64_t
#define MAXIDi64  (QU_I64_MAX / 2)
#endif



/*) module Lock
	This module provides a simple fast thread-safe persistent ID lock
	and counter.

	The lock file contains records. A record is simply a signed integer
	lock state.

	The possible values are:
	@code
		 0 = not locked
		>0 = number of concurrent shared locks
		-1 = write locked
		-2 = deleted

	@p
	IDs can simply be calculated as offsets from the start of the file.
	There is only two integer overhead values.

	@code
		last ID
		reserved (always 0)
		state ID-0
		state ID-1
		state ID-2
		...

	@p
	ID-0 is never used and is always write locked.

	Thus when first created the file would have these entries:
	@code
		0    last ID
		0    reserved
		-1   ID=0 always write locked because it is never used
		0
		0
		...

	@p
	The Lock module supports 32-bits integers and, if available, also
	64-bits integers.

	@warn
	Deleted IDs can not be reused.

	The Lock module does not do file locking.
	You must never access a lock file from outside the single Qu application
	that creates it. If you need to do that then you must create a lock daemon.
*/

/* the two integer overhead */
#define OVERHEADi32 (4 + 4)
#define OVERHEADi64 (8 + 8)

/* offset of an ID */
#define OFFSETi32(id) ((id) * 4 + OVERHEADi32)
#define OFFSETi64(id) ((id) * 8 + OVERHEADi64)

/* lock modes */
#define READLOCK  'r'
#define WRITELOCK 'w'

/* possible results */
#define OKAY        0
#define SYSERR      1
#define BADID       2
#define DELETED     3
#define NOTLOCKED   4
#define HASNOEXCL   5
#define CANTUNLOCK  6
#define TIMEOUT     7
#define EXHAUSTED   8
#define DAMAGED     9

/* throw on failure */
#define QU_LOCK_THROW(result) \
QU_BEGIN \
	lock__throw (result); \
	QU_RETHROW; \
QU_END

static int lock__throw (int result)
{
	assert (result != OKAY);
	switch (result)
	{
		case BADID:      QU_THROWZ (QuEIndex,   "Lock bad ID");
		case DELETED:    QU_THROWZ (QuENone,    "Lock ID deleted");
		case NOTLOCKED:  QU_THROWZ (QuECall,    "Lock ID not locked");
		case HASNOEXCL:  QU_THROWZ (QuECall,    "Lock not exclusively locked");
		case CANTUNLOCK: QU_THROWZ (QuEPanic,   "Lock cant unlock");
		case TIMEOUT:    QU_THROWZ (QuETimeout, "Lock timeout");
		case EXHAUSTED:  QU_THROWZ (QuEData,    "Lock exhausted");
		case DAMAGED:    QU_THROWZ (QuEPanic,   "Lock file damaged");
	}
	QU_THROWERRNO;
}


/* !!! the following macros need fd = file descriptor
   !!! and var is a local variable */

#define SEEK(ofs)     lseek (fd, ofs, SEEK_SET)
#define TELL          lseek (fd, 0, SEEK_CUR)
#define READi32(var)  (readfd  (fd, (char *) &var, 4) != 4)
#define WRITEi32(var) (writefd (fd, (char *) &var, 4) != 4)
#define READi64(var)  (readfd  (fd, (char *) &var, 8) != 8)
#define WRITEi64(var) (writefd (fd, (char *) &var, 8) != 8)

inline static ssize_t readfd (int fd, char *p, size_t z)
{
	ssize_t r;

	while ((r = read (fd, p, z)) < 0)
	{
		if (errno != EINTR)
			return -1;
	}
	return r;
}



inline static ssize_t writefd (int fd, const char *p, size_t z)
{
	ssize_t r;

	while ((r = write (fd, p, z)) < 0)
	{
		if (errno != EINTR)
			return -1;
	}
	return r;
}



#define SUITE(size) \
static int getnew##size (int fd, qu_##size##_t *id) \
{ \
	qu_##size##_t lastid; \
	off_t ofs; \
	\
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (lastid < MAXID##size) \
	{ \
		qu_##size##_t locked = -1; \
		*id = ++lastid; \
		ofs = OFFSET##size (lastid); \
		SEEK (ofs); \
		if (WRITE##size (locked)) \
			return SYSERR; \
		SEEK (0); \
		if (WRITE##size (lastid)) \
			return DAMAGED; \
	} \
	else \
	{ \
		return EXHAUSTED; \
	} \
	return OKAY; \
} \
static int lock##size (int fd, qu_##size##_t id, char mode, double timeout) \
{ \
	qu_##size##_t lastid, state; \
	off_t ofs; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (id < 1 || id > lastid) \
		return BADID; \
	timeout += qu_stream_now (); \
	ofs = OFFSET##size (id); \
	for (;;) \
	{ \
		SEEK (ofs); \
		if (READ##size (state)) \
			return SYSERR; \
		if (state < -1) \
			return DELETED; \
		if (mode == READLOCK) \
		{ \
			if (state >= 0) \
			{ \
				SEEK (ofs); \
				state++; \
				if (WRITE##size (state)) \
					return SYSERR; \
				return OKAY; \
			} \
		} \
		else \
		{ \
			if (state == 0) \
			{ \
				SEEK (ofs); \
				state = -1; \
				if (WRITE##size (state)) \
					return SYSERR; \
				return OKAY; \
			} \
		} \
		if (qu_stream_now () > timeout) \
			return TIMEOUT; \
		__qu_blocking_begin; \
		__qu_blocking_end; \
	} \
} \
static int unlock##size (int fd, qu_##size##_t id) \
{ \
	qu_##size##_t lastid, state; \
	off_t ofs; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (id < 1 || id > lastid) \
		return BADID; \
	ofs = OFFSET##size (id); \
	SEEK (ofs); \
	if (READ##size (state)) \
		return CANTUNLOCK; \
	if (state < -1) \
		return OKAY; \
	if (state == -1) \
	{ \
		SEEK (ofs); \
		state = 0; \
		if (WRITE##size (state)) \
			return CANTUNLOCK; \
	} \
	else if (state > 0) \
	{ \
		SEEK (ofs); \
		--state; \
		if (WRITE##size (state)) \
			return CANTUNLOCK; \
	} \
	return OKAY; \
} \
static int upgrade##size (int fd, qu_##size##_t id, double timeout) \
{ \
	qu_##size##_t lastid, state; \
	off_t ofs; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (id < 1 || id > lastid) \
		return BADID; \
	timeout += qu_stream_now (); \
	ofs = OFFSET##size (id); \
	for (;;) \
	{ \
		SEEK (ofs); \
		if (READ##size (state)) \
			return SYSERR; \
		if (state < -1) \
			return DELETED; \
		if (state == 1) \
		{ \
			SEEK (ofs); \
			state = -1; \
			if (WRITE##size (state)) \
				return SYSERR; \
			return OKAY; \
		} \
		else if (state == -1) \
		{ \
			return OKAY; \
		} \
		if (qu_stream_now () > timeout) \
			return TIMEOUT; \
		__qu_blocking_begin; \
		__qu_blocking_end; \
	} \
} \
static int downgrade##size (int fd, qu_##size##_t id) \
{ \
	qu_##size##_t lastid, state; \
	off_t ofs; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (id < 1 || id > lastid) \
		return BADID; \
	ofs = OFFSET##size (id); \
	SEEK (ofs); \
	if (READ##size (state)) \
		return SYSERR; \
	if (state < -1) \
		return DELETED; \
	if (state != -1) \
		return HASNOEXCL; \
	SEEK (ofs); \
	state = 1; \
	if (WRITE##size (state)) \
		return SYSERR; \
	return OKAY; \
} \
static int delete##size (int fd, qu_##size##_t id) \
{ \
	qu_##size##_t lastid, state; \
	off_t ofs; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (id < 1 || id > lastid) \
		return BADID; \
	ofs = OFFSET##size (id); \
	SEEK (ofs); \
	if (READ##size (state)) \
		return SYSERR; \
	if (state < -1) \
		return OKAY; \
	if (state != -1) \
		return HASNOEXCL; \
	SEEK (ofs); \
	state = -2; \
	if (WRITE##size (state)) \
		return SYSERR; \
	return OKAY; \
} \
static int getstate##size (int fd, qu_##size##_t id, int *res) \
{ \
	qu_##size##_t lastid, state; \
	off_t ofs; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (id < 1 || id > lastid) \
		return BADID; \
	ofs = OFFSET##size (id); \
	SEEK (ofs); \
	if (READ##size (state)) \
		return SYSERR; \
	if (state == 0) \
		*res =  0; \
	else if (state < -1) \
		*res = -2; \
	else if (state == -1) \
		*res = -1; \
	else \
		*res = 1; \
	return OKAY; \
} \
static int reset##size (int fd) \
{ \
	qu_##size##_t lastid, state, zero = 0; \
	SEEK (0); \
	if (READ##size (lastid)) \
		return SYSERR; \
	if (lastid < 0) \
		return DAMAGED; \
	if (READ##size (state)) \
		return SYSERR; \
	for (; lastid >= 0; --lastid) \
	{ \
		off_t ofs = TELL; \
		if (READ##size (state)) \
			return SYSERR; \
		if (state < -1) \
			continue; \
		SEEK (ofs); \
		if (WRITE##size (zero)) \
			return SYSERR; \
	} \
	return OKAY; \
}

SUITE (i32)
#ifdef qu_i64_t
SUITE (i64)
#endif



/*============================================================================
	functions
============================================================================*/

/*) sub create (path: String, size = 4, reserved = 1000) -> Null
	Create a new lock file.

	v{path} is the file path to create.

	v{size} is the size of the integer ID.
	It must be either 4 (32-bits) or 8 (64-bits).

	This function creates a new file with v{reserved} entries to improve
	run-time access. The minimum is 1000 and maximum 1000000000.

	@warn
	This will override any existing file so make sure to check for that
	before calling this function.

	If your system does not have 64-bits integer then v{size} must be 4.

	@exc
	.EArgValue
	.ESystem
	.ESecurity

	@xmp
		use Lock

		PATH = '/tmp/qutest.lock'
		Lock.create (PATH, 4)
*/
#define MINSIZE 1000
#define MAXSIZE 1000000000

static qu_null qu_lock_create (int argc, qu_any *argv)
{
	qu_string path;
	int       size = 4;
	int       reserve = MINSIZE;
	int       fd;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
	{
		QU_ARG_vint (1, size);
#ifdef qu_i64_t
		QU_ARG_CHECK (size == 4 || size == 8);
#else
		QU_ARG_CHECK (size == 4);
#endif
		if (argc > 2)
		{
			QU_ARG_vint (2, reserve);
			QU_ARG_CHECK (reserve >= 0);
			if (reserve < MINSIZE)
				reserve = MINSIZE;
			else if (reserve > MAXSIZE)
				reserve = MAXSIZE;
		}
	}

	fd = creat (path->str, 0666);
	if (fd < 0)
		QU_THROWERRNO;

#ifdef qu_i64_t
	if (size == 8)
	{
		qu_i64_t v;

		v = 0;
		if (WRITEi64 (v))
			goto nope;
		if (WRITEi64 (v))
			goto nope;

		v = -1;
		if (WRITEi64 (v))
			goto nope;

		v = 0;
		for (;reserve > 0; --reserve)
		{
			if (WRITEi64 (v))
				goto nope;
		}
	}
	else
#endif
	{
		qu_i32_t v;

		v = 0;
		if (WRITEi32 (v))
			goto nope;
		if (WRITEi32 (v))
			goto nope;

		v = -1;
		if (WRITEi32 (v))
			goto nope;

		v = 0;
		for (; reserve > 0; --reserve)
		{
			if (WRITEi32 (v))
				goto nope;
		}
	}

	close (fd);
	return QuNil;

  nope:
	{
		int e = errno;
	    close (fd);
		errno = e;
		QU_THROWERRNO;
	}
}


/*============================================================================
	class Lock
============================================================================*/

/*) class Lock
	Use this class to access a lock file created with !{create}.

	@xmp
		use Lock

		PATH = '/tmp/qutest.lock'

		Lock.create (PATH, 4)

		o = Lock ('/tmp/qu.lock', 4, 1)
		id = o.get
		println ('new ID = ' id)
		println ('state  = ' o.mode (id))  # write locked
		o.unlock (id)
		println ('state  = ' o.mode (id))  # unlocked

		o.close
		File.delete (PATH)
	@out
		1
		-1
		0
*/
qu_class QuLockClass;

typedef struct qu_lock_s qu_lock_s;
struct qu_lock_s
{
	QU_OBJECT_HEAD;
	int    size;
	int    fd;
	double timeout;
};
typedef struct qu_lock_s *qu_lock;

#define QU_LOCK_ENSURE_OPENED(self) \
	QU_REQUIREZ (QuECall, (self)->fd >= 0, "Lock already closed")



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (path: String, size = 4, timeout = 1.0) -> Lock
	Create a new Lock object.

	v{path} is an existing lock file created with !{create}.

	v{timeout} is the number of seconds to wait for a #{lock} before
	throwing ETimeout.

	@warn
	The v{size} must be the same as the v{size} used to create the file
	with !{create}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem   cant access file
*/
static qu_lock qu_lock__inst (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	double    timeout = 1.0;
	int       size = 4;
	int       fd;
	qu_lock   o;

	assert (self->size >= QuLockClass->size);

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 0)
	{
		QU_ARG_vint (1, size);
#ifdef qu_i64_t
		QU_ARG_CHECK (size == 4 || size == 8);
#else
		QU_ARG_CHECK (size == 4);
#endif
		if (argc > 2)
			QU_ARG_vdouble (2, timeout);
	}

	fd = open (path->str, O_RDWR | O_SYNC);
	if (fd < 0)
		QU_THROWERRNO;

	o = qu_gc_new (self);
	o->size = size;
	o->fd = fd;
	o->timeout = timeout;
	return o;
}



/*) sub class new (path: String, size = 4, timeout = 1.0) -> Lock
	Create a new Lock object.

	This simply calls #{__inst} then f{__init} if one is defined.
*/
static qu_lock qu_lock_new (qu_class self, int argc, qu_any *argv)
{
	qu_lock o = qu_lock__inst (self, argc, argv);
	QU_ENSURE (o);
	return qu_class_init_new (self, o, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu_lock__destroy (qu_lock self)
{
	if (self->fd >= 0)
		close (self->fd);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub close
	Close the lock file.

	@warn
	After a call to this method, you can no longer access the lock file.
	You would generally only call this method when exiting your application.

	@noexc
*/
static qu_null qu_lock_close (qu_lock self)
{
	if (self->fd >= 0)
	{
		close (self->fd);
		self->fd = -1;
	}
	return QuNil;
}


/*) sub get -> Integer
	Get a new ID.

	@warn
	The new ID is exclusively locked so you must #{unlock} it after you
	have finished using it.

	@exc
	.EPanic  lock file damaged
	.EData   lock file has become too big (exhausted)
	.ESystem I/O error
*/
static qu_integer qu_lock_get (qu_lock self)
{
	int      r;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		r = getnewi64 (self->fd, &id64);
		if (r == OKAY)
			return qu_integer_newi64 (id64);
	}
	else
#endif
	{
		r = getnewi32 (self->fd, &id32);
		if (r == OKAY)
			return qu_integer_newi32 (id32);
	}
	QU_LOCK_THROW (r);
}



/*) sub lock (id: Integer, mode = "r") -> Null
	Lock an ID.

	v{mode} must be either C{"r"} for shared lock, or C{"w"} for exclusive
	(write) lock.

	@exc
	.EPanic   lock file damaged
	.ETimeout lock timeout
	.EIndex   invalid ID
	.ENone    ID has been marked as deleted
	.ESystem  I/O error
*/
static qu_null qu_lock_lock (qu_lock self, int argc, qu_any *argv)
{
	char     mode = 'r';
	int      r;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
	QU_ARG_VCOUNT (1, 2);
	if (argc > 0)
	{
		qu_string s;

		QU_ARG_VTYPE (1, s, qu_string);
		mode = s->str[0];
		QU_ARG_CHECK (mode == 'r' || mode == 'w');
	}
#ifdef qu_i64_t
	if (self->size == 8)
	{
		QU_ARG_vi64 (0, id64);
		r = locki64 (self->fd, id64, mode, self->timeout);
	}
	else
#endif
	{
		QU_ARG_vi32 (0, id32);
		r = locki32 (self->fd, id32, mode, self->timeout);
	}
	if (r == OKAY)
		return QuNil;
	QU_LOCK_THROW (r);
}




/*) sub unlock (id: Integer) -> Null
	Unlock an ID previously locked with #{lock}.

	You must have called #{lock} before calling this method.

	If the ID is not locked then nothing happens.

	@exc
	.EPanic   lock file damaged or failed to unlock
	.EIndex   invalid ID
	.ESystem  I/O error
*/
static qu_null qu_lock_unlock (qu_lock self, qu_integer id)
{
	int      r;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		QU_ARG_i64 (id, id64);
		r = unlocki64 (self->fd, id64);
	}
	else
#endif
	{
		QU_ARG_i32 (id, id32);
		r = unlocki32 (self->fd, id32);
	}
	if (r == OKAY)
		return QuNil;
	QU_LOCK_THROW (r);
}



/*) sub upgrade (id: Integer) -> Null
	Upgrade a shared (read) lock to an exclusive (write) lock.

	You must have a shared lock on the ID.

	@exc
	.EPanic   lock file damaged
	.EIndex   invalid ID
	.ENone    ID has been deleted
	.ETimeout lock timeout
	.ESystem  I/O error
*/
static qu_null qu_lock_upgrade (qu_lock self, qu_integer id)
{
	int      r;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		QU_ARG_i64 (id, id64);
		r = upgradei64 (self->fd, id64, self->timeout);
	}
	else
#endif
	{
		QU_ARG_i32 (id, id32);
		r = upgradei32 (self->fd, id32, self->timeout);
	}
	if (r == OKAY)
		return QuNil;
	QU_LOCK_THROW (r);
}



/*) sub downgrade (id: Integer) -> Null
	Downgrade an exclusive (write) lock to a shared (read) lock.

	You must have an exclusive lock on the ID.

	@exc
	.EPanic   lock file damaged
	.EIndex   invalid ID
	.ENone    ID has been deleted
	.ECall    not exclusively locked
	.ETimeout lock timeout
	.ESystem  I/O error
*/
static qu_null qu_lock_downgrade (qu_lock self, qu_integer id)
{
	int      r;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		QU_ARG_i64 (id, id64);
		r = downgradei64 (self->fd, id64);
	}
	else
#endif
	{
		QU_ARG_i32 (id, id32);
		r = downgradei32 (self->fd, id32);
	}
	if (r == OKAY)
		return QuNil;
	QU_LOCK_THROW (r);
}



/*) sub delete (id: Integer) -> Null
	Mark ID as deleted.

	You must have an exclusive (write) lock on the ID.

	If the ID is already deleted then nothing happens.

	@exc
	.EPanic   lock file damaged
	.EIndex   invalid ID
	.ECall    ID not exclusively locked
	.ESystem  I/O error
*/
static qu_null qu_lock_delete (qu_lock self, qu_integer id)
{
	int      r;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		QU_ARG_i64 (id, id64);
		r = deletei64 (self->fd, id64);
	}
	else
#endif
	{
		QU_ARG_i32 (id, id32);
		r = deletei32 (self->fd, id32);
	}
	if (r == OKAY)
		return QuNil;
	QU_LOCK_THROW (r);
}



/*) sub getstate (id: Integer) -> String
	Get the lock state.

	On success, the returned value is one of:
	@ul
	.C{'r'} shared (read) lock
	.C{'w'} exclusive (write) lock
	.C{'d'} deleted
	.C{''}  not locked

	@exc
	.EPanic   lock file damaged
	.EIndex   invalid ID
	.ENone    ID deleted
	.ESystem  I/O error
*/
static qu_string qu_lock_getstate (qu_lock self, qu_integer id)
{
	int      r, m;
	qu_i32_t id32;
#ifdef qu_i64_t
	qu_i64_t id64;
#endif

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		QU_ARG_i64 (id, id64);
		r = getstatei64 (self->fd, id64, &m);
	}
	else
#endif
	{
		QU_ARG_i32 (id, id32);
		r = getstatei32 (self->fd, id32, &m);
	}
	if (r == OKAY)
	{
		switch (m)
		{
			case  0: return QuStringEMPTY;
			case -1: return qu_string_byte ('w');
			case -2: return qu_string_byte ('d');
		}
		return qu_string_byte ('r');
	}
	QU_LOCK_THROW (r);
}



/*) sub reset -> Null
	Unlock all.

	@warn
	Use this only when you have no lock, e.g. for repairing after a panic
	state.

	@exc
	.EPanic   lock file damaged
	.ESystem  I/O error
*/
static qu_null qu_lock_reset (qu_lock self)
{
	int r;

	QU_LOCK_ENSURE_OPENED (self);
#ifdef qu_i64_t
	if (self->size == 8)
	{
		r = reseti64 (self->fd);
	}
	else
#endif
	{
		r = reseti32 (self->fd);
	}
	if (r == OKAY)
		return QuNil;
	QU_LOCK_THROW (r);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Lock_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		functions
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	qu_module_add_function (m, "create", qu_lock_create, -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Lock
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	QuLockClass = qu_class_newz (m, "Lock", qu_lock_s, qu_lock__destroy, NULL);

	/* standard class methods */
	qu_class_method_class (QuLockClass, QU_NAME__inst, qu_lock__inst, -1);
	qu_class_method_class (QuLockClass, QU_NAME_new  , qu_lock_new  , -1);

	/* specific instance methods */
	qu_class_method_instz (QuLockClass, "close"    , qu_lock_close    ,  0);
	qu_class_method_instz (QuLockClass, "get"      , qu_lock_get      ,  0);
	qu_class_method_instz (QuLockClass, "lock"     , qu_lock_lock     , -1);
	qu_class_method_instz (QuLockClass, "unlock"   , qu_lock_unlock   ,  1);
	qu_class_method_instz (QuLockClass, "upgrade"  , qu_lock_upgrade  ,  1);
	qu_class_method_instz (QuLockClass, "downgrade", qu_lock_downgrade,  1);
	qu_class_method_instz (QuLockClass, "delete"   , qu_lock_delete   ,  1);
	qu_class_method_instz (QuLockClass, "getstate" , qu_lock_getstate ,  1);
	qu_class_method_instz (QuLockClass, "reset"    , qu_lock_reset    ,  0);

	/* finalize declaration */
	qu_class_enddecl (QuLockClass);

	return 1;
}
