/* Math.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Math
	This module provides standard math functions that work with any Qu
	Number objects including Complex, e.g. #{sign}, #{exp}, #{sin}, etc.

	Remember that NaN (Not A Number) in Qu is an error thus any operation
	that yield NaN will throw EMath.

	If the operation yield infinity then EInf or EHuge is thrown unless the
	global Number context allows infinities to be returned. If the operation
	yield a number too small which makes it impossible to be represented then
	ETiny is thrown unless the global Number context allows rounding to 0
	(in which case 0 is returned).

	See -{Number.contex} for more information.

	@warn
	Starting from version 1.09, the Math module no longer extend Number methods
	so you must explicitly say C{Math.foo (x)} instead of C{x.foo}. This will
	be the behavior for stable Qu versions.

	@xmp
		use Math
		x = 1.2
		println (Math.sin (x))

	@out
		0.932039

	@p
	The functions work with Complex numbers and return a Complex number if
	required. For example C{Math.sqrt (-9)} will return a C{0 + 3j}.

	You can prevent this by using the global -{Number.context}. If Complex
	numbers are not allowed then EComplex is thrown if the result is a Complex.


	@h3
	Precision
	@--------
	@indent
	Most functions let you specify a precision. If specified then the calculation
	is performed with Big numbers using that precision. If not, then the
	calculation is performed with Float numbers unless it involves Big numbers,
	in which case the default precision is used.

	Refer to the -{Big} class documentation for valid precision numbers.
	It must be in -{Big.MINPREC} .. -{Big.MAXPREC}.

	A precision value of 8 means the calculation is performed with
	8 * 4 = 32 significant digits.

	A precision value of 0 means to use the default precision. Which should
	be 8.

	When you compare the result of using a Float and Big object then you might
	be surprised. This should give you an idea.

	@code
		# C code
		printf ("%g\n", fmod (1e200, 10.0));
		printf ("%g\n", sin (1e200));

		8
		0.999824

		# Qu code
		println (Big ('1e200') % 10)
		println (Math.sin (Big ('1e200')))
		Big.precision (20)
		println (Math.sin (Big ('1e200')))

		0
		-0.59240528873948451483029329307569E+0
		0.917058182139891576...E+0

	@p
	The C code uses the hardware floating point and can not exactly represent
	the two numbers thus the "erroneous" result. The Qu code uses the Big object
	and does not have that problem. But that does not mean that the Qu code is
	the correct one because it depends on the precision you are using which
	has to do with the way #{sin} is calculated, i.e. it uses a series that
	blows up for large values and thus calculate the remainder of dividing
	the argument by #{PI}.

	This is an issue that is not easy to resolve, especially when it involves
	irrational numbers such as PI.
	@unindent


	@h3
	Interruption
	@-----------
	@indent
	Calculation of high precision values may take a while so most functions
	that needs to compute power series and such, check for interrupt and
	termination signals while performing the calculation. Thus they may
	throw EInterrupt or EQuit.
	@unindent


	@h3
	Performance
	@----------
	@indent
	Operation using Big numbers are much slower than using the hardware floating
	point number. How much slower it is depends, of course, on the precision you
	are using. The #{cos} function, for example, is about 150 times slower when
	using the default precision (8).

	Hopefully future versions will be faster ...
	@unindent
*/

/*
	So how to make it faster?

	- manipulate the qu_big struct directly, i.e. use assignable temporary
	  values to avoid allocation

	- better algorithm: those that do not need divisions
	  (division is about 4 times slower)
*/

#define MATH_ALLOW_EXTERN 0

#if MATH_ALLOW_EXTERN
#define API extern
#define DECLARE
#else
#define API static
#define DECLARE static
#endif


API qu_decimal Math_PI    (int argc, qu_any *argv);
API qu_decimal Math_PHI   (int argc, qu_any *argv);
API qu_decimal Math_E     (int argc, qu_any *argv);

API qu_number  Math_sqrt  (int argc, qu_any *argv);
API qu_real    Math_hypot (int argc, qu_any *argv);
API qu_real    Math_abs   (int argc, qu_any *argv);
API qu_number  Math_exp   (int argc, qu_any *argv);
API qu_number  Math_log   (int argc, qu_any *argv);

API qu_number  Math_sin   (int argc, qu_any *argv);
API qu_number  Math_cos   (int argc, qu_any *argv);
API qu_number  Math_tan   (int argc, qu_any *argv);

API qu_number  Math_sinh  (int argc, qu_any *argv);
API qu_number  Math_cosh  (int argc, qu_any *argv);
API qu_number  Math_tanh  (int argc, qu_any *argv);

API qu_number  Math_asin  (int argc, qu_any *argv);
API qu_number  Math_acos  (int argc, qu_any *argv);
API qu_number  Math_atan  (int argc, qu_any *argv);
API qu_number  Math_atan2 (int argc, qu_any *argv);

API qu_number  Math_asinh (int argc, qu_any *argv);
API qu_number  Math_acosh (int argc, qu_any *argv);
API qu_number  Math_atanh (int argc, qu_any *argv);

API qu_int     Math_rand  (int argc, qu_any *argv);


/*----------------------------------------------------------------------------
	constants
----------------------------------------------------------------------------*/

#define MATH_PI_FLOAT  3.14159265358979323846
#define MATH_PHI_FLOAT 1.6180339887498948482
#define MATH_E_FLOAT   2.7182818284590452354


/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

/* ensure the precision value is a valid one */
#define ENSURE_VALID_PRECISION(n) \
	QU_ARG_CHECK (QU_BIG_MINPREC <= (n) && (n) <= QU_BIG_MAXPREC)


/* check for interruption while calculating */
#define CHECK_INTERRUPT \
QU_BEGIN \
	if (__QuSigINT) \
		QU_THROW (QuEInterrupt); \
	if (__QuSigTERM) \
		QU_THROW (QuEQuit); \
QU_END


#define CHECK_INTERRUPT_GOTO(where) \
QU_BEGIN \
	if (__QuSigINT) \
	{ \
		qu_thread_throw (QuEInterruptClass, 0); \
		goto where; \
	} \
	if (__QuSigTERM) \
	{ \
		qu_thread_throw (QuEQuitClass, 0); \
		goto where; \
	} \
QU_END



/* do the switch */
#define MATH_NUMBER(type, name, x) \
QU_BEGIN \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return (type) Math_##name##_int      ((qu_int) x); \
		case QU_CLASS_ID_Long:     return (type) Math_##name##_long     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return (type) Math_##name##_rational ((qu_rational) x); \
		case QU_CLASS_ID_Float:    return (type) Math_##name##_float    ((qu_float) x); \
		case QU_CLASS_ID_Big:      return (type) Math_##name##_big      ((qu_big) x); \
		case QU_CLASS_ID_Complex:  return (type) Math_##name##_complex  ((qu_complex) x); \
		case QU_CLASS_ID_Inf:      return (type) Math_##name##_inf      ((qu_inf) x); \
	} \
	QU_ARG_THROW; \
QU_END



/* all functions with a precision argument take this form */
#define MATH_WITH_PREC(type, name) \
QU_BEGIN \
	qu_any x; \
	QU_ARG_VCOUNT (1, 2); \
	x = argv[0]; \
	if (argc == 2) \
	{ \
		long prec; \
		QU_ARG_vlong (1, prec); \
		if (prec) \
		{ \
			long p = __QuBigPrecision; \
			type r; \
			\
			ENSURE_VALID_PRECISION (prec); \
			__QuBigPrecision = prec; \
			switch (qu_classid (x)) \
			{ \
				case QU_CLASS_ID_Int: \
					r = (type) Math_##name##_big (qu_int_as_big ((qu_int) x)); \
					break; \
				case QU_CLASS_ID_Long: \
					r = (type) Math_##name##_big (qu_long_as_big ((qu_long) x)); \
					break; \
				case QU_CLASS_ID_Rational: \
					r = (type) Math_##name##_big (qu_rational_as_big ((qu_rational) x)); \
					break; \
				case QU_CLASS_ID_Float: \
					r = (type) Math_##name##_big (qu_float_as_big ((qu_float) x)); \
					break; \
				case QU_CLASS_ID_Big: \
					r = (type) Math_##name##_big ((qu_big) x); \
					break; \
				case QU_CLASS_ID_Inf: \
					r = (type) Math_##name##_inf ((qu_inf) x); \
					break; \
				case QU_CLASS_ID_Complex: \
					x = _math_complex_big ((qu_complex) x); \
					r = (type) Math_##name##_complex ((qu_complex) x); \
					break; \
				default: \
					QU_ARG_THROW; \
			} \
			__QuBigPrecision = p; \
			return r; \
		} \
	} \
	MATH_NUMBER(type, name, x); \
QU_END


/* call Math_sqrt with 1 argument assert the result is a qu_real */
#define Math_sqrt1r(x) ((qu_real) Math_sqrt1 (x))

/* call Math_xxx with 1 argument */
inline static qu_number Math_sqrt1  (qu_any x) { return Math_sqrt  (1, &x); }
inline static qu_number Math_exp1   (qu_any x) { return Math_exp   (1, &x); }
inline static qu_number Math_log1   (qu_any x) { return Math_log   (1, &x); }
inline static qu_number Math_sin1   (qu_any x) { return Math_sin   (1, &x); }
inline static qu_number Math_cos1   (qu_any x) { return Math_cos   (1, &x); }
inline static qu_number Math_tan1   (qu_any x) { return Math_tan   (1, &x); }
inline static qu_number Math_sinh1  (qu_any x) { return Math_sinh  (1, &x); }
inline static qu_number Math_cosh1  (qu_any x) { return Math_cosh  (1, &x); }
inline static qu_number Math_tanh1  (qu_any x) { return Math_tanh  (1, &x); }
inline static qu_number Math_asin1  (qu_any x) { return Math_asin  (1, &x); }
inline static qu_number Math_acos1  (qu_any x) { return Math_acos  (1, &x); }
inline static qu_number Math_atan1  (qu_any x) { return Math_atan  (1, &x); }
inline static qu_number Math_asinh1 (qu_any x) { return Math_asinh (1, &x); }
inline static qu_number Math_acosh1 (qu_any x) { return Math_acosh (1, &x); }
inline static qu_number Math_atanh1 (qu_any x) { return Math_atanh (1, &x); }


/* call Math_sqrt with a precision */
inline static qu_number Math_sqrt2  (qu_any x, int prec)
{
	qu_any argv[2] = {x, qu_int_new (prec) };
	return Math_sqrt (2, argv);
}


/* call Math_hypot with 2 arguments */
inline static qu_real Math_hypot2 (qu_real x, qu_real y)
{
	qu_any argv[2] = {x, y};
	return Math_hypot (2, argv);
}


/* call Math_atan2 with 2 arguments */
inline static qu_number Math_atan22 (qu_any x, qu_any y)
{
	qu_any argv[2] = {x, y};
	return Math_atan2 (2, argv);
}



/* whether or not the imaginary or real part is a big one */
#define COMPLEX_IS_BIG(x) (qu_big_is ((x)->real) || qu_big_is ((x)->imag))


/*
	Make the real and imaginary parts qu_big numbers.
	This is used prior to do the math with precision.
*/
static qu_complex _math_complex_big (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;
	int  k = 0;

	if (!qu_inf_is (r) && !qu_big_is (r))
	{
		k = 1;
		r = (qu_real) qu_real_as_big (r);
		assert (r && qu_big_is (r));
	}

	if (!qu_inf_is (i) && !qu_big_is (i))
	{
		k = 1;
		i = (qu_real) qu_real_as_big (i);
		assert (i && qu_big_is (i));
	}

	return k ? qu_complex_new (r, i) : x;
}




/*----------------------------------------------------------------------------
	PI
----------------------------------------------------------------------------*/

/*) sub PI (precision = 0) -> Decimal
	Return the ratio of a circle's circumference to its diameter.

	If you don't specify a v{precision} then a constant Float with a value
	of 3.14159265358979323846 is returned. Otherwise a Big is returned.

	The Math module keeps a PI value with 1000 digits so if you specify
	a precision less than or equal 250 then no calculation is performed.

	Otherwise this function calculates PI using the Machin's formula:
	@code

	    PI/4 = 4 atan (1/5) - atan (1/239)

	@p
	where #{atan} is calculated using the Maclaurin series:
	@code
		           oo
		          ___       n + 1
		atan(x) = \     (-1)       2n -1
		          /     --------- x
		          ---     2n - 1
		         n = 1

		              1  3   1  5   1  7
		        = x - - x  + - x  - - x  + ...  for  -1 < x < 1
		              3      5      7

	@p
	Calculating PI with a high precision value may take a while.
	On a 4GHz Pentium it takes about 8 seconds for a precision of 2500
	(10000 digits).

	@xmp
		use Math
		println (Math.PI)
		println (Math.PI (8))

	@out
		3.1415926535897931
		0.00031415926535897932384626433832E+4
*/
static qu_big   MathPiBig;
static qu_float MathPiFloat;


/*
	!!! except for interruption, this function always succeed
*/
static qu_big _math_pi_big ()
{
	if (__QuBigPrecision == MathPiBig->len)
	{
		return MathPiBig;
	}
	else if (__QuBigPrecision < MathPiBig->len)
	{
		return qu_big_make (MathPiBig->digits, __QuBigPrecision, MathPiBig->expo);
	}
	else
	{
		static qu_big b8 = 0;
		static qu_big b4 = 0;
		static qu_big b9 = 0;
		static qu_big b5 = 0;

		qu_big pi, d, t, k, p;
		int    ctx;

		if (b8 == 0)
		{
			b8 = qu_big_new_static ("-80");
			b4 = qu_big_new_static ("-0.04");
			b9 = qu_big_new_static ("956");
			b5 = qu_big_new_static ("-57121");
		}

	  __qu_number_context_begin (ctx, __QuNumberContext | QU_NUMBER_CONTEXT_ALLOW_TINY);

		p  = qu_big_make1 (1, -__QuBigPrecision);
		pi = QuBig0;

		t = b8;
		k = QuBig1;
		for (;;)
		{
			t = (qu_big) qu_big_mul_big (t, b4); assert (t && qu_big_is (t));
			d = (qu_big) qu_big_div_big (t, k);  assert (d && qu_big_is (d));

			if (qu_big_cmpabs_big (d, p) <= 0)
				break;

			pi = (qu_big) qu_big_plus_big (pi, d);     assert (pi && qu_big_is (pi));
			k  = (qu_big) qu_big_plus_big (k, QuBig2); assert (k && qu_big_is (k));

			CHECK_INTERRUPT_GOTO (interrupted);
		}

		t = b9;
		k = QuBig1;
		for (;;)
		{
			t = (qu_big) qu_big_div_big (t, b5); assert (t && qu_big_is (t));
			d = (qu_big) qu_big_div_big (t, k);  assert (d && qu_big_is (d));

			if (qu_big_cmpabs_big (d, p) <= 0)
				break;

			pi = (qu_big) qu_big_plus_big (pi, d);     assert (pi && qu_big_is (pi));
			k  = (qu_big) qu_big_plus_big (k, QuBig2); assert (k && qu_big_is (k));

			CHECK_INTERRUPT_GOTO (interrupted);
		}

	  __qu_number_context_end (ctx);
		return pi;

	  interrupted:

	  __qu_number_context_end (ctx);
		QU_RETHROW;

	}
}



DECLARE qu_decimal Math_PI (int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		long prec;

		QU_ARG_vlong (0, prec);
		if (prec)
		{
			long   old = __QuBigPrecision;
			qu_big pi;

			ENSURE_VALID_PRECISION (prec);
			__QuBigPrecision = prec;
			pi = _math_pi_big ();
			__QuBigPrecision = old;
			return (qu_decimal) pi;
		}
	}
	return (qu_decimal) MathPiFloat;
}



/*----------------------------------------------------------------------------
	PHI
----------------------------------------------------------------------------*/

/*) sub PHI (precision = 0) -> Decimal
	Return the I{golden ratio}.

	If you don't specify a v{precision} then a constant Float with a value
	of 1.6180339887498948482 is returned. Otherwise a Big is returned.

	The Math module keeps a PHI value with 1000 digits so if you specify
	a precision less than or equal 250 then no calculation is performed.

	Otherwise this function calculates
	@code

		      1        _
	    PHI = - (1 + \/5 )
		      2

	@p
	Calculating PHI with a high precision value may take a while.
	On a 4GHz Pentium it takes about 8 seconds for a precision of 2500
	(10000 digits).

	@xmp
		use Math
		println (Math.PHI)
		println (Math.PHI (8))

	@out
		1.6180339887498949
		0.00016180339887498948482045868343E+4
*/
static qu_big   MathPhiBig;
static qu_float MathPhiFloat;


/*
	!!! except for interruption, this function always succeed
*/
static qu_big _math_phi_big ()
{
	if (__QuBigPrecision == MathPhiBig->len)
	{
		return MathPhiBig;
	}
	else if (__QuBigPrecision < MathPhiBig->len)
	{
		return qu_big_make (MathPhiBig->digits, __QuBigPrecision, MathPhiBig->expo);
	}
	else
	{
		qu_big r;

		/* Remember! sqrt can get interrupted */

		r = (qu_big) Math_sqrt1      (QuBig5);      QU_ENSURE (r); assert (qu_big_is (r));
		r = (qu_big) qu_big_plus_big (r, QuBig1);   QU_ENSURE (r); assert (qu_big_is (r));
		r = (qu_big) qu_big_mul_big  (r, QuBig1_2); QU_ENSURE (r); assert (qu_big_is (r));

		return r;
	}
}



DECLARE qu_decimal Math_PHI (int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		long prec;

		QU_ARG_vlong (0, prec);
		if (prec)
		{
			long   old = __QuBigPrecision;
			qu_big phi;

			ENSURE_VALID_PRECISION (prec);
			__QuBigPrecision = prec;
			phi = _math_phi_big ();
			__QuBigPrecision = old;
			return (qu_decimal) phi;
		}
	}
	return (qu_decimal) MathPhiFloat;
}



/*----------------------------------------------------------------------------
	E
----------------------------------------------------------------------------*/

/*) sub E (precision = 0) -> Decimal
	Return the base of natural logarithm.

	If you don't specify a v{precision} then a constant Float with a value of
	2.7182818284590452354 is returned. Otherwise a Big is returned.

	The Math module keeps an I{e} value with 1000 digits so if you specify
	a precision <= 250 then no calculation is performed.

	Otherwise this function calculates I{e} using the Euler formula:
	@code
		     oo
		     ___
		     \    1
		e =  /    --
			 ---  k!
		     k=0

		  = 2.5 + 1/3! + 1/4! + 1/5! + ...

	@p
	This is a slow converging series so calculating I{e} with a high precision
	value may take a while. On a 4GHz Pentium it takes about 83 seconds for a
	precision of 1000 (8000 digits).

	You need about 120 terms to get I{e} correct to 200 digits.

	@xmp
		use Math
		println (Math.E)
		println (Math.E (10))

	@out
		2.7182818284590451
		0.0002718281828459045235360287471352662497E+4
*/
static qu_big   MathEBig;
static qu_float MathEFloat;

/*
	!!! except for interruption, this function always succeed
*/
static qu_big _math_e_big ()
{
	if (__QuBigPrecision == MathEBig->len)
	{
		return MathEBig;
	}
	else if (__QuBigPrecision < MathEBig->len)
	{
		return qu_big_make (MathEBig->digits, __QuBigPrecision, MathEBig->expo);
	}
	else
	{
		static qu_big b25 = 0;

		qu_big p = qu_big_make1 (1, -__QuBigPrecision);
		qu_big f = QuBig6;
		qu_big i = QuBig3;
		qu_big e, m;
		int ctx;

		if (b25 == 0)
			b25 = qu_big_new_static ("2.5");

	  __qu_number_context_begin (ctx, __QuNumberContext | QU_NUMBER_CONTEXT_ALLOW_TINY);

		e = b25;
		for (;;)
		{
			m = (qu_big) qu_big_div_big  (QuBig1, f); assert (m && qu_big_is (m));

			if (qu_big_cmpabs_big (m, p) <= 0)
				break;

			e = (qu_big) qu_big_plus_big (e, m);  assert (e && qu_big_is (e));
			i = (qu_big) qu_big_incr     (i);     assert (i && qu_big_is (i));
			f = (qu_big) qu_big_mul_big  (f, i);  assert (f && qu_big_is (f));

			CHECK_INTERRUPT_GOTO (interrupted);
		}

	  __qu_number_context_end (ctx);
		return e;

	  interrupted:
	  __qu_number_context_end (ctx);
		QU_RETHROW;
	}
}



DECLARE qu_decimal Math_E (int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		long prec;

		QU_ARG_vlong (0, prec);

		if (prec)
		{
			long   old = __QuBigPrecision;
			qu_big e;

			ENSURE_VALID_PRECISION (prec);
			__QuBigPrecision = prec;
			e = _math_e_big ();
			__QuBigPrecision = old;
			return (qu_decimal) e;
		}
	}
	return (qu_decimal) MathEFloat;
}



/*----------------------------------------------------------------------------
	hypot
----------------------------------------------------------------------------*/

/*) sub hypot (x: Real, y:Real, precision = 0) -> Real
	Return the distance of the point C{(x,  y)} from the origin.

	@code
	    y
		|    (x, y)
		|   /
		|  /
		| / hypot
		|/
		+--------- x

		hypot (x, y) =  sqrt (x * x + y * y)

	@p
	If either v{x} or v{y} is an infinity then this function returns infp
	or throws EInf if the -{Number.context} does not allow infinities to
	be returned.

	If both arguments are 0, then an exact value of Int (0) is returned.

	If one of the argument is 0, then the absolute value of the other is
	returned. This means that the result is as exact as the argument.
	No precision lost.

	Otherwise, the calculation is done using the above formula. The result
	depends on the #{sqrt} function. If it is able to return an exact result,
	e.g. C{sqrt (25 _/ 36) = 5 _/ 6} then the result is exact.

	If you specify a precision, then -{Big.precision} is set to this value
	before the calculation is performed. The outcome depends, again, on the
	#{sqrt} function. The multiplication and addition will still be carried
	normally.

	If the intermmediate result (the multiplication and addition) is too big
	or too small then EHuge, ETiny or EInf might be thrown.

	@xmp
		use Math
		println (Math.hypot (0, 5 _/ 6))
		println (Math.hypot (-1, 2))
		println (Math.hypot (1, 2, 10))

	@out
		(5 _/ 6)
		2.2360679774997898
		0.000223606797749978969640917366873099E+4
*/
static qu_real _math_hypot (qu_real x, qu_real y, int prec)
{
	if (qu_inf_is (x) || qu_inf_is (y))
	{
		return (qu_real) __qu_inf__ (1);
	}
	else
	{
		int sx = qu_real_sign (x);
		int sy = qu_real_sign (y);

		if (sx == 0)
			return (sy > 0) ? y : ((sy < 0) ? qu_real_neg (y) : (qu_real) QuInt0);

		else if (sy == 0)
			return (sx < 0) ? qu_real_neg (x) : x;

		else
		{
			/*
				note that we let QuEHuge or QuETiny propagate automatically
				e.g. if x * x yields QuEHuge or QuETiny then we simply rethrow it
				and if it yields 0 or infinity then we simply carry on
			*/

			x = qu_real_mul_real  (x, x); QU_ENSURE (x);
			y = qu_real_mul_real  (y, y); QU_ENSURE (y);
			x = qu_real_plus_real (x, y); QU_ENSURE (x);

			return prec ? (qu_real) Math_sqrt2 (x, prec) : (qu_real) Math_sqrt1 (x);
		}
	}
}



DECLARE qu_real Math_hypot (int argc, qu_any *argv)
{
	qu_real x, y;
	long    prec = 0;
	long    old = __QuBigPrecision;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, x, qu_real);
	QU_ARG_VTYPE  (1, y, qu_real);

	if (argc == 3)
	{
		QU_ARG_vlong (2, prec);
		ENSURE_VALID_PRECISION (prec);
		__QuBigPrecision = prec;
	}

	x = _math_hypot (x, y, prec);
	__QuBigPrecision = old;
	return x;
}



/*----------------------------------------------------------------------------
	abs
----------------------------------------------------------------------------*/

/*) sub abs (x: Number, precision = 0) -> Real
	This function returns the absolute value of v{x}.

	For Real numbers the result is a number with the same class. There is
	no "morphing". For example, C{abs (1.0)} returns a Float, not an Int.
	The v{precision} argument is ignored.

	Thus for Real numbers, there is no precision lost.
	@code
		abs (x) =  x  for x >= 0
		          -x  for x <  0

	For Complex numbers the value is mathematically defined as:
	@code
		abs (z) = hypot (z.real, z.imag)
	@p
	thus the v{precision} value is respected. See #{hypot] for details.

	@xmp
		use Math
		println (Math.abs (-4))
		println (Math.abs (-4 _/ 7))
		println (Math.abs (2 + 4j, 10))

		Number.context += 'I'
		println (Math.abs (infp))
		println (Math.abs (infn))

	@out
		4
		(4 _/ 7)
		0.000447213595499957939281834733746255E+4
		infp
		infp
*/
static qu_int      Math_abs_int      (qu_int      x) { return qu_int_isneg      (x) ? qu_int_neg      (x) : x; }
static qu_long     Math_abs_long     (qu_long     x) { return qu_long_isneg     (x) ? qu_long_neg     (x) : x; }
static qu_rational Math_abs_rational (qu_rational x) { return qu_rational_isneg (x) ? qu_rational_neg (x) : x; }
static qu_float    Math_abs_float    (qu_float    x) { return qu_float_isneg    (x) ? qu_float_neg    (x) : x; }
static qu_big      Math_abs_big      (qu_big      x) { return qu_big_isneg      (x) ? qu_big_neg      (x) : x; }
static qu_real     Math_abs_complex  (qu_complex  x) { return _math_hypot (x->real, x->imag, 0); }
static qu_inf      Math_abs_inf      (qu_inf      x) { return __qu_inf__ (1); }

DECLARE qu_real Math_abs (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_real, abs);
}



/*----------------------------------------------------------------------------
	sqrt
----------------------------------------------------------------------------*/

/*) sub sqrt (x: Number, precision = 0) -> Number
	Return the square root of v{x}.

	@code
		sqrt (R) = SQRT(R)   for R >= 0
		sqrt (Z) = r + i.j   for R <  0

		where:
			r = SQRT ((h + Z.real) / 2)
			i = SQRT ((h - Z.real) / 2) * sign (Z.imag)
			h = hypot (Z.real, Z.imag)

	@p
	The f{sqrt} function will try its best to return an exact result within
	reasonable performance limits.

	@xmp
		use Math
		println (Math.sqrt (9))
		println (Math.sqrt (25 _/ 9))
		println (Math.sqrt (15241578750190521))
		println (Math.sqrt (9.0))
		println (Math.sqrt (Big ('900.000')))

		Number.context += 'R'

		println (Math.sqrt (9 + 0j))
		println (Math.sqrt (-4))

		Number.context += 'I'
		println (Math.sqrt (infp))
		println (Math.sqrt (infn))

	@out
		3
		(5 _/ 3)
		123456789
		3
		30
		3
		(0 + 2j)
		infp
		(0 + infp.j)
*/
static qu_number Math_sqrtd         (double v);
static qu_number Math_sqrt_int      (qu_int x);
static qu_number Math_sqrt_long     (qu_long x);
static qu_number Math_sqrt_rational (qu_rational x);
static qu_number Math_sqrt_float    (qu_float x);
static qu_number Math_sqrt_big      (qu_big x);
static qu_number Math_sqrt_complex  (qu_complex x);
static qu_number Math_sqrt_inf      (qu_inf x);

/* the maximum integer that can be square-root-ed */
static double MathSqrtMaxInteger;

static qu_number Math_sqrtd (double v)
{
	if (v == 0)
	{
		return (qu_number) QuInt0;
	}
	else
	{
		int     g = (v < 0);
		double  q, m;
		qu_real r = 0;

		if (g)
			v = -v;

		q = sqrt (v);
		if (v <= MathSqrtMaxInteger && (modf (v, &m) == 0))
		{
			/*
				v is a whole number
				see if the square root is also a whole number
			*/

#ifdef HAVE_LONG_LONG
			long long t = (long long) q;
			if (t * t == (long long) v)
				r = (qu_real) qu_integer_newll (t);
#else
			long t = (long) q;
			if (t * t == (long) v)
				r = (qu_real) qu_integer_newl (t);
#endif
		}

		if (r == 0)
			r = (qu_real) qu_float_new (q);

		return g
			? (qu_number) __qu_complex__ ((qu_real) QuInt0, r)
			: (qu_number) r;
	}
}



/*
	See if we can determine whether the result is an qu_integer.
	r is the result, x is the original number.
*/
static qu_integer _math_sqrt_is_integer (qu_integer r, qu_real x)
{
	qu_integer m;

	m = (qu_integer) qu_integer_mul_integer (r, r); assert (m && qu_integer_is (m));

	if (qu_integer_eq_real (m, x))
	{
		return r;
	}
	else
	{
		/* deal with rounding :-( */

		qu_integer q;

		q = (qu_integer) qu_integer_incr (r);           assert (q && qu_integer_is (q));
		m = (qu_integer) qu_integer_mul_integer (q, q); assert (m && qu_integer_is (m));

		if (qu_integer_eq_real (m, x))
		{
			return q;
		}
		else
		{
			/* CHECKME - this should always fail? */

			q = (qu_integer) qu_integer_decr (r);           assert (q && qu_integer_is (q));
			m = (qu_integer) qu_integer_mul_integer (q, q); assert (m && qu_integer_is (m));

			if (qu_integer_eq_real (m, x))
			{
				return q;
			}
		}
	}
	return NULL;
}



static qu_number Math_sqrt_int (qu_int x)
{
	long v = qu_int_tol (x);

	if (v == 0)
	{
		return (qu_number) QuInt0;
	}
	else
	{
		int    g = (v < 0);
		qu_real   r;
		double q;

		if (g)
			v = -v;

		q = sqrt ((double) v);

		r = (qu_real) _math_sqrt_is_integer ((qu_integer) qu_int_new (q), (qu_real) x);
		if (r == 0)
			r = (qu_real) qu_float_new (q);

		return g
			? (qu_number) __qu_complex__ ((qu_real) QuInt0, r)
			: (qu_number) r;
	}
}



static qu_number Math_sqrt_long (qu_long x)
{
	long i = qu_long_toli (x);

	if (errno == 0)
	{
		/* this includes 0 */
		return Math_sqrt_int (qu_int_new (i));
	}
	else
	{
		int        g = qu_long_isneg (x);
		qu_integer q;
		qu_real    r;

		if (g)
			x = qu_long_neg (x);

		r = (qu_real) Math_sqrt_big (qu_long_as_big (x));
		assert (r && qu_real_is (r) && !qu_inf_is (r));

		q = _math_sqrt_is_integer (qu_real_as_integer (r), (qu_real) x);
		if (q)
			r = (qu_real) q;

		return g
			? (qu_number) __qu_complex__ ((qu_real) QuInt0, r)
			: (qu_number) r;
	}
}



static qu_number Math_sqrt_rational (qu_rational x)
{
	if (qu_rational_iszero (x))
	{
		return (qu_number) QuInt0;
	}
	else if (qu_rational_iswhole (x))
	{
		return qu_long_is (x->num)
			? Math_sqrt_long ((qu_long) x->num)
			: Math_sqrt_int  ((qu_int)  x->num);
	}
	else
	{
		int        g = qu_rational_isneg (x);
		qu_integer n = x->num;
		qu_integer d = x->den;
		qu_real    p, q;
		double     v;

		if (g)
			n = qu_integer_neg ((qu_integer) n);

		if (qu_int_is (n))
			p = (qu_real) Math_sqrt_int ((qu_int) n);
		else
			p = (qu_real) Math_sqrt_long ((qu_long) n);

		if (qu_integer_is (p))
		{
			if (qu_int_is (d))
				q = (qu_real) Math_sqrt_int ((qu_int) d);
			else
				q = (qu_real) Math_sqrt_long ((qu_long) d);

			if (qu_integer_is (q))
			{
				p = qu_rational_new_ex ((qu_integer) p, (qu_integer) q);
				return g
					? (qu_number) __qu_complex__ ((qu_real) QuInt0, p)
					: (qu_number) p;
			}
		}

		v = qu_rational_tod (x);
		return errno
			? Math_sqrt_big (qu_rational_as_big (x))
			: Math_sqrtd (v);
	}
}



static qu_number Math_sqrt_float (qu_float x)
{
	return Math_sqrtd (qu_float_tod (x));
}



/*
	!!! this calculates sqrt (0.xxxxxx) -- see Math_sqrt_big below
	!!! except for interruption, this always succeed
*/
static qu_big _math_sqrt_big_noexp (qu_big x)
{
	qu_big p = qu_big_make1 (1, -__QuBigPrecision);
	qu_big r, l, d;
	int    n = 0;

	assert (x->expo == 0);
	assert (qu_big_sign (x) > 0);

	r = (qu_big) qu_big_mul_big (x, QuBig1_2);
	assert (r && qu_big_is (r));

	for (;;)
	{
		l = r;
		r = (qu_big) qu_big_div_big   (x, r);        assert (r && qu_big_is (r));
		r = (qu_big) qu_big_plus_big  (r, l);        assert (r && qu_big_is (r));
		r = (qu_big) qu_big_mul_big   (r, QuBig1_2); assert (r && qu_big_is (r));
		d = (qu_big) qu_big_minus_big (r, l);        assert (d && qu_big_is (d));

		if (qu_big_cmpabs_big (d, p) <= 0)
			break;

		if (++n > 50)
			__QuBigPrecision++;

		CHECK_INTERRUPT;
	}
	return r;
}



/* !!! do not call this from outside Math_sqrt, call Math_sqrt1 instead */
static qu_number Math_sqrt_big (qu_big x)
{
	int sx = qu_big_sign (x);

	if (sx == 0)
	{
		return (qu_number) QuInt0;
	}
	else if (qu_big_isone (x))
	{
		return (sx < 0)
			? (qu_number) __qu_complex__ ((qu_real) QuInt0, (qu_real) QuInt1)
			: (qu_number) QuInt1;
	}
	else
	{
		/*
		|                     e
		|  b = 0.xxxxx * 10000
		|    _     _______      e/2
		|  \/b = \/0.xxxxx 10000    * (e % 2 ? 100 : 1)
		|
		|                         1      _     _______      0
		|      b = 0.xxxxx * 10000     \/b = \/0.xxxxx 10000  * 100
		|                         2      _     _______      1
		|      b = 0.xxxxx * 10000     \/b = \/0.xxxxx 10000
		|                         3      _     _______      1
		|      b = 0.xxxxx * 10000     \/b = \/0.xxxxx 10000  * 100
		|                         4      _     _______      2
		|      b = 0.xxxxx * 10000     \/b = \/0.xxxxx 10000
		|                         5      _     _______      2
		|      b = 0.xxxxx * 10000     \/b = \/0.xxxxx 10000  * 100
		|
		|      ... etc.
		*/

		qu_big  r;
		qu_real q;
		long    e, o;

		if (sx < 0)
			x = qu_big_neg (x);

		e = x->expo;
		x->expo = 0;
		o = __QuBigPrecision;
		r = _math_sqrt_big_noexp (x);
		__QuBigPrecision = o;
		x->expo = e;

		QU_ENSURE (r);

		if (r != QuBig0)
		{
			assert (r != x);
			r->expo += e / 2;
			if (e % 2)
			{
				static qu_big m = 0;
				if (m == 0)
					m = qu_big_new_static ("100");

				r = (qu_big) qu_big_mul_big (r, m); assert (r && qu_big_is (r));
			}
		}
		q = (qu_real) r;
		if (qu_big_iswhole (x))
		{
			qu_integer e = _math_sqrt_is_integer (qu_big_as_integer (r), (qu_real) x);
			if (e)
				q = (qu_real) e;
		}
		return (sx < 0)
			? (qu_number) __qu_complex__ ((qu_real) QuInt0, q)
			: (qu_number) q;
	}
}



static qu_number Math_sqrt_complex (qu_complex x)
{
	if (qu_real_iszero (x->imag))
	{
		if (qu_real_iszero (x->real))
		{
			return __qu_complex__ ((qu_real) QuInt0, (qu_real) QuInt0);
		}
		else
		{
			qu_real v = x->real;
			int     g = qu_real_isneg (v);

			if (g)
				v = qu_real_neg (v);

			v = (qu_real) Math_sqrt1 (v); QU_ENSURE (v); assert (qu_real_is (v));

			return g
				? __qu_complex__ ((qu_real) QuInt0, v)
				: __qu_complex__ (v, (qu_real) QuInt0);
		}
	}
	else if (qu_real_iszero (x->real))
	{
		qu_real r = x->imag;
		int     g = qu_real_isneg (r);

		if (g)
			r = qu_real_neg (r);

		r = qu_real_div_int (r, QuInt2); QU_ENSURE (r); assert (qu_real_is (r));
		r = Math_sqrt1r     (r);         QU_ENSURE (r); assert (qu_real_is (r));

		return g
			? __qu_complex__ (qu_real_neg (r), r)
			: __qu_complex__ (r, r);
	}
	else
	{
		qu_real r = x->real;
		qu_real i = x->imag;
		qu_real h;
		int     g = qu_real_isneg (i);

		h = Math_hypot2        (r, i);      if (h == 0) goto nope; assert (qu_real_is (h));
		i = qu_real_minus_real (h, r);      if (i == 0) goto nope; assert (qu_real_is (i));
		i = qu_real_div_int    (i, QuInt2); if (i == 0) goto nope; assert (qu_real_is (i));
		i = Math_sqrt1r        (i);         if (i == 0) goto nope; assert (qu_real_is (i));
		r = qu_real_plus_real  (h, r);      if (r == 0) goto nope; assert (qu_real_is (r));
		r = qu_real_div_int    (r, QuInt2); if (r == 0) goto nope; assert (qu_real_is (r));
		r = Math_sqrt1r        (r);         if (r == 0) goto nope; assert (qu_real_is (r));

		return __qu_complex__ (r, g ? qu_real_neg (i) : i);

	  nope:
		if (QU_EXCEPTION_LAST_IS (QuEInf))
			QU_RETHROW;
		QU_THROW (QuEMath);
	}
}



static qu_number Math_sqrt_inf (qu_inf x)
{
	if (x == QuInfPos)
		return (qu_number) __qu_inf__ (1);
	else
		return __qu_complex__ ((qu_real) QuInt0, (qu_real) QuInfPos);
}



DECLARE qu_number Math_sqrt (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, sqrt);
}




/*----------------------------------------------------------------------------
	exp
----------------------------------------------------------------------------*/

/*) sub exp (x: Number, precision = 0) -> Number
	Return the value of C{e} (the base of natural logarithms) raised to the
	power of v{x}.

	@code
		           R
		exp (R) = e        for -oo <= R <= +oo
		exp (Z) = r + i.j

		where:
			r = exp (Z.real) * cos (Z.imag)
			i = exp (Z.real) * sin (Z.imag)

	@p
	If required, computation uses the Maclaurin series:
	@code
		     oo
		     ---   1   n
		 x   \     -- x
		e  = /     n!
		     ---
		     n=0

		             1  2   1  3   1   4
		   = 1 + x + - x  + - x  + -- x  + ...  for 0 <= x < +oo
		             2      6      24

	@p
	Note that this is a slow converging series. On a 4GHz Pentium it takes about
	28 seconds to calculate C{Math.exp (Big ('1000000.0'))} using the default
	precision of 8 (32 digits).

	@xmp
		use Math
		println (Math.exp (2))
		println (Math.exp (10000))

		Number.context += 'R'
		println (Math.exp (1 + 0j))

		Number.context += 'I'
		println (Math.exp (infp))
		println (Math.exp (infn))

	@out
		7.3890560989306504
		0.08806818225662921587338715178973E+4344
		2.7182818284590451
		infp
		0
*/
static qu_real   Math_expd         (double v);
static qu_real   Math_exp_int      (qu_int x);
static qu_real   Math_exp_long     (qu_long x);
static qu_real   Math_exp_rational (qu_rational x);
static qu_real   Math_exp_float    (qu_float x);
static qu_real   Math_exp_big      (qu_big x);
static qu_number Math_exp_complex  (qu_complex x);
static qu_real   Math_exp_inf      (qu_inf x);



static qu_real Math_expd (double v)
{
	if (v == 0)
	{
		return (qu_real) QuInt0;
	}
	else
	{
		double r = exp (v);
		return (qu_double_isfinite (r) && r)
			? (qu_real) qu_float_new (r)
			: (qu_real) Math_exp_big (qu_big_newd (v));
	}
}



static qu_real Math_exp_int (qu_int x)
{
	return Math_expd (qu_int_tol (x));
}



static qu_real Math_exp_long (qu_long x)
{
	double v = qu_long_tod (x);
	return (errno)
		? (qu_real) Math_expd (v)
		: (qu_real) Math_exp_big (qu_long_as_big (x));
}



static qu_real Math_exp_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return (errno)
		? (qu_real) Math_expd (v)
		: (qu_real) Math_exp_big (qu_rational_as_big (x));
}



static qu_real Math_exp_float (qu_float x)
{
	return Math_expd (qu_float_tod (x));
}



/*
	We try to avoid throwing exceptions thus if this function returns QuNil
	then it means we couldn't do it. If it returns 0 then we were interupted.
*/
static qu_any _math_exp_big (qu_big x)
{
	qu_big r, m, i, f, p, t;

	assert (!qu_big_isneg (x));

	p = qu_big_make1 (1, -__QuBigPrecision);
	t = r = f = QuBig1;
	i = QuBig0;
	for (;;)
	{
		t = (qu_big) qu_big_mul_big (t, x); if (t == 0 || !qu_big_is (t)) return QuNil;
		i = (qu_big) qu_big_incr    (i);    if (i == 0 || !qu_big_is (i)) return QuNil;
		f = (qu_big) qu_big_mul_big (f, i); if (f == 0 || !qu_big_is (f)) return QuNil;
		m = (qu_big) qu_big_div_big (t, f); if (m == 0 || !qu_big_is (m)) return QuNil;

		if (qu_big_cmpabs_big (m, p) <= 0)
			break;

		r = (qu_big) qu_big_plus_big (r, m); if (r == 0 || !qu_big_is (r)) return QuNil;

		CHECK_INTERRUPT;
	}
	return r;
}



static qu_real Math_exp_big (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return (qu_real) QuInt1;
	}
	else if (qu_big_isone (x))
	{
		qu_big e = _math_e_big ();

		if (qu_big_isneg (x))
			return qu_big_div_big (QuBig1, e);
		else
			return (qu_real) e;
	}
	else
	{
		qu_any r;
		int    g = qu_big_isneg (x);

		if (g)
			x = qu_big_neg (x);

		r = _math_exp_big (x);
		assert (r != (qu_any) x);

		/* we were interrupted */
		QU_ENSURE (r);

		if (qu_real_is (r))
			return g ? qu_big_div_real (QuBig1, (qu_real) r) : (qu_real) r;

		/* couldn't do it */
		if (x->expo > 0)
			return g
				? (qu_real) __qu_tiny__ ((qu_real) QuInt0)
				: (qu_real) __qu_huge__ (1);

		else
			return g
				? (qu_real) __qu_huge__ (1)
				: (qu_real) __qu_tiny__ ((qu_real) QuInt0);
	}
}



static qu_number Math_exp_complex (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;

	if (qu_real_iszero (i))
	{
		r = (qu_real) Math_exp1 (r);
		i = (qu_real) QuInt0;
	}
	else if (qu_real_iszero (r))
	{
		r = (qu_real) Math_cos1 (i);
		i = (qu_real) Math_sin1 (i);
	}
	else
	{
		qu_real e, a, b;
		/*
			z.real = exp (x.real) * cos (x.imag)
			z.imag = exp (x.real) * sin (x.imag)
		*/
		a = (qu_real) Math_cos1 (i); QU_ENSURE (a); assert (qu_real_is (a));
		b = (qu_real) Math_sin1 (i); QU_ENSURE (b); assert (qu_real_is (b));
		e = (qu_real) Math_exp1 (r); QU_ENSURE (e); assert (qu_real_is (e));

		r = qu_real_mul_real (e, a);
		i = qu_real_mul_real (e, b);
	}

	QU_ENSURE (r); assert (qu_real_is (r));
	QU_ENSURE (i); assert (qu_real_is (i));

	return __qu_complex__ (r, i);
}



static qu_real Math_exp_inf (qu_inf x)
{
	return (x == QuInfPos) ? (qu_real) __qu_inf__ (1) : (qu_real) QuInt0;
}



DECLARE qu_number Math_exp (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, exp);
}



/*----------------------------------------------------------------------------
	log
----------------------------------------------------------------------------*/

/*) sub log (x: Number, precision = 0) -> Number
	Return the natural logarithm of v{x}.

	@code
		log (R) = LOG (R)  for R >= 0
		log (Z) = r + i.j

		where:
			r = log (hypot (Z.real, Z.imag))
			i = atan2 (Z.imag, Z.real)

	@p
	If required, computation uses the Maclaurin series:
	@code
		               oo
		               ---    2      2n - 1
		     1 + x     \   -------- x
		ln ( ----- ) = /   (2n - 1)
		     1 - x     ---
		               n=1

		                    2  3   2  5   2  7
		             = 2x + - x  + - x  + - x  + ...  for -1 < x < 1
		                    3      5      7

		solving for x = (y - 1) / (y + 1)

		                1  3   1  5   1  7
		ln (y) = 2 (x + - x  + - x  + - x + ... )
		                3      5      7

		observing:

		          n
		ln (x * 10 ) = ln (x) + ln (10) * n

	@xmp
		use Math
		println (Math.log (123))
		println (Math.log (Big ('123')))

		Number.context += 'R'
		println (Math.log (Math.E + 0j))

		Number.context += 'I'
		println (Math.log (infp))
		println (Math.log (infn))

	@out
		4.8121843553724171
		0.000481218435537241749524E+4
		1
		infp
		(infp + 3.1415926535897931j)
*/

/*
	log (-1) = PI.j
	log (0)  = infn
	log (1)  = 0
	log (e)  = 1
	log (j)  = PI/2.j
	log (-j) = -PI/2.j
*/

static qu_number Math_log_int      (qu_int x);
static qu_number Math_log_long     (qu_long x);
static qu_number Math_log_rational (qu_rational x);
static qu_number Math_log_float    (qu_float x);
static qu_number Math_log_big      (qu_big x);
static qu_number Math_log_complex  (qu_complex x);
static qu_number Math_log_inf      (qu_inf x);


static qu_number Math_logd (double v)
{
	if (v == 0)
	{
		return (qu_number) __qu_inf__ (-1);
	}
	else if (v == 1)
	{
		return (qu_number) QuInt0;
	}
	else if (v < 0)
	{
		QU_NUMBER_ENSURE_ALLOW_COMPLEX;
		return Math_log1 (qu_complex_new1 ((qu_real) qu_float_new (v)));
	}
	else
	{
		double r = log (v);
		return (qu_double_isfinite (r) && r)
			? (qu_number) qu_float_new (r)
			: (qu_number) Math_log_big (qu_big_newd (v));
	}
}



static qu_number Math_log_int (qu_int x)
{
	return Math_logd (qu_int_tod (x));
}



static qu_number Math_log_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_log_big (qu_long_as_big (x))
		: Math_logd (v);
}



static qu_number Math_log_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_log_big (qu_rational_as_big (x))
		: Math_logd (v);
}



static qu_number Math_log_float (qu_float x)
{
	return Math_logd (qu_float_tod (x));
}



/*
	this computes log (0.xxxxxxxx) for small exponent
	it always succeed except for interruption
*/
static qu_big _math_log_big (qu_big x)
{
	qu_big r, f, w, d, i, t, p;

	p = qu_big_make1 (1, -__QuBigPrecision);

	f = (qu_big) qu_big_minus_big (x, QuBig1); assert (f && qu_big_is (f));
	t = (qu_big) qu_big_plus_big  (x, QuBig1); assert (t && qu_big_is (t));
	f = (qu_big) qu_big_div_big   (f, t);      assert (f && qu_big_is (f));
	w = (qu_big) qu_big_mul_big   (f, f);      assert (w && qu_big_is (w));

	r = f;
	i = QuBig3;

	for (;;)
	{
		f = (qu_big) qu_big_mul_big (f, w); assert (f && qu_big_is (f));
		d = (qu_big) qu_big_div_big (f, i); assert (d && qu_big_is (d));

		if (qu_big_cmpabs_big (d, p) <= 0)
			break;

		r = (qu_big) qu_big_plus_big (r, d);      assert (r && qu_big_is (r));
		i = (qu_big) qu_big_plus_big (i, QuBig2); assert (i && qu_big_is (i));

		CHECK_INTERRUPT;
	}
	r = (qu_big) qu_big_mul_big (r, QuBig2); assert (r && qu_big_is (r));
	return r;
}



static qu_number Math_log_big (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return (qu_number) __qu_inf__ (-1);
	}
	else if (qu_big_is1 (x))
	{
		return (qu_number) QuInt0;
	}
	else if (qu_big_isneg (x))
	{
		QU_NUMBER_ENSURE_ALLOW_COMPLEX;
		return Math_log1 (qu_complex_new1 ((qu_real) x));
	}
	else
	{
		/*
			b = 0.xxx * 10000 ** E
			  = log (0.xxx) + log (10) * E * 4
		*/

		long   e = x->expo;
		int    g = e < 0;
		qu_big p = qu_big_newl ((g ? -e : e) * QU_BIG_FIG);
		qu_big m, r;

		m =          _math_log_big  (QuBig10); assert (m && qu_big_is (m));
		m = (qu_big) qu_big_mul_big (m, p);    assert (m && qu_big_is (m));

		x->expo = 0;
		r =  _math_log_big (x); assert (r && qu_big_is (r));
		x->expo = e;

		return g ? (qu_number) qu_big_minus_big (r, m) : (qu_number) qu_big_plus_big (r, m);
	}
}



static qu_number Math_log_complex (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;
	qu_real v;
	/*
		z.real = log (hypot (x.real, x.imag))
		z.imag = atan2 (x.imag, x.real)
	*/
	v = (qu_real) Math_hypot2 (r, i); QU_ENSURE (v); assert (qu_real_is (v));
	v = (qu_real) Math_log1   (v);    QU_ENSURE (v); assert (qu_real_is (v));
	i = (qu_real) Math_atan22 (i, r); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (v, i);
}



static qu_number Math_log_inf (qu_inf x)
{
	if (x == QuInfPos)
		return (qu_number) __qu_inf__ (1);
	else
	{
		qu_decimal pi = Math_PI (0, NULL);
		return __qu_complex__ ((qu_real) QuInfPos, (qu_real) pi);
	}
}



DECLARE qu_number Math_log (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, log);
}




/*----------------------------------------------------------------------------
	helper - reduce angle
----------------------------------------------------------------------------*/

inline static qu_big _math_reduce_angle (qu_big x)
{
	if (qu_big_cmpabs_big (x, QuBig7) >= 0)
	{
		qu_big p2 = _math_pi_big ();

		assert (p2);

		p2 = (qu_big) qu_big_mul_big (p2, QuBig2); QU_ENSURE (p2); assert (qu_big_is (p2));
		x  = (qu_big) qu_big_rem_big (x, p2);      QU_ENSURE (x);  assert (qu_big_is (x));
	}
	return x;
}



/*----------------------------------------------------------------------------
	sin
----------------------------------------------------------------------------*/

/*) sub sin (x: Number, precision = 0) -> Number
	Return the sine of v{x}, where v{x} is given in radians.

	@code
		sin (R) = SIN (R)   for -oo < R < +oo
		sin (Z) = r + i.j

		where:
			r = sin (Z.real) * cosh (Z.imag)
			i = cos (Z.real) * sinh (Z.imag)

		sin (infp)  ->  EMath
		sin (infn)  ->  EMath

	@p
	If required, computation uses the Maclaurin series:
	@code
		           oo        n
		           ---   (-1)     2n + 1
		           \   --------- x
		sin (x) =  /   (2n + 1)!
		           ---
		           n=0

		               1  3    1   5     1   7
		        = x -  - x  + --- x -  ---- x  + ... for -oo < x < oo
		               6      120      5040

	@p
	This series is only good for small v{x} values. For large v{x} values then
	this function calculates C{x %= 2 * pi} which may require the calculation
	of #{PI}.

	@xmp
		use Math
		println (Math.sin (0.2))
		println (Math.sin (0.2, 10))

		# remember that 1e200 can not be exactly stored in a Float
		println (Math.sin (1e200))
		println (Math.sin (Big ('1e200')))

	@out
		0.19866933079506122
		0.1986693307950612263403374309676868065228E+0
		0.9998238056375376
		-0.59240528873948451483029329307569E+0
*/
static qu_real   Math_sin_int      (qu_int x);
static qu_real   Math_sin_float    (qu_float x);
static qu_real   Math_sin_long     (qu_long x);
static qu_real   Math_sin_rational (qu_rational x);
static qu_real   Math_sin_big      (qu_big x);
static qu_number Math_sin_complex  (qu_complex x);
static qu_any    Math_sin_inf      (qu_inf x);



static qu_real Math_sind (double v)
{
	if (v == 0)
		return (qu_real) QuInt0;
	return (qu_real) qu_float_new (sin (v));
}



static qu_real Math_sin_int (qu_int x)
{
	return Math_sind (qu_int_tod (x));
}



static qu_real Math_sin_float (qu_float x)
{
	return Math_sind (qu_float_tod (x));
}



static qu_real Math_sin_long (qu_long x)
{
	qu_int y = qu_long_as_int (x);
	if (y)
		return Math_sin_int (y);
	else
	{
		double v = qu_long_tod (x);
		return errno
			? (qu_real) Math_sin_big (qu_long_as_big (x))
			: (qu_real) qu_float_new (sin (v));
	}
}



static qu_real Math_sin_rational (qu_rational x)
{
	if (qu_rational_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else if (qu_rational_iswhole (x))
	{
		if (qu_int_is (x->num))
			return Math_sin_int ((qu_int) x->num);
		else
			return Math_sin_long ((qu_long) x->num);
	}
	else
	{
		double v = qu_rational_tod (x);
		return errno
			? Math_sin_big (qu_rational_as_big (x))
			: (qu_real) qu_float_new (sin (v));
	}
}



static qu_real Math_sin_big (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		qu_big f, w, d, i, z, p;
		int g;

		x =          _math_reduce_angle (x);    assert (x && qu_big_is (x));
		w = (qu_big) qu_big_mul_big     (x, x); assert (w && qu_big_is (w));

		p = qu_big_make1 (1, -__QuBigPrecision);
		f = x;
		i = QuBig3;
		z = QuBig6;
		g = 0;

		for (;;)
		{
			f = (qu_big) qu_big_mul_big (f, w); assert (f && qu_big_is (f));
			d = (qu_big) qu_big_div_big (f, z); assert (d && qu_big_is (d));

			if (qu_big_cmpabs_big (d, p) <= 0)
				break;

			g = !g;
			if (g)
			{
				x = (qu_big) qu_big_minus_big (x, d); assert (x && qu_big_is (x));
			}
			else
			{
				x = (qu_big) qu_big_plus_big (x, d); assert (x && qu_big_is (x));
			}

			d = (qu_big) qu_big_plus_big (i, QuBig1); assert (d && qu_big_is (d));

			z = (qu_big) qu_big_mul_big  (z, d);      assert (z && qu_big_is (z));
			i = (qu_big) qu_big_plus_big (i, QuBig2); assert (i && qu_big_is (i));
			z = (qu_big) qu_big_mul_big  (z, i);      assert (z && qu_big_is (z));

			CHECK_INTERRUPT;
		}
		return (qu_real) x;
	}
}



static qu_number Math_sin_complex (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;

	if (qu_real_iszero (i))
	{
		/*
			z.real = sin (x.real) * 1
			z.imag = cos (x.real) * 0
		*/

		if (qu_real_iszero (r))
		{
			r = i = (qu_real) QuInt0;
		}
 		else
		{
			r = (qu_real) Math_sin1 (r);
			i = (qu_real) QuInt0;
		}
	}
	else if (qu_real_iszero (r))
	{
		/*
			z.real = 0 * cosh (x.imag)
			z.imag = 1 * sinh (x.imag)
		*/

		r = (qu_real) QuInt0;
		i = (qu_real) Math_sinh1 (i);
	}
	else
	{
		/*
			z.real = sin (x.real) * cosh (x.imag)
			z.imag = cos (x.real) * sinh (x.imag)
		*/

		qu_real s, c, sh, ch;

		s  = (qu_real) Math_sin1  (r); QU_ENSURE (s);  assert (qu_real_is (s));
		c  = (qu_real) Math_cos1  (r); QU_ENSURE (c);  assert (qu_real_is (c));
		sh = (qu_real) Math_sinh1 (i); QU_ENSURE (sh); assert (qu_real_is (sh));
		ch = (qu_real) Math_cosh1 (i); QU_ENSURE (ch); assert (qu_real_is (ch));

		r = (qu_real) qu_real_mul_real (s, ch);
		i = (qu_real) qu_real_mul_real (c, sh);
	}

	QU_ENSURE (r); assert (qu_real_is (r));
	QU_ENSURE (i); assert (qu_real_is (i));

	return __qu_complex__ (r, i);
}



static qu_any Math_sin_inf (qu_inf x)
{
	QU_THROW (QuEMath);
}



DECLARE qu_number Math_sin (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, sin);
}



/*----------------------------------------------------------------------------
	cos
----------------------------------------------------------------------------*/

/*) sub cos (x: Number, precision = 0) -> Number
	Return the cosine of v{x}, where v{x} is given in radians.

	@code
		cos (R) = COS (R)   for -oo < R < +oo
		cos (Z) = r + i.j

		where:
			r =  cos (Z.real) * cosh (Z.imag)
			i = -sin (Z.real) * sinh (Z.imag)

		cos (infp)  ->  EMath
		cos (infn)  ->  EMath

	@p
	If required, computation uses the Maclaurin series:
	@code
		           oo      n
		           --- (-1)   2n
		           \   ----- x
		cos (x) =  /   (2n)!
		           ---
		           n=0

		               1  2    1   4    1   6
		        = 1 -  - x  + --- x -  --- x  + ... for -oo < x < oo
		               2      24       720

	@p
	This series is only good for small v{x} values. For large v{x} values then
	this function calculates C{x %= 2 * pi} which may require the calculation
	of #{PI}.

	@xmp
		use Math
		println (Math.cos (0.2))
		println (Math.cos (0.2, 10))

	@out
		0.98006657784124163
		0.9800665778412416289185238633926826334941E+0
*/
static qu_real   Math_cos_int      (qu_int x);
static qu_real   Math_cos_float    (qu_float x);
static qu_real   Math_cos_long     (qu_long x);
static qu_real   Math_cos_rational (qu_rational x);
static qu_real   Math_cos_big      (qu_big x);
static qu_number Math_cos_complex  (qu_complex x);
static qu_any    Math_cos_inf      (qu_inf x);


static qu_real Math_cosd (double v)
{
	if (v == 0)
		return (qu_real) QuInt1;
	return (qu_real) qu_float_new (cos (v));
}



static qu_real Math_cos_int (qu_int x)
{
	return Math_cosd (qu_int_tod (x));
}



static qu_real Math_cos_float (qu_float x)
{
	return Math_cosd (qu_float_tod (x));
}



static qu_real Math_cos_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_cos_big (qu_long_as_big (x))
		: Math_cosd (v);
}



static qu_real Math_cos_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_cos_big (qu_rational_as_big (x))
		: Math_cosd (v);
}



static qu_real Math_cos_big (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return (qu_real) QuInt1;
	}
	else
	{
		int g;
		qu_any f, w, d, i, z, delta;

		x = _math_reduce_angle (x);    assert (x && qu_big_is (x));
		w = qu_big_mul_big     (x, x); assert (x && qu_big_is (x));

		x = f = QuBig1;
		i = z = QuBig2;
		g = 0;
		delta = qu_big_make1 (1, -__QuBigPrecision);

		for (;;)
		{
			f = (qu_big) qu_big_mul_big (f, w); assert (f && qu_big_is (f));
			d = (qu_big) qu_big_div_big (f, z); assert (d && qu_big_is (d));

			if (qu_big_cmpabs_big (d, delta) <= 0)
				break;

			g = !g;
			if (g)
			{
				x = (qu_big) qu_big_minus_big (x, d); assert (x && qu_big_is (x));
			}
			else
			{
				x = (qu_big) qu_big_plus_big  (x, d); assert (x && qu_big_is (x));
			}

			d = (qu_big) qu_big_plus_big (i, QuBig1); assert (d && qu_big_is (d));
			z = (qu_big) qu_big_mul_big  (z, d);      assert (z && qu_big_is (z));
			i = (qu_big) qu_big_plus_big (i, QuBig2); assert (i && qu_big_is (i));
			z = (qu_big) qu_big_mul_big  (z, i);      assert (z && qu_big_is (z));
		}
		return (qu_real) x;
	}
}



static qu_number Math_cos_complex (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;

	if (qu_real_iszero (i))
	{
		/*
			z.real =  cos (x.real) * 1
			z.imag = -sin (x.real) * 0
		*/

		if (qu_real_iszero (r))
		{
			r = (qu_real) QuInt1;
			i = (qu_real) QuInt0;
		}
 		else
		{
			r = (qu_real) Math_cos1 (r);
			i = (qu_real) QuInt0;
		}
	}
	else if (qu_real_iszero (r))
	{
		/*
			z.real = 1 * cosh (x.imag)
			z.imag = 0 * sinh (x.imag)
		*/

		r = (qu_real) Math_cosh1 (i);
		i = (qu_real) QuInt0;
	}
	else
	{
		/*
			z.real =  cos (x.real) * cosh (x.imag)
			z.imag = -sin (x.real) * sinh (x.imag)
		*/

		qu_real s, c, sh, ch;

		s  = (qu_real) Math_sin1   (r);  QU_ENSURE (s);  assert (qu_real_is (s));
		c  = (qu_real) Math_cos1   (r);  QU_ENSURE (c);  assert (qu_real_is (c));
		sh = (qu_real) Math_sinh1  (i);  QU_ENSURE (sh); assert (qu_real_is (sh));
		ch = (qu_real) Math_cosh1  (i);  QU_ENSURE (ch); assert (qu_real_is (ch));

		r = (qu_real) qu_real_mul_real (c, ch);
		i = (qu_real) qu_real_mul_real (s, sh);
	}

	QU_ENSURE (r); assert (qu_real_is (r));
	QU_ENSURE (i); assert (qu_real_is (i));

	return __qu_complex__ (r, qu_real_neg (i));
}



static qu_any Math_cos_inf (qu_inf x)
{
	QU_THROW (QuEMath);
}



DECLARE qu_number Math_cos (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, cos);
}



/*----------------------------------------------------------------------------
	tan
----------------------------------------------------------------------------*/

/*) sub tan (x: Number, precisionr = 0) -> Number
	Return the tangent of v{x}, where v{x} is given in radians.

	@code
		tan (X) = sin (X) / cos (X)

		tan (infp)  ->  EMath
		tan (infn)  ->  EMath

	@xmp
		use Math
		println (Math.tan (0.2))
		println (Math.tan (0.2, 10))

	@out
		0.2027100355086725
		0.20271003550867249487979424105214E+0

	@see
	#{sin}
	#{cos}
*/
static qu_real   Math_tand         (double v);
static qu_real   Math_tan_int      (qu_int x);
static qu_real   Math_tan_float    (qu_float x);
static qu_real   Math_tan_long     (qu_long x);
static qu_real   Math_tan_rational (qu_rational x);
static qu_real   Math_tan_big      (qu_big x);
static qu_number Math_tan_complex  (qu_complex x);
static qu_any    Math_tan_inf      (qu_inf x);



static qu_real Math_tand (double v)
{
	if (v == 0)
	{
		return (qu_real) QuInt0;
	}
	else
	{
		double r = tan (v);
		return qu_double_isfinite (r)
			? (qu_real) qu_float_new (r)
			: (qu_real) __qu_inf__ ((v > 0) ? 1 : -1);
	}
}



static qu_real Math_tan_int (qu_int x)
{
	return Math_tand (qu_int_tol (x));
}



static qu_real Math_tan_float (qu_float x)
{
	return Math_tand (qu_float_tod (x));
}



static qu_real Math_tan_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? (qu_real) Math_tan_big (qu_long_as_big (x))
		: (qu_real) Math_tand (v);
}



static qu_real Math_tan_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? (qu_real) Math_tan_big (qu_rational_as_big (x))
		: (qu_real) Math_tand (v);
}



static qu_real Math_tan_big (qu_big x)
{
	qu_real a, b;
	x = _math_reduce_angle (x); assert (x && qu_big_is (x));
	a = (qu_real) Math_sin1 (x); QU_ENSURE (a); assert (qu_real_is (a));
	b = (qu_real) Math_cos1 (x); QU_ENSURE (b); assert (qu_real_is (b));
	return qu_real_div_real (a, b);
}



static qu_number Math_tan_complex (qu_complex x)
{
	if (qu_complex_iszero (x))
	{
		return (qu_number) QuInt0;
	}
	else
	{
		qu_number a, b;
		a = Math_sin1 (x); QU_ENSURE (a);
		b = Math_cos1 (x); QU_ENSURE (b);
		return qu_number_div_number (a, b);
	}
}



static qu_any Math_tan_inf (qu_inf x)
{
	QU_THROW (QuEMath);
}



DECLARE qu_number Math_tan (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, tan);
}



/*----------------------------------------------------------------------------
	sinh
----------------------------------------------------------------------------*/

/*) sub sinh (x: Number, precision = 0) -> Number
	Return the hyperbolic sine of v{x}.

	@code
		            R     -R
		           e  -  e
		sinh(R) = ---------   for -oo <= R <= +oo
		              2

		sinh(Z) = r + i.j

		where:
			r = sinh (Z.real) * cos (Z.imag)
			i = cosh (Z.real) * sin (Z.imag)

		sinh (infp) -> infp
		sinh (infn) -> infn

	@xmp
		use Math
		println (Math.sinh (0.2))
		println (Math.sinh (0.2, 10))

	@out
		0.20133600254109399
		0.201336002541093998950584230440128878E+0

	@see
	#{exp}
	#{sin}
	#{cos}
	#{cosh}
*/
static qu_real Math_sinhr (qu_real x)
{
	if (qu_real_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		double v = qu_real_tod (x);
		qu_real   y;

		if (errno == 0)
		{
			v = sinh (v);
			if (qu_double_isfinite (v))
				return (qu_real) qu_float_new (v);
		}

		x = (qu_real) Math_exp1 (x);           QU_ENSURE (x); assert (qu_real_is (x));
		y = qu_float_div_real   (QuFloat1, x); QU_ENSURE (y);
		x = qu_real_minus_real  (x, y);        QU_ENSURE (x);
		x = qu_real_div_float   (x, QuFloat2); QU_ENSURE (x);
		return x;
	}
}


static qu_real Math_sinh_int (qu_int x)
{
	return Math_sinhr ((qu_real) x);
}



static qu_real Math_sinh_long (qu_long x)
{
	return Math_sinhr ((qu_real) x);
}



static qu_real Math_sinh_rational (qu_rational x)
{
	return Math_sinhr ((qu_real) x);
}



static qu_real Math_sinh_float (qu_float x)
{
	return Math_sinhr ((qu_real) x);
}



static qu_real Math_sinh_big (qu_big x)
{
	qu_real a = (qu_real) x;
	qu_real b;

	a = (qu_real) Math_exp1 (a);         QU_ENSURE (a); assert (qu_real_is (a));
	b = qu_big_div_real     (QuBig1, a); QU_ENSURE (b);
	a = qu_real_minus_real  (a, b);      QU_ENSURE (a);
	a = qu_real_div_big     (a, QuBig2); QU_ENSURE (a);

	return a;
}



static qu_number Math_sinh_complex (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;
	qu_real a, b, t;

	t = (qu_real) Math_sinh1 (r);     QU_ENSURE (t); assert (qu_real_is (t));
	a = (qu_real) Math_cos1  (i);     QU_ENSURE (a); assert (qu_real_is (a));
	a = qu_real_mul_real     (t, a);  QU_ENSURE (a);
	t = (qu_real) Math_cosh1 (r);     QU_ENSURE (t); assert (qu_real_is (t));
	b = (qu_real) Math_sin1  (i);     QU_ENSURE (b); assert (qu_real_is (b));
	b = qu_real_mul_real     (t, b);  QU_ENSURE (b);

	return __qu_complex__ (a, b);
}



static qu_inf Math_sinh_inf (qu_inf x)
{
	return __qu_inf__ (qu_inf_sign (x));
}



DECLARE qu_number Math_sinh (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, sinh);
}



/*----------------------------------------------------------------------------
	cosh
----------------------------------------------------------------------------*/

/*) sub cosh (x: Number, precision = 0) -> Number
	Return the hyperbolic cosine of v{x}.

	@code
		            R     -R
		           e  +  e
		cosh(R) = ---------    for -oo <= R <= +oo
		              2

		cosh(Z) = r + i.j

		where:
			r = cosh (Z.real) * cos (Z.imag)
			i = sinh (Z.real) * sin (Z.imag)

		cosh (infp) -> infp
		cosh (infn) -> infp

	@xmp
		use Math
		println (Math.cosh (0.2))
		println (Math.cosh (0.2, 10))

	@out
		1.0200667556190759
		0.000102006675561907584853078240870046E+4

	@see
	#{exp}
	#{sin}
	#{cos}
	#{sinh}
*/
static qu_real Math_coshr (qu_real x)
{
	if (qu_real_iszero (x))
	{
		return (qu_real) QuFloat1;
	}
	else
	{
		double  v = qu_real_tod (x);
		qu_real y;

		if (errno == 0)
		{
			v = cosh (v);
			if (qu_double_isfinite (v))
				return (qu_real) qu_float_new (v);
		}

		x = (qu_real) Math_exp1 (x);           QU_ENSURE (x); assert (qu_real_is (x));
		y = qu_float_div_real   (QuFloat1, x); QU_ENSURE (y);
		x = qu_real_plus_real   (x, y);        QU_ENSURE (x);
		x = qu_real_div_float   (x, QuFloat2); QU_ENSURE (x);
		return x;
	}
}



static qu_real Math_cosh_int (qu_int x)
{
	return Math_coshr ((qu_real) x);
}



static qu_real Math_cosh_long (qu_long x)
{
	return Math_coshr ((qu_real) x);
}



static qu_real Math_cosh_rational (qu_rational x)
{
	return Math_coshr ((qu_real) x);
}



static qu_real Math_cosh_float (qu_float x)
{
	return Math_coshr ((qu_real) x);
}



static qu_real Math_cosh_big (qu_big x)
{
	qu_real a = (qu_real) x;
	qu_real b;

	a = (qu_real) Math_exp1 (a);         QU_ENSURE (a); assert (qu_real_is (a));
	b = qu_big_div_real     (QuBig1, a); QU_ENSURE (b);
	a = qu_real_plus_real   (a, b);      QU_ENSURE (a);
	a = qu_real_div_big     (a, QuBig2); QU_ENSURE (a);
	return a;
}



static qu_number Math_cosh_complex (qu_complex x)
{
	qu_real r = x->real;
	qu_real i = x->imag;
	qu_real a, b, t;

	t = (qu_real) Math_cosh1 (r);    QU_ENSURE (t); assert (qu_real_is (t));
	a = (qu_real) Math_cos1  (i);    QU_ENSURE (a); assert (qu_real_is (a));
	a = qu_real_mul_real     (t, a); QU_ENSURE (a);
	t = (qu_real) Math_sinh1 (r);    QU_ENSURE (t); assert (qu_real_is (t));
	b = (qu_real) Math_sin1  (i);    QU_ENSURE (b); assert (qu_real_is (b));
	b = qu_real_mul_real     (t, b); QU_ENSURE (b);

	return __qu_complex__ (a, b);
}



static qu_inf Math_cosh_inf (qu_inf x)
{
	return __qu_inf__ (1);
}



DECLARE qu_number Math_cosh (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, cosh);
}



/*----------------------------------------------------------------------------
	tanh
----------------------------------------------------------------------------*/

/*) sub tanh (x: Number, precision = 0) -> Number
	Return the hyperbolic tangent of v{x}.

	@code
		tanh (X) = sinh (X) / cosh (X)

		tanh (infp) ->  1
		tanh (infn) -> -1

	@xmp
		use Math
		println (Math.tanh (0.2))
		println (Math.tanh (0.2, 10))

	@out
		0.19737532022490401
		0.197375320224904011407877792245269369E+0
*/
static qu_real Math_tanhr (qu_real x)
{
	if (qu_real_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		qu_real a, b;

		a = (qu_real) Math_sinh1 (x);    QU_ENSURE (a); assert (qu_real_is (a));
		b = (qu_real) Math_cosh1 (x);    QU_ENSURE (b); assert (qu_real_is (b));
		a = qu_real_div_real     (a, b); QU_ENSURE (a);
		return a;
	}
}



static qu_real Math_tanh_int (qu_int x)
{
	return Math_tanhr ((qu_real) x);
}



static qu_real Math_tanh_long (qu_long x)
{
	return Math_tanhr ((qu_real) x);
}



static qu_real Math_tanh_rational (qu_rational x)
{
	return Math_tanhr ((qu_real) x);
}



static qu_real Math_tanh_float (qu_float x)
{
	return Math_tanhr ((qu_real) x);
}



static qu_real Math_tanh_big (qu_big x)
{
	return Math_tanhr ((qu_real) x);
}



static qu_number Math_tanh_complex (qu_complex x)
{
	qu_number a, b;

	a = Math_sinh1 (x); QU_ENSURE (a);
	b = Math_cosh1 (x); QU_ENSURE (b);
	return qu_number_div_number (a, b);
}



static qu_int Math_tanh_inf (qu_inf x)
{
	return (x == QuInfPos) ? QuInt1 : QuInt_1;
}



DECLARE qu_number Math_tanh (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, tanh);
}



/*----------------------------------------------------------------------------
	asin
----------------------------------------------------------------------------*/

/*) sub asin (x: Number, precision = 0) -> Number
	Return the inverse sine of v{x}.

	@code
		asin (X) = -1j * log (1j * X + sqrt (1 - X * X))

	@xmp
		use Math
		println (Math.asin (0.2))
		println (Math.asin (0.2, 10))

	@out
		0.2013579207903308
		0.2013579207903308027862918481386094730007E+0

	@p
	As you can see from the series, for very small values of v{x} then the result
	is v{x} unless the v{precision} is smaller than v{x} itself.
*/
static qu_number Math_asind         (double v);
static qu_number Math_asin_int      (qu_int x);
static qu_number Math_asin_float    (qu_float x);
static qu_number Math_asin_long     (qu_long x);
static qu_number Math_asin_rational (qu_rational x);
static qu_number Math_asin_big      (qu_big x);
static qu_number Math_asin_complex  (qu_complex x);
static qu_number Math_asin_inf      (qu_inf x);



static qu_number Math_asind (double v)
{
	if (v == 0)
	{
		return (qu_number) QuInt0;
	}
	else
	{
		v = asin (v);
		if (qu_double_isnan (v))
			return Math_asin_complex (qu_complex_new1 ((qu_real) qu_float_new (v)));
		assert (qu_double_isfinite (v));
		return (qu_number) qu_float_new (v);
	}
}



static qu_number Math_asin_int (qu_int x)
{
	return Math_asind (qu_int_tol (x));
}



static qu_number Math_asin_float (qu_float x)
{
	return Math_asind (qu_float_tod (x));
}



static qu_number Math_asin_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_asin_big (qu_long_as_big (x))
		: Math_asind (v);
}



static qu_number Math_asin_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_asin_big (qu_rational_as_big (x))
		: Math_asind (v);
}



static qu_number Math_asin_big (qu_big x)
{
	if (qu_big_iszero ((qu_big) x))
	{
		/* x == 0 -> 0 */
		return (qu_number) QuInt0;
	}
	else
	{
		int c = qu_big_cmpabs_big (x, QuBig1);

		/* |x| > 1 -> qu_complex */
		if (c > 0)
			return Math_asin_complex (qu_complex_new1 ((qu_real) x));

		if (c == 0)
		{
			/*
				x ==  1 ->  PI/2
				x == -1 -> -PI/2
			*/

			x = _math_pi_big (); QU_ENSURE (x);
			x = (qu_big) qu_big_mul_big (x, QuBig1_2); QU_ENSURE (x); assert (qu_big_is (x));

			return qu_big_isneg (x) ? (qu_number) qu_big_neg (x) : (qu_number) x;
		}
		else
		{
			/*
				                    3         5           7
				                1  x     1 3 x     1 3 5 x
				asin(x) = x  +  -  -  +  - - -  +  - --- -  +  ...
				                2  3     2 4 5     2 4 6 7
			*/

			qu_big f, d, w, i, y, z, t, u, p;
			int ctx;

			w = (qu_big) qu_big_mul_big (x, x); assert (w && qu_big_is (w));
			f = x;
			d = QuBig0;
			i = y = QuBig1;
			z = QuBig2;

			p = qu_big_make1 (1, -__QuBigPrecision);

		  __qu_number_context_begin (ctx, __QuNumberContext | QU_NUMBER_CONTEXT_ALLOW_TINY);

			for (;;)
			{
				f = (qu_big) qu_big_mul_big  (f, w);      assert (f && qu_big_is (f));
				d = (qu_big) qu_big_mul_big  (f, y);      assert (d && qu_big_is (d));
				i = (qu_big) qu_big_plus_big (i, QuBig2); assert (i && qu_big_is (i));
				u = (qu_big) qu_big_mul_big  (z, i);      assert (u && qu_big_is (u));
				d = (qu_big) qu_big_div_big  (d, u);      assert (d && qu_big_is (d));

				if (qu_big_cmpabs_big (d, p) <= 0)
					break;

				x = (qu_big) qu_big_plus_big (x, d);  assert (x && qu_big_is (x));
				y = (qu_big) qu_big_mul_big  (y, i);  assert (y && qu_big_is (y));
				t = (qu_big) qu_big_incr     (i);     assert (t && qu_big_is (t));
				z = (qu_big) qu_big_mul_big  (z, t);  assert (z && qu_big_is (z));

				CHECK_INTERRUPT_GOTO (interrupted);
			}

		  __qu_number_context_end (ctx);
			return (qu_number) x;

		  interrupted:
		  __qu_number_context_end (ctx);
			QU_RETHROW;
		}
	}
}



static qu_number Math_asin_complex (qu_complex x)
{
	if (qu_real_iszero (x->imag))
	{
		qu_real a;

		a = (qu_real) Math_asin1 (x->real); QU_ENSURE (a); assert (qu_real_is (a));
		return __qu_complex__ (a, (qu_real) QuInt0);
	}
	else
	{
		qu_number a, b;

		a = qu_complex_mul_complex (x, x);           QU_ENSURE (a);
		a = qu_int_minus_number    (QuInt1, a);      QU_ENSURE (a);
		a = Math_sqrt1             (a);              QU_ENSURE (a);
		b = qu_complex_mul_complex (QuComplex01, x); QU_ENSURE (b);
		a = qu_number_plus_number  (a, b);           QU_ENSURE (a);
		a = Math_log1              (a);              QU_ENSURE (a);

		return qu_complex_mul_number (QuComplex0_1, a);
	}
}



static qu_number Math_asin_inf (qu_inf x)
{
	qu_decimal p;

	p = Math_PI (0, NULL); QU_ENSURE (p);
	p = (qu_decimal) qu_decimal_div_int (p, QuInt2); QU_ENSURE (p); assert (qu_decimal_is (p));

	if (x == QuInfNeg)
		p = (qu_decimal) qu_decimal_neg (p);

	return __qu_complex__ ((qu_real) p, (qu_real) QuInfPos);
}



DECLARE qu_number Math_asin (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, asin);
}



/*----------------------------------------------------------------------------
	acos
----------------------------------------------------------------------------*/

/*) sub acos (x: Number, precision = 0) -> Number
	Return the inverse cosine of v{x}.

	@code
		acos(X) = (PI / 2) - asin(X)

	@xmp
		use Math
		println (Math.acos (0.2))
		println (Math.acos (0.2, 10))

	@out
		1.3694384060045659
		0.0001369438406004565816445029843501140527E+4
*/
static qu_number Math_acosd         (double v);
static qu_number Math_acos_int      (qu_int x);
static qu_number Math_acos_float    (qu_float x);
static qu_number Math_acos_long     (qu_long x);
static qu_number Math_acos_rational (qu_rational x);
static qu_number Math_acos_big      (qu_big x);
static qu_number Math_acos_complex  (qu_complex x);
static qu_number Math_acos_inf      (qu_inf x);



static qu_number Math_acosd (double v)
{
	if (v == 1)
	{
		return (qu_number) QuInt0;
	}
	else
	{
		v = acos (v);
		if (qu_double_isnan (v))
			return Math_acos_complex (qu_complex_new1 ((qu_real) qu_float_new (v)));
		assert (qu_double_isfinite (v));
		return (qu_number) qu_float_new (v);
	}
}



static qu_number Math_acos_int (qu_int x)
{
	return Math_acosd (qu_int_tol (x));
}



static qu_number Math_acos_float (qu_float x)
{
	return Math_acosd (qu_float_tod (x));
}



static qu_number Math_acos_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_acos_big (qu_long_as_big (x))
		: Math_acosd (v);
}



static qu_number Math_acos_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_acos_big (qu_rational_as_big (x))
		: Math_acosd (v);
}



static qu_number Math_acos_big (qu_big x)
{
	qu_big  a;
	qu_real b;

	b = (qu_real) Math_asin1    (x);           QU_ENSURE (b); assert (qu_real_is (b));
	a = _math_pi_big            ();            assert (b && qu_big_is (b));
	a = (qu_big) qu_big_mul_big (a, QuBig1_2); QU_ENSURE (a); assert (qu_big_is (a));
	b = qu_big_minus_real       (a, b);        QU_ENSURE (b); assert (qu_real_is (b));
	return (qu_number) b;
}



static qu_number Math_acos_complex (qu_complex x)
{
	qu_number  a;
	qu_decimal p;

	p = COMPLEX_IS_BIG (x) ? (qu_decimal) _math_pi_big () : (qu_decimal) MathPiFloat;

	p = (qu_decimal) qu_decimal_div_int  (p, QuInt2); QU_ENSURE (p); assert (qu_decimal_is (p));
	a = Math_asin1 (x); QU_ENSURE (a);

	return (qu_number) qu_decimal_minus_number (p, a);
}



static qu_number Math_acos_inf (qu_inf x)
{
	if (x == QuInfPos)
		return __qu_complex__ ((qu_real) QuInt0, (qu_real) QuInfNeg);
	else
	{
		qu_decimal pi = Math_PI (0, NULL);

		QU_ENSURE (pi);
		return __qu_complex__ ((qu_real) pi, (qu_real) QuInfNeg);
	}
}



DECLARE qu_number Math_acos (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, acos);
}



/*----------------------------------------------------------------------------
	atan
----------------------------------------------------------------------------*/

/*) sub atan (x: Number, precision = 0) -> Number
	Return the arc tangent of v{x}.

	@code
		atan (R) = ATAN (R)
		atan (Z) = 1j * log ((1j + Z) / (1j - Z)) / 2

		atan (infp) ->  PI/2
		atan (infn) -> -PI/2

	@p
	If required, computation uses the Maclaurin series:
	@code
		           oo
		          ___       n + 1
		          \     (-1)       2n -1
		atan(x) = /     --------- x
		          ---     2n - 1
		          n=1

		              1  3   1  5   1  7
		        = x - - x  + - x  - - x  + ...  for  -1 < x < 1
		              3      5      7


		atan(x) = -PI/2 - atan (1/x) for x < 0
		        =  PI/2 - atan (1/x) for x > 0

	@p
	As you can see from the series, for very small values of v{x} then the result
	is v{x} unless the v{precision} is smaller than v{x} itself.

	@xmp
		use Math
		println (Math.atan (0.2))

	@out
		0.197396
*/
/*
	atan (infp) = PI/2
	atan (infn) = -PI/2
	atan (j)    = infp.j
	atan (-j)   = infn.j
	atan (0)    = 0
*/
static qu_float   Math_atand         (double v);
static qu_float   Math_atan_int      (qu_int x);
static qu_decimal Math_atan_long     (qu_long x);
static qu_decimal Math_atan_rational (qu_rational x);
static qu_float   Math_atan_float    (qu_float x);
static qu_big     Math_atan_big      (qu_big x);
static qu_number  Math_atan_complex  (qu_complex x);
static qu_decimal Math_atan_inf      (qu_inf x);



static qu_float Math_atand (double v)
{
	v = atan (v);
	assert (qu_double_isfinite (v));
	return qu_float_new (v);
}



static qu_float Math_atan_int (qu_int x)
{
	return Math_atand (qu_int_tod (x));
}



static qu_float Math_atan_float (qu_float x)
{
	return Math_atand (qu_float_tod (x));
}



static qu_decimal Math_atan_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? (qu_decimal) Math_atan_big (qu_long_as_big (x))
		: (qu_decimal) Math_atand (v);
}



static qu_decimal Math_atan_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? (qu_decimal) Math_atan_big (qu_rational_as_big (x))
		: (qu_decimal) Math_atand (v);
}



static qu_number Math_atan_complex (qu_complex x)
{
	if (qu_real_iszero (x->real) && qu_real_isone (x->imag))
	{
		if (qu_real_isneg (x->imag))
		{
			return __qu_complex__ ((qu_real) QuInt0, (qu_real) QuInfNeg);
		}
		else
		{
			return __qu_complex__ ((qu_real) QuInt0, (qu_real) QuInfPos);
		}
	}
	else
	{
		qu_number a, b;

		a = qu_complex_plus_complex  (QuComplex01, x);  QU_ENSURE (a);
		b = qu_complex_minus_complex (QuComplex01, x);  QU_ENSURE (b);
		a = qu_number_div_number     (a, b);            QU_ENSURE (a);
		a = Math_log1                (a);               QU_ENSURE (a);
		a = qu_complex_mul_number    (QuComplex01, a);  QU_ENSURE (a);

		return qu_number_div_int (a, QuInt2);
	}
}



static qu_big _math_atan_big1 (qu_big x)
{
	qu_big x2, i, p, m, prec;
	int g = 0;

	assert (qu_big_cmpabs_big (x, QuBig1));

	prec = qu_big_make1 (1, -__QuBigPrecision);
	i    = QuBig3;
	p    = x;
	x2   = (qu_big) qu_big_mul_big (x, x); QU_ENSURE (x2); QU_REQUIRE (QuEBrain, qu_big_is (x2));

	for (;;)
	{
		p = (qu_big) qu_big_mul_big (p, x2); QU_ENSURE (p); QU_REQUIRE (QuEBrain, qu_big_is (p));
		m = (qu_big) qu_big_div_big (p, i);  QU_ENSURE (m); QU_REQUIRE (QuEBrain, qu_big_is (m));

		g = !g;
		if (g)
		{
			x = (qu_big) qu_big_minus_big (x, m); QU_ENSURE (x); QU_REQUIRE (QuEBrain, qu_big_is (x));
		}
		else
		{
			x = (qu_big) qu_big_plus_big (x, m); QU_ENSURE (x); QU_REQUIRE (QuEBrain, qu_big_is (x));
		}

		if (qu_big_cmpabs_big (m, prec) <= 0)
			break;

		i = (qu_big) qu_big_plus_big (i, QuBig2); QU_ENSURE (i); QU_REQUIRE (QuEBrain, qu_big_is (i));

		CHECK_INTERRUPT;
	}
	return x;
}



static qu_big Math_atan_big (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return QuBig0;
	}
	else if (qu_big_isone (x))
	{
		qu_big p = _math_pi_big ();

		if (qu_big_isneg (x))
			return (qu_big) qu_big_mul_big (p, QuBig1_4);
		else
			return (qu_big) qu_big_mul_big (p, QuBig_1_4);
	}
	else if (qu_big_cmpabs_big (x, QuBig1) < 0)
	{
		return _math_atan_big1 (x);
	}
	else
	{
		qu_big p, r;

		r = (qu_big) qu_big_div_big  (QuBig1, x);   QU_ENSURE (r); QU_REQUIRE (QuEBrain, qu_big_is (r));
		r = _math_atan_big1          (r);           QU_ENSURE (r);
		p = _math_pi_big             ();            QU_ENSURE (p);
		p = (qu_big) qu_big_mul_big  (p, QuBig1_2); QU_ENSURE (p); QU_REQUIRE (QuEBrain, qu_big_is (p));

		if (qu_big_isneg (x))
		{
			r = (qu_big) qu_big_plus_big  (p, r); QU_ENSURE (r); QU_REQUIRE (QuEBrain, qu_big_is (r));
			r = qu_big_neg (r);
		}
		else
		{
			r = (qu_big) qu_big_minus_big (p, r); QU_ENSURE (r); QU_REQUIRE (QuEBrain, qu_big_is (r));
		}
		return r;
	}
}



static qu_decimal Math_atan_inf (qu_inf x)
{
	qu_decimal pi = Math_PI (0, NULL);

	if (x == QuInfPos)
		return (qu_decimal) qu_decimal_div_int (pi, QuInt2);

	else if (x == QuInfNeg)
		return (qu_decimal) qu_decimal_div_int (pi, QuInt_2);

	QU_THROW (QuEMath);
}



DECLARE qu_number Math_atan (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, atan);
}



/*----------------------------------------------------------------------------
	atan2
----------------------------------------------------------------------------*/

/*) sub atan2 (y: Number, x: Number, precision = 0) -> Number
	This is similar to calculating #{atan} except that the signs of both
	arguments are used to determine the quadrant of the result.

	@xmp
		use Math
		println (Math.atan2 (0.9, 1))
		println (Math.atan2 (0.9, -1))

	@out
		0.732815
*/
static qu_real Math_atan2_big (qu_big y, qu_big x)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	if (sx == 0)
	{
		if (sy == 0)
		{
			/* atan2 (0, 0) = 0 */
			return (qu_real) QuInt0;
		}
		else
		{
		  infinity:

			/*
				atan (infp) =  PI /  2
				atan (infn) =  PI / -2
			*/

			x = _math_pi_big (); QU_ENSURE (x);

			return qu_big_isneg (y)
				? qu_big_mul_big (x, QuBig_1_2)
				: qu_big_mul_big (x, QuBig1_2);
		}
	}
	else if (sy == 0)
	{
		return (sx < 0) ? (qu_real) _math_pi_big () : (qu_real) QuInt0;
	}
	else
	{
		/*
			We want to return +/- PI/2 if the result is infinity.
			We must not tweak the context: suppose the result is huge
			and the context says that infinity is not allowed, then this
			means the programmer does not consider that huge number as
			infinity and so we must treat it as an error as well.
		*/

		qu_real r = qu_big_div_big (y, x);

		QU_ENSURE (r); assert (qu_real_is (r));

		if (qu_inf_is (r))
		{
			goto infinity;
		}
		else
		{
			qu_big a = (qu_big) r;

			assert (qu_big_is (a));
			if (qu_big_isneg (a))
				a = qu_big_neg (a);

			a = Math_atan_big (a); QU_ENSURE (a);

			/*
				y+  x+  -> a
				y+  x-  -> PI - a
				y-  x+  -> -a
				y-  x-  -> a - PI
			*/

			if (sy < 0)
			{
				if (sx < 0)
				{
					qu_big pi = _math_pi_big (); QU_ENSURE (pi);
					return qu_big_minus_big (a, pi);
				}
				else
				{
					return (qu_real) qu_big_neg (a);
				}
			}
			else
			{
				if (sx < 0)
				{
					qu_big pi = _math_pi_big (); QU_ENSURE (pi);
					return qu_big_minus_big (pi, a);
				}
				else
				{
					return (qu_real) a;
				}
			}
		}
	}
}



static qu_number Math_atan2_complex (qu_complex x, qu_complex y)
{
	/*
		-1j * log ((x + 1j * y) / sqrt (x * x + y * y))
	*/

	qu_number a, b;

	a = qu_complex_mul_complex  (x, x);           QU_ENSURE (a);
	b = qu_complex_mul_complex  (y, y);           QU_ENSURE (b);
	b = qu_number_plus_number   (a, b);           QU_ENSURE (b);
	b = Math_sqrt1              (b);              QU_ENSURE (b);
	a = qu_complex_mul_complex  (QuComplex01, y); QU_ENSURE (a);
	a = qu_complex_plus_number  (x, a);           QU_ENSURE (a);
	a = qu_number_div_number    (a, b);           QU_ENSURE (a);
	a = Math_log1               (a);              QU_ENSURE (a);

	return qu_complex_mul_number (QuComplex01, a);
}




static qu_real Math_atan2_real (qu_real y, qu_real x)
{
	int sx  = qu_real_sign (x);
	int sy  = qu_real_sign (y);
	int big = qu_big_is (y) || qu_big_is (x);

	if (sx == 0)
	{
		if (sy == 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
		  infinity:

			{
				qu_decimal pi = big ? (qu_decimal) _math_pi_big () : (qu_decimal) MathPiFloat;
				QU_ENSURE (pi);

				return qu_real_isneg (y)
					? qu_decimal_div_int (pi, QuInt_2)
					: qu_decimal_div_int (pi, QuInt2);
			}
		}
	}
	else if (sy == 0)
	{
		if (sx < 0)
			return big ? (qu_real) _math_pi_big () : (qu_real) MathPiFloat;
		else
			return (qu_real) QuInt0;
	}
	else
	{
		qu_real r = qu_real_div_real (y, x);

		QU_ENSURE (r); assert (qu_real_is (r));

		if (qu_inf_is (r))
		{
			goto infinity;
		}
		else
		{
			qu_real angle;

			if (qu_real_isneg (r))
				r = qu_real_neg (r);

			angle = (qu_real) Math_atan1 (r);

			QU_ENSURE (angle);
			assert (qu_real_is (angle));

			/*
				y+  x+  -> angle
				y+  x-  -> PI - angle
				y-  x+  -> -angle
				y-  x-  -> angle - PI
			*/

			if (sy < 0)
			{
				if (sx < 0)
				{
					qu_decimal pi = (qu_decimal) MathPiFloat;

					if (big)
					{
						pi = (qu_decimal) _math_pi_big ();
						QU_ENSURE (pi);
					}
					return qu_real_minus_decimal (angle, pi);
				}
				else
				{
					return qu_real_neg (angle);
				}
			}
			else
			{
				if (sx < 0)
				{
					qu_decimal pi = (qu_decimal) MathPiFloat;

					if (big)
					{
						pi = (qu_decimal) _math_pi_big ();
						QU_ENSURE (pi);
					}
					return qu_decimal_minus_real (pi, angle);
				}
				else
				{
					return angle;
				}
			}
		}
	}
}



DECLARE qu_number Math_atan2 (int argc, qu_any *argv)
{
	qu_number x, y;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, y, qu_number);
	QU_ARG_VTYPE  (1, x, qu_number);

	if (argc == 3)
	{
		long prec = 0;

		QU_ARG_vlong (2, prec);

		if (prec)
		{
			long old = __QuBigPrecision;

			ENSURE_VALID_PRECISION (prec);
			__QuBigPrecision = prec;

			if (qu_complex_is (y) || qu_complex_is (x))
			{
				if (qu_complex_is (y))
					y = (qu_number) _math_complex_big ((qu_complex) y);
				else
					y = (qu_number) qu_complex_new1 ((qu_real) qu_real_as_big ((qu_real) y));

				if (qu_complex_is (x))
					x = (qu_number) _math_complex_big ((qu_complex) x);
				else
					x = (qu_number) qu_complex_new1 ((qu_real) qu_real_as_big ((qu_real) x));

				x = (qu_number) Math_atan2_complex ((qu_complex) y, (qu_complex) x);
			}
			else
			{
				assert (qu_real_is (x) && qu_real_is (y));
				x = (qu_number) Math_atan2_big (qu_real_as_big ((qu_real) y), qu_real_as_big ((qu_real) x));
			}
			__QuBigPrecision = old;
			return x;
		}
	}

	if (qu_complex_is (y) || qu_complex_is (x))
	{
		if (!qu_complex_is (y))
			y = (qu_number) qu_complex_new1 ((qu_real) qu_real_as_big ((qu_real) y));

		if (!qu_complex_is (x))
			x = (qu_number) qu_complex_new1 ((qu_real) qu_real_as_big ((qu_real) x));

		return (qu_number) Math_atan2_complex ((qu_complex) y, (qu_complex) x);
	}
	else
	{
		assert (qu_real_is (x) && qu_real_is (y));
		return (qu_number) Math_atan2_real ((qu_real) y, (qu_real) x);
	}
}



/*----------------------------------------------------------------------------
	asinh
----------------------------------------------------------------------------*/

/*) sub asinh (x: Number, precision = 0) -> Number
	Return the inverse hyperbolic sine of v{x}.

	@code
		asinh (X) = ln (X + sqrt (1 + X * X))

	@xmp
		use Math
		println (Math.asinh (0.5))
		println (Math.asinh (0.5, 10))

	@out
		0.48121182505960347
		0.4812118250596034474977589131E+0
*/
static qu_number Math_asinh_complex (qu_complex x);
static qu_number Math_asinh_big     (qu_big x);

static qu_number Math_asinhd (double v)
{
	if (v == 0)
	{
		return (qu_number) QuInt0;
	}
	else
	{
		double r = asinh (v);

		if (qu_double_isnan (r))
			return Math_asin_complex (qu_complex_new1 ((qu_real) qu_float_new (v)));

		if (qu_double_isfinite (r))
			return (qu_number) qu_float_new (r);

		return (qu_number) __qu_inf__ ((r < 0) ? -1 : 1);
	}
}


static qu_number Math_asinh_int (qu_int x)
{
	return Math_asinhd (qu_int_tod (x));
}



static qu_number Math_asinh_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_asinh_big (qu_long_as_big (x))
		: Math_asinhd (v);
}



static qu_number Math_asinh_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_asinh_big (qu_rational_as_big (x))
		: Math_asinhd (v);
}



static qu_number Math_asinh_float (qu_float x)
{
	return Math_asinhd (qu_float_tod (x));
}



static qu_number Math_asinh_big (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return (qu_number) QuInt0;
	}
	else
	{
		qu_real r;

		r = qu_big_mul_big   (x, x);  QU_ENSURE (r); assert (qu_real_is (r));
		r = qu_real_incr     (r);     QU_ENSURE (r); assert (qu_real_is (r));
		r = Math_sqrt1r      (r);     QU_ENSURE (r); assert (qu_real_is (r));
		r = qu_big_plus_real (x, r);  QU_ENSURE (r); assert (qu_real_is (r));

		return Math_log1 (r);
	}
}



static qu_number Math_asinh_complex (qu_complex x)
{
	qu_number r;

	r = (qu_number) qu_complex_mul_complex (x, x);      QU_ENSURE (r);
	r = (qu_number) qu_number_plus_int     (r, QuInt1); QU_ENSURE (r);
	r = (qu_number) Math_sqrt1             (r);         QU_ENSURE (r);
	r = (qu_number) qu_complex_plus_number (x, r);      QU_ENSURE (r);

	return Math_log1 (r);
}



static qu_number Math_asinh_inf (qu_inf x)
{
	return (qu_number) __qu_inf__ (qu_inf_sign (x));
}



DECLARE qu_number Math_asinh (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, asinh);
}



/*----------------------------------------------------------------------------
	acosh
----------------------------------------------------------------------------*/

/*) sub acosh (x: Number, precision = 0) -> Number
	Return the inverse hyperbolic cosine of v{x}.

	@code
		acosh (X) = ln (X + sqrt (X + 1) sqrt (X - 1))

		acosh (R) = ln (R - sqrt (R * R - 1)) for R < -1
		            ln (R + sqrt (R * R - 1)) for R > 0

	@xmp
		use Math
		println (Math.acosh (1.0))

	@out
		0.0
*/
static qu_number Math_acosh_big     (qu_big x);
static qu_number Math_acosh_complex (qu_complex x);

static qu_number Math_acoshd (double v)
{
	if (v == 1)
	{
		return (qu_number) QuInt0;
	}
	else
	{
		double r = acosh (v);

		if (qu_double_isnan (r))
			return Math_acosh_complex (qu_complex_new1 ((qu_real) qu_float_new (v)));

		assert (qu_double_isfinite (r));
		return (qu_number) qu_float_new (r);
	}
}



static qu_number Math_acosh_int (qu_int x)
{
	return Math_acoshd (qu_int_tod (x));
}



static qu_number Math_acosh_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_acosh_big (qu_long_as_big (x))
		: Math_acoshd (v);
}



static qu_number Math_acosh_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_acosh_big (qu_rational_as_big (x))
		: Math_acoshd (v);
}



static qu_number Math_acosh_float (qu_float x)
{
	return Math_acoshd (qu_float_tod (x));
}



static qu_number Math_acosh_big (qu_big x)
{
	if (qu_big_is1 (x))
	{
		return (qu_number) QuInt0;
	}
	else if (qu_big_cmp_big (x, QuBig_1) < 0)
	{
		qu_real r;

		r = qu_big_mul_big    (x, x);       QU_ENSURE (r); assert (r && qu_real_is (r));
		r = qu_real_minus_int (r, QuInt1);  QU_ENSURE (r); assert (r && qu_real_is (r));
		r = Math_sqrt1r       (r);          QU_ENSURE (r); assert (r && qu_real_is (r));
		r = qu_big_minus_real (x, r);       QU_ENSURE (r); assert (r && qu_real_is (r));

		return Math_log1 (r);
	}
	else if (qu_big_cmp_big (x, QuBig0) > 0)
	{
		qu_real r;

		r = qu_big_mul_big    (x, x);       QU_ENSURE (r); assert (r && qu_real_is (r));
		r = qu_real_minus_int (r, QuInt1);  QU_ENSURE (r); assert (r && qu_real_is (r));
		r = Math_sqrt1r       (r);          QU_ENSURE (r); assert (r && qu_real_is (r));
		r = qu_big_plus_real  (x, r);       QU_ENSURE (r); assert (r && qu_real_is (r));

		return Math_log1 (r);
	}
	else
	{
		return Math_acosh_complex (qu_complex_new1 ((qu_real) x));
	}
}



static qu_number Math_acosh_complex (qu_complex x)
{
	qu_number a, b;

	a = qu_complex_plus_int  (x, QuInt1); QU_ENSURE (a);
	a = Math_sqrt1           (a);         QU_ENSURE (a);
	b = qu_complex_minus_int (x, QuInt1); QU_ENSURE (b);
	b = Math_sqrt1           (b);         QU_ENSURE (b);

	a = qu_number_mul_number   (a, b); QU_ENSURE (a);
	b = qu_complex_plus_number (x, a); QU_ENSURE (b);

	return Math_log1 (b);
}



static qu_number Math_acosh_inf (qu_inf x)
{
	if (x == QuInfPos)
	{
		return (qu_number) __qu_inf__ (1);
	}
	else
	{
		qu_decimal pi = Math_PI (0, NULL);
		return __qu_complex__ ((qu_real) QuInfPos, (qu_real) pi);
	}
}



DECLARE qu_number Math_acosh (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, acosh);
}



/*----------------------------------------------------------------------------
	atanh
----------------------------------------------------------------------------*/

/*) sub atanh (x: Number, precision = 0) -> Number
	Return the inverse hyperbolic tangent of v{x}.

	@code
		atanh (X) = log ((1 + X) / (1 - X)) / 2

	@xmp
		use Math
		println (Math.atanh (0.5))
		println (Math.atanh (0.5, 10))

	@out
		0.54930614433405489
		0.5493061443340548456976226184612628491104E+0
*/
static qu_number Math_atanh_big     (qu_big x);
static qu_number Math_atanh_complex (qu_complex x);

static qu_number Math_atanhd (double v)
{
	if (v == 0)
	{
		return (qu_number) QuInt0;
	}
	else if (v == 1)
	{
		return (qu_number) __qu_inf__ (1);
	}
	else if (v == -1)
	{
		return (qu_number) __qu_inf__ (-1);
	}
	else
	{
		double r = atanh (v);

		if (qu_double_isnan (r))
			return Math_atanh_complex (qu_complex_new1 ((qu_real) qu_float_new (v)));

		assert (qu_double_isfinite (r));
		return (qu_number) qu_float_new (r);
	}
}



static qu_number Math_atanh_int (qu_int x)
{
	return Math_atanhd (qu_int_tod (x));
}



static qu_number Math_atanh_long (qu_long x)
{
	double v = qu_long_tod (x);
	return errno
		? Math_atanh_big (qu_long_as_big (x))
		: Math_atanhd (v);
}



static qu_number Math_atanh_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	return errno
		? Math_atanh_big (qu_rational_as_big (x))
		: Math_atanhd (v);
}



static qu_number Math_atanh_float (qu_float x)
{
	return Math_atanhd (qu_float_tod (x));
}



static qu_number Math_atanh_big (qu_big x)
{
	qu_big r, w, f, i, p, d;

	p = qu_big_make1 (1, -__QuBigPrecision);
	i = QuBig3;
	r = f = x;
	w = (qu_big) qu_big_mul_big (x, x); QU_ENSURE (w); QU_REQUIRE (QuEBrain, qu_big_is (w));

	for (;;)
	{
		f = (qu_big) qu_big_mul_big (f, w); QU_ENSURE (f); QU_REQUIRE (QuEBrain, qu_big_is (f));
		d = (qu_big) qu_big_div_big (f, i); QU_ENSURE (d); QU_REQUIRE (QuEBrain, qu_big_is (d));

		if (qu_big_cmpabs_big (d, p) <= 0)
			break;

		r = (qu_big) qu_big_plus_big (r, d);      QU_ENSURE (r); QU_REQUIRE (QuEBrain, qu_big_is (r));
		i = (qu_big) qu_big_plus_big (i, QuBig2); QU_ENSURE (i); QU_REQUIRE (QuEBrain, qu_big_is (i));

		CHECK_INTERRUPT;
	}
	return (qu_number) r;
}



static qu_number Math_atanh_complex (qu_complex x)
{
	qu_number a, b;

	a = qu_int_plus_complex  (QuInt1, x); QU_ENSURE (a);
	b = qu_int_minus_complex (QuInt1, x); QU_ENSURE (b);
	a = qu_number_div_number (a, b);      QU_ENSURE (a);
	a = Math_log1            (a);         QU_ENSURE (a);
	return qu_number_div_int (a, QuInt2);
}



static qu_number Math_atanh_inf (qu_inf x)
{
	qu_decimal i;
	qu_real    v;

	i = Math_PI (0, NULL);
	v = (qu_real) qu_decimal_neg (i);
	v = qu_real_div_int (v, QuInt2);
	return __qu_complex__ ((qu_real) QuInt0, (qu_real) v);
}



DECLARE qu_number Math_atanh (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, atanh);
}



/*----------------------------------------------------------------------------
	deg
----------------------------------------------------------------------------*/

/*) sub deg (x: Number, precision = 0) -> Number
	Convert v{x} radians into degrees.

	For Complex numbers both the real and imaginary parts are converted.

	@code
		deg (0)     ->  0
		deg (infp)  ->  infp
		deg (infn)  ->  infn

	@xmp
		use Math
		println (Math.deg (1))
		println (Math.deg (1 + 2j))

	@out
		57.295779513082323
		(57.295779513082323 + 114.59155902616465j)
*/
static qu_real Math_degr (qu_any xx)
{
	qu_real x = (qu_any) xx;

	assert (qu_real_is (x));

	if (qu_real_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		qu_decimal pi;
		qu_real    r;

		pi = Math_PI (0, NULL); QU_ENSURE (pi);
		r  = qu_int_div_decimal (qu_int_new (180), pi); assert (r && qu_real_is (r));
		return qu_real_mul_real (r, x);
	}
}

#define Math_deg_int      Math_degr
#define Math_deg_long     Math_degr
#define Math_deg_rational Math_degr
#define Math_deg_float    Math_degr
#define Math_deg_big      Math_degr


static qu_inf Math_deg_inf (qu_inf x)
{
	return __qu_inf__ (qu_inf_sign (x));
}



static qu_number Math_deg_complex (qu_complex x)
{
	qu_real r, i;
	r = Math_degr (x->real); QU_ENSURE (r);
	i = Math_degr (x->imag); QU_ENSURE (i);
	return __qu_complex__ (r, i);
}



DECLARE qu_number Math_deg (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, deg);
}



/*----------------------------------------------------------------------------
	rad
----------------------------------------------------------------------------*/

/*) sub rad (x: Number, precision = 0) -> Number
	Convert v{x} degrees into radians.

	For Complex numbers both the real and imaginary parts are converted.

	@code
		rad (0)     ->  0
		rad (infp)  ->  infp
		rad (infn)  ->  infn

	@xmp
		use Math
		println (Math.rad (57.295779513082323))
		println (Math.rad (57.295779513082323 + 114.59155902616465j))

	@out
		1
		(1 + 2j)
*/
static qu_real Math_radr (qu_any xx)
{
	qu_real x = (qu_any) xx;

	assert (qu_real_is (x));

	if (qu_real_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		qu_decimal pi;
		qu_real    r;

		pi = Math_PI (0, NULL); QU_ENSURE (pi);
		r  = qu_decimal_div_int (pi, qu_int_new (180)); assert (r && qu_real_is (r));
		return qu_real_mul_real (r, x);
	}
}



#define Math_rad_int      Math_radr
#define Math_rad_long     Math_radr
#define Math_rad_rational Math_radr
#define Math_rad_float    Math_radr
#define Math_rad_big      Math_radr



static qu_number Math_rad_complex (qu_complex x)
{
	qu_real r, i;
	r = Math_radr (x->real); QU_ENSURE (r);
	i = Math_radr (x->imag); QU_ENSURE (i);
	return __qu_complex__ (r, i);
}



static qu_inf Math_rad_inf (qu_inf x)
{
	return __qu_inf__ (qu_inf_sign (x));
}



DECLARE qu_number Math_rad (int argc, qu_any *argv)
{
	MATH_WITH_PREC (qu_number, rad);
}



/*----------------------------------------------------------------------------
	rand
----------------------------------------------------------------------------*/

/*) sub rand (*arg) -> Int
	@def
	sub rand -> Int
	sub rand (lo, hi = RAND_MAX) -> Int
	sub rand (hi) -> Int
	sub rand (r: Range) -> Int

	@p
	Return a random number.

	The first form returns a number between C{0 .. RAND_MAX} where v{RAND_MAX}
	is a system defined constant.

	The second form returns a number between v{lo} and v{hi} inclusive, while
	the third form returns a number between 0 and v{hi} inclusive.
	The Math module will interpret the values as best as possible to
	avoid any exception.

	The last form returns a number within the range. The number is guaranteed
	to be a value in the range, e.g. C{assert (result in range)} always succeed.

	@noexc

	@xmp
		use Math

		x = Math.rand (40)        # x = Math.rand (0 .. 40)

		r = Range (1, 100)
		x = Math.rand (r)

		r = Range (0, 10000, 10)
		x = Math.rand (r)         # x is a multiply of 10
*/
static int ChangeRandSeed = 10000;

static qu_int Math_rand (int argc, qu_any *argv)
{
	long r;

	if (++ChangeRandSeed > 10000)
	{
		ChangeRandSeed = 0;
		srand (time (NULL));
	}

	if (argc == 0)
	{
		r = (long) ((double) (RAND_MAX + 1.0) * rand() / (double) RAND_MAX);
	}
	else if (argc == 1 && qu_range_is (argv[0]))
	{
		qu_range g = argv[0];

		r = (long) ((double) g->count * rand() / (double) RAND_MAX);
		r = g->first + r * g->step;
	}
	else
	{
		qu_any   x;
		long  l = 0, h = LONG_MAX;

		x = argv[0];
		switch (qu_classid (x))
		{
			case QU_CLASS_ID_Int:      l = qu_int_tol      ((qu_int     ) x); break;
			case QU_CLASS_ID_Long:     l = qu_long_tol     ((qu_long    ) x); break;
			case QU_CLASS_ID_Rational: l = qu_rational_tol ((qu_rational) x); break;
			case QU_CLASS_ID_Float:    l = qu_float_tol    ((qu_float   ) x); break;
			case QU_CLASS_ID_Big:      l = qu_big_tol      ((qu_big     ) x); break;
			case QU_CLASS_ID_Complex:  l = qu_complex_tol  ((qu_complex ) x); break;
		}

		if (argc > 1)
		{
			x = argv[1];

			switch (qu_classid (x))
			{
				case QU_CLASS_ID_Int:      h = qu_int_tol      ((qu_int     ) x); break;
				case QU_CLASS_ID_Long:     h = qu_long_tol     ((qu_long    ) x); break;
				case QU_CLASS_ID_Rational: h = qu_rational_tol ((qu_rational) x); break;
				case QU_CLASS_ID_Float:    h = qu_float_tol    ((qu_float   ) x); break;
				case QU_CLASS_ID_Big:      h = qu_big_tol      ((qu_big     ) x); break;
				case QU_CLASS_ID_Complex:  h = qu_complex_tol  ((qu_complex ) x); break;
			}
		}
		else
		{
			h = l;
			l = 0;
		}
		r = l + (long) ((double) (h - l) * rand() / (double) RAND_MAX);
	}
	return qu_int_new (r);
}




/*----------------------------------------------------------------------------
	qu_complex
----------------------------------------------------------------------------*/

static qu_real Math_arg_complex (qu_complex x)
{
	return (qu_real) Math_atan22 (x->real, x->imag);
}



/*----------------------------------------------------------------------------
	__pow
----------------------------------------------------------------------------*/

/*
	x ** y = z

	where:

	z.real = r1 * cos (a1)
	z.imag = r1 * sin (a1)

	r0 = hypot (x.real, x.imag)
	a0 = atan2 (x.imag, x.real)
	r1 = exp (y.real * log (r0) - y.imag * a0)
	a1 = a0 * y.real + y.imag * log (r0);
*/
static qu_number Math_pow_complexr (qu_real ar, qu_real ai, qu_real br, qu_real bi)
{
	qu_any l0, t1, t2, cr, ci, r0, a0, r1, a1;

	r0 = Math_hypot2 (ar, ai); QU_ENSURE (r0);
	a0 = Math_atan22 (ai, ar); QU_ENSURE (a0);
	l0 = Math_log1   (r0);     QU_ENSURE (l0);

	t1 = qu_sys_mul (br, l0);   QU_ENSURE (t1);
	t2 = qu_sys_mul (bi, a0);   QU_ENSURE (t2);

	r1 = qu_sys_minus (t1, t2); QU_ENSURE (r1);
	r1 = Math_exp1 (r1);        QU_ENSURE (r1);

	t1 = qu_sys_mul (a0, br);   QU_ENSURE (t1);
	t2 = qu_sys_mul (bi, l0);   QU_ENSURE (t2);

	a1 = qu_sys_plus (t1, t2);  QU_ENSURE (a1);

	cr = Math_cos1 (a1);        QU_ENSURE (cr);
	ci = Math_sin1 (a1);        QU_ENSURE (ci);

	cr = qu_sys_mul (r1, cr);   QU_ENSURE (cr);
	ci = qu_sys_mul (r1, ci);   QU_ENSURE (ci);

	QU_REQUIRE (QuEMath, qu_real_is (cr));
	QU_REQUIRE (QuEMath, qu_real_is (ci));

	return __qu_complex__ (cr, ci);
}



static qu_number Math_pow_big2 (qu_big x, qu_big y)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	if (sx == 0)
	{
		if (sy == 0)
			return (qu_number) QuInt1;

		else if (sy < 0)
			return (qu_number) __qu_inf__ (1);

		else
			return (qu_number) QuInt0;
	}
	else if (sy == 0)
	{
		return (qu_number) QuInt1;
	}
	else if (qu_big_isone (y))
	{
		if (sy < 0)
			return (qu_number) qu_big_div_big (QuBig1, x);
		else
			return (qu_number) x;
	}
	else if (sx < 0 && !qu_big_iswhole (y))
	{
		return Math_pow_complexr ((qu_real) x, (qu_real) QuInt0, (qu_real) y, (qu_real) QuInt0);
	}
	else
	{
		qu_real r;
		int  sign = qu_big_isodd (y) ? -1 : 1;

		if (sx < 0)
			x = qu_big_neg (x);

		r = (qu_real) Math_log1       (x);    QU_ENSURE (r); assert (qu_real_is (r));
		r = (qu_real) qu_big_mul_real (y, r); QU_ENSURE (r); assert (qu_real_is (r));
		r = (qu_real) Math_exp1       (r);    QU_ENSURE (r); assert (qu_real_is (r));

		if (qu_inf_is (r))
		{
			if (sy < 0)
				return (qu_number) QuInt0;
			else
				return (qu_number) __qu_inf__ (sign);
		}

		if (qu_real_sign (r) != sign)
			r = qu_real_neg (r);
		return (qu_number) r;
	}
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Math_init (qu_module m)
{

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		determine the maximum integer that can be square-root-ed
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef HAVE_LONG_LONG
	MathSqrtMaxInteger = (double) ((long long) sqrt ((double) LONG_LONG_MAX));
#else
	MathSqrtMaxInteger = (double) ((long) sqrt ((double) LONG_MAX));
#endif
	MathSqrtMaxInteger = MathSqrtMaxInteger * MathSqrtMaxInteger;


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		PI PHI E
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	MathPiFloat = qu_float_new_static (MATH_PI_FLOAT);
	MathPiBig = qu_big_new_static (
	  "3.14159265358979323846264338327950288419716939937510"
		"58209749445923078164062862089986280348253421170679"
		"82148086513282306647093844609550582231725359408128"
		"48111745028410270193852110555964462294895493038196"
		"44288109756659334461284756482337867831652712019091"
		"45648566923460348610454326648213393607260249141273"
		"72458700660631558817488152092096282925409171536436"
		"78925903600113305305488204665213841469519415116094"
		"33057270365759591953092186117381932611793105118548"
		"07446237996274956735188575272489122793818301194912"
		"98336733624406566430860213949463952247371907021798"
		"60943702770539217176293176752384674818467669405132"
		"00056812714526356082778577134275778960917363717872"
		"14684409012249534301465495853710507922796892589235"
		"42019956112129021960864034418159813629774771309960"
		"51870721134999999837297804995105973173281609631859"
		"50244594553469083026425223082533446850352619311881"
		"71010003137838752886587533208381420617177669147303"
		"59825349042875546873115956286388235378759375195778"
		"185778053217122680661300192787661119590921642");

	MathPhiFloat = qu_float_new_static (MATH_PHI_FLOAT);
	MathPhiBig = qu_big_new_static (
	  "1.61803398874989484820458683436563811772030917980576"
		"28621354486227052604628189024497072072041893911374"
		"84754088075386891752126633862223536931793180060766"
		"72635443338908659593958290563832266131992829026788"
		"06752087668925017116962070322210432162695486262963"
		"13614438149758701220340805887954454749246185695364"
		"86444924104432077134494704956584678850987433944221"
		"25448770664780915884607499887124007652170575179788"
		"34166256249407589069704000281210427621771117778053"
		"15317141011704666599146697987317613560067087480710"
		"13179523689427521948435305678300228785699782977834"
		"78458782289110976250030269615617002504643382437764"
		"86102838312683303724292675263116533924731671112115"
		"88186385133162038400522216579128667529465490681131"
		"71599343235973494985090409476213222981017261070596"
		"11645629909816290555208524790352406020172799747175"
		"34277759277862561943208275051312181562855122248093"
		"94712341451702237358057727861600868838295230459264"
		"78780178899219902707769038953219681986151437803149"
		"97411069260886742962267575605231727775203536");

	MathEFloat = qu_float_new_static (MATH_E_FLOAT);
	MathEBig = qu_big_new_static (
	  "2.71828182845904523536028747135266249775724709369995"
		"95749669676277240766303535475945713821785251664274"
		"27466391932003059921817413596629043572900334295260"
		"59563073813232862794349076323382988075319525101901"
		"15738341879307021540891499348841675092447614606680"
		"82264800168477411853742345442437107539077744992069"
		"55170276183860626133138458300075204493382656029760"
		"67371132007093287091274437470472306969772093101416"
		"92836819025515108657463772111252389784425056953696"
		"77078544996996794686445490598793163688923009879312"
		"77361782154249992295763514822082698951936680331825"
		"28869398496465105820939239829488793320362509443117"
		"30123819706841614039701983767932068328237646480429"
		"53118023287825098194558153017567173613320698112509"
		"96181881593041690351598888519345807273866738589422"
		"87922849989208680582574927961048419844436346324496"
		"84875602336248270419786232090021609902353043699418"
		"49146314093431738143640546253152096183690888707016"
		"76839642437814059271456354906130310720851038375051"
		"0115747704171898610687396965521267154688957035");


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		functions
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	qu_module_add_function (m, "PI"   , Math_PI   , -1);
	qu_module_add_function (m, "PHI"  , Math_PHI  , -1);
	qu_module_add_function (m, "E"    , Math_E    , -1);

	qu_module_add_function (m, "abs"  , Math_abs  , -1);
	qu_module_add_function (m, "sqrt" , Math_sqrt , -1);
	qu_module_add_function (m, "hypot", Math_hypot, -1);
	qu_module_add_function (m, "atan2", Math_atan2, -1);

	qu_module_add_function (m, "exp"  , Math_exp  , -1);
	qu_module_add_function (m, "log"  , Math_log  , -1);

	qu_module_add_function (m, "sin"  , Math_sin  , -1);
	qu_module_add_function (m, "cos"  , Math_cos  , -1);
	qu_module_add_function (m, "tan"  , Math_tan  , -1);

	qu_module_add_function (m, "sinh" , Math_sinh , -1);
	qu_module_add_function (m, "cosh" , Math_cosh , -1);
	qu_module_add_function (m, "tanh" , Math_tanh , -1);

	qu_module_add_function (m, "asin" , Math_asin , -1);
	qu_module_add_function (m, "acos" , Math_acos , -1);
	qu_module_add_function (m, "atan" , Math_atan , -1);

	qu_module_add_function (m, "asinh", Math_asinh, -1);
	qu_module_add_function (m, "acosh", Math_acosh, -1);
	qu_module_add_function (m, "atanh", Math_atanh, -1);

	qu_module_add_function (m, "deg"  , Math_deg  , -1);
	qu_module_add_function (m, "rad"  , Math_rad  , -1);

	qu_module_add_function (m, "rand" , Math_rand , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		builtins
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	__QuBigPowBigMath         = Math_pow_big2;
	__QuComplexPowComplexMath = Math_pow_complexr;
	__QuComplexAbsMath        = Math_abs_complex;
	__QuComplexArgMath        = Math_arg_complex;

	return 1;
}
