/* Net.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Net
	This module provides a low-level interface to the system network library,
	including ${Socket}.

	@xmp
		use Net

		host = Net.gethostbyname ('localhost.localdomain')
		addr = host->addr_list[0]

		host = Net.gethostbyaddr (host->addrtype, addr)
		println (host->name)
		println (host->aliases)

		saddr = Net.sockaddr (Net.AF_INET, addr, 80)

	@out
		localhost.localdomain
		['foo']
		'\x02\x00\x00P\x7f\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00'

	@p
	Most functions are marked as k{__unsafe} so ESecurity is thrown if called
	by untrusted modules.

	The Net module supports IPv6 although the implementation may not be complete
	at this time. See #{ipv6}.
*/

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#ifdef HAVE_SYS_LIMITS_H
#include <sys/syslimits.h>
#endif

#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#if EAGAIN != EWOULDBLOCK
#define FIXEAGAIN \
QU_BEGIN \
	if (errno == EWOULDBLOCK) \
		errno = EAGAIN; \
QU_END
#else
#define FIXEAGAIN QU_BEGIN QU_END
#endif

#ifdef HAVE_NETPACKET_PACKET_H
#include <net/if.h>
#include <netpacket/packet.h>
#endif

#ifdef HAVE_BLUETOOTH_BLUETOOTH_H
#include <bluetooth/bluetooth.h>
#include <bluetooth/l2cap.h>
#include <bluetooth/rfcomm.h>
#include <bluetooth/sco.h>
#endif


/* runtime-able setting */
static int IPv6Enabled = 0;

/* what time is now? */
#define NOW qu_stream_now ()

/* union of supported socket addresses */
typedef union
{
	struct sockaddr         gen;
	struct sockaddr_in      in;
#ifdef AF_INET6
	struct sockaddr_in6     in6;
	struct sockaddr_storage storage;
#endif
#ifdef AF_UNIX
	struct sockaddr_un      un;
#endif
#ifdef HAVE_NETPACKET_PACKET_H
	struct sockaddr_ll      ll;
#endif
#if 0
#ifdef AF_X25
	struct sockaddr_x25     x25;
#endif
#endif
#ifdef HAVE_BLUETOOTH_BLUETOOTH_H
	struct sockaddr_l2      l2;
	struct sockaddr_rc      rc;
	struct sockaddr_sco     sco;
#endif
} sockaddr_s;




/*----------------------------------------------------------------------------
	enum, const, struct
----------------------------------------------------------------------------*/

#include "Netenum.c"
/*) include Netenum.c */


#include "Netconst.c"
/*) include Netconst.c */


#include "Netstruct.c"
/*) include Netstruct.c */



/*----------------------------------------------------------------------------
	String with the special flag QU_STRING_FLAG_SOCKADDR is a socket address
----------------------------------------------------------------------------*/

#define ENSURE_SOCKADDR(s) \
	QU_ARG_CHECKZ (qu_string_is (s) && qu_flag_test (s, QU_STRING_FLAG_SOCKADDR), "not a socket address")

#define SET_SOCKADDR_FLAG(s) \
	qu_flag_set (s, QU_STRING_FLAG_SOCKADDR)


/*----------------------------------------------------------------------------

	lockers

	Provide a lock for each function specifically so threads in __qu_blocking_begin
	state are not contending unessecarily.

----------------------------------------------------------------------------*/

/* for getaddrinfo */
static qu_thread_mutex_t AddrinfoLocker = QU_THREAD_MUTEX_INITIALIZER;
#define LOCK_ADDRINFO   qu_thread_mutex_lock (&AddrinfoLocker)
#define UNLOCK_ADDRINFO qu_thread_mutex_unlock (&AddrinfoLocker)

/* for those that access struct hostent */
static qu_thread_mutex_t HostentLocker = QU_THREAD_MUTEX_INITIALIZER;
#define LOCK_HOSTENT   qu_thread_mutex_lock (&HostentLocker)
#define UNLOCK_HOSTENT qu_thread_mutex_unlock (&HostentLocker)

/* for those that access struct protoent */
static qu_thread_mutex_t ProtoentLocker = QU_THREAD_MUTEX_INITIALIZER;
#define LOCK_PROTOENT   qu_thread_mutex_lock (&ProtoentLocker)
#define UNLOCK_PROTOENT qu_thread_mutex_unlock (&ProtoentLocker)

/* for those that access struct servent */
static qu_thread_mutex_t ServentLocker = QU_THREAD_MUTEX_INITIALIZER;
#define LOCK_SERVENT   qu_thread_mutex_lock (&ServentLocker)
#define UNLOCK_SERVENT qu_thread_mutex_unlock (&ServentLocker)



/*----------------------------------------------------------------------------
	general
----------------------------------------------------------------------------*/

/* helper for gethostbyname and gethostbyaddr */
static qu_record _net_record_from_hostent (struct hostent *he)
{
	int        l = he->h_length;
	char     **s;
	qu_array   a = qu_array_new (2);
	qu_array   d = qu_array_new (2);

	qu_any arg[5] =
	{
		qu_string_newz (he->h_name),
		a,
		qu_int_newi (he->h_addrtype),
		qu_int_newi (he->h_length),
		d
	};

	s = he->h_aliases;
	while (*s)
	{
		qu_array_push_quick (a, qu_string_newz (*s));
		s++;
	}
	qu_array_freeze (a);

	s = he->h_addr_list;
	while (*s)
	{
		qu_string addr = qu_string_new (*s, l);
		qu_array_push_quick (d, addr);
		s++;
	}
	qu_array_freeze (d);
	return qu_record_new (HostentStruct, 5, arg);
}



/*) sub gethostbyname (name: String) -> Record
	Get host by specifying the host name or IP string.

	The returned value is a Record of #{Hostent}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		use Net
		host = Net.gethostbyname ('127.0.0.1')
		addr = host->addr_list[0]
		println (addr.quote)
		println (Net.inet_ntop (host->addrtype, addr))

	@out
		'\x7f\x00\x00\x01'
		127.0.0.1

	@see
		#{gethostbyaddr}
*/
static qu_record Net_gethostbyname (qu_string name)
{
	struct hostent *he;
	qu_record r;

	__qu_unsafe__;

	QU_ARG_TYPE (name, qu_string);

	__qu_blocking_begin;
	LOCK_HOSTENT;
	he = gethostbyname (name->str);
	__qu_blocking_end;
	if (he == 0)
	{
		UNLOCK_HOSTENT;
		QU_THROWERRNO;
	}
	r = _net_record_from_hostent (he);
	UNLOCK_HOSTENT;
	return r;
}



/*) sub gethostbyaddr (type: Integer, addr: String) -> Record
	Get host by specifying the address v{addr} and the address v{type}.

	The returned value is a Record of #{Hostent}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		use Net
		host = Net.gethostbyname ('127.0.0.1')
		addr = host->addr_list[0]

		host = Net.gethostbyaddr (AF_INET, addr)
		println (host->name)
		println (host->aliases)

	@out
		localhost.localdomain
		['foo']

	@see
		#{gethostbyname}
*/
static qu_record Net_gethostbyaddr (qu_integer xtype, qu_string name)
{
	struct hostent *he;
	int       type;
	qu_record r;

	__qu_unsafe__;

	QU_ARG_int  (xtype, type);
	QU_ARG_TYPE (name, qu_string);

	__qu_blocking_begin;
	LOCK_HOSTENT;
	he = gethostbyaddr (name->str, name->len, type);
	__qu_blocking_end;

	if (he == 0)
	{
		UNLOCK_HOSTENT;
		QU_THROWERRNO;
	}
	r = _net_record_from_hostent (he);
	UNLOCK_HOSTENT;
	return r;
}



/*) sub gethosname -> String
	Get the host name of the current process.

	@exc
	.ESecurity
	.ESystem

	@xmp
		use Net
		println (Net.gethostname)

	@out
		localhost.localdomain

	@see
	C{man 2 gethosname)
	#{sethostname}
*/
static qu_string Net_gethostname ()
{
	char name [MAXHOSTNAMELEN + 1];

	__qu_unsafe__;

	if (gethostname (name, MAXHOSTNAMELEN) == 0)
	{
		name [MAXHOSTNAMELEN] = '\0';
		return qu_string_newz (name);
	}
	QU_THROWERRNO;
}



/*) sub getnameinfo (saddr: String, flags: Integer) -> Array
	Get information from the socket address v{saddr}.

	The returned value is an Array of two items: [host, service] which
	values depend on v{flags}. See #{NI} for details.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		host = Net.gethostbyname ('localhost.localdomain')
		addr = host->addr_list[0]
		sa = Net.sockaddr (host->addrtype, addr, 80)
		println (Net.getnameinfo (sa, Net.NI_NUMERICHOST))

	@out
		['127.0.0.1', 'http']

	@see
		C{man 3 getnameinfo}
		#{gethostbyname}
		#{gethostbyaddr}
*/
static qu_array Net_getnameinfo (qu_string saddr, qu_integer f)
{
	int  flags;
	char host [NI_MAXHOST], serv [NI_MAXSERV];

	__qu_unsafe__;

	ENSURE_SOCKADDR (saddr);
	QU_ARG_int (f, flags);

	if (getnameinfo ((struct sockaddr *) saddr->str, saddr->len,
			host, sizeof (host), serv, sizeof (serv), flags) != 0)
	{
		switch (errno)
		{
			case EAI_BADFLAGS : QU_ARG_THROWZ ("invalid flags");
			case EAI_FAMILY   : QU_ARG_THROWZ ("bad address");
			case EAI_AGAIN    : QU_THROW (QuETimeout);
			case EAI_MEMORY   : QU_THROW (QuEMemory);
			case EAI_NONAME   : QU_THROW (QuENone);
			case EAI_FAIL     : QU_THROW (QuESystem);
			/* case EAI_SYSTEM: */
		}
		QU_THROWERRNO;
	}
	else
	{
		return qu_array_new_ex (2, qu_string_newz (host), qu_string_newz (serv));
	}
}



/*) sub getip (saddr: String, timeout = 3.0) -> String
	Get the IP address from the socket address v{saddr}.

	This is just a convenient wrapper around #{getnameinfo}.

	If the address can not be determined within the specified v{timeout} then
	an empty string is returned, thus no exception is thrown except for
	invalid arguments.

	v{timeout} is in seconds.

	@exc
	.ESecurity
	.EArgValue
*/
static qu_string Net_getip (int argc, qu_any *argv)
{
	qu_string saddr;
	double    to = 3;
	char      host [NI_MAXHOST];

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	saddr = argv[0];

	ENSURE_SOCKADDR (saddr);

	if (argc > 1)
		QU_ARG_vdouble (1, to);

	to += NOW;

	while (getnameinfo ((struct sockaddr *) saddr->str, saddr->len,
				host, sizeof (host), NULL, 0, NI_NUMERICHOST) != 0)
	{
		/* this should never happen but nothing is said about it in the manual */
		if (errno == EAI_AGAIN)
		{
			if (NOW < to)
			{
				qu_thread_yield ();
				continue;
			}
		}
		return QuStringEMPTY;
	}
	return qu_string_newz (host);
}



/* helper for getprotobyname / getprotobynumber / getprotodict */
static qu_record _net_record_from_protoent (struct protoent *e)
{
	qu_any arg[3] =
	{
		qu_string_newz (e->p_name),
		qu_array_new (2),
		qu_int_new (e->p_proto)
	};

	char     *s, **p = e->p_aliases;
	qu_array  a = arg[2];

	while ((s = *p++) != NULL)
		qu_array_push_quick (a, qu_string_newz (s));

	return qu_record_new (ProtoentStruct, 3, arg);
}



/*) sub getprotobyname (name: String) -> Record
	Get protocol information by specifying the protocol name.

	The returned value is a #{Protoent} Record.

	@exc
	.EArgValue
	.ESecurity
	.ENone      protocol not found
	.ESystem

	@xmp
		use Net
		p = Net.getprotobyname ('TCP')
		println (p)
		println (p->proto)
		println (p->name)
		println

	@out
		['tcp', ['TCP'], 6]
		6
		tcp

	@see
		C{man 5 protocols}
		#{getprotobynumber}
		#{getprotodict}
		#{IPPROTO}
*/
static qu_record Net_getprotobyname (qu_string name)
{
	struct protoent *e;
	qu_record r;

	__qu_unsafe__;

	QU_ARG_TYPE (name, qu_string);

	__qu_blocking_begin;
	LOCK_PROTOENT;
	e = getprotobyname (name->str);
	__qu_blocking_end;

	if (e == NULL)
	{
		UNLOCK_PROTOENT;
		QU_THROWERRNO;
	}

	r = _net_record_from_protoent (e);
	UNLOCK_PROTOENT;
	return r;
}



/*) sub getprotobynumber (num: Integer) -> Record
	Get protocol information by specifying the protocol number.

	The returned value is a #{Protoent} Record.

	@exc
	.EArgValue
	.ESecurity
	.ESystem

	@xmp
		use Net
		p = Net.getprotobynumber (6)
		println (p)
		println (p->proto)
		println (p->name)
		println

	@out
		['tcp', ['TCP'], 6]
		6
		tcp

	@see
		#{getprotobyname}
		#{getprotodict}
		#{IPPROTO}
*/
static qu_record Net_getprotobynumber (qu_integer num)
{
	struct protoent *e;
	qu_record r;
	int       n;

	__qu_unsafe__;

	QU_ARG_int (num, n);

	__qu_blocking_begin;
	LOCK_PROTOENT;
	e = getprotobynumber (n);
	__qu_blocking_end;

	if (e == NULL)
	{
		UNLOCK_PROTOENT;
		QU_THROWERRNO;
	}

	r = _net_record_from_protoent (e);
	UNLOCK_PROTOENT;
	return r;
}




/*) sub getprotodict (keys = 0) -> Dict
	Get a Dict containing records of all protocols.

	If v{keys} is 0 then the keys are the protocol names, otherwise
	the keys are the protocol numbers.

	Note that since protocol names and numbers should be unique then
	the Dict values are always #{Protoent} records, thus unlike #{getservdict}
	which may be Arrays.

	@exc
	.ESecurity
	.ESystem

	@xmp
		use Net
		d = Net.getprotodict
		println (d.__len)
		n = 0
		for k, v in d
			break if ++n > 3
			println (k ' => ' v)
		;;

	@out
		129
		l2tp => ['l2tp', ['L2TP'], 115]
		ipv6-nonxt => ['ipv6-nonxt', ['IPv6-NoNxt'], 59]
		udp => ['udp', ['UDP'], 17]

	@see
		#{getprotobyname}
		#{getprotobynumber}
		#{IPPROTO}
*/
static qu_dict Net_getprotodict (int argc, qu_any *argv)
{
	struct protoent *e;
	qu_dict   d;
	int       i = 0;
	qu_record r;

	__qu_unsafe__;

	if (argc)
		QU_ARG_vint (0, i);

	__qu_blocking_begin;
	LOCK_PROTOENT;
	__qu_blocking_end;

	d = qu_dict_new (64);

	setprotoent (1);
	while ((e = getprotoent ()) != NULL)
	{
		r = _net_record_from_protoent (e);
		qu_dict_setitem (d, i ? r->objs[2] : r->objs[0], r);
	}
	endprotoent ();
	UNLOCK_PROTOENT;
	return d;
}



/* helper for getservbyname / getservbyport / getservdict */
static qu_record _net_record_from_servent (struct servent *se)
{
	char     **s;
	qu_array   a;

	qu_any arg[4] =
	{
		qu_string_newz (se->s_name),
		qu_array_new (2),
		qu_int_newi ((int) htons (se->s_port)),
		qu_string_newz (se->s_proto)
	};

	a = arg [2];

	s = se->s_aliases;
	while (*s)
	{
		qu_array_push_quick (a, qu_string_newz (*s));
		s++;
	}
	qu_array_freeze (a);

	return qu_record_new (ServentStruct, 4, arg);
}



/*) sub getservbyname (name: String, proto = '') -> Record
	Get service information by specifying the service and protocol name.

	If v{proto} is an empty string then any protocol is matched.

	The returned value is a #{Servent} Record.

	@exc
	.EArgValue
	.ESecurity
	.ESystem

	@xmp
		use Net
		p = Net.getservbyname ('www')
		println (p)
		println (p->name)
		println (p->proto)
		println (p->port)
		println

	@out
		['http', ['www', 'www-http'], 80, 'tcp']
		http
		tcp
		80

	@see
		C{man 3 getservbyname}
		#{getservebyport}
		#{IPPORT}
*/
static qu_record Net_getservbyname (int argc, qu_any *argv)
{
	struct servent *e;
	qu_string name;
	qu_string proto = NULL;
	qu_record r;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, name, qu_string);

	if (argc > 1)
	{
		QU_ARG_VTYPE (1, proto, qu_string);
		if (proto->len == 0)
			proto = NULL;
	}

	__qu_unsafe__;

	__qu_blocking_begin;
	LOCK_SERVENT;
	e = getservbyname (name->str, proto ? proto->str : NULL);
	__qu_blocking_end;
	if (e == NULL)
	{
		UNLOCK_SERVENT;
		QU_THROWERRNO;
	}
	r = _net_record_from_servent (e);
	UNLOCK_SERVENT;
	return r;
}



/*) sub getservbyport (port: Integer, proto = '') -> Record
	Get service information by specifying the port number.

	The returned value is a #{Servent} Record.

	If v{proto} is an empty string than any protocol is matched.

	@exc
	.EArgValue
	.ESecurity
	.ENone      service not found
	.ESystem

	@xmp
		use Net
		p = Net.getservbyport (80)
		println (p)
		println (p->name)
		println (p->proto)
		println (p->port)
		println

	@out
		['http', ['www', 'www-http'], 80, 'tcp']
		http
		tcp
		80

	@see
		#{getservebyname}
		#{IPPORT}
*/
static qu_record Net_getservbyport (int argc, qu_any *argv)
{
	struct servent *e;
	int       port;
	qu_string proto = 0;
	qu_record r;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vint (0, port);

	if (argc > 1)
	{
		QU_ARG_VTYPE (1, proto, qu_string);
		if (proto->len == 0)
			proto = NULL;
	}

	__qu_blocking_begin;
	LOCK_SERVENT;
	e = getservbyport (ntohs (port), proto ? proto->str : NULL);
	__qu_blocking_end;
	if (e == NULL)
	{
		UNLOCK_SERVENT;
		QU_THROWERRNO;
	}
	r = _net_record_from_servent (e);
	UNLOCK_SERVENT;
	return r;
}



/*) sub getservdict (keys: Integer) -> Dict
	Get a Dict containing records of all services.

	If v{keys} is 0 then the keys are the service names.
	If v{keys} is positive, the keys are the service ports.
	Otherwise the keys are the service protocol names.

	@exc
	.ESecurity
	.ESystem

	@warn
	Note that for duplicate keys, the values will be Array objects.

	@xmp
		use Net
		d = Net.getservdict (1)
		println (d.__len)
		for r in d[80]
			println (r)
		;;

	@out
		262
		['http', ['www', 'www-http'], 80, 'tcp']
		['http', ['www', 'www-http'], 80, 'udp']

	@see
		#{getservbyname}
		#{getservbynumber}
		#{IPPROTO}
*/
static qu_dict Net_getservdict (int argc, qu_any *argv)
{
	struct servent *e;
	qu_dict   d;
	int       i = 0;
	qu_record r;

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc)
		QU_ARG_vint (0, i);

	__qu_blocking_begin;
	LOCK_PROTOENT;
	__qu_blocking_end;

	d = qu_dict_new (64);

	setservent (1);
	while ((e = getservent ()) != NULL)
	{
		qu_any k, x;

		r = _net_record_from_servent (e);
		k = r->objs[(i == 0) ? 0 : ((i > 0) ? 2 : 3)];
		x = qu_dict_getitem (d, k);
		if (x != NULL)
		{
			if (!qu_array_is (x))
			{
				qu_array a = qu_array_new (2);
				qu_array_push_quick (a, x);
				qu_dict_setitem (d, k, a);
				x = a;
			}
			qu_array_push_quick (x, r);
		}
		else
		{
			qu_dict_setitem (d, k, r);
		}
	}
	endservent ();
	UNLOCK_PROTOENT;
	return d;
}



/*) sub htonl (x: Integer) -> Integer
	Convert from host byte order to network byte order.

	v{x} must be a positive 32 bits Integer.

	@exc
	.EArgValue
*/
static qu_integer Net_htonl (qu_integer x)
{
#if SIZEOF_LONG == 4
	unsigned long v;

	QU_ARG_ulong (x, v);
	v = htonl (v);
	return qu_integer_newul (v);
#else
	unsigned int v;

	QU_ARG_uint (x, v);
	v = htonl (v);
	return qu_integer_newui (v);
#endif
}



/*) sub htons (x: Integer) -> Integer
	Convert from host byte order to network byte order.

	v{x} must be a positive 16 bits Integer.

	@exc
	.EArgValue
*/
static qu_integer Net_htons (qu_integer x)
{
	unsigned short v;

	QU_ARG_ushort (x, v);
	v = htons (v);
	return qu_integer_newuh (v);
}



/*) sub inet_ntop (af: Integer, addr: String) -> String
	Convert the network address v{addr} in the internet address v{af} family
	into an IP string.

	v{af} must be AF_INET or AF_INET6.

	@exc
	.EArgValue
	.ESystem

	@xmp
		use Net
		host = Net.gethostbyname ('localhost.localdomain')
		addr = host->addr_list[0]
		println (Net.inet_ntop (host->addrtype, addr))

	@out
		127.0.0.1
*/
static qu_string Net_inet_ntop (qu_integer xaf, qu_string addr)
{
	int         af, l;
	char        ip [MAXIPLEN + 1];
	const char *r;

	QU_ARG_int  (xaf, af);
	QU_ARG_TYPE (addr, qu_string);
	switch (af)
	{
#ifdef AF_INET6
		case AF_INET6: l = sizeof (struct in6_addr); break;
#endif
		case AF_INET:  l = sizeof (struct in_addr);  break;

		default:
			QU_ARG_THROW;
	}
	QU_ARG_CHECK (addr->len == l);
	r = inet_ntop (af, addr->str, ip, sizeof (ip));
	if (r == 0)
		QU_THROWERRNO;
	return qu_string_newz (r);
}



/*) sub inet_pton (af: Integer, ip: String) -> String
	Convert the IP string v{ip} into an internet address in the v{af}
	address family.

	v{af} must be AF_INET or AF_INET6.

	@exc
	.EArgValue
	.ESystem

	@xmp
		use Net
		addr = Net.inet_pton (Net.AF_INET, '127.0.0.1')
		println (addr.quote)

	@out
		'\x7f\x00\x00\x01'
*/
static qu_string Net_inet_pton (qu_integer xaf, qu_string ip)
{
	int af, r, l = sizeof (struct in_addr);
#ifdef AF_INET6
	char addr [sizeof (struct in6_addr)];
#else
	char addr [sizeof (struct in_addr)];
#endif

	QU_ARG_int  (xaf, af);
	QU_ARG_TYPE (ip, qu_string);

#ifdef AF_INET6
	if (af == AF_INET6)
	{
		QU_ARG_CHECKZ (IPv6Enabled, "IPv6 is not enabled");
		l = sizeof (struct in6_addr);
	}
#endif

	r = inet_pton (af, ip->str, addr);
	if (r < 0)
		QU_THROWERRNO;
	QU_ARG_CHECKZ (r > 0, "bad IP string");
	return qu_string_new (addr, l);
}



/*) sub ipv6 (*arg) -> Boolean
	@def
	sub ipv6 -> Boolean
	sub ipv6 (setting: Boolean) -> Boolean
	@p
	Set or get whether IPv6 is enabled.

	If called with no argument, return the current setting.
	If called with one argument, set and return the old setting.

	If your system does not support IPv6 then this method always return
	false.

	@exc
	.EArgValue
	.ESecurity
*/
static qu_boolean Net_ipv6 (int argc, qu_any *argv)
{
#ifdef AF_INET6
	QU_ARG_VMAX (1);

	if (argc == 0)
		return IPv6Enabled ? QuTrue : QuFalse;
	else
	{
		int v, old = IPv6Enabled;

		__qu_unsafe__;

		QU_ARG_vbool (0, v);
		IPv6Enabled = v;
		return old ? QuTrue : QuFalse;
	}
#else
	return QuFalse;
#endif
}



/*) sub isip (s: String) -> Boolean
	Return true if v{s} is a valid IP string, false otherwise.

	@noexc

	@xmp
		use Net
		println (Net.isip ('foo.bar'))
		println (Net.isip ('127.0.0.1'))

	@out
		false
		true
*/
static qu_boolean Net_isip (qu_string s)
{
	if (!qu_string_is (s) || s->len < 7)
		return QuFalse;

#ifdef AF_INET6
	if (IPv6Enabled)
	{
		struct in6_addr a;
		if (inet_pton (AF_INET6, s->str, &a) > 0)
			return QuTrue;
	}
	else
#endif
	{
		struct in_addr a;
		if (inet_pton (AF_INET, s->str, &a) > 0)
			return QuTrue;
	}
	return QuFalse;
}



/*) sub isname (s: String) -> Boolean
	Return true if v{s} is a valid host name, false otherwise.

	@warn
	The result is slightly different with RFC-2396, i.e. we don't allow
	digits as the first letter, a name that ends with a dot, an empty string,
	and a string longer than #{MAXHOSTNAMELEN}.

	@xmp
		use Net

		println (Net.isname ('foo'))
		println (Net.isname ('foo.bar'))
		println (Net.isname ('foo.bar.baz'))

	@out
		true
		true
		true
*/
/*
	RFC-2396:
		hostname    = *( domainlabel "." ) toplabel [ "." ]
		domainlabel = alphanum | alphanum *( alphanum | "-" ) alphanum
		toplabel    = alpha | alpha *( alphanum | "-" ) alphanum

	We don't want  : 7foo.bar  foo.bar.
	These are okay : foo7.bar  foo.7bar
*/
static qu_boolean Net_isname (qu_string s)
{
	char *i;
	long  l;
	int   n;

	if (!qu_string_is (s) || s->len < 1 || s->len > MAXHOSTNAMELEN)
		return QuFalse;

	l = s->len;
	if (l < 1)
		return QuFalse;

	i = s->str;
	for (n = 0; n < l; n++)
	{
		if (n == 0)
		{
			if (!qu_ascii_isalpha (*i))
				return QuFalse;
		}
		else
		{
			if (!qu_ascii_isalnum (*i))
				return QuFalse;
		}
		i++;
		while (qu_ascii_isalnum (*i) || (*i == '-'))
			i++;
		if (i[-1] == '-')
			return QuFalse;
		if (*i == '\0')
			break;
		if (*i != '.')
			return QuFalse;
		i++;
	}
	return QuTrue;
}



/*) sub ntohl (x : Integer) -> Integer
	Convert from network byte order to host byte order.

	v{x} must be a positive 32 bits Integer.

	@exc
	.EArgValue
*/
static qu_integer Net_ntohl (qu_integer x)
{
#if SIZEOF_LONG == 4
	unsigned long v;

	QU_ARG_ulong (x, v);
	v = ntohl (v);
	return qu_integer_newul (v);
#else
	unsigned int v;

	QU_ARG_uint (x, v);
	v = ntohl (v);
	return qu_integer_newui (v);
#endif
}



/*) sub ntohs (x : Integer) -> Integer
	Convert from network byte order to host byte order.

	v{x} must be a positive 16 bits Integer.

	@exc
	.EArgValue
*/
static qu_integer Net_ntohs (qu_integer x)
{
	unsigned short v;

	QU_ARG_ushort (x, v);
	v = ntohs (v);
	return qu_integer_newuh (v);
}



/*) sub sethosname (name: String) -> Null
	Set the host name previously set by #{sethostname}.

	v{name} must be a valid #{isname}.

	@exc
	.EArgValue
	.ESecurity
	.ESystem

	@see
		C{man 2 sethosname)
		#{gethostname}
*/
static qu_null Net_sethostname (qu_string name)
{
	__qu_unsafe__;

	QU_ARG_CHECK (Net_isname (name) == QuTrue);
	if (sethostname (name->str, name->len) == 0)
		return QuNil;
	QU_THROWERRNO;
}



/*) sub sockaddr (type: Integer, addr, *arg) -> Any
	@def
	sub sockaddr (type: Integer, addr: String, *arg) -> String
	sub sockaddr (type: Integer, addr: Hostent, *arg) -> Array
	Build a socket address.

	@warn
	This is the only way to create a socket address.

	The returned string (or strings) are internally flagged as a socket addresses.
	Functions or methods in the Net module that requires a socket address check
	for this flag.

	@p
	v{type} is the address type of v{addr}, e.g. AF_INET.

	v{addr} is the network address or a #{Hostent} Record.
	You get this from, e,g, #{gethostbyname}.

	If it is a Hostent Record then an Array of socket addresses is returned
	based on the v{addr_list} entries.

	The v{arg} arguments depends on the address type.
	The following table shows the supported types and the required arguments.

	@table
	.| B{type}   | B{arg}
	.| AF_INET   | (Integer) port number
	.| AF_INET6  | (Integer) port number
	.| AF_UNIX   | none

	@p
	For AF_UNIX, v{addr} must be a valid path with length in 4 .. 108.
	It may also be a Hostent Record provided that the addresses in v{addr_list}
	are all valid paths.

	@xmp
		paths  = ['/foo/bar.baz', '/bar/baz.foo']
		addrs  = Hostent ("whatever", [], AF_UNIX, -1, paths)
		salist = Net.sockaddr (addrs)

	@exc
	.EArgValue
	.EImplement v{type} not supported
	.ESecurity
	.ESystem

	@xmp
		use Net

		host  = Net.gethostbyname ('www.foo.bar')
		addr  = host->addr_list[0]
		saddr = Net.sockaddr (Net.AF_INET, addr, 80)

		# or

		host   = Net.gethostbyname ('www.foo.bar')
		salist = Net.sockaddr (Net.AF_INET, host, 80)
*/
static qu_string _net_sockaddr1 (int type, qu_string addr, int argc, qu_any *argv)
{

#define ENSURE_VALID_PORT(p) \
	QU_ARG_CHECKZ (0 <= (p) && (p) <= 65535, "port number must be in 0..65535")

	qu_string saddr = 0;

	__qu_unsafe__;

	switch (type)
	{
		case AF_INET:
		{
			struct sockaddr_in sa;
			int port, len = sizeof (struct in_addr);

			QU_ARG_CHECK (addr->len == len);
			QU_ARG_VMIN (1);
			QU_ARG_vint (0, port);
			ENSURE_VALID_PORT (port);

			memset (&sa, 0, sizeof (sa));
			sa.sin_family = AF_INET;
			sa.sin_port = htons ((int) port);
			memcpy (&sa.sin_addr.s_addr, addr->str, len);
			saddr = qu_string_new ((char *) &sa, sizeof (sa));
			break;
		}

#ifdef AF_INET6
		case AF_INET6:
		{
			struct addrinfo hints, *r;
			struct sockaddr_in6 sa;
			int    err, port, len = sizeof (struct in6_addr);
			volatile qu_string ip;

			QU_ARG_CHECK (addr->len == len);
			QU_ARG_VMIN (1);
			QU_ARG_vint (0, port);
			ENSURE_VALID_PORT (port);

			ip = Net_inet_ntop ((qu_integer) qu_int_new (AF_INET6), addr);
			QU_ENSURE (ip);

		 	__qu_blocking_begin;
			memset (&sa, 0, sizeof (sa));
			memset (&hints, 0, sizeof(hints));
			hints.ai_family = AF_INET6;
			LOCK_ADDRINFO;
			err = getaddrinfo (ip->str, NULL, &hints, &r);
			__qu_blocking_end;
			UNLOCK_ADDRINFO;
			if (err)
				QU_THROWZ (QuESystem, gai_strerror (err));
			memcpy (&sa, r->ai_addr, (r->ai_addrlen < len) ? r->ai_addrlen : len);
			freeaddrinfo (r);
			sa.sin6_family = AF_INET6;
			sa.sin6_port = htons ((int) port);
			memcpy (&sa.sin6_addr.s6_addr, addr->str, len);
			saddr = qu_string_new ((char *) &sa, sizeof (sa));
			break;
		}
#endif

#ifdef AF_UNIX
		case AF_UNIX:
		{
			struct sockaddr_un sa;

			QU_ARG_CHECK (4 <= addr->len && addr->len <= 108 && qu_file_valid (addr, 0));
			memcpy (&sa.sun_path, addr->str, addr->len + 1);
			saddr = qu_string_new ((char *) &sa, sizeof (sa));
			break;
		}
#endif

		default:
			QU_THROW (QuEImplement);
	}

	SET_SOCKADDR_FLAG (saddr);
	return saddr;
}


static qu_any Net_sockaddr (int argc, qu_any *argv)
{
	int    type;
	qu_any addr;

	__qu_unsafe__;

	QU_ARG_VMIN (2);
	QU_ARG_vint (0, type);
	addr = argv[1];

	argc -= 2;
	argv += 2;

	if (qu_string_is (addr))
	{
		return _net_sockaddr1 (type, (qu_string) addr, argc, argv);
	}
	else
	{
		qu_array  a, r;
		qu_string sa;
		int       i;

		QU_ARG_CHECK (qu_record_of (addr, HostentStruct));
		a = ((qu_record) addr)->objs[4];
		QU_ARG_TYPE (a, qu_array);

		r = qu_array_new (a->len);
		for (i = 0; i < a->len; ++i)
		{
			addr = a->objs[i];
			QU_ARG_TYPE (addr, qu_string);
			sa = _net_sockaddr1 (type, (qu_string) addr, argc, argv);
			QU_ENSURE (sa);
			qu_array_push_quick (r, sa);
		}
		return r;
	}
}



/*----------------------------------------------------------------------------
	class Socket
----------------------------------------------------------------------------*/

/*) class Socket is Stream
	This class prodives an interface to sockets.

	Make sure you have read the documenation for the -{Stream} class.

	Most of the methods are inherited from Stream except those that are
	special to sockets.

	There are two kind of Socket objects: #{listener} and #{client}.

	In brief, a listener socket (or "server") listens for incoming connections
	then create a "client" socket to deal with the communication while a client
	socket connects to a listener socket to send requests and gets responses
	from an #{accept}ed connection by a listener socket.

	Here is a typical server program.

	@prog
		use Net

		server = Net.Socket.listener (Net.PF_INET, Net.SOCK_STREAM, Net.IPPROTO_IP)
		server.setsockopt (Net.SOL_SOCKET, Net.SO_REUSEADDR, true)

		iaddr = Net.inet_pton (Net.AF_INET, Net.INADDR_ANY)
		saddr = Net.sockaddr (Net.AF_INET, iaddr, 80)

		server.bind (saddr)
		server.listen

		loop
			server.timeout (3)
			try
				client = server.accept
			catch ETimeout
				... do something else ...
			catch
				__abort__ ("server error ", __err__)
			else
				Thread (@service, client)
			;;
		;;

		sub service
			println ("servicing " self)
			self.timeout (3600)

			... do the service ...

			self.close
		;;

	@p
	And here is a typical client program.

	@prog
		use Net

		host   = Net.gethostbyname ('www.foo.bar')
		salist = Net.sockaddr (Net.AF_INET, host, 80)
		sock   = Net.Socket.client (Net.PF_INET, Net.SOCK_STREAM, Net.IPPROTO_IP)

		sock.timeout (30)
		try
			sock.connect (salist)
		catch
			__abort__ ('could not connect: ', __err__)
		;;

		println ('connected')
		sock.timeout (3600)

		... do stuffs ...

		println ('disconnecting ...')
		sock.close
*/
static qu_class SocketClass;

/* the default timeout (seconds) -> now + 10800 */
#define SOCKET_DEFAULT_TIMEOUT 10800.0

typedef struct TSocket TSocket;
struct TSocket
{
	QU_STREAM_HEAD;      /* see qu_stream.h */
	qu_string sa;        /* socket address, sa->str points to sockaddr_s */
	int       domain;    /* domain, i.e. PF_INET */
	int       type;      /* type, i.e. SOCK_STREAM */
	int       protocol;  /* protocol, i.e. 0 (IP) */
};

typedef TSocket *Socket;

/* default qu_stream functions, set by init */
static qu_fstream_s SocketFuncs;

/* test for a Socket */
#define SocketIs(o) qu_object_instof (o, SocketClass)

/* default linebreak */
static qu_string Sbr;

/* flags*/
#define SocketLISTENER QU_FLAG_9 /* is a listener socket */

/* flag tests */
#define SOCKET_IS_LISTENER(self) qu_flag_test ((self), SocketLISTENER)


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	stuffs
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
	We always set non-blocking mode and take care of blocking stuffs ourselves.

	We should be using stream_blocking but I'm not sure if fnctl works while
	SO_NONBLOCK is defined.
*/
static int _socket_set_nonblock (int fd)
{
#ifdef SO_NONBLOCK
	int v = 1;
	socklen_t z = sizeof (v);
	return setsockopt (fd, SOL_SOCKET, SO_NONBLOCK, &v, z) == 0;
#else
	int f = fcntl (fd, F_GETFL, 0);
	return fcntl (fd, F_SETFL, f | O_NONBLOCK) == 0;
#endif
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	standard class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class __inst (domain: Integer, type: Integer, protocol: Integer, listener: Boolean) -> Socket __final
	Create a new Socket instance.

	v{domain} is one of %{Net!PF}.

	v{type} is one of %{Net!SOCK}.

	v{protocol} is the protocol number. See %{Net!IPPROTO}.

	v{listener} is true to create a listener socket, false to create a client
	socket. A listener socket can not call #{send}, #{recv}, etc., while a client
	socket can not call #{accept} and #{listen}.

	After creating a listener socket, you should call #{bind} then #{listen} and
	go into a loop calling #{accept}.

	After creating a client socket, you should call #{connect} instead.

	See C{man 2 socket} for details.

	@exc
	.ESecurity
	.EEnum
	.EArgValue
	.ESystem

	@xmp
		use Net
		sock = Net.Socket (Net.PF_INET, Net.SOCK_STREAM, Net.IPPROTO_IP, false)

	@see
	See #{listener} and #{client} for methods that are pronounced better.
*/
static Socket Socket__inst (qu_class self, qu_integer domain, qu_integer type,
                            qu_integer protocol, qu_boolean listener)
{
	int    fd, dom, tp, prot;
	Socket s;

	__qu_unsafe__;

	QU_ARG_int  (domain, dom);
	QU_ARG_int  (type, tp);
	QU_ARG_int  (protocol, prot);
	QU_ARG_TYPE (listener, qu_boolean);
	QU_ARG_int  (domain, dom);
	QU_ARG_int  (type, tp);
	QU_ARG_int  (protocol, prot);

	fd = socket (dom, tp, prot);
	if (fd < 0)
		QU_THROWERRNO;

	if (!_socket_set_nonblock (fd))
	{
		/* this should never happen! */
		while (close (fd) && errno == EINTR);
		QU_THROWZ (QuESystem, "cant set non-blocking mode");
	}

	assert (self->size >= SocketClass->size);

	s = (Socket) qu_stream_new (self);

	s->flags   |= /*QU_STREAM_FLAG_CANRECV | */ QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANSELECT | QU_STREAM_FLAG_CANCLOSE;
	s->domain   = dom;
	s->type     = tp;
	s->protocol = prot;
	s->fd       = fd;
	s->sa       = 0;
	s->br       = Sbr;
	s->timeout  = NOW + SOCKET_DEFAULT_TIMEOUT;
	s->funcs    = &SocketFuncs;

	if (listener == QuTrue)
		s->flags |= SocketLISTENER;

	return s;
}



/*) sub class new (domain: Integer, type: Integer, protocol: Integer, listener: Boolean) -> Socket
	Create a new Socket instance.

	This method simply calls #{__inst} with the same arguments.

	You would generally override this method in your Socket subclasses.

	@exc
	.ESecurity
	.EEnum
	.EArgValue
	.ESystem
*/
static Socket Socket_new (qu_class self, qu_integer domain, qu_integer type, qu_integer protocol, qu_boolean listener)
{
	return Socket__inst (self, domain, type, protocol, listener);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	specific class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class any (listeners: Array, result: Array, timeout: Integer = 1000) -> Array
	Wait until there are clients trying to connect.

	v{listeners} must be an Array of listener sockets.

	Listeners ready to #{accept} are collected in v{result} which must be an Array
	and the same as v{listeners}. This Array is returned as the result.

	If v{timeout} is the number of microsecond to wait. If it is non-positive then
	it means to wait indefinitely until there is at least one client.

	@warn
		If v{result} is frozen then calling this method will be useless.

	@exc
	.EArgValue
	.ESystem

	@xmp
		use Net
		use Socket
		use Ssl

		Listeners =
		[
			Socket.InetSocket.listener (80)
			Socket.InetSocket.listener (443, true,
				Ssl.Sslctx ('/opt/qu/demo/servercrt.pem', '/opt/qu/demo/serverkey.pem'))
		]

		ready = []
		loop
			for sock in Net.Socket.any (Listeners, ready.__clear)
				sock.timeout = 8
				try
					sock = sock.accept
				catch ETimeout
					# do nothing
				else
					# ... service sock ...
				;;
			;;
		;;

	@note
	This is the same as using -{Stream.fselect} except that this is more efficient.
*/
static qu_array Socket_any (qu_class self, int argc, qu_any *argv)
{
	qu_array   lst, res;
	long       timeout = 1000;
	long       l, i;
	qu_any    *p;
	qu_stream  x;
	int        r, fdmax = 0;
	struct     timeval v;
	fd_set     t;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, lst, qu_array);
	QU_ARG_VTYPE  (1, res, qu_array);
	QU_ARG_CHECK  (lst != res);

	if (argc > 2)
		QU_ARG_vlong (2, timeout);

	l = lst->len;
	if (l == 0)
		return res;

	FD_ZERO (&t);
	p = lst->objs;
	while (l--)
	{
		x = (qu_stream) *p++;
		QU_ARG_CHECK (SocketIs (x) && SOCKET_IS_LISTENER (x) && x->fd >= 0);
		if (x->fd > fdmax)
			fdmax = x->fd;
		FD_SET (x->fd, &t);
	}

	__qu_blocking_begin;
	if (timeout < 0)
	{
		v.tv_sec  = 0;
		v.tv_usec = 1000;
	}
	else
	{
		v.tv_sec  = timeout / 1000000;
		v.tv_usec = timeout % 1000000;
	}

	while ((r = select (fdmax + 1, &t, 0, 0, &v)) < 0)
	{
		if (errno != EINTR)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;

	if (r)
	{
		for (i = 0; i < lst->len; ++i)
		{
			x = (qu_stream) lst->objs[i];
			QU_ARG_CHECK (SocketIs (x) && SOCKET_IS_LISTENER (x) && x->fd >= 0);
			if (FD_ISSET (x->fd, &t))
				qu_array_push_quick (res, x);
		}
	}
	return res;
}



/*) sub class client (domain: Integer, type: Integer, protocol: Integer) -> Socket
	Create a new client Socket.

	This is better pronounced method that calls #{__inst} (domain, type, protocol, true).

	You would generally override this method in your Socket subclasses.

	@exc
	.ESecurity
	.EEnum
	.EArgValue
	.ESystem
*/
static Socket Socket_client (qu_class self, qu_integer domain, qu_integer type, qu_integer protocol)
{
	return Socket__inst (self, domain, type, protocol, QuFalse);
}



/*) sub class listener (domain: Integer, type: Integer, protocol: Integer) -> Socket
	Create a new listener Socket.

	This is better pronounced method that calls #{__inst} (domain, type, protocol, false).

	You would generally override this method in your Socket subclasses.

	@exc
	.ESecurity
	.EEnum
	.EArgValue
	.ESystem
*/
static Socket Socket_listener (qu_class self, qu_integer domain, qu_integer type, qu_integer protocol)
{
	int yes = 1;
	Socket s = Socket__inst (self, domain, type, protocol, QuTrue);
	QU_ENSURE (s);
	setsockopt (s->fd, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof (yes));
	return s;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	low-level instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void Socket__mark (Socket self)
{
	qu__stream__mark ((qu_stream) self);
	qu_gc_xmark (self->sa);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	specific instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub accept -> Socket
	Accept a connection on a socket, return a new #{client} Socket of this
	class.

	self must be a #{islistener} socket.

	The new Socket will inherit most of the underlying socket parameters
	including #{domain}, #{type}, #{protocol} and #{linebreak}. The #{timeout}
	is set as in the #{__inst} method and the socket is non-blocking.

	This method does not return until #{timeout} or a new connection has
	been accepted. You should use #{any} before calling this method unless
	you only have one listener.

	See C{man 2 accept} for details.

	@exc
	.ECall   socket is not a listener socket
	.EArgValue
	.ETimeout
	.ESystem
*/
static Socket Socket_accept (Socket self)
{
	sockaddr_s sa;
	socklen_t  salen;
	int        fd;
	Socket     s;

	QU_REQUIRE (QuECall, SOCKET_IS_LISTENER (self));

	__qu_blocking_begin;
	for (;;)
	{
		memset (&sa, 0, sizeof (sa));
		salen = sizeof (sa);
		fd = accept (self->fd, (struct sockaddr *) &sa, &salen);
		if (fd >= 0)
			break;
		if (errno != EINTR)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		if (NOW >= self->timeout)
		{
			__qu_blocking_end;
			QU_THROW (QuETimeout);
		}
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	if (!_socket_set_nonblock (fd))
	{
		/* this should never happen! */
		while (close (fd) && errno == EINTR);
    	__qu_blocking_end;
		QU_THROWZ (QuESystem, "cant set non-blocking mode");
	}
	__qu_blocking_end;

	s = (Socket) qu_stream_new (self->klass);
	s->flags   |= QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANSELECT | QU_STREAM_FLAG_CANCLOSE;
	s->domain   = self->domain;
	s->type     = self->type;
	s->protocol = self->protocol;
	s->fd       = fd;
	s->sa       = 0; /* not a mistake - to avoid  __qu_gc_critical_begin */
	s->sa       = qu_string_new ((char *) &sa, salen);
	s->br       = self->br;
	s->timeout  = NOW + SOCKET_DEFAULT_TIMEOUT;
	s->funcs    = &SocketFuncs;

	SET_SOCKADDR_FLAG (s->sa);
	return s;
}



/*) sub bind (saddr: String) -> Null __final
	Bind the Socket with the socket address v{saddr}.

	self must be a #{islistener} socket.

	See %{Net!sockaddr} for building a valid socket address.

	See C{man 2 bind} for details.

	@exc
	.ECall   socket already bind to an address or is a client socket
	.EArgValue
	.ESystem
*/
static qu_null Socket_bind (Socket self, qu_string saddr)
{
	QU_REQUIRE (QuECall, SOCKET_IS_LISTENER (self) && self->sa == 0);
	ENSURE_SOCKADDR (saddr);
	if (bind (self->fd, (struct sockaddr *) saddr->str, saddr->len) != 0)
		QU_THROWERRNO;
	self->sa = saddr;
	return QuNil;
}



/*) sub close -> Null
	Close the socket.

	If the socket is alread closed then nothing happens.

	@noexc
*/
static qu_null Socket_close (Socket self)
{
	int eno = errno;
	if (self->fd >= 0)
	{
		if (shutdown (self->fd, SHUT_WR) == 0)
		{
			fd_set  e;
			struct  timeval t;
			char    buf [512];

			__qu_blocking_begin;

			FD_ZERO (&e);
			FD_SET (self->fd, &e);

			t.tv_sec  = 0;
			t.tv_usec = 1000;

			while (select (self->fd + 1, &e, 0, 0, &t) < 0)
			{
				if (errno != EINTR)
					break;
			}

			while (read (self->fd, buf, sizeof (buf)) < 0)
			{
				if (errno != EINTR)
					break;
			}
			__qu_blocking_end;
		}

		while (close (self->fd) != 0)
		{
			if (errno != EINTR)
			{
				/* should never happen */
				QU_WARN ("cant close %d", self->fd);
				break;
			}
		}
		self->fd = -1;
	}
	errno = eno;
	return QuNil;
}



/*) sub connect (saddr) -> Null
	Initiate a connection on a client socket.

	self must U{not} be a #{islistener} socket.

	v{saddr} is the socket address to connect to, or an Array of possible
	socket addresses to connect to.

	If v{saddr} is an Array then connection is attempted to each address until
	one succeeds.

	See %{Net!sockaddr} on how to build a socket address.

	This method does not return until #{timeout} or success.
	On success, the #{socketaddr} property is set.

	See C{man 2 connect} for details.

	@exc
	.ECall     not a client socket
	.EArgValue
	.ETimeout
	.ESystem

	@xmp
		use Net

		host  = Net.gethostbyname ('www.foo.bar')
		addr  = host->addr_list[0]
		saddr = Net.sockaddr (Net.AF_INET, addr, 80)
		sock  = Net.Socket.client (Net.PF_INET, Net.SOCK_STREAM, Net.IPPROTO_IP)
		sock.connect (saddr)

		# and here is a simpler one

		host   = Net.gethostbyname ('www.foo.bar')
		salist = Net.sockaddr (Net.AF_INET, host, 80)
		sock   = Net.Socket.client (Net.PF_INET, Net.SOCK_STREAM, Net.IPPROTO_IP)
		sock.connect (salist)
*/
static qu_null _socket_connect (Socket self, qu_string sa)
{
	ENSURE_SOCKADDR (sa);

	__qu_blocking_begin;
	while (connect (self->fd, (struct sockaddr *) sa->str, sa->len) != 0)
	{
		if (errno != EINTR)
		{
			FIXEAGAIN;
			switch (errno)
			{
				case EAGAIN:
				case EINPROGRESS:
				case EALREADY:
					break;

				default:
					__qu_blocking_end;
					QU_THROWERRNO;
			}
		}
		if (NOW > self->timeout)
		{
			__qu_blocking_end;
			QU_THROW (QuETimeout);
		}
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;
	self->sa = sa;
	return QuNil;
}



static qu_null Socket_connect (Socket self, qu_any x)
{
	QU_REQUIRE (QuECall, !SOCKET_IS_LISTENER (self));

	if (qu_string_is (x))
	{
		return _socket_connect (self, x);
	}
	else
	{
		long   i;
		double to = self->timeout - NOW;

		QU_ARG_TYPE (x, qu_array);
		for (i = 0; i < ((qu_array) x)->len; ++i)
		{
			if (_socket_connect (self, ((qu_array) x)->objs[i]))
				return QuNil;
			self->timeout += to;
		}
		QU_RETHROW;
	}
}



/*) sub domain -> Integer __final
	Return the socket domain as passed to #{__inst}.

	The result is one of %{Net!PF}.

	@noexc
*/
static qu_integer Socket_domain (Socket self)
{
	return qu_integer_newi (self->domain);
}



/*) sub eof -> Boolean __final
	This method always return false since EOF is irrelevant with sockets.

	@noexc
*/
#define Socket_eof qu_object_method_false



/*) sub getsockopt (*arg) -> Any __final
	@def
	sub getsockopt (level: Integer, optname: Integer, len: Integer) -> String
	sub getsockopt (level: Integer, optname: Integer, klass = Boolean) -> Any
	sub getsockopt (level: Integer, optname: Integer, linger: Record) -> Boolean
	@p
	Get socket option.

	If called with two arguments, get an Integer value.

	If called with three arguments and the third argument is a positive Integer
	then get value of v{len} bytes into a String, otherwise you mean to return a
	Boolean value.

	v{level} must be one of %{Net!SOL}.

	v{optname} must be one of %{Net!SO}.

	If v{optname} is SO_LINGER then a special treatment is provided for convenience.
	You must pass a %{Net!Linger} Record and the fields will be filled in.
	The returned value is C{linger->l_onoff}.

	@warn
	Do not set blocking mode until you have read the documenation for the
	#{blocking} method.

	@p
	See C{man 2 getsockopt} for details.

	@exc
	.EArgValue
	.ESystem

	@xmp
		println (mysock.getsockopt (Net.SOL_SOCKET, Net.SO_RCVBUF))
		println (mysock.getsockopt (Net.SOL_SOCKET, Net.SO_KEEPALIVE, Boolean))
		linger = Net.Linger (false, 0)
		println (mysock.getsockopt (Net.SOL_SOCKET, Net.SO_LINGER, linger))
		println (linger->l_onoff)
		println (linger->l_linger)

	@out
		65535
		false
		false
		false
		0

	@see
	#{setsockopt}
*/
static qu_any Socket_getsockopt (Socket self, int argc, qu_any *argv)
{
	int lv, op;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_vint  (0, lv);
	QU_ARG_vint  (1, op);

	if (argc == 3)
	{
		/* handle special case for SO_LINGER */
		if (op == SO_LINGER)
		{
			qu_record r;
			struct linger v;
			socklen_t z = sizeof (v);

			QU_ARG_CHECK (lv == SOL_SOCKET);
			QU_ARG_vrecord (2, r, LingerStruct);
			if (getsockopt (self->fd, SOL_SOCKET, SO_LINGER, &v, &z) != 0)
				QU_THROWERRNO;
			r->objs[0] = v.l_onoff ? QuTrue : QuFalse;
			r->objs[1] = qu_int_newl (v.l_linger);
			return v.l_onoff ? QuTrue : QuFalse;
		}
		else
		{
			int v;
			socklen_t l;

			v = qu_integer_toi (argv[2]);
			if (errno == 0 && v > 0)
			{
				qu_string  s;
				char      *z;

				QU_MEM_CHECK (l <= QU_MEM_MAX);
				s = qu_string_newl (v + 1);
				z = s->str;
				if (getsockopt (self->fd, lv, op, z, &l) != 0)
					QU_THROWERRNO;
				s->str[l] = 0;
				if (l == 0)
					return QuStringEMPTY;
				if (l == 1)
					return qu_string_byte (s->str[0]);
				return s;
			}
			else
			{
				socklen_t z = sizeof (v);

				if (getsockopt (self->fd, lv, op, &v, &z) != 0)
					QU_THROWERRNO;
				return v ? QuTrue : QuFalse;
			}
		}
	}
	else
	{
		int v;
		socklen_t z = sizeof (v);

		if (getsockopt (self->fd, lv, op, &v, &z) != 0)
			QU_THROWERRNO;
		return qu_integer_newi (v);
	}
}



/*) sub keepalive (*arg) -> Any __final
	@def
	sub keepalive -> Boolean
	sub keepalive (setting: Boolean) -> Nil

	@p
	Get or set keepalive mode.

	This is a convenience function over #{getsockopt} and #{setsockopt}.
	It is provided because it spells nicely in the HTTP world.

	@exc
	.EArgValue
	.ESystem

	@xmp
		mysock.keepalive = true
*/
static qu_any Socket_keepalive (Socket self, int argc, qu_any *argv)
{
	int v;
	socklen_t z = sizeof (v);

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc == 0)
	{
		if (getsockopt (self->fd, SOL_SOCKET, SO_KEEPALIVE, &v, &z) != 0)
			QU_THROWERRNO;
		return v ? QuTrue : QuFalse;
	}
	else
	{
		QU_ARG_vbool (0, v);
		if (setsockopt (self->fd, SOL_SOCKET, SO_KEEPALIVE, &v, z) != 0)
			QU_THROWERRNO;
		return QuNil;
	}
}



/*) sub islistener -> Boolean __final
	Return true if self is a listener socket, false if it is a client
	socket.

	@noexc
*/
static qu_boolean Socket_islistener (Socket self)
{
	return SOCKET_IS_LISTENER (self) ? QuTrue : QuFalse;
}



/*) sub listen (backlog: Integer = SOMAXCONN) -> Null
	Listen for connections on a socket.

	self must be a #{islistener} socket.

	Note that the system may not respect v{backlog} and set its own value
	silently but it is a good idea to set it to a specific value nevertheless.

	See C{man 2 listen} for details.

	@exc
	.ECall  not a listener socket
	.EArgValue
	.ESystem
*/
static qu_null Socket_listen (Socket self, int argc, qu_any *argv)
{
	int backlog = SOMAXCONN;

	QU_REQUIRE (QuECall, SOCKET_IS_LISTENER (self));

	QU_ARG_VMAX (1);
	if (argc)
	{
		QU_ARG_vint (0, backlog);
		QU_ARG_CHECK (backlog > 0);
	}
	if (listen (self->fd, backlog) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub protocol -> Integer __final
	Return the socket protocol as passed to #{__inst}.

	The result is generally one of %{Net!IPPROTO}.

	@noexc
*/
static qu_integer Socket_protocol (Socket self)
{
	return qu_integer_newi (self->protocol);
}



/*) sub setsockopt (*arg) -> Any __final
	@def
	sub setsockopt (level: Integer, optname: Integer, val: Integer, bool: confirm = False) -> Any
	sub setsockopt (level: Integer, optname: Integer, val: String, bool: confirm = False) -> Any
	sub setsockopt (level: Integer, optname: Integer, val: Boolean, bool: confirm = False) -> Any
	sub setsockopt (level: Integer, optname: Integer, linger: Record) -> Boolean
	@p
	Set socket option.

	v{level} must be one of %{Net!SOL}.

	v{optname} must be one of %{Net!SO}.

	If the third argument is an Integer then set an Integer value.
	Otherwise set value of v{len} bytes from String.

	If v{optname} is SO_LINGER then a special treatment is provided for convenience.
	You must pass a %{Net!Linger} Record with the fields filled in as you desire.
	The returned value is C{linger->l_onoff}.

	If v{confirm} is true then return the actual value set, i.e. return the
	result of #{getsockopt}, return Nil otherwise. This might be useful if you
	suspect that the system might silently disrespect the value and you don't
	like it.

	See C{man 2 setsockopt} for details.

	@exc
	.EArgValue
	.ESystem

	@xmp
		println (mysock.getsockopt (Net.SOL_SOCKET, Net.SO_RCVBUF))
		println (mysock.setsockopt (Net.SOL_SOCKET, Net.SO_RCVBUF, 4096, true))

		linger = Net.Linger (true, 30)
		println (mysock.setsockopt (Net.SOL_SOCKET, Net.SO_LINGER, linger))

	@out
		65535
		8192
		true

	@see
	#{getsockopt}
*/
static qu_any Socket_setsockopt (Socket self, int argc, qu_any *argv)
{
	int lv, op, confirm = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (3, 4);
	QU_ARG_vint (0, lv);
	QU_ARG_vint (1, op);

	/* handle special case for SO_LINGER */
	if (op == SO_LINGER)
	{
		qu_record r;
		struct linger v;
		socklen_t z = sizeof (v);

		QU_ARG_CHECK (lv == SOL_SOCKET);
		QU_ARG_vrecord (2, r, LingerStruct);
		v.l_onoff = (r->objs[0] == QuTrue);
		v.l_linger = qu_int_tol (r->objs[1]);
		if (setsockopt (self->fd, SOL_SOCKET, SO_LINGER, &v, z) != 0)
			QU_THROWERRNO;
		return v.l_onoff ? QuTrue : QuFalse;
	}

	if (argc > 3)
		confirm = (argv[3] == QuTrue);

	if (qu_string_is (argv[2]))
	{
		qu_string  s = argv[2];
		char      *z = s->str;
		socklen_t  l;

		QU_ARG_CHECK (s->len < INT_MAX);

		l = (socklen_t) s->len;
		if (setsockopt (self->fd, lv, op, z, l) != 0)
			QU_THROWERRNO;
		if (confirm)
		{
			s = qu_string_newl ((long) l + 1);
			z = s->str;
			if (getsockopt (self->fd, lv, op, z, &l) != 0)
				QU_THROWERRNO;
			return s;
		}
	}
	else if (qu_boolean_is (argv[2]))
	{
		int v;
		socklen_t z = sizeof (v);

		v = (argv[2] == QuTrue);
		if (setsockopt (self->fd, lv, op, &v, z) != 0)
			QU_THROWERRNO;

		if (confirm)
		{
			if (getsockopt (self->fd, lv, op, &v, &z) != 0)
				QU_THROWERRNO;
			return v ? QuTrue : QuFalse;
		}
	}
	else
	{
		int v;
		socklen_t z = sizeof (v);

		QU_ARG_vint (2, v);
		if (setsockopt (self->fd, lv, op, &v, z) != 0)
			QU_THROWERRNO;

		if (confirm)
		{
			if (getsockopt (self->fd, lv, op, &v, &z) != 0)
				QU_THROWERRNO;
			return qu_integer_newi (v);
		}
	}
	return QuNil;
}




/*) sub shutdown (how: Integer) -> self __final
	Causes all or part of a full-duplex connection on the socket to be shut down.

	v{how} must be one of:
	@args
	.SHUT_RD   further receptions will be disallowed.
	.SHUT_WR   further transmissions will be disallowed.
	.SHUT_RDWR further receptions and transmissions will be disallowed.

	@p
	See C{man 2 shutdown} for details.

	@warn
	This does not close the socket.
	You must explicitly call #{close} for that.

	@exc
	.EEnum
	.ESystem
*/
static qu_null Socket_shutdown (Socket self, qu_integer h)
{
	int how;

	QU_ARG_int (h, how);
	if (shutdown (self->fd, how) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub sockaddr -> String __final
	Get the socket address.

	For a #{listener} socket, this is the local address set with #{bind}.

	For a #{client} socket, this is the host socket address connected to this
	Socket.

	@exc
	.ENone none is set
*/
static qu_string Socket_sockaddr (Socket self)
{
	QU_REQUIRE (QuENone, self->sa);
	return self->sa;
}



/*) sub type -> Integer __final
	Return the socket type as passed to #{__inst}.

	The result is one of %{Net!SOCK}.

	@noexc
*/
static qu_integer Socket_type (Socket self)
{
	return qu_integer_newi (self->type);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Net_init (qu_module m)
{
	_net_init_enum   (m);
	_net_init_const  (m);
	_net_init_struct (m);

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		general
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	qu_module_add_function (m, "gethostbyname"   , Net_gethostbyname   ,  1);
	qu_module_add_function (m, "gethostbyaddr"   , Net_gethostbyaddr   ,  2);
	qu_module_add_function (m, "gethostname"     , Net_gethostname     ,  0);
	qu_module_add_function (m, "getnameinfo"     , Net_getnameinfo     ,  2);
	qu_module_add_function (m, "getip"           , Net_getip           , -1);
	qu_module_add_function (m, "getprotobyname"  , Net_getprotobyname  ,  1);
	qu_module_add_function (m, "getprotobynumber", Net_getprotobynumber,  1);
	qu_module_add_function (m, "getprotodict"    , Net_getprotodict    , -1);
	qu_module_add_function (m, "getservbyname"   , Net_getservbyname   , -1);
	qu_module_add_function (m, "getservbyport"   , Net_getservbyport   , -1);
	qu_module_add_function (m, "getservdict"     , Net_getservdict     , -1);
	qu_module_add_function (m, "htonl"           , Net_htonl           ,  1);
	qu_module_add_function (m, "htons"           , Net_htons           ,  1);
	qu_module_add_function (m, "inet_ntop"       , Net_inet_ntop       ,  2);
	qu_module_add_function (m, "inet_pton"       , Net_inet_pton       ,  2);
	qu_module_add_function (m, "ipv6"            , Net_ipv6            , -1);
	qu_module_add_function (m, "isip"            , Net_isip            ,  1);
	qu_module_add_function (m, "isname"          , Net_isname          ,  1);
	qu_module_add_function (m, "ntohl"           , Net_ntohl           ,  1);
	qu_module_add_function (m, "ntohs"           , Net_ntohs           ,  1);
	qu_module_add_function (m, "sethostname"     , Net_sethostname     ,  1);
	qu_module_add_function (m, "sockaddr"        , Net_sockaddr        , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Socket
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	SocketClass = qu_class_newz (m, "Socket", TSocket,
		(qu_destroy_f) qu__stream__destroy, (qu_mark_f) Socket__mark);

	SocketClass->id = QU_CLASS_ID_Socket;

	SocketFuncs.rdr   = (qu_stream_rdr_f)   read;
	SocketFuncs.wrt   = (qu_stream_wrt_f)   write;
	SocketFuncs.ordr  = (qu_stream_ordr_f)  QuStreamDefaultFuncs.ordr;
	SocketFuncs.owrt  = (qu_stream_owrt_f)  QuStreamDefaultFuncs.owrt;
	SocketFuncs.acc   = (qu_stream_acc_f)   Socket_accept;
	SocketFuncs.con   = (qu_stream_con_f)   Socket_connect;
	SocketFuncs.seek  = (qu_stream_seek_f)  QuStreamDefaultFuncs.seek;
	SocketFuncs.oseek = (qu_stream_oseek_f) QuStreamDefaultFuncs.oseek;

	Sbr = qu_string_new_staticz ("\r\n");

	/* standard class methods */
	qu_class_method_class  (SocketClass, QU_NAME__inst , Socket__inst, 4);
	qu_class_method_class  (SocketClass, QU_NAME_new   , Socket_new  , 4);

	/* specific class methods */
	qu_class_method_classz (SocketClass, "any"     , Socket_any     , -1);
	qu_class_method_classz (SocketClass, "client"  , Socket_client  ,  3);
	qu_class_method_classz (SocketClass, "listener", Socket_listener,  3);

	/* specific instance methods */
	qu_class_method_instzf (SocketClass, "bind"      , Socket_bind      ,  1);
	qu_class_method_instz  (SocketClass, "close"     , Socket_close     ,  0);
	qu_class_method_instz  (SocketClass, "domain"    , Socket_domain    ,  0);
	qu_class_method_instzf (SocketClass, "eof"       , Socket_eof       ,  0);
	qu_class_method_instzf (SocketClass, "getsockopt", Socket_getsockopt, -1);
	qu_class_method_instzf (SocketClass, "islistener", Socket_islistener,  0);
	qu_class_method_instzf (SocketClass, "keepalive" , Socket_keepalive , -1);
	qu_class_method_instz  (SocketClass, "listen"    , Socket_listen    , -1);
	qu_class_method_instzf (SocketClass, "protocol"  , Socket_protocol  ,  0);
	qu_class_method_instzf (SocketClass, "setsockopt", Socket_setsockopt, -1);
	qu_class_method_instzf (SocketClass, "shutdown"  , Socket_shutdown  ,  1);
	qu_class_method_instzf (SocketClass, "sockaddr"  , Socket_sockaddr  ,  0);
	qu_class_method_instzf (SocketClass, "type"      , Socket_type      ,  0);

#if 0 /* dont override qu_stream's otherwise SSL wont work */
	qu_class_method_instz  (SocketClass, "accept"    , Socket_accept    ,  0);
	qu_class_method_instz  (SocketClass, "connect"   , Socket_connect   ,  1);
#endif

	/* finalize declaration */
	qu_class_set_super (SocketClass, QuStreamClass);
	qu_class_enddecl   (SocketClass);

	return 1;
}
