/* Os.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_UTIME_H
#include <utime.h>
#endif
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_GRP_H
#include <grp.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_LIMITS_H
#include <sys/limits.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif
#ifdef HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>
#endif
#ifdef HAVE_SYS_TIMES_H
#include <sys/times.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif


/*) module Os
	This module provides some functions related to the operating system.

	@xmp
		use Os
		println (Os.uname)

	@out
		['Linux', 'localhost.localdomain', '2.4.18-3', '#1 Thu Apr 18 07:37:53 EDT 2002', 'i686']

	@note
	Most functions are marked as k{__unsafe} thus can not be called from an
	untrusted module. See -{eval} for more information.
*/


/*----------------------------------------------------------------------------
	enum, const and struct
----------------------------------------------------------------------------*/

/*) enum E
		EPERM
		ENOENT
		ESRCH
		EINTR
		EIO
		ENXIO
		E2BIG
		ENOEXEC
		EBADF
		ECHILD
		EAGAIN
		ENOMEM
		EACCES
		EFAULT
		ENOTBLK
		EBUSY
		EEXIST
		EXDEV
		ENODEV
		ENOTDIR
		EISDIR
		EINVAL
		ENFILE
		EMFILE
		ENOTTY
		ETXTBSY
		EFBIG
		ENOSPC
		ESPIPE
		EROFS
		EMLINK
		EPIPE
		EDOM
		ERANGE
		EDEADLK
		ENAMETOOLONG
		ENOLCK
		ENOSYS
		ENOTEMPTY
		ELOOP
		EWOULDBLOCK
		ENOMSG
		EIDRM
		ECHRNG
		EL2NSYNC
		EL3HLT
		EL3RST
		ELNRNG
		EUNATCH
		ENOCSI
		EL2HLT
		EBADE
		EBADR
		EXFULL
		ENOANO
		EBADRQC
		EBADSLT
		EDEADLOCK
		EBFONT
		ENOSTR
		ENODATA
		ETIME
		ENOSR
		ENONET
		ENOPKG
		EREMOTE
		ENOLINK
		EADV
		ESRMNT
		ECOMM
		EPROTO
		EMULTIHOP
		EDOTDOT
		EBADMSG
		EOVERFLOW
		ENOTUNIQ
		EBADFD
		EREMCHG
		ELIBACC
		ELIBBAD
		ELIBSCN
		ELIBMAX
		ELIBEXEC
		EILSEQ
		ERESTART
		ESTRPIPE
		EUSERS
		ENOTSOCK
		EDESTADDRREQ
		EMSGSIZE
		EPROTOTYPE
		ENOPROTOOPT
		EPROTONOSUPPORT
		ESOCKTNOSUPPORT
		EOPNOTSUPP
		EPFNOSUPPORT
		EAFNOSUPPORT
		EADDRINUSE
		EADDRNOTAVAIL
		ENETDOWN
		ENETUNREACH
		ENETRESET
		ECONNABORTED
		ECONNRESET
		ENOBUFS
		EISCONN
		ENOTCONN
		ESHUTDOWN
		ETOOMANYREFS
		ETIMEDOUT
		ECONNREFUSED
		EHOSTDOWN
		EHOSTUNREACH
		EALREADY
		EINPROGRESS
		ESTALE
		EUCLEAN
		ENOTNAM
		ENAVAIL
		EISNAM
		EREMOTEIO
		EDQUOT
		ENOMEDIUM
		EMEDIUMTYPE
	;;
	Possible #{errno} value.
*/


/*) enum F_
		F_RDLCK
		F_WRLCK
		F_UNLCK
		F_DUPFD
		F_GETFD
		F_SETFD
		F_GETFL
		F_SETFL
		F_GETLK
		F_SETLK
		F_SETLKW
		F_GETLK64
		F_SETLK64
		F_SETLKW64
		F_SETOWN
		F_GETOWN
		F_SETSIG
		F_GETSIG
		F_SETLEASE
		F_GETLEASE
		F_NOTIFY

		F_LOCK
		F_TLOCK
		F_ULOCK
		F_TEST
	;;
	Constants for f{fcntl} and others.

	Note that some values may not be defined.
	See C{man 2 fcntl} for details.
*/


/*) enum O_
		O_RDONLY
		O_WRONLY
		O_RDWR
		O_CREAT
		O_EXCL
		O_NOCTTY
		O_TRUNC
		O_APPEND
		O_NONBLOCK
		O_SYNC
		O_NOFOLLOW
		O_DIRECTORY
		O_LARGEFILE
	;;
	Constants for f{open}.

	Note that some values may not be defined.
	See C{man 2 open} for details.
*/


/*) enum SEEK
		SEEK_SET
		SEEK_CUR
		SEEK_END
	;;
	Possible values for f{lseek}.
*/


/*) enum S_
		S_IFMT   : 0170000   # bitmask for the file type bitfields
		S_IFSOCK : 0140000   # socket
		S_IFLNK  : 0120000   # symbolic link
		S_IFREG  : 0100000   # regular file
		S_IFBLK  : 0060000   # block device
		S_IFDIR  : 0040000   # directory
		S_IFCHR  : 0020000   # character device
		S_IFIFO  : 0010000   # fifo
		S_ISUID  : 0004000   # set UID bit
		S_ISGID  : 0002000   # set GID bit (see below)
		S_ISVTX  : 0001000   # sticky bit (see below)
		S_IRWXU  : 00700     # mask for file owner permissions
		S_IRUSR  : 00400     # owner has read permission
		S_IWUSR  : 00200     # owner has write permission
		S_IXUSR  : 00100     # owner has execute permission
		S_IRWXG  : 00070     # mask for group permissions
		S_IRGRP  : 00040     # group has read permission
		S_IWGRP  : 00020     # group has write permission
		S_IXGRP  : 00010     # group has execute permission
		S_IRWXO  : 00007     # mask for permissions for others (not in group)
		S_IROTH  : 00004     # others have read permission
		S_IWOTH  : 00002     # others have write permisson
		S_IXOTH  : 00001     # others have execute permission
	;;
	Constants for f{open} and to test the v{st_mode} field of f{stat}.

	See C{man 2 open} and C{man 2 stat} for details.
*/


/*) enum SIG
       SIGHUP
       SIGINT
       SIGQUIT
       SIGILL
       SIGABRT
       SIGFPE
       SIGKILL
       SIGSEGV
       SIGPIPE
       SIGALRM
       SIGTERM
       SIGUSR1
       SIGUSR2
       SIGCHLD
       SIGCONT
       SIGSTOP
       SIGTSTP
       SIGTTIN
       SIGTTOU
       SIGBUS
       SIGPOLL
       SIGPROF
       SIGSYS
       SIGTRAP
       SIGURG
       SIGVTALRM
       SIGXCPU
       SIGXFSZ
       SIGIOT
       SIGEMT
       SIGSTKFLT
       SIGIO
       SIGCLD
       SIGPWR
       SIGINFO
       SIGLOST
       SIGWINCH
       SIGUNUSED
	;;
	Signal list.

	Note that some may not be defined.
	See C{man 7 signal} for details.
*/


/*) enum PRIO
		PRIO_PROCESS
		PRIO_PGRP
		PRIO_USER
	;;
	Possible values for v{which} in #{getpriority}.

	See #{man 2 getpriority} for details.
*/


/*) enum _PC
		_PC_LINK_MAX
		_PC_MAX_CANON
		_PC_MAX_INPUT
		_PC_NAME_MAX
		_PC_PATH_MAX
		_PC_PIPE_BUF
		_PC_CHOWN_RESTRICTED
		_PC_NO_TRUNC
		_PC_VDISABLE
	;;
	Possible values for v{name} in #{pathconf}.

	See #{man 3 pathconf} for details.
*/


/*) enum _SC
		_SC_ARG_MAX
		_SC_CHILD_MAX
		_SC_CLK_TCK
		_SC_STREAM_MAX
		_SC_TZNAME_MAX
		_SC_OPEN_MAX
		_SC_JOB_CONTROL
		_SC_SAVED_IDS
		_SC_VERSION
		_SC_BC_BASE_MAX
		_SC_BC_DIM_MAX
		_SC_BC_SCALE_MAX
		_SC_BC_STRING_MAX
		_SC_COLL_WEIGHTS_MAX
		_SC_EXPR_NEST_MAX
		_SC_LINE_MAX
		_SC_RE_DUP_MAX
		_SC_2_VERSION
		_SC_2_DEV
		_SC_2_FORT_DEV
		_SC_2_FORT_RUN
		_SC_2_LOCALEDEF
		_SC_2_SW_DEV
		_SC_PAGE_SIZE
		_SC_PHYS_PAGES
		_SC_AVPHYS_PAGES
	;;
	The possible values for v{name} in #{sysconf}.

	Note that some values may not be defined.
	See C{man 3 sysconf} for details.
*/


/*) enum WAIT
		WNOHANG
		WUNTRACED
		__WCLONE
		__WALL
		__WNOTHREAD
	;;
	Possible options for #{waitpid}.

	Note that the __WXXX options might not be available.
*/


/*) enum RLIMIT
       RLIMIT_CPU
       RLIMIT_FSIZE
       RLIMIT_DATA
       RLIMIT_STACK
       RLIMIT_CORE
       RLIMIT_RSS
       RLIMIT_NPROC
       RLIMIT_NOFILE
       RLIMIT_MEMLOCK
       RLIMIT_AS
	;;
	Possible value for v{resource} in #{getrlimit} etc.
*/


/*) const RLIM_INFINITY = -1
	A possible setting for #{setrlimit}.
*/
#ifndef RLIM_INFINITY
#define RLIM_INFINITY -1
#endif


/*) enum RUSAGE
		RUSAGE_SELF
		RUSAGE_CHILDREN
	;;
	The possible values for v{who} in #{getrusage}.
*/


/*) struct Usage
		utime     : Decimal # user time used
		stime     : Decimal # system time used
		maxrss    : Integer # maximum resident set size
		ixrss     : Integer # integral shared memory size
		idrss     : Integer # integral unshared data size
		isrss     : Integer # integral unshared stack size
		minflt    : Integer # page reclaims
		majflt    : Integer # page faults
		nswap     : Integer # swaps
		inblock   : Integer # block input operations
		oublock   : Integer # block output operations
		msgsnd    : Integer # messages sent
		msgrcv    : Integer # messages received
		nsignals  : Integer # signals received
		nvcsw     : Integer # voluntary context switches
		nivcsw    : Integer # involuntary context switches
	;;
	This -{Struct} is used to store information returned by #{getrusage}.
*/
#ifdef HAVE_GETRUSAGE
static qu_struct UsageStruct;
#endif


/*) const CLOCKS_PER_SEC = 1000000
	The number of clock ticks per second.

	@warn
	The value is set by the system.
	POSIX requires that it is 1000000 but you should probably not rely on that.
*/


/*) const __sig__ = const [:]
	This is a frozen Dict containing the current signal handlers.
	See #{signal}.
*/



/*----------------------------------------------------------------------------
	functions
----------------------------------------------------------------------------*/

/*) sub alarm (seconds: Integer) -> Null
	Arrange for a SIGALRM signal to be delivered to the process in v{seconds}.

	If seconds is zero, no new alarm is scheduled.

	In any event any previously set alarm is cancelled.

	See C{man 2 alarm} for details.

	@exc
	.EValue
	.ESystem
	.ESecurity
*/
static qu_integer Os_alarm (qu_integer seconds)
{
	unsigned int s;

	__qu_unsafe__;

	QU_ARG_uint (seconds, s);
	s = alarm (s);
	return qu_integer_newui (s);
}




/*) sub clock -> Integer
	Return an approximation of processor time used by the program in clock ticks.

	Use #{CLOCKS_PER_SEC} to get the number of seconds.

	@noexc

	@xmp
		use Os

		for i in 1, 100000
			i = 0
		;;
		println (Os.clock)

	@out
		2000
*/
static qu_integer Os_clock ()
{
	return qu_integer_newl ((long) clock ());
}



/*) sub errno (*arg) -> Integer
	@def
	sub errno -> Integer
	sub errno (value: Integer) -> Integer
	@p
	Get or set the value of the v{errno}.

	@exc
	.EValue
*/
static qu_integer Os_errno (int argc, qu_any *argv)
{
	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc)
	{
		int v;
		QU_ARG_vint (0, v);
		errno = v;
	}
	return qu_integer_newi (errno);
}



/*) sub execv (path: String, argv: Array, env: Array)
	Execute the program pointed to by v{path}.

	v{argv} is an Array of String arguments to pass to the new program.

	See C{man 2 execv} for details.

	On success, this function does not return.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.execv ('/usr/bin/qu', ['/opt/qu/demo/hello.qu'])
*/
static qu_any _os_exec (qu_string path, qu_array arg, qu_array env, int p)
{
	char    **argp, **envp = 0;
	long      i;
	qu_string s;

	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	QU_ARG_TYPE (arg, qu_array);

	argp = qu_mem_calloc ((arg->len + 2), sizeof (char *));
	if (env)
	{
		QU_ARG_TYPE (env, qu_array);
		envp = qu_mem_calloc ((env->len + 1), sizeof (char *));
	}

	argp[0] = path->str;
	for (i = 0; i < arg->len; i++)
	{
		s = arg->objs[i];
		if (!qu_string_is (s))
			goto nope;
		argp[i + 1] = s->str;
	}

	if (env)
	{
		for (i = 0; i < env->len; i++)
		{
			s = env->objs[i];
			if (!qu_string_is (s))
				goto nope;
			envp[i] = s->str;
		}
		execve (path->str, argp, envp);
	}
	else
	{
		if (p)
			execvp (path->str, argp);
		else
			execv  (path->str, argp);
	}

	qu_mem_free (argp);
	if (env)
		qu_mem_free (envp);
	QU_THROWERRNO;

  nope:
	qu_mem_free (argp);
	if (env)
		qu_mem_free (envp);
	QU_THROW (QuEValue);
}



static qu_any Os_execv (qu_string path, qu_array arg)
{
	return _os_exec (path, arg, NULL, 0);
}



/*) sub execve (filename: String, argv: Array, env: Array)
	Execute the program pointed to by v{filename}.

	v{argv} is an Array of String arguments to pass to the new program.
	v{env} is an Array to pass as environment to the new program, conventionally
	of the form key=value.

	See C{man 2 execve} for details.

	On success, this function does not return.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.execve ('/usr/bin/qu', ['/opt/qu/demo/hello.qu'], ['foo=bar'])
*/
static qu_any Os_execve (qu_string path, qu_array arg, qu_array env)
{
	return _os_exec (path, arg, env, 2);
}



/*) sub execvp (filename: String, argv: Array)
	Execute the program pointed to by v{filename}.

	v{argv} is an Array of String arguments to pass to the new program.

	See C{man 2 execvp} for details.

	On success, this function does not return.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.execvp ('qu', ['/opt/qu/demo/hello.qu'])
*/
static qu_any Os_execvp (qu_string path, qu_array arg)
{
	return _os_exec (path, arg, NULL, 1);
}



/*) sub getegid -> Integer
	Return the effective group ID of the current process.

	See C{man 2 getegid} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.getegid)

	@out
		123
*/
static qu_integer Os_getegid ()
{
	gid_t g;

	__qu_unsafe__;

	g = getegid ();
	return qu_integer_newl ((long) g);
}



/*) sub geteuid -> Integer
	Return the effective user ID of the current process.

	See C{man 2 geteuid} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.geteuid)

	@out
		123
*/
static qu_integer Os_geteuid ()
{
	gid_t g;

	__qu_unsafe__;

	g = geteuid ();
	return qu_integer_newl ((long) g);
}



/*) sub getgid -> Integer
	Return the real group ID of the current process.

	See C{man 2 getgid} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.getgid)
	@out
		123
*/
static qu_integer Os_getgid ()
{
	gid_t g;

	__qu_unsafe__;

	g = getgid ();
	return qu_integer_newl ((long) g);
}



/*) sub getgroups -> Array
	Return supplementary group IDs.

	See C{man 2 getgroups} for details.

	@exc
	.EImplement f{setgroups} is not available on your system
	.ESystem
	.ESecurity

	@xmp
		use Os
		println (Os.getgroups)

	@out
		[123, 456]
*/
static qu_array Os_getgroups ()
{
#ifdef HAVE_GETGROUPS
#  ifndef NGROUPS_MAX
#     define NGROUPS_MAX 64
#  endif
	qu_array a;
	gid_t    g [NGROUPS_MAX];
	int      i, l;

	__qu_unsafe__;

	l = getgroups (NGROUPS_MAX, g);
	if (l < 0)
		QU_THROWERRNO;
	a = qu_array_new (l);
	for (i = 0; i < l; ++i)
		qu_array_push (a, qu_integer_newl ((long) g[i]));
    return a;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub getlogin -> String
	Return the name of the user logged in on the controlling terminal of the process
	or throws ENone if this information cannot be determined.

	See C{man 3 getlogin} for details.

	@exc
	.ESecurity
	.ENone

	@xmp
		use Os
		println (Os.getlogin)

	@out
		foo
*/
static qu_string Os_getlogin ()
{
	char *s;

	__qu_unsafe__;

	s = getlogin ();
	QU_REQUIRE (QuENone, s);
	return qu_string_newz (s);
}



/*) sub fork -> Integer
	Create a child process that differs from the parent process only in its
	PID and PPID, and in the fact that resource utilizations are set to 0.
	File locks and pending signals are not inherited.

	See C{man 2 fork} for details.

	On success, the PID of the child process is returned in the parent's thread of
	execution, and a 0 is returned in the child's thread of execution.

	On failure an ESystem exception is thrown in the parent's context, no child
	process will be created,

	@exc
	.ESecurity
	.ESystem

	@warn
	The bevahior of executing f{fork} from within a Qu program is not very well
	tested. Please report if you have any problem.
*/
static qu_integer Os_fork ()
{
	pid_t p;

	__qu_unsafe__;

	p = fork ();
	if (p < 0)
		QU_THROWERRNO;
	return qu_integer_newl ((long) p);
}



/*) sub getpgrp -> Integer
	This is equivalent to getpgid (0).

	See C{man 2 getpgrp} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.getpgrp)

	@out
		123
*/
static qu_integer Os_getpgrp ()
{
	long p;

	__qu_unsafe__;

	p = (long) getpgrp ();
	return qu_integer_newl (p);
}



/*) sub getpgid (pid: Integer) -> Integer
	Return the process group ID of the process specified by v{pid}.
	If v{pid} is zero, the process ID of the current process is used.

	See C{man 2 getpgid} for details.

	@exc
	.EImplement f{getpgid} is not available on your system
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		println (Os.getpgid (0))

	@out
		123
*/
static qu_integer Os_getpgid (qu_integer pid)
{
#ifdef HAVE_GETPGID
	extern pid_t getpgid (pid_t pid);

	long p;

	__qu_unsafe__;

	QU_ARG_long (pid, p);
	QU_ARG_CHECK (p >= 0);
	p = (long) getpgid ((pid_t) p);
	if (p < 0)
		QU_THROWERRNO;
	return qu_integer_newl (p);
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub getpid -> Integer
	Return the process ID of the current process.

	See C{man 2 getpid} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.getpid)

	@out
		123
*/
static qu_integer Os_getpid ()
{
	long p;

	__qu_unsafe__;

	p = (long) getpid ();
	return qu_integer_newl (p);
}



/*) sub getppid -> Integer
	Return the process ID of the parent of the current process.

	See C{man 2 getppid} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.getppid)

	@out
		123
*/
static qu_integer Os_getppid ()
{
	long p;

	__qu_unsafe__;

	p = (long) getppid ();
	return qu_integer_newl (p);
}



/*) sub getpriority (which: Int, who: Int) -> Int
	Get the scheduling priority of the process, process group, or user, as
	indicated by v{which} and v{who}

	v{which} is one of #{PRIO} and v{who} is interpreted relative to v{which}.

	See C{man 2 getpriority} for details.

	@exc
	.EImplement f{getpriority} is not only available on your system
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		println (Os.getpriority (Os.PRIO_PROCESS, 0))

	@out
		0
*/
static qu_int Os_getpriority (qu_int which, qu_int who)
{
#ifdef HAVE_GETPRIORITY
	int w, o;

	__qu_unsafe__;

	QU_ARG_int (which, w);
	QU_ARG_int (who, o);

	errno = 0;
	o = getpriority (w, o);
	if (o == -1 && errno)
		QU_THROWERRNO;
	return qu_int_new (o);
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub getrlimit (resource: Integer) -> Array
	Get the current and maximum limit of the specified v{resource}.

	v{resource} must be one of #{RLIMIT}.

	See C{man 2 getrlimit} for details.

	The returned value is an Array of 2 items:
	@args
	.current the current setting
	.max     the maximum value

	@exc
	.EImplement f{getrlimit} is not only available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		(rcur, rmax) = Os.getrlimit (Os.RLIMIT_NOFILE)
		println (rcur)
		println (rmax)

	@out
		1024
		1024
*/
static qu_array Os_getrlimit (qu_integer res)
{
#ifdef HAVE_GETRLIMIT
	struct rlimit m;
	int r;

	__qu_unsafe__;

	QU_ARG_int (res, r);
	if (getrlimit (r, &m) < 0)
		QU_THROWERRNO;
	return qu_array_new_ex (2, qu_integer_newl ((long) m.rlim_cur), qu_integer_newl ((long) m.rlim_max));
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub getrusage (who: Integer) -> Record
	Return the current resource usages, for a v{who} of either RUSAGE_SELF
	or RUSAGE_CHILDREN.  The former asks for resources used by the current process,
	the  latter for resources used by those of its children that have terminated and
	have been waited for.

	The returned value is a Record of #{Usage}.

	@exc
	.EImplement f{getrusage} is not available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		u = Os.getrusage (Os.RUSAGE_SELF)
		println (u->ixrss)
*/
static qu_record Os_getrusage (qu_integer who)
{
#ifdef HAVE_GETRUSAGE
	struct rusage u;
	int w;

	__qu_unsafe__;

	QU_ARG_int (who, w);
	if (getrusage (w, &u) < 0)
		QU_THROWERRNO;

	return qu_record_new_ex (UsageStruct,
		qu_float_new ((double) u.ru_utime.tv_sec + (double) u.ru_utime.tv_usec / 1e6),
		qu_float_new ((double) u.ru_stime.tv_sec + (double) u.ru_stime.tv_usec / 1e6),
		qu_integer_newl (u.ru_maxrss),
		qu_integer_newl (u.ru_ixrss),
		qu_integer_newl (u.ru_idrss),
		qu_integer_newl (u.ru_isrss),
		qu_integer_newl (u.ru_minflt),
		qu_integer_newl (u.ru_majflt),
		qu_integer_newl (u.ru_nswap),
		qu_integer_newl (u.ru_inblock),
		qu_integer_newl (u.ru_oublock),
		qu_integer_newl (u.ru_msgsnd),
		qu_integer_newl (u.ru_msgrcv),
		qu_integer_newl (u.ru_nsignals),
		qu_integer_newl (u.ru_nvcsw),
		qu_integer_newl (u.ru_nivcsw));
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub getsid (pid: Integer) -> Integer
	Returns the
       session ID of the process with process ID p.	Return the session ID of the calling process.

	See C{man 2 getsid} for details.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		println (Os.getsid)

	@out
		123
*/
extern pid_t getsid (pid_t pid);
static qu_integer Os_getsid (qu_integer pid)
{
	long p;

	__qu_unsafe__;

	QU_ARG_long (pid, p);
	QU_ARG_CHECK (p >= 0);
	p = (long) getsid (p);
	if (p < 0)
		QU_THROWERRNO;
	return qu_integer_newl (p);
}



/*) sub getuid -> Integer
	Return the real user ID of the current process.

	See C{man 2 getuid} for details.

	@exc
	.ESecurity

	@xmp
		use Os
		println (Os.getuid)

	@out
		123
*/
static qu_integer Os_getuid ()
{
	gid_t g;

	__qu_unsafe__;

	g = getuid ();
	return qu_integer_newl ((long) g);
}



/*) sub initgroups (user: String, gid: Integer) -> Null
	Initialize the group access list.

	See C{man 3 initgrups} for details.

	@exc
	.EImplement f{initgroups} is not available on your system
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		Os.initgroups ('foo', 500)
*/
static qu_null Os_initgroups (qu_string user, qu_integer gid)
{
#ifdef HAVE_INITGROUPS
	long g;

	__qu_unsafe__;

	QU_ARG_TYPE  (user, qu_string);
	QU_ARG_long  (gid, g);
	QU_ARG_CHECK (gid >= 0);
	if (initgroups (user->str, (gid_t) g) != 0)
		QU_THROWERRNO;
	return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}




/*) sub kill (pid: Integer, sig: Integer) -> Null
	Send the signal v{sig} to the process v{pid}.

	v{sig} is one of #{SIG}.

	See C{man 2 kill} for details.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.kill (12345, Os.SIGTERM)
*/
static qu_null Os_kill (qu_integer pid, qu_integer sig)
{
	long p;
	int  s;

	__qu_unsafe__;

	QU_ARG_long (pid, p);
	QU_ARG_int  (sig, s);
	if (kill (p, s) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub killpg (pgrp: Integer, sig: Integer) -> Null
	Send the signal v{sig} to the process group v{pgrp}.

	v{sig} is one of #{SIG}.

	See C{man 2 killpg} for details.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.killpg (12345, Os.SIGTERM)
*/
static qu_null Os_killpg (qu_integer pgrp, qu_integer sig)
{
	long g;
	int  s;

	__qu_unsafe__;

	QU_ARG_long (pgrp, g);
	QU_ARG_int  (sig, s);
	if (killpg (g, s) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub nice (inc: Int) -> Int
	Add v{inc} to the nice value for the calling process and return the new priority.

	See #{man 2 nice} for details.

	@exc
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		println (Os.nice (3))
	@out
		3
*/
static qu_int Os_nice (qu_int inc)
{
	int r, i;

	__qu_unsafe__;

	QU_ARG_int (inc, i);
	errno = 0;
	r = nice (i);
	if (r == 0)
		r = getpriority (PRIO_PROCESS, 0);
	if (i == -1 && errno)
		QU_THROWERRNO;
	return qu_int_new (i);
}



/*) sub pathconf (path: String, Integer: name) -> Integer
	Get  a value for the configuration option v{name} for the file v{path}.

	v{name} is one of #{_PC}.

	See C{man 3 pathconf} for details.

	@exc
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		println (Os.pathconf ('/', Os._PC_PATH_MAX))

	@out
		4096
*/
static qu_integer Os_pathconf (qu_string path, qu_integer name)
{
	int  n;
	long v;

	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	QU_ARG_int  (name, n);

	errno = 0;
	v = pathconf (path->str, n);
	if (v == -1 && errno)
		QU_THROWERRNO;

	return qu_integer_newl (v);
}



/*) sub pathsep -> String
	Return the path separator.

	@noexc

	@xmp
		use Os
		println (Os.pathsep)

	@out
		/
*/
static qu_string Os_pathsep ()
{
	return QuFilePathSepStr;
}



/*) sub raise (sig: Integer) -> Null
	Send signal v{sig} to the current process.

	v{sig} is one of #{SIG}.

	See C{man 3 raise} for details.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.raise (Os.SIGTERM)
*/
static qu_null Os_raise (qu_integer sig)
{
	int s;

	__qu_unsafe__;

	QU_ARG_int (sig, s);
	errno = 0;
	if (raise (s) != 0)
	{
		if (errno)
			QU_THROWERRNO;
		QU_THROW (QuESystem);
	}
	return QuNil;
}



/*) sub setegid (Integer: uid) -> Null
	Set the effective group ID of the current process.

	See C{man 2 setegid} for details.

	@exc
	.EImplement f{setedig} is not available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setegid (123)
*/
static qu_null Os_setegid (qu_integer uid)
{
#ifdef HAVE_SETEGID
	long u;

	__qu_unsafe__;

	QU_ARG_long (uid, u);
	QU_ARG_CHECK (u >= 0);
	if (setegid ((uid_t) u) != 0)
		QU_THROWERRNO;
	return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub seteuid (Integer: euid) -> Null
	Set the effective user ID of the current process.

	See C{man 2 seteuid} for details.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.seteuid (123)
*/
static qu_null Os_seteuid (qu_integer uid)
{
	long u;

	__qu_unsafe__;

	QU_ARG_long (uid, u);
	QU_ARG_CHECK (u >= 0);
	if (seteuid ((uid_t) u) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub setgid (Integer: gid) -> Null
	Set the effective group ID of the current process.

	See C{man 2 setgid} for details.

	@exc
	.EImplement f{setgid} is not available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setgid (123)

	@warn
	This function is only available if your system has the f{setgid} function.
*/
static qu_null Os_setgid (qu_integer gid)
{
#ifdef HAVE_SETGID
	long g;

	__qu_unsafe__;

	QU_ARG_long (gid, g);
	QU_ARG_CHECK (g >= 0);
	if (setgid ((uid_t) g) != 0)
		QU_THROWERRNO;
	return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub setgroups (groups: Array) -> Null
	Set the supplementary group IDs for the process.

	Note that only the super-user may use this function.

	See C{man 2 setgroups} for details.

	v{groupd} must be an Array of valid group IDs.
	Its length must not exceed a certain limit, i.e. 64 by default.

	@exc
	.EImplement f{setgroups} is not available on your system
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		println (Os.getgroups)
	@out
		[500, 1]

	@warn
	This function is only available if your system has the f{setgroups} function.
*/
static qu_null Os_setgroups (qu_array x)
{
#ifdef HAVE_SETGROUPS
#  ifndef NGROUPS_MAX
#    define NGROUPS_MAX 64
#  endif
	gid_t   g [NGROUPS_MAX];
	long    i, l, v;
	qu_any *p;

	__qu_unsafe__;

	QU_ARG_TYPE (x, qu_array);
	l = x->len;
	QU_ARG_CHECK (l <= NGROUPS_MAX);
	p = x->objs;
	for (i = 0; i < l; ++i)
	{
		qu_any m = *p++;
		QU_ARG_long (m, v);
		QU_ARG_CHECK (v >= 0);
		g[i] = (gid_t) v;
	}
	if (setgroups ((size_t) l, g) != 0)
		QU_THROWERRNO;
    return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub setpriority (which: Int, who: Int, prio: Int) -> Int
	Set the scheduling priority of the process, process group, or user, as
	indicated by v{which} and v{who}

	v{which} is one of #{PRIO} and v{who} is interpreted relative to v{which}.

	See C{man 2 setpriority} for details.

	@exc
	.EImplement f{setpriority} is not available on your system
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		Os.setpriority (Os.PRIO_PROCESS, 0, 3)
*/
static qu_int Os_setpriority (qu_int which, qu_int who, qu_int prio)
{
#ifdef HAVE_SETPRIORITY
	int w, o, p;

	__qu_unsafe__;

	QU_ARG_int (which, w);
	QU_ARG_int (who, o);
	QU_ARG_int (prio, p);

	p = setpriority (w, o, p);
	if (p != 0)
		QU_THROWERRNO;
	return prio;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub setreuid (Integer: ruid, reuid: Integer) -> Null
	Set the real and effective user IDs of the current process.

	See C{man 2 setreuid} for details.

	@exc
	.EImplement f{setreuid} is not available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setreuid (123, 456)
*/
static qu_null Os_setreuid (qu_integer ruid, qu_integer reuid)
{
#ifdef HAVE_SETREUID
	long r, e;

	__qu_unsafe__;

	QU_ARG_long (ruid, r);
	QU_ARG_long (reuid, e);
	QU_ARG_CHECK (r >= 0);
	QU_ARG_CHECK (e >= 0);
	if (setreuid ((uid_t) r, (uid_t) e) != 0)
		QU_THROWERRNO;
	return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub setregid (Integer: rgid, regid: Integer) -> Null
	Set the real and effective group IDs of the current process.

	See C{man 2 setregid} for details.

	@exc
	.EImplement f{setregid} is not available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setregid (123, 456)
*/
static qu_null Os_setregid (qu_integer rgid, qu_integer regid)
{
#ifdef HAVE_SETREGID
	long r, e;

	__qu_unsafe__;

	QU_ARG_long (rgid, r);
	QU_ARG_long (regid, e);
	QU_ARG_CHECK (r >= 0);
	QU_ARG_CHECK (e >= 0);
	if (setregid ((uid_t) r, (uid_t) e) != 0)
		QU_THROWERRNO;
	return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub setrlimit (resource: Integer, max: Integer) -> Null
	Set the current maximum limit of the specified v{resource}.

	v{resource} must be one of #{RLIMIT}.

	See C{man 2 setrlimit} for details.

	@exc
	.EImplement f{setrlimit} is not available on your system
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setrlimit (Os.RLIMIT_NOFILE, Os.RLIM_INFINITY)
*/
static qu_null Os_setrlimit (qu_integer res, qu_integer max)
{
#ifdef HAVE_SETRLIMIT
	struct rlimit m;
	int r, x;

	__qu_unsafe__;

	QU_ARG_int (res, r);
	QU_ARG_int (max, x);
	QU_ARG_CHECK (x >= 0);
	m.rlim_cur = m.rlim_max = x;
	if (setrlimit (r, &m))
		QU_THROWERRNO;
	return QuNil;
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub setsid -> Integer
	Create a new  session if the calling process is not a process group leader.

	See C{man 2 setsid} for details.

	@exc
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setsid
*/
static qu_integer Os_setsid ()
{
	long u;

	__qu_unsafe__;
	u = (long) setsid ();
	if (u < 0)
		QU_THROWERRNO;
	return qu_integer_newl (u);
}



/*) sub setuid (Integer: uid) -> Null
	Set the effective user ID of the current process.

	See C{man 2 setuid} for details.

	@exc
	.EValue
	.ESecurity
	.ESystem

	@xmp
		use Os
		Os.setuid (123)
*/
static qu_null Os_setuid (qu_integer uid)
{
	long u;

	__qu_unsafe__;
	QU_ARG_long (uid, u);
	QU_ARG_CHECK (u >= 0);
	if (setuid ((uid_t) u) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub signal (signum: Integer, handler) -> Null
	Install or uninstall a signal handler.

	v{signum} is one of #{SIG}.

	If v{handler} is nil then the handler is uninstalled, otherwise it must
	be a static Sub that takes no argument to handle the signal.

	@exc
	.EValue
	.ESystem
	.ESecurity

	@warn
	You can only install handlers for the following signals since the others
	might interfere with Qu's internals: SIGCHLD SIGALRM SIGPIPE SIGUSR1 SIGUSR2

	The thread that catches the signal can be any thread.
	It is not necessarily the main thread.

	You must only do a minimal in the handler such as setting a flag.

	The handler is automatically disabled when it is called.
	You can call f{signal} in the handler to enable it again.

	@xmp
		use Os

		alarmed = 0

		Os.signal (Os.SIGALRM, @handler)
		Os.alarm (3)

		println (Os.__sig__)

		loop
			pass
		until alarmed

		println (alarmed)
		println (Os.__sig__)

		sub handler
			alarmed++
		;;

	@out
		[14:<Sub __main__.alarm_handler>]
		1
		[:]
*/
static qu_dict SigHandlers;

static void _os_sig_handle (int signum)
{
	qu_sub    u;
	int       b;
	qu_int    n;
	qu_thread t = qu_thread_self ();

	switch (signum)
	{
		case SIGCHLD: qu_sig_ack (CHLD); break;
		case SIGALRM: qu_sig_ack (ALRM); break;
		case SIGHUP:  qu_sig_ack (HUP ); break;
		case SIGPIPE: qu_sig_ack (PIPE); break;
		case SIGUSR1: qu_sig_ack (USR1); break;
		case SIGUSR2: qu_sig_ack (USR2); break;
	}

	assert (t);
	b = qu_thread_blocking (t);
	if (b)
		__qu_blocking_end;

	__qu_gc_critical_begin;

	n = qu_int_newi (signum);
	u = qu_dict_getitem (SigHandlers, n);

	qu_dict_unfreeze (SigHandlers);
	qu_dict_del      (SigHandlers, n);
	qu_dict_freeze   (SigHandlers);

	if (u)
		qu_sub_call0 (u, QuNil);

	__qu_gc_critical_end;

	if (b)
		__qu_blocking_begin;
}


static void _os_sig_CHLD () { _os_sig_handle (SIGCHLD); }
static void _os_sig_ALRM () { _os_sig_handle (SIGALRM); }
static void _os_sig_HUP  () { _os_sig_handle (SIGHUP);  }
static void _os_sig_PIPE () { _os_sig_handle (SIGPIPE); }
static void _os_sig_USR1 () { _os_sig_handle (SIGUSR1); }
static void _os_sig_USR2 () { _os_sig_handle (SIGUSR2); }


static qu_null Os_signal (qu_integer signum, qu_any handler)
{
	int s;

	__qu_unsafe__;

	QU_ARG_int (signum, s);
	if (handler == QuNil)
	{
		switch (s)
		{
			case SIGCHLD: __QuSigHookCHLD = NULL; break;
			case SIGALRM: __QuSigHookALRM = NULL; break;
			case SIGHUP:  __QuSigHookHUP  = NULL; break;
			case SIGPIPE: __QuSigHookPIPE = NULL; break;
			case SIGUSR1: __QuSigHookUSR1 = NULL; break;
			case SIGUSR2: __QuSigHookUSR2 = NULL; break;
			default:
				QU_THROW (QuEValue);
		}
		qu_dict_unfreeze (SigHandlers);
		qu_dict_del      (SigHandlers, qu_int_new (s));
		qu_dict_freeze   (SigHandlers);
	}
	else
	{
		QU_ARG_CHECK (qu_sub_is (handler) && qu_sub_is_static (handler) && ((qu_sub) handler)->argc == 0);
		switch (s)
		{
			case SIGCHLD: __QuSigHookCHLD = _os_sig_CHLD; break;
			case SIGALRM: __QuSigHookALRM = _os_sig_ALRM; break;
			case SIGHUP:  __QuSigHookHUP  = _os_sig_HUP;  break;
			case SIGPIPE: __QuSigHookPIPE = _os_sig_PIPE; break;
			case SIGUSR1: __QuSigHookUSR1 = _os_sig_USR1; break;
			case SIGUSR2: __QuSigHookUSR2 = _os_sig_USR2; break;
			default:
				QU_THROW (QuEValue);
		}
		qu_dict_unfreeze (SigHandlers);
		qu_dict_setitem  (SigHandlers, qu_int_new (s), handler);
		qu_dict_freeze   (SigHandlers);
	}
	return QuNil;
}



/*) sub sysconf (Integer: name) -> Integer
	Get system limits or options.

	v{name} is one of #{_SC}.

	See C{man 3 sysconf} for details.

	The value returned is the value of the system resource, 1 if a queried option
	is available, 0 if it is not, throw ESystem on error.

	@exc
	.EValue
	.ESystem
	.ESecutiry

	@xmp
		use Os
		println (Os.sysconf (Os._SC_ARG_MAX))

	@out
		131072
*/
static qu_integer Os_sysconf (qu_integer name)
{
	int  n;
	long v;

	__qu_unsafe__;

	QU_ARG_int (name, n);
	v = sysconf (n);
	if (v == -1)
		QU_THROWERRNO;
	return qu_integer_newl (v);
}



/*) sub system (cmd: String) -> Null
	Execute a command specified by v{cmd} and return after the command has been
	completed.

	See C{man 3 system} for details.

	@exc
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		Os.system ('echo "Hello World!"')
*/
static qu_null Os_system (qu_string cmd)
{
	int r;

	__qu_unsafe__;

	QU_ARG_TYPE (cmd, qu_string);
	__qu_blocking_begin;
	r = system (cmd->str);
	__qu_blocking_end;
	if (r != 0)
		QU_THROWERRNO;
	return QuNil;
}




/*) sub times -> Array
	Return the current process times.

	The result is an Array of 5 Integer seconds:
	@args
    .utime   user time
    .stime   system time
    .cutime  user time of children
    .cstime  system time of children
	.elapse  elapsed time since a point in the past (since boot time for Linux)

	@exc
	.ESystem

	@xmp
		use Os
		(u, s, c, m, e) = Os.times
		println (u)
		println (s)
		println (c)
		println (m)
		println (e)

	@out
		1000000
		0
		0
		0
		7379846000000
*/
static qu_array Os_times ()
{
	struct tms t;
	clock_t r;
	qu_int  f = qu_int_newl (CLOCKS_PER_SEC);

	r = times (&t);
	if (r < 0)
		QU_THROWERRNO;
	return qu_array_new_ex (5,
		qu_sys_mul (qu_integer_newl (t.tms_utime ), f),
		qu_sys_mul (qu_integer_newl (t.tms_stime ), f),
		qu_sys_mul (qu_integer_newl (t.tms_cutime), f),
		qu_sys_mul (qu_integer_newl (t.tms_cstime), f),
		qu_sys_mul (qu_integer_newl (r), f));
}




/*) sub umask (mask: Integer) -> Integer
	Sets the umask to mask & 0777 and return the previous mask.

	See C{man 2 umask} for details.

	@exc
	.EValue    v{mask} must be within 0 to 0777
	.ESecurity

	@xmp
		use Os
		Os.umask (0660)
*/
static qu_integer Os_umask (qu_integer mask)
{
	long m;

	__qu_unsafe__;

	QU_ARG_long (mask, m);
	QU_ARG_CHECK (0 <= m && m <= 0777);
	m = (long) umask ((mode_t) m);
	return qu_integer_newl (m);
}



/*) sub uname -> Array
	Return the operating system information.

	See C{man 2 uname} for details.

	The returned value is a frozen Array of 5 items:
	@ul
	.sysname  = the operating system name
	.nodename = the network node hostname
	.release  = the operating system release information
	.version  = the operating system version
	.machine  = the hardware type

	@noexc

	@xmp
		use Os
		(sysname, nodename, release, version, machine) = Os.uname
		println (sysname)
		println (nodename)
		println (release)
		println (version)
		println (machine)

	@out
		Linux
		localhost.localdomain
		#1 Thu Apr 18 07:37:53 EDT 2002
		2.4.18-3
		i686
*/
static qu_array Os_uname ()
{
	struct utsname u;
	uname (&u);
	return qu_array_new_ex (5,
		qu_string_newz (u.sysname),
		qu_string_newz (u.nodename),
		qu_string_newz (u.release),
		qu_string_newz (u.version),
		qu_string_newz (u.machine));
}



/*) sub utime (filename: String, atime: Real, mtime: Real) -> Null
	Change the access and modification times of the specified v{filename}
	to v{atime} and v{mtime} respectively.

	See C{man 2 utime} for details.

	@exc
	.EValue
	.ESystem
	.ESecurity

	@xmp
		use Os
		Os.utime ("/foo/bar.txt", 0, 0)
*/
static qu_null Os_utime (qu_string filename, qu_real atime, qu_real mtime)
{
	struct utimbuf u;
	double a, m;

	__qu_unsafe__;

	QU_ARG_TYPE (filename, qu_string);
	QU_ARG_double (atime, a);
	QU_ARG_double (mtime, m);

	u.actime  = (time_t) a;
	u.modtime = (time_t) m;
	if (utime (filename->str, &u) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub wait -> Array
	Suspend execution of the current process until a child has exited, or until
	a signal is delivered whose action is to terminate the current process or
	to call a signal handling function.

	See C{man 2 wait} for details.

	The returned value is an Array of two items:
	@args
	.pid    the child process ID
	.status the status

	@p
	Use #{WEXITSTATUS} #{WIFSIGNALED} #{WTERMSIG} #{WIFSTOPPED} #{WSTOPSIG} to
	evalue the status.

	@exc
	.ESecurity
	.ESystem
*/
static qu_array Os_wait ()
{
	pid_t p;
	int   status;

	__qu_unsafe__;

	p = wait (&status);
	if (p < 0)
		QU_THROWERRNO;
	return qu_array_new_ex (2, qu_integer_newl ((long) p), qu_integer_newi (status));
}



/*) sub waitpid (pid: Integer, options: Integer) -> Array
	Suspend execution of the current process until a child as specified by the
	v{pid} argument has exited, or until a signal is delivered whose action is
	to  terminate the current process or to call a signal handling function.

	v{options} is one of #{WAIT}.

	See C{man 2 waitpid} for details.

	The returned value is an Array of two items:
	@args
	.pid    the child process ID
	.status the status

	@p
	Use #{WEXITSTATUS} #{WIFSIGNALED} #{WTERMSIG} #{WIFSTOPPED} #{WSTOPSIG} to
	evalue the status.

	@exc
	.EValue
	.ESecurity
	.ESystem
*/
static qu_array Os_waitpid (qu_integer pid, qu_integer options)
{
	long p;
	int  op, status;

	__qu_unsafe__;

	QU_ARG_long (pid, p);
	QU_ARG_int  (options, op);
	p = waitpid (p, &status, op);
	if (p < 0)
		QU_THROWERRNO;
	return qu_array_new_ex (2, qu_integer_newl ((long) p), qu_integer_newi (status));
}



/*) sub WIFEXITED (status: Integer) -> Boolean
	Return true if the child exited normally, false otherwise.

	See #{wait}.

	@exc
	.EValue
*/
static qu_boolean Os_WIFEXITED (qu_integer status)
{
	long s;
	QU_ARG_long (status, s);
	return WIFEXITED (s) ? QuTrue : QuFalse;
}


/*) sub WEXITSTATUS (status: Integer) -> Integer
	Return the least significant eight bits of the return code of the
	child which terminated, which may have been set as the argument to a call
	to f{exit} or as the argument for a return statement in the main program.

	This function may only be called if #{WIFEXITED} returns true otherwise
	ECall is thrown.

	See #{wait}.

	@exc
	.EValue
	.ECall
*/
static qu_integer Os_WEXITSTATUS (qu_integer status)
{
	long s;
	QU_ARG_long (status, s);
	QU_REQUIRE (QuECall, WIFEXITED (status));
	s = WEXITSTATUS (s);
	return qu_integer_newl (s);
}



/*) sub WIFSIGNALED (status: Integer) -> Boolean
	Return true if the child process exited because of a signal which was
	not caught.

	See #{wait}.

	@exc
	.EValue
*/
static qu_boolean Os_WIFSIGNALED (qu_integer status)
{
	long s;
	QU_ARG_long (status, s);
	return WIFSIGNALED (s) ? QuTrue : QuFalse;
}



/*) sub WTERMSIG (status: Integer) -> Boolean
	Return true if the child process exited because of a signal which was
	not caught.

	This function may only be called if #{WIFSIGNALED} returns true otherwise
	ECall is thrown.

	See #{wait}.

	@exc
	.EValue
	.ECall
*/
static qu_integer Os_WTERMSIG (qu_integer status)
{
	long s;
	QU_ARG_long (status, s);
	QU_REQUIRE (QuECall, WIFSIGNALED (s));
	s = WTERMSIG (s);
	return qu_integer_newl (s);
}



/*) sub WIFSTOPPED (status: Integer) -> Boolean
	Return true if the child process which caused the return is currently
	stopped. This is only possible if the call was done using WUNTRACED.

	See #{wait}.

	@exc
	.EValue
	.ECall
*/
static qu_boolean Os_WIFSTOPPED (qu_integer status)
{
	long s;
	QU_ARG_long (status, s);
	return WIFSTOPPED (s) ? QuTrue : QuFalse;
}



/*) sub WSTOPSIG (status: Integer) -> Integer
	Return the number of the signal which caused the child to stop.

	This function may only be called if #{WIFSTOPPED} returns true otherwise
	ECall is thrown.

	See #{wait}.

	@exc
	.EValue
	.ECall
*/
static qu_integer Os_WSTOPSIG (qu_integer status)
{
	long s;
	QU_ARG_long (status, s);
	QU_REQUIRE (QuECall, WIFSTOPPED (s));
	s = WSTOPSIG (s);
	return qu_integer_newl (s);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Os_init (qu_module m)
{

#define e_(name) #name, name,
	qu_var_enum ((qu_symbol) m,  "E",
		e_(EPERM)
		e_(ENOENT)
		e_(ESRCH)
		e_(EINTR)
		e_(EIO)
		e_(ENXIO)
		e_(E2BIG)
		e_(ENOEXEC)
		e_(EBADF)
		e_(ECHILD)
		e_(EAGAIN)
		e_(ENOMEM)
		e_(EACCES)
		e_(EFAULT)
		e_(ENOTBLK)
		e_(EBUSY)
		e_(EEXIST)
		e_(EXDEV)
		e_(ENODEV)
		e_(ENOTDIR)
		e_(EISDIR)
		e_(EINVAL)
		e_(ENFILE)
		e_(EMFILE)
		e_(ENOTTY)
		e_(ETXTBSY)
		e_(EFBIG)
		e_(ENOSPC)
		e_(ESPIPE)
		e_(EROFS)
		e_(EMLINK)
		e_(EPIPE)
		e_(EDOM)
		e_(ERANGE)
		e_(EDEADLK)
		e_(ENAMETOOLONG)
		e_(ENOLCK)
		e_(ENOSYS)
		e_(ENOTEMPTY)
		e_(ELOOP)
		e_(EWOULDBLOCK)
		e_(ENOMSG)
		e_(EIDRM)
		e_(ECHRNG)
		e_(EL2NSYNC)
		e_(EL3HLT)
		e_(EL3RST)
		e_(ELNRNG)
		e_(EUNATCH)
		e_(ENOCSI)
		e_(EL2HLT)
		e_(EBADE)
		e_(EBADR)
		e_(EXFULL)
		e_(ENOANO)
		e_(EBADRQC)
		e_(EBADSLT)
		e_(EDEADLOCK)
		e_(EBFONT)
		e_(ENOSTR)
		e_(ENODATA)
		e_(ETIME)
		e_(ENOSR)
		e_(ENONET)
		e_(ENOPKG)
		e_(EREMOTE)
		e_(ENOLINK)
		e_(EADV)
		e_(ESRMNT)
		e_(ECOMM)
		e_(EPROTO)
		e_(EMULTIHOP)
		e_(EDOTDOT)
		e_(EBADMSG)
		e_(EOVERFLOW)
		e_(ENOTUNIQ)
		e_(EBADFD)
		e_(EREMCHG)
		e_(ELIBACC)
		e_(ELIBBAD)
		e_(ELIBSCN)
		e_(ELIBMAX)
		e_(ELIBEXEC)
		e_(EILSEQ)
		e_(ERESTART)
		e_(ESTRPIPE)
		e_(EUSERS)
		e_(ENOTSOCK)
		e_(EDESTADDRREQ)
		e_(EMSGSIZE)
		e_(EPROTOTYPE)
		e_(ENOPROTOOPT)
		e_(EPROTONOSUPPORT)
		e_(ESOCKTNOSUPPORT)
		e_(EOPNOTSUPP)
		e_(EPFNOSUPPORT)
		e_(EAFNOSUPPORT)
		e_(EADDRINUSE)
		e_(EADDRNOTAVAIL)
		e_(ENETDOWN)
		e_(ENETUNREACH)
		e_(ENETRESET)
		e_(ECONNABORTED)
		e_(ECONNRESET)
		e_(ENOBUFS)
		e_(EISCONN)
		e_(ENOTCONN)
		e_(ESHUTDOWN)
		e_(ETOOMANYREFS)
		e_(ETIMEDOUT)
		e_(ECONNREFUSED)
		e_(EHOSTDOWN)
		e_(EHOSTUNREACH)
		e_(EALREADY)
		e_(EINPROGRESS)
		e_(ESTALE)
		e_(EUCLEAN)
		e_(ENOTNAM)
		e_(ENAVAIL)
		e_(EISNAM)
		e_(EREMOTEIO)
		e_(EDQUOT)
		e_(ENOMEDIUM)
		e_(EMEDIUMTYPE)
		NULL);

	qu_var_enum ((qu_symbol) m,  "SIG",
		e_(SIGHUP)
		e_(SIGINT)
		e_(SIGQUIT)
		e_(SIGILL)
		e_(SIGABRT)
		e_(SIGFPE)
		e_(SIGKILL)
		e_(SIGSEGV)
		e_(SIGPIPE)
		e_(SIGALRM)
		e_(SIGTERM)
		e_(SIGUSR1)
		e_(SIGUSR2)
		e_(SIGCHLD)
		e_(SIGCONT)
		e_(SIGSTOP)
		e_(SIGTSTP)
		e_(SIGTTIN)
		e_(SIGTTOU)
#ifdef SIGBUS
		e_(SIGBUS)
#endif
#ifdef SIGPOLL
		e_(SIGPOLL)
#endif
#ifdef SIGPROF
		e_(SIGPROF)
#endif
#ifdef SIGSYS
		e_(SIGSYS)
#endif
#ifdef SIGTRAP
		e_(SIGTRAP)
#endif
#ifdef SIGURG
		e_(SIGURG)
#endif
#ifdef SIGVTALRM
		e_(SIGVTALRM)
#endif
#ifdef SIGXCPU
		e_(SIGXCPU)
#endif
#ifdef SIGXFSZ
		e_(SIGXFSZ)
#endif
#ifdef SIGIOT
		e_(SIGIOT)
#endif
#ifdef SIGEMT
		e_(SIGEMT)
#endif
#ifdef SIGSTKFLT
		e_(SIGSTKFLT)
#endif
#ifdef SIGIO
		e_(SIGIO)
#endif
#ifdef SIGCLD
		e_(SIGCLD)
#endif
#ifdef SIGPWR
		e_(SIGPWR)
#endif
#ifdef SIGINFO
		e_(SIGINFO)
#endif
#ifdef SIGLOST
		e_(SIGLOST)
#endif
#ifdef  SIGWINCH
		e_(SIGWINCH)
#endif
#ifdef SIGUNUSED
		e_(SIGUNUSED)
#endif
		NULL);

	qu_var_enum ((qu_symbol) m,  "F",
		e_(F_RDLCK)
		e_(F_WRLCK)
		e_(F_UNLCK)
		e_(F_DUPFD)
		e_(F_GETFD)
		e_(F_SETFD)
		e_(F_GETFL)
		e_(F_SETFL)
		e_(F_GETLK)
		e_(F_SETLK)
		e_(F_SETLKW)
		e_(F_GETLK64)
		e_(F_SETLK64)
		e_(F_SETLKW64)
		e_(F_SETOWN)
		e_(F_GETOWN)
#ifdef F_SETSIG
		e_(F_SETSIG)
#endif
#ifdef F_GETSIG
		e_(F_GETSIG)
#endif
#ifdef F_SETLEASE
		e_(F_SETLEASE)
#endif
#ifdef F_GETLEASE
		e_(F_GETLEASE)
#endif
#ifdef F_NOTIFY
		e_(F_NOTIFY)
#endif
		e_(F_LOCK)
		e_(F_TLOCK)
		e_(F_ULOCK)
		e_(F_TEST)
		NULL);

	qu_var_enum ((qu_symbol) m,  "O",
		e_(O_RDONLY)
		e_(O_WRONLY)
		e_(O_RDWR)
		e_(O_CREAT)
		e_(O_EXCL)
		e_(O_NOCTTY)
		e_(O_TRUNC)
		e_(O_APPEND)
		e_(O_NONBLOCK)
		e_(O_SYNC)
#ifdef O_NOFOLLOW
		e_(O_NOFOLLOW)
#endif
#ifdef O_DIRECTORY
		e_(O_DIRECTORY)
#endif
#ifdef O_LARGEFILE
		e_(O_LARGEFILE)
#endif
		NULL);

	qu_var_enum ((qu_symbol) m,  "SEEK",
		e_(SEEK_SET)
		e_(SEEK_CUR)
		e_(SEEK_END)
		NULL);

	qu_var_enum ((qu_symbol) m,  "S",
		e_(S_IFMT)
		e_(S_IFSOCK)
		e_(S_IFLNK)
		e_(S_IFREG)
		e_(S_IFBLK)
		e_(S_IFDIR)
		e_(S_IFCHR)
		e_(S_IFIFO)
		e_(S_ISUID)
		e_(S_ISGID)
		e_(S_ISVTX)
		e_(S_IRWXU)
		e_(S_IRUSR)
		e_(S_IWUSR)
		e_(S_IXUSR)
		e_(S_IRWXG)
		e_(S_IRGRP)
		e_(S_IWGRP)
		e_(S_IXGRP)
		e_(S_IRWXO)
		e_(S_IROTH)
		e_(S_IWOTH)
		e_(S_IXOTH)
		NULL);

	qu_var_enum ((qu_symbol) m,  "PRIO_",
		e_(PRIO_PROCESS)
		e_(PRIO_PGRP)
		e_(PRIO_USER)
		NULL);


	qu_var_enum ((qu_symbol) m,  "_PC_",
		e_(_PC_LINK_MAX)
		e_(_PC_MAX_CANON)
		e_(_PC_MAX_INPUT)
		e_(_PC_NAME_MAX)
		e_(_PC_PATH_MAX)
		e_(_PC_PIPE_BUF)
		e_(_PC_CHOWN_RESTRICTED)
		e_(_PC_NO_TRUNC)
		e_(_PC_VDISABLE)
		NULL);


	qu_var_enum ((qu_symbol) m,  "_SC_",
		e_(_SC_ARG_MAX)
		e_(_SC_CHILD_MAX)
		e_(_SC_CLK_TCK)
		e_(_SC_STREAM_MAX)
		e_(_SC_TZNAME_MAX)
		e_(_SC_OPEN_MAX)
		e_(_SC_JOB_CONTROL)
		e_(_SC_SAVED_IDS)
		e_(_SC_VERSION)
		e_(_SC_BC_BASE_MAX)
		e_(_SC_BC_DIM_MAX)
		e_(_SC_BC_SCALE_MAX)
		e_(_SC_BC_STRING_MAX)
		e_(_SC_COLL_WEIGHTS_MAX)
		e_(_SC_EXPR_NEST_MAX)
		e_(_SC_LINE_MAX)
		e_(_SC_RE_DUP_MAX)
		e_(_SC_2_VERSION)
#ifdef _SC_2_DEV
		e_(_SC_2_DEV)
#endif
#ifdef _SC_2_FORT_DEV
		e_(_SC_2_FORT_DEV)
#endif
#ifdef _SC_2_FORT_RUN
		e_(_SC_2_FORT_RUN)
#endif
#ifdef _SC_2_LOCALEDEF
		e_(_SC_2_LOCALEDEF)
#endif
#ifdef _SC_2_SW_DEV
		e_(_SC_2_SW_DEV)
#endif
#ifdef _SC_PAGE_SIZE
		e_(_SC_PAGE_SIZE)
#endif
#ifdef _SC_PHYS_PAGES
		e_(_SC_PHYS_PAGES)
#endif
#ifdef _SC_AVPHYS_PAGES
		e_(_SC_AVPHYS_PAGES)
#endif
		NULL);


	qu_var_enum ((qu_symbol) m,  "WAIT",
		e_(WNOHANG)
		e_(WUNTRACED)
#ifdef __WCLONE
		e_(__WCLONE)
#endif
#ifdef  __WALL
		e_(__WALL)
#endif
#ifdef __WNOTHREAD
		e_(__WNOTHREAD)
#endif
		NULL);


#ifdef HAVE_GETRLIMIT
	qu_var_enum ((qu_symbol) m,  "RLIMIT",
		e_(RLIMIT_CPU)
		e_(RLIMIT_FSIZE)
		e_(RLIMIT_DATA)
		e_(RLIMIT_STACK)
		e_(RLIMIT_RSS)
		e_(RLIMIT_NPROC)
		e_(RLIMIT_NOFILE)
		e_(RLIMIT_MEMLOCK)
		e_(RLIMIT_AS)
		NULL);
#endif


#ifdef HAVE_GETRUSAGE
	qu_var_enum ((qu_symbol) m,  "RUSAGE",
		e_(RUSAGE_SELF)
		e_(RUSAGE_CHILDREN)
		NULL);
#endif

#ifdef HAVE_GETRUSAGE
	UsageStruct = qu_struct_new_static (16,
		"utime"   , QuDecimalClass,
		"stime"   , QuDecimalClass,
		"maxrss"  , QuIntegerClass,
		"ixrss"   , QuIntegerClass,
		"idrss"   , QuIntegerClass,
		"isrss"   , QuIntegerClass,
		"minflt"  , QuIntegerClass,
		"majflt"  , QuIntegerClass,
		"nswap"   , QuIntegerClass,
		"inblock" , QuIntegerClass,
		"oublock" , QuIntegerClass,
		"msgsnd"  , QuIntegerClass,
		"msgrcv"  , QuIntegerClass,
		"nsignals", QuIntegerClass,
		"nvcsw"   , QuIntegerClass,
		"nivcsw"  , QuIntegerClass);
#endif


	SigHandlers = qu_dict_new (8);
	qu_dict_freeze (SigHandlers);
	qu_module_add_const (m, "__sig__", SigHandlers);

	qu_module_add_const (m, "CLOCKS_PER_SEC", qu_integer_newl (CLOCKS_PER_SEC));
	qu_module_add_const (m, "RLIM_INFINITY" , qu_integer_newl (RLIM_INFINITY));

	qu_module_add_function (m, "alarm"      , Os_alarm      ,  1);
	qu_module_add_function (m, "clock"      , Os_clock      ,  0);
	qu_module_add_function (m, "errno"      , Os_errno      , -1);
	qu_module_add_function (m, "execv"      , Os_execv      ,  2);
	qu_module_add_function (m, "execve"     , Os_execve     ,  3);
	qu_module_add_function (m, "execvp"     , Os_execvp     ,  2);
	qu_module_add_function (m, "fork"       , Os_fork       ,  0);
	qu_module_add_function (m, "getegid"    , Os_getegid    ,  0);
	qu_module_add_function (m, "geteuid"    , Os_geteuid    ,  0);
	qu_module_add_function (m, "getgid"     , Os_getgid     ,  0);
	qu_module_add_function (m, "getgroups"  , Os_getgroups  ,  0);
	qu_module_add_function (m, "getlogin"   , Os_getlogin   ,  0);
	qu_module_add_function (m, "getpgrp"    , Os_getpgrp    ,  0);
	qu_module_add_function (m, "getpgid"    , Os_getpgid    ,  1);
	qu_module_add_function (m, "getpid"     , Os_getpid     ,  0);
	qu_module_add_function (m, "getppid"    , Os_getppid    ,  0);
	qu_module_add_function (m, "getpriority", Os_getpriority,  2);
	qu_module_add_function (m, "getrlimit"  , Os_getrlimit  ,  1);
	qu_module_add_function (m, "getrusage"  , Os_getrusage  ,  1);
	qu_module_add_function (m, "getsid"     , Os_getsid     ,  1);
	qu_module_add_function (m, "getuid"     , Os_getuid     ,  0);
	qu_module_add_function (m, "initgroups" , Os_initgroups ,  2);
	qu_module_add_function (m, "kill"       , Os_kill       ,  2);
	qu_module_add_function (m, "killpg"     , Os_killpg     ,  2);
	qu_module_add_function (m, "nice"       , Os_nice       ,  1);
	qu_module_add_function (m, "pathconf"   , Os_pathconf   ,  2);
	qu_module_add_function (m, "pathsep"    , Os_pathsep    ,  0);
	qu_module_add_function (m, "raise"      , Os_raise      ,  1);
	qu_module_add_function (m, "setgid"     , Os_setgid     ,  1);
	qu_module_add_function (m, "setegid"    , Os_setegid    ,  1);
	qu_module_add_function (m, "seteuid"    , Os_seteuid    ,  1);
	qu_module_add_function (m, "setgroups"  , Os_setgroups  ,  1);
	qu_module_add_function (m, "setpriority", Os_setpriority,  3);
	qu_module_add_function (m, "setreuid"   , Os_setreuid   ,  2);
	qu_module_add_function (m, "setregid"   , Os_setregid   ,  2);
	qu_module_add_function (m, "setrlimit"  , Os_setrlimit  ,  3);
	qu_module_add_function (m, "setsid"     , Os_setsid     ,  0);
	qu_module_add_function (m, "setuid"     , Os_setuid     ,  1);
	qu_module_add_function (m, "signal"     , Os_signal     ,  2);
	qu_module_add_function (m, "sysconf"    , Os_sysconf    ,  1);
	qu_module_add_function (m, "system"     , Os_system     ,  1);
	qu_module_add_function (m, "times"      , Os_times      ,  0);
	qu_module_add_function (m, "umask"      , Os_umask      ,  1);
	qu_module_add_function (m, "uname"      , Os_uname      ,  0);
	qu_module_add_function (m, "utime"      , Os_utime      ,  3);
	qu_module_add_function (m, "wait"       , Os_wait       ,  0);
	qu_module_add_function (m, "waitpid"    , Os_waitpid    ,  2);

	qu_module_add_function (m, "WIFEXITED"  , Os_WIFEXITED  ,  1);
	qu_module_add_function (m, "WEXITSTATUS", Os_WEXITSTATUS,  1);
	qu_module_add_function (m, "WIFSIGNALED", Os_WIFSIGNALED,  1);
	qu_module_add_function (m, "WTERMSIG"   , Os_WTERMSIG   ,  1);
	qu_module_add_function (m, "WIFSTOPPED" , Os_WIFSTOPPED ,  1);
	qu_module_add_function (m, "WSTOPSIG"   , Os_WSTOPSIG   ,  1);

	return 1;
}
