/* Qu.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Qu
	This module provides various constants used by Qu, access to C standard library
	functions, and predefined static Pointer objects.

	@xmp
		use Qu
		__extern Qu.printf "-:ssi" ("%s and %d\n", "foo", 123)

	@out
		foo and 123

	@p
	This module also defines a number of preallocated static Pointer objects you can
	use to pass to C function. The intention is performance and efficiency. That is,
	you don't need to allocate new Pointer objects everytime.

	@warn
	You must wrap the code using the Pointers in an k{atomic} block otherwise other
	threads which happen to use them at the same time will screw the result.

	You must never f{acquire} these pointers.

	@xmp
		use Qu
		use Foo

		sub bar
			atomic
				__extern Foo.get_two_ints "-:pp" (Qu.INT1, Qu.INT2)
				return [Qu.INT1.get, Qu.INT2.get]
			;;
		;;

	@p
	Here is the list of available objects (5 objects each):
	@ul
	.CHAR1       .. CHAR5
	.BYTE1       .. BYTE5
	.UBYTE1      .. UBYTE5
	.SHORT1      .. SHORT5
	.USHORT1     .. USHORT5
	.INT1        .. INT5
	.UINT1       .. UINT5
	.INT161      .. INT165
	.UINT161     .. UINT165
	.INT321      .. INT325
	.UINT321     .. UINT325
	.LONG1       .. LONG5
	.ULONG1      .. ULONG5
	.LONGLONG1   .. LONGLONG5
	.ULONGLONG1  .. ULONGLONG5
	.FLOAT1      .. FLOAT5
	.DOUBLE1     .. DOUBLE5
	.LONGDOUBLE1 .. LONGDOUBLE5
	.SSIZE1      .. SSIZE5
	.SIZE1       .. SIZE5
	.OFF1        .. OFF5
	.BOOL1       .. BOOL5
	.PTR1        .. PTR5
	.PPTR1       .. PPTR5
	.ANY1        .. ANY5 (any Qu object, use with care)
*/


/* the DL handle */
static void *Handle = NULL;

/* list of all resolved Methods */
static qu_dict Resolved;

/* get the C function */
static qu_any qu_Qu_get (qu_module self, qu_string name, qu_string spec)
{
	return qu_module_get_cfunc (self, Handle, name, spec, Resolved);
}



/*) const NAME = 'qu'
	This is the Qu application name.
*/


/*) const VERSION = '1.09.05'
	This is the installed Qu version.
*/


/*) const PATH_SEP = '/'
	The file system path separator.
*/


/*) const CC = 'gcc'
	The C compiler used to compile Qu.
*/


/*) const CLDFLAGS = '-rdynamic'
	The flags passed to the C compiler when building the Qu application.
*/


/*) const CFLAGS = '-Wall'
	The general flags to pass to the C compiler.
*/


/*) const CFLAGSMOD = '-fPIC'
	Flags to pass to the C compiler to build object file.o.
*/


/*) const CFLAGSSHARED = '-shared'
	Flags to supply to the C linker to build shared library file.so.
*/


/*) const CLIBSTD = '-lm -ld -lpthread'
	The standard libraries to link to.
*/


/*) const EXT_MOD = '.qm'
	Qu module extension, with the dot.
*/


/*) const EXT_DL = '.so'
	Dynamic module extension, with the dot.
*/


/*) const SIZEOF_CHAR = 1
	The size of the C C{char}.
*/


/*) const SIZEOF_SHORT = 2
	The size of the C C{short}.
*/


/*) const SIZEOF_INT = 4
	The size of the C C{int}.
*/
int sizeof_int;


/*) const SIZEOF_LONG = 4
	The size of the C C{long}.
*/


/*) const SIZEOF_LONG_LONG = 8
	The size of the C C{long long}.

	@warn
	This is only available if the C C{long long} type is supported.
*/


/*) const SIZEOF_LONG_DOUBLE = 12
	The size of the C C{long double}.

	@warn
	This is only available if the C C{long double} type is supported.
*/


/*) const SIZEOF_DOUBLE = 8
	The size of the C C{double}.
*/


/*) const SIZEOF_VOID_P = 4
	The size of the C C{void *}.
*/


/*) const SIZEOF_POINTER = 4
	An alias of SIZEOF_VOID_P
*/


/*) const SIZEOF_OFF_T = 4
	The size of the C C{off_t}.
*/


/*) const SIZEOF_SIZE_T = 4
	The size of the C C{size_t}.
*/


/*) const SIZEOF_PTHREAD_T = 4
	The size of the C C{pthread_t}.
*/


/*) const DIR_TMP = '/tmp/'
	The absolute path of the temporary directory,
	slash terminated.
*/


/*) const DIR_EXE = '/usr/bin/'
	The absolute path of the executable directory,
	slash terminated.
*/


/*) const DIR_DEST = '/opt/qu/'
	The absolute path of the directory where Qu is installed,
	slash terminated.
*/


/*) const DIR_BIN = '/opt/qu/bin/'
	The absolute path of the directory where Qu binaries are installed,
	slash terminated.
*/


/*) const DIR_DEMO = '/opt/qu/demo/'
	The absolute path of Qu's C{demo} directory,
	slash terminated.

	This is where Qu demo files are installed.
*/


/*) const DIR_REMOVE = '/opt/qu/remove/'
	The absolute path of Qu's C{remove} directory,
	slash terminated.

	This is where Qu package installer programs are saved.
	They are called to uninstall packages.
*/


/*) const DIR_LIB = '/opt/qu/lib/'
	The absolute path of Qu's C{lib} directory,
	slash terminated.

	This is where Qu modules are installed.
*/


/*) const DIR_DOC = '/opt/qu/doc'
	The absolute path of Qu's C{doc} directory,
	slash terminated.

	This is where documentation files are installed.
*/


/*) const DIR_DOC_LIB = '/opt/qu/doc/lib'
	The absolute path of the library documentation directory,
	slash terminated.

	This is where Qu module documentation files are installed.
*/


/*) const DIR_DOC_API = '/opt/qu/doc/api'
	The absolute path of the API documentation directory,
	slash terminated.

	This is where API documentation files are installed.
*/


/*) const DIR_DOC_LANG = '/opt/qu/doc/lang'
	The absolute path of the language manual documentation directory,
	slash terminated.
*/


/*) const DIR_DOC_OTHERS = '/opt/qu/doc/others'
	The absolute path of other references documentation directory,
	slash terminated.
*/


/*) const DIR_DOC_TUTORIAL = '/opt/qu/doc/tutorial'
	The absolute path of the tutorial documentation directory,
	slash terminated.
*/


/*) const DIR_DOC_DEVEL = '/opt/qu/doc/devel'
	The absolute path of the developers documentation directory,
	slash terminated.
*/


/*) const OBJ_HEADER_SIZE = 16
	The header size of Qu objects.
*/




/* used to test the qu_proc class */
typedef char *(*foo_t) (int a, float b, char *c, double d);
static foo_t foo;
void QuSetIt (foo_t f)
{
	foo = f;
}
void QuCallIt ()
{
	char *r;
	printf ("C calling Qu function\n");
	r = foo (123, 0.1, "Hi there!", 4.56);
	printf ("C got: %s\n", r);
}

float QuTestArg1 (short a, int b, long c, char d, float e, double f, char *g, char h, int i)
{
	printf ("a = %hd\n", a);
	printf ("b = %d\n" , b);
	printf ("c = %ld\n", c);
	printf ("d = %c\n" , d);
	printf ("e = %g\n" , e);
	printf ("f = %f\n" , f);
	printf ("g = %s\n" , g);
	printf ("h = %c\n" , h);
	printf ("i = %d\n" , i);
	return 1.23;
}

#ifdef HAVE_LONG_LONG
long long QuTestArg2 (short a, int b, long long c, char d, float e, double f, char *g, char h, int i)
{
	printf ("a = %hd\n", a);
	printf ("b = %d\n" , b);
	printf ("c = %Ld\n", c);
	printf ("d = %c\n" , d);
	printf ("e = %g\n" , e);
	printf ("f = %g\n" , f);
	printf ("g = %s\n" , g);
	printf ("h = %c\n" , h);
	printf ("i = %d\n" , i);
	return 12345678901234567LL;
}
#endif

#ifdef HAVE_LONG_DOUBLE
long double QuTestArg3 (short a, int b, long long c, char d, float e, long double f, char *g, char h, int i)
{
	printf ("a = %hd\n", a);
	printf ("b = %d\n" , b);
	printf ("c = %Ld\n", c);
	printf ("d = %c\n" , d);
	printf ("e = %g\n" , e);
	printf ("f = %Lg\n" , f);
	printf ("g = %s\n" , g);
	printf ("h = %c\n" , h);
	printf ("i = %d\n" , i);
	return 1.23E400L;
}
#endif



/*----------------------------------------------------------------------------
	functions
----------------------------------------------------------------------------*/

/*) sub call (*arg) -> Any
	Call a C function.

	Use this to call functions with variable arguments.

	@xmp
		use Qu

		sub printstrings (*strings)
			n = strings.len
			format = "%s " * n
			args   = "-:s" _ "s" * n
			Qu.call ("printf", args, format, *strings)
		;;

		printstrings ('hello', 'world')

	@out
		hello world
*/
static qu_any qu_Qu_call (int argc, qu_any *argv)
{
	return qu_dlib_call (Handle, argc, argv);
}



/*) sub PPTRxget
	@def
	sub PPTR1get -> Pointer # returns PPTR1
	sub PPTR2get -> Pointer # returns PPTR2
	sub PPTR3get -> Pointer # returns PPTR3
	sub PPTR4get -> Pointer # returns PPTR4
	sub PPTR5get -> Pointer # returns PPTR5
	@p
	Set a PPTRx Pointer to contain the pointer contained in a PTRx pointer.
	This is useful if a C function requires a location to store a pointer and is more
	efficient than calling the pointer's f{get} method (which creates a new Pointer).

	@xmp
		# C function: void foo (void **p)

		atomic
			somehow_call_foo (Qu.PTR1)
			Qu.PPTR1get
				# now Qu.PPTR1 contains the pointer contained in Qu.PTR1
		;;
*/
static qu_pointer PTR1;
static qu_pointer PTR2;
static qu_pointer PTR3;
static qu_pointer PTR4;
static qu_pointer PTR5;
static qu_pointer PPTR1;
static qu_pointer PPTR2;
static qu_pointer PPTR3;
static qu_pointer PPTR4;
static qu_pointer PPTR5;

#define f_(xx) \
static qu_pointer qu_Qu_P##xx##get () \
{ \
	void *p = xx->ptr; \
	P##xx->ptr = p ? *((void **) p) : NULL; \
	return P##xx; \
}

f_(PTR1)
f_(PTR2)
f_(PTR3)
f_(PTR4)
f_(PTR5)

#undef f_


/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/
static void qu_Qu_atexit ()
{
	if (Handle)
		qu_dl_close (Handle);
}


int qu_Qu_init (qu_module m)
{
	Handle = qu_dl_open_ex ("libm", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (Handle);

	/* set the resolver */
	m->getcfunc = qu_Qu_get;

	/* strings are static, values are static Methods */
	Resolved = qu_dict_new_static (512);

	qu_module_add_function (m, "call", qu_Qu_call, -1);


#define i_(namez, val)   qu_module_add_int (m, namez, val)
#define s_(namez, strz)  qu_module_add_const (m, namez, qu_string_new_staticz (strz))
#define d_(namez, fmt)   qu_module_add_const (m, namez, qu_string_new_staticf ("%s" fmt QU_PATH_SEPSTR, QU_APP_HOMEDIR))

	/* package */
	s_("NAME"             , QU_APP_PACKAGE);
	s_("VERSION"          , QU_APP_VERSION);

	/* system stuffs */
	s_("PATH_SEP"         , QU_PATH_SEPSTR);

	/* compiler stuffs */
	s_("CC"               , COMPILE_CC);
	s_("CLDFLAGS"         , COMPILE_LDFLAGS);
	s_("CFLAGS"           , COMPILE_CFLAGS);
	s_("CFLAGSMOD"        , COMPILE_CFLAGSMOD);
	s_("CFLAGSSHARED"     , COMPILE_CFLAGSSHARED);
	s_("CLIBSTD"          , COMPILE_LIBSTD);
	i_("SIZEOF_CHAR"      , SIZEOF_CHAR);
	i_("SIZEOF_SHORT"     , SIZEOF_SHORT);
	i_("SIZEOF_INT"       , SIZEOF_INT);
	i_("SIZEOF_LONG"      , SIZEOF_LONG);
	i_("SIZEOF_DOUBLE"    , SIZEOF_DOUBLE);
	i_("SIZEOF_VOID_P"    , SIZEOF_VOID_P);
	i_("SIZEOF_POINTER"   , SIZEOF_VOID_P);
	i_("SIZEOF_OFF_T"     , SIZEOF_OFF_T);
	i_("SIZEOF_SIZE_T"    , SIZEOF_SIZE_T);
#ifdef HAVE_LONG_LONG
	i_("SIZEOF_LONG_LONG" , SIZEOF_LONG_LONG);
#endif
#ifdef HAVE_LONG_DOUBLE
	i_("SIZEOF_LONG_DOUBLE", SIZEOF_LONG_DOUBLE);
#endif
	i_("SIZEOF_PTHREAD_T"  , SIZEOF_PTHREAD_T);

	/* file extensions */
	s_("EXT_MOD"          , QU_MOD_EXT);
	s_("EXT_DL"           , QU_MOD_DLEXT);

	/* directories */
	s_("DIR_TMP"          , QU_APP_TMPDIR QU_PATH_SEPSTR);
	s_("DIR_EXE"          , QU_APP_BINDIR QU_PATH_SEPSTR);

	d_("DIR_DEST"         , "" );
	d_("DIR_BIN"          , QU_PATH_SEPSTR "bin");
	d_("DIR_DEMO"         , QU_PATH_SEPSTR "demo");
	d_("DIR_LIB"          , QU_PATH_SEPSTR "lib");
	d_("DIR_REMOVE"       , QU_PATH_SEPSTR "remove");
	d_("DIR_DOC"          , QU_PATH_SEPSTR "doc");
	d_("DIR_DOC_LIB"      , QU_PATH_SEPSTR "doc" QU_PATH_SEPSTR "lib");
	d_("DIR_DOC_API"      , QU_PATH_SEPSTR "doc" QU_PATH_SEPSTR "api");
	d_("DIR_DOC_LANG"     , QU_PATH_SEPSTR "doc" QU_PATH_SEPSTR "lang");
	d_("DIR_DOC_OTHERS"   , QU_PATH_SEPSTR "doc" QU_PATH_SEPSTR "others");
	d_("DIR_DOC_TUTORIAL" , QU_PATH_SEPSTR "doc" QU_PATH_SEPSTR "tutorial");
	d_("DIR_DOC_DEVEL"    , QU_PATH_SEPSTR "doc" QU_PATH_SEPSTR "devel");

	/* miscellaneous */
	i_("OBJ_HEADER_SIZE"  , sizeof (qu_object_s));

#undef i_
#undef s_
#undef d_


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Pointers
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (char), 'a'));
	p_("CHAR1")
	p_("CHAR2")
	p_("CHAR3")
	p_("CHAR4")
	p_("CHAR5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (1, 'y'));
	p_("BYTE1")
	p_("BYTE2")
	p_("BYTE3")
	p_("BYTE4")
	p_("BYTE5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (1, 'Y'));
	p_("UBYTE1")
	p_("UBYTE2")
	p_("UBYTE3")
	p_("UBYTE4")
	p_("UBYTE5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (short), 'h'));
	p_("SHORT1")
	p_("SHORT2")
	p_("SHORT3")
	p_("SHORT4")
	p_("SHORT6")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (short), 'H'));
	p_("USHORT1")
	p_("USHORT2")
	p_("USHORT3")
	p_("USHORT4")
	p_("USHORT5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (int), 'i'));
	p_("INT1")
	p_("INT2")
	p_("INT3")
	p_("INT4")
	p_("INT5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (int), 'I'));
	p_("UINT1")
	p_("UINT2")
	p_("UINT3")
	p_("UINT4")
	p_("UINT5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (2, 'j'));
	p_("INT161")
	p_("INT162")
	p_("INT163")
	p_("INT164")
	p_("INT165")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (2, 'J'));
	p_("UINT161")
	p_("UINT162")
	p_("UINT163")
	p_("UINT164")
	p_("UINT165")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (4, 'k'));
	p_("INT321")
	p_("INT322")
	p_("INT323")
	p_("INT324")
	p_("INT325")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (4, 'K'));
	p_("UINT321")
	p_("UINT322")
	p_("UINT323")
	p_("UINT324")
	p_("UINT325")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (long), 'l'));
	p_("LONG1")
	p_("LONG2")
	p_("LONG3")
	p_("LONG4")
	p_("LONG5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (long), 'L'));
	p_("ULONG1")
	p_("ULONG2")
	p_("ULONG3")
	p_("ULONG4")
	p_("ULONG5")
#undef p_

#ifdef HAVE_LONG_LONG
#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (long long), 'o'));
	p_("LONGLONG1")
	p_("LONGLONG2")
	p_("LONGLONG3")
	p_("LONGLONG4")
	p_("LONGLONG5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (long long), 'O'));
	p_("ULONGLONG1")
	p_("ULONGLONG2")
	p_("ULONGLONG3")
	p_("ULONGLONG4")
	p_("ULONGLONG5")
#undef p_
#endif

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (float), 'g'));
	p_("FLOAT1")
	p_("FLOAT2")
	p_("FLOAT3")
	p_("FLOAT4")
	p_("FLOAT5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (double), 'f'));
	p_("DOUBLE1")
	p_("DOUBLE2")
	p_("DOUBLE3")
	p_("DOUBLE4")
	p_("DOUBLE5")
#undef p_

#ifdef HAVE_LONG_DOUBLE
#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (long double), 'F'));
	p_("LONGDOUBLE1")
	p_("LONGDOUBLE2")
	p_("LONGDOUBLE3")
	p_("LONGDOUBLE4")
	p_("LONGDOUBLE5")
#undef p_
#endif

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (ssize_t), 'z'));
	p_("SSIZE1")
	p_("SSIZE2")
	p_("SSIZE3")
	p_("SSIZE4")
	p_("SSIZE5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (size_t), 'Z'));
	p_("SIZE1")
	p_("SIZE2")
	p_("SIZE3")
	p_("SIZE4")
	p_("SIZE5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (off_t), 'e'));
	p_("OFF1")
	p_("OFF2")
	p_("OFF3")
	p_("OFF4")
	p_("OFF5")
#undef p_

#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (int), 'b'));
	p_("BOOL1")
	p_("BOOL2")
	p_("BOOL3")
	p_("BOOL4")
	p_("BOOL5")
#undef p_

#define p_(x) \
	x = qu_pointer_predefined (sizeof (void *), 'p'); \
	qu_module_add_const (m, #x, x);
	p_(PTR1)
	p_(PTR2)
	p_(PTR3)
	p_(PTR4)
	p_(PTR5)
#undef p_

#define p_(x) \
	x = qu_pointer_new_static (NULL); \
	qu_module_add_const (m, #x, x);
	p_(PPTR1)
	p_(PPTR2)
	p_(PPTR3)
	p_(PPTR4)
	p_(PPTR5)
#undef p_

	qu_module_add_function (m, "PPTR1get", qu_Qu_PPTR1get, 0);
	qu_module_add_function (m, "PPTR2get", qu_Qu_PPTR2get, 0);
	qu_module_add_function (m, "PPTR3get", qu_Qu_PPTR3get, 0);
	qu_module_add_function (m, "PPTR4get", qu_Qu_PPTR4get, 0);
	qu_module_add_function (m, "PPTR5get", qu_Qu_PPTR5get, 0);


#define p_(z) qu_module_add_const (m, z, qu_pointer_predefined (sizeof (void *), 'x'));
	p_("ANY1")
	p_("ANY2")
	p_("ANY3")
	p_("ANY4")
	p_("ANY5")
#undef p_

	atexit (qu_Qu_atexit);
	return 1;
}
