/* Readline.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Readline
	This module provides a simple interface to the GNU I{readline} library.

	Note that you don't need to recompile Qu nor this module.
	Once you have installed the GNU I{readline} library then it will be automatically
	available.

	Check the C{qu.conf} in the directory where Qu is installed and adjust
	the v{readline} and v{curses} entries as necessary.

	@xmp
		use Readline
		Readline.history = true
		s = Readline.input ('prompt> ')
*/


/*---------------------------------------------------------------------------
	stuffs we need from the readline library
---------------------------------------------------------------------------*/
static int   (*rl_initialize          ) (void);
static void  (*rl_free_line_state     ) (void);
static void  (*rl_cleanup_after_signal) (void);
static char *(*readline               ) (const char *);
static void  (*add_history            ) (const char *);


/*---------------------------------------------------------------------------
	our own stuffs
---------------------------------------------------------------------------*/
/* just to catch SIGINT :-( */
static jmp_buf SigIntJmpBuf;


static int UseHistory = 1;

static void _readline_sigint_hook ()
{
	rl_free_line_state ();
	rl_cleanup_after_signal ();
	longjmp (SigIntJmpBuf, 1);
}


/*---------------------------------------------------------------------------
	functions
---------------------------------------------------------------------------*/
/*) sub input (*prompt) -> String
	Get input from terminal.

	v{prompt} items are concatenated as strings to produce the prompt.

	@warn
		The program is blocked while in this function.

	@exc
	.EBADF       -{File.stdin} not opened
	.EInterrupt  user pressed C{Ctrl-c} or C{Ctrl-d}

	@xmp
		use Readline
		Readline.history = true
		s = Readline.input ('prompt> ')
*/
static qu_string Readline_input (int argc, qu_any *argv)
{
	char      *p = NULL, *z;
	qu_string  s;

	if (argc)
	{
		s = qu_sys_cat (argc, argv);
		p = s->str;
	}

	/* !!! do not wrap in __qu_blocking_begin/end */

	__qu_sig_begin (INT, _readline_sigint_hook);
	{
		if (setjmp (SigIntJmpBuf) == 0)
		{
			z = readline (p);
			if (add_history && UseHistory && z && strlen (z))
				add_history (z);
		}
		else
		{
			z = NULL;
		}
	}
	__qu_sig_end (INT);

	if (qu_sig_got (INT) || z == NULL)
	{
		qu_sig_clear (INT);
		QU_THROW (QuEInterrupt);
	}

	s = qu_string_newz (z);
	free (z);
	return s;
}



/*) sub history (*arg)
	@def
	sub history (setting: Boolean) -> Boolean
	sub history -> Boolean
	@p
	Set or get whether or not to record the input to the history facility.

	If setting then this function return the old setting.

	@exc
	.EArgValue
*/
static qu_boolean Readline_history (int argc, qu_any *argv)
{
	if (add_history)
	{
		QU_ARG_VCOUNT (0, 1);
		if (argc)
		{
			int old = UseHistory;
			int setting;

			QU_ARG_vbool (0, setting);
			UseHistory = setting;
			return old ? QuTrue : QuFalse;
		}
		else
		{
			return UseHistory ? QuTrue : QuFalse;
		}
	}
	else
	{
		QU_THROWZ (QuEImplement, "history not supported");
	}
}



/*----------------------------------------------------------------------------
	DL stuffs
----------------------------------------------------------------------------*/
static void *QuCursesHandle;
static void *QuReadlineHandle;

static void qu_readline_atexit ()
{
	if (QuCursesHandle)
		qu_dl_close (QuCursesHandle);

	if (QuReadlineHandle)
		qu_dl_close (QuReadlineHandle);
}

static int qu_readline_failed (const char *z)
{
	qu_readline_atexit ();
	QU_THROWF (QuESystem, "cant initialize Readline missing symbol %s", z);
}


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/
int qu_Readline_init (qu_module m)
{
	/* in this order */
	QuCursesHandle   = qu_dl_open_ex ("curses"  , QU_DL_GLOBAL | QU_DL_NOW);
	QuReadlineHandle = qu_dl_open_ex ("readline", QU_DL_GLOBAL | QU_DL_NOW);

	if (QuCursesHandle == NULL || QuReadlineHandle == NULL)
	{
		qu_readline_atexit ();
		QU_RETHROW;
	}

#define f_(name) \
	name = qu_dl_sym (QuReadlineHandle, #name); \
	if (name == 0) \
		return qu_readline_failed ( #name);

	f_(rl_initialize          )
	f_(rl_free_line_state     )
	f_(rl_cleanup_after_signal)
	f_(readline               )
	f_(add_history            )

	/*rl_readline_name*/
	{
		char **v = (char **) qu_dl_sym (QuReadlineHandle, "rl_readline_name");
		if (v)
			*v = "qu";
	}

	/*rl_catch_signals*/
	{
		int *v = (int *) qu_dl_sym (QuReadlineHandle, "rl_catch_signals");
		if (v)
			*v = 0;
	}

	qu_module_add_function (m, "input", Readline_input, -1);

	if (add_history)
		qu_module_add_function (m, "history", Readline_history, -1);

	return 1;
}
