/* Regex.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_REGEX_H
#include <regex.h>


/*) module Regex
	This module provides an interface to the system POSIX regular expression
	library.

	See the ${Regex} class and C{man 7 regex} for details.

	It works with String and Buffer objects.

	@xmp
		use Regex
		use Buffer

		r = Regex ("[ABC]", 'i')
		s = 'aaa bbb ccc'
		println (r.match (s))

		b = Buffer (s)
		println (r.match (b))

	@out
		9
		9
*/


/*----------------------------------------------------------------------------
	class Regex
----------------------------------------------------------------------------*/

/*) class Regex __final
	This class provides methods to deal with regular expressions.

	See C{man 7 regex} for details.

	@xmp
		use Regex

		r = Regex ("[abc]")
		println (r.match ('aaa bbb ccc'))

	@out
		9
*/
static qu_class RegexClass;


#define MAXSUBS 512

typedef struct TRegex TRegex;
struct TRegex
{
	QU_OBJECT_HEAD;
	regex_t re;
};

typedef TRegex *Regex;

#define RegexIs(o) (qu_classof (o) == RegexClass)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	internals
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static int _regex_option (qu_string x)
{
	int   op = 0;
	long  l = x->len;
	char *s = x->str;

	while (l--)
	{
		switch (*s++)
		{
			case 'i': case 'I': op |= REG_ICASE;   break;
			case 'u': case 'U': op |= REG_NOSUB;   break;
			case 'l': case 'L': op |= REG_NEWLINE; break;
			case 'o': case 'O': op |= REG_NOTBOL;  break;
			case 'e': case 'E': op |= REG_NOTEOL;  break;
		}
	}
	return op;
}



static qu_string _regex_error (int errcode, const regex_t *preg)
{
	long      l = regerror (errcode, preg, NULL, 0);
	qu_string s;

	s = qu_string_newl (l);
	regerror (errcode, preg, s->str, l);
	return s;
}



static qu_stringy _regex_do_subst (Regex self, qu_stringy s, qu_stringy u, long max, int option)
{
	regex_t    *r = &self->re;
	char       *i = s->str;
	char       *k = i + s->len;
	qu_buffer   b = qu_buffer_alloc (s->len);
	char       *last = 0;
	regmatch_t  m [2];

	while (max-- > 0
		&& i < k
		&& i != last
		&& regexec (r, i, 1, m, option) != REG_NOMATCH)
	{
		last = i;
		if (m[0].rm_so > 0)
			QU_ENSURE (qu_buffer_put (b, i, m[0].rm_so));
		QU_ENSURE (qu_buffer_puts (b, u));
		i += m[0].rm_eo;
	}

	if (i < k)
		QU_ENSURE (qu_buffer_put (b, i, k - i));

	if (qu_buffer_is (s))
	{
		qu_buffer_exchange_tmp ((qu_buffer) s, b);
		return s;
	}
	return qu_buffer_tmp_to_string (b);
}



static qu_stringy _regex_do_substex (Regex self, qu_stringy s, qu_function f, long max, int option)
{
	qu_buffer  buf;
	qu_array   u;
	regex_t   *r = &self->re;
	char      *i = s->str;
	char      *k = i + s->len;
	char      *last = NULL;
	long       j, n;
	off_t      o;
	qu_any     theself = 0;
	qu_stringy x;
	regmatch_t m [MAXSUBS];
	int        issub = qu_sub_is (f);
	int        isbuf = qu_buffer_is (s);
	qu_any     scope = 0;

	if (isbuf)
		QU_BUFFER_PROTECT_BEGIN ((qu_buffer) s, scope);

	if (issub)
		theself = qu_interp_self ();
	else
		assert (qu_method_is (f));

	buf = qu_buffer_alloc (s->len);
	u = qu_array_new (MAXSUBS);
	qu_array_freeze (u);

	while (max-- > 0
		&& i < k
		&& i != last
		&& regexec (r, i, MAXSUBS, m, option) != REG_NOMATCH)
	{
		last = i;
		assert (u->cap >= MAXSUBS);
		for (j = 1, n = 0; j < MAXSUBS; j++, n++)
		{
			o = m[j].rm_so;
			if (o < 0)
				break;
			if (isbuf)
				x = (qu_stringy) qu_buffer_new (i + o, m[j].rm_eo - o);
			else
				x = (qu_stringy) qu_string_new (i + o, m[j].rm_eo - o);
			u->objs[j - 1] = x;
		}
		u->len = n;
		if (!qu_buffer_put (buf, i, m[0].rm_so))
			goto err;
		i += m[0].rm_eo;
		x = issub ? qu_sub_call1 ((qu_sub) f, theself, u) : qu_method_call1 ((qu_method) f, u);
		if (x == 0)
			goto err;
		if (!qu_stringy_is (x))
			x = qu_sys_string (x);
		if (!qu_buffer_puts (buf, x))
			goto err;
	}
	if (i < k)
	{
		if (!qu_buffer_put (buf, i, k - i))
			goto err;
	}
	if (isbuf)
	{
		qu_buffer_exchange_tmp ((qu_buffer) s, buf);
		return s;
	}
	return (qu_stringy) qu_buffer_tmp_to_string (buf);

  err:
	if (isbuf)
		QU_BUFFER_PROTECT_END ((qu_buffer) s, scope);
	QU_RETHROW;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	standard class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub class new (pattern: String, option = '') -> Regex
	Create and compile a new regular expression.

	v{pattern} must be a valid regex pattern String.

	v{option} is a combination of:
	@ul
	.C{"i"} (REG_ICASE)
	.C{"u"} (REG_NOSUB)
	.C{"l"} (REG_NEWLINE)
	.C{"o"} (REG_NOTBOL)
	.C{"e"} (REG_NOTEOL)

	@p
	Refer to C{man 7 regex} for valid options and their meanings.

	@exc
	.EArgValue
	.ESyntax

	@xmp
		use Regex
		r = Regex (`[abc]`. "il")
		println (r.match ('aaa bbb ccc'))

	@out
		9
*/
static Regex Regex_new (qu_class self, int argc, qu_any *argv)
{
	int       opt = 0;
	qu_string pat;
	int       err;
	Regex     r;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, pat, qu_string);
	QU_ARG_CHECK  (pat->len);
	if (argc > 1)
	{
		qu_string x;
		QU_ARG_VTYPE (1, x, qu_string);
		opt = _regex_option (x);
	}

	opt &= ~REG_NOSUB;
	opt |= REG_EXTENDED;

	r = qu_gc_new (RegexClass);
	err = regcomp (&r->re, pat->str, opt);
	if (err != 0)
		QU_THROWS (QuESyntax, _regex_error (err, &r->re));
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	low-level instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static void regex__free (Regex self)
{
	regfree (&self->re);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	specific instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub match (x, option = '') -> String
	@def
	sub match (x: String, option = '') -> String
	sub match (x: Buffer, option = '') -> Buffer
	@p
	Find pattern in the String or Buffer v{x}.

	See #{new} for valid v{option} value.

	If found then the first matching part is returned.
	If not, then an empty String or Buffer is returned.

	@exc
	.EArgValue

	@xmp
		use Regex
		r = Regex (`[abc]x`)
		println (r.match ('aaa ax xxx'))

	@out
		ax
*/
static qu_stringy regex_match (Regex self, int argc, qu_any *argv)
{
	int         opt = 0;
	qu_stringy  s;
	regmatch_t  m [2];

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, s, qu_stringy);
	if (argc > 1)
	{
		qu_string x;
		QU_ARG_VTYPE (1, x, qu_string);
		opt = _regex_option (x);
	}
	if (regexec (&self->re, s->str, 2, m, opt) != REG_NOMATCH)
	{
		if (qu_buffer_is (s))
			return (qu_stringy) qu_buffer_new (s->str + m[0].rm_so, m[0].rm_eo - m[0].rm_so);
		return qu_string_new (s->str + m[0].rm_so, m[0].rm_eo - m[0].rm_so);
	}
	else
	{
		if (qu_buffer_is (s))
			return (qu_stringy) qu_buffer_alloc (0);
		return QuStringEMPTY;
	}
}



/*) sub count (x, option = '') -> Int
	@def
	sub count (x: String, option = '') -> Int
	sub count (x: Buffer, option = '') -> Int
	@p
	Count the number of matches in the String or Buffer v{x}.

	See #{new} for valid v{option} value.

	@exc
	.EArgValue

	@xmp
		use Regex
		r = Regex (`[abc]`)
		println (r.count ('aaa bbb ccc'))

	@out
		9
*/
static qu_int regex_count (Regex self, int argc, qu_any *argv)
{
	int         opt = 0;
	qu_stringy  s;
	long        n;
	regex_t    *r;
	char       *i, *k, *last = NULL;
	regmatch_t  m [1];

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, s, qu_stringy);
	if (argc > 1)
	{
		qu_string x;
		QU_ARG_VTYPE (1, x, qu_string);
		opt = _regex_option (x);
	}
	n  = 0;
	r = &self->re;
	i = s->str;
	k = i + strlen (i);
	while (i < k
		&& i != last
		&& regexec (r, i, 1, m, opt) != REG_NOMATCH)
	{
		last = i;
		i += m[0].rm_eo;
		n++;
	}
	return qu_int_new (n);
}




/*) sub split (x, max = 0, option = '') -> Array
	@def
	sub split (x: String, max = 0, option = '') -> Array
	sub split (x: Buffer, max = 0, option = '') -> Array
	@p
	Split the String or Buffer v{x} using the regex pattern as the separator,
	optionally v{max} items at most.

	A non-positive v{max} means no limit.

	The result is an Array containing the items.
	If v{x} is a Buffer then the items are Buffer objects otherwise they are
	String objects.

	See #{new} for valid v{option} value.

	@exc
	.EArgValue

	@xmp
		use Regex

		r = Regex (`[[:space:]]*[,;][[:space:]]*`)
		a = r.split ('a, b ; c ,d')
		println (a)

	@out
		['a', 'b', 'c', 'd']
*/
static qu_array regex_split (Regex self, int argc, qu_any *argv)
{
	long        max = LONG_MAX;
	int         opt = 0;
	qu_stringy  s;
	qu_array    a;
	regex_t    *r;
	char       *i, *k, *last;
	regmatch_t  m [1];
	int         isbuf;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, s, qu_stringy);
	isbuf = qu_buffer_is (s);
	if (argc > 1)
	{
		QU_ARG_vlong (1, max);
		if (max <= 0)
			max = LONG_MAX;
		if (argc > 2)
		{
			qu_string x;
			QU_ARG_VTYPE (2, x, qu_string);
			opt = _regex_option (x);
		}
	}

	a = qu_array_new (8);
	r = &self->re;
	i = s->str;
	k = i + s->len;
	last = NULL;

	while (--max > 0
		&& i < k
		&& i != last
		&& regexec (r, i, 1, m, opt) != REG_NOMATCH)
	{
		last = i;
		if (isbuf)
			qu_array_push (a, qu_buffer_new (i, m[0].rm_so));
		else
			qu_array_push (a, qu_string_new (i, m[0].rm_so));
		i += m[0].rm_eo;
	}
	if (i < k)
	{
		if (isbuf)
			qu_array_push (a, qu_buffer_new (i, k - i));
		else
			qu_array_push (a, qu_string_new (i, k - i));
	}
	return a;
}



/*) sub splitex (x, max = 0, option = '') -> Array
	@def
	sub splitex (x: String, max = 0, option = '') -> Array
	sub splitex (x: Buffer, max = 0, option = '') -> Array
	@p
	This method does the same as #{split} except that the matching pattern is
	included in the Array.

	A non-positive v{max} means no limit.

	See #{new} for valid v{option} value.

	@exc
	.EArgValue

	@xmp
		use Regex

		r = Regex (`[[:space:]]*x[[:space:]]*`)
		a = r.splitex (' x a x  b x  x c d  ', 3)
		println (a)
		println (a.join.quote)

	@out
		['', ' x ', 'a', ' x  ', 'b x  x c d  ']
		' x a x  b x  x c d  '
*/
static qu_array regex_splitex (Regex self, int argc, qu_any *argv)
{
	long        max = LONG_MAX;
	int         opt = 0;
	qu_stringy  s;
	qu_array    a;
	regex_t    *r;
	char       *i, *k, *last;
	regmatch_t  m [2];
	int         isbuf;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, s, qu_stringy);
	isbuf = qu_buffer_is (s);
	if (argc > 1)
	{
		QU_ARG_vlong (1, max);
		if (max <= 0)
			max = LONG_MAX;
		if (argc > 2)
		{
			qu_string x;
			QU_ARG_VTYPE (2, x, qu_string);
			opt = _regex_option (x);
		}
	}

	a = qu_array_new (8);
	r = &self->re;
	i = s->str;
	k = i + s->len;
	last = NULL;
	while (--max > 0
		&& i < k
		&& i != last
		&& regexec (r, i, 1, m, opt) != REG_NOMATCH)
	{
		last = i;
		if (isbuf)
		{
			qu_array_push (a, qu_buffer_new (i, m[0].rm_so));
			qu_array_push (a, qu_buffer_new (i + m[0].rm_so, m[0].rm_eo - m[0].rm_so));
		}
		else
		{
			qu_array_push (a, qu_string_new (i, m[0].rm_so));
			qu_array_push (a, qu_string_new (i + m[0].rm_so, m[0].rm_eo - m[0].rm_so));
		}
		i += m[0].rm_eo;
	}
	if (i < k)
	{
		if (isbuf)
			qu_array_push (a, qu_buffer_new (i, k - i));
		else
			qu_array_push (a, qu_string_new (i, k - i));
	}
	return a;
}



/*) sub replace (x, rep = '', max = 0, option = '')
	@def
	sub replace (x: String, with = '', max = 0, option = '') -> String
	sub replace (x: Buffer, with = '', max = 0, option = '') -> Buffer
	@p
	Substitute matching parts of a String or Buffer.

	If v{x} is a String then a new String is returned.
	If v{x} is a Buffer then the replacement is done inplace and the Buffer object
	is returned.

	A non-positive v{max} means no limit.

	See #{new} for valid v{option} values.

	v{rep} may be a String, Buffer or Function.

	If v{rep} is a String or Buffer then it will replace the matching parts.

	If v{rep} is a Function then it will be called for each substitution with
	one argument: a frozen Array containing the matching groups or an empty one
	if the regex pattern does not contain groups. The Function must return a String
	or Buffer to use as the substitution or throw an exception in case of an error.
	If it throws EStop then substitution stops, otherwise the exception is rethrown.

	If v{x} is a String then the Array will contain String items.
	Otherwise it will contain Buffer items.

	@exc
	.EArgValue
	.EMemory
	.EModify
	.Exception anything thrown by the Sub or Method

	@xmp
		use Regex

		sub rep (a)
			println (a)
			return 'x'
		;;

		e = Regex (`([bc])(...)r`)
		s = '__b123r__c456r__'
		r = e.replace (s, @rep)
		println (r)

		r = e.replace (s, lambda (x) 'z')
		println (r)

		e = Regex ('[bc]')
		r = e.replace (s, 'x')
		println (r)

	@out
		['b', '123']
		['c', '456']
		__x__x__
		__z__z__
		__x123r__x456r__
*/
static qu_string regex_replace (Regex self, int argc, qu_any *argv)
{
	qu_stringy s;
	qu_any     u = QuStringEMPTY;
	long       max = LONG_MAX;
	int        opt = 0;

	QU_ARG_VCOUNT (1, 4);
	QU_ARG_VTYPE  (0, s, qu_string);
	if (argc > 1)
	{
		u = argv[1];
		if (argc > 2)
		{
			QU_ARG_vlong (2, max);
			if (max <= 0)
				max = LONG_MAX;
			if (argc > 3)
			{
				qu_string x;
				QU_ARG_VTYPE (3, x, qu_string);
				opt = _regex_option (x);
			}
		}
	}
	if (qu_function_is (u))
		return _regex_do_substex (self, s, (qu_function) u, max, opt);
	QU_ARG_TYPE (u, qu_string);
	return _regex_do_subst (self, s, (qu_string) u, max, opt);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Regex_init (qu_module m)
{
	RegexClass = qu_class_new_finalz (m, "Regex", TRegex, regex__free, 0);
	RegexClass->id = QU_CLASS_ID_Regex;

	/* standard class methods */
	qu_class_method_class (RegexClass, QU_NAME_new, Regex_new, -1);

	/* specific instance methods */
	qu_class_method_instz (RegexClass, "match"  , regex_match  , -1);
	qu_class_method_instz (RegexClass, "count"  , regex_count  , -1);
	qu_class_method_instz (RegexClass, "split"  , regex_split  , -1);
	qu_class_method_instz (RegexClass, "splitex", regex_splitex, -1);
	qu_class_method_instz (RegexClass, "replace", regex_replace, -1);

	/* finalize declaration */
	qu_class_enddecl (RegexClass);

	return 1;
}



#else /*HAVE_REGEX_H*/


int qu_Regex_init (qu_module m)
{
	QU_THROWZ (QuESystem, "POSIX regex library not available");
}


#endif /*HAVE_REGEX_H*/
