/* Ssl.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


#ifdef HAVE_OPENSSL_SSL_H /* up to the end */

#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>


/* stupid forward */
static qu_class SslClass;


/*) module Ssl
	This module provides an interface to the SSL library.

	Note that this is only available if you have OpenSSL installed on your
	system. Otherwise k{use}ing this module will throw a compile exception.

	There are only two classes provided by this module:
	${Ssl} which represent SSL connection objects and
	${Sslctx} which represent SSL context objects.

	Basically, all you have to do is create a Sslctx object then attach it
	to a Stream simply by creating a SSl object.

	The Ssl object will "attach" itself to the Stream by redirecting the
	Stream's I/O functions to SSL specific functions. You can then use the
	Stream exactly the same way as if SSL were not present.

	@warn
	The implementation is incomplete.
	A helping hand would be nice :-)
*/


/*) const __version = ''
	This holds the OpenSSL version string.
	The actual value is set by the Ssl module.
*/


/*) const __verno = ''
	This holds the OpenSSL version number.
	The actual value is set by the Ssl module.
*/


/*) sub isssl (x) -> Boolean
	Whether or not v{x} is a Stream and an SSL-enabled Stream.

	@noexc
*/
static qu_boolean Ssl_isssl (qu_stream x)
{
	return (qu_stream_is (x) && x->obj && ((qu_obj) x->obj)->klass == SslClass)
		? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	SSL error handling
----------------------------------------------------------------------------*/

/*) class ESSL is ESystem __final
	This is the exception class thrown by the Ssl module for specific SSL
	errors.

	Notice that it is a subclass of ESystem. Generally you only need to catch
	ESystem, unless of course, you want to find out whether the exception
	was SSL specific.

	@xmp
		use Ssl
		ctx = Ssl.Sslctx ('/foo', '/bar')

	@out
		[1024] __main__ (ESSL) error:02001002:system library:fopen:No such file or directory

	@p
	The following SSL errors are internally translated into an ESystem exception
	to make things simpler and more consistent with -{Stream} objects.
	@indent
	@rc
	.| SSL_ERROR_NONE              | EAGAIN
	.| SSL_ERROR_WANT_READ         | EAGAIN
	.| SSL_ERROR_WANT_WRITE        | EAGAIN
	.| SSL_ERROR_ZERO_RETURN       | ECOMM
	.| SSL_ERROR_WANT_X509_LOOKUP  | ENODATA
	.| SSL_ERROR_SSL               | ECONNABORTED
	.| SSL_ERROR_SYSCALL           | EAGAIN
	.| else                        | EPROTO
	@unindent

	@p
	Note that EINTR and EAGAIN are handled internally by Stream via the
	-{Stream.timeout} mechanism. In Qu, you never have to deal with them.
*/
static qu_class ESSLClass;


/*) var msg = 'SSL error'
	This is the default error message.
*/
#define ESSLDefaultMessage "SSL error"


/*
	Throw the last SSL error.
*/
#define THROWSSL \
	QU_THROWS (ESSL, _ssl_last_error ())

static qu_string _ssl_last_error ()
{
	char  buf [1024], *s;

	s = ERR_error_string (ERR_get_error (), buf);
	if (s == NULL)
		return qu_string_newz ("unknown SSL error");
	return qu_string_newz (s);
}



/*
	Translate SSL error into errno.
*/
static void _ssl_seterrno (SSL *s, int result)
{
	switch (SSL_get_error (s, result))
	{
		case SSL_ERROR_NONE:
			errno = EAGAIN;
			break;

		case SSL_ERROR_WANT_READ:
			errno = EAGAIN;
			break;

		case SSL_ERROR_WANT_WRITE:
			errno = EAGAIN;
			break;

		case SSL_ERROR_ZERO_RETURN:
			errno = ECOMM;
			break;

		case SSL_ERROR_WANT_X509_LOOKUP:
			errno = ENODATA;
			break;

		case SSL_ERROR_SSL:
			errno = ECONNABORTED;
			break;

		case SSL_ERROR_SYSCALL:
			if (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)
			{
				errno = EAGAIN;
				break;
			}
			/*nobreak*/

		default:
			errno = EPROTO;
	}
}



/*----------------------------------------------------------------------------
	class Sslctx
----------------------------------------------------------------------------*/

/*) class Sslctx __final
	A Sslctx object defines a SSL context.

	It is a wrapper around SSL_CTX.

	You must create a SSL context before you can create a SSL connection.
	See #{new} and ${Ssl.new}.
*/
static qu_class SslctxClass;

typedef struct TSslctx TSslctx;
struct TSslctx
{
	QU_OBJECT_HEAD;
	SSL_CTX   *ctx;   /* the SSL context */
	qu_string  cert;  /* certificate file  */
	qu_string  keys;  /* private keys file */
};

typedef TSslctx *Sslctx;

#define SslctxIs(o) (qu_classof (o) == SslctxClass)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Sslctx - standard class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub class new (cert: String, keys: String) -> Sslctx
	Create a new SSL context object.

	v{cert} is the certificate file and v{keys} is the private keys file.

	@warn
	Both must be asbolute paths otherwise EArgValue is thrown.

	@exc
	.EArgValue
	.ESecurity
	.ESystem

	@xmp
		use Ssl
		ctx = Ssl.Sslctx ('/foo/bar/baz.cert', '/foo/bar/goo.cert')
*/
static Sslctx Sslctx_new (qu_class self, qu_string cert, qu_string keys)
{
	SSL_CTX *c;
	Sslctx   x;

	__qu_unsafe__;

	QU_ARG_CHECK (qu_string_is (cert)
	      && qu_string_is (keys)
	      && cert->len
	      && keys->len
	      && qu_file_is_abs (cert)
	      && qu_file_is_abs (keys));

  __qu_blocking_begin;
	c = SSL_CTX_new (SSLv23_method ());
	if (c)
	{
		if (SSL_CTX_use_PrivateKey_file (c, keys->str, SSL_FILETYPE_PEM) < 1)
			goto nope;

		if (SSL_CTX_use_certificate_chain_file (c, cert->str) < 1)
			goto nope;

		if (SSL_CTX_set_options (c, SSL_OP_ALL) < 1)
			goto nope;

		SSL_CTX_set_verify (c, SSL_VERIFY_NONE, NULL);

	  __qu_blocking_end;
		x = qu_gc_new (SslctxClass);
		x->ctx  = c;
		x->cert = cert;
		x->keys = keys;
		return x;

	  nope:

		SSL_CTX_free (c);
	}
  __qu_blocking_end;
	THROWSSL;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Sslctx - low-level instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static void Sslctx_free (Sslctx self)
{
	SSL_CTX_free (self->ctx);
}



static void Sslctx_mark (Sslctx self)
{
	qu_gc_mark (self->cert);
	qu_gc_mark (self->keys);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Sslctx - standard instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub __like (x) -> Boolean
	Return true if v{x} is a Sslctx object and both the certifiacte
	and private keys files are the same, false otherwise.
*/
static qu_boolean Sslctx__like (Sslctx self, qu_any x)
{
	return (SslctxIs (x)
		&& qu_string_EQ (self->cert, ((Sslctx) x)->cert)
		&& qu_string_EQ (self->keys, ((Sslctx) x)->keys)) ? QuTrue : QuFalse;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Sslctx - specific instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub cert -> String
	Return the certificate file as passed to the #{new} method.

	@noexc
*/
static qu_string Sslctx_cert (Sslctx self)
{
	return self->cert;
}



/*) sub keys -> String
	Return the private keys file as passed to the #{new} method.

	@noexc
*/
static qu_string Sslctx_keys (Sslctx self)
{
	return self->keys;
}



/*) sub flush_sessions (time = 0) -> Null
	Flush sessions before a specified v{time}.

	@exc
	.EArgValue
*/
static qu_null Sslctx_flush_sessions (Sslctx self, int argc, qu_any *argv)
{
	long t = time (NULL);

	if (argc)
	{
		QU_ARG_vlong (0, t);
		if (t <= 0)
			t = 0;
	}
	SSL_CTX_flush_sessions (self->ctx, t);
	return QuNil;
}



/*----------------------------------------------------------------------------
	class Ssl
----------------------------------------------------------------------------*/

/*) class Ssl __final
	A Ssl object represents a Ssl connection.

	It requires a Stream object to perform I/O and timeouts.

	An Ssl object is not a Stream. It simply redirect the I/O routines of
	the underlying Stream to specific SSL routines. Thus to do any operation,
	you use the Stream object instead.
*/
/* static qu_class SslClass; */

typedef struct TSsl TSsl;
struct TSsl
{
	QU_OBJECT_HEAD;
	SSL          *ssl;
	Sslctx        ctx;
	qu_stream     stm;
	qu_fstream_s *old;
};

typedef TSsl *Ssl;

/* test for an Ssl object */
#define SslIs(o) (qu_classof (o) == SslClass)

/* flags */
#define SslACCEPT  QU_FLAG_0 /* accept already successful */
#define SslCONNECT QU_FLAG_1 /* connect already successful */

/* make sure the SSL object has not been freed */
#define Ssl_ENSURE_NOT_FREED \
	QU_REQUIREZ (QuECall, self->ssl, "Ssl object already freed")

/* what time is now? */
#define NOW qu_stream_now ()


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Ssl stream functions
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static qu_fstream_s SslFuncs;

static ssize_t _ssl_read (Ssl self, char *buf, size_t n)
{
	ssize_t r;

	if (self->ssl == 0)
	{
		errno = EBADF;
		return -1;
	}

	r = (ssize_t) SSL_read (self->ssl, buf, (int) n);
	if (r <= 0)
	{
		_ssl_seterrno (self->ssl, r);
		return -1;
	}
	return r;
}



static ssize_t _ssl_write (Ssl self, const char *buf, size_t n)
{
	ssize_t r;

	if (self->ssl == 0)
	{
		errno = EBADF;
		return -1;
	}

	r = (ssize_t) SSL_write (self->ssl, buf, (int) n);
	if (r <= 0)
	{
		_ssl_seterrno (self->ssl, r);
		return -1;
	}
	return r;
}



static qu_stream _ssl_accept (Ssl self)
{
	double     to;
	int        r;
	qu_stream  f;
	Ssl        o;
	SSL       *s;

	assert (self->stm && qu_stream_is (self->stm));
	assert (self->old && self->old->acc);
	assert (self->ssl == 0);

	f = self->old->acc (self->stm);
	QU_ENSURE (f);

	s = SSL_new (self->ctx->ctx);
	if (s == NULL)
	{
		qu_stream_close (f);
		THROWSSL;
	}

	SSL_set_fd (s, f->fd);

	to = self->stm->timeout;

  __qu_blocking_begin;
	while ((r = SSL_accept (s)) <= 0)
	{
		_ssl_seterrno (s, r);
		switch (errno)
		{
			case EINTR:
			case EAGAIN:
				break;

			default:
			  __qu_blocking_end;
				qu_stream_close (f);
				SSL_free (s);
				QU_THROWERRNO;
		}
		if (NOW > to)
		{
		  __qu_blocking_end;
			qu_stream_close (f);
			SSL_free (s);
			QU_THROW (QuETimeout);
		}
	  __qu_blocking_check_signal;
		qu_thread_yield ();
	}
  __qu_blocking_end;

	o        = qu_gc_new (SslClass);
	o->stm   = f;
	o->ssl   = s;
	o->ctx   = self->ctx;
	o->old   = f->funcs;
	f->obj   = o;
	f->funcs = &SslFuncs;

	return f;
}



static qu_null _ssl_connect (Ssl self, qu_any sa)
{
	double     to;
	int        r;
	qu_stream  f;
	Ssl        o;
	SSL       *s;

	assert (self && SslIs (self));
	assert (self->stm && qu_stream_is (self->stm));
	assert (self->old && self->old->acc);
	assert (self->ssl == 0);

	f = self->stm;
	QU_ENSURE (self->old->con (f, sa));

	s = SSL_new (self->ctx->ctx);
	if (s == NULL)
	{
		qu_stream_close (f);
		THROWSSL;
	}

	SSL_set_fd (s, f->fd);

	to = f->timeout;

  __qu_blocking_begin;
	while ((r = SSL_connect (s)) <= 0)
	{
		_ssl_seterrno (s, r);
		switch (errno)
		{
			case EINTR:
			case EAGAIN:
				break;

			default:
			  __qu_blocking_end;
				qu_stream_close (f);
				SSL_free (s);
				QU_THROWERRNO;
		}
		if (NOW > to)
		{
		  __qu_blocking_end;
			qu_stream_close (f);
			SSL_free (s);
			QU_THROW (QuETimeout);
		}
	  __qu_blocking_check_signal;
		qu_thread_yield ();
	}
  __qu_blocking_end;

	o        = qu_gc_new (SslClass);
	o->stm   = f;
	o->ssl   = s;
	o->ctx   = self->ctx;
	o->old   = f->funcs;
	f->obj   = o;
	f->funcs = &SslFuncs;

	return QuNil;
}



static qu_fstream_s SslFuncs =
{
	(qu_stream_rdr_f)   NULL,
	(qu_stream_wrt_f)   NULL,
	(qu_stream_ordr_f)  _ssl_read,
	(qu_stream_owrt_f)  _ssl_write,
	(qu_stream_acc_f)   _ssl_accept,
	(qu_stream_con_f)   _ssl_connect
};



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Ssl - standard class methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub class new (ctx: Sslctx, stm: Stream)
	Create a new SSL connection object.

	v{ctx} is an ${Sslctx} object and v{stm} is the Stream object to attach to.
	It must support f{accept} and f{connect}.

	This will redirect the Stream I/O routines to SSL specific routines.

	You generaly don't need the returned Ssl object since you can get it
	from -{Stream.obj}.

	@exc
	.EArgValue
	.EImplement
	.ESystem

	@xmp
		use Net, Ssl

		sock    = ... create a socket ...
		sslctx  = ... create a Sslctx ...
		Ssl (sslctx, sock)
*/
static Ssl Ssl_new (qu_class self, Sslctx c, qu_stream f)
{
	Ssl s;

	QU_ARG_CHECK (SslctxIs (c) && qu_stream_is (f) && f->fd >= 0);
	QU_REQUIRE (QuEImplement, f->funcs->acc && f->funcs->con);

	s        = qu_gc_new (SslClass);
	s->stm   = f;
	s->ssl   = 0;
	s->ctx   = c;
	s->old   = f->funcs;

	f->obj   = s;
	f->funcs = &SslFuncs;
	return s;
}




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Ssl - low-level instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static void Ssl__free (Ssl self)
{
	if (self->ssl)
		SSL_free (self->ssl);
}



static void Ssl__mark (Ssl self)
{
	qu_gc_xmark (self->ctx);
	qu_gc_mark  (self->stm);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Ssl - standard instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub __like (x) -> Boolean
	Return true if v{x} is a Ssl object and then underlying Stream objects
	are the same.
*/
static qu_boolean Ssl__like (Ssl self, qu_any x)
{
	return (SslIs (x) && self->stm == ((Ssl) x)->stm) ? QuTrue : QuFalse;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Ssl - specific instance methods
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub ctx -> Sslctx
	Return the ${Sslctx} object passed to the #{new} method.

	@noexc
*/
static Sslctx Ssl_ctx (Ssl self)
{
	assert (self->ctx && SslctxIs (self->ctx));
	return self->ctx;
}



/*) sub stream -> Stream
	Return the Stream object associated with this Ssl object.

	@noexc
*/
static qu_stream Ssl_stream (Ssl self)
{
	assert (self->ctx && qu_stream_is (self->stm));
	return self->stm;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Ssl_init (qu_module m)
{
	OpenSSL_add_all_algorithms ();
	ERR_load_crypto_strings ();
	SSL_load_error_strings ();


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		module
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	qu_module_add_const    (m, "__version", qu_string_new_staticz (OPENSSL_VERSION_TEXT));
	qu_module_add_const    (m, "__verno"  , qu_integer_newl (OPENSSL_VERSION_NUMBER));

	qu_module_add_function (m, "isssl", Ssl_isssl, 1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		ESSL
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	ESSLClass = qu_exception_class_new (m, QuESystemClass, "ESSL", ESSLDefaultMessage);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Sslctx
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	SslctxClass = qu_class_new_finalz (m, "Sslctx", TSslctx, Sslctx_free, Sslctx_mark);

	/* standard class methods */
	qu_class_method_class (SslctxClass, QU_NAME_new, Sslctx_new, 2);

	/* standard instance methods */
	qu_class_method_inst  (SslctxClass, QU_NAME__like, Sslctx__like, 1);

	/* specific instance methods */
	qu_class_method_instz (SslctxClass, "cert"          , Sslctx_cert          ,  0);
	qu_class_method_instz (SslctxClass, "keys"          , Sslctx_keys          ,  0);
	qu_class_method_instz (SslctxClass, "flush_sessions", Sslctx_flush_sessions, -1);

	/* finalize declaration */
	qu_class_enddecl (SslctxClass);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Ssl
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	SslClass = qu_class_new_finalz (m, "Ssl", TSsl, Ssl__free, Ssl__mark);

	/* standard class methods */
	qu_class_method_class (SslClass, QU_NAME_new, Ssl_new, 2);

	/* standard instance methods */
	qu_class_method_inst  (SslClass, QU_NAME__like, Ssl__like, 1);

	/* specific instance methods */
	qu_class_method_instz (SslClass, "ctx"   , Ssl_ctx   , 0);
	qu_class_method_instz (SslClass, "stream", Ssl_stream, 0);

	/* finalize declaration */
	qu_class_enddecl (SslClass);

	return 1;
}



#else /* HAVE_OPENSSL_SSL_H */


int qu_Ssl_init (qu_module m)
{
	QU_THROWZ (QuESystem, "OpenSSL is not available");
}


#endif /* HAVE_OPENSSL_SSL_H */
