/* Syslog.c - Copyright (c) Marc Krisnanto */

#include "qu.h"
#include <syslog.h>


/*) module Syslog
	This module provides an interface to the system logging.

	See C{man 3 syslog} for details.
*/


#define MAXIDENTLEN 64
static char Ident [MAXIDENTLEN + 1];


/*) __doc__ constants
	Various constants.

	See C{man 3 syslog} for details.

	The C{LOG_} prefix is removed so you should just say, e.g:
	C{Syslog.CONS}, C{Syslog.NDELAY}, etc.

	@code
	CONS
	NDELAY
	NOWAIT
	ODELAY
	PERROR
	PID

	AUTH
	AUTHPRIV
	CRON
	DAEMON
	FTP
	KERN
	LOCAL0
	LOCAL1
	LOCAL2
	LOCAL3
	LOCAL4
	LOCAL5
	LOCAL6
	LOCAL7
	LPR
	MAIL
	NEWS
	SYSLOG
	USER
	UUCP

	EMERG
	ALERT
	CRIT
	ERR
	WARNING
	NOTICE
	INFO
	DEBUG
*/


/*) sub Syslog (priority: Int, message: String) -> Null
	Generate a log message which will be distributed by syslogd(8).

	v{priority} is formed by bit-or-ing the facility and the level values.

	If v{message} is not a String then it's __string method is called.

	@xmp
		use Syslog

		Syslog (Syslog.ERR, "something wrong")

	@exc
	.ESecurity
*/
static qu_null qu_syslog_Syslog (qu_integer p, qu_stringy msg)
{
	int prio;

	__qu_unsafe__;

	QU_ARG_int (p, prio);
	QU_ARG_TYPE (msg, qu_stringy);
	syslog (prio, "%s", msg->str);
	return QuNil;
}



/*) sub open (ident: String, option = 0, facility = LOG_USER) -> Null
	Open a connection to the system logger for a program.

	v{ident} is prepended to every message and is typically set to the program name.
	It must not be longer than 64 bytes.

	v{option} specifies flags which control the operation of f{open} and subsequent
	calls to #{log}.

	v{facility} establishes a default to be used if none is specified in subsequent
	calls #{log}.

	@xmp
		use Syslog
		Syslog.open ("Foo")

	@exc
	.EArgValue
	.ESecurity
*/
static qu_null qu_syslog_open (int argc, qu_any *argv)
{
	qu_string ident;
	int       option = 0;
	int       facility = LOG_USER;

	static char *argx[] =
	{
		"s", "ident",
		":",
		"i", "option",
		"i", "facility",
		NULL
	};

	__qu_unsafe__;

	QU_ARG_XPARSE (&ident, &option, &facility);
	QU_ARG_CHECK (ident->len <= MAXIDENTLEN);
	memcpy (Ident, ident->str, ident->len + 1);
	openlog (Ident, option, facility);
	return QuNil;
}



/*) sub close -> Null
	Close the descriptor being used to write to the system logger.
	The use of f{close} is optional.

	@exc
	.ESecurity
*/
static qu_null qu_syslog_close ()
{
	__qu_unsafe__;

	closelog ();
	return QuNil;
}



/*) sub mask (*arg) -> Integer
	@def
	sub mask (newmask: Integer) -> Integer
	sub mask -> Integer
	@p
	Get/set the log mask.

	A process has a log priority mask that determines which calls to #{log}
	may be logged. All other calls will be ignored. Logging is enabled for the
	priorities that have the corresponding bit set in mask. The initial mask is
	such that logging is enabled for all priorities.

	If called with one argument f{logmask} sets this logmask for the current process,
	and returns the previous mask. If the mask argument is 0, the current logmask is
	not modified.

	@xmp
		use Syslog

		Syslog.mask = Syslog.EMERG | Syslog.ALERT

	@exc
	.EArgValue
	.ESecurity
*/
static qu_integer qu_syslog_mask (int argc, qu_any *argv)
{
	int m, v = 0;

	__qu_unsafe__;

	if (argc)
		QU_ARG_vint (0, v);

	m = setlogmask (v);
	return qu_integer_newi (m);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Syslog_init (qu_module m)
{
	/* constants */
#define c_(name) qu_module_add_const (m, #name, qu_integer_newi (LOG_##name))

	c_(CONS);
	c_(NDELAY);
	c_(NOWAIT);
	c_(ODELAY);
	c_(PERROR);
	c_(PID);

	c_(AUTH);
	c_(AUTHPRIV);
	c_(CRON);
	c_(DAEMON);
	c_(FTP);
	c_(KERN);
	c_(LOCAL0);
	c_(LOCAL1);
	c_(LOCAL2);
	c_(LOCAL3);
	c_(LOCAL4);
	c_(LOCAL5);
	c_(LOCAL6);
	c_(LOCAL7);
	c_(LPR);
	c_(MAIL);
	c_(NEWS);
	c_(SYSLOG);
	c_(USER);
	c_(UUCP);

	c_(EMERG);
	c_(ALERT);
	c_(CRIT);
	c_(ERR);
	c_(WARNING);
	c_(NOTICE);
	c_(INFO);
	c_(DEBUG);

	/* functions */
	qu_module_add_function (m, "Syslog" , qu_syslog_Syslog,  2);
	qu_module_add_function (m, "open"   , qu_syslog_open  , -1);
	qu_module_add_function (m, "close"  , qu_syslog_close , -1);
	qu_module_add_function (m, "mask"   , qu_syslog_mask  , -1);

	return 1;
}
