/* Time.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Time
	This module provides basic functions to deal with date/time.

	A "time value" (calendar time) is a Float number which value is the number
	of seconds since Epoch (00:00:00 UTC, January 1, 1970). This value is an
	independent value which is the same anywhere around the globe.

	Generally you do math with "time values" and convert the end result into
	a time within a specific timezone, e.g. #{localtime}, #{gmtime}.

	@note
	The Time module uses the C library thus imposes the same limitations,
	for example, you must not specify a time before Epoch.

	Hopefully this will change in the future to allow a larger time range.
*/


/*) struct Tm
		year  : Int  # year
		mon   : Int  # month 0 .. 11
		day   : Int  # day of the month 1 .. 31
		hour  : Int  # hours 0 .. 23
		min   : Int  # minutes 0 .. 59
		sec   : Int  # seconds 0 .. 59
		usec  : Int  # microsecond >= 0
		dst   : Int  # daylight saving time flag (see below)
		wday  : Int  # day of the week 0 .. 6
		yday  : Int  # day of the year 0 .. 365
	;;
	This represents a time value broken into its components.

	It is the essentially the same as C's v{struct tm} except for the v{usec}
	field and the v{year} field which contains the actual year (not the number
	of years since 1900 which is just ridiculous).

	The purpose of v{usec} is to match the precision used by the -{Stream}
	class v{timeout} value.

	v{isdst} is a flag that indicates whether daylight saving time is in
	effect at the time described. The value is positive if daylight saving
	time is in effect, zero if it is not, and negative if the information
	is not available.

	See C{man 3 ctime} for more information.
*/
static qu_struct TmStruct;


/*----------------------------------------------------------------------------
	internals
----------------------------------------------------------------------------*/

/* the current time - double */
#define _time_now() qu_stream_now ()


/*
	use our own struct
	e_usec (as in gettimeofday) is only good in relation with qu_stream->timeout
*/
typedef struct time_s time_s;
struct time_s
{
	struct tm m;
	long      e_usec;
};

#define e_year  m.tm_year
#define e_mon   m.tm_mon
#define e_day   m.tm_mday
#define e_hour  m.tm_hour
#define e_min   m.tm_min
#define e_sec   m.tm_sec
#define e_dst   m.tm_isdst
#define e_wday  m.tm_wday
#define e_yday  m.tm_yday


static double _struct2val (time_s *e)
{
	double v;

	e->e_year -= 1900;
	v = mktime ((struct tm *) e) + e->e_usec / 1e6;
	e->e_year += 1900;
	return v;
}


static void _val2struct (double v, time_s *e)
{
	time_t t = (time_t) v;

	localtime_r (&t, (struct tm *) e);
	e->e_year += 1900;
	e->e_usec = fmod (v, 1e6);
}


static void _val2local (double v, time_s *e)
{
	time_t t = (time_t) v;
	localtime_r (&t, (struct tm *) e);
	e->e_year += 1900;
}


static void _val2gm (double v, time_s *e)
{
	time_t t = (time_t) v;
	gmtime_r (&t, (struct tm *) e);
	e->e_year += 1900;
}


static qu_any _val2obj (double v)
{
	return qu_float_new (v);
}


static void _rec2struct (qu_record r, time_s *e)
{
	qu_any *p = r->objs;

	e->e_year = qu_int_tol (p[0]);
	e->e_mon  = qu_int_tol (p[1]);
	e->e_day  = qu_int_tol (p[2]);
	e->e_hour = qu_int_tol (p[3]);
	e->e_min  = qu_int_tol (p[4]);
	e->e_sec  = qu_int_tol (p[5]);
	e->e_usec = qu_int_tol (p[6]);
	e->e_dst  = qu_int_tol (p[7]);
	e->e_wday = qu_int_tol (p[8]);
	e->e_yday = qu_int_tol (p[9]);
}


#if 0 /* not used */
static double _rec2val (qu_record r)
{
	time_s e;
	qu_any   *p = r->objs;

	e.e_year = qu_int_tol (p[0]) - 1900;
	e.e_mon  = qu_int_tol (p[1]);
	e.e_day  = qu_int_tol (p[2]);
	e.e_hour = qu_int_tol (p[3]);
	e.e_min  = qu_int_tol (p[4]);
	e.e_sec  = qu_int_tol (p[5]);
	e.e_usec = qu_int_tol (p[6]);
	e.e_dst  = qu_int_tol (p[7]);

	return (double) mktime ((struct tm *) &e) + e.e_usec / 1e6;
}
#endif


/*----------------------------------------------------------------------------
	module functions
----------------------------------------------------------------------------*/


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	get field value
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

#define CODE_FOR_GET_FIELD(fieldname) \
	double v; \
	time_s e; \
	\
	QU_ARG_VCOUNT (0, 1); \
	if (argc == 0) \
		v = _time_now (); \
	else \
	{ \
		QU_ARG_vdouble (0, v); \
	} \
	_val2struct (v, &e); \
	return qu_int_new (e.fieldname) \



/*) sub year (timeval: Real = now) -> Int
	Get the year of a time value.

	@exc
	.EValue
*/
static qu_int Time_year (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_year);
}



/*) sub mon (timeval: Real = now) -> Int
	Get the month of a time value.

	@warn
	The value returned is within 0..11 where 0 means January.

	@exc
	.EValue
*/
static qu_int Time_mon (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_mon);
}



/*) sub day (timeval: Real = now) -> Int
	Get the day of the month of a time value.

	@exc
	.EValue
*/
static qu_int Time_day (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_day);
}



/*) sub hour (timeval: Real = now) -> Int
	Get the hour of a time value.

	@exc
	.EValue
*/
static qu_int Time_hour (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_hour);
}



/*) sub min (timeval: Real = now) -> Int
	Get the minute of a time value.

	@exc
	.EValue
*/
static qu_int Time_min (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_min);
}



/*) sub sec (timeval: Real = now) -> Int
	Get the second of a time value.

	@exc
	.EValue
*/
static qu_int Time_sec (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_sec);
}



/*) sub usec (timeval: Real = now) -> Int
	Get the microsecond of a time value.

	@exc
	.EValue
*/
static qu_int Time_usec (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_usec);
}



/*) sub isdst (timeval: Real = now) -> Int
	Get the daylight saving flag of a time value.

	@exc
	.EValue
*/
static qu_int Time_isdst (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_dst);
}



/*) sub wday (timeval: Real = now) -> Int
	Get the weekday of a time value.

	@exc
	.EValue
*/
static qu_int Time_wday (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_wday);
}



/*) sub yday (timeval: Real = now) -> Int
	Get the day of the year of a time value.

	@exc
	.EValue
*/
static qu_int Time_yday (int argc, qu_any *argv)
{
	CODE_FOR_GET_FIELD (e_yday);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	current time
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub now (*arg)
	@def
	sub now     -> Float
	sub now (x) -> Record
	@p
	Get the current time, i.e. the number of seconds since Epoch
	(00:00:00 UTC, January 1, 1970).

	If called with any argument then return a Record of Tm.
	Otherwise return the time value.

	@noexc

	@xmp
		use Time
		println (Time.now)
		println (Time.now (true))

	@out
		1122235488.713521
		[2005, 6, 25, 3, 4, 48, 235488, 0, 1, 205]
*/
static qu_any Time_now (int argc, qu_any *argv)
{
	if (argc == 1)
	{
		time_s e;

		_val2struct (_time_now (), &e);
		{
 			qu_any arg[10] =
			{
				qu_int_new (e.e_year),
				qu_int_new (e.e_mon),
				qu_int_new (e.e_day),
				qu_int_new (e.e_hour),
				qu_int_new (e.e_min),
				qu_int_new (e.e_sec),
				qu_int_new (e.e_usec),
				qu_int_new (e.e_dst),
				qu_int_new (e.e_wday),
				qu_int_new (e.e_yday)
			};
			return qu_record_new (TmStruct, 10, arg);
		}
	}
	return _val2obj (_time_now ());
}



/*) sub date -> Float
	Return the current date.

	This is the same as #{now} with hour = min = sec = usec = 0.

	@xmp
		use Time
		println (Time.now)
		println (Time.date)

	@out
		1122235517.7856021
		1122224400
*/
static qu_float Time_date ()
{
	time_s e;

	_val2struct (_time_now (), &e);
	e.e_hour = e.e_min = e.e_sec = e.e_usec = 0;
	return _val2obj (_struct2val (&e));
}



/*) sub clock -> Float
	Return an approximation of processor time used by the program in seconds.

	@xmp
		use Time

		t = Time.clock
		n = 100000
		--n while n
		println (Time.clock - t)

	@out
		0.01
*/
static qu_float Time_clock ()
{
	return qu_float_new ((double) clock () / (double) CLOCKS_PER_SEC);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	info
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub timezone -> Int
	Return the timezone value.

	See C{man 3 tzset} for details.

	@noexc
*/
static qu_int Time_timezone ()
{
	return qu_int_new (timezone);
}



/*) sub tzname (which = 0) -> String
	Return the timezone name.

	v{which} is 0 or 1. If it is not 1, then it is assumed to be 0.

	See C{man 3 tzset} for details.

	@noexc

	@xmp
		use Time
		println (Time.tzname)

	@out
		WIB
*/
static qu_string Time_tzname (int argc, qu_any *argv)
{
	int i = 0;

	if (argc)
	{
		i = qu_integer_toi (argv[0]);
		if (i < 0 || i > 1)
			i = 0;
	}
	return qu_string_newz (tzname[i]);
}



/*) sub daylight -> Int
	Return the daylight saving time.

	See C{man 3 tzset} for details.

	@xmp
		use Time
		println (Time.daylight)

	@out
		0
*/
static qu_int Time_daylight ()
{
	return qu_int_new (daylight);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	counting
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/*) sub ydays (year: Integer) -> Int
	Return the number of days in the specified v{year}.

	v{year} must be a positive integer value within the Int range.

	@exc
	.EValue

	@xmp
		use Time
		println (Time.ydays (2005))

	@out
		365
*/
static qu_int Time_ydays (qu_integer year)
{
	int y, r;

	QU_ARG_int (year, y);
	QU_ARG_CHECK (y >= 0);
	r = 365 + ((y % 4) == 0) - (((y % 100) == 0) && ((y % 400) != 0));
	return qu_int_newi (r);
}



/*) sub mdays (*arg) -> Int
	@def
	sub mdays (m: Integer, y: Integer = Time.year) -> Int
	sub mdays (t: Time.tm) -> Int
	@p
	Return the number of days in the specified month v{m} of year v{y}
	or #{Tm} Record.

	@exc
	.EValue

	@xmp
		use Time
		println (Time.mdays (1, 2004))

	@out
		29
*/
static qu_int Time_mdays (int argc, qu_any *argv)
{
	int y, m, d;

	QU_ARG_VCOUNT (1, 2);
	if (qu_record_is (argv[0]))
	{
		qu_record r;

		QU_ARG_vrecord (0, r, TmStruct);
		y = qu_int_tol (r->objs[0]);
		m = qu_int_tol (r->objs[1]);
	}
	else
	{
		QU_ARG_vint (0, m);
		if (argc > 1)
		{
			QU_ARG_vint (1, y);
		}
		else
		{
			time_s e;
			_val2struct (_time_now (), &e);
			y = e.e_year;
		}
	}
	QU_ARG_CHECK (y > 0 && 0 <= m && m < 12);
	++m;
	d = 30 + ((m % 2) ^ (m > 7)) - ((m != 2) ? 0 : 2 - ((y % 4) == 0) + (((y % 100) == 0) && ((y % 400) != 0)));
	return qu_int_newi (d);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	localtime & gmtime
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static qu_any _time_lgtime (int gm, int argc, qu_any *argv)
{
	time_s  e;
	double  v;
	int     i = -1;

	QU_ARG_VCOUNT (0, 2);
	if (argc == 0)
	{
		v = _time_now ();
	}
	else
	{
		if (qu_record_is (argv[0]))
		{
			qu_record r;

			QU_ARG_vrecord (0, r, TmStruct);
			_rec2struct (r, &e);
			v = _struct2val (&e);
		}
		else
		{
			QU_ARG_vdouble (0, v);
		}
		if (argc > 1)
		{
			QU_ARG_vint (1, i);
			QU_ENSURE_INDEX (i, 6);
		}
	}

	if (gm)
		_val2gm (v, &e);
	else
		_val2local (v, &e);

	switch (i)
	{
		case 0: return qu_int_new (e.e_year);
		case 1: return qu_int_new (e.e_mon);
		case 2: return qu_int_new (e.e_day);
		case 3: return qu_int_new (e.e_hour);
		case 4: return qu_int_new (e.e_min);
		case 5: return qu_int_new (e.e_sec);
		case 6: return qu_int_new (e.e_usec);
		case 7: return qu_int_new (e.e_dst);
		case 8: return qu_int_new (e.e_wday);
		case 9: return qu_int_new (e.e_yday);
	}
	/*begin*/
	{
		qu_any arg[10] =
		{
			qu_int_new (e.e_year),
			qu_int_new (e.e_mon),
			qu_int_new (e.e_day),
			qu_int_new (e.e_hour),
			qu_int_new (e.e_min),
			qu_int_new (e.e_sec),
			qu_int_new (e.e_usec),
			qu_int_new (e.e_dst),
			qu_int_new (e.e_wday),
			qu_int_new (e.e_yday)
		};
		return qu_record_new (TmStruct, 10, arg);
	}
}



/*) sub localtime (timeval: Real = Time.now, *index) -> Any
	Get the local time of the time value v{timeval}.

	If v{index} is not specified then return a #{Tm} Record. Otherwise
	it must be an Integer in 0..6 to get a particular field.

	@exc
	.EValue
	.EIndex

	@xmp
		use Time

		t = Time.localtime
		println (t)
		println (t->year)
		println (Time.localtime (Time.now, 0))

	@out
		[2005, 6, 25, 3, 6, 12, 134780282, 0, 1, 205]
		2005
		2005
*/
static qu_any Time_localtime (int argc, qu_any *argv)
{
	QU_RETURN (_time_lgtime (0, argc, argv));
}



/*) sub gmtime (timeval: Real = Time.now, *index) -> Any
	Get the GM time of the time value v{timeval}.

	If v{index} is not specified then return a #{Tm} Record. Otherwise
	it must be an Integer in 0..6 to get a particular field.

	@exc
	.EValue
	.EIndex

	@xmp
		use Time

		t = Time.gmtime
		println (t)
		println (t->year)
		println (Time.gmtime (Time.now, 0))

	@out
		[2005, 6, 24, 20, 6, 33, 134780282, 0, 0, 204]
		2005
		2005
*/
static qu_any Time_gmtime (int argc, qu_any *argv)
{
	QU_RETURN (_time_lgtime (1, argc, argv));
}




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	scan & build
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/* return 1 if pattern matches, 0 otherwise */
static int _ispattern (const char *s, const char *format)
{
	register char f, c;

	while ((f = *format++) != '\0')
	{
		if ((c = *s++) == '\0')
			return 0;

		switch (f)
		{
			case '%':
				if (qu_ascii_isdigit (c))
					continue;
				return 0;

			case '$':
				if (qu_ascii_isalpha (c))
					continue;
				return 0;

			case '!':
				if ((qu_ascii_isdigit (c)) || (c == ' '))
					continue;
				return 0;

			default:
				if (f != c)
					return 0;
		}
	}
	return 1;
}



/*) sub value (x, defval: Float = 0.0) -> Float
	Convert v{x} to a time value.

	v{x} may be a string or a #{Tm}.

	If the format is not understood, if v{defval} is supplied then it is returned,
	otherwise ESyntax is thrown.

	Formats understood:
	@ul
	. RFC-822 : Sun, 06 Nov 1994 08:49:37 GMT
	. RFC-850 : Sunday, 06-Nov-94 08:49:37 GMT
	. ANSI C asctime() : Sun Nov  6 08:49:37 1994
	. extended: basically d-m-y h:m:s am/pm GMT

	@exc
	.EValue
	.ESyntax format is not understood

	@xmp
		use Time

		v = Time.value ('9-Jun-2004')
		println (v)
		println (Time.year (v))
		println (Time.mon  (v))
		println (Time.day  (v))

	@out
		1086714000
		2004
		5
		9
*/
static qu_float Time_value (int argc, qu_any *argv)
{
	time_s e;
	double v;
	int    g = 0;

	QU_ARG_VCOUNT (1, 2);

	if (!qu_string_is (argv[0]))
	{
		qu_record x;
		QU_ARG_vrecord (0, x, TmStruct);
		_rec2struct (x, &e);
	}
	else
	{
		char *h, *m, *s = ((qu_string) argv[0])->str;

		e.e_usec = 0;

#define K(c) ((s[c] - '0') * 10 + (s[c+1] - '0'))
#define Y  e.e_year
#define M  e.e_mon
#define D  e.e_day
#define H  e.e_hour
#define N  e.e_min
#define S  e.e_sec

#define skipspace_(s)  QU_BEGIN while (qu_ascii_isspace (*(s))) (s)++; QU_END
#define skipto_(s, is) QU_BEGIN while (*(s)) { if (is (*(s))) break; (s)++; } QU_END

		skipspace_ (s);
		if (qu_ascii_isdigit (*s))
		{
			/*
				A bit more flexible format.
				Never allow the idiotic 10-11-03 notation!
				10...nov...[1995...[01...[19...[59...[GMT]]]]] */

			D = *s++ - '0';
			if (qu_ascii_isdigit (*s))
				D = D * 10 + *s++ - '0';

			skipto_ (s, qu_ascii_isalpha);
			m = (char *) s;
			while (qu_ascii_isalpha (*s))
				s++;
			if (m == s)
				goto nope;

			skipto_ (s, qu_ascii_isdigit);

			Y = 0;
			if (qu_ascii_isdigit (*s))
			{
				Y = *s++ - '0';
				if (qu_ascii_isdigit (*s))
				{
					Y = Y * 10 + *s++ - '0';
					if (qu_ascii_isdigit (*s))
					{
						Y = Y * 10 + *s++ - '0';
						if (!qu_ascii_isdigit (*s))
							goto nope;
						Y = Y * 10 + *s++ - '0';
					}
				}
			}
			if (Y == 0)
			{
				time_s w;

				_val2struct (_time_now (), &w);
				Y = w.e_year + 1900;
			}

			S = N = H = 0;

			if (qu_ascii_isdigit (*s))
			{
				H = *s++ - '0';
				if (qu_ascii_isdigit (*s))
					H = H * 10 + *s++ - '0';
				skipto_ (s, qu_ascii_isdigit);
				if (qu_ascii_isdigit (*s))
				{
					N = *s++ - '0';
					if (qu_ascii_isdigit (*s))
						N = N * 10 + *s++ - '0';
					skipto_ (s, qu_ascii_isdigit);
					if (qu_ascii_isdigit (*s))
					{
						S = *s++ - '0';
						if (qu_ascii_isdigit (*s))
							S = S * 10 + *s++ - '0';
					}
				}
				if ((strchr (s, 'P') != NULL) || (strchr (s, 'p') != NULL))
				{
					if (H < 12)
						H += 12;
				}
				else if ((strchr (s, 'A') != NULL) || (strchr (s, 'a') != NULL))
				{
					if (H == 12)
						H -= 12;
				}
			}
			g = ((strchr (s, 'G') != NULL) || (strchr (s, 'g') != NULL));
		}
		else
		{
			s = strchr (s, ' ');
			if (s == NULL)
				goto nope;

			++s;

			if (_ispattern (s, "%% $$$ %%%% %%:%%:%% GMT"))
			{
				/* RFC 822
				Sun, 06 Nov 1994 08:49:37 GMT */

				g = 1;
				h = (char *) s + 12;
				m = (char *) s + 3;
				D = K(0);
				Y = K(7) * 100 + K(9);
			}
			else if (_ispattern (s, "%%-$$$-%% %%:%%:%% GMT"))
			{
				/* RFC 850
				Sunday, 06-Nov-94 08:49:37 GMT */

				g = 1;
				h = (char *) s + 10;
				m = (char *) s + 3;
				D = K(0);
				if ((Y = K(7)) < 70)
					Y += 1900;
				else
					Y += 2000;
			}
			else if (_ispattern (s, "$$$ !% %%:%%:%% %%%%"))
			{
				/* ANSI C asctime()
				Sun Nov  6 08:49:37 1994 */

				g = 0;
				h = (char *) s + 7;
				m = (char *) s;
				D = (s[4] == ' ') ? (s[5] - '0') : K(4);
				Y = K(16) * 100 + K(18);
			}
			else
				goto nope;

			s = h;
			H = K(0);
			N = K(3);
			S = K(6);
		}

		switch (qu_ascii_tolower (m[0]))
		{
			case 'j': M = (qu_ascii_tolower (m[1]) == 'a') ? 0 : (qu_ascii_tolower (m[2]) == 'n') ? 5 : 6; break;
			case 'a': M = (qu_ascii_tolower (m[1]) == 'p') ? 3 : 7; break;
			case 'm': M = (qu_ascii_tolower (m[2]) == 'r') ? 2 : 4; break;
			case 's': M =  8; break;
			case 'f': M =  1; break;
			case 'o': M =  9; break;
			case 'n': M = 10; break;
			case 'd': M = 11; break;
			default : goto nope;
		}
	}

	v = _struct2val (&e) - (g ? timezone : 0);
	return _val2obj (v);

  nope:
	if (argc > 1)
	{
		qu_float x;
		QU_ARG_VTYPE (1, x, qu_float);
		return x;
	}
	QU_THROW (QuESyntax);

#undef S
#undef N
#undef H
#undef D
#undef M
#undef Y
#undef I

}



/*) sub make (y = year, m = mon, d = day, h = 0, n = 0, s = 0, u = 0) -> Float
	Build a time value from the arguments.

	Missing argument means to use the value of the current time.

	@xmp
		use Time
		println (Time.make (2004, 9, 1))
		println (Time.make (d:1 y:2004 m:9))

	@out
		1088615340
		1088615340
*/
static qu_float Time_make (int argc, qu_any *argv)
{
	static char *argx[] =
	{
		":",
		"i", "y",
		"i", "m",
		"i", "d",
		"i", "h",
		"i", "m",
		"i", "s",
		"i", "u",
		NULL
	};

	time_s e;

	_val2struct (_time_now (), &e);
	e.e_hour = e.e_min = e.e_sec = e.e_usec = 0;
	QU_ARG_XPARSE (&e.e_year, &e.e_mon, &e.e_day, &e.e_hour, &e.e_min, &e.e_sec, &e.e_usec);
	return _val2obj (_struct2val (&e));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/* the buffer size required to format a HTTP time */
#define TIME_WEBSIZE 64


/* month names for Time_web and Time_smtp */
static const char Months3[12][4] =
{
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

/* weekday names for Time_web and Time_smtp */
static const char Days3[7][4] =
{
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};


/*) sub format (fmt, x = Time.now) -> String
	Format the time v{x} using a specified format.

	v{x} may be a time value or a #{Tm}.

	Refer to C{man strftime} for format details.

	@exc
	.EValue

	@xmp
		use Time
		println (Time.format ('%d-%B-%Y', Time.now))

	@out
		25-July-2005
*/
static qu_string Time_format (int argc, qu_any *argv)
{
	time_s    e;
	char      b [4096];
	qu_string fmt;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, fmt, qu_string);

	if (argc > 1)
	{
		if (qu_record_is (argv[1]))
		{
			qu_record x;
			QU_ARG_vrecord (1, x, TmStruct);
			_rec2struct (x, &e);
		}
		else
		{
			double v;
			QU_ARG_vdouble (1, v);
			_val2struct (v, &e);
		}
	}
	else
	{
		_val2struct (_time_now (), &e);
	}
	e.e_year -= 1900;
	strftime (b, sizeof (b), ((qu_string) fmt)->str, (struct tm *) &e);
	b [sizeof (b) - 1] = '\0';
	return qu_string_newz (b);
}



/*) sub web (x = Time.now) -> String
	Return a formatted HTTP time.

	v{x} may be a time value or a #{Tm}.

	@exc
	.EValue

	@xmp
		use Time
		println (Time.web)

	@out
		Sun, 24 Jul 2005 20:10:41 GMT
*/
static qu_string Time_web (int argc, qu_any *argv)
{
	time_s e;
	double v;
	char   buf [TIME_WEBSIZE + 1];

	if (argc == 0)
	{
		v = _time_now ();
	}
	else
	{
		if (qu_record_is (argv[0]))
		{
			qu_record x;
			QU_ARG_vrecord (0, x, TmStruct);
			_rec2struct (x, &e);
			v = _struct2val (&e);
		}
		else
		{
			QU_ARG_vdouble (0, v);
		}
	}
	_val2gm (v, &e);
	snprintf (buf, TIME_WEBSIZE, "%s, %.2d %s %d %.2d:%.2d:%.2d GMT",
	          Days3[e.e_wday], e.e_day, Months3[e.e_mon],
	          e.e_year, e.e_hour, e.e_min, e.e_sec);
	buf [TIME_WEBSIZE] = '\0';
	return qu_string_newz (buf);
}



/*) sub smtp (timeval: Real = Time.now) -> String
	Return a formatted SMTP time.

	@exc
	.EValue

	@xmp
		use Time
		println (Time.smtp)

	@out
		Mon, 25 Jul 2005 03:10 +0700
*/
static qu_string Time_smtp (int argc, qu_any *argv)
{
	time_s e;
	double v;
	char   buf [TIME_WEBSIZE + 1];

	if (argc == 0)
	{
		v = _time_now ();
	}
	else
	{
		if (qu_record_is (argv[0]))
		{
			qu_record x;
			QU_ARG_vrecord (0, x, TmStruct);
			_rec2struct (x, &e);
			v = _struct2val (&e);
		}
		else
		{
			QU_ARG_vdouble (0, v);
		}
	}
	_val2local (v, &e);
	snprintf (buf, TIME_WEBSIZE, "%s, %d %s %d %02d:%02d %+03d00",
			  Days3[e.e_wday], e.e_day, Months3[e.e_mon],
	          e.e_year, e.e_hour, e.e_min, (int) - timezone / 3600);
	buf [TIME_WEBSIZE] = '\0';
	return qu_string_newz (buf);
}




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	sleep (interrupted) and wait (uninterrupted)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

#define TimeS2M 1000L        /* sec   -> milli    */
#define TimeS2U 1000000L     /* sec   -> micro    */
#define TimeS2N 1000000000L  /* sec   -> nano     */
#define TimeM2U 1000L        /* milli -> microsec */
#define TimeM2N 1000000L     /* milli -> nanosec  */
#define TimeU2N 1000L        /* micro -> nano     */


static qu_any _time_pause (qu_any len, int what, int nobrk)
{
	struct timespec req, rem;
	long n;
	int  r;

	QU_ARG_int (len, n);
	QU_ARG_CHECK (n > 0);

  __qu_blocking_begin;
	switch (what)
	{
		case 3: /*sec*/
			req.tv_sec  = n;
			req.tv_nsec = 0;
			break;

		case 2: /*milli*/
			req.tv_sec  = n / TimeS2M;
			req.tv_nsec = (n % TimeS2M) * TimeM2N + 1;
			break;

		case 1: /*micro*/
			req.tv_sec  = n / TimeS2U;
			req.tv_nsec = (n % TimeS2U) * TimeS2N + 1;
			break;

		default: /*nano*/
			req.tv_sec  = 0;
			req.tv_nsec = n;
	}
	for (;;)
	{
		r = nanosleep (&req, &rem);
		if (r < 0)
		{
			QU_ARG_CHECK (errno == EINTR);
			if (nobrk && (rem.tv_sec > 0 || rem.tv_nsec > 0))
			{

				req.tv_sec  = rem.tv_sec;
				req.tv_nsec = rem.tv_nsec;
				continue;
			}
			switch (what)
			{
				case 3: /*sec*/
					n = rem.tv_sec;
					break;

				case 2: /*milli*/
					n = rem.tv_sec * TimeS2M + rem.tv_nsec / TimeM2N;
					break;

				case 1: /*micro*/
					n = rem.tv_sec * TimeS2U + rem.tv_nsec / TimeU2N;
					break;

				default: /*nano*/
					n = rem.tv_nsec;
			}
		  __qu_blocking_end;
			return qu_int_new (n);
		}
		break;
	}
  __qu_blocking_end;
	return QuInt0;
}



/*) sub wait (secs: Integer) -> Null
	Do an uninterrupted sleep for the specified amount of seconds.

	v{secs} must be a positive integer value within the Int range.

	@exc
	.EValue
*/
static qu_null Time_wait (qu_any secs)
{
	QU_RETURN (_time_pause (secs, 3, 1));
}



/*) sub mwait (millisecs: Integer) -> Null
	Do an uninterrupted sleep for the specified amount of milli seconds.

	v{millisecs} must be a positive integer value within the Int range.

	@exc
	.EValue
*/
static qu_null Time_mwait (qu_any millisecs)
{
	QU_RETURN (_time_pause (millisecs, 2, 1));
}



/*) sub uwait (microsecs: Integer) -> Null
	Do an uninterrupted sleep for the specified amount of micro seconds.

	v{microsecs} must be a positive integer value within the Int range.

	@exc
	.EValue
*/
static qu_null Time_uwait (qu_any microsecs)
{
	QU_RETURN (_time_pause (microsecs, 1, 1));
}



/*) sub nwait (nanosecs: Integer) -> Null
	Do an uninterrupted sleep for the specified amount of nano seconds.

	v{nanosecs} must be a positive integer value within the Int range.

	@exc
	.EValue
*/
static qu_null Time_nwait (qu_any secs)
{
	QU_RETURN (_time_pause (secs, 0, 1));
}



/*) sub sleep (secs: Integer) -> Int
	Do an interrupted sleep for the specified amount of seconds.

	v{secs} must be a positive integer value within the Int range.

	Return zero or the time left after an interruption.

	@exc
	.EValue
*/
static qu_int Time_sleep (qu_int secs)
{
	QU_RETURN (_time_pause (secs, 3, 0));
}



/*) sub msleep (millisecs: Integer) -> Int
	Do an interrupted sleep for the specified amount of milli seconds.

	v{millisecs} must be a positive integer value within the Int range.

	Return zero or the time left after an interruption.

	@exc
	.EValue
*/
static qu_int Time_msleep (qu_int millisecs)
{
	QU_RETURN (_time_pause (millisecs, 2, 0));
}



/*) sub usleep (microsecs: Integer) -> Int
	Do an interrupted sleep for the specified amount of micro seconds.

	v{microsecs} must be a positive integer value within the Int range.

	Return zero or the time left after an interruption.

	@exc
	.EValue
*/
static qu_int Time_usleep (qu_int microsecs)
{
	QU_RETURN (_time_pause (microsecs, 1, 0));
}



/*) sub nsleep (nanosecs: Integer) -> Int
	Do an interrupted sleep for the specified amount of nano seconds.

	v{nanosecs} must be a positive integer value within the Int range.

	Return zero or the time left after an interruption.

	@exc
	.EValue
*/
static qu_int Time_nsleep (qu_int secs)
{
	QU_RETURN (_time_pause (secs, 0, 0));
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Time_init (qu_module m)
{
	time_t t = time (NULL);
	mktime (localtime (&t));


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Tm
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	TmStruct = qu_struct_new_static (10,
		"year" , QuIntClass,
		"mon"  , QuIntClass,
		"day"  , QuIntClass,
		"hour" , QuIntClass,
		"min"  , QuIntClass,
		"sec"  , QuIntClass,
		"usec" , QuIntClass,
		"isdst", QuIntClass,
		"wday" , QuIntClass,
		"yday" , QuIntClass);

	qu_module_add_const (m, "Tm", TmStruct);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		get field value
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "year" , Time_year , -1);
	qu_module_add_function (m, "mon"  , Time_mon  , -1);
	qu_module_add_function (m, "day"  , Time_day  , -1);
	qu_module_add_function (m, "hour" , Time_hour , -1);
	qu_module_add_function (m, "min"  , Time_min  , -1);
	qu_module_add_function (m, "sec"  , Time_sec  , -1);
	qu_module_add_function (m, "usec" , Time_usec , -1);
	qu_module_add_function (m, "isdst", Time_isdst, -1);
	qu_module_add_function (m, "wday" , Time_wday , -1);
	qu_module_add_function (m, "yday" , Time_yday , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		current time
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "now"  , Time_now  , -1);
	qu_module_add_function (m, "date" , Time_date ,  0);
	qu_module_add_function (m, "clock", Time_clock,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		info
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "timezone", Time_timezone,  0);
	qu_module_add_function (m, "tzname"  , Time_tzname  , -1);
	qu_module_add_function (m, "daylight", Time_daylight,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		counting
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "ydays", Time_ydays, -1);
	qu_module_add_function (m, "mdays", Time_mdays, -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		local and GM
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "localtime", Time_localtime, -1);
	qu_module_add_function (m, "gmtime"   , Time_gmtime   , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		scan & build
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "value", Time_value, -1);
	qu_module_add_function (m, "make" , Time_make , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		format
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "format", Time_format, -1);
	qu_module_add_function (m, "web"   , Time_web   , -1);
	qu_module_add_function (m, "smtp"  , Time_smtp  , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		sleep (interrupted) and wait (uninterrupted)
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "wait"  , Time_wait  , 1);
	qu_module_add_function (m, "mwait" , Time_mwait , 1);
	qu_module_add_function (m, "uwait" , Time_uwait , 1);
	qu_module_add_function (m, "nwait" , Time_nwait , 1);
	qu_module_add_function (m, "sleep" , Time_sleep , 1);
	qu_module_add_function (m, "msleep", Time_msleep, 1);
	qu_module_add_function (m, "usleep", Time_usleep, 1);
	qu_module_add_function (m, "nsleep", Time_nsleep, 1);

	return 1;
}
