/* Uri.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Uri
	This module provides a number of functions to deal with
	URI (I{Uniform Resource Index}) per RFC-2396.
*/


/*) struct URI
		scheme : String   # scheme, e.g. "HTTP"
		host   : String   # host name
		port   : Int      # port number
		path   : String   # resource path
		query  : String   # query string
		frag   : String   # fragment string
	;;
	This represents an URI.

	Note that we deliberatly do not provide the User and Password fields since
	they only introduce security holes.
*/
static qu_struct UriStruct;



/*----------------------------------------------------------------------------
	field names
----------------------------------------------------------------------------*/

static qu_string Sscheme;
static qu_string Shost;
static qu_string Sport;
static qu_string Spath;
static qu_string Squery;
static qu_string Sfrag;



/*----------------------------------------------------------------------------
	alphabet
----------------------------------------------------------------------------*/

static unsigned char Escaped [256] =
{
	/*   0 */    1,   1,   2,   3,   4,   5,   6,   8,
	/*   8 */    9,'\t','\n','\v','\f','\r',  14,  15,
	/*  16 */   16,  17,  18,  19,  20,  21,  22,  23,
	/*  24 */   24,  25,  26,  27,  28,  29,  30,  31,
	/*  32 */  ' ',   0, '"', '#', '$', '%', '&',   0,
	/*  40 */    0,   0,   0, '+',   0,   0,   0, '/',
	/*  48 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  56 */    0,   0, ':', ';', '<', '=', '>', '?',
	/*  64 */  '@',   0,   0,   0,   0,   0,   0,   0,
	/*  72 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  80 */    0,   0,   0,   0,   0,   0,   0,   0,
	/*  88 */    0,   0,   0, '[','\\', ']', '^',   0,
	/*  96 */  '`',   0,   0,   0,   0,   0,   0,   0,
	/* 104 */    0,   0,   0,   0,   0,   0,   0,   0,
	/* 112 */    0,   0,   0,   0,   0,   0,   0,   0,
	/* 120 */    0,   0,   0, '{', '|', '}',   0, 127,
	/* 128 */  128, 129, 130, 131, 132, 133, 134, 135,
	/* 136 */  136, 137, 138, 139, 140, 141, 142, 143,
	/* 144 */  144, 145, 146, 147, 148, 149, 150, 151,
	/* 152 */  152, 153, 154, 155, 156, 157, 158, 159,
	/* 160 */  160, 161, 162, 163, 164, 165, 166, 167,
	/* 168 */  168, 169, 170, 171, 172, 173, 174, 175,
	/* 176 */  176, 177, 178, 179, 180, 181, 182, 183,
	/* 184 */  184, 185, 186, 187, 188, 189, 190, 191,
	/* 192 */  192, 193, 194, 195, 196, 197, 198, 199,
	/* 200 */  200, 201, 202, 203, 204, 205, 206, 207,
	/* 208 */  208, 209, 210, 211, 212, 213, 214, 215,
	/* 216 */  216, 217, 218, 219, 220, 221, 222, 223,
	/* 224 */  224, 225, 226, 227, 228, 229, 230, 231,
	/* 232 */  232, 233, 234, 235, 236, 237, 238, 239,
	/* 240 */  240, 241, 242, 243, 244, 245, 246, 247,
	/* 248 */  248, 249, 250, 251, 252, 253, 254, 255
};

#define ESCAPED(c) Escaped [255 & (c)]



/*----------------------------------------------------------------------------
	basics
----------------------------------------------------------------------------*/

static long _uri_escape_len (char *s, long l)
{
	long n = 0;

	while (l--)
	{
		n += ESCAPED (*s) ? 3 : 1;
		if (n <= 0)
			return -1;
		++s;
	}
	return n;
}



static long _uri_escape (char *z, long l, char *b, int query)
{
	long n = 0;
	while (l--)
	{
		if (*z == ' ' && query)
		{
			*b++ = '+';
			++z;
			++n;
		}
		else if (ESCAPED (*z))
		{
			*b++ = '%';
			*b++ = QU_ASCII_DIGITU [(255 & *z) >> 4];
			*b++ = QU_ASCII_DIGITU [(255 & *z) & 0xF];
			++z;
			n += 3;
		}
		else
		{
			*b++ = *z++;
			++n;
		}
	}
	*b = '\0';
	return n;
}


static long _uri_unescape_len (char *z, long l)
{
	long n = 0;

	while (l)
	{
		if (z[0] == '%' && l > 2 && qu_ascii_isxdigit (z[1]) && qu_ascii_isxdigit (z[2]))
		{
			l -= 3;
			z += 3;
		}
		else
		{
			--l;
			++z;
		}
		++n;
	}
	return n;
}



static long _uri_unescape (char *z, long l, char *b, int query)
{
	long n = 0;

	while (l)
	{
		if (z[0] == '%' && l > 2 && qu_ascii_isxdigit (z[1]) && qu_ascii_isxdigit (z[2]))
		{
			*b++ = qu_ascii_hexval (z[1]) * 16 + qu_ascii_hexval (z[2]);
			l -= 3;
			z += 3;
		}
		else
		{
			if (*z == '+' && query)
			{
				*b++ = ' ';
				++z;
			}
			else
			{
				*b++ = *z++;
			}
			--l;
		}
		++n;
	}
	*b = '\0';
	return n;
}



static qu_string _uri_string_result (qu_string r, long l)
{
	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
		return qu_string_byte (r->str[0]);

	r->len = l;
	r->str[l] = '\0';
	return r;
}



static char *_uri_string_char (char *z, long l, char c)
{
	while (l--)
	{
		if (*z == c)
			return z;
		++z;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	functions
----------------------------------------------------------------------------*/

/*) sub escape (s: String)
	Escape string data to include in URI.

	See #{unescape} for the reverse.

	If no escaping is necessary then v{s} is returned.
	If escaping is necessary then a String is returned.

	@exc
	.EArgValue
	.EMemory the resulting string is too big

	@xmp
		use Uri
		println (Uri.escape ('a#b c'))

	@out
		a%23b+c
*/
qu_string Uri_escape (qu_string s)
{
	qu_string  r;
	char      *z;
	long       n, l;

	QU_ARG_TYPE (s, qu_string);
	l = s->len;
	z = s->str;
	n = _uri_escape_len (z, l);
	if (n == l && _uri_string_char (z, l, ' ') == NULL)
		return s;
	QU_MEM_CHECK (0 < n && n < QU_MEM_MAX);
	r = qu_string_newl (n);
	n = _uri_escape (z, l, r->str, 1);
	return (qu_string) _uri_string_result (r, n);
}



/*) sub unescape (s: String) -> String
	Does the reverse of #{escape}, i.e. unescape the string.

	If no escaping is necessary then v{s} is returned.
	If escaping is necessary then a String is returned.

	@exc
	.EArgValue

	@xmp
		use Uri
		println (Uri.unescape ('a%23b+c'))

	@out
		a#b c
*/
qu_string Uri_unescape (qu_string s)
{
	qu_string  r;
	char      *z;
	long       n, l;

	QU_ARG_TYPE (s, qu_string);
	l = s->len;
	z = s->str;
	n = _uri_unescape_len (z, l);
	if (n < 2)
		return s;
	QU_MEM_CHECK (0 < n && n < QU_MEM_MAX);
	r = qu_string_newl (n);
	n = _uri_unescape (z, l, r->str, 1);
	return (qu_string) _uri_string_result (r, n);
}



/*) sub dequery (s: String, d: Dict = [:]) -> Dict
	Parse a query string into a Dict.

	v{s} must be a string in the escaped form.

	If a field with the same name exists, then the value will be an Array.
	If a field lacks a value then the value is set to an empty string.

	You can optionally specify an existing Dict to merge with the result.

	@exc
	.EArgValue

	@xmp
		use Uri

		s = 'a=foo+bar&b=2&a=3'
		println (Uri.dequery (s, [c:3]))

	@out
		['a':['foo bar', '3'], 'b':'2', 'c':3]
*/
qu_dict Uri_dequery (int argc, qu_any *argv)
{
	qu_string s;
	qu_dict   d;
	char     *z;
	long      l;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, s, qu_string);

	if (argc > 1)
	{
		QU_ARG_VTYPE (1, d, qu_dict);
	}
	else
	{
		d = qu_dict_new (4);
	}

	l = s->len;
	z = s->str;

	while (l > 0)
	{
		char      *p = _uri_string_char (z, l, '&');
		char      *q;
		long       m, n, r;
		qu_string  k, v;
		qu_any     x;

		if (p == NULL)
			p = z + l;

		m = p - z;
		l -= (m + 1);
		q = _uri_string_char (z, m, '=');
		if (q)
		{
			m = q++ - z;
			r = p - q;
			n = _uri_unescape_len (q, r);
			v = qu_string_newl (n);
			n = _uri_unescape (q, r, v->str, 1);
			v = _uri_string_result (v, n);
		}
		else
		{
			v = QuStringEMPTY;
		}

		n = _uri_unescape_len (z, m);
		k = qu_string_newl (n);
		n = _uri_unescape (z, m, k->str, 1);
		k = _uri_string_result (k, n);

		x = qu_dict_getitem (d, k);
		if (x)
		{
			if (qu_array_is (x))
				QU_ENSURE (qu_array_push (x, v));
			else
				qu_dict_setitem (d, k, qu_array_new_ex (2, x, v));
		}
		else
		{
			qu_dict_setitem (d, k, v);
		}
		z = p + 1;
	}
	return d;
}



/*) sub enquery (d:Dict) -> String
	Encode a query to be included in an URI.

	If a field is an Array then it will be repeated for each of the items.

	@exc
	.EMemory
	.EArgValue

	@xmp
		use Uri

		x = ['a':['foo bar', '3'], 'b':'2', 'c':3]
		println (Uri.enquery (x))

	@out
		a=foo+bar&a=3&b=2&c=3
*/
qu_string Uri_enquery (qu_dict d)
{
	qu_any    k, v, x;
	long      b, i, j, l;
	qu_string r;

	r = qu_string_buffer_new (d->len);
	b = i = 0;
	l = d->len;
	while ((v = qu_dict_walk (d, &k, &b, &i)) != NULL)
	{
		k = Uri_escape ((qu_string) qu_sys_string (k));
		QU_ENSURE (k);

		if (qu_array_is (v))
		{
			for (j = 0; j < ((qu_array) v)->len; ++j)
			{
				x = ((qu_array) v)->objs[j];
				QU_ENSURE (qu_string_buffer_puts (r, k));
				QU_ENSURE (qu_string_buffer_putc (r, '='));
				x = Uri_escape ((qu_string) qu_sys_string (x));
				QU_ENSURE (x);
				QU_ENSURE (qu_string_buffer_puts (r, x));
				if (j < ((qu_array) v)->len - 1)
					QU_ENSURE (qu_string_buffer_putc (r, '&'));
			}
		}
		else
		{
			x = Uri_escape ((qu_string) qu_sys_string (v));
			QU_ENSURE (x);
			QU_ENSURE (qu_string_buffer_puts (r, k));
			QU_ENSURE (qu_string_buffer_putc (r, '='));
			QU_ENSURE (qu_string_buffer_puts (r, x));
		}

		if (--l > 0)
			QU_ENSURE (qu_string_buffer_putc (r, '&'));
	}
	return qu_string_buffer_return (r);
}



/*) sub split (s:String, i = -1) -> Record
	Split URI into its components.

	See #{make} for the reverse.

	Return a -{Record} of #{URI}.

	If v{i} is not -1 then it must be between 0..5 to get a particular field:
	@ul
	.0 : scheme
	.1 : host
	.2 : port
	.3 : path
	.4 : query
	.5 : fragment

	@p
	The user and password parts are thrown away since they serve only as
	security holes and must never be used in the first place.

	@exc
	.EArgValue

	@xmp
		use Uri

		r = Uri.split ('http://www.foo.bar:80/bar/baz.html?a=1&b=2#fox')
		println (r)

	@out
		[scheme:'http', host:'www.foo.bar', port:80, path:'/bar/baz.html', query:'a=1&b=2', fragment:'fox']
*/
/*
	scheme://user_info@host:port/path?query#fragment
	scheme://reg_name/path?query#fragment
	scheme:/path?query#fragment
	//user_info@host:port/path?query#fragment
	//reg_name/path?query#fragment
	/path?query#fragment
	path?query#fragment

	"http://www%2Esome%2Eorg/a%2Fb%2F/c"
	host = "www%2Esome%2Eorg" -> "www.some.org"
	path = "/a%2Fb%2F/c"      -> "/a/b/c"

	!!! "http://www%2esome%2eorg%3a%38%30%2fa%2fb"
	!!! host = "www.some.org:80/a/b"

	user_info is intentionally not returned. It's stupid!
*/
static int _do_split (char *u, char **scheme, char **host, int *port,
                      char **path, char **query, char **frag)
{
	register char *i;
	int abs = 0;

#define skipto_(s, c) QU_BEGIN while (*(s) && (*(s) != (c))) (s)++; QU_END

	*scheme = NULL;
	*host   = NULL;
	*path   = NULL;
	*query  = NULL;
	*frag   = NULL;
	*port   = 0;

	i = u;
	skipto_(i, '#');
	if (*i)
	{
		*i = '\0';
		*frag = i + 1;
	}

	i = u;
	if (qu_ascii_isalpha (*i))
	{
		i++;
		while (qu_ascii_isalnum (*i)
			|| (*i == '+') || (*i == '-') || (*i == '.'))
			++i;

		if (*i == ':')
		{
			*i = '\0';
			*scheme = u;
			u = i + 1;
		}
	}

	if (u[0] == '/')
	{
		if (u[1] != '/')
		{
			abs = 1;
			*path = u;
		}
		else
		{
			u += 2;
			i = u;
			skipto_(i, '/');
			if (*i)
			{
				*i = '\0';
				*path = i + 1;
			}

			i = u;
			skipto_(i, '@');
			if (*i)
			{
				*i = '\0';
				u = i + 1;
			}

			if (*u)
			{
				*host = u;
				skipto_(u, ':');
				if (*u)
				{
					int p;

					*u = '\0';
					p = strtol (u + 1, &u, 0);
					/* Port 0 is never valid */
					if (p > 0 && p <= 65535)
						*port = p;
				}
			}
		}
	}
	else
	{
		abs = 0;
		*path = u;
	}

	if (*path != NULL)
	{
		i = *path;
		skipto_(i, '?');
		if (*i)
		{
			*i = '\0';
			*query = i + 1;
		}
	}
	return abs;
}



static qu_string _uri_component (char *z, int abs)
{
	if (z)
	{
		long l = strlen (z);

		if (l)
		{
			long n = _uri_unescape_len (z, l);

			assert (n >= 0);
			if (n)
			{
				qu_string s = qu_string_newl (n + abs);

				if (abs)
				{
					s->str[0] = '/';
					n = 1 + _uri_unescape (z, l, s->str + 1, 0);
				}
				else
				{
					n = _uri_unescape (z, l, s->str, 0);
				}

				if (n == 1)
					return qu_string_byte (s->str[0]);

				else if (n)
				{
					s->len = n;
					s->str[n] = '\0';
					return s;
				}
			}
		}
	}
	return 0;
}


qu_any Uri_split (int argc, qu_any *argv)
{
	char      *scheme, *host, *path, *query, *frag;
	int        port, abs, i = -1;
	qu_stringy x;

	QU_ARG_VCOUNT (0, 1);
	QU_ARG_VTYPE  (0, x, qu_stringy);

	if (argc > 1)
	{
		QU_ARG_vint (0, i);
		QU_ARG_CHECK (i == -1 || (0 <= i && i <= 5));
	}

	x = qu_string_new (x->str, x->len);
	abs = _do_split (x->str, &scheme, &host, &port, &path, &query, &frag);

	switch (i)
	{
		case 0:
			x = _uri_component (scheme, 0);
			return x ? x : QuStringEMPTY;

		case 1:
			x = _uri_component (host, 0);
			return x ? x : QuStringEMPTY;

		case 2:
			return (port < 0) ? QuInt_1 : qu_int_new (port);

		case 3:
			x = _uri_component (path, !abs);
			return x ? x : QuStringEMPTY;

		case 4:
			if (query)
				return qu_string_newz (query);
			return QuStringEMPTY;

		case 5:
			x = _uri_component (frag, 0);
			return x ? x : QuStringEMPTY;

		case -1:
		{
			qu_any arg[6] =
			{
				QuStringEMPTY,
				QuStringEMPTY,
				QuInt0,
				QuStringEMPTY,
				QuStringEMPTY,
				QuStringEMPTY
			};

			if (scheme)
			{
				x = _uri_component (scheme, 0);
				if (x)
					arg[0] = x;
			}

			if (host)
			{
				x = _uri_component (host, 0);
				if (x)
					arg[1] = x;
			}

			if (port > 0)
				arg[2] = qu_int_new (port);

			x = _uri_component (path, !abs);
			if (x)
				arg[3] = x;

			if (query)
				arg[4] = qu_string_newz (query);

			x = _uri_component (frag, 0);
			if (x)
				arg[5] = x;

			return qu_record_new (UriStruct, 6, arg);
		}
	}
	QU_THROW (QuEIndex);
}



/*) sub make (*arg) -> String
	Does the reverse of #{split}, i.e. build an URI from the components.

	v{arg} must be a #{URI} Record object, a Dict with the same items,
	an Array with items as URI, arguments as URI, or an object with properties
	as URI.

	If v{port] is 0 or negative then it is not included.

	@exc
	.EArgValue
	.EMemory result string is too big

	@xmp
		use Uri
		s = Uri.uri ('http', 'www.foo.bar', 0, '/bar/baz.html', 'a=1&b=2', 'fox')
		a = s as Array
		d = s as Dict
		println (Uri.make (s))
		println (Uri.make (a))
		println (Uri.make (*a))
		println (Uri.make (d))

	@out
		http://www.foo.bar:80/bar/baz.html?a=1&b=2#fox
		http://www.foo.bar:80/bar/baz.html?a=1&b=2#fox
		http://www.foo.bar:80/bar/baz.html?a=1&b=2#fox
		http://www.foo.bar:80/bar/baz.html?a=1&b=2#fox
*/
qu_string Uri_make (int argc, qu_any *argv)
{
	qu_any      x;
	qu_string   r, scheme, host, path, query, frag;
	char       *z;
	qu_integer  port;
	int         p = 0;
	long        l, n;

	QU_ARG_VMIN (1);
	x = argv[0];

	if (argc == 1)
	{
		if (qu_dict_is (x))
		{
			scheme = qu_dict_getitem (x, Sscheme);
			host   = qu_dict_getitem (x, Shost);
			port   = qu_dict_getitem (x, Sport);
			path   = qu_dict_getitem (x, Spath);
			query  = qu_dict_getitem (x, Squery);
			frag   = qu_dict_getitem (x, Sfrag);
		}
		else if (qu_array_is (x))
		{
			qu_any *q = ((qu_array) x)->objs;
			n = ((qu_array) x)->len;

			scheme = (n < 1) ? (qu_string ) NULL : (qu_string ) q[0];
			host   = (n < 2) ? (qu_string ) NULL : (qu_string ) q[1];
			port   = (n < 3) ? (qu_integer) NULL : (qu_integer) q[2];
			path   = (n < 4) ? (qu_string ) NULL : (qu_string ) q[3];
			query  = (n < 5) ? (qu_string ) NULL : (qu_string ) q[4];
			frag   = (n < 6) ? (qu_string ) NULL : (qu_string ) q[5];
		}
		else if (qu_record_is (x))
		{
			qu_any *q = ((qu_record) x)->objs;

			QU_ARG_CHECK (qu_record_of (x, UriStruct));

			scheme = q[0];
			host   = q[1];
			port   = q[2];
			path   = q[3];
			query  = q[4];
			frag   = q[5];
		}
		else
		{
			qu_class c = qu_classof (x);

			scheme = c->get (x, Sscheme, 1);
			host   = c->get (x, Shost,   1);
			port   = c->get (x, Sport,   1);
			path   = c->get (x, Spath,   1);
			query  = c->get (x, Squery,  1);
			frag   = c->get (x, Sfrag,   1);
		}
	}
	else
	{
		scheme = x;
		host   = (argc < 2) ? (qu_string ) NULL : (qu_string ) argv[1];
		port   = (argc < 3) ? (qu_integer) NULL : (qu_integer) argv[2];
		path   = (argc < 4) ? (qu_string ) NULL : (qu_string ) argv[3];
		query  = (argc < 5) ? (qu_string ) NULL : (qu_string ) argv[4];
		frag   = (argc < 6) ? (qu_string ) NULL : (qu_string ) argv[5];
	}

	if (scheme)
		QU_ARG_TYPE (scheme, qu_string);
	else
		scheme = QuStringEMPTY;

	if (host)
		QU_ARG_TYPE (host, qu_string);
	else
		host = QuStringEMPTY;

	if (port)
		QU_ARG_int (port, p);
	QU_ARG_CHECK (0 <= p && p <= 65535);

	if (path)
		QU_ARG_TYPE (path, qu_string);
	else
		path = QuStringEMPTY;

	if (query)
		QU_ARG_TYPE (query, qu_string);
	else
		query = QuStringEMPTY;

	if (frag)
		QU_ARG_TYPE (frag, qu_string);
	else
		frag = QuStringEMPTY;


	l = 0;
	if (scheme->len)
	{
		n = _uri_escape_len (scheme->str, scheme->len);
		QU_ARG_CHECK (n > 0);
		QU_MEM_CHECK (n < QU_MEM_MAX - l - 3);
		l += n + 3;
	}

	if (host->len)
	{
		n = _uri_escape_len (host->str, host->len);
		QU_ARG_CHECK (n > 0);
		QU_MEM_CHECK (n < QU_MEM_MAX - l);
		l += n;

		if (p)
			l += 6;
	}

	if (path->len)
	{
		n = _uri_escape_len (path->str, path->len);
		QU_ARG_CHECK (n > 0);
		QU_MEM_CHECK (n < QU_MEM_MAX - l - 1);
		l += n + 1;
	}

	if (query->len)
	{
		n = _uri_escape_len (query->str, query->len);
		QU_ARG_CHECK (n >= 0);
		QU_MEM_CHECK (n < QU_MEM_MAX - l - 1);
		l += n + 1;
	}

	if (frag->len)
	{
		n = _uri_escape_len (frag->str, frag->len);
		QU_ARG_CHECK (n >= 0);
		QU_MEM_CHECK (n < QU_MEM_MAX - l - 1);
		l += n + 1;
	}

	r = qu_string_newl (l);
	z = r->str;

	if (scheme->len)
	{
		z += _uri_escape (scheme->str, scheme->len, z, 0);
		*z++ = ':';
		*z++ = '/';
		*z++ = '/';
	}

	if (host->len)
	{
		z += _uri_escape (host->str, host->len, z, 0);
		if (p)
		{
			char   b [8];
			size_t m;

			sprintf (b, ":%d", p);
			m = strlen (b);
			memcpy (z, b, m);
			z += m;
		}
	}

	if (path->len)
	{
		if (scheme->len || host->len)
			*z++ = '/';

		if (path->str[0] == '/')
		{
			z += _uri_escape (path->str + 1, path->len - 1, z, 0);
		}
		else
		{
			z += _uri_escape (path->str, path->len, z, 0);
		}
	}
	else
	{
		*z++ = '/';
	}

	if (query->len)
	{
		*z++ = '?';
		/* z += _uri_escape (query->str, query->len, z); */
		memcpy (z, query->str, query->len);
		z += query->len;
	}

	if (frag->len)
	{
		*z++ = '#';
		z += _uri_escape (frag->str, frag->len, z, 0);
	}

	n = z - r->str;
	assert (n <= l);

	return _uri_string_result (r, n);
}



/*) sub normalize (path: String) -> String
	Normalize a path, i.e. get rid of C{.././} and other weird stuffs.

	C{'..'} and C{'.'} are treated as directory entries but no actual directory
	look-up is made. This is just a string processing.

	@exc
	.EArgValue

	@xmp
		use Uri
		println (Uri.normalize ('.././foo/.././bar/.../...././baz'))

	@out
		bar/.../..../baz
*/
qu_string Uri_normalize (qu_string path)
{
	qu_string  r;
	char      *p;
	int        i, j, k, l;

	QU_ARG_TYPE (path, qu_string);

	if (path->len == 0)
		return QuStringEMPTY;

	r = qu_string_new (path->str, path->len);
	p = r->str;

	for (i = j = 0; p[i] != '\0';)
	{
		if ((p[i] == '/') && (p[i+1] == '/'))
			i++;
		else
			p[j++] = p[i++];
	}
	p[j] = '\0';

	for (i = j = 0; p[i] != '\0';)
	{
		if ((p[i] == '.') && (p[i+1] == '/')
				&& ((i == 0) || (p[i-1] == '/')))
			i += 2;
		else
			p[j++] = p[i++];
	}

	if (((j == 1) && (p[0] == '.'))
			|| ((j > 1) && (p[j-1] == '.') && (p[j-2] == '/')))
		j--;

	p[j] = '\0';

	i = 0;
	while (p[i] != '\0')
	{
		if ((p[i] == '.') && (p[i+1] == '.') && (p[i+2] == '/')
				&& ((i == 0) || (p[i-1] == '/')))
		{
			k = i + 3;
			i -= 2;
			if (i >= 0)
			{
				while ((i >= 0) && (p[i] != '/'))
					i--;
				i++;
			}
			else
				i = 0;

			l = i;
			while ((p[l] = p[k]))
			{
				l++;
				k++;
			}
		}
		else
			++i;
	}

	if ((i == 2) && (p[0] == '.') && (p[1] == '.'))
	{
		p[0] = '\0';
	}
	else if ((i > 2) && (p[i-1] == '.') && (p[i-2] == '.') && (p[i-3] == '/'))
	{
		i = i - 4;
		if (i < 1)
		{
			p[0] = '\0';
		}
		else
		{
			while ((i >= 0) && (p[i] != '/'))
				i--;
			p[i+1] = '\0';
		}
	}
	return _uri_string_result (r, strlen (r->str));
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Uri_init (qu_module m)
{
	UriStruct = qu_struct_new_static (6,
		"scheme", QuStringClass,
		"host"  , QuStringClass,
		"port"  , QuIntClass,
		"path"  , QuStringClass,
		"query" , QuStringClass,
		"frag"  , QuStringClass);

	qu_module_add_const (m, "URI", UriStruct);

	Sscheme = qu_string_new_staticz ("scheme");
	Shost   = qu_string_new_staticz ("host");
	Sport   = qu_string_new_staticz ("port");
	Spath   = qu_string_new_staticz ("path");
	Squery  = qu_string_new_staticz ("query");
	Sfrag   = qu_string_new_staticz ("frag");

	qu_module_add_function (m, "dequery"  , Uri_dequery  , -1);
	qu_module_add_function (m, "enquery"  , Uri_enquery  ,  1);
	qu_module_add_function (m, "escape"   , Uri_escape   ,  1);
	qu_module_add_function (m, "unescape" , Uri_unescape ,  1);
	qu_module_add_function (m, "normalize", Uri_normalize,  1);
	qu_module_add_function (m, "split"    , Uri_split    , -1);
	qu_module_add_function (m, "make"     , Uri_make     , -1);

	return 1;
}
