/* Utf8.c - Copyright (c) Marc Krisnanto */

/*) module Utf8
	This module provides additional methods to the String and Buffer classes
	to deal with UTF-8.

	@warn
	This module is experimental. Do not use in production environment.

	@xmp
		use Utf8

		x = 'foo'
		println (x)
		println (x.utoupper)
		println (x.isalpha)
		println (x.uispunct)

	@out
		foo
		FOO
		true
		false
*/

#include "qu.h"

typedef wint_t (*stringto_f) (wint_t c);
typedef int    (*stringis_f) (wint_t c);


/*----------------------------------------------------------------------------
	stuffs
----------------------------------------------------------------------------*/

static qu_boolean _utf8_string_is (qu_stringy self, stringis_f is, int argc, qu_any *argv)
{
	if (argc == 0)
	{
		long           l = self->len;
		volatile char *p = self->str;
		char          *e = (char *) p + l - 1;

		while (p <= e)
		{
			wint_t c = qu_utf8_getchar ((char **) &p, e);
			if (c == QU_WCHAR_INVALID || !is (c))
				return QuFalse;
		}
		return QuTrue;
	}
	else
	{
		long  i;
		char *p;

		QU_ARG_vlong (0, i);
		p = qu_utf8_pos (self->str, self->len, i);
		if (p)
		{
			wint_t c = qu_utf8_charat (p);
			if (c != QU_WCHAR_INVALID && is (c))
				return QuTrue;
		}
		return QuFalse;
	}
}



static qu_string _utf8_string_to (qu_stringy self, stringto_f to)
{
	/*
		volatile: stupid gcc bug optimizes the loop which result in the loop
		never entered.
	*/

	long           l = self->len;
	volatile char *p = self->str;
	char          *e = (char *) p + l - 1;
	long           w = 0;
	int            k = 0;
	wint_t         c = 0, d;

	while (p <= e)
	{
		c = qu_utf8_getchar ((char **) &p, e);
		QU_REQUIRE (QuESyntax, c != QU_WCHAR_INVALID);
		d = to (c);
		if (c != d)
			k = 1;
		w += qu_utf8_charlen (d);
	}

	if (k == 0)
	{
		return self;
	}
	if (w == 0)
	{
		return QuStringEMPTY;
	}
	else if (w == 1)
	{
		assert (0 <= c && c <= 127);
		return qu_string_byte (c);
	}
	else
	{
		qu_string  r = qu_string_newl (w);
		char      *q = r->str;

		p = self->str;
		while (p <= e)
		{
			c = qu_utf8_getchar ((char **) &p, e);
			assert (c != QU_WCHAR_INVALID);
			c = to (c);
			q = qu_utf8_putchar (q, c);
		}
		return r;
	}
}



/*----------------------------------------------------------------------------
	qu_string
----------------------------------------------------------------------------*/

/*) class String
	The Utf8 module add some methods to the String class.
*/


/*) sub ulen -> Int
	Return the number of wide characters.

	@exc
	.ESyntax
*/
static qu_int String_ulen (qu_string self)
{
	long l = qu_utf8_width (self->str, self->len);
	QU_REQUIRE (QuESyntax, l >= 0);
	assert (l <= QU_INT_MAX);
	return qu_int_new (l);
}



/*) sub utolower -> String
	Return a new String with all characters lowercased.

	@exc
	.ESyntax
*/
static qu_string String_utolower (qu_string self)
{
	return _utf8_string_to ((qu_stringy) self, qu_utf8_tolower);
}



/*) sub utotitle -> self
	Return a new String with all characters converted to title letters.

	@exc
	.ESyntax
*/
static qu_string String_utotitle (qu_string self)
{
	return _utf8_string_to ((qu_stringy) self, qu_utf8_totitle);
}



/*) sub utoupper -> String
	Return a new String with all characters uppercased.

	@exc
	.ESyntax
*/
static qu_string String_utoupper (qu_string self)
{
	return _utf8_string_to ((qu_stringy) self, qu_utf8_toupper);
}



/*) sub uisxxx -> Boolean
	@def
	sub uisxxx -> Boolean
	sub uisxxx (index) -> Boolean
	@p
	Methods to test for characters.

	The first form tests all characters.

	The second form test the character at the specific position.
	A negative v{index} means to count from the end.
	If the index is out of bound or the string is not a valid UTF-8 string
	then return false.

	@noexc

	@p
	The methods:
	@rc
	.| uisalpha | test for alphas
	.| uisalnum | test for alphas and digits
	.| uiscntrl | test for control characters
	.| uisgraph | test for graph characters
	.| uislower | test for lowercase letters
	.| uisprint | test for printing characters
	.| uispunct | test for punctuations
	.| uisspace | test for spaces
	.| uisupper | test for uppercase letters
	.| uiswide  | test for wide characters
*/
static qu_boolean String_uisalpha (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isalpha, argc, argv);
}



static qu_boolean String_uisalnum (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isalnum, argc, argv);
}



static qu_boolean String_uiscntrl (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_iscntrl, argc, argv);
}



static qu_boolean String_uisgraph (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isgraph, argc, argv);
}



static qu_boolean String_uislower (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_islower, argc, argv);
}



static qu_boolean String_uisprint (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isprint, argc, argv);
}



static qu_boolean String_uispunct (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_ispunct, argc, argv);
}



static qu_boolean String_uisspace (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isspace, argc, argv);
}



static qu_boolean String_uisupper (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isupper, argc, argv);
}



static qu_boolean String_uiswide (qu_string self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_iswide, argc, argv);
}



/*----------------------------------------------------------------------------
	qu_buffer
----------------------------------------------------------------------------*/

/*) class Buffer
	The Utf8 module add some methods to the Buffer class.
*/


/*) sub ulen -> Int
	Return the number of wide characters.

	@exc
	.ESyntax
*/
#define Buffer_ulen String_ulen




/*) sub utolower -> self
	Convert all characters to lowercase letters.

	@exc
	.ESyntax
*/
static qu_buffer Buffer_utolower (qu_buffer self)
{
	qu_string x = _utf8_string_to ((qu_stringy) self, qu_utf8_tolower);

	if (x != (qu_string) self)
		qu_buffer_exchange_tmp_string (self, x);
	return self;
}



/*) sub utotitle -> self
	Convert all characters to title letters.

	@exc
	.ESyntax
*/
static qu_buffer Buffer_utotitle (qu_buffer self)
{
	qu_string x = _utf8_string_to ((qu_stringy) self, qu_utf8_totitle);

	if (x != (qu_string) self)
		qu_buffer_exchange_tmp_string (self, x);
	return self;
}



/*) sub utoupper -> self
	Convert all characters to uppercase letters.

	@exc
	.ESyntax
*/
static qu_buffer Buffer_utoupper (qu_buffer self)
{
	qu_string x = _utf8_string_to ((qu_stringy) self, qu_utf8_toupper);

	if (x != (qu_string) self)
		qu_buffer_exchange_tmp_string (self, x);
	return self;
}



/*) sub uisxxx -> Boolean
	@def
	sub uisxxx -> Boolean
	sub uisxxx (index) -> Boolean
	@p
	Methods to test for characters.

	The first form tests all characters.

	The second form test the character at the specific position.
	A negative v{index} means to count from the end.
	If the index is out of bound or the string is not a valid UTF-8 string
	then return false.

	@noexc

	@p
	The methods:
	@rc
	.| uisalpha | test for alphas
	.| uisalnum | test for alphas and digits
	.| uiscntrl | test for control characters
	.| uisgraph | test for graph characters
	.| uislower | test for lowercase letters
	.| uisprint | test for printing characters
	.| uispunct | test for punctuations
	.| uisspace | test for spaces
	.| uisupper | test for uppercase letters
	.| uiswide  | test for wide characters
*/
static qu_boolean Buffer_uisalpha (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isalpha, argc, argv);
}



static qu_boolean Buffer_uisalnum (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isalnum, argc, argv);
}



static qu_boolean Buffer_uiscntrl (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_iscntrl, argc, argv);
}



static qu_boolean Buffer_uisgraph (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isgraph, argc, argv);
}



static qu_boolean Buffer_uislower (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_islower, argc, argv);
}



static qu_boolean Buffer_uisprint (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isprint, argc, argv);
}



static qu_boolean Buffer_uispunct (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_ispunct, argc, argv);
}



static qu_boolean Buffer_uisspace (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isspace, argc, argv);
}



static qu_boolean Buffer_uisupper (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_isupper, argc, argv);
}



static qu_boolean Buffer_uiswide (qu_buffer self, int argc, qu_any *argv)
{
	return _utf8_string_is ((qu_stringy) self, qu_utf8_iswide, argc, argv);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Utf8_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		qu_string
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_class_method_instz (QuStringClass, "ulen"    , String_ulen    ,  0);
	qu_class_method_instz (QuStringClass, "utolower", String_utolower,  0);
	qu_class_method_instz (QuStringClass, "utotitle", String_utotitle,  0);
	qu_class_method_instz (QuStringClass, "utoupper", String_utoupper,  0);

	qu_class_method_instz (QuStringClass, "uisalnum", String_uisalnum, -1);
	qu_class_method_instz (QuStringClass, "uisalpha", String_uisalpha, -1);
	qu_class_method_instz (QuStringClass, "uiscntrl", String_uiscntrl, -1);
	qu_class_method_instz (QuStringClass, "uisgraph", String_uisgraph, -1);
	qu_class_method_instz (QuStringClass, "uislower", String_uislower, -1);
	qu_class_method_instz (QuStringClass, "uisprint", String_uisprint, -1);
	qu_class_method_instz (QuStringClass, "uispunct", String_uispunct, -1);
	qu_class_method_instz (QuStringClass, "uisspace", String_uisspace, -1);
	qu_class_method_instz (QuStringClass, "uisupper", String_uisupper, -1);
	qu_class_method_instz (QuStringClass, "uiswide" , String_uiswide , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		qu_buffer
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_class_method_instz (QuBufferClass, "ulen"    , Buffer_ulen    ,  0);
	qu_class_method_instz (QuBufferClass, "utolower", Buffer_utolower,  0);
	qu_class_method_instz (QuBufferClass, "utotitle", Buffer_utotitle,  0);
	qu_class_method_instz (QuBufferClass, "utoupper", Buffer_utoupper,  0);

	qu_class_method_instz (QuBufferClass, "uisalnum", Buffer_uisalnum, -1);
	qu_class_method_instz (QuBufferClass, "uisalpha", Buffer_uisalpha, -1);
	qu_class_method_instz (QuBufferClass, "uiscntrl", Buffer_uiscntrl, -1);
	qu_class_method_instz (QuBufferClass, "uisgraph", Buffer_uisgraph, -1);
	qu_class_method_instz (QuBufferClass, "uislower", Buffer_uislower, -1);
	qu_class_method_instz (QuBufferClass, "uisprint", Buffer_uisprint, -1);
	qu_class_method_instz (QuBufferClass, "uispunct", Buffer_uispunct, -1);
	qu_class_method_instz (QuBufferClass, "uisspace", Buffer_uisspace, -1);
	qu_class_method_instz (QuBufferClass, "uisupper", Buffer_uisupper, -1);
	qu_class_method_instz (QuBufferClass, "uiswide" , Buffer_uiswide , -1);

	return 1;
}
