/* valid.c - Copyright (c) 2001 Marc Krisnanto */

#include "qu.h"


/*) module Valid
	This module provides a number of validator functions.

	The intention is to provide fast and I{common sense} validators.
	They are not perfect but they should be usable in most cases.

	@xmp
		use Valid

		struct Foo
			auth  : Valid.Auth
			email : Valid.Email
		;;
*/

#define STRINGIFY(x) \
	(qu_string_is (x) \
		? (qu_string) x \
		: (qu_buffer_is (x) \
			? qu_buffer_as_string ((qu_buffer) x) \
			: qu_sys_string (x))) \


/*-----------------------------------------------------------------------------
	general
-----------------------------------------------------------------------------*/

/*) sub Authority (x)
	Return k{false} if v{x} is not an -{Integer} within C{0..15}, the -{Int} value
	otherwise.

	@noexc
*/
static qu_any Valid_Authority (qu_any x)
{
	long v;

	v = qu_integer_tol (x);
	if (errno == 0 && 0 <= v && v < 16)
		return qu_int_newl (v);
	return QuFalse;
}



/*) sub Birth (x)
	Return false if v{x} is not a birthdate value, return an Integer in the form of
	C{yyyymmdd} otherwise.

	v{x} must be a String in the form C{yyyy-mmm-dd}.

	For example, C{'12-Jun-1929'} returns C{19290512}.

	@warn
	The year range is C{1000} to C{9999}.
*/
static qu_any Valid_Birth (qu_string x)
{
	char *s;
	int   y, m, d, mdays;

	if (!qu_string_is (x) || x->len != 11)
		return QuFalse;

	s = x->str;
	d = (s[1] - '0') * 10 + (s[1] - '0');
	y = (s[7] - '0') * 1000 + (s[8] - '0') * 100 + (s[9] - '0') * 10 + (s[10] - '0');

	if (y < 1000 || y > 9999 || d > 31)
		return QuFalse;

	switch (qu_ascii_tolower (s[3]))
	{
		case 'j': m = (qu_ascii_tolower (s[4]) == 'a') ? 1 : (qu_ascii_tolower (s[5]) == 'n') ? 6 : 7; break;
		case 'a': m = (qu_ascii_tolower (s[4]) == 'p') ? 4 : 8; break;
		case 'm': m = (qu_ascii_tolower (s[5]) == 'r') ? 3 : 5; break;
		case 's': m =  9; break;
		case 'f': m =  2; break;
		case 'o': m = 10; break;
		case 'n': m = 11; break;
		case 'd': m = 12; break;
		default : return QuFalse;
	}

	mdays = 30 + ((m % 2) ^ (m > 7)) - ((m != 2) ? 0 : 2 - ((y % 4) == 0) + (((y % 100) == 0) && ((y % 400) != 0)));
	if (d > mdays)
		return QuFalse;

	d = y * 10000 + m * 100 + d;
	return qu_int_newl (d);
}



/*) sub Codename (x)
	Return k{false} if v{x} is not a valid codename, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be an ASCII string 4 to 32 characters, start with an alpha and must
	contain only alphanumeric characters (thus underscores are invalid).

	@noexc
*/
static qu_any Valid_Codename (qu_stringy x)
{
	if (qu_stringy_is (x) && 4 <= x->len && x->len <= 32 && qu_ascii_isalpha (x->str[0]))
	{
		char *i = x->str;
		char *m = i + x->len;

		while (*i)
		{
			if (!qu_ascii_isalnum (*i))
				return QuFalse;
			++i;
		}

		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Email (x)
	Return k{false} if v{x} is not a valid email address, the -{String} value
	otherwise.

	v{x} may be a String or -{Buffer}.

	This function is only meant to verify simple email addresses in the form
	of C{aaa@bbb.ccc}.

	The string must be 6 to 64 characters ASCII string.

	@noexc
*/
static qu_any Valid_Email (qu_stringy x)
{
	if (qu_stringy_is (x) && x->len >= 6 && x->len <= 64)
	{
		char *i = x->str;
		char *m = i + x->len;
		int   k = 0;

		while (qu_ascii_isalnum (*i))
			++i;

		if (i == x->str || *i == '\0')
			return QuFalse;

	  dot1:
		if (*i == '.' || *i == '-')
		{
			if (*++i == '\0')
				return QuFalse;

			while (qu_ascii_isalnum (*i))
				++i;
			goto dot1;
		}

		if (*i != '@')
			return QuFalse;

		if (*++i == '\0')
			return QuFalse;

		while (qu_ascii_isalnum (*i))
			++i;

	  dot2:
		if (*i == '.' || *i == '-')
		{
			if (*++i == '\0')
				return QuFalse;

			while (qu_ascii_isalnum (*i))
				++i;

			if (++k > 2)
				return QuFalse;

			goto dot2;
		}

		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Identifier (x)
	Return k{false} if v{x} is not a valid identifier, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be an ASCII string which starts with an alpha or underscore, and
	contain only identifier characters. Thus a valid Qu symbol.

	@noexc
*/
static qu_any Valid_Identifier (qu_stringy x)
{
	if (qu_stringy_is (x)
			&& 0 < x->len && x->len <= 32
			&& (qu_ascii_isalpha (x->str[0]) || x->str[0] == '_'))
	{
		char *i = x->str;
		char *m = x->str + x->len;

		while (*i)
		{
			if (!qu_ascii_isident (*i))
				return QuFalse;
			++i;
		}
		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Keyword (x)
	Return k{false} if v{x} is not a valid keyword, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be 1 to 80 characters long, must not be spaces or digits only,
	and must contain only printable characters.

	@noexc

	@todo
	This works with ASCII strings only.
*/
static qu_any Valid_Keyword (qu_stringy x)
{
	if (qu_stringy_is (x) && 0 < x->len && x->len <= 80)
	{
		char *i = x->str;
		char *m = x->str + x->len;
		int   k = 0;

		while (*i)
		{
			if (!qu_ascii_isprint (*i))
				return QuFalse;

			if (!qu_ascii_isspace (*i))
			{
				if (!qu_ascii_isdigit (*i))
					k = 1;
			}
			++i;
		}
		if (i == m && k)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Name (x)
	Return k{false} if v{x} is not a valid name, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must only contain alpha, spaces, dots and single quotes.
	It must contain at least one alpha. Its length must be between C{2..64}.

	@noexc

	@todo
	This works with ASCII strings only.
*/
static qu_any Valid_Name (qu_stringy x)
{
	if (qu_stringy_is (x) && 1 < x->len && x->len <= 64)
	{
		char *i = x->str;
		char *m = i + x->len;
		int   k = 0;

		while (*i)
		{
			if (qu_ascii_isalpha (*i))
				k = 1;
			else if (*i != ' ' && *i != '.' && *i != '\'')
				return QuFalse;
			i++;
		}

		if (k && i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub NonEmptyString (x)
	Return k{false} if v{x} is an empty string, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	@noexc
*/
static qu_any Valid_NonEmptyString (qu_stringy x)
{
	if (qu_stringy_is (x) && x->len)
		return STRINGIFY (x);
	return QuFalse;
}



/*) sub Password (x)
	Return k{false} if v{x} is not a valid password, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be an 6 to 16 ASCII characters string and must satisfy
	-{String.ispwd},

	@noexc
*/
static qu_any Valid_Password (qu_stringy x)
{
	if (qu_stringy_is (x) && 6 <= x->len && x->len <= 16)
	{
		char *i = x->str;
		char *m = i + x->len;

		while (*i)
		{
			if (!qu_ascii_ispwd (*i))
				return QuFalse;
			i++;
		}

		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Phone (x)
	Return k{false} if v{x} is not a valid phone number(s), the -{String} value
	otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be an ASCII string with 1 to 80 characters.
	Each phone number is separated by a comma and optionally followed by spaces.
	Spaces are only allowed after the comma.

	A phone number may contain C{'-'} and alphanumeric characters.

	@noexc
*/
static qu_any Valid_Phone (qu_stringy x)
{
	if (qu_stringy_is (x) && 1 <= x->len && x->len <= 80)
	{
		long  l = x->len;
		char *i = x->str;
		int   n = 0;
		int   p = 0;

		while (l--)
		{
			switch (*i)
			{
				case '-':
					if (n == 0)
						return QuFalse;
					n = 0;
					break;

				case ' ':
					return QuFalse;

				case '+':
					if (n != 0)
						return QuFalse;
					break;

				case ',':
					if (n < 3)
						return QuFalse;
					i++;
					while (l && *i == ' ')
					{
						l--;
						i++;
					}
					n = 0;
					continue;

				default:
					if (qu_ascii_isalnum (*i))
					{
						p = 1;
						n++;
						if (n > 64)
							return QuFalse;
					}
					else
					{
						return QuFalse;
					}
			}
			i++;
		}
		if (p)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Port (x)
	Return k{false} if v{x} is not a port number, the -{Int} value otherwise.

	v{x} may be a Int or -{Long}.

	The number must be within 1 to 65535.

	@noexc
*/
static qu_any Valid_Port (qu_integer x)
{
	int v = qu_integer_toi (x);
	if (errno || v <= 0 || v > 65535)
		return QuFalse;
	return qu_int_newl (v);
}



/*) sub Text (x)
	Return k{false} if v{x} is not a valid text, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be 4 to 1MB characters long with all printable characters and
	must not be all spaces.

	@noexc

	@todo
	This only works with ASCII strings.
*/
static qu_any Valid_Text (qu_stringy x)
{
	if (qu_stringy_is (x) && 4 <= x->len && 1048576)
	{
		char *i = x->str;
		char *m = i + x->len;
		int   k = 0;

		while (*i)
		{
			if (!qu_ascii_istext (*i))
				return QuFalse;
			if (!qu_ascii_isspace (*i))
				k = 1;
			++i;
		}
		if (k && i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Title (x)
	Return k{false} if v{x} is a not a valid title, the ={String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be 2 to 80 characters long, must contain only printable characters
	except C{"|"}, and must not be all spaces.

	@noexc

	@todo
	This only works with ASCII strings.
*/
static qu_any Valid_Title (qu_stringy x)
{
	if (qu_stringy_is (x) && 2 <= x->len && x->len <= 80)
	{
		char *i = x->str;
		char *m = x->str + x->len;
		int   k = 0;

		while (*i)
		{
			if (!qu_ascii_isprint (*i) || *i == '|')
				return QuFalse;
			if (!qu_ascii_isspace (*i))
				k = 1;
			i++;
		}
		if (k && i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Username (x)
	Return k{false} if v{x} is not a valid username, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string be an ASCII string and start with an alpha or underscore, satisfy
	-{String.isident} and its length must be between C{2..32}.

	@noexc
*/
static qu_any Valid_Username (qu_stringy x)
{
	if (qu_stringy_is (x) && 2 <= x->len && x->len <= 32 && qu_ascii_isalpha (x->str[0]))
	{
		char *i = x->str;
		char *m = i + x->len;

		while (*i)
		{
			if (!qu_ascii_isident (*i))
				return QuFalse;
			++i;
		}
		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Version (x)
	Return k{false} if v{x} is not a version, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be an ASCII string in the form of C{d[.d[.d]]} where C{d} is
	one or two decimal digits.

	@noexc
*/
static qu_any Valid_Version (qu_stringy x)
{
	if (qu_stringy_is (x))
	{
		char *i = x->str;
		char *m = i + x->len;

		if (!qu_ascii_isdigit (*i))
			return QuFalse;
		++i;
		if (qu_ascii_isdigit (*i))
			++i;

		if (*i == '.')
		{
			++i;
			if (!qu_ascii_isdigit (*i))
				return QuFalse;
			++i;
			if (qu_ascii_isdigit (*i))
				++i;

			if (*i == '.')
			{
				++i;
				if (!qu_ascii_isdigit (*i))
					return QuFalse;
				++i;
				if (qu_ascii_isdigit (*i))
					++i;
			}
		}

		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Zip (x)
	Return k{false} if v{x} is not a valid zip code, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must be a 4 to 15 characters long and contain decimal digits only.

	@noexc
*/
static qu_any Valid_Zip (qu_stringy x)
{
	if (qu_stringy_is (x) && 4 <= x->len && x->len <= 15)
	{
		char *i = x->str;
		char *m = i + x->len;

		while (*i)
		{
			if (!qu_ascii_isdigit (*i))
				return QuFalse;
			++i;
		}
		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*-----------------------------------------------------------------------------
	file validators
-----------------------------------------------------------------------------*/

/*) sub Abspath (x)
	Return k{false} if v{x} is not a valid absolute path, the -{String} value
	otherwise.

	v{x} may be a String or -{Buffer}.

	The string must satisfy -{File.isabs} and -{File.valid} with no space.

	@noexc
*/
static qu_any Valid_Abspath (qu_stringy x)
{
	if (qu_stringy_is (x) && qu_file_is_abs (x) && qu_file_valid (x, 0))
		return STRINGIFY (x);
	return QuFalse;
}



/*) sub Absdir (x)
	Return k{false} if v{x} is not a valid absolute path, the -{String} value
	otherwise.

	v{x} may be a String or -{Buffer}.

	The string must satisfy -{File.isabs} and -{File.valid} with no space and
	must be an existing directory.

	@noexc
*/
static qu_any Valid_Absdir (qu_stringy x)
{
	if (qu_stringy_is (x) && qu_file_is_abs ((qu_string) x) && qu_file_valid ((qu_string) x, 0))
	{
		x = (qu_stringy) STRINGIFY (x);
		if (qu_file_is_dir ((qu_string) x))
			return x;
	}
	return QuFalse;
}



/*) sub Filename (x)
	Return k{false} if v{x} is a valid filename, the -{String} value
	otherwise.

	v{x} may be a String or -{Buffer}.

	The string must satisfy -{File.valid} and must not contain C{/}.

	@noexc
*/
static qu_any Valid_Filename (qu_stringy x)
{
	if (qu_stringy_is (x) && qu_file_valid ((qu_string) x, 0))
	{
		char *i = x->str;
		char *m = i + x->len;

		while (*i)
		{
			if (*i == '/' || /* just in case */ *i == '\\')
				return QuFalse;
			++i;
		}

		if (i == m)
			return STRINGIFY (x);
	}
	return QuFalse;
}



/*) sub Path (x)
	Return k{false} if v{x} is a valid filename, the -{String} value otherwise.

	v{x} may be a String or -{Buffer}.

	The string must satisfy -{File.valid} with no space.

	@noexc
*/
static qu_any Valid_Path (qu_stringy x)
{
	if (qu_stringy_is (x) && qu_file_valid ((qu_string) x, 0))
		return STRINGIFY (x);
	return QuFalse;
}



/*-----------------------------------------------------------------------------
	misc
-----------------------------------------------------------------------------*/

/*) sub Header (x)
	Return k{false} if v{x} is a valid header name, the -{String} value
	otherwise.

	v{x} may be a String or -{Buffer}.

	The length must be within 1..40 and contains only alphas, digits,
	or C{'-'} (dash).

	The string is converted to lowercase by means of -{String.tolower}.

	@noexc

	@xmp
		use Valid
		println (Valid.Header ('Content-Length'))
		println (Valid.Header ('Hello + World'))

	@out
		content-length
		false
*/
static qu_any Valid_Header (qu_stringy x)
{
	if (qu_stringy_is (x) && 2 <= x->len && x->len <= 40)
	{
		char *i = x->str;
		char *j, buf [40];
		int   k = 0;

		j = buf;
		while (*i)
		{
			if (!(qu_ascii_isalnum (*i) || *i == '-'))
				return QuFalse;
			*j++ = qu_ascii_tolower (*i++);
			if (qu_ascii_islower (*i))
				k = 1;
		}
		if (k == 0 && qu_string_is (x))
			return x;
		return qu_string_new (buf, j - buf);
	}
	return QuFalse;
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

int qu_Valid_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		general
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "Authority"     , Valid_Authority     , 1);
	qu_module_add_function (m, "Birth"         , Valid_Birth         , 1);
	qu_module_add_function (m, "Codename"      , Valid_Codename      , 1);
	qu_module_add_function (m, "Email"         , Valid_Email         , 1);
	qu_module_add_function (m, "Identifier"    , Valid_Identifier    , 1);
	qu_module_add_function (m, "Keyword"       , Valid_Keyword       , 1);
	qu_module_add_function (m, "Name"          , Valid_Name          , 1);
	qu_module_add_function (m, "NonEmptyString", Valid_NonEmptyString, 1);
	qu_module_add_function (m, "Password"      , Valid_Password      , 1);
	qu_module_add_function (m, "Phone"         , Valid_Phone         , 1);
	qu_module_add_function (m, "Port"          , Valid_Port          , 1);
	qu_module_add_function (m, "Text"          , Valid_Text          , 1);
	qu_module_add_function (m, "Title"         , Valid_Title         , 1);
	qu_module_add_function (m, "Username"      , Valid_Username      , 1);
	qu_module_add_function (m, "Version"       , Valid_Version       , 1);
	qu_module_add_function (m, "Zip"           , Valid_Zip           , 1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		file
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "Abspath"       , Valid_Abspath       , 1);
	qu_module_add_function (m, "Absdir"        , Valid_Absdir        , 1);
	qu_module_add_function (m, "Filename"      , Valid_Filename      , 1);
	qu_module_add_function (m, "Path"          , Valid_Path          , 1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		misc
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "Header"        , Valid_Header        , 1);

	return 1;
}
