/* Http.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Http
	This module provides numerous functions to deal with the HTTP
	protocol.

	The module is intentionally written in C to help boost performance.

	@xmp
		use Web.Http

	@xmp
		println (Http.reason (206))

	@out
		Partial Content
*/

/* valid status range */
#define HTTP_STATUS_MIN 100
#define HTTP_STATUS_MAX 599

/* the reason strings */
static qu_string QuHttpStatus [HTTP_STATUS_MAX - HTTP_STATUS_MIN + 1];

/* how to sort the headers -- Array of String */
static qu_array QuHttpHeaderSort;



/*) sub parse (s, h = [:], options = 'a') -> Dict
	Parse the headers String or Buffer v{s}.

	v{h} is returned as the result. It may be any object that supports the
	C{__getitem}, C{__getslice}, and C{__setitem} ala -{Dict}.

	v{options} is a combination of the following characters:
	@rc
	.| C{'a'} |
		Allow multiple header with the same name and collect in Array.
		Otherwise throw.

	.| C{':'} |
		Allow header without a semicolon.
		The key will be the whole header and the value is nil.

	@note
		The items are String objects, not Buffer. The field names are converted
		to ASCII lowercase characters.

		This function works for both C{'\n'} and C{'\r\n'} separators.

	@exc
		.EValue

	@xmp
		use Web.Header

		s = syscat (
				'Foo: Fooey\r\n'
				'   Mr.\r\n'
				'Bar: bar is\r\n'
				'Bar: tender\r\n\r\n')

		h = Header.parse (s, [:])

		println (k ' = ' v) for k, v in h

		@out
			foo = Fooey Mr.
			bar = ['bar is', 'tender']
*/
qu_dict qu_http_parse (int argc, qu_any *argv)
{
	qu_stringy s;
	qu_dict    h;
	int        multi = 1;
	int        nocolon = 0;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, s, qu_stringy);
	if (argc > 1)
	{
		QU_ARG_VTYPE (1, h, qu_dict);
		if (argc > 2)
		{
			qu_stringy op;
			QU_ARG_VTYPE (1, op, qu_stringy);
			multi   = strchr (op->str, 'a') != 0;
			nocolon = strchr (op->str, ':') != 0;
	}
	else
	{
		h = qu_dict_new (4);
	}

	while (*i)
	{
		/* find the end of the line */
		start = i;
		while (*i && *i != '\n')
			++i;




	for s in s.unfold.eachline
		if ':' in s
			(k, v) = s.split (':', 2)
		else
			continue if s.isspace
			throw EValue, "no semicolon" unless ':' in options
			k = s
			v = ''
		;;

		k = String (k.trim.tolower)
		v = String (v.trim)

		switch s = h [k, nil]
			case is String
				if options is String
					throw EValue, "duplicate header" unless 'a' in options
				else
					throw EValue, "duplicate header" unless k in options
				;;
				h [k] = [s, v]
				continue
			case is Array
				s <<< v
				continue
			default
				h [k] = v
		;;
	;;
	return h;
}



/*) sub write (dst, h, sort = nil, repr = const [:])
	Write headers to v{dst}.

	v{h} may be any object that supports the f{__getitem}, f{__getslice}, and
	f{__setitem} ala -{Dict}.

	v{sort} is an Array of String. If provided then the headers are sorted as
	specified, otherwise they appear in an arbitrary order.

	v{repr} is a Dict of (field names, representation). If supplied then the
	representation of field names are fecthed from the Dict, otherwise they
	are converted into their usual representation, e.g. C{content-length} becomes
	C{Content-Length}.

	If v{dst} is nil then return a String. Otherwise it must be an object that
	support the f{__push} method (such as -{Stream}, -{Buffer}, in which case
	v{dst} is returned as the result.

	@xmp
		use Web.Header

		h =
		[
			'foo': 'Fooey'
			'bar': ['bar is', 'tender']
		]

		r = ['foo': 'FOO']

		Header.write (__stdout__, h, ['bar', 'foo'], r)

	@out
		Bar: bar is
		Bar: tender
		FOO: Fooey
*/
	local k, v
	b = dst == nil ? Buffer (80) : dst

	sub push
		k = k in repr ? repr [k] : k.toheader
		if v is Array
			for x in v
				x = String (x) unless x is Buffer or x is String
				b <<< k <<< ': ' <<< x <<< '\r\n'
			;;
		else
			v = String (v) unless v is Buffer or v is String
			b <<< k <<< ': ' <<< v <<< '\r\n'
		;;
	;;

	if sort
		r = {}
		for k in sort
			if k in h
				r <<< k
				v = h[k]
				push
			;;
		;;
		for k, v in h
			push unless k in r
		;;
	else
		push for k, v in h
	;;

	return String (b) if dst is String
	return dst
;;

sub parse (s, h = [:])
	```
	This is a convenient function that calls %{Web/Header.parse}:
	@prog
		Header.parse (s, h, Multi)

	@p
	See #{Multi}.
	```
	return Header.parse (s, h, Multi)
;;


sub write (dst, headers)
	```
	This is a convenient function that calls %{Web/Header.write}:
	@prog
		Header.write (dst, headers, Sort, Repr)

	@p
	See #{Sort} and #{Repr}.
	```
	return Header.write (dst, headers, Sort, Repr)
;;



/*) sub nobody (status) -> Boolean
	Return true if the response must not return a body.

	v{status} may be an Integer or anything convertible to an Integer such as
	String, Buffer, Float, etc.

	If v{status} is not within C{100..599} then true is returned.

	@noexc
*/
qu_boolean qu_http_nobody (qu_integer status)
{
	int i;

	if (!qu_integer_is (x))
	{
		x = qu_sys_as (x, QuIntClass);
		if (x == 0)
			return QuHttpStatusUnknown;
	}
	i = qu_integer_toi (x);
	if (errno)
		return QuTrue;
	return (i == 100
		||  i == 101
		||  i == 204
		||  i == 205
		||  i == 304
	/*	||  i == 305*/) ? QuTrue : QuFalse;
}



/*) sub reason (status) -> String
	Get reason string.

	v{status} may be an Integer or anything convertible to an Integer such as
	String, Buffer, Float, etc.

	If v{status} is not within C{100..599} then C{"Unknown"} is returned.

	@noexc

	@xmp
		use Web.Http
		println (Http.reason (206))
		println (Http.reason ('206 xxxx'))

	@out
		Partial Content
		Partial Content
*/
qu_string qu_http_reason (qu_any x)
{
	static qu_string unknown = 0;
	int i;

	if (unknown == 0)
		unknown = qu_string_new_static ("Unknown");

	if (!qu_integer_is (x))
	{
		x = qu_sys_as (x, QuIntClass);
		if (x == 0)
			return unknown;
	}
	i = qu_integer_toi (x);
	if (errno == 0 && HTTP_STATUS_MIN <= i && i <= HTTP_STATUS_MAX)
	{
		x = QuHttpStatus [i - HTTP_STATUS_MIN];
		if (x)
			return x;
	}
	return unknown;
}



/*) sub kind (status) -> Int
	Return the kind of status.

	v{status} may be an Integer or anything responding to C{as Int} such as
	String, Buffer, Float, etc.

	If it is not within 100..599 then C{INFO} is returned, otherwise
	the returned value is one of:
	@args
	.1  informative  C{(100 .. 199)}
	.2  success      C{(200 .. 299)}
	.3  redirect     C{(300 .. 399)}
	.4  client error C{(400 .. 499)}
	.5  server error C{(500 .. 599)}

	@noexc

	@xmp
		use Web.Http
		k = Http.kind (206)
		println (k)

	@out
		2
*/
qu_int qu_http_kind (qu_any x)
{
	int v;

	if (!qu_integer_is (x))
	{
		x = qu_sys_as (x, QuIntClass);
		if (x == 0)
			return 1;
	}
	v = qu_integer_toi (x);
	if (errno == 0)
	{
		if (v < 200) return 1;
		if (v < 300) return 2;
		if (v < 400) return 3;
		if (v < 500) return 4;
		if (v < 600) return 5;
	}
	return 1;
}


# HttpHeader.qm - by Marc K.

use Web.Header


static
var Repr =
[
	'content-md5'      : 'Content-MD5'
	'etag'             : 'ETag'
	'te'               : 'TE'
	'www-authenticate' : 'WWW-Authenticate'

] of (Header.Field, Header.Repr)

	```
	This is a Dict of known headers with strange representation, that is, they
	do not follow -{String.toheader}.

	Note that HTTP header names are case-insensitive but it is common practice to
	write them as they are written in RFC-2616.
	```




var Multi =
{
	'cookie'
	'set-cookie'
	'set-cookie2'

} of Header.Field

	```This is a Set of headers that are allowed to appear more than once.```


var Single =
{
	'transfer-encoding'
	'upgrade'

	'accept-ranges'

	'etag'
	'location'
	'proxy-authenticate'
	'www-authenticate'

	'authorization'
	'expect'
	'host'
	'proxy-authorization'
	'content-encoding'
	'content-md5'
	'content-length'
	'content-location'
	'content-type'

} of Header.Field

	```This is a Set of headers that are only allowed to appear once.```


/*) sub notrail (s) -> Boolean
	Whether the header v{s} may not appear in trailers.

	This is:
	@ul
	.content-length
	.trailer
	.transfer-encoding

	@noexc

	@warn
	If v{s} is neither a String or Buffer then false is returned.
*/
qu_boolean qu_http_notrail (qu_stringy x)
{
	if (qu_stringy_is (x) && x->len < 128)
	{
		long  l = x->len;
		char *s = x->str;

		if (   (l == 14 && qu_strz_eqnc ("content-length"   , s, l))
			|| (l ==  7 && qu_strz_eqnc ("trailer"          , s, l))
			|| (l == 17 && qu_strz_eqnc ("transfer-encoding", s, l))
			return QuTrue;
	}
	return QuFalse;
}



/*) sub parserange (s, size: Integer) -> Array
	Parse a Range header.

	v{s} is the Range header value.

	v{size} is the size of the resource, e.g. file length.

	The returned value is an Array of Range objects.

	@exc
	.Exception any exception means an invalid range specifier

	@xmp
		use Web.Http

		x = "1-200, 4-, 300-700, -500"
		r = HttpHeader.parse_range (x, 900)
		println (r)

	@out
		[<Range 0..0>, <Range 1..200>, <Range 4..899>, <Range 300..700>, <Range 400..899>]

	@note
		-{Range} objects have -{Int} limit. This is intentional since huge file ranges
		are either bogus or stupid.
*/
qu_array qu_http_parserange (qu_stringy s, qu_integer size)
{
	long     z, i, j, n;
	int      k;
	char    *p, *q;
	qu_array r;

	QU_ARG_TYPE  (s, qu_stringy);
	QU_ARG_long  (size, z);
	QU_ARG_CHECK (z >= 0);

	r = qu_array_new (4);
	p = s->str;
	while (*p)
	{
		while (qu_ascii_isspace (*p))
			p++;
		if (*p == '-')
		{
			k = 1;
			++p;
		}
		i = 0;
		while (qu_ascii_isdigit (*p))
		{
			i = i * 10 + (*p - '0');
			++p;
		}
		while (qu_ascii_isspace (*p))
			p++;
		if (k == 1)
		{
			/* -500 */
			n = i;
			i = size - i;
		}
		else
		{





	for s in s.split (',')
		i = s.head ('-')
		a = s.tail ('-')
		if i == ''
			i = size - Int (a)
			a = size - 1
		else
			i = Int (i)
			a = a == '' ? size - 1 : Int (a)
		;;
		throw EIndex if i > a or i > size or a > size or i < 0 or a < 0
		r <<< Range (i, a)
	;;
	return r || throw ENone
;;




/*) sub rangesize (ranges: Array, size: Integer, boundary: String, mime: String) -> Integer
	Get the content-length of a range response.

	v{range} is an Array of Range objects as returned by #{parserange}.

	v{size} is the size of the resource, e.g. file length.

	v{boundary} is the boundary string to use.

	v{mime} is the content-type.

	See RFC-2616 C{19.2 Internet Media Type multipart/byteranges}.

	@exc
	.EArgValue
	.EValue total size too big
*/
static long _long_strlen (register long v)
{
	register long n = 1;
	while (v)
	{
		v /= 10;
		n++;
	}
	return n;
}


qu_integer qu_http_rangesize (qu_array ranges, qu_integer size,
                              qu_string boundary, qu_string mime)
{
	long lsize, lboundary, lmime, l = 0;

	QU_ARG_TYPE (ranges, qu_array);
	QU_ARG_long (size, lsize);
	QU_ARG_TYPE (boundary, qu_stringy);
	QU_ARG_TYPE (mime, qu_stringy);

	lsize     = _long_strlen (lsize);
	lboundary = boundary->len;
	lmime     = mime->len;

	for (i = 0; i < ranges->len; ++i)
	{
		qu_range r = qu_array_GETITEM (ranges, i);

		QU_ARG_TYPE (r, qu_range);

		l += lboundary + 2
		  + 14 + lmime + 2
		  + 21 + _long_strlen (r->first) + 1 + _long_strlen (r->last) + 1 + lsize + 2
		  + 2
		  + r->count + 2;

		QU_REQUIRE (QuEValue, l > 0);
	}
	l += lboundary + 4;
	return qu_integer_newl (l);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

int qu_Http_init (qu_module m)
{

#define s_(i, z) QuHttpStatus[i - HTTP_STATUS_MIN] = qu_string_new_static (z)

	s_(100, "Continue");
	s_(101, "Switching Protocols");
	s_(200, "OK");
	s_(201, "Created");
	s_(202, "Accepted");
	s_(203, "Non-Authoritative Information");
	s_(204, "No Content");
	s_(205, "Reset Content");
	s_(206, "Partial Content");
	s_(300, "Multiple Choices");
	s_(301, "Moved Permanently");
	s_(302, "Found");
	s_(303, "See Other");
	s_(304, "Not Modified");
	s_(305, "Use Proxy");
	s_(307, "Temporary Redirect");
	s_(400, "Bad Request");
	s_(401, "Unauthorized");
	s_(402, "Payment Required");
	s_(403, "Forbidden");
	s_(404, "Not Found");
	s_(405, "Method Not Allowed");
	s_(406, "Not Acceptable");
	s_(407, "Proxy Authentication Required");
	s_(408, "Request Timeout");
	s_(409, "Conflict");
	s_(410, "Gone");
	s_(411, "Length Required");
	s_(412, "Precondition Failed");
	s_(413, "Request Entity Too Large");
	s_(414, "Request-URI Too Large");
	s_(415, "Unsupported Media Type");
	s_(416, "Request Range Not Satisfiable");
	s_(417, "Expectation Failed");
	s_(500, "Internal Server Error");
	s_(501, "Not Implemented");
	s_(502, "Bad Gateway");
	s_(503, "Service Unavailable");
	s_(504, "Gateway Timeout");
	s_(505, "Protocol Version Not Supported");
	s_(506, "Variant Also Negotiates");

#undef s_

#define s_(z) qu_array_push (QuHttpHeaderSort, qu_string_new_staticz (z))

	QuHttpHeaderSort = qu_array_new_static (50);

	/* request */
	s_("cache-control");
	s_("connection");
	s_("keep-alive");
	s_("date");
	s_("pragma");
	s_("trailer");
	s_("transfer-encoding");
	s_("upgrade");
	s_("via");
	s_("warning");
	s_("cookie");

	/* response */
	s_("accept-ranges");
	s_("age");
	s_("etag");
	s_("location");
	s_("proxy-authenticate");
	s_("retry-after");
	s_("server");
	s_("vary");
	s_("www-authenticate");

	/* request */
	s_("accept");
	s_("accept-charset");
	s_("accept-encoding");
	s_("accept-language");
	s_("authorization");
	s_("expect");
	s_("from");
	s_("host");
	s_("if-match");
	s_("if-modified-since");
	s_("if-none-match");
	s_("if-range");
	s_("if-unmodified-since");
	s_("max-forwards");
	s_("proxy-authorization");
	s_("range");
	s_("referer");
	s_("te");
	s_("user-agent");

	/* entity */
	s_("content-encoding");
	s_("content-language");
	s_("content-length");
	s_("content-location");
	s_("content-md5");
	s_("content-range");
	s_("content-type");
	s_("expires");
	s_("last-modified");

#undef s_

	qu_module_add_function (m, "nobody"   , qu_http_nobody   ,  1);
	qu_module_add_function (m, "reason"   , qu_http_reason   ,  1);
	qu_module_add_function (m, "kind"     , qu_http_kind     ,  1);
	qu_module_add_function (m, "rangesize", qu_http_rangesize,  4);

	return 1;
}
