/* XmlParse.c - Copyright (c) Marc Krisnanto */

#include "qu.h"
#include "XmlParse.h"


/*) module XmlParse
	This module provides an interface to the Expat XML parser library.

	To create a parser, you subclass the ${XmlParse} class and define
	the methods (handlers) are required.

	@xmp
		use Xml.XmlParse

		class Demo is XmlParse

			sub _start (name, attrs)
				println ('\t' * self.depth name)
			;;

			sub _data (s)
				println ('\t' * self.depth '"' s '"') if s
			;;
		;;

		TheContent =
		'<soap:Envelope xmlns:soap="http://www.w3.org/2001/12/soap-envelope" soap:encodingStyle="http://www.w3.org/2001/12/soap-encoding">
			<soap:Body xmlns:m="http://www.example.org/stock">
				<m:GetStockPrice>
					<m:StockName>IBM</m:StockName>
				</m:GetStockPrice>
			</soap:Body>
		</soap:Envelope>
		'

		x = Demo (' ')
		x.parse (TheContent)
		println

		x = Demo ()
		x.parse (TheContent)

	@out
		soap:Envelope
			soap:Body
				m:GetStockPrice
					m:StockName
						"IBM"

		http://www.w3.org/2001/12/soap-envelope Envelope
			http://www.w3.org/2001/12/soap-envelope Body
				http://www.example.org/stock GetStockPrice
					http://www.example.org/stock StockName
						"IBM"

	@p
	If you don't have the Expat library installed on your system then this
	module simply don't work and Qu will give you an exception if you try
	to k{use} it. Once you have Expat installed then this module will work
	automatically. That is, you don't need to recompile Qu nor XmlParsec.c.
*/

/*) enum Error
		XML_ERROR_NONE
		XML_ERROR_NO_MEMORY
		XML_ERROR_SYNTAX
		XML_ERROR_NO_ELEMENTS
		XML_ERROR_INVALID_TOKEN
		XML_ERROR_UNCLOSED_TOKEN
		XML_ERROR_PARTIAL_CHAR
		XML_ERROR_TAG_MISMATCH
		XML_ERROR_DUPLICATE_ATTRIBUTE
		XML_ERROR_JUNK_AFTER_DOC_ELEMENT
		XML_ERROR_PARAM_ENTITY_REF
		XML_ERROR_UNDEFINED_ENTITY
		XML_ERROR_RECURSIVE_ENTITY_REF
		XML_ERROR_ASYNC_ENTITY
		XML_ERROR_BAD_CHAR_REF
		XML_ERROR_BINARY_ENTITY_REF
		XML_ERROR_ATTRIBUTE_EXTERNAL_ENTITY_REF
		XML_ERROR_MISPLACED_XML_PI
		XML_ERROR_UNKNOWN_ENCODING
		XML_ERROR_INCORRECT_ENCODING
		XML_ERROR_UNCLOSED_CDATA_SECTION
		XML_ERROR_EXTERNAL_ENTITY_HANDLING
		XML_ERROR_NOT_STANDALONE
		XML_ERROR_UNEXPECTED_STATE
	;;
	Possible error codes.
*/


/*) enum ContentType
		CTYPE_EMPTY
		CTYPE_ANY
		CTYPE_MIXED
		CTYPE_NAME
		CTYPE_CHOICE
		CTYPE_SEQ
	;;
	Possible content type value.
*/


/*) enum ContentQuant
		CQUANT_NONE
		CQUANT_OPT
		CQUANT_REP
		CQUANT_PLUS
	;;
	Possible quant value.
*/




/*----------------------------------------------------------------------------
	class XmlParse
----------------------------------------------------------------------------*/

/*) class XmlParse
	Use this class as a base to create your own XML parser class.

	To process the document you simply provide the methods (handlers).
	The methods are documented here but the XmlParse class does not actually
	implement them. Your class should implement them as necessary.

	See the %{XmlParse} module documentation for an example and also the
	#{parse} method.
*/
static qu_class QuXmlParseClass;

typedef struct qu_xmlparse_s qu_xmlparse_s;
struct qu_xmlparse_s
{
	QU_OBJECT_HEAD;
	XML_Parser parser;  /* the actual parser */
	int        depth;   /* current depth */
	int        err;     /* last error code */
	qu_class   eclass;  /* exception class */
	qu_any     emsg;    /* the message, may be 0 */
};

typedef qu_xmlparse_s *qu_xmlparse;

#define QU_XMLPARSE_FLAG_TRIM QU_FLAG_0
#define qu_xmlparse_is_trim(self)    qu_flag_test  (self, QU_XMLPARSE_FLAG_TRIM)
#define qu_xmlparse_set_trim(self)   qu_flag_set   (self, QU_XMLPARSE_FLAG_TRIM)
#define qu_xmlparse_unset_trim(self) qu_flag_unset (self, QU_XMLPARSE_FLAG_TRIM)

#define qu_xmlparse_is(x) qu_object_instof (x, QuXmlParseClass)

/* handler names */
static qu_string Sstart;
static qu_string Send;
static qu_string Sdata;
static qu_string Sinstruction;
static qu_string Scomment;
static qu_string Sstartcdata;
static qu_string Sendcdata;
static qu_string Sdefault;
static qu_string Sdefaultexpand;
static qu_string Sstartns;
static qu_string Sendns;
static qu_string Sxml;
static qu_string Sstartdoc;
static qu_string Senddoc;
static qu_string Selement;
static qu_string Sattlist;
static qu_string Sentity;
static qu_string Snotation;
static qu_string Snotalone;
static qu_string Sexternref;


static qu_sub qu_xmlparse__gethandler (qu_xmlparse self, qu_string name)
{
	qu_sub f = qu_object_getattr (self, name, 0);
	if (f && qu_sub_is (f) && !qu_sub_is_class (f))
		return f;
	return 0;
}


#define SKIPONERROR(self) \
QU_BEGIN \
	if ((self)->eclass) \
		return; \
QU_END

#define SKIPONERRORV(self, retval) \
QU_BEGIN \
	if ((self)->eclass) \
		return retval; \
QU_END


#define CALLHANDLER(self, expr) \
QU_BEGIN \
	if ((expr) == 0) \
	{ \
		qu_thread t = qu_thread_self (); \
		(self)->eclass = t->eclass; \
		(self)->emsg = t->emsg; \
	} \
QU_END


#define CALLHANDLERV(self, var, expr) \
QU_BEGIN \
	var = expr; \
	if ((var) == 0) \
	{ \
		qu_thread t = qu_thread_self (); \
		(self)->eclass = t->eclass; \
		(self)->emsg = t->emsg; \
	} \
QU_END



/*
	better use our own
*/
static void *qu_xmlparse_malloc  (size_t size)          { return qu_mem_alloc (size); }
static void *qu_xmlparse_realloc (void *p, size_t size) { return qu_mem_realloc (p, size); }
static void  qu_xmlparse_free    (void *p)              { return qu_mem_free (p); }

static XML_Memory_Handling_Suite qu_xmlparse_memsuite =
{
	qu_xmlparse_malloc,
	qu_xmlparse_realloc,
	qu_xmlparse_free
};



/*----------------------------------------------------------------------------
	handlers
----------------------------------------------------------------------------*/

/*) sub _startdoc (doctypename: String, sysid: Any, pubid: Any, has_internal_subset: Boolean) -> Null
	Called at the start of a DOCTYPE declaration, before any external
	or internal subset is parsed.

	Both v{sysid} and v{pubid} may be k{nil} or a String.

	The v{has_internal_subset} will be k{true} if the DOCTYPE declaration
	has an internal subset, k{false} otherwise.
*/
static void qu_xmlparse_startdoc (qu_xmlparse self, const char *doctypename,
	const char *sysid, const char *pubid, int has_internal_subset)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sstartdoc);

		if (f)
		{
			qu_any argv [4] =
			{
				qu_string_newz (doctypename),
				sysid ? (qu_any) qu_string_newz (sysid) : (qu_any) QuNil,
				pubid ? (qu_any) qu_string_newz (pubid) : (qu_any) QuNil,
				has_internal_subset ? QuTrue : QuFalse
			};
			CALLHANDLER (self, qu_sub_calln (f, self, 4, argv));
		}
	}
}



/*) sub _enddoc -> Null
	Called at the end of a DOCTYPE declaration, after parsing any external
	subset.
*/
static void qu_xmlparse_enddoc (qu_xmlparse self)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Senddoc);

		if (f)
			CALLHANDLER (self, qu_sub_call0 (f, self));
	}
}



/*) sub _start (name: String, attrs: Dict) -> Null
	Called when a start tag is encountered.

	I{name} is the tag name and v{attrs} is a Dict object containing the
	attributes.

	Note that an empty tag such as C{<foo bar=1/>} generates a call to both
	f{_start} and #{_end}.

	@xmp
		<foo bar=1>...

		--> name = 'foo'
			attrs = ['bar': '1']
*/
static void qu_xmlparse_start (qu_xmlparse self, const char *name, const char **attrs)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sstart);

		if (f)
		{
			qu_dict     d;
			int         i, n;
			const char *k, *v;

			for (i = n = 0; attrs [i]; i += 2, n++) {};
			d = qu_dict_new (n);

			for (i = 0; attrs [i]; i += 2)
			{
				k = attrs [i];
				v = attrs [i + 1];
				qu_dict_setitem (d, qu_string_newz (k), qu_string_newz (v));
			}
			CALLHANDLER (self, qu_sub_call2 (f, self, qu_string_newz (name), d));
		}
	}
	self->depth++;
}



/*) sub _end (name: String) -> Null
	Called when a closing tag is encountered.

	v{name} is the tag name.

	@xmp
		</foo>

		--> name = "foo"
*/
static void qu_xmlparse_end (qu_xmlparse self, const char *tag)
{
	SKIPONERROR (self);
	self->depth--;
	{
		qu_sub f = qu_xmlparse__gethandler (self, Send);

		if (f)
			CALLHANDLER (self, qu_sub_call1 (f, self, qu_string_newz (tag)));
	}
}



/*) sub _data (content: String) -> Null
	Called when data between two tags is encountered.

	v{content} is any characters between the tags.

	@warn
	The string is either trimmed or as is depending on the argument you
	pass to #{parse}.

	@xmp
		<foo> bar </foo>

		--> content = "bar"
*/
static void qu_xmlparse_data (qu_xmlparse self, const char *data, int l)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sdata);

		if (f)
		{
			char *s = (char *) data;

			if (qu_xmlparse_is_trim (self))
			{
				while (qu_ascii_isspace (*s))
				{
					++s;
					--l;
				}
				if (l > 0)
				{
					char *p = s + l;
					while (p > s && qu_ascii_isspace (p[-1]))
					{
						--l;
						--p;
					}
				}
			}
			CALLHANDLER (self, qu_sub_call1 (f, self, qu_string_new (s, l)));
		}
	}
}



/*) sub _instruction (target: String, data: String) -> Null
	Called when a processing instruction is encountered.

	v{target} is the first word in the processing instruction.

	v{data} is the rest of the characters in it after skipping all whitespace
	after the initial word.

	@xmp
		<?foo bar="1" baz="2"?>

		--> target = 'foo'
			data   = 'bar="1" baz="2"'
*/
static void qu_xmlparse_instruction (qu_xmlparse self, const char *target, const char *data)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sinstruction);

		if (f)
			CALLHANDLER (self, qu_sub_call2 (f, self, qu_string_newz (target), qu_string_newz (data)));
	}
}



/*) sub _comment (content: String) -> Null
	Called when a comment is encountered.

	v{content} is the content between the opening and closing tags.

	@xmp
		<!-- hello -->

			--> content = ' hello '
*/
static void qu_xmlparse_comment (qu_xmlparse self, const char *s)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Scomment);

		if (f)
			CALLHANDLER (self, qu_sub_call1 (f, self, qu_string_newz (s)));
	}
}



/*) sub _startcdata -> Null
	Called at the beginning of a CDATA section.

	@xmp
		<![CDATA[ ... ]]>
		|
		start
*/
static void qu_xmlparse_startcdata (qu_xmlparse self)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sstartcdata);

		if (f)
			CALLHANDLER (self, qu_sub_call0 (f, self));
	}
}



/*) sub _endcdata -> Null
	Called at the end of a CDATA section.

	@xmp
		<![CDATA[ ... ]]>
		              |
		              end
*/
static void qu_xmlparse_endcdata (qu_xmlparse self)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sendcdata);

		if (f)
			CALLHANDLER (self, qu_sub_call0 (f, self));
	}
}



/*) sub _default (content: String) -> Null
	Called for any characters in the document which would not otherwise be
	handled.

	This includes both data for which no handlers can be set (like some
	kinds of DTD declarations) and data which could be reported but which
	currently has no handler set. Note that a contiguous piece of data that
	is destined to be reported to the default handler may actually be reported
	over several calls to the handler. Setting the handler with this call has
	the side effect of turning off expansion of references to internally
	defined general entities. Instead these references are passed to the default
	handler.
*/
static void qu_xmlparse_default (qu_xmlparse self, const char *s, int len)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sdefault);
		if (f)
			CALLHANDLER (self, qu_sub_call1 (f, self, qu_string_new (s, len)));
	}
}



/*) sub _defaultexpand (content: String) -> Null
	The default handler but does not inhibit expansion of internal entities.
	The entity reference will not be passed to the default handler.

	@xmp
		<!DOCTYPE greeting SYSTEM "hello.dtd">

		--> '<!DOCTYPE'
		    ' '
			'greeting'
			' '
			'SYSTEM'
			' '
			'"hello.dtd"'
			'>'
*/
static void qu_xmlparse_defaultexpand (qu_xmlparse self, const char *s, int len)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sdefaultexpand);

		if (f)
			CALLHANDLER (self, qu_sub_call1 (f, self, qu_string_new (s, len)));
	}
}



/*) sub _startns (prefix: String, uri: String) -> Null
	Called when a namespace is declared.

	Namespace declarations occur inside start tags. But the namespace declaration
	start handler is called before the start tag handler for each namespace
	declared in that start tag.

	@xmp
		<soap:Envelope xmlns:soap="http://www.w3.org/2001/12/soap-envelope" soap:encodingStyle="http://www.w3.org/2001/12/soap-encoding">
		|
		+-- prefix = 'soap'
		    uri    = 'http://www.w3.org/2001/12/soap-envelope'
*/
static void qu_xmlparse_startns (qu_xmlparse self, const char *prefix, const char *uri)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sstartns);

		if (f)
			CALLHANDLER (self, qu_sub_call2 (f, self, qu_string_newz (prefix), qu_string_newz (uri)));
	}
}



/*) sub _endns (prefix: String) -> Null
	Called when leaving the scope of a namespace declaration.

	This will be called, for each namespace declaration, after the handler for
	the end tag of the element in which the namespace was declared.

	@xmp
		</soap:Envelope>

			--> prefix = 'soap'
*/
static void qu_xmlparse_endns (qu_xmlparse self, const char *prefix)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sendns);

		if (f)
			CALLHANDLER (self, qu_sub_call1 (f, self, qu_string_newz (prefix)));
	}
}



/*) sub _xml (version: Any, encoding: Any, standalone: Int) -> Null
	Called for XML declarations and also for text declarations discovered
	in external entities.

	The way to distinguish is that the v{version} parameter will be k{nil}
	for declarations, a String otherwise.

	The v{encoding} parameter may be k{nil} for an XML declaration, a String
	otherwise.

	The v{standalone} argument will contain -1, 0, or 1 indicating respectively
	that there was no standalone parameter in the declaration, that it was given
	as no, or that it was given as yes.

	@xmp
		<?xml version="1.0" encoding="US-ASCII"?

		--> version    = '1.0'
			encoding   = 'US-ASCII'
			standalone = -1
*/
static void qu_xmlparse_xml (qu_xmlparse self, const char *version, char *encoding, int standalone)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sxml);

		if (f)
			CALLHANDLER (self, qu_sub_call3 (f, self,
				version  ? (qu_any) qu_string_newz (version)  : (qu_any) QuNil,
				encoding ? (qu_any) qu_string_newz (encoding) : (qu_any) QuNil,
				qu_int_new (standalone)));
	}
}



/*) sub _element (name: String, content: Array)
	This is called for an element declaration.

	The Array will have 5 items:
	@ul
	.type        one of !{ContentType}
	.quant       one of !{ContentQuant}
	.name        (String)
	.numchildren (Int)
	.children    (Array)

	@p
	See the XML Expat documentation for details.
*/
static void qu_xmlparse__make_element (qu_array p, XML_Content *c, int depth)
{
	qu_array_push (p, qu_int_new (c->type));
	qu_array_push (p, qu_int_new (c->quant));
	qu_array_push (p, c->name ? (qu_any) qu_string_newz (c->name) : (qu_any) QuNil);
	qu_array_push (p, qu_int_new (c->numchildren));

#define MAX_ELEMENT_DEPTH 32
	if (depth > MAX_ELEMENT_DEPTH)
		qu_array_push (p, QuNil);
	else
	{
		qu_array     a = qu_array_new (5);
		int          i;
		XML_Content *g;

		qu_array_push (p, a);
		for (i = 0, g = c->children; i < c->numchildren; i++, g++)
			qu_xmlparse__make_element (a, g, depth + 1);
	}
}


static void qu_xmlparse_element (qu_xmlparse self, const char *name, XML_Content *c)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Selement);

		if (f)
		{
			volatile qu_array a = qu_array_new (5);

			qu_xmlparse__make_element (a, c, 0);
			XML_FreeContentModel (self->parser, c);
			CALLHANDLER (self, qu_sub_call2 (f, self, qu_string_newz (name), a));
		}
	}
}


/*) sub _attlist (elname: String, atname: String, atype: String, dflt: Any, required: Boolean) -> Null
	Called for attlist declarations in the DTD. This handler is called
	for each attribute. So a single attlist declaration with multiple
	attributes declared will generate multiple calls to this handler.

	The v{elname} parameter returns the name of the element for which the
	attribute is being declared. The attribute name is in the v{atname}
	parameter. The attribute type is in the v{atype} parameter.
	It is the string representing the type in the declaration with
	whitespace removed.

	The v{dflt} parameter holds the default value. It will be k{nil}
	in the case of "#IMPLIED" or "#REQUIRED" attributes. You can distinguish
	these two cases by checking the v{required} parameter, which will be k{true}
	in the case of "#REQUIRED" attributes. Attributes which are "#FIXED" will
	have also have a k{true} value, but they will have the non-empty string value
	in the v{dflt} parameter.
*/
static void qu_xmlparse_attlist (qu_xmlparse self, const char *elname, const char *attname,
	const char *attype, const char *dflt, int required)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sattlist);

		if (f)
		{
			qu_any argv [5] =
			{
				qu_string_newz (elname),
				qu_string_newz (attname),
				qu_string_newz (attype),
				dflt ? (qu_any) qu_string_newz (dflt) : QuNil,
				required ? QuTrue : QuFalse
			};
			CALLHANDLER (self, qu_sub_calln (f, self, 5, argv));
		}
	}
}



/*) sub _entity (name: String, isparam: Boolean, value: Any, base: Any, sysid: Any, pubid: Any, ntname: Any) -> Null
	Called for all entity declarations.

	The v{isparam} argument will be k{true} in the case of parameter entities
	and k{false} otherwise.

	For internal entities (<!ENTITY foo "bar">), v{value} will be non-empty
	strings and v{sysid}, v{pubid}, and v{ntname} will all be k{nil}.

	Do not use k{value}.len to test for internal entities, since it is legal
	to have zero-length values. Instead check for whether or not v{value}
	is k{nil}.

	The v{ntname} argument will be a string for unparsed entity declarations,
	k{nil} otherwise.
*/
static void qu_xmlparse_entity (qu_xmlparse self, const char *entname,
	int isparam, const char *value, int valen,
	const char *base, const char *sysid,
	const char *pubid, const char *ntname)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sentity);

		if (f)
		{
			qu_any argv [7] =
			{
				qu_string_newz (entname),
				isparam ? QuTrue : QuFalse,
				value ? (qu_any) qu_string_new (value, valen) : (qu_any) QuNil,
				qu_string_newz (base),
				qu_string_newz (sysid),
				pubid ? (qu_any) qu_string_newz (pubid) : (qu_any) QuNil,
				ntname ? (qu_any) qu_string_newz (ntname) : (qu_any) QuNil
			};
			CALLHANDLER (self, qu_sub_calln (f, self, 7, argv));
		}
	}
}



/*) sub _notation (name: String, base: String, sysid: String, pubid: String) -> Null
	Called to receive notation declarations.
*/
static void qu_xmlparse_notation (qu_xmlparse self, const char *name,
	const char *base, const char *sysid, const char *pubid)
{
	SKIPONERROR (self);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Snotation);

		if (f)
		{
			qu_any argv [4] =
			{
				qu_string_newz (name),
				qu_string_newz (base),
				qu_string_newz (sysid),
				qu_string_newz (pubid)
			};
			CALLHANDLER (self, qu_sub_calln (f, self, 4, argv));
		}
	}
}



/*) sub _notalone -> Boolean
	Called if the document is not "standalone".

	This happens when there is an external subset or a reference to a parameter
	entity, but does not have standalone set to "yes" in an XML declaration.
	If this method returns k{false}, then the parser will internally throw an
	XML_ERROR_NOT_STANDALONE error.

	The defaul method returns true.
*/
static int qu_xmlparse_notalone (qu_xmlparse self)
{
	SKIPONERRORV (self, 1);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Snotalone);

		if (f)
		{
			qu_any r;

			CALLHANDLERV (self, r, qu_sub_call0 (f, self));
			if (r == 0)
				return 0;
			if (r == QuTrue)
				return 1;
			if (r != QuFalse)
				QU_SYS_WARN_RETVAL (self, Snotalone->str, "Boolean");
		}
	}
	return 1;
}



/*) sub _externref (context: String, base: String, systemid: String, publicid: String) -> Boolean
	Called for a reference to an external parsed general entity.
	The referenced entity is not automatically parsed.
	Use #{external_entity_parser_create} for that.

	The v{systemid} argument is the system identifier as specified in the entity
	declaration. The base argument is the system identifier that should be used
	as the base for resolving v{systemid} if it was relative; this is set by
	#{base}; it may be nil.

	The v{publicid} argument is the public identifier as specified in the entity
	declaration, or nil if none was specified; the whitespace in the public
	identifier will have been normalized as required by the XML spec.

	The v{context} argument specifies the parsing context in the format
	expected by the context argument to #{external_entity_parser_create};
	context is valid only until the handler returns, so if the referenced
	entity is to be parsed later, it must be copied.

	The handler should return false if processing should not continue because of
	a fatal error in the handling of the external entity. In this case the
	calling parser will return an ERROR_EXTERNAL_ENTITY_HANDLING error.
*/


static int qu_xmlparse_externref (XML_Parser p, const XML_Char *context,
	const XML_Char *base, const XML_Char *sysid, const XML_Char *pubid)
{
	qu_xmlparse self = XML_GetUserData (p);

	SKIPONERRORV (self, 1);
	{
		qu_sub f = qu_xmlparse__gethandler (self, Sexternref);

		if (f)
		{
			qu_any r;

			CALLHANDLERV (self, r, qu_sub_call4 (f, self,
				qu_string_newz (context),
				base  ? (qu_any) qu_string_newz (base) : (qu_any) QuNil,
				sysid ? (qu_any) qu_string_newz (sysid) : (qu_any) QuNil,
				qu_string_newz (pubid)));

			if (r == 0)
				return 0;
			if (r == QuTrue)
				return 1;
			if (r != QuFalse)
				QU_SYS_WARN_RETVAL (self, Sexternref->str, "Boolean");
		}
	}
	return 1;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (namespace_separator = '', entity_parsing = "n", *arg) -> XmlParse __final
	Create a new XmlParse parser instance.

	If v{namespace_separator} is non-empty then namespace processing is
	in effect. The separator used is the first character. If it is C{'\0'}
	then the namespace URI and the local part will be concatenated without
	any separator.  When a namespace is not declared, the name and prefix
	will be passed through without expansion.

	v{entity_parsing} must be one of:
	@ul
	C{'n'}: never
	C{'a'}: always
	C{'s'}: if not standalone
*/
static qu_xmlparse qu_xmlparse__inst (qu_class self, int argc, qu_any *argv)
{
	XML_Parser  p;
	qu_xmlparse x;
	qu_string   sep = 0;
	int         ep = XML_PARAM_ENTITY_PARSING_NEVER;

	if (argc)
	{
		QU_ARG_VTYPE (0, sep, qu_string);
		if (sep->len == 0)
			sep = 0;
		if (argc > 1)
		{
			qu_string e;
			QU_ARG_VTYPE (1, e, qu_string);
			switch (e->str[0])
			{
				case 's': case 'S': ep = XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE; break;
				case 'a': case 'A': ep = XML_PARAM_ENTITY_PARSING_ALWAYS; break;
			}
		}
	}

#if 0
	if (sep)
		p = XML_ParserCreateNS (NULL, sep->str[0]);
	else
		p = XML_ParserCreate (NULL);
#else
	p = XML_ParserCreate_MM (NULL, &qu_xmlparse_memsuite, sep ? sep->str : 0);
#endif

	QU_REQUIRE (QuEMemory, p);

	x = qu_gc_new (self);
	x->parser = p;
	x->depth  = 0;
	x->eclass = 0;
	x->emsg   = 0;
	x->err    = 0;

	XML_SetUserData (p, x);
	XML_SetParamEntityParsing (p, ep);

	XML_SetStartElementHandler          (p, (void *) qu_xmlparse_start);
	XML_SetEndElementHandler            (p, (void *) qu_xmlparse_end);
	XML_SetProcessingInstructionHandler (p, (void *) qu_xmlparse_instruction);
	XML_SetCharacterDataHandler         (p, (void *) qu_xmlparse_data);
	XML_SetEntityDeclHandler            (p, (void *) qu_xmlparse_entity);
	XML_SetNotationDeclHandler          (p, (void *) qu_xmlparse_notation);
	XML_SetStartNamespaceDeclHandler    (p, (void *) qu_xmlparse_startns);
	XML_SetEndNamespaceDeclHandler      (p, (void *) qu_xmlparse_endns);
	XML_SetCommentHandler               (p, (void *) qu_xmlparse_comment);
	XML_SetStartCdataSectionHandler     (p, (void *) qu_xmlparse_startcdata);
	XML_SetEndCdataSectionHandler       (p, (void *) qu_xmlparse_endcdata);
	XML_SetDefaultHandler               (p, (void *) qu_xmlparse_default);
	XML_SetDefaultHandlerExpand         (p, (void *) qu_xmlparse_defaultexpand);
	XML_SetNotStandaloneHandler         (p, (void *) qu_xmlparse_notalone);
	XML_SetExternalEntityRefHandler     (p, (void *) qu_xmlparse_externref);
	XML_SetStartDoctypeDeclHandler      (p, (void *) qu_xmlparse_startdoc);
	XML_SetEndDoctypeDeclHandler        (p, (void *) qu_xmlparse_enddoc);
	XML_SetXmlDeclHandler               (p, (void *) qu_xmlparse_xml);
	XML_SetElementDeclHandler           (p, (void *) qu_xmlparse_element);
	XML_SetAttlistDeclHandler           (p, (void *) qu_xmlparse_attlist);
	return x;
}



/*) sub class new (*arg) -> Xml
	Create a new XML parser instance then call the f{__init} method passing
	v{arg} as the arguments.
*/
static qu_xmlparse qu_xmlparse_new (qu_class self, int argc, qu_any *argv)
{
	qu_xmlparse x = qu_xmlparse__inst (self, argc, argv);
	QU_ENSURE (x);
	QU_RETURN (qu_class_init_new (self, x, argc, argv));
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class extern (parser: XmlParse, context: String, encoding = '') -> XmlParse
	Creates a parser object that can parse an external general entity;

	v{context} is a string specifying the parse context.

	v{encoding} is a string giving the name of the externally specified encoding
	or empty if there is none.

	The context string consists of a sequence of tokens separated by formfeeds
	C{\f}; a token consisting of a name specifies that the general entity of
	the name is open; a token of the form C{prefix=uri} specifies the namespace
	for a particular prefix; a token of the C{form =uri} specifies the default
	namespace.  This can be called at any point after the first call to an
	#{_externref} handler as long as the parser has not yet been freed.

	The new parser is completely independent and may safely be used in a separate
	thread.

	@exc
	.EArgValue
*/
static qu_xmlparse qu_xmlparse_extern (qu_class self, int argc, qu_any *argv)
{
	qu_xmlparse parser, x;
	qu_string   context, encoding = 0;
	XML_Parser  p;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, parser, qu_xmlparse);
	QU_ARG_VTYPE (0, context, qu_string);
	if (argc > 2)
	{
		QU_ARG_VTYPE (2, encoding, qu_string);
		if (encoding->len == 0)
			encoding = 0;
	}

	p = XML_ExternalEntityParserCreate (parser->parser, context->str, encoding ? encoding->str : 0);
	QU_REQUIRE (QuEMemory, p);

	x = qu_gc_new (self);
	x->parser = p;
	x->depth  = 0;
	x->eclass = 0;
	x->emsg   = 0;
	x->err    = 0;

	XML_SetUserData (p, x);
	return x;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu_xmlparse__destroy (qu_xmlparse self)
{
	if (self->parser)
		XML_ParserFree (self->parser);
	qu_gc_xmark (self->eclass);
	qu_gc_xmark (self->emsg);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub attrcount -> Int
	Return the number of the attribute/value pairs passed in last call
	to the _start handler that were specified in the start-tag
	rather than defaulted. Each attribute/value pair counts as 2; thus
	this correspondds to an index into the v{attrs} array passed to the
	handler.

	@noexc
*/
static qu_int qu_xmlparse_attrcount (qu_xmlparse self)
{
	return qu_int_newi (XML_GetSpecifiedAttributeCount (self->parser));
}



/*) sub attrindex -> Int
	Returns the index of the ID attribute passed in the last call to
	#{_start}, or -1 if there is no ID attribute. Each attribute/value pair
	counts as 2; thus this correspondds to an index into the v{attrs} array
	passed to the handler.

	@noexc
*/
static qu_int qu_xmlparse_attrindex (qu_xmlparse self)
{
	return qu_int_newi (XML_GetIdAttributeIndex (self->parser));
}



/*) sub base (*arg) -> String
	@def
	sub base (url: String) -> String
	sub base -> String
	@p
	Get or set the base URL.

	@exc
	.EArgValue
	.EMemory
*/
static qu_string qu_xmlparse_base (qu_xmlparse self, int argc, qu_any *argv)
{
	if (argc)
	{
		qu_string base;
		QU_ARG_VTYPE (0, base, qu_string);
		QU_REQUIRE (QuEMemory, XML_SetBase (self->parser, base->str));
		return base;
	}
	else
	{
		const char *base = XML_GetBase (self->parser);
		if (base == 0)
			return QuStringEMPTY;
		return qu_string_newz (base);
	}
}



/*) sub bytecount -> Integer
	Return the number of bytes in the current event.

	Returns 0 if the event is in an internal entity.

	@noexc
*/
static qu_integer qu_xmlparse_bytecount (qu_xmlparse self)
{
	return qu_integer_newi (XML_GetCurrentByteCount (self->parser));
}



/*) sub depth -> Integer
	Return the depth of the current tag.

	@noexc
*/
static qu_integer qu_xmlparse_depth (qu_xmlparse self)
{
	return qu_integer_newi (self->depth);
}



/*) sub error -> Int
	Get the last error code.

	This is one of !{ERROR}.

	@noexc
*/
static qu_int qu_xmlparse_error (qu_xmlparse self)
{
	return qu_int_newi (self->err);
}



/*) sub parse (source: String, encoding = '', trim = true, triplet = false)
	Parse an XML document.

	v{source} is the source and must be a String. If the source contains errors
	then EValue is thrown with the appropriate XML error message.

	v{encoding} is the name of the encoding. An empty string means to use
	the default (UTF-8).

	v{trim} is whether or not to trim data string (remove leading and
	trailing blanks). See #{_data}.

	If v{triplet} is true and namespace processing is in effect, and
	a name has a prefix (i.e. an explicit namespace qualifier) then
	that name is returned as a triplet in a single
	string separated by the separator character specified when the parser
	was created: C{URI + sep + local_name + sep + prefix}.

	Otherwise namespace information is returned in the default manner
	C{URI + sep + local_name} whether or not the names has a prefix.

	If any of the handler throw an exception then parsing stops and
	that exception is rethrown. A special case is EStop which will
	simply cause a stop without throwing any exception.

	@exc
	.EArgValue
	.EValue     XML error (see #{error} for the code)
	.Exception  anything thrown by the handlers except EStop

	@xmp
		use Xml.XmlParse

		class Node
			var parent
			var childs
			var name

			sub __init (name, parent)
				self.name   = name
				self.parent = parent
				self.childs = []
			;;

			sub view (level = 0)
				print ('    ') for __ in 1, level
				println (self.name)
				c.view (level + 1) for c in self.childs
			;;
		;;

		class Demo is XmlParse
			var top
			var curnode

			sub __init (*arg)
				self.top = self.curnode = Node ('ROOT', nil)
			;;

			sub _start (name, attrs)
				c = self.curnode
				n = Node (name, c)
				c.childs <<< n
				self.curnode = n
			;;

			sub _end (name)
				c = self.curnode
				self.curnode = c.parent
				throw EStop
			;;
		;;

		TheContent =
		'<?xml version="1.0" encoding="US-ASCII"?>
		<soap:Envelope xmlns:soap="http://www.w3.org/2001/12/soap-envelope" soap:encodingStyle="http://www.w3.org/2001/12/soap-encoding">
			<soap:Body xmlns:m="http://www.example.org/stock">
				<m:GetStockPrice>
					<m:StockName>Foo</m:StockName>
					<m:StockName>Bar</m:StockName>
					<m:StockName>Baz</m:StockName>
				</m:GetStockPrice>
			</soap:Body>
		</soap:Envelope>
		'
		x = Demo ()
		x.parse (TheContent)
		x.top.view

	@top
		ROOT
			soap:Envelope
				soap:Body
					m:GetStockPrice
						m:StockName
*/
static qu_null qu_xmlparse_parse (qu_xmlparse self, int argc, qu_any *argv)
{
	static char *argx[] =
	{
		"s", "source",
		":",
		"s", "encoding",
		"b", "trim",
		"b", "triplet",
		NULL
	};

	qu_string source, encoding = 0;
	int       trim = 1, triplet = 0;

	QU_ARG_XPARSE (&source, &encoding, &trim, &triplet);

	XML_SetReturnNSTriplet (self->parser, triplet);
	if (encoding && encoding->len)
		XML_SetEncoding (self->parser, encoding->str);

	self->eclass = 0;
	self->emsg = 0;
	self->err = 0;

	if (trim)
		qu_xmlparse_set_trim (self);
	else
		qu_xmlparse_unset_trim (self);

	if (!XML_Parse (self->parser, source->str, source->len, 1))
	{
		self->err = XML_GetErrorCode (self->parser);
		QU_THROWF (QuEValue, "%s line %d",
			XML_ErrorString (self->err),
			XML_GetCurrentLineNumber (self->parser));
	}
	else if (self->eclass && self->eclass != QuEStopClass)
	{
		qu_thread__throw (self->eclass, self->emsg, __FUNCTION__);
		return 0;
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/
static void *QuXmlHandle;

void qu_xmlparse_atexit ()
{
	if (QuXmlHandle)
		qu_dl_close (QuXmlHandle);
}


int qu_XmlParse_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		expat
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuXmlHandle = qu_dl_open_ex ("expat", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (QuXmlHandle);

#define f_(name) name = qu_dl_sym (QuXmlHandle, #name)

	f_(XML_ParserCreate);
	f_(XML_ParserCreateNS);
	f_(XML_ParserCreate_MM);
	f_(XML_SetElementDeclHandler);
	f_(XML_SetAttlistDeclHandler);
	f_(XML_SetXmlDeclHandler);
	f_(XML_SetEntityDeclHandler);
	f_(XML_SetElementHandler);
	f_(XML_SetStartElementHandler);
	f_(XML_SetEndElementHandler);
	f_(XML_SetCharacterDataHandler);
	f_(XML_SetProcessingInstructionHandler);
	f_(XML_SetCommentHandler);
	f_(XML_SetCdataSectionHandler);
	f_(XML_SetStartCdataSectionHandler);
	f_(XML_SetEndCdataSectionHandler);
	f_(XML_SetDefaultHandler);
	f_(XML_SetDefaultHandlerExpand);
	f_(XML_SetDoctypeDeclHandler);
	f_(XML_SetStartDoctypeDeclHandler);
	f_(XML_SetEndDoctypeDeclHandler);
	f_(XML_SetNotationDeclHandler);
	f_(XML_SetNamespaceDeclHandler);
	f_(XML_SetStartNamespaceDeclHandler);
	f_(XML_SetEndNamespaceDeclHandler);
	f_(XML_SetNotStandaloneHandler);
	f_(XML_SetExternalEntityRefHandler);
	f_(XML_SetExternalEntityRefHandlerArg);
	f_(XML_SetUnknownEncodingHandler);
	f_(XML_DefaultCurrent);
	f_(XML_SetReturnNSTriplet);
	f_(XML_SetUserData);
	f_(XML_SetEncoding);
	f_(XML_UseParserAsHandlerArg);
	f_(XML_SetBase);
	f_(XML_GetBase);
	f_(XML_GetSpecifiedAttributeCount);
	f_(XML_GetIdAttributeIndex);
	f_(XML_Parse);
	f_(XML_GetBuffer);
	f_(XML_ParseBuffer);
	f_(XML_ExternalEntityParserCreate);
	f_(XML_SetParamEntityParsing);
	f_(XML_GetErrorCode);
	f_(XML_ErrorString);
	f_(XML_GetCurrentLineNumber);
	f_(XML_GetCurrentColumnNumber);
	f_(XML_GetCurrentByteIndex);
	f_(XML_GetCurrentByteCount);
	f_(XML_GetInputContext);
	f_(XML_ParserFree);
	f_(XML_ExpatVersion);
	f_(XML_FreeContentModel);

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		constants
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#define e_(name) #name, XML_##name,

	qu_var_enum ((qu_symbol) m, "Error",
		e_(ERROR_NONE)
		e_(ERROR_NO_MEMORY)
		e_(ERROR_SYNTAX)
		e_(ERROR_NO_ELEMENTS)
		e_(ERROR_INVALID_TOKEN)
		e_(ERROR_UNCLOSED_TOKEN)
		e_(ERROR_PARTIAL_CHAR)
		e_(ERROR_TAG_MISMATCH)
		e_(ERROR_DUPLICATE_ATTRIBUTE)
		e_(ERROR_JUNK_AFTER_DOC_ELEMENT)
		e_(ERROR_PARAM_ENTITY_REF)
		e_(ERROR_UNDEFINED_ENTITY)
		e_(ERROR_RECURSIVE_ENTITY_REF)
		e_(ERROR_ASYNC_ENTITY)
		e_(ERROR_BAD_CHAR_REF)
		e_(ERROR_BINARY_ENTITY_REF)
		e_(ERROR_ATTRIBUTE_EXTERNAL_ENTITY_REF)
		e_(ERROR_MISPLACED_XML_PI)
		e_(ERROR_UNKNOWN_ENCODING)
		e_(ERROR_INCORRECT_ENCODING)
		e_(ERROR_UNCLOSED_CDATA_SECTION)
		e_(ERROR_EXTERNAL_ENTITY_HANDLING)
		e_(ERROR_NOT_STANDALONE)
		e_(ERROR_UNEXPECTED_STATE)
		NULL);

	qu_var_enum ((qu_symbol) m, "ContentType",
		e_(CTYPE_EMPTY)
		e_(CTYPE_ANY)
		e_(CTYPE_MIXED)
		e_(CTYPE_NAME)
		e_(CTYPE_CHOICE)
		e_(CTYPE_SEQ)
		NULL);

	qu_var_enum ((qu_symbol) m, "ContentQuant",
		e_(CQUANT_NONE)
		e_(CQUANT_OPT)
		e_(CQUANT_REP)
		e_(CQUANT_PLUS)
		NULL);

#undef e_

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Xml
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	QuXmlParseClass = qu_class_newz (m, "XmlParse",qu_xmlparse_s,qu_xmlparse__destroy, 0);

	qu_class_method_class  (QuXmlParseClass, QU_NAME__inst, qu_xmlparse__inst    , -1);
	qu_class_method_class  (QuXmlParseClass, QU_NAME_new  , qu_xmlparse_new      , -1);
	qu_class_method_classz (QuXmlParseClass, "extern"     , qu_xmlparse_extern   , -1);

	qu_class_method_instz  (QuXmlParseClass, "attrcount"  , qu_xmlparse_attrcount,  0);
	qu_class_method_instz  (QuXmlParseClass, "attrindex"  , qu_xmlparse_attrindex,  0);
	qu_class_method_instz  (QuXmlParseClass, "base"       , qu_xmlparse_base     , -1);
	qu_class_method_instz  (QuXmlParseClass, "bytecount"  , qu_xmlparse_bytecount,  0);
	qu_class_method_instz  (QuXmlParseClass, "depth"      , qu_xmlparse_depth    ,  0);
	qu_class_method_instz  (QuXmlParseClass, "error"      , qu_xmlparse_error    ,  0);
	qu_class_method_instz  (QuXmlParseClass, "parse"      , qu_xmlparse_parse    , -1);

	Sstart         = qu_string_new_staticz ("_start");
	Send           = qu_string_new_staticz ("_end");
	Sdata          = qu_string_new_staticz ("_data");
	Sinstruction   = qu_string_new_staticz ("_instruction");
	Scomment       = qu_string_new_staticz ("_comment");
	Sstartcdata    = qu_string_new_staticz ("_startcdata");
	Sendcdata      = qu_string_new_staticz ("_endcdata");
	Sdefault       = qu_string_new_staticz ("_default");
	Sdefaultexpand = qu_string_new_staticz ("_defaultexpand");
	Sstartns       = qu_string_new_staticz ("_startns");
	Sendns         = qu_string_new_staticz ("_endns");
	Sxml           = qu_string_new_staticz ("_xml");
	Sstartdoc      = qu_string_new_staticz ("_startdoc");
	Senddoc        = qu_string_new_staticz ("_enddoc");
	Selement       = qu_string_new_staticz ("_element");
	Sattlist       = qu_string_new_staticz ("_attlist");
	Sentity        = qu_string_new_staticz ("_entity");
	Snotation      = qu_string_new_staticz ("_notation");
	Snotalone      = qu_string_new_staticz ("_notalone");
	Sexternref     = qu_string_new_staticz ("_externref");

	atexit (qu_xmlparse_atexit);
	return 1;
}
