/* Zlib.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Zlib
	This module provides an interface to the zlib library.

	@xmp
		use Zlib

		s = 'foobar' * 5
		println (s ' (' s.len ')')

		r = Zlib.compress (s)
		println (r.quote ' (' r.len ')')

		s = Zlib.decompress (r)
		println (s)

	@out
		foobarfoobarfoobarfoobarfoobar (30)
		'x\x9cK\xcb\xcfOJ,J\xc3A\x02\x00\xbf\xb3\f^' (17)
		foobarfoobarfoobarfoobarfoobar
*/

/*) __doc__ CONSTANTS
	See C{usr/include/zlib.h} for details.

	@ul
	.ZLIB_VERSION
	.MAX_WBITS
	.DEF_MEM_LEVEL
	.Z_DEFLATED
	.Z_BEST_SPEED
	.Z_BEST_COMPRESSION
	.Z_DEFAULT_COMPRESSION
	.Z_FILTERED
	.Z_HUFFMAN_ONLY
	.Z_DEFAULT_STRATEGY
	.Z_FINISH
	.Z_NO_FLUSH
	.Z_SYNC_FLUSH
	.Z_FULL_FLUSH
*/

/*----------------------------------------------------------------------------
	stuffs we need from zlib
----------------------------------------------------------------------------*/

#define ZLIB_VERSION "1.2.1.2"
#define ZLIB_VERNUM 0x1212

#ifndef Z_DEFLATED
#define Z_DEFLATED 8
#endif

#ifndef MAX_WBITS
#  define MAX_WBITS 15
#endif

#ifndef DEF_WBITS
#define DEF_WBITS MAX_WBITS
#endif

#ifndef MAX_MEM_LEVEL
#  ifdef MAXSEG_64K
#    define MAX_MEM_LEVEL 8
#  else
#    define MAX_MEM_LEVEL 9
#  endif
#endif

#if MAX_MEM_LEVEL >= 8
#  define DEF_MEM_LEVEL 8
#else
#  define DEF_MEM_LEVEL  MAX_MEM_LEVEL
#endif

typedef void *(*alloc_func) (void *opaque, unsigned int items, unsigned int size);
typedef void  (*free_func)  (void *opaque, void *address);

struct internal_state;

typedef struct z_stream z_stream;
struct z_stream
{
	char                  *next_in;
	unsigned int           avail_in;
	unsigned long          total_in;
	char                  *next_out;
	unsigned int           avail_out;
	unsigned long          total_out;
	char                  *msg;
	struct internal_state *state;
	alloc_func             zalloc;
	free_func              zfree;
	void                  *opaque;
	int                    data_type;
	unsigned long          adler;
	unsigned long          reserved;
};


#define Z_NO_FLUSH      0
#define Z_PARTIAL_FLUSH 1 /* will be removed, use Z_SYNC_FLUSH instead */
#define Z_SYNC_FLUSH    2
#define Z_FULL_FLUSH    3
#define Z_FINISH        4
#define Z_BLOCK         5

#define Z_OK            0
#define Z_STREAM_END    1
#define Z_NEED_DICT     2
#define Z_ERRNO        (-1)
#define Z_STREAM_ERROR (-2)
#define Z_DATA_ERROR   (-3)
#define Z_MEM_ERROR    (-4)
#define Z_BUF_ERROR    (-5)
#define Z_VERSION_ERROR (-6)

#define Z_NO_COMPRESSION         0
#define Z_BEST_SPEED             1
#define Z_BEST_COMPRESSION       9
#define Z_DEFAULT_COMPRESSION  (-1)

#define Z_FILTERED            1
#define Z_HUFFMAN_ONLY        2
#define Z_RLE                 3
#define Z_DEFAULT_STRATEGY    0

#define Z_BINARY   0
#define Z_ASCII    1
#define Z_UNKNOWN  2

#define Z_DEFLATED   8

#define Z_NULL  0  /* for initializing zalloc, zfree, opaque */


static int           (*deflate   ) (z_stream *s, int flush);
static int           (*deflateEnd) (z_stream *s);
static int           (*inflate   ) (z_stream *s, int flush);
static int           (*inflateEnd) (z_stream *s);
static unsigned long (*adler32   ) (unsigned long adler, const char *buf, unsigned int len);
static unsigned long (*crc32     ) (unsigned long crc, const char *buf, unsigned int len);

static int (*deflateInit_    ) (z_stream *s, int level, const char *version, int stream_size);
static int (*inflateInit_    ) (z_stream *s, const char *version, int stream_size);
static int (*deflateInit2_   ) (z_stream *s, int level, int method, int windowBits, int memLevel, int strategy, const char *version, int stream_size);
static int (*inflateInit2_   ) (z_stream *s, int windowBits, const char *version, int stream_size);
static int (*inflateBackInit_) (z_stream *strm, int windowBits, unsigned char *window, const char *version, int stream_size);

#define deflateInit(strm, level) \
        deflateInit_((strm), (level),       ZLIB_VERSION, sizeof(z_stream))

#define inflateInit(strm) \
        inflateInit_((strm),                ZLIB_VERSION, sizeof(z_stream))

#define deflateInit2(strm, level, method, windowBits, memLevel, strategy) \
        deflateInit2_((strm),(level),(method),(windowBits),(memLevel),\
                      (strategy),           ZLIB_VERSION, sizeof(z_stream))

#define inflateInit2(strm, windowBits) \
        inflateInit2_((strm), (windowBits), ZLIB_VERSION, sizeof(z_stream))

#define inflateBackInit(strm, windowBits, window) \
        inflateBackInit_((strm), (windowBits), (window), \
        ZLIB_VERSION, sizeof(z_stream))


/*----------------------------------------------------------------------------
	our own stuffs
----------------------------------------------------------------------------*/

/* forward */
static qu_class EZlibClass;

#define THROW_EZLIB(z) \
QU_BEGIN \
	if (z.msg) \
		QU_THROWZ (EZlib, z.msg); \
	else \
		QU_THROWZ (EZlib, "unknown Zlib error"); \
QU_END


/* for adler32 and crc32 */
typedef unsigned long (*checksum_f) (unsigned long initval, const char *buf, unsigned long len);

static qu_integer Zlib__checksum (checksum_f f, int argc, qu_any *argv)
{
	qu_stringy    s;
	long          ofs =  0;
	unsigned long v = f (0, 0, 0);

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, s, qu_stringy);
	if (argc > 1)
	{
		if (!qu_null_is (argv[1]))
			QU_ARG_vulong (1, v);

		if (argc > 2)
		{
			QU_ARG_vlong (1, ofs);
			QU_ENSURE_INDEX (ofs, s->len);
		}
	}
	v = f (v, s->str + ofs, s->len - ofs);
	return qu_integer_newul (v);
}



/*----------------------------------------------------------------------------
	functions
----------------------------------------------------------------------------*/

/*) sub compress (s, b = nil, level = 1) -> Any
	Compress (deflate) a String or Buffer.

	If v{b} is not nil then it must be a Buffer object to store the output
	In this case this function returns that Buffer on success.
	Otherwise this function returns a new String.

	@exc
	.EArgValue
	.EMemory
	.EZlib
*/
static qu_stringy Zlib_compress (int argc, qu_any *argv)
{
	qu_stringy in;
	qu_stringy out = 0;
	int        isbuf = 0;
	int        level = Z_DEFAULT_COMPRESSION;
	long       l, m;
    z_stream   z;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, in, qu_stringy);

	l = in->len;
	QU_ARG_CHECKZ (l <= QU_MEM_MAX - l / 1000 - 12 - 1, "input too big");
	m = l + l / 1000 + 12 + 1;

	if (argc > 1)
	{
		out = argv[1];
		if (qu_null_is (out))
		{
			out = 0;
		}
		else
		{
			QU_ARG_TYPE (out, qu_buffer);
			QU_BUFFER_ENSURE_NOT_FROZEN (out);
			QU_ENSURE (qu_buffer_make_space ((qu_buffer) out, m));
			isbuf = 1;
		}

		if (argc > 2)
			QU_ARG_vint (2, level);
	}

	if (l == 0)
	{
		if (out)
			return out;
		return QuStringEMPTY;
	}

	z.zalloc    = 0;
    z.zfree     = 0;
    z.avail_in  = l;
    z.next_in   = in->str;
	z.avail_out = m;

	if (isbuf)
	{
		z.next_out = out->str + out->len;
	}
	else
	{
		out = qu_string_buffer_new (m);
		z.next_out = out->str;
	}

	switch (deflateInit (&z, level))
	{
		case Z_OK:
			break;

		case Z_MEM_ERROR:
			QU_THROW (QuEMemory);

		case Z_STREAM_ERROR:
			QU_ARG_THROWZ ("invalid level");

		default:
			deflateEnd (&z);
			THROW_EZLIB (z);
    }

	if (deflate (&z, Z_FINISH) != Z_STREAM_END)
	{
		deflateEnd (&z);
		THROW_EZLIB (z);
    }

	if (deflateEnd (&z) != Z_OK)
		THROW_EZLIB (z);

	if (isbuf)
	{
		qu_buffer_set_len ((qu_buffer) out, z.total_out + out->len);
		return out;
	}
	else
	{
		qu_string_buffer_set_len (out, z.total_out);
		return qu_string_buffer_return (out);
	}
}



/*) sub decompress (s, b = nil, level = 1) -> Any
	Decompress (inflate) a String or Buffer.

	If v{b} is not nil then it must be a Buffer object to store the output
	In this case this function returns that Buffer on success.
	Otherwise this function returns a new String.

	@exc
	.EArgValue
	.EMemory
	.EZlib
*/
static qu_stringy Zlib_decompress (int argc, qu_any *argv)
{
	qu_stringy in;
	qu_stringy out = 0;
	int        isbuf = 0;
	int        wbits = DEF_WBITS;
	long       l, m;
    z_stream   z;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, in, qu_stringy);

	l = in->len;
	QU_ARG_CHECKZ (l <= QU_MEM_MAX / 4, "input too big");
	m = l * 2;

	if (argc > 1)
	{
		out = argv[1];
		if (qu_null_is (out))
			out = 0;
		else
		{
			QU_ARG_TYPE (out, qu_buffer);
			QU_BUFFER_ENSURE_NOT_FROZEN (out);
			QU_ENSURE (qu_buffer_make_space ((qu_buffer) out, m));
			isbuf = 1;
		}

		if (argc > 2)
			QU_ARG_vint (2, wbits);
	}

	if (l == 0)
	{
		if (out)
			return out;
		return QuStringEMPTY;
	}

	z.zalloc    = 0;
	z.zfree     = 0;
	z.avail_in  = l;
	z.next_in   = in->str;
	z.avail_out = m;

	if (isbuf)
	{
		z.next_out = out->str + out->len;
	}
	else
	{
		out = qu_string_buffer_new (m);
		z.next_out = out->str;
	}

	switch (inflateInit2 (&z, wbits))
	{
		case Z_OK:
			break;

		case Z_MEM_ERROR:
			QU_THROW (QuEMemory);

		default:
			inflateEnd (&z);
			THROW_EZLIB (z);
    }

  again:

	switch (inflate (&z, Z_FINISH))
	{
		case Z_STREAM_END:
			break;

		case Z_BUF_ERROR:
			if (z.avail_out > 0)
			{
				inflateEnd (&z);
				THROW_EZLIB (z);
			}
			/*nobreak*/

		case Z_OK:
			m = z.avail_in * 2;
			if (isbuf)
			{
				qu_buffer_set_len ((qu_buffer) out, z.total_out + out->len);
				if (!qu_buffer_make_space ((qu_buffer) out, m))
				{
					inflateEnd (&z);
					QU_THROW (QuEMemory);
				}
			}
			else
			{
				qu_string_buffer_set_len (out, z.total_out);
				if (!qu_string_buffer_make_space (out, m))
				{
					inflateEnd (&z);
					QU_THROW (QuEMemory);
				}
			}
			z.next_out = out->str + out->len;
			z.avail_out = m;
			goto again;

		default:
			inflateEnd (&z);
			THROW_EZLIB (z);
	}

	if (inflateEnd (&z) != Z_OK)
		THROW_EZLIB (z);

	if (isbuf)
	{
		qu_buffer_set_len ((qu_buffer) out, z.total_out + out->len);
		return out;
	}
	else
	{
		qu_string_buffer_set_len (out, z.total_out);
		return qu_string_buffer_return (out);
	}
}



/*) sub adler32 (s, initval = 0, ofs = 0) -> Integer
	Compute the Adler-32 checksum of the String or Buffer v{s} optionally start
	at offset v{ofs}.

	If v{ofs} is negative then count from the end.

	You can optionally provide an initial value. If v{initval} is nil then
	ignore it.

	@exc
	.EArgValue
	.EIndex

	@xmp
		use Zlib
		println (Zlib.adler32 ("hello world"))

	@out
		436929629
*/
static qu_integer Zlib_adler32 (int argc, qu_any *argv)
{
	return Zlib__checksum ((checksum_f) adler32,  argc, argv);
}



/*) sub crc32 (s, initval = 0, ofs = 0) -> Integer
	Compute the CRC-32 checksum of the String or Buffer v{s} optionally start
	at offset v{ofs}.

	If v{ofs} is negative then count from the end.

	You can optionally provide an initial value.

	@exc
	.EArgValue
	.EIndex

	@xmp
		use Zlib
		println (Zlib.crc32 ("hello world"))

	@out
		222957957
*/
static qu_integer Zlib_crc32 (int argc, qu_any *argv)
{
	return Zlib__checksum ((checksum_f) crc32,  argc, argv);
}



/*============================================================================
	class EZlib
============================================================================*/

/*) class EZlib is EStandard
	This is the standard exception thrown by the Zlib library.
*/



/*============================================================================
	class Deflate
============================================================================*/

/*) class Deflate
	This class provides a convenient way to compress multiple data.

	You start by creating a Deflate object via #{new}, call #{compress}
	for each data you need to compress, and finally call #{finish} to
	end the compression process.

	@xmp
		use Zlib

		d = Zlib.Deflate ()
		b = Buffer (10)
		for __ in 1, 5
			d.compress ('foobar', b)
		;;
		d.finish (b)
		println (b.quote)

		println (d.total_in)
		println (d.total_out)
		println (d.adler)

	@out
		'x\x9cK\xcb\xcfOJ,J\xc3A\x02\x00\xbf\xb3\f^'
		30
		17
		3216182366
*/
static qu_class DeflateClass;

typedef struct TDeflate TDeflate;
struct TDeflate
{
	QU_OBJECT_HEAD;
	z_stream z;
};

typedef struct TDeflate *Deflate;

#define DeflateFlagENDED QU_FLAG_0


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (*arg) -> Deflate __final
	Create a new Deflate object.

	All arguments are optional and given default values if not set.
	@arg
	.level     Z_DEFAULT_COMPRESSION
	.method    Z_DEFLATED
	.wbits     DEF_WBITS
	.memlevel  DEF_MEM_LEVEL
	.strategy  0

	@p
	See the C{man 3 zlib} for details of the arguments.

	@exc
	.EArgValue
	.EMemory
*/
static Deflate Deflate__inst (qu_class self, int argc, qu_any *argv)
{
	static char *argx[] =
	{
		":",
		"i", "level",
		"i", "method",
		"i", "wbits",
		"i", "memlevel",
		"i", "strategy",
		NULL
	};

	int level    = Z_DEFAULT_COMPRESSION;
	int method   = Z_DEFLATED;
	int wbits    = DEF_WBITS;
	int memlevel = DEF_MEM_LEVEL;
	int strategy = 0;

	Deflate o;

	QU_ARG_XPARSE (&level, &method, &wbits, &memlevel, &strategy);

	o = qu_gc_new (self);
	o->z.avail_in = 0;
	o->z.zalloc = 0;
	o->z.zfree  = 0;

	switch (deflateInit2 (&o->z, level, method, wbits, memlevel, strategy))
	{
		case Z_OK:
			break;

		case Z_MEM_ERROR:
			QU_THROW (QuEMemory);

		default:
			QU_ARG_THROW;
	}
    return o;
}



/*) sub class new (*arg) -> Deflate
	Create a new Deflate object.

	This default method simply calls __inst then __init with all the arguments
	if the class defines one.

	See #{__inst} for details.
*/
static Deflate Deflate_new (qu_class self, int argc, qu_any *argv)
{
	Deflate o = Deflate__inst (self, argc, argv);

	QU_ENSURE (o);
	return qu_class_init_new (self, o, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void Deflate__destroy (Deflate self)
{
	if (!qu_flag_test (self, DeflateFlagENDED))
		deflateEnd (&self->z);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub compress (s, b = nil, ofs = 0, len = s.len) -> Any
	Compress the String or Buffer v{s} optionally start at offset and a specified
	length.

	If v{ofs} is negative then count from the end.

	If v{b} is not nil then it must be a Buffer object to store the output
	In this case this method returns that Buffer on success.
	Otherwise this method returns a new String.

	Once you have completed compressing, you must call #{finish} to get the
	final result.

	@warn
	If v{b} is a Buffer then don't forget to consume it.

	@exc
	.EArgValue
	.ECall      #{finish} already called
	.EIndex
	.EMemory
	.EZlib
*/
static qu_stringy Deflate_compress (Deflate self, int argc, qu_any *argv)
{
	qu_stringy    s, out = 0;
	int           isbuf = 0;
	long          ofs = 0;
	long          len, size, l;
	unsigned long total = self->z.total_out;

	QU_REQUIREZ (QuECall, !qu_flag_test (self, DeflateFlagENDED), "already finished");

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, s, qu_stringy);
	len = s->len;
	if (argc > 1)
	{
		out = argv[1];
		if (qu_null_is (out))
			out = 0;
		else
		{
			QU_ARG_TYPE (out, qu_buffer);
			QU_BUFFER_ENSURE_NOT_FROZEN (out);
			isbuf = 1;
		}
		if (argc > 2)
		{
			long max;
			QU_ARG_vlong (2, ofs);
			QU_ENSURE_INDEX (ofs, len);
			if (argc > 3)
			{
				QU_ARG_vlong (3, len);
				QU_ARG_CHECK (len >= 0);
			}
			max = s->len - ofs;
			if (len > max)
				len = max;
		}
	}

	if (len == 0)
	{
		if (isbuf)
			return out;
		return QuStringEMPTY;
	}

	QU_MEM_CHECK (len < QU_MEM_MAX - len / 1000 - 12 - 1);
	size = len + len / 1000 + 12 + 1;

	if (isbuf)
	{
		QU_ENSURE (qu_buffer_make_space ((qu_buffer) out, size));
		self->z.avail_out = qu_buffer_avail ((qu_buffer) out);
		self->z.next_out  = out->str + out->len;
	}
	else
	{
		out = qu_string_buffer_new (size);
		self->z.avail_out = size;
		self->z.next_out  = out->str;
	}

	self->z.avail_in = len;
	self->z.next_in  = s->str + ofs;

  again:

	assert (self->z.avail_out > 0);
	switch (deflate (&self->z, Z_NO_FLUSH))
	{
		case Z_OK:
			if (self->z.avail_out)
				break;

			l = self->z.total_out - total;
			assert (size < QU_MEM_MAX - size);
			size <<= 1;
			if (isbuf)
			{
				l += ((qu_buffer) out)->len;
				qu_buffer_set_len (out, l);
				if (!qu_buffer_make_space ((qu_buffer) out, size))
					QU_THROW (QuEMemory);
			}
			else
			{
				qu_string_buffer_set_len (out, l);
				if (!qu_string_buffer_make_space (out, size))
					QU_THROW (QuEMemory);
			}
			self->z.next_out = out->str + out->len;
			self->z.avail_out = size;
			goto again;

		default:
			THROW_EZLIB (self->z);
	}

	l = self->z.total_out - total;
	if (isbuf)
	{
		l += out->len;
		qu_buffer_set_len (out, l);
		return out;
	}
	else
	{
		l = self->z.total_out - total;
		qu_string_buffer_set_len (out, l);
		return qu_string_buffer_return (out);
	}
}



/*) sub finish (b = nil) -> Any
	Finish compressing.

	This flushes the internal buffer and outputs all remaining data.

	If v{b} is not nil then it must be a Buffer object to store the output
	In this case this method returns that Buffer on success.
	Otherwise this method returns a new String.

	@exc
	.EArgValue
	.ECall      #{finish} already called
	.EMemory
	.EZlib
*/
static qu_stringy Deflate_finish (Deflate self, int argc, qu_any *argv)
{
	qu_stringy    out = 0;
	int           isbuf = 0;
	long          size, l;
	unsigned long total = self->z.total_out;

	QU_REQUIREZ (QuECall, !qu_flag_test (self, DeflateFlagENDED), "already finished");

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
	{
		out = argv[0];
		if (qu_null_is (out))
			out = 0;
		else
		{
			QU_ARG_TYPE (out, qu_buffer);
			QU_BUFFER_ENSURE_NOT_FROZEN (out);
			isbuf = 1;
		}
	}

	/* start small - it will grow pretty fast if necessary anyway */
	size = 8;

	if (isbuf)
	{
		QU_ENSURE (qu_buffer_make_space ((qu_buffer) out, size));
		self->z.avail_out = qu_buffer_avail ((qu_buffer) out);
		self->z.next_out  = out->str + out->len;
	}
	else
	{
		out = qu_string_buffer_new (size);
		self->z.avail_out = size;
		self->z.next_out  = out->str;
	}

	self->z.avail_in = 0;

  again:

	assert (self->z.avail_out > 0);
	switch (deflate (&self->z, Z_FINISH))
	{
		case Z_STREAM_END:
			break;

		case Z_OK:
			assert (size < QU_MEM_MAX - size);
			size <<= 1;
			l = self->z.total_out - total;
			if (isbuf)
			{
				l += ((qu_buffer) out)->len;
				qu_buffer_set_len ((qu_buffer) out, l);
				if (!qu_buffer_make_space ((qu_buffer) out, size))
					QU_THROW (QuEMemory);
			}
			else
			{
				qu_string_buffer_set_len (out, l);
				if (!qu_string_buffer_make_space (out, size))
					QU_THROW (QuEMemory);
			}
			self->z.next_out = out->str + out->len;
			self->z.avail_out = size;
			goto again;

		default:
			THROW_EZLIB (self->z);
	}

	qu_flag_set (self, DeflateFlagENDED);
	if (deflateEnd (&self->z) != Z_OK)
		THROW_EZLIB (self->z);

	l = self->z.total_out - total;
	if (isbuf)
	{
		l += ((qu_buffer) out)->len;
		qu_buffer_set_len (out, l);
		return out;
	}
	else
	{
		l = self->z.total_out - total;
		qu_string_buffer_set_len (out, l);
		return qu_string_buffer_return (out);
	}
}



/*) sub total_in -> Integer
	Return the total number of bytes passed so far.

	@noexc
*/
static qu_integer Deflate_total_in (Deflate self)
{
	return qu_integer_newul (self->z.total_in);
}



/*) sub total_out -> Integer
	Return the total number of bytes produced so far.

	@noexc
*/
static qu_integer Deflate_total_out (Deflate self)
{
	return qu_integer_newul (self->z.total_out);
}



/*) sub adler -> Integer
	Return the Adler-32 value of the uncompressed data.

	@noexc
*/
static qu_integer Deflate_adler (Deflate self)
{
	return qu_integer_newul (self->z.adler);
}



/*) sub data_type -> Int
	Best guess about the data type: Z_BINARY, Z_ASCII, or Z_UNKNOWN

	@noexc
*/
static qu_int Deflate_data_type(Deflate self)
{
	return qu_int_newi (self->z.data_type);
}



/*============================================================================

	class Inflate

============================================================================*/

/*) class Inflate
	This class provides a convenient way to compress multiple data.

	You start by creating a Inflate object via #{new}, call #{decompress}
	for each data you need to decompress, and finally call #{finish} to
	end the compression process. Check also #{unused} for unused data
	in case you pass data beyond compressed data.

	@xmp
		use Zlib

		# first, let's produce some compressed data

		d = Zlib.Deflate ()
		b = Buffer (10)
		s = 'foobar'
		for i in 1, 5
			d.compress (s, b)
		;;
		d.finish (b)

		# now we decompress it
		s = b.get
		a = [s[0, 4], s[4, 8], s[12,]]
		i = Zlib.Inflate ()
		for s in a
			i.decompress (s, b)
		;;
		i.finish (b)
		println (b)
		println (i.unused.len)

	@out
		foobarfoobarfoobarfoobarfoobar
		0
*/
static qu_class InflateClass;

typedef struct TInflate TInflate;
struct TInflate
{
	QU_OBJECT_HEAD;      /* \__ must match */
	z_stream  z;         /* /    TDeflate  */
	qu_string unused;    /* unused data    */
};

typedef struct TInflate *Inflate;

#define InflateFlagENDED DeflateFlagENDED


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (wbits = DEF_WBITS) -> Inflate __final
	Create a new Inflate object.

	See the C{man 3 zlib} for details of the argument.

	@exc
	.EArgValue
	.EMemory
*/
static Inflate Inflate__inst (qu_class self, int argc, qu_any *argv)
{
	static char *argx[] =
	{
		":",
		"i", "wbits",
		NULL
	};

	int wbits = DEF_WBITS;
	Inflate o;

	QU_ARG_XPARSE (&wbits);

	o = qu_gc_new (self);
	o->unused = 0;
	o->z.avail_in = 0;
	o->z.zalloc = 0;
	o->z.zfree  = 0;

	switch (inflateInit2 (&o->z, wbits))
	{
		case Z_OK:
			break;

		case Z_MEM_ERROR:
			QU_THROW (QuEMemory);

		default:
			QU_ARG_THROW;
	}
    return o;
}



/*) sub class new (*arg) -> Inflate
	Create a new Inflate object.

	This default method simply calls __inst then __init with all the arguments
	if the class defines one.

	See #{__inst} for details.
*/
static Inflate Inflate_new (qu_class self, int argc, qu_any *argv)
{
	Inflate o = Inflate__inst (self, argc, argv);

	QU_ENSURE (o);
	return qu_class_init_new (self, o, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

#define Inflate__destroy Deflate__destroy



static void Inflate__mark (Inflate self)
{
	qu_gc_xmark (self->unused);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub decompress (s, b = nil, ofs = 0, len = s.len) -> Any
	Decompress the String or Buffer v{s} optionally start at offset and a specified
	length.

	If v{ofs} is negative then count from the end.

	If v{b} is not nil then it must be a Buffer object to store the output
	In this case this method returns that Buffer on success.
	Otherwise this method returns a new String.

	Once you have completed decompressing, you must call #{finish} to get the
	final result.

	@warn
	If v{b} is a Buffer then don't forget to consume it.

	@exc
	.EArgValue
	.ECall      #{finish} already called
	.EIndex
	.EMemory
	.EZlib
*/
static qu_stringy Inflate_decompress (Inflate self, int argc, qu_any *argv)
{
	qu_stringy    s, out = 0;
	int           isbuf = 0;
	long          ofs = 0;
	long          len, size, l;
	unsigned long total = self->z.total_out;

	QU_REQUIREZ (QuECall, !qu_flag_test (self, InflateFlagENDED), "already finished");

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, s, qu_stringy);
	len = s->len;
	if (argc > 1)
	{
		out = argv[1];
		if (qu_null_is (out))
			out = 0;
		else
		{
			QU_ARG_TYPE (out, qu_buffer);
			QU_BUFFER_ENSURE_NOT_FROZEN (out);
			isbuf = 1;
		}
		if (argc > 2)
		{
			long max;
			QU_ARG_vlong (2, ofs);
			QU_ENSURE_INDEX (ofs, len);
			if (argc > 3)
			{
				QU_ARG_vlong (3, len);
				QU_ARG_CHECK (len >= 0);
			}
			max = s->len - ofs;
			if (len > max)
				len = max;
		}
	}

	if (len == 0)
	{
		if (isbuf)
			return out;
		return QuStringEMPTY;
	}

	QU_MEM_CHECK (len < QU_MEM_MAX - len / 1000 - 12 - 1);
	size = len + len / 1000 + 12 + 1;

	if (isbuf)
	{
		QU_ENSURE (qu_buffer_make_space ((qu_buffer) out, size));
		self->z.avail_out = qu_buffer_avail ((qu_buffer) out);
		self->z.next_out  = out->str + out->len;
	}
	else
	{
		out = qu_string_buffer_new (size);
		self->z.avail_out = size;
		self->z.next_out  = out->str;
	}

	self->z.avail_in = len;
	self->z.next_in  = s->str + ofs;

  again:

	assert (self->z.avail_out > 0);
	switch (inflate (&self->z, Z_NO_FLUSH))
	{
		case Z_OK:
			if (self->z.avail_out)
				break;

			l = self->z.total_out - total;
			assert (size < QU_MEM_MAX - size);
			size <<= 1;
			if (isbuf)
			{
				l += ((qu_buffer) out)->len;
				qu_buffer_set_len (out, l);
				if (!qu_buffer_make_space ((qu_buffer) out, size))
					QU_THROW (QuEMemory);
			}
			else
			{
				qu_string_buffer_set_len (out, l);
				if (!qu_string_buffer_make_space (out, size))
					QU_THROW (QuEMemory);
			}
			self->z.next_out = out->str + out->len;
			self->z.avail_out = size;
			goto again;

		case Z_STREAM_END:
			if (self->z.avail_in)
				self->unused = qu_string_new ((char *) self->z.next_in, self->z.avail_in);
			break;

		default:
			THROW_EZLIB (self->z);
	}

	l = self->z.total_out - total;
	if (isbuf)
	{
		l += out->len;
		qu_buffer_set_len (out, l);
		return out;
	}
	else
	{
		l = self->z.total_out - total;
		qu_string_buffer_set_len (out, l);
		return qu_string_buffer_return (out);
	}
}



/*) sub finish (b = nil) -> Any
	Finish decompressing.

	This flushes the internal buffer and outputs all remaining data.

	If v{b} is not nil then it must be a Buffer object to store the output
	In this case this method returns that Buffer on success.
	Otherwise this method returns a new String.

	@exc
	.EArgValue
	.ECall      #{finish} already called
	.EMemory
	.EZlib
*/
static qu_stringy Inflate_finish (Inflate self, int argc, qu_any *argv)
{
	qu_stringy    out = 0;
	int           isbuf = 0;
	long          size, l;
	unsigned long total = self->z.total_out;

	QU_REQUIREZ (QuECall, !qu_flag_test (self, InflateFlagENDED), "already finished");

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
	{
		out = argv[0];
		if (qu_null_is (out))
			out = 0;
		else
		{
			QU_ARG_TYPE (out, qu_buffer);
			QU_BUFFER_ENSURE_NOT_FROZEN (out);
			isbuf = 1;
		}
	}

	/* start small - it will grow pretty fast if necessary anyway */
	size = 8;

	if (isbuf)
	{
		QU_ENSURE (qu_buffer_make_space ((qu_buffer) out, size));
		self->z.avail_out = qu_buffer_avail ((qu_buffer) out);
		self->z.next_out  = out->str + out->len;
	}
	else
	{
		out = qu_string_buffer_new (size);
		self->z.avail_out = size;
		self->z.next_out  = out->str;
	}

	self->z.avail_in = 0;

  again:

	assert (self->z.avail_out > 0);
	switch (inflate (&self->z, Z_FINISH))
	{
		case Z_STREAM_END:
			break;

		case Z_OK:
			assert (size < QU_MEM_MAX - size);
			size <<= 1;
			l = self->z.total_out - total;
			if (isbuf)
			{
				l += ((qu_buffer) out)->len;
				qu_buffer_set_len ((qu_buffer) out, l);
				if (!qu_buffer_make_space ((qu_buffer) out, size))
					QU_THROW (QuEMemory);
			}
			else
			{
				qu_string_buffer_set_len (out, l);
				if (!qu_string_buffer_make_space (out, size))
					QU_THROW (QuEMemory);
			}
			self->z.next_out = out->str + out->len;
			self->z.avail_out = size;
			goto again;

		default:
			THROW_EZLIB (self->z);
	}

	qu_flag_set (self, InflateFlagENDED);
	if (inflateEnd (&self->z) != Z_OK)
		THROW_EZLIB (self->z);

	l = self->z.total_out - total;
	if (isbuf)
	{
		l += ((qu_buffer) out)->len;
		qu_buffer_set_len (out, l);
		return out;
	}
	else
	{
		l = self->z.total_out - total;
		qu_string_buffer_set_len (out, l);
		return qu_string_buffer_return (out);
	}
}



/*) sub unused -> String
	Return the unused data, an empty String if none.

	@noexc
*/
static qu_string Inflate_unused (Inflate self)
{
	if (self->unused)
		return (self->unused);
	return QuStringEMPTY;
}



/*) sub total_in -> Integer
	Return the total number of bytes passed so far.

	@noexc
*/
#define Inflate_total_in Deflate_total_in



/*) sub total_out -> Integer
	Return the total number of bytes produced so far.

	@noexc
*/
#define Inflate_total_out Deflate_total_out



/*) sub adler -> Integer
	Return the Adler-32 value of the compressed data.

	@noexc
*/
#define Inflate_adler Deflate_adler



/*) sub data_type -> Int
	Best guess about the data type: Z_BINARY, Z_ASCII, or Z_UNKNOWN

	@noexc
*/
#define Inflate_data_type Deflate_data_type



/*----------------------------------------------------------------------------
	DL stuffs
----------------------------------------------------------------------------*/
static void *QuZlibHandle;

static void qu_zlib_atexit ()
{
	if (QuZlibHandle)
		qu_dl_close (QuZlibHandle);
}

static int qu_zlib_failed (const char *z)
{
	qu_zlib_atexit ();
	QU_THROWF (QuESystem, "cant initialize Zlib missing symbol %s", z);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/
int qu_Zlib_init (qu_module m)
{
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		zlib library
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuZlibHandle = qu_dl_open_ex ("zlib", QU_DL_GLOBAL | QU_DL_NOW);
	QU_ENSURE (QuZlibHandle);


#define f_(name) \
	name = qu_dl_sym (QuZlibHandle, #name); \
	if (name == 0) \
		return qu_zlib_failed ( #name);

	f_(deflate         )
	f_(deflateEnd      )
	f_(inflate         )
	f_(inflateEnd      )
	f_(adler32         )
	f_(crc32           )
	f_(deflateInit_    )
	f_(inflateInit_    )
	f_(deflateInit2_   )
	f_(inflateInit2_   )
	f_(inflateBackInit_)

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		constants
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_string (m, "ZLIB_VERSION"         , ZLIB_VERSION);
	qu_module_add_int    (m, "MAX_WBITS"            , MAX_WBITS);
	qu_module_add_int    (m, "DEF_MEM_LEVEL"        , DEF_MEM_LEVEL);
	qu_module_add_int    (m, "Z_DEFLATED"           , Z_DEFLATED);
	qu_module_add_int    (m, "Z_BEST_SPEED"         , Z_BEST_SPEED);
	qu_module_add_int    (m, "Z_BEST_COMPRESSION"   , Z_BEST_COMPRESSION);
	qu_module_add_int    (m, "Z_DEFAULT_COMPRESSION", Z_DEFAULT_COMPRESSION);
	qu_module_add_int    (m, "Z_FILTERED"           , Z_FILTERED);
	qu_module_add_int    (m, "Z_HUFFMAN_ONLY"       , Z_HUFFMAN_ONLY);
	qu_module_add_int    (m, "Z_DEFAULT_STRATEGY"   , Z_DEFAULT_STRATEGY);
	qu_module_add_int    (m, "Z_FINISH"             , Z_FINISH);
	qu_module_add_int    (m, "Z_NO_FLUSH"           , Z_NO_FLUSH);
	qu_module_add_int    (m, "Z_SYNC_FLUSH"         , Z_SYNC_FLUSH);
	qu_module_add_int    (m, "Z_FULL_FLUSH"         , Z_FULL_FLUSH);
	qu_module_add_int    (m, "Z_BINARY"             , Z_BINARY);
	qu_module_add_int    (m, "Z_ASCII"              , Z_ASCII);
	qu_module_add_int    (m, "Z_UNKNOWN"            , Z_UNKNOWN);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		functions
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (m, "decompress", Zlib_decompress, -1);
	qu_module_add_function (m, "compress"  , Zlib_compress  , -1);
	qu_module_add_function (m, "adler32"   , Zlib_adler32   , -1);
	qu_module_add_function (m, "crc32"     , Zlib_crc32     , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class EZlib
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	EZlibClass = qu_exception_class_new (m, QuERuntimeClass, "EZlib", "Zlib error");


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Deflate
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	DeflateClass = qu_class_newz (m, "Deflate", TDeflate, Deflate__destroy, 0);

	qu_class_method_class (DeflateClass, QU_NAME__inst, Deflate__inst    , -1);
	qu_class_method_class (DeflateClass, QU_NAME_new  , Deflate_new      , -1);
	qu_class_method_instz (DeflateClass, "compress"   , Deflate_compress , -1);
	qu_class_method_instz (DeflateClass, "finish"     , Deflate_finish   , -1);
	qu_class_method_instz (DeflateClass, "total_in"   , Deflate_total_in ,  0);
	qu_class_method_instz (DeflateClass, "total_out"  , Deflate_total_out,  0);
	qu_class_method_instz (DeflateClass, "adler"      , Deflate_adler    ,  0);
	qu_class_method_instz (DeflateClass, "data_type"  , Deflate_data_type,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		class Inflate
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	InflateClass = qu_class_newz (m, "Inflate", TInflate, Inflate__destroy, Inflate__mark);

	qu_class_method_class (InflateClass, QU_NAME__inst, Inflate__inst     , -1);
	qu_class_method_class (InflateClass, QU_NAME_new  , Inflate_new       , -1);
	qu_class_method_instz (InflateClass, "decompress" , Inflate_decompress, -1);
	qu_class_method_instz (InflateClass, "finish"     , Inflate_finish    , -1);
	qu_class_method_instz (InflateClass, "unused"     , Inflate_unused    ,  0);
	qu_class_method_instz (InflateClass, "total_in"   , Inflate_total_in  ,  0);
	qu_class_method_instz (InflateClass, "total_out"  , Inflate_total_out ,  0);
	qu_class_method_instz (InflateClass, "adler"      , Inflate_adler     ,  0);
	qu_class_method_instz (InflateClass, "data_type"  , Inflate_data_type ,  0);

	atexit (qu_zlib_atexit);
	return 1;
}
