/* html.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module html
	This module provide support for Qu's HTML syntax.

	@warn
		You must not declare this module directly.
		Use C{pragma html} instead.

	@xmp
		pragma html

		choices = [1 2 3]
		<FORM ACTION="foo.qu" METHOD="POST" ENCTYPE="multipart/form-data">
			'Select one'<BR>
			<SELECT NAME='choice'>
				for x in choices
					<OPTION VALUE=x>(.x)</OPTION>
				;;
			</SELECT>
			<INPUT TYPE="submit" VALUE="Send"> &nbsp; <INPUT TYPE="reset">
		</FORM>

	@p
	Tag parameters are functions or variables that work with the k{ensure}
	mechanism. This means that you must not declare them in your module
	otherwise you will get a compile error.

	They are all named with a leading uppercase letter to avoid naming
	conflict with existing symbols. Here is a complete list:

	@ul
	.ABBR
	.ACCEPT
	.ACCEPT_CHARSET
	.ACCESSKEY
	.ACTION
	.ALIGN
	.ALINK
	.ALT
	.ARCHIVE
	.AXIS
	.BACKGROUND
	.BGCOLOR
	.BORDER
	.CELLPADDING
	.CELLSPACING
	.CHAR
	.CHAROFF
	.CHARSET
	.CHECKED
	.CITE
	.CLASS
	.CLASSID
	.CLEAR
	.CODE
	.CODEBASE
	.CODETYPE
	.COLOR
	.COLS
	.COLSPAN
	.COMPACT
	.CONTENT
	.COORDS
	.DATA
	.DATETIME
	.DECLARE
	.DEFER
	.DIR
	.DISABLED
	.ENCTYPE
	.FACE
	.FOR
	.FRAME
	.FRAMEBORDER
	.HALIGN
	.HEADERS
	.HEIGHT
	.HREF
	.HREFLANG
	.HSPACE
	.HTTP_EQUIV
	.ID
	.ISMAP
	.LABEL
	.LANGUAGE
	.LINK
	.LONGDESC
	.MARGINHEIGHT
	.MARGINWIDTH
	.MAXLENGTH
	.MEDIA
	.METHOD
	.MULTIPLE
	.NAME
	.NOHREF
	.NOERSIZE
	.NOSHADE
	.NOWRAP
	.OBJECT
	.ONBLUR
	.ONCHANGE
	.ONCLICK
	.ONDBLCLICK
	.ONFOCUS
	.ONKEYDOWN
	.ONKEYPRESS
	.ONKEYUP
	.ONLOAD
	.ONMOUSEDOWN
	.ONMOUSEMOVE
	.ONMOUSEOUT
	.ONMOUSEOVER
	.ONMOUSEUP
	.ONRESET
	.ONSELECT
	.ONSUBMIT
	.ONUNLOAD
	.PROFILE
	.PROMPT
	.READONLY
	.REL
	.REV
	.ROWS
	.ROWSPAN
	.RULES
	.SCHEME
	.SCOPE
	.SCROLLING
	.SELECTED
	.SHAPE
	.SIZE
	.SPAN
	.SRC
	.STANDBY
	.START
	.STYLE
	.SUMMARY
	.TABINDEX
	.TARGET
	.TEXT
	.TITLE
	.TYPE
	.USEMAP
	.VALIGN
	.VALUE
	.VALUETYPE
	.VERSION
	.VLINK
	.VSPACE
	.WIDTH
*/

/* list of user defined tags and parameters */
static qu_dict UserTags;


/*----------------------------------------------------------------------------
	stuff
----------------------------------------------------------------------------*/

inline static qu_array _html_ensure_current ()
{
	qu_thread me = qu_thread_me ();

	if (me->html == 0)
		me->html = qu_array_new_ex (2, QuStringEMPTY, qu_dict_new (64));
	return me->html;
}


static void _html_set_param (qu_any value, const char *label)
{
	qu_string s      = qu_string_new_staticz (label);
	qu_array  data   = _html_ensure_current ();
	qu_string tag    = data->objs[0];
	qu_dict   tags   = data->objs[1];
	qu_dict   params = qu_dict_getitem (tags, tag);

	if (tag->len)
	{
		if (params == 0)
		{
			params = qu_dict_new (4);
			qu_dict_setitem (tags, tag, params);
		}
		qu_dict_setitem (params, s, value);
	}
}


static void _html_asis (qu_any x, const char *label)
{
	qu_sys_printz (label, strlen (label));
	qu_sys_print  (x);
	qu_sys_printz ("\"", 1);
}


static int _html_escape (qu_string s, const char *label)
{
#define SMALL 1024
	long  l, m, n = 0, len = strlen (label);
	char *k, *i, *j;
	char  buf [SMALL];
	int   small;

	if (!qu_stringy_is (s))
		s = qu_sys_string (s);

	i = s->str;
	l = s->len;
	while (l--)
	{
		QU_MEM_CHECK (n <= QU_MEM_MAX - 8);
		switch (*i++)
		{
			case '<':
			case '>': n += 4; continue;
			case '&': n += 5; continue;
			case '"': n += 6; continue;
			default : ++n;
		}
	}

	if (n == s->len)
	{
		if (!qu_string_is (s))
			s = qu_sys_string (s);
		qu_sys_printz (label, len);
		qu_sys_print  (s);
		qu_sys_printz ("\"", 1);
		return 1;
	}

	i = s->str;
	l = s->len;

	m = n + len + 1;
	small = m < SMALL;
	if (small)
		k = buf;
	else
		k = qu_mem_alloc (m);

	strcpy (k, label);
	j = k + len;

	while (l--)
	{
		switch (*i)
		{
			case '<': ++i; *j++ = '&'; *j++ = 'l'; *j++ = 't'; *j++ = ';'; continue;
			case '>': ++i; *j++ = '&'; *j++ = 'g'; *j++ = 't'; *j++ = ';'; continue;
			case '&': ++i; *j++ = '&'; *j++ = 'a'; *j++ = 'm'; *j++ = 'p'; *j++ = ';'; continue;
			case '"': ++i; *j++ = '&'; *j++ = 'q'; *j++ = 'u'; *j++ = 'o'; *j++ = 't'; *j++ = ';'; continue;
			default : *j++ = *i++;
		}
	}
	*j++ = '"';
	*j++ = '\0';

	qu_sys_printz (k, m);
	if (!small)
		qu_mem_free (k);
	return 1;
}



#define V(name, label) \
static qu_null html_##name (qu_any x) \
{ \
	_html_asis (x, " " label "=\""); \
	_html_set_param (x, label); \
	return QuNil; \
}

#define E(name, label) \
static qu_null html_##name (qu_any x) \
{ \
	_html_escape (x, " " label "=\""); \
	_html_set_param (x, label); \
	return QuNil; \
}

#define T(name, label) \
static qu_null html_##name () \
{ \
	qu_sys_printz (" " label, strlen (" " label)); \
	return QuNil; \
}

V (ABBR           , "abbr")
V (ACCEPT         , "accept")
V (ACCEPT_CHARSET , "accept-charset")
V (ACCESSKEY      , "accesskey")
V (ACTION         , "action")
V (ALIGN          , "align")
V (ALINK          , "alink")
V (ALT            , "alt")
V (ARCHIVE        , "archive")
V (AXIS           , "axis")
V (BACKGROUND     , "background")
V (BGCOLOR        , "bgcolor")
V (BORDER         , "border")
V (CELLPADDING    , "cellpadding")
V (CELLSPACING    , "cellspacing")
V (CHAR           , "char")
V (CHAROFF        , "charoff")
V (CHARSET        , "charset")
T (CHECKED        , "checked")
V (CITE           , "cite")
V (CLASS          , "class")
V (CLASSID        , "classid")
V (CLEAR          , "clear")
V (CODE           , "code")
V (CODEBASE       , "codebase")
V (CODETYPE       , "codetype")
V (COLOR          , "color")
V (COLS           , "cols")
V (COLSPAN        , "colspan")
T (COMPACT        , "compact")
E (CONTENT        , "content")
V (COORDS         , "coords")
V (DATA           , "data")
V (DATETIME       , "datetime")
T (DECLARE        , "declare")
T (DEFER          , "defer")
V (DIR            , "dir")
T (DISABLED       , "disable")
V (ENCTYPE        , "enctype")
V (FACE           , "face")
V (FOR            , "for")
V (FRAME          , "frame")
V (FRAMEBORDER    , "frameborder")
V (HALIGN         , "halign")
V (HEADERS        , "headers")
V (HEIGHT         , "height")
V (HREF           , "href")
V (HREFLANG       , "hreflang")
V (HSPACE         , "hspace")
V (HTTP_EQUIV     , "http-equiv")
V (ID             , "id")
T (ISMAP          , "ismap")
V (LABEL          , "label")
V (LANGUAGE       , "language")
V (LINK           , "link")
V (LONGDESC       , "longdesc")
V (MARGINHEIGHT   , "marginheight")
V (MARGINWIDTH    , "marginwidth")
V (MAXLENGTH      , "maxlength")
V (MEDIA          , "media")
V (METHOD         , "method")
T (MULTIPLE       , "multiple")
V (NAME           , "name")
T (NOHREF         , "nohref")
T (NOERSIZE       , "noresize")
T (NOSHADE        , "noshade")
T (NOWRAP         , "nowrap")
V (OBJECT         , "object")
V (ONBLUR         , "onblur")
V (ONCHANGE       , "onchange")
V (ONCLICK        , "onclick")
V (ONDBLCLICK     , "ondblclick")
V (ONFOCUS        , "onfocus")
V (ONKEYDOWN      , "onkeydown")
V (ONKEYPRESS     , "onkeypress")
V (ONKEYUP        , "onkeyup")
V (ONLOAD         , "onload")
V (ONMOUSEDOWN    , "onmousedown")
V (ONMOUSEMOVE    , "onmousemove")
V (ONMOUSEOUT     , "onmouseout")
V (ONMOUSEOVER    , "onmouseover")
V (ONMOUSEUP      , "onmouseup")
V (ONRESET        , "onreset")
V (ONSELECT       , "onselect")
V (ONSUBMIT       , "onsubmit")
V (ONUNLOAD       , "onunload")
V (PROFILE        , "profile")
V (PROMPT         , "prompt")
T (READONLY       , "readonly")
V (REL            , "rel")
V (REV            , "rev")
V (ROWS           , "rows")
V (ROWSPAN        , "rowspan")
V (RULES          , "rules")
V (SCHEME         , "scheme")
V (SCOPE          , "scope")
V (SCROLLING      , "scrolling")
T (SELECTED       , "selected")
V (SHAPE          , "shape")
V (SIZE           , "size")
V (SPAN           , "span")
V (SRC            , "src")
V (STANDBY        , "standby")
V (START          , "start")
V (STYLE          , "style")
V (SUMMARY        , "summary")
V (TABINDEX       , "tabindex")
V (TARGET         , "target")
V (TEXT           , "text")
E (TITLE          , "title")
V (TYPE           , "type")
V (USEMAP         , "usemap")
V (VALIGN         , "valign")
E (VALUE          , "value")
V (VALUETYPE      , "valuetype")
V (VERSION        , "version")
V (VLINK          , "vlink")
V (VSPACE         , "vspace")
V (WIDTH          , "width")

#undef V
#undef E
#undef T


static qu_null html__open (qu_string tag, long op)
{
	int      closed = op & 0x1;
	int      linebreak = op & 0x10;
	qu_dict  all, params;
	qu_array handler, data;

	assert (qu_string_is (tag));
	data = _html_ensure_current ();
	data->objs[0] = tag;

	all = data->objs[1];
	params = qu_dict_getitem (all, tag);
	if (params)
	{
		assert (qu_dict_is (params));
		qu_dict_clear (params);
	}

	handler = qu_dict_getitem (UserTags, tag);
	if (handler)
	{
		qu_function f;

		assert (qu_array_is (handler));
		f = handler->objs[0];
		if (qu_method_is (f))
		{
			qu_sub u = ((qu_method) f)->sub;
			if (u->narg == 0)
				QU_ENSURE (qu_method_call0 ((qu_method) f));
			else
				QU_ENSURE (qu_method_call1 ((qu_method) f, closed ? QuTrue : QuFalse));
		}
		else if (qu_sub_is (f))
		{
			qu_any theself = qu_interp_self ();
			if (((qu_sub) f)->narg == 0)
				QU_ENSURE (qu_sub_call0 ((qu_sub) f, theself));
			else
				QU_ENSURE (qu_sub_call1 ((qu_sub) f, theself, closed ? QuTrue : QuFalse));
		}
		else
		{
			goto none;
		}
	}
	else
	{
	  none:
		qu_sys_printz ("<", 1);
		qu_sys_printz (tag->str, tag->len);
	}
	if (closed)
	{
		if (linebreak)
			qu_sys_printz (">\n", 2);
		else
			qu_sys_printz (">", 1);
	}
	return QuNil;
}


static qu_null html_open (int argc, qu_any *argv)
{
	qu_string tag;
	int       closed = 0;
	int       linebreak = 0;
	long      op = 0;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE  (0, tag, qu_string);
	if (argc > 1)
	{
		QU_ARG_vbool (1, closed);
		if (argc > 2)
			QU_ARG_vbool (2, linebreak);
	}
	if (closed)
		op = 1;
	if (linebreak)
		op |= 0x10;
	return html__open (tag, op);
}



static qu_null html__close (qu_string tag, int linebreak)
{
	qu_array handler;

	if (tag->len == 1 && tag->str[0] == '/')
	{
		qu_array data = _html_ensure_current ();
		tag = data->objs[0];

		handler = qu_dict_getitem (UserTags, tag);
		if (handler)
		{
			qu_any f;

			assert (qu_array_is (handler));
			f = handler->objs[1];
			if (qu_method_is (f))
			{
				QU_ENSURE (qu_method_call0 ((qu_method) f));
			}
			else if (qu_sub_is (f))
			{
				qu_any theself = qu_interp_self ();
				QU_ENSURE (qu_sub_call0 ((qu_sub) f, theself));
			}
			else
			{
				goto none1;
			}
		}
		else
		{
		  none1:
			qu_sys_printz ("/>", 2);
		}
	}
	else if (tag->len)
	{
		handler = qu_dict_getitem (UserTags, tag);
		if (handler)
		{
			qu_any f;

			assert (qu_array_is (handler));
			f = handler->objs[1];
			if (qu_method_is (f))
			{
				QU_ENSURE (qu_method_call0 ((qu_method) f));
			}
			else if (qu_sub_is (f))
			{
				qu_any theself = qu_interp_self ();
				QU_ENSURE (qu_sub_call0 ((qu_sub) f, theself));
			}
			else
			{
				goto none2;
			}
		}
		else
		{
		  none2:
			qu_sys_printz ("</", 2);
			qu_sys_printz (tag->str, tag->len);
			qu_sys_printz (">", 1);
		}
	}
	else
	{
		qu_sys_printz (">", 1);
	}
	if (linebreak)
		qu_sys_printz ("\n", 1);
	return QuNil;
}



static qu_null html_close (int argc, qu_any *argv)
{
	qu_string tag;
	int       linebreak = 0;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, tag, qu_string);
	if (argc > 1)
		QU_ARG_vbool (1, linebreak);
	return html__close (tag, linebreak);
}



static qu_null html_entity (qu_string x)
{
#define MAXENTITYLEN 32
	char b [MAXENTITYLEN + 4];
	long l;

	if (!qu_stringy_is (x))
		x = qu_sys_string (x);
	l = x->len;
	if (l > MAXENTITYLEN)
		l = MAXENTITYLEN;
	b[0] = '&';
	strncpy (b + 1, x->str, l);
	b[l + 1] = ';';
	qu_sys_printz (b, l + 2);
	return QuNil;
}



/*) sub tag (name: String, opener: Function, closer: Function) -> Null
	Register a new tag.

	v{name} is the name of the tag, case sensitive.

	v{opener} is the Function that will be called when C{<foo} is encountered.
	This Function must print the tag with the opening C{'<'} but not the closing
	C{'>'}. It must take at most one argument.
	@ul
	.v{closed} whether this is C{<foo>} or C{<foo}

	v{closer} is the Function that will be called when C{</foo>} is encountered.
	This Function must print the tag with the opening C{'</'} and the closing
	C{'>'}. It must take no argument.

	Both v{opener} and v{closer} might be nil if none is necessary.
	But one must not be nil.

	@exc
	.EArgValue
	.EKey      name already exists

	@xmp
		pragma html

		html.tag ("foo", @foo, @_foo)

		sub foo (closed)
			print ('<bar thing="1"')
		;;

		sub _foo
			print ('</bar>')
		;;

		<foo Value=2>
			'hello'
		</foo>

	@out
		<bar thing="1" value="2">
		hello
		</bar>
*/
static qu_null html_tag (qu_string name, qu_function opener, qu_any closer)
{
	QU_ARG_CHECK (qu_string_is (name)
		&& (qu_function_is (opener) || opener == (qu_any) QuNil)
		&& (qu_function_is (closer) || closer == (qu_any) QuNil)
		&& !(opener == (qu_any) QuNil && closer == (qu_any) QuNil));

	QU_REQUIREZ (QuEKey, qu_dict_getitem (UserTags, name) == 0, "tag already registered");

	qu_dict_unfreeze (UserTags);
	qu_dict_setitem  (UserTags, name, qu_array_new_ex (2, opener, closer));
	qu_dict_freeze   (UserTags);

	return QuNil;
}



/*) sub ctag -> String
	Return the current tag printed by this thread.

	@noexc

	@xmp
		pragma html

		<foo Value=2>
			println (html.ctag)
		</foo>

	@out
		<foo value="2">
		foo
		</foo>
*/
static qu_string html_ctag ()
{
	qu_array data = _html_ensure_current ();
	return data->objs[0];
}



/*) sub cparam -> Dict
	Return the tags and parameters printed by this thread so far.

	Note that you can alter the Dict but generally that is not a good idea.

	@noexc

	@xmp
		pragma html

		<FOO VALUE=2>
			println (html.cparam)
		</FOO>

	@out
		<FOO VALUE="2">
		['FOO':['VALUE': 2]]
		</FOO>
*/
static qu_dict html_cparam ()
{
	qu_array data = _html_ensure_current ();
	return data->objs[1];
}



/*) sub clear -> Null
	Clear all recorded tags and parameters values printed by this thread.

	If you call #{cparam} after this then you get an empty Dict.

	@noexc
*/
static qu_null html_clear ()
{
	qu_array data = _html_ensure_current ();
	qu_dict  param = data->objs[1];
	qu_dict_clear (param);
	return QuNil;
}



/*) sub param (name: String, value: Any) -> Null
	Set parameter value for the current tag.

	@exc
	.EArgValue
*/
static qu_null html_param (qu_string name, qu_any value)
{
	QU_ARG_TYPE (name, qu_string);
	_html_set_param (value, name->str);
	return QuNil;
}


int qu_html_init (qu_module m)
{

#define V(name, label) qu_module_add_function (m, #name, html_##name,  1);
#define E(name, label) qu_module_add_function (m, #name, html_##name,  1);
#define T(name, label) qu_module_add_function (m, #name, html_##name,  0);

	UserTags = qu_dict_new (8);
	qu_dict_freeze (UserTags);
	qu_module_add_const (m, "__tags__", UserTags);

	__QuModuleHtml_open   = html__open;
	__QuModuleHtml_close  = html__close;
	__QuModuleHtml_entity = html_entity;

	qu_module_add_function (m, "open"  , html_open  , -1);
	qu_module_add_function (m, "close" , html_close , -1);
	qu_module_add_function (m, "entity", html_entity,  1);
	qu_module_add_function (m, "tag"   , html_tag   ,  3);
	qu_module_add_function (m, "ctag"  , html_ctag  ,  0);
	qu_module_add_function (m, "cparam", html_cparam,  0);
	qu_module_add_function (m, "clear" , html_clear ,  0);
	qu_module_add_function (m, "param" , html_param ,  2);


V (ABBR           , "abbr")
V (ACCEPT         , "accept")
V (ACCEPT_CHARSET , "accept-charset")
V (ACCESSKEY      , "accesskey")
V (ACTION         , "action")
V (ALIGN          , "align")
V (ALINK          , "alink")
V (ALT            , "alt")
V (ARCHIVE        , "archive")
V (AXIS           , "axis")
V (BACKGROUND     , "background")
V (BGCOLOR        , "bgcolor")
V (BORDER         , "border")
V (CELLPADDING    , "cellpadding")
V (CELLSPACING    , "cellspacing")
V (CHAR           , "char")
V (CHAROFF        , "charoff")
V (CHARSET        , "charset")
T (CHECKED        , "checked")
V (CITE           , "cite")
V (CLASS          , "class")
V (CLASSID        , "classid")
V (CLEAR          , "clear")
V (CODE           , "code")
V (CODEBASE       , "codebase")
V (CODETYPE       , "codetype")
V (COLOR          , "color")
V (COLS           , "cols")
V (COLSPAN        , "colspan")
T (COMPACT        , "compact")
E (CONTENT        , "content")
V (COORDS         , "coords")
V (DATA           , "data")
V (DATETIME       , "datetime")
T (DECLARE        , "declare")
T (DEFER          , "defer")
V (DIR            , "dir")
T (DISABLED       , "disable")
V (ENCTYPE        , "enctype")
V (FACE           , "face")
V (FOR            , "for")
V (FRAME          , "frame")
V (FRAMEBORDER    , "frameborder")
V (HALIGN         , "halign")
V (HEADERS        , "headers")
V (HEIGHT         , "height")
V (HREF           , "href")
V (HREFLANG       , "hreflang")
V (HSPACE         , "hspace")
V (HTTP_EQUIV     , "http-equiv")
V (ID             , "id")
T (ISMAP          , "ismap")
V (LABEL          , "label")
V (LANGUAGE       , "language")
V (LINK           , "link")
V (LONGDESC       , "longdesc")
V (MARGINHEIGHT   , "marginheight")
V (MARGINWIDTH    , "marginwidth")
V (MAXLENGTH      , "maxlength")
V (MEDIA          , "media")
V (METHOD         , "method")
T (MULTIPLE       , "multiple")
V (NAME           , "name")
T (NOHREF         , "nohref")
T (NOERSIZE       , "noresize")
T (NOSHADE        , "noshade")
T (NOWRAP         , "nowrap")
V (OBJECT         , "object")
V (ONBLUR         , "onblur")
V (ONCHANGE       , "onchange")
V (ONCLICK        , "onclick")
V (ONDBLCLICK     , "ondblclick")
V (ONFOCUS        , "onfocus")
V (ONKEYDOWN      , "onkeydown")
V (ONKEYPRESS     , "onkeypress")
V (ONKEYUP        , "onkeyup")
V (ONLOAD         , "onload")
V (ONMOUSEDOWN    , "onmousedown")
V (ONMOUSEMOVE    , "onmousemove")
V (ONMOUSEOUT     , "onmouseout")
V (ONMOUSEOVER    , "onmouseover")
V (ONMOUSEUP      , "onmouseup")
V (ONRESET        , "onreset")
V (ONSELECT       , "onselect")
V (ONSUBMIT       , "onsubmit")
V (ONUNLOAD       , "onunload")
V (PROFILE        , "profile")
V (PROMPT         , "prompt")
T (READONLY       , "readonly")
V (REL            , "rel")
V (REV            , "rev")
V (ROWS           , "rows")
V (ROWSPAN        , "rowspan")
V (RULES          , "rules")
V (SCHEME         , "scheme")
V (SCOPE          , "scope")
V (SCROLLING      , "scrolling")
T (SELECTED       , "selected")
V (SHAPE          , "shape")
V (SIZE           , "size")
V (SPAN           , "span")
V (SRC            , "src")
V (STANDBY        , "standby")
V (START          , "start")
V (STYLE          , "style")
V (SUMMARY        , "summary")
V (TABINDEX       , "tabindex")
V (TARGET         , "target")
V (TEXT           , "text")
E (TITLE          , "title")
V (TYPE           , "type")
V (USEMAP         , "usemap")
V (VALIGN         , "valign")
E (VALUE          , "value")
V (VALUETYPE      , "valuetype")
V (VERSION        , "version")
V (VLINK          , "vlink")
V (VSPACE         , "vspace")
V (WIDTH          , "width")

#undef V
#undef E
#undef T

	return 1;
}
