/* Array.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Array is Container
	An Array object is a dynamic sequential object container which automatically
	grows and shrinks as necessary. But see the the #{fix} and #{freeze} methods.

	To create an Array you use the language syntax, the #{new}, #{dim} or #{make}
	methods.

	The index count starts from 0. Thus C{x [0]} is the first item and
	C{x [x.__len - 1]} is the last, or more conveniently C{x [-1]}.

	To ensure that all items are of a specified type, you use the k{of} keyword
	(which calls the #{__of} method).

	@xmp
		a = [1, 2, 3]
		a = Array (1, 2, 3)
		a = Array.dim (10)
		a = Array ('123456')
		a = '12.34.56'.split ('.')
		a = [] of Int              # same as a = Array().__of (Int)

		struct Person
			Int    age
			String name
		;;

		a = [] of Person

	@p
	To get a single item or items based on a -{Range} you use the C{[]} syntax
	which calls the #{__getitem} method. To get a slice you use the C{[,]}
	syntax which calls the #{__getslice} method.

	You get the number of items in an Array by calling the #{__len} method, or
	the more convenient alias, #{len}.

	@xmp
		a = [0, 1, 2, 3, 4]
		println (a.len)
		println (a[1] ' ' a[-2])
		println (a[2, 3])
		println (a[10 .. -3 << 2]) # 10, 8, 6, 4, 2, 0, -2

	@out
		5
		1 3
		[2, 3, 4]
		[4, 2, 0]

	@p
	You can fix the size of an Array with the #{fix} method, or you can make
	the Array immutable with the #{freeze} method. Be careful though.
	Once fixed, there is no way to unfix it. Once frozen there is no way to
	unfreeze it.

	You can also temporarily protect an Array from being modified. See
	-{Container.protect}.

	I{Immutable} here means that the Array is guaranteed to contain the same
	objects at any time. Those objects, however, do not necessarily have to
	be immutable. They can be container objects with variable items.

	Attempt to modify a #{frozen} Array or change the size of a #{fixed} array
	will throw EModify.

	You create an Array literal using the k{const} syntax. Literals are created at
	compile time, once only. Thus in the following example the literal is simply
	assigned to the variable.

	It is up to the compiler to define what constitutes a constant. Generally it
	must be a -{String}, -{Number}, references, nil, true, false or an expression
	that yields a literal.

	@xmp
		sub foo
			println ('foo')
		;;

		a = const [1, 2 * 3, @foo, 'bar', nil, true, false]

	@p
	You can use a literal Array as a I{validator}.

	@xmp
		#  +---- must ---+
		#  |             |
		const Fooish = const ['foo', 'bar', 'baz']
		var x: Fooish = 'bar'

		x = 'foo'
		println (x)

		y = 'blah'
		try
			# x = 'blah' #-> compile error
			x = y        #-> runtime error
		catch
			println (__err__)
		;;

	@out
		foo
		(EValue) invalid type or value
*/
qu_class QuArrayClass;


/* predefined frozen empty array */
qu_array QuArrayEmpty;

/* stupid forwards */
static qu_array   qu__array_select  (qu_array self, int argc, qu_any *argv);
static qu_string  qu__array_join    (qu_array self, int argc, qu_any *argv);
static qu_array   qu__array_push    (qu_array self, int argc, qu_any *argv);


/*----------------------------------------------------------------------------
	macros
----------------------------------------------------------------------------*/

/*
	Ensure the value x conforms with self->type.

	Note that qu_object_validate may go to program level but since we only allow to
	set the type once (once set it can not be modified) then there is no need
	to freeze the qu_array.

	Note that a validator may change the value thus x must be a variable.
*/
#define ENSURE_VALUE(self, x) \
QU_BEGIN \
	if ((self)->type) \
	{ \
		x = qu_object_validate (self, self->type, x); \
		QU_ENSUREZ (x, "invalid item for typed Array"); \
	} \
QU_END



/*----------------------------------------------------------------------------
	sizing
----------------------------------------------------------------------------*/

/* dont shrink below this limit */
#define SMALL_SIZE 16


/* perhaps shrink? */
#define MAYBE_SHRINK(self) \
QU_BEGIN \
	if ((self)->cap - (self)->len > SMALL_SIZE) \
		_array_resize (self, (self)->len); \
QU_END


/*
	Shrink or grow.
	!!! if growing, does not initialize the new items
*/
static void _array_resize (qu_array self, long cap)
{
	long c = self->cap;

	assert (cap >= 0);

	if (cap > c || c > SMALL_SIZE)
	{
		cap = qu_container_sizer (cap);
		if (cap != c)
		{
			self->objs = qu_mem_realloc (self->objs, cap * sizeof (qu_any));
			self->cap = cap;
		}
	}
}



/*----------------------------------------------------------------------------
	create
----------------------------------------------------------------------------*/

/*
	Initialize a new Array with a specified capacity.
	!!! make sure 0 <= cap && cap <= QU_GC_MAX_ITEMS before calling
*/
static qu_array _array_init (qu_array self, long cap)
{
	assert (0 <= cap && cap <= QU_GC_MAX_ITEMS);

	cap = qu_container_sizer (cap);

	self->len  = 0;
	self->cap  = cap;
	self->objs = qu_mem_alloc (cap * sizeof (qu_any));
	self->type = 0;

	QU_CONTAINER_INIT (self);
	return self;
}



/*
	Create a regular GC Array.
*/
qu_array qu_array_new (long len)
{
	return _array_init (qu_gc_new (QuArrayClass), len);
}



/*
	Create a root GC Array.
*/
qu_array qu_array_new_root (long len)
{
	return _array_init (qu_gc_new_root (QuArrayClass), len);
}



/*
	Create a static non-GC Array.
*/
qu_array qu_array_new_static (long len)
{
	return _array_init (qu_gc_new_static (QuArrayClass), len);
}



/*
	Create a regular non-GC Array with object arguments.
*/
qu_array qu_array_new_argv (int argc, qu_any *argv)
{
	qu_array  a = qu_array_new (argc);
	qu_any   *p = a->objs;

	a->len = argc;
	while (argc--)
		*p++ = *argv++;
	return a;
}



/*
	Create a regular GC Array by specifying the items.
*/
qu_array qu_array_new_ex (long len, ...)
{
	qu_array a = qu_array_new (len);
	long     i;
	va_list  va;

	va_start (va, len);
	for (i = 0; i < len; ++i)
	{
		qu_any x = va_arg (va, qu_any);
		a->objs [i] = x;
	}
	va_end (va);
	a->len = len;
	return a;
}



/*
	Create a static Array from an existing one.
*/
qu_array qu_array_to_static (qu_array self)
{
	qu_array a;
	long     l = self->len, i;
	qu_any  *p = self->objs, *q, x;

	if (qu_gc_is_static (self))
		return self;

	for (i = 0; i < l; ++i)
	{
		x = *p++;
		if (!qu_gc_is_static (x))
			return 0;
	}
	p = self->objs;
	a = qu_array_new_static (l);
	q = a->objs;
	for (i = 0; i < l; ++i)
		*q++ = qu_object_to_static (*p++);
	a->len = l;
	qu_array_freeze (a);
	return a;
}



/*----------------------------------------------------------------------------
	tools
----------------------------------------------------------------------------*/

/*
	Return the index of x optionally starting from offset i going forward.
	Return -1 if not found.

	!!! this might go to program level
*/
long qu_array_index (qu_array self, qu_any x, long i)
{
	for (; i < self->len; ++i)
	{
		if (qu_sys_eq (self->objs[i], x) == QuTrue)
			return i;
	}
	return -1;
}



/*
	Return the index of x optionally starting from offset i going backward.
	Return -1 if not found.

	!!! this might go to program level
*/
long qu_array_rindex (qu_array self, qu_any x, long i)
{
	if (i >= self->len)
		i = self->len - 1;

	for (; 0 <= i; --i)
	{
		/* this might go to program level */
		if (qu_sys_eq (self->objs[i], x) == QuTrue)
			return i;
	}
	return -1;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (*arg) -> Array
	Create a new Array with v{arg} as the items.

	If called without any arguments then a new empty Array is returned.

	@noexc
*/
static qu_array qu__array__inst (qu_class self, int argc, qu_any *argv)
{
	qu_array a = qu_array_new_argv (argc, argv);

	assert (self->size >= QuArrayClass->size);
	a->klass = self;
	return a;
}



/*) sub class new (*arg) -> Array
	Create a new Array with v{arg} as the items.

	This method calls #{__inst} then f{__init} if one is defined.

	@noexc

	@xmp
		x = Array (1, 2, 3)   # same as x = [1, 2, 3]
		x = Array ()          # same as x = []
*/
static qu_array qu__array_new (qu_class self, int argc, qu_any *argv)
{
	qu_array a = qu_array_new_argv (argc, argv);

	assert (self->size >= QuArrayClass->size);
	a->klass = self;
	return qu_class_init_new (self, a, argc, argv);
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class dim (size: Integer, value = nil, typed = false) -> Array
	Create a new Array with a specified size and initialize all items with
	v{value}, optionally set the #{type}.

	v{size} must be a positive integer.

	Note if v{typed} is true then v{val} may not be nil because the
	Array class does not allow setting the type to the Null class.

	@exc
	.EArgValue
	.EMemory

	@xmp
		a = Array.dim (4, 'x', true).fix
		println (a)
		println (a.type)
		println (a.fixed)

	@out
		['x', 'x', 'x', 'x']
		<Class sys.String>
		true

	@p
	If v{value} is a -{Record} then the Array will be typed with the record's
	-{Struct}, not the Record class. The logic is simple. It is meaningless to
	create an Array that may only contain the Record -{Class} object as the value.

	@xmp
		struct Person
			age  : Int
			name : String
		;;

		x = Person (0, '')
		a = Array.dim (4, x, true)
		println (a.type is Person)
		x = Person (99, 'Foo')
		a <<< x
		println (a)

		try
			a <<< 123
		catch
			println (__err__)
		;;

	@out
		true
		[[0, ''], [0, ''], [0, ''], [0, ''], [99, 'Foo']]
		(EValue) invalid type or value

	@note
	If you want a more flexible type such as using a I{validator} function then
	you have to create and fill the array first then use the k{of} keyword
	(the __of method). It is not possible to do that with the f{dim} method.
*/
static qu_array qu__array_dim (qu_class self, int argc, qu_any *argv)
{
	long     size;
	qu_any   val   = QuNil;
	int      typed = 0;
	qu_any   type  = 0;
	qu_array a;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_vlong (0, size);
	QU_ARG_CHECK  (size >= 0);
	QU_MEM_CHECK  (size <= QU_GC_MAX_ITEMS);

	if (argc > 1)
	{
		val = argv[1];
		if (argc > 2)
			QU_ARG_vbool (2, typed);
	}

	if (typed)
	{
		QU_ARG_CHECK (val != (qu_any) QuNil);
		if (qu_record_is (val))
			type = ((qu_record) val)->struc;
		else
			type = qu_classof (val);
		QU_ARG_CHECK (qu_object_valid_type (type));
	}

	a = qu_array_new (size);
	a->type = type;
	QU_OBJECT_LIST_SETVAL (a->objs, val, size);
	a->len = size;
	return a;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__array__destroy (qu_array self)
{
	qu_mem_free (self->objs);
}



static void qu__array__mark (qu_array self)
{
	if (self->type != QuIntClass)
	{
		qu_any x, *p = self->objs;
		long   n = self->len;

		while (n--)
		{
			x = *p++;
			qu_gc_mark (x);
		}
		qu_gc_xmark (self->type);
	}
	QU_CONTAINER_MARK (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be a one of the following -{Class} objects:
	@ops
	.| -{Dict} |
		Create a new Dict by assuming key/value pairs.
		If C{self.len} is an odd number then the last key will have
		a nil value. This is a non-atomic operation.

	.| -{Set} |
		Create a new Set by inserting items in self.
		This is a non-atomic operation.

	.| -{String}  | Delegate to #{__string}.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{Array}   | Return self.

	@exc
	.EArgValue

	@xmp
		x = ['foo', 1, 'bar', 2, 'baz']
		y = x as Dict
		println (y)

	@out
		['foo': 1, 'bar': 2, 'baz': nil]

	@p
	Non-atomic means if another thread or the f{__eq} or f{__hash} methods of
	the items in this Array modify this Array while this method is in progress
	then the result is unpredictable. That is, the new Set or Dict is not
	guaranteed to contain all items in this Array.

	Here is an example that demonstrates how this is possible even if your
	program is single threaded. Bad coding, example but possible nonetheless.

	@xmp
		ary = nil # make it known to Foo.__eq

		class Foo
			var bar

			sub __init (x)
				self.bar = x
			;;

			sub __hash
				println (__here__)
				return self.bar
			;;

			sub __eq (x)
				println (__here__)
				ary.pop            #! bad idea
				return x is Foo and x.bar == self.bar
			;;
		;;

		ary = [Foo (1), Foo (2), 'foo', 'bar']
		s = ary as Set
		println (s)

	@out
		__hash
		__hash
		__eq
		{<a Foo>, <a Foo>, 'foo'}
*/
qu_set qu_array_as_set (qu_array self)
{
	qu_set s = qu_set_new (8);
	long   i;

	for (i = 0; i < self->len; ++i)
	{
		qu_set_push (s, self->objs[i]);
	}
	return s;
}



qu_dict qu_array_as_dict (qu_array self)
{
	qu_dict d = qu_dict_new (8);
	long    i;

	for (i = 0; i < self->len; ++i)
	{
		qu_any k = self->objs[i++];
		qu_any v = (i < self->len) ? self->objs[i] : QuNil;
		qu_dict_setitem (d, k, v);
	}
	return d;
}



static qu_any qu__array__as (qu_array self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Set:     return qu_array_as_set     (self);
		case QU_CLASS_ID_Dict:    return qu_array_as_dict    (self);
		case QU_CLASS_ID_String:  return qu_array_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_array_as_boolean (self);
		case QU_CLASS_ID_Array:   return self;
	}
	QU_ARG_THROW;
}



/*) sub __clear -> self
	Remove all items.

	This is an atomic operation.
	After a call to this method, this Array is guaranteed to be empty.

	@exc
	.EModify

	@xmp
		a = [1, 2, 3]
		a.__clear
		println (a)

	@out
		[]
*/
/*) sub clear -> self
	This is an alias of #{__clear}.
*/
qu_array qu__array__clear (qu_array self)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	self->len = 0;
	_array_resize (self, 0);
	return self;
}



/*) sub __copy -> Array
	Create a shallow copy.

	Note that the #{type}, #{fixed}, and #{frozen} states are not copied
	to the new Array. You have to do it yourself as necessary.

	This is an atomic operation.
	The new Array is guaranteed to contain all items in this Array.

	@noexc

	@xmp
		x = [1 2 3]
		y = x.__copy
		println (y)
		println (x == y)
		println (x === y)

	@out
		[1, 2, 3]
		false
		true
*/
qu_array qu__array__copy (qu_array self)
{
	long     l = self->len;
	qu_array a = qu_array_new (l);

	QU_OBJECT_LIST_COPY (a->objs, self->objs, l);
	a->len = l;
	return a;
}



/*) sub __cmp (x) -> Int
	If v{x} is an Array then compare the number of items.
	Otherwise delegate to -{Object.__cmp}.

	@noexc

	@xmp
		x = [1, 2, 3]
		y = [3, 4, 5, 6]
		z = [7, 8, 9]

		println (x <=> y)
		println (x <=> z)

	@out
		-1
		0
*/
static qu_int qu__array__cmp (qu_array self, qu_any x)
{
	if (qu_array_is (x))
	{
		long l = self->len;
		long n = ((qu_array) x)->len;

		if (l > n) return QuInt1;
		if (l < n) return QuInt_1;
		return QuInt0;
	}
	return qu_object_default_cmp (self, x);
}



/*) sub __del (x) -> self
	Remove all items equal v{x} from this Array.

	This is an atomic operation unless the f{__eq} method of the items or another
	thread modifies this Array while this method is in progress.

	@exc
	.EModify

	@xmp
		x = [4, 5, 6, 7, 6, 8, 9, 6]
		x.__del (6)
		println (x)

	@out
		[4, 5, 7, 8, 9]
*/
static qu_array qu__array__del (qu_array self, qu_any x)
{
	long i, j;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	for (i = j = 0; i < self->len; i++)
	{
		qu_any o = self->objs[i];
		if (qu_sys_eq (o, x) != QuTrue)
		{
			if (j >= self->len)
				break;
			self->objs[j++] = o;
		}
	}
	if (j <= self->len)
	{
		self->len = j;
		MAYBE_SHRINK (self);
	}
	return self;
}



/*) sub __each -> Iterator
	Get a -{Iterator} object to iterate through the items.

	While creating the Iterator is atomic, the iteration process is not.
	The Array may be modified while iteration is in progress.

	@noexc

	@xmp
		a = [4, 5, 6]
		i = each a      # same as i = a.__each

		println (i.next)
		println (i.next)

	@out
		4
		5
*/
static qu_any qu__array__next (qu_array self, qu_iterator t)
{
	long i = t->ex1;
	QU_ITERATOR_STOP (0 <= i && i < self->len);
	t->ex1 = i + t->ex2;
	return self->objs [i];
}



static qu_iterator qu__array__each (qu_array self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, qu__array__next);
	i->ex2 = 1;
	return i;
}



/*) sub __eq (x) -> Boolean
	An Array is only equal to itself thus this method returns true if v{x}
	is self and false otherwise.

	@noexc
*/
/* qu__array__eq = qu_object.__eq */



/*) sub __getitem (x) -> Any
	@def
	sub __getitem (index:   Integer) -> Any
	sub __getitem (range:   Range  ) -> Array
	sub __getitem (indexes: Array  ) -> Array
	@p
	Get one or more items.

	If v{x} is an integer then get the item at that index.
	A negative value means counting from the end.

	If v{x} is a -{Range} to get a slice based on the range.
	Only range values that fall within this Array are collected.
	If there are none then an empty Array is returned.
	No exception thrown.

	If v{x} is an Array then collect items specified by the v{indexes}
	(same as the #{select} method).

	This is an atomic operation.

	@exc
	.EIndex

	@xmp
		a = [4 5 6]
		println (a[1] ', ' a[-2])

		a = [0 1 2 3 4 5 6]
		println (a [ 8 .. -2 << 2])  #  8, 6, 4, 2, 0, -2
		println (a [-2 ..  8 >> 2])  # -2, 0, 2, 4, 6,  8

	@out
		5, 5
		[6, 4, 2, 0]
		[0, 2, 4, 6]
*/
static qu_any qu__array__getitem (qu_array self, qu_any x)
{
	long i;

	/* the odd looking lable 'doint' is intentional since testing qu_int_is is
		faster than qu_classof, and qu_int_tol is faster than qu_integer_toli,
		and qu_int is usually the case */

	if (qu_int_is (x))
	{
		i = qu_int_tol (x);

	  doint:
		QU_ENSURE_INDEX (i, self->len);
	 	return self->objs[i];
	}
	else if (qu_range_is (x))
	{
		qu_array a;
		long     n = qu_range_cover ((qu_range) x, self->len, &i);

		/* this should never happen - see _range_init */
		QU_MEM_CHECK (n <= QU_GC_MAX_ITEMS);

		a = qu_array_new (n);
		if (n > 0)
		{
			long    z = ((qu_range) x)->step;
			qu_any *p = self->objs;
			qu_any *q = a->objs;

			a->len = n;
			for (; n; i += z, n--)
				*q++ = p[i];
		}
		return a;
	}
	else if (qu_array_is (x))
	{
		return qu__array_select (self, ((qu_array) x)->len, ((qu_array) x)->objs);
	}
	else
	{
		i = qu_integer_toli (x);
		QU_REQUIRE (QuEIndex, errno == 0);
		goto doint;
	}
}



/*) sub __getslice (idx: Integer, x)
	@def
	sub __getslice (idx: Integer, num: Integer) -> Array
	sub __getslice (idx: Integer, defval) -> Any
	@p
	Get at most v{num} objects starting from the specified index v{idx}
	or safely get an object without throwing EIndex.

	In the first form, v{idx} must be an integer index within this String.
	A negative value means counting from the end.
	If v{num} is negative then get all characters starting from v{idx}.

	The second form returns v{defval} if the index is out of bound,
	otherwise the object at v{idx}. The downside is that v{defval} can not
	be an Integer.

	This is an atomic operation.

	@exc
	.EIndex

	@xmp
		a = [1, 2, 3, 4, 5]
		println (a[1, 100])
		println (a[-3, 2])
		println (a[1, 0])
		println (a[1,])       # = a[1, -1]
		println (a[,2])       # = a[0, 2]
		println (a[100, 'x'])

	@out
		[2, 3, 4, 5]
		[3, 4]
		[]
		[2, 3, 4, 5]
		[1, 2]
		x
*/
static qu_array qu__array__getslice (qu_array self, qu_any idx, qu_any num)
{
	long l = self->len;
	long i, n;

	i = qu_integer_toli (idx);
	QU_REQUIRE (QuEIndex, errno == 0);

	n = qu_integer_toli (num);
	if (errno == EDOM)
	{
		if (i < 0)
			i += l;
		if (i < 0 || i > l)
			return num;
		return self->objs[i];
	}
	else
	{
		qu_array r;

		QU_REQUIRE (QuEIndex, errno == 0);
		QU_ENSURE_INDEX (i, l);
		l -= i;
		if (n < 0 || n > l)
			n = l;
		r = qu_array_new (n);
		QU_OBJECT_LIST_COPY (r->objs, self->objs + i, n);
		r->len = n;
		return r;
	}
}



/*) sub __iand (x: Array) -> Array
	Return a new Array containing items which are both in this Array and v{x}.
	Thus the intersection of self and v{x}.

	Note that this does not produce #{unique} items.

	This is an atomic operation unless the __eq method of the items in v{x}
	modifies the Arrays.

	@exc
	.EIand

	@xmp
		a = [2, 1, 2, 3, 4]
		b = [5, 6, 2, 1, 1]
		c = a & b
		println (c)

	@out
		[2, 1, 2]
*/
static qu_array qu__array__iand (qu_array self, qu_array x)
{
	QU_REQUIRE (QuEIand, qu_array_is (x));
	{
		qu_array r = qu_array_new (4);
		long     i;

		for (i = 0; i < self->len; ++i)
		{
			qu_any o = self->objs[i];
			if (qu_array_contains (x, o))
				qu_array_push_quick (r, o);
		}
		return r;
	}
}



/*) sub __in (x) -> Boolean
	Return true if v{x} is an item in this Array, false otherwise.

	This is an atomic operation unless the __eq method of the items or another
	thread modifies this Array while this method is in progress.

	@noexc

	@xmp
		a = [1, 2, 3]
		println (2 in a, ' ', 8 in a)

	@out
		true false

	@p
	The following example demonstrates how it is possible to get an "unexpected"
	non-atomic result.

	@xmp
		ary = nil # make it known to Foo

		class Foo
			sub __eq (x)
				println (__here__)
				ary.pop            #! bad idea
				return x is Foo
			;;
		;;

		ary = [ Foo(), 'foo', 'bar']
		println ('bar' in ary)

	@out
		__eq
		__eq
		false
*/
qu_boolean qu__array__in (qu_array self, qu_any x)
{
	long i;
	for (i = 0; i < self->len; i++)
	{
		if (qu_sys_eq (self->objs[i], x) == QuTrue)
			return QuTrue;
	}
	return QuFalse;
}



/*) sub __ior (x: Array) -> Array
	Return a new Array containing all items in this Array and v{x}.
	Thus the union of self and v{x}.

	Unlike the #{__plus} method, the items are pushed in an rotating manner.
	That is, item from self then item from v{x} then item from k{self], etc.
	If you don't care about the position then use __plus since it is more
	efficient.

	@exc
	.EIor
	.EMemory

	@xmp
		a = [1, 2]
		b = [4, 5, 6, 7]
		c = a | b
		println (c)

	@out
		[1, 4, 2, 5, 6, 7]
*/
static qu_array qu__array__ior (qu_array self, qu_array x)
{
	qu_array r;
	long     i, m;

	QU_REQUIRE (QuEIor, qu_array_is (x));
	QU_MEM_CHECK (self->len <= QU_GC_MAX_ITEMS - x->len);

	r = qu_array_new (self->len + x->len);
	m = self->len;
	if (x->len > m)
		m = x->len;
	for (i = 0; i < m; ++i)
	{
		if (i < self->len)
			qu_array_push_quick (r, self->objs[i]);

		if (i < x->len)
			qu_array_push_quick (r, x->objs[i]);
	}
	return r;
}



/*) sub __ixor (x: Array) -> Array
	Return a new Array containing items which are in this Array but not in v{x}
	and in v{x} but not in this Array. Thus the difference of self and v{x}.

	@exc
	.EIxor

	@xmp
		a = [4, 3, 1, 2]
		b = [2, 1, 5, 6, 7]
		c = a ^ b
		println (c)

	@out
		[4, 3, 5, 6, 7]
*/
static qu_array qu__array__ixor (qu_array self, qu_array x)
{
	qu_array r;
	qu_any   o;
	long     i;

	QU_REQUIRE (QuEIxor, qu_array_is (x));

	r = qu_array_new (4);
	for (i = 0; i < self->len; ++i)
	{
		o = self->objs[i];
		if (!qu_array_contains (x, o))
			qu_array_push_quick (r, o);
	}
	for (i = 0; i < x->len; ++i)
	{
		o = x->objs[i];
		if (!qu_array_contains (self, o))
			qu_array_push_quick (r, o);
	}
	return r;
}



/*) sub __like (x) -> Boolean
	Return true if v{x} is an Array and contains the same items in this
	Array in the same order.

	This is an atomic operation unless the __eq method of the items or another
	thread modifies the Arrays while this method is in progress.

	@xmp
		a = [1, 2, 3]
		b = [3, 2, 1]
		c = [1, 2, 3]
		d = {1, 2, 3}
		println (a === b)
		println (a === c)
		println (a === d)

	@out
		false
		true
		false
*/
qu_boolean qu__array__like (qu_array self, qu_any x)
{
	if (qu_array_is (x) && self->len == ((qu_array) x)->len)
	{
		long i;

		for (i = 0; i < self->len && i < ((qu_array) x)->len; ++i)
		{
			if (qu_sys_eq (self->objs[i], ((qu_array) x)->objs[i]) != QuTrue)
				return QuFalse;
		}

		/* not a mistake, by now the Arrays might have been modified */
		if (self->len == ((qu_array) x)->len)
			return QuTrue;
	}
	return QuFalse;
}



/*) sub __lshift (n: Integer) -> Array
	Return a new Array as the result of excluding the first v{n} items
	of this Array.

	v{n} must be a positive integer.

	See #{shift} for a version that modifies the Array.

	This is an atomic operation.

	@exc
	.ELshift

	@xmp
		a = [1, 2, 3, 4, 5]
		b = a << 2
		println (a)
		println (b)

	@out
		[1, 2, 3, 4, 5]
		[3, 4, 5]
*/
static qu_array qu__array__lshift (qu_array self, qu_integer num)
{
	long n, l = self->len;

	n = qu_integer_tol (num);
	QU_REQUIRE (QuELshift, errno == 0 && n >= 0);

	if (n >= l || l == 0)
		return qu_array_new (0);
	else
	{
		qu_array a;

		l -= n;
		a = qu_array_new (l);
		a->len = l;
		QU_OBJECT_LIST_COPY (a->objs, self->objs + n, l);
		return a;
	}
}



/*) sub __minus (x: Array) -> Array
	Return a new Array containing items in this Array which are not in v{x}.

	This is not an atomic operation.

	@exc
	.EMinus

	@xmp
		a = ['a', 'b', 'c', 'd', 'e']
		b = ['a', 'e']
		c = a - b
		println (c)

	@out
		['b', 'c', 'd']
*/
static qu_array qu__array__minus (qu_array self, qu_array x)
{
	qu_array r;
	qu_any   o;
	long     i;

	QU_REQUIRE (QuEMinus, qu_array_is (x));

	r = qu_array_new (4);
	for (i = 0; i < self->len; ++i)
	{
		o = self->objs[i];
		if (!qu_array_contains (x, o))
			qu_array_push_quick (r, o);
	}
	return r;
}



/*) sub __mul (x)
	@def
	sub __mul (n: Integer) -> Array
	sub __mul (s: String) -> String
	@p
	If called with an Integer argument then return a new Array containing v{n} times
	the items in this Array.

	If called with a String argument then this method returns C{self.join (s)}.

	@exc
	.EArgValue
	.EMemory

	@xmp
		x = [1, 2] * 3 * ':'
		println (x)

	@out
		1:2:1:2:1:2
*/
static qu_any qu__array__mul (qu_array self, qu_any x)
{
	if (qu_string_is (x))
	{
		return qu__array_join (self, 1, &x);
	}
	else
	{
		long n = qu_integer_toli (x);

		QU_REQUIRE (QuEMul, errno == 0 && n >= 0);

		if (n == 0)
		{
			return qu_array_new (0);
		}
		else
		{
			long l = self->len;

			QU_MEM_CHECK (l <= QU_GC_MAX_ITEMS / n);
			{
				qu_any   *p = self->objs;
				long      m = l * n;
				qu_array  r = qu_array_new (m);
				qu_any   *q = r->objs;

				while (n--)
				{
					QU_OBJECT_LIST_COPY (q, p, l);
					q += l;
				}
				r->len = m;
				return r;
			}
		}
	}
}



/*) sub __of (type) -> self
	Assign a I{validator} so that the Array only contain items conforming
	to v{type}.

	EValue is thrown if a #{type} is already assigned or the Array already
	contains incompatible items.

	This is an atomic operation.
	The Array is #{frozen} during the execution of this method.

	@warn
	The -{Null} class may not be used as a type.

	Once set, you can not unset the type.

	@exc
	.EModify
	.EArgValue
	.EValue

	@xmp
		sub foo (x)
			require (x in 1 .. 10 or x == 'foo')
		;;

		a = [1, 2, 'foo', 3] of @foo
		println (a.type)
		a[0] = 5
		try
			a[1] = 30
		catch
			println (__err__)
		;;

		a = [1, 2, 3] of Int
		println (a.type)
		try
			a[0] = 'foo'
		catch
			println (__err__)
		;;

	@out
		<Sub __main__.foo>
		(EValue) invalid type or value
		<Class sys.Int>
		(EValue) invalid type or value
*/
qu_array qu__array__of (qu_array self, qu_any type)
{
	QU_ARG_CHECK (qu_object_valid_type (type));
	if (qu_var_is (type) && !qu_var_is_typeof (type))
		type = ((qu_var) type)->value;

	if (self->type != type)
	{
		QU_REQUIRE (QuEModify, self->type == 0);
		if (self->len)
		{
			int ok;
		  __qu_container_protect_begin (self);
			ok = (qu__array_typeof (self, type) == QuTrue);
		  __qu_container_protect_end;
			QU_ARG_CHECK (ok);
		}
		self->type = type;
	}
	return self;
}




/*) sub __plus (x: Array) -> Array
	Return a new Array which contains items in this Array followed by items in v{x}.

	@exc
	.EPlus
	.EMemory

	@xmp
		a = [1, 2, 3]
		b = [4, 5]
		c = a + b
		println (c)

	@out
		[1, 2, 3, 4, 5]
*/
static qu_array qu__array__plus (qu_array self, qu_array x)
{
	long     n;
	qu_array r;

	QU_REQUIRE (QuEPlus, qu_array_is (x));
	QU_MEM_CHECK (self->len <= QU_GC_MAX_ITEMS - x->len);

	n = self->len + x->len;
	r = qu_array_new (n);
	QU_OBJECT_LIST_COPY (r->objs, self->objs, self->len);
	QU_OBJECT_LIST_COPY (r->objs + self->len, x->objs, x->len);
	r->len = n;
	return r;
}



/*) sub __print -> Null
	Print the Array to v{stdout}.

	See -{print} for possible exceptions.

	@xmp
		x = [1 2 'foo']
		print (x)        # calls this method

	@out
		[1, 2, 'foo']
*/
static qu_null qu__array__print (qu_array self)
{
	if (self->len < 1)
	{
		QU_ENSURE (qu_sys_printz ("[]", 2));
	}
	else if (qu_thread_state_exist (QU_THREAD_STATE_PRINT, self))
	{
		QU_ENSURE (qu_sys_printz ("[..]", 4));
	}
	else
	{
		qu_sys_printb_s b;
		long   i;
		int    r;
		qu_any x;

		qu_sys_printb_init (b);

	  __qu_thread_state_begin (QU_THREAD_STATE_PRINT, self);

		r = qu_sys_printb (b, "[", 1);
		for (i = 0; r && i < self->len; i++)
		{
			x = self->objs[i];
			if (i > 0)
				r = qu_sys_printb (b, ", ", 2);
			if (r)
			{
				x = QU_CONTAINER_QUOTE (x);
				r = qu_sys_printb_string (b, x);
			}
		}
		if (r)
			r = qu_sys_printb (b, "]", 1);
		if (r)
			r = qu_sys_printb_final (b);

	  __qu_thread_state_end;
		QU_ENSURE (r);
	}
	return QuNil;
}



/*) sub __pop -> Any
	Pop and return the last item.

	This is an atomic operation.

	@exc
	.EModify
	.ENone    Array is empty

	@xmp
		a = [1, 2, 3, 4, 5, 6]
		a >>> b
		println (b)
		println (a)

	@out
		6
		[1, 2, 3, 4, 5]
*/
static qu_any qu__array__pop (qu_array self)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	QU_REQUIRE (QuENone, self->len);
	{
		qu_any x = self->objs[--(self->len)];
		MAYBE_SHRINK (self);
		return x;
	}
}



/*) sub __push (x) -> self
	Append v{x} to the end of the Array.

	This is an atomic operation unless the Array has a #{type} that involves
	calling functions or methods that modifies the Array.

	@exc
	.EModify
	.EValue
	.EMemory

	@xmp
		a = [1, 2, 3]
		a <<< 'foo'
		println (a)

	@out
		[1, 2, 3, 'foo']
*/
qu_array qu__array__push (qu_array self, qu_any x)
{
	/* in this order */
	ENSURE_VALUE (self, x);
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	/*--- atomic below this point */

	QU_ENSURE (qu_array_make_space (self, 1));
	self->objs[self->len++] = x;
	return self;
}



/*) sub __rshift (n: Integer) -> Array
	Return a new Array containing the last v{n} items of this Array.

	v{n} must be a positive integer.

	This is an atomic operation.

	@exc
	.ERshift

	@xmp
		a = [1, 2, 3, 4, 5]
		b = a >> 2
		println (a)
		println (b)

	@out
		[1, 2, 3, 4, 5]
		[1, 2, 3]
*/
static qu_array qu__array__rshift (qu_array self, qu_integer num)
{
	long n = qu_integer_tol (num);

	QU_REQUIRE (QuELshift, errno == 0 && n >= 0);
	{
		long l = self->len;
		if (n >= l || l == 0)
			return qu_array_new (0);
		else
		{
			qu_array a = qu_array_new (n);
			a->len = l - n;
			QU_OBJECT_LIST_COPY (a->objs, self->objs, l - n);
			return a;
		}
	}
}



/*) sub __setitem (what, x) -> Any
	@def
	sub __setitem (index: Integer, x) -> Any
	sub __setitem (range: Range  , x) -> Any
	@p
	Set one or more items then return v{x} as the result.

	If v{what} is an integer, set the item at that index.
	A negative value means counting from the end.

	Otherwise, v{what} must be -{Range} to set the items based on the range.
	Range values that do not fall within the Array are simply ignored.
	No exception thrown.

	This is an atomic operation unless the Array has a #{type} that involves
	calling functions or methods that modifies the Array.

	@exc
	.EModify
	.EArgValue
	.EIndex
	.EValue

	@xmp
		a = [4, 5, 6]
		a[2] = 0
		println (a)

		a = [1, 2, 3, 4, 5, 6, 7]
		a [-3 .. 12 >> 2] = 0
		println (a)

	@out
		[4, 5, 0]
		[1, 0, 3, 0, 5, 0, 7]
*/
static qu_any qu__array__setitem (qu_array self, qu_any what, qu_any x)
{
	long i;

	/* in this order */
	ENSURE_VALUE (self, x);
	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

	/*--- atomic below this point */

	if (qu_int_is (what))
	{
	  doint:

		i = qu_int_tol (what);

		QU_ENSURE_INDEX (i, self->len);
		self->objs [i] = x;
	}
	else if (qu_range_is (what))
	{
		long    n  = qu_range_cover ((qu_range) what, self->len, &i);
		long    z = ((qu_range) what)->step;
		qu_any *p = self->objs;

		while (n-- > 0)
		{
			p [i] = x;
			i += z;
		}
	}
	else
	{
		i = qu_integer_toli (what);
		QU_REQUIRE (QuEIndex, errno == 0);
		goto doint;
	}
	return x;
}



/*) sub __setslice (idx: Integer, num: Integer, x) -> Any
	Set at most v{num} items starting from the specified index v{idx} with the
	value v{x} and return v{x} as the result.

	A negative v{idx} means counting from the end.
	A negative v{num} means to set all items starting from v{idx}.

	This is an atomic operation unless the Array has a #{type} that involves
	calling functions or methods that modifies the Array.

	@exc
	.EModify
	.EArgValue
	.EIndex
	.EValue

	@xmp
		a = [6, 7, 8]
		a[1, 2] = 0
		println (a)

	@out
		[6, 0, 0]
*/
static qu_any qu__array__setslice (qu_array self, qu_integer idx, qu_integer num, qu_any x)
{
	long i, n, l = self->len;

	i = qu_integer_toli (idx); QU_REQUIRE (QuEIndex, errno == 0);
	n = qu_integer_toli (num); QU_REQUIRE (QuEIndex, errno == 0);

	/* in this order */
	ENSURE_VALUE (self, x);
	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

	/*--- atomic below this point */

	if (n < 0)
		n = l;

	if (n > 0)
	{
		qu_any *p;
		long    m;

		QU_ENSURE_INDEX (i, l);
		p = self->objs + i;
		m = l - i;
		if (n > m)
			n = m;
		while (n--)
			*p++ = x;
	}
	return x;
}



/*) sub __string -> String
	Return a String representation of self.

	This is an atomic operation unless the items __string method modifies
	the Array.

	@exc
	.EMemory

	@xmp
		a = [1, 2, 3]
		a <<< a
		s = String (a)
		s[0] = '<'
		s[-1] = '>'
		println (s)

	@out
		<1, 2, 3, [..]>
*/
qu_string qu__array__string (qu_array self)
{
	if (self->len < 1)
	{
		return qu_string_new ("[]", 2);
	}
	else if (qu_thread_state_exist (QU_THREAD_STATE_STRING, self))
	{
		return qu_string_new ("[..]", 4);
	}
	else
	{
		qu_string s = qu_string_buffer_new (8);
		long      i;

	  __qu_thread_state_begin (QU_THREAD_STATE_STRING, self);

		qu_string_buffer_putc (s, '[');
		for (i = 0; i < self->len; i++)
		{
			qu_any x = self->objs[i];

			if (i > 0)
				qu_string_buffer_put (s, ", ", 2);
			qu_string_buffer_puts (s, QU_CONTAINER_QUOTE (x));
		}
		qu_string_buffer_putc (s, ']');

	  __qu_thread_state_end;

		return s;
	}
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub chop (*arg) -> Array
	@def
	sub chop (i: Integer, n: Integer) -> Array
	sub chop (n: Integer) -> Array
	@p
	Remove v{n} items from the Array.

	If called with one argument, remove items from the head or tail.
	If v{n} is positive then remove v{n} items from the tail.
	If v{n} is negative then remove v{n} items from the head.

	Otherwise remove v{n} items starting from the index v{i}.
	If v{i} is negative then count from the end.
	v{n} must be a non-negative number.

	This is an atomic operation.

	@exc
	.EArgValue
	.EIndex
	.EModify

	@xmp
		a = [1, 2, 3, 4, 5]
		a.chop (-3)
		println (a)

		a = [1, 2, 3, 4, 5 , 6, 7]
		a.chop (2, 3)
		println (a)

	@out
		[4, 5]
		[1, 2, 6, 7]
*/
static int qu__array__chop_args (int argc, qu_any *argv, long l, long *idx, long *num)
{
	long i, n;

	if (argc == 1)
	{
		QU_ARG_vlong (0, n);
		if (n >= 0)
		{
			if (n > l)
				n = l;
			i = n;
		}
		else
		{
			n = -n;
			if (n >= l)
				n = l;
			i = 0;
		}
	}
	else
	{
		QU_ARG_VEXACT (2);
		QU_ARG_vlong (0, i);
		QU_ARG_vlong (1, n);
		QU_ARG_CHECK  (n >= 0);
		QU_ENSURE_INDEX (i, l);
		if (n > l - i)
			n = l - i;
	}

	*idx = i;
	*num = n;
	return 1;
}



static qu_array qu__array_chop (qu_array self, int argc, qu_any *argv)
{
	long l = self->len;
	long i, n;

	QU_ENSURE (qu__array__chop_args (argc, argv, l, &i, &n));
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	if (n >= l)
	{
		self->len = 0;
		_array_resize (self, 0);
	}
	else if (i >= l || n == 0)
	{
		/* nothing to do */
	}
	else
	{
		l -= n;
		QU_OBJECT_LIST_COPY (self->objs + i, self->objs + i + n, l - i);
		self->len = l;
		MAYBE_SHRINK (self);
	}
	return self;
}



/*) sub copy (keeptype = false) -> Array
	Make a shallow copy of this Array.

	If v{keeptype} is true then the new Array is given the same #{type} as
	this Array. Otherwise this method is just an alias of #{__copy}.

	@exc
	.EArgValue
*/
static qu_array qu__array_copy (qu_array self, int argc, qu_any *argv)
{
	int      keep = 0;
	qu_array r;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, keep);
	r = qu__array__copy (self);
	if (keep)
		r->type = self->type;
	return r;
}



/*) sub delete (*arg) -> self
	Delete all items specified in v{arg} from the Array.

	See #{__del} for a version that deletes only one argument.

	If called without any argument then this method simply calls #{__pop}.

	This is an atomic operation unless the __eq methods of the items in the Array
	modifies the Array.

	@exc
	.EModify

	@xmp
		x = [4, 5, 6, 7, 6, 8, 9, 6]
		x.delete (4, 6)
		println (x)

	@out
		[5, 7, 8, 9]
*/
static qu_array qu__array_delete (qu_array self, int argc, qu_any *argv)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	if (argc == 0)
	{
		if (self->len)
		{
			self->len--;
			MAYBE_SHRINK (self);
		}
	}
	else
	{
		long i, j;

		QU_ARG_VMIN (1);

		for (i = j = 0; i < self->len; i++)
		{
			qu_any  o = self->objs[i];
			int     m = argc;
			qu_any *p = argv;

			while (m--)
			{
				if (qu_sys_eq (o, *p++) == QuTrue)
					goto next;
			}

			if (j >= self->len)
				break;
			self->objs[j++] = o;

		next:
			continue;
		}
		if (j <= self->len)
		{
			self->len = j;
			MAYBE_SHRINK (self);
		}
	}
	return self;
}



/*) sub grow (size: Integer, x = nil) -> self
	Grow the Array by filling in the gaps with v{x}.

	This is an atomic operation unless the Array has a #{type} that involves
	calling functions or methods that modifies the Array.

	@exc
	.EModify
	.EArgValue
	.EValue
	.EMemory

	@xmp
		a = [0 1 2]
		a.grow (5, 'foo')
		println (a)

	@out
		[0, 1, 2, 'foo', 'foo']
*/
static qu_array qu__array_grow (qu_array self, int argc, qu_any *argv)
{
	long    n;
	qu_any  x = QuNil;
	long    l, m;
	qu_any *p;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong (0, n);
	QU_ARG_CHECK (n >= 0);
	if (argc > 1)
		x = argv[1];

	/* in this order */
	ENSURE_VALUE (self, x);
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	/*--- atomic below this point */

	l = self->len;
	m = n - l;
	if (m > 0)
	{
		QU_MEM_CHECK (n <= QU_GC_MAX_ITEMS);
		_array_resize (self, n);
		p = self->objs + l;
		self->len += m;
		while (m--)
			*p++ = x;
	}
	return self;
}



/*) sub head (separator = nil) -> self
	Retain items before the v{separator}.

	If v{seperator} is not found then nothing happens.

	This is an atomic operation unless the __eq method of the items or another
	thread modifies this Array while this method is in progress.

 	@exc
	.EModify

	@xmp
		a = [1, 2, nil, 3]
		a.head
		println (a)

	@out
		[1, 2]
*/
static qu_array qu__array_head (qu_array self, int argc, qu_any *argv)
{
	qu_any sep = QuNil;
	long   i;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	QU_ARG_VMAX (1);
	if (argc > 0)
		sep = argv[0];
	i = qu_array_index (self, sep, 0);
	if (i >= 0)
	{
		self->len = i;
		MAYBE_SHRINK (self);
	}
	return self;
}



/*) sub index (x, offset = 0) -> Int
	Return the position of v{x} in the Array, optionally start at v{offset}.

	Return -1 if not found.

	This is an atomic operation unless the __eq method of the items or another
	thread modifies this Array while this method is in progress.

	@exc
	.EArgValue

	@xmp
		a = [6, 7, 7, 8]
		println (a.index (7))

	@out
		1
*/
static qu_int qu__array_index (qu_array self, int argc, qu_any *argv)
{
	long i = 0;

	QU_ARG_VCOUNT (1, 2);
	if (argc > 1)
		QU_ARG_vlong (1, i);
	i = qu_array_index (self, argv[0], i);
	return qu_int_new (i);
}



/*) sub insert (index: Integer, *arg) -> self
	Insert items in v{arg} starting from v{index}.

	This is an atomic operation.

	@exc
	.EModify
	.EArgValue
	.EValue
	.EMemory

	@xmp
		sub foo (x)
			return 'x' _ x
		;;

		a = [1, 2] of @foo

		a.insert (1, 'foo', 'bar')
		println (a)

	@out
		[1, 'xfoo', 'xbar', 2]
*/
static qu_array qu__array_insert_do (qu_array self, qu_any ii, int argc, qu_any *argv)
{
	long i;
	long l = self->len;

	QU_ARG_VMIN (1);
	QU_ARG_long (ii, i);

	if (i == self->len)
		return qu__array_push (self, argc, argv);

	QU_ENSURE_INDEX (i, l);

	if (self->type)
	{
		while (argc--)
		{
			qu_any x = *argv++;

			ENSURE_VALUE (self, x);
			QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
			l = self->len;
			QU_ENSURE_INDEX (i, l);
			QU_ENSURE (qu_array_make_space (self, 1));
			QU_OBJECT_LIST_RCOPY (self->objs + l, self->objs + i, l - i);
			self->objs[i++] = x;
			self->len++;
		}
	}
	else
	{
		qu_any *p;

		QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
		QU_ENSURE (qu_array_make_space (self, argc));
		self->len += argc;
		p = self->objs + i;
		QU_OBJECT_LIST_RCOPY (p + argc, p, l - i);
		while (argc--)
			*p++ = *argv++;
	}
	return self;
}



static qu_array qu__array_insert (qu_array self, int argc, qu_any *argv)
{
	QU_ARG_VMIN (2);
	return qu__array_insert_do (self, argv[0], argc - 1, argv + 1);
}



/*) sub inject (*arg) -> self
	This is a convenience method that calls C{self.insert (0, *arg)}.

	This is an atomic operation.

	@exc
	.EModify
	.EMemory
	.EValue

	@xmp
		a = [1, 2, 3]
		println (a.inject (4, 5))

	@out
		[4, 5, 1, 2, 3]
*/
/*) sub unshift (*arg) -> self
	This is an alias of #{inject}.
*/
static qu_array qu__array_inject (qu_array self, int argc, qu_any *argv)
{
	return qu__array_insert_do (self, QuInt0, argc, argv);
}



/*) sub join (*arg) -> String
	@def
	sub join (sep = '') -> String
	sub join (sep, range: Range) -> String
	sub join (sep, index = 0, count = self.__len) -> String
	sub join (range: Range) -> String
	sub join (index = 0, count = self.__len) -> String
	@p
	Concatenate items as strings using v{sep} as the item separator.

	If v{sep} is not specified then an empty string is assumed.

	v{index} is the index to start. A negative value means to count from
	the end.

	v{len} is the number of items to concatenate.

	If self is empty or the index or range falls outside of the array
	then an empty string is returned. No exception thrown.

	@exc
	.EArgValue
	.EMemory

	@xmp
		a = [1, [2, 3], 4]
		println (a.join (':'))

		a = [0, 1, 2, 3, 4, 5, 6]
		println (a.join (',', 3, 2))
		println (a.join (-1 .. 15 >> 2))

	@out
		1:[2, 3]:4
		3,4
		135
*/
static qu_string qu__array_join (qu_array self, int argc, qu_any *argv)
{
	long      l = self->len;
	long      i = 0;
	long      n = l;
	long      d = 1;
	qu_string sep = QuStringEMPTY;
	qu_string s;
	int       e, k;
	qu_any    x;

	if (argc > 0)
	{
		x = argv[0];
		if (qu_string_is (x))
		{
			sep = x;
			if (argc > 1)
			{
				if (qu_range_is (argv[1]))
				{
					qu_range r = argv[1];
					n = qu_range_cover (r, l, &i);
					d = r->step;
				}
				else
				{
					QU_ARG_vlong (1, i);
					if (argc > 2)
						QU_ARG_vlong (2, n);
					if (i < 0)
						i += l;
				}
			}
		}
		else if (qu_range_is (x))
		{
			qu_range r = x;
			n = qu_range_cover (r, l, &i);
			d = r->step;
		}
		else
		{
			QU_ARG_vlong (0, i);
			if (argc > 1)
				QU_ARG_vlong (1, n);
			if (i < 0)
				i += l;
		}
	}

	if (n == 0)
		return QuStringEMPTY;

	s = qu_string_buffer_new (self->len / 2 + 8);
	e = sep->len > 0;
	k = 0;
	while (n-- && 0 <= i && i < self->len)
	{
		if (k && e)
			QU_ENSURE (qu_string_buffer_puts (s, sep));
		k = 1;
		x = self->objs [i];
		if (x == self)
			QU_ENSURE (qu_string_buffer_put (s, "[..]", 4));
		else if (qu_string_is (x))
			QU_ENSURE (qu_string_buffer_puts (s, x));
		else
			QU_ENSURE (qu_string_buffer_puts (s, QU_CONTAINER_QUOTE (x)));
		i += d;
	}
	return qu_string_buffer_return (s);
}



/*) sub pop (*arg) -> Any
	@def
	sub pop -> Any
	sub pop (defval) -> Any
	@p
	Pop the last item out of the Array.

	The first form will throw ENone if the Array is empty while the second
	will return v{defval}.

	@exc
	.EModify
	.ENone

	@xmp
		x = [1, 2, 3]

		y = x.pop
		println (y ' ' x)

		y = x.pop
		y = x.pop
		println (y ' ' x)

		y = x.pop (nil)
		println (y ' ' x)

	@out
		3 [1, 2]
		1 []
		nil []
*/
static qu_any qu__array_pop (qu_array self, int argc, qu_any *argv)
{
	long l = self->len;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	QU_ARG_VCOUNT (0, 1);
	if (l < 1)
	{
		QU_REQUIRE (QuENone, argc);
		return argv[0];
	}
	else
	{
		qu_any x = self->objs[--self->len];
		MAYBE_SHRINK (self);
		return x;
	}
}



/*) sub popr (*arg) -> Any
	@def
	sub popr -> Any
	sub popr (defval) -> Any
	@p
	Pop the first item out of the Array.

	The first form will throw ENone if the Array is empty while the second
	will return v{defval}.

	@exc
	.EModify
	.ENone

	@xmp
		x = [1, 2, 3]

		y = x.popr
		println (y ' ' x)

		y = x.popr
		y = x.popr
		println (y ' ' x)

		y = x.popr (nil)
		println (y ' ' x)

	@out
		1 [2, 3]
		3 []
		nil []
*/
static qu_any qu__array_popr (qu_array self, int argc, qu_any *argv)
{
	long l = self->len;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	QU_ARG_VCOUNT (0, 1);
	if (l < 1)
	{
		QU_REQUIRE (QuENone, argc);
		return argv[0];
	}
	else
	{
		qu_any x = self->objs[0];
		QU_OBJECT_LIST_COPY (self->objs, self->objs + 1, --self->len);
		MAYBE_SHRINK (self);
		return x;
	}
}



/*) sub push (*arg) -> self
	Push all arguments into the Array.

	There must be at least one argument.

	@exc
	.EModify

	@xmp
		x = [1, 2, 3]

		x.push (4, 5)
		println (x)

	@out
		[1, 2, 3, 4, 5]
*/
static qu_array qu__array_push (qu_array self, int argc, qu_any *argv)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	QU_ARG_VMIN (1);
	QU_ENSURE (qu_array_make_space (self, argc));
	while (argc--)
		QU_ENSURE (qu_array_push (self, *argv++));
	return self;
}



/*) sub replace (x, rep, max = -1) -> self
	Replace items in this Array.

	v{x} is the item to replace, v{rep} is the replacement, and v{max] is
	the maximum number of replacements to perform

	If v{max} is non-positive then there is no limit.

	The Array is temporarily I{protected} while this method is in progress.

	@exc
	.EModify
	.EArgValue
	.EValue

	@xmp
		a = [1, 5, 5, 5, 7, 8]
		a.replace (5, 0, 2)
		println (a)

	@out
		[1, 0, 0, 5, 7, 8]
*/
static qu_array qu__array_replace (qu_array self, int argc, qu_any *argv)
{
	qu_any what, with;
	long   max = LONG_MAX;
	long   i;

	QU_ARG_VCOUNT (2, 3);
	what = argv[0];
	with = argv[1];
	if (argc > 2)
	{
		QU_ARG_vlong (2, max);
		if (max <= 0)
			max = LONG_MAX;
	}

	/* in this order */
	ENSURE_VALUE (self, with);
	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

  __qu_container_protect_begin (self);
	for (i = 0; max && i < self->len; ++i)
	{
		if (qu_sys_eq (self->objs[i], what) == QuTrue)
		{
			assert (i < self->len);
			self->objs[i] = with;
			--max;
		}
	}
  __qu_container_protect_end;
	return self;
}



/*) sub reverse -> self
	Reverse the items.

	This is an atomic operation.

	@exc
	.EModify

	@xmp
		a = [1 2 3]
		a.reverse
		println (a)

	@out
		[3, 2, 1]
*/
static qu_array qu__array_reverse (qu_array self)
{
	long l;

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

	l = self->len;
	if (l > 1)
	{
		register qu_any *i = self->objs;
		register qu_any *j = i + l - 1;
		register qu_any  x;

		l /= 2;
		while (l--)
		{
			x = *j;
			*j-- = *i;
			*i++ = x;
		}
	}
	return self;
}



/*) sub rindex (x, offset = self.__len - 1) -> Int
	Return the position of v{x} in the Array searching from the end and optionally
	start at v{offset}.

	Return -1 if not found.

	This is an atomic operation unless the __eq method of the items or another
	thread modifies this Array while this method is in progress.

	@exc
	.EArgValue

	@xmp
		a = [6, 7, 8, 8, 9]
		println (a.rindex (3))

	@out
		3
*/
static qu_int qu__array_rindex (qu_array self, int argc, qu_any *argv)
{
	long i = self->len - 1;

	QU_ARG_VMIN (1);
	if (argc > 1)
		QU_ARG_vlong (1, i);
	i = qu_array_rindex (self, argv[0], i);
	return qu_int_new (i);
}



/*) sub select (*indexes) -> Array
	Select items specified by v{indexes} into an Array.

	v{indexes} may be integer indexes or Range objects.

	@exc
	.EArgValue
	.EIndex
	.EMemory

	@xmp
		x = [1, 2, 3, 4, 5, 6, 7]
		println (x.select (0, 4, -1, 2..3))

	@out
		[1, 5, 7, 3, 4]
*/
static qu_array qu__array_select (qu_array self, int argc, qu_any *argv)
{
	qu_array a = qu_array_new (argc);

	while (argc--)
	{
		qu_any x = *argv++;
		long   i;

		if (qu_int_is (x))
		{
			i = qu_int_tol (x);

		  doint:
			QU_ENSURE_INDEX (i, self->len);
			QU_ENSURE (qu_array_make_space (a, 1));
		 	a->objs[a->len++] = self->objs[i];
		}
		else if (qu_range_is (x))
		{
			long n = qu_range_cover ((qu_range) x, self->len, &i);
			if (n > 0)
			{
				QU_ENSURE (qu_array_make_space (a, n));
				{
					long   z = ((qu_range) x)->step;
					qu_any *p = self->objs;
					qu_any *q = a->objs + a->len;

					a->len += n;
					for (; n; i += z, n--)
						*q++ = p[i];
				}
			}
		}
		else
		{
			i = qu_integer_toli (x);
			QU_REQUIRE (QuEIndex, errno == 0);
			goto doint;
		}
	}
	return a;
}



/*) sub shift (num = 1) -> Array
	Shift out the first v{num} items from this Array and return them in
	a new Array.

	This is an atomic operation.

	See #{chop} if you don't need the result.

	@exc
	.EModify
	.EArgValue
	.ENone

	@xmp
		a = [1, 2, 3, 4, 5]

		b = a.shift
		println (a)
		println (b)

		b = a.shift (2)
		println (a)
		println (b)

	@out
		[2, 3, 4, 5]
		[1]
		[4, 5]
		[2, 3]
*/
static qu_array qu__array_shift (qu_array self, int argc, qu_any *argv)
{
	long l = self->len;
	long n = 1;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	if (argc > 0)
	{
		QU_ARG_vlong (0, n);
		QU_ARG_CHECK (n > 0);
	}

	if (l)
	{
		qu_array r;

		l -= n;
		if (l <= 0)
		{
			long    x;
			qu_any *y;

			r = qu_array_new (0);

			x = self->cap;
			self->cap = r->cap;
			r->cap = x;

			y = self->objs;
			self->objs = r->objs;
			r->objs = y;

			r->len = self->len;
			self->len = 0;
		}
		else
		{
			r = qu_array_new (l);
			QU_OBJECT_LIST_COPY (r->objs, self->objs, n);
			r->len = n;

			QU_OBJECT_LIST_COPY (self->objs, self->objs + n, l);
			self->len = l;
			MAYBE_SHRINK (self);
		}
		return r;
	}
	else
	{
		return qu_array_new (0);
	}
}



/*) sub shrink (size: Integer) -> self
	Shrink self to v{size}.

	v{size} must be a non-negative integer.
	If it is bigger than the current size then nothing happens.

	See #{grow} to grow the Array.

	@exc
	.EModify
	.EArgValue

	@xmp
		a = [0 1 2 3 4]
		a.shrink (3)
		println (a)

	@out
		[0 1 2]
*/
static qu_array qu__array_shrink (qu_array self, qu_integer num)
{
	long l, n;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	QU_ARG_long (num, n);
	QU_ARG_CHECK (n >= 0);
	l = self->len;
	if (n < self->len)
	{
		self->len = n;
		MAYBE_SHRINK (self);
	}
	return self;
}



/*) sub sort (*what) -> self
	@def
	sub sort -> self
	sub sort (ascending: true) -> self
	sub sort (func: Function) -> self
	@p
	Sort the items in self inplace.

	If called without any arguments, sort the items in a ascending order.

	If called with one Boolean argument, it must be true to sort the items
	in ascending order or false to sort the items in a descending order.

	If called with a Function argument, the Function will be called to
	compare items. It must take at least two arguments and return a negative
	Int to indicate that the first argument is greater than the second,
	a positive Int to indicate that the first argument is smaller than the
	second, 0 to indiciate that they are equal.

	@warn
	The Function is called as a trusted function. This means that sorting an
	Array from within a method will expose self to the Function.

	The Array is temporarily frozen during the sort. Attempt to modify self
	will cause an EModify exception which is ignored by the f{sort} method.

	Any exception thrown by the Function causes the sort to stop and the
	exception rethrown.

	@exc
	.EModify
	.EArgValue
	.Exception anything thrown by the Function

	@xmp
		a = [4, 3, 5, 1, 2].sort (2 <=> 1)
		println (a)

	@out
		[5, 4, 3, 2, 1]

	@note
	If not using a Function then to sort items the __cmp method is called
	only if the two objects being compared are of the same class. Otherwise
	the method provided by the Object class is used. This is done because
	C{x <=> y} may return a different result than C{y <=> x} and we can not
	let that happen while sorting.
*/
static int qu__array_sort_asc (qu_any *i, qu_any *j)
{
	qu_any r = (qu_classof (*i) == qu_classof (*j)) ? qu_sys_cmp (*i, *j) : qu_object_default_cmp (*i, *j);
	QU_SYS_SORT_RETURN_RESULT (r);
}



static int qu__array_sort_dsc (qu_any *i, qu_any *j)
{
	qu_any r = (qu_classof (*i) == qu_classof (*j)) ? qu_sys_cmp (*i, *j) : qu_object_default_cmp (*i, *j);
	QU_SYS_SORT_RETURN_RESULT (r);
}



static int qu__array_sort_fun (qu_any *i, qu_any *j, qu_function f, qu_any self)
{
	qu_any r = qu_method_is (f) ? qu_method_call2 ((qu_method) f, *i, *j) : qu_sub_call2 ((qu_sub) f, self, *i, *j);
	QU_SYS_SORT_RETURN_RESULT (r);
}



static qu_array qu__array_sort (qu_array self, int argc, qu_any *argv)
{
	qu_any how = QuTrue;
	int    res;

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

	if (self->len < 2)
		return self;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
	{
		how = argv[0];
		if (!qu_boolean_is (how))
			QU_ARG_TYPE (how, qu_function);
	}

  __qu_container_protect_begin (self);

	if (how == QuTrue)
		res = qu_sys_sort ((qu_sys_sort_f) qu__array_sort_asc, self->objs, sizeof (qu_any), self->len);

	else if (how == QuFalse)
		res = qu_sys_sort ((qu_sys_sort_f) qu__array_sort_dsc, self->objs, sizeof (qu_any), self->len);

	else
		res = qu_sys_osort ((qu_sys_osort_f) qu__array_sort_fun, self->objs, sizeof (qu_any), self->len, how, self);

  __qu_container_protect_end;

	QU_ENSURE (res);
	return self;
}



/*) sub split (*arg)
	@def
	sub split (index: Integer) -> Array
	@p
	Split self into two new Array objects at the specified index.

	self is not modified.

	A negative v{index} means counting from the end.

	@exc
	.EArgValue
	.EIndex

	@xmp
		a = [1 2 3 4 5 6 7]
		(b, c) = a.split (3)
		println (b)
		println (c)

	@out
		[1, 2, 3]
		[4, 5, 6, 7]
*/
static qu_array qu__array_split (qu_array self, int argc, qu_any *argv)
{
	long     l = self->len;
	qu_array a = 0;

	if (argc == 1)
	{
		long i;

		QU_ARG_vlong (0, i);
		QU_ENSURE_INDEX (i, l);

		a = qu_array_new (2);
		if (i >= l)
		{
			qu_array_push_quick (a, qu__array__copy (self));
			qu_array_push_quick (a, qu_array_new (0));
		}
		else if (i < 1)
		{
			qu_array_push_quick (a, qu_array_new (0));
			qu_array_push_quick (a, qu__array__copy (self));
		}
		else
		{
			qu_array b, c;

			l -= i;
			b = qu_array_new (i);
			c = qu_array_new (l);
			QU_OBJECT_LIST_COPY (b->objs, self->objs, i);
			QU_OBJECT_LIST_COPY (c->objs, self->objs + i, l);
			b->len = i;
			c->len = l;
			qu_array_push_quick (a, b);
			qu_array_push_quick (a, c);
		}
	}
	else
	{
		/* --- put code for other future arguments here */

		QU_THROW (QuEArgCount);
	}
	return a;
}



/*) sub tail (separator = nil) -> self
	Retain items after v{separator}.

	If the seperator is not found then nothing happens.

	@exc
	.EModify

	@xmp
		x = [1, 2, nil, 3]
		x.tail
		println (x)

	@out
		[3]
*/
static qu_array qu__array_tail (qu_array self, int argc, qu_any *argv)
{
	qu_any sep = QuNil;
	long   i;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		sep = argv[0];

	i = qu_array_index (self, sep, 0);
	if (i >= 0)
	{
		++i;
		self->len -= i;
		memcpy (self->objs, self->objs + i, self->len * sizeof (qu_any));
	}
	return self;
}



/*) sub type -> Any
	Return the type of items in self or nil if none has been assigned.

	See the #{__of} method.

	@xmp
		a = [1, 2, 3] of Int
		println (a.type)

	@out
		<Class sys.Int>
*/
static qu_any qu__array_type (qu_array self)
{
	if (self->type == 0)
		return QuNil;
	return self->type;
}



/*) sub typeof (x) -> Boolean
	Check whether or not all items in self are of v{x}.

	v{x} may be any validator.
	If it is a -{Record} then its -{Struct} is used.
	Otherwise it is used as is.

	If self has been designated to contain items of v{x} (the #{type} method)
	then true is returned even if self is empty. Otherwise, if self is
	empty then false is returned.

	This is an atomic operation unless v{x} involves calling functions/methods
	that modifies this Array.

	@noexc

	@xmp
		a = [] of Int
		println (a.typeof (Int))

		a = [1, 2, 3]
		println (a.typeof (Int))

		a = []
		println (a.typeof (Int))

	@out
		true
		true
		false
*/
qu_boolean qu__array_typeof (qu_array self, qu_any x)
{
	if (self->type)
	{
		return qu_object_compat_type (self->type, x) ? QuTrue : QuFalse;
	}
	else if (self->len == 0)
	{
		return QuFalse;
	}
	else
	{
		long i;
		for (i = 0; i < self->len; i++)
		{
			if (!qu_object_validate (self, x, self->objs[i]))
				return QuFalse;
		}
		return QuTrue;
	}
}



/*) sub unique -> self
	Remove duplicate items.

	@exc
	.EModify

	@xmp
		x = [1, 2, 2, 4, 3, 4, 5, 7]
		x.unique
		println (x)

	@out
		[1, 2, 4, 3, 5, 7]
*/
static qu_array qu__array_unique (qu_array self)
{
	qu_set s;
	long   i, j;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	s = qu_array_as_set (self);
	QU_ENSURE (s);
	for (i = j = 0; i < self->len; ++i)
	{
		qu_any x = self->objs[i];
		if (s->len && qu_set_contains (s, x))
		{
			if (j >= self->len)
				break;
			self->objs[j++] = x;
			qu_set_del (s, x);
		}
	}
	self->len = j;
	return self;
}



/*-----------------------------------------------------------------------------
	interface
-----------------------------------------------------------------------------*/

/*
	Create a regular non-GC'ed qu_array for argument flattening.
*/
qu_array qu_array_flatten (int argc, qu_any *argv)
{
	int       m = 0;
	int       n = 0;
	int       j = 0;
	qu_any   *g = 0;
	qu_array  a = qu_array_new (argc);
	qu_array  f;
	qu_any    x;

	for (;;)
	{
		if (j < n)
		{
			x = g [j++];
		}
		else
		{
			if (m >= argc)
				break;

			x = argv[m++];
			if (qu_var_is (x))
			{
				assert (qu_var_is_flattener (x));
				f = (qu_array) ((qu_var) x)->value;
				assert (qu_array_is (f));
				n = f->len;
				j = 0;
				g = f->objs;
				continue;
			}
		}
		qu_array_push_quick (a, x);
	}
	return a;
}



/*
	To make sure the qu_array has enough room for 'more' items
	for a quick and dirty set.
*/
int qu_array_make_space (qu_array self, long more)
{
	if (self->cap - self->len >= more)
		return 1;
	QU_MEM_CHECK (self->len <= QU_GC_MAX_ITEMS - more);
	_array_resize (self, self->len + more);
	return 1;
}



/*
	Get a slice without throwing.
*/
qu_array qu_array_getslice (qu_array self, long i, long n)
{
	long     l = self->len;
	qu_array r;

	if (i < 0)
		i += l;
	if (i < 0 || i > l)
		return 0;
	if (n < 0 || n > l - i)
		n = l - i;
	r = qu_array_new (n);
	QU_OBJECT_LIST_COPY (r->objs, self->objs + i, n);
	r->len = n;
	return r;
}



/*
	Quick and dirty delete.
*/
void qu_array_del_quick (qu_array self, qu_any x)
{
	long    l = self->len;
	qu_any *p = self->objs;

	while (l--)
	{
		if (x == *p)
		{
			QU_OBJECT_LIST_COPY (p, p + 1, l);
			return;
		}
		++p;
	}
}



/*
	Quick and dirty push.
*/
void qu_array_push_quick (qu_array self, qu_any x)
{
	assert (self->len <= self->cap);
	if (self->len == self->cap)
	{
		assert (self->cap < QU_GC_MAX_ITEMS);
		_array_resize (self, self->cap + 1);
	}
	self->objs[self->len++] = x;
}




/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_array__setup ()
{
	qu_class_setup (QuArrayClass, qu_array_s, qu__array__destroy, qu__array__mark);

	QuArrayEmpty = qu_array_new_static (0);
	qu_array_freeze (QuArrayEmpty);
}



void qu_array__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuArrayClass, QU_NAME__inst, qu__array__inst, -1);
	qu_class_method_class  (QuArrayClass, QU_NAME_new  , qu__array_new  , -1);

	/* specific class methods */
	qu_class_method_classz (QuArrayClass, "dim", qu__array_dim, -1);

	/* standard instance methods */
	qu_class_method_inst  (QuArrayClass, QU_NAME__as      , qu__array__as      , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__clear   , qu__array__clear   , 0);
	qu_class_alias        (QuArrayClass, QU_NAME__clear   , QU_NAME_clear);
	qu_class_method_inst  (QuArrayClass, QU_NAME__cmp     , qu__array__cmp     , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__copy    , qu__array__copy    , 0);
	qu_class_method_inst  (QuArrayClass, QU_NAME__del     , qu__array__del     , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__each    , qu__array__each    , 0);
	qu_class_method_inst  (QuArrayClass, QU_NAME__getitem , qu__array__getitem , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__getslice, qu__array__getslice, 2);
	qu_class_method_inst  (QuArrayClass, QU_NAME__iand    , qu__array__iand    , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__in      , qu__array__in      , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__ior     , qu__array__ior     , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__ixor    , qu__array__ixor    , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__like    , qu__array__like    , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__lshift  , qu__array__lshift  , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__minus   , qu__array__minus   , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__mul     , qu__array__mul     , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__of      , qu__array__of      , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__plus    , qu__array__plus    , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__print   , qu__array__print   , 0);
	qu_class_method_inst  (QuArrayClass, QU_NAME__pop     , qu__array__pop     , 0);
	qu_class_method_inst  (QuArrayClass, QU_NAME__push    , qu__array__push    , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__rshift  , qu__array__rshift  , 1);
	qu_class_method_inst  (QuArrayClass, QU_NAME__setitem , qu__array__setitem , 2);
	qu_class_method_inst  (QuArrayClass, QU_NAME__setslice, qu__array__setslice, 3);
	qu_class_method_inst  (QuArrayClass, QU_NAME__string  , qu__array__string  , 0);

	/* specific instance methods */
	qu_class_method_instz (QuArrayClass, "chop"   , qu__array_chop   , -1);
	qu_class_method_instz (QuArrayClass, "copy"   , qu__array_copy   , -1);
	qu_class_method_instz (QuArrayClass, "delete" , qu__array_delete , -1);
	qu_class_method_instz (QuArrayClass, "grow"   , qu__array_grow   , -1);
	qu_class_method_instz (QuArrayClass, "head"   , qu__array_head   , -1);
	qu_class_method_instz (QuArrayClass, "index"  , qu__array_index  , -1);
	qu_class_method_instz (QuArrayClass, "inject" , qu__array_inject , -1);
	qu_class_aliasz       (QuArrayClass, "inject" , "unshift");
	qu_class_method_instz (QuArrayClass, "insert" , qu__array_insert , -1);
	qu_class_method_instz (QuArrayClass, "join"   , qu__array_join   , -1);
	qu_class_method_instz (QuArrayClass, "pop"    , qu__array_pop    , -1);
	qu_class_method_instz (QuArrayClass, "popr"   , qu__array_popr   , -1);
	qu_class_method_instz (QuArrayClass, "push"   , qu__array_push   , -1);
	qu_class_method_instz (QuArrayClass, "replace", qu__array_replace, -1);
	qu_class_method_instz (QuArrayClass, "reverse", qu__array_reverse,  0);
	qu_class_method_instz (QuArrayClass, "rindex" , qu__array_rindex , -1);
	qu_class_method_instz (QuArrayClass, "select" , qu__array_select , -1);
	qu_class_method_instz (QuArrayClass, "shift"  , qu__array_shift  , -1);
	qu_class_method_instz (QuArrayClass, "shrink" , qu__array_shrink ,  1);
	qu_class_method_instz (QuArrayClass, "sort"   , qu__array_sort   , -1);
	qu_class_method_instz (QuArrayClass, "split"  , qu__array_split  , -1);
	qu_class_method_instz (QuArrayClass, "tail"   , qu__array_tail   , -1);
	qu_class_method_instz (QuArrayClass, "type"   , qu__array_type   ,  0);
	qu_class_method_instz (QuArrayClass, "typeof" , qu__array_typeof ,  1);
	qu_class_method_instz (QuArrayClass, "unique" , qu__array_unique ,  0);

	/* finalize */
	qu_class_set_super (QuArrayClass, QuContainerClass);
	qu_class_enddecl   (QuArrayClass);
}
