/* qu_ast.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Ast __final
	A Ast object represents a node in a parsed source code tree.

	Ast stands for "a structured tree" because it is used loosely, for an
	"abstract syntax tree", "code tree", and whatever else it is suited for.
	The name is choosen because the first three letters offer a good degree
	of uniqueness.

	Ast is generated per module, not individually for Class/Sub objects.
	Unless you run the program with the C{-p} option, Ast objects are destroyed
	once compilation is successful.

	Note also that what you get is a tree after symbol resolution and perhaps
	a number of optimizations. Currently it is not possible to get the original
	parse tree nor any other intermmediate trees.

	@warn
	Don't forget to run the examples with the C{-p} option otherwise you will
	get a ENone exception.

	@xmp
		use Time
		sub foo (x)
			return ++x + 1
		;;
		println (__main__.__ast)

	@out
		\_0: module <Module __main__>
		. \_1: import <Module /Time>
		. \_0: nop
		. \_0: compound
		. . \_2: sub <Sub __main__.foo>
		. . . \_0: compound
		. . . . \_0: stmt
		. . . . . \_3: return
		. . . . . . \_3: binop op:+
		. . . . . . . \_3: unop op:++
		. . . . . . . . \_3: name <Var x>
		. . . . . . . \_3: literal 1
		. . \_0: stmt
		. . . \_5: call op:1
		. . . . \_5: name <Sub Sys.println> op:1
		. . . . \_5: getattr
		. . . . . \_5: name <Sub Sys.__main__>
		. . . . . \_5: literal __ast
*/
qu_class QuAstClass;


/* qu_ast names */
static qu_string AstNames [QU_AST_LAST + 1];


/* Flags */
#define QU_AST_FLAG_P QU_FLAG_0 /* production node? */


/*----------------------------------------------------------------------------
	stuffs
----------------------------------------------------------------------------*/

/*
	Get the name of the function of a binary operator.
	This is used to get a better AST representation.
*/
static const char *_ast_get_binop_name (qu_ast n)
{
	void *f = (void *) n->op;

#define n_(name, strz) (f == (void *) qu_sys_##name) return strz;
	if      n_(cmp    , "<=>")
	else if n_(div    , "/")
	else if n_(eq     , "==")
	else if n_(ge     , ">=")
	else if n_(gt     , ">")
	else if n_(iand   , "&")
	else if n_(idiv   , "\\")
	else if n_(ior    , "|")
	else if n_(ixor   , "^")
	else if n_(in     , "in")
	else if n_(is     , "is")
	else if n_(le     , "<=")
	else if n_(lshift , "<<")
	else if n_(like   , "===")
	else if n_(lt     , "<")
	else if n_(ne     , "!=")
	else if n_(minus  , "-")
	else if n_(mul    , "*")
	else if n_(plus   , "+")
	else if n_(pow    , "**")
	else if n_(ratio  , "\\_")
	else if n_(rem    , "%")
	else if n_(rin    , "rin")
	else if n_(rshift , ">>")
	else if n_(concat , "_")
	else if n_(range  , "..")
	else return "binop";
#undef n_
}



/*
	Get the name of the function of a unary operator.
	This is used to get a better AST representation.
*/
static const char *_ast_get_unop_name (qu_ast n)
{
	void *f = (void *) n->op;

#define n_(name, strz) (f == (void *) qu_sys_##name) return strz;
	if      n_(decr, "--")
	else if n_(incr, "++")
	else if n_(inot, "~")
	else if n_(neg , "(-)")
	else if n_(not , "!")
	else if n_(pos , "(+)")
	else if n_(true, "true")
	else return "unop";
#undef n_
}



/*----------------------------------------------------------------------------
	specific class properties
----------------------------------------------------------------------------*/

/*) const LIST = nil # set by the system
	This is a frozen -{Dict} containing all Ast names and types.

	The keys are the #{name} and the values are the #{type}.

	@xmp
		n = 0
		for k, v in Ast.LIST
			break unless ++n < 5
			println (v ': ' k)
		;;

	@out
		61: unop
		15: compound
		23: flat
		33: literal
*/


/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__ast__mark (qu_ast self)
{
	qu_gc_xmark (self->thing);
	qu_gc_xmark (self->next);
	qu_gc_xmark (self->child);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

static long _ast_get_lineno (qu_ast n)
{
	long l = n->line;

	if (l == 0)
	{
		if (n->child)
			return _ast_get_lineno (n->child);
	}
	return l;
}



static qu_null qu__ast__print_do (qu_ast n, int level)
{
	char      b [128];
	qu_string s;
	int       k = level;

	while (k--)
		QU_ENSURE (qu_sys_printz (". ", 2));

	s = qu_ast_name (n->type);
	sprintf (b, "\\_%ld: %.48s", _ast_get_lineno (n), s->str);

	QU_ENSURE (qu_sys_printz (b, strlen (b)));

	if (n->thing)
	{
		QU_ENSURE (qu_sys_printz (" ", 1));
		QU_ENSURE (qu_code_print_literal (n->thing));
	}

	if (n->op)
	{
		switch (n->type)
		{
			case QU_AST_binop:
				sprintf (b, " op:%s", _ast_get_binop_name (n));
				break;

			case QU_AST_unop:
				sprintf (b, " op:%s", _ast_get_unop_name (n));
				break;

			default:
				sprintf (b, " op:%ld", n->op);
		}
		QU_ENSURE (qu_sys_printz (b, strlen (b)));
	}
	QU_ENSURE (qu_sys_printz ("\n", 1));

	for (n = n->child; n != NULL; n = n->next)
		qu__ast__print_do (n, level + 1);

	return QuNil;
}



/*) sub __print
	Print this Ast object to v{stdout}.

	See -{print} for possible exceptions.
*/
static qu_null qu__ast__print (qu_ast self)
{
	return qu__ast__print_do (self, 0);
}



/*) sub __string
	Return a String representation.

	@noexc
*/
static qu_string qu__ast__string (qu_ast self)
{
 	qu_string s = qu_ast_name (self->type);
	char buf [1024];

	snprintf (buf, sizeof (buf), "<Ast %s>", s->str);
	return qu_string_newz (buf);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub child -> Any
	Return the child Ast of this Ast. If the

	@exc
	.ENone

	@xmp
		use Time
		println (__main__.__ast.child)

	@out
		\_1: import <Module /Time>
*/
static qu_ast qu__ast_child (qu_ast self)
{
	QU_REQUIREZ (QuENone, self->child, "Ast does not have a child");
	return self->child;
}



/*) sub line -> Int
	Return the line number where this Ast is in the source code.

	If the Ast does not have a line number information then C{0} is returned.

	@xmp
		use Time
		println (__main__.__ast.child.line)

	@out
		1
*/
static qu_int qu__ast_line (qu_ast self)
{
	assert (self->line <= QU_INT_MAX);
	return qu_int_new (self->line);
}



/*) sub name -> String
	Return the name of this Ast.

	@xmp
		use Time
		println (__main__.__ast.child.name)

	@out
		import
*/
static qu_string qu__ast_name (qu_ast self)
{
	return qu_ast_name (self->type);
}



/*) sub next -> Ast
	Return the sibling Ast of this Ast.

	@exc
	.ENone

	@xmp
		x = 1
		println (__main__.__ast.child.next.name)

	@out
		compound
*/
static qu_ast qu__ast_next (qu_ast self)
{
	QU_REQUIREZ (QuENone, self->next, "Ast does not have a sibling");
	return self->next;
}



/*) sub op -> Int
	Return an Int (a value, flag, option, etc.) associated with this Ast.

	@noexc

	@xmp
		x = 1
		println (__main__.__ast.child.op)

	@out
		0
*/
static qu_int qu__ast_op (qu_ast self)
{
	return qu_int_new (self->op);
}



/*) sub thing -> Any
	Return the object associated with this Ast.

	@exc
	.ENone

	@xmp
		use Time
		println (__main__.__ast.child.thing)

	@out
		<Module /Time>
*/
static qu_any qu__ast_thing (qu_ast self)
{
	QU_REQUIREZ (QuENone, self->thing, "Ast has no associated object");
	return self->thing;
}



/*) sub type -> Int
	Return the type.

	@xmp
		use Time
		println (__main__.__ast.child.type)

	@out
		30
*/
static qu_int qu__ast_type (qu_ast self)
{
	return qu_int_new (self->type);
}



/*----------------------------------------------------------------------------
	Interface
----------------------------------------------------------------------------*/

static char *RuleBug = "<bug: Ast R-node used for error message>";

qu_ast qu_ast_newr (int type)
{
	register qu_ast n;

	n         = qu_gc_new (QuAstClass);
	n->next   = n->child = 0;
	n->type   = type;
	n->cursor = RuleBug;
	n->line   = 0;
	n->thing  = 0;
	n->op     = 0;
	return n;
}



qu_ast qu_ast_newp (int type, const char *cursor, long line)
{
	register qu_ast n;

	n          = qu_gc_new (QuAstClass);
	n->flags  |= QU_AST_FLAG_P;
	n->next    = n->child = 0;
	n->type    = type;
	n->cursor  = cursor;
	n->line    = line;
	n->thing   = 0;
	n->op      = 0;
	return n;
}



qu_string qu_ast_name (int type)
{
	if (0 < type && type <= QU_AST_LAST)
		return AstNames [type];
	return AstNames [0];
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_ast__setup ()
{
	qu_class_setup_sys (QuAstClass, qu_ast_s, NULL, qu__ast__mark);
}



void qu_ast__init ()
{
	qu_dict d = qu_dict_new_static (64);

#define n_(s) \
	AstNames [QU_AST_##s] = qu_string_new_staticz (#s); \
	qu_dict_setitem (d, AstNames [QU_AST_##s], qu_int_new (QU_AST_##s));

#include "Asts.h"

#undef n_

	qu_dict_freeze (d);

	/* specific class properties */
	qu_class_constz (QuAstClass, "LIST", d);

	/* standard instance methods */
	qu_class_method_inst  (QuAstClass, QU_NAME__print , qu__ast__print , 0);
	qu_class_method_inst  (QuAstClass, QU_NAME__string, qu__ast__string, 0);

	/* specific instance methods */
	qu_class_method_instz (QuAstClass, "child", qu__ast_child, 0);
	qu_class_method_instz (QuAstClass, "line" , qu__ast_line , 0);
	qu_class_method_instz (QuAstClass, "name" , qu__ast_name , 0);
	qu_class_method_instz (QuAstClass, "next" , qu__ast_next , 0);
	qu_class_method_instz (QuAstClass, "op"   , qu__ast_op   , 0);
	qu_class_method_instz (QuAstClass, "thing", qu__ast_thing, 0);
	qu_class_method_instz (QuAstClass, "type" , qu__ast_type , 0);

	/* finalize declaration */
	qu_class_enddecl (QuAstClass);
}
