/* Big.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Big is Decimal __final
	A Big object is a decimal floating point number which can be as big as the
	available memory (with a few limits, of course), contrary to -{Float} which
	has a limited range.

	Refer to the -{Number} class for information on Qu numerics in general.

	The class hierarchy:
	@tree
	.Number
	.	-{Decimal}
	.		Big

	@p
	Generally you should make no distinction between a Float and a Big in
	your applications. The only reason why Float exists is performance.
	Once computers are very fast, Float and Big will be merged into -{Decimal}.

	Big numbers are always normalized, that is, they are internally represented
	as: U{+}C{0.dddd... * base ** exponent} where v{base} is 10000.
	See #{BASE}.

	The digits forming the mantissa are simply translated from their ASCII
	representation in a group of 4 with MSB (most significant bit) first.
	Thus the exponent value is a power of 10 and a multiplication of 4.
	See #{FIG}.

	This means that when you translate a decimal string into a Big, there is
	no digit lost.

	@xmp
		b = Big ('1234567890.1234')
		println (b)
		println (b as Array ' E ' b.expo)

	@out
		0.0012345678901234E+12
		[1234, 7890, 3456, 12] E 3

	@p
	The Big class provides a number of methods you can use to manipulate
	Big numbers, e.g. ={expo}, ={frac}, ={precision}, etc. They might be
	handy in some situations.

	The ={new} method provides a number of ways to create Big numbers.
	You can translate -{Float} and -{Rational} numbers, for example, or build
	your own number by providing the digits and the exponent.

	The following example might give you an idea on why you want to use Big
	instead of Float.

	@xmp
		# the correct answer is 1

		# C program
		printf ("%.18f\n", fmod (1e40, 11.0));

		# Qu
		printfln ("%.18f", 1e40 % 11)
		println (Big ('1e40') % 11)
		println (10000000000000000000000000000000000000000 % 11)

	@out
		0.000000000000000000
		-2.000000000000000000
		0.0001E+4
		1

	@note
	In Qu it is always possible to convert a Big into a Long and vice versa
	without throwing an exception, provided of course, there is enough memory.
*/
qu_class QuBigClass;

/*) const MINPREC = 2
	This the minimum precision value allowed.

	Note that the actual value is set by the system.
	A value of 2 means you can not calculate Big numbers using less than
	8 significant digits.
*/

/*) const MAXPREC = 67108863
	This the maximum precision value allowed.

	Note that the actual value is set by the system and is normally set
	to the maximum practical allocatable memory for the digits.

	A value of 100000 means you can not calculate Big numbers using more than
	400000 significant digits.
*/

/*) const DIGITSIZE = 4
	This is the number of bytes per digit.

	Note that the actual value is set by the system but is generally 4 on
	32-bits machines (it is a C unsigned long).
*/

/*) const BASE = 10000
	This is the base value used.
	There should be no reason why this value is not 10000.
*/

/*) const FIG = 4
	This is the number of ASCII characters in one digit.
	The value is always log10 (#{BASE}).
*/

/*) const MAXEXPO = 8388607
	This is the maximum possible absolute value of the exponent.
	It fits in and -{Int} object.

	The value is a multiplication of #{FIG}.
*/

/*) const MAXLEN = 134217727
	This is the maximum number of digits that can be practically allocated.

	Note that the actual value is set by the system according to memory
	limits. It is normally bigger than #{MAXPREC}.
*/


/* set and unset sign */
#define SETNEG(self)  qu_flag_set   (self, QU_BIG_FLAG_NEG)
#define SETPOS(self)  qu_flag_unset (self, QU_BIG_FLAG_NEG)

#define SETSIGN(self, sign) \
QU_BEGIN \
	if ((sign) >= 0) \
		SETPOS (self); \
	else \
		SETNEG (self); \
QU_END

/* current precision - in terms of QU_BIG_FIG thus 8 means 32 digits */
long __QuBigPrecision = 8;

/* predefined */
qu_big QuBig0;
qu_big QuBig1;
qu_big QuBig2;
qu_big QuBig3;
qu_big QuBig4;
qu_big QuBig5;
qu_big QuBig6;
qu_big QuBig7;
qu_big QuBig8;
qu_big QuBig9;
qu_big QuBig10;
qu_big QuBig16;
qu_big QuBig_1;
qu_big QuBig_2;
qu_big QuBig_4;
qu_big QuBig_8;
qu_big QuBig1_2;
qu_big QuBig_1_2;
qu_big QuBig1_4;
qu_big QuBig_1_4;


/* how many digits in a long? */
#if SIZEOF_LONG == 4
#define QU_BIG_FIG_LONG 3
#elif SIZEOF_LONG == 8
#define QU_BIG_FIG_LONG 5
#else
#define QU_BIG_FIG_LONG 20 /* should be enough */
#endif

/* how many digits in a double? */
#define QU_BIG_FIG_DOUBLE 20

/* to reduce polution */
#define DIGIT qu_big_digit_t

/* to make a temporary qu_big from a C long */
typedef struct TBigTmpl TBigTmpl;
struct TBigTmpl
{
	QU_OBJECT_HEAD;
	DIGIT *digits;
	long   len;
	long   expo;
	long   cap;
	DIGIT  them [QU_BIG_FIG_LONG];
};


/* to make a temporary qu_big from a C double */
typedef struct TBigTmpd TBigTmpd;
struct TBigTmpd
{
	QU_OBJECT_HEAD;
	DIGIT *digits;
	long   len;
	long   expo;
	long   cap;
	DIGIT  them [QU_BIG_FIG_DOUBLE];
};


/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

/*
	allocate a new qu_big
	- does not zero the digits
	- len = 0, expo = 0
*/
static qu_big _big_alloc (long cap)
{
	register qu_big g;

	/* not necessary
	assert (0 < cap && cap <= QU_BIG_MAXLEN); */

	g         = qu_gc_new (QuBigClass);
	g->cap    = cap++;
	g->len    = 0;
	g->expo   = 0;
	g->digits = qu_mem_calloc (cap, sizeof (DIGIT));
	return g;
}



/*
	-ditto- static
*/
static qu_big _big_alloc_static (long cap)
{
	register qu_big g;

	assert (0 < cap && cap <= QU_BIG_MAXLEN);

	g         = qu_gc_new_static (QuBigClass);
	g->cap    = cap++;
	g->len    = 0;
	g->expo   = 0;
	g->digits = qu_mem_static (cap * sizeof (DIGIT));

	memset (g->digits, 0, cap * sizeof (DIGIT));
	return g;
}



/*
	return a copy of a qu_big
*/
static qu_big qu_big_dup (qu_big self)
{
	register qu_big g;

	assert (self->len > 0);

	g         = qu_gc_new (QuBigClass);
	g->flags  = self->flags;

	qu_flag_unset (self, QU_FLAG_STATIC);

	g->cap    = self->len;
	g->len    = self->len;
	g->expo   = self->expo;
	g->digits = qu_mem_calloc (self->len + 1, sizeof (DIGIT));

	memcpy (g->digits, self->digits, self->len * sizeof (DIGIT));
	return g;
}



/*----------------------------------------------------------------------------
	temporary
----------------------------------------------------------------------------*/

/* !!! the result is not normalized */
static qu_big _big_tmp_clong (TBigTmpl *t, long g)
{
	DIGIT x, v, lo, hi, *d;
	long  e = 0, n = 0, l;

	d = t->digits = t->them;
	t->cap = QU_BIG_FIG_LONG;

	if (g == 0)
	{
		t->len = 0;
		t->flags = QU_BIG_FLAG_TMP;
		t->expo = 0;
		t->klass = QuBigClass;
		d[0] = 0;
		return (qu_big) t;
	}

	if (g < 0)
	{
		t->flags = QU_BIG_FLAG_NEG | QU_BIG_FLAG_TMP;
		g = -g;
	}
	else
	{
		t->flags = QU_BIG_FLAG_TMP;
	}

	x = (DIGIT) g;
	while (x > 0)
	{
		if (x)
		{
			hi = 1;
			lo = x;
			while (lo >= QU_BIG_BASE)
			{
				lo /= QU_BIG_BASE;
				hi *= QU_BIG_BASE;
			}
			x = x - hi * lo;
			v = lo;
		}
		else
		{
			v = 0;
		}
		*d++ = v;
		e++;
		n++;
	}
	l = n;
	while (l < QU_BIG_FIG_LONG)
	{
		l++;
		*d++ = 0;
	}
	t->expo = e;
	t->len = n;
	t->klass = QuBigClass;
	return (qu_big) t;
}



/* !!! the result is not normalized */
static qu_big _big_tmp_cdouble (TBigTmpd *t, double x)
{
	DIGIT *d, c;
	long   l = QU_BIG_FIG_DOUBLE, e = 0, n = 0, i;
	double v, w;

	assert (qu_double_isfinite (x));

	d = t->digits = t->them;
	t->cap = QU_BIG_FIG_DOUBLE;

	if (x == 0.0)
	{
		t->len = 0;
		t->flags = QU_BIG_FLAG_TMP;
		t->expo = 0;
		t->klass = QuBigClass;
		d[0] = 0;
		return (qu_big) t;
	}

	if (x < 0)
	{
		v = -x;
		t->flags = QU_BIG_FLAG_TMP | QU_BIG_FLAG_NEG;
	}
	else
	{
		v = x;
		t->flags = QU_BIG_FLAG_TMP;
	}

	if (v >= 1.0)
	{
		while (v >= 1.0)
		{
			v /= (double) QU_BIG_BASE;
			e++;
		}
	}
	else
	{
		w = 1.0 / (double) QU_BIG_BASE;
		while (v < w)
		{
			v *= (double) QU_BIG_BASE;
			--e;
		}
	}
	for (i = 0; v > 0.0 && i < l; i++, n++)
	{
		v *= (double) QU_BIG_BASE;
		c = (DIGIT) v;
		v -= (double) c;
		d[i] = c;
	}
	t->expo = e;
	t->len = n;
	t->klass = QuBigClass;
	return (qu_big) t;
}



/*----------------------------------------------------------------------------
	normalization
----------------------------------------------------------------------------*/

/*
	normalize by removing leading and trailing 0s
	00xxxxx0000 E4 => xxxxx E2

	!!! return QuBig0 if there are no significant digits
	!!! chop digits if there more digits than __QuBigPrecision
*/
static qu_big _big_normalize (qu_big self, long cut)
{
	long   l = self->len;
	DIGIT *d = self->digits;
	long   e;

	while (l && *d == 0)
	{
		l--;
		d++;
	}

	e = self->len - l;

	d += l;
	while (l && *--d == 0)
		l--;

	if (l == 0)
		return QuBig0;

	if (cut && l > __QuBigPrecision)
		l = __QuBigPrecision;

	if (l != self->len)
	{
		memcpy (self->digits, self->digits + e, l * sizeof (DIGIT));
#ifdef QU_DEBUG_LOCAL
		memset (self->digits + l, 0, (self->cap - l) * sizeof (DIGIT));
#endif
		self->len = l;
	}
	self->expo -= e;
	return self;
}



/*
	always use this to return the result to program level
	it checks for limits

	!!! must call _big_normalize before calling this function
*/
static qu_real _big_return (qu_big self)
{
	if (self)
	{
		if (self->expo < 0)
		{
			if (-self->expo > QU_BIG_MAXEXPO)
				return __qu_tiny__ ((qu_real) QuBig0);
		}
		else
		{
			if (self->expo > QU_BIG_MAXEXPO)
				return (qu_real) __qu_huge__ (qu_big_sign (self));
		}

		/* we don't want to sign 0 ... just in case */
		if (self->len == 0)
		{
			assert (!qu_big_isneg (self));
			return (qu_real) QuBig0;
		}

		if (qu_flag_test (self, QU_BIG_FLAG_TMP))
			return (qu_real) qu_big_dup (self);
	}
	return (qu_real) self;
}



/*----------------------------------------------------------------------------
	basics
----------------------------------------------------------------------------*/

/*
	add two qu_big ignoring the sign
	- require a >= b
	- do not call if 'a' or 'b' is 0
	- return a new qu_big or possibly 'a'
	- does not set sign
*/
static qu_big _big_basic_plus (qu_big a, qu_big b)
{
	long    e, la, lb, lc, i, j, k;
	DIGIT  *da, *db, *dc, carry;
	qu_big  c;

	assert (qu_big_cmpabs_big (a, b) >= 0);

	da = a->digits;
	db = b->digits;
	la = a->len;
	lb = b->len;

	assert (la > 0);
	assert (lb > 0);

	e = a->expo - b->expo;
	assert (e >= 0);

	if (lb + e >= la)
	{
		lc = lb + e;
		if (lc > __QuBigPrecision)
			lc = __QuBigPrecision;

		if (lc <= e)
		{
			/*
			| aa
			|      bbbbbbbbbbbbbbb
			| ccccc
			*/
			return a;
		}
		else if (la < lc)
		{
			if (e + lb >= lc)
			{
				/*
				| aa
				|  i
				|     bbbb
				|      j
				| cccccc
				*/
				i  = la - 1;
				j  = lc - 1 - e;
			}
			else
			{
				/*
				| aa
				|  i
				|     bb
				|      j
				| cccccccc
				*/
				i  = la - 1;
				j  = lb - 1;
			}
		}
		else
		{
			/*
			| aaaaaaaaaaa
			|       i
			|      bbbbbbbbb
			|       j
			| cccccccc
			*/
			i = lc - 1;
			j = lc - e - 1;
		}
	}
	else
	{
		lc = la;
		if (lc > __QuBigPrecision)
			lc = __QuBigPrecision;

		if (lc <= e)
		{
			/*
			| aaaaaaaaaaaaaa
			|      bbbbb
			| ccccc
			*/
			c       = _big_alloc (lc);
			c->len  = lc;
			c->expo = a->expo;

			memcpy (c->digits, a->digits, lc * sizeof (DIGIT));
			return c;
		}
		else if (lc < la)
		{
			if (lc >= e + lb)
			{
				/*
				| aaaaaaaaaaaaaaaaa
				|            i
				|      bbbbb
				|          j
				| cccccccccccc
				*/
				i = lc - 1;
				j = lb - 1;
			}
			else
			{
				/*
				| aaaaaaaaaaaaaaaaa
				|       i
				|      bbbbb
				|       j
				| ccccccc
				*/
				i = lc - 1;
				j = i - e;
			}
		}
		else
		{
			/*
			| aaaaaaaaaaaaaaaaa
			|      bbbbb
			| ccccccccccccccccc
			*/
			assert (lc == la);
			i = lc - 1;
			j = lb - 1;
		}
	}

	c       = _big_alloc (lc + 1);
	c->len  = lc + 1;
	c->expo = a->expo + 1;

	dc = c->digits;
	/* memset (dc, 0, (lc + 1) * sizeof (DIGIT)); */

	assert (0 <= j && j < lb);
	assert (0 <= i && i < la);

	for (k = e + j + 1; j >= 0; j--, k--)
	{
		assert (0 < k && k <= lc);

		dc[k] = db[j];
	}

	carry = 0;
	for (k = i + 1; i >= 0; i--, k--)
	{
		assert (0 <= i && i < la);
		assert (0 < k  && k <= lc);

		dc[k] += da[i] + carry;
		if (dc[k] >= QU_BIG_BASE)
		{
			dc[k] -= QU_BIG_BASE;
			carry = 1;
		}
		else
		{
			carry = 0;
		}
	}
	while (carry)
	{
		assert (0 <= k && k < lc);

		dc[k] += carry;
		if (dc[k] >= QU_BIG_BASE)
		{
			dc[k--] -= QU_BIG_BASE;
			carry = 1;
			continue;
		}
		break;
	}
	return _big_normalize (c, 1);
}



/*
	subtract two Bigs ignoring the sign
	- require a >= b
	- do not call if 'a' or 'b' is 0
	- return a new qu_big or possibly 'a'
	- does not set sign
*/
static qu_big _big_basic_minus (qu_big a, qu_big b)
{
	long    e, la, lb, lc, i, j;
	DIGIT  *da, *db, *dc, borrow;
	qu_big  c;

	assert (qu_big_cmpabs_big (a, b) >= 0);

	da = a->digits;
	db = b->digits;
	la = a->len;
	lb = b->len;

	assert (la > 0);
	assert (lb > 0);

	e = a->expo - b->expo;
	assert (e >= 0);

	if (lb + e >= la)
	{
		lc = lb + e;
		if (lc > __QuBigPrecision)
			lc = __QuBigPrecision;

		if (lc <= e)
		{
			/*
			| aa
			|      bbbbbbbbbbbbbbb
			| ccccc
			*/
			return a;
		}
		else if (la < lc)
		{
			if (e + lb >= lc)
			{
				/*
				| aa
				|  i
				|     bbbb
				|      j
				| cccccc
				*/
				i  = la - 1;
				j  = lc - 1 - e;
			}
			else
			{
				/*
				| aa
				|  i
				|     bb
				|      j
				| cccccccc
				*/
				i  = la - 1;
				j  = lb - 1;
			}
		}
		else
		{
			/*
			| aaaaaaaaaaa
			|       i
			|      bbbbbbbbb
			|       j
			| cccccccc
			*/
			i = lc - 1;
			j = lc - e - 1;
		}
	}
	else
	{
		lc = la;
		if (lc > __QuBigPrecision)
			lc = __QuBigPrecision;

		if (lc <= e)
		{
			/*
			| aaaaaaaaaaaaaa
			|      bbbbb
			| ccccc
			*/
			c       = _big_alloc (lc);
			c->len  = lc;
			c->expo = a->expo;

			memcpy (c->digits, a->digits, lc * sizeof (DIGIT));
			return c;
		}
		else if (lc < la)
		{
			if (lc >= e + lb)
			{
				/*
				| aaaaaaaaaaaaaaaaa
				|            i
				|      bbbbb
				|          j
				| cccccccccccc
				*/
				i = lc - 1;
				j = lb - 1;
			}
			else
			{
				/*
				| aaaaaaaaaaaaaaaaa
				|       i
				|      bbbbb
				|       j
				| ccccccc
				*/
				i = lc - 1;
				j = i - e;
			}
		}
		else
		{
			/*
			| aaaaaaaaaaaaaaaaa
			|      bbbbb
			| ccccccccccccccccc
			*/
			assert (lc == la);
			i = lc - 1;
			j = lb - 1;
		}
	}
	c       = _big_alloc (lc);
	c->len  = lc;
	c->expo = a->expo;

	dc = c->digits;
	/* memset (dc, 0, lc * sizeof (DIGIT)); */

	assert (0 <= j && j < lb);
	assert (0 <= i && i < la);

	for (; i >= 0; i--)
	{
		assert (0 <= i && i < lc);
		assert (0 <= i && i < la);

		dc[i] = da[i];
	}

	borrow = 0;
	for (i = e + j; j >= 0; j--, i--)
	{
		assert (0 <= j && j < lb);
		assert (0 <= i && i < lc);

		dc[i] -= db[j] + borrow;
		if (dc[i] >= QU_BIG_BASE)
		{
			dc[i] += QU_BIG_BASE;
			borrow = 1;
		}
		else
		{
			borrow = 0;
		}
	}
	while (borrow)
	{
		assert (0 <= i && i < lc);

		dc[i] -= borrow;
		if (dc[i] >= QU_BIG_BASE)
		{
			dc[i--] += QU_BIG_BASE;
			borrow = 1;
			continue;
		}
		break;
	}
	return _big_normalize (c, 1);
}



/*
	multiply two Bigs by disregarding sign
	- do not call if 'a' or 'b' is 0 or 1 or -1
	- return a new qu_big
	- does not set sign
*/
static qu_big _big_basic_mul (qu_big a, qu_big b)
{
	long    la, lb, lc, i, j, k, m, u;
	DIGIT  *da, *db, *dc, carry, v;
	qu_big  c;

	if (a->len < b->len)
	{
		qu_big t = a;
		a = b;
		b = t;
	}

	da = a->digits;
	db = b->digits;
	la = a->len;
	lb = b->len;

	assert (la > 0);
	assert (lb > 0);

	lc = la + lb + 1;
	if (lc > __QuBigPrecision)
		lc = __QuBigPrecision;

	c  = _big_alloc (lc);
	c->len = lc;
	c->expo = a->expo + b->expo + 1;
	dc = c->digits;
	/* memset (dc, 0, lc * sizeof (DIGIT)); */

	for (k = la + lb, j = lb - 1; j >= 0; j--, k--)
	{
		carry = 0;
		for (i = la - 1, m = k; i >= 0; i--, m--)
		{
			if (m < lc)
			{
				dc[m] += db[j] * da[i] + carry;
				if (dc[m] >= QU_BIG_BASE)
				{
					carry = dc[m] / QU_BIG_BASE;
					dc[m] %= QU_BIG_BASE;
				}
				else
				{
					carry = 0;
				}
			}
			else
			{
				v = db[j] * da[i] + carry;
				if (v >= QU_BIG_BASE)
				{
					carry = v / QU_BIG_BASE;
				}
				else
				{
					carry = 0;
				}
			}

			u = m;
			while (carry)
			{
				if (--u < lc)
				{
					assert (u >= 0);
					dc[u] += carry;
					if (dc[u] >= QU_BIG_BASE)
					{
						carry = dc[u] / QU_BIG_BASE;
						dc[u] %= QU_BIG_BASE;
					}
					else
					{
						carry = 0;
					}
				}
				else
				{
					v = carry;
					if (v >= QU_BIG_BASE)
					{
						carry = v / QU_BIG_BASE;
						v %= QU_BIG_BASE;
					}
					else
					{
						carry = 0;
					}
				}
			}
		}
	}
	return _big_normalize (c, 1);
}



/*
	divide two Bigs by disregarding sign
	- do not call if 'a' or 'b' is 0 or 1 or -1
	- return a new qu_big (does not return 'a' nor 'b')
	- rem is the residu, not the remainder
	- does not set sign

	Note: algorithm modified from BigDecimal
	Copyright(C) 2002 by Shigeo Kobayashi(shigeo@tinyforest.gr.jp)
*/
static qu_big _big_basic_div (qu_big a, qu_big b, qu_big *rem)
{
	qu_big  c, r;
	long    la, lb, lc, lr;
	long    i, n, ia, ib, ic, ir;
	long    nloop;
	DIGIT  *da, *db, *dc, *dr;
    DIGIT   q, b1, b1p1, b1b2, b1b2p1, r1r2;
    DIGIT   borrow, borrow1, borrow2, qb;

	assert (!qu_big_isone (b));

	da = a->digits;
	db = b->digits;
	la = a->len;
	lb = b->len;

	assert (la > 0);
	assert (lb > 0);

	lc = la;
	if (lc < lb)
		lc = lb;
	if (lc < __QuBigPrecision)
		lc = __QuBigPrecision;

	c = _big_alloc (lc);
	c->len = lc;
	c->expo = a->expo - b->expo + 2;
	dc = c->digits;
	/* memset (dc, 0, lc * sizeof (DIGIT)); */

	/* FIXME - I don't understand this algorithm, it requires a bigger lr than
		necessary and occassionally have ir >= lr at (4) */
	/*lr = lc + 1;
	if (lr < __QuBigPrecision)
		lr = __QuBigPrecision; */
	lr = lc + __QuBigPrecision;

#define CHECKRSPACE(i, where) \
	if ((i) >= lr) \
	{ \
		/* printf ("REALLOCATING (%d): %d >= lr=%d  la=%d lb=%d lc=%d\n", where, i, lr, la, lb, lc); */ \
		dr = qu_mem_calloc (lr + 1, sizeof (DIGIT)); \
		memcpy (dr, r->digits, lr * sizeof (DIGIT)); \
		lr++; \
		r->digits = dr; \
	}

	r = _big_alloc (lr + 1);
	r->len = lr;
	r->expo = a->expo + 2;
	dr = r->digits;
	/* memset (dr, 0, lr * sizeof (DIGIT)); */

	ir = 1;
	while (ir <= la)
	{
		assert (ir < lr);
		dr[ir] = da[ir - 1];
		++ir;
	}

	b1 = b1p1 = db[0];
	if (lb <= 1)
	{
		b1b2p1 = b1b2 = b1p1 * QU_BIG_BASE;
	}
	else
	{
		b1p1 = b1 + 1;
		b1b2p1 = b1b2 = b1 * QU_BIG_BASE + db[1];
		if (lb > 2)
			++b1b2p1;
	}

	n = lr - 1;
	nloop = lc;
	if (nloop > n)
		nloop = n;

	ic = 1;
	while (ic < nloop)
	{
		CHECKRSPACE (ic, 1);
		assert (ic < lr);

		if (dr[ic] == 0)
		{
			++ic;
			continue;
		}

		CHECKRSPACE (ic + 1, 2);
		assert (ic + 1 < lr);

		r1r2 = dr[ic] * QU_BIG_BASE + dr[ic + 1];
		if (r1r2 == b1b2)
		{
			ib = 2;
			ia = ic + 2;
			while (ib < lb)
			{
				assert (ib < lb);
				assert (ia < lr);

                if (dr[ia] < db[ib])
					goto div1;

                if (dr[ia] > db[ib])
					break;

                ++ia;
                ++ib;
            }

			borrow = 0;
			ib = lb - 1;
			ir = ic + ib;

			CHECKRSPACE (ir, 3);
			assert (ir < lr);

			n = ib;
			for (i = 0; i <= n; ++i, --ir, --ib)
			{
				assert (ir < lr);
				assert (ib < lb);

				if (dr[ir] < db[ib] + borrow)
				{
					dr[ir] += QU_BIG_BASE - db[ib] - borrow;
                    borrow = 1;
				}
				else
				{
					dr[ir] = dr[ir] - db[ib] - borrow;
					borrow = 0;
                }
			}
			assert (ic < lc);
			++dc[ic];
			goto carry;
		}

		if (r1r2 >= b1b2p1)
		{
			q = r1r2 / b1b2p1;
			dc[ic] += q;
			ir = lb + ic - 1;
			goto submul;
		}

	 div1:
		if (ic + 1 >= lc)
			goto done;

		q = r1r2 / b1p1;
		dc[ic + 1] += q;
		ir = lb + ic;

	  submul:

		borrow1 = borrow2 = 0;
		ib = lb - 1;
		n = ib;
		for (i = 0; i <= n; ++i)
		{
			assert (ib < lb);
			qb = q * db[ib];
			if (qb < QU_BIG_BASE)
			{
				borrow1 = 0;
			}
			else
			{
				borrow1 = qb / QU_BIG_BASE;
				qb -= borrow1 * QU_BIG_BASE;
			}

			CHECKRSPACE (ir, 4);
			assert (ir < lr);

			if (dr[ir] < qb)
			{
				dr[ir] += QU_BIG_BASE - qb;
				borrow2 = borrow2 + borrow1 + 1;
            }
			else
			{
				dr[ir] -= qb;
				borrow2 += borrow1;
			}
			--ir;
			if (borrow2)
			{
				CHECKRSPACE (ir, 5);
				assert (ir < lr);

				if (dr[ir] < borrow2)
				{
					dr[ir] += QU_BIG_BASE - borrow2;
					borrow2 = 1;
				}
				else
				{
					dr[ir] -= borrow2;
					borrow2 = 0;
				}
			}
			--ib;
		}

		CHECKRSPACE (ir, 6);
		assert (ir < lr);

		dr[ir] -= borrow2;

	  carry:

		ir = ic;

		CHECKRSPACE (ir, 7);
		assert (ir < lr);

		while (dc[ir] >= QU_BIG_BASE)
		{
			dc[ir] -= QU_BIG_BASE;
			--ir;
			assert (ir >= 0);
			++dc[ir];
		}
	}

  done:
	*rem = _big_normalize (r, 1);
	return _big_normalize (c, 1);

}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (*arg) -> Big
	@def
	sub class new (x) -> Big
	sub class new (digits: Array, len = digits.__len, expo = 0) -> Big
	@p
	Create a new Big object from another object.

	The following objects are supported:
	@ul
	.Number (Int, Float, Long, Rational, Big, Complex)
	.String
	.Array

	@p
	If you pass an Array then you can optionally set the exponent.
	The items in the Array must be Integer numbers with value between
	0 to 9999. The exponent must be in the range #{Int.MIN} and #{Int.MAX}.
	The first item is the MSB (most significant digit).

	If you pass a String then the ={precision} setting is ignored, that is,
	there is no rounding.

	@exc
	.EArgValue

	@xmp
		b = Big ([1230 1456].reverse, -3)
		println (b)

		b = Big ('12301456e-3')
		println (b)

	@out
		0.12301456E-12
		0.000123014560E+8
*/
static qu_big qu__big_new (qu_class self, int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VMIN (1);
	x = argv[0];

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_big_new_int      ((qu_int     ) x);
		case QU_CLASS_ID_Float:    return qu_big_new_float    ((qu_float   ) x);
		case QU_CLASS_ID_Long:     return qu_big_new_long     ((qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_big_new_rational ((qu_rational) x);
		case QU_CLASS_ID_Complex:  return qu_big_new_complex  ((qu_complex ) x);
		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Buffer:   return qu_big_new_string   ((qu_string  ) x);
		case QU_CLASS_ID_Big:      return (qu_big) x;

		case QU_CLASS_ID_Array:
		{
			long    l = ((qu_array) x)->len;
			qu_any *p = ((qu_array) x)->objs;
			DIGIT  *d;
			qu_big  r;
			long    n = l;

			if (l == 0)
				return QuBig0;

			if (argc == 2)
			{
				QU_ARG_vlong (1, n);
				QU_ARG_CHECK (n < l);
				if (n < 1)
					n = l;
			}

			QU_ARG_CHECK (n <= QU_BIG_MAXLEN);

			r = _big_alloc (n);
			r->len = n;
			d = r->digits;

			while (n--)
			{
				qu_any z = *p++;
				int v;

				QU_ARG_int (z, v);
				QU_ARG_CHECK (0 <= v && v <= QU_BIG_BASE);
				*d++ = v;
			}
			if (argc > 2)
			{
				int e;

				QU_ARG_vint (2, e);
				QU_ARG_CHECK (-QU_BIG_MAXEXPO <= e && e <= QU_BIG_MAXEXPO);
				r->expo = e;
			}
			return _big_normalize (r, 0);
		}
	}
	return qu_object_as (x, QuBigClass);
}



static qu_big _big_static1 (int v)
{
	qu_big b = qu_gc__new_static (QuBigClass, sizeof (qu_big_s) + sizeof (DIGIT));

	assert (v < QU_BIG_BASE);

	b->digits = (DIGIT *) (b + 1);
	if (v == 0)
	{
		b->len = 0;
		b->expo = 0;
	}
	else if (v < 0)
	{
		b->len = 1;
		b->digits[0] = -v;
		b->expo = 1;
		SETNEG (b);
	}
	else
	{
		b->len = 1;
		b->digits[0] = v;
		b->expo = 1;
	}
	return b;
}



/* sometimes C can be really annoying ... */
static qu_big _big_scan (char **str, int neg, int statik);

qu_big qu_big_new_static (const char *s)
{
	char *z = (char *) s;
	return _big_scan (&z, 0, 1);
}


qu_big qu_big_to_static (qu_big self)
{
	if (self->len == 0)
		return QuBig0;
	else
	{
		register qu_big x;

		x         = qu_gc_new_static (QuBigClass);
		x->cap    = self->cap;
		x->len    = self->len;
		x->expo   = self->expo;
		x->digits = qu_mem_static ((x->cap + 1) * sizeof (DIGIT));
		memcpy (x->digits, self->digits, self->len * sizeof (DIGIT));
		return x;
	}
}



qu_big qu_big_make (DIGIT *digits, long len, int expo)
{
	qu_big b;

	assert (len >= 0);
	assert (-QU_BIG_MAXEXPO <= expo && expo <= QU_BIG_MAXEXPO);

	if (len == 0)
		return QuBig0;

	b = _big_alloc (len);
	b->len = len;
	b->expo = expo;
	memcpy (b->digits, digits, len * sizeof (DIGIT));
	return b;
}



qu_big qu_big_make1 (DIGIT digit, int expo)
{
	qu_big b;

	assert (-QU_BIG_MAXEXPO <= expo && expo <= QU_BIG_MAXEXPO);

	if (digit == 0)
		return QuBig0;

	b = _big_alloc (1);
	b->len = 1;
	b->expo = expo;
	b->digits[0] = digit;
	return b;
}



qu_big qu_big_newd (double v)
{
	if (v == 0)
	{
		return QuBig0;
	}
	else if (v == 1)
	{
		return QuBig1;
	}
	else if (v == -1)
	{
		return QuBig_1;
	}
	else
	{
		TBigTmpd t;
		return (qu_big) _big_return (_big_normalize (_big_tmp_cdouble (&t, v), 1));
	}
}



qu_real qu_big_newdx (double v)
{
	if (qu_double_isfinite (v))
		return (qu_real) qu_big_newd (v);
	QU_REQUIREZ (QuEValue, !qu_double_isnan (v), "Big (double NaN)");
	return (qu_real) __qu_inf__ ((v < 0) ? -1 : 1);
}



#ifdef HAVE_LONG_DOUBLE
qu_big qu_big_newld (long double v)
{
	char buf [64], *b;

	assert (qu_longdouble_isfinite (v));

	snprintf (buf, sizeof (buf), "%.24Lg", v);
	b = buf;
	return qu_big_scan (&b, 0);
}

qu_real qu_big_newldx (long double v)
{
	if (qu_longdouble_isfinite (v))
		return (qu_real) qu_big_newld (v);
	QU_REQUIREZ (QuEValue, !qu_longdouble_isnan (v), "Big (long double NaN)");
	return (qu_real) __qu_inf__ ((v < 0) ? -1 : 1);
}
#endif



qu_big qu_big_newl (long v)
{
	switch (v)
	{
		case  0: return QuBig0;
		case  1: return QuBig1;
		case  2: return QuBig2;
		case -1: return QuBig_1;
		case -2: return QuBig_2;
	}
	/*default*/
	{
		TBigTmpl t;
		return (qu_big) _big_return (_big_normalize (_big_tmp_clong (&t, v), 1));
	}
}



#ifdef HAVE_LONG_LONG
qu_big qu_big_newll (long long v)
{
	switch (v)
	{
		case  0: return QuBig0;
		case  1: return QuBig1;
		case  2: return QuBig2;
		case -1: return QuBig_1;
		case -2: return QuBig_2;
	}
	/*default*/
	{
		char buf [128], *b;
		snprintf (buf, sizeof (buf), "%Ld", v);
		b = buf;
		return qu_big_scan (&b, 0);
	}
}



qu_big qu_big_newull (unsigned long long v)
{
	switch (v)
	{
		case 0: return QuBig0;
		case 1: return QuBig1;
		case 2: return QuBig2;
	}
	/*default*/
	{
		char buf [128], *b;
		snprintf (buf, sizeof (buf), "%Lu", v);
		b = buf;
		return qu_big_scan (&b, 0);
	}
}
#endif



qu_big qu_big_new_long (qu_long v)
{
	if (qu_long_iszero (v))
	{
		return QuBig0;
	}
	else if (qu_long_is1 (v))
	{
		return QuBig1;
	}
	else if (qu_long_isone (v))
	{
		return QuBig_1;
	}
	else
	{
		qu_string  s = qu_long_as_string (v);
		char      *z = s->str;

		return qu_big_scan (&z, 0);
	}
}



qu_big qu_big_new_rational (qu_rational x)
{
	if (qu_rational_iszero (x))
	{
		return QuBig0;
	}
	else if (qu_rational_isone (x))
	{
		return qu_rational_isneg (x) ? QuBig_1 : QuBig1;
	}
	else
	{
		/*
			note that since we can always convert an integer (qu_int or qu_long)
			into a qu_big and the smallest possible value of x->den is 1
			then this never fails.
		*/
		return (qu_big) qu_big_div_integer (qu_integer_as_big (x->num), x->den);
	}
}



qu_big qu_big_new_string (qu_string s)
{
	char   *z = s->str;
	qu_big  r = qu_big_scan (&z, 0);

	QU_REQUIREZ (QuESyntax, r, "Big (String) invalid syntax");
	return r;
}



qu_big qu_big_new (qu_any x)
{
	return qu__big_new (QuBigClass, 1, &x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	scan
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
	!!! scanning a decimal string must not loose digits
*/
static qu_big _big_scan (char **str, int neg, int statik)
{
	long    expo = 0;
	char   *cc, *c = *str;
	long    n, nd, shift, k;
	qu_big  b;
	DIGIT  *d, v;

	if (!neg)
	{
		while (qu_ascii_isspace (*c))
			c++;
		if (*c == '\0')
		{
			errno = EDOM;
			return 0;
		}
		if (*c == '-')
		{
			c++;
			neg = 1;
		}
		else if (*c == '+')
		{
			c++;
		}
	}

	if (*c == '\0')
	{
		errno = EDOM;
		return 0;
	}

	while (*c == '0')
		c++;

	if (*c == 0)
		return QuBig0;

	cc = c;
	while (qu_ascii_isdigit (*c))
	{
		expo++;
		c++;
	}

	if (expo == 0)
	{
		char *sd, *end;

		if (*c == '.')
		{
			c++;
			while (*c == '0')
			{
				expo--;
				c++;
			}
		}
		sd = c;
		nd = 0;
		while (qu_ascii_isdigit (*c))
		{
			c++;
			nd++;
		}
		if (*c == 'e' || *c == 'E' || *c == 'd' || *c == 'D')
		{
			int e = 0;
			int g = 0;

			c++;
			if (*c == '+')
			{
				c++;
			}
			else if (*c == '-')
			{
				c++;
				g = -1;
			}
			while (qu_ascii_isdigit (*c))
			{
				e = e * 10 + (*c - '0');
				if (e > QU_BIG_MAXEXPO * QU_BIG_FIG)
				{
					errno = ERANGE;
					return 0;
				}
				c++;
			}
			if (g)
				e = -e;
			expo += e;
		}
		end = c;

		k = nd / QU_BIG_FIG + 1;
		if (k >= QU_BIG_MAXLEN)
		{
			errno = ERANGE;
			return 0;
		}

		if (statik)
			b =	_big_alloc_static (k + 1);
		else
			b =	_big_alloc (k + 1);
		b->len = k + 1;

		shift = 0;
		while (expo % QU_BIG_FIG)
		{
			shift++;
			expo++;
		}
		b->expo = expo / QU_BIG_FIG;

		n = shift;
		v = 0;
		while (shift--)
			v *= 10;

		d = b->digits;
		k = 0;
		c = sd;
		while (nd && qu_ascii_isdigit (*c))
		{
			v = v * 10 + *c - '0';
			if (++n == QU_BIG_FIG)
			{
				d[k++] = v;
				n = 0;
				v = 0;
			}
			nd--;
			c++;
		}
		if (n)
		{
			while (n < QU_BIG_FIG)
			{
				n++;
				v *= 10;
			}
			d[k] = v;
		}
		c = end;
	}
	else
	{
		if (*c == '.')
			c++;

		while (qu_ascii_isdigit (*c))
			c++;

		if (*c == 'e' || *c == 'E' || *c == 'd' || *c == 'D')
		{
			int e = 0;
			int g = 0;

			c++;
			if (*c == '+')
			{
				c++;
			}
			else if (*c == '-')
			{
				c++;
				g = -1;
			}
			while (qu_ascii_isdigit (*c))
			{
				e = e * 10 + (*c - '0');
				if (e > QU_BIG_MAXEXPO * QU_BIG_FIG)
				{
					errno = ERANGE;
					return 0;
				}
				c++;
			}
			if (g)
				e = -e;
			expo += e;
		}

		k = (c - cc) / QU_BIG_FIG + 1;

		if (statik)
			b =	_big_alloc_static (k + 1);
		else
			b =	_big_alloc (k + 1);
		b->len = k + 1;

		shift = 0;
		while (expo % QU_BIG_FIG)
		{
			shift++;
			expo++;
		}

		b->expo = expo / QU_BIG_FIG;

		n = shift;
		v = 0;
		while (shift--)
			v *= 10;

		d = b->digits;
		k = 0;
		while (qu_ascii_isdigit (*cc))
		{
			v = v * 10 + *cc - '0';
			if (++n == QU_BIG_FIG)
			{
				d[k++] = v;
				n = 0;
				v = 0;
			}
			cc++;
		}
		if (*cc == '.')
		{
			cc++;
			while (qu_ascii_isdigit (*cc))
			{
				v = v * 10 + *cc - '0';
				if (++n == QU_BIG_FIG)
				{
					d[k++] = v;
					n = 0;
					v = 0;
				}
				cc++;
			}
		}
		if (n)
		{
			while (n < QU_BIG_FIG)
			{
				n++;
				v *= 10;
			}
			d[k] = v;
		}
	}

	*str = c;

	if (neg)
		SETNEG (b);

	errno = 0;
	return _big_normalize (b, 0);
}



qu_big qu_big_scan (char **str, int neg)
{
	return _big_scan (str, neg, 0);
}



qu_big qu_big_scan_string (qu_string s)
{
	char   *z = s->str;
	qu_big  x = qu_big_scan (&z, 0);

	QU_REQUIREZ (QuESyntax, x, "Big (String) invalid syntax");
	return x;
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class rand -> Big
	Get a random Big number using the current precision.
*/
static qu_big qu__big_rand (qu_class self)
{
	qu_big  b = _big_alloc (__QuBigPrecision);
	DIGIT  *d = b->digits;
	long    l = __QuBigPrecision;

	b->len = __QuBigPrecision;
	while (l--)
		*d++ = ((double) QU_BIG_BASE * rand() / ((double) RAND_MAX + 1.0));
	b->expo = ((double) QU_BIG_BASE * rand() / ((double) RAND_MAX + 1.0)) / QU_BIG_MAXEXPO;
	if (rand () & 1)
		SETNEG (b);
	return b;
}



/*) sub class precision (*arg) -> Int
	@def
	sub precision -> Int
	sub precision (prec: Int) -> Int
	@p
	Get or set the current precision.

	If called without any argument, return the current precision.

	If called with one argument, set the precision and return the previous
	precision. v{prec} must be a number between #{MINPREC} to #{MAXPREC}.

	Note that precision here means the number of significant digits, not
	the number of digits after the decimal point. The actual number is
	multiplied by 4 thus a precision value of 5 means 20 significant digits.

	The default precision is 8.

	@xmp
		println (Big.precision)
		println (Big.rand)
		old = Big.precision (10)
		println (Big.rand)
		Big.precision (old)

	@out
		8
		-0.84013943783079849116197533527682E+0
		0.4773628836475134952291616357717214166069E+0
*/
static qu_int qu__big_precision (qu_class self, int argc, qu_any *argv)
{
	if (argc == 0)
	{
		return qu_int_new (__QuBigPrecision);
	}
	else
	{
		long p, old;

		QU_ARG_VMAX (1);
		QU_ARG_vlong (0, p);
		QU_ARG_CHECKZ (QU_BIG_MINPREC <= p && p <= QU_BIG_MAXPREC, "invalid precision");

		old = __QuBigPrecision;
		__QuBigPrecision = p;
		return qu_int_new (old);
	}
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__big__destroy (qu_big self)
{
	qu_mem_free (self->digits);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_big_sign (qu_big x)
{
	return x->len ? (qu_flag_test (x, QU_BIG_FLAG_NEG) ? -1 : 1) : 0;
}



int qu_big_isone (qu_big x)
{
	return x->len == 1 && x->expo == 1 && x->digits[0] == 1;
}



int qu_big_is1 (qu_big x)
{
	return qu_big_isone (x) && !qu_big_isneg (x);
}



int qu_big_iswhole (qu_big x)
{
	return x->expo >= x->len;
}



int qu_big_isfrac (qu_big x)
{
	return x->len && x->expo < 1;
}



int qu_big_isodd (qu_big x)
{
	return (x->expo >= x->len) && x->len && (x->digits[x->len - 1] & 1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be one of the following classes:
	@ops
	.| -{Int}      | Return an Int if possible, throw EValue otherwise.
	.| -{Integer}  | Return an Int if possible, Long otherwise.
	.| -{Float}    | Return a Float if possible, throw EValue otherwise.
	.| -{Complex}  | Return C{self + 0j}.
	.| -{Rational} | Return a Rational number.
	.| -{Array}    | Return an Array containing the digits.
	.| -{Boolean}  | Delegate to #{__true}
	.| -{String}   | Delegate to #{__string}
	.| -{Big}
		-{Decimal}
		-{Real}
		-{Number}  | Return self

	@exc
	.EArgValue
	.EValue

	@xmp
		b = Big ('123.45')
		println (b as Int)
		println (b as Rational)
		println (b as Array)

	@out
		123
		(2469 _/ 20)
		[4500, 123]
*/
static qu_any qu__big__as (qu_big self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:      return qu_big_as_int      (self);
		case QU_CLASS_ID_Long:     return qu_big_as_long     (self);
		case QU_CLASS_ID_Integer:  return qu_big_as_integer  (self);
		case QU_CLASS_ID_Rational: return qu_big_as_rational (self);
		case QU_CLASS_ID_Float:    return qu_big_as_float    (self);
		case QU_CLASS_ID_Complex:  return qu_big_as_complex  (self);
		case QU_CLASS_ID_Array:    return qu_big_as_array    (self);
		case QU_CLASS_ID_Big:
		case QU_CLASS_ID_Decimal:
		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Number:   return self;
		case QU_CLASS_ID_String:   return qu_big_as_string   (self);
		case QU_CLASS_ID_Boolean:  return qu_big_as_boolean  (self);
	}
	QU_ARG_THROW;
}



long qu_big_tol (qu_big x)
{
	int    sx = qu_big_sign (x);
	long   l, e;
	DIGIT *d;
	long   v, m;

	errno = 0;

	if (sx == 0)
		return 0;

	e = x->expo;
	if (e <= 0)
		return 0;

	m = (sx < 0) ? (DIGIT) LONG_MIN : (DIGIT) LONG_MAX;

	l = x->len;
	d = x->digits;
	v = 0;
	while (l && e > 0)
	{
		DIGIT g;
		long     o;

		if (l)
			g = *d++;
		else
			g = 0;

		o = v;
		v = v * QU_BIG_BASE + g;
		if (v <= o || v < 0 || v > m)
			goto nope;

		e--;
		l--;
	}

	if (e == 0)
		return (sx < 0) ? -v : v;

  nope:
	errno = ERANGE;
	return qu_big_isneg (x) ? LONG_MIN : LONG_MAX;
}



unsigned long qu_big_toul  (qu_big x)
{
	int    sx = qu_big_sign (x);
	long   l, e;
	DIGIT *d;
	unsigned long v;

	errno = 0;

	if (sx == 0)
		return 0;

	if (sx < 0)
	{
		errno = ERANGE;
		return 0;
	}

	e = x->expo;
	if (e < 0)
		return 0;

	l = x->len;
	d = x->digits;
	v = 0;
	while (l && e)
	{
		DIGIT g;
		unsigned long o;

		if (l)
			g = *d++;
		else
			g = 0;

		o = v;
		v = o * QU_BIG_BASE + g;

		if (v <= o)
			goto nope;

		e--;
		l--;
	}

	if (e == 0)
		return v;

  nope:
	errno = ERANGE;
	return ULONG_MAX;
}



#ifdef HAVE_LONG_LONG
long long qu_big_toll (qu_big x)
{
	long long v, m;
	DIGIT *d;
	long   l, e;
	int    sx = qu_big_sign (x);

	errno = 0;

	if (sx == 0)
		return 0;

	e = x->expo;
	if (e < 0)
		return 0;

	m = (sx < 0) ?  LONG_LONG_MIN : LONG_LONG_MAX;

	l = x->len;
	d = x->digits;
	v = 0;
	while (l && e)
	{
		DIGIT  g;
		long long o;

		if (l)
			g = *d++;
		else
			g = 0;

		o = v;
		v = v * QU_BIG_BASE + g;

		if (v < o || v < 0 || v > m)
			goto nope;

		e--;
		l--;
	}
	if (e == 0)
		return qu_big_isneg (x) ? -v : v;

  nope:
	errno = ERANGE;
	return qu_big_isneg (x) ? LONG_LONG_MIN : LONG_LONG_MAX;
}



unsigned long long qu_big_toull (qu_big x)
{
	unsigned long long  v;
	DIGIT *d;
	long   l, e;

	errno = 0;

	if (qu_big_iszero (x))
		return 0;

	if (qu_big_isneg (x))
	{
		errno = ERANGE;
		return 0;
	}

	e = x->expo;
	if (e < 0)
		return 0;

	l = x->len;
	d = x->digits;
	v = 0;
	while (l && e)
	{
		DIGIT g;
		unsigned long long o;

		if (l)
			g = *d++;
		else
			g = 0;

		o = v;
		v = v * QU_BIG_BASE + g;
		if (v <= o)
			goto nope;

		e -= QU_BIG_FIG;
		l--;
	}
	if (e == 0)
		return qu_big_isneg (x) ? -v : v;

  nope:
	errno = ERANGE;
	return ULONG_LONG_MAX;
}
#endif



float qu_big_tof (qu_big x)
{
	double v = qu_big_tod (x);

	if (errno == 0)
	{
		if (v < 0)
		{
			if (v >= (double) FLT_MIN)
				return (float) v;
		}
		else
		{
			if (v <= (double) FLT_MAX)
				return (float) v;
		}
		errno = ERANGE;
	}
	return (v < 0) ? FLT_MIN : FLT_MAX;
}



double qu_big_tod (qu_big x)
{
	DIGIT *d;
	double v, e;
	long   l;

	errno = 0;

	if (qu_big_iszero (x))
		return 0;

	l = x->len;
	d = x->digits;
	if (l > QU_BIG_FIG_DOUBLE)
		l = QU_BIG_FIG_DOUBLE;
	e = 1.0;
	v = 0.0;
	while (l--)
	{
		e /= (double) QU_BIG_BASE;
		v += e * *d++;
	}
	v *= pow (10.0, (double) QU_BIG_FIG * x->expo);
	if (!qu_double_isfinite (v))
	{
		errno = ERANGE;
		return qu_big_isneg (x) ? -DBL_MAX : DBL_MAX;
	}
	return qu_big_isneg (x) ? -v : v;
}



#ifdef HAVE_LONG_DOUBLE
extern long double powl (long double BASE, long double POWER);
long double qu_big_told  (qu_big x)
{
	long double  v, e;
	DIGIT *d;
	long   l;

	errno = 0;

	if (qu_big_iszero (x))
		return 0;

	l = x->len;
	d = x->digits;
	if (l > 48)
		l = 48;
	e = 1.0;
	v = 0.0;
	while (l--)
	{
		e /= (double) QU_BIG_BASE;
		v += e * *d++;
	}
	v *= powl (10.0, (double) QU_BIG_FIG * x->expo);
	if (!qu_longdouble_isfinite (v))
	{
		errno = ERANGE;
		return qu_big_isneg (x) ? -LDBL_MAX : LDBL_MAX;
	}
	return qu_big_isneg (x) ? -v : v;
}
#endif




qu_int qu_big_as_int (qu_big x)
{
	long v = qu_big_toli (x);
	QU_REQUIREZ (QuEValue, errno == 0, "value out of Int range");
	return qu_int_new (v);
}



qu_rational qu_big_as_rational (qu_big x)
{
	if (qu_big_iszero (x))
	{
		return QuRational0;
	}
	else if (qu_big_is1 (x))
	{
		return QuRational1;
	}
	else if (qu_big_isone (x))
	{
		return QuRational_1;
	}
	else
	{
		qu_integer n, d;
		long       e = x->expo;

		x->expo = x->len;
		n = (qu_integer) qu_big_as_long (x);
		assert (n);
		x->expo = e;

		e = x->len - e;
		if (e == 0)
		{
			return qu_rational_new_num (n);
		}
		else if (e > 0)
		{
			d = (qu_integer) qu_long_pow_10 (e * QU_BIG_FIG);
			assert (d && qu_integer_is (d));
			return qu_rational_newr (n, (qu_integer) d);
		}
		else
		{
			d = (qu_integer) qu_long_pow_10 (-e * QU_BIG_FIG);
			assert (d && qu_integer_is (d));
			n = (qu_integer) qu_integer_mul_integer (n, d);
			assert (n && qu_integer_is (n));
			return qu_rational_new_num ((qu_integer) n);
		}
	}
}



qu_long qu_big_as_long (qu_big x)
{
	if (qu_big_iszero (x) || x->expo < 0)
	{
		return QuLong0;
	}
	else if (qu_big_is1 (x))
	{
		return QuLong1;
	}
	else if (qu_big_isone (x))
	{
		return QuLong_1;
	}
	else
	{
		long       l = x->len;
		DIGIT     *d = x->digits;
		long       e = x->expo;
		char      *z;
		qu_string  s;
		int        k = 1;

		s = qu_string_buffer_new ((e + 1) * QU_BIG_FIG);
		if (qu_big_isneg (x))
			qu_string_buffer_putc (s, '-');
		while (e)
		{
			if (l)
			{
				char     b [64];
				DIGIT v = *d++;

				if (k)
				{
					sprintf (b, "%lu", v);
					k = 0;
				}
				else
				{
					sprintf (b, QU_BIG_FORMAT, v);
				}
				l--;
				qu_string_buffer_putz (s, b);
			}
			else
			{
				qu_string_buffer_put (s, QU_BIG_FIG_ZERO, QU_BIG_FIG);
			}
			e--;
		}
		s = qu_string_buffer_return (s);
		z = s->str;
		return qu_long_scan (&z, 0);
	}
}



qu_float qu_big_as_float (qu_big x)
{
	double d = qu_big_tod (x);
	QU_REQUIREZ (QuEValue, errno == 0, "value out of Float range");
	return qu_float_new (d);
}



qu_integer qu_big_as_integer (qu_big self)
{
	long v = qu_big_toli (self);

	if (errno == 0)
		return (qu_integer) qu_int_new (v);
	return (qu_integer) qu_big_as_long (self);
}



qu_array qu_big_as_array (qu_big self)
{
	long      l = self->len;
	DIGIT    *d = self->digits;
	qu_array  a;

	assert (l <= QU_GC_MAX_ITEMS);
	a = qu_array_new (l);
	assert (a);
	while (l--)
		qu_array_push (a, qu_integer_newul (*d++));
	return a;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	Compare self with another object.

	If v{x} is a Number then their numerical values are compared.
	Refer to -{Number.__cmp} for details on how comparisson is made.

	The returned value is -1 if self is smaller than v{x}, 0 if self equals v{x},
	1 otherwise.

	@noexc

	@xmp
		println (Big ('1') <=> 0)
		println (Big ('1') <=> 1)
		println (Big ('1') <=> 2)

		class Foo
			var value = 10
			sub __cmp (x)
				switch x.__class
					case Int: return self.value <=> x
				;;
			;;
		;;

		x = Foo ()
		println (x <=> Big ('20'))

	@out
		1
		0
		-1
		-1
*/
static qu_int qu__big__cmp (qu_big self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_big_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_big_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_big_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_big_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_big_cmp_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_big_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_big_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}



int qu_big_cmp_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_cmp_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



int qu_big_cmp_long (qu_big x, qu_long y)
{
	return qu_big_cmp_big (x, qu_long_as_big (y));
}



int qu_big_cmp_rational (qu_big x, qu_rational y)
{
	return qu_big_cmp_big (x, qu_rational_as_big (y));
}



int qu_big_cmp_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_cmp_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



int qu_big_cmp_big (qu_big x, qu_big y)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	QU_NUMBER_COMPARE_SIGN (sx, sy);
	assert (sx == sy && sx);
	return (sx < 0) ? qu_big_cmpabs_big (y, x) : qu_big_cmpabs_big (x, y);
}



QU_NUMBER_BINOP_integer (qu_big, cmp, int)
QU_NUMBER_BINOP_decimal (qu_big, cmp, int)
QU_NUMBER_BINOP_real    (qu_big, cmp, int)
QU_NUMBER_BINOP_number  (qu_big, cmp, int)



/*
	|<=>|
*/

int qu_big_cmpabs_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_cmpabs_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



int qu_big_cmpabs_long (qu_big x, qu_long y)
{
	return qu_big_cmpabs_big (x, qu_long_as_big (y));
}



int qu_big_cmpabs_rational (qu_big x, qu_rational y)
{
	return qu_big_cmpabs_big (x, qu_rational_as_big (y));
}



int qu_big_cmpabs_big (qu_big x, qu_big y)
{
	long lx = x->len;
	long ly = y->len;

	if (lx == 0)
	{
		if (ly == 0)
			return 0;
		return -1;
	}

	if (ly == 0)
		return 1;

	if (x->expo > y->expo)
		return 1;

	if (x->expo < y->expo)
		return -1;

	/*else*/
	{
		DIGIT *dx = x->digits;
		DIGIT *dy = y->digits;

		while (lx && ly)
		{
			if (*dx > *dy)
				return 1;

			else if (*dx != *dy)
				return -1;

			dx++;
			dy++;
			lx--;
			ly--;
		}
		return QU_NUMBER_COMPARE_VALUE (lx, ly);
	}
}



int qu_big_cmpabs_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_cmpabs_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



QU_NUMBER_BINOP_integer (qu_big, cmpabs, int)
QU_NUMBER_BINOP_decimal (qu_big, cmpabs, int)
QU_NUMBER_BINOP_real    (qu_big, cmpabs, int)
QU_NUMBER_BINOP_number  (qu_big, cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Number
	Return a new Big by subtracting 1 from v{self}.

	If the result is huge and the -{Number.context} does not allow infinities
	then throws EHuge, otherwise returns infn.

	Note that ESmall is not possible since the subtractor is a whole number.

	@exc
	.EHuge

	@xmp
		a = Big (10)
		b = --a
		println (b)

	@out
		0.0009E+04
*/
static qu_real qu__big__decr (qu_big self)
{
	return qu_big_plus_big (self, QuBig_1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number
	Return the result of dividing self with v{x}.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{\} operator).

	Refer to -{Number.__div} for details on how the division is performed.

	@exc
	.EDiv
	.EHuge
	.ETiny
	.EInf
	.EComplex

	@xmp
		x = Big ('11')
		y = Big ('3')
		v = x / y
		i = x \ y
		r = x % y
		println (v)
		println (i)
		println (r)

	@out
		0.0003666666666666666666666666E+4
		3
		0.0002E+4

	@note
	Qu bothers to check whether self is v{x} (the same object) and return
	Big (1) if it is.
*/
static qu_number qu__big__div (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_big_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_big_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_big_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_big_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_big_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_big_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



/*
	REM (x, y) = SIGN (x) * (|x| - |y| * IDIV (|x|, |y|))

	For y = 0:

		REM (x, 0) = x - 0 * IDIV (x, 0)
		REM (x, 0) = x - NaN
		REM (x, 0) = NaN

	!!! if the divisor is zero, the remainder is undefined
	!!! thus *rem is set to NULL
*/
static qu_real _big_div_res (qu_big x, qu_big y, qu_real *rem)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	if (sx == 0)
	{
		if (rem)
			*rem = (qu_real) QuBig0;

		return (qu_real) QuBig0;
	}
	else if (sy == 0)
	{
		if (rem)
			*rem = NULL;

		return (qu_real) __qu_zerodiv__ (sx);
	}
	else if (qu_big_isone (y))
	{
		if (rem)
			*rem = (qu_real) QuBig0;
		if (sx * sy == sx)
			return _big_return (x);
		return (qu_real) qu_big_neg (x);
	}
	else
	{
		qu_big r, c;

		c = _big_basic_div (x, y, &r);
		if (r->len)
		{
			SETSIGN (r, sx);
			if (rem)
				*rem = _big_return (r);
		}
		else
		{
			if (rem)
				*rem = (qu_real) r;
		}
		if (c->len)
			SETSIGN (c, sx * sy);
		return _big_return (c);
	}
}



qu_real qu_big_div_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_div_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



qu_real qu_big_div_long (qu_big x, qu_long y)
{
	return qu_big_div_big (x, qu_long_as_big (y));
}



qu_real qu_big_div_rational (qu_big x, qu_rational y)
{
	return qu_big_div_big (x, qu_rational_as_big (y));
}



qu_real qu_big_div_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_div_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



qu_real qu_big_div_big (qu_big x, qu_big y)
{
	if (x == y)
		return (qu_real) QuBig1;
	return _big_div_res (x, y, NULL);
}



qu_big qu_big_div_inf (qu_big x, qu_inf y)
{
	QU_REQUIREZ (QuEDiv, !qu_big_iszero (x), "Big(0) / infinity");
	return QuBig0;
}



QU_NUMBER_BINOP_integer (qu_big, div, qu_real)
QU_NUMBER_BINOP_decimal (qu_big, div, qu_real)
QU_NUMBER_BINOP_real    (qu_big, div, qu_real)
QU_NUMBER_BINOP_number  (qu_big, div, qu_number)



/*
	/ reversed
*/

qu_real qu_big_rdiv_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_div_big (_big_tmp_clong (&t, qu_int_tol (y)), x);
}



qu_real qu_big_rdiv_long (qu_big x, qu_long y)
{
	return qu_big_div_big (qu_long_as_big (y), x);
}




qu_real qu_big_rdiv_rational (qu_big x, qu_rational y)
{
	return qu_big_div_big (qu_rational_as_big (y), x);
}




qu_real qu_big_rdiv_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_div_big (_big_tmp_cdouble (&t, qu_float_tod (y)), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if v{x} is a -{Number} and has the same numerical value
	as self, false otherwise.

	Refer to -{Number.__eq} for details on how the equality test is performed.

	@noexc

	@xmp
		x = Big ('1.0000')
		println (x == 1)

	@out
		true
*/
static qu_boolean qu__big__eq (qu_big self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_big_eq_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_big_eq_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_big_eq_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_big_eq_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_big_eq_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_big_eq_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



int qu_big_eq_int (qu_big x, qu_int y)
{
	if (qu_big_sign (x) == qu_int_sign (y) && qu_big_iswhole (x))
	{
		long v = qu_big_tol (x);
		if (errno == 0)
			return v == qu_int_tol (y);
	}
	return 0;
}



int qu_big_eq_long (qu_big x, qu_long y)
{
	if (qu_big_sign (x) == qu_long_sign (y) && qu_big_iswhole (x))
		return qu_big_cmp_big (x, qu_long_as_big (y)) == 0;
	return 0;
}



int qu_big_eq_rational (qu_big x, qu_rational y)
{
	if (qu_rational_iswhole (y))
	{
		return qu_big_eq_integer (x, y->num);
	}
	else if (qu_big_sign (x) == qu_rational_sign (y))
	{
		if (qu_big_iszero (x))
			return 1;
		else
			return qu_big_eq_big (x, qu_rational_as_big (y));
	}
	return 0;
}



int qu_big_eq_float (qu_big x, qu_float y)
{
	if (qu_big_sign (x) == qu_float_sign (y))
	{
		double v = qu_big_tod (x);
		if (errno == 0)
			return v == qu_float_tod (y);
	}
	return 0;
}



int qu_big_eq_big (qu_big x, qu_big y)
{
	return qu_big_cmp_big (x, y) == 0;
}



QU_NUMBER_BINOP_integer (qu_big, eq, int)
QU_NUMBER_BINOP_decimal (qu_big, eq, int)
QU_NUMBER_BINOP_real    (qu_big, eq, int)
QU_NUMBER_BINOP_number  (qu_big, eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this Big.

	See -{Number.__format} for details on the format.

	@exc
	.EFormat
	.EMemory

	@xmp
		f = Big (1234.56)
		println (f.__format ('%.3d'))
		printfln ('f = %.2,3m', f)
		printfln ('f = %10.3e', f)

	@out
		1234.559
		f = 1,234.55
		f =   1.234e+3
*/
static qu_string qu__big__format (qu_big self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_big_format (self, &f);
}



static qu_string _big_digits (qu_big x, long ndigits, long *expo)
{
	if (qu_big_iszero (x))
	{
		*expo = 0;
		return qu_string_byte ('0');
	}
	else
	{
		qu_string  s = qu_string_buffer_new (64);
		long       l = x->len;
		long       e = x->expo;
		DIGIT     *d = x->digits;
		char       b [64];
		int        k = 1;
		long       n;

		if (ndigits < QU_BIG_FIG)
			ndigits = QU_BIG_FIG;

		while (l > 0 && ndigits > 0)
		{
			l--;
			if (k)
			{
				k = 0;
				sprintf (b, "%lu", *d++);
			}
			else
			{
				sprintf (b, QU_BIG_FORMAT, *d++);
			}
			n = strlen (b);
			e--;
			ndigits -= n;
			QU_ENSURE (qu_string_buffer_put (s, b, n));
		}
		*expo = e * QU_BIG_FIG;
		return qu_string_buffer_return (s);
	}
}



qu_string qu_big_format (qu_big x, qu_format_s *f)
{
	long      n, e;
	qu_string s;

	switch (f->mode)
	{
		case 'y': return qu_format_yesno (f, x);
		case 'p': return qu_format_ptr (f, x);
	}

	n = f->width + f->prec;
	if (n <= 0)
		n = INT_MAX;
	s = _big_digits (x, n, &e);

	switch (f->mode)
	{
		case 'x':
		case 'o':
		case 'b': f->mode = 'd';
	}
	return qu_format_number (f, s, qu_big_isneg (x), e);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> Int
	Return the hash value of this Big.

	@noexc

	@xmp
		x = Big ('1234.567')
		println (x.__hash)
		x = Big ('1234.567e100')
		println (x.__hash)
		println (Int.MAX)

	@out
		1234
		1073741823
*/
static qu_int qu__big__hash (qu_big self)
{
	long v = qu_big_toli (self);
	return qu_int_new (v);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer
	Divide self with v{x} in an I{integer context}.

	@warn
	There is a big difference between the #{__div} method (the C{/} operator)
	and the __idiv method (the C{\} operator).

	Refer to -{Number.__idiv} for details on how the division is performed.

	@p
	This method finds the result by scaling the objects. In rare circumstances
	where the scaled objects are too big, then EBrain is thrown.

	@exc
	.EIdiv
	.EBrain

	@xmp
		a = Big ('10')
		b = Big ('3.9')
		println (a \ b)

	@out
		2

	@note
	Qu bothers to check whether self is v{x} (the same object) and return 1
	if it is.
*/
static qu_integer qu__big__idiv (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_big_idiv_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_big_idiv_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_big_idiv_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_big_idiv_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_big_idiv_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_big_idiv_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_big_idiv_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEIdiv);
}



qu_integer qu_big_idiv_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_idiv_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



qu_integer qu_big_idiv_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_idiv_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



qu_integer qu_big_idiv_long (qu_big x, qu_long y)
{
	return qu_big_idiv_big (x, qu_long_as_big (y));
}



qu_integer qu_big_idiv_rational (qu_big x, qu_rational y)
{
	return qu_big_idiv_big (x, qu_rational_as_big (y));
}



qu_integer qu_big_idiv_big (qu_big x, qu_big y)
{
	if (x == y)
	{
		return (qu_integer) QuInt1;
	}
	else
	{
		int cmp;

		QU_REQUIRE (QuEIdiv, !qu_big_iszero (y));

		cmp = qu_big_cmpabs_big (x, y);

		if (cmp < 0)
			return (qu_integer) QuInt0;

		else if (cmp == 0)
			return (qu_big_isneg (x) == qu_big_isneg (y)) ? (qu_integer) QuInt1 : (qu_integer) QuInt_1;

		else
		{
			/* TODO - there should be a better way of doing this */

			qu_integer a, b;
			long       ex = x->expo;
			long       ey = y->expo;
			long       lx = x->len;
			long       ly = y->len;

			long       l = (lx > ly) ? lx : ly;
			long       e = ex - ey;

			e += l;

			QU_REQUIRE (QuEBrain, x->expo <= QU_BIG_MAXEXPO - e);
			QU_REQUIRE (QuEBrain, y->expo <= QU_BIG_MAXEXPO - e);

			x->expo += e;
			y->expo += e;

			a = qu_big_as_integer (x);
			b = qu_big_as_integer (y);

			x->expo = ex;
			y->expo = ey;

			return qu_integer_idiv_integer (a, b);
		}
	}
}



qu_integer qu_big_idiv_inf (qu_big x, qu_inf y)
{
	QU_REQUIRE (QuEIdiv, !qu_big_iszero (x));
	return (qu_integer) QuInt0;
}



QU_NUMBER_BINOP_integer (qu_big, idiv, qu_integer)
QU_NUMBER_BINOP_decimal (qu_big, idiv, qu_integer)
QU_NUMBER_BINOP_real    (qu_big, idiv, qu_integer)
QU_NUMBER_BINOP_number  (qu_big, idiv, qu_integer)



/*
	\ reversed
*/

qu_integer qu_big_ridiv_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_idiv_big (_big_tmp_clong (&t, qu_int_tol (y)), x);
}



qu_integer qu_big_ridiv_long (qu_big x, qu_long y)
{
	return qu_big_idiv_big (qu_long_as_big (y), x);
}



qu_integer qu_big_ridiv_rational (qu_big x, qu_rational y)
{
	return qu_big_idiv_big (qu_rational_as_big (y), x);
}



qu_integer qu_big_ridiv_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_idiv_big (_big_tmp_cdouble (&t, qu_float_tod (y)), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Real
	Return self incremented by 1.

	If the result is huge and the -{Number.context} does not allow infinities
	then throws EHuge, otherwise returns infp.

	@exc
	.EHuge

	@xmp
		a = Big ('12.3')
		++a
		println (a)

	@out
		0.00133000E+4
*/
static qu_real qu__big__incr (qu_big self)
{
	return qu_big_plus_big (self, QuBig1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __len -> Int
	Return the number of digits used.

	@noexc

	@warn
	This value may depend on the internal implementation of the Big class
	(although this will probably never change).

	@xmp
		b = Big ('1234567.89')
		println (b.__len)

	@out
		5
*/
static qu_int qu__big__len (qu_big self)
{
	assert (self->len <= QU_INT_MAX);
	return qu_int_new (self->len);
}



/*) sub __like (x) -> Boolean
	If v{x} is a -{Number} and it has the same absolute numerical
	value as self then true is returned, false otherwise.

	Refer to -{Number.__like} for for details on how the comparisson is performed.

	@noexc

	@xmp
		println (Big ('-1.0000') === 1)

	@out
		true
*/
static qu_boolean qu__big__like (qu_big self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_big_like_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_big_like_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_big_like_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_big_like_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_big_like_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_big_like_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



QU_NUMBER_BINOP_integer (qu_big, like, int)
QU_NUMBER_BINOP_decimal (qu_big, like, int)
QU_NUMBER_BINOP_real    (qu_big, like, int)
QU_NUMBER_BINOP_number  (qu_big, like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Return the result of subtracting self with v{x}.

	Refer to -{Number.__minus} for for details on how the subtraction is performed.

	@exc
	.EMinus
	.EHuge
	.ESmall
	.EInf
	.EComplex

	@xmp
		x = Big ('1234')
		println (x - 4)

	@out
		0.1230E+4
*/
static qu_number qu__big__minus (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_big_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_big_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_big_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_big_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_big_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_big_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



qu_real qu_big_minus_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_minus_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



qu_real qu_big_minus_long (qu_big x, qu_long y)
{
	return qu_big_minus_big (x, qu_long_as_big (y));
}



qu_real qu_big_minus_rational (qu_big x, qu_rational y)
{
	return qu_big_minus_big (x, qu_rational_as_big (y));
}



qu_real qu_big_minus_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_minus_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



qu_real qu_big_minus_big (qu_big x, qu_big y)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	if (sx == 0)
	{
		if (sy == 0)
			return (qu_real) QuBig0;

		y = qu_big_dup (y);
		SETNEG (y);
		return (qu_real) y;
	}
	else if (sy == 0)
	{
		return _big_return (x);
	}
	else if (sx > 0)
	{
		int c = qu_big_cmpabs_big (x, y);

		/*
			+a - +b => |a| - |b|
			+a - -b => |a| + |b|
		*/
		if (sy > 0)
		{
			if (c > 0)
			{
				/* +a - +b => |a| - |b| */
				return _big_return (_big_basic_minus (x, y));
			}
			else if (c < 0)
			{
				/* +a - +b => -(|b| - |a|) */
				qu_big r = _big_basic_minus (y, x);

				if (r->len)
				{
					if (r == x || r == y)
						r = qu_big_dup (r);
					SETNEG (r);
				}
				return _big_return (r);
			}
			else
			{
				return (qu_real) QuBig0;
			}
		}
		else
		{
			if (c >= 0)
			{
				/* +a - -b => |a| + |b| */
				return _big_return (_big_basic_plus (x, y));
			}
			else
			{
				/* +a - -b => |b| + |a| */
				return _big_return (_big_basic_plus (y, x));
			}
		}
	}
	else
	{
		/*
			-a - +b => -(|a| + |b|)
			-a - -b => -(|a| - |b|)
		*/

		int    c = qu_big_cmpabs_big (x, y);
		qu_big r;

		if (sy > 0)
		{
			if (c >= 0)
			{
				/* -a - +b => -(|a| + |b|) */
				r = _big_basic_plus (x, y);
			}
			else
			{
				/* -a - +b => -(|b| + |a|) */
				r = _big_basic_plus (y, x);
			}
		}
		else
		{
			if (c > 0)
			{
				/* -a - -b => -(|a| - |b|) */
				r = _big_basic_minus (x, y);
			}
			else if (c < 0)
			{
				/* -a - -b => |b| - |a| */
				return _big_return (_big_basic_minus (y, x));
			}
			else
			{
				return (qu_real) QuBig0;
			}
		}

		if (r->len)
		{
			if (r == x || r == y)
				r = qu_big_dup (r);

			SETNEG (r);
		}
		return _big_return (r);
	}
}



QU_NUMBER_BINOP_integer (qu_big, minus, qu_real)
QU_NUMBER_BINOP_decimal (qu_big, minus, qu_real)
QU_NUMBER_BINOP_real    (qu_big, minus, qu_real)
QU_NUMBER_BINOP_number  (qu_big, minus, qu_number)



/*
	- reversed
*/

qu_real qu_big_rminus_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_minus_big (_big_tmp_clong (&t, qu_int_tol (y)), x);
}



qu_real qu_big_rminus_long (qu_big x, qu_long y)
{
	return qu_big_minus_big (qu_long_as_big (y), x);
}



qu_real qu_big_rminus_rational (qu_big x, qu_rational y)
{
	return qu_big_minus_big (qu_rational_as_big (y), x);
}



qu_real qu_big_rminus_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_minus_big (_big_tmp_cdouble (&t, qu_float_tod (y)), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Return the result of multiplying self with v{x}.

	Refer to -{Number.__mul} for for details on how the multiplication is performed.

	@exc
	.EMul
	.EHuge
	.ETiny
	.EInf
	.EComplex

	@xmp
		x = Big ('1.2')
		println (x * 2)

	@out
		0.00024000E+4
*/
static qu_number qu__big__mul (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_big_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_big_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_big_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_big_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_big_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_big_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



qu_real qu_big_mul_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_mul_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



qu_real qu_big_mul_long (qu_big x, qu_long y)
{
	return qu_big_mul_big (x, qu_long_as_big (y));
}



qu_real qu_big_mul_rational (qu_big x, qu_rational y)
{
	return qu_big_mul_big (x, qu_rational_as_big (y));
}



qu_real qu_big_mul_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_mul_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



qu_real qu_big_mul_big (qu_big x, qu_big y)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	if (sx == 0 || sy == 0)
	{
		return (qu_real) QuBig0;
	}
	else if (qu_big_isone (x))
	{
		sx *= sy;
		if (sx != sy)
		{
			qu_big r = qu_big_dup (y);
			SETSIGN (r, sx);
			return (qu_real) r;
		}
		return _big_return (y);
	}
	else if (qu_big_isone (y))
	{
		sy *= sx;
		if (sy != sx)
		{
			qu_big r = qu_big_dup (x);
			SETSIGN (r, sy);
			return (qu_real) r;
		}
		return _big_return (x);
	}
	else
	{
		qu_big r = _big_basic_mul (x, y);

		if (r->len)
		{
			assert (r != x && r != y);
			SETSIGN (r, sx * sy);
			return _big_return (r);
		}
		return (qu_real) r;
	}
}



QU_NUMBER_BINOP_integer (qu_big, mul, qu_real)
QU_NUMBER_BINOP_decimal (qu_big, mul, qu_real)
QU_NUMBER_BINOP_real    (qu_big, mul, qu_real)
QU_NUMBER_BINOP_number  (qu_big, mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Big
	Return the negation of self.

	Note that there is no conversion here. The result is always a Big even
	if the result fits in a smaller number.

	@noexc

	@xmp
		x = Big ('1.234')
		println (-x)

	@out
		-0.00012340E+4
*/
qu_big qu__big__neg (qu_big self)
{
	if (qu_big_iszero (self))
	{
		return QuBig0;
	}
	else
	{
		qu_big r = qu_big_dup (self);

		if (qu_big_isneg (self))
			SETPOS (r);
		else
			SETNEG (r);
		return r;
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Return the result of adding v{x} to self.

	Refer to -{Number.__plus} for for details on how the addition is performed.

	@exc
	.EPlus
	.EHuge
	.ESmall
	.EInf
	.EComplex

	@xmp
		x = Big ('2.34')
		println (x + 1)

	@out
		0.00033400E+4
*/
static qu_number qu__big__plus (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_big_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_big_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_big_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_big_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_big_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_big_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPlus);
}



qu_real qu_big_plus_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_plus_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



qu_real qu_big_plus_long (qu_big x, qu_long y)
{
	return qu_big_plus_big (x, qu_long_as_big (y));
}



qu_real qu_big_plus_rational (qu_big x, qu_rational y)
{
	return qu_big_plus_big (x, qu_rational_as_big (y));
}



qu_real qu_big_plus_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_plus_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



qu_real qu_big_plus_big (qu_big x, qu_big y)
{
	int sx = qu_big_sign (x);
	int sy = qu_big_sign (y);

	if (sx == 0)
	{
		return _big_return (y);
	}
	else if (sy == 0)
	{
		return _big_return (x);
	}
	else if (sx > 0)
	{
		int c = qu_big_cmpabs_big (x, y);

		/*
			+a + +b => |a| + |b|
			+a + -b => |a| - |b|
		*/
		if (sy > 0)
		{
			if (c >= 0)
			{
				/* +a + +b => |a| + |b| */
				return _big_return (_big_basic_plus (x, y));
			}
			else
			{
				/* +a + +b => |b| + |a| */
				return _big_return (_big_basic_plus (y, x));
			}
		}
		else
		{
			if (c > 0)
			{
				/* +a + -b => |a| - |b| */
				return _big_return (_big_basic_minus (x, y));
			}
			else if (c < 0)
			{
				/* +a + -b => -(|b| - |a|) */
				qu_big r = _big_basic_minus (y, x);

				if (r->len)
				{
					if (r == x || r == y)
						r = qu_big_dup (r);
					SETNEG (r);
				}
				return _big_return (r);
			}
			else
			{
				return (qu_real) QuBig0;
			}
		}
	}
	else
	{
		/*
			-a + +b => -(|a| - |b|)
			-a + -b => -(|a| + |b|)
		*/

		int    c = qu_big_cmpabs_big (x, y);
		qu_big r;

		if (sy > 0)
		{
			if (c > 0)
			{
				/* -a + +b => -(|a| - |b|) */
				r = _big_basic_minus (x, y);
			}
			else if (c < 0)
			{
				/* -a + +b => (|b| - |a|) */
				return _big_return (_big_basic_minus (y, x));
			}
			else
			{
				return (qu_real) QuBig0;
			}
		}
		else
		{
			if (c >= 0)
			{
				/* -a + -b => -(|a| + |b|) */
				r = _big_basic_plus (x, y);
			}
			else
			{
				/* -a + -b => -(|b| + |a|) */
				r = _big_basic_plus (y, x);
			}
		}
		if (r->len)
		{
			if (r == x || r == y)
				r = qu_big_dup (r);
			SETNEG (r);
		}
		return _big_return (r);
	}
}



QU_NUMBER_BINOP_integer (qu_big, plus, qu_real)
QU_NUMBER_BINOP_decimal (qu_big, plus, qu_real)
QU_NUMBER_BINOP_real    (qu_big, plus, qu_real)
QU_NUMBER_BINOP_number  (qu_big, plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub pos -> self
	This method always return self.
*/
#define qu__big__pos  qu_object_method_self



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x:Number) -> Number
	Return the result of raising self to the power of v{x}.

	Refer to -{Number.__pow} for for details on how the operation is performed.

	@exc
	.EPow
	.ETiny
	.EHuge
	.EInf
	.EComplex

	@xmp
		x = Big (2)
		y = x ** 2
		println (y)

	@out
		0.0004E+4
*/
static qu_number qu__big__pow (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_big_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_big_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_big_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_big_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_big_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_big_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}



/*
	!!! dont call if p <= 1
*/
static qu_real _big_pow_clong (qu_big x, long p)
{
	qu_big r = x, w = x;

	assert (p > 1);
	for (p--; p; p--)
	{
		while ((p & 1) == 0)
		{
			p >>= 1;
			w = (qu_big) qu_big_mul_big (w, w);
			if (w == 0 || !qu_big_is (w))
				return (qu_real) w;
		}
		r = (qu_big) qu_big_mul_big (r, w);
		if (r == 0 || !qu_big_is (r))
			return (qu_real) r;
	}
	return (qu_real) r;
}



qu_real qu_big_pow_int (qu_big x, qu_int y)
{
	if (qu_big_iswhole (x))
	{
		return qu_integer_pow_int (qu_big_as_integer (x), y);
	}
	else
	{
		long v = qu_int_tol (y);

		if (v == 0)
		{
			return (qu_real) QuBig1;
		}
		else if (v == 1)
		{
			return (qu_real) x;
		}
		else if (v == -1)
		{
			if (qu_big_iszero (x))
				return (qu_real) __qu_inf__ (1);

			else if (qu_big_isone (x))
				return qu_big_is1 (x) ? (qu_real) QuInt1 : (qu_real) QuInt_1;

			else
				return qu_big_div_big (QuBig1, x);
		}
		else if (v < 0)
		{
			if (qu_big_iszero (x))
			{
				return (qu_real) __qu_inf__ (1);
			}
			else if (qu_big_isone (x))
			{
				return (qu_big_isneg (x) && qu_int_isodd (y)) ? (qu_real) QuInt_1 : (qu_real) QuInt1;
			}
			else
			{
				qu_real r;

				v = -v;
				r = _big_pow_clong (x, v);
				if (r == 0 || qu_inf_is (r))
					return (qu_real) QuBig0;

				assert (qu_big_is (r));
				x = (qu_big_isneg (x) && (v & 1)) ? QuBig_1 : QuBig1;
				return qu_big_div_big (x, (qu_big) r);
			}
		}
		return _big_pow_clong (x, v);
	}
}



qu_real qu_big_pow_long (qu_big x, qu_long y)
{
	qu_int p = qu_long_as_int (y);
	if (p)
		return qu_big_pow_int (x, p);
	return (qu_real) qu_big_pow_big (x, qu_long_as_big (y));
}



qu_number qu_big_pow_rational (qu_big x, qu_rational y)
{
	if (qu_rational_iswhole (y))
		return (qu_number) qu_big_pow_integer (x, y->num);
	return qu_big_pow_big (x, qu_rational_as_big (y));
}



qu_number qu_big_pow_float (qu_big x, qu_float y)
{
	if (qu_float_iswhole (y))
		return (qu_number) qu_big_pow_integer (x, qu_float_as_integer (y));
	return qu_big_pow_big (x, qu_float_as_big (y));
}



qu_bigpowbig_f __QuBigPowBigMath = NULL;

qu_number qu_big_pow_big (qu_big x, qu_big y)
{
	if (qu_big_iswhole (y))
		return (qu_number) qu_big_pow_integer (x, qu_big_as_integer (y));
	else
	{
		QU_ENSURE (qu_module_math ());
		assert (__QuBigPowBigMath);
		return __QuBigPowBigMath (x, y);
	}
}



qu_real qu_big_pow_inf (qu_big x, qu_inf y)
{
	if (y == QuInfPos)
	{
		if (qu_big_iszero (x))
			return (qu_real) QuBig0;

		else if (qu_big_isone (x))
			return (qu_real) QuBig1;

		else
			return (qu_real) __qu_inf__ (qu_inf_sign (y));
	}
	else
	{
		if (qu_big_iszero (x))
			return (qu_real) __qu_inf__ (1);
		return qu_big_isone (x) ? (qu_real) QuBig1 : (qu_real) QuBig0;
	}
}



QU_NUMBER_BINOP_integer (qu_big, pow, qu_real)
QU_NUMBER_BINOP_decimal (qu_big, pow, qu_number)
QU_NUMBER_BINOP_real    (qu_big, pow, qu_number)
QU_NUMBER_BINOP_number  (qu_big, pow, qu_number)



/*
	** reversed
*/

qu_number qu_big_rpow_int (qu_big x, qu_int y)
{
	if (qu_big_iswhole (x))
		return qu_int_pow_integer (y, qu_big_as_integer (x));
	return qu__big__pow (qu_int_as_big (y), x);
}



qu_number qu_big_rpow_long (qu_big x, qu_long y)
{
	if (qu_big_iswhole (x))
		return (qu_number) qu_long_pow_integer (y, qu_big_as_integer (x));
	return qu__big__pow (qu_long_as_big (y), x);
}



qu_number qu_big_rpow_rational (qu_big x, qu_rational y)
{
	if (qu_big_iswhole (x))
		return (qu_number) qu_rational_pow_integer (y, qu_big_as_integer (x));
	return qu_big_pow_big (qu_rational_as_big (y), x);
}



qu_number qu_big_rpow_float (qu_big x, qu_float y)
{
	return qu_big_pow_big (qu_float_as_big (y), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print -> Null
	Print k[self} to -{Stream.stdout}.

	This simply prints the string returned by the ={__string} method.

	See the builtin -{print} function for possible errors.
*/
static qu_null qu__big__print (qu_big self)
{
	qu_string s = qu_big_as_string (self);
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The meaning of "remainder" may differ with other languages.
	Refer to -{Number.__rem} for details.

	@exc
	.ERem
	.EBrain

	@p
	The remainder of dividing by zero is undefined thus throws ERem.

	This method finds the remainder by scaling the numbers. In rare circumstances
	where the scaled numbers are too big, then EBrain is thrown.

	@xmp
		x = Big ('11')
		y = Big ('3')
		println (x / y)
		println (x \ y)
		println (x % y)
		println (Big ('1e6789') % 3)

	@out
		0.0003666666666666666666666666E+4
		3
		0.0002E+4
		0.0001E+4
*/
static qu_number qu__big__rem (qu_big self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_big_rem_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_big_rem_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_big_rem_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_big_rem_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_rem_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_big_rem_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_big_rem_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuERem);
}



qu_real qu_big_rem_int (qu_big x, qu_int y)
{
	TBigTmpl t;
	return qu_big_rem_big (x, _big_tmp_clong (&t, qu_int_tol (y)));
}



qu_real qu_big_rem_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_rem_big (x, _big_tmp_cdouble (&t, qu_float_tod (y)));
}



qu_real qu_big_rem_long (qu_big x, qu_long y)
{
	return qu_big_rem_big (x, qu_long_as_big (y));
}



qu_real qu_big_rem_rational (qu_big x, qu_rational y)
{
	return qu_big_rem_big (x, qu_rational_as_big (y));
}



qu_real qu_big_rem_big (qu_big x, qu_big y)
{
	int cmp;

	QU_REQUIRE (QuERem, !qu_big_iszero (y));

	cmp = qu_big_cmpabs_big (x, y);

	if (cmp == 0)
		return (qu_real) QuBig0;

	else if (cmp < 0)
		return (qu_real) x;

	else
	{
		/* TODO - there should be a better way of doing this */

		qu_integer a, b;
		long       ex = x->expo;
		long       ey = y->expo;
		long       lx = x->len;
		long       ly = y->len;

		long       l = (lx > ly) ? lx : ly;
		long       e = ex - ey;

		e += l;

		QU_REQUIRE (QuEBrain, x->expo <= QU_BIG_MAXEXPO - e);
		QU_REQUIRE (QuEBrain, y->expo <= QU_BIG_MAXEXPO - e);

		x->expo += e;
		y->expo += e;

		a = qu_big_as_integer (x);
		b = qu_big_as_integer (y);

		x->expo = ex;
		y->expo = ey;

		a = (qu_integer) qu_integer_rem_integer (a, b);
		assert (a && qu_integer_is (a));

		if (qu_integer_iszero (a))
			return (qu_real) QuBig0;
		else
		{
			qu_real r;
			qu_long v;

			QU_REQUIRE (QuEBrain, e < QU_BIG_MAXEXPO / QU_BIG_FIG);
			v = qu_long_pow_10 (e * QU_BIG_FIG);
			r = qu_integer_div_long (a, v);
			assert (r && qu_real_is (r));
			return (qu_real) qu_real_as_big (r);
		}
	}
}



qu_big qu_big_rem_inf (qu_big x, qu_inf y)
{
	QU_REQUIRE (QuERem, !qu_big_iszero (x));
	return x;
}



QU_NUMBER_BINOP_integer (qu_big, rem, qu_real)
QU_NUMBER_BINOP_decimal (qu_big, rem, qu_real)
QU_NUMBER_BINOP_real    (qu_big, rem, qu_real)
QU_NUMBER_BINOP_number  (qu_big, rem, qu_number)



/*
	% reversed
*/

qu_real qu_big_rrem_int (qu_big x, qu_int y)
{
	TBigTmpd t;
	return qu_big_rem_big (_big_tmp_cdouble (&t, qu_int_tod (y)), x);
}



qu_real qu_big_rrem_long (qu_big x, qu_long y)
{
	return qu_big_rem_big (qu_long_as_big (y), x);
}



qu_real qu_big_rrem_rational (qu_big x, qu_rational y)
{
	return qu_big_rem_big (qu_rational_as_big (y), x);
}



qu_real qu_big_rrem_float (qu_big x, qu_float y)
{
	TBigTmpd t;
	return qu_big_rem_big (_big_tmp_cdouble (&t, qu_float_tod (y)), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Get the standard string representation of this Big number.

	This is what the string printed by the builtin -{print} function.
*/
qu_string qu__big__string (qu_big self)
{
	long       l = self->len;
	DIGIT     *d = self->digits;
	char       b [64];
	qu_string  s;

	if (l == 0)
		return qu_string_byte ('0');

	s = qu_string_buffer_new (l * QU_BIG_FIG + 16);

	if (qu_big_isneg (self))
		qu_string_buffer_put (s, "-0.", 3);
	else
		qu_string_buffer_put (s, "0.", 2);

	while (l--)
	{
		sprintf (b, QU_BIG_FORMAT, *d++);
		qu_string_buffer_putz (s, b);
	}

	sprintf (b, "D%+ld", self->expo * QU_BIG_FIG);
	qu_string_buffer_putz (s, b);

	return qu_string_buffer_return (s);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	Return true if the number is non-zero, false otherwise.
*/
static qu_boolean qu__big__true (qu_big self)
{
	return qu_big_iszero (self) ? QuFalse : QuTrue;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub abs -> Big
	Return the absolute value.
*/
qu_big qu__big_abs (qu_big self)
{
	if (qu_big_isneg (self))
		return qu_big_neg (self);
	return self;
}



/*) sub ceil -> Real
	Return the result of rounding this number to the nearest integer towards
	positive infinity.

	@xmp
		x = Big ('-1.2')  println (x.ceil)
		x = Big (' 1.5')  println (x.ceil)

	@out
		-0.0001E+4
		0.0002E+4
*/
qu_real qu__big_ceil (qu_big x)
{
	if (qu_big_iswhole (x))
	{
		return (qu_real) x;
	}
	else if (qu_big_isfrac (x))
	{
		return (qu_real) (qu_big_isneg (x) ? QuBig0 : QuBig1);
	}
	else if (qu_big_isneg (x))
	{
		long   e = x->expo;
		qu_big r = _big_alloc (e);

		r->len = r->expo = e;
		memcpy (r->digits, x->digits, e * sizeof (DIGIT));
		SETNEG (r);
		return (qu_real) r;
	}
	else
	{
		long    e  = x->expo;
		DIGIT  *ds = x->digits;
		qu_big  r;

		assert (e < x->len);
		if (ds [e] < QU_BIG_BASE / 2)
		{
			r = _big_alloc (e);
			r->len = r->expo = e;
			memcpy (r->digits, x->digits, e * sizeof (DIGIT));
			return (qu_real) r;
		}
		else
		{
			DIGIT *dr, carry;
			long   i;

			r  = _big_alloc (e + 1);
			r->len = r->expo = e + 1;
			dr = r->digits;
			memcpy (r->digits + 1, x->digits, e * sizeof (DIGIT));
			for (i = e, carry = 1; i >= 0 && carry; --i)
			{
				dr[i] += carry;
				if (dr[i] >= QU_BIG_BASE)
				{
					dr[i] -= QU_BIG_BASE;
					carry = 1;
				}
				else
				{
					carry = 0;
				}
			}
			dr[i] += carry;
			return _big_return (_big_normalize (r, 0));
		}
	}
}



/*) sub div (x: Real) -> Array
	Internally, performing division of a Big yields the I{residual} at the
	same time. This method might save time and memory if you need the
	result of the division and the residual at the same time.

	Note however that v{x} may not be zero and may not be an infinity
	otherwise the residual can not be determined and so EDiv is thrown.

	Note also that the I{residual} is not the same as the I{remainder}.
	See #{__rem}.

	@exc
	.EDiv
	.ERem
	.ESmall
	.EHuge

	@xmp
		x = Big ('100')
		y = Big ('3')
		(v, r) = x.div (y)
		println (v)
		println (r)
		println (v * y + r)

	@out
		0.0033333333333333333333333333E+4
		0.0001E-16
		0.0100E+4

	@note
	The multiplication of the result with the divisor plus the residual
	should yield self exactly. So this method can also be used to
	determine whether self is an exact multiplication of v{x}.
*/
static qu_array qu__big_div (qu_big self, qu_any x)
{
	qu_big  y = 0;    /* initialized to avoid compiler warning */
	qu_real v, r = 0;

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      y = qu_int_as_big      ((qu_int     ) x); break;
		case QU_CLASS_ID_Long:     y = qu_long_as_big     ((qu_long    ) x); break;
		case QU_CLASS_ID_Rational: y = qu_rational_as_big ((qu_rational) x); break;
		case QU_CLASS_ID_Float:    y = qu_float_as_big    ((qu_float   ) x); break;
		case QU_CLASS_ID_Big:      y = qu_big_as_big      ((qu_big     ) x); break;
		default:
			QU_THROW (QuEDiv);
	}
	QU_REQUIRE (QuERem, !qu_big_iszero (y));
	v = _big_div_res (self, y, &r);
	QU_ENSURE (v);
	return qu_array_new_ex (2, v, r);
}



/*) sub expo (*arg) -> Int
	@def
	sub expo -> Int
	sub expo (Int: value) -> Int
	@p
	Return the exponent of this Big number or get a new Big number by
	changing the exponent of self.

	Note that the exponent is specified in terms of the internal representation,
	that is, it is a multiplication of #{FIG}.

	@exc
	.EArgValue

	@xmp
		b = Big ('123.45')
		println (b)
		println (b.expo)
		println (b.expo (7))

	@out
		0.012345000000E+4
		1
		0.012345000000E+28
*/
static qu_real qu__big_expo (qu_big self, int argc, qu_any *argv)
{
	if (argc == 0)
	{
		long v = self->expo;
		assert (QU_INT_MIN <= v && v <= QU_INT_MAX);
		return (qu_real) qu_int_new (v);
	}
	else
	{
		long e;

		QU_ARG_VMAX (1);
		QU_ARG_vlong (0, e);
		QU_ARG_CHECK (-QU_BIG_MAXEXPO <= e && e <= QU_BIG_MAXEXPO);

		if (qu_big_iszero (self))
		{
			return (qu_real) self;
		}
		else
		{
			qu_big r = qu_big_dup (self);
			r->expo = e;
			return (qu_real) r;
		}
	}
}



/*) sub floor -> Big
	Return the result of rounding this number to the nearest integer
	towards negative infinity.

	@xmp
		x = Big ('-1.5')  println (x.floor)
		x = Big ('1.7')   println (x.floor)

	@out
		-0.0002E+4
		0.0001E+4
*/
qu_real qu__big_floor (qu_big x)
{
	if (qu_big_iswhole (x))
	{
		return (qu_real) x;
	}
	else if (qu_big_isfrac (x))
	{
		return (qu_real) (qu_big_isneg (x) ? QuBig0 : QuBig1);
	}
	else if (qu_big_isneg (x))
	{
		long    e  = x->expo;
		DIGIT  *ds = x->digits;
		qu_big  r;

		assert (e < x->len);
		if (ds [e] < QU_BIG_BASE / 2)
		{
			r = _big_alloc (e);
			SETNEG (r);
			r->len = r->expo = e;
			memcpy (r->digits, x->digits, e * sizeof (DIGIT));
			return (qu_real) r;
		}
		else
		{
			DIGIT *dr, carry;
			long   i;

			r  = _big_alloc (e + 1);
			SETNEG (r);
			r->len = r->expo = e + 1;
			dr = r->digits;
			memcpy (r->digits + 1, x->digits, e * sizeof (DIGIT));
			for (i = e, carry = 1; i >= 0 && carry; --i)
			{
				dr[i] += carry;
				if (dr[i] >= QU_BIG_BASE)
				{
					dr[i] -= QU_BIG_BASE;
					carry = 1;
				}
				else
				{
					carry = 0;
				}
			}
			dr[i] += carry;
			return _big_return (_big_normalize (r, 0));
		}
	}
	else
	{
		long   e = x->expo;
		qu_big r = _big_alloc (e);

		r->len = r->expo = e;
		memcpy (r->digits, x->digits, e * sizeof (DIGIT));
		return (qu_real) r;
	}
}



/*) sub frac -> Big
	Return the fraction part.

	@noexc

	@xmp
		x = 1.23
		println (x.frac)

	@out
		0.23
*/
qu_big qu__big_frac (qu_big x)
{
	long e = x->expo;
	long n = x->len - e;

	if (n <= 0)
	{
		return QuBig0;
	}
	else
	{
		qu_big f = _big_alloc (n);

		f->len = n;
		f->expo = 0;
		memcpy (f->digits, x->digits + e, n * sizeof (DIGIT));
		if (qu_big_isneg (x))
			SETNEG (f);
		return f;
	}
}



/*) sub iseven ->Boolean
	Return true if self is an even number, false otherwise.

	Note that 0 is considered an even number.
*/
static qu_boolean qu__big_iseven (qu_big self)
{
	return qu_big_iswhole (self) ? (qu_big_isodd (self) ? QuFalse : QuTrue) : QuFalse;
}



/*) sub isfrac -> Boolean
	Return true if C{0 < self < 1}, false otherwise.
*/
static qu_boolean qu__big_isfrac (qu_big self)
{
	return qu_big_isfrac (self) ? QuTrue : QuFalse;
}



/*) sub isodd -> Boolean
	Return true if self is an odd number, false otherwise.

	Note that 0 is considered an even number.
*/
static qu_boolean qu__big_isodd (qu_big self)
{
	return qu_big_isodd (self) ? QuTrue : QuFalse;
}



/*) sub iswhole -> Boolean
	Return true if self is a I{whole} number, false otherwise.

	@noexc

	@xmp
		x = Big ('1.23')
		println (x.iswhole)
		x = Big ('1.0')
		println (x.iswhole)

	@out
		false
		true
*/
qu_boolean qu__big_iswhole (qu_big self)
{
	return qu_big_iswhole (self) ? QuTrue : QuFalse;
}



/*) sub prec (*arg)
	@def
	sub prec -> Int
	sub prec (value: Int) -> Big
	@p
	Get the precision of this Big number or get a new Big number by changing
	the precision.

	v{value} must be a number between 2 to 10000.
	See the ={precision} method.

	If the number is zero then self is returned because there is no
	sense to set the precision for 0.

	If the precision is the same, then self is returned.

	@xmp
		b = Big ('123.45')
		println (b.prec)
		println (b)
		a = b.prec (4)
		println (a)

	@out
		2
		0.00012345E+3
		0.0001234500000000E+3
*/
static qu_any qu__big_prec (qu_big self, int argc, qu_any *argv)
{
	if (argc == 0)
	{
		return qu_integer_newl (self->len);
	}
	else
	{
		long   p;
		qu_big r;

		QU_ARG_VMAX (1);
		QU_ARG_vlong (0, p);
		QU_ARG_CHECK (QU_BIG_MINPREC <= p && p <= QU_BIG_MAXPREC);

		if (qu_big_iszero (self) || p == self->len)
			return self;

		r = _big_alloc (p);
		r->len = p;
		r->expo = self->expo;
		memcpy (r->digits, self->digits, p * sizeof (DIGIT));
		return r;
	}
}



/*) sub round (precision: Int) -> Big
	Return the result of rounding this number to the specified
	v{precision} where v{precision} is the number of decimal digits.

	@xmp
		x = Big ('1.23456789')  println (x.round (6))

	@out
		0.000123456800E+4
*/
static qu_big qu__big_round (qu_big self, qu_integer prec)
{
	long p;
	QU_ARG_long (prec, p);
	return qu_big_round (self, p);
}



qu_big qu_big_round (qu_big x, long nd)
{
	DIGIT  *dr, d, v;
	long    k, n, o, l, i;
	qu_big  r;

	if (qu_big_iswhole (x) || nd < 1)
		return x;

	assert (x->len > 0);

	nd += (x->expo + 1) * QU_BIG_FIG;
	k = nd / QU_BIG_FIG;

	if (k < 0)
		return QuBig0;

	if (k > x->len)
		return x;

	r       = _big_alloc (k + 1);
	r->len  = k + 1;
	r->expo = x->expo + 1;
	dr = r->digits;
	memcpy (dr + 1, x->digits, k * sizeof (DIGIT));

	d = dr[k];
	o = nd - k * QU_BIG_FIG;
	n = l = QU_BIG_FIG - o;
	v = 5;
	while (--l > 0)
		v *= 10;

	d += v;
	if (d >= QU_BIG_BASE)
	{
		DIGIT carry = 1;
		for (i = k - 1; i <= 0; i--)
		{
			dr[i] += carry;
			if (dr[i] >= QU_BIG_BASE)
			{
				dr[i] -= QU_BIG_BASE;
				carry = 1;
			}
			else
			{
				carry = 0;
			}
		}
		dr[i] += carry;
	}

	l = n;
	while (l--)
		d /= 10;
	while (n--)
		d *= 10;
	dr[k] = d;

	if (qu_big_isneg (x))
		SETNEG (r);

	r = (qu_big) _big_return (_big_normalize (r, 0));
	QU_REQUIRE (QuEHuge, qu_big_is (r));
	return r;
}



/*) sub sign -> Int
	Return -1, 0, 1 if self is negative, zero and postive respectively.
*/
static qu_int qu__big_sign (qu_big self)
{
	return qu_big_isneg (self) ? QuInt_1 : (qu_big_iszero (self) ? QuInt0 : QuInt1);
}



/*) sub trunc -> Big
	Return the result of rounding this number to the nearest integer
	towards zero, thus removing the fractional part.

	@xmp
		x = -1.01  println (x.trunc)

	@out
		-1
*/
qu_big qu__big_trunc (qu_big x)
{
	if (qu_big_iswhole (x))
	{
		return x;
	}
	else if (qu_big_isfrac (x))
	{
		return QuBig0;
	}
	else
	{
		int    e = x->expo;
		qu_big r = _big_alloc (e);

		r->len = r->expo = e;
		memcpy (r->digits, x->digits, e * sizeof (DIGIT));
		if (qu_big_isneg (x))
			SETNEG (r);
		return r;
	}
}



/*----------------------------------------------------------------------------
	misc
----------------------------------------------------------------------------*/

/* what integer must be multiplied to x to get an integer? */
qu_integer qu_big_integer_factor (qu_big x)
{
	long f = x->len - x->expo;
	return (f < 0)
		? (qu_integer) QuInt1
		: (qu_integer) qu_long_pow_10 (f * QU_BIG_FIG);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_big__setup ()
{
	qu_class_setup_final (QuBigClass, qu_big_s, qu__big__destroy, NULL);

	QuBig0  = _big_static1 (0);
	QuBig1  = _big_static1 (1);
	QuBig2  = _big_static1 (2);
	QuBig3  = _big_static1 (3);
	QuBig4  = _big_static1 (4);
	QuBig5  = _big_static1 (5);
	QuBig6  = _big_static1 (6);
	QuBig7  = _big_static1 (7);
	QuBig8  = _big_static1 (8);
	QuBig9  = _big_static1 (9);
	QuBig10 = _big_static1 (10);
	QuBig16 = _big_static1 (16);
	QuBig_1 = _big_static1 (-1);
	QuBig_2 = _big_static1 (-2);
	QuBig_4 = _big_static1 (-4);
	QuBig_8 = _big_static1 (-8);

	QuBig1_2  = qu_big_new_static ("0.5");
	QuBig_1_2 = qu_big_new_static ("-0.5");
	QuBig1_4  = qu_big_new_static ("0.25");
	QuBig_1_4 = qu_big_new_static ("-0.25");
}



void qu_big__init ()
{
	/* specific class properties */
	qu_class_constz  (QuBigClass, "MINPREC"  , qu_int_new (QU_BIG_MINPREC));
	qu_class_constz  (QuBigClass, "MAXPREC"  , qu_int_new (QU_BIG_MAXPREC));
	qu_class_constz  (QuBigClass, "DIGITSIZE", qu_int_new (sizeof (DIGIT)));
	qu_class_constz  (QuBigClass, "BASE"     , qu_int_new (QU_BIG_BASE));
	qu_class_constz  (QuBigClass, "FIG"      , qu_int_new (QU_BIG_FIG));
	qu_class_constz  (QuBigClass, "MAXEXPO"  , qu_int_new (QU_BIG_MAXEXPO));
	qu_class_constz  (QuBigClass, "MAXLEN"   , qu_int_new (QU_BIG_MAXLEN));

	/* standard class methods */
	qu_class_method_class  (QuBigClass, QU_NAME_new, qu__big_new, -1);

	/* specific class methods */
	qu_class_method_classz (QuBigClass, "rand"     , qu__big_rand     ,  0);
	qu_class_method_classz (QuBigClass, "precision", qu__big_precision, -1);

	/* standard instance methods */
	qu_class_method_inst   (QuBigClass, QU_NAME__as    , qu__big__as    , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__cmp   , qu__big__cmp   , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__decr  , qu__big__decr  , 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__div   , qu__big__div   , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__eq    , qu__big__eq    , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__format, qu__big__format, 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__hash  , qu__big__hash  , 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__idiv  , qu__big__idiv  , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__incr  , qu__big__incr  , 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__len   , qu__big__len   , 0);
	qu_class_alias         (QuBigClass, QU_NAME__len   , QU_NAME_len);
	qu_class_method_inst   (QuBigClass, QU_NAME__like  , qu__big__like  , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__minus , qu__big__minus , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__mul   , qu__big__mul   , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__neg   , qu__big__neg   , 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__plus  , qu__big__plus  , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__pos   , qu__big__pos   , 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__pow   , qu__big__pow   , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__print , qu__big__print , 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__rem   , qu__big__rem   , 1);
	qu_class_method_inst   (QuBigClass, QU_NAME__string, qu__big__string, 0);
	qu_class_method_inst   (QuBigClass, QU_NAME__true  , qu__big__true  , 0);

	/* specific instance methods */
	qu_class_method_instz  (QuBigClass, "abs"    , qu__big_abs    ,  0);
	qu_class_method_instz  (QuBigClass, "ceil"   , qu__big_ceil   ,  0);
	qu_class_method_instz  (QuBigClass, "div"    , qu__big_div    ,  1);
	qu_class_method_instz  (QuBigClass, "expo"   , qu__big_expo   , -1);
	qu_class_method_instz  (QuBigClass, "floor"  , qu__big_floor  ,  0);
	qu_class_method_instz  (QuBigClass, "frac"   , qu__big_frac   ,  0);
	qu_class_method_instz  (QuBigClass, "iseven" , qu__big_iseven ,  0);
	qu_class_method_instz  (QuBigClass, "isfrac" , qu__big_isfrac ,  0);
	qu_class_method_instz  (QuBigClass, "isodd"  , qu__big_isodd  ,  0);
	qu_class_method_instz  (QuBigClass, "iswhole", qu__big_iswhole,  0);
	qu_class_method_instz  (QuBigClass, "prec"   , qu__big_prec   , -1);
	qu_class_method_instz  (QuBigClass, "round"  , qu__big_round  ,  1);
	qu_class_method_instz  (QuBigClass, "sign"   , qu__big_sign   ,  0);
	qu_class_method_instz  (QuBigClass, "trunc"  , qu__big_trunc  ,  0);

	/* finalize declaration */
	qu_class_set_super (QuBigClass, QuDecimalClass);
	qu_class_enddecl   (QuBigClass);
}
