/* Boolean.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Boolean __final
	The only two instances of this class are the system objects, C{true}
	and C{false}. The keywords k{true} and k{false} simply point to those objects.

	All objects have a boolean version of their value commonly known as
	I{the truth value}. To test whether or not an object is C{true} or C{false},
	its f{__true} method is called. The easiest way to do it is to use the
	#{new} method.

	@xmp
		println (Boolean ([1, 2, 3]))
		println (Boolean (''))

	@out
		true
		false

	@p
	@warn
	Qu ignores any exception thrown by the f{__true} method. If the method does
	not return true nor false then true is assumed. That is why you
	should not call the f{__true} method directly. If the method throws an exception,
	for example, then you will have an inconsistent behaviour with the language
	semantics.

	Unlike in some other languages, the Boolean class is not derived from an integer
	class. In Qu, Boolean and -{Integer} are two different animals.

	The number C{1} is I{true} when its truth value is tested, but it is not the same
	object as true so an equality test will fail.

	@p
	All classes that do not define the f{__true} method inherit the one
	defined by the -{Object} class, which happens to return true.

	@xmp
		println (1 == true ? 'different' : 'same')
		println (1 as Boolean == true ? 'yep' : 'nope')
		println (1 ? 'yes' : 'no')
		x = [1, 2, 3]
		y = 0
		x ? println ('x is true')
		x && y ? println ('never')
		x ? println ('yes')
		println (x or y)
		println (x || y)
		println (1 && 2 && 3)
		println (1 || 2 || 3)

	@out
		different
		yep
		yes
		x is true
		yes
		true
		[1, 2, 3]
		3
		1
*/
qu_class QuBooleanClass;

qu_boolean QuTrue;
qu_boolean QuFalse;

qu_string QuTrueSTR;
qu_string QuFalseSTR;


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (x) -> Boolean
	Get the truth value of v{x}.

	This method simply calls the object's f{__true} method via a builtin
	function to ensure that the value is either true or kfalse}.

	@noexc

	@xmp
		b = Boolean (1.23)
*/
static qu_boolean qu__boolean_new (qu_class self, qu_any x)
{
	if (qu_boolean_is (x))
		return x;
	return qu_sys_true (x);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

#ifdef QU_DEBUG_LOCAL
static void qu__boolean__destroy (qu_boolean self)
{
	QU_FATAL ("never");
}
#endif



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Int}
		-{Integer}
		-{Number}  | Return C{0} or C{1}.
	.| -{Long}     | Return C{0} or C{1}.
	.| -{Rational} | Return C{0 _/ 1} or C{1 _/ 1}.
	.| -{Float}
		-{Decimal} | Return C{0.0} or C{1.0}.
	.| -{Big}      | Return C{0.0} or C{1.0}.
	.| -{Complex}  | Return C{(0 + 0j)} or C{(1 + 0j)}.
	.| -{String}   | Delegate to #{__string}.
	.| -{Boolean}  | Return self.

	@exc
	.EArgValue

	@xmp
		a = true
		b = a as Long
		println (b)
		println (b is Long)

	@out
		1
		true
*/
static qu_any qu__boolean__as (qu_boolean self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Integer:
		case QU_CLASS_ID_Number:   return (self == QuTrue) ? QuInt1      : QuInt0;
		case QU_CLASS_ID_Long:     return (self == QuTrue) ? QuLong1     : QuLong0;
		case QU_CLASS_ID_Rational: return (self == QuTrue) ? QuRational1 : QuRational0;
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Decimal:  return (self == QuTrue) ? QuFloat1    : QuFloat0;
		case QU_CLASS_ID_Big:      return (self == QuTrue) ? QuBig1      : QuBig0;
		case QU_CLASS_ID_Complex:  return (self == QuTrue) ? QuComplex10 : QuComplex00;
		case QU_CLASS_ID_String:   return (self == QuTrue) ? QuTrueSTR   : QuFalseSTR;
		case QU_CLASS_ID_Boolean:  return self;
		/* case QU_CLASS_ID_Inf: */
	}
	QU_ARG_THROW;
}



/*) sub __cmp (x) -> Int
	Compare self with v{x}.

	If v{x} is a Boolean then this method compare the objects using this
	rule: false is less then true.

	Otherwise delegate to -{Object.__cmp}.

	@noexc

	@xmp
		println (true <=> false)
		println (true <=> 1)

	@out
		1
		-1
*/
static qu_int qu__boolean__cmp (qu_boolean self, qu_any x)
{
	if (qu_boolean_is (x))
	{
		if (self == x)
			return QuInt0;
		else if (self == QuTrue && x == QuFalse)
			return QuInt1;
		return QuInt_1;
	}
	return qu_object_default_cmp (self, x);
}



/*) sub __eq (x) -> Boolean
	A Boolean is only equal to itself thus thus method returns true if v{x} is
	the same as self, false otherwise.

	@noexc
*/
static qu_boolean qu__boolean__eq (qu_boolean self, qu_any x)
{
	return ((qu_any) self == x) ? QuTrue : QuFalse;
}



/*) sub __hash -> Int
	Return C{1} if self is true and C{0} otherwise.

	@noexc
*/
static qu_int qu__boolean__hash (qu_boolean self)
{
	return (self == QuTrue) ? QuInt1 : QuInt0;
}



/*) sub __like (x) -> Boolean
	This method returns the same result as the #{__eq} method.

	@noexc
*/
static qu_boolean qu__boolean__like (qu_boolean self, qu_any x)
{
	return ((qu_any) self == x) ? QuTrue : QuFalse;
}



/*) sub __print -> Null
	Print the string C{'true'} or C{'false'} to v{stdout}.

	See -{print} for possible exceptions.
*/
static qu_null qu__boolean__print (qu_boolean self)
{
	qu_string s = (self == QuTrue) ? QuTrueSTR : QuFalseSTR;
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/*) sub __string -> String
	Return the string C{'true'} if self is true, C{'false'} otherwise.

	@noexc

	@xmp
		x = true as String
		println (false, ' ', x)

	@out
		false true
*/
static qu_string qu__boolean__string (qu_boolean self)
{
	return (self == QuTrue) ? QuTrueSTR : QuFalseSTR;
}



/*) sub __true -> self
	Return self.

	@noexc
*/
static qu_boolean qu__boolean__true (qu_boolean self)
{
	return self;
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_boolean__setup ()
{

#ifdef QU_DEBUG_LOCAL
	qu_class_setup_final (QuBooleanClass, qu_object_s, qu__boolean__destroy, 0);
#else
	qu_class_setup_final (QuBooleanClass, qu_object_s, 0, 0);
#endif

	QuTrue  = qu_gc_new_static (QuBooleanClass);
	QuFalse = qu_gc_new_static (QuBooleanClass);

	QuTrueSTR  = qu_string_new_staticz ("true");
	QuFalseSTR = qu_string_new_staticz ("false");
}



void qu_boolean__init ()
{
	/* standard class methods */
	qu_class_method_class (QuBooleanClass, QU_NAME_new, qu__boolean_new, 1);

	/* standard instance methods */
	qu_class_method_inst  (QuBooleanClass, QU_NAME__as    , qu__boolean__as    , 1);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__cmp   , qu__boolean__cmp   , 1);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__eq    , qu__boolean__eq    , 1);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__hash  , qu__boolean__hash  , 0);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__like  , qu__boolean__like  , 1);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__print , qu__boolean__print , 0);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__string, qu__boolean__string, 0);
	qu_class_method_inst  (QuBooleanClass, QU_NAME__true  , qu__boolean__true  , 0);

	/* finalize declaration */
	qu_class_enddecl (QuBooleanClass);
}
