/* Buffer.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Buffer __final
	Buffer objects are used to store bytes much like -{String} except that
	you can modify the content.

	@warn
	Buffer objects are not strings.
	They look like strings but they are not.

	Most functions/methods outside of the Buffer class do not accept Buffer
	in place of String.

	@xmp
		b = Buffer (100)

		b <<< 'foo' <<< 'bar'
		println (b)

		b[1, 2] = '--'
		println (b)

		b = Buffer ('/foo/bar.txt')
		f = File.open (b)

	@out
		foobar
		f--bar
		[1024] __main__ (EArgValue) invalid argument

	@p
	Most String-like methods are also provided by the Buffer class.
	The names are the same but now they modify the Buffer.
	So be careful.

	@xmp
		b = Buffer (100)
		b <<< '  foo  '
		b.trim
		println (b.quote)

	@out
		'foo'

	@p
	Initially Buffer objects have a system hard limit.
	You can set this limit with the #{limit} method.
	Exceeding this limit will throw EOverflow.

	@xmp
		b = Buffer (80)
		b.limit = 100
		loop
			b <<< 'hello'
		;;

	@out
		[1024] __main__ (EOverflow) capacity overflow
*/
qu_class QuBufferClass;


/*----------------------------------------------------------------------------
	macros
----------------------------------------------------------------------------*/

/*
	Make sure we can modify the Buffer.
*/
#define ENSURE_NOT_FROZEN(self) \
	QU_REQUIREZ (QuEModify, !qu_flag_test  (self, QU_BUFFER_FLAG_FROZEN | QU_BUFFER_FLAG_PROTECT), \
		"Buffer is frozen or protected ")


/*----------------------------------------------------------------------------
	allocate
----------------------------------------------------------------------------*/

static qu_buffer _buffer_alloc (long cap, int statik)
{
	register qu_buffer b;

	assert (0 <= cap && cap < QU_MEM_MAX);

	cap = qu_container_sizer (cap);

	if (statik)
		b = qu_gc_new_static (QuBufferClass);
	else
		b = qu_gc_new (QuBufferClass);

	b->scope = 0;
	b->len   = 0;
	b->hash  = -1;
	b->cap   = cap;
	b->limit = QU_MEM_MAX;
	b->str   = qu_mem_alloc (cap + 1);
	b->str[0] = '\0';
	return b;
}



/*
	Create a new qu_buffer with a specified capacity.
*/
qu_buffer qu_buffer_alloc (long cap)
{
	return _buffer_alloc (cap, 0);
}


/*
	Create a new static Buffer.
*/
qu_buffer qu_buffer_new_static (long cap)
{
	return _buffer_alloc (cap, 1);
}



/*
	Create a new regular GC'ed qu_string.
*/
qu_buffer qu_buffer_new (const char *z, long len)
{
	qu_buffer b = qu_buffer_alloc (len);
	memcpy (b->str, z, len);
	b->len = len;
	b->str [len] = '\0';
	return b;
}



/*
	Same as qu_buffer_new except that we do the strlen.
*/
qu_buffer qu_buffer_newz (const char *z)
{
	return qu_buffer_new (z, strlen (z));
}



/*
	Same as qu_buffer_new but we do the dirty work to trim the qu_string.
*/
qu_buffer qu_buffer_new_trim (const char *s, long l)
{
	char *e;

	while (l && qu_ascii_isspace (*s))
	{
		--l;
		++s;
	}
	e = (char *) s + l;
	while (l && qu_ascii_isspace (e[-1]))
	{
		--l;
		--e;
	}
	return qu_buffer_new (s, l);
}



/*----------------------------------------------------------------------------
	basic routines
----------------------------------------------------------------------------*/

#define STRZEQ(az, bz, n, nocase) \
	((nocase) ? qu_strz_eqnc (az, bz, n) : qu_strz_eq (az, bz, n))


#define STRZCHR(az, bz, n, nocase) \
	((nocase) ? qu_strz_chrnc (az, bz, n) : qu_strz_chr (az, bz, n))


/*
	Make sure there is enough room for 'more' bytes.
*/
int qu_buffer_make_space (qu_buffer self, long more)
{
	long z;

	QU_REQUIRE (QuEOverflow, more < self->limit - self->len);
	z = qu_container_sizer (self->len + more);
	if (z > self->cap)
	{
		self->str = qu_mem_realloc (self->str, z + 1);
		self->cap = z;
	}
	return 1;
}



/*
	Put a C string.
*/
int qu_buffer_put (qu_buffer self, const char *s, long len)
{
	long c, l = self->len;

	QU_REQUIRE (QuEOverflow, len < self->limit - l);
	c = l + len;
	if (self->cap < c)
	{
		c = qu_container_sizer (c);
		self->str = qu_mem_realloc (self->str, c + 1);
		self->cap = c;
	}
	memcpy (self->str + l, s, len);
	self->len += len;
	self->str[self->len] = '\0';
	return 1;
}



/*
	Same as qu_buffer_put but we do the strlen.
*/
int qu_buffer_putz (qu_buffer self, const char *s)
{
	return qu_buffer_put (self, s, strlen (s));
}



/*
	Put a qu_string or qu_buffer.
*/
int qu_buffer_puts (qu_buffer self, qu_stringy s)
{
	return qu_buffer_put (self, s->str, s->len);
}



/*
	Put one character.
*/
int qu_buffer_putc (qu_buffer self, char c)
{
	return qu_buffer_put (self, &c, 1);
}



/*
	Delete n bytes starting from i where 0 <= i < self->len.
*/
long qu_buffer_del (qu_buffer self, long i, long n)
{
	long l = self->len;

	assert (0 <= i && i < l);
	if (n > l)
		n = l;
	if (n)
	{
		long m = l - i;
		if (n > m)
			n = m;
		memcpy (self->str + i, self->str + i + n, l - i - n + 1);
		self->len -= n;
		self->str[self->len] = '\0';
	}
	return n;
}



/*
	Delete l bytes starting from the start.
*/
long qu_buffer_consume (qu_buffer self, long n)
{
	long l = self->len;
	l -= n;
	if (l <= 0)
	{
		n = l;
		self->len = 0;
		self->str[0] = '\0';
	}
	else
	{
		memcpy (self->str, self->str + n, l + 1);
		self->len = l;
		self->str[l] = '\0';
	}
	return n;
}



/*
	Delete n bytes from head or tail.
*/
void qu_buffer_chop (qu_buffer self, long n)
{
	long l = self->len;

	if (n < 0)
	{
		n = -n;
		if (n < l)
		{
			l -= n;
			memcpy (self->str, self->str + n, l + 1);
		}
	}
	else
	{
		if (n < l)
			l -= n;
	}
	self->len = l;
	self->str [l] = '\0';
}



/*
	Insert string starting from i where 0 <= i < self->len.
*/
int qu_buffer_insert (qu_buffer self, long i, const char *s, long n)
{
	assert (0 <= i && i < self->len);
	if (n)
	{
		QU_ENSURE (qu_buffer_make_space (self, n));
		memmove (self->str + i + n, self->str + i, self->len - i + 1);
		memcpy  (self->str + i, s, n);
		self->len += n;
		assert (self->str[self->len] == '\0');
	}
	return 1;
}



/*
	StartWith/EndWith routines.
*/
typedef struct bufferwith_t bufferwith_t;
struct bufferwith_t
{
	int         nocase;
	const char *cursor;
	long        len;
	qu_stringy  match;
};


static int _buffer_startwith_callback (qu_stringy item, qu_any idx, bufferwith_t *t)
{
	if (qu_stringy_is (item)
		&& item->len <= t->len
		&& STRZEQ (item->str, t->cursor, item->len, t->nocase))
	{
		t->match = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}



static int _buffer_endwith_callback (qu_stringy item, qu_any idx, bufferwith_t *t)
{
	if (qu_stringy_is (item)
		&& item->len
		&& item->len <= t->len
		&& STRZEQ (t->cursor + t->len - item->len, item->str, item->len, t->nocase))
	{
		t->match = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}



static qu_stringy _buffer_startwith (const char *i, long l, int nocase, qu_any what, qu_any *value)
{
	*value = 0;

	if (qu_stringy_is (what))
	{
		qu_stringy x = (qu_stringy) what;

		if (x->len
			&& x->len <= l
			&& STRZEQ (i, x->str, x->len, nocase))
			return what;
	}
	else if (qu_array_is (what))
	{
		long       n = ((qu_array) what)->len;
		qu_any    *p = ((qu_array) what)->objs;
		qu_stringy x;

		while (n--)
		{
			x = (qu_stringy) *p++;
			if (qu_stringy_is (x)
					&& x->len
					&& x->len <= l
					&& STRZEQ (i, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_set_is (what))
	{
		qu_stringy x;
		long       b = 0, j = 0;

		while ((x = (qu_stringy) qu_set_walk ((qu_set) what, &b, &j)) != 0)
		{
			if (qu_stringy_is (x)
				&& x->len
				&& x->len <= l
				&& STRZEQ (i, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_dict_is (what))
	{
		qu_stringy k;
		qu_any     v;
		long       b = 0, j = 0;

		while ((v = qu_dict_walk ((qu_dict) what, (qu_any *) &k, &b, &j)) != 0)
		{
			if (qu_stringy_is (k)
				&& k->len
				&& k->len <= l
				&& STRZEQ (i, k->str, k->len, nocase))
			{
				*value = v;
				return k;
			}
		}
	}
	else
	{
		bufferwith_t t = {nocase, i, l, 0};
		if (qu_iterator_do (what, (qu_itercallback_f) _buffer_startwith_callback, &t) == QU_ITERATOR_CALLBACK_RESULT)
			return t.match;
	}
	return 0;
}



static qu_stringy _buffer_endwith (const char *i, long l, int nocase, qu_any what, qu_any *value)
{
	*value = 0;

	if (qu_stringy_is (what))
	{
		qu_stringy x = (qu_stringy) what;

		if (x->len
				&& x->len <= l
				&& STRZEQ (i + l - x->len, x->str, x->len, nocase))
			return what;
	}
	else if (qu_array_is (what))
	{
		long       n = ((qu_array) what)->len;
		qu_any    *p = ((qu_array) what)->objs;
		qu_stringy x;

		while (n--)
		{
			x = (qu_stringy) *p++;
			if (qu_stringy_is (x)
					&& x->len
					&& x->len <= l
					&& STRZEQ (i + l - x->len, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_set_is (what))
	{
		qu_stringy x;
		long       b = 0, j = 0;

		while ((x = (qu_stringy) qu_set_walk ((qu_set) what, &b, &j)) != 0)
		{
			if (qu_stringy_is (x)
					&& x->len
					&& x->len <= l
					&& STRZEQ (i + l - x->len, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_dict_is (what))
	{
		qu_stringy k;
		qu_any     v;
		long       b = 0, j = 0;

		while ((v = (qu_stringy) qu_dict_walk ((qu_dict) what, (qu_any *) &k, &b, &j)) != 0)
		{
			if (qu_stringy_is (k)
				&& k->len
				&& k->len <= l
				&& STRZEQ (i, k->str, k->len, nocase))
			{
				*value = v;
				return k;
			}
		}
	}
	else
	{
		bufferwith_t t = {nocase, i, l, 0};
		if (qu_iterator_do (what, (qu_itercallback_f) _buffer_endwith_callback, &t) == QU_ITERATOR_CALLBACK_RESULT)
			return t.match;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (x) -> Buffer
	@def
	sub class new (s: String) -> Buffer
	sub class new (size: Integer) -> Buffer
	@p
	Create a new Buffer object.

	The first form creates a new Buffer with the content of the String v{s}.

	The second form creates a new empty Buffer with an initial v{size}.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('abc')
		println (b.quote)

		b = Buffer (100)
		println (b.size)

	@out
		'abc'
		128
*/
static qu_buffer qu__buffer_new (qu_class self, qu_any x)
{
	if (qu_stringy_is (x))
	{
		return qu_buffer_new (((qu_string) x)->str, ((qu_string) x)->len);
	}
	else
	{
		long l;

		QU_ARG_long (x, l);
		QU_MEM_CHECK (0 <= l && l < QU_MEM_MAX);
		return qu_buffer_alloc (l);
	}
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/* none */



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__buffer__destroy (qu_buffer self)
{
	qu_mem_free (self->str);
}



static void qu__buffer__mark (qu_buffer self)
{
	qu_gc_xmark (self->scope);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (c: Class) -> Any
	Get a version of self based on v{x}.

	This works exactly the same as -{String.__as} thus as if you were saying
	C{String (self) as c}.
*/
static qu_any qu__buffer__as (qu_buffer self, qu_class c)
{
	qu_any r;

	__qu_buffer_protect_begin (self);
	r = qu__string__as ((qu_string) self, c);
	__qu_buffer_protect_end;
	QU_ENSURE (r);
	return r;
}



/*) sub __clear -> self
	Clear content.

	@noexc
*/
/*) sub clear -> self
	This is an alias of #{__clear}.
*/
qu_buffer qu__buffer__clear (qu_buffer self)
{
	ENSURE_NOT_FROZEN (self);

	self->len = 0;
	self->str[0] = '\0';
	return self;
}



/*) sub __cmp (x) -> Int
	Compare with another object.

	If v{x} is a String or Buffer then compare bytes, otherwise delegate to
	C{Object.__cmp}.

	@noexc

	@xmp
		a = Buffer ('abcde')
		b = 'bcde'
		println (a <=> b)

	@out
		-1
*/
static qu_int qu__buffer__cmp (qu_buffer self, qu_any x)
{
	if (qu_stringy_is (x))
	{
		int r = strcmp (self->str, ((qu_string) x)->str);
		return (r > 0) ? QuInt1 : ((r < 0) ? QuInt_1 : QuInt0);
	}
	return qu_object_default_cmp (self, x);
}



/*) sub __copy -> Buffer
	Make a copy.

	@noexc

	@xmp
		a = Buffer ('abcde')
		b = a.__copy
		println (b)

	@out
		abcde
*/
/*) sub copy -> Buffer
	This is an alias of #{__copy}.
*/
static qu_buffer qu__buffer__copy (qu_buffer self)
{
	qu_buffer b = qu_buffer_alloc (self->cap);

	memcpy (b->str, self->str, self->len + 1);
	b->len   = self->len;
	b->hash  = self->hash;
	b->limit = self->limit;
	return b;
}



/*) sub __del (x) -> self
	Remove v{x} from the Buffer.

	If v{x} is a String or Buffer then remove them from the Buffer.
	Otherwise v{x} must be an Integer, positive to remove the first v{x} bytes,
	negative to remove the last v{x} bytes.

	@exc
	.EModify
	.EArgValue

	@xmp
		b = Buffer ('foobar')
		b.__del ('o')
		println (b)

	@out
		fbar
*/
static qu_buffer qu__buffer__del (qu_buffer self, qu_any x)
{
	ENSURE_NOT_FROZEN (self);

	if (qu_stringy_is (x))
	{
		if (((qu_stringy) x)->len)
		{
			long  l = self->len;
			char *i = self->str;
			long  n = ((qu_stringy) x)->len;
			char *j = ((qu_stringy) x)->str;
			char  c = j[0];
			char *p = i;

			while (l)
			{
				if (*i == c)
				{
					char *u = i + 1;
					char *v = j + 1;
					long  m = n - 1;

					while (m--)
					{
						if (*u++ != *v++)
							goto next;
					}
					l -= n;
					i += n;
					continue;
				}

			 next:
				*p++ = *i++;
				--l;
			}
			self->len = p - self->str;
			self->str[self->len] = '\0';
		}
	}
	else
	{
		long n = qu_integer_toli (x);

		QU_ARG_CHECK (errno == 0);

		if (n >= 0)
		{
			if (n >= self->len)
				self->len = 0;
			else
			{
				self->len -= n;
				memcpy (self->str, self->str + n, self->len);
			}
		}
		else
		{
			n = -n;
			if (n >= self->len)
				self->len = 0;
			else
				self->len -= n;
		}
		self->str[self->len] = '\0';
	}
	return self;
}



/*) sub __each -> Iterator
	Return a -{Iterator} object to iterate through the bytes.

	The items yielded are String objects. Not Buffer objects.

	@noexc

	@xmp
		x = Buffer ('abcdef')
		for i in x
			print (i, '-')
		;;
		println

	@out
		a-b-c-d-e-f-
*/
static qu_string qu__buffer__next (qu_buffer self, qu_iterator i)
{
	long k = i->ex1++;

	QU_REQUIRE (QuEStop, 0 <= k && k < self->len);
	return qu_string_byte (self->str[k]);
}


static qu_iterator qu__buffer__each (qu_buffer self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, qu__buffer__next);
	return i;
}



/*) sub __eq (x) -> Boolean
	A Buffer is only equal to itself even if the contents are the same.

	Thus this method returns true if v{x} is self, returns false otherwise.

	@noexc

	@xmp
		a = Buffer ('abcde')
		b = Buffer ('abcde')
		println (a == b)

	@out
		false
*/
/* just inherit from qu_object */



/*) sub __getitem (x) -> Buffer
	@def
	sub __getitem (i: Integer) -> Buffer
	sub __getitem (r: Range  ) -> Buffer
	sub __getitem (a: Array  ) -> Buffer
	@p
	Get a byte or slice of this Buffer.

	If v{x} is an integer, get the character at that index.
	In this case v{i} must be a valid index within the string.
	A negative value means counting from the end.

	If v{x} is a -{Range} then collect bytes as specified by the Range.
	If an item in the range is out of bound then it is ignored.
	No exception thrown.

	If v{x} is an -{Array} then collect bytes as specified by the v{indexes}.

	@exc
	.EArgValue
	.EIndex

	@xmp
		s = Buffer ('abcdefghijklmn')
		println (s [1] ' ' s [-2])
		println (s [Range (8, -1, -2)])
		println (s [[1, 3, 10]])

	@out
		b m
		igeca
		['b', 'd', 'k']
*/
static qu_any qu__buffer__getitem (qu_buffer self, qu_any x)
{
	long l = self->len;
	long i = 0;

	/* the odd looking lable 'doint' is intentional since testing qu_int_is is
		faster than qu_classof, and qu_int_tol is faster than qu_integer_toli,
		and qu_int is usually the case */

	if (qu_int_is (x))
	{
		i = qu_int_tol (x);

	  doint:
		QU_ENSURE_INDEX (i, l);
		return qu_buffer_new (self->str + i, 1);
	}
	else if (qu_range_is (x))
	{
		long      n = qu_range_cover ((qu_range) x, l, &i);
		qu_buffer r = qu_buffer_alloc (n);
		long      z = ((qu_range) x)->step;
		char     *q = r->str;
		char     *p = self->str;

		for (; n; i += z, n--)
			*q++ = p[i];

		r->len = n;
		r->str[n] = '\0';
		return r;
	}
	else if (qu_array_is (x))
	{
		long       n = ((qu_array) x)->len;
		qu_any    *o = ((qu_array) x)->objs;
		qu_any     v;
		qu_buffer  r;
		char      *p;

		if (n == 0)
			return QuStringEMPTY;

		if (n == 1)
		{
			v = *o++;
			QU_ARG_long (v, i);
			QU_ENSURE_INDEX (i, l);
			return qu_buffer_new (self->str + i, 1);
		}

		r = qu_buffer_alloc (n);
		p = self->str;
		while (n--)
		{
			v = *o++;
			QU_ARG_long (v, i);
			QU_ENSURE_INDEX (i, l);
			qu_buffer_putc (r, p[i]);
		}
		return r;
	}
	else
	{
		i = qu_integer_toli (x);
		QU_REQUIRE (QuEIndex, errno == 0);
		goto doint;
	}
}



/*) sub __getslice (idx: Integer, x) -> Any
	@def
	sub __getslice (idx: Integer, num: Integer) -> Buffer
	sub __getslice (idx: Integer, defval) -> Any
	@p
	Get at most v{num} characters starting from the specified index v{idx}
	or safely get a byte without throwing EIndex.

	In the first form, v{idx} must be an integer index within this Buffer.
	A negative value means counting from the end.
	If v{num} is negative then get all characters starting from v{idx}.

	The second form returns v{defval} if the index is out of bound,
	otherwise the character at v{idx}.

	@exc
	.EIndex

	@xmp
		s = Buffer ('abcdef')
		println (s [2,])
		println (s [-4, 2])
		println (s [, 4])
		println (s [100, 'x'])

	@out
		cdef
		cd
		abcd
		x
*/
static qu_any qu__buffer__getslice (qu_buffer self, qu_integer idx, qu_any num)
{
	long l = self->len;
	long i, n;

	i = qu_integer_toli (idx);
	QU_REQUIRE (QuEIndex, errno == 0);

	n = qu_integer_toli (num);
	if (errno == EDOM)
	{
		if (i < 0)
			i += l;
		if (i < 0 || i > l)
			return num;
		return qu_buffer_new (self->str + i, 1);
	}
	else
	{
		QU_REQUIRE (QuEIndex, errno == 0);
		QU_ENSURE_INDEX (i, l);
		l -= i;
		if (n < 0 || n > l)
			n = l;
		return qu_buffer_new (self->str + i, n);
	}
}



/*) sub __in (x) -> Boolean
	Return true if this Buffer contains v{x}, return false otherwise.

	If v{x} is not a String or Buffer then false is returned.
	No exception thrown.

	@noexc

	@xmp
		b = Buffer ('abc-foo-def')
		println ('foo' in b)

	@out
		true
*/
static qu_boolean qu__buffer__in (qu_buffer self, qu_any x)
{
	return (qu_stringy_is (x)
			&& qu_strz_str (self->str, self->len, ((qu_string) x)->str, ((qu_string) x)->len))
		? QuTrue : QuFalse;
}



/*) sub __len -> Int
	Return the number of bytes in this Buffer.

	@noexc

	@xmp
		b = Buffer ('foo')
		println (b.__len)

	@out
		3
*/
/*) sub len -> Int
	This is an alias of #{__len}.
*/
static qu_int qu__buffer__len (qu_buffer self)
{
	assert (self->len <= QU_INT_MAX);
	return qu_int_new (self->len);
}



/*) sub __like (x) -> Boolean
	Return true if v{x} is a String or Buffer and its length and characters
	are the same as this Buffer, false otherwise.

	@noexc

	@xmp
		x = Buffer ('abc')
		y = Buffer ('abc')
		z = 'abc'
		println (x == y)
		println (x === z)

	@out
		false
		true
*/
static qu_boolean qu__buffer__like (qu_buffer self, qu_any x)
{
	if (qu_stringy_is (x)
			&& self->len == ((qu_string) x)->len
			&& memcmp (self->str, ((qu_string) x)->str, self->len) == 0)
		return QuTrue;
	return QuFalse;
}



/*) sub __print -> Null
	Print the content to v{stdout}.

	See -{print} for possible exceptions.

	The Buffer is protected while printing.
*/
static qu_null qu__buffer__print (qu_buffer self)
{
	qu_any r;

  __qu_buffer_protect_begin (self);
	r = qu_sys_printz (self->str, self->len);
  __qu_buffer_protect_end;
	QU_ENSURE (r);
	return QuNil;
}



/*) sub __pop -> String
	Return the content of the Buffer as a String and empty the Buffer.

	@exc
	.EModify

	@xmp
		b = Buffer (100)
		b <<< 'foo'
		b >>> x              # same as: x = String (b); b.clear
		println (b.len)
		println (x)

	@out
		0
		foo
*/
static qu_string qu__buffer__pop (qu_buffer self)
{
	ENSURE_NOT_FROZEN (self);
	/*begin*/
	{
		qu_string s = qu_string_new (self->str, self->len);
		self->len = 0;
		return s;
	}
}



/*) sub __push (x) -> self
	Append v{x} to the end of the Buffer.

	v{x} may be anything. If it is not a String then its __string method is called.

	@exc
	.EArgValue
	.EModify
	.EOverflow

	@xmp
		b = Buffer (100)
		b <<< 'foo' <<< 'bar'
		println (b)

	@out
		foobar
*/
static qu_buffer qu__buffer__push (qu_buffer self, qu_any x)
{
	ENSURE_NOT_FROZEN (self);

	if (!qu_stringy_is (x))
		x = qu_sys_string (x);
	QU_ENSURE (qu_buffer_puts (self, (qu_stringy) x));
	return self;
}



/*) sub __setitem (index: Integer, x)
	Substitute the content starting at the specified v{index} with v{s}.

	A negative value means counting from the end.

	v{x} must be a String or Buffer.
	If it is empty then this means to delete one byte.

	@warn
	The result of this method v{x}. Not self.

	@exc
	.EModify
	.EArgValue
	.EIndex

	@xmp
		b = Buffer (16)
		b <<< 'foobar'
		b[2] = 'XXX'
		println (b)

	@out
		foXXXbar
*/
qu_stringy qu__buffer__setitem (qu_buffer self, qu_integer idx, qu_stringy x)
{
	long l = self->len;
	long n;
	long i;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_long (idx, i);
	QU_ARG_TYPE (x, qu_stringy);

	QU_ENSURE_INDEX (i, l);

	n = x->len;
	if (l == 0)
	{
		if (n)
			QU_ENSURE (qu_buffer_put (self, x->str, n));
	}
	else if (n == 0)
	{
		qu_buffer_del (self, i, 1);
	}
	else if (n == 1)
	{
		self->str [i] = x->str[0];
	}
	else
	{
		QU_ENSURE (qu_buffer_insert (self, i + 1, x->str + 1, n - 1));
		self->str [i] = x->str[0];
	}
	return x;
}



/*) sub __setslice (index: Integer, len: Integer, s)
	Substitute the content starting at the specified slice with v{s}, then
	return v{s}.

	A negative index means counting from the end.
	A negative length means the whole content starting from v{index}.
	A zero length does nothing.

	v{s} must be a String or Buffer.
	If v{s} is empty then this means to delete v{len} bytes starting from v{index}.

	@warn
	The result of this method v{x}. Not self.

	@exc
	.EModify
	.EArgValue
	.EIndex
	.EOverflow

	@xmp
		b = Buffer (16)
		b <<< 'foobar'
		b[2, 3] = 'XXX'
		println (b)

	@out
		foXXXr
*/
qu_stringy qu__buffer__setslice (qu_buffer self, qu_integer ii, qu_integer nn, qu_stringy x)
{
	long l = self->len;
	long n;
	long i;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_long (ii, i);
	QU_ARG_long (nn, n);
	QU_ARG_TYPE (x, qu_stringy);

	QU_ENSURE_INDEX (i, l);

	if (n == 0)
		return x;

	if (n < 0)
		n = l - i;

	if (x->len == 0)
	{
		qu_buffer_del (self, i, n);
	}
	else if (x->len == n && i + n < self->cap)
	{
		char *b = self->str + i;
		char *z = x->str;

		while (n--)
			*b++ = *z++;
	}
	else
	{
		long r = l - n;
		QU_REQUIRE (QuEOverflow, r <= QU_MEM_MAX - x->len);
		r += x->len;
		qu_buffer_del (self, i, n);
		if (i >= self->len)
			qu_buffer_put (self, x->str, x->len);
		else
			qu_buffer_insert (self, i, x->str, x->len);
	}
	return x;
}



/*) sub __string -> self
	This method returns a String with the same content as self.

	@noexc
*/
static qu_string qu__buffer__string (qu_buffer self)
{
	return qu_string_new (self->str, self->len);
}



/*) sub __true -> Boolean
	Return true if this Buffer is not empty, false otherwise.

	@noexc

	@xmp
		s = Buffer (100)
		println (s ? 'yep' : 'nope')

	@out
		nope
*/
static qu_boolean qu__buffer__true (qu_buffer self)
{
	return self->len ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods - modifying
----------------------------------------------------------------------------*/

/*) sub capitalize -> self
	Capitalize the first non-space character.

	@exc
	.EModify

	@xmp
		b = Buffer ('  abc')
		b.capitalize
		println (b.quote)

	@out
		'   Abc'
*/
static qu_buffer qu__buffer_capitalize (qu_buffer self)
{
	long  l = self->len;
	char *p = self->str;

	ENSURE_NOT_FROZEN (self);

	while (l && qu_ascii_isspace (*p))
	{
		l--;
		p++;
	}
	if (l)
		*p = qu_ascii_toupper (*p);
	return self;
}



/*) sub chop (*arg) -> self
	@def
	sub chop -> self
	sub chop (idx: Integer, num: Integer) ->self
	sub chop (num: Integer) -> self
	@p
	Chop one or more characters.

	The first form chops any single last character.

	The second form takes a starting index and the number of bytes to chop.

	The third form takes a number of bytes to chop.
	A positive value means to chop bytes from the end while a negative value
	means to chop bytes from the start.

	@exc
	.EModify
	.EArgValue
	.EIndex

	@xmp
		b = Buffer ('fooxyz.')
		b.chop
		println (b)

		b.chop (2, 3)
		println (b)

	@out
		fooxyz
		foz
*/
static qu_buffer qu__buffer_chop (qu_buffer self, int argc, qu_any *argv)
{
	ENSURE_NOT_FROZEN (self);

	if (argc == 0)
	{
		qu_buffer_chop (self, 1);
	}
	else
	{
		qu_any x = argv[0];

		if (argc == 1)
		{
			long n;

			QU_ARG_long (x, n);
			qu_buffer_chop (self, n);
		}
		else
		{
			long i, n, l = self->len;

			QU_ARG_VEXACT (2);
			QU_ARG_long  (x, i);
			QU_ARG_vlong (1, n);
			QU_ARG_CHECK  (n >= 0);
			QU_ENSURE_INDEX (i, l);
			qu_buffer_del (self, i, n);
		}
	}
	return self;
}



/*) sub chomp (all = false) -> self
	Remove one ore all linbreaks at the end.

	The linebreak is either C{'\n'} or C{'\r\n'}.

	@exc
	.EModify
	.EArgValue

	@xmp
		b = Buffer ('123\n\r\n')
		b.chomp
		println (b.quote)

		b = Buffer ('123\n\n')
		b.chomp (true)
		println (b.quote)

	@out
		'123\n'
		123
*/
static qu_buffer qu__buffer_chomp (qu_buffer self, int argc, qu_any *argv)
{
	int   all = 0;
	long  l = self->len;
	char *s = self->str + l - 1;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, all);

	while (l && *s == '\n')
	{
		--l;
		--s;
		if (l && *s == '\r')
		{
			--l;
			--s;
		}
		if (!all)
			break;
	}
	self->len = l;
	self->str[l] = '\0';
	return self;
}



/*) sub consume (len = 1) -> Int
	Remove the first v{len} bytes.

	Return the actual number of bytes removed.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('foobar')
		println (b.consume (3))
		println (b)

	@out
		3
		bar
*/
static qu_int qu__buffer_consume (qu_buffer self, qu_integer nn)
{
	long l = self->len, n;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_long (nn, n);
	QU_ARG_CHECK (n >= 0);
	if (n < 1 || l == 0)
		return QuInt0;
	if (n >= l)
		n = l;
	self->len -= n;
	memcpy (self->str, self->str + n, self->len + 1);
	assert (self->str[self->len] == '\0');
	return qu_int_new (n);
}



/*) sub exchange (dst: Buffer) -> self
	Exchange content of self with v{dst}.

	Note that #{limit} is exchanged as well.

	@exc
	.EModify
	.EArgValue
*/
static qu_buffer qu__buffer_exchange (qu_buffer self, qu_buffer x)
{
	char *str;
	long  len, cap, lim, hsh;

	QU_ARG_TYPE (x, qu_buffer);

	ENSURE_NOT_FROZEN (self);
	ENSURE_NOT_FROZEN (x);

	str = self->str;
	len = self->len;
	hsh = self->hash;
	cap = self->cap;
	lim = self->limit;

	self->str   = x->str;
	self->len   = x->len;
	self->hash  = x->hash;
	self->cap   = x->cap;
	self->limit = x->limit;

	x->str   = str;
	x->len   = len;
	x->hash  = hsh;
	x->cap   = cap;
	x->limit = lim;

	return self;
}



/*) sub fixline (rep = '\n') -> self
	Replace all linebreak characters with v{rep}.

	All sequences of C{"\n"} and C{"\r\n"} are replaced.

	This method is more efficient than #{replace}.

	If v{rep} is an empty String then all linebreaks are removed.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('foo\nbar\rbaz\r\nfox')
		b.fixline
		println (b.quote)

	@out
		'foo\nbar\rbaz\nfox'
*/
static qu_buffer qu__buffer_fixline (qu_buffer self, int argc, qu_any *argv)
{
	qu_string rep = qu_string_byte ('\n');
	long      l = self->len;
	char     *s = self->str;
	long      lr;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_VTYPE (0, rep, qu_string);

	lr = rep->len;
	if (lr == 0)
	{
		char *p = s;

		while (l)
		{
			if (*s == '\n')
			{
				--l;
				++s;
			}
			else if (s[0] == '\r' && s[1] == '\n')
			{
				assert (l >= 2);
				l -= 2;
				s += 2;
			}
			else
			{
				--l;
				*p++ = *s++;
			}
		}
		self->len = p - self->str;
		self->str [self->len] = '\0';
	}
	else
	{
		if (qu_strz_chr (self->str, self->len, '\n'))
		{
			qu_buffer r = qu_buffer_alloc (self->len);
			while (l)
			{
				if (*s == '\n')
				{
					QU_ENSURE (qu_buffer_puts (r, rep));
					++s;
					--l;
				}
				else if (s[0] == '\r' && s[1] == '\n')
				{
					assert (l >= 2);
					l -= 2;
					s += 2;
					QU_ENSURE (qu_buffer_puts (r, rep));
				}
				else
				{
					QU_ENSURE (qu_buffer_putc (r, *s));
					--l;
					++s;
				}
			}
			qu_buffer_exchange_tmp (self, r);
		}
	}
	return self;
}



/*) sub flatten -> self
	Convert linebreaks into spaces.

	@exc
	.EModify

	@xmp
		b = Buffer ('1\n \r\n23')
		b.flatten
		println (b.quote)

	@out
		'1  \r 23'
*/
qu_buffer qu__buffer_flatten (qu_buffer self)
{
	char *s = self->str;
	long  l = self->len;

	ENSURE_NOT_FROZEN (self);

	while (l--)
	{
		if (*s == '\n')
			*s = ' ';
		++s;
	}
	return self;
}



/*) sub get (len = self.len) -> String
	Get at most v{len} bytes from the Buffer into a String.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('bar')
		s = b.get (2)
		println (s)
		println (b)

	@out
		ba
		r
*/
static qu_string qu__buffer_get (qu_buffer self, int argc, qu_any *argv)
{
	long      n = self->len;
	qu_string r;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		QU_ARG_vlong (0, n);
		QU_ARG_CHECK (n >= 0);
	}
	if (n == 0)
		return QuStringEMPTY;
	if (n >= self->len)
	{
		r = qu_string_new (self->str, self->len);
		self->len = 0;
	}
	else
	{
		r = qu_string_new (self->str, n);
		self->len -= n;
		memcpy (self->str, self->str + n, self->len);
	}
	self->str[self->len] = '\0';
	return r;
}



/*) sub ltrim -> self
	Remove leading spaces.

	@noexc

	@xmp
		b = Buffer ('  123 ')
		b.ltrim
		println (b.quote)

	@out
		'123 '
*/
static qu_buffer qu__buffer_ltrim (qu_buffer self)
{
	long  l = self->len;
	char *i = self->str;

	ENSURE_NOT_FROZEN (self);

	while (l && qu_ascii_isspace (*i))
	{
		--l;
		++i;
	}
	self->len = l;
	memcpy (self->str, i, l + 1);
	return self;
}



/*) sub makespace (more: Integer) -> self
	Ensure the Buffer has enough room for v{more} bytes.

	v{more} must be a non-negative integer.

	@exc
	.EArgValue
	.EModify
	.EOverflow

	@xmp
		b = Buffer ('123')
		b.makespace (10)
*/
static qu_buffer qu__buffer_makespace (qu_buffer self, qu_integer n)
{
	long more;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_long (n, more);
	QU_ARG_CHECK (n >= 0);
	QU_ENSURE (qu_buffer_make_space (self, more));
	return self;
}



/*) sub indent (num: Integer = 1, tab = '\t') -> self
	Indent the content.

	v{tab} must not be empty and v{num} must be a non-negative integer.

	@exc
	.EModify
	.EOverflow
	.EArgValue

	@p
	@code
		| xxxxxxxx      |    xxxxxxxx
		|   xxxxx   ->  |      xxxxx
		|xxxx           |   xxxx
*/
static qu_buffer qu__buffer_indent (qu_buffer self, int argc, qu_any *argv)
{
	char  tab = '\t';
	long  num = 1;
	char *i   = self->str;
	long  l   = self->len;
	long  m   = l;
	long  n;
	char *p;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (0, 2);
	if (argc)
	{
		QU_ARG_vlong (0, num);
		QU_ARG_CHECK  (num >= 0);
		if (argc > 1)
		{
			qu_stringy t = argv[1];
			QU_ARG_CHECK (qu_stringy_is (t) && t->len);
			tab = t->str[0];
		}
	}

	if (num == 0)
		return self;

	i = p = self->str;
	l = self->len;
	n = 0;
	m = QU_MEM_MAX - l;
	while (l)
	{
		if (*i == '\n')
		{
			--l;
			++i;
		}
		else
		{
			QU_REQUIRE (QuEOverflow, n < m);
			n += num;
			m -= num;
			while (l && *i != '\n')
			{
				--l;
				++i;
			}
		}
	}

	if (n == 0)
		return self;

	QU_ENSURE (qu_buffer_make_space (self, n));

	l = self->len;
	i = self->str + l - 1;
	p = i + n;
	p[1] = '\0';
	self->len = l + n;

	while (l)
	{
		while (l && *i == '\n')
		{
			--l;
			*p-- = *i--;
		}
		while (l && *i != '\n')
		{
			--l;
			*p-- = *i--;
		}
		n = num;
		while (n--)
			*p-- = tab;
		if (l)
		{
			--l;
			--i;
			*p-- = '\n';
		}
	}
	return self;
}



/*) sub inject (*arg) -> self
	Insert at the front of the Buffer.

	v{arg} must be a String or Buffer items.

	@exc
	.EModify
	.EArgValue

	@xmp
		b = Buffer ('foo')
		s = 'baz'
		b.inject (s, 'bar')
		println (b)

	@out
		bazbarfoo
*/
static qu_buffer qu__buffer_inject (qu_buffer self, int argc, qu_any *argv)
{
	int        n = argc;
	qu_any    *p = argv;
	qu_stringy x;
	long       m = 0;

	ENSURE_NOT_FROZEN (self);

	while (n--)
	{
		x = (qu_stringy) *p++;
		QU_ARG_TYPE (x, qu_stringy);
		m += x->len;
	}
	if (m)
	{
		char *z;

		QU_ENSURE (qu_buffer_make_space (self, m));
		z = self->str;
		memmove (z + m, z, self->len + 1);
		while (argc--)
		{
			x = (qu_stringy) *argv++;
			memcpy (z, x->str, x->len);
			z += x->len;
		}
		self->len += m;
	}
	return self;
}



/*) sub insert (x, idx = 0) -> self
	Insert v{x} starting from v{idx}.

	v{x} must be a String or Buffer.

	@exc
	.EModify
	.EArgValue
	.EIndex

	@xmp
		b = Buffer ('foo')
		b.insert ('bar', 1)
		println (b)

	@out
		fbaroo
*/
static qu_buffer qu__buffer_insert (qu_buffer self, int argc, qu_any *argv)
{
	qu_stringy x;
	long       i = 0;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, x, qu_stringy);
	if (argc > 1)
	{
		QU_ARG_vlong (1, i);
		QU_ENSURE_INDEX (i, self->len);
	}
	QU_ENSURE (qu_buffer_insert (self, i, x->str, x->len));
	return self;
}



/*) sub push (*arg) -> self
	Append to the end of Buffer.

	v{arg} must be String or Buffer items.

	@exc
	.EModify
	.EArgValue

	@xmp
		b = Buffer ('foo')
		s = 'baz'
		b.push (s, 'bar')
		println (b)

	@out
		foobazbar
*/
static qu_buffer qu__buffer_push (qu_buffer self, int argc, qu_any *argv)
{
	int        n = argc;
	qu_any    *p = argv;
	qu_stringy x;
	long       m = 0;

	ENSURE_NOT_FROZEN (self);

	while (n--)
	{
		x = (qu_stringy) *p++;
		QU_ARG_TYPE (x, qu_stringy);
		m += x->len;
	}
	if (m)
	{
		char *z;

		QU_ENSURE (qu_buffer_make_space (self, m));
		z = self->str + self->len;
		self->len += m;
		while (argc--)
		{
			x = (qu_stringy) *argv++;
			memcpy (z, x->str, x->len);
			z += x->len;
		}
		*z = '\0';
	}
	return self;
}



/*) sub put (x, ofs = 0, len = x.len) -> Int
	Append v{x} the end of Buffer.

	v{x} must be a String or Buffer.

	Return the actual number of bytes put on success.

	@exc
	.EModify
	.EArgValue

	@xmp
		b = Buffer ('foo')
		println (b.put ('bar', 1, 10))
		println (b)

	@out
		2
		fooar
*/
static qu_int qu__buffer_put (qu_buffer self, int argc, qu_any *argv)
{
	long       ofs = 0;
	long       len = 0, max;
	qu_stringy x;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, x, qu_stringy);
	len = x->len;
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
		{
			QU_ARG_vlong (2, len);
			QU_ARG_CHECK (len >= 0);
		}
	}
	QU_ENSURE_INDEX (ofs, x->len);
	max = x->len - ofs;
	if (len > max)
		len = max;
	if (len)
		QU_ENSURE (qu_buffer_put (self, x->str + ofs, len));
	return qu_int_new (len);
}



/*) sub replace (*arg) -> self
	@def
	sub replace (what, rep = '', max = -1, nocase = false) -> self
	sub replace (what: Dict, max = -1, nocase = false) -> self
	sub replace (pattern: Regex, *arg) -> self
	@p
	Replace parts specified by v{what}.

	The last form simply calls C{pattern.replace (self, *arg)}.

	v{what} may be a String or Buffer or an Array, Set or Dict that yields
	String or Buffer objects. In case of a Dict then the keys are used and
	the values ignored. Non String/Buffer items are ignored.

	v{rep} is the replacement String/Buffer.

	v{max} is the maximum number of replacements to perform.
	A non-positive value means no limit.

	v{nocase} is true to do a case insensitive match.

	@exc
	.EArgValue
	.EOverflow

	@xmp
		b = Buffer ('foo Foo BAR foo bar')
		b.replace (['foo', 'bar'], 'X', 3, true)
		println (b)

	@out
		X X X foo bar
*/
static qu_buffer qu__buffer_replace (qu_buffer self, int argc, qu_any *argv)
{
	qu_any     what;
	qu_stringy with = QuStringEMPTY;
	long       max = -1;
	int        nocase = 0;
	long       l;
	char      *i;
	qu_buffer  r;
	qu_any     v;
	int        err = 0;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (1, 4);
	what = argv[0];

	if (qu_classid (what) == QU_CLASS_ID_Regex)
	{
		static qu_string name = 0;
		qu_any res;

		if (name == 0)
			name = qu_string_new_staticz ("replace");

		if (argc == 1)
			res = qu_object_call1 (what, name, self);
		else if (argc == 2)
			res = qu_object_call2 (what, name, self, argv[1]);
		else if (argc == 3)
			res = qu_object_call3 (what, name, self, argv[1], argv[2]);
		else
			res = qu_object_call4 (what, name, self, argv[1], argv[2], argv[3]);

		/* this should never happen */
		QU_ENSURE (res);
		QU_REQUIRE (QuEImplement, res == (qu_any) self);

		return self;
	}

	if (argc > 1)
	{
		QU_ARG_VTYPE (1, with, qu_string);
		if (argc > 2)
		{
			QU_ARG_vlong (2, max);
			if (max <= 0)
				max = -1;
			if (argc > 3)
				QU_ARG_vbool (3, nocase);
		}
	}

	l = self->len;
	i = self->str;
	r = qu_buffer_alloc (8);
  __qu_buffer_protect_begin (self);
	while (l && max)
	{
		qu_stringy x = _buffer_startwith (i, l, nocase, what, &v);
		if (x)
		{
			--max;
			if (v)
			{
				if (!qu_stringy_is (v))
					v = qu_sys_string (v);
			}
			if (!qu_buffer_puts (r, v ? (qu_stringy) v : with))
			{
				err = 1;
				goto done;
			}
			l -= x->len;
			i += x->len;
		}
		else
		{
			if (!qu_buffer_putc (r, *i))
			{
				err = 1;
				goto done;
			}
			--l;
			++i;
		}
	}
	if (l)
	{
		if (!qu_buffer_put (r, i, l))
			err = 1;
	}

  done:
  __qu_buffer_protect_end;

	QU_ENSURE (err == 0);
	qu_buffer_exchange_tmp (self, r);
	return self;
}



/*) sub reverse -> self
	Reverse the bytes.

	@exc
	.EModify

	@xmp
		b = Buffer ('abcde')
		b.reverse
		println (b)

	@out
		edcba
*/
qu_buffer qu__buffer_reverse (qu_buffer self)
{
	long  l = self->len;
	char *s = self->str;
	char *p = s + l - 1;
	char  c;

	ENSURE_NOT_FROZEN (self);

	l /= 2;
	while (l--)
	{
		c = *s;
		*s++ = *p;
		*p-- = c;
	}
	return self;
}



/*) sub rtrim -> self
	Remove trailing spaces.

	@noexc

	@xmp
		x = Buffer (' 123 \r\t\v\f\n ')
		x.rtrim
		println (x.quote)

	@out
		' 123'
*/
static qu_buffer qu__buffer_rtrim (qu_buffer self)
{
	long  l = self->len;
	char *s = self->str + l - 1;

	ENSURE_NOT_FROZEN (self);

	while (l && qu_ascii_isspace (*s))
	{
		--l;
		--s;
	}
	self->len = l;
	*++s = '\0';
	return self;
}



/*) sub scrap (x, translate: Dict = const [:], nocase = false) -> Int __public
	Remove consecutive repetition of characters in v{x}.

	Optionally v{translate} the characters specified by a Dict before the
	equality test.

	@exc
	.EModify
	.EArgValue

	@xmp
		d =
		[
			'\t': ' '
			'\r': ' '
			';' : ','
		]
		b = Buffer ('a,,;  \t   b///c')
		b.scrap (' ,/', d)
		println (b)

	@out
		a, b/c
*/
qu_buffer qu__buffer_scrap (qu_buffer self, int argc, qu_any *argv)
{
	qu_stringy cset    = qu_string_byte (' ');
	qu_dict    trans   = 0;
	int        nocase  = 0;
	char       current = 0;
	int        matched = 0;
	char      *s, *q, c;
	long       l;
	qu_stringy x;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (0, 3);
	if (argc > 0)
	{
		QU_ARG_VTYPE (0, cset, qu_string);
		if (argc > 1)
		{
			QU_ARG_VTYPE (1, trans, qu_dict);
			if (trans->len == 0)
				trans = 0;
			if (argc > 2)
				QU_ARG_vbool (2, nocase);
		}
	}

  __qu_buffer_protect_begin (self);
	q = s = self->str;
	l = self->len;
	while (l--)
	{
		c = *s++;
		if (trans)
		{
			x = qu_dict_getitem (trans, qu_string_byte (c));
			if (x && qu_stringy_is (x) && x->len)
			{
				c = x->str[0];
				if (nocase)
					c = qu_ascii_toupper (c);
			}
		}
		if (matched && (c == current || (nocase && c == qu_ascii_toupper (current))))
		{
			continue;
		}
		else
		{
			if (STRZCHR (cset->str, cset->len, c, nocase))
			{
				current = c;
				matched = 1;
			}
			else
				matched = 0;
			*q++ = c;
		}
	}
	self->len = q - self->str;
	self->str[self->len] = '\0';
  __qu_buffer_protect_end;
	return self;
}



/*) sub title -> self
	Capitalize first letter of words.

	@noexc

	@xmp
		b = Buffer ('a boy-and a girl')
		b.title
		println (b)

	@out
		A Boy-And A Girl
*/
static qu_buffer qu__buffer_title (qu_buffer self)
{
	char  *s = self->str;
	long   l = self->len;

	ENSURE_NOT_FROZEN (self);

	while (l--)
	{
		if (!qu_ascii_isalpha (*s))
		{
			s++;
		}
		else
		{
			*s = qu_ascii_toupper (*s);
			++s;
			while (l && !qu_ascii_isspace (*s))
			{
				--l;
				++s;
			}
		}
	}
	return self;
}




/*) sub transfer (buf: Buffer, len = self.len) -> Int
	Transfer content to another Buffer.

	v{buf} may not be the same as self.

	Return the number of bytes actually transferred.

	@exc
	.EModify
	.EArgValue
	.EOverflow

	@xmp
		x = Buffer (100)
		y = Buffer (100)
		x <<< 'foobar'
		println (x.quote)
		println (y.quote)
		x.transfer (y)
		println (x.quote)
		println (y.quote)

	@out
		'foobar'
		''
		''
		'foobar'
*/
static qu_int qu__buffer_transfer (qu_buffer self, int argc, qu_any *argv)
{
	qu_buffer d;
	long   l = self->len;
	long   n = l;
	long   m;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, d, qu_buffer);
	QU_ARG_CHECK (d != self);

	ENSURE_NOT_FROZEN (d);

	if (argc > 1)
	{
		QU_ARG_vlong (1, n);
		QU_ARG_CHECK (n >= 0);
	}

	if (l == 0 || n == 0)
		return QuInt0;

	if (n > l)
		n = l;

	m = d->cap - d->len;
	QU_ENSURE (qu_buffer_put (d, self->str, n));
	qu_buffer_consume (self, n);
	return qu_int_new (n);
}



/*) sub trim -> self
	Removing leading and trailing spaces.
	Thus combines #{ltrim} and #{rtrim}.

	@noexc

	@xmp
		b = Buffer (' 123  ')
		b.trim
		println (b.quote)

	@out
		'123'
*/
static qu_buffer qu__buffer_trim (qu_buffer self)
{
	ENSURE_NOT_FROZEN (self);

	qu__buffer_ltrim (self);
	qu__buffer_rtrim (self);
	return self;
}



/*) sub unfold -> self
	Unfold C{'\r\n'} continuations.

	This method is useful for things such as HTTP headers.

	@exc
	.EModify

	@xmp
		b = Buffer ('foo\r\n  \t bar')
		b.unfold
		println (b.quote)

	@out
		'foo bar'
*/
static qu_buffer qu__buffer_unfold (qu_buffer self)
{
	char *i = self->str;
	long  l = self->len;
	char *p = i;

	ENSURE_NOT_FROZEN (self);

	while (l)
	{
		if (l > 2
			&& i[0] == '\r'
			&& i[1] == '\n'
			&& (    i[2] == ' '
				||  i[2] == '\t'
				||  i[2] == '\v'
				||  i[2] == '\f'
				||  i[2] == '\n'
				|| (i[2] == '\r' && l > 3 && i[3] != '\n')))
		{
			*p++ = ' ';
			i += 2;
			l -= 2;
			while (l
				&& (    i[0] == ' '
					||  i[0] == '\t'
					||  i[0] == '\v'
					||  i[0] == '\f'
					||  i[0] == '\n'
					|| (i[0] == '\r' && l > 1 && i[1] != '\n')))
			{
				--l;
				++i;
			}
		}
		else
		{
			--l;
			*p++ = *i++;
		}
	}

	*p = '\0';
	self->len = p - self->str;
	return self;
}



/*) sub unindent (max = -1, char = '\t') -> self
	Unindent string. v{char} must not be empty and v{max} must be a non-negative
	integer.

	@exc
	.EArgValue

	@p
	@code
		|		xxxxxxxx        |	xxxxxxxx
		|			xxxxx   ->  |		xxxxx
		|	xxxx                |xxxx
*/
static qu_buffer qu__buffer_unindent (qu_buffer self, int argc, qu_any *argv)
{
	char  tab = '\t';
	long  max = LONG_MAX;
	long  ind = LONG_MAX;
	char *i   = self->str;
	long  l   = self->len;
	long  n;
	char *p, *k;

	ENSURE_NOT_FROZEN (self);

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, max);
		if (max < 0)
			max = LONG_MAX;

		if (argc > 1)
		{
			qu_stringy t = argv[1];
			QU_ARG_CHECK (qu_stringy_is (t) && t->len);
			tab = t->str[0];
		}
	}

	i = p = self->str;
	l = self->len;
	ind = LONG_MAX;
	while (l)
	{
		if (*i == '\n')
		{
			--l;
			++i;
		}
		else
		{
			n = 0;
			while (l && *i == tab && n < max)
			{
				--l;
				++i;
				++n;
			}
			if (n < ind)
				ind = n;
			if (ind == 0)
				return self;
			while (l && *i != '\n')
			{
				--l;
				++i;
			}
		}
	}

	l = self->len;
	i = self->str;
	while (l)
	{
		n = ind;
		while (l && *i == tab && n)
		{
			i++;
			n--;
			l--;
		}
		k = i;
		while (l)
		{
			if (*i == '\n')
			{
				l--;
				i++;
				break;
			}
			l--;
			i++;
		}
		n = i - k;
		memcpy (p, k, n);
		p += n;
	}

	*p = '\0';
	self->len = p - self->str;
	return self;
}



/*----------------------------------------------------------------------------
	specific instance methods - conversion
----------------------------------------------------------------------------*/

/*) sub toident -> self
	Convert punctuations and funny characters to alpha or digits suitable for
	reference.

	@noexc

	@xmp
		x = Buffer ('F-o^ o  9')
		x.toident
		println (x)

	@out
		F-ov_o__9
*/
static qu_buffer qu__buffer_toident (qu_buffer self)
{
	char *s = (unsigned char *) self->str;
	long  l = self->len;

	ENSURE_NOT_FROZEN (self);

	while (l--)
	{
		*s = qu_ascii_toident (*s);
		++s;
	}
	return self;
}



/*) sub tolower -> self
	Convert characters to lowercase.

	@xmp
		x = Buffer ('ABC')
		x.tolower
		println (x)

	@out
		abc
*/
qu_buffer qu__buffer_tolower (qu_buffer self, int argc, qu_any *argv)
{
	char *s = self->str;
	long  l = self->len;

	ENSURE_NOT_FROZEN (self);

	while (l--)
	{
		*s = qu_ascii_tolower (*s);
		++s;
	}
	return self;
}



/*) sub toprint -> self
	Converting non-printable characters to spaces.

	Printable characters are C{' ' to '~'} plus C{'\t'} and C{'\n'}.
	Thus is different from f{isprint}.

	@xmp
		x = Buffer ('a\x00b\x01c')
		x.toprint
		println (x)

	@out
		a b c
*/
qu_buffer qu__buffer_toprint (qu_buffer self, int argc, qu_any *argv)
{
	char *s = (unsigned char *) self->str;
	long  l = self->len;

	ENSURE_NOT_FROZEN (self);

	while (l--)
	{
		*s = qu_ascii_toprint (*s);
		++s;
	}
	return self;
}



/*) sub toupper -> self
	Convert characters to uppercase.

	@xmp
		x = Buffer ('abc')
		x.toupper
		println (x)

	@out
		ABC
*/
static qu_buffer qu__buffer_toupper (qu_buffer self, int argc, qu_any *argv)
{
	char *s = (unsigned char *) self->str;
	long  l = self->len;

	ENSURE_NOT_FROZEN (self);

	while (l--)
	{
		*s = qu_ascii_toupper (*s);
		++s;
	}
	return self;
}



/*) sub toheader (i: Integer = 0, n: Integer = self.len) -> self
	Convert to header format optionally start from index v{i} and convert
	v{n} characters.

	If v{i} is negative then count from the end.
	If v{n} is negative it means all character from v{i}.

	@exc
	.EIndex

	@xmp
		x = Buffer ('foo-bar: 123')
		x.toheader
		println (x)

	@out
		Foo-Bar: 123
*/
static qu_buffer qu__buffer_toheader (qu_buffer self, int argc, qu_any *argv)
{
	char *s = self->str;
	long  l = self->len;
	long  i = 0;
	long  n = l;
	int   k = 0;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, i);
		QU_ENSURE_INDEX (i, l);
		if (argc > 1)
		{
			QU_ARG_vlong (1, n);
			if (n < 0 || n > l - i)
				n = l - i;
		}
	}
	while (n--)
	{
		char c = s[i++];

		if (c == ':')
			break;
		if (qu_ascii_ispunct (c) || qu_ascii_isspace (c))
		{
			if (n < 1)
				break;
			s[i] = qu_ascii_toupper (s[i]);
			--n;
		}
		else if (k == 0)
		{
			s[i - 1] = qu_ascii_toupper (c);
			k = 1;
		}
	}
	return self;
}



/*----------------------------------------------------------------------------
	specific instance methods - non-modifying
----------------------------------------------------------------------------*/

/*) sub avail -> Int
	@p
	Get the available space, that is, the number of bytes that can be stored
	until the Buffer needs to grow.

	@noexc

	@xmp
		b = Buffer (10)
		b <<< 'xxx'
		println (b.avail)
		b <<< 'xxx'
		println (b.avail)
		b <<< 'xxxxxxxxxxxxxxx'
		println (b.avail)

	@out
		13
		10
		11
*/
static qu_int qu__buffer_avail (qu_buffer self)
{
	long n = self->cap - self->len;
	assert (n <= QU_INT_MAX);
	return qu_int_new (n);
}



/*) sub contains (x, nocase = true) -> Bool
	Whether the String contains v{x}, optionally case insensitive.

	v{x} can be anything. If it is not a String ro Buffer then it's __string
	method is called.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('abc123def')
		println (b.contains (123))

	@out
		true
*/
static qu_boolean qu__buffer_contains (qu_buffer self, int argc, qu_any *argv)
{
	int        nocase = 1;
	qu_stringy what;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (!qu_stringy_is (what))
		what = (qu_stringy) qu_sys_string (what);

	if (argc > 1)
		QU_ARG_vbool (1, nocase);

	if (nocase)
	{
		if (qu_strz_strnc (self->str, self->len, what->str, what->len))
			return QuTrue;
	}
	else
	{
		if (qu_strz_str (self->str, self->len, what->str, what->len))
			return QuTrue;
	}
	return QuFalse;
}



/*) sub count (*arg)
	@def
	sub count (what, nocase = false) -> Int
	sub count (pattern: Regex, *arg) -> Int
	@p
	Count strings in the Buffer.

	v{what} may be a String or Buffer or an Array, Set or Dict that yields
	String or Buffer objects. In case of a Dict then the keys are used and
	the values ignored. Non String/Buffer items are ignored.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('abxdyfXgYh')
		println (b.count ({'x', 'y'}, true))

	@out
		4

	@p
	The second form expects a %{Regex} object.
	It simply calls C{pattern.count (self, *arg)}.

	@xmp
		use Regex

		r = Regex ('[abc]')
		b = Buffer ('foobar')
		println (b.count (r))

	@out
		2
*/
static qu_int qu__buffer_count (qu_buffer self, int argc, qu_any *argv)
{
	qu_any what;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];

	if (qu_classid (what) == QU_CLASS_ID_Regex)
	{
		static qu_string name = 0;

		if (name == 0)
			name = qu_string_new_staticz ("count");

		if (argc == 1)
			return qu_object_call1 (what, name, self);
		else
			return qu_object_call2 (what, name, self, argv[1]);
	}
	else
	{
		int     nocase = 0;
		long    l = self->len;
		char   *i = self->str;
		qu_any  v;
		long    count = 0;

		if (argc > 1)
			QU_ARG_vbool (1, nocase);

	  __qu_buffer_protect_begin (self);
		while (l)
		{
			qu_stringy x = _buffer_startwith (i, l, nocase, what, &v);
			if (x)
			{
				++count;
				l -= x->len;
				i += x->len;
			}
			else
			{
				--l;
				++i;
			}
		}
	  __qu_buffer_protect_end;
		return qu_int_new (count);
	}
}



/*) sub eachline -> Iterator
	Return a -{Iterator} object to iterate through each line.

	The lines returned are Buffer objects and do not contain the linebreak.

	@noexc

	@xmp
		s = Buffer ('abc\nde\nfgh\n')
		for l in s.eachline
			println (l)
		;;

	@out
		abc
		de
		fgh
*/
static qu_buffer qu__buffer_eachline_next (qu_buffer self, qu_iterator i)
{
	long p = i->ex1;
	long n = self->len - p;

	QU_REQUIRE (QuEStop, n > 0);
	/*begin*/
	{
		char *s = self->str;
		char *q = s + p;
		char *j = q;

		while (n && *j != '\n')
		{
			n--;
			j++;
		}
		n = j - q;
		i->ex1 = p + n + 1;
		return qu_buffer_new (q, n);
	}
}



static qu_iterator qu__buffer_eachline (qu_buffer self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, qu__buffer_eachline_next);
	return i;
}



/*) sub eachsection (marker) -> Iterator
	Return a -{Iterator} object to iterate through each section in self.

	Sections are divided by a line that starts with a special v{marker}
	character optionally preceeded by whitespaces.

	The objects yileded as Buffer objects are right-trimmed.
	Only non-empty ones are yielded.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('foo\n  \t.\n.bar\n  .baz')
		for x in b.eachsection ('.')
			println (x.quote)
		;;

	@out
		'foo'
		'bar'
		'baz'
*/
static qu_buffer qu__buffer_eachsection_next (qu_buffer self, qu_iterator i)
{
	long  p, l;
	char *s, *e, *q, *j, c;

  again:

	p = i->ex1;
	l = self->len;

	QU_REQUIRE (QuEStop, p < l);

	c = (char) i->ex2;
	s = self->str;
	e = s + l;
	q = s + p;
	j = q;

	if (p == 0)
	{
		while (j < e && qu_ascii_isspace (*j))
			j++;
		q = j;
	}
	while (j < e)
	{
		if (*j++ == '\n')
		{
			while (j < e && qu_ascii_iswhite (*j))
				j++;
			if (*j == c)
				break;
		}
	}
	i->ex1 = j - s;
	if (*q == c)
		++q;
	while (j > q && qu_ascii_isspace (j[-1]))
		--j;
	if (j <= q)
		goto again;

	return qu_buffer_new (q, j - q);
}



static qu_iterator qu__buffer_eachsection (qu_buffer self, qu_stringy marker)
{
	register qu_iterator i;

	QU_ARG_CHECK (qu_stringy_is (marker) && marker->len);
	QU_ITERATOR_NEW_EXTERN (i, self, qu__buffer_eachsection_next);
	i->ex2 = marker->str[0];
	return i;
}



/*) sub eachsegment (len: Integer) -> Iterator
	Return an -{Iterator} to iterate through segments which are v{len} long each.

	v{len} must be a positive integer.

	Items yielded are Buffer objects.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('abcdefghij')
		for x in b.eachsegment (4)
			println (x)
		;;

	@out
		abcd
		efgh
		ij
*/
static qu_buffer qu__buffer_eachsegment_next (qu_buffer self, qu_iterator i)
{
	long p = i->ex1;
	long l = self->len;

	QU_REQUIRE (QuEStop, p < l);
	/*begin*/
	{
		long len = i->ex2;
		long max = l - p;

		if (len > max)
			len = max;
		i->ex1 += len;
		return qu_buffer_new (self->str + p, len);
	}
}



static qu_iterator qu__buffer_eachsegment (qu_buffer self, qu_integer len)
{
	register qu_iterator i;
	long l;

	QU_ARG_long (len, l);
	QU_ARG_CHECK (l > 0);
	QU_ITERATOR_NEW_EXTERN (i, self, qu__buffer_eachsegment_next);
	i->ex2 = l;
	return i;
}



/*) sub endwith (what, nocase = false) -> Any
	Whether or not the Buffer ends with v{what}, optionally case insensitive.

	v{what} may be a String, or Buffer, or an Array, Set or Dict that yields
	String or Buffer objects. Non String/Buffer items are ignored. In case of
	a Dict then the keys are used and the corresponding value returned.

	Returns nil if none matches, the matched item otherwise.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('foo')
		println (b.endwith ('oo'))

		b = Buffer ('foo.PNG')
		println (b.endwith (['.gif', '.jpeg', '.png'], true))

	@out
		oo
		.png
*/
static qu_any qu__buffer_endwith (qu_buffer self, int argc, qu_any *argv)
{
	qu_any what;
	int    nocase = 0;
	qu_any v;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (argc > 1)
		QU_ARG_vbool (1, nocase);
	what = _buffer_endwith (self->str, self->len, nocase, what, &v);
	if (what)
		return v ? v : what;
	return QuNil;
}



/*) sub endswith (what, nocase = false) -> Boolean
	This method wraps #{endwith} and returns false if it returns nil,
	true otherwise.

	This method may be useful in switch statements.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('abcde')
		switch b
			case .endswith ('de')
				println (1)
				break

			case .startswith ('ab')
				println (2)
				break
		;;

	@out
		1
*/
static qu_any qu__buffer_endswith (qu_buffer self, int argc, qu_any *argv)
{
	qu_any x = qu__buffer_endwith (self, argc, argv);

	QU_ENSURE (x);
	return (x == (qu_any) QuNil) ? QuFalse : QuTrue;
}



/*) sub extract (opener: String, closer = opener) -> Array
	Extract parts of self enclosed by v{opener} and v{closer}.

	Both v{closer} and v{opener} must be non-empty Strings.

	The result is an -{Array} containing all the extracted String objects trimmed.

	@exc
	.EArgValue
	.EOverflow

	@xmp
		b = Buffer (' {: xxxx :} zzzz {: yyyy :} ')
		a = b.extract (s, '{:', ':}')
		for b in a
			println ('> 'b' <')
		;;
		println

		b = Buffer (' :: xxxx :: zzzz :: yyyy :: ')
		a = b.extract ('::')
		for b in a
			println ('> 'b' <')
		;;

	@out
		|xxxx|
		|yyyy|

		|xxxx|
		|zzzz|
		|yyyy|
*/
static qu_array qu__buffer_extract (qu_buffer self, int argc, qu_any *argv)
{
	qu_string opener, closer = 0;
	int       same = 1;

	QU_ARG_VCOUNT (1, 2);

	opener = argv[0];
	QU_ARG_CHECK (qu_string_is (opener) && opener->len);

	if (argc > 1)
	{
		closer = argv[1];
		QU_ARG_CHECK (qu_string_is (closer) && closer->len);
		same = qu_string_EQ (opener, closer);
	}
	else
	{
		closer = opener;
	}

	/*begin*/
	{
		long   l = self->len;
		char  *s = self->str;
		qu_array  a = qu_array_new (8);
		long   o = opener->len;
		long   c = closer->len;
		char  *u = opener->str;
		char  *v = closer->str;
		long   i = 0;
		long   d = same ? 0 : c;
		long   j;

		while (i < l - o)
		{
			i = qu_strz_index (s, l, u, o, i);
			if (i < 0 || i >= l - c)
				break;

			i += o;
			j = qu_strz_index (s, l, v, c, i);
			if (j < 0)
				break;

			qu_array_push (a, qu_buffer_new_trim (s + i, j - i));
			i = j + d;
		}
		return a;
	}
}



/*) sub head (sep = '', offset = 0, nocase = false) -> Buffer
	Return the part of the buffer before the separator v{sep}.

	Here v{sep} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	If v{sep} is an empty string, it means any space.

	Optionally start at a specified v{offset}.
	If this is a negative value then count from the end.
	If the v{offset} falls out of this String then an empty String is returned,
	thus no EIndex thrown.

	If the separator is not found then an empty buffer is returned.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('123\t 45678')
		x = b.head
		println (x)

	@out
		123
*/
static qu_buffer qu__buffer_head (qu_buffer self, int argc, qu_any *argv)
{
	qu_any    sep = 0;
	int       nocase = 0;
	long      ofs = 0;
	char     *i = self->str;
	long      l = self->len;
	qu_string x;
	qu_any    v;

	QU_ARG_VCOUNT (0, 3);
	if (argc > 0)
	{
		sep = argv[0];
		if (qu_stringy_is (sep) && ((qu_stringy) sep)->len == 0)
			sep = 0;

		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			if (argc > 2)
				QU_ARG_vbool (2, nocase);

			if (ofs < 0)
				ofs += l;
			if (ofs < 0)
				return qu_buffer_new ("", 0);
			if (ofs >= l)
				return qu_buffer_new (self->str, self->len);
		}
	}
	i += ofs;
	l -= ofs;

	if (sep == 0)
	{
		while (l)
		{
			if (qu_ascii_isspace (*i))
				return qu_buffer_new (self->str, i - self->str);
			++i;
			--l;
		}
	}
	else
	{
		while (l)
		{
			x = _buffer_startwith (i, l, nocase, sep, &v);
			if (x)
				return qu_buffer_new (self->str, i - self->str);
			++i;
			--l;
		}
	}
	return qu_buffer_new (self->str, self->len);
}



/*) sub index (what, offset = 0, nocase = false) -> Int
	Return the index where v{what} starts in self, -1 if not found.

	Here v{what} may be a String or Buffer or an Array, Set or Dict that yields
	String or Buffer objects. In case of a Dict then the keys are used and the
	values ignored. Non String/Buffer items are ignored.

	Optionally start looking from v{offset} and ignore case.
	If v{offset} is negative then count from the end.
	If it falls out of this String then -1 is returned, thus no EIndex thrown.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('..XY..ab')
		println (b.index (const ['xy', 'ab']))

	@out
		6
*/
static qu_int qu__buffer_index (qu_buffer self, int argc, qu_any *argv)
{
	qu_any      what;
	long        ofs = 0, res = -1;
	int         nocase = 0;
	long        l = self->len;
	char       *i;
	qu_stringy  x;
	qu_any      v;

	QU_ARG_VCOUNT (1, 3);
	what = argv[0];
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
			QU_ARG_vbool (2, nocase);

		if (ofs < 0)
			ofs += l;
		if (ofs < 0 || ofs >= l)
			return QuInt_1;
	}
	i = self->str + ofs;
	l -= ofs;
  __qu_buffer_protect_begin (self);
	while (l)
	{
		x = _buffer_startwith (i, l, nocase, what, &v);
		if (x)
		{
			res = i - self->str;
			break;
		}
		++i;
		--l;
	}
  __qu_buffer_protect_end;
	return qu_int_new (res);
}



/*) sub match (what) -> Buffer
	@def
	sub match (what, nocase= false) -> Buffer
	sub match (pattern: Regex, *arg) -> Buffer
	@p
	Find string in Buffer.

	Here v{what} may be a String or Buffer or an Array, Set or Dict that yields String
	or Buffer objects. In case of a Dict then the keys are used and the values ignored.
	Non String/Buffer items are ignored.

	The returned value is an empty Buffer if no match found, the part of the Buffer
	that matches otherwise.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('FooBarBaz')
		r = b.match ({'baz', 'bar'}, true)
		println (r)
		println (b.index (r))

		r = b.match (['baz', 'bar'])
		println (r.quote)
		println (b.index (r))

	@out
		Bar
		3
		''
		-1

	@p
	The second form expects a %{Regex} object.
	It simply calls C{pattern.match (self, *arg)}.

	@xmp
		use Regex

		r = Regex (`[abc]x`)
		b = Buffer ('aaa ax xxx')
		println (b.match (r))

	@out
		ax
*/
static qu_buffer qu__buffer_match (qu_buffer self, int argc, qu_any *argv)
{
	qu_any    what;
	qu_buffer res = 0;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];

	if (qu_classid (what) == QU_CLASS_ID_Regex)
	{
		static qu_string name = 0;

		if (name == 0)
			name = qu_string_new_staticz ("match");

		if (argc == 1)
			res = qu_object_call1 (what, name, self);
		else
			res = qu_object_call2 (what, name, self, argv[1]);

		QU_ENSURE (res);

		/* this should never happen */
		QU_REQUIRE (QuEImplement, qu_buffer_is (res));
	}
	else
	{
		int     nocase = 0;
		long    l = self->len;
		char   *i = self->str;
		qu_any  v;

		if (argc > 1)
			QU_ARG_vbool (1, nocase);

	  __qu_buffer_protect_begin (self);
		while (l)
		{
			qu_stringy x = _buffer_startwith (i, l, nocase, what, &v);
			if (x)
			{
				res = qu_buffer_new (i, x->len);
				break;
			}
			--l;
			++i;
		}
	  __qu_buffer_protect_end;
	}
	return res ? res : qu_buffer_alloc (0);
}



/*) sub quote (quote = "'") -> String
	Return a new quoted String with characaters escaped as necessary.

	You can optionally set the v{quote} as C{'} or C{"} or C{`} but no other.
	If v{quote} is C{`} and String contains non-printable characters then
	then the v{quote} is forced to C{'}.

	@exc
	.EArgValue
	.EOverflow

	@xmp
		s = 'foo\tbar'
		a = [1, s]
		println (a)
		println (s.quote)
		println (s)

	@out
		[1, 'foo\tbar']
		'foo\tbar'
		foo     bar
*/
static qu_string qu__buffer_quote (qu_buffer self, int argc, qu_any *argv)
{
	char quote = '\'';

	if (argc)
	{
		qu_stringy s;
		QU_ARG_VTYPE (0, s, qu_stringy);
		quote = s->str[0];
		QU_ARG_CHECK (quote == '\'' || quote == '"' || quote == '`');
	}
	return qu_strz_quote (self->str, self->len, quote);
}



/*) sub limit (*arg) -> Int
	@def
	sub limit -> Int
	sub limit (max: Integer) -> Int
	@p
	Set or get the capacity limit of this Buffer.

	If setting then a non-positive integer or a value bigger then the system hard
	limit, means to use the system hard limit.

	Note that this does not change the content of the Buffer.

	@exc
	.EArgValue

	@xmp
		b = Buffer (80)
		b.limit = 100
		loop
			b <<< 'hello'
		;;

	@out
		[1024] __main__ (EOverflow) capacity overflow
*/
static qu_int qu__buffer_limit (qu_buffer self, int argc, qu_any *argv)
{
	if (argc)
	{
		long m;

		QU_ARG_vlong (0, m);
		if (m <= 0)
			self->limit = QU_MEM_MAX;
		else
		{
			if (m > QU_MEM_MAX)
				m = QU_MEM_MAX;
			self->limit = m;
		}
	}
	assert (self->limit <= QU_INT_MAX);
	return qu_int_new (self->limit);
}



/*) sub rindex (what, offset = self.len - 1, nocase = false) -> Int
	Return the index where v{what} starts in self, -1 if not found.
	Optionally start looking from v{offset} and ignore case.

	Here v{what} may be a String or Buffer or an Array, Set or Dict that yields
	String or Buffer objects. In case of a Dict then the keys are used and the
	values ignored. Non String/Buffer items are ignored.

	If v{offset} is negative then count from the end.

	Unlike #{index}, this method does the search in reverse order towards the
	start of the string.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('..XY..ab')
		println (b.rindex (const ['xy', 'ab']))

	@out
		6
*/
static qu_int qu__buffer_rindex (qu_buffer self, int argc, qu_any *argv)
{
	long        l = self->len, res = -1;
	qu_any      what;
	long        ofs = l - 1;
	int         nocase = 0;
	char       *i;
	qu_stringy  x;
	qu_any      v;

	QU_ARG_VCOUNT (1, 3);
	what = argv[0];
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
			QU_ARG_vbool (2, nocase);

		if (ofs < 0)
			ofs += l;
		if (ofs < 0 || ofs >= l)
			return QuInt_1;
	}
	i = self->str + ofs;
	l = ofs + 1;
  __qu_buffer_protect_begin (self);
	while (l)
	{
		x = _buffer_startwith (i, l, nocase, what, &v);
		if (x)
		{
			res = i - self->str;
			break;
		}
		--i;
		--l;
	}
  __qu_buffer_protect_end;
	return qu_int_new (res);
}



/*) sub split (*arg) -> Array
	@def
	sub split -> Array
	sub split (delimiter, max = 0, options = 't') -> Array
	sub split (regex, max = 0, options = '') -> Array
	@p
	Split the Buffer into an Array of Buffers.

	The first form splits the Buffer by any space delimiter.

	@xmp
		b = Buffer (' a \t  b \n  c \r d \v ')
		a = b.split
		println (a)

	@out
		['a', 'b', 'c', 'd']

	@p
	In the second form, v{delimiter} may be a String, or Buffer, or an Array, Set
	or Dict that yields String or Buffer objects. In case of a Dict then the keys
	are used and the values ignored. Non String/Buffer items are ignored.

	v{max} is the maximum number of items to return.
	A non-positive v{max] means no limit.

	The possible options are:
	@ul
	.C{"n"} case insensitive
	.C{"s"} remove spaces surrounding the delimiter
	.C{"q"} account for quoted strings
	.C{"u"} unquote (remove quotes), only if C{"q"} specified
	.C{"t"} trim after removing quotes, only if C{"u"} specified

	@xmp
		b = Buffer ('  a, b ; c ,d  ')
		println (b.split (const [';' ',']))

		b = Buffer (' " fXo " X `bxr` x \'baz\'  ')
		println (b.split ('x', 0, "nsqut"))

	@out
		['a', 'b', 'c', 'd']
		['fXo', 'bar', 'baz']

	@p
	The last form simply calls C{regex.split (self, max, options)}.
	See the %{Regex} module for more information.

	@xmp
		use Regex

		b = Buffer ('  a, b ; c ,d  ')
		r = Regex (`[[:space:]]*[,;][[:space:]]*`)
		a = b.split (r)
		println (a)

	@out
		[' a', 'b', 'c', 'd  ']
*/
static qu_array qu__buffer_split_byspaces (qu_buffer self)
{
	qu_array  r = qu_array_new (4);
	long      l = self->len;
	char     *i = self->str;
	char     *j;

	for (;;)
	{
		while (l && qu_ascii_isspace (*i))
		{
			++i;
			--l;
		}
		if (l == 0)
			break;
		j = i;
		while (l && !qu_ascii_isspace (*i))
		{
			--l;
			++i;
		}
		QU_ENSURE (qu_array_push (r, qu_buffer_new (j, i - j)));
	}
	return r;
}


static qu_array qu__buffer_split_byregex (qu_buffer self, int argc, qu_any *argv)
{
	static qu_string name = 0;
	qu_array r;

	if (name == 0)
		name = qu_string_new_staticz ("split");

	assert (argc > 0);

	if (argc == 1)
		r = qu_object_call1 (argv[0], name, self);

	else if (argc == 2)
		r = qu_object_call2 (argv[0], name, self, argv[1]);

	else if (argc == 3)
		r = qu_object_call3 (argv[0], name, self, argv[1], argv[2]);

	else
		QU_THROW (QuEArgCount);

	QU_ENSURE (r);
	QU_REQUIRE (QuEImplement, qu_array_is (r));
	return r;
}


#define QU_BUFFER_SPLIT_NOCASE  0x01
#define QU_BUFFER_SPLIT_SPACE   0x02
#define QU_BUFFER_SPLIT_QUOTE   0x04
#define QU_BUFFER_SPLIT_UNQUOTE 0x08
#define QU_BUFFER_SPLIT_TRIM    0x10

static qu_any qu__buffer_split_bydelim_push (qu_array r, char *start, char *end, long op)
{
	if (op & QU_BUFFER_SPLIT_SPACE)
	{
		while (start < end && qu_ascii_isspace (*start))
			++start;
		while (end > start && qu_ascii_isspace (*end))
			--end;
	}
	if (op & QU_BUFFER_SPLIT_QUOTE)
	{
		if (op & QU_BUFFER_SPLIT_UNQUOTE)
		{
			char q = *start;
			if (end > start && *end == q && (q == '"' || q == '\'' || q == '`'))
			{
				++start;
				--end;
				if (op & QU_BUFFER_SPLIT_TRIM)
				{
					while (start < end && qu_ascii_isspace (*start))
						++start;
					while (end > start && qu_ascii_isspace (*end))
						--end;
				}
			}
		}
	}
	if (end >= start)
		return qu_array_push (r, qu_buffer_new (start, end - start + 1));
	return qu_array_push (r, qu_buffer_alloc(0));
}


static qu_array qu__buffer_split_bydelim (qu_buffer self, int argc, qu_any *argv)
{
	long        max = LONG_MAX;
	long        op  = QU_BUFFER_SPLIT_SPACE;
	qu_any      spec;
	qu_array    r;
	long        l;
	char       *i, *j, q;
	qu_stringy  delim = 0;
	qu_any      v;

	assert (argc > 0);
	spec = argv[0];

	QU_ARG_VMAX (3);
	if (argc > 1)
	{
		QU_ARG_vlong (1, max);
		if (max <= 0)
			max = LONG_MAX;
		if (argc > 2)
		{
			qu_stringy  s;
			long        l;
			char       *z;

			QU_ARG_VTYPE (2, s, qu_stringy);
			l = s->len;
			z = s->str;
			op = 0;
			while (l--)
			{
				switch (*z++)
				{
					case 'n': op |= QU_BUFFER_SPLIT_NOCASE;  break;
					case 's': op |= QU_BUFFER_SPLIT_SPACE;   break;
					case 'q': op |= QU_BUFFER_SPLIT_QUOTE;   break;
					case 'u': op |= QU_BUFFER_SPLIT_UNQUOTE; break;
					case 't': op |= QU_BUFFER_SPLIT_TRIM;    break;
				}
			}
		}
	}

  __qu_buffer_protect_begin (self);

	l = self->len;
	i = j = self->str;
	r = qu_array_new (4);

	max--;
	while (l && max)
	{
		while (l)
		{
			if (op & QU_BUFFER_SPLIT_QUOTE)
			{
				q = *i;
				if (q == '"' || q == '\'' || q == '`')
				{
					--l;
					++i;
					while (l)
					{
						if (*i == '\\' && q != '`')
						{
							--l;
							++i;
						}
						else if (*i == q)
						{
							--l;
							++i;
							break;
						}
						else
						{
							--l;
							++i;
						}
					}
				}
				if (l < 1)
					break;
			}

			delim = _buffer_startwith (i, l, op & QU_BUFFER_SPLIT_NOCASE, spec, &v);
			if (delim && delim->len)
				goto found;
			--l;
			++i;
		}
		break;

	found:
		if (!qu__buffer_split_bydelim_push (r, j, i - 1, op))
		{
			r = 0;
			goto done;
		}
		max--;
		l -= delim->len;
		i += delim->len;
		j = i;
	}
	i = self->str + self->len;
	if (i > j)
	{
		if (!qu__buffer_split_bydelim_push (r, j, i - 1, op))
			r = 0;
	}

  done:

  __qu_buffer_protect_end;
	QU_ENSURE (r);
	return r;
}



static qu_array qu__buffer_split (qu_buffer self, int argc, qu_any *argv)
{
	if (argc == 0)
		return qu__buffer_split_byspaces (self);

	else if (qu_classid (argv[0]) == QU_CLASS_ID_Regex)
		return qu__buffer_split_byregex (self, argc, argv);

	else
		return qu__buffer_split_bydelim (self, argc, argv);
}



/*) sub splitat (*indexes) -> Array
	@def
	sub splitat (*indexes) -> Array
	sub splitat (indexes: Array) -> Array
	@p
	Split the Buffer at the specified v{indexes}.

	The indexes must be sorted from low to high.
	Use -{Array.sort} if necessary.

	This is useful when you are scanning the Buffer and keep scanned indexes
	in an Array.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('abcdefghijkl')
		a = [1 4 7 100]
		r = b.splitat (a)    # same as r = b.splitat (*a)
		println (r)

	@out
		['a', 'bcd', 'efg', 'hijkl']
*/
static qu_array qu__buffer_splitat (qu_buffer self, int argc, qu_any *argv)
{
	long     l = self->len;
	char    *s = self->str;
	long     i = 0;
	long     j, m, n;
	qu_array r;
	qu_any   x;

	if (argc && qu_array_is (argv[0]))
	{
		argc = ((qu_array) argv[0])->len;
		argv = ((qu_array) argv[0])->objs;
	}

	r = qu_array_new (argc);
	while (argc--)
	{
		x = *argv++;
		QU_ARG_int (x, j);
		if (j >= l || j < i)
			break;
		n = j - i;
		m = l - j;
		if (n > m)
			n = m;
		QU_ENSURE (qu_array_push (r, qu_buffer_new (s + i, n)));
		i = j;
	}
	if (i < l)
		QU_ENSURE (qu_array_push (r, qu_buffer_new (s + i, l - i)));
	return r;
}



/*) sub splitnum (how = 'B') -> Array
	Break the string by separating integer numbers within the string with the others.

	Optionally set v{how} to:
	@ops
	.| C{b} | collect both, don't convert to Integer
	.| C{B} | collect both, convert to Integer
	.| C{n} | collect numbers only, don't convert to Integer
	.| C{N} | collect numbers only, convert to Integer
	.| C{s} | collect strings only

	@exc
	.EArgValue

	@xmp
		s = 'x-a12.100p'
		println (s.splitnum)
		println (s.splitnum ('N'))
		println (s.splitnum ('n'))
		println (s.splitnum ('s'))

	@out
		['x-a', 12, '.', 100, 'p']
		[12, 100]
		['12', '100']
		['x-a', '.', 'p']
*/
static qu_array qu__buffer_splitnum (qu_buffer self, int argc, qu_any *argv)
{
	char        op = 'B';
	long        l = self->len;
	char       *i = self->str;
	qu_array    a;
	char       *p, *k;
	qu_integer  v;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
	{
		qu_stringy x;
		QU_ARG_VTYPE (0, x, qu_stringy);
		op = x->str[0];
		QU_ARG_CHECK (op == 'B' || op == 'b' || op == 'n' || op =='N' || op == 's' || op == 'S');
	}

	if (l == 0)
		return qu_array_new (0);

	a = qu_array_new (4);
	p = i;
	while (l)
	{
		if (qu_ascii_isdigit (*i))
		{
			k = i;

			if (i > p && (op == 'B' || op == 'b' || op == 'S' || op == 's'))
				qu_array_push (a, qu_buffer_new (p, i - p));

			while (qu_ascii_isdigit (*i))
			{
				++i;
				--l;
			}

			if (op == 'b' || op == 'n')
			{
				qu_array_push (a, qu_buffer_new (k, i - k));
			}
			else if (op == 'B' || op == 'N')
			{
				/*HACK -- qu_int_scan errs when there is a dot after the digit */
				char c = *i;
				*i = '\0';
				v = (qu_integer) qu_int_scan (&k, 0);
				*i = c;
				if (v == 0)
				{
					*i = '\0';
					v = (qu_integer) qu_long_scan (&k, 0);
					*i = c;
					if (v == 0)
						/* this should never happen unless the string is extra huge */
						v = (qu_integer) QuInt0;
				}
				qu_array_push (a, v);
			}
			p = i;
		}
		else
		{
			++i;
			--l;
		}
	}
	if (i > p && (op == 'B' || op == 'b' || op == 'S' || op == 's'))
		qu_array_push (a, qu_buffer_new (p, i - p));
	return a;
}



/*) sub startwith (what, offset = 0, nocase = false) -> Any
	Whether or not the Buffer starts with v{what}, optionally case insensitive
	and from a specified v{offset}.

	v{what} may be a String, or Buffer, or an Array, Set or Dict that yields
	String or Buffer objects. Non String/Buffer items are ignored. In case of a
	Dict then the keys are used and the corresponding value returned.

	Returns nil if none matches, the matched item otherwise.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('foo')
		println (b.startwith ('fo'))

	@out
		fo
*/
static qu_any qu__buffer_startwith (qu_buffer self, int argc, qu_any *argv)
{
	qu_any  what;
	long    ofs = 0;
	int     nocase = 0;
	qu_any  v;

	QU_ARG_VCOUNT (1, 3);
	what = argv[0];
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
			QU_ARG_vbool (2, nocase);

		if (ofs < 0)
			ofs += self->len;
		if (ofs < 0 || ofs >= self->len)
			return QuNil;
	}
	what = _buffer_startwith (self->str + ofs, self->len - ofs, nocase, what, &v);
	if (what)
		return v ? v : what;
	return QuNil;
}



/*) sub startswith (what, offset = 0, nocase = false, word = false) -> Boolean
	This method wraps #{starwith} and returns false if it returns nil,
	true otherwise.

	It may be useful in a switch statement.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('bar')
		switch b
			case .startswith ('a')
				println (1)
				break

			case .startswith ('b')
				println (2)
				break
		;;

	@out
		2
*/
static qu_any qu__buffer_startswith (qu_buffer self, int argc, qu_any *argv)
{
	qu_any x = qu__buffer_startwith (self, argc, argv);

	QU_ENSURE (x);
	return (x == (qu_any) QuNil) ? QuFalse : QuTrue;
}



/*) sub tail (sep = '', ofs = 0, nocase = false) -> String
	Return the part of the Buffer after the separator v{sep}.

	v{sep} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	If v{sep} is an empty String then it means any number of spaces.

	Optionally start of v{offset}.
	If this is a negative value then count from the end.

	If the seperator is not found then an empty Buffer is returned.

	@exc
	.EArgValue

	@xmp
		b = Buffer ('123 \n \t 45678')
		x = b.tail
		println (x)

	@out
		45678
*/
static qu_buffer qu__buffer_tail (qu_buffer self, int argc, qu_any *argv)
{
	qu_any     sep = 0;
	int        nocase = 0;
	long       ofs = 0;
	char      *i = self->str;
	long       l = self->len;
	qu_string  x;
	qu_any     v;

	QU_ARG_VCOUNT (0, 3);
	if (argc > 0)
	{
		sep = argv[0];
		if (qu_stringy_is (sep) && ((qu_stringy) sep)->len == 0)
			sep = 0;
		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			if (argc > 2)
				QU_ARG_vbool (2, nocase);

			if (ofs < 0)
				ofs += l;
			if (ofs < 0)
				return qu_buffer_new (self->str, self->len);
			if (ofs >= l)
				return qu_buffer_new ("", 0);
		}
	}
	i += ofs;
	l -= ofs;

	if (sep == 0)
	{
		while (l)
		{
			if (qu_ascii_isspace (*i))
			{
				while (l && qu_ascii_isspace (*i))
				{
					--l;
					++i;
				}
				return qu_buffer_new (i, self->str + self->len - i);
			}
			++i;
			--l;
		}
	}
	else
	{
		while (l)
		{
			x = _buffer_startwith (i, l, nocase, sep, &v);
			if (x)
				return qu_buffer_new (i + x->len, l - x->len);
			++i;
			--l;
		}
	}
	return qu_buffer_new ("", 0);
}



/*-----------------------------------------------------------------------------
	specific instance methods - protect
-----------------------------------------------------------------------------*/

/*) sub protect -> Null
	Make this Buffer temporarily immutable.

	You can #{unprotect} the Buffer but only within the same I{scope} that
	calls this method. The I{scope} is a -{Class} if the caller is within a Class,
	a Module otherwise.

	If the I{scope} is a Module then only code at module level can unprotect the
	Container. If the I{scope} is a Class then only code within that Class can
	unprotect it, that is, not even subclasses can.

	@exc
	.EModify  already protected

	@xmp
		Buf = Buffer ('abcde')

		class Foo

			sub class protect
				return Buf.protect     # protected from within Foo
			;;

			sub class unprotect
				Buf.unprotect          # only Foo can unprotect it
			;;
		;;

		Foo.protect
		println (Buf.protector)

		try
			Buf[0] = 'X'
		catch
			println (__err__)
		;;

		try
			Buf.unprotect
		catch
			println (__err__)
		;;

		Foo.unprotect
		Buf[0] = 'X'
		println (Buf)

		Buf.protect                   # protected from within module
		try
			Buf[0] = 'Y'
		catch
			println (__err__)
		;;
		Buf.unprotect                 # only this module can unprotect it

	@out
		<Class __main__.Foo>
		(EModify) Buffer is frozen or protected
		(EModify) can not modify object
		Xbcde
		(EModify) Buffer is frozen or protected
*/
static qu_null qu__buffer_protect (qu_buffer self)
{
	QU_REQUIRE (QuEModify, self->scope == 0);
	self->scope = qu_interp_scope ();
	qu_flag_set (self, QU_BUFFER_FLAG_PROTECT);
	return QuNil;
}



/*) sub protector
	Return the Module or Class that called #{protect}.

	If none did then nil is returned.

	See #{protect} for more information.

	@noexc
*/
static qu_any qu__buffer_protector (qu_buffer self)
{
	if (self->scope == 0)
		return QuNil;
	return self->scope;
}



/*) sub unprotect -> Null
	Unprotect this Buffer.

	If the Buffer is not protected then nothing happens.

	You can only unprotect the Buffer from within the same I{scope}
	that called the f{protect} method.

	See #{protect} for more information.

	@exc
	.EModify invalid protection scope
*/
static qu_null qu__buffer_unprotect (qu_buffer self)
{
	if (self->scope)
	{
		qu_any x = qu_interp_scope ();
		QU_REQUIRE (QuEModify, x == self->scope);
		qu_flag_unset (self, QU_BUFFER_FLAG_PROTECT);
		self->scope = 0;
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

/*
	Exchange buffer with a temporary qu_buffer.
*/
void qu_buffer_exchange_tmp (qu_buffer self, qu_buffer tmp)
{
	char *b = self->str;

	self->str = tmp->str;
	self->len = tmp->len;
	self->cap = tmp->cap;

	tmp->len = 0;
	tmp->cap = self->cap;
	tmp->str = b;
	tmp->str[0] = '\0';
}



/*
	Exchange buffer with a temporary qu_string.
*/
void qu_buffer_exchange_tmp_string (qu_buffer self, qu_string tmp)
{
	char *b = self->str;

	/* this should rarely happen so we keep it simple */
	if (qu_string_is_static (tmp))
		tmp = qu_string_dup (tmp);

	self->str = tmp->str;
	self->len = self->cap = tmp->len;

	tmp->len = 0;
	tmp->str = b;
	tmp->str[0] = '\0';
}



/*
	Magically change the qu_buffer into a qu_string.
*/
qu_string qu_buffer_tmp_to_string (qu_buffer self)
{
	self->klass = QuStringClass;
	self->scope = 0;
	return (qu_string) self;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_buffer__setup ()
{
	qu_class_setup_final (QuBufferClass, qu_buffer_s, qu__buffer__destroy, qu__buffer__mark);
}


void qu_buffer__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuBufferClass, QU_NAME_new, qu__buffer_new, 1);

	/* standard instance methods */
	qu_class_method_inst  (QuBufferClass, QU_NAME__as      , qu__buffer__as      , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__clear   , qu__buffer__clear   , 0);
	qu_class_alias        (QuBufferClass, QU_NAME__clear   , QU_NAME_clear);
	qu_class_method_inst  (QuBufferClass, QU_NAME__cmp     , qu__buffer__cmp     , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__copy    , qu__buffer__copy    , 0);
	qu_class_alias        (QuBufferClass, QU_NAME__copy    , QU_NAME_copy);
	qu_class_method_inst  (QuBufferClass, QU_NAME__del     , qu__buffer__del     , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__each    , qu__buffer__each    , 0);
	qu_class_method_inst  (QuBufferClass, QU_NAME__getitem , qu__buffer__getitem , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__getslice, qu__buffer__getslice, 2);
	qu_class_method_inst  (QuBufferClass, QU_NAME__in      , qu__buffer__in      , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__len     , qu__buffer__len     , 0);
	qu_class_alias        (QuBufferClass, QU_NAME__len     , QU_NAME_len);
	qu_class_method_inst  (QuBufferClass, QU_NAME__like    , qu__buffer__like    , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__print   , qu__buffer__print   , 0);
	qu_class_method_inst  (QuBufferClass, QU_NAME__pop     , qu__buffer__pop     , 0);
	qu_class_method_inst  (QuBufferClass, QU_NAME__push    , qu__buffer__push    , 1);
	qu_class_method_inst  (QuBufferClass, QU_NAME__setitem , qu__buffer__setitem , 2);
	qu_class_method_inst  (QuBufferClass, QU_NAME__setslice, qu__buffer__setslice, 3);
	qu_class_method_inst  (QuBufferClass, QU_NAME__print   , qu__buffer__print   , 0);
	qu_class_method_inst  (QuBufferClass, QU_NAME__string  , qu__buffer__string  , 0);
	qu_class_method_inst  (QuBufferClass, QU_NAME__true    , qu__buffer__true    , 0);

	/* specific instance methods - modifying */
	qu_class_method_instz (QuBufferClass, "capitalize", qu__buffer_capitalize,  0);
	qu_class_method_instz (QuBufferClass, "chop"      , qu__buffer_chop      , -1);
	qu_class_method_instz (QuBufferClass, "chomp"     , qu__buffer_chomp     , -1);
	qu_class_method_instz (QuBufferClass, "consume"   , qu__buffer_consume   ,  1);
	qu_class_method_instz (QuBufferClass, "exchange"  , qu__buffer_exchange  ,  1);
	qu_class_method_instz (QuBufferClass, "fixline"   , qu__buffer_fixline   , -1);
	qu_class_method_instz (QuBufferClass, "flatten"   , qu__buffer_flatten   ,  0);
	qu_class_method_instz (QuBufferClass, "get"       , qu__buffer_get       , -1);
	qu_class_method_instz (QuBufferClass, "indent"    , qu__buffer_indent    , -1);
	qu_class_method_instz (QuBufferClass, "inject"    , qu__buffer_inject    , -1);
	qu_class_method_instz (QuBufferClass, "insert"    , qu__buffer_insert    , -1);
	qu_class_method_instz (QuBufferClass, "ltrim"     , qu__buffer_ltrim     ,  0);
	qu_class_method_instz (QuBufferClass, "makespace" , qu__buffer_makespace ,  1);
	qu_class_method_instz (QuBufferClass, "push"      , qu__buffer_push      , -1);
	qu_class_method_instz (QuBufferClass, "put"       , qu__buffer_put       , -1);
	qu_class_method_instz (QuBufferClass, "replace"   , qu__buffer_replace   , -1);
	qu_class_method_instz (QuBufferClass, "reverse"   , qu__buffer_reverse   ,  0);
	qu_class_method_instz (QuBufferClass, "rtrim"     , qu__buffer_rtrim     ,  0);
	qu_class_method_instz (QuBufferClass, "scrap"     , qu__buffer_scrap     , -1);
	qu_class_method_instz (QuBufferClass, "title"     , qu__buffer_title     ,  0);
	qu_class_method_instz (QuBufferClass, "transfer"  , qu__buffer_transfer  , -1);
	qu_class_method_instz (QuBufferClass, "trim"      , qu__buffer_trim      ,  0);
	qu_class_method_instz (QuBufferClass, "unfold"    , qu__buffer_unfold    ,  0);
	qu_class_method_instz (QuBufferClass, "unindent"  , qu__buffer_unindent  , -1);

	/* specific instance methods - conversion */
	qu_class_method_instz (QuBufferClass, "toident" , qu__buffer_toident ,  0);
	qu_class_method_instz (QuBufferClass, "tolower" , qu__buffer_tolower ,  0);
	qu_class_method_instz (QuBufferClass, "toprint" , qu__buffer_toprint ,  0);
	qu_class_method_instz (QuBufferClass, "toupper" , qu__buffer_toupper ,  0);
	qu_class_method_instz (QuBufferClass, "toheader", qu__buffer_toheader, -1);

	/* specific instance methods - non-modifying */
	qu_class_method_instz (QuBufferClass, "avail"      , qu__buffer_avail      ,  0);
	qu_class_method_instz (QuBufferClass, "contains"   , qu__buffer_contains   , -1);
	qu_class_method_instz (QuBufferClass, "count"      , qu__buffer_count      , -1);
	qu_class_method_instz (QuBufferClass, "eachline"   , qu__buffer_eachline   ,  0);
	qu_class_method_instz (QuBufferClass, "eachsection", qu__buffer_eachsection,  1);
	qu_class_method_instz (QuBufferClass, "eachsegment", qu__buffer_eachsegment,  1);
	qu_class_method_instz (QuBufferClass, "endwith"    , qu__buffer_endwith    , -1);
	qu_class_method_instz (QuBufferClass, "endswith"   , qu__buffer_endswith   , -1);
	qu_class_method_instz (QuBufferClass, "extract"    , qu__buffer_extract    , -1);
	qu_class_method_instz (QuBufferClass, "head"       , qu__buffer_head       , -1);
	qu_class_method_instz (QuBufferClass, "index"      , qu__buffer_index      , -1);
	qu_class_method_instz (QuBufferClass, "quote"      , qu__buffer_quote      , -1);
	qu_class_method_instz (QuBufferClass, "limit"      , qu__buffer_limit      , -1);
	qu_class_method_instz (QuBufferClass, "match"      , qu__buffer_match      , -1);
	qu_class_method_instz (QuBufferClass, "rindex"     , qu__buffer_rindex     , -1);
	qu_class_method_instz (QuBufferClass, "split"      , qu__buffer_split      , -1);
	qu_class_method_instz (QuBufferClass, "splitat"    , qu__buffer_splitat    , -1);
	qu_class_method_instz (QuBufferClass, "splitnum"   , qu__buffer_splitnum   , -1);
	qu_class_method_instz (QuBufferClass, "startwith"  , qu__buffer_startwith  , -1);
	qu_class_method_instz (QuBufferClass, "startswith" , qu__buffer_startswith , -1);
	qu_class_method_instz (QuBufferClass, "tail"       , qu__buffer_tail       , -1);

	/* protection */
	qu_class_method_instz (QuBufferClass, "protect"   , qu__buffer_protect  , 0);
	qu_class_method_instz (QuBufferClass, "protector" , qu__buffer_protector, 0);
	qu_class_method_instz (QuBufferClass, "unprotect" , qu__buffer_unprotect, 0);

	/* from String */
	qu_class_method_instz (QuBufferClass, "isalnum" , qu__string_isalnum , -1);
	qu_class_method_instz (QuBufferClass, "isalpha" , qu__string_isalpha , -1);
	qu_class_method_instz (QuBufferClass, "isbdigit", qu__string_isbdigit, -1);
	qu_class_method_instz (QuBufferClass, "isblank" , qu__string_isblank , -1);
	qu_class_method_instz (QuBufferClass, "iscntrl" , qu__string_iscntrl , -1);
	qu_class_method_instz (QuBufferClass, "isdigit" , qu__string_isdigit , -1);
	qu_class_method_instz (QuBufferClass, "isgraph" , qu__string_isgraph , -1);
	qu_class_method_instz (QuBufferClass, "ishi"    , qu__string_ishi    , -1);
	qu_class_method_instz (QuBufferClass, "isident" , qu__string_isident , -1);
	qu_class_method_instz (QuBufferClass, "isidnum" , qu__string_isidnum , -1);
	qu_class_method_instz (QuBufferClass, "islo"    , qu__string_islo    , -1);
	qu_class_method_instz (QuBufferClass, "islower" , qu__string_islower , -1);
	qu_class_method_instz (QuBufferClass, "isname"  , qu__string_isname  , -1);
	qu_class_method_instz (QuBufferClass, "isodigit", qu__string_isodigit, -1);
	qu_class_method_instz (QuBufferClass, "ispath"  , qu__string_ispath  , -1);
	qu_class_method_instz (QuBufferClass, "isprint" , qu__string_isprint , -1);
	qu_class_method_instz (QuBufferClass, "ispwd"   , qu__string_ispwd   , -1);
	qu_class_method_instz (QuBufferClass, "ispunct" , qu__string_ispunct , -1);
	qu_class_method_instz (QuBufferClass, "isspace" , qu__string_isspace , -1);
	qu_class_method_instz (QuBufferClass, "istext"  , qu__string_istext  , -1);
	qu_class_method_instz (QuBufferClass, "isupper" , qu__string_isupper , -1);
	qu_class_method_instz (QuBufferClass, "isxdigit", qu__string_isxdigit, -1);
	qu_class_method_instz (QuBufferClass, "iswhite" , qu__string_iswhite , -1);
	qu_class_method_instz (QuBufferClass, "isword"  , qu__string_isword  , -1);

	qu_class_method_instz (QuBufferClass, "asc", qu__string_asc, 0);
	qu_class_method_instz (QuBufferClass, "ord", qu__string_ord, 0);

	/* finalize declaration */
	qu_class_enddecl (QuBufferClass);
}
