/* Class.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/* standard attribute name */
#define QU_ATTRS_IN_SOURCE_DECL
#include "attrs.h"


/*
	!!! since 1.21.04 the Class class must no longer define any attribute
	!!! they have all been moved to Object

	This makes attribute lookup more efficient.
*/



/*) class Class __abstract
	All classes are instances of the Class class.

	The Class class is internally marked __final and __abstract. You can
	not create class objects directly. You must use the k{class} declaration
	syntax instead.

	@note
	Note that you can use k{eval} to build Class objects on the fly.

	@xmp
		s =
		`
			class Foo __final
				"""Foo documentation"""
				var bar
				sub baz (x)
					return x * 2
				;;
			;;
		`
		module = eval (s, "DT")
		println (module)

		Foo = module.Foo
		println (Foo)

		x = Foo ()
		println (x)
		println (Foo.__decl)
		println (Foo.__doc)
		println (Foo.__isfinal)

	@out
		<Module eval0>
		<Class eval0.Foo>
		<a Foo>
		class Foo __final
		Foo documentation
		true
*/
qu_class QuClassClass;


/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

/* allocation must only be done via this function */
static qu_class class_alloc (int id, qu_module mod, qu_string name, long flags, size_t size,
                              qu_destroy_f f, qu_mark_f m)
{
	register qu_class c;

	/*
		dont use qu_gc_new because there is one single time where QuClassClass is
		not yet initialized
	*/

	c          = (qu_class) qu_gc__new (QuClassClass, sizeof (qu_class_s));
	c->id      = id;
	c->flags   = flags;
	c->attrs   = qu_attr_alloc ();
	c->size    = size;
	c->name    = name;
	c->parent  = (qu_symbol) mod;
	c->super   = 0;
	c->dtor    = f;
	c->mark    = m;
	c->accel   = NULL;
	c->mixin   = NULL;
	c->get     = (qu_attrget_f) qu_attr_inst_get;
	c->set     = (qu_attrset_f) qu_attr_inst_set;
	c->init    = NULL;
	c->fnew    = NULL;
	return c;
}



/*----------------------------------------------------------------------------
	prototype
----------------------------------------------------------------------------*/

static int class_check_method_arg (qu_class self, qu_string name, qu_sub u, int narg)
{
	if ((narg != -1) && u->argc != narg)
	{
		QU_THROWF (QuECompile, "class %s: %s must take %d argument(s)",
			self->name->str, name->str, narg);
	}
	return 1;
}



static int class_check_throw (qu_class self, qu_string name, const char *msg)
{
	QU_THROWF (QuECompile, "class %s: %s must be a %s method",
		self->name->str, name->str, msg);
}



#if 0 /* not used */
static int class_check_static_public_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_static (a)
		|| qu_sub_is_private (a))
		return class_check_throw (self, name, "static public");
	return class_check_method_arg (self, name, a, narg);
}
#endif



#if 0 /* not used */
static int class_check_static_private_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_static (a)
		|| !qu_sub_is_private (a))
		return class_check_throw (self, name, "static private");
	return class_check_method_arg (self, name, a, narg);
}
#endif



#if 0 /* not used */
static int class_check_static_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_static (a))
		return class_check_throw (self, name, "static");
	return class_check_method_arg (self, name, a, narg);
}
#endif



static int class_check_class_public_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
			|| !qu_sub_is_class (a)
			|| qu_sub_is_private (a))
		return class_check_throw (self, name, "class public");
	return class_check_method_arg (self, name, a, narg);
}



static int class_check_class_private_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_class (a)
		|| !qu_sub_is_private (a))
		return class_check_throw (self, name, "class private");
	return class_check_method_arg (self, name, a, narg);
}



static int class_check_class_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_class (a))
		return class_check_throw (self, name, "class");
	return class_check_method_arg (self, name, a, narg);
}



#if 0 /* not used */
static int class_check_instance_public_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_inst (a)
		|| qu_sub_is_private (a))
		return class_check_throw (self, name, "instance public");
	return class_check_method_arg (self, name, a, narg);
}
#endif



static int class_check_instance_private_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_inst (a)
		|| !qu_sub_is_private (a))
		return class_check_throw (self, name, "instance private");
	return class_check_method_arg (self, name, a, narg);
}



static int class_check_instance_method (qu_class self, qu_string name, qu_any a, int narg)
{
	if (!qu_sub_is (a)
		|| !qu_sub_is_inst (a))
		return class_check_throw (self, name, "instance");
	return class_check_method_arg (self, name, a, narg);
}



qu_string qu_class_check_prototype (qu_class self, qu_string name, qu_symbol attr)
{
	int r = 1;

	name = qu_string_get_static (name);

	if (name == QU_NAME__inst)
	{
		/* help to make creation easier for C classes */
		if (qu_class_is_extern (self))
			qu_flag_set (attr, QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_FINAL);

		r = class_check_class_private_method (self, name, attr, -1);
	}
	else if (name == QU_NAME__call)
	{
		r = class_check_instance_method (self, name, attr, -1);
	}
	else if (name == QU_NAME__load)
	{
		r = class_check_class_method (self, name, attr, 1);
	}
	else if (name == QU_NAME_new)
		/* || name == QU_NAME__version) */
	{
		r = class_check_class_public_method (self, name, attr, -1);
	}
	else if (name == QU_NAME__init)
	{
		/* help to make creation easier for C classes */
		if (qu_class_is_extern (self))
			qu_flag_set (attr, QU_SUB_FLAG_PRIVATE);

		r = class_check_instance_private_method (self, name, attr, -1);
		self->init = (qu_sub) attr;
	}
	else if (name == QU_NAME__destroy)
	{
		r = class_check_instance_private_method (self, name, attr, 0);

		/*
			set the flag to help GC so it doesn't have to do an attribute lookup
			everytime an object is destroyed
		*/
		qu_flag_set (self, QU_CLASS_FLAG_DESTROY);
	}
	else if (name == QU_NAME__clear
	      || name == QU_NAME__copy
	      || name == QU_NAME__decr
		  || name == QU_NAME__each
		  || name == QU_NAME__hash
		  || name == QU_NAME__incr
		  || name == QU_NAME__inot
	      || name == QU_NAME__len
		  || name == QU_NAME__neg
		  || name == QU_NAME__pop
		  || name == QU_NAME__pos
		  || name == QU_NAME__print
		  || name == QU_NAME__string
		  || name == QU_NAME__true)
	{
		r = class_check_instance_method (self, name, attr, 0);
	}
	else if (name == QU_NAME__as
		  || name == QU_NAME__del
		  || name == QU_NAME__eq
		  || name == QU_NAME__getitem
		  || name == QU_NAME__format
		  || name == QU_NAME__push
		  || name == QU_NAME__cmp
		  || name == QU_NAME__div
		  || name == QU_NAME__iand
		  || name == QU_NAME__idiv
		  || name == QU_NAME__in
		  || name == QU_NAME__ior
		  || name == QU_NAME__ixor
		  || name == QU_NAME__like
		  || name == QU_NAME__lshift
		  || name == QU_NAME__minus
		  || name == QU_NAME__rem
		  || name == QU_NAME__mul
		  || name == QU_NAME__of
		  || name == QU_NAME__plus
		  || name == QU_NAME__pow
		  || name == QU_NAME__rshift
		  || name == QU_NAME__store)
	{
		r = class_check_instance_method (self, name, attr, 1);
	}
	else if (name == QU_NAME__setitem
		  || name == QU_NAME__getslice)
	{
		r = class_check_instance_method (self, name, attr, 2);
	}
	else if (name == QU_NAME__setslice)
	{
		r = class_check_instance_method (self, name, attr, 3);
	}
	else if (name == QU_NAME__get)
	{
		/* force private */
		qu_flag_set (attr, QU_SUB_FLAG_PRIVATE);
		r = class_check_instance_method (self, name, attr, 1);
	}
	else if (name == QU_NAME__set)
	{
		/* force private */
		qu_flag_set (attr, QU_SUB_FLAG_PRIVATE);
		r = class_check_instance_method (self, name, attr, 2);
	}

	QU_ENSURE (r);
	return name;
}



static void class_set_accel (qu_class self)
{
	qu_sub u;

#define a_(type, fname) \
	u = qu_attr_find (self->attrs, QU_NAME__##fname); \
	if (u == 0) \
		u = qu_attr_find (QuObjectClass->attrs, QU_NAME__##fname); \
	if (u && qu_sub_is (u) && qu_sub_is_extern (u)) \
	{ \
		if (self->accel == NULL) \
			self->accel = qu_mem_calloc (1, sizeof (qu_claccel_s)); \
		self->accel->f##fname = (type) u->value; \
	}

	a_(qu_sub1_f, as      )
	a_(qu_sub0_f, clear   )
	a_(qu_sub1_f, cmp     )
	a_(qu_sub0_f, copy    )
	a_(qu_sub0_f, decr    )
	a_(qu_sub1_f, del     )
	a_(qu_sub1_f, div     )
	a_(qu_sub0_f, each    )
	a_(qu_sub1_f, eq      )
	a_(qu_sub1_f, format  )
	a_(qu_sub1_f, getitem )
	a_(qu_sub2_f, getslice)
	a_(qu_sub0_f, hash    )
	a_(qu_sub1_f, iand    )
	a_(qu_sub1_f, idiv    )
	a_(qu_sub1_f, in      )
	a_(qu_sub0_f, incr    )
	a_(qu_subn_f, init    )
	a_(qu_sub0_f, inot    )
	a_(qu_sub1_f, ior     )
	a_(qu_sub1_f, ixor    )
	a_(qu_sub0_f, len     )
	a_(qu_sub1_f, like    )
	a_(qu_sub1_f, lshift  )
	a_(qu_sub1_f, minus   )
	a_(qu_sub1_f, mul     )
	a_(qu_sub0_f, neg     )
	a_(qu_sub1_f, of      )
	a_(qu_sub1_f, plus    )
	a_(qu_sub0_f, pop     )
	a_(qu_sub0_f, pos     )
	a_(qu_sub1_f, pow     )
	a_(qu_sub0_f, print   )
	a_(qu_sub1_f, push    )
	a_(qu_sub1_f, rem     )
	a_(qu_sub1_f, rshift  )
	a_(qu_sub2_f, setitem )
	a_(qu_sub3_f, setslice)
	a_(qu_sub0_f, string  )
	a_(qu_sub0_f, true    )

#undef a_

}



void qu_class_del_accel (qu_class self, qu_string name)
{
	qu_claccel_s *a = self->accel;

	if (a == NULL)
		return;

#define if_(type, fname) \
	if (name == QU_NAME__##fname) \
	{ \
		a->f##fname = NULL; \
	}

	name = qu_string_get_static (name);

	if_(qu_sub1_f, as      ) else
	if_(qu_sub0_f, clear   ) else
	if_(qu_sub1_f, cmp     ) else
	if_(qu_sub0_f, copy    ) else
	if_(qu_sub1_f, del     ) else
	if_(qu_sub0_f, decr    ) else
	if_(qu_sub1_f, div     ) else
	if_(qu_sub0_f, each    ) else
	if_(qu_sub1_f, eq      ) else
	if_(qu_sub1_f, format  ) else
	if_(qu_sub1_f, getitem ) else
	if_(qu_sub2_f, getslice) else
	if_(qu_sub0_f, hash    ) else
	if_(qu_sub1_f, iand    ) else
	if_(qu_sub1_f, idiv    ) else
	if_(qu_sub1_f, in      ) else
	if_(qu_sub0_f, incr    ) else
	if_(qu_subn_f, init    ) else
	if_(qu_sub0_f, inot    ) else
	if_(qu_sub1_f, ior     ) else
	if_(qu_sub1_f, ixor    ) else
	if_(qu_sub1_f, like    ) else
	if_(qu_sub0_f, len     ) else
	if_(qu_sub1_f, lshift  ) else
	if_(qu_sub1_f, minus   ) else
	if_(qu_sub1_f, mul     ) else
	if_(qu_sub0_f, neg     ) else
	if_(qu_sub1_f, of      ) else
	if_(qu_sub1_f, plus    ) else
	if_(qu_sub0_f, pop     ) else
	if_(qu_sub0_f, pos     ) else
	if_(qu_sub1_f, pow     ) else
	if_(qu_sub0_f, print   ) else
	if_(qu_sub1_f, push    ) else
	if_(qu_sub1_f, rem     ) else
	if_(qu_sub1_f, rshift  ) else
	if_(qu_sub2_f, setitem ) else
	if_(qu_sub3_f, setslice) else
	if_(qu_sub0_f, string  ) else
	if_(qu_sub0_f, true    )

#undef if_

}



/*----------------------------------------------------------------------------
	declaration and inheritance - compile time only
----------------------------------------------------------------------------*/
/*
	This only creates the Class object.

	Note that if we inherit the qu_object methods here then the compiler will
	have a problem when checking for duplicate symbols, i.e. it must do
	some special coding to avoid reporting error when an overridable method
	is declared (e.g. __init). The solution is to have the compiler do
	inheritance later. qu_code.c does this.
*/
/* compile time only */
qu_class qu_class_declare (qu_module m, qu_string name, long flags)
{
	assert (m);
	return class_alloc (QU_CLASS_ID_USER, m, name, flags, sizeof (qu_object_s), NULL, NULL);
}



/* compile time only */
int qu_class_decl_attr (qu_class self, qu_string name, qu_symbol v)
{
	assert (qu_var_is (v) || qu_sub_is (v));

	name = qu_class_check_prototype (self, name, (qu_symbol) v);
	QU_ENSURE (name);

	if (qu_class_is_mixin (self))
		QU_REQUIREZ (QuECompile, !qu_flag_test (v, QU_SYMBOL_FLAG_ABSTRACT),
			"mixin class can not have abstract attributes");

	qu_attr_set (self->attrs, name, v);

	if (qu_var_is (v))
		((qu_var) v)->name = name;
	else
		((qu_sub) v)->name = name;

	return 1;
}



/*----------------------------------------------------------------------------
	inheritance

	!!! We let C classes control everything.
	!!! Maybe this is not such a good idea, but if we dont do this
	!!! then overriding methods such as __inst will result in a error,
	!!! and it shouldn't be an error.

	... they can just set self->super = xxx anyway
	... and it does make things easier here ...

	!!! always set super class before mixins
	!!! this is what the compiler does

----------------------------------------------------------------------------*/

static int class_check_attr (qu_class self, qu_class sup)
{
	qu_string name;
	qu_symbol sa, a;
	long      i = 0;

	while ((sa = qu_attr_walk (sup->attrs, &name, &i)) != 0)
	{
		assert (qu_var_is (sa) || qu_sub_is (sa));

		if (qu_symbol_is_local (sa))
			continue;

		a = qu_attr_find (self->attrs, name);
		if (a)
			assert (qu_var_is (a) || qu_sub_is (a));

		if (a == 0 || a == sa)
		{
			/* RULE --- non-abstract class must redefine abstract methods */
			if (qu_symbol_is_abstract (sa) && !qu_class_is_abstract (self))
				QU_THROWF (QuECompile, "class %s: must override abstract method %s",
						self->name->str, name->str);
			continue;
		}

		/* just in case ... */
		if (sa == a)
			continue;

		/* RULE --- final attribute may not be redefined */
		if (qu_symbol_is_final (sa))
		{
			/* unless they came from a C class */
			if (! (qu_symbol_is_extern (a) && qu_symbol_is_extern (sa)))
				QU_THROWF (QuECompile, "class %s: cant redefine final attribute %s",
							self->name->str, name->str);
		}

		if (qu_var_is (sa))
		{
			if (qu_sub_is (a))
			{
				/* RULE -- consistency */
				if (((qu_sub) a)->argc != 0)
					QU_THROWF (QuECompile, "class %s: %s must be a property or a method with no arguments",
							self->name->str, name->str);
			}
			else
			{
				/* RULE -- consistency */
				if (qu_var_is_class (sa))
				{
					if (!qu_var_is_class (a))
						QU_THROWF (QuECompile, "class %s: %s must be a class or static property",
								self->name->str, name->str);
				}
				else
				{
					if (!qu_var_is_inst (a))
						QU_THROWF (QuECompile, "class %s: %s must be an instance property",
								self->name->str, name->str);
				}
#if 0
				/* RULE -- if typed, it must be typed too */
				if (((qu_var) sa)->type)
				{
					if (((qu_var) a)->type == 0)
						QU_THROWF (QuECompile, "class %s: %s must be typed",
								self->name->str, name->str);
				}
#endif
				/* ignore qu_object type */
				if (((qu_var) a)->type == QuObjectClass)
					((qu_var) a)->type = 0;
			}
		}
		else
		{
			/* RULE -- consistency */
			if (qu_sub_is_static (sa))
			{
#if 0 /* let be, it make things simpler and makes Object.__cmp, __eq, etc. work */
				if (!qu_sub_is_static (a))
					QU_THROWF (QuECompile, "class %s: %s must be a static method",
							self->name->str, name->str);
				*/
#endif
			}
			else if (qu_sub_is_class (sa))
			{
				if (!(qu_sub_is_class (a) || qu_sub_is_static (a)))
					QU_THROWF (QuECompile, "class %s: %s must be a class or static method",
							self->name->str, name->str);
			}
			else
			{
				if (!qu_sub_is_inst (a))
					QU_THROWF (QuECompile, "class %s: %s must be an instance method",
							self->name->str, name->str);
			}
		}
	}
	return 1;
}



/* compile time only */
int qu_class_set_super (qu_class self, qu_class sup)
{
	if (sup == QuClassClass || sup == QuObjectClass)
		QU_THROWF (QuECompile, "%s: cant use %s as super class",
			self->name->str, sup->name->str);

	if (qu_class_is_extern (self))
	{
		self->super = sup;
		return 1;
	}

	if (self->super)
	{
		QU_THROWF (QuECompile, "%s: already has super class %s",
			self->name->str, self->super->name->str);
	}

	/* RULE -- final classes can not be super classes */
	if (qu_class_is_final (sup))
	{
		QU_THROWF (QuECompile, "%s: cant extend final class %s",
			self->name->str, sup->name->str);
	}

	/* RULE -- class can not be the super class of itself or its subclass */
	{
		qu_class u;
		for (u = sup; u != 0; u = u->super)
		{
			if (self == u)
				QU_THROWF (QuECompile, "class %s: cant extend itself or its subclass",
					self->name->str);
		}
	}

	if (qu_class_is_mixin (self))
	{
		/* RULE -- __mixin classes may only have another __mixin class
			as the super class */
		if (!qu_class_is_mixin (sup))
			QU_THROWF (QuECompile, "class %s: __mixin class may only have another "
					"__mixin class as the super class",
				self->name->str);
	}
	else
	{
		/* RULE -- can not use an __mixin as a super class of a regular class */
		if (qu_class_is_mixin (sup))
			QU_THROWF (QuECompile, "class %s: cant use an __mixin class as a "
					"super class of a regular class",
				self->name->str);
	}

	/* RULE -- exception classes name must be EXxxx */
	{
		if (qu_class_isa (sup, QuExceptionClass))
		{
			if (self->name->len < 3
				|| self->name->str[0] != 'E'
				|| !qu_ascii_isupper (self->name->str[1]))
				QU_THROWF (QuECompile, "class %s: exception classes must be named EXx...",
					self->name->str);
		}
	}

#if 0 /* let qu_class_enddecl do it */
	QU_ENSURE (class_check_attr (self, sup));
#endif

	self->super = sup;

	/*
		!!! must do this in qu_class_enddecl as well since the super class may be
		!!! declared later

		use Net
		use Ssl

		class InetSocket is SslSocket
		;;

		class SslSocket is Net.Socket
		;;
	*/

	if (self->size < sup->size)
		self->size = sup->size;

	if (!qu_class_is_extern (self))
	{
		self->dtor = sup->dtor;
		self->mark = sup->mark;
	}
	return 1;
}



/* compile time only */
int qu_class_set_mixin (qu_class self, qu_class sup)
{
	if (sup == QuClassClass || sup == QuObjectClass)
		QU_THROWF (QuECompile, "%s: cant use %s as mixin class",
			self->name->str, sup->name->str);

	if (sup == self->super)
		QU_THROWF (QuECompile, "class %s: %s is already the super class",
			self->name->str, sup->name->str);

	if (self->mixin)
	{
		/* not more than once */
		if (qu_array_contains (self->mixin, sup))
			QU_THROWF (QuECompile, "class %s: %s already used",
				self->name->str, sup->name->str);
	}

	/* RULE --- final classes may not be used as mixins */
	if (qu_class_is_final (sup))
		QU_THROWF (QuECompile, "class %s: cant extend final class %s",
			self->name->str, sup->name->str);

	/* RULE --- C classes may not be used as mixins unless they truly are mixins */
	if (!qu_class_is_extern (self))
	{
		if (qu_class_is_extern (sup) && !qu_class_is_mixin (sup))
			QU_THROWF (QuECompile, "class %s: cant mix external class %s",
				self->name->str, sup->name->str);
	}

	/* RULE --- class can not mix itself or its subclass */
	{
		qu_class u;
		for (u = sup; u != 0; u = u->super)
		{
			if (self == u)
				QU_THROWF (QuECompile, "class %s: cant extend itself or its subclass",
					self->name->str);
		}
	}

#if 0 /* let qu_class_enddecl do it */
	QU_ENSURE (class_check_attr (self, sup));
#endif

	if (self->mixin == 0)
		self->mixin = qu_array_new (4);

	QU_ENSURE (qu_array_push (self->mixin, sup));
	return 1;
}



/* compile-time only */
qu_any qu_class_search_attr (qu_class self, qu_string name)
{
	qu_any x = qu_attr_search (self, name);

	if (x && !qu_flag_test (x, QU_SYMBOL_FLAG_LOCAL))
		return x;

#if 0 /* QuClassClass no longer defines any attribute since 1.21.04 */
	x = qu_attr_find (QuClassClass->attrs, name);
	if (x && !qu_flag_test (x, QU_SYMBOL_FLAG_LOCAL))
		return x;
#endif

	x = qu_attr_find (QuObjectClass->attrs, name);
	if (x) /* && !qu_flag_test (x, QU_SYMBOL_FLAG_LOCAL)) */
		return x;

	return 0;
}



static qu_class class_find_class (qu_class c, qu_string name, qu_class org)
{
	if (c != org)
	{
		if (qu_string_EQ (c->name, name))
			return c;

		if (c->mixin)
		{
			long   n = c->mixin->len;
			qu_class *p = (qu_class *) (c->mixin->objs + n);
			while (n--)
			{
				qu_class x = class_find_class (*--p, name, c);
				if (x)
					return x;
			}
		}
		if (c->super)
			return class_find_class (c->super, name, c);
	}
	return 0;
}



qu_class qu_class_find_class (qu_class self, qu_string name)
{
	return class_find_class (self, name, 0);
}


/*----------------------------------------------------------------------------
	merging attributes

	After qu_class_enddecl, self->attrs will contain all attributes from super
	classes and mixin classes, except from qu_class and qu_object.

	class2  - - - - - - - - - - already done above this line
	|
	|       smix111, ssmix112
	|         |
	class1, mix11, mix12

	for class1 set all defined by x unless already defined, where x walks through:

	class2 -> mix12 -> mix11 -> ssmix112 -> smix111

----------------------------------------------------------------------------*/

/* compile time only */
static void class_merge (qu_class self, qu_class u)
{
#if 0
	if (u->super)
	{
#if 0
		class_merge (self, u->super);
#else
		qu_class  c = u->super;
		long   i = 0;
		qu_any    x;
		qu_string name;

		while ((x = qu_attr_walk (c->attrs, &name, &i)) != 0)
		{
			if (qu_attr_find (self->attrs, name) == 0)
			{
#if 0
				printf ("%s.%s <- %s.%s\n", self->name->str, name->str, u->name->str, name->str);
#endif
				qu_attr_add (self->attrs, name, x);
			}
		}
#endif
	}

	if (u != self)
	{
#endif
		long i = 0;
		qu_any  x;
		qu_string  name;

		while ((x = qu_attr_walk (u->attrs, &name, &i)) != 0)
		{
			if (qu_attr_find (self->attrs, name) == 0)
			{
#if 0
				printf ("%s.%s <- %s.%s\n", self->name->str, name->str, u->name->str, name->str);
#endif
				qu_attr_add (self->attrs, name, x);
			}
		}
#if 0
	}

	if (u->mixin)
	{
		long   n = u->mixin->len;
		qu_class *c = (qu_class *) (u->mixin->objs + n);

		while (n--)
			class_merge (self, *--c);
	}
#endif
}


/* init & compile time only */
int qu_class_enddecl (qu_class self)
{

#if 0
	qu_any    x;
	qu_string name;
	long      i = 0;

	while ((x = qu_attr_walk (self->attrs, &name, &i)) != 0)
		printf ("%s.%s\n", self->name->str, name->str);
#endif

	/* don't doit when documenting */
	if (!qu_class_is_doc (self))
	{
		qu_class sup = self->super;
		qu_array mix = self->mixin;

		if (!qu_class_is_extern (self))
			QU_ENSURE (class_check_attr (self, QuObjectClass));

		if (mix)
		{
			long      n = mix->len;
			qu_class *c = (qu_class *) (mix->objs + n);

			while (n--)
				class_merge (self, *--c);
		}

		if (sup)
			class_merge (self, sup);

		/* now the final check */
		if (!qu_class_is_extern (self))
		{
			if (mix)
			{
				long      n = mix->len;
				qu_class *c = (qu_class *) (mix->objs + n);

				while (n--)
					QU_ENSURE (class_check_attr (self, *--c));
			}

			while (sup)
			{
				QU_ENSURE (class_check_attr (self, sup));
				mix = sup->mixin;
				if (mix)
				{
					long      n = mix->len;
					qu_class *c = (qu_class *) (mix->objs + n);

					while (n--)
						QU_ENSURE (class_check_attr (self, *--c));
				}
				sup = sup->super;
			}
		}

		/* and finally to speed things up a bit */
		if (self->init == 0)
		{
			self->init = qu_class_search_attr (self, QU_NAME__init);
			if (self->init)
				assert (qu_sub_is (self->init) && qu_sub_is_inst (self->init));
		}

		if (self->fnew == 0)
		{
			self->fnew = qu_class_search_attr (self, QU_NAME_new);
			if (self->fnew == 0)
				self->fnew = __QuObjectNew;
			else
			{
				assert (qu_sub_is (self->fnew));
				assert (qu_sub_is_class (self->fnew) || qu_sub_is_static (self->fnew));
			}
		}
	}

	if (qu_class_is_extern (self))
	{
#ifdef QU_DEBUG_LOCAL
		if (self->accel)
			QU_FATAL ("%s: accelerators must no longer be set by classes", self->name->str);
#endif
		class_set_accel (self);
	}

#if 0
	/* to see the difference */
	self->accel = NULL;
#endif

	/* !!! this might come later - see qu_class_set_super */
	if (!qu_class_is_extern (self) && self->super)
	{
		qu_class sup = self->super;

		if (self->size < sup->size)
			self->size = sup->size;

		self->dtor = sup->dtor;
		self->mark = sup->mark;
	}
#if 0
	printf ("%s => free = %p  mark = %p\n", self->name->str, self->dtor, self->mark);
#endif

	return 1;
}



int qu_class_can_access (qu_class self, qu_symbol from)
{
	qu_symbol owner = self->parent;

	if (owner == from)
		return 1;

	if (!qu_class_is_private (self))
		return 1;

	if (qu_module_is (from))
	{
		return 0;
	}
	else if (qu_class_is (from))
	{
		return owner == ((qu_class) from)->parent;
	}
	else
	{
		assert (qu_sub_is (from));
		return owner == (qu_symbol) qu_sub_get_module ((qu_sub) from);
	}
	return 0;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__class__mark (qu_class self)
{
	qu_gc_xmark (self->super);
	qu_gc_xmark (self->mixin);
	qu_gc_mark  (self->parent);
	qu_gc_mark  (self->name);
}



static void qu__class__destroy (qu_class self)
{
	if (self->accel)
		qu_mem_free (self->accel);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/* none - all moved to Object since 1.21.04 */



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/* none - all moved to Object since 1.21.04 */



/*----------------------------------------------------------------------------
	methods designed for C modules
----------------------------------------------------------------------------*/

qu_class qu_class_new (qu_module mod, qu_string name, long flags, size_t size, qu_destroy_f f, qu_mark_f m)
{
	qu_class c;

	assert (mod && qu_module_is (mod));
	assert (size >= sizeof (qu_object_s));
	assert (qu_module_getattr (mod, name) == 0);

	c = class_alloc (QU_CLASS_ID_USER, mod, name, flags | QU_CLASS_FLAG_EXTERN, size, f, m);
	qu_module_setattr (mod, c->name, (qu_symbol) c);
	return c;
}



qu_var qu_class_prop (qu_class self, qu_string name, qu_any value, long flags)
{
	qu_var v = qu_var_new ((qu_symbol) self, name, flags);

	name = qu_class_check_prototype (self, name, (qu_symbol) v);
	if (name == 0)
	{
		qu_string s = qu_thread_err_string ();
		QU_FATAL ("%s", s->str);
	}
	v->name = name;
	v->value = value;
	qu_attr_set (self->attrs, name, v);
	return v;
}



qu_sub qu_class_method (qu_class self, qu_string name, void *cfunc, int argc, long flags)
{
	qu_sub s = qu_sub_new ((qu_symbol) self, name, cfunc, argc, QU_SUB_FLAG_EXTERN | flags);

	name = qu_class_check_prototype (self, name, (qu_symbol) s);
	if (name == 0)
	{
		qu_string s = qu_thread_err_string ();
		QU_FATAL ("%s", s->str);
	}
	s->name = name;
	qu_attr_set (self->attrs, name, s);
	return s;
}



qu_any qu_class_alias (qu_class self, qu_string name, qu_string alias)
{
	qu_any x = qu_attr_find (self->attrs, name);

	if (x == 0)
		QU_FATAL ("class %s has no such symbol %s", self->name->str, name->str);

	if (qu_attr_find (self->attrs, alias))
		QU_FATAL ("class %s already has symbol %s", self->name->str, name->str);

	qu_attr_set (self->attrs, alias, x);
	return x;
}



qu_any qu_class_init_new (qu_class self, qu_any inst, int argc, qu_any *argv)
{
	qu_sub u;

	assert (!qu_class_is (inst));
	if (self->init)
	{
		u = self->init;
	}
	else
	{
		u = self->get (inst, QU_NAME__init, 0);
		if (u == 0)
			return inst;
	}
	assert (qu_sub_is_inst (u));
	if (u->parent != (qu_symbol) QuObjectClass)
		QU_ENSURE (qu_sub_calln (u, inst, argc, argv));
	return inst;
}



/*----------------------------------------------------------------------------
	methods for general purposes
----------------------------------------------------------------------------*/

/* this is provided for efficiency over qu__class_isa */
int qu_class_isa (qu_class self, qu_class x)
{
	if (x == QuClassClass || x == QuObjectClass)
		return 1;
	do
	{
		if (x == self)
			return 1;
		self = self->super;

	} while (self);

	return 0;
}



static int class_is_super_or_mixin (qu_class self, qu_class x)
{
	if (self->super)
	{
		if (self->super == x)
			return 1;
		if (class_is_super_or_mixin (self->super, x))
			return 1;
	}

	if (self->mixin)
	{
		long   n = self->mixin->len;
		qu_class *p = (qu_class *) self->mixin->objs;
		while (n--)
		{
			if (*p == x)
				return 1;
			if (class_is_super_or_mixin (*p, x))
				return 1;
			++p;
		}
	}
	return 0;
}



/* whether x is a mixin or superclass of self*/
int qu_class_is_super_or_mixin (qu_class self, qu_class x)
{
	return class_is_super_or_mixin (self, x);
}



/* get class depth in its hierarchy */
int qu_class_depth (qu_class self)
{
	int l = 0;
	while ((self = self->super) != 0)
		l++;
	return l;
}



/*----------------------------------------------------------------------------
	methods specifically designed for builtin classes
----------------------------------------------------------------------------*/

/*
	called from xxx__setup to initialize the qu_class_s struct.
*/
void qu_class__setup (qu_class self, long flags, size_t size, qu_destroy_f f, qu_mark_f m)
{
	assert (size >= sizeof (qu_object_s));

	self->parent = (qu_symbol) QuSysModule;
	self->size   = size;
	self->flags |= flags | QU_CLASS_FLAG_EXTERN;
	self->dtor   = f;
	self->mark   = m;
	self->accel  = NULL;
}



/*----------------------------------------------------------------------------
	set/get instance attributes - runtime
----------------------------------------------------------------------------*/

static qu_any class__get (qu_class self, qu_string name, int restrict)
{
	qu_any x;

	assert (qu_class_is (self));

	/*HACK - documentation */
	if (qu_class_is_doc (self))
	{
		if (strcmp (self->name->str, "Class") == 0
			|| strcmp (self->name->str, "Object") == 0)
		{
			x = qu_attr_find (QuObjectClass->attrs, name);
			if (x == 0)
				goto nope;
			goto okay;
		}
	}

	x = qu_attr_find (self->attrs, name);
	if (x == 0)
	{
		x = qu_attr_find (QuObjectClass->attrs, name);
		if (x == 0)
			goto nope;
	}

	if (qu_flag_test (x, QU_SYMBOL_FLAG_LOCAL | QU_SYMBOL_FLAG_INST)
		|| (restrict && qu_flag_test (x, QU_SYMBOL_FLAG_PRIVATE)))
		goto nope;

  okay:

	if (qu_var_is (x))
		return ((qu_var) x)->value;
	return x;

  nope:
	/* to prevent throwing expensive exception message */
	if (__QuAttrNoThrow)
		return 0;
	QU_THROWF (QuEImplement, "%s has no such attribute %s", self->name->str, name->str);
}



static qu_any class__set (qu_class self, qu_string name, qu_any val, int restrict)
{
	qu_any x;

	if (self == QuClassClass)
		QU_THROWF (QuEImplement, "Class has not settable attribute");

	assert (qu_class_is (self));
	x = qu_attr_find (self->attrs, name);
	if (x == 0)
	{
		x = qu_attr_find (QuObjectClass->attrs, name);
		if (x == 0)
			goto nope;
	}
	if (qu_flag_test (self, QU_SYMBOL_FLAG_LOCAL | QU_SYMBOL_FLAG_INST)
		|| (restrict && qu_flag_test (x, QU_SYMBOL_FLAG_PRIVATE)))
		goto nope;

	if (qu_var_is (x))
	{
 		if (qu_var_frozen (x) || (restrict && !qu_var_is_rdwr (x)))
			goto modify;

		if (((qu_var) x)->type)
		{
			val = qu_object_validate (self, name, val);
			if (val == 0)
				goto bad;
		}
		((qu_var) x)->value = val;
		return val;
	}
	else
	{
		assert (qu_sub_is (x));
#ifdef QU_SUB_ALLOW_ASSIGN
		return qu_sub_call1 (x, self, val);
#else
		goto modify;
#endif
	}

  nope:   QU_THROWF (QuEImplement, "%s has no such attribute %s", self->name->str, name->str);
  bad:    QU_THROWF (QuEValue    , "mismatch value type %s.%s"  , self->name->str, name->str);
  modify: QU_THROWF (QuEModify   , "cant modify %s.%s"          , self->name->str, name->str);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

static void class_pre (qu_class c, const char *name)
{
	c->name   = qu_string_new_staticz (name);
	c->parent = (qu_symbol) QuSysModule;
	qu_module_setattr (QuSysModule, c->name, (qu_symbol) c);
}


static void class_set_special_flags ()
{
	/* interp uses an internal iterator for these classes */
	qu_flag_set (QuStringClass  , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuArrayClass   , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuDictClass    , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuArgClass     , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuBufferClass  , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuSetClass     , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuFrameClass   , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuRangeClass   , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuIteratorClass, QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuRecordClass  , QU_CLASS_FLAG_AUTOITER);
	qu_flag_set (QuStructClass  , QU_CLASS_FLAG_AUTOITER);

	/* cant make weak references to instances of these classes */
	qu_flag_set (QuNullClass    , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuBooleanClass , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuStringClass  , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuIntClass     , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuLongClass    , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuRationalClass, QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuBigClass     , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuFloatClass   , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuInfClass     , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuComplexClass , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuRangeClass   , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuOpaqueClass  , QU_CLASS_FLAG_NOWEAK);
	qu_flag_set (QuWeakClass    , QU_CLASS_FLAG_NOWEAK);

	/* may call these abstract classes 'new' method as they return an
		instance of a real class */
	qu_flag_set (QuRealClass    , QU_CLASS_FLAG_CAST);
	qu_flag_set (QuDecimalClass , QU_CLASS_FLAG_CAST);
	qu_flag_set (QuIntegerClass , QU_CLASS_FLAG_CAST);
	qu_flag_set (QuNumberClass  , QU_CLASS_FLAG_CAST);

	/* whether the class is callable - see Interp CASE (CALL) */
	qu_flag_set (QuSubClass     , QU_CLASS_FLAG_CALLABLE);
	qu_flag_set (QuMethodClass  , QU_CLASS_FLAG_CALLABLE);
	qu_flag_set (QuClassClass   , QU_CLASS_FLAG_CALLABLE);
	qu_flag_set (QuStructClass  , QU_CLASS_FLAG_CALLABLE);

	/* whether the class is symbol */
	qu_flag_set (QuModuleClass  , QU_CLASS_FLAG_SYMBOL);
	qu_flag_set (QuClassClass   , QU_CLASS_FLAG_SYMBOL);
	qu_flag_set (QuSubClass     , QU_CLASS_FLAG_SYMBOL);
	qu_flag_set (QuVarClass     , QU_CLASS_FLAG_SYMBOL);

	/* whether instance can be made static */
	qu_flag_set (QuNullClass    , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuBooleanClass , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuIntClass     , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuLongClass    , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuRationalClass, QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuFloatClass   , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuBigClass     , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuInfClass     , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuComplexClass , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuStringClass  , QU_CLASS_FLAG_LITERAL);
	qu_flag_set (QuRangeClass   , QU_CLASS_FLAG_LITERAL);
}



void qu_class__init ()
{
	/*
		Allocate all builtin classes.
	*/

#define _alloc(id, u) \
	class_alloc (id, 0, 0, QU_CLASS_FLAG_EXTERN, sizeof (u), NULL, NULL)

	/* one time fooling gc assertion */
	{
		qu_class_s fool;
		QuClassClass = &fool;
		QuClassClass = _alloc (QU_CLASS_ID_Class, qu_class_s);
		QuClassClass->klass = QuClassClass;
	}

	QuObjectClass    = _alloc (QU_CLASS_ID_Object   , qu_object_s);
	QuArgClass       = _alloc (QU_CLASS_ID_Arg      , qu_dict_s);
	QuArrayClass     = _alloc (QU_CLASS_ID_Array    , qu_array_s);
	QuAstClass       = _alloc (QU_CLASS_ID_Ast      , qu_ast_s);
	QuBigClass       = _alloc (QU_CLASS_ID_Big      , qu_big_s);
	QuBooleanClass   = _alloc (QU_CLASS_ID_Boolean  , qu_object_s);
	QuBufferClass    = _alloc (QU_CLASS_ID_Buffer   , qu_buffer_s);
	QuCodeClass      = _alloc (QU_CLASS_ID_Code     , qu_code_s);
	QuComplexClass   = _alloc (QU_CLASS_ID_Complex  , qu_complex_s);
	QuContainerClass = _alloc (QU_CLASS_ID_Container, qu_container_s);  /* mixin */
	QuDecimalClass   = _alloc (QU_CLASS_ID_Decimal  , qu_object_s);     /* abstract */
	QuDictClass      = _alloc (QU_CLASS_ID_Dict     , qu_dict_s);
	QuDlibClass      = _alloc (QU_CLASS_ID_Dlib     , qu_dlib_s);
	QuExceptionClass = _alloc (QU_CLASS_ID_Exception, qu_exception_s);
	QuFileClass      = _alloc (QU_CLASS_ID_File     , qu_object_s);     /* qu_file_s hidden */
	QuFloatClass     = _alloc (QU_CLASS_ID_Float    , qu_float_s);
	QuFrameClass     = _alloc (QU_CLASS_ID_Frame    , qu_frame_s);
	QuFunctionClass  = _alloc (QU_CLASS_ID_Function , qu_object_s);     /* abstract */
	QuIntClass       = _alloc (QU_CLASS_ID_Int      , qu_object_s);     /* fake size */
	QuIntegerClass   = _alloc (QU_CLASS_ID_Integer  , qu_object_s);     /* abstract */
	QuInfClass       = _alloc (QU_CLASS_ID_Inf      , qu_object_s);     /* abstract */
	QuIteratorClass  = _alloc (QU_CLASS_ID_Iterator , qu_iterator_s);
	QuLongClass      = _alloc (QU_CLASS_ID_Long     , qu_long_s);
	QuMethodClass    = _alloc (QU_CLASS_ID_Method   , qu_method_s);
	QuModuleClass    = _alloc (QU_CLASS_ID_Module   , qu_module_s);
	QuNullClass      = _alloc (QU_CLASS_ID_Null     , qu_object_s);
	QuNumberClass    = _alloc (QU_CLASS_ID_Number   , qu_object_s);     /* abstract */
	QuOpaqueClass    = _alloc (QU_CLASS_ID_Opaque   , qu_opaque_s);
	QuProcClass      = _alloc (QU_CLASS_ID_Proc     , qu_proc_s);
	QuPointerClass   = _alloc (QU_CLASS_ID_Pointer  , qu_pointer_s);
	QuQueueClass     = _alloc (QU_CLASS_ID_Queue    , qu_queue_s);
	QuRangeClass     = _alloc (QU_CLASS_ID_Range    , qu_range_s);
	QuRationalClass  = _alloc (QU_CLASS_ID_Rational , qu_rational_s);
	QuRealClass      = _alloc (QU_CLASS_ID_Real     , qu_object_s);     /* abstract */
	QuRecordClass    = _alloc (QU_CLASS_ID_Record   , qu_record_s);
	QuSetClass       = _alloc (QU_CLASS_ID_Set      , qu_set_s);
	QuStreamClass    = _alloc (QU_CLASS_ID_Stream   , qu_stream_s);
	QuStringClass    = _alloc (QU_CLASS_ID_String   , qu_string_s);
	QuStructClass    = _alloc (QU_CLASS_ID_Struct   , qu_struct_s);
	QuSubClass       = _alloc (QU_CLASS_ID_Sub      , qu_sub_s);
	QuThreadClass    = _alloc (QU_CLASS_ID_Thread   , qu_thread_s);
	QuTokenizerClass = _alloc (QU_CLASS_ID_Tokenizer, qu_tokenizer_s);
	QuVarClass       = _alloc (QU_CLASS_ID_Var      , qu_var_s);
	QuWeakClass      = _alloc (QU_CLASS_ID_Weak     , qu_weak_s);

#undef _allocate

	/* set this early */
	class_set_special_flags ();

	/*
		In this order.
	*/
	qu_string__setup   ();
	qu_module__setup   ();
	qu_function__setup ();
	qu_sub__setup      ();
	qu_var__setup      ();

	/*
		Create standard attribute name strings.
	*/
#define QU_ATTRS_IN_SOURCE_INIT
#include "attrs.h"

	/*
		Setup Class.

		Since 1.21.04 the Class class must not define any attribute.
		They are all moved to Object.
		This makes better attribute lookup logic.
	*/
	class_pre (QuClassClass, "Class");
	QuClassClass->flags |= QU_CLASS_FLAG_EXTERN | QU_CLASS_FLAG_FINAL | QU_CLASS_FLAG_ABSTRACT;
	QuClassClass->klass  = QuClassClass;
	QuClassClass->mark   = (qu_mark_f   ) qu__class__mark;
	QuClassClass->dtor   = (qu_destroy_f) qu__class__destroy;
	QuClassClass->get    = (qu_attrget_f) class__get;
	QuClassClass->set    = (qu_attrset_f) class__set;

	/*
		Setup qu_object.
	*/
	class_pre (QuObjectClass, "Object");
	qu_object__setup ();

	/*
		Set accelerators for Object and Class.
		!!! in this order so the accelerators are correctly set for Class
	*/
	class_set_accel (QuObjectClass);
	class_set_accel (QuClassClass);

	/*
		First time setup
	*/
	class_pre (QuArgClass      , "Arg");
	class_pre (QuArrayClass    , "Array");
	class_pre (QuAstClass      , "Ast");
	class_pre (QuBigClass      , "Big");
	class_pre (QuBooleanClass  , "Boolean");
	class_pre (QuBufferClass   , "Buffer");
	class_pre (QuCodeClass     , "Code");
	class_pre (QuComplexClass  , "Complex");
	class_pre (QuContainerClass, "Container");
	class_pre (QuDictClass     , "Dict");
	class_pre (QuDecimalClass  , "Decimal");
	class_pre (QuDlibClass     , "Dlib");
	class_pre (QuExceptionClass, "Exception");
	class_pre (QuFileClass     , "File");
	class_pre (QuFloatClass    , "Float");
	class_pre (QuFrameClass    , "Frame");
	class_pre (QuFunctionClass , "Function");
	class_pre (QuIntClass      , "Int");
	class_pre (QuIntegerClass  , "Integer");
	class_pre (QuInfClass      , "Inf");
	class_pre (QuIteratorClass , "Iterator");
	class_pre (QuLongClass     , "Long");
	class_pre (QuMethodClass   , "Method");
	class_pre (QuModuleClass   , "Module");
	class_pre (QuNullClass     , "Null");
	class_pre (QuNumberClass   , "Number");
	class_pre (QuOpaqueClass   , "Opaque");
	class_pre (QuProcClass     , "Proc");
	class_pre (QuPointerClass  , "Pointer");
	class_pre (QuQueueClass    , "Queue");
	class_pre (QuRangeClass    , "Range");
	class_pre (QuRationalClass , "Rational");
	class_pre (QuRealClass     , "Real");
	class_pre (QuRecordClass   , "Record");
	class_pre (QuSetClass      , "Set");
	class_pre (QuStreamClass   , "Stream");
	class_pre (QuStringClass   , "String");
	class_pre (QuStructClass   , "Struct");
	class_pre (QuSubClass      , "Sub");
	class_pre (QuThreadClass   , "Thread");
	class_pre (QuTokenizerClass, "Tokenizer");
	class_pre (QuVarClass      , "Var");
	class_pre (QuWeakClass     , "Weak");

	/*
		Have builtin classes setup their qu_class_s struct.
		Order is important!
	*/
	qu_null__setup      ();
	qu_boolean__setup   ();
	qu_container__setup ();
	qu_number__setup    ();
	qu_real__setup      ();
	qu_integer__setup   ();
	qu_decimal__setup   ();
	qu_inf__setup       ();
	qu_int__setup       ();
	qu_long__setup      ();
	qu_float__setup     ();
	qu_big__setup       ();
	qu_rational__setup  ();
	qu_complex__setup   ();
	qu_exception__setup ();
	qu_stream__setup    ();

	qu_array__setup     ();
	qu_ast__setup       ();
	qu_dict__setup      ();
	qu_dlib__setup      ();
	qu_iterator__setup  ();
	qu_range__setup     ();
	qu_set__setup       ();
	qu_method__setup    ();
	qu_opaque__setup    ();
	qu_proc__setup      ();
	qu_pointer__setup   ();
	qu_queue__setup     ();
	qu_code__setup      ();
	qu_frame__setup     ();
	qu_thread__setup    ();
	qu_file__setup      (); /* must be after qu_stream */
	qu_buffer__setup    (); /* -ditto- */
	qu_struct__setup    ();
	qu_record__setup    ();
	qu_weak__setup      ();

	/*
		Let builtin classes override methods inherited from qu_object
		and do their own initialization.
		Order is important!
	*/
	qu_string__init     ();

	qu_null__init       ();
	qu_boolean__init    ();
	qu_container__init  ();
	qu_number__init     ();
	qu_real__init       ();
	qu_integer__init    ();
	qu_decimal__init    ();
	qu_inf__init        ();
	qu_int__init        ();
	qu_long__init       ();
	qu_float__init      ();
	qu_big__init        ();
	qu_rational__init   ();
	qu_complex__init    (); /* must be last after all other qu_number classes */
	qu_function__init   ();
	qu_sub__init        ();
	qu_var__init        ();
	qu_module__init     ();
	qu_exception__init  ();
	qu_stream__init     ();
	qu_weak__init       ();

	qu_array__init      ();
	qu_ast__init        ();
	qu_dict__init       ();
	qu_dlib__init       ();
	qu_iterator__init   ();
	qu_range__init      ();
	qu_set__init        ();
	qu_method__init     ();
	qu_opaque__init     ();
	qu_proc__init       ();
	qu_pointer__init    ();
	qu_queue__init      ();
	qu_code__init       ();
	qu_frame__init      ();
	qu_thread__init     ();
	qu_file__init       (); /* must be after qu_stream */
	qu_buffer__init     (); /* -ditto- */
	qu_struct__init     ();
	qu_record__init     ();
	qu_tokenizer__setup ();
	qu_tokenizer__init  ();

#ifdef QU_SUPPORT_GLIB
	QuGObjectClass       = _alloc (QU_CLASS_ID_GObject  , qu_gobject_s);
	QuGObjectClass->name = qu_string_new_staticz ("GObject");
#endif

}
