/* Code.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Code __abstract
	A Code object contains compiled bytecode which is run by the
	interpreter. It holds some information such as the block which owns
	the code, the required stack size etc.

	@xmp
		c = __main__.__code
		println (c.stacksize)
		println (c.__len)

	@out
		16
		104
*/
qu_class QuCodeClass;

/* instruction names */
qu_string __QuCodePcName [QU_PC_LAST_INSTRUCTION + 1];

/* argument types */
int __QuCodePcType [QU_PC_LAST_INSTRUCTION + 1];

/* binop function list */
qu_binop_f QU_CODE_BINOPS [QU_PC_LAST_INSTRUCTION + 1];

/* JIT code hooks */
qu_codejitgen_f   qu_code_jit_generate = NULL;
qu_codejitprint_f __QuCodeJitPrint     = NULL;


/*----------------------------------------------------------------------------
	compilation is one module at a time, one Sub at a time
----------------------------------------------------------------------------*/

static long           CoLine;        /* last line number */
static unsigned char *CoBytes;       /* the bytecode */
static int            CoSize;        /* allocated capacity of CoBytes */
static int            CoNextOfs;     /* offset of next instruction */
static int            CoPcLast;      /* last instruction */
static int            CoPcLastOfs;   /* offset of last instruction */
static int            CoStackSize;   /* required stack size */
static int            CoSp;          /* current stack depth */
static jmp_buf        CoJmpbuf;      /* to bail out on error */
static qu_module      CoModule;      /* the module object being compiled */
static qu_symbol      CoMaster;      /* current master block (module, class or sub) */
static int            CoMasterLevel; /* current master level */
static qu_symbol      CoMasterStack [QU_COMPILE_MAXLEVEL];

#define CoPath     CoModule->name
#define CoSource   CoModule->source
#define CoLiterals CoModule->literals

/* initial size of bytecode to allocate */
#define INIT_BYTECODE_SIZE 64

/* account for an instruction plus arguments when growing bytecode */
#define GROWEXTRA (SIZEOF_LONG * 8)

/* grow bytecode if necessary */
#define CHECKGROW(ofs) \
QU_BEGIN \
	if ((ofs) >= (CoSize - GROWEXTRA)) \
	{ \
		assert (CoSize < (QU_MEM_MAX >> 1)); \
		CoSize <<= 1; \
		CoBytes = qu_mem_realloc (CoBytes, CoSize); \
	} \
QU_END

/* instruction sizes */
#define L0 1
#define L1 (1 + sizeof (qu_any))
#define L2 (1 + sizeof (qu_any) + sizeof (qu_any))

#define PSIZE(pc) (1 + (__QuCodePcType [pc] & 0xF) * sizeof (long))


/*
	Start new Code.
	Must call for Module and Sub.
*/
static void code_begin (qu_any master)
{
	CoPcLast      = 0;
	CoPcLastOfs   = 0;

	CoMasterLevel = 0;
	CoMaster      = CoMasterStack[0] = master;
	CoLine        = 0;
	CoNextOfs     = 0;
	CoSp          = 0;
	CoStackSize   = 0;
	CoSize        = INIT_BYTECODE_SIZE;
	CoBytes       = qu_mem_alloc (INIT_BYTECODE_SIZE);
}



/*
	Cleanup after compiling module or error.
	!!! only qu_code_generate may call this
*/
static void code_end ()
{
	if (CoBytes)
		qu_mem_free (CoBytes);
}



/*
	Must call to enter a master block: Module, Class and Sub.
*/
static void code_enter_master (qu_any m)
{
	assert (qu_class_is (m) || qu_sub_is (m));
	CoMasterStack[++CoMasterLevel] = CoMaster = m;
}



/*
	Must call sometimes later after calling code_enter_master.
*/
static void code_leave_master ()
{
	CoMaster = CoMasterStack[--CoMasterLevel];
}



/*
	Optimize jumps.
*/
#define jaddr(ip) (*(long *) (cs + ip + 1))

static void code_optimize_jump1 (qu_code c)
{
	register unsigned char *cs = c->bytes;
	register int            ip = 0, im;
	register int            len = c->len;
	register int            pc;
	register long           j;

	while (ip < len)
	{
		pc = cs [ip];
		switch (pc)
		{
			case 0:
				return;

			case QU_PC_JMP:
				j = jaddr (ip);
				while (cs [j] == QU_PC_JMP)
					j = jaddr (j);
				jaddr (ip) = j;
				im = cs [j];
				if (im == QU_PC_RET || im == QU_PC_END)
				{
					register int k;
					cs [ip] = im;
					for (k = 1; k <= sizeof (long); k++)
						cs [ip + k] = QU_PC_NOP;
				}
				break;
		}
		ip += PSIZE (pc);
	}
}



static void code_optimize_jump2 (qu_code c)
{
	register unsigned char *cs = c->bytes;
	register int            ip = 0;
	register int            len = c->len;
	register int            pc;
	register long           j;

	while (ip < len)
	{
		pc = cs [ip];
		switch (pc)
		{
			case 0:
				return;

			case QU_PC_JMPFALSE:
			case QU_PC_JMPTRUE:
			case QU_PC_JMPFALSEX:
			case QU_PC_JMPTRUEX:
			case QU_PC_CATCH:
			case QU_PC_ENTER:
			case QU_PC_PARAM:
				j = jaddr (ip);
				if (cs [j] == QU_PC_JMP)
					jaddr (ip) = jaddr (j);
				break;
		}
		ip += PSIZE (pc);
	}
}

#undef jaddr


/*
	Make the Code object.
*/
static qu_code code_make ()
{
	qu_code c;

	c            = qu_gc_new (QuCodeClass);
	c->bytes     = CoBytes;
	c->parent    = CoMaster;
	c->len       = CoNextOfs;
	c->stacksize = (CoStackSize | 0xF) + 1;

	if (qu_sub_is (CoMaster))
		c->stacksize += ((qu_sub) CoMaster)->nlocal;

	/* always put the end-of-code marker - required by optimizers */
	c->bytes [c->len] = 0;

	/* optimize away :-/ */
	code_optimize_jump1 (c);
	code_optimize_jump2 (c);

	/* so code_end does not free the allocated bytecode */
	CoBytes = NULL;

	return c;
}



/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

/*
	Binary operators that never fail and always return a Boolean.
*/
static int code_binop_return_bool (qu_binop_f f)
{
	return f == (qu_binop_f) qu_sys_eq
		|| f == (qu_binop_f) qu_sys_ge
		|| f == (qu_binop_f) qu_sys_gt
		|| f == (qu_binop_f) qu_sys_is
		|| f == (qu_binop_f) qu_sys_le
		|| f == (qu_binop_f) qu_sys_like
		|| f == (qu_binop_f) qu_sys_lt
		|| f == (qu_binop_f) qu_sys_ne;
}



/*
	Unary operators that never fail and always return a Boolean.
*/
static int code_unop_return_bool (qu_unop_f f)
{
	return f == (qu_unop_f) qu_sys_true;
}



/*
	Binary operator of two literals (String or Number) that always return
	either a String or Number.
*/
static int code_binop_literal_return_literal (qu_binop_f f)
{
	return f == (qu_binop_f) qu_sys_cmp
		|| f == (qu_binop_f) qu_sys_div
		|| f == (qu_binop_f) qu_sys_eq
		|| f == (qu_binop_f) qu_sys_ge
		|| f == (qu_binop_f) qu_sys_gt
		|| f == (qu_binop_f) qu_sys_iand
		|| f == (qu_binop_f) qu_sys_idiv
		|| f == (qu_binop_f) qu_sys_ior
		|| f == (qu_binop_f) qu_sys_ixor
		|| f == (qu_binop_f) qu_sys_in
		|| f == (qu_binop_f) qu_sys_is
		|| f == (qu_binop_f) qu_sys_le
		|| f == (qu_binop_f) qu_sys_lshift
		|| f == (qu_binop_f) qu_sys_like
		|| f == (qu_binop_f) qu_sys_lt
		|| f == (qu_binop_f) qu_sys_ne
		|| f == (qu_binop_f) qu_sys_minus
		|| f == (qu_binop_f) qu_sys_mul
		|| f == (qu_binop_f) qu_sys_plus
		|| f == (qu_binop_f) qu_sys_pow
		|| f == (qu_binop_f) qu_sys_ratio
		|| f == (qu_binop_f) qu_sys_rem
		|| f == (qu_binop_f) qu_sys_rin
		|| f == (qu_binop_f) qu_sys_rshift
		|| f == (qu_binop_f) qu_sys_concat
		|| f == (qu_binop_f) qu_sys_range;
}



/*
	Whether or not the unary operator of a literal (String, Number)
	returns a literal too.
*/
static int code_unop_literal_return_literal (qu_unop_f f)
{
	return f == (qu_unop_f) qu_sys_inot
		|| f == (qu_unop_f) qu_sys_neg
		|| f == (qu_unop_f) qu_sys_not
		|| f == (qu_unop_f) qu_sys_pos
		|| f == (qu_unop_f) qu_sys_true;
}



/*-----------------------------------------------------------------------------
	Error
-----------------------------------------------------------------------------*/

/*
	Make error message
*/
#define ERRMSG(n, msg) code_errmsg (n, msg)
static qu_string code_errmsg (qu_ast n, const char *msg)
{
	qu_string s = CoSource;

	if (s && (s->str <= n->cursor) && (n->cursor < s->str + s->len))
	{
		char *p, buf [81];
		int   l = strlen (n->cursor);

		if (l >= sizeof (buf) - 1)
			l = sizeof (buf) - 1;

		strncpy (buf, n->cursor, l);
		buf[l] = '\0';

		/* display until the end of the line only */
		p = strchr (buf, '\n');
		if (p != 0)
			*p = '\0';

		return qu_string_newf ("%s line %d: %s near %s", CoPath->str, n->line, msg, buf);
	}
	return qu_string_newf ("%s: %s", CoPath->str, msg);
}



/*
	Error with our own message.
*/
#define CTHROW(n, s) code_throw (n, s)
static void code_throw (qu_ast n, const char *s)
{
	qu_thread_throw (QuECompileClass, ERRMSG (n, s));
	longjmp (CoJmpbuf, 1);
}



/*
	Error with message as thrown by library functions.
*/
#define CRETHROW(n) code_rethrow (n)
static void code_rethrow (qu_ast n)
{
	qu_string s = qu_thread_err_string ();
	CTHROW (n, s->str);
}



/*----------------------------------------------------------------------------
	Reduce
----------------------------------------------------------------------------*/

#define QU_REDUCE_IN_CODE_C 1
#include "Reduce.c"



/*----------------------------------------------------------------------------
	Emit
----------------------------------------------------------------------------*/

#define QU_EMIT_IN_CODE_C 1
#include "Emit.c"



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class pcname (type: Integer) -> String
	Get the name of a bytecode.

	@exc
	.EArgValue
	.ENone     no such bytecode

	@xmp
		println ('type name        args')
		i = 0
		loop
			i += 1
			try
				printfln ('%4d %-12s %2d', i, Code.pcname (i), Code.pctype (i))
			catch ENone
				break
			;;
		;;

	@out
		type name        args
		   1 ARRAY        -1
		   2 ARGC          0
		   3 AS            0
		  ...
		  91 THROW        -1
		  92 TRUE          0
		  93 XCHG          0
*/
static qu_string qu__code_pcname (qu_class self, qu_integer type)
{
	qu_string s;
	int       i;

	QU_ARG_int (type, i);
	s = qu_code_pcname (i);
	QU_REQUIRE (QuENone, s);
	return s;
}



/*) sub class pctype (type: Integer) -> String
	Get the type of arguments of a bytecode.

	@exc
	.EArgValue
	.ENone    no such bytecode

	On success, an Int is returned:
	@ops
	.| 0x000 | no argument
	.| 0x011 | one C long argument
	.| 0x112 | two C long arguments
	.| 0x122 | one C long argument, one object argument
	.| 0x021 | one object argument
	.| 0x212 | one object argument, one C long argument
	.| 0x222 | two object arguments

	@xmp
		println (Code.pctype (12))

	@out
		-1
*/
static qu_int qu__code_pctype (qu_class self, qu_integer type)
{
	int    i;
	qu_int r;

	QU_ARG_int (type, i);
	r = qu_code_pctype (i);
	QU_REQUIRE (QuENone, r);
	return r;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__code__destroy (qu_code self)
{
	qu_mem_free (self->bytes);
}



static void qu__code__mark (qu_code self)
{
	qu_gc_mark (self->parent);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __each -> Iterator
	Return an -{Iterator} object to iterate through each bytecode.

	Each item is returned as an -{Array} of the following items:
	@ul
	. offset
	. type
	. argument type
	. argument-1 (if any)
	. argument-2 (if any)

	@exc
	.EImplement the code is JIT
*/
static qu_array qu__code__next (qu_code self, qu_iterator i)
{
	unsigned char *b = self->bytes;
	long           k, v;
	int            type, argtype;
	qu_any         x;
	qu_array       a;

	assert (i->obj == (qu_any) self);
	k = i->ex1;
	QU_ITERATOR_STOP (0 <= k && k < self->len);

	type = b[k];
	x = qu_code_pctype (type);
	QU_REQUIREZ (QuEValue, x, "unknown instruction in code");

	argtype = qu_int_tol (x);

	a = qu_array_new (6);
	qu_array_push (a, qu_int_new (k));
	qu_array_push (a, qu_int_new (type));
	qu_array_push (a, x);

	k++;
	switch (argtype)
	{
		case QU_PC_0:
			break;

		case QU_PC_L:
			v = *(long *) (b + k);
			qu_array_push (a, qu_int_new (v));
			k += sizeof (long);
			break;

		case QU_PC_LL:
			v = *(long *) (b + k);
			qu_array_push (a, qu_int_new (v));
			k += sizeof (long);
			v = *(long *) (b + k);
			qu_array_push (a, qu_int_new (v));
			k += sizeof (long);
			break;

		case QU_PC_LA:
			v = *(long *) (b + k);
			qu_array_push (a, qu_int_new (v));
			k += sizeof (long);
			x = *(qu_any *) (b + k);
			qu_array_push (a, x);
			k += sizeof (qu_any);
			break;

		case QU_PC_A:
			x = *(qu_any *) (b + k);
			qu_array_push (a, x);
			k += sizeof (qu_any);
			break;

		case QU_PC_AL:
			x = *(qu_any *) (b + k);
			qu_array_push (a, x);
			k += sizeof (qu_any);
			v = *(long *) (b + k);
			qu_array_push (a, qu_int_new (v));
			k += sizeof (long);
			break;

		case QU_PC_AA:
			x = *(qu_any *) (b + k);
			qu_array_push (a, x);
			k += sizeof (qu_any);
			x = *(qu_any *) (b + k);
			qu_array_push (a, x);
			k += sizeof (qu_any);
			break;

		default:
			QU_THROWZ (QuEValue, "invalid instruction argument type in code");
	}
	i->ex1 = k;
	return a;
}



static qu_iterator qu__code__each (qu_code self)
{
	register qu_iterator i;
	QU_REQUIREZ (QuEImplement, !qu_code_is_jit (self), "cant iterate through JIT code");
	QU_ITERATOR_NEW_EXTERN (i, self, qu__code__next);
	return i;
}



/*) sub __len -> Int
	Return the length of the bytecode. Note that the result may vary depending
	on the compiler.

	@noexc

	@xmp
		sub foo
			return 1
		;;

		f = @foo
		println (f.code.__len)

	@out
		17
*/
static qu_int qu__code__len (qu_code self)
{
	assert (self->len <= QU_INT_MAX);
	return qu_int_new (self->len);
}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	Note that the output of the following example may not be as shown here.
	It depends on the compiler.

	@xmp
		a = -123
		b = a * 45.6
		println (__main__.__code)

	@out
		 0: LITERAL      <Int -123>
		 5: SETGLOBAL    <Var __main__.a>
		10: DISCARD
		11: GETGLOBAL    <Var __main__.a>
		16: LITERAL      <Float 4.56>
		21: MUL
		22: SETGLOBAL    <Var __main__.b>
		27: DISCARD
		28: LITERAL      <Sub Sys.println>
		33: LITERAL      <Sub Sys.__main__>
		38: CALL         0
		43: GETATTRX     <String __code>
		48: AUTOATTR
		49: CALL         1
		54: DISCARD
		55: END
*/
#define p_(z)    if (!qu_sys_printz (z, strlen (z))) return 0
#define s_(x)    if (!code_print_string (x)) return 0
#define e_(x)    if (!qu_code_print_literal (x)) return 0
#define l_(v)    if (!code_print_long (v)) return 0
#define i_(i, x) if (!code_print_inst (i, x)) return 0

static void code_nice_strz (char *buf, long l)
{
	while (l--)
	{
		if (*buf < 32 || *buf > 126)
			*buf = ' ';
		buf++;
	}
}


static int code_print_string (qu_string x)
{
	char buf [40];
	long l = x->len;

	if (l <= 32)
	{
		memcpy (buf, x->str, l + 1);
		code_nice_strz (buf, l);
		p_(buf);
	}
	else
	{
		memcpy (buf, x->str, 32);
		buf [32] = '\0';
		code_nice_strz (buf, 32);
		p_(buf);
		p_("...");
	}
	return 1;
}



int qu_code_print_literal (qu_any x)
{
	char buf [128];

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_String:
			p_("<String ");
			s_(x);
			p_(">");
			break;

		case QU_CLASS_ID_Int:
			sprintf (buf, "<Int %ld>", qu_int_tol (x));
			p_(buf);
			break;

		case QU_CLASS_ID_Float:
			sprintf (buf, "<Float %g>", ((qu_float) x)->value);
			p_(buf);
			break;

		case QU_CLASS_ID_Null:
			p_("nil");
			break;

		case QU_CLASS_ID_Boolean:
			if (x == (qu_any) QuTrue)
			{
				p_("true");
			}
			else
			{
				p_("false");
			}
			break;

		case QU_CLASS_ID_Inf:
			if (x == (qu_any) QuInfPos)
			{
				p_("infp");
			}
			else
			{
				p_("infn");
			}
			break;

		case QU_CLASS_ID_Module:
		case QU_CLASS_ID_Class:
		case QU_CLASS_ID_Sub:
		case QU_CLASS_ID_Var:
			if (!qu_sys_print (x))
				return 0;
			break;

		default:
			p_("<");
			s_(qu_classof (x)->name);
			p_(" ");
			sprintf (buf, "%p", x);
			p_(buf);
			p_(">");
	}
	return 1;
}



static int code_print_long (long v)
{
	char buf [128];

	sprintf (buf, "%ld", v);
	p_(buf);
	return 1;
}



static int code_print_inst (long i, qu_string name)
{
	char buf [128];

	snprintf (buf, sizeof (buf), "%5ld: %-15s ", i, name->str);
	buf [sizeof (buf) - 1] = '\0';
	p_(buf);
	return 1;
}



static int code_print (qu_code self)
{
	unsigned char *b = self->bytes;
	int            l = self->len;
	int            i, p, t;
	qu_string      s;
	qu_any         x;
	long           y;

	for (i = 0; i < l;)
	{
		p = b[i];
		s = qu_code_pcname (p);
		QU_REQUIREZ (QuEValue, s, "unknown instruction in code");
		x = qu_code_pctype (p);
		assert (x);
		t = qu_int_tol (x);
		i_(i, s);
		i++;
		switch (t)
		{
			case QU_PC_0:
				break;

			case QU_PC_L:
				y = *(long *) (b + i);
				l_(y);
				i += sizeof (long);
				break;

			case QU_PC_LL:
				y = *(long *) (b + i);
				i += sizeof (long);
				l_(y);
				y = *(long *) (b + i);
				p_(" ");
				l_(y);
				i += sizeof (long);
				break;

			case QU_PC_LA:
				y = *(long *) (b + i);
				i += sizeof (long);
				l_(y);
				x = *(qu_any *) (b + i);
				p_(" ");
				e_(x);
				i += sizeof (qu_any);
				break;

			case QU_PC_A:
				x = *(qu_any *) (b + i);
				e_(x);
				i += sizeof (qu_any);
				break;

			case QU_PC_AL:
				x = *(qu_any *) (b + i);
				e_(x);
				i += sizeof (qu_any);
				y = *(long *) (b + i);
				p_(" ");
				l_(y);
				i += sizeof (long);
				break;

			case QU_PC_AA:
				x = *(qu_any *) (b + i);
				e_(x);
				i += sizeof (qu_any);
				x = *(qu_any *) (b + i);
				p_(" ");
				e_(x);
				i += sizeof (qu_any);
				break;

			default:
				QU_THROWZ (QuEValue, "invalid instruction argument type in code");
		}
		p_("\n");
	}
	return 1;
}

#undef p_
#undef s_
#undef e_
#undef l_
#undef i_


qu_null qu__code__print (qu_code self)
{
	if (qu_code_is_jit (self))
	{
		if (__QuCodeJitPrint)
			__QuCodeJitPrint (self);
		else
			qu_sys_printz ("JITCODE", strlen ("JITCODE"));
	}
	else
	{
		QU_ENSURE (code_print (self));
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub stacksize -> Int
	Return the required stack size.

	Note that the result may vary depending on the compiler.
	Usually the size is rounded to some aligned value.

	@noexc

	@xmp
		sub foo
			return 1
		;;

		f = @foo
		println (f.code.stacksize)

	@out
		16
*/
static qu_int qu__code_stacksize (qu_code self)
{
	assert (self->stacksize <= QU_INT_MAX);
	return qu_int_new (self->stacksize);
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

int qu_code_generate (qu_module mod)
{
	static volatile int here = 0;
	int result = 0;

	assert (here == 0);
	here = 1;

	assert (mod->tree);

	CoModule = mod;
	code_begin (mod);
	if (setjmp (CoJmpbuf) == 0)
	{
		reduce_module (mod->tree);
		emit_module ();
		qu_module_set_compiled (mod);
		result = 1;
	}
	code_end ();
	here = 0;
	return result;
}



qu_string qu_code_pcname (int type)
{
	if (type < 0 || type > QU_PC_LAST_INSTRUCTION)
		return 0;
	return __QuCodePcName [type];
}



qu_int qu_code_pctype (int type)
{
	if (type < 0 || type > QU_PC_LAST_INSTRUCTION)
		return 0;
	return qu_int_new (__QuCodePcType [type]);
}



qu_module qu_code_get_module (qu_code self)
{
	qu_any b = self->parent;

	if (qu_module_is (b))
		return (qu_module) b;
	assert (qu_sub_is (b));
	return qu_sub_get_module (b);
}


#if 0 /* save later for optimizing Sub */
int qu_code_does_nothing (qu_code self)
{
	return self->bytes[0] == QU_PC_END
		|| (self->bytes[6] == QU_PC_END
			&& self->bytes[5] == QU_PC_DISCARD
			&& self->bytes[0] == QU_PC_LITERAL);
	/* TODO - the Sub should be optimized to just contain QU_END */
}
#endif



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_code__setup ()
{
	qu_class_setup_sys (QuCodeClass, qu_code_s, qu__code__destroy, qu__code__mark);
}



void qu_code__init ()
{

#define p_(name, type) \
	__QuCodePcName [QU_PC_##name] = qu_string_new_staticz (#name); \
	__QuCodePcType [QU_PC_##name] = type;

#include "Codes.h"

#undef p_

	/* specific class methods */
	qu_class_method_classz (QuCodeClass, "pcname", qu__code_pcname, 1);
	qu_class_method_classz (QuCodeClass, "pctype", qu__code_pctype, 1);

	/* standard instance methods */
	qu_class_method_inst   (QuCodeClass, QU_NAME__each , qu__code__each , 0);
	qu_class_method_inst   (QuCodeClass, QU_NAME__len  , qu__code__len  , 0);
	qu_class_alias         (QuCodeClass, QU_NAME__len  , QU_NAME_len);
	qu_class_method_inst   (QuCodeClass, QU_NAME__print, qu__code__print, 0);

	/* specific instance methods */
	qu_class_method_instz (QuCodeClass, "stacksize", qu__code_stacksize, 0);

	/* finalize declaration */
	qu_class_enddecl (QuCodeClass);

	/* setup binop list */
	memset (QU_CODE_BINOPS, 0, sizeof (QU_CODE_BINOPS));

	QU_CODE_BINOPS [QU_PC_IN    ] = (qu_binop_f) qu_sys_in;
	QU_CODE_BINOPS [QU_PC_EQ    ] = (qu_binop_f) qu_sys_eq;
	QU_CODE_BINOPS [QU_PC_IS    ] = (qu_binop_f) qu_sys_is;
	QU_CODE_BINOPS [QU_PC_LT    ] = (qu_binop_f) qu_sys_lt;
	QU_CODE_BINOPS [QU_PC_GT    ] = (qu_binop_f) qu_sys_gt;
	QU_CODE_BINOPS [QU_PC_NE    ] = (qu_binop_f) qu_sys_ne;
	QU_CODE_BINOPS [QU_PC_LE    ] = (qu_binop_f) qu_sys_le;
	QU_CODE_BINOPS [QU_PC_GE    ] = (qu_binop_f) qu_sys_ge;
	QU_CODE_BINOPS [QU_PC_CMP   ] = (qu_binop_f) qu_sys_cmp;
	QU_CODE_BINOPS [QU_PC_LIKE  ] = (qu_binop_f) qu_sys_like;
	QU_CODE_BINOPS [QU_PC_RIN   ] = (qu_binop_f) qu_sys_rin;
	QU_CODE_BINOPS [QU_PC_PLUS  ] = (qu_binop_f) qu_sys_plus;
	QU_CODE_BINOPS [QU_PC_MINUS ] = (qu_binop_f) qu_sys_minus;
	QU_CODE_BINOPS [QU_PC_MUL   ] = (qu_binop_f) qu_sys_mul;
	QU_CODE_BINOPS [QU_PC_DIV   ] = (qu_binop_f) qu_sys_div;
	QU_CODE_BINOPS [QU_PC_IDIV  ] = (qu_binop_f) qu_sys_idiv;
	QU_CODE_BINOPS [QU_PC_CONCAT] = (qu_binop_f) qu_sys_concat;
	QU_CODE_BINOPS [QU_PC_AS    ] = (qu_binop_f) qu_sys_as;
	QU_CODE_BINOPS [QU_PC_PUSH  ] = (qu_binop_f) qu_sys_push;
	QU_CODE_BINOPS [QU_PC_IOR   ] = (qu_binop_f) qu_sys_ior;
	QU_CODE_BINOPS [QU_PC_IAND  ] = (qu_binop_f) qu_sys_iand;
	QU_CODE_BINOPS [QU_PC_IXOR  ] = (qu_binop_f) qu_sys_ixor;
	QU_CODE_BINOPS [QU_PC_LSHIFT] = (qu_binop_f) qu_sys_lshift;
	QU_CODE_BINOPS [QU_PC_RSHIFT] = (qu_binop_f) qu_sys_rshift;
	QU_CODE_BINOPS [QU_PC_RATIO ] = (qu_binop_f) qu_sys_ratio;
	QU_CODE_BINOPS [QU_PC_REM   ] = (qu_binop_f) qu_sys_rem;
	QU_CODE_BINOPS [QU_PC_POW   ] = (qu_binop_f) qu_sys_pow;
	QU_CODE_BINOPS [QU_PC_DROP1 ] = (qu_binop_f) qu_sys_rep;
	QU_CODE_BINOPS [QU_PC_OF    ] = (qu_binop_f) qu_sys_of;


	/*
		This is used to simulate yield from C code.

		sub __yield__ (f, data)
			loop
				yield f (data)
			;;
		;;


		Here is an example on how to use it

		C module Foo
		-------------
		static qu_sub YieldIterator;
		static qu_any _yield_iterator (qu_iterator i)
		{
			assert (qu_iterator_is (i));
			return qu__iterator_next (i);
		}

		static qu_frame Foo_gen (qu_any x)
		{
			qu_iterator i = qu_sys_each (x);
			QU_ENSURE (i);
			return qu_sub_call2 (QuSubYield, qu_interp_self (), YieldIterator, i);
		}

		static void Foo_init (qu_module m)
		{
			YieldIterator =
				qu_module_add_sub (m, "yield", _yield_iterator, 1,
					QU_SUB_FLAG_CFUNC | QU_SUB_FLAG_PRIVATE);

			qu_module_add_sub (m, "gen", Foo_gen, 1, QU_SUB_FLAG_CFUNC);
		}

		Qu program
		----------
		use Foo
		for x in Foo.gen ({1, 2, 3})
			println (x)
		;;
	*/
	{
		int j;

		QuSubYield = qu_sub_new ((qu_symbol) QuSysModule,
				qu_string_new_staticz ("yield"), NULL, 2,
				QU_SUB_FLAG_GENERATOR | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_STATIC);

		code_begin (QuSubYield);
		emit_begin ();

		j = P1 (GETLEXICAL, 0);
		    P1 (GETLEXICAL, 1);
		    P1 (CALL, 1);
		    P0 (YIELD);
		    P0 (DISCARD);
			P1 (JMP, j);

		emit_end ();
		QuSubYield->value = code_make ();
		assert (QuSubYield->root = QuSubYield);
		qu_sub_enddecl (QuSubYield);
		QuSubYield->nlocal= 2;
		code_end ();
	}
}
