/* Compile.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/*
	Remember
	--------
	- we keep all non-eval modules in memory using Weak references
	- we don't want to compile the same module if it's already in memory
	- modules may be inter-dependent
	- so it's difficult to do a multi-threaded compile
	- so we dont do that
	- compilation is done one thread at a time, with a lock

	This compiler does the parsing and at the same time creates Symbol
	objects (Module, Class, Sub, Var). Implicit variables are created
	later by Code.

	You might see this process as a combination of parsing to a literal
	tree then convert the literal tree into an object tree at the same
	time.

	Note on labels
	--------------
	Labels are glued to the start of a do, loop, while, etc.
	They are not independent such as in C, i.e. there is no goto.

	   _____ glued
	  /  \
	Foo: while foo
		...
*/

/* we only need one tokenizer for everything */
static qu_tokenizer CoTokenizer = 0;
#define CoToken  CoTokenizer->token /* the current token */

static jmp_buf   CoJmpbuf;      /* to bail out on error */
static qu_module CoModule;      /* the module object being compiled */
static qu_string CoPath;        /* module->name, for efficiency */
static qu_string CoSource;      /* module->source, for efficiency */
static int       CoDepth;       /* current expression depth */
static int       CoBlockLevel;  /* current block depth */
static int       CoMasterLevel; /* current master block depth */
static qu_any    CoBlock;       /* current block */
static qu_symbol CoMaster;      /* current master block */
static int       CoInList;      /* in a list? */
static int       CoInAtomic;    /* in atomic block? */
static int       CoInEnsure;    /* in ensure sub, cant use 'yield' */
static int       CoInLambda;    /* in lambda expression */
static int       CoInCurlySub;  /* in {} sub */
static int       CoInIf;        /* in if block */
static int       CoInConst;     /* in const expression */
static qu_string CoLabel;       /* to set label, for while, loop, etc. */
static int       CoUseIndent;   /* use indent style */
static int       CoDocumenting; /* include doumentation strings? */
static qu_dict   CoDocDict;     /* documentation entries */

static qu_any    CoBlockStack  [QU_COMPILE_MAXLEVEL + 1]; /* the blocks */
static qu_symbol CoMasterStack [QU_COMPILE_MAXLEVEL + 1]; /* master blocks */
static qu_string CoLabelStack  [QU_COMPILE_MAXLEVEL + 1]; /* block labels */

#define peek        qu_tokenizer_next (CoTokenizer)
#define peekn(n)    qu_tokenizer_peek (CoTokenizer, n)
#define eat         qu_tokenizer_take (CoTokenizer)
#define match(t)    qu_tokenizer_match (CoTokenizer, t)
#define req(t,z)    QU_BEGIN if (qu_tokenizer_need (CoTokenizer, t) == 0) { if (peek) eat; croak (z); } QU_END
#define sameline    qu_tokenizer_nobr (CoTokenizer)
#define endofsource qu_tokenizer_eos (CoTokenizer)

#define nobrk(z) \
QU_BEGIN \
	if (!sameline) \
		croak ("no linebreak " z); \
QU_END


#define identifier(z)  req (QU_T_NAME, "identifier expected " z)
#define indent(z)      req (QU_T_INDENT, "tab indentation expected to start " z )
#define dedent(z)      req (QU_T_DEDENT, "tab deindentation expected to close " z)

#define skipindent \
QU_BEGIN \
	if (CoUseIndent) \
		while (match (QU_T_INDENT) || match (QU_T_DEDENT)) {} \
QU_END

#define skipdedent \
QU_BEGIN \
	if (CoUseIndent) \
		while (match (QU_T_DEDENT)) {} \
QU_END

#define untrusted(z) \
QU_BEGIN \
	if (!qu_module_is_trusted (CoModule)) \
		croak (z); \
QU_END



/*----------------------------------------------------------------------------
	statements
----------------------------------------------------------------------------*/

static qu_ast _module   ();
static qu_ast _stmtexpr ();
static qu_ast _compound ();
static qu_ast _use      ();
static qu_ast _alias    ();
static qu_ast _class    ();
static qu_ast _type     ();
static qu_ast _sub      ();
static qu_ast _var      ();
static qu_ast _local    ();
static qu_ast _static   ();
static qu_ast _const    ();
static qu_ast _if       ();
static qu_ast _unless   ();
static qu_ast _switch   ();
static qu_ast _do       ();
static qu_ast _loop     ();
static qu_ast _while    ();
static qu_ast _for_head ();
static qu_ast _for      ();
static qu_ast _try      ();
static qu_ast _atomic   ();
static qu_ast _enum     ();
static qu_ast _print    (int html);
static qu_ast _struct   ();
static qu_ast _with     ();
static qu_ast _assert   ();
static qu_ast _doc      ();
static qu_ast _htopen   (int closed, int linebreak);
static qu_ast _htclose  (int linebreak);
static qu_ast _htentity ();

/*----------------------------------------------------------------------------
	expressions - from lowest to highest precedence.
----------------------------------------------------------------------------*/

/* expressions start with "or" thus this is just an alias */
#define _expr _bor

static qu_ast _bor      ();
static qu_ast _band     ();
static qu_ast _not      ();
static qu_ast _assign   ();
static qu_ast _quest    ();
static qu_ast _or       ();
static qu_ast _and      ();
static qu_ast _cmp      ();
static qu_ast _bitor    ();
static qu_ast _bitand   ();
static qu_ast _shift    ();
static qu_ast _arith    ();
static qu_ast _term     ();
static qu_ast _power    ();
static qu_ast _unary    ();
static qu_ast _trailer  ();
static qu_ast _factor   ();
static qu_ast _return   ();
static qu_ast _yield    ();
static qu_ast _multicat ();
static qu_ast _swap     ();
static qu_ast _throw    ();
static qu_ast _break    ();
static qu_ast _continue ();
static qu_ast _redo     ();
static qu_ast _symbol   ();
static qu_ast _array    ();
static qu_ast _dictarg  ();
static qu_ast _set      ();
static qu_ast _constant ();


/* this is to keep the C stack from crashing */
#define __enter_expression \
QU_BEGIN \
	if (CoDepth >= QU_COMPILE_MAXDEPTH) \
		croak ("too many nested expressions"); \
	CoDepth++; \
QU_END

#define __leave_expression \
QU_BEGIN \
	CoDepth--; \
QU_END

/* get name of a SYMBOL node */
#define symbol_name(n) \
	((n)->child \
		? ((n)->child->child ? (n)->child->child->thing : (n)->child->thing) \
		: (n)->thing)


/*----------------------------------------------------------------------------
	documentation
----------------------------------------------------------------------------*/

/* add documentation entry */
#define docentry(obj, decl, doc) \
	qu_dict_setitem (CoDocDict, obj, qu_array_new_ex (2, decl, doc))



/*----------------------------------------------------------------------------
	error
----------------------------------------------------------------------------*/

/* call this to bail out on error */
#define croak(strz) compile_error (strz)

/* call this to rethrow exception thrown by someone else then bail out */
#define rethrow compile_rethrow ()

/*
	check for maximum number of arguments
	nice side-effect: we never have to check for integer overflow
	note that this macro has no side-effect
*/
#define argmax(n, z) \
QU_BEGIN \
	if ((n) >= QU_COMPILE_MAXARGS) \
		croak (z); \
QU_END


/* make error message - always at the current token */
static qu_string errmsg (qu_string msg)
{
	if (endofsource)
	{
		return qu_string_newf ("%s at eof %s", CoPath->str, msg->str);
	}
	else if (CoSource->str <= CoToken->cursor && CoToken->cursor < CoSource->str + CoSource->len)
	{
		char *p, buf [81];
		int   l = strlen (CoToken->cursor);

		if (l >= sizeof (buf) - 1)
			l = sizeof (buf) - 1;

		strncpy (buf, CoToken->cursor, l);
		buf[l] = '\0';

		/* display until the end of the line only */
		p = strchr (buf, '\n');
		if (p)
			*p = '\0';

		return qu_string_newf ("%s line %d: %s near %s", CoPath->str, CoToken->line, msg->str, buf);
	}
	else
	{
		return msg;
	}
}



/* error with our own message */
static void compile_error (const char *z)
{
	qu_thread_throw (QuECompileClass, errmsg (qu_string_new_staticz (z)));
	longjmp (CoJmpbuf, 1);
}



/* error with exception thrown by someone else */
static void compile_rethrow ()
{
	qu_string s;

	assert (__QuThreadCurrent);

	s = __QuThreadCurrent->emsg;
	if (s == 0)
	{
		assert (__QuThreadCurrent->eclass);
		s = qu_exception_defmsg (__QuThreadCurrent->eclass);
	}
	qu_thread_throw (__QuThreadCurrent->eclass, errmsg (qu_sys_string (s)));
	longjmp (CoJmpbuf, 1);
}



/*----------------------------------------------------------------------------
	blocks

	master blocks     : Module, Class, and Sub
	non-master blocks : while, do, loop, for, etc.

	sub foo
		a = 0
		while a < 10
			b = 1    -> b belongs to 'foo', not the 'while'
			a += b
		;;
		print (b)    -> 1
	;;
----------------------------------------------------------------------------*/

#define CAN_BREAK     0x10000
#define CAN_CONTINUE  0x20000
#define CAN_REDO      0x40000

/* these are qu_int objects */
#define BLOCK_SWITCH  ((qu_any) (0x011 | CAN_BREAK))
#define BLOCK_DO      ((qu_any) (0x021 | CAN_BREAK | CAN_REDO))
#define BLOCK_LOOP    ((qu_any) (0x041 | CAN_BREAK | CAN_CONTINUE))
#define BLOCK_WHILE   ((qu_any) (0x081 | CAN_BREAK | CAN_CONTINUE))
#define BLOCK_FOR     ((qu_any) (0x101 | CAN_BREAK | CAN_CONTINUE))
#define BLOCK_ATOMIC  ((qu_any) (0x201))
#define BLOCK_TRY     ((qu_any) (0x401))

/* is it a master block (Module, Class, Sub) ? */
#define ismaster(b) (!qu_int_is (b))


/* must call when entering a block */
static void enter (qu_any b)
{
	if (CoBlockLevel >= QU_COMPILE_MAXLEVEL)
		croak ("too many nested blocks");

	CoBlockStack [++CoBlockLevel] = CoBlock = b;
	if (ismaster (b))
		CoMasterStack [++CoMasterLevel] = CoMaster = (qu_symbol) b;
	CoLabelStack [CoBlockLevel] = 0;
}



/* must call when leaving a block */
static void leave (qu_any b)
{
	assert (CoBlockLevel > 0);
	assert (CoBlock == b);

	CoBlock = CoBlockStack [--CoBlockLevel];
	if (ismaster (b))
		CoMaster = CoMasterStack [--CoMasterLevel];
}



/* do, while, etc. must call this function before calling enter */
static void setlabel ()
{
	CoLabelStack[CoBlockLevel] = CoLabel;
	CoLabel = 0;
}



/*----------------------------------------------------------------------------
	node - get familiar with this first
----------------------------------------------------------------------------*/

/* create a rule node (has no cursor position information) */
#define rnode(type) \
	qu_ast_newr (type)

/* create a production node (has cursor position information) */
#define pnode(type) \
	qu_ast_newp (type, CoToken->cursor, CoToken->line)

/*
	set node->thing from the current token
	CoToken->thing is always a literal (Int, Long, Float, String, etc.)
*/
#define grab(n) \
QU_BEGIN \
	(n)->thing = CoToken->thing; \
QU_END

/*
	add sibling node maintained with variable

	... - o     ->   ... - o - x
	      ^var                 ^var
*/
#define rpl(var, x) \
QU_BEGIN \
	var->next = x; \
	var = var->next; \
QU_END

/*
	add child node maintained with variable

	we want 'var' to always point to the last child node of 'top'

	top = pnode (xxx);
	var = NULL;

	x = _foo ();
	adc (top, var, x);

		top
		|
		x
		^var

	x = _bar ();
	adc (top, var, x);

		top
		|
		o - x
		    ^var
*/
#define adc(top, var, x) \
QU_BEGIN \
	if (var == NULL) \
		var = top->child = x; \
	else \
		rpl (var, x); \
QU_END

/*
	upgrade to parent node

	create parent node, make n the child node of this node, then
	set the variable n to that parent node

	x      ->    new
	^var         | ^var
	             x
*/
#define par(variable, type) \
QU_BEGIN \
	register qu_ast TmP = pnode (type); \
	TmP->child = variable; \
	TmP->cursor = variable->cursor; \
	TmP->line = variable->line; \
	variable = TmP; \
QU_END



/*----------------------------------------------------------------------------
	symbol
----------------------------------------------------------------------------*/

/*
	Make sure this is a unique symbol in the current block, i.e. the
	same symbol is not declared more than once.

	Doing this in a class block arises a problem. Should we inherit methods
	from Object directly when the class is created, then 'unique' will incorrectly
	fail when an overridable Object attribute is declared (e.g. __init). Our
	solution is to inherit from Object later and so unique works correctly.
*/
#define unique(name) \
QU_BEGIN \
	if (((qu_obj) CoMaster)->attrs && qu_attr_find (((qu_obj) CoMaster)->attrs, name)) \
		croak ("redefinition of symbol"); \
QU_END


/* set symbol in the current master block */
static void addsymbol (qu_string name, qu_symbol o)
{
	qu_class c = qu_classof (CoMaster);

	assert (qu_class_is (o) || qu_sub_is (o) || qu_var_is (o));

	if (c == QuClassClass)
	{
		/*
			qu_class_decl_attr checks for signatures, e.g. if the name
			happens to be a standard method name then it throws if the
			the signature doesn't match the predefined one.
		*/
		if (qu_class_decl_attr ((qu_class) CoMaster, name, o) == 0)
			rethrow;
	}
	else if (c == QuModuleClass)
	{
		/*
			Modules, on the other hand, dont care.
			That is why unique *should* be called before addsymbol.
		*/
		qu_module_setattr ((qu_module) CoMaster, name, o);
	}
	else
	{
		/*
			Make sure to use qu_sub_setattr so local variables are indexed
			and the local count is adjusted properly.
		*/
		assert (c == QuSubClass);
		qu_sub_setattr ((qu_sub) CoMaster, name, o);
	}
}



/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

/*
	To name anonymous subs (closures and ensures) so they have a unique
	entry in an attribute table (attr).
*/
static qu_string anon ()
{
	char buf [128];

#ifdef HAVE_LONG_LONG
	static unsigned long long count = 0;
	sprintf (buf, "<%llu>", count++);
#else
	static unsigned long count = 0;
	sprintf (buf, "<%lu>", count++);
#endif

	return qu_string_newz (buf);
}



/* get operator name */
static qu_string opname ()
{
	switch (peek)
	{
		case QU_T_BITOR:   return QU_NAME__ior;
		case QU_T_BITXOR:  return QU_NAME__ixor;
		case QU_T_BITAND:  return QU_NAME__iand;
		case QU_T_LSHIFT:  return QU_NAME__lshift;
		case QU_T_RSHIFT:  return QU_NAME__rshift;
		case QU_T_PLUS:    return QU_NAME__plus;
		case QU_T_MINUS:   return QU_NAME__minus;
		case QU_T_MUL:     return QU_NAME__mul;
		case QU_T_DIV:     return QU_NAME__div;
		case QU_T_IDIV:    return QU_NAME__idiv;
		case QU_T_REM:     return QU_NAME__rem;
		case QU_T_POW:     return QU_NAME__pow;

		case QU_T_LSQUARE:
			/*
				[]   __getitem
				[]=  __setitem
				[,]  __getslice
				[,]= __setslice
			*/
			switch (peekn (2))
			{
				case QU_T_RSQUARE:
					eat;
					eat;
					if (match (QU_T_ASSIGN))
						return QU_NAME__setitem;
					return QU_NAME__getitem;

				case QU_T_COMMA:
					if (peekn (3) == QU_T_RSQUARE)
					{
						eat;
						eat;
						eat;
						if (match (QU_T_ASSIGN))
							return QU_NAME__setslice;
						return QU_NAME__getslice;
					}
					break;
			}
			break;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	statement
----------------------------------------------------------------------------*/

#ifdef QU_HAS_UNIVERSE
static void _universe ()
{
	qu_string name;
	qu_set    set;

	untrusted ("universe declaration is prohibited");

	nobrk ("after universe");
	identifier ("after universe");
	name = CoToken->thing;

	set = qu_dict_getitem (QuUniverses, name);
	if (set == 0)
	{
		set = qu_set_new (1);
		qu_dict_setitem (QuUniverses, name, set);
	}
	qu_set_add (set, CoModule);

	if (CoModule->universes == NULL)
		CoModule->universes = qu_array_new (1);
	qu_array_push_quick (CoModule->universes, name);
}
#endif



/*
	todo
	|
	(tag | import)*
*/
static void _todo (qu_ast n)
{
	qu_ast todo = CoModule->tree->child;
	n->next = todo->child;
	todo->child = n;
}


/*
	module (thing: Module object)
	|
	todo - compound
*/
static qu_ast _module ()
{
	qu_ast top = CoModule->tree;
	qu_ast c   = NULL;

	top->thing = CoModule;

	/* necessary to put all IMPORT nodes up front and a place to store
		things to do */
	adc (top, c, rnode (QU_AST_todo));

	if (CoDocumenting)
	{
		/*HACK - documentation
			by setting this flag we don't bother to check for valid
			typed values at compile time */

		qu_module_set_doc (CoModule);
		CoDocDict = CoModule->doc = qu_dict_new (32);
	}

	if (match (QU_T_DOC))
	{
		if (CoDocumenting)
			docentry (CoModule, QuStringEMPTY, (qu_string) CoToken->thing);
	}
	else
	{
		if (CoDocumenting)
			docentry (CoModule, QuStringEMPTY, QuStringEMPTY);
	}

	while (match (QU_T_pragma))
	{
		qu_string p;

		nobrk ("after pragma");
		switch (peek)
		{
			case QU_T_NAME:
			case QU_T_STRING:
				eat;
				break;

			default:
				croak ("name or string expected after pragma");
		}

		p = CoToken->thing;

		if (strcmp (p->str, "indent") == 0)
		{
			CoUseIndent = 1;
			qu_tokenizer_set_use_indent (CoTokenizer);
		}
		else if (strcmp (p->str, "html") == 0)
		{
			static qu_string as = 0;
			qu_module m = (qu_module) qu_module_html ();

			if (as == 0)
				as = qu_string_new_staticz ("html");

			if (m == 0)
				croak ("cant load html");

			qu_module_add_use (CoModule, m, as);
			qu_module_import_all (CoModule, m);
			qu_tokenizer_set_use_html (CoTokenizer);
		}
		else
		{
			croak ("unknown pragma");
		}
	}

#ifdef QU_HAS_UNIVERSE
	while (match (QU_T_universe))
	{
		do
		{
			_universe ();

		} while (match (QU_T_COMMA));
	}
#endif

	adc (top, c, _compound ());

	if (qu_tokenizer_has_error (CoTokenizer))
		rethrow;

	switch (peek)
	{
		case QU_T_INDENT:
		case QU_T_DEDENT:
		case 0:
			break;

		default:
			croak ("end of source expected");
	}

	return top;
}


/*
	compound
	|
	(stmt | class | sub | ... etc)*
*/
static qu_ast _compound ()
{
	qu_ast top, c = 0, x = 0;

	top = rnode (QU_AST_compound);
	for (;;)
	{
		switch (peek)
		{
			case QU_T_END:
			case QU_T_end:
			case QU_T_elif:
			case QU_T_else:
			case QU_T_case:
			case QU_T_default:
			case QU_T_until:
			case QU_T_catch:
			case QU_T_finally:
				if (CoUseIndent)
					croak ("unexpected indentation");
			case 0:
				return top;

			case QU_T_RCURLY:
				if (CoInCurlySub)
					return top;
				goto expr;

			case QU_T_DEDENT:
				assert (CoUseIndent);
				if (CoBlockLevel == 0)
				{
					eat;
					continue;
				}
				return top;

			case QU_T_INDENT:
				assert (CoUseIndent);
				eat;
				croak ("bad indentation");

			case QU_T_SEMI:
				eat;
				continue;

			case QU_T_class:  eat; x = _class  (); break;
			case QU_T_sub:    eat; x = _sub    (); break;
			case QU_T_var:    eat; x = _var    (); break;
			case QU_T_if:     eat; x = _if     (); break;
			case QU_T_unless: eat; x = _unless (); break;
			case QU_T_switch: eat; x = _switch (); break;
			case QU_T_do:     eat; x = _do     (); break;
			case QU_T_loop:   eat; x = _loop   (); break;
			case QU_T_while:  eat; x = _while  (); break;
			case QU_T_for:    eat; x = _for    (); break;
			case QU_T_try:    eat; x = _try    (); break;
			case QU_T_atomic: eat; x = _atomic (); break;
			case QU_T_struct: eat; x = _struct (); break;
			case QU_T_with:   eat; x = _with   (); break;
			case QU_T_enum:   eat; x = _enum   (); break;

			case QU_T__doc:
				eat;
				_doc ();
				continue;

			case QU_T_static:
				eat;
				do
				{
					x = _static ();
					if (x)
						adc (top, c, x);

				} while (match (QU_T_COMMA));
				continue;

			case QU_T_const:
				eat;
				do
				{
					x = _const ();
					if (x)
						adc (top, c, x);

				} while (match (QU_T_COMMA));
				continue;

			case QU_T_local:
				eat;
				do
				{
					x = _local ();
					if (x)
						adc (top, c, x);

				} while (match (QU_T_COMMA));
				continue;

			case QU_T_use:
				eat;
				do
				{
					x = _use ();
					if (x)
						_todo (x);

				} while (match (QU_T_COMMA));
				continue;

			case QU_T_alias:
				eat;
				x = _alias ();
				if (x == 0)
					continue;
				break;

			case QU_T_assert:
				eat;
				x = _assert ();
				if (x == 0)
					continue;
				break;

			case QU_T_DOC:
				eat;
				croak ("misplaced documentation string");

			case QU_T_LPRINT:          eat; x = _print    (0);    break;
			case QU_T_LPRINTHTML:      eat; x = _print    (1);    break;
			case QU_T_HTMLOPEN:        eat; x = _htopen   (0, 0); break;
			case QU_T_HTMLOPENCLOSE:   eat; x = _htopen   (1, 0); break;
			case QU_T_HTMLOPENCLOSEBR: eat; x = _htopen   (1, 1); break;
			case QU_T_HTMLCLOSE:       eat; x = _htclose  (0);    break;
			case QU_T_HTMLCLOSEBR:     eat; x = _htclose  (1);    break;
			case QU_T_HTMLENTITY:      eat; x = _htentity ();     break;

			case QU_T_NAME:
				if (peekn (2) == QU_T_COLON)
				{
					eat;
					CoLabel = (qu_string) CoToken->thing;
					eat;
					switch (peek)
					{
						case QU_T_switch: eat; x = _switch (); break;
						case QU_T_do:     eat; x = _do     (); break;
						case QU_T_loop:   eat; x = _loop   (); break;
						case QU_T_while:  eat; x = _while  (); break;
						case QU_T_for:    eat; x = _for    (); break;

						default:
							croak ("switch, do, loop, while, or for expected after label");
					}
					break;
				}
				/*nobreak*/

			default:
			  expr:
				x = _stmtexpr ();
				break;
		}
		adc (top, c, x);
	}
	QU_FATAL ("never");
	return 0;
}



/*
	stmt
	|
	<expr>
*/
static qu_ast _stmtexpr ()
{
	qu_ast x = rnode (QU_AST_stmt);

	x->child = _expr ();

	if (x->child->type == QU_AST_literal)
	{
		/*CHECKME - a single literal statement prints

			stmt          print (thing: object)
			|        ->
			literal
		*/
		x->type = QU_AST_print;
		x->thing = x->child->thing;
		x->op = !sameline;
		x->child->thing = 0;
		x->child = 0;
	}

	if (sameline)
	{
		int t = peek;

		if (t == QU_T_if)
		{
			/*
				y if z

				stmt        if
				 x          |
				 |    =>    z - compound
				 y                |
				                stmt
				                  x
			*/
			qu_ast e, i, c;

			eat;
			e = _expr ();
			i = rnode (QU_AST_if);
			i->child = e;
			i->child->next = c = rnode (QU_AST_compound);
			c->child = x;
			x = i;
		}
		else if (t == QU_T_unless)
		{
			/*
				y unless z

				stmt        if
				 x          |
				 |    =>   not - compound
				 y          |       |
				            z      stmt
				                     x
			*/
			qu_ast e, i, n, c;

			eat;
			e = _expr ();
			i = rnode (QU_AST_if);
			i->child = n = rnode (QU_AST_unop);
			n->op = (long) qu_sys_not;
			n->child = e;
			n->next = c = rnode (QU_AST_compound);
			c->child = x;
			x = i;
		}
		else if (t == QU_T_while)
		{
			/*
				y while z

				stmt      while
				 x          |
				 |    =>    z - compound
				 y                |
				                stmt
				                  x
			*/
			qu_ast e, c, p = x;

			eat;
			x = pnode (QU_AST_while);
			e = x->child = _expr ();
			e->next = c = rnode (QU_AST_compound);
			c->child = p;
		}
		else if (t == QU_T_until)
		{
			/*
				y until z

				stmt        loop
				 x            |
				 |    =>    compound - z
				 y            |
				            stmt
				              x
			*/
			qu_ast c, p = x;

			eat;
			x = pnode (QU_AST_loop);
			x->child = c = rnode (QU_AST_compound);
			c->child = p;
			c->next = _expr ();
		}
		else if (t == QU_T_for)
		{
			/*
				y for z

				stmt         for
				 x            |
				 |    =>    compound - z
				 y            |
				            stmt
				              x
			*/
			qu_ast e, c, p = x;

			eat;
			x = _for_head ();
			e = x->child->next;
			e->next = c = rnode (QU_AST_compound);
			c->child = p;
		}
	}
	return x;
}



/*
	import (thing: Array of Module (no import))

	import (thing: Module, op: 0=import nothing 1=import all)

	import (thing: Module)
	|
	name (thing:name String to import)+

	!!! this function may return NULL
*/
static qu_string _getpath ()
{
	qu_string r;
	int       t;

	switch (peek)
	{
		case QU_T_DOT:
			eat;
			nobrk ("after .");
			identifier ("after .");
			r = qu_string_buffer_new (16);
			qu_string_buffer_putc (r, '/');
			goto next;

		case QU_T_NAME:
			eat;
			if (sameline)
			{
				t = peekn (1);
				if (t == QU_T_DOT)
				{
					r = qu_string_buffer_new (16);

				   next:
					qu_string_buffer_puts (r, CoToken->thing);

					while (sameline && match (QU_T_DOT))
					{
						nobrk ("after .");
						identifier ("after .");
						qu_string_buffer_putc (r, '/');
						qu_string_buffer_puts (r, CoToken->thing);
					}
					return qu_string_buffer_return (r);
				}
			}
			return CoToken->thing;

		case QU_T_STRING:
			eat;
			return CoToken->thing;
	}
	croak ("unexpected token");
	return 0;
}



static void _import (qu_ast top)
{
	if (match (QU_T_import))
	{
		if (match (QU_T_MUL))
		{
			top->op = 1;
		}
		else if (sameline)
		{
			qu_ast x, c = 0;

			do
			{
				skipindent;
				identifier ("for imported symbol");
				x = pnode (QU_AST_name);
				grab (x);
				adc (top, c, x);

			} while (match (QU_T_COMMA));

			skipdedent;
		}
	}
	else if (sameline && match (QU_T_LPAREN))
	{
		if (match (QU_T_MUL))
		{
			top->op = 1;
			req (QU_T_RPAREN, ") expected to close import list");
		}
		else
		{
			qu_ast x, c = 0;

			skipindent;
			do
			{
				identifier ("for imported symbol");
				x = pnode (QU_AST_name);
				grab (x);
				adc (top, c, x);
				match (QU_T_COMMA);
				skipindent;

			} while (!match (QU_T_RPAREN));

			skipdedent;
		}
	}
}



static qu_ast _use ()
{
	qu_string path, missing = 0, as = 0;
	qu_ast    top;
	qu_any    use;
	int       ignore;
	int       package = 0;
	int       standard = 0;

	untrusted ("use is prohibited");

	if (!qu_module_is (CoBlock))
		croak ("use is only allowed in module level");

	if (match (QU_T__package))
		package = 1;
	else if (match (QU_T__std))
		standard = 1;

	ignore = match (QU_T_try);

	/* documenting use __package xxx -> use __package try xxx */
	if (package && qu_module_is_doc (CoModule))
		ignore = 1;

	path = _getpath ();

#if 0
	if (sameline && match (QU_T_for))
	{
		/*HACK - documentation
			use __package xxx for yyy -> use yyy
			documentation uses the yyy module instead
		*/
		if (package && qu_module_is_doc (CoModule))
		{
			path = _getpath ();
			package = 0;
		}
		else
		{
			_getpath ();
		}
	}
#endif

	if (match (QU_T_or))
	{
		switch (peek)
		{
			case QU_T_NAME:
			case QU_T_STRING:
				missing = _getpath ();
				break;

			case QU_T_nil:
				eat;
				ignore = 1;
				break;

			default:
				croak ("name, string, or nil expected after or");
		}
	}

	if (match (QU_T_as))
	{
		identifier ("after as");
		as = CoToken->thing;
	}

	use = qu_module_use (CoModule, path, as, package ? QU_MODULE_USE_PACKAGE : (standard ? QU_MODULE_USE_STANDARD : QU_MODULE_USE_DEFAULT));

	if (use == 0)
	{
		if (missing == 0)
		{
		  ignored:

			if (as == 0)
				as = qu_file_name (path);

			if (ignore)
			{
				/* make it public as is the case when use succeeds
					but make it read-only */
   				qu_var v = qu_var_new (CoMaster, as, 0);
				addsymbol (as, (qu_symbol) v);
				return 0;
			}
			rethrow;
		}

		use = qu_module_use (CoModule, missing, as, package);
		if (use == 0)
		{
			if (ignore)
				goto ignored;
			rethrow;
		}
	}

	top = pnode (QU_AST_import);
	top->thing = use;

	_import (top);
	return top;
}



/*
	alias (thing: name)
	|
	symbol
*/
static qu_ast _alias ()
{
	qu_ast     top;
	qu_string  name;
	char      *di = (char *) CoToken->cursor;
	long       dl = 0;

	untrusted ("alias is prohibited");

	if (!(qu_module_is (CoMaster) || qu_class_is (CoMaster)))
		croak ("alias is only allowed at module or class level");

	nobrk ("after alias");
	identifier ("after alias");
	name = CoToken->thing;
	unique (name);

	match (QU_T_COLON);

	top = pnode (QU_AST_alias);
	top->thing = name;
	top->child = _symbol ();

	if (match (QU_T_DOC) && CoDocumenting)
	{
		/*HACK - documentation for alias */
		qu_var v = qu_var_new (CoMaster, name, QU_VAR_FLAG_ALIAS);

		dl = CoToken->cursor - di + CoToken->len;
		docentry (v, qu_string_new (di, dl), (qu_string) CoToken->thing);
		addsymbol (name, (qu_symbol) v);
	}
	return top;
}



/*
	class (thing: Class object)
	|
	[isa] - mixin* - compound

	isa (thing: name)     mixin (thing: name)
	|                     |
	[name]                [name]
*/
static void _class_decorators (qu_class c)
{
	int  p = 0;
	long f = 0;

	for (;;)
	{
		switch (peek)
		{
			case QU_T__public:
				eat;
				if (f & QU_CLASS_FLAG_PRIVATE)
					croak ("cant mix __public and __private");
				p = 1;
				continue;

			case QU_T__private:
				eat;
				if (p)
					croak ("cant mix __public and __private");
				f |= QU_CLASS_FLAG_PRIVATE | QU_CLASS_FLAG_LOCAL;
				continue;

			case QU_T__final:
				eat;
				if (f & QU_CLASS_FLAG_ABSTRACT)
					croak ("cant mix __final and __abstract");
				f |= QU_CLASS_FLAG_FINAL;
				continue;

			case QU_T__abstract:
				eat;
				if (f & QU_CLASS_FLAG_FINAL)
					croak ("cant mix __final and __abstract");
				if (f & QU_CLASS_FLAG_MIXIN)
					croak ("cant mix __mixin and __abstract");
				if (f & QU_CLASS_FLAG_AUTO)
					croak ("cant mix __auto and __abstract");
				f |= QU_CLASS_FLAG_ABSTRACT;
				continue;

			case QU_T__auto:
				eat;
				if (f & QU_CLASS_FLAG_MIXIN)
					croak ("cant mix __mixin and __auto");
				if (f & QU_CLASS_FLAG_ABSTRACT)
					croak ("cant mix __auto and __abstract");
				f |= QU_CLASS_FLAG_AUTO;
				continue;

			case QU_T__mixin:
				eat;
				if (f & QU_CLASS_FLAG_ABSTRACT)
					croak ("cant mix __mixin and __abstract");
				if (f & QU_CLASS_FLAG_AUTO)
					croak ("cant mix __mixin and __auto");
				f |= QU_CLASS_FLAG_MIXIN;
				continue;

			case QU_T__pure:
				eat;
				f |= QU_CLASS_FLAG_PURE;
				continue;

			case QU_T__uncatchable:
				eat;
				f |= QU_CLASS_FLAG_NOCATCH;
				continue;
		}
		break;
	}

	if (!p && c->name->str[0] == '_')
		f |= QU_CLASS_FLAG_PRIVATE | QU_CLASS_FLAG_LOCAL;

	c->flags |= f;
}



static qu_ast _class ()
{
	qu_ast     top, x, c = 0;
	qu_string  name;
	qu_class   kls;
	char      *di = (char *) CoToken->cursor;
	long       dl = 0;

	untrusted ("class declaration is prohibited");

	if (CoMasterLevel || CoInIf)
		croak ("class declaration is only allowed at module level");

	nobrk ("after class");
	identifier ("after class");
	name = CoToken->thing;
	unique (name);

	top = pnode (QU_AST_class);
	top->thing = kls = qu_class_declare ((qu_module) CoMaster, name, 0);
	addsymbol (name, (qu_symbol) kls);

	if (match (QU_T_is))
	{
		x = _symbol ();
		x->type = QU_AST_isa;
		adc (top, c, x);
	}

	if (match (QU_T_LPAREN))
	{
		do
		{
			x = _symbol ();
			x->type = QU_AST_mixin;
			adc (top, c, x);
			match (QU_T_COMMA);

		} while (!match (QU_T_RPAREN));
	}

	_class_decorators (kls);

	if (CoUseIndent)
		indent ("class");

	if (CoDocumenting)
	{
		/*HACK - documentation

			Doc.qm -> building documentation of a C module extracts the
			strings from the C file then does an eval. C classes can do
			anything. Program level classes can not, for example, override
			the __init method (see qu_class_set_mixin). Thus when the produced
			program is eval-ed, we get a compile error.

			Faking the QU_CLASS_FLAG_EXTERN flag solves the problem without adding
			additional burden to the compiler (has to specifically check for
			a QU_CLASS_FLAG_DOC).
		*/
		qu_flag_set (kls, QU_CLASS_FLAG_EXTERN | QU_CLASS_FLAG_DOC);
		dl = CoToken->cursor - di + CoToken->len;
	}
	if (match (QU_T_DOC))
	{
		if (CoDocumenting)
			docentry (kls, qu_string_new (di, dl), (qu_string) CoToken->thing);
	}
	else
	{
		if (CoDocumenting)
			docentry (kls, qu_string_new (di, dl), QuStringEMPTY);
	}

	enter (kls);
	x = _compound ();
	adc (top, c, x);

	if (CoUseIndent)
	{
		dedent ("class");
	}
	else
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_class, "end class expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close class");
		}
	}

	leave (kls);
	return top;
}




/*
	x.y.z
	x.y of z

	symbol
	|
	[name]
	|
	[name]
*/
static qu_ast _type ()
{
	if (match (QU_T_self))
	{
		/* just ignore, we haven't got the mechanism to check this yet */
		return 0;
	}
	else if (qu_module_is_doc (CoModule))
	{
		_symbol ();
		if (match (QU_T_of))
			identifier ("after of");
		return 0;
	}
	else
	{
		qu_ast top = _symbol ();

		top->op = 1;

		if (match (QU_T_of))
		{
			qu_ast c = top->child;

			if (c && c->child && c->child->child)
				croak ("of is impossible here");

			identifier ("after of");
			c = top;
			while (c->child)
				c = c->child;
			c->child = pnode (QU_AST_name);
			c->child->thing = CoToken->thing;
		}
		return top;
	}
}



/*
	tag (thing: name, op: Sub or Var)
	|
	<expr>

	using op to store a pointer is a bad idea but we have a reference
	already so what the heck ... creating an Array is too expensive
*/
static void _set_tags (qu_symbol x)
{
	assert (qu_var_is (x) || qu_sub_is (x));

	while (!match (QU_T_TAG))
	{
		qu_string name;
		qu_ast    value;

		identifier ("after ---");
		name = CoToken->thing;

		match (QU_T_COLON);
		value = _expr ();
		if (value->type != QU_AST_literal)
		{
			qu_ast t = rnode (QU_AST_tag);
			t->thing = name;
			t->op    = (long) x;
			t->child = value;
			_todo (t);
		}
		else
		{
			if (qu_var_is (x))
				qu_var_settag ((qu_var) x, name, value->thing);
			else
				qu_sub_settag ((qu_sub) x, name, value->thing);
		}
	}
}



/*
	sub (thing: Sub) or if dotted (thing: [Sub, Ast symbol])
	|
	[symbol] - param* - compound

	param (thing: Var)
	|
	[symbol] - [<expr>]
*/
static void _subargs (qu_ast top, qu_sub u, int closer, int allowdef)
{
	qu_ast c   = NULL;
	int    def = 0;
	int    n   = 0;

	while (!match (closer))
	{
		qu_ast type = 0;
		long   f;
		qu_var v;
		qu_ast x;

		argmax (++n, "too many parameters");

		f = QU_VAR_FLAG_ARG | QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_LEXICAL;
		if (match (QU_T_MUL))
			f |= QU_VAR_FLAG_VARARG;

		identifier ("as sub parameter");

		if (qu_sub_getattr (u, CoToken->thing))
			croak ("duplicate parameter");

		v = qu_var_new (CoMaster, (qu_string) CoToken->thing, f);

		if (match (QU_T_COLON))
		{
			type = _symbol ();

			/* ignore if qu_module_is_doc, to make things simpler */
			if (QU_APP_NOVALIDATION || qu_module_is_doc (CoModule))
				type = 0;
			else
				qu_flag_set (u, QU_SUB_FLAG_ARGCHECK);
		}

		/*
			Make sure to use qu_sub_setattr so the variable is indexed and the
			sub's local count is adjusted properly.
		*/
		qu_sub_setattr (u, v->name, (qu_symbol) v);

		if (allowdef && match (QU_T_ASSIGN))
		{
			qu_ast p = _expr ();

			def = 1;

			/* ignore if documenting, to make things simpler */
			if (!qu_module_is_doc (CoModule)
				/*CHECKME - no need to set a nil default value */
				&& !(p->type == QU_AST_literal && p->thing == QuNil))
			{
				x = pnode (QU_AST_param);
				x->thing = v;

				if (type)
				{
					x->child = type;
					type->next = p;
				}
				else
				{
					x->child = p;
				}
				adc (top, c, x);
			}
		}
		else
		{
			if (def && !(f & QU_VAR_FLAG_VARARG))
				croak ("non-default must not follow default parameters");

			u->argc++;

			if (type)
			{
				x = pnode (QU_AST_param);
				x->thing = v;
				x->child = type;
				adc (top, c, x);
			}
		}
		if (f & QU_VAR_FLAG_VARARG)
		{
			req (closer, "expect to close parameters");
			u->flags |= QU_SUB_FLAG_VARARG;
			u->argc--;
			break;
		}
		if (def)
			qu_flag_set (u, QU_SUB_FLAG_ARGMIN);

		match (QU_T_COMMA);
	}
}



static long _sub_scope (int dotted)
{
	if (dotted)
	{
		if (match (QU_T_local))
			croak ("local is not applicable with dotted sub");

		if (match (QU_T_static))
			return QU_SUB_FLAG_STATIC;

		if (match (QU_T_class))
			return QU_SUB_FLAG_CLASS;

		return QU_SUB_FLAG_INST;
	}
	else if (qu_class_is (CoBlock))
	{
		if (match (QU_T_class))
			return QU_SUB_FLAG_CLASS;

		if (match (QU_T_static))
			return QU_SUB_FLAG_STATIC;

		if (match (QU_T_local))
			return QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL;

		return QU_SUB_FLAG_INST;
	}
	else if (qu_module_is (CoBlock) || qu_sub_is (CoBlock))
	{
		if (match (QU_T_class))
			croak ("class is applicable in class block only");

		if (match (QU_T_static))
			croak ("static is applicable in class block only");

		if (qu_sub_is (CoBlock))
		{
			if (match (QU_T_local))
				croak ("local is applicable in class or module block only");

			return QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL;
		}

		if (match (QU_T_local))
			return QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL;

		return QU_SUB_FLAG_STATIC;
	}
	else
	{
		croak ("cant declare sub in loop, while, try, etc.");
		return -1; /*never*/
	}
}



static void _sub_type (qu_ast top)
{
	if (match (QU_T_ARROW))
	{
		qu_ast t = _type ();

		if (t)
		{
			qu_ast r = top->child;

			top->op = 1;
			top->child = t;
			top->child->next = r;
		}
	}
}



static void _sub_decorators (qu_sub u)
{
	int  p = 0;
	long f = 0;

	for (;;)
	{
		switch (peek)
		{
			case QU_T__public:
				eat;
				if (f & QU_SUB_FLAG_PRIVATE)
					croak ("cant mix __public and __private");
				p = 1;
				continue;

			case QU_T__private:
				eat;
				if (p)
					croak ("cant mix __public and __private");
				f |= QU_SUB_FLAG_PRIVATE;
				continue;

			case QU_T__final:
				eat;
				if (!qu_class_is (u->parent))
					croak ("__final is applicable in class only");
				if (f & QU_SUB_FLAG_ABSTRACT)
					croak ("cant mix __final and __abstract");
				/*if (f & QU_SUB_FLAG_STATIC)
					croak ("cant mix __final and static");*/
				f |= QU_SUB_FLAG_FINAL;
				continue;

			case QU_T__abstract:
				eat;
				if (!qu_class_is (u->parent))
					croak ("__abstract is applicable in class block only");
				if (!qu_class_is_abstract (u->parent))
					croak ("cant mix __abstract method in non-abstract class");
				if (qu_class_is_mixin (u->parent))
					croak ("__mixin class can not have __abstract methods");
				if (f & QU_SUB_FLAG_FINAL)
					croak ("cant mix __final and __abstract");
				f |= QU_SUB_FLAG_ABSTRACT;
				continue;

			case QU_T__unsafe:
				eat;
				if (f & QU_SUB_FLAG_ABSTRACT)
					croak ("__unsafe and __abstract is useless");
				f |= QU_SUB_FLAG_UNSAFE;
				continue;

			case QU_T__generator:
				eat;
				if (f & QU_SUB_FLAG_ONCE)
					croak ("cant mix __generator and __once");
				f |= QU_SUB_FLAG_GENERATOR;
				continue;

			case QU_T__once:
				eat;
				if (f & QU_SUB_FLAG_GENERATOR)
					croak ("cant mix __generator and __once");
				f |= QU_SUB_FLAG_ONCE;
				continue;

			case QU_T__accumulator:
				eat;
				croak ("__accumulator is only applicable to closures");

			case QU_T__args:
				eat;
				f |= QU_SUB_FLAG_ARGS;
				continue;

			case QU_T__pure:
				eat;
				f |= QU_SUB_FLAG_NOJIT;
				continue;
		}
		break;
	}

	if (!p && u->name->str[0] == '_')
		f |= QU_SUB_FLAG_PRIVATE;

	u->flags |= f;
}



static qu_ast _sub ()
{
	qu_symbol blk = CoBlock;
	qu_ast    top, c;
	qu_string name;
	qu_sub    sub;
	long      f;
	int       single = 0;
	char     *di = (char *) CoToken->cursor;
	long      dl = 0;

	/*EXPERIMENTAL - dotted sub */
	qu_ast    y = 0;
	int       dotted;

	nobrk ("after sub");

	name = opname ();
	if (name)
	{
		if (!qu_class_is (CoMaster))
			croak ("sub with operator name is applicable in class block only");
		unique (name);
		f = QU_SUB_FLAG_INST;
		dotted = 0;
	}
	else
	{
		if (peek == QU_T_NAME)
			dotted = peekn (2) == QU_T_DOT;
		else
			dotted = peekn (3) == QU_T_DOT;

		f = _sub_scope (dotted);

		if (dotted)
		{
			/*EXPERIMENTAL - dotted sub
				extend or replace methods:

				sub Foo.Bar.baz (x)
					...
				;;
			*/

			untrusted ("dotted sub is prohibited");

			if (!qu_module_is (CoBlock))
				croak ("dotted sub must be declared at module level");

			y = _symbol ();
			name = symbol_name (y);

			/* remove the last dotted name */
			assert (y->child);
			y->child->child = NULL;
		}
		else
		{
			identifier ("after sub");
			name = (qu_string) CoToken->thing;
			unique (name);
		}
	}

	top = pnode (QU_AST_sub);
	top->thing = sub = qu_sub_new ((qu_symbol) CoBlock, name, 0, 0, f);

	enter (sub);

	/* this must be done within the sub, that's why enter first. */
	if (sameline && match (QU_T_LPAREN))
		_subargs (top, sub, QU_T_RPAREN, 1);

	/*DANG
		module level __destroy method - must check exclusively

		# module
		sub __destroy
			...
		;;
	*/
	if (sub->narg && qu_module_is (blk) && qu_string_EQ (name, QU_NAME__destroy))
		croak ("__destroy must take no argument");

	_sub_type (top);
	_sub_decorators (sub);

	/*
		can't put __tags within Sub since they are not executed within
		the call, i.e. closures can not capture lexicals

		         +------------------+
		         |                  | cant do this!
		sub foo (x)                 |
			__tags (bar: lambda (y) x + y)
			...
		;;

		thus leave (u) first then _set_tags
	*/
	leave (sub);
	if (match (QU_T_TAG))
		_set_tags ((qu_symbol) sub);

	/*CHECKME - dotted don't get inserted into CoMaster->attrs
		we rely marking of the object from the Ast */
	if (!dotted)
		addsymbol (name, (qu_symbol) sub);

	enter (sub);

	if (CoUseIndent)
		indent ("sub");

	dl = CoToken->cursor - di + CoToken->len;
	match (QU_T_COLON);
	if (match (QU_T_DOC))
	{
		if (CoDocumenting)
			docentry (sub, qu_string_new (di, dl), (qu_string) CoToken->thing);
	}
	else
	{
		if (CoDocumenting)
			docentry (sub, qu_string_new (di, dl), QuStringEMPTY);
	}

	c = top->child;
	while (c && c->next)
		c = c->next;

	if (match (QU_T_ASSIGN))
	{
		qu_ast x = _expr ();
		qu_ast m = rnode (QU_AST_compound);
		m->child = x;
		adc (top, c, m);
		single = 1;
	}
	else
	{
		adc (top, c, _compound ());
	}

	if (CoUseIndent)
	{
		dedent ("sub");
	}
	else if (!single)
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_sub, "sub expected after end");
		}
		else
		{
			req (QU_T_END, ";; expected to close sub block");
		}
	}

	leave (sub);
	if (dotted)
		top->thing = qu_array_new_ex (2, sub, y);
	return top;
}



/*
	anonsub (thing: Sub)
	|
	[symbol] - param* - compound
*/
static void _anonsub_decorators (qu_sub u)
{
	long f = 0;

	for (;;)
	{
		switch (peek)
		{
			case QU_T__accumulator:
				eat;
				f |= QU_SUB_FLAG_ACCUM;
				continue;
		}
		break;
	}
	u->flags |= f;
}



static qu_ast _anonsub ()
{
	qu_ast    top, c;
	qu_sub    sub;
	qu_string name = 0;
	long      f = QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL;

	if (match (QU_T_NAME))
		name = (qu_string) CoToken->thing;
	else
		name = anon ();

	unique (name);
	sub = qu_sub_new (CoMaster, name, 0, 0, f);
	addsymbol (name, (qu_symbol) sub);

	top = pnode (QU_AST_anonsub);
	top->thing = sub;

	enter (sub);
	if (sameline && match (QU_T_LPAREN))
		_subargs (top, sub, QU_T_RPAREN, 1);

	_sub_type (top);
	_anonsub_decorators (sub);

	if (CoUseIndent)
		indent ("sub");

	c = top->child;
	while (c && c->next)
		c = c->next;

	match (QU_T_COLON);
	adc (top, c, _compound ());

	if (CoUseIndent)
	{
		dedent ("sub");
	}
	else
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_sub, "sub expected after end");
		}
		else
		{
			req (QU_T_END, ";; expected to close sub block");
		}
	}

	leave (sub);
	return top;
}



/*
	lambda (thing: Sub)
	|
	[symbol] - param* - return
	                       |
	                     <expr>
*/
static qu_ast _lambda ()
{
	qu_ast    top, c;
	qu_sub    sub;
	qu_string name;
	long      f = QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL;

	if (match (QU_T_NAME))
		name = (qu_string) CoToken->thing;
	else
		name = anon ();

	unique (name);
	sub = qu_sub_new (CoMaster, name, 0, 0, f);
	addsymbol (name, (qu_symbol) sub);

	top = pnode (QU_AST_anonsub);
	top->thing = sub;

	enter (sub);
	++CoInLambda;

	if (sameline && match (QU_T_LPAREN))
		_subargs (top, sub, QU_T_RPAREN, 1);

	_sub_type (top);
	_anonsub_decorators (sub);

	c = top->child;
	while (c && c->next)
		c = c->next;

	/*CHECKME - no need for QU_AST_stmt */
	adc (top, c, rnode (QU_AST_return));
	match (QU_T_COLON);
	c->child = _expr ();

	--CoInLambda;
	leave (sub);
	return top;
}



/*
	{|subargs| compound}
	{: compound}

	anonsub (thing: Sub)
	|
	[symbol] - param* - compound
*/
static qu_ast _curlsub (int witharg)
{
	/* anonymous subs are implicitly private */

	qu_ast    top, c;
	qu_sub    sub;
	qu_string name = 0;
	long      f = QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL;

	name = anon ();
	unique (name);
	sub = qu_sub_new (CoMaster, name, 0, 0, f);
	addsymbol (name, (qu_symbol) sub);

	top = pnode (QU_AST_anonsub);
	top->thing = sub;

	enter (sub);
	++CoInCurlySub;

	if (witharg)
		_subargs (top, sub, QU_T_BITOR, 0);

	_anonsub_decorators (sub);
	_sub_type (top);

	c = top->child;
	while (c && c->next)
		c = c->next;

	match (QU_T_COLON);
	adc (top, c, _compound ());

	c = c->child;
	if (c == 0)
		croak ("body can not be empty");

	if (c->type == QU_AST_stmt && c->next == 0)
	{
		/* auto return if a single expression */
		if (c->child->type != QU_AST_return)
			c->type = QU_AST_return;
	}

	req (QU_T_RCURLY, "} expected to close curly sub");

	--CoInCurlySub;
	leave (sub);
	return top;
}



/*
	var (thing: Var) or if dotted (thing:[Var, Ast symbol])
	|
	[symbol] - [<expr>] - [sub]
	                     (ensure)

	EXPERIMENTAL - currying type

		v: @foo (1, 2)

	var
	|
	ref (op: 1)
*/
static qu_ast _var_type ()
{
	qu_ast type;

	/*EXPERIMENTAL - currying */
	if (peek == QU_T_AT && peekn (3) == QU_T_LPAREN)
	{
		type = _factor ();
		if (type->type != QU_AST_ref || type->child == NULL || type->child->type != QU_AST_list)
			croak ("invalid type specification");
		type->op = 1;
	}
	else
	{
		type = _type ();
	}

	/* ignore if qu_module_is_doc, to make things simpler */
	if (QU_APP_NOVALIDATION || qu_module_is_doc (CoModule))
		type = 0;
	return type;
}


static qu_ast _var ()
{
	qu_ast    top, c, x, type = 0;
	qu_string name;
	qu_var    var;
	long      f      = 0;
	int       pub    = 0;
	int       ensure = 0;
	int       rdonly = 0;
	char     *di     = (char *) CoToken->cursor;
	long      dl     = 0;
	int       doc    = 0;

	/*EXPERIMENTAL - dotted var */
	qu_ast    y = 0;
	int       dotted = 0;

	nobrk ("after var");

	if (peek == QU_T_NAME)
		dotted = peekn (2) == QU_T_DOT;
	else
		dotted = peekn (3) == QU_T_DOT;

	if (qu_class_is (CoBlock) || dotted)
	{
		if (match (QU_T_class))
			f = QU_VAR_FLAG_CLASS;
		else
			f = QU_VAR_FLAG_INST;
	}
	else if (qu_module_is (CoBlock))
	{
		if (match (QU_T_class))
			croak ("var class in module level must be dotted");
	}
	else
	{
		croak ("var is applicable in module or class block only");
	}

	/*
		!!! We know we dont inherit properties from Object.
		!!! If we do, we need to modify this to make sure that the
		!!! Object property is not overridden.
	*/

	if (dotted)
	{
		y = _symbol ();
		name = symbol_name (y);

		/* remove the last dotted name */
		assert (y->child);
		y->child->child = NULL;
	}
	else
	{
		identifier ("after var");
		name = (qu_string) CoToken->thing;
		unique (name);
	}

	if (match (QU_T_COLON))
		type = _var_type ();

	for (;;)
	{
		switch (peek)
		{
			case QU_T__public:
				eat;
				if (f & QU_VAR_FLAG_PRIVATE)
					croak ("cant mix __public and __private");
				pub = 1;
				continue;

			case QU_T__private:
				eat;
				if (pub)
					croak ("cant mix __public and __private");
				f |= QU_VAR_FLAG_PRIVATE;
				continue;

			case QU_T__final:
				eat;
				if (qu_module_is (CoBlock) && !dotted)
					croak ("__final is applicable in class only");
				f |= QU_VAR_FLAG_FINAL;
				continue;

			case QU_T__readonly:
				eat;
				if (f & QU_VAR_FLAG_RDWR)
					croak ("__readonly cant be combined with __readwrite");
				rdonly = 1;
				continue;

			case QU_T__readwrite:
				eat;
				if (rdonly)
					croak ("__readwrite cant be combined with __readonly");
				f |= QU_VAR_FLAG_RDWR;
				continue;
		}
		break;
	}

	if (!pub && name->str[0] == '_')
		f |= QU_VAR_FLAG_PRIVATE;

	var = qu_var_new (CoMaster, name, f);

	/*CHECKME - dotted don't get inserted into CoMaster->attrs
		we rely marking of the object from the Ast */
	if (!dotted)
		addsymbol (name, (qu_symbol) var);

	top = pnode (QU_AST_var);
	top->thing = var;
	c = NULL;

	if (type)
		adc (top, c, type);

	switch (peek)
	{
		case QU_T_ASCOLON:
			if (rdonly)
				croak ("__readonly cant be combined with __readwrite");
			var->flags |= QU_VAR_FLAG_RDWR;
			/*nobreak*/

		case QU_T_ASSIGN:
			eat;
			x = _expr ();
			/*
				TODO - setting an initial value of nil is always acceptable
				even if the variable is typed.

				this has a bad effect:

					var Foo: qu_boolean = nil

				but it saves us from creating a node

				consider also:

					var Foo: Bar.Baz := nil

				the last thing the programmer wants is to create a new
				object for nothing
			*/
			if (x->type == QU_AST_literal && (type == 0 || x->thing == QuNil))
				var->value = x->thing;
			else
				adc (top, c, x);
	}

	if (match (QU_T_TAG))
		_set_tags ((qu_symbol) var);

	dl = CoToken->cursor + CoToken->len - di;

	if (CoUseIndent)
	{
		if (peek == QU_T_INDENT && peekn (2) == QU_T_DOC)
		{
			doc = 1;
			eat;
			eat;
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
		}
		else
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}
	else
	{
		if (match (QU_T_DOC))
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
		}
		else
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}

	if (CoUseIndent)
	{
		if (doc)
		{
			if (match (QU_T_ensure))
			{
				ensure = 1;
				match (QU_T_COLON);
			}
		}
		else
		{
			if (peek == QU_T_INDENT && peekn (2) == QU_T_ensure)
			{
				ensure = 1;
				eat;
				eat;
				match (QU_T_COLON);
			}
		}

		if (ensure)
			indent ("ensure");
	}
	else
	{
		ensure = match (QU_T_ensure);
		match (QU_T_COLON);
	}

	if (ensure)
	{
		/*
			ensure Sub is named xxx.ensure where xxx is the name of the Var
			it takes one argument which is named as the name of the Var
		*/
		qu_string ename = qu_string_cat (name, qu_string_byte ('.'), qu_tokenizer_name (QU_T_ensure), 0);
		qu_sub    ens;
		qu_var    arg;

		if (type)
			croak ("cant specify both type and ensure");

		ens = qu_sub_new (CoMaster, ename, 0, 1, QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_ENSURE);

		if (f & QU_VAR_FLAG_STATIC)
			ens->flags |= QU_SUB_FLAG_STATIC;

		else if (f & QU_VAR_FLAG_CLASS)
			ens->flags |= QU_SUB_FLAG_CLASS;

		else if (f & QU_VAR_FLAG_INST)
			ens->flags |= QU_SUB_FLAG_INST;

		/* ignore if qu_module_is_doc, to make things simpler */
		if (!(QU_APP_NOVALIDATION || qu_module_is_doc (CoModule)))
			var->type = ens;

		arg = qu_var_new ((qu_symbol) ens, name, QU_VAR_FLAG_ARG | QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_LEXICAL);
		qu_sub_setattr (ens, name, (qu_symbol) arg);

		x = pnode (QU_AST_sub);
		x->thing = ens;

		enter (ens);
		CoInEnsure = 1;

		x->child = _compound ();
		adc (top, c, x);

		if (CoUseIndent)
			dedent ("ensure");
		else
		{
			if (match (QU_T_end))
			{
				nobrk ("after end");
				if (!(match (QU_T_var) || match (QU_T_ensure)))
					croak ("end var or end ensure expected");
			}
			else
			{
				req (QU_T_END, ";; expected to close ensure");
			}
		}

		CoInEnsure = 0;
		leave (ens);
	}

	if (CoUseIndent && (doc || ensure))
		dedent ("ensure");

	if (dotted)
		top->thing = qu_array_new_ex (2, var, y);

	return top;
}



/*
	!!! This function may return NULL

	const (thing: Var)
	|
	<expr>

	we don't allow a type here because the value must be a literal and
	the bytecode will produce a QU_PC_LITERAL
*/
static qu_ast _const ()
{
	qu_ast    top, x;
	qu_string name;
	qu_var    var;
	int       f   = 0;
	int       pub = 0;
	char     *di  = di = (char *) CoToken->cursor;
	long      dl  = 0;

	if (!qu_module_is (CoBlock) && !qu_class_is(CoBlock))
		croak ("constants may only be declared in module or class block");

	identifier ("after const");
	name = (qu_string) CoToken->thing;
	unique (name);

	if (qu_class_is (CoBlock))
		f = QU_VAR_FLAG_CLASS;

	for (;;)
	{
		switch (peek)
		{
			case QU_T__public:
				eat;
				if (f & QU_VAR_FLAG_PRIVATE)
					croak ("cant mix __public and __private");
				pub = 1;
				continue;

			case QU_T__private:
				eat;
				if (pub)
					croak ("cant mix __public and __private");
				f |= QU_VAR_FLAG_PRIVATE;
				continue;

			case QU_T__final:
				eat;
				if (qu_module_is (CoBlock))
					croak ("__final is applicable in class only");
				f |= QU_VAR_FLAG_FINAL;
				continue;
		}
		break;
	}

	nobrk ("before literal constant");
	match (QU_T_ASSIGN);

	if (!pub && name->str[0] == '_')
		f |= QU_VAR_FLAG_PRIVATE;

	var = qu_var_new (CoMaster, name, f);

	x = _expr ();
	if (x->type == QU_AST_literal)
	{
		var->value = x->thing;
		qu_var_freeze (var);
		top = 0;
	}
	else
	{
		top = rnode (QU_AST_const);
		top->thing = var;
		top->child = x;
	}
	addsymbol (name, (qu_symbol) var);

	if (match (QU_T_TAG))
		_set_tags ((qu_symbol) var);

	dl = CoToken->cursor - di + CoToken->len;

	if (CoUseIndent)
	{
		if (peek == QU_T_INDENT && peekn (2) == QU_T_DOC)
		{
			eat;
			eat;
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
			dedent ("documentation string");
		}
		else
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}
	else
	{
		if (match (QU_T_DOC))
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
		}
		else
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}
	return top;
}



/*
	!!! This function may return NULL

	local (thing: Var, op: 1=has symbol 0:not)
	|
	[symbol] - <expr>
*/
static qu_ast _local ()
{
	qu_ast    top = 0, c = 0;
	qu_string name;
	qu_var    var;
	int       f = 0;

	if (!ismaster (CoBlock))
		croak ("cant declare local in while, loop, etc.");

	if (qu_sub_is (CoMaster))
		f = QU_VAR_FLAG_LEXICAL;

	identifier ("after local");
	name = (qu_string) CoToken->thing;
	unique (name);

	if (match (QU_T_COLON))
		c = _var_type ();

	var = qu_var_new (CoMaster, name, QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_PRIVATE | f);
	addsymbol (name, (qu_symbol) var);

	if (c)
	{
		top = pnode (QU_AST_local);
		top->thing = var;
		top->child = c;
		top->op = 1;
	}

	switch (peek)
	{
		case QU_T_ASCOLON:
			/* just ignore */
		case QU_T_ASSIGN:
			eat;
			if (c)
			{
				c->next = _expr ();
			}
			else
			{
				top = pnode (QU_AST_local);
				top->thing = var;
				top->child = _expr ();
			}
	}
	return top;
}



/*
	!!! This function may return NULL

	static (thing: Var)
	|
	[<expr>]

	we don't allow a type here because the value must be a literal and
	the bytecode will produce a QU_PC_LITERAL
*/
static qu_ast _static ()
{
	qu_ast    top, x;
	qu_string name;
	qu_var    var;

	if (!qu_sub_is (CoMaster))
		croak ("static variable is only applicable in sub");

	identifier ("after static");
	name = (qu_string) CoToken->thing;
	unique (name);

	var = qu_var_new (CoMaster, name, QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_STATIC);
	addsymbol (name, (qu_symbol) var);

	if (match (QU_T_ASSIGN))
	{
		x  = _expr ();
		if (x->type == QU_AST_literal)
		{
			var->value = x->thing;
			return 0;
		}
		top = pnode (QU_AST_static);
		top->thing = var;
		top->child = x;
		return top;
	}
	return 0;
}



/*
	enum (thing: Var, op: number of items)
	|
	enumitem+

	enumitem (thing: Var)
	|
	[<expr>]
*/
static qu_ast _enum ()
{
	qu_string ename;
	qu_string name;
	long      count;
	qu_var    var, v;
	long      f;
	char     *di = (char *) CoToken->cursor;
	long      dl = 0;
	int       gotdoc = 0;
	qu_ast    top, c = 0, e, x;

	if (!qu_module_is (CoBlock) && !qu_class_is(CoBlock))
		croak ("enum may only be declared in module or class block");

	nobrk ("after enum");
	identifier ("after enum");
	ename = (qu_string) CoToken->thing;
	unique (ename);

	if (match (QU_T__public))
		f = 0;
	else if (match (QU_T__private))
		f = QU_VAR_FLAG_PRIVATE;
	else
		f = (ename->str[0] == '_') ? QU_VAR_FLAG_PRIVATE : 0;

	if (qu_class_is (CoBlock))
		f |= QU_VAR_FLAG_CLASS;

	var = qu_var_new (CoMaster, ename, f);
	addsymbol (ename, (qu_symbol) var);

	top = pnode (QU_AST_enum);
	top->thing = var;

	match (QU_T_COLON);

	if (CoUseIndent)
		indent ("enum");

	for (count = 0; ;count++)
	{
		if (CoUseIndent)
		{
			if (match (QU_T_DEDENT))
				break;

			if (peek == QU_T_DOC)
			{
				dl = CoToken->cursor - di + CoToken->len;
				eat;
				if (CoDocumenting)
				{
					docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
					gotdoc = 1;
				}
				dedent ("documentation string");
				break;
			}
		}
		else
		{
			if (match (QU_T_end))
			{
				nobrk ("after end");
				req (QU_T_enum, "end enum expected");
				break;
			}
			else if (match (QU_T_END))
			{
				break;
			}
		}

		identifier ("for enum item");
		name = (qu_string) CoToken->thing;
		unique (name);

		x = 0;
		if (match (QU_T_COLON) || match (QU_T_ASSIGN))
			x = _expr ();

		v = qu_var_new (CoMaster, name, f);
		addsymbol (name, (qu_symbol) v);

		e = pnode (QU_AST_enumitem);
		e->thing = v;
		e->child = x;
		adc (top, c, e);

		match (QU_T_COMMA);
	}

	if (count == 0)
		croak ("enum cant be empty");
	top->op = count;

	dl = CoToken->cursor - di + CoToken->len;
	if (CoUseIndent)
	{
		if (CoDocumenting && !gotdoc)
			docentry (var, qu_string_new (di, dl), QuStringEMPTY);
	}
	else
	{
		if (match (QU_T_DOC))
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
		}
		else
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}
	return top;
}



/*
   (. x y z) -> print (x, y, z)
   (\ x y z) -> same but html escape, i.e. '<' -> &lt; etc.

	printex (op: 0=print 1=html)
	|
	<expr>+
*/
static qu_ast _print (int html)
{
	qu_ast top, c = 0;

	top = pnode (QU_AST_printex);
	top->op = html;
	do
	{
		skipindent;
		adc (top, c, _expr ());
		match (QU_T_COMMA);
	}
	while (!match (QU_T_RPAREN));
	return top;
}



/*
	struct (thing: Var, op: number of fields)
	|
	(literal - symbol)+

	literal (thing: name)
*/
static qu_ast _struct ()
{
	qu_ast    top, type, c = 0;
	qu_string name;
	qu_var    var;
	long      f = 0;
	int       n = 0;
	char     *di = di = (char *) CoToken->cursor;
	long      dl = 0;
	int       gotdoc = 0;

	if (!qu_module_is (CoBlock) && !qu_class_is (CoBlock))
		croak ("structs may only be declared in module or class level block");

	identifier ("after struct");
	name = (qu_string) CoToken->thing;
	unique (name);

	if (name->str[0] == '_')
		f = QU_VAR_FLAG_PRIVATE;

	if (match (QU_T__public))
		f = 0;
	else if (match (QU_T__private))
		f = QU_VAR_FLAG_PRIVATE;

	if (qu_class_is (CoBlock))
		f |= QU_VAR_FLAG_CLASS;

	var = qu_var_new (CoMaster, name, f | QU_VAR_FLAG_STRUCT | QU_VAR_FLAG_FROZEN);
	addsymbol (name, (qu_symbol) var);

	top = pnode (QU_AST_struct);
	top->thing = var;

	match (QU_T_COLON);

	if (CoUseIndent)
		indent ("struct");

	for (;;)
	{
		if (CoUseIndent)
		{
			if (match (QU_T_DEDENT))
				break;

			if (peek == QU_T_DOC)
			{
				dl = CoToken->cursor - di + CoToken->len;
				eat;
				if (CoDocumenting)
				{
					docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
					gotdoc = 1;
				}
				dedent ("struct");
				break;
			}
		}
		else
		{
			if (match (QU_T_end))
			{
				nobrk ("after end");
				req (QU_T_struct, "end struct expected");
				break;
			}
			else if (match (QU_T_END))
			{
				break;
			}
		}

		argmax (++n, "too many struct items");

		identifier ("as struct item");
		adc (top, c, pnode (QU_AST_literal));
		grab (c);
		match (QU_T_COLON);

		type = _type ();
		if (type == 0)
		{
			/* we're in "documentation" mode */
			/* just ignore and set to qu_object */
			type = pnode (QU_AST_symbol);
			type->thing = QuObjectClass->name;
		}
		adc (top, c, type);
		match (QU_T_COMMA);
	}
	top->op = n;

	if (CoUseIndent)
	{
		if (CoDocumenting && !gotdoc)
		{
			dl = CoToken->cursor - di + CoToken->len;
			docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}
	else
	{
		dl = CoToken->cursor - di + CoToken->len;
		if (match (QU_T_DOC))
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), (qu_string) CoToken->thing);
		}
		else
		{
			if (CoDocumenting)
				docentry (var, qu_string_new (di, dl), QuStringEMPTY);
		}
	}

	/* preallocate the struct so nested structs and late declared structs work */
	var->value = qu_struct_compile_prealloc ();
	return top;
}



/*
	if
	|
	<expr> - compound -  [if*] - [compound (else)]

	if
	|
	unop (op: qu_sysnot) - compound -  [if*] - [compound (else)]
	|
	<expr>
*/
static qu_ast _if ()
{
	qu_ast top, c;

	nobrk ("after if/unless");
	++CoInIf;

	top = pnode (QU_AST_if);
	top->child = c = _expr ();

	match (QU_T_COLON);

	if (CoUseIndent)
		indent ("if/unless");

	rpl (c, _compound ());

	if (CoUseIndent)
		dedent ("if/unless");

	while (match (QU_T_elif))
	{
		qu_ast x;

		nobrk ("after elif");
		x = pnode (QU_AST_if);
		x->child = _expr ();
		rpl (c, x);

		match (QU_T_COLON);

		if (CoUseIndent)
			indent ("elif");

		x->child->next = _compound ();

		if (CoUseIndent)
			dedent ("elif");
	}

	if (match (QU_T_else))
	{
		match (QU_T_COLON);

		if (CoUseIndent)
			indent ("else");

		rpl (c, _compound ());

		if (CoUseIndent)
			dedent ("else");
	}

	if (!CoUseIndent)
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_if, "end if expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close if block");
		}
	}

	--CoInIf;
	return top;
}



static qu_ast _unless ()
{
	qu_ast top = _if ();
	qu_ast c   = top->child;
	qu_ast u   = rnode (QU_AST_unop);

	u->op    = (long) qu_sys_not;
	u->child = c;
	top->child = u;
	u->next = c->next;
	return top;
}



/*
	switch
	|
	<expr> - case | nop - case*
	          (default)

	case (op: C function address 0 for default)
	|
	[<expr> | compound] - compound

	case 1, 2, 3      -> test 1, 2, or 3
	case (1, 2, 3)    -> test Array (1, 2, 3)
	case in (1, 2, 3) -> test 1, 2, or 3
*/
static qu_ast _switch ()
{
	qu_ast top, c, x, g, d;
	int    t, n = 0, k = 0;

	nobrk ("after switch");

	top = pnode (QU_AST_switch);
	top->child = c = _expr ();

	match (QU_T_COLON);

	d = pnode (QU_AST_nop);
	rpl (c, d);

	enter (BLOCK_SWITCH);
	setlabel ();

	if (CoUseIndent)
		indent ("switch");

	for (;;)
	{
		switch (peek)
		{
			case QU_T_DEDENT:
				eat;
				assert (CoUseIndent);
				break;

			case QU_T_end:
				if (CoUseIndent)
					croak ("cant use end or ;; with indentation style");
				nobrk ("after end");
				req (QU_T_switch, "end switch expected");
				/*nobreak*/

			case QU_T_END:
				if (CoUseIndent)
					croak ("cant use end or ;; with indentation style");
				if (n == 0)
					croak ("switch block cant be empty");
				eat;
				break;

			case QU_T_default:
				eat;
				match (QU_T_COLON);
				if (k)
					croak ("duplicate default case");
				k = 1;
				d->type = QU_AST_case;

				if (CoUseIndent)
					indent ("default");

				d->child = _compound ();

				if (CoUseIndent)
					dedent ("default");

				if (peek == QU_T_case)
					croak ("default case must be last");
				continue;

			case QU_T_case:
				eat;
				nobrk ("after case");
				n++;
				x = pnode (QU_AST_case);
				switch (t = peek)
				{
					case QU_T_LT:    eat; x->op = (long) qu_sys_lt;   break;
					case QU_T_GT:    eat; x->op = (long) qu_sys_gt;   break;
					case QU_T_LE:    eat; x->op = (long) qu_sys_le;   break;
					case QU_T_GE:    eat; x->op = (long) qu_sys_ge;   break;
					case QU_T_CONTRA:
					case QU_T_NE:    eat; x->op = (long) qu_sys_ne;   break;
					case QU_T_in:    eat; x->op = (long) qu_sys_rin;  break;
					case QU_T_LIKE:  eat; x->op = (long) qu_sys_like; break;
					case QU_T_is:    eat; x->op = (long) qu_sys_is;   break;

					case QU_T_QUEST:
						/*
							case ? expr:

						        case
						          |
						         unop (thing: nil, op: qu_sys_true)
						          |
						        <expr>
						*/
						eat;
						x->child = pnode (QU_AST_unop);
						x->child->op = (long) qu_sys_true;
						x->child->thing = QuNil;
						x->child->child = _expr ();
						rpl (c, x);
						goto colon;

					case QU_T_DOT:
						eat;
						identifier ("after .");

						if (sameline && peek == QU_T_LPAREN)
						{
							/*
								case .method (a, b, c):

							        case
							          |
							         call (op: narg, thing: nil)
							        / | \
							  literal .....

							*/
							qu_ast name = pnode (QU_AST_literal);
							qu_ast call, tmp;

							name->thing = CoToken->thing;
							call = _factor ();
							if (call->type != QU_AST_list)
							{
								par (call, QU_AST_call);
								call->op = 1;
							}
							else
							{
								call->type = QU_AST_call;
							}
							tmp = call->child;
							call->child = name;
							name->next = tmp;
							x->child = call;
							call->thing = QuNil;
						}
						else
						{
							/*
								case .method:

								case
								|
								literal (op: 1)
							*/
							x->child = pnode (QU_AST_literal);
							x->child->thing = CoToken->thing;
							x->child->op = 1;
						}
						rpl (c, x);
						goto colon;

					case QU_T_EQ:
						eat;
					default:
						x->op = (long) qu_sys_eq;
				}
				rpl (c, x);
				x = _expr ();
				if (match (QU_T_COMMA))
				{
					g = c->child = pnode (QU_AST_compound);
					g->child = x;
					do
					{
						rpl (x, _expr ());

					} while (match (QU_T_COMMA));
				}
				else
				{
					c->child = x;
				}

			  colon:

				match (QU_T_COLON);

				if (CoUseIndent)
					indent ("case");

				c->child->next = _compound ();

				if (CoUseIndent)
					dedent ("case");

				continue;

			default:
				croak ("case or default expected");
		}
		break;
	}
	leave (BLOCK_SWITCH);
	return top;
}



/*
	do
	|
	if | switch | loop | while | for | compound
*/
static qu_ast _do ()
{
	qu_ast top, c = 0, x = 0;
	int    plain = 0;

	top = pnode (QU_AST_do);
	enter (BLOCK_DO);
	setlabel ();
	if (sameline)
	{
		switch (peek)
		{
			case QU_T_if:     eat; x = _if     (); break;
			case QU_T_switch: eat; x = _switch (); break;
			case QU_T_loop:   eat; x = _loop   (); break;
			case QU_T_while:  eat; x = _while  (); break;
			case QU_T_for:    eat; x = _for    (); break;
			case QU_T_try:    eat; x = _try    (); break;
			case QU_T_unless: eat; x = _unless (); break;

			default:
				croak ("if, unless, switch, loop, while, for, or try expected as do modifier");
		}
	}
	else
	{
		plain = 1;
		match (QU_T_COLON);

		if (CoUseIndent)
			indent ("do");

		x = _compound ();
	}
	adc (top, c, x);

	if (plain)
	{
		if (CoUseIndent)
		{
			dedent ("do");
		}
		else
		{
			if (match (QU_T_end))
			{
				nobrk ("after end");
				req (QU_T_do, "end do expected");
			}
			else
			{
				req (QU_T_END, ";; expected to close do");
			}
		}
	}
	leave (BLOCK_DO);
	return top;
}



static qu_ast _ifmodifier ()
{
	if (sameline)
	{
		int t = peek;

		switch (t)
		{
			case QU_T_if:
				eat;
				return _expr ();

			case QU_T_unless:
			{
				qu_ast c;

				eat;
				c = pnode (QU_AST_unop);
				c->op = (long) qu_sys_not;
				c->child = _expr ();
				return c;
			}
		}
	}
	return NULL;
}



/*
	loop
	|
	[<expr>] - compound - [<expr>]
*/
static qu_ast _loop ()
{
	qu_ast top, c = _ifmodifier ();

	match (QU_T_COLON);

	top = pnode (QU_AST_loop);
	top->child = c;

	enter (BLOCK_LOOP);
	setlabel ();

	if (CoUseIndent)
		indent ("loop");

	adc (top, c, _compound ());

	if (CoUseIndent)
		dedent ("loop");

	if (match (QU_T_until))
	{
		nobrk ("after until");
		rpl (c, _expr ());
	}
	else
	{
		if (!CoUseIndent)
		{
			if (match (QU_T_end))
			{
				nobrk ("after end");
				req (QU_T_loop, "end loop expected");
			}
			else
			{
				req (QU_T_END, ";; expected to close loop");
			}
		}
	}
	leave (BLOCK_LOOP);
	return top;
}



/*
	while
	|
	<expr> - compound
*/
static qu_ast _while ()
{
	qu_ast top, c;

	nobrk ("after while");
	c = _expr ();

	match (QU_T_COLON);
	top = pnode (QU_AST_while);
	top->child = c;
	enter (BLOCK_WHILE);
	setlabel ();

	if (CoUseIndent)
		indent ("while");

	rpl (c, _compound ());

	if (CoUseIndent)
	{
		dedent ("while");
	}
	else
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_while, "end while expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close while");
		}
	}
	leave (BLOCK_WHILE);
	return top;
}



/*
	for
	|
	(name | list) - <expr> - compound

	list (op: number of items)
	|
	name+ - [eat]

	'name' in the first form instead of 'symbol' is intentional:
	we must only use local block symbol to store the value of an iterator.
*/
static qu_ast _for_head ()
{
	qu_ast top, x, c = 0;
	int    par = 0;
	int    k = 0, n = 0, m;
	qu_ast l, i;

	nobrk ("after for");
	top = pnode (QU_AST_for);

	/* loosen up a bit ... */
	match (QU_T_each);

	par = match (QU_T_LPAREN);

	l = pnode (QU_AST_list);
	adc (top, c, l);
	x = NULL;

	do
	{
		argmax (++n, "too many for arguments");

		if (match (QU_T_MUL))
		{
			if (k == 0)
				croak ("* cant be the first argument");
			identifier ("after *");
			i = pnode (QU_AST_eat);
			grab (i);
			adc (l, x, i);
			if (peek == QU_T_NAME)
			{
				eat;
				croak ("no more argument after *");
			}
			if (par)
				req (QU_T_RPAREN, ") expected to close parenthesized for arguments");
			break;
		}
		k = 1;
		identifier ("as for arguments");
		i = pnode (QU_AST_name);
		grab (i);
		adc (l, x, i);

		match (QU_T_COMMA);

		if (par)
		{
			if (match (QU_T_RPAREN))
				break;
		}

		m = peek;

	} while (m == QU_T_MUL || m == QU_T_NAME);

	l->op = n;

	if (n == 1)
	{
		c = l->child;
		top->child = c;
	}

	switch (peek)
	{
		case QU_T_in:
		case QU_T_as:
		case QU_T_COLON:
		case QU_T_ASSIGN:
		case QU_T_ARROW:
			eat;
			break;

		default:
			eat;
			croak ("unexpected token");
	}

	x = _expr ();
	if (x->type != QU_AST_list && match (QU_T_COMMA))
	{
		adc (top, c, pnode (QU_AST_list));
		c->child = x;
		adc (c, x, _expr ());
		c->op = 2;
		if (match (QU_T_COMMA))
		{
			adc (c, x, _expr ());
			c->op = 3;
		}
		if (match (QU_T_COMMA))
			croak ("for iteration needs at most three arguments");
	}
	else
	{
		adc (top, c, x);
	}
	return top;
}



static qu_ast _for ()
{
	qu_ast top, c;

	top = _for_head ();
	c = top->child->next;

	match (QU_T_COLON);

	enter (BLOCK_FOR);
	setlabel ();

	if (CoUseIndent)
		indent ("for");

	adc (top, c, _compound ());

	if (CoUseIndent)
	{
		dedent ("for");
	}
	else
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_for, "end for expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close for");
		}
	}
	leave (BLOCK_FOR);
	return top;
}



/*
	try
	|
	[<expr>] - compound [- catch* - [else] - [finally]]

	catch (op:number of exceptions, thing: 0 or name to store the exception)
	|
	symbol* - compound

	else      finally
	|         |
	block     block

	Note:

		try
			foo
		;;

	means:

		try
			foo
		catch
			# do nothing
		;;
*/
static qu_ast _catch (qu_ast try)
{
	qu_ast top, c = 0;
	int    n = 0;

	top = pnode (QU_AST_catch);
	if (sameline && peek != QU_T_COLON)
	{
		if (try->op)
			croak ("catch expression cant follow default catch");

		if (peek == QU_T_ARROW)
			goto default_catch;

		do
		{
			argmax (++n, "too many catch arguments");
			adc (top, c, _symbol ());

		} while (match (QU_T_COMMA));
	}
	else
	{
	  default_catch:
		if (try->op)
			croak ("cant specify more than one default catch");
		try->op = 1;
	}
	top->op = (long) n;

	if (match (QU_T_ARROW))
	{
		identifier ("after ->");
		grab (top);
	}

	match (QU_T_COLON);

	if (CoUseIndent)
		indent ("catch");

	adc (top, c, _compound ());

	if (CoUseIndent)
		dedent ("catch");

	return top;
}



static qu_ast _try ()
{
	qu_ast top = pnode (QU_AST_try);
	qu_ast c   = _ifmodifier ();
	qu_ast x;

	top->child = c;
	match (QU_T_COLON);

	enter (BLOCK_TRY);

	if (CoUseIndent)
		indent ("try");

	adc (top, c, _compound ());

	if (CoUseIndent)
		dedent ("try");

	if (match (QU_T_catch))
	{
		do
		{
			adc (top, c, _catch (top));

		} while (match (QU_T_catch));

		x = pnode (QU_AST_catch);
		adc (top, c, x);
		x->child = pnode (QU_AST_stmt);
		x->child->child = pnode (QU_AST_throw);

	  proceed:

		if (match (QU_T_else))
		{
			match (QU_T_COLON);

			if (CoUseIndent)
				indent ("else");

			x = _compound ();

			if (CoUseIndent)
				dedent ("else");

			x->type = QU_AST_else;
			rpl (c, x);
		}

		if (match (QU_T_finally))
		{
			match (QU_T_COLON);

			if (CoUseIndent)
				indent ("finally");

			x = _compound ();

			if (CoUseIndent)
				dedent ("finally");

			x->type = QU_AST_finally;
			rpl (c, x);
		}
	}
	else
	{
		adc (top, c, rnode (QU_AST_catch));
		c->child = rnode (QU_AST_compound);
		goto proceed;
	}

	if (!CoUseIndent)
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_try, "end try expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close try");
		}
	}
	leave (BLOCK_TRY);
	return top;
}



/*
	atomic
	|
	stmt*
*/
static qu_ast _atomic ()
{
	qu_ast top;

	if (CoInAtomic)
		croak ("cant nest atomic block");
	CoInAtomic = 1;

	match (QU_T_COLON);
	enter (BLOCK_ATOMIC);

	if (CoUseIndent)
		indent ("atomic");

	top = _compound ();
	top->type = QU_AST_atomic;

	if (CoUseIndent)
	{
		dedent ("atomic");
	}
	else
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_atomic, "end atomic expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close atomic");
		}
	}

	leave (BLOCK_ATOMIC);
	CoInAtomic = 0;
	return top;
}



/*
	with name in expr
		compound
	;;

	translates to:

	for x in _anon_ = expr
		try
			compound
		catch
			_anon_.__stop (__err__)
		;;
	;;

	for
	|
	(name | list) - assign - compound
	                 / \      |
	            name   expr  try
	           (anon)         |
	                      compound - catch
	                                   |
	                                compound
	                                   |
	                                  stmt
	                                    |
	                                  call (op:1)
	                                  /  \
	                           getattr    literal (thing: __QuSys__err__)
	                             /   \
	                         name  literal (thing: '__stop')
	                        (anon)
*/
static qu_ast _with ()
{
	qu_ast    top, a, e, c;
	qu_string name;

	if (CoUseIndent)
		indent ("with");

	name = anon ();
	top = _for_head ();
	e = top->child->next;

	top->child->next = a = rnode (QU_AST_assign);
	a->child = rnode (QU_AST_name);
	a->child->thing = name;
	a->child->next = e;
	e->next = 0;
	a->next = c = rnode (QU_AST_compound);

	c->child = rnode (QU_AST_try);
	c = c->child;
	c->op = 1;

	c->child = _compound ();
	c = c->child;

	c->next = rnode (QU_AST_catch);
	c = c->next;

	c->child = rnode (QU_AST_compound);
	c = c->child;

	c->child = rnode (QU_AST_stmt);
	c = c->child;

	c->child = rnode (QU_AST_call);
	c = c->child;
	c->op = 1;

	a = c->child = rnode (QU_AST_getattr);
	a->op = 1;

	a->child = rnode (QU_AST_name);
	a->child->thing = name;
	a->child->next = rnode (QU_AST_literal);
	a->child->next->thing = QU_NAME__stop;

	a->next = rnode (QU_AST_literal);
	a->next->thing = __QuSys__err__;

	if (CoUseIndent)
	{
		dedent ("with");
	}
	else
	{
		if (match (QU_T_end))
		{
			nobrk ("after end");
			req (QU_T_with, "end with expected");
		}
		else
		{
			req (QU_T_END, ";; expected to close with");
		}
	}
	return top;
}



/*
	assert (thing: expression String)
	|
	<expr>

	note: there is no stmt on top of assert
*/
static qu_ast _assert ()
{
	if (QU_APP_DEBUG)
	{
		char   *di = (char *) CoToken->cursor + CoToken->len;
		char   *dl = di;
		qu_ast  top = pnode (QU_AST_assert);

		top->child = _expr ();

		dl = (char *) CoToken->cursor + CoToken->len;
		while (qu_ascii_isspace (*di))
			++di;
		while (dl > di && qu_ascii_isspace (*dl))
			--dl;
		top->thing = qu_string_new (di, dl - di + 1);
		return top;
	}
	else
	{
		_expr ();
		return 0;
	}
}



/*
	__doc__

	This does not produce a node. The function always return NULL.
*/
static qu_ast _doc ()
{
	qu_string name = 0;
	long      f = 0;
	qu_var    v;

	if (qu_class_is (CoBlock))
		f = QU_VAR_FLAG_CLASS | QU_VAR_FLAG_FINAL | QU_VAR_FLAG_FROZEN;
	else if (qu_module_is (CoBlock))
		f = QU_VAR_FLAG_FROZEN;
	else
		croak ("__doc__ may only be declared in Class or Module level");

	nobrk ("after __doc__");

	switch (peek)
	{
		case QU_T_NAME:
		case QU_T_STRING:
			eat;
			name = (qu_string) CoToken->thing;
			break;

		default:
			croak ("identifier or string expected after __doc__");
	}
	unique (name);

	if (CoUseIndent)
		indent ("__doc__");

	req (QU_T_DOC, "documentation string expected");

	if (CoDocumenting)
	{
		v = qu_var_new (CoMaster, name, f | QU_VAR_FLAG_DOCENTRY);
		addsymbol (name, (qu_symbol) v);
		v->value = CoToken->thing;
		docentry (v, v->name, v->value);
	}

	if (CoUseIndent)
		dedent ("__doc__");

	return NULL;
}



/*
	<tag             -> html.open ("tag", false, false)
	<tag>            -> html.open ("tag", true, false)
	<tag>[linebreak] -> html.open ("tag", true, true)

	htopen (thing:String, op: closed|linebreak)
*/
static qu_ast _htopen (int closed, int linebreak)
{
	qu_ast x = pnode (QU_AST_htopen);

	grab (x);
	x->op = closed ? 0x01 : 0;
	if (linebreak)
		x->op |= 0x10;
	return x;
}


/*
	>                 -> html.close ("", false)
	>[lineberak]      -> html.close ("", true)
	/>                -> html.close ("/", false)
	/>[lineberak]     -> html.close ("/", true)
	</tag>            -> html.close ("tag", false)
	</tag>[linebreak] -> html.close ("tag", true)

	htclose (thing:String op:linbreak)
*/
static qu_ast _htclose (int linebreak)
{
	qu_ast x = pnode (QU_AST_htclose);

	x->thing = CoToken->thing ? CoToken->thing : QuStringEMPTY;
	x->op = linebreak;
	return x;
}


/*
	&foo; -> html.entity ("foo")

	htentity (thing:String)
*/
static qu_ast _htentity (int linebreak)
{
	qu_ast x = pnode (QU_AST_htentity);
	grab (x);
	return x;
}



/*----------------------------------------------------------------------------
	EXPRESSION
----------------------------------------------------------------------------*/

static qu_ast _unop (qu_ast a, qu_unop_f f)
{
	par (a, QU_AST_unop);
	a->op = (long) f;
	return a;
}



static qu_ast _binop (qu_ast a, qu_ast b, qu_binop_f f)
{
	a->next = b;
	par (a, QU_AST_binop);
	a->op = (long) f;
	return a;
}



/*
	a or b

	binop (op: qu_sys_bor)
	 / \
	a   b
*/
static qu_ast _bor ()
{
	qu_ast top;

	__enter_expression;
	top = _band ();
	while (match (QU_T_or))
	{
		top = _binop (top, _band (), (qu_binop_f) qu_sys_bor);
	}
	__leave_expression;
	return top;
}



/*
	a and b

	binop (op: qu_sys_band)
	 / \
	a   b
*/
static qu_ast _band ()
{
	qu_ast top;

	top = _not ();
	while (match (QU_T_and))
	{
		top = _binop (top, _not (), (qu_binop_f) qu_sys_band);
	}
	return top;
}



/*
	not a

	unop (op: qu_sys_not)
	|
	a
*/
static qu_ast _not ()
{
	qu_ast top;
	int    k;

	k = match (QU_T_not);
	top = _assign ();
	if (k)
		top = _unop (top, (qu_unop_f) qu_sys_not);
	return top;
}



/*
	a  = b
	a += b
	a -= b
	a *= b
	a /= b
	a |= b
	a &= b
	a ^= b
	a ||= b
	a &&= b

	assign (op: 0 | qu_sys_concat | qu_sys_plus | etc.)
	 / \
	a   b
*/
static qu_ast _assign ()
{
	qu_ast top, c;
	int    t;
	long   op;

	top = _quest ();
	for (;;)
	{
		switch (t = peek)
		{
		case QU_T_ASSIGN:
		case QU_T_ASCONCAT:
		case QU_T_ASPLUS:
		case QU_T_ASMINUS:
		case QU_T_ASMUL:
		case QU_T_ASDIV:
		case QU_T_ASIDIV:
		case QU_T_ASBITOR:
		case QU_T_ASBITXOR:
		case QU_T_ASBITAND:
		case QU_T_ASOR:
		case QU_T_ASAND:
		case QU_T_ASLSHIFT:
		case QU_T_ASRSHIFT:
		case QU_T_ASREM:
			eat;
			switch (top->type)
			{
				case QU_AST_name:
				case QU_AST_getattr:
				case QU_AST_getitem:
				case QU_AST_getslice:
				case QU_AST_assign:
					break;

				case QU_AST_list:
					if (t != QU_T_ASSIGN)
						croak ("list assignment only supports the = operator");
					for (c = top->child; c != NULL; c = c->next)
					{
						switch (c->type)
						{
							case QU_AST_name:
							case QU_AST_getattr:
							case QU_AST_getitem:
							case QU_AST_getslice:
							case QU_AST_assign:
							case QU_AST_eat:
								break;

							case QU_AST_list:
								croak ("cant do nested list assignment");

							case QU_AST_literal:
								if (c->thing == QuNil)
									break;

							default:
								croak ("invalid assignment");
						}
					}
					break;

				default:
					croak ("invalid assignment");
			}

			__enter_expression;
			top->next = _assign ();
			__leave_expression;

			par (top, QU_AST_assign);
			switch (t)
			{
				case QU_T_ASCONCAT: op = (long) qu_sys_concat; break;
				case QU_T_ASPLUS:   op = (long) qu_sys_plus;   break;
				case QU_T_ASMINUS:  op = (long) qu_sys_minus;  break;
				case QU_T_ASMUL:    op = (long) qu_sys_mul;    break;
				case QU_T_ASDIV:    op = (long) qu_sys_div;    break;
				case QU_T_ASIDIV:   op = (long) qu_sys_idiv;   break;
				case QU_T_ASBITOR:  op = (long) qu_sys_ior;    break;
				case QU_T_ASBITXOR: op = (long) qu_sys_ixor;   break;
				case QU_T_ASBITAND: op = (long) qu_sys_iand;   break;
				case QU_T_ASOR:     op = (long) qu_sys_or;     break;
				case QU_T_ASAND:    op = (long) qu_sys_and;    break;
				case QU_T_ASLSHIFT: op = (long) qu_sys_lshift; break;
				case QU_T_ASRSHIFT: op = (long) qu_sys_rshift; break;
				case QU_T_ASREM:    op = (long) qu_sys_rem;    break;
				default:         op = 0;
			}
			top->op = op;
			continue;
		}
		break;
	}
	return top;
}



/*
	a ? b [: c]

	 quest
	 / | \
	a  b [c]
*/
static qu_ast _quest ()
{
	qu_ast top;

	top = _or ();
	while (match (QU_T_QUEST))
	{
		par (top, QU_AST_quest);
		top->child->next = _or ();
		if (match (QU_T_COLON))
			top->child->next->next = _or ();
	}
	return top;
}



/*
	a || b

	binop (op: qu_sys_or)
	 / \
	a   b
*/
static qu_ast _or ()
{
	qu_ast top;

	top = _and ();
	while (match (QU_T_OR))
	{
		top = _binop (top, _and (), (qu_binop_f) qu_sys_or);
	}
	return top;
}



/*
	a && b

	binop (op: qu_sys_and)
	 / \
	a   b
*/
static qu_ast _and ()
{
	qu_ast top;

	top = _cmp ();
	while (match (QU_T_AND))
	{
		top = _binop (top, _cmp (), (qu_binop_f) qu_sys_and);
	}
	return top;
}



/*
	a === b
	a == b
	a != b
	a >  b
	a >= b
	a <  b
	a <= b

	binop (op: qu_sys_eq | qu_sys_like | etc.)
	 / \
	a   b
*/
static qu_ast _cmp ()
{
	qu_ast top;

	top = _bitor ();
	for (;;)
	{
		switch (peek)
		{
			case QU_T_EQ:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_eq);
				continue;

			case QU_T_CONTRA:
			case QU_T_NE:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_ne);
				continue;

			case QU_T_LT:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_lt);
				continue;

			case QU_T_LE:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_le);
				continue;

			case QU_T_GT:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_gt);
				continue;

			case QU_T_GE:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_ge);
				continue;

			case QU_T_CMP:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_cmp);
				continue;

			case QU_T_LIKE:
				eat;
				top = _binop (top, _bitor (), (qu_binop_f) qu_sys_like);
				continue;
		}
		break;
	}
	return top;
}



/*
	a | b
	a ^ b

	binop (op: qu_sys_ior | qu_sys_ixor)
	 / \
	a   b
*/
static qu_ast _bitor ()
{
	qu_ast top;

	top = _bitand ();
	for (;;)
	{
		switch (peek)
		{
			case QU_T_BITOR:
				eat;
				top = _binop (top, _bitand (), qu_sys_ior);
				continue;

			case QU_T_BITXOR:
				eat;
				top = _binop (top, _bitand (), qu_sys_ixor);
				continue;
		}
		break;
	}
	return top;
}



/*
	a & b

	binop (op: qu_sys_iand)
	 / \
	a   b
*/
static qu_ast _bitand ()
{
	qu_ast top;

	top = _shift ();
	while (match (QU_T_BITAND))
	{
		top = _binop (top, _shift (), qu_sys_iand);
	}
	return top;
}



/*
	a << b
	a >> b

	binop (op: qu_sys_lshift | qu_sys_rshift)
	 / \
	a   b
*/
static qu_ast _shift ()
{
	qu_ast top;

	top = _arith ();
	for (;;)
	{
		switch (peek)
		{
			case QU_T_LSHIFT:
				eat;
				top = _binop (top, _arith (), qu_sys_lshift);
				continue;

			case QU_T_RSHIFT:
				eat;
				top = _binop (top, _arith (), qu_sys_rshift);
				continue;
		}
		break;
	}
	return top;
}



/*
	a + b
	a - b
	a _ b

	binop (op: qu_sys_plus | qu_sys_minus | qu_sys_concat)
	 / \
	a   b
*/
static qu_ast _arith ()
{
	qu_ast top;

	top = _term ();
	for (;;)
	{
		switch (peek)
		{
			case QU_T_PLUS:
				eat;
				top = _binop (top, _term (), qu_sys_plus);
				continue;

			case QU_T_MINUS:
				eat;
				top = _binop (top, _term (), qu_sys_minus);
				continue;

			case QU_T_CONCAT:
				eat;
				top = _binop (top, _term (), (qu_binop_f) qu_sys_concat);
				continue;
		}
		break;
	}
	return top;
}



/*
	a *  b
	a /  b
	a %  b

	binop (op: qu_sys_mul | qu_sys_div | etc.)
	 / \
	a   b

	a in b
	a not in b
	a is b
	a is not b

	[unop]
	  |
	binop
	 / \
	b   a
*/
static qu_ast _term ()
{
	qu_ast top;

	top = _power ();
	for (;;)
	{
		switch (peek)
		{
			case QU_T_MUL:
				eat;
				top = _binop (top, _power (), qu_sys_mul);
				continue;

			case QU_T_DIV:
				eat;
				top = _binop (top, _power (), qu_sys_div);
				continue;

			case QU_T_IDIV:
				eat;
				top = _binop (top, _power (), qu_sys_idiv);
				continue;

			case QU_T_REM:
				eat;
				top = _binop (top, _power (), qu_sys_rem);
				continue;

			case QU_T_in:
				eat;
				top = _binop (_power (), top, (qu_binop_f) qu_sys_in);
				continue;

			case QU_T_is:
			{
				int k;

				eat;
				k = match (QU_T_not);
				top = _binop (top, _power (), (qu_binop_f) qu_sys_is);
				if (k)
				{
					top = _unop (top, (qu_unop_f) qu_sys_not);
				}
				continue;
			}

			case QU_T_not:
				if (peekn (2) == QU_T_in)
				{
					eat;
					eat;
					top = _binop (_power (), top, (qu_binop_f) qu_sys_in);
					top = _unop (top, (qu_unop_f) qu_sys_not);
					continue;
				}
				break;

			case QU_T_are:
			{
				int k;

				eat;
				k = match (QU_T_not);
				top = _binop (top, _power (), (qu_binop_f) qu_sys_are);
				if (k)
				{
					top = _unop (top, (qu_unop_f) qu_sys_not);
				}
				continue;
			}
		}
		break;
	}
	return top;
}



/*
	a ** b
	a .. b
	a %% b
	a <<< b
	a >>> b
	a :: b
	a of b
	a _/ b

	binop (op: qu_sys_pow | qu_sys_range | etc.)
	 / \
	a   b
*/
static qu_ast _power ()
{
	qu_ast top;

	top = _unary ();
	for (;;)
	{
		switch (peek)
		{
			case QU_T_POW:
				eat;
				top = _binop (top, _unary (), qu_sys_pow);
				continue;

			case QU_T_RATIO:
				eat;
				top = _binop (top, _unary (), qu_sys_ratio);
				continue;

			case QU_T_DOT2:
				eat;
				top = _binop (top, _unary (), qu_sys_range);
				continue;

			case QU_T_POP:
				/*
					 assign (op: 0)
					 /    \
					b    unop (op: qu_sys_pop)
					      |
					      a
				*/
				eat;
				{
					qu_ast g = pnode (QU_AST_assign);
					qu_ast u = pnode (QU_AST_unop);

					g->child = _unary ();
					g->child->next = u;
					u->child = top;
					u->op = (long) qu_sys_pop;
					top = g;
				}
				continue;

			case QU_T_PUSH:
				eat;
				top = _binop (top, _unary (), qu_sys_push);
				continue;

			case QU_T_XCHG:
				eat;
				top = _binop (top, _unary (), qu_sys_xchg);
				continue;

			case QU_T_FORMAT:
				eat;
				top->next = _unary ();
				par (top, QU_AST_method);
				top->op = (long) qu_sys_format;
				continue;

			case QU_T_as:
				eat;
				top = _binop (top, _unary (), qu_sys_as);
				continue;

			case QU_T_of:
				eat;
				par (top, QU_AST_binop);
				top->op = (long) qu_sys_of;
				if (match (QU_T_LPAREN))
				{
					qu_ast l = rnode (QU_AST_list);
					l->op = 2;
					top->child->next = l;
				  match (QU_T_AT); /* backward compatible */
					l->child = _symbol ();
					match (QU_T_COMMA);
				  match (QU_T_AT);
					l->child->next = _symbol ();
					req (QU_T_RPAREN, ") expected");
				}
				else
				{
				  match (QU_T_AT);
					top->child->next = _symbol ();
					if (match (QU_T_COMMA))
					{
						qu_ast l = rnode (QU_AST_list);
						l->op = 2;
						l->child = top->child->next;
						top->child->next = l;
					  match (QU_T_AT);
						l->child->next = _symbol ();
					}
				}
				continue;

			case QU_T_COLON2:
				eat;
				top = _binop (top, _unary (), qu_sys_rep);
				continue;
		}
		break;
	}
	return top;
}



/*
	-a
	+a
	!a
	~a
	++a
	--a

	unop (op: qu_sys_neg | qu_sys_pos | etc.)
	|
	a
*/
static qu_ast _unary ()
{
	qu_ast n;
	int  t;

	switch (t = peek)
	{
		case QU_T_MINUS:
			eat;
			return _unop (_trailer (), qu_sys_neg);

		case QU_T_PLUS:
			eat;
			return _unop (_trailer (), qu_sys_pos);

		case QU_T_NOT:
			eat;
			return _unop (_trailer (), (qu_unop_f) qu_sys_not);

		case QU_T_BITNOT:
			eat;
			return _unop (_trailer (), qu_sys_inot);

		case QU_T_INCR:
		case QU_T_DECR:
			eat;
			n = pnode (QU_AST_unop);
			n->op = (t == QU_T_INCR) ? (long) qu_sys_incr : (long) qu_sys_decr;
			n->child = _trailer ();
			switch (n->child->type)
			{
				case QU_AST_name:
				case QU_AST_getattr:
				case QU_AST_getitem:
					break;

				default:
					croak ("invalid ++ or -- operand");
			}
			return n;
	}
	return _trailer ();
}



/*
	a.name
	a.(expr)

		getattr (op: 1=followed by call() 0=not)
		 /  \
		a    (LITERAL | expr)

		LITERAL
		|
		op = name


	a[b]
	a`b

	    getitem
		 /  \
		a    b

	a(b, c, d)

		call (op: number of arguments)
		|
		a - args

		args = (<expr> | flat)*

		flat
		|
		<expr>

	Example:

		foo.bar
		foo [bar * 2]

		foo (a, b)

			    call (op: 2)
			   / | \
			foo  a  b

		foo (a, *b)

				call (op: 2)
			   / |  \
			foo  a  flat
			          |
			          b
*/
static void _call_check_extern (qu_ast top)
{
	qu_ast     e = top->child;
	int        n = top->op;
	qu_array   a = e->thing;
	qu_string  s = a->objs[2];
	int        m = qu_dlib_check_spec (s->str, s->len);

	switch (m)
	{
		case -1: croak ("invalid argument specification");
		case -2: croak ("invalid argument specification missing :");
		case -3: croak ("too many arguments");
	}

	if (n != m)
		croak ("mismatched number of arguments");
}


static qu_ast _trailer ()
{
	qu_ast top, c, x;
	int    t, n;

	top = _factor ();
	while (sameline)
	{
		switch (t = peek)
		{
			case QU_T_INCR:
			case QU_T_DECR:
				eat;
				switch (top->type)
				{
					case QU_AST_name:
					case QU_AST_getattr:
					case QU_AST_getitem:
						break;

					default:
						croak ("invalid ++ or -- operand");
				}
				par (top, QU_AST_unop);
				top->op = (t == QU_T_INCR) ? (long) qu_sys_incr : (long) qu_sys_decr;
				top->thing = QuNil;
				continue;

			case  QU_T_DOT:
				eat;

				{
					/*
						b = a.+(2)  -> b = a.__plus (2)
					*/

					qu_string op = opname ();

					if (op)
					{
						if (!(sameline && peek == QU_T_LPAREN))
							nobrk ("before call arguments");
						par (top, QU_AST_getattr);
						x = pnode (QU_AST_literal);
						x->thing = op;
						goto dotset;
					}
				}

				par (top, QU_AST_getattr);
				if (match (QU_T_NAME))
				{
					x = pnode (QU_AST_literal);
					grab (x);
				}
				else
				{
					req (QU_T_LPAREN, "identifier or ( expected after .");
					x = _expr ();
					req (QU_T_RPAREN, ") expected");
					if (x->type == QU_AST_literal)
					{
						if (!qu_string_is (x->thing))
							croak ("attribute literal must be a string");
					}
				}

			  dotset:

				top->child->next = x;
				if (sameline && peek == QU_T_LPAREN)
					top->op = 1;
				continue;

			case QU_T_FIELD:
				eat;
				top->next = c = pnode (QU_AST_literal);
				grab (c);
				par (top, QU_AST_getitem);
				continue;

			case QU_T_ARROW:
				eat;
				identifier ("after ->");
				top->next = c = pnode (QU_AST_literal);
				grab (c);
				par (top, QU_AST_getitem);
				continue;

			case QU_T_LSQUARE:
				eat;
				if (match (QU_T_COMMA))
				{
					c = top->next = pnode (QU_AST_literal);
					c->thing = QuInt0;
					par (top, QU_AST_getslice);
					c->next = _expr ();
				}
				else
				{
					c = top->next = _expr ();
					par (top, QU_AST_getitem);
					if (peek != QU_T_RSQUARE)
					{
						top->type = QU_AST_getslice;
						match (QU_T_COMMA);
						if (peek == QU_T_RSQUARE)
						{
							x = pnode (QU_AST_literal);
							x->thing = QuInt_1;
						}
						else
						{
							x = _expr ();
						}
						c->next = x;
					}
				}
				req (QU_T_RSQUARE, "] expected");
				continue;

			case QU_T_LPAREN:
				switch (peekn (2))
				{
					case QU_T_DOT:
					case QU_T_COLON:
						return top;
				}

				eat;
				skipindent;
				c = top;
				par (top, QU_AST_call);

				if (match (QU_T_CONCAT))
				{
					c->next = _multicat ();
					top->op = 1;
					if (top->child->type == QU_AST_extern)
						_call_check_extern (top);
					continue;
				}

				n = 0;
				if (!match (QU_T_RPAREN))
				{
					if (peek == QU_T_NAME && peekn (2) == QU_T_COLON)
					{
						n = 1;
						c->next = _dictarg ();
					}
					else
					{
						for (;;)
						{
							skipindent;
							if (match (QU_T_RPAREN))
								break;

							argmax (++n, "too many arguments");
							if (match (QU_T_MUL))
							{
								x = _expr ();
								if (x->type == QU_AST_literal)
									croak ("cant flatten literal value");
								par (x, QU_AST_flat);
							}
							else
							{
								x = _expr ();
							}
							rpl (c, x);
							match (QU_T_COMMA);
						}
					}
				}
				top->op = n;
				if (top->child->type == QU_AST_extern)
					_call_check_extern (top);
				continue;
		}
		break;
	}
	return top;
}



static qu_ast _factor ()
{
	qu_ast top = 0;

	switch (peek)
	{
		case QU_T_LPAREN:
		{
			/*
				(x + 10)  -> expr
				(x, y, z) -> list
				()        -> *** syntax error

				list (op: number of expressions)
				|
				<expr>+
			*/
			eat;
			if (match (QU_T_CONCAT))
			{
				top = _multicat ();
				break;
			}

			top = _expr ();
			if (peek != QU_T_RPAREN)
			{
				qu_ast c = top, x;
				int  k = 1;
				int  e = 0;

				if (CoInList)
					croak ("nested list is not allowed");
				match (QU_T_COMMA);
				CoInList = 1;
				par (top, QU_AST_list);
				while (!match (QU_T_RPAREN))
				{
					argmax (++k, "too many items in list");

					/* (a, b, *c) */
					if (match (QU_T_MUL))
					{
						if (e)
							croak (") expected");
						e = 1;
						identifier ("after *");
						x = pnode (QU_AST_eat);
						grab (x);
					}
					else
					{
						x = _expr ();
					}
					rpl (c, x);
					match (QU_T_COMMA);
				}
				CoInList = 0;
				top->op = k;
			}
			else
			{
				req (QU_T_RPAREN, ") expected");
			}
			break;
		}

		case QU_T_NAME:
			/*
				name (thing: String, op: 1=followed by call() 0=not)
			*/
			eat;
			top = pnode (QU_AST_name);
			grab (top);
			if (sameline && peek == QU_T_LPAREN)
				top->op = 1;
			break;


		/*
			literal (thing: object)
		*/

		case QU_T_true:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = QuTrue;
			break;

		case QU_T_false:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = QuFalse;
			break;

		case QU_T_nil:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = QuNil;
			break;

		case QU_T_infp:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = QuInfPos;
			break;

		case QU_T_infn:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = QuInfNeg;
			break;

		case QU_T_null:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = QuStreamNull;
			break;

		case QU_T_INT:
		case QU_T_FLOAT:
		case QU_T_BIG:
		case QU_T_RATIONAL:
		case QU_T_LONG:
		case QU_T_COMPLEX:
			eat;
			top = pnode (QU_AST_literal);
			grab (top);
			break;

		case QU_T_STRING:
		case QU_T_TEXT:
			eat;
			top = pnode (QU_AST_literal);
			grab (top);
			break;

		case QU_T_me:
			eat;
			top = pnode (QU_AST_me);
			break;

		case QU_T_self:
			eat;
			if (qu_module_is (CoMaster))
				croak ("use me in module level instead of self");
			if (qu_class_is (CoMaster))
				croak ("self is only applicable in sub");
			top = pnode (QU_AST_self);
			break;

		case QU_T_AT:
			eat;

			if (match (QU_T_self))
			{
				/* @self.foo => self.__attrs ("foo")
					to overcome the autocall method feature problem */

				top = pnode (QU_AST_ref);
				req (QU_T_DOT, ". expected after @self");
				identifier ("after .");
				top->thing = CoToken->thing;
				top->op = 2;
				break;
			}


			top = _symbol ();
			top->type = QU_AST_ref;
#if 1
			/*EXPERIMENTAL - currying */
			if (sameline && match (QU_T_LPAREN))
			{
				qu_ast l, c = 0, x;
				int    k = 0;

				l = rnode (QU_AST_list);
				while (!match (QU_T_RPAREN))
				{
					argmax (++k, "too many items in list");
					x = _expr ();
#if 0
					if (x->type != QU_AST_literal)
						croak ("can only use literals for curry arguments");
#endif
					adc (l, c, x);
					match (QU_T_COMMA);
				}
				l->next = top->child;
				l->op = k;
				top->child = l;
				/* cant have trailers here */
				if (sameline)
				{
					switch (peek)
						case QU_T_INCR:
						case QU_T_DECR:
						case QU_T_DOT:
						case QU_T_FIELD:
						case QU_T_ARROW:
						case QU_T_LSQUARE:
						case QU_T_LPAREN:
							croak ("cant have trailers here");
				}
				return top;
			}
#endif
			break;

		/*
			super (thing: this class or Array [this class, superclass/mixin name])

				and eventually:

				  getattr
					/  \
				super   ...
		*/
		case QU_T_super:
		{
			qu_array a = 0;

			eat;
			if (match (QU_T_LPAREN))
			{
				qu_string name;
				identifier ("after super(");
				name = CoToken->thing;
				req (QU_T_RPAREN, ") expected after super(identifier");
				{
					qu_any argv[2] = {QuNil, name};
					a = qu_array_new_argv (2, argv);
				}
			}

			if (peek != QU_T_DOT)
				croak (". expected after super");

			if (qu_sub_is (CoMaster))
			{
				qu_sub u = ((qu_sub) CoMaster)->root;

				if (qu_class_is (u->parent))
				{
					top = pnode (QU_AST_super);
					if (a)
					{
						top->thing = a;
						a->objs[0] = u->parent;
					}
					else
					{
						top->thing = u->parent;
					}
					break;
				}
			}
			croak ("super is only applicable in methods");
		}


		/*
			Array, Dict, Set
		*/
		case QU_T_LSQUARE:
			eat;
			top = _array ();
			break;

		case QU_T_LCURLY:
			eat;
			if (match (QU_T_BITOR))
			{
				top = _curlsub (1);
			}
			else if (match (QU_T_COLON))
			{
				top = _curlsub (0);
			}
			else
			{
				top = _set ();
			}
			break;

		case QU_T_const:
			eat;
			top = _constant ();
			break;


		/*
			statements allowed in expressions
		*/
		case QU_T_pass:
			eat;
			if (CoInAtomic)
				croak ("dont use pass in atomic");
			top = pnode (QU_AST_pass);
			break;

		case QU_T_redo:
			eat;
			top = _redo ();
			break;

		case QU_T_break:
			eat;
			top = _break ();
			break;

		case QU_T_continue:
			eat;
			top = _continue ();
			break;

		case QU_T_return:
			eat;
			top = _return ();
			break;

		case QU_T_yield:
			eat;
			top = _yield ();
			break;

		case QU_T__swap:
			eat;
			top = _swap ();
			break;

		case QU_T_throw:
			eat;
			top = _throw ();
			break;

		/*
			others
		*/
		case QU_T__frame:
			eat;
			if (!qu_sub_is (CoBlock) && !qu_flag_test (CoBlock, QU_SUB_FLAG_GENERATOR))
				croak ("__frame__ not in generator");
			top = pnode (QU_AST_frame);
			break;

		case QU_T__argc:
			eat;
			if (!qu_sub_is (CoMaster))
				croak ("__argc__ not in sub");
			top = pnode (QU_AST_argc);
			break;

		case QU_T__line:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = qu_int_new (CoToken->line);
			break;

		case QU_T__file:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = CoPath;
			break;

		case QU_T__here:
			eat;
			top = pnode (QU_AST_literal);
			top->thing = CoMaster->name;
			break;

		case QU_T_sub:
			eat;
			top = _anonsub ();
			break;

		case QU_T_lambda:
			eat;
			top = _lambda ();
			break;

		case QU_T_each:
			eat;
			top = _expr ();
			par (top, QU_AST_unop);
			top->op = (long) qu_sys_each;
			break;

		case QU_T_DOLLAR:
			/*
				$foo means self.foo
				$("f" _ "oo") means self.("foo")
				$self means self
			*/
			eat;
			if (peek == QU_T_self)
			{
				eat;
				top = pnode (QU_AST_self);
				break;
			}

			top = pnode (QU_AST_getattr);
			top->child = pnode (QU_AST_self);
			switch (peek)
			{
				case QU_T_LPAREN:
					eat;
					top->child->next = _expr ();
					req (QU_T_RPAREN, ") expected to close $(");
					break;

				case QU_T_NAME:
					eat;
					top->child->next = pnode (QU_AST_literal);
					top->child->next->thing = CoToken->thing;
					break;

				default:
					croak ("identifier or ( expected after $");
			}
			if (sameline && peek == QU_T_LPAREN)
				top->op = 1;
			break;

		case QU_T__extern:
		{
			/*
				__extern Gtk.gtk_window_set_title "-:Wsi" (...)

				EXTERN -> thing: [Module name, symbol name, spec]
			*/
			qu_string m, s, a;

			untrusted ("__extern is prohibited");
			eat;
			top = pnode (QU_AST_extern);
			switch (peek)
			{
				case QU_T_me:
					eat;
					top->thing = CoModule;
					break;

				default:
					req (QU_T_NAME, "identifier or me expected after __extern");
					grab (top);
			}
			m = top->thing;
			req (QU_T_DOT, ". expected");
			switch (peek)
			{
				case QU_T_NAME:
				case QU_T_STRING:
					eat;
					break;

				default:
					croak ("identifier or string expected after __extern");
			}
			grab (top);
			s = top->thing;
			match (QU_T_COMMA);
			req (QU_T_STRING, "argument string expected");
			grab (top);
			a = top->thing;
			top->thing = qu_array_new_ex (3, m, s, a);

			if (qu_sub_is (CoMaster))
				qu_flag_set (CoMaster, QU_SUB_FLAG_NOJIT);

			break;
		}

		case 0:
			croak ("syntax error");

		default:
			eat;
			croak ("unexpected token in expression");
    }
	return top;
}



/*
	redo     (op: number of blocks to break)
	break    (op: number of blocks to break)
	continue (op: number of blocks to break)
*/
static qu_ast _rbcr (int can, int type)
{
	qu_ast    top;
	qu_string l, name = 0;
	qu_any    b;
	int       i, n;

	n = peek;

	if (sameline && match (QU_T_NAME))
		name = CoToken->thing;

	for (i = CoBlockLevel, n = 1; i > 0; i--, n++)
	{
		b = CoBlockStack [i];
		if (ismaster (b))
			break;

		if ((long) b & can)
		{
			if (name == 0)
				goto okay;
			l = CoLabelStack [i];
			if (l && qu_string_EQ (l, name))
				goto okay;
		}
	}
	croak ("target not found");

  okay:
	top = pnode (type);
	top->op = n;
	return top;
}



static qu_ast _redo ()
{
	return _rbcr (CAN_REDO, QU_AST_redo);
}



static qu_ast _break ()
{
	return _rbcr (CAN_BREAK, QU_AST_break);
}



static qu_ast _continue ()
{
	return _rbcr (CAN_CONTINUE, QU_AST_continue);
}



/*
	return
	|
	[<expr>]
*/
static qu_ast _return ()
{
	qu_ast top;

	if (CoInLambda)
		croak ("cant use return in lambda");

	if (!qu_sub_is (CoMaster))
		croak ("return not in sub");

	top = pnode (QU_AST_return);
	if (sameline)
	{
		switch (peek)
		{
			case 0:
			case QU_T_SEMI:
			case QU_T_END:
			case QU_T_end:
			case QU_T_COLON:
			case QU_T_RCURLY:
			case QU_T_RPAREN:
			case QU_T_RSQUARE:
			case QU_T_if:
			case QU_T_unless:
			case QU_T_INDENT:
			case QU_T_DEDENT:
				break;

			default:
				top->child = _expr ();
		}
	}
	return top;
}



/*
	yield
	|
	<expr>+
*/
static qu_ast _yield ()
{
	qu_ast top, c = NULL;
	int    n = 0;

	if (CoInEnsure)
		croak ("cant use yield in ensure");

	if (!qu_sub_is (CoMaster))
		croak ("yield not in sub");

	if (qu_flag_test (CoMaster, QU_SUB_FLAG_ONCE))
		croak ("yield cant be mixed with __once");

	nobrk ("after yield");

	qu_flag_set (CoMaster, QU_SUB_FLAG_GENERATOR);
	top = pnode (QU_AST_yield);
	do
	{
		argmax (++n, "too many yields");
		adc (top, c, _expr ());

	} while (match (QU_T_COMMA));
	return top;
}



/*
	(_ a b c)

	             call (op:3)
	              / / \  \
	       literal  a  b  c
	__QuSys_syscat
*/
static qu_ast _multicat ()
{
	qu_ast top = pnode (QU_AST_call), c;
	int    n = 0;

	top->child = c = pnode (QU_AST_literal);
	c->thing = __QuSys_syscat;

	while (!match (QU_T_RPAREN))
	{
		skipindent;
		argmax (n, "too many arguments");
		++n;
		adc (top, c, _expr ());
		match (QU_T_COMMA);
	}
	top->op = n;
	return top;
}



/*
	swap
	|
	<expr>
*/
static qu_ast _swap ()
{
	qu_ast top;

	if (CoInEnsure)
		croak ("cant use __swap__ in ensure");

	if (!qu_sub_is (CoMaster))
		croak ("__swap__ not in sub");

	nobrk ("after __swap__");

	top = pnode (QU_AST_swap);
	top->child = _expr ();
	return top;
}



/*
	throw
	|
	[<expr> - [<expr>]]
*/
static qu_ast _throw ()
{
	qu_ast top;

	top = pnode (QU_AST_throw);
	if (sameline)
	{
		top->child = _expr ();
		if (match (QU_T_COMMA))
			top->child->next = _expr ();
	}
	return top;
}



/*
	symbol (thing: String)
	|
	[name (op: String)]
	|
	[name (op: String)]
*/
static qu_ast _symbol ()
{
	qu_ast top, c, g;

	identifier ("as symbol");
	top = pnode (QU_AST_symbol);
	grab (top);
	if (match (QU_T_DOT))
	{
		identifier ("after .");
		c = pnode (QU_AST_name);
		grab (c);
		top->child = c;
		if (match (QU_T_DOT))
		{
			g = pnode (QU_AST_name);
			identifier ("after .");
			grab (g);
			c->child = g;
			if (match (QU_T_DOT))
				croak ("another . is not possible");
		}
	}
	return top;
}



/*
	a = {for i in 1, 10: i}

	a = (sub __anon
			x = {}
			for i in 1, 10
				x <<< i + 10
			;;
			return x
		;;)()

	                     call (op:0)
	                       |
	                    anonsub
	                       |
	                    compound
	         ______________|______________
	        /              |              \
	    stmt              for            return
	     |                / \              |
	  assign    (name|list)  compound     name
	   /  \                     |
	name  <expr>            stmt | if
	                            m

	     m                  m
	    stmt                if
	     |                  |
	   binop push           <expr> - compound
	    / \                            |
	name  <expr>                     binop push
	                                  / \
	                              name   <expr>

	a = [for i in 4, 6: [for j in 1, 2: i + j]]

	a = (sub <0>
			<1> = []
			for i in 4, 6
				<1> <<< ((sub <2>
							<3> = []
							for j in 1, 2
								<3> <<< (i + j)
							;;
							return <3>
						;;)())
			;;
			return <1>
			;;)()

*/
static qu_ast _builder_stmt (qu_string name, qu_ast expr)
{
	qu_ast m, x, c, b;
	int    t = peek;

	if (t == QU_T_if || t == QU_T_unless)
	{
		/*
			expr if|unless (s)

			if                 if
			|                  |
			s - compound       not - cmopound
			       |           |         |
			     binop push    s       binop push
			      / \                    / \
			  name  <expr>           name   <expr>
		*/
		eat;
		x = _expr ();
		m = rnode (QU_AST_if);
		if (t == QU_T_unless)
		{
			m->child = rnode (QU_AST_unop);
			m->child->op = (long) qu_sys_not;
			m->child->child = x;
		}
		else
		{
			m->child = x;
		}
		m->child->next = c = rnode (QU_AST_compound);
		c->child = b = rnode (QU_AST_binop);
	}
	else
	{
		/*
			expr

			   stmt
			    |
			   binop push
		        / \
			name  <expr>
		*/
		m = rnode (QU_AST_stmt);
		m->child = b = rnode (QU_AST_binop);
	}
	b->op = (long) qu_sys_push;
	b->child = rnode (QU_AST_name);
	b->child->thing = name;
	b->child->next = expr;

	return m;
}



static qu_ast _builder_for (qu_ast top, int closer)
{
	qu_ast    a, c, s, f, x, sub, expr, m;
	qu_string v;
	qu_sub    u;

	if (CoInConst)
		croak ("cant use const with builders");

	v = anon ();
	sub = rnode (QU_AST_anonsub);
	sub->thing = u = qu_sub_new (CoMaster, v, 0, 0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL);
	addsymbol (v, (qu_symbol) u);
	enter (u);

	f = _for_head ();
	v = anon ();

	match (QU_T_COLON);

	enter (BLOCK_FOR);
	expr = _expr ();
	m = _builder_stmt (v, expr);

	leave (BLOCK_FOR);

	if (closer == QU_T_RCURLY)
		req (QU_T_RCURLY, "} expected to close Array or Dict builder");
	else
		req (QU_T_RSQUARE, "] expected to close Set builder");

	sub->child = c = rnode (QU_AST_compound);
	c->child = s = rnode (QU_AST_stmt);
	s->child = a = rnode (QU_AST_assign);
	s->next = f;
	f->next = x = rnode (QU_AST_return);
	x->child = rnode (QU_AST_name);
	x->child->thing = v;

	a->child = rnode (QU_AST_name);
	a->child->thing = v;
	a->child->next = top;

	f->child->next->next = x = rnode (QU_AST_compound);
	x->child = m;

	top = rnode (QU_AST_call);
	top->child = sub;

	leave (u);
	return top;
}



/*
	just the same except for the for/while node

	                     call (op:0)
	                       |
	                    anonsub
	                       |
	                    compound
	         ______________|______________
	        /              |              \
	    stmt             while           return
	     |                / \              |
	  assign         <expr>  compound     name
	   /  \                     |
	name  <expr>              stmt
	                            |
	                          binop push
	                           / \
	                       name  <expr>
*/
static qu_ast _builder_while (qu_ast top, int closer)
{
	qu_ast    a, c, s, f, x, sub, expr, m;
	qu_string v;
	qu_sub    u;

	if (CoInConst)
		croak ("cant use const with builders");

	v = anon ();
	sub = rnode (QU_AST_anonsub);
	sub->thing = u = qu_sub_new (CoMaster, v, 0, 0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_LOCAL);
	addsymbol (v, (qu_symbol) u);
	enter (u);

	f = rnode (QU_AST_while);
	f->child = _expr ();

	v = anon ();

	match (QU_T_COLON);
	expr = _expr ();

	m = _builder_stmt (v, expr);

	if (closer == QU_T_RCURLY)
		req (QU_T_RCURLY, "} expected to close Array or Dict builder");
	else
		req (QU_T_RSQUARE, "] expected to close Set builder");

	sub->child = c = rnode (QU_AST_compound);
	c->child = s = rnode (QU_AST_stmt);
	s->child = a = rnode (QU_AST_assign);
	s->next = f;
	f->next = x = rnode (QU_AST_return);
	x->child = rnode (QU_AST_name);
	x->child->thing = v;

	a->child = rnode (QU_AST_name);
	a->child->thing = v;
	a->child->next = top;

	f->child->next = x = rnode (QU_AST_compound);
	x->child = m;

	top = rnode (QU_AST_call);
	top->child = sub;

	leave (u);
	return top;
}



/*
	dict (thing:0 op:number of items)
	|
	(<expr> - <expr>)*

	array (thing:0 op:number of items)
	|
	<expr>*
*/
static qu_ast _array ()
{
	qu_ast top = pnode (QU_AST_array);
	qu_ast c   = NULL;
	int    n   = 0;
	qu_ast x;

	skipindent;
	switch (peek)
	{
		case QU_T_RSQUARE:
			eat;
			break;

		case QU_T_COLON:
			eat;
			top->type = QU_AST_dict;
			/* skipindent; */
			if (match (QU_T_RSQUARE))
				break;

			if (match (QU_T_for))
				return _builder_for (top, QU_T_RSQUARE);

			if (match (QU_T_while))
				return _builder_while (top, QU_T_RSQUARE);

			goto exprform;

		case QU_T_for:
			eat;
			return _builder_for (top, QU_T_RSQUARE);

		case QU_T_while:
			eat;
			return _builder_while (top, QU_T_RSQUARE);

		case QU_T_NAME:
			if (peekn (2) == QU_T_COLON)
			{
				top->type = QU_AST_dict;
				for (;;)
				{
					skipindent;
					if (match (QU_T_RSQUARE))
						break;

					argmax (++n, "too many items");

					x = pnode (QU_AST_literal);
					if (n == 1)
						adc (top, c, x);
					else
						rpl (c, x);

					identifier ("in this Dict declaration");
					grab  (c);
					req (QU_T_COLON, ": expected after Dict key item");

					skipindent;
					rpl (c, _expr ());
					n++;

					match (QU_T_COMMA);
				}
				break;
			}
			/*nobreak*/

		default:
		  exprform:
			for (;;)
			{
				skipindent;

				if (match (QU_T_RSQUARE))
					break;

				argmax (++n, "too many items");

				x = _expr ();

				if (n == 1)
				{
					adc (top, c, x);
					if (top->type == QU_AST_dict)
						; /* req (QU_T_COLON, ": expected after Dict key item"); */
					else if (peek == QU_T_COLON)
						top->type = QU_AST_dict;
				}
				else
					rpl (c, x);

				if (top->type == QU_AST_dict)
				{
					req (QU_T_COLON, ": expected after Dict key item");
					rpl  (c, _expr ());
					n++;
				}

				match (QU_T_COMMA);
			}
	}
	top->op = n;
	return top;
}



/*
	dict (thing: 0, op: umber of items)
	|
	(<expr> - <expr>)*
*/
static qu_ast _dictarg ()
{
	qu_ast top = pnode (QU_AST_arg);
	qu_ast c   = NULL;
	int    n   = 0;
	int    k   = (peek == QU_T_NAME);
	qu_ast x;

	for (;;)
	{
		skipindent;
		if (match (QU_T_RPAREN))
			break;

		if (k)
		{
			identifier ("as Dict key");
			x = pnode (QU_AST_literal);
			grab (x);
		}
		else
			x = _expr ();

		argmax (++n, "too many arguments");

		if (n == 1)
			adc (top, c, x);
		else
			rpl (c, x);

		req (QU_T_COLON, ": expected after argument name");
		skipindent;
		rpl (c, _expr ());
		n++;

		match (QU_T_COMMA);
	}
	top->op = n;
	return top;
}



/*
	set (thing: 0, op: umber of items)
	|
	<expr>*
*/
static qu_ast _set ()
{
	qu_ast top = pnode (QU_AST_set);
	qu_ast c   = NULL;
	int    n   = 0;
	qu_ast x;

	if (match (QU_T_for))
		return _builder_for (top, QU_T_RCURLY);

	if (match (QU_T_while))
		return _builder_while (top, QU_T_RCURLY);

	for (;;)
	{
		skipindent;
		if (match (QU_T_RCURLY))
			break;
		argmax (++n, "too many items");
		x = _expr ();
		if (n == 1)
			adc (top, c, x);
		else
			rpl (c, x);
		match (QU_T_COMMA);
	}

	top->op = n;
	return top;
}



/*
	array | dict | set (thing: = QuInt1)
	|
	<expr>*
*/
static qu_ast _constant ()
{
	qu_ast top = 0;

	CoInConst++;

	if (match (QU_T_LSQUARE))
		top = _array ();

	else if (match (QU_T_LCURLY))
		top = _set ();

	else
		croak ("Array, Dict or Set declaration expected after const");

	CoInConst--;

	top->thing = QuInt1;
	return top;
}



/*----------------------------------------------------------------------------
	compile
----------------------------------------------------------------------------*/

static int _compile_parse_one (qu_module m)
{
	CoDocDict = 0;
	CoUseIndent = 0;

	CoModule = m;
	CoPath   = m->name;
	CoSource = m->source;
	CoDepth  = CoBlockLevel = CoMasterLevel = 0;
	CoInList = CoInAtomic = CoInEnsure = CoInIf = CoInLambda = CoInCurlySub = CoInConst = 0;
	CoBlock  = CoBlockStack [0] = m;
	CoMaster = CoMasterStack [0] = (qu_symbol) m;
	CoLabelStack[0] = 0;

	qu_tokenizer_reset (CoTokenizer, m->literals, CoSource, CoDocumenting ? QU_TOKENIZER_FLAG_DOC : 0);

	if (qu_module_is_eval (m))
		qu_tokenizer_set_resident (CoTokenizer);
	else
		qu_tokenizer_unset_resident (CoTokenizer);

	if (setjmp (CoJmpbuf) == 0)
	{
		m->tree = rnode (QU_AST_module);
		_module ();
		qu_module_set_parsed (m);
		return 1;
	}
	return 0;
}



static int _compile_one (qu_module m)
{
	return qu_code_generate (m);
}



static int _compile_parse (qu_module m)
{
	if (_compile_parse_one (m) == 0)
		return 0;

	if (m->use)
	{
		qu_module u;
		long      b = 0, i = 0;

		while ((u = (qu_module) qu_set_walk (m->use, &b, &i)) != 0)
		{
			assert (qu_module_is (u));
			if (!qu_module_parsed (u))
			{
				qu_module_set_parsed (u);
				if (_compile_parse (u) == 0)
				{
					qu_module_unset_parsed (u);
					return 0;
				}
			}
		}
	}
	return 1;
}



static int _compile_do (qu_module m)
{
	if (m->use)
	{
		long   b = 0, i = 0;
		qu_module u;

		while ((u = (qu_module) qu_set_walk (m->use, &b, &i)) != 0)
		{
			assert (qu_module_is (u));
			if (!qu_module_compiled (u))
			{
				qu_module_set_compiled (u);
				if (_compile_do (u) == 0)
				{
					qu_module_unset_compiled (u);
					return 0;
				}
			}
		}
	}
	return _compile_one (m);
}



static void _compile_failed (qu_module m)
{
	if (!qu_module_haserror (m))
	{
		qu_module_set_haserror (m);
		qu_module_failed (m);

		if (m->use)
		{
			qu_module u;
			long      b = 0, i = 0;

			while ((u = (qu_module) qu_set_walk (m->use, &b, &i)) != 0)
			{
				assert (qu_module_is (u));
				if (!qu_module_is_ready (u))
					_compile_failed (u);
			}
		}
	}
}



static void _compile_jit (qu_attr_s *a)
{
	long      iter = 0;
	qu_string ignore;
	qu_any    x;

	assert (QU_APP_OPTIMIZE_CODE);

	while ((x = qu_attr_walk (a, &ignore, &iter)) != NULL)
	{
		/* don't do it for
			- qu_sub_is_extern   : it's C code
			- qu_sub_is_jit      : already done
			- qu_sub_is_abstract : it's useless
			- qu_sub_is_once     : it's a waste

			__destroy -> might be tricky so don't do it either
		*/
		if (qu_sub_is (x)
			&& ! (qu_sub_is_abstract (x)
			        || qu_flag_test (x, QU_SUB_FLAG_NOJIT)
					|| qu_sub_is_jit (x)
					|| qu_sub_is_extern (x)
					|| qu_sub_is_once (x)))
		{
			if (((qu_sub) x)->name != QU_NAME__destroy)
			{
				assert (strcmp (((qu_sub) x)->name->str, "__destroy") != 0);
				if (qu_code_jit_generate (x))
				{
					if (((qu_sub) x)->attrs)
						_compile_jit (((qu_sub) x)->attrs);
				}
			}
		}
		else if (qu_class_is (x))
		{
			_compile_jit (((qu_class) x)->attrs);
		}
	}
}



static void _compile_ready (qu_module m)
{
	if (!qu_module_is_ready (m))
	{
		assert (qu_module_compiled (m));
		qu_module_set_ready (m);

		if (QU_APP_OPTIMIZE_CODE
			&& qu_code_jit_generate
			&& qu_module_is_trusted (m)
			&& !qu_module_is_doc (m))
		{
			_compile_jit (m->attrs);
		}

		if (!QU_APP_KEEP_TREE)
		{
			m->tree = 0;
		}

		if (m->use)
		{
			qu_module u;
			long      b = 0, i = 0;

			while ((u = (qu_module) qu_set_walk (m->use, &b, &i)) != 0)
			{
				assert (qu_module_is (u));
				_compile_ready (u);
			}
		}
	}
}



/*
	!!! must have lock
*/
static qu_module _compile_module (qu_module m)
{
	if (_compile_parse (m))
	{
		if (_compile_do (m))
		{
			_compile_ready (m);
			return m;
		}
	}
	_compile_failed (m);
	return 0;
}



/*----------------------------------------------------------------------------
	lock
----------------------------------------------------------------------------*/

int volatile __QuCompileIsCompiling = 1;

static qu_thread_mutex_t LockMutex = QU_THREAD_MUTEX_INITIALIZER;
static qu_thread_t       LockOwner = 0; /* assuming there is never a thread-id 0 */


static int _compile_lock ()
{
	qu_thread_t tid = qu_thread_id ();

  __qu_blocking_begin;
	qu_thread_mutex_lock (&LockMutex);
	if (tid == LockOwner)
	{
		qu_thread_mutex_unlock (&LockMutex);
	  __qu_blocking_end;
		QU_THROWZ (QuECompile, "currently compiling");
	}
	LockOwner = tid;
  __qu_blocking_end;

	__QuCompileIsCompiling = 1;
	__QuGcDisable++;
	return 1;
}


static void _compile_unlock ()
{
	LockOwner = 0;
	__QuCompileIsCompiling = 0;
	__QuGcDisable--;
	qu_thread_mutex_unlock (&LockMutex);
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

static int OldNumberContext;

#define SAVE_NUMBER_CONTEXT \
  __qu_number_context_begin (OldNumberContext, \
		  QU_NUMBER_CONTEXT_ALLOW_INF \
		| QU_NUMBER_CONTEXT_ALLOW_TINY \
		| QU_NUMBER_CONTEXT_ALLOW_HUGE \
		| QU_NUMBER_CONTEXT_ALLOW_ZERODIV \
		| QU_NUMBER_CONTEXT_ALLOW_COMPLEX)

#define RESTORE_NUMBER_CONTEXT \
  __qu_number_context_end (OldNumberContext)


qu_module qu_compile_main (qu_string s, int ispath)
{
	qu_module m = qu_module_new_main (s, ispath);

	QU_ENSURE (m);

	QU_ENSURE (_compile_lock ());
	{
		CoDocumenting = QU_APP_WITH_DOC;
		if (CoTokenizer == 0)
			CoTokenizer = qu_tokenizer_new_root (m->literals, m->source, 0);
		SAVE_NUMBER_CONTEXT;
		m = _compile_module (m);
		RESTORE_NUMBER_CONTEXT;
	}
	_compile_unlock ();

	QU_ENSURE (m);
	if (!QU_APP_KEEP_TREE)
		m->tree = 0;
	return m;
}



/*
	!!! must only be called by Sys.eval
*/
qu_module qu_compile_eval (qu_module m, int docstring)
{
	QU_ENSURE (_compile_lock ());
	{
		CoDocumenting = (docstring != 0);
		SAVE_NUMBER_CONTEXT;
		m = _compile_module (m);
		RESTORE_NUMBER_CONTEXT;
	}
	_compile_unlock ();

	QU_ENSURE (m);
	if (!QU_APP_KEEP_TREE)
		m->tree = 0;
	return m;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_compile__init ()
{
	return;
}
