/* Complex.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Complex is Number __final
	This class provide support for complex numbers.

	The class hierarchy:
	@tree
	.-{Number}
	.	Complex

	@p
	A complex number is defined as:
	@def
		Complex (z) = (z.real + z.imag * i)

	  # where i is mathematically defined as:
		      __
		i = \/-1

	@p
	Qu does not have an Imaginary class. An Imaginary number is simply
	a Complex which real part is zero.

	Both the real and imaginary parts may be any -{Real} numbers, i.e.
	-{Int}, -{Long}, -{Float}, -{Big} or -{Rational}, including infinities
	infp and infn.

	To create a Complex number you use the #{new} method or the special
	I{j} syntax.

	@warn
	There must be no space between the imaginary number and the C{"j"} character.

	@xmp
		c = Complex (1, 2.3)
		d = 1 + 2.3j
		println (c)
		println (d)
		println (c == d)
		println (c is Number)
		println (c is Complex)

	@out
		(1 + 2.3j)
		(1 + 2.3j)
		true
		true
		true

	@p
	The global -{Number.context} lets you to specify whether or not Complex
	results are allowed and whether a -{Real} number should be returned when
	the imaginary part of a Complex result is zero.

	@xmp
		a = 1 + 1j
		b = 2 - 1j

		Number.context |= 'C'
		println (a + b)

		Number.context |= 'CR'
		println (a + b)

		Number.context -= 'CR'
		println (a + b)

	@out
		(3 + 0j)
		3
		[1024] __main__ (EComplex) result is a Complex

	@p
	When you perform an operation that mixes a Complex number with a Real number,
	Qu regards a Real number as a special case of a Complex number with the
	imaginary part being zero.

	@xmp
		a = 1 + 2j
		b = 2
		println (a + b)

	@out
		(3 + 2j)
*/
qu_class QuComplexClass;

qu_complex QuComplex00;
qu_complex QuComplex01;
qu_complex QuComplex0_1;
qu_complex QuComplex10;
qu_complex QuComplex11;
qu_complex QuComplex1_1;
qu_complex QuComplex_10;
qu_complex QuComplex_11;
qu_complex QuComplex_1_1;

qu_complex QuComplexBig00;
qu_complex QuComplexBig01;
qu_complex QuComplexBig0_1;
qu_complex QuComplexBig10;
qu_complex QuComplexBig11;
qu_complex QuComplexBig1_1;
qu_complex QuComplexBig_10;
qu_complex QuComplexBig_11;
qu_complex QuComplexBig_1_1;

qu_complex QuComplexInfp;
qu_complex QuComplexInfn;



/*
	always use this to return a complex number
*/
qu_number __qu_complex__ (qu_real r, qu_real i)
{
	if (QU_NUMBER_REALIFY_COMPLEX && qu_real_iszero (i))
		return (qu_number) r;
	QU_REQUIRE (QuEComplex, QU_NUMBER_ALLOW_COMPLEX);
	return (qu_number) qu_complex_newr (r, i);
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (real: Real, imag: Real = 0) -> Complex
	Create a new Complex number by specifying the real part and optionally
	the imaginary part.

	Both the real and imaginary parts may be any -{Real} numbers, including
	infp ir infn provided that you allow infinities to be
	returned via -{Number.context}.

	If called with one argument and v{real} is not a -{Number} then this
	method will call C{real.__as (Complex)} ensuring that the returned
	value is a Complex.

	@exc
	.EArgValue

	@xmp
		x = Complex (1, 2)
		println (x)

		class Foo
			sub __as (what)
				switch what
					case Complex: return 3 + 4j
				;;
				throw EImplement
			;;
		;;

		x = Foo ()
		println (Complex (x))

		Number.context |= 'I'
		x = Complex (1, infn)
		println (x)

	@out
		(1 + 2j)
		(3 + 4j)
		(1 + infn.j)
*/
static qu_complex qu__complex_new (qu_class self, int argc, qu_any *argv)
{
	/* TODO - from qu_string/qu_buffer */

	qu_any r, i;

	QU_ARG_VCOUNT (1, 2);
	if (argc == 1)
	{
		r = argv[0];
		i = QuInt0;

		if (qu_real_is (r))
		{
			return qu_complex_newr (r, i);
		}
		else
		{
			return qu_object_as (r, QuComplexClass);
		}
	}
	else
	{
		QU_ARG_VTYPE (0, r, qu_real);
		QU_ARG_VTYPE (1, i, qu_real);
		return qu_complex_newr (r, i);
	}
}



qu_complex qu_complex_new (qu_real r, qu_real i)
{
	register qu_complex z;

	assert (r);
	assert (i);
	assert (qu_real_is (r));
	assert (qu_real_is (i));

	z = qu_gc_new (QuComplexClass);
	z->real = r;
	z->imag = i;
	return z;
}



qu_complex qu_complex_newr (qu_real r, qu_real i)
{
	if (!QU_NUMBER_ALLOW_INF)
		QU_REQUIRE (QuEInf, !qu_inf_is (r) && !qu_inf_is (i));
	return qu_complex_new (r, i);
}



qu_complex qu_complex_new1 (qu_real r)
{
	register qu_complex z;

	assert (r);
	assert (qu_real_is (r));

	if (!QU_NUMBER_ALLOW_INF)
		QU_REQUIRE (QuEInf, !qu_inf_is (r));

	z = qu_gc_new (QuComplexClass);
	z->real = r;
	z->imag = (qu_real) QuInt0;
	return z;
}



qu_complex qu_complex_new_number (qu_number r)
{
	QU_ARG_TYPE (r, qu_real);
	return qu_complex_new1 ((qu_real) r);
}



static qu_complex qu__complex_static_int (long r, long i)
{
	register qu_complex f;

	f = qu_gc_new_static (QuComplexClass);
	f->real = (qu_real) qu_int_new (r);
	f->imag = (qu_real) qu_int_new (i);
	return f;
}



static qu_complex qu__complex_static_big (qu_big r, qu_big i)
{
	register qu_complex f;

	f = qu_gc_new_static (QuComplexClass);
	f->real = (qu_real) r;
	f->imag = (qu_real) i;
	return f;
}



static qu_complex qu__complex_static_inf (qu_inf r)
{
	register qu_complex f;

	f = qu_gc_new_static (QuComplexClass);
	f->real = (qu_real) r;
	f->imag = (qu_real) QuInt0;
	return f;
}



qu_complex qu_complex_newli (long v)
{
	QU_REQUIREZ (QuEValue, QU_INT_MIN <= v && v <= QU_INT_MAX, "Complex (long) ouf of Int Range");
	return qu_complex_new1 ((qu_real) qu_int_new (v));
}



qu_complex qu_complex_to_static (qu_complex self)
{
	qu_complex r = qu_gc_new_static (QuComplexClass);

	switch (qu_classid (self->real))
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Inf:
			r->real = self->real;
			break;

		case QU_CLASS_ID_Long:
			r->real = (qu_real) qu_long_to_static ((qu_long) self->real);
			break;

		case QU_CLASS_ID_Float:
			r->real = (qu_real) qu_float_to_static ((qu_float) self->real);
			break;

		case QU_CLASS_ID_Big:
			r->real = (qu_real) qu_big_to_static ((qu_big) self->real);
			break;

		case QU_CLASS_ID_Rational:
			r->real = (qu_real) qu_rational_to_static ((qu_rational) self->real);
			break;

		default:
			QU_FATAL ("real not a Real");
	}

	switch (qu_classid (self->imag))
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Inf:
			r->imag = self->imag;
			break;

		case QU_CLASS_ID_Long:
			r->imag = (qu_real) qu_long_to_static ((qu_long) self->imag);
			break;

		case QU_CLASS_ID_Float:
			r->imag = (qu_real) qu_float_to_static ((qu_float) self->imag);
			break;

		case QU_CLASS_ID_Big:
			r->imag = (qu_real) qu_big_to_static ((qu_big) self->imag);
			break;

		case QU_CLASS_ID_Rational:
			r->imag = (qu_real) qu_rational_to_static ((qu_rational) self->imag);
			break;

		default:
			QU_FATAL ("imag not a Real");
	}

	return r;
}



/*-----------------------------------------------------------------------------
	specific class methods
-----------------------------------------------------------------------------*/

/*) sub class ret (r:Real, i:Real) -> Number
	If the global -{Number.context} does not specify ALLOW_COMPLEX then this
	method throws EComplex.

	If the context does not specify ALLOW_INF and either v{r} or v{i} is an
	infinity then throws EInf.

	Otherwise if the context specifies REALIFY_COMPLEX and v{i} is zero then
	returns v{r}. A Complex otherwise.

	@exc
	.EComplex
	.EArgValue
	.EInf
*/
static qu_number qu__complex_ret (qu_class self, qu_real r, qu_real i)
{
	QU_ARG_TYPE (r, qu_real);
	QU_ARG_TYPE (i, qu_real);
	return __qu_complex__ (r, i);
}


qu_number qu_complex_return (qu_any r, qu_any i)
{
	return qu__complex_ret (QuComplexClass, r, i);
}




/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

static void qu__complex__mark (qu_complex self)
{
	qu_gc_mark (self->real);
	qu_gc_mark (self->imag);
}



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_complex_iszero (qu_complex x)
{
	return qu_real_iszero (x->real) && qu_real_iszero (x->imag);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x as y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following classes:
	@ops
	.| -{Complex} -{Number} | Return self.
	.| -{Boolean}           | Delegate to the #{__true} method.
	.| -{String}            | Delegate to the #{__string} method.
	.| -{Real}              | Return the real part.
	.| -{Number}            | Delegate to the real part object.

	@exc
	.EImplement v{x} none of the above classes
	.EArgValue

	@xmp
		z = 7 + 3j
		println (z as Real)

	@out
		7
*/
static qu_any qu__complex__as (qu_complex self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:      return qu_complex_as_int      (self);
		case QU_CLASS_ID_Long:     return qu_complex_as_long     (self);
		case QU_CLASS_ID_Rational: return qu_complex_as_rational (self);
		case QU_CLASS_ID_Float:    return qu_complex_as_float    (self);
		case QU_CLASS_ID_Big:      return qu_complex_as_big      (self);
		case QU_CLASS_ID_Complex:
		case QU_CLASS_ID_Number:   return self;
		case QU_CLASS_ID_Decimal:  return qu_complex_as_decimal  (self);
		case QU_CLASS_ID_Real:     return self->real;
		case QU_CLASS_ID_Boolean:  return qu_complex_as_boolean  (self);
		case QU_CLASS_ID_String:   return qu_complex_as_string   (self);
	}
	QU_ARG_THROW;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x <=> y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	If v{x} is a -{Number} then comparisson is made by converting v{x} into a
	Complex, otherwise C{x.__cmp (self)} is called and the result negated.

	@noexc

	@p
	This method returns -1 if self < v{x}, 0 if self == v{x}, 1 otherwise.

	@xmp
		a = 1 + 2j
		b = 1 + 3j
		println (a <=> b)

	@out
		-1
*/
static qu_int qu__complex__cmp (qu_complex self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_complex_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_complex_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_complex_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_complex_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_complex_cmp_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_complex_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_complex_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}

#define COMPLEX_CMP1(ytype) \
int qu_complex_cmp_##ytype (qu_complex x, qu_##ytype y) \
{ \
	int r = qu_real_cmp_##ytype (x->real, y); \
	return r ? r : qu_real_sign (x->imag); \
}


COMPLEX_CMP1 (int)
COMPLEX_CMP1 (long)
COMPLEX_CMP1 (rational)
COMPLEX_CMP1 (float)
COMPLEX_CMP1 (big)


int qu_complex_cmp_complex (qu_complex x, qu_complex y)
{
	int r = qu_real_cmp_real (x->real, y->real);
	return r ? r : qu_real_cmp_real (x->imag, y->imag);
}



int qu_complex_cmp_inf (qu_complex x, qu_inf y)
{
	int r = qu_inf_cmp_real (y, x->real);
	return r ? -r : qu_real_sign (x->imag);
}



QU_NUMBER_BINOP_integer (qu_complex, cmp, int)
QU_NUMBER_BINOP_decimal (qu_complex, cmp, int)
QU_NUMBER_BINOP_real    (qu_complex, cmp, int)
QU_NUMBER_BINOP_number  (qu_complex, cmp, int)



#define COMPLEX_CMPABS1(ytype) \
int qu_complex_cmpabs_##ytype (qu_complex x, qu_##ytype y) \
{ \
	int r = qu_real_cmpabs_##ytype (x->real, y); \
	return r ? r : (qu_real_iszero (x->imag) ? 0 : 1); \
}


COMPLEX_CMPABS1 (int);
COMPLEX_CMPABS1 (long);
COMPLEX_CMPABS1 (rational);
COMPLEX_CMPABS1 (float);
COMPLEX_CMPABS1 (big);


int qu_complex_cmpabs_complex (qu_complex x, qu_complex y)
{
	int r = qu_real_cmpabs_real (x->real, y->real);
	return r ? r : qu_real_cmpabs_real (x->imag, y->imag);
}



int qu_complex_cmpabs_inf (qu_complex x, qu_inf y)
{
	int r = qu_inf_cmpabs_real (y, x->real);
	return r ? -r : 1;
}



QU_NUMBER_BINOP_integer (qu_complex, cmpabs, int)
QU_NUMBER_BINOP_decimal (qu_complex, cmpabs, int)
QU_NUMBER_BINOP_real    (qu_complex, cmpabs, int)
QU_NUMBER_BINOP_number  (qu_complex, cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Complex
	Return a decremented version of self. That is, subtract 1 from the
	real and imaginary part.

	This method ignores REALIFY_COMPLEX and ALLOW_COMPLEX,
	but ALLOW_INF is respected thus if the result contains an infinity
	then EInf is thrown unless the global -{Number.context} specifies ALLOW_INF.

	@exc
	.EHuge
	.EInf

	@xmp
		a = 2 + 3j
		println (--a)

	@out
		(1 + 2j)
*/
qu_complex qu__complex__decr (qu_complex self)
{
	qu_real r, i;

	r = qu_real_decr (self->real); QU_ENSURE (r);
	i = qu_real_decr (self->imag); QU_ENSURE (i);
	return qu_complex_new (r, i);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x / y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number
	Return the result of dividing self with v{x}.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{\} operator).

	Refer to -{Number.__div} for more information.

	@p
	When dividing a Complex number with another then the other is converted
	into a Complex.
	@def
		x * y = z

	  # where:

		z.real = (x.real * y.real + x.imag * y.imag) / p
		z.imag = (x.imag * y.real - x.real * y.imag) / p

		p = y.real * y.real + y.imag * y.imag

	@p
	The result depends on whether the -{Number.context} specifies REALIFY_COMPLEX.
	If z.imag is zero and REALIFY_COMPLEX is specified then a -{Real} is returned,
	a Complex otherwise.

	@exc
	.EDiv
	.EComplex
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 2 + 4j
		b = 1 + 2j
		println (a / b)

	@out
		(2 + 0j)
*/
static qu_number qu__complex__div (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



static qu_number qu__complex_divri (qu_real xr, qu_real xi, qu_real yr, qu_real yi)
{
	qu_real d, r, i, t;

	d = qu_real_mul_real  (yr, yr);  QU_ENSURE (d); assert (qu_real_is (d));
	t = qu_real_mul_real  (yi, yi);  QU_ENSURE (t); assert (qu_real_is (t));
	d = qu_real_plus_real (d, t);    QU_ENSURE (d); assert (qu_real_is (d));

	r = qu_real_mul_real  (xr, yr);  QU_ENSURE (r); assert (qu_real_is (r));
	t = qu_real_mul_real  (xi, yi);  QU_ENSURE (t); assert (qu_real_is (t));
	r = qu_real_plus_real (r, t);    QU_ENSURE (r); assert (qu_real_is (r));
	r = qu_real_div_real  (r, d);    QU_ENSURE (r); assert (qu_real_is (r));

	i = qu_real_mul_real   (xi, yr); QU_ENSURE (i); assert (qu_real_is (i));
	t = qu_real_mul_real   (xr, yi); QU_ENSURE (t); assert (qu_real_is (t));
	i = qu_real_minus_real (i, t);   QU_ENSURE (i); assert (qu_real_is (i));
	i = qu_real_div_real   (i, d);   QU_ENSURE (i); assert (qu_real_is (i));

	return __qu_complex__ (r, i);
}



#define COMPLEX_DIV1(ytype) \
qu_number qu_complex_div_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu__complex_divri (x->real, x->imag, (qu_real) y, (qu_real) QuInt0); \
}


COMPLEX_DIV1 (int)
COMPLEX_DIV1 (long)
COMPLEX_DIV1 (rational)
COMPLEX_DIV1 (float)
COMPLEX_DIV1 (big)



qu_number qu_complex_div_complex (qu_complex x, qu_complex y)
{
	return qu__complex_divri (x->real, x->imag, y->real, y->imag);
}



qu_number qu_complex_div_inf (qu_complex x, qu_inf y)
{
	return qu__complex_divri (x->real, x->imag, (qu_real) y, (qu_real) QuInt0);
}



QU_NUMBER_BINOP_integer (qu_complex, div, qu_number)
QU_NUMBER_BINOP_decimal (qu_complex, div, qu_number)
QU_NUMBER_BINOP_real    (qu_complex, div, qu_number)
QU_NUMBER_BINOP_number  (qu_complex, div, qu_number)



/*
	y \ x
*/
#define COMPLEX_RDIV1(ytype) \
qu_number qu_complex_rdiv_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu__complex_divri ((qu_real) y, (qu_real) QuInt0, x->real, x->imag); \
}


COMPLEX_RDIV1 (int)
COMPLEX_RDIV1 (long)
COMPLEX_RDIV1 (rational)
COMPLEX_RDIV1 (float)
COMPLEX_RDIV1 (big)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x == y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if v{x} is the same number, false otherwise.

	If v{x} is a -{Real} then equality can only be satisfied when self.imag
	is zero. If v{x} is a Complex then equality can only be satisfied if
	both real and imaginary parts are equal.

	@noexc

	@xmp
		a = 2 + 0j
		b = 2
		println (a == b)

	@out
		true
*/
static qu_boolean qu__complex__eq (qu_complex self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_complex_eq_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_complex_eq_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_complex_eq_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_complex_eq_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_complex_eq_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_complex_eq_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_complex_eq_inf      (self, (qu_inf     ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



#define COMPLEX_EQ1(ytype) \
int qu_complex_eq_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu_real_iszero (x->imag) && qu_real_eq_##ytype (x->real, y); \
}


COMPLEX_EQ1 (int)
COMPLEX_EQ1 (long)
COMPLEX_EQ1 (rational)
COMPLEX_EQ1 (float)
COMPLEX_EQ1 (big)



int qu_complex_eq_complex (qu_complex x, qu_complex y)
{
	return qu_real_eq_real (x->real, y->real) && qu_real_eq_real (x->imag, y->imag);
}



int qu_complex_eq_inf (qu_complex x, qu_inf y)
{
	return qu_real_iszero (x->imag) && qu_inf_eq_real (y, x->real);
}



QU_NUMBER_BINOP_integer (qu_complex, eq, int)
QU_NUMBER_BINOP_decimal (qu_complex, eq, int)
QU_NUMBER_BINOP_real    (qu_complex, eq, int)
QU_NUMBER_BINOP_number  (qu_complex, eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this Complex.

	The format applies to both the real and imaginary part except that
	the width is not passed. See the class documentation for each part.

	@xmp
		x = 1.2 - 3j
		printfln ('|%20.4f|', x)

	@out
		|    1.1999 - 3.0000j|

	@note
	If you want specific formats for the real and imaginary part then you
	will have to do something like:

	@xmp
		printfln ("%.3f %2d", x.real, x.imag)
*/
static qu_string qu__complex__format (qu_complex self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_complex_format (self, &f);
}



qu_string qu_complex_format (qu_complex x, qu_format_s *f)
{
	switch (f->mode)
	{
		case 'y': return qu_format_yesno (f, x);
		case 'p': return qu_format_ptr (f, x);
	}
	/*default*/
	{
		int       w;
		qu_string s, r, i;
		char      c = qu_real_isneg (x->imag) ? '-' : '+';
		qu_string g = qu_string_byte (c);

		w = f->width;
		f->width = 0;
		r = qu_real_format (x->real, f); QU_ENSURE (r);
		f->sign = -1;
		i = qu_real_format (x->imag, f); QU_ENSURE (i);
		f->width = w;
		s = qu_string_cat (r, qu_string_byte (' '), g, qu_string_byte (' '), i, qu_string_byte ('j'), NULL);
		return qu_format_align (f, s);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> Int
	Return the hash value. Only the real part is taken into account. Thus
	this is the same as -{Float.__hash} of the real part.

	@noexc

	@xmp
		a = 2 + 3j
		b = 2.0
		c = 2
		println (a.__hash)
		println (b.__hash)
		println (c.__hash)

	@out
		2
		2
		2
*/
static qu_int qu__complex__hash (qu_complex self)
{
	return qu_int_newl (qu_object_hash (self->real));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x \ y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer
	Divide self with v{x} in an I{integer context}.

	This method simply returns C{self.real \ x}.
	The imaginary part is ignored.

	@warn
	There is a big difference between the #{__div} method (the C{/} operator)
	and the __idiv method (the C{\} operator).

	Refer to -{Number.__idiv} for more information.

	@exc
	.EIdiv
	.EBrain

	@xmp
		a = 3 + 7j
		println (a \ 2)

	@out
		1
*/
static qu_integer qu__complex__idiv (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_idiv_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_idiv_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_idiv_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_idiv_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_idiv_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_idiv_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_idiv_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEIdiv);
}



#define COMPLEX_IDIV1(ytype) \
qu_integer qu_complex_idiv_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu_real_idiv_real (x->real, (qu_real) y); \
}


COMPLEX_IDIV1 (int)
COMPLEX_IDIV1 (long)
COMPLEX_IDIV1 (rational)
COMPLEX_IDIV1 (float)
COMPLEX_IDIV1 (big)



qu_integer qu_complex_idiv_complex (qu_complex x, qu_complex y)
{
	return qu_real_idiv_real (x->real, y->real);
}



qu_integer qu_complex_idiv_inf (qu_complex x, qu_inf y)
{
	return qu_real_idiv_inf (x->real, y);
}



QU_NUMBER_BINOP_integer (qu_complex, idiv, qu_integer)
QU_NUMBER_BINOP_decimal (qu_complex, idiv, qu_integer)
QU_NUMBER_BINOP_real    (qu_complex, idiv, qu_integer)
QU_NUMBER_BINOP_number  (qu_complex, idiv, qu_integer)



#define COMPLEX_RIDIV1(ytype) \
qu_integer qu_complex_ridiv_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu_real_idiv_real ((qu_real) y, x->real); \
}


COMPLEX_RIDIV1 (int)
COMPLEX_RIDIV1 (long)
COMPLEX_RIDIV1 (rational)
COMPLEX_RIDIV1 (float)
COMPLEX_RIDIV1 (big)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Complex
	Return an incremented version of self. That is, add 1 to the real
	and imaginary part.

	This method ignores REALIFY_COMPLEX and ALLOW_COMPLEX,
	but ALLOW_INF is respected thus if the result contains an infinity
	then EInf is thrown unless the global -{Number.context} specifies ALLOW_INF.

	@exc
	.EHuge
	.EInf

	@xmp
		a = 2 + 3j
		println (++a)

	@out
		(3 + 4j)
*/
qu_complex qu__complex__incr (qu_complex self)
{
	qu_real r, i;

	r = qu_real_incr (self->real); QU_ENSURE (r);
	i = qu_real_incr (self->imag); QU_ENSURE (i);
	return qu_complex_new (r, i);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x: Any) -> Boolean
	If v{x} is a -{Number} and have the same absolute numerical value then
	this method returns true, false otherwise.

	@noexc

	@xmp
		a = 2 + 0j
		b = -2
		println (a === b)

	@out
		true
*/
static qu_boolean qu__complex__like (qu_complex self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_complex_like_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_complex_like_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_complex_like_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_complex_like_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_complex_like_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_complex_like_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_complex_like_inf      (self, (qu_inf     ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



#define COMPLEX_LIKE1(ytype) \
int qu_complex_like_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu_real_iszero (x->imag) && qu_real_like_##ytype (x->real, y); \
}


COMPLEX_LIKE1 (int)
COMPLEX_LIKE1 (long)
COMPLEX_LIKE1 (rational)
COMPLEX_LIKE1 (float)
COMPLEX_LIKE1 (big)



int qu_complex_like_complex (qu_complex x, qu_complex y)
{
	return qu_real_like_real (x->real, y->real) && qu_real_like_real (x->imag, y->imag);
}



int qu_complex_like_inf (qu_complex x, qu_inf y)
{
	return qu_real_iszero (x->imag) && qu_inf_like_real (y, x->real);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Return the result of subtracting self with v{x}.

	When subtracting a Complex number with another then the other is converted
	into a Complex.
	@def
		x - y = z

	  # where:

		z.real = x.real - y.real
		z.imag = x.imag - y.imag

	@p
	The result depends on whether the -{Number.context} specifies REALIFY_COMPLEX.
	If z.imag is zero and REALIFY_COMPLEX is specified then a -{Real} is returned,
	a Complex otherwise.

	@exc
	.EMinus
	.EComplex
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 1 + 3j
		b = 2 - 4j
		println (a - b)

	@out
		(-1 + 7j)
*/
static qu_number qu__complex__minus (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



#define COMPLEX_MINUS1(ytype) \
qu_number qu_complex_minus_##ytype (qu_complex x, qu_##ytype y) \
{ \
	qu_real r = qu_real_minus_##ytype (x->real, y); \
	QU_ENSURE (r); \
	return __qu_complex__ (r, x->imag); \
}


COMPLEX_MINUS1 (int);
COMPLEX_MINUS1 (long);
COMPLEX_MINUS1 (rational);
COMPLEX_MINUS1 (float);
COMPLEX_MINUS1 (big);



qu_number qu_complex_minus_complex (qu_complex x, qu_complex y)
{
	qu_real r, i;
	r = qu_real_minus_real (x->real, y->real); QU_ENSURE (r); assert (qu_real_is (r));
	i = qu_real_minus_real (x->imag, y->imag); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (r, i);
}



qu_number qu_complex_minus_inf (qu_complex x, qu_inf y)
{
	qu_real r = qu_real_minus_inf (x->real, y);
	QU_ENSURE (r);
	return __qu_complex__ (r, x->imag);
}



QU_NUMBER_BINOP_integer (qu_complex, minus, qu_number)
QU_NUMBER_BINOP_decimal (qu_complex, minus, qu_number)
QU_NUMBER_BINOP_real    (qu_complex, minus, qu_number)
QU_NUMBER_BINOP_number  (qu_complex, minus, qu_number)



/*
	y - x
*/

#define COMPLEX_RMINUS1(ytype) \
qu_number qu_complex_rminus_##ytype (qu_complex x, qu_##ytype y) \
{ \
	qu_real r, i; \
	r = qu_##ytype##_minus_real (y, x->real); QU_ENSURE (r); \
	i = qu_real_neg (x->imag); QU_ENSURE (i); \
	return __qu_complex__ (r, i); \
}


COMPLEX_RMINUS1 (int);
COMPLEX_RMINUS1 (long);
COMPLEX_RMINUS1 (rational);
COMPLEX_RMINUS1 (float);
COMPLEX_RMINUS1 (big);



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Return the result of multiplying self with v{x}.

	When subtracting a Complex number with another then the other is converted
	into a Complex.
	@def
		x * y = z

	  # where:

		z.real = x.real * y.real - x.imag * y.imag
		z.imag = x.real * y.imag + x.imag * y.real

	@p
	The result depends on whether the -{Number.context} specifies REALIFY_COMPLEX.
	If z.imag is zero and REALIFY_COMPLEX is specified then a -{Real} is returned,
	a Complex otherwise.

	@exc
	.EMul
	.EComplex
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 2 + 0j
		b = 1 + 2j
		println (a * b)

	@out
		(2 + 4j)
*/
static qu_number qu__complex__mul (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



static qu_number qu__complex_mulri (qu_real xr, qu_real xi, qu_real yr, qu_real yi)
{
	qu_real r, i, t;

	r = qu_real_mul_real   (xr, yr); QU_ENSURE (r); assert (qu_real_is (r));
	t = qu_real_mul_real   (xi, yi); QU_ENSURE (t); assert (qu_real_is (t));
	r = qu_real_minus_real (r, t);   QU_ENSURE (r); assert (qu_real_is (r));

	i = qu_real_mul_real   (xr, yi); QU_ENSURE (i); assert (qu_real_is (i));
	t = qu_real_mul_real   (xi, yr); QU_ENSURE (t); assert (qu_real_is (t));
	i = qu_real_plus_real  (i, t);   QU_ENSURE (i); assert (qu_real_is (i));

	return __qu_complex__ (r, i);
}



#define COMPLEX_MUL1(ytype) \
qu_number qu_complex_mul_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu__complex_mulri (x->real, x->imag, (qu_real) y, (qu_real) QuInt0); \
}



COMPLEX_MUL1 (int)
COMPLEX_MUL1 (long)
COMPLEX_MUL1 (rational)
COMPLEX_MUL1 (float)
COMPLEX_MUL1 (big)



qu_number qu_complex_mul_complex (qu_complex x, qu_complex y)
{
	return qu__complex_mulri (x->real, x->imag, y->real, y->imag);
}



qu_number qu_complex_mul_inf (qu_complex x, qu_inf y)
{
	return qu__complex_mulri (x->real, x->imag, (qu_real) y, (qu_real) QuInt0);
}



QU_NUMBER_BINOP_integer (qu_complex, mul, qu_number)
QU_NUMBER_BINOP_decimal (qu_complex, mul, qu_number)
QU_NUMBER_BINOP_real    (qu_complex, mul, qu_number)
QU_NUMBER_BINOP_number  (qu_complex, mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Complex
	Return the negation of self.

	This method ignores ALLOW_COMPLEX, REALIFY_COMPLEX and
	ALLOW_INF because __neg is an I{identity} methods which must
	always succeed and return an object of the same class.

	Both the real and imaginary parts negated.

	@noexc

	@xmp
		a = 2 + 3j
		println (-a)

	@out
		(-2 - 3j)
*/
qu_complex qu__complex__neg (qu_complex self)
{
	return qu_complex_newr (qu_real_neg (self->real), qu_real_neg (self->imag));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x + y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Return the result of adding v{x} to self.

	When adding a Complex number with another then the other is converted
	into a Complex and the calculation uses this formula:

	@def
		x - y = z

	  # where:

		z.real = x.real + y.real
		z.imag = x.imag + y.imag

	@p
	The result depends on whether the -{Number.context} specifies REALIFY_COMPLEX.
	If z.imag is zero and REALIFY_COMPLEX is specified then a -{Real} is returned,
	a Complex otherwise.

	@exc
	.EPlus
	.EComplex
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 2 + 3j
		b = 7 - 1j
		println (a + b)

	@out
		9 + 2j
*/
static qu_number qu__complex__plus (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPlus);
}


#define COMPLEX_PLUS1(ytype) \
qu_number qu_complex_plus_##ytype (qu_complex x, qu_##ytype y) \
{ \
	qu_real r = qu_##ytype##_plus_real (y, x->real); \
	QU_ENSURE (r); \
	return __qu_complex__ (r, x->imag); \
}


COMPLEX_PLUS1 (int);
COMPLEX_PLUS1 (long);
COMPLEX_PLUS1 (rational);
COMPLEX_PLUS1 (float);
COMPLEX_PLUS1 (big);



qu_number qu_complex_plus_complex (qu_complex x, qu_complex y)
{
	qu_real r, i;
	r = qu_real_plus_real (x->real, y->real); QU_ENSURE (r); assert (qu_real_is (r));
	i = qu_real_plus_real (x->imag, y->imag); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (r, i);
}



qu_number qu_complex_plus_inf (qu_complex x, qu_inf y)
{
	qu_real r = qu_real_plus_inf (x->real, y);
	QU_ENSURE (r);
	return __qu_complex__ (r, x->imag);
}



QU_NUMBER_BINOP_integer (qu_complex, plus, qu_number)
QU_NUMBER_BINOP_decimal (qu_complex, plus, qu_number)
QU_NUMBER_BINOP_real    (qu_complex, plus, qu_number)
QU_NUMBER_BINOP_number  (qu_complex, plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pos -> self
	Return a positive version of self.

	This method ignores ALLOW_COMPLEX, REALIFY_COMPLEX and
	ALLOW_INF because __pos is an I{identity} method which must
	always succeed and return an object of the same class.

	Thus this method always return self.

	@noexc

	@xmp
		c = 1 + 2j
		println (+c)

	@out
		1 + 2j
*/
#define qu__complex__pos qu_object_method_self




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x ** y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x:Number) -> Number
	Return the result of raising self to the power of v{x}.

	When raising a Complex number with another then the other is converted
	into a Complex.

	@def
		x ** y = z

	  # where:

		z.real = r1 * cos (a1)
		z.imag = r1 * sin (a1)

		r0 = hypot (x.real, x.imag)
		a0 = atan2 (x.imag, x.real)
		r1 = exp (y.real * log (r0) - y.imag * a0)
		a1 = a0 * y.real + y.imag * log (r0);

	@p
	The result depends on whether the -{Number.context} specifies REALIFY_COMPLEX.
	If z.imag is zero and REALIFY_COMPLEX is specified then a -{Real} is returned,
	a Complex otherwise.

	@exc
	.EPlus
	.EComplex
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 2 + 7j
		b = 2 + 0j

		println (a ** b)
		println (a * a)

	@out
		(-45 + 28j)
		(-45 + 28j)

	@note
	This method automatically loads the %{Math} module if required.
*/
static qu_number qu__complex__pow (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}


qu_complexpowbig_f __QuComplexPowComplexMath = NULL;

#define COMPLEX_POW1(ytype) \
qu_number qu_complex_pow_##ytype (qu_complex x, qu_##ytype y) \
{ \
	QU_ENSURE (qu_module_math ()); \
	assert (__QuComplexPowComplexMath); \
	return __QuComplexPowComplexMath (x->real, x->imag, (qu_real) y, (qu_real) QuInt0); \
}

COMPLEX_POW1 (int)
COMPLEX_POW1 (long)
COMPLEX_POW1 (rational)
COMPLEX_POW1 (float)
COMPLEX_POW1 (big)
COMPLEX_POW1 (inf)


qu_number qu_complex_pow_complex (qu_complex x, qu_complex y)
{
	QU_ENSURE (qu_module_math ());
	assert (__QuComplexPowComplexMath);
	return __QuComplexPowComplexMath (x->real, x->imag, y->real, y->imag);
}



QU_NUMBER_BINOP_integer (qu_complex, pow, qu_number)
QU_NUMBER_BINOP_decimal (qu_complex, pow, qu_number)
QU_NUMBER_BINOP_real    (qu_complex, pow, qu_number)
QU_NUMBER_BINOP_number  (qu_complex, pow, qu_number)



#define COMPLEX_RPOW1(ytype) \
qu_number qu_complex_rpow_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu__complex__pow (qu_complex_new1 ((qu_real) y), x); \
}


COMPLEX_RPOW1 (int)
COMPLEX_RPOW1 (long)
COMPLEX_RPOW1 (rational)
COMPLEX_RPOW1 (float)
COMPLEX_RPOW1 (big)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print -> Null
	Print k[self} to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		println (Complex (1, 2.3))

	@out
		(1 + 2.3j)
*/
static qu_null qu__complex__print (qu_complex self)
{
	qu_string s = qu_complex_as_string (self);
	return qu_sys_printz (s->str, s->len);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	x % y
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The meaning of "remainder" may differ with other languages.
	Refer to -{Number.__rem} for details.

	@def
		REM(x, y) = z

	  # where

		z = x - y * IDIV(x, y)

	@p
	The result depends on whether the -{Number.context} specifies REALIFY_COMPLEX.
	If z.imag is zero and REALIFY_COMPLEX is specified then a -{Real} is returned,
	a Complex otherwise.

	@exc
	.ERem
	.EComplex
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 2 + 3j
		b = 1 - 4j
		c = a \ b
		d = a % b
		e = c * b + d
		println (c)
		println (d)
		println (e)

	@out
		2
		(0 + 11j)
		(2 + 3j)
*/
static qu_number qu__complex__rem (qu_complex self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_complex_rem_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_complex_rem_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_complex_rem_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_complex_rem_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_complex_rem_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_rem_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_complex_rem_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuERem);
}



/*
	REM(a, b) = a - b * IDIV(a, b)
*/
static qu_number qu__complex_remri (qu_real xr, qu_real xi, qu_real yr, qu_real yi)
{
	qu_integer v;
	qu_number  m;

	v = qu_real_idiv_real (xr, yr); QU_ENSURE (v); assert (qu_integer_is (v));
	m = qu__complex_mulri (yr, yi, (qu_real) v, (qu_real) QuInt0); QU_ENSURE (m);

	if (qu_complex_is (m))
	{
		xr = qu_real_minus_real (xr, ((qu_complex) m)->real); QU_ENSURE (xr);
		xi = qu_real_minus_real (xi, ((qu_complex) m)->imag); QU_ENSURE (xi);
	}
	else
	{
		xr = qu_real_minus_real (xr, (qu_real) m); QU_ENSURE (xr);
	}
	return __qu_complex__ (xr, xi);
}



#define COMPLEX_REM1(ytype) \
qu_number qu_complex_rem_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu__complex_remri (x->real, x->imag, (qu_real) y, (qu_real) QuInt0); \
}


COMPLEX_REM1 (int)
COMPLEX_REM1 (long)
COMPLEX_REM1 (rational)
COMPLEX_REM1 (float)
COMPLEX_REM1 (big)



qu_number qu_complex_rem_complex (qu_complex x, qu_complex y)
{
	return qu__complex_remri (x->real, x->imag, y->real, y->imag);
}



qu_number qu_complex_rem_inf (qu_complex x, qu_inf y)
{
	return qu__complex_remri (x->real, x->imag, (qu_real) y, (qu_real) QuInt0);
}



QU_NUMBER_BINOP_integer (qu_complex, rem, qu_number)
QU_NUMBER_BINOP_decimal (qu_complex, rem, qu_number)
QU_NUMBER_BINOP_real    (qu_complex, rem, qu_number)
QU_NUMBER_BINOP_number  (qu_complex, rem, qu_number)



/*
	y % x
*/

#define COMPLEX_RREM1(ytype) \
qu_number qu_complex_rrem_##ytype (qu_complex x, qu_##ytype y) \
{ \
	return qu__complex_remri ((qu_real) y, (qu_real) QuInt0, x->real, x->imag); \
}


COMPLEX_RREM1 (int)
COMPLEX_RREM1 (long)
COMPLEX_RREM1 (rational)
COMPLEX_RREM1 (float)
COMPLEX_RREM1 (big)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	string
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Return a String representation of self.

	@noexc

	@xmp
		c = 1 + 2.3j
		s = String (c)
		println (s)

	@out
		(1 + 2.3j)
*/
qu_string qu__complex__string (qu_complex self)
{
	qu_real   g = self->imag;
	int       s = qu_real_isneg (g);
	qu_string r, i;

	if (s)
		g = qu_real_neg (g);

	r = qu_sys_string (self->real);
	i = qu_sys_string (g);

	return qu_string_cat (qu_string_byte ('('), r, qu_string_byte (' '),
		s ? qu_string_byte ('-') : qu_string_byte ('+'), qu_string_byte (' '),
		i, qu_inf_is (g) ? qu_string_byte ('.') : QuStringEMPTY,
		qu_string_byte ('j'), qu_string_byte (')'), 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	If both the real and imaginary parts are zero then false is returned,
	true otherwise.

	@noexc

	@xmp
		a = 0 + 1j
		println (a ? 'T' : 'F')

	@out
		T
*/
static qu_boolean qu__complex__true (qu_complex self)
{
	return (qu_real_iszero (self->real) && qu_real_iszero (self->imag)) ? QuFalse : QuTrue;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub abs -> Real
	Return the absolute value.

	The value is mathematically defined as:

	@def
		abs (z) = hypot (z.real, z.imag)

	@xmp
		x = 1 + 2j
		println (x.abs)

	@out
		2.23606797749979

	@note
	This method automatically loads the %{Math} module if required.
*/
qu_complexabs_f __QuComplexAbsMath;

qu_real qu__complex_abs (qu_complex self)
{
	/* TODO - try without the Math module */

	QU_ENSURE (qu_module_math ());
	assert (__QuComplexAbsMath);
	return __QuComplexAbsMath (self);
}



/*) sub arg -> Real
	Return the I{argument} of the complex number which is defined by the
	formula: atan2 C{(self.real, self.imag)}.

	@xmp
		x = 2 + 3j
		println (x.arg)

	@out
		0.5880026035475675

	@note
	This method automatically loads the %{Math} module if required.
*/
qu_complexarg_f __QuComplexArgMath;
static qu_real qu__complex_arg (qu_complex self)
{
	/* TODO - try without the Math module */

	QU_ENSURE (qu_module_math ());
	assert (__QuComplexArgMath);
	return __QuComplexArgMath (self);
}



/*) sub ceil -> Number
	Return the result of rounding both real and imaginary parts to the nearest
	integer towards positive infinity.
*/
qu_number qu__complex_ceil (qu_complex x)
{
	qu_real r, i;
	r = qu_real_ceil (x->real); QU_ENSURE (r); assert (qu_real_is (r));
	i = qu_real_ceil (x->imag); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (r, i);
}



/*) sub conj -> Complex
	Return the I{conjugate} of self, i.e. a Complex number with the same
	real part and negated imaginary part.

	@xmp
		a = -2 - 3j
		println (a.conj)

	@out
		(-2 + 3j)
*/
qu_complex qu__complex_conj (qu_complex self)
{
	return qu_complex_newr (self->real, qu_real_neg (self->imag));
}



/*) sub floor -> Number
	Return the result of rounding both real and imaginary parts to the nearest
	integer towards negative infinity.
*/
qu_number qu_complex_floor (qu_complex x)
{
	qu_real r, i;
	r = qu_real_floor (x->real); QU_ENSURE (r); assert (qu_real_is (r));
	i = qu_real_floor (x->imag); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (r, i);
}



/*) sub frac -> Number
	Return a new Complex number by converting both real and imaginary parts to
	their fractions.

	@xmp
		x = 1.2 + 3.4j
		println (x.frac)

	@out
		(0.2 + 0.4j)
*/
qu_number qu__complex_frac (qu_complex x)
{
	qu_real r, i;
	r = qu_real_frac (x->real); QU_ENSURE (r); assert (qu_real_is (r));
	i = qu_real_frac (x->imag); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (r, i);
}



/*) sub imag -> Real
	Return the imaginary part.

	@noexc

	@xmp
		a = 1 + 2j
		println (a.imag)

	@out
		2
*/
static qu_real qu__complex_imag (qu_complex self)
{
	return self->imag;
}



/*) sub iseven -> Boolean
	Return true if the real part is an even number, false otherwise.
*/
static qu_boolean qu__complex_iseven (qu_complex self)
{
	return qu_real_isodd (self->real) ? QuFalse : QuTrue;
}



/*) sub isfrac -> Boolean
	Return true if the real part is a fraction, false otherwise.
*/
static qu_boolean qu__complex_isfrac (qu_complex self)
{
	return qu_real_isfrac (self->real) ? QuTrue : QuFalse;
}



/*) sub isodd -> Boolean
	Return true if the real part is a whole and odd number, false otherwise.
*/
static qu_boolean qu__complex_isodd (qu_complex self)
{
	return qu_real_isodd (self->real) ? QuTrue : QuFalse;
}



/*) sub iswhole -> Boolean
	Return true if the real part is a I{whole} number, false otherwise.
*/
static qu_boolean qu__complex_iswhole (qu_complex self)
{
	return qu_real_iswhole (self->real) ? QuTrue : QuFalse;
}



/*) sub product -> Real
	Return the result of C{self.real * self.real + self.imag * self.imag}.

	@xmp
		a = 2 + 4j
		x = a.product
		println (x)

	@out
		20
*/
static qu_real qu__complex_product (qu_complex self)
{
	qu_real a, b;

	a = qu_real_mul_real (self->real, self->real);
	QU_ENSURE (a);
	assert (qu_real_is (a));

	b = qu_real_mul_real (self->imag, self->imag);
	QU_ENSURE (b);
	assert (qu_real_is (b));

	return qu_real_plus_real (a, b);
}



/*) sub real -> Real
	Return the real part.

	@noexc

	@xmp
		a = 1 + 2j
		println (a.real)

	@out
		1
*/
static qu_real qu__complex_real (qu_complex self)
{
	return self->real;
}




/*) sub round (n: Integer) -> Number
	Return a new Complex by rounding the real part.
*/
static qu_number qu__complex_round (qu_complex self, qu_integer prec)
{
	int p;
	QU_ARG_int (prec, p);
	return qu_complex_round (self, p);
}



qu_number qu_complex_round (qu_complex self, int prec)
{
	qu_real r= qu_real_round (self->real, prec);
	QU_ENSURE (r);
	assert (qu_real_is (r));
	return __qu_complex__ (r, self->imag);
}



/*) sub sign -> Int
	Return -1, 0, 1 if the real part is negatve, zero, or positive respectively.
*/
static qu_int qu__complex_sign (qu_complex self)
{
	long v = qu_real_sign (self->real);
	return (v < 0) ? QuInt_1 : ((v > 0) ? QuInt1 : QuInt0);
}



/*) sub trunc -> Number
	Return the result of rounding both the real and imaginary parts to the nearest
	integer towards zero, thus removing the fractional part.
*/
qu_number qu__complex_trunc (qu_complex x)
{
	qu_real r, i;
	r = qu_real_trunc (x->real); QU_ENSURE (r); assert (qu_real_is (r));
	i = qu_real_trunc (x->imag); QU_ENSURE (i); assert (qu_real_is (i));
	return __qu_complex__ (r, i);
}


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_complex__setup ()
{
	qu_class_setup_final (QuComplexClass, qu_complex_s, NULL, qu__complex__mark);
}


void qu_complex__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuComplexClass, QU_NAME_new, qu__complex_new, -1);

	/* specific class methods */
	qu_class_method_classz (QuComplexClass, "ret", qu__complex_ret, 2);

	/* standard instance methods */
	qu_class_method_inst   (QuComplexClass, QU_NAME__as    , qu__complex__as    , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__cmp   , qu__complex__cmp   , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__decr  , qu__complex__decr  , 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__div   , qu__complex__div   , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__eq    , qu__complex__eq    , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__format, qu__complex__format, 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__hash  , qu__complex__hash  , 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__idiv  , qu__complex__idiv  , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__incr  , qu__complex__incr  , 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__like  , qu__complex__like  , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__minus , qu__complex__minus , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__mul   , qu__complex__mul   , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__neg   , qu__complex__neg   , 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__plus  , qu__complex__plus  , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__pos   , qu__complex__pos   , 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__pow   , qu__complex__pow   , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__print , qu__complex__print , 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__rem   , qu__complex__rem   , 1);
	qu_class_method_inst   (QuComplexClass, QU_NAME__string, qu__complex__string, 0);
	qu_class_method_inst   (QuComplexClass, QU_NAME__true  , qu__complex__true  , 0);

	/* specific instance methods */
	qu_class_method_instz  (QuComplexClass, "abs"    , qu__complex_abs    , 0);
	qu_class_method_instz  (QuComplexClass, "arg"    , qu__complex_arg    , 0);
	qu_class_method_instz  (QuComplexClass, "ceil"   , qu__complex_ceil   , 0);
	qu_class_method_instz  (QuComplexClass, "conj"   , qu__complex_conj   , 0);
	qu_class_method_instz  (QuComplexClass, "floor"  , qu__complex_floor  , 0);
	qu_class_method_instz  (QuComplexClass, "frac"   , qu__complex_frac   , 0);
	qu_class_method_instz  (QuComplexClass, "imag"   , qu__complex_imag   , 0);
	qu_class_method_instz  (QuComplexClass, "iseven" , qu__complex_iseven , 0);
	qu_class_method_instz  (QuComplexClass, "isfrac" , qu__complex_isfrac , 0);
	qu_class_method_instz  (QuComplexClass, "isodd"  , qu__complex_isodd  , 0);
	qu_class_method_instz  (QuComplexClass, "iswhole", qu__complex_iswhole, 0);
	qu_class_method_instz  (QuComplexClass, "real"   , qu__complex_real   , 0);
	qu_class_method_instz  (QuComplexClass, "product", qu__complex_product, 0);
	qu_class_method_instz  (QuComplexClass, "round"  , qu__complex_round  , 1);
	qu_class_method_instz  (QuComplexClass, "sign"   , qu__complex_sign   , 0);
	qu_class_method_instz  (QuComplexClass, "trunc"  , qu__complex_trunc  , 0);

	/* finalize declaration */
	qu_class_set_super (QuComplexClass, QuNumberClass);
	qu_class_enddecl   (QuComplexClass);

	QuComplex00      = qu__complex_static_int ( 0,  0);
	QuComplex01      = qu__complex_static_int ( 0,  1);
	QuComplex0_1     = qu__complex_static_int ( 0, -1);
	QuComplex10      = qu__complex_static_int ( 1,  0);
	QuComplex11      = qu__complex_static_int ( 1,  1);
	QuComplex1_1     = qu__complex_static_int ( 1, -1);
	QuComplex_10     = qu__complex_static_int (-1,  0);
	QuComplex_11     = qu__complex_static_int (-1,  1);
	QuComplex_1_1    = qu__complex_static_int (-1, -1);

	QuComplexBig00   = qu__complex_static_big (QuBig0 , QuBig0);
	QuComplexBig01   = qu__complex_static_big (QuBig0 , QuBig1);
	QuComplexBig0_1  = qu__complex_static_big (QuBig0 , QuBig_1);
	QuComplexBig10   = qu__complex_static_big (QuBig1 , QuBig0);
	QuComplexBig11   = qu__complex_static_big (QuBig1 , QuBig1);
	QuComplexBig1_1  = qu__complex_static_big (QuBig1 , QuBig_1);
	QuComplexBig_10  = qu__complex_static_big (QuBig_1, QuBig0);
	QuComplexBig_11  = qu__complex_static_big (QuBig_1, QuBig1);
	QuComplexBig_1_1 = qu__complex_static_big (QuBig_1, QuBig_1);

	QuComplexInfp    = qu__complex_static_inf (QuInfPos);
	QuComplexInfn    = qu__complex_static_inf (QuInfNeg);
}
