/* qu_container.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Container __abstract
	This is the parent class of the following classes:
	@tree
	.Container
	.	-{Array}
	.	-{Set}
	.	-{Dict}
	.	-{Queue}
	.	-{Record}

	@xmp
		map = [{|x| x * 2}, "__incr"]
		a = [1, 2, 3]
		a.apply (map)
		println (a)

		a = [1 2 3 4]
		a.every (@println)

	@out
		[3, 5, 7]
		1
		2
		3
		4

	@warn
	If the Container is a Dict then the methods iterate through the keys,
	not the values.

	@xmp
		map = ["tolower", "capitalize", @println]
		d = [fOO: 1, baR: 2]
		d.every (map)

	@out
		Foo
		Bar
*/
qu_class QuContainerClass;



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/
/*none*/



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/
/* none */



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/
/* none */



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __len -> Int
	Return the number of items in this Container.

	@noexc

	@p
	Note that there is also the more convenient #{len} method.
*/
/*) sub len -> Int
	This is an alias of the __len method.
*/
static qu_int qu__container__len (qu_container self)
{
	assert (0 <= self->len && self->len <= QU_GC_MAX_ITEMS && self->len <= QU_INT_MAX);
	return qu_int_new (self->len);
}



/*) sub __true -> Boolean
	Return false if the Container is empty, true otherwise.
*/
static qu_boolean qu__container__true (qu_container self)
{
	return qu_container_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub apply (map) -> self
	For all items in self process them through v{x} then apply to self.

	Thus this is the same as #{map} except that the items are used to modify
	self instead of collected.

	This method calls -{sysmap}, thus:
	@ul
	.If v{x} is a String then the item's method named v{x} is called.

	.Else if v{x} is a -{callable} then it is called with the item as the
	 only argument.

	.Else v{x} is an Array of String or -{callable}. Each are called
	 consecutively to process the item. Note that it is possible for the Array
	 to dynamically change while in the loop. It is probably not desirable,
	 possible nonetheless.


	@xmp
		map = [{|x| x * 2}, "__incr"]
		a = [1, 2, 3]
		a.apply (map)
		println (a)

	@out
		[3, 5, 7]
*/
static int qu__container_apply_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any self = argv[0];
	qu_any map  = argv[1];

	item = qu_sys_map (item, map);
	if (item == 0)
	{
		if (qu_thread_err_is_estop ())
			return QU_ITERATOR_CALLBACK_END;
		return QU_ITERATOR_CALLBACK_ERROR;
	}
	if (qu_sys_setitem (self, idx, item) == 0)
		return QU_ITERATOR_CALLBACK_ERROR;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_apply (qu_container self, qu_any ap)
{
	qu_any arg[2] = {self, ap};
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_apply_callback, arg)
		!= QU_ITERATOR_CALLBACK_ERROR);
	return self;
}



/*) sub count (what, like = false) -> Int
	Count objects in Container.

	v{what} may be an Array, Set or Dict to match any object within it, or
	any other object to match that particular object.
	In case of a Dict then the keys are used and the values ignored.

	If v{like} is true then the __like method is used to do the matching,
	the __eq method otherwise.

	@exc
	.EArgValue

	@xmp
		a = {'foo', 'BAR', 'bar'}

		println (a.count ('bar', true))

		b = ['bir', 'bar', 'bor']
		println (a.count (b))

	@out
		2
		1
*/
typedef struct count_t count_t;
struct count_t
{
	qu_container self;
	qu_any       what;
	long         count;
};


static int _container_callback_count_like_item (qu_any item, qu_any idx, count_t *f)
{
	if (qu_sys_like (f->what, item) == QuTrue)
		f->count++;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int qu__container_callback_count_like (qu_any item, qu_any idx, count_t *f)
{
	count_t m = {f->self, item, 0};
	qu_iterator_do (f->self, (qu_itercallback_f) _container_callback_count_like_item, &m);
	f->count += m.count;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int _container_callback_count_eq_item (qu_any item, qu_any idx, count_t *f)
{
	if (qu_sys_eq (item, f->what) == QuTrue)
		f->count++;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int qu__container_callback_count_eq (qu_any item, qu_any idx, count_t *f)
{
	if (qu_sys_in (f->self, item) == QuTrue)
		f->count++;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_count (qu_container self, int argc, qu_any *argv)
{
	qu_any what;
	int    like = 0;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (argc > 1)
		QU_ARG_vbool (1, like);
	/*begin*/
	{
		count_t f = {self, what, 0};
		if (qu_array_is (what) || qu_dict_is (what) || qu_set_is (what))
		{
			if (like)
				qu_iterator_do (what, (qu_itercallback_f) qu__container_callback_count_like, &f);
			else
				qu_iterator_do (what, (qu_itercallback_f) qu__container_callback_count_eq, &f);
		}
		else
		{
			if (like)
				qu_iterator_do (self, (qu_itercallback_f) _container_callback_count_like_item, &f);
			else
				qu_iterator_do (self, (qu_itercallback_f) _container_callback_count_eq_item, &f);
		}
		return qu_int_new (f.count);
	}
}



/*) sub cross (x) -> Any
	Return the item in self which is also in v{x}, return nil if there is
	none.

	v{x} must support the __in method.

	@xmp
		a = [1 2 3]
		b = {4 7 2}
		c = [foo:8 bar:2]
		d = [1:'foo' 2:'bar']
		println (a.cross (b))
		println (a.cross (c))
		println (a.cross (d))

	@out
		2
		nil
		1
*/
static int qu__container_cross_callback (qu_any item, qu_any idx, qu_any *x)
{
	qu_boolean r = qu_sys_in (*x, item);
	if (r == QuTrue)
	{
		*x = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_boolean qu__container_cross (qu_container self, qu_any x)
{
	return (qu_iterator_do (self, (qu_itercallback_f) qu__container_cross_callback, &x)
		== QU_ITERATOR_CALLBACK_RESULT) ? x : (qu_any) QuNil;
}



/*) sub diff (x, res = {})
	Collect items in self which are not in v{x} and items in v{x} which are
	not in self. The items are collected in v{res} and v{res} is returned
	as the result.

	v{res} must support the __push method.

	v{x} must support the __each and __in method.

	@xmp
		a = {1 2 3 4}
		b = [1 5 6 4]
		r = a.diff (b)
		println (r)

		a.diff (b, __stdout__)
		println

	@out
		{2, 3, 5, 6}
		2356
*/
static int qu__container_diff_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any     x   = argv[0];
	qu_any     res = argv[1];
	qu_boolean r;

	r = qu_sys_in (x, item);
	if (r == QuFalse)
	{
		if (qu_sys_push (res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_diff (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	/*begin*/
	{
		qu_any res    = (argc > 1) ? argv[1] : (qu_any) qu_set_new (0);
		qu_any arg[2] = {argv[0], res};

		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_diff_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		arg[0] = self;
		QU_ENSURE (qu_iterator_do (argv[0], (qu_itercallback_f) qu__container_diff_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		return res;
	}
}



/*) sub every (x) -> self
	For all items in self process them through v{x}.

	This method calls -{sysmap}, thus:
	@ul
	. If v{x} is a String then the item's method named v{x} is called.

	. Else if v{x} is a -{callable} then it is called with the item as the
	  only argument.

	. Else v{x} is an Array of String or callable objects. Each are called
	  consecutively to process the item. Note that it is possible for the Array
	  to dynamically change while in the loop. It is probably not desirable,
	  possible nonetheless.

	@p
	If EStop is thrown within the loop then iteration stops with no error.
	Any other exceptions are rethrown.

	@xmp
		x = ["lcase", @println]
		a = ['fOo', 'bAr']
		a.every (x)

	@out
		foo
		bar
*/
static int qu__container_every_callback (qu_any item, qu_any idx, qu_any ap)
{
	item = qu_sys_map (item, ap);
	if (item == 0)
	{
		if (qu_thread_err_is_estop ())
			return QU_ITERATOR_CALLBACK_END;
		return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


qu_container qu__container_every (qu_container self, qu_any ap)
{
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_every_callback, ap)
		!= QU_ITERATOR_CALLBACK_ERROR);
	return self;
}



/*) sub fill (x) -> self
	Fill self with v{x}.

	If self is empty then nothing happens.

	If v{x} is a Dict the this method fills the values.
	It does not change the keys.

	@xmp
		a = [1, 2, 3].fill ('foo')
		println (a)

		a = [foo:1, bar:2, baz:3].fill (5)
		println (a)

	@out
		['foo', 'foo', 'foo']
		['foo':5, 'bar':5, 'baz':5]
*/
static int qu__container_fill_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any self = argv[0];
	qu_any fill = argv[1];

	if (qu_dict_is (self))
	{
		if (qu_dict_setitem (self, item, fill) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	else
	{
		if (qu_sys_setitem (self, idx, fill) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_container qu__container_fill (qu_container self, qu_any fill)
{
	qu_any arg[2] = {self, fill};
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_fill_callback, arg)
		!= QU_ITERATOR_CALLBACK_ERROR);
	return self;
}



/*) sub filter (x, res = []) -> Any
	For all items in self filter them through v{x} and collect them
	in v{res}.

	v{res} must support the __push method.

	This method calls -{sysfilter}, thus:
	@ul
	.If v{x} is a String then the item's method named v{x} is called.

	.Else if v{x} is a -{callable} then it is called with the item as the
	 only argument.

	.Else v{x} is an Array of String or -{callable}. Each are called
	 consecutively to process the item. Note that it is possible for the Array
	 to dynamically change while in the loop. It is probably not desirable,
	 possible nonetheless.

	@p
	I{Filtering} here means to test whether the item is qualified to be
	included in v{res}.

	The method or function is expected to return true or false.

	If EStop is thrown within the loop then iteration stops with no error.
	Any other exceptions are rethrown.

	@xmp
		map = [lambda (x) x.len > 3, "isalpha"]

		a = ['foo', 'fooey', 'foo123']
		r = a.filter (map)
		println (r)

	@out
		['fooey']

	@p
	Since this method makes use of -{sysfilter} then it is aware of the
	Regex module.

	@xmp
		use Regex

		r = Regex (`[^abrz]`)
		a = ['foo', 'bar', 'baz']
		r = a.filter (r)
		println (r)

	@out
		['foo']
*/
static int qu__container_filter_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any     flt = argv[0];
	qu_any     res = argv[1];
	qu_boolean r;

	r = qu_sys_filter (item, flt);
	if (r == 0)
	{
		if (qu_thread_err_is_estop ())
			return QU_ITERATOR_CALLBACK_END;
		return QU_ITERATOR_CALLBACK_ERROR;
	}
	if (r == QuTrue)
	{
		if (qu_sys_push (res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_filter (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	/*begin*/
	{
		qu_any arg[2] = {argv[0], (argc > 1) ? argv[1] : (qu_any) qu_array_new (0)};
		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_filter_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		return arg[1];
	}
}



/*) sub grab (x, like = false, res = []) -> Any
	Collect matching objects.

	v{x} may be an Array, Set or Dict to match any object within it, or
	any other object to match that particular object.
	In case of a Dict then the keys are used and the values ignored.

	If v{like} is true then the __like method is used to do the matching,
	the __eq method otherwise.

	@exc
	.EArgValue

	@xmp
		a = ['foo', 'BAR', 'bar']
		b = a.grab ('bar', true)
		println (b)

	@out
		[BAR, bar]
*/
typedef struct grab_t grab_t;
struct grab_t
{
	qu_container self;
	qu_any       what;
	int          like;
	qu_any       res;
};


static int _container_callback_grab_like_item (qu_any item, qu_any idx, grab_t *f)
{
	if ((f->like ? qu_sys_like (f->what, item) : qu_sys_eq (f->what, item)) == QuTrue)
	{
		if (qu_sys_push (f->res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int qu__container_callback_grab_like (qu_any item, qu_any idx, grab_t *f)
{
	grab_t m = {f->self, item, f->like, f->res};
	qu_iterator_do (f->self, (qu_itercallback_f) _container_callback_grab_like_item, &m);
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int _container_callback_grab_eq_item (qu_any item, qu_any idx, grab_t *f)
{
	if (qu_sys_eq (item, f->what) == QuTrue)
	{
		if (qu_sys_push (f->res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int qu__container_callback_grab_eq (qu_any item, qu_any idx, grab_t *f)
{
	if (qu_sys_in (f->self, item) == QuTrue)
	{
		if (qu_sys_push (f->res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_grab (qu_container self, int argc, qu_any *argv)
{
	qu_any what;
	int    like = 0;
	int    r;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (argc > 1)
		QU_ARG_vbool (1, like);
	/*begin*/
	{
		grab_t f = {self, what, like, (argc > 2) ? argv[2] : (qu_any) qu_array_new (0)};
		if (qu_array_is (what) || qu_dict_is (what) || qu_set_is (what))
		{
			if (like)
				r = qu_iterator_do (what, (qu_itercallback_f) qu__container_callback_grab_like, &f);
			else
				r = qu_iterator_do (what, (qu_itercallback_f) qu__container_callback_grab_eq, &f);
		}
		else
		{
			if (like)
				r = qu_iterator_do (self, (qu_itercallback_f) _container_callback_grab_like_item, &f);
			else
				r = qu_iterator_do (self, (qu_itercallback_f) _container_callback_grab_eq_item, &f);
		}
		QU_ENSURE (r != QU_ITERATOR_CALLBACK_ERROR);
		return f.res;
	}
}



/*) sub intersect (x, res = {}) -> Any
	For all items in self which are also in k{x}, collect them in v{res}.

	k{x} must support the __in method.

	@xmp
		x = [1 2 2 4 5]
		y = {4 7 2 5 8}
		r = x.intersect (y)
		println (r)

	@out
		{2, 4, 5}
*/
static int qu__container_intersect_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any     x   = argv[0];
	qu_any     res = argv[1];
	qu_boolean r;

	r = qu_sys_in (x, item);
	if (r == QuTrue)
	{
		if (qu_sys_push (res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_intersect (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	/*begin*/
	{
		qu_any arg[2] = {argv[0], (argc > 1) ? argv[1] : (qu_any) qu_set_new (0)};
		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_intersect_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		return arg[1];
	}
}



/*) sub istrue (m: String = '__true') -> Any
	Find an item in Container which value is true, return nil if none are.

	v{m} is the method to call without any argument thus it may be a property
	instead of a method.

	@exc
	.Exception any exception thrown by the method

	@xmp
		x = [false, nil, 123]
		p = x.istrue

	@out
		123
*/
/*) sub isfalse (m: String = '__true') -> Any
	This works like #{istrue} except that a false value is looked up.
*/
static int qu__container_isbool_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_string m = argv[0];
	qu_any r;

	if (m == 0)
		r = qu_sys_true (item);
	else
	{
		r = qu_object_call0 (item, m);
		if (r == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
		if (!qu_boolean_is (r))
			r = qu_sys_true (r);
	}

	if (r == argv[1])
	{
		argv[2] = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_isbool (qu_container self, int argc, qu_any *argv, qu_any val)
{
	qu_string s = 0;

	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		QU_ARG_VTYPE (0, s, qu_string);
		if (qu_string_EQ (s, QU_NAME__true))
			s = 0;
	}
	/*begin*/
	{
		qu_any arg[3] = {s, val, 0};
		switch (qu_iterator_do (self, (qu_itercallback_f) qu__container_isbool_callback, arg))
		{
			case QU_ITERATOR_CALLBACK_RESULT:
				return arg[2];

			case QU_ITERATOR_CALLBACK_ERROR:
				QU_RETHROW;
		}
		return QuNil;
	}
}



static qu_any qu__container_istrue (qu_container self, int argc, qu_any *argv)
{
	return qu__container_isbool (self, argc, argv, QuTrue);
}



static qu_any qu__container_isfalse (qu_container self, int argc, qu_any *argv)
{
	return qu__container_isbool (self, argc, argv, QuFalse);
}



/*) sub match (x, like = false) -> Any
	Find object in Container.

	v{x} may be an Array, Set or Dict to match any object within it, or
	any other object to match that particular object.
	In case of a Dict then the keys are used and the values ignored.

	If v{like} is true then the __like method is used to do the matching,
	the __eq method otherwise.

	The returned value is nil if no match found, the object that matches otherwise.
	Thus you should probably not use this method if self or v{x} contains nil
	objects.

	@exc
	.EArgValue

	@xmp
		a = ['foo', 'BAR', 'baz']

		println (a.match ('bar', true))

		b = {'bir', 'bar', 'bor'}
		println (a.match (b, true))

	@out
		BAR
		BAR
*/
typedef struct match_t match_t;
struct match_t
{
	qu_container self;
	qu_any       what;
};


static int _container_callback_match_like_item (qu_any item, qu_any idx, match_t *f)
{
	if (qu_sys_like (f->what, item) == QuTrue)
	{
		f->what = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int qu__container_callback_match_like (qu_any item, qu_any idx, match_t *f)
{
	match_t m = {0, item};
	if (qu_iterator_do (f->self, (qu_itercallback_f) _container_callback_match_like_item, &m) == QU_ITERATOR_CALLBACK_RESULT)
	{
		f->what = m.what;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static int qu__container_callback_match_eq (qu_any item, qu_any idx, match_t *f)
{
	if (qu_sys_in (f->self, item) == QuTrue)
	{
		f->what = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_match (qu_container self, int argc, qu_any *argv)
{
	qu_any what;
	int    like = 0;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (argc > 1)
		QU_ARG_vbool (1, like);

	if (qu_array_is (what) || qu_dict_is (what) || qu_set_is (what))
	{
		match_t f = {self, what};
		if (like)
		{
			if (qu_iterator_do (what, (qu_itercallback_f) qu__container_callback_match_like, &f) == QU_ITERATOR_CALLBACK_RESULT)
				return f.what;
		}
		else
		{
			if (qu_iterator_do (what, (qu_itercallback_f) qu__container_callback_match_eq, &f) == QU_ITERATOR_CALLBACK_RESULT)
				return f.what;
		}
	}
	else
	{
		if (like)
		{
			match_t f = {0, what};
			if (qu_iterator_do (self, (qu_itercallback_f) _container_callback_match_like_item, &f) == QU_ITERATOR_CALLBACK_RESULT)
				return f.what;
		}
		else
		{
			/* CHECKME - this is not exactly right since we are returning 'what'
				instead of the item in 'self' */
			if (qu_sys_in (self, what))
				return what;
		}
	}
	return QuNil;
}



/*) sub map (x, res = []) -> Any
	For all items in self process them through v{x}, collect them in v{res}
	then return v{res}.

	This method calls -{sysmap}, thus:
	@ul
	.If v{x} is a String then the item's method named v{x} is called.

	.Else if v{x} is a -{callable} then it is called with the item as the
	 only argument.

	.Else v{x} is an Array of String or -{callable}. Each are called
	 consecutively to process the item. Note that it is possible for the Array
	 to dynamically change while in the loop. It is probably not desirable,
	 possible nonetheless.

	@xmp
		sub converter (x)
			static count = 0
			return ++count _ '-' _ x.capitalize
		;;

		x = ["tolower", @converter]
		a = ['fOo', 'bAr']
		r = a.map (x)
		println (r)

	@out
		['1-Foo', '2-Bar']
*/
static int qu__container_map_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any map = argv[0];
	qu_any res = argv[1];

	item = qu_sys_map (item, map);
	if (item == 0)
	{
		if (qu_thread_err_is_estop ())
			return QU_ITERATOR_CALLBACK_END;
		return QU_ITERATOR_CALLBACK_ERROR;
	}
	if (qu_sys_push (res, item) == 0)
		return QU_ITERATOR_CALLBACK_ERROR;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_map (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	{
		qu_any arg[2] = {argv[0], (argc > 1) ? argv[1] : (qu_any) qu_array_new (0)};
		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_map_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		return arg[1];
	}
}



/*) sub max (cmp = nil) -> Any
	Find the maximum value in self.

	If v{cmp} is nil then simply compare the item using the C{<=>}
	operator. Otherwise call the -{Function} v{cmp} to do the comparisson.

	The result of f{max} is compared with 0, i.e. C{result <=> 0}, thus it does
	not matter what kind of value the function returns.

	@xmp
		a = [4, 7, 3]
		m = a.max ({|x, y| return x <=> y})  # same as a.max
		println (m)

	@out
		7
*/
static int qu__container_max_callback (qu_any item, qu_any idx, qu_any *arg)
{
	qu_any res = arg[0];
	qu_any ap  = arg[1];
	qu_any r;

	if (res == 0)
	{
		arg[0] = item;
	}
	else
	{
		if (ap)
		{
			r = qu_object_call_x2 (ap, item, res);
			if (r == 0)
				return QU_ITERATOR_CALLBACK_ERROR;
			if (qu_sys_gt (r, QuInt0) == QuTrue)
				arg[0] = item;
		}
		else
		{
			if (qu_sys_gt (item, res) == QuTrue)
				arg[0] = item;
		}
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_max (qu_container self, int argc, qu_any *argv)
{
	qu_any arg[2] = {0, 0};
	if (argc > 0)
	{
		qu_any cmp = argv[0];
		if (cmp != (qu_any) QuNil)
		{
			QU_ARG_TYPE (cmp, qu_function);
			arg[1] = cmp;
		}
	}
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_max_callback, arg)
		!= QU_ITERATOR_CALLBACK_ERROR);
	QU_REQUIRE (QuENone, arg[0]);
	return arg[0];
}



/*) sub merge (x) -> self
	Merge items from v{x} into self.

	v{x} must support the __each method.

	@xmp
		a = [4, 7, 3]
		b = {5, 8, 2}
		b.merge (a)
		println (b)

	@out
		{2, 3, 4, 5, 7, 8}
*/
static int qu__container_merge_callback (qu_any item, qu_any idx, qu_any self)
{
	if (qu_dict_is (self))
	{
		if (qu_dict_setitem (self, item, idx) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	else
	{
		if (qu_sys_push (self, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_container qu__container_merge (qu_container self, qu_any x)
{
	QU_ENSURE (qu_iterator_do (x, (qu_itercallback_f) qu__container_merge_callback, self)
		!= QU_ITERATOR_CALLBACK_ERROR);
	return self;
}



/*) sub mergex (x) -> self
	Merge items from v{x} into self only if the item is not already in v{self}.

	v{x} must support the __each method.

	@xmp
		a = [foo:1 bar:2 baz:3]
		b = [foo:10 box:4]
		a.mergex (b)
		println (a)

	@out
		['foo':1, 'bar':2, 'baz':3, 'box':4]
*/
static int qu__container_mergex_callback (qu_any item, qu_any idx, qu_any self)
{
	if (qu_dict_is (self))
	{
		if (qu_dict_getitem (self, item) == 0)
		{
			if (qu_dict_setitem (self, item, idx) == 0)
				return QU_ITERATOR_CALLBACK_ERROR;
		}
	}
	else
	{
		if (qu_sys_in (self, item) != QuTrue)
		{
			if (qu_sys_push (self, item) == 0)
				return QU_ITERATOR_CALLBACK_ERROR;
		}
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_container qu__container_mergex (qu_container self, qu_any x)
{
	QU_ENSURE (qu_iterator_do (x, (qu_itercallback_f) qu__container_mergex_callback, self)
		!= QU_ITERATOR_CALLBACK_ERROR);
	return self;
}



/*) sub min (cmp = nil) -> Any
	This does the same as #{max} except that it returns the minimum value.

	@xmp
		a = [4, 7, 3]
		m = a.min ({|x, y| return x <=> y})  # same as a.min
		println (m)

	@out
		3
*/
static int qu__container_min_callback (qu_any item, qu_any idx, qu_any *arg)
{
	qu_any res = arg[0];
	qu_any ap  = arg[1];
	qu_any r;

	if (res == 0)
	{
		arg[0] = item;
	}
	else
	{
		if (ap)
		{
			r = qu_object_call_x2 (ap, item, res);
			if (r == 0)
				return QU_ITERATOR_CALLBACK_ERROR;
			if (qu_sys_lt (r, QuInt0) == QuTrue)
				arg[0] = item;
		}
		else
		{
			if (qu_sys_lt (item, res) == QuTrue)
				arg[0] = item;
		}
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_min (qu_container self, int argc, qu_any *argv)
{
	qu_any arg[2] = {0, 0};
	if (argc > 0)
	{
		qu_any cmp = argv[0];
		if (cmp != (qu_any) QuNil)
		{
			QU_ARG_TYPE (cmp, qu_function);
			arg[1] = cmp;
		}
	}
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_min_callback, arg)
		!= QU_ITERATOR_CALLBACK_ERROR);
	QU_REQUIRE (QuENone, arg[0]);
	return arg[0];
}



/*) sub move (x) -> self
	Move all items from self to v{x}.

	k{x} must support the __push method.

	This method works as follows:
	@prog
		for item in self
			x <<< item
		;;
		self.clear

	@p
	Thus if an error occurs in the middle of the process, then self will not
	be empty while v{x} may be modified.

	@xmp
		a = {1, 2, 3, 4, 5}
		b = [1, 2]
		a.move (b)
		println (a)
		println (b)

	@out
		{}
		[1, 2, 1, 2, 3, 4, 5]

	@p
	A special treatment is performed when self and/or v{x} is a -{Dict}.

	@xmp
		a = [foo:1 bar:2]
		b = [foo:10]
		a.move (b)
		println (a)
		println (b)

	@out
		[:]
		['foo':1, 'bar':2]

	@note
	Since Stream objects support the __push method then, for example,
	saying C{a.move (null)} has the same effect as C{a.clear}.

	@xmp
		a = [1, 2, 3]
		a.move (__stdout__)
		println
		println (a)

	@out
		123
		[]
*/
static int qu__container_move_callback (qu_any item, qu_any idx, qu_any x)
{
	if (qu_dict_is (x))
	{
		if (qu_dict_setitem (x, idx, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	else
	{
		if (qu_sys_push (x, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_container qu__container_move (qu_container self, qu_any x)
{
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_move_callback, x)
		!= QU_ITERATOR_CALLBACK_ERROR);
	QU_ENSURE (qu_sys_clear (self));
	return self;
}



/*) sub pair (*arg) -> Array
	Return an Array of pairs taken consecutively from iterable objects in v{arg}.

	@xmp
		a = [1 2 3 4 5]
		b = {4 5 6}
		c = 3..10
		r = a.pair (b, c)
		println (r)

	@out
		[[1, 4, 3], [2, 5, 4], [3, 6, 5]]
*/
static qu_array qu__container_pair (qu_container self, int argc, qu_any *argv)
{
	qu_array    a, r;
	qu_iterator i, j, *p;
	long        k, n = argc + 1;

	QU_ARG_VMIN (1);

	i = qu_iterator_from (self);
	QU_ENSURE (i);

	a = qu_array_new (n);
	qu_array_push (a, i);

	while (argc--)
	{
		j = qu_iterator_from (*argv++);
		QU_ENSURE (j);
		qu_array_push (a, j);
	}
	r = qu_array_new (0);
	for (;;)
	{
		qu_array m = qu_array_new (n);
		qu_any   x;

		p = (qu_iterator *) a->objs;
		for (k = 0; k < n; ++k)
		{
			i = *p++;
			x = qu_iterator_next (i);
			if (x == 0)
			{
				if (qu_thread_err_is_estop ())
					goto done;
				QU_RETHROW;
			}
			QU_ENSURE (qu_array_push (m, x));
		}
		QU_ENSURE (qu_array_push (r, m));
	}
  done:
	return r;
}



/*) sub product (op, ap = nil) -> Any
	Calculate the product of self against v{op}.

	First, the result is set as the first item in self.
	Thus there must be at least one item in self, otherwise ENone is thrown.

	Then for the rest of the items in self, apply v{op} and update the result.
	If called with two arguments, then the result is further processed via v{ap}.

	v{op} may be a callable object taking two arguments, or a Qu binary operator
	name.

	Like #{map}, v{ap} may be a String, callable or an Array of String and callable.
	If nil then it is ignored.

	If EStop is thrown within the loop then iteration stops with no error.
	Any other exceptions are rethrown.

	The end result is returned.

	@xmp
		x = [1, 2, 3, 4]
		println (x.product ('+'))

		println

		ap = ['__incr', {|x| println (x); return x}]
		r = x.product ('+', ap)
		println
		println (r)

	@out
		10

		2
		5
		9
		14

		14
*/
static int qu__container_product_callback (qu_any item, qu_any idx, qu_any *arg)
{
	qu_any     res   = arg[0];
	qu_any     op    = arg[1];
	qu_boolean binop = arg[2];
	qu_any     ap    = arg[3];

	if (res == 0)
	{
		res = item;
	}
	else
	{
		if (binop == QuTrue)
			res = ((qu_binop_f) op) (res, item);
		else
			res = qu_object_call_ex1 (res, op, item);

		if (res == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	if (ap)
	{
		res = qu_sys_map (res, ap);
		if (res == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			return QU_ITERATOR_CALLBACK_END;
		}
	}
	arg[0] = res;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_product_do (qu_container self, qu_any op, int argc, qu_any *argv)
{
	qu_any     ap = 0;
	qu_boolean binop = QuFalse;

	QU_ARG_VCOUNT (0, 1);

	if (qu_string_is (op))
	{
		qu_binop_f f = (qu_any) qu_sys_binop_get (op);
		if (f)
		{
			op = (qu_any) f;
			binop = QuTrue;
		}
	}

	if (argc > 0)
	{
		ap = argv[0];
		if (ap == QuNil)
			ap = 0;
	}

	/*begin*/
	{
		qu_any arg[4] = {0, op, binop, ap};

		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_product_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		QU_REQUIRE (QuENone, arg[0]);
		return arg[0];
	}
}


qu_any qu__container_product (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VMIN (1);
	return qu__container_product_do (self, argv[0], argc - 1, argv + 1);
}



/*) sub repeat (n: Integer, res = []) -> Any
	Push items in self into v{res} for v{n} times consecutively.

	v{res} must support the __push method.

	v{n} must be non-negative integer

	@xmp
		a = {1 2 3}
		b = a.repeat (3, [])
		println (b)

	@out
		[1, 2, 3, 1, 2, 3, 1, 2, 3]
*/
static int qu__container_repeat_callback (qu_any item, qu_any idx, qu_any res)
{
	if (qu_sys_push (res, item) == 0)
		return QU_ITERATOR_CALLBACK_ERROR;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_repeat (qu_container self, int argc, qu_any *argv)
{
	qu_any res;
	long   n;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong (0, n);
	QU_ARG_CHECK  (n >= 0);
	res = (argc > 1) ? argv[1] : (qu_any) qu_array_new (0);
	while (n-- > 0)
	{
		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_repeat_callback, res)
			!= QU_ITERATOR_CALLBACK_ERROR);
	}
	return res;
}



/*) sub rfilter (x, res = []) -> Any
	This does the reverse of #{filter} thus collect items not collected
	by f{filter}.

	@xmp
		map = [lambda (x) x.len > 3, "isalpha"]

		a = ['foo', 'fooey', 'foo123']
		r = a.rfilter (map)
		println (r)

	@out
		['foo', 'foo123']
*/
static int qu__container_rfilter_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any     flt = argv[0];
	qu_any     res = argv[1];
	qu_boolean r;

	r = qu_sys_filter (item, flt);
	if (r == 0)
	{
		if (qu_thread_err_is_estop ())
			return QU_ITERATOR_CALLBACK_END;
		return QU_ITERATOR_CALLBACK_ERROR;
	}
	if (r != QuTrue)
	{
		if (qu_sys_push (res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_rfilter (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	/*begin*/
	{
		qu_any arg[2] = {argv[0], (argc > 1) ? argv[1] : (qu_any) qu_array_new (0)};
		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_rfilter_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		return arg[1];
	}
}



/*) sub subset (x) -> Bool
	Return true if all items in self are found in v{x}, false otherwise.
	If one or both are empty then false is returned.

	v{x} must support the __each and __in method.

	@xmp
		a = {1 2 3 4}
		b = [5 6 7 4 3 1 2]
		println (a.subset (b))

	@out
		true
*/
static int qu__container_subset_callback (qu_any item, qu_any idx, qu_any x)
{
	qu_boolean r = qu_sys_in (x, item);
	if (r == 0)
		return QU_ITERATOR_CALLBACK_ERROR;
	if (r == QuFalse)
		return QU_ITERATOR_CALLBACK_RESULT;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_boolean qu__container_subset (qu_container self, qu_any x)
{
	int r = qu_iterator_do (self, (qu_itercallback_f) qu__container_subset_callback, x);
	QU_ENSURE (r != QU_ITERATOR_CALLBACK_ERROR);
	return (r == QU_ITERATOR_CALLBACK_RESULT) ? QuFalse : QuTrue;
}



/*) sub superset (x) -> Bool
	Return true if all items in k{x} are found in self, false otherwise.
	If one or both are empty then false is returned.

	v{x} must support the __in method.

	@xmp
		a = [1 2 3 4]
		b = {5 6 7 4 3 1 2}
		println (b.superset (a))

	@out
		true
*/
static qu_boolean qu__container_superset (qu_container self, qu_any x)
{
	QU_RETURN (qu__container_subset (x, self));
}



/*) sub sum (*arg) -> Any
	Calculate the sum of all items in self.

	This is just a convenience method that calls #{product}C{('+', *arg)}.

	@xmp
		a = [1, 2, 3]
		println (a.sum)

	@out
		6
*/
static qu_any qu__container_sum (qu_container self, int argc, qu_any *argv)
{
	return qu__container_product_do (self, qu_string_byte ('+'), argc, argv);
}



/*) sub union (x, res = {}) -> Any
	For all unique items in self and v{x} collect them in v{res}.

	@xmp
		a = 'abcd' as Set
		b = 'cdef' as Set
		r = a.union (b)
		println (r)

		a.union (b, @print)

	@out
		{'a', 'b', 'c', 'd', 'e', 'f'}
		abcdef
*/
static int qu__container_union_callback (qu_any item, qu_any idx, qu_any *argv)
{
	qu_any x = argv[0];
	qu_any r = argv[1];

	if (x)
	{
		qu_boolean b = qu_sys_in (x, item);
		if (b == QuTrue)
			return QU_ITERATOR_CALLBACK_CONTINUE;
	}
	if (qu_sys_push (r, item) == 0)
		return QU_ITERATOR_CALLBACK_ERROR;
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_union (qu_container self, int argc, qu_any *argv)
{
	qu_any x, r;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	r = (argc > 1) ? argv[1] : (qu_any) qu_set_new (0);
	/*begin*/
	{
		qu_any arg[2] = {0, r};

		QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_union_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		arg[0] = self;
		QU_ENSURE (qu_iterator_do (x, (qu_itercallback_f) qu__container_union_callback, arg)
			!= QU_ITERATOR_CALLBACK_ERROR);
		return r;
	}
}



/*) sub within (x, defval = nil) -> Any
	Match v{x} against items in self.

	If self is a Dict then match the keys and return the corresponding value.
	Otherwise return true or false.

	For Array, Dict, Set, Range, and Queue items, matching uses the k{in}
	operator. For others the k{==} operator is used.

	If there is no match then v{defval} is returned.

	@noexc

	@xmp
		d = [1..99:'foo', 100..200:'bar', 'baz':10]
		println (d.within (20))
		println (d.within ('baz'))

	@out
		foo
		10
*/
static int qu__container_within_callback (qu_any item, qu_any v, qu_any *argv)
{
	switch (qu_classid (item))
	{
		case QU_CLASS_ID_Array:
		case QU_CLASS_ID_Dict:
		case QU_CLASS_ID_Set:
		case QU_CLASS_ID_Range:
		case QU_CLASS_ID_Queue:
			if (qu_sys_in (item, argv[0]) == QuTrue)
			{
				if (!argv[1])
					argv[1] = v;
				return QU_ITERATOR_CALLBACK_RESULT;
			}
			break;

		default:
			if (qu_sys_eq (item, argv[0]) == QuTrue)
			{
				if (!argv[1])
					argv[1] = v;
				return QU_ITERATOR_CALLBACK_RESULT;
			}
			break;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_within (qu_container self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	{
		qu_any a[2] = {argv[0], qu_dict_is (self) ? (qu_any) 0 : (qu_any) QuTrue};

		if (qu_iterator_do (self, (qu_itercallback_f) qu__container_within_callback, a)
				== QU_ITERATOR_CALLBACK_RESULT)
			return a[1];
	}
	return (argc > 1) ? argv[1] : (qu_any) QuNil;
}



/*) sub without (x, res = {})
	Collect items in self in v{res} except those in v{x}.

	v{x} must support the __in method.

	@xmp
		a = {1 2 3 4}
		b = [1 4 5]
		r = a.without (b)
		println (r)

	@out
		{2, 3}
*/
static int qu__container_without_callback (qu_any item, qu_any ignore, qu_any *argv)
{
	qu_any     x   = argv[0];
	qu_any     res = argv[1];
	qu_boolean r;

	r = qu_sys_in (x, item);
	if (r == 0)
		return QU_ITERATOR_CALLBACK_ERROR;

	if (r == QuFalse)
	{
		if (qu_sys_push (res, item) == 0)
			return QU_ITERATOR_CALLBACK_ERROR;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}


static qu_any qu__container_without (qu_container self, int argc, qu_any *argv)
{
	qu_any arg [2];

	QU_ARG_VCOUNT (1, 2);
	arg[0] = argv[0];
	arg[1] = (argc > 1) ? argv[1] : (qu_any) qu_set_new (0);
	QU_ENSURE (qu_iterator_do (self, (qu_itercallback_f) qu__container_without_callback, arg)
		!= QU_ITERATOR_CALLBACK_ERROR);
	return arg[1];
}



/*-----------------------------------------------------------------------------
	specific instance methods - fix/freeze
-----------------------------------------------------------------------------*/

/*) sub fix -> self
	Make the Container permantently unresizable.

	@warn
	Once #{fixed} there is no way to unfix it.

	@exc
	.EModify  already frozen or fixed
*/
static qu_container qu__container_fix (qu_container self)
{
	QU_REQUIRE (QuEModify, !(qu_container_frozen (self) || qu_container_fixed (self)));
	qu_container_fix (self);
	return self;
}



/*) sub fixed -> Boolean
	Whether or not the Container is permantently unresizable.

	See #{fix}.

	@noexc
*/
static qu_boolean qu__container_fixed (qu_container self)
{
	return qu_container_fixed (self) ? QuTrue : QuFalse;
}



/*) sub freeze -> self
	Make the Container permantently immutable.

	Thus unlike #{protect}, this is permanent.

	@warn
	Once #{frozen} there is no way to unfreeze it.

	@exc
	.EModify  already frozen
*/
static qu_container qu__container_freeze (qu_container self)
{
	QU_REQUIRE (QuEModify, !qu_container_frozen (self));
	qu_container_freeze (self);
	return self;
}



/*) sub frozen -> Boolean
	Whether or not the Container is permantently immutable.

	See #{freeze}.

	@noexc
*/
static qu_boolean qu__container_frozen (qu_container self)
{
	return qu_container_frozen (self) ? QuTrue : QuFalse;
}



/*-----------------------------------------------------------------------------
	specific instance methods - protect
-----------------------------------------------------------------------------*/
#ifdef QU_CONTAINER_WITH_PROTECT

/*) sub protect -> self
	Make the Container temporarily immutable.

	Thus unlike #{freeze}, this is not permanent.

	You can #{unprotect} the Container but only within the same I{scope} that
	calls this method. The I{scope} is a -{Class} if the caller is within a Class,
	a Module otherwise.

	If the I{scope} is a Module then only code at module level can unprotect the
	Container. If the I{scope} is a Class then only code within that Class can
	unprotect it, that is, not even subclasses can.

	@exc
	.EModify  already protected

	@xmp
		Ary = [1, 2, 3]

		class Foo

			sub class protect
				return Ary.protect     # protected from within Foo
			;;

			sub class unprotect
				Ary.unprotect          # only Foo can unprotect it
			;;
		;;

		Foo.protect
		println (Ary.protector)

		try
			Ary[0] = 456
		catch
			println (__err__)
		;;

		try
			Ary.unprotect
		catch
			println (__err__)
		;;

		Foo.unprotect
		Ary[0] = 789
		println (Ary)

		Ary.protect                   # protected from within module
		try
			Ary[0] = 456
		catch
			println (__err__)
		;;
		Ary.unprotect                 # only this module can unprotect it

	@out
		<Class __main__.Foo>
		(EModify) can not modify object
		(EModify) invalid protection scope
		[789, 2, 3]
		(EModify) can not modify object
*/
static qu_container qu__container_protect (qu_container self)
{
	QU_REQUIRE (QuEModify, self->scope == 0);
	self->scope = qu_interp_scope ();
	qu_flag_set (self, QU_CONTAINER_FLAG_PROTECT);
	return self;
}



/*) sub protector
	Return the Module or Class that called #{protect}.

	If none did then nil is returned.

	See #{protect} for more information.

	@noexc
*/
static qu_any qu__container_protector (qu_container self)
{
	if (self->scope == 0)
		return QuNil;
	return self->scope;
}



/*) sub unprotect -> self
	Unprotect Container.

	If the Container is not protected then nothing happens.

	You can only unprotect the Container from within the same I{scope}
	that called the f{protect} method.

	See #{protect} for more information.

	@exc
	.EModify invalid protection scope
*/
static qu_container qu__container_unprotect (qu_container self)
{
	if (self->scope)
	{
		qu_any x = qu_interp_scope ();
		QU_REQUIRE (QuEModify, x == self->scope);
		qu_flag_unset (self, QU_CONTAINER_FLAG_PROTECT);
		self->scope = 0;
	}
	return self;
}



#endif /*QU_CONTAINER_WITH_PROTECT*/

/*-----------------------------------------------------------------------------
	interface
-----------------------------------------------------------------------------*/

/*
	Get a nice size.
*/
long qu_container_sizer (long cap)
{
	assert (cap >= 0);

	if (cap >= QU_GC_MAX_ITEMS / 2)
		return cap;
	else if (cap < QU_CONTAINER_SIZE_MIN)
		return QU_CONTAINER_SIZE_MIN;
	else
	{
		register long c = 1;
		while (c < cap)
			c <<= 1;
		return c;
	}
}


/*
	Get a nice mask value for a requested capacity.
*/
long qu_container_mask_sizer (long cap)
{
	assert (cap <= QU_GC_MAX_ITEMS);

	if (cap > QU_INT_MAX / 2)
		return QU_CONTAINER_MASK_MAX;
	else
	{
		register long c = QU_CONTAINER_MASK_MIN + 1;
		register long m = cap * 3;

		while  (c < m)
			c <<= 1;
		return c - 1;
	}
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_container__setup ()
{
	qu_class_setup_abstract (QuContainerClass, qu_container_s);
	qu_flag_set (QuContainerClass, QU_CLASS_FLAG_FINAL);
}



void qu_container__init ()
{
	/* standard class methods */
	/* none */

	/* specific class methods */
	/* none */

	/* standard instance methods */
	qu_class_method_inst  (QuContainerClass, QU_NAME__len , qu__container__len , 0);
	qu_class_alias        (QuContainerClass, QU_NAME__len , QU_NAME_len);
	qu_class_method_inst  (QuContainerClass, QU_NAME__true, qu__container__true, 0);

	/* specific instance methods */
	qu_class_method_instz (QuContainerClass, "apply"     , qu__container_apply     ,  1);
	qu_class_method_instz (QuContainerClass, "count"     , qu__container_count     , -1);
	qu_class_method_instz (QuContainerClass, "cross"     , qu__container_cross     ,  1);
	qu_class_method_instz (QuContainerClass, "diff"      , qu__container_diff      , -1);
	qu_class_method_instz (QuContainerClass, "every"     , qu__container_every     ,  1);
	qu_class_method_instz (QuContainerClass, "fill"      , qu__container_fill      ,  1);
	qu_class_method_instz (QuContainerClass, "filter"    , qu__container_filter    , -1);
	qu_class_method_instz (QuContainerClass, "grab"      , qu__container_grab      , -1);
	qu_class_method_instz (QuContainerClass, "intersect" , qu__container_intersect , -1);
	qu_class_method_instz (QuContainerClass, "istrue"    , qu__container_istrue    , -1);
	qu_class_method_instz (QuContainerClass, "isfalse"   , qu__container_isfalse   , -1);
	qu_class_method_instz (QuContainerClass, "match"     , qu__container_match     , -1);
	qu_class_method_instz (QuContainerClass, "map"       , qu__container_map       , -1);
	qu_class_method_instz (QuContainerClass, "max"       , qu__container_max       , -1);
	qu_class_method_instz (QuContainerClass, "merge"     , qu__container_merge     ,  1);
	qu_class_method_instz (QuContainerClass, "mergex"    , qu__container_mergex    ,  1);
	qu_class_method_instz (QuContainerClass, "min"       , qu__container_min       , -1);
	qu_class_method_instz (QuContainerClass, "move"      , qu__container_move      ,  1);
	qu_class_method_instz (QuContainerClass, "pair"      , qu__container_pair      , -1);
	qu_class_method_instz (QuContainerClass, "product"   , qu__container_product   , -1);
	qu_class_method_instz (QuContainerClass, "repeat"    , qu__container_repeat    , -1);
	qu_class_method_instz (QuContainerClass, "rfilter"   , qu__container_rfilter   , -1);
	qu_class_method_instz (QuContainerClass, "subset"    , qu__container_subset    ,  1);
	qu_class_method_instz (QuContainerClass, "superset"  , qu__container_superset  ,  1);
	qu_class_method_instz (QuContainerClass, "sum"       , qu__container_sum       , -1);
	qu_class_method_instz (QuContainerClass, "union"     , qu__container_union     , -1);
	qu_class_method_instz (QuContainerClass, "within"    , qu__container_within    , -1);
	qu_class_method_instz (QuContainerClass, "without"   , qu__container_without   , -1);

	qu_class_method_instz (QuContainerClass, "fix"       , qu__container_fix       ,  0);
	qu_class_method_instz (QuContainerClass, "fixed"     , qu__container_fixed     ,  0);
	qu_class_method_instz (QuContainerClass, "freeze"    , qu__container_freeze    ,  0);
	qu_class_method_instz (QuContainerClass, "frozen"    , qu__container_frozen    ,  0);

#ifdef QU_CONTAINER_WITH_PROTECT
	qu_class_method_instz (QuContainerClass, "protect"   , qu__container_protect   ,  0);
	qu_class_method_instz (QuContainerClass, "protector" , qu__container_protector ,  0);
	qu_class_method_instz (QuContainerClass, "unprotect" , qu__container_unprotect ,  0);
#endif

	/* finalize declaration */
	qu_class_enddecl  (QuContainerClass);
}
