/* qu_decimal.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Decimal is Number __abstract
	This is the parent class of the -{Float} and -{Big} classes.

	@tree
	.-{Number}
	.	Decimal
	.		-{Float} system native double precision
	.		-{Big}   arbitrary size floating point

	@xmp
		i = 1.2
		println (i is Decimal)

	@out
		true

	@p
	The Decimal class provides a convenient way to convert an object into
	a Float or a Big via the #{new] method.

	@xmp
		d = Decimal ('1.2')
		println (d, ' ', d.__class.__name)

		d = Decimal (12.34)
		println (d, ' ', d.__class.__name)

	@out
		1.2 Float
		12.34 Float
*/
qu_class QuDecimalClass;


/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

#define UNOP(name, rtype) \
rtype qu_decimal_##name (qu_decimal x) \
{ \
	return qu_float_is (x) \
		? (rtype) qu_float_##name ((qu_float) x) \
		: (rtype) qu_big_##name   ((qu_big) x); \
}


#define BINOP_BASIC(name, rtype, ytype) \
rtype qu_decimal_##name##_##ytype (qu_decimal x, qu_##ytype y) \
{ \
	return qu_float_is (x) \
		? (rtype) qu_float_##name##_##ytype ((qu_float) x, y) \
		: (rtype) qu_big_##name##_##ytype   ((qu_big)   x, y); \
}

#define BINOP_int(name, rtype)      BINOP_BASIC (name, rtype, int)
#define BINOP_long(name, rtype)     BINOP_BASIC (name, rtype, long)
#define BINOP_rational(name, rtype) BINOP_BASIC (name, rtype, rational)
#define BINOP_float(name, rtype)    BINOP_BASIC (name, rtype, float)
#define BINOP_big(name, rtype)      BINOP_BASIC (name, rtype, big)
#define BINOP_complex(name, rtype)  BINOP_BASIC (name, rtype, complex)
#define BINOP_inf(name, rtype)      BINOP_BASIC (name, rtype, inf)

#define BINOP_integer(name, rtype) \
rtype qu_decimal_##name##_integer (qu_decimal x, qu_integer y) \
{ \
	return qu_float_is (x) \
		? (qu_int_is (y) \
			? (rtype) qu_float_##name##_int  ((qu_float) x, (qu_int) y) \
			: (rtype) qu_float_##name##_long ((qu_float) x, (qu_long) y)) \
		: (qu_int_is (y) \
			? (rtype) qu_big_##name##_int  ((qu_big) x, (qu_int) y) \
			: (rtype) qu_big_##name##_long ((qu_big) x, (qu_long) y)); \
}

#define BINOP_decimal(name, rtype) \
rtype qu_decimal_##name##_decimal (qu_decimal x, qu_decimal y) \
{ \
	return qu_float_is (x) \
		? (qu_float_is (y) \
			? (rtype) qu_float_##name##_float ((qu_float) x, (qu_float) y) \
			: (rtype) qu_float_##name##_big   ((qu_float) x, (qu_big) y)) \
		: (qu_float_is (y) \
			? (rtype) qu_big_##name##_float ((qu_big) x, (qu_float) y) \
			: (rtype) qu_big_##name##_big   ((qu_big) x, (qu_big) y)); \
}

#define BINOP_real(name, rtype) \
	QU_NUMBER_BINOP_real (qu_decimal, name, rtype)

#define BINOP_number(name, rtype) \
	QU_NUMBER_BINOP_number (qu_decimal, name, rtype)



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (x) -> Decimal
	You can use this method to convert v{x} into a Decimal number, that is,
	a -{Float} or a -{Big}.

	It simply return the value returned by C{x.__as (Decimal)} provided that
	it is a Decimal number. Otherwise EImplement is thrown.

	This method will try to get a Float if possible, a Big otherwise.
	If, however, v{x} is a Big, then this method does not try to convert it
	into a Float.

	If you specifically want a Big or Float, then say C{Big (x)} or C{Float (x)}
	instead.

	@exc
	.ESyntax
	.EImplement

	@xmp
		r = Decimal ('123')
		println (r ' ' r.__class.__name)

		r = Decimal (1 _/ 3)
		println (r ' ' r.__class.__name)

	@out
		123.0 Float
		0.333333 Float
*/
static qu_decimal qu__decimal_new (qu_class self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      
			return (qu_decimal) qu_int_as_decimal ((qu_int) x);

		case QU_CLASS_ID_Long:     
			return (qu_decimal) qu_long_as_decimal ((qu_long) x);

		case QU_CLASS_ID_Rational: 
			return (qu_decimal) qu_rational_as_decimal ((qu_rational) x);

		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Big:
			return (qu_decimal) x;

		case QU_CLASS_ID_Boolean:
			return (qu_decimal) ((qu_sys_true (x) == QuTrue) ? QuFloat1 : QuFloat0);

		case QU_CLASS_ID_String:
			x = qu_number_scan_string (x, QU_NUMBER_SCAN_SKIPSPACE);
			QU_REQUIREZ (QuESyntax, x && qu_decimal_is (x), "Number (String) invalid syntax");
			return (qu_decimal) x;
	}
	return qu_object_as (x, QuDecimalClass);
}



#ifdef HAVE_LONG_DOUBLE
qu_decimal qu_decimal_newld (long double v)
{
	assert (qu_longdouble_isfinite (v));
	if (-DBL_MAX <= v && v <= DBL_MAX)
		return qu_decimal_newd (v);
	return (qu_decimal) qu_big_newld (v);
}

qu_real qu_decimal_newldx (long double v)
{
	if (qu_longdouble_isfinite (v))
		return (qu_real) qu_decimal_newld (v);
	QU_REQUIREZ (QuEValue, !qu_longdouble_isnan (v), "Decimal (long double NaN)");
	return (qu_real) __qu_inf__ ((v < 0) ? -1 : 1);
}
#endif



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (sign   , int)
UNOP (iszero , int)
UNOP (isneg  , int)
UNOP (isone  , int)
UNOP (is1    , int)
UNOP (iswhole, int)
UNOP (isfrac , int)
UNOP (isodd  , int)

UNOP (neg , qu_decimal)
UNOP (incr, qu_real)
UNOP (decr, qu_real)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#define TOC(name, rtype) \
rtype qu_decimal_to##name (qu_any x) \
{ \
	errno = 0; \
	if (qu_float_is (x)) \
		return qu_float_to##name ((qu_float) x); \
	else if (qu_big_is (x)) \
		return qu_big_to##name ((qu_big) x); \
	errno = EDOM; \
	return 0; \
}


TOC (l  , long)
TOC (ul , unsigned long)
TOC (f  , float)
TOC (d  , double)
#ifdef HAVE_LONG_LONG
TOC (ll , long long)
TOC (ull, unsigned long long)
#endif
#ifdef HAVE_LONG_DOUBLE
TOC (ld , long double)
#endif

#undef TOC


#define AS(type) \
qu_##type qu_decimal_as_##type (qu_decimal x) \
{ \
	if (qu_float_is (x)) \
	{ \
		return qu_float_as_##type ((qu_float) x); \
	} \
	else \
	{ \
		assert (qu_big_is (x)); \
		return qu_big_as_##type ((qu_big) x); \
	} \
}


AS (int)
AS (long)
AS (rational)
AS (float)
AS (big)
AS (integer)

#undef AS


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (eq, int)
BINOP_long     (eq, int)
BINOP_rational (eq, int)
BINOP_float    (eq, int)
BINOP_big      (eq, int)
BINOP_complex  (eq, int)
BINOP_inf      (eq, int)
BINOP_integer  (eq, int)
BINOP_decimal  (eq, int)
BINOP_real     (eq, int)
BINOP_number   (eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (like, int)
BINOP_long     (like, int)
BINOP_rational (like, int)
BINOP_float    (like, int)
BINOP_big      (like, int)
BINOP_complex  (like, int)
BINOP_inf      (like, int)
BINOP_integer  (like, int)
BINOP_decimal  (like, int)
BINOP_real     (like, int)
BINOP_number   (like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (cmp, int)
BINOP_long     (cmp, int)
BINOP_rational (cmp, int)
BINOP_float    (cmp, int)
BINOP_big      (cmp, int)
BINOP_complex  (cmp, int)
BINOP_inf      (cmp, int)
BINOP_integer  (cmp, int)
BINOP_decimal  (cmp, int)
BINOP_real     (cmp, int)
BINOP_number   (cmp, int)

/*
	|<=>|
*/
BINOP_int      (cmpabs, int)
BINOP_long     (cmpabs, int)
BINOP_rational (cmpabs, int)
BINOP_float    (cmpabs, int)
BINOP_big      (cmpabs, int)
BINOP_complex  (cmpabs, int)
BINOP_inf      (cmpabs, int)
BINOP_integer  (cmpabs, int)
BINOP_decimal  (cmpabs, int)
BINOP_real     (cmpabs, int)
BINOP_number   (cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (plus, qu_real)
BINOP_long     (plus, qu_real)
BINOP_rational (plus, qu_real)
BINOP_float    (plus, qu_real)
BINOP_big      (plus, qu_real)
BINOP_complex  (plus, qu_complex)
BINOP_inf      (plus, qu_real)
BINOP_integer  (plus, qu_real)
BINOP_decimal  (plus, qu_real)
BINOP_real     (plus, qu_real)
BINOP_number   (plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (minus, qu_real)
BINOP_long     (minus, qu_real)
BINOP_rational (minus, qu_real)
BINOP_float    (minus, qu_real)
BINOP_big      (minus, qu_real)
BINOP_complex  (minus, qu_complex)
BINOP_inf      (minus, qu_real)
BINOP_integer  (minus, qu_real)
BINOP_decimal  (minus, qu_real)
BINOP_real     (minus, qu_real)
BINOP_number   (minus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (mul, qu_real)
BINOP_long     (mul, qu_real)
BINOP_rational (mul, qu_real)
BINOP_float    (mul, qu_real)
BINOP_big      (mul, qu_real)
BINOP_complex  (mul, qu_complex)
BINOP_inf      (mul, qu_real)
BINOP_integer  (mul, qu_real)
BINOP_decimal  (mul, qu_real)
BINOP_real     (mul, qu_real)
BINOP_number   (mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (div, qu_real)
BINOP_long     (div, qu_real)
BINOP_rational (div, qu_real)
BINOP_float    (div, qu_real)
BINOP_big      (div, qu_real)
BINOP_complex  (div, qu_complex)
BINOP_inf      (div, qu_real)
BINOP_integer  (div, qu_real)
BINOP_decimal  (div, qu_real)
BINOP_real     (div, qu_real)
BINOP_number   (div, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (idiv, qu_integer)
BINOP_long     (idiv, qu_integer)
BINOP_rational (idiv, qu_integer)
BINOP_float    (idiv, qu_integer)
BINOP_big      (idiv, qu_integer)
BINOP_complex  (idiv, qu_integer)
BINOP_inf      (idiv, qu_integer)
BINOP_integer  (idiv, qu_integer)
BINOP_decimal  (idiv, qu_integer)
BINOP_real     (idiv, qu_integer)
BINOP_number   (idiv, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (rem, qu_real)
BINOP_long     (rem, qu_real)
BINOP_rational (rem, qu_real)
BINOP_float    (rem, qu_real)
BINOP_big      (rem, qu_real)
BINOP_complex  (rem, qu_complex)
BINOP_inf      (rem, qu_decimal)
BINOP_integer  (rem, qu_real)
BINOP_decimal  (rem, qu_real)
BINOP_real     (rem, qu_real)
BINOP_number   (rem, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (pow, qu_real)
BINOP_long     (pow, qu_real)
BINOP_rational (pow, qu_number)
BINOP_float    (pow, qu_number)
BINOP_big      (pow, qu_number)
BINOP_complex  (pow, qu_complex)
BINOP_inf      (pow, qu_real)
BINOP_integer  (pow, qu_number)
BINOP_decimal  (pow, qu_number)
BINOP_real     (pow, qu_number)
BINOP_number   (pow, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_string qu_decimal_format (qu_decimal x, qu_format_s *f)
{
	if (qu_float_is (x))
		return qu_float_format ((qu_float) x, f);
	assert (qu_big_is (x));
	return qu_big_format ((qu_big) x, f);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_decimal__setup ()
{
	qu_class_setup_base (QuDecimalClass, qu_object_s);
}



void qu_decimal__init ()
{
	/* standard class methods */
	qu_class_method_class (QuDecimalClass, QU_NAME_new, qu__decimal_new, 1);

	/* finalize declaration */
	qu_class_set_super (QuDecimalClass, QuRealClass);
	qu_class_enddecl   (QuDecimalClass);
}
