/* Dict.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Dict
	A Dict object is a container of key/value pairs. The keys are unique but
	not the values. Both keys and values may be any object, i.e. the keys are not
	necessarily strings.

	To create a Dict you use the syntax or the #{new} method. The syntax is
	a bit unusual. If the U{first key is an indentifier} then it is interpreted
	as a string and so are the rest of the keys. Otherwise they are interpreted
	as expressions.

	All of the following lines produce the same object.
	@prog
		d = [foo:0, bar:2, baz:3 foo:1]
		d = ['foo':1, 'bar':2, 'baz':3]
		x = 'bar'; d = ['foo':1, x:2, 'baz':3]
		d = Dict ('foo', 1, 'bar', 2, 'baz', 3)

	@p
	To force the compiler to read the keys as expressions you can do this:
	@prog
		foo = 'a'
		bar = 'b'
		d = [: foo:1, bar:2]    # same as d = [a:1 b:2]
		d = Dict (foo 1, bar 2) # same thing

	@p
	EKey is thrown by methods requiring a key but the Dict does not contain
	the key. For example, you say C{MyDict['foo']} when the Dict does not
	contain the key C{'foo'}.

	You can make a Dict immutable with the #{_freeze} method.
	You can also fix the keys with the #{_fix} method, which means that you
	can only change the values later on.
	@prog
		x = [foo:1, bar:2]._fix

	@warn
	Once frozen there is no way to unfreeze it.
	Once fixed there is no way to unfix it.

	@p
	Immutable here means that the Dict is guaranteed to contain the same
	objects at any time. Those objects, however, do not necessarily have
	to be immutable. They may be container objects with variable items.

	You can also create frozen Dict objects using the k{const} syntax.
	@prog
		x = const [name:'Foo', age:99, id:123]

	@p
	The k{const} syntax is not the same as the #{_freeze} method.
	A literal object is created at compile time once only. Thus the above code
	will simply assign the literal to the variable. Note however that all keys
	and values must be literals, i.e. strings, numbers, references, or anything
	defined as literals by the compiler.

	The following does not create a new Dict.
	It refers to a predefined system Dict.
	This is useful to intialize variables or properties.
	@prog
		var x = const [:]

	@p
	To ensure that all keys or values are of a specified type, use the k{of} keyword.
	To find out the type, use the #{type} method.

	@prog
		d = [a:1, b:2, c:3] of (String, Int)
		println (d.type)

	@out
		[<Class Sys.String>, <Class Sys.Int>]
*/
qu_class QuDictClass;


/* predefined frozen empty Dict */
qu_dict QuDictEMPTY;


/* an entry in the bucket */
#define item_s qu_dictentry_s


/*----------------------------------------------------------------------------
	macros
----------------------------------------------------------------------------*/

/*
	The string used to represent an empty and recursive qu_dict item.
*/
#define DICT_EMPTY_STR     "[:]"
#define DICT_EMPTY_STRLEN  3
#define DICT_RECUR_STR     "[::]"
#define DICT_RECUR_STRLEN  4


/*----------------------------------------------------------------------------
	sizing
----------------------------------------------------------------------------*/

static void _dict_resize (qu_dict self, long newmask)
{
	register item_s **b, *i, *s = &self->sentinel;
	register long     k;

	b = qu_mem_calloc (newmask + 1, sizeof (item_s *));
	for (i = s->next; i != s; i = i->next)
	{
		k = i->hash & newmask;
		i->right = b [k];
		b [k] = i;
	}
	qu_mem_free (self->buckets);
	self->buckets = b;
	self->mask = newmask;
}



/*----------------------------------------------------------------------------
	allocate
----------------------------------------------------------------------------*/

/*
	initialize buckets
*/
#define CLEAR_BUCKETS(self) \
	memset (self->buckets, 0, (self->mask + 1) * sizeof (item_s *))



/*
	Free all items but retain the buckets.
*/
static void _dict_free_items (qu_dict self)
{
	item_s *s = &self->sentinel;
	item_s *i, *x;
	for (i = s->next; i != s; i = x)
	{
		x = i->next;
		qu_mem_free (i);
	}
	s->prev = s->next = s;
}



/*
	Initialize or reinitialize a Dict.
*/
static qu_dict _dict_setup (register qu_dict self, long cap)
{
	long    m;
	item_s *i = &self->sentinel;

	assert (0 <= cap && cap <= QU_GC_MAX_ITEMS);

	m = qu_container_mask_sizer (cap);

	self->buckets = qu_mem_calloc (m + 1, sizeof (item_s *));
	self->mask    = m;
	self->len     = 0;
	self->ktype   = 0;
	self->vtype   = 0;

	i->prev = i->next = i;

	QU_CONTAINER_INIT (self);
	return self;
}



/*
	Create a regular GC Dict.
*/
qu_dict qu_dict_new (long cap)
{
	return _dict_setup (qu_gc_new (QuDictClass), cap);
}



/*
	Create a root GC Dict.
*/
qu_dict qu_dict_new_root (long cap)
{
	return _dict_setup (qu_gc_new_root (QuDictClass), cap);
}



/*
	Create a system static Dict.
*/
qu_dict qu_dict_new_static (long cap)
{
	return _dict_setup (qu_gc_new_static (QuDictClass), cap);
}



/*
	For the interpreter.
	To create a Dict from objects on the stack.
*/
qu_dict qu_dict_new_argv (int argc, qu_any *argv)
{
	return qu__dict_new (QuDictClass, argc, argv);
}



/*
	For the interpreter.
	Create an Arg object to pass named arguments to functions/methods.
*/
qu_dict qu_arg_new (int argc, qu_any *argv)
{
	return qu__dict_new (QuArgClass, argc, argv);
}



/*
	Create a static Dict from an existing one.
*/
qu_dict qu_dict_to_static (qu_dict self)
{
	qu_dict d;
	long    i, j;
	qu_any  k, v;

	if (qu_gc_is_static (self))
		return self;

	i = j = 0;
	while ((v = qu_dict_walk (self, &k, &i, &j)) != 0)
	{
		if (!qu_gc_is_static (k) || !qu_gc_is_static (v))
			return 0;
	}

	d = qu_dict_new_static (self->len);
	i = j = 0;
	while ((v = qu_dict_walk (self, &k, &i, &j)) != 0)
		qu_dict_setitem (d, qu_object_to_static (k), qu_object_to_static (v));
	qu_dict_freeze (d);
	return d;
}



/*----------------------------------------------------------------------------
	find
----------------------------------------------------------------------------*/

/*
	Find an entry by specifying a key object.
	!!! Dict must be protected since qu_sys_eq might go to program level.
*/
static item_s *_dict_lookup (qu_dict self, qu_any key, long h)
{
	register item_s *i;

	for (i = self->buckets[h & self->mask]; i; i = i->right)
	{
		/*
			This must be i->key == key, not key == i->key, i.e. the __eq
			method to be called must be the one of the item in the qu_dict.
			Not of the object that is looked up.
		*/
		if (qu_sys_eq (i->key, key) == QuTrue)
			return i;
	}
	return 0;
}



/*
	Find an entry by specifying a C string key.
	Note that unlike _dict_lookup, this is an atomic operation.
*/
static item_s *_dict_lookupz (qu_dict self, const char *key, long len, long h)
{
	register item_s    *i;
	register qu_string  k;

	for (i = self->buckets[h & self->mask]; i; i = i->right)
	{
		k = (qu_string) i->key;
		if (qu_string_is (k) && k->len == len && memcmp (k->str, key, len) == 0)
			return i;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	include / exclude
----------------------------------------------------------------------------*/

/*
	Add entry to the Dict. No questions asked!

	This function add the object without checking whether or not the same
	key is already in the Dict. The caller must do that by itself to make
	sure the Dict contains unique keys.
*/
static void _dict_include (qu_dict self, qu_any key, qu_any val, long h)
{
	item_s **b = self->buckets, *s = &self->sentinel, *i;
	long     m = self->mask;
	long     k = h & m;

	assert (0 <= self->len && self->len < QU_GC_MAX_ITEMS);

	i        = qu_mem_alloc (sizeof (item_s));
	i->key   = key;
	i->value = val;
	i->prev  = s->prev;
	i->next  = s;
	i->hash  = h;
	i->right = b[k];

	b[k] = i;

	s->prev->next = i;
	s->prev = i;

	self->len++;

	m = qu_container_mask_grow (m, self->len);
	if (m)
		_dict_resize (self, m);
}



/*
	Remove an entry from the qu_dict.
	!!! Dict must be protected since qu_sys_eq might go to program level.
*/
static void _dict_exclude (qu_dict self, qu_any key, long h)
{
	register item_s *i, *x, *p = 0;
	long m = self->mask;

	for (i = self->buckets[h & m]; i; p = i, i = x)
	{
		x = i->right;
		if (qu_sys_eq (i->key, key) == QuTrue)
		{
			if (p)
				p->right = x;
			else
				self->buckets[h & m] = x;

			i->prev->next = i->next;
			i->next->prev = i->prev;

			qu_mem_free (i);

			self->len--;

			m = qu_container_mask_shrink (m, self->len);
			if (m)
				_dict_resize (self, m);
			return;
		}
	}
}



/*
	Remove an arbitrary entry.
*/
void qu_dict_del_any (qu_dict self)
{
	if (self->len)
	{
		register item_s *i;
		register long h, m = self->mask;

		i = (&self->sentinel)->next;
		h = i->hash;
		i = self->buckets [h & m];
		self->buckets [h & m] = i->right;
		i->prev->next = i->next;
		i->next->prev = i->prev;
		qu_mem_free (i);
		self->len--;
	}
}



/*----------------------------------------------------------------------------
	type
----------------------------------------------------------------------------*/

static int _dict_keyof (qu_dict self, qu_any ktype)
{
	if (self->ktype)
	{
		if (qu_object_compat_type (self->ktype, ktype))
			return 1;
	}
	else if (self->len)
	{
		long b = 0, i = 0;
		qu_any  k;

		while (qu_dict_walk (self, &k, &b, &i) != 0)
		{
			if (!qu_object_validate (self, ktype, k))
				return 0;
		}
		return 1;
	}
	return 0;
}



static int _dict_valueof (qu_dict self, qu_any vtype)
{
	if (self->vtype)
	{
		if (qu_object_compat_type (self->vtype, vtype))
			return 1;
	}
	else if (self->len)
	{
		long b = 0, i = 0;
		qu_any  k, v;

		while ((v = qu_dict_walk (self, &k, &b, &i)) != 0)
		{
			if (!qu_object_validate (self, vtype, v))
				return 0;
		}
		return 1;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (*arg) -> Array
	Create a new Dict with v{arg} as the items.

	v{arg} must be pairs of key and value.

	If called without any arguments then a new empty Dict is returned.

	@exc
	.EArgCount must be even
*/
static qu_dict qu__dict__inst (qu_class self, int argc, qu_any *argv)
{
	if (argc == 1)
	{
		qu_any x = argv[0];

		switch (qu_classid (x))
		{
			case QU_CLASS_ID_Array: return qu_array_as_dict ((qu_array) x);
			case QU_CLASS_ID_Set:   return qu_set_as_dict ((qu_set) x);
		}
		QU_ARG_THROW;
	}
	else
	{
		qu_dict d;

		QU_REQUIRE (QuEArgCount, (argc & 1) == 0);

		d = qu_dict_new (argc);
		assert (self->size >= QuDictClass->size);
		d->klass = self;
		while (argc--)
		{
			_dict_include (d, argv[0], argv[1], qu_object_hash (argv[0]));
			--argc;
			++argv;
			++argv;
		}
		return d;
	}
}



/*) sub class new (*arg) -> Dict
	Create a new Dict with v{arg} as the items.

	This method calls #{__inst} then f{__init} if one is defined.

	@exc
	.EArgCount must be even

	@xmp
		d = Dict ('foo', 1, 'bar', 2, 'baz', 3)
		# same as d = ['foo':1, 'bar':2, 'baz':3]
*/
qu_dict qu__dict_new (qu_class self, int argc, qu_any *argv)
{
	qu_dict d = qu__dict__inst (self, argc, argv);

	QU_ENSURE (d);
	return qu_class_init_new (self, d, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__dict__destroy (qu_dict self)
{
	_dict_free_items (self);
	qu_mem_free (self->buckets);
}



static void qu__dict__mark (qu_dict self)
{
	if (self->ktype != QuIntClass)
	{
		item_s *i, *s = &self->sentinel;

		for (i = s->next; i != s; i = i->next)
			qu_gc_mark (i->key);

		qu_gc_xmark (self->ktype);
	}
	if (self->vtype != QuIntClass)
	{
		item_s *i, *s = &self->sentinel;

		for (i = s->next; i != s; i = i->next)
			qu_gc_mark (i->value);

		qu_gc_xmark (self->vtype);
	}
	QU_CONTAINER_MARK (self);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be a one of the following -{Class} objects:
	@ops
	.| -{Array}   | Return an Array containing flattened keys and values.
	.| -{Set}     | Return a Set containing the values.
	.| -{String}  | Delegate to the #{__string} method.
	.| -{Boolean} | Delegate to the #{__true} method.
	.| -{Dict}    | Return self.

	@exc
	.EArgValue

	@xmp
		x = [foo:3, bar:2, baz:3]
		y = x as Array
		z = y as Dict
		println (y)
		println (z)
		println (x == z)
		println (x === z)

		y = x as Set
		println (y)

	@out
		['foo', 3, 'bar', 2, 'baz', 3]
		['foo':3, 'bar':2, 'baz':3]
		false
		true
		{2, 3}
*/
static qu_any qu__dict__as (qu_dict self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return qu_dict_as_array   (self);
		case QU_CLASS_ID_Set:     return qu_dict_as_set     (self);
		case QU_CLASS_ID_String:  return qu_dict_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_dict_as_boolean (self);
		case QU_CLASS_ID_Dict:    return self;
	}
	QU_ARG_THROW;
}



qu_array qu_dict_as_array (qu_dict self)
{
	qu_array  a;
	qu_any   *p;
	item_s   *s, *i;
	long      l = self->len;

	QU_MEM_CHECK (l <= QU_GC_MAX_ITEMS / 2);

	l *= 2;
	a = qu_array_new (l);
	a->len = l;
	p = a->objs;
	s = &self->sentinel;
	for (i = s->next; i != s; i = i->next)
	{
		*p++ = i->key;
		*p++ = i->value;
	}
	return a;
}



qu_set qu_dict_as_set (qu_dict self)
{
	qu_set s;
	qu_any k, v;
	long   b = 0, i = 0;

	assert (self->len <= QU_GC_MAX_ITEMS);

	s = qu_set_new (self->len);
	while ((v = qu_dict_walk (self, &k, &b, &i)) != 0)
		qu_set_push (s, v);
	return s;
}



/*) sub __clear -> self
	Delete all entries.

	This is an atomic operation.
	On success, the Dict is guaranteed to be empty after a call to this method.

	@exc
	.EModify

	@xmp
		d = [foo:1 bar:2]
		d.__clear
		println (d)
		println (d.len)

	@out
		[:]
		0
*/
qu_dict qu__dict__clear (qu_dict self)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	if (self->len)
	{
		_dict_free_items (self);
		CLEAR_BUCKETS (self);
		self->len = 0;
	}
	return self;
}



/*) sub clear -> self
	This is an alias of #{__clear}.
*/



/*) sub __copy -> Dict
	Make a shallow copy of self.

	@noexc

	@warn
	The #{type}, #{frozen} and #{fixed} properties are not inherited.
	You have to do it yourself as necessary.

	@xmp
		x = [foo:1 bar:2 baz: 3]
		y = x.__copy
		println (y)
		println (x == y)
		println (x === y)

	@out
		['foo':1, 'bar':2, 'baz':3]
		false
		true
*/
qu_dict qu__dict__copy (qu_dict self) /*atomic*/
{
	item_s *s = &self->sentinel, *i, **u, *e, *j;
	long    m = self->mask, h;
	qu_dict d;

	d           = qu_gc__new (self->klass, self->klass->size);
	d->buckets  = u = qu_mem_calloc (m + 1, sizeof (item_s *));
	d->mask     = m;
	d->len      = self->len;
	d->ktype    = 0;
	d->vtype    = 0;

	QU_CONTAINER_INIT (d);

	e = &d->sentinel;
	e->prev = e->next = e;

	for (i = s->next; i != s; i = i->next)
	{
		j        = qu_mem_alloc (sizeof (item_s));
		j->key   = i->key;
		j->value = i->value;
		j->hash  = h = i->hash;
		j->prev  = e->prev;
		j->next  = e;

		h &= m;
		j->right = u[h];
		u[h] = j;

		e->prev->next = j;
		e->prev = j;
	}
	return d;
}



/*) sub __del (key) -> self
	Remove v{key}.

	If the key does not exist then nothing happens.

	The Dict is temporarily I{protected} while this method is progress.

	@exc
	.EModify

	@xmp
		d = ['foo':1 'bar':2]
		d.delete ('foo')
		println (d)

	@out
		['bar':2]
*/
static qu_dict qu__dict__del (qu_dict self, qu_any key)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
  __qu_container_protect_begin (self);
	{
		_dict_exclude (self, key, qu_object_hash (key));
	}
  __qu_container_protect_end;
	return self;
}



/*) sub __each -> Iterator
	Return a -{Iterator} object to iterate through the keys.

	Note that C{for k, v in d} performs much better than C{for k in each d}
	and does not require extra memory nor an index lookup to get the value.

	@noexc

	@xmp
		d = [a:1, b:2, c:3]

		for k in each d            # calls d.__each
			print (k ' ' d[k] ' ')
		;;
		println

		for k, v in d              # interpreter uses an internal iterator
			print (k ' ' v ' ')
		;;
		println

	@out
		a 1 b 2 c 3
		a 1 b 2 c 3

	@p
	If you are sharing the Dict between threads then the result may be
	unpredictable. For example, if the other thread is modifying the same
	Dict then you may not get all the items you think you should have.
*/
static qu_any _dict_next (qu_dict self, qu_iterator i)
{
	qu_any k, v;

	v = qu_dict_walk (self, &k, &i->ex1, &i->ex2);
	QU_ITERATOR_STOP (v);
	if (qu_iterator_yield_keys (i))
		return k;
	else if (qu_iterator_yield_values (i))
		return v;
	else
	{
		qu_any argv[2] = {k, v};
		return qu_array_new_argv (2, argv);
	}
}



static qu_iterator qu__dict__each (qu_dict self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, _dict_next);
	qu_iterator_set_yield_keys (i);
	return i;
}



/*) sub __getitem (x) -> Any
	Return the associated value of the key v{x}.

	The Dict is temporarily I{protected} while this method is in progress.

	Note that the __eq method called are the methods of the items in
	the Dict. Not C{x.__eq}.

	@exc
	.EKey

	@xmp
		a = [foo:7, bar:8]
		println (a['foo'])

	@out
		7
*/
static qu_any qu__dict__getitem (qu_dict self, qu_any x)
{
	x = qu_dict_getitem (self, x);
	QU_REQUIRE (QuEKey, x);
	return x;
}



/*) sub __getslice (key, defval) -> Any
	Return the associated value of the key v{x}.

	This is the same as the #{__getitem} method except that if the key does
	not exist then v{defval} is returned.

	The Dict is temporarily I{protected} while this method is in progress.

	@noexc

	@xmp
		d = [foo:7, bar:8]
		println (d ['baz', 123])

	@out
		123

	@warn
	Because of how the slice operator works, if you do not specify a default
	value then -1 is returned.

	@xmp
		a = [foo:7, bar:8]
		println (a ['foo',])

	@out
		-1

	@see
	See the #{get} method for an alternative.
*/
static qu_any qu__dict__getslice (qu_dict self, qu_any x, qu_any defval)
{
	item_s *i;

  __qu_container_protect_begin (self);
	{
		i = _dict_lookup (self, x, qu_object_hash (x));
	}
  __qu_container_protect_end;

	if (i)
		return i->value;
	return defval;
}



/*) sub __in (x) -> Boolean
	Return true if v{x} is a key in self, false otherwise.

	This is not an atomic operation.
	__eq methods might need to be called to compare items.

	@noexc

	@xmp
		a = [foo:7, bar:8]
		println ('baz' in a)

	@out
		false
*/
qu_boolean qu__dict__in (qu_dict self, qu_any x)
{
	return qu_dict_getitem (self, x) ? QuTrue : QuFalse;
}



/*) sub __like (x) -> Boolean
	If v{x} is Dict and all keys in self are found in v{x} then return
	true, false otherwise.

	Thus C{x === y} and C{y === x} may yield different results.

	@noexc

	@xmp
		d = [foo:1 bar:2 gaz:3]
		e = [foo:4 gaz:6 bar:5 gee:4]
		println (d === e)
		println (e === d)

	@out
		true
		false
*/
static qu_boolean qu__dict__like (qu_dict self, qu_any x)
{
	if (qu_dict_is (x))
	{
		qu_any k;
		long   b = 0, i = 0;

		while (qu_dict_walk (self, &k, &b, &i))
		{
			if (!qu_dict_getitem ((qu_dict) x, k))
				return QuFalse;
		}
		return QuTrue;
	}
	return QuFalse;
}



/*) sub __of (type) -> self
	Set the type of the keys and/or items then return self.

	To set the type of the keys, call with one v{type} argument.

	To set the type of the keys and values, call with one Array argument.

	To set the type of the values, call with one Array argument with the
	first item set to nil.

	If the type is already set then EModify is thrown.

	The Dict is temporarily I{protected} while this method is in progress.

	@warn
	The -{Null} class may not be used as a type.

	Once set, you can not unset the type.

	@exc
	.EModify
	.EArgValue
	.EValue

	@xmp
		d = [:] of String
		println (d.type)

		d = [:] of (String, Int)
		println (d.type)

		d = [:] of (nil, Int)
		println (d.type)

	@out
		[<Class Sys.String>, nil]
		[<Class Sys.String>, <Class Sys.Int>]
		[nil, <Class Sys.Int>]

	@note
	The %{Valid} module provides some common validators for Dict.
*/
static void _dict_get_type_arg (qu_any x, qu_any *ktype, qu_any *vtype)
{
	qu_any k = 0, v = 0;

	if (qu_array_is (x))
	{
		qu_array a = (qu_array) x;

		if (a->len > 0)
		{
			k = a->objs[0];
			if (k == (qu_any) QuNil)
				k = 0;

			if (a->len > 1)
			{
				v = a->objs[1];
				if (v == (qu_any) QuNil)
					v = 0;
			}
		}
	}
	else
	{
		k = x;
	}

	*ktype = k;
	*vtype = v;
}


qu_dict qu__dict__of (qu_dict self, qu_any x)
{
	qu_any ktype;
	qu_any vtype;
	int    err = 0;

	_dict_get_type_arg (x, &ktype, &vtype);

	QU_ARG_CHECK (ktype || vtype);

	if (ktype)
	{
		QU_ARG_CHECK (qu_object_valid_type (ktype));
		if (qu_var_is (ktype) && !qu_var_is_typeof (ktype))
			ktype = ((qu_var) ktype)->value;
	}

	if (vtype)
	{
		QU_ARG_CHECK (qu_object_valid_type (vtype));
		if (qu_var_is (vtype) && !qu_var_is_typeof (vtype))
			vtype = ((qu_var) vtype)->value;
	}

  __qu_container_protect_begin (self);
	{
		if (ktype)
		{
			QU_ARG_CHECK (qu_object_valid_type (ktype));
			if (self->ktype && qu_var_is (self->ktype))
			{
			}

			if (self->ktype != ktype)
			{
				if (self->ktype)
					err = 1;
				else if (!qu_object_valid_type (ktype))
					err = 2;
			}
		}


		if (vtype)
		{
			if (self->vtype != vtype)
			{
				if (self->vtype)
					err = 1;
				else if (!qu_object_valid_type (vtype))
					err = 2;
			}
		}

		if (err == 0 && self->len)
		{
			if (ktype)
			{
				if (!_dict_keyof (self, ktype))
					err = 3;
			}
			if (vtype)
			{
				if (!_dict_valueof (self, vtype))
					err = 3;
			}
		}
	}
  __qu_container_protect_end;

	switch (err)
	{
		case 1: QU_THROW (QuEModify);
		case 2: QU_THROW (QuEArgValue);
		case 3: QU_THROW (QuEValue);
	}

	if (ktype)
		self->ktype = ktype;

	if (vtype)
		self->vtype = vtype;

	return self;
}



/*) sub __plus (x: Dict) -> Dict
	Return a new Dict which contains items in both.

	@exc
	.EPlus
	.EMemory

	@xmp
		a = [foo:1, bar:2]
		b = [baz:3]
		c = a + b
		println (c)

	@out
		[foo:1, bar:2, baz:3]
*/
static qu_dict qu__dict__plus (qu_dict self, qu_dict x)
{
	long    b, i;
	qu_any  k, v;
	qu_dict r;

	QU_REQUIRE (QuEPlus, qu_dict_is (x));
	QU_MEM_CHECK (self->len <= QU_GC_MAX_ITEMS - x->len);
	r = qu_dict_copy (self);
	b = i = 0;
	while ((v = qu_dict_walk (x, &k, &b, &i)) != 0)
		qu_dict_setitem (r, k, v);
	return r;
}



/*) sub __push (item) -> self
	Add a new item into this Dict.

	v{item} must be an Array of key/value pairs or a Dict.

	@exc
	.EArgValue
	.EModify

	@xmp
		d = [:]
		d <<< ('foo', 1, 'bar', 2)
		println (d)

	@out
		['foo':1, 'bar':2]
*/
qu_dict qu__dict__push (qu_dict self, qu_any x)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

	if (qu_array_is (x))
	{
		qu_array a = (qu_array) x;
		long     i;

		for (i = 0; i < a->len; i += 2)
		{
			QU_ARG_CHECK (a->len >= 2);
			QU_ENSURE (qu_dict_setitem (self, a->objs[i], a->objs[i + 1]));
		}
	}
	else if (qu_dict_is (x))
	{
		long   b = 0, i = 0;
		qu_any k, v;

		while ((v = qu_dict_walk ((qu_dict) x, &k, &b, &i)) != 0)
		{
			QU_ENSURE (qu_dict_setitem (self, k, v));
		}
	}
	else
	{
		QU_ARG_THROW;
	}
	return self;
}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		x = [a:1 b:2 c:3]
		x['d'] = x
		println (x)

	@out
		['a':1, 'b':2, 'c':3, 'd':[::]]
*/
static qu_null qu__dict__print (qu_dict self)
{
	if (self->len < 1)
		return qu_sys_printz (DICT_EMPTY_STR, DICT_EMPTY_STRLEN);

	else if (qu_thread_state_exist (QU_THREAD_STATE_PRINT, self))
		return qu_sys_printz (DICT_RECUR_STR, DICT_RECUR_STRLEN);

	else
	{
		qu_null r;
		long    b, i;
		qu_any  k, v;
		int     m = 0;

	  __qu_thread_state_begin (QU_THREAD_STATE_PRINT, self);
		r = qu_sys_printz ("[", 1);
		b = i = 0;
		while (r && ((v = qu_dict_walk (self, &k, &b, &i)) != 0))
		{
			if (m++)
				r = qu_sys_printz (", ", 2);
			r = QU_CONTAINER_PRINT_QUOTED (k);
			if (r)
				r = qu_sys_printz (":", 1);
			if (r)
				r = QU_CONTAINER_PRINT_QUOTED (v);
		}
		if (r)
			r = qu_sys_printz ("]", 1);
	  __qu_thread_state_end;
		QU_ENSURE (r);
		return r;
	}
}



/*) sub __setitem (key, val) -> Any
	If v{key} is already a key in the Dict then set its value, otherwise add
	a new entry. v{val} is returned as the result.

	The Dict is temporarily #{frozen} while this method is progress.

	@exc
	.EModify
	.EValue  #{type} mismatch

	@xmp
		d = [a:1, b:2]
		d['c'] = 3
		println (d)

	@out
		['a':1, 'b':2, 'c':3]
*/
qu_any qu__dict__setitem (qu_dict self, qu_any k, qu_any v)
{
	item_s *i;
	long    h;
	int     err = 1;

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

  __qu_container_protect_begin (self);
	{
		if (self->ktype)
			k = qu_object_validate (self, self->ktype, k);

		if (k && self->vtype)
			v = qu_object_validate (self, self->vtype, v);

		if (k && v)
		{
			err = 0;
			h = qu_object_hash (k);
			i = _dict_lookup (self, k, h);
			if (i)
			{
				i->value = v;
			}
			else
			{
				if (self->len >= QU_GC_MAX_ITEMS)
					err = 2;
				else
					_dict_include (self, k, v, h);
			}
		}
	}
  __qu_container_protect_end;

	if (err == 1)
		QU_THROW (QuEValue);

	if (err == 2)
		QU_THROW (QuEMemory);

	return v;
}



/*) sub __string -> String
	Return a String representation of self.

	@exc
	.EMemory

	@xmp
		d = ["foo":1 "bar":2]
		s = d as String
		println (s)

	@out
		['foo':1, 'bar':2]
*/
qu_string qu__dict__string (qu_dict self)
{
	qu_string s = qu_string_buffer_new (8);

	if (self->len < 1)
		qu_string_buffer_put (s, DICT_EMPTY_STR, DICT_EMPTY_STRLEN);

	else if (qu_thread_state_exist (QU_THREAD_STATE_STRING, self))
		qu_string_buffer_put (s, DICT_RECUR_STR, DICT_RECUR_STRLEN);

	else
	{
		int  r;
		long b, i;
		qu_any  k, v;
		int  m = 0;

	  __qu_thread_state_begin (QU_THREAD_STATE_STRING, self);
		r = qu_string_buffer_putc (s, '[');
		b = i = 0;
		while (r && ((v = qu_dict_walk (self, &k, &b, &i)) != 0))
		{
			if (m++) r = qu_string_buffer_put  (s, ", ", 2);
			         r = qu_string_buffer_puts (s, QU_CONTAINER_QUOTE (k));
			if (r)   r = qu_string_buffer_putc (s, ':');
			if (r)   r = qu_string_buffer_puts (s, QU_CONTAINER_QUOTE (v));
		}
		if (r)
			r = qu_string_buffer_putc (s, ']');
	  __qu_thread_state_end;
		QU_ENSURE (r);
	}
	return qu_string_buffer_return (s);
}



/*----------------------------------------------------------------------------
	specific instance methods - non-modifying
----------------------------------------------------------------------------*/

/*) sub add (key, value) -> self
	Add item.

	If v{key} exists, then make the value into an Array unless it is already
	an Array, then push v{value} into the Array.

	Otherwise just add the item.

	@exc
	.EValue
	.EModify

	@xmp
		d = [foo:1, bar:2]
		d.addquery ('foo', 2)
		d.addquery ('baz', 3)
		println (d)

	@out
		['foo': [1, 2], 'bar': 2, 'baz': 3]
*/
static qu_any qu__dict_add (qu_dict self, qu_any k, qu_any v)
{
	qu_any x = qu_dict_getitem (self, k);

	if (x == 0)
	{
		QU_ENSURE (qu_dict_setitem (self, k, v));
	}
	else
	{
		if (qu_array_is (x))
		{
			QU_ENSURE (qu_array_push (x, v));
		}
		else
		{
			qu_array a = qu_array_new (2);
			a->objs[0] = x;
			a->objs[1] = v;
			a->len = 2;

			QU_ENSURE (qu_dict_setitem (self, k, a));
		}
	}
	return self;
}



/*) sub any (defval = nil) -> Any
	Return an arbitrary key in the Dict.

	If the Dict is empty and v{defval} is given then v{defval} is returned,
	otherwise ENone is thrown

	@exc
	.ENone

	@xmp
		d = ['a':1, 'b':2, 'c':3]
		println (d.any)

	@out
		a
*/
qu_any qu__dict_any (qu_dict self, int argc, qu_any *argv)
{
	if (self->len == 0)
	{
		if (argc)
			return argv[0];
		QU_THROW (QuENone);
	}
	else
	{
		item_s *s = &self->sentinel;
		return s->next->key;
	}
}



/*) sub collect (what, res = {}) -> Any
	Collect keys with the same values into v{res}.

	v{res} must support the __push method.

	@exc
	.EArgValue

	@xmp
		d = [foo:1, bar:2, baz:1]
		s = d.collect (1)
		println (s)

	@out
		{'foo', 'baz'}
*/
static qu_any qu__dict_collect (qu_dict self, int argc, qu_any *argv)
{
	qu_any  what, into;
	long    b, i;
	qu_any  k, v;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (argc > 1)
		into = argv[1];
	else
		into = qu_set_new (4);

	b = i = 0;
	while ((v = qu_dict_walk (self, &k, &b, &i)) != 0)
	{
		if (qu_sys_eq (v, what) == QuTrue)
			QU_ENSURE (qu_sys_push (into, k));
	}
	return into;
}



/*) sub copy (keeptype = false) -> Dict
	Make a shallow copy of this Dict.

	If v{keeptype} is true then the new Dict is given the same #{type} as
	this Dict. Otherwise this method is just an alias of #{__copy}.

	@exc
	.EArgValue
*/
static qu_dict qu__dict_copy (qu_dict self, int argc, qu_any *argv)
{
	qu_dict r;
	int     keep = 0;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, keep);
	r = qu__dict__copy (self);
	if (keep)
	{
		r->ktype = self->ktype;
		r->vtype = self->vtype;
	}
	return r;
}



/*) sub delete (*arg) -> self
	Remove all keys specified in v{arg}.

	If called without any argument then remove any arbitrary key.

	@exc
	.EModify

	@xmp
		d = [foo:1, bar:2, baz:3]
		d.delete ('foo', 'bar')
		println (d)

	@out
		[baz: 3]
*/
static qu_dict qu__dict_delete (qu_dict self, int argc, qu_any *argv)
{
	if (argc == 0)
	{
		QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
		qu_dict_del_any (self);
	}
	else
	{
		while (argc--)
			QU_ENSURE (qu__dict__del (self, *argv++));
	}
	return self;
}



/*) sub eachvalue -> Iterator
	Return a -{Iterator} object to iterate through the values.

	@noexc

	@xmp
		d = ['a':1, 'b':2, 'c':3]
		for x in d.eachvalue
			print (x)
		;;

	@out
		1
		2
		3
*/
qu_iterator qu__dict_eachvalue (qu_dict self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, _dict_next);
	qu_iterator_set_yield_values (i);
	return i;
}



/*) sub find (dotspec: String) -> Any
	Find value with a dot notation.

	@exc
	.ENone

	@xmp
		a = [foo:[bar:[goo:'hi', giz:2], biz:4]]
		println (a.find ('foo.bar.goo'))

	@out
		hi
*/
qu_any qu__dict_find (qu_dict self, qu_string dot)
{
	qu_dict  d = self;
	char    *i, *j;
	long     l;
	qu_any   x = 0;

	QU_ARG_TYPE (dot, qu_string);

	l = dot->len;
	i = j = dot->str;
	while (l)
	{
		if (*j == '.')
		{
			x = qu_dict_getitemz (d, i, j - i);
			if (l > 1)
			{
				QU_REQUIRE (QuENone, x && qu_dict_is (x));
				d = x;
				i = j + 1;
			}
		}
		++j;
		--l;
	}
	if (j > i)
		x = qu_dict_getitemz (d, i, j - i);
	QU_REQUIRE (QuENone, x);
	return x;
}



/*) sub haskey (*arg) -> Boolean
	Whether the Dict has any key specified in v{arg}.

	@noexc

	@xmp
		d = [foo:1, bar:2]
		println (d.haskey ('fox'))
		println (d.haskey ('bax', 'bar'))

	@out
		false
		true
*/
static qu_boolean qu__dict_haskey (qu_dict self, int argc, qu_any *argv)
{
	while (argc--)
	{
		if (qu__dict__in (self, *argv++) == QuTrue)
			return QuTrue;
	}
	return QuFalse;
}



/*) sub hasvalue (*arg) -> Boolean
	Return true if any item in v{arg} is a value in the Dict.

	@noexc

	@xmp
		d = [foo:1, bar:2]
		println (d.hasvalue ('foo'))
		println (d.hasvalue (1, 7))

	@out
		false
		true
*/
static qu_boolean qu__dict_hasvalue (qu_dict self, int argc, qu_any *argv)
{
	while (argc--)
	{
		long i = 0, b = 0;
		qu_any  k, v, x = *argv++;

		while ((v = qu_dict_walk (self, &k, &b, &i)) != 0)
		{
			if (qu_sys_eq (v, x) == QuTrue)
				return QuTrue;
		}
	}
	return QuFalse;
}



/*) sub invert -> Dict
	Create a new Dict from self by inverting the key/value pairs, i.e.
	the keys become values and values become keys.

	If there are more than one values with the same value then the result
	is arbitrary.

	@noexc

	@xmp
		d = [a:1, b:2, c:1, d:1]
		e = d.invert
		println (e)

	@out
		[1:'d', 2:'b']
*/
qu_dict qu__dict_invert (qu_dict self)
{
	qu_dict d;
	long    b, i;
	qu_any  k, v;

	d = qu_dict_new (self->len);
	b = i = 0;
	while ((v = qu_dict_walk (self, &k, &b, &i)) != 0)
		qu_dict_setitem (d, v, k);
	return d;
}



/*) sub isenum -> Boolean
	Return true if this comes from an k{enum} statement.

	An k{enum} statement produces a #{frozen} Dict and read-only -{Var} objects
	as demonstrated in the following example.

	@xmp
		enum Weekdays
			Sun
			Mon
			Tue
			Wed
			Thu
			Fri
			Sat
		;;
		println (Weekdays.isenum)
		println (Weekdays)
		println (Weekdays.frozen)
		println (Mon)
		v = @Sun
		println (v)
		println (v.isconst)

	@out
		true
		[0:'Sun', 1:'Mon', 2:'Tue', 3:'Wed', 4:'Thu', 5:'Fri', 6:'Sat']
		true
		1
		<Var __main__.Sun>
		true
*/
qu_boolean qu__dict_isenum (qu_dict self)
{
	return qu_dict_is_enum (self) ? QuTrue : QuFalse;
}



/*) sub join (kvsep = ':', isep = ', ') -> String
	Concatenate the items using v{kvsep} as the value separator and v{isep}
	as the item separator.

	@exc
	.EMemory
	.EArgValue

	@xmp
		d = [a:1 b:2 c:3]
		println (d.join (' = ', ' :: '))

	@out
		a = 1 :: b = 2 :: c = 3
			# or maybe b = 2 :: a = 1 :: c = 3
			# or maybe c = 3 :: b = 2 :: a = 1
			# ... etc.
*/
qu_string qu__dict_join (qu_dict self, int argc, qu_any *argv)
{
	qu_string isep  = QuStringEMPTY;
	qu_string kvsep = QuStringEMPTY;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_VTYPE (0, kvsep, qu_string);
		if (argc > 1)
			QU_ARG_VTYPE (1, isep, qu_string);
	}

	if (self->len < 1)
	{
		return QuStringEMPTY;
	}
	else
	{
		qu_any    k, v;
		long      i, j;
		int       e = 0;
		int       z, w;
		qu_string b;

		QU_MEM_CHECK (self->len < QU_MEM_MAX / (isep->len + kvsep->len + 2));

		b = qu_string_buffer_new (self->len);
		i = j = 0;
		z = (isep->len != 0);
		w = (kvsep->len != 0);
		while ((v = qu_dict_walk (self, (qu_any *) &k, &i, &j)) != 0)
		{
			if (e && z)
				QU_ENSURE (qu_string_buffer_puts (b, isep));
			e = 1;
			if (k == self)
				QU_ENSURE (qu_string_buffer_put (b, DICT_RECUR_STR, DICT_RECUR_STRLEN));
			else
				QU_ENSURE (qu_string_buffer_puts (b, QU_CONTAINER_QUOTE (k)));

			if (w)
				QU_ENSURE (qu_string_buffer_puts (b, kvsep));

			if (v == self)
				QU_ENSURE (qu_string_buffer_put (b, DICT_RECUR_STR, DICT_RECUR_STRLEN));
			else
				QU_ENSURE (qu_string_buffer_puts (b, QU_CONTAINER_QUOTE (v)));
		}
		return qu_string_buffer_return (b);
	}
}



/*) sub joinkey (sep = ', ') -> String
	Concatenate the keys using v{sep} as separator.

	@exc
	.EMemory
	.EArgValue

	@xmp
		d = [a:1 b:2 c:3]
		println (d.joinkey (' : '))

	@out
		a : c : b
*/
qu_string qu__dict_joinkey (qu_dict self, int argc, qu_any *argv)
{
	qu_string sep = QuStringEMPTY;

	if (argc > 0)
		QU_ARG_VTYPE (0, sep, qu_string);

	if (self->len > 0)
	{
		qu_any     k, v;
		long       i, j;
		int        z, e = 0;
		qu_string  b;

		QU_MEM_CHECK (self->len < QU_MEM_MAX / (sep->len + 2));

		b = qu_string_buffer_new (self->len);
		i = j = 0;
		z = (sep->len != 0);
		while ((v = qu_dict_walk (self, (qu_any *) &k, &i, &j)) != 0)
		{
			if (e && z)
				QU_ENSURE (qu_string_buffer_puts (b, sep));
			e = 1;
			if (k == self)
				QU_ENSURE (qu_string_buffer_put (b, DICT_RECUR_STR, DICT_RECUR_STRLEN));
			else
				QU_ENSURE (qu_string_buffer_puts (b, QU_CONTAINER_QUOTE (k)));
		}
		return qu_string_buffer_return (b);
	}
	return QuStringEMPTY;
}



/*) sub joinvalue (sep = ', ') -> String
	Concatenate the values using v{sep} as separator.

	@exc
	.EMemory
	.EArgValue

	@xmp
		d = [a:1 b:2 c:3]
		println (d.joinvalue (' : '))

	@out
		1 : 2 : 3
*/
qu_string qu__dict_joinvalue (qu_dict self, int argc, qu_any *argv)
{
	qu_string sep = QuStringEMPTY;

	if (argc > 0)
		QU_ARG_VTYPE (0, sep, qu_string);

	if (self->len > 0)
	{
		qu_any    k, v;
		long      i, j;
		int       z, e = 0;
		qu_string b;

		QU_MEM_CHECK (self->len < QU_MEM_MAX / (sep->len + 2));

		b = qu_string_buffer_new (self->len);
		i = j = 0;
		z = (sep->len > 0);
		while ((v = qu_dict_walk (self, (qu_any *) &k, &i, &j)) != 0)
		{
			if (e && z)
				QU_ENSURE (qu_string_buffer_puts (b, sep));
			e = 1;
			if (v == self)
				QU_ENSURE (qu_string_buffer_put (b, DICT_RECUR_STR, DICT_RECUR_STRLEN));
			else
				QU_ENSURE (qu_string_buffer_puts (b, QU_CONTAINER_QUOTE (v)));
		}
		return qu_string_buffer_return (b);
	}
	return QuStringEMPTY;
}



/*) sub keys -> Array
	Return an -{Array} containing all keys.

	@exc
	.EMemory

	@xmp
		d = ['foo':1, 'bar':2, 'baz':3]
		a = d.keys
		println (a)

	@out
		['bar', 'baz', 'foo']
*/
qu_array qu__dict_keys (qu_dict self)
{
	item_s   *s = &self->sentinel, *i;
	qu_array  a = qu_array_new (self->len);
	qu_any   *p = a->objs;

	for (i = s->next; i != s; i = i->next)
		*p++ = i->key;
	a->len = self->len;
	return a;
}



/*) sub replace (value, replacement, max = -1) -> self
	Replace values specified by v{value} with v{replacement}.

	v{max} is the maximum number of replacements to perform.
	If v{max} is non-positive then there is no limit.

	Since items in a Dict are not sorted then specifying v{max} does not
	make much sense.

	@warn
	The Dict is frozen while replacement is performed.

	@exc
	.EModify
	.EValue

	@xmp
		d = [foo:1, bar:2, baz:1]
		d.replace (1, 4)
		println (d)

	@out
		['foo':4, 'bar':2, 'baz':4]
*/
qu_dict qu__dict_replace (qu_dict self, int argc, qu_any *argv)
{
	qu_any what, with;
	long   max = LONG_MAX;

	QU_ARG_VCOUNT (2, 3);
	what = argv[0];
	with = argv[1];
	if (argc > 2)
	{
		QU_ARG_vlong (2, max);
		if (max <= 0)
			max = LONG_MAX;
	}

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

  __qu_container_protect_begin (self);
	{
		if (self->vtype)
			with = qu_object_validate (self, self->vtype, with);

		if (with)
		{
			item_s *i, *s = &self->sentinel;
			for (i = s->next; i != s; i = i->next)
			{
				if (qu_sys_eq (i->value, what) == QuTrue)
					i->value = with;
			}
		}
	}
  __qu_container_protect_end;

	QU_ARG_CHECK (with);
	return self;
}



/*) sub select (*keys) -> Array
	Select values specified by v{keys} into an Array.

	@exc
	.EArgValue
	.EIndex
	.EMemory

	@xmp
		x = [foo:1, bar:2, baz:3]
		y = ['foo', 'bar']
		z = x.select (*y)
		println (y)

	@out
		[1, 2]
*/
static qu_array qu__dict_select (qu_dict self, int argc, qu_any *argv)
{
	qu_array a = qu_array_new (argc);

	while (argc--)
	{
		qu_any x = qu_dict_getitem (self, *argv++);
		if (x)
			qu_array_push (a, x);
	}
	return a;
}



/*) sub type -> Array
	Return the type of the keys and values as an Array of two items.
	The first item is the type of the keys.

	If not set then the item is set to Nil.

	See the #{__of} method for an example.
*/
qu_any qu__dict_type (qu_dict self)
{
	qu_any argv[2] =
	{
		self->ktype ? (qu_any) self->ktype : (qu_any) QuNil,
		self->vtype ? (qu_any) self->vtype : (qu_any) QuNil
	};
	return qu_array_new_argv (2, argv);
}



/*) sub typeof (x) -> Boolean
	Check whether or not all keys/values are of a specific type.

	To check for the keys, call with one argument.

	To check for the values, call with one Array argument with the first item
	set to Nil.

	To check for both, call with on Array argument. The first item is the
	type of the keys.

	@noexc

	@xmp
		d = [a:1, b:2, c:3] of (String, Int)
		println (d.typeof ((String, Int)))

	@out
		true
*/
static qu_boolean qu__dict_typeof (qu_dict self, qu_any x)
{
	qu_any ktype, vtype;

	_dict_get_type_arg (x, &ktype, &vtype);

	if (ktype)
	{
		if (!_dict_keyof (self, ktype))
			return QuFalse;
	}
	if (vtype)
	{
		if (!_dict_valueof (self, vtype))
			return QuFalse;
	}
	return (ktype || vtype) ? QuTrue : QuFalse;
}



/*) sub values -> Array
	Return an -{Array} containing all values.

	@noexc

	@xmp
		d = [foo:6, bar:7, baz:8]
		a = d.values
		println (a)

	@out
		[8, 7, 6]
*/
qu_array qu__dict_values (qu_dict self)
{
	item_s   *s = &self->sentinel, *i;
	qu_array  a = qu_array_new (self->len);
	qu_any   *p = a->objs;

	for (i = s->next; i != s; i = i->next)
		*p++ = i->value;
	a->len = self->len;
	return a;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

int qu_dict_del (qu_dict self, qu_any key)
{
	QU_ENSURE (qu__dict__del (self, key));
	return 1;
}



qu_any qu_dict_getitem (qu_dict self, qu_any key)
{
	item_s *i;
	long    h = qu_object_hash (key);

	if (qu_string_is (key))
		i = _dict_lookupz (self, ((qu_string) key)->str, ((qu_string) key)->len, h);
	else
	{
	  __qu_container_protect_begin (self);
		i = _dict_lookup (self, key, h);
	  __qu_container_protect_end;
	}
	return i ? i->value : 0;
}



qu_any qu_dict_getitemz (qu_dict d, const char *key, long len)
{
	item_s *i = _dict_lookupz (d, key, len, qu_strz_hash (key, len));
	return i ? i->value : 0;
}



void qu_dict_setitemz (qu_dict d, const char *key, long len, qu_any val)
{
	long    h = qu_strz_hash (key, len);
	item_s *i = _dict_lookupz (d, key, len, h);

	if (i)
		i->value = val;
	else
	{
		qu_string k = qu_string_new (key, len);
		k->hash = h;
		_dict_include (d, k, val, h);
	}
}



qu_any qu_dict_walk (qu_dict self, qu_any *key, long *bucket, long *index)
{
	item_s *e, **b;
	long    i, j, k, m;

	b = self->buckets;
	m = self->mask;
	for (k = *bucket, i = *index; k <= m; k++, i = 0)
	{
		for (e = b[k], j = 0; e; e = e->right, j++)
		{
			if (j == i)
			{
				*bucket = k;
				*index = i + 1;
				*key = e->key;
				return e->value;
			}
		}
	}
	return 0;
}



int qu_dict_like (qu_dict self, qu_dict other)
{
	return qu__dict__like (self, other) == QuTrue;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

qu_class QuArgClass;

void qu_dict__setup ()
{
	qu_class_setup       (QuDictClass, qu_dict_s, qu__dict__destroy, qu__dict__mark);
	qu_class_setup_final (QuArgClass , qu_dict_s, qu__dict__destroy, qu__dict__mark);

	QuDictEMPTY = qu_dict_new_static (0);
	qu_dict_freeze (QuDictEMPTY);
}



void qu_dict__init ()
{
	/* standard class methods */
	qu_class_method_class (QuDictClass, QU_NAME__inst, qu__dict__inst, -1);
	qu_class_method_class (QuDictClass, QU_NAME_new  , qu__dict_new  , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuDictClass, QU_NAME__as      , qu__dict__as      , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__clear   , qu__dict__clear   , 0);
	qu_class_alias        (QuDictClass, QU_NAME__clear   , QU_NAME_clear);
	qu_class_method_inst  (QuDictClass, QU_NAME__copy    , qu__dict__copy    , 0);
	qu_class_method_inst  (QuDictClass, QU_NAME__del     , qu__dict__del     , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__each    , qu__dict__each    , 0);
	qu_class_method_inst  (QuDictClass, QU_NAME__getitem , qu__dict__getitem , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__getslice, qu__dict__getslice, 2);
	qu_class_method_inst  (QuDictClass, QU_NAME__in      , qu__dict__in      , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__like    , qu__dict__like    , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__of      , qu__dict__of      , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__plus    , qu__dict__plus    , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__push    , qu__dict__push    , 1);
	qu_class_method_inst  (QuDictClass, QU_NAME__print   , qu__dict__print   , 0);
	qu_class_method_inst  (QuDictClass, QU_NAME__setitem , qu__dict__setitem , 2);
	qu_class_method_inst  (QuDictClass, QU_NAME__string  , qu__dict__string  , 0);

	/* specific instance methods */
	qu_class_method_instz (QuDictClass, "add"      , qu__dict_add      ,  2);
	qu_class_method_instz (QuDictClass, "any"      , qu__dict_any      , -1);
	qu_class_method_instz (QuDictClass, "collect"  , qu__dict_collect  , -1);
	qu_class_method_instz (QuDictClass, "copy"     , qu__dict_copy     , -1);
	qu_class_method_instz (QuDictClass, "delete"   , qu__dict_delete   , -1);
	qu_class_method_instz (QuDictClass, "eachvalue", qu__dict_eachvalue, -1);
	qu_class_method_instz (QuDictClass, "find"     , qu__dict_find     ,  1);
	qu_class_method_instz (QuDictClass, "haskey"   , qu__dict_haskey   , -1);
	qu_class_method_instz (QuDictClass, "hasvalue" , qu__dict_hasvalue , -1);
	qu_class_method_instz (QuDictClass, "invert"   , qu__dict_invert   ,  0);
	qu_class_method_instz (QuDictClass, "isenum"   , qu__dict_isenum   ,  0);
	qu_class_method_instz (QuDictClass, "join"     , qu__dict_join     , -1);
	qu_class_method_instz (QuDictClass, "joinkey"  , qu__dict_joinkey  , -1);
	qu_class_method_instz (QuDictClass, "joinvalue", qu__dict_joinvalue, -1);
	qu_class_method_instz (QuDictClass, "keys"     , qu__dict_keys     ,  0);
	qu_class_method_instz (QuDictClass, "replace"  , qu__dict_replace  , -1);
	qu_class_method_instz (QuDictClass, "select"   , qu__dict_select   , -1);
	qu_class_method_instz (QuDictClass, "type"     , qu__dict_type     ,  0);
	qu_class_method_instz (QuDictClass, "typeof"   , qu__dict_typeof   ,  1);
	qu_class_method_instz (QuDictClass, "values"   , qu__dict_values   ,  0);

	/* finalize declaration */
	qu_class_set_super (QuDictClass, QuContainerClass);
	qu_class_enddecl   (QuDictClass);

	qu_class_set_super (QuArgClass, QuDictClass);
	qu_class_enddecl   (QuArgClass);
}
