/* Dlib.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Dlib
	Dlib enables you to access shared (dynamic) library functions and variables
	from within your Qu programs.

	@warn
	Using this class requires some knowledge of C.
	Give the wrong type and your program will crash.

	You can not pass more than 16 arguments to a library function.

	@xmp
		libm = Dlib ('/usr/lib/libm.so')
		r = libm.call ("sin", 'f:f', 1.234)
		println (r)
		libm.close

	@out
		0.94381820937463368

	@p
	Unless you know how Qu internals work, you must not attempt to open Qu libraries
	because they are not static libraries. They may need special treatments, i.e.
	initialized, load and compile the ".qm" part at the same time, open other
	dependent libraries, etc.
*/
qu_class QuDlibClass;


/* whether or not the qu_dlib object is opened */
#define DLIB_OPENED(self) ((self)->handle != NULL)

/* make sure the qu_dlib object is opened otherwise throw QuECall */
#define DLIB_ENSURE_OPENED(self) \
QU_BEGIN \
	if (!DLIB_OPENED (self)) \
		QU_THROWZ (QuECall, "library is closed"); \
QU_END


/*
	Check for valid specification.
	Return 1 if valid, throw and return 0 otherwise.
	Call with argc = -1 to check the specification only.
*/
static int _dlib_check_spec (const char *spec, int argc);

/*
	Call a function.
	Make sure to call _dlib_check_spec before you call this function.
*/
static qu_any _dlib_call (void *handle, const char *spec, int argc, qu_any *argv);

/* no argument spec */
static qu_string SpecNoArg; /* "-:" */


/*
	Cfunc is not created until it is actually used.

	How it works:

	m = qu__dlib_func (lib, name, spec)

		m->sub  = qu__cfunc_caller
		m->self = the qu_cfunc object

	m (1, 2) -> qu__cfunc_caller (m->self, 1, 2)
*/
static qu_class QuCfuncClass = 0;

typedef struct qu_cfunc_s qu_cfunc_s;
struct qu_cfunc_s
{
	QU_OBJECT_HEAD;
	qu_string   name;   /* the symbol name */
	qu_dlib     lib;    /* the corresponding qu_dlib */
	qu_string   spec;   /* calling specification */
	void       *handle; /* address of C function */

	/* the 'name' field has no functional meaning */
};

typedef qu_cfunc_s *qu_cfunc;

/* whether the object is a Cfunc */
#define qu_cfunc_is(o) (qu_classof (o) == QuCfuncClass)

static void qu__cfunc__mark (qu_cfunc self)
{
	qu_gc_xmark (self->lib);
	qu_gc_mark  (self->spec);
	qu_gc_mark  (self->name);
}

/* the caller */
static qu_sub _sub_cfunc_caller;

static qu_any qu__cfunc_caller (qu_cfunc self, int argc, qu_any *argv)
{
	long n;

	/* just in case */
	QU_REQUIRE (QuECall, qu_cfunc_is (self));

	if (self->lib)
		DLIB_ENSURE_OPENED (self->lib);

	n = self->spec->len - 2;
	if (n != argc)
		QU_THROWF (QuEArgCount, "%s requires %d argument(s) but %d given", self->name->str, n, argc);
	return _dlib_call (self->handle, self->spec->str, argc, argv);
}


/* create the Cfunc class - done at runtime by qu__dlib_func */
static void qu__cfunc__setup ()
{
	QuCfuncClass = qu_class_newz (QuSysModule, "Cfunc", qu_cfunc_s, 0, qu__cfunc__mark);
	QuCfuncClass->id = QU_CLASS_ID_Cfunc;

	_sub_cfunc_caller =
	qu_class_method_staticz (QuCfuncClass, "__cfunc__", qu__cfunc_caller, -1);

	/* finalize declaration */
	qu_class_enddecl (QuCfuncClass);
}


/*----------------------------------------------------------------------------
	types and arguments
----------------------------------------------------------------------------*/

/* the maximum number of arguments that can be passed to a library function */
#define DLIB_MAX_ARGS  16

/* the arguments */
#define DLIB_CALL_ARGS long, long, long, long, long, long, long, long, long, long, long, long, long, long, long, long

/* how to call */
#define DLIB_CALL(f, arg) (f) (arg[0], arg[1], arg[2], arg[3], arg[4], arg[5], arg[6], arg[7], arg[8], arg[9], arg[10], arg[11], arg[12], arg[13], arg[14], arg[15])

/* functions with no arguments */
typedef void               (*f0_f           ) ();
typedef char               (*f0_char_f      ) ();
typedef int                (*f0_byte_f      ) ();
typedef unsigned int       (*f0_ubyte_f     ) ();
typedef short              (*f0_short_f     ) ();
typedef unsigned short     (*f0_ushort_f    ) ();
typedef int                (*f0_int_f       ) ();
typedef unsigned int       (*f0_uint_f      ) ();
typedef short              (*f0_int16_f     ) ();
typedef unsigned short     (*f0_uint16_f    ) ();
#if SIZEOF_INT == 4
typedef int                (*f0_int32_f     ) ();
typedef unsigned int       (*f0_uint32_f    ) ();
#elif SIZEOF_LONG == 4
typedef long               (*f0_int32_f     ) ();
typedef unsigned long      (*f0_uint32_f    ) ();
#endif
typedef long               (*f0_long_f      ) ();
typedef unsigned long      (*f0_ulong_f     ) ();
#ifdef HAVE_LONG_LONG
typedef long long          (*f0_longlong_f  ) ();
typedef unsigned long long (*f0_ulonglong_f ) ();
#endif
typedef float              (*f0_float_f     ) ();
typedef double             (*f0_double_f    ) ();
#ifdef HAVE_LONG_DOUBLE
typedef long double        (*f0_longdouble_f) ();
#endif
typedef ssize_t            (*f0_ssize_t_f   ) ();
typedef size_t             (*f0_size_t_f    ) ();
typedef off_t              (*f0_off_t_f     ) ();
typedef int                (*f0_Boolean_f   ) ();
typedef long               (*f0_Int_f       ) ();
typedef void *             (*f0_Pointer_f   ) ();
typedef char *             (*f0_String_f    ) ();
typedef qu_proc            (*f0_Proc_f      ) ();
typedef qu_any             (*f0_Any_f       ) ();
#if QU_SUPPORT_GLIB
typedef qu_gobject         (*f0_GObject_f   ) ();
#endif

/* functions with arguments */
typedef void               (*fn_f           ) (DLIB_CALL_ARGS);
typedef char               (*fn_char_f      ) (DLIB_CALL_ARGS);
typedef int                (*fn_byte_f      ) (DLIB_CALL_ARGS);
typedef unsigned int       (*fn_ubyte_f     ) (DLIB_CALL_ARGS);
typedef short              (*fn_short_f     ) (DLIB_CALL_ARGS);
typedef unsigned short     (*fn_ushort_f    ) (DLIB_CALL_ARGS);
typedef int                (*fn_int_f       ) (DLIB_CALL_ARGS);
typedef unsigned int       (*fn_uint_f      ) (DLIB_CALL_ARGS);
typedef short              (*fn_int16_f     ) (DLIB_CALL_ARGS);
typedef unsigned short     (*fn_uint16_f    ) (DLIB_CALL_ARGS);
#if SIZEOF_INT == 4
typedef int                (*fn_int32_f     ) (DLIB_CALL_ARGS);
typedef unsigned int       (*fn_uint32_f    ) (DLIB_CALL_ARGS);
#elif SIZEOF_LONG == 4
typedef long               (*fn_int32_f     ) (DLIB_CALL_ARGS);
typedef unsigned long      (*fn_uint32_f    ) (DLIB_CALL_ARGS);
#endif
typedef long               (*fn_long_f      ) (DLIB_CALL_ARGS);
typedef unsigned long      (*fn_ulong_f     ) (DLIB_CALL_ARGS);
#ifdef HAVE_LONG_LONG
typedef long long          (*fn_longlong_f  ) (DLIB_CALL_ARGS);
typedef unsigned long long (*fn_ulonglong_f ) (DLIB_CALL_ARGS);
#endif
typedef float              (*fn_float_f     ) (DLIB_CALL_ARGS);
typedef double             (*fn_double_f    ) (DLIB_CALL_ARGS);
#ifdef HAVE_LONG_DOUBLE
typedef long double        (*fn_longdouble_f) (DLIB_CALL_ARGS);
#endif
typedef ssize_t            (*fn_ssize_t_f   ) (DLIB_CALL_ARGS);
typedef size_t             (*fn_size_t_f    ) (DLIB_CALL_ARGS);
typedef off_t              (*fn_off_t_f     ) (DLIB_CALL_ARGS);
typedef int                (*fn_Boolean_f   ) (DLIB_CALL_ARGS);
typedef long               (*fn_Int_f       ) (DLIB_CALL_ARGS);
typedef char *             (*fn_String_f    ) (DLIB_CALL_ARGS);
typedef void *             (*fn_Pointer_f   ) (DLIB_CALL_ARGS);
typedef qu_proc            (*fn_Proc_f      ) (DLIB_CALL_ARGS);
typedef qu_any             (*fn_Any_f       ) (DLIB_CALL_ARGS);
#if QU_SUPPORT_GLIB
typedef qu_gobject         (*fn_GObject_f   ) (DLIB_CALL_ARGS);
#endif



/*----------------------------------------------------------------------------
	check specification
----------------------------------------------------------------------------*/

static int _dlib_supported_type (char c, int *n)
{
	switch (c)
	{
		case '-':
		case QU_STRUCT_ID_char:
		case QU_STRUCT_ID_byte:
		case QU_STRUCT_ID_ubyte:
		case QU_STRUCT_ID_short:
		case QU_STRUCT_ID_ushort:
		case QU_STRUCT_ID_int:
		case QU_STRUCT_ID_uint:
#if SIZEOF_SHORT == 2
		case QU_STRUCT_ID_int16:
		case QU_STRUCT_ID_uint16:
#endif
#if SIZEOF_INT == 4
		case QU_STRUCT_ID_int32:
		case QU_STRUCT_ID_uint32:
			*n += sizeof (int);
			return 1;
#elif SIZEOF_LONG == 4
		case QU_STRUCT_ID_int32:
		case QU_STRUCT_ID_uint32:
			*n += sizeof (long);
			return 1;
#endif

		case QU_STRUCT_ID_long:
		case QU_STRUCT_ID_ulong:
			*n += sizeof (long);
			return 1;

#ifdef HAVE_LONG_LONG
		case QU_STRUCT_ID_longlong:
		case QU_STRUCT_ID_ulonglong:
			*n += sizeof (long long);
			return 1;
#endif
		case QU_STRUCT_ID_float:
			*n += sizeof (float);
			return 1;

		case QU_STRUCT_ID_double:
			*n += sizeof (double);
			return 1;

#ifdef HAVE_LONG_DOUBLE
		case QU_STRUCT_ID_longdouble:
			*n += sizeof (long double);
			return 1;

#endif
#if SIZEOF_SIZE_T <= SIZEOF_LONG
		case QU_STRUCT_ID_ssize_t:
		case QU_STRUCT_ID_size_t:
			*n += sizeof (size_t);
			return 1;

#endif
#if SIZEOF_OFF_T <= SIZEOF_LONG
		case QU_STRUCT_ID_off_t:
			*n += sizeof (off_t);
			return 1;

#endif
		case QU_STRUCT_ID_Boolean:
		case QU_STRUCT_ID_Int:
		case QU_STRUCT_ID_Pointer:
		case QU_STRUCT_ID_String:
		case QU_STRUCT_ID_Proc:
		case QU_STRUCT_ID_Any:
#ifdef QU_SUPPORT_GLIB
		case QU_STRUCT_ID_GObject:
#endif
			*n += sizeof (void *);
			return 1;
	}
	return 0;
}


static int _dlib_check_spec (const char *spec, int argc)
{
	int n = 0;
	int m = sizeof (long) * DLIB_MAX_ARGS;

	switch (*spec)
	{
		case '-':
			break;

		case '+':
			QU_ARG_CHECKZ (argc, "+ requires at least one argument");
			break;

		case '\0':
			QU_ARG_THROWZ ("empty specification");

		default:
			if (!_dlib_supported_type (*spec, &n))
				QU_ARG_THROWF ("return type \"%c\" not supported", *spec);
	}

	++spec;
	if (*spec == '\0')
	{
	  zero:
		if (argc == 0 || argc == -1)
			return 1;
		QU_ARG_THROWF ("require %d arguments but %d given", 0, argc);
	}

	if (*spec != ':')
		QU_ARG_THROWZ ("missing \":\"");

	++spec;
	n = strlen (spec);
	if (n == 0)
		goto zero;

	if (n != argc && argc != -1)
		QU_ARG_THROWF ("require %d arguments but %d given", n, argc);

	n = 0;
	for (; *spec; ++spec)
	{
		if (!_dlib_supported_type (*spec, &n))
			QU_ARG_THROWF ("argument type \"%c\" not supported", *spec);
		if (n > m)
			QU_ARG_THROWZ ("size of arguments too big");
	}
	return 1;
}



/*----------------------------------------------------------------------------
	pushing the arguments
----------------------------------------------------------------------------*/

static char *_dlib_push_char (char *p, qu_any x)
{
	QU_ARG_TYPE  (x, qu_stringy);
	QU_ARG_CHECK (((qu_string) x)->len);
	*((int *) p) = (int) ((qu_string) x)->str[0];
	p += sizeof (int);
	return p;
}



static char *_dlib_push_byte (char *p, qu_any x)
{
	int v;
	QU_ARG_byte (x, v);
	*((int *) p) = (int) v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_ubyte (char *p, qu_any x)
{
	unsigned char v;
	QU_ARG_ubyte (x, v);
	*((unsigned int *) p) = (unsigned int) v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_short (char *p, qu_any x)
{
	short v;
	QU_ARG_short (x, v);
	*((int *) p) = (int) v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_ushort (char *p, qu_any x)
{
	unsigned short v;
	QU_ARG_ushort (x, v);
	*((unsigned int *) p) = (unsigned int) v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_int (char *p, qu_any x)
{
	int v;
	QU_ARG_int (x, v);
	*((int *) p) = (int) v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_uint (char *p, qu_any x)
{
	unsigned int v;
	QU_ARG_uint (x, v);
	*((unsigned int *) p) = (unsigned int) v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_int16 (char *p, qu_any x)
{
#if SIZEOF_SHORT == 2
	short v;
	QU_ARG_short (x, v);
	*((int *) p) = (int) v;
	p += sizeof (long);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_uint16 (char *p, qu_any x)
{
#if SIZEOF_SHORT == 2
	unsigned short v;
	QU_ARG_ushort (x, v);
	*((unsigned int *) p) = (unsigned int) v;
	p += sizeof (int);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_int32 (char *p, qu_any x)
{
#if SIZEOF_INT == 4
	int v;
	QU_ARG_int (x, v);
	*((int *) p) = (int) v;
	p += sizeof (int);
	return p;
#elif SIZEOF_LONG == 4
	long v;
	QU_ARG_long (x, v);
	*((long *) p) = v;
	p += sizeof (long);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_uint32 (char *p, qu_any x)
{
#if SIZEOF_INT == 4
	unsigned int v;
	QU_ARG_uint (x, v);
	*((unsigned int *) p) = (unsigned int) v;
	p += sizeof (int);
	return p;
#elif SIZEOF_LONG == 4
	unsigned long v;
	QU_ARG_ulong (x, v);
	*((unsigned long *) p) = v;
	p += sizeof (long);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_long (char *p, qu_any x)
{
	long v;
	QU_ARG_long (x, v);
	*((long *) p) = v;
	p += sizeof (long);
	return p;
}



static char *_dlib_push_ulong (char *p, qu_any x)
{
	unsigned long v;
	QU_ARG_ulong (x, v);
	*((unsigned long *) p) = v;
	p += sizeof (long);
	return p;
}



static char *_dlib_push_longlong (char *p, qu_any x)
{
#ifdef HAVE_LONG_LONG
	long long v;
	QU_ARG_longlong (x, v);
	*((long long *) p) = v;
	p += sizeof (long long);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_ulonglong (char *p, qu_any x)
{
#ifdef HAVE_LONG_LONG
	unsigned long long v;
	QU_ARG_ulonglong (x, v);
	*((unsigned long long *) p) = v;
	p += sizeof (long long);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_float (char *p, qu_any x)
{
	float v;
	QU_ARG_float (x, v);
	*((float *) p) = v;
	p += sizeof (float);
	return p;
}



static char *_dlib_push_double (char *p, qu_any x)
{
	double v;
	QU_ARG_double (x, v);
	*((double *) p) = v;
	p += sizeof (double);
	return p;
}



static char *_dlib_push_longdouble (char *p, qu_any x)
{
#ifdef HAVE_LONG_DOUBLE
	long double v;
	QU_ARG_longdouble (x, v);
	*((long double *) p) = v;
	p += sizeof (long double);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_ssize_t (char *p, qu_any x)
{
#if SIZEOF_SIZE_T == SIZEOF_INT
	ssize_t v;
	QU_ARG_int (x, v);
	*((ssize_t *) p) = v;
	p += sizeof (ssize_t);
	return p;
#elif SIZEOF_SIZE_T == SIZEOF_LONG
	ssize_t v;
	QU_ARG_long (x, v);
	*((ssize_t *) p) = v;
	p += sizeof (ssize_t);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_size_t (char *p, qu_any x)
{
#if SIZEOF_SIZE_T == SIZEOF_INT
	size_t v;
	QU_ARG_uint (x, v);
	*((size_t *) p) = v;
	p += sizeof (size_t);
	return p;
#elif SIZEOF_SIZE_T == SIZEOF_LONG
	size_t v;
	QU_ARG_ulong (x, v);
	*((size_t *) p) = v;
	p += sizeof (size_t);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_off_t (char *p, qu_any x)
{
#if SIZEOF_OFF_T == SIZEOF_INT
	off_t v;
	QU_ARG_int (x, v);
	*((off_t *) p) = v;
	p += sizeof (off_t);
	return p;
#elif SIZEOF_OFF_T == SIZEOF_LONG
	off_t v;
	QU_ARG_long (x, v);
	*((off_t *) p) = v;
	p += sizeof (off_t);
	return p;
#else
	return 0;
#endif
}



static char *_dlib_push_Null (char *p, qu_any x)
{
	/* allow any argument */
	*((void **) p) = NULL;
	p += sizeof (void *);
	return p;
}



static char *_dlib_push_Boolean (char *p, qu_any x)
{
	int v;
	QU_ARG_bool (x, v);
	*((int *) p) = v;
	p += sizeof (int);
	return p;
}



static char *_dlib_push_Int (char *p, qu_any x)
{
	long v;
	QU_ARG_long (x, v);
	*((long *) p) = v;
	p += sizeof (long);
	return p;
}



static char *_dlib_push_String (char *p, qu_any x)
{
	if (qu_stringy_is (x))
	{
		*((char **) p) = ((qu_string) x)->str;
	}
#if 0
	else if (qu_null_is (x))
	{
		*((char **) p) = NULL;
	}
	else if (qu_pointer_is (x))
	{
		*((char **) p) = (char *) ((qu_pointer) x)->ptr;
	}
#endif
	else
	{
		return 0;
	}
	p += sizeof (char *);
	return p;
}



static char *_dlib_push_Pointer (char *p, qu_any x)
{
	if (qu_pointer_is (x))
	{
		*((void **) p) = ((qu_pointer) x)->ptr;
	}
	else if (qu_null_is (x))
	{
		*((void **) p) = NULL;
	}
	else if (qu_string_is (x))
	{
		*((void **) p) = (void *) ((qu_string) x)->str;
	}
#ifdef QU_SUPPORT_GLIB
	else if (qu_gobject_is (x))
	{
		*((void **) p) = (void *) ((qu_gobject) x)->ptr;
	}
#endif
	p += sizeof (void *);
	return p;
}



static void *_dlib_push_Proc (char *p, qu_proc x)
{
	if (!qu_proc_is (x))
		return 0;
	*((void **) p) = (void *) x->code;
	p += sizeof (void *);
	return p;
}



static void *_dlib_push_Any (char *p, qu_any x)
{
	if (x == QuNil)
		*((void **) p) = NULL;
	else
		*((void **) p) = x;
	p += sizeof (void *);
	return p;
}



#ifdef QU_SUPPORT_GLIB
static void *_dlib_push_GObject (char *p, qu_any x)
{
	if (qu_gobject_is (x))
	{
		*((void **) p) = ((qu_gobject) x)->ptr;
	}
	else if (x == QuNil)
	{
		*((void **) p) = NULL;
	}
	else if (qu_pointer_is (x))
	{
		*((void **) p) = ((qu_pointer) x)->ptr;
	}
	else
	{
		return 0;
	}
	p += sizeof (void *);
	return p;

}
#endif



/*----------------------------------------------------------------------------
	do the call
----------------------------------------------------------------------------*/

static qu_any _dlib_call (void *handle, const char *spec, int argc, qu_any *argv)
{
	long    arg [DLIB_MAX_ARGS * 2];
	qu_any *args = argv;
	char   *s;
	int     i;
	char   *p;

	assert (_dlib_check_spec (spec, argc));

	s = (char *) spec + 2;
	p = (char *) arg;

	for (i = 1; *s; i++, s++, args++)
	{
		qu_any x = *args;

		switch (*s)
		{
			case QU_STRUCT_ID_char:       p = _dlib_push_char       (p, x); break;
			case QU_STRUCT_ID_byte:       p = _dlib_push_byte       (p, x); break;
			case QU_STRUCT_ID_ubyte:      p = _dlib_push_ubyte      (p, x); break;
			case QU_STRUCT_ID_short:      p = _dlib_push_short      (p, x); break;
			case QU_STRUCT_ID_ushort:     p = _dlib_push_ushort     (p, x); break;
			case QU_STRUCT_ID_int:        p = _dlib_push_int        (p, x); break;
			case QU_STRUCT_ID_uint:       p = _dlib_push_uint       (p, x); break;
			case QU_STRUCT_ID_int16:      p = _dlib_push_int16      (p, x); break;
			case QU_STRUCT_ID_uint16:     p = _dlib_push_uint16     (p, x); break;
			case QU_STRUCT_ID_int32:      p = _dlib_push_int32      (p, x); break;
			case QU_STRUCT_ID_uint32:     p = _dlib_push_uint32     (p, x); break;
			case QU_STRUCT_ID_long:       p = _dlib_push_long       (p, x); break;
			case QU_STRUCT_ID_ulong:      p = _dlib_push_ulong      (p, x); break;
			case QU_STRUCT_ID_longlong:   p = _dlib_push_longlong   (p, x); break;
			case QU_STRUCT_ID_ulonglong:  p = _dlib_push_ulonglong  (p, x); break;
			case QU_STRUCT_ID_float:      p = _dlib_push_float      (p, x); break;
			case QU_STRUCT_ID_double:     p = _dlib_push_double     (p, x); break;
			case QU_STRUCT_ID_longdouble: p = _dlib_push_longdouble (p, x); break;
			case QU_STRUCT_ID_size_t:     p = _dlib_push_size_t     (p, x); break;
			case QU_STRUCT_ID_ssize_t:    p = _dlib_push_ssize_t    (p, x); break;
			case QU_STRUCT_ID_off_t:      p = _dlib_push_off_t      (p, x); break;
			case QU_STRUCT_ID_Null:       p = _dlib_push_Null       (p, x); break;
			case QU_STRUCT_ID_Boolean:    p = _dlib_push_Boolean    (p, x); break;
			case QU_STRUCT_ID_Int:        p = _dlib_push_Int        (p, x); break;
			case QU_STRUCT_ID_String:     p = _dlib_push_String     (p, x); break;
			case QU_STRUCT_ID_Pointer:    p = _dlib_push_Pointer    (p, x); break;
			case QU_STRUCT_ID_Proc:       p = _dlib_push_Proc       (p, x); break;
			case QU_STRUCT_ID_Any:        p = _dlib_push_Any        (p, x); break;
#ifdef QU_SUPPORT_GLIB
			case QU_STRUCT_ID_GObject:
				p = _dlib_push_GObject (p, x);
				if (p == NULL)
					QU_ARG_THROWZ ("cant pass gobject already destroyed");
				break;
#endif
			default:
				p = NULL;
		}

		if (p == NULL)
			QU_ARG_THROWF ("cant convert argument %d \"%c\" to C type", i, *s);
	}

	if (argc == 0)
	{
		switch (*spec)
		{
			case '-':
			{
				((f0_f) handle) ();
				return QuNil;
			}

			case '+':
				QU_ARG_THROWZ ("+ requires at least one argument");

			case QU_STRUCT_ID_char:
			{
				char r = ((f0_char_f) handle) ();
				return qu_string_byte (r);
			}

			case QU_STRUCT_ID_byte:
			{
				int r = ((f0_byte_f) handle) ();
				return qu_int_newi (r);
			}

			case QU_STRUCT_ID_ubyte:
			{
				unsigned int r = ((f0_ubyte_f) handle) ();
				return qu_int_newui (r);
			}

			case QU_STRUCT_ID_short:
			{
				short r = ((f0_short_f) handle) ();
				return qu_int_newh (r);
			}

			case QU_STRUCT_ID_ushort:
			{
				unsigned short r = ((f0_ushort_f) handle) ();
				return qu_int_newuh (r);
			}

			case QU_STRUCT_ID_int:
			{
				int r = ((f0_int_f) handle) ();
				return qu_integer_newi (r);
			}

			case QU_STRUCT_ID_uint:
			{
				unsigned int r = ((f0_uint_f) handle) ();
				return qu_integer_newui (r);
			}

			case QU_STRUCT_ID_int16:
			{
				int r = (int) ((f0_int16_f) handle) ();
				return qu_int_newi (r);
			}

			case QU_STRUCT_ID_uint16:
			{
				unsigned int r = (unsigned int) ((f0_uint16_f) handle) ();
				return qu_int_newui (r);
			}

			case QU_STRUCT_ID_int32:
			{
				long r = (long) ((f0_int32_f) handle) ();
				return qu_integer_newl (r);
			}

			case QU_STRUCT_ID_uint32:
			{
				unsigned long r = (unsigned long) ((f0_uint32_f) handle) ();
				return qu_integer_newul (r);
			}

			case QU_STRUCT_ID_long:
			{
				long r = ((f0_long_f) handle) ();
				return qu_integer_newl (r);
			}

			case QU_STRUCT_ID_ulong:
			{
				unsigned long r = ((f0_ulong_f) handle) ();
				return qu_integer_newul (r);
			}

#ifdef HAVE_LONG_LONG
			case QU_STRUCT_ID_longlong:
			{
				long long r = ((f0_longlong_f) handle) ();
				return qu_integer_newll (r);
			}

			case QU_STRUCT_ID_ulonglong:
			{
				unsigned long long r = ((f0_ulonglong_f) handle) ();
				return qu_integer_newull (r);
			}
#endif

			case QU_STRUCT_ID_float:
			{
				double r = (double) ((f0_float_f) handle) ();
				return qu_float_newdx (r);
			}

			case QU_STRUCT_ID_double:
			{
				double r = ((f0_double_f) handle) ();
				return qu_float_newdx (r);
			}

#ifdef HAVE_LONG_DOUBLE
			case QU_STRUCT_ID_longdouble:
			{
				long double r = ((f0_longdouble_f) handle) ();
				return qu_decimal_newldx (r);
			}
#endif

			case QU_STRUCT_ID_size_t:
			{
#if SIZEOF_SIZE_T == SIZEOF_INT
				unsigned int r = ((f0_uint_f) handle) ();
				return qu_integer_newui (r);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
				unsigned long r = ((f0_ulong_f) handle) ();
				return qu_integer_newul (r);
#else
				QU_THROWZ (QuEImplement, "cant find the integer size of size_t");
#endif
			}


			case QU_STRUCT_ID_ssize_t:
			{
#if SIZEOF_SIZE_T == SIZEOF_INT
				int r = ((f0_uint_f) handle) ();
				return qu_integer_newi (r);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
				long r = ((f0_long_f) handle) ();
				return qu_integer_newl (r);
#else
				QU_THROWZ (QuEImplement, "cant find the integer size of ssize_t");
#endif
			}


			case QU_STRUCT_ID_off_t:
			{
#if SIZEOF_OFF_T == SIZEOF_INT
				int r = ((f0_int_f) handle) ();
				return qu_integer_newi (r);
#elif SIZEOF_OFF_T == SIZEOF_LONG
				long r = ((f0_long_f) handle) ();
				return qu_integer_newl (r);
#else
				QU_THROWZ (QuEImplement, "cant find the integer size of off_t");
#endif
			}


			case QU_STRUCT_ID_Boolean:
			{
				int r = ((f0_Boolean_f) handle) ();
				return r ? QuTrue : QuFalse;
			}

			case QU_STRUCT_ID_Int:
			{
				long r = ((f0_Int_f) handle) ();
				return qu_integer_newl (r);
			}

			case QU_STRUCT_ID_String:
			{
				char *r = ((f0_String_f) handle) ();
				if (r == NULL)
					return QuStringEMPTY;
				return qu_string_newz (r);
			}

			case QU_STRUCT_ID_Pointer:
			{
				void *r = ((f0_Pointer_f) handle) ();
				return qu_pointer_new (r);
			}

			case QU_STRUCT_ID_Proc:
			{
				qu_proc r = ((f0_Proc_f) handle) ();
				if (r == NULL)
					return QuNil; /* QuProcNULL */
				return r;
			}

			case QU_STRUCT_ID_Any:
			{
				qu_any r = ((f0_Any_f) handle) ();
				if (r == NULL)
					return QuNil;
				return r;
			}

#ifdef QU_SUPPORT_GLIB
			case QU_STRUCT_ID_GObject:
			{
				void *r = ((f0_GObject_f) handle) ();
				if (r == NULL)
					return QuNil;
				return __QuGObjectNew (r);
			}
#endif
		}
	}
	else
	{

#define f_(type) DLIB_CALL ((type) handle, arg)

		switch (*spec)
		{
			case '-':
			{
				f_(fn_f);
				return QuNil;
			}

			case '+':
			{
				f_(fn_f);
				return argv[0];
			}

			case QU_STRUCT_ID_char:
			{
				char r = f_(fn_char_f);
				return qu_string_byte (r);
			}

			case QU_STRUCT_ID_byte:
			{
				int r = f_(fn_byte_f);
				return qu_int_newi (r);
			}

			case QU_STRUCT_ID_ubyte:
			{
				unsigned int r = f_(fn_ubyte_f);
				return qu_int_newui (r);
			}

			case QU_STRUCT_ID_short:
			{
				short r = f_(fn_short_f);
				return qu_int_newh (r);
			}

			case QU_STRUCT_ID_ushort:
			{
				unsigned short r = f_(fn_ushort_f);
				return qu_int_newuh (r);
			}

			case QU_STRUCT_ID_int:
			{
				int r = f_(fn_int_f);
				return qu_integer_newi (r);
			}

			case QU_STRUCT_ID_uint:
			{
				unsigned int r = f_(fn_uint_f);
				return qu_integer_newui (r);
			}

			case QU_STRUCT_ID_int16:
			{
				int r = (int) f_(fn_int16_f);
				return qu_int_newi (r);
			}

			case QU_STRUCT_ID_uint16:
			{
				unsigned int r = (unsigned int) f_(fn_uint16_f);
				return qu_int_newui (r);
			}

			case QU_STRUCT_ID_int32:
			{
				long r = (long) f_(fn_int32_f);
				return qu_integer_newl (r);
			}

			case QU_STRUCT_ID_uint32:
			{
				unsigned long r = (unsigned long) f_(fn_uint32_f);
				return qu_integer_newul (r);
			}

			case QU_STRUCT_ID_long:
			{
				long r = f_(fn_long_f);
				return qu_integer_newl (r);
			}

			case QU_STRUCT_ID_ulong:
			{
				unsigned long r = f_(fn_ulong_f);
				return qu_integer_newul (r);
			}

#ifdef HAVE_LONG_LONG
			case QU_STRUCT_ID_longlong:
			{
				long long r = f_(fn_longlong_f);
				return qu_integer_newll (r);
			}

			case QU_STRUCT_ID_ulonglong:
			{
				unsigned long long r = f_(fn_ulonglong_f);
				return qu_integer_newull (r);
			}
#endif

			case QU_STRUCT_ID_float:
			{
				double r = (double) f_(fn_float_f);
				return qu_float_newdx (r);
			}

			case QU_STRUCT_ID_double:
			{
				double r = f_(fn_double_f);
				return qu_float_newdx (r);
			}

#ifdef HAVE_LONG_DOUBLE
			case QU_STRUCT_ID_longdouble:
			{
				long double r = f_(fn_longdouble_f);
				return qu_decimal_newldx (r);
			}
#endif

			case QU_STRUCT_ID_size_t:
			{
#if SIZEOF_SIZE_T == SIZEOF_INT
				unsigned int r = f_(fn_uint_f);
				return qu_integer_newui (r);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
				unsigned long r = f_(fn_uint_f);
				return qu_integer_newul (r);
#else
				QU_THROWZ (QuEImplement, "cant find the integer size of size_t");
#endif
			}


			case QU_STRUCT_ID_ssize_t:
			{
#if SIZEOF_SIZE_T == SIZEOF_INT
				int r = f_(fn_int_f);
				return qu_integer_newi (r);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
				long r = f_(fn_int_f);
				return qu_integer_newl (r);
#else
				QU_THROWZ (QuEImplement, "cant find the integer size of ssize_t");
#endif
			}


			case QU_STRUCT_ID_off_t:
			{
#if SIZEOF_OFF_T == SIZEOF_INT
				int r = f_(fn_int_f);
				return qu_integer_newi (r);
#elif SIZEOF_OFF_T == SIZEOF_LONG
				long r = f_(fn_int_f);
				return qu_integer_newl (r);
#else
				QU_THROWZ (QuEImplement, "cant find the integer size of off_t");
#endif
			}


			case QU_STRUCT_ID_Boolean:
			{
				int r = f_(fn_Boolean_f);
				return r ? QuTrue : QuFalse;
			}

			case QU_STRUCT_ID_Int:
			{
				long r = f_(fn_Int_f);
				return qu_integer_newl (r);
			}

			case QU_STRUCT_ID_String:
			{
				char *r = f_(fn_String_f);
				if (r == NULL)
					return QuStringEMPTY;
				return qu_string_newz (r);
			}

			case QU_STRUCT_ID_Pointer:
			{
				void *r = f_(fn_Pointer_f);
				return qu_pointer_new (r);
			}

			case QU_STRUCT_ID_Proc:
			{
				qu_proc r = f_(fn_Proc_f);
				if (r == NULL)
					return QuNil; /* QuProcNULL? */
				return r;
			}

			case QU_STRUCT_ID_Any:
			{
				qu_any r = f_(fn_Any_f);
				if (r == NULL)
					return QuNil;
				return r;
			}

#ifdef QU_SUPPORT_GLIB
			case QU_STRUCT_ID_GObject:
			{
				void *r = f_(fn_GObject_f);
				if (r == NULL)
					return QuNil;
				return __QuGObjectNew (r);
			}
#endif
		}

#undef f_

	}

	QU_FATAL ("invalid return type \"%c\"", *spec);
	return 0; /*never*/
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (path: String) -> Dlib
	Open a shared dynamic library.

	v{path} is the absolute path to the library file.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		libm = Dlib ('/usr/lib/libm.so')
		do_something_with_libm
		libm.close
*/
static qu_dlib qu__dlib_new (qu_class self, qu_string path)
{
	qu_dlib  d;
	void    *h;

	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	h = dlopen (path->str, RTLD_NOW | RTLD_GLOBAL);
	if (h == NULL)
		QU_THROWF (QuESystem, "%s", dlerror ());
	d = qu_gc_new (QuDlibClass);
	d->handle = h;
	return d;
}




/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__dlib__destroy (qu_dlib self)
{
	if (DLIB_OPENED (self))
		dlclose (self->handle);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __hash -> Int
	Return the hash value of the handler.

	@noexc
*/
static qu_int qu__dlib__hash  (qu_dlib self)
{
	return qu_int_new ((long) self->handle);
}



/*) sub __like (x) -> Boolean
	Return k{true] if v{x} is a Dlib that points to the same library,
	false otherwise.

	@noexc
*/
static qu_boolean qu__dlib__like (qu_dlib self, qu_any x)
{
	return (qu_dlib_is (x) && self->handle == ((qu_dlib) x)->handle) ? QuTrue : QuFalse;
}



/*) sub __true -> Boolean
	Return true if the library is opened, false otherwise.

	@noexc

	@xmp
		libm = Dlib ('/usr/lib/libm.so')
		libm.close
		println (libm ? 'yes' : 'no')

	@out
		no
*/
static qu_boolean qu__dlib__true (qu_dlib self)
{
	return DLIB_OPENED (self) ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub call (name: String, spec = "-:", *arg) -> Any
	Call a function in the shared library.

	v{name} is the name of the function.

	v{spec} is a string that must take this format:
	C{<return type>:<arguments>}.

	See -{Struct.spec} for valid I{return type} and I{arguments}.

	For the I{return type} you can specify C{"-"} to indicate that the function
	does not return a value e.g. C{void foo (int a, double b)}, in which
	case the result returned to the Qu program will be nil.

	You can also specify C{"+"} to return the first argument. In which case
	you must at least pass one argument.

	In addition you can also specify "n" to pass a NULL pointer argument.

	For a Pointer argument, the value passed is the underlying C pointer.
	Not the address of the Pointer object.

	Note that long long and long double are also supported provided that your
	system support it.

	@warn
	Be careful. Pass the wrong argument and your program may crash.

	@exc
	.EArgValue
	.EImplement the symbol does not exist

	@xmp
		libm = Dlib ('/usr/lib/libm.so')
		r = libm.call ('sin', 'f:f', 1.234)
		println (r)

	@out
		0.943818

	@p
	If you plan to use the function over and over again, then use #{func}
	instead of f{call}. It is more convenient and efficient.

	@note
	Use -{errno} to check or set the system v{errno} varaible.

	Currently it is not possible to pass NaN or infinities because Qu does not
	use floating point values to store infinities and NaN in Qu is an error.

	If the C function returns NaN then you get a EArgValue exception, and if it
	returns an infinity you get EInf, or possibly -{__infp__} or -{__infn__}
	depending on the -{Number.context}, i.e. whether infinities are allowed.

	@xmp
		libm = Dlib ('/usr/lib/libm.so')
		r = libm.call ('asin', 'f:f', 1.2)

	@out
		[1024] __main__ (EArgValue) invalid type or value

	@p
	As of 1.21.05 you can pass the following value for:
	@ul
	.C{"s"} : String Buffer Nil Pointer
	.C{"p"} : String Buffer Nil Pointer GObject
*/
static qu_any qu__dlib_call (qu_dlib self, int argc, qu_any *argv)
{
	qu_string  name, spec = SpecNoArg;
	void      *h;

	DLIB_ENSURE_OPENED (self);

	QU_ARG_VMIN  (1);
	QU_ARG_VTYPE (0, name, qu_string);
	if (argc > 1)
		QU_ARG_VTYPE (1, spec, qu_string);

	h = dlsym (self->handle, name->str);
	if (h == 0)
		QU_THROWF (QuEImplement, "symbol %s not found", name->str);

	QU_ENSURE (_dlib_check_spec (spec->str, argc - 2));
	return _dlib_call (h, spec->str, argc - 2, argv + 2);
}



/*) sub close -> Null
	Close this library object.

	If the library is already closed then nothing happens, thus this method
	never throw an exception.

	@noexc
*/
static qu_null qu__dlib_close (qu_dlib self)
{
	if (DLIB_OPENED (self))
	{
		dlclose (self->handle);
		self->handle = NULL;
	}
	return QuNil;
}



/*) sub func (name: String, spec = "-:") -> Method
	This creates a callable method thus allowing you to call it directly instead
	of using the #{call} method.

	It is much more efficient than #{call} because the specification is only
	checked once and there is no attribute lookup involved.
	It is generally about 3 times faster.

	@exc
	.ECall      library closed
	.EImplement symbol not found
	.EArgValue

	@xmp
		libm = Dlib ('/usr/lib/libm.so')
		sin = libm.func ('sin', 'f:f')
		r = sin (1.234)
		println (r)
		libm.close

	@out
		0.94381820937463368
*/
static qu_method qu__dlib_func (qu_dlib self, int argc, qu_any *argv)
{
	qu_string  name, spec = SpecNoArg;
	void      *h;
	qu_cfunc   f;

	DLIB_ENSURE_OPENED (self);

	if (QuCfuncClass == 0)
		qu__cfunc__setup ();

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE  (0, name, qu_string);
	if (argc > 1)
		QU_ARG_VTYPE (1, spec, qu_string);

	h = dlsym (self->handle, name->str);
	if (h == 0)
		QU_THROWF (QuEImplement, "symbol %s not found", name->str);

	QU_ENSURE (_dlib_check_spec (spec->str, -1));

	f         = qu_gc_new (QuCfuncClass);
	f->lib    = self;
	f->name   = name;
	f->spec   = spec;
	f->handle = h;

	return qu_method_new (_sub_cfunc_caller, f);
}



/*) sub get (name: String, type: String) -> Any
	Get the value of a variable.

	v{name} is the variable name and v{type} is the C type.

	See -{Struct.spec} for valid v{type}.

	@exc
	.ECall      library closed
	.EArgValue
	.EImplement symbol not found

	@warn
	Be careful. Pass the wrong type and your program may crash.

	@xmp
		# C library /foo/bar.so

			double TheValue = 1.234

		# Qu program

			d = Dlib ('/foo/bar.so')
			v = d.get ('TheValue', 'f')
			assert (v is Float)
			println (v)
			d.close

	@out
		1.234
*/
static qu_any qu__dlib_get (qu_dlib self, qu_string name, qu_any type)
{
	void *h;
	char  t;

	QU_ARG_TYPE (name, qu_string);
	DLIB_ENSURE_OPENED (self);

	t = qu_struct_type_to_char (type);
	QU_ARG_CHECKZ (t, "invalid type");

	h = dlsym (self->handle, name->str);
	if (h == 0)
		QU_THROWF (QuEImplement, "symbol %s not found", name->str);

	return qu_pointer_get (h, t);
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

int qu_dlib_check_spec (char *s, long len)
{
	qu_struct_type_s *i;
	char c;
	int  n;

	if (len < 1)
		return -1; /* invalid spec */

	c = s[0];
	if (!(c == '-' || c == '+'))
	{
		i = qu_struct_type_info (c);
		if (i->size == 0)
			return -1; /* invalid spec */
	}

	if (len >= 2)
	{
		c = s [1];
		if (c != ':')
			return -2; /* missing ':' */
		if (len == 2)
			return 0; /* OK, 0 argument */
	}

	len -= 2;
	n = len;

	if (n > DLIB_MAX_ARGS)
		return -3;

	s += 2;

	while (len--)
	{
		c = *s++;
		i = qu_struct_type_info (c);
		if (i->size == 0)
			return -1; /* invalid spec */
	}
	return n;
}



qu_any qu_dlib_call (void *h, int argc, qu_any *argv)
{
	qu_string name, spec = SpecNoArg;

	QU_ARG_VMIN  (1);
	QU_ARG_VTYPE (0, name, qu_string);
	if (argc > 1)
		QU_ARG_VTYPE (1, spec, qu_string);

	h = dlsym (h, name->str);
	if (h == 0)
		QU_THROWF (QuEImplement, "symbol %s not found", name->str);

	QU_ENSURE (_dlib_check_spec (spec->str, argc - 2));
	return _dlib_call (h, spec->str, argc - 2, argv + 2);
}


qu_method qu_dlib_cfunc_static (void *h, qu_string name, qu_string spec)
{
	qu_cfunc f;

	if (QuCfuncClass == 0)
		qu__cfunc__setup ();

	QU_ENSURE (_dlib_check_spec (spec->str, -1));

	f         = qu_gc_new_static (QuCfuncClass);
	f->lib    = 0;
	f->name   = qu_string_to_static (name);
	f->spec   = qu_string_to_static (spec);
	f->handle = h;

	return qu_method_new_static (_sub_cfunc_caller, f);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

#ifdef QU_SUPPORT_GLIB
qu_class QuGObjectClass;
qu_gobject_new_f __QuGObjectNew;
#endif

void qu_dlib__setup ()
{
	qu_class_setup_final (QuDlibClass, qu_dlib_s, qu__dlib__destroy, 0);
}



void qu_dlib__init ()
{
	SpecNoArg = qu_string_new_staticz ("-:");

	/* standard class methods */
	qu_class_method_class (QuDlibClass, QU_NAME_new, qu__dlib_new, 1);

	/* standard instance methods */
	qu_class_method_inst  (QuDlibClass, QU_NAME__hash, qu__dlib__hash, 0);
	qu_class_method_inst  (QuDlibClass, QU_NAME__like, qu__dlib__like, 1);
	qu_class_method_inst  (QuDlibClass, QU_NAME__true, qu__dlib__true, 0);

	/* specific instance methods */
	qu_class_method_instz (QuDlibClass, "call" , qu__dlib_call , -1);
	qu_class_method_instz (QuDlibClass, "close", qu__dlib_close,  0);
	qu_class_method_instz (QuDlibClass, "func" , qu__dlib_func , -1);
	qu_class_method_instz (QuDlibClass, "get"  , qu__dlib_get  ,  2);

	/* finalize declaration */
	qu_class_enddecl (QuDlibClass);
}
