/* Exception.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Exception
	This is the base class of all exceptions classes.

	Following is a list of all the builtin exception classes.

	@tree
	.#{EStandard}                  standard error
	.	#{ECompile}                compile error
	.		#{ESyntax}             syntax error
	.		#{EEos}                unexpected end of source
	.	#{ERuntime}                program error
	.		#{EAbstract}           object is abstract
	.		#{EAssert}             uncacthable failed assertion
	.			#{EProgram}        cacthable failed assertion
	.		#{ECall}               invalid call
	.			#{EArgCount}       bad number of arguments
	.			#{EArgValue}       invalid argument
	.		#{EData}               data error
	.			#{EDatabase}       database error
	.			#{ECorrupt}        corrupted data
	.			#{ETransaction}    transaction error
	.				#{ECommit}     commit error
	.				#{EConflict}   conflicting data
	.				#{EUndo}       undo operation failed
	.		#{EExit}               uncacthable exit
	.			#{EQuit}           cacthable exit
	.		#{EImplement}          not implemented
	.		#{EInterrupt}          keyboard interrupt
	.		#{EMemory}             out of memory
	.		#{EOverflow}           capacity overflow
	.		#{EPanic}              system in panic state
	.		#{ERecursion}          recursion limit exceeded
	.		#{ESecurity}           security error
	.		#{EStop}               stop iteration
	.		#{ETimeout}            operation timed-out
	.		#{EThread}             thread error
	.			#{EDeadlock}       deadlock condition
	.		#{EValue}              invalid type or value
	.			#{EAssign}         object not assignable
	.			#{EEnum}           value not in enumeration
	.			#{EFormat}         format error
	.			#{EIndex}          invalid index
	.				#{EKey}        no such key
	.			#{EMath}           math error
	.				#{EBrain}      unable to perform task
	.				#{ECmp}        objects not comparable
	.				#{EComplex}    result is a Complex
	.				#{EDiv}        invalid division
	.				#{EHuge}       result is too big
	.				#{EIand}       invalid bit-and operand
	.				#{EIdiv}       invalid integer division
	.				#{EIor}        invalid bit-or operand
	.				#{EInf}        result is infinity
	.				#{EIxor}       invalid bit-xor operand
	.				#{EMinus}      invalid subtraction
	.				#{EMul}        invalid multiplication
	.				#{EPlus}       invalid addition
	.				#{EPow}        invalid power
	.				#{ERem}        invalid remainder
	.				#{EShift}      invalid shift operand
	.					#{ELshift} invalid left shift operand
	.					#{ERshift} invalid right shift operand
	.				#{ETiny}       result is too small
	.			#{EModify}         can not modify object
	.			#{ENone}           no result
	.			#{EUndef}          result is undefined
	.#{ESystem}                    system error
	.	EPERM                      operation not permitted
	.	ENOENT                     no such file or directory
	.	...                        refer to C{man errno}

	@warn
	The exception default message is an instance property so to correctly
	set a default message you must do as demonstrated in the following
	example.

	Exception classes name must start with the uppercase letter C{'E'}
	followed by another uppercase letter then the rest. Otherwise the
	compiler will complain.

	@xmp
		class EFoo is ESystem
			var msg = 'default EFoo message'
		;;

		try
			throw EFoo
		catch
			println (__err__)
		;;

		try
			throw EFoo, "specialized EFoo message"
		catch
			println (__err__)
		;;

		try
			throw EFoo, [1, 2, 3]
		catch
			println (__err__)
		;;

	@out
		(EFoo) default EFoo message
		(EFoo) specialized EFoo message
		(EFoo) [1, 2, 3]

	@p
	#{EValue} and #{EArgCount} are generally not documented by functions
	or methods. If you get #{EArgCount}, then you know you are passing an
	invalid number of arguments to the function or method. If you get #{EValue}
	then you know you are passing an invalid value.

	@warn
	Conversely, if a function or method takes a number of arguments, then
	you know that #{EArgCount} or #{EValue} may be thrown even if the function
	or method does not document it.

	@p
	The exception classes under #{EMath} are designed to help debugging. The
	preferred exception thrown for an invalid addition operation was EValue
	because the operand is simply an argument to f{__plus}. But this does not help
	much if you are in a middle of an expression. Now when you get #{EPlus}, you
	know that a C{+} operation in the expression is invalid.

	@see
	-{__trace__} as a debugging aid.
*/
qu_class QuExceptionClass;


/*) include Exception.txt */


#define QU_EXCEPTION_IN_SOURCE_DECL
#include "Exceptions.h"

/* mapping errno to the corresponding classes */
static qu_class ErrnoClasses [LASTERRNO + 1];

/* the exit code */
volatile int __QuExceptionExitCode = -1;

/* strings */
static qu_string Smsg; /* "msg" */


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst -> Exception __final
	Create a new exception instance.

	@noexc
*/
static qu_exception qu__exception__inst (qu_class self)
{
	assert (qu_class_isa (self, QuExceptionClass));
	return qu_gc__new (self, self->size);
}



/*) sub class new (*arg) -> Exception
	Create a new exception instance then call the #{__init} method to
	initialize the object.

	See the #{__init} method for an example.

	@noexc

	@warn
	Any exception thrown by the f{__init} method is ignored. Thus this method
	never fails.
*/
qu_exception qu__exception_new (qu_class self, int argc, qu_any *argv)
{
	qu_exception e;
	qu_any       r;

	assert (qu_class_isa (self, QuExceptionClass));
	e = qu_gc__new (self, self->size);
	r = qu_class_init_new (self, e, argc, argv);
	if (r == 0)
		QU_SYS_WARN ("%s.__init threw an exception", self->name->str);
	return e;
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class defmsg -> Any
	Get the class default message. If the class does not specify one then
	the name of the class is returned.

	@noexc

	@xmp
		class EFoo is EStandard
			var msg = 'default foo message'
		;;

		println (EFoo.defmsg)

	@out
		deault foo message
*/
qu_any qu__exception_defmsg (qu_class self)
{
	qu_var v = qu_attr_find (self->attrs, Smsg);
	if (v && qu_var_is (v) && v->parent == (qu_symbol) self)
		return v->value;
	return self->name;
}



/*----------------------------------------------------------------------------
	specific static methods
----------------------------------------------------------------------------*/

/*) sub static getmsg -> Any
	Get the message associated with this exception instance.

	Note that the message may be anything, not necessarily a string. If none is
	specified then the class default message is returned.

	@noexc

	@xmp
		m = nil
		try
			throw EValue, [1, 2, 3]
		catch
			m = __err__.getmsg
		;;
		println (m)

	@out
		[1, 2, 3]
*/
qu_any qu__exception_getmsg (qu_exception x)
{
	/* this will get the "msg" attribute of *any* class but who cares ... */

	qu_class c = qu_classof (x);
	qu_any   v;

	__qu_attr_nothrow_begin;
	v = c->get (x, Smsg, 0);
	__qu_attr_nothrow_end;

	if (v == 0)
		return c->name;
	return v;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

/* none */



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __init (msg) -> Null
	Initialize a newly created exception instance.

	Note that v{msg} can be any arbitrary object, not just String.

	@noexc

	@xmp
		class EFoo is Exception
			var msg = 'foo error'
			sub __init (msg)
				println ('I am called')
				self.msg = msg
			;;
		;;
		try
			throw EFoo, 'hello'
		catch
			println (__err__)
		;;

	@out
		I am called
		(EFoo) hello
*/
static qu_null qu__exception__init (qu_exception self, int argc, qu_any *argv)
{
	if (argc)
	{
		if (self->attrs == 0)
			self->attrs = qu_attr_alloc ();
		qu_attr_set (self->attrs, Smsg, argv[0]);
	}
	return QuNil;
}



/*) sub __true -> Boolean __final
	This method always return true.
*/
#define qu__exception__true qu_object_method_true



/*) sub __print -> Null
	Print the exception to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		class EFoo is Exception
			var msg = 'foo error'
			sub __init (msg)
				self.msg = msg
			;;
			sub __print
				print ('Hey!')
			;;
		;;
		try
			throw EFoo, 'hello'
		catch
			println (__err__)
		;;

	@out
		Hey!
*/
static qu_null qu__exception__print (qu_exception self)
{
	qu_class  c;
	qu_any    m;
	qu_string s;

	c = qu_classof (self);
	m = qu__exception_getmsg (self);
	s = qu_sys_string (m);

	QU_ENSURE (qu_sys_printz ("(", 1));
	QU_ENSURE (qu_sys_printz (c->name->str, c->name->len));
	QU_ENSURE (qu_sys_printz (") ", 2));
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/*) sub __string -> String
	Get a String representation.

	@noexc
*/
qu_string qu__exception__string (qu_exception self)
{
	qu_class  c;
	qu_any    m;
	qu_string s;

	c = qu_classof (self);
	m = qu__exception_getmsg (self);
	s = qu_sys_string (m);

	return qu_string_newf ("(%s) %s", c->name->str, s->str);
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

qu_class qu_exception_errno (int eno)
{
	if (0 < eno && eno <= LASTERRNO)
	{
		switch (eno)
		{
			case ETIMEDOUT:  return QuETimeoutClass;
			case ENOMEM:     return QuEMemoryClass;
			default:         return ErrnoClasses [eno];
		}
	}
	return QuESystemClass;
}



qu_class qu_exception_class_new (qu_module m, qu_class sup, const char *name, const char *doc)
{
	qu_class c;

	if (!qu_exception_class_is (sup))
		QU_FATAL ("super class is not an Exception class");

	c = qu_class_newz (m, name, qu_exception_s, 0, 0);
	qu_class_set_super (c, sup);
	qu_class_prop (c, Smsg, qu_string_new_staticz (doc), QU_VAR_FLAG_INST);
	qu_class_enddecl (c);
	return c;
}



qu_string qu_exception_defmsg (qu_class ex)
{
	if (ex)
	{
		qu_any x = qu__exception_defmsg (ex);
		return qu_string_is (x) ? (qu_string) x : qu_sys_string (x);
	}
	else
	{
		return QuExceptionClass->name;
	}
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_exception__setup ()
{
	qu_class_setup (QuExceptionClass, qu_exception_s, 0, 0);
	Smsg = qu_string_new_staticz ("msg");
}



void qu_exception__init ()
{
	/* standard class methods */
	qu_class_method_class   (QuExceptionClass, QU_NAME__inst, qu__exception__inst, -1);
	qu_class_method_class   (QuExceptionClass, QU_NAME_new  , qu__exception_new  , -1);

	/* specific class methods */
	qu_class_method_classz  (QuExceptionClass, "defmsg", qu__exception_defmsg, 0);

	/* specific static methods */
	qu_class_method_staticz (QuExceptionClass, "getmsg", qu__exception_getmsg, 0);

	/* standard instance methods */
	qu_class_method_inst    (QuExceptionClass, QU_NAME__init  , qu__exception__init  , -1);
	qu_class_method_inst    (QuExceptionClass, QU_NAME__print , qu__exception__print ,  0);
	qu_class_method_inst    (QuExceptionClass, QU_NAME__string, qu__exception__string,  0);
	qu_class_method_inst    (QuExceptionClass, QU_NAME__true  , qu__exception__true  ,  0);

	/* specific instance properties */
	qu_class_prop_inst      (QuExceptionClass, Smsg, QuExceptionClass->name);

	/* finalize declaration */
	qu_class_enddecl (QuExceptionClass);

#define QU_EXCEPTION_IN_SOURCE_INIT
#include "Exceptions.h"

	/* set the uncatchable classes */
	qu_flag_set (QuEExitClass  , QU_CLASS_FLAG_NOCATCH);
	qu_flag_set (QuEAssertClass, QU_CLASS_FLAG_NOCATCH);
}
