/* File.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#error Get a real system
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#ifdef HAVE_GRP_H
#include <grp.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#ifdef HAVE_SYS_LIMITS_H
#include <sys/limits.h>
#endif

/* Fix EWOULDBLOCK = EAGAIN */
#ifdef EWOULDBLOCK
#  if (EAGAIN == EWOULDBLOCK)
#    define FIXEAGAIN QU_BEGIN QU_END
#  else
#    define FIXEAGAIN \
#       QU_BEGIN if (errno == EWOULDBLOCK) errno = EAGAIN; QU_END
#  endif
#else
#    define FIXEAGAIN
#endif

#ifndef PATH_MAX
#  if defined(_POSIX_PATH_MAX)
#    define PATH_MAX _POSIX_PATH_MAX
#  else
#    define PATH_MAX 4096
#  endif
#endif

#ifndef BUFSIZ
#define BUFSIZ 512
#endif

#ifndef S_IFHIDDEN
#define S_IFHIDDEN 1  /* crap! */
#endif



/*) class File is Stream  __final
	The File class deal with files.

	Make sure you have read the documenation for the -{Stream} class.

	File objects are instance of this class. You create them via the
	#{new} method.

	Once opened you can use the instance methods such as #{read} to read
	from the file, #{readln} to read a line, #{close} to close the file,
	#{seek} to set the file position, #{stat} to get file information, etc.

	File also provides a number of class methods such as #{list} to
	get a directory content, #{load} to load a file into memory, #{mkdir}
	to create directories recursively, etc.

	@warn
	All methods that take a v{path} argument accept an absolute path or
	a relative path. If you supply a relative path then it is taken
	relative to the current working directory as returned by #{getcwd}.

	Methods that change the file system are marked as k{__unsafe} thus
	untrusted modules can not access them.

	@p
	Always keep in mind that when you deal with the file system, then you are
	faced with race conditions. The file system may be modified by someone
	else at any time. Even by some thread in your own application without
	you being aware of it (you should, but you might forget).

	Race conditions may make the result of most class methods unreliable.
	For example, the #{exist} method might return true but a microsecond
	later someone deleted or renamed it. The #{compare} method might return
	true but a microsecond later someone else modified the files.

	The #{debug} method is a handy tool when you are debugging. Setting it
	to true prevents some methods from changing the file system.
	The #{dump} method for example, will not actually create the file,
	the #{chmod} method will not actually change the file mode, etc.
	Please check individual methods to see whether or not they support the
	debug mode.

	Note that initially #{debug} is set to true if you started Qu with the
	C{-d} option, false otherwise.

	@note
	Large files are not yet supported since they are generally a bad idea.
*/
qu_class QuFileClass;

struct qu_file_s
{
	QU_STREAM_HEAD; /* see qu_stream.h */
};


/*) const PATHSEP = '/'
	This is the path separator.
*/
qu_string QuFilePathSepStr;


/* da dot :-| */
static qu_string Sdot;

/* debug */
volatile int __QuFileDebug;


/*----------------------------------------------------------------------------
	for operations with a timeout
----------------------------------------------------------------------------*/

/* what time is now? */
#define NOW \
	time (NULL)



/*----------------------------------------------------------------------------
	current directory
----------------------------------------------------------------------------*/

static char CurDir [PATH_MAX + 1];
static int  CurDirLen;

static void _file_set_curdir ()
{
	if (getcwd (CurDir, sizeof (CurDir)) == NULL)
		QU_FATAL ("getcwd failed");
	CurDirLen = strlen (CurDir);
}



/*----------------------------------------------------------------------------
	permission
----------------------------------------------------------------------------*/

/* default permission used when creating new files */
#define DEFPERM    (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)
#define DEFPERMDIR (S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP| S_IXGRP | S_IROTH | S_IWOTH | S_IXOTH)

static const char *_file_get_rwx (const char *s, int *p, int r, int w, int x)
{
	int pr = 0, pw = 0, px = 0;

	while (*s)
	{
		switch (*s)
		{
			case 'r':
			case 'R':
				if (pr)
					return NULL;
				pr = 1;
				*p |= r;
				++s;
				continue;

			case 'w':
			case 'W':
				if (pw)
					return NULL;
				pw = 1;
				*p |= w;
				++s;
				continue;

			case 'x':
			case 'X':
				if (px)
					return NULL;
				px = 1;
				*p |= x;
				++s;
				continue;
		}
		break;
	}
	return s;
}



int qu_file_get_permission_mode (const char *s, int *p)
{
	int u = 0, g = 0, o = 0;
	*p = 0;

	while (*s)
	{
		switch (*s)
		{
			case 'u':
			case 'U':
				if (u)
					return 0;
				u = 1;
				s = _file_get_rwx (s + 1, p, S_IRUSR, S_IWUSR, S_IXUSR);
				break;

			case 'g':
			case 'G':
				if (g)
					return 0;
				g = 1;
				s = _file_get_rwx (s + 1, p, S_IRGRP, S_IWGRP, S_IXGRP);
				break;

			case 'o':
			case 'O':
				if (o)
					return 0;
				o = 1;
				s = _file_get_rwx (s + 1, p, S_IROTH, S_IWOTH, S_IXOTH);
				break;

			case ' ':
				++s;
				continue;

			default:
				return 0;
		}

		if (s == NULL)
			return 0;
	}
	return 1;
}



/*----------------------------------------------------------------------------
	open mode
----------------------------------------------------------------------------*/

int qu_file_get_open_mode (const char *s, int *mode, int *perm)
{
	int m, p = 0;
	int r, w, t, a, c, l, n, x, b, y;

	m = r = w = t = a = c = l = n = x = b = y = 0;
	while (*s)
	{
		switch (*s)
		{
			case ' ': ++s; continue;
			case 'r': case 'R': r++; ++s; continue;
			case 'w': case 'W': w++; ++s; continue;
			case 't': case 'T': t++; ++s; continue;
			case 'a': case 'A': a++; ++s; continue;
			case 'c': case 'C': c++; ++s; continue;
			case 'l': case 'L': l++; ++s; continue;
			case 'n': case 'N': n++; ++s; continue;
			case 'x': case 'X': x++; ++s; continue;
			case 'b': case 'B': b++; ++s; continue;
			case 's': case 'S': y++; ++s; continue;
			case '|': case ':': p=1; ++s; continue;
			default : return 0;
		}
	}

	if (r > 1
		|| w > 1
		|| t > 1
		|| a > 1
		|| c > 1
		|| l > 1
		|| n > 1
		|| x > 1
		|| b > 1
		|| y > 1)
		return 0;

	m = (r && w) ? O_RDWR : (w ? O_WRONLY : O_RDONLY);
	if (t) m |= O_TRUNC;
	if (a) m |= O_APPEND;
	if (c) m |= O_CREAT;
#ifdef O_NOFOLLOW
	if (l) m |= O_NOFOLLOW;
#endif
	if (n) m |= O_NONBLOCK;
	if (x) m |= O_EXCL;
#ifdef O_BINARY
	if (b) m |= O_BINARY;
#endif
#ifdef O_SYNC
	if (y) m |= O_SYNC;
#endif

	*mode = m;
	if (p)
		return qu_file_get_permission_mode (s, perm);
	return 1;
}



/*----------------------------------------------------------------------------
	filter mode
----------------------------------------------------------------------------*/

int qu_file_get_filter_mode (const char *s, int *filter)
{
	int f = 0;

	while (*s)
	{
		switch (*s)
		{
			case 's': case 'S': ++s; f |= S_IFSOCK;   break;
			case 'l': case 'L': ++s; f |= S_IFLNK;    break;
			case 'f': case 'F': ++s; f |= S_IFREG;    break;
			case 'b': case 'B': ++s; f |= S_IFBLK;    break;
			case 'd': case 'D': ++s; f |= S_IFDIR;    break;
			case 'c': case 'C': ++s; f |= S_IFCHR;    break;
			case 'i': case 'I': ++s; f |= S_IFIFO;    break;
			case 'h': case 'H': ++s; f |= S_IFHIDDEN; break;

			case 'v': case 'V':
#ifdef S_ISVTX
				++s; f |= S_ISVTX;
#endif
				break;

			case 'u': case 'U':
				++s;
				switch (*s)
				{
					case 'r': case 'R':
					case 'w': case 'W':
					case 'x': case 'X':
						s = _file_get_rwx (s, &f, S_IRUSR, S_IWUSR, S_IXUSR);
						if (s == NULL)
							return -1;
						break;
#ifdef S_ISUID
					default :
						f |= S_ISUID;
#endif
				}
				break;


			case 'g': case 'G':
				++s;
				switch (*s)
				{
					case 'r': case 'R':
					case 'w': case 'W':
					case 'x': case 'X':
						s = _file_get_rwx (s, &f, S_IRGRP, S_IWGRP, S_IXGRP);
						if (s == NULL)
							return -1;
						break;
#ifdef S_ISGID
					default :
						f |= S_ISGID;
#endif
				}
				break;


			case 'o': case 'O':
				++s;
				switch (*s)
				{
					case 'r': case 'R':
					case 'w': case 'W':
					case 'x': case 'X':
						s = _file_get_rwx (s, &f, S_IROTH, S_IWOTH, S_IXOTH);
						if (s == NULL)
							return -1;
						break;

					default:
						return -1;
				}
				break;

			case ' ':
				++s;
				continue;

			default:
				return -1;
		}
	}
	*filter = f;
	return 1;
}



/*-----------------------------------------------------------------------
	pattern matching
-----------------------------------------------------------------------*/

int qu_file_match (const char *pat, long lp, const char *name, long ln)
{
	register char *p, *s;

	p = (char *) pat;
	s = (char *) name;
	while (lp > 0)
	{
		switch (*p)
		{
			case '*':
				do
				{
					p++;
					lp--;

				} while (lp && (*p == '*' || *p == '?'));

				if (lp == 0)
					return 1;

				while (ln > 0)
				{
					if (*s == *p)
						goto next;
					ln--;
					s++;
				}
				return 0;

			case '?':
				if (ln < 1 || *s == '.')
					return 0;
				break;

			default:
				if (ln < 1 || *s != *p)
					return 0;
		}
	  next:
		s++;
		ln--;
		p++;
		lp--;
	}
	return (ln == 0);
}



/*----------------------------------------------------------------------------
	path
----------------------------------------------------------------------------*/

static int _file_stat_path (const char *path, struct stat *fs)
{
	while (stat (path, fs) != 0)
	{
		if (errno != EINTR)
			return -1;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	tools
----------------------------------------------------------------------------*/

#define CLOSEFD(fd) QU_BEGIN while (close (fd) && errno == EINTR) {} QU_END

static int _file_stat (int fd, struct stat *fs)
{
	while (fstat (fd, fs))
	{
		if (errno != EINTR)
			return -1;
	}
	return 0;
}


static size_t _file_size (int fd)
{
	struct stat fs;

	if (_file_stat (fd, &fs) != 0)
		return -1;
	return fs.st_size;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (path: String, mode = 'r', timeout = 30) -> File
	Open an existing file or create a new file.

	v{mode} must be one of:
	@ops
	.| C{'r'}  | for read           | O_RDONLY
	.| C{'w'}  | for write          | O_WRONLY
	.| C{'rw'} | for read and write | O_RDWR
	.| C{'a'}  | for append only    | O_APPEND

	@p
	possibly combined with:
	@ops
	.| C{'t'} | truncate after opened    | O_TRUNC
	.| C{'c'} | create if does not exist | O_CREAT
	.| C{'l'} | dont follow link         | O_NOFOLLOW
	.| C{'x'} | exclusive                | O_EXCL
	.| C{'s'} | auto synchronize         | O_SYNC

	@p
	And if you specify a creation mode then followed by ':' and permissions
	characters. See #{chmod} for valid values.

	You can optionally specify the number of seconds to v{timeout}.

	All newly opened files have the #{linebreak} marker set to C{'\n'}
	and a #{timeout} set to a huge value.

	All files are opened in non-blocking mode as this is the preferred
	way to work with Qu -{Stream} objects.

	@exc
	.ESecurity
	.EArgValue
	.ETimeout
	.ESystem

	@xmp
		f = File.open ('/foo/bar.txt', 'rwcxln:urwxgrwor', 30)
*/
static qu_file qu__file_new (qu_class self, int argc, qu_any *argv)
{
	qu_string path, m = qu_string_byte ('r');
	int       mode = O_RDONLY;
	int       perm = DEFPERM;
	long      to = 30;
	int       fd;
	qu_file   f;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
	{
		QU_ARG_VTYPE (1, m, qu_string);
		QU_ARG_CHECK (qu_file_get_open_mode (m->str, &mode, &perm));
		if (argc > 2)
			QU_ARG_vlong (2, to);
	}

	mode |= O_NONBLOCK;
	to += NOW;

	__qu_blocking_begin;
	for (;;)
	{
		if (mode & O_CREAT)
			fd = open (path->str, mode, perm);
		else
			fd = open (path->str, mode);

		if (fd >= 0)
			break;

		if (errno != EINTR)
		{
			FIXEAGAIN;
			switch (errno)
			{
				case ETXTBSY:
				case EAGAIN:
					break;

				default:
					__qu_blocking_end;
					QU_THROWERRNO;
			}
		}
		if (NOW > to)
		{
			__qu_blocking_end;
			QU_THROW (QuETimeout);
		}
		__qu_blocking_check_signal;
	}
	__qu_blocking_end;
	f = (qu_file) qu_stream_new (QuFileClass);
	f->flags |= QU_STREAM_FLAG_CANEOF | QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANSELECT | QU_STREAM_FLAG_CANCLOSE;
	f->fd = fd;
	return f;
}



/*) sub class temp -> Array
	Create a new empty temporary file.

	The returned value is an Array of two items: the opened File object
	then the absolute path.

	See C{man mkstemp} for details.

	@exc
	.ESystem

	@xmp
		(f, path) = File.temp
*/
static qu_array qu__file_temp (qu_class self)
{
	char    template [PATH_MAX] = QU_APP_TMPDIR QU_PATH_SEPSTR "QuXXXXXX";
	int     fd;
	qu_file f;

	__qu_blocking_begin;
	fd = mkstemp (template);
	if (fd < 0)
	{
		__qu_blocking_end;
		QU_THROWERRNO;
	}
	__qu_blocking_end;

	f         = (qu_file) qu_stream_new (QuFileClass);
	f->flags |= QU_STREAM_FLAG_CANEOF | QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANSELECT | QU_STREAM_FLAG_CANCLOSE;
	f->fd     = fd;
	return qu_array_new_ex (2, f, qu_string_newz (template));
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class agecmp (src: String, dst: String, timeout = Int.MAX) -> Int
	Compare the modification date of two files.

	This method return -1 if v{src} is older than v{dst}, 0 if they are the
	same, 1 otherwise.

	You can optionally specify the number of seconds to v{timeout}.

	Note that this method works with directories as well.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		r = File.agecmp ('/foo/bar.txt', '/bar/foo.txt')
*/
static qu_int qu__file_agecmp (qu_class self, int argc, qu_any *argv)
{
	qu_string   src, dst;
	long        to = 0;
	struct stat fs, fd;

	__qu_unsafe__;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, src, qu_string);
	QU_ARG_VTYPE (1, dst, qu_string);

	if (argc > 2)
	{
		QU_ARG_vlong (2, to);
		to += NOW;
	}

	while (stat (src->str, &fs) < 0)
	{
		if (errno != EINTR)
			QU_THROWERRNO;
		QU_REQUIRE (QuETimeout, to == 0 || NOW <= to);
	}

	while (stat (dst->str, &fd) < 0)
	{
		if (errno != EINTR)
			QU_THROWERRNO;
		QU_REQUIRE (QuETimeout, to == 0 || NOW <= to);
	}

	if (fs.st_mtime < fd.st_mtime)
		return QuInt_1;
	else if (fs.st_mtime == fd.st_mtime)
		return QuInt0;
	else
		return QuInt1;
}



/*) sub class chdir (path: String) -> Null
	Change the current directory to v{path}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		File.chdir ('/foo/bar')
		File.chdir ('baz')

	@see
	#{getcwd}
*/
static qu_null qu__file_chdir (qu_class self, qu_string path)
{
	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	if (chdir (path->str) != 0)
		QU_THROWERRNO;
	_file_set_curdir ();
	return QuNil;
}



/*) sub class chroot (path: String) -> Null
	Change the root directory for the whole application.

	@warn
	Use with care. You should generally never call this function unless you
	have considered all the consequences.

	See C{man 2 chroot} for details.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		File.chroot ('/foo/bar')

	@see
	#{chdir}
*/
static qu_null qu__file_chroot (qu_class self, qu_string path)
{
	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	if (chroot (path->str) != 0)
		QU_THROWERRNO;
	return QuNil;
}



/*) sub class compare (src: String, dst: String, timeout = Int.MAX) -> Boolean
	Return true if both file content are the same, false otherwise.

	v{src} is the source file path and v{dst} is the destination file path.

	You can optionally specify the number of seconds to v{timeout}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		println ('nope') unless File.compare ('/foo/bar.txt', '/bar/foo.txt')
*/
static qu_boolean qu__file_compare (qu_class self, int argc, qu_any *argv)
{
	qu_string  src, dst;
	char       bs[BUFSIZ], bd[BUFSIZ];
	qu_stream  fs, fd;
	size_t     ns, nd;
	long       ls, ld;
	long       to = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, src, qu_string);
	QU_ARG_VTYPE (1, dst, qu_string);

	if (argc > 2)
		QU_ARG_vlong (2, to);

	fs = (qu_stream) qu__file_new (QuFileClass, 1, (qu_any *) &src);
	if (fs == 0)
		goto err0;

	fd = (qu_stream) qu__file_new (QuFileClass, 1, (qu_any *) &dst);
	if (fd == 0)
		goto err1;

	ns = _file_size (fs->fd);
	if (ns < 0)
		goto err2;

	nd = _file_size (fd->fd);
	if (nd < 0)
		goto err2;

	if (ns != nd)
		goto nope;

	if (argc > 2)
		fd->timeout = fs->timeout = NOW + to;

	for (;;)
	{
		ls = qu_stream_read (fs, bs, BUFSIZ, 0);
		if (ls < 0)
			goto err2;

		if (ls == 0)
			break;

		ld = qu_stream_read (fd, bd, ls, 1);
		if (ld < 0)
			goto err2;

		if (ls != ld)
			goto nope;

		if (memcmp (bs, bd, ls))
			goto nope;
	}
	qu_stream_close (fs);
	qu_stream_close (fd);
	return QuTrue;

  nope:
	qu_stream_close (fs);
	qu_stream_close (fd);
	return QuFalse;

  err2: qu_stream_close (fd);
  err1: qu_stream_close (fs);
  err0: QU_RETHROW;
}



/*) sub class contains (path: String, s: String, timeout = Int.MAX) -> Boolean
	Whether or not the file contains the specified string.

	You can optionally specify a v{timeout} in seconds.

	@exc
	.ESecurity
	.ESystem
	.EArgValue
*/
static qu_boolean qu__file_contains (qu_class self, int argc, qu_any *argv)
{
	qu_string  path, s;
	qu_stream  f;
	qu_boolean r;
	long       to = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	QU_ARG_VTYPE (1, s, qu_string);
	if (argc > 2)
		QU_ARG_vlong (2, to);

	f = (qu_stream) qu__file_new (self, 1, (qu_any *) &path);
	QU_ENSURE (f);

	if (argc > 2)
		f->timeout = NOW + to;

	r = qu_stream_contains (f, s);
	qu_stream_close (f);
	return r;
}



/*) sub class copy (src: String, dst: String, timeout = Int.MAX) -> Null
	Copy file v{src} to v{dst}.

	v{src} is the source file path and v{dst} is the destination file path.

	You can optionally specify a v{timeout} in seconds.

	@warn
	If v{dst} does not exist, it will be created. If it does, it will be
	truncated before the copy operation.

	If a ETimeout occurs in the middle of the process, then v{dst} is not
	deleted and will contain part of v{src}.

	@exc
	.ESecurity
	.EArgValue
	.EValue     v{src} is the same as v{dst}
	.ETimeout
	.ESystem

	@xmp
		File.copy ('/foo/bar.txt', '/bar/foo.txt')

	@note
	If #{debug} is true then copy is not actually performed and success
	is assumed. Note however that an exception is still thrown if you pass
	the wrong arguments.
*/
static qu_null qu__file_copy (qu_class self, int argc, qu_any *argv)
{
	qu_string src, dst;
	qu_stream fs, fd;
	long      ls, ld;
	char      buf [BUFSIZ];
	long      to = 0;
	qu_any    arg[2];

	__qu_unsafe__;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, src, qu_string);
	QU_ARG_VTYPE (1, dst, qu_string);
	QU_ARG_CHECK (!qu_string_EQ (src, dst));

	if (argc > 2)
		QU_ARG_vlong (2, to);

	if (__QuFileDebug)
		return QuNil;

	fs = (qu_stream) qu__file_new (QuFileClass, 1, (qu_any *) &src);
	if (fs == 0)
		goto err0;

	arg[0] = dst;
	arg[1] = qu_string_newz ("rwct");
	fd = (qu_stream) qu__file_new (QuFileClass, 2, arg);
	if (fd == 0)
		goto err1;

	if (argc > 2)
		fd->timeout = fs->timeout = NOW + to;

	for (;;)
	{
		ls = qu_stream_read (fs, buf, BUFSIZ, 0);
		if (ls < 0)
			goto err2;
		if (ls == 0)
			break;
		ld = qu_stream_write (fd, buf, ls, 1);
		if (ld < 0)
			goto err2;
		if (ls != ld)
		{
			qu_thread_throw (QuEIOClass, 0);
			goto err2;
		}
	}
	qu_stream_close (fs);
	qu_stream_close (fd);
	return QuNil;

  err2: qu_stream_close (fd);
  err1: qu_stream_close (fs);
  err0: QU_RETHROW;
}



/*) sub class debug (setting: Boolean) -> Boolean
	Set whether or not File operations that modify the filesystem should
	actually be performed.

	If true then some methods such as #{chmod}, #{dump}, #{copy} etc. are not
	actually performed and a success is assumed as if nothing went wrong.

	Note however that read/write operations are not affected by f{debug} because
	the logic is, if you open a file for writine then it means you intend to
	actually do some writing.

	The initial value is the same as -{__debug__}, that is, if you started
	Qu with the C{-d} option then the initial value is set true,
	and false otherwise.

	This method returns the previous setting.

	@warn
	Since the value may be changed by any other threads, you must generally
	not rely on this method to protect your application from doing harm to
	the file system  unless you are certain that you are in total control.

	Consider using -{__debug__} instead. The -{__debug__} value is set by
	Qu and can not be changed by applications.

	@xmp
		File.copy (foo, bar) unless __debug__

		# is probably better than

		File.debug (true)
		File.copy (foo, bar)

	@exc
	.ESecurity
	.EArgValue
*/
static qu_boolean qu__file_debug (qu_class self, int argc, qu_any *argv)
{
	qu_boolean old = __QuFileDebug ? QuTrue : QuFalse;

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc > 0)
		QU_ARG_vbool (0, __QuFileDebug);
	return old;
}



/*) sub class delete (path: String) -> Null
	Delete the file specified by v{path}.

	This function checks for the root directory by means of #{isroot}.
	If f{isroot} returns true then this function will throw EPERM.

	Note that you can not delete a directory with this method. Use the
	#{rmdir} method for that.

	See C{man unlink}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		File.delete ('/foo/bar.txt')

	@note
	If #{debug} is true then delete is not actually performed and success
	is assumed. Note however that an exception is still thrown if you pass
	the wrong arguments.
*/
static qu_null qu__file_delete (qu_class self, qu_string path)
{
	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	QU_REQUIRE (QuEPERM, !qu_file_is_root (path));
	if (!__QuFileDebug)
	{
		__qu_blocking_begin;
		while (unlink (path->str) != 0)
		{
			if (errno != EINTR)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class dir (path: String) -> String
	Get the directory part from v{path}.

	@exc
	.EArgValue

	@xmp
		println (File.dir ('foo/bar/baz.txt'))
		println (File.dir ('/foo/bar/baz.txt'))
		println (File.dir ('foo.txt'))
		println (File.dir ('/foo.txt'))
		println (File.dir ('/'))

	@out
		foo/bar
		/foo/bar

		/
		/
*/
static qu_string qu__file_dir (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_file_dir (path);
}



/*) sub class dump (path: String, s: String, timeout = Int.MAX) -> Null
	Store the content of v{s} to file.

	v{path} is the absolute path of the file.

	You can optionally set a v{timeout} in seconds.

	@warn
	If the file does not exist, it will be created. If it does, it will be
	truncated.

	If a ETimeout occurs in the middle of the process then only part of
	v{s} is written.

	@exc
	.ESecurity
	.EArgValue
	.EIndex
	.ETimeout
	.ESystem

	@xmp
		File.dump ('/foo/hello.txt', 'Hello World!')

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
qu_null qu__file_dump (qu_class self, int argc, qu_any *argv)
{
	static qu_string rwct = 0;

	qu_string path, s;
	qu_stream f;
	long      to = 0;
	qu_any    arg[2];

	__qu_unsafe__;

	if (rwct == 0)
		rwct = qu_string_new_staticz ("rwct");

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	QU_ARG_VTYPE (1, s, qu_string);
	if (argc > 2)
		QU_ARG_vlong (2, to);

	if (__QuFileDebug)
		return QuNil;

	arg[0] = path;
	arg[1] = rwct;

	f = (qu_stream) qu__file_new (QuFileClass, 2, arg);
	QU_ENSURE (f);

	if (argc > 2)
		f->timeout = NOW + to;

	to = qu_stream_write (f, s->str, s->len, 1);
	qu_stream_close (f);
	QU_ENSURE (to >= 0);
	return QuNil;
}



/*) sub class eachfile (path: String) -> Iterator
	Return an -{Iterator} object to iterate through the files in the directory
	specified by v{path}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		for f in File.eachfile ('/usr/bin')
			f.startwith ('info') or continue
			println (f)
		;;

	@out
		infocmp
		infotocap
		info
*/
typedef struct qu_dir_s qu_dir_s;
struct qu_dir_s
{
	QU_OPAQUE_HEAD;
	DIR *dir;
};

typedef qu_dir_s *Dir;


static void _dir_free (Dir self)
{
	if (self->dir)
		closedir (self->dir);
}


static qu_any _file_each_next (Dir self, qu_iterator i)
{
	struct dirent *entry;

	QU_REQUIRE (QuEStop, self->dir);
	entry = readdir (self->dir);
	if (entry == NULL)
	{
		self->dir = NULL;
		QU_THROW (QuEStop);
	}
	return qu_string_newz (entry->d_name);
}



static qu_iterator qu__file_eachfile (qu_class self, qu_string path)
{
	DIR        *dir;
	Dir         d;
	qu_iterator i;

	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	dir = opendir (path->str);
	if (dir == NULL)
		QU_THROWERRNO;
	d = (Dir) qu_opaque_new (NULL, _dir_free, sizeof (qu_dir_s));
	d->dir = dir;
	QU_ITERATOR_NEW_EXTERN (i, d, _file_each_next);
	return i;
}



/*) sub class exist (path: String, type = 'f') -> Boolean
	Return true if the file v{path} exist, false otherwise.

	v{type} is a filter. See #{istype} for valid values.

	Note that false is returned if the file can not be accessed.
	See C{man 2 access} for more information.

	@exc
	.ESecurity
	.EArgValue

	@xmp
		unless File.exist ('/foo/bar.txt')
			println ('did not exist a while ago')
		;;
*/
static qu_boolean qu__file_exist (qu_class self, int argc, qu_any *argv)
{
	qu_string path, type = qu_string_byte ('f');

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
		QU_ARG_VTYPE (1, type, qu_string);
	return qu_file_exist (path, type) ? QuTrue : QuFalse;
}



/*) sub class ext (path: String) -> String
	Get file extension from v{path}.

	@exc
	.EArgValue

	@xmp
		p = 'foo/bar/baz.tar.gz'
		println (File.ext (p))

	@out
		.tar.gz
*/
static qu_string qu__file_ext (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_file_ext (path);
}



/*) sub class fchmod (path: String, perm: String) -> Null
	Set file permission.

	v{path} is the file to modify.

	See #{chmod} for valid v{perm} values.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		File.fchmod ('/foo/bar.txt', 'urwx gwr or')

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static qu_null qu__file_fchmod (qu_class self, qu_string path, qu_string perm)
{
	int p;

	__qu_unsafe__;

	QU_ARG_TYPE (path, qu_string);
	QU_ARG_TYPE (perm, qu_string);
	QU_REQUIRE (QuEPERM, !qu_file_is_root (path));
	QU_ARG_CHECK (qu_file_get_permission_mode (perm->str, &p));

	if (!__QuFileDebug)
	{
		__qu_blocking_begin;
		if (chmod (path->str, p) != 0)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class fchown (*arg) -> Null
	@def
	sub class fchown (path: String, user: String) -> Null
	sub class fchown (path: String, user: String, group: String) -> Null

	@p
	Set file owner.

	v{path} is the file to modify.

	v{user} and v{group} must be valid user and group name. If called with
	two arguments then v{group} is the user group as returned by the system
	function f{getpwnam}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		File.fchown ('/foo/bar.txt', 'Foo', 'foes')

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static qu_null qu__file_fchown (qu_class self, int argc, qu_any *argv)
{
	qu_string path, usr, grp = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	QU_ARG_VTYPE (1, usr, qu_string);
	if (argc > 2)
		QU_ARG_VTYPE (2, grp, qu_string);

	if (qu_file_is_root (path))
		QU_THROW (QuEPERM);

	if (!__QuFileDebug)
	{
		struct passwd *u;

		__qu_blocking_begin;
		u = getpwnam (usr->str);
		if (u != NULL)
		{
			uid_t uid = u->pw_uid;
			gid_t gid;

			if (grp == 0 || grp->len == 0)
			{
				gid = u->pw_gid;
			}
			else
			{
				struct group *g = getgrnam (grp->str);

				if (g == NULL)
				{
					__qu_blocking_end;
					QU_THROWERRNO;
				}
				gid = g->gr_gid;
			}
			if (chown (path->str, uid, gid) != 0)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class fixdir (path: String) -> String
	Return v{path} that ends with a directory separator.

	@exc
	.EArgValue
	.ESystem

	@xmp
		p = File.fixdir ('foo/bar')
		println (p)

	@out
		foo/bar/
*/
static qu_string qu__file_fixdir (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	if (path->str[path->len - 1] == QU_PATH_SEPCHR)
		return path;
	return qu_string_cat (path, qu_string_byte (QU_PATH_SEPCHR), NULL);
}



/*) sub class ftruncate (path: String, len = 0) -> Null
	Truncate the file to a specified length.

	v{len} must be a non-negative integer.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static qu_null qu__file_ftruncate (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	long      len = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
	{
		QU_ARG_vlong (1, len);
		QU_ARG_CHECK  (len >= 0);
	}
	if (!__QuFileDebug)
	{
		__qu_blocking_begin;
		if (truncate (path->str, (off_t) len) < 0)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class getcwd -> String
	Get the current directory.

	@xmp
		println (File.getcwd)

	@see
	#{chdir}
*/
static qu_string qu__file_getcwd (qu_class self)
{
	return qu_string_newz (CurDir);
}



/*) sub class isabs (path: String) -> Boolean
	Return true if v{path} is an asbsolute path, v{false} otherwise.

	@exc
	.EArgValue

	@xmp
		println (File.isabs ('/foo/bar.txt'))
		println (File.isabs ('foo/bar.txt'))

	@out
		true
		false

	@see
	#{isdir} #{istype} #{isroot} #{toabs}
*/
static qu_boolean qu__file_isabs (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_file_is_abs (path) ? QuTrue : QuFalse;
}



/*) sub class isroot (path: String) -> Boolean
	Return true if v{path} is the root directory, false otherwise.

	Note that the meaning of the "root" directory is C{'/'} unless you
	have changed it with #{chroot}.

	This function account for back-references and the current working
	directory as set by #{chdir}.

	@exc
	.EArgValue

	@xmp
		File.chdir ('/')
		println (File.isroot ('/'))
		println (File.isroot (' / '))
		File.chdir ('/tmp')
		println (File.isroot ('../'))

	@out
		true
		false
		true

	@see
	#{isdir} #{isabs} #{istype}
*/
static qu_boolean qu__file_isroot (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_file_is_root (path) ? QuTrue : QuFalse;
}



/*) sub class istype (f, filtermode: String, strict = true) -> Boolean
	Check whether a file is of a certain type.

	v{f} may be a File or a path String.

	If v{strict} is true then throw if the file does not exist, return false
	otherwise.

	v{filtermode} must be a combination of:
	@ops
	.| C{s} | socket
	.| C{l} | symbolic link
	.| C{f} | regular file
	.| C{b} | block device
	.| C{d} | directory
	.| C{c} | character device
	.| C{i} | fifo
	.| C{h} | hidden file
	.| C{v} | sticky bit
	.| C{u} | set UID bit, or user permission if followed by a combination of rwx
	.| C{g} | set GID bit, or group permission if followed by a combination rwx
	.| C{o} | others permission followed by a combination of rwx

	@p
	You can separate any characters with any number of spaces and uppercase
	letters are interpreted the same as lowercase letters.

	Types not understood by the system are simply ignored, but using characters
	other than listed above is an error.

	Note that the permission charaters uses the same format as #{chmod}.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@xmp
		println (File.istype ('/tmp', 'f'))
		println (File.istype ('/tmp', 'd'))
		println (File.istype ('/tmp', 'fd'))
		println (File.istype ('/tmp', 'F D Urwx Grw'))

	@out
		false
		true
		true
		true

	@see
	#{isdir} #{isabs} #{isroot}
*/
static qu_boolean qu__file_istype (qu_class self, int argc, qu_any *argv)
{
	qu_any    f;
	qu_string filtertype;
	int       strict = 1, r;

	__qu_unsafe__;

	QU_ARG_VCOUNT (2, 3);
	f = argv[0];
	QU_ARG_VTYPE (1, filtertype, qu_string);
	if (argc > 2)
		QU_ARG_vbool (2, strict);

	if (qu_string_is (f))
	{
		r = qu_file_is_type ((qu_string) f, filtertype);
		if (errno)
		{
			if (strict)
				QU_RETHROW;
			return QuFalse;
		}
		return r ? QuTrue : QuFalse;
	}
	else
	{
		struct stat fs;
		int fd, m;

		QU_ARG_CHECK (qu_file_is (f));
		QU_ARG_CHECK (qu_file_get_filter_mode (filtertype->str, &m));

		fd = ((qu_file) f)->fd;
		if (fd < 0)
		{
			if (strict)
			QU_THROW (QuEBADF);
			return QuFalse;
		}
		else
		{
			__qu_blocking_begin;
			r = _file_stat (fd, &fs);
			__qu_blocking_end;
			if (r != 0)
			{
				if (strict)
					QU_THROWERRNO;
				return QuFalse;
			}
			if (fs.st_mode & m)
				return QuTrue;
		}
		return QuFalse;
	}
}



/*) sub class isdir (f, strict = true)
	Return true if v{f} is a directory, false otherwise.

	This is a convenience method that calls #{istype}C{(f, 'd', strict)}.
*/
static qu_boolean qu__file_isdir (qu_class self, int argc, qu_any *argv)
{
	QU_ARG_VMIN (1);
	{
		qu_any arg[3] = {argv[0], qu_string_byte ('d'), (argc > 1) ? argv[1] : QuTrue};
		return qu__file_istype (self, 3, arg);
	}
}



/*) sub class list (path: String, filtermode = '', filtername = '*', strip = false) -> Set
	List all files in the directory v{path}.

	v{filtername} is a file pattern such as C{'foo.*'}, C{'f?o.ba?'}, etc.

	See #{istype} for valid v{filtermode}.
	@ops
	.| C{f} | regular file  | S_IFREG
	.| C{d} | directory     | S_IFDIR
	.| C{l} | symbolic link | S_IFLNK
	.| C{h} | hidden file   | S_IFHIDDEN

	@exc
	.EArgValue
	.ESecurity
	.ESystem

	@xmp
		for f in File.list ('/usr/bin')
			println (f)
		;;
*/
static qu_set qu__file_list (qu_class self, int argc, qu_any *argv)
{
	qu_string path, filtertype = 0, filtername = 0;
	int       strip = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 4);
	path = argv[0];
	if (argc > 1)
	{
		filtertype = argv[1];
		if (filtertype == (qu_string) QuNil)
			filtertype = 0;

		if (argc > 2)
		{
			filtername = argv[2];
			if (filtername == (qu_string) QuNil)
				filtername = 0;

			if (argc > 3)
				QU_ARG_vbool (3, strip);
		}
	}

	QU_ARG_CHECK (qu_string_is (path)
		&& (filtertype == 0 || qu_string_is (filtertype))
		&& (filtername == 0 || qu_string_is (filtername)));

	if (filtertype == 0)
		filtertype = qu_string_byte ('f');
	return qu_file_list (path, filtertype, filtername, strip);
}



/*) sub class load (path: String, max = -1, timeout = 180) -> String
	Read an entire file into a String.

	Optionally set a timeout in seconds.

	@exc
	.ESecurity
	.EArgValue
	.EMemory
	.ESystem

	@xmp
		s = File.load ('/foo/bar.txt')

	@see
	-{Stream.__pop}
*/
static qu_string qu__file_load (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	long      to = 180;
	long      max = -1;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
	{
		QU_ARG_vlong (1, max);
		if (argc > 2)
			QU_ARG_vlong (2, to);

		if (max <= 0 || max > QU_MEM_MAX)
			max = -1;
	}
	return qu_file_load_ex (path, NOW + to, max, NULL, NULL);
}



/*) sub class make (path: String, defdir = '', defext = '') -> String
	Make a file path.

	If v{path} is an absolute path then v{defdir} is ignored.
	Otherwise v{defdir} is prepended unless it is empty. If v{path} already
	contains a file extension then v{defext} is ignored. Otherwise it is
	appended unless it is empty.

	@exc
	.EArgValue
*/
static qu_string qu__file_make (qu_class self, int argc, qu_any *argv)
{
	qu_string path, defdir = 0, defext = 0;

	QU_ARG_VCOUNT (1, 3);
	path = argv[0];
	if (argc > 1)
	{
		defdir = argv[1];
		if (argc > 2)
			defext = argv[2];
	}
	QU_ARG_CHECK (qu_string_is (path)
		&& (defdir == 0 || qu_string_is (defdir))
		&& (defext == 0 || qu_string_is (defext)));
	return qu_file_make (path, defdir, defext);
}



/*) sub class match (pattern: String, name: String) -> Boolean
	Return true if v{name} matches v{pattern}, false otherwise.

	v{pattern} is a file matching pattern, e.g. C{'*.foo'}, C{'f?o.b?r'}.

	@exc
	.EArgValue
*/
static qu_boolean qu__file_match (qu_class self, qu_string pattern, qu_string name)
{
	QU_ARG_TYPE (pattern, qu_string);
	QU_ARG_TYPE (name, qu_string);
	return qu_file_match (pattern->str, pattern->len, name->str, name->len) ? QuTrue : QuFalse;
}



/*) sub class mkdir (path: String, all = false) -> Null
	Make directory.

	If v{all} is true then make all sub-directories in v{path}.

	The permission used to create new directories is C{'urwxgrwxor'}.

	Note that you can change the owner and permission with the #{chown} and #{chmod}
	methods.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static qu_null qu__file_mkdir (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	int       all = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, path, qu_string);
	QU_ARG_CHECK (path->len <= PATH_MAX);
	if (argc > 1)
		QU_ARG_vbool (1, all);

	if (!__QuFileDebug)
	{
		char buf [PATH_MAX + 1];

		if (all)
		{
			char *i = buf;

			strncpy (buf, path->str, path->len);
			buf [path->len] = '\0';
			while ((i = index (i + 1, QU_PATH_SEPCHR)) != NULL)
			{
				*i = '\0';
				if (mkdir (buf, DEFPERMDIR) != 0 && errno != EEXIST)
					QU_THROWERRNO;
				*i = QU_PATH_SEPCHR;
			}
		}

		if (mkdir (path->str, DEFPERMDIR) != 0 && errno != EEXIST)
			QU_THROWERRNO;
	}
	return QuNil;
}



/*) sub class name (path: String, keepext = true) -> String
	Get the filename part from path.

	@exc
	.EArgValue

	@xmp
		println (Filename ('foo/bar/baz'))
		println (Filename ('foo/bar/baz.txt'))

	@out
		baz
		baz.txt
*/
static qu_string qu__file_name (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	int       keep = 1;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 1)
		QU_ARG_vbool (1, keep);
	if (keep)
		return qu_file_name_ext (path);
	return qu_file_name (path);
}



/*) sub class normalize (path: String) -> String
	Get rid of weird stuffs in v{path}.

	Note that this method does not check against existing files nor
	directories. It simply gets rid of any directory entry that starts
	with a dot.

	This function might be useful in some situations such as to avoid abuse
	of paths received from the net.

	@exc
	.EArgValue

	@xmp
		println (File.normalize ('/../.foo.////bar...baz/..///'))
		println (File.normalize ( '../.foo.////bar...baz/..'))

	@out
		/bar...baz/
		/bar...baz

	@see
	#{toabs}
*/
static qu_string qu__file_normalize (qu_class self, qu_string path)
{
	long       l;
	char      *p, *i;
	qu_string  s;

	QU_ARG_TYPE (path, qu_string);
	s = qu_string_dup ((qu_string) path);
	p = s->str;
	l = path->len;
	i = path->str;

	while (l && *i == '.')
	{
		i++;
		l--;
	}
	while (l)
	{
		if (*i == QU_PATH_SEPCHR)
		{
			while (l > 1 && i[1] == QU_PATH_SEPCHR)
			{
				i++;
				l--;
			}
			if (l > 1 && i[1] == '.')
			{
				while (--l && *++i != QU_PATH_SEPCHR) {};
				continue;
			}
		}
		*p++ = *i++;
		l--;
	}
	l = p - s->str;
	if (l == 0)
		return QuStringEMPTY;
	if (l == 1)
		return qu_string_byte (s->str[0]);
	s->len = l;
	return s;
}




/*) sub class open (path: String, mode = 'r') -> File
	This is an alias of the #{new} method.
*/
#define qu__file_open qu__file_new



/*) sub class relative (src: String, dst: String) -> String
	Return the relative path from v{src} to {dst}.

	This is useful to create links.

	@exc
	.EArgValue

	@xmp
		src = 'foo/bar/baz/xxx'
		dst = 'foo/bar/yyy'
		println (File.relative (src, dst))

	@out
		../yyy
*/
static qu_string qu__file_relative (qu_class self, qu_string src, qu_string dst)
{
	char      *i, *j, *k, *m;
	qu_string  r;

	QU_ARG_TYPE (src, qu_string);
	QU_ARG_TYPE (dst, qu_string);

	src = qu_file_toabs (src, 0);
	dst = qu_file_toabs (dst, 0);

	i = src->str;
	j = dst->str;

	m = rindex (i, '/');
	assert (m);

	while (*i == *j && i <= m)
	{
		++i;
		++j;
	}

	if (*i == '/' && *j != '/')
		--i;
	else if (*i != '/' && *j == '/')
		--j;

	k = dst->str;
	while (j > k && *j != '/')
		--j;

	k = src->str;
	while (i > k && *i != '/')
		--i;

	r = qu_string_buffer_new (i - src->str);
	while (*i && i < m)
	{
		if (*i == '/')
			qu_string_buffer_put (r, "../", 3);
		++i;
	}

	qu_string_buffer_putz (r, j + 1);
	return qu_string_buffer_return (r);
}



/*) sub class rename (path: String, newpath: String)
	Rename or move a file.

	v{path} is the file to move. v{newpath} is the destination file.

	@warn
	Remember that if you supply a relative directory then it means it is
	relative to the directory returned by #{getcwd}.

	If either v{path} or v{newpath} is the root directory (checked by
	#{isroot}) then this function throws EArgValue.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static qu_null qu__file_rename (qu_class self, qu_string path, qu_string newpath)
{
	__qu_unsafe__;

	QU_ARG_CHECK (qu_string_is (path) && qu_string_is (newpath));

	if (qu_file_is_root (path) || qu_file_is_root (newpath))
		QU_THROW (QuEPERM);

	if (!__QuFileDebug)
	{
		__qu_blocking_begin;
		if (rename (path->str, newpath->str) != 0)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class repext (path: String, ext: String) -> String
	Return a new string by replacing the file extension in v{path}
	with v{ext}.

	v{ext} may or may not start with a dot. It does not matter.

	If v{path} has no file extension then v{ext} is simply appended.

	@exc
	.EArgValue

	@xmp
		p = 'foo'
		q = File.repext (p, 'bar')
		println (q)

	@out
		foo.bar
*/
static qu_string qu__file_repext (qu_class self, qu_string path, qu_string ext)
{
	QU_ARG_CHECK (qu_string_is (path) && qu_string_is (ext));
	return qu_file_repext (path, ext);
}



/*) sub class rmdir (path: String, recursive = false) -> Null
	Delete directory.

	If v{recursive} is true then all files and subdirectories of
	v{path} are deleted. Otherwise only v{path} is deleted, in which case
	it must be empty.

	@warn
	Be careful. This is a powerful yet dangerous method.

	@p
	Note that this method may fail for a number of reasons. Refer to
	C{man 2 rmdir} for details.

	@exc
	.ESecurity
	.EArgValue
	.ESystem

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static int _File_rmdir_recur (const char *path)
{
	long    lpath, lmax, lname;
	DIR    *dir;
	char    p [PATH_MAX], *np, *name;
	struct  dirent *entry;
	struct  stat t;

	lpath = strlen (path);
	lmax  = PATH_MAX - lpath - 1;
	QU_REQUIRE (QuENAMETOOLONG, lmax > 4);
	strncpy (p, path, lpath);
	p [lpath] = '\0';
	dir = opendir (p);
	if (dir == NULL)
		QU_THROWERRNO;
	np = p + lpath;
	*np++ = QU_PATH_SEPCHR;
	while ((entry = readdir (dir)) != NULL)
	{
		name = entry->d_name;
		if (name[0] == '.' && (name[1] == '.' || name[1] == '\0'))
			continue;
		lname = strlen (name);
		if (lname >= lmax)
			QU_THROW (QuENAMETOOLONG);
		strcpy (np, name);
		if (_file_stat_path (p, &t) != 0)
			QU_THROWERRNO;
		if (S_ISDIR (t.st_mode))
		{
			if (!_File_rmdir_recur (p))
				return 0;
		}
		else
		{
			if (unlink (p) != 0)
				QU_THROWERRNO;
		}
	}
	closedir (dir);
	if (rmdir (path) != 0)
		QU_THROWERRNO;
	return 1;
}



static qu_null qu__file_rmdir (qu_class self, int argc, qu_any *argv)
{
	qu_string path;
	int       recur = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, path, qu_string);
	if (argc > 0)
		QU_ARG_vbool (1, recur);
	QU_REQUIRE (QuEPERM, !qu_file_is_root (path));
	if (!__QuFileDebug)
	{
		__qu_blocking_begin;
		if (recur)
		{
			if (!_File_rmdir_recur (path->str))
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		else
		{
			if (rmdir (path->str) != 0)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class split (path: String) -> Array
	Get directory, name and extension parts from path. The returned
	value is an -{Array} of 3 elements: (directory, name, extension).

	@exc
	.EArgValue

	@xmp
		(dir, name, ext) = File.split ('/foo/bar/baz.tar.gz')
		println (dir)
		println (name)
		println (ext)

	@out
		/foo/bar
		baz
		.tar.gz
*/
static qu_array qu__file_split (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_array_new_ex (3, qu_file_dir (path), qu_file_name (path), qu_file_ext (path));
}



/*) sub class strip (path: String) -> String
	Return a new string by stripping the file extension from path.

	@exc
	.EArgValue

	@xmp
		println (File.strip ('foo/bar.tar.gz'))

	@out
		foo/bar
*/
static qu_string qu__file_strip (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_file_strip (path);
}



/*) sub class stripdir (path: String) -> String
	Return a new string by stripping the directory part from path.

	@exc
	.EArgValue

	@xmp
		println (File.stripdir ('foo/bar.tar.gz'))

	@out
		bar.tar.gz
*/
static qu_string qu__file_stripdir (qu_class self, qu_string path)
{
	long e;

	QU_ARG_TYPE (path, qu_string);
	e = qu_string_rindex (path, QuFilePathSepStr, 0);
	if (e < 0)
		return path;
	return qu_string_newz (path->str + e + 1);
}



/*) sub class symlink (src: String, dst: String) -> Null
	Make a symbolic link.

	v{dst} is the path of the link to create.

	v[src} is the target where v{dst} points to.

	@exc
	.EArgValue
	.ESystem
	.ESecurity

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that an exception may still be
	thrown if you pass the wrong arguments.
*/
static qu_null qu__file_symlink (qu_class self, qu_string src, qu_string dst)
{
	__qu_unsafe__;

	QU_ARG_CHECK (qu_string_is (src) && qu_string_is (dst));
	if (!__QuFileDebug)
	{
		__qu_blocking_begin;
		if (symlink (src->str, dst->str) != 0)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub class toabs (path: String) -> String
	Return the absolute path of v{path}.

	This method removes multiple C{'/'} and resolves all C{'/./'} and C{'/../'}
	components. If the C{'/../'} causes the path to pass the root directory then
	EArgValue is thrown.

	This method does not check for a valid existing path. It simply transforms
	v{path} to an absolute v{path}.

	Note that v{path} may start with C{'/'}.

	@exc
	.EArgValue
	.ESystem

	@xmp
		# assuming the current directory is '/opt/qu/demo'
		println (File.toabs (''))
		println (File.toabs ('/'))
		println (File.toabs ('foo'))
		println (File.toabs ('foo.bar'))
		println (File.toabs ('./foo.bar'))
		println (File.toabs ('.././..///foo.bar'))
		println (File.toabs ('/././//./foo.bar'))
		println (File.toabs ('/xxx/.././foo.bar'))

	@out
		/opt/qu/demo/
		/
		/opt/qu/demo/foo
		/opt/qu/demo/foo.bar
		/opt/qu/demo/foo.bar
		/opt/foo.bar
		/foo.bar
		/foo.bar
*/
static qu_string qu__file_toabs (qu_class self, qu_string path)
{
	QU_ARG_TYPE (path, qu_string);
	return qu_file_toabs (path, 0);
}



/*) sub class valid (path: Any, allowspace = false) -> Boolean
	Return true if path is a valid file path, false otherwise.

	If v{path} is not a String or an empty one then false is returned.

	This method does not check for an existing path. It simply checks
	for invalid characters in v{path}.

	This method considers a space (ASCII 32) an invalid character unless
	v{allowspace} is true.

	See -{String.ispath} for valid characters.

	The system may consider other characters to be valid but using these
	characters might save you a lot of pain.

	The File methods do not call #{valid} to validate paths so you can pass
	anything acceptable to the system.

	@noexc
*/
static qu_boolean qu__file_valid (qu_class self, int argc, qu_any *argv)
{
	return (argc < 1)
		? QuFalse
		: (qu_file_valid (argv[0], (argc > 1) ? (argv[1] == QuTrue) : 0) ? QuTrue : QuFalse);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/
/* none */



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/
/* all inherited from qu_stream */


/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub chmod (perm: String, timeout = 30) -> self
	Set file permission of this file.

	You can optionally set a timeout argument in seconds.

	v{perm} must be a combination of the following characters:
	@ops
	.| C{u} | user
	.| C{g} | group
	.| C{o} | others

	@p
	each followed by a combination of the following characters:
	@ops
	.| C{r} | has read permission
	.| C{w} | has write permission
	.| C{x} | has execute permission

	@p
	You can separate any characters with any number of spaces and uppercase
	letters are interpreted the same as lowercase letters.

	@exc
	.EArgValue
	.ESystem

	@xmp
		File.chmod ('/foo/bar.txt', 'UrwxGwrOr')
		File.chmod ('/foo/bar.txt', 'urwx gwr or')

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed.
*/
static qu_file qu__file_chmod (qu_file self, int argc, qu_any *argv)
{
	long      to = 30;
	qu_string perm;
	int       p;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, perm, qu_string);
	QU_ARG_CHECK (qu_file_get_permission_mode (perm->str, &p));
	if (argc > 1)
		QU_ARG_vlong (1, to);

	if (!__QuFileDebug)
	{
		to += NOW;

		__qu_blocking_begin;
		while (fchmod (self->fd, p) < 0)
		{
			if (errno != EINTR)
			{
				FIXEAGAIN;
				if (errno != EAGAIN)
				{
					__qu_blocking_end;
					QU_THROWERRNO;
				}
			}
			if (NOW > to)
			{
				__qu_blocking_end;
				QU_THROW (QuETimeout);
			}
			__qu_blocking_check_signal;
		}
		__qu_blocking_end;
	}
	return self;
}



/*) sub chown (user, *arg) -> self
	@def
	sub chown (user: String, timeout = 30) -> self
	sub chown (user: String, group: String, timeout = 30) -> self

	@p
	Set file owner of this file.

	v{user} and v{group} must be valid user and group name. If called with
	two arguments then v{group} is the user group as returned by the
	system function f{getpwnam}.

	You can optionally set at timeout in seconds.

	@exc
	.EArgValue
	.ESystem

	@note
	If #{debug} is true then the operation is not actually performed
	and success is assumed. Note however that the arguments are still validated
	including v{user} and v{group}.
*/
static qu_file qu__file_chown (qu_file self, int argc, qu_any *argv)
{
	qu_string user, grp = QuStringEMPTY;
	long      to = 30;
	struct passwd *u;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, user, qu_string);
	if (argc > 1)
	{
		if (qu_string_is (argv[1]))
		{
			grp = argv[1];
			if (argc > 2)
				QU_ARG_vlong (2, to);
		}
		else
		{
			QU_ARG_vlong (1, to);
		}
	}

	__qu_blocking_begin;
	u = getpwnam (user->str);
	__qu_blocking_end;

	if (u == NULL)
		QU_THROWERRNO;
	else
	{
		uid_t uid = u->pw_uid;
		gid_t gid;

		if (grp->len == 0)
		{
			gid = u->pw_gid;
		}
		else
		{
			struct group *g;

			__qu_blocking_begin;
			g = getgrnam (grp->str);
			__qu_blocking_end;
			if (g == NULL)
				QU_THROWERRNO;
			gid = g->gr_gid;
		}

		if (!__QuFileDebug)
		{
			to += NOW;

			__qu_blocking_begin;
			while (fchown (self->fd, uid, gid) < 0)
			{
				if (errno != EINTR)
				{
					FIXEAGAIN;
					if (errno != EAGAIN)
					{
						__qu_blocking_end;
						QU_THROWERRNO;
					}
				}
				if (NOW > to)
				{
					__qu_blocking_end;
					QU_THROW (QuETimeout);
				}
				__qu_blocking_check_signal;
			}
			__qu_blocking_end;
		}
	}
	return self;
}



/*-----------------------------------------------------------------------
	interface
-----------------------------------------------------------------------*/

qu_string qu_file_dir (qu_string path)
{
	long p = qu_string_rindex (path, QuFilePathSepStr, 0);

	if (p < 0)
		return QuStringEMPTY;
	else if (p == 0)
		return QuFilePathSepStr;
	else
		return qu_string_getslice (path, 0, p);
}



qu_string qu_file_ext (qu_string path)
{
	long i;

	i = qu_string_rindex (path, QuFilePathSepStr, 0);
	i = qu_string_index (path, Sdot, (i < 0 ? 0 : i));
	if (i <= 0)
		return QuStringEMPTY;
	else
		return qu_string_getslice (path, i, -1);
}



int qu_file_exist (qu_string path, qu_string type)
{
	int r;

	__qu_blocking_begin;
	r = access (path->str, F_OK);
	__qu_blocking_end;
	if (r != 0)
		return 0;
	r = qu_file_is_type (path, type);
	return r && (errno == 0);
}



int qu_file_is_abs (qu_string path)
{
	return (path->len && path->str[0] == QU_PATH_SEPCHR);
}



int qu_file_is_dir (qu_string path)
{
	struct stat fs;

	errno = 0;
	if (_file_stat_path (path->str, &fs) != 0)
		return 0;
	return S_ISDIR (fs.st_mode);
}



static int _file_path_depth (const char *s)
{
	int n = 0;

	while (*s == QU_PATH_SEPCHR)
		s++;
	while (*s)
	{
		if (*s == QU_PATH_SEPCHR)
		{
			while (s[1] == QU_PATH_SEPCHR)
				s++;
			if (s[1] == '.')
			{
				if (s[2] == '.' && (s[3] == QU_PATH_SEPCHR || s[3] == '\0'))
				{
					s += 2;
					n--;
					continue;
				}
				if (s[2] == QU_PATH_SEPCHR || s[2] == '\0')
				{
					s++;
					continue;
				}
			}
			n++;
		}
		s++;
	}
	return n;
}



int qu_file_is_root (qu_string path)
{
	char p [PATH_MAX];

	if (path->len && path->str[0] == QU_PATH_SEPCHR)
	{
		if (path->len < PATH_MAX)
		{
			memcpy (p, path->str, path->len);
			p [path->len] = '\0';
			if (_file_path_depth (p) == -1)
				return 1;
		}
	}
	else
	{
		if (CurDirLen + path->len < PATH_MAX)
		{
			strcpy (p, CurDir);
			p [CurDirLen] = QU_PATH_SEPCHR;
			memcpy (p + CurDirLen + 1, path->str, path->len);
			p[CurDirLen + 1 + path->len] = '\0';
			if (_file_path_depth (p) == -1)
				return 1;
		}
	}
	return 0;
}



int qu_file_is_type (qu_string path, qu_string filtertype)
{
	struct stat fs;
	int m, r;

	QU_ARG_CHECK (qu_file_get_filter_mode (filtertype->str, &m));
	__qu_blocking_begin;
	r = _file_stat_path (path->str, &fs);
	__qu_blocking_end;
	if (r != 0)
		QU_THROWERRNO;
	errno = 0;
	return (fs.st_mode & m) != 0;
}



qu_set qu_file_list (qu_string spath, qu_string filtertype, qu_string filtername, int strip)
{
	volatile qu_string path;
	int     filter = 0;
	long    lpath, lmax, lname;
	DIR    *dir;
	char    p [PATH_MAX], *np, *name;
	struct  dirent *entry;
	struct  stat t;
	volatile qu_set a;

	path = qu_file_toabs (spath, 0);
	QU_ENSURE (path);
	lpath = strlen (path->str);
	QU_ARG_CHECK (lpath > 0);

	if (filtertype)
	{
		if (filtertype->len == 0)
			filtertype = 0;
		else
			QU_ARG_CHECK (qu_file_get_filter_mode (filtertype->str, &filter));
	}
	if (filtername)
	{
		if (filtername->len == 0)
			filtername = 0;
	}
	lmax  = PATH_MAX - lpath - 1;
	QU_REQUIRE (QuENAMETOOLONG, lmax > 4);
	strncpy (p, path->str, lpath);
	p [lpath] = '\0';

	__qu_blocking_begin;
	dir = opendir (p);
	__qu_blocking_end;

	if (dir == NULL)
		QU_THROWERRNO;

	a = qu_set_new (8);
	np = p + lpath;
	*np++ = QU_PATH_SEPCHR;

	__qu_blocking_begin;
	while ((entry = readdir (dir)) != NULL)
	{
		name = entry->d_name;
		if (name[0] == '.')
		{
			if (name[1] == '.'
				|| name[1] == '\0'
				|| !(filter & S_IFHIDDEN))
				continue;
		}
		lname = strlen (name);
		if (lname >= lmax)
			continue;
		if (filtername && !qu_file_match (filtername->str, filtername->len, name, lname))
			continue;
		if (filtertype)
		{
			strcpy (np, name);
			if (_file_stat_path (p, &t) != 0)
				continue;
			if (S_ISREG (t.st_mode) && !(filter & S_IFREG))
				continue;
			if (S_ISDIR (t.st_mode) && !(filter & S_IFDIR))
				continue;
			if (S_ISLNK (t.st_mode) && !(filter & S_IFLNK))
				continue;
		}
		__qu_blocking_end;
		{
			qu_string entry = qu_string_newz (name);
			if (strip)
				entry = qu_file_strip (entry);
			qu_set_push (a, entry);
		}
		__qu_blocking_begin;
	}
	closedir (dir);
	__qu_blocking_end;
	return a;
}



int qu_file_info (qu_string path, long *size, double *mtime)
{
	struct stat fs;

	__qu_blocking_begin;
	if (_file_stat_path (path->str, &fs) != 0)
	{
		__qu_blocking_end;
		QU_THROWERRNOF ("cant stat %s", path->str);
	}

	assert (fs.st_size <= LONG_MAX);
	if (size)
		*size = (long) fs.st_size;
	if (mtime)
		*mtime = (double) fs.st_mtime;
	return 1;
}



/* we want to do this as quick as possible */
qu_string qu_file_load_ex (qu_string path, long timeout, long max, long *size, double *mtime)
{
	char *p = path->str;
	long  len;
	int   fd;
	volatile qu_string s;
	struct stat fs;

	__qu_blocking_begin;
	for (;;)
	{
		fd = open (p, O_RDONLY);
		if (fd >= 0)
			break;

		switch (errno)
		{
			case ETXTBSY:
			case EMFILE:
			case ENFILE:
			case EAGAIN:
				if (NOW > timeout)
				{
					__qu_blocking_end;
					goto errtimeout;
				}

			case EINTR:
				continue;

			default:
				__qu_blocking_end;
				goto erropen;
		}
	}

	while (fstat (fd, &fs) != 0)
	{
		if (errno != EINTR)
		{
			CLOSEFD (fd);
			__qu_blocking_end;
		 	goto errstat;
		}
	}

	if (max <= 0 && fs.st_size >= QU_MEM_MAX)
	{
		CLOSEFD (fd);
		__qu_blocking_end;
		goto errtoobig;
	}

	len = fs.st_size;

	if (size)
		*size = len;

	if (mtime)
		*mtime = (double) fs.st_mtime;

	if (max > 0 && len > max)
		len = max;

	__qu_blocking_end;

	s = qu_string_newl (len);

	__qu_blocking_begin;
	for (;;)
	{
		ssize_t r = read (fd, s->str, (size_t) len);
		if (r < 0)
		{
			switch (errno)
			{
				case EAGAIN:
					if (NOW > timeout)
					{
						CLOSEFD (fd);
						__qu_blocking_end;
						goto errtimeout;
					}

				case EINTR:
					continue;
			}
			CLOSEFD (fd);
			__qu_blocking_end;
			goto errio;
		}

		if (r != (ssize_t) len)
		{
			CLOSEFD (fd);
			__qu_blocking_end;
			errno = EIO;
			goto errio;
		}
		break;
	}
	CLOSEFD (fd);
	__qu_blocking_end;

	return s;

	errtimeout: QU_THROWF (QuETimeout, "timeout while trying to load %s", p);
	errtoobig:  QU_THROWF (QuEMemory, "file %s is too big to load into memory", p);
	errio:      QU_THROWERRNOF ("I/O error while loading %s", p);
	erropen:    QU_THROWERRNOF ("cant open %s", p);
	errstat:    QU_THROWERRNOF ("cant stat %s", p);
}



qu_string qu_file_load (qu_string path)
{
	return qu_file_load_ex (path, 180, -1, NULL, NULL);
}



/* for qu_sub.c __interface */
int qu_file_modified (qu_string path, double lasttime)
{
	struct stat fs;

	if (_file_stat_path (path->str, &fs))
		return 1;
	return fs.st_mtime > lasttime;
}



qu_string qu_file_make (qu_string path, qu_string defdir, qu_string defext)
{
	int e = strrchr (path->str, '.') != NULL;

	if (defdir == 0)
		defdir = QuStringEMPTY;

	if (defext == 0)
		defext = QuStringEMPTY;

	if (qu_file_is_abs (path) || defdir->len == 0)
	{
		if (e || defext->len == 0)
			return path;

		if (defext->str[0] != '.')
			return qu_string_cat (path, qu_string_byte ('.'), defext, NULL);

		return qu_string_cat (path, defext, NULL);
	}
	if (e || defext->len == 0)
	{
		if (defdir->str [defdir->len - 1] == QU_PATH_SEPCHR)
			return qu_string_cat (defdir, path, NULL);

		return qu_string_cat (defdir, QuFilePathSepStr, path, NULL);
	}
	if (defdir->str [defdir->len - 1] == QU_PATH_SEPCHR)
	{
		if (defext->str[0] != '.')
			return qu_string_cat (defdir, path, qu_string_byte ('.'), defext, NULL);

		return qu_string_cat (defdir, path, defext, NULL);
	}
	if (defext->str[0] != '.')
		return qu_string_cat (defdir, QuFilePathSepStr, path, qu_string_byte ('.'), defext, NULL);

	return qu_string_cat (defdir, QuFilePathSepStr, path, defext, NULL);
}



qu_string qu_file_name (qu_string path)
{
	if (path->str [path->len - 1] == QU_PATH_SEPCHR)
	{
		return QuStringEMPTY;
	}
	else
	{
		long sep = qu_string_rindex (path, QuFilePathSepStr, 0);

		if (sep < 0)
			return qu_file_strip (path);
		else
		{
			long dot = qu_string_index (path, Sdot, sep);
			if (dot > sep)
				return qu_string_getslice (path, sep + 1, dot - sep - 1);
			return qu_string_getslice (path, sep + 1, -1);
		}
	}
}



qu_string qu_file_name_ext (qu_string path)
{
	if (path->str [path->len - 1] == QU_PATH_SEPCHR)
	{
		return QuStringEMPTY;
	}
	else
	{
		long sep = qu_string_rindex (path, QuFilePathSepStr, 0);

		if (sep < 0)
			return qu_file_strip (path);
		else
			return qu_string_getslice (path, sep + 1, -1);
	}
}



qu_string qu_file_repext (qu_string path, qu_string ext)
{
	int  k = (ext->str[0] == '.') ? 1 : 0;
	long e = qu_string_rindex (path, QuFilePathSepStr, 0);

	e = qu_string_index (path, Sdot, e < 0 ? 0 : e);
	if (e < 0)
	{
		if (k)
			return qu_string_cat ((qu_string) path, ext, 0);
		else
			return qu_string_cat ((qu_string) path, Sdot, ext, 0);
	}
	else
	{
		if (e >= QU_INT_MAX - ext->len - 1)
		{
			QU_THROW (QuENAMETOOLONG);
		}
		else
		{
			qu_string s = qu_string_newl (++e + ext->len - k);

			memcpy (s->str, path->str, e);
			memcpy (s->str + e, ext->str + k, ext->len - k);
			assert (s->str [s->len] == '\0');
			return s;
		}
	}
}



qu_string qu_file_strip (qu_string path)
{
	long e = qu_string_rindex (path, QuFilePathSepStr, 0);
	e = qu_string_index (path, Sdot, e < 0 ? 0 : e);
	if (e < 0)
		return path;
	return qu_string_new (path->str, e);
}



qu_string qu_file_toabs (qu_string path, qu_string defdir)
{
	long lp = path->len;
	char d [PATH_MAX + 1], *s, *p, *q;

	if (qu_file_is_abs (path))
		return path;

	if (path->str[0] == QU_PATH_SEPCHR)
	{
		QU_REQUIRE (QuENAMETOOLONG, lp < sizeof (d) - 1);
		memcpy (d, path->str, lp + 1);
	}
	else
	{
		long  ls = QuFilePathSepStr->len;
		long  ld;
		char *sd;

		if (defdir)
		{
			ld = defdir->len;
			sd = defdir->str;
		}
		else
		{
			ld = CurDirLen;
			sd = CurDir;
		}

		QU_REQUIRE (QuENAMETOOLONG, lp < sizeof (d) - ls - ld);
		memcpy (d, sd, ld);
		memcpy (d + ld, QuFilePathSepStr->str, ls);
		memcpy (d + ld + ls, path->str, lp + 1);
	}

	/* remove multiple slashes /// -> / */
	p = q = d;
	while (*q)
	{
		while (*q == QU_PATH_SEPCHR && q[1] == QU_PATH_SEPCHR)
			q++;
		*p++ = *q++;
	}
	*p = '\0';

	/* special case :-( */
	if (d[0] == QU_PATH_SEPCHR && d[1] == '\0')
		return QuFilePathSepStr;

	/* mark the filename */
	s = d + strlen (d) - 1;
	while (s > d && *s != QU_PATH_SEPCHR)
		s--;
	if (s == d)
		return qu_string_newf ("%s" QU_PATH_SEPSTR "%s", CurDir, s);

	*s++ = '\0';

	/* translate /./ and /../ */
	p = q = d;
	while (*q)
	{
		if (q[0] == QU_PATH_SEPCHR && q[1] == '.' && (q[2] == QU_PATH_SEPCHR || q[2] == '\0'))
		{
			q += 2;
			continue;
		}
		if (q[0] == QU_PATH_SEPCHR && q[1] == '.' && q[2] == '.' && (q[3] == QU_PATH_SEPCHR || q[3] == '\0'))
		{
			if (*p == QU_PATH_SEPCHR)
				p--;
				while (p > d && *p != QU_PATH_SEPCHR)
					p--;
			QU_ARG_CHECK (p >= d);
			q += 3;
			continue;
		}
		*p++ = *q++;
	}

	/* build it */
	if (*p != QU_PATH_SEPCHR)
		*p++ = QU_PATH_SEPCHR;
	else
		p++;

	while (*s)
		*p++ = *s++;
	*p = '\0';

	return qu_string_newz (d);
}



int qu_file_valid (qu_any path, int allowspace)
{
	if (qu_string_is (path))
	{
		long l = ((qu_string) path)->len;

		if (0 < l && l <= PATH_MAX)
		{
			char *s = ((qu_string) path)->str;
			int   k = 0;

			while (l--)
			{
				if (*s == ' ')
				{
					if (!allowspace)
						return 0;
				}
				else if (qu_ascii_ispath (*s))
				{
					k = 1;
				}
				s++;
			}
			return k;
		}
	}
	return 0;
}



/*-----------------------------------------------------------------------
	init
-----------------------------------------------------------------------*/

void qu_file__setup ()
{
	qu_class_setup_final (QuFileClass, qu_file_s, qu__stream__destroy, qu__stream__mark);
	QuFilePathSepStr = qu_string_new_staticz (QU_PATH_SEPSTR);
	Sdot = qu_string_byte ('.');
}


void qu_file__init ()
{
	__QuFileDebug = QU_APP_DEBUG;

	_file_set_curdir ();

	/* class property */
	qu_class_constz  (QuFileClass, "PATHSEP", qu_string_new_staticz (QU_PATH_SEPSTR));

	/* standard class methods */
	qu_class_method_class  (QuFileClass, QU_NAME_new, qu__file_new , -1);
	qu_class_method_classz (QuFileClass, "temp"     , qu__file_temp,  0);

	/* specific class methods */
	qu_class_method_classz (QuFileClass, "agecmp"   , qu__file_agecmp   , -1);
	qu_class_method_classz (QuFileClass, "chdir"    , qu__file_chdir    ,  1);
	qu_class_method_classz (QuFileClass, "chroot"   , qu__file_chroot   ,  1);
	qu_class_method_classz (QuFileClass, "compare"  , qu__file_compare  , -1);
	qu_class_method_classz (QuFileClass, "contains" , qu__file_contains , -1);
	qu_class_method_classz (QuFileClass, "copy"     , qu__file_copy     , -1);
	qu_class_method_classz (QuFileClass, "debug"    , qu__file_debug    , -1);
	qu_class_method_classz (QuFileClass, "delete"   , qu__file_delete   ,  1);
	qu_class_method_classz (QuFileClass, "dir"      , qu__file_dir      ,  1);
	qu_class_method_classz (QuFileClass, "dump"     , qu__file_dump     , -1);
	qu_class_method_classz (QuFileClass, "eachfile" , qu__file_eachfile ,  1);
	qu_class_method_classz (QuFileClass, "exist"    , qu__file_exist    , -1);
	qu_class_method_classz (QuFileClass, "ext"      , qu__file_ext      ,  1);
	qu_class_method_classz (QuFileClass, "fchmod"   , qu__file_fchmod   ,  2);
	qu_class_method_classz (QuFileClass, "fchown"   , qu__file_fchown   , -1);
	qu_class_method_classz (QuFileClass, "fixdir"   , qu__file_fixdir   ,  1);
	qu_class_method_classz (QuFileClass, "ftruncate", qu__file_ftruncate, -1);
	qu_class_method_classz (QuFileClass, "getcwd"   , qu__file_getcwd   ,  0);
	qu_class_method_classz (QuFileClass, "isabs"    , qu__file_isabs    ,  1);
	qu_class_method_classz (QuFileClass, "isdir"    , qu__file_isdir    , -1);
	qu_class_method_classz (QuFileClass, "isroot"   , qu__file_isroot   ,  1);
	qu_class_method_classz (QuFileClass, "istype"   , qu__file_istype   , -1);
	qu_class_method_classz (QuFileClass, "list"     , qu__file_list     , -1);
	qu_class_method_classz (QuFileClass, "load"     , qu__file_load     , -1);
	qu_class_method_classz (QuFileClass, "make"     , qu__file_make     , -1);
	qu_class_method_classz (QuFileClass, "match"    , qu__file_match    ,  2);
	qu_class_method_classz (QuFileClass, "mkdir"    , qu__file_mkdir    , -1);
	qu_class_method_classz (QuFileClass, "name"     , qu__file_name     , -1);
	qu_class_method_classz (QuFileClass, "normalize", qu__file_normalize,  1);
	qu_class_method_classz (QuFileClass, "open"     , qu__file_open     , -1);
	qu_class_method_classz (QuFileClass, "relative" , qu__file_relative ,  2);
	qu_class_method_classz (QuFileClass, "rename"   , qu__file_rename   ,  2);
	qu_class_method_classz (QuFileClass, "repext"   , qu__file_repext   ,  2);
	qu_class_method_classz (QuFileClass, "rmdir"    , qu__file_rmdir    , -1);
	qu_class_method_classz (QuFileClass, "split"    , qu__file_split    ,  1);
	qu_class_method_classz (QuFileClass, "strip"    , qu__file_strip    ,  1);
	qu_class_method_classz (QuFileClass, "stripdir" , qu__file_stripdir ,  1);
	qu_class_method_classz (QuFileClass, "symlink"  , qu__file_symlink  ,  2);
	qu_class_method_classz (QuFileClass, "toabs"    , qu__file_toabs    ,  1);
	qu_class_method_classz (QuFileClass, "valid"    , qu__file_valid    , -1);

	/* standard instance methods */
	/* all inherited from qu_stream */

	/* specific instance methods */
	qu_class_method_instz  (QuFileClass, "chmod", qu__file_chmod, -1);
	qu_class_method_instz  (QuFileClass, "chown", qu__file_chown, -1);

	/* finalize declaration */
	qu_class_set_super (QuFileClass, QuStreamClass);
	qu_class_enddecl   (QuFileClass);
}
