/* Float.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Float is Decimal __final
	A Float object is a floating point number such as C{-1.234} in
	the range of Float.#{MIN} .. Float.#{MAX}, contrary to -{Big} which
	can be as big as the available memory.

	Refer to the -{Number} class for information on Qu numerics in general.

	The class hierarchy:
	@tree
	.-{Number}
	.	-{Decimal}
	.		Float

	@p
	Generally you should make no distinction between a Float and a Big in
	your applications. The only reason why Float exists is performance and
	memory saving. Once computers are very fast and memory is very cheap
	then Float and Big will be merged into -{Decimal}.

	@xmp
		println (Float.MIN)
		println (Float.MAX)

	@out
		2.22507e-308
		1.79769e+308

	@p
	Unlike in C, Float objects are never infinity nor NaN. In Qu NaN is an
	error and you get an EMath exception as soon as the outcome of an operation
	leads to one, e.g. C{0 / 0}. Infinities are defined by two system objects
	infp and infn and are -{Inf} objects, not Float objects.
*/
qu_class QuFloatClass;


/* predefined */
qu_float QuFloat0;
qu_float QuFloat1;
qu_float QuFloat2;
qu_float QuFloat_1;
qu_float QuFloat_2;


/*) const MIN = 2.2250e-308
	This the minimum possible value of a Float.

	@warn
	The actual value is set by the system.
*/
qu_float QU_FLOAT_MIN;



/*) const MAX = 1.7976e+308
	This the maximum possible value of a Float.

	@warn
	The actual value is set by the system.
*/
qu_float QU_FLOAT_MAX;



/*) const PACKSIZE = 8
	The length of the string returned by the #{pack} method.

	@warn
	The actual value is set by the system.
*/



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (x) -> Float
	Create a new Float object.

	If v{x} is a Float then v{x} is returned, otherwise this method calls
	C{x.__as (Float)} ensuring that the returned value is a Float.

	@exc
	.EArgValue
	.EImplement
	.ESyntax
	.Exception  anything thrown by the __as method

	@xmp
		i = Float ('1.23')
		println (i)

		class Foo
			sub __as (what)
				switch what
					case Float: return 4.56
				;;
				throw EImplement
			;;
		;;

		x = Foo ()
		println (Float (x))

	@out
		1.23
		4.56
*/
static qu_float qu__float_new (qu_class self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_int_as_float      ((qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_long_as_float     ((qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_rational_as_float ((qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_float) x;
		case QU_CLASS_ID_Big:      return qu_big_as_float      ((qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_as_float  ((qu_complex ) x);

		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Buffer:
			return qu_float_new_string (x);

		case QU_CLASS_ID_Boolean:
			return qu_float_new_boolean (x);

		case QU_CLASS_ID_Inf:
			QU_ARG_THROWZ ("Float (Inf)");
	}
	return qu_object_as (x, QuFloatClass);
}



qu_real qu_float_newdx (double v)
{
	if (qu_double_isfinite (v))
		return (qu_real) qu_float_newd (v);
	QU_ARG_CHECKZ (!qu_double_isnan (v), "Float (double NaN)");
	return (qu_real) __qu_inf__ ((v < 0) ? -1 : 1);
}



qu_float qu_float_newd (double v)
{
	assert (qu_double_isfinite (v));
	{
		qu_float f = qu_gc_new (QuFloatClass);
		f->value = v;
		return f;
	}
}



#ifdef HAVE_LONG_DOUBLE
qu_float qu_float_newld (long double v)
{
	QU_ARG_CHECKZ ((long double) -DBL_MAX <= v && v <= (long double) DBL_MAX, "Float (long double) out of range");
	return qu_float_newd ((double) v);
}

qu_real qu_float_newldx (long double v)
{
	if (qu_longdouble_isfinite (v))
		return (qu_real) qu_float_newld (v);
	QU_ARG_CHECKZ (!qu_longdouble_isnan (v), "Float (long double NaN)");
	return (qu_real) __qu_inf__ ((v < 0) ? -1 : 1);
}
#endif



qu_float qu_float_new_long (qu_long x)
{
	double v = qu_long_tod (x);
	QU_ARG_CHECKZ (errno == 0, "Float (Long) out of range");
	return qu_float_newd (v);
}



qu_float qu_float_new_rational (qu_rational x)
{
	double v = qu_rational_tod (x);
	QU_ARG_CHECKZ (errno == 0, "Float (Rational) out of range");
	return qu_float_newd (v);
}



qu_float qu_float_new_integer (qu_integer x)
{
	return qu_int_is (x) ? qu_float_new_int ((qu_int) x) : qu_float_new_long ((qu_long) x);
}



qu_float qu_float_new_decimal (qu_decimal x)
{
	return qu_float_is (x) ? (qu_float) x : qu_float_new_big ((qu_big) x);
}



qu_float qu_float_new_real (qu_real x)
{
	QU_ARG_CHECKZ (!qu_inf_is (x), "Float (Inf)");
	return qu_integer_is (x) ? qu_float_new_integer ((qu_integer) x) : qu_float_new_decimal ((qu_decimal) x);
}



qu_float qu_float_new_number (qu_number x)
{
	return qu_complex_is (x) ? qu_float_new_complex ((qu_complex) x) : qu_float_new_real ((qu_real) x);
}



qu_float qu_float_new_string (qu_string s)
{
	char *z = s->str;
	qu_float x = qu_float_scan (&z, 0);
	QU_REQUIREZ (QuESyntax, x, "Float (String) invalid syntax");
	return x;
}



qu_float qu_float_new_static (double v)
{
	assert (qu_double_isfinite (v));
	{
		qu_float f  = qu_gc_new_static (QuFloatClass);
		f->value = v;
		return f;
	}
}



qu_float qu_float_to_static (qu_float self)
{
	qu_float f = qu_gc_new_static (QuFloatClass);
	f->value = self->value;
	return f;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	scan
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

double qu_float_scand (char **s, int neg)
{
	char  *c = *s;
	double v;

	if (!neg)
	{
		while (qu_ascii_isspace (*c))
			c++;
		if (*c == '\0')
			goto syntax;

		if (*c == '-')
		{
			c++;
			neg = 1;
		}
		else if (*c == '+')
		{
			c++;
		}
		while (qu_ascii_isspace (*c))
			c++;
	}

	if (*c == '\0')
		goto syntax;

	errno = 0;
	v = strtod (c, &c);
	if (errno == 0)
		*s = c;
	return v;

  syntax:
	errno = EDOM;
	return 0.0;
}



qu_float qu_float_scan (char **s, int neg)
{
	double v;

	v = qu_float_scand (s, neg);
	QU_REQUIRE (QuESyntax, errno == 0);
	return qu_float_newd (v);
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class unpack (s: String, ofs = 0) -> Int
	Unpack a Float from a String optionally start at a specified v{offset} in
	the String.

	@exc
	.EIndex
	.EArgValue
	.EImplement  your Int native size does not support packing

	@xmp
		i = 1.23
		s = i.pack
		println (s->len)
		i = Float.unpack (s)
		println (i)

	@out
		8
		1.23
*/
static qu_float qu__float_unpack (qu_class self, int argc, qu_any *argv)
{
#if SIZEOF_DOUBLE == 8
	qu_string  s;
	long       ofs = 0;
	double    *p;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, s, qu_string);

	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		QU_ENSURE_INDEX (ofs, s->len);
	}

	p = (double *) (s->str + ofs);

	QU_ARG_CHECK ((s->len - ofs) >= 8);
	return qu_float_newd (*p);
#else
	QU_THROW (QuEImplement);
#endif
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/
/* none */



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_float_sign (qu_float x)
{
	return (x->value > 0) ? 1 : ((x->value < 0) ? -1 : 0);
}



int qu_float_isone (qu_float x)
{
	return x->value == 1 || x->value == -1;
}



int qu_float_isfrac (qu_float x)
{
	double v = x->value;
	if (v < 0)
		v = -v;
	return 0 < v && v < 1;
}



double __QuFloatIsWholeHelper;


int qu_float_isodd (qu_float x)
{
	int    e;
	double f;

	f = frexp (x->value, &e);
	return (e < 1) ? 0 : (fmod (x->value, 2) != 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following classes:
	@ops
	.| Float
		-{Decimal}
		-{Real}
		-{Number}  | Return self.
	.| -{Int}      | Convert to Int if possible, throw EValue otherwise.
	.| -{Long}     | Convert to Long.
	.| -{Rational} | Convert to Rational.
	.| -{Integer}  | Convert to Int if possible, Long otherwise.
	.| -{Big}      | Convert to Big.
	.| -{Complex}  | Return C{(self + 0j)}.
	.| -{Boolean}  | Delegate to #{__true}.
	.| -{String}   | Delegate to #{__string}.

	@exc
	.EArgValue
	.EValue

	@xmp
		x = 1.2
		println (x as Int)

	@out
		1
*/
static qu_any qu__float__as (qu_float self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:      return qu_float_as_int      (self);
		case QU_CLASS_ID_Long:     return qu_float_as_long     (self);
		case QU_CLASS_ID_Rational: return qu_float_as_rational (self);
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Decimal:
		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Number:   return self;
		case QU_CLASS_ID_Big:      return qu_float_as_big      (self);
		case QU_CLASS_ID_Complex:  return qu_float_as_complex  (self);
		case QU_CLASS_ID_Integer:  return qu_float_as_integer  (self);
		case QU_CLASS_ID_Boolean:  return qu_float_as_boolean  (self);
		case QU_CLASS_ID_String:   return qu_float_as_string   (self);
	}
	QU_ARG_THROW;
}



long qu_float_tol (qu_float x)
{
#ifdef HAVE_LONG_LONG
	long long v = (long long) x->value;
	if ((long long) LONG_MIN <= v && v <= (long long) LONG_MAX)
#else
	double v = x->value;
	if ((double) LONG_MIN <= v && v <= (double) LONG_MAX)
#endif
	{
		errno = 0;
		return (long) v;
	}
	errno = ERANGE;
	return (v < 0) ? LONG_MIN : LONG_MAX;
}



unsigned long qu_float_toul (qu_float x)
{
#ifdef HAVE_LONG_LONG
	long long v = (long long) x->value;
	if (0 <= v && v <= (long long) ULONG_MAX)
#else
	double v = x->value;
	if (0 <= v && v <= (double) ULONG_MAX)
#endif
	{
		errno = 0;
		return (unsigned int) v;
	}
	errno = ERANGE;
	return (v < 0) ? 0 : ULONG_MAX;
}



#ifdef HAVE_LONG_LONG
long long qu_float_toll (qu_float x)
{
	double v = x->value;
	if ((double) LONG_LONG_MIN <= v && v <= (double) LONG_LONG_MAX)
	{
		errno = 0;
		return (long long) v;
	}
	errno = ERANGE;
	return (v < 0) ? LONG_LONG_MIN : LONG_LONG_MAX;
}



unsigned long long qu_float_toull (qu_float x)
{
	double v = x->value;
	if (0 <= v && v <= (double) LONG_LONG_MAX)
	{
		errno = 0;
		return (unsigned long long) v;
	}
	errno = ERANGE;
	return (v < 0) ? 0 : LONG_LONG_MAX;
}
#endif



float qu_float_tof (qu_float x)
{
	double v = x->value;

	errno = 0;
	if (v < 0)
	{
		if (v >= (double) FLT_MIN)
			return (float) v;
	}
	else
	{
		if (v <= (double) FLT_MAX)
			return (float) v;
	}
	errno = ERANGE;
	return (v < 0) ? FLT_MIN : FLT_MAX;
}



#ifdef HAVE_LONG_DOUBLE
long double qu_float_told (qu_float x)
{
	errno = 0;
	return (long double) x->value;
}
#endif



qu_int qu_float_as_int (qu_float x)
{
	long r = (long) x->value;
	QU_ARG_CHECKZ (QU_INT_MIN <= r && r <= QU_INT_MAX, "value out out Int range");
	return qu_int_new (r);
}



qu_integer qu_float_as_integer (qu_float x)
{
	long r = qu_float_tol (x);
	return (QU_INT_MIN <= r && r <= QU_INT_MAX)
		? (qu_integer) qu_int_new (r)
		: (qu_integer) qu_float_as_long (x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	Compare self with another object.

	This method return a negative Int, zero, or positive, if self is less than,
	the same, or greater than v{x} respectively.

	@noexc

	@xmp
		println (1.1 <=> 0.1)
		println (1.1 <=> 1.1)
		println (1.1 <=> 2.1)

	@out
		1
		0
		-1
*/
static qu_int qu__float__cmp (qu_float self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_float_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Float:    r = qu_float_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Long:     r = qu_float_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_float_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Big:      r = qu_float_cmp_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_float_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_float_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}



int qu_float_cmp_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	return QU_NUMBER_COMPARE_VALUE (a, b);
}



int qu_float_cmp_long (qu_float x, qu_long y)
{
	int sx = qu_float_sign (x);
	int sy = qu_long_sign (y);

	QU_NUMBER_COMPARE_SIGN (sx, sy);
	assert (sx == sy && sx);
	/*else*/
	{
		double b = qu_long_tod (y);
		if (errno == 0)
		{
			double a = x->value;
			return QU_NUMBER_COMPARE_VALUE (a, b);
		}
		return -sx;
	}
}



int qu_float_cmp_rational (qu_float x, qu_rational y)
{
	int sx = qu_float_sign (x);
	int sy = qu_rational_sign (y);

	QU_NUMBER_COMPARE_SIGN (sx, sy);
	assert (sx == sy && sx);
	/*else*/
	{
		double b = qu_rational_tod (y);
		if (errno == 0)
		{
			double a = x->value;
			return QU_NUMBER_COMPARE_VALUE (a, b);
		}
		return -sx;
	}
}



int qu_float_cmp_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;
	return QU_NUMBER_COMPARE_VALUE (a, b);
}



QU_NUMBER_BINOP_integer (qu_float, cmp, int)
QU_NUMBER_BINOP_decimal (qu_float, cmp, int)
QU_NUMBER_BINOP_real    (qu_float, cmp, int)
QU_NUMBER_BINOP_number  (qu_float, cmp, int)



/*
	|<=>|
*/

int qu_float_cmpabs_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	return QU_NUMBER_COMPARE_VALUE (a, b);
}



int qu_float_cmpabs_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		if (a < 0)
			a = -a;
		if (b < 0)
			b = -b;
		return QU_NUMBER_COMPARE_VALUE (a, b);
	}
	return -1;
}



int qu_float_cmpabs_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		if (a < 0)
			a = -a;
		if (b < 0)
			b = -b;
		return QU_NUMBER_COMPARE_VALUE (a, b);
	}
	return -1;
}



int qu_float_cmpabs_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;
	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	return QU_NUMBER_COMPARE_VALUE (a, b);
}



QU_NUMBER_BINOP_integer (qu_float, cmpabs, int)
QU_NUMBER_BINOP_decimal (qu_float, cmpabs, int)
QU_NUMBER_BINOP_real    (qu_float, cmpabs, int)
QU_NUMBER_BINOP_number  (qu_float, cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Decimal
	Return self decremented by 1.

	If the result can be represented with a Float then a Float is returned,
	a Big otherwise.

	@noexc

	@xmp
		i = 1.1
		println (i.decr)

	@out
		0.1
*/
qu_decimal qu__float__decr (qu_float self)
{
	double r = self->value - 1.0;
	return qu_double_isfinite (r)
		? (qu_decimal) qu_float_newd (r)
		: (qu_decimal) qu_big_decr (qu_float_as_big (self));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number
	Return the result of dividing self with v{x}.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{\} operator).

	Refer to -{Number.__div} for details on how the division is performed.

	@exc
	.EDiv
	.ETiny
	.EHuge
	.EInf
	.EComplex

	@xmp
		x = 11.0
		y = 3.0
		v = x / y
		i = x \ y
		r = x % y
		println (v)
		println (i)
		println (r)

	@out
		3.6666666666666665
		3
		2

	@note
	Qu bothers to check whether self is v{x} (the same object) and return
	Float (1) if it is.
*/
static qu_number qu__float__div (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_float_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_float_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_float_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_float_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_float_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_float_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



#define FLOAT_TRY_DOUBLE_DIV(type, a, b) \
QU_BEGIN \
	if ((a) == 0) \
	{ \
		QU_REQUIRE (QuEDiv, b); \
		return (type) QuFloat0; \
	} \
	else if ((b) == 0) \
	{ \
		return (qu_real) __qu_zerodiv__ (((a) > 0) ? 1 : -1); \
	} \
	else \
	{ \
		double r = (a) / (b); \
		if (qu_double_isfinite (r) && r) \
			return (type) qu_float_newd (r); \
	} \
QU_END



qu_real qu_float_div_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
	return (qu_real) qu_big_rdiv_float (qu_int_as_big (y), x);
}



qu_real qu_float_div_long (qu_float x, qu_long y)
{
	double b = qu_int_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
	}
	return (qu_real) qu_big_rdiv_float (qu_long_as_big (y), x);
}



qu_real qu_float_div_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
	}
	return (qu_real) qu_big_rdiv_float (qu_rational_as_big (y), x);
}



qu_real qu_float_div_float (qu_float x, qu_float y)
{
	if (x == y)
	{
		return (qu_real) QuFloat1;
	}
	else
	{
		double a = x->value;
		double b = y->value;
		FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
		return (qu_real) qu_big_rdiv_float (qu_float_as_big (y), x);
	}
}



qu_float qu_float_div_inf (qu_float x, qu_inf y)
{
	QU_REQUIRE (QuEDiv, !qu_float_iszero (x));
	return QuFloat0;
}



QU_NUMBER_BINOP_integer (qu_float, div, qu_real)
QU_NUMBER_BINOP_decimal (qu_float, div, qu_real)
QU_NUMBER_BINOP_real    (qu_float, div, qu_real)
QU_NUMBER_BINOP_number  (qu_float, div, qu_number)



/*
	/ reverse
*/

qu_real qu_float_rdiv_int (qu_float x, qu_int y)
{
	double a = qu_int_tod (y);
	double b = x->value;
	FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
	return (qu_real) qu_big_div_float (qu_int_as_big (y), x);
}



qu_real qu_float_rdiv_long (qu_float x, qu_long y)
{
	double a = qu_long_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
	}
	return (qu_real) qu_big_div_float (qu_long_as_big (y), x);
}



qu_real qu_float_rdiv_rational (qu_float x, qu_rational y)
{
	double a = qu_rational_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_DIV (qu_real, a, b);
	}
	return (qu_real) qu_big_div_float (qu_rational_as_big (y), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if v{x} is a -{Number} and has the same numerical value
	as self, false otherwise.

	Refer to -{Number.__eq} for details on how the equality test is performed.

	@noexc

	@xmp
		x = 1.0
		println (x == 1)

	@out
		true
*/
static qu_boolean qu__float__eq (qu_float self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_float_eq_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Float:    r = qu_float_eq_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Long:     r = qu_float_eq_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_float_eq_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Big:      r = qu_float_eq_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_float_eq_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



int qu_float_eq_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	return errno == 0 && x->value == b;
}



int qu_float_eq_rational (qu_float x, qu_rational y)
{
	if (qu_rational_iswhole (y))
	{
		return qu_float_eq_integer (x, y->num);
	}
	else if (qu_float_sign (x) == qu_rational_sign (y))
	{
		if (qu_float_iszero (x))
			return 1;
		else
		{
			double d = qu_rational_tod (y);
			return errno ? 0 : (d == x->value);
		}
	}
	return 0;
}



QU_NUMBER_BINOP_integer (qu_float, eq, int)
QU_NUMBER_BINOP_decimal (qu_float, eq, int)
QU_NUMBER_BINOP_real    (qu_float, eq, int)
QU_NUMBER_BINOP_number  (qu_float, eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (s:String) -> String
	Format this Float.

	See -{Number.__format} for details on the format.

	@exc
	.EFormat
	.EMemory

	@xmp
		f = 1234.56
		println (f.__format ('%.3d'))
		printfln ('f = %.2,3m', f)
		printfln ('f = %10.3e', f)

	@out
		1234.559
		f = 1,234.55
		f =   1.234e+3
*/
static qu_string qu__float__format (qu_float self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_float_format (self, &f);
}



qu_string qu_float_format (qu_float x, qu_format_s *f)
{
	double  v = x->value;
	char    b[512];
	char   *p, *e;
	int     l, expo;

	if (f->mode == 'y')
		return qu_format_yesno (f, x);

	if (v < 0)
		v = -v;

	sprintf (b, "%.17E", v);
	/* assert b[1] is the decimal point
		(which can be anything depending on the locale)
		and we are assuming that it is a single character */
	e = strchr (b, 'E');
	assert (e);
	expo = atoi (e + 1);
	*e = '\0';
	l = strlen (b);
	if (b[0] == '0')
	{
		expo--;
		p = b + 2;
		while (*p == '0')
		{
			p++;
			expo--;
			l--;
		}
		memmove (b, p, l);
		++l;
	}
	else
	{
		expo -= (l - 2);
		memmove (b + 1, b + 2, l);
		--l;
	}
	switch (f->mode)
	{
		case 'x':
		case 'o':
		case 'b': f->mode = 'd';
	}
	return qu_format_number (f, qu_string_new (b, l), qu_float_isneg (x), expo);
}



/*) sub __hash -> Int
	Return the hash value of this Float.

	@noexc

	@xmp
		x = 1.234
		println (x.__hash)

	@out
		1
*/
static qu_int qu__float__hash (qu_float self)
{
	return qu_int_new (qu_float_hashd (self->value));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Number
	Divide self with v{x} in an I{integer context}.

	@warn
	There is a big difference between the #{__div} method (the C{/} operator)
	and the __idiv method (the C{\} operator).

	Refer to -{Number.__idiv} for details on how the division is performed.

	@exc
	.EIdiv

	@xmp
		a = 10.0
		b = 3.9
		println (a \ b)

	@out
		2

	@note
	Qu bothers to check whether self is v{x} (the same object) and return
	Int (1) if it is.
*/
static qu_integer qu__float__idiv (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_float_idiv_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_float_idiv_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_float_idiv_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_float_idiv_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_float_idiv_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_float_idiv_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_float_idiv_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEIdiv);
}



#define FLOAT_DOUBLE_IDIV(a, b) \
QU_BEGIN \
	QU_REQUIRE (QuEIdiv, b); \
	if ((a) == 0) \
	{ \
		return (qu_integer) QuInt0; \
	} \
	else \
	{ \
		double r = (a) / (b); \
		if (qu_double_isfinite (r) && r) \
			return qu_integer_newd (r); \
	} \
QU_END



qu_integer qu_float_idiv_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	FLOAT_DOUBLE_IDIV (a, b);
	return qu_big_idiv_int (qu_float_as_big (x), y);
}



qu_integer qu_float_idiv_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno)
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		double a = x->value;
		FLOAT_DOUBLE_IDIV (a, b);
		return qu_big_idiv_long (qu_float_as_big (x), y);
	}
}



qu_integer qu_float_idiv_rational(qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno)
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		double a = x->value;
		FLOAT_DOUBLE_IDIV (a, b);
		return qu_big_idiv_rational (qu_float_as_big (x), y);
	}
}



qu_integer qu_float_idiv_float (qu_float x, qu_float y)
{
	if (x == y)
	{
		return (qu_integer) QuInt1;
	}
	else
	{
		double a = x->value;
		double b = y->value;
		FLOAT_DOUBLE_IDIV (a, b);
		return qu_big_idiv_float (qu_float_as_big (x), y);
	}
}



qu_integer qu_float_idiv_inf (qu_float x, qu_inf y)
{
	QU_REQUIRE (QuEIdiv, !qu_float_iszero (x));
	return (qu_integer) QuInt0;
}



QU_NUMBER_BINOP_integer (qu_float, idiv, qu_integer)
QU_NUMBER_BINOP_decimal (qu_float, idiv, qu_integer)
QU_NUMBER_BINOP_real    (qu_float, idiv, qu_integer)
QU_NUMBER_BINOP_number  (qu_float, idiv, qu_integer)



/*
	\ reversed
*/

qu_integer qu_float_ridiv_int(qu_float x, qu_int y)
{
	double a = qu_int_tod (y);
	double b = x->value;
	FLOAT_DOUBLE_IDIV (a, b);
	return qu_big_idiv_float (qu_int_as_big (y), x);
}



qu_integer qu_float_ridiv_long (qu_float x, qu_long y)
{
	double a = qu_long_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_DOUBLE_IDIV (a, b);
	}
	return qu_big_idiv_float (qu_long_as_big (y), x);
}



qu_integer qu_float_ridiv_rational (qu_float x, qu_rational y)
{
	double a = qu_rational_tod (y);
	if (errno)
	{
		double b = x->value;
		FLOAT_DOUBLE_IDIV (a, b);
	}
	return qu_big_idiv_float (qu_rational_as_big (y), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Decimal
	Return self incremented by 1.

	If the result can be represented with a Float then a Float is returned,
	a Big otherwise.

	@noexc

	@xmp
		i = 30.2
		println (i.incr)

	@out
		31.2
*/
qu_decimal qu__float__incr (qu_float self)
{
	double r = self->value + 1.0;
	return qu_double_isfinite (r)
		? (qu_decimal) qu_float_newd (r)
		: (qu_decimal) qu_big_incr (qu_float_as_big (self));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x) -> Boolean
	If v{x} is a number and they have the same absolute values then true
	is returned, false otherwise.

	@xmp
		println (-1.0 === 1)

	@out
		true
*/
qu_boolean qu__float__like (qu_float self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_float_like_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Float:    r = qu_float_like_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Long:     r = qu_float_like_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_float_like_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Big:      r = qu_float_like_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_float_like_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



int qu_float_like_rational (qu_float x, qu_rational y)
{
	if (qu_rational_iswhole (y))
	{
		return qu_float_like_integer (x, y->num);
	}
	else if (qu_float_sign (x) == qu_rational_sign (y))
	{
		if (qu_float_iszero (x))
			return 1;
		else
		{
			double b = qu_rational_tod (y);
			if (errno == 0)
			{
				double a = x->value;
				if (a < 0)
					a = -a;
				if (b < 0)
					b = -b;
				return (a == b);
			}
		}
	}
	return 0;
}



int qu_float_like_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;

	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	return a == b;
}



QU_NUMBER_BINOP_integer (qu_float, like, int)
QU_NUMBER_BINOP_decimal (qu_float, like, int)
QU_NUMBER_BINOP_real    (qu_float, like, int)
QU_NUMBER_BINOP_number  (qu_float, like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Return the result of subtracting self with v{x}.

	Refer to -{Number.__minus} for for details on how the subtraction is performed.

	@exc
	.EMinus
	.ETiny
	.EHuge
	.EInf
	.EComplex

	@xmp
		println (1.1 - 3.2)

	@out
		-2.1
*/
static qu_number qu__float__minus (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_float_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_float_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_float_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_float_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_float_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_float_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



#define FLOAT_TRY_DOUBLE_MINUS(type, a, b) \
QU_BEGIN \
	if ((a) == (b)) \
		return (type) QuFloat0; \
	else \
	{ \
		double r = (a) - (b); \
		if (qu_double_isfinite (r) && r) \
			return (type) qu_float_newd (r); \
	} \
QU_END



qu_decimal qu_float_minus_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	FLOAT_TRY_DOUBLE_MINUS (qu_decimal, a, b);
	return (qu_decimal) qu_big_minus_int (qu_float_as_big (x), y);
}



qu_real qu_float_minus_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_MINUS (qu_real, a, b);
	}
	return (qu_real) qu_big_minus_long (qu_big_new_float (x), y);
}



qu_real qu_float_minus_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_MINUS (qu_real, a, b);
	}
	return (qu_real) qu_big_minus_rational (qu_big_new_float (x), y);
}



qu_decimal qu_float_minus_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;
	FLOAT_TRY_DOUBLE_MINUS (qu_decimal, a, b);
	return (qu_decimal) qu_big_minus_float (qu_float_as_big (x), y);
}



QU_NUMBER_BINOP_integer (qu_float, minus, qu_real)
QU_NUMBER_BINOP_decimal (qu_float, minus, qu_real)
QU_NUMBER_BINOP_real    (qu_float, minus, qu_real)
QU_NUMBER_BINOP_number  (qu_float, minus, qu_number)



/*
	- reverse
*/

qu_decimal qu_float_rminus_int (qu_float x, qu_int y)
{
	double a = qu_int_tod (y);
	double b = x->value;
	FLOAT_TRY_DOUBLE_MINUS (qu_decimal, a, b);
	return (qu_decimal) qu_big_minus_float (qu_int_as_big (y), x);
}



qu_real qu_float_rminus_long (qu_float x, qu_long y)
{
	double a = qu_long_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_MINUS (qu_real, a, b);
	}
	return (qu_real) qu_big_minus_float (qu_long_as_big (y), x);
}



qu_real qu_float_rminus_rational (qu_float x, qu_rational y)
{
	double a = qu_rational_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_MINUS (qu_real, a, b);
	}
	return (qu_real) qu_big_minus_float (qu_rational_as_big (y), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Return the result of multiplying self with v{x}.

	Refer to -{Number.__mul} for for details on how the multiplication is performed.

	@exc
	.EMul
	.EHuge
	.ETiny
	.EInf
	.EComplex

	@xmp
		println (8.1 * 3.1)

	@out
		25.11
*/
static qu_number qu__float__mul (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_float_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_float_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_float_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_float_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_float_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_float_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



#define FLOAT_TRY_DOUBLE_MUL(type, a, b) \
QU_BEGIN \
	if ((a) == 0 || (b) == 0) \
		return (type) QuFloat0; \
	else \
	{ \
		double r = (a) * (b); \
		if (qu_double_isfinite (r) && r) \
			return (type) qu_float_newd (r); \
	} \
QU_END



qu_decimal qu_float_mul_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	FLOAT_TRY_DOUBLE_MUL (qu_decimal, a, b);
	return (qu_decimal) qu_big_mul_float (qu_int_as_big (y), x);
}



qu_real qu_float_mul_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_MUL (qu_real, a, b);
	}
	return (qu_real) qu_big_mul_float (qu_long_as_big (y), x);
}



qu_real qu_float_mul_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_MUL (qu_real, a, b);
	}
	return (qu_real) qu_big_mul_float (qu_rational_as_big (y), x);
}



qu_decimal qu_float_mul_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;
	FLOAT_TRY_DOUBLE_MUL (qu_decimal, a, b);
	return (qu_decimal) qu_big_mul_float (qu_float_as_big (y), x);
}



QU_NUMBER_BINOP_integer (qu_float, mul, qu_real)
QU_NUMBER_BINOP_decimal (qu_float, mul, qu_real)
QU_NUMBER_BINOP_real    (qu_float, mul, qu_real)
QU_NUMBER_BINOP_number  (qu_float, mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Float
	Return the negation of self.

	@xmp
		x = 1.1
		println (-x)

	@out
		-1.1
*/
qu_float qu__float__neg (qu_float self)
{
	double v = self->value;
	return v ? qu_float_newd (-v) : QuFloat0;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Return the result of adding v{x} to self.

	Refer to -{Number.__plus} for for details on how the addition is performed.

	@exc
	.EPlus
	.EHuge
	.ESmall
	.EInf
	.EComplex

	@xmp
		println (8.1 + 3.1)

	@out
		11.2
*/
static qu_number qu__float__plus (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_float_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_float_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_float_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_float_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_float_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_float_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPlus);
}



#define FLOAT_TRY_DOUBLE_PLUS(type, a, b) \
QU_BEGIN \
	if ((a) == -(b)) \
		return (type) QuFloat0; \
	else \
	{ \
		double r = (a) + (b); \
		if (qu_double_isfinite (r) && r) \
			return (type) qu_float_newd (r); \
	} \
QU_END



qu_decimal qu_float_plus_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);

	FLOAT_TRY_DOUBLE_PLUS (qu_decimal, a, b);
	return (qu_decimal) qu_big_plus_int (qu_float_as_big (x), y);
}



qu_real qu_float_plus_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_PLUS (qu_real, a, b);
	}
	return (qu_real) qu_big_plus_long (qu_float_as_big (x), y);
}



qu_real qu_float_plus_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_PLUS (qu_real, a, b);
	}
	return (qu_real) qu_big_plus_rational (qu_float_as_big (x), y);
}



qu_decimal qu_float_plus_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;
	FLOAT_TRY_DOUBLE_PLUS (qu_decimal, a, b);
	return (qu_decimal) qu_big_plus_float (qu_float_as_big (x), y);
}



QU_NUMBER_BINOP_integer (qu_float, plus, qu_real)
QU_NUMBER_BINOP_decimal (qu_float, plus, qu_real)
QU_NUMBER_BINOP_real    (qu_float, plus, qu_real)
QU_NUMBER_BINOP_number  (qu_float, plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub pos -> self
	This method always return self.

	@noexc
*/
#define qu__float__pos  qu_object_method_self



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x: Number) -> Number
	Return the result of raising self to the power of v{x}.

	Refer to -{Number.__pow} for for details on how the operation is performed.

	@exc
	.EPow
	.ETiny
	.EHuge
	.EInf
	.EComplex

	@xmp
		x = 2.0
		y = x ** 2.0
		println (y)

	@out
		4
*/
static qu_number qu__float__pow (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_float_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_float_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_float_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_float_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_float_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_float_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}



#define FLOAT_TRY_DOUBLE_POW(a, b) \
QU_BEGIN \
	if ((a) == 0) \
	{ \
		if ((b) < 0) \
			return (qu_number) __qu_inf__ (1); \
		return (qu_number) QuFloat0; \
	} \
	else \
	{ \
		double r; \
		errno = 0; \
		r = pow (a, b); \
		if (errno == 0) \
		{ \
			if (qu_double_isnan (r)) \
				QU_THROW (QuEPow); \
			else if (qu_double_isfinite (r) && r) \
				return (qu_number) qu_float_newd (r); \
		} \
	} \
QU_END



qu_number qu_float_pow_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	FLOAT_TRY_DOUBLE_POW (a, b);
	if (errno == EDOM)
		return (qu_number) qu_complex_rpow_float (qu_int_as_complex (y), x);
	return (qu_number) qu_big_pow_int (qu_float_as_big (x), y);
}



qu_number qu_float_pow_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_POW (a, b);
		if (errno == EDOM)
			return (qu_number) qu_complex_rpow_float (qu_long_as_complex (y), x);
	}
	return (qu_number) qu_big_pow_long (qu_float_as_big (x), y);
}



qu_number qu_float_pow_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_POW (a, b);
		if (errno == EDOM)
			return (qu_number) qu_complex_rpow_float (qu_rational_as_complex (y), x);
	}
	return (qu_number) qu_big_pow_rational (qu_float_as_big (x), y);
}



qu_number qu_float_pow_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;

	FLOAT_TRY_DOUBLE_POW (a, b);
	if (errno == EDOM)
		return (qu_number) qu_complex_rpow_float (qu_float_as_complex (y), x);
	return (qu_number) qu_big_pow_float (qu_float_as_big (x), y);
}



qu_real qu_float_pow_inf (qu_float x, qu_inf y)
{
	if (y == QuInfPos)
	{
		if (qu_float_iszero (x))
			return (qu_real) QuFloat0;

		else if (qu_float_isone (x))
			return (qu_real) QuFloat1;

		else
			return (qu_real) __qu_inf__ (1);
	}
	else
	{
		if (qu_float_iszero (x))
			return (qu_real) __qu_inf__ (1);

		return qu_float_isone (x) ? (qu_real) QuFloat1 : (qu_real) QuFloat0;
	}
}



QU_NUMBER_BINOP_integer (qu_float, pow, qu_number)
QU_NUMBER_BINOP_decimal (qu_float, pow, qu_number)
QU_NUMBER_BINOP_real    (qu_float, pow, qu_number)
QU_NUMBER_BINOP_number  (qu_float, pow, qu_number)



/*
	** reversed
*/

qu_number qu_float_rpow_int (qu_float x, qu_int y)
{
	double a = qu_int_tod (y);
	double b = x->value;
	FLOAT_TRY_DOUBLE_POW (a, b);
	if (errno == EDOM)
		return (qu_number) qu_complex_pow_float (qu_int_as_complex (y), x);
	return (qu_number) qu_big_pow_float (qu_int_as_big (y), x);
}



qu_number qu_float_rpow_long (qu_float x, qu_long y)
{
	double a = qu_long_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_POW (a, b);
		if (errno == EDOM)
			return (qu_number) qu_complex_pow_float (qu_long_as_complex (y), x);
	}
	return (qu_number) qu_big_pow_float (qu_long_as_big (y), x);
}



qu_number qu_float_rpow_rational (qu_float x, qu_rational y)
{
	double a = qu_rational_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_POW (a, b);
		if (errno == EDOM)
			return (qu_number) qu_complex_pow_float (qu_rational_as_complex (y), x);
	}
	return (qu_number) qu_big_pow_float (qu_rational_as_big (y), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	This simply prints the string returned by the ={__string} method.

	See the builtin -{print} function for possible errors.
*/
static qu_null qu__float__print (qu_float self)
{
	char buf [64];
	sprintf (buf, "%.16g", self->value);
	return qu_sys_printz (buf, strlen (buf));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The meaning of "remainder" may differ with other languages.
	Refer to -{Number.__rem} for details.

	@p
	The remainder of dividing by zero is undefined thus throws ERem.

	Note that the remainder is not an integral number. If you want the
	same effect as integer rem then you will specifically have to
	use the m{Math} module f{floor} function or whatever you think is
	the correct answer for your application.

	@exc
	.ERem
	.EInf
	.EBrain

	@xmp
		use Math
		a = 8.1 % 3
		println (a)
		println (a.floor)

	@out
		2.0999999999999996
		2
*/
static qu_number qu__float__rem (qu_float self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_float_rem_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_rem_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_float_rem_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_float_rem_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_float_rem_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_float_rem_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_float_rem_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuERem);
}



#define FLOAT_TRY_DOUBLE_REM(type, a, b) \
QU_BEGIN \
	if ((b) == 0) \
		QU_THROW (QuERem); \
	else if ((a) == 0) \
	{ \
		return (type) QuFloat0; \
	} \
	else \
	{ \
		double r = fmod (a, b); \
		if (qu_double_isfinite (r) && r) \
			return (type) qu_float_newd (r); \
	} \
QU_END



qu_real qu_float_rem_int (qu_float x, qu_int y)
{
	double a = x->value;
	double b = qu_int_tod (y);
	FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	return qu_big_rrem_float (qu_int_as_big (y), x);
}



qu_real qu_float_rem_long (qu_float x, qu_long y)
{
	double b = qu_long_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	}
	return qu_big_rrem_float (qu_long_as_big (y), x);
}



qu_real qu_float_rem_rational (qu_float x, qu_rational y)
{
	double b = qu_rational_tod (y);
	if (errno == 0)
	{
		double a = x->value;
		FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	}
	return qu_big_rrem_float (qu_rational_as_big (y), x);
}



qu_real qu_float_rem_float (qu_float x, qu_float y)
{
	double a = x->value;
	double b = y->value;
	FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	return qu_big_rrem_float (qu_float_as_big (y), x);
}



qu_float qu_float_rem_inf (qu_float x, qu_inf y)
{
	QU_REQUIRE (QuERem, !qu_float_iszero (x));
	return x;
}



QU_NUMBER_BINOP_integer (qu_float, rem, qu_real)
QU_NUMBER_BINOP_decimal (qu_float, rem, qu_real)
QU_NUMBER_BINOP_real    (qu_float, rem, qu_real)
QU_NUMBER_BINOP_number  (qu_float, rem, qu_number)



/*
	% reversed
*/

qu_real qu_float_rrem_int (qu_float x, qu_int y)
{
	double a = qu_int_tod (y);
	double b = x->value;
	FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	return (qu_real) qu_big_rem_float (qu_int_as_big (y), x);
}



qu_real qu_float_rrem_long (qu_float x, qu_long y)
{
	double a = qu_long_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	}
	return (qu_real) qu_big_rem_float (qu_long_as_big (y), x);
}



qu_real qu_float_rrem_rational (qu_float x, qu_rational y)
{
	double a = qu_rational_tod (y);
	if (errno == 0)
	{
		double b = x->value;
		FLOAT_TRY_DOUBLE_REM (qu_real, a, b);
	}
	return (qu_real) qu_big_rem_float (qu_rational_as_big (y), x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Return a string representation of self.

	@xmp
		s = 1.234 as String
		println (s)

	@out
		1.234
*/
qu_string qu__float__string (qu_float self)
{
	char buf [64];

	sprintf (buf, "%.16g", self->value);
	return qu_string_newz (buf);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	If self is exactly 0 then false is returned, true otherwise.

	@noexc

	@xmp
		x = 0
		println (Boolean (x))

	@out
		false
*/
qu_boolean qu__float__true (qu_float self)
{
	return qu_float_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub abs -> Float
	Return the absolute value.
*/
static qu_float qu__float_abs (qu_float self)
{
	if (self->value >= 0)
		return self;
	return qu_float_new (-self->value);
}



/*) sub ceil -> Decimal
	Return the result of rounding this number to the nearest integer towards
	positive infinity.
*/
qu_decimal qu__float_ceil (qu_float x)
{
	double v = ceil (x->value);
	if (qu_double_isfinite (v))
		return (qu_decimal) qu_float_newd (v);
	return (qu_decimal) qu_big_ceil (qu_float_as_big (x));
}



/*) sub floor -> Decimal
	Return the result of rounding this number to the nearest integer
	towards negative infinity.
*/
qu_decimal qu__float_floor (qu_float x)
{
	double v = floor (x->value);
	if (qu_double_isfinite (v))
		return (qu_decimal) qu_float_newd (v);
	return (qu_decimal) qu_big_floor (qu_float_as_big (x));
}



/*) sub frac -> Float
	Return the fraction part.
*/
qu_float qu__float_frac (qu_float x)
{
	double i, f;
	f = modf (x->value, &i);
	return f ? qu_float_new (f) : QuFloat0;
}



/*) sub iseven -> Boolean
	Return true if self is whole and even number, false otherwise.
*/
static qu_boolean qu__float_iseven (qu_float self)
{
	return qu_float_isodd (self) ? QuFalse : QuTrue;
}



/*) sub isfrac -> Boolean
	Return true if C{0 < self < 1}, false otherwise.
*/
static qu_boolean qu__float_isfrac (qu_float self)
{
	return qu_float_isfrac (self) ? QuTrue : QuFalse;
}



/*) sub isodd -> Boolean
	Return true if self is an odd number, false otherwise.
*/
static qu_boolean qu__float_isodd (qu_float self)
{
	return qu_float_isodd (self) ? QuTrue : QuFalse;
}



/*) sub iswhole -> Boolean
	Return true if self is a I{whole} number, false otherwise.
*/
static qu_boolean qu__float_iswhole (qu_float self)
{
	return qu_float_iswhole (self) ? QuTrue : QuFalse;
}



/*) sub pack -> String
	Pack this Float into a String.

	@exc
	.EImplement native size is not supported
*/
qu_string qu__float_pack (qu_float self)
{
#if SIZEOF_DOUBLE == 8
	double v = self->value;
	return qu_string_new ((char *) &v, 8);
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub round (precision: Int) -> Float
	Return the result of rounding this number to the specified
	v{precision} where v{precision} is the number of decimal digits.
*/
static qu_float qu__float_round (qu_float self, qu_integer prec)
{
	int p;
	QU_ARG_int (prec, p);
	return qu_float_round (self, p);
}



qu_float qu_float_round (qu_float x, int prec)
{
	char f[16], b[512];

	if (qu_float_iswhole (x) || prec < 1 || prec > 16)
		return x;

	sprintf (f, "%%.%df", prec);
	sprintf (b, f, x->value);
	return qu_float_new (atof (b));
}



/*) sub sign -> Int
	Return -1, 0, 1 if self is negative, zero and postive respectively.
*/
static qu_int qu__float_sign (qu_float self)
{
	double v = self->value;
	return (v < 0) ? QuInt_1 : ((v > 0) ? QuInt0 : QuInt1);
}



/*) sub trunc -> Float
	Return the result of rounding this number to the nearest integer
	towards zero, thus removing the fractional part.
*/
qu_float qu__float_trunc (qu_float x)
{
	double v = x->value;
	v = (v < 0) ? ceil (v) : floor (v);
	assert (qu_double_isfinite (v));
	return qu_float_newd (v);
}



/*----------------------------------------------------------------------------
	misc
----------------------------------------------------------------------------*/

/* what integer must be multiplied to x to get an integer? */
qu_integer qu_float_integer_factor (qu_float x)
{
	if (qu_float_iswhole (x))
		return (qu_integer) QuInt1;
	else
	{
		double v = x->value;
		char   buf [64], *s;
		int    e, l;

		if (v < 0)
			v = -v;
		sprintf (buf, "%.17e", v);
		assert (buf[1] == '.');
		s = strchr (buf, 'e');
		l = s - buf + 1;
		assert (s);
		e = atoi (s + 1);

		/* TODO - should have qu_int_pow10 that possibly returns an qu_int */
		return (qu_integer) qu_long_pow_10 (l - e);
	}
}



/*----------------------------------------------------------------------------
	crappy C stuffs
----------------------------------------------------------------------------*/

double __QuDoubleInfp;
double __QuDoubleInfn;
int qu_double__ne (double a, double b)
{
	return a != b;
}

#ifdef HAVE_LONG_DOUBLE
long double __QuLongDoubleInfp;
long double __QuLongDoubleInfn;
int qu_longdouble__ne (long double a, long double b)
{
	return a != b;
}
#endif



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_float__setup ()
{
	__QuDoubleInfp = 1.0 / 0.0;
	__QuDoubleInfn =  -__QuDoubleInfp;

#ifdef HAVE_LONG_DOUBLE
	__QuLongDoubleInfp = (long double) 1.0 / (long double) 0.0;
	__QuLongDoubleInfn = -__QuLongDoubleInfp;
#endif

	qu_class_setup_final (QuFloatClass, qu_float_s, 0, 0);

	QU_FLOAT_MIN = qu_float_new_static (-DBL_MAX);
	QU_FLOAT_MAX = qu_float_new_static (DBL_MAX);
	QuFloat0     = qu_float_new_static (0.0);
	QuFloat1     = qu_float_new_static (1.0);
	QuFloat2     = qu_float_new_static (2.0);
	QuFloat_1    = qu_float_new_static (-1.0);
	QuFloat_2    = qu_float_new_static (-2.0);
}



void qu_float__init ()
{
	/* constants */
	qu_class_constz  (QuFloatClass, "MIN"     , QU_FLOAT_MIN);
	qu_class_constz  (QuFloatClass, "MAX"     , QU_FLOAT_MAX);
	qu_class_constz  (QuFloatClass, "PACKSIZE", qu_int_new (8));

	/* standard class methods */
	qu_class_method_class  (QuFloatClass, QU_NAME_new, qu__float_new, 1);

	/* specific class methods */
	qu_class_method_classz (QuFloatClass, "unpack", qu__float_unpack, -1);

	/* standard instance methods */
	qu_class_method_inst   (QuFloatClass, QU_NAME__as     , qu__float__as     ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__cmp    , qu__float__cmp    ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__decr   , qu__float__decr   ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__div    , qu__float__div    ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__eq     , qu__float__eq     ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__format , qu__float__format ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__hash   , qu__float__hash   ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__idiv   , qu__float__idiv   ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__incr   , qu__float__incr   ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__like   , qu__float__like   ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__minus  , qu__float__minus  ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__mul    , qu__float__mul    ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__neg    , qu__float__neg    ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__plus   , qu__float__plus   ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__pos    , qu__float__pos    ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__pow    , qu__float__pow    ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__print  , qu__float__print  ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__rem    , qu__float__rem    ,  1);
	qu_class_method_inst   (QuFloatClass, QU_NAME__string , qu__float__string ,  0);
	qu_class_method_inst   (QuFloatClass, QU_NAME__true   , qu__float__true   ,  0);

	/* specific instance methods */
	qu_class_method_instz  (QuFloatClass, "abs"     , qu__float_abs     ,  0);
	qu_class_method_instz  (QuFloatClass, "ceil"    , qu__float_ceil    ,  0);
	qu_class_method_instz  (QuFloatClass, "floor"   , qu__float_floor   ,  0);
	qu_class_method_instz  (QuFloatClass, "frac"    , qu__float_frac    ,  0);
	qu_class_method_instz  (QuFloatClass, "iseven"  , qu__float_iseven  ,  0);
	qu_class_method_instz  (QuFloatClass, "isfrac"  , qu__float_isfrac  ,  0);
	qu_class_method_instz  (QuFloatClass, "isodd"   , qu__float_isodd   ,  0);
	qu_class_method_instz  (QuFloatClass, "iswhole" , qu__float_iswhole ,  0);
	qu_class_method_instz  (QuFloatClass, "pack"    , qu__float_pack    ,  0);
	qu_class_method_instz  (QuFloatClass, "round"   , qu__float_round   ,  1);
	qu_class_method_instz  (QuFloatClass, "sign"    , qu__float_sign    ,  0);
	qu_class_method_instz  (QuFloatClass, "trunc"   , qu__float_trunc   ,  0);

	/* finalize declaration */
	qu_class_set_super (QuFloatClass, QuDecimalClass);
	qu_class_enddecl   (QuFloatClass);
}
