/* Frame.c - Copyright (c) Marc Krisnanto */


#include "qu.h"


/*) class Frame __final
	Frame objects are created to execute -{Code} objects, i.e. initializing
	module and calling program level functions/methods (calling C level functions
	does not require a Frame).

	You can not create Frame objects from program level directly but
	you can get one when you call a generator. There are a number of methods
	provided to get information on Frame objects such as #{code}, #{mod},
	#{done}, etc.
*/
qu_class QuFrameClass;


/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__frame__destroy (qu_frame self)
{
	assert (!qu_frame_is_cframe (self));
	qu_mem_free (self->ss);
}



static void qu__frame__mark (qu_frame self)
{
	if (qu_frame_done (self)) /* || qu_frame_is_cframe (self)) */
		return;

	qu_gc_xmark (self->result);
	qu_gc_mark  (self->code);
	qu_gc_mark  (self->self);
	qu_gc_xmark (self->eclass);

	/* mark stack */
	{
		qu_any  *p = self->ss;
		qu_any *sp = self->sp;

		for (; p < sp; p++)
		{
			qu_gc_mark (*p);
		}

		/* top of stack may be NULL */
		qu_gc_xmark (*p);
	}

	/* mark block */
	{
		int d;
		qu_any x;

		for (d = self->level; d >= 0; d--)
		{
			x = self->blocks[d].obj;
			qu_gc_xmark (x);
		}
	}
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Array} |
		Collect all yielded values in an Array.
		Frame is exhausted after a call to this method.

	.| -{Set} |
		Collect all yielded in a Set.
		Frame is exhausted after a call to this method.

	.| -{Boolean} | Delegate to #{__true}.
	.| -{String}  | Delegate to #{__string}.
	.| -{Frame}   | Return self.

	@exc
	.EArgValue

	@xmp
		sub foo
			yield 1
			yield 2
			yield 1
		;;

		a = foo as Array
		println (a)

		a = foo as Set
		println (a)

	@out
		[1, 2, 1]
		{1, 2}
*/
static qu_any qu__frame__as (qu_frame self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return qu_frame_as_array   (self);
		case QU_CLASS_ID_Set:     return qu_frame_as_set     (self);
		case QU_CLASS_ID_Dict:    return qu_frame_as_dict    (self);
		case QU_CLASS_ID_String:  return qu_frame_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_frame_as_boolean (self);
		case QU_CLASS_ID_Frame:   return self;
	}
	QU_ARG_THROW;
}


qu_array qu_frame_as_array (qu_frame self)
{
	qu_array a = qu_array_new (4);
	for (;;)
	{
		qu_any x = qu_frame_next (self);
		if (x == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			break;
		}
		QU_ENSURE (qu_array_push (a, x));
	}
	return a;
}



qu_set qu_frame_as_set (qu_frame self)
{
	qu_set a = qu_set_new (4);
	for (;;)
	{
		qu_any x = qu_frame_next (self);
		if (x == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			break;
		}
		QU_ENSURE (qu_set_push (a, x));
	}
	return a;
}



qu_dict qu_frame_as_dict (qu_frame self)
{
	qu_dict a = qu_dict_new (4);
	for (;;)
	{
		qu_any x = qu_frame_next (self);
		if (x == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			break;
		}
		if (qu_array_is (x) && ((qu_array) x)->len == 2)
			QU_ENSURE (qu_dict_setitem (a, ((qu_array) x)->objs[0], ((qu_array) x)->objs[1]));
		else
			QU_ENSURE (qu_dict_setitem (a, x, QuNil));
	}
	return a;
}



/*) sub __call (*arg) -> Any
	Call a Frame.

	This simply performs #{__stop}C{(*arg)} and return the result of #{next}.

	@exc
	.Exception anything thrown by #{__stop} or #{next}.
*/
qu_any qu__frame__call (qu_frame self, int argc, qu_any *argv)
{
	QU_ENSURE (qu_frame_stop (self, argc, argv));
	return qu_frame_next (self);
}



/*) sub __copy -> Frame
	Create a new Frame with the current state of this Frame.

	Note that you can copy the Frame while it is running as demonstrated
	in the following I{not_such_a_good_idea} example.

	@exc
	.ENone Frame has finished running

	@xmp
		g = gen (1)
		h = nil

		sub gen (x)
			yield x
			if h == nil
				h = g.__copy
			;;
			yield x + 1
			yield x + 2
		;;

		for i in g
			print (i, ' ')
		;;
		println
		for i in h
			print (i, ' ')
		;;
		println

	@out
		1 2 3
		2 3
*/
/*) sub copy
	This is just an alias of #{__copy}.
*/
static qu_frame qu__frame__copy (qu_frame self)
{
	QU_REQUIRE (QuENone, !qu_frame_done (self));
	QU_REQUIRE (QuENone,  qu_frame_is_sub (self));
	{
		qu_sub   u = (qu_sub) self->code->parent;
		qu_frame f;

		QU_REQUIRE (QuENone, qu_sub_is (u));
		f         = qu__frame_new (self->code, self->self, u->nlocal);
		f->eclass = self->eclass;
		f->flags  = self->flags;
		assert (qu_gc_is_static (f));
		f->ip     = self->ip;
		f->sp     = f->ss + (self->sp - self->ss);
		f->argc   = self->argc;
		f->line   = self->line;
		f->level  = self->level;
		f->atomic = self->atomic;
		f->result = self->result;

		QU_OBJECT_LIST_COPY (f->ss, self->ss, self->sp - self->ss + 1);
		memcpy (f->blocks, self->blocks, sizeof (qu_frblk_s) * f->level);

		return f;
	}
}



/*) sub __each -> Iterator
	Create an iterator to run the Frame one step at a time.

	@noexc
*/
static qu_any qu__frame__each_next (qu_frame self, qu_iterator t)
{
	return qu_frame_next (self);
}



static qu_iterator qu__frame__each (qu_frame self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, qu__frame__each_next);
	i->ex2 = 1;
	return i;
}



/*  sub __in (x: qu_any) -> qu_boolean
	Whether or not v{x} is a value in the iterated object.

	@warn
	After a call to this method, the qu_frame object may be exhausted.

	In fact, since the qu_frame may have already stopped, the result of this
	method is not reliable unless you know what you are doing.

	@noexc

	@xmp
		sub foo
			yield 1
			yield 2
			yield 3
		;;

		x = foo
		println (2 in x)
		println (x.next)

	@out
		true
		3
*/
static qu_boolean qu__frame__in (qu_frame self, qu_any x)
{
	/* Remember! the __in method may not throw */

	if (qu_frame_running (self) || qu_frame_done (self) || qu_frame_thrown (self))
		return QuFalse;

	if (!qu_frame_is_generator (self))
		return QuFalse;

	for (;;)
	{
		qu_any r = qu_frame_next (self);
		if (r == 0)
			break;
		if (qu_sys_eq (x, r) == QuTrue)
			return QuTrue;
	}
	return QuFalse;
}



/*  sub __like (x: qu_any) -> qu_boolean
	Return true if v{x} is a qu_frame and has the same -{qu_code} object,
	false otherwise.

	@noexc

	@xmp
		sub foo
			yield 1
			yield 2
			yield 3
		;;

		x = foo
		y = foo
		println (x == y)
		println (x === y)

	@out
		false
		true
*/
static qu_boolean qu__frame__like (qu_frame self, qu_any x)
{
	return (qu_frame_is (x) && ((qu_frame) x)->code == self->code) ? QuTrue : QuFalse;
}



/*) sub __next -> Any
	Get the next value yielded by the Frame.

	@exc
	.ERuntime the Frame is currently running
	.EStop    already done
	.ECall    not a generator

	@xmp
		sub foo
			yield 1
			yield 2
			yield 3
		;;

		f = foo
		i = each f
		println (f)
		println (i)
		println (i.__next)
		println (i.__next)

	@out
		<a Frame>
		<a Iterator>
		1
		2
*/
/*) sub next
	This is just an alias of #{__next}.
*/
qu_any qu__frame__next (qu_frame self)
{
	/*
		!!! QU_FRAME_FLAG_THROWN has no association with qu_thread->eclass
		!!! because another thread may be executing this frame.

		We don't record exceptions in qu_frame objects for efficiency.
	*/
	QU_REQUIRE (QuECall   ,  qu_frame_is_generator (self));
	QU_REQUIRE (QuERuntime, !qu_frame_running   (self));
	QU_REQUIRE (QuEStop   , !qu_frame_done      (self));
#if 0
	QU_REQUIRE (QuECall   , !qu_frame_thrown    (self));
#endif
	if (qu_frame_thrown (self))
	{
		qu_class x = self->eclass;
		if (x == 0)
			x = QuERuntimeClass;
		qu_thread_throwex (x, 0, self->code->parent, self->line);
		return 0;
	}
	return qu_interp_frame_next (self);
}



/*  sub __print -> qu_null
	Print this qu_frame to v{stdout}.

	See -{print} for possible exceptions.

	@xmp
		sub foo
			yield 1
			yield 2
		;;

		f = foo
		println (f)

	@out
		<qu_frame foo>
		true
*/
static qu_null qu__frame__print (qu_frame self)
{
	qu_string name = self->code->parent->name;
	QU_ENSURE (qu_sys_printz ("<Frame ", 7));
	QU_ENSURE (qu_sys_printz (name->str, name->len));
	QU_ENSURE (qu_sys_printz (">", 1));
	return QuNil;
}



/*) sub __stop (exc = EStop, msg = nil) -> Null
	Stop execution of a generator by throwing an exception.

	@exc
	.EArgValue
	.ECall     not a generator or Frame has stopped

	@xmp
		sub foo (x)
			yield x
			yield x + 1
			yield x + 2
		;;

		f = foo (1)
		println (f.next)
		f.__stop (EValue)
		try
			f.next
		catch
			println (__err__)
		;;

	@out
		1
		(EValue) invalid type or value
*/
/*) sub stop
	This is just an alias of #{__stop}.
*/
qu_null qu__frame__stop (qu_frame self, int argc, qu_any *argv)
{
	qu_class c = QuEStopClass;
	qu_any   m = 0;

	/* TODO - qu_frame->emsg? */

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		qu_any e = argv[0];
		if (qu_exception_is (e))
		{
			c = ((qu_exception) e)->klass;
		}
		/*HACK - 'with' statement */
		else if (e == __QuSys__err__)
		{
			e = qu__sys__err__ ();
			if (e && qu_exception_is (e))
				c = ((qu_exception) e)->klass;
		}
		else
		{
			QU_ARG_CHECK (qu_exception_class_is (e));
			c = e;
		}

		if (argc > 1)
			m = argv[1];
	}
	qu_flag_set (self, QU_FRAME_FLAG_THROWN);
	self->eclass = c;
	return QuNil;
}



/*) sub __string -> String
	Return a String representation of this Frame.

	@noexc
*/
qu_string qu__frame__string (qu_frame self)
{
	qu_string name = self->code->parent->name;
	return qu_string_newf ("<Frame %s>", name->str);
}



/*) sub __true -> Boolean
	Return false if the Frame has stopped, true otherwise.

	@noexc
*/
qu_boolean qu__frame__true (qu_frame self)
{
	return qu_frame_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub argc -> Int
	Return the number of arguments passed to this Frame.

	@noexc

	@xmp
		sub foo (x)
			yield x
		;;

		f = foo (3)
		println (f.argc)

	@out
		1
*/
static qu_int qu__frame_argc (qu_frame self)
{
	assert (self->argc <= QU_INT_MAX);
	return qu_int_new (self->argc);
}



/*) sub code -> Code
	Return the -{Code} object associated with this Frame.

	@exc
	.ENone Frame has finished running

	@xmp
		sub foo
			yield 1
			yield 2
			yield 3
		;;

		f = foo
		println (f)
		println (f.code)
		println (f.code.stacksize)

	@out
		<a Frame>
		 0: LITERAL      1
		 5: YIELD
		 6: RESTORE
		 7: LITERAL      2
		12: YIELD
		13: RESTORE
		14: LITERAL      3
		19: YIELD
		20: RESTORE
		21: LITERAL      nil
		26: RETURN

		16
*/
static qu_code qu__frame_code (qu_frame self)
{
	QU_REQUIRE (QuENone, !qu_frame_done (self));
	return self->code;
}



/*) sub done -> Boolean
	Whether or not the Frame has finished execution.

	@xmp
		g = gen

		sub gen
			yield 1
			println (g.running)
			yield 2
			yield 3
		;;

		println (g.done)
		println (g.running)

		println ('---')
		for i in g
			println (g.running, ' ', i)
		;;
		println ('---')

		println (g.running)
		println (g.done)

	@out
		false
		false
		---
		false 1
		true
		false 2
		false 3
		---
		false
		true
*/
static qu_boolean qu__frame_done (qu_frame self)
{
	return qu_frame_done (self) ? QuTrue : QuFalse;
}



/*) sub every (what) -> Null
	@def
	sub every (what: String) -> Null
	sub every (what: Sub) -> Null

	@p
	For each yielded value apply v{what}.

	If v{what} is a String then call the method named v{what} for each item
	without any argument. Otherwise v{what} must be a static -{Sub} or a
	-{Method} that will be called with each item as the only argument.

	If the method or function throws EStop then iteration stops.
	If it throws any other exception then that exception is rethrown.

	@warn
	After a call to this method, the Frame object may be exhausted.

	In fact, since the Frame may have already stopped, the result of this
	method is not reliable unless you know what you are doing.

	@exc
	.EArgValue
	.Exception as thrown by the method/function except C{EStop}

	@xmp
		sub three
			yield 1
			yield 2
			yield 3
		;;

		three.every (@print)
		println

	@out
		123
*/
static qu_null qu__frame_every (qu_frame self, qu_any what)
{
	qu_class c = qu_classof (what);
	qu_any   x;

	if (c == QuStringClass)
	{
		for (;;)
		{
			x = qu_frame_next (self);
			if (x == 0)
				goto stop;
			x = qu_object_call0 (x, what);
			if (x == 0)
				goto stop;
		}
	}
	else if (c == QuSubClass)
	{
		qu_any theself;

		QU_ARG_CHECK (qu_sub_is_static (what));
		theself = qu_interp_self ();
		for (;;)
		{
			x = qu_frame_next (self);
			if (x == 0)
				goto stop;
			x = qu_sub_call1 (what, theself, x);
			if (x == 0)
				goto stop;
		}
	}
	else if (c == QuMethodClass)
	{
		for (;;)
		{
			x = qu_frame_next (self);
			if (x == 0)
				goto stop;
			x = qu_method_call1 (what, x);
			if (x == 0)
				goto stop;
		}
	}
	else
	{
		QU_ARG_THROW;
	}
	return QuNil;

  stop:
	QU_ENSURE (qu_thread_err_is_estop ());
	return QuNil;
}



/*) sub run (callback = nil) -> Null
	Keep calling #{next} until the Frame is exhausted.

	If a v{callback} is given then it is called for each yielded value passing
	that value as the only argument. If the v{callback} throws EStop then
	f{run} stops and if it throws any other exception then that exception is
	rethrown. The value returned by this callback is the value returned by
	this method.

	@exc
	.ERuntime  the Frame is currently running
	.ECall     not a generator
	.Exception anything thrown by v{callback} except EStop

	@xmp
		sub foo (*arg)
			yield x for x in arg
		;;

		sub bar (i)
			println (i)
			return i * 2
		;;

		g = foo (1, 2, 3)
		println (g.run (@bar))

	@out
		1
		2
		3
		6
*/
static qu_any qu__frame_run (qu_frame self, int argc, qu_any *argv)
{
	qu_any r = QuNil;
	qu_any f = 0;

	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		f = argv[0];
		if (f == QuNil)
			f = 0;
	}

	for (;;)
	{
		qu_any x = qu_frame_next (self);
		if (x == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			break;
		}
		if (f)
		{
			x = qu_object_call_x1 (f, x);
			if (x == 0)
				QU_ENSURE (qu_thread_err_is_estop ());
		}
		r = x;
	}
	return r;
}



/*) sub running -> Boolean
	Whether or not the Frame is currently running.

	@noexc

	See #{Frame.done} for an example.
*/
static qu_boolean qu__frame_running (qu_frame self)
{
	return qu_frame_running (self) ? QuTrue : QuFalse;
}



/*) sub selfof -> Any
	Return the self object associated with this Frame.

	@exc
	.ENone Frame has finished running

	@xmp
		class Foo
			sub gen
				yield 1
				yield 2
			;;
		;;

		x = Foo ()
		f = x.gen
		println (f.selfof)

	@out
		<a Foo>
*/
static qu_any qu__frame_selfof (qu_frame self)
{
	QU_REQUIRE (QuENone, !qu_frame_done (self));
	return self->self;
}



/*) sub sp -> Int
	Return the current stack pointer of this Frame.

	@noexc

	@xmp
		f = foo

		sub foo
			println (f.sp)
			yield 'foo'
			println (f.sp)
			yield 'bar'
			println (f.sp)
			yield 'baz'
		;;

		for x in f
			println (x)
		;;

	@out
		3
		foo
		3
		bar
		3
		baz
*/
static qu_int qu__frame_sp (qu_frame self)
{
	return qu_int_new (self->sp - self->ss);
}



/*) sub where -> String
	Get a string that tells you the whereabouts of this Frame.

	@exc
	.ENone Frame has finished running

	@xmp
		sub foo
			yield 1
			yield 2
			yield 3
		;;

		f = foo
		println (f)
		println (f.where)

	@out
		<a Frame>
		__main__.foo
*/
qu_string qu__frame_where (qu_frame self)
{
	QU_REQUIRE (QuENone, !qu_frame_done (self));
	return qu_object_where_string (self->code->parent, "", "");
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

qu_frame qu__frame_new (qu_code c, qu_any theself, int nlocal)
{
	register qu_frame f = qu_gc_new (QuFrameClass);

	f->self   = theself;
	f->eclass = 0;
	f->code   = c;
	f->argc   = 0;
	f->line   = 0;
	f->ip     = 0;
	f->level  = f->atomic = -1;
	f->ss     = qu_mem_alloc (c->stacksize * sizeof (qu_any));
	f->sp     = f->bp = f->ss + nlocal;
	f->sp[0]  = QuNil;
	f->result = QuNil;

	return f;
}




/* rethrow the last exception */
void qu_frame_rethrow (qu_frame self)
{
	qu_thread_throwex (__QuThreadCurrent->eclass, __QuThreadCurrent->emsg, self->code->parent, self->line);
	qu_frame_set_thrown (self);
	self->eclass = __QuThreadCurrent->eclass;
}


/* throw an exception */
void qu_frame_throw (qu_frame self, qu_class c)
{
	qu_thread_throwex (c, 0, self->code->parent, self->line);
	qu_frame_set_thrown (self);
	self->eclass = c;
}



static qu_frame _frame_setarg (qu_frame f, qu_sub u, int argc, qu_any *argv)
{
	f->argc = argc;

	if (u->nlocal)
	{
		int  n = u->nlocal;
		qu_any *p = f->ss + u->first;

		if (qu_sub_is_closure (u))
		{
			assert (u->clex != 0);
			assert (u->clex->len == n);
			QU_OBJECT_LIST_COPY (f->ss, u->clex->objs, n);
		}
		else
		{
			if (u == u->root)
				QU_OBJECT_LIST_SETVAL (f->ss, QuNil, n);
			else
			{
				assert (__QuInterpFrame);
				QU_OBJECT_LIST_COPY (f->ss, __QuInterpFrame->ss, n);
			}
		}

		if (!qu_sub_is_vararg (u))
		{
			QU_OBJECT_LIST_COPY (p, argv, argc);
		}
		else
		{
			int   m = u->narg;
			qu_array a = QuArrayEmpty;

			n = m - 1;
			if (argc < n)
			{
				QU_OBJECT_LIST_COPY (p, argv, argc);
			}
			else
			{
				QU_OBJECT_LIST_COPY (p, argv, n);
				argc -= n;
				if (argc)
				{
					argv += n;
					a = qu_array_new (argc);
					a->len = argc;
					QU_OBJECT_LIST_COPY (a->objs, argv, argc);
				}
			}
			p [m - 1] = a;
		}
	}
	return f;
}



qu_frame qu_frame_new_c (qu_frame f, qu_any *stack, qu_sub u, qu_any theself, int argc, qu_any *argv)
{
	assert (! (qu_sub_is_extern (u) || qu_sub_is_generator (u)));
	assert (f && stack && ((qu_code) u->value)->stacksize <= QU_FRAME_CSTACK_MAX);

	QU_ENSURE (qu_sub_arg_check (u, argc, argv, 1));

	f->flags  = QU_FRAME_FLAG_SUB | QU_FRAME_FLAG_CFRAME;
	f->klass  = QuFrameClass;
	f->attrs  = NULL;
	f->self   = theself;
	f->eclass = 0;
	f->code   = u->value;
	f->argc   = 0;
	f->line   = 0;
	f->ip     = 0;
	f->level  = f->atomic = -1;
	f->ss     = stack;
	f->sp     = f->bp = f->ss + u->nlocal;
	f->sp[0]  = QuNil;
	f->result = QuNil;

	return _frame_setarg (f, u, argc, argv);
}



qu_frame qu_frame_new_a (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
	qu_frame f;

	assert (! (qu_sub_is_extern (u) || qu_sub_is_generator (u)));

	QU_ENSURE (qu_sub_arg_check (u, argc, argv, 1));

	f = qu__frame_new (((qu_code) u->value), theself, u->nlocal);
	f->flags |= QU_FRAME_FLAG_SUB;
	return _frame_setarg (f, u, argc, argv);
}



qu_frame qu_frame_new_g (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
	qu_frame f;

	assert (!qu_sub_is_extern (u));
	assert (qu_sub_is_generator (u));

	QU_ENSURE (qu_sub_arg_check (u, argc, argv, 1));

	f = qu__frame_new (((qu_code) u->value), theself, u->nlocal);
	f->flags |= QU_FRAME_FLAG_SUB | QU_FRAME_FLAG_GENERATOR;

	return _frame_setarg (f, u, argc, argv);
}



qu_module qu_frame_get_module (qu_frame self)
{
	qu_any b = self->code->parent;

	switch (qu_classid (b))
	{
		case QU_CLASS_ID_Module:
			return (qu_module) b;

		case QU_CLASS_ID_Class:
			return qu_class_get_module ((qu_class) b);

		case QU_CLASS_ID_Sub:
			return qu_sub_get_module ((qu_sub) b);
	}
	QU_FATAL ("Frame without module");
	return 0; /*never*/
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_frame__setup ()
{
	qu_class_setup_sys (QuFrameClass, qu_frame_s, qu__frame__destroy, qu__frame__mark);
}



void qu_frame__init ()
{
	/* standard instance methods */
	qu_class_method_inst  (QuFrameClass, QU_NAME__as    , qu__frame__as    ,  1);
	qu_class_method_inst  (QuFrameClass, QU_NAME__call  , qu__frame__call  , -1);
	qu_class_method_inst  (QuFrameClass, QU_NAME__copy  , qu__frame__copy  ,  0);
	qu_class_alias        (QuFrameClass, QU_NAME__copy  , QU_NAME_copy);
	qu_class_method_inst  (QuFrameClass, QU_NAME__each  , qu__frame__each  ,  0);
	qu_class_method_inst  (QuFrameClass, QU_NAME__in    , qu__frame__in    ,  1);
	qu_class_method_inst  (QuFrameClass, QU_NAME__like  , qu__frame__like  ,  1);
	qu_class_method_inst  (QuFrameClass, QU_NAME__next  , qu__frame__next  ,  0);
	qu_class_aliasz       (QuFrameClass, "__next", "next");
	qu_class_method_inst  (QuFrameClass, QU_NAME__print , qu__frame__print ,  0);
	qu_class_method_inst  (QuFrameClass, QU_NAME__string, qu__frame__string,  0);
	qu_class_method_inst  (QuFrameClass, QU_NAME__stop  , qu__frame__stop  , -1);
	qu_class_aliasz       (QuFrameClass, "__stop", "stop");
	qu_class_method_inst  (QuFrameClass, QU_NAME__true  , qu__frame__true  ,  0);

	/* specific instance methods */
	qu_class_method_instz (QuFrameClass, "argc"   , qu__frame_argc   ,  0);
	qu_class_method_instz (QuFrameClass, "code"   , qu__frame_code   ,  0);
	qu_class_method_instz (QuFrameClass, "done"   , qu__frame_done   ,  0);
	qu_class_method_instz (QuFrameClass, "every"  , qu__frame_every  ,  1);
	qu_class_method_instz (QuFrameClass, "run"    , qu__frame_run    , -1);
	qu_class_method_instz (QuFrameClass, "running", qu__frame_running,  0);
	qu_class_method_instz (QuFrameClass, "selfof" , qu__frame_selfof ,  0);
	qu_class_method_instz (QuFrameClass, "sp"     , qu__frame_sp     ,  0);
	qu_class_method_instz (QuFrameClass, "where"  , qu__frame_where  ,  0);

	/* finalize declaration */
	qu_class_enddecl (QuFrameClass);
}
