/* Inf.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Inf __abstract
	Like -{Boolean} this class has two instances only, C{infp} and C{infn}
	which represent positive and negative infinity respectively.

	Like k{true} and k{false}, k{infp} and k{infn} are keywords that simply
	point to those objects.

	The class hierarchy:
	@tree
	.-{Number}
	.	-{Real}
	.		Inf

	@p
	Methods/functions must throw EInf if the global -{Number.context} does not
	allow infinities.

	@xmp
		Number.context |= 'I'
		x = 1 / 0
		println (x)

	@out
		infp
*/
qu_class QuInfClass;

qu_inf QuInfPos;
qu_inf QuInfNeg;

qu_string QuInfPosSTR;
qu_string QuInfNegSTR;



/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

#define BINOP_integer(name, rtype) QU_NUMBER_BINOP_integer (qu_inf, name, rtype)
#define BINOP_decimal(name, rtype) QU_NUMBER_BINOP_decimal (qu_inf, name, rtype)
#define BINOP_real(name, rtype)    QU_NUMBER_BINOP_real    (qu_inf, name, rtype)
#define BINOP_number(name, rtype)  QU_NUMBER_BINOP_number  (qu_inf, name, rtype)



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (x) -> Inf
	This method returns infp if C{x > 0}, infn otherwise.

	Note that this method always succeed even v{x} is not a Number and the
	global -{Number.context} does not allow infinities.

	@noexc

	@xmp
		p = Inf (1)
		println (p)

	@out
		infp
*/
static qu_inf qu__inf_new (qu_class self, qu_any x)
{
	return (qu_sys_cmp (x, QuInt0) == QuInt_1) ? QuInfNeg : QuInfPos;
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class ret (r) -> Inf
	If the context does not allow infinities then throws EInf otherwise
	returns an infinity.

	v{x} may be an infinity (infp or infn) or a number < 0 to return
	infn and >= 0 to return infp.

	@exc
	.EInf
*/
static qu_inf qu__inf_ret (qu_class self, qu_any x)
{
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	if (qu_inf_is (x))
		return (qu_inf) x;
	return (qu_sys_lt (x, QuInt0) == QuTrue) ? QuInfNeg : QuInfPos;
}


qu_inf qu_inf_return (qu_any x)
{
	return qu__inf_ret (QuInfClass, x);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following classes:
	@ops
	.| -{Inf}
		-{Number}
		-{Real}   | Return self
	.| -{Complex} | Return C{(self + 0j)}
	.| -{Boolean} | Delegate to #{__true}
	.| -{String}  | Delegate to #{__string}

	@exc
	.EArgValue

	@xmp
		println (infp as String)

	@out
		infp
*/
static qu_any qu__inf__as (qu_inf self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Inf:
		case QU_CLASS_ID_Number:
		case QU_CLASS_ID_Real:
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return self;

		case QU_CLASS_ID_Complex:
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (self == QuInfPos) ? QuComplexInfp : QuComplexInfn;

		case QU_CLASS_ID_String:
			return qu_inf_as_string (self);

		case QU_CLASS_ID_Boolean:
			return qu_inf_as_boolean (self);
	}
	QU_ARG_THROW;
}



short qu_inf_toh (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? SHORT_MIN : SHORT_MAX;
}



unsigned short qu_inf_touh (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? 0 : USHORT_MAX;
}



int qu_inf_toi (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? INT_MIN : INT_MAX;
}



unsigned int qu_inf_toui (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? 0 : UINT_MAX;
}



long qu_inf_tol (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? LONG_MIN : LONG_MAX;
}



unsigned long qu_inf_toul (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? 0 : ULONG_MAX;
}



long qu_inf_toli (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? QU_INT_MIN : QU_INT_MAX;
}



float qu_inf_tof (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? -FLT_MAX : FLT_MAX;
}


double qu_inf_tod (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? -DBL_MAX : DBL_MAX;
}


#ifdef HAVE_LONG_LONG
long long qu_inf_toll (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? LONG_LONG_MIN : LONG_LONG_MAX;
}



unsigned long long qu_inf_toull (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? 0 : ULONG_LONG_MAX;
}
#endif



#ifdef HAVE_LONG_DOUBLE
long double qu_inf_told  (qu_inf x)
{
	errno = ERANGE;
	return (x == QuInfNeg) ? -LDBL_MAX : LDBL_MAX;
}
#endif



qu_any qu__inf__toc ()
{
	QU_THROW (QuEValue);
}



qu_complex qu_inf_as_complex (qu_inf x)
{
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return (x == QuInfPos) ? QuComplexInfp : QuComplexInfn;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	Compare self to v{x}.

	See -{Number.__cmp} for more information.

	@noexc

	@xmp
		println (infp <=> 1)
		println (infn <=> 1)

	@out
		1
		-1
*/
static qu_int qu__inf__cmp (qu_inf self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_inf_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_inf_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_inf_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_inf_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_inf_cmp_big      (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_inf_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_inf_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}



int qu_inf_cmp_complex (qu_inf x, qu_complex y)
{
	return qu_complex_cmp_complex ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);
}



int qu_inf_cmp_inf (qu_inf x, qu_inf y)
{
	return (x == y) ? 0 : ((x == QuInfPos) ? 1 : -1);
}



BINOP_integer (cmp, int)
BINOP_decimal (cmp, int)
BINOP_real    (cmp, int)
BINOP_number  (cmp, int)



/*
	|<=>|
*/

int qu_inf_cmpabs_complex (qu_inf x, qu_complex y)
{
	return qu_complex_cmpabs_complex ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);
}



BINOP_integer (cmpabs, int)
BINOP_decimal (cmpabs, int)
BINOP_real    (cmpabs, int)
BINOP_number  (cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> self
	Return self or throw EInf if the global -{Number.context} does not
	allow infinities.

	See -{Number.__decr} for more information.

	@exc
	.EInf

	@xmp
		Number.context |= 'I'

		x = infp
		println (--x)

		x = infn
		println (--x)

	@out
		infp
		infn
*/
static qu_inf qu__inf__decr (qu_inf self)
{
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return self;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number)
	Divide self with v{x}.

	If v{x} is an infinity then EDiv is thrown.

	@p
	See -{Number.__div} for more information.

	@exc
	.EDiv

	@xmp
		Number.context |= 'I'

		x = infp
		println (x / 10)

		x = infn
		println (x / 3)

		try
			println (x / infn)
		catch
			println (__err__)
		;;

	@out
		infp
		infn
		(EDiv) invalid division
*/
static qu_number qu__inf__div (qu_inf self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_inf_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_inf_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_inf_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_inf_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_inf_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_inf_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_inf_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



qu_inf qu__inf__divr (qu_inf x, int sign)
{
	if (sign == 0)
		return __qu_zerodiv__ (sign);
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return __qu_inf__ (qu_inf_sign (x) * (sign | 1));
}



qu_number qu_inf_div_complex (qu_inf x, qu_complex y)
{
	return qu_complex_div_complex ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);
}



qu_any qu_inf_div_inf (qu_inf x, qu_inf y)
{
	QU_THROW (QuEDiv);
}



BINOP_integer (div, qu_inf)
BINOP_decimal (div, qu_inf)
BINOP_real    (div, qu_inf)
BINOP_number  (div, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if self equals v{x}, false otherwise.

	Note that infinities are only equal to themselves.

	See -{Number.__eq} for more information.

	@noexc

	@xmp
		println (infp == infn)
		println (infp == infp)
		println (infn == infn)

	@out
		false
		true
		true
*/
static qu_boolean qu__inf__eq (qu_inf self, qu_any x)
{
	return ((qu_any) self == x) ? QuTrue : QuFalse;
}



int qu_inf_eq_complex (qu_inf x, qu_complex y)
{
	return qu_real_iszero (y->imag) && (qu_real) x == y->real;
}



BINOP_integer (eq, int)
BINOP_decimal (eq, int)
BINOP_real    (eq, int)
BINOP_number  (eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this infinity.

	See -{Number.__format} for details.
*/
static qu_string qu__inf__format (qu_inf self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_inf_format (self, &f);
}



qu_string qu_inf_format (qu_inf x, qu_format_s *f)
{
	return qu_format_align (f, (x == QuInfPos) ? QuInfPosSTR : QuInfNegSTR);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> Int
	Return the hash value of self.

	This method returns 1 for #{infp} and -1 for #{infn}.

	See -{Number.__hash} for more information.

	@xmp
		println (infp.__hash)
		println (infn.__hash)

	@out
		1
		-1
*/
qu_int qu__inf__hash (qu_inf self)
{
	return (self == QuInfPos) ? QuInt1 : QuInt_1;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer
	Divide self with v{x} in an -{Integer} context.

	This method always throw EIDiv because dividing an infinity with anything
	in an integer context is not possible.

	See -{Number.__idiv} for more information.

	@exc
	.EIdiv

	@xmp
		p = infp
		try
			x = p \ 0
		catch
			println (__err__)
		;;

	@out
		(EIdiv) invalid integer division
*/
static qu_integer qu__inf__idiv (qu_inf self, qu_any x)
{
	QU_THROW (QuEIdiv);
}



qu_integer qu__inf__idivr ()
{
	QU_THROW (QuEIdiv);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> self
	Return self or throw EInf if the global -{Number.context} does not
	allow infinities.

	See -{Number.__incr} for more information.

	@exc
	.EInf

	@xmp
		Number.context |= 'I'

		x = infp
		println (++x)

		x = infn
		println (++x)

	@out
		infp
		infn
*/
static qu_inf qu__inf__incr (qu_inf self)
{
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return self;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x) -> Boolean
	Return true if v{x} is an infinity, false otherwise.

	See -{Number.__like} for more information.

	@noexc

	@xmp
		println (infp === infn)

	@out
		true
*/
static qu_boolean qu__inf__like (qu_inf self, qu_any x)
{
	return qu_inf_is (x) ? QuTrue : QuFalse;
}



int qu_inf_like_complex (qu_inf x, qu_complex y)
{
	return qu_real_iszero (y->imag) && qu_inf_is (y->real);
}



BINOP_integer (like, int)
BINOP_decimal (like, int)
BINOP_real    (like, int)
BINOP_number  (like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Subtract self with v{x}.

	Subtracting two infinities of the same sign throws EMinus.

	@p
	See -{Number.__minus} for more information.

	@exc
	.EMinus
	.EInf

	@xmp
		Number.context |= 'I'
		x = infp
		println (x - 1)
		println (x - infn)
		try
			println (x - x)
		catch
			println (__err__)
		;;

	@out
		infp
		infp
		(EMinus) invalid subtraction
*/
static qu_number qu__inf__minus (qu_inf self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_inf_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_inf_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_inf_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_inf_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_inf_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_inf_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_inf_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



qu_inf qu__inf__minusr (qu_inf x)
{
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return x;
}



qu_number qu_inf_minus_complex (qu_inf x, qu_complex y)
{
	return qu_complex_minus_complex ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);
}



qu_inf qu_inf_minus_inf (qu_inf x, qu_inf y)
{
	int s = qu_inf_sign (x);
	QU_REQUIRE (QuEMinus, s != qu_inf_sign (y));
	return __qu_inf__ (s);
}



BINOP_integer (minus, qu_inf)
BINOP_decimal (minus, qu_inf)
BINOP_real    (minus, qu_inf)
BINOP_number  (minus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Multiply self with v{x}.

	Multiplying an infinity with 0 throws EMul.

	@p
	See -{Number.__mul} for more information.

	@exc
	.EMul
	.EInf

	@xmp
		Number.context |= 'I'
		x = infp
		println (x * 1.23)
		println (x * infn)
		println (x * x)
		try
			println (x * 0)
		catch
			println (__err__)
		;;

	@out
		infp
		infn
		infp
		(EMul) invalid multiplication
*/
static qu_number qu__inf__mul (qu_inf self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_inf_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_inf_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_inf_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_inf_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_inf_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_inf_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_inf_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



qu_number qu__inf__mulr (qu_inf x, int sign)
{
	QU_REQUIRE (QuEMul, sign);
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return (qu_number) __qu_inf__ (sign * qu_inf_sign (x));
}



qu_inf qu_inf_mul_inf (qu_inf x, qu_inf y)
{
	return __qu_inf__ (qu_inf_sign (x) * qu_inf_sign (y));
}



BINOP_integer (mul, qu_inf)
BINOP_decimal (mul, qu_inf)
BINOP_real    (mul, qu_inf)
BINOP_number  (mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Inf
	This method returns infn if self is infp, and infp otherwise.
	It ignores the global -{Number.context}.

	See -{Number.__neg} for more information.

	@noexc

	@xmp
		println (-infp)
		println (-infn)

	@out
		infn
		infp
*/
static qu_inf qu__inf__neg (qu_inf self)
{
	return (self == QuInfPos) ? QuInfNeg : QuInfPos;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Add self with v{x}.

	Adding two infinities of different sign throws EPlus.

	See -{Number.__plus} for more information.

	@exc
	.EPlus
	.EInf

	@xmp
		Number.context |= 'I'
		x = infp
		println (x + 1)
		println (x + infp)
		try
			println (x + infn)
		catch
			println (__err__)
		;;

	@out
		infp
		infp
		(EMinus) invalid subtraction
*/
static qu_number qu__inf__plus (qu_inf self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_inf_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_inf_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_inf_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_inf_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_inf_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_inf_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_inf_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



qu_inf qu__inf__plusr (qu_inf x)
{
	QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
	return x;
}



qu_number qu_inf_plus_complex (qu_inf x, qu_complex y)
{
	return qu_complex_plus_complex ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);
}



qu_inf qu_inf_plus_inf (qu_inf x, qu_inf y)
{
	int s = qu_inf_sign (x);
	QU_REQUIRE (QuEPlus, s == qu_inf_sign (y));
	return __qu_inf__ (s);
}



BINOP_integer (plus, qu_inf)
BINOP_decimal (plus, qu_inf)
BINOP_real    (plus, qu_inf)
BINOP_number  (plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pos -> self
	This method always return self.
	It ignores the global -{Number.context}.

	See -{Number.__pos} for more information.

	@noexc

	@xmp
		println (+infp)
		println (+infn)

	@out
		infp
		infn
*/
#define qu__inf__pos qu_object_method_self




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x: Number) -> Number
	Return self raised to the power of v{x}.

	See -{Number.__pow} for more information.

	Note that raising infn to the power of non whole number is an error
	because even in the Complex domain the imaginary part is undefined.

	@exc
	.EPow
	.EInf

	@xmp
		Number.context |= 'I'
		println (infp ** 1.23)
		println (infn ** -4)
		try
			x = infn
			println (x ** 1.2)
		catch
			println (__err__)
		;;

	@out
		infp
		0
		(EPow) invalid power
*/
static qu_number qu__inf__pow (qu_inf self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_inf_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_inf_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_inf_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_inf_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_inf_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_inf_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_inf_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}



qu_real qu_inf_pow_int (qu_inf x, qu_int y)
{
	int sy = qu_int_sign (y);

	if (sy == 0)
	{
		return (qu_real) QuInt1;
	}
	else
	{
		if (sy < 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_real) ((x == QuInfPos) ? x : (qu_int_isodd (y) ? QuInfNeg: QuInfPos));
		}
	}
}



qu_real qu_inf_pow_long (qu_inf x, qu_long y)
{
	int sy = qu_long_sign (y);

	if (sy == 0)
	{
		return (qu_real) QuInt1;
	}
	else
	{
		if (sy < 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_real) ((x == QuInfPos) ? x : (qu_long_isodd (y) ? QuInfNeg: QuInfPos));
		}
	}
}



qu_number qu_inf_pow_rational (qu_inf x, qu_rational y)
{
	int sy = qu_rational_sign (y);

	if (sy == 0)
	{
		return (qu_number) QuInt1;
	}
	else
	{
		if (x == QuInfNeg && !qu_rational_iswhole (y))
			return qu_complex_pow_rational ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);

		if (sy < 0)
		{
			return (qu_number) QuInt0;
		}
		else
		{
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_number) ((x == QuInfPos) ? x : (qu_rational_isodd (y) ? QuInfNeg: QuInfPos));
		}
	}
}



qu_number qu_inf_pow_float (qu_inf x, qu_float y)
{
	int sy = qu_float_sign (y);

	if (sy == 0)
	{
		return (qu_number) QuInt1;
	}
	else
	{
		if (x == QuInfNeg && !qu_float_iswhole (y))
			return qu_complex_pow_float ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);

		if (sy < 0)
		{
			return (qu_number) QuInt0;
		}
		else
		{
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_number) ((x == QuInfPos) ? x : (qu_float_isodd (y) ? QuInfNeg: QuInfPos));
		}
	}
}



qu_number qu_inf_pow_big (qu_inf x, qu_big y)
{
	int sy = qu_big_sign (y);

	if (sy == 0)
	{
		return (qu_number) QuInt1;
	}
	else
	{
		if (x == QuInfNeg && !qu_big_iswhole (y))
			return qu_complex_pow_big ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);

		if (sy < 0)
		{
			return (qu_number) QuInt0;
		}
		else
		{
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_number) ((x == QuInfPos) ? x : (qu_big_isodd (y) ? QuInfNeg: QuInfPos));
		}
	}
}



qu_number qu_inf_pow_complex (qu_inf x, qu_complex y)
{
	return qu_complex_pow_complex ((x == QuInfPos) ? QuComplexInfp : QuComplexInfn, y);
}



qu_real qu_inf_pow_inf (qu_inf x, qu_inf y)
{
	if (y == QuInfNeg)
	{
		return (qu_real) QuInt0;
	}
	else
	{
		QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
		return (qu_real) QuInfPos;
	}
}



BINOP_integer (pow, qu_real)
BINOP_decimal (pow, qu_number)
BINOP_real    (pow, qu_number)
BINOP_number  (pow, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print -> Null
	Print self to stderr.

	See the builtin -{print} method for possible exceptions.
*/
static qu_null qu__inf__print (qu_inf self)
{
	qu_string s = (self == QuInfPos) ? QuInfPosSTR : QuInfNegSTR;
	return qu_sys_printz (s->str, s->len);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number)
	This method always throw ERem because the #{__idiv} method always fail.

	See #{__idiv}.

	@exc
	.ERem

	@xmp
		x = infp
		try
			y = x % infp
		catch
			println (__err__)
		;;

	@out
		0
		(ERem) invalid remainder
*/
static qu_number qu__inf__rem (qu_inf self, qu_any x)
{
	QU_THROW (QuERem);
}



qu_any qu_inf__remr ()
{
	QU_THROW (QuERem);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	This method returns the string "infp" if self is #{infp},
	"infn" otherwise.
*/
qu_string qu__inf__string (qu_inf self)
{
	return qu_inf_as_string (self);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	This method always return true.
*/
qu_boolean qu__inf__true (qu_inf self)
{
	return qu_inf_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub ceil
	This method always throw EMath.
*/
qu_any qu__inf_ceil (qu_inf self)
{
	QU_THROW (QuEMath);
}



/*) sub floor
	This method always throw EMath.
*/
qu_any qu__inf_floor (qu_inf self)
{
	QU_THROW (QuEMath);
}



/*) sub frac
	This method always throw EMath.
*/
qu_any qu__inf_frac (qu_inf self)
{
	QU_THROW (QuEMath);
}



/*) sub iseven -> Boolean
	This method always return false.
*/
static qu_any qu__inf_iseven (qu_inf self)
{
	return QuFalse;
}



/*) sub isfrac -> Boolean
	This method always return false.
*/
static qu_any qu__inf_isfrac (qu_inf self)
{
	return QuFalse;
}



/*) sub isodd -> Boolean
	This method always return false.
*/
static qu_any qu__inf_isodd (qu_inf self)
{
	return QuFalse;
}



/*) sub iswhole -> Boolean
	This method always return false.
*/
static qu_any qu__inf_iswhole (qu_inf self)
{
	return QuFalse;
}



/*) sub round (n: Int)
	This method always throw EMath.
*/
qu_any qu__inf_round (qu_inf self, qu_integer n)
{
	QU_THROW (QuEMath);
}



/*) sub sign -> Int
	Return -1 if the number is negative infinity, 1 otherwise.
*/
static qu_int qu__inf_sign (qu_inf self)
{
	return (self == QuInfPos) ? QuInt1 : QuInt_1;
}



/*) sub trunc
	This method always throw EMath.
*/
qu_any qu__inf_trunc (qu_inf self)
{
	QU_THROW (QuEMath);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_inf__setup ()
{
	qu_class_setup_final (QuInfClass, qu_object_s, 0, 0);

	QuInfPos = qu_gc_new_static (QuInfClass);
	QuInfNeg = qu_gc_new_static (QuInfClass);

	QuInfPosSTR = qu_string_new_staticz ("infp");
	QuInfNegSTR = qu_string_new_staticz ("infn");
}



void qu_inf__init ()
{
	/* specific class constants */
	qu_class_const (QuInfClass, QuInfPosSTR, QuInfPos);
	qu_class_const (QuInfClass, QuInfNegSTR, QuInfNeg);

	/* standard class methods */
	qu_class_method_class  (QuInfClass, QU_NAME_new, qu__inf_new,  1);

	/* specific class methods */
	qu_class_method_classz (QuInfClass, "ret", qu__inf_ret,  1);

	/* standard instance methods */
	qu_class_method_inst   (QuInfClass, QU_NAME__as    , qu__inf__as    , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__cmp   , qu__inf__cmp   , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__decr  , qu__inf__decr  , 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__div   , qu__inf__div   , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__eq    , qu__inf__eq    , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__format, qu__inf__format, 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__hash  , qu__inf__hash  , 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__idiv  , qu__inf__idiv  , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__incr  , qu__inf__incr  , 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__like  , qu__inf__like  , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__minus , qu__inf__minus , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__mul   , qu__inf__mul   , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__neg   , qu__inf__neg   , 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__plus  , qu__inf__plus  , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__pos   , qu__inf__pos   , 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__pow   , qu__inf__pow   , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__print , qu__inf__print , 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__rem   , qu__inf__rem   , 1);
	qu_class_method_inst   (QuInfClass, QU_NAME__string, qu__inf__string, 0);
	qu_class_method_inst   (QuInfClass, QU_NAME__true  , qu__inf__true  , 0);

	/* specific instance methods */
	qu_class_method_instz  (QuInfClass, "ceil"   , qu__inf_ceil   , 0);
	qu_class_method_instz  (QuInfClass, "floor"  , qu__inf_floor  , 0);
	qu_class_method_instz  (QuInfClass, "frac"   , qu__inf_frac   , 0);
	qu_class_method_instz  (QuInfClass, "iseven" , qu__inf_iseven , 0);
	qu_class_method_instz  (QuInfClass, "isfrac" , qu__inf_isfrac , 0);
	qu_class_method_instz  (QuInfClass, "isodd"  , qu__inf_isodd  , 0);
	qu_class_method_instz  (QuInfClass, "iswhole", qu__inf_iswhole, 0);
	qu_class_method_instz  (QuInfClass, "round"  , qu__inf_round  , 1);
	qu_class_method_instz  (QuInfClass, "sign"   , qu__inf_sign   , 0);
	qu_class_method_instz  (QuInfClass, "trunc"  , qu__inf_trunc  , 0);

	/* finalize declaration */
	qu_class_set_super (QuInfClass, QuRealClass);
	qu_class_enddecl   (QuInfClass);
}
