/* Int.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Int is Integer __final
	Int objects are integer numbers within the range Int.#{MIN} .. Int.#{MAX},
	contrary to -{Long} that can be as big as the available memory (well,
	sort of).

	Refer to the -{Number} class for information on Qu numerics in general.

	The class hierarchy:
	@tree
	.-{Number}
	.	-{Real}
	.		-{Integer}
	.			Int

	@p
	Int objects do not allocate memory. They are C I{long} values minus one
	bit disguised as pointers. Thus on 32-bits machines they are 31-bits long.

	Int.#{MIN} and Int.#{MAX} have the same absolute value. This is intentional.
	It reduces a lot of non-sense complications.

	@xmp
		println (Int.MIN)
		println (Int.MAX)

		a = 7654321
		b = 8765432
		c = a * b
		d = a as Float * b

		println (c ' ' c.__name)
		println (d ' ' d.__name)

	@out
		-1073741823
		1073741823
		67093430231672 Long
		6.70934e+13 Float

	@p
	Generally you should make no distinction between an Int and a Long in
	your applications. The only reason why Int exists is performance.
	Once computers are very fast and memory is very cheap then Int and Long
	will be merged into -{Integer} or perhaps -{Int}.

	Integers are "exact" entities and Qu tries to treat them as such. If you
	divide two Integers with the C{/} operator, the result is "exact". If
	there is no remainder then the result is an Integer, a Rational otherwise.

	Raising an Integer to the power of another Integer gives you a Rational
	or an Integer as well. See the #{__pow} method.

	@xmp
		x = 2
		println (x / 3)
		println (x / 3.0)
		println (x ** -3)
		println (x ** 3.0)

	@out
		(2 / 3)
		0.666667
		(1 / 8)
		8.0
*/
qu_class QuIntClass;


/*----------------------------------------------------------------------------
	specific class properties
----------------------------------------------------------------------------*/

/*) const MIN = -1073741823
	This is the minimum possible value of a Int.
	The valus is exactly -#{MAX}.

	@warn
	The actual value is set by the system.
	On a 64-bits machine, for example, it may be -9223372036854775807.
*/

/*) const MAX = 1073741823
	This is the maximum possible value of a Int.
	The valus is exactly -#{MIN}.

	@warn
	The actual value is set by the system.
	On a 64-bits machine, for example, it may be 9223372036854775807.
*/



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (*arg) -> Int
	@def
	sub class new (x) -> Int
	sub class new (x:String, base: Integer) -> Int
	sub class new (x:String, range: Range, defval: Int) -> Int
	@p
	Create a new Int object.

	The first form method simply calls C{x.__as (Int)} and makes sure that the
	returned value is an Int otherwise EImplement is thrown.

	The second form only supports base 2, 8, 10 and 16.

	The third form requires the result to be within the specified v{range}.
	Otherwise, if v{defval} is given then v{defval} is returned. Otherwise,
	ESyntax is thrown.

	@exc
	.EArgValue
	.EImplement
	.ESyntax
	.Exception  anything thrown by the __as method

	@p
	If v{x} is an infinity then EValue is thrown because it is not possible
	to represent infinity with a Int. If v{x} is a -{Number} except infinity,
	then this call always succeed.

	If v{x} is a -{String} then ESyntax is thrown unless the string correctly
	represents a Int as written in the Qu language syntax.

	If v{x} is a -{Boolean} then Int (1) is returned if it is true and
	Int (0) otherwise.

	@xmp
		i = Int ('123')
		println (i)

		class Foo
			sub __as (what)
				switch what
					case Int: return 456
				;;
				throw EImplement
			;;
		;;

		x = Foo ()
		println (Int (x))

		i = Int ('0a', 16)
		println (i)

	@out
		123
		456
		10
*/
static qu_int qu__int_new (qu_class self, int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VCOUNT (1, 3);
	x = argv[0];

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_int) x;
		case QU_CLASS_ID_Long:     return qu_int_new_long     ((qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_int_new_rational ((qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_int_new_float    ((qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_int_new_big      ((qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_int_new_complex  ((qu_complex ) x);
		case QU_CLASS_ID_Boolean:  return qu_int_new_boolean  ((qu_boolean ) x);

		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Buffer:
			if (argc >= 2)
			{
				qu_any r = argv[1];

				if (qu_integer_is (r))
				{
					char *s = ((qu_string) x)->str;
					int   base;

					QU_ARG_vint (1, base);
					switch (base)
					{
						case  2: return qu_int_scan_binary (&s, 0);
						case  8: return qu_int_scan_oct    (&s, 0);
						case 16: return qu_int_scan_hex    (&s, 0);
						case 10: return qu_int_scan        (&s, 0);
						default: QU_ARG_THROWZ ("unsupported base");
					}
				}
				else
				{
					qu_int v, def = 0;

					QU_ARG_TYPE (r, qu_range);

					if (argc > 2)
					{
						def = argv[2];
						QU_ARG_TYPE (def, qu_int);
					}

					v = qu_int_new_string (x);
					if (v == 0)
					{
						QU_REQUIRE (QuESyntax, def);
						return def;
					}

					if (qu_sys_in (r, v) != QuTrue)
					{
						QU_ARG_CHECK (def);
						return def;
					}
					return v;
				}
			}
			return qu_int_new_string (x);

		case QU_CLASS_ID_Inf:
			QU_ARG_THROWZ ("Int (Inf)");

	}
	return qu_object_as (x, QuIntClass);
}



qu_int qu_int_newul (unsigned long v)
{
	QU_REQUIREZ (QuEValue, v <= (unsigned long) QU_INT_MAX, "value out out Int range");
	return qu_int_new ((long) v);
}



qu_int qu_int_newd (double v)
{
	long x = (long) v;
	QU_REQUIREZ (QuEValue, QU_INT_MIN <= x && x <= QU_INT_MAX, "value out out Int range");
	return qu_int_new (x);
}



#ifdef HAVE_LONG_LONG
qu_int qu_int_newll (long long v)
{
	QU_REQUIREZ (QuEValue, (long long) QU_INT_MIN <= v && v <= (long long) QU_INT_MAX, "value out out Int range");
	return qu_int_new (v);
}



qu_int qu_int_newull (unsigned long long v)
{
	QU_REQUIREZ (QuEValue, v <= (long long) QU_INT_MAX, "value out out Int range");
	return qu_int_new (v);
}
#endif /*HAVE_LONG_LONG*/



#ifdef HAVE_LONG_DOUBLE
qu_int qu_int_newld  (long double v)
{
	QU_REQUIREZ (QuEValue, (long double) QU_INT_MIN <= v && v <= (long double) QU_INT_MAX, "value out out Int range");
	return qu_int_new (v);
}
#endif /*HAVE_LONG_DOUBLE*/



qu_int qu_int_new_string (qu_string s)
{
	char *z = s->str;
	qu_int   x = qu_int_scan (&z, 0);

	QU_REQUIREZ (QuESyntax, x, "Int (String) invalid syntax");
	return x;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	scan
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_int _int_scan (char **s, int radix, const unsigned char *charset, int exp, int neg)
{

/* FIXME - this is ugly */
#define _wide_ double
#ifdef HAVE_LONG_LONG
#  if SIZEOF_LONG_LONG >= (SIZEOF_LONG + SIZEOF_LONG)
#    undef _wide_
#    define _wide_ long long
#  endif
#endif

	char  *c = *s;
	long   v = 0;
	int    k = 0;
	int    z = 0;
	int    d;
	_wide_ r;

	if (!neg)
	{
		while (qu_ascii_isspace (*c))
			c++;
		if (*c == '-')
		{
			neg = 1;
			c++;
		}
		while (qu_ascii_isspace (*c))
			c++;
	}
	if (*c == '\0')
		return 0;

	while (*c)
	{
		switch (*c)
		{
			case '_':
				c++;
				continue;

			default:
				d = *c;
				if (charset [d] == 0)
					break;
				c++;
				if (d == '0' && z == 0)
				{
					k = 1;
					continue;
				}
				z = 1;
				k = 1;
				if ('0' <= d && d <= '9')
					d -= 48;
				else
					d = qu_ascii_toupper (d) - 55;
				r = (_wide_) v * radix + d;
				if (neg)
				{
					if (-r < (_wide_) QU_INT_MAX)
						goto nope;
				}
				else
				{
					if (r > (_wide_) QU_INT_MAX)
						goto nope;
				}
				v = v * radix + d;
				continue;

			  nope:
				while (charset [(int) *c])
					c++;

				errno = ERANGE;
				switch (*c)
				{
					case 'j':
						goto done;

					case 'e':
					case 'E':
						if (exp)
							errno = EDOM;
						break;

					case '.':
						if (exp && (c[1] == 'e' || c[1] == 'E' || qu_ascii_isdigit (c[1])))
							errno = EDOM;
						break;
				}
				return 0;
		}
		break;
	}

	if (k == 0)
		return 0;

	switch (*c)
	{
		case 'j':
			goto done;

		case 'e':
		case 'E':
			if (exp)
			{
				errno = EDOM;
				return 0;
			}
			break;

		case '.':
			if (exp && (c[1] == 'e' || c[1] == 'E' || qu_ascii_isdigit (c[1])))
			{
				errno = EDOM;
				return 0;
		}
	}

  done:

	*s = c;
	if (neg)
		v = -v;
	return qu_int_new (v);

#undef _wide_

}



qu_int qu_int_scan_binary (char **s, int neg)
{
	return _int_scan (s, 2, QU_ASCII_ISBDIGIT, 0, neg);
}



qu_int qu_int_scan_oct (char **s, int neg)
{
	return _int_scan (s, 8, QU_ASCII_ISODIGIT, 0, neg);
}




qu_int qu_int_scan_hex (char **s, int neg)
{
	return _int_scan (s, 16, QU_ASCII_ISXDIGIT, 0, neg);
}



qu_int qu_int_scan (char **s, int neg)
{
	return _int_scan (s, 10, QU_ASCII_ISDIGIT, 1, neg);
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class unpack (s: String, ofs = 0) -> Int
	Unpack an Int from a String optionally start at a specified v{offset} in
	the String.

	@exc
	.EIndex
	.EArgValue
	.EImplement  your Int native size does not support packing

	@xmp
		i = 123
		s = i.pack
		println (s.len)
		i = Int.unpack (s)
		println (i)

	@out
		8
		123
*/
static qu_int qu__int_unpack (qu_class self, int argc, qu_any *argv)
{
	qu_string  s;
	long       ofs = 0;
	long      *p, v;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, s, qu_string);

	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		QU_ENSURE_INDEX (ofs, s->len);
	}

	p = (long *) (s->str + ofs);

#if SIZEOF_LONG == 8

	QU_ARG_CHECK ((s->len - ofs) >= 8);
	v = *p;

#elif SIZEOF_LONG == 4

	QU_ARG_CHECK ((s->len - ofs) >= 4 && p[1] == 0);
	v = *p;

#else

	QU_THROW (QuEImplement);

#endif

	return qu_int_new (v);
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

#ifdef QU_DEBUG_LOCAL
static void qu__int__destroy (qu_int self)
{
	QU_FATAL ("never");
}
#endif



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_int_sign (qu_int x)
{
	long v = qu_int_tol (x);
	return (v > 0) ? 1 : (v ? -1 : 0);
}



int qu_int_isone (qu_int x)
{
	return (x == QuInt1) || (x == QuInt_1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x : Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following classes:
	@ops
	.| -{Int}
		-{Integer}
		-{Real}
		-{Number}   | Return self.
	.| -{Float}
		-{Decimal}  | Convert to Float.
	.| -{Long}      | Convert to Long.
	.| -{Rational}  | Return C{self _/ 1}.
	.| -{Big}       | Convert to Big.
	.| -{Complex}   | Return C{(self + 0j)}.
	.| -{Boolean}   | Delegate to #{__true}.
	.| -{String}    | Delegate to #{__string}.

	@p
	Note that for all of the above classes, this method always succeed.

	@exc
	.EArgValue

	@xmp
		x = 2
		println (x as Rational)

	@out
		(2 _/ 1)
*/
static qu_any qu__int__as (qu_int self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Integer:
		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Number:   return self;
		case QU_CLASS_ID_Long:     return qu_int_as_long (self);
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Decimal:  return qu_int_as_float    (self);
		case QU_CLASS_ID_Big:      return qu_int_as_big      (self);
		case QU_CLASS_ID_Rational: return qu_int_as_rational (self);
		case QU_CLASS_ID_Complex:  return qu_int_as_complex  (self);
		case QU_CLASS_ID_Boolean:  return qu_int_as_boolean  (self);
		case QU_CLASS_ID_String:   return qu_int_as_string   (self);
		case QU_CLASS_ID_Range:    return qu_int_as_range    (self);
	}
	QU_ARG_THROW;
}



#ifndef QU_RSHIFT_PRESERVES_SIGNBIT
long qu__int_tol (long i)
{
	return (i < 0) ? ~(~i >> 1) : (i >> 1);
}
#endif



unsigned long qu_int_toul (qu_int x)
{
	long v = qu_int_tol (x);
	if (v >= 0)
	{
		errno = 0;
		return (unsigned long) v;
	}
	errno = ERANGE;
	return ULONG_MIN;
}



qu_range qu_int_as_range (qu_int x)
{
	long v = qu_int_tol (x);
	return (v < 0) ? qu_range_new (v, 0, 1) : qu_range_new (0, v, 1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	Compare self with another object.

	If v{x} is a Number then their numerical values are compared.
	Refer to -{Number.__cmp} for details on how comparisson is made.

	The returned value is -1 if self is smaller than v{x}, 0 if self equals
	v{x}, 1 otherwise.

	@noexc

	@xmp
		println (1 <=> 0)
		println (1 <=> 1)
		println (1 <=> 2)

		class Foo
			var value = 10
			sub __cmp (x)
				switch x.__class
					case Int: return self.value <=> x
				;;
			;;
		;;

		x = Foo ()
		println (x <=> 20)

	@out
		1
		0
		-1
		-1
*/
static qu_int qu__int__cmp (qu_int self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_int_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_int_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_int_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_int_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_int_cmp_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_int_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_int_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}



int qu_int_cmp_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);
	return QU_NUMBER_COMPARE_VALUE (a, b);
}



QU_NUMBER_BINOP_integer (qu_int, cmp, int)
QU_NUMBER_BINOP_decimal (qu_int, cmp, int)
QU_NUMBER_BINOP_real    (qu_int, cmp, int)
QU_NUMBER_BINOP_number  (qu_int, cmp, int)



/*
	|<=>|
*/

int qu_int_cmpabs_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);
	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	return QU_NUMBER_COMPARE_VALUE (a, b);
}



QU_NUMBER_BINOP_integer (qu_int, cmpabs, int)
QU_NUMBER_BINOP_decimal (qu_int, cmpabs, int)
QU_NUMBER_BINOP_real    (qu_int, cmpabs, int)
QU_NUMBER_BINOP_number  (qu_int, cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Integer
	Return self decremented by 1.

	Note that this method always succeed. If the result fits in an Int then
	an Int is returned, a -{Long} otherwise.

	@noexc

	@xmp
		a = 10
		b = --a
		println (b)

	@out
		9
*/
qu_integer qu__int__decr (qu_int self)
{
	long v = qu_int_tol (self) - 1;
	if (QU_INT_MIN <= v && v <= QU_INT_MAX)
		return (qu_integer) qu_int_new (v);
	return (qu_integer) qu_long_newl (v);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number
	Return the result of dividing self with v{x}.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{\} operator).

	Refer to -{Number.__div} for details on how the division is performed.

	@p
	When you divide an -{Integer} with another Integer, Qu delivers an "exact"
	result. If the division has a remainder, then a -{Rational} number is
	returned, an -{Integer} otherwise.

	@exc
	.EDiv
	.EInf
	.EHuge
	.ETiny

	@xmp
		println (9 / 3)
		println (7 / 3)
		println (7 \ 3)
		println (7.0 / 3)

	@out
		3
		(7 _/ 3)
		2
		2.33333
*/
static qu_number qu__int__div (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_int_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_int_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_int_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_int_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_int_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_int_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



qu_real qu_int_div_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	if (b == 0)
	{
		QU_REQUIRE (QuEDiv, a);
		return (qu_real) __qu_zerodiv__ (qu_int_sign (x));
	}

	if ((a % b) == 0)
		return (qu_real) qu_int_new (a / b);

	return (qu_real) qu_rational_newr ((qu_integer) x, (qu_integer) y);
}



qu_int qu_int_div_inf (qu_int x, qu_inf y)
{
	QU_REQUIRE (QuEDiv, !qu_int_iszero (x));
	return QuInt0;
}



QU_NUMBER_BINOP_integer (qu_int, div, qu_real)
QU_NUMBER_BINOP_decimal (qu_int, div, qu_real)
QU_NUMBER_BINOP_real    (qu_int, div, qu_real)
QU_NUMBER_BINOP_number  (qu_int, div, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if v{x} is a -{Number} and has the same numerical value
	as self, false otherwise.

	Refer to -{Number.__eq} for details on how the equality test is performed.

	@noexc

	@xmp
		println (1 == 1.0)

	@out
		true
*/
static qu_boolean qu__int__eq (qu_int self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_int_eq_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_int_eq_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_int_eq_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_int_eq_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_int_eq_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_int_eq_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



QU_NUMBER_BINOP_integer (qu_int, eq, int)
QU_NUMBER_BINOP_decimal (qu_int, eq, int)
QU_NUMBER_BINOP_real    (qu_int, eq, int)
QU_NUMBER_BINOP_number  (qu_int, eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this Int.

	See -{Number.__format} for details on the format.

	@exc
	.EFormat
	.EMemory

	@xmp
		i = 12345
		println (i.__format ('%.3d'))
		printfln ('i = %.2,3m', i)
		printfln ('i = %10.3e', i)

	@out
		12345.000
		i = 12,345.00
		i =   1.234e+4
*/
static qu_string qu__int__format (qu_int self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_int_format (self, &f);
}



qu_string qu_int_format (qu_int x, qu_format_s *f)
{
	char  b [128];
	long  v;
	char *z;
	int   base = 10;
	char *digs = QU_ASCII_DIGITL;

	switch (f->mode)
	{
		case 'x': base = 16; if (f->ucase) digs = QU_ASCII_DIGITU; break;
		case 'o': base =  8; break;
		case 'b': base =  2; break;
		case 'y': return qu_format_yesno (f, x);
		case 'p': return qu_format_ptr (f, x);
	}
	v = qu_int_tol (x);
	if (v < 0)
		v = -v;
	z = b + sizeof (b);
	*--z = '\0';
	do
	{
		*--z = digs[v % base];
		v /= base;

	} while (v);
	return qu_format_number (f, qu_string_newz (z), qu_int_isneg (x), 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> self
	Return the hash value of this Int.

	The returned value is the same as the Int value.

	@noexc

	@xmp
		i = 1234
		println (i.__hash)

	@out
		1234
*/
#define qu__int__hash qu_object_method_self



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	&
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __iand (x: Integer) -> Integer
	Return the result of the bitwise-and operation on self with v{x}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIand

	@xmp
		a = -7
		b = ~a
		println (b)
		println (a & b)

	@out
		-1073741816
		0
*/
static qu_integer qu__int__iand (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:  return (qu_integer) qu_int_iand_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long: return (qu_integer) qu_int_iand_long (self, (qu_long) x);
	}
	QU_THROW (QuEIand);
}



qu_int qu_int_iand_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	if (a < 0)
	{
		if (b < 0)
			a = -a & -b;
		else
			a = -(-a & b);
	}
	else
	{
		if (b < 0)
			a = -(a & -b);
		else
			a &= b;
	}
	return qu_int_new (a);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer
	Divide self with v{x} in an I{integer context}.

	@warn
	There is a big difference between the #{__div} method (the C{/} operator)
	and the __idiv method (the C{\} operator).

	Refer to -{Number.__idiv} for details on how the division is performed.

	@exc
	.EIdiv

	@xmp
		a = 13
		b = 4
		println (a \ b)

	@out
		3
*/
static qu_integer qu__int__idiv (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_integer) qu_int_idiv_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_integer) qu_int_idiv_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_integer) qu_int_idiv_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_integer) qu_int_idiv_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_integer) qu_int_idiv_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_integer) qu_int_idiv_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_integer) qu_int_idiv_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEIdiv);
}



qu_int qu_int_idiv_int (qu_int x, qu_int y)
{
	QU_REQUIRE (QuEIdiv, !qu_int_iszero (y));
	return qu_int_new (qu_int_tol (x) / qu_int_tol (y));
}



qu_int qu_int_idiv_inf (qu_int x, qu_inf y)
{
	QU_REQUIRE (QuEIdiv, !qu_int_iszero (x));
	return QuInt0;
}



QU_NUMBER_BINOP_integer (qu_int, idiv, qu_integer)
QU_NUMBER_BINOP_decimal (qu_int, idiv, qu_integer)
QU_NUMBER_BINOP_real    (qu_int, idiv, qu_integer)
QU_NUMBER_BINOP_number  (qu_int, idiv, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Integer
	Return self incremented by 1.

	Note that this method always succeed. If the result fits in an Int then
	an Int is returned, a -{Long} otherwise.

	@noexc

	@xmp
		i = 10
		println (i.incr)

	@out
		11
*/
qu_integer qu__int__incr (qu_int self)
{
	long v = qu_int_tol (self) + 1;
	if (QU_INT_MIN <= v && v <= QU_INT_MAX)
		return (qu_integer) qu_int_new (v);
	return (qu_integer) qu_long_newl (v);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	~x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __inot -> Int
	Return the bitwise inversion of self.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@noexc

	@xmp
		i = 5
		j = ~i
		println (i)
		println (j)

		i = -5
		j = ~i
		println (i)
		println (j)

	@out
		5
		1073741818
		-5
		-1073741818
*/
qu_int qu__int__inot (qu_int self)
{
	long a = qu_int_tol (self);
	int  g = (a < 0);

	if (g)
		a = -a;
	a = ~a;
	a = (unsigned long) a & ((unsigned long) -1 >> 2);
	if (g)
		a = -a;
	return qu_int_new (a);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	|
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __ior (x: Integer) -> Integer
	Return the bitwise-or of self and v{x}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIor

	@xmp
		println (2 | -4)

	@out
		-6
*/
static qu_integer qu__int__ior (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:   return (qu_integer) qu_int_ior_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long:  return (qu_integer) qu_int_ior_long (self, (qu_long) x);
	}
	QU_THROW (QuEIor);
}



qu_int qu_int_ior_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	if (a < 0)
	{
		if (b < 0)
			a = -a | -b;
		else
			a = -(-a | b);
	}
	else
	{
		if (b < 0)
			a = -(a | -b);
		else
			a |= b;
	}
	return qu_int_new (a);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	^
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __ixor (x: Integer) -> Integer
	Return the bitwise-xor of self and v{x}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIxor

	@xmp
		println (1 ^ -3)

	@out
		-2
*/
static qu_integer qu__int__ixor (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:   return (qu_integer) qu_int_ixor_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long:  return (qu_integer) qu_int_ixor_long (self, (qu_long) x);
	}
	QU_THROW (QuEIxor);
}



qu_int qu_int_ixor_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	if (a < 0)
	{
		if (b < 0)
			a = -a ^ -b;
		else
			a = -(-a ^ b);
	}
	else
	{
		if (b < 0)
			a = -(a ^ -b);
		else
			a ^= b;
	}
	return qu_int_new (a);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x) -> Boolean
	If v{x} is a -{Number} and it has the same absolute numerical
	value as self then true is returned, false otherwise.

	Refer to -{Number.__like} for for details on how the comparisson is performed.

	@noexc

	@xmp
		println (-1 === 1.0)

	@out
		true
*/
static qu_boolean qu__int__like (qu_int self, qu_any x)
{
	register int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_int_like_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_int_like_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_int_like_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_int_like_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_int_like_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_int_like_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



QU_NUMBER_BINOP_integer (qu_int, like, int)
QU_NUMBER_BINOP_decimal (qu_int, like, int)
QU_NUMBER_BINOP_real    (qu_int, like, int)
QU_NUMBER_BINOP_number  (qu_int, like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<<
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __lshift (x: Integer) -> Integer
	Return the result of left-shifting the bits of self by the amount of
	v{x}.

	v{x} must be a positive -{Integer} otherwise ELshift is thrown.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.ELshift
	.EHuge

	@xmp
		println (-1 << 3)

	@out
		-8
*/
static qu_real qu__int__lshift (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:   return (qu_real) qu_int_lshift_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long:  return (qu_real) qu_int_lshift_long (self, (qu_long) x);
	}
	QU_THROW (QuELshift);
}



qu_real qu_int_lshift_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	QU_REQUIRE (QuELshift, b >= 0);

	if (a == 0 || b == 0)
	{
		return (qu_real) x;
	}
	else
	{
		if (b < 8 * SIZEOF_LONG)
		{
			int           g = (a < 0);
			unsigned long v = g ? -a : a;
			unsigned long r = v << b;

			if (((r >> b) == v) && (r < QU_INT_MAX))
				return (qu_real) qu_int_new (g ? -((long) r) : (long) r);
		}

		return qu_long_rlshift_int (qu_int_as_long (x), y);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Return the result of subtracting self with v{x}.

	Note that subtracting an Int with an Int always succeed.
	The result is either an Int or a Long.

	Refer to -{Number.__minus} for for details on how the subtraction is performed.

	@exc
	.EMinus
	.EHuge
	.EInf

	@xmp
		println (1 - 3)

	@out
		-2
*/
static qu_number qu__int__minus (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_int_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_int_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_int_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_int_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_int_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_int_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



qu_integer qu_int_minus_int (qu_int x, qu_int y)
{
	long r = qu_int_tol (x) - qu_int_tol (y);
	if (QU_INT_MIN <= r && r <= QU_INT_MAX)
		return (qu_integer) qu_int_new (r);
	return (qu_integer) qu_long_newl (r);
}



QU_NUMBER_BINOP_integer (qu_int, minus, qu_real)
QU_NUMBER_BINOP_decimal (qu_int, minus, qu_real)
QU_NUMBER_BINOP_real    (qu_int, minus, qu_real)
QU_NUMBER_BINOP_number  (qu_int, minus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Return the result of multiplying self with v{x}.

	Note that multiplying an Int with an Int always succeed.
	The result is either an Int or a Long.

	Refer to -{Number.__mul} for for details on how the multiplication is performed.

	@exc
	.EMul
	.EInf
	.EHuge
	.ETiny

	@xmp
		println (8 * 3)

	@out
		24
*/
static qu_number qu__int__mul (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_int_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_int_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_int_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_int_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_int_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_int_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



qu_integer qu_int_mul_int (qu_int x, qu_int y)
{
#ifdef HAVE_LONG_LONG
	long long r = (long long) qu_int_tol (x) * qu_int_tol (y);

	if ((long long) QU_INT_MIN <= r && r <= (long long) QU_INT_MAX)
		return (qu_integer) qu_int_newl ((long) r);

	return (qu_integer) qu_long_newll (r);
#else
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);
	long r = a * b;

	if (a < 0)
	{
		if (b < 0)
		{
			if (-b >= QU_INT_MAX / -a)
				goto aslong;
		}
		else
		{
			if (b >= QU_INT_MAX / -a)
				goto aslong;
		}
	}
	else if (a > 0)
	{
		if (b < 0)
		{
			if (-b >= -QU_INT_MAX / a)
				goto aslong;
		}
		else
		{
			if (b >= QU_INT_MAX / a)
				goto aslong;
		}
	}
	return (qu_integer) qu_int_new (r);

  aslong:

	return (qu_integer) qu_long_mul_int (qu_int_as_long (x), y);
#endif
}



QU_NUMBER_BINOP_integer (qu_int, mul, qu_real)
QU_NUMBER_BINOP_decimal (qu_int, mul, qu_real)
QU_NUMBER_BINOP_real    (qu_int, mul, qu_real)
QU_NUMBER_BINOP_number  (qu_int, mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Int
	Return the negation of self.

	@noexc

	@xmp
		i = 10
		println (-i)

	@out
		-10
*/
qu_int qu__int__neg (qu_int self)
{
	return qu_int_new (-qu_int_tol (self));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Return the result of adding v{x} to self.

	Note that adding an Int with an Int always succeed.
	The result is either an Int or a Long.

	Refer to -{Number.__plus} for for details on how the addition is performed.

	@exc
	.EPlus
	.EHuge
	.EInf

	@xmp
		println (8 + 3)

	@out
		11
*/
static qu_number qu__int__plus (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_int_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_int_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_int_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_int_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_int_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_int_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPlus);
}



qu_integer qu_int_plus_int (qu_int x, qu_int y)
{
	long r = qu_int_tol (x) + qu_int_tol (y);
	if (QU_INT_MIN <= r && r <= QU_INT_MAX)
		return (qu_integer) qu_int_new (r);
	return (qu_integer) qu_long_newl (r);
}



QU_NUMBER_BINOP_integer (qu_int, plus, qu_real)
QU_NUMBER_BINOP_decimal (qu_int, plus, qu_real)
QU_NUMBER_BINOP_real    (qu_int, plus, qu_real)
QU_NUMBER_BINOP_number  (qu_int, plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub pos -> self
	This method always return self.
*/
#define qu__int__pos  qu_object_method_self



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x: Number) -> Number
	Return the result of self raised to the power of v{x}.

	Note that if v{x} is an -{Integer} then the result is "exact", that is,
	an Integer. If it is negative then a -{Rational} number is returned.

	Refer to -{Number.__pow} for for details on how the operation is performed.

	@noexc

	@xmp
		a = 5
		println (a ** 3)
		println (a ** -3)

	@out
		125
		(1 _/ 125)
*/
static qu_number qu__int__pow (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_int_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_int_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_int_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_int_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_int_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_int_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}



static qu_real _int_powl (long v, long p)
{
	qu_real w, r;

	assert (p > 1);

	w = r = (qu_real) qu_int_new (v);

	for (p--; p; p--)
	{
		while ((p & 1) == 0)
		{
			p >>= 1;
			w = qu_real_mul_real (w, w);
			QU_ENSURE (w);
			if (qu_inf_is (w))
				return w;
		}
		r = qu_real_mul_real (w, r);
		QU_ENSURE (r);
		if (qu_inf_is (r))
			return r;
	}
	return r;
}



qu_real qu_int_pow_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	if (a == 0)
	{
		/* note: Math 0 ** 0 -> undefined */
		if (b == 0)
			return (qu_real) QuInt1;

		if (b > 0)
			return (qu_real) QuInt0;

		return (qu_real) __qu_inf__ (1);
	}
	else if (b == 0)
	{
		return (qu_real) QuInt1;
	}
	else if (b == 1)
	{
		return (qu_real) x;
	}
	else if (b == -1)
	{
		return (qu_real) qu_rational_newr ((qu_integer) QuInt1, (qu_integer) x);
	}
	else if (b < 0)
	{
		if (a == 1)
			return (qu_real) QuInt1;

		if (a == -1)
			return (b & 1) ? (qu_real) QuInt_1 : (qu_real) QuInt1;

		if (b >= -24)
		{
			qu_real r = _int_powl (a, -b);

			if (r == 0 || qu_inf_is (r))
				return (qu_real) QuInt0;

			assert (qu_integer_is (r));

			return (qu_real) qu_rational_new_ex ((qu_integer) QuInt1, (qu_integer) r);
		}
		return qu_long_pow_int (qu_int_as_long (x), y);
	}
	else
	{
		if (a == 1)
			return (qu_real) QuInt1;

		if (a == -1)
			return (b & 1) ? (qu_real) QuInt_1 : (qu_real) QuInt1;

		if (b <= 24)
		{
			qu_real r = _int_powl (a, b);
			QU_ENSURE (r);
			return ((a < 0) && (b & 1)) ? qu_real_neg (r) : r;
		}
		return qu_long_pow_int (qu_int_as_long (x), y);
	}
}



qu_real qu_int_pow_inf (qu_int x, qu_inf y)
{
	if (y == QuInfPos)
	{
		if (qu_int_iszero (x))
			return (qu_real) x;

		else if (qu_int_isone (x))
			return (qu_real) QuInt1;

		else
		{
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_real) y;
		}
	}
	else
	{
		if (qu_int_iszero (x))
			return (qu_real) __qu_inf__ (1);

		else if (qu_int_isone (x))
			return (qu_real) QuInt1;

		else
			return (qu_real) QuInt0;
	}
}



QU_NUMBER_BINOP_integer (qu_int, pow, qu_number)
QU_NUMBER_BINOP_decimal (qu_int, pow, qu_number)
QU_NUMBER_BINOP_real    (qu_int, pow, qu_number)
QU_NUMBER_BINOP_number  (qu_int, pow, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print
	Print self to -{Stream.stdout}.

	This method simply prints the string returned by the #{__string} method.

	See -{print} for possible exceptions.

	@xmp
		a = 123
		println (a)

	@out
		123
*/
static qu_null qu__int__print (qu_int self)
{
	char buf [64];
	sprintf (buf, "%ld", qu_int_tol (self));
	return qu_sys_printz (buf, strlen (buf));
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The meaning of "remainder" may differ with other languages.
	Refer to -{Number.__rem} for details.

	@p
	The remainder of dividing by zero is undefined thus throws ERem.

	@exc
	.ERem

	@xmp
		println (10 % 6)
		println (10 % -6)
		println (-10 % 6)
		println (-10 % -6)

	@out
		4
		4
		-4
		-4
*/
static qu_number qu__int__rem (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_rem_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_int_rem_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_int_rem_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_int_rem_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_int_rem_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_int_rem_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_int_rem_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuERem);
}



qu_int qu_int_rem_int (qu_int x, qu_int y)
{
	if (qu_int_iszero (y))
		QU_THROW (QuERem);

	/* CHECKME - C does it the way we want it */
	return qu_int_new (qu_int_tol (x) % qu_int_tol (y));
}



qu_int qu_int_rem_inf (qu_int x, qu_inf y)
{
	QU_REQUIRE (QuERem, !qu_int_iszero (x));
	return x;
}



QU_NUMBER_BINOP_integer (qu_int, rem, qu_integer)
QU_NUMBER_BINOP_decimal (qu_int, rem, qu_real)
QU_NUMBER_BINOP_real    (qu_int, rem, qu_real)
QU_NUMBER_BINOP_number  (qu_int, rem, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	>>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rshift (x: Integer) -> Int
	Return the result of right-shifting the bits of self by the amount of
	v{x}. Thus the result is always an Int and never a -{Long}.

	v{x} must be a positive -{Integer} otherwise ELshift is thrown.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.ERshift

	@xmp
		println (-8 >> 3)

	@out
		-1
*/
static qu_int qu__int__rshift (qu_int self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:   return qu_int_rshift_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long:  return qu_int_rshift_long (self, (qu_long) x);
	}
	QU_THROW (QuERshift);
}



qu_int qu_int_rshift_int (qu_int x, qu_int y)
{
	long a = qu_int_tol (x);
	long b = qu_int_tol (y);

	QU_REQUIRE (QuERshift, b >= 0);

#ifdef QU_RSHIFT_PRESERVES_SIGNBIT
	return qu_int_new (a >> b);
#else
	if (a >= 0)
		return qu_int_new (a >> b);
	return qu_int_new (-(-a >> b));
#endif
}



qu_int qu_int_rshift_long (qu_int x, qu_long y)
{
	QU_REQUIRE (QuERshift, !qu_long_isneg (y));
	{
		long n = qu_long_tol (y);

		if (errno || n >= (SIZEOF_LONG * 8) - 1)
			return QuInt0;
		return qu_int_new (qu_int_tol (x) >> n);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Return a string representation of self.

	@noexc

	@xmp
		s = 123 as String
		println (s)

	@out
		123
*/
qu_string qu__int__string (qu_int self)
{
	char buf [128];
	sprintf (buf, "%ld", qu_int_tol (self));
	return qu_string_newz (buf);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	This method returns false if self is C{0}, true otherwise.

	@noexc
*/
static qu_boolean qu__int__true (qu_int self)
{
	return (self != QuInt0) ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub abs -> Int
	Return the absolute value.
*/
static qu_int qu__int_abs (qu_int self)
{
	long v = qu_int_tol (self);
	if (v < 0)
	{
		v = -v;
		return qu_int_new (self);
	}
	return self;
}



/*) sub bin -> String
	Return a binary representation.

	@xmp
		x = 17
		println (x.bin)

	@out
		0b10001
*/
qu_string qu__int_bin (qu_int x)
{
	char buf [128], *p;
	long v = qu_int_tol (x);
	int  neg = v < 0;

	if (neg)
		v = -v;
	p = buf + sizeof (buf) - 1;
	*p-- = '\0';
	for (;;)
	{
		*p = '0' + (v & 1);
		v >>= 1;
		if (v == 0)
			break;
		p--;
	}
	*--p = 'b';
	*--p = '0';
	if (neg)
		*--p = '-';
	return qu_string_newz (p);
}



/*) sub ceil -> self
	This method always return self.
*/
#define qu__int_ceil qu_object_method_self



/*) sub den -> Int
	Return the denominator.

	This method always return 1.
*/
static qu_int qu__int_den (qu_int self)
{
	return QuInt1;
}



/*) sub digitsum -> Int
	Return the sum of the digits.

	@xmp
		x = 1234
		println (x.digitsum)

	@out
		10
*/
qu_int qu__int_digitsum (qu_int x)
{
	long v = qu_int_tol (x);
	long s = 0;

	while (v)
	{
		s += v % 10;
		v /= 10;
	}
	assert (0 <= s && s <= QU_INT_MAX);
	return qu_int_newl (s);
}



/*) sub floor -> self
	This method always return self.
*/
#define qu__int_floor qu_object_method_self



/*) sub frac -> Int
	This method always return 0.
*/
static qu_int qu__int_frac (qu_int self)
{
	return QuInt0;
}



/*) sub hex -> String
	Return a hexadecimal representation.
*/
qu_string qu__int_hex (qu_int x)
{
	char buf [64];
	long v = qu_int_tol (x);

	if (v < 0)
		sprintf (buf, "-0x%lX", -v);
	else
		sprintf (buf, "0x%lX", v);
	return qu_string_newz (buf);
}



/*) sub iseven -> Boolean
	Return true if the number is even, false otherwise.
*/
static qu_boolean qu__int_iseven (qu_int self)
{
	return qu_int_isodd (self) ? QuFalse : QuTrue;
}


/*) sub isfrac -> Boolean
	This method always return false.
*/
#define qu__int_isfrac qu_object_method_false



/*) sub isodd -> Boolean
	Return true if self is an odd number, false otherwise.

	Note that 0 is considered an even number.
*/
static qu_boolean qu__int_isodd (qu_int self)
{
	return qu_int_isodd (self) ? QuTrue : QuFalse;
}



/*) sub iswhole -> Boolean
	This method always return true.
*/
#define qu__int_iswhole qu_object_method_true



/*) sub num -> self
	Return the numerator.

	This method always return self.
*/
static qu_int qu__int_num (qu_int self)
{
	return self;
}



/*) sub oct -> String
	Return an octal representation.
*/
qu_string qu__int_oct (qu_int x)
{
	char buf [64];
	long v = qu_int_tol (x);

	if (v < 0)
		sprintf (buf, "-0%lo", -v);
	else
		sprintf (buf, "0%lo", v);
	return qu_string_newz (buf);
}



/*) sub strlen (base = 10, sign = true, prefix = false) -> Int
	The length of the Int as String.

	If v{sign} is false then don't include the sign.

	v{base} must be one of: C{2}, C{8}, C{16} or C{10}.

	If v{prefix} is k{true} then include the base prefix.

	@xmp
		x = -100
		println (x.strlen)

	@out
		4
*/
static qu_int qu__int_strlen (qu_int self, int argc, qu_any *argv)
{
	int  base = 10, sign = 1, prefix = 0;
	long r = 0;
	long v;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vint (0, base);
		QU_ARG_CHECK (base == 10 || base == 16 || base == 8 || base == 2);

		if (argc > 1)
		{
			QU_ARG_vbool (0, sign);
			if (argc > 2)
				QU_ARG_vbool (2, prefix);
		}
	}

	v = qu_int_tol (self);
	if (v < 0)
	{
		if (sign)
			++r;
		v = -v;
	}

	do
	{
		r++;
		v /= base;

	} while (v);

	if (prefix)
	{
		switch (base)
		{
			case  2: r += 2; break;
			case  8: r += 1; break;
			case 16: r += 2; break;
		}
	}

	return qu_int_newl (r);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub pack -> String
	Pack this Int into a String.

	See the #{unpack} method to do the reverse.

	@exc
	.EImplement  method not implemented for your Int native size

	@xmp
		i = 123
		s = i.pack
		println (s.len)
		i = Int.unpack (s)
		println (i)

	@out
		8
		123
*/
static qu_string qu__int_pack (qu_int self)
{

#if SIZEOF_LONG == 4

	long v[2] = {qu_int_tol (self), 0};
	return qu_string_new ((char *) &v, 8);

#elif SIZEOF_LONG == 8

	long v = qu_int_tol (self);
	return qu_string_new ((char *) &v, 8);

#else

	QU_THROW (QuEImplement);

#endif

}



/*) sub round (n: Int) -> self
	This method always return self.
*/
static qu_int qu__int_round (qu_int self, qu_integer n)
{
	return self;
}



/*) sub sign -> Int
	Return -1, 0, 1 if the number is negatve, zero, or positive respectively.
*/
static qu_int qu__int_sign (qu_int self)
{
	long v = qu_int_tol (self);
	return (v < 0) ? QuInt_1 : ((v > 0) ? QuInt1 : QuInt0);
}



/*) sub trunc -> self
	This method always return self.
*/
#define qu__int_trunc qu_object_method_self



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_int__setup ()
{
#ifdef QU_DEBUG_LOCAL
	qu_class_setup_final (QuIntClass, /*fake*/ qu_object_s, qu__int__destroy, 0);
#else
	qu_class_setup_final (QuIntClass, /*fake*/ qu_object_s, 0, 0);
#endif
}



void qu_int__init ()
{
	/* class constants */
	qu_class_constz  (QuIntClass, "MIN", qu_int_new (QU_INT_MIN));
	qu_class_constz  (QuIntClass, "MAX", qu_int_new (QU_INT_MAX));

	/* standard class methods */
	qu_class_method_class  (QuIntClass, QU_NAME_new, qu__int_new, -1);

	/* specific class methods */
	qu_class_method_classz (QuIntClass, "unpack", qu__int_unpack, -1);

	/* standard instance methods */
	qu_class_method_inst   (QuIntClass, QU_NAME__as    , qu__int__as    , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__cmp   , qu__int__cmp   , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__decr  , qu__int__decr  , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__div   , qu__int__div   , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__eq    , qu__int__eq    , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__format, qu__int__format, 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__hash  , qu__int__hash  , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__iand  , qu__int__iand  , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__idiv  , qu__int__idiv  , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__incr  , qu__int__incr  , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__inot  , qu__int__inot  , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__ior   , qu__int__ior   , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__ixor  , qu__int__ixor  , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__like  , qu__int__like  , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__lshift, qu__int__lshift, 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__minus , qu__int__minus , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__mul   , qu__int__mul   , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__neg   , qu__int__neg   , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__plus  , qu__int__plus  , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__pos   , qu__int__pos   , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__pow   , qu__int__pow   , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__print , qu__int__print , 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__rem   , qu__int__rem   , 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__rshift, qu__int__rshift, 1);
	qu_class_method_inst   (QuIntClass, QU_NAME__string, qu__int__string, 0);
	qu_class_method_inst   (QuIntClass, QU_NAME__true  , qu__int__true  , 0);

	/* specific instance methods */
	qu_class_method_instz  (QuIntClass, "abs"     , qu__int_abs     ,  0);
	qu_class_method_instz  (QuIntClass, "bin"     , qu__int_bin     ,  0);
	qu_class_method_instz  (QuIntClass, "den"     , qu__int_den     ,  0);
	qu_class_method_instz  (QuIntClass, "ceil"    , qu__int_ceil    ,  0);
	qu_class_method_instz  (QuIntClass, "digitsum", qu__int_digitsum,  0);
	qu_class_method_instz  (QuIntClass, "floor"   , qu__int_floor   ,  0);
	qu_class_method_instz  (QuIntClass, "frac"    , qu__int_frac    ,  0);
	qu_class_method_instz  (QuIntClass, "hex"     , qu__int_hex     ,  0);
	qu_class_method_instz  (QuIntClass, "iseven"  , qu__int_iseven  ,  0);
	qu_class_method_instz  (QuIntClass, "isfrac"  , qu__int_isfrac  ,  0);
	qu_class_method_instz  (QuIntClass, "isodd"   , qu__int_isodd   ,  0);
	qu_class_method_instz  (QuIntClass, "iswhole" , qu__int_iswhole ,  0);
	qu_class_method_instz  (QuIntClass, "num"     , qu__int_num     ,  0);
	qu_class_method_instz  (QuIntClass, "oct"     , qu__int_oct     ,  0);
	qu_class_method_instz  (QuIntClass, "pack"    , qu__int_pack    ,  0);
	qu_class_method_instz  (QuIntClass, "round"   , qu__int_round   ,  1);
	qu_class_method_instz  (QuIntClass, "sign"    , qu__int_sign    ,  0);
	qu_class_method_instz  (QuIntClass, "strlen"  , qu__int_strlen  , -1);
	qu_class_method_instz  (QuIntClass, "trunc"   , qu__int_trunc   ,  0);

	/* finalize declaration */
	qu_class_set_super (QuIntClass, QuIntegerClass);
	qu_class_enddecl   (QuIntClass);
}
