/* Integer.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Integer is Number __abstract
	This is the parent class of the -{Int} and -{Long} classes.

	@tree
	.-{Number}
	.	Integer
	.		-{Int}   system native integer
	.		-{Long}  arbitrary size integer

	@xmp
		i = 1.2
		println (i is Integer)

	@out
		false
*/
qu_class QuIntegerClass;


/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

#define UNOP(name, rtype) \
rtype qu_integer_##name (qu_integer x) \
{ \
	return qu_int_is (x) \
		? (rtype) qu_int_##name  ((qu_int) x) \
		: (rtype) qu_long_##name ((qu_long) x); \
}

#define BINOP_BASIC(name, rtype, ytype) \
rtype qu_integer_##name##_##ytype (qu_integer x, qu_##ytype y) \
{ \
	return qu_int_is (x) \
		? (rtype) qu_int_##name##_##ytype  ((qu_int)  x, y) \
		: (rtype) qu_long_##name##_##ytype ((qu_long) x, y); \
}

#define BINOP_int(name, rtype)      BINOP_BASIC (name, rtype, int)
#define BINOP_long(name, rtype)     BINOP_BASIC (name, rtype, long)
#define BINOP_rational(name, rtype) BINOP_BASIC (name, rtype, rational)
#define BINOP_float(name, rtype)    BINOP_BASIC (name, rtype, float)
#define BINOP_big(name, rtype)      BINOP_BASIC (name, rtype, big)
#define BINOP_complex(name, rtype)  BINOP_BASIC (name, rtype, complex)
#define BINOP_inf(name, rtype)      BINOP_BASIC (name, rtype, inf)

#define BINOP_integer(name, rtype) \
rtype qu_integer_##name##_integer (qu_integer x, qu_integer y) \
{ \
	return qu_int_is (x) \
		? (qu_int_is (y) \
			? (rtype) qu_int_##name##_int  ((qu_int) x, (qu_int) y) \
			: (rtype) qu_int_##name##_long ((qu_int) x, (qu_long) y)) \
		: (qu_int_is (y) \
			? (rtype) qu_long_##name##_int  ((qu_long) x, (qu_int) y) \
			: (rtype) qu_long_##name##_long ((qu_long) x, (qu_long) y)); \
}

#define BINOP_decimal(name, rtype) \
rtype qu_integer_##name##_decimal (qu_integer x, qu_decimal y) \
{ \
	return qu_int_is (x) \
		? (qu_float_is (y) \
			? (rtype) qu_int_##name##_float ((qu_int) x, (qu_float) y) \
			: (rtype) qu_int_##name##_big   ((qu_int) x, (qu_big) y)) \
		: (qu_float_is (y) \
			? (rtype) qu_long_##name##_float ((qu_long) x, (qu_float) y) \
			: (rtype) qu_long_##name##_big   ((qu_long) x, (qu_big) y)); \
}

#define BINOP_real(name, rtype) \
	QU_NUMBER_BINOP_real (qu_integer, name, rtype)

#define BINOP_number(name, rtype) \
	QU_NUMBER_BINOP_number (qu_integer, name, rtype)



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (*arg) -> Integer
	@def
	sub class new (x) -> Integer
	sub class new (x: String, base = 10, offset = 0) -> Integer
	@p
	Use this method to convert v{x} into an Integer number.

	The first form simply returns the value returned by C{x.__as (Integer)}
	provided that it is a Integer number. Otherwise EImplement is thrown.

	The second form allows you to interpret a string with a specified v{base}
	and optionally start with a specified v{offset}. A negative v{offset}
	means to count from the end of the string.

	@warn
	Only base 2, 8, 10, and 16 are supported.

	@exc
	.ESyntax
	.EImplement
	.EArgValue
	.EIndex

	@xmp
		r = Integer ('123.4')
		println (r ' ' r.__name)

		r = Integer ('=41', 1, 16)
		println (r ' ' r.__name)

	@out
		123 Int
		65 Int
*/
static qu_integer _integer_value (int argc, qu_any *argv)
{
	long       ofs  = 0;
	int        base = 10;
	long       l;
	char      *z;
	qu_string  s;
	qu_integer v;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, s, qu_string);
	l = s->len;
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		QU_ENSURE_INDEX (ofs, l);
		if (argc > 2)
		{
			QU_ARG_vint (2, base);
			QU_ARG_CHECK (base == 2 || base == 8 || base == 10 || base == 16);
		}
	}

	z = s->str + ofs;

	switch (base)
	{
		case 2:
			v = (qu_integer) qu_int_scan_binary (&z, 0);
			if (v == 0)
				v = (qu_integer) qu_long_scan_binary (&z, 0);
			break;

		case 8:
			v = (qu_integer) qu_int_scan_oct (&z, 0);
			if (v == 0)
				v = (qu_integer) qu_long_scan_oct (&z, 0);
			break;

		case 10:
			v = (qu_integer) qu_int_scan (&z, 0);
			if (v == 0)
				v = (qu_integer) qu_long_scan (&z, 0);
			break;

		case 16:
			v = (qu_integer) qu_int_scan_hex (&z, 0);
			if (v == 0)
				v = (qu_integer) qu_long_scan_hex (&z, 0);
			break;

		default:
			QU_FATAL ("never");
			v = 0;
	}

	QU_REQUIREZ (QuESyntax, v, "Integer (String) invalid syntax");

	if (qu_long_is (v))
	{
		l = qu_long_toli ((qu_long) v);
		if (errno == 0)
			return (qu_integer) qu_int_new (l);
	}
	return v;
}



static qu_integer qu__integer_new (qu_class self, int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VMIN (1);
	x = argv[0];

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Long:
		case QU_CLASS_ID_Integer:
			return (qu_integer) x;

		case QU_CLASS_ID_Float:
			return (qu_integer) qu_float_as_integer ((qu_float) x);

		case QU_CLASS_ID_Rational:
			return (qu_integer) qu_rational_as_integer ((qu_rational) x);

		case QU_CLASS_ID_Big:
			return (qu_integer) qu_big_as_integer ((qu_big) x);

		case QU_CLASS_ID_Boolean:
			return (qu_integer) ((qu_sys_true (x) == QuTrue) ? QuInt1 : QuInt0);

		case QU_CLASS_ID_String:
			return _integer_value (argc, argv);
	}
	return qu_object_as (x, QuIntegerClass);
}



qu_integer qu_integer_newl (long v)
{
	if (v < QU_INT_MIN || v > QU_INT_MAX)
		return (qu_integer) qu_long_newl (v);
	return (qu_integer) qu_int_new (v);
}



qu_integer qu_integer_newul (unsigned long v)
{
	if (v > (unsigned long) QU_INT_MAX)
		return (qu_integer) qu_long_newul (v);
	return (qu_integer) qu_int_new ((long) v);
}



qu_integer qu_integer_newd (double v)
{
	long x = (long) v;

	if (QU_INT_MIN <= x && x <= QU_INT_MAX)
		return (qu_integer) qu_int_new (x);
	return (qu_integer) qu_long_newd (v);
}



#if HAVE_LONG_LONG
qu_integer qu_integer_newll (long long v)
{
	if (v < (long long) QU_INT_MIN || v > (long long) QU_INT_MAX)
		return (qu_integer) qu_long_newll (v);
	return (qu_integer) qu_int_new ((long) v);
}



qu_integer qu_integer_newull (unsigned long long v)
{
	if (v > (unsigned long long) QU_INT_MAX)
		return (qu_integer) qu_long_newull (v);
	return (qu_integer) qu_int_new (v);
}
#endif /*HAVE_LONG_LONG*/



#ifdef HAVE_LONG_DOUBLE
qu_integer qu_integer_newld (long double v)
{
	long x = (long) v;
	if (QU_INT_MIN <= x && x <= QU_INT_MAX)
		return (qu_integer) qu_int_new (x);
	return (qu_integer) qu_long_newld (v);
}
#endif



/*-----------------------------------------------------------------------------
	specific class methods
-----------------------------------------------------------------------------*/

/*) sub gcd2 (x: Integer, y: Integer) -> Integer
	Return the greatest common divisor (GCD) of the two numbers.

	@exc
	.EDiv v{y} is zero

	@see
	#{gcd1}
*/
static qu_integer qu__integer_gcd2 (qu_class self, qu_integer x, qu_integer y)
{
	QU_ARG_CHECK (qu_integer_is (x) && qu_integer_is (y));
	return qu_integer_gcd (x, y);
}



long qu_integer_gcdl (long n, long d)
{
	register long a = n;
	register long b = d;
	register long r;

	assert (d != 0);

	if (a < 0)
		a = -a;

	if (b < 0)
		b = -b;

	if (a == 0)
		return 1;

	r = a % b;
	while (r)
	{
		a = b;
		b = r;
        r = a % b;
	}
	return b;
}



qu_integer qu_integer_gcd (qu_integer n, qu_integer d)
{
	QU_REQUIREZ (QuEDiv, !qu_integer_iszero (d), "gcd (x/0)");

	if (qu_int_is (n) && qu_int_is (d))
	{
		long g = qu_integer_gcdl (qu_int_tol (n), qu_int_tol (d));
		assert (g <= QU_INT_MAX);
		return (qu_integer) qu_int_new (g);
	}
	else
	{
		qu_integer r;

		if (qu_integer_isneg (n))
			n = qu_integer_neg (n);

		if (qu_integer_isneg (d))
			d = qu_integer_neg (d);

		r = qu_integer_rem_integer (n, d);
		while (!qu_integer_iszero (r))
		{
			n = d;
			d = r;
			r = qu_integer_rem_integer (n, d);
		}
		return d;
	}
}



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (sign   , int)
UNOP (iszero , int)
UNOP (isneg  , int)
UNOP (isone  , int)
UNOP (is1    , int)
UNOP (iswhole, int)
UNOP (isfrac , int)
UNOP (isodd  , int)


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

long qu_integer_tol (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_tol ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_tol ((qu_long) x);
	errno = EDOM;
	return 0;
}



unsigned long qu_integer_toul (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_toul ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_toul ((qu_long) x);
	errno = EDOM;
	return 0;
}



float qu_integer_tof (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_tof ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_tof ((qu_long) x);
	errno = EDOM;
	return 0.0;
}



double qu_integer_tod (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_tod ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_tod ((qu_long) x);
	errno = EDOM;
	return 0.0;
}



#ifdef HAVE_LONG_LONG
long long qu_integer_toll (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_toll ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_toll ((qu_long) x);
	errno = EDOM;
	return 0LL;
}



unsigned long long qu_integer_toull (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_toull ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_toull ((qu_long) x);
	errno = EDOM;
	return 0ULL;
}
#endif /*HAVE_LONG_LONG*/



#ifdef HAVE_LONG_DOUBLE
long double qu_integer_told (qu_any x)
{
	errno = 0;
	if (qu_int_is (x))
		return qu_int_told ((qu_int) x);
	if (qu_long_is (x))
		return qu_long_told ((qu_long) x);
	errno = EDOM;
	return 0.0;
}
#endif




qu_int qu_integer_as_int (qu_integer x)
{
	return qu_int_is (x) ? (qu_int) x : qu_long_as_int ((qu_long) x);
}



qu_long qu_integer_as_long (qu_integer x)
{
	return qu_int_is (x) ? qu_int_as_long ((qu_int) x) : (qu_long) x;
}



qu_float qu_integer_as_float (qu_integer x)
{
	return qu_int_is (x) ? qu_int_as_float ((qu_int) x) : qu_long_as_float ((qu_long) x);
}



qu_big qu_integer_as_big (qu_integer x)
{
	return qu_int_is (x) ? qu_int_as_big ((qu_int) x) : qu_long_as_big ((qu_long) x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (neg , qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (incr, qu_real)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (decr, qu_real)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	~x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (inot, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (as_string, qu_string)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (eq, int)
BINOP_long     (eq, int)
BINOP_rational (eq, int)
BINOP_float    (eq, int)
BINOP_big      (eq, int)
BINOP_complex  (eq, int)
BINOP_inf      (eq, int)
BINOP_integer  (eq, int)
BINOP_decimal  (eq, int)
BINOP_real     (eq, int)
BINOP_number   (eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (like, int)
BINOP_long     (like, int)
BINOP_rational (like, int)
BINOP_float    (like, int)
BINOP_big      (like, int)
BINOP_complex  (like, int)
BINOP_inf      (like, int)
BINOP_integer  (like, int)
BINOP_decimal  (like, int)
BINOP_real     (like, int)
BINOP_number   (like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (cmp, int)
BINOP_long     (cmp, int)
BINOP_rational (cmp, int)
BINOP_float    (cmp, int)
BINOP_big      (cmp, int)
BINOP_complex  (cmp, int)
BINOP_inf      (cmp, int)
BINOP_integer  (cmp, int)
BINOP_decimal  (cmp, int)
BINOP_real     (cmp, int)
BINOP_number   (cmp, int)

/*
	|<=>|
*/

BINOP_int      (cmpabs, int)
BINOP_long     (cmpabs, int)
BINOP_rational (cmpabs, int)
BINOP_float    (cmpabs, int)
BINOP_big      (cmpabs, int)
BINOP_complex  (cmpabs, int)
BINOP_inf      (cmpabs, int)
BINOP_integer  (cmpabs, int)
BINOP_decimal  (cmpabs, int)
BINOP_real     (cmpabs, int)
BINOP_number   (cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (plus, qu_real)
BINOP_long     (plus, qu_real)
BINOP_rational (plus, qu_real)
BINOP_float    (plus, qu_real)
BINOP_big      (plus, qu_real)
BINOP_complex  (plus, qu_complex)
BINOP_inf      (plus, qu_real)
BINOP_integer  (plus, qu_real)
BINOP_decimal  (plus, qu_real)
BINOP_real     (plus, qu_real)
BINOP_number   (plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (minus, qu_real)
BINOP_long     (minus, qu_real)
BINOP_rational (minus, qu_real)
BINOP_float    (minus, qu_real)
BINOP_big      (minus, qu_real)
BINOP_complex  (minus, qu_complex)
BINOP_inf      (minus, qu_real)
BINOP_integer  (minus, qu_real)
BINOP_decimal  (minus, qu_real)
BINOP_real     (minus, qu_real)
BINOP_number   (minus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (mul, qu_real)
BINOP_long     (mul, qu_real)
BINOP_rational (mul, qu_real)
BINOP_float    (mul, qu_real)
BINOP_big      (mul, qu_real)
BINOP_complex  (mul, qu_complex)
BINOP_inf      (mul, qu_real)
BINOP_integer  (mul, qu_real)
BINOP_decimal  (mul, qu_real)
BINOP_real     (mul, qu_real)
BINOP_number   (mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (div, qu_real)
BINOP_long     (div, qu_real)
BINOP_rational (div, qu_real)
BINOP_float    (div, qu_real)
BINOP_big      (div, qu_real)
BINOP_complex  (div, qu_complex)
BINOP_inf      (div, qu_real)
BINOP_integer  (div, qu_real)
BINOP_decimal  (div, qu_real)
BINOP_real     (div, qu_real)
BINOP_number   (div, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (idiv, qu_integer)
BINOP_long     (idiv, qu_integer)
BINOP_rational (idiv, qu_integer)
BINOP_float    (idiv, qu_integer)
BINOP_big      (idiv, qu_integer)
BINOP_complex  (idiv, qu_integer)
BINOP_inf      (idiv, qu_integer)
BINOP_integer  (idiv, qu_integer)
BINOP_decimal  (idiv, qu_integer)
BINOP_real     (idiv, qu_integer)
BINOP_number   (idiv, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (rem, qu_real)
BINOP_long     (rem, qu_real)
BINOP_rational (rem, qu_real)
BINOP_float    (rem, qu_real)
BINOP_big      (rem, qu_real)
BINOP_complex  (rem, qu_complex)
BINOP_inf      (rem, qu_integer)
BINOP_integer  (rem, qu_integer)
BINOP_decimal  (rem, qu_real)
BINOP_real     (rem, qu_real)
BINOP_number   (rem, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (pow, qu_real)
BINOP_long     (pow, qu_real)
BINOP_rational (pow, qu_number)
BINOP_float    (pow, qu_number)
BINOP_big      (pow, qu_number)
BINOP_complex  (pow, qu_complex)
BINOP_inf      (pow, qu_real)
BINOP_integer  (pow, qu_number)
BINOP_decimal  (pow, qu_number)
BINOP_real     (pow, qu_number)
BINOP_number   (pow, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	&
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int     (iand, qu_integer)
BINOP_long    (iand, qu_integer)
BINOP_integer (iand, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	|
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int     (ior, qu_integer)
BINOP_long    (ior, qu_integer)
BINOP_integer (ior, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	^
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int     (ixor, qu_integer)
BINOP_long    (ixor, qu_integer)
BINOP_integer (ixor, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<<
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int     (lshift, qu_integer)
BINOP_long    (lshift, qu_integer)
BINOP_integer (lshift, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	>>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int     (rshift, qu_integer)
BINOP_long    (rshift, qu_integer)
BINOP_integer (rshift, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_string qu_integer_format (qu_integer x, qu_format_s *f)
{
	if (qu_int_is (x))
		return qu_int_format ((qu_int) x, f);
	assert (qu_long_is (x));
	return qu_long_format ((qu_long) x, f);
}



/*-----------------------------------------------------------------------------
	specific instance methods
-----------------------------------------------------------------------------*/

/*) sub gcd (x: Integer) -> Integer
	Return the greatest common divisor (GCD) of this Integer and another.

	@exc
	.EArgValue
	.EDiv v{x} is zero

	@see
	#{gcd2}
*/
qu_integer qu__integer_gcd (qu_integer self, qu_integer x)
{
	QU_ARG_TYPE (x, qu_integer);
	return qu_integer_gcd (self, x);
}



/*) sub inv -> Rational
	Return a Rational number which numerator is 1 and denominator is this Integer.

	@exc
	.EValue self is zero
*/
static qu_rational qu__integer_inv (qu_integer self)
{
	QU_ARG_CHECKZ (!qu_integer_iszero (self), "inv (0) is not possible");
	return qu_rational_newr ((qu_integer) QuInt1, self);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_integer__setup ()
{
	qu_class_setup_base (QuIntegerClass, qu_object_s);
}



void qu_integer__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuIntegerClass, QU_NAME_new, qu__integer_new, -1);

	/* specific class methods */
	qu_class_method_classz (QuIntegerClass, "gcd2", qu__integer_gcd2, 2);

	/* specific instance methods */
	qu_class_method_instz  (QuIntegerClass, "gcd", qu__integer_gcd, 1);
	qu_class_method_instz  (QuIntegerClass, "inv", qu__integer_inv, 0);

	/* finalize declaration */
	qu_class_set_super (QuIntegerClass, QuRealClass);
	qu_class_enddecl   (QuIntegerClass);
}
