/* Iterator.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Iterator __final
	Iterator objects are used to iterate through a collection of items.

	The interpreter automatically creates an iterator as required by the
	language syntax but you can also create custom iterators yourself
	with the #{new} method.

	The k{each} keyword calls the object's __each method. The method must
	return an iterable object, e.g. an Iterator.

	@xmp
		class Foo

			var items

			sub __init
				self.items = [1, 2, 3]
			;;

			sub __each
				println ('I am called')
				return each self.items
			;;
		;;

		x = Foo ()
		for i in x
			print (i)
		;;

	@out
		I am called
		123

	@p
	Here is an example of a custom iterator.
	@prog
		sub bar
			return Iterator (nil, @next, 7)
		;;

		sub next (i)
			x = i.pos
			throw EStop if x > 9
			i.pos (x + 1)
			return x
		;;

		for x in bar
			print (x)
		;;

		println

	@out
		789
*/
qu_class QuIteratorClass;



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (*arg) -> Iterator
	@def
	sub class new (x) -> Iterator
	sub class new (x, func, pos: Any = 0) -> Iterator
	@p
	Create a new Iterator object.

	v{x} is the object to iterate through.

	If v{x} is the only argument then this method simply calls the object's
	__each method and makes sure the returned value is a Iterator, convert it
	if necessary.

	@xmp
		i = Iterator ([1 2 3])
		loop
			try
				print (i.next)
			catch EStop
				break
			;;
		;;

	@out
		123

	@p
	v{func} is the name of the object method to call, a static Sub or a Method.

	v{pos} is the initial position.

	The method or function will be called to get the next item passing the
	iterator object as the only argument. If there is no more value, it
	must throw EStop.

	@warn
	The method/function is called as a trusted function so self is exposed
	to this function. Be careful if you use this from within a method.

	@xmp
		MyArray = [1, 2, 3, 4, 5, 6]

		sub foo (iter)
			i = iter.pos
			try
				r = MyArray[i]
			catch EIndex
				throw EStop
			;;
			iter.pos (i + 2)
			return r
		;;

		for v in Iterator (nil, @foo)
			print (v ' ')
		;;
		println

	@out
		1 3 5

	@p
	Here is an "interesting" example.

	@xmp
		class Foo
			sub doit (it)
				i = it.pos
				try
					r = it.obj [i]
				catch EIndex
					throw EStop
				;;
				it.pos (i + 1)
				println (self ' ' it.obj ' -> ' r)
				return r
			;;
		;;

		x = Foo ()
		i = Iterator ([1, 2, 3], Method (x, "doit"))
		for x in i
			println ('---')
		;;

	@out
		<a Foo> [1, 2, 3] -> 1
		---
		<a Foo> [1, 2, 3] -> 2
		---
		<a Foo> [1, 2, 3] -> 3
		---
*/
static qu_iterator qu__iterator_new (qu_class self, int argc, qu_any *argv)
{
	if (argc == 1)
	{
		return qu_iterator_from (argv[0]);
	}
	else
	{
		qu_any      o, u, p = QuInt0;
		qu_iterator i;

		QU_ARG_VCOUNT (2, 3);
		o = argv[0];
		u = argv[1];
		if (argc > 2)
			p = argv[2];

		switch (qu_classid (u))
		{
			case QU_CLASS_ID_String:
				u = qu_classof (o)->get (o, (qu_string) u, 0);
				QU_ENSURE (u);
				QU_ARG_CHECK (qu_sub_is (u) && ((qu_sub) u)->argc == 1);
				if (qu_sub_is_class (u) && !qu_class_is (o))
					o = qu_classof (o);
				break;

			case QU_CLASS_ID_Sub:
				QU_ARG_CHECK (qu_sub_is_static (u) && ((qu_sub) u)->argc == 1);
				break;

			case QU_CLASS_ID_Method:
				QU_ARG_CHECK (((qu_method) u)->sub->argc == 1);
				break;

			default:
				QU_ARG_THROW;
		}

		i      = qu_gc_new (QuIteratorClass);
		i->obj = o;
		i->pos = p;
		i->met = u;
		i->ex1 = i->ex2 = 0;

		if (qu_sub_is_generator (u))
		{
			qu_frame f;

			f = qu_sub_call1 ((qu_sub) u, o, i);
			QU_ENSURE (f);
			assert (qu_frame_is (f));
			i->met = f;
		}
		return i;
	}
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class nest (*iterables) -> Iterator
	Create a new Iterator object which iterates through v{iterables} in a
	nesting manner.

	There must be at least one argument.

	@exc
	.EImplement object is not an iterable

	@xmp
		for a, b, c in Iterator.nest ([1 2], {5 6}, 7..9)
			println (a b c)
		;;

		# has the same effect as:
		#
		# for a in [1, 2]
		#     for b in {5, 6}
		#         for c in 7..9
		#             println (a b c)
		#         ;;
		#     ;;
		# ;;

	@out
		157
		158
		159
		167
		168
		169
		257
		258
		259
		267
		268
		269

	@p
	Thus if one of the iterables is empty then nothing happens.
*/
static qu_any qu__iterator_nest_next (qu_array iters, qu_iterator self)
{
	qu_array    p = self->pos;
	qu_iterator i;
	qu_any      x;

	if (self->ex1 >= iters->len)
	{
		self->ex1 = iters->len - 1;
		qu_array_pop_quick (p);
	}

	for (; self->ex1 < iters->len;)
	{
		i = iters->objs[self->ex1];
		x = qu_iterator_next (i);
		if (x)
		{
			qu_array_push_quick (p, x);

			++self->ex1;
		}
		else
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			qu_array_pop_quick (p);
			if (--self->ex1 < 0)
				QU_THROW (QuEStop);
			i = iters->objs[self->ex1 + 1];
			i = qu_iterator_from (i->obj);
			QU_ENSURE (i);
			iters->objs[self->ex1 + 1] = i;
		}
	}
	return p;
}


static qu_iterator qu__iterator_nest (qu_class self, int argc, qu_any *argv)
{
	qu_iterator  i;
	qu_array iters;

	QU_ARG_VMIN (1);
	iters = qu_array_new (argc);
	while (argc--)
	{
		i = qu_iterator_from (*argv++);
		QU_ENSURE (i);
		qu_array_push (iters, i);
	}
	i      = qu_gc_new (QuIteratorClass);
	i->obj = iters;
	i->pos = qu_array_new (iters->len);
 	i->met = qu__iterator_nest_next;
	i->ex1 = i->ex2 = 0;

	qu_iterator_set_extern (i);
	return i;
}



/*) sub class seq (*iterables) -> Iterator
	Create a new Iterator object which iterates through v{iterables} in a
	sequential manner.

	There must be at least one argument.

	@exc
	.EImplement object is not an iterable

	@xmp
		for a, b, c, d in Iterator.seq ([1 2 3], {3 4}, [a:'foo' b:'bar'], 5..10)
			println (a b c d)
		;;

		# has the same effect as:
		#
		# a = each [1 2 3]
		# b = each {3 4}
		# c = each [a:'foo' b:'bar]
		# d = each 5..10
		#
		# loop
		#     try
		#         s = a.next _ b.next _ c.next _ d.next
		#     catch EStop
		#         break
		#     ;;
		#     println (s)
		# ;;

	@out
		13a5
		24b6
*/
static qu_any qu__iterator_seq_next (qu_array iters, qu_iterator self)
{
	qu_array    r = 0;
	long        j;
	qu_iterator i;
	qu_any      x;

	assert (qu_array_is (iters));
	for (j = 0; j < iters->len; ++j)
	{
		i = iters->objs[j];
		assert (qu_iterator_is (i));
		x = qu_iterator_next (i);
		QU_ENSURE (x);
		if (r == 0)
			r = qu_array_new (iters->len);
		QU_ENSURE (qu_array_push (r, x));
	}
	if (r == 0)
		return qu_array_new (0);
	return r;
}



static qu_iterator qu__iterator_seq (qu_class self, int argc, qu_any *argv)
{
	qu_iterator i;
	qu_array    iters;

	QU_ARG_VMIN (1);
	iters = qu_array_new (argc);
	while (argc--)
	{
		i = qu_iterator_from (*argv++);
		QU_ENSURE (i);
		qu_array_push (iters, i);
	}

	i      = qu_gc_new (QuIteratorClass);
	i->obj = iters;
	i->pos = QuInt0;
	i->met = qu__iterator_seq_next;
	i->ex1 = i->ex2 = 0;

	qu_iterator_set_extern (i);
	return i;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__iterator__mark (qu_iterator self)
{
	qu_gc_mark (self->obj);
	qu_gc_mark (self->pos);
	if (!qu_iterator_is_extern (self))
		qu_gc_mark (self->met);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Array} |
		Collect all items in an Array.
		 Iterator is exhausted after a call to this method.

	.| -{Set} |
		Collect all items in a Set.
		Iterator is exhausted after a call to this method.

	.| -{Boolean}  | Delegate to #{__true}.
	.| -{String}   | Delegate to #{__string}.
	.| -{Iterator} | Return self.

	@exc
	.EArgValue

	@xmp
		class Foo
			sub __each
				return const [1, 2, 1]
			;;
		;;

		f = Foo ()
		a = (each f) as Array
		println (a)

		a = (each f) as Set
		println (a)

	@out
		[1, 2, 1]
		{1, 2}
*/
static qu_any qu__iterator__as (qu_iterator self, qu_class c)
{
	QU_ARG_CHECK (qu_class_is (c));

	switch (c->id)
	{
		case QU_CLASS_ID_Array:
		{
			qu_array a = qu_array_new (4);
			for (;;)
			{
				qu_any x = qu_iterator_next (self);
				if (x == 0)
				{
					QU_ENSURE (qu_thread_err_is_estop ());
					break;
				}
				QU_ENSURE (qu_array_push (a, x));
			}
			return a;
		}

		case QU_CLASS_ID_Set:
		{
			qu_set a = qu_set_new (4);
			for (;;)
			{
				qu_any x = qu_iterator_next (self);
				if (x == 0)
				{
					QU_ENSURE (qu_thread_err_is_estop ());
					break;
				}
				QU_ENSURE (qu_set_push (a, x));
			}
			return a;
		}

		case QU_CLASS_ID_String:
			return qu_iterator_as_string (self);

		case QU_CLASS_ID_Boolean:
			return qu_iterator_as_boolean (self);

		case QU_CLASS_ID_Frame:
			return self;
	}
	QU_ARG_THROW;
}



/*) sub __next -> Any
	Get the next item.

	@exc
	.EStop     no more values
	.Exception anything thrown by the method or function called

	@xmp
		i = each [1, 2, 3]
		println (i.__next, i.__next, i.__next)
		try
			println (i.__next)
		catch
			println (__err__)
		;;

	@out
		123
		(EStop) stop iteration
*/
/*) sub next
	This is an alias of #{__next}.
*/
qu_any qu__iterator__next (qu_iterator self)
{
	qu_any x = self->met;

	if (qu_iterator_is_extern (self))
		return ((qu_iterator_f) x) (self->obj, self);

	if (qu_sub_is (x))
		return qu_sub_call1 (x, self->obj, self);

	if (qu_method_is (x))
		return qu_method_call1 (x, self);

	assert (qu_frame_is (x));
	return qu_interp_frame_next (x);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub obj -> Any
	Get the iterated object, i.e. the object passed to the #{new} method).

	@xmp
		i = each [1 2 3]
		println (i)
		println (i.obj)
		println (i.next)

	@out
		<a Iterator>
		[1, 2, 3]
		1
*/
static qu_any qu__iterator_obj (qu_iterator self)
{
	return self->obj;
}



/*) sub pos (*arg) -> Any
	@def
	sub pos -> Any
	sub pos (newpos: Any) -> Null

	@p
	Set or get the current position.

	Without any argument, return the current position. With one argument,
	set the current position and return nil.

	v{newpos} may be anything. It is up to the iterator method or function
	to deal with it.

	See the #{new} method for an example.

	@note
	Some builtin iterators do not allow read/write of the current position.
	If they do not, then this method throws ENone.

	@exc
	.ENone can not set position
*/
static qu_any qu__iterator_pos (qu_iterator self, int argc, qu_any *argv)
{
	QU_REQUIRE (QuENone, !qu_iterator_is_extern (self));
	if (argc > 0)
	{
		QU_ARG_VMIN (1);
		self->pos = argv[0];
	}
	return self->pos;
}



/*----------------------------------------------------------------------------
	C level iterator
----------------------------------------------------------------------------*/

qu_iterator qu_iterator_from (qu_any x)
{
	if (qu_iterator_is (x))
		return x;
	else
	{
		x = qu_sys_each (x);
		QU_ENSURE (x);
		if (!qu_iterator_is (x))
		{
			switch (qu_classid (x))
			{
				case QU_CLASS_ID_String:
				case QU_CLASS_ID_Array:
				case QU_CLASS_ID_Dict:
				case QU_CLASS_ID_Set:
				case QU_CLASS_ID_Frame:
				case QU_CLASS_ID_Range:
				case QU_CLASS_ID_Record:
				case QU_CLASS_ID_Struct:
				{
					qu_class c = qu_classof (x);
					assert (c->accel->feach);
					return c->accel->feach (x);
				}
			}
			QU_THROW (QuEImplement);
		}
	}
	return x;
}



int qu_iterator_do (qu_any x, qu_itercallback_f f, void *data)
{
	long i = 0, j = 0, n = 0;
	int  r;
	qu_any  item, idx;

  again:

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Buffer:
		case QU_CLASS_ID_Array:
		case QU_CLASS_ID_Dict:
		case QU_CLASS_ID_Arg:
		case QU_CLASS_ID_Set:
		case QU_CLASS_ID_Frame:
		case QU_CLASS_ID_Range:
		case QU_CLASS_ID_Iterator:
		case QU_CLASS_ID_Record:
		case QU_CLASS_ID_Struct:
			break;

		default:
			if (i == 1)
				QU_THROW (QuEImplement);
			x = qu_sys_each (x);
			QU_ENSURE (x);
			i = 1;
			goto again;
	}

	for (i = 0; ;++i)
	{
		idx = qu_int_new (i);
		switch (qu_classid (x))
		{
			case QU_CLASS_ID_String:
			case QU_CLASS_ID_Buffer:
				if (i >= ((qu_string) x)->len)
					return QU_ITERATOR_CALLBACK_END;
				item = qu_string_byte (((qu_string) x)->str[i]);
				break;

			case QU_CLASS_ID_Array:
				if (i >= ((qu_array) x)->len)
					return QU_ITERATOR_CALLBACK_END;
				item = ((qu_array) x)->objs[i];
				break;

			case QU_CLASS_ID_Dict:
			case QU_CLASS_ID_Arg:
				if ((idx = qu_dict_walk (x, &item, &j, &n)) == NULL)
					return QU_ITERATOR_CALLBACK_END;
				break;

			case QU_CLASS_ID_Set:
				if ((item = qu_set_walk (x, &j, &n)) == NULL)
					return QU_ITERATOR_CALLBACK_END;
				break;

			case QU_CLASS_ID_Frame:
				item = qu_frame_next (x);
				if (item == 0)
				{
					if (qu_thread_err_is_estop ())
						return QU_ITERATOR_CALLBACK_END;
					return QU_ITERATOR_CALLBACK_ERROR;
				}
				break;

			case QU_CLASS_ID_Range:
				if (i >= ((qu_range) x)->count)
					return QU_ITERATOR_CALLBACK_END;
				item = qu_int_new (((qu_range) x)->first + ((qu_range) x)->step * i);
				break;

			case QU_CLASS_ID_Iterator:
				item = qu_iterator_next (x);
				if (item == 0)
				{
					if (qu_thread_err_is_estop ())
						return QU_ITERATOR_CALLBACK_END;
					return QU_ITERATOR_CALLBACK_ERROR;
				}
				break;

			case QU_CLASS_ID_Record:
				if (i >= ((qu_record) x)->len)
					return QU_ITERATOR_CALLBACK_END;
				item = ((qu_record) x)->objs[i];
				break;

			case QU_CLASS_ID_Struct:
			{
				qu_structentry_s *e;

				if (i >= qu_struct_len ((qu_struct) x))
					return QU_ITERATOR_CALLBACK_END;
				e = &((qu_struct) x)->items[i++];
				item = e->name;
				idx  = e->type;
				break;
			}

			default:
				QU_FATAL ("never");
				item = idx = 0;
		}
		r = f (item, idx, data);
		if (r != QU_ITERATOR_CALLBACK_CONTINUE)
			return r;
	}
	QU_FATAL ("never");
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_iterator__setup ()
{
	qu_class_setup_final (QuIteratorClass, qu_iterator_s, 0, qu__iterator__mark);
}



void qu_iterator__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuIteratorClass, QU_NAME_new, qu__iterator_new, -1);

	/* specific class methods */
	qu_class_method_classz (QuIteratorClass, "seq" , qu__iterator_seq , -1);
	qu_class_method_classz (QuIteratorClass, "nest", qu__iterator_nest, -1);

	/* standard instance methods */
	qu_class_method_inst   (QuIteratorClass, QU_NAME__as  , qu__iterator__as  , 1);
	qu_class_method_inst   (QuIteratorClass, QU_NAME__next, qu__iterator__next, 0);
	qu_class_aliasz        (QuIteratorClass, "__next", "next");

	/* specific instance methods */
	qu_class_method_instz  (QuIteratorClass, "obj", qu__iterator_obj,  0);
	qu_class_method_instz  (QuIteratorClass, "pos", qu__iterator_pos, -1);

	/* finalize declaration */
	qu_class_enddecl (QuIteratorClass);
}
