/* Long.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Long is Integer __final
	A Long object is an integer which can be as big as the available memory
	(with a few limits, of course), contrary to -{Int} which has a limited
	range.

	Refer to the -{Number} class for information on Qu numerics in general.

	The class hierarchy:
	@tree
	.-{Number}
	.	-{Real}
	.		-{Integer}
	.			Long

	@p
	Generally you should make no distinction between an Int and a Long in
	your applications. The only reason why Int exists is performance.
	Once computers are very fast, Int and Long will be merged into -{Integer}.

	For example, if the product of two Int objects does not fit in an Int
	then a Long is produced. If the division of two Long objects fits in
	an Int then a Int is produced.

	@xmp
		a = 1234567
		b = Long (1234567)
		c = a * b
		d = c / b
		println (a.__name ' ' a)
		println (b.__name ' ' b)
		println (c.__name ' ' c)
		println (d.__name ' ' d)

		a = 123456789876543212346789
		println (Float (a))

	@out
		Int 1234567
		Long 1234567
		Long 1524155677489
		Int 1234567
		1.23457e+23

	@p
	Integers are "exact" entities and Qu truly treats them as such. If you divide
	two Integers, the result is exact. If there is no remainder then the result
	is an Integer, a Rational otherwise. See the #{__div} and #{__idiv} methods
	for more information.

	Raising an Integer to the power of another Integer gives you a Rational
	or an Integer as well. See the #{__pow} method.

	When you mix an integer with a floating point number then the result is a
	floating point number.

	@xmp
		a = Long (2)
		println (a / 3)
		println (a ** -3)
		println (a / 3.0)
		println (a ** 3.0)

	@out
		(2 _/ 3)
		(1 _/ 8)
		0.666667
		8.0

	@note
	In Qu it is always possible to convert a Long into a Big and vice versa
	without throwing an exception, provided of course, there is enough memory.

	@warn
	Operation on huge Long number may take a while so it is generally not a
	good idea to write constants that are computed during compile time because
	your program may seem to start slowly.

	On a 4GHz Pentium, calculating 2 ** 65537 takes several minutes.

	@xmp
		x = 2 ** 65537   # done at compile time

		println ('starting ...')
		x = 2
		y = x ** 65537   # done at run-time
*/
qu_class QuLongClass;

/* predefined */
qu_long QuLong0;
qu_long QuLong1;
qu_long QuLong2;
qu_long QuLong_1;
qu_long QuLong_2;
qu_long QuLong10;

/* stupid forwards */
static qu_string _long_digits (qu_long x, int base, int sign, int ucase);


/*----------------------------------------------------------------------------
	definition & basic macros
----------------------------------------------------------------------------*/

/* because I like it short ... */
#define DIGIT   qu_long_digit_t
#define DIGIT2  qu_long_digit2_t
#define SDIGIT2 QU_LONG_SDIGIT2
#define SIZE    QU_LONG_SIZE

#ifdef QU_LONG_2ISLL

#ifdef QU_RSHIFT_PRESERVES_SIGNBIT /* see config.h */
#  define S2RSHIFT(x, n) ((x) >> (n))
#else
#  define S2RSHIFT(x, n) _long_dang_shift (x, n)
   inline static long long _long_dang_shift (long long i, int n)
   {
       return (i < 0) ? ~(~i >> n) : (i >> n);
   }
#endif

#else /*QU_LONG_2ISLL*/

#ifdef QU_RSHIFT_PRESERVES_SIGNBIT
#  define S2RSHIFT(x, n) ((x) >> (n))
#else
#  define S2RSHIFT(x, n) _long_dang_shift (x, n)
   inline static long _long_dang_shift (long i, int n)
   {
       return (i < 0) ? ~(~i >> n) : (i >> n);
   }
#endif

#endif /*QU_LONG_2ISLL*/


/*
	Digit stuffs.
*/
#define BITS       (SIZE * 8)
#define SHIFT      BITS
#define RADIX      ((DIGIT2) 1 << BITS)
#define MASK       ((DIGIT) -1)
#define HIBITMASK  (1 << (BITS - 1))

/*
	qu_set and unset sign.
*/
#define SETNEG(o)        qu_flag_set   (o, QU_LONG_FLAG_NEG)
#define SETPOS(o)        qu_flag_unset (o, QU_LONG_FLAG_NEG)
#define SETSIGN(o, sign) QU_BEGIN if (sign > 0) SETPOS (o); else SETNEG (o); QU_END



/*
	For temporary local C long - to avoid allocating memory.
*/
typedef struct TLongLocal TLongLocal;
struct TLongLocal
{
	QU_OBJECT_HEAD;
	DIGIT *digits;
	long   len;
	DIGIT  digit[8]; /* should be enough for short=2 -> longlong=16 */
};


/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

/*
	allocate a new qu_long object with uninitialized digits
*/
static qu_long _long_alloc (long n)
{
	register qu_long g = qu_gc_new (QuLongClass);

	g->len = n++;
	g->digits = qu_mem_alloc (n * SIZE);
	return g;
}



/*
	allocate a new qu_long object with zeroed digits
*/
static qu_long _long_alloc_zero (long n)
{
	register qu_long g = qu_gc_new (QuLongClass);

	g->len = n++;
	g->digits = qu_mem_alloc (n * SIZE);
	memset (g->digits, 0, n * sizeof (DIGIT));
	return g;
}



/*
	allocate a new static qu_long object with zeroed digits
*/
static qu_long _long_alloc_static (long n)
{
	register qu_long g = qu_gc_new_static (QuLongClass);

	g->len = n++;
	g->digits = qu_mem_static (n * SIZE);
	memset (g->digits, 0, n * sizeof (DIGIT));
	return g;
}



/*
	return an exact copy of self
*/
static qu_long _long_dup (qu_long x)
{
	long    n = x->len;
	qu_long y = _long_alloc (n);

	y->flags = x->flags;
	qu_flag_unset (y, QU_FLAG_STATIC);
	qu_flag_unset (y, QU_LONG_FLAG_TMP);
	memcpy (y->digits, x->digits, n * sizeof (DIGIT));
	return y;
}



/*
	return a copy of self with 'n' extra zeroed digits
*/
static qu_long _long_dup_ex (qu_long x, long n)
{
	long    l = x->len;
	long    m = l + n;
	qu_long y;

	assert (n <= INT_MAX - l);

	y        = _long_alloc (m);
	y->len   = l;
	y->flags = x->flags;
	qu_flag_unset (y, QU_FLAG_STATIC);
	qu_flag_unset (y, QU_LONG_FLAG_TMP);

	memcpy (y->digits, x->digits, l * sizeof (DIGIT));
	memset (y->digits + l, 0, n * sizeof (DIGIT));

	return y;
}



/*
	make temporary qu_long object from an qu_int - avoiding memory allocation
	!!! this object must never go to script level
	!!! if returned as a result then it must be transformed with _long_dup
*/
static qu_long _long_local (TLongLocal *x, qu_int y)
{
	long v = qu_int_tol (y);

	x->flags  = QU_LONG_FLAG_TMP;
	x->digits = &x->digit[0];
	x->klass  = QuLongClass;
	x->len    = 0;

	if (v)
	{
#if SIZE == SIZEOF_LONG
		x->len = 1;
		if (v < 0)
		{
			assert (v > LONG_MIN);
			x->digit[0] = (DIGIT) -v;
			x->flags |= QU_LONG_FLAG_NEG;
		}
		else
		{
			x->digit[0] = (DIGIT) v;
		}
#else
		DIGIT *g = x->digits;
		if (v < 0)
		{
			v = -v;
			x->flags |= QU_LONG_FLAG_NEG;
		}
		while (v)
		{
			*g++ = (DIGIT) v;
			v >>= SHIFT;
			x->len++;
		}
#endif
	}
	return (qu_long) x;
}



/*----------------------------------------------------------------------------
	normalization

	We must not make any "difference" between an qu_int and a qu_long so always
	return an qu_int whenever possible.

----------------------------------------------------------------------------*/

/*
	remove leading 0s
*/
static qu_long _long_normalize (qu_long x)
{
	long   l = x->len;
	DIGIT *d = x->digits + l;

	while (l && *--d == 0)
		l--;
	x->len = l;
	return x;
}



/*
	return qu_int or qu_long
*/
static qu_integer _long_return (qu_long x)
{
	long v = qu_long_toli (x);

	if (errno)
		return (qu_integer) x;
	return (qu_integer) qu_int_new (v);
}



/*----------------------------------------------------------------------------
	inline arithmetics with one digit
	!!! must precalculate the size to accommodate possible carry
----------------------------------------------------------------------------*/

/*
	left shift by one digit, return 1 if overflow, 0 otherwise
*/
static int _long_lshift1 (qu_long x, DIGIT n)
{
	long   l = x->len;
	DIGIT *b = x->digits;
	DIGIT  c = 0;
	DIGIT2 u;

	assert (n <= BITS);

	while (l--)
	{
		u = (DIGIT2) *b << n;
		*b++ = (DIGIT) ((u & MASK) | c);
		c = (DIGIT) (u >> SHIFT);
	}
	*b = c;
	return c != 0;
}



/*
	right shift by one digit
*/
static void _long_rshift1 (qu_long x, DIGIT n)
{
	long   l = x->len;
	DIGIT *b = x->digits + l - 1;
	DIGIT  c = 0;
	DIGIT2 u;

	assert (n <= BITS);

	while (l--)
	{
		u = (DIGIT2) *b << SHIFT;
		u >>= n;
		*b-- = (DIGIT) ((u >> SHIFT) | c);
		c = (DIGIT) u;
	}
}



/*
	add with a single digit - return 1 if overflow, 0 otherwise
*/
static int _long_plus1 (qu_long x, DIGIT v)
{
	long   n = x->len;
	DIGIT *b = x->digits;
	DIGIT2 u;

	for (; v; b++, n--)
	{
		u = (DIGIT2) *b + v;
		*b = (DIGIT) u;
		v = (DIGIT) (u >> SHIFT);
	}
	return n < 0;
}



/*
	subtract with a single digit - return 1 if overflow, 0 otherwise
*/
static int _long_minus1 (qu_long x, DIGIT v)
{
	long    l = x->len;
	DIGIT  *d = x->digits;
	SDIGIT2 c = (SDIGIT2) v;

	while (c && l)
	{
		l--;
		c = (SDIGIT2) *d - c;
		*d++ = (DIGIT) (c & MASK);
		c >>= SHIFT;
		c &= 1;
	}
	return c != 0;
}



/*
	multiply with a single digit - return 1 if overflow, 0 otherwise
*/
static int _long_mul1 (qu_long x, DIGIT v)
{
	long   n = x->len;
	DIGIT *b = x->digits;
	DIGIT2 c = 0;

	while (n--)
	{
		c += (DIGIT2) *b * v;
		*b++ = (DIGIT) c;
		c >>= SHIFT;
	}
	*b = c;
	return c != 0;
}



/*
	divide with a single digit, return the remainder
	to check length reduction, test x->digits [x->len - 1] == 0
*/
static DIGIT _long_div1 (qu_long x, DIGIT v)
{
	long   n = x->len;
	DIGIT *b = x->digits + n;
	DIGIT2 c = 0;

	while (n--)
	{
		c = (c << SHIFT) | *--b;
		*b = (DIGIT) (c / v);
		c %= v;
	}
	return (DIGIT) c;
}



/*----------------------------------------------------------------------------
	basic arithmetics - ignoring signs
----------------------------------------------------------------------------*/

/*
	!!! this function requires x->len >= y->len
	!!! returns a new qu_long object
*/
static qu_long _long_basic_plus (qu_long x, qu_long y)
{
	qu_long r;

	assert (x->len > 0 && y->len > 0 && x->len >= y->len);

	if (y->len == 1)
	{
		r = _long_dup_ex (x, 1);
		r->len += _long_plus1 (r, y->digits[0]);
	}
	else
	{
		long   lx = x->len;
		long   ly = y->len;
		DIGIT *dx = x->digits;
		DIGIT *dy = y->digits;
		DIGIT *dr;
		DIGIT2 c = 0;

		r = _long_alloc (lx + 1);
		dr = r->digits;
		for (; ly > 0; dr++, dx++, dy++, ly--, lx--)
		{
			c += (DIGIT2) *dx + *dy;
			*dr = (DIGIT) c;
			c >>= SHIFT;
		}
		for (; lx > 0; lx--, dr++, dx++)
		{
			c += (DIGIT2) *dx;
			*dr = (DIGIT) c;
			c >>= SHIFT;
		}
		*dr = c;
	}
	return _long_normalize (r);
}



/*
	!!! this function requires |x| >= |y| so make sure to compare
	!!! the value to the digits -> use qu_long_cmpabs_long
	!!! returns a new qu_long object
*/
static qu_long _long_basic_minus (qu_long x, qu_long y)
{
	qu_long r;

	assert (x->len > 0 && y->len > 0 && x->len >= y->len);

	if (y->len == 1)
	{
		r = _long_dup (x);
		_long_minus1 (r, y->digits[0]);
	}
	else
	{
		long    lx = x->len;
		long    ly = y->len;
		DIGIT  *dx = x->digits;
		DIGIT  *dy = y->digits;
		DIGIT  *dr;
		SDIGIT2 c = 0;

		r = _long_alloc (lx);
		dr = r->digits;

		for (; ly > 0; dx++, dy++, dr++, ly--, lx--)
		{
			c += (SDIGIT2) *dx - *dy;
			*dr = (DIGIT) c;
			c = S2RSHIFT (c, SHIFT);
		}
		for (; c && lx; lx--, dr++, dx++)
		{
			c += *dx;
			*dr = (DIGIT) c;
			c = S2RSHIFT (c, SHIFT);
		}
		while (lx--)
			*dr++ = *dx++;
	}
	return _long_normalize (r);
}



/*
	!!! this function requires x->len >= y->len
	!!! returns a new qu_long object
*/
static qu_long _long_basic_mul (qu_long x, qu_long y)
{
	qu_long r;

	assert (x->len > 0 && y->len > 0 && x->len >= y->len);

	if (y->len == 1)
	{
		r = _long_dup_ex (x, 1);
		r->len += _long_mul1 (r, y->digits[0]);
	}
	else
	{
		long   lx = x->len;
		long   ly = y->len;
		DIGIT *dx = x->digits;
		DIGIT *dy = y->digits;
		DIGIT *dr;

		r = _long_alloc_zero (lx + ly + 1);
		dr = r->digits;

		for (; lx; lx--, dx++, dr++)
		{
			int    l = ly;
			DIGIT  d = *dx;
			DIGIT *p = dr;
			DIGIT *q = dy;
			DIGIT2 c = 0;

			for (; l; l--, p++, q++)
			{
				c += (DIGIT2) *q * d + *p;
				*p = (DIGIT) c;
				c >>= SHIFT;
			}
			for (; c; p++)
			{
				c += (DIGIT2) *p;
				*p = (DIGIT) c;
				c >>= SHIFT;
			}
		}
	}
	return _long_normalize (r);
}



/*
	!!! this function requires x >= y so make sure to compare the value
	!!! to the digits -> use qu_long_cmpabs_long
	!!! returns a new qu_long object
*/
static qu_long _long_basic_div (qu_long x, qu_long y, qu_long *rem)
{
	assert (x->len > 0 && y->len > 0 && x->len >= y->len);

	if (y->len == 1)
	{
		DIGIT c;
		qu_long  r = _long_dup_ex (x, 1);
		c = _long_div1 (r, y->digits[0]);
		if (rem)
		{
			if (c == 0)
			{
				*rem = QuLong0;
			}
			else
			{
				qu_long e = _long_alloc (1);
				e->digits[0] = c;
				if (qu_long_isneg (x))
					SETNEG (e);
				*rem = e;
			}
		}
		return _long_normalize (r);
	}
	else
	{
		qu_long  r, v, w;
		long     lv, lw;
		DIGIT   *dv, *dw, *dr;
		long     i, k;
		DIGIT    shift = 0;

		/*so that the most significant bit of y is set */
		{
			DIGIT m = HIBITMASK;
			DIGIT d = y->digits [y->len - 1];

			for (; (shift < BITS) && !(d & m); shift++, m >>= 1);

			v = _long_dup_ex (x, 1);
			w = y;

			if (shift)
			{
				if (_long_lshift1 (v, shift))
					v->len++;

				w = _long_dup (y);
				k = _long_lshift1 (w, shift);
				assert (k == 0);
			}
			/* otherwise there is no need to dup y since we are not modifying it */
		}

		dv = v->digits;
		lv = v->len;
		dw = w->digits;
		lw = w->len;

		r = _long_alloc (lv - lw + 2);
		dr = r->digits;
		k = --r->len;
		dr [k] = 0;
		dr [--k] = 0;

		for (i = lv; i >= lw; i--, k--)
		{
			DIGIT quot;
			DIGIT d = (i < lv) ? dv [i] : 0;

			assert (k >= 0);

			/* find quotient digit */
			if (d == dw [lw - 1])
				quot = MASK;
			else
				quot = (((DIGIT2) d << SHIFT) + dv [i - 1]) / dw [lw - 1];

			/* if it's zero then it *is* zero */
			if (quot)
			{
				DIGIT2  h;
				SDIGIT2 c;
				int     j;

				/* multiply and subtract */
				h = 0;
				c = 0;
				for (j = 0; j < lw; j++)
				{
					DIGIT2 r;

					h += (DIGIT2) dw [j] * quot;
					r = c - (h & MASK);
					c = r + dv [j + k];
					if (r)
						dv [j + k] = (DIGIT) (c & MASK);
					c = S2RSHIFT (c, SHIFT);
					h >>= SHIFT;
				}
				c += dv [j + k] - h;

				while (c)
				{
					/* quotient was too big so adjust and add back */
					quot--;
					c = 0;
					for (j = 0; j < lw; j++)
					{
						DIGIT2 r = c + dw [j];
						c = r + dv [j + k];
						if (r)
							dv [j + k] = (DIGIT) (c & MASK);
						c = S2RSHIFT (c, SHIFT);
					}
					c--;
				}

				/* clear high bit --- for the remainder */
				dv [j + k] = 0;

				/* store in result */
				dr [k] = quot;
			}
		}

		if (rem)
		{
			if (shift)
				_long_rshift1 (v, shift);
			*rem = _long_normalize (v);
		}
		return _long_normalize (r);
	}
}



/*
	!!! dont call if p <= 1
*/
static qu_long _long_basic_pow (qu_long x, long p)
{
	qu_long r = x, w = x;

	assert (p > 1);
	for (p--; p; p--)
	{
		while ((p & 1) == 0)
		{
			p >>= 1;

			if (w->len > QU_LONG_MAXLEN - w->len - 1)
				return 0;

			w = _long_basic_mul (w, w);
		}

		if (w->len > QU_LONG_MAXLEN - r->len - 1)
			return 0;

		if (w->len > r->len)
			r = _long_basic_mul (w, r);
		else
			r = _long_basic_mul (r, w);
	}
	return r;
}



/*-----------------------------------------------------------------------------
	specific class properties
-----------------------------------------------------------------------------*/

/*) const DIGITSIZE = 4
	The number of bytes used to represent a digit in a Long number.

	@warn
	This value may depend on the internal implementation of the Long class.

	@noexc

	@xmp
		println (Long.DIGITSIZE)

	@out
		2
*/



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (*arg) -> Int
	@def
	sub class new (x) -> Int
	sub class new (x:String, base: Integer) -> Int
	@p
	Create a new Long object.

	The first form method simply calls C{x.__as (Long)} and makes sure that the
	returned value is an Int otherwise EImplement is thrown.

	The second form only supports base 2, 8, 10 and 16.

	@exc
	.EArgValue
	.EImplement
	.ESyntax
	.Exception  anything thrown by the __as method

	@p
	If v{x} is an infinity then EArgValue is thrown because it is not possible
	to represent infinity with a Long. If v{x} is a -{Number} except infinity,
	then this call always succeed.

	If v{x} is a -{String} then ESyntax is thrown unless the string correctly
	represents a Long as written in the Qu language syntax.

	If v{x} is a -{Boolean} then Long (1) is returned if it is true and
	Long (0) otherwise.

	@xmp
		i = Long ('123')
		j = Long (true)
		println (i)
		println (j)

		class Foo
			sub __as (what)
				switch what
					case Long: return Long (456)
				;;
				throw EImplement
			;;
		;;

		x = Foo ()
		println (Long (x))

	@out
		123
		1
		456
*/
qu_long qu__long_new (qu_class self, int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_int_as_long      ((qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_long) x;
		case QU_CLASS_ID_Rational: return qu_rational_as_long ((qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_float_as_long    ((qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_big_as_long      ((qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_complex_as_long  ((qu_complex ) x);

		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Buffer:
			if (argc > 1)
			{
				char *s = ((qu_string) x)->str;
				int   base;

				QU_ARG_vint (1, base);
				switch (base)
				{
					case  2: return qu_long_scan_binary (&s, 0);
					case  8: return qu_long_scan_oct    (&s, 0);
					case 16: return qu_long_scan_hex    (&s, 0);
					case 10: return qu_long_scan       (&s, 0);
					default: QU_ARG_THROWZ ("unsupported base");
				}
			}
			return qu_long_scan_string (x);

		case QU_CLASS_ID_Boolean:
			return qu_long_new_boolean (x);

		case QU_CLASS_ID_Inf:
			QU_ARG_THROW;
	}
	return qu_object_as (x, QuLongClass);
}



qu_long qu_long_newl (long v)
{
	int     neg = v < 0;
	qu_long g;

	if (v == 0)
		return QuLong0;

	if (v == LONG_MIN)
	{
#if SIZE == SIZEOF_LONG
		g = _long_alloc (1);
		g->digits[0] = ULONG_MAX;
		SETNEG (g);
#else
		g = _long_alloc (2);
		g->digits[0] = USHORT_MAX;
		g->digits[1] = USHORT_MAX;
		SETNEG (g);
#endif
	}
	else
	{
		if (neg)
			v = -v;
#if SIZE == SIZEOF_LONG
		g = _long_alloc (1);
		g->digits[0] = (DIGIT) v;
#else
		if ((v >> SHIFT) == 0)
		{
			g = _long_alloc (1);
		}
		else
		{
			g = _long_alloc (2);
			g->digits[1] = (DIGIT) (v >> SHIFT);
		}
		g->digits[0] = (DIGIT) v;
#endif
		if (neg)
			SETNEG (g);
	}
	return g;
}



qu_long qu_long_newul (unsigned long v)
{
	qu_long g;

	if (v == 0)
		return QuLong0;

#if SIZE == SIZEOF_LONG
	g = _long_alloc (1);
	g->digits[0] = (DIGIT) v;
#else
	if ((v >> SHIFT) == 0)
	{
		g = _long_alloc (1);
	}
	else
	{
		g = _long_alloc (2);
		g->digits[1] = (DIGIT) (v >> SHIFT);
	}
	g->digits[0] = (DIGIT) v;
#endif
	return g;
}



#ifdef HAVE_LONG_LONG
qu_long qu_long_newll (long long v)
{
	int      n = SIZEOF_LONG_LONG / SIZE;
	qu_long  r;
	DIGIT   *d;

	if (v == 0)
		return QuLong0;

	r = _long_alloc (n);
	if (v < 0)
	{
		SETNEG (r);
		v = -v;
	}
	d = r->digits;
	while (n--)
	{
		*d++ = (DIGIT) (v & MASK);
		v >>= SHIFT;
    }
	_long_normalize (r);
	return r;
}



qu_long qu_long_newull (unsigned long long v)
{
	int      n = SIZEOF_LONG_LONG / SIZE;
	qu_long  r;
	DIGIT   *d;

	if (v == 0)
		return QuLong0;

	r = _long_alloc (n);
	d = r->digits;
	while (n--)
	{
		*d++ = (DIGIT) (v & MASK);
		v >>= SHIFT;
    }
	_long_normalize (r);
	return r;
}
#endif /*HAVE_LONG_LONG*/



qu_long qu_long_newd (double v)
{
	double frac;
	int    ex, neg = 0;

	assert (qu_double_isfinite (v));

	if (v < 0)
	{
		neg = 1;
		v = -v;
	}

	frac = frexp (v, &ex);
	if (ex <= 0)
	{
		return QuLong0;
	}
	else
	{
		int      len = ((ex - 1) / BITS) + 1;
		qu_long  g;
		SDIGIT2  b;
		DIGIT  *dg;

		g = _long_alloc (len);
		if (neg)
			SETNEG (g);
		dg = g->digits + len - 1;
		frac = ldexp (frac, ((ex - 1) % BITS) + 1);
		for (; len > 0; len--, dg--)
		{
			b = frac;
			*dg = b;
			frac -= b;
			frac = ldexp (frac, BITS);
		}
		return g;
	}
}



#ifdef HAVE_LONG_DOUBLE
extern long double frexpl (long double VALUE, int *EXPONENT);
extern long double ldexpl (long double VALUE, int EXPONENT);
qu_long qu_long_newld (long double v)
{
	long double frac;
	int         ex, neg = 0;

	assert (qu_longdouble_isfinite (v));

	if (v < 0)
	{
		neg = 1;
		v = -v;
	}

	frac = frexpl (v, &ex);
	if (ex <= 0)
	{
		return QuLong0;
	}
	else
	{
		int      len = ((ex - 1) / BITS) + 1;
		qu_long  g;
		SDIGIT2  b;
		DIGIT   *dg;

		g = _long_alloc (len);
		if (neg)
			SETNEG (g);
		dg = g->digits + len - 1;
		frac = ldexpl (frac, ((ex - 1) % BITS) + 1);
		for (; len > 0; len--, dg--)
		{
			b = frac;
			*dg = b;
			frac -= b;
			frac = ldexpl (frac, BITS);
		}
		return g;
	}
}
#endif



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	scan
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_long _long_scan (char **s, DIGIT base, const unsigned char *charset, int neg, int statik)
{
	char   *i = *s;
	char   *start, *end;
	int     digits, cap, d;
	qu_long g;

	if (!neg)
	{
		while (qu_ascii_isspace (*i))
			i++;
		if (*i == '-')
		{
			neg = 1;
			i++;
		}
		while (qu_ascii_isspace (*i))
			i++;
	}
	if (*i == '\0')
	{
		errno = EDOM;
		return 0;
	}

	start = i;
	digits = 0;
	while (*i)
	{
		if (*i == '_')
		{
			i++;
			continue;
		}
		d = *i;
		if (charset [d] == 0)
			break;
		if (d != '0' || digits != 0)
			digits++;
		i++;
	}
	if (digits == 0)
		return 0;

	*s = i;
	end = i;
	i = start;

	/* FIXME - way too much */
	cap = (digits * SIZE) / SIZE + 1;
	if (cap >= QU_LONG_MAXLEN)
	{
		errno = ERANGE;
		return 0;
	}

	if (statik)
		g = _long_alloc_static (cap);
	else
		g = _long_alloc_zero (cap);

	if (neg)
		SETNEG (g);

	g->len = 1;
	while (i < end)
	{
		if (*i == '_')
		{
			i++;
			continue;
		}
		g->len += _long_mul1 (g, base);
		d = *i++;
		if ('0' <= d && d <= '9')
			d -= 48;
		else
			d = qu_ascii_toupper (d) - 55;
		if (d > 0)
			g->len += _long_plus1 (g, d);
	}
	return g;
}



qu_long qu_long_scan_binary (char **s, int neg)
{
	return _long_scan (s, 2, QU_ASCII_ISBDIGIT, neg, 0);
}



qu_long qu_long_scan_oct (char **s, int neg)
{
	return _long_scan (s, 8, QU_ASCII_ISODIGIT, neg, 0);
}



qu_long qu_long_scan_hex (char **s, int neg)
{
	return _long_scan (s, 16, QU_ASCII_ISXDIGIT, neg, 0);
}



qu_long qu_long_scan (char **s, int neg)
{
	return _long_scan (s, 10, QU_ASCII_ISDIGIT, neg, 0);
}



qu_long qu_long_scan_string (qu_string s)
{
	char    *z = s->str;
	qu_long  x = qu_long_scan (&z, 0);

	QU_REQUIREZ (QuESyntax, x, "Long (String) invalid syntax");
	return x;
}



qu_long qu_long_new_static (const char *s)
{
	char *z = (char *) s;
	return _long_scan (&z, 10, QU_ASCII_ISDIGIT, 0, 1);
}


qu_long qu_long_new_statici  (int v)
{
	char buf [64];
	sprintf (buf, "%d", v);
	return qu_long_new_static (buf);
}



qu_long qu_long_new_staticui (unsigned int v)
{
	char buf [64];
	sprintf (buf, "%u", v);
	return qu_long_new_static (buf);
}



qu_long qu_long_new_staticl  (long v)
{
	char buf [64];
	sprintf (buf, "%ld", v);
	return qu_long_new_static (buf);
}



qu_long qu_long_new_staticul (unsigned long v)
{
	char buf [64];
	sprintf (buf, "%lu", v);
	return qu_long_new_static (buf);
}



qu_long qu_long_to_static (qu_long self)
{
	if (self->len == 0)
		return QuLong0;
	else
	{
		qu_long x = qu_gc_new_static (QuLongClass);

		x->len = self->len;
		x->digits = qu_mem_static (self->len * SIZE);
		memcpy (x->digits, self->digits, self->len * sizeof (DIGIT));
		return x;
	}
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/* none */



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

static void qu__long__destroy (qu_long self)
{
	assert (!qu_flag_test (self, QU_LONG_FLAG_TMP));
	qu_mem_free (self->digits);
}



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_long_sign (qu_long x)
{
	return x->len ? (qu_flag_test (x, QU_LONG_FLAG_NEG) ? -1 : 1) : 0;
}


int qu_long_isone (qu_long x)
{
	return x->len == 1 && x->digits[0] == 1;
}



int qu_long_is1 (qu_long x)
{
	return qu_long_isone (x) && !qu_long_isneg (x);
}



int qu_long_isodd (qu_long x)
{
	return x->len && (x->digits[0] & 1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following classes:
	@ops
	.| -{Int}     | Return an Int if possible, throw EValue otherwise.
	.| -{Integer}
		-{Real}
		-{Number} | Return self.
	.| -{Float}   | Return a Float if possible, throw EValue otherwise.
	.| -{Big}     | Return a Big.
	.| -{Decimal} | Return a Float if possible, Big otherwise.
	.| -{Complex} | Return C{(self + 0j)}.
	.| -{Array}   | Return an Array containing the digits.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{String}  | Delegate to #{__string}.
	.| -{Long}    | Return self.

	@exc
	.EArgValue
	.EValue

	@xmp
		x = 12345678901234567890
		println (x as Float)

	@out
		1.23457e+19
*/
qu_any qu__long__as (qu_long self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:      return qu_long_as_int      (self);
		case QU_CLASS_ID_Long:
		case QU_CLASS_ID_Integer:
		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Number:   return self;
		case QU_CLASS_ID_Rational: return qu_long_as_rational (self);
		case QU_CLASS_ID_Float:    return qu_long_as_float    (self);
		case QU_CLASS_ID_Big:      return qu_long_as_big      (self);
		case QU_CLASS_ID_Complex:  return qu_long_as_complex  (self);
		case QU_CLASS_ID_Decimal:  return qu_long_as_decimal  (self);
		case QU_CLASS_ID_String:   return qu_long_as_string   (self);
		case QU_CLASS_ID_Boolean:  return qu_long_as_boolean  (self);
		case QU_CLASS_ID_Array:    return qu_long_as_array    (self);
	}
	QU_ARG_THROW;
}



long qu_long_tol (qu_long x)
{
	long l = x->len;

	errno = 0;
	if (l == 0)
	{
		return 0;
	}
	else
	{
		int neg = qu_long_isneg (x);

#if SIZE == SIZEOF_LONG
		assert (SIZE <= SIZEOF_LONG);
		if (l == 1)
		{
			DIGIT2 d = (DIGIT2) x->digits[0];
			if (neg)
			{
				if (d <= (DIGIT2) LONG_MAX + 1)
					return -((long) d);
			}
			else
			{
				if (d <= (DIGIT2) LONG_MAX)
					return (long) d;
			}
		}
#else
		DIGIT2 v = 0;
		DIGIT *g = x->digits;

		while (l--)
		{
			v = (v << SHIFT) + (DIGIT2) *g++;
			if (neg)
			{
				if (v > ((DIGIT2) LONG_MAX + 1))
					goto nope;
			}
			else
			{
				if (v > (DIGIT2) LONG_MAX)
					goto nope;
			}
		}
		return (long) v;
    nope:

#endif
	}
	errno = ERANGE;
	return qu_long_isneg (x) ? LONG_MIN : LONG_MAX;
}



unsigned long qu_long_toul (qu_long x)
{
	long l = x->len;

	if (qu_long_isneg (x))
	{
		errno = ERANGE;
		return 0;
	}

	errno = 0;
	if (l == 0)
	{
		return 0;
	}
	else
	{
#if SIZE == SIZEOF_LONG
		assert (SIZE <= SIZEOF_LONG);
		if (l == 1)
			return (unsigned long) x->digits[0];
#else
		DIGIT2 v = 0;
		DIGIT *g = x->digits;

		while (l--)
		{
			v = (v << SHIFT) + (DIGIT2) *g++;
			if (v > (DIGIT2) ULONG_MAX)
				goto nope;
		}
		return (unsigned long) v;
    nope:
		v = v;
#endif
	}
	errno = ERANGE;
	return ULONG_MAX;
}



float qu_long_tof (qu_long x)
{
	double v = qu_long_tod (x);
	if (errno == 0)
	{
		if (v < 0)
		{
			if (v >= (double) FLT_MIN)
				return (float) v;
		}
		else
		{
			if (v <= (double) FLT_MAX)
				return (float) v;
		}
		errno = ERANGE;
	}
	return (float) v;
}



double qu_long_tod (qu_long x)
{
	double r = 0.0;
	long   n = x->len;
	DIGIT *b = x->digits + n;
	int    f;

	while (n--)
		r = r * (double) RADIX + (double) *--b;

	f = qu_double_isinf (r);
	if (f)
	{
		errno = ERANGE;
		return qu_long_isneg (x) ? -DBL_MAX : DBL_MAX;
	}
	errno = 0;
	return qu_long_isneg (x) ? -r : r;
}



#ifdef HAVE_LONG_DOUBLE
long double qu_long_told (qu_long x)
{
	long double r = 0.0;
	long   n = x->len;
	DIGIT *b = x->digits + n;

	while (n--)
		r = r * (long double) RADIX + (long double) *--b;

	if (qu_longdouble_isfinite (r))
	{
		errno = 0;
		return qu_long_isneg (x) ? -r : r;
	}
	errno = ERANGE;
	return qu_long_isneg (x) ? -LDBL_MAX : LDBL_MAX;
}
#endif



#ifdef HAVE_LONG_LONG
long long qu_long_toll (qu_long x)
{
	long l = x->len;

	errno = 0;
	if (l == 0)
	{
		return 0;
	}
	else if (l <= SIZEOF_LONG_LONG / SIZE)
	{
		DIGIT *d = x->digits;
		unsigned long long v = 0;

		while (l--)
			v = (v << SHIFT) | d[l];

		if (!((long long) v < 0 && (qu_long_isneg (x) || (long long) v != LONG_LONG_MIN)))
		{
			if (qu_long_isneg (x))
				return -((long long) v);
			return (long long) v;
		}
	}
	errno = ERANGE;
	return qu_long_isneg (x) ? LONG_LONG_MIN : LONG_LONG_MAX;
}



unsigned long long qu_long_toull (qu_long x)
{
	long l = x->len;

	errno = 0;
	if (l == 0)
	{
		return 0;
	}
	else if (l <= SIZEOF_LONG_LONG / SIZE)
	{
		unsigned long long v = 0;
		DIGIT *d = x->digits;

		while (l--)
			v = (v << SHIFT) | d[l];

		if (!qu_long_isneg (x))
			return v;
	}
	errno = ERANGE;
	return qu_long_isneg (x) ? LONG_LONG_MIN : LONG_LONG_MAX;
}
#endif /*HAVE_LONG_LONG*/



qu_int qu_long_as_int (qu_long x)
{
	long l = qu_long_toli (x);
	QU_REQUIREZ (QuEValue, errno == 0, "value out out Int range");
	return qu_int_new (l);
}



qu_decimal qu_long_as_decimal (qu_long x)
{
	double d = qu_long_tod (x);
	return errno ? (qu_decimal) qu_long_as_big (x) : (qu_decimal) qu_float_newd (d);
}



qu_array qu_long_as_array (qu_long self)
{
	int       l = self->len;
	DIGIT    *d = self->digits;
	qu_array  a = qu_array_new (l);
	qu_any   *p = a->objs;

	while (l--)
	{
		a->len++;
#if QU_LONG_FORMAT == 2
		*p++ = qu_integer_newul (*d++);
#elif QU_LONG_FORMAT == 1
		*p++ = qu_integer_newui (*d++);
#else
		*p++ = qu_integer_newuh (*d++);
#endif
	}
	return a;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	Compare self with another object.

	If v{x} is a Number then their numerical values are compared.
	Refer to -{Number.__cmp} for details on how comparisson is made.

	The returned value is -1 if self is smaller than v{x}, 0 if self equals v{x},
	1 otherwise.

	@noexc

	@xmp
		x = 1 as Long
		println (x <=> 1.2)

	@out
		-1
*/
qu_int qu__long__cmp (qu_long self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_long_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_long_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_long_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_long_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_long_cmp_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_long_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_long_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}



int qu_long_cmp_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	return qu_long_cmp_long (x, _long_local (&tmp, y));
}



int qu_long_cmp_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	QU_NUMBER_COMPARE_SIGN (sx, sy);
	assert (sx == sy && sx);
	return sx * qu_long_cmpabs_long (x, y);
}



QU_NUMBER_BINOP_integer (qu_long, cmp, int)
QU_NUMBER_BINOP_decimal (qu_long, cmp, int)
QU_NUMBER_BINOP_real    (qu_long, cmp, int)
QU_NUMBER_BINOP_number  (qu_long, cmp, int)



/*
	|<=>|
*/

int qu_long_cmpabs_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	return qu_long_cmpabs_long (x, _long_local (&tmp, y));
}



int qu_long_cmpabs_long (qu_long x, qu_long y)
{
	long lx = x->len;
	long ly = y->len;

	if (lx > ly)
	{
		return 1;
	}
	else if (lx != ly)
	{
		return -1;
	}
	else
	{
		DIGIT *dx = x->digits + lx;
		DIGIT *dy = y->digits + lx;

		while (lx--)
		{
			if (*--dx > *--dy)
				return 1;

			else if (*dx != *dy)
				return -1;
		}
	}
	return 0;
}



QU_NUMBER_BINOP_integer (qu_long, cmpabs, int)
QU_NUMBER_BINOP_decimal (qu_long, cmpabs, int)
QU_NUMBER_BINOP_real    (qu_long, cmpabs, int)
QU_NUMBER_BINOP_number  (qu_long, cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Number
	Return self decremented by 1.

	If the result is huge and the -{Number.context} does not allow infinities
	then throws EHuge, otherwise returns infn.

	Note that ESmall is not possible since the subtractor is a whole number.

	@exc
	.EHuge the result is too big

	@xmp
		i = 123456789123456789
		println (i.decr)

	@out
		123456789123456788
*/
qu_real qu__long__decr (qu_long self)
{
	return qu_long_minus_long (self, QuLong1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number
	Return the result of dividing self with v{x}.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{\} operator).

	Refer to -{Number.__div} for details on how the division is performed.

	@p
	When you divide an -{Integer} with another Integer, Qu delivers an "exact"
	result. If the division has a remainder, then a -{Rational} number is
	returned. An -{Integer} otherwise.

	@exc
	.EDiv
	.EInf
	.EHuge
	.ETiny

	@xmp
		x = 1234567891234567891234
		println (x / 2)
		println (x / 3)
		println (x \ 3)
		println (x / 3.0)

		x = Long (1)
		y = x / 3
		println (y ' ' y.__class.__name)
		z = y * 3
		println (z ' ' z.__class.__name)

	@out
		617283945617283945617
		(1234567891234567891234 _/ 3)
		411522630411522630411
		4.11523e+20
		(1 _/ 3) Rational
		1 Int
*/
qu_number qu__long__div (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_long_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_long_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_long_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_long_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_long_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_long_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



qu_real qu_long_div_int (qu_long x, qu_int y)
{
	int sx = qu_long_sign (x);
	int sy = qu_int_sign  (y);

	if (sy == 0)
	{
		QU_REQUIRE (QuEDiv, sx);
		return (qu_real) __qu_zerodiv__ (sx);
	}
	else if (sx == 0)
	{
		return (qu_real) QuInt0;
	}
	else if (qu_int_isone (y))
	{
		qu_integer r = _long_return (x);
		int        s = sx * sy;

		if (qu_long_is (r))
		{
			if (qu_long_sign ((qu_long) r) != s)
				return (qu_real) qu_long_neg ((qu_long) r);
		}
		else
		{
			if (qu_int_sign ((qu_int) r) != s)
				return (qu_real) qu_int_neg ((qu_int) r);
		}
		return (qu_real) r;
	}
	else
	{
		qu_rational r = qu_rational_newr ((qu_integer) x, (qu_integer) y);
		if (qu_rational_iswhole (r))
			return (qu_real) r->num;
		return (qu_real) r;
	}
}



qu_real qu_long_div_long (qu_long x, qu_long y)
{
	if (x == y)
	{
		return (qu_real) QuInt1;
	}
	else
	{
		int sx = qu_long_sign (x);
		int sy = qu_long_sign (y);

		if (sy == 0)
		{
			QU_REQUIRE (QuEDiv, sx);
			return (qu_real) __qu_zerodiv__ (sx);
		}
		else if (sx == 0)
		{
			return (qu_real) QuInt0;
		}
		else if (qu_long_isone (y))
		{
			qu_integer r = _long_return (x);
			int        s = sx * sy;

			if (qu_long_is (r))
			{
				if (qu_long_sign ((qu_long) r) != s)
					return (qu_real) qu_long_neg ((qu_long) r);
			}
			else
			{
				if (qu_int_sign ((qu_int) r) != s)
					return (qu_real) qu_int_neg ((qu_int) r);
			}
			return (qu_real) r;
		}
		else
		{
			qu_rational r = qu_rational_newr ((qu_integer) x, (qu_integer) y);
			if (qu_rational_iswhole (r))
				return (qu_real) r->num;
			return (qu_real) r;
		}
	}
}



qu_int qu_long_div_inf (qu_long x, qu_inf y)
{
	QU_REQUIRE (QuEDiv, !qu_long_iszero (x));
	return QuInt0;
}



QU_NUMBER_BINOP_integer (qu_long, div, qu_real)
QU_NUMBER_BINOP_decimal (qu_long, div, qu_real)
QU_NUMBER_BINOP_real    (qu_long, div, qu_real)
QU_NUMBER_BINOP_number  (qu_long, div, qu_number)



/*
	/ reversed
*/

qu_real qu_long_rdiv_int (qu_long x, qu_int y)
{
	int sx = qu_long_sign (x);
	int sy = qu_int_sign  (y);

	if (sx == 0)
	{
		QU_REQUIRE (QuEDiv, sy);
		return (qu_real) __qu_zerodiv__ (sy);
	}
	else if (sy == 0)
	{
		return (qu_real) QuInt0;
	}
	else if (qu_long_isone (x))
	{
		if (sx * sy < 0)
			y = qu_int_neg (y);
		return (qu_real) y;
	}
	else
	{
		qu_rational r = qu_rational_newr ((qu_integer) y, (qu_integer) x);
		if (qu_rational_iswhole (r))
			return (qu_real) r->num;
		return (qu_real) r;
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if v{x} is a -{Number} and has the same numerical value
	as self, false otherwise.

	Refer to -{Number.__eq} for details on how the equality test is performed.

	@noexc

	@xmp
		x = 1 as Long
		println (x == 1)

	@out
		true
*/
qu_boolean qu__long__eq (qu_long self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_long_eq_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_long_eq_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_long_eq_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_long_eq_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_long_eq_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_long_eq_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



int qu_long_eq_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	return qu_long_eq_long (x, _long_local (&tmp, y));
}



int qu_long_eq_long (qu_long x, qu_long y)
{
	if (qu_long_isneg (x) == qu_long_isneg (y))
	{
		long l = x->len;

		if (l == y->len)
		{
			DIGIT *dx = x->digits;
			DIGIT *dy = y->digits;

			while (l--)
			{
				if (*dx++ != *dy++)
					return 0;
			}
			return 1;
		}
	}
	return 0;
}



QU_NUMBER_BINOP_integer (qu_long, eq, int)
QU_NUMBER_BINOP_decimal (qu_long, eq, int)
QU_NUMBER_BINOP_real    (qu_long, eq, int)
QU_NUMBER_BINOP_number  (qu_long, eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	[x]
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __getitem (idx: Integer) -> Integer
	Return the digit at the specified index.

	Note that C{0} is the least significant digit.

	@warn
	This method is intended for debugging.
	Do not use in production environment.

	@exc
	.EValue
	.EIndex

	@xmp
		x = 12345678901234567890123
		println (x[2])

	@out
		669
*/
qu_integer qu__long__getitem (qu_long self, qu_integer idx)
{
	long i;
	QU_ARG_long (idx, i);
	QU_ENSURE_INDEX (i, self->len);
#if QU_LONG_FORMAT == 2
	return qu_integer_newul (self->digits [i]);
#elif QU_LONG_FORMAT == 1
	return qu_integer_newui (self->digits [i]);
#else
	return qu_integer_newuh (self->digits [i]);
#endif
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this Long.

	See -{Number.__format} for details on the format.

	@exc
	.EFormat
	.EMemory

	@xmp
		i = Long (12345)
		println (i.__format ('%.3d'))
		printfln ('i = %.2,3m', i)
		printfln ('i = %10.3e', i)

	@out
		12345.000
		i = 12,345.00
		i =   1.234e+4
*/
qu_string qu__long__format (qu_long self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_long_format (self, &f);
}



static qu_string _long_digits (qu_long x, int base, int sign, int ucase)
{
	if (qu_long_iszero (x))
	{
		return qu_string_byte ('0');
	}
	else
	{
		long       l = x->len;
		qu_string  s = qu_string_buffer_new (l + 4);
		qu_long    g = _long_dup (x);
		DIGIT     *d = g->digits;
		DIGIT2     h = base * base;
		char       b[1024];
		char      *p = b;
		char      *m = p + sizeof (b) - 1;
		char      *digs = ucase ? QU_ASCII_DIGITU : QU_ASCII_DIGITL;

#if SIZE == 8
		h *= h;
		h *= h;
#elif SIZE == 4
		h *= h;
#endif
		while (l)
		{
			DIGIT *u = d + l;
			DIGIT2 c = 0;
			int    n = l;

			while (n--)
			{
				c = (c << SHIFT) + *--u;
				*u = c / h;
				c %= h;
			}

			if (d [l - 1] == 0)
				l--;

			n = SIZE;
			while (n--)
			{
				if (p == m)
				{
					QU_ENSURE (qu_string_buffer_put (s, b, sizeof (b)));
					p = b;
				}
				*p++ = digs[c % base];
				c /= base;
				if (c == 0 && l == 0)
					break;
			}
		}
		if (p != b)
			QU_ENSURE (qu_string_buffer_put (s, b, p - b));

		switch (base)
		{
			case  2: QU_ENSURE (qu_string_buffer_put (s, "b0", 2)); break;
			case  8: QU_ENSURE (qu_string_buffer_put (s, "0" , 1)); break;
			case 16: QU_ENSURE (qu_string_buffer_put (s, "x0", 2)); break;
		}

		if (sign && qu_long_isneg (x))
			QU_ENSURE (qu_string_buffer_putc (s, '-'));

		qu_string_buffer_reverse (s);
		return qu_string_buffer_return (s);
	}
}



qu_string qu_long_format (qu_long x, qu_format_s *f)
{
	qu_string s;

	switch (f->mode)
	{
		default:  s = _long_digits (x, 10, 0, 0); break;
		case 'b': s = _long_digits (x,  2, 0, 0); break;
		case 'o': s = _long_digits (x,  8, 0, 0); break;
		case 'x': s = _long_digits (x, 16, 0, f->ucase); break;
		case 'y': return qu_format_yesno (f, x);
		case 'p': return qu_format_ptr (f, x);
	}
	QU_ENSURE (s);
	return qu_format_number (f, s, qu_long_isneg (x), 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> Int
	Return the hash value of this Long.

	@noexc

	@xmp
		i = Long (1234)
		println (i.__hash)

	@out
		1234
*/
qu_int qu__long__hash (qu_long self)
{
	long v = qu_long_toli (self);
	return qu_int_new (v);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	&
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __iand (x: Integer) -> Integer
	Return the bitwise-and of self and v{x}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIand

	@xmp
		x = -123456789012345
		println (x & 1)
		println (x & ~x)

		x = -x
		println (x & 1)

	@out
		-1
		0
		1
*/
static qu_integer qu__long__iand (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:  return (qu_integer) qu_long_iand_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long: return (qu_integer) qu_long_iand_long (self, (qu_long) x);
	}
	QU_THROW (QuEIand);
}



qu_int qu_long_iand_int (qu_long x, qu_int y)
{
	int sx = qu_long_sign (x);
	int sy = qu_int_sign (y);

	if (sx == 0 || sy == 0)
	{
		return QuInt0;
	}
	else
	{
		long v = qu_int_tol (y);
		if (sy < 0)
			v = -v;
		v = (long) ((DIGIT) v & x->digits[0]);
		if (sx != sy)
			v = -v;
		return qu_int_new (v);
	}
}



qu_integer qu_long_iand_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0 || sy == 0)
	{
		return (qu_integer) QuLong0;
	}
	else
	{
		long     l;
		qu_long  r;
		DIGIT   *dr, *dy;

		if (x->len > y->len)
		{
			qu_long t = x;
			x = y;
			y = t;
		}

		r = _long_dup (x);
		dr = r->digits;
		dy = y->digits;
		l = r->len;
		while (l--)
			*dr++ &= *dy++;
		SETSIGN (r, sx * sy);
		return (qu_integer) _long_normalize (r);
	}
}



QU_NUMBER_BINOP_integer (qu_long, iand, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer
	Divide self with v{x} in an I{integer context}.

	@warn
	There is a big difference between the #{__div} method (the C{/} operator)
	and the __idiv method (the C{\} operator).

	Refer to -{Number.__idiv} for details on how the division is performed.

	@exc
	.EIdiv

	@xmp
		a = Long (13)
		b = Long (4)
		println (a // b)

	@out
		3
*/
static qu_integer qu__long__idiv (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_integer) qu_long_idiv_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_integer) qu_long_idiv_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_integer) qu_long_idiv_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_integer) qu_long_idiv_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_integer) qu_long_idiv_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_integer) qu_long_idiv_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_integer) qu_long_idiv_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEIdiv);
}



qu_integer qu_long_idiv_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_integer r = qu_long_idiv_long (x, _long_local (&tmp, y));

	if (r == (qu_integer) &tmp)
		return (qu_integer) _long_dup ((qu_long) r);
	return r;
}



qu_integer qu_long_idiv_long (qu_long x, qu_long y)
{
	if (x == y)
	{
		return (qu_integer) QuInt1;
	}
	else
	{
		int sx = qu_long_sign (x);
		int sy = qu_long_sign (y);
		int cmp;

		QU_REQUIRE (QuEIdiv, sy);

		cmp = qu_long_cmpabs_long (x, y);
		if (cmp < 0)
		{
			return (qu_integer) QuInt0;
		}
		else if (cmp > 0)
		{
			qu_long r = _long_basic_div (x, y, NULL);
			SETSIGN (r, sx * sy);
			return _long_return (r);
		}
		else
		{
			return (qu_integer) ((sx == sy) ? QuInt1 : QuInt_1);
		}
	}
}



qu_int qu_long_idiv_inf (qu_long x, qu_inf y)
{
	QU_REQUIRE (QuEIdiv, !qu_long_iszero (x));
	return QuInt0;
}



QU_NUMBER_BINOP_integer (qu_long, idiv, qu_integer)
QU_NUMBER_BINOP_decimal (qu_long, idiv, qu_integer)
QU_NUMBER_BINOP_real    (qu_long, idiv, qu_integer)
QU_NUMBER_BINOP_number  (qu_long, idiv, qu_integer)



/*
	\ reversed
*/

qu_integer qu_long_ridiv_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_integer r = qu_long_idiv_long (_long_local (&tmp, y), x);

	if (r == (qu_integer) &tmp)
		return (qu_integer) _long_dup ((qu_long) r);
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Number
	Return self incremented by 1.

	If the result is huge and the -{Number.context} does not allow infinities
	then throws EHuge, otherwise returns infp.

	@exc
	.EHuge

	@xmp
		a = 123456789123456789
		++a
		println (a)

	@out
		123456789123456790
*/
qu_real qu__long__incr (qu_long self)
{
	return qu_long_plus_long (self, QuLong1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	~
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __inot -> Integer
	Return the bitwise inversion of self.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@noexc

	@xmp
		x = 12345678901234
		println (~x)

		x = -x
		println (~x)

	@out
		18446731728030650381
		-18446731728030650381
*/
qu_integer qu__long__inot (qu_long self)
{
	if (qu_long_iszero (self))
		return (qu_integer) QuInt1;
	else
	{
		DIGIT   *d;
		qu_long  x = self;
		int      n = x->len;

		x = _long_dup_ex (x, 1);
		d = x->digits;
		while (n--)
		{
			*d = ~*d;
			d++;
		}
		return _long_return (x);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	|
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __ior (x: Integer) -> Integer
	Return the bitwise-or of self and v{x}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIor

	@xmp
		x = 12345678901234
		println (x | 4)

	@out
		12345678901238
*/
static qu_integer qu__long__ior (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:  return qu_long_ior_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long: return qu_long_ior_long (self, (qu_long) x);
	}
	QU_THROW (QuEIor);
}



qu_integer qu_long_ior_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0)
	{
		return (qu_integer) y;
	}
	else if (sy == 0)
	{
		return (qu_integer) x;
	}
	else
	{
		long     l;
		qu_long  r;
		DIGIT   *dr, *dy;

		if (x->len < y->len)
		{
			qu_long t = x;
			x = y;
			y = t;
		}
		r = _long_dup (x);
		dr = r->digits;
		dy = y->digits;
		l = y->len;
		while (l--)
			*dr++ |= *dy++;
		SETSIGN (r, sx * sy);
		return (qu_integer) r;
	}
}



qu_integer qu_long_ior_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_integer r = qu_long_ior_long (x, _long_local (&tmp, y));

	if (r == (qu_integer) &tmp)
		return (qu_integer) _long_dup ((qu_long) r);
	return r;
}



QU_NUMBER_BINOP_integer (qu_long, ior, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	^
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __ixor (x: Integer) -> Integer
	Return the bitwise-xor of self and v{x}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIxor

	@xmp
		x = 12345678901234
		println (x ^ 16)

	@out
		12345678901218
*/
static qu_integer qu__long__ixor (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:  return qu_long_ixor_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long: return qu_long_ixor_long (self, (qu_long) x);
	}
	QU_THROW (QuEIxor);
}



qu_integer qu_long_ixor_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0 || sy == 0)
	{
		return (qu_integer) QuInt1;
	}
	else
	{
		long     l;
		qu_long  r;
		DIGIT   *dr, *dy;

		if (x->len < y->len)
		{
			qu_long t = x;
			x = y;
			y = t;
		}
		r = _long_dup (x);
		dr = r->digits;
		dy = y->digits;
		l = y->len;
		while (l--)
			*dr++ ^= *dy++;
		SETSIGN (r, sx *sy);
		return (qu_integer) _long_normalize (r);
	}
}



qu_integer qu_long_ixor_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_integer r = qu_long_ixor_long (x, _long_local (&tmp, y));

	if (r == (qu_integer) &tmp)
		return (qu_integer) _long_dup ((qu_long) r);
	return r;
}



QU_NUMBER_BINOP_integer (qu_long, ixor, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__len
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __len -> Int
	Return the number of digits used.

	@noexc

	@warn
	This method is intended for debugging.
	Do not use in production environment.

	The result may depend on the internal implementation of the Long class.

	@xmp
		l = -1234567889012345678890
		println (l.__len)

	@out
		5
*/
static qu_int qu__long__len (qu_long self)
{
	assert (self->len <= QU_INT_MAX);
	return qu_int_new (self->len);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x) -> Boolean
	If v{x} is a -{Number} and it has the same absolute numerical
	value as self then true is returned, false otherwise.

	Refer to -{Number.__like} for for details on how the comparisson is performed.

	@noexc

	@xmp
		l = 1 as Long
		println (l === -1)

	@out
		true
*/
qu_boolean qu__long__like (qu_long self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_long_like_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_long_like_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_long_like_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_long_like_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_long_like_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_long_like_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_long_like_inf      (self, (qu_inf     ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



QU_NUMBER_BINOP_integer (qu_long, like, int)
QU_NUMBER_BINOP_decimal (qu_long, like, int)
QU_NUMBER_BINOP_real    (qu_long, like, int)
QU_NUMBER_BINOP_number  (qu_long, like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<<
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __lshift (x: Integer) -> Integer
	Return the result of left-shifting the bits of self by the amount of
	v{x}.

	v{x} must be a positive -{Integer} otherwise ELshift is thrown.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.ELshift
	.EHuge

	@xmp
		x = Long (-1)
		println (x << 3)

	@out
		-8
*/
static qu_real qu__long__lshift (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:  return qu_long_lshift_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long: return qu_long_lshift_long (self, (qu_long) x);
	}
	QU_THROW (QuELshift);
}



qu_real qu_long_lshift_long (qu_long x, qu_long y)
{
	long n = qu_long_toli (x);

	if (errno)
		return (qu_real) __qu_huge__ (qu_long_sign (x));
	return qu_long_lshift_int (x, qu_int_new (n));
}



qu_real qu_long_lshift_int (qu_long x, qu_int y)
{
	int n = qu_int_tol (y);

	if (n < 0)
	{
		QU_THROW (QuELshift);
	}
	else if (n == 0)
	{
		return (qu_real) _long_return (x);
	}
	else
	{
		long nd  = n / BITS;
		long len = x->len;

		if (len < ((INT_MAX - sizeof (qu_long_s)) / sizeof (DIGIT) - nd - 1))
		{
			DIGIT   *dr, *dx = x->digits;
			DIGIT2   c = 0;
			qu_long  r = _long_alloc (len + nd + 1);

			dr = r->digits;
			while (nd--)
				*dr++ = 0;
			n %= BITS;
			while (len--)
			{
				c |= (DIGIT2) *dx++ << n;
				*dr++ = c;
				c >>= SHIFT;
			}
			if (c)
				*dr = c;
			else
				r->len--;
			if (qu_long_isneg (x))
				SETNEG (r);
			return (qu_real) r;
		}
		return (qu_real) __qu_huge__ (qu_long_sign (x));
	}
}



qu_real qu_long_rlshift_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_lshift_long (_long_local (&tmp, y), x);

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



QU_NUMBER_BINOP_integer (qu_long, lshift, qu_real)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Return the result of subtracting self with v{x}.

	Refer to -{Number.__minus} for for details on how the subtraction is performed.

	@exc
	.EMinus
	.EHuge
	.EInf

	@xmp
		x = 1234567890123456
		println (x - 1)

	@out
		1234567890123455
*/
static qu_number qu__long__minus (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_long_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_long_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_long_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_long_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_long_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_long_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



/*----------------------------------------------------------------------------
	-
----------------------------------------------------------------------------*/

qu_real qu_long_minus_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_minus_long (x, _long_local (&tmp, y));

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



qu_real qu_long_minus_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0)
	{
		qu_integer r = _long_return (y);
		if (qu_long_is (r))
			return (qu_real) qu_long_neg (y);
		else
			return (qu_real) qu_int_neg ((qu_int) r);
	}
	else if (sy == 0)
	{
		return (qu_real) _long_return (x);
	}
	else
	{
		qu_long r;

		if (sx != sy)
		{
			if (x->len > QU_LONG_MAXLEN - y->len + 1)
				return (qu_real) __qu_huge__ (sx);

			if (x->len > y->len)
			{
				r = _long_basic_plus (x, y);
			}
			else
			{
				r = _long_basic_plus (y, x);
				sx = sy;
			}
		}
		else
		{
			int cmp = qu_long_cmpabs_long (x, y);

			if (cmp > 0)
			{
				r = _long_basic_minus (x, y);
			}
			else if (cmp < 0)
			{
				r = _long_basic_minus (y, x);
			}
			else
			{
				return (qu_real) QuInt0;
			}
		}

		if (r->len)
		{
			SETSIGN (r, sx);
			return (qu_real) _long_return (r);
		}
		return (qu_real) QuInt0;
	}
}



QU_NUMBER_BINOP_integer (qu_long, minus, qu_real)
QU_NUMBER_BINOP_decimal (qu_long, minus, qu_real)
QU_NUMBER_BINOP_real    (qu_long, minus, qu_real)
QU_NUMBER_BINOP_number  (qu_long, minus, qu_number)



/*
	- reversed
*/

qu_real qu_long_rminus_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_minus_long (_long_local (&tmp, y), x);

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Return the result of multiplying self with v{x}.

	Refer to -{Number.__mul} for for details on how the multiplication is performed.

	@exc
	.EMul
	.EHuge
	.ETiny
	.EInf

	@xmp
		x = 1234567890123456
		println (x * 2)

	@out
		2469135780246912
*/
static qu_number qu__long__mul (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_long_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_long_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_long_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_long_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_long_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_long_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



qu_real qu_long_mul_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_mul_long (x, _long_local (&tmp, y));

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



qu_real qu_long_mul_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0 || sy == 0)
	{
		return (qu_real) QuInt0;
	}
	else
	{
		long lx, ly;

		if (x->len < y->len)
		{
			qu_long t = x;
			x = y;
			y = t;
		}

		lx = x->len;
		ly = y->len;

		if (lx > QU_LONG_MAXLEN - ly - 1)
		{
			return (qu_real) __qu_huge__ (sx * sy);
		}
		else
		{
			qu_long r = _long_basic_mul (x, y);
			if (r->len)
				SETSIGN (r, sx * sy);
			return (qu_real) _long_return (r);
		}
	}
}



QU_NUMBER_BINOP_integer (qu_long, mul, qu_real)
QU_NUMBER_BINOP_decimal (qu_long, mul, qu_real)
QU_NUMBER_BINOP_real    (qu_long, mul, qu_real)
QU_NUMBER_BINOP_number  (qu_long, mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Long
	Return the negation of self.

	Note that there is no conversion here. The result is always a Long even
	if the result fits in an Int.

	@noexc

	@xmp
		x = 1234567890123456
		println (-x)

	@out
		-1234567890123456
*/
qu_long qu__long__neg (qu_long self)
{
	if (qu_long_iszero (self))
	{
		return self;
	}
	else
	{
		qu_long l = _long_dup (self);
		if (qu_long_isneg (self))
			SETPOS (l);
		else
			SETNEG (l);
		return l;
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Return the result of adding v{x} to self.

	Refer to -{Number.__plus} for for details on how the addition is performed.

	@exc
	.EPlus
	.EHuge
	.EInf

	@xmp
		x = 1234567890123456
		println (x + 1)

	@out
		1234567890123457
*/
static qu_number qu__long__plus (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_long_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_long_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_long_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_long_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_long_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_long_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPlus);
}



qu_real qu_long_plus_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_plus_long (x, _long_local (&tmp, y));

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



qu_real qu_long_plus_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0)
	{
		return (qu_real) _long_return (y);
	}
	else if (sy == 0)
	{
		return (qu_real) _long_return (x);
	}
	else
	{
		qu_long r;

		if (sx == sy)
		{
			if (x->len > QU_LONG_MAXLEN - y->len + 1)
				return (qu_real) __qu_huge__ (sx);

			if (x->len > y->len)
			{
				r = _long_basic_plus (x, y);
			}
			else
			{
				r = _long_basic_plus (y, x);
			}
		}
		else
		{
			int cmp = qu_long_cmpabs_long (x, y);
			if (cmp > 0)
			{
				r = _long_basic_minus (x, y);
			}
			else if (cmp < 0)
			{
				r = _long_basic_minus (y, x);
				sx = sy;
			}
			else
			{
				return (qu_real) QuInt0;
			}
		}

		if (r->len)
		{
			SETSIGN (r, sx);
			return (qu_real) _long_return (r);
		}
		return (qu_real) QuInt0;
	}
}



QU_NUMBER_BINOP_integer (qu_long, plus, qu_real)
QU_NUMBER_BINOP_decimal (qu_long, plus, qu_real)
QU_NUMBER_BINOP_real    (qu_long, plus, qu_real)
QU_NUMBER_BINOP_number  (qu_long, plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub pos -> self
	This method always return self.
*/
#define qu__long__pos  qu_object_method_self



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x: Number) -> Number
	Return the result of raising self to the power of v{x}.

	Note that if v{x} is an -{Integer} then the result is "exact", that is,
	an Integer. If it is negative then a -{Rational} number is returned.

	Refer to -{Number.__pow} for for details on how the operation is performed.

	@exc
	.EPow
	.EHuge
	.ETiny
	.EInf

	@xmp
		a = Long (3)
		println (a ** 2)
		println (a ** -2)

	@out
		9
		(1 _/ 9)
*/
static qu_number qu__long__pow (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_long_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_long_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_long_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_long_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_long_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_long_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}



qu_real qu_long_pow_int (qu_long x, qu_int y)
{
	/* dont use temporary long because qu_long_pow_long might create a qu_rational */
	return qu_long_pow_long (x, qu_int_as_long (y));
}



qu_real qu_long_pow_long (qu_long x, qu_long y)
{
	int sx = qu_long_sign (x);
	int sy = qu_long_sign (y);

	if (sx == 0)
	{
		if (sy == 0)
			return (qu_real) QuInt1;

		if (sy > 0)
			return (qu_real) QuInt0;

		return (qu_real) __qu_inf__ (1);
	}
	else if (sy == 0)
	{
		return (qu_real) QuInt1;
	}
	else if (qu_long_isone (x))
	{
		return (qu_real) ((qu_long_isneg(x) && qu_long_isodd (y)) ? QuInt_1 : QuInt1);
	}
	else
	{
		long p;

		p = qu_long_tol (y);
		if (errno)
		{
			if (sy < 0)
				return __qu_tiny__ ((qu_real) QuInt0);
			return (qu_real) __qu_huge__ ((qu_long_isodd (y) && (sx < 0)) ? -1 : 1);
		}
		else if (p == 0)
		{
			return (qu_real) QuInt1;
		}
		else if (p == 1)
		{
			return (qu_real) x;
		}
		else if (p == -1)
		{
			return (qu_real) qu_rational_newr ((qu_integer) QuInt1, (qu_integer) x);
		}
		else if (p < 0)
		{
			x = _long_basic_pow (x, -p);
			if (x == 0)
				return __qu_tiny__ ((qu_real) QuInt0);
			else
			{
				qu_integer d = _long_return (x);
				qu_int     n = ((p & 1) && (sx < 0)) ? QuInt_1 : QuInt1;
				return (qu_real) qu_rational_newr ((qu_integer) n, (qu_integer) d);
			}
		}
		else
		{
			x = _long_basic_pow (x, p);
			if (x == 0)
				return (qu_real) __qu_huge__ (((p & 1) && (sx < 0)) ? -1 : 1);
			if ((p & 1) && (sx < 0))
				SETSIGN (x, -1);
			return (qu_real) _long_return (x);
		}
	}
}



qu_real qu_long_pow_inf (qu_long x, qu_inf y)
{
	if (y == QuInfPos)
	{
		if (qu_long_iszero (x))
			return (qu_real) QuInt0;

		else if (qu_long_isone (x))
			return (qu_real) QuInt1;

		else
			return (qu_real) __qu_inf__ (1);
	}
	else
	{
		if (qu_long_iszero (x))
			return (qu_real) __qu_inf__ (1);

		else if (qu_long_isone (x))
			return (qu_real) QuInt1;

		else
			return (qu_real) QuInt0;
	}
}



QU_NUMBER_BINOP_integer (qu_long, pow, qu_real)
QU_NUMBER_BINOP_decimal (qu_long, pow, qu_real)
QU_NUMBER_BINOP_real    (qu_long, pow, qu_real)
QU_NUMBER_BINOP_number  (qu_long, pow, qu_number)



qu_long qu_long_pow_10 (int p)
{
	assert (p >= 0);

	if (p == 0)
	{
		return QuLong1;
	}
	else if (p == 1)
	{
		return QuLong10;
	}
	else
	{
		qu_long r, w, t;

		/* TODO - should be able to calculate the required length then do
			a shift of 3 and a plus */

		r = w = QuLong10;
		for (p--; p; p--)
		{
			while ((p & 1) == 0)
			{
				p >>= 1;

				if (w->len > QU_LONG_MAXLEN - w->len - 1)
					return 0;

				t = _long_basic_mul (w, w);
				_long_normalize (t);
				w = t;
			}

			if (w->len > QU_LONG_MAXLEN - r->len - 1)
				return 0;

			t = (w->len > r->len) ? _long_basic_mul (w, r) : _long_basic_mul (r, w);
			_long_normalize (t);
			r = t;
		}
		return r;
	}
}



/*
	** reversed
*/

qu_real qu_long_rpow_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_pow_long (_long_local (&tmp, y), x);

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print
	Print self to -{Stream.stdout}.

	This method simply prints the string returned by the #{__string} method.

	See -{print} for possible exceptions.

	@xmp
		a = Long (123)
		println (a)

	@out
		123

	@warn
	Printing a huge Long number may take a while.
*/
static qu_null qu__long__print (qu_long self)
{
	qu_string s = qu_long_as_string (self);
	return qu_sys_printz (s->str, s->len);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The meaning of "remainder" may differ with other languages.
	Refer to -{Number.__rem} for details.

	@p
	The remainder of dividing by zero is undefined thus throws ERem.

	@exc
	.ERem

	@xmp
		x = 1234567890123457
		println (x % 2)

	@out
		1
*/
static qu_number qu__long__rem (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_long_rem_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_rem_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_long_rem_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_long_rem_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_long_rem_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_long_rem_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_long_rem_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuERem);
}



qu_real qu_long_rem_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_rem_long (x, _long_local (&tmp, y));

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



qu_real qu_long_rem_long (qu_long x, qu_long y)
{
	if (qu_long_iszero (y))
	{
		QU_THROW (QuERem);
	}
	else if (qu_long_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		int cmp = qu_long_cmpabs_long (x, y);
		if (cmp < 0)
		{
			return (qu_real) x;
		}
		else if (cmp > 0)
		{
			qu_long m;
			_long_basic_div (x, y, &m);
			return (qu_real) _long_return (m);
		}
		else
		{
			return (qu_real) QuInt0;
		}
	}
}



qu_integer qu_long_rem_inf (qu_long x, qu_inf y)
{
	QU_REQUIRE (QuERem, !qu_long_iszero (x));
	return _long_return (x);
}



QU_NUMBER_BINOP_integer (qu_long, rem, qu_real)
QU_NUMBER_BINOP_decimal (qu_long, rem, qu_real)
QU_NUMBER_BINOP_real    (qu_long, rem, qu_real)
QU_NUMBER_BINOP_number  (qu_long, rem, qu_number)



/*
	% reversed
*/

qu_real qu_long_rrem_int (qu_long x, qu_int y)
{
	TLongLocal tmp;
	qu_real r = qu_long_rem_long (_long_local (&tmp, y), x);

	if (r == (qu_real) &tmp)
		return (qu_real) _long_dup ((qu_long) r);
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	>>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rshift (x: Integer) -> Integer
	Return the bitwise-right-shift of self and v{x}.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.ERshift

	@xmp
		x = 1234567890123456
		println (x >> 1)

	@out
		617283945061728
*/
static qu_integer qu__long__rshift (qu_long self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:  return (qu_integer) qu_long_rshift_int  (self, (qu_int ) x);
		case QU_CLASS_ID_Long: return (qu_integer) qu_long_rshift_long (self, (qu_long) x);
	}
	QU_THROW (QuERshift);
}



qu_integer qu_long_rshift_long (qu_long x, qu_long y)
{
	qu_int b = qu_long_as_int (y);

	if (errno)
		return (qu_integer) QuInt0;
	return qu_long_rshift_int (x, b);
}



qu_integer qu_long_rshift_int (qu_long x, qu_int y)
{
	long n = qu_int_tol (y);

	if (n < 0)
	{
		QU_THROW (QuERshift);
	}
	else if (n == 0)
	{
		return (qu_integer) x;
	}
	else
	{
		long nd  = n / BITS;
		long len = x->len - nd;

		if (len > 0)
		{
			qu_long r = _long_alloc (len);

			memcpy (r->digits, x->digits + nd, len * SIZE);
			n %= BITS;
			if (n)
			{
				DIGIT *b = r->digits + len - 1;
				DIGIT  c = 0;
				DIGIT2 u;

				while (len--)
				{
					u = (DIGIT2) *b << SHIFT;
					u >>= n;
					*b-- = (u >> SHIFT) | c;
					c = u;
				}
				if (r->digits [r->len - 1] == 0)
					r->len--;
			}
			if (qu_long_isneg (x))
				SETNEG (r);

			return (qu_integer) _long_return (r);
		}
		return (qu_integer) QuInt0;
	}
}



QU_NUMBER_BINOP_integer (qu_long, rshift, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Return a string representation of self.

	@noexc

	@xmp
		s = Long (123) as String
		println (s)

	@out
		123

	@warn
	Creating a string for a huge Long number may take a while.
*/
qu_string qu__long__string (qu_long self)
{
	if (qu_long_iszero (self))
	{
		return qu_string_byte ('0');
	}
	else
	{
		long      l = self->len;
		qu_string s = qu_string_buffer_new (l + 4);
		qu_long   g = _long_dup (self);
		DIGIT    *d = g->digits;
#if SIZE == 8
		DIGIT2 h = 100000000;
#elif SIZE == 4
		DIGIT2 h = 10000;
#else
		DIGIT2 h = 100;
#endif
		while (l)
		{
			DIGIT *b = d + l;
			DIGIT2 c = 0;
			int    n = l;

			while (n--)
			{
				c = (c << SHIFT) + *--b;
				*b = c / h;
				c %= h;
			}

			if (d [l - 1] == 0)
				l--;

			n = SIZE;
			while (n--)
			{
				if (qu_string_buffer_putc (s, '0' + (c % 10)) == 0)
					goto nope;
				c /= 10;
				if (c == 0 && l == 0)
					break;
			}
		}

		if (qu_long_isneg (self))
		{
			if (qu_string_buffer_putc (s, '-') == 0)
				goto nope;
		}
		qu_string_buffer_reverse (s);
		return qu_string_buffer_return (s);

	  nope:
		return qu_string_newz ("*Long*");
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	Return false if self is 0, true otherwise.

	@noexc

	@xmp
		l = Long (-1)
		println (l ? 'yep' : 'nope')

	@out
		yep
*/
static qu_boolean qu__long__true (qu_long self)
{
	return qu_long_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub abs -> Long
	Return the absolute value.
*/
static qu_long qu__long_abs (qu_long self)
{
	if (qu_long_isneg (self))
		return qu_long_neg (self);
	return self;
}



/*) sub bin -> String
	Return a binary representation.

	@xmp
		x = Long (17)
		println (x.bin)

	@out
		0b10001
*/
qu_string qu__long_bin (qu_long x)
{
	return _long_digits (x, 2, 1, 0);
}



/*) sub ceil -> self
	This method always return self.
*/
#define qu__long_ceil qu_object_method_self



/*) sub den -> Long
	Return the denominator.

	This method always return Long (1).
*/
static qu_long qu__long_den (qu_long self)
{
	return QuLong1;
}


/*) sub digitsum -> Int
	Return the sum of the digits.

	@xmp
		x = Long (1234)
		println (x.digitsum)

	@out
		10
*/
qu_integer qu__long_digitsum (qu_long x)
{
	if (qu_long_iszero (x))
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		qu_string      s = qu_long_as_string (x);
		unsigned char *z = (unsigned char *) s->str;
		long           l = s->len;
		unsigned long  v = 0;

		/* most unlikely going to exceed LONG_MAX */

		if (l == 0)
			return (qu_integer) QuInt0;

		while (!qu_ascii_isdigit (*z))
		{
			--l;
			++z;
		}

		while (l--)
			v += *z++ - '0';

		return qu_integer_newul (v);
	}
}



/*) sub floor -> self
	This method always return self.
*/
#define qu__long_floor qu_object_method_self



/*) sub frac -> Long
	This method always return Long (0).
*/
static qu_long qu__long_frac (qu_long self)
{
	return QuLong0;
}



/*) sub hex -> String
	Return a hexadecimal representation.
*/
qu_string qu__long_hex (qu_long x)
{
	return _long_digits (x, 16, 1, 1);
}



/*) sub iseven -> Boolean
	Return true if the number is even, false otherwise.
*/
static qu_boolean qu__long_iseven (qu_long self)
{
	return qu_long_isodd (self) ? QuFalse : QuTrue;
}


/*) sub isfrac -> Boolean
	This method always return false.
*/
#define qu__long_isfrac qu_object_method_false



/*) sub isodd -> Boolean
	Return true if self is an odd number, false otherwise.

	Note that 0 is considered an even number.
*/
static qu_boolean qu__long_isodd (qu_long self)
{
	return qu_long_isodd (self) ? QuTrue : QuFalse;
}



/*) sub iswhole -> Boolean
	This method always return true.
*/
#define qu__long_iswhole qu_object_method_true



/*) sub num -> self
	Return the numerator.

	This method always return self.
*/
#define qu__long_num qu_object_method_self



/*) sub oct -> String
	Return an octal representation.
*/
qu_string qu__long_oct (qu_long x)
{
	return _long_digits (x, 8, 1, 0);
}



/*) sub round (n: Int) -> self
	This method always return self.
*/
static qu_long qu__long_round (qu_long self, qu_integer n)
{
	return self;
}



/*) sub sign -> Int
	Return -1, 0, 1 if the number is negatve, zero, or positive respectively.
*/
static qu_int qu__long_sign (qu_long self)
{
	int v = qu_long_sign (self);
	return (v < 0) ? QuInt_1 : ((v > 0) ? QuInt1 : QuInt0);
}



/*) sub strlen (sign = true, base = 10, prefix = false) -> Int
	The length of the Long as String.

	If v{sign} is false then don't include the sign.

	v{base} must be one of: C{2}, C{8}, C{16} or C{10}.

	If v{prefix} is k{true} then include the base prefix.

	@xmp
		x = Long (-100)
		println (x.strlen)

	@out
		4
*/
static qu_int qu__long_strlen (qu_long x, int argc, qu_any *argv)
{
	int base = 10, sign = 1, prefix = 0;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vint (0, base);
		QU_ARG_CHECK (base == 10 || base == 16 || base == 8 || base == 2);

		if (argc > 1)
		{
			QU_ARG_vbool (0, sign);
			if (argc > 2)
				QU_ARG_vbool (2, prefix);
		}
	}

	if (qu_long_iszero (x))
	{
		if (prefix)
		{
			switch (base)
			{
				case  2: return QuInt2;
				case  8: return QuInt1;
				case 16: return QuInt2;
			}
		}
		return QuInt1;
	}
	else
	{
		long     l = x->len;
		long     r = 0;
		qu_long  g = _long_dup (x);
		DIGIT   *d = g->digits;
		DIGIT2   h = base * base;
		char     b[1024];
		char    *p = b;
		char    *m = p + sizeof (b) - 1;
		char    *digs = QU_ASCII_DIGITU;

#if SIZE == 8
		h *= h;
		h *= h;
#elif SIZE == 4
		h *= h;
#endif
		while (l)
		{
			DIGIT *u = d + l;
			DIGIT2 c = 0;
			int    n = l;

			while (n--)
			{
				c = (c << SHIFT) + *--u;
				*u = c / h;
				c %= h;
			}

			if (d [l - 1] == 0)
				l--;

			n = SIZE;
			while (n--)
			{
				if (p == m)
				{
					r += sizeof (b);
					p = b;
				}
				*p++ = digs[c % base];
				c /= base;
				if (c == 0 && l == 0)
					break;
			}
		}
		if (p != b)
			r += p - b;

		switch (base)
		{
			case  2: r += 2; break;
			case  8: r += 1; break;
			case 16: r += 2; break;
		}

		if (sign && qu_long_isneg (x))
			r++;

		return qu_int_newl (r);
	}
}



/*) sub trunc -> self
	This method always return self.
*/
#define qu__long_trunc qu_object_method_self



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_long__setup ()
{
	qu_class_setup_final (QuLongClass, qu_long_s, qu__long__destroy, 0);

	/*
		special case - otherwise qu_long_new_static will return QuLong0 which is NULL
		at this point of time
	*/
	QuLong0  = qu_gc_new_static (QuLongClass);
	QuLong0->digits = qu_mem_static (SIZE);
	QuLong0->digits[0] = 0;
	QuLong0->len = 0;

	QuLong1  = qu_long_new_static ("1");
	QuLong_1 = qu_long_new_static ("-1");
	QuLong2  = qu_long_new_static ("2");
	QuLong_2 = qu_long_new_static ("-2");
	QuLong10 = qu_long_new_static ("10");
}



void qu_long__init ()
{
	/* specific class properties */
	qu_class_constz (QuLongClass, "DIGITSIZE", qu_int_new (sizeof (DIGIT)));

	/* standard class methods */
	qu_class_method_class (QuLongClass, QU_NAME_new, qu__long_new, -1);

	/* standard instance methods */
	qu_class_method_inst  (QuLongClass, QU_NAME__as     , qu__long__as     ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__cmp    , qu__long__cmp    ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__decr   , qu__long__decr   ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__div    , qu__long__div    ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__eq     , qu__long__eq     ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__getitem, qu__long__getitem,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__format , qu__long__format ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__hash   , qu__long__hash   ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__iand   , qu__long__iand   ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__idiv   , qu__long__idiv   ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__incr   , qu__long__incr   ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__inot   , qu__long__inot   ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__ior    , qu__long__ior    ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__ixor   , qu__long__ixor   ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__len    , qu__long__len    ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__like   , qu__long__like   ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__lshift , qu__long__lshift ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__minus  , qu__long__minus  ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__mul    , qu__long__mul    ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__neg    , qu__long__neg    ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__plus   , qu__long__plus   ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__pos    , qu__long__pos    ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__pow    , qu__long__pow    ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__print  , qu__long__print  ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__rem    , qu__long__rem    ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__rshift , qu__long__rshift ,  1);
	qu_class_method_inst  (QuLongClass, QU_NAME__string , qu__long__string ,  0);
	qu_class_method_inst  (QuLongClass, QU_NAME__true   , qu__long__true   ,  0);

	/* specific instance methods */
	qu_class_method_instz (QuLongClass, "abs"     , qu__long_abs     ,  0);
	qu_class_method_instz (QuLongClass, "bin"     , qu__long_bin     ,  0);
	qu_class_method_instz (QuLongClass, "ceil"    , qu__long_ceil    ,  0);
	qu_class_method_instz (QuLongClass, "den"     , qu__long_den     ,  0);
	qu_class_method_instz (QuLongClass, "digitsum", qu__long_digitsum,  0);
	qu_class_method_instz (QuLongClass, "floor"   , qu__long_floor   ,  0);
	qu_class_method_instz (QuLongClass, "frac"    , qu__long_frac    ,  0);
	qu_class_method_instz (QuLongClass, "hex"     , qu__long_hex     ,  0);
	qu_class_method_instz (QuLongClass, "iseven"  , qu__long_iseven  ,  0);
	qu_class_method_instz (QuLongClass, "isfrac"  , qu__long_isfrac  ,  0);
	qu_class_method_instz (QuLongClass, "isodd"   , qu__long_isodd   ,  0);
	qu_class_method_instz (QuLongClass, "iswhole" , qu__long_iswhole ,  0);
	qu_class_method_instz (QuLongClass, "num"     , qu__long_num     ,  0);
	qu_class_method_instz (QuLongClass, "oct"     , qu__long_oct     ,  0);
	qu_class_method_instz (QuLongClass, "round"   , qu__long_round   ,  1);
	qu_class_method_instz (QuLongClass, "sign"    , qu__long_sign    ,  0);
	qu_class_method_instz (QuLongClass, "strlen"  , qu__long_strlen  , -1);
	qu_class_method_instz (QuLongClass, "trunc"   , qu__long_trunc   ,  0);

	/* finalize declaration */
	qu_class_set_super (QuLongClass, QuIntegerClass);
	qu_class_enddecl   (QuLongClass);
}
