/* Method.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Method is Function __final
	A Method object is simply a -{Sub} object bind to a self object.

	Method can enhance performance and can be used to I{curry} functions
	or methods.

	Once created, you can call a Method just as you would call a Sub.

	@xmp
		class Foo
			sub hey (x)
				println ('Hey ' x '!')
			;;
		;;
		x = Foo.new
		m = Method (x, "hey")
		m ('World')

	@out
		Hey World!

	@p
	Following is an example on how to I{curry} a function and method.

	The only limitation is that the I{curried} arguments can only be specified
	before the formal arguments.

	@xmp
		sub foo (prefix)
			return prefix _ self.__len
		;;

		x = [1 2 3]
		m = Method (x, @foo)
		println (m ('len = '))


		sub bar (a, b, c)
			println (a b c)
		;;

		m = Method (@bar, 1, 2)
		m (3)  # same as bar (1, 2, 3)


		class Foo
			sub bar (a, b)
				println (a ' ' b)
			;;
		;;

		f = Foo ()
		m = Method (f, "bar", 'Hello')
		m ('World!')   # same as f.bar ('Hello', 'World!')

	@out
		len = 3
		123
		Hello World!
*/
qu_class QuMethodClass;


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (obj, what, *args) -> Method
	@def
	sub class new (obj, what: String, *args) -> Method
	sub class new (obj, what: Sub, *args) -> Method
	sub class new (what: Sub, *args) -> Method
	sub class new (what: Method, *args) -> Method

	@p
	Create a new Method object.

	v{obj} may be any object.

	v{what} is a either a method name of v{obj} or a static Sub (function) or
	a Method.

	v{args} is the arguments to pass before the formal arguments.

	See the -{Method} class documentation for examples.

	@exc
	.EArgValue
	.EImplement method does not exist
*/
static qu_method _method_new_method (qu_method x, int argc, qu_any *argv)
{
	qu_method m = qu_method_new (x->sub, x->self);
	qu_array  a = 0;

	if (qu_flag_test (x, QU_METHOD_FLAG_SETSELF))
		qu_flag_set (m, QU_METHOD_FLAG_SETSELF);

	if (x->args)
		a = qu_array_copy (x->args);

	if (argc)
	{
		if (a == 0)
			a = qu_array_new (argc);

		while (argc--)
			qu_array_push (a, *argv++);
	}

	m->args = a;
	return m;
}



static qu_method qu__method_new (qu_class self, int argc, qu_any *argv)
{
	qu_any    x, theself = QuNil;
	int       set = 0;
	qu_method m;

	QU_ARG_VMIN (1);
	x = argv[0];

	if (qu_sub_is (x))
	{
		QU_ARG_CHECKZ (!qu_sub_is_inst (x), "cant create Method on instance method");

		if (qu_sub_is_class (x))
			theself = qu_sub_get_class (x);
		else
			set = 1;

		--argc;
		++argv;
	}
	else if (qu_method_is (x))
	{
		return _method_new_method (x, argc - 1, argv + 1);
	}
	else
	{
		theself = x;

		QU_ARG_VMIN (2);
		x = argv[1];

		if (qu_string_is (x))
		{
			qu_class c = qu_classof (theself);

			x = c->get (theself, (qu_string) x, 0);
			QU_ENSURE (x);
			QU_REQUIREZ (QuEImplement, qu_sub_is (x), "not a Sub");
			if (qu_sub_is_class (x))
			{
				if (!qu_class_is (theself))
					theself= qu_sub_get_class (x);
				else
					set = 1;
			}
			else if (qu_sub_is_inst (x))
			{
				QU_ARG_CHECKZ (qu_object_instof (theself, (qu_class) ((qu_sub) x)->parent), "invalid instance object");
			}
		}
		else
		{
			QU_ARG_CHECKZ (qu_sub_is (x) && !qu_sub_is_inst (x), "Sub or String argument expected");
		}

		argc -= 2;
		argv += 2;
	}

	m = qu_method_new (x, theself);
	if (set)
		qu_flag_set (m, QU_METHOD_FLAG_SETSELF);

	if (argc)
	{
		qu_array  a = qu_array_new (argc);
		qu_any   *p = a->objs;

		a->len = argc;
		while (argc--)
			*p++ = *argv++;
		m->args = a;
	}
	return m;
}



/* QU_WARN_EXPERIMENTAL - currying */
qu_method qu_method_curry_compile_time (qu_any x, qu_ast list)
{
	qu_any    theself = QuNil;
	qu_method m;

	QU_REQUIREZ (QuECompile, qu_sub_is (x) && !qu_sub_is_inst (x), "invalid curry");

	if (qu_sub_is_class (x))
		theself = qu_sub_get_class (x);

	m = qu_method_new (x, theself);
	if (theself == QuNil)
		qu_flag_set (m, QU_METHOD_FLAG_SETSELF);

	if (list)
	{
		long n = list->op;
		if (n)
		{
			qu_array  a = qu_array_new (n);
			qu_any   *p = a->objs;
			qu_ast    c;

			a->len = n;
			for (c = list->child; c; c = c->next)
			{
				QU_REQUIREZ (QuECompile, c->type == QU_AST_literal
					|| ((c->type == QU_AST_name || c->type == QU_AST_ref)
						&& qu_var_is (c->thing)), "non-literal curry argument");
				*p++ = c->thing;
			}
			m->args = a;
		}
	}
	return m;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__method__mark (qu_method self)
{
	qu_gc_mark  (self->sub);
	qu_gc_mark  (self->self);
	qu_gc_xmark (self->args);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __cmp (x) -> Int
	If v{x} is a Method then compare otherwise delegate to -{Object.__cmp}.

	This method compares the -{Sub} objects first, then the self objects.

	@xmp
		a = 'foo'
		b = 'bar'
		f = Method (a, "quote")
		g = Method (b, "quote")
		println (f <=> g)

	@out
		0
*/
static qu_int qu__method__cmp (qu_method self, qu_any x)
{
	if (qu_method_is (x))
	{
		if (self->sub == ((qu_method) x)->sub)
			return qu_object_default_cmp (self->self, ((qu_method) x)->self);
		return qu_object_default_cmp (self->sub, ((qu_method) x)->sub);
	}
	return qu_object_default_cmp (self, x);
}



/*) sub __getitem (name) -> Any
	See -{Sub.__getitem}.
*/
static qu_any qu__method__getitem (qu_method self, qu_string name)
{
	return qu__sub__getitem (self->sub, name);
}



/*) sub __getslice (name, defval) -> Any
	See -{Sub.__getslice}.
*/
static qu_any qu__method__getslice (qu_method self, qu_string name, qu_any defval)
{
	return qu__sub__getslice (self->sub, name, defval);
}


/*) sub __like (x) -> Boolean
	If v{x} is a Method then this method returns true if both Sub objects
	are the same, false otherwise.

	@noexc

	@xmp
		a = 'foo'
		b = 'bar'
		f = Method (a, "quote")
		g = Method (b, "quote")
		println (f === g)

	@out
		true
*/
static qu_boolean qu__method__like (qu_method self, qu_any x)
{
	if (qu_method_is (x))
		return (self->sub == ((qu_method) x)->sub) ? QuTrue : QuFalse;
	return QuFalse;
}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@noexc

	@xmp
		a = [1 2 3]
		f = Method (a, "push")
		println (f)

	@out
		<Method sys.Array.push>
*/
static qu_null qu__method__print (qu_method self)
{
	qu_string s = qu_method_as_string (self);
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/* sub __push (x) -> self
	Call the qu_method with one argument then return it.

	@exc
	.qu_exception anything thrown by the qu_sub

	@xmp
		sub foo (a b)
			println (a ' ' b)
		;;

		f = qu_method (@foo, 'hello')
		f <<< 'world' <<< 'there'

	@out
		hello world
		hello there
*/
static qu_method qu__method__push (qu_method self, qu_any x)
{
	QU_ENSURE (qu_method_call1 (self, x));
	return self;
}



/*) sub __setitem (name, val) -> Any
	See -{Sub.__setitem}.
*/
static qu_any qu__method__setitem (qu_method self, qu_string name, qu_any val)
{
	return qu__sub__setitem (self->sub, name, val);
}



/*) sub __string -> String
	Return a string representation of self.

	@noexc

	@xmp
		a = [1 2 3]
		f = Method (a, "push")
		s = f as String
		println (s)

	@out
		<Method sys.Array.push>
*/
qu_string qu__method__string (qu_method self)
{
	return qu_object_where_string ((qu_symbol) self->sub, "<Method ", ">");
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub argcheck (*arg) -> self
	This does the same as #{argmatch} except that if the arguments do not
	match then an exception is thrown.

	Unlike #{argmatch}, if the argument contains validators then the validators
	are actually called.

	@exc
	.EArgValue
	.EArgCount
	.EAbstract
*/
static qu_method qu__method_argcheck (qu_method self, int argc, qu_any *argv)
{
	QU_ENSURE (qu_sub_arg_check (self->sub, argc, argv, 1));
	return self;
}



/*) sub argmatch (*arg) -> Int
	Return 0 if v{arg} is a valid argument list for this Sub, otherwise:
	@ul
	.-1      insufficient argument
	.integer position of the invalid argument, first is 1

	@noexc

	@xmp
		a = [1 2 3]
		f = Method (a, "push")
		println (f.argmatch (1, 2, 3))

	@out
		0
*/
static qu_int qu__method_argmatch (qu_method self, int argc, qu_any *argv)
{
	return qu__sub_argmatch (self->sub, argc, argv);
}



/*) sub argnames -> Any
	See -{Sub.argnames}.
*/
static qu_any qu__method_argnames (qu_method self)
{
	return qu__sub_argnames (self->sub);
}



/*) sub argtypes -> Any
	See -{Sub.args}.
*/
static qu_any qu__method_argtypes (qu_method self)
{
	return qu__sub_argtypes (self->sub);
}



/*) sub code -> Code
	Return the -{Code} object of this Method.

	This simply calls C{self.subof.code}.
	See -{Sub.code} for details.
*/
static qu_code qu__method_code (qu_method self)
{
	return qu__sub_code (self->sub);
}



/*) sub classof -> Class
	Return the Class object where the Sub is declared.

	This simply calls C{self.subof.classof}.
	See -{Sub.classof} for details.
*/
static qu_class qu__method_classof (qu_method self)
{
	qu_class c = qu_sub_get_class (self->sub);
	QU_REQUIREZ (QuENone, c, "not a method");
	return c;
}



/*) sub ident -> String
	Return the identifier name of the Sub.

	This simply calls C{self.subof.ident}.
	See -{Sub.classof} for details.
*/
static qu_string qu__method_ident (qu_method self)
{
	return self->sub->name; /* cheat */
}



/*) sub isclass -> Boolean
	Return true if the Sub is a class method, false otherwise.

	This simply calls C{self.subof.isclass}.
	See -{Sub.isclass} for details.
*/
static qu_boolean qu__method_isclass (qu_method self)
{
	return qu_sub_is_class (self->sub) ? QuTrue : QuFalse;
}



/*) sub isextern -> Boolean
	Return true if the Sub is a C Method.

	This simply calls C{self.subof.isextern}.
	See -{Sub.isextern} for details.
*/
static qu_boolean qu__method_isextern (qu_method self)
{
	return qu_sub_is_extern (self->sub) ? QuTrue : QuFalse;
}



/*) sub isfunction -> Boolean
	Return true if the Sub is a module function, false otherwise.
	The result of this method is the reverse of #{ismethod}.

	This simply calls C{self.subof.isfunction}.
	See -{Sub.isfunction} for details.
*/
static qu_boolean qu__method_isfunction (qu_method self)
{
	return qu_sub_get_class (self->sub) ? QuFalse : QuTrue;
}



/*) sub isinst -> Boolean
	Return true if the Sub is an instance method, false otherwise.

	This simply calls C{self.subof.isinst}.
	See -{Sub.isinst} for details.
*/
static qu_boolean qu__method_isinst (qu_method self)
{
	return qu_sub_is_inst (self->sub) ? QuTrue : QuFalse;
}



/*) sub ismethod -> Boolean
	Return true if the Sub is a method, false otherwise.
	The result of this method is the reverse of #{isfunction}.


	This simply calls C{self.subof.ismethod}.
	See -{Sub.isinst} for details.
*/
static qu_boolean qu__method_ismethod (qu_method self)
{
	return qu_sub_get_class (self->sub) ? QuTrue : QuFalse;
}



/*) sub isstatic -> Boolean
	Return true if the Sub is static, false otherwise.

	This simply calls C{self.subof.isstatic}.
	See -{Sub.isstatic} for details.
*/
static qu_boolean qu__method_isstatic (qu_method self)
{
	return qu_sub_is_static (self->sub) ? QuTrue : QuFalse;
}



/*) sub selfof -> Any
	Get the self object bound to this Method.

	@noexc

	@xmp
		a = [1 2 3]
		f = Method (a, "push")
		println (f.selfof)

	@out
		[1, 2, 3]
*/
static qu_any qu__method_selfof (qu_method self)
{
	return self->self;
}



/*) sub subof -> Sub
	Return the -{Sub} object of this Method.

	@exc
	.ESecurity

	@xmp
		a = [1 2 3]
		f = Method (a, "push")
		println (f.subof)

	@out
		<Sub sys.Array.push>

	@warn
	Use with care. It is possible to deceive Qu and call the method with
	the wrong instance if you save the value returned by this method in
	an instance attribute.

	@xmp
		class Foo
			var met
			sub __init
				x = Method ([], "push")
				self.met = x.subof
			;;
		;;

		x = Foo()
		x.met (0)

	@p
	The last line calls the Array.push method with the wrong instance
	i.e. v{x} wich is not an Array. If your program crashes, you are lucky.
	If not, you might be looking for a very difficult bug to locate.

	If you are clever, however, this might be a magical door.
*/
static qu_sub qu__method_subof (qu_method self)
{
	__qu_unsafe__;
	return self->sub;
}



/*) sub typeof -> Any
	Return the type returned by this Method.

	@exc
	.ENone
*/
static qu_any qu__method_typeof (qu_method self)
{
	QU_REQUIREZ (QuENone, self->sub->type, "Method not typed");
	return self->sub->type;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

qu_method qu_method_new (qu_sub sub, qu_any self)
{
	register qu_method m;

	m       = qu_gc_new (QuMethodClass);
	m->sub  = sub;
	m->self = self;
	m->args = 0;
	return m;
}


qu_method qu_method_new_static (qu_sub sub, qu_any self)
{
	register qu_method m;

	m       = qu_gc_new_static (QuMethodClass);
	m->sub  = sub;
	m->self = self;
	m->args = 0;
	return m;
}


qu_any qu_method_calln (qu_method self, int argc, qu_any *argv)
{
	qu_any theself;

	if (qu_sub_is_class (self->sub))
	{
		theself = self->sub->parent;
	}
	else if (qu_sub_is_inst (self->sub))
	{
		theself = self->self;
		/* assert self->self is an instance of (qu_class) self->sub->parent */
	}
	else
	{
		if (qu_flag_test (self, QU_METHOD_FLAG_SETSELF))
			theself = qu_interp_self ();
		else
			theself = self->self;
	}

	if (self->args)
	{
		qu_array g = self->args;
		qu_any  *p = g->objs;
		long     l = g->len;
		long     n = l + argc;

#define MAXCSTACKARG 32
		if (n <= MAXCSTACKARG)
		{
			qu_any args [MAXCSTACKARG];
			qu_any *q = args;

			while (l--)
			{
				qu_any x = *p++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			while (argc--)
			{
				qu_any x = *argv++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			return qu_sub_calln (self->sub, theself, n, args);
		}
		else
		{
			qu_array a = qu_array_new (n);
			qu_any  *q = a->objs;

			while (l--)
			{
				qu_any x = *p++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			while (argc--)
			{
				qu_any x = *argv++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			a->len = n;
			return qu_sub_calln (self->sub, theself, n, q);
		}
	}
	return qu_sub_calln (self->sub, theself, argc, argv);
}



qu_any qu_method_call4 (qu_method self, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv[4] = {a1, a2, a3, a4};
	return qu_method_calln (self, 4, argv);
}



qu_any qu_method_call3 (qu_method self, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv[3] = {a1, a2, a3};
	return qu_method_calln (self, 3, argv);
}



qu_any qu_method_call2 (qu_method self, qu_any a1, qu_any a2)
{
	qu_any argv[2] = {a1, a2};
	return qu_method_calln (self, 2, argv);
}



qu_any qu_method_call1 (qu_method self, qu_any a1)
{
	return qu_method_calln (self, 1, &a1);
}



qu_any qu_method_call0 (qu_method self)
{
	return qu_method_calln (self, 0, 0);
}



/* FIXME - duplicate code w/ qu_method_calln */
qu_any qu_method_calln_attr (qu_method self, qu_any theself, int argc, qu_any *argv)
{
	if (qu_flag_test (self->sub, QU_SUB_FLAG_CLASS | QU_SUB_FLAG_INST)
			|| !qu_flag_test (self, QU_METHOD_FLAG_SETSELF))
		return qu_method_calln (self, argc, argv);

	if (self->args)
	{
		qu_array g = self->args;
		qu_any  *p = g->objs;
		long     l = g->len;
		long     n = l + argc;

		if (n <= MAXCSTACKARG)
		{
			qu_any args [MAXCSTACKARG];
			qu_any *q = args;

			while (l--)
			{
				qu_any x = *p++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			while (argc--)
			{
				qu_any x = *argv++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			return qu_sub_calln (self->sub, theself, n, args);
		}
		else
		{
			qu_array a = qu_array_new (n);
			qu_any  *q = a->objs;

			while (l--)
			{
				qu_any x = *p++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			while (argc--)
			{
				qu_any x = *argv++;
				if (qu_var_is (x))
					*q++ = ((qu_var) x)->value;
				else
					*q++ = x;
			}
			a->len = n;
			return qu_sub_calln (self->sub, theself, n, q);
		}
	}
	return qu_sub_calln (self->sub, theself, argc, argv);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_method__setup ()
{
	qu_class_setup_final (QuMethodClass, qu_method_s, 0, qu__method__mark);
}



void qu_method__init ()
{
	/* standard class methods */
	qu_class_method_class (QuMethodClass, QU_NAME_new, qu__method_new, -1);

	/* standard instance methods */
	qu_class_method_inst  (QuMethodClass, QU_NAME__cmp     , qu__method__cmp     , 1);
	qu_class_method_inst  (QuMethodClass, QU_NAME__getitem , qu__method__getitem , 1);
	qu_class_method_inst  (QuMethodClass, QU_NAME__getslice, qu__method__getslice, 2);
	qu_class_method_inst  (QuMethodClass, QU_NAME__like    , qu__method__like    , 1);
	qu_class_method_inst  (QuMethodClass, QU_NAME__print   , qu__method__print   , 0);
	qu_class_method_inst  (QuMethodClass, QU_NAME__push    , qu__method__push    , 1);
	qu_class_method_inst  (QuMethodClass, QU_NAME__setitem , qu__method__setitem , 2);
	qu_class_method_inst  (QuMethodClass, QU_NAME__string  , qu__method__string  , 0);

	/* specific instance methods */
	qu_class_method_instz (QuMethodClass, "argcheck"  , qu__method_argcheck  , -1);
	qu_class_method_instz (QuMethodClass, "argmatch"  , qu__method_argmatch  , -1);
	qu_class_method_instz (QuMethodClass, "argnames"  , qu__method_argnames  ,  0);
	qu_class_method_instz (QuMethodClass, "argtypes"  , qu__method_argtypes  ,  0);
	qu_class_method_instz (QuMethodClass, "classof"   , qu__method_classof   ,  0);
	qu_class_method_instz (QuMethodClass, "code"      , qu__method_code      ,  0);
	qu_class_method_instz (QuMethodClass, "ident"     , qu__method_ident     ,  0);
	qu_class_method_instz (QuMethodClass, "isclass"   , qu__method_isclass   ,  0);
	qu_class_method_instz (QuMethodClass, "isextern"  , qu__method_isextern  ,  0);
	qu_class_method_instz (QuMethodClass, "isfunction", qu__method_isfunction,  0);
	qu_class_method_instz (QuMethodClass, "isinst"    , qu__method_isinst    ,  0);
	qu_class_method_instz (QuMethodClass, "ismethod"  , qu__method_ismethod  ,  0);
	qu_class_method_instz (QuMethodClass, "isstatic"  , qu__method_isstatic  ,  0);
	qu_class_method_instz (QuMethodClass, "selfof"    , qu__method_selfof    ,  0);
	qu_class_method_instz (QuMethodClass, "subof"     , qu__method_subof     ,  0);
	qu_class_method_instz (QuMethodClass, "typeof"    , qu__method_typeof    ,  0);

	/* finalize declaration */
	qu_class_set_super (QuMethodClass, QuFunctionClass);
	qu_class_enddecl   (QuMethodClass);
}
