/* Module.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/*
	define to use weak references
	this might be a command line option in the future
*/
#undef QU_APP_WEAKMODULE


/*) class Module __final
	The main program, modules, and k{eval} are all instances of the Module class.

	The main program, the builtin -{Sys} and all non-eval module live in memory
	forever until it Qu exits. Eval modules only live as long as they are referenced.

	The main program Module object is always available via the builtin -{__main__}
	function. The k{me} keyword reveals the Module object where the keyword appears.

	@xmp
		println (me)         # this module
		println (__main__)   # the main program

		s =
		`
			println ('module initialized')
			class Foo
				println ('class initialized')
				var bar
				sub __init (x)
					self.bar = x
				;;
			;;
		`
		mod = eval (s, "MT")
		println (mod)

		Foo = mod.Foo
		println (Foo)
		x = Foo ('fooey')
		println (x)

	@out
		<Module __main__>
		<Module __main__>
		module initialized
		class initialized
		<Module eval0>
		<Class eval0.Foo>
		<a Foo>

	@note
	Methods of the Module class are named with a leading double underscore.
	You should never name your class/function/methods with a leading double
	underscore or you will not be able to access the Module methods.

	@p
	If you define a function named f{__init} then it will be automatically
	called by Qu upon when the module is initialized. Likewise, if you define
	a function named f{__destroy} then it will be automatically called by
	Qu upon exit.

	You should always tag these function with the C{__once} decorator to
	make sure that they are only called once.

	@warn
	In the main program, you must call C{__init} yourself if there is one,
	because it does not make much sense for Qu to automatically call it.

	@xmp
		# module Foo

			sub __init __once
				println (__where__)
			;;

			sub __destroy __once
				println (__where__)
			;;


		# main program

			use Foo

			println ('Hello')

			sub __init __once
				println (__where__)
			;;

			sub __destroy __once
				println (__where__)
			;;

	@out
		Foo.__init
		Hello
		Foo.__destroy
		__main__.__destroy
*/
qu_class QuModuleClass;


/* the library directory, e.g. '/opt/qu/lib' -- without the terminating slash */
qu_string QuModuleDir;


/* Qu module file extension, e.g. '.qm' -- includes the dot */
qu_string QuModuleExt;
#ifndef MOD_EXT
#define MOD_EXT ".qm"
#endif


/* shared library file extension, e.g. '.so' -- includes the dot */
qu_string QuModuleDlExt;
#ifndef MOD_DLEXT
#define MOD_DLEXT ".so"
#endif


/*
	The main program Module object.
	It is created at early initialization phase.
*/
qu_module QuModuleMain;


/*
	C init function prototype for external modules.
*/
typedef int (*qu_dlinit_f) (qu_module self);


/*
	Modules, main program, and eval blocks are all Module objects.

	Modules (e.g. "/foo/bar.qm", etc.) are shared thus loaded, compiled,
	and initialized once only, freed when no longer referenced.

	Residents is a Dict of Weak references to Module objects.
*/
static qu_dict Residents = NULL;

#ifdef GC_IS_MULTITHREADED
static qu_thread_mutex_t ResidentsLock = QU_THREAD_MUTEX_INITIALIZER;
#define LOCK_RESIDENTS   qu_thread_mutex_lock   (&ResidentsLock)
#define UNLOCK_RESIDENTS qu_thread_mutex_unlock (&ResidentsLock)
#else
#define LOCK_RESIDENTS
#define UNLOCK_RESIDENTS
#endif

#ifdef QU_HAS_UNIVERSE
qu_dict QuUniverses;
#endif


/* strings */
static qu_string SRegType;  /* f    - regular file filter */
static qu_string SModPat;   /* *.qm - module file pattern */
static qu_string SSoPat;    /* *.so - shared library file pattern */

/*----------------------------------------------------------------------------
	dl
----------------------------------------------------------------------------*/

void *qu_dl_open (const char *filename, int flag)
{
	return dlopen (filename, flag);
}


qu_string qu_dl_config (const char *name)
{
	static qu_string conf = 0;
	qu_string  c;
	char      *s, *p;

	if (conf == 0)
		conf = qu_string_new_staticz (QU_APP_PREFIX QU_PATH_SEPSTR QU_APP_PACKAGE QU_PATH_SEPSTR "qu.conf");

	c = qu_file_load (conf);
	QU_REQUIREF (QuESystem, c, "cant load ", conf->str);
	s = c->str;
	while (*s)
	{
		while (*s && *s != '\n')
			++s;
		++s;
		if (!qu_ascii_isalpha (*s))
			continue;
		p = (char *) name;
		while (*s == *p)
		{
			++s;
			++p;
		}
		if (*p == '\0')
		{
			while (qu_ascii_isspace (*s))
				++s;
			if (*s != '=')
				continue;
			++s;
			while (qu_ascii_isspace (*s))
				++s;
			p = s;
			while (!qu_ascii_isspace (*s))
				++s;
			*s = '\0';
			return qu_string_newz (p);
		}
	}
	QU_THROWF (QuESystem, "cant find %s in qu.conf", name);
}



void *qu_dl_open_ex (const char *name, int flag)
{
	qu_string  p = qu_dl_config (name);
	void      *h;

	QU_ENSURE (p);
	h = dlopen (p->str, flag);
	if (h == 0)
		QU_THROWF (QuESystem, "cant open %s", dlerror ());
	return h;
}



void *qu_dl_error ()
{
	return dlerror ();
}


void *qu_dl_sym (void *handle, char *symbol)
{
	return dlsym (handle, symbol);
}


int qu_dl_close (void *handle)
{
	return dlclose (handle);
}



/* get a C function */
qu_any qu_module_get_cfunc (qu_module self, void *handle, qu_string name, qu_string spec, qu_dict resolved)
{
	qu_any m = qu_dict_getitem (resolved, name);

	if (m == 0)
	{
		void *h = qu_dl_sym (handle, name->str);

		if (h == 0)
			QU_THROWF (QuEImplement, "symbol %s not found", name->str);

		m = qu_dlib_cfunc_static (h, name, spec);
		if (m == 0)
			QU_THROWF (QuEValue, "%s: %s\n", name->str, qu_thread_err_string ()->str);

		qu_dict_setitem (resolved, qu_string_to_static (name), m);
	}
	return m;
}



/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

/*
	Get/set attribute -- no questions asked
*/
#define GETATTR(self, name)         qu_attr_find ((self)->attrs, name)
#define SETATTR(self, name, symbol) qu_attr_set  ((self)->attrs, name, symbol)


/* must have LOCK_RESIDENT */
#ifdef QU_APP_WEAKMODULE
qu_module module_get_resident (qu_string path)
{
	qu_weak w = qu_dict_getitem (Residents, path);

	if (w)
	{
		assert (qu_weak_is (w));
		if (w->ref)
			return (qu_module) w->ref;
		qu_dict_del (Residents, path);
	}
	return 0;
}
#else
#define module_get_resident(path) qu_dict_getitem (Residents, path)
#endif


/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

static int ModuleAfterSetup = 0;

static qu_module module_alloc (qu_string name, long flags, int root)
{
	qu_module m;

	/* can't use anything fancy here since we might be in an early init stage */

	m           = root ? qu_gc_new_root (QuModuleClass) : qu_gc_new (QuModuleClass);
	m->name     = ModuleAfterSetup ? qu_file_strip (name) : name;
	m->parent   = 0;
	m->flags   |= flags;
	m->attrs    = qu_attr_alloc ();
	m->literals = 0;
	m->use      = 0;
	m->dlhandle = 0;
	m->code     = 0;
	m->source   = 0;
	m->tree     = 0;
	m->doc      = 0;
	m->getcfunc = 0;
#ifdef QU_HAS_UNIVERSE
	m->universes = 0;
#endif
	return m;
}



/*----------------------------------------------------------------------------
	internals
----------------------------------------------------------------------------*/

/*
	From path spelled in the "use" statement to absolute path.

	use "/foo/Bar"  -> /foo/Bar
	use "Foo/Bar"   -> /opt/qu/lib/Foo/Bar
	use "../Foo"    ->  need the module's dir to get the absolute path
*/
static qu_string module_use_toabs (qu_string dir, qu_string use, int package)
{
	if (qu_file_is_abs (use))
	{
		return use;
	}
	else
	{
		if (use->str[0] == '.' && (use->str[1] == '/' || use->str[1] == '.'))
			return qu_file_toabs (use, dir);

		if (package)
		{
			assert (QuModuleMain);
			if (QU_APP_USEPACKAGE)
			{
				static qu_string p = 0;
				if (p == 0)
					p = qu_string_new_staticz (QU_APP_USEPACKAGE);
				return qu_file_toabs (use, p);
			}
			return qu_file_toabs (use, qu_file_dir (QuModuleMain->name));
		}

		return qu_file_toabs (use, QuModuleDir);
	}
}



/*
	Add Module object and attribute.
*/
void qu_module_add_use (qu_module self, qu_module m, qu_string as)
{
	if (self->use == 0)
		self->use = qu_set_new (8);
	qu_set_unfreeze (self->use);
	qu_set_push (self->use, m);
	qu_set_freeze (self->use);
	SETATTR (self, as, m);
}



/*
	Load one module.
*/
static qu_module module_load_one (qu_string abspath, int depth)
{
	qu_module  m;
	void      *h;
	qu_string  p, s = 0;

	if (depth > 2)
		QU_THROWF (QuEValue, "module %s has cyclic dependency", abspath->str);

	/*
		Try load the .so module.
	*/
	p = qu_file_repext (abspath, QuModuleDlExt);
	h = dlopen (p->str, RTLD_NOW | RTLD_GLOBAL);

	if (h)
	{
		/*
			Also load the Qu module if there is one.
		*/
		if (qu_file_exist (abspath, SRegType))
		{
			s = qu_file_load (abspath);
			if (s == 0)
			{
				dlclose (h);
				QU_RETHROW;
			}
		}
		else
		{
			p = qu_file_repext (abspath, QuModuleExt);
			if (qu_file_exist (p, SRegType))
			{
				s = qu_file_load (p);
				if (s == 0)
				{
					dlclose (h);
					QU_RETHROW;
				}
			}
		}
	}
	else
	{
		/*
			If the .so module exists but could not be loaded then something is
			wrong with it. Just throw an ESystem.
		*/
		if (qu_file_exist (p, SRegType))
			QU_THROWZ (QuESystem, dlerror ());

		/*
			Load the Qu module.
		*/
		if (qu_file_exist (abspath, SRegType))
		{
			s = qu_file_load (abspath);
		}
		else
		{
			p = qu_file_repext (abspath, QuModuleExt);
			s = qu_file_load (p);
		}
		if (s == 0)
			QU_RETHROW;
	}

	/*
		Now we can create the Module object.
	*/
	m = module_alloc (abspath, QU_MODULE_FLAG_RESIDENT | QU_MODULE_FLAG_TRUSTED, 0);
	m->literals = QuSysModule->literals;

	/*
		If the Qu module does not exist then no compilation is required.
		Otherwise m->source contains the source code.
	*/
	m->source = s;
	if (s == 0)
		m->flags |= QU_MODULE_FLAG_PARSED | QU_MODULE_FLAG_COMPILED | QU_MODULE_FLAG_READY | QU_MODULE_FLAG_INIT;

	if (h)
	{
		/*
			If the .so module exists, call the init function.

			The init function may or may not exist.
			The name must be qu_xxx_init where xxx is the filename.

			If it exists it must raise an exception in case of failure.

				.../foo/Bar.so

				int qu_Bar_init (qu_module m)
				{
					...

					if (failed-to-initialize)
						QU_THROW (...

					return 1;
				}
		*/
		qu_string name = qu_file_name (m->name);

		if (name->len <= QU_OBJECT_MAXIDENTLEN)
		{
			char        buf [QU_OBJECT_MAXIDENTLEN + 16];
			qu_dlinit_f ini;

			buf [0] = 'q';
			buf [1] = 'u';
			buf [2] = '_';
			memcpy (buf + 3, name->str, name->len);
			strcpy (buf + 3 + name->len, "_init");

			ini = (qu_dlinit_f) dlsym (h, buf);
			if (ini)
			{
				if (ini (m) == 0)
				{
					dlclose (h);
					QU_RETHROW;
				}
			}
		}
	}
	return m;
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class list -> Set
	Return a -{Set} of all static modules loaded. This does not include the
	main program.

	@noexc

	@xmp
		use Foo
		println (Module.list)

	@out
		{<Module Foo>, <Module Sys>}
*/
static qu_set qu__module_list (qu_class self)
{
#ifdef QU_APP_WEAKMODULE
	qu_set  e = qu_set_new (Residents->len);
	long    b, i;
	qu_any  k;
	qu_weak w;

	b = i = 0;
	while ((w = (qu_weak) qu_dict_walk (Residents, &k, &b, &i)) != 0)
	{
		if (w->ref)
			qu_set_push (e, w->ref);
	}
	return e;
#else
	return qu_dict_as_set (Residents);
#endif
}



/*) sub class loaded (path: String) -> Module
	Get a loaded module.

	v{path} is the module path. If relative then it is relative to the
	Qu's C{lib} directory.

	If the module is loaded then return that module.
	Otherwise ENone is thrown.

	@exc
	.EArgValue
	.ENone     module not loaded

	@xmp
		use Time

		m = Module.loaded ('Time')
		assert (m == Time)
*/
static qu_module qu__module_loaded (qu_class self, qu_string path)
{
	qu_module m;

	QU_ARG_TYPE (path, qu_string);
	path = qu_file_make (path, QuModuleDir, 0);
	m = qu_module_loaded (qu_file_strip (path));
	QU_REQUIRE (QuENone, m);
	return m;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__module__destroy (qu_module self)
{
	if (self->dlhandle)
		dlclose (self->dlhandle);

	if (QU_APP_VERBOSE)
		printf ("module freed: %p %s\n", self, self->name->str);
}



static void qu__module__mark (qu_module self)
{
	qu_gc_xmark (self->parent);
	qu_gc_mark  (self->name);
	qu_gc_xmark (self->literals);
	qu_gc_xmark (self->use);
	qu_gc_xmark (self->code);
	qu_gc_xmark (self->source);
	qu_gc_xmark (self->tree);
	qu_gc_xmark (self->doc);
#ifdef QU_HAS_UNIVERSE
	qu_gc_xmark (self->universes);
#endif
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __cmp (x) -> Int
	If v{x} is not a Module then delegate to -{Object.__cmp}. Otherwise the
	compare the #{__kind} (main program > static module > eval) then the
	#{__path}.

	@noexc

	@xmp
		use Foo

		a = __main__
		b = Foo

		println (a <=> b)

	@out
		-1
*/
static qu_int qu__module__cmp (qu_module self, qu_any x)
{
	if (qu_module_is (x))
	{
		unsigned long a = self->flags & (QU_MODULE_FLAG_RESIDENT | QU_MODULE_FLAG_EVAL);
		unsigned long b = ((qu_module) x)->flags & (QU_MODULE_FLAG_RESIDENT | QU_MODULE_FLAG_EVAL);

		if (a == b)
			return qu_string_cmpint (self->name, ((qu_module) x)->name);
		if (a < b)
			return QuInt_1;
		return QuInt1;
	}
	return qu_object_default_cmp (self, x);
}



/*) sub __string -> String
	Return a string representation of self.

	@noexc

	@xmp
		s = __main__ as String
		println (s)

	@out
		<Module __main__>
*/
static qu_string qu__module__string (qu_module self)
{
	return qu_object_where_string ((qu_symbol) self, "<Module ", ">");
}



/*) sub __print
	Print this module to -{Stream.stdout}.

	@exc
	.Exception as thrown by -{print}.
*/
static qu_null qu__module__print (qu_module self)
{
	qu_string s = qu__module__string (self);
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub __code -> Code
	Get the Code object of this module.

	@exc
	.ENone module does not have a Code object or has been deleted

	@note
	External modules do not have Code objects.

	The interpreter deletes the Code object once the module has been
	initialized, except of course for the main program.

	@xmp
		println (__main__.__code)

	@warn
	The output depends on the compiler.

	@out
		 0: LITERAL      <Sub Sys.println>
		 5: LITERAL      <Sub Sys.__main__>
		10: CALL         0 0
		19: GETATTR      __code
		24: AUTOATTR
		25: CALL         1 0
		34: RESTORE
		35: LITERAL      nil
		40: RETURN
*/
static qu_code qu__module__code (qu_module self)
{
	QU_REQUIRE (QuENone, self->code);
	return self->code;
}



/*) sub __docdict -> Dict
	Return a -{Dict} containing the documentation info of the module.
	This is only available if the module was loaded via an eval module
	with documentation strings enabled or with the -D option.

	The Dict will contain all objects in the module that defines a
	documentation string.

	The keys are the objects and the values are -{Array}s of two items:
	(the declaration string, the documentation string).

	@exc
	.ENone module does not have documentation info

	@xmp
		# run with the -D option

		"""The Foo module"""

		sub foo (x)
			"""This function prints x."""
			println ('Foo!')
		;;

		for k, v in __main__.__docdict
			println (k ' : ' v)
		;;

	@out
		<Module __main__> : ['', 'The Foo module']
		<Sub __main__.foo> : ['sub foo (x)', 'This function prints x.']
*/
static qu_dict qu__module__docdict (qu_module self)
{
	QU_REQUIRE (QuENone, self->doc);
	return self->doc;
}



/*) sub __getitem (name: String) -> Any
	Get a reference to a module object.

	This works like #{__getslice} except that ENone is thrown if the symbol does
	not exist.

	@exc
	.EArgValue
	.ESecurity
	.ENone

	@xmp
		use Foo

		_foo = 10
		println (__main__['_foo'])
		println (__main__['Foo'])

	@out
		<Var __main__._foo>
		<Module Foo>
*/
static qu_symbol qu__module__getitem (qu_module self, qu_string name)
{
	qu_any x;

	__qu_unsafe__;

	QU_ARG_TYPE (name, qu_string);
	x = qu_attr_find (self->attrs, name);
	QU_REQUIRE (QuENone, x);
	return x;
}



/*) sub __getslice (name: String, defval) -> Any
	Get a reference to a module object.

	Unlike the -{sysref} function, this method works for all objects, private,
	local, etc.

	If the object does not exist, then v{defval} is returned.

	@exc
	.EArgValue
	.ESecurity

	@xmp
		use Foo

		_foo = 10
		println (__main__['_foo'])
		println (__main__['Foo'])

	@out
		<Var __main__._foo>
		<Module Foo>
*/
static qu_any qu__module__getslice (qu_module self, qu_string name, qu_any defval)
{
	qu_any x;

	__qu_unsafe__;

	QU_ARG_TYPE (name, qu_string);
	x = qu_attr_find (self->attrs, name);
	return x ? x : defval;
}



/*) sub __kind -> String
	Return the kind of the module:
	@ul
	.C{'_'} : main program
	.C{'m'} : static module
	.C{''}  : (empty string) eval module

	@noexc

	@xmp
		use Foo

		println (__main__.__kind)
		println (Foo.__kind)

	@out
		_
		m
*/
static qu_string qu__module__kind (qu_module self)
{
	if (qu_module_is_resident (self))
		return qu_string_byte ('m');
	if (self == QuModuleMain)
		return qu_string_byte ('_');
	return QuStringEMPTY;
}



/*) sub __path (*path) -> String
	@def
	sub __path -> String
	sub __path (path: String) -> String

	@p
	If called with one argument and the module is an eval module then sets
	the path. This may be needed for documentation processors.

	Otherwise return the absolute path of the module without the extension
	part. If the module is an eval module then a string in the form of
	C{"evalX"} where v{X} is an arbitrary number is returned.
	If the module is the builtin %{Sys} module then C{'Sys'} is returned.

	@exc
	.EArgValue
	.ECall      can only set eval module

	@xmp
		println (__main__.__path)

	@out
		/opt/qu/demo/test
*/
static qu_string qu__module__path (qu_module self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		qu_string p = argv[0];

		QU_ARG_TYPE (p, qu_string);
		QU_REQUIRE (QuECall, qu_module_is_eval (self));
		self->name = p;
	}
	return self->name;
}



/*) sub __callable
	Get class or function which has the same name as the module.

	@exc
	.EImplement
*/
static qu_symbol qu__module__callable (qu_module self)
{
	return qu_module_getattr_callable_name (self, 1);
}



/*) sub __stab (all = false) -> Array
	Return an -{Array} containing attributes defined by this module
	(-{Class}, -{Sub} and -{Var} objects).

	If v{all} is true then include private symbols as well.

	@exc
	.ESecurity

	@xmp
		var _foo = 10

		sub bar
			return 1
		;;

		for y in __main__.__stab (true)
			println (y)
		;;

	@out
		<Var __main__._foo>
		<Sub __main__.bar>
*/
static qu_array qu__module__stab (qu_module self, int argc, qu_any *argv)
{
	long      all = QU_SYMBOL_FLAG_PRIVATE;
	qu_array  a;
	qu_any    o;
	qu_string k;
	qu_any   *p;
	long   i, n;

	__qu_unsafe__;

	if (argc && argv[0] == QuTrue)
		all = 0;

	n = qu_attr_len (self->attrs);
	assert (n <= QU_GC_MAX_ITEMS);
	a = qu_array_new (n);
	p = a->objs;
	i = n = 0;
	while ((o = qu_attr_walk (self->attrs, &k, &i)) != 0)
	{
		qu_class c = qu_classof (o);
		if ((c == QuClassClass || c == QuSubClass || c == QuVarClass)
			&& !qu_flag_test (o, QU_SYMBOL_FLAG_LOCAL | all))
		{
			*p++ = o;
			n++;
		}
	}
	a->len = n;
	return a;
}



/*) sub __ref (name: String, defval) -> Any
	Get a reference to a module object.

	@warn
	This method is deprecated. Use #{__getslice} instead.

	@p
	Unlike the -{sysref} function, this method works for all objects, private,
	local, etc.

	If the object does not exist, if v{defval} is given then it is returned
	otherwise ENone is thrown.

	@exc
	.EArgValue
	.ESecurity
	.ENone

	@xmp
		use Foo

		_foo = 10
		println (__main__.__ref ('_foo'))
		println (__main__.__ref ('Foo'))

	@out
		<Var __main__._foo>
		<Module Foo>
*/
static qu_any qu__module__ref (qu_module self, int argc, qu_any *argv)
{
	qu_any x;

	__qu_unsafe__;

	QU_SYS_WARN ("Module.__ref is deprecated, use __getitem/__getslice instead");

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, x, qu_string);
	x = qu_attr_find (self->attrs, (qu_string) x);
	if (x)
		return x;
	if (argc > 1)
		return argv[1];
	QU_THROW (QuENone);
}



/*) sub __ast -> Ast
	Get the parsed root -{Ast} object.

	Note that you must run the program with the C{-p} option, otherwise
	parsed trees are deleted once compilation is successfull.

	@exc
	.ENone     module does not have a -{Code} object
	           (such as is the case with external modules)
	.ESecurity

	@xmp
		# run with the -p option
		use Foo
		println (__main__.__ast)

	@warn
		The output depends on the parser/compiler.

	@out
		\_0: module <Module __main__>
		. \_2: import <Module Foo>
		. \_0: nop
		. \_0: compound
		. . \_0: stmt
		. . . \_3: call op:1
		. . . . \_3: object <Sub Sys.println> op:1
		. . . . \_3: getattr
		. . . . . \_3: object <Sub Sys.__main__>
		. . . . . \_3: literal __tree
*/
static qu_ast qu__module__ast (qu_module self)
{
	__qu_unsafe__;

	QU_REQUIRE (QuENone, self->tree);
	return self->tree;
}



/*) sub __usepath -> String
	Return the path notation as spelled in the k{use} statement without
	the file extension.

	If the module is an eval module then an empty string is returned.
	If the module is the builtin %{Sys} module then C{'Sys'} is returned.

	@xmp
		/opt/qu/lib/foo/bar.qm -> foo/bar
		/foo/bar/baz.xx        -> /foo/bar/baz
*/
static qu_string qu__module__usepath (qu_module self)
{
	static qu_string ModuleDirWithSlash = 0;
	if (ModuleDirWithSlash == 0)
		ModuleDirWithSlash = qu_string_new_staticf ("%s" QU_PATH_SEPSTR, QuModuleDir->str);

	if (qu_module_is_eval (self))
		return QuStringEMPTY;
	return qu_string_delete (self->name, ModuleDirWithSlash);
}



/*) sub __uses -> Set
	Return a -{Set} containing Module objects used by this module.

	@noexc

	@xmp
		use Time, Math

		println (__main__.__uses)

	@out
		{<Module Math>, <Module Time>}
*/
static qu_set qu__module__uses (qu_module self)
{
	if (self->use)
		return self->use;
	return QuSetEMPTY;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

/*
	Get attribute.
*/
qu_symbol qu_module_getattr (qu_module self, qu_string name)
{

#ifdef QU_DEBUG_LOCAL
	qu_symbol x = GETATTR (self, name);
	if (x)
	{
		assert (qu_symbol_is (x));
	}
	return x;
#else
	return GETATTR (self, name);
#endif

}



/*
	Set attribute.

	!!! If an attribute with the same name already exists, it is replaced.
	!!! Thus the caller must call qu_module_getattr if necessary.
*/
void qu_module_setattr (qu_module self, qu_string name, qu_symbol x)
{
	assert (qu_symbol_is (x));
	SETATTR (self, name, x);
}



/*
	Get class or function which has the same name as the module.

	module /foo/bar/Baz

	class Baz
	sub Baz
*/
qu_symbol qu_module_getattr_callable_name (qu_module self, int runtime)
{
	qu_string s = qu_file_strip (qu_file_name (self->name));
	qu_symbol x = GETATTR (self, s);

	if (x && !qu_flag_test (x, QU_SYMBOL_FLAG_PRIVATE | QU_SYMBOL_FLAG_LOCAL))
	{
		if (qu_var_is (x))
		{
			if (!runtime && !qu_var_frozen (x))
				goto nope;
			x = ((qu_var) x)->value;
		}
		if (qu_class_is (x) || qu_sub_is (x))
			return x;
	}
  nope:
	QU_THROWF (QuEImplement, "module %s has no callable attribute %s", self->name->str, s->str);
}



/*
	The compiler must call this function if a resident module failed.

	What we do is remove the failed module from the Resident list so the
	next time a fresh version is loaded.

	If nothing were changed when the module is reloaded, then the same
	failure will be experienced again, of course, and that is desirable.
*/
void qu_module_failed (qu_module self)
{
	if (qu_module_is_resident (self))
	{
#ifdef QU_APP_WEAKMODULE
		qu_weak w;
#endif
		if (QU_APP_VERBOSE)
			printf ("module failed: %p %s\n", self, self->name->str);

		LOCK_RESIDENTS;
#ifdef QU_APP_WEAKMODULE
		w = qu_dict_getitem (Residents, self->name);
		qu_weak_unref (w);
#endif
		qu_dict_unfreeze (Residents);
		qu_dict_del      (Residents, self->name);
		qu_dict_unfreeze (Residents);

#ifdef QU_HAS_UNIVERSE
		{
			long   b, i;
			qu_any name;
			qu_set s;

			b = i = 0;
			while ((s = qu_dict_walk (QuUniverses, &name, &b, &i)) != NULL)
			{
				assert (qu_set_is (s));
				qu_set_del (s, self);
			}
			/* TODO - remove empty universes */
		}
#endif

		UNLOCK_RESIDENTS;
	}
}



/*
	Use one module.
*/
static qu_module module_use_one (qu_module self, qu_string abspath, qu_string as)
{
	qu_string name;
	qu_module m;

	if (GETATTR (self, as))
		QU_THROWF (QuEValue, "duplicate symbol %s in module %s", as->str, abspath->str);

	name = qu_file_strip (abspath);

	if (qu_string_EQ (name, self->name))
		QU_THROWF (QuEValue, "module %s cant use itself (%s = %s)", abspath->str, name->str, self->name->str);

	LOCK_RESIDENTS;

	m = module_get_resident (name);
	if (m)
	{
		if (self->use && qu_set_contains (self->use, m))
		{
			UNLOCK_RESIDENTS;
			QU_THROWF (QuEValue, "module %s already uses %s", self->name->str, name->str);
		}
	}
	else
	{
		m = module_load_one (abspath, 0);
		if (m == 0)
		{
			UNLOCK_RESIDENTS;
			QU_RETHROW;
		}

		qu_dict_unfreeze (Residents);
#ifdef QU_APP_WEAKMODULE
		qu_dict_setitem  (Residents, m->name, qu_weak_new (m));
#else
		qu_dict_setitem  (Residents, m->name, m);
#endif
		qu_dict_freeze   (Residents);

		if (QU_APP_VERBOSE)
			printf ("module loaded: %p %ld %s\n", m, Residents->len - 1, m->name->str);
	}

	qu_module_add_use (self, m, as);
	UNLOCK_RESIDENTS;
	return m;
}



qu_obj qu_module_use (qu_module self, qu_string use, qu_string as, int option)
{
	if (qu_string_EQ (use, QuSysModule->name))
	{
		/*
			Special case - the builtin Sys module must be spelled as: use Sys
			We don't care how many times it is spelled.
		*/

		QU_REQUIREZ (QuECompile, as == 0, "the Sys module must always be used as Sys");

		if (GETATTR (self, QuSysModule->name) == 0)
			qu_module_add_use (self, QuSysModule, QuSysModule->name);
		return (qu_obj) QuSysModule;
	}
	else
	{
		/*
			This is the logic of how we locate modules.
		*/

		int isabs = qu_file_is_abs (use);

		if (isabs)
		{
			/* If it is an absolute path then just use it */
		}
		else if (option == QU_MODULE_USE_PACKAGE)
		{
			/* If it specifies __package then we know it must be relative to the
				main program path */

			use = module_use_toabs (qu_file_dir (self->name), use, 1);
		}
		else if (option == QU_MODULE_USE_STANDARD)
		{
			/* If it specifies __std then we know it must be relative to the
				lib directory */

			use = module_use_toabs (qu_file_dir (self->name), use, 0);
		}
		else
		{
			qu_string dir = qu_file_dir (self->name);
			qu_string p1, p2, m1, m2;

			/* 1. try relative to the main program */
			p1 = module_use_toabs (dir, use, 1);
			if (!qu_string_EQ (self->name, p1))
			{
				m1 = qu_file_repext (p1, QuModuleExt);
				if (qu_file_exist (m1, qu_string_byte ('f')))
				{
					use = p1;
					goto found;
				}

				m1 = qu_file_repext (p1, QuModuleDlExt);
				if (qu_file_exist (m1, qu_string_byte ('f')))
				{
					use = p1;
					goto found;
				}
			}

			/* 2. try relative to /opt/qu/lib */
			p2 = module_use_toabs (dir, use, 0);
			if (!qu_string_EQ (self->name, p2))
			{
				m2 = qu_file_repext (p2, QuModuleExt);
				if (qu_file_exist (m2, qu_string_byte ('f')))
				{
					use = p2;
					goto found;
				}

				m2 = qu_file_repext (p2, QuModuleDlExt);
				if (qu_file_exist (m2, qu_string_byte ('f')))
				{
					use = p2;
					goto found;
				}
			}
#if 0
			/* 3. Try relative to the module directory. */
			/* !!! doesn't work for Kata's relative modules */
			p1 = qu_file_toabs (use, dir);
			m1 = qu_file_repext (p1, QuModuleExt);
			if (qu_file_exist (m1, qu_string_byte ('f')))
			{
				use = p1;
				goto found;
			}

			m1 = qu_file_repext (p1, QuModuleDlExt);
			if (qu_file_exist (m1, qu_string_byte ('f')))
			{
				use = p1;
				goto found;
			}
#endif
		}

	  found:

		if (as == 0)
			as = qu_file_name (use);
		return (qu_obj) module_use_one (self, use, as);
	}
}




/*
	ispath = 1
		's' must be the full absolute path to the file including the
		extension, e.g. ../foo/bar/baz.xx. Note that we dont force a standard
		file extension except for modules.

	ispath = 0 (for the -E option)
		's' is the string program
*/
qu_module qu_module_new_main (qu_string s, int ispath)
{
	qu_string path;

	if (ispath)
	{
		path = s;
		s = qu_file_load (path);
		QU_ENSURE (s);
	}
	else
	{
		path = qu_file_toabs (qu_string_newz ("__eval__.qu"), 0);
	}

	QuModuleMain = module_alloc (path, QU_MODULE_FLAG_TRUSTED | QU_MODULE_FLAG_RESIDENT, 1);
	QuModuleMain->literals = QuSysModule->literals;
	QuModuleMain->source = s;
	return QuModuleMain;
}



/*
	Create an "eval" module.

	!!! Remember: the module object is just an ordinary object.
	!!! It is not rooted.
*/
qu_module qu_module_new_eval (qu_module self, qu_string src, qu_string path)
{
	char      buf [128];
	qu_module m;

#ifdef HAVE_LONG_LONG
	static volatile unsigned long long count = 0;
	if (path == 0)
	{
		sprintf (buf, "eval<%llu>", count++);
		path = qu_string_newz (buf);
	}
#else
	static volatile unsigned long count = 0;
	if (path == 0)
	{
		sprintf (buf, "eval<%lu>", count++);
		path = qu_string_newz (buf);
	}
#endif

	m = module_alloc (path, QU_MODULE_FLAG_EVAL, 0);
	m->literals = qu_set_new (16);
	m->source = src;
	m->parent = (qu_symbol) self;
	return m;
}



/*
	Import a symbol from another module.
	!!! Does not care whether or not 'u' is a resident module.
*/
int qu_module_import (qu_module self, qu_module u, qu_string name)
{
	qu_any o = qu_attr_find (u->attrs, name);

	if (o != 0)
	{
		qu_class c = qu_classof (o);

		if (c == QuClassClass || c == QuSubClass || c == QuVarClass)
		{
			if (qu_flag_test (o, QU_SYMBOL_FLAG_PRIVATE | QU_SYMBOL_FLAG_LOCAL))
				goto nope;
		}
		else
		{
			/* Dont import modules used by the module */
			goto nope;
		}

		if (qu_attr_find (self->attrs, name) != 0)
			QU_THROWF (QuEValue, "%s: conflicting import symbol '%s'", u->name->str, name->str);

		qu_attr_add (self->attrs, name, o);
	}
	return 1;

  nope:
	QU_THROWF (QuEValue, "symbol '%s' not found", name->str);
}



/*
	Import all symbols from another module.
	!!! Does not care whether or not 'u' is a resident module.
*/
int qu_module_import_all (qu_module self, qu_module u)
{
	qu_any    sym;
	qu_string name;
	long      i = 0;

	while ((sym = qu_attr_walk (u->attrs, &name, &i)) != 0)
	{
		qu_class c = qu_classof (sym);
		qu_any   y;

		if (c == QuClassClass || c == QuSubClass || c == QuVarClass)
		{
			if (qu_flag_test (sym, QU_SYMBOL_FLAG_PRIVATE | QU_SYMBOL_FLAG_LOCAL))
				continue;
		}
		else if (c == QuModuleClass)
		{
			/* don't import modules used by the module */
			continue;
		}

		y = qu_attr_find (self->attrs, name);
		if (y)
		{
			/* QU_THROWF (QuEValue, "%s: conflicting import symbol '%s'", u->name->str, name->str); */
			qu_attr_set (self->attrs, name, sym);
		}
		else
		{
			qu_attr_add (self->attrs, name, sym);
		}
	}
	return 1;
}




/* whether or not a module is loaded, return the object if so, NULL otherwise */
qu_module qu_module_loaded (qu_string path)
{
	qu_module m;

	LOCK_RESIDENTS;
	m = module_get_resident (path);
	UNLOCK_RESIDENTS;
	return m;
}



/* must only be used to load C modules */
qu_module qu_module_load (qu_string path)
{
	qu_module m;

	LOCK_RESIDENTS;
	m = module_get_resident (path);
	if (m == 0)
	{
		m = module_load_one (path, 0);
		if (m == 0)
		{
			UNLOCK_RESIDENTS;
			QU_THROWF (QuESystem, "cant load %s", path->str);
		}
		qu_dict_unfreeze (Residents);
#ifdef QU_APP_WEAKMODULE
		qu_dict_setitem (Residents, m->name, qu_weak_new (m));
#else
		qu_dict_setitem (Residents, m->name, m);
#endif
		qu_dict_freeze (Residents);
	}
	UNLOCK_RESIDENTS;
	return m;
}


qu_module qu_module_math ()
{
#ifdef QU_APP_WEAKMODULE
	static qu_string p = 0;
	qu_module m;

	if (p == 0)
		p = qu_string_newf ("%s" QU_PATH_SEPSTR "Math", QuModuleDir->str);

	m = module_get_resident (p);
	if (m)
		return m;
	return qu_module_load (p);
#else
	static qu_module m = 0;

	if (m == 0)
	{
		qu_string p = qu_string_newf ("%s" QU_PATH_SEPSTR "Math", QuModuleDir->str);
		m = module_get_resident (p);
		if (m == 0)
			m = qu_module_load (p);
	}
	return m;
#endif
}


htopen_f   __QuModuleHtml_open;
htclose_f  __QuModuleHtml_close;
htentity_f __QuModuleHtml_entity;

qu_module qu_module_html ()
{
#ifdef QU_APP_WEAKMODULE
	static qu_string p = 0;
	qu_module m;

	if (p == 0)
		p = qu_string_newf ("%s" QU_PATH_SEPSTR "html", QuModuleDir->str);

	m = module_get_resident (p);
	if (m)
		return m;
	return qu_module_load (p);
#else
	static qu_module m = 0;

	if (m == 0)
	{
		qu_string p = qu_string_newf ("%s" QU_PATH_SEPSTR "html", QuModuleDir->str);
		m = module_get_resident (p);
		if (m == 0)
			m = qu_module_load (p);
	}
	return m;
#endif
}



/*----------------------------------------------------------------------------
	for C modules
----------------------------------------------------------------------------*/

static void module_check_attr_new (qu_module self, qu_string name)
{
	if (qu_module_getattr (self, name))
		QU_FATAL ("module %s: %s already declared", self->name->str, name->str);
}



qu_sub qu_module_add_sub (qu_module self, const char *name, void *func, int argc, long flags)
{
	qu_string s = qu_string_new_staticz (name);
	qu_sub    u;

	module_check_attr_new (self, s);

	u = qu_sub_new ((qu_symbol) self, s, func, argc, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_EXTERN | flags);
	qu_module_setattr (self, s, (qu_symbol) u);
	return u;
}



qu_var qu_module_add_var (qu_module self, const char *name, qu_any value)
{
	qu_string s = qu_string_new_staticz (name);
	qu_var    v;

	module_check_attr_new (self, s);

	v = qu_var_new_rdwr ((qu_symbol) self, s);
	v->value = value;
	qu_module_setattr (self, s, (qu_symbol) v);
	return v;
}



qu_var qu_module_add_const (qu_module self, const char *name, qu_any value)
{
	qu_string s = qu_string_new_staticz (name);
	qu_var    v;

	module_check_attr_new (self, s);

	v = qu_var_new_const ((qu_symbol) self, s);
	v->value = value;
	qu_module_setattr (self, s, (qu_symbol) v);
	return v;
}



qu_symbol qu_module_add_alias (qu_module self, const char *name, const char *alias)
{
	qu_string n = qu_string_new_staticz (name);
	qu_string a = qu_string_new_staticz (alias);

	qu_symbol x = qu_module_getattr (self, n);

	if (x == 0)
		QU_FATAL ("module %s: no such symbol %s", self->name->str, name);

	module_check_attr_new (self, a);

	qu_module_setattr (self, a, x);
	return x;
}



qu_var qu_module_add_float (qu_module self, const char *name, double val)
{
	return qu_module_add_const (self, name, qu_float_new (val));
}



qu_var qu_module_add_int (qu_module self, const char *name, long val)
{
	if (val < QU_INT_MIN || val > QU_INT_MAX)
		QU_FATAL ("module %s: %s Int value out of range %ld",
			self->name->str, name, val);
	return qu_module_add_const (self, name, qu_int_new (val));
}



qu_var qu_module_add_string (qu_module self, const char *name, const char *val)
{
	return qu_module_add_const (self, name, qu_string_new_staticz (val));
}


/*----------------------------------------------------------------------------
	set/get instance attribute - runtime
----------------------------------------------------------------------------*/

static qu_any module__get (qu_module self, qu_string name, int restrict)
{
	qu_any x;

	assert (qu_module_is (self));
	x = qu_attr_find (self->attrs, name);
	if (x == 0)
	{
		x = qu_attr_find (QuModuleClass->attrs, name);
		if (x == 0)
		{
			x = qu_attr_find (QuObjectClass->attrs, name);
			if (x == 0)
				goto not_found;
		}
	}

	assert (qu_symbol_is (x));

	if (qu_module_is (x))
		return x;

	if (restrict && qu_flag_test (x, QU_SYMBOL_FLAG_PRIVATE | QU_SYMBOL_FLAG_LOCAL))
		goto not_found;

	if (qu_var_is (x))
		return ((qu_var) x)->value;

	return x;

	not_found: QU_THROWF (QuEImplement, "module %s has no such attribute %s", self->name->str, name->str);
}



static qu_any module__set (qu_module self, qu_string name, qu_any val, int restrict)
{
	qu_any x;

	assert (qu_module_is (self));
	x = qu_attr_find (self->attrs, name);
	if (x == 0)
	{
		x = qu_attr_find (QuModuleClass->attrs, name);
		if (x == 0)
		{
			x = qu_attr_find (QuObjectClass->attrs, name);
			if (x == 0)
				goto not_found;
		}
	}

	assert (qu_symbol_is (x));

	if (qu_module_is (x))
		goto cant_modify;

	if (restrict && qu_flag_test (x, QU_SYMBOL_FLAG_PRIVATE | QU_SYMBOL_FLAG_LOCAL))
		goto not_found;

	if (qu_var_is (x))
	{
		if (qu_var_frozen (x) || (restrict && !qu_var_is_rdwr (x)))
			goto cant_modify;

		QU_ENSURE (qu_var_set_value (x, val));

		/* don't return val since the validator may change the value assigned */
		return ((qu_var) x)->value;
	}
	else if (qu_sub_is (x))
	{
		return qu_sub_call1 (x, self, val);
	}

	cant_modify: QU_THROWF (QuEModify   , "cant modify %s.%s"                 , self->name->str, name->str);
	not_found:   QU_THROWF (QuEImplement, "module %s has no such attribute %s", self->name->str, name->str);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_module__setup ()
{
	qu_class_setup_sys (QuModuleClass, qu_module_s, qu__module__destroy, qu__module__mark);

	QuModuleClass->get = (qu_attrget_f) module__get;
	QuModuleClass->set = (qu_attrset_f) module__set;

	QuSysModule = module_alloc (qu_string_new_staticz ("Sys"), QU_MODULE_FLAG_RESIDENT | QU_MODULE_FLAG_TRUSTED, 1);
	qu_flag_set (QuSysModule, QU_MODULE_FLAG_PARSED | QU_MODULE_FLAG_COMPILED | QU_MODULE_FLAG_READY | QU_MODULE_FLAG_INIT);

	QuModuleDir   = qu_string_new_staticf ("%s" QU_PATH_SEPSTR "lib", QU_APP_HOMEDIR);
	QuModuleExt   = qu_string_new_staticz (MOD_EXT);
	QuModuleDlExt = qu_string_new_staticz (MOD_DLEXT);

	SModPat  = qu_string_new_staticz ("*" MOD_EXT);
	SSoPat   = qu_string_new_staticz ("*" MOD_DLEXT);
	SRegType = qu_string_byte    ('f');

	ModuleAfterSetup = 1;
}



void qu_module__init ()
{
	QuSysModule->literals = qu_set_new (128);

#ifdef QU_HAS_UNIVERSE
	QuUniverses = qu_dict_new_root (128);
	{
		qu_set s = qu_set_new (1);
		qu_set_add (s, QuSysModule);
		qu_dict_setitem (QuUniverses, QuSysModule->name, s);
	}
#endif

	Residents = qu_dict_new_root (128);
#ifdef QU_APP_WEAKMODULE
	qu_dict_setitem (Residents, QuSysModule->name, qu_weak_new (QuSysModule));
#else
	qu_dict_setitem (Residents, QuSysModule->name, QuSysModule);
#endif
	qu_dict_freeze (Residents);
	qu_module_add_const (QuSysModule, "__modules__", Residents);

	/* the Any symbol = the Object Class object */
	qu_module_setattr (QuSysModule, qu_string_new_staticz ("Any"), (qu_symbol) QuObjectClass);

	/* class specific methods */
	qu_class_method_classz (QuModuleClass, "list"  , qu__module_list  , 0);
	qu_class_method_classz (QuModuleClass, "loaded", qu__module_loaded, 1);

	/* standard instance methods */
	qu_class_method_instz  (QuModuleClass, "__ast"          , qu__module__ast     ,  0);
	qu_class_method_instz  (QuModuleClass, "__code"         , qu__module__code    ,  0);
	qu_class_method_inst   (QuModuleClass, QU_NAME__cmp     , qu__module__cmp     ,  1);
	qu_class_method_instz  (QuModuleClass, "__docdict"      , qu__module__docdict ,  0);
	qu_class_method_inst   (QuModuleClass, QU_NAME__getitem , qu__module__getitem ,  1);
	qu_class_method_inst   (QuModuleClass, QU_NAME__getslice, qu__module__getslice,  2);
	qu_class_method_instz  (QuModuleClass, "__kind"         , qu__module__kind    ,  0);
	qu_class_method_instz  (QuModuleClass, "__callable"     , qu__module__callable,  0);
	qu_class_method_instz  (QuModuleClass, "__path"         , qu__module__path    , -1);
	qu_class_method_instz  (QuModuleClass, "__ref"          , qu__module__ref     , -1);
	qu_class_method_instz  (QuModuleClass, "__stab"         , qu__module__stab    , -1);
	qu_class_method_inst   (QuModuleClass, QU_NAME__string  , qu__module__string  ,  0);
	qu_class_method_inst   (QuModuleClass, QU_NAME__print   , qu__module__print   ,  0);
	qu_class_method_instz  (QuModuleClass, "__usepath"      , qu__module__usepath ,  0);
	qu_class_method_instz  (QuModuleClass, "__uses"         , qu__module__uses    ,  0);

	/* finalize declaration */
	qu_class_enddecl (QuModuleClass);
}
