/* Null.c - Copyright (c) Marc Krisnanto */


#include "qu.h"


/*) class Null __final
	The sole instance of this class is the v{nil} object. The keyword
	k{nil} simply points to this object.

	The Null class is internally marked as k{final} and k{abstract}
	so it can not be extended nor instantiated, that is, you can not create
	new instances of the Null class.
*/
qu_class QuNullClass;

qu_null   QuNil;
qu_string QuNilSTR;


/*-----------------------------------------------------------------------------
	standard class methods
-----------------------------------------------------------------------------*/
/* none */


/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

#ifdef QU_DEBUG_LOCAL
static void qu__null__destroy (qu_null self)
{
	QU_FATAL ("never");
}
#endif



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Int}
		-{Integer}
		-{Number}  | Return Int C{0}.
	.| -{Long}     | Return Long C{0}.
	.| -{Rational} | Return Rational C{0 _/ 1}.
	.| -{Float}
		-{Decimal} | Return Float C{0.0}.
	.| -{Big}      | Return Big C{0.0}.
	.| -{Complex}  | Return Complex C{(0 + 0j)}.
	.| -{String}   | Return String C{'nil'}.
	.| -{Boolean}  | Return false.
	.| Null        | Return nil.

	@exc
	.EArgValue

	@xmp
		i = nil as Number
		println (i)

	@out
		0
*/
static qu_any null__as (qu_null self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Integer:
		case QU_CLASS_ID_Number:   return QuInt0;
		case QU_CLASS_ID_Long:     return QuLong0;
		case QU_CLASS_ID_Rational: return QuRational0;
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Decimal:  return QuFloat0;
		case QU_CLASS_ID_Big:      return QuBig0;
		case QU_CLASS_ID_Complex:  return QuComplex00;
		case QU_CLASS_ID_Boolean:  return QuFalse;
		case QU_CLASS_ID_String:   return QuNilSTR;
		case QU_CLASS_ID_Null:     return self;
	}
	QU_ARG_THROW;
}



/*) sub __cmp (x) -> Int
	Compare nil to v{x}.

	If v{x} is nil then 0 is returned. Otherwise delegate to -{Object.__cmp}.

	@noexc

	@xmp
		println (0 <=> nil)

	@out
		1
*/
static qu_int null__cmp (qu_null self, qu_any x)
{
	if (self == x)
		return QuInt0;
	return qu_object_default_cmp (self, x);
}



/*) sub __eq (x) -> Boolean
	Return true if v{x} is nil, false otherwise.

	@noexc

	@xmp
		println (0 == nil)

	@out
		false
*/
static qu_boolean null__eq (qu_null self, qu_any x)
{
	return ((qu_null) x == QuNil) ? QuTrue : QuFalse;
}



/*) sub __hash
	Return the hash value of nil which is 0.

	@noexc

	@xmp
		println (nil.__hash)

	@out
		0
*/
static qu_int null__hash (qu_null self)
{
	return QuInt0;
}



/*) sub __like (x) -> Boolean
	This method returns the same result as the #{__eq} method.

	@noexc
*/
static qu_boolean null__like (qu_null self, qu_any x)
{
	return ((qu_null) x == QuNil) ? QuTrue : QuFalse;
}



/*) sub __print
	Print the string C{'nil'} to v{stdout}.

	See -{print} for possible exceptions.
*/
static qu_null null__print (qu_null self)
{
	return qu_sys_printz (QuNilSTR->str, QuNilSTR->len);
}



/*) sub __push (x) -> Null
	This method does nothing then return nil.

	@noexc

	@xmp
		println (nil <<< 1 <<< 'foo')

	@out
		nil
*/
static qu_null null__push (qu_null self, qu_any x)
{
	return QuNil;
}



/*) sub __string -> String
	Return the string C{'nil'}.

	@noexc

	@xmp
		println (nil as String)

	@out
		nil
*/
static qu_string null__string (qu_null self)
{
	return QuNilSTR;
}



/*) sub __true -> Boolean
	Return false.

	@noexc

	@xmp
		if nil
			__abort__ ('never')
		else
			println ('ok')
		;;

	@out
		ok
*/
static qu_boolean null__true (qu_null self)
{
	return QuFalse;
}



/*----------------------------------------------------------------------------
	Init
----------------------------------------------------------------------------*/

void qu_null__setup ()
{

#ifdef QU_DEBUG_LOCAL
	qu_class_setup_sys (QuNullClass, qu_object_s, qu__null__destroy, 0);
#else
	qu_class_setup_sys (QuNullClass, qu_object_s, 0, 0);
#endif

	QuNil    = qu_gc_new_static      (QuNullClass);
	QuNilSTR = qu_string_new_staticz ("nil");
}



void qu_null__init ()
{
	/* standard instance methods */
	qu_class_method_inst (QuNullClass, QU_NAME__as    , null__as    , 1);
	qu_class_method_inst (QuNullClass, QU_NAME__cmp   , null__cmp   , 1);
	qu_class_method_inst (QuNullClass, QU_NAME__eq    , null__eq    , 1);
	qu_class_method_inst (QuNullClass, QU_NAME__hash  , null__hash  , 0);
	qu_class_method_inst (QuNullClass, QU_NAME__like  , null__like  , 1);
	qu_class_method_inst (QuNullClass, QU_NAME__print , null__print , 0);
	qu_class_method_inst (QuNullClass, QU_NAME__push  , null__push  , 1);
	qu_class_method_inst (QuNullClass, QU_NAME__string, null__string, 0);
	qu_class_method_inst (QuNullClass, QU_NAME__true  , null__true  , 0);

	/* finalize declaration */
	qu_class_enddecl (QuNullClass);
}
