/* Number.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Number __abstract
	This is the parent class of all number classes, -{Integer}, -{Decimal}
	and -{Complex}.

	The class hierarchy:
	@tree
	.Number
	.	-{Real}          real numbers
	.		-{Integer}   integer
	.			-{Int}   system native integer
	.			-{Long}  arbitrary size integer
	.		-{Decimal}   floating point
	.			-{Float} system native double precision
	.			-{Big}   arbitrary size floating point
	.		-{Rational}  rational number
	.	-{Complex}       complex number

	@p
	Creating a number system is extremely difficult. There are many opened issues
	and different opinions. The lack of universal standards and the present state
	of our technology makes this somewhat non-trivial. We have to compromise
	with hardware and math to get reasonable performance.

	Qu basically adopts IEEE except for some that are irrelevant or make no sense.
	For example, in IEEE NaN is an entity that may be propagated to enable a program
	to check the result of an operation later. It appears that this is believed to
	be more eficient than having to check for a NaN after each operation. Qu does
	not share this idea. NaN is an error telling us that something is not mathematically
	defined. In Qu, as soon as a NaN yields from a Number operation, an EMath exception
	is thrown. It is foolish to assume that NaN is harmless. You might be changing the
	state of a number of things and when you later find out that there was an error
	it's already too late.

	The Number class provides a global context which determines things such as whether
	or not an exception should be thrown for special conditions. See the #{context}
	method for details.

	When you perform an operation one two Number objects of different classes,
	you meet the terms "coercion" and "promotion".

	In Qu "coercion" happens automatically but Qu also lets you do this on your
	own. Converting a Long integer to a Float is not "promotion". Precision
	may be lost and you get "demotion" instead. Thus "promotion" gives you
	the wrong impression that somehow accuracy is increased. That may not
	always be the case. "morphing" is probably a more appropriate term.

	Morphing does occur in Qu and it works this way:
	@def
		Integer -> Decimal -> Complex

	@p
	The opposite direction does not happen unless you deliberately do it yourself.

	@xmp
		a = 1
		b = 2.0
		c = a + b       #-> Float
		c = a + Int (b) #-> Int

	@p
	When you mix an -{Integer} with a -{Decimal}, the Integer number is morphed
	into a Decimal. The result is either a Decimal or infinity.

	@warn
	This behavior might change in future versions. New #{context} options may
	be introduced and might give you an Int(3) out of Float(1) + Big(2).

	@p
	Infinity in Qu is neither an Integer nor Decimal and certainly not a -{Float}.
	infp and infn are Qu's sytem defined -{Inf} objects which can appear
	as the real or imaginary part of a Complex number.

	When you mix an Integer or a Decimal with a -{Complex} then the Integer or
	Decimal is morphed into a Complex. In this case, the result a Complex unless
	the #{context} specifies #{REALIFY_COMPLEX}. If infinity comes in, it will
	simply be either the real or imaginary part or both, but still a Complex.

	For each Number classes, you must carefully inspect their behaviors by
	inspecting their methods. See for example #{__div} and #{__idiv} and
	observe how they operate differently.

	Sometimes you must write your expressions in certain ways to achieve what
	you want. Qu gives you the tool to convert numbers of different classes.
	All Number classes support the f{new} which take any Number object, and
	perhaps other objects, as the argument.

	@xmp
		a = 1 / 3
		println (a)

		b = Float (a)
		println (b)

		Big.precision (4)
		c = Big (b)
		println (c)

		d = Big ('1234.5')
		e = Float ('123E45')

	@out
		(1 _/ 3)
		0.333333
		0.3333333333333333E+0


	@h3
	Exceptions
	@---------
	@indent
	EMinus, EPlus, EMul, etc. are thrown if v{x} is not a Number.

	EInf is thrown if the result is infinity the #{context} does not allow
	infinities to be returned.

	EHuge is thrown if the number is too big and can not be represented
	using an -{Integer} or -{Decimal}. This happens, for example, when you
	divide a huge number with a tiny number.

	ETiny is thrown in -{Decimal} context if the number is too small and can
	not be represented using a -{Decimal}. This happens for example, when
	you divide a tiny -{Big} with a huge Big.

	EBrain is thrown in situations when an intermediate result of a calculation
	throws an exception which makes it imposible to determine the outcome.
	For example, C{x = foo (y + y)} where y is a huge number and the #{context}
	does not allow infinities. This means that y + y will throw an exception
	and so the result of v{foo} is undefined.
	@unindent
*/
qu_class QuNumberClass;

/* global context */
volatile long __QuNumberContext = QU_NUMBER_CONTEXT_DEFAULT;



/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

#define UNOP(name, rtype) \
rtype qu_number_##name (qu_number x) \
{ \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) qu_int_##name      ((qu_int) x); \
		case QU_CLASS_ID_Long:     return (rtype) qu_long_##name     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return (rtype) qu_rational_##name ((qu_rational) x); \
		case QU_CLASS_ID_Float:    return (rtype) qu_float_##name    ((qu_float) x); \
		case QU_CLASS_ID_Big:      return (rtype) qu_big_##name      ((qu_big) x); \
		case QU_CLASS_ID_Complex:  return (rtype) qu_complex_##name  ((qu_complex) x); \
		case QU_CLASS_ID_Inf:      return (rtype) qu_inf_##name      ((qu_inf) x); \
	} \
	QU_FATAL ("x not a Number object"); \
	return 0; /* never */ \
}



#define BINOP(name, rtype, ytype) \
rtype qu_number_##name##_##ytype (qu_number x, qu_##ytype y) \
{ \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) qu_int_##name##_##ytype      ((qu_int) x, y); \
		case QU_CLASS_ID_Long:     return (rtype) qu_long_##name##_##ytype     ((qu_long) x, y); \
		case QU_CLASS_ID_Rational: return (rtype) qu_rational_##name##_##ytype ((qu_rational) x, y); \
		case QU_CLASS_ID_Float:    return (rtype) qu_float_##name##_##ytype    ((qu_float) x, y); \
		case QU_CLASS_ID_Big:      return (rtype) qu_big_##name##_##ytype      ((qu_big) x, y); \
		case QU_CLASS_ID_Complex:  return (rtype) qu_complex_##name##_##ytype  ((qu_complex) x, y); \
		case QU_CLASS_ID_Inf:      return (rtype) qu_inf_##name##_##ytype      ((qu_inf) x, y); \
	} \
	QU_FATAL ("x is not a Real object"); \
	return 0; \
}

#define BINOP_int(name, rtype)      BINOP (name, rtype, int)
#define BINOP_float(name, rtype)    BINOP (name, rtype, float)
#define BINOP_long(name, rtype)     BINOP (name, rtype, long)
#define BINOP_rational(name, rtype) BINOP (name, rtype, rational)
#define BINOP_big(name, rtype)      BINOP (name, rtype, big)
#define BINOP_complex(name, rtype)  BINOP (name, rtype, complex)
#define BINOP_inf(name, rtype)      BINOP (name, rtype, inf)
#define BINOP_integer(name, rtype)  QU_NUMBER_BINOP_integer  (qu_number, name, rtype)
#define BINOP_decimal(name, rtype)  QU_NUMBER_BINOP_decimal  (qu_number, name, rtype)
#define BINOP_real(name, rtype)     QU_NUMBER_BINOP_real     (qu_number, name, rtype)
#define BINOP_number(name, rtype)   QU_NUMBER_BINOP_number   (qu_number, name, rtype)



/*----------------------------------------------------------------------------
	specific class constants
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (x) -> Number __abstract
	Create a new Number object from another.

	This method is implemented by all Number classes.

	@xmp
		a = 1
		b = Float (a)
		c = Rational (b)
		d = Big ('1.234')

		println (a)
		println (b)
		println (c)
		println (d)

		x = Number ('123E400')
		println (x ' ' x.__name)

	@out
		1
		1.0
		(1 _/ 1)
		0.00012340E+4
		0.0123E+404 Big

	@see
	-{Int.new}
	-{Long.new}
	-{Rational.new}
	-{Float.new}
	-{Big.new}
	-{Complex.new}
*/
static qu_number qu__number_new (qu_class self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Long:
		case QU_CLASS_ID_Rational:
		case QU_CLASS_ID_Big:
			return (qu_number) x;

		case QU_CLASS_ID_Inf:
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_number) x;

		case QU_CLASS_ID_Boolean:
			return ((qu_boolean) x == QuTrue) ? QuNumber1 : QuNumber0;

		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Buffer:
			x = qu_number_scan_string ((qu_string) x, QU_NUMBER_SCAN_COMPLEX | QU_NUMBER_SCAN_SKIPSPACE);
			QU_ENSURE (x);
			if (qu_inf_is (x))
				QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return x;
	}
	return qu_object_as (x, QuNumberClass);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	scan
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_real _number_scan_real (char **s, int skipspace)
{
	char   *c   = *s;
	int     neg = 0;
	qu_any  x;
	char   *e;

	if (skipspace)
	{
		while (qu_ascii_isspace (*c))
			++c;

		if (*c == '-')
		{
			neg = 1;
			++c;
		}
		else if (*c == '+')
		{
			++c;
		}

		while (qu_ascii_isspace (*c))
			++c;
	}

	if (*c == '\0')
	{
		errno = EDOM;
		return 0;
	}

	if (*c == '0')
	{
		if (c[1] == '\0')
		{
		  zero:
			c++;
			x = QuInt0;
			goto done;
		}
		switch (c[1])
		{
			case 'j':
				goto zero;

			case '.':
				/* account for range operator .. */
				if (c[2] == '.')
					goto zero;
				e = c + 2;
				while (qu_ascii_isdigit (*e))
					++e;
				if (*e == 'D' || *e == 'd')
					x = qu_big_scan (&c, neg);
				else
					x = qu_float_scan (&c, neg);
				goto done;

			case 'E':
			case 'e':
				x = qu_float_scan (&c, neg);
				goto done;

			case 'D':
			case 'd':
				x = qu_big_scan (&c, neg);
				goto done;

			case 'B':
			case 'b':
				c += 2;
				x = qu_int_scan_binary (&c, neg);
				if (x == 0 && errno == ERANGE)
					x = qu_long_scan_binary (&c, neg);
				goto done;

			case 'X':
			case 'x':
				c += 2;
				x = qu_int_scan_hex (&c, neg);
				if (x == 0 && errno == ERANGE)
					x = qu_long_scan_hex (&c, neg);
				goto done;

			default:
				c++;
				if (!qu_ascii_isodigit (*c))
				{
					x = QuInt0;
				}
				else
				{
					x = qu_int_scan_oct (&c, neg);
					if (x == 0 && errno == ERANGE)
						x = qu_long_scan_oct (&c, neg);
				}
				goto done;
		}
	}

	if (c[1] == 'E' || c[1] == 'e')
	{
		x = qu_float_scan (&c, neg);
	}
	else if (c[1] == 'D' || c[1] == 'd')
	{
		x = qu_big_scan (&c, neg);
	}
	else
	{
		char *i = c;
		int   f = 0;

		while (qu_ascii_isdigit (*i))
			++i;

		if (*i == '.' && i[1] != '.')
		{
			f = 1;
			++i;
			while (qu_ascii_isdigit (*i))
				++i;
		}

		if (*i == 'E' || *i == 'e')
		{
			x = qu_float_scan (&c, neg);
		}
		if (*i == 'D' || *i == 'd')
		{
			x = qu_big_scan (&c, neg);
		}
		else if (f)
		{
			/* make the best of it */
			x = qu_float_scan (&c, neg);
			if (x == 0 && errno == ERANGE)
				x = qu_big_scan (&c, neg);
		}
		else
		{
			x = qu_int_scan (&c, neg);
			if (x == 0 && errno == ERANGE)
				x = qu_long_scan (&c, neg);
		}
	}

  done:
	if (x)
		*s = c;
	return x;
}



qu_number qu_number_scan (char **s, int options)
{
	char    *c = *s;
	qu_real  r = _number_scan_real (&c, options & QU_NUMBER_SCAN_SKIPSPACE);
	qu_real  j;

	QU_ENSURE (r);

	if (*c == 'j')
	{
		++c;
		if (options & QU_NUMBER_SCAN_COMPLEX)
		{
			char *i = c;
			j = r;
			r = _number_scan_real (&i, 1);
			if (r)
				c = i;
			else
				r = (qu_real) QuInt0;
			*s = c;
			return (qu_number) qu_complex_newr (r, j);
		}
		else
		{
			*s = c;
			return (qu_number) qu_complex_newr ((qu_real) QuInt0, r);
		}
	}
	else if (*c == '_' && c[1] == '/')
	{
		c += 2;
		j = _number_scan_real (&c, 1);
		QU_ENSURE (j);
		QU_REQUIRE (QuESyntax, qu_integer_is (j));
		QU_REQUIRE (QuEMath, !qu_integer_iszero ((qu_integer) j));
		return (qu_number) qu_rational_new_ex ((qu_integer) r, (qu_integer) j);
	}
	else
	{
		*s = c;
		if (options & QU_NUMBER_SCAN_COMPLEX)
		{
			char *i = c;

			while (qu_ascii_isspace (*i))
				i++;

			if (qu_ascii_isdigit (*i) || *i == '+' || *i == '-')
			{
				j = _number_scan_real (&i, 1);
				if (j && *i == 'j')
				{
					*s = i + 1;
					return (qu_number) qu_complex_newr (r, j);
				}
			}
		}
		return (qu_number) r;
	}
}



qu_number qu_number_scan_string (qu_string s, int options)
{
	char     *z = s->str;
	qu_number x = qu_number_scan (&z, options);

	if (x == 0)
		QU_THROWZ (QuESyntax, "Number (String) invalid syntax");
	return x;
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class context (*arg) -> Int
	@def
	sub class context -> String
	sub class context (setting: String) -> String

	@p
	Use this to set or get the global Number context.

	If called with no arguments, return the current setting. If called with
	one argument, set and return the old setting.

	The setting is as follows:
	@table
	.| B{flag} | B{default}  | B{description}

	.| T | on |
		Whether or not ETiny should be thrown when the result is too small
		which makes it impossible to represent the number (ALLOW_TINY).

		If this flag is is set, then 0 is returned instead of throwing ETiny.

		See -{__tiny__}.

	.| H | on |
		Whether or not EHuge should be thrown when the result is too big
		which makes it impossible to represent the number (ALLOW_HUGE).

		If this flag is set, then an infinity is returned instead of
		throwing EHuge.

		See -{__huge__}.

	.| I | off |
		Whether or not infinities are allowed (ALLOW_INF).

		If not allowed then EInf is thrown if the result is infinity.

		See -{__infp__}, -{__infn__}, and -{__inf__}.

	.| Z | on |
		Whether or not division by zero should throw EDiv or return
		an infinity (ALLOW_ZERODIV).

		See -{__zerodiv__}.

	.| C | on |
		Whether or not EComplex should be thrown when the result is a Complex
		(ALLOW_COMPLEX).

		Note that currently this does not have any effect on all Number methods.
		It might, however, affect the Math module.

		See -{__complex__}.

	.| R | off |
		Whether or not to convert a Complex result into a Real when the
		imaginary part is zero (REALIFY_COMPLEX).

	.| E | off |
		Whether or not to try get "exact" results (EXACT).

		Currently this has no effect.

	@exc
	.EValue invalid setting

	@xmp
		println (Number.context)
		Number.context -= 'C'
		println (Number.context)
		Number.context += 'CRI'
		println (Number.context)

	@out
		CHT
		HT
		CHIRT

	@note
	When compiling, Qu temporarily sets the C flag on.

	This means, for example, that when you write a constant expression which
	yields infinity, then the compiler will complain.

	@xmp
		x = 1 / 0   #-> compile error

		x = 1
		x = x / 0   #-> no compile error

	@see
	-{__context__}
*/
qu_string qu__number_context (qu_class self, int argc, qu_any *argv)
{
	char buf [16], *p;

	QU_ARG_VCOUNT (0, 1);
	if (argc)
	{
		qu_string  s;
		char      *z;
		long       l;
		long       v = 0;

		QU_ARG_VTYPE (0, s, qu_string);

		z = s->str;
		l = s->len;
		while (l--)
		{
			switch (*z++)
			{
				case 'C': case 'c': v |= QU_NUMBER_CONTEXT_ALLOW_COMPLEX;   break;
				case 'E': case 'e': v |= QU_NUMBER_CONTEXT_EXACT;           break;
				case 'H': case 'h': v |= QU_NUMBER_CONTEXT_ALLOW_HUGE;      break;
				case 'I': case 'i': v |= QU_NUMBER_CONTEXT_ALLOW_INF;       break;
				case 'R': case 'r': v |= QU_NUMBER_CONTEXT_REALIFY_COMPLEX; break;
				case 'T': case 't': v |= QU_NUMBER_CONTEXT_ALLOW_TINY;      break;
				case 'Z': case 'z': v |= QU_NUMBER_CONTEXT_ALLOW_ZERODIV;   break;
				default: QU_ARG_THROWZ ("bad flag");
			}
		}
		__QuNumberContext = v;
	}

	p = buf;
	if (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_COMPLEX  ) *p++ = 'C';
	if (__QuNumberContext & QU_NUMBER_CONTEXT_EXACT          ) *p++ = 'E';
	if (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_HUGE     ) *p++ = 'H';
	if (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_INF      ) *p++ = 'I';
	if (__QuNumberContext & QU_NUMBER_CONTEXT_REALIFY_COMPLEX) *p++ = 'R';
	if (__QuNumberContext & QU_NUMBER_CONTEXT_ALLOW_TINY     ) *p++ = 'T';

	*p = '\0';
	return qu_string_newz (buf);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

/* none */



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_number_is (qu_any x)
{
	qu_class c = qu_classof (x);
	return (c == QuIntClass
		||  c == QuLongClass
		||  c == QuRationalClass
		||  c == QuFloatClass
		||  c == QuBigClass
		||  c == QuInfClass
		||  c == QuComplexClass);
}

UNOP (sign   , int)
UNOP (iszero , int)
UNOP (isneg  , int)
UNOP (isone  , int)
UNOP (is1    , int)
UNOP (iswhole, int)
UNOP (isodd  , int)




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any __abstract
	Get a version of self based on v{x}.

	This method is implemented by all Number classes.

	All Number classes support the following classes:
	@ul
	.Int
	.Long
	.Rational
	.Float
	.Big
	.Complex
	.Integer
	.Real
	.Number

	@p
	Refer to each class on how this method is implemented.

	@xmp
		x = 123
		println (x as Float)

	@out
		123.0

	@see
	-{Int.__as}
	-{Long.__as}
	-{Rational.__as}
	-{Float.__as}
	-{Big.__as}
	-{Complex.__as}
*/
/* abstract */



#define TO(name, rtype) \
rtype qu_number_to##name (qu_any x) \
{ \
	errno = 0; \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) qu_int_to##name      ((qu_int) x); \
		case QU_CLASS_ID_Long:     return (rtype) qu_long_to##name     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return (rtype) qu_rational_to##name ((qu_rational) x); \
		case QU_CLASS_ID_Float:    return (rtype) qu_float_to##name    ((qu_float) x); \
		case QU_CLASS_ID_Big:      return (rtype) qu_big_to##name      ((qu_big) x); \
		case QU_CLASS_ID_Complex:  return (rtype) qu_complex_to##name  ((qu_complex) x); \
		case QU_CLASS_ID_Inf:      return (rtype) qu_inf_to##name      ((qu_inf) x); \
	} \
	errno = EDOM; \
	return 0; \
}


TO (l  , long)
TO (ul , unsigned long);
TO (li , long);
TO (d  , double);
#ifdef HAVE_LONG_LONG
TO (ll , long long);
TO (ull, unsigned long long);
#endif
#ifdef HAVE_LONG_DOUBLE
TO (ld , long double);
#endif

#undef TO


#define AS(rtype) \
qu_##rtype qu_number_as_##rtype (qu_any x) \
{ \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return qu_int_as_##rtype      ((qu_int) x); \
		case QU_CLASS_ID_Long:     return qu_long_as_##rtype     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return qu_rational_as_##rtype ((qu_rational) x); \
		case QU_CLASS_ID_Float:    return qu_float_as_##rtype    ((qu_float) x); \
		case QU_CLASS_ID_Big:      return qu_big_as_##rtype      ((qu_big) x); \
		case QU_CLASS_ID_Complex:  return qu_complex_as_##rtype  ((qu_complex) x); \
		case QU_CLASS_ID_Inf:      return qu_inf_as_##rtype      ((qu_inf) x); \
	} \
	QU_THROW (QuEValue); \
}


AS (int)
AS (float);
AS (long);
AS (rational);
AS (big);
AS (complex);
AS (integer);
AS (decimal);
AS (real);
AS (number);

#undef AS


char qu_number__ltob (long v)
{
	if (errno == 0)
	{
		if (-127 <= v && v <= 127)
			return (char) v;
		errno = ERANGE;
	}
	return (v < 0) ? -127 : 127;
}



unsigned char qu_number__ultoub (unsigned long v)
{
	if (errno == 0)
	{
		if (v <= 255)
			return (unsigned char) v;
		errno = ERANGE;
	}
	return 255;
}



short qu_number__ltoh (long v)
{
	if (errno == 0)
	{
		if ((long) SHORT_MIN <= v && v <= (long) SHORT_MAX)
			return (short) v;
		errno = ERANGE;
	}
	return (v < 0) ? SHORT_MIN : SHORT_MAX;
}


unsigned short qu_number__ultouh (unsigned long v)
{
	if (errno == 0)
	{
		if (v <= (unsigned long) USHORT_MAX)
			return (unsigned short) v;
		errno = ERANGE;
	}
	return USHORT_MAX;
}



int qu_number__ltoi16 (long v)
{
	if (errno == 0)
	{
		if (-32768 <= v && v <= 32767)
			return (int) v;
		errno = ERANGE;
	}
	return (v < 0) ? -32768 : 32767;
}



unsigned int qu_number__ultoui16 (unsigned long v)
{
	if (errno == 0)
	{
		if (v <= 65535)
			return (unsigned int) v;
		errno = ERANGE;
	}
	return 65535;
}



long qu_number__ltoi32 (long v)
{
	if (errno == 0)
	{
		/* how more stupid can it be ... */
		if ((-2147483647L - 1L) <= v && v <= 2147483647L)
			return v;
		errno = ERANGE;
	}
	return (v < 0) ? (-2147483647L - 1L) : 2147483647L;
}



unsigned long qu_number__ultoui32 (unsigned long v)
{
	if (errno == 0)
	{
		if (v <= 4294967295UL)
			return (unsigned int) v;
		errno = ERANGE;
	}
	return 4294967295UL;
}



long qu_number__ltoli (long v)
{
	if (errno == 0)
	{
		if ((long) QU_INT_MIN <= v && v <= (long) QU_INT_MAX)
			return v;
		errno = ERANGE;
	}
	return (v < 0) ? QU_INT_MIN : QU_INT_MAX;
}



long qu_number__ltouli (long v)
{
	if (errno == 0)
	{
		if ((long) 0 <= v && v <= (long) QU_INT_MAX)
			return v;
		errno = ERANGE;
	}
	return (v < 0) ? 0 : QU_INT_MAX;
}



#if SIZEOF_INT != SIZEOF_LONG
int qu_number__ltoi (long v)
{
	if (errno == 0)
	{
		if ((long) INT_MIN <= v && v <= (long) INT_MAX)
			return (int) v;
		errno = ERANGE;
	}
	return (v < 0) ? INT_MIN : INT_MAX;
}


unsigned int qu_number__ultoui (unsigned long v)
{
	if (errno == 0)
	{
		if (v <= (unsigned long) UINT_MAX)
			return (unsigned short) v;
		errno = ERANGE;
	}
	return UINT_MAX;
}
#endif



float qu_number__dtof (double v)
{
	if (errno == 0)
	{
		if (v < 0)
		{
			if (v >= (double) -FLT_MAX)
				return (float) v;
		}
		else
		{
			if (v <= (double) FLT_MAX)
				return (float) v;
		}
		errno = ERANGE;
	}
	return (v < 0) ? -FLT_MAX : FLT_MAX;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int __abstract
	Compare self with another object.

	This method is implemented by all Number classes.

	If v{x} is a Number then their numerical values are compared.
	Otherwise C{x.__cmp (self)} is called and the result negated.
	The call is done via an internal delegator that makes sure that the result
	is an -{Int}, otherwise calls -{Obj.__cmp}.

	This method returns -1 if self < v{x}, 0 if self == v{x}, 1 otherwise.

	See the #{cmp} method for a version that throws when the argument is not
	a Number.

	@noexc

	@xmp
		println (123 <=> 456)

	@out
		-1

	@see
	-{Int.__cmp}
	-{Long.__cmp}
	-{Rational.__cmp}
	-{Float.__cmp}
	-{Big.__cmp}
	-{Complex.__cmp}
*/
/* abstract */



BINOP_int      (cmp, int)
BINOP_float    (cmp, int)
BINOP_long     (cmp, int)
BINOP_rational (cmp, int)
BINOP_big      (cmp, int)
BINOP_complex  (cmp, int)
BINOP_inf      (cmp, int)
BINOP_integer  (cmp, int)
BINOP_decimal  (cmp, int)
BINOP_real     (cmp, int)
BINOP_number   (cmp, int)

/*
	<=>
*/
BINOP_int      (cmpabs, int)
BINOP_float    (cmpabs, int)
BINOP_long     (cmpabs, int)
BINOP_rational (cmpabs, int)
BINOP_big      (cmpabs, int)
BINOP_complex  (cmpabs, int)
BINOP_inf      (cmpabs, int)
BINOP_integer  (cmpabs, int)
BINOP_decimal  (cmpabs, int)
BINOP_real     (cmpabs, int)
BINOP_number   (cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Number __abstract
	Return the result of subtracting 1 from self.

	This method is implemented by all Number classes.

	Note that for -{Decimal} numbers, the result is a Decimal even if the
	number has no fractional part.

	If the result is huge and the Number.#{context} does not specify ALLOW_INF
	then this method throws EHuge.

	Notice that ETiny is not possible.

	@exc
	.EHuge
	.EInf

	@xmp
		a = 3
		--a
		println (a)

	@out
		2

	@p
	Special cases:
	@ul
	.C{infp.__decr --> infp}
	.C{infn.__decr --> infn}

	@see
	-{Int.__decr}
	-{Long.__decr}
	-{Rational.__decr}
	-{Float.__decr}
	-{Big.__decr}
	-{Complex.__decr}
*/
/* abstract */



UNOP (decr, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number __abstract
	Return the result of dividing self with v{x}.

	This method is implemented by all Number classes.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{\} operator).

	@p
	The __div method works in a I{floating point} context. The remainder of
	the division is included in the result. The #{__idiv} method, on the
	other hand, works in an I{integer} context. The remainder is trashed
	yielding an -{Integer} result.

	When you divide an -{Integer} with another Integer, Qu delivers an "exact"
	result. If the division has a remainder, then a -{Rational} number is
	returned. An -{Integer} otherwise.

	Thus this is completely different with the C{/} operator in many other
	languages such as C/C++, Java, Pascal, etc.

	Here is a table that shows you all the possible outcome of this method.

	@code
	      +-------+-------+-------+-------+-------+-------+-------+
	   /  |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  Io   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  Lo   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  Ro   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  | FBN*  | FBN*  | FBN*  | FBN*  |  BN*  |  Fo   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  Bo   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |  N*   |  N*   |  N*   |  N*   |  N*   |   *   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception
	Io=Int(0) Lo=Long(0) Ro=Rational(0) Fo=Float(0) Bo=Big(0)

	@exc
	.EDiv
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		x = 17
		println (x / 2)
		println (x \ 2)
		println (x / 2.0)

		x = 1
		y = x / 3
		println (y ' ' y.__class.__name)
		z = y * 3
		println (z ' ' z.__class.__name)

	@out
		(17 _/ 2)
		8
		8.5
		(1 _/ 3) Rational
		1 Int

	@see
	-{Int.__div}
	-{Long.__div}
	-{Rational.__div}
	-{Float.__div}
	-{Big.__div}
	-{Complex.__div}
*/
/* abstract */



BINOP_int      (div, qu_number)
BINOP_float    (div, qu_number)
BINOP_long     (div, qu_number)
BINOP_rational (div, qu_number)
BINOP_big      (div, qu_number)
BINOP_complex  (div, qu_complex)
BINOP_inf      (div, qu_number)
BINOP_integer  (div, qu_number)
BINOP_decimal  (div, qu_number)
BINOP_real     (div, qu_number)
BINOP_number   (div, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x: Number) -> Boolean __abstract
	This method returns true if v{x} is a Number and has the same numerical
	value as self, false otherwise.

	This method is implemented by all Number classes.

	If v{x} is not of the same Number class as self either self or v{x}
	is I{morphed} into the class of the other as explained in the -{Number}
	class documentation above.

	If v{x} is not a Number then C{x.__cmp (self)} is called via an internal
	delegator that makes sure that the result is a -{Boolean} otherwise calls
	-{Obj.__eq}.

	@p
	See the #{eq} method for a version that throws if v{x} is not a Number.

	@noexc

	@xmp
		x = 1
		y = 1.0
		println (x == y)

	@out
		true

	@see
	-{Int.__eq}
	-{Long.__eq}
	-{Rational.__eq}
	-{Float.__eq}
	-{Big.__eq}
	-{Complex.__eq}
*/
/* abstract */



BINOP_int      (eq, int)
BINOP_float    (eq, int)
BINOP_long     (eq, int)
BINOP_rational (eq, int)
BINOP_big      (eq, int)
BINOP_complex  (eq, int)
BINOP_inf      (eq, int)
BINOP_integer  (eq, int)
BINOP_decimal  (eq, int)
BINOP_real     (eq, int)
BINOP_number   (eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this Number.

	This method is implemented by all Number classes.

	@warn
	Since 1.09 Qu no longer uses the C formatting convention.

	Unlike C, all Qu Number objects share the same format specifications.
	For example, you can format a Float or an Int using the same format
	specification.

	@exc
	.EFormat  invalid format specification
	.EArgValue   v{fmt} is not a -{String}

	@p
	v{fmt} is: C{["%"]["-"][sign][width]["."prec][","group][mode]}

	The first format character C{"%"} is optional and is completely ignored
	but is used by the builtin -{printf} functions family.

	I{width} is the amount of spaces to use. If specified and C{"-"} is
	specified then the string is aligned to the left, the right otherwise.
	If I{width} is less then the length of the produced string then you get
	a bunch of star characters letting you know that you have an error.

	I{sign} is one of:
	@arg
	.<space> print a space if the number is positive
	.+       print the plus sign if the number is positive
	.*       do not print the sign at all

	@p
	I{prec} is the precision, i.e. the number of decimal digits.

	I{group} is the number of mantissa digits to group. If you want a thousand
	separator, for example, then you set this to 3.

	I{mode} is the formatting mode and must be one of the following:
	@arg
	.d   format without an exponent: [U{+}]xxx.xxx
	.m   monetary format: [U{+}]x,xxx,xxx.xxx
	.e   engineering format: [U{+}]x.xxxxE+x
	.E   same as "e" but use an uppercase letter for the exponent
	.f   format as decimal
	.F   same as "F" but use an uppercase letter for the exponent
	.o   format as octal
	.b   format as binary
	.X   format as hexadecimal using uppercase letters
	.x   same as "x" but use lowercase letters
	.s   use the result of the __string method
	.y   use the __true method and print yes or no

	@p
	"o", "b", "x" and "X" can only be used with -{Integer} objects.

	The "f" mode tries the "d" mode. If that fails, then tries the "e" mode.
	Thus an exponent might or might not be present.

	When used with the "e" mode, v{group} is ignored because there is always
	only one digit before the decimal point.

	The "m" mode uses a constant grouping, that is, you can not print something
	like 12,3456,78,9.00 but you can print 12,34,56,78.0000 or 1,2345,6789.00.
	Thus only the decimal point and the group separator is used despite how the
	locale defines the grouping.

	-{Rational} numbers interpret "f" and "e" as an instruction to print the
	number as a floating point number rather than a pair of numerator and
	denominator.

	@xmp
		a = 16
		println ('|', a.__format ('%8d'), '|')
		printfln ('|%8X|', a)

		a = 123456789.0
		printfln ('|%.2,3d|', a)

		a = 1 _/ 3
		printfln ("%d", a)
		printfln ("%.5f", a)

	@out
		|      16|
		|    0x10|
		|123,456,789.00|
		(1 _/ 3)
		0.33333

	@note
	This method respects the current locale if possible.
	See the m{Locale} module for more information.

	@see
	-{Int.__format}
	-{Long.__format}
	-{Rational.__format}
	-{Float.__format}
	-{Big.__format}
	-{Complex.__format}
*/
static qu_string qu__number__format (qu_number self, qu_string x)
{
	qu_format_s f;
	QU_ARG_TYPE (x, qu_string);
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_number_format (self, &f);
}



qu_string qu_number_format (qu_number x, qu_format_s *f)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_int_format      ((qu_int     ) x, f);
		case QU_CLASS_ID_Long:     return qu_long_format     ((qu_long    ) x, f);
		case QU_CLASS_ID_Rational: return qu_rational_format ((qu_rational) x, f);
		case QU_CLASS_ID_Float:    return qu_float_format    ((qu_float   ) x, f);
		case QU_CLASS_ID_Big:      return qu_big_format      ((qu_big     ) x, f);
		case QU_CLASS_ID_Complex:  return qu_complex_format  ((qu_complex ) x, f);
		case QU_CLASS_ID_Inf:      return qu_inf_format      ((qu_inf     ) x, f);
	}
	QU_FATAL ("x is not a Number object");
	return 0; /* never */
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> Int
	Return the hash value of this Number.

	This method is implemented by all Number classes and always succeed
	and return an -{Int}.

	Theoretically the result should be the same for all Number classes with
	the same numerical value. Unfortunately, due to the inexact nature of
	floating point numbers, this is not always the case.

	Thus it is generally a bad idea to use floating point numbers as keys
	in a -{Dict}.

	@noexc

	@xmp
		a = Big ('1.4')
		b = Float (a)
		printfln ('%.8f', a)
		printfln ('%.8f', b)
		println (a.__hash)
		println (b.__hash)
		d = [a: 'Hello']
		println (d.haskey (b))

	@out
		1.40000000
		1.39999999
		1
		1
		false

	@see
	-{Int.__hash}
	-{Long.__hash}
	-{Rational.__hash}
	-{Float.__hash}
	-{Big.__hash}
	-{Complex.__hash}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	&
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __iand (x: Integer) -> Integer __abstract
	Return the bitwise-and of self and v{x}.

	This method is only implemented by -{Int} and -{Long}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIand

	@xmp
		x = -123456789012345
		println (x & 1)
		println (x & ~x)
		println (-x & 1)

	@out
		-1
		0
		1
	@see
	-{Int.__iand}
	-{Long.__iand}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer __abstract
	Divide self with v{x} in an I{integer context} (the I{quotient} in
	mathematic).

	This method is implemented by all Number classes.

	@warn
	There is a big difference between the #{__div} method (the C{/} operator)
	and the __idiv method (the C{\} operator).

	There is a correlation between the __idiv and __rem methods.
	See #{__rem} for the formula.

	@p
	The __idiv method divides the numbers in an I{integer context} and always
	yield an Integer or throws EIdiv, i.e. it does not return an infinity
	when the divisor is 0.

	If v{x} is not of the same Number class as self either self or v{x}
	is I{morphed} into the class of the other as explained in the -{Number}
	class documentation above.

	The following table shows all the possible outcome of this method.

	@code
	      +-------+-------+-------+-------+-------+-------+-------+
	  \   |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  |  I*   |  I*   |  IL*  |  IL*  |  IL*  |   *   |  IL*  |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  |  IL*  |  IL*  |  IL*  |  IL*  |  IL*  |   *   |  IL*  |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  |  IL*  |  IL*  |  IL*  |  IL*  |  IL*  |   *   |  IL*  |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  |  IL*  |  IL*  |  IL*  |  IL*  |  IL*  |   *   |  IL*  |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  IL*  |  IL*  |  LI*  |  IL*  |  IL*  |   *   |  IL*  |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |   *   |   *   |   *   |   *   |   *   |   *   |   *   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  IL*  |  IL*  |  IL*  |  IL*  |  IL*  |  IL*  |  IL*  |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception

	@exc
	.EIdiv
	.EBrain

	@p
	Note that EIdiv is also thrown if the divisor is 0 or the result is too big
	which makes it impossible to return an -{Integer}.

	@xmp
		a = Long (13)
		b = Long (4)
		println (a \ b)

	@out
		3

	@see
	-{Int.__idiv}
	-{Long.__idiv}
	-{Rational.__idiv}
	-{Float.__idiv}
	-{Big.__idiv}
	-{Complex.__idiv}
*/
/* abstract */



BINOP_int      (idiv, qu_integer)
BINOP_float    (idiv, qu_integer)
BINOP_long     (idiv, qu_integer)
BINOP_rational (idiv, qu_integer)
BINOP_big      (idiv, qu_integer)
BINOP_complex  (idiv, qu_integer)
BINOP_inf      (idiv, qu_integer)
BINOP_integer  (idiv, qu_integer)
BINOP_decimal  (idiv, qu_integer)
BINOP_real     (idiv, qu_integer)
BINOP_number   (idiv, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Number __abstract
	Return the result of adding 1 to self.

	This method is implemented by all Number classes.

	Note that for -{Decimal} numbers, the result is a Decimal even if the number
	has no fractional part.

	If the result is huge and the Number.#{context} does not specify ALLOW_INF
	then this method throws EHuge.

	Notice that ETiny is not possible.

	@exc
	.EHuge
	.EInf

	@xmp
		a = 1
		++a
		println (a)

	@out
		2

	@p
	Special cases:
	@ul
	.C{infp.__incr --> infp}
	.C{infn.__incr --> infn}

	@see
	-{Int.__incr}
	-{Long.__incr}
	-{Rational.__incr}
	-{Float.__incr}
	-{Big.__incr}
	-{Complex.__incr}
*/
/* abstract */



UNOP (incr, qu_number)


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	~
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __inot -> Integer __abstract
	Return the bitwise inversion of self.

	This method is only implemented by -{Int} and -{Long}.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@noexc

	@xmp
		x = 12345678901234
		println (~x)

		x = -x
		println (~x)

	@out
		18446731728030650381
		-18446731728030650381

	@see
	-{Int.__inot}
	-{Long.__inot}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	|
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __ior (x: Integer) -> Integer __abstract
	Return the bitwise-or of self and v{x}.

	This method is only implemented by -{Int} and -{Long}.

	If v{x} is not an -{Integer} or is negative then EIor is thrown.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIor

	@xmp
		x = 12345678901234
		println (x | 4)

	@out
		12345678901238

	@see
	-{Int.__ior}
	-{Long.__ior}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	^
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __ixor (x: Integer) -> Integer __abstract
	Return the bitwise-xor of self and v{x}.

	This method is only implemented by -{Int} and -{Long}.

	If v{x} is not an -{Integer} or is negative then EIxor is thrown.

	If the result fits in an -{Int} then an Int is returned.
	A Long otherwise.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.EIxor

	@xmp
		x = 12345678901234
		println (x ^ 16)

	@out
		12345678901218

	@see
	-{Int.__ixor}
	-{Long.__ixor}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x) -> Boolean __abstract
	This method returns true if v{x} is a Number and has the same absolute
	numerical value as the absolute value of self, false otherwise.

	This method is implemented by all Number classes.

	If v{x} is not of the same Number class as self either self or v{x}
	is I{morphed} into the class of the other as explained in the -{Number}
	class documentation above.

	If v{x} is not a Number then C{x.__cmp (self)} is called via an internal
	delegator that makes sure that the result is a -{Boolean} otherwise calls
	-{Obj.__like}.

	@p
	See the #{like} method for a version that throws if v{x} is not a Number.

	@noexc

	@xmp
		x = 1
		y = -1.0
		println (x === y)

	@out
		true

	@see
	-{Int.__like}
	-{Long.__like}
	-{Rational.__like}
	-{Float.__like}
	-{Big.__like}
	-{Complex.__like}
*/
/* abstract */



BINOP_int      (like, int)
BINOP_float    (like, int)
BINOP_long     (like, int)
BINOP_rational (like, int)
BINOP_big      (like, int)
BINOP_complex  (like, int)
BINOP_inf      (like, int)
BINOP_integer  (like, int)
BINOP_decimal  (like, int)
BINOP_real     (like, int)
BINOP_number   (like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<<
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __lshift (x: Integer) -> Integer __abstract
	Return the result of left-shifting the bits of self by the amount of
	v{x}.

	If v{x} is not an -{Integer} or is negative then ELshift is thrown.

	This method is only implemented by -{Int} and -{Long}.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.ELshift
	.EHuge

	@xmp
		x = Long (-1)
		println (x << 3)

	@out
		-8

	@see
	-{Int.__lshift}
	-{Long.__lshift}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number __abstract
	Return the result of subtracting self with v{x}.

	If v{x} is not of the same Number class as self then either self
	or v{x} is morphed into the other class as explained in the -{Number}
	class documentation above.

	The following table shows all the possible outcome of this method.
	Note that this is the same as the #{__plus} method.

	Notice that Int - Int always succeed.

	@code
	      +-------+-------+-------+-------+-------+-------+-------+
	   -  |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  |  IL   | ILN*  | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  | ILN*  | ILN*  | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  |  FBN* |  FBN* | FBN*  |  BN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |  N*   |  N*   |  N*   |  N*   |  N*   |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception

	@exc
	.EMinus
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 5.2
		b = 3
		println (a - b)

	@out
		2.2

	@see
	-{Int.__minus}
	-{Long.__minus}
	-{Rational.__minus}
	-{Float.__minus}
	-{Big.__minus}
	-{Complex.__minus}
*/
/* abstract */



BINOP_int      (minus, qu_number)
BINOP_float    (minus, qu_number)
BINOP_long     (minus, qu_number)
BINOP_rational (minus, qu_number)
BINOP_big      (minus, qu_number)
BINOP_complex  (minus, qu_complex)
BINOP_inf      (minus, qu_number)
BINOP_integer  (minus, qu_number)
BINOP_decimal  (minus, qu_number)
BINOP_real     (minus, qu_number)
BINOP_number   (minus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number __abstract
	Return the result of multiplying self with v{x}.

	This method is implemented by all Number classes.

	If v{x} is not of the same Number class as self then either self
	or v{x} is morphed into the other class as explained in the -{Number}
	class documentation above.

	The following table shows all the possible outcome of this method.
	Note that Int * Int, Int * Float, Float * Float always succeed.

	@code
	      +-------+-------+-------+-------+-------+-------+-------+
	   *  |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  |  IL   | ILN*  | ILRN* |  FB   |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  | ILN*  | ILN*  | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  |  FB   |  FBN* | FBN*  |  BN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |  N*   |  N*   |  N*   |  N*   |  N*   |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception

	@exc
	.EMul
	.EHuge
	.ETiny
	.EInf
	.EBrain

	@xmp
		a = 1.2
		b = 3
		println (a * b)

	@out
		3.6

	@see
	-{Int.__mul}
	-{Long.__mul}
	-{Rational.__mul}
	-{Float.__mul}
	-{Big.__mul}
	-{Complex.__mul}
*/
/* abstract */



BINOP_int      (mul, qu_number)
BINOP_float    (mul, qu_number)
BINOP_long     (mul, qu_number)
BINOP_rational (mul, qu_number)
BINOP_big      (mul, qu_number)
BINOP_complex  (mul, qu_complex)
BINOP_inf      (mul, qu_number)
BINOP_integer  (mul, qu_number)
BINOP_decimal  (mul, qu_number)
BINOP_real     (mul, qu_number)
BINOP_number   (mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> self __abstract
	Return the negation of self.

	This method is implemented by all Number classes.

	Qu considers __neg as an I{identity} method.
	The method always succeed and the result is always the same class as self.

	@noexc

	@xmp
		x = 1.2
		println (-x)

	@out
		-1.23

	@see
	-{Int.__neg}
	-{Long.__neg}
	-{Rational.__neg}
	-{Float.__neg}
	-{Big.__neg}
	-{Complex.__neg}
*/
/* abstract */



UNOP (neg, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number __abstract
	Return the result of adding v{x} to self.

	This method is implemented by all Number classes.

	If v{x} is not of the same Number class as self then either self
	or v{x} is morphed into the other class as explained in the -{Number}
	class documentation above.

	The following table shows you all the possible outcome of this method.
	Notice that Int + Int always succeed.

	@code

	      +-------+-------+-------+-------+-------+-------+-------+
	   +  |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  |  IL   | ILN*  | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  | ILN*  |  LN*  | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  |  FB   | FBN*  | FBN*  | FBN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |  N*   |  N*   |  N*   |  N*   |  N*   |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception

	@exc
	.EPlus
	.EHuge
	.EInf
	.EBrain

	@xmp
		x = 1234567890123456
		println (x + 1)

	@out
		1234567890123457

	@see
	-{Int.__plus}
	-{Long.__plus}
	-{Rational.__plus}
	-{Float.__plus}
	-{Big.__plus}
	-{Complex.__plus}
*/
/* abstract */



BINOP_int      (plus, qu_number)
BINOP_float    (plus, qu_number)
BINOP_long     (plus, qu_number)
BINOP_rational (plus, qu_number)
BINOP_big      (plus, qu_number)
BINOP_complex  (plus, qu_complex)
BINOP_inf      (plus, qu_number)
BINOP_integer  (plus, qu_number)
BINOP_decimal  (plus, qu_number)
BINOP_real     (plus, qu_number)
BINOP_number   (plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pos -> self
	Return a positive version of self.

	This method is implemented by all Number classes.
	It always succeed and always return self.

	@xmp
		a = 10
		b = +a
		println (b)

	@out
		10

	@see
	-{Int.__pos}
	-{Long.__pos}
	-{Rational.__pos}
	-{Float.__pos}
	-{Big.__pos}
	-{Complex.__pos}
*/
qu_number qu__number__pos (qu_number self)
{
	return self;
}



UNOP (pos, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x: Number) -> Number __abstract
	Return the result of raising self to the power of v{x}.

	This method is implemented by all Number classes.

	If v{x} is not of the same Number class as self then either self
	or v{x} is morphed into the other class as explained in the -{Number}
	class documentation above.

	The following table shows you all the possible outcome of this method.

	@code

	      +-------+-------+-------+-------+-------+-------+-------+
	  **  |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  IN*  |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  LN*  |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  IN*  |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  |  FB   | FBN*  | FBN*  | FBN*  |  BN*  |  FN*  |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |  IN*  |  LN*  |  IN*  |  FN*  |  BN*  |  N*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception

	@exc
	.EPow
	.EHuge
	.ETiny
	.EInf

	@xmp
		a = 3
		println (a ** 2)
		println (a ** -2)

	@out
		9
		(1 _/ 9)

	@note
	This method may require the Math module. If it is not already loaded then
	it will be automatically loaded so you do not need to specifically
	declare C{"use Math"}.

	@see
	-{Int.__pow}
	-{Long.__pow}
	-{Rational.__pow}
	-{Float.__pow}
	-{Big.__pow}
	-{Complex.__pow}
*/
/* abstract */



BINOP_int      (pow, qu_number)
BINOP_float    (pow, qu_number)
BINOP_long     (pow, qu_number)
BINOP_rational (pow, qu_number)
BINOP_big      (pow, qu_number)
BINOP_complex  (pow, qu_complex)
BINOP_inf      (pow, qu_number)
BINOP_integer  (pow, qu_number)
BINOP_decimal  (pow, qu_number)
BINOP_real     (pow, qu_number)
BINOP_number   (pow, qu_number)


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print -> Null __abstract
	Print self to -{Stream.stdout}.

	This method simply prints the string returned by the #{__string} method.

	See -{print} for possible exceptions.

	@xmp
		a = 123
		println (a)

	@out
		123

	@see
	-{Int.__print}
	-{Long.__print}
	-{Rational.__print}
	-{Float.__print}
	-{Big.__print}
	-{Complex.__print}
*/
/* abstract */


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The definition of "remainder" might differ with other languages.

	@p
	The remainder is calculated using this formula:
	@def
		REM(a, b) = a - b * IDIV(a, b)

	@p
	You might want to notice that the result always carry the sign of v{a}
	because the sign of v{b} eliminates in the second term:

	@def
		REM(a, b) = SGN(a) * |a| - SGN(a) * |b| * IDIV(|a|, |b|)
		          = SGN(a) * (|a| - |b| * IDIV(|a| - |b|))

	@code
	      +-------+-------+-------+-------+-------+-------+-------+
	   %  |   I   |   L   |   R   |   F   |   B   |   N   |   C   |
	+=====+=======+=======+=======+=======+=======+=======+=======+
	|  I  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  I*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  L  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  L*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  R  | ILRN* | ILRN* | ILRN* | FBN*  |  BN*  |  R*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  F  | FBN*  | FBN*  | FBN*  | FBN*  |  BN*  |  F*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  B  |  BN*  |  BN*  |  BN*  |  BN*  |  BN*  |  B*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  N  |   *   |   *   |   *   |   *   |   *   |   *   |   *   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	|  C  |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |  C*   |
	+-----+-------+-------+-------+-------+-------+-------+-------+
	I=Int L=Long R=Rational F=Float B=Big N=infinity C=Complex *=exception

	@exc
	.EDiv
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 10
		println (a % 3)
		a = -10
		println (a % 4)

	@out
		1
		-2

	@see
	-{Int.__rem}
	-{Long.__rem}
	-{Rational.__rem}
	-{Float.__rem}
	-{Big.__rem}
	-{Complex.__rem}
*/
/* abstract */



BINOP_int      (rem, qu_number)
BINOP_float    (rem, qu_number)
BINOP_long     (rem, qu_number)
BINOP_rational (rem, qu_number)
BINOP_big      (rem, qu_number)
BINOP_complex  (rem, qu_complex)
BINOP_inf      (rem, qu_number)
BINOP_integer  (rem, qu_number)
BINOP_decimal  (rem, qu_number)
BINOP_real     (rem, qu_number)
BINOP_number   (rem, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	>>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rshift (x: Integer) -> Integer
	Return the bitwise-right-shift of self and v{x}.

	If v{x} is not an -{Integer} or is negative then ERshift is thrown.

	This method is only implemented by -{Int} and -{Long}.

	@warn
	In Qu the sign bit is never included in bitwise operations.

	@exc
	.ERshift

	@xmp
		x = 1234567890123456
		println (x >> 1)

	@out
		617283945061728

	@see
	-{Int.__rshift}
	-{Long.__rshift
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Return a string representation of self.

	This method is implemented by all Number classes.

	@noexc

	@xmp
		s = Long (123) as String
		println (s)

	@out
		123

	@see
	-{Int.__string}
	-{Long.__string}
	-{Rational.__string}
	-{Float.__string}
	-{Big.__string}
	-{Complex.__string}
*/
/* abstract */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	Return false if self is 0, true otherwise.

	This method is implemented by all Number classes.

	@noexc

	@xmp
		l = -1
		println (l ? 'yep' : 'nope')

	@out
		yep

	@see
	-{Int.__true}
	-{Long.__true}
	-{Rational.__true}
	-{Float.__true}
	-{Big.__true}
	-{Complex.__true}
*/
/* abstract */



/*----------------------------------------------------------------------------
	specific instance methods - comparison
----------------------------------------------------------------------------*/

#define STRICTCMP(self, x, func) \
QU_BEGIN \
	QU_REQUIRE (QuECmp, qu_number_is (x)); \
	return func (self, x); \
QU_END


/*) sub cmp (x) -> Int
	This method does the same thing as the C{<=>} operator (the #{__cmp} method)
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x <=> 2)
		println (x <=> [])
		println (x.cmp (2))
		println (x.cmp ([]))

	@out
		-1
		-1
		-1
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
qu_int qu__number_cmp (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_cmp);
}



/*) sub eq (x) -> Boolean
	This method does the same thing as the C{==} operator (the #{__eq} method)
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x == 2)
		println (x == [])
		println (x.eq (2))
		println (x.eq ([]))

	@out
		false
		false
		false
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
static qu_boolean qu__number_eq (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_eq);
}



/*) sub ge (x) -> Boolean
	This method does the same thing as the C{>=} operator
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x >= 2)
		println (x >= [])
		println (x.ge (2))
		println (x.ge ([]))

	@out
		false
		false
		false
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
static qu_boolean qu__number_ge (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_ge);
}



/*) sub gt (x) -> Boolean
	This method does the same thing as the C{>} operator
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x > 2)
		println (x > [])
		println (x.gt (2))
		println (x.gt ([]))

	@out
		false
		false
		false
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
static qu_boolean qu__number_gt (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_gt);
}



/*) sub le (x) -> Boolean
	This method does the same thing as the C{<=} operator
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x <= 2)
		println (x <= [])
		println (x.le (2))
		println (x.le ([]))

	@out
		true
		true
		true
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
static qu_boolean qu__number_le (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_le);
}



/*) sub lt (x) -> Boolean
	This method does the same thing as the C{<} operator
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x < 2)
		println (x < [])
		println (x.lt (2))
		println (x.lt ([]))

	@out
		true
		true
		true
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
static qu_boolean qu__number_lt (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_lt);
}



/*) sub like (x) -> Boolean
	This method does the same thing as the C{===} operator (the #{__like} method)
	except that it throws ECmp if v{x} is not a Number.

	@exc
	.ECmp

	@xmp
		x = 1
		println (x === 2)
		println (x === [])
		println (x.like (2))
		println (x.like ([]))

	@out
		false
		false
		false
		[1024] __main__ line 5 (ECmp) objects not comparable
*/
static qu_boolean qu__number_like (qu_number self, qu_number x)
{
	STRICTCMP (self, x, qu_sys_like);
}



/*----------------------------------------------------------------------------
	rounding
----------------------------------------------------------------------------*/

UNOP (floor, qu_number)
UNOP (ceil , qu_number)
UNOP (trunc, qu_number)
UNOP (frac , qu_number)


qu_number qu_number_round (qu_number x, int prec)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_int_round      ((qu_int     ) x, prec);
		case QU_CLASS_ID_Long:     return (qu_number) qu_long_round     ((qu_long    ) x, prec);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_round ((qu_rational) x, prec);
		case QU_CLASS_ID_Float:    return (qu_number) qu_float_round    ((qu_float   ) x, prec);
		case QU_CLASS_ID_Big:      return (qu_number) qu_big_round      ((qu_big     ) x, prec);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_inf_round      ((qu_inf     ) x, prec);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_complex_round  ((qu_complex ) x, prec);
	}
	QU_FATAL ("x is not a Number object");
	return 0;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_number__setup ()
{
	qu_class_setup_base (QuNumberClass, qu_object_s);
}



void qu_number__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuNumberClass, QU_NAME_new, qu__number_new, 1);

	/* specific class methods */
	qu_class_method_classz (QuNumberClass, "context", qu__number_context, -1);

	/* standard instance methods */
	qu_class_method_inst   (QuNumberClass, QU_NAME__pos   , qu__number__pos   , 0);
	qu_class_method_inst   (QuNumberClass, QU_NAME__format, qu__number__format, 1);

	/* specific instance methods */
	qu_class_method_instz  (QuNumberClass, "cmp" , qu__number_cmp , 1);
	qu_class_method_instz  (QuNumberClass, "eq"  , qu__number_eq  , 1);
	qu_class_method_instz  (QuNumberClass, "like", qu__number_like, 1);
	qu_class_method_instz  (QuNumberClass, "le"  , qu__number_le  , 1);
	qu_class_method_instz  (QuNumberClass, "lt"  , qu__number_lt  , 1);
	qu_class_method_instz  (QuNumberClass, "ge"  , qu__number_ge  , 1);
	qu_class_method_instz  (QuNumberClass, "gt"  , qu__number_gt  , 1);

	/* finalize declaration */
	qu_class_enddecl (QuNumberClass);
}
