/* Object.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*
	!!! The Object class must only define methods, no properties.
	!!! Otherwise Compile.c:_var must do extra check.
*/


/*) class Object __final
	The Object class is implicitly the base of all class.
	It provides default methods inherited by all classes.

	The class is marked k{final} so you can not use this class as a super class.
	Note however that the k{is} operator test against the Object class always
	return true, but not the other way around.

	@xmp
		a = [1, 2, 3]
		println (Array is Object)
		println (a is Object)
		println (Object is a)

	@output
		true
		true
		false

	@note
	You can create instances of this class directly via the #{new} method
	although this is probably not a very good idea.
*/
qu_class QuObjectClass;


/*-----------------------------------------------------------------------------
	standard static methods
-----------------------------------------------------------------------------*/

/*) sub static __attrs (*args) -> Any __public __final
	@def
	sub static __attrs (name: String) -> Any
	sub static __attrs -> Dict
	@p
	Get attribute or a -{Dict} containing all attributes of self.

	If called with one string argument then if the attribute is a method,
	a -{Method} object is returned. Otherwise the property value is returned.

	@exc
	.EArgValue
	.EImplement
	.ESecurity

	@warn
	Be careful when using the objects. You must have a good knowledge of Qu
	internals to manipulate them.

	Be extra careful when you use -{Sub} objects. You should not call
	them nor assign them to arbitrary objects or your program may crash.
	C methods do not check for a valid self argument. They just take
	it for granted that self will have the expected value.

	@p
	This has to do with performance. If C methods need to check for a valid
	self everytime they are called then performance drops.

	That is why this method is marked as k{__unsafe}, i.e. an untrusted
	k{eval} module will not be able to call this method.

	Here is a simple example that will cause a crash (if it does not then
	just know that it should).

	@xmp
		class Foo
			var func
			sub __init (f)
				self.func = f
			;;
		;;

		f = Array.__attrs ['push']  # ref to a C method
		x = Foo (f)
		x.func (123)

	@p
	So why does it crash?

	@prog
		Any array_push (Array self, Any x)
		{
			...
		}

	@p
	The f{array_push} function will be called with a self object that is not
	an -{Array} but the function thinks it is. It simply takes it for granted
	that self is an Array with a valid and initialized C struct.

	Apparently not. self is a v{Foo}.

	Naturally we could force the interpreter to do a check, or make the
	function safe like so:
	@prog

		Any Array_push (Array self, Any x)
		{
			REQUIRE (ECall, ArrayIs (self));
			...
		}

	@p
	But we don't, because that would be a terrible idea, a pain in the neck
	for developers, and a performance loss.
*/
static qu_any qu__object__attrs (qu_any self, int argc, qu_any *argv)
{
	__qu_unsafe__;

	QU_ARG_VCOUNT (0, 1);

	if (argc == 0)
	{
		qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
		qu_dict  d = qu_dict_new (0);

		qu_attr_merge_to_dict (c->attrs, d);
		return d;
	}
	else
	{
		qu_string name;
		qu_class  c;
		qu_any    x;

		QU_ARG_VTYPE (0, name, qu_string);
		if (qu_class_is (self))
		{
			c = (qu_class) self;
			x = QuClassClass->get (self, name, 0);
		}
		else
		{
			c = qu_classof (self);
			x = c->get (self, name, 0);
		}
		QU_ENSURE (x);

		if (qu_sub_is (x))
		{
			if (qu_sub_is_class (x))
			{
				x = qu_method_new (x, c);
			}
			else if (qu_sub_is_inst (x))
			{
				if (qu_class_is (self))
						QU_THROWZ (QuEImplement, "cant access instance attribute via class");
				x = qu_method_new (x, self);
			}
			qu_method_set_delegate (x);
/*! c->get already retrieves the value in case of qu_var */
#if 0
			return x;
#endif
		}
#if 0
		else if (qu_var_is (x))
		{
			return ((qu_var) x)->value;
		}
		QU_THROW (QuEImplement);
#endif
		return x;
	}
}



/*) sub static __block -> Any __public __final
	Return the block where the class or object is declared.

	The returned value depends on what self is:
	@ops
	.| -{Module} | Return self.
	.| -{Class}  | Return its Module object.

	.| -{Sub} |
		Return its Class if it is a method, its Module if it is a
		module level function, its parent Sub if it is a nested function.

	.| -{Var} |
		Return its Class if it is a property, its Module if it is
		a module level variable, its Sub if it is a lexical variable.

	.| -{Code} |  return its Class or Module

	@noexc

	@xmp
		# main program /opt/qu/demo/test.qu
		class Foo
			var bar
			sub baz
			;;
		;;

		sub foo
		;;

		a = Foo.__attrs
		println (Foo.__block)
		println (a['bar'].__block)
		println (a['baz'].__block)
		f = @foo
		println (f.__block)

	@out
		<Module __main__>
		<Class __main__.Foo>
		<Class __main__.Foo>
		<Module __main__>
*/
static qu_symbol qu__object__block (qu_any self)
{
	qu_class c = qu_classof (self);

	if      (c == QuClassClass ) return (qu_symbol) ((qu_class) self)->parent;
	else if (c == QuModuleClass) return (qu_symbol) self;
	else if (c == QuVarClass   ) return ((qu_var) self)->parent;
	else if (c == QuSubClass   ) return ((qu_sub) self)->parent;
	else if (c == QuCodeClass  ) return ((qu_code) self)->parent;
	else                         return (qu_symbol) c;
}



/*) sub static __class __public __final
	Return the Class object of self.

	@noexc

	@xmp
		a = [1 2 3]
		println (a.__class)
		println (a.__class.__class)
		println (Array.__class)
		println (Class.__class)

	@out
		<Class Sys.Array>
		<Class Sys.Class>
		<Class Sys.Class>
		<Class Sys.Class>
*/
static qu_class qu__object__class (qu_any self)
{
	return qu_classof (self);
}



/*) sub __decl -> String __final
	Return the declaration string if any, an empty string if none.

	@noexc

	@xmp
		# run with the -D option
		sub foo (x)
			"""foo documentation"""
			return 1
		;;

		f = @foo
		println (f.__decl)

	@out
		sub foo (x)
*/
static qu_string qu__object__decl (qu_any self)
{
	qu_module m = qu_object_get_module (self);

	if (m->doc)
	{
		qu_array a = (qu_array) qu_dict_getitem (m->doc, self);

		if (a && qu_array_is (a) && a->len)
			return a->objs[0];
	}
	return QuStringEMPTY;
}



/*) sub static __defines (*names) -> Boolean __public __final
	Return true if it the object defines all the specified attributes,
	false otherwise.

	Any attribute will be reported, including private ones.

	@note
	This method does not give any indication whether the attribute is
	a property, method, static, class or instance attribute. You have to
	use the methods provided by each class if necessary, e.g. -{Sub.isclass}.

	@noexc

	@xmp
		sub once foo
			return 1
		;;

		f = @foo
		println (f.__defines ('ident', '__attrs'))

	@out
		true
*/
static qu_boolean qu__object__defines (qu_any self, int argc, qu_any *argv)
{
	qu_class c = qu_classof (self);
	qu_any   x;

	QU_ARG_VMIN (1);
	while (argc--)
	{
		x = *argv++;
		if (!qu_string_is (x))
			return QuFalse;

		__qu_attr_nothrow_begin;
		x = c->get (self, (qu_string) x, 0);
		__qu_attr_nothrow_end;
		if (x == 0)
			return QuFalse;
	}
	return QuTrue;
}



/*) sub __doc -> String __final
	Return the documentation string if any, an empty string if none.

	@noexc

	@xmp
		# run with the -D option
		sub foo (x)
			"""foo documentation"""
			return 1
		;;

		f = @foo
		println (f.__doc)

	@out
		foo documentation
*/
static qu_string qu__object__doc (qu_any self)
{
	qu_module m = qu_object_get_module (self);
	if (m->doc)
	{
		qu_array a = (qu_array) qu_dict_getitem (m->doc, self);
		if (a && qu_array_is (a) && a->len > 1)
			return a->objs[1];
	}
	return QuStringEMPTY;
}



/*) sub static __ident -> String
	Return the identifier associated with this object.

	For a Class, Sub, and Var then return the name.
	For a Module, return C{"__main__"} if it is the main module otherwise
	return the name of the Module.

	For others, EImplement is thrown.
*/
static qu_string qu__object__ident (qu_any self)
{
	switch (qu_classid (self))
	{
		case QU_CLASS_ID_Class:
		case QU_CLASS_ID_Sub:
		case QU_CLASS_ID_Var:
			return ((qu_symbol) self)->name;

		case QU_CLASS_ID_Module:
			if (self == (qu_any) QuModuleMain)
				return qu_string_newz ("__main__");
			else
			{
				char *name  = ((qu_module) self)->name->str;
				char *dot   = rindex (name, '.');
				char *slash = rindex (name, '/');

				if (dot == 0)
					dot = name + ((qu_module) self)->name->len;
				if (slash == 0)
					slash = name;
				else
					++slash;
				return qu_string_new (slash, dot - slash);
			}
	}
	QU_THROW (QuEImplement);
}



/*) sub static __isa (x) -> Boolean __public __final
	This method returns true if the object is a v{x}, false otherwise.

	v{x} may be anything. If it is not a class then its class is used.
	If self is a -{Record} of -{Struct} v{x} then true is returned.

	@noexc

	@xmp
		class Foo
		;;

		class Bar is Foo
		;;

		println (Bar.__isa (Class))
		println (Bar.__isa (Foo))
		x = Bar ()
		println (x.__isa (Foo))

	@out
		true
		true
		true
*/
static qu_boolean qu__object__isa (qu_any self, qu_any x)
{
	return qu_object_isa (self, x) ? QuTrue : QuFalse;
}



/*) sub static __isabstract -> Boolean __public
	Return true if this is an __abstract class, false otherwise.

	@noexc
*/
static qu_boolean qu__object__isabstract (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	return qu_class_is_abstract (c) ? QuTrue : QuFalse;
}



/*) sub static __isauto -> Boolean __public
	Return true if self is an __auto class, false otherwise.

	If self is not a Class then its class is used.

	@noexc
*/
static qu_boolean qu__object__isauto (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	return qu_class_is_auto (c) ? QuTrue : QuFalse;
}



/*) sub static __isextern -> Boolean __public
	Return true if self is a __final class, false otherwise.

	If self is not a Class then its class is used.

	External classes are builtin classes and those from external modules
	(modules written in C or other language).

	@noexc
*/
static qu_boolean qu__object__isextern (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	return qu_class_is_extern (c) ? QuTrue : QuFalse;
}



/*) sub static __isfinal -> Boolean __public
	Return true if self is a __final class, false otherwise.

	If self is not a Class then its class is used.

	@noexc
*/
static qu_boolean qu__object__isfinal (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	return qu_class_is_final (c) ? QuTrue : QuFalse;
}



/*) sub static __ismixin -> Boolean __public
	Return true if self is a __mixin class, false otherwise.

	If self is not a Class then its class is used.

	@noexc
*/
static qu_boolean qu__object__ismixin (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	return qu_class_is_mixin (c) ? QuTrue : QuFalse;
}



/*) sub static __isprivate -> Boolean __public
	Return true if the object is k{private}, false otherwise.

	@noexc
*/
static qu_boolean qu__object__isprivate (qu_any self)
{
	if (qu_class_is (self) || qu_sub_is (self) || qu_var_is (self))
	{
		return qu_symbol_is_private (self) ? QuTrue : QuFalse;
	}
	else if (qu_method_is (self))
	{
		self = ((qu_method) self)->sub;
		return qu_symbol_is_private (self) ? QuTrue : QuFalse;
	}
	else
	{
		qu_class c = qu_classof (self);
		return qu_class_is_private (c) ? QuTrue : QuFalse;
	}
}



/*) sub static __issuper (x) -> Boolean __public
	Return true if the v{x} is a super class of this class, false otherwise.

	If self is not a Class then its class is used.
	If v{x} is not a Class then its class is used.

	@exc
	.EArgValue
*/
static qu_boolean qu__object__issuper (qu_any self, qu_any x)
{
	qu_class c = (qu_class_is (self) ? (qu_class) self : qu_classof (self));
	qu_class k = (qu_class_is (x)    ? (qu_class) x    : qu_classof (x));
	qu_class u = c->super;

	while (u)
	{
		if (u == k)
			return QuTrue;
		u = u->super;
	}
	return QuFalse;
}



/*) sub static __iterable -> Boolean __public __final
	Whether the object is iterable.

	@noexc

	@xmp
		a = [1 2 3]
		println (a.__iterable)
		println (Array.__iterable)

	@out
		true
		false
*/
static qu_boolean qu__object__iterable (qu_any self)
{
	return qu_object_iterable (self) ? QuTrue : QuFalse;
}



/*) sub static __master -> Any __public __final
	Return the master block where the object/class is declared.

	Unless the object is a nested function or local variable then this
	method returns the same result as the #{__block} method.

	@noexc

	@xmp
		sub foo (x)
			sub bar
				println ('Hello ' x '!')
			;;
			return @bar
		;;

		f = foo ('World')
		f ()
		println (f.__master)

	@out
		Hello World!
		<Sub __main__.foo>
*/
static qu_symbol qu__object__master (qu_any self)
{
	qu_class c = qu_classof (self);
	switch (c->id)
	{
		case QU_CLASS_ID_Module: return (qu_symbol) self;
		case QU_CLASS_ID_Class:  return (qu_symbol) ((qu_class) self)->parent;
		case QU_CLASS_ID_Sub:    return ((qu_sub) self)->parent;
		case QU_CLASS_ID_Method: return ((qu_method) self)->sub->parent;
		case QU_CLASS_ID_Var:    return ((qu_var) self)->parent;
		case QU_CLASS_ID_Frame:  return ((qu_frame) self)->code->parent;
	}
	return (qu_symbol) c;
}



/*) sub static __mixins -> Array __public __final
	Return the classes used as mixins by this class or the class of this object.
	The returned value is a frozen Array.

	@noexc

	@xmp
		class Foo
		;;

		class Bar
		;;

		class Baz (Foo, Bar)
		;;

		println (Bar.__mixins)
		x = Baz ()
		println (x.__mixins)

	@out
		[<Class __main__.Foo>, <Class __main__.Bar>]
		[<Class __main__.Foo>, <Class __main__.Bar>]
*/
static qu_array qu__object__mixins (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);

	if (c->mixin == 0)
		return QuArrayEmpty;

	/* CHECKME - class->mixins
		if we're here, there is no longer a possibility to change self->mixins */
	qu_array_freeze (c->mixin);
	return c->mixin;
}



/*) sub __mixes (x: Class) -> Boolean __public
	Return true if the class v{x} is used as a mixin of self, false otherwise.

	If self is not a class then its class is used.

	This is more efficient than saying C{x in self.__mixin}.

	@exc
	.EArgValue
*/
qu_boolean qu__object__mixes (qu_any self, qu_class x)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);

	QU_ARG_TYPE (x, qu_class);
	return (c->mixin && qu_array_contains (c->mixin, x)) ? QuTrue : QuFalse;
}



/*) sub static __module -> Any __public __final
	Return the -{Module} object where the class or object is declared
	or originated from.

	@exc
	.ENone in case of a Frame and the Frame has stopped running

	@xmp
		class Foo
		;;

		println (Foo.__module)
		println (Foo().__module)

	@out
		<Module __main__>
		<Module __main__>
*/
qu_module qu__object__module (qu_any self)
{
	qu_class c = qu_classof (self);

	if (c == QuModuleClass) return (qu_module) self;
	if (c == QuClassClass ) return qu_class_get_module ((qu_class) self);
	if (c == QuSubClass   ) return qu_sub_get_module ((qu_sub) self);
	if (c == QuMethodClass) return qu_sub_get_module (((qu_method) self)->sub);
	if (c == QuVarClass   ) return qu_var_get_module ((qu_var) self);
	if (c == QuCodeClass  ) return qu_code_get_module ((qu_code) self);
	if (c == QuFrameClass ) return qu_frame_get_module ((qu_frame) self);
	return qu_class_get_module (c);
}



/*) sub static __name -> String __final
	Return the name of the class of this object.

	@noexc

	@out
		a = [1 2 3]
		println (Array.__name)
		println (a.__name)

	@out
		Array
		Array
*/
qu_string qu__object__name (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	return c->name;
}



/*) sub __rank -> Int
	Return the rank of the object.

	The value must not be taken literally as it may change in the future.
	The only purpose of this function is for sorting.

	Builtin classes f{__cmp} methods that can not do a comparisson call
	the Object.#{__cmp} which uses this method to do the comparisson.

	@noexc

	@xmp
		for c in [Array, Dict, 1, [1,2], {1, 2}]
			println (c.__rank)
		;;
		println (Array <=> Dict)

	@out
		1
		1
		33
		51
		53
		-1
*/
static qu_int qu__object__rank (qu_any self)
{
	return qu_int_new (qu_object_rank (self));
}



/*) sub static __root -> Class __public __final
	Return the root Class object of self.

	@noexc

	@xmp
		class Foo ;;
		class Bar is Foo ;;
		class Baz is Bar ;;

		x = Baz ()
		println (x.__root)
		println (Baz.__root)

	@out
		<Class __main__.Foo>
		<Class __main__.Foo>
*/
static qu_class qu__object__root (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);

	while (c->super)
		c = c->super;
	return c;
}



/*) sub static __storable -> Boolean __final
	Return true if the object is storable, i.e. it defines the standard
	methods f{__load} and f{__store}.

	@noexc
*/
static qu_boolean qu__object__storable (qu_any self)
{
	qu_class c = qu_classof (self);
	int      r;

	__qu_attr_nothrow_begin;
	r = c->get (self, QU_NAME__load, 0) && c->get (self, QU_NAME__store, 0);
	__qu_attr_nothrow_end;
	if (r)
		return QuTrue;
	return QuFalse;
}



/*) sub static __super -> Any __final
	Return the super class of this object, nil if none.

	@noexc

	@out
		class Foo
		;;

		class Bar is Foo
		;;

		println (Bar.__super)
		println (Bar().__super)
		println (Array.__super)

	@out
		<Class __main__.Foo>
		<Class __main__.Foo>
		nil
*/
static qu_any qu__object__super (qu_any self)
{
	if (!qu_class_is (self))
		self = qu_classof (self);
	if (((qu_class) self)->super)
		return ((qu_class) self)->super;
	return QuNil;
}



/*) sub static __tree -> Array __public __final
	Return an -{Array} consiting all the classes in this class's hierachy.

	The first item is this class.

	@noexc
*/
static qu_array qu__object__tree (qu_any self)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	qu_array a = qu_array_new (4);

	while (c)
	{
		qu_array_push_quick (a, c);
		c = c->super;
	}
	return a;
}



/*) sub __where -> String
	Return a String representation of the whereabout of the object.

	@noexc
*/
static qu_string qu__object__where (qu_any self)
{
	qu_class c = qu_classof (self);

	if (c == QuModuleClass
		|| c == QuClassClass
		|| c == QuSubClass
		|| c == QuVarClass)
	{
		return qu_object_where_string (self, "", "");
	}
	else if (c == QuMethodClass)
	{
		return qu_object_where_string (((qu_method) self)->sub, "", "");
	}
	else
	{
		return qu_object_where_string (c, "", "");
	}
}



/*-----------------------------------------------------------------------------
	class methods
-----------------------------------------------------------------------------*/

/*) sub class __inst -> Any __final
	Create a new empty instance of this class.

	You use this method if your class supports multiple constructors.

	This method is private so it is only accessible from within the class.

	@exc
	.Exception anything thrown by the constructor method

	@xmp
		class Foo
			var _bar
			var baz

			sub class new (x)
				i = self.__inst   # Qu allows you to access any private
				i._bar = x        # instance attributes from within a class sub
				i.baz = 1         # provided they are in the same class hierarchy
				return i
			;;

			sub class new2 (x, y)
				i = self.__inst
				i._bar = x
				i.baz = y
				return i
			;;
		;;

		a = Foo.new  (1)
		b = Foo.new2 (1, 2)
*/
static qu_any qu__object__inst (qu_class self)
{
	assert (!qu_class_is_extern (self));

	if (qu_class_is_abstract (self) || qu_class_is_mixin (self))
		QU_THROWF (QuEAbstract, "%s: can not instantiate class", self->name->str);
	return qu_gc__new (self, self->size);
}



/*) sub class new (*arg) -> Any
	This is the default constructor method and also a constructor for
	instances of the Object class.

	The method is smart enough to know whether to create an instance of
	the Object class or of a class that inherits this method.

	In case the class is not Object then it creates an instance of the
	class then call the f{__init} instance method, if defined.

	In case of the Object class the you must supply exactly one -{Dict}
	argument containing description of methods and properties.

	@exc
	.EArgValue
	.Exception as thrown by the f{__init} method.

	@xmp
		sub _show
			println ('the name is ' self.name)
		;;

		o = Object ('show': @_show, 'name': 'Foo')
		o.show
		println (o.name)

	@output
		the name is Foo
		Foo
*/
static qu_any qu__object_new (qu_class self, int argc, qu_any *argv)
{
	if (self == QuObjectClass)
	{
		qu_obj     o;
		qu_dict    d;
		qu_any     k, v;
		long    i, b;
		qu_attr_s *a;

		if (argc != 1 || !(qu_dict_is (argv[0]) || qu_arg_is (argv[0])))
			QU_THROWZ (QuEArgCount, "requires one Dict argument.");

		d = argv[0];
		a = qu_attr_alloc ();
		i = b = 0;

		while ((v = qu_dict_walk (d, &k, &i, &b)) != 0)
		{
			if (!qu_string_is (k))
			{
				qu_attr_free (a);
				QU_ARG_THROWZ ("non String key in Dict");
			}
			if (qu_sub_is (v))
			{
				if (!qu_sub_is_static (v))
				{
					qu_attr_free (a);
					QU_ARG_THROWZ ("non static Sub value in Dict");
				}
			}
			qu_attr_set (a, (qu_string) k, v);
		}

		o = qu_gc_new (QuObjectClass);
		if (qu_attr_len (a) != 0)
			o->attrs = a;
		else
			qu_attr_free (a);
		return o;
	}
	else
	{
		if (qu_class_is_abstract (self) || qu_class_is_mixin (self))
			QU_THROWF (QuEAbstract, "cant instantiate %s", self->name->str);

		/* C classes must define their own 'new' method */
		if (qu_class_is_extern (self))
		{
			/*HACK - documentation - see Compile.c */
			if (!qu_class_is_doc (self))
			{
				qu_sub u = QuClassClass->get (self, QU_NAME_new, 0);

				if (u == 0 || u->value == qu__object_new)
					QU_THROWF (QuEImplement, "%s.new is not implemented", self->name->str);

				assert (u && qu_sub_is (u) && qu_sub_is_class (u));
				return qu_sub_calln (u, self, argc, argv);
			}
		}

		QU_RETURN (qu_class_init_new (self, qu_gc__new (self, self->size), argc, argv));
	}
}



/*-----------------------------------------------------------------------------
	low-level instance method
-----------------------------------------------------------------------------*/

/* none */



/*-----------------------------------------------------------------------------
	instance methods
-----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{String}  | Delegate to #{__string}.
	.| -{Boolean} | Delegate to #{__true}.

	@p
	Otherwise v{x} must satisfy C{self is x} to return self.

	@exc
	.EImplement none of the above

	@xmp
		class Foo
		;;

		f = Foo ()
		println (f as Foo)

	@out
		true
		<a Foo>
*/
static qu_any qu__object__as (qu_any self, qu_any x)
{
	if (x == QuStringClass)
	{
		return qu_sys_string (self);
	}
	else if (x == QuBooleanClass)
	{
		return qu_sys_true (self);
	}
	else if (qu_sys_is (self, x) == QuTrue)
	{
		return self;
	}
	else
	{
		QU_THROWF (QuEImplement, "cant do %s as %s",
			qu_classof (self)->name->str, qu_classof (x)->name->str);
	}
}



/*) sub __cmp (x) -> Int
	If self and v{x} are the very same objects then 0 is returned.

	Otherwise the objects are compared using the #{__rank} method.

	@noexc

	@xmp
		class Foo
		;;

		f = Foo ()
		println (f <=> 'bar')

	@out
		1
*/
static qu_int qu__object__cmp (qu_any self, qu_any x)
{
	if (qu_class_is (self) && qu_class_is (x))
		return qu_string_cmpint (((qu_class) self)->name, ((qu_class) x)->name);

	return qu_object_default_cmp (self, x);
}


int qu_object_default_cmpi (qu_any self, qu_any x)
{
	/* !!! must return -1, 0, 1
		not just any arbitrary negative/positive value */

	if (self != x)
	{
		int a = qu_object_rank (self);
		int b = qu_object_rank (x);

		if (a > b)
			return 1;
		if (a < b)
			return -1;
	}
	return 0;
}



/*) sub __eq (x) -> Boolean
	This method returns true if self and k{x} are the very same object,
	false otherwise.

	@noexc

	@xmp
		class Foo
		;;

		f = Foo ()
		g = Foo ()
		println (f == g)

	@out
		false
*/
qu_boolean qu__object__eq (qu_any self, qu_any x)
{
	if (self == x && qu_classof (self) == qu_classof (x))
		return QuTrue;
	return QuFalse;
}



/*) sub __hash -> Int
	Return the hash value of this object.

	@noexc

	@xmp
		class Foo
		;;

		f = Foo ()
		println (f.__hash)

	@out
		135645172
*/
static qu_int qu__object__hash (qu_any self)
{
	return qu_int_newl ((long) self);
}


long qu_object_hash (qu_any self)
{
	static int here = 0;
	qu_class c = qu_classof (self);

	if (c->accel)
	{
		/* this already includes c == QuClassClass */
		assert (QuClassClass->accel && QuClassClass->accel->fhash);

		if (c->accel->fhash)
			return qu_int_tol (c->accel->fhash (self));
	}
	else if (here)
	{
		here = 0;
		QU_WARN ("cyclic hash");
	}
	else
	{
		qu_sub x = qu_attr_inst_getnt (self, QU_NAME__hash, 0);
		if (x)
		{
			if (x->parent != (qu_symbol) QuObjectClass)
			{
				qu_any h;

				here = 1;
				h = qu_object_call0 (self, QU_NAME__hash);
				here = 0;
				if (h)
				{
					if (qu_int_is (h))
						return qu_int_tol (h);
				}
				QU_SYS_WARN_RETVAL (c, "__hash", "Int");
			}
		}
	}
	assert (!qu_int_is (self));
	return (long) self;
}



/*) sub __init (*arg) -> Null
	This method does absolutely nothing then return nil.

	@noexc

	@xmp
		class Foo
		;;

		f = Foo ()   # actually calls Object.__init
*/
static qu_any qu__object__init (qu_any self, int argc, qu_any *argv)
{
	return QuNil;
}



/*) sub __like (x) -> Boolean
	This method returns true if self and k{x} are the very same object,
	false otherwise. Thus by default, f{__like} returns the same result
	as the f{__eq} method.

	@noexc

	@xmp
		class Foo
		;;

		f = Foo ()
		println (f == 'bar')

	@out
		false
*/
static qu_boolean qu__object__like (qu_any self, qu_any x)
{
	if (qu_class_is (self))
	{
		if (!qu_class_is (x))
			x = qu_classof (x);

		return ((self == x)
			|| qu_class_isa ((qu_class) self, (qu_class) x)
			|| qu_class_isa ((qu_class) x, (qu_class) self)) ? QuTrue : QuFalse;
	}
	else
	{
		return qu__object__eq (self, x);
	}
}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		class Foo
		;;

		x = Foo ()
		println (x)

	@out
		<a Foo>
*/
qu_null qu__object__print (qu_any self)
{
	qu_string s;

	if (qu_class_is (self))
		s = qu_class_as_string (self);
	else
		s = qu_sys_string (self);
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/*) sub __string -> String
	Return a string representation of self.

	Note that this method may be called as a fallback when an object's f{__string}
	method fails or does not return a -{String}. Failure includes, for example,
	if the string is too big.

	@noexc

	@xmp
		class Foo
		;;

		x = Foo ()
		s = x as String
		println (s)

	@out
		<a Foo>
*/
qu_string qu__object__string (qu_any self)
{
	if (qu_class_is (self))
	{
		return qu_object_where_string ((qu_symbol) self, "<Class ", ">");
	}
	else
	{
		qu_class c = qu_classof (self);
		return qu_string_cat (qu_string_byte ('<'), qu_string_byte ('a'), qu_string_byte (' '), c->name, qu_string_byte ('>'), 0);
	}
}



/*) sub __true -> Boolean
	This method always return true.

	@noexc

	@xmp
		class Foo
		;;

		x = Foo ()
		if x
			println ('yes')
		else
			println ('no')
		;;

	@out
		'yes'
*/
static qu_boolean qu__object__true (qu_any self)
{
	return qu_object_default_true (self);
}



/*) sub static __version -> String __public
	Return the version.
	If not overridden then the Qu version is returned.

	@noexc
*/
static qu_string qu__object__version (qu_any self)
{
	static qu_string v = 0;
	if (v == 0)
		v = qu_string_new_staticz (QU_APP_VERSION);
	return v;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

/* none */



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

#define STRICTCMP(self, x, func) \
QU_BEGIN \
	QU_REQUIRE (QuECmp, qu_object_isa ((x), (self))); \
	return func (self, x); \
QU_END



/*) sub cmp (x) -> Int
	This method compares self with v{x}.

	The <=> operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{cmp} does the same as the __cmp method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_int qu__object_cmp (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_cmp);
}



/*) sub eq (x) -> Boolean
	This method compares self with v{x} for equality.

	The == operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{eq} does the same as the __eq method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_boolean qu__object_eq (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_eq);
}



/*) sub ge (x) -> Boolean
	This method compares self with v{x}.

	The >= operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{ge} does the same as the __ge method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_boolean qu__object_ge (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_ge);
}



/*) sub gt (x) -> Boolean
	This method compares self with v{x}.

	The > operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{gt} does the same as the __gt method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_boolean qu__object_gt (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_gt);
}



/*) sub le (x) -> Boolean
	This method compares self with v{x}.

	The <= operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{le} does the same as the __le method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_boolean qu__object_le (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_le);
}



/*) sub lt (x) -> Boolean
	This method compares self with v{x}.

	The < operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{lt} does the same as the __lt method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_boolean qu__object_lt (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_lt);
}



/*) sub like (x) -> Boolean
	This method compares self with v{x} for flexible equality.

	The === operator in Qu does not throw an exception.
	If this is not desirable then you can use this method.

	f{like} does the same as the __like method except that ECmp is thrown if
	v{x} is not a compatible object, i.e. C{x is self} must be true.

	@exc
	.ECmp
*/
static qu_boolean qu__object_like (qu_any self, qu_any x)
{
	STRICTCMP (self, x, qu_sys_like);
}



/*-----------------------------------------------------------------------------
	Convert to static
-----------------------------------------------------------------------------*/

/*) sub __static __final
	Convert self into a static object, that is, one that is never garbage
	colected nor marked.

	If the object can not be made static then nothing happens and self
	if returned.

	@exc
	.ESecurity
*/
qu_any qu__object__static (qu_any x)
{
#ifdef QU_APP_WEAKMODULE
	return x;
#else
	qu_any y = 0;

	if (qu_gc_is_static (x))
		return x;

	__qu_unsafe__;

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Null:
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Boolean:
		case QU_CLASS_ID_Inf:
			return x;

		case QU_CLASS_ID_String:   return qu_string_to_static (x);
		case QU_CLASS_ID_Long:     return qu_long_to_static (x);
		case QU_CLASS_ID_Rational: return qu_rational_to_static (x);
		case QU_CLASS_ID_Float:    return qu_float_to_static (x);
		case QU_CLASS_ID_Big:      return qu_big_to_static (x);
		case QU_CLASS_ID_Range:    return qu_range_to_static (x);
		case QU_CLASS_ID_Array:    y = qu_array_to_static (x); break;
		case QU_CLASS_ID_Dict:     y = qu_dict_to_static  (x); break;
		case QU_CLASS_ID_Set:      y = qu_set_to_static   (x); break;
	}
	return y ? y : x;
#endif
}



/*-----------------------------------------------------------------------------
	calls

	- we allow calling properties with no argument
	- we also allow calling class methods via an instance of the class
-----------------------------------------------------------------------------*/

int qu_object_callable (qu_any self)
{
	switch (qu_classid (self))
	{
		case QU_CLASS_ID_Sub:
		case QU_CLASS_ID_Method:
		case QU_CLASS_ID_Class:
			return 1;
	}
	return 0;
}



qu_any qu_object_calln (qu_any self, qu_string name, int argc, qu_any *argv)
{
	qu_any x = qu_classof (self)->get (self, name, 0);
	QU_ENSURE (x);
	if (qu_sub_is (x))
	{
		if (qu_sub_is_class (x) && !qu_class_is (self))
			self = qu_classof (self);
		return ((qu_sub) x)->call->fn (x, self, argc, argv);
	}

	if (qu_method_is (x) && qu_method_is_delegate (x))
		return qu_method_calln (x, argc, argv);

	QU_THROW (QuECall);
}



qu_any qu_object_call4 (qu_any self, qu_string name, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any x = qu_classof (self)->get (self, name, 0);
	QU_ENSURE (x);
	if (qu_sub_is (x))
	{
		if (qu_sub_is_class (x) && !qu_class_is (self))
			self = qu_classof (self);
		return ((qu_sub) x)->call->f4 (x, self, a1, a2, a3, a4);
	}

	if (qu_method_is (x) && qu_method_is_delegate (x))
		return qu_method_call4 (x, a1, a2, a3, a4);

	QU_THROW (QuECall);
}



qu_any qu_object_call3 (qu_any self, qu_string name, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any x = qu_classof (self)->get (self, name, 0);
	QU_ENSURE (x);
	if (qu_sub_is (x))
	{
		if (qu_sub_is_class (x) && !qu_class_is (self))
			self = qu_classof (self);
		return ((qu_sub) x)->call->f3 (x, self, a1, a2, a3);
	}

	if (qu_method_is (x) && qu_method_is_delegate (x))
		return qu_method_call3 (x, a1, a2, a3);

	QU_THROW (QuECall);
}



qu_any qu_object_call2 (qu_any self, qu_string name, qu_any a1, qu_any a2)
{
	qu_any x = qu_classof (self)->get (self, name, 0);
	QU_ENSURE (x);
	if (qu_sub_is (x))
	{
		if (qu_sub_is_class (x) && !qu_class_is (self))
			self = qu_classof (self);
		return ((qu_sub) x)->call->f2 (x, self, a1, a2);
	}

	if (qu_method_is (x) && qu_method_is_delegate (x))
		return qu_method_call2 (x, a1, a2);

	QU_THROW (QuECall);
}



qu_any qu_object_call1 (qu_any self, qu_string name, qu_any a1)
{
	qu_any x = qu_classof (self)->get (self, name, 0);
	QU_ENSURE (x);
	if (qu_sub_is (x))
	{
		if (qu_sub_is_class (x) && !qu_class_is (self))
			self = qu_classof (self);
		return ((qu_sub) x)->call->f1 (x, self, a1);
	}

	if (qu_method_is (x) && qu_method_is_delegate (x))
		return qu_method_call1 (x, a1);

	QU_THROW (QuECall);
}



qu_any qu_object_call0 (qu_any self, qu_string name)
{
	qu_any x = qu_classof (self)->get (self, name, 0);
	QU_ENSURE (x);
	if (qu_sub_is (x))
	{
		if (qu_sub_is_class (x) && !qu_class_is (self))
			self = qu_classof (self);
		return ((qu_sub) x)->call->f0 (x, self);
	}

	if (qu_method_is (x) && qu_method_is_delegate (x))
		return qu_method_call0 (x);

	return x;
}



qu_any qu_object_call_xn (qu_any self, int argc, qu_any *argv)
{
	switch (qu_classid (self))
	{
		case QU_CLASS_ID_Sub:
		{
			qu_class c = qu_sub_get_class (self);
			if (c)
			{
				if (qu_sub_is_inst (self))
					break;
				return qu_sub_calln (self, c, argc, argv);
			}
			else
			{
				return qu_sub_calln (self, qu_interp_self (), argc, argv);
			}
		}

		case QU_CLASS_ID_Method:
			return qu_method_calln (self, argc, argv);

		case QU_CLASS_ID_Class:
			return qu__object_new (self, argc, argv);
	}
	QU_THROW (QuECall);
}


qu_any qu_object_call_x4 (qu_any self, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv[4] = {a1, a2, a3, a4};
	return qu_object_call_xn (self, 4, argv);
}


qu_any qu_object_call_x3 (qu_any self, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv[3] = {a1, a2, a3};
	return qu_object_call_xn (self, 3, argv);
}


qu_any qu_object_call_x2 (qu_any self, qu_any a1, qu_any a2)
{
	qu_any argv[2] = {a1, a2};
	return qu_object_call_xn (self, 2, argv);
}


qu_any qu_object_call_x1 (qu_any self, qu_any a1)
{
	return qu_object_call_xn (self, 1, &a1);
}



qu_any qu_object_call_x0 (qu_any self)
{
	return qu_object_call_xn (self, 0, NULL);
}



qu_any qu_object_call_exn (qu_any self, qu_any how, int argc, qu_any *argv)
{
	switch (qu_classid (how))
	{
		case QU_CLASS_ID_String:
			return qu_object_calln (self, how, argc, argv);

		case QU_CLASS_ID_Sub:
		{
			qu_class c = qu_sub_get_class (how);
			if (c)
			{
				if (qu_sub_is_inst (how))
					QU_REQUIRE (QuECall, qu_object_instof (self, c));
				else if (qu_sub_is_class (how))
					self = c;
				return qu_sub_calln (how, self, argc, argv);
			}
			else
			{
				switch (argc)
				{
					case 0: return qu_sub_call1 (how, qu_interp_self (), self);
					case 1: return qu_sub_call2 (how, qu_interp_self (), self, argv[0]);
					case 2: return qu_sub_call3 (how, qu_interp_self (), self, argv[0], argv[1]);
					case 3: return qu_sub_call4 (how, qu_interp_self (), self, argv[0], argv[1], argv[2]);
					default:
					{
						qu_array a = qu_array_new (argc + 1);
						QU_OBJECT_LIST_COPY (a->objs + 1, argv, argc);
						a->objs[0] = self;
						return qu_sub_calln (how, qu_interp_self (), argc + 1, a->objs);
					}
				}
			}
		}

		case QU_CLASS_ID_Method:
		{
			qu_array a = qu_array_new (argc + 1);
			QU_OBJECT_LIST_COPY (a->objs + 1, argv, argc);
			a->objs[0] = self;
			return qu_method_calln (how, argc + 1, a->objs);
		}

		case QU_CLASS_ID_Class:
			return qu__object_new (how, argc, argv);
	}
	QU_THROW (QuECall);
}



qu_any qu_object_call_ex3 (qu_any self, qu_any how, qu_any a1, qu_any a2, qu_any a3)
{
	switch (qu_classid (how))
	{
		case QU_CLASS_ID_String:
			return qu_object_call3 (self, how, a1, a2, a3);

		case QU_CLASS_ID_Sub:
		{
			qu_class c = qu_sub_get_class (how);
			if (c)
			{
				if (qu_sub_is_inst (how))
					QU_REQUIRE (QuECall, qu_object_instof (self, c));
				else if (qu_sub_is_class (how))
					self = c;
				return qu_sub_call3 (how, self, a1, a2, a3);
			}
			else
			{
				return qu_sub_call4 (how, qu_interp_self (), self, a1, a2, a3);
			}
		}

		case QU_CLASS_ID_Method:
			return qu_method_call4 (how, self, a1, a2, a3);

		case QU_CLASS_ID_Class:
		{
			qu_any argv[4] = {self, a1, a2, a3};
			return qu__object_new (how, 4, argv);
		}
	}
	QU_THROW (QuECall);
}


qu_any qu_object_call_ex2 (qu_any self, qu_any how, qu_any a1, qu_any a2)
{
	switch (qu_classid (how))
	{
		case QU_CLASS_ID_String:
			return qu_object_call2 (self, how, a1, a2);

		case QU_CLASS_ID_Sub:
		{
			qu_class c = qu_sub_get_class (how);
			if (c)
			{
				if (qu_sub_is_inst (how))
					QU_REQUIRE (QuECall, qu_object_instof (self, c));
				else if (qu_sub_is_class (how))
					self = c;
				return qu_sub_call2 (how, self, a1, a2);
			}
			else
			{
				return qu_sub_call3 (how, qu_interp_self (), self, a1, a2);
			}
		}

		case QU_CLASS_ID_Method:
			return qu_method_call3 (how, self, a1, a2);

		case QU_CLASS_ID_Class:
		{
			qu_any argv[3] = {self, a1, a2};
			return qu__object_new (how, 3, argv);
		}
	}
	QU_THROW (QuECall);
}



qu_any qu_object_call_ex1 (qu_any self, qu_any how, qu_any a1)
{
	switch (qu_classid (how))
	{
		case QU_CLASS_ID_String:
			return qu_object_call1 (self, how, a1);

		case QU_CLASS_ID_Sub:
		{
			qu_class c = qu_sub_get_class (how);
			if (c)
			{
				if (qu_sub_is_inst (how))
/* FIXME - as long as the method is declared by self's class or any superclass or mixin class */
					QU_REQUIRE (QuECall, qu_object_instof (self, c));
				else if (qu_sub_is_class (how))
					self = c;
				return qu_sub_call1 (how, self, a1);
			}
			else
			{
				return qu_sub_call2 (how, qu_interp_self (), self, a1);
			}
		}

		case QU_CLASS_ID_Method:
			return qu_method_call2 (how, self, a1);

		case QU_CLASS_ID_Class:
		{
			qu_any argv[2] = {self, a1};
			return qu__object_new (how, 2, argv);
		}
	}
	QU_THROW (QuECall);
}



qu_any qu_object_call_ex0 (qu_any self, qu_any how)
{
	switch (qu_classid (how))
	{
		case QU_CLASS_ID_String:
			return qu_object_call0 (self, how);

		case QU_CLASS_ID_Sub:
		{
			qu_class c = qu_sub_get_class (how);
			if (c)
			{
				if (qu_sub_is_inst (how))
					QU_REQUIRE (QuECall, qu_object_instof (self, c));
				else if (qu_sub_is_class (how))
					self = c;
				return qu_sub_call0 (how, self);
			}
			else
			{
				return qu_sub_call1 (how, qu_interp_self (), self);
			}
		}

		case QU_CLASS_ID_Method:
			return qu_method_call1 (how, self);

		case QU_CLASS_ID_Class:
			return qu__object_new (how, 1, &self);
	}
	QU_THROW (QuECall);
}



/*-----------------------------------------------------------------------------
	instance specific

	"instance specific" is an instance attribute not declared in the class
	!!! the class must have the QU_CLASS_FLAG_DYNAMIC flag set
-----------------------------------------------------------------------------*/

static void _object_ensure_specific (qu_object self)
{
	qu_class c = qu_classof (self);

	if (c == QuClassClass)
		QU_FATAL ("class cant have specific attribute");

	if (!qu_class_is_dynamic (c))
		QU_FATAL ("the ClassFlagDYNAMIC flag must be set");
}



/* create instance specific property */
qu_any qu_object_specific_prop (qu_obj self, qu_string name, qu_any value)
{
	_object_ensure_specific (self);
	if (self->attrs == NULL)
	{
		self->attrs = qu_attr_alloc ();
		qu_attr_add (self->attrs, name, value);
	}
	else
	{
		qu_attr_set (self->attrs, name, value);
	}
	return value;
}



/* create instance specific method */
qu_sub qu_object_specific_method (qu_obj self, qu_string name, void *func, int argc, long flags)
{
	qu_sub u;

	_object_ensure_specific (self);
	u = qu_sub_new ((qu_symbol) qu_classof (self), name, func, argc, QU_SUB_FLAG_EXTERN | QU_SUB_FLAG_INST | flags);
	return (qu_sub) qu_object_specific_prop (self, name, u);
}



/* delete instance specific attribute */
qu_null qu_object_specific_delete (qu_obj self, qu_string name)
{
	_object_ensure_specific (self);
	if (self->attrs != NULL)
		qu_attr_del (self->attrs, name);
	return QuNil;
}



/*-----------------------------------------------------------------------------
	miscellaneous
-----------------------------------------------------------------------------*/

qu_any     qu_object_method_self  (qu_any self) { return self;    }
qu_boolean qu_object_method_true  (qu_any self) { return QuTrue;  }
qu_boolean qu_object_method_false (qu_any self) { return QuFalse; }
qu_null    qu_object_method_nil   (qu_any self) { return QuNil;   }
qu_int     qu_object_method_int0  (qu_any self) { return QuInt0;  }



qu_any qu_object_as (qu_any self, qu_any x)
{
	qu_any r;

	if (!qu_class_is (x))
		x = qu_classof (x);

	r = qu_sys_as (self, x);
	QU_ENSURE (r);

	if (qu_object_instof (r, x))
		return r;

	QU_SYS_WARN ("%s.__as(%s) did not return %s",
		qu_classof (self)->name->str, ((qu_class) x)->name->str, ((qu_class) x)->name->str);

	QU_THROW (QuEImplement);
}



/* whether or not the object defines the specified attribute */
int qu_object_defines (qu_any self, qu_string name)
{
	int r;

	__qu_attr_nothrow_begin;
	r = qu_classof (self)->get (self, name, 0) != 0;
	__qu_attr_nothrow_end;

	return r;
}



/* whether an object is iterable */
int qu_object_iterable (qu_any self)
{
	int r;

	/* not a very wise thing to do but we want to avoid a call to c->get */
	switch (qu_classid (self))
	{
		case QU_CLASS_ID_Arg:
		case QU_CLASS_ID_Array:
		case QU_CLASS_ID_Dict:
		case QU_CLASS_ID_File:
		case QU_CLASS_ID_Frame:
		case QU_CLASS_ID_Iterator:
		case QU_CLASS_ID_Queue:
		case QU_CLASS_ID_Range:
		case QU_CLASS_ID_Record:
		case QU_CLASS_ID_Set:
		case QU_CLASS_ID_Stream:
		case QU_CLASS_ID_String:
		case QU_CLASS_ID_Struct:
		case QU_CLASS_ID_Tokenizer:
			return 1;

		case QU_CLASS_ID_Sub:
			if (qu_sub_is_generator (self))
				return 1;
			/*nobreak*/

		case QU_CLASS_ID_Big:
		case QU_CLASS_ID_Boolean:
		case QU_CLASS_ID_Cfunc:
		case QU_CLASS_ID_Class:
		case QU_CLASS_ID_Code:
		case QU_CLASS_ID_Complex:
		case QU_CLASS_ID_Decimal:
		case QU_CLASS_ID_Dlib:
		case QU_CLASS_ID_Exception:
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Function:
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Integer:
		case QU_CLASS_ID_Inf:
		case QU_CLASS_ID_Long:
		case QU_CLASS_ID_Method:
		case QU_CLASS_ID_Module:
		case QU_CLASS_ID_Ast:
		case QU_CLASS_ID_Null:
		case QU_CLASS_ID_Number:
		case QU_CLASS_ID_Object:
		case QU_CLASS_ID_Opaque:
		case QU_CLASS_ID_Proc:
		case QU_CLASS_ID_Pointer:
		case QU_CLASS_ID_Rational:
		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Thread:
		case QU_CLASS_ID_Var:
			return 0;
	}

	__qu_attr_nothrow_begin;
	r = qu_classof (self)->get (self, QU_NAME__each, 0) != 0;
	__qu_attr_nothrow_end;

	return r;
}



/* get object rank */
int qu_object_rank (qu_any self)
{
	qu_class c = qu_classof (self);

	if (c == QuModuleClass   ) return 0;
	if (c == QuClassClass    ) return 1;
	if (c == QuSubClass      ) return 2;
	if (c == QuVarClass      ) return 3;

	/* ------ others must be over 9 */

	if (c == QuThreadClass   ) return 10;
	if (c == QuMutexClass    ) return 11;
	if (c == QuCondClass     ) return 12;

	if (c == QuCodeClass     ) return 20;
	if (c == QuFrameClass    ) return 21;
	if (c == QuAstClass      ) return 22;
	if (c == QuOpaqueClass   ) return 23;

	if (c == QuExceptionClass) return 30;
	if (c == QuNullClass     ) return 31;
	if (c == QuNumberClass   ) return 32;
	if (c == QuIntegerClass  ) return 33;
	if (c == QuComplexClass  ) return 34;
	if (c == QuBooleanClass  ) return 35;
	if (c == QuIntClass      ) return 36;
	if (c == QuLongClass     ) return 37;
	if (c == QuFloatClass    ) return 38;
	if (c == QuStringClass   ) return 39;

	if (c == QuFileClass     ) return 50;

	if (c == QuRangeClass    ) return 60;
	if (c == QuArrayClass    ) return 61;
	if (c == QuDictClass     ) return 62;
	if (c == QuSetClass      ) return 63;
	if (c == QuStructClass   ) return 64;
	if (c == QuRecordClass   ) return 65;

	if (c == QuIteratorClass ) return 70;

	return 99;
}



/*
	(part of) the "is" operator (whether or not self is a x)
	this does not take into account Struct objects
*/
int qu_object_isa (qu_any self, qu_any x)
{
	if (self == x || x == (qu_any) QuObjectClass)
		return 1;

	if (qu_class_is (self))
	{
		if (!qu_class_is (x))
			return 0;

		if (qu_class_is_mixin (x))
			return qu__object__mixes (self, x) == QuTrue;

		return qu_class_isa ((qu_class) self, (qu_class) x);
	}

	if (x == QuClassClass)
		return 0;

	if (qu_class_is (x))
	{
		if (qu_class_is_mixin (x))
			return qu__object__mixes (self, x) == QuTrue;

		return qu_class_isa (qu_classof (self), (qu_class) x);
	}

	/* return qu_class_isa (qu_classof (self), qu_classof (x)); */
	return 0;
}



static char *_object_wherestr_mod (char *buf, qu_module m)
{
	char *e;

	snprintf (buf, PATH_MAX, "%s", m->name->str);
	buf [PATH_MAX] = '\0';
	if (strncmp (buf, QuModuleDir->str, QuModuleDir->len) == 0)
		buf += QuModuleDir->len + 1;
	e = strrchr (buf, QU_PATH_SEPCHR);
	if (e == NULL)
		e = buf;
	else
		e++;
	e = strchr (e, '.');
	if (e)
		*e = '\0';
	return buf;
}



qu_string qu_object_where_string (qu_any sym, const char *prefix, const char *suffix)
{
	char buf [PATH_MAX + 1];

#define _object_where_mod(m) \
	((m == QuSysModule) ? "Sys" : ((m == QuModuleMain) ? "__main__" : _object_wherestr_mod (buf, m)))

	qu_class c = qu_classof (sym);

	if (c == QuModuleClass)
	{
		return qu_string_newf ("%s%s%s", prefix, _object_where_mod (sym), suffix);
	}
	else if (c == QuClassClass)
	{
		return qu_string_newf ("%s%s.%s%s", prefix, _object_where_mod (qu_class_get_module ((qu_class) sym)), ((qu_class) sym)->name->str, suffix);
	}
	else if (c == QuSubClass)
	{
	  issub:
		c = qu_sub_get_class (sym);
		if (c != 0)
		{
			return qu_string_newf ("%s%s.%s.%s%s", prefix, _object_where_mod (qu_sub_get_module (sym)), c->name->str, ((qu_sub) sym)->name->str, suffix);
		}
		else if (qu_sub_is (((qu_sub) sym)->parent))
		{
			return qu_string_newf ("%s%s%s", prefix, ((qu_sub) sym)->name->str, suffix);
		}
		else
		{
			return qu_string_newf ("%s%s.%s%s", prefix, _object_where_mod (qu_sub_get_module (sym)), ((qu_sub) sym)->name->str, suffix);
		}
	}
	else if (c == QuMethodClass)
	{
		sym = ((qu_method) sym)->sub;
		goto issub;
	}
	else if (c == QuVarClass)
	{
		c = qu_var_get_class (sym);
		if (c != 0)
		{
			return qu_string_newf ("%s%s.%s.%s%s", prefix, _object_where_mod (qu_var_get_module (sym)), c->name->str, ((qu_var) sym)->name->str, suffix);
		}
		else if (qu_var_is_lexical (sym))
		{
			return qu_string_newf ("%s%s%s", prefix, ((qu_var) sym)->name->str, suffix);
		}
		else
		{
			return qu_string_newf ("%s%s.%s%s", prefix, _object_where_mod (qu_var_get_module (sym)), ((qu_var) sym)->name->str, suffix);
		}
	}
	else
	{
		/* should never happen */
		return QuStringEMPTY;
	}
}



/*-----------------------------------------------------------------------------
	validators ("type")
-----------------------------------------------------------------------------*/

/*
	compile-time only
	whether x is a valid validator
*/
static int _object_valid_type (qu_any x, int level)
{
	if (level > 8)
		return 0;

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Class:
			return (qu_class) x != QuNullClass && !qu_class_is_mixin (x);

		case QU_CLASS_ID_Struct:
		case QU_CLASS_ID_Range:
			return 1;

		case QU_CLASS_ID_Array:
			return qu_array_frozen (x);

		case QU_CLASS_ID_Set:
			return qu_set_frozen ((qu_set) x);

		case QU_CLASS_ID_Dict:
			return qu_dict_frozen ((qu_dict) x);

		case QU_CLASS_ID_Sub:
			if (qu_sub_is_generator (x) || qu_sub_is_once (x) || qu_sub_is_abstract (x))
				return 0;
			/* it typed, don't allow qu_null */
			if (((qu_sub) x)->type == (qu_any) QuNullClass)
				return 0;
			if (qu_sub_is_inst (x))
				return ((qu_sub) x)->argc == 0 || ((qu_sub) x)->argc == -1;
			return ((qu_sub) x)->argc == 1 || ((qu_sub) x)->argc == -1;

		/* QU_WARN_EXPERIMENTAL - allow for dynamic setting --> qu_var.type */
		case QU_CLASS_ID_Method:
			return 1;

		case QU_CLASS_ID_Var:
			if (qu_flag_test (x, QU_VAR_FLAG_ENUM | QU_VAR_FLAG_STRUCT | QU_VAR_FLAG_TYPEOF))
				return 1;
			if (qu_var_frozen (x))
				return _object_valid_type (((qu_var) x)->value, level + 1);
			/*break;*/
			return 1;
	}
	return 0;
}



int qu_object_valid_type (qu_any x)
{
	return _object_valid_type (x, 0);
}



/* compile-time only */
qu_any qu_object_fix_type (qu_any type)
{
	if (qu_var_is (type))
	{
		/* so we don't need to keep the qu_var object at run-time */

		if (qu_var_is_enum (type) || qu_var_is_struct (type))
		{
			return ((qu_var) type)->value;
		}
		else if (qu_var_is_typeof (type))
		{
			/*
				qu_var_is_typeof comes from

				var foo: qu_array of xxx
				var foo: qu_set   of xxx
				var foo: qu_dict  of xxx
			*/
			qu_var subtype = ((qu_var) type)->type;
			if (qu_var_is (subtype))
			{
				if (qu_var_is_enum (subtype) || qu_var_is_struct (subtype))
					((qu_var) type)->type = subtype->value;
			}
		}
	}
	return type;
}



static int _object_check_value (qu_any type, qu_any val, int level)
{
	if (level > 8)
		return 0;

	switch (qu_classid (type))
	{
		case QU_CLASS_ID_Class:
			return qu_object_instof (val, type);

		case QU_CLASS_ID_Struct:
			return qu_record_of (val, type);

		case QU_CLASS_ID_Range:
			return qu_range_contains (type, val);

		case QU_CLASS_ID_Array:
			return qu_array_contains (type, val);

		case QU_CLASS_ID_Set:
			return qu_set_contains (type, val);

		case QU_CLASS_ID_Dict:
			return qu_dict_getitem (type, val) != 0;

		case QU_CLASS_ID_Sub:
#if 0
			/* this is wrong - the value may be validated into a different type */
			if (type == val)
				return 1;
			else if (((qu_sub) type)->type)
				return _object_check_value ((((qu_sub) type)->type), val, level + 1);
#endif
			/* do it at runtime */
			return 1;

		/* QU_WARN_EXPERIMENTAL - allow for dynamic setting --> qu_var.type */
		case QU_CLASS_ID_Method:
			/* FIXME - should check for method->sub->type */
			return 1;

		case QU_CLASS_ID_Var:
			if (qu_flag_test (type, QU_VAR_FLAG_ENUM | QU_VAR_FLAG_STRUCT | QU_VAR_FLAG_TYPEOF))
				return 1;
			if (qu_var_frozen (type))
				return _object_check_value (((qu_var) type)->value, val, level + 1);
			/* break; */
			return 1;
	}
	return 0;
}



/*
	compile-time only
	whether or not 'val' is a valid value of validator 'type'
*/
int qu_object_check_value (qu_any type, qu_any val)
{
	return _object_check_value (type, val, 0);
}



/*
	run-time only
	check whether 'val' is a valid value for 'self' based on self's 'type'
	!!! cant be used with properties -- use setattr for that
*/
qu_any qu_object_validate (qu_any self, qu_any type, qu_any val)
{
	switch (qu_classid (type))
	{
		case QU_CLASS_ID_Class:
			if (qu_object_instof (val, type))
				return val;
			break;

		case QU_CLASS_ID_Struct:
			/* CHECKME - may be nil */
			if (val == QuNil || qu_record_of (val, type))
				return val;
			break;

		case QU_CLASS_ID_Range:
			if (qu_range_contains (type, val))
				return val;
			break;

		case QU_CLASS_ID_Array:
			if (qu_array_contains (type, val))
				return val;
			break;

		case QU_CLASS_ID_Set:
			if (qu_set_contains (type, val))
				return val;
			break;

		case QU_CLASS_ID_Dict:
			if (qu_dict_getitem (type, val))
				return val;
			break;

		case QU_CLASS_ID_Sub:
		{
			qu_any   x;
			qu_class c = (qu_class) ((qu_sub) type)->parent;

			if (c == 0 || !qu_class_is (c))
			{
				x = qu_sub_call1 (type, self, val);
			}
			else
			{
				if (qu_sub_is_inst (type))
				{
					if (qu_sub_is_ensure (type))
					{
						x = qu_sub_call1 (type, self, val);
					}
					else
					{
						if (!qu_object_instof (val, c))
							break;

						x = qu_sub_call0 (type, val);
						if (x == QuTrue)
						{
							return val;
						}
						else if (x == QuFalse)
						{
							break;
						}
						else
						{
							if (x && qu_object_instof (x, c))
								return x;
						}
					}
				}
				else if (qu_sub_is_class (type))
				{
					x = qu_sub_call1 (type, (qu_class) ((qu_sub) type)->parent, val);
				}
				else
				{
					x = qu_sub_call1 (type, self, val);
				}
			}

			if (x == QuTrue)
			{
				return val;
			}
			else if (x != QuFalse)
			{
				return x;
			}
			break;
		}

		case QU_CLASS_ID_Var:
			if (qu_var_is_typeof (type))
			{
				qu_var   v = type;
				qu_class c = v->value;

				assert (v->type);
				if (c == QuArrayClass)
				{
					if (qu_array_is (val) && qu_array_of (val, v->type))
						return val;
				}
				else if (c == QuSetClass)
				{
					if (qu_set_is (val) && qu_set_of (val, v->type))
						return val;
				}
				else if (c == QuDictClass)
				{
					if (qu_dict_is (val) && qu_dict_of (val, v->type))
						return val;
				}
			}
			else
			{
				return qu_object_validate (self, ((qu_var) type)->value, val);
			}
			break;

		/* QU_WARN_EXPERIMENTAL - allow for dynamic setting --> qu_var.type */
		case QU_CLASS_ID_Method:
		{
			qu_any x = qu_method_call1 (type, val);
			if (x)
				return x;
			break;
		}
	}
	QU_THROW (QuEValue);
}



static int _object_compat_type (qu_any type, qu_any othertype, int level)
{
	if (type == othertype)
		return 1;

	if (level > 8)
		return 0;

	switch (qu_classid (type))
	{
		case QU_CLASS_ID_Class:
			return qu_class_is (othertype) && qu_class_isa (type, othertype);

		case QU_CLASS_ID_Struct:
			return qu_struct_is (othertype)
				&& qu_struct_like (type, othertype);

		case QU_CLASS_ID_Dict:
			return qu_dict_is (othertype) && qu_dict_frozen (othertype)
				&& qu_dict_like (type, othertype);

		case QU_CLASS_ID_Array:
			return qu_array_is (othertype) && qu_array_frozen (othertype)
				&& qu_array_like (type, othertype);

		case QU_CLASS_ID_Set:
			return qu_set_is (othertype) && qu_set_frozen (othertype)
				&& qu_set_like (type, othertype);

		case QU_CLASS_ID_Range:
			return qu_range_is (othertype) && qu_range_eq (type, othertype);

		case QU_CLASS_ID_Sub:
			if (((qu_sub) type)->type)
				return _object_compat_type (type, othertype, level + 1);
			break;

		case QU_CLASS_ID_Var:
			if (qu_var_is (othertype)
				&& qu_flag_test (othertype, QU_VAR_FLAG_ENUM | QU_VAR_FLAG_STRUCT | QU_VAR_FLAG_TYPEOF)
				&& ((qu_var) type)->value == ((qu_var) othertype)->value
				&& ((qu_var) type)->type
				&& ((qu_var) othertype)->type)
				return _object_compat_type (((qu_var) type)->type, ((qu_var) othertype)->type, level + 1);
			break;
	}
	return 0;
}



/*
	run-time only
	whether 'othertype' is a compatible type for 'type'
*/
int qu_object_compat_type (qu_any type, qu_any othertype)
{
	return _object_compat_type (type, othertype, 0);

}




/*----------------------------------------------------------------------------
	Symbol = Module, Class, Sub, Var
----------------------------------------------------------------------------*/

qu_module qu_symbol_get_module (qu_symbol self)
{
	switch (qu_classid (self))
	{
		case QU_CLASS_ID_Class:
			return qu_class_get_module ((qu_class) self);

		case QU_CLASS_ID_Sub:
			return qu_sub_get_module ((qu_sub) self);

		case QU_CLASS_ID_Var:
			return qu_var_get_module ((qu_var) self);
	}
	return 0;
}


qu_class qu_symbol_get_class (qu_symbol self)
{
	switch (qu_classid (self))
	{
		case QU_CLASS_ID_Sub:
			return qu_sub_get_class ((qu_sub) self);

		case QU_CLASS_ID_Var:
			return qu_var_get_class ((qu_var) self);
	}
	return 0;
}



/*
	Find symbol in block, then parent blocks, then the builtin Sys module.
*/
qu_symbol qu_symbol_find (qu_symbol from, qu_string name)
{
	qu_symbol x;

	if (from->attrs)
	{
		x = (qu_symbol) qu_attr_find (from->attrs, name);
		if (x)
			goto found;
	}

	if (qu_module_is (from))
	{
		x = qu_attr_find (QuSysModule->attrs, name);
	}
	else if (qu_class_is (from))
	{
		x = qu_symbol_find ((qu_symbol) ((qu_class) from)->parent, name);
	}
	else
	{
		x = qu_symbol_find (((qu_sub) from)->parent, name);
	}

	if (x == 0)
		return 0;

  found:
	assert (qu_symbol_is (x));
	return x;
}



/*-----------------------------------------------------------------------------
	get/set instance attributes - runtime
-----------------------------------------------------------------------------*/

#define qu__object__get qu_attr_inst_get
#define qu__object__set qu_attr_inst_set


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/
qu_sub __QuObjectNew;

void qu_object__setup ()
{
	QuObjectClass->flags |= QU_CLASS_FLAG_EXTERN | QU_CLASS_FLAG_FINAL | QU_CLASS_FLAG_AUTO;
	QuObjectClass->get    = (qu_attrget_f) qu__object__get;
	QuObjectClass->set    = (qu_attrset_f) qu__object__set;

#define m_(name, cfunc, argc, flags) \
	qu_attr_add (QuObjectClass->attrs, name, qu_sub_new ((qu_symbol) QuObjectClass, name, \
		(void *) cfunc, argc, QU_SUB_FLAG_EXTERN | flags))

#define z_(namez, cfunc, argc, flags) \
QU_BEGIN \
	qu_string name = qu_string_new_staticz (namez); \
	qu_attr_add (QuObjectClass->attrs, name, qu_sub_new ((qu_symbol) QuObjectClass, name, \
		(void *) cfunc, argc, QU_SUB_FLAG_EXTERN | flags)); \
QU_END

	/* standard class methods */
	m_(QU_NAME__inst, qu__object__inst,  0, QU_SUB_FLAG_CLASS | QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_FINAL);
	m_(QU_NAME_new  , qu__object_new  , -1, QU_SUB_FLAG_CLASS);

	/* standard static methods */
	m_(QU_NAME__attrs     , qu__object__attrs     , -1, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__block     , qu__object__block     ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__class     , qu__object__class     ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__defines   , qu__object__defines   , -1, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__decl      , qu__object__decl      ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__doc       , qu__object__doc       ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__ident     , qu__object__ident     ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__isa       , qu__object__isa       ,  1, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__isabstract, qu__object__isabstract,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__isauto    , qu__object__isauto    ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__isextern  , qu__object__isextern  ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__isfinal   , qu__object__isfinal   ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__ismixin   , qu__object__ismixin   ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__isprivate , qu__object__isprivate ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__issuper   , qu__object__issuper   ,  1, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__iterable  , qu__object__iterable  ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__master    , qu__object__master    ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__mixins    , qu__object__mixins    ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__mixes     , qu__object__mixes     ,  1, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__module    , qu__object__module    ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__name      , qu__object__name      ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__rank      , qu__object__rank      ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__root      , qu__object__root      ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__static    , qu__object__static    ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__storable  , qu__object__storable  ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__super     , qu__object__super     ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__tree      , qu__object__tree      ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
	m_(QU_NAME__where     , qu__object__where     ,  0, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);

	/* standard instance methods */
	m_(QU_NAME__as        , qu__object__as        ,  1, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__cmp       , qu__object__cmp       ,  1, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__eq        , qu__object__eq        ,  1, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__hash      , qu__object__hash      ,  0, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__init      , qu__object__init      , -1, QU_SUB_FLAG_INST);
	m_(QU_NAME__like      , qu__object__like      ,  1, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__print     , qu__object__print     ,  0, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__string    , qu__object__string    ,  0, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__true      , qu__object__true      ,  0, QU_SUB_FLAG_STATIC);
	m_(QU_NAME__version   , qu__object__version   ,  0, QU_SUB_FLAG_STATIC);

	/* specific instance methods */

	z_("cmp" , qu__object_cmp , 1, QU_SUB_FLAG_INST);
	z_("eq"  , qu__object_eq  , 1, QU_SUB_FLAG_INST);
	z_("like", qu__object_like, 1, QU_SUB_FLAG_INST);
	z_("le"  , qu__object_le  , 1, QU_SUB_FLAG_INST);
	z_("lt"  , qu__object_lt  , 1, QU_SUB_FLAG_INST);
	z_("ge"  , qu__object_ge  , 1, QU_SUB_FLAG_INST);
	z_("gt"  , qu__object_gt  , 1, QU_SUB_FLAG_INST);

	/* to speed things up a bit */
	__QuObjectNew = qu_attr_find (QuObjectClass->attrs, QU_NAME_new);
	assert (__QuObjectNew);
}
