/* Pointer.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Pointer
	This class provides a way to wrap C pointers, allocate memory, and
	set/get values in convenient ways.

	@warn
	This is a dangerous class that requires some C knowledge.
	Use it wrong and your program will crash.

	@p
	The main purpose of this class is to pass data from/to a Qu program
	to/from a C program. This is handy when you work with -{Proc} or
	have to call functions in a shared library.

	You can get and set values with the #{get} and #{set} methods.
	See #{spec} for convenient ways to get/set values.

	Like number objects, you can not modify the pointer address inside
	a Pointer, but you can do simple arithmetics with them to produce another Pointer.

	To test whether or not two Pointer objects point to the same memory address
	you use the C{===} operator (the #{__like} method). The C{==} operator
	(the __eq method) will compare the address of the Pointer objects instead
	of the pointers they hold.

	@xmp
		p1 = Pointer (50)      # allocate 50 bytes
		p2 = p1 + 5
		p3 = p1 + 0
		assert (p3 === p1)
		assert (p3 != p1)
		println (p1)
		println (p2)
		println (p3)
		p1.free

	@warn
	Be careful if the content contains a Qu object since it will not be referenced.
	You must keep the object alive as long as the Pointer is used.
*/
qu_class QuPointerClass;


/*) var class NULL
	This is a class constant. It is a Pointer object with a NULL pointer.
	Use it whenever you need a NULL pointer.
*/
qu_pointer QuPointerNULL;


/* flags */
#define qu_pointer_set_allocated(self)   qu_flag_set   (self, QU_POINTER_FLAG_ALLOCATED)
#define qu_pointer_unset_allocated(self) qu_flag_unset (self, QU_POINTER_FLAG_ALLOCATED)
#define qu_pointer_set_malloc(self)      qu_flag_set   (self, QU_POINTER_FLAG_MALLOC)
#define qu_pointer_unset_malloc(self)    qu_flag_unset (self, QU_POINTER_FLAG_MALLOC)


/* stupid forwards */
static qu_any qu__pointer_spec (qu_pointer self, int argc, qu_any *argv);


#define ENSURE_NOT_FROZEN(self) QU_REQUIREZ (QuEModify, !qu_pointer_frozen (self), "Pointer is frozen")


/* stupid forwards */
static qu_any qu__pointer__setitem (qu_pointer self, qu_any x, qu_any v);


/*----------------------------------------------------------------------------
	stuffs
----------------------------------------------------------------------------*/

/* get the type and offset of x */
static char qu_pointer__ofs (qu_any spec, qu_any x, long *ofs)
{
	long      i;
	qu_string v;
	char      type;

	switch (qu_classid (spec))
	{
		case QU_CLASS_ID_Struct:
			i = qu_struct_index_ex (spec, x);
			QU_REQUIRE (QuEIndex, i >= 0);
			*ofs = qu_struct_offset_ex (spec, i, &type);
			QU_REQUIRE (QuENone, *ofs >= 0);
			return type;

		case QU_CLASS_ID_String:
			QU_ARG_long (x, i);
			QU_REQUIRE (QuEIndex, 0 <= i && i < ((qu_string) spec)->len);
			*ofs = qu_struct_offset_ex (spec, i, &type);
			QU_REQUIRE (QuENone, *ofs >= 0);
			return type;

		case QU_CLASS_ID_Dict:
			v = qu_dict_getitem (spec, x);
			QU_REQUIRE (QuEIndex, v);
			QU_REQUIRE (QuENone, qu_string_is (v) && v->len >= 2);
			*ofs = atol (v->str + 1);
			return v->str[0];
	}
	return 0;
}



/*
	If the class specifies a Struct named as the class then return that Struct.
*/
static qu_struct qu__pointer_getstruct (qu_class self, int throws)
{
	qu_var    v;
	qu_struct u = 0;

	v = qu_attr_find (self->attrs, self->name);
	if (v && qu_var_is (v))
		u = v->value;
	if (u && qu_struct_is (u))
		return u;
	if (throws)
		QU_ARG_THROWF ("class %s does not define struct", self->name->str);
	return NULL;
}



/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

qu_pointer qu_pointer_new_static (void *m)
{
	qu_pointer p = qu_gc_new_static (QuPointerClass);
	p->ptr   = m;
	p->delta = 1;
	p->size  = 0;
	p->spec  = 0;
	return p;
}



qu_pointer qu_pointer_predefined (long size, char spec)
{
	qu_pointer p = qu_gc_new_static (QuPointerClass);
	p->ptr   = qu_mem_calloc (1, size);
	p->delta = size;
	p->size  = size;
	p->spec  = qu_string_byte (spec);
	return p;
}



qu_pointer qu_pointer_new (void *m)
{
	if (m == NULL)
	{
		return QuPointerNULL;
	}
	else
	{
		qu_pointer p = qu_gc_new (QuPointerClass);
		p->ptr   = m;
		p->delta = 1;
		p->size  = 0;
		p->spec  = 0;
		return p;
	}
}



qu_pointer qu_pointer_new_allocated (void *m, long size)
{
	qu_pointer p = qu_pointer_new (m);

	assert (m);
	p->size = size;
	qu_pointer_set_allocated (p);
	return p;
}



qu_pointer qu_pointer_new_alloc (long size)
{
	return qu_pointer_new_allocated (qu_mem_alloc (size), size);
}



qu_pointer qu_pointer_new_calloc (long size)
{
	qu_pointer p = qu_pointer_new_alloc (size);

	QU_ENSURE (p);
	memset (p->ptr, 0, size);
	return p;
}



static qu_pointer qu_pointer_new_integer (qu_integer size)
{
	long z;

	QU_ARG_long (size, z);
	QU_ARG_CHECK (z >= 0);
	QU_MEM_CHECK (z <= QU_MEM_MAX);
	return qu_pointer_new_calloc (z);
}



static qu_pointer qu_pointer_new_struct (qu_struct u)
{
	qu_pointer p;
	long       z = qu_struct_size (u);

	assert (0 < z && z <= QU_MEM_MAX);
	p = qu_pointer_new_calloc (z);
	p->spec = u;
	p->delta = z;
	return p;
}



static qu_pointer qu_pointer_new_string (qu_string s, int copy)
{
	qu_pointer p;

	if (copy)
	{
		long l = s->len + 1;
		p = qu_pointer_new_alloc (l);
		memcpy (p->ptr, s->str, l);
	}
	else
	{
		p = qu_pointer_new (s->str);
	}
	p->delta = 1;
	return p;
}



qu_pointer qu_pointer_new_record (qu_record r)
{
	qu_struct  u = r->struc;
	long       z = qu_struct_size (u);
	qu_pointer p = qu_pointer_new_alloc (z);

	p->spec = u;
	p->delta = z;
	QU_ENSURE (qu_pointer_set_record (p->ptr, r));
	return p;
}



static qu_pointer qu_pointer_new_pointer (qu_class self, qu_pointer p, long ofs)
{
	qu_pointer r;
	qu_struct  u;
	long       z;

	assert (p->ptr);

	u = qu__pointer_getstruct (self, 1);
	QU_ENSURE (u);

	z = qu_struct_size_ex (u);

	r        = qu_pointer_new_calloc (z);
	r->spec  = u;
	r->delta = qu_struct_size (u);
	r->klass = self;

	memcpy (r->ptr, p->ptr + ofs, z);
	return r;
}



static qu_pointer qu_pointer_new_flat_spec (qu_stringy s, long argc, qu_any *argv)
{
	char              y = s->str[0];
	qu_struct_type_s *t = qu_struct_type_info (y);
	long              z = t->size;
	qu_pointer        p;
	void             *m;

	QU_ARG_VMIN (1);
	QU_ARG_CHECK (t->type);
	QU_MEM_CHECK (z < QU_MEM_MAX / argc / 2);
	p = qu_pointer_new_alloc (z * (argc + 1));
	p->delta = z;
	QU_ENSURE (p);
	m = p->ptr;
	while (argc--)
	{
		QU_ENSURE (qu_pointer_set (m, y, *argv++));
		m += z;
	}
	memset (m, 0, z);
	return p;
}



static qu_pointer qu_pointer_new_flat_record (long argc, qu_any *argv)
{
	qu_record   r;
	qu_struct   u;
	long        z;
	qu_pointer  p;
	void       *m;

	QU_ARG_VMIN (1);
	r = argv[0];
	QU_ARG_TYPE (r, qu_record);
	u = r->struc;
	z = qu_struct_size (u);
	QU_MEM_CHECK (z < QU_MEM_MAX / argc / 2);
	p = qu_pointer_new_alloc (z * (argc + 1));
	p->spec = u;
	p->delta = z;
	m = p->ptr;
	while (argc--)
	{
		r = *argv++;
		QU_ARG_CHECK (qu_record_is (r) && r->struc == u);
		qu_pointer_set_record (m, r);
		m += z;
	}
	memset (m, 0, z);
	return p;
}



static qu_pointer qu_pointer_new_flat_pointer (qu_class self, int argc, qu_any *argv)
{
	qu_any      spec;
	long        z;
	qu_pointer  r, p;
	void       *m;
	qu_class    c;

	QU_ARG_CHECKZ (argc, "cant flat an empty Array");
	p = argv[0];
	QU_ARG_TYPE (p, qu_pointer);
	spec = p->spec;
	if (spec == 0)
	{
		spec = qu__pointer_getstruct (self, 1);
		QU_ENSURE (spec);
	}
	z = qu_struct_size_ex (spec);
	QU_MEM_CHECK (z < QU_MEM_MAX / argc / 2);
	r = qu_pointer_new_alloc (z * (argc + 1));
	m = r->ptr;
	while (argc--)
	{
		p = *argv++;
		c = qu_classof (p);
		QU_ARG_CHECK (qu_class_isa (c, self));
		memcpy (m, p->ptr, z);
		m += z;
	}
	memset (m, 0, z);
	return r;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (*arg)
	@def
	sub class __inst (size: Int) -> Pointer
	sub class __inst (u: Struct) -> Pointer
	sub class __inst (r: Record) -> Pointer
	sub class __inst (p: Pointer, ofs = 0) -> Pointer
	sub class __inst (s: String, copy = false) -> Pointer
	@p
	Create a new Pointer.

	@def
	sub class __inst (size: Int) -> Pointer
	@p
	Allocate a specified amount of memory.
	v{size} must be a positive value and not bigger than some hard limit
	(which is a huge value and should never be specified in the first place).
	The #{delta} value is set to 1.

	@def
	sub class __inst (u: Struct) -> Pointer
	@p
	Bind a -{Struct} with a -{Pointer} and allocate
	the necessary zero initialized space. See the #{spec} method for more.
	The #{delta} value is set to the size of the Struct.

	@def
	sub class __inst (r: Record) -> Pointer
	@p
	Create a pointer which contains a -{Struct} which value are set with the
	value of the -{Record}.
	The #{delta} value is set to the size of the Struct.

	@def
	sub class __inst (s: String) -> Pointer
	@p
	Create a pointer which contains a pointer to the content of the String.
	Make sure the String lives as long as the Pointer.
	The #{delta} value is set 1.

	@def
	sub class __inst (p: Pointer, ofs = 0) -> Pointer
	@p
	Create a pointer from a pointer to a C struct which optionally starts at
	a specified v{offset}.

	@xmp
		struct Foo
		    a : _int_
		    b : _float_
		;;

		p = Pointer (Foo)
		p->a = 1
		p->b = 2.3

	@exc
	.ESecurity
	.EArgValue
	.EMemory
*/
static qu_any qu__pointer__inst (qu_class self, int argc, qu_any *argv)
{
	qu_any     x;
	qu_pointer p = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];

	if (qu_integer_is (x))
	{
		p = qu_pointer_new_integer (x);
	}
	else if (qu_struct_is (x))
	{
		p = qu_pointer_new_struct (x);
	}
	else if (qu_string_is (x))
	{
		int copy = 0;

		if (argc > 1)
			QU_ARG_vbool (1, copy);

		p = qu_pointer_new_string (x, copy);
	}
	else if (qu_record_is (x))
	{
		p = qu_pointer_new_record (x);
	}
	else if (qu_pointer_is (x))
	{
		long ofs = 0;

		if (argc > 1)
			QU_ARG_vlong (1, ofs);

		if (((qu_pointer) x)->ptr == 0)
			return QuPointerNULL;

		p = qu_pointer_new_pointer (self, x, ofs);
	}
	else
	{
		QU_THROW (QuEArgValue);
	}
	p->klass = self;
	return p;
}



/*) sub class new (*arg) -> Pointer
	@def
	sub class new (x: Null, *arg) -> Pointer
	sub class new (x: Pointer, ofs = 0) -> Pointer
	sub class new (*arg) -> Pointer
	@p
	If v{x} is nil or NULL then return #{NULL} and ignore any other arguments.

	Otherwise if this class is the Pointer class then call #{__inst}.

	Otherwise this method finds a struct definition in this class which name
	is the same as the class, then set #{spec} to that struct and initialize the
	fields with v{arg}.

	If the first argument is a Pointer then #{acquire} that pointer.
	The other arguments are ignored.

	Otherwise EArgValue is thrown.

	@exc
	.EArgValue

	@xmp
		class Dot is Pointer __final
			struct dot
				x: _int_
				y: _int_
			;;
		;;

		p = Dot (10, 20)
		println (p)

		p = Dot ()
		println (p)

	@out
		<Dot 10 20>
		<Dot 0 0>
*/
static qu_pointer qu__pointer_new (qu_class self, int argc, qu_any *argv)
{
	qu_pointer p;
	qu_any     x = argc ? argv[0] : 0;

	if (x == (qu_any) QuNil || x == (qu_any) QuPointerNULL)
	{
		return QuPointerNULL;
	}
	else if (self == QuPointerClass)
	{
		return qu__pointer__inst (self, argc, argv);
	}
	else
	{
		qu_struct u = qu__pointer_getstruct (self, 1);
		long      z;

		QU_ENSURE (u);
		z = qu_struct_size (u);

		if (x && qu_pointer_is (x))
		{
 			p = (qu_pointer) x;
			ENSURE_NOT_FROZEN (p);
		}
		else
		{
			long n = qu_struct_len (u);
			long i;

			QU_REQUIRE (QuEArgCount, n >= (long) argc);
			p = qu_pointer_new_calloc (z);
			p->spec = u;
			for (i = 0; i < argc; ++i)
				QU_ENSURE (qu__pointer__setitem (p, qu_int_new (i), *argv++));
		}
		p->spec  = u;
		p->size  = p->delta = z;
		p->klass = self;
		return p;
	}
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class structof -> Struct
	If the class specifies a -{Struct} which name is the name of the class then
	return that Struct, otherwise throw EImplement.

	@exc
	.EImplement

	@xmp
		class Foo is Pointer
			struct Foo
				a: Cint
				b: Cfloat
			;;
		;;

		println (Foo.structof)

	@out
		[a: Cint, b: Cfloat]
*/
static qu_struct qu__pointer_structof (qu_class self)
{
	qu_struct u = qu__pointer_getstruct (self, 0);
	QU_REQUIREF (QuEImplement, u, "class %s does not define a struct", self->name->str);
	return u;
}



/*) sub class sizeof -> Struct
	If the class specifies a -{Struct} which name is the name of the class then
	return the size of that Struct, otherwise throw EImplement.

	@exc
	.EImplement

	@xmp
		class Foo is Pointer
			struct Foo
				a: Cint
				b: Cdouble
			;;
		;;

		println (Foo.sizeof)

	@out
		12
*/
static qu_int qu__pointer_sizeof (qu_class self)
{
	qu_struct u = qu__pointer_getstruct (self, 0);
	long      z;

	QU_REQUIREF (QuEImplement, u, "class %s does not define a struct", self->name->str);
	z = qu_struct_size (u);
	return qu_int_new (z);
}



/*) sub class acquire (p: Pointer) -> Pointer
	@p
	If v{p} is a NULL pointer then return nil otherwise set its class ot this class
	then return it.

	If the class specifies a -{Struct} which name is the name of the class then
	the #{spec}, #{size} and #{delta} fields are set.

	@exc
	.EArgValue
	.EModify

	@xmp
		class Point
			struct point
				x: Cint
				y: Cint
			;;
		;;

		p = call_c_function_which_returns_a_pointer
		Point.acquire (p)
*/
static qu_pointer qu__pointer_acquire (qu_class self, qu_pointer p)
{
	QU_ARG_TYPE (p, qu_pointer);
	ENSURE_NOT_FROZEN (p);
	if (p->ptr)
	{
		qu_struct u = qu__pointer_getstruct (self, 0);
		p->klass = self;
		if (u)
		{
			p->size = p->delta = qu_struct_size (u);
			p->spec = u;
		}
	}
	return p;
}



/*) sub class flat (*x) -> Pointer
	@def
	sub class flat (p1: Pointer, p2: Pointer, ...) -> Pointer
	sub class flat (a: Array of Pointer) -> Pointer
	sub class flat (a: Array of Record) -> Pointer
	sub class flat (a: Array, spec: String) -> Pointer
	@p
	Create a C pointer from an Array of Qu Pointers.

	In the first and second form, all Pointers must be of this class and may not be NULL.
	In the third and fourth form, all items must of the same Record or specification.

	v{spec} must be one of the characters specified by -{Struct.spec}.

	Note that a NULL record entry is appended at the end of the Array.

	@exc
	.EArgValue

	@xmp
		class Foo is Pointer
			struct foo
				x: Cint
				y: Cint
			;;
		;;

		f1 = Foo (10, 20)
		f2 = Foo (30, 40)

		p = Foo.flat (f1, f2)    # pointer to memory containing 10 20 30 40
		p = Foo.flat ([f1, f2])  # same thing
*/
static qu_pointer qu__pointer_flat (qu_class self, int argc, qu_any *argv)
{
	qu_any     x;
	qu_pointer p = 0;

	QU_ARG_VMIN (1);
	x = argv[0];

	if (qu_pointer_is (x))
	{
		p = qu_pointer_new_flat_pointer (self, argc, argv);
	}
	else if (qu_array_is (x))
	{
		qu_array a = x;

		QU_ARG_CHECKZ (a->len, "array can not be empty");
		if (argc > 1)
		{
			qu_stringy s;

			QU_ARG_VTYPE (1, s, qu_stringy);
			p = qu_pointer_new_flat_spec (s, a->len, a->objs);
		}
		else
		{
			x = a->objs[0];
			if (qu_record_is (x))
			{
				p = qu_pointer_new_flat_record (a->len, a->objs);
			}
			else
			{
				p = qu_pointer_new_flat_pointer (self, a->len, a->objs);
			}
		}
	}
	else
	{
		QU_THROW (QuEArgValue);
	}
	return p;
}



/*) sub class array (p: Pointer, n: Integer) -> Array
	Create a new Array of Qu Pointer objects from a flat C pointer.
	Thus the reverse of #{flat}.

	The class must define a -{Struct} named as the class and the items can be
	directly copied into the struct.

	@exc
	.EArgValue

	@xmp
		# C code
			typedef struct Point Point;
			struct Point
			{
				int x;
				int y;
			};

			static Point ThreePoints [3] = {{1, 2}, {3, 4}, {5, 6}};

			static void *foo ()
			{
				return ThreePoints;
			}

		# Qu code

			class Point is Pointer
				struct Point
					x : Cint
					y : Cint
				;;
			;;

			p = somehow_call_foo
			a = Point.array (p, 3)
*/
static qu_array qu__pointer_array (qu_class self, qu_pointer p, qu_integer l)
{
	long       n, z;
	qu_array   a;
	void      *m;
	qu_struct  u;

	QU_ARG_TYPE (p, qu_pointer);
	QU_ARG_long (l, n);
	QU_ARG_CHECK (n >= 0);

	if (n == 0)
		return qu_array_new (0);

	QU_MEM_CHECK (n <= QU_GC_MAX_ITEMS);

	u = qu__pointer_getstruct (self, 1);
	QU_ENSURE (u);

	z = qu_struct_size (u);
	a = qu_array_new (n);
	m = p->ptr;
	while (n--)
	{
		p = qu_pointer_new_alloc (z);
		p->klass = self;
		p->spec = u;
		p->delta = p->size = z;
		memcpy (p->ptr, m, z);
		m += z;
		qu_array_push_quick (a, p);
	}
	return a;
}



/*) sub class addr (x, spec = nil) -> Pointer
	Return a Pointer object with the pointer points to v{x}.

	You can optionally set a specification.
	See #{spec}.

	@exc
	.ESecurity
	.EArgValue    the object is an Int

	@xmp
		use Qu

		v = 4.567
		printfln ('%p', v)
		p = Pointer.addr (v)
		println (p)
		r = p.get ('f', Qu.OBJ_HEADER_SIZE)
		println (r)

	@out
		0x400c8098
		<Pointer 0x400c8098>
		4.567
*/
static qu_pointer qu__pointer_addr (qu_class self, int argc, qu_any *argv)
{
	qu_pointer p;
	qu_any     x, spec = QuNil;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (argc > 1)
		spec = argv[1];
	QU_ARG_CHECK (!qu_int_is (x));
	p = qu_pointer_new (x);
	p->klass = self;
	if (spec != QuNil)
		QU_ENSURE (qu__pointer_spec (p, 1, &spec));
	return p;
}



/*) sub class addrof (x) -> Integer
	Return the memory address of v{x} as an Integer value.

	@exc
	.ESecurity
	.EArgValue     the object is an Int

	@xmp
		s = 'foo'
		p = Pointer.addr (s)
		println (p.value)
		println (Pointer.addrof (s))

	@out
		1074651320
		1074651320
*/
static qu_integer qu__pointer_addrof (qu_class self, qu_any x)
{
	__qu_unsafe__;

	QU_ARG_CHECK (!qu_int_is (x));
	return qu_integer_newl ((long) x);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

void qu__pointer__destroy (qu_pointer self)
{
	if (qu_pointer_allocated (self))
	{
		assert (self->ptr);
		qu_mem_free (self->ptr);
	}
}



void qu__pointer__mark (qu_pointer self)
{
	qu_gc_xmark (self->spec);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x) -> Any
	Convert the underlying pointer to the class v{x}.

	v{x} must be one of the following -{Class} object:
	@ops
	.| -{Int}     | Convert address to Int if possible, throw EValue otherwise.
	.| -{Long}    | Convert address to Long.
	.| -{Integer} | Convert address to Int if possible, Long otherwise.
	.| -{Pointer} | Return self.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{String}  | Delegate to #{__string}.

	@exc
	.EArgValue
	.EValue
*/
static qu_any qu__pointer__as (qu_pointer self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:
		{
			qu_any x = qu_integer_newul ((unsigned long) self->ptr);
			QU_ARG_CHECK (qu_int_is (x));
			return x;
		}

		case QU_CLASS_ID_Long:
		case QU_CLASS_ID_Integer:
			return qu_integer_newul ((unsigned long) self->ptr);

		case QU_CLASS_ID_Boolean: return qu_pointer_as_boolean (self);
		case QU_CLASS_ID_String:  return qu_pointer_as_string (self);
		case QU_CLASS_ID_Pointer: return self;
	}
	QU_ARG_THROW;
}



/*) sub __cmp (x) -> Int
	If v{x} is a Pointer then compare the pointer addresses otherwise delegate
	to C{Object.__cmp}.

	@noexc
*/
static qu_int qu__pointer__cmp (qu_pointer self, qu_any x)
{
	if (qu_pointer_is (x))
	{
		if (self->ptr > ((qu_pointer) x)->ptr)
			return QuInt1;

		else if (self->ptr < ((qu_pointer) x)->ptr)
			return QuInt_1;

		else
			return QuInt0;
	}
	return qu_object_default_cmp (self, x);
}



/*) sub __clear -> self
	Set all bytes to 0.

	@exc
	.ECall   pointer is NULL
	.EModify
*/
static qu_pointer qu__pointer__clear (qu_pointer self)
{
	ENSURE_NOT_FROZEN (self);
	QU_REQUIRE (QuECall, self->ptr && self->size >= 0);
	memset (self->ptr, 0, self->size);
	return self;
}



/*) sub __decr -> Pointer
	Return a new Pointer by subtracting #{delta} from this pointer.

	The new Pointer object will inherit C{self.delta}.

	@exc
	.EOverflow result is NULL
*/
static qu_pointer qu__pointer__decr (qu_pointer self)
{
	void       *m = self->ptr - self->delta;
	qu_pointer  p;

	QU_REQUIRE (QuEOverflow, m);
	p = qu_pointer_new (m);
	p->delta = self->delta;
	p->klass = self->klass;
	return p;
}



/*) sub __hash -> Int
	Return a hash value of the pointer.

	@noexc
*/
static qu_int qu__pointer__hash (qu_pointer self)
{
	return qu_int_new (self->ptr);
}



/*) sub __getitem (x) -> Any
	Get value of a particular field as specified by #{spec}.

	The pointer must not be NULL.

	@exc
	.EArgValue
	.EIndex
	.ENone  you have not set a #{spec} yet.

	@xmp
		struct TString
			klass : Class
			flags : Clong
			attrs : Pointer
			len   : Clong
			hash  : Clong
			str   : String
		;;

		println (TString.ofs ('str'))
		s = 'Hello!'
		printfln ("%p", s)
		p = Pointer.addr (s)
		p.spec (TString)
		println (p->str)
*/
static qu_any qu__pointer__getitem (qu_pointer self, qu_any x)
{
	char type;
	long ofs;

	QU_ARG_CHECK (self->ptr);
	QU_REQUIRE (QuENone, self->spec);
	type = qu_pointer__ofs (self->spec, x, &ofs);
	QU_ENSURE (type);
	return qu_pointer_get (self->ptr + ofs, type);
}



/*) sub __getslice (idx: Integer, len: Integer) -> Array
	Get v{len} field values string from v{idx}.

	The pointer must not be NULL and #{spec} must have been set.

	@exc
	.EArgValue
	.EIndex
	.ENone  you have not set a #{spec} yet.

	@xmp
		struct TString
			klass : Class
			flags : Clong
			attrs : Pointer
			len   : Clong
			hash  : Clong
			str   : String
		;;

		p = Pointer.addr ('Hello!')
		p.spec (TString)
		(l, h, s) = p[3, 3]
		println (l)
		println (h)
		println (s)

	@out
		6
		-1
		Hello!
*/
static qu_array qu__pointer__getslice (qu_pointer self, qu_integer idx, qu_integer len)
{
	long     i, l;
	qu_array a;

	QU_REQUIRE (QuENone, self->spec);
	QU_ARG_long (idx, i);
	QU_ARG_long (len, l);
	QU_ARG_CHECK (idx >= 0 && len >= 0);
	a = qu_array_new (4);
	while (l--)
	{
		qu_any x;
		char   type;
		long   ofs = qu_struct_offset_ex (self->spec, i++, &type);

		if (ofs < 0)
			break;

		x = qu_pointer_get (self->ptr + ofs, type);
		qu_array_push (a, x);
	}
	return a;
}



/*) sub __incr -> Pointer
	Return a new Pointer by adding #{delta} to this pointer.

	The new Pointer object will inherit C{self.delta}.

	@noexc
	.EOverflow result is NULL
*/
static qu_pointer qu__pointer__incr (qu_pointer self)
{
	void       *m = self->ptr + self->delta;
	qu_pointer  p;

	QU_REQUIRE (QuEOverflow, m);
	p = qu_pointer_new (m);
	p->delta = self->delta;
	p->klass = self->klass;
	return p;
}



/*) sub __len -> Int
	Return the memory size allocated or 0 if none.

	@noexc
*/
/*) sub len -> Int
	This is an alias of #{__len}.
*/
/*) sub size -> Int
	This is an alias of #{__len}.
*/
static qu_int qu__pointer__len (qu_pointer self)
{
	return qu_int_new (self->size);
}



/*) sub __like (x) -> Boolean
	Return true if v{x} is a Pointer and points to the same memory location
	as self. Return false otherwise.

	@noexc
*/
static qu_boolean qu__pointer__like (qu_pointer self, qu_any x)
{
	return (qu_pointer_is (x) && ((qu_pointer) x)->ptr == self->ptr) ? QuTrue : QuFalse;
}



/*) sub __minus (x) -> Any
	If v{x} is an -{Integer} then return a new Pointer with a memory address which
	is the result of subtracting self with v{x} bytes multiplied by #{delta}.

	The new Pointer object will inherit C{self.delta}.

	Otherwise v{x} must be a Pointer in which case the address difference is returned,
	thus an Integer.

	@exc
	.EMinus
	.EOverflow  result is NULL

	@xmp
		p1 = Pointer (10)
		p2 = Pointer (20)
		println (p1 - p2)
		p1.free
		p2.free

	@out
		-32
*/
static qu_any qu__pointer__minus (qu_pointer self, qu_any x)
{
	if (qu_integer_is (x))
	{
		long       i;
		void      *m;
		qu_pointer p;

		QU_ARG_long (x, i);
		m = self->ptr + i * self->delta;
		QU_REQUIRE (QuEOverflow, m);
		p = qu_pointer_new (m);
		p->delta = self->delta;
		p->klass = self->klass;
		return p;
	}
	else
	{
		long d;
		QU_REQUIRE (QuEMinus, qu_pointer_is (x));
		d = self->ptr - ((qu_pointer) x)->ptr;
		return qu_integer_newl (d);
	}
}



/*) sub __plus (x) -> Any
	Return a new Pointer with a memory address which is the result of adding
	self memory address with v{x}.

	If v{x} is an -{Integer} then v{x} is multiplied by #{delta} bytes.
	Otherwise it must be a Pointer object in which case the addresses are added.

	The new Pointer object will inherit C{self.delta}.

	@exc
	.EPlus
	.EOverflow  result is NULL

	@xmp
		p = Pointer (100)
		p.delta (8)
		p1 = p++
		println (p.hex)
		println (p1.hex)

	@out
		400af9ac
		400af9a4
*/
static qu_pointer qu__pointer__plus (qu_pointer self, qu_any x)
{
	void       *m;
	qu_pointer  p;

	if (qu_integer_is (x))
	{
		long i;

		QU_ARG_long (x, i);
		m = self->ptr + i * self->delta;
	}
	else
	{
		QU_REQUIRE (QuEPlus, qu_pointer_is (x));
		m = (void *) ((unsigned long) self->ptr + (unsigned long) ((qu_pointer) x)->ptr);
	}

	QU_REQUIRE (QuEOverflow, m);
	p = qu_pointer_new (m);
	p->delta = self->delta;
	p->klass = self->klass;
	return p;
}



/*) sub __string -> String */
qu_string qu__pointer__string (qu_pointer self)
{
	qu_string s = self->klass->name;
	qu_struct u = self->spec;

	if (self->ptr == 0)
	{
		return qu_string_newf ("<%s NULL>", s->str);
	}
	else if (self->klass == QuPointerClass || u == 0 || !qu_struct_is (u))
	{
		return qu_string_newf ("<%s %p>", s->str, self->ptr);
	}
	else
	{
		long   i, n = qu_struct_len (u);
		qu_any x;

		__qu_buffer__;
		qu_sys_printz ("<", 1);
		qu_sys_printz (s->str, s->len);
		for (i = 0; i < n; ++i)
		{
			qu_sys_printz (" ", 1);
			x = qu__pointer__getitem (self, qu_int_new (i));
			assert (x);
			qu_sys_print (x);
		}
		qu_sys_printz (">", 1);
		return __qu_collect__;
	}
}



/*) sub __print -> Null */
static qu_null qu__pointer__print (qu_pointer self)
{
	qu_string s = qu__pointer__string (self);
	return qu_sys_printz (s->str, s->len);
}



/*) sub __setitem (x, v) -> Any
	Set value of a particular field as specified by #{spec}.

	The pointer must not be NULL.

	@exc
	.EModify
	.EArgValue
	.ENone  you have not set a #{spec} yet.

	@p
	Here is an example you must never do.

	@xmp
		struct TString
			klass : Class
			flags : Clong
			attrs : Pointer
			len   : Clong
			hash  : Clong
			str   : String
		;;

		println (TString.spec)
		println (TString.ofs ('str'))
		s = 'Hello!'
		printfln ("%p", s)
		p = Pointer.addr (s, TString)
		println (s.__hash)
		p->hash = -1
		println (p->hash)

	@out
		Clplls
		20
		0x40071fd8
		794045173
		-1
*/
static qu_any qu__pointer__setitem (qu_pointer self, qu_any x, qu_any v)
{
	long ofs;
	char type;

	ENSURE_NOT_FROZEN (self);
	QU_ARG_CHECK (self->ptr);
	QU_REQUIRE (QuENone, self->spec);
	type = qu_pointer__ofs (self->spec, x, &ofs);
	QU_ENSURE (type);
	QU_ENSURE (qu_pointer_set (self->ptr + ofs, type, v));
	return v;
}



/*) sub __setslice (idx: Integer, len: Integer, val: Array) -> Array
	Set v{len} field values string from v{idx}.

	The pointer must not be NULL and #{spec} must have been set.

	@exc
	.EArgValue
	.ENone  you have not set a #{spec} yet.

	@xmp
		struct TFoo
			foo : Cint
			bar : Clong
			baz : Cdouble
		;;

		p = Pointer (TFoo)
		p[0, 3] = [1, 2, 3.4]
		println (p->foo)
		println (p->bar)
		println (p->baz)
		p.free

	@out
		1
		2
		3.3999999999999999
*/
static qu_array qu__pointer__setslice (qu_pointer self, qu_integer idx, qu_integer len, qu_array val)
{
	long i, l, j = 0;

	ENSURE_NOT_FROZEN (self);
	QU_REQUIRE (QuENone, self->spec);
	QU_ARG_long (idx, i);
	QU_ARG_long (len, l);
	QU_ARG_CHECK (qu_array_is (val));
	QU_REQUIRE (QuEIndex, i >= 0);
	for (; l && j < val->len; --l, ++j)
	{
		qu_any x = val->objs[j];
		char   type;
		long   ofs = qu_struct_offset_ex (self->spec, i++, &type);

		if (ofs < 0)
			break;

		QU_ENSURE (qu_pointer_set (self->ptr + ofs, type, x));
	}
	return val;
}



/*) sub __true -> Boolean
	Return false if the underlying pointer is NULL, true otherwise.
*/
static qu_boolean qu__pointer__true (qu_pointer self)
{
	return qu_pointer_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub bin -> String
	Return a binary representation of the pointer.

	@noexc

	@xmp
		println (myptr.bin)

	@out
		1000000000011010011010101000100
*/
static qu_string qu__pointer_bin (qu_pointer self)
{
	qu_any x = qu__pointer__as (self, QuIntegerClass);
	if (qu_int_is (x))
		return qu_int_bin ((qu_int) x);
	assert (qu_long_is (x));
	return qu_long_bin ((qu_long) x);
}



/*) sub clear (value = 0)
	Set all bytes to v{value}.

	@exc
	.ECall     pointer is NULL
	.EArgValue
*/
static qu_pointer qu__pointer_clear (qu_pointer self, int argc, qu_any *argv)
{
	int v = 0;

	ENSURE_NOT_FROZEN (self);
	QU_REQUIRE (QuECall, self->ptr && self->size >= 0);
	QU_ARG_VCOUNT (0, 1);
	if (argc)
		QU_ARG_vint (0, v);
	memset (self->ptr, v, self->size);
	return self;
}



/*) sub delta (*arg)
	@def
	sub delta -> Int
	sub delta (value: Integer) -> self
	@p
	Get or set the incremental value for this Pointer.

	After setting the value then all arithmetic methods such as #{__incr}
	and #{__plus} will use this value. The resulting new Pointer value will
	inherit this value.

	When you first create a Pointer object, it's v{delta} value is 1.

	@exc
	.EArgValue

	@xmp
		p = original = Pointer (20)
		p.delta = 4
		println (p++)
		println (p)
		q = p + 1
		println (q)
		original.free

	@out
		<Pointer 0x400dfba4>
		<Pointer 0x400dfba8>
		<Pointer 0x400dfbac>
*/
static qu_any qu__pointer_delta (qu_pointer self, int argc, qu_any *argv)
{
	if (argc == 0)
		return qu_int_new (self->delta);
	else
	{
		long d;

		ENSURE_NOT_FROZEN (self);
		QU_ARG_vlong (0, d);
		QU_ARG_CHECK  (d >= 0);
		self->delta = d;
		return self;
	}
}



/*) sub free -> Null
	Free an allocated pointer created with the #{new} method.

	See the #{new] method for an example.

	Once freed, you can no longer use the pointer with the #{set} and
	#{get} methods.

	@exc
	.ECall the pointer was not allocated

	@xmp
		p = Pointer (20)
		...
		p.free
*/
static qu_null qu__pointer_free (qu_pointer self)
{
	QU_REQUIRE (QuECall, qu_pointer_allocated (self));

	assert (self->ptr);
	qu_mem_free (self->ptr);
	qu_pointer_unset_allocated (self);
	self->ptr = NULL;
	return QuNil;
}



/*) sub freeze -> self
	Freeze this pointer so it can not be modified.

	@warn
	Once frozen, there is no way to unfreeze it.

	@noexc
*/
static qu_pointer qu__pointer_freeze (qu_pointer self)
{
	qu_pointer_freeze (self);
	return self;
}



/*) sub get (type = self.spec, ofs = 0) -> Any
	Get a value from the pointer.

	This is generally used to access a C struct field but you can use
	it for other purposes, of course.

	v{ofs} is the offset from the pointer. The default value is 0.

	v{type} may be a single character specification as explained in -{Struct.spec},
	or a Class object representing that specification, or a -{Struct}, or
	a -{Record}.

	If v{type} is a -{Struct} then a new -{Record} is returned which contains the
	values fetched from the pointer.

	If v{type} is a -{Record} then the values are fetched from the pointer.

	@exc
	.ECall  NULL or freed allocated pointer
	.EArgValue

	@xmp
		m = Pointer (20)

		m.set ('f', 1.2)
		x = m.get ('f')
		println (x)

		m.set (Cint, 12)
		x = m.get (Cint)
		println (x)

		m.free

	@out
		1.2
		12
*/
static qu_any qu__pointer_get (qu_pointer self, int argc, qu_any *argv)
{
	qu_any x = self->spec;
	char   type;
	int    ofs = 0;

	QU_REQUIRE (QuECall, self->ptr);
	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		x = argv[0];
		if (argc > 1)
			QU_ARG_vint (1, ofs);
	}
	else
	{
		QU_ARG_CHECKZ (x, "Pointer has no specification");
	}

	if (qu_struct_is (x))
	{
		qu_record r = qu_record_alloc (x);
		qu_pointer_get_record (self->ptr + ofs, r);
		return r;
	}
	else if (qu_record_is (x))
	{
		qu_record r = (qu_record) x;
		QU_CONTAINER_ENSURE_NOT_FROZEN (r);
		qu_pointer_get_record (self->ptr + ofs, r);
		return r;
	}
	else
	{
		type = qu_struct_type_to_char (x);
		QU_ARG_CHECK (type);
		return qu_pointer_get (self->ptr + ofs, type);
	}
}



/*) sub gets (len = -1, ofs = 0) -> String
	Get a string from the pointer.

	This is different from #{get}. The pointer #{ptr} field is assumed to point to
	a nil terminated string thus not to a pointer to a string.

	@exc
	.ECall  NULL or freed allocated pointer
	.EArgValue

	@xmp
		p = Pointer (20)

		z = 0
		for c in 'Hello'
			p.set ("a", c, z++)
		;;

		s = p.gets (1)
		println (s)

		s = p.gets (3, 1)
		println (s)

	@out
		ello
		ell
*/
/*) sub getstring (len = -1, ofs = 0) -> String
	This is an alias of #{gets}.
*/
static qu_string qu__pointer_gets (qu_pointer self, int argc, qu_any *argv)
{
	int  ofs = 0;
	long len = -1;

	QU_REQUIRE (QuECall, self->ptr);
	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, len);
		if (argc > 1)
			QU_ARG_vint (1, ofs);
	}
	if (len < 0)
		return qu_string_newz (self->ptr + ofs);
	return qu_string_new (self->ptr + ofs, len);
}



/*) sub geta (type: String, len = -1, ofs = 0) -> Array
	Get an Array from a flat C pointer optionally start at a specified offset.

	v{type} is the type of the value.
	See -{Struct.spec} for the specifications.

	If v{len} is negative then calculate the length by assuming an array of pointers.
	A null pointer marks the end of the array.

	@exc
	.ECall  NULL or freed allocated pointer
	.EArgValue

	@xmp
		a = ['foo', 'bar', 'baz']
		p = Pointer (a, 's')
		b = p.geta ('s', 3)
		println (b)

	@out
		['foo', 'bar', 'baz']
*/
/*) sub getarray (type: String, len = -1, ofs = 0) -> Array
	This is an alias of #{geta}.
*/
static qu_array qu__pointer_geta (qu_pointer self, int argc, qu_any *argv)
{
	qu_array          a;
	qu_stringy        s;
	int               ofs = 0;
	long              z, l = -1;
	char              y;
	qu_struct_type_s *t;
	void             *p;

	QU_REQUIRE (QuECall, self->ptr);
	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, s, qu_stringy);
	if (argc > 1)
	{
		QU_ARG_vlong (1, l);
		if (argc > 2)
			QU_ARG_vint (2, ofs);
	}

	y = s->str[0];
	t = qu_struct_type_info (y);
	QU_ARG_CHECK (t->type);

	if (l < 0)
	{
		p = self->ptr + ofs;
		l = 0;
		while (*(void **) p)
		{
			p += SIZEOF_VOID_P;
			++l;
		}
	}

	QU_MEM_CHECK (l <= QU_GC_MAX_ITEMS);
	a = qu_array_new (l);
	p = self->ptr + ofs;
	z = t->size;
	while (l--)
	{
		qu_any x = qu_pointer_get (p, y);
		QU_ENSURE (x);
		qu_array_push_quick (a, x);
		p += z;
	}
	return a;
}



/*) sub hex -> String
	Return a hexadecimal representation of the pointer.

	@noexc

	@xmp
		println (myptr.hex)

	@out
		400d3544
*/
static qu_string qu__pointer_hex (qu_pointer self)
{
	qu_any x = qu__pointer__as (self, QuIntegerClass);
	if (qu_int_is (x))
		return qu_int_hex ((qu_int) x);
	assert (qu_long_is (x));
	return qu_long_hex ((qu_long) x);
}



/*) sub index (x) -> Integer
	Return the index of the field v{x}.

	The class must define a -{Struct} named as the class.

	@exc
	.EIndex
*/
static qu_integer qu__pointer_index (qu_pointer self, qu_any x)
{
	qu_struct u = qu__pointer_getstruct (self->klass, 1);
	long      i;

	QU_ENSURE (u);
	i = qu_struct_index_ex (u, x);
	QU_REQUIRE (QuEIndex, i >= 0);
	return qu_integer_newl (i);
}



/*) sub isnull -> Boolean
	Return true if this is the NULL pointer or contains a pointer which value is NULL,
	false otherwise.

	@noexc
*/
static qu_boolean qu__pointer_isnull (qu_pointer self)
{
	if (self->ptr == NULL || *((void **) self->ptr) == NULL)
		return QuTrue;
	return QuFalse;
}



/*) sub oct -> String
	Return an octal representation of the pointer.

	@noexc

	@xmp
		println (myptr.oct)

	@out
		10003232504
*/
static qu_string qu__pointer_oct (qu_pointer self)
{
	qu_any x = qu__pointer__as (self, QuIntegerClass);
	if (qu_int_is (x))
		return qu_int_oct ((qu_int) x);
	assert (qu_long_is (x));
	return qu_long_oct ((qu_long) x);
}



/*) sub pack -> String
	Return the memory pointed by the Pointer as a String.

	@exc
	.ECall pointer must not be null and must have a #{size}

	@xmp
		p = Pointer (20)
		p.set ('l', 12345)
		p.sets ('Hello', Struct.sizeof ('l'))
		s = p.pack
		println (s.quote)

	@out
		'90\x00\x00Hello\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
*/
static qu_string qu__pointer_pack (qu_pointer self)
{
	QU_REQUIRE (QuECall, self->size && self->ptr);
	return qu_string_new (self->ptr, self->size);
}



/*) sub put (p: Pointer, ofs = 0) -> Int
	Put the values into a C pointer at a specified offset then return the
	offset for the next item.

	@warn
	Currently this only works if #{spec} is a Struct.

	@exc
	.EArgValue
*/
static qu_integer qu__pointer_put (qu_pointer self, int argc, qu_any *argv)
{
	qu_pointer p;
	qu_struct  u;
	void      *m;
	long       i, n, ofs = 0;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, p, qu_pointer);
	if (argc > 1)
		QU_ARG_vlong (1, ofs);

	u = self->spec;
	QU_ARG_CHECK (u && qu_struct_is (u));
	n = qu_struct_len (u);

	m = p->ptr + ofs;
	for (i = 0; i < n; ++i)
	{
		qu_any            x = qu__pointer__getitem (self, qu_int_new (i));
		qu_structentry_s *e = &u->items [i];
		char              c = qu_struct_type_to_char (e->type);
		qu_struct_type_s *f = qu_struct_type_info (c);

		assert (x);
		QU_ARG_CHECK (c);
		QU_ENSURE (qu_pointer_set (m, c, x));
		m += f->size;
	}
	ofs += m - p->ptr;
	return qu_integer_newl (ofs);
}



/*) sub set (*arg) -> self
	@def
	sub set (value: Record, offset = 0) -> self
	sub set (type, value, offset = 0) -> self
	@p
	Set value in the memory address pointed by the pointer.

	See the #{get} method for the arguments and example.

	You sould probably not modify the memory unless you created
	the pointer yourself.

	@exc
	.ECall  NULL or freed allocated pointer
	.EArgValue
*/
static qu_pointer qu__pointer_set (qu_pointer self, int argc, qu_any *argv)
{
	qu_any x;
	int    ofs = 0;
	char   type;

	ENSURE_NOT_FROZEN (self);
	QU_REQUIRE (QuECall, self->ptr);
	QU_ARG_VCOUNT (1, 3);
	x = argv[0];

	if (qu_record_is (x))
	{
		if (argc > 1)
			QU_ARG_vint (1, ofs);
		QU_ENSURE (qu_pointer_set_record (self->ptr + ofs, x));
	}
	else
	{
		type = qu_struct_type_to_char (argv[0]);
		QU_ARG_CHECK (type);
		if (argc > 2)
			QU_ARG_vint (2, ofs);
		QU_ENSURE (qu_pointer_set (self->ptr + ofs, type, argv[1]));
	}
	return self;
}



/*) sub sets (x: String, ofs = 0) -> self
	Set string.

	This is different from #{set}. The String content will be copied to the
	address #{ptr} with a nil terminator. Optionally start at offset v{ofs}
	from the address.

	@exc
	.ECall  NULL or freed allocated pointer
	.EIndex
	.EArgValue

	@xmp
		p = Pointer (20)
		p.sets ('Hello')
		s = p.gets
		println (s)

	@out
		Hello
*/
static qu_pointer qu__pointer_sets (qu_pointer self, int argc, qu_any *argv)
{
	qu_string x;
	int       ofs = 0;

	ENSURE_NOT_FROZEN (self);
	QU_REQUIRE (QuECall, self->ptr);
	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, x, qu_string);
	if (argc > 1)
		QU_ARG_vint (1, ofs);
	memcpy (self->ptr + ofs, x->str, x->len + 1);
	return self;
}



/*) sub spec (*arg) -> Any
	@def
	sub spec -> Struct
	sub spec (u: Struct) -> self
	sub spec (u: String) -> self
	@p
	Set or get the C struct specification.

	If called with no arguments, return the specification.
	If called with one argument, set the specification.

	Once set you use the #{__getitem} or #{__setitem} to get or set particular
	values.

	If you pass an String then see the #{get} method for valid types. The __getitem
	and __setitem can only be used with Integer indexes. The value offsets are
	automatically calculated.

	If you pass a Struct then you can use __getitem and __setitem with Integer
	or named indexes.

	If you pass a Dict then see the #{get} method for valid types. The __getitem
	and __setitem must be used named indexes. The keys of the dict are the field
	names and the values are strings of the form "<type><offset>".

	@warn
	This property is not inherited to new pointers resulting from an arithmetic
	operation.

	@exc
	.ECall  NULL or freed allocated pointer
	.EArgValue
	.ENone

	@xmp
		struct Foo
			bar: Cint
			baz: Cdouble
		;;

		p.spec (Foo)
		p[bar] = 1
		x = p[baz]

		p.spec ('ii')
		p[0] = 1
		x = p[1]

		p.spec (bar: 'i0', baz: 'F4')
		p[bar] = 1
		x = p[baz]
*/
static qu_any qu__pointer_spec (qu_pointer self, int argc, qu_any *argv)
{
	QU_REQUIRE (QuECall, self->ptr);

	if (argc == 0)
	{
		QU_REQUIRE (QuENone, self->spec);
		return self->spec;
	}
	else
	{
		qu_any x = argv[0];
		ENSURE_NOT_FROZEN (self);
		QU_ARG_CHECK (qu_string_is (x) || qu_struct_is (x) || qu_dict_is (x));
		self->spec = x;
		return self;
	}
}



/*) sub value -> Integer
	Return the address pointed by this pointer as an Integer.

	@noexc

	@xmp
		println (Pointer.NULL.value)

	@out
		0
*/
static qu_integer qu__pointer_value (qu_pointer self)
{
	long v = (long) self->ptr;
	return qu_integer_newl (v);
}



/*) sub __dump -> Null
	Dump the hex content of the pointer to v{stderr}.

	This method is provided for debugging purposes.

	@noexc
*/
static qu_null qu__pointer__dump (qu_pointer self)
{

	unsigned char *s = (unsigned char *) self->ptr;
	long  z = self->size;

	fprintf (stderr, "(%ld) ", z);

	if (s == NULL)
		fprintf (stderr, "nil");
	else
	{
		assert (z >= 0);
		while (z--)
			fprintf (stderr, "%02x ", *s++);
	}
	return QuNil;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

qu_any qu_pointer_get (void *p, char type)
{
	switch (type)
	{
		case QU_STRUCT_ID_char:
		{
			char v = *((char *) p);
			return qu_string_byte (v);
		}

		case QU_STRUCT_ID_byte:
		{
			char v = *((char *) p);
			return qu_int_newuh ((unsigned short) v);
		}

		case QU_STRUCT_ID_ubyte:
		{
			unsigned char v = *((unsigned char *) p);
			return qu_int_newh ((unsigned short) v);
		}

		case QU_STRUCT_ID_short:
		{
			short v = *((short *) p);
			return qu_integer_newh (v);
		}

		case QU_STRUCT_ID_ushort:
		{
			unsigned short v = *((unsigned short *) p);
			return qu_integer_newuh (v);
		}

		case QU_STRUCT_ID_int:
		{
			int v = *((int *) p);
			return qu_integer_newi (v);
		}

		case QU_STRUCT_ID_uint:
		{
			unsigned int v = *((unsigned int *) p);
			return qu_integer_newui (v);
		}

#if SIZEOF_SHORT == 2
		case QU_STRUCT_ID_int16:
		{
			short v = *((short *) p);
			return qu_integer_newh (v);
		}

		case QU_STRUCT_ID_uint16:
		{
			unsigned short v = *((unsigned short *) p);
			return qu_integer_newuh (v);
		}
#endif

#if SIZEOF_INT == 4
		case QU_STRUCT_ID_int32:
		{
			int v = *((int *) p);
			return qu_integer_newi (v);
		}

		case QU_STRUCT_ID_uint32:
		{
			unsigned int v = *((unsigned int *) p);
			return qu_integer_newui (v);
		}
#endif

		case QU_STRUCT_ID_long:
		{
			long v = *((long *) p);
			return qu_integer_newl (v);
		}


		case QU_STRUCT_ID_ulong:
		{
			unsigned long v = *((unsigned long *) p);
			return qu_integer_newul (v);
		}

#ifdef HAVE_LONG_LONG
		case QU_STRUCT_ID_longlong:
		{
			long long v = *((long long *) p);
			return qu_integer_newll (v);
		}

		case QU_STRUCT_ID_ulonglong:
		{
			unsigned long long v = *((unsigned long long *) p);
			return qu_integer_newull (v);
		}
#endif

		case QU_STRUCT_ID_float:
		{
			float v = *((float *) p);
			return qu_float_new (v);
		}

		case QU_STRUCT_ID_double:
		{
			double v = *((double *) p);
			return qu_float_new (v);
		}

#ifdef HAVE_LONG_DOUBLE
		case QU_STRUCT_ID_longdouble:
		{
			long double v = *((long double *) p);
			return qu_big_newld (v);
		}
#endif

		case QU_STRUCT_ID_ssize_t:
		{
#if SIZEOF_SIZE_T == SIZEOF_INT
			int v = (int) *((ssize_t *) p);
			return qu_integer_newi (v);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
			long v = (long) *((ssize_t *) p);
			return qu_integer_newl (v);
#else
            long long v = (long long) *((ssize_t *) p);
			return qu_integer_newll (v);
#endif
		}

		case QU_STRUCT_ID_size_t:
		{
#if SIZEOF_SIZE_T == SIZEOF_INT
			unsigned int v = (unsigned int) *((size_t *) p);
			return qu_integer_newui (v);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
			unsigned long v = (unsigned long) *((size_t *) p);
			return qu_integer_newul (v);
#else
            unsigned long long v = (unsigned long long) *((size_t *) p);
			return qu_integer_newull (v);
#endif
		}

		case QU_STRUCT_ID_off_t:
		{
#if SIZEOF_OFF_T == SIZEOF_INT
			int v = (int) *((off_t *) p);
			return qu_integer_newi (v);
#elif SIZEOF_OFF_T == SIZEOF_LONG
			long v = (long) *((off_t *) p);
			return qu_integer_newl (v);
#else
            long long v = (long long) *((off_t *) p);
			return qu_integer_newll (v);
#endif
		}

		case QU_STRUCT_ID_Null:
			return QuNil;

		case QU_STRUCT_ID_Boolean:
		{
			int v = *((int *) p);
			return v ? QuTrue : QuFalse;
		}

		case QU_STRUCT_ID_Int:
		{
			long v = *((long *) p);
			return qu_integer_newli (v);
		}

		case QU_STRUCT_ID_void_p:
		case QU_STRUCT_ID_Pointer:
		{
			void *v = *((void **) p);
			if (v == NULL)
				return QuPointerNULL;
			return qu_pointer_new (v);
		}

		case QU_STRUCT_ID_char_p:
		case QU_STRUCT_ID_String:
		{
			char *v = *((char **) p);
			if (v == NULL)
				return QuStringEMPTY;
			return qu_string_newz (v);
		}

		case QU_STRUCT_ID_Array:
		{
			qu_array v = *((qu_array *) p);
			if (v && qu_array_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Dict:
		{
			qu_dict v = *((qu_dict *) p);
			if (v && qu_dict_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Set:
		{
			qu_set v = *((qu_set *) p);
			if (v && qu_set_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Range:
		{
			qu_range v = *((qu_range *) p);
			if (v && qu_range_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Class:
		{
			qu_class v = *((qu_class *) p);
			if (v && qu_class_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Queue:
		{
			qu_queue v = *((qu_queue *) p);
			if (v && qu_queue_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Stream:
		{
			qu_stream v = *((qu_stream *) p);
			if (v && qu_stream_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Proc:
		{
			qu_proc v = *((qu_proc *) p);
			if (v && qu_proc_is (v))
				return v;
			break;
		}

		case QU_STRUCT_ID_Any:
		{
			qu_any v = *((qu_any *) p);
			if (v == NULL)
				return QuNil;
			return v;
		}

#ifdef QU_SUPPORT_GLIB
		case QU_STRUCT_ID_GObject:
		{
			void *v = *((void **) p);
			if (v == NULL)
				return QuNil;
			return __QuGObjectNew (v);
		}
#endif

	}
	QU_ARG_THROWF ("cant get type \"%c\"", type);
}



int qu_pointer_get_record (void *p, qu_record r)
{
	qu_struct u = r->struc;
	long      n = qu_struct_len (u);
	long      i, ofs;
	char      type;
	qu_any    x;

	for (i = 0; i < n; ++i)
	{
		ofs = qu_struct_offset_ex (u, i, &type);
		QU_ARG_CHECK (ofs >= 0);
		x = qu_pointer_get (p + ofs, type);
		QU_ENSURE (x);
		x = qu_object_validate (r, u->items[i].type, x);
		QU_ENSURE (x);
		r->objs[i] = x;
	}
	return 1;
}



int qu_pointer_set_record (void *p, qu_record r)
{
	qu_struct  u = r->struc;
	long       n = qu_struct_len (u);
	qu_any    *o = r->objs;
	char       type;
	long       ofs, i;

	for (i = 0; i < n; ++i, ++o)
	{
		ofs = qu_struct_offset_ex (u, i, &type);
		QU_ENSURE (ofs >= 0 && qu_pointer_set (p + ofs, type, *o));
	}
	return 1;
}



int qu_pointer_set (void *p, char type, qu_any x)
{
	switch (type)
	{
		case QU_STRUCT_ID_char:
			if (x == (qu_any) QuNil)
			{
				*((char *) p) = 0;
				return 1;
			}
			else if (qu_string_is (x) && ((qu_string) x)->len > 0)
			{
				*((char *) p) = ((qu_string) x)->str[0];
				return 1;
			}
			break;

		case QU_STRUCT_ID_byte:
			if (x == (qu_any) QuNil)
			{
				*((char *) p) = 0;
				return 1;
			}
			else
			{
				short v = qu_number_toh (x);
				if (errno == 0 && -127 <= v && v <= 127)
				{
					*((char *) p) = (char) v;
					return 1;
				}
			}
			break;

		case QU_STRUCT_ID_ubyte:
			if (x == (qu_any) QuNil)
			{
				*((unsigned char *) p) = 0;
				return 1;
			}
			else
			{
				unsigned short v = qu_number_touh (x);
				if (errno == 0 && v <= 255)
				{
					*((unsigned char *) p) = (unsigned char) v;
					return 1;
				}
			}
			break;

		case QU_STRUCT_ID_short:
			if (x == (qu_any) QuNil)
			{
				*((short *) p) = 0;
				return sizeof (short);
			}
			else
			{
				short v = qu_number_toh (x);
				if (errno == 0)
				{
					*((short *) p) = v;
					return sizeof (short);
				}
			}
			break;

		case QU_STRUCT_ID_ushort:
			if (x == (qu_any) QuNil)
			{
				*((short *) p) = 0;
				return sizeof (short);
			}
			else
			{
				unsigned short v = qu_number_touh (x);
				if (errno == 0)
				{
					*((unsigned short *) p) = v;
					return sizeof (short);
				}
			}
			break;

		case QU_STRUCT_ID_int:
			if (x == (qu_any) QuNil)
			{
				*((int *) p) = 0;
				return sizeof (int);
			}
			else
			{
				int v = qu_number_toi (x);
				if (errno == 0)
				{
					*((int *) p) = v;
					return sizeof (int);
				}
			}
			break;

		case QU_STRUCT_ID_uint:
			if (x == (qu_any) QuNil)
			{
				*((unsigned int *) p) = 0;
				return sizeof (int);
			}
			else
			{
				unsigned int v = qu_number_toui (x);
				if (errno == 0)
				{
					*((unsigned int *) p) = v;
					return sizeof (int);
				}
			}
			break;

#if SIZEOF_SHORT == 2
		case QU_STRUCT_ID_int16:
			if (x == (qu_any) QuNil)
			{
				*((short *) p) = 0;
				return 2;
			}
			else
			{
				short v = qu_number_toh (x);
				if (errno == 0)
				{
					*((short *) p) = v;
					return 2;
				}
			}
			break;

		case QU_STRUCT_ID_uint16:
			if (x == (qu_any) QuNil)
			{
				*((unsigned short *) p) = 0;
				return 2;
			}
			else
			{
				unsigned short v = qu_number_touh (x);
				if (errno == 0)
				{
					*((unsigned short *) p) = v;
					return 2;
				}
			}
			break;
#endif

#if SIZEOF_INT == 4
		case QU_STRUCT_ID_int32:
			if (x == (qu_any) QuNil)
			{
				*((int *) p) = 0;
				return 4;
			}
			else
			{
				int v = qu_number_toi (x);
				if (errno == 0)
				{
					*((int *) p) = v;
					return 4;
				}
			}
			break;

		case QU_STRUCT_ID_uint32:
			if (x == (qu_any) QuNil)
			{
				*((int *) p) = 0;
				return 4;
			}
			else
			{
				unsigned int v = qu_number_toui (x);
				if (errno == 0)
				{
					*((unsigned int *) p) = v;
					return 4;
				}
			}
			break;
#endif

		case QU_STRUCT_ID_long:
			if (x == (qu_any) QuNil)
			{
				*((long *) p) = 0;
				return sizeof (long);
			}
			else
			{
				long v = qu_number_tol (x);
				if (errno == 0)
				{
					*((long *) p) = v;
					return sizeof (long);
				}
			}
			break;

		case QU_STRUCT_ID_ulong:
			if (x == (qu_any) QuNil)
			{
				*((unsigned long *) p) = 0;
				return sizeof (long);
			}
			else
			{
				unsigned long v = qu_number_toul (x);
				if (errno == 0)
				{
					*((unsigned long *) p) = v;
					return sizeof (long);
				}
			}
			break;

#ifdef HAVE_LONG_LONG
		case QU_STRUCT_ID_longlong:
			if (x == (qu_any) QuNil)
			{
				*((long long *) p) = 0;
				return sizeof (long long);
			}
			else
			{
				long long v = qu_number_toll (x);
				if (errno == 0)
				{
					*((long long *) p) = v;
					return sizeof (long long);
				}
			}
			break;

		case QU_STRUCT_ID_ulonglong:
			if (x == (qu_any) QuNil)
			{
				*((unsigned long long *) p) = 0;
				return sizeof (long long);
			}
			else
			{
				unsigned long long v = qu_number_toll (x);
				if (errno == 0)
				{
					*((unsigned long long *) p) = v;
					return sizeof (long long);
				}
			}
			break;
#endif

		case QU_STRUCT_ID_float:
			if (x == (qu_any) QuNil)
			{
				*((float *) p) = 0;
				return sizeof (float);
			}
			else
			{
				double v = qu_number_tod (x);
				if (errno == 0)
				{
					double f = v;

					if (v < 0)
						f = -v;

					if ((double) FLT_MIN <= f && f <= (double) FLT_MAX)
					{
						*((float *) p) = (float) v;
						return sizeof (float);
					}
				}
			}
			break;

		case QU_STRUCT_ID_double:
			if (x == (qu_any) QuNil)
			{
				*((double *) p) = 0;
				return sizeof (double);
			}
			else
			{
				double v = qu_number_tod (x);
				if (errno == 0)
				{
					*((double *) p) = v;
					return sizeof (double);
				}
			}
			break;

#ifdef HAVE_LONG_DOUBLE
		case QU_STRUCT_ID_longdouble:
			if (x == (qu_any) QuNil)
			{
				*((long double *) p) = 0;
				return sizeof (long double);
			}
			else
			{
				long double v = qu_number_told (x);
				if (errno == 0)
				{
					*((long double *) p) = v;
					return sizeof (long double);
				}
			}
			break;
#endif

		case QU_STRUCT_ID_ssize_t:
			if (x == (qu_any) QuNil)
			{
				*((ssize_t *) p) = 0;
				return sizeof (ssize_t);
			}
			else
#if SIZEOF_SIZE_T == SIZEOF_INT
			{
				int v = qu_number_toi (x);
				if (errno == 0)
				{
					*((ssize_t *) p) = (ssize_t) v;
					return sizeof (ssize_t);
				}
			}
#elif SIZEOF_SIZE_T == SIZEOF_LONG
			{
				long v = qu_number_tol (x);
				if (errno == 0)
				{
					*((ssize_t *) p) = (ssize_t) v;
					return sizeof (ssize_t);
				}
			}
#else
			{
				long long v = qu_number_toll (x);
				if (errno == 0)
				{
					*((ssize_t *) p) = (ssize_t) v;
					return sizeof (ssize_t);
				}
			}
#endif
			break;


		case QU_STRUCT_ID_size_t:
			if (x == (qu_any) QuNil)
			{
				*((size_t *) p) = 0;
				return sizeof (size_t);
			}
			else
			{
#if SIZEOF_SIZE_T == SIZEOF_INT
				unsigned int v = qu_number_toui (x);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
				unsigned long v = qu_number_toul (x);
#else
	            unsigned long long v = qu_number_toull (x);
#endif
				if (errno == 0)
				{
					*((size_t *) p) = (size_t) v;
					return sizeof (size_t);
				}
			}
			break;

		case QU_STRUCT_ID_off_t:
			if (x == (qu_any) QuNil)
			{
				*((size_t *) p) = 0;
				return sizeof (size_t);
			}
			else
			{
#if SIZEOF_OFF_T == SIZEOF_INT
				int v = qu_number_toi (x);
#elif SIZEOF_OFF_T == SIZEOF_LONG
				long v = qu_number_tol (x);
#else
	            long long v = qu_number_toll (x);
#endif
				if (errno == 0)
				{
					*((off_t *) p) = (off_t) v;
					return sizeof (off_t);
				}
			}
			break;

		case QU_STRUCT_ID_Null:
			*((void **) p) = NULL;
			return sizeof (void *);

		case QU_STRUCT_ID_Boolean:
			*((int *) p) = (qu_sys_true (x) == QuTrue);
			return sizeof (int);

		case QU_STRUCT_ID_Int:
			if (x == (qu_any) QuNil)
			{
				*((long *) p) = 0;
				return sizeof (long);
			}
			else
			{
				long v = qu_number_toli (x);
				if (errno == 0)
				{
					*((long *) p) = v;
					return sizeof (long);
				}
			}
			break;

		case QU_STRUCT_ID_void_p:
		case QU_STRUCT_ID_Pointer:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (void *);
			}
			else if (qu_pointer_is (x))
			{
				*((void **) p) = ((qu_pointer) x)->ptr;
				return sizeof (void *);
			}
			else if (qu_string_is (x))
			{
				*((void **) p) = ((qu_string) x)->str;
				return sizeof (void *);
			}
#ifdef QU_SUPPORT_GLIB
			else if (qu_gobject_is (x))
			{
				*((void **) p) = ((qu_gobject) x)->ptr;
				return sizeof (void *);
			}
#endif
			break;

		case QU_STRUCT_ID_char_p:
		case QU_STRUCT_ID_String:
			if (x == (qu_any) QuNil)
			{
				*((char **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_string_is (x))
			{
				*((char **) p) = ((qu_string) x)->str;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Array:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_array_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Dict:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_dict_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Set:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_set_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Range:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_range_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Class:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_class_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Queue:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_queue_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Stream:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_stream_is (x))
			{
				*((void **) p) = x;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Proc:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else if (qu_proc_is (x))
			{
				*((void **) p) = ((qu_proc) x)->code;
				return sizeof (qu_any);
			}
			break;

		case QU_STRUCT_ID_Any:
 			*((qu_any *) p) = x;
			return sizeof (qu_any);

#ifdef QU_SUPPORT_GLIB
		case QU_STRUCT_ID_GObject:
			if (x == (qu_any) QuNil)
			{
				*((void **) p) = NULL;
				return sizeof (qu_any);
			}
			else
			{
				QU_ARG_CHECKZ (((qu_gobject) x)->ptr, "GObject already destroyed");
				*((qu_any *) p) = ((qu_gobject) x)->ptr;
				return sizeof (qu_any);
			}
			break;
#endif

		default:
			QU_ARG_THROWF ("cant set type \"%c\"", type);
	}
	QU_ARG_THROWF ("mismatch %s for type \"%c\"", qu_classof (x)->name->str, type);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_pointer__setup ()
{
	qu_class_setup (QuPointerClass, qu_pointer_s, qu__pointer__destroy, qu__pointer__mark);
}



void qu_pointer__init ()
{
	QuPointerNULL = qu_pointer_new_static (NULL);
	qu_pointer_freeze (QuPointerNULL);

	/* specific class property */
	qu_class_constz (QuPointerClass, "NULL", QuPointerNULL);

	/* standard class methods */
	qu_class_method_class  (QuPointerClass, QU_NAME__inst, qu__pointer__inst   , -1);
	qu_class_method_class  (QuPointerClass, QU_NAME_new  , qu__pointer_new     , -1);
	qu_class_method_classz (QuPointerClass, "structof"   , qu__pointer_structof,  0);
	qu_class_method_classz (QuPointerClass, "sizeof"     , qu__pointer_sizeof  ,  0);
	qu_class_method_classz (QuPointerClass, "acquire"    , qu__pointer_acquire ,  1);
	qu_class_method_classz (QuPointerClass, "flat"       , qu__pointer_flat    , -1);
	qu_class_method_classz (QuPointerClass, "array"      , qu__pointer_array   ,  2);

	/* specific class methods */
	qu_class_method_classz (QuPointerClass, "addr"  , qu__pointer_addr  , -1);
	qu_class_method_classz (QuPointerClass, "addrof", qu__pointer_addrof,  1);

	/* standard instance methods */
	qu_class_method_inst   (QuPointerClass, QU_NAME__as      , qu__pointer__as      , 1);
	qu_class_method_inst   (QuPointerClass, QU_NAME__cmp     , qu__pointer__cmp     , 1);
	qu_class_method_inst   (QuPointerClass, QU_NAME__clear   , qu__pointer__clear   , 0);
	qu_class_method_inst   (QuPointerClass, QU_NAME__decr    , qu__pointer__decr    , 0);
	qu_class_method_inst   (QuPointerClass, QU_NAME__hash    , qu__pointer__hash    , 0);
	qu_class_method_inst   (QuPointerClass, QU_NAME__getitem , qu__pointer__getitem , 1);
	qu_class_method_inst   (QuPointerClass, QU_NAME__getslice, qu__pointer__getslice, 2);
	qu_class_method_inst   (QuPointerClass, QU_NAME__incr    , qu__pointer__incr    , 0);
	qu_class_method_inst   (QuPointerClass, QU_NAME__like    , qu__pointer__like    , 1);
	qu_class_method_inst   (QuPointerClass, QU_NAME__len     , qu__pointer__len     , 0);
	qu_class_alias         (QuPointerClass, QU_NAME__len     , QU_NAME_len);
	qu_class_aliasz        (QuPointerClass, "__len"          , "size");
	qu_class_method_inst   (QuPointerClass, QU_NAME__minus   , qu__pointer__minus   , 1);
	qu_class_method_inst   (QuPointerClass, QU_NAME__plus    , qu__pointer__plus    , 1);
	qu_class_method_inst   (QuPointerClass, QU_NAME__print   , qu__pointer__print   , 0);
	qu_class_method_inst   (QuPointerClass, QU_NAME__setitem , qu__pointer__setitem , 2);
	qu_class_method_inst   (QuPointerClass, QU_NAME__setslice, qu__pointer__setslice, 3);
	qu_class_method_inst   (QuPointerClass, QU_NAME__string  , qu__pointer__string  , 0);
	qu_class_method_inst   (QuPointerClass, QU_NAME__true    , qu__pointer__true    , 0);

	/* specific instance methods */
	qu_class_method_instz  (QuPointerClass, "bin"    , qu__pointer_bin   ,  0);
	qu_class_method_instz  (QuPointerClass, "clear"  , qu__pointer_clear , -1);
	qu_class_method_instz  (QuPointerClass, "delta"  , qu__pointer_delta , -1);
	qu_class_method_instz  (QuPointerClass, "free"   , qu__pointer_free  ,  0);
	qu_class_method_instz  (QuPointerClass, "freeze" , qu__pointer_freeze,  0);
	qu_class_method_instz  (QuPointerClass, "get"    , qu__pointer_get   , -1);
	qu_class_method_instz  (QuPointerClass, "gets"   , qu__pointer_gets  , -1);
	qu_class_method_instz  (QuPointerClass, "geta"   , qu__pointer_geta  , -1);
	qu_class_method_instz  (QuPointerClass, "hex"    , qu__pointer_hex   ,  0);
	qu_class_method_instz  (QuPointerClass, "index"  , qu__pointer_index ,  1);
	qu_class_method_instz  (QuPointerClass, "isnull" , qu__pointer_isnull,  0);
	qu_class_method_instz  (QuPointerClass, "oct"    , qu__pointer_oct   ,  0);
	qu_class_method_instz  (QuPointerClass, "pack"   , qu__pointer_pack  ,  0);
	qu_class_method_instz  (QuPointerClass, "put"    , qu__pointer_put   , -1);
	qu_class_method_instz  (QuPointerClass, "set"    , qu__pointer_set   , -1);
	qu_class_method_instz  (QuPointerClass, "sets"   , qu__pointer_sets  , -1);
	qu_class_method_instz  (QuPointerClass, "spec"   , qu__pointer_spec  , -1);
	qu_class_method_instz  (QuPointerClass, "value"  , qu__pointer_value ,  0);

	qu_class_method_instz  (QuPointerClass, "__dump" , qu__pointer__dump ,  0);

	/* finalize declaration */
	qu_class_enddecl (QuPointerClass);
}
