/* Proc.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


#ifdef LIGHTNING_CPU_NONE

qu_class QuProcClass;


static qu_proc qu__proc_new (qu_class self, int argc, qu_any *argv)
{
	QU_THROWZ (QuEImplement, "Proc requires GNU Lightning which unfortunately does not support your hardware");
}


void qu_proc__setup ()
{
	qu_class_setup_final (QuProcClass, qu_proc_s, 0, 0);
}


void qu_proc__init ()
{
	qu_class_method_class (QuProcClass, QU_NAME_new, qu__proc_new, -1);
}



#else /*LIGHTNING_CPU_NONE*/


#include "lightning.h"


#define QU_PROC_CALL_MAXARGS           0x00080 /* = 128 */
#define QU_PROC_CALL_FLAG_SELFPOS_MASK 0x000FF /* to get the index of self */
#define QU_PROC_CALL_FLAG_SAFE         0x10000 /* check for external thread */
#define QU_PROC_CALL_FLAG_LAST         0x20000 /* set self to last argument, first otherwise */

int qu_proc_cando_type (char type)
{
	switch (type)
	{
		case 0:
		case QU_STRUCT_ID_longlong:
		case QU_STRUCT_ID_ulonglong:
		case QU_STRUCT_ID_longdouble:
#if SIZEOF_SIZE_T > SIZEOF_LONG
		case QU_STRUCT_ID_size_t:
		case QU_STRUCT_ID_ssize_t:
#endif
#if SIZEOF_OFF_T > SIZEOF_LONG
		case QU_STRUCT_ID_off_t:
#endif
			return 0;
	}
	return 1;
}



/*----------------------------------------------------------------------------
	wrappers

	the generated qu_proc functions simply contain native code that call these
	wrappers which then do the call to Qu qu_sub objects and return the result

	PUSH argv
	PUSH argc
	CALL _proc_call_Any
----------------------------------------------------------------------------*/

#define THESELF \
	(argc \
		? ((flag & QU_PROC_CALL_FLAG_LAST) ? argv [argc - 1] : argv[flag & QU_PROC_CALL_FLAG_SELFPOS_MASK]) \
		: QuNil)

#define ENTER \
	qu_thread tHr; \
	int       wAsBlOcKiNg; \
	if (flag & QU_PROC_CALL_FLAG_SAFE) \
	{ \
		if (qu_thread_check_external ()) \
			return 0; \
	} \
	tHr = qu_thread_self (); \
	wAsBlOcKiNg = qu_thread_blocking (tHr); \
	if (wAsBlOcKiNg) \
		__qu_blocking_end

#define LEAVE \
	if (wAsBlOcKiNg) \
		__qu_blocking_begin


int qu_proc_call_char (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_string r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_string_is (r))
	{
		LEAVE;
		return r->str[0];
	}
	LEAVE;
	return 0;
}



int qu_proc_call_byte (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		int v = qu_integer_toi (r);
		if (errno == 0 && -127 <= v && v <= 127)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



int qu_proc_call_ubyte (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		int v = qu_integer_toi (r);
		if (errno == 0 && 0 <= v && v <= 255)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



short qu_proc_call_short (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		short v = qu_integer_toh (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



unsigned short qu_proc_call_ushort (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		unsigned short v = qu_integer_touh (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



int qu_proc_call_int (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		int v = qu_integer_toi (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	QU_SYS_WARN ("%s did not return int", u->name->str);
	LEAVE;
	return 0;
}



unsigned int qu_proc_call_uint (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		unsigned int v = qu_integer_toui (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



int qu_proc_call_int16 (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		int v = qu_integer_toi (r);
		if (errno == 0 && -23768 <= v && v <= 23377)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



unsigned int qu_proc_call_uint16 (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		unsigned int v = qu_integer_toui (r);
		if (errno == 0 && v <= 65535)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



long qu_proc_call_int32 (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		long v = qu_integer_tol (r);
		if (errno == 0 && (-2147483647L - 1L) <= v && v <= 2147483647L)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



unsigned long qu_proc_call_uint32 (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		unsigned long v = qu_integer_toul (r);
		if (errno == 0 && v <= 4294967295UL)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



long qu_proc_call_long (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		long v = qu_integer_tol (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



unsigned long qu_proc_call_ulong (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		unsigned long v = qu_integer_toul (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



#ifdef HAVE_LONG_LONG
long long qu_proc_call_longlong (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		long long v = qu_integer_toll (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}



unsigned long long qu_proc_call_ulonglong (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		unsigned long long v = qu_integer_toull (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0;
}
#endif /*HAVE_LONG_LONG*/



float qu_proc_call_float (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_number r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		double v = qu_number_tod (r);
		if (errno == 0)
		{
			double f = v;
			if (f < 0)
				f = -f;
			if ((double) FLT_MIN <= f && f <= (double) FLT_MAX)
			{
				LEAVE;
				return (float) v;
			}
		}
	}
	LEAVE;
	return 0.0;
}



double qu_proc_call_double (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_number r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		double v = qu_number_tod (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0.0;
}



#ifdef HAVE_LONG_DOUBLE
long double qu_proc_call_longdouble (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_number r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		long double v = qu_number_told (r);
		if (errno == 0)
		{
			LEAVE;
			return v;
		}
	}
	LEAVE;
	return 0.0;
}
#endif



#if SIZEOF_SIZE_T == SIZEOF_INT
#define qu_proc_call_ssize_t         qu_proc_call_int
#define qu_proc_call_size_t          qu_proc_call_uint
#elif SIZEOF_SIZE_T == SIZEOF_LONG
#define qu_proc_call_ssize_t         qu_proc_call_long
#define qu_proc_call_size_t          qu_proc_call_ulong
#else
#define qu_proc_call_ssize_t         qu_proc_call_longlong
#define qu_proc_call_size_t          qu_proc_call_ulonglong
#endif


#if SIZEOF_OFF_T == SIZEOF_INT
#define qu_proc_call_off_t           qu_proc_call_int
#elif SIZEOF_OFF_T == SIZEOF_LONG
#define qu_proc_call_off_t           qu_proc_call_long
#else
#define qu_proc_call_off_t           qu_proc_call_longlong
#endif



void *qu_proc_call_Null (qu_sub u, int flag, int argc, qu_any *argv)
{
	ENTER;
	qu_sub_calln (u, THESELF, argc, argv);
	LEAVE;
	return NULL;
}



int qu_proc_call_Boolean (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_boolean r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	LEAVE;
	return r == QuTrue;
}



long qu_proc_call_Int (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_integer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r)
	{
		long v = qu_integer_toli (r);
		if (errno == 0 && QU_INT_MIN <= v && v <= QU_INT_MAX)
		{
			LEAVE;
			return v;
		}
	}
	QU_SYS_WARN ("%s did not return Int", u->name->str);
	LEAVE;
	return 0;
}



#define qu_proc_call_void_p qu_proc_call_Pointer
void *qu_proc_call_Pointer (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_pointer r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_pointer_is (r))
	{
		LEAVE;
		return r->ptr;
	}
	LEAVE;
	return NULL;
}



#define qu_proc_call_char_p qu_proc_call_String
char *qu_proc_call_String (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_string r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_string_is (r))
	{
		LEAVE;
		return r->str;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Array (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_array r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_array_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Dict (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_dict r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_dict_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Set (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_set r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_set_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Range (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_range r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_range_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Class (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_class r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_class_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Queue (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_queue r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_queue_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Stream (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_stream r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_stream_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}



void *qu_proc_call_Proc (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_proc r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_proc_is (r))
	{
		LEAVE;
		return r;
	}
	LEAVE;
	return NULL;
}


#ifdef QU_SUPPORT_GLIB
void *qu_proc_call_GObject (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_gobject r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	if (r && qu_gobject_is (r))
	{
		LEAVE;
		return r->ptr;
	}
	LEAVE;
	return NULL;
}
#endif



qu_any qu_proc_call_Any (qu_sub u, int flag, int argc, qu_any *argv)
{
	qu_any r;

	ENTER;
	r = qu_sub_calln (u, THESELF, argc, argv);
	LEAVE;
	return r;
}



/*----------------------------------------------------------------------------
	C value to Qu object
----------------------------------------------------------------------------*/

qu_string qu_proc_cto_char (char v)
{
	return qu_string_byte (v);
}


qu_int qu_proc_cto_byte (char v)
{
	return qu_int_newi ((int) v);
}


qu_int qu_proc_cto_ubyte (unsigned char v)
{
	return qu_int_newi ((unsigned int) v);
}


qu_integer qu_proc_cto_short (short v)
{
	return qu_integer_newh (v);
}


qu_integer qu_proc_cto_ushort (unsigned short v)
{
	return qu_integer_newuh (v);
}


qu_integer qu_proc_cto_int (int v)
{
	return qu_integer_newi (v);
}


qu_integer qu_proc_cto_uint (unsigned int v)
{
	return qu_integer_newui (v);
}


#if SIZEOF_SHORT == 2
#define qu_proc_cto_int16             qu_proc_cto_short
#define qu_proc_cto_uint16            qu_proc_cto_ushort
#else
#define qu_proc_cto_int16             qu_proc_cto_int
#define qu_proc_cto_uint16            qu_proc_cto_uint
#endif


#if SIZEOF_INT == 4
#define qu_proc_cto_int32             qu_proc_cto_short
#define qu_proc_cto_uint32            qu_proc_cto_ushort
#else
#define qu_proc_cto_int32             qu_integer_newl
#define qu_proc_cto_uint32            qu_integer_newul
#endif


#define qu_proc_cto_long              qu_integer_newl
#define qu_proc_cto_ulong             qu_integer_newul
#define qu_proc_cto_longlong          qu_integer_newll
#define qu_proc_cto_ulonglong         qu_integer_newull


qu_any qu_proc_cto_float (float v)
{
	qu_any x = qu_float_newdx ((double) v);
	if (x == 0)
		/* let it generate an exception */
		return QuNil;
	return x;
}



qu_any qu_proc_cto_double (double v)
{
	qu_any x = qu_float_newdx (v);
	if (x == 0)
		/* let it generate an exception */
		return QuNil;
	return x;
}



#ifdef HAVE_LONG_DOUBLE
qu_any qu_proc_cto_longdouble (long double v)
{
	qu_any x = qu_decimal_newldx (v);
	if (x == 0)
		/* let it generate an exception */
		return QuNil;
	return x;
}
#endif



#if SIZEOF_SIZE_T == SIZEOF_INT
#define qu_proc_cto_ssize_t           qu_proc_cto_int
#define qu_proc_cto_size_t            qu_proc_cto_uint
#elif SIZEOF_SIZE_T == SIZEOF_LONG
#define qu_proc_cto_ssize_t           qu_proc_cto_long
#define qu_proc_cto_size_t            qu_proc_cto_ulong
#else
#define qu_proc_cto_ssize_t           qu_proc_cto_longlong
#define qu_proc_cto_size_t            qu_proc_cto_ulonglong
#endif

#if SIZEOF_OFF_T == SIZEOF_INT
#define qu_proc_cto_off_t             qu_proc_cto_int
#elif SIZEOF_OFF_T == SIZEOF_LONG
#define qu_proc_cto_off_t             qu_proc_cto_long
#else
#define qu_proc_cto_off_t             qu_proc_cto_longlong
#endif


qu_null qu_proc_cto_Null (void *v)
{
	return QuNil;
}


qu_boolean qu_proc_cto_Boolean (int v)
{
	return v ? QuTrue : QuFalse;
}


#define qu_proc_cto_Int               qu_integer_newl
#define qu_proc_cto_Pointer           qu_pointer_new


qu_string qu_proc_cto_String (char *v)
{
	if (v)
		return qu_string_newz (v);
	return QuStringEMPTY;
}


#define qu_proc_cto_Array           qu_proc_cto_Any
#define qu_proc_cto_Dict            qu_proc_cto_Any
#define qu_proc_cto_Set             qu_proc_cto_Any
#define qu_proc_cto_Range           qu_proc_cto_Any
#define qu_proc_cto_Class           qu_proc_cto_Any
#define qu_proc_cto_Queue           qu_proc_cto_Any
#define qu_proc_cto_Stream          qu_proc_cto_Any
#define qu_proc_cto_Proc            qu_proc_cto_Any


#ifdef QU_SUPPORT_GLIB
qu_any qu_proc_cto_GObject (void *v)
{
	if (v && __QuGObjectNew)
	{
		qu_any x = __QuGObjectNew (v);
		if (x)
			return x;
	}
	/* let it generate an exception */
	return QuNil;
}
#endif


qu_any qu_proc_cto_Any (void *v)
{
	if (v)
		return v;
	return QuNil;
}



/*) class Proc __final
	The Proc class provides a means to wrap Qu functions and enable them
	to be called by C functions. This is useful when you have to pass a
	C function as an argument to another C function, for example,
	event handlers in a GUI application.

	Currently only Qu functions (non-method) are supported and they must
	have a fixed number of typed arguments and typed return value.
	The type must be a Class or a standard validator (_int_, _float_, etc.).

	See the -{Pointer} class to wrap C pointers.

	@warn
	Be careful. Pass the wrong types and your program may crash.

	@p
	See #{new} for details.

	@xmp
		# C library /foo/bar.so

			typedef char *(*foo_t) (int a, float b, char *c, double d);

			static foo_t foo;

			void SetIt (foo_t f)
			{
				foo = f;
			}

			void CallIt ()
			{
				char *r;
				printf ("C calling Qu function\n");
				r = foo (123, 0.1, "Hi there!", 4.56);
				printf ("C got: %s\n", r);
			}


		# Qu program

			sub foo (a:_int_, b:_float_, c:_string_, d:_double_) -> _string_
				println ('a = ', a)
				println ('b = ', b)
				println ('c = ', c)
				println ('d = ', d)
				return 'Hello from Qu'
			;;

			f = Proc (@foo)
			l = Dlib ('/foo/bar.so')
			l.call ("SetIt" , '-:P', f)
			l.call ("CallIt", '-:')

	@out
		C calling Qu function
		a = 123
		b = 0.10000000149011612
		c = Hi there!
		d = 4.5599999999999996
		C got: Hello from Qu
*/
qu_class QuProcClass;


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (u: Sub, iself = -1, safe = false) -> Proc
	Create a Proc object from a Sub object. A Proc object is a wrapper around
	the Sub which makes it possible for C code to call the Sub directly.

	This is handy if you need to pass a C function to another C function that
	needs to call a handler, such as signal handlers in a GUI application.

	v{u} must be a static Sub object. It can not be a class nor instance
	method. The reason for this is because there is no way for the C caller
	to know what self should be. Thus in the Qu handler, self can be
	anything.

	For convenience, you can tell Proc to set self from the v{iself}th
	argument passed to the Sub. First argument index is 0. A value of -1 means
	to use the last argument or, in the case the Sub does not have any argument,
	set self to nil.

	The Qu handler is assumed to have knowledge whether or not self means
	anything.

	v{u} must also be fully typed, that is, all of its arguments and its
	returned value must be declared with types as expected by the C caller.
	It must not have more than 128 arguments.

	See -{Struct.spec} for supported types.

	Note however that some types may not be supported depending on what is
	available on your system. In particular, the C long long and long double
	types are not supported if used as arguments.

	Here is an example of the corresponding C and Qu functions:
	@prog
		C:  double func (void *a, long b, char *c, float d)
		Qu: sub    func (a:_ptr_, b:_long_, c:_string_, d:_float_) -> _double_

	@p
	Note that for C{char *} it is an error for the C caller to pass a NULL
	pointer. If the C caller can pass a NULL pointer for C{char *} then you
	must use Pointer instead.

	If the Qu handler throws an exception then a "zero" return value is assumed.
	Zero here means 0, NULL, 0.0, etc. This might be handy if you specify a
	pointer return value such as _string_ and want to return NULL.

	@warn
	Qu does not have (yet) have a mechanism to allow external thread to enter
	the interpreter. You must make sure that the C function calling the Proc
	is a Qu thread otherwise you'll be in chaos. This is how the flow must go:

	@ul
	. Qu program calls a C function
	. C function calls a Proc
	. Proc executed
	. Thread goes back to C function
	. Thread goes back to Qu program

	@p
	If you are not sure then you can set v{safe} to true to have Proc check
	for this. This requires an extra call to an internal Qu function to make
	sure that the thread is a Qu thread. If the thread is an external thread
	then a warning will be printed to v{stderr} and 0 returned.

	Note that currently it is not possible to pass/get NaN and infinities
	because Qu does not use floating point values to store infinities.
	Qu uses -{__infp__} and -{__infn__}, and NaN in Qu is an error.

	If the C function passes a NaN or infinity, then it is translated into
	nil which will generate an exception thus the Qu function will never
	be called.

	@warn
	If you receive Qu objects, make sure they are Qu objects created under your
	application. Qu objects created by another Qu application is not acceptable
	because they live in a different memory space. Well, except -{Int} of course.

	You must make sure that the Proc object and the associated Qu object, if
	any, lives as long as the C function needs to call it.

	Here is a possible example of a class in a GUI application.

	@xmp
		use Gui

		class Foo

			local instances   = {}
			local bar_handler = Proc (@bar, -1)
			local baz_handler = Proc (@baz, -1)
			local destroyer   = Proc (@destroy, 0)

			sub __init (a, b, c)
				...
				...
				Gui.connect (self, "bar_signal" , bar_handler)
				Gui.connect (self, "baz_signal" , baz_handler)
				Gui.connect (self, "destroy"    , destroyer)
				...
				...
				instances <<< self
			;;

			sub local destroy (inst)
				instances.delete (inst)
			;;

			sub local bar (s:_string_, theself:_any_) -> _int_
				...
			;;

			sub local baz (a: _int_, b:_double, theself:_any_) -> _string_
				...
			;;
		;;

	@p
	If you ever need to free memory of a pointer passed by a C function then
	that library should provide you with a way to do so. If it tells you to
	do it with C's f{free} function then just use -{Pointer.sysfree}.

	If a library requires you to allocate memory with C's f{malloc} function
	then -{Pointer.new} provides you with just that.

	As a final note, -{Dlib.errno} lets you set and get the value of v{errno}.

	@xmp
		sub local handler (i: _int_, p:_ptr_) -> _int_
			...
			... do something with the pointer
			...
			p.sysfree
		;;

	@exc
	.ESecurity
	.EArgValue
	.EIndex    self index out of range
*/
static qu_proc _proc_new_throw (const char *msg, qu_any type)
{
	qu_string s;

	if (qu_sub_is (type))
		s = ((qu_sub) type)->name;
	else if (qu_class_is (type))
		s = ((qu_class) type)->name;
	else
		s = qu_sys_string (type);
	QU_ARG_THROWF ("%s %s is not supported", msg, s->str);
}



static qu_proc qu__proc_new (qu_class self, int argc, qu_any *argv)
{
	qu_sub    u;
	int       selfpos = -1;
	int       safe = 0;
	int       flag;
	int       idx  [QU_PROC_CALL_MAXARGS];
	char      type [QU_PROC_CALL_MAXARGS];
	jit_insn *code, *max, *j;
	char      ret;
	qu_proc   p;
	int       n, i;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, u, qu_sub);
	QU_ARG_CHECKZ (qu_sub_is_static (u), "must be a static sub");

	if (argc > 1)
	{
		QU_ARG_vint (1, selfpos);
		if (argc > 2)
			QU_ARG_vbool (2, safe);
	}

	n = u->argc;
	if (qu_sub_is_argmin (u)
			|| qu_sub_is_vararg (u)
			|| u->ltype == NULL
			|| u->type == NULL
			|| n < 0)
		QU_ARG_THROWZ ("must have fixed typed arguments and return value");

	if (n > QU_PROC_CALL_MAXARGS)
		QU_ARG_THROWZ ("too many arguments");

	if (selfpos >= 0 && selfpos >= n)
		QU_THROWZ (QuEIndex, "self index out of range");

	ret = qu_struct_type_to_char (u->type);
	if (!qu_proc_cando_type (ret))
		return _proc_new_throw ("return value", u->type);

	for (i = 0; i < n; i++)
	{
		qu_any x = u->ltype [i];
		type[i] = qu_struct_type_to_char (x);
		if (!qu_proc_cando_type (type[i]))
			return _proc_new_throw ("argument", x);
	}

	/* FIXME - get a better size? */
	code = qu_mem_alloc (128 + n * 32);
	max  = code + 128 * n * 32 - 1;

	/* JIT start */
	jit_set_ip (code);
	jit_prolog (n);

	/* save stack pointer - we restore it after the call */
	jit_movr_p (JIT_V2, JIT_SP);

	/* get the argument indexes */
	for (i = 0; i < n; i++)
	{
		switch (type[i])
		{
			case QU_STRUCT_ID_char:       idx[i] = jit_arg_c  (); break;
			case QU_STRUCT_ID_byte:       idx[i] = jit_arg_c  (); break;
			case QU_STRUCT_ID_ubyte:      idx[i] = jit_arg_uc (); break;
			case QU_STRUCT_ID_short:      idx[i] = jit_arg_s  (); break;
			case QU_STRUCT_ID_ushort:     idx[i] = jit_arg_us (); break;
			case QU_STRUCT_ID_int:        idx[i] = jit_arg_i  (); break;
			case QU_STRUCT_ID_uint:       idx[i] = jit_arg_ui (); break;
			case QU_STRUCT_ID_int16:      idx[i] = jit_arg_i  (); break;
			case QU_STRUCT_ID_uint16:     idx[i] = jit_arg_ui (); break;
			case QU_STRUCT_ID_int32:      idx[i] = jit_arg_i  (); break;
			case QU_STRUCT_ID_uint32:     idx[i] = jit_arg_ui (); break;
			case QU_STRUCT_ID_long:       idx[i] = jit_arg_l  (); break;
			case QU_STRUCT_ID_ulong:      idx[i] = jit_arg_ul (); break;
			case QU_STRUCT_ID_longlong:
			case QU_STRUCT_ID_ulonglong:  QU_FATAL ("never");
			case QU_STRUCT_ID_float:      idx[i] = jit_arg_f (); break;
			case QU_STRUCT_ID_double:     idx[i] = jit_arg_d (); break;
			case QU_STRUCT_ID_longdouble: QU_FATAL ("never");
#if SIZEOF_SIZE_T == SIZEOF_INT
			case QU_STRUCT_ID_ssize_t:    idx[i] = jit_arg_i  (); break;
			case QU_STRUCT_ID_size_t:     idx[i] = jit_arg_ui (); break;
#elif SIZEOF_SIZE_T == SIZEOF_LONG
			case QU_STRUCT_ID_ssize_t:    idx[i] = jit_arg_l  (); break;
			case QU_STRUCT_ID_size_t:     idx[i] = jit_arg_ul (); break;
#endif
#if SIZEOF_OFF_T == SIZEOF_INT
			case QU_STRUCT_ID_off_t:      idx[i] = jit_arg_i  (); break;
#elif SIZEOF_OFF_T == SIZEOF_LONG
			case QU_STRUCT_ID_off_t:      idx[i] = jit_arg_l  (); break;
#endif
			case QU_STRUCT_ID_Boolean:    idx[i] = jit_arg_i  (); break;
			case QU_STRUCT_ID_Int:        idx[i] = jit_arg_l  (); break;

			case QU_STRUCT_ID_char_p:
			case QU_STRUCT_ID_void_p:
			case QU_STRUCT_ID_Pointer:
			case QU_STRUCT_ID_String:
			case QU_STRUCT_ID_Array:
			case QU_STRUCT_ID_Dict:
			case QU_STRUCT_ID_Set:
			case QU_STRUCT_ID_Range:
			case QU_STRUCT_ID_Class:
			case QU_STRUCT_ID_Queue:
			case QU_STRUCT_ID_Stream:
			case QU_STRUCT_ID_Proc:
#ifdef QU_SUPPORT_GLIB
			case QU_STRUCT_ID_GObject:
#endif
			case QU_STRUCT_ID_Any:        idx[i] = jit_arg_p (); break;
			default:                      QU_FATAL ("never");
		}
	}

	/*
		push the arguments on the stack
		convert them into the proper Qu objects
	*/
	for (i = n - 1; i >= 0; i--)
	{
		int k = idx[i];

		switch (type[i])
		{
			case QU_STRUCT_ID_char:
				jit_getarg_c   (JIT_R0, k);
				jit_prepare    (1);
				jit_pusharg_p  (JIT_R0);
				jit_finish     (qu_proc_cto_char);
				jit_pushr_p    (JIT_RET);
				break;

			case QU_STRUCT_ID_byte:
				jit_getarg_i   (JIT_R0, k);
				jit_prepare    (1);
				jit_pusharg_i  (JIT_R0);
				jit_finish     (qu_proc_cto_byte);
				jit_pushr_p    (JIT_RET);
				break;

			case QU_STRUCT_ID_ubyte:
				jit_getarg_i   (JIT_R0, k);
				jit_prepare    (1);
				jit_pusharg_i  (JIT_R0);
				jit_finish     (qu_proc_cto_ubyte);
				jit_pushr_p    (JIT_RET);
				break;

			case QU_STRUCT_ID_short:
				jit_prepare   (1);
				jit_getarg_i  (JIT_R0, k);
				jit_pusharg_s (JIT_R0);
				jit_finish    (qu_proc_cto_short);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_ushort:
				jit_prepare    (1);
				jit_getarg_us  (JIT_R0, k);
				jit_pusharg_us (JIT_R0);
				jit_finish     (qu_proc_cto_ushort);
				jit_pushr_p    (JIT_RET);
				break;

			case QU_STRUCT_ID_int:
				jit_prepare   (1);
				jit_getarg_i  (JIT_R0, k);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_int);
				jit_pushr_i   (JIT_RET);
				break;

			case QU_STRUCT_ID_uint:
				jit_prepare    (1);
				jit_getarg_ui  (JIT_R0, k);
				jit_pusharg_ui (JIT_R0);
				jit_finish     (qu_proc_cto_uint);
				jit_pushr_ui   (JIT_RET);
				break;

			case QU_STRUCT_ID_int16:
				jit_prepare   (1);
				jit_getarg_i  (JIT_R0, k);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_int16);
				jit_pushr_i   (JIT_RET);
				break;

			case QU_STRUCT_ID_uint16:
				jit_prepare   (1);
				jit_getarg_i  (JIT_R0, k);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_uint16);
				jit_pushr_ui  (JIT_RET);
				break;

			case QU_STRUCT_ID_int32:
				jit_prepare   (1);
				jit_getarg_i  (JIT_R0, k);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_int32);
				jit_pushr_l   (JIT_RET);
				break;

			case QU_STRUCT_ID_uint32:
				jit_prepare   (1);
				jit_getarg_i  (JIT_R0, k);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_uint32);
				jit_pushr_ul  (JIT_RET);
				break;

			case QU_STRUCT_ID_long:
				jit_prepare   (1);
				jit_getarg_l  (JIT_R0, k);
				jit_pusharg_l (JIT_R0);
				jit_finish    (qu_proc_cto_long);
				jit_pushr_l   (JIT_RET);
				break;

			case QU_STRUCT_ID_ulong:
				jit_prepare    (1);
				jit_getarg_ul  (JIT_R0, k);
				jit_pusharg_ul (JIT_R0);
				jit_finish     (qu_proc_cto_ulong);
				jit_pushr_ui   (JIT_RET);
				break;

			case QU_STRUCT_ID_longlong:
			case QU_STRUCT_ID_ulonglong:
				QU_FATAL ("never");

			case QU_STRUCT_ID_float:
				jit_getarg_f  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_f (JIT_R0);
				jit_finish    (qu_proc_cto_float);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_double:
				jit_getarg_d  (JIT_FPR0, k);
				jit_prepare   (1);
				jit_pusharg_d (JIT_FPR0);
				jit_finish    (qu_proc_cto_double);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_longdouble:
				QU_FATAL ("never");

#if SIZEOF_SIZE_T == SIZEOF_INT
			case QU_STRUCT_ID_ssize_t:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_ssize_t);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_size_t:
				jit_getarg_p   (JIT_R0, k);
				jit_prepare    (1);
				jit_pusharg_ui (JIT_R0);
				jit_finish     (qu_proc_cto_size_t);
				jit_pushr_p    (JIT_RET);
				break;

#elif SIZEOF_SIZE_T == SIZEOF_LONG
			case QU_STRUCT_ID_ssize_t:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_l (JIT_R0);
				jit_finish    (qu_proc_cto_ssize_t);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_size_t:
				jit_getarg_p   (JIT_R0, k);
				jit_prepare    (1);
				jit_pusharg_ul (JIT_R0);
				jit_finish     (qu_proc_cto_size_t);
				jit_pushr_p    (JIT_RET);
				break;
#else
			case QU_STRUCT_ID_ssize_t:
			case QU_STRUCT_ID_size_t:
				QU_FATAL ("never");
#endif

#if SIZEOF_OFF_T == SIZEOF_INT
			case QU_STRUCT_ID_off_t:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_i (JIT_R0);
				jit_finish    (qu_proc_cto_off_t);
				jit_pushr_p   (JIT_RET);
				break;
#elif SIZEOF_OFF_T == SIZEOF_LONG
			case QU_STRUCT_ID_off_t:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_l (JIT_R0);
				jit_finish    (qu_proc_cto_off_t);
				jit_pushr_p   (JIT_RET);
				break;
#else
			case QU_STRUCT_ID_off_t:
				QU_FATAL ("never");
#endif

			case QU_STRUCT_ID_Null:
				jit_getarg_i  (JIT_R0, k);
				jit_movi_p    (JIT_R1, QuNil);
				jit_pushr_p   (JIT_R1);
				break;

			case QU_STRUCT_ID_Boolean:
				jit_getarg_i  (JIT_R0, k);
				jit_movi_p    (JIT_R1, QuTrue);
			j = jit_bnei_i    (jit_forward (), JIT_R0, 0);
				jit_movi_p    (JIT_R1, QuFalse);
				jit_patch     (j);
				jit_pushr_p   (JIT_R1);
				break;

			case QU_STRUCT_ID_Int:
				jit_prepare    (1);
				jit_getarg_l   (JIT_R0, k);
				jit_pusharg_l  (JIT_R0);
				jit_finish     (qu_proc_cto_Int);
				jit_pushr_p    (JIT_RET);
				break;

			case QU_STRUCT_ID_void_p:
			case QU_STRUCT_ID_Pointer:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_p (JIT_R0);
				jit_finish    (qu_proc_cto_Pointer);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_char_p:
			case QU_STRUCT_ID_String:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_p (JIT_R0);
				jit_finish    (qu_proc_cto_String);
				jit_pushr_p   (JIT_RET);
				break;

			case QU_STRUCT_ID_Array:
			case QU_STRUCT_ID_Dict:
			case QU_STRUCT_ID_Set:
			case QU_STRUCT_ID_Range:
			case QU_STRUCT_ID_Class:
			case QU_STRUCT_ID_Queue:
			case QU_STRUCT_ID_Stream:
			case QU_STRUCT_ID_Proc:
			case QU_STRUCT_ID_Any:
				jit_getarg_p   (JIT_R0, k);
			j = jit_bnei_p     (jit_forward (), JIT_R0, 0);
				jit_movi_p     (JIT_R0, QuNil); /* let it generate an exception */
				jit_patch      (j);
				jit_pusharg_p  (JIT_R0);
				break;

#ifdef QU_SUPPORT_GLIB
			case QU_STRUCT_ID_GObject:
				jit_getarg_p  (JIT_R0, k);
				jit_prepare   (1);
				jit_pusharg_p (JIT_R0);
				jit_finish    (qu_proc_cto_GObject);
				jit_pushr_p   (JIT_RET);
				break;
#endif

			default:
				QU_FATAL ("never");
		}
	}

	/*
		call the proper wrapper
	*/
	flag = 0;
	if (safe)
		flag |= QU_PROC_CALL_FLAG_SAFE;
	if (selfpos < 0)
		flag |= QU_PROC_CALL_FLAG_LAST;
	else
		flag |= selfpos;

	jit_prepare   (4);
	jit_pusharg_p (JIT_SP);
	jit_movi_i    (JIT_R0, n);
	jit_pusharg_i (JIT_R0);
	jit_movi_i    (JIT_R0, flag);
	jit_pusharg_i (JIT_R0);
	jit_movi_p    (JIT_R0, u);
	jit_pusharg_p (JIT_R0);

#define c_(type) case QU_STRUCT_ID_##type: jit_finish (qu_proc_call_##type); break;
	switch (ret)
	{
		case '-':
		c_(char)
		c_(byte)
		c_(ubyte)
		c_(short)
		c_(ushort)
		c_(int)
		c_(uint)
		c_(int16)
		c_(uint16)
		c_(int32)
		c_(uint32)
		c_(long)
		c_(ulong)
		c_(longlong)
		c_(ulonglong)
		c_(float)
		c_(double)
		c_(longdouble)
		c_(ssize_t)
		c_(size_t)
		c_(off_t)
		c_(char_p)
		c_(void_p)
		c_(Null)
		c_(Boolean)
		c_(Int)
		c_(Pointer)
		c_(String)
		c_(Array)
		c_(Dict)
		c_(Set)
		c_(Range)
		c_(Class)
		c_(Queue)
		c_(Stream)
		c_(Proc)
		c_(Any)
#ifdef QU_SUPPORT_GLIB
		c_(GObject)
#endif
		default : QU_FATAL ("never");
	}
#undef c_

	/* restore the stack */
	jit_movr_p  (JIT_SP, JIT_V2);

	/* and return */
	jit_ret ();

	/* make sure we did allocate enough space */
	if (jit_get_label () > max)
		QU_FATAL ("must fix Proc_new size allocated too small");

#ifdef QU_DEBUG_LOCAL
	fprintf (stderr, "Proc %s size = %d\n", u->name->str, jit_get_label () - code);
#endif

	jit_flush_code (code, jit_get_ip().ptr);

	p       = qu_gc_new (QuProcClass);
	p->code = (char *) code;
	p->sub  = u;

	return p;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__proc__destroy (qu_proc self)
{
	qu_mem_free (self->code);
}



static void qu__proc__mark (qu_proc self)
{
	qu_gc_mark (self->sub);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __like (x) -> Boolean
	If v{x} is a Proc then return true if both point to the same Sub,
	false otherwise.

	@noexc
*/
static qu_boolean qu__proc__like (qu_proc self, qu_any x)
{
	return (qu_proc_is (x) && self->sub == ((qu_proc) x)->sub) ? QuTrue : QuFalse;
}



/*) sub __print -> Null
	Print whatever is returned by the __string method to v{stdout}.

	See -{print} for possible exceptions.
*/
static qu_null qu__proc__print (qu_proc self)
{
	qu_string s = qu_proc_as_string (self);
	return qu_sys_printz (s->str, s->len);
}



/*) sub __string -> String
	Return a string representation of self including the name of the
	Sub object.

	@noexc
*/
qu_string qu__proc__string (qu_proc self)
{
	return qu_string_cat (qu_string_byte ('<'), QuProcClass->name, qu_string_byte (' '),
		qu_sub_as_string (self->sub), qu_string_byte ('>'), NULL);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub subof -> Sub
	Return the corresponding Sub object.

	@noexc
*/
static qu_sub qu__proc_subof (qu_proc self)
{
	return self->sub;
}



/*----------------------------------------------------------------------------
	for GTK
----------------------------------------------------------------------------*/

qu_proc qu_proc_glib_new (qu_any sub)
{
	return qu__proc_new (QuProcClass, 1, &sub);
}


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_proc__setup ()
{
	qu_class_setup_final (QuProcClass, qu_proc_s, qu__proc__destroy, qu__proc__mark);
}



void qu_proc__init ()
{
	/* standard class methods */
	qu_class_method_class (QuProcClass, QU_NAME_new, qu__proc_new, -1);

	/* standard instance methods */
	qu_class_method_inst  (QuProcClass, QU_NAME__like  , qu__proc__like  ,  1);
	qu_class_method_inst  (QuProcClass, QU_NAME__print , qu__proc__print ,  0);
	qu_class_method_inst  (QuProcClass, QU_NAME__string, qu__proc__string,  0);

	/* specific instance methods */
	qu_class_method_instz (QuProcClass, "subof", qu__proc_subof, 0);

	/* finalize declaration */
	qu_class_enddecl (QuProcClass);
}

#endif /*LIGHTNING_CPU_NONE*/
