/* Queue.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Queue is Container __final
	A Queue object is a container which stores objects in a queue.

	It is more efficient than using an Array because Queue objects free and
	allocate memory for each item popped and pushed and the size of memory
	to allocate for each item is well suited for Qu's memory allocator.

	If you pop the first item from a huge Array, the process is slow since
	all the rest of the items must be shifted. Queues does not suffer from
	this. Popping from a Queue is atomic and is very fast.

	Queues are not designed for positional queries thus the method __getitem
	and __getslice are not supported. If you need it then you will have to
	device your own functions, which, is a very bad idea.

	Queues are particularly useful when the objects are short term objects,
	such as clients in a server application.

	@xmp
		q = Queue ("F", 100)
		q <<< 10 <<< 20 <<< 30

		loop
			try
				q >>> x
				println (x)
			catch
				println (__err__)
				break
			;;
		;;

	@out
		10
		20
		30
		(ENone) no result
*/
qu_class QuQueueClass;


/* to reduce polution */
#define item_s qu_queueitem_s



/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

static qu_queue _queue_alloc (qu_class self, long flags, long size)
{
	qu_queue q;

	assert (size <= QU_GC_MAX_ITEMS);
	assert (self->size >= QuQueueClass->size);

	q         = qu_gc__new (self, self->size);
	q->flags |= flags;
	q->len    = 0;
	q->size   = (size < 1) ? QU_GC_MAX_ITEMS : size;
	q->first  = q->last = 0;

	QU_CONTAINER_INIT (q);
	return q;
}



qu_queue qu_queue_new (long flags, long size)
{
	return _queue_alloc (QuQueueClass, flags, size);
}



/*----------------------------------------------------------------------------
	stringify
----------------------------------------------------------------------------*/

/*
	Put what should be printed in v{buf} which must be QUEUE_STRING_BUFSIZE.
*/
#define QUEUE_STRING_BUFSIZE 128
static long _queue_sprint (qu_queue self, char *buf)
{
	snprintf (buf, QUEUE_STRING_BUFSIZE, "<%s %s %ld/%ld>",
		self->klass->name->str,
		qu_queue_is_fifo (self) ? "FIFO" : "LIFO",
		self->len,
		self->size);

	buf [QUEUE_STRING_BUFSIZE - 1] = '\0';
	return strlen (buf);
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (type: String, size = -1) -> Queue
	Create a new Queue Object.

	v{size} is the maximum capacity, a non-positive value means to use the
	system hard limit.

	v{type} is one of:
	@ops
	.| C{"F"} | FIFO (First In First Out)
	.| C{"L"} | LIFO (Last In Lirst Out)

	@p
	Note that the maximum capacity has a system hard limit set by Qu as reasonable
	as possible depending on your system. This limit is huge and is unlikely to be
	reached since you would most probably run out of memory before this happens.

	The capacity may or may not be exact. If you use the #{putback} method then no
	matter what the capacity is, the value is pushed back into the Queue, provided
	of course, the hard limit is not reached.

	@exc
	.EArgValue
*/
static qu_queue qu__queue__inst (qu_class self, int argc, qu_any *argv)
{
	long      flags;
	qu_string stype;
	long      size = -1;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, stype, qu_string);

	switch (stype->str[0])
	{
		case 'F': case 'f': flags = QU_QUEUE_FLAG_FIFO; break;
		case 'L': case 'l': flags = QU_QUEUE_FLAG_LIFO; break;
		default : QU_ARG_THROW;
	}

	if (argc > 1)
		QU_ARG_vlong (1, size);

	return _queue_alloc (self, flags, size);
}



/*) sub class new (type: String, size = -1) -> Queue
	Create a new Queue Object.

	This calls #{__inst} and f{__init} if one is defined.

	@exc
	.EArgValue

	@xmp
		q = Queue ("F", 100)
*/
static qu_queue qu__queue_new (qu_class self, int argc, qu_any *argv)
{
	qu_queue q = qu__queue__inst (self, argc, argv);
	return qu_class_init_new (self, q, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__queue__destroy (qu_queue self)
{
	register item_s *i, *x;

	for (i = self->first; i; i = x)
	{
		x = i->next;
		qu_mem_free (i);
	}
}



static void qu__queue__mark (qu_queue self)
{
	register item_s *i;

	for (i = self->first; i; i = i->next)
		qu_gc_mark (i->obj);

	QU_CONTAINER_MARK (self);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be a one of the following -{Class} objects:
	@ops
	.| -{Array} |
		Return an Array containing all items in the order they are popped.
		This is an atomic operation and the Array is guaranteed to contain
		all items in the Queue at the time this method is called.

	.| -{String}  | Delegate to #{__string}.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{Queue}   | Return self.

	@exc
	.EArgValue

	@xmp
		q = Queue ("L", 100)
		q <<< 1 <<< 2 <<< 3
		a = q as Array
		println (a)

	@out
		[3, 2, 1]
*/
qu_array queue_as_array (qu_queue self)
{
	qu_array  a = qu_array_new (self->len);
	qu_any   *p = a->objs;
	item_s   *i;

	for (i = self->first; i; i = i->next)
		*p++ = i->obj;

	a->len = self->len;
	return a;
}



static qu_any qu__queue__as (qu_queue self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return queue_as_array      (self);
		case QU_CLASS_ID_String:  return qu_queue_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_queue_as_boolean (self);
		case QU_CLASS_ID_Queue:   return self;
	}
	QU_ARG_THROW;
}



/*) sub __clear -> self
	Remove everything from the Queue.

	Note that this is an atomic operation.

	@noexc
*/
qu_queue qu__queue__clear (qu_queue self)
{
	qu__queue__destroy (self);
	self->len = 0;
	self->first = self->last = 0;
	return self;
}



/*) sub __copy -> Set
	Make a shallow copy of self.

	The new Queue will have the same #{type} and #{size} as self.

	Note that this is an atomic operation.
	The new Queue is guaranteed to contain all items in the Queue at the
	time this method is called.

	@noexc
*/
qu_queue qu__queue__copy (qu_queue self)
{
	qu_queue  q;
	item_s   *i, *j;
	int       k = 0;

	q = _queue_alloc (self->klass, qu_queue_is_fifo (self) ? QU_QUEUE_FLAG_FIFO : QU_QUEUE_FLAG_LIFO, self->size);
	for (j = self->first; j; j = j->next)
	{
		i = qu_mem_alloc (sizeof (item_s));
		i->obj = j->obj;
		if (k == 0)
		{
			q->first = q->last = i;
			k = 1;
		}
		else
		{
			q->last->next = i;
			q->last = i;
			i->next = 0;
		}
	}
	q->len = self->len;
	return q;
}



/*) sub __each -> Iterator
	Return an -{Iterator} object to iterate through the items.

	Note that this is a very slow and non-atomic operation.
	If another thread is modifying the Queue then not all items are yielded.
	It is even possible that the same item is yielded more than once.

	Queue objects do not keep the items in an array thus it is not possible
	to keep a positional state. Each iteration requires counting from the
	beginning.

	You would normally use this method for inspection only.

	@noexc

	@xmp
		q = Queue (Queue.FIFO)
		q <<< 10 <<< 20

		for x in q
			println (x)
		;;

	@out
		10
		20

	@p
	The following example demonstrates how you might get all items at any particular
	time.

	@xmp
		atomic
			for x in q
				...
			;;
		;;

		for x in q as Array
			...
		;;
*/
static qu_any qu__queue__next (qu_queue self, qu_iterator i)
{
	long    k = i->ex1;
	long    j;
	item_s *m = self->first;

	QU_ITERATOR_STOP (k < self->len);

	for (j = 0; j < k; j++)
		m = m->next;
	i->ex1++;
	return m->obj;
}


static qu_iterator qu__queue__each (qu_queue self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, qu__queue__next);
	return i;
}



/*) sub __pop -> Any
	Pop item from the Queue.

	If the Queue is FIFO then the first object pushed is returned. Otherwise
	(the Queue is LIFO) the last object pushed is returned.

	Note that this is an atomic operation.

	@exc
	.ENone the queue is empty

	@xmp
		q = Queue (Queue.FIFO)
		q <<< 10 <<< 20

		q >>> x
		println (x)

	@out
		10
*/
qu_any qu__queue__pop (qu_queue self)
{
	item_s *i;
	qu_any  x;

	QU_REQUIRE (QuENone, self->len);
	i = self->first;
	self->first = i->next;
	if (--self->len == 0)
		self->last = self->first = 0;
	x = i->obj;
	qu_mem_free (i);
	return x;
}



/*) sub __print -> Null
	Print Queue to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		q = Queue ("L", 100)
		q <<< 1 <<< 2 <<< 3
		println (q)

	@out
		<Queue LIFO 3/100>
*/
static qu_null qu__queue__print (qu_queue self)
{
	char buf [QUEUE_STRING_BUFSIZE];
	long len = _queue_sprint (self, buf);
	QU_ENSURE (qu_sys_printz (buf, len));
	return QuNil;
}



/*) sub __push (x) -> self
	Push an object into the queue.

	@exc
	.EOverflow

	@xmp
		q = Queue (Queue.FIFO)
		q <<< 10 <<< 20
		println (q.__len)

	@out
		2
*/
qu_queue qu__queue__push (qu_queue self, qu_any x)
{
	QU_REQUIRE (QuEOverflow, self->len < self->size);
	return qu_queue_putback (self, x);
}



/*) sub __string -> String
	Return a String representation of self.

	@exc
	.EMemory result is too big

	@xmp
		q = Queue ('L', 100)
		q <<< 1 <<< 2 <<< 3
		println (q as String)

	@out
		<Queue LIFO 3/100>
*/
qu_string qu__queue__string (qu_queue self)
{
	char buf [QUEUE_STRING_BUFSIZE];
	long len = _queue_sprint (self, buf);

	return qu_string_new (buf, len);
}



/*) sub __true -> Boolean
	Return true if the Queue is not empty, false otherwise.

	@noexc
*/
static qu_boolean qu__queue__true (qu_queue self)
{
	return qu_queue_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods - non-modifying
----------------------------------------------------------------------------*/

/*) sub first -> Any
	Return the first item in the queue without popping it.

	This is the next item that you get when you call #{__pop}.

	@exc
	.ENone queue is empty
*/
static qu_any qu__queue_first (qu_queue self)
{
	QU_REQUIRE (QuENone, self->len);
	return self->first->obj;
}



/*) sub last -> Any
	Return the last item in the queue without popping it.

	This is the item that will be popped last from the queue.

	@exc
	.ENone queue is empty
*/
static qu_any qu__queue_last (qu_queue self)
{
	QU_REQUIRE (QuENone, self->len);
	return self->last->obj;
}



/*) sub size -> Int
	Return the current allocated capacity.

	@noexc
*/
static qu_int qu__queue_size (qu_queue self)
{
	return qu_int_new (self->size);
}



/*) sub type -> String
	Return "F" for FIFO Queue, "L" for LIFO Queue.

	@noexc
*/
static qu_string qu__queue_type (qu_queue self)
{
	return qu_queue_is_fifo (self) ? qu_string_byte ('F') : qu_string_byte ('L');
}



/*----------------------------------------------------------------------------
	specific instance methods - modifying
----------------------------------------------------------------------------*/

/*) sub putback (x) -> self
	Put back an object into the queue. No questions asked.

	This method is useful if you must push back an object for more processing
	such as a multiclient application with fixed threads.

	Like #{__push} and #{__get}, this is an atomic operation.

	@exc
	.EOverflow only if the hard limit is reached
*/
qu_queue qu__queue_putback (qu_queue self, qu_any x)
{
	register item_s *i;

	QU_REQUIRE (QuEOverflow, self->len < QU_GC_MAX_ITEMS);

	i = qu_mem_alloc (sizeof (item_s));
	i->obj = x;
	if (self->len == 0)
	{
		self->first = self->last = i;
		i->next = 0;
	}
	else
	{
		if (qu_queue_is_fifo (self))
		{
			self->last->next = i;
			self->last = i;
			i->next = 0;
		}
		else
		{
			i->next = self->first;
			self->first = i;
		}
	}
	self->len++;
	return self;
}



/*) sub rotate -> self
	Rotate items.

	This is an atomic operation and has the same effect as calling #{__pop}
	and #{__push} the popped item in an k{atomic} block.

	If the queue is empty than nothing happens.

	@noexc

	@xmp
		q = Queue ("L", 100)
		q <<< 1 <<< 2 <<< 3 <<< 4

		print (i) for i in q
		println

		q.rotate

		print (i) for i in q
		println

	@out
		4321
		3214
*/
qu_queue qu__queue_rotate (qu_queue self)
{
	if (self->len)
	{
		register item_s *i;

		i = self->first;
		self->first = i->next;
		self->last->next = i;
		self->last = i;
		i->next = NULL;
	}
	return self;
}


/*----------------------------------------------------------------------------
	These may not be done from Qu programs.
	Must be defined so error message reporting works
----------------------------------------------------------------------------*/

static qu_any qu__queue_fix       (qu_queue self) { QU_THROW (QuEImplement); }
static qu_any qu__queue_fixed     (qu_queue self) { QU_THROW (QuEImplement); }
static qu_any qu__queue_freeze    (qu_queue self) { QU_THROW (QuEImplement); }
static qu_any qu__queue_frozen    (qu_queue self) { QU_THROW (QuEImplement); }

#ifdef QU_CONTAINER_WITH_PROTECT
static qu_any qu__queue_protect   (qu_queue self) { QU_THROW (QuEImplement); }
static qu_any qu__queue_protector (qu_queue self) { QU_THROW (QuEImplement); }
static qu_any qu__queue_unprotect (qu_queue self) { QU_THROW (QuEImplement); }
#endif


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_queue__setup ()
{
	qu_class_setup (QuQueueClass, qu_queue_s, qu__queue__destroy, qu__queue__mark);
}



void qu_queue__init ()
{
	/* standard class methods */
	qu_class_method_class (QuQueueClass, QU_NAME__inst, qu__queue__inst, -1);
	qu_class_method_class (QuQueueClass, QU_NAME_new  , qu__queue_new  , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuQueueClass, QU_NAME__as    , qu__queue__as    , 1);
	qu_class_method_inst  (QuQueueClass, QU_NAME__clear , qu__queue__clear , 0);
	qu_class_method_inst  (QuQueueClass, QU_NAME__copy  , qu__queue__copy  , 0);
	qu_class_alias        (QuQueueClass, QU_NAME__copy  , QU_NAME_copy);
	qu_class_method_inst  (QuQueueClass, QU_NAME__each  , qu__queue__each  , 0);
	qu_class_method_inst  (QuQueueClass, QU_NAME__pop   , qu__queue__pop   , 0);
	qu_class_method_inst  (QuQueueClass, QU_NAME__print , qu__queue__print , 0);
	qu_class_method_inst  (QuQueueClass, QU_NAME__push  , qu__queue__push  , 1);
	qu_class_method_inst  (QuQueueClass, QU_NAME__string, qu__queue__string, 0);
	qu_class_method_inst  (QuQueueClass, QU_NAME__true  , qu__queue__true  , 0);

	/* specific instance methods */
	qu_class_method_instz (QuQueueClass, "first"   , qu__queue_first  , 0);
	qu_class_method_instz (QuQueueClass, "last"    , qu__queue_last   , 0);
	qu_class_method_instz (QuQueueClass, "putback" , qu__queue_putback, 1);
	qu_class_method_instz (QuQueueClass, "rotate"  , qu__queue_rotate , 0);
	qu_class_method_instz (QuQueueClass, "size"    , qu__queue_size   , 0);
	qu_class_method_instz (QuQueueClass, "type"    , qu__queue_type   , 0);

	qu_class_method_instz (QuQueueClass, "fix"     , qu__queue_fix    , 0);
	qu_class_method_instz (QuQueueClass, "fixed"   , qu__queue_fixed  , 0);
	qu_class_method_instz (QuQueueClass, "freeze"  , qu__queue_freeze , 0);
	qu_class_method_instz (QuQueueClass, "frozen"  , qu__queue_frozen , 0);

#ifdef QU_CONTAINER_WITH_PROTECT
	qu_class_method_instz (QuQueueClass, "protect"  , qu__queue_protect  , 0);
	qu_class_method_instz (QuQueueClass, "protector", qu__queue_protector, 0);
	qu_class_method_instz (QuQueueClass, "unprotect", qu__queue_unprotect, 0);
#endif

	/* finalize declaration */
	qu_class_set_super (QuQueueClass, QuContainerClass);
	qu_class_enddecl   (QuQueueClass);
}
