/* Range.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Range __final
	A Range object is a range of Int items.
	It is never empty and contains at least one item.

	You create a Range with the syntax or the #{new} method.

	@xmp
		r = 1 .. 10  # same as r = Range (1, 10,  1)
		r = 10 .. 1  # same as r = Range (10, 1, -1)

	@p
	Range objects are opaque objects, except those created by the special method
	#{_local}.

	The equality test of two range objects will return true if the two
	objects iterate through the same items. However, Range objects created
	with the f{_local} method are only equal to themselves.

	@xmp
		a = 1 .. 10
		b = Range (1, 10)
		println (a == b)

	@out
		true

	@warn
	The #{first}, #{last}, and #{len} (the total count of the items)
	must all fall within the -{Int} range.

	@xmp
		try
			r = Range (0, Int.MAX)
		catch
			println (__err__)
		;;

	@out
		(EArgValue) Range too big
*/
qu_class QuRangeClass;


/* predefined 0 .. QU_INT_MAX - 1 */
static qu_range RangeMAX;


/*-----------------------------------------------------------------------
	allocate
-----------------------------------------------------------------------*/

static qu_range _range_alloc (qu_class self, long first, long last, long step, long count)
{
	register qu_range r;

	assert (self->size >= sizeof (qu_range_s));

	r        = qu_gc__new (self, self->size);
	r->first = first;
	r->step  = step;
	r->count = count;
	r->last  = last;
	return r;
}


inline static qu_range _range_copy (qu_range self)
{
	return _range_alloc (self->klass, self->first, self->last, self->step, self->count);
}


qu_range qu_range_to_static (qu_range self)
{
	register qu_range r;

	r        = qu_gc_new_static (QuRangeClass);
	r->first = self->first;
	r->step  = self->step;
	r->count = self->count;
	r->last  = self->last;
	return r;

}



/*-----------------------------------------------------------------------
	stuffs
-----------------------------------------------------------------------*/

/*
	A "local" Range is created via the _local method and can be modified.
	It's purpose is to save memory allocation.
*/
#define QU_RANGE_FLAG_LOCAL     QU_FLAG_0
#define qu_range_is_local(self) qu_flag_test (self, QU_RANGE_FLAG_LOCAL)


/*
	Here, we assume that step is intended.
	First and last may be interchanged based on the step sign.
*/
#define RANGE_ADJUST_KEEP_STEP(first, last, step) \
QU_BEGIN \
	if (((step) < 0 && (first) < (last)) || ((step) > 0 && (first) > (last))) \
	{ \
		long tmp = first; \
		first = last; \
		last = tmp; \
	} \
QU_END


/*
	Here, we assume that first and last are intended.
	The step sign may be adjusted.
*/
#define RANGE_ADJUST_CHANGE_STEP(first, last, step) \
QU_BEGIN \
	if ((first) < (last)) \
	{ \
		if ((step) < 0) \
			step = -(step); \
	} \
	else \
	{ \
		if ((step) > 0) \
			step = -(step); \
	} \
QU_END



/*
	Make sure the range is a good one.
	This takes care of setting the fields correctly.

	Here, we are assuming that step is intended.
	first and last may be interchanged based on the step sign.
*/
static qu_range _range_init (qu_range self, long first, long last, long step)
{
	long f, delta, count;

	assert (step != 0);

	RANGE_ADJUST_KEEP_STEP (first, last, step);

	f = first;
	if (f < 0)
		f = -f;

	QU_ARG_CHECKZ (
		   QU_INT_MIN <= first && first <= QU_INT_MAX
		&& QU_INT_MIN <= last  && last  <= QU_INT_MAX
		&& last < QU_GC_MAX_ITEMS - f, "Range too big");

	delta = (step < 0) ? -step : step;
	count = last - first;
	if (count < 0)
		count = -count;
	count = (count + delta) / delta;

	self->first = first;
	self->step  = step;
	self->count = count;
	self->last  = first + (count - 1) * step;

	return self;
}



/*
	Put what should be printed in 'buf' which must be RANGE_STRING_BUFSIZ big.

	Range objects are rarely printed thus performance is the last thing
	to consider.
*/
#define RANGE_STRING_BUFSIZ 128
static long _range_sprint (qu_range self, char *buf)
{
	long step = self->step;

	if (step < 0)
		step = -step;

	if (step == 1 || step == -1)
		snprintf (buf, RANGE_STRING_BUFSIZ, "<%s %ld..%ld>",
			self->klass->name->str,
			self->first,
			self->last);
	else
		snprintf (buf, RANGE_STRING_BUFSIZ, "<%s %ld..%ld %s %ld>",
			self->klass->name->str,
			self->first,
			self->last,
			(self->step < 0) ? "<<" : ">>",
			step);

	buf [RANGE_STRING_BUFSIZ - 1] = '\0';
	return strlen (buf);
}



/*-----------------------------------------------------------------------
	standard class methods
-----------------------------------------------------------------------*/

/*) sub class new (*arg) -> Range
	@def
	sub class new (first: Integer, last: Integer, step = 1) -> Range
	sub class new (x) -> Range
	@p
	Create a new Range object.

	The first form creates a Range by specifying the v{first} item, the target
	v{last} item and optionally a v{step} value.

	If called with two arguments then step is set to C{1} if C{first <= last}
	and to C{-1} otherwise.

	If called with three arguments, v{step} must be a positive Integer if
	C{first <= last}, negative otherwise.

	v{last} is adjusted so that it reflects the actual last value.

	@exc
	.EArgValue

	@xmp
		r = Range (1, 8, 2)
		println (r)

	@out
		<Range 1..7 >> 2>

	@p
	The second form creates a Range which starts from 0 and stops at the last
	index of the object with a v{step} value of 1,
	thus C{Range (0, x.__len - 1, 1)}.

	The object must support the __len method.

	If it is empty then a special predefined system Range which iterates from
	0 to a maximum hard limit value with a v{step} value of 1 is returned.
	Thus something like C{Range ('')} reveals this limit.

	@exc
	.EArgValue
	.EArgValue
	.EImplement

	@xmp
		a = [1 2 3]
		r = Range (a)
		println (r)

		r = Range ('')
		println (r)

	@out
		<Range 0..2>
		<Range 0..1073741822>
*/
qu_range qu__range_new (qu_class self, int argc, qu_any *argv)
{
	qu_any x;

	QU_ARG_VCOUNT (1, 3);
	x = argv[0];

	if (qu_integer_is (x))
	{
		long first, last, step = 1;

		QU_ARG_ilong (x, first);

		if (argc == 1)
		{
			long count;

			QU_ARG_CHECKZ (first, "single argument integer can not be zero");

			if (first < 0)
			{
				last = 0;
				count = -first + 1;
			}
			else
			{
				last = first;
				first = 0;
				count = last + 1;
			}
			return _range_alloc (self, first, last, step, count);
		}
		else
		{
			QU_ARG_vilong (1, last);
			if (argc > 2)
			{
				QU_ARG_vilong (2, step);
				QU_ARG_CHECKZ (step, "step value can not be zero");
			}
			else
			{
				step = (first <= last) ? 1 : -1;
			}
			return _range_init (qu_gc__new (self, self->size), first, last, step);
		}
	}
	else
	{
		qu_int l = qu_sys_len (argv[0]);
		long   n;

		QU_ENSURE (l);

		n = qu_int_tol (l);
		QU_ARG_CHECKZ (n <= QU_GC_MAX_ITEMS, "Range too big");

		if (n == 0)
			return RangeMAX;
		return _range_alloc (self, 0, n - 1, 1, n);
	}
}



/*-----------------------------------------------------------------------
	specific class methods
-----------------------------------------------------------------------*/

/*) sub class _local (first: Integer, last: Integer, step = 1) -> Range
	Create a new Range object that can be modified later on.

	See #{new} for the arguments.

	Ranges created with this method are not distinguishable from the others
	except to f{_modify} and #{__eq}.

	See #{modify} for an example.

	@exc
	.EArgValue
*/
static qu_range qu__range__local (qu_class self, int argc, qu_any *argv)
{
	qu_range r = qu__range_new (self, argc, argv);

	QU_ENSURE (r);

	/* careful - qu__range_new may return RangeMAX */
	if (r == RangeMAX)
		r = _range_copy (r);

	r->flags |= QU_RANGE_FLAG_LOCAL;
	return r;
}



/*-----------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------*/
/* none */



/*-----------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Array}   | Return an Array containing all items in self.
	.| -{Set}     | Return a Set containing all items in self.
	.| -{String}  | Delegate to #{__string}.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{Range}   | Return self.

	@exc
	.EValue    range too big
	.EArgValue

	@xmp
		a = (1..5) as Array
		println (a)

		sub make (x, *arg)
			for y in arg
				x <<< y for y in y
			;;
			return x
		;;

		s = make ({}, 1..3, 8..10, 15..18)
		println (s)

	@out
		[1, 2, 3, 4, 5]
		{16, 17, 1, 18, 2, 3, 8, 9, 10, 15}

	@warn
	Be careful when you call this method. If you don't have enough memory to
	create the Array or Set then Qu will give you an unrecoverable error.
*/
qu_array qu_range_as_array (qu_range self)
{
	long     l = self->count;
	long     i = self->first;
	long     d = self->step;
	long     n = l;
	qu_array a;
	qu_any  *p;

	QU_MEM_CHECK (l <= QU_GC_MAX_ITEMS);
	a = qu_array_new (l);
	p = a->objs;
	while (n--)
	{
		*p++ = qu_int_new (i);
		i += d;
	}
	a->len = l;
	return a;
}



qu_set qu_range_as_set (qu_range self)
{
	/*
		This is not really atomic.
		We just know it is because qu_set will never have to go to program level
		since all items are qu_int.
	*/

	long   l = self->count;
	long   i = self->first;
	long   d = self->step;
	qu_set s;

	QU_MEM_CHECK (l <= QU_GC_MAX_ITEMS);
	s = qu_set_new (l);
	while (l--)
	{
		qu_set_push (s, qu_int_new (i));
		i += d;
	}
	return s;
}



static qu_any qu__range__as (qu_range self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return qu_range_as_array   (self);
		case QU_CLASS_ID_Set:     return qu_range_as_set     (self);
		case QU_CLASS_ID_String:  return qu_range_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_range_as_boolean (self);
		case QU_CLASS_ID_Range:   return self;
	}
	QU_ARG_THROW;
}



/*) sub __cmp (x) -> Int
	If v{x} is a Range then compare the number of iterated items, otherwise
	delegate to -{Object.__cmp}.

	@noexc

	@xmp
		a = 0 .. 10
		b = 0 .. 11
		println (a <=> b)

	@out
		-1
*/
static qu_int qu__range__cmp (qu_range self, qu_any x)
{
	if (qu_range_is (x))
		return QU_INT_CMP (self->count, ((qu_range) x)->count);
	return qu_object_default_cmp (self, x);
}



/*) sub __each
	Return an -{Iterator} object to iterate through the items.

	Note that the interpreter does not create an Iterator object for the k{for}
	loop syntax unless you specifically call this method via the k{each}
	keyword. It uses an internal iterator which is much more efficient.

	@noexc

	@xmp
		r = 1 .. 4

		for x in r         # interpreter uses an internal iterator
			print (x)
		;;
		println

		for x in each r    # calls r.__each
			print (x)
		;;
		println

	@out
		1234
		1234
*/
static qu_int qu__range__next (qu_range self, qu_iterator i)
{
	long k = i->ex1;
	QU_REQUIRE (QuEStop, k < self->count);
	i->ex1 = k + 1;
	return qu_int_new (self->first + self->step * k);
}



static qu_iterator qu__range__each (qu_range self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, qu__range__next);
	return i;
}



/*) sub __eq (x) -> Boolean
	Return true if v{x} is a Range and iterate through the same values
	as self, false otherwise.

	Note that a Range created with #{_local} is not equal with anything
	except itself.

	@noexc

	@xmp
		a = 1..10
		b = 1..10
		println (a == b)

	@out
		true
*/
static qu_boolean qu__range__eq (qu_range self, qu_any x)
{
	return (qu_range_is (x) && qu_range_eq (self, x)) ? QuTrue : QuFalse;
}



/*) sub __getitem (i: Integer) -> Int
	Get the value at a specified position.

	v{i} must be an integer within the Range.
	A negative value means counting from the end.

	@exc
	.EArgValue
	.EIndex

	@xmp
		r = Range (2, 17, 3)  # 2, 5, 8, 11, 14, 17
		println (r [4])
		println (r [-1])

	@out
		14
		17
*/
static qu_int qu__range__getitem (qu_range self, qu_integer idx)
{
	long i, n = self->count;

	QU_ARG_long (idx, i);
	QU_ENSURE_INDEX (i, n);
	return qu_int_new (self->first + self->step * i);
}



/*) sub __hash -> Int
	Return the hash value of this Range.

	Note that Ranges created with #{_local} have fixed hash value.
	This means that they are not well suited for use as keys in a Dict
	or items in a Set.

	@noexc
*/
static qu_int qu__range__hash (qu_range self)
{
	if (qu_range_is_local (self))
		return qu_int_new (QU_CLASS_ID_Range);
	return qu_int_new (self->first | self->last | self->step);
}



/*) sub __in (x) -> Boolean
	Return true if v{x} is an item in the Range, false otherwise.

	If v{x} is not an integer then false is returned, no exception thrown.

	@noexc

	@xmp
		r = Range (2, 11, 3)  # 2, 5, 8, 11
		println (4 in r)

	@out
		false
*/
qu_boolean qu__range__in (qu_range self, qu_any x)
{
	long k;

	if (qu_int_is (x))
	{
		k = qu_int_tol (x);

	  doint:
		if (self->step > 0)
		{
			if (self->first <= k && k <= self->last)
			{
				if (self->first < 0)
				{
					if (((self->first - k) % self->step) == 0)
						return QuTrue;
				}
				else
				{
					if (((k - self->first) % self->step) == 0)
						return QuTrue;
				}
			}
		}
		else
		{
			if (self->last <= k && k <= self->first)
			{
				if (self->last < 0)
				{
					if (((self->last - k) % self->step) == 0)
						return QuTrue;
				}
				else
				{
					if (((k - self->last) % self->step) == 0)
						return QuTrue;
				}
			}
		}
	}
	else if (qu_long_is (x))
	{
		k = qu_long_tol (x);
		if (errno == 0)
			goto doint;
	}
	return QuFalse;
}



/*) sub __len -> Int
	Return the number of items in the range.

	@noexc

	@xmp
		r = Range (2, 17, 3)  # 2, 5, 8, 11, 14, 17
		println (r.__len)

	@out
		6
*/
static qu_int qu__range__len (qu_range self)
{
	assert (self->count <= QU_INT_MAX);
	return qu_int_new (self->count);
}



/*) sub len -> Int
	This is an alias of #{__len}.
*/



/*) sub __like (x) -> Boolean
	Return true if v{x} is a Range and the length are the same.

	@noexc

	@xmp
		a = 1..10
		b = 2..11
		println (a === b)

	@out
		true
*/
qu_boolean qu__range__like (qu_range self, qu_any x)
{
	return (qu_range_is (x) && self->count == ((qu_range) x)->count)
		? QuTrue : QuFalse;
}



/*) sub __lshift (x: Integer) -> Range
	Return a new Range by modifying the step value of self to v{x}
	so it iterates in a descending direction.

	@exc
	.ELshift

	@xmp
		a = 1..10
		b = a << 2
		println (a)
		println (b)

	@out
		<Range 1..10>
		<Range 10..2 << 2>
*/
static qu_range qu__range__lshift (qu_range self, qu_integer x)
{
	qu_range r;
	long     step = qu_integer_toli (x);

	QU_REQUIRE (QuELshift, errno == 0 && step != 0);
	r = _range_copy (self);
	QU_REQUIRE (QuELshift, _range_init (r, r->first, r->last, -step));
	return r;
}



/*) sub __rshift (x: Integer) -> Range
	Return a new Range by modifying the step value of self to v{x}
	so it iterates in an anscending direction.

	@exc
	.ERshift

	@xmp
		a = 1..10
		b = a >> 2
		println (a)
		println (b)

	@out
		<Range 1..10>
		<Range 1..9 >> 2>
*/
static qu_range qu__range__rshift (qu_range self, qu_integer x)
{
	qu_range r;
	long     step = qu_integer_toli (x);

	QU_REQUIRE (QuERshift, errno == 0 && step != 0);
	r = _range_copy (self);
	QU_REQUIRE (QuERshift, _range_init (r, r->first, r->last, step));
	return r;
}



/*) sub __print -> Null
	Print this Range to v{stdout}.

	See -{print} for possible exceptions.

	@xmp
		r = 1..10
		println (r)

	@out
		<Range 1..10>
*/
static qu_null qu__range__print (qu_range self)
{
	char buf [RANGE_STRING_BUFSIZ];
	long len = _range_sprint (self, buf);
	return qu_sys_printz (buf, len);
}



/*) sub __string -> String
	Return a string representation of this Range.

	@noexc

	@xmp
		r = 1..10
		s = r as String
		println (s)

	@out
		<Range 1..10>
*/
qu_string qu__range__string (qu_range self)
{
	char buf [RANGE_STRING_BUFSIZ];
	long len = _range_sprint (self, buf);
	return qu_string_new (buf, len);
}



/*) sub __true -> Boolean
	Return true if the Range is ascending, false otherwise.

	@noexc

	@xmp
		r = Range (1, 10, -1)
		println (r ? '->' : '<-')

	@out
		->
*/
static qu_boolean qu__range__true (qu_range self)
{
	return qu_range_as_boolean (self);
}



/*-----------------------------------------------------------------------
	specific instance methods - non-modifying
-----------------------------------------------------------------------*/

/*) sub cover (x) -> Range
	Return a new Range to cover v{x} based on this Range.

	v{x} may be an Integer to denote the number of items to cover, or any
	object that supports the __len method.

	@exc
	.EArgValue
	.EImplement
	.ENone       nothing to cover (zero length)

	@xmp
		s = 'abcdefg'

		r = Range (-3, 10, 2)
			# -3 -1 1 3 5 7 9
			#       ^^^^^

		g = r.cover (s)   # same as r.cover (s.len)
		println (g)

	@out
		<Range 1..5 >> 2>
*/
static qu_range qu__range_cover (qu_range self, qu_any x)
{
	long n, i;

	if (qu_integer_is (x))
	{
		QU_ARG_long (x, n);
		QU_ARG_CHECK (n >= 0);
	}
	else
	{
		qu_int l = qu_sys_len (x);

		QU_ENSURE (l);
		n = qu_int_tol (l);
	}
	n = qu_range_cover (self, n, &i);
	QU_REQUIREZ (QuENone, n > 0, "cant cover zero length");
	return qu_range_new (i, i + (n - 1) * self->step, self->step);
}



/*) sub first -> Int
	Return the first item.

	@noexc

	@xmp
		r = 2 .. 7
		println (r.first)

	@out
		2
*/
static qu_int qu__range_first (qu_range self)
{
	return qu_int_new (self->first);
}



/*) sub last -> Int
	Return the last item.

	@noexc

	@xmp
		r = 7 .. 2
		println (r.last)

	@out
		2
*/
static qu_int qu__range_last (qu_range self)
{
	return qu_int_new (self->last);
}



/*) sub reverse -> Range
	Return a new Range which iterates in the reverse direction.

	Same as C{Range (self.last, self.first, self.step)}.

	@noexc

	@xmp
		a = 1..10
		b = a.reverse
		println (b)

	@out
		<Range 10..1>
*/
static qu_range qu__range_reverse (qu_range self)
{
	return qu_range_new (self->last, self->first, self->step);
}



/*) sub step -> Int
	Return the step value in the range.

	@noexc

	@xmp
		r = Range (2, 18, 4)
		println (r.step)

	@out
		4
*/
static qu_int qu__range_step (qu_range self)
{
	return qu_int_new (self->step);
}



/*-----------------------------------------------------------------------
	specific instance methods - modifying
-----------------------------------------------------------------------*/

/*) sub _modify (first: Integer, last = self.last, step = self.step) -> self
	Modify a range created with #{_local}.

	This is an efficient method that might save you some memory.

	@warn
	Use with care. If you are using the Range for something else then
	you should probably not modify it. You would normally only use this
	locally within a function or method.

	@exc
	.EModify
	.EArgValue

	@xmp
		sub setinterval (a, n, *arg)
			range = Range._local (0, a.len - 1)
			a [range._modify (i, i + n - 1)] = 1 for i in arg
		;;

		a = Array.dim (10, 0)
		println (a)

		setinterval (a, 2, 0, 4, 7)
		println (a)

	@out
		[0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
		[1, 1, 0, 0, 1, 1, 0, 1, 1, 0]
*/
static qu_range qu__range__modify (qu_range self, int argc, qu_any *argv)
{
	long first, last = self->last, step = self->step;

	QU_REQUIREZ (QuEModify, qu_range_is_local (self), "not a local Range");

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_vilong (0, first);
	if (argc > 1)
	{
		QU_ARG_vilong (1, last);
		if (argc > 2)
		{
			QU_ARG_vilong (2, step);
			QU_ARG_CHECKZ (step, "step value may not be zero");
		}
	}
	return _range_init (self, first, last, step);
}



/*-----------------------------------------------------------------------
	interface
-----------------------------------------------------------------------*/

/*
	Create a new qu_range - may throw QuEArgValue.
*/
qu_range qu_range_new (long first, long last, long step)
{
	RANGE_ADJUST_CHANGE_STEP (first, last, step);
	{
		qu_any argv [3] = {qu_integer_newl (first), qu_integer_newl (last), qu_integer_newl (step)};
		return qu__range_new (QuRangeClass, 3, argv);
	}
}



/*
	Create a new qu_range - may throw QuEArgValue.
	This is called by qu_sys_range.
*/
qu_range qu_range_new_ex (qu_any x, qu_any y)
{
	long first, last, step = 1;

	QU_ARG_ilong (x, first);
	QU_ARG_ilong (y, last);

	RANGE_ADJUST_CHANGE_STEP (first, last, step);
	{
		qu_any argv [3] = {qu_integer_newl (first), qu_integer_newl (last), qu_integer_newl (step)};
		return qu__range_new (QuRangeClass, 3, argv);
	}
}



/*
	Get starting index and the number of items covered.
*/
long qu_range_cover (qu_range self, long len, long *start)
{
	long i = self->first;
	long t = self->last;
	long d = self->step;
	long n;

	if (d > 0)
	{
		if (t < 0)
			return 0;
		if (i < 0)
			i += d * ((d - i - 1) / d);
		if (i >= len)
			return 0;
		if (t >= --len)
			t = len;
		n = (t - i + d) / d;
	}
	else
	{
		if (t >= len)
			return 0;
		if (i >= len)
			i += d * ((i - len - d) / -d);
		if (i < 0)
			return 0;
		if (t < 0)
			t = 0;
		n = (i - t - d) / -d;
	}
	*start = i;
	return n;
}



int qu_range_contains (qu_range self, qu_any x)
{
	return qu__range__in (self, x) == QuTrue;
}



int qu_range_eq (qu_range self, qu_range x)
{
	if (self == x)
		return 1;

	if (qu_range_is_local (self) || qu_range_is_local (x))
		return 0;

	if (self->first == x->first
			&& self->last  == x->last
			&& self->step  == x->step)
		return 1;

	return 0;
}



/*-----------------------------------------------------------------------
	init
-----------------------------------------------------------------------*/

void qu_range__setup ()
{
	qu_class_setup_final (QuRangeClass, qu_range_s, 0, 0);

	RangeMAX        = qu_gc_new_static (QuRangeClass);
	RangeMAX->first = 0;
	RangeMAX->last  = QU_INT_MAX- 1;
	RangeMAX->step  = 1;
	RangeMAX->count = QU_INT_MAX;
}



void qu_range__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuRangeClass, QU_NAME_new, qu__range_new, -1);

	/* specific class methods */
	qu_class_method_classz (QuRangeClass, "_local", qu__range__local, -1);

	/* standard instance methods */
	qu_class_method_inst   (QuRangeClass, QU_NAME__as     , qu__range__as     , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__cmp    , qu__range__cmp    , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__each   , qu__range__each   , 0);
	qu_class_method_inst   (QuRangeClass, QU_NAME__eq     , qu__range__eq     , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__getitem, qu__range__getitem, 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__hash   , qu__range__hash   , 0);
	qu_class_method_inst   (QuRangeClass, QU_NAME__in     , qu__range__in     , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__len    , qu__range__len    , 0);
	qu_class_alias         (QuRangeClass, QU_NAME__len    , QU_NAME_len);
	qu_class_method_inst   (QuRangeClass, QU_NAME__like   , qu__range__like   , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__lshift , qu__range__lshift , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__print  , qu__range__print  , 0);
	qu_class_method_inst   (QuRangeClass, QU_NAME__rshift , qu__range__rshift , 1);
	qu_class_method_inst   (QuRangeClass, QU_NAME__string , qu__range__string , 0);
	qu_class_method_inst   (QuRangeClass, QU_NAME__true   , qu__range__true   , 0);

	/* specific instance methods - non-modifying */
	qu_class_method_instz  (QuRangeClass, "cover"  , qu__range_cover  , 1);
	qu_class_method_instz  (QuRangeClass, "first"  , qu__range_first  , 0);
	qu_class_method_instz  (QuRangeClass, "last"   , qu__range_last   , 0);
	qu_class_method_instz  (QuRangeClass, "reverse", qu__range_reverse, 0);
	qu_class_method_instz  (QuRangeClass, "step"   , qu__range_step   , 0);

	/* specific instance methods - modifying */
	qu_class_method_instz (QuRangeClass, "_modify", qu__range__modify, -1);

	/* finalize declaration */
	qu_class_enddecl (QuRangeClass);
}
