/* Rational.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Rational is Number __final
	This class provide support for rational numbers.

	Refer to the -{Number} class for information on Qu numerics in general.

	The class hierarchy:
	@tree
	.-{Number}
	.	Rational

	@p
	To create a Rational number you use the #{new} method or the _/ operator.

	Both the numerator and denominator must be -{Integer} objects (-{Int} or
	-{Long}) and the denominator may not be 0.

	Rational numbers are always I{normalized}. The sign is carried by the
	numerator thus the denominator is always a positive integer. If the
	numerator is 0, the denominator is set to Int (1).

	@xmp
		a = Rational (1, 2)
		b = 1 _/ 2
		c = 1
		d = c _/ 2
		println (a)
		println (b)
		println (a == b)
		println (a is Number)
		println (a is Rational)

		a = 0 _/ -8
		println (a)

	@out
		(1 _/ 2)
		(1 _/ 2)
		true
		true
		true
		0 _/ 1

	@warn
	There must be a space between the numerator and the _/ operator otherwise
	Qu will interpret it as c_ / 2.

	@p
	It is not possible to use the complex number syntax with a Rational
	denominator, i.e. C{1 _/ 2j} because Qu will interpret it as C{1 _/ (2j)}
	and give you an error because the denominator is not an Integer.
	Use this syntax instead:
	@prog
		c = Complex (1, 2 _/ 3)
	@p
	or, with the Math module:
	@prog
		use Math
		c = 1 + (2 _/ 3).j
*/
qu_class QuRationalClass;


/* predefined */
qu_rational QuRational0;
qu_rational QuRational1;
qu_rational QuRational_1;


/* make sure we return an Integer if the denominator is 1 */
#define RATIONAL_RESULT(type, x) \
	(qu_rational_iswhole(x) ? (type) ((x)->num) : (type) (x))



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	new
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub class new (*arg) -> Rational
	@def
	sub class new (numerator: Real, denominator: Real) -> Rational
	sub class new (x) -> Rational
	@p
	Create a new Rational number by specifying the numerator and denominator,
	or create one from another object.

	If called with one argument then this method simply calls C{x.__as (Rational)}
	and makes sure that the returned value is a Rational otherwise EImplement
	is thrown.

	If called with two arguments then both arguments must be -{Integer} or
	-{Decimal} objects. The denominator may not be zero.

	Remember that if you use Decimal numbers then the result may not be
	accurate. Although the numbers appear to be "exact" on your screen or
	editor, internally they may not be stored as such.

	@exc
	.EArgValue
	.EImplement
	.ESyntax
	.Exception  anything thrown by the __as method

	@xmp
		class Foo
			sub __as (what)
				switch what
					case Rational: return 3 _/ 4
				;;
				throw EImplement
			;;
		;;

		x = Rational (1, 2)
		println (x)

		x = Foo ()
		println (Rational (x))

		x = Rational (1.3, 2)
		println (x)

		println (1.2 _/ 3.6)
		println (3.3 _/ 2.1)

		x = 1.23456789E-30
		println (x _/ x)

		println (2.3 _/ 4.9)
		println (2.3 _/ Big ('4.9'))

	@out
		(1 _/ 2)
		(3 _/ 4)
		(13 _/ 20)
		(1 _/ 3)
		(11 _/ 7)
		(1 _/ 1)
		(7019042968749999 _/ 14953613281250002)
		(7019042968749999 _/ 14953613281250000)
*/
static qu_rational qu__rational_new (qu_class self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	if (argc == 1)
	{
	  one:
		{
			qu_any x = argv[0];
			switch (qu_classid (x))
			{
				case QU_CLASS_ID_Int:      return qu_int_as_rational     ((qu_int    ) x);
				case QU_CLASS_ID_Long:     return qu_long_as_rational    ((qu_long   ) x);
				case QU_CLASS_ID_Rational: return (qu_rational) x;
				case QU_CLASS_ID_Float:    return qu_float_as_rational   ((qu_float  ) x);
				case QU_CLASS_ID_Big:      return qu_big_as_rational     ((qu_big    ) x);
				case QU_CLASS_ID_Complex:  return qu_complex_as_rational ((qu_complex) x);

				case QU_CLASS_ID_String:
					/* TODO - String as Rational */
					QU_ARG_THROW;

				case QU_CLASS_ID_Boolean:
					return qu_rational_new_boolean (x);

				case QU_CLASS_ID_Inf:
					QU_ARG_THROW;
			}
			return qu_object_as (x, QuRationalClass);
		}
	}
	else
	{
		qu_real x = argv[0];
		qu_real y = argv[1];
		qu_real v;
		int     cmp;

		QU_ARG_CHECK ((qu_integer_is (x) || qu_decimal_is (x))
			&& (qu_integer_is (y) || qu_decimal_is (y)) && !qu_real_iszero (y));

		if (qu_integer_is (y) && qu_integer_is1 ((qu_integer) y))
			goto one;

		if (qu_real_iszero (x))
			return QuRational0;

		cmp = qu_real_cmpabs_real ((qu_real) x, (qu_real) y);
		if (cmp == 0)
			return QuRational1;

		if (qu_integer_is (x) && qu_integer_is (y))
			return qu_rational_newr ((qu_integer) x, (qu_integer) y);

		if (1)
		{
			qu_integer fn, fd, f;
			qu_real    d, n;

			if (qu_integer_is (x))
				fn = (qu_integer) x;
			else if (qu_float_is (x))
				fn = qu_float_integer_factor ((qu_float) x);
			else
				fn = qu_big_integer_factor ((qu_big) x);

			if (qu_integer_is (y))
				fd = (qu_integer) y;
			else if (qu_float_is (y))
				fd = qu_float_integer_factor ((qu_float) y);
			else
				fd = qu_big_integer_factor ((qu_big) y);

			f = (qu_integer_cmp_integer (fn, fd) >= 0) ? fn : fd;

			if (f && qu_integer_is (f))
			{
				d = qu_real_mul_integer ((qu_real) x, f);
				if (d && !qu_inf_is (d))
				{
					n = qu_real_mul_integer ((qu_real) y, f);
					if (n && !qu_inf_is (n))
						return qu_rational_newr (qu_real_as_integer (d), qu_real_as_integer (n));
				}
			}
		}

		v = (qu_real) qu_real_div_real ((qu_real) x, (qu_real) y);
		assert (v && qu_real_is (v));

		if (cmp > 0)
		{
			if (qu_real_iswhole (v))
				return qu_rational_new_num (qu_real_as_integer (v));
		}
		else
		{
			qu_real g = (qu_real) qu_real_div_real ((qu_real) y, (qu_real) x);

			assert (g && qu_real_is (g));
			if (qu_real_iswhole (g))
			{
				assert (!qu_real_iszero (g));
				return qu_rational_new_den (qu_real_as_integer (g));
			}
		}

		return qu_real_as_rational (v);
	}
}



qu_rational qu_rational_new_ratio (qu_any x, qu_any y)
{
	qu_any argv[2] = {x, y};
	return qu__rational_new (QuRationalClass, 2, argv);
}



qu_rational qu_rational_newl (long n, long d)
{
	register qu_rational r;

	QU_REQUIREZ (QuEValue, d, "zero denominator");

	if (d < 0)
	{
		n = -n;
		d = -d;
	}
	r = qu_gc_new (QuRationalClass);
	r->num = qu_integer_newl (n);
	r->den = qu_integer_newl (d);
	return r;
}



qu_rational qu_rational_newr (qu_integer n, qu_integer d)
{
	qu_rational r;
	qu_integer  g;

	assert (qu_integer_is (n));
	assert (qu_integer_is (d));

	QU_REQUIREZ (QuEValue, !qu_integer_iszero (d), "zero denominator");

	if (qu_integer_iszero (n))
		return QuRational0;

	if (qu_integer_isneg (d))
	{
		n = qu_integer_neg (n);
		d = qu_integer_neg (d);
	}

	if (qu_integer_isone (n))
	{
		if (qu_integer_isone (d))
			return qu_integer_isneg (n) ? QuRational_1 : QuRational1;

		n = qu_integer_isneg (n) ? (qu_integer) QuInt_1 : (qu_integer) QuInt1;
	}
	else if (qu_integer_isone (d))
	{
		d = (qu_integer) QuInt1;
	}
	else
	{
		/* !!! don't use qu_integer_div_integer otherwise we loop forever */
		g = qu_integer_gcd (n, d);
		n = (qu_integer) qu_integer_idiv_integer (n, g);
		d = (qu_integer) qu_integer_idiv_integer (d, g);
	}
	r = qu_gc_new (QuRationalClass);
	r->num = n;
	r->den = d;
	return r;
}



qu_rational qu_rational_new_static (qu_integer n, qu_integer d)
{
	qu_rational r;
	qu_integer  g;

	assert (qu_integer_is (n));
	assert (qu_integer_is (d));

	QU_REQUIREZ (QuEValue, !qu_integer_iszero (d), "zero denominator");

	if (qu_integer_iszero (n))
		return QuRational0;

	if (qu_integer_isneg (d))
	{
		n = qu_integer_neg (n);
		d = qu_integer_neg (d);
	}

	if (qu_integer_isone (n))
	{
		if (qu_integer_isone (d))
			return qu_integer_isneg (n) ? QuRational_1 : QuRational1;

		n = qu_integer_isneg (n) ? (qu_integer) QuInt_1 : (qu_integer) QuInt1;
	}
	else if (qu_integer_isone (d))
	{
		d = (qu_integer) QuInt1;
	}
	else
	{
		/* !!! don't use qu_integer_div_integer otherwise we loop forever */
		g = qu_integer_gcd (n, d);
		n = (qu_integer) qu_integer_idiv_integer (n, g);
		d = (qu_integer) qu_integer_idiv_integer (d, g);
	}
	r = qu_gc_new_static (QuRationalClass);
	r->num = n;
	r->den = d;
	return r;
}



qu_rational qu_rational_new_staticl (long n, long d)
{
	qu_rational r;
	long        g;

	QU_REQUIREZ (QuEValue, d, "zero denominator");

	if (n == 0)
		return QuRational0;

	if (d < 0)
	{
		n = -n;
		d = -d;
	}

	if (n == 1 || n == -1)
	{
		if (d == 1)
			return (n < 0) ? QuRational_1 : QuRational1;
	}
	else
	{
		g = qu_integer_gcdl (n, d);
	}
	r = qu_gc_new_static (QuRationalClass);
	r->num = qu_integer_newl (n);
	r->den = qu_integer_newl (d);
	return r;
}



qu_rational qu_rational_to_static (qu_rational self)
{
	qu_rational r = qu_gc_new_static (QuRationalClass);

	if (qu_int_is (self->num))
		r->num = self->num;
	else
		r->num = (qu_integer) qu_long_to_static ((qu_long) self->num);

	if (qu_int_is (self->den))
		r->den = self->den;
	else
		r->den = (qu_integer) qu_long_to_static ((qu_long) self->den);
	return r;
}



qu_real qu_rational_new_ex (qu_integer n, qu_integer d)
{
	qu_rational r;
	qu_integer  g;
	int         sn, sd;

	assert (qu_integer_is (n));
	assert (qu_integer_is (d));
	assert (!qu_integer_iszero (d));

	sn = qu_integer_sign (n);
	sd = qu_integer_sign (d);
	assert (sd);

	if (sn == 0)
		return (qu_real) QuInt0;

	if (qu_integer_isone (n))
	{
		int s = sn * sd;

		if (qu_integer_isone (d))
			return (s < 0) ? (qu_real) QuInt_1 : (qu_real) QuInt1;

		if (sd < 0)
			d = qu_integer_neg (d);
		n = (s < 0) ? (qu_integer) QuInt_1 : (qu_integer) QuInt1;
	}
	else if (qu_integer_isone (d))
	{
		return (qu_real) n;
	}
	else
	{
		/* !!! don't use qu_integer_div_integer otherwise we loop forever */
		g = qu_integer_gcd (n, d);
		n = (qu_integer) qu_integer_idiv_integer (n, g);
		d = (qu_integer) qu_integer_idiv_integer (d, g);

		if (qu_integer_isone (d))
			return (qu_real) n;

		if (sd < 0)
		{
			d = qu_integer_neg (d);
			n = qu_integer_neg (n);
		}
	}
	r = qu_gc_new (QuRationalClass);
	r->num = n;
	r->den = d;
	return (qu_real) r;
}



qu_rational qu_rational_new_num (qu_integer n)
{
	register qu_rational r;

	assert (qu_integer_is (n));

	if (qu_integer_iszero (n))
		return QuRational0;

	r = qu_gc_new (QuRationalClass);
	r->num = n;
	r->den = (qu_integer) QuInt1;
	return r;
}




qu_rational qu_rational_new_den (qu_integer n)
{
	register qu_rational r;

	assert (qu_integer_is (n));
	assert (!qu_integer_iszero (n));

	r = qu_gc_new (QuRationalClass);
	r->num = (qu_integer) QuInt1;
	r->den = n;
	return r;
}




qu_rational qu_rational_newd (double v)
{
	static qu_long d = 0;

	int        e;
	double     f;
	qu_integer x, y;

	if (d == 0)
		d = qu_long_new_static ("100000000000000000");

	f = frexp (v, &e) * 1E17;
	x = (qu_integer) qu_long_newd (f);
	assert (x && qu_long_is (x));

	if (e == 0)
		return qu_rational_newr ((qu_integer) x, (qu_integer) d);

	if (e > 0)
	{
		x = (qu_integer) qu_integer_lshift_int (x, qu_int_new (e));
		assert (x && qu_integer_is (x));
		return qu_rational_newr (x, (qu_integer) d);
	}
	else
	{
		y = (qu_integer) qu_long_lshift_int (d, qu_int_new (-e));
		assert (y && qu_integer_is (y));
		return qu_rational_newr (x, y);
	}
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

static void qu__rational__mark (qu_rational self)
{
	qu_gc_mark (self->num);
	qu_gc_mark (self->den);
}



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_rational_isone (qu_rational x)
{
	return qu_rational_iswhole (x) && qu_integer_isone (x->num);
}



int qu_rational_is1 (qu_rational x)
{
	return qu_rational_isone (x) && !qu_rational_isneg (x);
}



int qu_rational_isodd (qu_rational x)
{
	return qu_rational_iswhole (x) && qu_integer_isodd (x->num);
}



int qu_rational_isfrac (qu_rational x)
{
	return !qu_rational_iszero (x) && qu_integer_cmpabs_integer (x->num, x->den) < 0;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Rational}
		 -{Real}
		-{Number} | Return self.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{String}  | Delegate to #{__string}.
	.| -{Int}     | Return an Int if possible, throw EValue otherwise.
	.| -{Long}    | Return a Long.
	.| -{Integer} | Return an Int if possible, Long otherwise.
	.| -{Float}   | Return a Float if possible, throw EValue otherwise.
	.| -{Big}     | Return a Big.
	.| -{Decimal} | Return a Float if possible, Big otherwise.

	@exc
	.EArgValue
	.EValue
*/
static qu_any qu__rational__as (qu_rational self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Int:      return qu_rational_as_int  (self);
		case QU_CLASS_ID_Long:     return qu_rational_as_long (self);
		case QU_CLASS_ID_Rational:
		case QU_CLASS_ID_Real:
		case QU_CLASS_ID_Number:   return self;
		case QU_CLASS_ID_Float:    return qu_rational_as_float   (self);
		case QU_CLASS_ID_Big:      return qu_rational_as_big     (self);
		case QU_CLASS_ID_Complex:  return qu_rational_as_complex (self);
		case QU_CLASS_ID_Integer:  return qu_rational_as_integer (self);
		case QU_CLASS_ID_Decimal:  return qu_rational_as_decimal (self);
		case QU_CLASS_ID_Boolean:  return qu_rational_as_boolean (self);
		case QU_CLASS_ID_String:   return qu_rational_as_string  (self);
	}
	QU_ARG_THROW;
}



/*----------------------------------------------------------------------------
	convert
----------------------------------------------------------------------------*/

long qu_rational_tol (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_tol (x->num);
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		long       v;

		assert (r);
		v = qu_integer_tol (r);

		if (errno == 0)
			return v;

		if (qu_rational_isneg (x))
			return 0;
		return LONG_MAX;
	}
}



unsigned long qu_rational_toul (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_toul (x->num);
	}
	else if (qu_rational_isneg (x))
	{
		errno = ERANGE;
		return 0;
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		long       v;

		assert (r);
		v = qu_integer_toul (r);
		return errno ? 0 : v;
	}
}



float qu_rational_tof (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_tof (x->num);
	}
	else
	{
		float n = qu_integer_tof (x->num);

		if (errno == 0)
		{
			float d = qu_integer_tof (x->den);

			if (errno == 0)
				return n / d;
		}
		return qu_big_tof (qu_rational_as_big (x));
	}
}



double qu_rational_tod (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_tod (x->num);
	}
	else
	{
		double n = qu_integer_tod (x->num);

		if (errno == 0)
		{
			double d = qu_integer_tod (x->den);

			if (errno == 0)
				return n / d;
		}
		return qu_big_tod (qu_rational_as_big (x));
	}
}



#ifdef HAVE_LONG_LONG
long long qu_rational_toll (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_toll (x->num);
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		long long  v;

		assert (r);
		v = qu_integer_toll (r);

		if (errno == 0)
			return v;

		if (qu_rational_isneg (x))
			return LONG_LONG_MIN;
		return LONG_LONG_MAX;
	}
}



unsigned long long qu_rational_toull (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_toull (x->num);
	}
	else if (qu_rational_isneg (x))
	{
		errno = ERANGE;
		return 0;
	}
	else
	{
		qu_integer         r = qu_integer_idiv_integer (x->num, x->den);
		unsigned long long v;

		assert (r);
		v = qu_integer_toull (r);
		return errno ? 0 : v;
	}
}
#endif /*HAVE_LONG_LONG*/



#ifdef HAVE_LONG_DOUBLE
long double qu_rational_told (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_told (x->num);
	}
	else
	{
		long double n = qu_integer_told (x->num);

		if (errno == 0)
		{
			long double d = qu_integer_told (x->den);

			if (errno == 0)
				return n / d;
		}
		return qu_big_told (qu_rational_as_big (x));
	}
}
#endif /*HAVE_LONG_DOUBLE*/



qu_int qu_rational_as_int (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_as_int (x->num);
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);

		assert (r);
		return qu_integer_as_int (r);
	}
}



qu_long qu_rational_as_long (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_as_long (x->num);
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);

		assert (r);
		return qu_integer_as_long (r);
	}
}



qu_integer qu_rational_as_integer (qu_rational x)
{
	if (qu_rational_iswhole (x))
		return x->num;
	else
		return qu_integer_idiv_integer (x->num, x->den);
}



qu_decimal qu_rational_as_decimal (qu_rational x)
{
	double d = qu_rational_tod (x);
	return errno ? (qu_decimal) qu_rational_as_big (x) : (qu_decimal) qu_float_new (d);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __cmp (x) -> Int
	Compare self with another object.

	If v{x} is a Number then their numerical values are compared.
	Refer to -{Number.__cmp} for details on how comparisson is made.

	The returned value is:
	@args
	.-1  self is smaller than v{x}
	.0   self equals v{x}
	.1   self is greater than v{x}

	@noexc

	@xmp
		a = 1 _/ 2
		b = 1.0 / 3
		println (a <=> b)

	@out
		1
*/
static qu_int qu__rational__cmp (qu_rational self, qu_any x)
{
	int r;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_rational_cmp_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_rational_cmp_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_rational_cmp_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_rational_cmp_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_rational_cmp_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_rational_cmp_complex  (self, (qu_complex ) x); break;
		case QU_CLASS_ID_Inf:      r = qu_rational_cmp_inf      (self, (qu_inf     ) x); break;

		default: return qu_object_default_cmp (self, x);
	}
	return QU_INT_CMPVAL (r);
}



int qu_rational_cmp_int (qu_rational x, qu_int y)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_cmp_int (x->num, y);
	}
	else
	{
		int sx = qu_rational_sign (x);
		int sy = qu_int_sign (y);

		QU_NUMBER_COMPARE_SIGN (sx, sy);
		assert (sx == sy && sx);
		/*else*/
		{
			double a = qu_rational_tod (x);
			if (errno == 0)
			{
				double b = qu_int_tod (y);
				return QU_NUMBER_COMPARE_VALUE (a, b);
			}
			return sx;
		}
	}
}



int qu_rational_cmp_long (qu_rational x, qu_long y)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_cmp_long (x->num, y);
	}
	else
	{
		int sx = qu_rational_sign (x);
		int sy = qu_long_sign (y);

		QU_NUMBER_COMPARE_SIGN (sx, sy);
		assert (sx == sy && sx);

		if (qu_rational_isfrac (x))
			return -sx;

		return qu_big_cmp_long (qu_rational_as_big (x), y);
	}
}



int qu_rational_cmp_rational (qu_rational x, qu_rational y)
{
	int sx = qu_rational_sign (x);
	int sy = qu_rational_sign (y);
	int cn, cd;

	QU_NUMBER_COMPARE_SIGN (sx, sy);
	assert (sx == sy && sx);

	cn = qu_integer_cmp_integer (x->num, y->num);
	cd = qu_integer_cmp_integer (x->den, y->den);

	if (cn > 0)
	{
		if (cd <= 0)
			/* 9/5  2/7 */
			return sx;
	}
	else if (cn < 0)
	{
		if (cd >= 0)
			/* 2/7  9/5 */
			return -sx;
	}
	else
	{
		if (cd > 0)
			/* 2/7  2/5 */
			return -sx;

		else if (cd < 0)
			/* 2/5  2/7 */
			return sx;

		else
			return 0;
	}
	return qu_decimal_cmp_decimal (qu_rational_as_decimal (x), qu_rational_as_decimal (y));
}



QU_NUMBER_BINOP_integer (qu_rational, cmp, int)
QU_NUMBER_BINOP_decimal (qu_rational, cmp, int)
QU_NUMBER_BINOP_real    (qu_rational, cmp, int)
QU_NUMBER_BINOP_number  (qu_rational, cmp, int)



/*
	|<=>|
*/

int qu_rational_cmpabs_int (qu_rational x, qu_int y)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_cmpabs_int (x->num, y);
	}
	else
	{
		double a = qu_rational_tod (x);
		if (errno == 0)
		{
			double b = qu_int_tod (y);
			if (a < 0)
				a = -a;
			if (b < 0)
				b = -b;
			return QU_NUMBER_COMPARE_VALUE (a, b);
		}
		return 1;
	}
}



int qu_rational_cmpabs_long (qu_rational x, qu_long y)
{
	if (qu_rational_iswhole (x))
	{
		return qu_integer_cmpabs_long (x->num, y);
	}
	else
	{
		if (qu_rational_isfrac (x))
			return -1;

		return qu_big_cmpabs_long (qu_rational_as_big (x), y);
	}
}



int qu_rational_cmpabs_rational (qu_rational x, qu_rational y)
{
	int cn = qu_integer_cmpabs_integer (x->num, y->num);
	int cd = qu_integer_cmpabs_integer (x->den, y->den);

	if (cn > 0)
	{
		if (cd <= 0)
			return 1;
	}
	else if (cn < 0)
	{
		if (cd >= 0)
			return -1;
	}
	else
	{
		if (cd > 0)
			return -1;

		else if (cd < 0)
			return 1;

		else
			return 0;
	}
	return qu_decimal_cmpabs_decimal (qu_rational_as_decimal (x), qu_rational_as_decimal (y));
}



QU_NUMBER_BINOP_integer (qu_rational, cmpabs, int)
QU_NUMBER_BINOP_decimal (qu_rational, cmpabs, int)
QU_NUMBER_BINOP_real    (qu_rational, cmpabs, int)
QU_NUMBER_BINOP_number  (qu_rational, cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __decr -> Number
	Return self decremented by 1.

	If the result is huge and the -{Number.context} does not allow infinities
	then throws EHuge, otherwise returns infn.

	@exc
	.EHuge the result is too big

	@xmp
		i = 1 _/ 2
		println (--i)

	@out
		(-1 _/ 2)
*/
qu_real qu__rational__decr (qu_rational self)
{
	return qu_rational_minus_int (self, QuInt1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __div (x: Number) -> Number
	Return the result of dividing self with v{x}.

	@warn
	There is a big difference between the __div method (the C{/} operator)
	and the #{__idiv} method (the C{//} operator).

	Refer to -{Number.__div} for details on how the division is performed.

	@exc
	.EDiv
	.EInf
	.EHuge
	.ETiny
	.EBrain

	@xmp
		a = 7 _/ 2
		println (a / 3)
		println (a / 7 _/ 2)

	@out
		(7 _/ 6)
		1

	@note
	Qu bothers to check whether self is v{x} (the same object) and return 1
	if it is.
*/
static qu_number qu__rational__div (qu_rational self, qu_number x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_rational_div_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_rational_div_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_div_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_rational_div_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_rational_div_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_rational_div_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_rational_div_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEDiv);
}



static qu_real _rational_div_1 (qu_integer a, qu_integer b, qu_integer d)
{
	/*
		a   1
		- * -
		b   d
	*/
	qu_integer g;

	assert (!qu_integer_iszero (a));
	assert (!qu_integer_iszero (b));
	assert (!qu_integer_iszero (d));

	g = qu_integer_gcd (a, d);
	a = (qu_integer) qu_integer_idiv_integer (a, g); assert (a && qu_integer_is (a));
	d = (qu_integer) qu_integer_idiv_integer (d, g); assert (d && qu_integer_is (d));
	b = (qu_integer) qu_integer_mul_integer (b, d);
	if (b && qu_integer_is (b))
		return qu_rational_new_ex (a, b);
	QU_THROW (QuEBrain);
}



qu_real qu_rational_div_int (qu_rational x, qu_int y)
{
	if (qu_int_iszero (y))
	{
		QU_REQUIRE (QuEDiv, !qu_rational_iszero (x));
		return (qu_real) __qu_zerodiv__ (qu_rational_sign (x));
	}
	else if (qu_rational_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	return _rational_div_1 (x->num, x->den, (qu_integer) y);
}



qu_real qu_rational_div_long (qu_rational x, qu_long y)
{
	if (qu_long_iszero (y))
	{
		QU_REQUIRE (QuEDiv, !qu_rational_iszero (x));
		return (qu_real) __qu_zerodiv__ (qu_rational_sign (x));
	}
	else if (qu_rational_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	return _rational_div_1 (x->num, x->den, (qu_integer) y);
}



static qu_real _rational_mul (qu_integer a, qu_integer b, qu_integer c, qu_integer d);

qu_real qu_rational_div_rational (qu_rational x, qu_rational y)
{
	if (x == y)
	{
		return (qu_real) QuInt1;
	}
	else if (qu_rational_iszero (y))
	{
		QU_REQUIRE (QuEDiv, !qu_rational_iszero (x));
		return (qu_real) __qu_zerodiv__ (qu_rational_sign (x));
	}
	else if (qu_rational_iszero (x))
	{
		return (qu_real) QuInt0;
	}
	return _rational_mul (x->num, x->den, y->den, y->num);
}



qu_int qu_rational_div_inf (qu_rational x, qu_inf y)
{
	QU_REQUIRE (QuEDiv, !qu_rational_iszero (x));
	return QuInt0;
}



QU_NUMBER_BINOP_integer (qu_rational, div, qu_real)
QU_NUMBER_BINOP_decimal (qu_rational, div, qu_real)
QU_NUMBER_BINOP_real    (qu_rational, div, qu_real)
QU_NUMBER_BINOP_number  (qu_rational, div, qu_number)



/*
	/ reversed
*/

static qu_real _rational_div_r1 (qu_integer a, qu_integer c, qu_integer d)
{
	/*
		a   c
		- * -
		1   d
	*/
	qu_integer g;

	assert (!qu_integer_iszero (a));
	assert (!qu_integer_iszero (c));
	assert (!qu_integer_iszero (d));

	g = qu_integer_gcd (a, d);
	a = (qu_integer) qu_integer_idiv_integer (a, g); assert (a && qu_integer_is (a));
	d = (qu_integer) qu_integer_idiv_integer (d, g); assert (d && qu_integer_is (d));
	a = (qu_integer) qu_integer_mul_integer (a, c);
	if (a && qu_integer_is (a))
		return qu_rational_new_ex (a, d);
	QU_THROW (QuEBrain);
}



qu_real qu_rational_rdiv_int (qu_rational x, qu_int y)
{
	if (qu_rational_iszero (x))
	{
		QU_REQUIRE (QuEDiv, !qu_int_iszero (y));
		return (qu_real) __qu_zerodiv__ (qu_int_sign (y));
	}
	else if (qu_int_iszero (y))
	{
		return (qu_real) QuInt0;
	}
	return _rational_div_r1 ((qu_integer) y, x->den, x->num);
}



qu_real qu_rational_rdiv_long (qu_rational x, qu_long y)
{
	if (qu_rational_iszero (x))
	{
		QU_REQUIRE (QuEDiv, !qu_long_iszero (y));
		return (qu_real) __qu_zerodiv__ (qu_long_sign (y));
	}
	else if (qu_long_iszero (y))
	{
		return (qu_real) QuInt0;
	}
	return _rational_div_r1 ((qu_integer) y, x->den, x->num);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __eq (x) -> Boolean
	Return true if v{x} is a -{Number} and has the same numerical value
	as self, false otherwise.

	Refer to -{Number.__eq} for details on how the equality test is performed.

	@noexc

	@xmp
		x = 7 _/ 1
		println (x == 7)

	@out
		true
*/
static qu_boolean qu__rational__eq (qu_rational self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_rational_eq_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_rational_eq_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_rational_eq_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_rational_eq_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_rational_eq_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_rational_eq_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



int qu_rational_eq_int (qu_rational x, qu_int y)
{
	return qu_rational_iswhole (x) && qu_integer_eq_int (x->num, y);
}



int qu_rational_eq_long (qu_rational x, qu_long y)
{
	return qu_rational_iswhole (x) && qu_integer_eq_long (x->num, y);
}



int qu_rational_eq_rational (qu_rational x, qu_rational y)
{
	return qu_integer_eq_integer (x->den, y->den) && qu_integer_eq_integer (x->num, y->num);
}



QU_NUMBER_BINOP_integer (qu_rational, eq, int)
QU_NUMBER_BINOP_decimal (qu_rational, eq, int)
QU_NUMBER_BINOP_real    (qu_rational, eq, int)
QU_NUMBER_BINOP_number  (qu_rational, eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __format (fmt: String) -> String
	Format this Rational.

	See -{Number.__format} for details on the format.

	Note that unless you specify 'f' then the format is used for both the
	numerator and denominator so if you want different formats for each then
	do something like C{printfln ("(%4d _/ %8d)", x.num, xden)}.

	@exc
	.EFormat
	.EMemory

	@xmp
		r = 3 _/ 7
		println (r.__format ('%d'))
		printfln ('%.6f', r)

	@out
		3 _/ 7
		0.428571
*/
static qu_string qu__rational__format (qu_rational self, qu_string x)
{
	qu_format_s f;

	QU_REQUIRE (QuEFormat, qu_string_is (x));
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_rational_format (self, &f);
}



qu_string qu_rational_format (qu_rational x, qu_format_s *f)
{
	qu_string s;

	switch (f->mode)
	{
		case 'y': return qu_format_yesno (f, x);
		case 'p': return qu_format_ptr (f, x);

		case 'f':
		case 'e':
		{
			qu_any r = qu_rational_as_float (x);

			if (r)
			{
				return qu_float_format ((qu_float) r, f);
			}
			else
			{
				r = qu_rational_as_big (x);
				assert (r);
				return qu_big_format ((qu_big) r, f);
			}
		}
	}
	/*default*/
	{
		int  w;
		qu_string  n, d;

		w = f->width;
		f->width = 0;
		n = qu_integer_format (x->num, f); QU_ENSURE (n);
		d = qu_integer_format (x->den, f); QU_ENSURE (d);
		f->width = w;
		s = qu_string_cat (n, qu_string_byte (' '), qu_string_byte ('_'), qu_string_byte ('/'),
			qu_string_byte (' '), d, NULL);
	}
	return qu_format_align (f, s);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __hash -> Int
	Return the hash value of this Rational.

	@noexc

	@xmp
		i = 7 _/ 3
		println (i.__hash)

	@out
		2
*/
static qu_int qu__rational__hash (qu_rational self)
{
	qu_integer x = qu_integer_idiv_integer (self->num, self->den);
	return qu_int_is (x) ? qu_int_hash ((qu_int) x) : qu_long_hash ((qu_long) x);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __idiv (x: Number) -> Integer
	Divide self with v{x} in an I{integer context}.

	@warn
	There is a big difference between the #{__div} method (the "/" operator)
	and the __idiv method (the "//" operator).

	Refer to -{Number.__idiv} for for details on how the division is performed.

	@exc
	.EIdiv
	.EBrain

	@xmp
		a = 11 _/ 3
		println (a // 2)

	@out
		1
*/
static qu_integer qu__rational__idiv (qu_rational self, qu_number x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_rational_idiv_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return qu_rational_idiv_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return qu_rational_idiv_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return qu_rational_idiv_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return qu_rational_idiv_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return qu_rational_idiv_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return qu_rational_idiv_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEIdiv);
}



static qu_integer _rational_idiv_1 (qu_integer a, qu_integer b, qu_integer d)
{
	/*
		a   1
		- * -
		b   d
	*/
	qu_integer g;

	assert (!qu_integer_iszero (a));
	assert (!qu_integer_iszero (b));
	assert (!qu_integer_iszero (d));

	g = qu_integer_gcd (a, d);
	a = (qu_integer) qu_integer_idiv_integer (a, g); assert (a && qu_integer_is (a));
	d = (qu_integer) qu_integer_idiv_integer (d, g); assert (d && qu_integer_is (d));
	b = (qu_integer) qu_integer_mul_integer  (b, d);
	if (b && qu_integer_is (b))
		return qu_integer_idiv_integer (a, b);
	QU_THROW (QuEBrain);
}



qu_integer qu_rational_idiv_int (qu_rational x, qu_int y)
{
	if (qu_int_iszero (y))
	{
		QU_THROW (QuEIdiv);
	}
	else if (qu_rational_iszero (x))
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		return _rational_idiv_1 (x->num, x->den, (qu_integer) y);
	}
}



qu_integer qu_rational_idiv_long (qu_rational x, qu_long y)
{
	if (qu_long_iszero (y))
	{
		QU_THROW (QuEIdiv);
	}
	else if (qu_rational_iszero (x))
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		return _rational_idiv_1 (x->num, x->den, (qu_integer) y);
	}
}



qu_integer qu_rational_idiv_rational (qu_rational x, qu_rational y)
{
	if (x == y)
	{
		return (qu_integer) QuInt1;
	}
	else if (qu_rational_iszero (y))
	{
		QU_THROW (QuEIdiv);
	}
	else if (qu_rational_iszero (x))
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		/*
			a   d
			- * -
			b   c
		*/
		qu_integer a = x->num;
		qu_integer b = x->den;
		qu_integer c = y->num;
		qu_integer d = y->den;
		qu_integer g = qu_integer_gcd (a, c);
		qu_integer h = qu_integer_gcd (b, d);

		a = (qu_integer) qu_integer_idiv_integer (a, g); assert (a && qu_integer_is (a));
		c = (qu_integer) qu_integer_idiv_integer (c, g); assert (c && qu_integer_is (c));

		b = (qu_integer) qu_integer_idiv_integer (b, h); assert (b && qu_integer_is (b));
		d = (qu_integer) qu_integer_idiv_integer (d, h); assert (d && qu_integer_is (d));

		a = (qu_integer) qu_integer_mul_integer (a, d);
		if (a && qu_integer_is (a))
		{
			b = (qu_integer) qu_integer_mul_integer (b, c);
			if (b && qu_integer_is (b))
				return qu_integer_idiv_integer (a, b);
		}
		QU_THROW (QuEBrain);
	}
}



qu_integer qu_rational_idiv_inf (qu_rational x, qu_inf y)
{
	QU_REQUIRE (QuEIdiv, !qu_rational_iszero (x));
	return (qu_integer) QuInt0;
}



QU_NUMBER_BINOP_integer (qu_rational, idiv, qu_integer)
QU_NUMBER_BINOP_decimal (qu_rational, idiv, qu_integer)
QU_NUMBER_BINOP_real    (qu_rational, idiv, qu_integer)
QU_NUMBER_BINOP_number  (qu_rational, idiv, qu_integer)



/*
	\ reversed
*/

static qu_integer _rational_idiv_r1 (qu_integer a, qu_integer c, qu_integer d)
{
	/*
		a   d
		- * -
		1   c
	*/
	qu_integer g;

	assert (!qu_integer_iszero (a));
	assert (!qu_integer_iszero (c));
	assert (!qu_integer_iszero (d));

	g = qu_integer_gcd (a, c);
	a = (qu_integer) qu_integer_idiv_integer (a, g); assert (a && qu_integer_is (a));
	c = (qu_integer) qu_integer_idiv_integer (c, g); assert (c && qu_integer_is (c));
	a = (qu_integer) qu_integer_mul_integer  (a, d);
	if (a && qu_integer_is (a))
		return qu_integer_idiv_integer (a, c);
	QU_THROW (QuEBrain);
}



qu_integer qu_rational_ridiv_int (qu_rational x, qu_int y)
{
	if (qu_rational_iszero (x))
	{
		QU_THROW (QuEIdiv);
	}
	else if (qu_int_iszero (y))
	{
		return (qu_integer) y;
	}
	else
	{
		return _rational_idiv_r1 ((qu_integer) y, x->num, x->den);
	}
}



qu_integer qu_rational_ridiv_long (qu_rational x, qu_long y)
{
	if (qu_rational_iszero (x))
	{
		QU_THROW (QuEIdiv);
	}
	else if (qu_long_iszero (y))
	{
		return (qu_integer) y;
	}
	else
	{
		return _rational_idiv_r1 ((qu_integer) y, x->num, x->den);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __incr -> Number
	Return self incremented by 1.

	If the result is huge and the -{Number.context} does not allow infinities
	then throws EHuge, otherwise returns infp.

	@exc
	.EHuge

	@xmp
		a = 1 _/ 2
		println (++a)

	@out
		(3 _/ 2)
*/
qu_real qu__rational__incr (qu_rational self)
{
	return qu_rational_plus_int (self, QuInt1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __like (x) -> Boolean
	If v{x} is a -{Number} and it has the same absolute numerical
	value as self then true is returned, false otherwise.

	Refer to -{Number.__like} for for details on how the comparisson is performed.

	@noexc

	@xmp
		x = 2 _/ 1
		println (x === -2.0)

	@out
		true
*/
static qu_boolean qu__rational__like (qu_rational self, qu_any x)
{
	int r = 0;
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      r = qu_rational_like_int      (self, (qu_int     ) x); break;
		case QU_CLASS_ID_Long:     r = qu_rational_like_long     (self, (qu_long    ) x); break;
		case QU_CLASS_ID_Rational: r = qu_rational_like_rational (self, (qu_rational) x); break;
		case QU_CLASS_ID_Float:    r = qu_rational_like_float    (self, (qu_float   ) x); break;
		case QU_CLASS_ID_Big:      r = qu_rational_like_big      (self, (qu_big     ) x); break;
		case QU_CLASS_ID_Complex:  r = qu_rational_like_complex  (self, (qu_complex ) x); break;
	}
	return r ? QuTrue : QuFalse;
}



int qu_rational_like_int (qu_rational x, qu_int y)
{
	return qu_rational_iswhole (x) && qu_integer_like_int (x->num, y);
}



int qu_rational_like_long (qu_rational x, qu_long y)
{
	return qu_rational_iswhole (x) && qu_integer_like_long (x->num, y);
}



int qu_rational_like_rational (qu_rational x, qu_rational y)
{
	return qu_integer_like_integer (x->den, y->den) && qu_integer_like_integer (x->num, y->num);
}



QU_NUMBER_BINOP_integer (qu_rational, like, int)
QU_NUMBER_BINOP_decimal (qu_rational, like, int)
QU_NUMBER_BINOP_real    (qu_rational, like, int)
QU_NUMBER_BINOP_number  (qu_rational, like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __minus (x: Number) -> Number
	Return the result of subtracting self with v{x}.

	Refer to -{Number.__minus} for for details on how the subtraction is performed.

	@exc
	.EMinus
	.EHuge
	.EInf
	.EBrain

	@xmp
		x = 6 _/ 7
		println (x - 1)

	@out
		(-1 _/ 7)
*/
static qu_number qu__rational__minus (qu_rational self, qu_number x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_rational_minus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_rational_minus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_minus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_rational_minus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_rational_minus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_rational_minus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_rational_minus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMinus);
}



static qu_real _rational_minus (qu_integer a, qu_integer b, qu_integer c, qu_integer d)
{
	/*
		a   c
		- - -
		b   d
	*/
	qu_integer r;

	r = (qu_integer) qu_integer_mul_integer (b, d);
	if (r && qu_integer_is (r))
	{
		a = (qu_integer) qu_integer_mul_integer (a, d);
		if (a && qu_integer_is (a))
		{
			c = (qu_integer) qu_integer_mul_integer (c, b);
			if (c && qu_integer_is (c))
			{
				a = (qu_integer) qu_integer_minus_integer (a, c);
				if (a && qu_integer_is (a))
					return qu_rational_new_ex (a, r);
			}
		}
	}
	QU_THROW (QuEBrain);
}



qu_real qu_rational_minus_int (qu_rational x, qu_int y)
{
	if (qu_rational_iszero (x))
	{
		return (qu_real) qu_int_neg (y);
	}
	else if (qu_int_iszero (y))
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	else
	{
		return _rational_minus (x->num, x->den, (qu_integer) y, (qu_integer) QuInt1);
	}
}



qu_real qu_rational_minus_long (qu_rational x, qu_long y)
{
	if (qu_rational_iszero (x))
	{
		return (qu_real) qu_long_neg (y);
	}
	else if (qu_long_iszero (y))
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	else
	{
		return _rational_minus (x->num, x->den, (qu_integer) y, (qu_integer) QuInt1);
	}
}



qu_real qu_rational_minus_rational (qu_rational x, qu_rational y)
{
	if (qu_rational_iszero (x))
	{
		if (qu_rational_iswhole (y))
		{
			return (qu_real) qu_integer_neg (y->num);
		}
		else
		{
			y = qu_rational_neg (y);
			return RATIONAL_RESULT (qu_real, y);
		}
	}
	else if (qu_rational_iszero (y))
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	return _rational_minus (x->num, x->den, y->num, y->den);
}



QU_NUMBER_BINOP_integer (qu_rational, minus, qu_real)
QU_NUMBER_BINOP_decimal (qu_rational, minus, qu_real)
QU_NUMBER_BINOP_real    (qu_rational, minus, qu_real)
QU_NUMBER_BINOP_number  (qu_rational, minus, qu_number)



/*
	- reversed
*/

qu_real qu_rational_rminus_int (qu_rational x, qu_int y)
{
	return _rational_minus ((qu_integer) y, (qu_integer) QuInt1, x->num, x->den);
}



qu_real qu_rational_rminus_long (qu_rational x, qu_long y)
{
	return _rational_minus ((qu_integer) y, (qu_integer) QuInt1, x->num, x->den);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __mul (x: Number) -> Number
	Return the result of multiplying self with v{x}.

	Refer to -{Number.__mul} for for details on how the multiplication is performed.

	@exc
	.EMul
	.EHuge
	.ETiny
	.EInf
	.EBrain

	@xmp
		x = 3 _/ 8
		println (x * 8)

	@out
		3
*/
static qu_number qu__rational__mul (qu_rational self, qu_number x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_rational_mul_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_rational_mul_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_mul_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_rational_mul_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_rational_mul_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_rational_mul_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_rational_mul_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEMul);
}



static qu_real _rational_mul_1 (qu_integer a, qu_integer b, qu_integer c)
{
	/*
		a   c
		- * -
		b   1
	*/
	qu_integer h;

	assert (!qu_integer_iszero (a));
	assert (!qu_integer_iszero (b));
	assert (!qu_integer_iszero (c));

	h = qu_integer_gcd (c, b);
	c = (qu_integer) qu_integer_idiv_integer (c, h); assert (c && qu_integer_is (c));
	b = (qu_integer) qu_integer_idiv_integer (b, h); assert (b && qu_integer_is (b));
	a = (qu_integer) qu_integer_mul_integer  (a, c);
	if (a && qu_integer_is (a))
		return qu_rational_new_ex (a, b);
	QU_THROW (QuEBrain);
}



static qu_real _rational_mul (qu_integer a, qu_integer b, qu_integer c, qu_integer d)
{
	/*
		a   c
		- * -
		b   d
	*/
	qu_integer g, h;

	assert (!qu_integer_iszero (a));
	assert (!qu_integer_iszero (b));
	assert (!qu_integer_iszero (c));
	assert (!qu_integer_iszero (d));

	g = qu_integer_gcd (a, d);
	h = qu_integer_gcd (c, b);

	a = (qu_integer) qu_integer_idiv_integer (a, g); assert (a && qu_integer_is (a));
	d = (qu_integer) qu_integer_idiv_integer (d, g); assert (d && qu_integer_is (d));

	c = (qu_integer) qu_integer_idiv_integer (c, h); assert (c && qu_integer_is (c));
	b = (qu_integer) qu_integer_idiv_integer (b, h); assert (b && qu_integer_is (b));

	a = (qu_integer) qu_integer_mul_integer (a, c);
	if (a && qu_integer_is (a))
	{
		b = (qu_integer) qu_integer_mul_integer (b, d);
		if (b && qu_integer_is (b))
			return qu_rational_new_ex (a, b);
	}
	QU_THROW (QuEBrain);
}



qu_real qu_rational_mul_int (qu_rational x, qu_int y)
{
	if (qu_rational_iszero (x) || qu_int_iszero (y))
	{
		return (qu_real) QuInt0;
	}
	else if (qu_rational_iswhole (x))
	{
		return qu_integer_mul_int (x->num, y);
	}
	else
	{
		return _rational_mul_1 (x->num, x->den, (qu_integer) y);
	}
}



qu_real qu_rational_mul_long (qu_rational x, qu_long y)
{
	if (qu_rational_iszero (x) || qu_long_iszero (y))
	{
		return (qu_real) QuInt0;
	}
	else if (qu_rational_iswhole (x))
	{
		return qu_integer_mul_long (x->num, y);
	}
	else
	{
		return _rational_mul_1 (x->num, x->den, (qu_integer) y);
	}
}



qu_real qu_rational_mul_rational (qu_rational x, qu_rational y)
{
	if (qu_rational_iszero (x) || qu_rational_iszero (y))
	{
		return (qu_real) QuInt0;
	}
	else
	{
		return _rational_mul (x->num, x->den, y->num, y->den);
	}
}



QU_NUMBER_BINOP_integer (qu_rational, mul, qu_real)
QU_NUMBER_BINOP_decimal (qu_rational, mul, qu_real)
QU_NUMBER_BINOP_real    (qu_rational, mul, qu_real)
QU_NUMBER_BINOP_number  (qu_rational, mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __neg -> Rational
	Return the negation of self.

	Note that there is no conversion here. The result is always a Rational even
	if the denominator is 1.

	@noexc

	@xmp
		x = 1 _/ 2
		println (-x)

	@out
		(-1 _/ 2)
*/
qu_rational qu__rational__neg (qu_rational self)
{
	if (qu_rational_iszero (self))
	{
		return self;
	}
	else
	{
		/* no need to normalize since it's already is */
		qu_rational r = qu_gc_new (QuRationalClass);
		r->num = qu_integer_neg (self->num);
		r->den = self->den;
		return r;
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __plus (x: Number) -> Number
	Return the result of adding v{x} to self.

	Refer to -{Number.__plus} for for details on how the addition is performed.

	@exc
	.EPlus
	.EHuge
	.EInf
	.EBrain

	@xmp
		x = 1 _/ 2
		println (x + 1)

	@out
		(3 _/ 2)
*/
static qu_number qu__rational__plus (qu_rational self, qu_number x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_rational_plus_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_rational_plus_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_plus_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_rational_plus_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_rational_plus_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_rational_plus_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_rational_plus_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPlus);
}



static qu_real _rational_plus (qu_integer a, qu_integer b, qu_integer c, qu_integer d)
{
	/*
		a   c
		- + -
		b   d
	*/
	qu_integer r;

	r = (qu_integer) qu_integer_mul_integer (b, d);
	if (r && qu_integer_is (r))
	{
		a = (qu_integer) qu_integer_mul_integer (a, d);
		if (a && qu_integer_is (a))
		{
			c = (qu_integer) qu_integer_mul_integer (c, b);
			if (c && qu_integer_is (c))
			{
				a = (qu_integer) qu_integer_plus_integer (a, c);
				if (a && qu_integer_is (a))
					return qu_rational_new_ex (a, r);
			}
		}
	}
	QU_THROW (QuEBrain);
}



qu_real qu_rational_plus_int (qu_rational x, qu_int y)
{
	if (qu_rational_iszero (x))
	{
		return (qu_real) y;
	}
	else if (qu_int_iszero (y))
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	return _rational_plus (x->num, x->den, (qu_integer) y, (qu_integer) QuInt1);
}



qu_real qu_rational_plus_long (qu_rational x, qu_long y)
{
	if (qu_rational_iszero (x))
	{
		return (qu_real) y;
	}
	else if (qu_long_iszero (y))
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	else
	{
		return _rational_plus (x->num, x->den, (qu_integer) y, (qu_integer) QuInt1);
	}
}



qu_real qu_rational_plus_rational (qu_rational x, qu_rational y)
{
	if (qu_rational_iszero (x))
	{
		return (qu_real) y;
	}
	else if (qu_rational_iszero (y))
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	else
	{
		return _rational_plus (x->num, x->den, y->num, y->den);
	}
}



QU_NUMBER_BINOP_integer (qu_rational, plus, qu_real)
QU_NUMBER_BINOP_decimal (qu_rational, plus, qu_real)
QU_NUMBER_BINOP_real    (qu_rational, plus, qu_real)
QU_NUMBER_BINOP_number  (qu_rational, plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub pos -> self
	This method always return self.
*/
#define qu__rational__pos  qu_object_method_self



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __pow (x: Number) -> Rational
	Return the result of raising self to the power of v{x}.

	Refer to -{Number.__pow} for for details on how the operation is performed.

	@exc
	.EPow
	.EHuge
	.ETiny
	.EInf

	@xmp
		a = 1 _/ 3
		println (a ** 2)
		println (a ** -2)

	@out
		(1 _/ 9)
		9
*/
static qu_number qu__rational__pow (qu_rational self, qu_number x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_rational_pow_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_rational_pow_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_pow_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_rational_pow_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_rational_pow_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_rational_pow_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_rational_pow_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuEPow);
}



qu_real qu_rational_pow_int (qu_rational x, qu_int y)
{
	long v = qu_int_tol (y);

	if (qu_rational_iswhole (x))
	{
		/* includes x == 0 */
		return qu_integer_pow_int (x->num, y);
	}
	else if (v == 0)
	{
		return (qu_real) QuInt1;
	}
	else if (v == -1)
	{
		if (qu_integer_isone (x->num))
		{
			if (qu_integer_isneg (x->num))
			{
				return (qu_real) qu_integer_neg (x->den);
			}
			else
			{
				return (qu_real) x->den;
			}
		}
		else
		{
			x = qu_rational_inv (x);
			assert (x);
			return RATIONAL_RESULT (qu_real, x);
		}
	}
	else if (v == 1)
	{
		return RATIONAL_RESULT (qu_real, x);
	}
	else if (v < 0)
	{
		qu_integer n, d;

		y = qu_int_neg (y);
		n = (qu_integer) qu_integer_pow_int (x->den, y); QU_REQUIRE (QuEPow, n && qu_integer_is (n));
		d = (qu_integer) qu_integer_pow_int (x->num, y); QU_REQUIRE (QuEPow, d && qu_integer_is (d));
		return qu_rational_new_ex (n, d);
	}
	else
	{
		qu_integer n, d;

		n = (qu_integer) qu_integer_pow_int (x->num, y); QU_REQUIRE (QuEPow, n && qu_integer_is (n));
		d = (qu_integer) qu_integer_pow_int (x->den, y); QU_REQUIRE (QuEPow, d && qu_integer_is (d));
		return qu_rational_new_ex (n, d);
	}
}



qu_real qu_rational_pow_long (qu_rational x, qu_long y)
{
	if (qu_rational_iswhole (x))
	{
		/* includes x == 0 */
		return qu_integer_pow_long (x->num, y);
	}
	else if (qu_long_iszero (y))
	{
		return (qu_real) QuInt1;
	}
	else if (qu_long_isone (y))
	{
		if (qu_long_isneg (y))
		{
			if (qu_integer_isone (x->num))
			{
				if (qu_integer_isneg (x->num))
				{
					return (qu_real) qu_integer_neg (x->den);
				}
				else
				{
					return (qu_real) x->den;
				}
			}
			else
			{
				x = qu_rational_inv (x);
				assert (x);
				return RATIONAL_RESULT (qu_real, x);
			}
		}
		else
		{
			return RATIONAL_RESULT (qu_real, x);
		}
	}
	else if (qu_long_isneg (y))
	{
		qu_integer n, d;

		y = qu_long_neg (y);
		n = (qu_integer) qu_integer_pow_long (x->den, y); QU_REQUIRE (QuEPow, n && qu_integer_is (n));
		d = (qu_integer) qu_integer_pow_long (x->num, y); QU_REQUIRE (QuEPow, d && qu_integer_is (d));
		return qu_rational_new_ex (n, d);
	}
	else
	{
		qu_integer n, d;

		n = (qu_integer) qu_integer_pow_long (x->num, y); QU_REQUIRE (QuEPow, n && qu_integer_is (n));
		d = (qu_integer) qu_integer_pow_long (x->den, y); QU_REQUIRE (QuEPow, d && qu_integer_is (d));
		return qu_rational_new_ex (n, d);
	}
}



qu_number qu_rational_pow_rational (qu_rational x, qu_rational y)
{
	if (qu_rational_iswhole (y))
	{
		return (qu_number) qu_rational_pow_integer (x, y->num);
	}
	else
	{
		qu_decimal a = qu_rational_as_decimal (x);
		qu_decimal b = qu_rational_as_decimal (y);

		return qu_decimal_pow_decimal (a, b);
	}
}



qu_real qu_rational_pow_inf (qu_rational x, qu_inf y)
{
	if (y == QuInfPos)
	{
		if (qu_rational_iszero (x))
		{
			return (qu_real) QuRational0;
		}
		else if (qu_rational_isone (x))
		{
			return (qu_real) QuRational1;
		}
		else
		{
			return (qu_real) __qu_inf__ (qu_inf_sign (y));
		}
	}
	else
	{
		if (qu_rational_iszero (x))
		{
			return (qu_real) __qu_inf__ (1);
		}
		else
		{
			return qu_rational_isone (x) ? (qu_real) QuRational1 : (qu_real) QuRational0;
		}
	}
}



QU_NUMBER_BINOP_integer (qu_rational, pow, qu_real)
QU_NUMBER_BINOP_decimal (qu_rational, pow, qu_real)
QU_NUMBER_BINOP_real    (qu_rational, pow, qu_real)
QU_NUMBER_BINOP_number  (qu_rational, pow, qu_number)



/*
	** reversed
*/

qu_number qu_rational_rpow_int (qu_rational x, qu_int y)
{
	if (qu_rational_iswhole (x))
	{
		return (qu_number) qu_int_pow_integer (y, x->num);
	}
	else
	{
		qu_float r = qu_rational_as_float (x);

		if (r)
			return qu_int_pow_float (y, r);
		else
			return qu_int_pow_big (y, qu_rational_as_big (x));
	}
}



qu_number qu_rational_rpow_long (qu_rational x, qu_long y)
{
	if (qu_rational_iswhole (x))
	{
		return (qu_number) qu_long_pow_integer (y, x->num);
	}
	else
	{
		qu_float r = qu_rational_as_float (x);

		if (r)
			return qu_long_pow_float (y, r);
		else
			return qu_long_pow_big (y, qu_rational_as_big (x));
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	print
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __print -> Null
	Print k[self} to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		println (1 _/ 2)

	@out
		(1 _/ 2)
*/
static qu_null qu__rational__print (qu_rational self)
{
	qu_string s = qu_rational_as_string (self);
	return qu_sys_printz (s->str, s->len);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __rem (x: Number) -> Number
	Return the remainder of dividing self with v{x}.

	@warn
	The meaning of "remainder" may differ with other languages.
	Refer to -{Number.__rem} for details.

	@p
	The remainder of dividing by zero is undefined thus throws ERem.

	@exc
	.ERem

	@xmp
		x = 8 _/ 3
		println (x % 2)

	@out
		(2 _/ 3)
*/
static qu_number qu__rational__rem (qu_rational self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_number) qu_rational_rem_int      (self, (qu_int     ) x);
		case QU_CLASS_ID_Long:     return (qu_number) qu_rational_rem_long     (self, (qu_long    ) x);
		case QU_CLASS_ID_Rational: return (qu_number) qu_rational_rem_rational (self, (qu_rational) x);
		case QU_CLASS_ID_Float:    return (qu_number) qu_rational_rem_float    (self, (qu_float   ) x);
		case QU_CLASS_ID_Big:      return (qu_number) qu_rational_rem_big      (self, (qu_big     ) x);
		case QU_CLASS_ID_Complex:  return (qu_number) qu_rational_rem_complex  (self, (qu_complex ) x);
		case QU_CLASS_ID_Inf:      return (qu_number) qu_rational_rem_inf      (self, (qu_inf     ) x);
	}
	QU_THROW (QuERem);
}



qu_real qu_rational_rem_int (qu_rational x, qu_int y)
{
	if (qu_int_iszero (y))
	{
		QU_THROW (QuERem);
	}
	else
	{
		int cmp = qu_rational_cmpabs_int (x, y);

		if (cmp < 0)
		{
			return RATIONAL_RESULT (qu_real, x);
		}
		else if (cmp == 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			qu_integer b;

			b = qu_rational_idiv_int (x, y);         assert (b && qu_integer_is (b));
			b = (qu_integer) qu_integer_mul_int (b, y); assert (b && qu_integer_is (b));
			return qu_rational_minus_integer (x, b);
		}
	}
}



qu_real qu_rational_rem_long (qu_rational x, qu_long y)
{
	if (qu_long_iszero (y))
	{
		QU_THROW (QuERem);
	}
	else
	{
		int cmp = qu_rational_cmpabs_long (x, y);

		if (cmp < 0)
		{
			return RATIONAL_RESULT (qu_real, x);
		}
		else if (cmp == 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			qu_integer b;

			b = qu_rational_idiv_long (x, y);         assert (b && qu_integer_is (b));
			b = (qu_integer) qu_long_mul_integer (y, b); assert (b && qu_integer_is (b));
			return qu_rational_minus_integer (x, b);
		}
	}
}



qu_real qu_rational_rem_rational (qu_rational x, qu_rational y)
{
	if (qu_rational_iszero (y))
	{
		QU_THROW (QuERem);
	}
	else
	{
		int cmp = qu_rational_cmpabs_rational (x, y);

		if (cmp < 0)
		{
			return RATIONAL_RESULT (qu_real, x);
		}
		else if (cmp == 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			qu_real    r;
			qu_integer b;

			b = qu_rational_idiv_rational (x, y); QU_REQUIRE (QuERem, b);
			r = qu_rational_mul_integer   (y, b); QU_REQUIRE (QuERem, r);
			return qu_rational_minus_real (x, r);
		}
	}
}



qu_rational qu_rational_rem_inf (qu_rational x, qu_inf y)
{
	QU_REQUIRE (QuERem, !qu_rational_iszero (x));
	return x;
}



QU_NUMBER_BINOP_integer (qu_rational, rem, qu_real)
QU_NUMBER_BINOP_decimal (qu_rational, rem, qu_real)
QU_NUMBER_BINOP_real    (qu_rational, rem, qu_real)
QU_NUMBER_BINOP_number  (qu_rational, rem, qu_number)



/*
	% reversed
*/

qu_real qu_rational_rrem_int (qu_rational x, qu_int y)
{
	if (qu_rational_iszero (x))
	{
		QU_THROW (QuERem);
	}
	else
	{
		int cmp = qu_int_cmpabs_rational (y, x);

		if (cmp < 0)
		{
			return (qu_real) y;
		}
		else if (cmp == 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			qu_integer b;
			qu_real    r;

			b = qu_int_idiv_rational    (y, x); assert  (b && qu_integer_is (b));
			r = qu_integer_mul_rational (b, x); QU_REQUIRE (QuERem, r);
			return qu_int_minus_real (y, r);
		}
	}
}



qu_real qu_rational_rrem_long (qu_rational x, qu_long y)
{
	if (qu_rational_iszero (x))
	{
		QU_THROW (QuERem);
	}
	else
	{
		int cmp = qu_long_cmpabs_rational (y, x);

		if (cmp < 0)
		{
			return (qu_real) y;
		}
		else if (cmp == 0)
		{
			return (qu_real) QuInt0;
		}
		else
		{
			qu_integer b;
			qu_real    r;

			b = qu_long_idiv_rational   (y, x); assert  (b && qu_integer_is (b));
			r = qu_integer_mul_rational (b, x); QU_REQUIRE (QuERem, r);
			return qu_long_minus_real (y, r);
		}
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __string -> String
	Return a string representation of self.

	@noexc

	@xmp
		r = Rational (1, 2)
		s = String (r)
		println (s)

	@out
		(1 _/ 2)
*/
qu_string qu__rational__string (qu_rational x)
{
	qu_string n = qu_integer_as_string (x->num);
	qu_string d = qu_integer_as_string (x->den);

	return qu_string_cat (qu_string_byte ('('), n, qu_string_byte (' '), qu_string_byte ('_'), qu_string_byte ('/'),
		qu_string_byte (' '),  d, qu_string_byte (')'), 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*) sub __true -> Boolean
	Return false if self is 0, true otherwise.

	@noexc

	@xmp
		a = 0 _/ 1
		println (a ? 'T' : 'F')

	@out
		F
*/
qu_boolean qu__rational__true (qu_rational self)
{
	return qu_rational_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub abs -> Rational
	Return the absolute value.
*/
static qu_rational qu__rational_abs (qu_rational self)
{
	if (qu_rational_isneg (self))
		return qu_rational_neg (self);
	return self;
}



/*) sub ceil -> Rational
	Return the result of rounding this number to the nearest integer towards
	positive infinity.
*/
qu_rational qu__rational_ceil (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return x;
	}
	else if (qu_rational_isfrac (x))
	{
		return qu_rational_isneg (x) ? QuRational0 : QuRational1;
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		assert (r && qu_integer_is (r));
		if (qu_rational_isneg (x))
		{
			r = (qu_integer) qu_integer_plus_int (r, QuInt1);
			QU_REQUIRE (QuEMath, r && qu_integer_is (r));
		}
		return qu_rational_new_num (r);
	}
}



#if 0
qu_rational qu_rational_ceil_integer (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return x->num;
	}
	else if (qu_rational_isfrac (x))
	{
		return qu_rational_isneg (x) ? (qu_integer) QuInt0 : (qu_integer) QuInt1;
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		return qu_rational_isneg (x) ? r : (qu_integer) qu_integer_plus_int (r, QuInt1);
	}
}
#endif



/*) sub den -> Integer
	Return the denominator.

	@noexc

	@xmp
		a = 1 _/ 2
		println (a.den)

	@out
		2
*/
static qu_integer qu__rational_den (qu_rational self)
{
	return self->den;
}



/*) sub floor -> Rational
	Return the result of rounding this number to the nearest integer
	towards negative infinity.
*/
qu_rational qu__rational_floor (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return x;
	}
	else if (qu_rational_isfrac (x))
	{
		return qu_rational_isneg (x) ? QuRational_1 : QuRational0;
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		assert (r && qu_integer_is (r));
		if (qu_rational_isneg (x))
		{
			r = (qu_integer) qu_integer_minus_int (r, QuInt1);
			QU_REQUIRE (QuEMath, r && qu_integer_is (r));
		}
		return qu_rational_new_num (r);
	}
}


#if 0
qu_integer qu_rational_floor_integer (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return x->num;
	}
	else if (qu_rational_isfrac (x))
	{
		return qu_rational_isneg (x) ? (qu_integer) QuInt0 : (qu_integer) QuInt1;
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		return qu_rational_isneg (x) ? (qu_integer) qu_integer_minus_int (r, QuInt1) : r;
	}
}
#endif



/*) sub frac -> Rational
	Return the fraction part.
*/
qu_rational qu__rational_frac (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return QuRational0;
	}
	else
	{
		qu_integer f, m;

		f =           qu_integer_idiv_integer  (x->num, x->den); assert (f && qu_integer_is (f));
		m = (qu_integer) qu_integer_mul_integer   (f, x->den);      assert (m && qu_integer_is (m));
		f = (qu_integer) qu_integer_minus_integer (x->num, m);      assert (f && qu_integer_is (f));

		return qu_rational_newr (f, x->den);
	}
}



/*) sub inv -> Rational
	Return a new Rational by inverting the numerator and denominator.

	If the numerator is zero then EArgValue is thrown.

	@exc
	EArgValue

	@xmp
		a = 1 _/ 2
		println (a.inv)

		b = 1 / a
		println (b)

	@out
		(2 _/ 1)
		2
*/
qu_rational qu__rational_inv (qu_rational self)
{
	qu_rational g;

	QU_ARG_CHECKZ (!qu_rational_iszero (self), "inv (0/x)");

	g = qu_gc_new (QuRationalClass);
	g->den = self->num;
	g->num = self->den;
	return g;
}



/*) sub iseven -> Boolean
	Return true if self is an even number, false otherwise.
*/
static qu_boolean qu__rational_iseven (qu_rational self)
{
	return qu_rational_isodd (self) ? QuFalse : QuTrue;
}



/*) sub isfrac -> Boolean
	Return true if C{0 < self < 1}, false otherwise.
*/
static qu_boolean qu__rational_isfrac (qu_rational self)
{
	return qu_rational_isfrac (self) ? QuTrue : QuFalse;
}



/*) sub isodd -> Boolean
	Return true if self is an odd number, false otherwise.
*/
static qu_boolean qu__rational_isodd (qu_rational self)
{
	return qu_rational_isodd (self) ? QuTrue : QuFalse;
}



/*) sub iswhole -> Boolean
	Return true if self is a I{whole} number, false otherwise.
*/
static qu_boolean qu__rational_iswhole (qu_rational self)
{
	return qu_rational_iswhole (self) ? QuTrue : QuFalse;
}



/*) sub num -> Integer
	Return the numerator.

	@noexc

	@xmp
		a = 1 _/ 2
		println (a.num)

	@out
		1
*/
static qu_integer qu__rational_num (qu_rational self)
{
	return self->num;
}



/*) sub round (n: Int) -> Rational
	Return the result of rounding this number to the specified
	v{precision} where v{precision} is the number of decimal digits.

	@warn
	The result may not be exact since it involves changing the Rational into a
	Decimal, round it, and change the result back into a Rational.
*/
qu_rational qu_rational_round (qu_rational self, int prec)
{
	qu_any x;

	x = qu_rational_as_decimal (self);
	QU_ENSURE (x);
	if (qu_float_is (x))
	{
		x = qu_float_round (x, prec);
		QU_ENSURE (x);
		return qu_float_as_rational ((qu_float) x);
	}
	else
	{
		assert (qu_big_is (x));
		x = qu_big_round (x, prec);
		QU_ENSURE (x);
		return qu_big_as_rational ((qu_big) x);
	}
}



static qu_rational qu__rational_round (qu_rational self, qu_integer prec)
{
	int p;
	QU_ARG_int (prec, p);
	return qu_rational_round (self, p);
}



/*) sub sign -> Int
	Return -1, 0, 1 if the number is negatve, zero, or positive respectively.
*/
static qu_int qu__rational_sign (qu_rational self)
{
	long v = qu_rational_sign (self);
	return (v < 0) ? QuInt_1 : ((v > 0) ? QuInt1 : QuInt0);
}



/*) sub trunc -> Rational
	Return the result of rounding this number to the nearest integer
	towards zero, thus removing the fractional part.
*/
qu_rational qu__rational_trunc (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return x;
	}
	else if (qu_rational_isfrac (x))
	{
		return QuRational0;
	}
	else
	{
		qu_integer r = qu_integer_idiv_integer (x->num, x->den);
		assert (r && qu_integer_is (r));
		return qu_rational_new_num (r);
	}
}

#if 0
qu_integer qu_rational_trunc_integer (qu_rational x)
{
	if (qu_rational_iswhole (x))
	{
		return x->num;
	}
	else if (qu_rational_isfrac (x))
	{
		return (qu_integer) QuInt0;
	}
	else
	{
		return qu_integer_idiv_integer (x->num, x->den);
	}
}
#endif



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

static qu_rational _rational_sys (long n, long d)
{
	qu_rational r;

	assert (d);

	r = qu_gc_new_static (QuRationalClass);
	r->num = qu_integer_newl (n);
	r->den = qu_integer_newl (d);
	return r;
}



void qu_rational__setup ()
{
	qu_class_setup_final (QuRationalClass, qu_rational_s, 0, qu__rational__mark);

	QuRational0  = _rational_sys (  0, 1);
	QuRational1  = _rational_sys (  1, 1);
	QuRational_1 = _rational_sys ( -1, 1);
}



void qu_rational__init ()
{
	/* standard class methods */
	qu_class_method_class (QuRationalClass, QU_NAME_new    , qu__rational_new    , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuRationalClass, QU_NAME__as    , qu__rational__as    ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__cmp   , qu__rational__cmp   ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__decr  , qu__rational__decr  ,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__div   , qu__rational__div   ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__eq    , qu__rational__eq    ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__format, qu__rational__format,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__hash  , qu__rational__hash  ,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__idiv  , qu__rational__idiv  ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__incr  , qu__rational__incr  ,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__like  , qu__rational__like  ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__minus , qu__rational__minus ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__rem   , qu__rational__rem   ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__mul   , qu__rational__mul   ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__neg   , qu__rational__neg   ,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__plus  , qu__rational__plus  ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__pos   , qu__rational__pos   ,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__pow   , qu__rational__pow   ,  1);
	qu_class_method_inst  (QuRationalClass, QU_NAME__print , qu__rational__print ,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__string, qu__rational__string,  0);
	qu_class_method_inst  (QuRationalClass, QU_NAME__true  , qu__rational__true  ,  0);

	/* specific instance methods */
	qu_class_method_instz (QuRationalClass, "abs"    , qu__rational_abs    , 0);
	qu_class_method_instz (QuRationalClass, "ceil"   , qu__rational_ceil   , 0);
	qu_class_method_instz (QuRationalClass, "den"    , qu__rational_den    , 0);
	qu_class_method_instz (QuRationalClass, "floor"  , qu__rational_floor  , 0);
	qu_class_method_instz (QuRationalClass, "frac"   , qu__rational_frac   , 0);
	qu_class_method_instz (QuRationalClass, "inv"    , qu__rational_inv    , 0);
	qu_class_method_instz (QuRationalClass, "iseven" , qu__rational_iseven , 0);
	qu_class_method_instz (QuRationalClass, "isfrac" , qu__rational_isfrac , 0);
	qu_class_method_instz (QuRationalClass, "isodd"  , qu__rational_isodd  , 0);
	qu_class_method_instz (QuRationalClass, "iswhole", qu__rational_iswhole, 0);
	qu_class_method_instz (QuRationalClass, "num"    , qu__rational_num    , 0);
	qu_class_method_instz (QuRationalClass, "round"  , qu__rational_round  , 1);
	qu_class_method_instz (QuRationalClass, "sign"   , qu__rational_sign   , 0);
	qu_class_method_instz (QuRationalClass, "trunc"  , qu__rational_trunc  , 0);

	/* finalize declaration */
	qu_class_set_super (QuRationalClass, QuRealClass);
	qu_class_enddecl   (QuRationalClass);
}
