/* Real.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Real __abstract
	This is the parent class of the -{Integer}, -{Rational}, -{Decimal}
	and -{Inf}.

	The class hierarchy:
	@tree
	.-{Number}
	.	Real
*/
qu_class QuRealClass;


/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

#define UNOP(name, rtype) \
rtype qu_real_##name (qu_real x) \
{ \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) qu_int_##name      ((qu_int) x); \
		case QU_CLASS_ID_Float:    return (rtype) qu_float_##name    ((qu_float) x); \
		case QU_CLASS_ID_Long:     return (rtype) qu_long_##name     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return (rtype) qu_rational_##name ((qu_rational) x); \
		case QU_CLASS_ID_Big:      return (rtype) qu_big_##name      ((qu_big) x); \
		case QU_CLASS_ID_Inf:      return (rtype) qu_inf_##name      ((qu_inf) x); \
	} \
	QU_FATAL ("x is not a Real object"); \
	return 0; \
}


#define BINOP(name, rtype, ytype) \
rtype qu_real_##name##_##ytype (qu_real x, qu_##ytype y) \
{ \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return (rtype) qu_int_##name##_##ytype      ((qu_int) x, y); \
		case QU_CLASS_ID_Float:    return (rtype) qu_float_##name##_##ytype    ((qu_float) x, y); \
		case QU_CLASS_ID_Long:     return (rtype) qu_long_##name##_##ytype     ((qu_long) x, y); \
		case QU_CLASS_ID_Rational: return (rtype) qu_rational_##name##_##ytype ((qu_rational) x, y); \
		case QU_CLASS_ID_Big:      return (rtype) qu_big_##name##_##ytype      ((qu_big) x, y); \
		case QU_CLASS_ID_Inf:      return (rtype) qu_inf_##name##_##ytype      ((qu_inf) x, y); \
	} \
	QU_FATAL ("x is not a Real object"); \
	return 0; \
}

#define BINOP_int(name, rtype)      BINOP (name, rtype, int)
#define BINOP_float(name, rtype)    BINOP (name, rtype, float)
#define BINOP_long(name, rtype)     BINOP (name, rtype, long)
#define BINOP_rational(name, rtype) BINOP (name, rtype, rational)
#define BINOP_big(name, rtype)      BINOP (name, rtype, big)
#define BINOP_complex(name, rtype)  BINOP (name, rtype, complex)
#define BINOP_inf(name, rtype)      BINOP (name, rtype, inf)
#define BINOP_integer(name, rtype)  QU_NUMBER_BINOP_integer (qu_real, name, rtype)
#define BINOP_decimal(name, rtype)  QU_NUMBER_BINOP_decimal (qu_real, name, rtype)
#define BINOP_real(name, rtype)     QU_NUMBER_BINOP_real    (qu_real, name, rtype)
#define BINOP_number(name, rtype)   QU_NUMBER_BINOP_number  (qu_real, name, rtype)



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (x) -> Real
	You can use this method to convert v{x} into a Real number.

	This method simply returns the value returned by C{x.__as (Real)} provided
	that it is a Real number. Otherwise a warning is issued and EImplement
	is thrown.

	@exc
	.ESyntax
	.EImplement

	@xmp
		s = '123'
		r = Real (s)
		println (r ' ' r.__class.__name)

	@out
		123 Int
*/
static qu_real qu__real_new (qu_class self, qu_any x)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:
		case QU_CLASS_ID_Float:
		case QU_CLASS_ID_Long:
		case QU_CLASS_ID_Rational:
		case QU_CLASS_ID_Big:
			return (qu_real) x;

		case QU_CLASS_ID_Inf:
			QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return (qu_real) x;

		case QU_CLASS_ID_Boolean:
			return ((qu_boolean) x == QuTrue) ? (qu_real) QuInt1 : (qu_real) QuInt0;

		case QU_CLASS_ID_String:
			x = qu_number_scan_string ((qu_string) x, QU_NUMBER_SCAN_SKIPSPACE);
			QU_ENSURE (x);
			if (qu_inf_is (x))
				QU_REQUIRE (QuEInf, QU_NUMBER_ALLOW_INF);
			return x;
	}
	return qu_object_as (x, QuRealClass);
}



qu_real qu_real_new_number (qu_number x)
{
	return qu_complex_is (x) ? qu_complex_as_real ((qu_complex) x) : qu_real_as_real ((qu_real) x);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	identity
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_real_is (qu_any x)
{
	qu_class c = qu_classof (x);
	return (c == QuIntClass
		||  c == QuLongClass
		||  c == QuRationalClass
		||  c == QuFloatClass
		||  c == QuBigClass
		||  c == QuInfClass);
}

UNOP (sign   , int);
UNOP (iszero , int);
UNOP (isneg  , int);
UNOP (isone  , int);
UNOP (is1    , int);
UNOP (iswhole, int);
UNOP (isfrac , int);
UNOP (isodd  , int);



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	as
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#define TOC(name, type) \
type qu_real_to##name (qu_any x) \
{ \
	errno = 0; \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return qu_int_to##name      ((qu_int) x); \
		case QU_CLASS_ID_Long:     return qu_long_to##name     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return qu_rational_to##name ((qu_rational) x); \
		case QU_CLASS_ID_Float:    return qu_float_to##name    ((qu_float) x); \
		case QU_CLASS_ID_Big:      return qu_big_to##name      ((qu_big) x); \
		case QU_CLASS_ID_Inf:      return qu_inf_to##name      ((qu_inf) x); \
	} \
	errno = EDOM; \
	return 0; \
}


TOC (h  , short);
TOC (uh , unsigned short);
TOC (i  , int);
TOC (ui , unsigned int);
TOC (l  , long);
TOC (ul , unsigned long);
TOC (li , long);
TOC (f  , float);
TOC (d  , double);
#ifdef HAVE_LONG_LONG
TOC (ll , long long);
TOC (ull, unsigned long long);
#endif
#ifdef HAVE_LONG_DOUBLE
TOC (ld , long double);
#endif

#undef TOC


#define AS(name) \
qu_##name qu_real_as_##name (qu_any x) \
{ \
	switch (qu_classid (x)) \
	{ \
		case QU_CLASS_ID_Int:      return qu_int_as_##name      ((qu_int) x); \
		case QU_CLASS_ID_Long:     return qu_long_as_##name     ((qu_long) x); \
		case QU_CLASS_ID_Rational: return qu_rational_as_##name ((qu_rational) x); \
		case QU_CLASS_ID_Float:    return qu_float_as_##name    ((qu_float) x); \
		case QU_CLASS_ID_Big:      return qu_big_as_##name      ((qu_big) x); \
		case QU_CLASS_ID_Inf:      return qu_inf_as_##name      ((qu_inf) x); \
	} \
	QU_THROW (QuEValue); \
}


AS (int);
AS (long);
AS (rational);
AS (float);
AS (big);
AS (complex);
AS (integer);
AS (decimal);
AS (real);
AS (number);

#undef AS


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	<=>
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (cmp, int)
BINOP_float    (cmp, int)
BINOP_long     (cmp, int)
BINOP_rational (cmp, int)
BINOP_big      (cmp, int)
BINOP_complex  (cmp, int)
BINOP_inf      (cmp, int)
BINOP_integer  (cmp, int)
BINOP_decimal  (cmp, int)
BINOP_real     (cmp, int)
BINOP_number   (cmp, int)



/*
	|<=>|
*/

BINOP_int      (cmpabs, int)
BINOP_float    (cmpabs, int)
BINOP_long     (cmpabs, int)
BINOP_rational (cmpabs, int)
BINOP_big      (cmpabs, int)
BINOP_complex  (cmpabs, int)
BINOP_inf      (cmpabs, int)
BINOP_integer  (cmpabs, int)
BINOP_decimal  (cmpabs, int)
BINOP_real     (cmpabs, int)
BINOP_number   (cmpabs, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	--
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (decr, qu_real);



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	/
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (div, qu_real)
BINOP_float    (div, qu_real)
BINOP_long     (div, qu_real)
BINOP_rational (div, qu_real)
BINOP_big      (div, qu_real)
BINOP_complex  (div, qu_complex)
BINOP_inf      (div, qu_real)
BINOP_integer  (div, qu_real)
BINOP_decimal  (div, qu_real)
BINOP_real     (div, qu_real)
BINOP_number   (div, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	==
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (eq, int)
BINOP_float    (eq, int)
BINOP_long     (eq, int)
BINOP_rational (eq, int)
BINOP_big      (eq, int)
BINOP_complex  (eq, int)
BINOP_inf      (eq, int)
BINOP_integer  (eq, int)
BINOP_decimal  (eq, int)
BINOP_real     (eq, int)
BINOP_number   (eq, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_string qu_real_format (qu_real x, qu_format_s *f)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return qu_int_format      ((qu_int) x, f);
		case QU_CLASS_ID_Float:    return qu_float_format    ((qu_float) x, f);
		case QU_CLASS_ID_Long:     return qu_long_format     ((qu_long) x, f);
		case QU_CLASS_ID_Rational: return qu_rational_format ((qu_rational) x, f);
		case QU_CLASS_ID_Big:      return qu_big_format      ((qu_big) x, f);
		case QU_CLASS_ID_Inf:      return qu_inf_format      ((qu_inf) x, f);
	}
	QU_FATAL ("x is not a Real object");
	return 0;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	hash
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	\
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (idiv, qu_integer)
BINOP_float    (idiv, qu_integer)
BINOP_long     (idiv, qu_integer)
BINOP_rational (idiv, qu_integer)
BINOP_big      (idiv, qu_integer)
BINOP_complex  (idiv, qu_integer)
BINOP_inf      (idiv, qu_integer)
BINOP_integer  (idiv, qu_integer)
BINOP_decimal  (idiv, qu_integer)
BINOP_real     (idiv, qu_integer)
BINOP_number   (idiv, qu_integer)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	++
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (incr, qu_real);



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	===
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (like, int)
BINOP_float    (like, int)
BINOP_long     (like, int)
BINOP_rational (like, int)
BINOP_big      (like, int)
BINOP_complex  (like, int)
BINOP_inf      (like, int)
BINOP_integer  (like, int)
BINOP_decimal  (like, int)
BINOP_real     (like, int)
BINOP_number   (like, int)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (minus, qu_real)
BINOP_float    (minus, qu_real)
BINOP_long     (minus, qu_real)
BINOP_rational (minus, qu_real)
BINOP_big      (minus, qu_real)
BINOP_complex  (minus, qu_complex)
BINOP_inf      (minus, qu_real)
BINOP_integer  (minus, qu_real)
BINOP_decimal  (minus, qu_real)
BINOP_real     (minus, qu_real)
BINOP_number   (minus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (mul, qu_real)
BINOP_float    (mul, qu_real)
BINOP_long     (mul, qu_real)
BINOP_rational (mul, qu_real)
BINOP_big      (mul, qu_real)
BINOP_complex  (mul, qu_complex)
BINOP_inf      (mul, qu_real)
BINOP_integer  (mul, qu_real)
BINOP_decimal  (mul, qu_real)
BINOP_real     (mul, qu_real)
BINOP_number   (mul, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (neg, qu_real);



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (plus, qu_real)
BINOP_float    (plus, qu_real)
BINOP_long     (plus, qu_real)
BINOP_rational (plus, qu_real)
BINOP_big      (plus, qu_real)
BINOP_complex  (plus, qu_complex)
BINOP_inf      (plus, qu_real)
BINOP_integer  (plus, qu_real)
BINOP_decimal  (plus, qu_real)
BINOP_real     (plus, qu_real)
BINOP_number   (plus, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	+x
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

UNOP (pos, qu_real);



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	**
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (pow, qu_real)
BINOP_float    (pow, qu_real)
BINOP_long     (pow, qu_real)
BINOP_rational (pow, qu_real)
BINOP_big      (pow, qu_real)
BINOP_complex  (pow, qu_complex)
BINOP_inf      (pow, qu_real)
BINOP_integer  (pow, qu_real)
BINOP_decimal  (pow, qu_real)
BINOP_real     (pow, qu_real)
BINOP_number   (pow, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	%
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

BINOP_int      (rem, qu_real)
BINOP_float    (rem, qu_real)
BINOP_long     (rem, qu_real)
BINOP_rational (rem, qu_real)
BINOP_big      (rem, qu_real)
BINOP_complex  (rem, qu_complex)
BINOP_inf      (rem, qu_real)
BINOP_integer  (rem, qu_real)
BINOP_decimal  (rem, qu_real)
BINOP_real     (rem, qu_real)
BINOP_number   (rem, qu_number)



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	str
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	true
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub j -> Complex
	Return a -{Complex} number which real part is Int(0) and imaginary
	part is self.

	@noexc

	Note that this method does not respect the global -{Number.context}.
	It always succeed.

	@xmp
		x = 10
		println (x.j)

	@out
		(0 + 10j)
*/
static qu_complex qu__real_j (qu_real x)
{
	return qu_complex_new ((qu_real) QuInt0, x);
}



/*----------------------------------------------------------------------------
	rounding
----------------------------------------------------------------------------*/

UNOP (floor, qu_real)
UNOP (ceil , qu_real)
UNOP (trunc, qu_real)
UNOP (frac , qu_real)


qu_real qu_real_round (qu_real x, int prec)
{
	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Int:      return (qu_real) qu_int_round      ((qu_int     ) x, prec);
		case QU_CLASS_ID_Long:     return (qu_real) qu_long_round     ((qu_long    ) x, prec);
		case QU_CLASS_ID_Rational: return (qu_real) qu_rational_round ((qu_rational) x, prec);
		case QU_CLASS_ID_Float:    return (qu_real) qu_float_round    ((qu_float   ) x, prec);
		case QU_CLASS_ID_Big:      return (qu_real) qu_big_round      ((qu_big     ) x, prec);
		case QU_CLASS_ID_Inf:      return (qu_real) qu_inf_round      ((qu_inf     ) x, prec);
	}
	QU_FATAL ("x is not a Real object");
	return 0;
}



/*----------------------------------------------------------------------------
	misc
----------------------------------------------------------------------------*/

qu_real __qu_tiny__ (qu_real zero)
{
	assert (qu_real_iszero (zero));
	QU_NUMBER_ENSURE_ALLOW_TINY;
	return zero;
}



qu_inf __qu_huge__ (int sign)
{
	QU_NUMBER_ENSURE_ALLOW_HUGE;
	return (sign < 0) ? QuInfNeg : QuInfPos;
}



qu_inf __qu_inf__ (int sign)
{
	QU_NUMBER_ENSURE_ALLOW_INF;
	return (sign < 0) ? QuInfNeg : QuInfPos;
}



qu_inf __qu_zerodiv__ (int sign)
{
	QU_REQUIRE (QuEDiv, sign);
	QU_NUMBER_ENSURE_ALLOW_ZERODIV;
	return (sign < 0) ? QuInfNeg : QuInfPos;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_real__setup ()
{
	qu_class_setup_base (QuRealClass, qu_object_s);
}



void qu_real__init ()
{
	/* standard class methods */
	qu_class_method_class (QuRealClass, QU_NAME_new, qu__real_new, 1);

	/* specific instance methods */
	qu_class_method_instz (QuRealClass, "j", qu__real_j, 0);

	/* finalize declaration */
	qu_class_set_super (QuRealClass, QuNumberClass);
	qu_class_enddecl   (QuRealClass);
}
