/* Record.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Record
	A Record object is a list of values which conform to a -{Struct}
	object specification.

	To access the values you can use the C{[]} index syntax or the dot
	syntax as you would when accessing a property.

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Record (Person, 'Joe', 98)
			# or simply: Joe = Person ('Joe', 98)

		println (Joe.name ' = ' Joe->name ' = ' Joe[0] ' = ' Joe['name'])
		println (Joe.name ' = ' Joe->age  ' = ' Joe[1] ' = ' Joe['age'])

		try
			Joe->age = 'boo'
		catch
			println (__err__)
		;;

		println (Joe._struct == Person)

	@out
		Joe = Joe = Joe = Joe
		98 = 98 = 98 = 98
		(EValue) invalid type or value
		true

	@note
	Methods of the Record class are named with a leading underscore or double
	underscores. This is done so that when you access the fields with the dot
	notation, you don't mistakenly access a method.

	This means that you should not name the fields of a -{Struct} with a leading
	underscore unless you know that you will never have to access the Record
	class methods.

	@xmp
		struct Person
			_name   : String  # bad idea
			__age   : Int     # ditto
			_struct : Int     # very bad idea
			__len   : String  # pathetic
		;;
*/
qu_class QuRecordClass;

/* to reduce polution */
#define item_s qu_structentry_s

/* strings */
static qu_string Srecursive;


/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

qu_record qu_record_alloc (qu_struct u)
{
	register qu_record r = qu_gc_new (QuRecordClass);

	r->struc = u;
	r->len   = qu_struct_len (u);
	r->objs  = qu_mem_alloc (r->len * sizeof (qu_any));
	QU_OBJECT_LIST_SETVAL (r->objs, QuNil, r->len);
	QU_CONTAINER_INIT (r);
	return r;
}


qu_record qu_record_new (qu_struct u, int argc, qu_any *argv)
{
	long      n = (long) argc;
	qu_record r;
	long      i;
	qu_any   *a;

	QU_REQUIRE (QuEArgCount, n <= qu_struct_len (u));

	r = qu_record_alloc (u);
	a = r->objs;
	for (i = 0; i < n; ++i)
	{
		qu_any x = qu_object_validate (u, u->items[i].type, argv[i]);
		QU_ENSURE (x);
		a[i] = x;
	}
	return r;
}



qu_record qu_record_new_ex (qu_struct u, ...)
{
	va_list    va;
	long       n = qu_struct_len (u);
	qu_record  r;
	qu_any    *a, x;

	r = qu_record_alloc (u);
	a = r->objs;
	va_start (va, u);
	while (n--)
	{
		x = va_arg (va, qu_any);
		assert (x);
		*a++ = x;
	}
	va_end (va);
	return r;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (struc: Struct, *values) -> Record
	Create a new Record.

	v{values} must match the types specified by the -{Struct} object v{struc}.

	@exc
	.EArgValue

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Record (Person, 'Joe', 98)
		println (Joe)

	@out
		['Joe', 98]

	@warn
	It is generally better to use the -{Struct} constructor.

	@xmp
		Joe = Person ('Joe', 98)
*/
qu_record qu__record__inst (qu_class self, int argc, qu_any *argv)
{
	qu_struct u;
	qu_record r;

	QU_ARG_VMIN (2);
	QU_ARG_VTYPE (0, u, qu_struct);
	r = qu_record_new (u, argc - 1, argv + 1);
	r->klass = self;
	return r;
}



/*) sub class new (struc: Struct, *values) -> Record
	This simply calls #{__inst} then f{__init} if one is defined.
*/
qu_record qu__record_new (qu_class self, int argc, qu_any *argv)
{
	qu_record r = qu__record__inst (self, argc, argv);
	QU_ENSURE (r);
	return qu_class_init_new (self, r, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__record__destroy (qu_record self)
{
	qu_mem_free (self->objs);
}



static void qu__record__mark (qu_record self)
{
	long    n = self->len;
	qu_any *p = self->objs;

	qu_gc_mark (self->struc);
	while (n--)
	{
		qu_gc_mark (*p);
		p++;
	}
	QU_CONTAINER_MARK (self);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be a one of the following -{Class} objects:
	@ops
	.| -{Array}   | Return a new Array containing all field values.
	.| -{Set}     | Return a new Set containing all field values.
	.| -{Dict}    | Return a new Dict containing all field names and values.
	.| -{String}  | Delegate to #{__string}.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{Pointer} | Return a new Pointer with the record as content.
	.| -{Record}  | Return self.

	@exc
	.EArgValue

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		r = Person ('Foo', 99)
		println (r as Array)
		println (r as Dict)

	@out
		['Foo', 99]
		['name':'Foo', 'age':99]
*/
static qu_any qu__record__as (qu_record self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return qu_record_as_array   (self);
		case QU_CLASS_ID_Set:     return qu_record_as_set     (self);
		case QU_CLASS_ID_Dict:    return qu_record_as_dict    (self);
		case QU_CLASS_ID_String:  return qu_record_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_record_as_boolean (self);
		case QU_CLASS_ID_Record:  return self;
		case QU_CLASS_ID_Pointer: return qu_pointer_new_record (self);
	}
	QU_ARG_THROW;
}


qu_array qu_record_as_array (qu_record self)
{
	qu_array a;
	qu_any  *p, *q = self->objs;
	long     l = self->len;

	a = qu_array_new (l);
	a->len = l;
	p = a->objs;
	while (l--)
		*p++ = *q++;
	return a;
}



qu_set qu_record_as_set (qu_record self)
{
	qu_any *q = self->objs;
	long    l = self->len;
	qu_set  a;

	a = qu_set_new (l);
	while (l--)
		qu_set_push (a, *q++);
	return a;
}



qu_dict qu_record_as_dict (qu_record self)
{
	item_s  *i = self->struc->items;
	long     l = self->len;
	qu_any  *q = self->objs;
	qu_dict  a;

	a = qu_dict_new (l);
	while (l--)
	{
		qu_dict_setitem (a, i->name, *q++);
		i++;
	}
	return a;
}



/*) sub __copy -> Record
	Return a new copy of this Record.

	@noexc
*/
/*) sub copy -> Record
	This is just an alias of #{__copy}.
*/
static qu_record qu__record__copy (qu_record self)
{
	qu_struct u = self->struc;
	long      n = qu_struct_len (u);
	qu_record r;

	r        = qu_gc_new (self->klass);
	r->struc = u;
	r->len   = n;
	r->objs  = qu_mem_alloc (n * sizeof (qu_any));

	QU_OBJECT_LIST_COPY (r->objs, self->objs, n);
	QU_CONTAINER_INIT (r);

	return r;
}



/*) sub __each -> Iterator
	Return a -{Iterator} object to iterate through the values.

	Note that unless you call this method explicitly, the interperter
	uses an internal iterator with a k{for each} loop which is much
	more efficient.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		r = Person ('Foo', 99)

		println (x)       for x in each r # calls r.__each
		println (k ':' v) for k, v in r   # interpreter internal iterator
		println (x)       for x in r      # interpreter internal iterator

	@out
		Foo
		99
		name:Foo
		age:99
		Foo
		99
*/
static qu_any qu__record__next (qu_record self, qu_iterator i)
{
	long k = i->ex1;

	QU_ITERATOR_STOP (k < self->len);
	i->ex1++;

	if (qu_iterator_yield_keys (i))
		return self->struc->items[k].name;

	else if (qu_iterator_yield_values (i))
		return self->objs[k];

	else
	{
		qu_any argv[2] = {self->struc->items[k].name, self->objs[k]};
		return qu_array_new_argv (2, argv);
	}
}



static qu_iterator qu__record__each (qu_record self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, qu__record__next);
	qu_iterator_set_yield_values (i);
	return i;
}



/*) sub __in (x) -> Boolean
	Whether v{x} is an item in the Record.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Person ('Foo', 99)
		println (99 in Joe)

	@out
		true
*/
static qu_boolean qu__record__in (qu_record self, qu_any x)
{
	long i, l = qu_struct_len (self->struc);

	for (i = 0; i < l; ++i)
	{
		if (qu_sys_eq (self->objs[i], x))
			return QuTrue;
	}
	return QuFalse;
}



/*) sub __getitem (x) -> Any
	@def
	sub __getitem (x: Integer) -> Any
	sub __getitem (x: String) -> Any

	@p
	Get a value by name or index.

	You can access a value in a Record by specifying the name of the field
	or an integer index.

	@exc
	.EKey

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Person ('Foo', 99)

		println (Joe [0])
		println (Joe ['name'])
		println (Joe [-1])

	@out
		Foo
		Foo
		99
*/
static qu_any qu__record__getitem (qu_record self, qu_any x)
{
	long i = qu_struct_index_ex (self->struc, x);
	QU_REQUIRE (QuEKey, i >= 0);
	return self->objs[i];
}



/*) sub __len -> Int
	Return the number of items in the Record.

	This is the same as: C{self._struct.len}.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		a = Person ('Foo', 99)
		println (a.len)

	@out
		2
*/
/*) sub len -> Int
	This is an alias of #{__len}.
*/
static qu_int qu__record__len (qu_record self)
{
	long l = qu_struct_len (self->struc);
	assert (l <= QU_INT_MAX);
	return qu_int_newl (l);
}



/*) sub __like (x) -> Boolean
	Return true if v{x} is a Record and has the same -{Struct}.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		a = Person ('Foo', 99)
		b = Person ('Bar', 98)

		println (a === b)

	@out
		true
*/
static qu_boolean qu__record__like (qu_record self, qu_any x)
{
	return (qu_record_is (x) && ((qu_record) x)->struc == self->struc) ? QuTrue : QuFalse;
}



/*) sub __print -> Null
	Print this Record.

	See -{print} for possible exceptions.

	@xmp
		struct Person
			name : String
			age  : Int
		;;
		Joe = Person ('Joe', 99)
		println (Joe)

	@out
		['Joe', 99]
*/
static qu_null qu__record__print (qu_record self)
{
	if (qu_thread_state_exist (QU_THREAD_STATE_PRINT, self))
	{
		QU_ENSURE (qu_sys_printz (Srecursive->str, Srecursive->len));
	}
	else
	{
		long    i;
		qu_null r;

	  __qu_thread_state_begin (QU_THREAD_STATE_PRINT, self);
		r = qu_sys_printz ("[", 1);
		for (i = 0; r && i < self->len; ++i)
		{
			if (i)
				r = qu_sys_printz (", ", 2);
			if (r)
			{
				qu_any x = self->objs[i];
				r = QU_CONTAINER_PRINT_QUOTED (x);
			}
		}
		r = qu_sys_printz ("]", 1);
	  __qu_thread_state_end;
		QU_ENSURE (r);
	}
	return QuNil;
}




/*) sub __setitem (i, v) -> Any
	Set a value by name or index.

	Return v{v} on success.

	@exc
	.EValue
	.EKey
	.EModify self is frozen

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Person ('Joey', 99)
		println (Joe ['name'])
		println (Joe->name)

	@out
		Joey
		Joey
*/
static qu_any qu__record__setitem (qu_record self, qu_any idx, qu_any x)
{
	long i;

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);
	i = qu_struct_index_ex (self->struc, idx);
	QU_REQUIRE (QuEKey, i >= 0);
	x = qu_object_validate (self, self->struc->items[i].type, x);
	QU_ENSURE (x);
	self->objs[i] = x;
	return x;
}



/*) sub __string
	Return a string representation of self.

	@exc
	.EMemory

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Person ('Joe', 99)
		s = String (Joe)
		println (s)

	@out
		['Joe', 99]
*/
qu_string qu__record__string (qu_record self)
{
	if (qu_thread_state_exist (QU_THREAD_STATE_STRING, self))
		return Srecursive;
	else
	{
		qu_string s = qu_string_buffer_new (32);
		long      i;
		int       r;

	  __qu_thread_state_begin (QU_THREAD_STATE_STRING, self);
		r = qu_string_buffer_putc (s, '[');
		for (i = 0; r && i < self->len; ++i)
		{
			if (i)
				r = qu_string_buffer_put (s, ", ", 2);
			if (r)
			{
				qu_any x = self->objs[i];
				r = qu_string_buffer_puts (s, QU_CONTAINER_QUOTE (x));
			}
		}
		if (r)
			r = qu_string_buffer_putc (s, ']');
	  __qu_thread_state_end;
		QU_ENSURE (r);
		return qu_string_buffer_return (s);
	}
}



/*) sub __true -> Boolean
	Return false if all fields are nil, return true otherwise.

	@noexc

	@xmp
		struct Person
			name : Any
			age  : Any
		;;

		Joe = Person (nil, nil)
		println (Joe ? 'yep' : 'nope')

	@out
		nope
*/
qu_boolean qu__record__true  (qu_record self)
{
	long    n = self->len;
	qu_any *p = self->objs;

	while (n--)
	{
		if (*p != (qu_any) QuNil)
			return QuTrue;
		p++;
	}
	return QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub __setclass (c: Class) -> self
	Set the class to a subclass of this Record class.

	@exc
	.EArgValue
*/
static qu_record qu__record__setclass (qu_record self, qu_class c)
{
	QU_ARG_CHECK (qu_class_is (c) && qu_class_isa (c, self->klass));
	self->klass = c;
	return self;
}



/*) sub _struct
	Return the associated -{Struct} object.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		Joe = Person ('Joe', 99)
		println (Joe._struct)

	@out
		[name:String, age:Int]
*/
static qu_struct qu__record__struct (qu_record self)
{
	return self->struc;
}



/*) sub acquire (r: Record) -> self
	Copy the values from v{r} to self.

	v{r} must have the same Struct object.

	@exc
	.EArgValue
*/
static qu_record qu__record_acquire (qu_record self, qu_record x)
{
	qu_any *a, *b;
	long    l;

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);
	QU_ARG_CHECK (qu_record_is (x) && x->struc == self->struc);
	a = self->objs;
	b = x->objs;
	l = qu_struct_len (self->struc);
	while (l--)
		*a++ = *b++;
	return self;
}



/*----------------------------------------------------------------------------
	set/get instance attributes - runtime
----------------------------------------------------------------------------*/

static qu_any qu__record__get (qu_record self, qu_string name, int /*ignored*/ restrict)
{
	qu_any x;
	long   n = qu_struct_index (((qu_record) self)->struc, (qu_string) name);

	if (n >= 0)
		return ((qu_record) self)->objs[n];

	if (((qu_record) self)->struc->attrs)
	{
		x = qu_attr_find (((qu_record) self)->struc->attrs, name);
		if (x)
		{
			assert (qu_sub_is (x));
			return x;
		}
	}

	x = qu_attr_find (QuRecordClass->attrs, name);
	if (x)
	{
		/* the Record class has no instance property */
		assert (qu_sub_is (x));
		return x;
	}

	x = qu_attr_find (QuObjectClass->attrs, name);
	if (x)
		return x;

	if (__QuAttrNoThrow)
		return 0;

	QU_THROWF (QuEImplement, "Record has no such attribute %s", name->str);
}



static qu_any qu__record__set (qu_record self, qu_string name, qu_any val, /*ignored*/ int restrict)
{
	assert (qu_record_is (self));

	QU_CONTAINER_ENSURE_NOT_FROZEN (self);

	/* the Record class has no settable instance property */
	return qu__record__setitem (self, name, val);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_record__setup ()
{
	qu_class_setup (QuRecordClass, qu_record_s, qu__record__destroy, qu__record__mark);

	QuRecordClass->get = (qu_attrget_f) qu__record__get;
	QuRecordClass->set = (qu_attrset_f) qu__record__set;

	Srecursive = qu_string_new_staticz ("[..]");
}



void qu_record__init ()
{
	/* standard class methods */
	qu_class_method_class (QuRecordClass, QU_NAME__inst, qu__record__inst, -1);
	qu_class_method_class (QuRecordClass, QU_NAME_new  , qu__record_new  , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuRecordClass, QU_NAME__as     , qu__record__as     , 1);
	qu_class_method_inst  (QuRecordClass, QU_NAME__copy   , qu__record__copy   , 0);
	qu_class_alias        (QuRecordClass, QU_NAME__copy   , QU_NAME_copy);
	qu_class_method_inst  (QuRecordClass, QU_NAME__each   , qu__record__each   , 0);
	qu_class_method_inst  (QuRecordClass, QU_NAME__in     , qu__record__in     , 1);
	qu_class_method_inst  (QuRecordClass, QU_NAME__getitem, qu__record__getitem, 1);
	qu_class_method_inst  (QuRecordClass, QU_NAME__len    , qu__record__len    , 0);
	qu_class_alias        (QuRecordClass, QU_NAME__len    , QU_NAME_len);
	qu_class_method_inst  (QuRecordClass, QU_NAME__like   , qu__record__like   , 1);
	qu_class_method_inst  (QuRecordClass, QU_NAME__print  , qu__record__print  , 0);
	qu_class_method_inst  (QuRecordClass, QU_NAME__setitem, qu__record__setitem, 2);
	qu_class_method_inst  (QuRecordClass, QU_NAME__string , qu__record__string , 0);
	qu_class_method_inst  (QuRecordClass, QU_NAME__true   , qu__record__true   , 0);

	/* specific instance methods */
	qu_class_methodz      (QuRecordClass, "__setclass", qu__record__setclass,  1, QU_SUB_FLAG_INST | QU_SUB_FLAG_PRIVATE);
	qu_class_method_instz (QuRecordClass, "_struct"   , qu__record__struct  ,  0);
	qu_class_method_instz (QuRecordClass, "acquire"   , qu__record_acquire  ,  1);

	/* finalize declaration */
	qu_class_set_super (QuRecordClass, QuContainerClass);
	qu_class_enddecl   (QuRecordClass);
}
