/* Reduce.c - Copyright (c) Marc Krisnanto */

#ifndef QU_REDUCE_IN_CODE_C
#error This file is part of Code.c
#endif

/*
	Reduce might remove and modify some nodes for efficiency so printing the
	tree after Reduce will not print the actual parsed tree.

	What Reduce does:
	- resolve symbols
	- perform some validations
	- finalize declarations
	- modify or delete some nodes
	- constant folding

	Examples:
	- WHILE nodes might be turned into NOP
	- ISA and MIXIN nodes are resolved then removed
	- qu_class_set_super and qu_class_set_mixin are called
	- STRUCT are created and turned into NOP
	- etc.

	After Reduce:
	- all SYMBOL nodes are changed to LITERAL
	- all NAME nodes are resolved, the 'thing' field will contain
	  the Module, Class, Sub or Var object
*/

static int ReduceSymbolAllowInst = 0;


/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

/*
	Create local variable.

	name = symbol name
	n    = the node to use to report error
*/
static qu_var reduce_auto_local_create (qu_string name, qu_ast n)
{
	qu_var v;

	if (qu_module_is (CoMaster))
	{
		v = qu_var_new (CoMaster, name, QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_LOCAL);
		qu_module_setattr ((qu_module) CoMaster, name, (qu_symbol) v);
	}
	else if (qu_class_is (CoMaster))
	{
		/*
			qu_class_decl_attr checks for signatures, e.g. if the name happens
			to be a standard method name then it will throw an exception.
		*/
		v = qu_var_new (CoMaster, name, QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_LOCAL);
		if (qu_class_decl_attr ((qu_class) CoMaster, name, (qu_symbol) v) == 0)
			CRETHROW (n);
	}
	else
	{
		/*
			Make sure to use qu_sub_setattr so the variable is indexed and the
			sub local count is adjusted properly.
		*/
		assert (qu_sub_is (CoMaster));
		v = qu_var_new (CoMaster, name, QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_LEXICAL);
		qu_sub_setattr ((qu_sub) CoMaster, name, (qu_symbol) v);
	}
	return v;
}



static void reduce_auto_local_node (qu_string name, qu_ast n)
{
	qu_any x = qu_symbol_find (CoMaster, name);

	if (x == 0)
	{
		x = reduce_auto_local_create (name, n);
	}
	else
	{
		if (!qu_var_is (x))
			CTHROW (n, "must use variable");

		if (!(qu_var_is_local (x) || qu_var_is_lexical (x)))
			CTHROW (n, "must use local variable");

		if (qu_var_frozen (x))
			CTHROW (n, "cant use constant here");
	}
	assert (x);
	n->thing = x;
}



/*
	Make sure all childs are literals.
	Used by ARRAY, DICT and SET.
*/
static void reduce_ensure_literal_childs (qu_ast n)
{
	for (n = n->child; n; n = n->next)
	{
		if (n->type == QU_AST_literal || n->type == QU_AST_ref)
			continue;
		CTHROW (n, "non literal value in constant");
	}
}



/*
	Make sure the thing is an exception class.
	Used by THROW and CATCH.
*/
static void reduce_ensure_literal_exception (qu_ast n)
{
	assert (n->type == QU_AST_literal);

	if (!qu_class_is (n->thing))
		CTHROW (n, "not a class");

	if (!qu_class_isa (n->thing, QuExceptionClass))
		CTHROW (n, "not an exception class");
}



/*----------------------------------------------------------------------------
	more helpers
----------------------------------------------------------------------------*/

/* process one node - forward declaration */
static void reduce_node (qu_ast n);



/* process child nodes */
inline static void reduce_child_nodes (qu_ast n)
{
	for (n = n->child; n; n = n->next)
		reduce_node (n);
}



/* process node starting from n rightwards */
inline static void reduce_nodes_from (qu_ast n)
{
	for (; n; n = n->next)
		reduce_node (n);
}



/*----------------------------------------------------------------------------
	forwards
----------------------------------------------------------------------------*/

#define     reduce_anonsub reduce_sub
static void reduce_array   (qu_ast n);
static void reduce_assert  (qu_ast n);
static void reduce_assign  (qu_ast n);
static void reduce_binop   (qu_ast n);
static void reduce_catch   (qu_ast n);
static void reduce_class   (qu_ast n);
static void reduce_const   (qu_ast n);
static void reduce_dict    (qu_ast n);
static void reduce_eat     (qu_ast n);
static void reduce_enum    (qu_ast n);
static void reduce_extern  (qu_ast n);
static void reduce_for     (qu_ast n);
static void reduce_getattr (qu_ast n);
static void reduce_import  (qu_ast n);
static void reduce_if      (qu_ast n);
static void reduce_isa     (qu_ast n);
static void reduce_local   (qu_ast n);
static void reduce_loop    (qu_ast n);
#define     reduce_mixin   reduce_isa
static void reduce_module  (qu_ast n);
static void reduce_name    (qu_ast n);
static void reduce_param   (qu_ast n);
static void reduce_ref     (qu_ast n);
static void reduce_return  (qu_ast n);
static void reduce_set     (qu_ast n);
static void reduce_static  (qu_ast n);
static void reduce_struct  (qu_ast n);
static void reduce_sub     (qu_ast n);
static void reduce_symbol  (qu_ast n);
static void reduce_switch  (qu_ast n);
static void reduce_unop    (qu_ast n);
static void reduce_var     (qu_ast n);
static void reduce_while   (qu_ast n);
static void reduce_with    (qu_ast n);
static void reduce_yield   (qu_ast n);



/*----------------------------------------------------------------------------
	special case of SYMBOL -> type

	SYMBOL -> thing = name
	|
	...

----------------------------------------------------------------------------*/

static void reduce_type (qu_ast n)
{
	qu_string s = n->thing;
	qu_ast    c = n->child;
	qu_class  x = 0;

	/*
		All this to handle special cases of containers to eliminate the
		need to create class validators, for example:

		var foo: Array of String
		var bar: Dict of @baz

		Thus instead of having to create validator methods, e.g. Array.String,
		Set.Int, Dict.Array, etc., we use a more flexible validator Var object
		which fields are:

		- name   = '*' (not important)
		- value  = QuArrayClass | QuSetClass | QuDictClass
		- type   = any valid validator
		- flag  |= QU_VAR_FLAG_TYPEOF

	*/
	if (s == QuArrayClass->name || strcmp (s->str, "Array") == 0)
		x = QuArrayClass;

	else if (s == QuSetClass->name || strcmp (s->str, "Set") == 0)
		x = QuSetClass;

	else if (s == QuDictClass->name || strcmp (s->str, "Dict") == 0)
		x = QuDictClass;

	if (x == 0)
	{
		ReduceSymbolAllowInst = 1;
		reduce_symbol (n);
		if (n->type != QU_AST_literal)
			CTHROW (n, "undeclared symbol");

		/* If type is Module get a symbol in that Module with the same name

			FIXME - there are a number of places in Reduce.c that do this
			independently. They should no longer do it because we do it
			here. Type related stuffs is still messy.
		*/
		if (qu_module_is (n->thing))
		{
			qu_module m = (qu_module) n->thing;
			qu_string s = qu_file_name (m->name);
			qu_any    t = qu_module_getattr (m, s);

			if (t == 0)
				CTHROW (n, "undeclared symbol");

			if (!(qu_class_is (t) || qu_sub_is (t) || qu_var_is (t)))
				CTHROW (n, "invalid type");

			n->thing = t;
			n->child = NULL;
		}

		ReduceSymbolAllowInst = 0;
	}
	else if (c == 0)
	{
		n->type = QU_AST_literal;
		n->thing = x;
		n->child = NULL;
	}
	else
	{
		qu_var v;
		qu_any type = 0;

		if (c)
		{
			c->type = QU_AST_symbol;
			reduce_node (c);
			type = c->thing;
		}

		v = qu_var_new ((qu_symbol) CoModule, qu_string_byte ('*'), QU_VAR_FLAG_PRIVATE | QU_VAR_FLAG_FROZEN);
		v->value = x;
		v->type = type;

		if (c)
			qu_flag_set (v, QU_VAR_FLAG_TYPEOF);

		n->type = QU_AST_literal;
		n->thing = v;
		n->child = NULL;
	}
}



/*----------------------------------------------------------------------------

	ALIAS -> thing = name
	|
	SYMBOL

	Action
	------
	Turn SYMBOL to LITERAL.

	Result
	------
		ALIAS
			==> LITERAL -> thing = symbol object

----------------------------------------------------------------------------*/

static void reduce_alias (qu_ast n)
{
	qu_string name = n->thing;
	qu_ast    c = n->child;
	qu_any    x;

	ReduceSymbolAllowInst = 1;
	reduce_node (c);
	ReduceSymbolAllowInst = 0;

	if (qu_module_is (CoMaster))
	{
		if (c->type != QU_AST_literal)
			CTHROW (n, "undeclared symbol");

		x = c->thing;
		switch (qu_classid (x))
		{
			case QU_CLASS_ID_Class:
				break;

			case QU_CLASS_ID_Sub:
				if (qu_sub_is_class (x) || qu_sub_is_inst (x) || qu_sub_is_local (x))
					CTHROW (n, "cant alias a local or class attribute from module level");
				break;

			case QU_CLASS_ID_Var:
				if (qu_var_is_class (x) || qu_var_is_inst (x) || qu_var_is_lexical (x))
					CTHROW (n, "cant alias a local or class attribute from module level");
				break;

			default:
				CTHROW (n, "invalid alias");
		}
		qu_module_setattr ((qu_module) CoMaster, name, (qu_symbol) x);
		n->type = QU_AST_nop;
		n->child = NULL;
	}
	else
	{
		assert (qu_class_is (CoMaster));
		/* darn! must wait for qu_class_enddecl, etc. -> see _emit_class */
	}
}



/*----------------------------------------------------------------------------

	ARRAY -> thing = 0:variable, otherwise:constant
	         op    = number of items

	Action
	------
	If constant, make a literal.

	Result
	------
	The node may be changed to LITERAL.

		ARRAY
			==> LITERAL -> thing = qu_array object

----------------------------------------------------------------------------*/

static void reduce_array (qu_ast n)
{

	if (n->thing)
	{
		qu_array a;

		if (n->op)
		{
			qu_ast c;

			reduce_child_nodes (n);
			reduce_ensure_literal_childs (n);

#ifdef QU_APP_WEAKMODULE
			a = qu_array_new (n->op);
#else
			if (qu_module_is_resident (CoModule))
				a = qu_array_new_static (n->op);
			else
				a = qu_array_new (n->op);
#endif

			for (c = n->child; c; c = c->next)
				qu_array_push (a, c->thing);

			qu_array_freeze (a);

#ifdef QU_APP_WEAKMODULE
			qu_set_add (CoLiterals, a);
#else
			if (!qu_module_is_resident (CoModule))
				qu_set_add (CoLiterals, a);
#endif
		}
		else
		{
			a = QuArrayEmpty;
		}

		n->type  = QU_AST_literal;
		n->thing = a;
		n->child = 0;
	}
	else
	{
		reduce_child_nodes (n);
	}
}




/*----------------------------------------------------------------------------

	ASSERT -> thing = string expression
	  |
	 expr

	Action
	------
	Turn into NOP if __debug__ is false.

----------------------------------------------------------------------------*/

static void reduce_assert (qu_ast n)
{
	if (QU_APP_DEBUG)
	{
		reduce_node (n->child);
		n->thing = qu_set_addif (CoLiterals, n->thing);
	}
	else
	{
		n->type = QU_AST_nop;
	}
}



/*----------------------------------------------------------------------------

	ASSIGN
	 /  \
	a    b

	Action
	------
	Auto-create variable
	If 'a' is a symbol, make sure it is assignable

		var x = 10  <-- x created by Compile.c
		x = 10      <-- x auto-created here

	Result
	------
	Symbol in 'a' are resolved.
	'a' is assignable.

	Note
	----
	Resolve the left side of the assignment first.

----------------------------------------------------------------------------*/

/* make sure the thing is assignable */
static void reduce_ensure_assignable (qu_ast n)
{
	qu_any x = n->thing;

	if (qu_var_is (x))
	{
		if (qu_var_frozen (x))
			CTHROW (n, "cant modify constant");
	}
	else if (qu_sub_is (x))
	{
#ifndef QU_SUB_ALLOW_ASSIGN
		CTHROW (n, "invalid assignment");
#endif
	}
	else
	{
		CTHROW (n, "invalid assignment");
	}

	if (qu_symbol_is_inst (x))
		CTHROW (n, "cant access instance attribute directly");
}



static void reduce_assign_name (qu_ast n)
{
	qu_string name = (qu_string) n->thing;
	qu_any    x    = qu_symbol_find (CoMaster, name);

	if (x == 0)
		x = reduce_auto_local_create (name, n);

	assert (qu_module_is (x) || qu_class_is (x) || qu_sub_is (x) || qu_var_is (x));

	n->thing = x;
	n->op    = 0; /* no autocall for assignment */

	reduce_ensure_assignable (n);
}



/* QU_APP_MORE_COMPILE_CHECK */
static void reduce_setattr_check_rdonly_property (qu_ast a)
{
	/*
		check for attempt to set read-only property from outside of
		the class or setting class constant via variable

		TODO - from self
	*/

	qu_any x = 0;

	if (a->child->type == QU_AST_self)
	{
		return;
	}

	if (a->child->thing && a->child->next->type == QU_AST_literal)
	{
		qu_any    c    = a->child->thing;
		qu_string name = a->child->next->thing;

		if (qu_class_is (c))
		{
			x = qu_class_search_attr (x, name);
			if (x && qu_var_is (x))
				goto check;
		}
		else if (qu_var_is (c) && !qu_var_is_rdwr (c))
		{
			qu_var v = (qu_var) c;
			if (v->type && qu_class_is (v->type))
			{
				x = qu_class_search_attr (v->type, name);
				if (x && qu_var_is (x))
					goto check;
			}
		}
	}
	return;

  check:

	if (qu_var_frozen (x))
	{
		CTHROW (a, "cant modify constant");
	}
	else if (!qu_var_is_rdwr (x))
	{
		qu_class owner  = qu_var_get_class (x);
		qu_class setter = 0;

		if (owner && owner != (qu_class) CoMaster)
		{
			if (qu_sub_is (CoMaster))
				setter = qu_sub_get_class ((qu_sub) CoMaster);
			else if (qu_class_is (CoMaster))
				setter = (qu_class) CoMaster;

			if (setter == 0)
				/* this includes from module-level, i.e. qu_module_is (CoMaster) */
				CTHROW (a, "cant set read-only property from here");

			if (!qu_class_isa (setter, owner))
				CTHROW (a, "cant set read-only property");
		}
	}
}


static void reduce_assign (qu_ast n)
{
	qu_ast a = n->child;
	qu_ast b = a->next;
	qu_ast c;

	switch (a->type)
	{
		case QU_AST_literal:
			CTHROW (a, "invalid assignment");

		case QU_AST_name:
			reduce_assign_name (a);
			break;

		case QU_AST_list:
			if (n->op)
				CTHROW (n, "can only use the '=' operator with list assignment");

			for (c = a->child; c; c = c->next)
			{
				switch (c->type)
				{
					case QU_AST_literal:
						if (c->thing != QuNil)
							CTHROW (c, "invalid assignment");
						continue;

					case QU_AST_name:
						reduce_assign_name (c);
						continue;

					case QU_AST_eat:
						reduce_eat (c);
						continue;

					default:
						reduce_node (c);
				}
			}
			break;

		default:
			reduce_node (a);

			if (a->type == QU_AST_name)
				reduce_ensure_assignable (a);

			else if (QU_APP_MORE_COMPILE_CHECK && a->type == QU_AST_getattr)
				reduce_setattr_check_rdonly_property (a);
	}

	if (b->type == QU_AST_list)
	{
		for (c = b->child; c; c = c->next)
		{
			if (c->type == QU_AST_eat)
				CTHROW (n, "cant use '*' on the right side of an assignment");
		}
	}

	reduce_node (b);
}



/*----------------------------------------------------------------------------

	BINOP -> op = qu_sys_plus | qu_sys_mul | etc.
	 / \
	a   b

	Action
	------
	Constant folding.

	Result
	------
	The node may be changed to LITERAL or CAT

		BINOP
			==> LITERAL -> thing = the result

		BINOP
			==> CAT -> op = number of items

----------------------------------------------------------------------------*/

static void reduce_binop (qu_ast n)
{
	qu_ast     a = n->child;
	qu_ast     b = a->next;
	qu_binop_f f = (qu_binop_f) n->op;

	reduce_node (a);
	reduce_node (b);

	if (a->type == QU_AST_literal
		&& b->type == QU_AST_literal
		&& code_binop_literal_return_literal (f))
	{
		qu_any r = f (a->thing, b->thing);

		if (r == 0)
			CRETHROW (n);

		if (!qu_gc_is_static (r))
		{
			/* must use qu_set_addif_ex otherwise storing, e.g. a Big and a Float,
				which are equal will only store one of them */
			r = qu_set_addif_ex (CoLiterals, r);
		}

		n->type  = QU_AST_literal;
		n->thing = r;
		n->child = 0;
		n->op    = 0;
	}
	else if (f == (qu_binop_f) qu_sys_concat
		&& ((a->type == QU_AST_binop && (qu_binop_f) a->op == f)
			|| (b->type == QU_AST_binop && (qu_binop_f) b->op == f)))
	{
		qu_ast c;
		int    l = 2;

		/*
			Note - do not do this sort of thing for other operators.

			a() _ b() _ c() will only fail if there is not enough memory.
			Following Qu's optimistic memory strategy policy, then prematurely
			evaluating c is justified. Incorrect but justified.

			a() + b() + c() can fail at any point, e.g. after a() + b()
			so prematurely evaluating c() is incorrect.
		*/

		while (a->type == QU_AST_binop && (qu_binop_f) a->op == f)
		{
			a = a->child;
			n->child = a;
			c = a->next;
			c->next = b;
			b = c;
			l++;
		}
		a = n->child;
		for (c = a->next; c; a = c, c = c->next)
		{
			while (c->type == QU_AST_binop && (qu_binop_f) c->op == f)
			{
				b = c->child;
				a->next = b;
				b->next->next = c->next;
				l++;
				c = b;
			}
		}
		n->type  = QU_AST_cat;
		n->op = l;
	}
}



/*----------------------------------------------------------------------------

	CATCH ->op = number of exceptions
	|
	SYMBOL* - COMPOUND

	SYMBOL -> thing = name

	Action
	------
	Resolve all SYMBOL nodes

	Result
	------
	SYMBOL nodes are gone. The exceptions are store in the 'thing' field.

		CATCH
			==> CATCH -> op = number of exceptions
			    |     -> thing = exception to catch or [exception to catch, local variable]
			    |
			    COMPOUND

	The 'op' field will be:
	- 0 : 'thing' is exception
	- 1 : 'thing' is Array [exception, variable to set]

	exception is one of the following values:
	- nil   -> this is the default exception
	- Class -> one exception class
	- Array -> array of exception classes


----------------------------------------------------------------------------*/

static void reduce_catch (qu_ast n)
{
	qu_ast c;
	int    k = 0;
	qu_any local = n->thing;

	if (local)
	{
		reduce_auto_local_node (local, n);
		local = n->thing;
		assert (qu_var_is (local));
		if (((qu_var) local)->type)
			CTHROW (n, "cant use typed variable with catch");
	}

	reduce_child_nodes (n);

	for (c = n->child; c->type == QU_AST_literal; c = c->next)
	{
		reduce_ensure_literal_exception (c);
		k++;
	}

	if (k == 0)
	{
		n->thing = QuNil;
	}
	else if (k == 1)
	{
		c = n->child;
		n->thing = c->thing;
		n->child = c->next;
	}
	else
	{
		qu_array a = qu_array_new (k);

		for (c = n->child; c && c->type == QU_AST_literal; c = c->next)
		{
			qu_array_push (a, c->thing);
		}
		n->child = c;
		n->thing = a;

		qu_array_freeze (a);
		qu_set_add (CoLiterals, a);
	}

	n->op = 0;
	if (local)
	{
		n->op = 1;
		n->thing = qu_array_new_ex (2, n->thing, local);
	}
}



/*----------------------------------------------------------------------------

	CLASS -> thing = class object
	|
	[ISA] - MIXIN* - COMPOUND

	ISA -> thing = name      MIXIN -> thing = name
	|                        |
	[NAME]                   [NAME]

	Action
	------
	Resolve ISA and MIXIN.
	Call qu_class_set_super and qu_class_set_mixin.

	Result
	------
	The ISA and MIXIN nodes are removed.

		CLASS
			==> CLASS -> thing = class object
			    |
			    COMPOUND

----------------------------------------------------------------------------*/

static void reduce_class (qu_ast n)
{
	qu_ast c;

	assert (qu_class_is (n->thing));

	code_enter_master (n->thing);

	reduce_child_nodes (n);

	c = n->child;
	if (c->type == QU_AST_isa)
	{
		assert (qu_class_is (c->thing));
		if (qu_class_set_super ((qu_class) CoMaster, c->thing) == 0)
			CRETHROW (c);

		c = c->next;
		n->child = c;
	}

	while (c && c->type == QU_AST_mixin)
	{
		assert (qu_class_is (c->thing));

		if (qu_class_set_mixin ((qu_class) CoMaster, c->thing) == 0)
			CRETHROW (c);

		c = c->next;
		n->child = c;
	}

	code_leave_master ();
}



/*----------------------------------------------------------------------------

	const foo = const [1, 2, 3]

	CONST -> thing = Var object
	 |
	 a

	Action
	------
	Build a constant.

	Result
	------
	The node is changed to NOP.

		CONST
			==> NOP

----------------------------------------------------------------------------*/

static void reduce_const (qu_ast n)
{
	qu_ast c = n->child;
	qu_var v;

	reduce_node (c);

	if (!(c->type == QU_AST_literal || c->type == QU_AST_ref))
		CTHROW (c, "value is not a constant");

	v = n->thing;
	v->value = c->thing;
	qu_flag_set (v, QU_VAR_FLAG_FROZEN);

	n->type  = QU_AST_nop;
	n->child = 0;
}



/*----------------------------------------------------------------------------

	DICT -> thing = 0:variable, otherwise:constant
	        op    = number of items  (8 means 4 key/value pairs)

	Action
	------
	If constant, make a literal.

	Result
	------
	The node may be changed to LITERAL.

		DICT
			==> LITERAL -> thing = Dict object

----------------------------------------------------------------------------*/

static void reduce_dict (qu_ast n)
{
	if (n->thing)
	{
		qu_dict a;

		if (n->op)
		{
			qu_ast c;

			reduce_child_nodes (n);
			reduce_ensure_literal_childs (n);

#ifdef QU_APP_WEAKMODULE
			a = qu_dict_new (n->op / 2);
#else
			if (qu_module_is_resident (CoModule))
				a = qu_dict_new_static (n->op / 2);
			else
				a = qu_dict_new (n->op / 2);
#endif

			for (c = n->child; c; c = c->next->next)
				qu_dict_setitem (a, c->thing, c->next->thing);

			qu_dict_freeze (a);

#ifdef QU_APP_WEAKMODULE
			qu_set_add (CoLiterals, a);
#else
			if (!qu_module_is_resident (CoModule))
				qu_set_add (CoLiterals, a);
#endif
		}
		else
		{
			a = QuDictEMPTY;
		}

		n->type  = QU_AST_literal;
		n->thing = a;
		n->child = 0;
	}
	else
	{
		reduce_child_nodes (n);
	}
}



/*----------------------------------------------------------------------------

	This node is found in list assignment and for each list.
	Thus always in a LIST node.

	EAT  -> thing = name

	Action
	------
	Resolve symbol.
	Automatically create a local variable if the symbol is not found.

	Result
	------
	The 'thing' field will contain the local Var object.

		EAT
			==> EAT -> thing = local Var object

----------------------------------------------------------------------------*/

static void reduce_eat (qu_ast n)
{
	qu_var v = reduce_auto_local_create (n->thing, n);
	n->thing = v;
}



/*----------------------------------------------------------------------------

	enum

	ENUM -> thing = Var object
	|       op = number of items
	|
	ENUMITEM+

	ENUMITEM -> thing = Var
	|
	[expr]

	Action
	------
	Create constants.

	Result
	------
	The enum node is turned into a NOP.

----------------------------------------------------------------------------*/

static void reduce_enum (qu_ast n)
{
	qu_ast  c, g;
	qu_dict d;
	qu_var  v, w;
	qu_any  val = QuInt0;

	for (c = n->child; c; c = c->next)
	{
		assert (c->type = QU_AST_enumitem);
		w = c->thing;
		g = c->child;
		if (g)
		{
			reduce_node (g);

			if (g->type != QU_AST_literal)
				CTHROW (g, "non-literal in enum");

			w->value = g->thing;
			g->thing = 0;

			if (qu_int_is (w->value))
			{
				val = qu_sys_incr (w->value);
				assert (val);
			}
		}
		else
		{
			qu_var_freeze (w);
			w->value = val;
			val = qu_sys_incr (val);
			assert (val);
		}
		c->child = 0;
	}

	v = (qu_var) n->thing;
	assert (qu_var_is (v));

	d = qu_dict_new (n->op);
	v->value = d;
	qu_flag_set (v, QU_VAR_FLAG_ENUM);
	qu_var_freeze (v);

	for (c = n->child; c; c = c->next)
	{
		w = c->thing;
		qu_var_freeze (w);
		qu_flag_set (w, QU_VAR_FLAG_ENUMITEM);
		qu_dict_setitem (d, w->value, w->name);
	}
	qu_dict_freeze (d);

	n->type  = QU_AST_nop;
	n->child = 0;
	n->thing = 0;
}



/*----------------------------------------------------------------------------

	EXTERN -> thing = [module name, function name]

	Action
	------
	Set thing = [Module object, function name]

----------------------------------------------------------------------------*/

static void reduce_extern (qu_ast n)
{
	qu_array a = n->thing;
	qu_any   x = a->objs[0];

	if (qu_string_is (x))
	{
		x = qu_symbol_find (CoMaster, a->objs[0]);
		if (x == 0)
			CTHROW (n, "undeclared symbol");

		if (!qu_module_is (x))
			CTHROW (n, "symbol not a module");
	}

	assert (qu_module_is (x));

	/* yet another documentation hack */
	if (!qu_module_is_doc (CoModule))
	{
		if (((qu_module) x)->getcfunc == 0)
			CTHROW (n, "module does not support __extern");
	}
	a->objs[0] = x;
}



/*----------------------------------------------------------------------------

	FOR
	|
	(NAME | LIST) - expr - COMPOUND
			|
	        NAME+ - [EAT]

	NAME -> thing = name
	EAT  -> thing = name

	Action
	------
	Make sure NAME/EAT are local variables. If it does not exist in the
	current block then automatically create a local variable.

	sub foo
		local x           # x created by Compile.c
		for y, z* in ...  # y z auto-created here
					...

	Magic
	-----
	If the name is '__' then this always means the global variable defined by
	the builtin Sys module.

		for __ in ...

	Result
	------
	The 'thing' field will be the local Var object.

		NAME
			==> NAME -> thing = local Var object

		EAT
			==> EAT -> thing = local Var object

	The 'expr' node might be turned into a LIST node.

		for x in Range (foo, bar, baz)
			-> for x in (foo, bar, baz)

		for x in foo .. bar
			-> for x in (foo, bar)

----------------------------------------------------------------------------*/

static void reduce_for (qu_ast n)
{
	qu_ast p = n;
	qu_ast c = n->child;

	if (c->type == QU_AST_list)
		p = c;

	if (c->next->type == QU_AST_list
		|| (c->next->type == QU_AST_binop && (qu_binop_f) c->next->op == (qu_binop_f) qu_sys_range))
	{
		if (c->type == QU_AST_list)
			CTHROW (c, "cant use list for ranges");
	}

	for (c = p->child; c; c = c->next)
	{
		qu_string s;
		qu_any    x = 0;

		assert (c->type == QU_AST_name || c->type == QU_AST_eat);
		s = c->thing;

		if (CoMaster->attrs)
			x = qu_attr_find (CoMaster->attrs, s);

		if (x == 0)
		{
			if (s->len == 2 && s->str[0] == '_' && s->str[1] == '_')
				x = QuSys__ignore;
			else
				x = reduce_auto_local_create (s, n);
		}
		else
		{
			if (!qu_var_is (x))
				CTHROW (n, "must use variable");

			if (qu_var_frozen (x))
				CTHROW (n, "cant modify constant");

			if (!(qu_var_is_local (x) || qu_var_is_lexical (x)))
				CTHROW (n, "must only use local variable");
		}

		c->thing = x;

		if (p == n)
			break;
	}

	c = n->child->next;

	if (c->type == QU_AST_binop && (qu_binop_f) c->op == (qu_binop_f) qu_sys_range)
	{
		reduce_child_nodes (c);
		c->type = QU_AST_list;
		c->op = 2;
		reduce_node (c->next);
	}
	else
	{
		reduce_nodes_from (c);
		if (c->type == QU_AST_call && c->child->type == QU_AST_name && c->child->thing == QuRangeClass)
		{
			if (c->op > 1)
			{
				c->type = QU_AST_list;
				c->child = c->child->next;
			}
		}
	}
}



/*----------------------------------------------------------------------------

	GETATTR -> op = 1: followed by call(), 0:not
	 /   \
	a     b

	Action
	------
	Try to resolve symbols.

	Result
	------
	The node may be changed to NAME or LITERAL.

		GETATTR
			==> NAME -> thing = Class, Sub or Var object
			            op    = 1: followed by call(), 0:not

			==> LITERAL -> thing = value

----------------------------------------------------------------------------*/

static void reduce_getattr (qu_ast n)
{
	qu_ast    a = n->child;
	qu_ast    b = a->next;
	qu_string s;
	qu_any    x;

#ifdef QU_HAS_UNIVERSE
	if (a->type == QU_AST_name
		&& (b->type == QU_AST_name
			|| (b->type == QU_AST_literal && qu_string_is (b->thing))))
	{
		qu_array u = CoModule->universes;
		if (u)
		{
			if (qu_array_contains (u, a->thing))
			{
				qu_string name = b->thing;
				qu_set    set  = qu_dict_getitem (QuUniverses, a->thing);

				if (set)
				{
					long      b = 0, i = 0;
					qu_module m;

					while ((m = qu_set_walk (set, &b, &i)) != NULL)
					{
						if (qu_module_getattr (m, name))
						{
							a->thing = m;
							a->type = QU_AST_literal;
							goto ufound;
						}
					}
				}
			}
		}
	}
#endif

	reduce_node (a);
	reduce_node (b);

#ifdef QU_HAS_UNIVERSE
  ufound:
#endif

	if (b->type != QU_AST_literal || a->type != QU_AST_name)
		return;

	s = b->thing;
	if (!qu_string_is (s))
		CTHROW (b, "attribute must be a string or identifier");

	x = a->thing;

	if (qu_module_is (x))
	{
		x = qu_module_getattr (x, s);
		if (x == 0)
		{
			/* Foo module has Class named Foo -> Foo.Foo.bar -> Foo.bar
				provided that bar is not an instance attribute */

			x = qu_module_getattr_callable_name (a->thing, 0);
			if (x && qu_class_is (x))
			{
#if 0
				/* this is a problem so don't do it:

					# module Color
					class Color is Pointer

					If we resolve the method, say "array" into a Sub then it will be Pointer.array.
					When the interpreter calls it, the class object will be Pointer instead of
					Color. Which is wrong.
				*/
				x = qu_class_search_attr (x, s);
				if (x)
				{
					if (!qu_symbol_is_inst (x))
						goto okay;
				}
#else
				a->thing = x;
				a->child = 0;
				return;
#endif
			}

			x = qu_attr_find (QuModuleClass->attrs, s);
			/* stay GETATTR for run-time */
			if (x == 0)
				return;
		}

#if 0
	  okay:
#endif
		assert (qu_module_is (x) || qu_class_is (x) || qu_sub_is (x) || qu_var_is (x));

		if (qu_symbol_is_local (x) || qu_symbol_is_private (x))
			CTHROW (n, "no such attribute");

		/* if instance method of the qu_module class then leave it to runtime */
		if (qu_symbol_is_inst (x))
			return;

		/*
			sub remote
			;;

			class Foo
				const x = @remote
			;;

			Foo.x -> optimizing this will not call remote so don't do it.
		*/

		if (qu_var_is (x) && qu_var_frozen (x)
			&& !(qu_sub_is (((qu_var) x)->value) || qu_method_is (((qu_var) x)->value)))
		{
			n->type  = QU_AST_literal;
			n->thing = ((qu_var) x)->value;
			n->child = NULL;
		}
		else
		{
			n->type  = QU_AST_name;
			n->thing = x;
			n->child = NULL;
		}
	}
	else if (qu_class_is (x))
	{
		x = qu_class_search_attr (x, s);
		if (x == 0)
		{
			/* might be accessing from a class which is not yet completely
				declared (qu_class_enddecl not yet called), so leave it to run-time */
			return;

		  nope:
			CTHROW (a, "class has no such public attribute");
		}

		if (qu_symbol_is_inst (x))
			CTHROW (a, "cant access instance attribute directly");

		if (qu_var_is (x))
		{
			if (!qu_var_can_access (x, CoMaster))
				goto nope;
		}
		else
		{
			assert (qu_sub_is (x));
			if (!qu_sub_can_access (x, CoMaster))
				goto nope;
		}

#if 0
		if (qu_sub_is (x))
		{
			n->type  = QU_AST_name;
			n->thing = x;
			n->child = NULL;
		}
		else
#endif
		if (qu_var_is (x) && qu_var_frozen (x)
			&& !(qu_sub_is (((qu_var) x)->value) || qu_method_is (((qu_var) x)->value)))
		{
			n->type  = QU_AST_literal;
			n->thing = ((qu_var) x)->value;
			n->child = NULL;
		}
	}
}



/*----------------------------------------------------------------------------

	(1) IMPORT -> thing = Array of Module objects (no symbol import)

		This form comes from a use statement where the path is a directory,
		thus use all modules found in that directory. It can not import
		symbols, i.e. accessing symbols from a module must use the module
		prefix.

	(2) IMPORT -> thing = Module object
			op = 0, nothing to import
			op = 1, import all

		This form comes from a use statement where the path is a module.
		It either import all symbols or none at all.

		use foo (*)
		use foo

	(3) IMPORT -> thing = Module object
		|
		NAME+ -> thing = symbol name (String) to import

		This form comes from a use statement where the path is a module
		with a list of symbol names to import.

	Action
	------
	Import the symbols.

	Result
	------
	No nodes are modified because we need the IMPORT nodes in Emit to produce
	QU_PC_INITMOD.

----------------------------------------------------------------------------*/

static void reduce_import (qu_ast n)
{
	qu_ast c = n->child;
	qu_any u = n->thing;

	if (qu_array_is (u))
		return;

	assert (qu_module_is (u));
	if (c)
	{
		for (; c; c = c->next)
		{
			assert (c->type == QU_AST_name);
			if (qu_module_import (CoModule, u, c->thing) == 0)
				CRETHROW (c);
		}
	}
	else
	{
		if (n->op)
		{
			if (qu_module_import_all (CoModule, u) == 0)
				CRETHROW (n);
		}
	}
}



/*----------------------------------------------------------------------------

	IF
	|
	expr - COMPOUND -  [NOP | IF*] - [COMPOUND (else)]
	                    |
	                    expr - COMPOUND

	                    !!! IF might be turned into a NOP
	Action
	------
	Change into NOP if possible.

	Result
	------
	The IF node may be changed into NOP or COMPOUND

		IF
			==> NOP | COMPOUND
----------------------------------------------------------------------------*/

static void reduce_if (qu_ast n)
{
	qu_ast c = n->child;

	reduce_nodes_from (c);

  again:

	c = n->child;
	if (c->type == QU_AST_literal)
	{
		if (qu_sys_true (c->thing) == QuTrue)
		{
			assert (c->next->type == QU_AST_compound);
			n->type = QU_AST_compound;
			n->child = c->next->child;
			/*reduce_node (n);*/
			return;
		}
		c = c->next->next;
		if (c == 0)
		{
			n->type  = QU_AST_nop;
			n->child = 0;
			return;
		}

		if (c->type == QU_AST_compound)
		{
			n->type = QU_AST_compound;
			n->child = c->child;
			/* reduce_node (n); */
			return;
		}
		else if (c->type == QU_AST_nop)
		{
			while (c && c->type == QU_AST_nop)
				c = c->next;

			if (c == 0)
			{
				n->type = QU_AST_nop;
				n->child = 0;
				return;
			}
			else if (c->type == QU_AST_compound)
			{
				n->type = QU_AST_compound;
				n->child = c->child;
				return;
			}
		}

		assert (c->type == QU_AST_if);

		n->child  = c->child;
		n->cursor = c->cursor;
		n->line   = c->line;
		c->child->next->next = c->next;
		goto again;
	}
	/* reduce_nodes_from (c->next); */
}



/*----------------------------------------------------------------------------

	ISA    -> thing = name
	|
	[NAME]


	Action
	------
	Resolve symbol.

	Result
	------
	The 'thing' field will contain a Class object

		ISA
			==> ISA -> thing = Class object

	Magic
	-----
	If a module defines a class named as the name of the module then we
	perform some magic:

	# module Foo
		class Foo
			...
		;;

	# module Bar
		use Foo
		class Bar is Foo  #--> we're smart enough to know that this can
			...           #    only mean: class Bar is Foo.Foo
		;;

----------------------------------------------------------------------------*/

static void reduce_isa (qu_ast n)
{
	qu_any x = qu_symbol_find (CoMaster, n->thing);
	qu_ast c;

	if (x == 0)
		CTHROW (n, "undeclared symbol");

	c = n->child;
	if (c)
	{
		assert (c->type == QU_AST_name);

		if (!qu_module_is (x))
			CTHROW (n, "module name expected");

		x = qu_module_getattr (x, c->thing);
		if (x == 0)
			CTHROW (c, "no such symbol");
	}

	if (!qu_class_is (x))
	{
		if (qu_module_is (x))
		{
			x = qu_module_getattr (x, qu_file_strip (qu_file_name (((qu_module) x)->name)));
			if (x == 0)
				CTHROW (n, "not such symbol");

			if (!qu_class_is (x))
				CTHROW (n, "not a class");
		}
		else
		{
			CTHROW (n, "not a class");
		}
	}
	n->thing = x;
}



/*----------------------------------------------------------------------------

	LOCAL -> thing = Var object
	|        op = 1:has SYMBOL 0:not
	|
	[SYMBOL] - [EXPR]


	Action
	------
	Resolve symbol.

	Result
	------
	The SYMBOL node, if present, is removed and var->type set.
	If EXPR is not present, the node is changed into a NOP.

		LOCAL
			==> LOCAL thing = Var object
			    |
			    EXPR


		LOCAL
			==> NOP

----------------------------------------------------------------------------*/

static void reduce_local (qu_ast n)
{
	qu_ast c = n->child;

#if 1
	if (c && c->type == QU_AST_ref && c->op)
	{
		/*EXPERIMENTAL - currying type

			sub foo
				local v: @bar (1, 2)

			VAR
			|
			REF -> op = 1
		*/

		qu_var v = n->thing;
		qu_any type;

		assert (qu_var_is (v));

		reduce_node (c);
		assert (c->type == QU_AST_literal);

		type = c->thing;
		assert (qu_method_is (type));

		/* check anyway */
		if (!qu_object_valid_type (type))
			CTHROW (c, "invalid type");

		v->type = type;
		c = c->next;
		n->child = c;

		if (c)
			reduce_nodes_from (c);
	}
	else
	{
#endif

	assert (c);
	reduce_child_nodes (n);

	if (n->op)
	{
		qu_var v = n->thing;
		qu_any type;

		assert (qu_var_is (v));
		assert (c->type == QU_AST_literal);
		type = c->thing;
		if (!qu_object_valid_type (type))
			CTHROW (c, "invalid type");
		v->type = qu_object_fix_type (type);

		n->op = 0;
		n->child = c->next;
		c->next = NULL;

		c = n->child;
		if (c && c->type == QU_AST_literal)
		{
			if (!qu_object_check_value (v->type, c->thing))
				CTHROW (c, "invalid value");
		}
	}
#if 1
	}
#endif
	if (c == 0)
	{
		n->type = QU_AST_nop;
		n->thing = 0;
	}
}



/*----------------------------------------------------------------------------

	LOOP
	|
	[expr] - COMPOUND - [expr]
	                    (until)

	Action
	------
	Optimize.

	Result
	------
	The LOOP node may be changed.

		LOOP
			==> NOP

			==> LOOP
				|
				COMPOUND

				LOOP
				|
				expr - COMPOUND

----------------------------------------------------------------------------*/

static void reduce_loop (qu_ast n)
{
	qu_ast c, u;

	reduce_child_nodes (n);

	c = n->child;
	if (c->type != QU_AST_compound)
	{
		if (c->type == QU_AST_literal)
		{
			if (qu_sys_true (c->thing) == QuTrue)
			{
				c = c->next;
				n->child = c;
			}
			else
			{
				n->type  = QU_AST_nop;
				n->child = 0;
				return;
			}
		}
	}

	u = c->next;
	if (u && u->type == QU_AST_literal)
	{
		if (qu_sys_true (u->thing) == QuTrue)
			u->thing = QuTrue;
		else
			c->next = 0;
	}
}



/*----------------------------------------------------------------------------

	MODULE
	|
	...

	Reduce all nodes.
----------------------------------------------------------------------------*/

static void reduce_module (qu_ast n)
{
	ReduceSymbolAllowInst = 0;
	reduce_child_nodes (n);
}



/*----------------------------------------------------------------------------

	NAME -> thing = name
	        op    = 1:followed by call  0:not

	Action
	------
	Resolve symbol.

	Result
	------
	If the symbol is a constant qu_var then NAME is changed to LITERAL.

		NAME
			==> LITERAL -> thing = value

	Otherwise the 'thing' field is the symbol object.

		NAME
			==> NAME -> thing = symbol object (Module, Class, Sub or Var)
			            op    = 1:followed by call  0:not

----------------------------------------------------------------------------*/

static void reduce_name (qu_ast n)
{
	qu_string name = (qu_string) n->thing;
	qu_any    x    = qu_symbol_find (CoMaster, name);

	if (x == 0)
	{
		x = qu_attr_find (QuSysModule->attrs, name);
		if (x == 0)
		{
			if (!qu_module_is_doc (CoModule))
				CTHROW (n, "undeclared symbol");

			/*HACK - documentation

				If documenting and the symbol is not declared then we point to
				Sys.__undocumented. This is to avoid problems when a symbol is
				not declared until run-time, for example, a module creates a new
				attribute in the Sys module such as done by lib/Kata/System.c.
			*/
			x = __QuSys__undocumented__;
		}
	}

	assert (qu_module_is (x) || qu_class_is (x) || qu_sub_is (x) || qu_var_is (x));

	if ((qu_var_is (x) || qu_sub_is (x)) && qu_symbol_is_inst (x))
		CTHROW (n, "cant refer to instance attribute directly");

	if (qu_var_is (x) && qu_var_frozen (x))
	{
		n->type  = QU_AST_literal;
		n->thing = ((qu_var) x)->value;
	}
	else
	{
		n->thing = x;
	}
}



/*----------------------------------------------------------------------------

	See reduce_sub:

	PARAM -> thing = Var object
	|
	[SYMBOL] - [expr]

	Action
	------
	Resolve and remove SYMBOL.
	Set the type of the parameters.
	If expr is a LITERAL, make sure it conforms to the type.

	Result
	------
	The SYMBOL node is removed.
	The QU_SUB_FLAG_ARGCHECK flag for the Sub object may be set.

		PARAM
			==> PARAM -> thing = Var object
			    |
			    [expr]

----------------------------------------------------------------------------*/

static void reduce_param (qu_ast n)
{
	qu_ast c = n->child;

	if (c->type == QU_AST_symbol)
	{
		qu_var v = (qu_var) n->thing;
		qu_any type;

		assert (qu_var_is (v));

		reduce_type (c);
		assert (c->type == QU_AST_literal);
		type = c->thing;
		if (!qu_object_valid_type (type))
			CTHROW (c, "invalid type");

		v->type = qu_object_fix_type (type);

		c = c->next;
		n->child = c;

		if (c)
		{
			reduce_node (c);
			if (c->type == QU_AST_literal)
			{
				if (!qu_object_check_value (v->type, c->thing))
					CTHROW (c, "invalid default value");
			}
		}
		else
		{
			n->type = QU_AST_nop;
		}

		assert (qu_sub_is (CoMaster));
		qu_flag_set (CoMaster, QU_SUB_FLAG_ARGCHECK);
	}
	else
	{
		reduce_child_nodes (n);
	}
}



/*----------------------------------------------------------------------------

	REF -> thing = name
	|
	[NAME] -> thing = name

	Action
	------
	Resolve the symbol and make sure it is a valid reference.

	Result
	------
	The 'thing' field contains a Sub or Var object.

----------------------------------------------------------------------------*/

static void reduce_ref (qu_ast n)
{
#if 1
	if (n->child && n->child->type == QU_AST_list)
	{
		/*EXPERIMENTAL - currying

			sub foo ...

			x = @foo (1, 2, 3)
		*/

		qu_ast    curry = n->child;
		qu_method m;

		n->child = curry->next;

		n->type = QU_AST_symbol;
		reduce_symbol (n);
		assert (n->type == QU_AST_literal);

		if (curry->child)
			reduce_child_nodes (curry);

		m = qu_method_curry_compile_time (n->thing, curry);
		if (m == 0)
			CTHROW (n, "can't curry this");

		/* don't forget to store this so it gets GC'ed */
		qu_set_add (CoLiterals, m);

		n->type = QU_AST_literal;
		n->thing = m;
		n->child = 0;
	}
	else
#endif
	{
		if (n->child == 0 && n->op == 2)
		{
			/* @self.name */
			return;
		}
		else
		{
			qu_any x;

			n->type = QU_AST_symbol;
			reduce_symbol (n);
			assert (n->type == QU_AST_literal);
			n->type = QU_AST_ref;
			x = n->thing;

			if (qu_sub_is (x))
			{
				if (qu_sub_is_inst (x))
					CTHROW (n, "cant make reference to instance attribute");
			}
			else if (qu_var_is (x))
			{
				if (qu_var_is_inst (x) || qu_var_is_lexical (x) || qu_var_is_local (x))
					CTHROW (n, "cant make reference to instance property or local variable");
				else
				{
					if (CoMaster != (qu_symbol) qu_var_get_module (x) && (qu_var_is_private (x) || qu_var_is_local (x)))
						CTHROW (n, "cant make reference to private variable");
				}
			}
			else
			{
				CTHROW (n, "invalid reference");
			}
		}
	}
}



/*----------------------------------------------------------------------------

	QU_RETURN
	|
	[expr]

	Action
	------
	Check return type if possible.

	Result
	------
	Nothing changed.

----------------------------------------------------------------------------*/

static void reduce_return (qu_ast n)
{
	qu_ast c = n->child;
	qu_any type;

	assert (qu_sub_is (CoMaster));

	if (c)
		reduce_node (c);

	type = ((qu_sub) CoMaster)->type;
	if (type)
	{
		qu_any v;

		if (c)
		{
			/* TODO - should be able to do better than just testing literals */
			if (c->type != QU_AST_literal)
				return;

			v = c->thing;
		}
		else
		{
			v = QuNil;
		}

		if (!qu_object_check_value (type, v))
			CTHROW (n, "invalid return value");
	}
}



/*----------------------------------------------------------------------------

	SET -> thing = 0:variable, otherwise:constant
	       op    = number of items

	Action
	------
	If constant, make a literal.

	Result
	------
	The node may be changed to LITERAL.

		SET
			==> LITERAL -> thing = Set object

----------------------------------------------------------------------------*/

static void reduce_set (qu_ast n)
{
	if (n->thing)
	{
		qu_set a;

		if (n->op)
		{
			qu_ast c;

			reduce_child_nodes (n);
			reduce_ensure_literal_childs (n);

#ifdef QU_APP_WEAKMODULE
			a = qu_set_new (n->op);
#else
			if (qu_module_is_resident (CoModule))
				a = qu_set_new_static (n->op);
			else
				a = qu_set_new (n->op);
#endif

			for (c = n->child; c; c = c->next)
				qu_set_push (a, c->thing);

			qu_set_freeze (a);

#ifdef QU_APP_WEAKMODULE
			qu_set_add (CoLiterals, a);
#else
			if (!qu_module_is_resident (CoModule))
				qu_set_add (CoLiterals, a);
#endif
		}
		else
		{
			a = QuSetEMPTY;
		}

		n->type  = QU_AST_literal;
		n->thing = a;
		n->child = 0;
	}
	else
	{
		reduce_child_nodes (n);
	}
}



/*----------------------------------------------------------------------------

	STATIC -> thing = Var object
	|
	EXPR

	Action
	------
	Create a literal.

	Result
	------
	The node is changed into a NOP.

		STATIC
			==> NOP

----------------------------------------------------------------------------*/

static void reduce_static (qu_ast n)
{
	qu_var v = n->thing;
	qu_ast c;

	assert (qu_var_is (v));

	c = n->child;
	switch (c->type)
	{
		case QU_AST_array:
		{
			qu_array a;

			reduce_child_nodes (c);
			reduce_ensure_literal_childs (c);
			a = qu_array_new (c->op);
			for (c = c->child; c; c = c->next)
				qu_array_push (a, c->thing);
			v->value = a;
			break;
		}

		case QU_AST_dict:
		{
			qu_dict a;

			reduce_child_nodes (c);
			reduce_ensure_literal_childs (c);
			a = qu_dict_new (c->op);
			for (c = c->child; c; c = c->next)
			{
				qu_dict_setitem (a, c->thing, c->next->thing);
				c = c->next;
			}
			v->value = a;
			break;
		}

		case QU_AST_set:
		{
			qu_set a;

			reduce_child_nodes (c);
			reduce_ensure_literal_childs (c);
			a = qu_set_new (c->op);
			for (c = c->child; c; c = c->next)
				qu_set_push (a, c->thing);
			v->value = a;
			break;
		}

		case QU_AST_literal:
			v->value = c->thing;
			break;

		default:
			CTHROW (n, "cant assign non-literal value to static variable");
	}
	n->type  = QU_AST_nop;
	n->thing = 0;
	n->child = 0;
}



/*----------------------------------------------------------------------------

	STRUCT -> thing = Var object, op = number of fields
	|
	(STRING - SYMBOL)+

	Action
	------
	Create the qStruct object.

	Result
	------
	The node is changed into a NOP.

		STRUCT
			==> NOP

----------------------------------------------------------------------------*/

static void reduce_struct (qu_ast n)
{
	qu_var v = n->thing;
	qu_ast c;

	assert (qu_var_is (v));
	for (c = n->child; c; c = c->next)
	{
		c = c->next;
		assert (c->type = QU_AST_symbol);
		reduce_type (c);
	}
	if (qu_struct_compile_setup (v, n) == 0)
		CTHROW (n, "invalid struct");

	n->type  = QU_AST_nop;
	n->thing = 0;
	n->child = 0;
}



/*----------------------------------------------------------------------------

	SUB -> thing = Sub object or Array [Sub object, Ast symbol] (dotted sub)
	|
	[SYMBOL] - PARAM* - COMPOUND

	Action
	------
	Resolve and remove SYMBOL.
	Resolve everything in this Sub.
	Call qu_sub_reindex to reindex lexicals.

	Result
	------
	The SYMBOL node is removed.

		SUB
			==> SUB -> thing = Sub object
			    |
			    PARAM* - COMPOUND

----------------------------------------------------------------------------*/

/*EXPERIMENTAL - dotted sub */
static void reduce_dotted_sub (qu_ast n)
{
	/*
		sub Foo.Bar.baz (x)
			...
		;;

		We must do the replacement now otherwise later reference to this Sub
		will fail.

		The bad thing is: if later compilation failed then we will get a weird
		"cant call sub at compile time" exception when the sub is called.

		See _compile_jit -> this sub will not get jit-ed :-(
	*/

	qu_array a = n->thing;
	qu_sub   u = a->objs[0];
	qu_ast   y = a->objs[1];
	qu_sub   old;
	qu_ast   c;
	qu_class k;
	qu_any   x;

	assert (a->len == 2);
	assert (qu_sub_is (u));
	assert (qu_ast_is (y));

	x = qu_symbol_find (CoMaster, y->thing);
	if (x == 0)
		CTHROW (y, "no such symbol");

	if (qu_module_is (x))
	{
		qu_module m = x;

		c = y->child;
		if (c == 0)
			CTHROW (y, "identifier expected after module name");

		x = qu_module_getattr (m, c->thing);
		if (c->child)
		{
			if (x == 0)
				CTHROW (y, "no such symbol");
		}
		else
		{
			old = x;
			if (old)
			{
				if (!qu_sub_is (old))
					CTHROW (n, "cant override variable or class");

				if (qu_sub_is_final (old))
					CTHROW (n, "cant override more than once or final method");

				old->name = qu_string_cat (u->name, qu_string_byte ('_'), NULL);
				qu_attr_setex (m->attrs, old->name, old);
				qu_flag_set (u, QU_SUB_FLAG_FINAL);
			}
			qu_attr_setex (m->attrs, u->name, u);

			/* FIXME - perhaps we shouldn't move it in case of qu_module_is_doc */
			u->parent = (qu_symbol) m;

			n->thing = u;
			qu_flag_unset (u, QU_SUB_FLAG_INST);
			qu_flag_unset (u, QU_SUB_FLAG_CLASS);
			qu_flag_set   (u, QU_SUB_FLAG_STATIC | QU_SUB_FLAG_FINAL);
			return;
		}
	}

	if (!qu_class_is (x))
		CTHROW (y, "no such class");

	k = x;

	/*
		The old one, if any, will be suffixed with "_" 	so it can still be called.
		Problem is, it won't work more than once.

		new         old
		-------     --------
		String.foo  String.foo_

		suppose String.foo calls the old one, which has become String.foo_
		now let's replace it again

		new         old          older
		-------     --------     -----
		String.foo  String.foo_  gone!

		so String.foo_ will call String.foo_ which now is itself :-(
	*/

	old = qu_attr_find (k->attrs, u->name);
	if (old)
	{
		if (!qu_sub_is (old))
			CTHROW (n, "cant override property with method or vice versa");

		if (qu_sub_is_final (old))
			CTHROW (n, "cant override method");

		if (qu_sub_is_inst (old))
		{
			if (!qu_sub_is_inst (u))
				CTHROW (n, "method must be an instance  method");
		}
		else if (qu_sub_is_class (old))
		{
			if (!qu_sub_is_class (u))
				CTHROW (n, "method must be a class method");
		}
		else if (qu_sub_is_static (old))
		{
			if (!qu_sub_is_static (u))
				CTHROW (n, "method must be a static method");
		}

		old->name = qu_string_cat (u->name, qu_string_byte ('_'), NULL);
		qu_attr_setex (k->attrs, old->name, old);
		qu_flag_set (u, QU_SUB_FLAG_FINAL);
	}

	/* FIXME - perhaps we shouldn't move it in case of qu_module_is_doc */
	u->parent = (qu_symbol) k;

	/*
		must do this now otherwise later reference to this Sub will fail
		use qu_attr_setex anyway
	*/
	qu_attr_setex (k->attrs, u->name, u);

	/*
		must remove from accelerators :-(
		because accelerators must be C functions ...
	*/
	qu_class_del_accel (k, u->name);

	n->thing = u;
}



static void reduce_sub (qu_ast n)
{
	qu_ast c = n->child;
	qu_sub u;

	if (qu_array_is (n->thing))
		reduce_dotted_sub (n);

	u = n->thing;
	assert (qu_sub_is (u));

	if (c->type == QU_AST_symbol)
	{
		qu_any type;

		reduce_type (c);
		assert (c->type == QU_AST_literal);
		type = c->thing;
		if (!qu_object_valid_type (type))
			CTHROW (c, "invalid type");
		u->type = qu_object_fix_type (type);

		n->child = c->next;
		c->next = NULL;
	}

	code_enter_master  (u);
	reduce_child_nodes (n);

	if (u->root == u)
		qu_sub_reindex (u);

	/* so we can say u->nlocal instead of u->root->nlocal at run-time */
	u->nlocal = u->root->nlocal;
	code_leave_master ();
}



/*----------------------------------------------------------------------------

	SYMBOL -> thing = name
	|      -> op = 1:came from a type, 0:not
	[NAME]
	|
	[NAME]

	Action
	------
	Resolve symbol.

	Result
	------
	SYMBOL is changed into LITERAL.

		SYMBOL
			==> LITERAL -> thing = Module, Class, Sub or Var object


	!!! SYMBOL nodes are childs of other nodes.
	!!! Those nodes have their own reduce and emit code.
	!!! After Reduce, all SYMBOL nodes are changed to LITERAL.

	Nodes using SYMBOL:
	- ALIAS
	- SUB
	- PARAM
	- VAR
	- QU_THROW
	- CATCH
	- STRUCT

----------------------------------------------------------------------------*/

static void reduce_symbol (qu_ast n)
{
	qu_any x = qu_symbol_find (CoMaster, n->thing);
	qu_ast c = n->child;

	if (x == 0)
	{
#ifdef QU_HAS_UNIVERSE
		qu_array u = CoModule->universes;
		if (u && c && c->child == NULL && qu_array_contains (u, n->thing))
		{
			qu_string name = c->thing;
			qu_set    set  = qu_dict_getitem (QuUniverses, n->thing);

			if (set)
			{
				long      b = 0, i = 0;
				qu_module m;

				while ((m = qu_set_walk (set, &b, &i)) != NULL)
				{
					if (qu_module_getattr (m, name))
					{
						x = m;
						goto ufound;
					}
				}
			}
		}
#endif

		if (ReduceSymbolAllowInst)
			return;

		CTHROW (n, "undeclared symbol");
	}

#ifdef QU_HAS_UNIVERSE
  ufound:
#endif

	if (c)
	{
		qu_ast g = c->child;

		assert (c->type == QU_AST_name);

		if (qu_class_is (x))
		{
			if (g)
				goto nope;

			x = qu_class_search_attr ((qu_class) x, (qu_string) c->thing);
		}
		else if (qu_module_is (x))
		{
			if (g && g->child)
				goto nope;

			x = qu_module_getattr (x, c->thing);
			if (x && g)
			{
				if (!qu_class_is (x))
					goto nope;

				x = qu_class_search_attr ((qu_class) x, (qu_string) g->thing);
			}
		}
		else
		{
			CTHROW (n, "module or class name expected");
		}

		if (x == 0)
		{
		  nope:
			CTHROW (c, "undeclared symbol");
		}

		if (qu_var_is (x))
		{
			if (!qu_var_is_lexical (x))
			{
				if (!qu_var_can_access (x, CoMaster))
					CTHROW (n, "symbol not accessible from here");
			}
		}
		else if (qu_sub_is (x))
		{
			if (!qu_sub_can_access (x, CoMaster))
				CTHROW (n, "symbol not accessible from here");
		}
	}

	if (qu_var_is (x) || qu_sub_is (x))
	{
		if (!ReduceSymbolAllowInst)
		{
			if (qu_symbol_is_inst (x))
				CTHROW (n, "cant use instance attribute");
		}
	}

	n->type  = QU_AST_literal;
	n->thing = x;
	n->child = 0;
}



/*============================================================================

	SWITCH
	|       (default)
	expr - CASE | NOP - CASE*

	CASE -> op = C function address, 0 for default
	|
	[expr | COMPOUND] - COMPOUND

	ACTION

============================================================================*/

static void reduce_switch (qu_ast n)
{
	qu_ast c, g;

	c = n->child;
	reduce_node (c);
	reduce_node (c->next);

	for (c = c->next->next; c; c = c->next)
	{
		assert (c->type == QU_AST_case);

		g = c->child;
		if (g->type == QU_AST_compound)
		{
			reduce_node (g);
		}
		else if (g->type == QU_AST_literal && g->op == 1)
		{
			/*
				case .method:

				case
				|
				literal (op: 1)
			*/

			/* nothing to do */
		}
		else if (g->type == QU_AST_call && g->thing == QuNil)
		{
			/*
				case .method (a, b, c):

			        case
			          |
			         call (op: narg, thing: nil)
			        / | \
			  literal .....
			*/

			reduce_child_nodes (g->child);
		}
		else if (g->type == QU_AST_unop && g->thing == QuNil)
		{
			/*
				case ? expr:

			        case
			          |
			         unop (thing: nil, op: qu_sys_true)
			          |
			        <expr>
			*/
			reduce_node (g->child);
		}
		else
		{
			reduce_node (g);
		}

		reduce_node (g->next);
	}

	c = n->child;
	if (c->type == QU_AST_case)
		reduce_node (c->child);
}



/*----------------------------------------------------------------------------

	UNOP -> op = qu_sys_not | qu_sys_true | etc.
	 |
	 a

	Action
	------
	Constant folding.

	Result
	------
	The node may be changed to LITERAL.

		UNOP
			==> LITERAL -> thing = the result

----------------------------------------------------------------------------*/

static void reduce_unop (qu_ast n)
{
	qu_ast    a = n->child;
	qu_unop_f f = (qu_unop_f) n->op;

	reduce_node (a);

	if (a->type == QU_AST_literal && code_unop_literal_return_literal (f))
	{
		qu_any r = f (a->thing);

		if (r == 0)
			CRETHROW (n);

		if (!qu_gc_is_static (r))
		{
			/* must use qu_set_addif_ex otherwise storing, e.g. a Big and a Float,
				which are equal will only store one of them */
			r = qu_set_addif_ex (CoLiterals, r);
		}

		n->type  = QU_AST_literal;
		n->thing = r;
		n->child = 0;
		n->op    = 0;
	}
}



/*----------------------------------------------------------------------------

	VAR -> thing = Var object  or Array [Var object, Ast symbol] (dotted var)
	|
	[SYMBOL] - [expr] - [SUB]

	Action
	------
	Resolve and remove SYMBOL.         dotted var => create attribute
	qu_set the type of the variable.

	Result
	------
	The SYMBOL node is removed.

		VAR
			==> VAR -> thing = Var object, op = 1:with SYMBOL, 0:not
			           |
			           [expr] - [SUB]

----------------------------------------------------------------------------*/

/*EXPERIMENTAL - dotted var */
static void reduce_dotted_var (qu_ast n)
{
	/*
		var Foo.Bar.baz

			SYMBOL -> thing = Foo
			|
			Bar

		var Foo.bar

			SYMBOL -> thing = Foo
	*/

	qu_array a = n->thing;
	qu_ast   y, c;
	qu_class k;
	qu_any   x;
	qu_var   v;

	assert (a->len == 2);

	v = a->objs[0]; assert (qu_var_is (v));
	y = a->objs[1]; assert (qu_ast_is (y));

	x = qu_symbol_find (CoMaster, y->thing);

	if (x == 0)
		CTHROW (y, "no such symbol");

	if (qu_module_is (x))
	{
		c = y->child;
		if (c == 0)
			CTHROW (y, "class name expected");

		x = qu_module_getattr ((qu_module) x, c->thing);
		if (x == 0)
			CTHROW (y, "no such symbol");
	}

	if (!qu_class_is (x))
		CTHROW (y, "no such class");

	k = x;

	/* unlike methods, we don't allow overriding properties */
	if (qu_class_search_attr (k, v->name))
		CTHROW (y, "cant override existing attribute with property");

	/*
		must do this now otherwise later reference to this Var will fail
		use qu_attr_setex anyway
	*/
	qu_attr_setex (k->attrs, v->name, v);

	n->thing = v;
}



static void reduce_var (qu_ast n)
{
	qu_var v;
	qu_ast c = n->child;

#if 1
	if (c && c->type == QU_AST_ref && c->op)
	{
		/*EXPERIMENTAL - currying type

			var v: @foo (1, 2)

			VAR
			|
			REF -> op = 1
		*/

		qu_any type;

		v = n->thing;
		assert (qu_var_is (v));

		reduce_node (c);
		assert (c->type == QU_AST_literal);

		type = c->thing;
		assert (qu_method_is (type));

		/* check anyway */
		if (!qu_object_valid_type (type))
			CTHROW (c, "invalid type");

		v->type = type;
		c = c->next;
		n->child = c;

		if (c)
			reduce_nodes_from (c);
		return;
	}
#endif

	/*EXPERIMENTAL - dotted var */
	if (qu_array_is (n->thing))
		reduce_dotted_var (n);

	v = n->thing;

	if (c && c->type == QU_AST_symbol)
	{
		qu_any type;

		reduce_type (c);
		assert (c->type == QU_AST_literal);
		type = c->thing;

		if (!qu_object_valid_type (type))
			CTHROW (c, "invalid type");
		v->type = qu_object_fix_type (type);

		c = c->next;
		n->child = c;

		if (c)
		{
			reduce_nodes_from (c);

			if (c->type == QU_AST_literal)
			{
				if (!qu_object_check_value (v->type, c->thing))
					CTHROW (c, "invalid default value");
			}
		}
	}
	else
	{
		reduce_child_nodes (n);
	}
}



/*----------------------------------------------------------------------------

	WHILE
	|
	[expr] - COMPOUND

	Action
	------
	Optimize.

	Result
	------
	The WHILE node may be changed to NOP or LOOP

		WHILE
			==> NOP

		WHILE
			==> LOOP
			    |
			    COMPOUND

----------------------------------------------------------------------------*/

static void reduce_while (qu_ast n)
{
	qu_ast c;

	reduce_child_nodes (n);

	c = n->child;
	if (c->type == QU_AST_literal)
	{
		if (qu_sys_true (c->thing) == QuTrue)
		{
			n->child = c->next;
			n->type  = QU_AST_loop;
		}
		else
		{
			n->type  = QU_AST_nop;
			n->child = 0;
		}
	}
}



/*----------------------------------------------------------------------------

	WITH
	|


	Action
	------


	Result
	------

----------------------------------------------------------------------------*/

static void reduce_with (qu_ast n)
{
	/* Currently we don't produce a WITH node. Compile makes a tree of
		other nodes instead. */

	QU_FATAL ("NEVER");
}



/*----------------------------------------------------------------------------

	YIELD
	|
	expr+

	Action
	------
	Check return type if possible.

	Result
	------
	Nothing changed.

----------------------------------------------------------------------------*/

static void reduce_yield (qu_ast n)
{
	qu_ast c;
	qu_any type;

	assert (qu_sub_is (CoMaster));

	reduce_child_nodes (n);

	type = ((qu_sub) CoMaster)->type;
	if (type)
	{
		for (c = n->child; c; c = c->next)
		{
			/* TODO - should be able to do better than just testing literals */
			if (c->type == QU_AST_literal)
			{
				if (!qu_object_check_value (type, c->thing))
					CTHROW (n, "invalid return value");
			}
		}
	}
}



/*----------------------------------------------------------------------------
	reduce node
----------------------------------------------------------------------------*/

typedef void (*reduce_f) (qu_ast n);

static reduce_f ReduceFunctions [QU_AST_LAST + 1] =
{
	/* 0        */ 0,
	/* alias    */ reduce_alias,
	/* anonsub  */ reduce_anonsub,
	/* arg      */ 0,
	/* argc     */ 0,
	/* array    */ reduce_array,
	/* assert   */ reduce_assert,
	/* assign   */ reduce_assign,
	/* atomic   */ 0,
	/* binop    */ reduce_binop,
	/* break    */ 0,
	/* call     */ 0,
	/* case     */ 0,
	/* cat      */ 0,
	/* catch    */ reduce_catch,
	/* class    */ reduce_class,
	/* compound */ 0,
	/* continue */ 0,
	/* const    */ reduce_const,
	/* dict     */ reduce_dict,
	/* do       */ 0,
	/* eat      */ reduce_eat,
	/* enum     */ reduce_enum,
	/* enumitem */ 0,
	/* else     */ 0,
	/* extern   */ reduce_extern,
	/* finally  */ 0,
	/* flat     */ 0,
	/* for      */ reduce_for,
	/* frame    */ 0,
	/* getattr  */ reduce_getattr,
	/* getitem  */ 0,
	/* getslice */ 0,
	/* htclose  */ 0,
	/* htentity */ 0,
	/* htopen   */ 0,
	/* if       */ reduce_if,
	/* import   */ reduce_import,
	/* isa      */ reduce_isa,
	/* list     */ 0,
	/* literal  */ 0,
	/* local    */ reduce_local,
	/* loop     */ reduce_loop,
	/* me       */ 0,
	/* method   */ 0,
	/* mixin    */ reduce_mixin,
	/* module   */ reduce_module,
	/* name     */ reduce_name,
	/* nop      */ 0,
	/* param    */ reduce_param,
	/* pass     */ 0,
	/* print    */ 0,
	/* printex  */ 0,
	/* quest    */ 0,
	/* redo     */ 0,
	/* ref      */ reduce_ref,
	/* return   */ reduce_return,
	/* self     */ 0,
	/* set      */ reduce_set,
	/* stacktop */ 0,
	/* static   */ reduce_static,
	/* stmt     */ 0,
	/* struct   */ reduce_struct,
	/* sub      */ reduce_sub,
	/* super    */ 0,
	/* swap     */ 0,
	/* switch   */ reduce_switch,
	/* symbol   */ reduce_symbol,
	/* tag      */ 0,
	/* throw    */ 0,
	/* todo     */ 0,
	/* try      */ 0,
	/* unop     */ reduce_unop,
	/* var      */ reduce_var,
	/* while    */ reduce_while,
	/* with     */ reduce_with,
	/* yield    */ reduce_yield,
};



static void reduce_node (qu_ast n)
{
	reduce_f f = ReduceFunctions [n->type];

#if 0
	qu_string s = qu_ast_name (n->type);
	fprintf (stderr, "REDUCE %s %ld %s\n", CoModule->name->str, n->line, s->str);
#endif

	if (f)
	{
		f (n);
	}
	else
	{
		if (n->child)
			reduce_child_nodes (n);
	}
}
