/* qu_set.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Set is Container
	A Set is a dynamic container like an -{Array}. The only difference
	is that a Set does not keep the items in a particular order. Instead,
	hash values of the items are used to quickly locate objects. Thus it is
	a variation of a -{Dict} with no values.

	To create a Set you use the syntax or the #{new} method.

	@xmp
		s = {1, 'foo', 2.3}
		s = Set (1, 'foo', 2.3)

	@xmp
		s = {}.type (Int)
		s <<< 1 <<< 2
		println (s)
		try
			s <<< 'foo'
		catch
			println (__err__)
		;;

	@out
		{1, 2}
		(EValue) invalid type or value

	@p
	You can make a Set immutable with the -{Container.freeze} method.
	You can also create frozen sets using the k{const} syntax.

	v{Immutable} here means that the Set is guaranteed to contain
	the same objects at any time. Those objects, however, do not
	necessarily have to be immutable. They may be container objects
	with variable items.

	@xmp
		s = const {1, 2, 3}

	@p
	The k{const} syntax has an advantage. A literal object is created at compile
	time, once only. Thus the above code will simply assign the literal to the
	variable. Note that all items must be literals, i.e. numbers or strings.
*/
qu_class QuSetClass;


/* frozen empty Set */
qu_set QuSetEMPTY;


/* strings */
static qu_string Sempty;
static qu_string Srecursive;


/* item in the bucket */
#define item_s qu_setentry_s
struct item_s
{
	item_s *next;  /* Next entry in bucket */
	qu_any  obj;   /* qu_object */
	long    hash;  /* Cached hash value */
};



/*----------------------------------------------------------------------------
	macros
----------------------------------------------------------------------------*/

/*
	Validate x to conform to self->type.

	If type is a subroutine validator then x may be changed
	and the returned value is processed.
*/
#define VALIDATE_ITEM(self, x) \
QU_BEGIN \
	if ((self)->type) \
		x = qu_object_validate ((self), (self)->type, x); \
QU_END


/*----------------------------------------------------------------------------
	sizing
----------------------------------------------------------------------------*/

static void _set_resize (qu_set self, long newmask)
{
	item_s **b, **u, *i, *x;
	long     k, h, m;

	/*
		This is ugly. It requires allocating new memory chunk while the
		old one kept alive.
	*/

	u = self->buckets;
	m = self->mask;
	b = qu_mem_calloc (newmask + 1, sizeof (item_s *));
	for (k = 0; k <= m; ++k)
	{
		for (i = u [k]; i; i = x)
		{
			x = i->next;
			h = i->hash & newmask;
			i->next = b [h];
			b [h] = i;
		}
	}
	qu_mem_free (u);
	self->buckets = b;
	self->mask = newmask;
}



/*----------------------------------------------------------------------------
	allocation
----------------------------------------------------------------------------*/

/*
	Free all entries.
	Does not free nor clear buckets.
*/
static void _set_free_items (qu_set self, int freebuckets)
{
	item_s **b = self->buckets, *i, *x;
	long     m = self->mask, k;

	for (k = 0; k <= m; ++k)
	{
		for (i = b [k]; i; i = x)
		{
			x = i->next;
			qu_mem_free (i);
		}
	}
	if (freebuckets)
		qu_mem_free (b);
}



/*
	Initialize the buckets.
*/
#define CLEAR_BUCKETS(self) \
	memset (self->buckets, 0, (self->mask + 1) * sizeof (item_s *))



/*
	Initialize or reinitialize.
	Allocate and initialize buckets.
*/
static qu_set _set_setup (qu_set self, long mask)
{
	self->len     = 0;
	self->mask    = mask;
	self->buckets = qu_mem_calloc (mask + 1, sizeof (item_s *));
	self->type    = 0;

	QU_CONTAINER_INIT (self);
	return self;
}



/*
	Create a regular GC Set.
*/
qu_set qu_set_new (long cap)
{
	return _set_setup (qu_gc_new (QuSetClass), qu_container_mask_sizer (cap));
}



/*
	Create a root Set.
*/
qu_set qu_set_new_root (long cap)
{
	return _set_setup (qu_gc_new_root (QuSetClass), qu_container_mask_sizer (cap));
}



/*
	Create a static Set.
*/
qu_set qu_set_new_static (long cap)
{
	return _set_setup (qu_gc_new_static (QuSetClass), qu_container_mask_sizer (cap));
}



/*
	Create a static Set from an existing one.
*/
qu_set qu_set_to_static (qu_set self)
{
	qu_set s;
	long   i, j;
	qu_any x;

	if (qu_gc_is_static (self))
		return self;

	i = j = 0;
	while ((x = qu_set_walk (self, &i, &j)) != 0)
	{
		if (!qu_gc_is_static (x))
			return 0;
	}

	s = qu_set_new_static (self->len);
	i = j = 0;
	while ((x = qu_set_walk (self, &i, &j)) != 0)
		qu_set_add (s, qu_object_to_static (x));
	qu_set_freeze (x);
	return s;
}



/*----------------------------------------------------------------------------
	find

	Equality test must use the Set item's __eq method, not of the foreign
	object.
----------------------------------------------------------------------------*/

/*
	Find an object.
*/
static item_s *_set_find (qu_set self, qu_any o, long hash)
{
	item_s *i;

	for (i = self->buckets [hash & self->mask]; i; i = i->next)
	{
		if (qu_sys_eq (i->obj, o) == QuTrue)
			return i;
	}
	return 0;
}



/*
	Find an object with the same class.

	This is used by the compiler so that the Set stores, for example, an Int and
	a Long with the same value as two different objects.
*/
static item_s *_set_find_ex (qu_set self, qu_any o, long hash)
{
	item_s *i;

	for (i = self->buckets [hash & self->mask]; i; i = i->next)
	{
		if (qu_classof (i->obj) == qu_classof (o) && qu_sys_eq (i->obj, o) == QuTrue)
			return i;
	}
	return 0;
}



int qu_set_contains (qu_set self, qu_any x)
{
	return _set_find (self, x, qu_object_hash (x)) != 0;
}



/*----------------------------------------------------------------------------
	include/exclude
----------------------------------------------------------------------------*/

/*
	Add object. No questions asked!

	This function add the object without checking whether or not the same
	object is already in the Set. The caller must do that by itself to make
	sure the Set contains unique objects.
*/
static int _set_include (qu_set self, qu_any o, long h)
{
	register item_s *i;

	QU_MEM_CHECK (self->len < QU_GC_MAX_ITEMS);

	i       = qu_mem_alloc (sizeof (item_s));
	i->obj  = o;
	i->hash = h;

	h &= self->mask;
	i->next = self->buckets [h];
	self->buckets [h] = i;
	self->len++;

	h = qu_container_mask_grow (self->mask, self->len);
	if (h)
		_set_resize (self, h);
	return 1;
}



/*
	Remove object.
*/
static int _set_exclude (qu_set self, qu_any o, long h)
{
	register item_s *i, *p;
	register long    k;

	k = h & self->mask;
	for (i = self->buckets [k], p = NULL; i; p = i, i = i->next)
	{
		if (qu_sys_eq (i->obj, o) == QuTrue)
		{
			if (p)
				p->next = i->next;
			else
				self->buckets [k] = i->next;
			qu_mem_free (i);
			self->len--;

			k = qu_container_mask_shrink (self->mask, self->len);
			if (k)
				_set_resize (self, k);
			return 1;
		}
	}
	return 0;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (*arg) -> Array
	Create a new Set with v{arg} as the items.

	If called without any arguments then a new empty Set is returned.

	Note that duplicate items are automatically ignored.

	@noexc
*/
static qu_set qu__set__inst (qu_class self, int argc, qu_any *argv)
{
	qu_set s;

	assert (self->size >= QuSetClass->size);
	s = _set_setup (qu_gc__new (self, self->size), qu_container_mask_sizer (argc));

	while (argc--)
	{
		qu_any x = *argv++;
		long   h = qu_object_hash (x);

		if (!_set_find (s, x, h))
			_set_include (s, x, h);
	}
	return s;
}



/*) sub class new (*arg) -> Set
	Create a new Set with v{arg} as the items.

	This calls #{__inst} then f{__init} if one is defined.

	@noexc

	@xmp
		s = Set (1, 2, 1, 3, 1, 7)   #= s = {1, 2, 1, 3, 1, 7}
		println (s)

	@out
		{2, 3, 1, 7}
*/
qu_set qu__set_new (qu_class self, int argc, qu_any *argv)
{
	qu_set s = qu__set__inst (self, argc, argv);
	return qu_class_init_new (self, s, argc, argv);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__set__destroy (qu_set self)
{
	_set_free_items (self, 1);
}



static void qu__set__mark (qu_set self)
{
	if (self->type != QuIntClass)
	{
		item_s *i, **b = self->buckets;
		long    k,   m = self->mask;
		qu_any  x;

		for (k = 0; k <= m; ++k)
		{
			for (i = b [k]; i; i = i->next)
			{
				x = i->obj;
				qu_gc_mark (x);
			}
		}
		qu_gc_xmark (self->type);
	}
	QU_CONTAINER_MARK (self);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self base on v{x}.

	v{x} must be a one of the following -{Class} objects:
	@ops
	.| -{Array} |
		Return an Array containing all items in self.
		Note that this is an atomic operation thus the new Array is guaranteed
		to contain all items in self.

	.| -{Dict} |
		Return a Dict containing all items in self. The keys are items in the
		Set and the values are all set to true. Note that this is an atomic
		operation unless the __hash or __eq method of the objects modify the
		Set while this method is in progress.

	.| -{String}  | Delegate to #{__string}.
	.| -{Boolean} | Delegate to #{__true}.
	.| -{Set}     | Return self.

	@exc
	.EArgValue

	@xmp
		x = {2 1 3}
		y = x as Array
		println (x)

	@out
		[1, 2, 3]
*/
qu_array qu_set_as_array (qu_set self)
{
	qu_array   a = qu_array_new (self->len);
	item_s   **b = self->buckets, *i;
	long       m = self->mask, k;
	qu_any    *p = a->objs;

	for (k = 0; k <= m; ++k)
		for (i = b [k]; i; i = i->next)
			*p++ = i->obj;

	a->len = self->len;
	return a;
}


qu_dict qu_set_as_dict (qu_set self)
{
	QU_MEM_CHECK (self->len <= QU_GC_MAX_ITEMS / 2);
	{
		qu_dict d = qu_dict_new (self->len);
		long    b = 0, i = 0;
		qu_any  x;

		while ((x = qu_set_walk (self, &b, &i)) != 0)
			qu_dict_setitem (d, x, QuTrue);
		return d;
	}
}


static qu_any qu__set__as (qu_set self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return qu_set_as_array   (self);
		case QU_CLASS_ID_Dict:    return qu_set_as_dict    (self);
		case QU_CLASS_ID_String:  return qu_set_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_set_as_boolean (self);
		case QU_CLASS_ID_Set:     return self;
	}
	QU_ARG_THROW;
}



/*) sub __clear -> self
	Remove all items from the Set.

	Note that this is an atomic operation.
	It does not modify the allocated capacity.

	@exc
	.EModify

	@xmp
		x = {1, 2, 3}
		println (x.len)

		x.__clear
		println (x.len)

	@out
		3
		0
*/
qu_set qu__set__clear (qu_set self)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	_set_free_items (self, 0);
	CLEAR_BUCKETS (self);
	self->len = 0;
	return self;
}



/*) sub clear
	This is an alias of #{__clear}.
*/



/*) sub __copy -> Set
	Return a shallow copy of self.

	Note that this is an atomic operation, thus the new Set is guaranteed to
	contain all items in self by the time this method is called.

	@warn
	The #{type} property is not inherited.
	You must call f{type} yourself as necessary.

	@noexc

	@xmp
		x = {1, 2, 3}
		y = x.__copy
		println (y)
		println (x == y)
		println (x === y)

	@out
		{1, 2, 3}
		false
		true
*/
static qu_set qu__set__copy (qu_set self)
{
	item_s **b = self->buckets, **u, *i;
	long     m = self->mask, k;
	qu_set   s;

	s          = qu_gc__new (self->klass, self->klass->size);
	s->len     = self->len;
	s->mask    = m;
	s->buckets = u = qu_mem_calloc (m + 1, sizeof (item_s *));
	s->type    = 0;

	QU_CONTAINER_INIT (s);

	for (k = 0; k <= m; ++k)
	{
		for (i = b [k]; i; i = i->next)
		{
			item_s *j = qu_mem_alloc (sizeof (item_s));
			long    h = i->hash;

			j->obj  = i->obj;
			j->hash = h;

			h &= m;
			j->next = u [h];
			u[h] = i;
		}
	}
	return s;
}



/*) sub __del (x) -> self
	Remove v{x} from the Set.

	@warn
	The Set is temporarily protected during deletion.

	@exc
	.EModify

	@xmp
		a = {1 2 3 4}
		println (a.__del (5))
		println (a)

	@out
		0
		{1, 2, 3, 4}
*/
qu_set qu__set__del (qu_set self, qu_any x)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
  __qu_container_protect_begin (self);
	_set_exclude (self, x, qu_object_hash (x));
  __qu_container_protect_end;
	return self;
}



/*) sub __each -> Iterator
	Return an -{Iterator} object to iterate through the items.

	Since a Set does not keep the items in any particular order then the
	items will appear in an arbitrary sequence.

	@warn
	The Set is not frozen during iteration.

	@noexc

	@xmp
		x = {3 'foo' 1.2}
		for i in each x          # calls x.__each
			print (i ' ')
		;;
		println

		for i in x               # interpreter uses an internal iterator
			print (i ' ')
		;;
		println

	@out
		1.2 foo 3
		1.2 foo 3
*/
static qu_any qu__set__next (qu_set self, qu_iterator i)
{
	qu_any x = qu_set_walk (self, &i->ex1, &i->ex2);
	QU_REQUIRE (QuEStop, x);
	return x;
}


static qu_iterator qu__set__each (qu_set self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, qu__set__next);
	return i;
}



/*) sub __in (x) -> Boolean
	Return true if v{x} is in the Set, v{false} otherwise.

	@warn
	The Set is temporarily protected during the search.

	@noexc

	@xmp
		a = {1 2 3}
		println (2 in a)

	@out
		true
*/
static qu_boolean qu__set__in (qu_set self, qu_any x)
{
	item_s *i;

  __qu_container_protect_begin (self);
	i = _set_find (self, x, qu_object_hash (x));
  __qu_container_protect_end;
	return i ? QuTrue : QuFalse;
}



/*) sub __like (x) -> Boolean
	If v{x} is Set and contains the same items as self then return true,
	false otherwise.

	@noexc

	@warn
	Both Sets are not frozen during the comparison.

	@xmp
		a = {1 2 3}
		b = {3 2 1}
		println (a == b)
		println (a === b)

	@out
		false
		true
*/
qu_boolean qu__set__like (qu_set self, qu_any x)
{
	if (qu_set_is (x))
	{
		long   b = 0, i = 0;
		qu_any y;

		while ((y = qu_set_walk (self, &b, &i)) != 0)
		{
			if (qu__set__in ((qu_set) x, y) != QuTrue)
				return QuFalse;
		}
		return QuTrue;
	}
	return QuFalse;
}



/*) sub __of (type) -> Any
	Assign self to contain only items of the specified type v{x} then return
	self. If a type is already assigned then EModify is thrown.

	@warn
	The -{Null} class may not be used as a type.

	Once set, you can not unset the type.

	@exc
	.EModify

	@xmp
		s = {1, 3} of Int
		println (s.type)
		try
			s <<< 'foo'
		catch
			println (__err__)
		;;

	@out
		<Class sys.Int>
		(EValue) invalid type or value
*/
qu_set qu__set__of (qu_set self, qu_any type)
{
	QU_ARG_CHECK (qu_object_valid_type (type));
	if (qu_var_is (type) && !qu_var_is_typeof (type))
		type = ((qu_var) type)->value;

	if (self->type != type)
	{
		QU_REQUIRE (QuEModify, self->type == 0);
		if (self->len)
		{
			int ok;
		  __qu_container_protect_begin (self);
			ok = (qu__set_typeof (self, type) == QuTrue);
		  __qu_container_protect_end;
			QU_ARG_CHECK (ok);
		}
	}
	self->type = type;
	return self;
}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		t = {1, 'foo'}
		t <<< t
		println (t)

	@out
		{1, 'foo', {..}}
*/
static qu_null qu__set__print (qu_set self)
{
	if (self->len < 1)
	{
		QU_ENSURE (qu_sys_printz (Sempty->str, Sempty->len));
	}
	else if (qu_thread_state_exist (QU_THREAD_STATE_PRINT, self))
	{
		QU_ENSURE (qu_sys_printz (Srecursive->str, Srecursive->len));
	}
	else
	{
		qu_sys_printb_s b;
		int     r;
		qu_any  x;
		long    i = 0, j = 0, k = 0;

		qu_sys_printb_init (b);

	  __qu_thread_state_begin (QU_THREAD_STATE_PRINT, self);
		r = qu_sys_printb (b, "{", 1);
		while (r && (x = qu_set_walk (self, &i, &j)))
		{
			if (k++)
				r = qu_sys_printb (b, ", ", 2);
			if (r)
			{
				x = QU_CONTAINER_QUOTE (x);
				r = qu_sys_printb_string (b, x);
			}
		}
		if (r)
			r = qu_sys_printb (b, "}", 1);
		if (r)
			r = qu_sys_printb_final (b);
	  __qu_thread_state_end;
		QU_ENSURE (r);
	}
	return QuNil;
}



/*) sub __push (x) -> self
	Include v{x} in the Set then return self.

	If v{x} is already in the Set then nothing happens.

	@warn
	The Set is temporarily protected during insertion.

	@exc
	.EModify
	.EValue  value does not conform to #{type}

	@xmp
		t = {'foo'}
		t <<< 3 <<< 'bar' <<< 2
		println (t)

	@out
		{'foo', 2, 'bar', 3}
*/
qu_set qu__set__push (qu_set self, qu_any x)
{
	long h;
	int  ok = 1;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
  __qu_container_protect_begin (self);
	VALIDATE_ITEM (self, x);
	if (x == 0)
	{
		ok = 0;
	}
	else
	{
		h = qu_object_hash (x);
		if (!_set_find (self, x, h))
			ok = _set_include (self, x, h);
	}
  __qu_container_protect_end;
	QU_ENSURE (ok);
	return self;
}



/*) sub __pop -> Any
	Pop an arbitrary item from the Set.

	@exc
	.EModify

	@xmp
		t = {'foo'}
		t >>> x
		println (t)
		println (x)

	@out
		{}
		foo
*/
qu_any qu__set__pop (qu_set self)
{
	item_s **b, *i;
	long     k, m;
	qu_any   x;

	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	m = self->mask;
	b = self->buckets;
	for (k = 0; k <= m; ++k)
	{
		i = b [k];
		if (i)
		{
			x = i->obj;
			b [k] = i->next;
			qu_mem_free (i);
			self->len--;
			k = qu_container_mask_shrink (m, self->len);
			if (k)
				_set_resize (self, k);
			return x;
		}
	}
	QU_THROW (QuENone);
}



/*) sub __string -> String
	Return a string representation of self.

	@exc
	.EMemory

	@xmp
		a = {1 2 3}
		b = a as String
		println (b)

	@out
		{1, 2, 3}
*/
qu_string qu__set__string (qu_set self)
{
	if (self->len < 1)
	{
		return Sempty;
	}
	else if (qu_thread_state_exist (QU_THREAD_STATE_STRING, self))
	{
		return Srecursive;
	}
	else
	{
		long      b = 0, i = 0;
		qu_string s = qu_string_buffer_new (32);
		int       r, k = 0;
		qu_any    x;

	  __qu_thread_state_begin (QU_THREAD_STATE_STRING, self);
		r = qu_string_buffer_putc (s, '{');
		while (r && (x = qu_set_walk (self, &b, &i)))
		{
			if (k++)
				r = qu_string_buffer_put (s, ", ", 2);
			if (r)
				r = qu_string_buffer_puts (s, QU_CONTAINER_QUOTE (x));
		}
		if (r)
			r = qu_string_buffer_putc (s, '}');

	  __qu_thread_state_end;
		QU_ENSURE (r);
		return qu_string_buffer_return (s);
	}
}



/*) sub __true -> Boolean
	Return false if the Set is empty, true otherwise.

	@noexc

	@xmp
		a = {}
		println (a ? 'something' : 'none')

	@out
		none
*/
static qu_boolean qu__set__true (qu_set self)
{
	return qu_set_as_boolean (self);
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub any (defval = nil) -> Any
	Return an arbitrary item in the Set.

	If the Set is empty and v{defval} is given then v{defval} is returned,
	otherwise ENone is thrown

	@exc
	.ENone

	@xmp
		s = {'foo', 'bar', 'gaz'}
		println (s.any)

	@out
		gaz
*/
qu_any qu__set_any (qu_set self, int argc, qu_any *argv)
{
	if (self->len == 0)
	{
		if (argc)
			return argv[0];
		QU_THROW (QuENone);
	}
	else
	{
		item_s **b, *i;
		long     k, m;

		b = self->buckets;
		m = self->mask;
		for (k = 0; k <= m; ++k)
		{
			i = b [k];
			if (i)
				return i->obj;
		}
		QU_FATAL ("never");
		return 0;
	}
}



/*) sub copy (keeptype = false) -> Set
	Make a shallow copy of this Set.

	If v{keeptype} is true then the new Set is given the same #{type} as
	this Set. Otherwise this method is just an alias of #{__copy}.

	@exc
	.EArgValue
*/
static qu_set qu__set_copy (qu_set self, int argc, qu_any *argv)
{
	qu_set r;
	int    keep = 0;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, keep);
	r = qu__set__copy (self);
	if (keep)
		r->type = self->type;
	return r;
}



/*) sub delete (*arg) -> self
	Delete all items specified in v{arg}.

	If called without any argument then delete any arbitrary item.

	@warn
	The Set is temporarily protected while deleting.

	@exc
	.EModify

	@xmp
		s = {1 2 3 4}
		s.delete (1, 4)
		println (s)

	@out
		{2, 3}
*/
static qu_set qu__set_delete (qu_set self, int argc, qu_any *argv)
{
	QU_CONTAINER_ENSURE_NOT_FROZEN_FIXED (self);
	if (argc == 0)
	{
		qu_set_del_any (self);
	}
	else
	{
		__qu_container_protect_begin (self);
		while (argc--)
		{
			qu_any x = *argv++;
			_set_exclude (self, x, qu_object_hash (x));
		}
		__qu_container_protect_end;
	}
	return self;
}



/*) sub join (sep: String) -> String
	Concatenate the items as strings using v{sep} as the item separator.

	If the Set is empty then an empty string is returned.

	@exc
	.EArgValue
	.EMemory

	@xmp
		t = {1, 2, 3}
		println (t.join (':'))

	@out
		1:3:2
*/
static qu_string qu__set_join (qu_set self, qu_string sep)
{
	QU_ARG_TYPE (sep, qu_string);

	if (self->len == 0)
		return QuStringEMPTY;
	else
	{
		long       l = sep->len;
		char      *s = sep->str;
		qu_string  b;
		long       i = 0, j = 0;
		int        k = 0;
		qu_any     x;

		QU_MEM_CHECK (self->len < QU_MEM_MAX / (sep->len + 2));

		b = qu_string_buffer_new (self->len);

		while ((x = qu_set_walk (self, &i, &j)) != 0)
		{
			if (k > 0 && l)
				QU_ENSURE (qu_string_buffer_put (b, s, l));
			k = 1;
			if (x == self)
				QU_ENSURE (qu_string_buffer_puts (b, Srecursive));
			else
				QU_ENSURE (qu_string_buffer_puts (b, qu_sys_string (x)));
		}
		return qu_string_buffer_return (b);
	}
}



/*) sub type -> Any
	Return the type of items in self or k{Nil} if none set.

	@noexc

	@xmp
		s = {} of Int
		println (s.type)

	@out
		<Class sys.Int>
*/
static qu_any qu__set_type (qu_set self)
{
	return self->type ? self->type : QuNil;
}



/*) sub typeof (x) -> Boolean
	Check whether or not all items are of a specific type.

	v{x} may be any validator. If it is a -{Record} then its -{Struct} is used.
	Otherwise it is used as is.

	If self has been designated to contain items of v{x} (the #{type} method)
	then true is returned even if v{self} is empty. Otherwise, if self is
	empty then false is returned.

	@noexc

	@xmp
		a = {} of Int
		println (a.typeof (1))
		println (a.typeof (Integer))

	@out
		true
		true
*/
qu_boolean qu__set_typeof (qu_set self, qu_any type)
{
	if (self->len == 0)
		return QuFalse;

	if (self->type)
	{
		if (!qu_object_compat_type (self->type, type))
			return QuFalse;
	}

	/*begin*/
	{
		long b = 0, i = 0;
		qu_any  o;

		while ((o = qu_set_walk (self, &b, &i)) != 0)
		{
			if (!qu_object_validate (self, type, o))
				return QuFalse;
		}
	}
	return QuTrue;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

void qu_set_exchange (qu_set self, qu_set x)
{
	long     l = self->len;
	long     m = self->mask;
	item_s **b = self->buckets;

	self->len     = x->len;
	self->mask    = x->mask;
	self->buckets = x->buckets;

	x->len     = l;
	x->mask    = m;
	x->buckets = b;
}


qu_any qu_set_add (qu_set self, qu_any o)
{
	long h = qu_object_hash (o);

	assert (_set_find (self, o, h) == NULL);
	QU_ENSURE (_set_include (self, o, h));
	return o;
}



qu_any qu_set_addif (qu_set self, qu_any o)
{
	item_s *i;
	long    h = qu_object_hash (o);

	i = _set_find (self, o, h);
	if (i)
		return i->obj;
	QU_ENSURE (_set_include (self, o, h));
	return o;
}



qu_any qu_set_addif_ex (qu_set self, qu_any o)
{
	item_s *i;
	long    h = qu_object_hash (o);

	i = _set_find_ex (self, o, h);
	if (i)
		return i->obj;
	QU_ENSURE (_set_include (self, o, h));
	return o;
}



qu_string qu_set_addz (qu_set self, const char *z, long len)
{
	item_s    *i;
	qu_string  s;
	long       h;
	qu_any     o;

	h = qu_strz_hash (z, len);
	for (i = self->buckets [h & self->mask]; i; i = i->next)
	{
		o = i->obj;
		if (qu_string_is (o) && (((qu_string) o)->len == len)
				&& memcmp (((qu_string) o)->str, z, len) == 0)
			return (qu_string) o;
	}
	s = qu_string_new (z, len);
	QU_ENSURE (_set_include (self, s, h));
	return s;
}



qu_float qu_set_add_float (qu_set self, double v)
{
	item_s   *i;
	qu_float  f;
	long      h;
	qu_any    o;

	h = qu_float_hashd (v);
	for (i = self->buckets [h & self->mask]; i; i = i->next)
	{
		o = i->obj;
		if (qu_float_is (o) && ((qu_float) o)->value == v)
			return o;
	}
	f = qu_float_new (v);
	QU_ENSURE (_set_include (self, f, h));
	return f;
}



/*
	Remove an arbitrary entry.
*/
void qu_set_del_any (qu_set self)
{
	if (self->len)
	{
		item_s **b, *i;
		long     k, m;

		b = self->buckets;
		m = self->mask;
		for (k = 0; k <= m; ++k)
		{
			i = b [k];
			if (i)
			{
				b [k] = i->next;
				qu_mem_free (i);
				self->len--;
				k = qu_container_mask_shrink (self->mask, self->len);
				if (k)
					_set_resize (self, k);
				break;
			}
		}
	}
}



/*
	Note that this iterator survives modification while iterating.

	However, the items yielded is arbitrary if modification is performed
	during the iteration. It is theoretically possible that the same
	item is yielded more than once.
*/
qu_any qu_set_walk (qu_set self, long *bucket, long *index)
{
	item_s *e, **b;
	long    i, j, k, m;

	b = self->buckets;
	m = self->mask;

	for (k = *bucket, i = *index; k <= m; ++k, i = 0)
	{
		for (e = b[k], j = 0; e; e = e->next, ++j)
		{
			if (j == i)
			{
				*index  = i + 1;
				*bucket = k;
				return e->obj;
			}
		}
	}
	return 0;
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_set__setup ()
{
	qu_class_setup (QuSetClass, qu_set_s, qu__set__destroy, qu__set__mark);

	QuSetEMPTY = qu_set_new_static (0);
	qu_set_freeze (QuSetEMPTY);

	Sempty     = qu_string_new_staticz ("{}");
	Srecursive = qu_string_new_staticz ("{..}");
}



void qu_set__init ()
{
	/* standard class methods */
	qu_class_method_class (QuSetClass, QU_NAME__inst, qu__set__inst, -1);
	qu_class_method_class (QuSetClass, QU_NAME_new  , qu__set_new  , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuSetClass, QU_NAME__as    , qu__set__as    , 1);
	qu_class_method_inst  (QuSetClass, QU_NAME__clear , qu__set__clear , 0);
	qu_class_alias        (QuSetClass, QU_NAME__clear , QU_NAME_clear);
	qu_class_method_inst  (QuSetClass, QU_NAME__copy  , qu__set__copy  , 0);
	qu_class_method_inst  (QuSetClass, QU_NAME__del   , qu__set__del   , 1);
	qu_class_method_inst  (QuSetClass, QU_NAME__each  , qu__set__each  , 0);
	qu_class_method_inst  (QuSetClass, QU_NAME__in    , qu__set__in    , 1);
	qu_class_method_inst  (QuSetClass, QU_NAME__like  , qu__set__like  , 1);
	qu_class_method_inst  (QuSetClass, QU_NAME__of    , qu__set__of    , 1);
	qu_class_method_inst  (QuSetClass, QU_NAME__print , qu__set__print , 0);
	qu_class_method_inst  (QuSetClass, QU_NAME__push  , qu__set__push  , 1);
	qu_class_method_inst  (QuSetClass, QU_NAME__pop   , qu__set__pop   , 0);
	qu_class_method_inst  (QuSetClass, QU_NAME__string, qu__set__string, 0);
	qu_class_method_inst  (QuSetClass, QU_NAME__true  , qu__set__true  , 0);

	/* specific instance methods */
	qu_class_method_instz (QuSetClass, "any"   , qu__set_any    , -1);
	qu_class_method_instz (QuSetClass, "copy"  , qu__set_copy   , -1);
	qu_class_method_instz (QuSetClass, "delete", qu__set_delete , -1);
	qu_class_method_instz (QuSetClass, "join"  , qu__set_join   ,  1);
	qu_class_method_instz (QuSetClass, "type"  , qu__set_type   ,  0);
	qu_class_method_instz (QuSetClass, "typeof", qu__set_typeof ,  1);

	/* finalize declaration */
	qu_class_set_super (QuSetClass, QuContainerClass);
	qu_class_enddecl   (QuSetClass);
}
