/* Stream.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#error Get a real system
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#ifdef HAVE_SYS_LIMITS_H
#include <sys/limits.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif


/* Fix EWOULDBLOCK = EAGAIN */
#ifdef EWOULDBLOCK
#  if (EAGAIN == EWOULDBLOCK)
#    define FIXEAGAIN QU_BEGIN QU_END
#  else
#    define FIXEAGAIN QU_BEGIN if (errno == EWOULDBLOCK) errno = EAGAIN; QU_END
#  endif
#else
#    define FIXEAGAIN
#endif


#ifndef BUFSIZ
#define BUFSIZ 512
#endif



/*) class Stream
	This is the base class of all I{stream} classes, e.g. -{File}.

	The Stream class provides default methods which can be used by sublcasses,
	e.g. #{fd}, #{canread}, #{close}, etc.

	When your program starts, Qu assigns four opened streams. These are known
	as the I{standard streams}: #{stderr}, #{stdin}, #{stdout}, and k{null}.

	The output of the -{print} family functions go to #{stdout}, the -{input}
	function gets its input from #{stdin}, and -{__warning__} goes to #{stderr}.

	The k{null} Stream is always opened and available. Closing the k{null}
	Stream has no effect. Except for a few things, you can use the k{null}
	Stream as any other Stream objects.

	Note that k{null} is a keyword in Qu as much as true and false.
	You can also say C{Stream.__null__} but k{null} is more convenient.

	The Stream class is designed to work with non-blocking streams as well
	as blocking streams. Non-blocking streams are preferable.

	Generally Stream objects are created as non-blocking streams by means
	of #{fcntl} with an O_NONBLOCK parameter. You can change this via the
	convenient #{blocking} method.

	All Stream objects are equipped with a #{timeout} you can set at any time
	before performing a timed operation. This, of course, has probably no effect
	on blocking streams.

	Timed Stream methods will "block" when called, that is, control is given
	to other threads. This has the same effect as the k{pass} keyword). When
	the #{timeout} is reached then you get ETimeout. You never have to deal
	with EINTR and EAGAIN.

	@warn
	#{stderr}, #{stdin}, and #{stdout} can timeout as well as any other streams.
	When first created, they are simply given a huge #{timeout} value that will
	probably last even after you no longer exist. Unless of course, you set
	their values to less than or the same as your earthly lifetime period.

	@note
	As with any other object, Qu does not keep a lock around a Stream.
	If you close a Stream while another thread is trying to read from it, or
	more than one thread try to read/write from the same Stream at the same time,
	or you supply the same buffer to two threads doing #{recv} as the same
	time, then strange things can happen.

	So sharing the same Stream with other threads is basically a bad idea.

	The builtin -{print} function family enclose the write operation in an
	k{atomic} block but the #{read}, #{write} etc, methods do not.

	You must also remember that it is possible to get partial read/write
	because an error occurs at the middle of a read/write process.
	For example, the buffer might be filled-in by #{recv} and then
	an error, such as a socket disconnection, occurs.
*/
qu_class QuStreamClass;


/*) const __stderr__ = nil
	This is the standard error Stream.
	It is set by the Stream class on initialization.

	The output of -{__warning__} goes here.

	You can access or change it via #{stdin}.
*/
qu_stream QuStreamStderr;


/*) const __stdin__ = nil
	This is the standard input Stream.
	It is set by the Stream class on initialization.

	The -{input} function reads from this Stream.

	You can access or change it via #{stdin}.
*/
qu_stream QuStreamStdin;


/*) const __stdout__ = nil
	This is the standard output Stream.
	It is set by the Stream class on initialization.

	The output of the -{print} function family go here.

	You can access or change it via #{stdout}.
*/
qu_stream QuStreamStdout;


/*) const __null__ = nil
	This is the k{null} Stream.

	You access it with the k{null} keyword.

	k{null} is always opened and available.
	Closing and writing to the k{null} Stream has no effect.
*/
qu_stream QuStreamNull;


/*) struct Stat
		st_dev     : Integer  # device
		st_ino     : Integer  # inode
		st_mode    : Integer  # protection
		st_nlink   : Integer  # number of hard links
		st_uid     : Integer  # user ID of owner
		st_gid     : Integer  # group ID of owner
		st_rdev    : Integer  # device type if inode device
		st_size    : Integer  # total size in bytes
		st_blksize : Integer  # blocksize for filesystem I/O
		st_blocks  : Integer  # number of blocks allocated
		st_atime   : Float    # time of last access
		st_mtime   : Float    # time of last modification
		st_ctime   : Float    # time of last change
	;;
	This -{Struct} is used to store file information.

	See the #{stat} method and C{man 2 stat}.

	You can use the #{ismode} method to test for the v{st_mode} field.
*/

/*) enum ST_
		ST_DEV
		ST_INO
		ST_MODE
		ST_NLINK
		ST_UID
		ST_GID
		ST_RDEV
		ST_SIZE
		ST_BLKSIZE
		ST_BLOCKS
		ST_ATIME
		ST_MTIME
		ST_CTIME
	;;
	Field number of #{Stat}.

	Use these with #{stat} to get a particular field.

	@xmp
		println (Stream.stderr.stat (Stream.ST_MODE))
*/
static qu_struct StatStruct;


/* stupid forwards */
static qu_any     qu__stream_fstat   (qu_class self, int argc, qu_any *argv);
static qu_integer qu__stream__len    (qu_stream self);
static qu_any     qu__stream_stat    (qu_stream self, int argc, qu_any *argv);
static qu_string  qu__stream_read    (qu_stream self, int argc, qu_any *argv);
static qu_string  qu__stream_readln  (qu_stream self, int argc, qu_any *argv);
static qu_integer qu__stream_search  (qu_stream self, int argc, qu_any *argv);
static qu_integer qu__stream_seek    (qu_stream self, int argc, qu_any *argv);
static qu_integer qu__stream_tell    (qu_stream self);
static qu_integer qu__stream_write   (qu_stream self, int argc, qu_any *argv);
static qu_integer qu__stream_writeln (qu_stream self, int argc, qu_any *argv);


/*----------------------------------------------------------------------------
	internal stuffs
----------------------------------------------------------------------------*/

/* what time is now? */
double qu_stream_now ()
{
	struct timeval v;
	gettimeofday (&v, NULL);
	return (double) v.tv_sec + (double) v.tv_usec / 1e6;
}

#define __qu_blocking_check_timeout(timeout) \
QU_BEGIN \
	if (__qu_now__ >= (timeout)) \
	{ \
		__qu_blocking_end; \
		QU_THROW (QuETimeout); \
	} \
QU_END

#define __qu_blocking_check_timeoutv(timeout, retval) \
QU_BEGIN \
	if (__qu_now__ >= (timeout)) \
	{ \
		__qu_blocking_end; \
		QU_THROWV (QuETimeout, retval); \
	} \
QU_END



/*
	If the bytes read is smaller than this, then we keep 'buf' and return
	a new string instead -- see read.

	It should be somewhat nice to Gc.
*/
#define SMALL_LEN 80


/*
	When doing a read/write operation, we do a select before actually
	calling the read/write function. The impact on performance is generally
	noticeable.

	-1 means to use the default value 1000 microsec.
	Dont change this!
*/
#define SELECT_READY -1



/*
	Get a stat item.
	!!! Watch the assertion.
*/
static qu_any _stream_stat_item (struct stat *s, long i)
{
	assert (0 <= i && i <= 12);

	switch (i)
	{
		case 0:
#if SIZEOF_DEV_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_dev);
#else
			return qu_integer_newul ((unsigned long) s->st_dev);
#endif

		case 1:
#if SIZEOF_INO_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_ino);
#else
			return qu_integer_newul ((unsigned long) s->st_ino);
#endif

		case 2:
			return qu_integer_newul (s->st_mode);

		case 3:
#if SIZEOF_NLINK_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_nlink);
#else
			return qu_integer_newul ((unsigned long) s->st_nlink);
#endif

		case 4:
#if SIZEOF_UID_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_uid);
#else
			return qu_integer_newul ((unsigned long) s->st_uid);
#endif

		case 5:
#if SIZEOF_GID_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_gid);
#else
			return qu_integer_newul ((unsigned long) s->st_gid);
#endif

		case 6:
#if SIZEOF_DEV_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_rdev);
#else
			return qu_integer_newul ((unsigned long) s->st_rdev);
#endif

		case 7:
#if SIZEOF_OFF_T > SIZEOF_LONG
			return qu_integer_newull ((unsigned long long) s->st_size);
#else
			return qu_integer_newul ((unsigned long) s->st_size);
#endif

		case 8:
#ifdef HAVE_LONG_LONG
			/* cant get the size of this from configure :-( */
			if (sizeof (s->st_blksize) > sizeof (long))
				return qu_integer_newull ((unsigned long long) s->st_blksize);
#endif
			/* thus this is incorrect but do it anyway */
			return qu_integer_newul ((unsigned long) s->st_blksize);

		case 9:
#ifdef HAVE_LONG_LONG
			/* cant get the size of this from configure :-( */
			if (sizeof (s->st_blocks) > sizeof (long))
				return qu_integer_newull ((unsigned long long) s->st_blocks);
#endif
			/* thus this incorrect but do it anyway */
			return qu_integer_newul ((unsigned long) s->st_blocks);

		case 10: return qu_float_new ((double) s->st_atime);
		case 11: return qu_float_new ((double) s->st_mtime);
		case 12: return qu_float_new ((double) s->st_ctime);
	}
	QU_FATAL ("never");
	return 0;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class __inst (fd: Integer, option = '') -> Stream
	Create a new Stream object.

	v{fd} is the file descriptor.
	It may be -1 to indicate a closed stream.

	The Stream will have the #{linebreak} set to C{'\n'} and #{timeout} to
	a huge value probably until the end of the world.

	v{option} must be a combination of the following characters:
	@ops
	.| C{c} | can close
	.| C{b} | can set blocking mode
	.| C{e} | end-of-file is relevant

	@warn
	You should generally never call this method and let the Stream or its
	subclass takes care of it. Messing with a descriptor is a very bad idea.

	See #{fd} for more information.

	@exc
	.ESecurity
	.EArgValue
*/
static qu_stream qu__stream__inst (qu_class self, int argc, qu_any *argv)
{
	qu_string op = QuStringEMPTY;
	qu_stream s;
	int       fd;

	__qu_unsafe__;

	QU_ARG_VMIN (1);
	QU_ARG_vint (0, fd);
	s = qu_stream_new (self);
	s->fd = fd;
	if (argc > 1)
	{
		char *z;

		QU_ARG_VTYPE (0, op, qu_string);
		z = op->str;
		while (*z)
		{
			switch (*z++)
			{
				case 'c': case 'C': s->flags |= QU_STREAM_FLAG_CANCLOSE; break;
				case 'b': case 'B': s->flags |= QU_STREAM_FLAG_CANBLOCK; break;
				case 'e': case 'E': s->flags |= QU_STREAM_FLAG_CANEOF;   break;
			}
		}
	}
	return s;
}



/*) sub class new (fd: Integer, option = '', *arg) -> Stream
	Create and open a new Stream object.

	Subclasses may override this method.

	This method creates a new Stream and calls the __init method if it exists
	passing all the arguments.

	See #{__inst} for the arguments.
*/
static qu_stream qu__stream_new (qu_class self, int argc, qu_any *argv)
{
	qu_stream s;

	QU_REQUIRE (QuEAbstract, !qu_class_is_abstract (self));
	s = qu__stream__inst (self, argc, argv);
	QU_ENSURE (s);
	return qu_class_init_new (self, s, argc, argv);
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class fmode (x, timeout = 30)
	This is a convenience method to get the v{st_mode} field of #{fstat}.

	@exc
	.EArgValue
	.ESystem
*/
static qu_any qu__stream_fmode (qu_class self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	{
		qu_any arg[3] = {argv[0], QuInt2, qu_int_new (30)};

		if (argc > 1)
			arg[2] = argv[1];

		return qu__stream_fstat (self, 3, arg);
	}
}



/*) sub class fmtime (x, timeout = 30)
	This is a convenience method to get the v{st_mtime} field of #{fstat}.

	@exc
	.EArgValue
	.ESystem
*/
static qu_any qu__stream_fmtime (qu_class self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	{
		qu_any arg[3] = {argv[0], QuInt10, qu_int_new (30)};

		if (argc > 1)
			arg[2] = argv[1];

		return qu__stream_fstat (self, 3, arg);
	}
}



/*) sub fselect (sread: Set, swrite: Set, serr: Set, timeout = 1000) -> Integer
	Wait for a number of Streams to change status.

	v{sread}, v{swrite} and v{serr} must be Set objects contaning only Stream
	objects. At least one of them must have one item.

	On success v{sread}, v{swrite}, and v{serr} are modified and contain the
	Stream objects that have changed status.  Thus you may not use frozen
	Set objects and the three Sets must be different objects.

	v{timeout} is the number of microseconds to wait.

	See C{man 2 select} for details.

	@exc
	.EArgValue
	.ESystem

	@xmp
		n = Stream.fselect (r = {mysock, myfile}, w = {}, e = {})
		println (n)
		println (r)
		println (w)
		println (e)

	@out
		1
		{3}
		{}
		{}

	@note
	Note that if all Stream objects are valid then f{select} should not throw any
	exception except ENOMEM (insufficient memory), which is rare and most probably
	never happen.

	It is bad idea for an operating system to allow ENOMEM but that is the
	situation at hand.
*/
static int _stream_select_set_fd (qu_set s, fd_set *e, int *fdmax)
{
	qu_stream x;
	long      b = 0, i = 0;

	FD_ZERO (e);
	while ((x = qu_set_walk (s, &b, &i)) != 0)
	{
		if (!qu_stream_is (x))
			return 0;
		if (x->fd >= 0)
		{
			FD_SET (x->fd, e);
			if (x->fd > *fdmax)
				*fdmax = x->fd;
		}
	}
	return 1;
}


static void _stream_select_get_fd (qu_set s, fd_set *e)
{
	if (s->len)
	{
		static qu_set r = 0;

		qu_stream x;
		long      b, i;

		if (r == 0)
			r = qu_set_new_static (8);

		qu_set_clear (r);
		b = i = 0;
		while ((x = qu_set_walk (s, &b, &i)) != 0)
		{
			assert (qu_stream_is (x));
			if (x->fd >= 0)
			{
				if (FD_ISSET (x->fd, e))
					qu_set_add (r, x);
			}
		}
		qu_set_exchange (r, s);
	}
}


static qu_integer qu__stream_fselect (qu_class self, int argc, qu_any *argv)
{
	qu_set rs, ws, es;
	long   usec = 1000;
	fd_set r, w, e;
	struct timeval t;
	int    res, max = 0;

	__qu_unsafe__;

	QU_ARG_VCOUNT (3, 4);
	rs = argv[0];
	ws = argv[1];
	es = argv[2];

	QU_ARG_CHECK (qu_set_is (rs) && qu_set_is (ws) && qu_set_is (es)
		&& rs != ws && rs != es && ws != es
		&& (rs->len || ws->len || es->len)
		&& !qu_set_frozen (rs)
		&& !qu_set_frozen (ws)
		&& !qu_set_frozen (es));

	if (argc > 3)
		QU_ARG_vlong (3, usec);

	QU_ARG_CHECK (_stream_select_set_fd (rs, &r, &max)
	       && _stream_select_set_fd (ws, &w, &max)
	       && _stream_select_set_fd (es, &e, &max));

	t.tv_sec  = usec / 1000000;
	t.tv_usec = usec % 1000000;

	__qu_blocking_begin;
	while ((res = select (max + 1, &r, &w, &e, &t)) < 0)
	{
		if (errno != EINTR)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;

	_stream_select_get_fd (rs, &r);
	_stream_select_get_fd (ws, &w);
	_stream_select_get_fd (es, &e);

	return qu_integer_newi (res);
}



/*) sub class fsize (x, timeout = 30)
	This is a convenience method to get the v{st_size} field of #{fstat}.

	@exc
	.EArgValue
	.ESystem
*/
static qu_any qu__stream_fsize (qu_class self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (1, 2);
	{
		qu_any arg[3] = {argv[0], QuInt7, qu_int_new (30)};

		if (argc > 1)
			arg[2] = argv[1];

		return qu__stream_fstat (self, 3, arg);
	}
}



/*) sub class fstat (x, index = 0, timeout = 30)
	Get information on v{x}.

	If v{x} is a Stream then this method simply calls #{stat}.

	Otherwise v{x} must be a String path to a file in which case you can
	optionally set the number of seconds to v{timeout}.

	You can use the #{ismode} method to test for the v{st_mode} field.

	@exc
	.EArgValue
	.ESystem
*/
static qu_any qu__stream_fstat (qu_class self, int argc, qu_any *argv)
{
	qu_any x;
	long   i = -1;
	double to = 30;
	struct stat s;

	QU_ARG_VCOUNT (1, 3);
	x = argv[0];

	if (qu_stream_is (x))
	{
		if (argc == 1)
			return qu__stream_stat ((qu_stream) x, 0, NULL);
		else
		{
			qu_any arg = argv[1];
			return qu__stream_stat ((qu_stream) x, 1, &arg);
		}
	}
	else
	{
		QU_ARG_TYPE (x, qu_string);
		if (argc > 1)
		{
			QU_ARG_vlong (1, i);
			QU_REQUIRE (QuEIndex, 0 <= i && i <= 12);

			if (argc > 2)
				QU_ARG_vdouble (2, to);
		}

		__qu_blocking_begin;
		to += __qu_now__;
		while (stat (((qu_string) x)->str, &s) != 0)
		{
			if (errno != EINTR)
			{
				FIXEAGAIN;
				if (errno != EAGAIN)
				{
					__qu_blocking_end;
					QU_THROWERRNO;
				}
			}
			__qu_blocking_check_timeout (to);
			__qu_blocking_check_signal;
			qu_thread_yield ();
		}
		__qu_blocking_end;

		if (i >= 0)
			return _stream_stat_item (&s, i);
		else
		{
			qu_any arg[13] =
			{
				_stream_stat_item (&s,  0),
				_stream_stat_item (&s,  1),
				_stream_stat_item (&s,  2),
				_stream_stat_item (&s,  3),
				_stream_stat_item (&s,  4),
				_stream_stat_item (&s,  5),
				_stream_stat_item (&s,  6),
				_stream_stat_item (&s,  7),
				_stream_stat_item (&s,  8),
				_stream_stat_item (&s,  9),
				_stream_stat_item (&s, 10),
				_stream_stat_item (&s, 11),
				_stream_stat_item (&s, 12)
			};
			return qu_record_new (StatStruct, 13, arg);
		}
	}
}



/*) sub class fsync -> Null
	Synchronize all Stream with hardware.

	See C{man 2 sync} for details.

	@exc
	.ESecurity
*/
static qu_null qu__stream_fsync (qu_class self)
{
	__qu_unsafe__;

	__qu_blocking_begin;
	sync ();
	__qu_blocking_end;

	return QuNil;
}



/*) sub class ismode (mode: Integer, filter: String) -> Boolean
	This method is used to test the v{st_mode} field of #{stat} or #{fstat}.

	See -{File.istype} for valid v{filter} value.

	@exc
	.EArgValue

	@xmp
		m = Stream.fmode ('/tmp')
		if (Stream.ismode (m, 'd urwx'))
			println ('directory, user have rwx permission')
		;;
*/
static qu_boolean qu__stream_ismode (qu_class self, qu_integer mode, qu_string f)
{
#define NFILTERS 19
	static int list[NFILTERS] =
	{
		S_IFSOCK,
		S_IFLNK,
		S_IFREG,
		S_IFBLK,
		S_IFDIR,
		S_IFCHR,
		S_IFIFO,
		S_ISUID,
		S_ISGID,
		S_ISVTX,
		S_IRUSR,
		S_IWUSR,
		S_IXUSR,
		S_IRGRP,
		S_IWGRP,
		S_IXGRP,
		S_IROTH,
		S_IWOTH,
		S_IXOTH
	};

	int i, m, v;

	QU_ARG_int (mode, m);
	QU_ARG_CHECK (qu_string_is (f) && qu_file_get_filter_mode (f->str, &v));

	for (i = 0; i < NFILTERS; ++i)
	{
		int x = list[i];

		if ((v & x) && !(m & x))
			return QuFalse;
	}
	return QuTrue;
}



/*) sub class stderr (*arg) -> Stream
	@def
	sub class stderr -> Stream
	sub class stderr (x: Stream) -> Stream

	@p
	Get or set the standard error Stream object.

	If called without any argument, return the current standard error
	Stream object.

	If called with one argument, set the standard error and return
	the previous Stream object. The argument must be a Stream object.

	@exc
	.EArgValue
	.ESecurity

	@see
	#{__stderr__} #{stdin} #{stdout}
*/
static qu_stream qu__stream_stderr (qu_class self, int argc, qu_any *argv)
{
	qu_stream old = QuStreamStderr;

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc)
	{
		qu_stream f;
		QU_ARG_VTYPE (0, f, qu_stream);
		QuStreamStderr = f;
	}
	return old;
}



/*) sub class stdin (*arg) -> Stream
	@def
	sub class stdin -> Stream
	sub class stdin (x: Stream) -> Stream

	@p
	Get or set the standard input Stream object.

	If called without any argument, return the current standard input
	Stream object.

	If called with one argument, set the standard input and return
	the previous object. The argument must be a Stream object.

	@exc
	.EArgValue
	.ESecurity

	@see
	#{__stdin__} #{stderr} #{stdout}
*/
static qu_stream qu__stream_stdin (qu_class self, int argc, qu_any *argv)
{
	qu_stream old = QuStreamStdin;

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc)
	{
		qu_stream f;
		QU_ARG_VTYPE (0, f, qu_stream);
		QuStreamStdin = f;
	}
	return old;
}



/*) sub class stdout (*arg) -> Stream
	@def
	sub class stdout -> Stream
	sub class stdout (x: Stream) -> Stream

	@p
	Get or set the standard output Stream object.

	If called without any argument, return the current standard output
	Stream object.

	If called with one argument, set the standard output and return
	the previous object. The argument must be a Stream object.

	@exc
	.EArgValue
	.ESecurity

	@xmp
		println ('starting ...')
		f = File.open ('/foo/bar.txt', 'w')
		old = Stream.stdout (f)
		println ('Hello World!')
		Stream.stdout (old)
		f.close
		println ('done!')

	@see
	#{__stdout__} #{stderr} #{stdin}
*/
static qu_stream qu__stream_stdout (qu_class self, int argc, qu_any *argv)
{
	qu_stream old = QuStreamStdout;

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc)
	{
		qu_stream f;
		QU_ARG_VTYPE (0, f, qu_stream);
		QuStreamStdout = f;
	}
	return old;
}



/*) sub class ftty (x, klass = Boolean) -> Any
	Check whether v{x} is connected to a I{tty} device (terminal).

	v{x} may be a Stream object or a descriptor.

	If v{klass} is Boolean then this method returns true or false,
	otherwise the I{tty} name is returned.

	If v{klass} is not Boolean and v{x} is not connected to a I{tty} then an
	empty string is returned.

	See C{man 3 isatty} and C{man 3 ttyname} for details.

	@exc
	.EArgValue

	@xmp
		println (Stream.tty (Stream.stdout))
		println (Stream.tty (Stream.stdout, String))

	@out
		true
		/dev/pts/4
*/
static qu_any qu__stream_ftty (qu_class self, int argc, qu_any *argv)
{
	int    fd;
	qu_any x;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];

	if (qu_stream_is (x))
		fd = ((qu_stream) x)->fd;
	else
		QU_ARG_int (x, fd);

	if (argc > 1 && argv[1] != QuBooleanClass)
	{
		char *s = ttyname (fd);
		return s ? qu_string_newz (s) : QuStringEMPTY;
	}
	return (fd >= 0 && isatty (fd)) ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

void qu__stream__destroy (qu_stream self)
{
	if (self->fd >= 0)
	{
		while (close (self->fd) != 0)
		{
			if (errno != EINTR)
			{
				/* should never happen */
				QU_WARN ("cant close %d", self->fd);
				break;
			}
		}
	}
}


void qu__stream__mark (qu_stream self)
{
	qu_gc_xmark (self->br);
	qu_gc_xmark (self->obj);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __each -> Iterator
	Return an -{Iterator} object to iterate through each byte of the Stream.

	Each byte is reported as a -{String} object containing a single character.

	@warn
	If you close the Stream or change the position within the loop then the
	result is unpredictable.

	@noexc
*/
static qu_any qu__stream__next (qu_stream self, qu_iterator i)
{
	qu_any    arg[2] = { QuInt1, QuTrue };
	qu_string s = qu__stream_read (self, 2, arg);
	QU_ENSURE (s);
	QU_ITERATOR_STOP (s->len > 0);
	return s;
}


static qu_iterator qu__stream__each (qu_stream self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, qu__stream__next);
	return i;
}



/*) sub __eq (x) -> Boolean
	Whether self and v{x} are the very same Stream object.

	@noexc
*/
static qu_boolean qu__stream__eq (qu_stream self, qu_stream x)
{
	return (self == x && self->sid == x->sid) ? QuTrue : QuFalse;
}


/*) sub __getitem (pos: Integer) -> String
	Read the character at a specific position v{pos}.

	The Stream must support #{seek}.

	A negative position means counting from the end of the Stream.

	@warn
	After this method, the read/write position will be set after the
	character read.

	@exc
	.EArgValue
	.EIndex
	.ESystem
*/
static qu_string qu__stream__getitem (qu_stream self, qu_integer pos)
{
	qu_any     arg[2] = { QuInt1, QuTrue };
	qu_integer len;

	len = qu__stream__len (self);
	QU_ENSURE (len);
	if (qu_sys_lt (pos, QuInt0) == QuTrue)
	{
		pos = qu_sys_plus (pos, len);
		QU_ARG_CHECK (pos);
	}
	QU_REQUIRE (QuEIndex, qu_sys_ge (pos, QuInt0) == QuTrue && qu_sys_lt (pos, len) == QuTrue);
	QU_ENSURE (qu__stream_seek (self, 1, (qu_any *) &pos));
	return qu__stream_read (self, 2, arg);
}



/*) sub __getslice (pos: Integer, max: Integer) -> String
	Read characters at a specific position v{pos}.

	The Stream must support #{seek}.

	A negative position means counting from the end of the Stream.

	The result is a string at most v{max} long.

	@warn
	You can not start at a position beyond the end of the Stream.

	After this method, the read/write position will be set after the characters read.

	@exc
	.EArgValue
	.EIndex
	.ESystem
*/
static qu_string qu__stream__getslice (qu_stream self, qu_integer pos, qu_integer max)
{
	qu_any     arg[2] = { max, QuTrue };
	qu_integer len;

	len = qu__stream__len (self);
	QU_ENSURE (len);
	if (qu_sys_lt (pos, QuInt0) == QuTrue)
	{
		pos = qu_sys_plus (pos, len);
		QU_ARG_CHECK (pos);
	}
	QU_REQUIRE (QuEIndex, qu_sys_ge (pos, QuInt0) == QuTrue && qu_sys_lt (pos, len) == QuTrue);
	len = qu_sys_minus (len, pos);
	QU_ENSURE (len);
	if (qu_sys_lt (len, max) == QuTrue)
		arg[0] = len;
	QU_ENSURE (qu__stream_seek (self, 1, (qu_any *) &pos));
	return qu__stream_read (self, 2, arg);
}



/*) sub __hash -> Int
	Return the hash value of this Stream.

	@noexc
*/
static qu_int qu__stream__hash (qu_stream self)
{
	return qu_int_newl (self->sid);
}



/*) sub __in (x) -> Boolean
	Return true if the file contains the string v{x}, false otherwise.

	This method simply calls #{search} and ignores any exception. Thus if
	a file error false is returned and you can not know whether the result
	was caused by an error or the string was not actually found.

	If v{x} is not a -{String} then false is returned.

	@noexc

	@warn
	After a call to this method the read/write position is changed unless the
	Stream or string is empty.
*/
qu_boolean qu__stream__in (qu_stream self, qu_string s)
{
	if (!qu_string_is (s))
		return QuFalse;
	else
	{
		qu_any     argv[2] = {s, QuFalse};
		qu_integer r = qu__stream_search (self, 2, argv);
		return (r && r != (qu_integer) QuInt_1) ? QuTrue : QuFalse;
	}
}



/*) sub __len -> Integer
	This is a convenience method to get the v{st_size} field of #{stat}.

	Note that you can not get the size of the k{null} Stream.
*/
static qu_integer qu__stream__len (qu_stream self)
{
	if (self == QuStreamNull)
		return (qu_integer) QuInt0;
	else
	{
		qu_any i = qu_int_new (7);
		return qu__stream_stat (self, 1, &i);
	}
}



/*) sub __like (x) -> Boolean
	Return true if v{x} is a Stream and both have the same opened descriptor,
	false otherwise.

	@noexc
*/
static qu_boolean qu__stream__like (qu_stream self, qu_any x)
{
	if (qu_stream_is (x))
	{
		if (self == (qu_stream) x || (self->fd == ((qu_stream) x)->fd && self->fd >= 0))
			return QuTrue;
	}
	return QuFalse;
}



/*) sub __pop -> String
	Read everything from the current position to the end of the Stream.

	The Stream must support #{seek}.

	@exc
	.EMemory
	.ESystem
*/
static qu_string qu__stream__pop (qu_stream self)
{
	if (self == QuStreamNull)
	{
		return QuStringEMPTY;
	}
	else
	{
		qu_any     arg [2] = {QuInt1, QuTrue};
		qu_integer len, ofs;

		len = qu__stream__len (self);     QU_ENSURE (len);
		ofs = qu__stream_tell (self);     QU_ENSURE (ofs);
		len = qu_sys_minus    (len ,ofs); QU_ENSURE (len);

		QU_MEM_CHECK (qu_sys_le (len, qu_int_new (QU_INT_MAX)) == QuTrue);

		arg[0] = len;
		return qu__stream_read (self, 2, arg);
	}
}



/*) sub __print -> Null
	Print the String returned by #{__string} to -{__stdout}.

	See -{print} for possible exceptions.
*/
static qu_null qu__stream__print (qu_stream self)
{
	char buf [256];

	sprintf (buf, "<%s %d>", self->klass->name->str, self->fd);
	QU_ENSURE (qu_sys_printz (buf, strlen (buf)));
	return QuNil;
}



/*) sub __push (x) -> self
	Write v{x} to Stream.

	If v{x} is not a String nor a Buffer then it's __string method is called.

	If self is the k{null} Stream then this method always return self
	and never throw an exception. Otherwise see #{write} for possible exceptions.

	@xmp
		__stdout__ <<< 123 <<< '\n'

	@out
		123
*/
static qu_stream qu__stream__push (qu_stream self, qu_stringy x)
{
	if (self != QuStreamNull)
	{
		if (qu_buffer_is (x))
		{
			QU_ENSURE (qu_stream_send (self, (qu_buffer) x, x->len, 1) >= 0);
		}
		else
		{
			if (!qu_string_is (x))
				x = qu_sys_string (x);
			QU_ENSURE (qu_stream_write (self, x->str, x->len, 1) >= 0);
		}
	}
	return self;
}



/*) sub __setitem (pos: Integer, s: String)
	Write a single character at a specific position v{pos}.

	The Stream must support #{seek}.

	A negative position means counting from the end of the Stream.

	v{s} must be at least one character long.

	On success the v{s} is returned.

	@warn
	After this method, the read/write position will be set after the character
	written.

	Remember that if you opened the Stream for read-only then you will get
	a EBADF exception.

	@exc
	.EArgValue
	.EIndex
	.ESystem
*/
static qu_string qu__stream__setitem (qu_stream self, qu_integer pos, qu_string c)
{
	qu_integer len;
	char       ch;

	QU_ARG_CHECK (qu_string_is (c) && c->len > 0);
	ch = c->str[0];
	len = qu__stream__len (self);
	QU_ENSURE (len);
	if (qu_sys_lt (pos, QuInt0) == QuTrue)
	{
		pos = qu_sys_plus (pos, len);
		QU_ARG_CHECK (pos);
	}
	QU_REQUIRE (QuEIndex, qu_sys_ge (pos, QuInt0) == QuTrue && qu_sys_lt (pos, len) == QuTrue);
	QU_ENSURE (qu__stream_seek (self, 1, (qu_any *) &pos));
	QU_ENSURE (qu_stream_write (self, &ch, 1, 1) >= 0);
	return c;
}



/*) sub __setslice (pos: Integer, max: Integer, s: String)
	Write characters at a specific position v{pos}.

	The Stream must support #{seek}.

	A negative position means counting from the end of the Stream.

	If v{s} is smaller than v{max} then only C{s.__len} bytes are written.
	Otherwise v{max} bytes are written.

	On success, v{s} is returned.

	@warn
	After this method, the read/write position will be set after the characters
	written.

	Remember that if you opened the Stream for read-only then you will get
	a EBADF exception.

	@exc
	.EArgValue
	.EIndex
	.ESystem
*/
static qu_string qu__stream__setslice (qu_stream self, qu_integer opos, qu_integer omax, qu_string s)
{
	qu_integer olen;
	long       pos, max, len, m;

	QU_ARG_CHECK (qu_string_is (s));
	QU_ARG_long (opos, pos);
	QU_ARG_long (omax, max);
	olen = qu__stream__len (self);
	QU_ENSURE (olen);
	QU_ARG_long (olen, len);
	QU_ENSURE_INDEX (pos, len);
	opos = qu_integer_newl (pos);
	QU_ENSURE (qu__stream_seek (self, 1, (qu_any *) &opos));

	m = len - pos;
	if (m > max)
		max = max;
	if (max > s->len)
		max = s->len;

	QU_ENSURE (qu_stream_write (self, s->str, max, 1) >= 0);
	return s;
}



/*) sub __string -> String
	Return a String representation of this Stream.

	@noexc
*/
static qu_string qu__stream__string (qu_stream self)
{
	char buf [256];

	sprintf (buf, "<%s %d>", self->klass->name->str, self->fd);
	return qu_string_newz (buf);
}



/*) sub __true -> Boolean
	Return true if the Stream is opened, false otherwise.

	If self is the k{null} Stream then this method always return true.

	@noexc
*/
static qu_boolean qu__stream__true (qu_stream self)
{
	if (self == QuStreamNull)
		return QuTrue;
	return (self->fd >= 0) ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub blocking (*arg) -> Boolean __final
	@def
	sub blocking -> Boolean
	sub blocking (setting: Boolean) -> Boolean
	@p
	Get or set Stream blocking mode.

	If called with no arguments, return whether or not the Stream is a blocking
	stream.

	If called with one argument, set the blocking mode.

	If the Stream is closed or does not support blocking mode then false is
	returned.

	@exc
	.EArgValue
	.ESystem

	@note
	The #{fcntl} method is used to set the mode with a O_NONBLOCK parameter.
*/
static qu_boolean qu__stream_blocking (qu_stream self, int argc, qu_any *argv)
{
	QU_ARG_VMAX (1);

	if (self->fd < 0 || !qu_stream_can_block (self))
		return QuFalse;

	if (argc == 0)
	{
		int v = fcntl (self->fd, F_GETFL, 0);
		return (v & O_NONBLOCK) ? QuFalse : QuTrue;
	}
	else
	{
		int r, v;

		QU_ARG_vbool (0, v);
		r = fcntl (self->fd, F_GETFL, 0);
		r = fcntl (self->fd, F_SETFL, v ? (r & ~O_NONBLOCK) : (r | O_NONBLOCK));
		if (r != 0)
			QU_THROWERRNO;
		return v ? QuTrue : QuFalse;
	}
}



/*) sub can (option: String) -> Boolean
	Get stream capability.

	See #{__inst} for the options.

	@warn
	An unknown option is ignored, so you better make sure there are none.

	@noexc
*/
static qu_boolean qu__stream_can (qu_stream self, qu_string op)
{
	char *z;

	QU_ARG_TYPE (op, qu_string);
	z = op->str;
	while (*z)
	{
		switch (*z++)
		{
			case 'c': case 'C': if (!qu_stream_can_close (self)) return QuFalse; continue;
			case 'b': case 'B': if (!qu_stream_can_block (self)) return QuFalse; continue;
			case 'e': case 'E': if (!qu_stream_can_eof   (self)) return QuFalse; continue;
		}
	}
	return QuTrue;
}



/*) sub close -> Null
	Close the Stream.

	This method does not throw any exception.
	If the stream is already closed, or is the k{null} Stream, then nothing happens.

	@noexc

	@warn
	If you mess with descriptors, say you close a Stream descriptor without
	going through this method, then you might get a -{__warning__}.

	... and that is a terrible idea.

	See the #{fd} method for more warnings.
*/
qu_null qu__stream_close (qu_stream self)
{
	if (qu_stream_can_close (self))
	{
		int eno = errno;

		if (self->fd >= 0)
		{
			while (close (self->fd) != 0)
			{
				if (errno != EINTR)
				{
					/* should never happen */
					QU_WARN ("cant close %d", self->fd);
					break;
				}
			}
			self->fd = -1;
		}
		errno = eno;
	}
	return QuNil;
}



/*) sub eachline -> Iterator
	Return an -{Iterator} object to iterate through the lines.

	The Stream must be opened and readable.

	@warn
	If you close the Stream or change the file position within the loop then
	the result is unpredictable.

	@noexc

	@xmp
		f = File.open ('/foo/bar.txt')
		for line in f.eachline
			print (line)
		;;

	@see
	#{readln}
*/
static qu_any qu__stream_eachline_next (qu_stream self, qu_iterator i)
{
	qu_string s = qu__stream_readln (self, 0, NULL);
	QU_ENSURE (s);
	QU_REQUIRE (QuEStop, s->len);
	return s;
}


static qu_iterator qu__stream_eachline (qu_stream self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, qu__stream_eachline_next);
	return i;
}



/*) sub eof -> Boolean
	Return true if the EOF (I{end of the file}) is reached, false otherwise.

	The Stream must support #{seek} otherwise false is returned.

	@noexc
*/
static qu_boolean qu__stream_eof (qu_stream self)
{
	qu_integer len = qu__stream__len (self);
	qu_integer pos;

	if (len == 0)
		return QuFalse;

	pos = qu__stream_tell (self);
	if (pos == 0)
		return QuFalse;

	return qu_sys_ge (pos, len);
}



/*) sub get (len: Integer) -> String
	This is a convenience method that calls #{read}C{(len, true)}.
*/
static qu_string qu__stream_get (qu_stream self, qu_integer len)
{
	qu_any argv[2] = {len, QuTrue};
	return qu__stream_read (self, 2, argv);
}



/*) sub fd -> Int
	Return the underlying file descriptor.

	@exc
	.EBADF Stream is closed or is the k{null} Stream

	@xmp
		println (Stream.stdin.fd)
		println (Stream.stdout.fd)
		println (Stream.stderr.fd)

	@out
		0
		1
		2

	@warn
	You must never mess with the file descriptor directly unless you know exactly
	what you are doing.

	@p
	Consider this scenario:
	@ul
	.you create a file, the file descriptor is, say 7
	.you manage to close the file descriptor without going through #{close}
	.you open another file, the descriptor is 7
	.you call the old file method, say #{write}, and you'll be writing to
	 to the new file instead of the old

	@p
	Qu could easily prevent this from happening and never let you know what
	the file descriptor is but that would limit experienced programmers and
	someone might just write a C module that lets you do it anyway.
*/
static qu_int qu__stream_fd (qu_stream self)
{
	QU_REQUIRE (QuEBADF, self->fd >= 0);
	assert (0 <= self->fd && self->fd <= QU_INT_MAX);
	return qu_int_new (self->fd);
}



/*) sub fcntl (cmd: Integer, *arg) -> Integer __final
	@def
	sub fcntl (cmd: Integer) -> Integer
	sub fcntl (cmd: Integer, arg: Integer) -> Integer
	@p
	Perform one of various miscellaneous operations on the Stream.

	See C{man 2 fcntl} for details.

	@exc
	.EArgValue
	.ESystem

	@note
	Use the %{Os} module for the constants.
*/
static qu_any qu__stream_fcntl (qu_stream self, int argc, qu_any *argv)
{
	double to = self->timeout;
	long   arg = 0;
	int    cmd, r;

	QU_ARG_VMAX (2);
	QU_ARG_vint (0, cmd);
	if (argc > 1)
		QU_ARG_vlong (2, arg);

	__qu_blocking_begin;
	for (;;)
	{
		if (argc > 1)
			r = fcntl (self->fd, cmd, arg);
		else
			r = fcntl (self->fd, cmd);

		if (r >= 0)
			break;

		if (errno != EINTR)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_check_timeout (to);
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;
	return qu_integer_newi (r);
}



/*) sub flush -> Null
	Flush the stream buffer.

	Subclasses must override this method if flush is supported.

	This method must throw an exception if flush fails, return nil otherwise.

	This default method does nothing and return nil.

	@noexc
*/
static qu_null qu__stream_flush (qu_stream self)
{
	return QuNil;
}



/*) sub ioctl (request: Integer, *arg) -> Integer __final
	Manipulate the underlying device parameters of the Stream.

	See C{man 2 ioctl) for details.

	v{arg} is 0 to 15 Pointer arguments.
	The pointers are passed as the arguments to the system f{ioctl} function.

	@warn
	This is an extremely sensitive method.
	Make sure you know what you are doing.

	@exc
	.EImplement f{ioctl} is not available on your system
	.EArgValue

	@note
	Use the %{Os} module for the constants.
*/
static qu_integer qu__stream_ioctl (qu_stream self, int argc, qu_any *argv)
{
#ifdef HAVE_SYS_IOCTL_H
	double to = self->timeout;
	int    fd = self->fd;
	int    r, i;
	void  *x[16];

	QU_ARG_VCOUNT (1, 16);
	QU_ARG_vint (0, r);
	++argv;
	--argc;
	for (i = 0; i < argc; i++)
	{
		qu_pointer p = *argv++;
		QU_ARG_TYPE (p, qu_pointer);
		x[i] = p->ptr;
	}
	__qu_blocking_begin;
	for (;;)
	{
		switch (argc)
		{
			case  1: r = ioctl (fd, r, x[0]); break;
			case  2: r = ioctl (fd, r, x[0], x[1]); break;
			case  3: r = ioctl (fd, r, x[0], x[1], x[2]); break;
			case  4: r = ioctl (fd, r, x[0], x[1], x[2], x[3]); break;
			case  5: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4]); break;
			case  6: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5]); break;
			case  7: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6]); break;
			case  8: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7]); break;
			case  9: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8]); break;
			case 10: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9]); break;
			case 11: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10]); break;
			case 12: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11]); break;
			case 13: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11], x[12]); break;
			case 14: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11], x[12], x[13]); break;
			case 15: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11], x[12], x[13], x[14]); break;
			case 16: r = ioctl (fd, r, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11], x[12], x[13], x[14], x[15]); break;
		}

		if (r >= 0)
			break;

		if (errno != EINTR)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_check_timeout (to);
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;
	return qu_integer_newi (r);
#else
	QU_THROW (QuEImplement);
#endif
}



/*) sub linebreak (*arg) -> String
	@def
	sub linebreak -> String
	sub linebreak (marker: String) -> String

	@p
	Subclasses must override this method if #{readln} and/or #{writeln}
	is supported and setting the linebreak is relevant.

	If called without any argument then this method must return the linebreak
	marker.

	If called with one argument then this method must set the linebreak marker
	or throw EImplement if setting is not possible.

	This default method accepts any linebreak provided it is a String.

	@exc
	.EArgValue v{marker} not a String

	@xmp
		println (Stream.stdin.linebreak.quote)
		Stream.stdin.linebreak = "\r\n"
		println (Stream.stdin.linebreak.quote)

	@out
		'\n'
		'\r\n'
*/
static qu_string qu__stream_linebreak (qu_stream self, int argc, qu_any *argv)
{
	QU_ARG_VMAX (1);
	if (argc)
	{
		qu_string s;
		QU_ARG_VTYPE (0, s, qu_string);
		self->br = s;
	}
	return self->br;
}



/*) sub len -> Int
	This is an alias of #{__len}.
*/



/*) sub lock (mode = 'r') -> self
	Lock a Stream.

	The Stream must be opened and supports locking.

	v{mode} must be one of the following:
	@ops
	.| C{r} | shared lock
	.| C{w} | exclusive lock

	@p
	Note that ETimeout is thrown if the lock can not be acquired until
	#{timeout}.

	Note that you can not lock the k{null} Stream.

	@exc
	.EArgValue
	.EImplement
	.ESystem
*/
static qu_null qu__stream_lock (qu_stream self, int argc, qu_any *argv)
{
	double to   = self->timeout;
	int    mode = F_RDLCK;
	struct flock k;

	if (self == QuStreamNull)
		return QuNil;

	QU_ARG_VMAX (1);
	if (argc)
	{
		qu_string s;
		QU_ARG_VTYPE (0, s, qu_string);
		switch (s->str[0])
		{
			case 'r': break;
			case 'w': mode = F_WRLCK; break;
			default : QU_ARG_THROW;
		}
	}

	k.l_type   = mode;
	k.l_whence = SEEK_SET;
	k.l_start  = k.l_len = 0;

	__qu_blocking_begin;
	while (fcntl (self->fd, F_SETLK, &k) != 0)
	{
		if (errno != EINTR)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_check_timeout (to);
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;
	return QuNil;
}



/*) sub locked -> Boolean
	Return true if the file is locked, false otherwise.

	If this method throws then it may be because the stream does not
	support locking or the locking information can not be acquired.

	Note that ETimeout is possible if the lock information could not be
	acquired until #{timeout}.

	If this is the k{null} Stream then this method returns false.

	@exc
	.ESystem
*/
static qu_boolean qu__stream_locked (qu_stream self)
{
	if (self == QuStreamNull)
		return QuFalse;
	else
	{
		double to = self->timeout;
		struct flock k;

		k.l_type   = F_WRLCK;
		k.l_whence = SEEK_SET;
		k.l_start  = k.l_len = 0;

		__qu_blocking_begin;
		while (fcntl (self->fd, F_GETLK, &k) != 0)
		{
			if (errno != EINTR)
			{
				FIXEAGAIN;
				if (errno != EAGAIN)
				{
					__qu_blocking_end;
					QU_THROWERRNO;
				}
			}
			__qu_blocking_check_timeout (to);
			__qu_blocking_check_signal;
			qu_thread_yield ();
		}
		__qu_blocking_end;
		return (k.l_type == F_UNLCK) ? QuFalse : QuTrue;
	}
}



/*) sub mode -> Float
	This is a convenience method to get the v{st_mode} field of #{stat}.
*/
static qu_float qu__stream_mode (qu_stream self)
{
	qu_any i = QuInt2;
	return qu__stream_stat (self, 1, &i);
}



/*) sub mtime -> Float
	This is a convenience method to get the v{st_mtime} field of #{stat}.
*/
static qu_float qu__stream_mtime (qu_stream self)
{
	qu_any i = QuInt10;
	return qu__stream_stat (self, 1, &i);
}



/*) sub obj -> Any
	Return the underlying object.

	Normally this an object that attached itself to the Stream and redirected
	the Stream's I/O methods.

	For a Socket stream, this may be an Ssl object.

	@exc
	.ENone there is none

	@note
	In Qu you can redirect a Stream's I/O methods but currently this can only
	be done in C level.
*/
static qu_any qu__stream_obj (qu_stream self)
{
	QU_REQUIRE (QuENone, self->obj);
	return self->obj;
}



/*) sub opened -> Boolean
	Return true if the stream is opened, false otherwise.

	This method must not throw any exception.

	If self is the k{null} Stream then this method returns true.

	@exc
*/
static qu_boolean qu__stream_opened (qu_stream self)
{
	if (self == QuStreamNull)
		return QuTrue;
	return (self->fd >= 0) ? QuTrue : QuFalse;
}



/*) sub print (*arg) -> Null
	Write all arguments to Stream.

	The items in v{arg} are converted to -{String} by calling their f{__string}
	method as necessary.

	If self is the k{null} Stream the nothing happens and this method
	returns nil.

	@exc
	.ESystem

	@xmp
		Stream.stdout.print ('foo', 'bar', 'baz')

	@out
		foobarbaz
*/
static qu_null qu__stream_print (qu_stream self, int argc, qu_any *argv)
{
	while (argc--)
	{
		qu_any x = qu_sys_string (*argv++);
		QU_ENSURE (qu__stream_write (self, 1, &x));
	}
	return QuNil;
}



/*) sub println (*arg) -> Null
	This simply calls the #{print} method twice. Once for the arguments
	and once for the #{linebreak}.

	@exc
	.ESystem

	@xmp
		Stream.stdout.println ('foo', 'bar', 'baz')
		Stream.stdout.println ('done')

	@out
		foobarbaz
		done
*/
static qu_null qu__stream_println (qu_stream self, int argc, qu_any *argv)
{
	qu_any s = self->br;
	QU_ENSURE (qu__stream_print (self, argc, argv));
	QU_ENSURE (qu__stream_print (self, 1, &s));
	return QuNil;
}



/*) sub put (s: String) -> Integer
	This is a convenience method that calls #{write}C{(s)}.
*/
static qu_integer qu__stream_put (qu_stream self, qu_string s)
{
	return qu__stream_write (self, 1, (qu_any *) &s);
}



/*) sub read (len = 1, all = false) -> String
	If v{all} is true then read exactly v{len} bytes, otherwise read at
	most v{len} bytes.

	If called without any arguments, it means to read at least one byte
	until #{timeout}.

	v{len} must not be bigger than the system SSIZE_MAX value. This is a
	large value and should never be applied. As long as you use a "reasonable"
	size then you don't need to know what SSIZE_MAX is.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.EArgValue
	.EMemory
	.ESystem
*/
static qu_string qu__stream_read (qu_stream self, int argc, qu_any *argv)
{
	int  all = 0;
	long l = 1;

	QU_ARG_VMAX (2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, l);
		QU_ARG_CHECK (0 <= l && l <= SSIZE_MAX && l < QU_MEM_MAX);

		if (argc > 1)
			QU_ARG_vbool (1, all);
	}
	return qu_stream_reads (self, l, all);
}



/*) sub readall (len: Integer) -> String
	This is just a convenient function that calls #{read}C{(len, true)}.
*/
static qu_string qu__stream_readall (qu_stream self, qu_integer n)
{
	long l;

	l = qu_integer_touli (n);
	QU_ARG_CHECK (errno == 0 && l <= SSIZE_MAX && l < QU_MEM_MAX);
	return qu_stream_reads (self, l, 1);
}



/*) sub readln (max = 8192) -> String
	Read a line from the Stream.

	This method keeps reading until #{linebreak} is encountered or v{max} bytes
	have been read, or throws ETimeout if #{timeout} is reached.

	v{max} should never be a huge number or you risk out of memory.

	If there is an error then the error is thrown if nothing has been read,
	otherwise the error is ignored and anything read until the error is
	returned. You can use #{select} to test whether the stream has an error.

	Note that the returned value may or may not contain the #{linebreak} marker
	depending on v{max}.

	If self is the k{null} Stream then EBADF is thrown.

	@note
	You can test whether or not the result ends with the #{linebreak} with
	-{String.endwith}.

	@exc
	.EArgValue
	.EMemory
	.ESystem
*/
static qu_string qu__stream_readln (qu_stream self, int argc, qu_any *argv)
{
	long max = BUFSIZ;

	QU_ARG_VMAX (1);
	if (argc)
	{
		QU_ARG_vlong (0, max);
		QU_MEM_CHECK (0 <= max && max <= SSIZE_MAX);
	}

	return qu_stream_readsln (self, max);
}



/*) sub readv (spec) -> Any
	Read data.

	v{spec} is the specification of the arguments v{arg}. It may be a String
	or a Struct. See -{Struct.spec} for details.

	If reading one data then return that data. If reading more than one then
	return an Array.

	@exc
	.EArgValue
	.ESystem
*/
static qu_any qu__stream_readv (qu_stream self, qu_any spec)
{
	/* TODO - qu_struct.c needs more functions */

	char     buf [128];
	qu_array a = 0;
	qu_any   x;

	if (qu_string_is (spec))
	{
		char *p = ((qu_string) spec)->str;
		long  l = ((qu_string) spec)->len;

		if (l > 1)
			a = qu_array_new (l);

		while (l--)
		{
			char              t = *p++;
			qu_struct_type_s *i = qu_struct_type_info (t);
			long              z = i->size;

			QU_ARG_CHECK (z);
			assert (z < sizeof (buf));
			QU_ENSURE (qu_stream_read (self, buf, z, 1) == z);
			x = qu_pointer_get (buf, t);
			QU_ENSURE (x);
			if (a == 0)
				return x;
			qu_array_push (a, x);
		}
		return a;
	}
	else
	{
		QU_ARG_TYPE (spec, qu_struct);
		{
			qu_struct         u = spec;
			long              l = qu_struct_len (u);
			long              i;
			qu_structentry_s *t = u->items;

			for (i = 0; i < l; ++i)
			{
				qu_structentry_s *e = &t[i];
				char              t = qu_struct_type_to_char (e->type);
				qu_struct_type_s *i = qu_struct_type_info (t);
				long              z = i->size;

				QU_ARG_CHECK (z);
				assert (z < sizeof (buf));
				QU_ENSURE (qu_stream_read (self, buf, z, 1) == z);
				x = qu_pointer_get (buf, t);
				QU_ENSURE (x);
				if (a == 0)
					return x;
				qu_array_push (a, x);
			}
			return a;
		}
	}
	return self;
}



/*) sub read8 -> Int
	Read a 8-bits signed integer.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.ESystem
*/
static qu_int qu__stream_read8 (qu_stream self)
{
	char v;
	long n = qu_stream_read (self, &v, 1, 1);
	QU_ENSURE (n == 1);
	return qu_int_newi ((int) v);
}



/*) sub read16 -> Int
	Read a 16-bits signed integer.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.ESystem
*/
static qu_int qu__stream_read16 (qu_stream self)
{
#if SIZEOF_SHORT == 2
	short v;
	long  n = qu_stream_read (self, (char *) &v, 2, 1);
	QU_ENSURE (n == 2);
	return qu_int_newh (v);
#else
	QU_THROWZ (QuEImplement, "native-16 bits integer not available");
#endif
}



/*) sub read32 -> Integer
	Read a 32-bits signed integer.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.ESystem
*/
static qu_integer qu__stream_read32 (qu_stream self)
{
#if SIZEOF_INT == 4
	int  v;
	long n = qu_stream_read (self, (char *) &v, 4, 1);
	QU_ENSURE (n == 4);
	return qu_integer_newi (v);
#elif SIZEOF_LONG == 4
	long v;
	long n = qu_stream_read (self, (char *) &v, 4, 1);
	QU_ENSURE (n == 4);
	return qu_integer_newl (v);
#else
	QU_THROWZ (QuEImplement, "native-32 bits integer not available");
#endif
}



/*) sub readu8 -> Int
	Read a 8-bits unsigned integer.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.ESystem
*/
static qu_int qu__stream_readu8 (qu_stream self)
{
	unsigned char v;
	QU_ENSURE (qu_stream_read (self, (char *) &v, 1, 1) == 1);
	return qu_int_newui ((unsigned int) v);
}



/*) sub readu16 -> Int
	Read a 16-bits unsigned integer.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.ESystem
*/
static qu_int qu__stream_readu16 (qu_stream self)
{
#if SIZEOF_SHORT == 2
	unsigned short v;
	QU_ENSURE (qu_stream_read (self, (char *) &v, 2, 1) == 2);
	return qu_int_newuh (v);
#else
	QU_THROWZ (QuEImplement, "native-16 bits integer not available");
#endif
}



/*) sub readu32 -> Integer
	Read a 32-bits unsigned integer.

	If self is the k{null} Stream then EBADF is thrown.

	@exc
	.ESystem
*/
static qu_integer qu__stream_readu32 (qu_stream self)
{
#if SIZEOF_INT == 4
	unsigned int v;
	QU_ENSURE (qu_stream_read (self, (char *) &v, 4, 1) == 4);
	return qu_integer_newui (v);
#elif SIZEOF_LONG == 4
	unsigned long v;
	QU_ENSURE (qu_stream_read (self, (char *) &v, 4, 1) == 4);
	return qu_integer_newul (v);
#else
	QU_THROWZ (QuEImplement, "native-32 bits integer not available");
#endif
}



/*) sub recv (b: Buffer, n: Integer, all = false) -> Int
	Read from Stream to Buffer.

	@exc
	.EArgValue
	.EOverflow
	.ESystem

	@xmp
		b = Buffer (80)
		f = File ('/foo/bar.txt')
		n = f.recv (b, 10)
		f.close
*/
static qu_int qu__stream_recv (qu_stream self, int argc, qu_any *argv)
{
	qu_buffer b;
	long   n;
	int    all = 0;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_VTYPE  (0, b, qu_buffer);
	QU_ARG_vlong (1, n);
	QU_ARG_CHECK  (n >= 0);
	if (argc > 2)
		QU_ARG_vbool (2, all);
	if (n == 0)
		return QuInt0;
	n = qu_stream_recv (self, b, n, all);
	QU_ENSURE (n >= 0);
	return qu_int_new (n);
}



/*) sub recvln (b: Buffer, max: Integer) -> Int
	Read a line from Stream to Buffer.

	@exc
	.EArgValue
	.EOverflow
	.ESystem

	@xmp
		b = Buffer (80)
		f = File ('/foo/bar.txt')
		n = f.recvln (b, 80)
		f.close
*/
static qu_int qu__stream_recvln (qu_stream self, qu_buffer b, qu_integer max)
{
	long n;

	QU_ARG_CHECK (qu_buffer_is (b));
	QU_ARG_long (max, n);
	QU_ARG_CHECK (n >= 0);
	if (n == 0)
		return QuInt0;
	n = qu_stream_recvln (self, b, n);
	QU_ENSURE (n >= 0);
	return qu_int_new (n);
}



/*) sub rewind -> self
	Set read/write position to the beginning of Stream.

	This is a convenience method that simply calls #{seek} (0).

	Note that rewinding the k{null} Stream has no effect.

	@exc
	.ESystem
*/
static qu_stream qu__stream_rewind (qu_stream self)
{
	if (self != QuStreamNull)
	{
		if (lseek (self->fd, 0, SEEK_SET) != 0)
			QU_THROWERRNO;
	}
	return self;
}



/*) sub search (s: String) -> Integer
	Search Stream starting from the current position.

	If the string is empty or not found then -1 is returned, otherwise
	the offset from the start of the file where the string is found is returned.

	@exc
	.EArgValue
	.ESystem

	@warn
	After a call to this method the read/write position is changed unless the
	Stream or string is empty.
*/
static qu_integer qu__stream_search (qu_stream self, int argc, qu_any *argv)
{
	/*
		TODO - need a better algorithm
	*/

	qu_string  s;
	int     nocase = 0;
	long    l;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, s, qu_string);
	if (argc > 1)
		QU_ARG_vbool (1, nocase);

	l = s->len;
	if (l)
	{
		qu_any     aseek [2] = {QuInt0, qu_string_byte ('c')};
		qu_any     aread [2] = {qu_int_newl (l), QuTrue};
		char      *z = s->str;
		qu_integer start;
		qu_string  r;

		start = qu__stream_seek (self, 2, aseek);
		QU_ENSURE (start);

		aseek [0] = start;
		aseek [1] = qu_string_byte ('s');

		for (;;)
		{
			r = qu__stream_read (self, 2, aread);
			if (r == 0 || r->len < l)
				break;

			if (nocase)
			{
				if (qu_strz_eqnc (r->str, z, l))
					return start;
			}
			else
			{
				if (qu_strz_eq (r->str, z, l))
					return start;
			}

			start = qu_sys_incr (start);
			QU_ENSURE (start);
			aseek[0] = start;
			QU_ENSURE (qu__stream_seek (self, 2, aseek));
		}
	}
	return (qu_integer) QuInt_1;
}



/*) sub seek (ofs: Integer, whence = 's') -> Integer
	Set file read/write position then return the position.

	v{whence} must be one of:
	@ops
	.| C{s} | set from start of file
	.| C{e} | set from end of file
	.| C{c} | set from current position

	@p
	If self is the k{null} Stream then 0 is returned.

	@warn
	f{seek} can place the read/write position beyond the end of the Stream.

	@exc
	.EArgValue
	.ESystem

	@xmp
		f = File.open ('/foo/bar.txt')
		f.seek (-10, 'e') # move to 10 bytes before the end of the file
*/
static qu_integer qu__stream_seek (qu_stream self, int argc, qu_any *argv)
{
	off_t pos;
	int   whence = SEEK_SET;

#if SIZEOF_OFF_T > SIZEOF_LONG /* implies HAVE_LONG_LONG */
	long long  ofs;
	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlonglong (0, ofs);
#else
	long ofs;
	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong (0, ofs);
#endif

	if (argc > 1)
	{
		qu_string x;
		QU_ARG_VTYPE (1, x, qu_string);
		switch (x->str[0])
		{
			case 's': whence = SEEK_SET; break;
			case 'e': whence = SEEK_END; break;
			case 'c': whence = SEEK_CUR; break;

			default: QU_ARG_THROW;
		}
	}

	if (self == QuStreamNull)
		return (qu_integer) QuInt0;

	if (self->obj)
		pos = self->funcs->oseek (self->obj, (off_t) ofs, whence);
	else
		pos = self->funcs->seek (self->fd, (off_t) ofs, whence);

	if (pos < 0)
		QU_THROWERRNO;

#if SIZEOF_OFF_T > SIZEOF_LONG
	return qu_integer_newll ((long long) pos);
#else
	return qu_integer_newl ((long) pos);
#endif

}



/*) sub select (how: String, timeout = 1000) -> Boolean __final
	Check for socket status.

	See C{man 2 select} for more information.

	v{timeout} is the time to wait in microsecond.

	v{how} is one of:
	@ops
	.| C{r} | whether the Stream is ready for reading
	.| C{w} | whether the Stream is ready for writing
	.| C{e} | whether the Stream has error

	@p
	Note that if the Stream is valid (not closed) then f{select} should not throw any
	exception except ENOMEM (insufficient memory), which is rare and most probably
	never happen.

	@exc
	.EArgValue
	.ESystem

	@see
	#{fselect} to check for multiple Streams.
*/
static qu_boolean qu__stream_select (qu_stream self, int argc, qu_any *argv)
{
	if (self == QuStreamNull)
		return QuTrue;
	else
	{
		qu_string how;
		int       r;
		long      usec = 1000;
		char      w;

		QU_ARG_VCOUNT (1, 2);
		QU_ARG_VTYPE (0, how, qu_string);
		if (argc > 1)
			QU_ARG_vlong (1, usec);
		w = how->str[0];
		QU_ARG_CHECK (w == 'r' || w == 'w' || w == 'e');
		__qu_blocking_begin;
		r = qu_stream_select (self, w, usec);
		__qu_blocking_end;
		if (r < 0)
			QU_THROWERRNO;
		return r ? QuTrue : QuFalse;
	}
}



/*) sub send (b:Buffer, len: Integer = self.len, all = true) -> Int
	Write Buffer content to Stream.

	The Buffer is consumed.

	If v{len} is negative it means C{b.len}.

	@exc
	.EArgValue
	.ESystem

	@xmp
		b = Buffer ('Hello\n')
		println (__stdout__.send (b))
		println (b.len)

	@out
		Hello
		6
		0
*/
static qu_int qu__stream_send (qu_stream self, int argc, qu_any *argv)
{
	qu_buffer b;
	long      len;
	int       all = 1;

	QU_ARG_VCOUNT (1, 3);
	QU_ARG_VTYPE (0, b, qu_buffer);
	len = b->len;
	if (argc > 1)
	{
		QU_ARG_vlong (1, len);
		if (len < 0)
			len = b->len;
		else
			QU_ARG_CHECK (len >= 0);
		if (argc > 2)
			QU_ARG_vbool (2, all);
	}
	if (len == 0)
		return QuInt0;
	len = qu_stream_send (self, b, len , all);
	QU_ENSURE (len >= 0);
	return qu_int_new (len);
}



/*) sub stat (*arg) -> Any
	@def
	sub stat (index: Integer) -> Any
	sub stat -> Record
	@p
	Get information on the Stream.

	If called with an Integer argument then only get a specific field of
	#{Stat}, otherwise return a Stat -{Record}.

	You can use the #{ismode} method to test for the v{st_mode} field.

	Note that you can not stat the k{null} Stream.

	@exc
	.EArgValue
	.EIndex
	.ESystem
*/
static qu_any qu__stream_stat (qu_stream self, int argc, qu_any *argv)
{
	double to = self->timeout;
	long   i  = -1;
	struct stat s;

	QU_ARG_VMAX (1);
	if (argc)
	{
		QU_ARG_vlong (0, i);
		QU_REQUIRE (QuEIndex, 0 <= i && i <= 12);
	}

	__qu_blocking_begin;
	while (fstat (self->fd, &s) != 0)
	{
		if (errno != EINTR)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_check_timeout (to);
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;

	if (i >= 0)
		return _stream_stat_item (&s, i);
	else
	{
		qu_any arg[13] =
		{
			_stream_stat_item (&s,  0),
			_stream_stat_item (&s,  1),
			_stream_stat_item (&s,  2),
			_stream_stat_item (&s,  3),
			_stream_stat_item (&s,  4),
			_stream_stat_item (&s,  5),
			_stream_stat_item (&s,  6),
			_stream_stat_item (&s,  7),
			_stream_stat_item (&s,  8),
			_stream_stat_item (&s,  9),
			_stream_stat_item (&s, 10),
			_stream_stat_item (&s, 11),
			_stream_stat_item (&s, 12)
		};
		return qu_record_new (StatStruct, 13, arg);
	}
}



/*) sub sync -> self
	Synchronize Stream with hardware.

	See C{man 2 fsync} for details.

	@exc
	.ESystem
*/
static qu_stream qu__stream_sync (qu_stream self)
{
	if (fsync (self->fd) != 0)
		QU_THROWERRNO;
	return self;
}



/*) sub tell -> Integer
	Return the current stream read/write position.

	The Stream must support #{seek}.

	If self is the k{null} Stream the 0 is returned.

	@exc
	.ESystem
*/
static qu_integer qu__stream_tell (qu_stream self)
{
	qu_any arg[2] = {QuInt0, qu_string_byte ('c')};
	return qu__stream_seek (self, 2, arg);
}



/*) sub timeout (*arg) -> Real
	@def
	sub timeout -> Real
	sub timeout (delta: Real) -> Real
	@p
	Set or get the timeout, return the current value.

	Setting means to set it v{timeout} seconds away from now. An infp value
	means some thousands year from now.

	If this is a small value (such as 0 or negative) then you will instantly
	get a ETimeout exception when performing a timed operation.

	Suppose the current time is 01:23:45AM. Then setting a value of 10 means
	that the Stream will be timed-out at 01:23:55AM.

	Many Stream methods are timed operation such as #{read}, #{write}, etc.

	@exc
	.EArgValue

	@xmp
		println ("connecting ...")
		mysock.timeout (10)
		try
			mysock.connect (some_socket_address)
		catch ETimeout
			println ("cant connect within 10 seconds")
			throw
		;;
		println ("connected")
*/
static qu_real qu__stream_timeout (qu_stream self, int argc, qu_any *argv)
{
	QU_ARG_VMAX (1);

	if (argc > 0)
	{
		qu_any x = argv[0];
		double v;

		if (x == QuInfPos)
		{
			v = DBL_MAX;
		}
		else
		{
			QU_ARG_vdouble (0, v);
			v = __qu_now__ + v;
		}
		self->timeout = v;
	}

	if (self->timeout <= (double) QU_INT_MAX)
		return (qu_real) qu_int_newl ((long) self->timeout);
	return (qu_real) qu_float_new (self->timeout);
}



/*) sub transfer (dst: Stream, len: Integer = -1) -> Null
	Transfer all data from self to v{dst}.

	Optionally v{len} bytes. If v{len} is negative then it means until EOF.

	@exc
	.EArgValue
	.ESystem
*/
qu_null qu__stream_transfer (qu_stream self, int argc, qu_any *argv)
{
	qu_stream to;
	long      n = -1;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, to, qu_stream);
	if (argc > 1)
		QU_ARG_vlong (1, n);
	QU_ENSURE (qu_stream_transfer (self, to, n));
	return QuNil;
}



/*) sub truncate (len = -1) -> self
	Truncate the Stream to a specified size.

	The Stream must support #{seek}.

	If v{len} is not supplied or is a negative value then it means to
	use the current file position.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_truncate (qu_stream self, int argc, qu_any *argv)
{
	double     to = self->timeout;
	qu_integer len;
#if SIZEOF_OFF_T > SIZEOF_LONG
	long long size;
#else
	long size;
#endif

	if (argc > 0)
	{
		len = argv[0];
	}
	else
	{
		len = qu__stream_tell (self);
		QU_ENSURE (len);
	}

#if SIZEOF_OFF_T > SIZEOF_LONG
	QU_ARG_longlong (len, size);
#else
	QU_ARG_long (len, size);
#endif

	__qu_blocking_begin;
	while (ftruncate (self->fd, (off_t) size))
	{
		if (errno != EINTR)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		__qu_blocking_check_timeout (to);
		__qu_blocking_check_signal;
		qu_thread_yield ();
	}
	__qu_blocking_end;
	return self;
}



/*) sub tty (klass = Boolean) -> Any
	Check whether this Stream is connected to a I{tty} device (terminal).

	This method simply calls #{ftty} providing it with this Stream's descriptor.

	@noexc

	@xmp
		println (Stream.stdout.tty (String))

	@out
		/dev/pts/4
*/
static qu_any qu__stream_tty (qu_stream self, int argc, qu_any *argv)
{
	qu_any arg[2] = {qu_int_new (self->fd), QuBooleanClass};

	QU_ARG_VMAX (1);
	if (argc)
		arg[1] = argv[0];
	return qu__stream_ftty (QuStreamClass, 2, arg);
}



/*) sub unlock -> Null
	Unlock the Stream.

	The Stream must support locking and previously locked otherwise an ESystem
	exception is thrown.

	Note that ETimeout is possible, meaning that unlocking could not be performed
	until #{timeout}.

	Note that you can not unlock the k{null} Stream.

	@exc
	.ESystem
*/
static qu_null qu__stream_unlock (qu_stream self)
{
	if (self != QuStreamNull)
	{
		double to = self->timeout;
		struct flock k;

		k.l_type   = F_UNLCK;
		k.l_whence = SEEK_SET;
		k.l_start  = 0;
		k.l_len    = 0;

		__qu_blocking_begin;
		while (fcntl (self->fd, F_SETLK, &k) != 0)
		{
			if (errno != EINTR)
			{
				FIXEAGAIN;
				if (errno != EAGAIN)
				{
					__qu_blocking_end;
					QU_THROWERRNO;
				}
			}
			__qu_blocking_check_timeout (to);
			__qu_blocking_check_signal;
			qu_thread_yield ();
		}
		__qu_blocking_end;
	}
	return QuNil;
}



/*) sub write (s: String, len = s.__len, ofs = 0, all = true) -> Integer
	Write v{s} to Stream.

	v{len} is the number of bytes to write.

	v{ofs} is the offset in the string to start. If negative then it means to
	count from the end of the string.

	If v{all} is true then all bytes are written, otherwise at most
	v{len} bytes are written.

	On success, the actual number of bytes written is returned.

	If an error occured in the middle of the writing and something has been
	written, then the error is ignored and the actual number of bytes written
	is returned. If, otherwise, nothing has been written, then an exception
	is thrown.

	If, however, #{timeout} is exceeded then an ETimeout exception is thrown
	no matter how many bytes have been written. If this happens then you should
	#{close} the Stream because it becomes invalid.

	@exc
	.EArgValue
	.EMemory
	.ESystem
*/
static qu_integer qu__stream_write (qu_stream self, int argc, qu_any *argv)
{
	long      ofs = 0;
	int       all = 1;
	long      len = -1;
	qu_string s;

	QU_ARG_VCOUNT (1, 4);
	QU_ARG_VTYPE (0, s, qu_string);
	if (argc > 1)
	{
		QU_ARG_vlong (1, len);
		QU_ARG_CHECK (len >= 0);
		if (argc > 2)
		{
			QU_ARG_vlong (2, ofs);
			if (argc > 3)
				QU_ARG_vbool (3, all);
		}
	}
	if (s->len == 0)
		return (qu_integer) QuInt0;

	len = qu_stream_writes (self, s, ofs, len, all);
	QU_ENSURE (len >= 0);
	return qu_integer_newl (len);
}



/*) sub writeln (s: String, len = s.__len, ofs = 0, all = true) -> Integer
	This method simply calls #{write} to write the string v{s} then call it
	again to write a #{linebreak}.

	On success, the returned value is the actual number of bytes written including
	the linebreak.

	See #{write} for the arguments.

	@warn
	You should generally leave v{all} to true otherwise f{writeln} would not
	make much sense.

	@exc
	.EArgValue
	.ESystem
*/
static qu_integer qu__stream_writeln (qu_stream self, int argc, qu_any *argv)
{
	qu_integer l = qu__stream_write (self, argc, argv);
	qu_string  s = self->br;

	QU_ENSURE (l);
	if (s->len)
	{
		QU_ENSURE (qu__stream_write (self, 1, (qu_any *) &s));
		return qu_sys_plus (l, qu_int_new (s->len));
	}
	return l;
}



/*) sub writev (spec, *arg) -> self
	Write data.

	v{spec} is the specification of the arguments v{arg}. It may be a String
	or a Struct. See -{Struct.spec} for details.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_writev (qu_stream self, int argc, qu_any *argv)
{
	/* TODO - qu_struct.c needs more functions */

	qu_any spec;
	char   buf [128];
	long   n;

	QU_ARG_VMIN (2);
	spec = *argv++;
	argc--;

	if (qu_string_is (spec))
	{
		char *p = ((qu_string) spec)->str;
		long  l = ((qu_string) spec)->len;

		QU_REQUIRE (QuEArgCount, argc == l);
		while (l--)
		{
			n = qu_pointer_set (buf, *p++, *argv++);
			QU_ENSURE (n);
			assert (n <= sizeof (buf));
			argc--;
			QU_ENSURE (qu_stream_write (self, buf, n, n) == n);
		}
	}
	else
	{
		QU_ARG_TYPE (spec, qu_struct);
		{
			qu_struct         u = spec;
			long              l = qu_struct_len (u);
			long              i;
			qu_structentry_s *t = u->items;

			for (i = 0; i < l; ++i)
			{
				qu_structentry_s *e = &t[i];
				char              c = qu_struct_type_to_char (e->type);

				QU_ARG_CHECK (c);
				n = qu_pointer_set (buf, c, *argv++);
				assert (n <= sizeof (buf));
				argc--;
				QU_ENSURE (qu_stream_write (self, buf, n, n) == n);
			}
		}
	}
	return self;
}



/*) sub write8 (x: Integer) -> self
	Write a 8-bits signed integer.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_write8 (qu_stream self, qu_integer x)
{
	char v;
	QU_ARG_byte (x, v);
	QU_ENSURE (qu_stream_write (self, &v, 1, 1) == 1);
	return self;
}



/*) sub write16 (x: Integer) -> self
	Write a 16-bits signed integer.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_write16 (qu_stream self, qu_integer x)
{
#if SIZEOF_SHORT == 2
	short v;
	QU_ARG_short (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 2, 1) == 2);
	return self;
#else
	QU_THROWZ (QuEImplement, "native-16 bits integer not available");
#endif
}



/*) sub write32 (x: Integer) -> self
	Write a 32-bits signed integer.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_write32 (qu_stream self, qu_integer x)
{
#if SIZEOF_INT == 4
	int v;
	QU_ARG_int (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 4, 1) == 4);
	return self;
#elif SIZEOF_LONG == 4
	long v;
	QU_ARG_long (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 4, 1) == 4);
	return self;
#else
	QU_THROWZ (QuEImplement, "native-32 bits integer not available");
#endif
}



/*) sub writeu8 (x: Integer) -> self
	Write a 8-bits unsigned integer.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_writeu8 (qu_stream self, qu_integer x)
{
	unsigned char v;
	QU_ARG_ubyte (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 1, 1) == 1);
	return self;
}



/*) sub writeu16 (x: Integer) -> self
	Write a 16-bits unsigned integer.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_writeu16 (qu_stream self, qu_integer x)
{
#if SIZEOF_SHORT == 2
	unsigned short v;
	QU_ARG_ushort (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 2, 1) == 2);
	return self;
#else
	QU_THROWZ (QuEImplement, "native-16 bits integer not available");
#endif
}



/*) sub writeu32 (x: Integer) -> self
	Write a 32-bits unsigned integer.

	@exc
	.EArgValue
	.ESystem
*/
static qu_stream qu__stream_writeu32 (qu_stream self, qu_integer x)
{
#if SIZEOF_INT == 4
	unsigned int v;
	QU_ARG_uint (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 4, 1) == 4);
	return self;
#elif SIZEOF_LONG == 4
	unsigned long v;
	QU_ARG_ulong (x, v);
	QU_ENSURE (qu_stream_write (self, (char *) &v, 4, 1) == 4);
	return self;
#else
	QU_THROWZ (QuEImplement, "native-32 bits integer not available");
#endif
}



/*-----------------------------------------------------------------------------
	QuStreamDefaultFuncs
-----------------------------------------------------------------------------*/

static ssize_t _stream_ordr (qu_any obj, char *buf, size_t len)
{
	errno = EBADF;
	return -1;
}

static ssize_t _stream_owrt (qu_any obj, const char *buf, size_t len)
{
	errno = EBADF;
	return -1;
}

static qu_stream qu__stream_accept (qu_stream self)
{
	QU_REQUIRE (QuEImplement, self->funcs->acc);
	return self->funcs->acc (self->obj ? self->obj : (qu_any) self);
}

static qu_null qu__stream_connect (qu_stream self, qu_any addr)
{
	QU_REQUIRE (QuEImplement, self->funcs->con);
	return self->funcs->con (self->obj ? self->obj : (qu_any) self, addr);
}


static ssize_t _stream_seek (int fd, ssize_t len, int whence)
{
	return lseek (fd, len, whence);
}

static ssize_t _stream_oseek (qu_any obj, ssize_t len, int whence)
{
	errno = EBADF;
	return -1;
}

qu_fstream_s QuStreamDefaultFuncs =
{
	(qu_stream_rdr_f)     read,
	(qu_stream_wrt_f)     write,
	(qu_stream_ordr_f)   _stream_ordr,
	(qu_stream_owrt_f)   _stream_owrt,
	(qu_stream_acc_f)    qu__stream_accept,
	(qu_stream_con_f)    qu__stream_connect,
	(qu_stream_seek_f)   _stream_seek,
	(qu_stream_oseek_f)  _stream_oseek
};



/*-----------------------------------------------------------------------------
	interface
-----------------------------------------------------------------------------*/

/*
	Create a new qu_stream --- this is the only allocation routine.
*/
static long __QuStreamLastId = 0;
qu_stream qu_stream_new (qu_class self)
{
	register qu_stream s;

	assert (self->size >= sizeof (qu_stream_s));

	s          = qu_gc__new (self, self->size);
	s->sid     = __QuStreamLastId++;
	s->fd      = -1;
	s->br      = qu_string_byte ('\n');
	s->funcs   = &QuStreamDefaultFuncs;
	s->obj     = 0;
	s->timeout = DBL_MAX;
	return s;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	select
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
	Check whether the qu_stream is ready for I/O.
*/
int qu_stream_select (qu_stream self, char how, long timeout)
{
	int    r, fd = self->fd;
	struct timeval v;
	fd_set t;

	if (!qu_stream_can_select (self))
		return 1;

	/* FD_SET crashes without this, unbelievable eh? */
	if (fd < 0)
	{
		errno = EBADF;
		return -1;
	}

	if (timeout < 0)
	{
		v.tv_sec  = 0;
		v.tv_usec = 1000;
	}
	else
	{
		v.tv_sec  = timeout / 1000000;
		v.tv_usec = timeout % 1000000;
	}

	FD_ZERO (&t);
	FD_SET (fd, &t);

	while ((r = select (fd + 1,
				how == 'r' ? &t : 0,
				how == 'w' ? &t : 0,
				how == 'e' ? &t : 0, &v)) < 0)
	{
		if (errno != EINTR)
			return -1;
	}
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	read
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
	Read - primitive
	If all, then read n bytes or time-out.
	Otherwise read at least 1 byte or time-out.

	!!! Since we throw on error, and we write to stderr when we throw
	!!! (see qu_thread_throw) then there is a chance that we are called in
	!!! an infinite loop. Here we don't care and leave the tracking to
	!!! qu_thread.c.
*/
long qu_stream_read (qu_stream self, char *buf, long n, int all)
{
	qu_fstream_s *funcs = self->funcs;
	int           fd    = self->fd;
	double        to    = self->timeout;
	qu_any        obj   = self->obj;
	long          res   = 0;
	long          r;
	int           test;

	if (self == QuStreamNull)
		QU_THROWV (QuEBADF, -1);

	assert (n >= 0);
	assert (n <= SSIZE_MAX);

	__qu_blocking_begin;
	while (n > 0)
	{
		if (!all && res)
			break;

		__qu_blocking_check_signalv (-1);
		__qu_blocking_check_timeoutv (to, -1);

		test = qu_stream_select (self, 'r', SELECT_READY);

		if (test < 0)
		{
			__qu_blocking_end;
			QU_THROWERRNOV (-1);
		}

		if (test == 0)
		{
			qu_thread_yield ();
			continue;
		}

		do
		{
			if (obj)
				r = (long) funcs->ordr (obj, buf + res, (size_t) n);
			else
				r = (long) funcs->rdr (fd, buf + res, (size_t) n);

		} while (r < 0 && errno == EINTR);

		if (r < 0)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNOV (-1);
			}
		}
		else if (r == 0)
		{
			if (qu_stream_can_eof (self))
			{
				if (all)
				{
					__qu_blocking_end;
					QU_THROWV (QuENODATA, -1);
				}
				break;
			}

			__qu_blocking_end;
			QU_THROWV (QuEPIPE, -1);
		}
		else
		{
			n -= r;
			res += r;
		}
	}
	__qu_blocking_end;
	return res;
}



/*
	Read - return a new qu_string
*/
qu_string qu_stream_reads (qu_stream self, long n, int all)
{
	assert (n >= 0);
	assert (n <= SSIZE_MAX);

	if (all || n > BUFSIZ)
	{
		qu_string s = qu_string_buffer_new (n);
		long      r = qu_stream_read (self, s->str, n, all);

		QU_ENSURE (r >= 0);
		s->len = r;
		return qu_string_buffer_return (s);
	}
	else
	{
		char buf [BUFSIZ];
		long r = qu_stream_read (self, buf, n, all);

		QU_ENSURE (r >= 0);
		return qu_string_new (buf, r);
	}
}



/*
	Read a line (until self->br or EOF or max)
*/
long qu_stream_readln (qu_stream self, char *buf, long max)
{
	qu_fstream_s *funcs = self->funcs;
	int           fd    = self->fd;
	double        to    = self->timeout;
	qu_any        obj   = self->obj;
	long          l     = 0;
	qu_string     br    = self->br;
	char         *z     = br->str;
	char          m     = br->len;
	char         *p     = z;
	long          n     = 0;
	long          i     = 0;
	char         *b, c;
	long          j, r;
	int           test;

	if (self == QuStreamNull)
		QU_THROWV (QuEBADF, -1);

	assert (max >= 0);
	assert (max <= SSIZE_MAX);

	__qu_blocking_begin;
	while (l < max)
	{
		__qu_blocking_check_signalv (-1);
		__qu_blocking_check_timeoutv (to, -1);

		test = qu_stream_select (self, 'r', SELECT_READY);

		if (test < 0)
		{
			__qu_blocking_end;
			QU_THROWERRNOV (-1);
		}

		if (test == 0)
		{
			qu_thread_yield ();
			continue;
		}

		do
		{
			if (obj)
				r = (long) funcs->ordr (obj, &c, 1);
			else
				r = (long) funcs->rdr (fd, &c, 1);

		} while (r < 0 && errno == EINTR);

		if (r < 0)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNOV (-1);
			}
		}
		else if (r == 0)
		{
			if (qu_stream_can_eof (self))
				break;

			__qu_blocking_end;
			QU_THROWV (QuEPIPE, -1);
		}
		else
		{
			assert (r == 1);

			/* this is more efficient than calling qu_string_buffer_putc */
			buf[l++] = c;

			if (c == *p)
			{
				if (++n == m)
					break;
				p++;
			}
			else if (n)
			{
				p = z;
				n = 0;
				b = buf;
				j = ++i;
				while (j < l)
				{
					if (b[j] == *p)
					{
						++n;
						++p;
					}
					else if (n)
					{
						p = z;
						n = 0;
					}
					else
					{
						++i;
					}
					++j;
				}
			}
			else
			{
				++i;
			}
		}
	}
	__qu_blocking_end;
	return l;
}



/*
	Read a line (until self->br or EOF or max) - return a new qu_string.
*/
qu_string qu_stream_readsln (qu_stream self, long max)
{
	qu_fstream_s *funcs = self->funcs;
	int           fd    = self->fd;
	double        to    = self->timeout;
	qu_any        obj   = self->obj;
	long          l     = 0;
	qu_string     br    = self->br;
	char         *z     = br->str;
	char          m     = br->len;
	char         *p     = z;
	long          n     = 0;
	long          i     = 0;
	char         *b, c;
	long          j, r;
	int           test;
	volatile qu_string buf;

	if (self == QuStreamNull)
		QU_THROW (QuEBADF);

	assert (max >= 0);
	assert (max <= SSIZE_MAX);

	/*
		always start small - is this a good value?
		8 -> 32 -> 128 -> 512 -> 2048 -> 8192 ...
	*/
	buf = qu_string_buffer_new (8);

	__qu_blocking_begin;
	while (l < max)
	{
		__qu_blocking_check_signal;
		__qu_blocking_check_timeout (to);

		test = qu_stream_select (self, 'r', SELECT_READY);
		if (test < 0)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}

		if (test == 0)
		{
			qu_thread_yield ();
			continue;
		}

		do
		{
			if (obj)
				r = (long) funcs->ordr (obj, &c, 1);
			else
				r = (long) funcs->rdr (fd, &c, 1);

		} while (r < 0 && errno == EINTR);

		if (r < 0)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
		}
		else if (r == 0)
		{
			if (qu_stream_can_eof (self))
				break;

			__qu_blocking_end;
			QU_THROW (QuEPIPE);
		}
		else
		{
			assert (r == 1);

			/* don't escape blocking mode unless we really have to */
			if (++l >= qu_string_buffer_avail (buf))
			{
				long more;

				__qu_blocking_end;
				if (l < QU_MEM_MAX / 4)
					more = l * 3;
				else if (l < QU_MEM_MAX / 2)
					more = BUFSIZ;
				else
					more = 1;
				QU_ENSURE (qu_string_buffer_make_space (buf, more));
				__qu_blocking_begin;
			}

			/* this is more efficient than calling qu_string_buffer_putc */
			buf->str[l - 1] = c;
			buf->len = l;

			if (c == *p)
			{
				if (++n == m)
					break;
				p++;
			}
			else if (n)
			{
				p = z;
				n = 0;
				b = buf->str;
				j = ++i;
				while (j < l)
				{
					if (b[j] == *p)
					{
						++n;
						++p;
					}
					else if (n)
					{
						p = z;
						n = 0;
					}
					else
					{
						++i;
					}
					++j;
				}
			}
			else
			{
				++i;
			}
		}
	}
	__qu_blocking_end;

	/* don't forget to put the nil terminator */
	buf->str[buf->len] = '\0';
	return buf;
}



/*
	Read into a qu_buffer.
*/
long qu_stream_recv (qu_stream self, qu_buffer b, long n, int all)
{
	qu_fstream_s *funcs = self->funcs;
	int           fd    = self->fd;
	double        to    = self->timeout;
	qu_any        obj   = self->obj;
	long          res   = 0;
	long          r;
	char         *buf;
	int           test;
	volatile qu_any scope;

	assert (n >= 0);
	assert (n <= SSIZE_MAX);

	if (self == QuStreamNull)
		QU_THROWV (QuEBADF, -1);

	if (qu_buffer_frozen (b))
		QU_THROWV (QuEModify, -1);

	if (!qu_buffer_make_space (b, n))
		QU_THROWV (QuEOverflow, -1);

	QU_BUFFER_PROTECT_BEGIN (b, scope);
	buf = b->str + b->len;

	__qu_blocking_begin;
	while (n > 0)
	{
		if (!all && res)
			break;

		if (__QuSigTERM || __QuSigINT)
		{
			__qu_blocking_end;
			QU_BUFFER_PROTECT_END (b, scope);
			if (__QuSigTERM)
				QU_THROWV (QuEQuit, -1);
			QU_THROWV (QuEInterrupt, -1);
		}

		if (__qu_now__ >= to)
		{
			__qu_blocking_end;
			QU_BUFFER_PROTECT_END (b, scope);
			QU_THROWV (QuETimeout, -1);
		}

		test = qu_stream_select (self, 'r', SELECT_READY);

		if (test < 0)
		{
			__qu_blocking_end;
			QU_BUFFER_PROTECT_END (b, scope);
			QU_THROWERRNOV (-1);
		}

		if (test == 0)
		{
			qu_thread_yield ();
			continue;
		}

		do
		{
			if (obj)
				r = (long) funcs->ordr (obj, buf, (size_t) n);
			else
				r = (long) funcs->rdr (fd, buf, (size_t) n);

		} while (r < 0 && errno == EINTR);

		if (r < 0)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_BUFFER_PROTECT_END (b, scope);
				QU_THROWERRNOV (-1);
			}
		}
		else if (r == 0)
		{
			if (qu_stream_can_eof (self))
			{
				if (all)
				{
					__qu_blocking_end;
					QU_BUFFER_PROTECT_END (b, scope);
					QU_THROWV (QuENODATA, -1);
				}
				break;
			}

			__qu_blocking_end;
			QU_BUFFER_PROTECT_END (b, scope);
			QU_THROWV (QuEPIPE, -1);
		}
		else
		{
			n -= r;
			res += r;
			buf += res;
			b->len += res;
			b->str[b->len] = '\0';
			if (n <= 0)
				break;
		}
	}
	__qu_blocking_end;
	QU_BUFFER_PROTECT_END (b, scope);
	return res;
}



long qu_stream_recvln (qu_stream self, qu_buffer b, long max)
{
	long r;

	if (!qu_buffer_make_space (b, max))
		QU_THROWV (QuEOverflow, -1);

  __qu_buffer_protect_begin (b);
	r = qu_stream_readln (self, b->str + b->len, max);
  __qu_buffer_protect_end;
	if (r > 0)
	{
		b->len += r;
		b->str[b->len] = '\0';
	}
	return r;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	write
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
	Write primitive.
*/
long qu_stream_write (qu_stream self, const char *b, long n, int all)
{
	qu_fstream_s *funcs = self->funcs;
	int           fd    = self->fd;
	double        to    = self->timeout;
	qu_any        obj   = self->obj;
	long          res   = 0;
	long          r;
	int           test;
	double        last  = qu_stream_now ();

	if (self == QuStreamNull)
		return n;

	assert (n >= 0);
	assert (n <= SSIZE_MAX);

	__qu_blocking_begin;
	while (n > 0)
	{
		if (!all && res)
			break;

		__qu_blocking_check_signalv (-1);
		__qu_blocking_check_timeoutv (to, -1);

		test = qu_stream_select (self, 'w', SELECT_READY);

		if (test < 0)
		{
			__qu_blocking_end;
			QU_THROWERRNOV (-1);
		}

		if (test == 0)
		{
			/*
				This is the only way I know how to force an EPIPE
				when the other side of a socket has been closed.

				The problem here is that 'select' does not give any indication
				of a broken pipe. The solution is to force a write. If we write
				0 bytes then we assume a broken pipe.

				Thus for each number of seconds, we insist on writing anyway,
				even when select says the socket is not ready.
			*/
#define CHECKWRITEPERIOD 8.0
			double t = qu_stream_now ();
			if (t > last + CHECKWRITEPERIOD)
				last = t;
			else
			{
				qu_thread_yield ();
				continue;
			}
		}

		do
		{
			if (obj)
				r = (long) funcs->owrt (obj, b + res, (size_t) n);
			else
				r = (long) funcs->wrt (fd, b + res, (size_t) n);

		} while (r < 0 && errno == EINTR);

		if (r < 0)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNOV (-1);
			}
		}
		else if (r == 0)
		{
			__qu_blocking_end;
			QU_THROWV (QuEPIPE, -1);
		}
		else
		{
			res += r;
			n -= r;
		}
	}
	__qu_blocking_end;
	return res;
}



/*
	This just calls qu_stream_write twice.
	Once to write the buffer, then self->br.
*/
long qu_stream_writeln (qu_stream self, const char *b, long n)
{
	n = qu_stream_write (self, b, n, 1);
	if (n < 0)
		return -1;

	if (self->br->len)
	{
		if (qu_stream_write (self, self->br->str, self->br->len, 1) < 0)
			return -1;

		assert (n <= QU_INT_MAX - self->br->len);
		n += self->br->len;
	}
	return n;
}



long qu_stream_writes (qu_stream self, qu_string x, long ofs, long len, int all)
{
	long  l = x->len;
	long  res = 0;
	long  max, r;
	char *p;

	if (ofs < 0)
		ofs += l;

	if (ofs < 0 || ofs >= l)
		QU_THROWV (QuEIndex, -1);

	max = l - ofs;

	if (len < 0)
		len = max;
	else if (len > max)
		len = max;

	if (len == 0)
		return 0;

	p = x->str + ofs;
	while (len)
	{
		r = qu_stream_write (self, p, len, all);
		if (r < 0)
			return -1;
		res += r;
		if (!all && res)
			break;
		p += r;
		len -= r;
	}
	return res;
}



long qu_stream_writesln (qu_stream self, qu_string x, long ofs, long len)
{
	long r = qu_stream_writes (self, x, ofs, len, 1);

	if (r < 0)
		return -1;
	if (qu_stream_writes (self, self->br, 0, self->br->len, 1) < 0)
		return -1;
	return r + self->br->len;
}



long qu_stream_send (qu_stream self, qu_buffer b, long len, int all)
{
	long res;

	if (qu_buffer_frozen (b))
		QU_THROWV (QuEModify, -1);

	if (len > b->len)
		len = b->len;

	if (len == 0)
		return 0;

  __qu_buffer_protect_begin (b);
	res = qu_stream_write (self, b->str, len, all);
  __qu_buffer_protect_end;
	if (res > 0)
		qu_buffer_del (b, 0, res);
	return res;
}



qu_null qu_stream_print (qu_stream self, qu_any x)
{
	return qu__stream_print (self, 1, &x);
}



qu_null qu_stream_println (qu_stream self, qu_any x)
{
	return qu__stream_println (self, 1, &x);
}



qu_null qu_stream_printz (qu_stream self, const char *s, long len)
{
	QU_ENSURE (qu_stream_write (self, s, len, 1) >= 0);
	return QuNil;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	transfer
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int qu_stream_transfer (qu_stream self, qu_stream to, long len)
{
	qu_fstream_s *sfuncs = self->funcs;
	int           sfd    = self->fd;
	double        sto    = self->timeout;
	qu_any        sobj   = self->obj;

	qu_fstream_s *tfuncs = to->funcs;
	int           tfd    = to->fd;
	double        tto    = to->timeout;
	qu_any        tobj   = to->obj;

	int   all = len < 0;
	char  buf [BUFSIZ];
	long  r, w;
	int   test;

	if (self == QuStreamNull)
		QU_THROW (QuEBADF);

	__qu_blocking_begin;
	for (;;)
	{
		/*
			read
		*/

		__qu_blocking_check_signal;
		__qu_blocking_check_timeout (sto);
		test = qu_stream_select (self, 'w', SELECT_READY);

		if (test < 0)
		{
			__qu_blocking_end;
			QU_THROWERRNO;
		}

		if (test == 0)
		{
			qu_thread_yield ();
			continue;
		}

		if (all)
			w = sizeof (buf);
		else
		{
			w = len;
			if (w > sizeof (buf))
				w = sizeof (buf);
		}

		do
		{
			if (sobj)
				r = (long) sfuncs->ordr (sobj, buf, w);
			else
				r = (long) sfuncs->rdr (sfd, buf, w);

		} while (r < 0 && errno == EINTR);

		if (r < 0)
		{
			FIXEAGAIN;
			if (errno != EAGAIN)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}
			continue;
		}

		if (r == 0)
		{
			if (all && qu_stream_can_eof (self))
				break;
			QU_THROW (QuENODATA);
		}

		if (!all)
			len -= r;

		if (to == QuStreamNull)
			continue;

		/*
			write
		*/

		while (r)
		{
			__qu_blocking_check_signal;
			__qu_blocking_check_timeout (tto);

			test = qu_stream_select (to, 'w', SELECT_READY);

			if (test < 0)
			{
				__qu_blocking_end;
				QU_THROWERRNO;
			}

			if (test == 0)
			{
				qu_thread_yield ();
				continue;
			}

			do
			{
				if (tobj)
					w = (long) tfuncs->owrt (tobj, buf, r);
				else
					w = (long) tfuncs->wrt (tfd, buf, r);

			} while (w < 0 && errno == EINTR);

			if (w < 0)
			{
				FIXEAGAIN;
				if (errno != EAGAIN)
				{
					__qu_blocking_end;
					QU_THROWERRNO;
				}
			}
			else if (w == 0)
			{
				__qu_blocking_end;
				QU_THROW (QuEIO);
			}
			else
			{
				r -= w;
			}
		}
	}
	__qu_blocking_end;
	return 1;
}


/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_stream__setup ()
{
	qu_class_setup (QuStreamClass, qu_stream_s, qu__stream__destroy, qu__stream__mark);
}



void qu_stream__init ()
{
	StatStruct = qu_struct_new_static (13,
		"st_dev"    , QuIntegerClass,
		"st_ino"    , QuIntegerClass,
		"st_mode"   , QuIntegerClass,
		"st_nlink"  , QuIntegerClass,
		"st_uid"    , QuIntegerClass,
		"st_gid"    , QuIntegerClass,
		"st_rdev"   , QuIntegerClass,
		"st_size"   , QuIntegerClass,
		"st_blksize", QuIntegerClass,
		"st_blocks" , QuIntegerClass,
		"st_atime"  , QuFloatClass,
		"st_mtime"  , QuFloatClass,
		"st_ctime"  , QuFloatClass);

	qu_var_enum ((qu_symbol) QuStreamClass, "ST_",
		"ST_DEV"    , 0,
		"ST_INO"    , 1,
		"ST_MODE"   , 2,
		"ST_NLINK"  , 3,
		"ST_UID"    , 4,
		"ST_GID"    , 5,
		"ST_RDEV"   , 6,
		"ST_SIZE"   , 7,
		"ST_BLKSIZE", 8,
		"ST_BLOCKS" , 9,
		"ST_ATIME"  , 10,
		"ST_MTIME"  , 11,
		"ST_CTIME"  , 12,
		NULL);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		QuStreamStderr QuStreamStdin QuStreamStdout
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	QuStreamStderr          = qu_stream_new (QuStreamClass);
	QuStreamStderr->flags  |= QU_STREAM_FLAG_CANEOF | QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANCLOSE;
	QuStreamStderr->fd      = STDERR_FILENO;

	QuStreamStdin           = qu_stream_new (QuStreamClass);
	QuStreamStdin->flags   |= QU_STREAM_FLAG_CANEOF | QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANCLOSE;
	QuStreamStdin->fd       = STDIN_FILENO;
	QuStreamStdin->timeout  = DBL_MAX;

	QuStreamStdout          = qu_stream_new (QuStreamClass);
	QuStreamStdout->flags  |= QU_STREAM_FLAG_CANEOF | QU_STREAM_FLAG_CANBLOCK | QU_STREAM_FLAG_CANCLOSE;
	QuStreamStdout->fd      = STDOUT_FILENO;
	QuStreamStdout->timeout = DBL_MAX;

	QuStreamNull           = qu_stream_new (QuStreamClass);
	QuStreamNull->timeout  = DBL_MAX;

	/* specific class properties */
	qu_class_constz  (QuStreamClass, "Stat"        , StatStruct);
	qu_class_constz  (QuStreamClass, "__stderr__"  , QuStreamStderr);
	qu_class_constz  (QuStreamClass, "__stdin__"   , QuStreamStdin);
	qu_class_constz  (QuStreamClass, "__stdout__"  , QuStreamStdout);
	qu_class_constz  (QuStreamClass, "__null__"    , QuStreamNull);

	/* standard class methods */
	qu_class_method_class  (QuStreamClass, QU_NAME_new       , qu__stream_new       , -1);
	qu_class_method_class  (QuStreamClass, QU_NAME__inst     , qu__stream__inst     , -1);

	/* specific class methods */
	qu_class_method_classz (QuStreamClass, "fmode"     , qu__stream_fmode     , -1);
	qu_class_method_classz (QuStreamClass, "fmtime"    , qu__stream_fmtime    , -1);
	qu_class_method_classz (QuStreamClass, "fselect"   , qu__stream_fselect   , -1);
	qu_class_method_classz (QuStreamClass, "fsize"     , qu__stream_fsize     , -1);
	qu_class_method_classz (QuStreamClass, "fstat"     , qu__stream_fstat     , -1);
	qu_class_method_classz (QuStreamClass, "fsync"     , qu__stream_fsync     ,  0);
	qu_class_method_classz (QuStreamClass, "ismode"    , qu__stream_ismode    ,  2);
	qu_class_method_classz (QuStreamClass, "stderr"    , qu__stream_stderr    , -1);
	qu_class_method_classz (QuStreamClass, "stdin"     , qu__stream_stdin     , -1);
	qu_class_method_classz (QuStreamClass, "stdout"    , qu__stream_stdout    , -1);
	qu_class_method_classz (QuStreamClass, "ftty"      , qu__stream_ftty      , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuStreamClass, QU_NAME__each     , qu__stream__each     ,  0);
	qu_class_method_inst  (QuStreamClass, QU_NAME__eq       , qu__stream__eq       ,  1);
	qu_class_method_inst  (QuStreamClass, QU_NAME__getitem  , qu__stream__getitem  ,  1);
	qu_class_method_inst  (QuStreamClass, QU_NAME__getslice , qu__stream__getslice ,  2);
	qu_class_method_inst  (QuStreamClass, QU_NAME__hash     , qu__stream__hash     ,  0);
	qu_class_method_inst  (QuStreamClass, QU_NAME__in       , qu__stream__in       ,  1);
	qu_class_method_inst  (QuStreamClass, QU_NAME__len      , qu__stream__len      ,  0);
	qu_class_alias        (QuStreamClass, QU_NAME__len      , QU_NAME_len);
	qu_class_method_inst  (QuStreamClass, QU_NAME__like     , qu__stream__like     ,  1);
	qu_class_method_inst  (QuStreamClass, QU_NAME__pop      , qu__stream__pop      ,  0);
	qu_class_method_inst  (QuStreamClass, QU_NAME__print    , qu__stream__print    ,  0);
	qu_class_method_inst  (QuStreamClass, QU_NAME__push     , qu__stream__push     ,  1);
	qu_class_method_inst  (QuStreamClass, QU_NAME__setitem  , qu__stream__setitem  ,  2);
	qu_class_method_inst  (QuStreamClass, QU_NAME__setslice , qu__stream__setslice ,  3);
	qu_class_method_inst  (QuStreamClass, QU_NAME__string   , qu__stream__string   ,  0);
	qu_class_method_inst  (QuStreamClass, QU_NAME__true     , qu__stream__true     ,  0);

	/* specific instance methods */
	qu_class_method_instz (QuStreamClass, "blocking"  , qu__stream_blocking  , -1);
	qu_class_method_instz (QuStreamClass, "can"       , qu__stream_can       ,  1);
	qu_class_method_instz (QuStreamClass, "close"     , qu__stream_close     ,  0);
	qu_class_method_instz (QuStreamClass, "eachline"  , qu__stream_eachline  ,  0);
	qu_class_method_instz (QuStreamClass, "eof"       , qu__stream_eof       ,  0);
	qu_class_method_instz (QuStreamClass, "fcntl"     , qu__stream_fcntl     , -1);
	qu_class_method_instz (QuStreamClass, "fd"        , qu__stream_fd        ,  0);
	qu_class_method_instz (QuStreamClass, "flush"     , qu__stream_flush     ,  0);
	qu_class_method_instz (QuStreamClass, "get"       , qu__stream_get       ,  1);
	qu_class_method_instz (QuStreamClass, "ioctl"     , qu__stream_ioctl     , -1);
	qu_class_method_instz (QuStreamClass, "linebreak" , qu__stream_linebreak , -1);
	qu_class_method_instz (QuStreamClass, "lock"      , qu__stream_lock      , -1);
	qu_class_method_instz (QuStreamClass, "locked"    , qu__stream_locked    ,  0);
	qu_class_method_instz (QuStreamClass, "mode"      , qu__stream_mode      ,  0);
	qu_class_method_instz (QuStreamClass, "mtime"     , qu__stream_mtime     ,  0);
	qu_class_method_instz (QuStreamClass, "obj"       , qu__stream_obj       ,  0);
	qu_class_method_instz (QuStreamClass, "opened"    , qu__stream_opened    ,  0);
	qu_class_method_instz (QuStreamClass, "print"     , qu__stream_print     , -1);
	qu_class_method_instz (QuStreamClass, "println"   , qu__stream_println   , -1);
	qu_class_method_instz (QuStreamClass, "put"       , qu__stream_put       ,  1);
	qu_class_method_instz (QuStreamClass, "read"      , qu__stream_read      , -1);
	qu_class_method_instz (QuStreamClass, "readall"   , qu__stream_readall   ,  1);
	qu_class_method_instz (QuStreamClass, "readln"    , qu__stream_readln    , -1);
	qu_class_method_instz (QuStreamClass, "readv"     , qu__stream_readv     ,  1);
	qu_class_method_instz (QuStreamClass, "read8"     , qu__stream_read8     ,  0);
	qu_class_method_instz (QuStreamClass, "read16"    , qu__stream_read16    ,  0);
	qu_class_method_instz (QuStreamClass, "read32"    , qu__stream_read32    ,  0);
	qu_class_method_instz (QuStreamClass, "readu8"    , qu__stream_readu8    ,  0);
	qu_class_method_instz (QuStreamClass, "readu16"   , qu__stream_readu16   ,  0);
	qu_class_method_instz (QuStreamClass, "readu32"   , qu__stream_readu32   ,  0);
	qu_class_method_instz (QuStreamClass, "recv"      , qu__stream_recv      , -1);
	qu_class_method_instz (QuStreamClass, "recvln"    , qu__stream_recvln    ,  2);
	qu_class_method_instz (QuStreamClass, "rewind"    , qu__stream_rewind    ,  0);
	qu_class_method_instz (QuStreamClass, "search"    , qu__stream_search    , -1);
	qu_class_method_instz (QuStreamClass, "seek"      , qu__stream_seek      , -1);
	qu_class_method_instz (QuStreamClass, "select"    , qu__stream_select    , -1);
	qu_class_method_instz (QuStreamClass, "send"      , qu__stream_send      , -1);
	qu_class_method_instz (QuStreamClass, "stat"      , qu__stream_stat      , -1);
	qu_class_method_instz (QuStreamClass, "sync"      , qu__stream_sync      ,  0);
	qu_class_method_instz (QuStreamClass, "tell"      , qu__stream_tell      ,  0);
	qu_class_method_instz (QuStreamClass, "timeout"   , qu__stream_timeout   , -1);
	qu_class_method_instz (QuStreamClass, "truncate"  , qu__stream_truncate  , -1);
	qu_class_method_instz (QuStreamClass, "transfer"  , qu__stream_transfer  , -1);
	qu_class_method_instz (QuStreamClass, "tty"       , qu__stream_tty       , -1);
	qu_class_method_instz (QuStreamClass, "unlock"    , qu__stream_unlock    ,  0);
	qu_class_method_instz (QuStreamClass, "write"     , qu__stream_write     , -1);
	qu_class_method_instz (QuStreamClass, "writeln"   , qu__stream_writeln   , -1);
	qu_class_method_instz (QuStreamClass, "writev"    , qu__stream_writev    , -1);
	qu_class_method_instz (QuStreamClass, "write8"    , qu__stream_write8    ,  1);
	qu_class_method_instz (QuStreamClass, "write16"   , qu__stream_write16   ,  1);
	qu_class_method_instz (QuStreamClass, "write32"   , qu__stream_write32   ,  1);
	qu_class_method_instz (QuStreamClass, "writeu8"   , qu__stream_writeu8   ,  1);
	qu_class_method_instz (QuStreamClass, "writeu16"  , qu__stream_writeu16  ,  1);
	qu_class_method_instz (QuStreamClass, "writeu32"  , qu__stream_writeu32  ,  1);

	qu_class_method_instz (QuStreamClass, "accept"    , qu__stream_accept    ,  0);
	qu_class_method_instz (QuStreamClass, "connect"   , qu__stream_connect   ,  1);

	/* finalize declaration */
	qu_class_enddecl (QuStreamClass);
}
