/* Struct.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*
		FIXME - problem when the field is a validator
		such as enum, Array, Range, etc.
*/


/*) class Struct __final
	A Struct defines field names and the associated types of which the
	field values must conform to.

	You create instances from a class, but you create -{Record} objects
	from a Struct.

	You would normally create a Struct with the language syntax or the
	#{new} method.

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		joe = Person ('Joe', 98)

		Animal = Struct ('legs', Int, 'name', String)
		dog = Animal (4, 'Dog')

	@p
	Naturally, you can not modify a Struct object once created.

	The field names are frozen Strings. That is, frozen copies are created
	if necessary.

	You can access the fields of the Record by name or index, or even
	the dot operator.

	The k{is} operator can also be used to test for the Struct of a -{Record}.

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		joe = Person ('Joe', 98)

		println (joe.name ' = ' joe->name ' = ' joe[0] ' = ' joe['name'])
		println (joe.age  ' = ' joe->age  ' = ' joe[1] ' = ' joe['age'])

		try
			joe->age = 'boo'
		catch
			println (__err__)
		;;

		println (joe.__struct == Person)
		println (joe is Person)

	@out
		Joe = Joe = Joe = Joe
		98 = 98 = 98 = 98
		(EValue) invalid value
		true
		true

	@p
	Qu allows you to create a Struct which field is itself. The catch is that
	you have to use the k{struct} syntax, i.e. it is not possible to do this
	with the #{new} method because the Struct has not yet been created,
	of course.

	A -{Record} field value which type is Struct may be set to nil indicating
	that it is an empty Record.

	@xmp
		struct More
			email : String
			phone : String
		;;

		struct Person
			age    : Int
			name   : String
			father : Person
			info   : More
		;;

		println (Person)
		p = Person (98, 'Foo', nil, More ('foo@bar.baz', '1234567'))
		println (p)

	@out
		[age:Int, name:String, father:Struct, info:Struct]
		[98, 'Foo', nil, ['foo@bar.baz', '1234567']]

	@p
	The combination of Struct and -{Record} objects are slightly more efficient
	than -{Dict} objects. They also ensure that all field values of a Record
	are of the specified types.

	Note that the field names must be unique and the -{Null} class may not be
	used as a field type. If you need to specify a field that may contain any
	object then use the -{Object} class or k{Any}.

	@xmp
		struct Foo
			anything : Object
			num      : Int
		;;

		struct Foo
			anything : Any
			num      : Int
		;;

	@p
	You may use an k{enum} to define the type of the field.
	In this case, the interpreter will make sure that the value assigned to
	the field is one value in the enumeration.

	You can even use I{validators}. Whenever a field is going to be set then
	the Sub is called.

	@xmp
		sub _age_ (x)
			println ('validating age ...')
			require (x is Int and x in 0..100)
		;;

		enum Foo
			ZERO  # 0
			ONE   # 1
			TWO   # 2
		;;

		struct Bar
			num   : Foo
			thing : String
			age   : _age_
		;;

		# using the new method, you will have to use references
		# Bar = Struct ('num', @Foo, 'thing', String, 'age', @_age_)

		x = Bar (1, 'hey', 99)
		println (x)

		try
			x = Bar (7, 'hey', 99)
		catch
			println (__err__)
		;;

		try
			x = Bar (1, 'hey', -1)
		catch
			println (__err__)
		;;

	@out
		validating age ...
		[1, 'hey', 99]
		(EArgValue) invalid argument
		validating age ...
		(EArgValue) invalid argument
*/
qu_class QuStructClass;


/* I like it short ... */
#define item_s qu_structentry_s


/*----------------------------------------------------------------------------
	allocate
----------------------------------------------------------------------------*/

qu_struct _struct_alloc (int nfields, qu_attr_s *a, int statik)
{
	register qu_struct u = statik ? qu_gc_new_static (QuStructClass) : qu_gc_new (QuStructClass);
	u->indexes = a ? a : qu_attr_alloc ();
	u->items = qu_mem_alloc (nfields * sizeof (item_s));
	return u;
}



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (*arg) -> Struct
	Create a new Struct instance.

	v{arg} must be pairs of String/type objects.

	The type may not be -{Null}.

	@exc
	.EArgValue

	@xmp
		Person = Struct ('name', String, 'age', @_int_)
		println (Person)

	@out
		[name:String, age:_int_]
*/
/*
	This should be 'sub static new', but Qu only allows 'sub class new'.

	By doing this, we can say : foo = Person ('Foo', 123)
	instead of : foo = qu_record (Person, 'Foo', 123)
*/

static qu_struct qu__struct_new (qu_any self, int argc, qu_any *argv)
{
	if (qu_struct_is (self))
		return (qu_any) qu_record_new (self, argc, argv);

	QU_ARG_CHECK (self == QuStructClass);
	QU_REQUIRE (QuEArgCount, argc && (argc & 1) == 0);
	/*begin*/
	{
		int        n = argc;
		qu_any    *v = argv;
		int        k = 0;
		qu_any     name, type;
		qu_attr_s *a = qu_attr_alloc ();

		while (n--)
		{
			n--;
			name = *v++;
			type = *v++;
			if (!qu_string_is (name) || !qu_object_valid_type (type))
			{
				qu_attr_free (a);
				QU_ARG_THROW;
			}
			if (qu_attr_find (a, (qu_string) name) != 0)
			{
				qu_attr_free (a);
				QU_ARG_THROW;
			}
			qu_attr_add (a, (qu_string) name, qu_int_new (k));
			k++;
		}
		/*begin*/
		{
			qu_struct  u = _struct_alloc (argc, a, 0);
			item_s *i = u->items;
			while (argc--)
			{
				argc--;
				i->name = *argv++;
				i->type = qu_object_fix_type (*argv++);
				i++;
			}
			return u;
		}
	}
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class sizeof (spec: String) -> Int
	Return the size of a C struct specification.

	@exc
	.EArgValue

	@xmp
		println (Struct.sizeof ('aiis'))

	@out
		16
*/
static qu_int qu__struct_sizeof (qu_class self, qu_string spec)
{
	long z;

	QU_ARG_CHECK (qu_string_is (spec) && spec->len);
	z = qu_struct_size_ex (spec);
	QU_ARG_CHECK (z > 0);
	return qu_int_new (z);
}



/*) sub class offsetof (spec: String, index: Integer) -> Int
	Return the offset of a member in a C struct specification.

	See #{spec} for details of specifications.

	@exc
	.EArgValue

	@xmp
		println (Struct.offsetof ('aiis', 2))

	@out
		8
*/
static qu_int qu__struct_offsetof (qu_class self, qu_string spec, qu_integer idx)
{
	long ofs;
	long i;
	char type;

	QU_ARG_CHECK (qu_string_is (spec) && spec->len);
	QU_ARG_long (idx, i);
	ofs = qu_struct_offset_ex (spec, i, &type);
	QU_REQUIRE (QuEIndex, ofs >= 0);
	return qu_int_new (ofs);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__struct__mark (qu_struct self)
{
	long    n = qu_struct_len (self);
	item_s *i = self->items;

	while (n--)
	{
		qu_gc_mark (i->name);
		qu_gc_mark (i->type);
		i++;
	}
	qu_attr_mark (self->indexes);
}



static void qu__struct__destroy (qu_struct self)
{
	if (self->items)
		qu_mem_free (self->items);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} objects:
	@ops
	.| -{Dict} |
		Return a new Dict which keys are the field names and values are the
		types. Note that the order of the fields is lost because a Dict does
		not store its items in a particular order.

	.| -{Array} |
		Return an Array of consecutive pairs: name, type, name, type, ...

	.| -{Boolean} | Delegate to #{__true}
	.| -{String}  | Delegate to #{__string}
	.| -{Struct}  | Return self.

	@exc
	.EArgValue

	@xmp
		struct Person
			age : Int
			name: String
		;;

		println (Person as Dict)
		println (Person as Array)

	@out
		['name':<Class Sys.String>, 'age':<Class Sys.Int>]
		['age', <Class Sys.Int>, 'name', <Class Sys.String>]
*/
qu_dict qu_struct_as_dict (qu_struct self)
{
	long     n = qu_struct_len (self);
	qu_dict  d = qu_dict_new (qu_struct_len (self));
	item_s  *i = self->items;

	while (n--)
	{
		qu_dict_setitem (d, i->name, i->type);
		i++;
	}
	return d;
}



qu_array qu_struct_as_array (qu_struct self)
{
	long      n = qu_struct_len (self);
	qu_array  a = qu_array_new (n * 2);
	item_s   *i = self->items;

	while (n--)
	{
		qu_array_push (a, i->name);
		qu_array_push (a, i->type);
		i++;
	}
	return a;
}



static qu_any qu__struct__as (qu_struct self, qu_class c)
{
	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Dict:    return qu_struct_as_dict    (self);
		case QU_CLASS_ID_Array:   return qu_struct_as_array   (self);
		case QU_CLASS_ID_String:  return qu_struct_as_string  (self);
		case QU_CLASS_ID_Boolean: return qu_struct_as_boolean (self);
		case QU_CLASS_ID_Struct:  return self;
	}
	QU_ARG_THROW;
}



/*) sub __each -> Iterator
	Return a -{Iterator} object to iterate through the items.

	Note that unless you call this method explicitly, the interperter
	uses an internal iterator with a k{for each} loop.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		for x in each Person       # calls Person.__each
			println (x)
		;;

		for k, v in Person         # interpreter uses an internal
			println (k ' ' v)      # iterator
		;;

	@out
		name
		age
		name <Class Sys.String>
		age <Class Sys.Int>
*/
static qu_any _struct_next (qu_struct self, qu_iterator i)
{
	long k = i->ex1;
	long n = qu_struct_len (self);

	QU_ITERATOR_STOP (k < n);
	i->ex1++;

	if (qu_iterator_yield_keys (i))
		return self->items[k].name;

	else if (qu_iterator_yield_values (i))
		return self->items[k].type;

	else
	{
		qu_any argv[2] = {self->items[k].name, self->items[k].type};
		return qu_array_new_argv (2, argv);
	}
}


static qu_iterator qu__struct__each (qu_struct self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, _struct_next);
	qu_iterator_set_yield_keys (i);
	return i;
}



/*) sub __getitem (x) -> Any
	@def
	sub __getitem (x: Integer) -> Any
	sub __getitem (x: String) -> Any

	@p
	If v{x} is an integer then return the field name at that index.
	Oherwise v{x} must be a String thus return the index of that field name.

	@exc
	.EIndex

	@xmp
		struct Person
			name : String
			age  : Int
		;;
		println (Person ['name'])
		println (Person [1])

	@out
		0
		age
*/
qu_any qu__struct__getitem (qu_struct self, qu_any x)
{
	if (qu_string_is (x))
	{
		x = qu_attr_find (self->indexes, (qu_string) x);
		QU_REQUIRE (QuEIndex, x);
		assert (qu_int_is (x));
		return x;
	}
	else
	{
		long i, l = qu_struct_len (self);

		QU_ARG_long (x, i);
		QU_ENSURE_INDEX (i, l);
		return self->items[i].name;
	}
}



/*) sub __len -> Int
	Return the number of fields in the Struct.

	@noexc

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		println (Person.__len)

	@out
		2
*/
static qu_int qu__struct__len (qu_struct self)
{
	return qu_int_new (qu_struct_len (self));
}



/*) sub __like (x) -> Boolean
	If v{x} is a Struct and has the same specifications as self then
	return true, false otherwise.

	@noexc

	@xmp
		struct Foo
			bar : Int
			baz : String
		;;

		struct Person
			age  : Int
			name : String
		;;

		println (Foo == Person)
		println (Foo === Person)

	@out
		false
		true
*/
qu_boolean qu__struct__like (qu_struct self, qu_struct x)
{
	if (qu_struct_is (x))
	{
		long n = qu_struct_len (self);

		if (n == qu_struct_len (x))
		{
			item_s *i = self->items;
			item_s *j = x->items;

			while (n--)
			{
				if (i->type != j->type)
					return QuFalse;
				i++;
				j++;
			}
			return QuTrue;
		}
	}
	return QuFalse;
}




/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		s = Struct ('name', String, 'age', Int)
		println (s)

	@out
		[name:String, age:Int]
*/
static qu_null qu__struct__print (qu_struct self)
{
	if (qu_thread_state_exist (QU_THREAD_STATE_PRINT, self))
	{
		QU_ENSURE (qu_sys_printz ("<..>", 4));
	}
	else
	{
		long    i;
		qu_null r;

	  __qu_thread_state_begin (QU_THREAD_STATE_PRINT, self);
		r = qu_sys_printz ("[", 1);
		for (i = 0; r && i < qu_struct_len (self); ++i)
		{
			if (i)
				r = qu_sys_printz (", ", 2);
			if (r)
			{
				item_s *e = &self->items[i];
				r = qu_sys_print (e->name);
				if (r)
					r = qu_sys_printz (": ", 2);
				if (r)
				{
					qu_any t = e->type;
					if (qu_class_is (t))
						r = qu_sys_printz (((qu_class) t)->name->str, ((qu_class) t)->name->len);
					else if (qu_sub_is (t))
						r = qu_sys_printz (((qu_sub) t)->name->str, ((qu_sub) t)->name->len);
					else
					{
						qu_class c = qu_classof (t);
						if (c == QuVarClass && qu_var_is_typeof (t))
							c = ((qu_var) t)->value;
						r = qu_sys_printz (c->name->str, c->name->len);
					}
				}
			}
		}
		r = qu_sys_printz ("]", 1);
	  __qu_thread_state_end;
		QU_ENSURE (r);
	}
	return QuNil;
}



/*) sub __string -> String
	Return a string representation of self.

	@noexc
		struct Person
			name : String
			age  : Int
		;;

		s = Person as String
		println (s)

	@out
		[name:String, age:Int]
*/
qu_string qu__struct__string (qu_struct self)
{
	if (qu_thread_state_exist (QU_THREAD_STATE_STRING, self))
	{
		return qu_string_new ("<..>", 4);
	}
	else
	{
		long      i;
		int       r;
		qu_string s = qu_string_buffer_new (32);

	  __qu_thread_state_begin (QU_THREAD_STATE_STRING, self);
		r = qu_string_buffer_put (s, "[", 1);
		for (i = 0; r && i < qu_struct_len (self); ++i)
		{
			if (i)
				r = qu_string_buffer_put (s, ", ", 2);
			if (r)
			{
				item_s *e = &self->items[i];
				r = qu_string_buffer_puts (s, e->name);
				if (r)
					r = qu_string_buffer_put (s, ": ", 2);
				if (r)
				{
					qu_any t = e->type;
					if (qu_class_is (t))
						r = qu_string_buffer_puts (s, ((qu_class) t)->name);
					else if (qu_sub_is (t))
						r = qu_string_buffer_puts (s, ((qu_sub) t)->name);
					else
					{
						qu_class c = qu_classof (t);
						if (c == QuVarClass && qu_var_is_typeof (t))
							c = ((qu_var) t)->value;
						r = qu_string_buffer_puts (s, c->name);
					}
				}
			}
		}
		r = qu_string_buffer_putc (s, ']');
	  __qu_thread_state_end;
		QU_ENSURE (r);
		return qu_string_buffer_return (s);
	}
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub len -> Int
	This is an alias of #{__len}.
*/



/*) sub method (name: String, func: Sub) -> self
	Set specialized method.

	v{name} must be a -{String} and v{func} must be a static -{Sub}.
	When called, the function will receive the -{Record} object as self.

	This method returns self.

	@warn
	Once set, you can not modify it.

	@exc
	.EModify
	.EArgValue

	@xmp
		struct Person
			name : String
			age  : Int
		;;

		sub _Person_do (what)
			println (self.name ' (' self.age ') is ' what)
		;;

		Person.method ('doit', @_Person_do)

		x = Person ('Foo', 99)
		x.doit ('sleeping')

	@out
		Foo (99) is sleeping
*/
static qu_struct qu__struct_method (qu_struct self, qu_string name, qu_sub f)
{
	QU_ARG_CHECK (qu_string_is (name) && qu_sub_is (f) && qu_sub_is_static (f));

	if (self->attrs == 0)
		self->attrs = qu_attr_alloc ();
	else
		QU_REQUIRE (QuEModify, qu_attr_find (self->attrs, name) == 0);

	qu_attr_add (self->attrs, name, f);
	return self;
}



/*) sub ofs (index) -> Int
	Get the offset of the value if the Struct were a C struct.

	v{index} may be a field name or an integer index.

	If the Struct contains values not convertible to C types then ENone
	is thrown.

	@exc
	.EIndex
	.EValue

	@xmp
		struct Foo
			bar : _int_
			baz : _double_
		;;

		println (Foo.ofs ('baz'))

	@out
		4
*/
static qu_int qu__struct_ofs (qu_struct self, qu_any x)
{
	long i = qu_struct_index_ex (self, x);
	char type;

	QU_REQUIRE (QuEIndex, 0 <= i && i < qu_struct_len (self));
	i = qu_struct_offset (self, i, &type);
	return qu_int_new (i);
}



/*) sub size -> Int
	Return the size of the Struct if it were a C struct.

	@noexc

	@xmp
		struct Foo
			bar : _int_
			baz : _float_
			bag : String
		;;

		println (Foo.size)

	@out
		12
*/
static qu_int qu__struct_size (qu_struct self)
{
	long z = qu_struct_size (self);
	assert (z <= QU_INT_MAX);
	return qu_int_new (z);
}



/*) sub spec -> String
	Return the specification of the Struct if it were a C struct.

	The following table shows the list of character types and the corresponding
	C type, Qu "type", and Qu object.

	Note that the same letters are also used internally by Qu for parsing
	arguments passed to C functions/methods.

	@table
	.| C{a} | char               | -{Cchar}       | -{String}
	.| C{y} | byte               | -{Cbyte}       | -{Int}
	.| C{Y} | unsigned byte      | -{Cubyte}      | -{Int}
	.| C{h} | short              | -{Cshort}      | -{Int}
	.| C{H} | unsigned short     | -{Cushort}     | -{Int}
	.| C{i} | int                | -{Cint}        | -{Integer}
	.| C{I} | unsigned int       | -{Cuint}       | -{Integer}
	.| C{j} | int16              | -{Cint16}      | -{Int}
	.| C{J} | unsigned int16     | -{Cuint16}     | -{Int}
	.| C{k} | int32              | -{Cint32}      | -{Integer}
	.| C{K} | unsigned int32     | -{Cuint32}     | -{Integer}
	.| C{l} | long               | -{Clong}       | -{Integer}
	.| C{L} | unsigned long      | -{Culong}      | -{Integer}
	.| C{o} | long long          | -{Clonglong}   | -{Integer}
	.| C{O} | unsigned long long | -{Culonglong}  | -{Integer}
	.| C{g} | float              | -{Cfloat}      | -{Float}
	.| C{f} | double             | -{Cdouble}     | -{Float}
	.| C{F} | long double        | -{Clongdouble} | -{Decimal} (Float or Big)
	.| C{S} | char *             | -{Cstring}     | -{String}
	.| C{p} | void *             | -{Cpointer}    | -{Pointer}
	.| C{b} | int                | -{Boolean}     | -{Boolean}
	.| C{q} | long               | -{Clong}       | -{Int}
	.| C{z} | ssize_t            | -{Cssize_t}    | -{Integer} (Int or Long)
	.| C{Z} | size_t             | -{Csize_t}     | -{Integer} (Int or Long)
	.| C{e} | off_t              | -{Coff_t}      | -{Integer} (Int or Long)
	.| C{s} | String             | -{String}      | -{String}
	.| C{A} | Array              | -{Array}       | -{Array}
	.| C{D} | Dict               | -{Dict}        | -{Dict}
	.| C{E} | Set                | -{Set}         | -{Set}
	.| C{R} | Range              | -{Range}       | -{Range}
	.| C{C} | Class              | -{Class}       | -{Class}
	.| C{Q} | Queue              | -{Queue}       | -{Queue}
	.| C{T} | Stream             | -{Stream}      | -{Stream}
	.| C{P} | Proc               | -{Proc}        | -{Proc}
	.| C{x} | Any                | -{Any}         | -{Object}

	@warn
	'f' stands for double, not float (which is 'g').

	The reason is because 'f' stands for Qu's Float which happens to use double.
	Also, if we use 'd' then we have to use 'D' for long double which means
	that we will have to use an unatural letter for Dict.

	The real reason, of course, just because.

	@p
	By the way, 'q' stands for Qu.

	@exc
	.EIndex
	.EArgValue

	@xmp
		struct Foo
			bar : _int_
			baz : _double_
		;;

		println (Foo.spec)

	@out
		if
*/
static qu_string qu__struct_spec (qu_struct self)
{
	long       l = qu_struct_len (self);
	qu_string  s = qu_string_newl (l);
	char      *z = s->str;
	long       i;
	item_s    *t = self->items;

	for (i = 0; i < l; ++i)
	{
		item_s *e = &t[i];
		z[i] = qu_struct_type_to_char (e->type);
		QU_ARG_CHECK (z[i]);
	}
	return s;
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

static qu_struct _struct_newv (int nelt, va_list va, int statik)
{
	qu_struct  u = _struct_alloc (nelt, 0, statik);
	qu_attr_s *a = u->indexes;
	char      *s;
	qu_any     type;
	qu_string  name;
	item_s    *i;
	int        k = 0;

	u->items = i = u->items;
	u->indexes = a;

	while (nelt--)
	{
		s = va_arg (va, char *);
		name = qu_string_new_staticz (s);
		type = va_arg (va, qu_any);
		assert (qu_class_is (type) || qu_sub_is (type) || qu_var_is (type) || qu_struct_is (type));
		assert (qu_attr_find (a, name) == 0);
		qu_attr_add (a, name, qu_int_new (k));
		i->name = name;
		i->type = type;
		i++;
		k++;
	}
	return u;
}



qu_struct qu_struct_new (int nelt, ...)
{
	va_list   va;
	qu_struct u;

	va_start (va, nelt);
	u = _struct_newv (nelt, va, 0);
	va_end (va);
	return u;
}



qu_struct qu_struct_new_static (int nelt, ...)
{
	va_list   va;
	qu_struct u;

	va_start (va, nelt);
	u = _struct_newv (nelt, va, 1);
	va_end (va);
	return u;
}



long qu_struct_index (qu_struct self, qu_string x)
{
	qu_int i = qu_attr_find (self->indexes, (qu_string) x);
	if (i == 0)
		return -1;
	return qu_int_tol (i);
}



long qu_struct_index_ex (qu_struct self, qu_any x)
{
	if (qu_string_is (x))
		return qu_struct_index (self, x);

	if (qu_integer_is (x))
	{
		long i = qu_integer_tol (x);
		if (errno == 0)
		{
			if (i < 0)
				i += qu_struct_len (self);
			if (0 <= i && i < qu_struct_len (self))
				return i;
		}
	}
	return -1;
}



/* one of those places where C really sucks ... */
typedef struct { char c; char        x; } qu_struct_char_s;
typedef struct { char c; short       x; } qu_struct_short_s;
typedef struct { char c; int         x; } qu_struct_int_s;
typedef struct { char c; char p,     x; } qu_struct_int16_s;
typedef struct { char c; char p,q,r, x; } qu_struct_int32_s;
typedef struct { char c; long        x; } qu_struct_long_s;
typedef struct { char c; float       x; } qu_struct_float_s;
typedef struct { char c; double      x; } qu_struct_double_s;
typedef struct { char c; void       *x; } qu_struct_ptr_s;
typedef struct { char c; size_t      x; } qu_struct_size_t_s;
typedef struct { char c; ssize_t     x; } qu_struct_ssize_t_s;
typedef struct { char c; off_t       x; } qu_struct_off_t_s;
#ifdef HAVE_LONG_LONG
typedef struct { char c; long long   x; } qu_struct_llong_s;
#endif
#ifdef HAVE_LONG_DOUBLE
typedef struct { char c; long double x; } qu_struct_ldouble_s;
#endif

#define ALIGN_BYTE        (sizeof (qu_struct_char_s   ) - 1)
#define ALIGN_SHORT       (sizeof (qu_struct_short_s  ) - sizeof (short))
#define ALIGN_INT         (sizeof (qu_struct_int_s    ) - sizeof (int))
#define ALIGN_INT16       (sizeof (qu_struct_int16_s  ) - 2)
#define ALIGN_INT32       (sizeof (qu_struct_int32_s  ) - 4)
#define ALIGN_LONG        (sizeof (qu_struct_long_s   ) - sizeof (long))
#define ALIGN_FLOAT       (sizeof (qu_struct_float_s  ) - sizeof (float))
#define ALIGN_DOUBLE      (sizeof (qu_struct_double_s ) - sizeof (double))
#define ALIGN_PTR         (sizeof (qu_struct_ptr_s    ) - sizeof (void *))
#define ALIGN_SIZE_T      (sizeof (qu_struct_size_t_s ) - sizeof (size_t))
#define ALIGN_SSIZE_T     (sizeof (qu_struct_ssize_t_s) - sizeof (ssize_t))
#define ALIGN_OFF_T       (sizeof (qu_struct_off_t_s  ) - sizeof (off_t))
#ifdef HAVE_LONG_LONG
#define ALIGN_LONG_LONG   (sizeof (qu_struct_llong_s  ) - sizeof (long long))
#endif
#ifdef HAVE_LONG_DOUBLE
#define ALIGN_LONG_DOUBLE (sizeof (qu_struct_ldouble_s) - sizeof (long double))
#endif

qu_struct_type_s __QuStructTypeList [256];


static int top_id (qu_class type)
{
	while (type->super)
		type = type->super;
	return type->id;
}


char qu_struct_type_to_char (qu_any type)
{
	if (qu_string_is (type))
	{
		char t = ((qu_string) type)->str[0];
		qu_struct_type_s *m = qu_struct_type_info (t);
		return m->type ? t : 0;
	}

	switch (qu_classid (type))
	{
		case QU_CLASS_ID_Sub:
		{
			qu_sub y = type;

			if (y == QuType_char      ) return QU_STRUCT_ID_char;
			if (y == QuType_byte      ) return QU_STRUCT_ID_byte;
			if (y == QuType_ubyte     ) return QU_STRUCT_ID_ubyte;
			if (y == QuType_short     ) return QU_STRUCT_ID_short;
			if (y == QuType_ushort    ) return QU_STRUCT_ID_ushort;
			if (y == QuType_int       ) return QU_STRUCT_ID_int;
			if (y == QuType_uint      ) return QU_STRUCT_ID_uint;
			if (y == QuType_int16     ) return QU_STRUCT_ID_int16;
			if (y == QuType_uint16    ) return QU_STRUCT_ID_uint16;
			if (y == QuType_int32     ) return QU_STRUCT_ID_int32;
			if (y == QuType_uint32    ) return QU_STRUCT_ID_uint32;
			if (y == QuType_long      ) return QU_STRUCT_ID_long;
			if (y == QuType_ulong     ) return QU_STRUCT_ID_ulong;
			if (y == QuType_longlong  ) return QU_STRUCT_ID_longlong;
			if (y == QuType_ulonglong ) return QU_STRUCT_ID_ulonglong;
			if (y == QuType_float     ) return QU_STRUCT_ID_float;
			if (y == QuType_double    ) return QU_STRUCT_ID_double;
			if (y == QuType_longdouble) return QU_STRUCT_ID_longdouble;
			if (y == QuType_ssize_t   ) return QU_STRUCT_ID_ssize_t;
			if (y == QuType_size_t    ) return QU_STRUCT_ID_size_t;
			if (y == QuType_off_t     ) return QU_STRUCT_ID_off_t;
			if (y == QuType_char_p    ) return QU_STRUCT_ID_char_p;
			if (y == QuType_void_p    ) return QU_STRUCT_ID_void_p;
#ifdef QU_SUPPORT_GLIB
			if (y == QuType_GObject   ) return QU_STRUCT_ID_GObject;
#endif
			break;
		}

		case QU_CLASS_ID_Class:
		{
			switch (top_id ((qu_class) type))
			{
				case QU_CLASS_ID_USER:    return QU_STRUCT_ID_Class;
				case QU_CLASS_ID_Float:   return QU_STRUCT_ID_double;
				case QU_CLASS_ID_Null:    return QU_STRUCT_ID_Null;
				case QU_CLASS_ID_Boolean: return QU_STRUCT_ID_Boolean;
				case QU_CLASS_ID_Int:     return QU_STRUCT_ID_Int;
				case QU_CLASS_ID_Pointer: return QU_STRUCT_ID_Pointer;
				case QU_CLASS_ID_String:  return QU_STRUCT_ID_String;
				case QU_CLASS_ID_Array:   return QU_STRUCT_ID_Array;
				case QU_CLASS_ID_Dict:    return QU_STRUCT_ID_Dict;
				case QU_CLASS_ID_Set:     return QU_STRUCT_ID_Set;
				case QU_CLASS_ID_Range:   return QU_STRUCT_ID_Range;
				case QU_CLASS_ID_Class:   return QU_STRUCT_ID_Class;
				case QU_CLASS_ID_Queue:   return QU_STRUCT_ID_Queue;
				case QU_CLASS_ID_Stream:  return QU_STRUCT_ID_Stream;
				case QU_CLASS_ID_Proc:    return QU_STRUCT_ID_Proc;
				case QU_CLASS_ID_Object:  return QU_STRUCT_ID_Any;
#ifdef QU_SUPPORT_GLIB
				case QU_CLASS_ID_GObject: return QU_STRUCT_ID_GObject;
#endif
			}
		}
	}
	return 0;
}


int qu_struct_valid_spec (char *s, long len)
{
	qu_struct_type_s *i;

	while (len--)
	{
		i = qu_struct_type_info (*s);
		if (i->size == 0)
			return 0;
	}
	return 1;
}



long qu_struct_offset (qu_struct self, long k, char *type)
{
	long    l   = qu_struct_len (self);
	item_s *t   = self->items;
	long    ofs = 0;
	long    size;
	long    i;
	item_s *e;
	char    y;
	qu_struct_type_s *m;

	if (k < 0 || k >= l)
		return -1;

	for (i = 0; i < k; ++i, ofs += size)
	{
		e = &t[i];
		y = qu_struct_type_to_char (e->type);
		assert (y);
		m = qu_struct_type_info (y);
		size = m->size;
		while (((ofs + sizeof (void *)) % m->align))
			++ofs;
	}
	e = &t[k];
	y = qu_struct_type_to_char (e->type);
	assert (y);
	m = qu_struct_type_info (y);
	while (((ofs + sizeof (void *)) % m->align))
		++ofs;
	y = qu_struct_type_to_char (e->type);
	assert (y);
	*type = y;
	return ofs;
}



long qu_struct_offset_ex (qu_any spec, long k, char *type)
{
	if (qu_string_is (spec))
	{
		long  l = ((qu_string) spec)->len;
		char *z = ((qu_string) spec)->str;
		long  size;
		long  ofs = 0;
		long  i;
		qu_struct_type_s *m;

		if (k < 0 || k >= l)
			return -1;

		for (i = 0; i < k; ++i, ofs += size)
		{
			m = qu_struct_type_info (z[i]);
			if (m->type == 0)
				return -1;
			size = m->size;
			while (((ofs + sizeof (void *)) % m->align))
				++ofs;
		}
		m = qu_struct_type_info (z[k]);
		if (m->type == 0)
			return -1;
		while (((ofs + sizeof (void *)) % m->align))
			++ofs;
		*type = z[k];
		return ofs;
	}
	else if (qu_struct_is (spec))
	{
		return qu_struct_offset (spec, k, type);
	}
	return -1;
}



long qu_struct_size (qu_struct self)
{
	char type;
	long i = qu_struct_len (self) - 1;
	long z = qu_struct_offset (self, i, &type);

	z += __QuStructTypeList [(unsigned char) type].size;
	return z + (z % sizeof (void *));
}



long qu_struct_size_ex (qu_any x)
{
	if (qu_struct_is (x))
	{
		return qu_struct_size (x);
	}
	else if (qu_string_is (x))
	{
		char type;
		long i = ((qu_string) x)->len - 1;
		long size = qu_struct_offset_ex (x, i, &type);
		qu_struct_type_s *m;

		if (size < 0)
			return -1;

		m = qu_struct_type_info (((qu_string) x)->str [i]);
		size += m->size;
		size += (size % sizeof (void *));
		return size;
	}
	return -1;
}



char qu_struct_type (qu_struct self, long i)
{
	long    l = qu_struct_len (self);
	item_s *t = self->items;
	item_s *e;

	if (i < 0 || i >= l)
		return 0;
	e = &t[i];
	return qu_struct_type_to_char (e->type);
}



/* compile-time only */
int qu_struct_compile_setup (qu_var v, qu_ast n)
{
	/*
		struct Foo
			name   : qu_string
			friend : Foo
					   \__ qu_var->value = nil because Foo has not been
					       created yet, but it has the VarSTRUCT flag
	*/
	int        i, l = 0;
	qu_struct  u = v->value;
	qu_string  name;
	qu_any     type;
	qu_attr_s *a;
	item_s    *e;
	qu_ast     c;

	assert (u && qu_struct_is (u) && u->items == 0);

	for (c = n->child; c; c = c->next)
	{
		c = c->next;
		l++;
	}

	u->items = qu_mem_alloc (l * sizeof (item_s));

	a = u->indexes;
	e = u->items;

	for (i = 0, c = n->child; c; c = c->next, ++i, ++e)
	{
		name = c->thing;
		assert (qu_string_is (name));

		c = c->next;
		type = qu_object_fix_type (c->thing);
		if (type == QuNil)
			type = u;

		if (!qu_object_valid_type (type))
			goto nope;

		if (qu_attr_find (a, name))
			goto nope;

		qu_attr_add (a, name, qu_int_new (i));
		e->name = name;
		e->type = type;
	}
	return 1;

  nope:
	qu_attr_free (a);
	QU_THROW (QuECompile);
}



/* compile-time only */
qu_struct qu_struct_compile_prealloc ()
{
	register qu_struct u = qu_gc_new (QuStructClass);
	u->indexes = qu_attr_alloc ();
	u->items = 0;
	return u;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

qu_sub QuStructMethodNew;


void qu_struct__setup ()
{
	qu_class_setup_final (QuStructClass, qu_struct_s, qu__struct__destroy, qu__struct__mark);
}



void qu_struct__init ()
{
	/* standard class methods */
	QuStructMethodNew = qu_class_method_class  (QuStructClass, QU_NAME_new, qu__struct_new, -1);
		qu_flag_unset (QuStructMethodNew, QU_SUB_FLAG_CLASS);
		qu_flag_set   (QuStructMethodNew, QU_SUB_FLAG_STATIC);

	/* specific class methods */
	qu_class_method_classz (QuStructClass, "offsetof", qu__struct_offsetof, 2);
	qu_class_method_classz (QuStructClass, "sizeof"  , qu__struct_sizeof  , 1);

	/* standard instance methods */
	qu_class_method_inst   (QuStructClass, QU_NAME__as     , qu__struct__as     , 1);
	qu_class_method_inst   (QuStructClass, QU_NAME__each   , qu__struct__each   , 0);
	qu_class_method_inst   (QuStructClass, QU_NAME__getitem, qu__struct__getitem, 1);
	qu_class_method_inst   (QuStructClass, QU_NAME__len    , qu__struct__len    , 0);
	qu_class_alias         (QuStructClass, QU_NAME__len    , QU_NAME_len);
	qu_class_method_inst   (QuStructClass, QU_NAME__like   , qu__struct__like   , 1);
	qu_class_method_inst   (QuStructClass, QU_NAME__print  , qu__struct__print  , 0);
	qu_class_method_inst   (QuStructClass, QU_NAME__string , qu__struct__string , 0);

	/* specific instance methods */
	qu_class_method_instz  (QuStructClass, "method", qu__struct_method, 2);
	qu_class_method_instz  (QuStructClass, "ofs"   , qu__struct_ofs   , 1);
	qu_class_method_instz  (QuStructClass, "spec"  , qu__struct_spec  , 0);
	qu_class_method_instz  (QuStructClass, "size"  , qu__struct_size  , 0);

	/* finalize declaration */
	qu_class_enddecl (QuStructClass);
}



static void _structtypeentry (char ch, qu_any type, int align, int size, const char *name)
{
	qu_struct_type_s *e = qu_struct_type_info (ch);

	e->type  = type;
	e->align = align;
	e->size  = size;
	e->name  = name;
}



void qu_struct_sys_init ()
{

#define e_(ch, type, align, size, name) \
	_structtypeentry (ch, type, align, size, name)

	memset (__QuStructTypeList, 0, sizeof (__QuStructTypeList));

	e_(QU_STRUCT_ID_char      , QuType_char      , ALIGN_BYTE       , 1                   , "Cchar");
	e_(QU_STRUCT_ID_byte      , QuType_byte      , ALIGN_BYTE       , 1                   , "Cbyte");
	e_(QU_STRUCT_ID_ubyte     , QuType_ubyte     , ALIGN_BYTE       , 1                   , "Cubyte");
	e_(QU_STRUCT_ID_short     , QuType_short     , ALIGN_SHORT      , sizeof (short)      , "Cshort");
	e_(QU_STRUCT_ID_ushort    , QuType_ushort    , ALIGN_SHORT      , sizeof (short)      , "Cushort");
	e_(QU_STRUCT_ID_int       , QuType_int       , ALIGN_INT        , sizeof (int)        , "Cint");
	e_(QU_STRUCT_ID_uint      , QuType_uint      , ALIGN_INT        , sizeof (int)        , "Cuint");
	e_(QU_STRUCT_ID_int16     , QuType_int16     , ALIGN_INT16      , 2                   , "Cint16");
	e_(QU_STRUCT_ID_uint16    , QuType_uint16    , ALIGN_INT16      , 2                   , "Cuint16");
	e_(QU_STRUCT_ID_int32     , QuType_int32     , ALIGN_INT32      , 4                   , "Cint32");
	e_(QU_STRUCT_ID_uint32    , QuType_uint32    , ALIGN_INT32      , 4                   , "uint32");
	e_(QU_STRUCT_ID_long      , QuType_long      , ALIGN_LONG       , sizeof (long)       , "Clong");
	e_(QU_STRUCT_ID_ulong     , QuType_ulong     , ALIGN_LONG       , sizeof (long)       , "Culong");
#ifdef HAVE_LONG_LONG
	e_(QU_STRUCT_ID_longlong  , QuType_longlong  , ALIGN_LONG_LONG  , sizeof (long long)  , "Clonglong");
	e_(QU_STRUCT_ID_ulonglong , QuType_ulonglong , ALIGN_LONG_LONG  , sizeof (long long)  , "Culonglong");
#endif
	e_(QU_STRUCT_ID_float     , QuType_float     , ALIGN_FLOAT      , sizeof (float)      , "Cfloat");
	e_(QU_STRUCT_ID_double    , QuType_double    , ALIGN_DOUBLE     , sizeof (double)     , "Cdouble");
#ifdef HAVE_LONG_DOUBLE
	e_(QU_STRUCT_ID_longdouble, QuType_longdouble, ALIGN_LONG_DOUBLE, sizeof (long double), "Clongdouble");
#endif
	e_(QU_STRUCT_ID_ssize_t   , QuType_ssize_t   , ALIGN_SSIZE_T    , sizeof (ssize_t)    , "Cssize_t");
	e_(QU_STRUCT_ID_size_t    , QuType_size_t    , ALIGN_SIZE_T     , sizeof (size_t)     , "Csize_t");
	e_(QU_STRUCT_ID_off_t     , QuType_off_t     , ALIGN_OFF_T      , sizeof (off_t)      , "Coff_t");
	e_(QU_STRUCT_ID_char_p    , QuType_char_p    , ALIGN_PTR        , sizeof (char *)     , "Cstring");
	e_(QU_STRUCT_ID_void_p    , QuType_void_p    , ALIGN_PTR        , sizeof (void *)     , "Cpointer");

	e_(QU_STRUCT_ID_Null      , QuNullClass      , ALIGN_PTR        , sizeof (void *)     , "NULL");
	e_(QU_STRUCT_ID_Boolean   , QuBooleanClass   , ALIGN_INT        , sizeof (int)        , "boolean");
	e_(QU_STRUCT_ID_Int       , QuIntClass       , ALIGN_LONG       , sizeof (long)       , "Int");
	e_(QU_STRUCT_ID_Pointer   , QuPointerClass   , ALIGN_PTR        , sizeof (void *)     , "void *");
	e_(QU_STRUCT_ID_String    , QuStringClass    , ALIGN_PTR        , sizeof (char *)     , "char *");
	e_(QU_STRUCT_ID_Array     , QuArrayClass     , ALIGN_PTR        , sizeof (qu_array)   , "Array");
	e_(QU_STRUCT_ID_Dict      , QuDictClass      , ALIGN_PTR        , sizeof (qu_dict)    , "Dict");
	e_(QU_STRUCT_ID_Set       , QuSetClass       , ALIGN_PTR        , sizeof (qu_set)     , "Set");
	e_(QU_STRUCT_ID_Range     , QuRangeClass     , ALIGN_PTR        , sizeof (qu_range)   , "Range");
	e_(QU_STRUCT_ID_Class     , QuClassClass     , ALIGN_PTR        , sizeof (qu_class)   , "Class");
	e_(QU_STRUCT_ID_Queue     , QuQueueClass     , ALIGN_PTR        , sizeof (qu_queue)   , "Queue");
	e_(QU_STRUCT_ID_Stream    , QuStreamClass    , ALIGN_PTR        , sizeof (qu_stream)  , "Stream");
	e_(QU_STRUCT_ID_Proc      , QuProcClass      , ALIGN_PTR        , sizeof (qu_proc)    , "Proc");
	e_(QU_STRUCT_ID_Any       , QuObjectClass    , ALIGN_PTR        , sizeof (qu_any)     , "Any");
#ifdef QU_SUPPORT_GLIB
	e_(QU_STRUCT_ID_GObject   , QuGObjectClass   , ALIGN_PTR        , sizeof (qu_gobject) , "GObject");
#endif

#undef e_

}
