/* Sys.c - Copyright (c) Marc Krisnanto */

#include "qu.h"
#include "Inlines.h"


/*) include Array.c */
/*) include Ast.c */
/*) include Big.c */
/*) include Boolean.c */
/*) include Buffer.c */
/*) include Class.c */
/*) include Code.c */
/*) include Complex.c */
/*) include Container.c */
/*) include Decimal.c */
/*) include Dict.c */
/*) include Dlib.c */
/*) include Exception.c */
/*) include File.c */
/*) include Float.c */
/*) include Frame.c */
/*) include Function.c */
/*) include Int.c */
/*) include Integer.c */
/*) include Stream.c */
/*) include Iterator.c */
/*) include Long.c */
/*) include Method.c */
/*) include Module.c */
/*) include Null.c */
/*) include Number.c */
/*) include Object.c */
/*) include Opaque.c */
/*) include Proc.c */
/*) include Pointer.c */
/*) include Queue.c */
/*) include Real.c */
/*) include Range.c */
/*) include Record.c */
/*) include Set.c */
/*) include String.c */
/*) include Struct.c */
/*) include Sub.c */
/*) include Thread.c */
/*) include Tokenizer.c */
/*) include Var.c */
/*) include Weak.c */


/*) module Sys
	The Sys module provides all the builtin classes and a number of functions.

	You don't need to C{use Sys} unless there are name conflicts in your module
	or program.

	@xmp
		use Sys

		sub print (x)
			Sys.print ('x is ', x)
		;;

		print (10)

	@p
	Some functions simply exist for convenience, some are provided because it
	is not possible to immitate their functionality in program level, and some
	are provided for performance reasons (they can be immitated in program level
	but the performance is probably unsatisfactory).

	@note
	Some functions are named with leading and trailing double underscores, e.g.
	#{__abort__}, #{__warning__}, etc. The intention is to make them more visible
	since they perform something special. A color editor should render them
	differently so it would be easy to spot them.



	@h3
	__memory__
		@indent
		Memory is handled automatically in Qu. You never have to allocate nor free
		memory yourself. If you create new objects, memory is allocated. When an
		object is no longer referenced, it is disposed by the Garbage Collector.

		Qu follows an optimistic memory allocation strategy.
		If you run out of memory then Qu will abort with a fatal error message.

		Qu considers running out of memory as a programmatic error.
		In other words, you better make sure that your program has enough memory
		before it runs.

		At first you might think that this is a bad approach. But in practice,
		there is almost nothing you can do if you run out of memory. Even if you
		try, you will most probably never succeed and in most cases it will simply
		lead to unecessary complications and sometimes even to disaster.

		If you get a EMemory exception in Qu, it simply means that you are
		trying to allocate an impossible amount of memory (too big for the
		system f{malloc} function). It does not mean that you actually run out
		of memory.

		So now that you know all that ...

		If your main program defines a function named __memory__ then that
		function will be called when memory allocation fails.

		This function must be a static function that takes no argument or the
		the interpreter will refuse to run with a fatal error.

		@warn
		Make sure you do the minimum amount of work there and don't allocate
		memory otherwise another fatal error will terminate it.

		@xmp
			sub __memory__

				# ... do not create new objects
				# ... only call functions/methods that
				# ... do not create new objects

				println ("*** dang! we're out of mem'ry ****");
				__abort__
			;;

		@p
		You don't need to call #{__abort__} but it's your choice.
		Qu will abort anyway.
		@unindent


	@h3
	Importing symbols to the Sys module
		@indent
		External modules may import symbols to the Sys module provided that:
		@ops
		.| __xxx__ |
			All lowercase symbols that start and end with a double underscore
			are reserved by the Sys module.

		.| _xxx_ |
			All lowercase symbols that start and end with a single underscore
			must only be used for validators.
		@unindent
*/
qu_module QuSysModule;



/*----------------------------------------------------------------------------
	handlers
----------------------------------------------------------------------------*/

static qu_boolean _sys_ensure_handler (qu_any x)
{
	__qu_unsafe__;
	if (x == QuNil || (qu_sub_is (x) && qu_sub_is_static (x) && ((qu_sub) x)->narg == 0))
		return QuTrue;
	return QuFalse;
}


/*) var __print__ = nil
	#{print} and family may throw an exception, for example, if -{Stream.stdout}
	is closed or points to a file and an I/O error occured.

	You can trap this exception by specifying a handler via this variable.

	The v{handler} must return nil to ignore the exception or throw any
	exception otherwise.

	@xmp
		sub handler
			__warning__ ('*** ' __err__.msg ' ***')
			throw
		;;

		__print__ = @handler
		println ('closing stdout')
		Stream.stdout.close
		println ('Hello')

	@out
		closing stdout
		WARNING [1024] in __main__.handler: *** Bad file descriptor ***
		[1024] __main__ line 5 (EBADF) Bad file descriptor
*/
static qu_var QuSys__print__;



/*) var __keyboard__ = nil
	You can trap keyboard interrupts by specifying a handler via this variable.

	The v{handler} must be a static Sub with no arguments. It must return
	nil to ignore the exception or throw any exception otherwise.

	@xmp
		sub handler
			println ('*** keyboard interrupt ***')
			throw EInterrupt
		;;

		__keyboard__ = @handler
		loop
			# press ctrl-c here
			pass
		;;

*/
qu_var QuSys__keyboard__;



/*) var __quit__ __public = nil
	EExit and EAssert are not catchable. They will propagate to the main program
	and causes an abnormal termination by calling the system function f{exit}.

	You can set a function to call before f{exit} is called via this variable.
	The function must be a static Sub with no arguments.

	@xmp
		sub handler
			println ('*** exit ***')
		;;

		__quit__ = @handler

		loop
			# open another terminal and send a kill TERM to this program
			pass
		;;

		# or alternatively: __exit__ (1)
*/
qu_var QuSys__quit__;



/*) const __debug__ = false
	This is a global constant. It's value is true if you start Qu with
	the C{-d} option, false otherwise.
*/
qu_var QuSys__debug__;



/*) const __daemon__ = false
	This is a global constant. It's value is treu if you start Qu with the
	C{-S} option (run as daemon), false otherwise.
*/
static qu_var QuSys__daemon__;


/*) var __arg__ := []
	This is an -{Array} of String objects passed from the command line.

	The Array is frozen so it can not be modified. If you want to modify
	it then you will have to copy it.

	@xmp
		# qu test.qu 10 baz -p -rt -w100 foo/bar -g 123
		println (__arg__)

		TheArgs = __arg__.__copy

	@out
		['10', 'baz', '-p', '-rt', '-w100', 'foo/bar', '-g', '123']
*/
qu_var QuSys__arg__;



/*) var __
	This is the only global read/write variable that you may use in k{for}
	statements to ignore the items.

	@xmp
	for __ in something
		dothings
	;;
*/
qu_var QuSys__ignore;



/* list of exceptions to print */
int volatile QuSysTrace = 0;
qu_var QuSys__tracefilter__ = 0;
qu_var QuSys__tracehow__ = 0;


/*----------------------------------------------------------------------------
	warnings and errors
----------------------------------------------------------------------------*/

void qu_sys__warn (const char *fmt, ...)
{
	va_list   va;
	qu_string s;

	va_start (va, fmt);
	fprintf (stderr, "[" QU_THREAD_IDFORMAT "] WARNING", qu_thread_id ());
	s = qu_interp_where_string ();
	if (s && s->len)
		fprintf (stderr, " in %s", s->str);
	else
		fprintf (stderr, " in __main__");
	fprintf (stderr, ": ");
	vfprintf (stderr, fmt, va);
	fprintf (stderr, "\n");
	va_end (va);
}



void qu_sys__warn_class (qu_any self, const char *msg)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	fprintf (stderr, "[" QU_THREAD_IDFORMAT "] WARNING class %s: %s\n",
		qu_thread_id (), c->name->str, msg);
}



void qu_sys__warn_retval (qu_any self, const char *name, const char *expected)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	fprintf (stderr, "[" QU_THREAD_IDFORMAT "] WARNING %s.%s did not return a %s\n",
		qu_thread_id (), c->name->str, name, expected);
}



void qu_sys__eimplement (qu_any self, const char *name, const char *func)
{
	qu_class c = qu_class_is (self) ? (qu_class) self : qu_classof (self);
	qu_thread__throw (QuEImplementClass, qu_string_newf ("%s.%s is not implemented", c->name->str, name), func);
}



/*----------------------------------------------------------------------------
	print internals
----------------------------------------------------------------------------*/

qu_null qu_sys_print (qu_any x)
{
	qu_class c = qu_classof (x);
	qu_array u = __QuThreadCurrent->buffers;

	if (u && u->len)
	{
		/*
			We are buffering the print.

			Buffering is per-thread basis so there can be no conflict, i.e.
			if it were done by qu_stream.stdout then the print will be a mess.

			This happens with a code such as:

				__buffer__
				print ('foo bar baz')
				...

			or indirectly, see qu_sys_format for example.
		*/

		qu_string b = u->objs [u->len - 1];
		qu_string s;

		assert (qu_string_is (b));
		s =  qu_sys_string (x);
		QU_ENSURE (qu_string_buffer_puts (b, s));
	}
	else
	{
		/*
			Not buffering, just print.
			The __print method is expected to call qu_sys_printz or qu_sys_print.

			!!! If qu_sys_print is used then the method must not pass self again
			!!! unless it can stop recursive calls (e.g. qu_array.__print).
			!!! Otherwise it should call qu_sys_print with a qu_string object.
			!!! qu_string.__print calls qu_sys_printz so it stops recursive-ness.
		*/

		if (c->accel)
		{
			if (c->accel->fprint)
				x = c->accel->fprint (x);
			else
				x = qu_object_default_print (x);
		}
		else
		{
			x = qu_object_call0 (x, QU_NAME__print);
		}
		QU_ENSURE (x);
	}
	return QuNil;
}



/*
	This is just a variation of qu_sys_print.
	's' must point to static memory or stack.
	It must not come from, say, SomeString->str.
*/
qu_null qu_sys_printz (const char *s, long len)
{
	qu_array u = __QuThreadCurrent->buffers;

	if (u && u->len)
	{
		qu_string b = u->objs [u->len - 1];
		assert (qu_string_is (b));
		QU_ENSURE (qu_string_buffer_put (b, s, len));
	}
	else
	{
		QU_ENSURE (qu_stream_printz (QuStreamStdout, s, len));
	}
	return QuNil;
}



/*
	TODO - program level function "parseformat"?
*/
static qu_string FormatBadStr;


/* note that this must be Qu atomic -- see qu_format_parse */
static char _format_info (int what)
{
	char *s = "_";

	switch (what)
	{
		/*- - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
		case 0:
#if defined(HAVE_LANGINFO_H) && defined (THOUSANDS_SEP)
			s = nl_langinfo (THOUSANDS_SEP);
			if (s == NULL)
#elif defined(HAVE_LANGINFO_H) && defined (THOUSEP)
			s = nl_langinfo (THOUSEP);
			if (s == NULL)
#else
			{
				struct lconv *l = localeconv ();
				s = l->thousands_sep;
			}
#endif
			if (s == NULL || s[0] == '\0')
				s = ",";
			break;

		/*- - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
 		case 1:
#if defined(HAVE_LANGINFO_H) && defined (DECIMAL_POINT)
			s = nl_langinfo (DECIMAL_POINT);
			if (s == NULL)
#else
			{
				struct lconv *l = localeconv ();
				s = l->decimal_point;
			}
#endif
			if (s == NULL || s[0] == '\0')
				s = ".";
			break;

		/*- - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
		case 2:
#if defined(HAVE_LANGINFO_H) && defined (MON_THOUSANDS_SEP)
			s = nl_langinfo (MON_THOUSANDS_SEP);
			if (s == NULL)
#else
			{
				struct lconv *l = localeconv ();
				s = l->mon_thousands_sep;
			}
#endif
			if (s == NULL || s[0] == '\0')
				s = ",";
			break;

		/*- - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
		case 3:
#if defined(HAVE_LANGINFO_H) && defined (MON_DECIMAL_POINT)
			s = nl_langinfo (MON_DECIMAL_POINT);
			if (s == NULL)
#else
			{
				struct lconv *l = localeconv ();
				s = l->mon_thousands_sep;
			}
#endif
			if (s == NULL || s[0] == '\0')
				s = ".";
			break;
	}
	return s[0];
}



/*
	Parse a format specification.
	Note that this must be Qu atomic so that v{i} may come from a non-frozen qu_string.
*/
int qu_format_parse (const char *i, qu_format_s *f)
{
	f->mode   = 'd';
	f->align  = 0;
	f->sign   = 0;
	f->zero   = 0;
	f->group  = -1;
	f->width  = 0;
	f->prec   = 0;
	f->ucase  = 1;
	f->sep    = _format_info (0);
	f->dec    = _format_info (1);

	if (*i == '%')
		i++;

	switch (*i)
	{
		case '-' :
		case '<' :
			f->align = 1;
			i++;
			break;
	}

	switch (*i)
	{
		case '+' :
			f->sign = 1;
			i++;
			break;

		case ' ' :
			f->sign = 2;
			i++;
			break;

		case '*' :
			f->sign = -1;
			i++;
			break;
	}

	if (qu_ascii_isdigit (*i))
	{
		int w = 0;
		if (*i == '0')
			f->zero = 1;
		do
		{
			w = w * 10 + (*i++ - '0');
			QU_REQUIRE (QuEFormat, w <= QU_FORMAT_MAX_WIDTH);

		} while (qu_ascii_isdigit (*i));
		f->width = w;
	}

	if (*i == '.')
	{
		int p = 0;
		i++;
		/* CHECKME - ignore this --- see _sys_printf */
		if (*i == '-')
			i++;
		while (qu_ascii_isdigit (*i))
		{
			p = p * 10 + (*i++ - '0');
			QU_REQUIRE (QuEFormat, p <= QU_BIG_MAXPREC);
		}
		f->prec = p;
	}

	if (*i == ',')
	{
		int g = 0;
		i++;
		while (qu_ascii_isdigit (*i))
		{
			g = g * 10 + (*i++ - '0');
			QU_REQUIRE (QuEFormat, g < QU_INT_MAX);
		}
		f->group = g;
	}

	switch (*i)
	{
		case 'd':
			f->ucase = 0;
		case 'D':
			f->mode = 'd';
			break;

		case 'e':
			f->ucase = 0;
		case 'E':
			f->mode = 'e';
			break;

		case 'f':
			f->ucase = 0;
		case 'F':
			f->mode = 'f';
			break;

		case 'm':
			f->ucase = 0;
		case 'M':
			f->sep = _format_info (2);
			f->dec = _format_info (3);
			f->mode = 'm';
			break;

		case 'x':
			f->ucase = 0;
		case 'X':
			f->mode = 'x';
			break;

		case 'b':
			f->ucase = 0;
		case 'B':
			f->mode = 'b';
			break;

		case 'o':
			f->ucase = 0;
		case 'O':
			f->mode = 'o';
			break;

		case 'p':
			f->ucase = 0;
		case 'P':
			f->mode = 'p';
			break;

		case 's':
		case 'S':
			f->mode = 's';
			break;

		case 'y':
			f->ucase = 0;
		case 'Y':
			f->mode = 'y';
			break;

		default:
			if (*i)
				f->mode = *i;
	}

	/* what a total crap */
	if (f->dec == '.')
		f->sep = ',';
	else if (f->dec == ',')
		f->sep = '.';
	else
		f->sep = ' ';
	return 1;
}



qu_string qu_format_number (qu_format_s *f, qu_string s, int neg, int expo)
{
	/*
		TODO - the dang 'f' mode
	*/
	qu_string  sgn, pre, zro, inte, dec, frac0, frac, frac1, exo;
	char      *z;
	int        l;
	char       mode = f->mode;
	int        prec = f->prec;
	int        sign = f->sign;

  again:
	z = s->str;
	l = s->len;

	sgn = QuStringEMPTY;
	if (neg)
	{
		if (sign >= 0)
			sgn = qu_string_byte ('-');
	}
	else
	{
		if (sign == 2)
			sgn = qu_string_byte (' ');
		else if (sign == 1)
			sgn = qu_string_byte ('+');
	}

	inte = (qu_string) s;
	zro = frac = frac0 = frac1 = exo = dec = QuStringEMPTY;

	if (prec)
		dec = qu_string_byte (f->dec);

	if (mode == 'd' || mode == 'm' || mode == 'f')
	{
		/*
			format with no exponent
			the only difference between 'd' and 'm' is the decimal point
			and the group separator

			12345678e-3 -> 123.4567
		*/

		if (expo == 0)
		{
			/*
				12345678e+0 -> 12345678
				12345678e+0 -> 12345678.000
			*/
			inte = (qu_string) s;
			if (prec > 0)
				frac = qu_string_newc ('0', prec);
		}
		else if (expo > 0)
		{
			/*
				12345678e+2 -> 1234567800
			*/
			qu_string p = qu_string_newc ('0', expo);
			inte = qu_string_cat (inte, p, NULL);
			if (prec > 0)
				frac = qu_string_newc ('0', prec);
		}
		else
		{
			expo = -expo;
			if (expo > l)
			{
				/*
					12345678e-9      -> 0.012345678
					12345678e-11     -> 0.00012345678
					12345678e-100000 -> 0.000000000000000...12345678
				*/
				inte = qu_string_byte ('0');
				if (prec > 0)
				{
					int n = expo - l;
					if (n >= prec)
					{
						frac = qu_string_newc ('0', prec);
					}
					else
					{
						qu_string p = qu_string_newc ('0', n);
						int m = prec - n;
						if (m < l)
						{
							frac  = p;
							frac1 = qu_string_new (z, m);
						}
						else if (m > l)
						{
							frac0 = p;
							frac  = (qu_string) s;
							frac1 = qu_string_newc ('0', m - l);
						}
						else
						{
							frac  = p;
							frac1 = (qu_string) s;
						}
					}
				}
			}
			else if (expo < l)
			{
				/*
					12345678e-6
					-> 12.34567800
					-> 12.345678
					-> 12.3456
				*/
				int m = l - expo;
				inte = qu_string_new (z, m);
				if (prec > 0)
				{
					if (prec > expo)
					{
						frac = qu_string_new (z + m, expo);
						frac1 = qu_string_newc ('0', prec - expo);
					}
					else
					{
						frac = qu_string_new (z + m, prec);
					}
				}
			}
			else
			{
				/*
					12345678e-8
					-> 0.1234567800
					-> 0.12345678
					-> 0.123456
				*/
				inte = qu_string_byte ('0');
				if (f->prec > 0)
				{
					frac = (qu_string) s;
					if (prec > l)
					{
						frac1 = qu_string_newc ('0', prec - l);
					}
					else if (prec < l)
					{
						frac = qu_string_new (z, prec);
					}
				}
			}
		}
	}
	else if (mode == 'e')
	{
		/*
			engineering format with exponent
			always one mantissa digit

			12345678e12
			-> 1.234e19
		*/
		expo += --l;
		inte = qu_string_byte (z[0]);
		if (prec > l)
		{
			frac  = qu_string_new (z + 1, l);
			frac1 = qu_string_newc ('0', prec - l);
		}
		else if (prec < l)
		{
			frac = qu_string_new (z + 1, prec);
		}
		else
		{
			frac = qu_string_new (z + 1, l);
		}
		/* exponent */
		{
			char b [64];
			sprintf (b, f->ucase ? "E%+d" : "e%+d", expo);
			exo = qu_string_newz (b);
		}
	}

	if (f->group > 0 && inte->len > f->group)
	{
		inte = qu_string_interval (inte, (qu_string) qu_string_byte (f->sep), f->group);
	}

	pre = QuStringEMPTY;
#if 0
	switch (mode)
	{
		case 'x':
		{
			static qu_string p = 0;
			if (p == 0)
				p = qu_string_new_staticz ("0x");
			pre = p;
			break;
		}

		case 'b':
		{
			static qu_string p = 0;
			if (p == 0)
				p = qu_string_new_staticz ("0b");
			pre = p;
			break;
		}
		case 'o':
			pre = qu_string_byte ('0');
			break;
	}
#endif

	/* don't print the decimal point if there's no fraction digits */
	if (frac0->len + frac->len + frac1->len == 0)
		dec = QuStringEMPTY;

	if (f->width > 0)
	{
		l = f->width - pre->len - sgn->len - inte->len - dec->len
		  - frac0->len - frac->len - frac1->len - exo->len;

		if ((mode == 'd' || mode == 'x' || mode == 'o' || mode == 'b') && f->zero && l > 0)
		{
			zro = qu_string_newc ('0', l);
			goto end;
		}

		if (l < 0)
		{
			if (mode == 'f')
			{
				mode = 'e';
				goto again;
			}
			return qu_string_newc ('*', f->width);
		}

		if (l > 0)
		{
			qu_string p = qu_string_newc (' ', l);
			if (f->align != 0)
				return qu_string_cat (sgn, pre, inte, dec, frac0, frac, frac1, exo, p, NULL);
			return qu_string_cat (p, sgn, pre, inte, dec, frac0, frac, frac1, exo, NULL);
		}
	}
  end:
	return qu_string_cat (sgn, pre, zro, inte, dec, frac0, frac, frac1, exo, NULL);
}



qu_string qu_format_ptr (qu_format_s *f, qu_any x)
{
	char      buf [128];
	long      l;
	qu_string s;

	sprintf (buf, "%p", x);
	s = qu_string_newz (buf);
	if (f->width > 0)
		l = f->width - strlen (buf);
	else
		l = 0;
	if (l < 0)
	{
		return FormatBadStr;
	}
	else if (l > 0)
	{
		qu_string p = qu_string_newc (' ', l);
		if (p == 0)
			return FormatBadStr;
		s = f->align ? qu_string_cat (s, p, 0) : qu_string_cat (p, s, 0);
		if (s == 0)
			return FormatBadStr;
	}
	return s;
}



qu_string qu_format_yesno (qu_format_s *f, qu_any x)
{
	static qu_string y = 0;
	static qu_string n = 0;

	if (y == 0)
	{
		/* TODO use locale */
		y = qu_string_new_staticz ("yes");
		n = qu_string_new_staticz ("no");
	}
	return qu_format_align (f, (qu_sys_true (x) == QuTrue) ? (qu_string) y : (qu_string) n);
}



qu_string qu_format_string (qu_format_s *f, qu_any x)
{
	qu_string s = qu_string_is (x) ? (qu_string) x : qu_sys_string (x);

	if (f->width > 0)
	{
		long l = f->width - s->len;
		if (l < 0)
		{
			return FormatBadStr;
		}
		else if (l > 0)
		{
			qu_string p = qu_string_newc (' ', l);
			if (p == 0)
				return FormatBadStr;
			s = f->align ? qu_string_cat (s, p, 0) : qu_string_cat (p, s, 0);
			if (s == 0)
				return FormatBadStr;
		}
	}
	return s;
}



qu_string qu_format_align (qu_format_s *f, qu_string s)
{
	if (f->width > 0)
	{
		long l = f->width - s->len;
		if (l < 0)
			return qu_string_newc ('*', f->width);
		if (l > 0)
		{
			qu_string p = qu_string_newc (' ', l);
			QU_ENSURE (p);
			if (f->align != 0)
				return qu_string_cat ((qu_string) s, p, NULL);
			return qu_string_cat ((qu_string) p, s, NULL);
		}
	}
	return s;
}



static qu_null _sys_printf (qu_any self, int argc, qu_any *argv)
{
	char       *i, *p, *f, fmt[64];
	int         n;
	qu_any      v;
	qu_class    c;
	qu_sub      u;
	qu_format_s m;
	volatile qu_string s;

	if (self)
	{
		QU_ARG_VMIN (1);
		v = self;
	}
	else
	{
		QU_ARG_VMIN (2);
		v = argv[0];
		argc--;
		argv++;
	}

	QU_ARG_TYPE (v, qu_string);

	i = ((qu_string) v)->str;
	p = i;

	while (*i)
	{
		if (*i != '%')
		{
			i++;
			continue;
		}

		if (i[1] == '%')
		{
			if (i > p)
				QU_ENSURE (qu_sys_printz (p, i - p));
			QU_ENSURE (qu_sys_printz ("%", 1));
			i += 2;
			p = i;
			continue;
		}

		if (i > p)
			QU_ENSURE (qu_sys_printz (p, i - p));

		QU_ARG_VMIN (1);
		argc--;
		v = *argv++;

		i++;
		n = sizeof (fmt) - 2;
		f = fmt;
		*f++ = '%';

		/* options */
		switch (*i)
		{
			case '#' :
			case '\'':
				/* don't entertain these ridiculous things! */
				goto bad;

			case '<' : /* \__ both  mean */
			case '-' : /* /   left align */
			case '+' :
			case ' ' :
				*f++ = *i++;
				n--;
		}

		/* width */
		while (qu_ascii_isdigit (*i))
		{
			if (n < 3)
				goto bad;
			*f++ = *i++;
			n--;
		}

		/* precision */
		if (*i == '.')
		{
			if (n < 2)
				goto bad;
			*f++ = *i++;
			n--;

			if (*i == '-')
			{
				if (n < 3)
					goto bad;
				*f++ = *i++;
				n--;
			}

			while (qu_ascii_isdigit (*i))
			{
				if (n < 3)
					goto bad;
				*f++ = *i++;
				n--;
			}
		}

		/* group */
		if (*i == ',')
		{
			if (n < 2)
				goto bad;
			*f++ = *i++;
			n--;
			while (qu_ascii_isdigit (*i))
			{
				if (n < 3)
					goto bad;
				*f++ = *i++;
				n--;
			}
		}

		/* mode */
		if (*i == '\0')
			goto bad;
		*f++ = *i++;
		*f = '\0';

		/* do it */
		QU_ENSURE (qu_format_parse (fmt, &m));
		if (f[-1] == 'p' || f[-1] == 'P')
		{
			s = qu_format_ptr (&m, v);
		}
		else if (f[-1] == 's' || f[-1] == 'S')
		{
			if (qu_string_is (v))
				s = qu_string_format (v, &m);
			else
				s = qu_format_string (&m, v);
		}
		else if (f[-1] == 'y' || f[-1] == 'Y')
		{
			s = qu_format_yesno (&m, v);
		}
		else
		{
			switch (qu_classid (v))
			{
				case QU_CLASS_ID_Int:
					s = qu_int_format ((qu_int) v, &m);
					break;

				case QU_CLASS_ID_Long:
					s = qu_long_format ((qu_long) v, &m);
					break;

				case QU_CLASS_ID_Rational:
					s = qu_rational_format ((qu_rational) v, &m);
					break;

				case QU_CLASS_ID_Float:
					s = qu_float_format ((qu_float) v, &m);
					break;

				case QU_CLASS_ID_Big:
					s = qu_big_format ((qu_big) v, &m);
					break;

				case QU_CLASS_ID_Complex:
					s = qu_complex_format ((qu_complex) v, &m);
					break;

				case QU_CLASS_ID_Inf:
					s = qu_inf_format ((qu_inf) v, &m);
					break;

				default:
					c = qu_classof (v);
					s = qu_string_newz (fmt);
					if (c->accel && c->accel->fformat)
					{
						s = c->accel->fformat (v, s);
					}
					else
					{
						__qu_attr_nothrow_begin;
						u = c->get (v, QU_NAME__format, 0);
						__qu_attr_nothrow_end;

						if (u && qu_sub_is (u) && qu_sub_is_inst (u))
						{
							s = qu_sub_call1 (u, v, s);
							if (s == 0 || !qu_string_is (s))
							{
								QU_SYS_WARN_RETVAL (v, "__format", "String");
								s = qu_object_default_string (v);
							}
						}
						else
						{
							s = qu_sys_string (v);
						}
					}
			}
		}

		if (s == 0)
			s = FormatBadStr;

		QU_ENSURE (qu_sys_printz (s->str, s->len));
		p = i;
	}
	if (i > p)
		QU_ENSURE (qu_sys_printz (p, i - p));
	return QuNil;

  bad:
	QU_THROW (QuEFormat);
}



qu_null qu_sys_printf (qu_string format, int argc, ...)
{
	va_list va;
	int     i;
	qu_any  x, argv [QU_SYS_PRINT_MAXARG];

	va_start (va, argc);
	assert (argc <= QU_SYS_PRINT_MAXARG);
	for (i = 0; i < argc; i++)
	{
		x = va_arg (va, qu_any *);
		assert (x != NULL);
		argv [i] = x;
	}
	va_end (va);
	return _sys_printf (format, argc, argv);
}



qu_string qu_sys_sprintf (qu_string format, int argc, ...)
{
	va_list va;
	int     i;
	qu_any     argv [QU_SYS_PRINT_MAXARG], x;

	assert (argc <= QU_SYS_PRINT_MAXARG);
	va_start (va, argc);
	for (i = 0; i < argc; i++)
	{
		x = va_arg (va, qu_any *);
		assert (x != NULL);
		argv [i] = x;
	}
	va_end (va);
	QU_ENSURE (__qu_buffer__);
	if (_sys_printf (format, argc, argv) == 0)
	{
		/*
			this function is called by C code so it should not fail and
			we should never be here, but we'll not make this QU_FATAL
		*/
		__qu_trash__;
		return qu_string_new_staticz ("*FORMAT*");
	}
	return __qu_collect__;
}



/*----------------------------------------------------------------------------
	generic
----------------------------------------------------------------------------*/

qu_any qu_sys_binop (qu_any self, qu_any x, int op)
{
	switch (op)
	{
		case QU_T_are: return qu_sys_are (self, x);

	}
	QU_THROWZ (QuEImplement, "operator not implemented");
}



/* x are y */
static int _sys_are_callback (qu_any item, qu_any idx, qu_any x)
{
	return (qu_sys_is (item, x) == QuTrue) ? QU_ITERATOR_CALLBACK_CONTINUE : QU_ITERATOR_CALLBACK_RESULT;
}

qu_boolean qu_sys_are (qu_any self, qu_any x)
{
	if (qu_object_iterable (self))
	{
		switch (qu_classid (self))
		{
			case QU_CLASS_ID_Range:
				return (x == QuIntClass || x == QuIntegerClass) ? QuTrue : QuFalse;

			case QU_CLASS_ID_String:
				return (x == QuStringClass) ? QuTrue : QuFalse;

			default:
				return (qu_iterator_do (self, (qu_itercallback_f) _sys_are_callback, x) == QU_ITERATOR_CALLBACK_RESULT)
					? QuFalse : QuTrue;
		}
	}
	return qu_object_instof (self, x) ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	operators
----------------------------------------------------------------------------*/

/*
	self && x
*/
qu_any qu_sys_and (qu_any self, qu_any x)
{
	return (qu_sys_true (self) == QuTrue) ? x : self;
}



/*
	self as x
*/
qu_any qu_sys_as (qu_any self, qu_any x)
{
	return qu_isys_as (self, x);
}



/*
	self and x
*/
qu_boolean qu_sys_band (qu_any self, qu_any x)
{
	return (qu_sys_true (self) == QuTrue && qu_sys_true (x) == QuTrue) ? QuTrue : QuFalse;
}



/*
	self or x
*/
qu_boolean qu_sys_bor (qu_any self, qu_any x)
{
	return (qu_sys_true (self) == QuTrue || qu_sys_true (x) == QuTrue) ? QuTrue : QuFalse;
}



/*
	self.__clear
*/
qu_any qu_sys_clear (qu_any self)
{
	return qu_isys_clear (self);
}



/*
	self <=> x  ->  self.__cmp (x)
*/
qu_int qu_sys_cmp (qu_any self, qu_any x)
{
	return qu_isys_cmp (self, x);
}



/*
	self _ x  ->  self.__string.__plus (x.__string)
*/
qu_string qu_sys_concat (qu_any self, qu_any x)
{
	return qu_isys_concat (self, x);
}



/*
	self.__copy
*/
qu_any qu_sys_copy (qu_any self)
{
	return qu_isys_copy (self);
}



/*
	self.__decr
*/
qu_any qu_sys_decr (qu_any self)
{
	return qu_isys_decr (self);
}



/*
	self.__del (x)
*/
qu_int qu_sys_del (qu_any self, qu_any x)
{
	return qu_isys_del (self, x);
}



/*
	self / x  ->  self.__div (x)
*/
qu_any qu_sys_div (qu_any self, qu_any x)
{
	return qu_isys_div (self, x);
}



/*
	self == x  ->  self.__eq (x)
*/
qu_boolean qu_sys_eq (qu_any self, qu_any x)
{
	return qu_isys_eq (self, x);
}



/*
	self.__each
*/
qu_any qu_sys_each (qu_any self)
{
	return qu_isys_each (self);
}



/*
	self %% (*arg)
*/
qu_any qu_sys_format (qu_any self, int argc, qu_any *argv)
{
	QU_ARG_TYPE (self, qu_string);

	QU_ENSURE (__qu_buffer__);
	if (_sys_printf (self, argc, argv) == 0)
	{
		__qu_trash__;
		QU_RETHROW;
	}
	return __qu_collect__;
}


/* dang! */
qu_string qu_sys_frmt (qu_any self, qu_any x)
{
	qu_class c = qu_classof (self);

	if (c->accel && c->accel->fformat)
		return c->accel->fformat (self, x);
	else
	{
		qu_sub    u;
		qu_string s;

		__qu_attr_nothrow_begin;
		u = c->get (self, QU_NAME__format, 0);
		__qu_attr_nothrow_end;

		if (u && qu_sub_is (u) && qu_sub_is_inst (u))
		{
			s = qu_sub_call1 (u, self, x);
			if (s == 0 || !qu_string_is (s))
			{
				QU_SYS_WARN_RETVAL (self, "__format", "String");
				s = qu_object_default_string (self);
			}
		}
		else
		{
			s = qu_sys_string (self);
		}
		return s;
	}
}



/*
	self >= x  ->  self.__cmp (x) >= 0
*/
qu_boolean qu_sys_ge (qu_any self, qu_any x)
{
	return qu_isys_ge (self, x);
}



/*
	self [x]  ->  self.__getitem (x)
*/
qu_any qu_sys_getitem (qu_any self, qu_any x)
{
	return qu_isys_getitem (self, x);
}



/*
	self [from, len]  ->  self.__getslice (from, len)
*/
qu_any qu_sys_getslice (qu_any self, qu_any i, qu_any len)
{
	return qu_isys_getslice (self, i, len);
}



/*
	self > x  ->  self.__cmp (x) > 0
*/
qu_boolean qu_sys_gt (qu_any self, qu_any x)
{
	return qu_isys_gt (self, x);
}



/*
	self & x  ->  self.__iand (x)
*/
qu_any qu_sys_iand (qu_any self, qu_any x)
{
	return qu_isys_iand (self, x);
}



/*
	self \ x  ->  self.__idiv (x)
*/
qu_any qu_sys_idiv (qu_any self, qu_any x)
{
	return qu_isys_idiv (self, x);
}



/*
	x in self  ->  self.__in (x)
*/
qu_boolean qu_sys_in (qu_any self, qu_any x)
{
	return qu_isys_in (self, x);
}



/*
	self.__incr
*/
qu_any qu_sys_incr (qu_any self)
{
	return qu_isys_incr (self);
}



/*
	~self  ->  self.__inot
*/
qu_any qu_sys_inot (qu_any self)
{
	return qu_isys_inot (self);
}



/*
	self | x  ->  self.__ior (x)
*/
qu_any qu_sys_ior (qu_any self, qu_any x)
{
	return qu_isys_ior (self, x);
}



/*
	self is x
*/
qu_boolean qu_sys_is (qu_any self, qu_any x)
{
	return qu_isys_is (self, x);
}



/*
	self ^ x  ->  self.__ixor (x)
*/
qu_any qu_sys_ixor (qu_any self, qu_any x)
{
	return qu_isys_ixor (self, x);
}



/*
	self <= x  ->  self.__cmp (x) <= 0
*/
qu_boolean qu_sys_le (qu_any self, qu_any x)
{
	return qu_isys_le (self, x);
}



/*
	self.__len
*/
qu_int qu_sys_len (qu_any self)
{
	return qu_isys_len (self);
}



/*
	self === x  ->  self.__like (x)
*/
qu_boolean qu_sys_like (qu_any self, qu_any x)
{
	return qu_isys_like (self, x);
}



/*
	self << x  ->  self.__lshift (x)
*/
qu_any qu_sys_lshift (qu_any self, qu_any x)
{
	return qu_isys_lshift (self, x);
}



/*
	self < x  ->  self.__cmp (x) < 0
*/
qu_boolean qu_sys_lt (qu_any self, qu_any x)
{
	return qu_isys_lt (self, x);
}



/*
	self - x  ->  self.__minus (x)
*/
qu_any qu_sys_minus (qu_any self, qu_any x)
{
	return qu_isys_minus (self, x);
}



/*
	self * x  ->  self.__mul (x)
*/
qu_any qu_sys_mul (qu_any self, qu_any x)
{
	return qu_isys_mul (self, x);
}



/*
	self != x  -> not self.__eq (x)
*/
qu_boolean qu_sys_ne (qu_any self, qu_any x)
{
	return qu_isys_ne (self, x);
}



/*
	-self  ->  self.__neg
*/
qu_any qu_sys_neg (qu_any self)
{
	return qu_isys_neg (self);
}



/*
	not self
	! self
*/
qu_boolean qu_sys_not (qu_any self)
{
	return qu_isys_not (self);
}



/*
	self of x
*/
qu_any qu_sys_of (qu_any self, qu_any x)
{
	return qu_isys_of (self, x);
}



/*
	self || x
*/
qu_any qu_sys_or (qu_any self, qu_any x)
{
	return (qu_sys_true (self) == QuTrue) ? self : x;
}



/*
	self + x  -> self.__plus (x)
*/
qu_any qu_sys_plus (qu_any self, qu_any x)
{
	return qu_isys_plus (self, x);
}



/*
	self >>> x ->  self.__pop
*/
qu_any qu_sys_pop (qu_any self)
{
	return qu_isys_pop (self);
}



/*
	+self  ->  self.__pos
*/
qu_any qu_sys_pos (qu_any self)
{
	return qu_isys_pos (self);
}



/*
	self ** x  ->  self.__pow (x)
*/
qu_any qu_sys_pow (qu_any self, qu_any x)
{
	return qu_isys_pow (self, x);
}



/*
	self <<< x  ->  self.__push (x)
*/
qu_any qu_sys_push (qu_any self, qu_any x)
{
	return qu_isys_push (self, x);
}



/*
	self .. x  ->  qu_range.new (self, x)
*/
qu_any qu_sys_range (qu_any self, qu_any x)
{
	return qu_isys_range (self, x);
}



/*
	self _/ x  ->  qu_rational.new (self, x)
*/
qu_any qu_sys_ratio (qu_any self, qu_any x)
{
	return qu_isys_ratio (self, x);
}



/*
	self % x  ->  self.__rem (x)
*/
qu_any qu_sys_rem (qu_any self, qu_any x)
{
	return qu_isys_rem (self, x);
}



/*
	self :: x
*/
qu_any qu_sys_rep (qu_any self, qu_any x)
{
	return x;
}



/*
	self in x  ->  x.__in (self)
*/
qu_boolean qu_sys_rin (qu_any self, qu_any x)
{
	x = qu_isys_in (self, x);
	if (x == QuTrue)
		return QuFalse;
	if (x == QuFalse)
		return QuTrue;
	return x;
}



/*
	self >> x  ->  self.__rshift (x)
*/
qu_any qu_sys_rshift (qu_any self, qu_any x)
{
	return qu_isys_rshift (self, x);
}



/*
	self [x] = v  ->  self.__setitem (x, v)
*/
qu_any qu_sys_setitem (qu_any self, qu_any x, qu_any v)
{
	return qu_isys_setitem (self, x, v);
}



/*
	self [from, len] = v  ->  self.__setslice (from, len, v)
*/
qu_any qu_sys_setslice (qu_any self, qu_any from, qu_any len, qu_any v)
{
	return qu_isys_setslice (self, from, len, v);
}



/*
	self.__string
*/
qu_string qu_sys_string (qu_any self)
{
	return qu_isys_string (self);
}



/*
	self.__true
*/
qu_boolean qu_sys_true (qu_any self)
{
	return qu_isys_true (self);
}



/*
	This function is just an address marker. The compiler does not generate
	code that call this function directly.
*/
qu_any qu_sys_xchg (qu_any self, qu_any x)
{
	QU_ARG_THROW;
}



/*----------------------------------------------------------------------------
	print
----------------------------------------------------------------------------*/

qu_null qu_sys_printh (qu_any x)
{
	volatile qu_string s;
	char b [1024], *i, *p = b, *m = b + sizeof (b) - 8;
	long l;

	s = (qu_string) x;
	if (!qu_string_is (s))
		s = qu_sys_string (s);
	l = s->len;
	i = s->str;
	while (l--)
	{
		if (p > m)
		{
			QU_ENSURE (qu_sys_printz (b, p - b));
			p = b;
		}
		switch (*i)
		{
			case '<': ++i; *p++ = '&'; *p++ = 'l'; *p++ = 't'; *p++ = ';'; continue;
			case '>': ++i; *p++ = '&'; *p++ = 'g'; *p++ = 't'; *p++ = ';'; continue;
			case '&': ++i; *p++ = '&'; *p++ = 'a'; *p++ = 'm'; *p++ = 'p'; *p++ = ';'; continue;
			case '"': ++i; *p++ = '&'; *p++ = 'q'; *p++ = 'u'; *p++ = 'o'; *p++ = 't'; *p++ = ';'; continue;
			default : *p++ = *i++;
		}
	}
	if (p > b)
		QU_ENSURE (qu_sys_printz (b, p - b));
	return QuNil;
}



int qu_sys_print_buff (qu_sys_printb_s *b, const char *z, long n)
{
	char *p = b->buf;
	char *s = b->ptr;
	long  m = QU_SYS_PRINTB_BUFSIZ - (s - p);
	long  w;

	while (n)
	{
		w = n;
		if (w > m)
			w = m;
		memcpy (s, z, w);
		z += w;
		s += w;
		n -= w;
		if (n)
		{
			if (qu_sys_printz (p, s - p) == 0)
				return 0;
			m = QU_SYS_PRINTB_BUFSIZ;
			s = p;
		}
	}
	b->ptr = s;
	return 1;
}



/*) sub print (*arg) -> Null
	Print the arguments by calling their __print methods, then return nil.

	@exc
	.ESystem
		As reported by -{Stream.print}. There are a lot of reasons why a simple
		print may fail. For example, you have closed -{Stream.stdout}, you
		have redirected v{stdout} to a file and the disk is full, and many
		others.

	.EMemory
		This exception is thrown if you are in #{__buffer__} and the buffer becomes
		too big.

	@xmp
		print (1 2 3)
		print (4 5 6)

	@out
		123456
*/
static qu_null qu__sys_print (int argc, qu_any *argv)
{
	__qu_atomic_begin;
	while (argc--)
	{
		if (qu_sys_print (*argv++) == 0)
		{
			__qu_atomic_end;
			goto err;
		}
	}
	__qu_atomic_end;
	return QuNil;

  err:
	if (QuSys__print__->value != QuNil)
	{
		if (qu_sub_call0 (QuSys__print__->value, QuNil))
			return QuNil;
	}
	QU_RETHROW;
}



/*) sub println (*arg) -> Null
	This function simply calls #{print} twice. Once for the arguments,
	and once to print a line-break.

	See #{print} for details and possible exceptions.

	@xmp
		println (1 2 3)
		println (4 5 6)

	@out
		123
		456
*/
static qu_null qu__sys_println (int argc, qu_any *argv)
{
	volatile qu_string br = QuStreamStdout->br;

	__qu_atomic_begin;
	while (argc--)
	{
		if (qu_sys_print (*argv++) == 0)
			goto err;
	}
	if (qu_sys_printz (br->str, br->len) == 0)
		goto err;
	__qu_atomic_end;
	return QuNil;

  err:
	__qu_atomic_end;
	if (QuSys__print__->value != QuNil)
	{
		if (qu_sub_call0 (QuSys__print__->value, QuNil))
			return QuNil;
	}
	QU_RETHROW;
}



/*) sub printc (translate, *arg) -> Null
	Print the arguments by calling their __string methods and use v{translate}
	as a list of character replacements.

	See -{String.translate} for the v{translate} argument.

	See -{print} for possible exceptions.

	@xmp
		rep = [' ':'SPC'  '\t':'TAB'  'f':lambda (x) x _ '-']
		printc (rep, 'foo\tbar and baz')
		println

	@out
		f-ooTABbarSPCandSPCbaz
*/
static qu_null qu__sys_printc (int argc, qu_any *argv)
{
	qu_any    trans;
	long   l;
	char  *i, c;
	qu_null   res = 0;
	volatile qu_string s, x;

	QU_ARG_VMIN (1);
	trans = *argv++;

	__qu_atomic_begin;
	while (--argc)
	{
		s = qu_sys_string (*argv++);
		l = s->len;
		i = s->str;
		while (l--)
		{
			c = *i++;
			x = qu_strz_translate (c, trans);
			if (x)
			{
				if (qu_sys_printz (x->str, x->len) == 0)
					goto done;
			}
			else
			{
				if (qu_sys_printz (&c, 1) == 0)
					goto done;
			}
		}
	}
	res = QuNil;

  done:
	__qu_atomic_end;
	return res;
}



/*) sub printcln (translate, *arg) -> Null
	This function does the same as f{printc} then prints a linebreak.

	See #{printc} for details.
*/
static qu_null qu__sys_printcln (int argc, qu_any *argv)
{
	volatile qu_string br = QuStreamStdout->br;

	if (qu__sys_printc (argc, argv))
	{
		if (qu_sys_printz (br->str, br->len))
			return QuNil;
	}
	if (QuSys__print__->value != QuNil)
	{
		if (qu_sub_call0 (QuSys__print__->value, QuNil))
			return QuNil;
	}
	QU_RETHROW;
}



/*) sub printf (format: String, *arg)
	Print formatted arguments using v{format} as the format specification
	and v{arg} as the arguments.

	The format is the same as C's printf with a few unsupported formats.
	Refer to C{man printf} for now.

	@exc
	.ESystem see #{print}
	.EMemory see #{print}
	.EFormat bad format specification or the arguments don't match

	@xmp
		printf ("%-10s %d", 'foo', 1)

	@out
		foo        1

	@todo
		Explain the format!
*/
static qu_null qu__sys_printf (int argc, qu_any *argv)
{
	qu_any r;

	__qu_atomic_begin;
	r = _sys_printf (0, argc, argv);
	__qu_atomic_end;
	return r;
}



/*) sub printfln (format: String, *arg) -> Null
	This function calls f{printf} then prints a linebreak.

	Refer to #{printf} for details and possible exceptions.
*/
static qu_null qu__sys_printfln (int argc, qu_any *argv)
{
	volatile qu_string br = QuStreamStdout->br;
	qu_any r = 0;

	__qu_atomic_begin;
	if (_sys_printf (0, argc, argv) == 0)
		goto done;
	if (qu_sys_printz (br->str, br->len) == 0)
		goto done;
	r = QuNil;

  done:
	__qu_atomic_end;
	return r;
}



/*) sub prints (format: String, *arg) -> String
	This function does the same as #{printf} except that the result is
	returned as a -{String} instead of printed.

	This is similar with the C language f{sprintf} function.

	@exc
	.ESystem see #{print}
	.EMemory see #{print}
	.EFormat bad format specification or the arguments don't match

	@xmp
		s = prints ("%s %d", 'foo', 1)
		println (s)

	@out
		foo 1
*/
static qu_string qu__sys_prints (int argc, qu_any *argv)
{
	QU_ENSURE (__qu_buffer__);
	if (_sys_printf (0, argc, argv) == 0)
	{
		__qu_trash__;
		QU_RETHROW;
	}
	return __qu_collect__;
}



/*) sub printsln (format: String, *arg) -> String
	This function does the same as f{prints} except that it adds a linebreak.

	Refer to #{prints} for details and possible exceptions.
*/
static qu_string qu__sys_printsln (int argc, qu_any *argv)
{
	volatile qu_string br = QuStreamStdout->br;

	QU_ENSURE (__qu_buffer__);
	if (_sys_printf (0, argc, argv) == 0)
	{
		__qu_trash__;
		QU_RETHROW;
	}
	if (qu_sys_printz (br->str, br->len) == 0)
	{
		__qu_trash__;
		QU_RETHROW;
	}
	return __qu_collect__;
}



/*) sub __buffer__ (*arg) -> Null __public
	@def
	sub __buffer__ (buf: String) -> Null
	sub __buffer__ -> Null
	@p
	Start a new print buffer.

	Each -{Thread} can have their own print buffers. If a buffer is defined,
	the builtin printing functions will print to this buffer instead of
	-{Stream.stdout}.

	Buffers can be nested. To clear the current buffer you use the #{__clear__}
	function. To retrieve what has been printed in the current buffer you use
	the #{__collect__} function.

	__collect__ return the content then switch to the previous buffer, if any,
	-{Stream.stdout} otherwise.

	You can optionally supply your own String buffer but you will still have to
	call __collect__ eventually. Also you must not freeze it otherwise you will
	get nothing.

	This function may be useful when you work with templates.

	@warn
	Make sure you eventually call __collect__ after a __buffer__ or your
	prints will be messed up. You might want to wrap function calls in between
	them in a k{try} block.

	Print buffers are per-thread specific. Be careful not to mix them when your
	application is multi-threaded.

	@exc
	.EMemory the buffer becomes too big

	@xmp
		__buffer__
			print ('foo')
			__buffer__
				print ('bah')
				# oops
				__clear__
				try
					do_something
				catch
					print ('***')
				;;
				print ('bar')
			print (__collect__)
		println (__collect__)

	@out
		foobar
*/
qu_null qu__sys__buffer__ ()
{
	qu_string b = qu_string_buffer_new (512);

	if (__QuThreadCurrent->buffers == 0)
	{
		__QuThreadCurrent->buffers = qu_array_new (8);
		__QuThreadCurrent->buffers->objs[0] = b;
		__QuThreadCurrent->buffers->len = 1;
	}
	else
	{
#define MAXTHREADBUFFERS 80
		if (__QuThreadCurrent->buffers->len > MAXTHREADBUFFERS)
			QU_THROWZ (QuEMemory, "__buffer__ nested too deep");
		qu_array_push (__QuThreadCurrent->buffers, b);
	}
	return QuNil;
}



/*) sub __clear__ -> Null __public
	Clear the content of the current print buffer.

	See #{__buffer__} for more information.
*/
qu_null qu__sys__clear__ ()
{
	qu_array u = __QuThreadCurrent->buffers;

	QU_REQUIREZ (QuENone, u, "__clear__ has no associated __buffer__");
	if (u->len)
	{
		qu_string b = u->objs [u->len - 1];
		assert (qu_string_is (b));
		qu_string_buffer_clear (b);
	}
	return QuNil;
}



/*) sub __collect__ -> String __public
	Get the content of the current print buffer and switch to the previous
	buffer, if any, -{Stream.stdout} otherwise.

	See #{__buffer__} for more information.

	@exc
	.ENone not paired with __buffer__
*/
qu_string qu__sys__collect__ ()
{
	qu_array  u = __QuThreadCurrent->buffers;
	qu_string b;

	QU_REQUIREZ (QuENone, u && u->len, "__collect__ has no associated __buffer__");
	b = u->objs [--u->len];
	assert (qu_string_is (b));
	return qu_string_buffer_return (b);
}



/*) sub __trash__ -> Null __public
	Trash the content of the current print buffer and switch to the previous
	buffer, if any, -{Stream.stdout} otherwise.

	See #{__buffer__} for more information.
*/
qu_null qu__sys__trash__ ()
{
	qu_array u = __QuThreadCurrent->buffers;
	QU_REQUIREZ (QuENone, u, "__trash__ has no associated __buffer__");
	if (u->len)
		--u->len;
	return QuNil;
}



/*----------------------------------------------------------------------------
	error
----------------------------------------------------------------------------*/

/*) sub errno (*arg) -> Integer
	@def
	sub errno -> Integer
	sub errno (value: Integer) -> Integer
	@p
	Get or set the value of v{errno}.

	@exc
	.EArgValue

	@note
	Check out the Os module if you need more than just v{errno}.
*/
static qu_integer qu__sys_errno (int argc, qu_any *argv)
{
	int v;

	__qu_unsafe__;

	if (argc)
	{
		QU_ARG_vint (0, v);
		errno = v;
	}
	return qu_integer_newi (errno);
}



/*) sub __verbose__ (*arg) -> Null
	Prints message to the system f{stdout} stream if the C{-o} option
	is used, does nothing otherwise.

	@noexc
*/
static qu_null qu__sys__verbose__ (int argc, qu_any *argv)
{
	if (QU_APP_VERBOSE)
	{
		while (argc--)
		{
			qu_string x = (qu_string) *argv++;
			if (!qu_string_is (x))
				x = qu_sys_string (x);
			printf ("%s", x->str);
		}
		printf ("\n");
	}
	return QuNil;
}



/*) sub require (condition: Any, *arg) -> Null
	@def
	sub require (condition: Any, e: Exception, m: Any) -> Null
	sub require (condition: Any, m: Any, e: Exception) -> Null
	sub require (condition: Any, e: Exception) -> Null
	sub require (condition: Any, m: Any) -> Null

	@p
	If v{condition} is false then throw, otherwise return nil.

	v{arg} must be an -{Exception} class and a message in any order

	If the exception is not specified then then EValue is assumed.
	If the message is not specified, then the default exception message
	is assumed.

	@warn
	Note that the message may be any object but for this function to work
	properly, it should not be an exception class object.

	This function does not throw an exception by itself even if you pass
	invalid arguments to it. Doing so would defeat its purpose.

	@xmp
		sub foo (x)
			require (x is Int and x > 0)
			...
		;;
*/
static qu_null qu__sys_require (int argc, qu_any *argv)
{
	if (argc && qu_sys_true (argv[0]) != QuTrue)
	{
		qu_any   first = 0, second = 0;
		qu_class klass = 0;
		qu_any   msg   = 0;

		if (argc > 1)
		{
			first = argv[1];
			if (argc > 2)
				second = argv[2];
		}
		if (first)
		{
			if (second == 0)
			{
				if (qu_class_is (first) && qu_exception_class_is (first))
					klass = first;
				else
				{
					klass = QuEValueClass;
					msg = first;
				}
			}
			else
			{
				if (qu_class_is (first) && qu_exception_class_is (first))
				{
					klass = first;
					msg = second;
				}
				else if (qu_class_is (second) && qu_exception_class_is (second))
				{
					klass = second;
					msg = first;
				}
			}
		}

		if (klass == 0)
			klass = QuEValueClass;

		qu_thread_throw (klass, msg);
		return 0;
	}
	return QuNil;
}



/*) sub raise (*arg) -> Null
	@def
	sub raise ()            # rethrow the last exception
	sub raise (String)      # throw ERuntime with this message
	sub raise (Class)       # throw exception class with default message
	sub raise (Class, Any)  # throw exception class with message
	sub raise (nil, Any)    # throw last exception with this message
	sub raise (Exception)   # rethrow this Exception object

	@p
	Throw an exception.

	Unlike the keyword k{throw}, this function does not fail at compile time
	when the argument is not an -{Exception} class.

	@p
	Note that f{raise} does not throw an exception by itself even if you pass
	invalid arguments to it (i.e. the first argument is neither a String,
	an Exception class, nor nil).

	@xmp
		raise
		raise ("foo bar")        # = raise (ERuntime, "foo bar")
		raise (EFoo)
		raise (EFoo, "bar")
		raise (EFoo, [1, 2, 3])

		try
			throw EFoo, "foo"
		;;
		raise (nil, "bar")       # = raise (EFoo, "bar")

		try
			...
		catch
			e = __err__
			...
			raise (e)
		;;
*/
qu_any qu__sys_raise (int argc, qu_any *argv)
{
	qu_any x;

	if (argc == 0)
	{
		qu_thread_rethrow ();
	}
	else if (argc == 1)
	{
		x = argv[0];

		if (qu_exception_class_is (x))
		{
			qu_thread_throw ((qu_class) x, qu_exception_getmsg (x));
		}
		else if (qu_exception_is (x))
		{
			qu_thread_throw (qu_classof (x), qu_exception_getmsg (x));
		}
		else if (__QuThreadCurrent->eclass)
		{
			qu_thread_throw (__QuThreadCurrent->eclass, x);
		}
		else
		{
			qu_thread_throw (QuERuntimeClass, x);
		}
	}
	else if (argc == 2)
	{
		x = argv[0];
		if (x == QuNil)
		{
			x = __QuThreadCurrent->eclass;
		}
		else
		{
			if (!qu_class_is (x))
				x = qu_classof (x);
		}
		if (!qu_exception_class_is (x))
			x = QuERuntimeClass;
		qu_thread_throw ((qu_class) x, argv[1]);
	}
	return 0;
}



/*) sub __abort__ (*arg) __public
	Abort program.

	This function never returns. It calls the system f{abort} function.

	Optionally print a message to -{Stream.stderr} before aborting. The message is
	the result of concatenating v{arg} as strings.

	@xmp
		__abort__
		__abort__ ('foo ' 123 ' bar')
*/
static qu_null qu__sys__abort__ (int argc, qu_any *argv)
{
	qu_string s, w;

	__qu_unsafe__;

	if (argc > 0)
	{
		w = qu_interp_where_string ();
		s = qu_sys_cat (argc, argv);
		qu_stream_printz (QuStreamStderr, "*** ", 4);
		qu_stream_printz (QuStreamStderr, w->str, w->len);
		qu_stream_printz (QuStreamStderr, ": ", 3);
		qu_stream_printz (QuStreamStderr, s->str, s->len);
		qu_stream_printz (QuStreamStderr, "\n", 1);
	}
	abort ();
	return QuNil; /*never*/
}



/*) sub __err__ -> Any __public
	Get the last exception thrown.

	This function returns an instance of the last exception class thrown by
	this thread by calling the class f{new} method supplying the message.

	@warn
	Each call to f{__err__} creates a new exception instance so you might want
	to call it once only and save the result in a variable.

	@noexc

	@xmp
		try
			throw EAGAIN, "try again next year"
		catch
			e = __err__
			try
				throw EModify
			catch
				println (__err__)
			;;
			println (e)
		;;

	@out
		(EModify) cant modify object
		(EAGAIN) try again next year
*/
qu_sub __QuSys__err__;

qu_any qu__sys__err__ ()
{
	qu_class c;
	qu_any   m;
	qu_any   x;

	assert (__QuThreadCurrent);
	c = __QuThreadCurrent->eclass;
	m = __QuThreadCurrent->emsg;
	if (c == 0)
		c = QuExceptionClass;
	if (m == 0)
		m = qu_exception_defmsg (c);
	x = qu__exception_new (c, 1, &m);
	if (x == 0)
	{
		QU_SYS_WARN_CLASS (c, "the 'new' method throws an exception");
		x = qu__exception_new (QuExceptionClass, 1, &m);
		assert (x != 0);
	}
	return x;
}



/*) sub __errmsg__ -> String __public
	Get the last exception thrown as a string.

	@noexc

	@xmp
		try
			throw ENone, "try again next year"
		catch
			println (__errmsg__)
		;;

	@out
		try again next year
*/
qu_any qu__sys__errmsg__ ()
{
	qu_class c;
	qu_any   m;

	assert (__QuThreadCurrent);
	c = __QuThreadCurrent->eclass;
	m = __QuThreadCurrent->emsg;
	if (c == 0)
		c = QuExceptionClass;
	if (m == 0)
		m = qu_exception_defmsg (c);
	return qu_string_newf ("%s", qu_sys_string (m)->str);
}



static qu_string _sys_log (const char *prefix, int argc, qu_any *argv)
{
	qu_string s = QuStringEMPTY;

	if (argc)
	{
		s = qu_sys_cat (argc, argv);
		s = qu_sys_sprintf ((qu_string) qu_string_new_staticz ("%s [%d] in %s: %s\n"), 4,
				qu_string_new_staticz (prefix), qu_int_new (__QuThreadCurrent->tid), qu_interp_where_string (), s);

		if (!qu_interp_unsafe ())
			qu_stream_printz (QuStreamStderr, s->str, s->len);
	}
	return s;
}



/*) sub __warning__ (*arg) -> String __public
	Print a warning message to -{Stream.stderr}.

	The message is the result of concatenating the arguments as strings
	prepended with C{'WARNING'}.

	The message is returned as the result for convenience just in case you
	want to use it for something else.

	@xmp
		__warning__ ("foo is ", 1, " or ", 2)
*/
static qu_string qu__sys__warning__ (int argc, qu_any *argv)
{
	__qu_unsafe__;
	return _sys_log ("WARNING", argc, argv);
}



/*) sub __remark__ (*arg) -> String __public
	Print a remark message to -{Stream.stderr}.

	The message is the result of concatenating the arguments as strings
	prepended with C{'REMARK'}.

	The message is returned as the result for convenience just in case you
	want to use it to do something else.

	@xmp
		__remark__ ("foo is ", 1, " or ", 2)
*/
static qu_string qu__sys__remark__ (int argc, qu_any *argv)
{
	__qu_unsafe__;
	return _sys_log ("REMARK", argc, argv);
}



/*) sub __trace__ (enable = true, filt = nil, show = true) -> Boolean __public
	Whether or not the interpreter should print exceptions to -{Stream.stderr}.

	This is useful for debugging.

	You can optionally specify which exception classes that should or should
	not be displayed in v{filt}. If v{filt} is nil then all classes are
	displayed and v{show} is ignored.

	Otherwise v{filt} must be a Set of exception classes.
	If v{show} is true then show exceptions in v{filt},
	if false then show if exception is not in v{filt}.

	@exc
	.EArgValue

	@xmp
		__trace__ (true, {EStop, EIndex})
		println (__trace__ (false))

	@out
		true

	@note
	You can see the setting by looking at the constants __tracefilter__ and
	__tracehow__.
*/
static qu_boolean qu__sys__trace__ (int argc, qu_any *argv)
{
	int enable = 0;
	int show   = 1;
	qu_any filter = QuNil;
	int old;

	__qu_unsafe__;

	QU_ARG_VCOUNT (0, 3);
	if (argc > 0)
	{
		QU_ARG_vbool (0, enable);
		if (argc > 1)
		{
			filter = argv[1];
			if (argc > 2)
				QU_ARG_vbool (2, show);
		}
	}

	if (filter == QuNil)
	{
		QuSys__tracefilter__->value = QuSetEMPTY;
	}
	else
	{
		QU_ARG_CHECK (qu_set_is (filter));
		QuSys__tracefilter__->value = filter;
		QuSys__tracehow__->value = show ? QuTrue : QuFalse;
	}
	old = QuSysTrace;
	QuSysTrace = enable;
	return old ? QuTrue : QuFalse;
}



/*) sub __exit__ (code: Int) __public
	Sets the exit code then throw EExit.

	The code must be an Int.

	@note
	EExit is not catchable but your #{__quit__} handler will be called before exiting.

	Note that #{eval} turns EExit into the catchable EQuit.

	@xmp
		__exit__ (1)
*/
static qu_any qu__sys__exit__ (int argc, qu_any *argv)
{
	int e = 0;

	__qu_unsafe__;

	QU_ARG_VMAX (1);
	if (argc)
		QU_ARG_vint (0, e);

	__QuExceptionExitCode = e;
	qu_thread__throw (QuEExitClass, qu_int_newi (e), __FUNCTION__);
	return 0;
}



/*----------------------------------------------------------------------------
	info
----------------------------------------------------------------------------*/

/*) sub __main__ -> Module __public
	Return the main program -{Module} object.

	@noexc

	@xmp
		# qu /foo/bar/baz.qu
		println (__main__.__path)

	@out
		/foo/bar/baz
*/
static qu_module qu__sys__main__ ()
{
	__qu_unsafe__;
	return QuModuleMain;
}



/*) sub __thread__ -> Thread __public
	Return this -{Thread} object.

	@noexc

	@xmp
		println (__thread__.id)

	@out
		1024   # or whatever the id is
*/
static qu_thread qu__sys__thread__ ()
{
	return __QuThreadCurrent;
}



/*) sub __where__ -> String __public
	Return a string indicating where in the source code this thread is currently
	running.

	@noexc

	@xmp
		# test.qu
		sub foo
			println (__where__)
		;;
		foo
	@out
		__main__.foo
*/
static qu_string qu__sys__where__ ()
{
	return qu_interp_where_string ();
}



/*) sub classof (x) -> Class
	This is a convenience function that simply returns C{x.__class}, that is,
	the Class object of v{x}.

	@noexc

	@xmp
		a = 10
		println (classof (a))

	@out
		<Class sys.Int>
*/
static qu_class qu__sys_classof (qu_any x)
{
	return qu_classof (x);
}



/*) sub nameof (x) -> String
	This is a convenience function that simply returns C{x.__name}, that is,
	the name of the Class object of v{x}.

	@noexc

	@xmp
		a = 10
		println (nameof (a))

	@out
		Int
*/
static qu_string qu__sys_nameof (qu_any x)
{
	return qu_object_class_name (x);
}



/*) sub anyof (x:Class, *arg) -> Boolean
	Return true if any item of v{arg} is an instance of class v{x} or
	its subclasses, false otherwise.

	@exc
	.EArgCount
	.EArgValue

	@xmp
		println (anyof (Int, 1, 2.2, 'foo'))

	@out
		true
*/
static qu_boolean qu__sys_anyof (int argc, qu_any *argv)
{
	qu_class c;

	QU_ARG_VMIN (2);
	c = (qu_class) *argv++;
	QU_ARG_TYPE (c, qu_class);
	while (--argc)
	{
		qu_any x = *argv++;
		if (qu_object_instof (x, c))
			return QuTrue;
	}
	return QuFalse;
}



/*) sub callable (*arg) -> Boolean
	Whether or not all items in v{arg} are callable objects.

	Callable objects in Qu are: Class, Sub, and Method.

	If called without any arguments then this function returns false,
	thus no exception thrown.

	@noexc

	@xmp
		println (callable (Array, @print))

	@out
		true
*/
static qu_boolean qu__sys_callable (int argc, qu_any *argv)
{
	if (argc == 0)
		return QuFalse;

	while (argc--)
	{
		qu_any x = *argv++;
		if (!qu_object_callable (x))
			return QuFalse;
	}
	return QuTrue;
}



/*) sub iterable (*x) -> Boolean
	Whether or not v{x} are all iterable objects.

	Iterable objects must define the __each method.

	Class objects are not iterable.

	@noexc

	@xmp
		println (iterable (Array, Dict))
		println (iterable ([1, 2], [foo:1, bar:2]))

	@out
		false
		true
*/
static qu_boolean qu__sys_iterable (int argc, qu_any *argv)
{
	if (argc == 0)
		return QuFalse;
	while (argc--)
	{
		if (!qu_object_iterable (*argv++))
			return QuFalse;
	}
	return QuTrue;
}



/*----------------------------------------------------------------------------
	minimal shell

	These are provided so that installers can do minimal shell stuffs
	without having to depend on any module, e.g. the main installer.
----------------------------------------------------------------------------*/

/*) sub sysclear (*arg) -> Null
	Clear the screen then optionally print v{arg} by concatenating the
	items as strings.

	f{sysclear} does not print a linebreak by itself.

	@exc
	.ESecurity
	.ESystem   as reported by -{Stream.stdout} (only if v{arg} is not empty)

	@xmp
		sysclear
		sysclear ('hello')
*/
static qu_null qu__sys_sysclear (int argc, qu_any *argv)
{
	__qu_unsafe__;

	system ("clear");
	while (argc--)
	{
		qu_string x = *argv++;
		x = qu_sys_string (x);
		QU_ENSURE (qu_stream_printz (QuStreamStdout, x->str, x->len));
	}
	return QuNil;
}



/*) sub sysinput (*prompt) -> String
	Get input from the system v{stdin}, optionally print v{prompt}
	before prompting.

	CPU is not given up to other threads.

	The result is right trimmed, thus if only spaces are entered then an empty
	string is returned.

	@warn
	This function only reads up to 2048 characters.

	@exc
	.ESecurity
	.EInterrupt
	.ESystem

	@xmp
		s = sysinput ('foo> ')
*/
static qu_string qu__sys_sysinput (int argc, qu_any *argv)
{
	char s [2048], *i;

	if (QuStreamStdin->fd < 0)
		QU_THROW (QuEBADF);

	while (argc--)
		QU_ENSURE (qu_stream_print (QuStreamStdin, *argv++));

	memset (s, sizeof (s), 0);

	/* must not get into __qu_blocking_begin/end */
	i = fgets (s, sizeof (s) - 1, stdin);
	if (__QuSigINT || i == 0)
	{
		/* this is necessary otherwise we will throw twice */
		__QuSigINT = 0;
		QU_THROW (QuEInterrupt);
	}
	i = s + strlen (s);
	while (i > s && qu_ascii_isspace (i[-1]))
		i--;
	*i = '\0';
	return qu_string_newz (s);
}



/*) sub sysexec (*arg) -> Null
	Execute a shell command.

	The command is the result of concatenating v{arg} as strings.

	@warn
	Be careful, no checks are made!

	@exc
	.ESecurity
	.ESystem

	@xmp
		what = 1
		sysexec ('foo -r ' what ' bar')
		# = sysexec ('foo -r 1 bar')
*/
static qu_any qu__sys_sysexec (int argc, qu_any *argv)
{
	volatile qu_string s;
	int r;

	__qu_unsafe__;

	QU_ARG_VMIN (1);
	QU_ARG_VTYPE (0, s, qu_string);
	s = qu_sys_cat (argc, argv);
	QU_ARG_CHECK (s->len > 0);
	__qu_blocking_begin;
	r = system (s->str);
	__qu_blocking_end;
	if (r != 0)
		QU_THROW (QuESystem);
	return QuNil;
}



/*----------------------------------------------------------------------------
	safe call
----------------------------------------------------------------------------*/

/*) sub syscall (func: Sub, theself: Any, *arg) -> Any
	Call a function.

	By default Qu calls functions by silently passing self to the function.
	Thus if called from methods then the object's internals are exposed since
	self may access anything. Despite its usefulness this may be dangerous.

	If you don't trust the function, use f{syscall} and set self to whatever
	is necessary. Usually nil.

	v{func} must be a static -{Sub}. It can not be a -{Method} because a
	Method is already bound to a self object.

	@exc
	.EArgValue
	.Exception anything thrown by the function

	@xmp
		sub dangerous
			println (self._private)
		;;

		class Foo
			var _private

			sub __init (x)
				self._private = x
			;;

			sub trusted
				dangerous
			;;

			sub untrusted
				try
					syscall (@dangerous, nil)
				catch
					println (__err__)
				;;
			;;
		;;

		x = Foo.new ('my private stuff')
		x.trusted
		x.untrusted

	@out
		my private stuff
		(EImplement) Null._private is not implemented
*/
static qu_any qu__sys_syscall (int argc, qu_any *argv)
{
	qu_sub f;

	QU_ARG_VMIN (2);
	f = argv[0];
	QU_ARG_CHECK (qu_sub_is (f) && qu_sub_is_static (f));
	return qu_sub_calln (f, argv[1], argc - 2, argv + 2);
}



/*----------------------------------------------------------------------------
	miscellaneous
----------------------------------------------------------------------------*/

/*) sub this (retval = nil, *arg) -> Any
	Does absolutely nothing then return v{retval}.

	This is handy for shortcut lovers and probably for some other things
	But that is not the reason.

	The reason is to remind you that I{this} is not a keyword and you should
	not use it in the first place.

	In Qu it's self, not I{this}. A color editor should render them differently.

	@noexc

	@xmp
		x = this (1, y = 2, z = 3)
		return this (10, Foo = 1, Bar = 2)
*/
static qu_any qu__sys_this (int argc, qu_any *argv)
{
	if (argc)
		return argv[0];
	return QuNil;
}



/*) sub sysdelay (expr) -> Any
	Does absolutely nothing then return v{expr}.

	This function is specifically designed to delay initial values of
	variables/properties. It forces the expression to be evaluated at
	run-time rather than at compile-time.

	It may be useful in situations where you are not sure whether an initial
	value is a valid value at compile-time.

	@noexc

	@xmp
		sub _thing_ (x)
			return x in 1 .. 10
		;;

		n = 100
		println ('starting ...')

		var x: thing = sysdelay (n)

		println ('continuing ...')

	@out
		starting ...
		[1024] __main__ (EValue) x type mismatch
*/
static qu_any qu__sys_sysdelay (qu_any x)
{
	return x;
}



/*) sub syscat (*arg) -> String
	Concatenate all arguments as String by calling their __string method as
	necessary.

	When you are concatenating a long list of arguments, sometimes using
	f{syscat} makes your code less prone to error than using the concatenation
	operator '_'.

	@exc
	.EMemory

	@xmp
		s = a _ b c         # oops, forgot the _ between b and c
		s = syscat (a b c)  # cleaner, maybe
*/
qu_sub __QuSys_syscat;

qu_string qu__sys_syscat (int argc, qu_any *argv)
{
	qu_string  s;
	qu_string  x;
	long    n = argc;
	qu_any    *p = argv;
	char   *i;
	long    l = 0;

	while (n--)
	{
		x = *p++;
		if (!qu_stringy_is (x))
			goto slow;
		QU_MEM_CHECK (l < QU_MEM_MAX - x->len);
		l += x->len;
	}
	if (l == 0)
		return QuStringEMPTY;
	s = qu_string_newl (l);
	i = s->str;
	while (argc--)
	{
		x = *argv++;
		memcpy (i, x->str, x->len);
		i += x->len;
	}
	return s;

  slow:
	s = qu_string_buffer_new (l);
	while (argc--)
	{
		x = *argv++;
		if (!qu_stringy_is (x))
			x = (qu_string) qu_sys_string (x);
		QU_ENSURE (qu_string_buffer_puts (s, x));
	}
	return qu_string_buffer_return (s);
}



/*) sub sysjoin (delim = '', *arg) -> String
	Concatenate all arguments as String by calling their __string method as
	necessary.

	v{delim} is inserted between two items.

	@exc
	.EArgValue   v{delim} must be a String
	.EMemory

	@xmp
		s = a _ '\n' _ b '\n' c _ '\n'  # oops, forgot the _ before c
		s = sysjoin ('\n' a b c)        # cleaner, maybe
*/
static qu_string qu__sys_sysjoin (int argc, qu_any *argv)
{
	qu_string  s, x, delim;
	long       m, n = argc;
	qu_any    *p = argv;
	char      *i, *d;
	long       l = 0;
	int        k;

	QU_ARG_VMIN (2);
	QU_ARG_VTYPE (0, delim, qu_string);

	m = delim->len;
	while (--n)
	{
		x = *++p;
		if (!qu_string_is (x))
			goto slow;
		QU_MEM_CHECK (l < QU_MEM_MAX - x->len - m);
		l += x->len + m;
	}
	if (l == 0)
		return QuStringEMPTY;
	s = qu_string_newl (l);
	i = s->str;
	d = delim->str;
	k = 0;
	while (argc--)
	{
		if (k)
		{
			memcpy (i, d, m);
			i += m;
		}
		k = 1;
		x = *argv++;
		memcpy (i, x->str, x->len);
		i += x->len;
	}
	*i = '\0';
	s->len = i - s->str;
	return s;

  slow:
	s = qu_string_buffer_new (l);
	k = 0;
	while (--argc)
	{
		if (k)
		{
			QU_ENSURE (qu_string_buffer_puts (s, delim));
			i += m;
		}
		k = 1;
		x = *++argv;
		if (!qu_string_is (x))
			x = (qu_string) qu_sys_string (x);
		QU_ENSURE (qu_string_buffer_puts (s, x));
	}
	return qu_string_buffer_return (s);
}



/*) sub sysref (name: String, x = me) -> Any
	Make a reference to a variable or function.

	If v{x} is a -{Module} object then v{name} must be an attribute name in
	the module. The attribute must exist and it must be public (unless v{x}
	happens to be k{me}) otherwise ENone is thrown.

	Else v{x} is a -{Class} object. The attribute must be a public class or
	static method/property.

	@exc
	.ESecurity
	.EArgValue
	.ENone

	@xmp
		# module abc
		sub foo (x, y, z)
			println (x, y, z)
		;;

		# main program
		use abc as x
		f = sysref (x, 'foo')
		f(1, 2, 3)

	@out
		123

	@note
	Since 1.10 you can also say, for example, C{@Foo.Bar.baz} to refer to the
	baz property of class Bar in module Foo.
*/
static qu_any qu__sys_sysref (int argc, qu_any *argv)
{
	qu_string name;
	qu_any x, r;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, name, qu_string);
	if (argc > 1)
	{
		x = argv[1];
		QU_ARG_CHECK (qu_module_is (x) || qu_class_is (x));
	}
	else
	{
		x = qu_interp_get_module ();
	}

	if (qu_module_is (x))
	{
		r = qu_module_getattr (x, name);
		QU_REQUIRE (QuENone, r && (qu_class_is (r) || qu_var_is (r) || qu_sub_is (r)));
		if (qu_flag_test (r, QU_SYMBOL_FLAG_PRIVATE))
			QU_REQUIRE (QuENone, x == qu_interp_get_module ());
	}
	else
	{
		r = QuClassClass->get (x, name, 1);
		QU_REQUIRE (QuENone, r && (qu_class_is (r) || qu_var_is (r) || qu_sub_is (r)) && !qu_flag_test (r, QU_SYMBOL_FLAG_INST));
	}
	return r;
}



/*) sub sysfilter (x, filter) -> Boolean
	Filter v{x} through v{filter}.

	I{Filtering} here means to test whether v{x} is qualified for further
	processing through v{filter}.

	If v{filter} is a string then call C{x.(filter)}.

	Else if v{filter} is a #{callable} then call C{filter (x)}.

	Else, v{filter} is an Array of string/callable to be called consecutively passing
	v{x} everytime. If one returns false then stop and return false.

	The result is converted to a Boolean, that is, the __true method is called
	if necessary.

	Any exception, except EStop, is rethrown.

	@xmp
		ap = ["__incr", lambda (x) x < 10]
		println (sysfilter (0, ap))

	@out
		true

	@p
	f{sysfilter} is aware of the Regex module.

	@xmp
		use Regex

		re = Regex (`[ab]`)
		for x in ['foo', 'bar', 'baz']
			println (x) if sysfilter (x, re)
		;;

	@out
		bar
		baz
*/

/* so it doesn't get GC'ed in case Regex is not a resident module */
static qu_var RegexMatchSub;

static qu_sub _sys_get_regmatch (qu_any x)
{
	qu_sub match;

	__qu_attr_nothrow_begin;
	match = qu_classof (x)->get (x, qu_string_newz ("match"), 0);
	__qu_attr_nothrow_end;

	if (match)
	{
		if (!(qu_sub_is (match) && qu_sub_is_inst (match)))
			QU_THROWZ (QuEImplement, "bad Regex object");
	}
	RegexMatchSub->value = match;
	return match;
}

qu_boolean qu__sys_sysfilter (qu_any x, qu_any ap)
{
	qu_any last = QuFalse;
	qu_any r;
	long   i;

	if (qu_array_is (ap))
	{
		for (i = 0; i < ((qu_array) ap)->len; ++i)
		{
			r = ((qu_array) ap)->objs[i];

			if (qu_classid (r) == QU_CLASS_ID_Regex)
			{
				if (RegexMatchSub->value == QuNil)
					QU_ENSURE (_sys_get_regmatch (r));
				r = qu_sub_call1 (RegexMatchSub->value, r, x);
			}
			else
			{
				r = qu_object_call_ex0 (x, r);
			}
			if (r == 0)
			{
				QU_ENSURE (qu_thread_err_is_estop ());
				return last;
			}
			if (qu_sys_true (r) != QuTrue)
				return QuFalse;
			last = QuTrue;
		}
	}
	else
	{
		if (qu_classid (ap) == QU_CLASS_ID_Regex)
		{
			if (RegexMatchSub->value == QuNil)
				QU_ENSURE (_sys_get_regmatch (ap));
			r = qu_sub_call1 (RegexMatchSub->value, ap, x);
		}
		else
		{
			r = qu_object_call_ex0 (x, ap);
		}
		if (r == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			return last;
		}
		if (qu_sys_true (r) != QuTrue)
			return QuFalse;
	}
	return QuTrue;
}



/*) sub sysmap (x, ap) -> Any
	Transform v{x} through v{ap}.

	If v{ap} is a string then call C{x.(ap)}.

	Else if v{ap} is a #{callable} then call C{ap (x)}.

	Else, v{ap} is an Array of string/callable to be called consecutively.

	Any exception, except EStop, is rethrown.

	@xmp
		ap = ['__incr', lambda (x) x * 10]
		println (sysmap (10, ap))

	@out
		110
*/
qu_any qu__sys_sysmap (qu_any x, qu_any ap)
{
	qu_any last = x;

	if (qu_array_is (ap))
	{
		long i;
		for (i = 0; i < ((qu_array) ap)->len; ++i)
		{
			x = qu_object_call_ex0 (x, ((qu_array) ap)->objs[i]);
			if (x == 0)
			{
				QU_ENSURE (qu_thread_err_is_estop ());
				return last;
			}
			last = x;
		}
		return last;
	}
	else
	{
		x = qu_object_call_ex0 (x, ap);
		if (x == 0)
		{
			QU_ENSURE (qu_thread_err_is_estop ());
			return last;
		}
		return x;
	}
}



/*) sub syslen (x) -> Int
	This is a convenience function that calls C{x.__len}.

	@exc
	.EImplement the __len method did not return an Int or __len is not implemented
	.Exception  anything thrown by the __len method

	@xmp
		x = [1 2 3]
		println (syslen (x))

	@out
		3
*/
static qu_int qu__sys_syslen (qu_any x)
{
	return qu_isys_len (x);
}



/*) sub sysmax (*arg) -> Any
	Return the greatest argument.

	@exc
	.EArgCount 1 at least

	@xmp
		x = 1
		y = 2.3
		z = 4
		println (sysmax (x, y, z))

	@out
		4
*/
static qu_any qu__sys_sysmax (int argc, qu_any *argv)
{
	qu_any r, x;

	QU_ARG_VMIN (1);
	r = argv[0];
	while (--argc)
	{
		x = *++argv;
		if (qu_sys_gt (x, r) == QuTrue)
			r = x;
	}
	return r;
}



/*) sub sysmin (*arg) -> Any
	Return the smallest argument.

	@exc
	.EArgCount 1 at least

	@xmp
		x = [1, 2.3, 4]
		println (sysmin (*x))

	@out
		1
*/
static qu_any qu__sys_sysmin (int argc, qu_any *argv)
{
	qu_any r, x;

	QU_ARG_VMIN (1);
	r = argv[0];
	while (--argc)
	{
		x = *++argv;
		if (qu_sys_lt (x, r) == QuTrue)
			r = x;
	}
	return r;
}


/*) sub __package__ -> String __public
	Return the string set with the C{-K} option or an empty String if none is
	specified.

	@noexc
*/
static qu_string qu__sys__package ()
{
	if (QU_APP_USEPACKAGE)
		return qu_string_newz (QU_APP_USEPACKAGE);
	return QuStringEMPTY;
}



/*----------------------------------------------------------------------------
	format
----------------------------------------------------------------------------*/

/*) struct Format
		mode   : String   # mode: 'g', 'f', etc.
		left   : Boolean  # false: align right, otherwise: align left
		sign   : Int      # 0:"+"/"-" 1:"-" 2:" "/"-" -1:dont print sign
		zero   : Boolean  # pad 0 on front
		group  : Int      # number of characters to group, -1 means none
		width  : Int      # total available space
		prec   : Int      # number of decimal digits
		ucase  : Boolean  # use uppercase letters
		sep    : String   # group separator character
		dec    : String   # decimal point character
	;;
	A Record of this Struct is returned by #{sysparseformat}.
	You use this to format an object.
*/
static qu_struct FormatStruct;


/*) sub sysparseformat (s: String) -> Record
	Parse a format string.

	The returned value is a -{Record} of #{Format}.

	@exc
	.EArgValue
	.EFormat

	@xmp
		r = sysparseformat ('%30s')
		println (r)

	@out
		['s', true, 0, false, -1, 30, 0, true, ',', '.']
*/
static qu_record qu__sys_parseformat (qu_string x)
{
	qu_format_s f;

	QU_ARG_TYPE (x, qu_string);
	QU_ENSURE (qu_format_parse (x->str, &f));
	/*begin*/
	{
		qu_any arg[10] =
		{
			qu_string_byte (f.mode),
			f.align ? QuFalse : QuTrue,
			qu_int_new (f.sign),
			f.zero ? QuTrue : QuFalse,
			qu_int_new (f.group),
			qu_int_new (f.width),
			qu_int_new (f.prec),
			f.ucase ? QuTrue : QuFalse,
			qu_string_byte (f.sep),
			qu_string_byte (f.dec)
		};
		return qu_record_new (FormatStruct, 10, arg);
	}
}



/*----------------------------------------------------------------------------
	context
----------------------------------------------------------------------------*/

/*) sub __context__ (*arg) __public
	Currently this simply calls #{Number.context} (*arg).
*/
static qu_any qu__sys__context__ (int argc, qu_any *argv)
{
	return qu__number_context (QuNumberClass, argc, argv);
}



/*----------------------------------------------------------------------------
	numeric
----------------------------------------------------------------------------*/

/*) sub __complex__ (r:Real, i:Real) -> Number __public
	Call this function when returning a Complex.
	It allows you to be consistent with the global -{Number.context}.

	This is a convenience function that simply calls -{Complex.ret} (r, i).

	If the context specifies REALIFY_COMPLEX and v{i} is 0 then this function
	returns v{r}. Otherwise if the context specifies ALLOW_COMPLEX then a Complex
	is returned, EComplex thrown otherwise.

	@xmp
		sub foo
			return __complex__ (1, 0)
			...
		;;
*/
static qu_number qu__sys__complex__ (qu_any r, qu_any i)
{
	return qu_complex_return (r, i);
}



/*) sub __inf__ (x) -> Inf __public
	Call this function when returning an infinity.
	It allows you to be consistent with the global -{Number.context}.

	This is a convenience function that simply calls -{Inf.ret} (x).

	If the -{Number.context} does not specify ALLOW_INF then this function
	throws EInf, returns an infinity based on v{x} otherwise (see -{Inf.ret}
	for what v{x} should be).

	@xmp
		sub foo
			return __inf__ (infp) if some_condition
			...
		;;
*/
static qu_inf qu__sys__inf__ (qu_any x)
{
	return qu_inf_return (x);
}



/*) sub __infp__ (x) -> Inf __public
	Call this function when returning an infinity.
	It allows you to be consistent with the global -{Number.context}.

	This is a convenience function that simply calls -{Inf.ret} (infp).

	If the context does not specify ALLOW_INF then this function throws EInf,
	returns infp otherwise.

	@xmp
		sub foo
			return __infp__ if some_condition
			...
		;;
*/
static qu_inf qu__sys__infp__ ()
{
	return qu_inf_return (QuInfPos);
}



/*) sub __infn__ (x) -> Inf __public
	Call this function when returning an infinity.
	It allows you to be consistent with the global -{Number.context}.

	This is a convenience function that simply calls -{Inf.ret} (infn).

	If the context does not specify ALLOW_INF then this function throws EInf,
	return infn otherwise.

	@xmp
		sub foo
			return __infn__ if some_condition
			...
		;;
*/
static qu_inf qu__sys__infn__ ()
{
	return qu_inf_return (QuInfNeg);
}



/*) sub __huge__ (x) -> Inf __public
	Call this function when returning a huge number that can not be represented
	by neither a Big nor Long.
	It allows you to be consistent with the global -{Number.context}.

	If the context does not specify ALLOW_HUGE then this function throws EHuge,
	returns an infinity based on v{x} otherwise.

	If v{x} is an infinity then v{x} is returned. Otherwise infn is returned
	if C{x < 0}, infp otherwise.

	@xmp
		sub foo
			return __huge__ (infp) if some_condition
			...
		;;
*/
static qu_inf qu__sys__huge__ (qu_any x)
{
	QU_NUMBER_ENSURE_ALLOW_HUGE;
	if (qu_inf_is (x))
		return (qu_inf) x;
	return (qu_int_tol (qu_sys_cmp (x, QuInt0)) < 0) ? QuInfNeg : QuInfPos;
}



/*) sub __tiny__ (x: Real = 0) -> Real __public
	Call this function when returning a tiny number that can not be represented
	by a Big.
	It allows you to be consistent with the global -{Number.context}.

	If the context does not specify ALLOW_TINY then this function throws ETiny,
	returns zero based on v{x}

	@warn
	v{x} must be a zero otherwise this function throws EValue and will be very
	confusing.

	@xmp
		sub foo
			return __tiny__ (0) if some_condition
			...
		;;
*/
static qu_real qu__sys__tiny (int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	QU_NUMBER_ENSURE_ALLOW_TINY;
	if (argc == 0)
	{
		return (qu_real) QuInt0;
	}
	else
	{
		qu_real x = argv[0];
		QU_ARG_CHECK (qu_real_is (x) && qu_real_iszero (x));
		return x;
	}
}



/*) sub __zerodiv__ (x) -> Inf __public
	Call this function when returning a division by zero result.
	It allows you to be consistent with the global -{Number.context}.

	If the context does not specify ALLOW_ZERODIV then this function throws EDiv,
	returns an infinity based on v{x} otherwise.

	If v{x} is an infinity then v{x} is returned. Otherwise infn is returned
	if C{x < 0}, infp otherwise.

	@xmp
		sub foo (x, y)
			return __zerodiv__ (x) if some_condition
			...
		;;
*/
static qu_inf qu__sys__zerodiv__ (qu_any x)
{
	QU_NUMBER_ENSURE_ALLOW_ZERODIV;
	if (qu_inf_is (x))
		return (qu_inf) x;
	return (qu_int_tol (qu_sys_cmp (x, QuInt0)) < 0) ? QuInfNeg : QuInfPos;
}



/*----------------------------------------------------------------------------
	stream
----------------------------------------------------------------------------*/

/*) sub __stderr__ -> Stream
	This is just a convenience method that returns -{Stream.__stderr__}.

	@noexc
*/
static qu_stream qu__sys__stderr__ ()
{
	return QuStreamStderr;
}



/*) sub __stdin__ -> Stream
	This is just a convenience method that returns -{Stream.__stdin__}.

	@noexc
*/
static qu_stream qu__sys__stdin__ ()
{
	return QuStreamStdin;
}



/*) sub __stdout__ -> Stream
	This is just a convenience method that returns -{Stream.__stdout__}.

	@noexc
*/
static qu_stream qu__sys__stdout__ ()
{
	return QuStreamStdout;
}



/*-----------------------------------------------------------------------------
	C type validators --- the things I do for C ...
----------------------------------------------------------------------------*/

qu_sub QuType_char;
qu_sub QuType_byte;
qu_sub QuType_ubyte;
qu_sub QuType_short;
qu_sub QuType_ushort;
qu_sub QuType_int;
qu_sub QuType_uint;
qu_sub QuType_int16;
qu_sub QuType_uint16;
qu_sub QuType_int32;
qu_sub QuType_uint32;
qu_sub QuType_long;
qu_sub QuType_ulong;
qu_sub QuType_longlong;
qu_sub QuType_ulonglong;
qu_sub QuType_float;
qu_sub QuType_double;
qu_sub QuType_longdouble;
qu_sub QuType_size_t;
qu_sub QuType_ssize_t;
qu_sub QuType_off_t;
qu_sub QuType_char_p;
qu_sub QuType_void_p;
#ifdef QU_SUPPORT_GLIB
qu_sub QuType_GObject;
#endif


/*) sub Cchar (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{String} and
	contains a single character (including the nil character).
*/
static qu_boolean qu_type_char (qu_any x)
{
	return (qu_string_is (x) && ((qu_string) x)->len == 1) ? QuTrue : QuFalse;
}



/*) sub Cbyte (x) -> Boolean __public
	This is a validator to ensure that v{x} is a signed byte -{Integer}.
*/
static qu_boolean qu_type_byte (qu_any x)
{
	qu_integer_tob (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cubyte (x) -> Boolean __public
	This is a validator to ensure that v{x} is an unsigned byte -{Integer}.
*/
static qu_boolean qu_type_ubyte (qu_any x)
{
	qu_integer_toub (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cshort (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{short} range.
*/
static qu_boolean qu_type_short (qu_any x)
{
	qu_integer_toh (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cushort (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{unsigned short} range.
*/
static qu_boolean qu_type_ushort (qu_any x)
{
	qu_integer_touh (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cint (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{int} range.
*/
static qu_boolean qu_type_int (qu_any x)
{
	qu_integer_toi (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cuint (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{unsigned int} range.
*/
static qu_boolean qu_type_uint (qu_any x)
{
	qu_integer_toui (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cint16 (x) -> Boolean __public
	This is a validator to ensure that v{x} is a signed 16 bits -{Integer}.
*/
static qu_boolean qu_type_int16 (qu_any x)
{
	qu_integer_toi16 (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cuint16 (x) -> Boolean __public
	This is a validator to ensure that v{x} is an unsigned 16 bits -{Integer}.
*/
static qu_boolean qu_type_uint16 (qu_any x)
{
	qu_integer_toui16 (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cint32 (x) -> Boolean __public
	This is a validator to ensure that v{x} is a signed 32 bits -{Integer}.
*/
static qu_boolean qu_type_int32 (qu_any x)
{
	qu_integer_toi32 (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cuint32 (x) -> Boolean __public
	This is a validator to ensure that v{x} is an unsigned 32 bits -{Integer}.
*/
static qu_boolean qu_type_uint32 (qu_any x)
{
	qu_integer_toui32 (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Clong (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{long} range.
*/
static qu_boolean qu_type_long (qu_any x)
{
	qu_integer_tol (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Culong (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{unsigned long} range.
*/
static qu_boolean qu_type_ulong (qu_any x)
{
	qu_integer_toul (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Clonglong (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{long long} range.

	@warn
	This is only available if your system does have the C long long type.
*/
static qu_boolean qu_type_longlong (qu_any x)
{
#ifdef HAVE_LONG_LONG
	qu_integer_toll (x);
#else
	qu_integer_tol (x);
#endif
	return errno ? QuFalse : QuTrue;
}



/*) sub Culonglong (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{unsigned long long} range.

	@warn
	This is only available if your system does have the C long long type.
*/
static qu_boolean qu_type_ulonglong (qu_any x)
{
#ifdef HAVE_LONG_LONG
	qu_integer_toull (x);
#else
	qu_integer_toul (x);
#endif
	return errno ? QuFalse : QuTrue;
}



/*) sub Cfloat (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Real} in the
	C I{float} range.
*/
static qu_boolean qu_type_float (qu_any x)
{
	qu_number_tof (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Cdouble (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Real} in the
	C I{double} range.
*/
static qu_boolean qu_type_double (qu_any x)
{
	qu_number_tod (x);
	return errno ? QuFalse : QuTrue;
}



/*) sub Clongdouble (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{long double} range.

	@warn
	This is only available if your system does have the C long double type.
*/
static qu_boolean qu_type_longdouble (qu_any x)
{
#ifdef HAVE_LONG_DOUBLE
	qu_number_told (x);
#else
	qu_number_tod (x);
#endif
	return errno ? QuFalse : QuTrue;
}



/*) sub Cssize_t (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{ssize_t} range.
*/
static qu_boolean qu_type_ssize_t (qu_any x)
{
#if SIZEOF_SIZE_T == SIZEOF_INT
	qu_integer_toi (x);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
	qu_integer_tol (x);
#else
	qu_integer_toll (x);
#endif
	return errno ? QuFalse : QuTrue;
}



/*) sub Csize_t (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{size_t} range.
*/
static qu_boolean qu_type_size_t (qu_any x)
{
#if SIZEOF_SIZE_T == SIZEOF_INT
	qu_integer_toui (x);
#elif SIZEOF_SIZE_T == SIZEOF_LONG
	qu_integer_toul (x);
#else
	qu_integer_toull (x);
#endif
	return errno ? QuFalse : QuTrue;
}



/*) sub Coff_t (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{Integer} in the
	C I{off_t} range.
*/
static qu_boolean qu_type_off_t (qu_any x)
{
#if SIZEOF_OFF_T == SIZEOF_INT
	qu_integer_toi (x);
#elif SIZEOF_OFF_T == SIZEOF_LONG
	qu_integer_tol (x);
#else
	qu_integer_toll (x);
#endif
	return errno ? QuFalse : QuTrue;
}



/*) sub Cstring (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{String} in the.

	It is used to differentiate a C C{char *} from a Qu String in structs.
*/
static qu_boolean qu_type_char_p (qu_any x)
{
	return qu_string_is (x) ? QuTrue : QuFalse;
}



/*) sub Cpointer (x) -> Boolean __public
	This is a validator to ensure that v{x} is a -{String} in the.

	It is used to differentiate a C {void *} from a Qu Pointer in structs.
*/
static qu_boolean qu_type_void_p (qu_any x)
{
	return qu_pointer_is (x) ? QuTrue : QuFalse;
}



#ifdef QU_SUPPORT_GLIB
/*) sub GObject (x) -> Any __public
	This validator is used to support the Gtk module.
*/
static qu_boolean qu_type_GObject (qu_any x)
{
	return qu_gobject_is (x) ? QuTrue : QuTrue;
}
#endif



/*----------------------------------------------------------------------------
	eval
----------------------------------------------------------------------------*/

/*) sub eval (code: String, options = 'e', imports = nil, path = nil) -> Any
	Evaluate code.

	v{code} is the source code or in case of v{options} is C{f} then the
	path of the source code file.

	If v{path} is a String then set the module's path.

	v{imports} is either nil or a -{Dict} to transfer symbol values
	back and forth to the k{eval} module. The items in Dict will become
	private variables to the k{eval} module (keys are the names of the
	variables, values are the variable values).

	v{options} is a combinaton of:
	@ops
	.| C{e} | compile as expression, return the result
	.| C{m} | compile as Module, initialize and return the Module object
	.| C{f} | same as v{m} but v{code} is the path to the file to load
	.| C{d} | compile for documentation, do not initialize and return the Module object

	@p
	Possibly combined with one of:
	@ops
	.| C{t} | this is a trusted source, allow everything
	.| C{i} | read back values into v{imports}

	@warn
	The C{m} option does not call the C{__init} function. You have to do it
	yourself as necessary.

	Specifying an untrusted module does not mean that you are safe.
	The module will be denied calls to f{eval} and __unsafe functions but
	it can still abuse resources by doing, for example, infinite loops, eat
	memory by constructing huge objects, etc.

	@note
	f{eval} turn uncacthable exceptions into:
	@ul
	.EAssert into EProgram
	.EExit   into EQuit

	@exc
	.EArgValue

	@xmp
		sub foo (x)
			println (x)
		;;

		code =
		`
			foo (bar)
			bar = 3
		`

		stab = [foo:@foo, bar:'hello']
		eval (code, 'ei', stab)
		println (stab['bar'])

	@out
		hello
		3
*/
static qu_any qu__sys_eval (int argc, qu_any *argv)
{
	qu_string src, path = 0, sop = qu_string_byte ('e');
	qu_any    imp = 0, res;
	long      opt = 0;
	int       isfile = 0;
	char     *z;

	__qu_unsafe__;

	QU_ARG_VCOUNT (1, 4);
	QU_ARG_VTYPE (0, src, qu_string);
	if (argc > 1)
	{
		QU_ARG_VTYPE (1, sop, qu_string);
		if (argc > 2)
		{
			imp = argv[2];
			if (imp == QuNil)
				imp = 0;
			else
				QU_ARG_CHECK (qu_dict_is (imp));

			if (argc > 3)
				QU_ARG_VTYPE (3, path, qu_string);
		}
	}

	z = sop->str;
	switch (z[0])
	{
		case 'f': case 'F': isfile = 1;
		case 'm': case 'M': opt = QU_INTERP_EVAL_MODULE; break;
		case 'e': case 'E': opt = QU_INTERP_EVAL_EXPR;   break;
		case 'd': case 'D': opt = QU_INTERP_EVAL_DOC;    break;
		default : QU_ARG_THROW;
	}
	while (*++z)
	{
		switch (*z)
		{
			case 't' : case 'T': opt |= QU_INTERP_EVAL_TRUSTED;   break;
			case 'i' : case 'I': opt |= QU_INTERP_EVAL_GETIMPORT; break;
			default: QU_ARG_THROW;
		}
	}

	if (isfile)
	{
		path = src;
		src = qu_file_load (path);
		QU_ENSURE (src);
	}

	res = qu_interp_eval (qu_interp_get_module (), src, imp, opt, path);
	if (res == 0)
	{
		if (__QuThreadCurrent->eclass == QuEAssertClass)
		{
			__QuThreadCurrent->eclass = QuEProgramClass;
		}
		else if (__QuThreadCurrent->eclass == QuEExitClass)
		{
			__QuThreadCurrent->eclass = QuEQuitClass;
		}
	}
	return res;
}




/*----------------------------------------------------------------------------
	sort
----------------------------------------------------------------------------*/

#define SORT_FIXED_TYPE_WITH_FUNCTION(name, type) \
static int _sys_osort_##name (qu_sys_osort_f fcmp, type *a, long i, long j, \
                              qu_any obj, qu_any self) \
{ \
	type  p = a[i]; \
	long  l = i; \
	long  r = j; \
	int   cmp; \
	\
	while (i < j) \
	{ \
		while (i < j) \
		{ \
			cmp = fcmp ((void *) a + j * sizeof (type), &p, obj, self); \
			if (cmp == QU_SYS_SORT_ERROR) \
				return 0; \
			if (cmp == QU_SYS_SORT_SMALLER) \
				break; \
			--j; \
		} \
		if (i != j) \
		{ \
			a[i] = a[j]; \
			++i; \
		} \
		while (i < j) \
		{ \
			cmp = fcmp ((void *) a + i * sizeof (type), &p, obj, self); \
			if (cmp == QU_SYS_SORT_ERROR) \
				return 0; \
			if (cmp > 0) \
				break; \
			++i; \
		} \
		if (i != j) \
		{ \
			a[j] = a[i]; \
			--j; \
		} \
	} \
	a[i] = p; \
	if (l < i) \
	{ \
		if (_sys_osort_##name (fcmp, a, l, i - 1, obj, self) == 0) \
			return 0; \
	} \
	if (r > i) \
	{ \
		if (_sys_osort_##name (fcmp, a, i + 1, r, obj, self) == 0) \
			return 0; \
	} \
	return 1; \
}


#define SORT_FIXED_TYPE_WITHOUT_FUNCTION(name, type) \
static int _sys_sort_##name (qu_sys_sort_f fcmp, type *a, long i, long j) \
{ \
	type  p = a[i]; \
	long  l = i; \
	long  r = j; \
	int   cmp; \
	\
	while (i < j) \
	{ \
		while (i < j) \
		{ \
			cmp = fcmp ((void *) a + j * sizeof (type), &p); \
			if (cmp == QU_SYS_SORT_ERROR) \
				return 0; \
			if (cmp == QU_SYS_SORT_SMALLER) \
				break; \
			--j; \
		} \
		if (i != j) \
		{ \
			a[i] = a[j]; \
			++i; \
		} \
		while (i < j) \
		{ \
			cmp = fcmp ((void *) a + i * sizeof (type), &p); \
			if (cmp == QU_SYS_SORT_ERROR) \
				return 0; \
			if (cmp > 0) \
				break; \
			++i; \
		} \
		if (i != j) \
		{ \
			a[j] = a[i]; \
			--j; \
		} \
	} \
	a[i] = p; \
	if (l < i) \
	{ \
		if (_sys_sort_##name (fcmp, a, l, i - 1) == 0) \
			return 0; \
	} \
	if (r > i) \
	{ \
		if (_sys_sort_##name (fcmp, a, i + 1, r) == 0) \
			return 0; \
	} \
	return 1; \
}


SORT_FIXED_TYPE_WITH_FUNCTION (chr, char)
SORT_FIXED_TYPE_WITH_FUNCTION (ptr, void *)
SORT_FIXED_TYPE_WITH_FUNCTION (dbl, double)
SORT_FIXED_TYPE_WITH_FUNCTION (itg, int)

SORT_FIXED_TYPE_WITHOUT_FUNCTION (chr, char)
SORT_FIXED_TYPE_WITHOUT_FUNCTION (ptr, void *)
SORT_FIXED_TYPE_WITHOUT_FUNCTION (dbl, double)
SORT_FIXED_TYPE_WITHOUT_FUNCTION (itg, int)


static int _sys_osort_generic (qu_sys_osort_f fcmp, char *data, long size, long i, long j,
                               qu_any obj, qu_any self)
{

#define a_(i) (data + i * size)

	char p [QU_SYS_SORT_MAXSIZE];
	long l = i;
	long r = j;
	int  cmp;

	if (size == 1)
		p[0] = data[i];
	else
		memcpy (p, a_(i), size);

	while (i < j)
	{
		while (i < j)
		{
			cmp = fcmp (a_(j), p, obj, self);
			if (cmp == QU_SYS_SORT_ERROR)
				return 0;
			if (cmp == QU_SYS_SORT_SMALLER)
				break;
			--j;
		}
		if (i != j)
		{
			memcpy (a_(i), a_(j), size);
			++i;
		}
		while (i < j)
		{
			cmp = fcmp (a_(i), p, obj, self);
			if (cmp == QU_SYS_SORT_ERROR)
				return 0;
			if (cmp > 0)
				break;
			++i;
		}
		if (i != j)
		{
			memcpy (a_(j), a_(i), size);
			--j;
		}
	}
	memcpy (a_(i), p, size);
	if (l < i)
	{
		if (_sys_osort_generic (fcmp, data, size, l, i - 1, obj, self) == 0)
			return 0;
	}
	if (r > i)
	{
		if (_sys_osort_generic (fcmp, data, size, i + 1, r, obj, self) == 0)
			return 0;
	}
	return 1;

#undef a_
}



static int _sys_sort_generic (qu_sys_sort_f fcmp, char *data, long size, long i, long j)
{

#define a_(i) (data + i * size)

	char p [QU_SYS_SORT_MAXSIZE];
	long l = i;
	long r = j;
	int  cmp;

	if (size == 1)
		p[0] = data[i];
	else
		memcpy (p, a_(i), size);

	while (i < j)
	{
		while (i < j)
		{
			cmp = fcmp (a_(j), p);
			if (cmp == QU_SYS_SORT_ERROR)
				return 0;
			if (cmp == QU_SYS_SORT_SMALLER)
				break;
			--j;
		}
		if (i != j)
		{
			memcpy (a_(i), a_(j), size);
			++i;
		}
		while (i < j)
		{
			cmp = fcmp (a_(i), p);
			if (cmp == QU_SYS_SORT_ERROR)
				return 0;
			if (cmp > 0)
				break;
			++i;
		}
		if (i != j)
		{
			memcpy (a_(j), a_(i), size);
			--j;
		}
	}
	memcpy (a_(i), p, size);
	if (l < i)
	{
		if (_sys_sort_generic (fcmp, data, size, l, i - 1) == 0)
			return 0;
	}
	if (r > i)
	{
		if (_sys_sort_generic (fcmp, data, size, i + 1, r) == 0)
			return 0;
	}
	return 1;

#undef a_
}



int qu_sys_osort (qu_sys_osort_f fcmp, void *data, long size, long count, qu_any obj, qu_any self)
{
	if (size > QU_SYS_SORT_MAXSIZE)
		QU_FATAL ("size > QU_SYS_SORT_MAXSIZE");

	if (size == 1)
		return _sys_osort_chr (fcmp, data, 0, count - 1, obj, self);
	else if (size == sizeof (void *))
		return _sys_osort_ptr (fcmp, data, 0, count - 1, obj, self);
	else if (size == sizeof (double))
		return _sys_osort_dbl (fcmp, data, 0, count - 1, obj, self);
	else if (size == sizeof (int))
		return _sys_osort_itg (fcmp, data, 0, count - 1, obj, self);
	else
		return _sys_osort_generic (fcmp, data, size, 0, count - 1, obj, self);
}



int qu_sys_sort (qu_sys_sort_f fcmp, void *data, long size, long count)
{
	if (size > QU_SYS_SORT_MAXSIZE)
		QU_FATAL ("size > QU_SYS_SORT_MAXSIZE");

	if (size == 1)
		return _sys_sort_chr (fcmp, data, 0, count - 1);
	else if (size == sizeof (void *))
		return _sys_sort_ptr (fcmp, data, 0, count - 1);
	else if (size == sizeof (double))
		return _sys_sort_dbl (fcmp, data, 0, count - 1);
	else if (size == sizeof (int))
		return _sys_sort_itg (fcmp, data, 0, count - 1);
	else
		return _sys_sort_generic (fcmp, data, size, 0, count - 1);
}



/*----------------------------------------------------------------------------
	operators
----------------------------------------------------------------------------*/

qu_binop_f qu_sys_binop_get (qu_string op)
{
	qu_int i = qu_dict_getitem (QuTokenizerByName, op);

	switch (qu_int_tol (i))
	{
		case QU_T_are:      return (qu_binop_f) qu_sys_are;

		case QU_T_as:       return (qu_binop_f) qu_sys_as;
		case QU_T_AND:      return (qu_binop_f) qu_sys_and;
		case QU_T_and:      return (qu_binop_f) qu_sys_band;
		case QU_T_or:       return (qu_binop_f) qu_sys_bor;
		case QU_T_CMP:      return (qu_binop_f) qu_sys_cmp;
		case QU_T_CONCAT:   return (qu_binop_f) qu_sys_concat;
		case QU_T_DIV:      return (qu_binop_f) qu_sys_div;
		case QU_T_EQ:       return (qu_binop_f) qu_sys_eq;
		case QU_T_FORMAT:   return (qu_binop_f) qu_sys_frmt;
		case QU_T_GE:       return (qu_binop_f) qu_sys_ge;
		case QU_T_GT:       return (qu_binop_f) qu_sys_gt;
		case QU_T_BITAND:   return (qu_binop_f) qu_sys_iand;
		case QU_T_IDIV:     return (qu_binop_f) qu_sys_idiv;
		case QU_T_in:       return (qu_binop_f) qu_sys_in;
		case QU_T_BITOR:    return (qu_binop_f) qu_sys_ior;
		case QU_T_is:       return (qu_binop_f) qu_sys_is;
		case QU_T_BITXOR:   return (qu_binop_f) qu_sys_ixor;
		case QU_T_LE:       return (qu_binop_f) qu_sys_le;
		case QU_T_LIKE:     return (qu_binop_f) qu_sys_like;
		case QU_T_LSHIFT:   return (qu_binop_f) qu_sys_lshift;
		case QU_T_LT:       return (qu_binop_f) qu_sys_lt;
		case QU_T_MINUS:    return (qu_binop_f) qu_sys_minus;
		case QU_T_MUL:      return (qu_binop_f) qu_sys_mul;
		case QU_T_NE:       return (qu_binop_f) qu_sys_ne;
		case QU_T_of:       return (qu_binop_f) qu_sys_of;
		case QU_T_OR:       return (qu_binop_f) qu_sys_or;
		case QU_T_PLUS:     return (qu_binop_f) qu_sys_plus;
		case QU_T_POW:      return (qu_binop_f) qu_sys_pow;
		case QU_T_PUSH:     return (qu_binop_f) qu_sys_push;
		case QU_T_DOT2:     return (qu_binop_f) qu_sys_range;
		case QU_T_RATIO:    return (qu_binop_f) qu_sys_ratio;
		case QU_T_REM:      return (qu_binop_f) qu_sys_rem;
		case QU_T_COLON2:   return (qu_binop_f) qu_sys_rep;
		case QU_T_RSHIFT:   return (qu_binop_f) qu_sys_rshift;
	}
	return NULL;
}




/*----------------------------------------------------------------------------
	others
----------------------------------------------------------------------------*/

/*HACK - documentation*/
qu_sub __QuSys__undocumented__;


/*) sub __undocumented__ __public
	This is to solve documentation problem when a symbol is only defined
	at run-time. For example, a module that register new symbols in the
	Sys module.

	This function does absolute nothing then return nil.
*/
static qu_null sys__undocumented (int argc, qu_any *argv)
{
	return QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_sys__init ()
{
	FormatBadStr = qu_string_new_staticz ("***");

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		handlers
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuSys__print__    = qu_module_add_var (QuSysModule, "__print__"   , QuNil);
	QuSys__keyboard__ = qu_module_add_var (QuSysModule, "__keyboard__", QuNil);
	QuSys__quit__     = qu_module_add_var (QuSysModule, "__quit__"    , QuNil);

	QuSys__print__   ->type =
	QuSys__keyboard__->type =
	QuSys__quit__    ->type =

		qu_module_add_sub (QuSysModule, "__handlers__", _sys_ensure_handler, 1,
			QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_ENSURE | QU_SUB_FLAG_CFUNC);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		documentation hack
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	__QuSys__undocumented__ =
		qu_module_add_sub (QuSysModule, "__undocumented__", sys__undocumented, 1,
			QU_SUB_FLAG_PRIVATE | QU_SUB_FLAG_CFUNC);

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		print
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "print"      , qu__sys_print     , -1);
	qu_module_add_function (QuSysModule, "printc"     , qu__sys_printc    , -1);
	qu_module_add_function (QuSysModule, "printf"     , qu__sys_printf    , -1);
	qu_module_add_function (QuSysModule, "prints"     , qu__sys_prints    , -1);
	qu_module_add_function (QuSysModule, "println"    , qu__sys_println   , -1);
	qu_module_add_function (QuSysModule, "printcln"   , qu__sys_printcln  , -1);
	qu_module_add_function (QuSysModule, "printfln"   , qu__sys_printfln  , -1);
	qu_module_add_function (QuSysModule, "printsln"   , qu__sys_printsln  , -1);
	qu_module_add_function (QuSysModule, "__buffer__" , qu__sys__buffer__ ,  0);
	qu_module_add_function (QuSysModule, "__clear__"  , qu__sys__clear__  ,  0);
	qu_module_add_function (QuSysModule, "__collect__", qu__sys__collect__,  0);
	qu_module_add_function (QuSysModule, "__trash__"  , qu__sys__trash__  ,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		info
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "classof"    , qu__sys_classof   ,  1);
	qu_module_add_function (QuSysModule, "nameof"     , qu__sys_nameof    ,  1);
	qu_module_add_function (QuSysModule, "anyof"      , qu__sys_anyof     , -1);
	qu_module_add_function (QuSysModule, "callable"   , qu__sys_callable  , -1);
	qu_module_add_function (QuSysModule, "iterable"   , qu__sys_iterable  , -1);
	qu_module_add_function (QuSysModule, "__main__"   , qu__sys__main__   ,  0);
	qu_module_add_function (QuSysModule, "__thread__" , qu__sys__thread__ ,  0);
	qu_module_add_function (QuSysModule, "__where__"  , qu__sys__where__  ,  0);

	QuSys__arg__ = qu_module_add_var (QuSysModule, "__arg__", qu_array_new (8));


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		minimal shell
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "sysclear" , qu__sys_sysclear, -1);
	qu_module_add_function (QuSysModule, "sysinput" , qu__sys_sysinput, -1);
	qu_module_add_function (QuSysModule, "sysexec"  , qu__sys_sysexec , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		safe call
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "syscall", qu__sys_syscall, -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		miscellaneous
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "this"     , qu__sys_this      , -1);
	qu_module_add_function (QuSysModule, "sysdelay" , qu__sys_sysdelay  ,  1);

	__QuSys_syscat =
	qu_module_add_function (QuSysModule, "syscat"   , qu__sys_syscat    , -1);

	qu_module_add_function (QuSysModule, "sysjoin"  , qu__sys_sysjoin   , -1);
	qu_module_add_function (QuSysModule, "sysref"   , qu__sys_sysref    , -1);
	qu_module_add_function (QuSysModule, "sysfilter", qu__sys_sysfilter ,  2);
	qu_module_add_function (QuSysModule, "sysmap"   , qu__sys_sysmap    ,  2);
	qu_module_add_function (QuSysModule, "sysmax"   , qu__sys_sysmax    , -1);
	qu_module_add_function (QuSysModule, "sysmin"   , qu__sys_sysmin    , -1);
	qu_module_add_function (QuSysModule, "syslen"   , qu__sys_syslen    ,  1);

	RegexMatchSub = qu_module_add_const (QuSysModule, "__regmatch__", QuNil);

	qu_module_add_function (QuSysModule, "__package__", qu__sys__package,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		format
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	FormatStruct = qu_struct_new_static (10,
		"mode"      , QuStringClass,
		"left"      , QuBooleanClass,
		"sign"      , QuIntClass,
		"zero"      , QuBooleanClass,
		"group"     , QuIntClass,
		"width"     , QuIntClass,
		"prec"      , QuIntClass,
		"ucase"     , QuBooleanClass,
		"sep"       , QuStringClass,
		"dec"       , QuStringClass);

	qu_module_add_function (QuSysModule, "sysparseformat" , qu__sys_parseformat, 1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		error
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "errno"      , qu__sys_errno     , -1);
	qu_module_add_function (QuSysModule, "require"    , qu__sys_require   , -1);
	qu_module_add_function (QuSysModule, "raise"      , qu__sys_raise     , -1);
	qu_module_add_function (QuSysModule, "__abort__"  , qu__sys__abort__  , -1);

	__QuSys__err__ =
	qu_module_add_function (QuSysModule, "__err__"    , qu__sys__err__    ,  0);
	qu_module_add_function (QuSysModule, "__errmsg__" , qu__sys__errmsg__ ,  0);
	qu_module_add_function (QuSysModule, "__verbose__", qu__sys__verbose__, -1);
	qu_module_add_function (QuSysModule, "__warning__", qu__sys__warning__, -1);
	qu_module_add_function (QuSysModule, "__remark__" , qu__sys__remark__ , -1);
	qu_module_add_function (QuSysModule, "__trace__"  , qu__sys__trace__  , -1);
	qu_module_add_function (QuSysModule, "__exit__"   , qu__sys__exit__   , -1);

	QuSys__tracefilter__ = qu_module_add_const (QuSysModule, "__tracefilter__", QuSetEMPTY);
	QuSys__tracehow__    = qu_module_add_const (QuSysModule, "__tracehow__"   , QuSetEMPTY);
	QuSys__debug__       = qu_module_add_const (QuSysModule, "__debug__"      , QU_APP_DEBUG ? QuTrue : QuFalse);
	QuSys__daemon__      = qu_module_add_const (QuSysModule, "__daemon__"     , QU_APP_DAEMON ? QuTrue : QuFalse);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		context
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "__context__", qu__sys__context__, -1);

	QuSys__ignore = qu_module_add_var (QuSysModule, "__", QuNil);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		numeric
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "__complex__" , qu__sys__complex__ ,  2);
	qu_module_add_function (QuSysModule, "__inf__"     , qu__sys__inf__     ,  1);
	qu_module_add_function (QuSysModule, "__infp__"    , qu__sys__infp__    ,  0);
	qu_module_add_function (QuSysModule, "__infn__"    , qu__sys__infn__    ,  0);
	qu_module_add_function (QuSysModule, "__tiny__"    , qu__sys__tiny      , -1);
	qu_module_add_function (QuSysModule, "__huge__"    , qu__sys__huge__    ,  1);
	qu_module_add_function (QuSysModule, "__zerodiv__" , qu__sys__zerodiv__ ,  1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		stream
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "__stderr__" , qu__sys__stderr__ ,  0);
	qu_module_add_function (QuSysModule, "__stdin__"  , qu__sys__stdin__  ,  0);
	qu_module_add_function (QuSysModule, "__stdout__" , qu__sys__stdout__ ,  0);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		eval
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_module_add_function (QuSysModule, "eval" , qu__sys_eval , -1);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		C types
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	QuType_char       = qu_module_add_function (QuSysModule, "Cchar"      , qu_type_char      , 1);
	QuType_byte       = qu_module_add_function (QuSysModule, "Cbyte"      , qu_type_byte      , 1);
	QuType_ubyte      = qu_module_add_function (QuSysModule, "Cubyte"     , qu_type_ubyte     , 1);
	QuType_short      = qu_module_add_function (QuSysModule, "Cshort"     , qu_type_short     , 1);
	QuType_ushort     = qu_module_add_function (QuSysModule, "Cushort"    , qu_type_ushort    , 1);
	QuType_int        = qu_module_add_function (QuSysModule, "Cint"       , qu_type_int       , 1);
	QuType_uint       = qu_module_add_function (QuSysModule, "Cuint"      , qu_type_uint      , 1);
	QuType_int16      = qu_module_add_function (QuSysModule, "Cint16"     , qu_type_int16     , 1);
	QuType_uint16     = qu_module_add_function (QuSysModule, "Cuint16"    , qu_type_uint16    , 1);
	QuType_int32      = qu_module_add_function (QuSysModule, "Cint32"     , qu_type_int32     , 1);
	QuType_uint32     = qu_module_add_function (QuSysModule, "Cuint32"    , qu_type_uint32    , 1);
	QuType_long       = qu_module_add_function (QuSysModule, "Clong"      , qu_type_long      , 1);
	QuType_ulong      = qu_module_add_function (QuSysModule, "Culong"     , qu_type_ulong     , 1);
	QuType_longlong   = qu_module_add_function (QuSysModule, "Clonglong"  , qu_type_longlong  , 1);
	QuType_ulonglong  = qu_module_add_function (QuSysModule, "Culonglong" , qu_type_ulonglong , 1);
	QuType_float      = qu_module_add_function (QuSysModule, "Cfloat"     , qu_type_float     , 1);
	QuType_double     = qu_module_add_function (QuSysModule, "Cdouble"    , qu_type_double    , 1);
	QuType_longdouble = qu_module_add_function (QuSysModule, "Clongdouble", qu_type_longdouble, 1);
	QuType_size_t     = qu_module_add_function (QuSysModule, "Csize_t"    , qu_type_size_t    , 1);
	QuType_ssize_t    = qu_module_add_function (QuSysModule, "Cssize_t"   , qu_type_ssize_t   , 1);
	QuType_off_t      = qu_module_add_function (QuSysModule, "Coff_t"     , qu_type_off_t     , 1);
	QuType_char_p     = qu_module_add_function (QuSysModule, "Cstring"    , qu_type_char_p    , 1);
	QuType_void_p     = qu_module_add_function (QuSysModule, "Cpointer"   , qu_type_void_p    , 1);
#ifdef QU_SUPPORT_GLIB
	QuType_GObject    = qu_module_add_function (QuSysModule, "GObject"    , qu_type_GObject   , 1);
#endif


#define e_(u) qu_flag_set (u, QU_SUB_FLAG_NOAUTO)
	/* this makes things such as: foo (Cint) to not call Cint but return
		the Sub object instead */
	e_(QuType_char       );
	e_(QuType_byte       );
	e_(QuType_ubyte      );
	e_(QuType_short      );
	e_(QuType_ushort     );
	e_(QuType_int        );
	e_(QuType_uint       );
	e_(QuType_int16      );
	e_(QuType_uint16     );
	e_(QuType_int32      );
	e_(QuType_uint32     );
	e_(QuType_long       );
	e_(QuType_ulong      );
	e_(QuType_longlong   );
	e_(QuType_ulonglong  );
	e_(QuType_float      );
	e_(QuType_double     );
	e_(QuType_longdouble );
	e_(QuType_size_t     );
	e_(QuType_ssize_t    );
	e_(QuType_off_t      );
	e_(QuType_char_p     );
	e_(QuType_void_p     );
#undef e_

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		special inits
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	qu_struct_sys_init ();
}
