/* Thread.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/* the final classes */
qu_class QuThreadClass;
qu_class QuMutexClass;
qu_class QuCondClass;

/* the main thread */
qu_thread __QuThreadMain;

/* the current thread */
qu_thread volatile __QuThreadCurrent = 0;

/* set to 1 once another thread is created */
int volatile __QuThreadMulti = 0;

/* the key to use with thread_specific */
qu_thread_key_t __QuThreadKey;

/* keep track of the number of live threads, incl. __QuThreadMain */
static int volatile NumThreads = 1;


/*=============================================================================

	system thread

=============================================================================*/

/*
	Create a new thread.
*/
int qu_thread_create (qu_thread_t *tid, void *(*start)(void *), void *arg,
                      int stacksize, int detach)
{
    pthread_attr_t attr;

#if (SIZEOF_PTHREAD_T == SIZEOF_LONG) || (SIZEOF_PTHREAD_T == SIZEOF_INT)
	int stackdef;
#else
	long stackdef;
#endif

	errno = pthread_attr_init (&attr);
	if (errno)
		return -1;

	if (stacksize > 0)
	{
		if (pthread_attr_getstacksize (&attr, &stackdef) == 0)
		{
			if (stacksize < QU_THREAD_STACK_MIN)
				stacksize = QU_THREAD_STACK_MIN;
			else if (stacksize > stackdef)
				stacksize = stackdef;
			pthread_attr_setstacksize (&attr, stacksize);
		}
	}

	if (detach)
	{
		errno = pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
		if (errno)
			return -1;
	}

	errno = pthread_create (tid, &attr, start, arg);
	if (errno)
		return -1;

	return 0;
}



/*
	Give up CPU to other thread.
*/
#ifndef HAVE_SCHED_YIELD
void qu_thread_yield ()
{
	struct timespec req, rem;
	req.tv_sec = 0;
	req.tv_nsec = 0;
	nanosleep (&req, &rem);
}
#endif



/*=============================================================================

	qu_thread

=============================================================================*/

/*) class Thread __final
	The Thread class allows you to create multithread applications.

	See also the -{Mutex} and -{Cond} classes.

	@xmp
		sub foo (x, y)
			println (__thread__.id ' entering self=' self ' x=' x ' y=' y)
			pass
			# do something ...
			println (__thread__.id ' leaving')
		;;

		t1 = Thread (@foo, nil, 1, 2)
		t2 = Thread (@foo, 'Hey', 3, 4)

		t1.join
		t2.join
		println ('done!')

	@out
		1026 entering self=nil x=1 y=2
		2051 entering self=Hey x=3 y=4
		1026 leaving
		2051 leaving
		done!
*/

/*-----------------------------------------------------------------------------
	macros
-----------------------------------------------------------------------------*/

/*
	Flag accessors that should only be used within qu_thread.c
*/
#define qu_thread_alive(self)        qu_flag_test  (self, QU_THREAD_FLAG_ALIVE)
#define qu_thread_set_alive(self)    qu_flag_set   (self, QU_THREAD_FLAG_ALIVE)
#define qu_thread_unset_alive(self)  qu_flag_unset (self, QU_THREAD_FLAG_ALIVE)

#define qu_thread_detached(self)     qu_flag_test  (self, QU_THREAD_FLAG_DETACHED)
#define qu_thread_set_detached(self) qu_flag_set   (self, QU_THREAD_FLAG_DETACHED)
/*      qu_thread_unset_alive(self)  undefined since the process is irreversible */



/*-----------------------------------------------------------------------------
	internals
-----------------------------------------------------------------------------*/

struct arg_s
{
	qu_thread self;     /* qu_thread object */
	qu_any    theself;  /* The self object to pass to the function */
	qu_sub    sub;      /* The starting function */
	qu_array  args;     /* Arguments to pass to the function */
	int       go;       /* To synchronize with qu__thread_new */
};


/*
	All threads born and die here.
	Except of course, the main thread.
*/
static void *qu__thread_start (struct arg_s *a)
{
	qu_thread self;
	qu_any    theself;

	while (a->go != 1)
		qu_thread_yield ();

	self = a->self;
	theself = a->theself;

	qu_thread_set_blocking (self);

	/* We store our qu_thread object in the os thread. */
	if (qu_thread_specific_set (__QuThreadKey, self) != 0)
		QU_FATAL ("thread_specific_set failed");

	/* See qu_sig_init_worker for what this means */
	qu_sig_init_worker ();

	/*
		Grab the args.
		Once qu__thread_new is done (by setting a->go = 0), they are gone.
	*/

	self->args = a->args;
	self->sub = a->sub;

	/* Run! */
	a->go = 0;
	NumThreads++;
	qu_interp_run_thread (self, theself);
	NumThreads--;

	/* these are no longer needed */
	self->stacklo = 0;
	self->stackhi = 0;
	self->buffers = 0;
	self->state   = 0;
	self->extra   = 0;
	self->html    = 0;

	if (qu_thread_detached (self))
	{
		self->eclass = 0;
		self->emsg   = 0;
		self->ewhere = 0;
	}

	qu_thread_unset_alive (self);
	return NULL;
}



static qu_thread qu__thread_alloc (qu_thread_t tid)
{
	qu_thread t;

	t           = qu_gc_new (QuThreadClass);
	t->tid      = tid;
	t->eclass   = 0;
	t->emsg     = 0;
	t->ewhere   = 0;
	t->eline    = 0;
	t->ticks    = 0;
	t->depth    = 0;
	t->result   = 0;
	t->atomic   = 0;
	t->args     = 0;
	t->sub      = 0;
	t->buffers  = 0;
	t->stacklo  = 0;
	t->stackhi  = 0;
	t->data     = 0;
	t->state    = 0;
	t->extra    = 0;
	t->html     = 0;

	qu_thread_set_alive (t);
	return t;
}



qu_any qu_thread_data_get (qu_thread self, qu_any key)
{
	if (self->data == 0)
		return 0;
	return qu_dict_getitem (self->data, key);
}



qu_any qu_thread_data_set (qu_thread self, qu_any key, qu_any val)
{
	if (self->data == NULL)
		self->data = qu_dict_new (1);
	return qu_dict_setitem (self->data, key, val);
}



/*-----------------------------------------------------------------------------
	standard class methods
-----------------------------------------------------------------------------*/

/*) sub class new (f, theself = nil, *arg) -> Thread
	@def
	sub class new (f: String, theself = nil, *arg) -> Thread
	sub class new (f: Sub, theself = nil, *arg) -> Thread

	@p
	Create a new thread.

	v{f} must be a top level static -{Sub} or a method name of v{theself}.
	In case of a static Sub, it will receive v{theself} as self.

	The newly created Thread will enter v{f} passing v{arg} as the arguments.
	Leaving v{f} automatically kills the thread (but not the -{Thread} object).

	@warn
	All threads are created in a joinable state (not detached). You can detach
	a thread using the #{detach} method.

	@exc
	.ESecurity
	.EArgValue

	@see
		-{__thread__}
*/
static qu_thread qu__thread_new (qu_class self, int argc, qu_any *argv)
{
	qu_thread_t  tid;
	qu_thread    t;
	qu_sub       u;
	struct arg_s a;
	qu_array     g;
	qu_any       theself;

	__qu_unsafe__;

	QU_ARG_VMIN (1);

	u = argv[0];
	theself = (argc > 1) ? argv[1] : (qu_any) QuNil;

	switch (qu_classid (u))
	{
		case QU_CLASS_ID_Sub:
			break;

		case QU_CLASS_ID_String:
			u = qu_classof (theself)->get (theself, (qu_string) u, 0);
			QU_ENSURE (u);
			QU_ARG_TYPE (u, qu_sub);
			if (qu_sub_is_class (u))
			{
				if (!qu_class_is (theself))
					theself = qu_classof (theself);
			}
			break;

		default:
			QU_ARG_THROW;
	}

	if (u->root != u)
		QU_ARG_THROWZ ("cant use nested function as thread function");

	if (argc > 1)
	{
		argc -= 2;
		argv++;
	}
	else
	{
		argc--;
	}

	if (argc == 0)
	{
		a.args = QuArrayEmpty;
	}
	else
	{
		g = qu_array_new (argc);
		while (argc--)
		{
			argv++;
			qu_array_push (g, *argv);
		}
		a.args = g;
	}

	a.sub = u;
	a.go = 0;
	a.theself = theself;

	if (qu_thread_create (&tid, (void *(*)(void *)) qu__thread_start, (void *) &a, QU_THREAD_STACK_MIN, 0) != 0)
		QU_THROWZ (QuEMemory, "cant create system thread");

	__QuThreadMulti = 1;

	a.self = t = qu__thread_alloc (tid);
	a.go = 1;

	while (a.go != 0)
		qu_thread_yield ();

	return t;
}



/*) sub class create (n: Integer, f, theself = nil, *arg) -> Array
	This does the same as #{new} except that it creates a number of threads
	in one go.

	Note that if you create too many threads that can be created by the system
	then this method may return an Array with less than v{n} items.

	@exc
	.ESecurity
	.EArgValue
*/
static qu_array qu__thread_create (qu_class self, int argc, qu_any *argv)
{
	qu_thread_t tid;
	qu_thread   t;
	qu_sub      u;
	qu_array    g, r;
	qu_any      theself;
	int         n;

	__qu_unsafe__;

	QU_ARG_VMIN (2);
	QU_ARG_vint (0, n);
	QU_ARG_CHECK (n > 0);
	QU_MEM_CHECK (n <= QU_GC_MAX_ITEMS);

	u = argv[1];
	theself = (argc > 2) ? argv[2] : (qu_any) QuNil;

	switch (qu_classid (u))
	{
		case QU_CLASS_ID_Sub:
			break;

		case QU_CLASS_ID_String:
			u = qu_classof (theself)->get (theself, (qu_string) u, 0);
			QU_ENSURE (u);
			QU_ARG_TYPE (u, qu_sub);
			if (qu_sub_is_class (u))
			{
				if (!qu_class_is (theself))
					theself = qu_classof (theself);
			}
			break;

		default:
			QU_ARG_THROW;
	}

	if (u->root != u)
		QU_ARG_THROWZ ("cant use nested function as thread function");

	if (argc > 2)
	{
		argc -= 3;
		argv += 2;
	}
	else
	{
		argc -= 2;
	}

	if (argc == 0)
	{
		g = QuArrayEmpty;
	}
	else
	{
		g = qu_array_new (argc);
		while (argc--)
		{
			argv++;
			qu_array_push (g, *argv);
		}
	}

	r = qu_array_new (n);
	while (n--)
	{
		struct arg_s a;

		a.args = g;
		a.sub = u;
		a.go = 0;
		a.theself = theself;

		if (qu_thread_create (&tid, (void *(*)(void *)) qu__thread_start, (void *) &a, QU_THREAD_STACK_MIN, 0) != 0)
			break;

		__QuThreadMulti = 1;

		a.self = t = qu__thread_alloc (tid);
		a.go   = 1;

		while (a.go != 0)
			qu_thread_yield ();

		qu_array_push (r, t);
	}
	return r;
}



/*) sub class this
	Return this Thread object. This method returns the same result as the
	builtin -{__thread__} function.

	@noexc

	@xmp
		sub foo
			println (Thread.this.id)
		;;

		t1 = Thread (@foo, nil)
		t1.join

	@out
		1026
*/
static qu_thread qu__thread_this (qu_class self)
{
	return __QuThreadCurrent;
}



/*) sub class count
	(Unreliable) Return the number of alive threads.

	@noexc

	@xmp
		sub foo
			println (Thread.this.id ' : ' Thread.count)
		;;

		t = []
		for i in 0..4
			t.push (Thread (@foo, nil))
			pass
		;;

		for i in t
			i.join
		;;

	@out
		1026 : 2
		2051 : 2
		3074 : 2
		4098 : 2
		5122 : 2
*/
static qu_int qu__thread_count (qu_class self)
{
	assert (NumThreads <= QU_INT_MAX);
	return qu_int_new (NumThreads);
}



/*) sub class joinall (*threads) -> Array
	This does the same as #{join} except that all v{threads} are joined.

	v{threads} may be an Array of Thread objects or a list of Thread objects.
	Non Thread items are ignored.

	The returned value is an Array of the result of threads. If the thread
	threw an exception then the exception class is the result.

	@noexc

	@xmp
		sub foo
			return 'hey'
		;;

		t = Thread.create (3, @foo)
		r = Thread.joinall (t)
		println (r)

	@out
		['hey', 'hey', 'hey']
*/
static qu_array qu__thread__joinall (int terminate, int argc, qu_any *argv)
{
	qu_thread me = __QuThreadCurrent;
	long      i;
	volatile qu_array r = 0;

	QU_ARG_VMIN (1);

	if (argc == 1 && qu_array_is (argv[0]))
	{
		qu_array a = argv[0];

		r = qu_array_new (a->len);

		__qu_blocking_begin;
		for (i = 0; i < a->len; ++i)
		{
			qu_thread t = (qu_thread) a->objs[i];
			qu_any    x = QuNil;

			if (qu_thread_is (t)
				&& me != t
				&& !qu_thread_detached (t)
				&& qu_thread_join (t->tid, 0) == 0)
			{
				x = t->result;
				if (x == 0)
					x = t->eclass;
			}
			r->objs[i] = x;
			r->len++;

			if (!terminate)
				__qu_blocking_check_signal;
		}
		__qu_blocking_end;
	}
	else
	{
		r = qu_array_new (argc);

		__qu_blocking_begin;
		i = 0;
		while (argc--)
		{
			qu_thread t = (qu_thread) *argv++;
			qu_any    x = QuNil;

			if (qu_thread_is (t)
				&& me != t
				&& !qu_thread_detached (t)
				&& qu_thread_join (t->tid, 0) == 0)
			{
				x = t->result;
				if (x == 0)
					x = t->eclass;
			}
			r->objs[i++] = x;
			r->len++;
			if (!terminate)
				__qu_blocking_check_signal;
		}
		__qu_blocking_end;
	}
	return r;
}



static qu_array qu__thread_joinall (qu_class self, int argc, qu_any *argv)
{
	return qu__thread__joinall (0, argc, argv);
}



/*) sub class exit (*threads) -> Array
	This does the same as #{joinall} except that termination signals are
	ignored.
*/
static qu_array qu__thread_exit (qu_class self, int argc, qu_any *argv)
{
	QU_RETURN (qu__thread__joinall (1, argc, argv));
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

static void qu__thread__mark (qu_thread self)
{
	qu_gc_xmark (self->args);
	qu_gc_xmark (self->sub);
	qu_gc_xmark (self->buffers);
	qu_gc_xmark (self->eclass);
	qu_gc_xmark (self->emsg);
	if (!qu_thread_is_efunction (self))
		qu_gc_xmark (self->ewhere);
	qu_gc_xmark (self->result);
	qu_gc_xmark (self->data);
	qu_gc_xmark (self->state);
	qu_gc_xmark (self->extra);
	qu_gc_xmark (self->html);
	if (qu_thread_alive (self))
		qu_gc_mark_cstack (self, self->stacklo, self->stackhi);
}


#ifdef QU_DEBUG_LOCAL
static void qu__thread__destroy (qu_thread self)
{
	fprintf (stderr, "thread freed: " QU_THREAD_IDFORMAT "\n", qu_thread_id ());
}
#endif



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*) sub __getitem (key) -> Any
	Get thread specific #{data}.

	@exc
	.ENone  #{data} not set

	@p
	See -{Dict.__getitem} for other exceptions.

	@xmp
		t = __thread__
		t['foo'] = 1
		println (t['foo'])

	@out
		1
*/
qu_any qu__thread__getitem (qu_thread self, qu_any key)
{
	qu_any v = qu_thread_data_get (self, key);
	QU_REQUIRE (QuENone, v);
	return v;
}



/*) sub __getslice (key, defval) -> Any
	This does the same as #{__getitem} except that if the data can not be found
	then v{defval} is returned.
*/
static qu_any qu__thread__getslice (qu_thread self, qu_any key, qu_any defval)
{
	qu_any v;

	QU_REQUIRE (QuENone, self->data);
	v = qu_dict_getitem (self->data, key);
	return v ? v : defval;
}



/*) sub __setitem (key, value) -> Any
	Set thread specific #{data} then return v{value}.

	See -{Dict.__setitem} for possible exceptions.

	@xmp
		t = __thread__
		t['foo'] = 1
*/
qu_any qu__thread__setitem (qu_thread self, qu_any key, qu_any val)
{
	return qu_thread_data_set (self, key, val);
}



/*-----------------------------------------------------------------------------
	specific instance methods
-----------------------------------------------------------------------------*/

/*) sub alive
	(Unreliable) Whether or not the thread is alive.

	@noexc

	@xmp
		sub foo
			println ('hello')
		;;

		t1 = Thread (@foo, nil)
		println (t1.alive)
		t1.join

	@out
		true
*/
static qu_boolean qu__thread_alive (qu_thread self)
{
	return qu_thread_alive (self) ? QuTrue : QuFalse;
}



/*) sub data (*arg) -> Dict
	@def
	sub data -> Dict
	sub data (d: Dict) -> Dict
	@p
	Get or set thread specific data.

	The Thread class allows you to store/get per-thread specific data using
	a Dict.

	If called with no arguments, return the Dict used to store the data.
	If none has been set then ENone is thrown.

	If called with one argument, set the Dict.

	@exc
	.EArgValue

	@xmp
		t = __thread__
		t.data ([foo:1, bar:2])
		println (t.data)
		println (t.data['foo'])
		t['bar'] = 20
		println (t.data)
		println (t['bar'])

	@out
		[foo:1, bar:2]
		1
		[foo:1, bar:20]
		20
*/
static qu_dict qu__thread_data (qu_thread self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		QU_REQUIRE (QuENone, self->data);
	}
	else
	{
		qu_dict x;
		QU_ARG_VTYPE (0, x, qu_dict);
		self->data = x;
	}
	return self->data;
}



/*) sub detach -> Null
	Set thread to detached state.
	If it is already #{detached} then nothing happens.

	Note that a detached thread can not be #{join}ed.

	@exc
	.ECall thread is dead

	@xmp
		sub foo
			println ('hello')
		;;

		t1 = Thread (@foo, nil)
		t1.detach
		pass
		try
			t1.join
		catch
			println (__err__)
		;;

	@out
		hello
		(EArgValue) cant join a detached thread
*/
static qu_null qu__thread_detach (qu_thread self)
{
	if (!qu_thread_alive (self))
		QU_THROWZ (QuECall, "thread is dead");

	if (!qu_thread_detached (self))
	{
		qu_thread_set_detached (self);
		qu_thread_detach (self->tid);
	}
	return QuNil;
}



/*) sub detached -> Boolean
	(Unreliable) Return true if the thread is in a detached state, false
	otherwise.

	@noexc

	@xmp
		sub foo
			println ('hello')
		;;

		t1 = Thread (@foo, nil)
		println (t1.detached)
		t1.join

	@out
		false
		hello
*/
static qu_boolean qu__thread_detached (qu_thread self)
{
	return qu_thread_detached (self) ? QuTrue : QuFalse;
}



/*) sub id -> Integer
	Return this thread id.

	@noexc

	@xmp
		sub foo
			println (Thread.this.id)
		;;

		t1 = Thread (@foo, nil)
		t1.join

	@out
		1026
*/
static qu_integer qu__thread_id (qu_thread self)
{
#if SIZEOF_PTHREAD_T <= SIZE_OF_INT
	return qu_integer_newi (self->tid);
#elif SIZEOF_PTHREAD_T <= SIZE_OF_LONG
	return qu_integer_newl (self->tid);
#else
	/* implies HAVE_LONG_LONG */
	return qu_integer_newll (self->tid);
#endif
}



/*) sub join -> Any
	Join this thread.

	This method blocks the calling thread until the thread to join has completed.

	If the thread threw an exception then calling this method will rethrow that
	exception. Otherwise the value returned is the value returned by the thread
	function.

	@exc
	.ENone     thread is dead or detached
	.EDeadlock

	@xmp
		sub foo
			return 'hey'
		;;

		t1 = Thread (@foo, nil)
		r = t1.join
		println (r)

	@out
		hey
*/
static qu_any qu__thread_join (qu_thread self)
{
	qu_thread me = __QuThreadCurrent;

	if (me == self)
		QU_THROWZ (QuEDeadlock, "cant join self");

	if (qu_thread_detached (self))
		QU_THROWZ (QuENone, "thread is detached");

	__qu_blocking_begin;
	switch (qu_thread_join (self->tid, 0))
	{
		case 0:
			break;

		case ESRCH:
			QU_THROW (QuENone); /* this should be a QU_FATAL error */

		case EINVAL:
			QU_THROWZ (QuEDeadlock, "another thread is waiting");

		default:
			QU_THROW (QuEDeadlock); /* should never happen */
	}
	__qu_blocking_end;
	if (self->result == 0)
	{
		if (qu_thread_is_efunction (self))
			qu_thread__throw (self->eclass, self->emsg, self->ewhere);
		else
			qu_thread_throwex (self->eclass, self->emsg, self->ewhere, self->eline);
		return 0;
	}
	return self->result;
}



/*=============================================================================

	qu_mutex

=============================================================================*/

/*) class Mutex __final
	A Mutex objects is used to lock resources from being accessed by other
	threads.

	See also the -{Thread} and -{Cond} classes.

	@xmp
		v = 1
		m = Mutex ()

		sub foo
			m.lock
			v++
			println (__thread__.id ' : ' v)
			m.unlock
		;;

		t1 = Thread (@foo)
		t2 = Thread (@foo)

		t1.join
		t2.join

	@out
		1026 : 2
		2051 : 3
*/
typedef struct qu_mutex_s qu_mutex_s;
struct qu_mutex_s
{
	QU_OBJECT_HEAD;
	qu_thread_mutex_t mutex;
};

typedef qu_mutex_s *qu_mutex;


/*-----------------------------------------------------------------------------
	standard class methods
-----------------------------------------------------------------------------*/

/*) sub class new -> Mutex
	Create a new Mutex object.

	@exc
	.ESecurity
*/
static qu_mutex qu__mutex_new (qu_class self)
{
	qu_mutex m;

	__qu_unsafe__;

	m = qu_gc_new (QuMutexClass);
	qu_thread_mutex_init (&m->mutex);
	return m;
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

/* none */



/*-----------------------------------------------------------------------------
	specific instance methods
-----------------------------------------------------------------------------*/

/*) sub lock -> Null
	Lock the mutex.

	@warn
	Since Qu-1.20 there is no automatic deadlock detection so be careful.

	@noexc
*/
static qu_null qu__mutex_lock (qu_mutex self)
{
	__qu_blocking_begin;
	qu_thread_mutex_lock (&self->mutex);
	__qu_blocking_end;
	return QuNil;
}



/*) sub trylock -> Boolean
	Lock the mutex but only if it is not locked by another thread.

	The returned values is:
	@ul
	.false means that the mutex is locked by another thread.
	.true  means that the calling thread has acquired the lock

	@warn
	Since Qu-1.20 there is no automatic deadlock detection so be careful.

	@noexc

	@xmp
		m = Mutex ()
		if m.trylock
			println ('got it!')
			m.unlock
		else
			println ('someone else has the lock')
		;;
*/
static qu_boolean qu__mutex_trylock (qu_mutex self)
{
	int r;

	__qu_blocking_begin;
	r = qu_thread_mutex_trylock (&self->mutex);
	__qu_blocking_end;
	return r ? QuFalse : QuTrue;
}



/*) sub unlock -> Null
	Unlock the mutex.

	@warn
	Since Qu-1.20 there is no automatic deadlock detection so be careful.

	@noexc
*/
static qu_null qu__mutex_unlock (qu_mutex self)
{
	__qu_blocking_begin;
	qu_thread_mutex_unlock (&self->mutex);
	__qu_blocking_end;
	return QuNil;
}



/*=============================================================================

	qu_cond

=============================================================================*/

/*) class Cond __final
	Cond (condition variable) objects are used as a mean to control the execution
	of threads.

	@warn
	Since Qu-1.20 condition variables already contain their own mutexes.

	@xmp
		c = Cond ()
		k = 0

		sub foo
			atomic
				k += 1
			;;
			println ('thread ' __thread__.id ' waiting ')
			c.wait
			println ('thread ' __thread__.id ' leaving')
		;;

		t1 = Thread (@foo)
		t2 = Thread (@foo)
		t3 = Thread (@foo)

		loop
			pass
		until k == 3

		println ('broadcasting ...')
		c.broadcast
		Thread.joinall (t1, t2, t3)
		println ('done')

	@out
		thread 1026 waiting
		thread 2051 waiting
		thread 3076 waiting
		broadcasting ...
		thread 1026 leaving
		thread 2051 leaving
		thread 3076 leaving
		done
*/

typedef struct qu_cond_s qu_cond_s;
struct qu_cond_s
{
	QU_OBJECT_HEAD;
	long              count;
	qu_thread_cond_t  cond;
	qu_thread_mutex_t mutex;
};

typedef qu_cond_s *qu_cond;


/*-----------------------------------------------------------------------------
	standard class methods
-----------------------------------------------------------------------------*/

/*) sub class new -> Cond
	Create a new -{Cond} object.

	@exc
	.ESecurity
*/
static qu_cond qu__cond_new (qu_class self)
{
	qu_cond c;

	__qu_unsafe__;
	c = qu_gc_new (QuCondClass);
	c->count = 0;
	qu_thread_mutex_init (&c->mutex);
	qu_thread_cond_init (&c->cond);
	return c;
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

/* none */



/*-----------------------------------------------------------------------------
	specific instance methods
-----------------------------------------------------------------------------*/

/*) sub broadcast -> Null
	Broadcast the -{Cond} object, i.e. release all threads that are currently
	waiting.

	If no threads are waiting then nothing happens.

	@noexc

	@xmp
		c = Cond ()
		k = 0

		sub foo
			atomic
				k += 1
			;;
			println ('thread ' __thread__.id ' waiting ')
			c.wait
			println ('thread ' __thread__.id ' leaving')
		;;

		t1 = Thread (@foo)
		t2 = Thread (@foo)
		t3 = Thread (@foo)

		loop
			pass
		until k == 3

		println ('signaling ...')
		c.broadcast
		Thread.joinall (t1, t2, t3)
		println ('done')

	@out
		thread 1026 waiting
		thread 2051 waiting
		thread 3076 waiting
		signaling ...
		thread 1026 leaving
		thread 2051 leaving
		thread 3076 leaving
		done
*/
static qu_null qu__cond_broadcast (qu_cond self)
{
	qu_thread_cond_broadcast (&self->cond);
	return QuNil;
}



/*) sub count -> Int
	Return the number of thread waiting.

	@noexc
*/
static qu_int qu__cond_count (qu_cond self)
{
	return qu_int_newl (self->count);
}



/*) sub signal (microsec: Integer = 0) -> Bool
	Signal the -{Cond} object, i.e. release one thread, if any is waiting.

	If v{microsec} is non-positive then wait until one thread is released.
	If positive, then wait but not longer than the specified value.

	Return true if any thread has been released, false otherwise.

	@exc
	.EArgValue

	@xmp
		c = Cond ()
		k = 0

		sub foo
			atomic
				k += 1
			;;
			println ('thread ' __thread__.id ' waiting ')
			c.wait
			println ('thread ' __thread__.id ' leaving')
		;;

		t1 = Thread (@foo)
		t2 = Thread (@foo)
		t3 = Thread (@foo)

		loop
			pass
		until k == 3

		println ('signaling ...')
		c.signal
		pass

		println ('signaling ...')
		c.signal
		pass

		println ('signaling ...')
		c.signal
		pass

		Thread.joinall (t1, t2, t3)
		println ('done')

	@out
		thread 1026 waiting
		thread 2051 waiting
		thread 3076 waiting
		signaling ...
		thread 1026 leaving
		signaling ...
		thread 2051 leaving
		signaling ...
		thread 3076 leaving
		done
*/
static qu_boolean qu__cond_signal (qu_cond self, int argc, qu_any *argv)
{
	long t = 0;
	long r = self->count;

	QU_ARG_VCOUNT (0, 1);
	if (argc)
		QU_ARG_vlong (0, t);

	if (r > 0)
	{
		qu_thread_cond_signal (&self->cond);
	}
	else
	{
		double to = qu_stream_now () + (double) t;
		while (self->count == 0)
		{
			__qu_blocking_begin;
			if (t > 0 && qu_stream_now () > to)
			{
				__qu_blocking_end;
				return QuFalse;
			}
			__qu_blocking_check_signal;
			qu_thread_yield ();
			__qu_blocking_end;
		}
		qu_thread_cond_signal (&self->cond);
	}
	return QuTrue;
}



/*) sub wait (microsec: Integer) -> Null
	Wait for a condition, i.e. block the calling thread until the -{Cond}
	variable is signaled.

	You can optionally set a timeout value.

	@exc
	.EArgValue
	.ETimeout
*/
static qu_null qu__cond_wait (qu_cond self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		__qu_blocking_begin;
		qu_thread_mutex_lock   (&self->mutex);
		self->count++;
		qu_thread_cond_wait    (&self->cond, &self->mutex);
		self->count--;
		qu_thread_mutex_unlock (&self->mutex);
		__qu_blocking_end;
	}
	else
	{
		struct timespec p;
		long t;

		QU_ARG_vlong (0, t);
		QU_ARG_CHECK (t >= 0);

		__qu_blocking_begin;
		p.tv_sec  = t / 1000000000L;
		p.tv_nsec = t % 1000000000L;
		qu_thread_mutex_lock (&self->mutex);
		self->count++;
		for (;;)
		{
			if (pthread_cond_timedwait (&self->cond, &self->mutex, &p) != EINTR)
			{
				__qu_blocking_end;
				QU_THROW (QuETimeout);
			}
		}
		self->count--;
		qu_thread_mutex_unlock (&self->mutex);
		__qu_blocking_end;
	}
	return QuNil;
}



/*-----------------------------------------------------------------------------
	interface
-----------------------------------------------------------------------------*/

static char *qu__thread_ewhere_string (char *buf, int *rem)
{
	qu_any     o = (qu_obj) __QuThreadCurrent->ewhere;
	qu_module  m = qu_object_get_module (o);
	char      *p = buf, *z;
	qu_class   c;
	long       n;
	int        l = *rem;

	if (m == QuModuleMain)
	{
		if (l <= 8)
			goto done;
		strcpy (p, "__main__");
		p += 8;
		l -= 8;
	}
	else
	{
		n = strlen (m->name->str);
		if (l <= n)
			goto done;
		strcpy (p, m->name->str);
		p += n;
		l -= n;
	}

	c = qu_classof (o);
	if (c == QuSubClass)
	{
		if (l <= 1)
			goto done;
		*p++ = '.';
		l--;
		c = qu_sub_get_class ((qu_sub) o);
		if (c != 0)
		{
			z = c->name->str;
			n = strlen (z);
			if (l <= n + 1)
				goto done;
			strcpy (p, z);
			p += n;
			*p++ = '.';
			l -= (n + 1);
		}
		z = ((qu_sub) o)->name->str;
		n = strlen (z);
		if (l <= n)
			goto done;
		strcpy (p, z);
		p += n;
		l -= n;
	}
	if (c == QuClassClass)
	{
		z = ((qu_class) o)->name->str;
		n = strlen (z);
		if (l <= n + 1)
			goto done;
		*p++ = '.';
		strcpy (p, z);
		p += n;
		l -= n;
	}

  done:
	*rem = l;
	return p;
}


/*
	QU_THROW and family record the C function name that throws the exception
	or the block object where the exception occurs.
*/
static qu_string qu__thread_errstr ()
{
	char   buf [2048];
	int    size = sizeof (buf);
	qu_obj b;
	qu_any x;
	char   num [64];
	char  *ptr = buf;
	int    rem = size - 1;
	int    dep;

#define prz_(strz) \
QU_BEGIN \
	long lEn = strlen (strz); \
	if (rem > lEn) \
	{ \
		strcpy (ptr, strz); \
		rem -= lEn; \
		ptr += lEn; \
	} \
QU_END

#define prl_(strz, len) \
QU_BEGIN \
	long lEn = len; \
	if (rem > lEn) \
	{ \
		strcpy (ptr, strz); \
		rem -= lEn; \
		ptr += lEn; \
	} \
QU_END

	assert (__QuThreadCurrent);

	sprintf (num, "[" QU_THREAD_IDFORMAT "]", __QuThreadCurrent->tid);
	prz_(num);
	prz_(" ");

	dep = __QuThreadCurrent->depth;
	if (dep > 8)
	{
		dep = 8;
		while (--dep > 0)
			prz_(". ");
		prz_("... ");
	}
	else
	{
		while (--dep > 0)
			prz_(". ");
	}

	if (qu_thread_is_efunction (__QuThreadCurrent))
	{
		char *z = (char *) __QuThreadCurrent->ewhere;
		if (z)
		{
			char fbuf [128];
			qu_string_cfunc_name (z, fbuf, sizeof (fbuf));
			prz_(fbuf);
		}
	}
	else
	{
		b = (qu_obj) __QuThreadCurrent->ewhere;
		if (b)
			ptr = qu__thread_ewhere_string (ptr, &rem);
		if (QU_APP_LINE_NUMBER)
		{
			sprintf (num, " line %ld", __QuThreadCurrent->eline);
			prz_(num);
		}
	}

	if (__QuThreadCurrent->eclass == 0)
		__QuThreadCurrent->eclass = QuERuntimeClass;

	prz_(" (");
	prz_(__QuThreadCurrent->eclass->name->str);
	prz_(")");

	x = __QuThreadCurrent->emsg;
	if (x == 0)
		x = qu_exception_defmsg (__QuThreadCurrent->eclass);
	x = qu_sys_string (x);
	prz_(" ");
	prz_(((qu_string) x)->str);
	prz_("\n");

#undef prz_
#undef prl_

	return qu_string_newz (buf);
}



/*
	Get the last exception as a string.
*/
qu_string qu_thread_err_string ()
{
	if (__QuThreadCurrent == 0)
	{
		qu_string m = qu_exception_defmsg (QuExceptionClass);
		return qu_string_newf ("(%s) %s", QuExceptionClass->name->str, m->str);
	}
	return qu__thread_errstr ();
}




/*
	Print the last exception.
*/
void qu_thread_print_err ()
{
	qu_string s = qu_thread_err_string ();

	if (__QuThreadCurrent == 0)
	{
		fprintf (stderr, "%s\n", s->str);
	}
	else
	{
		static int busy = 0;

	  __qu_atomic_begin;
		if (busy == 0)
		{
			busy = 1;
			qu_stream_printz (QuStreamStderr, s->str, s->len);
			busy = 0;
		}
	  __qu_atomic_end;
	}
}



void qu_thread_print_trace ()
{
	if (__QuThreadCurrent
		&& (QuSys__tracefilter__->value == QuSetEMPTY
			|| (qu_set_contains (QuSys__tracefilter__->value, __QuThreadCurrent->eclass)
			       == ((QuSys__tracehow__->value == QuTrue) ? 1 : 0))))
	{
			qu_thread_print_err ();
	}
}



static void qu__thread_cfunc_msg (qu_class kls, qu_any msg, const char *cfunc)
{
	if (cfunc && (msg == 0 || qu_string_is (msg)))
	{
		char buf [64];

		if (msg == 0)
			msg = qu_exception_defmsg (kls);
		qu_string_cfunc_name (cfunc, buf, sizeof (buf));
		__QuThreadCurrent->emsg = qu_string_newf ("%s: %s", buf, ((qu_string) msg)->str);
	}
}


void qu_thread__throw (qu_class kls, qu_any msg, const char *cfunc)
{
	if (kls == 0)
		kls = QuERuntimeClass;
	else
		assert (qu_exception_class_is (kls));

	if (__QuThreadCurrent)
	{
		assert (kls);
		__QuThreadCurrent->eclass = kls;
		__QuThreadCurrent->emsg   = msg;
		__QuThreadCurrent->ewhere = (char *) cfunc;
		qu_thread_set_efunction (__QuThreadCurrent);

		if (QU_APP_DEBUG)
			qu__thread_cfunc_msg (kls, msg, cfunc);

		if (QuSysTrace && !__QuCompileIsCompiling)
			qu_thread_print_trace ();
	}
	else
	{
		if (kls == 0)
			QU_FATAL ("error at initialization before Exception initialized");

		if (msg == 0)
		{
			QU_FATAL ("(%s) unspecified error at initialization", kls->name->str);
		}
		else
		{
			qu_string s = qu_sys_string (msg);
			QU_FATAL ("(%s) %s", kls->name->str, s->str);
		}
	}
}



void qu_thread__rethrow (const char *cfunc)
{
	if (__QuThreadCurrent)
	{
		__QuThreadCurrent->ewhere = (char *) cfunc;
		qu_thread_set_efunction (__QuThreadCurrent);

		if (QU_APP_DEBUG)
			qu__thread_cfunc_msg (__QuThreadCurrent->eclass, __QuThreadCurrent->emsg, cfunc);

		if (QuSysTrace && !__QuCompileIsCompiling)
			qu_thread_print_trace ();
	}
	else
	{
		QU_FATAL ("(%s) error at initialization", cfunc);
	}
}



void qu_thread_throwex (qu_class kls, qu_any msg, qu_any obj, long line)
{
	__QuThreadCurrent->eclass = kls;
	__QuThreadCurrent->emsg   = msg;
	__QuThreadCurrent->ewhere = obj;
	__QuThreadCurrent->eline  = line;
	qu_thread_set_eobject (__QuThreadCurrent);

	if (QU_APP_DEBUG && obj && (msg == 0 || qu_string_is (msg)))
	{
		char  buf [1024];
		char *p   = buf;
		int   rem = sizeof (buf) - 1;

		p = qu__thread_ewhere_string (buf, &rem);
		*p = '\0';

		if (msg == 0)
			msg = qu_exception_defmsg (kls);
		__QuThreadCurrent->emsg = qu_string_newf ("%s: %s", buf, ((qu_string) msg)->str);
	}

	if (QuSysTrace && !__QuCompileIsCompiling)
		qu_thread_print_trace ();
}



int qu_thread_check_external ()
{
	if (qu_thread_specific_get (__QuThreadKey))
	{
		/*
			just assume it's our thread
			.... we could call _gc_is_live_object (thread) but ... naahhh
		*/
		return 1;
	}
	fprintf (stderr, "*** WARNING: external thread entering Qu\n");
	return 0;
}



void qu_thread_state_begin (qu_string key, qu_any obj)
{
	qu_array state = __QuThreadCurrent->state;

	if (state == 0)
		state = __QuThreadCurrent->state = qu_array_new (8);

	/* we are confidence that this always succeed */
	qu_array_make_space (state, 2);
	assert (state->cap >= state->len + 2);

	state->objs[state->len++] = key;
	state->objs[state->len++] = obj;
}



int qu_thread_state_exist (qu_string key, qu_any obj)
{
	qu_thread self = qu_thread_self ();

	if (self && self->state)
	{
		register long    n = self->state->len;
		register qu_any *p = self->state->objs + n - 2;

		while (n)
		{
			assert (n >= 2);
			if (p[0] == key && p[1] == obj)
				return 1;
			p -= 2;
			n -= 2;
		}
	}
	return 0;
}



void qu_thread_state_end (qu_string key, qu_any obj)
{
	qu_thread self = qu_thread_self ();

	if (self && self->state)
	{
		register qu_array a = self->state;
		register long     n = a->len;
		register qu_any  *p = a->objs + n - 2;

		while (n)
		{
			assert (n >= 2);
			if (p[0] == key && p[1] == obj)
			{
				QU_OBJECT_LIST_COPY (p, p + 2, a->len - n);
				a->len -= 2;
				break;
			}
			n -= 2;
			p -= 2;
		}
	}
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_thread__setup ()
{
#ifdef QU_DEBUG_LOCAL
	qu_class_setup_final (QuThreadClass, qu_thread_s, qu__thread__destroy, qu__thread__mark);
#else
	qu_class_setup_final (QuThreadClass, qu_thread_s, 0, qu__thread__mark);
#endif

	if (qu_thread_specific_create (&__QuThreadKey, NULL))
		QU_FATAL ("thread_specific_create failed");
}



void qu_thread__init ()
{
	__QuThreadMain = qu__thread_alloc (qu_thread_id ());
	if (qu_thread_specific_set (__QuThreadKey, (void *) __QuThreadMain) != 0)
		QU_FATAL ("thread_specific_set failed");

	__QuThreadCurrent = __QuThreadMain;

	/* make sure it's never GC'ed */
	qu_class_constz (QuThreadClass, "__mainthread__", __QuThreadMain);

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Thread
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	/* methods */
	qu_class_method_class  (QuThreadClass, QU_NAME_new      , qu__thread_new      , -1);
	qu_class_method_classz (QuThreadClass, "create"         , qu__thread_create   , -1);
	qu_class_method_classz (QuThreadClass, "count"          , qu__thread_count    , -1);
	qu_class_method_classz (QuThreadClass, "joinall"        , qu__thread_joinall  , -1);
	qu_class_method_classz (QuThreadClass, "exit"           , qu__thread_exit     , -1);
	qu_class_method_classz (QuThreadClass, "this"           , qu__thread_this     ,  0);
	qu_class_method_inst   (QuThreadClass, QU_NAME__getitem , qu__thread__getitem ,  1);
	qu_class_method_inst   (QuThreadClass, QU_NAME__getslice, qu__thread__getslice,  2);
	qu_class_method_inst   (QuThreadClass, QU_NAME__setitem , qu__thread__setitem ,  2);
	qu_class_method_instz  (QuThreadClass, "alive"          , qu__thread_alive    ,  0);
	qu_class_method_instz  (QuThreadClass, "data"           , qu__thread_data     , -1);
	qu_class_method_instz  (QuThreadClass, "detach"         , qu__thread_detach   ,  0);
	qu_class_method_instz  (QuThreadClass, "detached"       , qu__thread_detached ,  0);
	qu_class_method_instz  (QuThreadClass, "id"             , qu__thread_id       ,  0);
	qu_class_method_instz  (QuThreadClass, "join"           , qu__thread_join     ,  0);

	/* finalize declaration */
	qu_class_enddecl (QuThreadClass);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Mutex
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	QuMutexClass = qu_class_new_finalz (QuSysModule, "Mutex", qu_mutex_s, 0, 0);

	/* methods */
	qu_class_method_class  (QuMutexClass, QU_NAME_new, qu__mutex_new     ,  0);
	qu_class_method_instz  (QuMutexClass, "lock"     , qu__mutex_lock    , -1);
	qu_class_method_instz  (QuMutexClass, "trylock"  , qu__mutex_trylock ,  0);
	qu_class_method_instz  (QuMutexClass, "unlock"   , qu__mutex_unlock  ,  0);

	/* finalize declaration */
	qu_class_enddecl (QuMutexClass);


	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Cond
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	QuCondClass = qu_class_new_finalz (QuSysModule, "Cond", qu_cond_s, 0, 0);

	/* methods */
	qu_class_method_class  (QuCondClass, QU_NAME_new , qu__cond_new      ,  0);
	qu_class_method_instz  (QuCondClass, "broadcast" , qu__cond_broadcast,  0);
	qu_class_method_instz  (QuCondClass, "count"     , qu__cond_count    ,  0);
	qu_class_method_instz  (QuCondClass, "signal"    , qu__cond_signal   , -1);
	qu_class_method_instz  (QuCondClass, "wait"      , qu__cond_wait     , -1);

	/* finalize declaration */
	qu_class_enddecl (QuCondClass);
}
