/* Tokenizer.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Tokenizer
	A Tokenizer object interprets a source code into a series of tokens.
	The source code must be a -{String} object (thus an ASCII source).

	Note that this Tokenizer is specific to the Qu language. You can use
	it for any purpose as long as the source is something similar with a Qu
	program including the documentation strings, the k{__end__} and
	k{__test__} keywords.

	@xmp
		literals = {}
		source = '
			sub foo (x)
				bar = 1.2
				return x + bar
			;;'

		T = Tokenizer.new (literals, source, 0)

		println ('tokens:')
		while t = T.next
			T.take
			println (t ' = ' Tokenizer.name (t))
		;;

		println
		println ('literals:')
		println (literals)

	@out
		tokens:
		94 = sub
		1 = NAME
		17 = (
		1 = NAME
		18 = )
		1 = NAME
		35 = =
		3 = FLOAT
		132 = return
		1 = NAME
		61 = +
		1 = NAME
		30 = ;;

		literals:
		{1.2, 'foo', 'bar'}
*/
qu_class QuTokenizerClass;

/* list of tokens */
qu_dict QuTokenizerByName;
qu_dict QuTokenizerById;

/* out of list */
static qu_string Sname;
static qu_string Sint;
static qu_string Sfloat;
static qu_string Sstr;
static qu_string Sfield;
static qu_string Sdoc;
static qu_string Stext;
static qu_string Slong;
static qu_string Scomplex;
static qu_string Sindent;
static qu_string Sdedent;

/* throw */
#define RAISEZ(kls, msg) \
QU_BEGIN \
	qu_tokenizer_set_error (self); \
	QU_THROWZ (kls, msg); \
QU_END

#define RAISE(kls) \
QU_BEGIN \
	qu_tokenizer_set_error (self); \
	QU_THROW (kls); \
QU_END



/*-----------------------------------------------------------------------------
	internals
-----------------------------------------------------------------------------*/

static int _tokenizer_skipwhite (qu_tokenizer self)
{
	register char *c = self->cursor;
	register long  l = self->line;

  again:

	while (qu_ascii_isspace (*c))
	{
		if (*c == '\n')
		{
			if (qu_tokenizer_use_indent (self))
			{
				if (c[1] == '\t' || !qu_ascii_isspace (c[1]))
					goto done;
			}
			++l;
		}
		++c;
	}
	if (c[0] == '/' && c[1] == '*')
	{
		c += 2;
		while (*c)
		{
			if (c[0] == '*' && c[1] == '/')
			{
				c += 2;
				goto again;
			}
			if (*c++ == '\n')
				++l;
		}
	}
	else if (*c == '#')
	{
		++c;
		while (*c)
		{
			if (*c == '\n')
				goto again;
			++c;
		}
	}

	if (c[0] == '.' && c[1] == '.' && c[2] == '.' && c[3] == '\n')
	{
		c += 4;
		++l;
		self->token->nobr = 1;
		goto again;
	}

  done:

	self->cursor = c;
	self->line   = l;
	return 1;
}



static int _tokenizer_getoperator (qu_tokenizer self)
{
	register char *c = self->cursor;
	qu_token_s    *t;
	int            p;

	t         = self->last->next;
	t->nobr   = 0;
	t->cursor = c;
	t->line   = self->line;
	t->thing  = 0;

	switch (*c++)
	{
		case '{' : p = QU_T_LCURLY;  break;
		case '}' : p = QU_T_RCURLY;  break;
		case '[' : p = QU_T_LSQUARE; break;
		case ']' : p = QU_T_RSQUARE; break;

		case '(' :
			if (*c == '.')
			{
				c++;
				p = QU_T_LPRINT;
				break;
			}
			else if (*c == '\\')
			{
				c++;
				p = QU_T_LPRINTHTML;
				break;
			}
			p = QU_T_LPAREN;
			break;

		case ')' : p = QU_T_RPAREN;  break;
		case ',' : p = QU_T_COMMA;   break;
		case '?' : p = QU_T_QUEST;   break;
		case '~' : p = QU_T_BITNOT;  break;
		case '@' : p = QU_T_AT;      break;
		case '$' : p = QU_T_DOLLAR;  break;

		case '%':
			if (*c == '%')
			{
				c++;
				p = QU_T_FORMAT;
				break;
			}
			else if (*c == '=')
			{
				c++;
				p = QU_T_ASREM;
				break;
			}
			p = QU_T_REM;
			break;

		case '.' :
			if (*c == '.')
			{
				c++;
				p = QU_T_DOT2;
				break;
			}
			p = QU_T_DOT;
			break;


		case ';' :
			if (*c == ';')
			{
				c++;
				p = QU_T_END;
				break;
			}
			p = QU_T_SEMI;
			break;

		case ':' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASCOLON;
				break;
			}
			if (*c == ':')
			{
				c++;
				p = QU_T_COLON2;
				break;
			}
			p = QU_T_COLON;
			break;

		case '=' :
			if (*c == '=')
			{
				c++;
				if (*c == '=')
				{
					c++;
					p = QU_T_LIKE;
					break;
				}
				p = QU_T_EQ;
				break;
			}
			p = QU_T_ASSIGN;
			break;

		case '!' :
			if (*c == '=')
			{
				c++;
				p = QU_T_NE;
				break;
			}
			p = QU_T_NOT;
			break;

		case '<' :
			if (*c == '=')
			{
				c++;
				if (*c == '>')
				{
					c++;
					p = QU_T_CMP;
					break;
				}
				p = QU_T_LE;
				break;
			}
			if (*c == '>')
			{
				c++;
				p = QU_T_CONTRA;
				break;
			}
			if (*c == '<')
			{
				c++;
				if (*c == '<')
				{
					c++;
					p = QU_T_PUSH;
					break;
				}
				if (*c == '=')
				{
					c++;
					p = QU_T_ASLSHIFT;
					break;
				}
				p = QU_T_LSHIFT;
				break;
			}
			if (*c == '-' && c[1] == '>')
			{
				c += 2;
				p = QU_T_XCHG;
				break;
			}

			if (qu_tokenizer_use_html (self) && *c == '/' && qu_ascii_isalpha (c[1]))
			{
				char buf [64], *q = buf;
				long l = 0;

				++c;
				while (qu_ascii_isalnum (*c))
				{
					*q++ = *c++;
					if (++l > 32)
						RAISEZ (QuESyntax, "html tag too long");
				}
				if (*c != '>')
					RAISEZ (QuESyntax, "html tag '>' missing");
				++c;
				p = QU_T_HTMLCLOSE;
				if (*c == '\n')
					p = QU_T_HTMLCLOSEBR;
				*q = '\0';

				if (qu_tokenizer_is_resident (self))
					t->thing = qu_string_static_dup (buf, q - buf);
				else
					t->thing = qu_set_addif (self->literals, qu_string_new (buf, q - buf));
				break;
			}

			if (qu_tokenizer_use_html (self) && qu_ascii_isalpha (*c))
			{
				char buf [64], *q = buf;
				long l = 0;

				while (qu_ascii_isalnum (*c))
				{
					*q++ = *c++;
					if (++l > 32)
						RAISEZ (QuESyntax, "html tag too long");
				}
				p = QU_T_HTMLOPEN;
				if (*c == '>')
				{
					p = QU_T_HTMLOPENCLOSE;
					++c;
					if (*c == '\n')
						p = QU_T_HTMLOPENCLOSEBR;
				}
				*q = '\0';
				if (qu_tokenizer_is_resident (self))
					t->thing = qu_string_static_dup (buf, q - buf);
				else
					t->thing = qu_set_addif (self->literals, qu_string_new (buf, q - buf));
				break;
			}

			p = QU_T_LT;
			break;

		case '>' :
			if (*c == '=')
			{
				c++;
				p = QU_T_GE;
				break;
			}
			if (*c == '>')
			{
				c++;
				if (*c == '=')
				{
					c++;
					p = QU_T_ASRSHIFT;
					break;
				}
				if (*c == '>')
				{
					c++;
					p = QU_T_POP;
					break;
				}
				p = QU_T_RSHIFT;
				break;
			}

			if (qu_tokenizer_use_html (self)
				&& c > self->source->str
				&& !qu_ascii_isspace (c[-2]))
			{
				p = QU_T_HTMLCLOSE;
				if (*c == '\n')
					p = QU_T_HTMLCLOSEBR;
				break;
			}

			p = QU_T_GT;
			break;

		case '&' :
			if (*c == '&')
			{
				c++;
				if (*c == '=')
				{
					c++;
					p = QU_T_ASAND;
					break;
				}
				p = QU_T_AND;
				break;
			}
			if (*c == '=')
			{
				c++;
				p = QU_T_ASBITAND;
				break;
			}

			if (qu_tokenizer_use_html (self) && qu_ascii_isalpha (*c))
			{
				char buf [32], *q = buf;
				long l = 0;

				while (qu_ascii_isalpha (*c))
				{
					*q++ = *c++;
					if (++l > 16)
						RAISEZ (QuESyntax, "html entity too long");
				}
				if (*c != ';')
					RAISEZ (QuESyntax, "html entity ';' missing");
				++c;
				*q = '\0';
				p = QU_T_HTMLENTITY;
				if (qu_tokenizer_is_resident (self))
					t->thing = qu_string_static_dup (buf, q - buf);
				else
					t->thing = qu_set_addif (self->literals, qu_string_new (buf, q - buf));
				break;
			}

			p = QU_T_BITAND;
			break;

		case '|' :
			if (*c == '|')
			{
				c++;
				if (*c == '=')
				{
					c++;
					p = QU_T_ASOR;
					break;
				}
				p = QU_T_OR;
				break;
			}
			if (*c == '=')
			{
				c++;
				p = QU_T_ASBITOR;
				break;
			}
			p = QU_T_BITOR;
			break;

		case '+' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASPLUS;
				break;
			}
			if (*c == '+')
			{
				c++;
				p = QU_T_INCR;
				break;
			}
			p = QU_T_PLUS;
			break;

		case '-' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASMINUS;
				break;
			}
			if (*c == '-')
			{
				c++;
				if (*c == '-')
				{
					c++;
					p = QU_T_TAG;
					break;
				}
				p = QU_T_DECR;
				break;
			}
			if (*c == '>')
			{
				c++;
				p = QU_T_ARROW;
				break;
			}
			p = QU_T_MINUS;
			break;

		case '_' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASCONCAT;
				break;
			}
			else if (*c == '/')
			{
				c++;
				p = QU_T_RATIO;
				break;
			}
			p = QU_T_CONCAT;
			break;

		case '*' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASMUL;
				break;
			}
			if (*c == '*')
			{
				c++;
				p = QU_T_POW;
				break;
			}
			p = QU_T_MUL;
			break;

		case '/' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASDIV;
				break;
			}

			if (qu_tokenizer_use_html (self) && *c == '>')
			{
				c++;
				t->thing = qu_string_byte ('/');
				p = QU_T_HTMLCLOSE;
				if (*c == '\n')
					p = QU_T_HTMLCLOSEBR;
				break;
			}

			p = QU_T_DIV;
			break;

		case '\\' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASIDIV;
				break;
			}
			p = QU_T_IDIV;
			break;

		case '^' :
			if (*c == '=')
			{
				c++;
				p = QU_T_ASBITXOR;
				break;
			}
			p = QU_T_BITXOR;
			break;

		default:
			RAISEZ (QuESyntax, "unrecognized operator");
	}
	t->len  = c - self->cursor;
	t->type = p;
	self->cursor = c;
	return p;
}



static int _tokenizer_getname (qu_tokenizer self)
{
	register char *c = self->cursor;
	qu_token_s    *t;
	int            p;
	long           l;
	qu_string      s;

	t         = self->last->next;
	t->nobr   = 0;
	t->line   = self->line;
	t->cursor = c;

	while (qu_ascii_isidnum (*c))
		c++;

	l = c - self->cursor;

	/*
		this is a must --- it is required by qu_object and is a sane
		check on untrusted sources.
	*/
	if (l > QU_OBJECT_MAXIDENTLEN)
		RAISEZ (QuESyntax, "identifier name too long");

	if (l > 1)
	{
		p = qu_tokenizer_keyword (t->cursor, l);
		if (p != 0)
		{
			/* recognize only these three in Config mode */
			if (qu_flag_test (self, QU_TOKENIZER_FLAG_CONFIG))
			{
				if (!(p == QU_T_true || p == QU_T_false || p == QU_T_nil))
					goto proceed;
			}

			t->thing = 0;
			goto done;
		}
	}

  proceed:

	p = QU_T_NAME;

	if (qu_tokenizer_is_resident (self))
		s = qu_string_static_dup (t->cursor, l);
	else
	{
		s = qu_string_static_findz (t->cursor, l);
		if (s == 0)
			s = qu_set_addz (self->literals, t->cursor, l);
	}
	t->thing = s;

  done:
	t->type = p;
	t->len  = l;

	self->cursor = c;
	return p;
}



static int _tokenizer_getnumber (qu_tokenizer self)
{
	qu_token_s *t;
	qu_any      x;
	int         p = 0;
	char       *cursor = self->cursor;

	t         = self->last->next;
	t->nobr   = 0;
	t->cursor = cursor;
	t->line   = self->line;

	x = qu_number_scan (&cursor, 0);
	QU_ENSURE (x);

	if (qu_int_is (x))
	{
		p = QU_T_INT;
		t->thing = x;
	}
	else
	{
		/* how ugly! */
		switch (qu_classid (x))
		{
			case QU_CLASS_ID_Long:     p = QU_T_LONG;     break;
			case QU_CLASS_ID_Rational: p = QU_T_RATIONAL; break;
			case QU_CLASS_ID_Float:    p = QU_T_FLOAT;    break;
			case QU_CLASS_ID_Big:      p = QU_T_BIG;      break;
			case QU_CLASS_ID_Inf:      p = QU_T_INF;      break;
			case QU_CLASS_ID_Complex:  p = QU_T_COMPLEX;  break;

			default:
				QU_FATAL ("bug: unexpected number class");
		}

		/* must use qu_set_addif_ex otherwise storing, e.g. a Big and a Float
			which are equal will only store one of them */

		if (qu_tokenizer_is_resident (self))
		{
			switch (qu_classid (x))
			{
				case QU_CLASS_ID_Long:     x = qu_long_to_static     (x); break;
				case QU_CLASS_ID_Rational: x = qu_rational_to_static (x); break;
				case QU_CLASS_ID_Float:    x = qu_float_to_static    (x); break;
				case QU_CLASS_ID_Big:      x = qu_big_to_static      (x); break;
				case QU_CLASS_ID_Inf:                                     break;
				case QU_CLASS_ID_Complex:  x = qu_complex_to_static  (x); break;

				default:
					QU_FATAL ("bug: unexpected number class");
			}
			t->thing = x;
		}
		else
		{
			t->thing = qu_set_addif_ex (self->literals, x);
		}
	}
	t->type = p;
	t->len = cursor - t->cursor;
	self->cursor = cursor;
	return p;
}



static int _tokenizer_getstring (qu_tokenizer self)
{
	register char *c = self->cursor;
	register long  l = self->line;
	char           q = self->cursor[0];
	qu_token_s    *t = self->last->next;
	int            k;
	char          *start;
	long           len;

	t->cursor = c;
	t->nobr   = 0;
	t->line   = l;
	t->type   = QU_T_STRING;

	start = ++c;

	k = 0;
	while (*c)
	{
		switch (*c)
		{
			case '\\':
				if (*++c)
					c++;
				k = 1;
				continue;

			case '\n':
				c++;
				l++;
				continue;

			default:
				if (*c == q)
					break;
				c++;
				continue;
		}
		break;
	}

	if (*c == '\0')
		RAISEZ (QuESyntax, "unclosed string");

	t->len = c - t->cursor + 1;
	len = t->len - 2;

	if (len == 0)
	{
		t->thing = QuStringEMPTY;
	}
	else if (len == 1)
	{
		t->thing = qu_string_byte (start[0]);
	}
	else if (k == 0)
	{
		qu_string s;

		if (qu_tokenizer_is_resident (self))
			s = qu_string_static_dup (start, len);
		else
		{
			s = qu_string_static_findz (start, len);
			if (s == 0)
				s = qu_set_addz (self->literals, start, len);
		}
		t->thing = s;
	}
	else
	{
		qu_string s = qu_string_source (start, len);

		if (s == 0)
			RAISEZ (QuESyntax, "bad string literal");
		if (s->len == 0)
			t->thing = QuStringEMPTY;
		else if (s->len == 1)
			t->thing = qu_string_byte (s->str[0]);
		else
		{
			if (qu_tokenizer_is_resident (self))
				t->thing = qu_string_static_dup (s->str, s->len);
			else
				t->thing = qu_set_addif (self->literals, s);
		}
	}

	self->cursor = c + 1;
	self->line = l;
	return QU_T_STRING;
}



static int _tokenizer_gettick (qu_tokenizer self)
{
	register char *c = self->cursor;
	register long  n = 0;
	qu_token_s    *t = self->last->next;
	qu_string      r = 0;
	register char *z;
	char          *lastc = c + 1;
	int            br = 0;

	t->nobr   = 0;
	t->cursor = c;
	t->line   = self->line;
	t->type   = QU_T_STRING;

	for (;;)
	{
		assert (*c == '`');
		++c;
		while (*c && *c != '`')
		{
			++c;
			if (++n >= QU_MEM_MAX)
				RAISEZ (QuEMemory, "string too big");
		}
		if (*c != '`')
			RAISEZ (QuESyntax, "unclosed string");

		lastc = ++c;

	 blank1:
		while (qu_ascii_isspace (*c))
		{
			if (*c == '\n')
				br = 1;
			++c;
		}
		if (*c == '#')
		{
			while (*c && *c != '\n')
				++c;
			if (*c == '\n')
			{
				++c;
				br = 1;
				goto blank1;
			}
			break;
		}
		else if (*c == '/' && c[1] == '*')
		{
			c += 2;
			while (*c)
			{
				if (*c == '\n')
				{
					br = 1;
				}
				else if (*c == '*' && c[1] == '/')
				{
					c += 2;
					goto blank1;
				}
				++c;
			}
			break;
		}

		if (*c != '`')
			break;

		n += br;
	}

	if (n == 0)
		goto done;

	r = qu_string_newl (n);
	z = r->str;
	c = self->cursor;

	br = 0;
	for (;;)
	{
		assert (*c == '`');
		++c;
		while (*c && *c != '`')
			*z++ = *c++;
		++c;

	 blank2:
		while (qu_ascii_isspace (*c))
		{
			if (*c == '\n')
				br = 1;
			++c;
		}
		if (*c == '#')
		{
			while (*c && *c != '\n')
				++c;
			if (*c == '\n')
			{
				++c;
				br = 1;
				goto blank2;
			}
			break;
		}
		else if (*c == '/' && c[1] == '*')
		{
			c += 2;
			while (*c)
			{
				if (*c == '\n')
				{
					br = 1;
				}
				else if (*c == '*' && c[1] == '/')
				{
					c += 2;
					goto blank2;
				}
				++c;
			}
			break;
		}

		if (*c != '`')
			break;

		if (br)
			*z++ = '\n';
	}

	*z = '\0';
	r->len = z - r->str;

  done:

	if (n == 0)
		r = QuStringEMPTY;
	else if (n == 1)
		r = qu_string_byte (r->str[0]);
	else
	{
		if (qu_tokenizer_is_resident (self))
			r = qu_string_static_dup (r->str, r->len);
		else
			r = qu_set_addif (self->literals, r);
	}

	t->len = lastc - t->cursor;
	t->thing = r;

	/*line*/
	{
		register long l = self->line;
		c = self->cursor;
		while (c != lastc)
		{
			if (*c++ == '\n')
				++l;
		}
		self->line = l;
	}

	self->cursor = lastc;
	return QU_T_STRING;
}



static int _tokenizer_gettext (qu_tokenizer self, char quot, int type)
{
	register char *c = self->cursor;
	register long  l = self->line;
	qu_token_s    *t;
	char          *start;

	t         = self->last->next;
	t->nobr   = 0;
	t->cursor = c;
	t->line   = l;
	t->type   = type;

	c += 3;
	start = c;

	while (*c)
	{
		if (c[0] == quot && c[1] == quot && c[2] == quot)
			break;

		if (*c++ == '\n')
			++l;
	}
	if (*c == '\0')
		RAISEZ (QuESyntax, "unclosed text string");

	t->len = c - t->cursor + 3;

	if (type == QU_T_TEXT)
	{
		long      len = t->len - 6;
		qu_string x   = qu_utf8_new (start, len);

		if (x == 0)
			RAISEZ (QuESyntax, "bad UTF-8 string");

		if (qu_tokenizer_is_resident (self))
			x = qu_string_static_dup (x->str, x->len);
		else
			x = qu_set_addif (self->literals, x);
		t->thing = x;
	}
	else
	{
		if (qu_tokenizer_include_doc (self))
		{
			long len = t->len - 6;
			if (len == 0)
				t->thing = QuStringEMPTY;
			else if (len == 1)
				t->thing = qu_string_byte (start[0]);
			else
			{
				if (qu_tokenizer_is_resident (self))
					t->thing = qu_string_static_dup (start, len);
				else
					t->thing = qu_set_addz (self->literals, start, len);
			}
		}
		else
		{
			t->thing = QuStringEMPTY;
		}
	}

	self->cursor = c + 3;
	self->line = l;
	return type;
}



/*
	0      : error
	-1     : not an indent, just get the next token
	others : is QU_T_DEDENT or QU_T_INDENT
*/
static int _tokenizer_getindent (qu_tokenizer self)
{
	char       *c = self->cursor;
	int         n = 0;
	qu_token_s *t;

	while (*c == '\t')
	{
		c++;
		n++;
	}

	while (*c == ' ')
		++c;

	if (n == self->indent)
	{
		self->cursor = c;
		return -1;
	}

	if (n < self->indent)
	{
		do
		{
			t = self->last->next;

			if (t == self->token)
				/* should never happen */
				RAISEZ (QuESyntax, "too many indentation level");

			t->type   = QU_T_DEDENT;
			t->cursor = c;
			t->line   = self->line;
			t->thing  = 0;

			self->last = t;

		} while (--self->indent > n);

		self->cursor = c;
		return QU_T_DEDENT;
	}

	if (n == self->indent + 1)
	{
		t         = self->last->next;
		t->nobr   = 0;
		t->type   = QU_T_INDENT;
		t->cursor = c;
		t->line   = self->line;
		t->thing  = 0;

		self->cursor = c;
		self->last   = t;
		self->indent++;
		return QU_T_INDENT;
	}

	RAISEZ (QuESyntax, "indentation error");
}



static int _tokenizer_getnext (qu_tokenizer self)
{
	register char *c;
	int p;

  again:

	if (_tokenizer_skipwhite (self) == 0)
		RAISEZ (QuESyntax, "syntax error");

	c = self->cursor;

	if (qu_flag_test (self, QU_TOKENIZER_FLAG_INDENT) && *c == '\n')
	{
		self->cursor++;
		self->line++;
		p = _tokenizer_getindent (self);
		if (p != -1)
			goto done;
		goto again;
	}

	if (*c == '\0')
		return 0;

	if (qu_ascii_isident (*c))
	{
		if (c[0] == '_' && !qu_ascii_isidnum (c[1]))
		{
			p = _tokenizer_getoperator (self);
		}
		else
		{
			p = _tokenizer_getname (self);
			if (p == QU_T__end)
			{
			  theend:
				while (*c)
					c++;
				self->cursor = c;
				return 0;
			}
			if (p == QU_T__test)
			{
				if (QU_APP_INCLUDE_TEST)
					goto again;
				goto theend;
			}
		}
	}
	else if (qu_ascii_isdigit (*c))
	{
		p = _tokenizer_getnumber (self);
	}
	else if (*c == '\'')
	{
		if (c[1] == '\'' && c[2] == '\'')
			p = _tokenizer_gettext (self, '\'', QU_T_TEXT);
		else
			p = _tokenizer_getstring (self);
	}
	else if (*c == '"')
	{
		if (c[1] == '"' && c[2] == '"')
			p = _tokenizer_gettext (self, '"', QU_T_DOC);
		else
			p = _tokenizer_getstring (self);
	}
	else if (*c == '`')
	{
		if (c[1] == '`' && c[2] == '`')
			p = _tokenizer_gettext (self, '`', QU_T_DOC);
		else
			p = _tokenizer_gettick (self);
	}
	else
	{
		p = _tokenizer_getoperator (self);
	}

	if (p != 0)
		self->last = self->last->next;

  done:
#if 0
	fprintf (stderr, "%p ::::: %ld %d %d %s\n", self->last, self->last->line, self->indent, p, qu_tokenizer_name (p)->str);
#endif
	return p;
}



/*-----------------------------------------------------------------------------
	standard class methods
-----------------------------------------------------------------------------*/

/*) sub class __inst (literals: Set, source: String, options = 0) -> Tokenizer __final
	Create a new Tokenizer.

	v{literals} is a mutable -{Set} where to store literals.

	v{source} is the source code.

	Available v{options} values:
	@ul
	.Tokenizer.#{WITHDOC}   = include documentation strings
	.Tokenizer.#{USEINDENT} = use indentation
	.Tokenizer.#{CONFIG}    = Config mode
*/
static qu_tokenizer qu__tokenizer__inst (qu_class self, int argc, qu_any *argv)
{
	qu_tokenizer t;
	qu_set       literals;
	qu_string    source;
	long         options = 0;

	QU_ARG_VCOUNT (2, 3);
	literals = argv[0];
	source   = argv[1];

	QU_ARG_CHECK (qu_set_is (literals) && !qu_set_frozen (literals));
	QU_ARG_CHECK (qu_string_is (source));

	if (argc > 2)
		QU_ARG_vlong (2, options);

	t = qu_tokenizer_new (literals, source, options);
	t->klass = self;
	return t;
}



/*) sub class new (literals: Set, source: String, options: Integer, *arg)
	This is the default f{new} method. It creates a new Tokenizer
	then call the f{__init} method passing all arguments.

	See the #{__inst} method for the arguments.
*/
static qu_tokenizer qu__tokenizer_new (qu_class self, int argc, qu_any *argv)
{
	qu_tokenizer t = qu__tokenizer__inst (self, argc, argv);

	QU_ENSURE (t);
	QU_RETURN (qu_class_init_new (self, t, argc, argv));
}



/*-----------------------------------------------------------------------------
	specific class methods
-----------------------------------------------------------------------------*/

/*) sub class name (id: Integer)
	Return a string representing the name of the specified token v{id}.

	@exc
	.EArgValue
	.ENone not a valid token

	@xmp
		println (Tokenizer.name (123))
*/
static qu_string qu__tokenizer_name (qu_class self, qu_integer i)
{
	int       id;
	qu_string s;

	QU_ARG_int (i, id);
	s = qu_tokenizer_name (id);
	QU_REQUIRE (QuENone, s);
	return s;
}



/*) sub class keyword (s: String) -> Int
	Return token type if v{s} is a keyword, 0 otherwise.

	@exc
	.EArgValue

	@xmp
		println (Tokenizer.keyword ('use'))
*/
static qu_int qu__tokenizer_keyword (qu_class self, qu_string s)
{
	QU_ARG_TYPE (s, qu_string);
	return qu_int_new (qu_tokenizer_keyword (s->str, (size_t) s->len));
}



/*) sub class list (byname = true) -> Dict
	Return a -{Dict} object containing the tokens.

	If v{byname} is true return the Dict which keys are the names,
	otherwise the keyws are the ids.

	@exc
	.EArgValue
*/
static qu_dict qu__tokenizer_list (qu_class self, int argc, qu_any *argv)
{
	int byname = 1;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, byname);

	return byname ? QuTokenizerByName : QuTokenizerById;
}



/*) sub class listkey (byname = true) -> Dict
	Return a -{Dict} object containing the keywords.

	If v{byname} is true return the Dict which keys are the names,
	otherwise the keyws are the ids.

	@exc
	.EArgValue
*/
#define QU_TOKENIZER_FIRST_KEYWORD  QU_T_as
#define QU_TOKENIZER_FIRST_OPERATOR QU_T_LCURLY

static qu_dict qu__tokenizer_listkey (qu_class self, int argc, qu_any *argv)
{
	int     byname = 1;
	qu_dict r;
	long    b, i;
	qu_any  k, v;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, byname);

	r = qu_dict_new (128);
	b = i = 0;
	while ((v = qu_dict_walk (QuTokenizerByName, &k, &b, &i)) != 0)
	{
		long p = qu_int_tol (v);

		if (p < QU_TOKENIZER_FIRST_KEYWORD)
			continue;

		if (byname)
			qu_dict_setitem (r, k, v);
		else
			qu_dict_setitem (r, v, k);
	}
	return r;
}



/*) sub class listop (byname = true) -> Dict
	Return a -{Dict} object containing the operators.

	If v{byname} is true return the Dict which keys are the names,
	otherwise the keyws are the ids.

	@exc
	.EArgValue
*/
static qu_dict qu__tokenizer_listop (qu_class self, int argc, qu_any *argv)
{
	int     byname = 1;
	qu_dict r;
	long    b, i;
	qu_any  k, v;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, byname);

	r = qu_dict_new (128);
	b = i = 0;
	while ((v = qu_dict_walk (QuTokenizerByName, &k, &b, &i)) != 0)
	{
		long p = qu_int_tol (v);

		if (p < QU_TOKENIZER_FIRST_OPERATOR || p >= QU_TOKENIZER_FIRST_KEYWORD)
			continue;

		if (byname)
			qu_dict_setitem (r, k, v);
		else
			qu_dict_setitem (r, v, k);
	}
	return r;
}



/*-----------------------------------------------------------------------------
	low-level instance methods
-----------------------------------------------------------------------------*/

static void qu__tokenizer__mark (qu_tokenizer self)
{
	qu_gc_mark (self->source);
	qu_gc_mark (self->literals);
	/* no need to mark the tokens since the only object is 'thing'
		and that is kept in the 'literals' Set */
}



/*-----------------------------------------------------------------------------
	standard instance methods
-----------------------------------------------------------------------------*/

/*) sub __each -> Iterator
	Get an -{Iterator} object to iterate through all tokens.

	@noexc
*/
static qu_int qu__tokenizer__next (qu_tokenizer self, qu_iterator i)
{
	int t = qu_tokenizer_next (self);

	QU_ITERATOR_STOP (t);

	qu_tokenizer_take (self);
	return qu_int_new (t);
}



static qu_iterator qu__tokenizer__each (qu_tokenizer self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, qu__tokenizer__next);
	return i;
}



/*-----------------------------------------------------------------------------
	specific instance methods
-----------------------------------------------------------------------------*/

/*) sub cursor -> Int
	Return the cursor at the current token.

	@exc
	.ENone there is no current token
*/
static qu_int qu__tokenizer_cursor (qu_tokenizer self)
{
	QU_REQUIRE (QuENone, self->cursor != self->source->str);
	return qu_int_new (self->token->cursor - self->source->str);
}



/*) sub eos -> Boolean
	Return true if the cursor is at the end of the source, false
	otherwise.

	@noexc
*/
static qu_boolean qu__tokenizer_eos (qu_tokenizer self)
{
	return self->cursor[0] ? QuFalse : QuTrue;
}



/*) sub line -> Int
	Return the line number at the current token.

	@exc
	.ENone there is no current token
*/
static qu_int qu__tokenizer_line (qu_tokenizer self)
{
	QU_REQUIRE (QuENone, self->cursor != self->source->str);
	return qu_int_new (self->token->line);
}



/*) sub match (type: Integer) -> Boolean
	Return true if the next token is of the specified v{type}, false
	otherwise. If it is then make it current.

	If the end of the source is reached then false is returned.

	@exc
	.EArgValue
*/
static qu_boolean qu__tokenizer_match (qu_tokenizer self, qu_integer x)
{
	int type;

	QU_ARG_int (x, type);
	return qu_tokenizer_match (self, type) ? QuTrue : QuFalse;
}



/*) sub need (type: Integer) -> Null
	Ensure that the next token is of the specified v{type}. If it is then
	make it current, otherwise throw ESyntax.

	@exc
	.EArgValue
	.ESyntax
*/
static qu_null qu__tokenizer_need (qu_tokenizer self, qu_integer type)
{
	int tp;

	QU_ARG_int (type, tp);
	QU_ENSURE (qu_tokenizer_need (self, tp));
	return QuNil;
}



/*) sub next -> Int
	Return the next token type (but does not make it current).

	If the end of the source is reached, returns 0.

	@noexc
*/
static qu_int qu__tokenizer_next (qu_tokenizer self)
{
	if (self->token == self->last)
		return qu_int_new (_tokenizer_getnext (self));
	return qu_int_new (self->token->next->type);
}



/*) sub nobr -> Boolean
	Return true if the next token is on the same line with the current
	token, false otherwise.

	If the end of the source is reached then false is returned.

	@noexc
*/
static qu_boolean qu__tokenizer_nobr (qu_tokenizer self)
{
	return qu_tokenizer_nobr (self) ? QuTrue : QuFalse;
}



/*) sub peek (n: Integer) -> Int
	Return the v{n}-th token type ahead.

	v{n} must be within C{1..Tokenizer.}#{MAXPEEK}.

	@exc
	.EArgValue
*/
static qu_int qu__tokenizer_peek (qu_tokenizer self, qu_integer nn)
{
	int n;

	QU_ARG_int (nn, n);
	QU_ARG_CHECK (0 < n && n <= QU_TOKENIZER_MAXPEEK);
	return qu_int_new (qu_tokenizer_peek (self, n));
}



/*) sub reset -> Null
	Reset the cursor position at the start of the source.

	@noexc
*/
static qu_null qu__tokenizer_reset (qu_tokenizer self)
{
	qu_tokenizer_reset (self, self->literals, self->source, -1);
	return QuNil;
}



/*) sub source -> String
	Return the source string.

	@noexc
*/
static qu_string qu__tokenizer_source (qu_tokenizer self)
{
	return self->source;
}



/*) sub take -> Null
	Make the next token current.

	If there is no next token then ENone is thrown so make sure to
	call #{next} before calling this method.

	@exc
	.ENone
*/
static qu_null qu__tokenizer_take (qu_tokenizer self)
{
	QU_REQUIRE (QuENone, self->token != self->last);
	self->token = self->token->next;
	return QuNil;
}



/*) sub thing -> Any
	Get the thing associated with the token.

	An STR token will return a -{String} object,
	an INT token will return an -{Int} object.
	etc.

	Keywords and operators do not have a 'thing'.

	@warn
	Make sure you have a current token otherwise you get an arbitrary
	result. No checks are made.

	@exc
	.ENone
*/
static qu_any qu__tokenizer_thing (qu_tokenizer self)
{
	QU_REQUIRE (QuENone, self->cursor != self->source->str && self->token->thing);
	return self->token->thing;
}



/*) sub type -> Int
	Get the type of the current token.

	@warn
	Make sure you have a current token otherwise you get an arbitrary
	result. No checks are made.

	@noexc
*/
static qu_any qu__tokenizer_type (qu_tokenizer self)
{
	return qu_int_new (self->last->type);
}



/*) sub toklen -> Int
	Get the length of the current token.

	@warn
	Make sure you have a current token otherwise you get an arbitrary
	result. No checks are made.

	@noexc
*/
static qu_int qu__tokenizer_toklen (qu_tokenizer self)
{
	return qu_int_new (self->last->len);
}



/*) enum TokenType
		NAME     # identifier
		INT      # Int value
		FLOAT    # Float value
		STR      # String value from 'xxx' or "xxx"
		DOC      # String documentation from ```xxx``` or """xxx"""
		TEXT     # String value from `xxx`
		LONG     # Long value
		COMPLEX  # Complex value
		INDENT   # indentation

		# ... etc. see the Tokens.h file in the include directory
	;;
	List of token types.

	@exc
	.ENone
*/

static void _init_token_type ()
{
#define t_(name, strz) #name, QU_T_##name,
	qu_var_enum
	(
		(qu_symbol) QuTokenizerClass, "TokenType",
#include "Tokens.h"
		NULL, NULL
	);
#undef t_
};



/*) const MAXPEEK = 7
	The maximum number of look-ahead tokens.

	@warn
	The actual value is set by Qu.
	It is not necessarily 7.
*/


/*) const WITHDOC = 1
	Whether or not to include documentation strings. If not then all
	documentation strings will be treated as empty strings.

	You use this as an option to the #{new} method.

	@warn
	The actual value is set by Qu.
	It is not necessarily 1.
*/


/*) const USEINDENT = 2
	Whether or not to return indentations as INDENT tokens.

	You use this as an option to the #{new} method.

	An indentation is a series of tabs after a linebreak.
	The INDENT token will have #{len} = the number of tabs.

	@warn
	The actual value is set by Qu.
	It is not necessarily 2.

	This Tokenizer only recognizes tabs as indent characters.
*/


/*) const CONFIG = 4
	For Qu style configuration files.

	If this is set then the Tokenizer will only recognize three keywords:
	true, false and nil.

	@warn
	The actual value is set by Qu.
	It is not necessarily 4.
*/




/*-----------------------------------------------------------------------------
	interface
-----------------------------------------------------------------------------*/

qu_tokenizer qu_tokenizer_new (qu_set literals, qu_string source, long op)
{
	qu_tokenizer T;
	qu_token_s  *t, *first;
	int          i;

	T = qu_gc_new (QuTokenizerClass);
	first = t = T->tokbuf;
	for (i = 0; i < QU_TOKENIZER_MAXPEEK; i++, t++)
		t->next = t + 1;
	t->next = first;
	qu_tokenizer_reset (T, literals, source, op);
	return T;
}



qu_tokenizer qu_tokenizer_new_root (qu_set literals, qu_string source, long op)
{
	qu_tokenizer T;
	qu_token_s  *t, *first;
	int          i;

	T = qu_gc_new_root (QuTokenizerClass);
	first = t = T->tokbuf;
	for (i = 0; i < QU_TOKENIZER_MAXPEEK; i++, t++)
		t->next = t + 1;
	t->next = first;
	qu_tokenizer_reset (T, literals, source, op);
	return T;
}



void qu_tokenizer_reset (qu_tokenizer self, qu_set literals, qu_string source, long op)
{
	assert (!qu_set_frozen (literals));

	self->literals = literals;
	self->source   = source;
	self->cursor   = source->str;
	self->line     = 1;
	self->last     = self->token = self->tokbuf;
	self->indent   = 0;

	qu_flag_unset (self, QU_TOKENIZER_FLAG_ERROR);

	if (op >= 0)
	{
		if (op & QU_TOKENIZER_FLAG_DOC)
			qu_tokenizer_set_include_doc (self);
		else
			qu_tokenizer_unset_include_doc (self);

		if (op & QU_TOKENIZER_FLAG_INDENT)
			qu_tokenizer_set_use_indent (self);
		else
			qu_tokenizer_unset_use_indent (self);

		if (op & QU_TOKENIZER_FLAG_HTML)
			qu_tokenizer_set_use_html (self);
		else
			qu_tokenizer_unset_use_html (self);

		if (op & QU_TOKENIZER_FLAG_CONFIG)
			qu_flag_set (self, QU_TOKENIZER_FLAG_CONFIG);
		else
			qu_flag_unset (self, QU_TOKENIZER_FLAG_CONFIG);
	}
}



int qu_tokenizer_eos (qu_tokenizer self)
{
	return self->cursor[0] == '\0';
}



int qu_tokenizer_peek (qu_tokenizer self, int n)
{
	qu_token_s *t;

	assert (0 < n && n <= QU_TOKENIZER_MAXPEEK);
	t = self->token;
	while (n > 0)
	{
		if (t == self->last)
			QU_ENSURE (_tokenizer_getnext (self));
		t = t->next;
		n--;
	}
	return t->type;
}



void qu_tokenizer_take (qu_tokenizer self)
{
	assert (self->token != self->last);
	self->token = self->token->next;
}



int qu_tokenizer_match (qu_tokenizer self, int type)
{
	if (self->token == self->last)
	{
		if (_tokenizer_getnext (self) != type)
			return 0;
	}
	else
	{
		if (self->token->next->type != type)
			return 0;
	}
	self->token = self->token->next;
	return 1;
}



int qu_tokenizer_next (qu_tokenizer self)
{
	if (self->token == self->last)
		return _tokenizer_getnext (self);
	return self->token->next->type;
}



int qu_tokenizer_need (qu_tokenizer self, int type)
{
	if (self->token == self->last)
	{
		if (_tokenizer_getnext (self) != type)
			RAISEZ (QuESyntax, "unexpected token or end of file");
	}
	else
	{
		if (self->token->next->type != type)
			RAISEZ (QuESyntax, "unexpected token");
	}
	self->token = self->token->next;
	return 1;
}



int qu_tokenizer_nobr (qu_tokenizer self)
{
	if (self->token == self->last)
	{
		if (_tokenizer_getnext (self) == 0)
			return 0;
	}
	return self->token->nobr || self->token->line == self->token->next->line;
}



/*
	Return token type if 's' is a keyword, 0 otherwise.
	Ugly but faster than using a hash table because we dont need to
	calculate a hash value.
*/
int qu_tokenizer_keyword (const char *s, size_t len)
{
#define K2(t, c0, c1) \
	if (s[0] == c0  \
	&&  s[1] == c1) \
	return t

#define K3(t, c0, c1, c2) \
	if (s[0] == c0  \
	&&  s[1] == c1  \
	&&  s[2] == c2) \
	return t

#define K4(t, c0, c1, c2, c3) \
	if (s[0] == c0  \
	&&  s[1] == c1  \
	&&  s[2] == c2  \
	&&  s[3] == c3) \
	return t

#define K5(t, c0, c1, c2, c3, c4) \
	if (s[0] == c0  \
	&&  s[1] == c1  \
	&&  s[2] == c2  \
	&&  s[3] == c3  \
	&&  s[4] == c4) \
	return t

#define K6(t, c0, c1, c2, c3, c4, c5) \
	if (s[0] == c0  \
	&&  s[1] == c1  \
	&&  s[2] == c2  \
	&&  s[3] == c3  \
	&&  s[4] == c4  \
	&&  s[5] == c5) \
	return t

#define K7(t, c0, c1, c2, c3, c4, c5, c6) \
	if (s[0] == c0  \
	&&  s[1] == c1  \
	&&  s[2] == c2  \
	&&  s[3] == c3  \
	&&  s[4] == c4  \
	&&  s[5] == c5  \
	&&  s[6] == c6) \
	return t

#define K8(t, c0, c1, c2, c3, c4, c5, c6, c7) \
	if (s[0] == c0  \
	&&  s[1] == c1  \
	&&  s[2] == c2  \
	&&  s[3] == c3  \
	&&  s[4] == c4  \
	&&  s[5] == c5  \
	&&  s[6] == c6  \
	&&  s[7] == c7) \
	return t

	switch (len)
	{
		case 2:
			K2 (QU_T_as, 'a','s');
			K2 (QU_T_do, 'd','o');
			K2 (QU_T_if, 'i','f');
			K2 (QU_T_in, 'i','n');
			K2 (QU_T_is, 'i','s');
			K2 (QU_T_me, 'm','e');
			K2 (QU_T_of, 'o','f');
			K2 (QU_T_or, 'o','r');
			break;

		case 3:
			K3 (QU_T_and, 'a','n','d');
			K3 (QU_T_are, 'a','r','e');
			K3 (QU_T_end, 'e','n','d');
			K3 (QU_T_for, 'f','o','r');
			K3 (QU_T_nil, 'n','i','l');
			K3 (QU_T_not, 'n','o','t');
			K3 (QU_T_sub, 's','u','b');
			K3 (QU_T_try, 't','r','y');
			K3 (QU_T_use, 'u','s','e');
			K3 (QU_T_var, 'v','a','r');
			K3 (QU_T_xor, 'x','o','r');
			break;

		case 4:
			K4 (QU_T_case, 'c','a','s','e');
			K4 (QU_T_each, 'e','a','c','h');
			K4 (QU_T_elif, 'e','l','i','f');
			K4 (QU_T_else, 'e','l','s','e');
			K4 (QU_T_enum, 'e','n','u','m');
			K4 (QU_T_infn, 'i','n','f','n');
			K4 (QU_T_infp, 'i','n','f','p');
			K4 (QU_T_loop, 'l','o','o','p');
			K4 (QU_T_null, 'n','u','l','l');
			K4 (QU_T_pass, 'p','a','s','s');
			K4 (QU_T_redo, 'r','e','d','o');
			K4 (QU_T_self, 's','e','l','f');
			K4 (QU_T_true, 't','r','u','e');
			K4 (QU_T_with, 'w','i','t','h');
			break;

		case 5:
			K5 (QU_T_alias, 'a','l','i','a','s');
			K5 (QU_T_break, 'b','r','e','a','k');
			K5 (QU_T_catch, 'c','a','t','c','h');
			K5 (QU_T_class, 'c','l','a','s','s');
			K5 (QU_T_const, 'c','o','n','s','t');
			K5 (QU_T_false, 'f','a','l','s','e');
			K5 (QU_T_local, 'l','o','c','a','l');
			K5 (QU_T_super, 's','u','p','e','r');
			K5 (QU_T_throw, 't','h','r','o','w');
			K5 (QU_T_until, 'u','n','t','i','l');
			K5 (QU_T_while, 'w','h','i','l','e');
			K5 (QU_T_yield, 'y','i','e','l','d');
			K5 (QU_T__std , '_','_','s','t','d');
			break;

		case 6:
			K6 (QU_T_assert, 'a','s','s','e','r','t');
			K6 (QU_T_atomic, 'a','t','o','m','i','c');
			K6 (QU_T_ensure, 'e','n','s','u','r','e');
			K6 (QU_T_import, 'i','m','p','o','r','t');
			K6 (QU_T_lambda, 'l','a','m','b','d','a');
			K6 (QU_T_pragma, 'p','r','a','g','m','a');
			K6 (QU_T_return, 'r','e','t','u','r','n');
			K6 (QU_T_static, 's','t','a','t','i','c');
			K6 (QU_T_struct, 's','t','r','u','c','t');
			K6 (QU_T_switch, 's','w','i','t','c','h');
			K6 (QU_T_unless, 'u','n','l','e','s','s');
			K6 (QU_T__args,  '_','_','a','r','g','s');
			K6 (QU_T__auto,  '_','_','a','u','t','o');
			K6 (QU_T__once,  '_','_','o','n','c','e');
			K6 (QU_T__pure,  '_','_','p','u','r','e');
			break;

		case 7:
			K7 (QU_T_default, 'd','e','f','a','u','l','t');
			K7 (QU_T_finally, 'f','i','n','a','l','l','y');
			K7 (QU_T__end   , '_','_','e','n','d','_','_');
			K7 (QU_T__doc   , '_','_','d','o','c','_','_');
			K7 (QU_T__final , '_','_','f','i','n','a','l');
			K7 (QU_T__mixin , '_','_','m','i','x','i','n');
			break;

		case 8:
			K8 (QU_T_continue, 'c','o','n','t','i','n','u','e');
			K8 (QU_T_universe, 'u','n','i','v','e','r','s','e');
			K8 (QU_T__argc   , '_','_','a','r','g','c','_','_');
			K8 (QU_T__line   , '_','_','l','i','n','e','_','_');
			K8 (QU_T__file   , '_','_','f','i','l','e','_','_');
			K8 (QU_T__here   , '_','_','h','e','r','e','_','_');
			K8 (QU_T__swap   , '_','_','s','w','a','p','_','_');
			K8 (QU_T__test   , '_','_','t','e','s','t','_','_');
			K8 (QU_T__public , '_','_','p','u','b','l','i','c');
			K8 (QU_T__unsafe , '_','_','u','n','s','a','f','e');
			K8 (QU_T__extern , '_','_','e','x','t','e','r','n');
			break;

		case 9:
			if (strncmp (s, "__frame__", 9) == 0)
				return QU_T__frame;
			else if (strncmp (s, "__private", 9) == 0)
				return QU_T__private;
			else if (strncmp (s, "__package", 9) == 0)
				return QU_T__package;
			break;

		case 10:
			if (strncmp (s, "__abstract", 10) == 0)
				return QU_T__abstract;
			else if (strncmp (s, "__readonly", 10) == 0)
				return QU_T__readonly;
			break;

		case 11:
			if (strncmp (s, "__generator", 11) == 0)
				return QU_T__generator;
			else if (strncmp (s, "__readwrite", 11) == 0)
				return QU_T__readwrite;
			else if (strncmp (s, "__interface", 11) == 0)
				return QU_T__interface;
			break;

		case 13:
			if (strncmp (s, "__accumulator", 13) == 0)
				return QU_T__accumulator;
			else if (strncmp (s, "__uncatchable", 13) == 0)
				return QU_T__uncatchable;
			break;
	}
	return 0;
}



qu_string qu_tokenizer_name (int id)
{
	qu_string s;

	switch (id)
	{
		case 0: goto nope;

		case QU_T_NAME:    return Sname;
		case QU_T_INT:     return Sint;
		case QU_T_FLOAT:   return Sfloat;
		case QU_T_STRING:  return Sstr;
		case QU_T_FIELD:   return Sfield;
		case QU_T_DOC:     return Sdoc;
		case QU_T_TEXT:    return Stext;
		case QU_T_LONG:    return Slong;
		case QU_T_COMPLEX: return Scomplex;
 		case QU_T_INDENT:  return Sindent;
 		case QU_T_DEDENT:  return Sdedent;
	}
	s = qu_dict_getitem (QuTokenizerById, qu_int_new (id));
	if (s)
	{
		assert (qu_string_is (s));
		return s;
	}
	else
	{
		nope:
		{
			static qu_string none = 0;
			if (none == 0)
				none = qu_string_new_staticz ("*?*");
			return none;
		}
	}
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

void qu_tokenizer__setup ()
{
	qu_class_setup (QuTokenizerClass, qu_tokenizer_s, 0, qu__tokenizer__mark);
}



void qu_tokenizer__init ()
{
	qu_string name;

	Sname    = qu_string_new_staticz ("NAME");
	Sint     = qu_string_new_staticz ("INT");
	Sfloat   = qu_string_new_staticz ("FLOAT");
	Sstr     = qu_string_new_staticz ("STR");
	Sfield   = qu_string_new_staticz ("FIELD");
	Sdoc     = qu_string_new_staticz ("DOC");
	Stext    = qu_string_new_staticz ("TEXT");
	Slong    = qu_string_new_staticz ("LONG");
	Scomplex = qu_string_new_staticz ("COMPLEX");
	Sindent  = qu_string_new_staticz ("INDENT");
	Sdedent  = qu_string_new_staticz ("DEDENT");

	QuTokenizerByName = qu_dict_new_static (511);
	QuTokenizerById   = qu_dict_new_static (511);

#define QU_TOKENIZER_IN_INIT
#define t_(id, strz) \
	name = qu_string_new_staticz (strz); \
	qu_dict_setitem (QuTokenizerByName, name, qu_int_new (QU_T_##id)); \
	qu_dict_setitem (QuTokenizerById, qu_int_new (QU_T_##id), name);
#include "Tokens.h"
#undef QU_TOKENIZER_IN_INIT
#undef t_

	qu_dict_freeze (QuTokenizerByName);
	qu_dict_freeze (QuTokenizerById);

	_init_token_type ();

	/* class properties */
	qu_class_constz (QuTokenizerClass, "MAXPEEK"  , qu_int_new (QU_TOKENIZER_MAXPEEK));
	qu_class_constz (QuTokenizerClass, "WITHDOC"  , qu_int_new (QU_TOKENIZER_FLAG_DOC));
	qu_class_constz (QuTokenizerClass, "USEINDENT", qu_int_new (QU_TOKENIZER_FLAG_INDENT));
	qu_class_constz (QuTokenizerClass, "USEHTML"  , qu_int_new (QU_TOKENIZER_FLAG_HTML));
	qu_class_constz (QuTokenizerClass, "CONFIG"   , qu_int_new (QU_TOKENIZER_FLAG_CONFIG));

	/* standard class methods */
	qu_class_method_class  (QuTokenizerClass, QU_NAME__inst, qu__tokenizer__inst, -1);
	qu_class_method_class  (QuTokenizerClass, QU_NAME_new  , qu__tokenizer_new  , -1);

	/* specific class methods */
	qu_class_method_classz (QuTokenizerClass, "name"   , qu__tokenizer_name    ,  1);
	qu_class_method_classz (QuTokenizerClass, "keyword", qu__tokenizer_keyword ,  1);
	qu_class_method_classz (QuTokenizerClass, "list"   , qu__tokenizer_list    , -1);
	qu_class_method_classz (QuTokenizerClass, "listkey", qu__tokenizer_listkey , -1);
	qu_class_method_classz (QuTokenizerClass, "listop" , qu__tokenizer_listop  , -1);

	/* standard instance methods */
	qu_class_method_inst  (QuTokenizerClass, QU_NAME__each, qu__tokenizer__each, 0);

	/* specific instance methods */
	qu_class_method_instz (QuTokenizerClass, "cursor", qu__tokenizer_cursor, 0);
	qu_class_method_instz (QuTokenizerClass, "eos"   , qu__tokenizer_eos   , 0);
	qu_class_method_instz (QuTokenizerClass, "match" , qu__tokenizer_match , 1);
	qu_class_method_instz (QuTokenizerClass, "need"  , qu__tokenizer_need  , 1);
	qu_class_method_instz (QuTokenizerClass, "next"  , qu__tokenizer_next  , 0);
	qu_class_method_instz (QuTokenizerClass, "nobr"  , qu__tokenizer_nobr  , 0);
	qu_class_method_instz (QuTokenizerClass, "line"  , qu__tokenizer_line  , 0);
	qu_class_method_instz (QuTokenizerClass, "peek"  , qu__tokenizer_peek  , 1);
	qu_class_method_instz (QuTokenizerClass, "reset" , qu__tokenizer_reset , 0);
	qu_class_method_instz (QuTokenizerClass, "source", qu__tokenizer_source, 0);
	qu_class_method_instz (QuTokenizerClass, "take"  , qu__tokenizer_take  , 0);
	qu_class_method_instz (QuTokenizerClass, "toklen", qu__tokenizer_toklen, 0);
	qu_class_method_instz (QuTokenizerClass, "thing" , qu__tokenizer_thing , 0);
	qu_class_method_instz (QuTokenizerClass, "type"  , qu__tokenizer_type  , 0);

	/* finalize declaration */
	qu_class_enddecl (QuTokenizerClass);
}
