/* Utf8.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/* signed char to unsigned char */
#define UCHR(c) (255 & (c))


int qu_utf8_charlen (wint_t c)
{
	if (c <= 0x7F      ) return 1;
	if (c <= 0x7FF     ) return 2;
	if (c <= 0xFFFF    ) return 3;
	if (c <= 0x1FFFFF  ) return 4;
	if (c <= 0x3FFFFFF ) return 5;
	return 6;
}



/*
	Put one wide character into a UTF-8 string.
	!!! make sure there is enough room.
*/
char *qu_utf8_putchar (char *s, wint_t c)
{
	unsigned char *p = (unsigned char *) s;

	if (c <= 0x7F)
	{
		*p++ = (unsigned char) c;
	}
	else if (c <= 0x7FF)
	{
		*p++ = (0xC0 | (c >> 6));
		*p++ = (0x80 | (c & 0x3F));
	}
	else if (c <= 0xFFFF)
	{
		*p++ = (0xE0 | (c >> 12));
		*p++ = (0x80 | ((c >> 6) & 0x3F));
		*p++ = (0x80 | (c & 0x3F));
	}
	else if (c <= 0x1FFFFF)
	{
		*p++ = (0xF0 | (c >> 18));
		*p++ = (0x80 | ((c >> 12) & 0x3F));
		*p++ = (0x80 | ((c >> 6) & 0x3F));
		*p++ = (0x80 | (c & 0x3F));
	}
	else if (c <= 0x3FFFFFF)
	{
		*p++ = (0xF8 | (c >> 24));
		*p++ = (0x80 | ((c >> 18) & 0x3F));
		*p++ = (0x80 | ((c >> 12) & 0x3F));
		*p++ = (0x80 | ((c >> 6) & 0x3F));
		*p++ = (0x80 | (c & 0x3F));
	}
	else if (c <= 0x7FFFFFFF)
	{
		*p++ = (0xFC | (c >> 30));
		*p++ = (0x80 | ((c >> 24) & 0x3F));
		*p++ = (0x80 | ((c >> 18) & 0x3F));
		*p++ = (0x80 | ((c >> 12) & 0x3F));
		*p++ = (0x80 | ((c >> 6) & 0x3F));
		*p++ = (0x80 | (c & 0x3F));
	}
	else
	{
		QU_FATAL ("never");
	}
	return (char *) p;
}




/*
	Read wide character.
	!!! s must be within the string
	!!! e must point to the last byte
	!!! return -1 if invalid.
*/
wint_t qu_utf8_getchar (char **s, char *e)
{
	unsigned char *p = *((unsigned char **) s);
	wint_t c;
	int    n;

	c = (wint_t) *p++;
	if (c > 0x7F)
	{
		if ((c & 0xC0) == 0xC0)
		{
			c &= ~0xC0;
			n = 2;
		}
		else if ((c & 0xE0) == 0xE0)
		{
			c &= ~0xE0;
			n = 3;
		}
		else if ((c & 0xF0) == 0xF0)
		{
			c &= ~0xF0;
			n = 4;
		}
		else if ((c & 0xF8) == 0xF8)
		{
			c &= ~0xF8;
			n = 5;
		}
		else
		{
			assert ((c & 0xFC) == 0xFC);
			c &= ~0xFC;
			n = 6;
		}

		if (p + n >= (unsigned char *) e)
			return QU_WCHAR_INVALID;

		while (--n)
		{
			wint_t w = (wint_t) *p++;
			c = ((c << 6) | (w & 0x3F));
		}
	}
	*s = (char *) p;
	return c;
}



/*
	Read wide character.
	!!! assumes vaild pointer and UTF-8 string
*/
wint_t qu_utf8_charat (char *s)
{
	return qu_utf8_getchar (&s, s + 8);
}



/*
	Create a new UTF-8 string.
*/
qu_string qu_utf8_new (const char *z, long len)
{
	/*
		volatile: stupid gcc bug optimizes the loop which result in the loop
		never entered.
	*/

	volatile char *p;
	char     *q, *e;
	long      l = 0;
	long      n;
	wint_t    c;
	qu_string t;

	if (len == 0)
		return QuStringEMPTY;

	p = (char *) z;
	e = (char *) p + len - 1;
	n = len;
	while ((char *) p <= e)
	{
		c = qu_utf8_getchar ((char **) &p, e);
		if (c == QU_WCHAR_INVALID)
			QU_THROWZ (QuESyntax, "not a valid UTF-8 string");
		l += qu_utf8_charlen (c);
	}
	t = qu_string_newl (l);
	q = t->str;
	n = len;
	p = (char *) z;
	while (p <= e)
	{
		c = qu_utf8_getchar ((char **) &p, e);
		q = qu_utf8_putchar (q, c);
	}
	return t;
}



/* allocate a single UTF-8 character string */
qu_string qu_utf8_new_char (wint_t c)
{
	if (c < 0x80)
	{
		return qu_string_byte (c);
	}
	else
	{
		qu_string r = qu_string_newl (qu_utf8_charlen (c));
		r->flags |= QU_STRING_FLAG_UTF8;
		qu_utf8_putchar (r->str, c);
		return r;
	}
}



/*
	Count the number of wide characters in the string.
*/
long qu_utf8_width (const char *z, long l)
{
	volatile char *p = (char *) z;
	char *e = (char *) p + l - 1;
	long  w = 0;

	while (p <= e)
	{
		wint_t c = qu_utf8_getchar ((char **) &p, e);
		if (c == QU_WCHAR_INVALID)
			return -1;
		w++;
	}
	return w;
}



/*
	Get the position of the i-th character, i may be negative
*/
char *qu_utf8_pos (const char *z, long l, long i)
{
	volatile char *p = (char *) z;

	if (i >= 0)
	{
		char *e = (char *) p + l;
		while (i-- && (char *) p < e)
		{
			if (qu_utf8_getchar ((char **) &p, e) == QU_WCHAR_INVALID)
				return 0;
		}
		if (p >= e)
			return 0;
		return (char *) p;
	}
	else
	{
		char *e = (char *) p + l - 1;
		while (i++ && e > (char *) p)
		{
			if (UCHR (*e) <= 0x7F)
				--e;
			else
			{
				while ((0x80 <= UCHR (*e) && UCHR (*e) <= 0xbf))
					--e;
			}
		}
		if (i < 0)
			return 0;
		return e;
	}
}



/*
	Get the position of the last character.
	!!! assuming a valid UTF-8 string.
*/
char *qu_utf8_last (const char *z, long l)
{
	if (l < 1)
	{
		return 0;
	}
	else
	{
		z += (l - 1);
		while ((0x80 <= UCHR (*z) && UCHR (*z) <= 0xbf))
			--z;
		return (char *) z;
	}
}



/*
	return a qu_string of the i-th character
	i may be negative
*/
qu_string qu_utf8_getitem (qu_string self, long i)
{
	char  *p = qu_utf8_pos (self->str, self->len, i);
	wint_t c;

	QU_REQUIRE (QuEIndex, p);
	c = qu_utf8_getchar (&p, p + self->len - 1);
	QU_REQUIRE (QuESyntax, c != QU_WCHAR_INVALID);
	return qu_utf8_new_char (c);
}



/*
	get a slice of the string starting at the i-th chracter, n characters long
	i may be negative
*/
qu_string qu_utf8_getslice (qu_string self, long i, long n)
{
	char      *p = qu_utf8_pos (self->str, self->len, i);
	long       w;
	char      *e;
	qu_string  r;
	volatile char *j;

	QU_REQUIRE (QuEIndex, p);

	w = 0;
	j = p;
	e = self->str + self->len - 1;

	for (; n && j < e; --n, ++w)
		qu_utf8_getchar ((char **) &j, e);

	r = qu_string_newl (j - p);
	r->flags |= QU_STRING_FLAG_UTF8;
	memcpy (r->str, self->str, r->len);
	return r;
}



/*----------------------------------------------------------------------------
	character test
----------------------------------------------------------------------------*/

typedef struct pair_t pair_t;
struct pair_t
{
	wint_t x;
	wint_t y;
};

#include "Utf8set.c"

static int _utf8_inset (wint_t c, pair_t *set, int j)
{
	int i = 0;
	int k;

	while ((j - i) > 1)
	{
		k = (i + j) / 2;
		if (c > set[k].y)
			i = k;
		else if (c < set[k].x)
			j = k;
		else
			return 1;
	}
	return (set[i].x <= c && c <= set[i].y) || (set[j].x <= c && c <= set[j].y);
}



int qu_utf8_isalpha (wint_t c)
{
	return _utf8_inset (c, IsAlpha, ((sizeof (IsAlpha) / sizeof (pair_t)) - 1));
}



int qu_utf8_isalnum (wint_t c)
{
	return qu_utf8_isalpha (c) || qu_utf8_isdigit (c);
}



int qu_utf8_isblank (wint_t c)
{
	return _utf8_inset (c, IsBlank, ((sizeof (IsBlank) / sizeof (pair_t)) - 1));
}



int qu_utf8_iscntrl (wint_t c)
{
	return _utf8_inset (c, IsCntrl, ((sizeof (IsCntrl) / sizeof (pair_t)) - 1));
}



int qu_utf8_isdigit (wint_t c)
{
	return _utf8_inset (c, IsDigit, ((sizeof (IsDigit) / sizeof (pair_t)) - 1));
}



int qu_utf8_isxdigit (wint_t c)
{
	return _utf8_inset (c, IsXdigit, ((sizeof (IsXdigit) / sizeof (pair_t)) - 1));
}



int qu_utf8_isgraph (wint_t c)
{
	return _utf8_inset (c, IsGraph, ((sizeof (IsGraph) / sizeof (pair_t)) - 1));
}



int qu_utf8_islower (wint_t c)
{
	return _utf8_inset (c, IsLower, ((sizeof (IsLower) / sizeof (pair_t)) - 1));
}



int qu_utf8_isprint (wint_t c)
{
	return _utf8_inset (c, IsPrint, ((sizeof (IsPrint) / sizeof (pair_t)) - 1));
}



int qu_utf8_ispunct (wint_t c)
{
	return _utf8_inset (c, IsPunct, ((sizeof (IsPunct) / sizeof (pair_t)) - 1));
}



int qu_utf8_isspace (wint_t c)
{
	return _utf8_inset (c, IsSpace, ((sizeof (IsSpace) / sizeof (pair_t)) - 1));
}


int qu_utf8_isupper (wint_t c)
{
	return _utf8_inset (c, IsUpper, ((sizeof (IsUpper) / sizeof (pair_t)) - 1));
}



int qu_utf8_iswide (wint_t c)
{
	return c > (wint_t) 0x7F;
}



/*----------------------------------------------------------------------------
	conversion
----------------------------------------------------------------------------*/

static wint_t _utf8_charto (pair_t *r, int j, wint_t c)
{
	int i = 0;
	int k, v;

	while ((j - i) > 1)
	{
		k = (i + j) / 2;
		v = r[k].x;
		if (c > v)
			i = k;
		else if (c < v)
			j = k;
		else
			return r[k].y;
	}
	if (r[i].x == c)
		return r[i].y;

	if (r[j].x == c)
		return r[j].y;

	return c;
}



wint_t qu_utf8_toupper (wint_t c)
{
	return _utf8_charto (ToUpper, ((sizeof (ToUpper) / sizeof (pair_t)) - 1), c);
}



wint_t qu_utf8_tolower (wint_t c)
{
	return _utf8_charto (ToLower, ((sizeof (ToLower) / sizeof (pair_t)) - 1), c);
}



wint_t qu_utf8_totitle (wint_t c)
{
	return _utf8_charto (ToTitle, ((sizeof (ToTitle) / sizeof (pair_t)) - 1), c);
}
