/* Var.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Var __final
	Variables and properties are instances of this class.

	You can not create Var objects directly. They are created by the compiler
	from your source code.

	Note that instance properties, once created, are stored in the instance
	property storage as the value, not as a Var object.

	Also, all local Var objects are freed once the Sub code is generated.

	@xmp
		class Foo
		    var bar     #-> Var object
		;;

		x = Foo ()
		x.bar           #-> x.bar value is in x's property storage


		sub foo (x, y)  # x and y are Var objects at compile time.
		    ...         # At run-time, they are gone because they are
		;;              # no longer required.
*/
qu_class QuVarClass;


/* stupid forwards */
static qu_string qu__var__string (qu_var self);


/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__var__mark (qu_var self)
{
	qu_gc_mark  (self->name);
	qu_gc_mark  (self->parent);
	qu_gc_mark  (self->value);
	qu_gc_xmark (self->type);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __cmp (x) -> Int
	If v{x} is a Var then do a comparisson suitable for documentation.
	Otherwise delegate to -{Object.__cmp}.

	@xmp
		var abc = 2
		var def = 1

		println (@abc <=> @def)

	@out
		-1
*/
static int qu__var_rank (qu_var self)
{
	if (qu_var_is_docentry (self)) return 5;
	if (qu_var_is_alias    (self)) return 4;
	if (qu_var_is_enum     (self)) return 3;
	if (qu_var_is_struct   (self)) return 2;
	if (qu_var_frozen      (self)) return 1;
	return 0;
}

static qu_int qu__var__cmp (qu_var self, qu_any x)
{
	qu_class c = qu_classof (x);

	if (c == QuClassClass || c == QuSubClass)
	{
		return QuInt_1;
	}
	else if (c == QuVarClass)
	{
		int a = qu__var_rank (self);
		int b = qu__var_rank ((qu_var) x);

		if (a > b)
			return QuInt1;
		if (a < b)
			return QuInt_1;
		return qu_string_cmpint (self->name, ((qu_var) x)->name);
	}
	else
	{
		return qu_object_default_cmp (self, x);
	}
}



/*) sub __getitem (name) -> Any
	Get a value in the tag list.

	@exc
	.EKey

	@xmp
		var foo = 1
			---
			hello: 'hi'
			---

		v = @foo
		println (v->hello)

	@out
		hi
*/
static qu_any qu__var__getitem (qu_var self, qu_string name)
{
	if (qu_string_is (name))
	{
		qu_any x = qu_var_gettag (self, name);
		if (x)
			return x;
	}
	QU_THROWZ (QuEKey, "tag not found");
}



/*) sub __getslice (name, defval: Any) -> Any
	Get a value in the tag list.

	This does the same as #{__getitem} except that if the name does not
	exist then v{defval} is returned instead of throwing an exception.

	@noexc
*/
static qu_any qu__var__getslice (qu_var self, qu_string name, qu_any defval)
{
	if (qu_string_is (name))
	{
		qu_any x = qu_var_gettag (self, name);
		if (x)
			return x;
	}
	return defval;
}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	@xmp
		var a = 10
		println (@a)

	@out
		<Var __main__.a>
*/
static qu_null qu__var__print (qu_var self)
{
	qu_string s = qu__var__string (self);
	QU_ENSURE (qu_sys_printz (s->str, s->len));
	return QuNil;
}



/*) sub __setitem (name: String, val: Any) -> Any
	Set value in the tag list.

	@exc
	.EArgValue name is not a String

	@xmp
		var foo = 1
			---
			hello   = 'hi'
			---

		v = @foo
		v->goodbye = 'bye'
		println (v->goodbye)

	@out
		bye
*/
static qu_any qu__var__setitem (qu_var self, qu_string name, qu_any val)
{
	QU_ARG_TYPE (name, qu_string);
	qu_var_settag (self, name, val);
	return val;
}



/*) sub __string -> String
	Return a string representation of self.

	@xmp
		var a = @a as String
		println (a)

	@out
		<Var __main__.a>
*/
static qu_string qu__var__string (qu_var self)
{
	return qu_object_where_string (self, "<Var ", ">");
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub classof -> Class
	Return the Class object where the Var is declared.

	@exc
	.ENone not a property

	@xmp
		class Foo
			local x
			println ((@x).classof)
		;;

	@out
		<Class __main__.Foo>
*/
static qu_class qu__var_classof (qu_var self)
{
	qu_class c = qu_var_get_class (self);
	QU_REQUIRE (QuENone, c);
	return c;
}



/*) sub deref -> Any
	Get the value until the value is not a Var object.

	@noexc

	@xmp
		var a = 1
		var b = @a
		var c = @b
		var d = @c
		println (d.deref)

	@out
		1
*/
static qu_any qu__var_deref (qu_var self)
{
	while (qu_var_is (self->value))
		self = self->value;
	return self->value;
}



/*) sub ident -> String
	Return the identifier name.

	@noexc

	@xmp
		foo = 1
		x = @foo
		println (x.ident)

	@out
		foo
*/
static qu_string qu__var_ident (qu_var self)
{
	return self->name;
}



/*HACK - documentation */
/*) sub isalias -> Boolean
	Return true if the variable is an alias declaration.

	This is a documentation hack. Never use this unless within documentation
	engines such as the %{Doc} module.

	@noexc

	@xmp
		alias moment: Time.now
*/
static qu_boolean qu__var_isalias (qu_var self)
{
	return qu_var_is_alias (self) ? QuTrue : QuFalse;
}



/*) sub isconst -> Boolean
	Return true if the variable is a constant, false otherwise.

	@noexc

	@xmp
		const foo = 1
		v = @foo
		println (v.isconst)

	@out
		true
*/
static qu_boolean qu__var_isconst (qu_var self)
{
	return qu_var_frozen (self) ? QuTrue : QuFalse;
}



/*) sub isclass -> Boolean
	Return true if self is a class property, false otherwise.

	@noexc

	@xmp
		class Foo
			var class bar   #-> isclass
		;;
*/
static qu_boolean qu__var_isclass (qu_var self)
{
	return qu_var_is_class (self) ? QuTrue : QuFalse;
}



/*) sub isdoc -> Boolean
	Return true if this comes from a k{__doc__} entry, false otherwise.

	@noexc

	@xmp
		# run with the -D option

		__doc__ Foo
			"""Hello World!"""

		v = @Foo
		println (v.isdoc)

	@out
		true
*/
static qu_boolean qu__var_isdoc (qu_var self)
{
	return qu_var_is_docentry (self) ? QuTrue : QuFalse;
}



/*) sub isenum -> Boolean
	Return true if this comes from an k{enum} statement, false otherwise.

	@noexc

	@xmp
		enum Thing
			ONE
			TWO
			THREE
		;;
		v = @Thing
		println (v.isenum)

	@out
		true
*/
static qu_boolean qu__var_isenum (qu_var self)
{
	return qu_var_is_enum (self) ? QuTrue : QuFalse;
}



/*) sub isenumitem -> Boolean
	Return true if this is an k{enum} item.

	@noexc

	@xmp
		enum Thing
			ONE
			TWO
			THREE
		;;
		v = @ONE
		println (v.isenumitem)

	@out
		true
*/
static qu_boolean qu__var_isenumitem (qu_var self)
{
	return qu_var_is_enumitem (self) ? QuTrue : QuFalse;
}



/*) sub isprivate -> Boolean
	Return true if the Var is private.

	@noexc
*/
static qu_boolean qu__var_isprivate (qu_var self)
{
	return qu_var_is_private (self) ? QuTrue : QuFalse;
}



/*) sub isreadonly -> Boolean
	Return true if the Var is read-only.

	@noexc
*/
static qu_boolean qu__var_isreadonly (qu_var self)
{
	return qu_var_is_rdwr (self) ? QuFalse : QuTrue;
}



/*) sub isinst -> Boolean
	Return true if self is an instance property, false otherwise.

	@noexc

	@xmp
		class Foo
			var bar   #-> isinst
		;;
*/
static qu_boolean qu__var_isinst (qu_var self)
{
	return qu_var_is_inst (self) ? QuTrue : QuFalse;
}



/*) sub isstruct -> Boolean
	Return true if this comes from a k{struct} statement, false otherwise.

	@noexc

	@xmp
		struct Thing    #-> isstruct isconst
			foo: Int
			bar: String
		;;
		v = @Thing
		println (v.isstruct)
		println (v.isconst)

	@out
		true
		true
*/
static qu_boolean qu__var_isstruct (qu_var self)
{
	return qu_var_is_struct (self) ? QuTrue : QuFalse;
}



/*) sub value (*arg) -> Any
	@def
	sub value -> Any
	sub value (val: Any) -> Any
	@p
	Set or get value.

	@exc
	.EModify   is an instance property or a constant
	.Exception anything thrown by the validator, if any
	.ENone     can't access from here
	.EValue

	@prog
		var x = 10
		v = @x
		v.value = 20
		println (x ' ' v.value)

	@out
		20 20
*/
static qu_any _var_value_set (qu_var self, qu_any val)
{
	if (qu_var_frozen (self))
		QU_THROWF (QuEModify, "cant modify constant %s", self->name->str);

	if (qu_var_is_inst (self))
		QU_THROWF (QuEModify, "cant modify instance initial value %s", self->name->str);

	if (!qu_var_is_rdwr (self))
	{
		qu_class owner = qu_symbol_get_class ((qu_symbol) self);

		if (owner && !qu_var_is_inst (self))
		{
			qu_class from = qu_symbol_get_class (__QuInterpFrame->code->parent);
			if (from != owner)
				goto nope;
		}

		if (!qu_var_can_access (self, __QuInterpFrame->code->parent))
		{
		  nope:
			QU_THROWF (QuEModify, "cant modify %s", self->name->str);
		}
	}

	if (self->type)
	{
		val = qu_object_validate (QuNil, self->type, val);
		if (val == 0)
			QU_THROWF (QuEValue, "%s type mismatch", self->name->str);
	}

	self->value = val;
	return val;
}


static qu_any _var_value_get (qu_var self)
{
	if (!qu_var_can_access (self, __QuInterpFrame->code->parent))
		QU_THROWF (QuENone, "cant access %s", self->name->str);
	return self->value;
}


static qu_any qu__var_value (qu_var self, int argc, qu_any *argv)
{
	assert (__QuInterpFrame);

	if (argc == 0)
	{
		return _var_value_get (self);
	}
	else
	{
		QU_ARG_VCOUNT (0, 1);
		return _var_value_set (self, argv[0]);
	}
}



/*) sub tag (*args) -> Any
	@def
	sub tag -> Any
	sub tag (items: Dict) -> self
	@p
	Set tag values or get an item in the tag list.

	If getting and the Var does not have any tag then nil is returned.

	@exc
	.EArgValue items is not a Dict or contains non-String key(s)

	@xmp
		var foo
			---
			hello = 0
			---

		(@foo).tag (bar:1, baz:2)
		println ((@foo).tag)

	@out
		['hello':0, 'bar':1, 'baz':2]
*/
static qu_any qu__var_tag (qu_var self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc == 0)
	{
		qu_dict d = qu_var_gettag_dict (self);
		if (d)
			return d;
		return QuNil;
	}
	else
	{
		qu_dict d;
		long    b = 0, i = 0;
		qu_any  k, v;

		QU_ARG_VTYPE (0, d, qu_dict);
		while ((v = qu_dict_walk (d, &k, &b, &i)) != 0)
		{
			QU_ARG_CHECKZ (qu_string_is (k), "tag name in Dict not a String");
			qu_var_settag (self, k, v);
		}
		return self;
	}
}



/*) sub type (*arg) -> Any
	@def
	sub type -> Any
	sub type (setting: Any) -> Any
	@p
	Get or set the variable type.

	If getting and the variable is not typed, then nil is returned.

	If setting, EModify is thrown if a type is already assigned.
	v{setting} must be a valid I{validator}.

	@exc
	.EModify
	.EValue

	@xmp
		var x

		v = @x

		println (v.type)
		v.type = Int

		x = 10
		println (x)

		try
			x = 'foo'
		catch
			println (__err__)
		;;

	@out
		10
		(EValue) x type mismatch
*/
static qu_any qu__var_type (qu_var self, int argc, qu_any *argv)
{
	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
	{
		qu_any type = argv[0];

		QU_REQUIREZ (QuEModify, !self->type, "variable already typed");
		QU_REQUIREZ (QuEValue, qu_object_valid_type (type), "not a valid type");

		self->type = type;
		return type;
	}
	else
	{
		if (self->type)
			return self->type;
		return QuNil;
	}
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

qu_var qu_var_new (qu_symbol block, qu_string name, long flags)
{
	register qu_var v;

	if (flags & (QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_LEXICAL))
		flags |= QU_VAR_FLAG_PRIVATE;

	v = qu_gc_new (QuVarClass);
	v->flags |= flags;
	v->name   = name;
	v->index  = -1;
	v->value  = QuNil;
	v->type   = 0;
	v->parent = block;
	return v;
}



qu_var qu_var_new_ex (qu_symbol block, qu_string name, long flags, qu_any type)
{
	qu_var v;

	assert (qu_class_is (type) || qu_sub_is (type) || qu_var_is (type));
	v = qu_var_new (block, name, flags);
	v->type = type;
	return v;
}



qu_module qu_var_get_module (qu_var self)
{
	qu_symbol block = self->parent;

	if (qu_sub_is (block))
	{
		return qu_sub_get_module ((qu_sub) block);
	}
	else if (qu_class_is (block))
	{
		return qu_class_get_module ((qu_class) block);
	}
	else
	{
		assert (qu_module_is (block));
		return (qu_module) block;
	}
}



qu_class qu_var_get_class (qu_var self)
{
	qu_any block = self->parent;

	if (qu_class_is (block))
	{
		return (qu_class) block;
	}
	else if (qu_sub_is (block))
	{
		return qu_sub_get_class ((qu_sub) block);
	}
	else
	{
		return 0;
	}
}



qu_any qu_var_set_value (qu_var self, qu_any val)
{
	assert (!qu_var_frozen (self));

	if (self->type)
	{
		val = qu_object_validate (QuNil, self->type, val);
		if (val == 0)
		{
			qu_string s = qu__var__string (self);
			QU_THROWF (QuEValue, "%s type mismatch", s->str);
		}
	}
	self->value = val;
	return val;
}



qu_any qu_var_gettag (qu_var self, qu_string name)
{
	if (self->attrs)
		return qu_attr_find (self->attrs, name);
	return 0;
}



void qu_var_settag (qu_var self, qu_string name, qu_any value)
{
	if (self->attrs == 0)
		self->attrs = qu_attr_alloc ();
	qu_attr_set (self->attrs, name, value);
}



qu_dict qu_var_gettag_dict (qu_var self)
{
	if (self->attrs == 0)
	{
		return 0;
	}
	else
	{
		qu_dict d = qu_dict_new (qu_attr_len (self->attrs));
		qu_attr_merge_to_dict (self->attrs, d);
		return d;
	}
}


static void _var_enum_set (qu_symbol where, qu_var v)
{
	if (qu_class_is (where))
	{
		if (qu_class_decl_attr ((qu_class) where, v->name, (qu_symbol) v) == 0)
			QU_FATAL ("bad or duplicate symbol %s in class %s",
				v->name->str, ((qu_class) where)->name->str);
	}
	else if (qu_module_is (where))
	{
		if (qu_module_getattr ((qu_module) where, v->name) != 0)
			QU_FATAL ("bad or duplicate symbol %s in class %s",
				v->name->str, ((qu_module) where)->name->str);
		qu_module_setattr ((qu_module) where, v->name, (qu_symbol) v);
	}
	else
	{
		QU_FATAL ("VarEnum called with 'where' not a Class nor Module");
	}
}



static qu_var _var_static (qu_symbol block, const char *name, long flags)
{
	register qu_var v;

	assert (block != 0);

	if (flags & (QU_VAR_FLAG_LOCAL | QU_VAR_FLAG_LEXICAL))
		flags |= QU_VAR_FLAG_PRIVATE;

	v = qu_gc_new_static (QuVarClass);
	v->flags |= flags;
	v->name   = qu_string_new_staticz (name);
	v->index  = -1;
	v->value  = QuNil;
	v->type   = 0;
	v->parent = block;
	return v;
}



/* !!! we use qu_string_new_staticz - not qu_string_new */
qu_var qu_var_enum (qu_symbol where, const char *name, const char *first, ...)
{
	long    flags = QU_VAR_FLAG_FROZEN | (qu_class_is (where) ? QU_VAR_FLAG_CLASS : 0);
	qu_dict d = qu_dict_new_static (8);
	qu_var  e = _var_static (where, name, flags);
	qu_var  v;
	char   *key;
	int     val;
	va_list va;

	e->flags |= QU_VAR_FLAG_ENUM | QU_VAR_FLAG_FROZEN;

	_var_enum_set (where, e);
	va_start (va, first);
	key = (char *) first;
	do
	{
		val = va_arg (va, int);
		v = _var_static (where, key, flags | QU_VAR_FLAG_ENUMITEM);
		v->value = qu_integer_newi (val);
		_var_enum_set (where, v);
		qu_dict_setitem (d, v->value, v->name);
		key = va_arg (va, char *);

	} while (key);
	va_end (va);
	qu_dict_freeze (d);
	e->value = d;
	return e;
}



/* !!! we use qu_string_new_staticz - not qu_string_new */
qu_var qu_var_enum_ul (qu_symbol where, const char *name, const char *first, ...)
{
	long          flags = QU_VAR_FLAG_FROZEN | (qu_class_is (where) ? QU_VAR_FLAG_CLASS : 0);
	qu_dict       d = qu_dict_new_static (8);
	qu_var        e = _var_static (where, name, flags);
	qu_var        v;
	char         *key;
	unsigned long val;
	va_list       va;

	e->flags |= QU_VAR_FLAG_ENUM | QU_VAR_FLAG_FROZEN;

	_var_enum_set (where, e);
	va_start (va, first);
	key = (char *) first;
	do
	{
		val = va_arg (va, unsigned long);
		v = _var_static (where, key, flags | QU_VAR_FLAG_ENUMITEM);
		if (val <= (unsigned long) QU_INT_MAX)
			v->value = qu_int_new (val);
		else
			v->value = qu_long_new_staticul (val);
		_var_enum_set (where, v);
		qu_dict_setitem (d, v->value, v->name);
		key = va_arg (va, char *);

	} while (key);
	va_end (va);
	qu_dict_freeze (d);
	e->value = d;
	return e;
}



/* !!! we use qu_string_new_staticz - not qu_string_new */
qu_var qu_var_enum_ex (qu_symbol where, const char *name, const char *spec, ...)
{
	long    flags = QU_VAR_FLAG_FROZEN | (qu_class_is (where) ? QU_VAR_FLAG_CLASS : 0);
	qu_dict d = qu_dict_new_static (8);
	qu_var  e = _var_static (where, name, flags);
	qu_var  v;
	char   *k;
	qu_any  x;
	va_list va;

	e->flags |= QU_VAR_FLAG_ENUM | QU_VAR_FLAG_FROZEN;

	_var_enum_set (where, e);
	va_start (va, spec);
	while (*spec)
	{
		k = va_arg (va, char *);
		v = _var_static (where, k, flags | QU_VAR_FLAG_ENUMITEM);
		switch (*spec)
		{
			case QU_STRUCT_ID_char:
			{
				int val = va_arg (va, int);
				x = qu_string_byte (val);
				break;
			}

			case QU_STRUCT_ID_Boolean:
			{
				int val = va_arg (va, int);
				x = val ? QuTrue : QuFalse;
				break;
			}

			case QU_STRUCT_ID_Int:
			{
				long val = va_arg (va, long);
				x = qu_int_newl (val);
				break;
			}

			case QU_STRUCT_ID_int:
			{
				long val = (long) va_arg (va, int);
				if (QU_INT_MIN <= val && QU_INT_MAX <= val)
					x = qu_int_newl (val);
				else
					x = qu_long_new_staticl (val);
				break;
			}

			case QU_STRUCT_ID_uint:
			{
				unsigned int val = va_arg (va, unsigned int);
				if ((unsigned long) val <= QU_INT_MAX)
					x = qu_int_newul (val);
				else
					x = qu_long_new_staticul (val);
				break;
			}

			case QU_STRUCT_ID_long:
			{
				long val = va_arg (va, long);
				if (QU_INT_MIN <= val && QU_INT_MAX <= val)
					x = qu_int_newl (val);
				else
					x = qu_long_new_staticl (val);
				break;
			}

			case QU_STRUCT_ID_ulong:
			{
				unsigned long val = va_arg (va, unsigned long);
				if (val <= (unsigned long) QU_INT_MAX)
					x = qu_int_newl (val);
				else
					x = qu_long_new_staticul (val);
				break;
			}

			case QU_STRUCT_ID_double:
			{
				double val = va_arg (va, double);
				x = qu_float_new_static (val);
				break;
			}

			case QU_STRUCT_ID_String:
			{
				char *val = va_arg (va, char *);
				x = qu_string_new_staticz (val);
				break;
			}

			case QU_STRUCT_ID_Any:
				x = va_arg (va, qu_any);
				if (!qu_gc_is_static (x))
					QU_FATAL ("object not static");
				break;

			default:
				QU_FATAL ("type not supported");
				x = 0; /* stupid compiler */
		}
		v->value = x;
		_var_enum_set (where, v);
		qu_dict_setitem (d, v->value, v->name);
		spec++;
	}
	va_end (va);
	qu_dict_freeze (d);
	e->value = d;
	return e;
}



int qu_var_can_access (qu_var self, qu_symbol from)
{
	qu_symbol owner = self->parent;

	if (owner == from)
		return 1;

	if (qu_var_is_lexical (self) || qu_var_is_local (self))
	{
		if (qu_sub_is (owner) && qu_sub_is (from) && ((qu_sub) from)->root == (qu_sub) owner)
			return 1;
		return 0;
	}
	else if (!qu_var_is_private (self))
	{
		return 1;
	}
	else
	{
		if (qu_module_is (owner))
		{
			if (qu_module_is (from))
			{
				return 0;
			}
			else if (qu_class_is (from))
			{
				from = (qu_symbol) ((qu_class) from)->parent;
			}
			else
			{
				assert (qu_sub_is (from));
				from = (qu_symbol) qu_sub_get_module ((qu_sub) from);
			}
			return owner == from;
		}
		else if (qu_class_is (owner))
		{
			switch (qu_classid (from))
			{
				case QU_CLASS_ID_Sub:
					from = (qu_symbol) qu_sub_get_class ((qu_sub) from);
					if (from == 0 || from == owner)
						break;
					/*nobreak*/
				case QU_CLASS_ID_Class:
					return qu_class_isa ((qu_class) from, (qu_class) owner);
			}
		}
		return 0;
	}
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_var__setup ()
{
	qu_class_setup_sys (QuVarClass, qu_var_s, 0, qu__var__mark);
}



void qu_var__init ()
{
	/* standard instance methods */
	qu_class_method_inst  (QuVarClass, QU_NAME__cmp     , qu__var__cmp      ,  1);
	qu_class_method_inst  (QuVarClass, QU_NAME__getitem , qu__var__getitem  ,  1);
	qu_class_method_inst  (QuVarClass, QU_NAME__getslice, qu__var__getslice ,  2);
	qu_class_method_inst  (QuVarClass, QU_NAME__print   , qu__var__print    ,  0);
	qu_class_method_inst  (QuVarClass, QU_NAME__setitem , qu__var__setitem  ,  2);
	qu_class_method_inst  (QuVarClass, QU_NAME__string  , qu__var__string   ,  0);

	/* specific instance methods */
	qu_class_method_inst  (QuVarClass, QU_NAME_classof  , qu__var_classof   ,  0);
	qu_class_method_inst  (QuVarClass, QU_NAME_ident    , qu__var_ident     ,  0);
	qu_class_method_instz (QuVarClass, "deref"          , qu__var_deref     ,  0);
	qu_class_method_instz (QuVarClass, "isalias"        , qu__var_isalias   ,  0);
	qu_class_method_instz (QuVarClass, "isconst"        , qu__var_isconst   ,  0);
	qu_class_method_instz (QuVarClass, "isclass"        , qu__var_isclass   ,  0);
	qu_class_method_instz (QuVarClass, "isdoc"          , qu__var_isdoc     ,  0);
	qu_class_method_instz (QuVarClass, "isenum"         , qu__var_isenum    ,  0);
	qu_class_method_instz (QuVarClass, "isenumitem"     , qu__var_isenumitem,  0);
	qu_class_method_instz (QuVarClass, "isprivate"      , qu__var_isprivate ,  0);
	qu_class_method_instz (QuVarClass, "isreadonly"     , qu__var_isreadonly,  0);
	qu_class_method_instz (QuVarClass, "isinst"         , qu__var_isinst    ,  0);
	qu_class_method_instz (QuVarClass, "isstruct"       , qu__var_isstruct  ,  0);
	qu_class_method_instz (QuVarClass, "value"          , qu__var_value     , -1);
	qu_class_method_instz (QuVarClass, "tag"            , qu__var_tag       , -1);
	qu_class_method_instz (QuVarClass, "type"           , qu__var_type      , -1);

	/* finalize declaration */
	qu_class_enddecl (QuVarClass);
}
