/* Weak.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class Weak __final
	This class allows you to make I{weak references} to objects.

	A I{weak reference} is a reference to an object but does not prevent the
	object from being garbage collected.

	When the referred object is garbage collected then the referer is notified
	which causes the reference to the object invalidated.

	@xmp
		use Gc

		x = [1, 2, 3]
		y = Weak (x)
		x = nil
		println (y.ref)
		Gc.reclaim
		Gc.reclaim
		println (y.ref)

	@out
		[1, 2, 3]
		[1024] (ENone) weak reference gone
*/
qu_class QuWeakClass;


/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (x) -> Weak
	Make a I{weak reference} to an object.

	If v{x} is already a Weak then nothing happens and v{x} is returned.

	Qu does not allow you to create references to the following objects:
	@ul
	.Null
	.Boolean
	.String
	.Int
	.Long
	.Rational
	.Big
	.Float
	.Inf
	.Complex
	.Range
	.Opaque
	.Weak

	@exc
	.EArgValue
	.ESecurity
	.EMemory
*/
qu_weak qu__weak_new (qu_class self, qu_any x)
{
	qu_class c = qu_classof (x);
	long     h;
	qu_weak  w;

	__qu_unsafe__;

	QU_REQUIREF (QuEArgValue, !qu_class_noweak (c), "cant make references to %s", c->name->str);
	h = qu_object_hash (x);
	w = qu_gc_new (self);
	w->ref = x;
	w->hash = h;
	if (!qu_gc_is_static (x))
		QU_ENSURE (qu_gc_weak_set (x, w));
	return w;
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

/* none - we don't mark self->ref, of course */

#ifdef QU_DEBUG_LOCAL
static void qu__weak__mark (qu_weak self)
{
	printf ("Weak mark: %p", self->ref);
	if (self->ref)
	{
		printf (" (%s)", qu_classof (self->ref)->name->str);
		if (qu_module_is (self->ref))
			printf (" %s", ((qu_module) self->ref)->name->str);
	}
	printf ("\n");
}
#endif



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __like (x) -> Boolean
	Return true if v{x} is a Weak and refer to the same object as self,
	return k{false] otherwise.

	@noexc
*/
static qu_boolean qu__weak__like  (qu_weak self, qu_any x)
{
	return (qu_weak_is (x) && self->ref && self->ref == ((qu_weak) x)->ref)
		? QuTrue : QuFalse;

}



/*) sub __print -> Null
	Print self to -{Stream.stdout}.

	See -{print} for possible exceptions.

	@xmp
		a = [1, 2, 3]
		w = Weak (a)
		println (w)

	@out
		<Weak Array 0x400781b8>
*/
#define WEAK_DEAD_STR    "<Weak (dead)>"
#define WEAK_DEAD_STRLEN strlen (WEAK_DEAD_STR)

#define WEAK_SPRINTF_ALIVE_BUFSIZ 128

#define WEAK_SPRINTF(self, buf) \
QU_BEGIN \
	snprintf (buf, WEAK_SPRINTF_ALIVE_BUFSIZ, \
		"<Weak %s %p>", qu_classof ((self)->ref)->name->str, (self)->ref); \
	buf[WEAK_SPRINTF_ALIVE_BUFSIZ - 1] = '\0'; \
QU_END

static qu_null qu__weak__print (qu_weak self)
{
	if (self->ref)
	{
		char buf [WEAK_SPRINTF_ALIVE_BUFSIZ];
		WEAK_SPRINTF (self, buf);
		QU_ENSURE (qu_sys_printz (buf, strlen (buf)));
	}
	else
	{
		QU_ENSURE (qu_sys_printz (WEAK_DEAD_STR, WEAK_DEAD_STRLEN));
	}
	return QuNil;
}



/*) sub __string -> String
	Return a string representation of self.

	@noexc

	@xmp
		a = [1, 2, 3]
		w = Weak (a)
		println (String (w))

	@out
		<Weak Array 0x400781b8>
*/
static qu_string qu__weak__string (qu_weak self)
{
	if (self->ref)
	{
		char buf [WEAK_SPRINTF_ALIVE_BUFSIZ];
		WEAK_SPRINTF (self, buf);
		return qu_string_newz (buf);
	}
	else
	{
		return qu_string_newz (WEAK_DEAD_STR);
	}
}



/*) sub __true -> Boolean
	Return false if #{unref} has been called or the referenced object no
	longer exist, true otherwise.

	@noexc

	@xmp
		a = [1, 2, 3]
		w = Weak (a)
		w.unref
		println ('none') unless w

	@out
		none
*/
static qu_boolean qu__weak__true  (qu_weak self)
{
	return self->ref ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub ref -> Any
	Return the referenced object.

	@exc
	.ENone object no longer exist or #{unref} has been called

	@xmp
		a = [1, 2, 3]
		w = Weak (a)
		println (w.ref)

	@out
		[1, 2, 3]
*/
static qu_any qu__weak_ref (qu_weak self)
{
	QU_REQUIREZ (QuENone, self->ref, "weak reference gone");
	return self->ref;
}



/*) sub unref -> Nil
	Unreference the object.

	If the object is already unreferenced or no longer exist, then nothing happens.

	@noexc

	@xmp
		a = [1, 2, 3]
		w = Weak (a)
		w.unref
		println (w.xref)

	@out
		nil
*/
qu_null qu__weak_unref (qu_weak self)
{
	if (self->ref)
	{
		qu_gc_weak_delete (self->ref, self);
		QU_WEAK_UNREF (self);
	}
	return QuNil;
}



/*) sub xref -> Any
	This does the same as #{ref} except that no exception is thrown.

	If the referenced object no longer exists or #{unref} has been called,
	then nil is returned.

	@noexc

	@xmp
		a = [1, 2, 3]
		w = Weak (a)
		println (w.xref)
		w.unref
		println (w.xref)

	@out
		[1, 2, 3]
		nil
*/
static qu_any qu__weak_xref (qu_weak self)
{
	return self->ref ? self->ref : (qu_any) QuNil;
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_weak__setup ()
{
#ifdef QU_DEBUG_LOCAL
	qu_class_setup_final (QuWeakClass, qu_weak_s, 0, qu__weak__mark);
#else
	qu_class_setup_final (QuWeakClass, qu_weak_s, 0, 0);
#endif
}



void qu_weak__init ()
{
	/* standard class methods */
	qu_class_method_class (QuWeakClass, QU_NAME_new, qu__weak_new, 1);

	/* standard instance methods */
	qu_class_method_inst  (QuWeakClass, QU_NAME__like  , qu__weak__like  , 1);
	qu_class_method_inst  (QuWeakClass, QU_NAME__print , qu__weak__print , 0);
	qu_class_method_inst  (QuWeakClass, QU_NAME__string, qu__weak__string, 0);
	qu_class_method_inst  (QuWeakClass, QU_NAME__true  , qu__weak__true  , 0);

	/* specific instance methods */
	qu_class_method_instz (QuWeakClass, "ref"  , qu__weak_ref  , 0);
	qu_class_method_instz (QuWeakClass, "unref", qu__weak_unref, 0);
	qu_class_method_instz (QuWeakClass, "xref" , qu__weak_xref , 0);

	/* finalize declaration */
	qu_class_enddecl (QuWeakClass);
}
