/* attr.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

/* initial capacity */
#define INITCAP 4

/* initial number of buckets must be a power of 2 */
#define INITBUCKETS 8

/* flags */
#define FLAG_FIXED     0x1  /* cant insert/delete */
#define FLAG_INSTFLAT  0x2  /* is instance attribute table flat array? */
#define FLAG_ARRANGED  0x4  /* class->attrs: var->index already arranged? */


typedef struct entry_s entry_s;
struct entry_s
{
	qu_any    value;   /* value can be anything */

	/* not allocated in case of FLAG_INSTFLAT */
	qu_string key;     /* key must be a String object (immutable) */
	int       index;   /* class->attrs: var->index instance property */
	long      right;   /* next entry in bucket */
};

struct qu_attr_s
{
	entry_s *items;   /* the entries */
	long    *buckets; /* buckets, indexes into items */
	long     count;   /* number of elements */
	long     mask;    /* current mask */
	long     cap;     /* numbers of allocated items */
	int      flags;   /* see below */
	int      nip;     /* number of instance properties */
};


static void _attr_grow (qu_attr_s *t)
{
	if (t->mask < (INT_MAX >> 2))
	{
		entry_s *e = t->items;
		long     n = t->count;
		long     k = ((t->mask + 1) << 1) - 1;
		long    *b = qu_mem_calloc (k + 1, sizeof (long));
		long     i;

		for (i = 0; i < n; i++, e++)
		{
			register long h = qu_string_HASH (e->key) & k;
			e->right = b [h];
			b [h] = i + 1;
		}

		qu_mem_free (t->buckets);
		t->buckets = b;
		t->mask = k;
	}
}



static void _attr_add (qu_attr_s *t, qu_string key, qu_any val, long hash)
{
	register entry_s *e;

	if (t->count >= t->cap)
	{
		t->cap *= 2;
		t->items = qu_mem_realloc (t->items, t->cap * sizeof (entry_s));
	}
	e = t->items + t->count;
	e->key = key;
	e->value = val;
	hash &= t->mask;
	e->right = t->buckets [hash];
	t->buckets [hash] = ++t->count;
	if (t->count > ((t->mask << 1) / 3))
		_attr_grow (t);
}



static entry_s *_attr_find (qu_attr_s *t, qu_string key, long hash)
{
	register entry_s *e;
	register long     i;
	register long     l = key->len;
	register char    *s = key->str;

	for (i = t->buckets [hash & t->mask]; i; i = e->right)
	{
		e = t->items + i - 1;
		if (key == e->key /* <-- migh be hit a lot */
			|| (l == e->key->len && memcmp (s, e->key->str, l) == 0))
			return e;
	}
	return 0;
}



static entry_s *_attr_findz (qu_attr_s *t, const char *key, long len, long hash)
{
	register entry_s *e;
	register long     i;

	for (i = t->buckets [hash & t->mask]; i; i = e->right)
	{
		e = t->items + i - 1;
		if (e->key->len == len && memcmp (e->key->str, key, len) == 0)
			return e;
	}
	return 0;
}



static int _attr_del (qu_attr_s *t, qu_string key, long hash)
{
	entry_s *e, *p;
	long     i;
	long     l = key->len;
	char    *s = key->str;
	long     k = hash & t->mask;

	for (i = t->buckets [k], p = NULL; i; p = e, i = e->right)
	{
		e = t->items + i - 1;
		if (e->key->len == l && memcmp (e->key->str, s, l) == 0)
		{
			if (p)
				p->right = e->right;
			else
				t->buckets [k] = e->right;

			memmove (e + 1, e, sizeof (entry_s) * (t->count - i));
			t->count--;
			return 1;
		}
	}
	return 0;
}



void qu_attr_free (qu_attr_s *t)
{
	if (!(t->flags & FLAG_INSTFLAT))
		qu_mem_free (t->buckets);
	qu_mem_free (t->items);
	qu_mem_free (t);
}



void qu_attr_mark (qu_attr_s *t)
{
	entry_s *e = t->items;
	long     n = t->count;

	if (t->flags & FLAG_INSTFLAT)
	{
		while (n--)
		{
			qu_gc_xmark (e->value);
			e = (entry_s *) ((void *) e + sizeof (qu_any));
		}
	}
	else
	{
		while (n--)
		{
			qu_gc_mark (e->value);
			qu_gc_mark (e->key);
			e++;
		}
	}
}


qu_attr_s *qu_attr_alloc ()
{
	register qu_attr_s *t = qu_mem_alloc (sizeof (qu_attr_s));

	t->count   = 0;
	t->cap     = INITCAP;
	t->mask    = INITBUCKETS - 1;
	t->items   = qu_mem_alloc (INITCAP * sizeof (entry_s));
	t->buckets = qu_mem_calloc (INITBUCKETS, sizeof (long));
	t->flags   = 0;

	return t;
}



qu_any qu_attr_find (qu_attr_s *t, qu_string key)
{
	register entry_s *i;

#ifdef QU_DEBUG_LOCAL
	if (t->flags & FLAG_INSTFLAT)
		QU_FATAL ("qu_attr_find cant be used with FLAG_INSTFLAT");
#endif

	i = _attr_find (t, key, qu_string_HASH (key));
	if (i)
		return i->value;
	return 0;
}



qu_any qu_attr_findz (qu_attr_s *t, const char *key, long len)
{
	register entry_s *i;

#ifdef QU_DEBUG_LOCAL
	if (t->flags & FLAG_INSTFLAT)
		QU_FATAL ("qu_attr_findz cant be used with FLAG_INSTFLAT");
#endif
	i = _attr_findz (t, key, len, qu_strz_hash (key, len));
	if (i)
		return i->value;
	return 0;
}



void qu_attr_add (qu_attr_s *t, qu_string key, qu_any val)
{

#ifdef QU_DEBUG_LOCAL
	if (t->flags & (FLAG_FIXED | FLAG_INSTFLAT))
		QU_FATAL ("qu_attr_add cant be used with FLAG_FIXED | FLAG_INSTFLAT");
#endif

	_attr_add (t, key, val, qu_string_HASH (key));
}



int qu_attr_del (qu_attr_s *t, qu_string key)
{

#ifdef QU_DEBUG_LOCAL
	if (t->flags & (FLAG_FIXED | FLAG_INSTFLAT))
		QU_FATAL ("qu_attr_del cant be used with FLAG_FIXED | FLAG_INSTFLAT");
#endif

	return _attr_del (t, key, qu_string_HASH (key));
}



int qu_attr_set (qu_attr_s *t, qu_string key, qu_any val)
{
	long     h = qu_string_HASH (key);
	entry_s *i = _attr_find (t, key, h);

#ifdef QU_DEBUG_LOCAL
	if (t->flags & (FLAG_FIXED | FLAG_INSTFLAT))
		QU_FATAL ("qu_attr_set cant be used with FLAG_FIXED | FLAG_INSTFLAT");
#endif

	if (i)
	{
		i->value = val;
		return 1;
	}
	else
	{
		_attr_add (t, key, val, h);
		return 0;
	}
}



/*EXPERIMENTAL - dotted sub*/
void qu_attr_setex (qu_attr_s *t, qu_string key, qu_any val)
{
	long     h = qu_string_HASH (key);
	entry_s *i = _attr_find (t, key, h);

	if (i)
		i->value = val;
	else
		_attr_add (t, key, val, h);
}



long qu_attr_len (qu_attr_s *t)
{
	return t->count;
}



qu_any qu_attr_walk (qu_attr_s *t, qu_string *key, long *iter)
{
#ifdef QU_DEBUG_LOCAL
	if (t->flags & FLAG_INSTFLAT)
		QU_FATAL ("qu_attr_walk can not be used with FLAG_INSTFLAT");
#endif

	if (*iter >= t->count)
	{
		return 0;
	}
	else
	{
		entry_s *e = &t->items[*iter];

		*iter = *iter + 1;
		*key = e->key;
		return e->value;
	}
}



void qu_attr_merge_to_dict (qu_attr_s *a, qu_dict d)
{
	qu_any    v;
	qu_string k;
	long      i = 0;

	while ((v = qu_attr_walk (a, &k, &i)) != 0)
		qu_dict_setitem (d, k, v);
}



void qu_attr_end_declsub (qu_sub u)
{
	qu_attr_s *t = u->attrs;
	entry_s   *e = t->items;
	long       n = t->count;
	long       k = n;
	long       i;
	qu_any     x;

	for (i = 0; i < n; i++, e++)
	{
		x = e->value;
		if (qu_var_is (x) && qu_var_is_lexical (x))
			k--;
	}

	if (k == 0)
	{
		/* this will be the case most of the time, i.e. unless the sub
			has nested subs */

		qu_attr_free (t);
		u->attrs = NULL;
	}
	else
	{
		/* look-up speed is no longer important so use minimal memory */

		long    *b = qu_mem_calloc (2, sizeof (long));
		entry_s *j = qu_mem_alloc (k * sizeof (entry_s));
		entry_s *p = j;

		n = t->count;
		e = t->items;

		for (i = 0; i < n; i++, e++)
		{
			long h = qu_string_HASH (e->key) & 1;
			x =  e->value;
			if (!(qu_var_is (x) && qu_var_is_lexical (x)))
			{
				p->value = e->value;
				p->key = e->key;
				p++;
				e->right = b [h];
				b [h] = i + 1;
			}
		}

		qu_mem_free (t->buckets);
		qu_mem_free (t->items);

		t->buckets = b;
		t->items = j;
		t->count = t->cap = k;
		t->mask = 1;
		t->flags |= FLAG_FIXED;
	}
}



/*
	Search declared attribute in class, including mixins and superclasses.
*/
static entry_s *_attr_search (qu_class c, qu_string name, qu_class org, long h)
{
	if (c != org)
	{
		entry_s *i = _attr_find (c->attrs, name, h);

		if (i)
			return i;

		if (c->mixin)
		{
			long      n = c->mixin->len;
			qu_class *m = (qu_class *) (c->mixin->objs + n);

			while (n--)
			{
				i = _attr_search (*--m, name, c, h);
				if (i)
					return i;
			}
		}

		if (c->super)
			return _attr_search (c->super, name, c, h);
	}
	return 0;
}


qu_any qu_attr_search (qu_class c, qu_string name)
{
	entry_s *i = _attr_search (c, name, 0, qu_string_HASH (name));
	if (i)
		return i->value;
	return NULL;
}


/*----------------------------------------------------------------------------
	instance attributes

	An instance attribute table is a special case.
	It references the class attribute table and stores instance properties
	values in a flat array.

	Note that we don't need to mark the class nor the class attribute table
	since obj->klass already references those.

	!!! This assumes that all class attributes including the super class
	!!! and mixin classes are merged into class->attrs.
	!!! See qu_class_enddecl.

	Remember, since 1.21.04 the Class class no longer defines any attribute.
	They are all moved to Object so there is no need to do a lookup such
	as qu_attr_find (QuClassClass->attrs, name).

----------------------------------------------------------------------------*/

/* arrange for better performance */
inline static void _attr_arrange (qu_attr_s *a)
{
	register long     m = a->count;
	register entry_s *e = a->items;
	register int      j = 0;
	register qu_any   x;

	while (m--)
	{
		x = e->value;
		if (qu_var_is (x) && qu_var_is_inst (x))
			e->index = j++;
		e++;
	}
	a->nip = j + 1;
	a->flags |= FLAG_ARRANGED;
}



/* allocate instance attribute table */
static void _attr_inst_alloc (qu_obj self)
{
	qu_class   c = qu_classof (self);
	qu_attr_s *t = c->attrs;
	long       n;

	if (!(t->flags & FLAG_ARRANGED))
		_attr_arrange (t);

	n = t->nip;
	assert (n > 0);

	t          = qu_mem_alloc (sizeof (qu_attr_s));
	t->items   = qu_mem_calloc (n, sizeof (qu_any));
	t->count   = n;
	t->cap     = n;
	t->mask    = 0;
	t->buckets = 0;
	t->flags   = FLAG_INSTFLAT;

	self->attrs = t;
}



/* find instance attribute table */
static qu_any _attr_inst_get (qu_obj self, qu_string key, int restrict)
{
	qu_class c = qu_classof (self);
	long     h = qu_string_HASH (key);
	qu_any   x;
	entry_s *i;

	if (qu_class_is_auto (c) || qu_class_is_dynamic (c))
	{
		if (self->attrs)
		{
			i = _attr_find (self->attrs, key, h);
			if (i)
				return i->value;
		}

		i = _attr_find (c->attrs, key, h);
		if (i == NULL)
		{
			if (qu_class_is_pure (c))
				return 0;

			i = _attr_find (QuObjectClass->attrs, key, h);
			if (i == NULL)
				return 0;
		}

		x = i->value;
		if (qu_symbol_is_local (x) || (restrict && qu_symbol_is_private (x)))
			return 0;

		if (qu_var_is (x))
			return ((qu_var) x)->value;
	}
	else
	{
		i = _attr_find (c->attrs, key, h);
		if (i == NULL)
		{
			if (qu_class_is_pure (c))
				return 0;

			i = _attr_find (QuObjectClass->attrs, key, h);
			if (i == NULL)
			{
				/*EXPERIMENTAL - dotted sub
					because now attributes may be added later on */
				i = _attr_search (c, key, NULL, h);
				if (i == NULL)
					return 0;
			}
		}

		x = i->value;
		if (qu_symbol_is_local (x) || (restrict && qu_symbol_is_private (x)))
			return 0;

		if (qu_var_is (x))
		{
			if (qu_var_is_inst (x))
			{
				if (self->attrs)
				{
					entry_s *e = (entry_s *) ((void *) self->attrs->items + i->index * sizeof (qu_any));
					if (e->value)
						return e->value;
				}
			}
			return ((qu_var) x)->value;
		}
	}
	return x;
}



/* find instance attribute without throwing (except if __get is called) */
qu_any qu_attr_inst_getnt (qu_obj self, qu_string key, int restrict)
{
	qu_any x = _attr_inst_get (self, key, restrict);

	if (x == 0)
	{
		qu_sub x = _attr_inst_get (self, QU_NAME__get, 0);

		if (x && qu_sub_is (x) && qu_sub_is_inst (x))
			return qu_sub_call1 (x, self, key);
		return 0;
	}
	return x;
}



/* to prevent throwing expensive exception message */
volatile int __QuAttrNoThrow;


/* find instance attribute respecting __QuAttrNoThrow */
qu_any qu_attr_inst_get (qu_obj self, qu_string key, int restrict)
{
	qu_any x = _attr_inst_get (self, key, restrict);

	if (x == 0)
	{
		qu_sub x = _attr_inst_get (self, QU_NAME__get, 0);

		if (x && qu_sub_is (x) && qu_sub_is_inst (x))
			return qu_sub_call1 (x, self, key);

		if (__QuAttrNoThrow)
			return 0;

		QU_THROWF (QuEImplement, "object %s has no attribute %s", qu_classof (self)->name->str, key->str);
	}
	return x;
}



/* set instance attribute */
qu_any qu_attr_inst_set (qu_obj self, qu_string key, qu_any val, int restrict)
{
	qu_class  c = qu_classof (self);
	long      h = qu_string_HASH (key);
	entry_s  *i = _attr_find (c->attrs, key, h);
	qu_any    x;

	if (i == NULL)
		i = _attr_find (QuObjectClass->attrs, key, h);

	if (qu_class_is_auto (c))
	{
		if (i)
		{
			x = i->value;
			if (qu_symbol_is_local (x))
				goto nope;

			if (qu_var_is (x))
			{
				if (restrict)
				{
					if (qu_symbol_is_private (x))
						goto nope;

					if (!qu_var_is_rdwr (x))
						goto modify;
				}

				if (qu_var_is_class (x))
				{
					if (((qu_var) x)->type)
					{
						val = qu_object_validate (c, ((qu_var) x)->type, val);
						if (val == 0)
							goto bad;
					}
					((qu_var) x)->value = val;
				}
				else
				{
					if (((qu_var) x)->type)
					{
						val = qu_object_validate (self, ((qu_var) x)->type, val);
						if (val == 0)
							goto bad;
					}
					goto setit;
				}
			}
			else
			{
#ifdef QU_SUB_ALLOW_ASSIGN
				assert (qu_sub_is (x));

				if (restrict && qu_symbol_is_private (x))
					goto nope;

				if (qu_sub_is_class (x))
					return qu_sub_call1 (x, c, val);
				return qu_sub_call1 (x, self, val);
#else
				goto modify;
#endif
			}
		}
		else
		{

		  setit:

			if (self->attrs == NULL)
			{
				self->attrs = qu_attr_alloc ();
				qu_attr_add (self->attrs, key, val);
			}
			else
			{
				qu_attr_set (self->attrs, key, val);
			}
		}
		return val;
	}
	else if (qu_class_is_dynamic (c))
	{
		if (i)
		{
			x = i->value;
			if (qu_symbol_is_local (x))
				goto nope;

			if (qu_var_is (x))
			{
				if (restrict)
				{
					if (qu_symbol_is_private (x))
						goto nope;
				}
				else
				{
					if (!qu_var_is_rdwr (x))
						goto modify;
				}

				if (qu_var_is_class (x))
				{
					if (((qu_var) x)->type)
					{
						val = qu_object_validate (c, ((qu_var) x)->type, val);
						if (val == 0)
							goto bad;
					}
					((qu_var) x)->value = val;
				}
				else
				{
					if (((qu_var) x)->type)
					{
						val = qu_object_validate (self, ((qu_var) x)->type, val);
						if (val == 0)
							goto bad;
					}
					goto setit;
				}
			}
			else
			{
#ifdef QU_SUB_ALLOW_ASSIGN
				assert (qu_sub_is (x));

				if (restrict && qu_symbol_is_private (x))
					goto nope;

				if (qu_sub_is_class (x))
					return qu_sub_call1 (x, c, val);
				return qu_sub_call1 (x, self, val);
#else
				goto modify;
#endif
			}
			return val;
		}
	}
	else if (i)
	{
		x = i->value;
		if (qu_symbol_is_local (x))
			goto nope;

		if (qu_var_is (x))
		{
			if (restrict)
			{
				if (qu_symbol_is_private (x))
					goto nope;

				if (!qu_var_is_rdwr (x))
					goto modify;
			}

			if (qu_var_is_class (x))
			{
				if (((qu_var) x)->type)
				{
					val = qu_object_validate (c, ((qu_var) x)->type, val);
					if (val == 0)
						goto bad;
				}
				((qu_var) x)->value = val;
				return val;
			}
			else
			{
				if (((qu_var) x)->type)
				{
					val = qu_object_validate (self, ((qu_var) x)->type, val);
					if (val == 0)
						goto bad;
				}
				if (self->attrs == NULL)
				{
					/* no need to set if it equals the default value
						- only quickly test for address equality
						- this might be a big win in some cases */
					if (val == ((qu_var) x)->value)
						return val;

					_attr_inst_alloc (self);
				}
				i = (entry_s *) ((void *) self->attrs->items + i->index * sizeof (qu_any));
				i->value = val;
				return val;
			}
		}
		else
		{
#ifdef QU_SUB_ALLOW_ASSIGN
			assert (qu_sub_is (x));

			if (restrict)
			{
				if (qu_symbol_is_private (x))
					goto nope;
			}
			if (qu_sub_is_class (x))
				return qu_sub_call1 (x, c, val);
			return qu_sub_call1 (x, self, val);
#else
			goto modify;
#endif
		}
	}

  nope:
	{
		qu_sub x = _attr_inst_get (self, QU_NAME__set, 0);
		if (x && qu_sub_is (x) && qu_sub_is_inst (x))
		{
			QU_ENSURE (qu_sub_call2 (x, self, key, val));
			return val;
		}
	}

	QU_THROWF (QuEImplement, "object %s has no attribute %s", c->name->str, key->str);

  bad:
	QU_THROWF (QuEValue, "mismatch value/type %s.%s", c->name->str, key->str);

  modify:
	QU_THROWF (QuEModify, "cant modify %s.%s", c->name->str, key->str);
}
