/* gc.c - Copyright (c) Marc Krisnanto */

#undef GC_DEBUG_HEAVY

#ifdef GC_DEBUG_HEAVY
#define QU_DEBUG_LOCAL
#endif

#undef GC_PRINT_CHAR
#ifdef GC_PRINT_CHAR
#define GC_PRINT_CHAR_ARENA_BEGIN   "A"
#define GC_PRINT_CHAR_ARENA_END     "a"
#define GC_PRINT_CHAR_RECLAIM_BEGIN "R"
#define GC_PRINT_CHAR_RECLAIM_END   "r"
#endif


#include "qu.h"

/*
	Arena = memory block allocated via malloc with a size of ARENA_SIZE.

	Arenas are divided into pages and pages are divided into memory chunks
	of the same size.

	Small memory chunks are allocated from a page. qu_big chunks are allocated
	via malloc.

	- - - -+-----------------------------------------------+- - - -
	       | Arena                                         |
	       | +---------------------+---------------------+ |
	       | | Page                | Page                | |
	       | | +-----+-----+-----+ | +-----+-----+-----+ | |
	       | | |chunk|chunk|chunk| | |chunk|chunk|chunk| | |
	       | | +-----+-----+-----+ | +-----+-----+-----+ | |
	       | +---------------------+---------------------+ |
	- - - -+-----------------------------------------------+- - - -

	Pages are PAGE_SIZE big and aligned to that size. Aligned because we
	want to be able to mask out the low bits of a pointer to get the start of
	the page. The start of the page (the page header) contains information of
	the size, the arena it belongs to, etc.

	When a page is empty, it is put in FreePage, a single-link-list.

	When all pages of an arena are freed, the pages are retrieved out from the
	FreePage list then the arena is freed. But we only do this if the total
	number of arenas is more than ARENA_LIMIT.

	We keep a list of used pages for a number of defined chunk sizes in pools.
	Pools simply keeps track of pages of the same chunk size and a pointer
	to a free chunk.

	We don't keep a free-chunk list. We keep track of the number of allocated
	chunks instead. When a chunk is freed we simply decrement the count.
	If the count reaches zero, the page is put back in the FreePage list
	for reuse. The following illustrates this:


	  after the            after                        after
	1st qu_mem_alloc    2x qu_mem_alloc               qu_mem_free C, A
	---------------    --------------               --------------
	count=0               count=3                      count=1
	+-----+               +-----+                      +-----+
	|  D  |               |  D  | <-- free             |  D  | <-- free
	+-----+               +-----+                      +-----+
	|  C  |               |::C::|                      |**C**|
	+-----+               +-----+                      +-----+
	|  B  | <-- free      |::B::|                      |::B::|
	+-----+               +-----+                      +-----+
	|::A::|               |::A::|                      |**A**|
	+-----+               +-----+                      +-----+
	  P1                    P1                            P1

	   after                    after                   after
	1x qu_mem_alloc          1x qu_mem_alloc          qu_mem_free B, D
	--------------          --------------          --------------
	count=2               count=2    count=1           count=1
	+-----+               +-----+ -> +-----+           +-----+
	|::D::|               |::D::|    |  H  |           |  H  |
	+-----+               +-----+    +-----+           +-----+
	|**C**|               |**C**|    |  G  |           |  G  |
	+-----+               +-----+    +-----+           +-----+
	|::B::|               |::B::|    |  F  | <-- free  |  F  | <-- free
	+-----+               +-----+    +-----+           +-----+
	|**A**|               |**A**|    |::E::|           |::E::|
	+-----+               +-----+    +-----+           +-----+
	  P1                    P1         P2                P2
	free = NULL
	                                                   FreePage
	                                                   +-----+ -> ...
	                                                   |     |
	                                                   +-----+
	                                                   |     |
	                                                   +-----+
	                                                   |     |
	                                                   +-----+
	                                                   |     |
	                                                   +-----+
	                                                     P1

	This means that there is a chance of over-allocation, i.e. when a page
	contains a chunk that is never freed while other chunks are freed. The
	freed chunks will never be used again.

	We let this happen because:
	- constant objects are rare
	- otherwise we will have to maintain a free chunk list and that will drop
	  performance by an untolerable amount

	Note that it is possible to free a page while there are free chunks in it.

	+-----+              After qu_mem_free A and B, the page is
	|     |              freed (put in the FreePage list) even
	+-----+              if there are still free chunks in it.
	|     | <- free
	+-----+              If this happens, we set pool->free to NULL.
	|::B::|              See qu_mem_free.
	+-----+
	|::A::|
	+-----+

	Static objects are objects that do not refer to other objects and are
	never freed. They are allocated via a static block or malloc or perhaps
	qu_mem_alloc. These are never freed even at exit. This relieves the GC
	from marking them. There are at least 1600 static objects created by Qu
	so this is a win.

	Root objects (TYPE_ROOT) are objects that must never be freed. They are
	containers that refer to other objects.

	Regular objects (TYPE_REGULAR) are non-root non-static object. They may,
	of course, refer to other objects. They are automatically freed by the GC
	(qu_gc_reclaim) when they are no longer referenced.

	Like previous versions of Qu, memory (TYPE_MEMORY) are not reclaimed by
	the GC. It is the responsibility of the objects to do so. Allocating
	memory does not trigger GC. It may only set a flag to tell the GC that
	it is time to collect garbage.

	Note
	----
	Since we are using mark and sweep, the efficiency of the GC is practically
	determined on how fast we can mark the C stack. If you want to improve,
	try focus on that.
*/


/*----------------------------------------------------------------------------
	__QuGcMutex - the global lock
----------------------------------------------------------------------------*/

/*
	This it the global GC lock. Only a single thread holds this lock at
	any given time.

	Note: years ago I experimented with a lock-per-object scheme.
	The performance is outrageously disappointing. If you know better,
	let me know. -marc-
*/
qu_thread_mutex_t __QuGcMutex = QU_THREAD_MUTEX_INITIALIZER;


#ifdef DEBUG
#define ASSERT_THREAD_NOT_BLOCKING \
	if (__QuThreadCurrent && qu_thread_blocking (__QuThreadCurrent)) \
		QU_FATAL ("thread " QU_THREAD_IDFORMAT " allocating or freeing memory while blocking", qu_thread_id ());
#else
#define ASSERT_THREAD_NOT_BLOCKING
#endif



/*----------------------------------------------------------------------------
	definitions - some are tunable
----------------------------------------------------------------------------*/

/* define this to get object statistics */
#undef COLLECT_STATISTICS

/* define this to free everything at exit - except, of course, static memory */
#undef FREE_AT_EXIT

/* define to hunt for free chunks when we run out of memory */
#undef HUNT_MEMORY

/* size of one page including page header */
#define PAGE_SIZE 8192

/* PAGESIZE size must be such that a page must at least contains 2 chunks */
#if PAGE_SIZE < (QU_GC_SIZE_LIMIT + QU_GC_SIZE_LIMIT)
#error PAGE_SIZE is too damn small
#endif

/* pages per arena - makes one arena about 256KB */
#define PAGES_PER_ARENA 32

/* maximum number of arenas to hold - totals to about 16MB */
#define ARENA_LIMIT 64

/* size per-allocation */
#define ARENA_SIZE (PAGES_PER_ARENA * PAGE_SIZE)

/* mask to get the start of the page */
#define PAGE_MASK ((unsigned long) (PAGE_SIZE - 1))

/* from pointer to page */
#define PTR_TO_PAGE(m) ((page_s *) ((unsigned long) (m) & ~PAGE_MASK))

/* number of chunks/objects per page */
#define CHUNKS_PER_PAGE(size) ((PAGE_SIZE - sizeof (page_s)) / (size))

/* address of first chunk in a page */
#define PAGE_FIRST_CHUNK_ADDR(g)  ((void *)(g) + sizeof (page_s))

/* address of last chunk in a page */
#define PAGE_LAST_CHUNK_ADDR(g, size) (PAGE_FIRST_CHUNK_ADDR(g) + (CHUNKS_PER_PAGE(size) - 1) * (size))

/* memory types */
#define TYPE_ROOT    3 /* root objects */
#define TYPE_REGULAR 2 /* regular objects, not root nor static */
#define TYPE_MEMORY  1 /* memory, not object */
#define TYPE_NONE    0 /* nothing */

typedef struct arena_s arena_s;
typedef struct page_s  page_s;
typedef struct pool_s  pool_s;
typedef struct mem_s   mem_s;

struct arena_s
{
	arena_s *next;  /* next in list */
	arena_s *prev;  /* previous in list */
	int      count; /* counter - indicates to free the arena */
	void    *mem;   /* allocated block with malloc (ARENA_SIZE) */
	page_s  *first; /* first page address so we don't need to recalculate it */
	page_s  *last;  /* last -ditto- */
};

#define PAGEHEAD \
	page_s  *next;  /* next page in list */ \
	page_s  *prev;  /* previous page in list - while in used pool */ \
	arena_s *arena; /* arena this page belongs to */ \
	int      count; /* allocation counter - indicates to free the page */ \
	size_t   size;  /* size of chunk */ \
	int      type   /* one of TYPE_XXX */

struct page_s
{
	PAGEHEAD;
	/* memory chunks starts after the struct */
};

struct pool_s
{
	PAGEHEAD;
	void *free;  /* pointer to free chunk - type MUST be (void *) */
	int   total; /* total number of chunks = CHUNKS_PER_PAGE(size) */
};


/* must match QU_OBJECT_HEAD since we're using the same arenas for both */
struct mem_s
{
	union
	{
		size_t size;  /* chunk size - including sizeof (mem_s) */
		qu_class  klass;
	} m;

	long    flags;
	qu_set  weak;

#ifdef QU_DEBUG_SEGV
	const char *func; /* function that allocated the memory */
#else
	qu_attr_s *attrs;
#endif

	/* user memory starts after the struct */
	/* if QU_DEBUG_SEGV, we also put a unsigned long magic at the end */
};

#define MAGIC 0x6B756366UL
#define MEM_END_MAGIC_PTR(m, size) ((long *)((void *) (m) + size - sizeof (long)))


#define m_size m.size
#if SIZEOF_SIZE_T == SIZEOF_VOID_P
#define m_free m.size
#else
#define m_free m.klass
#endif

/* from mem_s to void* and vice versa */
#define M2P(m) ((void *) (m + 1))
#define P2M(p) (((mem_s *) p) - 1)

/* the total amount of active memory */
unsigned long volatile __QuGcMemAllocated = 0;

/*
	Setting RECLAIM_THRESHOLD to a big value does not help performance
	since it will simply delay GC and when GC does execute then it will
	take more time because there are more objects.
*/
#define RECLAIM_THRESHOLD (1UL * 1024UL * 1024UL)
#define RECLAIM_MAX       (128UL * 1024UL * 1024UL)
static unsigned long GcReclaimLimit = RECLAIM_THRESHOLD;

/* setting this to 0 will disable GC entirely */
int volatile __QuGcDisable = 0;

/* do we need to reclaim? */
int volatile __QuGcNeedReclaim = 0;

/* the sizes */
#define SIZE_NONE 0
#define NUMSIZES  7
static int Sizes[NUMSIZES] =
    { 16, 32, 64, 128, 256, 512, 1024 };
	                            /* |
                        +----------+
                        | */
#if QU_GC_SIZE_LIMIT > 1024
#error QU_GC_SIZE_LIMIT need to be corrected
#endif

/* the pools */
#define P_(type) \
	{ NULL, NULL, NULL, 0,   16, type , NULL, CHUNKS_PER_PAGE (  16)}, \
	{ NULL, NULL, NULL, 0,   32, type , NULL, CHUNKS_PER_PAGE (  32)}, \
	{ NULL, NULL, NULL, 0,   64, type , NULL, CHUNKS_PER_PAGE (  64)}, \
	{ NULL, NULL, NULL, 0,  128, type , NULL, CHUNKS_PER_PAGE ( 128)}, \
	{ NULL, NULL, NULL, 0,  256, type , NULL, CHUNKS_PER_PAGE ( 256)}, \
	{ NULL, NULL, NULL, 0,  512, type , NULL, CHUNKS_PER_PAGE ( 512)}, \
	{ NULL, NULL, NULL, 0, 1024, type , NULL, CHUNKS_PER_PAGE (1024)},

static pool_s PoolRoot[NUMSIZES] = { P_(TYPE_ROOT)    };
static pool_s PoolReg [NUMSIZES] = { P_(TYPE_REGULAR) };
static pool_s PoolMem [NUMSIZES] = { P_(TYPE_MEMORY)  };

/* quick map size to pool index */
static int *PoolIndex;

/* free empty pages */
static page_s *FreePage = NULL;

/* list of arenas */
static arena_s *Arenas;

/* total number of arenas */
static int ArenaCount = 0;

/* lowest and highest address of arenas */
static page_s *ArenaPageLo = (void *)((unsigned long) -1);
static page_s *ArenaPageHi = 0;

#ifdef QU_DEBUG_LOCAL
#define GC_STATE_IDLE           0
#define GC_STATE_MARK_ROOT      1
#define GC_STATE_MARK_REGULAR   2
#define GC_STATE_MARK_STACK     3
#define GC_STATE_MARK_REGISTERS 4
#define GC_STATE_SWEEP          5
static int     GcState = GC_STATE_IDLE; /* current state */
static page_s *GcSweepTrackPage;        /* to prove the algorithm */
#endif

/* pre-allocated static memory */
#define STATIC_MEMORY_SIZE (128 * 1024)
static void  *StaticMem;
static void  *StaticPtr;
static size_t StaticRem = STATIC_MEMORY_SIZE;

/*----------------------------------------------------------------------------
	some statistics
----------------------------------------------------------------------------*/

#ifdef COLLECT_STATISTICS
static long volatile NumRegulars = 0;
static long volatile NumRoots    = 0;
static long volatile NumStatics  = 0;

long qu_gc_num_regulars () { return NumRegulars; }
long qu_gc_num_roots    () { return NumRoots;    }
long qu_gc_num_statics  () { return NumStatics;  }

#else

long qu_gc_num_regulars () { return 0; }
long qu_gc_num_roots    () { return 0; }
long qu_gc_num_statics  () { return 0; }

#endif /*COLLECT_STATISTICS*/


void qu_gc_stat ()
{
	/* TODO - more info? */
#ifdef COLLECT_STATISTICS
	fprintf (stderr, "Objects\n");
	fprintf (stderr, "- regular = %ld\n", NumRegulars);
	fprintf (stderr, "- root    = %ld\n", NumRoots);
	fprintf (stderr, "- static  = %ld\n", NumStatics);
	fprintf (stderr, "\n");
#endif
	fprintf (stderr, "Memory\n");
	fprintf (stderr, "- arena size = %ld\n", (long) ARENA_SIZE);
	fprintf (stderr, "- page size  = %ld\n", (long) PAGE_SIZE);
	fprintf (stderr, "- arenas     = %ld\n", (long) ArenaCount);
	fprintf (stderr, "- allocated  = %ld\n", (long) ArenaCount * ARENA_SIZE);
	fprintf (stderr, "- active     = %ld\n", (long) __QuGcMemAllocated);
	fprintf (stderr, "\n");
}



/*----------------------------------------------------------------------------
	trap
----------------------------------------------------------------------------*/

/* current trap */
qu_gcmemtrap_f __QuGcMemTrap = 0;

static void _gc_mem_trap ()
{
	static int here = 0;

	if (here == 0 && __QuGcMemTrap)
	{
		__QuGcMemTrap ();
		here = 1;
	}
	QU_FATAL ("out of memory");
}



/*----------------------------------------------------------------------------
	critical
----------------------------------------------------------------------------*/

#define QU_USE_CRITICAL_MEMORY 1

#ifdef QU_USE_CRITICAL_MEMORY

/* pre-allocated critical memory - size should be a command line? */
#define CRITICAL_MEMORY_SIZE (128 * 1024)
static void  *CriticalMem;
static void  *CriticalPtr;
static size_t CriticalRem = CRITICAL_MEMORY_SIZE;
static int    Critical = 0;


static void *_gc_mem_alloc_critical (size_t size)
{
	void *p;

	Critical = 1;
	ASSERT_THREAD_NOT_BLOCKING
	if (CriticalRem > size)
	{
		if (qu_int_is (CriticalPtr))
		{
			++CriticalPtr;
			--CriticalRem;
		}
		p = CriticalPtr;
		CriticalPtr += size;
		CriticalRem -= size;
		return p;
	}
	return 0;
}
#endif /*QU_USE_CRITICAL_MEMORY*/



/*----------------------------------------------------------------------------
	setup
----------------------------------------------------------------------------*/

static void _gc_setup ()
{
	int i, j, z;

	Arenas = calloc (1, sizeof (arena_s));
	if (Arenas == NULL)
		_gc_mem_trap ();
	Arenas->prev = Arenas->next = Arenas;

	for (i = 0; i < NUMSIZES; i++)
	{
		PoolRoot[i].prev = PoolRoot[i].next = (page_s *) &PoolRoot[i];
		PoolReg [i].prev = PoolReg [i].next = (page_s *) &PoolReg [i];
		PoolMem [i].prev = PoolMem [i].next = (page_s *) &PoolMem [i];
	}

	PoolIndex = malloc ((QU_GC_SIZE_LIMIT + 1) * sizeof (int));
	if (PoolIndex == NULL)
		_gc_mem_trap ();

	for (i = z = 0; i < NUMSIZES; z = Sizes[i++] + 1)
		for (j = z; j <= Sizes[i]; j++)
			PoolIndex [j] = i;
}



/*----------------------------------------------------------------------------
	debugging tools
----------------------------------------------------------------------------*/

#ifdef QU_DEBUG_LOCAL
static int _gc_page_ensure_size (page_s *g)
{
	int i;
	for (i = 0; i < NUMSIZES; i++)
	{
		if (Sizes[i] == g->size)
			return 1;
	}
	return 0;
}



static int _gc_page_ensure_count (page_s *g)
{
	if (0 <= g->count && g->count <= CHUNKS_PER_PAGE (g->size))
		return 1;
	return 0;
}



static int _gc_page_ensure_in_regular_pool (page_s *g)
{
	register pool_s *r;
	register page_s *x;
	register int     i;

	for (i = 0; i < NUMSIZES; i++)
	{
		r = &PoolReg [i];
		for (x = r->next; x != (page_s *) r; x = x->next)
		{
			if (g == x)
				return 1;
		}
	}
	return 0;
}



static int _gc_page_ensure_in_mem_pool (page_s *g)
{
	pool_s *r;
	page_s *x;
	int     i;

	for (i = 0; i < NUMSIZES; i++)
	{
		r = &PoolMem [i];
		for (x = r->next; x != (page_s *) r; x = x->next)
		{
			if (g == x)
				return 1;
		}
	}
	return 0;
}


static int _gc_page_in_free_list (page_s *g)
{
	page_s *x;

	for (x = FreePage; x; x = x->next)
	{
		if (g == x)
			return 1;
	}
	return 0;
}


static void _gc_page_ensure_setup (page_s *g, size_t size)
{
	register int    n = CHUNKS_PER_PAGE (size);
	register mem_s *e = PAGE_FIRST_CHUNK_ADDR (g);

	while (n--)
	{
		if (e->m_free)
			QU_FATAL ("chunk not marked as freed");
		e = (mem_s *) ((void *) e + size);
	}
}

#endif /*QU_DEBUG_LOCAL*/



/*----------------------------------------------------------------------------
	helpers
----------------------------------------------------------------------------*/

inline static int _gc_page_in_arena_extremes (page_s *g)
{
	return ArenaPageLo <= g && g <= ArenaPageHi;
}



inline static int _gc_page_in_arena (page_s *p)
{
	arena_s *a;

	for (a = Arenas->prev; a != Arenas; a = a->prev)
	{
		if (a->first <= p && p <= a->last)
			return 1;
	}
	return 0;
}



inline static int _gc_addr_is_chunk (page_s *g, void *o)
{
	return ((void *) g + sizeof (page_s)) <= o
		&& o <= ((void *) g + PAGE_SIZE - g->size)
		&& (((unsigned long) o - (unsigned long) g - sizeof (page_s)) % g->size) == 0;
}




/*----------------------------------------------------------------------------
	arena allocation
----------------------------------------------------------------------------*/

/*
	This function is called when the FreePage list is empty.
	It allocates a new arena then divide it into pages.
	Returns >0 on success, 0 on failure.
*/
static int _gc_more_arena ()
{
	arena_s      *a;
	page_s       *p;
	unsigned long d;
	int           n;

#ifdef GC_PRINT_CHAR
	fprintf (stderr, GC_PRINT_CHAR_ARENA_BEGIN);
	fflush (stderr);
#endif

	assert (FreePage == NULL);

	a = malloc (sizeof (arena_s));
	if (a == NULL)
		return 0;

	p = malloc (ARENA_SIZE);
	if (p == NULL)
	{
		free (a);
		return 0;
	}

	a->mem   = p;
	a->count = 0;
	a->prev  = Arenas;
	a->next  = Arenas->next;

	Arenas->next->prev = a;
	Arenas->next = a;

	/* pages must be aligned to PAGE_SIZE */
	n = ARENA_SIZE / PAGE_SIZE;
	d = ((unsigned long) p) & PAGE_MASK;
	if (d != 0)
	{
		p = (page_s *) ((void *) p + PAGE_SIZE - d);
		--n;
	}

	a->first = FreePage = p;

	/* setup single-link-list pages */
	while (--n)
	{
		p->next  = (page_s *) ((void *) p + PAGE_SIZE);
		p->arena = a;
		p->size  = SIZE_NONE;
		p->type  = TYPE_NONE;
		p = p->next;
	}
	p->next  = NULL;
	p->arena = a;
	p->size  = SIZE_NONE;
	p->type  = TYPE_NONE;
	a->last  = p;

	if (a->first < ArenaPageLo)
		ArenaPageLo = a->first;

	if (a->last > ArenaPageHi)
		ArenaPageHi = a->last;

#ifdef QU_DEBUG_LOCAL
	fprintf (stderr, "arena allocated (%p) %p .. %p\n", a, a->first, a->last);
	fprintf (stderr, "   Arenas         : %d\n" , ArenaCount + 1);
	fprintf (stderr, "   GcMemAllocated : %ld\n", __QuGcMemAllocated);
	fprintf (stderr, "   GcReclaimLimit : %ld\n", GcReclaimLimit);
	fprintf (stderr, "   GcNeedReclaim  : %d\n" , __QuGcNeedReclaim);
#endif

#ifdef GC_PRINT_CHAR
	printf (GC_PRINT_CHAR_ARENA_END);
	fflush (stdout);
#endif

	return ++ArenaCount;
}



/*
	When all pages in an arena are in the FreePage list (they are all
	not used) and there are more arenas then ARENA_LIMIT then we free
	the arena.

	Thus we must retrieve all the arena's pages from the FreePage list.
*/
static void _gc_free_arena (arena_s *a)
{
	register page_s *lo = a->first;
	register page_s *hi = a->last;
	register page_s *r, *q, *x;

	while (lo <= FreePage && FreePage <= hi)
		FreePage = FreePage->next;

	for (r = q = FreePage; q; q = x)
	{
		x = q->next;
		if (q < lo || q > hi)
		{
			r->next = q;
			r = q;
		}
#ifdef QU_DEBUG_LOCAL
		else if (GcState == GC_STATE_SWEEP)
		{
			if (q == GcSweepTrackPage)
				QU_FATAL ("wrong algorithm - sweep has lost the next page");
		}
#endif
	}

	if (r)
		r->next = NULL;

	a->prev->next = a->next;
	a->next->prev = a->prev;

	free (a->mem);
	free (a);

#ifdef QU_DEBUG_LOCAL
	fprintf (stderr, "arena freed %p\n", a);
#endif
	--ArenaCount;
}



/*
	Pages might be used by memory (TYPE_MEMORY) or by objects.

	We mark chunks as being freed by setting the first field to 0.

	This is required not only by _gc_mem_hunt but also by the object
	allocator and garbage collector to mark that an object is freed so
	it does not get marked.

	!!! Only call this if the page size is not the same as the intended
	!!! size. If it is the same, it means that it has already been done.
*/
inline static void _gc_page_setup (page_s *g, size_t size)
{
	int    n = CHUNKS_PER_PAGE (size);
	mem_s *e = PAGE_FIRST_CHUNK_ADDR (g);

	assert (g->size != size);

	while (n--)
	{
		e->m_free = 0;
		e = (mem_s *) ((void *) e + size);
	}
	g->size = size;
}



/*----------------------------------------------------------------------------
	memory allocation (TYPE_MEMORY)
----------------------------------------------------------------------------*/

#ifdef QU_DEBUG_SEGV
void qu_mem__free (void *p, const char *func)
#else
void qu_mem_free (void *p)
#endif
{
	mem_s  *m;
	size_t  size;

	ASSERT_THREAD_NOT_BLOCKING

	assert (p);

	m = P2M (p);
	size = m->m_size;

#ifdef QU_DEBUG_SEGV
	if (*MEM_END_MAGIC_PTR (m, size) != MAGIC)
	{
		char *z = (char *) m;
		int   n = 100;
		while (n--)
		{
			if (*z < ' ' || *z > 126)
				fprintf (stderr, " ");
			else
				fprintf (stderr, "%c", *z);
			++z;
		}
		fprintf (stderr, "\n");
		/* lets hope this works */
		fprintf (stderr, "allocator: %s\n", m->func);
		QU_FATAL ("MEMORY MAGIC OVERIDDEN %s", func);
	}
#endif


	assert (size > sizeof (mem_s));

	__QuGcMemAllocated -= size;

	if (size <= QU_GC_SIZE_LIMIT)
	{
		page_s *g = PTR_TO_PAGE (m);

		/* this marks the chunk as free - see _gc_page_setup */
		m->m_free = 0;

		assert (g->count > 0);
		if (--(g->count) == 0)
		{
			arena_s *a = g->arena;
			pool_s  *p = &PoolMem[PoolIndex[size]];

			assert (p->type == TYPE_MEMORY);
			assert (size == p->size);

			/*
				put back page in FreePage list
			*/
			g->type = TYPE_NONE;
			g->next->prev = g->prev;
			g->prev->next = g->next;
			g->next = FreePage;
			FreePage = g;

			/*
				the page might be freed when the free chunk pointer still
				points in it, so take care of that
			*/
			if (PTR_TO_PAGE (p->free) == g)
				p->free = NULL;

			/*
				if all pages in the arena are in the free pages list
				then free the arena unless we haven't allocated
				the specified limit
			*/
			if (--(a->count) == 0 && ArenaCount > ARENA_LIMIT)
				_gc_free_arena (a);
		}
	}
	else
	{
		/* it was allocated with malloc */
		free (m);
	}
}



/*
	When we run out of memory, we try to hunt into a pool to
	find a free chunk.

	TODO - look into pools with higher sizes too?
*/
#ifdef HUNT_MEMORY
inline static void *_gc_mem_hunt (pool_s *p)
{
	page_s *s = p->next;
	int     n = p->total;
	size_t  z = p->size;
	page_s *g;

	for (g = s; g != s; g = g->next)
	{
		if (g->count < n)
		{
			register int    k = n;
			register void * m = PAGE_FIRST_CHUNK_ADDR (g);

			while (k--)
			{
				if (((mem_s *) m)->m_free == 0)
				{
					g->count++;
					return (void *) m;
				}
				m += z;
			}
		}
	}
	return NULL;
}
#endif



#ifdef QU_DEBUG_SEGV
void *qu_mem__alloc (size_t size, const char *func)
#else
void *qu_mem_alloc (size_t size)
#endif
{
	mem_s *m;

	ASSERT_THREAD_NOT_BLOCKING
	assert (0 < size && size <= QU_MEM_MAX);

#ifdef QU_DEBUG_SEGV
	size += sizeof (mem_s) + sizeof (long);
#else
	size += sizeof (mem_s);
#endif

#ifdef QU_USE_CRITICAL_MEMORY
	if (Critical)
	{
		m = _gc_mem_alloc_critical (size);
		if (m)
			goto done;
		_gc_mem_trap ();
	}
#endif

	if (size <= QU_GC_SIZE_LIMIT)
	{
		pool_s *p = &PoolMem[PoolIndex[size]];
		page_s *g;

		assert (size <= p->size);
		size = p->size;

		if (p->free == NULL)
		{
			if (FreePage == NULL)
			{
				if (!_gc_more_arena ())
				{
#ifdef HUNT_MEMORY
					m = _gc_mem_hunt (p);
					if (m)
						goto done;
#endif
#ifdef QU_USE_CRITICAL_MEMORY
					Critical = 1;
				    m = _gc_mem_alloc_critical (size);
					if (m)
						goto done;
#endif
					_gc_mem_trap ();
					QU_FATAL ("never");
				}
			}

			g = FreePage;
			FreePage = g->next;
			m = (mem_s *) PAGE_FIRST_CHUNK_ADDR (g);
			g->arena->count++;

			if (g->size != size)
				_gc_page_setup (g, size);
#ifdef QU_DEBUG_LOCAL
			_gc_page_ensure_setup (g, size);
#endif

			g->count = 1;
			g->type  = TYPE_MEMORY;
			g->prev  = p->prev;
			g->next  = (page_s *) p;
			p->prev->next = g;
			p->prev = g;

			p->free = (void *) m + size;
		}
		else
		{
			m = p->free;
			g = PTR_TO_PAGE (m);
			g->count++;
			p->free += size;
			if (p->free > PAGE_LAST_CHUNK_ADDR (g, size))
				p->free = NULL;
		}
	}
	else
	{
		m = malloc (size);
		if (m == NULL)
		{
#if 0
			if (!__QuGcDisable)
			{
				qu_gc_reclaim ();
				m = malloc (size);
				if (m)
					goto done;
			}
#endif
#ifdef QU_USE_CRITICAL_MEMORY
			Critical = 1;
		    m = _gc_mem_alloc_critical (size);
			if (m)
				goto done;
#endif
			_gc_mem_trap ();
			QU_FATAL ("never");
		}
	}

#if defined(HUNT_MEMORY) || defined (QU_USE_CRITICAL_MEMORY)
  done:
#endif

	__QuGcMemAllocated += size;
	if (__QuGcMemAllocated >= GcReclaimLimit)
		__QuGcNeedReclaim = 1;

	m->m_size = size;

#ifdef QU_DEBUG_SEGV
	m->func = func;
	*MEM_END_MAGIC_PTR (m, size) = MAGIC;
#endif

	return M2P (m);
}



void *qu_mem_static (size_t size)
{
	void *p;

	ASSERT_THREAD_NOT_BLOCKING
	assert (0 < size && size <= QU_MEM_MAX);

	if (StaticRem > size)
	{
		if (qu_int_is (StaticPtr))
		{
			++StaticPtr;
			--StaticRem;
		}
		p = StaticPtr;
		StaticPtr += size;
		StaticRem -= size;
		return p;
	}
	p = malloc (size);
	if (p == 0)
		p = qu_mem_alloc (size);
	return p;
}



#ifdef QU_DEBUG_SEGV
void *qu_mem__realloc (void *p, size_t size, const char *func)
#else
void *qu_mem_realloc (void *p, size_t size)
#endif
{
	mem_s  *m;
	size_t  z;
	void   *r;

	ASSERT_THREAD_NOT_BLOCKING

	assert (p);
	assert (0 < size && size < QU_MEM_MAX);

	m = P2M (p);
	assert (m->m_size > sizeof (mem_s));
	z = m->m_size - sizeof (mem_s);

	/* lets be reasonable */
	if (z / 2 <= size && size <= z)
		return p;

#ifdef QU_DEBUG_SEGV
	r = qu_mem__alloc (size, func);
#else
	r = qu_mem_alloc (size);
#endif
	memcpy (r, p, (size < z) ? size : z);
	qu_mem_free (p);
	return r;
}



#ifdef QU_DEBUG_SEGV
void *qu_mem__calloc (size_t nelt, size_t size, const char *func)
#else
void *qu_mem_calloc (size_t nelt, size_t size)
#endif
{
	ASSERT_THREAD_NOT_BLOCKING

	assert (size > 0 && nelt > 0 && nelt < (QU_MEM_MAX - nelt - sizeof (mem_s)) / size);
	size *= nelt;
#ifdef QU_DEBUG_SEGV
	return memset (qu_mem__alloc (size, func), 0, size);
#else
	return memset (qu_mem_alloc (size), 0, size);
#endif
}


#ifdef QU_DEBUG_SEGV
void *qu_mem__copy (void *p, const char *func)
#else
void *qu_mem_copy (void *p)
#endif
{
	mem_s  *m;
	size_t  z;
	void   *r;

	ASSERT_THREAD_NOT_BLOCKING

	assert (p);
	m = P2M (p);
	assert (m->m_size > sizeof (mem_s));
	z = m->m_size - sizeof (mem_s);
#ifdef QU_DEBUG_SEGV
	r = qu_mem__alloc (z, func);
#else
	r = qu_mem_alloc (z);
#endif
	memcpy (r, p, z);
	return r;
}



/*----------------------------------------------------------------------------
	weak reference
----------------------------------------------------------------------------*/


int qu_gc_weak_set (qu_obj self, qu_weak w)
{
	if (self->weak == NULL)
	{
		self->weak = qu_set_new (4);
		qu_set_add (self->weak, w);
	}
	else
	{
		QU_ENSURE (qu_set_addif (self->weak, w));
	}
	return 1;
}



void qu_gc_weak_delete (qu_obj self, qu_weak w)
{
	assert (self->weak);
	qu_set_del (self->weak, w);
}



static void qu_gc_weak_destroy (qu_obj self)
{
	qu_weak w;
	long    b = 0, i = 0;

	assert (self->weak);
	assert (qu_set_is (self->weak));
	while ((w = qu_set_walk (self->weak, &b, &i)) != 0)
	{
		assert (qu_weak_is (w));
		QU_WEAK_UNREF (w);
	}
	self->weak = 0;
}




/*----------------------------------------------------------------------------
	object allocation
----------------------------------------------------------------------------*/

inline static void _gc_destroy (qu_obj o)
{
	page_s *g;
	size_t  size;

	assert (o);
	assert (!qu_int_is (o));

	if (o->weak)
		qu_gc_weak_destroy (o);

	g = PTR_TO_PAGE (o);
	size = g->size;
	__QuGcMemAllocated -= size;

	/* this marks the object as being freed - see _gc_page_setup */
	((mem_s *) o)->m_free = 0;

#ifdef COLLECT_STATISTICS
	NumRegulars--;
#endif

	assert (g->count > 0);
	if (--(g->count) == 0)
	{
		arena_s *a = g->arena;
		pool_s  *p = &PoolReg[PoolIndex[size]];

		g->prev->next = g->next;
		g->next->prev = g->prev;
		g->next = FreePage;
		FreePage = g;

		if (PTR_TO_PAGE (p->free) == g)
			p->free = NULL;

		if (--(a->count) == 0 && ArenaCount > ARENA_LIMIT)
			_gc_free_arena (a);
	}
}



void qu_gc_delete (qu_obj o)
{
	assert (!qu_int_is (o));

	ASSERT_THREAD_NOT_BLOCKING

#ifdef QU_DEBUG_LOCAL
	if (GcState != GC_STATE_IDLE)
		QU_FATAL ("wrong implementation - cant GcDelete while reclaiming");
#endif

	qu_attr_xfree (o->attrs);
	if (o->klass->dtor)
		o->klass->dtor (o);
	_gc_destroy (o);
}


static qu_any _gc_alloc_object (pool_s *p, qu_class c, int type)
{
	size_t  size = p->size;
	page_s *g;
	qu_obj  o;

	/*
		dont put this in the middle otherwise there may be situations where
		memory is never reclaimed because when qu_mem_alloc allocates new
		arena then p->free might never be NULL until memory is exhausted
	*/
	if (__QuGcNeedReclaim)
		qu_gc_reclaim ();

	if (p->free == NULL)
	{
		if (FreePage == NULL)
		{
			if (_gc_more_arena () == 0)
			{
#ifdef HUNT_MEMORY
				o = _gc_mem_hunt (p);
				if (o)
					goto done;
#endif
				_gc_mem_trap ();
				QU_FATAL ("never");
			}
		}

		g = FreePage;
		FreePage = g->next;

		g->arena->count++;

		if (g->size != size)
			_gc_page_setup (g, size);
#ifdef QU_DEBUG_LOCAL
		_gc_page_ensure_setup (g, size);
#endif

		g->count = 1;
		g->type  = type;
		g->prev  = p->prev;
		g->next  = (page_s *) p;

		p->prev->next = g;
		p->prev = g;

		o = (qu_obj) PAGE_FIRST_CHUNK_ADDR (g);
		p->free = (void *) o + size;
	}
	else
	{
		o = p->free;
		g = PTR_TO_PAGE (o);
		assert (g->type == type);
		g->count++;
		p->free += size;
		if (p->free > PAGE_LAST_CHUNK_ADDR (g, size))
			p->free = NULL;
	}

#ifdef HUNT_MEMORY
  done:
#endif

	__QuGcMemAllocated += size;
	if (__QuGcMemAllocated >= GcReclaimLimit)
		__QuGcNeedReclaim = 1;

	o->klass = c;
	o->flags = 0;
	o->attrs = 0;
	o->weak  = 0;

	return (qu_any) o;
}



qu_any qu_gc__new (qu_class c, size_t size)
{
	ASSERT_THREAD_NOT_BLOCKING

	assert (size && size <= QU_GC_SIZE_LIMIT);
#ifdef COLLECT_STATISTICS
	NumRegulars++;
#endif
	return _gc_alloc_object (&PoolReg[PoolIndex[size]], c, TYPE_REGULAR);
}



qu_any qu_gc__new_root (qu_class c, size_t size)
{
	ASSERT_THREAD_NOT_BLOCKING

	assert (size && size <= QU_GC_SIZE_LIMIT);
#ifdef COLLECT_STATISTICS
	NumRoots++;
#endif
	return _gc_alloc_object (&PoolRoot[PoolIndex[size]], c, TYPE_ROOT);
}



qu_any qu_gc__new_static (qu_class c, size_t size)
{
	qu_obj o;

	ASSERT_THREAD_NOT_BLOCKING

	assert (size && size <= QU_GC_SIZE_LIMIT);

	if (StaticRem > size)
	{
		if (qu_int_is (StaticPtr))
		{
			++StaticPtr;
			--StaticRem;
		}
		o = StaticPtr;
		StaticPtr += size;
		StaticRem -= size;
		goto okay;
	}

	o = malloc (size);
	if (o == NULL)
		o = qu_mem_alloc (size);

  okay:
	o->klass = c;
	o->flags = QU_FLAG_STATIC;
	o->attrs = NULL;
	o->weak  = NULL;

#ifdef COLLECT_STATISTICS
	NumStatics++;
#endif
	return o;
}



int qu_gc_is_static (qu_any o)
{
	return qu_int_is (o) || qu_flag_test (o, QU_FLAG_STATIC);
}



/*----------------------------------------------------------------------------
	garbage collection
----------------------------------------------------------------------------*/

/*
	Marking macros.
	We know that o is not an qu_int.
*/
#define MARKED(o) qu_flag_test  (o, QU_FLAG_MARK)
#define MARK(o)   qu_flag_set   (o, QU_FLAG_MARK)
#define UNMARK(o) qu_flag_unset (o, QU_FLAG_MARK)


/*
	All "mark" methods must call qu_gc_mark to mark referenced objects.
	They dont need to mark the object itself, the klass and attrs
	fields because we do it here for efficiency.

	We still have to mark the class because if the object happens to
	come from an eval module, there might be situations when the module
	and class is not referenced (only the object is).

	Observe:

		x = eval (
			`
				class Foo
				;;
				Foo
			`
			, nil, "et")

		x = x()
*/
static int _gc_is_live_object (qu_any o);

#define QU_MARK_UNLESS_STATIC(o) \
QU_BEGIN \
	if (!qu_static_is (o)) \
		_gc_mark (o); \
QU_END

static void _gc_mark (qu_obj o)
{
	qu_class c = o->klass;

	assert (c);

	if (!MARKED (c))
	{
		MARK (c);
		qu_attr_mark (c->attrs);
		QuClassClass->mark ((qu_obj) c);
	}

	MARK (o);

	if (o->weak)
	{
		assert (qu_set_is (o->weak));
		qu_gc_mark (o->weak);
	}

	if (o->attrs)
		qu_attr_mark (o->attrs);

	if (c->mark)
		c->mark (o);
}



/*
	Mark object.
*/
void qu_gc_mark (qu_any o)
{
	if (!qu_int_is (o) && !MARKED (o))
	{
		_gc_mark (o);
	}
}



/*
	If o is possibly 0, use this instead of qu_gc_mark.
*/
void qu_gc_xmark (qu_any o)
{
	if (o && !qu_int_is (o) && !MARKED (o))
	{
		_gc_mark (o);
	}
}



/*
	The (unreliable) current limit when GC should start.
*/
long __QuGcLimit ()
{
	return (long) GcReclaimLimit;
}



/*
	This function is designed to be called when marking.
	Thus there is no need to detect root and static objects.
*/
static int _gc_is_live_object (qu_any o)
{
	register page_s *g = PTR_TO_PAGE (o);

	return
		/* this also invalidates zeroes and hopefully most qu_int objects */
		_gc_page_in_arena_extremes (g)

		/* must not touch memory until we know it's ours */
		&& _gc_page_in_arena (g)

		/* a quick check */
		&& g->size != SIZE_NONE

		/* yet another quick check */
		&& g->type == TYPE_REGULAR

		/* it must be chunk aligned */
		&& _gc_addr_is_chunk (g, o)

		/* must not be freed */
		&& ((qu_obj) o)->klass;
}


/* must only be called by qu__thread__mark */
void qu_gc_mark_cstack (qu_thread self, qu_any *lo, qu_any *hi)
{
#ifdef QU_DEBUG_LOCAL
	GcState = GC_STATE_MARK_STACK;
	fprintf (stderr, "................ MARKING STACK %ld ..................\n", self->tid);
#endif

	assert (lo && hi);

	if (lo > hi)
	{
		qu_any *tmp = hi;
		hi = lo;
		lo = tmp;
	}

#ifdef GC_DEBUG_HEAVY
	fprintf (stderr, "%ld  %p .. %p = %d\n", self->tid, lo, hi, (hi - lo) / sizeof (void *));
#endif

	while (lo < hi)
	{
		qu_any o = *lo++;

		if (_gc_is_live_object (o) && !MARKED (o))
		{
#ifdef GC_DEBUG_HEAVY
			fprintf (stderr, "%p (%p)\n", o, qu_classof (o));
			fprintf (stderr, "    (%s)\n", qu_classof (o)->name->str);
#endif
			_gc_mark (o);
		}
	}
}



/*
	The collector - plain simple mark & sweep algorithm
*/
#if defined(sparc) || defined(__sparc__)
inline static void flush_register_windows ()
{
    asm
#ifdef __GNUC__
	volatile
#endif
#if defined(__sparc_v9__) || defined(__sparcv9) || defined(__arch64__)
	("flushw");
#elif defined(linux) || defined(__linux__)
	("ta 0x83");
#else
	("ta 0x03");
#endif
}
#else
#define flush_register_windows() QU_BEGIN QU_END
#endif


void qu_gc_reclaim ()
{
	int       i;
	qu_thread t;

#ifdef QU_DEBUG_SEGV
	long count  = 0;
	long memory = 0; /* just roughly */
#endif

	if (__QuGcDisable)
		return;

#ifdef QU_DEBUG_SEGV
	printf ("\nGC START " QU_THREAD_IDFORMAT " %ld %d --------------------------------\n", qu_thread_id (), __QuGcMemAllocated, ArenaCount);
#endif

#ifdef GC_PRINT_CHAR
	fprintf (stderr, GC_PRINT_CHAR_RECLAIM_BEGIN);
	fflush (stderr);
#endif

#ifdef QU_DEBUG_LOCAL
	fprintf (stderr, "\n");
	fprintf (stderr, "====================== RECLAIM ======================\n");
	fprintf (stderr, "Arenas         : %d\n", ArenaCount);
	fprintf (stderr, "Arena extremes : %p .. %p\n", ArenaPageLo, ArenaPageHi);
	fprintf (stderr, "GcMemAllocated : %ld\n", __QuGcMemAllocated);
	fprintf (stderr, "GcReclaimLimit : %ld\n", GcReclaimLimit);
	fprintf (stderr, "GcNeedReclaim  : %d\n" , __QuGcNeedReclaim);
#endif

	t = qu_thread_self ();
	t->stacklo = (void *) &i;

#ifdef QU_DEBUG_LOCAL
	GcState = GC_STATE_MARK_ROOT;
	fprintf (stderr, "................. MARKING ROOTS .....................\n");
#endif

	for (i = 0; i < NUMSIZES; i++)
	{
		pool_s *p = &PoolRoot[i];
		int     z = p->size;
		int     k = p->total;
		page_s *g;

		for (g = p->next; g != (page_s *) p; g = g->next)
		{
			qu_obj o = (qu_obj) ((void *) g + sizeof (page_s));
			int n = k;

			while (n--)
			{
				if (o->klass)
				{
					_gc_mark (o);
				}
				o = (qu_obj) ((void *) o + z);
			}
		}
	}

#ifdef QU_DEBUG_LOCAL
	GcState = GC_STATE_MARK_REGULAR;
	fprintf (stderr, "................ MARKING REGULARS ...................\n");
#endif

	for (i = 0; i < NUMSIZES; i++)
	{
		pool_s *p = &PoolReg[i];
		int     z = p->size;
		int     k = p->total;
		page_s *g;

		for (g = p->next; g != (page_s *) p; g = g->next)
		{
			qu_obj o = (qu_obj) ((void *) g + sizeof (page_s));
			int    n = k;

			while (n--)
			{
				if (o->klass && !MARKED (o))
				{
					if (qu_flag_test (o, QU_FLAG_DONTGC))
					{
						qu_gc_mark (o);
					}
					else if (!qu_flag_test (o, QU_FLAG_FIN))
					{
						if (qu_class_has_destroy (o->klass))
						{
							assert (o->klass != QuClassClass);
							qu_flag_set (o, QU_FLAG_FIN);
							qu_gc_mark (o);
							qu_interp_add_finalizer (o);
						}
					}
				}
				o = (qu_obj) ((void *) o + z);
			}
		}
	}

#ifdef QU_DEBUG_LOCAL
	GcState = GC_STATE_MARK_REGISTERS;
	fprintf (stderr, "................ MARKING REGISTERS ..................\n");
#endif

	flush_register_windows ();
	{
		jmp_buf  regs;
		qu_any  *p;
		int      n = sizeof (regs) / sizeof (void *);

		memset (&regs, 0, sizeof (regs));
		setjmp (regs);
		p = (qu_any *) &regs;
		while (n--)
		{
			qu_any o = *p++;

			if (_gc_is_live_object (o) && !MARKED (o))
			{
#ifdef GC_DEBUG_HEAVY
				fprintf (stderr, "%p     \n", o);
				fprintf (stderr, "   (%p)\n", qu_classof (o));
				fprintf (stderr, "        (%s)\n", qu_classof (o)->name->str);
#endif
				_gc_mark (o);
			}
		}
	}

#ifdef QU_DEBUG_LOCAL
	GcState = GC_STATE_SWEEP;
	fprintf (stderr, ".................... SWEEP ..........................\n");
#endif

	for (i = 0; i < NUMSIZES; i++)
	{
		pool_s *p = &PoolReg[i];
		int     z = p->size;
		int     k = p->total;
		page_s *g, *x;

		/*
			During sweeping, a page might be paged-out to the FreePage list
			so we need 'x' to point to the next page before we call Gc_Free.

			Observe that it is impossible that the next page 'x' is freed
			after Gc_Free. So 'x' is always valid.
		*/
		for (g = p->next; g != (page_s *) p; g = x)
		{
			qu_obj o = (qu_obj) ((void *) g + sizeof (page_s));
			int    n = k;

			x = g->next;

#ifdef QU_DEBUG_LOCAL
			GcSweepTrackPage = x;
#endif
			while (n--)
			{
				if (o->klass)
				{
					if (MARKED (o))
					{
						UNMARK (o);
					}
					else
					{
#ifdef QU_DEBUG_SEGV
						count++;
						memory += o->klass->size;
#endif
						if (o->weak)
							qu_gc_weak_destroy (o);

						qu_attr_xfree (o->attrs);
						if (o->klass->dtor)
							o->klass->dtor (o);
						o->klass = 0;
						g->count--;
						__QuGcMemAllocated -= z;
					}
				}
				o = (qu_obj) ((void *) o + z);
			}

			if (g->count == 0)
			{
				/* all objects in the page are freed */

				arena_s *a = g->arena;

				g->prev->next = g->next;
				g->next->prev = g->prev;
				g->next = FreePage;
				FreePage = g;

				if (PTR_TO_PAGE (p->free) == g)
					p->free = NULL;

				if (--(a->count) == 0 && ArenaCount > ARENA_LIMIT)
					_gc_free_arena (a);
			}
		}
	}

#ifdef QU_DEBUG_LOCAL
	GcState = GC_STATE_IDLE;
#endif

	GcReclaimLimit = __QuGcMemAllocated + RECLAIM_THRESHOLD;
	if (GcReclaimLimit > RECLAIM_MAX)
		GcReclaimLimit = RECLAIM_MAX;
	else if (GcReclaimLimit < RECLAIM_THRESHOLD)
		GcReclaimLimit = RECLAIM_THRESHOLD;

	__QuGcNeedReclaim = 0;

#ifdef QU_DEBUG_LOCAL
	fprintf (stderr, "Arenas         : %d\n", ArenaCount);
	fprintf (stderr, "Arena extremes : %p .. %p\n", ArenaPageLo, ArenaPageHi);
	fprintf (stderr, "GcMemAllocated : %ld\n", __QuGcMemAllocated);
	fprintf (stderr, "GcReclaimLimit : %ld\n", GcReclaimLimit);
	fprintf (stderr, "GcNeedReclaim  : %d\n" , __QuGcNeedReclaim);
	fprintf (stderr, "=====================================================\n");
#endif

#ifdef GC_PRINT_CHAR
	fprintf (stderr, GC_PRINT_CHAR_RECLAIM_END);
	fflush (stderr);
#endif

#ifdef QU_DEBUG_SEGV
	printf ("GC END - - - - - - - - - - - - - - - - - - %ld / %ld\n", count, memory);
#endif
}


/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_gc__init ()
{
	_gc_setup ();

	StaticMem = StaticPtr = malloc (STATIC_MEMORY_SIZE);
	if (StaticMem == NULL)
		QU_FATAL ("cant allocate static memory");

#ifdef QU_USE_CRITICAL_MEMORY
	CriticalMem = CriticalPtr = malloc (CRITICAL_MEMORY_SIZE);
	if (CriticalMem == NULL)
		QU_FATAL ("cant allocate critical memory");
#endif

#ifdef FREE_AT_EXIT
	atexit (_gc_free_all);
#endif

}
