/* interp.c - Copyright (c) Marc Krisnanto */

#undef QU_DEBUG_LOCAL
#include "qu.h"


/* the main module */
static volatile qu_module MainModule = 0;

/* must process signal */
volatile int __QuInterpGotSignal = 0;

/* the current running frame */
volatile qu_frame __QuInterpFrame = 0;

/* whether we are exiting the main program */
volatile int __QuInterpExiting;


/*----------------------------------------------------------------------------
	thread stuffs
----------------------------------------------------------------------------*/

/*
	Locking and unlocking the interpreter.
*/
#ifdef DEBUG /* see qu_thread.c -- it sets the qu_thread BLOCKING flag */
#define INTERP_DEBUG_LOCK
#else
#undef INTERP_DEBUG_LOCK
#endif

#ifdef INTERP_DEBUG_LOCK

#define INTERP_LOCK(t) \
QU_BEGIN \
	if (!qu_thread_blocking (t)) \
		QU_FATAL ("thread is not blocking while acquiring GcLock"); \
	qu_gc_lock (); \
	qu_thread_unset_blocking (t); \
	__QuThreadCurrent = t; \
QU_END

#define INTERP_UNLOCK \
QU_BEGIN \
	if (!__QuThreadCurrent) \
		QU_FATAL ("ThreadCurrent == NULL while releasing GcLock"); \
	if (qu_thread_blocking (__QuThreadCurrent)) \
		QU_FATAL ("thread is blocking while releasing GcLock"); \
	qu_thread_set_blocking (__QuThreadCurrent); \
	__QuThreadCurrent = 0; \
	qu_gc_unlock (); \
	qu_thread_yield (); \
QU_END


#else /*INTERP_DEBUG_LOCK*/


#define INTERP_LOCK(t) \
QU_BEGIN \
	qu_gc_lock (); \
	qu_thread_unset_blocking (t); \
	__QuThreadCurrent = t; \
QU_END

#define INTERP_UNLOCK \
QU_BEGIN \
	qu_thread_set_blocking (__QuThreadCurrent); \
	qu_gc_unlock (); \
	qu_thread_yield (); \
QU_END


#endif /*INTERP_DEBUG_LOCK*/


void qu_interp_lock ()
{
	qu_thread t = qu_thread_self ();

	assert (t->atomic >= 0);
	if (t->atomic == 0)
	{
		INTERP_LOCK (t);
	}
}

void qu_interp_unlock ()
{
	qu_thread t = qu_thread_self ();

	assert (__QuThreadCurrent == t);
	assert (t->atomic >= 0);
	if (t->atomic == 0)
	{
		t->stacklo = (qu_any) &t;
		INTERP_UNLOCK;
	}
}



/*----------------------------------------------------------------------------
	qu_interp_run_frame - run frame
----------------------------------------------------------------------------*/

#undef pCodeDEBUG
#include "Inlines.h"

#ifdef __GNUC__
#define USE_LABEL
#endif


/* The maximum recursive depth into the interpreter loop */
#define MAX_RECURSIVE 8192


/* size of instructions */
#define L0 (0)
#define L1 (0 + sizeof (qu_any))
#define L2 (0 + sizeof (qu_any) + sizeof (qu_any))

/* goto the next instruction */
#define NEXT2 IP += L2
#define NEXT1 IP += L1
#define NEXT0 IP += L0

/* get instruction argument */
#define ARGJMP   (*(long *) (CS + IP + L0))
#define ARGLONG  (*(long *) (CS + IP + L0))
#define ARGANY   (*(qu_any  *) (CS + IP + L0))
#define ARGLONG2 (*(long *) (CS + IP + L1))
#define ARGANY2  (*(qu_any  *) (CS + IP + L1))

/* rethrow the last exception */
#define RERAISE      QU_BEGIN qu_frame_rethrow (FRM); goto leave; QU_END

/* throw exception */
#define RAISE(kls)   QU_BEGIN qu_frame_throw (FRM, kls##Class); goto leave; QU_END

/* stack base pointer */
#define BP      (FRM->bp)

/* current depth */
#define LEVEL   (FRM->level)

/* the 'self' object */
#define SELF    (FRM->self)

/* current line */
#define LINE    (FRM->line)

/* atomic state */
#define ATOMIC  (FRM->atomic)

/*
	going to the outside world might trigger GC so we need to update SP
	we also need to tell the world the current running frame
*/
#define __out   { FRM->sp = SP; __QuInterpFrame = FRM;

/* back from the outside world */
#define __back  }

/*
	check for signals -- only the main thread handles signals

	note: the use of __qu_atomic_begin/end might cause a deadlock if the
	signal handler goes into __qu_blocking_begin state but I consider
	that as a programmer error
*/
#define CHECKSIGNALS \
QU_BEGIN \
	if (__QuInterpGotSignal && __QuThreadCurrent == __QuThreadMain) \
	{ \
		__QuInterpGotSignal = 0; \
		if (!__QuInterpExiting) \
		{ \
			if (__QuSigTERM) \
				RAISE (QuEQuit); \
			if (__QuSigINT) \
			{ \
				qu_any u = QuSys__keyboard__->value; \
				__QuSigINT = 0; \
				if (u != QuNil) \
				{ \
					assert (qu_sub_is (u)); \
					assert (qu_sub_is_static (u)); \
					__qu_atomic_begin; \
				  __out \
					u = qu_sub_call0 (u, QuNil); \
				  __back \
					__qu_atomic_end; \
					if (u == 0) \
						RAISE (QuEInterrupt); \
				} \
				else \
				{ \
					RAISE (QuEInterrupt); \
				} \
			} \
		} \
	} \
QU_END


#define OPTIMIZE_TAIL_RECURSION

#ifdef OPTIMIZE_TAIL_RECURSION
inline static qu_any _interp_frame_tail (qu_frame f, qu_sub u, long argc, qu_any *argv)
{
	assert (qu_sub_is (u));
	assert (! (qu_sub_is_extern (u) || qu_sub_is_generator (u) || qu_sub_is_vararg (u)));

	QU_ENSURE (qu_sub_arg_check (u, argc, argv, 1));

	QU_OBJECT_LIST_SETVAL (f->ss, QuNil, u->nlocal);
	QU_OBJECT_LIST_COPY   (f->ss + u->first, argv, argc);

	f->argc  = argc;
	f->line  = 0;
	f->ip    = 0;
	f->level = f->atomic = -1;
	f->sp    = f->bp;
	f->sp[0] = QuNil;
	return f;
}
#endif



qu_any qu_interp_run_frame (qu_frame FRM)
{

#undef SHOWME

#ifdef USE_LABEL
	static const void *LABELS[] =
	{
#define p_(name, whatever) &&_##name##_,
#include "Codes.h"
	};

#ifdef SHOWME
#define CASE(inst) _##inst##_: \
	fprintf (stderr, "%d %s\n", IP, qu_code_pcname(CS[IP-1])->str);
#else
#define CASE(inst) _##inst##_:
#endif
#define PROCEED    goto *LABELS[CS[IP++]]


#else /*USE_LABEL*/

#ifdef SHOWME
#define CASE(inst) case PC_##inst: \
	fprintf (stderr, "%d %s\n", IP, qu_code_pcname(CS[IP])->str);
#else
#define CASE(inst) case PC_##inst:
#endif
#define PROCEED    goto jmploop

#endif /*USE_LABEL*/


	register unsigned char *CS = FRM->code->bytes;
	register int            IP = FRM->ip;
	register qu_any        *SP = FRM->sp;
	register qu_any        *LS = FRM->ss;
	register qu_frblk_s    *BLOCKS = FRM->blocks;

#ifdef QU_DEBUG_LOCAL

	int     max_ip = FRM->code->len - 1;
	qu_any *max_sp = FRM->ss + FRM->code->stacksize - 1;

	if (__QuCompileIsCompiling)
		QU_FATAL ("bug: calling script while compiling");

	if (qu_frame_running (FRM))
		QU_FATAL ("bug: frame already running");
#endif

	assert (__QuThreadCurrent);

	if (__QuThreadCurrent->depth > MAX_RECURSIVE)
	{
		qu_thread_throwex (QuERecursionClass, 0, FRM->code->parent, FRM->line);
		qu_frame_set_thrown (FRM);
		qu_frame_set_done (FRM);
		return 0;
	}

#if 0
	if (qu_frame_thrown (FRM))
	{
		qu_any x = FRM->eclass;
		if (x == 0)
			x = QuERuntimeClass;
		qu_thread_throwex (x, 0, FRM->code->parent, FRM->line);
		return 0;
	}
#endif

	qu_frame_set_running (FRM);


#if defined(OPTIMIZE_TAIL_RECURSION) || !defined(USE_LABEL)
  jmploop:
#endif

#ifdef QU_DEBUG_LOCAL
	{
		qu_string s = qu_interp_where_string ();

		if (s == 0)
			s = qu_string_byte ('?');

		if (IP > max_ip)
			QU_FATAL ("%s: IP overflow %d", s->str, IP);

		if (IP < 0)
			QU_FATAL ("%s: IP underflow %d", s->str, IP);

		if (SP > max_sp)
		{
			qu_string b = qu_code_pcname (CS[IP]);
			QU_FATAL ("%s %d: SP overflow %d line:%ld %s ", s->str, IP, SP - BP, LINE, b->str);
		}
		if (SP < BP)
		{
			qu_string b = qu_code_pcname (CS[IP]);
			QU_FATAL ("%s %d: SP underflow %d line:%ld %s", s->str, IP, SP - BP, LINE, b->str);
		}
	}
#endif



#ifdef USE_LABEL
	PROCEED;
#else
	switch (CS[IP++])
	{
#endif

	/*----------------------------------------------------------------------*/
	CASE (ARRAY)
	{
		register long n = ARGLONG - 1;

	  __out
		SP[-n] = qu_array_new_argv (n + 1, SP - n);
	  __back
		SP -= n;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (ARG)
	{
		register long n = ARGLONG - 1;
		qu_dict d;

	  __out
		d = qu_arg_new (n + 1, SP - n);
	  __back
		SP -= n;
		SP[0] = d;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (ARGC)
	{
		++SP;
		SP[0] = qu_int_new (FRM->argc);
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (AS)
	{
		register qu_any a;

	  __out
		a = qu_isys_as (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (ASLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_as (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (ASLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_as (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (AUTOATTR)
	{
		register qu_any a    = SP[0];
		register qu_any c    = qu_classof (a);
		register qu_any sElf = SP[-1];

		if (c == QuSubClass)
		{
			if (!qu_sub_noauto (a))
			{
				if (qu_sub_is_class (a))
				{
					if (!qu_class_is (sElf))
						sElf = qu_classof (sElf);
				}
				else if (qu_sub_is_inst (a))
				{
					if (qu_class_is (sElf))
						RAISE (QuECall);
				}

			  __out
				a = qu_sub_call0 (a, sElf);
			  __back

				if (a == 0)
					RERAISE;
			}
		}
		else if (c == QuMethodClass)
		{
		  __out
			a = qu_method_calln_attr (a, sElf, 0, NULL);
		  __back

			if (a == 0)
				RERAISE;
		}
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (AUTOATTREX)
	{
		register qu_any a    = SP[0];
		register qu_any c    = qu_classof (a);
		register qu_any sElf = SP[-1];

		if (c == QuSubClass)
		{
			if (!qu_sub_noauto (a))
			{
				if (qu_sub_is_class (a))
				{
					if (!qu_class_is (sElf))
						sElf = qu_classof (sElf);
				}
				else if (qu_sub_is_inst (a))
				{
					if (qu_class_is (sElf))
						RAISE (QuECall);
				}

			  __out
				a = qu_sub_call0 (a, sElf);
			  __back

				if (a == 0)
					RERAISE;

				SP[0] = a;
			}
		}
		else if (c == QuMethodClass)
		{
		  __out
			a = qu_method_calln_attr (a, sElf, 0, NULL);
		  __back

			if (a == 0)
				RERAISE;

			SP[0] = a;
		}
		NEXT0;
		PROCEED;
	}


	CASE (AUTOATTREX2)
	{
		register qu_any   a = SP[0];
		register qu_class c = qu_classof (a);

		if (c == QuSubClass)
		{
			if (!qu_sub_noauto (a))
			{
				register qu_any sElf = SP[-2];

				if (qu_sub_is_class (a))
				{
					if (!qu_class_is (sElf))
						sElf = qu_classof (sElf);
				}
				else if (qu_sub_is_inst (a))
				{
					if (qu_class_is (sElf))
						RAISE (QuECall);
				}

			  __out
				a = qu_sub_call0 (a, sElf);
			  __back

				if (a == 0)
					RERAISE;

				SP[0] = a;
			}
		}
		else if (c == QuMethodClass)
		{
			qu_sub u = ((qu_method) a)->sub;

		  __out
			if (qu_sub_is_inst (u))
				a = qu_method_call0 (a);
			else
				a = qu_sub_call0 (u, SP[-2]);
		  __back

			if (a == 0)
				RERAISE;

			SP[0] = a;
		}
		NEXT0;
		PROCEED;
	}


	/*----------------------------------------------------------------------*/
	CASE (BINOP)
	{
		register qu_any a = qu_sys_binop (SP[-1], SP[0], ARGLONG);
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (BREAK)
	{
		LEVEL -= (ARGLONG - 1);
		if (LEVEL < ATOMIC)
		{
			__qu_atomic_end;
			ATOMIC = -1;
		}
		goto leave;
	}

	/*----------------------------------------------------------------------*/
	CASE (CALL)
	{
		register long     n = ARGLONG;
		register qu_any   a = SP[-n];
		register qu_class c = qu_classof (a);

	  callagain:

		if (c == QuSubClass)
		{
			if (qu_sub_is_class (a))
			{
				/*
					since we know we are calling a class method then we know
					self must be the Class object so let's allow it.
				*/
			  __out
				a = qu_sub_calln (a, qu_sub_get_class (a), n, SP - n + 1);
			  __back

				if (a == 0)
					RERAISE;

				SP -= n;
				SP[0] = a;
				NEXT1;
				PROCEED;
			}
			else
			{
#ifdef OPTIMIZE_TAIL_RECURSION
				qu_binop_f f = 0;

				if (!qu_sub_is_static (a))
					RAISE (QuECall);

				if ((CS[IP + L1] == QU_PC_RET
						|| ((f = QU_CODE_BINOPS [CS[IP + L1]]) && CS[IP + L1 + 1] == QU_PC_RET))
					&& LEVEL < 0
					&& !(qu_sub_is_extern (a) || qu_sub_is_generator (a) || qu_sub_is_vararg (a))
					&& ((qu_sub) a)->value == (void *) FRM->code
					&& ((qu_sub) a)->root == (qu_sub) a)
				{
					qu_any r = 0;

					if (f)
						r = SP[-n-1];

				__out
					a = _interp_frame_tail (FRM, a, n, SP - n + 1);
				__back

					if (a == 0)
						RERAISE;

					if (f)
					{
						/* binop tail recursion - the idea is to use the C stack
							to store the result thus deep recursion may overflow
							the C stack and we crash but we'll take that risk */

						/* careful! FRM->sp is set by qu_interp_run_frame */
						a = qu_interp_run_frame (FRM);
						if (a)
							a = f (r, a);
						SP = BP;
						if (a == 0)
							qu_frame_rethrow (FRM);
						else
							SP[0] = a;
						FRM->result = a;
						goto end;
					}
					else
					{
						SP = FRM->sp;
						IP = FRM->ip;
						goto jmploop;
					}
				}
#endif
			  __out
				a = qu_sub_calln (a, SELF, n, SP - n + 1);
			  __back
			}

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuMethodClass)
		{
		  __out
			a = qu_method_calln (a, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuStructClass)
		{
		  __out
			a = qu_record_new (a, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuClassClass)
		{
			assert (((qu_class) a)->fnew);
		  __out
			a = qu_sub_calln (((qu_class) a)->fnew, a, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuModuleClass)
		{
		  __out
			a = qu_module_getattr_callable_name (a, 1);
		  __back

			if (a == 0)
				RERAISE;

			c = qu_classof (a);
			assert (qu_class_callable (c));
			goto callagain;
		}
		else
		{
			qu_any f;

		  __out
			f = c->get (a, QU_NAME__call, 0);
		  __back

			if (f == 0)
				RAISE (QuECall);

			assert (qu_sub_is (f) && qu_sub_is_inst (f));

		  __out
			a = qu_sub_calln (f, a, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
	}

	/*----------------------------------------------------------------------*/
	CASE (CALLFLAT)
	{
		register long     n = ARGLONG;
		register qu_any   a = SP[-n];
		register qu_class c = qu_classof (a);
		volatile qu_array arg;

	  callflatagain:

		if (c == QuSubClass)
		{
			if (!qu_sub_is_static (a))
				RAISE (QuECall);

		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_sub_calln (a, SELF, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuMethodClass)
		{
		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_method_calln (a, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuStructClass)
		{
		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_record_new (a, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuClassClass)
		{
			assert (((qu_class) a)->fnew);
		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_sub_calln (((qu_class) a)->fnew, a, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuModuleClass)
		{
		  __out
			a = qu_module_getattr_callable_name (a, 1);
		  __back

			if (a == 0)
				RERAISE;

			c = qu_classof (a);
			assert (qu_class_callable (c));
			goto callflatagain;
		}
		else
		{
			register qu_any b;

		  __out
			b = c->get (a, QU_NAME__call, 0);
		  __back

			if (b == 0)
				RAISE (QuECall);

			assert (qu_sub_is (b) && qu_sub_is_inst (b));

		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_sub_calln (b, a, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
	}

	/*----------------------------------------------------------------------*/
	CASE (CALLATTR)
	{
		register long     n = ARGLONG;
		register qu_any   a = SP[-n];
		register qu_any   sElf = SP [-n - 1];
		register qu_class c = qu_classof (a);

	  callattragain:

		if (c == QuSubClass)
		{
			if (qu_sub_is_class (a))
			{
				if (!qu_class_is (sElf))
					sElf = qu_classof (sElf);
			}
			else if (qu_sub_is_inst (a))
			{
				if (qu_class_is (sElf))
					RAISE (QuECall);
			}
		  __out
			a = qu_sub_calln (a, sElf, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuMethodClass)
		{
		  __out
			a = qu_method_calln_attr (a, sElf, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuClassClass)
		{
			assert (((qu_class) a)->fnew);
		  __out
			a = qu_sub_calln (((qu_class) a)->fnew, a, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuStructClass)
		{
		  __out
			a = qu_record_new (a, n, SP - n + 1);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuModuleClass)
		{
			/* this is a consequence of "universe" */
		  __out
			a = qu_module_getattr_callable_name (a, 1);
		  __back

			if (a == 0)
				RERAISE;

			c = qu_classof (a);
			assert (qu_class_callable (c));
			goto callattragain;
		}
		else
		{
			/* allow calling property with no argument */
			if (n)
				RAISE (QuECall);

			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
	}

	/*----------------------------------------------------------------------*/
	CASE (CALLATTRFLAT)
	{
		register long     n = ARGLONG;
		register qu_any   a = SP[-n];
		register qu_any   sElf = SP [-n - 1];
		register qu_class c = qu_classof (a);
		volatile qu_array arg;

	  callattrflatagain:

		if (c == QuSubClass)
		{
			if (qu_sub_is_class (a))
			{
				if (!qu_class_is (sElf))
					sElf = qu_classof (sElf);
			}
			else if (qu_sub_is_inst (a))
			{
				if (qu_class_is (sElf))
					RAISE (QuECall);
			}

		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_sub_calln (a, sElf, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuMethodClass)
		{
		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_method_calln_attr (a, sElf, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuClassClass)
		{
			assert (((qu_class) a)->fnew);
		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_sub_calln (((qu_class) a)->fnew, a, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuStructClass)
		{
		  __out
			arg = qu_array_flatten (n, SP - n + 1);
			a = qu_record_new (a, arg->len, arg->objs);
		  __back

			if (a == 0)
				RERAISE;

			SP -= n;
			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
		else if (c == QuModuleClass)
		{
			/* this is a consequence of "universe" */
		  __out
			a = qu_module_getattr_callable_name (a, 1);
		  __back

			if (a == 0)
				RERAISE;

			c = qu_classof (a);
			assert (qu_class_callable (c));
			goto callattrflatagain;
		}
		else
		{
			/* allow calling property with no argument */
			if (n)
				RAISE (QuECall);

			--SP;
			SP[0] = a;
			NEXT1;
			PROCEED;
		}
	}

	/*----------------------------------------------------------------------*/
	CASE (CATCH)
	{
		register qu_any   e = ARGANY2;
		register qu_class c = FRM->eclass;

		assert (c != NULL);
		assert (qu_class_is (c));

		if (e == QuNil || (qu_class) e == c)
		{
		  catch:

			qu_frame_unset_thrown (FRM);
			FRM->eclass = 0;
			NEXT2;
			PROCEED;
		}

		if (qu_array_is (e))
		{
			register long    n = ((qu_array) e)->len;
			register qu_any *p = ((qu_array) e)->objs;

			while (n--)
			{
				register qu_any x = *p++;

				assert (qu_class_is (x));
				if (qu_class_isa (c, x))
					goto catch;
			}
		}
		else
		{
			assert (qu_class_is (e));
			if (qu_class_isa (c, e))
				goto catch;
		}
		IP = ARGJMP;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (CATCHSET)
	{
		register qu_any v = ARGANY;
		register qu_any x = __qu_err__;

		assert (x);

		if (qu_int_is (v))
			LS[qu_int_tol (v)] = x;
		else
		{
			((qu_var) ARGANY)->value = x;
		}
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (CLOSURE)
	{
		++SP;
		SP[0] = qu_sub_new_closure (ARGANY, FRM);
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (CMP)
	{
	  __out
		SP[-1] = qu_isys_cmp (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (CMPLL)
	{
	  __out
		SP[1] = qu_isys_cmp (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (CMPLX)
	{
	  __out
		SP[1] = qu_isys_cmp (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (CONCAT)
	{
		register qu_any a;

	  __out
		a = qu_isys_concat (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (CONCATLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_concat (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (CONCATLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_concat (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (CONTINUE)
	{
		register long n = ARGLONG - 1;

		LEVEL -= n;
		if (LEVEL < ATOMIC)
		{
			__qu_atomic_end;
			ATOMIC = -1;
		}
		SP = BP;
		IP = BLOCKS[LEVEL].cont;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (DECR)
	{
		register qu_any a = SP[0];

		if (qu_int_is (a))
		{
			register long v = qu_int_tol (a) - 1;

			if (v >= QU_INT_MIN)
			{
				SP[0] = qu_int_new (v);
				NEXT0;
				PROCEED;
			}
		}

	  __out
		a = qu_isys_decr (SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (DECRLPOST)
	{
		register long   i = ARGLONG;
		register qu_any a = LS[i];
		register qu_any b;

		if (qu_int_is (a))
		{
			register long v = qu_int_tol (a) - 1;

			if (v >= QU_INT_MIN)
			{
				LS[i] = qu_int_new (v);
				SP++;
				SP[0] = a;
				NEXT1;
				PROCEED;
			}
		}

	  __out
		b = qu_isys_decr (a);
	  __back
		if (b == 0)
			RERAISE;
		LS[i] = b;
		++SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (DECRLPRE)
	{
		register long   i = ARGLONG;
		register qu_any a = LS[i];

		if (qu_int_is (a))
		{
			register long v = qu_int_tol (a) - 1;

			if (v >= QU_INT_MAX)
			{
				a = qu_int_new (v);
				LS[i] = a;
				++SP;
				SP[0] = a;
				NEXT1;
				PROCEED;
			}
		}

	  __out
		a = qu_isys_decr (a);
	  __back
		if (a == 0)
			RERAISE;
		LS[i] = a;
		++SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (DICT)
	{
		register long n = ARGLONG - 1;

	  __out
		SP[-n] = qu_dict_new_argv (n + 1, SP - n);
	  __back
		SP -= n;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (DISCARD)
	{
		/*
			note: this is the only place where we check for signals
			it is possible that a statement never reach this instruction
			and so the signal is never caught but I consider this as
			a programmer error

			loop
				continue if true
				...
			;;
		*/

		CHECKSIGNALS;

		qu_interp_run_finalize ();

		if (__QuThreadMulti)
		{
			if (++__QuThreadCurrent->ticks >= QU_INTERP_THREAD_TICKS)
			{
				if (__QuThreadCurrent->atomic)
				{
					--__QuThreadCurrent->ticks;
				}
				else
				{
					__QuThreadCurrent->ticks = 0;
				  __out
					__qu_blocking_begin;
					/* zzzz... */
					__qu_blocking_end;
				  __back
				}
			}
		}

		/*
			!!! this is intentionally put here so that
			!!! the top of the stack is not reclaimed
			!!! it is necessary for eval, otherwise the
			!!! result might be GC'ed before collected
			!!! by the caller
		*/
		--SP;

		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (DIV)
	{
		register qu_any a;

	  __out
		a = qu_isys_div (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (DIVLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_div (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (DIVLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_div (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (DROP1)
	{
		--SP;
		SP[0] = SP[1];
		NEXT0;
		PROCEED;
	}

	CASE (DROP1LL)
	{
		/* yeah, ARGLONG is not used -- Emit did a generalized case */
		--SP;
		SP[0] = LS[ARGLONG2];
		NEXT2;
		PROCEED;
	}

	CASE (DROP1LX)
	{
		/* yeah, ARGLONG is not used -- Emit did a generalized case */
		--SP;
		SP[0] = ARGANY2;
		NEXT2;
		PROCEED;
	}

	CASE (DROP2)
	{
		SP -= 2;
		SP[0] = SP[2];
		NEXT0;
		PROCEED;
	}

	CASE (DROP3)
	{
		SP -= 3;
		SP[0] = SP[3];
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (DUP1)
	{
		++SP;
		SP[0] = SP[-1];
		NEXT0;
		PROCEED;
	}

	CASE (DUP2)
	{
		SP += 2;
		SP[0]  = SP[-2];
		SP[-1] = SP[-3];
		NEXT0;
		PROCEED;
	}

	CASE (DUP3)
	{
		SP += 3;
		SP[0]  = SP[-3];
		SP[-1] = SP[-4];
		SP[-2] = SP[-5];
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (EACH)
	{
		register qu_any a;

	  __out
		a = qu_isys_each (SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (END)
	{

	  end:

		if (qu_frame_thrown (FRM))
		{
			FRM->result = 0;
		}
		else
		{
			if (qu_frame_is_generator (FRM))
			{
				qu_frame_throw (FRM, QuEStopClass);
				FRM->result = 0;
				goto end_done;
			}
			else
			{
				qu_any x = FRM->code->parent;
				if (qu_sub_is (x))
				{
					if (((qu_sub) x)->type)
					{
					  __out
						FRM->result = qu_object_validate (QuNil, ((qu_sub) x)->type, FRM->result);
					  __back
						if (FRM->result == 0)
						{
							qu_frame_throw (FRM, QuEValueClass);
							goto end_done;
						}
					}
				}
				else
				{
					assert (qu_module_is (x));
				}
			}
		}

	  end_done:

		qu_frame_unset_running (FRM);
		qu_frame_set_done (FRM);
		FRM->sp = SP;
		return FRM->result;
	}

	/*--------------------------------------------------------------------------*/
	CASE (ENTER)
	{
		register qu_frblk_s *blk = &BLOCKS [++LEVEL];

		blk->nextip = ARGJMP;
		blk->cont   = IP + L2;
		blk->obj    = 0; /* must do this for qu__frame_mark */

		switch (ARGLONG2)
		{
			case QU_CODE_BLOCK_ATOMIC:
			{
				__qu_atomic_begin;
				ATOMIC = LEVEL;
				NEXT2;
				PROCEED;
			}

			case QU_CODE_BLOCK_RANGE:
			{
				register qu_any c = SP[0];  /* step */
				register qu_any b = SP[-1]; /* last */
				register qu_any a = SP[-2]; /* first */

				if (!qu_int_is (a) || !qu_int_is (b) || !qu_int_is (c))
					RAISE (QuEValue);

				blk->obj = c;
				blk->ex1 = qu_int_tol (a);
				blk->ex2 = qu_int_tol (b);
				SP -= 3;
				NEXT2;
				PROCEED;
			}

			case QU_CODE_BLOCK_FOR:
			{
				register qu_any a = SP[0];
				register qu_any c = qu_classof (a);

				if (qu_class_autoiter (c))
				{
					blk->obj = a;
					blk->ex1 = blk->ex2 = 0;
					--SP;
					NEXT2;
					PROCEED;
				}

			  __out
				a = qu_isys_each (a);
			  __back

				if (a == 0)
				{
					--LEVEL;
					RERAISE;
				}

				blk->obj = a;
				blk->ex1 = blk->ex2 = 0;
				--SP;
				NEXT2;
				PROCEED;
			}

			case QU_CODE_BLOCK_SWITCH:
			{
				blk->obj = SP[0];
				--SP;
				NEXT2;
				PROCEED;
			}

			case QU_CODE_BLOCK_ELSETRY:
			{
				/*
					!!! special case
					we get here from a JMP at the end of a try block without
					a LEAVE so we have to fake that LEAVE
				*/
				blk         = &BLOCKS [--LEVEL];
				blk->nextip = ARGJMP;
				blk->cont   = IP + L2;
				NEXT2;
				PROCEED;
			}
		}
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (EQ)
	{
	  __out
		SP[-1] = qu_isys_eq (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (EQLL)
	{
	  __out
		SP[1] = qu_isys_eq (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (EQLX)
	{
	  __out
		SP[1] = qu_isys_eq (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (EXTERN)
	{
		/*
			 EXTERN  Module object
			 LITERAL "gtk_window_set_title"
			 LITERAL '+:Ws'

			... <- jumps here if Module object has been turned into a Method
		*/
		register qu_any a = ARGANY;

		if (qu_module_is (a))
		{
			qu_module m = a;
		 	qu_string name, spec;

			if (m->getcfunc == 0)
			{
				qu_thread_throwex (QuEImplementClass, 0, FRM->code->parent, FRM->line);
				RERAISE;
			}

			name = *(qu_string *) (CS + IP + 1 + L1);
			spec = *(qu_string *) (CS + IP + 1 + L1 + 1 + L1);

			a = m->getcfunc (m, name, spec);
			if (a == 0)
				RERAISE;
			/*
				!!! we are changing the code on the fly
				!!! the result is a static Cfunc
			*/
			ARGANY = a;
		}
		++SP;
		SP[0] = a;
		IP += L1 + 1 + L1 + 1 + L1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (FORMAT)
	{
		register long   n = ARGLONG;
		register qu_any a;

	  __out
		a = qu_isys_format (SP[-n], n, SP - n + 1);
	  __back
		if (a == 0)
			RERAISE;
		SP -= n;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (FLAT)
	{
		register qu_any a = SP[0];

		if (!qu_array_is (a))
			RAISE (QuEValue);

	  __out
		SP[0] = qu_var_new_flattener ();
		((qu_var) SP[0])->value = a;
	  __back
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (FRAME)
	{
		++SP;
		SP[0] = FRM;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (GE)
	{
	  __out
		SP[-1] = qu_isys_ge (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (GELL)
	{
	  __out
		SP[1] = qu_isys_ge (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (GELX)
	{
	  __out
		SP[1] = qu_isys_ge (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (GETATTR)
	{
		register qu_any a;

	  __out
		a = qu_isys_getattr (SP[-1], SP[0], FRM);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (GETATTR2)
	{
		register qu_any a;

	  __out
		a = qu_isys_getattr (SP[-1], SP[0], FRM);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (GETATTRX)
	{
		register qu_any a;

	  __out
		a = qu_isys_getattrx (SP[0], ARGANY, FRM);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (GETATTRSELF)
	{
		register qu_any a;

	  __out
		a = qu_classof (SELF)->get (SELF, ARGANY, 0);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = SELF;
		++SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (GETITEM)
	{
		register qu_any a;

	  __out
		a = qu_isys_getitem (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (GETITEM2)
	{
		register qu_any a;

	  __out
		a = qu_isys_getitem (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (GETITEMX)
	{
		register qu_any a;

	  __out
		a = qu_isys_getitem (SP[0], ARGANY);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (GETITEML)
	{
		register qu_any a;

	  __out
		a = qu_isys_getitem (LS[ARGLONG], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (GETITEMLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_getitem (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (GETGLOBAL)
	{
		++SP;
		SP[0] = ((qu_var) ARGANY)->value;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (GETLEXICAL)
	{
		++SP;
		SP[0] = LS[ARGLONG];
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (GETSLICE)
	{
		register qu_any a;

	  __out
		a = qu_isys_getslice (SP[-2], SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP -= 2;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (GETSLICE2)
	{
		register qu_any a;

	  __out
		a = qu_isys_getslice (SP[-2], SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (GETSWITCH)
	{
		++SP;
		SP[0] = BLOCKS [LEVEL].obj;
		NEXT0;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (GT)
	{
	  __out
		SP[-1] = qu_isys_gt (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (GTLL)
	{
	  __out
		SP[1] = qu_isys_gt (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (GTLX)
	{
	  __out
		SP[1] = qu_isys_gt (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (HTOPEN)
	{
		assert (__QuModuleHtml_open);
	  __out
		__QuModuleHtml_open (ARGANY, ARGLONG2);
	  __back

		NEXT2;
		PROCEED;
	}

	CASE (HTCLOSE)
	{
		assert (__QuModuleHtml_close);
	  __out
		__QuModuleHtml_close (ARGANY, ARGLONG2);
	  __back

		NEXT2;
		PROCEED;
	}

	CASE (HTENTITY)
	{
		assert (__QuModuleHtml_entity);
	  __out
		__QuModuleHtml_entity (ARGANY);
	  __back

		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (IAND)
	{
		register qu_any a;

	  __out
		a = qu_isys_iand (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (IANDLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_iand (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (IANDLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_iand (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
		RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (IDIV)
	{
		register qu_any a;

	  __out
		a = qu_isys_idiv (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (IDIVLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_idiv (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (IDIVLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_idiv (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (IOR)
	{
		register qu_any a;

	  __out
		a = qu_isys_ior (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (IORLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_ior (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (IORLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_ior (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (IN)
	{
		register qu_any a;

	  __out
		a = qu_isys_in (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (INLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_in (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (INLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_in (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (INCR)
	{
		register qu_any a = SP[0];

		if (qu_int_is (a))
		{
			register long v = qu_int_tol (a) + 1;

			if (v <= QU_INT_MAX)
			{
				SP[0] = qu_int_new (v);
				NEXT0;
				PROCEED;
			}
		}
	  __out
		a = qu_isys_incr (a);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (INCRLPOST)
	{
		register long   i = ARGLONG;
		register qu_any a = LS[i];
		register qu_any b;

		if (qu_int_is (a))
		{
			register long v = qu_int_tol (a) + 1;

			if (v <= QU_INT_MAX)
			{
				LS[i] = qu_int_new (v);
				SP++;
				SP[0] = a;
				NEXT1;
				PROCEED;
			}
		}

	  __out
		b = qu_isys_incr (a);
	  __back
		if (b == 0)
			RERAISE;
		LS[i] = b;
		++SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (INCRLPRE)
	{
		register long   i = ARGLONG;
		register qu_any a = LS[i];

		if (qu_int_is (a))
		{
			register long v = qu_int_tol (a) + 1;

			if (v <= QU_INT_MAX)
			{
				a = qu_int_new (v);
				LS[i] = a;
				++SP;
				SP[0] = a;
				NEXT1;
				PROCEED;
			}
		}
	  __out
		a = qu_isys_incr (a);
	  __back
		if (a == 0)
			RERAISE;
		LS[i] = a;
		++SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (INOT)
	{
		register qu_any a;

	  __out
		a = qu_isys_inot (SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (INITMOD)
	{
		qu_module m = ARGANY;
		qu_frame  f;
		qu_any    r;

		assert (qu_module_is (m));

		if (qu_module_initialized (m))
		{
			NEXT1;
			PROCEED;
		}

		/*
			!!! No other thread may interfere or we risk initializing the
			!!! same module more than once at the same time.
		*/

		if (QU_APP_VERBOSE)
			fprintf (stderr, "init: %s\n", m->name->str);

		assert (m->code);
	  __out
		f = qu_frame_new_module (m);
		qu_module_set_initialized (m);
		__qu_atomic_begin;
		r = qu_interp_run_frame (f);
		__qu_atomic_end;
	  __back

		/*
			CHECKME
			There is no way for script level to refer to this qu_frame.
			We dont make __QuInterpFrame public.
		*/
		qu_gc_delete ((qu_obj) f);

		/* no longer need the qu_code object, unref it so it can be GC'ed */
		m->code = 0;

		if (r == 0)
			RERAISE;

		/* automatically call __init if defined */
		r = qu_attr_find (m->attrs, QU_NAME__init);
		if (r && qu_sub_is (r))
		{
			__qu_atomic_begin;
			r = qu_sub_call0 (r, m);
			__qu_atomic_end;
			if (r == 0)
				RERAISE;
		}

		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (IS)
	{
	  __out
		SP[-1] = qu_isys_is (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (ISLL)
	{
	  __out
		SP[1] = qu_isys_is (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (ISLX)
	{
	  __out
		SP[1] = qu_isys_is (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (ITERNEXT)
	{
		register qu_frblk_s *blk = &BLOCKS [LEVEL];
		qu_any a = blk->obj;
		qu_any b = 0;
		qu_any thing;

		switch (qu_classid (a))
		{

		case QU_CLASS_ID_Int:
		{
			register long v = blk->ex1;
			register long n = qu_int_tol (a);

			if (n >= 0)
			{
				if (v > blk->ex2)
					goto leave;
			}
			else
			{
				if (v < blk->ex2)
					goto leave;
			}
			a = qu_int_new (v);
			blk->ex1 += n;
			break;
		}

		case QU_CLASS_ID_String:
		{
			if (blk->ex1 >= ((qu_string) a)->len)
				goto leave;
			a = qu_string_byte (((qu_string) a)->str[blk->ex1++]);
			break;
		}

		case QU_CLASS_ID_Buffer:
		{
			if (blk->ex1 >= ((qu_buffer) a)->len)
				goto leave;
			a = qu_buffer_new (((qu_buffer) a)->str + blk->ex1++, 1);
			break;
		}

		case QU_CLASS_ID_Array:
		{
			if (blk->ex1 >= ((qu_array) a)->len)
				goto leave;
			a = ((qu_array) a)->objs[blk->ex1++];
			break;
		}

		case QU_CLASS_ID_Dict:
		case QU_CLASS_ID_Arg:
		{
			qu_any k;
			b = qu_dict_walk ((qu_dict) a, &k, &blk->ex1, &blk->ex2);
			if (b == 0)
				goto leave;
			a = k;
			break;
		}

		case QU_CLASS_ID_Set:
		{
			a = qu_set_walk ((qu_set) a, &blk->ex1, &blk->ex2);
			if (a == 0)
				goto leave;
			break;
		}

		case QU_CLASS_ID_Range:
		{
			if (blk->ex1 >= ((qu_range) a)->count)
				goto leave;
			a = qu_int_new (((qu_range) a)->first + ((qu_range) a)->step * blk->ex1++);
			break;
		}

		case QU_CLASS_ID_Frame:
		{
			assert (qu_frame_is_generator (a));

			if (qu_frame_done (a))
				goto leave;

			if (qu_frame_running (a))
				RAISE (QuERuntime);

			assert (((qu_frame) a)->code);

	      __out
			a = qu_interp_frame_next (a);
		  __back

			if (a == 0)
			{
				if (qu_thread_err_is_estop ())
					goto leave;
				RERAISE;
			}
			break;
		}

		case QU_CLASS_ID_Record:
		{
			long n = blk->ex1;

			if (n >= ((qu_record) a)->len)
				goto leave;

			blk->ex1++;
			b = ((qu_structentry_s *)((qu_record) a)->struc->items)[n].name;
			a = ((qu_record) a)->objs[n];
			break;
		}

		case QU_CLASS_ID_Struct:
		{
			long n = blk->ex1;

			if (n >= qu_struct_len ((qu_struct) a))
				goto leave;

			blk->ex1++;
			b = ((qu_structentry_s *)((qu_struct) a)->items)[n].type;
			a = ((qu_structentry_s *)((qu_struct) a)->items)[n].name;
			break;
		}

		default:
		{
			assert (qu_iterator_is (a));
		  __out
			a = qu_iterator_next (a);
		  __back

			if (a == 0)
			{
				if (qu_thread_err_is_estop ())
					goto leave;
				RERAISE;
			}
			break;
		}

		}


		/* a = the next value, b = the next key in case of qu_dict/qu_record

			Things to set is either a:
			- qu_int -> lexical, the value is the index
			- qu_var -> global variable
			- qu_array of qu_int/qu_var objects
		*/

#define set_iter_item_(v, val) \
QU_BEGIN \
	if (qu_int_is (v)) \
	{ \
		int j = qu_int_tol (v); \
		if (qu_frame_is_sub (FRM)) \
		{ \
			qu_sub u = (qu_sub) FRM->code->parent; \
			if (u->ltype && u->ltype[j]) \
			{ \
				val = qu_object_validate (QuNil, u->ltype[j], val); \
				if (val == 0) \
					RERAISE; \
			} \
		} \
		LS [j] = val; \
	} \
	else \
	{ \
		assert (qu_var_is (v)); \
		if (((qu_var) v)->type) \
		{ \
			val = qu_var_set_value (v, val); \
			if (val == 0) \
				RERAISE; \
		} \
		else \
		{ \
			((qu_var) v)->value = val; \
		} \
	} \
QU_END

		thing = ARGANY;

		if (qu_array_is (thing))
		{
			long    len  = ((qu_array) thing)->len;
			int     eat  = qu_array_forlast (thing);
			qu_any *vars = ((qu_array) thing)->objs;
			qu_any  what = blk->obj;
			qu_any  var;

			assert (len > 0);

			if (qu_dict_is (what) || qu_arg_is (what) || qu_record_is (what) || qu_struct_is (what))
			{
				/* dang! */
				if (len == 2 && qu_record_is (what))
				{
					register qu_any x = a;
					a = b;
					b = x;
				}

				/* for a, b, c in [foo:1, bar:2] is wrong */
				if (len > 2 && !eat)
					RAISE (QuENone);

				var = *vars++;
				set_iter_item_(var, a);
				if (len == 2)
				{
					var = *vars++;
					if (eat)
						b = qu_array_new_argv (1, &b);
					set_iter_item_(var, b);
				}
			}
			else if (qu_array_is (a))
			{
				long i;

			  __out
				for (i = 0; i < len; i++)
				{
					if (eat && i == len - 1)
					{
						b = qu_array_getslice (a, i, len);
						if (b == 0)
							b = qu_array_new (0);
					}
					else
					{
						if (0 <= i && i < ((qu_array) a)->len)
							b = qu_array_GETITEM ((qu_array) a, i);
						else
							b = QuNil;
					}
					var = *vars++;
					set_iter_item_(var, b);
				}
			  __back
			}
			else
			{
				RAISE (QuENone);
			}
		}
		else if (thing != QuSys__ignore)
		{
			set_iter_item_(thing, a);
		}

		NEXT1;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (IXOR)
	{
		register qu_any a;

	  __out
		a = qu_isys_ixor (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (IXORLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_ixor (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (IXORLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_ixor (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (JMP)
	{
		IP = ARGLONG;
		PROCEED;
	}

	CASE (JMPFALSE)
	{
		if (SP[0] == QuTrue)
		{
			--SP;
			NEXT1;
			PROCEED;
		}
		else if (SP[0] == QuFalse)
		{
			--SP;
			IP = ARGJMP;
			PROCEED;
		}
		else if (qu_isys_true (SP[0]) == QuTrue)
		{
			--SP;
			NEXT1;
			PROCEED;
		}
		else
		{
			--SP;
			IP = ARGJMP;
			PROCEED;
		}
	}

	CASE (JMPTRUE)
	{
		if (SP[0] == QuTrue)
		{
			--SP;
			IP = ARGJMP;
			PROCEED;
		}
		else if (SP[0] == QuFalse)
		{
			--SP;
			NEXT1;
			PROCEED;
		}
		else if (qu_isys_true (SP[0]) == QuTrue)
		{
			--SP;
			IP = ARGJMP;
			PROCEED;
		}
		else
		{
			--SP;
			NEXT1;
			PROCEED;
		}
	}

	CASE (JMPFALSEX)
	{
		if (SP[0] == QuTrue)
		{
			NEXT1;
			PROCEED;
		}
		else if (SP[0] == QuFalse)
		{
			IP = ARGJMP;
			PROCEED;
		}
		else if (qu_isys_true (SP[0]) == QuTrue)
		{
			NEXT1;
			PROCEED;
		}
		else
		{
			IP = ARGJMP;
			PROCEED;
		}
	}

	CASE (JMPTRUEX)
	{
		if (SP[0] == QuTrue)
		{
			IP = ARGJMP;
			PROCEED;
		}
		else if (SP[0] == QuFalse)
		{
			NEXT1;
			PROCEED;
		}
		else if (qu_isys_true (SP[0]) == QuTrue)
		{
			IP = ARGJMP;
			PROCEED;
		}
		else
		{
			NEXT1;
			PROCEED;
		}
	}

	/*----------------------------------------------------------------------*/
	CASE (LE)
	{
	  __out
		SP[-1] = qu_isys_le (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (LELL)
	{
	  __out
		SP[1] = qu_isys_le (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (LELX)
	{
	  __out
		SP[1] = qu_isys_le (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (LEAVE)
	{

	  leave:

		SP = BP;
		while (LEVEL >= 0)
		{
			if (LEVEL == ATOMIC)
			{
				__qu_atomic_end;
				ATOMIC = -1;
			}

			IP = BLOCKS[LEVEL--].nextip;

			while (CS[IP] == QU_PC_ENTER)
			{
				switch (*(long *) (CS + IP + 1 + sizeof (long)))
				{
					case QU_CODE_BLOCK_FINALLY:
						PROCEED;

					case QU_CODE_BLOCK_CATCH:
						if (qu_frame_thrown (FRM))
						{
							assert (FRM->eclass);
							if (qu_class_catchable (FRM->eclass))
								PROCEED;
						}
						IP = *(long *) (CS + IP + 1);
						continue;
				}
				break;
			}

			if (! (qu_frame_thrown (FRM) || qu_frame_returning (FRM)))
				PROCEED;
		}
		goto end;
	}

	/*----------------------------------------------------------------------*/
	CASE (LIKE)
	{
	  __out
		SP[-1] = qu_isys_like (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (LIKELL)
	{
	  __out
		SP[1] = qu_isys_like (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (LIKELX)
	{
	  __out
		SP[1] = qu_isys_like (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (LINE)
	{
		LINE = ARGLONG;
		NEXT1;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (LITERAL)
	{
		++SP;
		SP[0] = ARGANY;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (LSHIFT)
	{
		register qu_any a;

	  __out
		a = qu_isys_lshift (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (LSHIFTLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_lshift (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (LSHIFTLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_lshift (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (LT)
	{
	  __out
		SP[-1] = qu_isys_lt (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (LTLL)
	{
	  __out
		SP[1] = qu_isys_lt (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (LTLX)
	{
	  __out
		SP[1] = qu_isys_lt (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (MINUS)
	{
		register qu_any a;

	  __out
		a = qu_isys_minus (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (MINUSLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_minus (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (MINUSLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_minus (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (MUL)
	{
		register qu_any a;

	  __out
		a = qu_isys_mul (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (MULLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_mul (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (MULLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_mul (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*--------------------------------------------------------------------------*/
	CASE (NE)
	{
	  __out
		SP[-1] = qu_isys_ne (SP[-1], SP[0]);
	  __back
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (NELL)
	{
	  __out
		SP[1] = qu_isys_ne (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	CASE (NELX)
	{
	  __out
		SP[1] = qu_isys_ne (LS[ARGLONG], ARGANY2);
	  __back
		++SP;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (NEG)
	{
		register qu_any a;

	  __out
		a = qu_isys_neg (SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (NOP)
	{
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (NOT)
	{
	  __out
		SP[0] = qu_isys_not (SP[0]);
	  __back
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (OF)
	{
		register qu_any a;

	  __out
		a = qu_isys_of (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (OFLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_of (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (OFLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_of (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (PARAM)
	{
		if (FRM->argc > ARGLONG2)
		{
			IP = ARGJMP;
			PROCEED;
		}
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (PASS)
	{
	  __out
		__qu_blocking_begin;
		qu_thread_yield ();
		__qu_blocking_end;
	  __back
		__QuThreadCurrent->ticks = 0;
		++SP;
		SP[0] = QuNil;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (PLUS)
	{
		register qu_any a;

	  __out
		a = qu_isys_plus (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (PLUSLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_plus (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (PLUSLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_plus (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (POP)
	{
		register qu_any a;

	  __out
		a = qu_isys_pop (SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (POS)
	{
		register qu_any a;

	  __out
		a = qu_isys_pos (SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (POW)
	{
		register qu_any a;

	  __out
		a = qu_isys_pow (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (POWLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_pow (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (POWLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_pow (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (PRINT)
	{
		if (qu_sys_print (ARGANY) == 0)
			RERAISE;
		/* !!! must not change the stack */
		NEXT1;
		PROCEED;
	}

	CASE (PRINTTOP)
	{
		if (qu_sys_print (SP[0]) == 0)
			RERAISE;
		/* !!! discards */
		--SP;
		NEXT0;
		PROCEED;
	}

	CASE (PRINTHTML)
	{
		if (qu_sys_printh (SP[0]) == 0)
			RERAISE;
		/* !!! discards */
		--SP;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (PUSH)
	{
		register qu_any a;

	  __out
		a = qu_isys_push (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (PUSHLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_push (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (PUSHLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_push (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (RANGE)
	{
		register qu_any a;

	  __out
		a = qu_range_new_ex (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (RATIO)
	{
		register qu_any a;

	  __out
		a = qu_rational_new_ratio (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (RATIOLL)
	{
		register qu_any a;

	  __out
		a = qu_rational_new_ratio (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (RATIOLX)
	{
		register qu_any a;

	  __out
		a = qu_rational_new_ratio (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
		RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (REM)
	{
		register qu_any a;

	  __out
		a = qu_isys_rem (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (REMLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_rem (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (REMLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_rem (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (RET)
	{
		FRM->result = SP[0];
		qu_frame_set_returning (FRM);
		goto leave;
	}

	/*----------------------------------------------------------------------*/
	CASE (RIN)
	{
		register qu_any a;

	  __out
		a = qu_isys_rin (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (RINLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_rin (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (RINLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_rin (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (ROT3)
	{
		register qu_any a = SP[-2];

		SP[-2] = SP[-1];
		SP[-1] = SP[0];
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (ROT4)
	{
		register qu_any a = SP[-3];

		SP[-3] = SP[-2];
		SP[-2] = SP[-1];
		SP[-1] = SP[0];
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (ROT4X)
	{
		register qu_any a = SP[0];

		SP[ 0] = SP[-1];
		SP[-1] = SP[-2];
		SP[-2] = SP[-3];
		SP[-3] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (RSHIFT)
	{
		register qu_any a;

	  __out
		a = qu_isys_rshift (SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (RSHIFTLL)
	{
		register qu_any a;

	  __out
		a = qu_isys_rshift (LS[ARGLONG], LS[ARGLONG2]);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	CASE (RSHIFTLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_rshift (LS[ARGLONG], ARGANY2);
	  __back
		if (a == 0)
			RERAISE;
		++SP;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SELF)
	{
		++SP;
		SP[0] = SELF;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SET)
	{
		register long n = ARGLONG - 1;

	  __out
		SP[-n] = qu__set_new (QuSetClass, n + 1, SP - n);
	  __back
		SP -= n;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SETATTR)
	{
		register qu_any a;

	  __out
		a = qu_isys_setattr (SP[-2], SP[-1], SP[0], FRM);
	  __back
		if (a == 0)
			RERAISE;
		SP -= 2;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (SETATTRX)
	{
		register qu_any a;

	  __out
		a = qu_isys_setattrx (SP[-1], ARGANY, SP[0], FRM);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (SETATTRSELF)
	{
		register qu_any a;

	  __out
		a = qu_classof(SELF)->set (SELF, ARGANY, SP[0], 0);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
#if 0
	CASE (SETGLOBAL)
	{
		((qu_var) ARGANY)->value = SP[0];
		NEXT1;
		PROCEED;
	}
#else
	/*
		Now that we are allowing dynamic setting of global variables types,
		then we have no use of SETGLOBALX --- but let's not throw it
		away just yet.

		We might just throw away the dynamic setting instead because checking
		for v->type is rather expensive.

		See also Lightning.c SETGLOBAL and observe the expensive change.
	*/

	CASE (SETGLOBAL)
	{
		if (((qu_var) ARGANY)->type == 0)
		{
			((qu_var) ARGANY)->value = SP[0];
			NEXT1;
			PROCEED;
		}
		/*nobreak*/
	}
#endif

	CASE (SETGLOBALX)
	{
		register qu_any a;

	  __out
		a = qu_var_set_value (ARGANY, SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SETITEM)
	{
		register qu_any a;

	  __out
		a = qu_isys_setitem (SP[-2], SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP -= 2;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (SETITEMX)
	{
		register qu_any a;

	  __out
		a = qu_isys_setitem (SP[-1], ARGANY, SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (SETITEML)
	{
		register qu_any a;

	  __out
		a = qu_isys_setitem (LS[ARGLONG], SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		--SP;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	CASE (SETITEMLX)
	{
		register qu_any a;

	  __out
		a = qu_isys_setitem (LS[ARGLONG], ARGANY2, SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP[0] = a;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SETLEXICAL)
	{
		LS[ARGLONG] = SP[0];
		NEXT1;
		PROCEED;
	}

	CASE (SETLEXICALX)
	{
		register qu_any a = SP[0];
		a = qu_object_validate (QuNil, ARGANY2, a);
		if (a == 0)
			RAISE (QuEValue);
		LS[ARGLONG] = a;
		NEXT2;
		PROCEED;
	}

	CASE (SETLEXICALLL)
	{
		++SP;
		SP[0] = LS[ARGLONG] = LS[ARGLONG2];
		NEXT2;
		PROCEED;
	}

	CASE (SETLEXICALLX)
	{
		++SP;
		SP[0] = LS[ARGLONG] = ARGANY2;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SETSLICE)
	{
		register qu_any a;

	  __out
		a = qu_isys_setslice (SP[-3], SP[-2], SP[-1], SP[0]);
	  __back
		if (a == 0)
			RERAISE;
		SP -= 3;
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SUPER)
	{
		register qu_any x = SP[0];

		if (!qu_string_is (x))
			RAISE (QuEValue);

	  __out
		if (qu_class_is (SELF))
			x = QuClassClass->get ((qu_class) ARGANY, x, 0);
		else
			x = ((qu_class) ARGANY)->get (SELF, x, 0);
	  __back

		if (x == 0)
			RERAISE;

		SP[0] = SELF;
		++SP;
		SP[0] = x;
		NEXT1;
		PROCEED;
	}

	CASE (SUPERX)
	{
		register qu_any x;

	  __out
		if (qu_class_is (SELF))
			x = QuClassClass->get ((qu_class) ARGANY, ARGANY2, 0);
		else
			x = ((qu_class) ARGANY)->get (SELF, ARGANY2, 0);
	  __back

		if (x == 0)
			RERAISE;

		++SP;
		SP[0] = SELF;
		++SP;
		SP[0] = x;
		NEXT2;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (SWAP)
	{
		register qu_any a = SP[0];

		if (!qu_frame_is (a) || !qu_frame_is_generator (a))
			RAISE (QuEValue);

		if (qu_frame_running (a))
			/* including when attempting to swap to this frame */
			RAISE (QuERuntime);

		if (qu_frame_done (a))
		{
			NEXT0;
			PROCEED;
		}

		qu_frame_unset_running (FRM);
		qu_frame_set_swapping (FRM);

		FRM->ip = IP;
		FRM->sp = SP;
		return a;
	}

	/*----------------------------------------------------------------------*/
	CASE (THROW)
	{
		register long n = ARGLONG;

		if (n == 1)
		{
			if (qu_exception_class_is (SP[0]))
			{
				__QuThreadCurrent->eclass = SP[0];
				__QuThreadCurrent->emsg   = 0;
			}
			else
			{
				qu_sys_raise (1, SP);
			}
		}
		else if (n == 2)
		{
			if (qu_exception_class_is (SP[0]))
			{
				__QuThreadCurrent->eclass = SP[0];
				__QuThreadCurrent->emsg   = SP[-1];
			}
			else
			{
				qu_sys_raise (2, SP - 1);
			}
		}
		RERAISE;
	}

	/*----------------------------------------------------------------------*/
	CASE (TRUE)
	{
	  __out
		SP[0] = qu_isys_true (SP[0]);
	  __back
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
#if 0
	CASE (UNSAFE)
	{
		qu_module m = qu_frame_get_module (FRM);
		assert (m);

		if (!qu_module_is_trusted (m))
			RAISE (QuESecurity);
		NEXT0;
		PROCEED;
	}
#endif

	/*----------------------------------------------------------------------*/
	CASE (XCAT)
	{
		register long n = ARGLONG - 1;
		register qu_any  a;

	  __out
		a = qu_sys_cat (n + 1, SP - n);
	  __back
		if (a == 0)
			RERAISE;
		SP -= n;
		SP[0] = a;
		NEXT1;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (XCHG)
	{
		register qu_any a = SP[0];

		SP[0] = SP[-1];
		SP[-1] = a;
		NEXT0;
		PROCEED;
	}

	CASE (XCHG3)
	{
		register qu_any a = SP[-3];

		SP[-3] = SP[0];
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	CASE (XCHG4)
	{
		register qu_any a = SP[-4];

		SP[-4] = SP[0];
		SP[0] = a;
		NEXT0;
		PROCEED;
	}

	/*----------------------------------------------------------------------*/
	CASE (YIELD)
 	{
		register qu_any a = SP[0];
		register qu_any b = FRM->code->parent;

		assert (qu_frame_is_generator (FRM));
		assert (qu_sub_is (b));

		b = ((qu_sub) b)->type;
		if (b)
		{
		  __out
			a = qu_object_validate (QuNil, b, a);
		  __back

			if (a == 0)
				RAISE (QuEValue);
		}
		qu_frame_unset_running (FRM);
		FRM->ip = IP + L0;
		FRM->sp = SP;
		return a;
	}

#ifdef USE_LABEL
  _end_marker_:
#else
	}
#endif

	QU_FATAL ("unknown byte code: %d", (int) CS[IP]);
	return 0; /*never*/
}



/*----------------------------------------------------------------------------
	out of memory trap
----------------------------------------------------------------------------*/

static qu_string MemTrapName = 0; /* __memory__ */
static void _interp_memtrap ()
{
	if (MemTrapName)
	{
		qu_sub u = qu_attr_find (QuModuleMain->attrs, MemTrapName);
		MemTrapName = 0;
		if (u && qu_sub_is (u) && qu_sub_is_static (u))
			qu_sub_call0 (u, QuNil);
	}
}



/*----------------------------------------------------------------------------
	finalizers
----------------------------------------------------------------------------*/

volatile qu_array __QuInterpFinalizeList;
volatile int      __QuInterpInFinalize = 0;

void qu__interp_run_finalize ()
{
	qu_class lastclass = 0;
	qu_sub   lastsub = 0;

	assert (__QuThreadCurrent);
	assert (__QuInterpInFinalize);

	while (__QuInterpFinalizeList->len)
	{
		qu_class c;
		qu_sub   u;
		qu_any   x;

		x = __QuInterpFinalizeList->objs [0];
		c = qu_classof (x);

		if (c == lastclass)
		{
			u = lastsub;
		}
		else
		{
			u = c->get (x, QU_NAME__destroy, 0);
			lastclass = c;
			lastsub = u;
		}
		if (u)
		{
			assert (qu_sub_is_inst (u));
			assert (u == u->root);

			x = qu_sub_call0 (u, x);
			if (x == 0)
				QU_SYS_WARN ("%s", qu_thread_err_string ()->str);
		}
		qu_array_DELFIRST (__QuInterpFinalizeList);
	}
}



/*----------------------------------------------------------------------------
	eval
----------------------------------------------------------------------------*/

qu_any qu_interp_eval (qu_module mod, qu_string src, qu_dict imports, long options, qu_string path)
{
	volatile qu_module eval;

	QU_REQUIRE (QuECall, __QuInterpFrame);
	QU_REQUIRE (QuESecurity, qu_module_is_trusted (mod));

	eval = qu_module_new_eval (mod, src, path);
	if (options & QU_INTERP_EVAL_TRUSTED)
		qu_module_set_trusted (eval);

	if (imports)
	{
		qu_any k, v;
		long   i, j;

		i = j = 0;
		while ((v = qu_dict_walk (imports, &k, &i, &j)) != 0)
		{
			if (qu_string_is (k))
			{
#if 0
				qu_var y = qu_var_new ((qu_symbol) eval, (qu_string) k, QU_VAR_FLAG_PRIVATE);
				y->value = v;
				qu_module_setattr (eval, (qu_string) k, (qu_symbol) y);
#else
				if (qu_var_is (v) || qu_sub_is (v) || qu_class_is (v))
				{
					qu_module_setattr (eval, (qu_string) k, (qu_symbol) v);
				}
				else
				{
					qu_var y = qu_var_new ((qu_symbol) eval, (qu_string) k, QU_VAR_FLAG_PRIVATE);
					y->value = v;
					qu_module_setattr (eval, (qu_string) k, (qu_symbol) y);
				}
#endif
			}
		}
	}

	eval = qu_compile_eval (eval, options & QU_INTERP_EVAL_DOC);
	if (eval == 0)
		return 0;

	if (options & QU_INTERP_EVAL_DOC)
	{
		return eval;
	}
	else
	{
		volatile qu_frame f = qu_frame_new_module (eval);
		volatile qu_frame g = __QuInterpFrame;

		__QuInterpFrame = f;
		if (qu_interp_frame_next (f) == 0)
		{
			__QuInterpFrame = g;
			return 0;
		}
		__QuInterpFrame = g;

		if (imports && (options & QU_INTERP_EVAL_GETIMPORT))
		{
			qu_any k;
			long   i, j;

			i = j = 0;
			while (qu_dict_walk (imports, &k, &i, &j) != 0)
			{
				if (qu_string_is (k))
				{
					qu_var y = (qu_var) qu_module_getattr (eval, (qu_string) k);
					if (y && qu_var_is (y))
						qu_dict_setitem (imports, k, y->value);
				}
			}
		}

		if (options & QU_INTERP_EVAL_MODULE)
		{
			qu_gc_delete ((qu_obj) f);
			return eval;
		}
		else
		{
			qu_any r = f->sp[1];

			qu_gc_delete ((qu_obj) f);
			if (r == 0)
				return QuNil;
			return r;
		}
	}

	/*never*/
	QU_THROW (QuESystem);
}



/*----------------------------------------------------------------------------
	calls

	TODO - no need to call if the qu_sub does nothing
----------------------------------------------------------------------------*/

/*
	CHECKME - copy back values of a nested sub

	sub foo
		local xxx = 0
		sub bar
			xxx = 1
		;;
		bar
		println (xxx)   # 1
	;;

	If we dont do this, println (xxx) will print 0.

	!!! NOT A BUG: see qu_code.c
	We do not allow a child sub calling a nested parent sub because this
	generally means bad design. This eliminates the need to create closures.
*/
#define COPY_BACK \
	if (qu_sub_is_closure (u)) \
	{ \
		if (qu_sub_is_accum (u) && u->clex) \
			QU_OBJECT_LIST_COPY (u->clex->objs, f->ss, u->clex->len); \
	} \
	else if (u->root != u) \
	{ \
		qu_sub c = (qu_sub) p->code->parent; \
		assert (qu_sub_is (c)); \
		if (c != u) \
			QU_OBJECT_LIST_COPY (p->ss, f->ss, u->nlocal); \
	} \


#define RUN_FRAME \
	__QuThreadCurrent->depth++; \
	result = qu_interp_run_frame (f); \
	__QuThreadCurrent->depth--; \
	assert (!qu_frame_swapping (f)); \
	COPY_BACK


/* make sure __unsafe is not called by untrusted source */
#define ENSURE_SAFE(sub) \
QU_BEGIN \
	if (qu_sub_is_unsafe (sub)) \
	{ \
		qu_module m; \
		assert (__QuInterpFrame); \
		m = qu_frame_get_module (__QuInterpFrame); \
		assert (m); \
		QU_REQUIRE (QuESecurity, qu_module_is_trusted (m)); \
	} \
QU_END


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	non-generators using C stack
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_any qu_interp_call_cn (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
	typedef struct TCFrame TCFrame;
	struct TCFrame
	{
		qu_frame_s dummy;
		int pad;
	};

#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		TCFrame  frm;
		qu_any   stk [QU_FRAME_CSTACK_MAX];
		qu_frame p = __QuInterpFrame;
		qu_frame f = (qu_frame) &frm;
		qu_any   result;

		assert (!qu_int_is (f));
		assert (__QuInterpFrame);
		assert (! (qu_sub_is_extern (u) || qu_sub_is_jit (u) || qu_sub_is_generator (u)));

		f = qu_frame_new_c (f, stk, u, theself, argc, argv);
		QU_ENSURE (f);
		assert (qu_frame_is_cframe (f));

		__QuInterpFrame = f;
		RUN_FRAME;
		__QuInterpFrame = p;

		return result;
	}
}



qu_any qu_interp_call_c4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_cn (u, theself, 4, argv);
}



qu_any qu_interp_call_c3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_cn (u, theself, 3, argv);
}



qu_any qu_interp_call_c2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_cn (u, theself, 2, argv);
}



qu_any qu_interp_call_c1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_cn (u, theself, 1, &a1);
}



qu_any qu_interp_call_c0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_cn (u, theself, 0, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	non-generators using allocated frame
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_any qu_interp_call_an (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		volatile qu_frame f;
		qu_frame p = __QuInterpFrame;
		qu_any   result;

		assert (__QuInterpFrame);
		assert (! (qu_sub_is_extern (u) || qu_sub_is_jit (u) || qu_sub_is_generator (u)));

		f = qu_frame_new_a (u, theself, argc, argv);
		QU_ENSURE (f);

		assert (!qu_frame_is_cframe (f));

		RUN_FRAME;
		qu_gc_delete ((qu_obj) f);
		return result;
	}
}



qu_any qu_interp_call_a4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_an (u, theself, 4, argv);
}



qu_any qu_interp_call_a3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_an (u, theself, 3, argv);
}



qu_any qu_interp_call_a2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_an (u, theself, 2, argv);
}



qu_any qu_interp_call_a1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_an (u, theself, 1, &a1);
}



qu_any qu_interp_call_a0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_an (u, theself, 0, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	generators
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_frame qu_interp_call_gn (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		assert (__QuInterpFrame);
		assert (qu_sub_is_generator (u));

		return qu_frame_new_g (u, theself, argc, argv);
	}
}


qu_frame qu_interp_call_g4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_gn (u, theself, 4, argv);
}


qu_frame qu_interp_call_g3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_gn (u, theself, 3, argv);
}


qu_frame qu_interp_call_g2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_gn (u, theself, 2, argv);
}


qu_frame qu_interp_call_g1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_gn (u, theself, 1, &a1);
}


qu_frame qu_interp_call_g0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_gn (u, theself, 0, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__once functions using C stack
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_any qu_interp_call_once_cn (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		qu_any result;

		assert (qu_sub_is_once (u));

		result = qu_interp_call_cn (u, theself, argc, argv);
		if (result);
			qu_sub_set_once (u, result);
		return result;
	}
}



qu_any qu_interp_call_once_c4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_once_cn (u, theself, 4, argv);
}



qu_any qu_interp_call_once_c3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_once_cn (u, theself, 3, argv);
}



qu_any qu_interp_call_once_c2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_once_cn (u, theself, 2, argv);
}



qu_any qu_interp_call_once_c1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_once_cn (u, theself, 1, &a1);
}



qu_any qu_interp_call_once_c0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_once_cn (u, theself, 0, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__once functions using allocated frame
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_any qu_interp_call_once_an (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		qu_any result;

		assert (qu_sub_is_once (u));

		result = qu_interp_call_an (u, theself, argc, argv);
		if (result)
			qu_sub_set_once (u, result);
		return result;
	}
}



qu_any qu_interp_call_once_a4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_once_an (u, theself, 4, argv);
}



qu_any qu_interp_call_once_a3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_once_an (u, theself, 3, argv);
}



qu_any qu_interp_call_once_a2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_once_an (u, theself, 2, argv);
}



qu_any qu_interp_call_once_a1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_once_an (u, theself, 1, &a1);
}



qu_any qu_interp_call_once_a0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_once_an (u, theself, 0, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	non-generators JIT using C stack
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#define RUN_JIT_FRAME \
	__QuThreadCurrent->depth++; \
	result = qu_code_jit_call (f); \
	__QuThreadCurrent->depth--; \
	assert (!qu_frame_swapping (f)); \
	COPY_BACK


qu_any qu_interp_call_jit_cn (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
	typedef struct TCFrame TCFrame;
	struct TCFrame
	{
		qu_frame_s dummy;
		int pad;
	};

#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		TCFrame  frm;
		qu_any   stk [QU_FRAME_CSTACK_MAX];
		qu_frame p = __QuInterpFrame;
		qu_frame f = (qu_frame) &frm;
		qu_any   result;

		assert (!qu_int_is (f));
		assert (__QuInterpFrame);
		assert (! (qu_sub_is_extern (u) || qu_sub_is_generator (u)));
		assert (qu_sub_is_jit (u));

		f = qu_frame_new_c (f, stk, u, theself, argc, argv);

		QU_ENSURE (f);
		assert (qu_frame_is_cframe (f));
		qu_frame_set_jit (f);

		__QuInterpFrame = f;
		RUN_JIT_FRAME;
		__QuInterpFrame = p;

		return result;
	}
}



qu_any qu_interp_call_jit_c4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_jit_cn (u, theself, 3, argv);
}



qu_any qu_interp_call_jit_c3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_jit_cn (u, theself, 3, argv);
}



qu_any qu_interp_call_jit_c2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_jit_cn (u, theself, 2, argv);
}



qu_any qu_interp_call_jit_c1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_jit_cn (u, theself, 1, &a1);
}



qu_any qu_interp_call_jit_c0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_jit_cn (u, theself, 0, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	non-generators JIT using allocated frame
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

qu_any qu_interp_call_jit_an (qu_sub u, qu_any theself, int argc, qu_any *argv)
{
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace (u);
#endif

	ENSURE_SAFE (u);
	{
		volatile qu_frame f;
		qu_frame p = __QuInterpFrame;
		qu_any   result;

		assert (__QuInterpFrame);
		assert (! (qu_sub_is_extern (u) || qu_sub_is_generator (u)));
		assert (qu_sub_is_jit (u));

		f = qu_frame_new_a (u, theself, argc, argv);

		QU_ENSURE (f);
		assert (!qu_frame_is_cframe (f));
		qu_frame_set_jit (f);

		__QuInterpFrame = f;
		RUN_JIT_FRAME;
		__QuInterpFrame = p;

		qu_gc_delete ((qu_obj) f);
		return result;
	}
}



qu_any qu_interp_call_jit_a4 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3, qu_any a4)
{
	qu_any argv [4] = {a1, a2, a3, a4};
	return qu_interp_call_jit_an (u, theself, 4, argv);
}



qu_any qu_interp_call_jit_a3 (qu_sub u, qu_any theself, qu_any a1, qu_any a2, qu_any a3)
{
	qu_any argv [3] = {a1, a2, a3};
	return qu_interp_call_jit_an (u, theself, 3, argv);
}



qu_any qu_interp_call_jit_a2 (qu_sub u, qu_any theself, qu_any a1, qu_any a2)
{
	qu_any argv [2] = {a1, a2};
	return qu_interp_call_jit_an (u, theself, 2, argv);
}



qu_any qu_interp_call_jit_a1 (qu_sub u, qu_any theself, qu_any a1)
{
	return qu_interp_call_jit_an (u, theself, 1, &a1);
}



qu_any qu_interp_call_jit_a0 (qu_sub u, qu_any theself)
{
	return qu_interp_call_jit_an (u, theself, 0, 0);
}



/*----------------------------------------------------------------------------
	interface
----------------------------------------------------------------------------*/

qu_any qu_interp_frame_next (qu_frame f)
{
	qu_any result;

#if 0
	/* not necessarily - might be a module init code -> see qu_interp_eval */
	assert (qu_frame_is_generator (f));
#endif

  again:

	__QuThreadCurrent->depth++;

	if (qu_frame_is_jit (f))
		result = qu_code_jit_call (f);
	else
		result = qu_interp_run_frame (f);

	__QuThreadCurrent->depth--;

	if (result == 0)
	{
		/*
			Returning with an exception.
			Dont use QU_RETHROW otherwise the exception location will be displayed
			as qu_interp_frame_next instead of the original location.
		*/
		return 0;
	}

	if (qu_frame_swapping (f))
	{
		/*
			We return here because the frame was swapped (the SWAP instruction).
		*/

		qu_frame_unset_swapping (f);

		f = result;
		assert (qu_frame_is (f));
		assert (qu_frame_is_generator (f));
		assert (!qu_frame_running (f));
		assert (!qu_frame_done (f));

		goto again;
	}

	return result;
}



static void _interp_finalize_module_collect (qu_array a, qu_module m)
{
	qu_flag_set (m, QU_MODULE_FLAG_FINALIZED);

	if (m->use)
	{
		qu_module u;
		long      b = 0, i = 0;

		while ((u = (qu_module) qu_set_walk (m->use, &b, &i)) != 0)
		{
			assert (qu_module_is (u));
			if (!qu_flag_test (u, QU_MODULE_FLAG_FINALIZED))
				_interp_finalize_module_collect (a, u);
		}
	}

	if (m != MainModule)
		qu_array_push (a, m);
}



static void _interp_finalize_module (qu_module m)
{
	volatile qu_array a = qu_array_new (8);
	qu_sub f;
	long   i;

	_interp_finalize_module_collect (a, m);
	for (i = a->len - 1; i >= 0; --i)
	{
		m = a->objs[i];
		f = qu_attr_find (m->attrs, QU_NAME__destroy);
		if (f && qu_sub_is (f))
		{
			if (QU_APP_VERBOSE)
				fprintf (stderr, "exit: %s\n", m->name->str);
			qu_sub_call0 (f, m);
		}
	}

	/* do the main module last */
	if (QU_APP_VERBOSE)
		fprintf (stderr, "exit: main\n");

	f = qu_attr_find (MainModule->attrs, QU_NAME__destroy);
	if (f && qu_sub_is (f))
	{
		qu_sub_call0 (f, MainModule);
	}
}



/*
	Run the main program.
*/
qu_any qu_interp_run_main (qu_module m)
{
	qu_thread t = __QuThreadMain; /* for marking */
	volatile qu_frame f;
	qu_any result;

	assert (__QuInterpFrame == 0);
	assert (__QuThreadCurrent);

	/* set up the main thread stack base - everything below is garbage */
	t->stackhi = (qu_any *) &t;

	/* clear possible previous exceptions */
	t->eclass = 0;
	t->emsg   = 0;
	t->ewhere = 0;
	t->eline  = 0;

	/* get rid of them to get a better startup time */

	if (__QuGcNeedReclaim)
		qu_gc_reclaim ();

	MainModule = m;

	f = qu_frame_new_module (m);
	result = qu_interp_run_frame (f);

	__QuInterpExiting = 1;
	__QuSigINT = __QuSigTERM = 0;

	if (result == 0
		&& (__QuThreadCurrent->eclass == QuEExitClass || __QuThreadCurrent->eclass == QuEAssertClass)
		&& QuSys__quit__->value != QuNil)
	{
		qu_sub u = QuSys__quit__->value;

		assert (qu_sub_is (u));
		assert (!qu_sub_is_extern (u));
		assert (!qu_sub_is_generator (u));
		assert (qu_sub_is_static (u));

		__qu_atomic_begin;
		qu_sub_call0 (u, QuNil);
		__qu_atomic_end;
	}

	/* run all finalizers on exit */
	qu_interp_run_finalize ();

	/* run all module __destroy function */
	_interp_finalize_module (MainModule);

	return result;
}



/*
	Run a new thread - called by qu_thread.c.
*/
/*
	Must keep Threads rooted until they're dead otherwise Threads not kept in
	global variables might be freed prematurely.

	sub blah
		for __ in 1, 10000
			__thread__.id     --> otherwise this might cause a crash
			pass
		;;
	;;

	qu_thread (@doit)  \
	qu_thread (@doit)   \__ not store in global variables
	qu_thread (@doit)   /
	qu_thread (@doit)  /

	This is something nobody except interp.c can do because the removal must happen
	*before* the global lock is released.

	TODO - need to tweak this for better performance
*/
static qu_array InterpThreadList;

void qu_interp_run_thread (qu_thread t, qu_any self)
{
	assert (t);
	assert (t == qu_thread_self ());

	/* mark the thread stack - everything below is garbage */
	t->stacklo = t->stackhi = (qu_any *) &t;
	assert (__QuInterpFrame);

	INTERP_LOCK (t);
	qu_array_push_quick (InterpThreadList, t); /* <-- must be done after we have the lock */

	t->result = qu_sub_calln (t->sub, self, t->args->len, t->args->objs);
	t->sub  = 0;
	t->args = 0;

	qu_array_del_quick (InterpThreadList, t);  /* <-- must be done before releasing the lock */
	INTERP_UNLOCK;
}



/*
	Return the qu_module or qu_class object of the current running qu_frame.
	!!! It is an error to call this function before the main program runs.
*/
qu_symbol qu_interp_where ()
{
	assert (__QuInterpFrame);
	return __QuInterpFrame->code->parent;
}



/*
	Return the associated 'self' object of the current running qu_frame.
	!!! It is an error to call this function before the main program runs.
*/
qu_any qu_interp_self ()
{
	assert (__QuInterpFrame);
	return __QuInterpFrame->self;
}



/*
	Return the associated qu_module object of the current running qu_frame.
*/
qu_module qu_interp_get_module ()
{
	if (__QuInterpFrame)
		return qu_frame_get_module (__QuInterpFrame);
	return QuModuleMain;
}



/*
	Get the qu_class or qu_module.
	!!! It is an error to call this function before the main program runs.
*/
qu_symbol qu_interp_scope ()
{
	qu_symbol x = qu_interp_where ();

	assert (x);
	if (qu_sub_is (x))
	{
		qu_class c = qu_sub_get_class ((qu_sub) x);
		if (c)
			return (qu_symbol) c;
	}
	x = (qu_symbol) qu_interp_get_module ();
	assert (x);
	return x;
}



/*
	Get a string representation of where the current qu_frame is running.
*/
qu_string qu_interp_where_string ()
{
	if (__QuInterpFrame)
	{
		qu_string s = qu_frame_where (__QuInterpFrame);
		if (s)
			return s;
	}
	{
		static qu_string s = 0;
		if (s == 0)
			s = qu_string_new_staticz ("__main__");
		return s;
	}
}



/*
	Returns 1 if the caller comes from an untrusted qu_module, 0 otherwise.
	At initialization phase, this function always return 0.
*/
int qu_interp_unsafe ()
{
	qu_frame f = __QuInterpFrame;
	if (f)
	{
		qu_module m = qu_frame_get_module (f);
		assert (m);
		return !qu_module_is_trusted (m);
	}
	return 0;
}



/*
	Add a finalization function to run by the interpreter.
*/
void qu_interp_add_finalizer (qu_any self)
{
	assert (!qu_class_is (self));
	qu_array_push (__QuInterpFinalizeList, self);
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

void qu_interp__init ()
{
	qu_thread_set_blocking (__QuThreadMain);
	__QuThreadCurrent = 0;
	qu_interp_lock ();

	assert (__QuThreadCurrent = __QuThreadMain);
	assert (qu_thread_self () == __QuThreadMain);
	assert (!qu_thread_blocking (__QuThreadMain));

	__QuInterpFinalizeList = qu_array_new_root (128);
	InterpThreadList       = qu_array_new_root (128);

	__QuGcMemTrap = _interp_memtrap;
	MemTrapName   = qu_string_new_staticz ("__memory__");
}
