/* sig.c - Copyright (c) Marc Krisnanto */

#include "qu.h"

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif


/*
	When a signal is caught, we increment the corresponding counter.
	When acknowledged, we decrement it.
*/
volatile int __QuSigTERM = 0;
volatile int __QuSigINT  = 0;
volatile int __QuSigHUP  = 0;
volatile int __QuSigALRM = 0;
volatile int __QuSigPIPE = 0;
volatile int __QuSigCHLD = 0;
volatile int __QuSigUSR1 = 0;
volatile int __QuSigUSR2 = 0;

/* hooks */
volatile qu_sighook_f __QuSigHookINT  = NULL;
volatile qu_sighook_f __QuSigHookCHLD = NULL;
volatile qu_sighook_f __QuSigHookALRM = NULL;
volatile qu_sighook_f __QuSigHookHUP  = NULL;
volatile qu_sighook_f __QuSigHookPIPE = NULL;
volatile qu_sighook_f __QuSigHookUSR1 = NULL;
volatile qu_sighook_f __QuSigHookUSR2 = NULL;


/* whether or not signals are blocked */
static int BlockSig = 0;


/*-----------------------------------------------------------------------------
	helpers
-----------------------------------------------------------------------------*/

static void _sig_setsig (int s, void (*func)(int))
{
	struct sigaction sa, old;

	sa.sa_handler = func;
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;

	if (sigaction (s, &sa, &old))
		QU_FATAL ("sigaction failed");
}



/*-----------------------------------------------------------------------------
	handlers
-----------------------------------------------------------------------------*/

#define _perr(sig) \
	fprintf (stderr, "***" sig " " QU_THREAD_IDFORMAT "***\n", qu_thread_id ())

static void _sig_SEGV (int s)
{
	_perr ("SEGV");
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace_show ();
#endif
	abort ();
}


#ifdef SIGBUS
static void _sig_BUS (int s)
{
	_perr ("BUS");
#ifdef QU_DEBUG_SEGV
	__qu_sub_trace_show ();
#endif
	abort ();
}
#endif


static void _sig_ILL (int s)
{
	_perr ("ILL");
	abort ();
}



static void _sig_QUIT (int s)
{
	_perr ("QUIT");
	abort ();
}



static void _sig_TERM (int s)
{
	_sig_setsig (s, _sig_TERM);
	__QuSigTERM++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("TERM");
}



static void _sig_INT  (int s)
{
	_sig_setsig (s, _sig_INT);
	__QuSigINT++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("INT");
	if (__QuSigHookINT)
		__QuSigHookINT ();
}



static void _sig_HUP (int s)
{
	_sig_setsig (s, _sig_HUP);
	__QuSigHUP++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("HUP");
	if (__QuSigHookHUP)
		__QuSigHookHUP ();
}



static void _sig_ALRM (int s)
{
	_sig_setsig (s, _sig_ALRM);
	__QuSigALRM++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("ALRM");
	if (__QuSigHookALRM)
		__QuSigHookALRM ();
}



static void _sig_PIPE (int s)
{
	_sig_setsig (s, _sig_PIPE);
	__QuSigPIPE++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("PIPE");
	if (__QuSigHookPIPE)
		__QuSigHookPIPE ();
}



#ifndef __WALL
#define __WALL 0
#endif

static void _sig_CHLD  (int s)
{
	_sig_setsig (s, _sig_CHLD);

#if 0
	/* don't do this - it's annoying */
	if (QU_APP_DEBUG)
		_perr ("CHLD");
#endif

	__QuSigCHLD++;
	while (waitpid (-1, NULL, WNOHANG | __WALL) > 0);
	__QuSigCHLD--;
	if (__QuSigHookCHLD)
		__QuSigHookCHLD ();
}



static void _sig_USR1 (int s)
{
	_sig_setsig (s, _sig_USR1);
	__QuSigUSR1++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("USR1");
	if (__QuSigHookUSR1)
		__QuSigHookUSR1 ();
}



static void _sig_USR2 (int s)
{
	_sig_setsig (s, _sig_USR2);
	__QuSigUSR2++;
	__QuInterpGotSignal = 1;
	if (QU_APP_DEBUG)
		_perr ("USR2");
	if (__QuSigHookUSR2)
		__QuSigHookUSR2 ();
}



/*-----------------------------------------------------------------------------
	init
-----------------------------------------------------------------------------*/

/*
	All worker threads must call Sig_initworker before they do their
	work, i.e. to make them deaf to signals they are not suppose to hear.
*/
void qu_sig_init_worker ()
{
	if (BlockSig)
	{
		sigset_t mask;

		sigemptyset (&mask);
		/*sigaddset (&mask, SIGPIPE);*/
		qu_thread_sigmask (SIG_SETMASK, &mask, NULL);
	}
}



void qu_sig__init (int blocksig)
{
	BlockSig = blocksig;

	if (!blocksig)
		return;

#ifdef SIGBUS
	_sig_setsig (SIGBUS  , _sig_BUS );
#endif
	_sig_setsig (SIGSEGV , _sig_SEGV);
	_sig_setsig (SIGILL  , _sig_ILL );
	_sig_setsig (SIGTERM , _sig_TERM);
	_sig_setsig (SIGHUP  , _sig_HUP );
	_sig_setsig (SIGALRM , _sig_ALRM);
	_sig_setsig (SIGPIPE , _sig_PIPE);
	_sig_setsig (SIGCHLD , _sig_CHLD);
	_sig_setsig (SIGINT  , _sig_INT );

	_sig_setsig (SIGQUIT , _sig_QUIT);
	_sig_setsig (SIGUSR1 , _sig_USR1);
	_sig_setsig (SIGUSR2 , _sig_USR2);
	_sig_setsig (SIGFPE  , SIG_IGN  );
#ifdef SIGWINCH
	_sig_setsig (SIGWINCH, SIG_IGN  );
#endif
}
