/* Lightning.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) module Lightning
	This module provides a JIT (just in time) compiler for Qu.
	It uses GNU Lighting by Paolo Bonzini.

	To optimize your code, all you have to do is run your program with the
	C{-O} option. Since 1.10 there is no need to say C{use Lightning}.

	Try run the C{demo/sort} program with and without C{-O} option to see the
	difference.

	@note
	Currently Lightning can not optimize generators and nested Sub (i.e. Sub
	which contains other Sub).

	Qu is already a fast and efficient interpreter (well, being dynamic and
	all). The difference between bytecode and JIT code is generally in the
	range 1 to 8 depending on the functions.

	Module and class intialization codes are not optimized since they only
	run once.

	The current internal settings are to optimize everything. This makes JIT
	code much longer (and of course, faster). So if you don't really need
	speed, using the -O options is not what you want.

	If you like, you can edit C{Lightning.c} and set the C{OPTIMIZE_XXX}
	macros to your preference.

	If you invoke your program with the C{-d} option, Lightning will display
	the original bytecode size and JIT code size for every Sub it translates.

	@warn
		Currently Lightning only works with the current processors:
		x86, sparc, power pc.
*/


#ifdef LIGHTNING_CPU_NONE


int Lightning_init (qu_module m)
{
	QU_SYS_WARN ("Lightning is not available for your platform");
	return 1;
}



#else /*LIGHTNING_CPU_NONE*/


#include "lightning.h"
#include "Inlines.h"


#define STACK_DIRECTION 1

#if STACK_DIRECTION != 1
#error "QuLightning works with stack growing downward only"
#endif


/*
	all OPTIMIZE_XXX make instructions much longer
	you generally don't need it
	see the module documentation below
*/


/* define to inline Int methods */
#ifdef QU_RSHIFT_PRESERVES_SIGNBIT
#define OPTIMIZE_INT
#endif

/* define to optimize tail call */
#define OPTIMIZE_TAIL_CALL

/* define to optimize more */
#define OPTIMIZE_MORE


/*----------------------------------------------------------------------------
	debug
-----------------------------------------------------------------------------*/

/* define to debug */
#undef DEBUG_JIT

#ifdef DEBUG_JIT
#undef assert
#define assert(cond) \
QU_BEGIN \
	if (!(cond)) \
		QU_FATAL ("%s line %d assertion failed " #cond, __FUNCTION__, __LINE__); \
QU_END
#endif /*DEBUG_JIT*/


/* define to show instructions as they are executed */
#undef SHOW


/*----------------------------------------------------------------------------
	mapping jumps
-----------------------------------------------------------------------------*/

/*
	For mapping jump targets.
	We allocate Jmp and Ip permanently, resize them as necessary.
*/
typedef struct jmp_s jmp_s;
struct jmp_s
{
	int       type; /* 0:patch_at  1:patch_movi */
	int       ip;   /* bytecode ip */
	jit_insn *ofs;  /* jit code where to patch */
};

static jmp_s     *Jmp;
static int        JmpSize = 0;
static int        JmpCount;

static jit_insn **Ip;
static int        IpSize = 0;


static jmp_s *_add_jmp (int type, int ip)
{
	jmp_s *j;

	if (JmpSize == 0)
	{
		JmpSize = 1024;
		Jmp = qu_mem_alloc (JmpSize * sizeof (jmp_s));
	}
	else if (JmpCount >= JmpSize)
	{
		JmpSize += 1024;
		Jmp = qu_mem_realloc (Jmp, JmpSize * sizeof (jmp_s));
	}

	j       = &Jmp[JmpCount++];
	j->type = type;
	j->ip   = ip;
	return j;
}



static void _map_ip (int ip, jit_insn *j)
{
	if (IpSize == 0)
	{
		IpSize = 8192;
		Ip = qu_mem_alloc (IpSize * sizeof (jit_insn *));
	}
	else if (ip >= IpSize)
	{
		IpSize += 8192;
		Ip = qu_mem_realloc (Ip, IpSize * sizeof (jit_insn *));
	}
	Ip[ip] = j;
}



/*----------------------------------------------------------------------------
	block info
-----------------------------------------------------------------------------*/

/*
	We store block information in the JIT code directly.
	This is due to the way Qu block and exception works.
*/

typedef struct block_s block_s;
struct block_s
{
	int       nextblk; /* the type of the block if nextip points to one, 0 otherwise */
	jit_insn *contip;  /* where to 'continue' */
	jit_insn *nextip;  /* where to jump when leaving the block, JIT code */
	long      ip;      /* where to jump when leaving the block, Qu bytecode */
	block_s  *ofs;     /* JIT code offest where to store the info */
};

#define MAKE_ROOM_FOR_BLOCK_STRUCT \
QU_BEGIN \
	jit_movi_p (R0, 0); \
	jit_movi_p (R0, 0); \
	jit_movi_p (R0, 0); \
	jit_movi_p (R0, 0); \
	jit_movi_p (R0, 0); \
QU_END

/* !!! this must match qu_frblk_s of qu_frame.h */
typedef struct jitfrblk_s jitfrblk_s;
struct jitfrblk_s
{
	qu_any   obj;    /* switch value or for-in iterator object or 0 */
	long     ex1;    /* \__ for builtin class iterators */
	long     ex2;    /* /   used internally by interp.c */
	long     cont;   /* where to 'continue' */
	block_s *block;  /* this block's info       <-- the only difference */
};

/*
	We allocate Block permanently, resize it as necessary.
*/
static block_s *Block;
static int      BlockSize = 0;
static int      BlockCount = 0;

static block_s *_add_block (block_s *ofs, int nextip)
{
	block_s *b;

	if (BlockSize == 0)
	{
		BlockSize = 1024;
		Block = qu_mem_alloc (BlockSize * sizeof (block_s));
	}
	else if (BlockCount >= BlockSize)
	{
		BlockSize += 1024;
		Block = qu_mem_realloc (Block, BlockSize * sizeof (block_s));
	}

	b          = &Block[BlockCount++];
	b->ip      = nextip;
	b->ofs     = ofs;
	b->nextip  = 0;
	b->contip  = 0;
	b->nextblk = 0;
	return b;
}



/*----------------------------------------------------------------------------
	registers
	GNU Lightning only guarantees these across platforms
----------------------------------------------------------------------------*/

#define R0   JIT_R0
#define R1   JIT_R1
#define R2   JIT_R2
#define I0   JIT_V0

#define RET  JIT_RET
#define SP   JIT_SP

/* qu_sub_is_jit */
#define FRM  JIT_V1   /* the Frame object all the way through */
#define LP   JIT_V2   /* start of locals = Frame->ss */

/* qu_sub_is_extern */
#define BSP  JIT_V1   /* the original stack pointer */
#define LP   JIT_V2   /* start of locals */

/*
	!!! DO NOT USE V2 and V1
	GNU Lightning guarantees that V0, V1, and V2 are preserved across function
	calls. We use V2 to get local variables (Frame->ss) all the way through
	and FRM to contain the qu_frame object.

	!!! jit_setlocal CHANGES R2!
	Unfortunately jit_stxi_p does not work -> I get a SEGV with it.
	So jit__setlocal uses jit_str_p and R2 to store an intermmediate value.

	Apparently the jit_ret instruction does not restore the stack so we
	have to do it ourselves. Annoying ...
*/



/*----------------------------------------------------------------------------
	addresses, for the current compiled code
-----------------------------------------------------------------------------*/

/* whether generating qu_sub_is_extern,
	i.e. qu_any foo (qu_any self, int argc, qu_any *argv) */
static int GeneratingExtern;

/* the Sub we are currently generating */
static qu_sub TheSub;

/* start of JIT code */
static jit_insn *StartOfCode;

/* qu_sub_is_jit: index of the Frame object, used with jit_getarg */
static int ArgFrame;

/* qu_sub_is_extern : index of the 'self' argument */
static int ArgSelf;

/* qu_sub_is_extern : index of the 'argc' argument */
static int ArgArgc;

/* qu_sub_is_extern : index of the 'argv' argument */
static int ArgArgv;

/* code to leave a block starts here */
static jit_insn *LeaveAddr;

/* code to rethrow exception starts here */
static jit_insn *ThrowAddr;

/* END code starts here */
static jit_insn *EndAddr;

/* tail call jumps here */
static jit_insn *TailAddr;

/* whether we must set __QuInterpFrame for every call */
static int MustSetInterpFrame = 1;

/* 1 = measuring the length of instructions at initialization time */
static int Measuring = 1;

/* offset of block_s info from the block ENTER address */
static long BlockOfs = 0;


#ifdef SHOW
static void show_me (const char *f)
{
	fprintf (stderr, "::: %s\n", f);
}

#define SHOWME \
QU_BEGIN \
	jit_prepare   (1); \
	jit_movi_p    (R0, __FUNCTION__); \
	jit_pusharg_p (R0); \
	jit_finish    (show_me); \
QU_END

static void show_reg (const char *msg, qu_any x)
{
	fprintf (stderr, "::: %s = %p\n", msg, x);
}

#define SHOWREG(msg, reg) \
QU_BEGIN \
	jit_prepare   (2); \
	jit_pusharg_p (reg); \
	jit_movi_p    (R0, msg); \
	jit_pusharg_p (R0); \
	jit_finish    (show_reg); \
QU_END

#else /*SHOW*/

#define SHOWME            QU_BEGIN QU_END
#define SHOWREG(msg, reg) QU_BEGIN QU_END

#endif /*SHOW*/



/*----------------------------------------------------------------------------
	our own extended instructions to simplify things a bit
----------------------------------------------------------------------------*/


	/********************************************************

	    be careful with jit__callxxx -> they may change R2

	*********************************************************/


/* to make patched targets more obvious */
#define __later__                 jit_forward ()

/* leave a block */
#define jit__leave                jit_jmpi (LeaveAddr)

/* goto END */
#define jit__end                  jit_jmpi (EndAddr)

/* increment SP */
#define jit__incr                 jit_subi_p (SP, SP, sizeof (qu_any))
#define jit__up(n)                jit_subi_p (SP, SP, (n) * sizeof (qu_any))

/* decrement SP */
#define jit__decr                 jit_addi_p (SP, SP, sizeof (qu_any))
#define jit__down(n)              jit_addi_p (SP, SP, (n) * sizeof (qu_any))

/* get value of 'self' */
#define jit__getself(reg) \
QU_BEGIN \
	if (!GeneratingExtern || Measuring) \
		{ jit_ldxi_p (reg, FRM, offsetof (qu_frame_s, self)); } \
	else \
		{ jit_getarg_p (reg, ArgSelf); } \
QU_END

/* get local variable */
#define jit__getlocal(reg, i)     jit_ldxi_p (reg, LP, (i) * sizeof (qu_any))

/* set local variable */
#define jit__setlocal(i, reg)     jit_addi_p (R2, LP, (i) * sizeof (qu_any)); \
                                  jit_str_p  (R2, reg)

#define jit__setInterpFrame \
QU_BEGIN \
	if (MustSetInterpFrame) \
	{ \
		if (!Measuring) \
			assert (!GeneratingExtern); \
		jit_movi_p (R2, &__QuInterpFrame); \
		jit_str_p  (R2, FRM); \
	} \
QU_END


/* call function with arguments, may throw exception */
#define jit__call(func)           jit__setInterpFrame; \
                                  jit_finish  (func); \
                                  jit_beqi_ul (ThrowAddr, RET, 0)

/* call register without arguments, may throw exception */
#define jit__callr(reg)           jit__setInterpFrame; \
                                  jit_finishr (reg); \
                                  jit_beqi_ul (ThrowAddr, RET, 0)

/* call function without arguments, may throw exception */
#define jit__call0(func)          jit__setInterpFrame; \
                                  jit_calli   (func); \
                                  jit_beqi_ul (ThrowAddr, RET, 0)

/* call register without arguments, may throw exception */
#define jit__call0r(reg)          jit__setInterpFrame; \
                                  jit_callr   (reg); \
                                  jit_beqi_ul (ThrowAddr, RET, 0)

/* call function with arguments, always succeeds */
#define jit__callsafe(func)       jit__setInterpFrame; \
                                  jit_finish (func)

/* call register with arguments, always succeeds */
#define jit__callsafer(reg)       jit__setInterpFrame; \
                                  jit_finishr (reg)

/* call function without arguments, always succeeds */
#define jit__callsafe0(func)      jit__setInterpFrame; \
                                  jit_calli  (func)

/* call register without arguments, always succeeds */
#define jit__callsafe0r(reg)      jit__setInterpFrame; \
                                  jit_callr  (func)


/*----------------------------------------------------------------------------
	reversing the stack content
----------------------------------------------------------------------------*/

#define REVERSE(n, sp) \
QU_BEGIN \
	if ((n) > 1) \
		_reverse (n, sp); \
QU_END

static void _reverse (int n, qu_any *sp)
{
	assert (n > 1);

	if (n == 2)
	{
		register qu_any t = sp[0];
		sp[0] = sp[1];
		sp[1] = t;
	}
	else if (n == 3)
	{
		register qu_any t = sp[0];
		sp[0] = sp[2];
		sp[2] = t;
	}
	else
	{
		register qu_any *i = sp;
		register qu_any *j = i + n;
		register qu_any  t;

		n /= 2;
		while (n--)
		{
			t = *--j;
			*j = *i;
			*i++ = t;
		}
	}
}



/*----------------------------------------------------------------------------
	emit
----------------------------------------------------------------------------*/

#ifdef OPTIMIZE_MORE
static int _throw_ECall ()
{
	QU_THROW (QuECall);
}
#endif



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ARG
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pARG (int n, qu_any *sp)
{
	REVERSE (n, sp);
	return qu_arg_new (n, sp);
}


static void _emit_ARG (int n)
{
	SHOWME;

	jit_prepare    (2);
	jit_pusharg_p  (SP);
	jit_movi_i     (I0, n);
	jit_pusharg_i  (I0);
	jit__callsafe  (pARG);
	jit__down      (n);
	jit_pushr_p    (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ARGC
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_ARGC ()
{
	int ext = GeneratingExtern;

	SHOWME;

	if (Measuring)
		ext = 0;

	if (!ext)
		jit_ldxi_l   (R0, FRM, offsetof (qu_frame_s, argc));
	else
		jit_getarg_i (R0, ArgArgc);

	jit_lshi_l  (R0, R0, 1);
	jit_ori_l   (R0, R0, 1);
	jit_pushr_p (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ARRAY
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pARRAY (int n, qu_any *sp)
{
	REVERSE (n, sp);
	return qu_array_new_argv (n, sp);
}



static void _emit_ARRAY (int n)
{
	SHOWME;

	if (Measuring)
		n = 1;

	if (n)
	{
		jit_prepare    (2);
		jit_pusharg_p  (SP);
		jit_movi_i     (I0, n);
		jit_pusharg_p  (I0);
		jit__callsafe  (pARRAY);
		jit__down      (n);
		jit_pushr_p    (RET);
	}
	else
	{
		jit_prepare    (1);
		jit_movi_i     (I0, n);
		jit_pusharg_i  (I0);
		jit__callsafe  (qu_array_new);
		jit_pushr_p    (RET);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	AS
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_AS ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_as);
	jit_pushr_p   (RET);
}



static void _emit_ASLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_as);
	jit_pushr_p   (RET);
}



static void _emit_ASLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_as);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	AUTOATTR AUTOATTREX AUTOATTREX2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pAUTOATTR (qu_any o, qu_any x)
{
	if (qu_sub_is (x))
	{
		if (!qu_sub_noauto (x))
		{
			if (qu_sub_is_class (x))
			{
				if (!qu_class_is (o))
					o = qu_classof (o);
			}
			else if (qu_sub_is_inst (x))
			{
				if (qu_class_is (o))
					QU_THROW (QuECall);
			}
			return qu_sub_call0 (x, o);
		}
	}
	else if (qu_method_is (x))
	{
		return qu_method_calln_attr (x, o, 0, NULL);
	}
	return x;
}



static void _emit_AUTOATTR ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (pAUTOATTR);
	jit_pushr_p   (RET);
}



static void _emit_AUTOATTREX ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_pushr_p   (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (pAUTOATTR);
	jit_pushr_p   (RET);
}



static void _emit_AUTOATTREX2 ()
{
	SHOWME;

	jit_popr_p    (R2);
	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_pushr_p   (R0);
	jit_pushr_p   (R1);
	jit_prepare   (2);
	jit_pusharg_p (R2);
	jit_pusharg_p (R0);
	jit__call     (pAUTOATTR);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	BINOP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_BINOP (long i)
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit_movi_l    (I0, i);
	jit_pusharg_l (I0);
	jit__call     (qu_sys_binop);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	BREAK
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifndef OPTIMIZE_MORE

static void pBREAK (qu_frame F, int n)
{
	F->level -= (n - 1);
	if (F->level < F->atomic)
	{
		__qu_atomic_end;
		F->atomic = -1;
	}
}

#endif


static void _emit_BREAK (int n)
{
#ifdef OPTIMIZE_MORE

	jit_insn *j;

	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	    jit_addi_p (R0, FRM, offsetof (qu_frame_s, level));
	    jit_ldr_i  (I0, R0);
	    jit_subi_i (I0, I0, n - 1);
	    jit_str_i  (R0, I0);
	    jit_addi_p (R2, FRM, offsetof (qu_frame_s, atomic));
	    jit_ldr_i  (R1, R2);
	j = jit_bgtr_i (__later__, I0, R1);
	    jit_movi_p (R0, &__QuThreadCurrent);
	    jit_ldr_p  (R0, R0);
	    jit_addi_p (R0, R0, offsetof (qu_thread_s, atomic));
	    jit_ldr_i  (I0, R0);
	    jit_subi_i (I0, I0, 1);
	    jit_str_i  (R0, I0);
	    jit_movi_i (I0, -1);
	    jit_str_i  (R2, I0);
	    jit_patch  (j);
	    jit__leave;

#else

	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_prepare   (2);
	jit_movi_i    (I0, n);
	jit_pusharg_i (I0);
	jit_pusharg_p (FRM);
	jit__callsafe (pBREAK);
	jit__leave;

#endif

}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CALL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pCALL (qu_any self, int n, qu_any *sp)
{
	qu_any   x = sp [n];
	qu_class c = qu_classof (x);

	REVERSE (n, sp);

  callagain:

	if (c == QuSubClass)
	{
		assert (!qu_sub_is_class (x) && !qu_sub_is_inst (x));
		return qu_sub_calln (x, self, n, sp);
	}
	else if (c == QuMethodClass)
	{
		return qu_method_calln (x, n, sp);
	}
	else if (c == QuStructClass)
	{
		return qu_record_new (x, n, sp);
	}
	else if (c == QuClassClass)
	{
#if 0
		qu_any b = c->get (x, QU_NAME_new, 0);

		QU_ENSURE (b);
		assert (qu_sub_is (b) && qu_sub_is_class (b));

		return qu_sub_calln (b, x, n, sp);
#else
		assert (c->fnew);
		return qu_sub_calln (c->fnew, x, n, sp);
#endif
	}
	else if (c == QuModuleClass)
	{
		x = qu_module_getattr_callable_name (x, 1);
		if (x == 0)
			return 0;
		c = qu_classof (x);
		assert (qu_class_callable (c));
		goto callagain;
	}
	else
	{
		qu_any b = qu_classof (x)->get (x, QU_NAME__call, 0);

		if (b == 0)
			QU_THROW (QuECall);

		assert (qu_sub_is (b) && qu_sub_is_inst (b));

		return qu_sub_calln (b, x, n, sp);
	}
	QU_THROW (QuECall);
}



#ifdef OPTIMIZE_MORE

static void _emit_reverse_args (int n)
{
	if (Measuring)
		n = 5;

	if (n > 1)
	{
		if (n == 2)
		{
			jit_addi_p  (R2, SP, sizeof (qu_any));
			jit_ldr_p   (R0, SP);
			jit_ldr_p   (R1, R2);
			jit_str_p   (R2, R0);
			jit_str_p   (SP, R1);
		}
		else if (n == 3)
		{
			jit_addi_p  (R2, SP, 2 * sizeof (qu_any));
			jit_ldr_p   (R0, SP);
			jit_ldr_p   (R1, R2);
			jit_str_p   (R2, R0);
			jit_str_p   (SP, R1);
		}
		else
		{
			jit_insn *j;

				jit_movr_p    (R0, SP);
			    jit_addi_p    (R1, R0, (n - 1) * sizeof (qu_any));
			j = jit_get_label ();
			    jit_ldr_p     (R2, R0);
		        jit_ldr_p     (I0, R1);
			    jit_str_p     (R0, I0);
			    jit_str_p     (R1, R2);
			    jit_subi_p    (R1, R1, sizeof (qu_any));
			    jit_addi_p    (R0, R0, sizeof (qu_any));
			    jit_bltr_p    (j, R0, R1);
		}
	}
}



static void _emit_CALL_do (int n)
{
	int ext = GeneratingExtern;

	jit_insn *j1, *j2, *j3, *j4, *j5, *j6, *j7, *j8, *j9, *j10, *j11;

	_emit_reverse_args (n);

	     jit_ldxi_p    (R1, SP, n * sizeof (qu_any));

	/* R1 = x  (the thing to call) */

	     jit_andi_ul   (R0, R1, 1);
	j1 = jit_bnei_ul   (__later__, R0, 1);
	j2 = jit_get_label ();
	     jit__call0    (_throw_ECall);

	     jit_patch     (j1);
	     jit_ldxi_p    (R0, R1, offsetof (qu_object_s, klass));

	/* R0 = qu_classof (x) */

	j3 = jit_bnei_p    (__later__, R0, QuSubClass);

	     /* return x->call->fn (x, self, n, sp) */

	     jit_ldxi_p    (R0, R1, offsetof (qu_sub_s, call));
	     jit_ldxi_p    (R0, R0, offsetof (qu_subcall_s, fn));
	     jit_prepare   (4);
	     jit_pusharg_p (SP);
	     jit_movi_i    (R2, n);
	     jit_pusharg_i (R2);
	     jit__getself  (R2);
	     jit_pusharg_p (R2);
	     jit_pusharg_p (R1);
	     jit__callr    (R0);
	j4 = jit_jmpi      (__later__);


	     jit_patch     (j3);
	j5 = jit_bnei_p    (__later__, R0, QuMethodClass);

	     /* return qu_method_calln (x, n, sp) */

	     jit_prepare   (3);
	     jit_pusharg_p (SP);
	     jit_movi_i    (I0, n);
	     jit_pusharg_i (I0);
	     jit_pusharg_p (R1);
	     jit__call     (qu_method_calln);
	j6 = jit_jmpi      (__later__);

	     jit_patch     (j5);
	j7 = jit_bnei_p    (__later__, R0, QuStructClass);

		 /* return qu_record_new (x, n, sp) */
	     jit_prepare   (3);
	     jit_pusharg_p (SP);
	     jit_movi_i    (I0, n);
	     jit_pusharg_i (I0);
	     jit_pusharg_p (R1);
	     jit__call     (qu_record_new);
	j8 = jit_jmpi      (__later__);

	     jit_patch     (j7);
	j9 = jit_bnei_p    (__later__, R0, QuClassClass);

			/*
				qu_any b = c->get (x, QU_NAME_new, 0);
				QU_ENSURE (b);
				assert (qu_sub_is (b) && qu_sub_is_class (b));
				return b->call->fn (b, x, n, sp);
			*/

	      jit_pushr_p   (R1);

	      jit_prepare   (3);
	      jit_movi_i    (I0, 0);
	      jit_pusharg_i (I0);
	      jit_movi_p    (R2, QU_NAME_new);
	      jit_pusharg_p (R2);
	      jit_pusharg_p (R1);
	      jit_ldxi_p    (R0, R0, offsetof (qu_class_s, get));
	      jit__callr    (R0);
	      jit_movr_p    (R2, RET);

	      jit_popr_p    (R1);

	      jit_ldxi_p    (R0, R2, offsetof (qu_sub_s, call));
	      jit_ldxi_p    (R0, R0, offsetof (qu_subcall_s, fn));
	      jit_prepare   (4);
	      jit_pusharg_p (SP);
	      jit_movi_i    (I0, n);
	      jit_pusharg_i (I0);
	      jit_pusharg_p (R1);
	      jit_pusharg_p (R2);
	      jit__callr    (R0);
	j10 = jit_jmpi      (__later__);

	      jit_patch     (j9);
	      jit_bnei_p    (j2, R0, QuModuleClass);
	      jit_prepare   (3);
	      jit_pusharg_p (SP);
	      jit_movi_i    (I0, n);
	      jit_pusharg_i (I0);
	      jit__getself  (R0);
	      jit_pusharg_p (R0);
	      jit__call     (pCALL);

	      jit_patch     (j4);
	      jit_patch     (j6);
	      jit_patch     (j8);
	      jit_patch     (j10);
	      jit__down     (n + 1);
	      jit_pushr_p   (RET);
}

#endif /*OPTIMIZE_MORE*/



#ifdef OPTIMIZE_TAIL_CALL

static int pPrepareTailJit (qu_frame f, int n, qu_any *sp)
{
	qu_any   x = sp [n];
	qu_class c = qu_classof (x);

	REVERSE (n, sp);
	QU_ENSURE (qu_sub_arg_check (x, n, sp, 1));

	QU_OBJECT_LIST_SETVAL (f->ss, QuNil, ((qu_sub) x)->nlocal);
	QU_OBJECT_LIST_COPY   (f->ss + ((qu_sub) x)->first, sp, n);

	f->argc  = n;
	f->line  = 0;
	f->ip    = 0;
	f->level = f->atomic = -1;
	f->sp    = f->bp;
	f->sp[0] = QuNil;
	return 1;
}



#ifdef OPTIMIZE_MORE

static void _emit_set_locals (int n)
{
	jit_insn *j;

	if (Measuring)
		n = 1;
	else
		assert (TheSub->nlocal >= n);

	    jit_movr_p    (R1, LP);

	if (n)
	{
	    jit_movr_p    (R2, SP);
	    jit_movi_i    (I0, n);
	j = jit_get_label ();
	    jit_ldr_p     (R0, R2);
	    jit_str_p     (R1, R0);
	    jit_addi_p    (R1, R1, sizeof (qu_any));
	    jit_addi_p    (R2, R2, sizeof (qu_any));
	    jit_subi_i    (I0, I0, 1);
	    jit_bgti_i    (j, I0, 0);
	}

	n = TheSub->nlocal - n;
	if (Measuring)
		n = 1;

	if (n)
	{
	    jit_movi_i    (I0, n);
	    jit_movi_p    (R0, QuNil);
	j = jit_get_label ();
	    jit_str_p     (R1, R0);
	    jit_addi_p    (R1, R1, sizeof (qu_any));
	    jit_subi_i    (I0, I0, 1);
	    jit_bgti_i    (j, I0, 0);
	}
}


#else /*OPTIMIZE_MORE*/


static int pPrepareTailExtern (int nlocal, qu_any *ss, int n, qu_any *sp)
{
	qu_any   x = sp [n];
	qu_class c = qu_classof (x);

	REVERSE (n, sp);
	QU_ENSURE (qu_sub_arg_check (x, n, sp, 1));

	while (n--)
		*ss++ = *sp++;

	n = nlocal - n;
	while (n--)
		*ss++ = QuNil;

	return 1;
}


#endif /*OPTIMIZE_MORE*/

#endif /*OPTIMIZE_TAIL_CALL*/


static void _emit_CALL (int argc, int tail)
{

#ifdef OPTIMIZE_TAIL_CALL

	int extr = GeneratingExtern;

	SHOWME;

	if (Measuring)
	{
		tail = 1;
		extr = 1;
	}

	if (tail && TheSub->nlocal >= argc)
	{
		if (extr)
		{
			jit_insn *p;

		    jit_ldxi_p    (R0, SP, argc * sizeof (qu_any));
		    jit_ldxi_p    (R0, R0, offsetof (qu_sub_s, value));
		p = jit_bnei_p    (__later__, R0, StartOfCode);

#ifdef OPTIMIZE_MORE

			_emit_reverse_args (argc);
			if (Measuring || qu_sub_need_arg_check (TheSub) || TheSub->narg != argc)
			{
			    jit_prepare   (4);
			    jit_ldxi_p    (R0, SP, argc * sizeof (qu_any));
			    jit_pusharg_p (SP);
			    jit_movi_i    (I0, argc);
			    jit_pusharg_p (I0);
			    jit_pusharg_p (R0);
			    jit_movi_i    (I0, 1);
			    jit_pusharg_i (I0);
			    jit__call     (qu_sub_arg_check);
			}
			_emit_set_locals (argc);
#else
		    jit_prepare   (4);
		    jit_pusharg_p (SP);
		    jit_movi_i    (I0, argc);
		    jit_pusharg_p (I0);
		    jit_pusharg_p (LP);
		    jit_movi_i    (I0, TheSub->nlocal);
		    jit_pusharg_p (I0);
		    jit__call     (pPrepareTailExtern);

#endif /*OPTIMIZE_MORE*/

		    /* reset SP to the start of the locals */
		    jit_movr_p    (SP, LP);
		    jit_jmpi      (TailAddr);
		    jit_patch     (p);
		}
		else
		{
			jit_insn *p;

		    jit_ldxi_p    (R1, SP, argc * sizeof (qu_any));
		    jit_ldxi_p    (R0, FRM, offsetof (qu_frame_s, code));
		    jit_ldxi_p    (R0, R0, offsetof (qu_code_s, parent));
		p = jit_bner_p    (__later__, R0, R1);

		    jit_prepare   (3);
		    jit_pusharg_p (SP);
		    jit_movi_i    (I0, argc);
		    jit_pusharg_i (I0);
		    jit_pushr_p   (FRM);
		    jit__call     (pPrepareTailJit);

			/* restore SP */
		    jit_ldxi_p    (R0, FRM, offsetof (qu_frame_s, bp));
		    jit_movr_p    (SP, R0);

		    jit_jmpi      (TailAddr);
		    jit_patch     (p);
		}
	}

#endif /*OPTIMIZE_TAIL_CALL*/

#ifdef OPTIMIZE_MORE
	_emit_CALL_do (argc);
#else
	jit_prepare   (3);
	jit_pusharg_p (SP);
	jit_movi_i    (I0, argc);
	jit_pusharg_i (I0);
	jit__getself  (R0);
	jit_pusharg_p (R0);
	jit__call     (pCALL);
	jit__down     (argc + 1);
	jit_pushr_p   (RET);
#endif

}



static qu_any pCALLFLAT (qu_any self, int n, qu_any *sp)
{
	qu_any   x = sp [n];
	qu_class c = qu_classof (x);
	volatile qu_array g;

	REVERSE (n, sp);
	g = qu_array_flatten (n, sp);

  callagain:

	if (c == QuSubClass)
	{
		return qu_sub_calln (x, self, g->len, g->objs);
	}
	else if (c == QuMethodClass)
	{
		return qu_method_calln (x, g->len, g->objs);
	}
	else if (c == QuStructClass)
	{
		return qu_record_new (x, g->len, g->objs);
	}
	else if (c == QuClassClass)
	{
#if 0
		qu_any b = c->get (x, QU_NAME_new, 0);

		QU_ENSURE (b);
		assert (qu_sub_is (b) && qu_sub_is_class (b));

		return qu_sub_calln (b, x, g->len, g->objs);
#else
		assert (c->fnew);
		return qu_sub_calln (c->fnew, x, g->len, g->objs);
#endif
	}
	else if (c == QuModuleClass)
	{
		x = qu_module_getattr_callable_name (x, 1);
		if (x == 0)
			return 0;
		c = qu_classof (x);
		assert (qu_class_callable (x));
		goto callagain;
	}
	else
	{
		qu_any b = qu_classof (x)->get (x, QU_NAME__call, 0);

		if (b == 0)
			QU_THROW (QuECall);

		assert (qu_sub_is (b) && qu_sub_is_inst (b));

		return qu_sub_calln (b, x, g->len, g->objs);
	}
	QU_THROW (QuECall);
}



static void _emit_CALLFLAT (int argc)
{
	SHOWME;

	jit_prepare   (3);
	jit_pusharg_p (SP);
	jit_movi_i    (I0, argc);
	jit_pusharg_i (I0);
	jit__getself  (R0);
	jit_pusharg_p (R0);
	jit__call     (pCALLFLAT);
	jit__down     (argc + 1);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CALLATTR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
	on entry to pCALLATTR  after _emit_CALLATTR

	hi  o                  result  <- SP
	:   x
	:   arg-3
	:   arg-2
	lo  arg-1  <- SP
*/

#if 1
/*ifndef OPTIMIZE_MORE*/

static qu_any pCALLATTR (int n, qu_any *sp)
{
	qu_any   x = sp [n];
	qu_any   o = sp [n + 1];
	qu_class c = qu_classof (x);

  callattragain:

	if (c == QuSubClass)
	{
		if (qu_sub_is_class (x))
		{
			if (!qu_class_is (o))
				o = qu_classof (o);
		}
		else if (qu_sub_is_inst (x))
		{
			if (qu_class_is (o))
				QU_THROW (QuECall);
		}
		REVERSE (n, sp);
		return qu_sub_calln (x, o, n, sp);
	}
	else if (c == QuMethodClass)
	{
		REVERSE (n, sp);
		return qu_method_calln_attr (x, o, n, sp);
	}
	else if (c == QuClassClass)
	{
#if 0
		o = QuClassClass->get (x, QU_NAME_new, 0);
		if (o == 0)
			QU_RETHROW;
		assert (qu_sub_is (o) && qu_sub_is_class (o));
		REVERSE (n, sp);
		return qu_sub_calln (o, x, n, sp);
#else
		REVERSE (n, sp);
		assert (((qu_class) x)->fnew);
		return qu_sub_calln (((qu_class) x)->fnew, x, n, sp);
#endif
	}
	else if (c == QuStructClass)
	{
		x = qu_record_new (x, n, sp);
	}
	else if (c == QuModuleClass)
	{
		/* this is a consequence of "universe" */
		x = qu_module_getattr_callable_name (x, 1);
		if (x == 0)
			return 0;
		c = qu_classof (x);
		assert (qu_class_callable (c));
		goto callattragain;
	}
	else
	{
		/* allow calling property with no argument */
		if (n)
			QU_THROW (QuECall);
	}
	return x;
}

#endif /*OPTIMIZE_MORE*/



static void _emit_CALLATTR (int argc)
{

#if 0 /* FIXME - qu_method_calln is not implemented correctly here */
/*ifdef OPTIMIZE_MORE*/

	jit_insn *j1, *j2, *j3, *j4, *j5, *j6, *j7, *j8, *j9, *ja;
	jit_insn *jx, *jy;

	SHOWME;

	if (Measuring)
		argc == 2;

	if (argc > 1)
	     _emit_reverse_args (argc);

	     jit_ldxi_p  (R2, SP, argc * sizeof (qu_any));

	     /* R2 = x */

	     jit_andi_ul (I0, R2, 1);
	j1 = jit_bnei_ul (__later__, I0, 0);
	     jit_ldxi_p  (R0, R2, offsetof (qu_object_s, klass));
	j2 = jit_beqi_p  (__later__, R0, QuSubClass);

	jx = jit_bnei_p  (__later__, R0, QuMethodClass);

	if (argc > 1)
	{
	     /* R2 = x, R1 = o */
	     jit_ldxi_p  (R2, SP, argc * sizeof (qu_any));
	}
	     /* R2 = ((qu_method) x)->sub  R1 = ((qu_method) x)->self */
	     jit_ldxi_p  (R1, R2, offsetof (qu_method_s, self));
	     jit_ldxi_p  (R2, R2, offsetof (qu_method_s, sub));
	jy = jit_jmpi    (__later__);

	     /* allow calling property with no argument */
	     jit_patch   (jx);
         jit_patch   (j1);
         jit_movi_i  (I0, argc);
    j3 = jit_bnei_i  (__later__, I0, 0);
	     jit__call0  (_throw_ECall);
	     jit_patch   (j3);
	     jit__down   (argc + 2);
	     jit_pushr_p (R2);
	j4 = jit_jmpi    (__later__);

	     /* R2 = x, R1 = o */

	     jit_patch   (j2);
	     jit_ldxi_p  (R1, SP, (argc + 1) * sizeof (qu_any));

	     jit_andi_ul (I0, R1, 1);
	     jit_movi_p  (R0, QuIntClass);
	j5 = jit_beqi_ul (__later__, I0, 1);
		 jit_ldxi_p  (R0, R1, offsetof (qu_object_s, klass));
	     jit_patch   (j5);

	     /* R2 = x, R1 = o, R0 = qu_classof (o) */

		 jit_ldxi_ul (I0, R2, offsetof (qu_sub_s, flags));
	     jit_andi_ul (I0, I0, QU_SUB_FLAG_CLASS);
	j6 = jit_beqi_ul (__later__, I0, 0);
	j7 = jit_beqi_p  (__later__, R0, QuClassClass);
	     jit_movr_p  (R1, R0);
	j8 = jit_jmpi    (__later__);

	     jit_patch   (j6);
		 jit_ldxi_ul (I0, R2, offsetof (qu_sub_s, flags));
	     jit_andi_ul (I0, I0, QU_SUB_FLAG_INST);
	j9 = jit_beqi_ul (__later__, I0, 0);
	ja = jit_bnei_p  (__later__, R0, QuClassClass);
	     jit__call0  (_throw_ECall);

	     jit_patch   (ja);
	     jit_patch   (j9);
	     jit_patch   (j8);
	     jit_patch   (j7);

	if (argc > 1)
	{
	     jit_ldxi_p    (R2, SP, argc * sizeof (qu_any));
	     jit_ldxi_p    (R1, SP, (argc + 1) * sizeof (qu_any));
	}
	     /* R2 = x, R1 = o */

	     /* x->call->fn (x, o, argc, SP) */

	     jit_patch     (jy);

	     jit_ldxi_p    (R0, R2, offsetof (qu_sub_s, call));
	     jit_ldxi_p    (R0, R0, offsetof (qu_subcall_s, fn));
	     jit_prepare   (4);
	     jit_pusharg_p (SP);
	     jit_movi_i    (I0, argc);
	     jit_pusharg_p (I0);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R2);
	     jit__callr    (R0);
	     jit__down     (argc + 2);
	     jit_pushr_p   (RET);

	     jit_patch     (j4);
#else

	SHOWME;

    jit_prepare   (2);
	jit_pusharg_p (SP);
	jit_movi_i    (I0, argc);
	jit_pusharg_i (I0);
	jit__call     (pCALLATTR);
	jit__down     (argc + 2);
	jit_pushr_p   (RET);

#endif
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CALLATTRFLAT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pCALLATTRFLAT (int n, qu_any *sp)
{
	qu_any   x = sp [n];
	qu_any   o = sp [n + 1];
	qu_class c = qu_classof (x);
	volatile qu_array g;

  callattrflatagain:

	if (c == QuSubClass)
	{
		if (qu_sub_is_class (x))
		{
			if (!qu_class_is (o))
				o = qu_classof (o);
		}
		else if (qu_sub_is_inst (x))
		{
			if (qu_class_is (o))
				QU_THROW (QuECall);
		}

		REVERSE (n, sp);
		g = qu_array_flatten (n, sp);
		return qu_sub_calln (x, o, g->len, g->objs);
	}
	else if (c == QuMethodClass)
	{
		REVERSE (n, sp);
		g = qu_array_flatten (n, sp);
		return qu_method_calln_attr (x, o, g->len, g->objs);
	}
	else if (c == QuClassClass)
	{
#if 0
		o = QuClassClass->get (x, QU_NAME_new, 0);
		if (o == 0)
			QU_RETHROW;
		assert (qu_sub_is (o) && qu_sub_is_class (o));
		REVERSE (n, sp);
		g = qu_array_flatten (n, sp);
		return qu_sub_calln (o, x, n, sp);
#else
		assert (((qu_class) x)->fnew);
		REVERSE (n, sp);
		g = qu_array_flatten (n, sp);
		return qu_sub_calln (((qu_class) x)->fnew, x, n, sp);
#endif
	}
	else if (c == QuStructClass)
	{
		REVERSE (n, sp);
		g = qu_array_flatten (n, sp);
		return qu_record_new (x, g->len, g->objs);
	}
	else if (c == QuModuleClass)
	{
		/* this is a consequence of "universe" */
		x = qu_module_getattr_callable_name (x, 1);
		if (x == 0)
			return 0;
		c = qu_classof (x);
		assert (qu_class_callable (c));
		goto callattrflatagain;
	}
	else
	{
		/* allow calling property with no argument */
		if (n)
			QU_THROW (QuECall);
	}
	return x;
}



static void _emit_CALLATTRFLAT (int argc)
{
	SHOWME;

    jit_prepare   (2);
	jit_pusharg_p (SP);
	jit_movi_i    (I0, argc);
	jit_pusharg_i (I0);
	jit__call     (pCALLATTRFLAT);
	jit__down     (argc + 2);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CATCH
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* returns 1:catched 0:otherwise */
static int pCATCH (qu_frame F, qu_any e)
{
	qu_class c = F->eclass;

	assert (c);
	assert (qu_class_is (c));

	if (e == QuNil || (qu_class) e == c)
	{
	  catch:

		/* clear exception */
		qu_frame_unset_thrown (F);
		F->eclass = 0;
		return 1;
	}

	if (qu_array_is (e))
	{
		register long    n = ((qu_array) e)->len;
		register qu_any *p = ((qu_array) e)->objs;

		while (n--)
		{
			register qu_any x = *p++;

			assert (qu_class_is (x));
			if (qu_class_isa (c, x))
				goto catch;
		}
	}
	else
	{
		assert (qu_class_is (e));
		if (qu_class_isa (c, e))
			goto catch;
	}
	return 0;
}



static void _emit_CATCH (long nextip, qu_any things)
{
	jmp_s *j = _add_jmp (0, nextip);

	if (!Measuring)
		assert (!GeneratingExtern);

	SHOWME;

	         jit_prepare   (2);
	         jit_movi_p    (R0, things);
	         jit_pusharg_p (R0);
	         jit_pusharg_p (FRM);
	         jit__callsafe (pCATCH);
	j->ofs = jit_beqi_i    (__later__, RET, 0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CATCHSET
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_CATCHSET (qu_any v)
{
	if (Measuring)
		v = QuSys__ignore;

	jit_prepare   (0);
	jit__callsafe (qu__sys__err__);

	if (qu_int_is (v))
	{
		long i = qu_int_tol (v);
		jit__setlocal (i, RET);
	}
	else
	{
		assert (qu_var_is (v));

		jit_movi_p (R0, v);
	    jit_addi_p (R0, R0, offsetof (qu_var_s, value));
	    jit_str_p  (R0, RET);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CLOSURE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_CLOSURE (qu_sub x)
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_prepare   (2);
	jit_pusharg_p (FRM);
	jit_movi_p    (R0, x);
	jit_pusharg_p (R0);
	jit__callsafe (qu_sub_new_closure);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CMP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_CMP_do ()
{
	/* R0 <=> R1 */

#ifdef OPTIMIZE_INT
	jit_insn *j1, *j2, *j3, *j4, *j5, *j6;

	/* optimizing qu_int <-> qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_movi_p    (R2, QuInt1);
	j3 = jit_bgtr_l    (__later__, R0, R1);

	     jit_movi_p    (R2, QuInt_1);
	j4 = jit_bltr_l    (__later__, R0, R1);

	     jit_movi_p    (R2, QuInt0);

	     jit_patch     (j3);
	     jit_patch     (j4);
	     jit_pushr_p   (R2);
	j5 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_cmp);
	     jit_pushr_p   (RET);
	     jit_patch     (j5);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_cmp);
	     jit_pushr_p   (RET);
#endif
}


static void _emit_CMP ()
{
	SHOWME;

	jit_popr_p   (R1);
	jit_popr_p   (R0);
	_emit_CMP_do ();
}



static void _emit_CMPLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_CMP_do  ();
}



static void _emit_CMPLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_CMP_do  ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CONCAT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_CONCAT ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_concat);
	jit_pushr_p   (RET);
}



static void _emit_CONCATLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_concat);
	jit_pushr_p   (RET);
}



static void _emit_CONCATLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_concat);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	CONTINUE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifndef OPTIMIZE_MORE
static jit_insn *pCONTINUE (qu_frame F, int n)
{
	F->level -= (n - 1);
	if (F->level < F->atomic)
	{
	  __qu_atomic_end;
		F->atomic = -1;
	}
	F->sp = F->bp;
	return (jit_insn *) F->blocks[F->level].cont;
}
#endif


static void _emit_CONTINUE (int n)
{

#ifdef OPTIMIZE_MORE

	jit_insn *j;

	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	    jit_addi_p (R0, FRM, offsetof (qu_frame_s, level));
	    jit_ldr_i  (I0, R0);
	    jit_subi_i (I0, I0, n - 1);
	    jit_str_i  (R0, I0);
	    jit_addi_p (R2, FRM, offsetof (qu_frame_s, atomic));
	    jit_ldr_i  (R1, R2);
	j = jit_bgtr_i (__later__, I0, R1);
	    jit_movi_p (R0, &__QuThreadCurrent);
	    jit_ldr_p  (R0, R0);
	    jit_addi_p (R0, R0, offsetof (qu_thread_s, atomic));
	    jit_ldr_i  (R1, R0);
	    jit_subi_i (R1, R1, 1);
	    jit_str_i  (R0, R1);
	    jit_movi_i (R1, -1);
	    jit_str_i  (R2, R1);
	    jit_patch  (j);
	    jit_addi_p (R0, FRM, offsetof (qu_frame_s, blocks));
	    jit_muli_i (R1, I0, sizeof (qu_frblk_s));
	    jit_addr_p (R0, R0, R1);
	    jit_ldxi_p (R0, R0, offsetof (qu_frblk_s, cont));
	    jit_ldxi_p (R1, FRM, offsetof (qu_frame_s, bp));
	    jit_movr_p (SP, R1);
	    jit_jmpr   (R0);

#else

	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_prepare   (2);
	jit_movi_i    (I0, n);
	jit_pusharg_i (I0);
	jit_pusharg_p (FRM);
	jit__callsafe (pCONTINUE);
	jit_jmpr      (RET);

#endif

}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DECR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_DECR_do ()
{
	/* R0 += 1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4;

	/* optimizing qu_int only */

	     jit_andi_ul   (R1, R0, 1);
	j1 = jit_bnei_ul   (__later__, R1, 1);

	     jit_rshi_l    (R1, R0, 1);
	     jit_subi_l    (R1, R1, 1);
	j2 = jit_blti_l    (__later__, R1, QU_INT_MIN);

	     jit_lshi_l    (R1, R1, 1);
	     jit_ori_l     (R0, R1, 1);
	j3 = jit_jmpi      (__later__);

	     jit_patch     (j2);
	     jit_prepare   (1);
	     jit_pushr_p   (R1);
	     jit__callsafe (qu_long_newl);
	j4 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_decr);
	     jit_patch     (j4);
	     jit_movr_p    (R0, RET);
	     jit_patch     (j3);

#else
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_decr);
	     jit_movr_p    (R0, RET);
#endif
}



static void _emit_DECR ()
{
	SHOWME;

	jit_popr_p    (R0);
	_emit_DECR_do ();
	jit_pushr_p   (R0);
}



static void _emit_DECRLPOST (int i)
{
	SHOWME;

	jit__getlocal (R0, i);
	jit_pushr_p   (R0);
	_emit_DECR_do ();
	jit__setlocal (i, R0);
}



static void _emit_DECRLPRE (int i)
{
	SHOWME;

	jit__getlocal (R0, i);
	_emit_DECR_do ();
	jit_pushr_p   (R0);
	jit__setlocal (i, R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DICT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pDICT (int n, qu_any *sp)
{
	REVERSE (n, sp);
	return qu_dict_new_argv (n, sp);
}



static void _emit_DICT (int n)
{
	SHOWME;

	if (Measuring)
		n = 2;

	if (n)
	{
		jit_prepare    (2);
		jit_pusharg_p  (SP);
		jit_movi_i     (I0, n);
		jit_pusharg_i  (I0);
		jit__callsafe  (pDICT);
		jit__down      (n);
		jit_pushr_p    (RET);
	}
	else
	{
		jit_prepare    (1);
		jit_movi_i     (I0, 0);
		jit_pusharg_i  (I0);
		jit__callsafe  (qu_dict_new);
		jit_pushr_p    (RET);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DISCARD
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static int pGOTSIGNAL ()
{
	__QuInterpGotSignal = 0;

	if (__QuSigTERM)
		QU_THROW (QuEQuit);

	if (__QuSigINT)
	{
		qu_any u = QuSys__keyboard__->value;

		__QuSigINT = 0;
		if (u != QuNil)
		{
			assert (qu_sub_is (u));
			assert (qu_sub_is_static (u));
			assert (!qu_sub_is_extern (u));

			__qu_atomic_begin;
			u = qu_sub_call0 (u, QuNil);
			__qu_atomic_end;

			if (u == 0)
				QU_THROW (QuEInterrupt);
		}
		else
		{
			QU_THROW (QuEInterrupt);
		}
	}
	return 1;
}



static void pUPDATETICKER ()
{
	if (++__QuThreadCurrent->ticks >= QU_INTERP_THREAD_TICKS)
	{
		if (__QuThreadCurrent->atomic)
		{
			--__QuThreadCurrent->ticks;
		}
		else
		{
			__QuThreadCurrent->ticks = 0;
			__qu_blocking_begin;
			/* zzzz... */
			__qu_blocking_end;
		}
	}
}



static void _emit_DISCARD ()
{
	jit_insn *j1, *j2, *j3, *j4;

	SHOWME;

	     jit_ldi_i      (R0, &__QuInterpGotSignal);
	j1 = jit_beqi_i     (__later__, R0, 0);
	     jit_ldi_p      (R0, &__QuThreadCurrent);
	     jit_ldi_p      (R1, &__QuThreadMain);
	j2 = jit_bner_i     (__later__, R0, R1);
	     jit__call0     (pGOTSIGNAL);
	     jit_patch      (j1);
	     jit_patch      (j2);
	     jit_ldi_i      (R0, &__QuThreadMulti);
	j3 = jit_beqi_i     (__later__, R0, 0);
	     jit__callsafe0 (pUPDATETICKER);
	     jit_patch      (j3);
	     jit_ldi_p      (R0, &__QuInterpFinalizeList);
	     jit_ldxi_l     (R0, R0, offsetof (qu_array_s, len));
	j4 = jit_beqi_i     (__later__, R0, 0);
	     jit__callsafe0 (qu__interp_run_finalize);
	     jit_patch      (j4);
	     jit__down      (1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DIV
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_DIV_do ()
{
	/* R0 / R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3;

	/* optimizing qu_int only */

	     jit_andi_ul   (R2, R0, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R1, 1);
	j2 = jit_bnei_ul   (__later__, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_rshi_l    (R1, R1, 1);
	     jit_divr_l    (R0, R0, R1);
	     jit_lshi_l    (R0, R0, 1);
	     jit_ori_l     (R0, R0, 1);
	     jit_pushr_p   (R0);
	j3 = jit_jmpi      (__later__);
	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_div);
	     jit_pushr_p   (RET);
	     jit_patch     (j3);

#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_div);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_DIV ()
{
	SHOWME;

	jit_popr_p   (R1);
	jit_popr_p   (R0);
	_emit_DIV_do ();
}



static void _emit_DIVLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_DIV_do ();
}



static void _emit_DIVLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_DIV_do  ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DROP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_DROP1 ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	jit_pushr_p (R1);
}



static void _emit_DROP1LL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R0, i1);
	jit_pushr_p   (R0);
}



static void _emit_DROP1LX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p  (R0, x);
	jit_pushr_p (R0);
}



static void _emit_DROP2 ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit__down   (2);
	jit_pushr_p (R1);
}



static void _emit_DROP3 ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit__down   (3);
	jit_pushr_p (R1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	DUP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_DUP1 ()
{
	SHOWME;

	jit_popr_p  (R0);
	jit_pushr_p (R0);
	jit_pushr_p (R0);
}



static void _emit_DUP2 ()
{
	SHOWME;

	jit_popr_p  (R0);
	jit_popr_p  (R1);
	jit_pushr_p (R1);
	jit_pushr_p (R0);
	jit_pushr_p (R1);
	jit_pushr_p (R0);
}


static void _emit_DUP3 ()
{
	SHOWME;

	jit_popr_p  (R0);
	jit_popr_p  (R1);
	jit_popr_p  (R2);
	jit_pushr_p (R2);
	jit_pushr_p (R1);
	jit_pushr_p (R0);
	jit_pushr_p (R2);
	jit_pushr_p (R1);
	jit_pushr_p (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	EACH
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_EACH ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_each);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	END
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifndef OPTIMIZE_MORE

static qu_any pEND (qu_frame f, qu_any *sp)
{

#ifdef SHOW
	fprintf (stderr, "::: in pEND: f=%p  f->bp=%p  f->sp=%p  SP=%p\n", f,  f->bp, f->sp, sp);
#endif

	if (qu_frame_thrown (f))
	{
		f->result = 0;
	}
	else
	{
		if (qu_frame_is_generator (f))
		{
			qu_frame_throw (f, QuEStopClass);
			f->result = 0;
			goto end_done;
		}
		else
		{
			qu_symbol x = qu_frame_get_symbol (f);
			if (qu_sub_is (x))
			{
				if (((qu_sub) x)->type)
				{
					f->result = qu_object_validate (QuNil, ((qu_sub) x)->type, f->result);
					if (f->result == 0)
					{
						qu_frame_throw (f, QuEValueClass);
						goto end_done;
					}
				}
			}
			else
			{
				assert (qu_module_is (x));
			}
		}
	}

  end_done:

	qu_frame_unset_running (f);
	qu_frame_set_done (f);

#ifdef SHOW
	fprintf (stderr, "::: in pEND: result = %p\n", f->result);
#endif

	return f->result;
}

#endif



static void _emit_end_code ()
{

#ifdef OPTIMIZE_MORE

	jit_insn *j1, *j2, *j3, *j4, *j5, *j6, *j7;

	if (!Measuring)
		assert (!GeneratingExtern);

	     jit_ldxi_p    (I0, FRM, offsetof (qu_frame_s, flags));
	     jit_andi_ul   (R0, I0, QU_FRAME_FLAG_THROWN);
	j1 = jit_bnei_ul   (__later__, R0, 0);
	     jit_andi_ul   (R0, I0, QU_FRAME_FLAG_GENERATOR);
	j2 = jit_bnei_ul   (__later__, R0, 0);
	     jit_ldxi_p    (R0, FRM, offsetof (qu_frame_s, code));
	     jit_ldxi_p    (R0, R0, offsetof (qu_code_s, parent));
	     jit_ldxi_p    (R1, R0, offsetof (qu_object_s, klass));
	j3 = jit_bnei_p    (__later__, R1, QuSubClass);
	     jit_ldxi_p    (R1, R0, offsetof (qu_sub_s, type));
	j4 = jit_bnei_p    (__later__, R1, 0);
	     jit_patch     (j3);
	     jit_ldxi_p    (R0, FRM, offsetof (qu_frame_s, result));
	     jit_movr_p    (RET, R0);

	j5 = jit_get_label ();
	     jit_addi_p    (R2, FRM, offsetof (qu_frame_s, flags));
	     jit_ldr_ul    (I0, R2);
	     jit_andi_ul   (I0, I0, ~QU_FRAME_FLAG_RUNNING);
	     jit_ori_ul    (I0, I0, QU_FRAME_FLAG_DONE);
	     jit_str_ul    (R2, I0);
	     jit_ret       ();

	     jit_patch     (j4);
	     jit_prepare   (3);
	     jit_ldxi_p    (R0, FRM, offsetof (qu_frame_s, result));
	     jit_pusharg_p (R0);
	     jit_pusharg_p (R1);
	     jit_movi_p    (R0, QuNil);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_object_validate);
	j6 = jit_beqi_p    (__later__, RET, 0);
	     jit_jmpi      (j5);

	     jit_patch     (j6);
	     jit_movi_p    (R0, QuEValueClass);
	j7 = jit_get_label ();
	     jit_prepare   (2);
	     jit_pusharg_p (R0);
	     jit_pusharg_p (FRM);
	     jit__callsafe (qu_frame_throw);

	     jit_patch     (j2);
	     jit_movi_p    (R0, QuEStopClass);
	     jit_jmpi      (j7);

	     jit_patch     (j1);
	     jit_movi_p    (RET, 0);
	     jit_ret       ();
#else

	jit_prepare   (2);
	jit_pusharg_p (SP);
	jit_pusharg_p (FRM);
	jit__callsafe (pEND);
	jit_ret       ();

#endif

}



static void _emit_END ()
{
	int ext = GeneratingExtern;

	SHOWME;

	if (Measuring)
		ext = 1;

	if (!ext)
	{
		jit__end;
	}
	else
	{
		/* if we reach END, then we have not 'return' thus result is QuNil */
		/* restore SP */
		jit_movr_p (SP, BSP);
		jit_movi_p (RET, QuNil);
		jit_ret    ();
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ENTER
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static int pENTER (qu_frame F, int type, block_s *b, qu_any *sp)
{
	jitfrblk_s *blk = (jitfrblk_s *) &F->blocks [++F->level];

#ifdef SHOW
	fprintf (stderr, "::: ENTERING BLOCK type=%ld\n", type);
	fprintf (stderr, "::: level = %ld\n", F->level);
	fprintf (stderr, "::: F = %p  sp = %p\n", F, sp);
#endif

	blk->block = b;
	blk->obj   = 0; /* must do this for qu__frame_mark */
	blk->cont  = (long) b->contip;

	switch (type)
	{
		case QU_CODE_BLOCK_ATOMIC:
		{
			__qu_atomic_begin;
			F->atomic = F->level;
			return 1;
		}

		case QU_CODE_BLOCK_RANGE:
		{
			qu_any c = sp[0]; /* step */
			qu_any b = sp[1]; /* last */
			qu_any a = sp[2]; /* first */
#ifdef SHOW
	fprintf (stderr, "range %ld, %ld, %ld\n", qu_int_tol (a), qu_int_tol (b), qu_int_tol (c));
#endif
			if (!qu_int_is (a) || !qu_int_is (b) || !qu_int_is (c))
				QU_THROW (QuEValue);

			blk->obj = c;
			blk->ex1 = qu_int_tol (a);
			blk->ex2 = qu_int_tol (b);
			return 1;
		}

		case QU_CODE_BLOCK_FOR:
		{
			register qu_any a = sp[0];
			register qu_any c = qu_classof (a);

			if (qu_class_autoiter (c))
			{
				blk->obj = a;
				blk->ex1 = blk->ex2 = 0;
				return 1;
			}

			a = qu_isys_each (a);
			if (a == 0)
			{
				--F->level;
				return 0;
			}

			blk->obj = a;
			blk->ex1 = blk->ex2 = 0;
			return 1;
		}

		case QU_CODE_BLOCK_SWITCH:
		{
			blk->obj = sp[0];
			return 1;
		}

		case QU_CODE_BLOCK_ELSETRY:
		{
			/*
				!!! special case
				we get here from a JMP at the end of a try block without
				a LEAVE so we have to fake that LEAVE
			*/
			blk = (jitfrblk_s *) &F->blocks [--F->level];
			blk->block = b;
			blk->cont  = (long) b->contip;
			return 1;
		}
	}
	return 1;
}


static int _emit_ENTER (long nextip, int type)
{
	jit_insn *i, *j, *o;
	block_s  *b;
	int       nextblk = 0;

	SHOWME;

	if (!Measuring)
	{
		unsigned char *b = ((qu_code) TheSub->value)->bytes;

		if (b[nextip] == QU_PC_ENTER)
			nextblk = *(long *)(b + nextip + 1 + sizeof (long));

		assert (!GeneratingExtern);
	}

	i = jit_get_label ();
	j = jit_jmpi      (__later__);
	o = jit_get_label ();

	if (Measuring)
		BlockOfs = o - i;

	/* embed block info here */
	b = _add_block ((block_s *) o, nextip);
	b->nextblk = nextblk;
	MAKE_ROOM_FOR_BLOCK_STRUCT;

	    jit_patch      (j);
	    jit_prepare    (4);
	    jit_pusharg_p  (SP);
	    jit_movi_p     (R0, b->ofs);
	    jit_pusharg_p  (R0);
	    jit_movi_i     (I0, type);
	    jit_pusharg_i  (I0);
	    jit_pusharg_p  (FRM);
	    jit__call      (pENTER);

	switch (type)
	{
		case QU_CODE_BLOCK_RANGE:
			jit__down (3);
			break;

		case QU_CODE_BLOCK_FOR:
		case QU_CODE_BLOCK_SWITCH:
			jit__down (1);
			break;
	}

	b->contip = jit_get_label ();
	return 1;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	EQ
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_EQ_do ()
{
	/* R0 == R1 */

#ifdef OPTIMIZE_INT
	jit_insn *j1, *j2, *j3, *j4, *j5;

	j1 = jit_bner_p    (__later__, R0, R1);
	     jit_movi_p    (R2, QuTrue);
	     jit_pushr_p   (R2);
	j2 = jit_jmpi      (__later__);

	     jit_andi_ul   (R2, R1, 1);
	j3 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j4 = jit_bnei_ul   (__later__, R2, 1);
	     jit_movi_p    (R2, QuFalse);
	     jit_pushr_p   (R2);
	j5 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j3);
	     jit_patch     (j4);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_eq);
	     jit_pushr_p   (RET);
	     jit_patch     (j2);
	     jit_patch     (j5);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_eq);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_EQ ()
{
	SHOWME;

    jit_popr_p  (R1);
    jit_popr_p  (R0);
	_emit_EQ_do ();
}



static void _emit_EQLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_EQ_do   ();
}



static void _emit_EQLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_EQ_do   ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	EXTERN
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static int pEXTERN (qu_module m, qu_string name, qu_string spec)
{
	QU_FATAL ("EXTERN is not yet available");
}


static void _emit_EXTERN (qu_module m)
{
	SHOWME;

	jit_popr_p    (R1); /* spec */
	jit_popr_p    (R0); /* name */
	jit_prepare   (3);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit_movi_p    (R0, m);
	jit_pusharg_p (R0);
	jit__call     (pEXTERN);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	FORMAT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pFORMAT (long n, qu_any *sp)
{
	/*
		hi   format
		:    arg-3      n = 3
		:    arg-2
		lo   arg-1 <- SP
	*/

	REVERSE (n, sp);
	return qu_isys_format (sp[n], n, sp);
}


static void _emit_FORMAT (long n)
{
	SHOWME;

	jit_prepare   (2);
	jit_pusharg_p (SP);
	jit_movi_l    (I0, n);
	jit_pusharg_l (I0);
	jit__call     (pFORMAT);
	jit__down     (n + 1);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	FLAT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pFLAT (qu_any x)
{
	qu_var v;

	QU_ARG_TYPE (x, qu_array);
	v = qu_var_new_flattener ();
	v->value = x;
	return v;
}



static void _emit_FLAT ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (pFLAT);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	FRAME
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_FRAME ()
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_pushr_p (FRM);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GE_do ()
{
	/* R0 >= R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5;

	/* optimizing qu_int + qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_rshi_l    (R1, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_movi_p    (R2, QuTrue);
	j3 = jit_bger_l    (__later__, R0, R1);
	     jit_movi_p    (R2, QuFalse);
	     jit_patch     (j3);
	     jit_pushr_p   (R2);
	j4 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_ge);
	     jit_pushr_p   (RET);
	     jit_patch     (j4);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_ge);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_GE ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	_emit_GE_do ();
}



static void _emit_GELL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_GE_do   ();
}



static void _emit_GELX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_GE_do   ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GETATTR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pGETATTR (qu_any o, qu_any x, qu_frame f)
{
	return qu_isys_getattr (o, x, f);
}


static void _emit_GETATTR ()
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_pushr_p   (R0);
	jit_prepare   (3);
	jit_pusharg_p (FRM);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (pGETATTR);
	jit_pushr_p   (RET);
}



static void _emit_GETATTR2 ()
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_pushr_p   (R0);
	jit_pushr_p   (R1);
	jit_prepare   (3);
	jit_pusharg_p (FRM);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (pGETATTR);
	jit_pushr_p   (RET);
}



static void _emit_GETATTRX (qu_string name)
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_popr_p    (R0);
	jit_pushr_p   (R0);
	jit_prepare   (3);
	jit_pusharg_p (FRM);
	jit_movi_p    (R1, name);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (pGETATTR);
	jit_pushr_p   (RET);
}



static void _emit_GETATTRSELF (qu_string name)
{
	SHOWME;

	/* qu_classof (self)->get (self, name, 0) */
	/* self is never an qu_int */

	jit__getself  (R2);
	jit_pushr_p   (R2);
	jit_ldxi_p    (R0, R2, offsetof (qu_object_s, klass));
	jit_ldxi_p    (R1, R0, offsetof (qu_class_s, get));
	jit_prepare   (3);
	jit_movi_i    (I0, 0);
	jit_pushr_p   (I0);
	jit_movi_p    (R0, name);
	jit_pushr_p   (R0);
	jit_pushr_p   (R2);
	jit__callr    (R1);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GETITEM
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GETITEM_do ()
{
	/* R0 [R1] */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5, *j6;

	/* optimizing qu_array [qu_int] only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_ldxi_p    (R2, R0, offsetof (qu_object_s, klass));
	j2 = jit_bnei_p    (__later__, R2, QuArrayClass);

	     jit_movr_p    (I0, R1);
	     jit_ldxi_l    (R2, R0, offsetof (qu_array_s, len));
	     jit_rshi_l    (I0, I0, 1);
	j3 = jit_bgei_l    (__later__, I0, 0);
	     jit_addr_l    (I0, I0, R2);
	     jit_patch     (j3);
	j4 = jit_blti_l    (__later__, I0, 0);
	j5 = jit_bger_l    (__later__, I0, R2);
	     jit_ldxi_p    (R2, R0, offsetof (qu_array_s, objs));
	     jit_muli_l    (I0, I0, sizeof (qu_any));
	     jit_addr_p    (R2, R2, I0);
	     jit_ldr_p     (R0, R2);
	     jit_pushr_p   (R0);
	j6 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_patch     (j4);
	     jit_patch     (j5);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_getitem);
	     jit_pushr_p   (RET);
	     jit_patch     (j6);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_getitem);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_GETITEM ()
{
	SHOWME;

	jit_popr_p (R1);
	jit_popr_p (R0);
	_emit_GETITEM_do ();
}



static void _emit_GETITEM2 ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	jit_pushr_p (R0);
	jit_pushr_p (R1);
	_emit_GETITEM_do ();
}


static void _emit_GETITEMX (qu_any i)
{
	SHOWME;

	jit_popr_p (R0);
	jit_movi_p (R1, i);
	_emit_GETITEM_do ();
}



static void _emit_GETITEML (int i)
{
	SHOWME;

	jit_popr_p    (R1);
	jit__getlocal (R0, i);
	_emit_GETITEM_do ();
}



static void _emit_GETITEMLX (int i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_GETITEM_do ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GETGLOBAL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GETGLOBAL (qu_var v)
{
	SHOWME;

	jit_movi_p  (R0, v);
	jit_ldxi_p  (R0, R0, offsetof (qu_var_s, value));
	jit_pushr_p (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GETLEXICAL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GETLEXICAL (int i)
{
	SHOWME;

	jit__getlocal (R0, i);
	jit_pushr_p   (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GETSLICE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GETSLICE ()
{
	SHOWME;

	jit_popr_p    (R2);
	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (R2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_getslice);
	jit_pushr_p   (RET);
}


static void _emit_GETSLICE2 ()
{
	SHOWME;

	jit_popr_p    (R2);
	jit_popr_p    (R1);
	jit_popr_p    (R0);

	jit_pushr_p   (R0);
	jit_pushr_p   (R1);
	jit_pushr_p   (R2);

	jit_prepare   (3);
	jit_pusharg_p (R2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_getslice);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GETSWITCH
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GETSWITCH ()
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	/* F->blocks[F->level].obj */
	jit_addi_p  (R0, FRM, offsetof (qu_frame_s, blocks));
	jit_ldxi_i  (I0, FRM, offsetof (qu_frame_s, level));
	jit_muli_i  (I0, I0, sizeof (qu_frblk_s));
	jit_addr_p  (R0, R0, I0);
	jit_ldxi_p  (R0, R0, offsetof (qu_frblk_s, obj));
	jit_pushr_p (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	GT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_GT_do ()
{
	/* R0 > R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5;

	/* optimizing qu_int + qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_rshi_l    (R1, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_movi_p    (R2, QuTrue);
	j3 = jit_bgtr_l    (__later__, R0, R1);
	     jit_movi_p    (R2, QuFalse);
	     jit_patch     (j3);
	     jit_pushr_p   (R2);
	j4 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_gt);
	     jit_pushr_p   (RET);
	     jit_patch     (j4);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_gt);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_GT ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	_emit_GT_do ();
}



static void _emit_GTLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_GT_do   ();
}



static void _emit_GTLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_GT_do   ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	HTOPEN HTCLOSE HTENTITY
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_HTOPEN (qu_string s, long op)
{
	SHOWME;

	if (!Measuring)
		assert (__QuModuleHtml_open);

	jit_prepare   (2);
	jit_movi_l    (R0, op);
	jit_pusharg_p (R0);
	jit_movi_p    (R0, s);
	jit_pusharg_p (R0);
	jit__call     (__QuModuleHtml_open);
}



static void _emit_HTCLOSE (qu_string s, long op)
{
	SHOWME;

	if (!Measuring)
		assert (__QuModuleHtml_close);

	jit_prepare   (2);
	jit_movi_l    (R0, op);
	jit_pusharg_p (R0);
	jit_movi_p    (R0, s);
	jit_pusharg_p (R0);
	jit__call     (__QuModuleHtml_close);
}



static void _emit_HTENTITY (qu_string s)
{
	SHOWME;

	if (!Measuring)
		assert (__QuModuleHtml_entity);

	jit_prepare   (1);
	jit_movi_p    (R0, s);
	jit_pusharg_p (R0);
	jit__call     (__QuModuleHtml_entity);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	IAND
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_IAND ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_iand);
	jit_pushr_p   (RET);
}



static void _emit_IANDLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_iand);
	jit_pushr_p   (RET);
}



static void _emit_IANDLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_iand);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	IDIV
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_IDIV ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_idiv);
	jit_pushr_p   (RET);
}



static void _emit_IDIVLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_idiv);
	jit_pushr_p   (RET);
}



static void _emit_IDIVLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_idiv);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	IOR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_IOR ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_ior);
	jit_pushr_p   (RET);
}



static void _emit_IORLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_ior);
	jit_pushr_p   (RET);
}



static void _emit_IORLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_ior);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	IN
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_IN ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_in);
	jit_pushr_p   (RET);
}



static void _emit_INLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_in);
	jit_pushr_p   (RET);
}



static void _emit_INLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_in);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	INCR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_INCR_do ()
{
	/* R0 += 1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4;

	/* optimizing qu_int only */

	     jit_andi_ul   (R1, R0, 1);
	j1 = jit_bnei_ul   (__later__, R1, 1);

	     jit_rshi_l    (R1, R0, 1);
	     jit_addi_l    (R1, R1, 1);
	j2 = jit_bgti_l    (__later__, R1, QU_INT_MAX);

	     jit_lshi_l    (R1, R1, 1);
	     jit_ori_l     (R0, R1, 1);
	j3 = jit_jmpi      (__later__);

	     jit_patch     (j2);
	     jit_prepare   (1);
	     jit_pushr_p   (R1);
	     jit__callsafe (qu_long_newl);
	j4 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_incr);
	     jit_patch     (j4);
	     jit_movr_p    (R0, RET);
	     jit_patch     (j3);

#else
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_incr);
	     jit_movr_p    (R0, RET);
#endif
}



static void _emit_INCR ()
{
	SHOWME;

	jit_popr_p    (R0);
	_emit_INCR_do ();
	jit_pushr_p   (R0);
}



static void _emit_INCRLPOST (int i)
{
	SHOWME;

    jit__getlocal (R0, i);
    jit_pushr_p   (R0);
	_emit_INCR_do ();
	jit__setlocal (i, R0);
}


static void _emit_INCRLPRE (int i)
{
	SHOWME;

	jit__getlocal (R0, i);
	_emit_INCR_do ();
	jit_pushr_p   (R0);
	jit__setlocal (i, R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	INOT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_INOT ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_inot);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	INITMOD
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void pINITMOD ()
{
	QU_FATAL ("INITMOD should never be emitted in a Sub");
}



static void _emit_INITMOD (qu_module m)
{
	jit_calli (pINITMOD);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	IS
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_IS ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
    jit_prepare   (2);
    jit_pusharg_p (R1);
    jit_pusharg_p (R0);
    jit__callsafe (qu_sys_is);
    jit_pushr_p   (R0);
}



static void _emit_ISLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__callsafe (qu_sys_is);
	jit_pushr_p   (RET);
}



static void _emit_ISLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__callsafe (qu_sys_is);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ITERNEXT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* returns 0:error 1:okay 2:no more */
static int pITERNEXT (qu_frame F, qu_any thing)
{
	register qu_frblk_s *blk = &F->blocks [F->level];
	qu_any a = blk->obj;
	qu_any b = 0;

	switch (qu_classid (a))
	{

		case QU_CLASS_ID_Int:
		{
			register long v = blk->ex1;
			register long n = qu_int_tol (a);

			if (n >= 0)
			{
				if (v > blk->ex2)
					return 2;
			}
			else
			{
				if (v < blk->ex2)
					return 2;
			}
			a = qu_int_new (v);
			blk->ex1 += n;
			break;
		}

		case QU_CLASS_ID_String:
		{
			if (blk->ex1 >= ((qu_string) a)->len)
				return 2;
			a = qu_string_byte (((qu_string) a)->str[blk->ex1++]);
			break;
		}

		case QU_CLASS_ID_Buffer:
		{
			if (blk->ex1 >= ((qu_buffer) a)->len)
				return 2;
			a = qu_buffer_new (((qu_buffer) a)->str + blk->ex1++, 1);
			break;
		}

		case QU_CLASS_ID_Array:
		{
			if (blk->ex1 >= ((qu_array) a)->len)
				return 2;
			a = ((qu_array) a)->objs[blk->ex1++];
			break;
		}

		case QU_CLASS_ID_Dict:
		case QU_CLASS_ID_Arg:
		{
			qu_any k;

			b = qu_dict_walk ((qu_dict) a, &k, &blk->ex1, &blk->ex2);
			if (b == 0)
				return 2;
			a = k;
			break;
		}

		case QU_CLASS_ID_Set:
		{
			a = qu_set_walk ((qu_set) a, &blk->ex1, &blk->ex2);
			if (a == 0)
				return 2;
			break;
		}

		case QU_CLASS_ID_Range:
		{
			if (blk->ex1 >= ((qu_range) a)->count)
				return 2;
			a = qu_int_new (((qu_range) a)->first + ((qu_range) a)->step * blk->ex1++);
			break;
		}

		case QU_CLASS_ID_Frame:
		{
			assert (qu_frame_is_generator (a));

			if (qu_frame_done (a))
				return 2;

			if (qu_frame_running (a))
				QU_THROW (QuERuntime);

			assert (((qu_frame) a)->code);

			a = qu_interp_frame_next (a);

			if (a == 0)
			{
				if (qu_thread_err_is_estop ())
					return 2;
				QU_RETHROW;
			}
			break;
		}

		case QU_CLASS_ID_Record:
		{
			long n = blk->ex1;

			if (n >= ((qu_record) a)->len)
				return 2;

			blk->ex1++;
			b = ((qu_structentry_s *)((qu_record) a)->struc->items)[n].name;
			a = ((qu_record) a)->objs[n];
			break;
		}

		case QU_CLASS_ID_Struct:
		{
			long n = blk->ex1;

			if (n >= qu_struct_len ((qu_struct) a))
				return 2;

			blk->ex1++;
			b = ((qu_structentry_s *)((qu_struct) a)->items)[n].type;
			a = ((qu_structentry_s *)((qu_struct) a)->items)[n].name;
			break;
		}

		default:
		{
			assert (qu_iterator_is (a));

			a = qu_iterator_next (a);
			if (a == 0)
			{
				if (qu_thread_err_is_estop ())
					return 2;
				QU_RETHROW;
			}
			break;
		}

	}

	/* a = the next value, b = the next key in case of qu_dict/qu_record

		Things to set is either a:
		- qu_int -> lexical, the value is the index
		- qu_var -> global variable
		- qu_array of qu_int/qu_var objects
	*/

#define set_iter_item_(v, val) \
QU_BEGIN \
	if (qu_int_is (v)) \
	{ \
		int j = qu_int_tol (v); \
		qu_sub u = (qu_sub) qu_frame_get_symbol (F); \
		assert (qu_sub_is (u)); \
		if (u->ltype && u->ltype[j]) \
		{ \
			val = qu_object_validate (QuNil, u->ltype[j], val); \
			if (val == 0) \
				return 0; \
		} \
		F->ss [j] = val; \
	} \
	else \
	{ \
		assert (qu_var_is (v)); \
		if (((qu_var) v)->type) \
		{ \
			val = qu_var_set_value (v, val); \
			if (val == 0) \
				return 0; \
		} \
		else \
		{ \
			((qu_var) v)->value = val; \
		} \
	} \
QU_END

	if (qu_array_is (thing))
	{
		long    len  = ((qu_array) thing)->len;
		int     eat  = qu_array_forlast (thing);
		qu_any *vars = ((qu_array) thing)->objs;
		qu_any  what = blk->obj;
		qu_any  var;

		assert (len > 0);

		if (qu_dict_is (what) || qu_arg_is (what) || qu_record_is (what) || qu_struct_is (what))
		{
			/* dang! */
			if (len == 2 && qu_record_is (what))
			{
				register qu_any x = a;
				a = b;
				b = x;
			}

			/* for a, b, c in [foo:1, bar:2] is wrong */
			if (len > 2 && !eat)
				QU_THROW (QuENone);

			var = *vars++;
			set_iter_item_(var, a);
			if (len == 2)
			{
				var = *vars++;
				if (eat)
					b = qu_array_new_argv (1, &b);
				set_iter_item_(var, b);
			}
		}
		else if (qu_array_is (a))
		{
			long i;

			for (i = 0; i < len; i++)
			{
				if (eat && i == len - 1)
				{
					b = qu_array_getslice (a, i, len);
					if (b == 0)
						b = qu_array_new (0);
				}
				else
				{
					if (0 <= i && i < ((qu_array) a)->len)
						b = qu_array_GETITEM ((qu_array) a, i);
					else
						b = QuNil;
				}
				var = *vars++;
				set_iter_item_(var, b);
			}
		}
		else
		{
			QU_THROW (QuENone);
		}
	}
	else if (thing != QuSys__ignore)
	{
		set_iter_item_(thing, a);
	}
	return 1;
}



static void _emit_ITERNEXT (qu_any thing)
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_prepare   (2);
	jit_movi_p    (R1, thing);
	jit_pusharg_p (R1);
	jit_pusharg_p (FRM);
	jit__call     (pITERNEXT);
	jit_bnei_i    (LeaveAddr, RET, 1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	IXOR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_IXOR ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_ixor);
	jit_pushr_p   (RET);
}



static void _emit_IXORLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_ixor);
	jit_pushr_p   (RET);
}



static void _emit_IXORLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_ixor);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	JMP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_JMP (int ip)
{
	jmp_s *j = _add_jmp (0, ip);

	SHOWME;

	j->ofs = jit_jmpi (__later__);
}



static void _emit_JMPFALSE_do (int ip, int ex)
{

#ifdef OPTIMIZE_MORE
	jit_insn *l1, *l2, *l3, *l4, *l5;
	jmp_s    *j = _add_jmp (0, ip);

	SHOWME;

	         jit_popr_p    (R0);
	if (ex)
	         jit_pushr_p   (R0);

	l1     = jit_beqi_p    (__later__, R0, QuTrue);
	l2     = jit_beqi_p    (__later__, R0, QuFalse);
	l3     = jit_beqi_p    (__later__, R0, QuInt0);
	         jit_andi_ul   (R2, R0, 1);
	l4     = jit_beqi_ul   (__later__, R0, 1);
	         jit_prepare   (1);
             jit_pusharg_p (R0);
	         jit__callsafe (qu_sys_true);
	l5     = jit_bnei_p    (__later__, RET, QuFalse);
	         jit_patch     (l2);
	         jit_patch     (l3);
	j->ofs = jit_jmpi      (__later__);
	         jit_patch     (l1);
	         jit_patch     (l4);
	         jit_patch     (l5);
#else
	jit_insn *l;
	jmp_s    *j = _add_jmp (0, ip);

	SHOWME;

	         jit_popr_p    (R0);
	if (ex)
	         jit_pushr_p   (R0);

	         jit_prepare   (1);
             jit_pusharg_p (R0);
	         jit__callsafe (qu_sys_true);
	l      = jit_beqi_p    (__later__, RET, QuTrue);
	j->ofs = jit_jmpi      (__later__);
	         jit_patch     (l);
#endif
}



static void _emit_JMPFALSE (int ip)
{
	_emit_JMPFALSE_do (ip, 0);
}



static void _emit_JMPFALSEX (int ip)
{
	_emit_JMPFALSE_do (ip, 1);
}



static void _emit_JMPTRUE_do (int ip, int ex)
{

#ifdef OPTIMIZE_MORE
	jit_insn *l1, *l2, *l3, *l4, *l5;
	jmp_s    *j = _add_jmp (0, ip);

	SHOWME;

	         jit_popr_p    (R0);
	if (ex)
	         jit_pushr_p   (R0);

	l1     = jit_beqi_p    (__later__, R0, QuFalse);
	l2     = jit_beqi_p    (__later__, R0, QuTrue);
	l3     = jit_beqi_p    (__later__, R0, QuInt0);
	         jit_andi_ul   (R2, R0, 1);
	l4     = jit_beqi_ul   (__later__, R0, 1);
	         jit_prepare   (1);
             jit_pusharg_p (R0);
	         jit__callsafe (qu_sys_true);
	l5     = jit_bnei_p    (__later__, RET, QuTrue);
	         jit_patch     (l2);
	         jit_patch     (l4);
	j->ofs = jit_jmpi      (__later__);
	         jit_patch     (l1);
	         jit_patch     (l3);
	         jit_patch     (l5);
#else
	jit_insn *l;
	jmp_s    *j = _add_jmp (0, ip);

	SHOWME;

	         jit_popr_p    (R0);
	if (ex)
	         jit_pushr_p   (R0);

	         jit_prepare   (1);
             jit_pusharg_p (R0);
	         jit__callsafe (qu_sys_true);
	l      = jit_bnei_p    (__later__, RET, QuTrue);
	j->ofs = jit_jmpi      (__later__);
	         jit_patch     (l);
#endif
}



static void _emit_JMPTRUE (int ip)
{
	_emit_JMPTRUE_do (ip, 0);
}



static void _emit_JMPTRUEX (int ip)
{
	_emit_JMPTRUE_do (ip, 1);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_LE_do ()
{
	/* R0 <= R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5;

	/* optimizing qu_int + qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_rshi_l    (R1, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_movi_p    (R2, QuTrue);
	j3 = jit_bler_l    (__later__, R0, R1);
	     jit_movi_p    (R2, QuFalse);
	     jit_patch     (j3);
	     jit_pushr_p   (R2);
	j4 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_le);
	     jit_pushr_p   (RET);
	     jit_patch     (j4);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_le);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_LE ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	_emit_LE_do ();
}



static void _emit_LELL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_LE_do   ();
}



static void _emit_LELX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_LE_do   ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LEAVE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static jit_insn *pLEAVE (qu_frame F)
{
	while (F->level >= 0)
	{
		jitfrblk_s *blk;
		block_s    *b;

		if (F->level == F->atomic)
		{
			__qu_atomic_end;
			F->atomic = -1;
		}

		blk = (jitfrblk_s *) &F->blocks[F->level--];
		b = blk->block;

		while (b->nextblk)
		{
			switch (b->nextblk)
			{
				case QU_CODE_BLOCK_FINALLY:
					return b->nextip;

				case QU_CODE_BLOCK_CATCH:
					if (qu_frame_thrown (F))
					{
						assert (F->eclass);
						if (qu_class_catchable (F->eclass))
							return b->nextip;
					}
					b = (block_s *) ((char *) b->nextip + BlockOfs);
					continue;
			}
			break;
		}

		if (! (qu_frame_thrown (F) || qu_frame_returning (F)))
			return b->nextip;
	}
	return 0;
}



static void _emit_leave_code ()
{
	jit_insn *j;

	if (!Measuring)
		assert (!GeneratingExtern);

		/* restore SP */
		jit_ldxi_p     (R0, FRM, offsetof (qu_frame_s, bp));
		jit_movr_p     (SP, R0);

	    jit_prepare    (1);
	    jit_pusharg_p  (FRM);
	    jit__callsafe  (pLEAVE);
	j = jit_beqi_p     (__later__, RET, 0);
	    jit_jmpr       (RET);
	    jit_patch      (j);
	     /* end code follows */
}



static void _emit_LEAVE ()
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit__leave;
}





/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LIKE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_LIKE ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
    jit_prepare   (2);
    jit_pusharg_p (R1);
    jit_pusharg_p (R0);
    jit__callsafe (qu_sys_like);
    jit_pushr_p   (R0);
}



static void _emit_LIKELL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__callsafe (qu_sys_like);
	jit_pushr_p   (RET);
}



static void _emit_LIKELX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__callsafe (qu_sys_like);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LINE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_LINE (long lineno)
{
	int ext = GeneratingExtern;

	SHOWME;

	if (Measuring)
		ext = 0;

	if (!ext)
	{
		jit_addi_p (R0, FRM, offsetof (qu_frame_s, line));
		jit_movi_l (I0, lineno);
		jit_str_l  (R0, I0);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LITERAL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_LITERAL (qu_any x)
{
	SHOWME;

	jit_movi_p  (R0, x);
	jit_pushr_p (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LSHIFT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_LSHIFT ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_lshift);
	jit_pushr_p   (RET);
}



static void _emit_LSHIFTLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_lshift);
	jit_pushr_p   (RET);
}



static void _emit_LSHIFTLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_lshift);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	LT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_LT_do ()
{
	/* R0 < R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5;

	/* optimizing qu_int + qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_rshi_l    (R1, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_movi_p    (R2, QuTrue);
	j3 = jit_bltr_l    (__later__, R0, R1);
	     jit_movi_p    (R2, QuFalse);
	     jit_patch     (j3);
	     jit_pushr_p   (R2);
	j4 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_lt);
	     jit_pushr_p   (RET);
	     jit_patch     (j4);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_lt);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_LT ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	_emit_LT_do ();
}



static void _emit_LTLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_LT_do ();
}



static void _emit_LTLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_LT_do   ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	MINUS
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_MINUS_do ()
{
	/* R0 - R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5, *j6;

	/* optimizing qu_int + qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_rshi_l    (R1, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_subr_l    (R2, R0, R1);
	j3 = jit_bgti_l    (__later__, R2, QU_INT_MAX);
	j4 = jit_blti_l    (__later__, R2, QU_INT_MIN);
	     jit_lshi_l    (R2, R2, 1);
	     jit_ori_l     (R2, R2, 1);
	     jit_pushr_p   (R2);
	j5 = jit_jmpi      (__later__);

	     jit_patch     (j3);
	     jit_patch     (j4);
	     jit_lshi_l    (R1, R1, 1);
	     jit_lshi_l    (R0, R0, 1);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_int_minus_int);
	j6 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_minus);
	     jit_patch     (j6);
	     jit_pushr_p   (RET);
	     jit_patch     (j5);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_minus);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_MINUS ()
{
	SHOWME;

	jit_popr_p     (R1);
	jit_popr_p     (R0);
	_emit_MINUS_do ();
}



static void _emit_MINUSLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal  (R1, i1);
	jit__getlocal  (R0, i0);
	_emit_MINUS_do ();
}



static void _emit_MINUSLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p     (R1, x);
	jit__getlocal  (R0, i);
	_emit_MINUS_do ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	MUL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_MUL ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_mul);
	jit_pushr_p   (RET);
}



static void _emit_MULLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_mul);
	jit_pushr_p   (RET);
}



static void _emit_MULLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_mul);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	NE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_NE_do ()
{
	/* R0 != R1 */

	jit_insn *j1, *j2;

	j1 = jit_beqr_p    (__later__, R0, R1);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_ne);
	     jit_pushr_p   (RET);
	j2 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_movi_p    (R2, QuFalse);
	     jit_pushr_p   (R2);
	     jit_patch     (j2);
}



static void _emit_NE ()
{
	SHOWME;

	jit_popr_p  (R1);
	jit_popr_p  (R0);
	_emit_NE_do ();
}



static void _emit_NELL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R1, i1);
	jit__getlocal (R0, i0);
	_emit_NE_do   ();
}



static void _emit_NELX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_NE_do   ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	NEG
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_NEG ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_neg);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	NOP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_NOP ()
{
	/* do nothing */
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	NOT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_NOT ()
{
#ifdef OPTIMIZE_MORE
	jit_insn *j1, *j2, *j3, *j4, *j5, *j6, *j7;

	SHOWME;

	     jit_popr_p    (R0);
	j1 = jit_beqi_p    (__later__, R0, QuTrue);
	j2 = jit_beqi_p    (__later__, R0, QuFalse);
	j3 = jit_beqi_p    (__later__, R0, QuInt0);
	     jit_andi_ul   (R1, R0, 1);
	j4 = jit_beqi_ul   (__later__, R1, 1);
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_not);
	j5 = jit_jmpi      (__later__);
	     jit_patch     (j2);
	     jit_patch     (j3);
	     jit_movi_p    (RET, QuTrue);
	j6 = jit_jmpi      (__later__);
	j7 = jit_get_label ();
	     jit_patch     (j1);
	     jit_patch     (j4);
	     jit_movi_p    (RET, QuFalse);
         jit_patch     (j5);
         jit_patch     (j6);
	     jit_pushr_p   (RET);
#else
	SHOWME;

	     jit_popr_p    (R0);
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_not);
	     jit_pushr_p   (RET);
#endif
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	PARAM
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_PARAM (long ip, long delta)
{
	jmp_s *jp = _add_jmp (0, ip);

	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	          jit_ldxi_i (R0, FRM, offsetof (qu_frame_s, argc));
	jp->ofs = jit_bgti_i (__later__, R0, delta);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	PASS
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_PASS ()
{
	SHOWME;

	jit__callsafe0 (qu_interp_unlock);
	/* zzzz ... */
	jit__callsafe0 (qu_interp_lock);
	jit_ldi_p      (R0, &__QuThreadCurrent);
	jit_addi_p     (R0, R0, offsetof (qu_thread_s, ticks));
	jit_movi_ul    (R1, 0);
	jit_str_p      (R0, R1);
	jit_movi_p     (R0, QuNil);
	jit_pushr_p    (R0);
}




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	PLUS
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_PLUS_do ()
{
	/* R0 + R1 */

#ifdef OPTIMIZE_INT

	jit_insn *j1, *j2, *j3, *j4, *j5, *j6;

	/* optimizing qu_int + qu_int only */

	     jit_andi_ul   (R2, R1, 1);
	j1 = jit_bnei_ul   (__later__, R2, 1);
	     jit_andi_ul   (R2, R0, 1);
	j2 = jit_bnei_ul   (__later__, R2, 1);

	     jit_rshi_l    (R1, R1, 1);
	     jit_rshi_l    (R0, R0, 1);
	     jit_addr_l    (R2, R0, R1);
	j3 = jit_bgti_l    (__later__, R2, QU_INT_MAX);
	j4 = jit_blti_l    (__later__, R2, QU_INT_MIN);
	     jit_lshi_l    (R2, R2, 1);
	     jit_ori_l     (R2, R2, 1);
	     jit_pushr_p   (R2);
	j5 = jit_jmpi      (__later__);

	     jit_patch     (j3);
	     jit_patch     (j4);
	     jit_lshi_l    (R1, R1, 1);
	     jit_lshi_l    (R0, R0, 1);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_int_plus_int);
	j6 = jit_jmpi      (__later__);

	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_plus);
	     jit_patch     (j6);
	     jit_pushr_p   (RET);
	     jit_patch     (j5);
#else
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_sys_plus);
	     jit_pushr_p   (RET);
#endif
}



static void _emit_PLUS ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	_emit_PLUS_do ();
}



static void _emit_PLUSLL (int i0, int i1)
{
	SHOWME;

     jit__getlocal (R1, i1);
     jit__getlocal (R0, i0);
	_emit_PLUS_do ();
}



static void _emit_PLUSLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R1, x);
	jit__getlocal (R0, i);
	_emit_PLUS_do ();
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	POP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_POP ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_pop);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	POS
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_POS ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_pos);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	POW
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_POW ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_pow);
	jit_pushr_p   (RET);
}



static void _emit_POWLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_pow);
	jit_pushr_p   (RET);
}



static void _emit_POWLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_pow);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	PRINT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_PRINT (qu_any x)
{
	SHOWME;

	jit_prepare   (1);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__call     (qu_sys_print);
	/* !!! must not change the stack */
}


static void _emit_PRINTTOP ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_print);
	/* !!! must not change the stack */
}


static void _emit_PRINTHTML ()
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_printh);
	/* !!! must not change the stack */
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	PUSH
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_PUSH ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_push);
	jit_pushr_p   (RET);
}



static void _emit_PUSHLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_push);
	jit_pushr_p   (RET);
}



static void _emit_PUSHLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_push);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	RANGE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_RANGE ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
    jit__call     (qu_sys_range);
	jit_pushr_p   (RET);
}




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	REM
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_REM ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rem);
	jit_pushr_p   (RET);
}



static void _emit_REMLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rem);
	jit_pushr_p   (RET);
}



static void _emit_REMLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rem);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	RET
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_RET ()
{
	int ext = GeneratingExtern;

	SHOWME;

	if (Measuring)
		ext = 0;

	if (!ext)
	{
		jit_addi_p  (R0, FRM, offsetof (qu_frame_s, flags));
		jit_ldr_ul  (R1, R0);
		jit_ori_ul  (R1, R1, QU_FRAME_FLAG_RETURNING);
		jit_str_ul  (R0, R1);
		jit_addi_p  (R0, FRM, offsetof (qu_frame_s, result));
		jit_popr_p  (R1);
		jit_str_p   (R0, R1);
		jit__leave;
	}
	else
	{
		SHOWREG ("<<<<< EXTERN ret  SP", SP);
		SHOWREG ("<<<<< EXTERN ret BSP", BSP);

		jit_popr_p (RET);
		jit_movr_p (SP, BSP);
		jit_ret    ();
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	RIN
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_RIN ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rin);
	jit_pushr_p   (RET);
}



static void _emit_RINLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rin);
	jit_pushr_p   (RET);
}



static void _emit_RINLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rin);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	ROT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_ROT3 ()
{
	SHOWME;

	jit_popr_p  (R2);
	jit_popr_p  (R1);
	jit_popr_p  (R0);

	jit_pushr_p (R2);
	jit_pushr_p (R0);
	jit_pushr_p (R1);
}




static void _emit_ROT4 ()
{
	SHOWME;

	jit_popr_p  (I0);
	jit_popr_p  (R2);
	jit_popr_p  (R1);
	jit_popr_p  (R0);

	jit_pushr_p (R1);
	jit_pushr_p (R2);
	jit_pushr_p (I0);
	jit_pushr_p (R0);
}



static void _emit_ROT4X ()
{
	SHOWME;

	jit_popr_p  (I0);
	jit_popr_p  (R2);
	jit_popr_p  (R1);
	jit_popr_p  (R0);

	jit_pushr_p (I0);
	jit_pushr_p (R0);
	jit_pushr_p (R1);
	jit_pushr_p (R2);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	RSHIFT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_RSHIFT ()
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rshift);
	jit_pushr_p   (RET);
}



static void _emit_RSHIFTLL (long i0, long i1)
{
	SHOWME;

	jit_prepare   (2);
	jit__getlocal (R1, i1);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i0);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rshift);
	jit_pushr_p   (RET);
}



static void _emit_RSHIFTLX (long i, qu_any x)
{
	SHOWME;

	jit_prepare   (2);
	jit_movi_p    (R1, x);
	jit_pusharg_p (R1);
	jit__getlocal (R0, i);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_rshift);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SELF
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_SELF ()
{
	SHOWME;

	jit__getself (R0);
	jit_pushr_p  (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SET
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_SET (int n)
{
	SHOWME;

	if (Measuring)
		n = 1;

	if (n)
	{
		/* the order does not matter */

		jit_prepare    (3);
		jit_pusharg_p  (SP);
		jit_movi_p     (R0, qu_int_new (n));
		jit_pusharg_p  (R0);
		jit_movi_p     (R0, QuSetClass);
		jit_pusharg_p  (R0);
		jit__callsafe  (qu__set_new);
		jit__down      (n);
		jit_pushr_p    (RET);
	}
	else
	{
		jit_prepare    (1);
		jit_movi_i     (I0, n);
		jit_pusharg_i  (I0);
		jit__callsafe  (qu_set_new);
		jit_pushr_p    (RET);
	}
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SETATTR
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pSETATTR (qu_any obj, qu_any attr, qu_any val, qu_frame f)
{
	return qu_isys_setattr (obj, attr, val, f);
}



static void _emit_SETATTR ()
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_popr_p     (R2);
	jit_popr_p     (R1);
	jit_popr_p     (I0);
	jit_prepare    (4);
	jit_pusharg_p  (FRM);
	jit_pusharg_p  (R2);
	jit_pusharg_p  (R1);
	jit_pusharg_p  (I0);
	jit__call      (pSETATTR);
	jit_pushr_p    (RET);
}



static qu_any pSETATTRX (qu_any obj, qu_string name, qu_any val, qu_frame f)
{
	return qu_isys_setattrx (obj, name, val, f);
}



static void _emit_SETATTRX (qu_string name)
{
	SHOWME;

	if (!Measuring)
		assert (!GeneratingExtern);

	jit_popr_p    (R2);
	jit_popr_p    (R1);
	jit_prepare   (4);
	jit_pusharg_p (FRM);
	jit_pusharg_p (R2);
	jit_movi_p    (R0, name);
	jit_pusharg_p (R0);
	jit_pusharg_p (R1);
	jit__call     (pSETATTRX);
	jit_pushr_p   (RET);
}



static void _emit_SETATTRSELF (qu_string name)
{
	SHOWME;

	/* qu_classof(self)->set (self, name, val, 0) */
	/* self is never an qu_int */

	jit__getself  (R2);
	jit_ldxi_p    (R0, R2, offsetof (qu_object_s, klass));
	jit_ldxi_p    (R1, R0, offsetof (qu_class_s, set));
	jit_popr_p    (R0);
	jit_prepare   (4);
	jit_movi_i    (I0, 0);
	jit_pushr_p   (I0);
	jit_pushr_p   (R0);
	jit_movi_p    (R0, name);
	jit_pushr_p   (R0);
	jit_pushr_p   (R2);
	jit__callr    (R1);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SETGLOBAL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_SETGLOBAL (qu_var v)
{
#if 0
	SHOWME;

	jit_popr_p  (R1);
	jit_movi_p  (R0, v);
	jit_addi_p  (R0, R0, offsetof (qu_var_s, value));
	jit_str_p   (R0, R1);
	jit_pushr_p (R1);
#else

	/* see interp.c SETGLOBAL for why we are doing this */

	jit_insn *j1, *j2;

	SHOWME;

	     jit_popr_p    (R1);
	     jit_movi_p    (R0, v);
	     jit_ldxi_p    (R2, R0, offsetof (qu_var_s, type));
	j1 = jit_beqi_p    (__later__, R2, 0);
	     jit_prepare   (2);
	     jit_pusharg_p (R1);
	     jit_pusharg_p (R0);
	     jit__call     (qu_var_set_value);
	     jit_pushr_p   (RET);
	j2 = jit_jmpi      (__later__);
	     jit_patch     (j1);
	     jit_addi_p    (R0, R0, offsetof (qu_var_s, value));
         jit_str_p     (R0, R1);
         jit_pushr_p   (R1);
	     jit_patch     (j2);
#endif
}



static void _emit_SETGLOBALX (qu_var v)
{
	SHOWME;

	jit_popr_p    (R1);
	jit_prepare   (2);
	jit_pusharg_p (R1);
	jit_movi_p    (R0, v);
	jit_pusharg_p (R0);
	jit__call     (qu_var_set_value);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SETITEM
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_SETITEM ()
{
	SHOWME;

	jit_popr_p    (R2);
	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (R2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
    jit__call     (qu_sys_setitem);
	jit_pushr_p   (RET);
}



static void _emit_SETITEMX (qu_any i)
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (R1);
	jit_movi_p    (R2, i);
	jit_pusharg_p (R2);
	jit_pusharg_p (R0);
    jit__call     (qu_sys_setitem);
	jit_pushr_p   (RET);
}



static void _emit_SETITEML (int idx)
{
	SHOWME;

	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__getlocal (R0, idx);
	jit_pusharg_p (R0);
    jit__call     (qu_sys_setitem);
	jit_pushr_p   (RET);
}



static void _emit_SETITEMLX (int idx, qu_any i)
{
	SHOWME;

	jit_popr_p    (R1);
	jit_prepare   (3);
	jit_pusharg_p (R1);
	jit_movi_p    (R0, i);
	jit_pusharg_p (R0);
	jit__getlocal (R0, idx);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_setitem);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SETLEXICAL
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_SETLEXICAL (int i)
{
	SHOWME;

	jit_popr_p    (R0);
	jit__setlocal (i, R0);
	jit_pushr_p   (R0);
}



static void _emit_SETLEXICALX (int i, qu_any type)
{
	SHOWME;

	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (R0);
	jit_movi_p    (R0, type);
	jit_pusharg_p (R0);
	jit_movi_p    (R0, QuNil);
	jit_pusharg_p (R0);
	jit__call     (qu_object_validate);
	jit__setlocal (i, RET);
	jit_pushr_p   (RET);
}



static void _emit_SETLEXICALLL (long i0, long i1)
{
	SHOWME;

	jit__getlocal (R0, i1);
	jit__setlocal (i0, R0);
	jit_pushr_p   (R0);
}



static void _emit_SETLEXICALLX (long i, qu_any x)
{
	SHOWME;

	jit_movi_p    (R0, x);
	jit__setlocal (i, R0);
	jit_pushr_p   (R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SETSLICE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_SETSLICE ()
{
	SHOWME;

	jit_popr_p    (I0);
	jit_popr_p    (R2);
	jit_popr_p    (R1);
	jit_popr_p    (R0);
	jit_prepare   (3);
	jit_pusharg_p (I0);
	jit_pusharg_p (R2);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit__call     (qu_sys_setslice);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SUPER
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pSUPER (qu_class sup, qu_any self, qu_any x)
{
	QU_ARG_TYPE (x, qu_string);
	return sup->get (self, x, 0);
}



static void _emit_SUPER (qu_class sup)
{
	SHOWME;

	jit_popr_p    (R1);
	jit__getself  (R0);
	jit_pushr_p   (R0);
	jit_prepare   (3);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit_movi_p    (R2, sup);
	jit_pusharg_p (R2);
	jit__call     (pSUPER);
	jit_pushr_p   (RET);
}



static qu_any _emit_SUPERX (qu_class sup, qu_string name)
{
	SHOWME;

	jit_prepare   (3);
	jit__getself  (R0);
	jit_pushr_p   (R0);
	jit_movi_p    (R1, name);
	jit_pusharg_p (R1);
	jit_pusharg_p (R0);
	jit_movi_p    (R2, sup);
	jit_pusharg_p (R2);
	jit__call     (pSUPER);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	SWAP
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* returns 0:error 1:swap  2:ignore */
static int pSWAP (jit_insn *ip)
{
	QU_FATAL ("SWAP is not yet available");
}


static void _emit_SWAP ()
{
	jit_insn *j, *g;

	SHOWME;

		jit_prepare    (1);
	j = jit_movi_p     (R0, __later__);
		jit_pusharg_p  (R0);
		jit__call      (pSWAP);
	g = jit_bnei_i     (__later__, RET, 1);
		/* WRONG --- must restore SP */
	    jit_ret        ();
		jit_patch      (g);
		jit_patch_movi (j, jit_get_label ());
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	THROW
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_THROW (long n)
{
	SHOWME;

	if (Measuring)
		n = 1;

	if (n)
	{
		jit_popr_p (R1);
		jit_ldi_p  (R2, &__QuThreadCurrent);
		jit_addi_p (R0, R2, offsetof (qu_thread_s, eclass));
		jit_str_p  (R0, R1);

	  if (n == 2)
		jit_popr_p (R1);
	  else
		jit_movi_p (R1, 0);

		jit_addi_p (R0, R2, offsetof (qu_thread_s, emsg));
		jit_str_p  (R0, R1);
	}

	if (!GeneratingExtern || Measuring)
	{
		jit_prepare   (1);
		jit_pusharg_p (FRM);
		jit__callsafe (qu_frame_rethrow);
		jit__leave;
	}
	else
	{
		jit__leave;
	}
}




/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	TRUE
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_TRUE ()
{

#ifdef OPTIMIZE_MORE

	jit_insn *j1, *j2;

	SHOWME;

	     jit_popr_p    (R0);
    j1 = jit_beqi_p    (__later__, R0, QuTrue);
    j2 = jit_beqi_p    (__later__, R0, QuFalse);
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_true);
	     jit_movr_p    (R0, RET);
	     jit_patch     (j1);
	     jit_patch     (j2);
	     jit_pushr_p   (R0);
#else
	     jit_popr_p    (R0);
	     jit_prepare   (1);
	     jit_pusharg_p (R0);
	     jit__callsafe (qu_sys_true);
	     jit_pushr_p   (RET);
#endif

}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	XCAT
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static qu_any pXCAT (int n, qu_any *sp)
{
	REVERSE (n, sp);
	return qu_sys_cat (n, sp);
}


static void _emit_XCAT (int n)
{
	SHOWME;

	jit_prepare   (2);
	jit_pusharg_p (SP);
	jit_movi_i    (I0, n);
	jit_pusharg_i (I0);
	jit__call     (pXCAT);
	jit__down     (n);
	jit_pushr_p   (RET);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	XCHG
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void _emit_XCHG ()
{
	SHOWME;

	jit_popr_p  (R0);
	jit_popr_p  (R1);
	jit_pushr_p (R0);
	jit_pushr_p (R1);
}



static void _emit_XCHG3 ()
{
	SHOWME;

	jit_popr_p  (R0);
	jit_popr_p  (R1);
	jit_popr_p  (R2);
	jit_popr_p  (I0);

	jit_pushr_p (R0);
	jit_pushr_p (R2);
	jit_pushr_p (R1);
	jit_pushr_p (I0);
}



static void _emit_XCHG4 ()
{
	SHOWME;

	jit_ldxi_p (R0, SP, 4 * sizeof (qu_any));
	jit_ldxi_p (R1, SP, 0);
	jit_movi_p (R2, SP);
	jit_addi_p (R2, R2, 4 * sizeof (qu_any));
	jit_str_p  (R2, R1);
	jit_str_p  (R1, R0);
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	YIELD
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* returns 0:error 1:yield 2:do nothing */
static int pYIELD (jit_insn *nextip)
{
	QU_FATAL ("YIELD is not yet available");
	return 0;
}


static void _emit_YIELD ()
{
	jit_insn *j;

	SHOWME;

		jit_prepare    (1);
	j = jit_movi_p     (R0, __later__);
		jit_pusharg_p  (R0);
		jit__call      (pYIELD);
		/* WRONG --- must restore SP */
	    jit_ret        ();
		jit_patch_movi (j, jit_get_label ());
}



/*----------------------------------------------------------------------------
	instruction lengths
----------------------------------------------------------------------------*/

#define arglong    (*(long *) (cs + ip + 1))
#define argany     (*(qu_any  *) (cs + ip + 1))
#define arglong2   (*(long *) (cs + ip + 1 + sizeof (qu_any)))
#define argany2    (*(qu_any  *) (cs + ip + 1 + sizeof (qu_any)))
#define pcsize(pc) (1 + (__QuCodePcType [pc] & 0xF) * sizeof (long))

static long InstLen [QU_PC_LAST_INSTRUCTION + 1];
static long HeadSize;
static long TailSize;

/* forward */
static void JitProlog ();

static void JitMeasure ()
{
	jit_insn cbuf [65536];
	jit_insn *j, *start;
	qu_sub_s fake;

#ifdef DEBUG_JIT

#define set_(i) \
	assert (0 <= (i) && (i) <= QU_PC_LAST_INSTRUCTION); \
	j = jit_get_label (); \
	InstLen[i] = j - start + 1; \
	if (i) \
		fprintf (stderr, "PC_%-15s = %d\n", qu_code_pcname (i)->str, InstLen[i]); \
	start = j

#else

#define set_(i) \
	j = jit_get_label (); \
	InstLen[i] = j - start + 1; \
	start = j

#endif /*DEBUG_JIT*/

#define po_(name)  _emit_##name ();     set_(QU_PC_##name);
#define pl_(name)  _emit_##name (2);    set_(QU_PC_##name);
#define pll_(name) _emit_##name (2, 2); set_(QU_PC_##name);
#define pla_(name) _emit_##name (2, 0); set_(QU_PC_##name);
#define pal_(name) _emit_##name (0, 2); set_(QU_PC_##name);
#define pa_(name)  _emit_##name (0);    set_(QU_PC_##name);
#define paa_(name) _emit_##name (0, 0); set_(QU_PC_##name);

	start = j = (jit_insn *) jit_set_ip (cbuf).ptr;

	jit_prolog  (3);

	TheSub = &fake;
	Measuring = 1;
	MustSetInterpFrame = 1;
	GeneratingExtern = 0;

	JitProlog ();

	HeadSize = jit_get_label () - j + 128;
	TailSize = 128;

#ifdef DEBUG_JIT
	fprintf (stderr, "HeadSize = %d\n", HeadSize);
	fprintf (stderr, "TailSize = %d\n", TailSize);
#endif

	set_(0);

	pl_ (ARG)
	po_ (ARGC)
	pl_ (ARRAY)
	po_ (AS)
	pll_(ASLL)
	pla_(ASLX)
	po_ (AUTOATTR)
	po_ (AUTOATTREX)
	po_ (AUTOATTREX2)
	pl_ (BINOP)
	pl_ (BREAK)

	/*pl_ (CALL) faked*/ pll_ (CALL)

	pl_ (CALLFLAT)
	pl_ (CALLATTR)
	pl_ (CALLATTRFLAT)
	pla_(CATCH)
	pa_ (CATCHSET)
	pa_ (CLOSURE)
	po_ (CMP)
	pll_(CMPLL)
	pla_(CMPLX)
	po_ (CONCAT)
	pll_(CONCATLL)
	pla_(CONCATLX)
	pl_ (CONTINUE)
	po_ (DECR)
	pl_ (DECRLPOST)
	pl_ (DECRLPRE)
	pl_ (DICT)
	po_ (DISCARD)
	po_ (DIV)
	pll_(DIVLL)
	pla_(DIVLX)
	po_ (DROP1)
	pll_(DROP1LL)
	pla_(DROP1LX)
	po_ (DROP2)
	po_ (DROP3)
	po_ (DUP1)
	po_ (DUP2)
	po_ (DUP3)
	po_ (EACH)
	po_ (END)
	pll_(ENTER)
	po_ (EQ)
	pll_(EQLL)
	pla_(EQLX)
	pa_ (EXTERN)
	pl_ (FORMAT)
	po_ (FLAT)
	po_ (FRAME)
	po_ (GE)
	pll_(GELL)
	pla_(GELX)
	po_ (GETATTR)
	po_ (GETATTR2)
	pa_ (GETATTRX)
	pa_ (GETATTRSELF)
	po_ (GETITEM)
	po_ (GETITEM2)
	pa_ (GETITEMX)
	pl_ (GETITEML)
	pla_(GETITEMLX)
	pa_ (GETGLOBAL)
	pl_ (GETLEXICAL)
	po_ (GETSLICE)
	po_ (GETSLICE2)
	po_ (GETSWITCH)
	po_ (GT)
	pll_(GTLL)
	pla_(GTLX)
	pal_(HTOPEN)
	pal_(HTCLOSE)
	pa_ (HTENTITY)
	po_ (IAND)
	pll_(IANDLL)
	pla_(IANDLX)
	po_ (IDIV)
	pll_(IDIVLL)
	pla_(IDIVLX)
	po_ (IOR)
	pll_(IORLL)
	pla_(IORLX)
	po_ (IN)
	pll_(INLL)
	pla_(INLX)
	po_ (INCR)
	pl_ (INCRLPOST)
	pl_ (INCRLPRE)
	po_ (INOT)
	pa_ (INITMOD)
	po_ (IS)
	pll_(ISLL)
	pla_(ISLX)
	pa_ (ITERNEXT)
	po_ (IXOR)
	pll_(IXORLL)
	pla_(IXORLX)
	pl_ (JMP)
	pl_ (JMPFALSE)
	pl_ (JMPFALSEX)
	pl_ (JMPTRUE)
	pl_ (JMPTRUEX)
	po_ (LE)
	pll_(LELL)
	pla_(LELX)
	po_ (LEAVE)
	po_ (LIKE)
	pll_(LIKELL)
	pla_(LIKELX)
	pl_ (LINE)
	pa_ (LITERAL)
	po_ (LSHIFT)
	pll_(LSHIFTLL)
	pla_(LSHIFTLX)
	po_ (LT)
	pll_(LTLL)
	pla_(LTLX)
	po_ (MINUS)
	pll_(MINUSLL)
	pla_(MINUSLX)
	po_ (MUL)
	pll_(MULLL)
	pla_(MULLX)
	po_ (NE)
	pll_(NELL)
	pla_(NELX)
	po_ (NEG)
	po_ (NOP)
	po_ (NOT)
	pll_(PARAM)
	po_ (PASS)
	po_ (PLUS)
	pll_(PLUSLL)
	pla_(PLUSLX)
	po_ (POP)
	po_ (POS)
	po_ (POW)
	pll_(POWLL)
	pla_(POWLX)
	pa_ (PRINT)
	po_ (PRINTTOP)
	po_ (PRINTHTML)
	po_ (PUSH)
	pll_(PUSHLL)
	pla_(PUSHLX)
	po_ (RANGE)
	po_ (REM)
	pll_(REMLL)
	pla_(REMLX)
	po_ (RET)
	po_ (RIN)
	pll_(RINLL)
	pla_(RINLX)
	po_ (ROT3)
	po_ (ROT4)
	po_ (ROT4X)
	po_ (RSHIFT)
	pll_(RSHIFTLL)
	pla_(RSHIFTLX)
	po_ (SELF)
	pl_ (SET)
	po_ (SETATTR)
	pa_ (SETATTRX)
	pa_ (SETATTRSELF)
	pa_ (SETGLOBAL)
	pa_ (SETGLOBALX)
	po_ (SETITEM)
	pa_ (SETITEMX)
	pl_ (SETITEML)
	pla_(SETITEMLX)
	pl_ (SETLEXICAL)
	pla_(SETLEXICALX)
	pll_(SETLEXICALLL)
	pla_(SETLEXICALLX)
	po_ (SETSLICE)
	pa_ (SUPER)
	paa_(SUPERX)
	po_ (SWAP)
	pl_ (THROW)
	po_ (TRUE)
	pl_ (XCAT)
	po_ (XCHG)
	po_ (XCHG3)
	po_ (XCHG4)
	po_ (YIELD)

#undef po_
#undef pl_
#undef pll_
#undef pla_
#undef pal_
#undef pa_
#undef paa_

	if (jit_get_label () - start >= sizeof (cbuf))
		QU_FATAL ("insufficient size of cbuf");

	Measuring = 0;
}



static long JitGetSize (qu_code self)
{
	unsigned char *cs = self->bytes;
	long size = 0;
	int  ip   = 0;
	int  en   = 0;
	int  pc;

	JmpCount = 0;
	BlockCount = 0;

	for (ip = 0; (pc = cs[ip]) != 0; ip += pcsize (pc))
	{

#ifdef DEBUG_JIT
	printf (">>> JIT %s\n", qu_code_pcname (pc)->str);
#endif
		switch (pc)
		{
			case QU_PC_END:
				en = 1;
				break;

			case QU_PC_CLOSURE:
				MustSetInterpFrame = 1;
				/*nobreak*/
			case QU_PC_FRAME:
			case QU_PC_GETATTR:
			case QU_PC_GETATTR2:
			case QU_PC_GETATTRX:
			case QU_PC_SETATTR:
			case QU_PC_SETATTRX:
			case QU_PC_PARAM:
			case QU_PC_ENTER: /* implies QU_PC_SWITCH and QU_PC_ITERNEXT */
				GeneratingExtern = 0;
				break;

			default:
				if (pc < 0 || pc > QU_PC_LAST_INSTRUCTION)
				{
					QU_SYS_WARN ("unknown instruction %d\n", pc);
					return -1;
				}
		}
		size += InstLen [pc];
		assert (size >= 0);
	}
	if (en == 0)
	{
		QU_SYS_WARN ("no END instruction\n");
		return -1;
	}

	JmpCount = 0;
	BlockCount = 0;

	return size + HeadSize + TailSize;
}



/*----------------------------------------------------------------------------
	builder
----------------------------------------------------------------------------*/

static int JitMustSetInterpFrame (qu_sub u)
{
	qu_attr_s *a = u->attrs;

	if (a)
	{
		long      e = 0;
		qu_string s;
		qu_any    x;

		while ((x = qu_attr_walk (a, &s, &e)) != 0)
		{
			if (qu_sub_is (x))
				return 1;
		}
	}
	return 0;
}



static void _emit_setup_local_extern ()
{
	/*
		case-1: nlocal == 0
		------

			...      <-- LP = BSP = SP     (LP will never be used anyway)


		case-2: nlocal == argc, both > 0
		------
			...      <-- BSP = SP

		  use caller stack directly:      (yes, the qu_sub may change the value
		    arg-2                          of the arguments, so what! )
		    arg-1    <-- LP
			...

		case-3: nlocal > 0, argc == 0
		------

			local-3                       (locals are initialized to QuNil)
			local-2
			local-1  <-- LP = BSP = SP


		case-4: nlocal > 0, argc > 0, nlocal > argc
		------

			local-3                      (locals are initialized to QuNil)
			local-2
			local-1
			arg-2
			arg-1    <-- LP = BSP = SP
	*/

	SHOWME;

	if (Measuring)
		TheSub->nlocal = 2;

	if (TheSub->nlocal == 0)
	{
	     /* assert argc == 0 */
	     jit_movr_p (LP, SP);
	}
	else
	{
	     jit_insn *j1, *j2, *j3, *j4, *j5, *j6;

	     jit_movi_i    (R0, TheSub->nlocal);
	     jit_getarg_i  (I0, ArgArgc);
	j1 = jit_bner_i    (__later__, R0, I0);

	     /* nlocal == argc, both > 0 */

	     jit_getarg_p  (LP, ArgArgv);
	j2 = jit_beqi_i    (__later__, I0, 0);

	     /* nlocal > argc, argc might be 0 */

	     jit_patch     (j1);
	     jit__up       (TheSub->nlocal);
	     jit_movr_p    (LP, SP);
	     jit_movr_p    (R1, LP);
	     jit_getarg_p  (R2, ArgArgv);
	j3 = jit_get_label ();
	j4 = jit_blei_i    (__later__, I0, 0);
	     jit_ldr_p     (R0, R2);
	     jit_str_p     (R1, R0);
	     jit_addi_p    (R1, R1, sizeof (qu_any));
	     jit_addi_p    (R2, R2, sizeof (qu_any));
	     jit_subi_i    (I0, I0, 1);
	     jit_jmpi      (j3);

	     jit_patch     (j4);
	     jit_movi_i    (I0, TheSub->nlocal);
	     jit_getarg_i  (R0, ArgArgc);
	     jit_subr_i    (I0, I0, R0);
	     jit_movi_p    (R0, QuNil);
	j5 = jit_get_label ();
	j6 = jit_blei_i    (__later__, I0, 0);
	     jit_str_p     (R1, R0);
	     jit_addi_p    (R1, R1, sizeof (qu_any));
	     jit_subi_i    (I0, I0, 1);
	     jit_jmpi      (j5);

	     jit_patch     (j6);
	     jit_patch     (j2);
	}
}


static void JitProlog ()
{
	int ext = GeneratingExtern;
	jit_insn *j;

	if (Measuring)
		ext = 1;

	if (!ext)
	{
		/* use FRM as the qu_frame object */
		jit_getarg_p  (FRM, ArgFrame);

		/* use LP as the local stack pointer  = qu_frame->ss */
		jit_ldxi_p    (LP, FRM, offsetof (qu_frame_s, ss));

		/* save current stack pointer to qu_frame->bp */
		jit_addi_p    (R0, FRM, offsetof (qu_frame_s, bp));
		jit_str_p     (R0, SP);

		SHOWREG (">>>>> JIT entry  SP = ", SP);
		SHOWREG (">>>>> JIT entry FRM = ", FRM);
	}
	else
	{
		/* save stack pointer */
		jit_movr_p    (BSP, SP);

		_emit_setup_local_extern ();

		SHOWREG (">>>>> EXTERN entry  SP = ", SP);
		SHOWREG (">>>>> EXTERN entry BSP = ", BSP);

#if 0
		/* get the current frame */
		jit_movi_p (R0, &__QuInterpFrame);
		jit_ldxi_p (FRM, R0, 0);
#endif

	}

	j = jit_jmpi (__later__);

	if (Measuring)
		ext = 0;

	if (!ext)
	{
		/* leave code */
		LeaveAddr = jit_get_label ();

#ifdef SHOW
		jit_prepare (1);
		jit_movi_p  (R0, "--- leave code JIT");
		jit_pushr_p (R0);
		jit_finish  (show_me);
#endif
		_emit_leave_code ();

		EndAddr = jit_get_label ();

#ifdef SHOW
		jit_prepare (1);
		jit_movi_p  (R0, "--- end code JIT");
		jit_pushr_p (R0);
		jit_finish  (show_me);
#endif
		SHOWREG ("<<<<< JIT end code  SP", SP);
		SHOWREG ("<<<<< JIT end code FRM", FRM);

		_emit_end_code ();
		jit_calli (abort); /* never */


		/* throw code */
		ThrowAddr = jit_get_label ();

#ifdef SHOW
		jit_prepare (1);
		jit_movi_p  (R0, "--- throw code JIT");
		jit_pushr_p (R0);
		jit_finish  (show_me);
#endif

		SHOWREG ("<<<<< JIT throw code  SP", SP);
		SHOWREG ("<<<<< JIT throw code FRM", FRM);

		jit_prepare   (1);
		jit_pusharg_p (FRM);
		jit_finish    (qu_frame_rethrow);
		jit_jmpi      (LeaveAddr);
	}
	else
	{
		LeaveAddr = jit_get_label ();
		ThrowAddr = LeaveAddr;
		EndAddr   = LeaveAddr;

		SHOWREG ("<<<<< EXTERN leaving  SP", SP);
		SHOWREG ("<<<<< EXTERN leaving BSP", BSP);

		jit_movr_p (SP, BSP);
		jit_movi_p (RET, 0);
		jit_ret    ();
	}

	jit_patch (j);

#ifdef SHOW
	jit_prepare (1);
	jit_movi_p  (R0, "--- after prolog");
	jit_pushr_p (R0);
	jit_finish  (show_me);
#endif

}



static int JitGenerate (qu_sub u)
{
	qu_code        c = u->value;
	unsigned char *cs = c->bytes;
	int            ip = 0;
	int            pc;
	long           n, size;

	jit_insn *j, *g, *max;

	assert (sizeof (jit_insn) == 1);

	/* currently we don't do generators  nor those that have __extern */
	if (qu_sub_is_generator (u))
	{
		if (qu_sys_debugging ())
			fprintf (stderr, "Lightning: cant do generators yet %s %s\n",
				qu_sub_get_module (u)->name->str, u->name->str);
		return 0;
	}

	TheSub = u;

	MustSetInterpFrame = (u != u->root) || JitMustSetInterpFrame (u);
	GeneratingExtern = !MustSetInterpFrame;

	if (u->type || qu_sub_is_argmin(u) || qu_sub_is_vararg (u))
		GeneratingExtern = 0;

#ifdef SHOW
	fprintf (stderr, "JitGenerate: %s\n", u->name->str);
#endif

	size = JitGetSize (c);
	if (size <= 0)
		return 0;

#ifdef SHOW
	fprintf (stderr, "::: size = %ld\n", size);
	fprintf (stderr, "::: generating = %s\n", GeneratingExtern ? "EXTERN" : "JIT");
#endif

	if (qu_sys_debugging ())
		fprintf (stderr, "Lightning: %s %s (%s) %d -> %d\n",
			qu_sub_get_module (u)->name->str,
			u->name->str,
			GeneratingExtern ? "EXTERN" : "JIT",
			((qu_code) u->value)->len,
			size);

	StartOfCode = qu_mem_alloc (size);
	max = StartOfCode + size;

#ifdef SHOW
	fprintf (stderr, "::: code = %p\n", StartOfCode);
#endif

    jit_set_ip (StartOfCode);

	if (!GeneratingExtern)
	{
		jit_prolog (1);

		ArgFrame = jit_arg_p ();
#ifdef SHOW
		fprintf (stderr, "::: ArgFrame = %d\n", ArgFrame);
#endif
	}
	else
	{
		jit_prolog (3);

		ArgSelf = jit_arg_p ();
		ArgArgc = jit_arg_i ();
		ArgArgv = jit_arg_p ();
#ifdef SHOW
		fprintf (stderr, "::: ArgSelf = %d\n", ArgSelf);
		fprintf (stderr, "::: ArgArgc = %d\n", ArgArgc);
		fprintf (stderr, "::: ArgArgv = %d\n", ArgArgv);
#endif
	}

#ifdef SHOW
	jit_prepare (1);
	jit_movi_p  (R0, "--- entering");
	jit_pushr_p (R0);
	jit_finish  (show_me);
#endif

	JitProlog ();

	TailAddr = jit_get_label ();

	for (ip = 0; (pc = cs[ip]) != 0; ip += pcsize (pc))
	{
		_map_ip (ip, jit_get_label ());

		switch (pc)
		{

#define po_(name)  case QU_PC_##name: _emit_##name ();                  continue;
#define pl_(name)  case QU_PC_##name: _emit_##name (arglong);           continue;
#define pll_(name) case QU_PC_##name: _emit_##name (arglong, arglong2); continue;
#define pla_(name) case QU_PC_##name: _emit_##name (arglong, argany2);  continue;
#define pa_(name)  case QU_PC_##name: _emit_##name (argany);            continue;
#define pal_(name) case QU_PC_##name: _emit_##name (argany, arglong2);  continue;
#define paa_(name) case QU_PC_##name: _emit_##name (argany, argany2);   continue;

	pl_ (ARG)
	po_ (ARGC)
	pl_ (ARRAY)
	po_ (AS)
	pll_(ASLL)
	pla_(ASLX)
	po_ (AUTOATTR)
	po_ (AUTOATTREX)
	po_ (AUTOATTREX2)
	pl_ (BINOP)
	pl_ (BREAK)

	/*
	pl_ (CALL)
	*/
	case QU_PC_CALL:
		_emit_CALL (arglong, cs[ip + pcsize (pc)] == QU_PC_RET);
		break;

	pl_ (CALLFLAT)
	pl_ (CALLATTR)
	pl_ (CALLATTRFLAT)
	pla_(CATCH)
	pa_ (CATCHSET)
	pa_ (CLOSURE)
	po_ (CMP)
	pll_(CMPLL)
	pla_(CMPLX)
	po_ (CONCAT)
	pll_(CONCATLL)
	pla_(CONCATLX)
	pl_ (CONTINUE)
	po_ (DECR)
	pl_ (DECRLPOST)
	pl_ (DECRLPRE)
	pl_ (DICT)
	po_ (DISCARD)
	po_ (DIV)
	pll_(DIVLL)
	pla_(DIVLX)
	po_ (DROP1)
	pll_(DROP1LL)
	pla_(DROP1LX)
	po_ (DROP2)
	po_ (DROP3)
	po_ (DUP1)
	po_ (DUP2)
	po_ (DUP3)
	po_ (EACH)
	po_ (END)
	pll_(ENTER)
	po_ (EQ)
	pll_(EQLL)
	pla_(EQLX)
	pa_ (EXTERN)
	pl_ (FORMAT)
	po_ (FLAT)
	po_ (FRAME)
	po_ (GE)
	pll_(GELL)
	pla_(GELX)
	po_ (GETATTR)
	po_ (GETATTR2)
	pa_ (GETATTRX)
	pa_ (GETATTRSELF)
	po_ (GETITEM)
	po_ (GETITEM2)
	pa_ (GETITEMX)
	pl_ (GETITEML)
	pla_(GETITEMLX)
	pa_ (GETGLOBAL)
	pl_ (GETLEXICAL)
	po_ (GETSLICE)
	po_ (GETSLICE2)
	po_ (GETSWITCH)
	po_ (GT)
	pll_(GTLL)
	pla_(GTLX)
	pal_(HTOPEN)
	pal_(HTCLOSE)
	pa_ (HTENTITY)
	po_ (IAND)
	pll_(IANDLL)
	pla_(IANDLX)
	po_ (IDIV)
	pll_(IDIVLL)
	pla_(IDIVLX)
	po_ (IOR)
	pll_(IORLL)
	pla_(IORLX)
	po_ (IN)
	pll_(INLL)
	pla_(INLX)
	po_ (INCR)
	pl_ (INCRLPOST)
	pl_ (INCRLPRE)
	po_ (INOT)
	pa_ (INITMOD)
	po_ (IS)
	pll_(ISLL)
	pla_(ISLX)
	pa_ (ITERNEXT)
	po_ (IXOR)
	pll_(IXORLL)
	pla_(IXORLX)
	pl_ (JMP)
	pl_ (JMPFALSE)
	pl_ (JMPFALSEX)
	pl_ (JMPTRUE)
	pl_ (JMPTRUEX)
	po_ (LE)
	pll_(LELL)
	pla_(LELX)
	po_ (LEAVE)
	po_ (LIKE)
	pll_(LIKELL)
	pla_(LIKELX)
	pl_ (LINE)
	pa_ (LITERAL)
	po_ (LSHIFT)
	pll_(LSHIFTLL)
	pla_(LSHIFTLX)
	po_ (LT)
	pll_(LTLL)
	pla_(LTLX)
	po_ (MINUS)
	pll_(MINUSLL)
	pla_(MINUSLX)
	po_ (MUL)
	pll_(MULLL)
	pla_(MULLX)
	po_ (NE)
	pll_(NELL)
	pla_(NELX)
	po_ (NEG)
	po_ (NOP)
	po_ (NOT)
	pll_(PARAM)
	po_ (PASS)
	po_ (PLUS)
	pll_(PLUSLL)
	pla_(PLUSLX)
	po_ (POP)
	po_ (POS)
	po_ (POW)
	pll_(POWLL)
	pla_(POWLX)
	pa_ (PRINT)
	po_ (PRINTTOP)
	po_ (PRINTHTML)
	po_ (PUSH)
	pll_(PUSHLL)
	pla_(PUSHLX)
	po_ (RANGE)
	po_ (REM)
	pll_(REMLL)
	pla_(REMLX)
	po_ (RET)
	po_ (RIN)
	pll_(RINLL)
	pla_(RINLX)
	po_ (ROT3)
	po_ (ROT4)
	po_ (ROT4X)
	po_ (RSHIFT)
	pll_(RSHIFTLL)
	pla_(RSHIFTLX)
	po_ (SELF)
	pl_ (SET)
	po_ (SETATTR)
	pa_ (SETATTRX)
	pa_ (SETATTRSELF)
	pa_ (SETGLOBAL)
	pa_ (SETGLOBALX)
	po_ (SETITEM)
	pa_ (SETITEMX)
	pl_ (SETITEML)
	pla_(SETITEMLX)
	pl_ (SETLEXICAL)
	pla_(SETLEXICALX)
	pll_(SETLEXICALLL)
	pla_(SETLEXICALLX)
	po_ (SETSLICE)
	pa_ (SUPER)
	paa_(SUPERX)
	po_ (SWAP)
	pl_ (THROW)
	po_ (TRUE)
	pl_ (XCAT)
	po_ (XCHG)
	po_ (XCHG3)
	po_ (XCHG4)
	po_ (YIELD)

#undef po_
#undef pl_
#undef pll_
#undef pla_
#undef pal_
#undef pa_
#undef paa_

		default:
			QU_FATAL ("unknown bytecode %d", ip);

		}
	}

	j = jit_get_label ();
	if (j >= max)
		QU_FATAL ("wrong size: allocated %ld - creating %d", size, j - StartOfCode);

	for (n = 0; n < JmpCount; n++)
	{
		jmp_s *jp = &Jmp[n];

		j = jp->ofs;
		g = Ip [jp->ip];

		if (jp->type == 0)
			jit_patch_at (j, g);
		else
			jit_patch_movi (j, g);
	}

	for (n = 0; n < BlockCount; n++)
	{
		block_s *b = &Block[n];
		block_s *e;

		e = b->ofs;
		e->nextblk = b->nextblk;
		e->nextip = Ip [b->ip];
		e->contip = b->contip;
	}

	jit_flush_code (StartOfCode, jit_get_ip().ptr);


	if (GeneratingExtern)
	{
		u->argc = -1;
		u->value = StartOfCode;
		qu_sub_jit_set_extern (u);
	}
	else
	{
		qu_mem_free (c->bytes);
		c->bytes = (unsigned char *) StartOfCode;
		qu_sub_jit_set (u);
	}

#ifdef SHOW
	fprintf (stderr, "::: Generating done\n");
#endif

	return 1;
}

#undef argany
#undef arglong
#undef arglong2
#undef argany2
#undef pcsize



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

#if 0
static qu_any JitCall (qu_frame f)
{

#ifdef SHOW
	qu_any result;

	fprintf (stderr, "JitCall ::::: %s\n", ((qu_sub) qu_frame_get_symbol (f))->name->str);
	fprintf (stderr, "   Frame=%p  ss=%p  sp=%p\n", f, f->ss, f->sp);
	result = ((qu_subc1_f) f->code->bytes) (f);
	fprintf (stderr, "JitCall ::::: RESULT = %p\n", result);
	return result;
#else
	return ((qu_subc1_f) f->code->bytes) (f);
#endif
}
#endif


int qu_Lightning_init (qu_module m)
{

#ifdef DEBUG_JIT
	fprintf (stderr, "Jit measuring instruction sizes\n");
#endif

	JitMeasure ();

	qu_code_jit_generate = JitGenerate;

#if 0
	qu_code_jit_call = JitCall;
/*	__QuCodeJitPrint = JitPrint; TODO */
#endif

	return 1;
}


#endif /*LIGHTNING_CPU_NONE*/
