/* qu_string.c - Copyright (c) Marc Krisnanto */

#include "qu.h"


/*) class String __final
	String objects are immutable byte character based strings.

	String objects are literals much like -{Number} objects.
	You can not modify the content of a String as much as you can not change
	the number 1 into something else.

	If you want "mutable" strings, see -{Buffer}.

	Strings may contain any byte characters including zeroes ("nils").
	Like C strings, they are "nil" terminated.
	There is always one 0x00 byte appended at the end.

	You create String objects with the syntax or via the #{new} method.
	Refer to the I{syntax} manual for more information.

	@xmp
		a = 'Foo'
		b = "Bar"
		c = `Baz`
		d = a _ b _ c
		println (d)
		println (d.len)
		println (d[3])
		println (d[2, 2])

		a = String (123)    # same as a = (123).__string
		println (a)

		a = 'foo\x00bar'
		println (a.quote)

	@out
		FooBarBaz
		9
		B
		oB
		123
		'foo\x00bar'

	@p
	The maximum size of a String is a huge value defined by Qu as reasonable as
	possible depending on your system. This is probably in the range of a quarter
	billion bytes. Attempt to create strings over this limit will throw EMemory.

	Remember that in Qu, EMemory does not mean that you are out of memory.
	It simply means that you are trying to create an object with an unacceptable
	size. If you are I{really} out of memory then Qu will throw an unrecoverable
	error. See Sys documentation and look for __memory__.

	Methods requiring an index will throw EIndex if the index is not within the
	String. Most methods will take a negative value to mean counting from the end
	of the String, thus C{s[-1]} is valid and yields the last byte.

	@warn
	By default builtin methods that need the information of a character set use
	the US-ASCII character set.

	@xmp
		x = 'foo'
		println (x.toupper)

	@out
		FOO

	@p
	Empty and single byte String objects are preallocated system strings.
	They are not recreated. All String methods use these whenever possible.
*/
qu_class QuStringClass;


/* single character Strings */
qu_string __QuStringByteList [256];


/* the empty string */
qu_string QuStringEMPTY;


/*----------------------------------------------------------------------------
	static string

	Note that we can not use a qu_set because no class are initialized when
	static strings are created. See qu_class__init.
----------------------------------------------------------------------------*/

/*
	Initial size of table must be n^2.
	8192 seems to be fine.
*/
#define STRING_STATIC_INIT_SIZE 8192

typedef struct entry_s entry_s;
struct entry_s
{
	entry_s   *next;
	entry_s   *right;
	qu_string  str;
};

static long      StaticCount   = 0;
static long      StaticMask    = STRING_STATIC_INIT_SIZE - 1;
static entry_s  *StaticFirst   = NULL;
static entry_s **StaticBuckets = NULL;


/*
	Add a new static qu_string entry.
*/
static qu_string _string_static_add (qu_string s, long hash)
{
	register entry_s *e;
	register long     h = hash & StaticMask;

	assert (qu_string_is_static (s));

	e        = qu_mem_alloc (sizeof (entry_s));
	e->str   = s;
	e->next  = StaticFirst;
	e->right = StaticBuckets [h];

	StaticBuckets [h] = StaticFirst = e;

	/* resize table */
	if (++StaticCount > ((StaticMask << 1) / 3))
	{
		long      m = ((StaticMask + 1) << 1) - 1;
		entry_s **b = qu_mem_calloc (m + 1, sizeof (entry_s *));

		for (e = StaticFirst; e; e = e->next)
		{
			h = qu_string_HASH (e->str) & m;
			e->right = b [h];
			b [h] = e;
		}
		qu_mem_free (StaticBuckets);
		StaticBuckets = b;
		StaticMask = m;
	}
	return s;
}



/*
	Find an existing static qu_string.

	This is called by e.g. the qu_tokenizer thus avoid creating the same static
	qu_string object if one already exists.

	This should happen very often since, for example, most programmers will
	write a lot of local symbols with the same name. Not to mention standard
	method names.
*/
static qu_string _string_static_findz (const char *z, long len, long hash)
{
	entry_s *e;

	for (e = StaticBuckets [hash & StaticMask]; e; e = e->right)
	{
		if (e->str->len == len && memcmp (e->str->str, z, len) == 0)
			return e->str;
	}
	return 0;
}



/*
	Find an existing static qu_string.
	This time by specifying an already created qu_string.

	The idea here is that if a static qu_string exists, then get rid of the
	other qu_string, that is, let it be collected by the GC later on.
*/
static qu_string _string_static_find (qu_string s)
{
	entry_s *e;

	for (e = StaticBuckets [qu_string_HASH (s) & StaticMask]; e; e = e->right)
	{
		if (e->str == s /* <- might hit a lot of times */
			|| (e->str->len == s->len
					&& memcmp (e->str->str, s->str, s->len) == 0))
			return e->str;
	}
	return 0;
}



/*
	Find a static version and return it if it exists.
	Return s otherwise.
*/
qu_string qu_string_get_static (qu_string s)
{
	if (!qu_flag_test (s, QU_STRING_FLAG_STATIC))
	{
		qu_string t = _string_static_find (s);
		if (t)
			return t;
	}
	return s;
}



/*
	Variation of qu_string_get_static.
*/
qu_string qu_string_static_findz (const char *z, long len)
{
	if (len == 0)
		return QuStringEMPTY;

	else if (len == 1)
		return qu_string_byte (*z);

	else
		return _string_static_findz (z, len, qu_strz_hash (z, len));
}



/*----------------------------------------------------------------------------
	static constructors
----------------------------------------------------------------------------*/

/*
	Create a static non-GC'ed qu_string.
	Or return an existing one.
*/
qu_string qu_string_new_static (const char *z, long len)
{
	assert (0 <= len && len < QU_MEM_MAX);

	if (len == 0)
		return QuStringEMPTY;

	else if (len == 1)
		return qu_string_byte (*z);

	else
	{
		long      h = qu_strz_hash (z, len);
		qu_string s = _string_static_findz (z, len, h);

		if (s == 0)
		{
			s         = qu_gc_new_static (QuStringClass);
			s->flags |= QU_STRING_FLAG_STATIC;
			s->len    = len;
			s->str    = (char *) z;
			s->hash   = h;

			_string_static_add (s, h);
		}
		return s;
	}
}



/*
	Same thing with qu_string_new_static but we do the strlen.
*/
qu_string qu_string_new_staticz (const char *z)
{
	return qu_string_new_static (z, strlen (z));
}



/*
	Same as qu_string_new_static but we dup the string.
*/
qu_string qu_string_static_dup (const char *z, long len)
{
	assert (0 <= len && len < QU_MEM_MAX);

	if (len == 0)
		return QuStringEMPTY;

	else if (len == 1)
		return qu_string_byte (*z);

	else
	{
		long      h = qu_strz_hash (z, len);
		qu_string s = _string_static_findz (z, len, h);

		if (s == 0)
		{
			s         = qu_gc_new_static (QuStringClass);
			s->flags |= QU_STRING_FLAG_STATIC;
			s->len    = len;
			s->str    = qu_mem_alloc (len + 1);
			s->hash   = h;

			memcpy (s->str, z, len);
			s->str[len] = '\0';
			_string_static_add (s, h);
		}
		return s;
	}
}



/*
	Same as qu_string_static_dup but we calclate the length.
*/
qu_string qu_string_static_dupz (const char *z)
{
	return qu_string_static_dup (z, strlen (z));
}


/*
	Same as qu_string_new_static except the we use a sprintf formatting facility.
*/
qu_string qu_string_new_staticv (const char *fmt, va_list va)
{
	char      *z;
	long       len;
	size_t     size;
	qu_string  s;

	size = strlen (fmt) * 4 + 2;
	if (size < 48)
		size = 48;
	for (;;)
	{
		QU_MEM_CHECK (size < QU_MEM_MAX);
		z = qu_mem_alloc (size);
		len = vsnprintf (z, size, fmt, va);
		if (len < 0)
			size *= 2;
		else if (len < size)
			break;
		else
			size = len + 1;
		qu_mem_free (z);
	}

	if (len == 0)
	{
		qu_mem_free (z);
		return QuStringEMPTY;
	}
	else if (len == 1)
	{
		char c = *z;
		qu_mem_free (z);
		return qu_string_byte (c);
	}
	else
	{
		s         = qu_gc_new_static (QuStringClass);
		s->flags |= QU_STRING_FLAG_STATIC;
		s->len    = len;
		s->hash   = -1;
		s->str    = z;
		return s;
	}
}



/*
	Same as qu_string_new_static except the we use a sprintf formatting facility.
*/
qu_string qu_string_new_staticf (const char *fmt, ...)
{
	va_list   va;
	qu_string  s;

	va_start (va, fmt);
	s = qu_string_new_staticv (fmt, va);
	va_end (va);
	return s;
}



/*----------------------------------------------------------------------------
	allocate
----------------------------------------------------------------------------*/

/*
	Create a new regular GC'ed qu_string.
*/
qu_string qu_string_new (const char *z, long len)
{
	if (len == 0)
		return QuStringEMPTY;

	else if (len == 1)
		return qu_string_byte (*z);

	else
	{
		register qu_string s;

		assert (0 <= len && len < QU_MEM_MAX);

		s       = qu_gc_new (QuStringClass);
		s->len  = len;
		s->hash = -1;
		s->str  = qu_mem_alloc (len + 1);
		memcpy (s->str, z, len);
		s->str [len] = '\0';
		return s;
	}
}



/*
	Same as qu_string_new except that we do the strlen.
*/
qu_string qu_string_newz (const char *z)
{
	return qu_string_new (z, strlen (z));
}



/*
	Same as qu_string_new but we do the dirty work to trim the qu_string.
*/
qu_string qu_string_new_trim (const char *s, long l)
{
	char *e;

	while (l && qu_ascii_isspace (*s))
	{
		--l;
		++s;
	}
	e = (char *) s + l;
	while (l && qu_ascii_isspace (e[-1]))
	{
		--l;
		--e;
	}

	if (l == 0)
		return QuStringEMPTY;
	else if (l == 1)
		return qu_string_byte (*s);
	else
		return qu_string_new (s, l);
}



/*
	Same as qu_string_new but with non-initialized content.

	Note that we put the terminating nil because most often the caller
	already knows exactly how long the qu_string will be.
*/
qu_string qu_string_newl (long len)
{
	register qu_string s;

	assert (0 <= len && len < QU_MEM_MAX);

	s       = qu_gc_new (QuStringClass);
	s->len  = len;
	s->hash = -1;
	s->str  = qu_mem_alloc (len + 1);
	s->str [len] = '\0';
	return s;
}



/*
	Same as qu_string_new except the we use a sprintf formatting facility.
*/
qu_string qu_string_newv (const char *fmt, va_list va)
{
	register char   *z;
	register int     len;
	register size_t  size;
	qu_string        s;

	size = strlen (fmt) * 4 + 2;
	if (size < 48)
		size = 48;
	for (;;)
	{
		QU_MEM_CHECK (size < QU_MEM_MAX);
		z = qu_mem_alloc (size);
		len = vsnprintf (z, size, fmt, va);
		if (len < 0)
			size *= 2;
		else if (len < size)
			break;
		else
			size = len + 1;
		qu_mem_free (z);
	}

	if (len == 0)
	{
		qu_mem_free (z);
		return QuStringEMPTY;
	}
	else if (len == 1)
	{
		char c = *z;
		qu_mem_free (z);
		return qu_string_byte (c);
	}
	else
	{
		s       = qu_gc_new (QuStringClass);
		s->len  = len;
		s->hash = -1;
		s->str  = z;
		return s;
	}
}



/*
	Same as qu_string_new except the we use a sprintf formatting facility.
*/
qu_string qu_string_newf (const char *fmt, ...)
{
	va_list   va;
	qu_string s;

	va_start (va, fmt);
	s = qu_string_newv (fmt, va);
	va_end (va);
	return s;
}



/*
	Same as qu_string_new which content is a repetition of a specified byte.
*/
qu_string qu_string_newc (char c, long n)
{
	QU_MEM_CHECK (n < QU_MEM_MAX);

	if (n == 0)
		return QuStringEMPTY;
	else if (n == 1)
		return qu_string_byte (c);
	else
	{
		qu_string s = qu_string_newl (n);
		memset (s->str, c, n);
		return s;
	}
}



/*
	Same as qu_string_new but from a syntax source, i.e. the qu_tokenizer.

	\a = 0x07
	\b = 0x08
	\t = 0x09
	\n = 0x0a
	\v = 0x0b
	\f = 0x0c
	\r = 0x0d
	\e = 0x1b
	\xnn -> hex 2 digits
	\0nn -> oct 2 digits
*/
qu_string qu_string_source (const char *z, long l)
{
	char      *i, *j;
	qu_string  s;

	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
		return qu_string_byte (*z);

	QU_MEM_CHECK (l < QU_MEM_MAX);

	s = qu_string_newl (l);
	j = s->str;
	i = (char *) z;

	while (l--)
	{
		if (*i != '\\')
		{
			*j++ = *i++;
		}
		else
		{
			QU_REQUIRE (QuESyntax, l);
			++i;
			--l;
			switch (*i++)
			{
				case 'n': *j++ = '\n'; continue;
				case 't': *j++ = '\t'; continue;
				case 'r': *j++ = '\r'; continue;
				case 'v': *j++ = '\v'; continue;
				case 'f': *j++ = '\f'; continue;
				case 'a': *j++ = '\a'; continue;
				case 'b': *j++ = '\b'; continue;
				case 'e': *j++ = '\e'; continue;
				case 'x':
					QU_REQUIRE (QuESyntax, l >= 2 && qu_ascii_isxdigit (i[0]) && qu_ascii_isxdigit (i[1]));
					*j++ = (qu_ascii_hexval (i[0]) << 4) + qu_ascii_hexval (i[1]);
					i += 2;
					l -= 2;
					continue;
				case '0':
					QU_REQUIRE (QuESyntax, l >= 2 && qu_ascii_isodigit (i[0]) && qu_ascii_isodigit (i[1]));
					*j++ = ((i[0] - '0') << 3) + (i[1] - '0');
					i += 2;
					l -= 2;
					continue;
				default:
					*j++ = i[-1];
			}
		}
	}
	*j = '\0';
	s->len = j - s->str;
	if (s->len == 1)
		return qu_string_byte (s->str[0]);
	return s;
}



/*
	Concatenate Strings.
*/
qu_string qu_string_cat (qu_string s, ...)
{
	va_list   va;
	qu_string r = s, x = s;
	long      l = 0, n;

	va_start (va, s);
	do
	{
		/*
			We keep track of a single character string just in case it is
			the only non-empty string so we can just return it and avoid
			recreating it.
		*/

		assert (qu_string_is (x));
		n = x->len;
		if (l > QU_MEM_MAX - n)
		{
			va_end (va);
			QU_THROW (QuEMemory);
		}
		if (n == 1)
			r = x;
		l += n;
		x = va_arg (va, qu_string);

	} while (x);
	va_end (va);

	if (l == 0)
	{
		return QuStringEMPTY;
	}
	else if (l > 1)
	{
		register char *i, *j;

		r = qu_string_newl (l);
		i = r->str;
		va_start (va, s);
		do
		{
			j = s->str;
			l = s->len;
			while (l--)
				*i++ = *j++;
			s = va_arg (va, qu_string);

		} while (s);
		*i = '\0';
		va_end (va);
	}
	return r;
}



/*
	Make a working duplicate.
	-- we want this function to be as quick as possible
*/
qu_string qu_string_dup (qu_string self)
{
	register qu_string s = qu_gc_new (QuStringClass);
	register long      l = self->len;

	s->hash = self->hash;
	s->len  = l;
	s->str  = qu_mem_alloc (l + 1);
	memcpy (s->str, self->str, l + 1);
	assert (s->str [l] == '\0');
	return s;
}



qu_string qu_string_to_static (qu_string self)
{
	qu_string s = qu_string_get_static (self);

	if (s)
		return s;
	return qu_string_new_static (self->str, self->len);
}



/*----------------------------------------------------------------------------
	qu_string_buffer routines
----------------------------------------------------------------------------*/

#if 0
#define STRING_BUFFER_MINLEN 8
static long _string_buffer_sizer (long cap)
{
	assert (cap >= 0);

	if (cap >= QU_GC_MAX_ITEMS / 2)
		return cap;
	else if (cap < STRING_BUFFER_MINLEN)
		return STRING_BUFFER_MINLEN;
	else
	{
		register long c = 1;
		while (c < cap)
			c <<= 1;
		return c;
	}
}
#else
#define _string_buffer_sizer qu_container_sizer
#endif


/*
	Create a new qu_string buffer with a specified capacity.
*/
qu_string qu_string_buffer_new (long cap)
{
	register qu_string s;

	assert (0 <= cap && cap < QU_MEM_MAX);

	cap = _string_buffer_sizer (cap);

	s       = qu_gc_new (QuStringClass);
	s->len  = 0;
	s->hash = cap;
	s->str  = qu_mem_alloc (cap + 1);
	return s;
}



/*
	Clear content.
*/
void qu_string_buffer_clear (qu_string self)
{
	self->len = 0;
}



/*
	Make sure there is enough room for 'more' bytes.
*/
int qu_string_buffer_make_space (qu_string self, long more)
{
	QU_MEM_CHECK (more < QU_MEM_MAX - self->len);
	self->hash = _string_buffer_sizer (self->len + more);
	self->str = qu_mem_realloc (self->str, self->hash + 1);
	return 1;
}



/*
	Get the current capacity.
*/
long qu_string_buffer_avail (qu_string self)
{
	return self->hash;
}



/*
	Put a C string.
*/
int qu_string_buffer_put (qu_string self, const char *s, long len)
{
	QU_MEM_CHECK (len < QU_MEM_MAX - self->len);
	if (self->hash < self->len + len)
	{
	 	self->hash = _string_buffer_sizer (self->len + len);
		self->str = qu_mem_realloc (self->str, self->hash + 1);
	}
	memcpy (self->str + self->len, s, len);
	self->len += len;
	return 1;
}




/*
	Same as qu_string_buffer_put but we do the strlen.
*/
int qu_string_buffer_putz (qu_string self, const char *s)
{
	return qu_string_buffer_put (self, s, (long) strlen (s));
}




/*
	Put a qu_string.
*/
int qu_string_buffer_puts (qu_string self, qu_string s)
{
	return qu_string_buffer_put (self, s->str, s->len);
}



/*
	Put one character.
*/
int qu_string_buffer_putc (qu_string self, char c)
{
	return qu_string_buffer_put (self, &c, 1);
}




/*
	qu_set the length.
*/
void qu_string_buffer_set_len (qu_string self, long len)
{
	assert (len <= self->hash);
	self->len = len;
}



/*
	Reverse the content.
*/
void qu_string_buffer_reverse (qu_string self)
{
	long  l = self->len;
	char *p = self->str;
	char *q = p + l - 1;

	l /= 2;
	while (l--)
	{
		char c = *p;
		*p = *q;
		*q = c;
		p++;
		q--;
	}
}



/*
	Must call this when done constructing.
*/
qu_string qu_string_buffer_return (qu_string self)
{
	long l = self->len;

	assert (l <= self->hash);

	if (l == 0)
		return QuStringEMPTY;
	else if (l == 1)
		return qu_string_byte (self->str[0]);
	else
	{
		self->str[l] = '\0';
		self->hash = -1;
		return self;
	}
}



/*----------------------------------------------------------------------------
	qu_strz routines
----------------------------------------------------------------------------*/

/*
	Compare memory, case-sensitive or not.
*/
#define STRZEQ(az, bz, n, nocase) \
	((nocase) ? qu_strz_eqnc (az, bz, n) : qu_strz_eq (az, bz, n))


/*
	Character position, case-sensitive or not.
*/
#define STRZINCHR(z, l, c, nocase) \
	(((nocase) ? qu_strz_chrnc (z, l, c) : qu_strz_chrnc (z, l, c)) != 0)



long qu_strz_hash (const char *z, long len)
{
	register long h = 0;

#define STRING_HASH_CUT_LEN 32
	if (len > STRING_HASH_CUT_LEN)
		len = STRING_HASH_CUT_LEN;
	while (len--)
		h = h * 33 + *z++;
	if (h != -1)
		return h;
	return -2;
}



/* qu_strz_quote helper */
static qu_string _strz_quote_tick (const char *s, long l)
{
	qu_string  r;
	char      *q;

	QU_MEM_CHECK (l < QU_MEM_MAX - 2);
	r = qu_string_newl (l + 2);
	q = r->str;
	*q++ = '`';
	memcpy (q, s, l);
	q += l;
	*q++ = '`';
	*q = '\0';
	r->len = l + 2;
	return r;
}



/* qu_strz_quote helper */
static qu_string _strz_quote_single_or_double (const char *s, long l, char quot)
{
	long       n;
	qu_string  r;
	char      *q, *v, c;

	QU_MEM_CHECK (l < QU_MEM_MAX - 2);
	r = qu_string_buffer_new (l ? (l + 2) : 4);
	q = v = r->str;

	*q++ = quot;
	while (l--)
	{
		/* account for the longest escape sequence and the closing quote */
		n = q - v;
		if (n > r->hash - 5)
		{
			r->len = n;
			QU_ENSURE (qu_string_buffer_make_space (r, l + 5));
			v = r->str;
			q = v + r->len;
		}

		c = *s++;
		switch (c)
		{
			default:
				if (32 <= c && c <= 126) *q++ = c;
				else { sprintf (q, "\\x%02x", (unsigned char) c); q += 4; }
				continue;

			case '\'':
				if (quot == '\'') { *q++ = '\\'; *q++ = '\''; }
				else *q++ = '\'';
				continue;

			case '"':
				if (quot == '"') { *q++ = '\\'; *q++ = '"'; }
				else *q++ = '"';
				continue;

			case '\\': *q++ = '\\'; *q++ = '\\'; continue;
			case '\t': *q++ = '\\'; *q++ = 't';  continue;
			case '\v': *q++ = '\\'; *q++ = 'v';  continue;
			case '\f': *q++ = '\\'; *q++ = 'f';  continue;
			case '\r': *q++ = '\\'; *q++ = 'r';  continue;
			case '\n': *q++ = '\\'; *q++ = 'n';  continue;
			case '\a': *q++ = '\\'; *q++ = 'a';  continue;
			case '\b': *q++ = '\\'; *q++ = 'b';  continue;
			case '\e': *q++ = '\\'; *q++ = 'e';  continue;
		}
	}
	*q++ = quot;
	*q = '\0';
	r->len = q - r->str;
	return qu_string_buffer_return (r);
}



/*
	Make a quoted string.
*/
qu_string qu_strz_quote (const char *s, long l, char quot)
{
	if (l == 0)
	{
		char buf [2] = {quot, quot};
		return qu_string_new (buf, 2);
	}
	else if (quot == '`')
	{
		long  n = l;
		char *i = (char *) s;

		while (n--)
		{
			if (!qu_ascii_istext (*i))
				return _strz_quote_single_or_double (s, l, '\'');
			++i;
		}
		return _strz_quote_tick (s, l);
	}
	else
	{
		return _strz_quote_single_or_double (s, l, quot);
	}
}



/*
	Compare 2 strings.
*/
int qu_strz_cmp (const char *a, long la, const char *b, long lb)
{
	while (la && lb)
	{
		if (*a > *b)
			return 1;
		if (*a < *b)
			return -1;
		++a;
		++b;
		--la;
		--lb;
	}
	return (la > lb) ? 1: ((la < lb) ? -1 : 0);
}



/*
	Compare 2 strings case insensitive.
*/
int qu_strz_casecmp (const char *a, long la, const char *b, long lb)
{
	while (la && lb)
	{
		char c = qu_ascii_tolower (*a);
		char d = qu_ascii_tolower (*b);

		if (c > d)
			return 1;
		if (c < d)
			return -1;
		++a;
		++b;
		--la;
		--lb;
	}
	return (la > lb) ? 1: ((la < lb) ? -1 : 0);
}



/*
	Equality test.
*/
int qu_strz_eq (const char *a, const char *b, long n)
{
	while (n--)
	{
		if (*a != *b)
			return 0;
		++a;
		++b;
	}
	return 1;
}



/*
	Equality test - case insensitive.
*/
int qu_strz_eqnc (const char *a, const char *b, long n)
{
	while (n--)
	{
		if (qu_ascii_tolower (*a) != qu_ascii_tolower (*b))
			return 0;
		++a;
		++b;
	}
	return 1;
}



/*
	Character position.
*/
char *qu_strz_chr (const char *z, long n, char c)
{
	while (n--)
	{
		if (*z == c)
			return (char *) z;
		++z;
	}
	return 0;
}



/*
	Character position - case insensitive.
*/
char *qu_strz_chrnc (const char *z, long n, char c)
{
	c = qu_ascii_tolower (c);
	while (n--)
	{
		if (qu_ascii_tolower (*z) == c)
			return (char *) z;
		++z;
	}
	return 0;
}



/*
	qu_string position.
*/
char *qu_strz_str (const char *z, long l, const char *x, long m)
{
	if (m)
	{
		char u = *x++;
		while (l-- >= m)
		{
			if (*z++ == u)
			{
				char *j = (char *) z;
				char *k = (char *) x;
				long  n = m - 1;
				while (n && *j++ == *k++)
					--n;
				if (n == 0)
					return (char *) z - 1;
			}
		}
	}
	return 0;
}



/*
	qu_string position - case insensitive.
*/
char *qu_strz_strnc (const char *z, long l, const char *x, long m)
{
	if (m)
	{
		char u = qu_ascii_tolower (*x++);
		while (l-- >= m)
		{
			if (qu_ascii_tolower (*z++) == u)
			{
				char *j = (char *) z;
				char *k = (char *) x;
				long  n = m - 1;
				while (n && qu_ascii_tolower (*j++) == qu_ascii_tolower (*k++))
					--n;
				if (n == 0)
					return (char *) z;
			}
		}
	}
	return 0;
}



/*
	Fill cset[256] with character set in z.
*/
void qu_strz_set_cset (const unsigned char *z, long l, unsigned char *cset,
					   unsigned char *cmin, unsigned char *cmax)
{
	unsigned char c, i = 255, a = 0;

	while (l--)
	{
		c = *z++;
		if (c < i)
			i = c;
		if (c > a)
			a = c;
		cset [c] = 1;
	}
	*cmin = i;
	*cmax = a;
}



/*
	Fill cset[256] with character set in z - case insensitive.
*/
void qu_strz_set_csetnc (const unsigned char *z, long l, unsigned char *cset,
					     unsigned char *cmin, unsigned char *cmax)
{
	unsigned char c, o, i = 255, a = 0;

	while (l--)
	{
		c = *z++;
		if (qu_ascii_islower (c) || qu_ascii_isupper (c))
		{
			o = qu_ascii_tolower (c);
			c = qu_ascii_toupper (c);
			cset [o] = 1;
			cset [c] = 1;
			if (c < i)
				i = c;
			if (o > a)
				a = o;
		}
		else
		{
			cset [c] = 1;
			if (c < i)
				i = c;
			if (c > a)
				a = c;
		}
	}
	*cmin = i;
	*cmax = a;
}



/*
	Copy bytes.
*/
long qu_strz_copy (const char *z, long l, char *x, long len)
{
	if (len >= l)
		len = l;
	memcpy (x, z, len);
	return len;
}



/*
	Get a slice in self, i may be negative, throws EIndex.
*/
qu_string qu_strz_slice (const char *z, long l, long i, long n)
{
	QU_ENSURE_INDEX (i, l);
	l -= i;
	if (n < 0 || n > l)
		n = l;
	return qu_string_new (z + i, n);
}



/*
	Return index of x in z or -1 if not found.
	ofs is the starting offset and may be out of the string.
*/
long qu_strz_index (const char *z, long l, const char *x, long lx, long ofs)
{
	if (0 <= ofs && ofs < l && lx > 0)
	{
		const char *p;
		for (l = l - ofs - lx, p = z + ofs; l >= 0; --l, ++p)
		{
			if (qu_strz_eq (p, x, lx))
				return p - z;
		}
	}
	return -1;
}



/*
	Same as qu_strz_index - case insensitive.
*/
long qu_strz_indexnc (const char *z, long l, const char *x, long lx, long ofs)
{
	if (0 <= ofs && ofs < l && lx > 0)
	{
		const char *p;
		for (l = l - ofs - lx, p = z + ofs; l >= 0; --l, ++p)
		{
			if (qu_strz_eqnc (p, x, lx))
				return p - z;
		}
	}
	return -1;
}



/*
	Same as qu_strz_index - backward.
*/
long qu_strz_rindex (const char *s, long l, const char *x, long m, long ofs)
{
	if (ofs < 0)
		ofs += l;

	if (l && m && m <= l && 0 <= ofs && ofs < l)
	{
		char *z = (char *) s + ofs;
		char *p = (char *) s + (l - m);
		char *r = (char *) x;

		while (p >= z)
		{
			if (qu_strz_eq (p, r, m))
				return p - s;
			p--;
		}
	}
	return -1;
}



/*
	Same as qu_strz_rindex - case insensitive.
*/
long qu_strz_rindexnc (const char *s, long l, const char *x, long m, long ofs)
{
	if (ofs < 0)
		ofs += l;

	if (l && m && m <= l && 0 <= ofs && ofs < l)
	{
		char *z = (char *) s + ofs;
		char *p = (char *) s + (l - m);
		char *r = (char *) x;

		while (p >= z)
		{
			if (qu_strz_eqnc (p, r, m))
				return p - s;
			p--;
		}
	}
	return -1;
}



/*
	Translate character based on x.
*/
qu_string qu_strz_translate (unsigned char c, qu_any x)
{
	qu_any p = 0;

	switch (qu_classid (x))
	{
		case QU_CLASS_ID_Array:
			if ((long) c < ((qu_array) x)->len)
				p = ((qu_array) x)->objs[c];
			break;

		case QU_CLASS_ID_Dict:
			p = qu_dict_getitem ((qu_dict) x, qu_string_byte (c));
			break;

		case QU_CLASS_ID_String:
			if (c < ((qu_string) x)->len)
				return qu_string_byte (((qu_string) x)->str[c]);
			break;

		case QU_CLASS_ID_Sub:
			if (!qu_sub_is_inst (x))
				p = x;
			break;

		case QU_CLASS_ID_Method:
			p = x;
			break;
	}
	if (p)
	{
		if (qu_sub_is (p))
			p = qu_sub_call1 (p, QuNil, qu_string_byte (c));
		else if (qu_method_is (p))
			p = qu_method_call1 (p, qu_string_byte (c));
	}
	if (p && qu_string_is (p))
		return (qu_string) p;
	return 0;
}



/*
	StartWith/EndWith routines.
*/
typedef struct stringwith_t stringwith_t;
struct stringwith_t
{
	int         nocase;
	const char *cursor;
	long        len;
	qu_string   match;
};


static int StrzStartWith_callback (qu_string item, qu_any idx, stringwith_t *t)
{
	if (qu_string_is (item)
		&& item->len <= t->len
		&& STRZEQ (item->str, t->cursor, item->len, t->nocase))
	{
		t->match = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}



static int StrzEndWith_callback (qu_string item, qu_any idx, stringwith_t *t)
{
	if (qu_string_is (item)
		&& item->len
		&& item->len <= t->len
		&& STRZEQ (t->cursor + t->len - item->len, item->str, item->len, t->nocase))
	{
		t->match = item;
		return QU_ITERATOR_CALLBACK_RESULT;
	}
	return QU_ITERATOR_CALLBACK_CONTINUE;
}



qu_string qu_strz_startwith (const char *i, long l, int nocase, qu_any what, qu_any *value)
{
	*value = 0;

	if (qu_string_is (what))
	{
		qu_string x = (qu_string) what;

		if (x->len
			&& x->len <= l
			&& STRZEQ (i, x->str, x->len, nocase))
			return what;
	}
	else if (qu_array_is (what))
	{
		long       n = ((qu_array) what)->len;
		qu_any    *p = ((qu_array) what)->objs;
		qu_string  x;

		while (n--)
		{
			x = (qu_string) *p++;
			if (qu_string_is (x)
					&& x->len
					&& x->len <= l
					&& STRZEQ (i, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_set_is (what))
	{
		long      b = 0, j = 0;
		qu_string x;

		while ((x = (qu_string) qu_set_walk ((qu_set) what, &b, &j)) != 0)
		{
			if (qu_string_is (x)
				&& x->len
				&& x->len <= l
				&& STRZEQ (i, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_dict_is (what))
	{
		long      b = 0, j = 0;
		qu_string k, v;

		while ((v = (qu_string) qu_dict_walk ((qu_dict) what, (qu_any *) &k, &b, &j)) != 0)
		{
			if (qu_string_is (k)
				&& k->len
				&& k->len <= l
				&& STRZEQ (i, k->str, k->len, nocase))
			{
				*value = v;
				return k;
			}
		}
	}
	else
	{
		stringwith_t t = {nocase, i, l, 0};
		if (qu_iterator_do (what, (qu_itercallback_f) StrzStartWith_callback, &t) == QU_ITERATOR_CALLBACK_RESULT)
			return t.match;
	}
	return 0;
}



qu_string qu_strz_endwith (const char *i, long l, int nocase, qu_any what, qu_any *value)
{
	*value = 0;

	if (qu_string_is (what))
	{
		qu_string x = (qu_string) what;

		if (x->len
				&& x->len <= l
				&& STRZEQ (i + l - x->len, x->str, x->len, nocase))
			return what;
	}
	else if (qu_array_is (what))
	{
		long       n = ((qu_array) what)->len;
		qu_any    *p = ((qu_array) what)->objs;
		qu_string  x;

		while (n--)
		{
			x = (qu_string) *p++;
			if (qu_string_is (x)
					&& x->len
					&& x->len <= l
					&& STRZEQ (i + l - x->len, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_set_is (what))
	{
		long      b = 0, j = 0;
		qu_string x;

		while ((x = (qu_string) qu_set_walk ((qu_set) what, &b, &j)) != 0)
		{
			if (qu_string_is (x)
					&& x->len
					&& x->len <= l
					&& STRZEQ (i + l - x->len, x->str, x->len, nocase))
				return x;
		}
	}
	else if (qu_dict_is (what))
	{
		long      b = 0, j = 0;
		qu_string k, v;

		while ((v = (qu_string) qu_dict_walk ((qu_dict) what, (qu_any *) &k, &b, &j)) != 0)
		{
			if (qu_string_is (k)
				&& k->len
				&& k->len <= l
				&& STRZEQ (i, k->str, k->len, nocase))
			{
				*value = v;
				return k;
			}
		}
	}
	else
	{
		stringwith_t t = {nocase, i, l, 0};
		if (qu_iterator_do (what, (qu_itercallback_f) StrzEndWith_callback, &t) == QU_ITERATOR_CALLBACK_RESULT)
			return t.match;
	}
	return 0;
}



/*----------------------------------------------------------------------------
	specific class constants
----------------------------------------------------------------------------*/

/*) doc constants
	The String class defines a number of String constants.

	@ops
	.| ALL8   | 256 8-bits characters
	.| ALL7   | 128 7-bits characters
	.| ALNUM  | alphanumeric characters
	.| ALPHA  | alpha characters
	.| CNTRL  | control characters
	.| BDIGIT | binary digits characters
	.| BLANK  | blank characters
	.| DIGIT  | decimal digit characters
	.| GRAPH  | graph characters
	.| IDENT  | valid identifier characters
	.| IDNUM  | valid identifier and numeric characters
	.| LOWER  | lowercase characters
	.| NAME   | valid name characters
	.| ODIGIT | octal digit characters
	.| PATH   | valid file path characters
	.| PRINT  | printable characters
	.| PUNCT  | punctuation characters
	.| PWD    | valid password characters
	.| SPACE  | space characters
	.| UPPER  | uppercase characters
	.| XDIGIT | hexadecimal digit characters
	.| WHITE  | whitespace characters
	.| WORD   | word characters
*/
static char *ALL8   = "\x00\x01\x02\x03\x04\x05\x06\a\b\t\n\v\f\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\e\x1c\x1d\x1e\x1f !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff";
static char *ALL7   = "\x00\x01\x02\x03\x04\x05\x06\a\b\t\n\v\f\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\e\x1c\x1d\x1e\x1f !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f";
static char *ALNUM  = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
static char *ALPHA  = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
static char *CNTRL  = "\x01\x02\x03\x04\x05\x06\a\b\t\n\v\f\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\e\x1c\x1d\x1e\x1f\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff";
static char *BDIGIT = "01";
static char *BLANK  = "\t ";
static char *DIGIT  = "0123456789";
static char *GRAPH  = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
static char *IDENT  = "ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";
static char *IDNUM  = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";
static char *LOWER  = "abcdefghijklmnopqrstuvwxyz";
static char *NAME   = " .ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
static char *ODIGIT = "01234567";
static char *PATH   = " -./0123456789=@ABCDEFGHIJKLMNOPQRSTUVWXYZ^_abcdefghijklmnopqrstuvwxyz~";
static char *PRINT  = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
static char *PUNCT  = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~";
static char *PWD    = "!$&()*+-./0123456789:;=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_abcdefghijklmnopqrstuvwxyz~";
static char *SPACE  = "\t\n\v\f\r ";
static char *UPPER  = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static char *XDIGIT = "0123456789ABCDEFabcdef";
static char *WHITE  = "\t\v\f\r ";
static char *WORD   = "\x01\x02\x03\x04\x05\x06\a\b\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\e\x1c\x1d\x1e\x1f!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff";

qu_string QuStringALL8;
qu_string QuStringALL7;
qu_string QuStringALNUM;
qu_string QuStringALPHA;
qu_string QuStringCNTRL;
qu_string QuStringBDIGIT;
qu_string QuStringBLANK;
qu_string QuStringDIGIT;
qu_string QuStringGRAPH;
qu_string QuStringIDENT;
qu_string QuStringIDNUM;
qu_string QuStringLOWER;
qu_string QuStringNAME;
qu_string QuStringODIGIT;
qu_string QuStringPATH;
qu_string QuStringPRINT;
qu_string QuStringPUNCT;
qu_string QuStringPWD;
qu_string QuStringSPACE;
qu_string QuStringUPPER;
qu_string QuStringXDIGIT;
qu_string QuStringWHITE;
qu_string QuStringWORD;



/*----------------------------------------------------------------------------
	standard class methods
----------------------------------------------------------------------------*/

/*) sub class new (x) -> String
	Create a new String object.

	If v{x} is already a String then v{x} is returned.

	Otherwise this simply calls the object's f{__string} method via an
	internal function that never fails. That is, if the __string method
	does not return a String or throws an exception, then the default
	Object.__string method is called.

	@noexc

	@xmp
		s = String (1.23)
		println (s.quote)

		s = String ([1, 2, 3])
		println (s.quote)

	@out
		'1.23'
		'[1, 2, 3]'
*/
static qu_string qu__string_new (qu_class self, qu_any x)
{
	if (qu_string_is (x))
		return x;
	else if (qu_buffer_is (x))
		return qu_string_new (((qu_buffer) x)->str, ((qu_buffer) x)->len);
	else
		return qu_sys_string (x);
}



/*----------------------------------------------------------------------------
	specific class methods
----------------------------------------------------------------------------*/

/*) sub class chr (x: Integer) -> String
	Return the character of the ASCII value v{x}.

	v{x} may be signed or unsigned.
	If signed then the value is truncated in -128..127.
	Otherwise it is truncated in 0..255.

	@exc
	.EArgValue

	@xmp
		println (String.chr (65))

	@out
		A
*/
static qu_string qu__string_chr (qu_class self, qu_integer x)
{
	long c;

	QU_ARG_long (x, c);
	if (c < 0)
	{
		if (c < -128)
			c = -128;
		return __QuStringByteList [256 + c];
	}
	else
	{
		if (c > 255)
			c = 255;
		return __QuStringByteList [c];
	}
}



/*) sub class make (*arg) -> String
	Construct a new String by concatenating parts of String objects.

	The arguments must be String objects optionally followed by a slice
	specification.

	Slices with an index out of the String are ignored.

	@exc
	.EArgValue
	.EMemory

	@xmp
		s = String.make ('foo', 0, 2, 'bar', 'baz', 2, -1)
		println (s)

	@out
		fobarz
*/
static qu_string qu__string_make (qu_class self, int argc, qu_any *argv)
{
	qu_string r = qu_string_buffer_new (8);

	while (argc--)
	{
		qu_string x = *argv++;

		QU_ARG_TYPE (x, qu_string);

		if (argc && qu_integer_is (argv[0]))
		{
			long l, i, n, m;

			QU_ARG_VMIN (2);
			QU_ARG_vlong (0, i);
			QU_ARG_vlong (1, n);

			argc -= 2;
			argv += 2;

			l = x->len;
			if (i < 0)
				i += l;
			if (0 <= i && i < l)
			{
				if (n < 0)
					n = l;
				m = l - i;
				if (n > m)
					n = m;
				QU_ENSURE (qu_string_buffer_put (r, x->str + i, n));
			}
		}
		else
		{
			QU_ENSURE (qu_string_buffer_puts (r, x));
		}
	}
	return qu_string_buffer_return (r);
}



/*----------------------------------------------------------------------------
	low-level instance methods
----------------------------------------------------------------------------*/

static void qu__string__destroy (qu_string self)
{
	assert (!qu_flag_test (self, QU_STRING_FLAG_STATIC));
	qu_mem_free (self->str);
}



/*----------------------------------------------------------------------------
	standard instance methods
----------------------------------------------------------------------------*/

/*) sub __as (x: Class) -> Any
	Get a version of self based on v{x}.

	v{x} must be one of the following -{Class} object:
	@ops
	.| -{Int} |
		Get a Int by interpreting this String using the language syntax.
		If it can not be interpreted or the value is too big then ESyntax is thrown.
		Note that a plus sign is considered an error and there may not be blanks
		after the minus sign.

	.| -{Long} |
		Get a Long by interpreting this String using the language syntax.
		If it can not be interpreted or the value is too big then ESyntax is thrown.
		Note that a plus sign is considered an error and there may not be blanks
		after the minus sign.

	.| -{Float} |
		Get a Float by interpreting this String using the language syntax.
		If it can not be interpreted then ESyntax is thrown.
		Note that a plus sign is considered an error and there may not be blanks
		after the minus sign.

	.| -{Big} |
		Get a Big by interpreting this String using the language syntax.
		If it can not be interpreted then ESyntax is thrown.
		Note that a plus sign is considered an error and there may not be blanks
		after the minus sign.

	.| -{Complex} |
		Get a Complex by interpreting this String using the language syntax.
		If it can not be interpreted then ESyntax is thrown.
		Note that a plus sign is considered an error and there may not be blanks
		after the minus sign.

	.| -{Number} |
		If the content of this String does not represent any number then ESyntax
		is thrown. Otherwise a -{Number} object (-{Int}, -{Long}, -{Float},
		or -{Complex}) is returned.

	.| -{Array} |
		Return an Array containing all characters in this String.

	.| -{Set} |
		Collect characters in a Set thus the Set will contain all the
		unique characters from this String.

	.| -{Dict} |
		Collect characters in a Dict. The keys will contain unique characters
		and values set to the number of those characters.

	.| -{Buffer} |
		Return a new Buffer containing the same characters as the String.

	.| -{String}  | Return self.
	.| -{Boolean} | Delegate to #{__true}.

	@exc
	.EArgValue
	.ESyntax

	@xmp
		s = '  -1_234_567xyz  '
		println (s as Int)

		s = '  -1_234_567_890_123_456xyz  '
		println (s as Long)

		s = ' -1234.56xyz '
		println (s as Float)

		s = '  -1 - 2.3jxyz '
		println (s as Complex)

		println ('abc' as Array)

		a = 'Foo and Bar' as Set
		println (a)

	@out
		-1234567
		-1234567890123456
		-1234.56
		(-1 - 2.3j)
		['a', 'b', 'c']
		{' ', 'a', 'B', 'd', 'F', 'n', 'o', 'r'}
*/
qu_array qu_string_as_array (qu_string self)
{
	long     l = self->len;
	char    *s = self->str;
	qu_any  *p;
	qu_array a;

	assert (l <= QU_GC_MAX_ITEMS);
	a = qu_array_new (l);
	p = a->objs;
	a->len = l;
	while (l--)
		*p++ = qu_string_byte (*s++);
	return a;
}



qu_set qu_string_as_set (qu_string self)
{
	long   l = self->len;
	char  *s = self->str;
	long   z = (l / 3) * 2;
	qu_set t;

	assert (z < QU_GC_MAX_ITEMS);
	t = qu_set_new (z);
	while (l--)
		qu_set_push (t, qu_string_byte (*s++));
	return t;
}



qu_dict qu_string_as_dict (qu_string self)
{
	long    l = self->len;
	char   *s = self->str;
	long    z = (l / 3) * 2;
	qu_dict t;

	assert (z < QU_GC_MAX_ITEMS);
	t = qu_dict_new (z);
	while (l--)
	{
		qu_string x = qu_string_byte (*s++);
		qu_int    c = qu_dict_getitem (t, x);

		if (c == 0)
			qu_dict_setitem (t, x, QuInt1);
		else
			qu_dict_setitem (t, x, qu_sys_incr (c));
	}
	return t;
}



qu_any qu__string__as (qu_string self, qu_class c)
{
	char   *s = self->str;
	qu_any  x = 0;

	QU_ARG_TYPE (c, qu_class);
	switch (c->id)
	{
		case QU_CLASS_ID_Array:   return qu_string_as_array   (self);
		case QU_CLASS_ID_Set:     return qu_string_as_set     (self);
		case QU_CLASS_ID_Dict:    return qu_string_as_dict    (self);
		case QU_CLASS_ID_Boolean: return qu_string_as_boolean (self);
		case QU_CLASS_ID_Buffer:  return qu_string_as_buffer  (self);
		case QU_CLASS_ID_String:  return self;

		case QU_CLASS_ID_Int:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_int (x);
			break;

		case QU_CLASS_ID_Long:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_long (x);
			break;

		case QU_CLASS_ID_Integer:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_integer (x);
			break;

		case QU_CLASS_ID_Float:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_float (x);
			break;

		case QU_CLASS_ID_Big:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_big (x);
			break;

		case QU_CLASS_ID_Decimal:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_decimal (x);
			break;

		case QU_CLASS_ID_Complex:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_COMPLEX | QU_NUMBER_SCAN_SKIPSPACE);
			if (x)
				return qu_number_as_complex (x);
			break;

		case QU_CLASS_ID_Number:
			x = qu_number_scan (&s, QU_NUMBER_SCAN_COMPLEX | QU_NUMBER_SCAN_SKIPSPACE);
			break;

		default:
			QU_ARG_THROW;
	}

	QU_REQUIRE (QuESyntax, x);
	return x;
}



/*) sub __cmp (x) -> Int
	Compare this String with another object.

	If v{x} is a String then compare bytes, otherwise delegate to -{Object.__cmp}.

	@noexc

	@xmp
		a = 'abcde'
		b = 'bcde'
		println (a <=> b)

	@out
		-1
*/
qu_int qu__string__cmp (qu_string self, qu_any x)
{
	if (qu_string_is (x))
	{
		int r = strcmp (self->str, ((qu_string) x)->str);
		return (r > 0) ? QuInt1 : ((r < 0) ? QuInt_1 : QuInt0);
	}
	return qu_object_default_cmp (self, x);
}


/*) sub __each -> Iterator
	Return a -{Iterator} object to iterate through the bytes.

	Note that the k{for} syntax does not create an iterator unless
	you specifically call the f{__each} method because the interpreter
	uses a more efficient internal iterator.

	@noexc

	@xmp
		for i in each 'abcdef'  # calls __each
			print (i, '-')
		;;
		println


		for i in 'abcdef'       # interpreter uses an internal iterator
			print (i, '-')
		;;
		println

	@out
		a-b-c-d-e-f-
		a-b-c-d-e-f-
*/
static qu_string _string_next (qu_string self, qu_iterator i)
{
	long k = i->ex1++;

	QU_REQUIRE (QuEStop, 0 <= k && k < self->len);
	return qu_string_byte (self->str[k]);
}


static qu_iterator qu__string__each (qu_string self)
{
	register qu_iterator i;

	QU_ITERATOR_NEW_EXTERN (i, self, _string_next);
	return i;
}



/*) sub __eq (x) -> Boolean
	If v{x} is a String and the bytes are the same as self then return
	true, return false otherwise.

	@noexc

	@xmp
		a = 'abcde'
		b = 'Abcde'
		println (a == b)

	@out
		false
*/
static qu_boolean qu__string__eq (qu_string self, qu_any x)
{
	if (qu_string_is (x) && self->len == ((qu_string) x)->len)
	{
		if (memcmp (self->str, ((qu_string) x)->str, self->len) == 0)
			return QuTrue;
	}
	return QuFalse;
}



/*) sub __format (fmt: String) -> String
	Format this String.

	Refer to -{Number.format} for details of the format.

	The String class recognizes the following extension modes:
	@ops
	.| C{s} | print the whole string as is
	.| C{c} | print the first character
	.| C{U} | print the string in uppercase letters
	.| C{L} | print the string in lowercase letters
	.| C{h} | use #{htescape}
	.| C{q} | escape C{'} to C{\'} and C{"} to C{\"}

	@exc
	.EArgValue
	.EFormat

	@xmp
		printfln ("|%-20s|", "foo")

		# translates to:
		# print ('|')
		# print ("foo".__format ('%-20s'))
		# print ('|')
		# println

	@out
		|foo                 |
*/
static qu_string qu__string__format (qu_string self, qu_string x)
{
	qu_format_s f;

	QU_ARG_TYPE (x, qu_string);
	QU_ENSURE (qu_format_parse (x->str, &f));
	return qu_string_format (self, &f);
}



/*) sub __getitem (x) -> String
	@def
	sub __getitem (i: Integer) -> String
	sub __getitem (r: Range  ) -> String
	sub __getitem (a: Array  ) -> String
	@p
	Get a byte or slice of this String.

	If v{x} is an integer, get the character at that index.
	In this case v{i} must be a valid index within the string.
	A negative value means counting from the end.

	If v{x} is a -{Range} then collect bytes as specified by the Range.
	If an item in the range is out of bound then it is ignored.
	No exception thrown.

	If v{x} is an -{Array} then collect bytes as specified by the v{indexes}
	thus the same as #{bytes}.

	@exc
	.EArgValue
	.EIndex

	@xmp
		s = 'abcdefghijklmn'
		println (s [1] ' ' s [-2])
		println (s [Range (8, -1, -2)])
		println (s [[1, 3, 10]])

	@out
		b m
		igeca
		['b', 'd', 'k']
*/
static qu_string qu__string_bytes (qu_string self, int argc, qu_any *argv);
static qu_any qu__string__getitem (qu_string self, qu_any x)
{
	long i = 0;
	long l = self->len;

	/* the odd looking lable 'doint' is intentional since testing qu_int_is is
		faster than qu_classof, and qu_int_tol is faster than qu_integer_toli,
		and qu_int is usually the case */

	if (qu_int_is (x))
	{
		i = qu_int_tol (x);

	  doint:
		QU_ENSURE_INDEX (i, self->len);
		return qu_string_byte (self->str[i]);
	}
	else if (qu_range_is (x))
	{
		long n;

		QU_REQUIRE (QuEIndex, qu_range_is (x));

		n = qu_range_cover ((qu_range) x, l, &i);

		if (n == 0)
			return QuStringEMPTY;
		else if (n == 1)
			return qu_string_byte (self->str[i]);
		else
		{
			qu_string  r = qu_string_newl (n);
			long       z = ((qu_range) x)->step;
			char      *q = r->str;
			char      *p = self->str;

			for (; n; i += z, n--)
				*q++ = p[i];

			return r;
		}
	}
	else if (qu_array_is (x))
	{
		return qu__string_bytes (self, ((qu_array) x)->len, ((qu_array) x)->objs);
	}
	else
	{
		i = qu_integer_toli (x);
		QU_REQUIRE (QuEIndex, errno == 0);
		goto doint;
	}
}



/*) sub __getslice (idx: Integer, x) -> Any
	@def
	sub __getslice (idx: Integer, num: Integer) -> String
	sub __getslice (idx: Integer, defval) -> Any
	@p
	Get at most v{num} characters starting from the specified index v{idx}
	or safely get a byte without throwing EIndex.

	In the first form, v{idx} must be an integer index within this String.
	A negative value means counting from the end.
	If v{num} is negative then get all characters starting from v{idx}.

	The second form returns v{defval} if the index is out of bound,
	otherwise the character at v{idx}.

	@exc
	.EIndex

	@xmp
		s = 'abcdef'
		println (s [2,])
		println (s [-4, 2])
		println (s [, 4])
		println (s [100, 'x'])

	@out
		cdef
		cd
		abcd
		x
*/
static qu_any qu__string__getslice (qu_string self, qu_integer idx, qu_any num)
{
	long l = self->len;
	long i, n;

	i = qu_integer_toli (idx);
	QU_REQUIRE (QuEIndex, errno == 0);

	n = qu_integer_toli (num);
	if (errno == EDOM)
	{
		if (i < 0)
			i += l;
		if (i < 0 || i > l)
			return num;
		return qu_string_byte (self->str[i]);
	}
	else
	{
		QU_REQUIRE (QuEIndex, errno == 0);
		QU_ENSURE_INDEX (i, l);
		l -= i;
		if (n < 0 || n > l)
			n = l;
		return qu_string_new (self->str + i, n);
	}
}



/*) sub __hash -> Int
	Return the hash value of this String.

	@noexc

	@xmp
		s = 'abc'
		println (s.__hash)

	@out
		108966
*/
static qu_int qu__string__hash (qu_string self)
{
	if (self->hash == -1)
		self->hash = qu_strz_hash (self->str, self->len);
	return qu_int_new (self->hash);
}



/*) sub __iand (x: String) -> String
	Return a new String containing bytes which are both in this String and v{x}.
	Thus the intersection of self and v{x}.

	The result is sorted.

	@exc
	.EIand

	@xmp
		a = 'bacd'
		b = 'gfba'
		println (a & b)

	@out
		ab
*/
static qu_string qu__string__iand (qu_string self, qu_string x)
{
	unsigned char bs [256], bx [256], i, is, ix, j, ms, mx, m, c = 0;
	long       l;
	qu_string  r;
	char      *s;

	QU_REQUIRE (QuEIand, qu_string_is (x));

	memset (bs, '\0', sizeof (bs));
	memset (bx, '\0', sizeof (bx));

	qu_strz_set_cset ((unsigned char *) self->str, self->len, bs, &is, &ms);
	qu_strz_set_cset ((unsigned char *) x->str, x->len, bx, &ix, &mx);

	j = (is < ix) ? is : ix;
	m = (ms > mx) ? ms : mx;

	l = 0;
	for (i = j; i <= m; ++i)
	{
		if (bs[i] && bx[i])
		{
			c = i;
			++l;
		}
	}

	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
		return qu_string_byte (c);

	r = qu_string_newl (l);
	s = r->str;
	for (i = j; i <= m; ++i)
		if (bs[i] && bx[i])
			*s++ = i;
	return r;
}



/*) sub __in (x) -> Boolean
	Return true if this String contains v{x}, return false otherwise.

	v{x} may be a String or an Array, Set or Dict that yields String objects.
	Non String items/keys are ignored. In case of a Dict then the keys are used
	and the values ignored.

	@exc
	.EArgValue

	@xmp
		s = 'foo'
		println (s.startwith ('fo'))

	@out
		fo

	@see
	#{endwith}
*/
static qu_boolean qu__string__in (qu_string self, qu_any x)
{
	if (qu_string_is (x))
	{
		if (qu_strz_str (self->str, self->len, ((qu_string) x)->str, ((qu_string) x)->len))
			return QuTrue;
	}
	else
	{
		char   *z = self->str;
		long    l = self->len;
		qu_any  v;

		while (l)
		{
			if (qu_strz_startwith (z, l, 0, x, &v))
				return QuTrue;
			++z;
			--l;
		}
	}
	return QuFalse;
}



/*) sub __ior (x: String) -> String
	Return a new String containing all characters in this String and v{x}.
	Thus the union of self and v{x}.

	The result is sorted.

	@exc
	.EIor

	@xmp
		a = 'cba'
		b = 'egf'
		println (a | b)

	@out
		abcefg
*/
static qu_string qu__string__ior (qu_string self, qu_string x)
{
	unsigned char buf [256], i, is, ix, j, ms, mx, m, c = 0;
	long       l;
	qu_string  r;
	char      *s;

	QU_REQUIRE (QuEIor, qu_string_is (x));

	memset (buf, '\0', sizeof (buf));

	qu_strz_set_cset ((unsigned char *) self->str, self->len, buf, &is, &ms);
	qu_strz_set_cset ((unsigned char *) x->str, self->len, buf, &ix, &mx);

	j = (is < ix) ? is : ix;
	m = (ms > mx) ? ms : mx;

	l = 0;
	for (i = j; i <= m; ++i)
	{
		if (buf[i])
		{
			c = i;
			++l;
		}
	}

	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
		return qu_string_byte (c);

	r = qu_string_newl (l);
	s = r->str;
	for (i = j; i <= m; ++i)
		if (buf[i])
			*s++ = i;
	return r;
}



/*) sub __ixor (x: String) -> String
	Return a new String containing all characters which are in this String
	but not in v{x} and in v{x} but not in this String.
	Thus the difference of self and v{x}.

	The result is sorted.

	@exc
	.EIxor

	@xmp
		a = 'dcba'
		b = 'baefg'
		println (a ^ b)

	@out
		cdefg
*/
static qu_string qu__string__ixor (qu_string self, qu_string x)
{
	unsigned char bs [256], bx [256], i, is, ix, j, ms, mx, m, c = 0;
	long       l;
	qu_string  r;
	char      *s;

	QU_REQUIRE (QuEIxor, qu_string_is (x));

	memset (bs, '\0', sizeof (bs));
	memset (bx, '\0', sizeof (bx));

	qu_strz_set_cset ((unsigned char *) self->str, self->len, bs, &is, &ms);
	qu_strz_set_cset ((unsigned char *) x->str, x->len, bx, &ix, &mx);

	j = (is < ix) ? is : ix;
	m = (ms > mx) ? ms : mx;

	l = 0;
	for (i = j; i <= m; ++i)
	{
		if (bs[i])
		{
			if (!bx[i])
			{
				c = i;
				++l;
			}
		}
		else if (bx[i])
		{
			c = i;
			++l;
		}
	}

	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
		return qu_string_byte (c);

	r = qu_string_newl (l);
	s = r->str;
	for (i = 0; i <= m; ++i)
	{
		if (bs[i])
		{
			if (!bx[i])
				*s++ = i;
		}
		else if (bx[i])
		{
			*s++ = i;
		}
	}
	return r;
}



/*) sub __len -> Int
	Return the number of bytes in this String.

	@noexc

	@xmp
		s = 'foo'
		println (s.__len)

	@out
		3
*/
/*) sub len -> Int
	This is an alias of #{__len}.
*/
static qu_int qu__string__len (qu_string self)
{
	assert (self->len <= QU_INT_MAX);
	return qu_int_new (self->len);
}



/*) sub __like (x) -> Boolean
	Return true if v{x} is a String and its length and characters
	are case-insensitive the same as this String, false otherwise.

	@noexc

	@xmp
		println ('ABC' ==  'abc')
		println ('ABC' === 'abc')

	@out
		false
		true
*/
static qu_boolean qu__string__like (qu_string self, qu_any x)
{
	return (qu_string_is (x)
			&& self->len == ((qu_string) x)->len
			&& qu_strz_eqnc (self->str, ((qu_string) x)->str, self->len))
		? QuTrue : QuFalse;
}



/*) sub __lshift (n: Integer) -> String
	Return a new String by excluding the first v{n} characters in this String.

	v{n} must be a non-negative -{Integer}.

	@exc
	.ELshift

	@xmp
		println ('abcdefg' << 2)

	@out
		bcdefg
*/
static qu_string qu__string__lshift (qu_string self, qu_integer num)
{
	long l = self->len;
	long n = qu_integer_touli (num);

	QU_REQUIRE (QuELshift, errno == 0);
	if (n >= l)
		n = l;
	return qu_string_new (self->str + n, l - n);
}



/*) sub __minus (x: String) -> String
	Return a new String containing characters in this String which are not in v{x}.

	The result is sorted.

	@exc
	.EMinus

	@xmp
		a = 'abcde'
		b = 'ae'
		println (a - b)

	@out
		bcd
*/
static qu_string qu__string__minus (qu_string self, qu_string x)
{
	unsigned char bs [256], bx [256], i, is, ix, j, ms, mx, m, c = 0;
	long       l;
	qu_string  r;
	char      *s;

	QU_REQUIRE (QuEMinus, qu_string_is (x));

	memset (bs, '\0', sizeof (bs));
	memset (bx, '\0', sizeof (bx));

	qu_strz_set_cset ((unsigned char *) self->str, self->len, bs, &is, &ms);
	qu_strz_set_cset ((unsigned char *) x->str, x->len, bx, &ix, &mx);

	j = (is < ix) ? is : ix;
	m = (ms > mx) ? ms : mx;

	l = 0;
	for (i = j; i <= m; ++i)
	{
		if (bs[i] && !bx[i])
		{
			c = i;
			++l;
		}
	}

	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
		return qu_string_byte (c);

	r = qu_string_newl (l);
	s = r->str;
	for (i = j; i <= m; ++i)
	{
		if (bs[i] && !bx[i])
			*s++ = i;
	}
	return r;
}




/*) sub __mul (n: Integer) -> String
	Return a new String which is a v{n} repetitions of this String.

	v{n} must be a non-negative integer. If it is C{0] then an empty string
	is returned. If it is C{1} then self is returned.

	@exc
	.EMul
	.EMemory

	@xmp
		println ('foo' * 2)

	@out
		foofoo
*/
static qu_string qu__string__mul (qu_string self, qu_integer nn)
{
	long l = self->len;
	long n;

	n = qu_integer_toli (nn);
	QU_REQUIRE (QuEMul, errno == 0);

	if (n == 0 || l == 0)
	{
		return QuStringEMPTY;
	}
	else if (n == 1)
	{
		return self;
	}
	else
	{
		qu_string  s;
		char      *i, *p;

		QU_REQUIRE (QuEMul, n > 0);
		QU_MEM_CHECK (n < QU_MEM_MAX / l);

		s = qu_string_newl (n * l);
		p = s->str;
		i = self->str;
		while (n--)
		{
			memcpy (p, i, l);
			p += l;
		}
		return s;
	}
}



/*) sub __plus (x: String) -> String
	Return a new string as the result of concatenating this String with v{x}.

	@exc
	.EPlus
	.EMemory

	@xmp
		a = 'abc'
		b = 'bcde'
		println (a + b)

	@out
		abcbcde
*/
static qu_string qu__string__plus (qu_string self, qu_string x)
{
	QU_REQUIRE (QuEPlus, qu_string_is (x));

	if (x->len == 0)
	{
		return self;
	}
	else if (self->len == 0)
	{
		return x;
	}
	else
	{
		qu_string s;

		QU_MEM_CHECK (self->len < QU_MEM_MAX - x->len);
		s = qu_string_newl (self->len + x->len);
		memcpy (s->str, self->str, self->len);
		memcpy (s->str + self->len, x->str, x->len);
		return s;
	}
}



/*) sub __print -> Null
	Print this String to v{stdout}.

	See -{print} for possible exceptions.
*/
qu_null qu__string__print (qu_string self)
{
	return qu_sys_printz (self->str, self->len);
}



/*) sub __rem (rep) -> String
	Replace sequences of H{$}{xxx} with the corresponding values in v{rep}.

	v{rep} may be an Array, Dict or Record. Anything else is treated as
	a single value and used repeatedly.

	Non-string values are converted to strings by means of their __string method.
	Values not found are substituted with C{?}.

	@noexc

	@xmp
		s = 'My name is ${name} and I am ${age} years old.'

		x = s % ['Fooey', 99]
		println (x)

		x = s % [name: 'Fooey', age: 99]
		println (x)

		struct Person
			name: String
			age : Int
		;;

		x = s % Person ('Fooey', 99)
		println (x)

		x = s % 'Fooey'
		println (x)

	@out
		My name is Fooey and I am 99 years old.
		My name is Fooey and I am 99 years old.
		My name is Fooey and I am 99 years old.
		My name is Fooey and I am Fooey years old.

	@note
	You might want to check out the %{Template} module.
*/
static qu_string qu__string__rem (qu_string self, qu_any x)
{
	qu_string  r = qu_string_buffer_new (8);
	long       l = self->len;
	char      *i = self->str;
	char      *j = i;
	qu_string  y;

	if (qu_array_is (x))
	{
		long k = 0;

		while (l--)
		{
			if (i[0] == '$' && i[1] == '{')
			{
				if (i > j)
					qu_string_buffer_put (r, j, i - j);

				if (k < ((qu_array) x)->len)
				{
					y = ((qu_array) x)->objs[k++];
					if (!qu_string_is (y))
						y = qu_sys_string (y);
					qu_string_buffer_puts (r, y);
				}
				else
				{
					qu_string_buffer_putc (r, '?');
				}

				i += 2;
				l -= 2;
				while (l && *i != '}')
				{
					--l;
					++i;
				}
				j = ++i;
			}
			else
			{
				++i;
				continue;
			}
		}
		if (i > j)
			qu_string_buffer_put (r, j, i - j);
	}
	else if (qu_dict_is (x))
	{
		while (l--)
		{
			if (i[0] == '$' && i[1] == '{')
			{
				if (i > j)
					qu_string_buffer_put (r, j, i - j);

				i += 2;
				l -= 2;
				j = i;
				while (l && *i != '}')
				{
					--l;
					++i;
				}

				y = qu_dict_getitem ((qu_dict) x, qu_string_new (j, i - j));
				if (y)
				{
					if (!qu_string_is (y))
						y = qu_sys_string (y);
					qu_string_buffer_puts (r, y);
				}
				else
				{
					qu_string_buffer_putc (r, '?');
				}
				j = ++i;
			}
			else
			{
				++i;
				continue;
			}
		}
		if (i > j)
			qu_string_buffer_put (r, j, i - j);
	}
	else if (qu_record_is (x))
	{
		long      k;
		qu_struct u = ((qu_record) x)->struc;

		while (l--)
		{
			if (i[0] == '$' && i[1] == '{')
			{
				if (i > j)
					qu_string_buffer_put (r, j, i - j);

				i += 2;
				l -= 2;
				j = i;
				while (l && *i != '}')
				{
					--l;
					++i;
				}

				k = qu_struct_index (u, qu_string_new (j, i - j));
				if (k >= 0)
				{
					y = ((qu_record) x)->objs[k];
					if (!qu_string_is (y))
						y = qu_sys_string (y);
					qu_string_buffer_puts (r, y);
				}
				else
				{
					qu_string_buffer_putc (r, '?');
				}
				j = ++i;
			}
			else
			{
				++i;
				continue;
			}
		}
		if (i > j)
			qu_string_buffer_put (r, j, i - j);
	}
	else
	{
		y = qu_string_is (x) ? (qu_string) x : qu_sys_string (x);
		while (l--)
		{
			if (i[0] == '$' && i[1] == '{')
			{
				if (i > j)
					qu_string_buffer_put (r, j, i - j);

				i += 2;
				l -= 2;
				j = i;
				while (l && *i != '}')
				{
					--l;
					++i;
				}
				qu_string_buffer_puts (r, y);
				j = ++i;
			}
			else
			{
				++i;
				continue;
			}
		}
		if (i > j)
			qu_string_buffer_put (r, j, i - j);
	}
	return qu_string_buffer_return (r);
}



/*) sub __rshift (n: Integer) -> String
	Return a new String by excluding the last v{n} characters of this String.

	v{n} must be a non-negative -{Integer}.

	@exc
	.ERshift

	@xmp
		println ('abcdefg' >> 2)

	@out
		abcde
*/
static qu_string qu__string__rshift (qu_string self, qu_integer num)
{
	long l = self->len;
	long n = qu_integer_touli (num);

	QU_REQUIRE (QuERshift, errno == 0);
	if (n > l)
		n = l;
	return qu_string_new (self->str, l - n);
}



/*) sub __string -> self
	This method simply returns self.
*/
static qu_string qu__string__string (qu_string self)
{
	return self;
}



/*) sub __true -> Boolean
	Return true if this String is not empty, false otherwise.

	@noexc

	@xmp
		s = ''
		println (s ? 'yep' : 'nope')

	@out
		nope
*/
static qu_boolean qu__string__true (qu_string self)
{
	return self->len ? QuTrue : QuFalse;
}



/*----------------------------------------------------------------------------
	specific instance methods
----------------------------------------------------------------------------*/

/*) sub asc -> Int
	Get the signed 8-bits of the first character.

	The returned value is an Int in -128..127.

	See #{ord} to get the unsigned value.

	@exc
	.EArgValue

	@xmp
		c = '\xff'
		println (c.asc)

	@out
		-128
*/
qu_int qu__string_asc (qu_string self)
{
	return qu_int_newi ((int) self->str[0]);
}



/*) sub bytes (*indexes) -> String
	Collect bytes specified by v{indexes} into a new String.

	v{indexes} may be integer indexes or Range objects.

	@exc
	.EIndex
	.EArgValue
	.EMemory

	@xmp
		println ('abcdefg'.bytes (0, 4, 5, 2..3))

	@out
		aefcd
*/
static qu_string qu__string_bytes (qu_string self, int argc, qu_any *argv)
{
	qu_string  a = qu_string_buffer_new (argc);
	long       l = self->len;
	char      *s = self->str;

	while (argc--)
	{
		qu_any x = *argv++;
		long   i;

		if (qu_int_is (x))
		{
			i = qu_int_tol (x);

		  doint:
			QU_ENSURE_INDEX (i, l);
			QU_ENSURE (qu_string_buffer_putc (a, s[i]));
		}
		else if (qu_range_is (x))
		{
			long n = qu_range_cover ((qu_range) x, self->len, &i);
			if (n > 0)
			{
				QU_ENSURE (qu_string_buffer_make_space (a, n));
				{
					long  z = ((qu_range) x)->step;
					char *q = a->str + a->len;

					a->len += n;
					for (; n; i += z, n--)
						*q++ = s[i];
				}
			}
		}
		else
		{
			i = qu_integer_toli (x);
			QU_REQUIRE (QuEIndex, errno == 0);
			goto doint;
		}
	}
	return qu_string_buffer_return (a);
}



/*) sub capitalize -> String
	Return a new string with the first non-space character capitalized.

	@noexc

	@xmp
		x = '  abc'
		y = x.capitalize
		println (y.quote)

	@out
		'   Abc'
*/
static qu_string qu__string_capitalize (qu_string self)
{
	long l = self->len;

	if (l == 0)
		return QuStringEMPTY;

	else if (l == 1)
		return qu_string_byte (qu_ascii_toupper (self->str[0]));

	else
	{
		char *p = self->str;

		while (l && qu_ascii_isspace (*p))
		{
			l--;
			p++;
		}
		if (l && qu_ascii_islower (*p))
		{
			qu_string s = qu_string_dup (self);

			s->str [p - self->str] = qu_ascii_toupper (*p);
			return s;
		}
		return self;
	}
}



/*) sub center (width: Integer, fill = ' ') -> String
	Return a new String centered in v{width} by padding v{fill}.

	@exc
	.EArgValue

	@xmp
		s = 'foo'
		println (s.center(5).quote)

	@out
		' foo '
*/
static qu_string qu__string_center (qu_string self, int argc, qu_any *argv)
{
	long width, left, marg;
	char fill = ' ';

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong (0, width);
	if (argc > 1)
	{
		qu_string x = argv[1];
		QU_ARG_CHECK (qu_string_is (x) && x->len);
		fill = x->str[0];
	}

	if (self->len >= width)
		return self;

	marg = width - self->len;
	left = marg / 2 + (marg & width & 1);
	return qu_string_pad (self, left, marg - left, fill);
}



/*) sub chop (*arg) -> String
	@def
	sub chop -> String
	sub chop (what, nocase = false) -> String
	sub chop (idx: Integer, num: Integer) -> String
	sub chop (num: Integer) -> String
	@p
	Return a new String by chopping one or more characters.

	The first form chops any single last character.

	The second form chops any last string specified by v{what}.
	Here v{what} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	The third form takes a starting index and the number of bytes to chop.

	The fourth form takes a number of bytes to chop.
	A positive value means to chop bytes from the end while a negative value
	means to chop bytes from the start.

	@exc
	.EArgValue
	.EIndex

	@xmp
		s = 'fooxyz'
		println (s.chop)
		println (s.chop ('z'))
		println (s.chop ('x'))
		println (s.chop (['foo', 'bar', 'xyz']))
		println (s.chop (3, 1))
		println (s.chop (-3))

	@out
		fooxy
		fooxy
		fooxyz
		foo
		fooyz
		xyz
*/
static qu_string _string_chop (qu_string self, long n)
{
	long l = self->len;

	if (n < 0)
	{
		n = -n;
		if (n < l)
			return qu_string_new (self->str + n, l - n);
	}
	else
	{
		if (n < l)
			return qu_string_new (self->str, l - n);
	}
	return QuStringEMPTY;
}



static qu_string _string_splice (qu_string self, long i, long n)
{
	assert (0 <= i && i < self->len);
	assert (n >= 0);

	if (n == 0)
	{
		return self;
	}
	else
	{
		long l = self->len;
		long m = l - i;

		if (n >= m)
		{
			return qu_string_new (self->str, i);
		}
		else
		{
			qu_string r;

			l -= n;
			r = qu_string_newl (l);
			memcpy (r->str, self->str, i);
			memcpy (r->str + i, self->str + i + n, l - i);
			return r;
		}
	}
}



static qu_string _string_chop_spec (qu_string self, qu_any spec, int nocase)
{
	qu_any    v;
	qu_string x = qu_strz_endwith (self->str, self->len, nocase, spec, &v);

	if (x == 0)
		return self;
	return qu_string_new (self->str, self->len - x->len);
}



static qu_string qu__string_chop (qu_string self, int argc, qu_any *argv)
{
	if (argc == 0)
		return _string_chop (self, 1);
	else
	{
		qu_any x = argv[0];

		if (qu_integer_is (x))
		{
			if (argc == 1)
			{
				long n;

				QU_ARG_long (x, n);
				return _string_chop (self, n);
			}
			else
			{
				long i, n, l = self->len;

				QU_ARG_VEXACT   (2);
				QU_ARG_long     (x, i);
				QU_ARG_vlong    (1, n);
				QU_ARG_CHECK    (n >= 0);
				QU_ENSURE_INDEX (i, l);
				return _string_splice (self, i, n);
			}
		}
		else
		{
			int nocase = 0;

			QU_ARG_VMAX (2);
			if (argc > 1)
				QU_ARG_vbool (1, nocase);

			return _string_chop_spec (self, x, nocase);
		}
	}
}



/*) sub chomp (all = false) -> String
	Return a new String by removing one ore all linbreaks at the end of self.

	The linebreak is either C{'\n'} or C{'\r\n'}.

	@exc
	.EArgValue

	@xmp
		println ('123\n\r\n'.chomp.quote)
		println ('123\n\n'.chomp (true))

	@out
		'123\n'
		123
*/
static qu_string qu__string_chomp (qu_string self, int argc, qu_any *argv)
{
	int   all = 0;
	long  l = self->len;
	char *s = self->str + l - 1;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vbool (0, all);

	while (l && *s == '\n')
	{
		--l;
		--s;
		if (l && *s == '\r')
		{
			--l;
			--s;
		}
		if (!all)
			break;
	}
	if (l == self->len)
		return self;
	return qu_string_new (self->str, l);
}



/*) sub contains (x, nocase = true) -> Bool
	Whether the String contains v{x}, optionally case insensitive.

	v{x} can be anything. If it is not a String then it's __string method
	is called.

	@exc
	.EArgValue

	@xmp
		s = 'abc123def'
		println (s.contains (123))

	@out
		true
*/
static qu_boolean qu__string_contains (qu_string self, int argc, qu_any *argv)
{
	qu_stringy what;
	int        nocase = 1;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];

	if (!qu_stringy_is (what))
		what = (qu_stringy) qu_sys_string (what);

	if (argc > 1)
		QU_ARG_vbool (1, nocase);

	if (nocase)
	{
		if (qu_strz_strnc (self->str, self->len, what->str, what->len))
			return QuTrue;
	}
	else
	{
		if (qu_strz_str (self->str, self->len, what->str, what->len))
			return QuTrue;
	}
	return QuFalse;
}



/*) sub count (*arg)
	@def
	sub count (what, nocase = false) -> Int
	sub count (pattern: Regex, *arg) -> Int
	@p
	Count strings in the String.

	v{what} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	@exc
	.EArgValue

	@xmp
		s = 'abxdyfXgYh'
		println (s.count ({'x', 'y'}, true))

	@out
		4

	@p
	The second form expects a %{Regex} object.
	It simply calls C{pattern.count (self, *arg)}.

	@xmp
		use Regex

		r = Regex ('[abc]')
		s = 'foobar'
		println (s.count (r))

	@out
		2
*/
static qu_int qu__string_count (qu_string self, int argc, qu_any *argv)
{
	qu_any what;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];

	if (qu_classid (what) == QU_CLASS_ID_Regex)
	{
		static qu_string name = 0;

		if (name == 0)
			name = qu_string_new_staticz ("count");

		if (argc == 1)
			return qu_object_call1 (what, name, self);
		else
			return qu_object_call2 (what, name, self, argv[1]);
	}
	else
	{
		int     nocase = 0;
		long    l = self->len;
		char   *i = self->str;
		qu_any  v;
		long    count = 0;

		if (argc > 1)
			QU_ARG_vbool (1, nocase);

		while (l)
		{
			qu_string x = qu_strz_startwith (i, l, nocase, what, &v);
			if (x)
			{
				++count;
				l -= x->len;
				i += x->len;
			}
			else
			{
				--l;
				++i;
			}
		}
		return qu_int_new (count);
	}
}



/*) sub eachline -> Iterator
	Return a -{Iterator} object to iterate through each line.

	The lines returned do not contain the linebreak.

	@noexc

	@xmp
		s = 'abc\nde\nfgh\n'
		for l in s.eachline
			println (l)
		;;

	@out
		abc
		de
		fgh
*/
static qu_string _string_eachline_next (qu_string self, qu_iterator i)
{
	long p = i->ex1;
	long n = self->len - p;

	QU_REQUIRE (QuEStop, n > 0);
	/*begin*/
	{
		char *s = self->str;
		char *q = s + p;
		char *j = q;

		while (n && *j != '\n')
		{
			n--;
			j++;
		}
		n = j - q;
		i->ex1 = p + n + 1;
		return qu_string_new (q, n);
	}
}



static qu_iterator qu__string_eachline (qu_string self)
{
	register qu_iterator i;
	QU_ITERATOR_NEW_EXTERN (i, self, _string_eachline_next);
	return i;
}



/*) sub eachsection (marker: String) -> Iterator
	Return a -{Iterator} object to iterate through each section in self.

	Sections are divided by a line that starts with a special v{marker}
	character optionally preceeded by whitespaces.

	The Strings yielded are right-trimmed.
	Only non-empty Strings are yielded.

	@exc
	.EArgValue

	@xmp
		s ='foo \n  \t. \n  . bar \n  .baz '
		for x in s.eachsection ('.')
			println (x.quote)
		;;

	@out
		'foo'
		'.  bar'
		'.baz'
*/
static qu_string _string_eachsection_next (qu_string self, qu_iterator i)
{
	long  p, l;
	char *s, *e, *q, *j, c;

  again:

	p = i->ex1;
	l = self->len;

	QU_REQUIRE (QuEStop, p < l);

	c = (char) i->ex2;
	s = self->str;
	e = s + l;
	q = s + p;
	j = q;

	if (p == 0)
	{
		while (j < e && qu_ascii_isspace (*j))
			j++;
		q = j;
	}
	while (j < e)
	{
		if (*j++ == '\n')
		{
			while (j < e && qu_ascii_iswhite (*j))
				j++;
			if (*j == c)
				break;
		}
	}
	i->ex1 = j - s;
	while (j > q && qu_ascii_isspace (j[-1]))
		--j;
	if (j <= q || (*q == c && j == q + 1))
		goto again;
	return qu_string_new (q, j - q);
}



static qu_iterator qu__string_eachsection (qu_string self, qu_string marker)
{
	register qu_iterator i;

	QU_ARG_CHECK (qu_string_is (marker) && marker->len);
	QU_ITERATOR_NEW_EXTERN (i, self, _string_eachsection_next);
	i->ex2 = marker->str[0];
	return i;
}



/*) sub eachsegment (len: Integer) -> Iterator
	Return an -{Iterator} to iterate through segments which are v{len} long each.

	v{len} must be a positive integer.

	@exc
	.EArgValue

	@xmp
		for i in 'abcdefghij'.eachsegment (4)
			println (i)
		;;

	@out
		abcd
		efgh
		ij
*/
static qu_string _string_eachsegment_next (qu_string self, qu_iterator i)
{
	long p = i->ex1;
	long l = self->len;

	QU_REQUIRE (QuEStop, p < l);
	/*begin*/
	{
		long len = i->ex2;
		long max = l - p;

		if (len > max)
			len = max;
		i->ex1 += len;
		return qu_string_new (self->str + p, len);
	}
}



static qu_iterator qu__string_eachsegment (qu_string self, qu_integer len)
{
	register qu_iterator i;
	long l;

	QU_ARG_long  (len, l);
	QU_ARG_CHECK (l > 0);
	QU_ITERATOR_NEW_EXTERN (i, self, _string_eachsegment_next);
	i->ex2 = l;
	return i;
}



/*) sub endwith (what, nocase = false) -> Any
	Whether or not the String ends with v{what}, optionally case insensitive.

	Here v{what} may be a String or an Array, Set or Dict that yields String objects.
	Non String items are ignored. In case of a Dict then the keys are used and the
	corresponding value returned.

	Returns nil if none matches, the matched item otherwise.

	@exc
	.EArgValue

	@xmp
		s = 'foo'
		println (s.endwith ('oo'))

		s = 'foo.PNG'
		println (s.endwith (['.gif', '.jpeg', '.png'], true))

	@out
		oo
		.png

	@see
	#{startwith} can also start at a specified index
*/
static qu_any qu__string_endwith (qu_string self, int argc, qu_any *argv)
{
	int    nocase = 0;
	qu_any what, v;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];
	if (argc > 1)
		QU_ARG_vbool (1, nocase);
	what = qu_strz_endwith (self->str, self->len, nocase, what, &v);
	if (what)
		return v ? v : what;
	return QuNil;
}



/*) sub endswith (what, nocase = false) -> Boolean
	This method wraps #{endwith} and returns false if it returns nil,
	true otherwise.

	This method may be useful in switch statements.

	@exc
	.EArgValue

	@xmp
		s = 'abcde'
		switch s
			case .endswith ('de')
				println (1)
				break

			case .startswith ('ab')
				println (2)
				break
		;;

	@out
		1
*/
static qu_any qu__string_endswith (qu_string self, int argc, qu_any *argv)
{
	qu_any x = qu__string_endwith (self, argc, argv);

	QU_ENSURE (x);
	return (x == (qu_any) QuNil) ? QuFalse : QuTrue;
}



/*) sub extract (opener: String, closer = opener) -> Array
	Extract parts of self enclosed by v{opener} and v{closer}.

	Both v{closer} and v{opener} must be non-empty Strings.

	The result is an -{Array} containing all the extracted String objects trimmed.

	@exc
	.EArgValue
	.EMemory

	@xmp
		s = ' {: xxxx :} zzzz {: yyyy :} '
		a = s.extract ('{:', ':}')
		for s in a
			println ('> 's' <')
		;;
		println

		s = ' :: xxxx :: zzzz :: yyyy :: '
		a = s.extract ('::')
		for s in a
			println ('> 's' <')
		;;

	@out
		|xxxx|
		|yyyy|

		|xxxx|
		|zzzz|
		|yyyy|
*/
static qu_array qu__string_extract (qu_string self, int argc, qu_any *argv)
{
	qu_string opener, closer = 0;
	int       same = 1;

	QU_ARG_VCOUNT (1, 2);

	opener = argv[0];
	QU_ARG_CHECK (qu_string_is (opener) && opener->len);

	if (argc > 1)
	{
		closer = argv[1];
		QU_ARG_CHECK (qu_string_is (closer) && closer->len);
		same = qu_string_EQ (opener, closer);
	}
	else
	{
		closer = opener;
	}

	/*begin*/
	{
		long      l = self->len;
		char     *s = self->str;
		qu_array  a = qu_array_new (8);
		long      o = opener->len;
		long      c = closer->len;
		char     *u = opener->str;
		char     *v = closer->str;
		long      i = 0;
		long      d = same ? 0 : c;
		long      j;

		while (i < l - o)
		{
			i = qu_strz_index (s, l, u, o, i);
			if (i < 0 || i >= l - c)
				break;

			i += o;
			j = qu_strz_index (s, l, v, c, i);
			if (j < 0)
				break;
			qu_array_push (a, qu_string_new_trim (s + i, j - i));
			i = j + d;
		}
		return a;
	}
}



/*) sub fixline (rep = '\n') -> String
	Return a new String by replacing all linebreak characters with v{rep}.

	All sequences of C{"\n"} and C{"\r\n"} are replaced.

	This method is more efficient than #{replace}.

	@exc
	.EArgValue

	@xmp
		s = 'foo\nbar\rbaz\r\nfox'
		s = s.fixline
		println (s.quote)

	@out
		'foo\nbar\rbaz\nfox'
*/
static qu_string qu__string_fixline (qu_string self, int argc, qu_any *argv)
{
	long       l = self->len;
	char      *s = self->str;
	long       n = 0;
	int        k = 0;
	qu_string  rep = qu_string_byte ('\n');
	long       lr;
	qu_string  r;
	char      *p, *q;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_VTYPE (0, rep, qu_string);

	lr = rep->len;
	while (l--)
	{
		switch (*s++)
		{
			case '\n':
				n += lr;
				k = 1;
				break;

			case '\r':
				if (l && *s == '\n')
				{
					--l;
					++s;
					n += lr;
					k = 1;
					break;
				}
				/*nobreak*/

			default:
				++n;
		}
	}

	if (n == 0)
		return QuStringEMPTY;

	if (k == 0)
		return self;

	r = qu_string_newl (n);
	p = r->str;
	l = self->len;
	s = self->str;
	q = rep->str;

	while (l--)
	{
		switch (*s++)
		{
			case '\n':
			  doit:
				memcpy (p, q, lr);
				p += lr;
				break;

			case '\r':
				if (l && *s == '\n')
				{
					--l;
					++s;
					goto doit;
				}
				/*nobreak*/
			default:
				*p++ = s[-1];
		}
	}

	if (n == 1)
		return qu_string_byte (r->str[0]);

	return r;
}



/*) sub flatten -> String
	Return a new String by converting linebreaks into spaces.

	@exc
	.EModify

	@xmp
		x = '1\n \r\n23'
		y = x.flatten
		println (y.quote)

	@out
		'1  \r 23'
*/
qu_string qu__string_flatten (qu_string self)
{
	qu_string  r = qu_string_new (self->str, self->len);
	char      *s = r->str;
	long       l = r->len;

	while (l--)
	{
		if (*s == '\n')
			*s = ' ';
		++s;
	}
	return r;
}



/*) sub head (sep = '', offset = 0, nocase = false) -> String
	Return the part of the string before the separator v{sep}.

	Here v{sep} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	If v{sep} is an empty string, it means any space.

	Optionally start at a specified v{offset}.
	If this is a negative value then count from the end.
	If the v{offset} falls out of this String then an empty String is returned,
	thus no EIndex thrown.

	If the separator is not found then self is returned.

	@exc
	.EArgValue

	@xmp
		s = '123\t 45678'
		println (s.head)

	@out
		123
*/
static qu_string qu__string_head (qu_string self, int argc, qu_any *argv)
{
	qu_any    sep = 0;
	int       nocase = 0;
	long      ofs = 0;
	char     *i = self->str;
	long      l = self->len;
	qu_string x;
	qu_any    v;

	QU_ARG_VMAX (3);
	if (argc > 0)
	{
		sep = argv[0];
		if (qu_stringy_is (sep) && ((qu_stringy) sep)->len == 0)
			sep = 0;

		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			if (argc > 2)
				QU_ARG_vbool (2, nocase);

			if (ofs < 0)
				ofs += l;
			if (ofs < 0)
				return QuStringEMPTY;
			if (ofs >= l)
				return self;
		}
	}
	i += ofs;
	l -= ofs;

	if (sep == 0)
	{
		while (l)
		{
			if (qu_ascii_isspace (*i))
				return qu_string_new (self->str, i - self->str);
			++i;
			--l;
		}
	}
	else
	{
		while (l)
		{
			x = qu_strz_startwith (i, l, nocase, sep, &v);
			if (x)
				return qu_string_new (self->str, i - self->str);
			++i;
			--l;
		}
	}
	return self;
}



/*) sub htescape -> String
	Return a new String by escaping HTML characters:
	@ul
	.C{<} to C{&lt;}
	.C{>} to C{&gt;}
	.C{;} to C{&amp;}
	.C{"} to C{&quot;}

	@exc
	.EMemory

	@xmp
		s = '<foo>'
		println (s.htescape)

	@out
		&lt;foo&gt;
*/
static qu_string qu__string_htescape (qu_string self)
{
	long       l = self->len;
	char      *s = self->str;
	long       n = 0;
	qu_string  r;
	char      *p;

	while (l--)
	{
		QU_MEM_CHECK (n <= QU_MEM_MAX - 8);
		switch (*s++)
		{
			case '<':
			case '>': n += 4; continue;
			case '&': n += 5; continue;
			case '"': n += 6; continue;
			default : n++;
		}
	}

	r = qu_string_newl (n);
	p = r->str;

	l = self->len;
	s = self->str;
	while (l--)
	{
		switch (*s)
		{
			case '<': s++; *p++ = '&'; *p++ = 'l'; *p++ = 't'; *p++ = ';'; continue;
			case '>': s++; *p++ = '&'; *p++ = 'g'; *p++ = 't'; *p++ = ';'; continue;
			case '&': s++; *p++ = '&'; *p++ = 'a'; *p++ = 'm'; *p++ = 'p'; *p++ = ';'; continue;
			case '"': s++; *p++ = '&'; *p++ = 'q'; *p++ = 'u'; *p++ = 'o'; *p++ = 't'; *p++ = ';'; continue;
			default : *p++ = *s++;
		}
	}
	return r;
}



/*) sub index (what, offset = 0, nocase = false) -> Int
	Return the index where v{what} starts in self, -1 if not found.

	Here v{what} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	Optionally start looking from v{offset} and ignore case.
	If v{offset} is negative then count from the end.
	If it falls out of this String then -1 is returned, thus no EIndex thrown.

	@exc
	.EArgValue

	@xmp
		println ('..xy..xy'.index ('xy'))
		println ('..XY..ab'.index (const ['xy', 'ab']))

	@out
		2
		6

	@see
	#{rindex}
*/
static qu_int qu__string_index (qu_string self, int argc, qu_any *argv)
{
	qu_any     what;
	long       ofs = 0;
	int        nocase = 0;
	long       l = self->len;
	char      *i;
	qu_string  x;
	qu_any     v;

	QU_ARG_VCOUNT (1, 3);
	what = argv[0];
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
			QU_ARG_vbool (2, nocase);

		if (ofs < 0)
			ofs += l;
		if (ofs < 0 || ofs >= l)
			return QuInt_1;
	}
	i = self->str + ofs;
	l -= ofs;
	while (l)
	{
		x = qu_strz_startwith (i, l, nocase, what, &v);
		if (x)
			return qu_int_new (i - self->str);
		++i;
		--l;
	}
	return QuInt_1;
}



/*) sub inset (set: String, nocase = false)
	Whether all characters of the string are in v{s}.

	@exc
	.EArgValue

	@xmp
		s = 'ABC'
		println ('abbbcca'.inset (s))
		println ('abbbcca'.inset (s, true))

	@out
		false
		true
*/
static qu_boolean qu__string_inset (qu_string self, int argc, qu_any *argv)
{
	qu_string x;
	int       nocase = 0;
	long      l = self->len;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, x, qu_string);
	if (argc > 1)
		QU_ARG_vbool (1, nocase);

	if (l && x->len)
	{
		/* I think that converting x into a 256-character set is slower
			since programmers generally supply short strings */

		char *z = self->str;
		char *s = x->str;
		long  n = x->len;

		if (nocase)
		{
			while (l--)
			{
				if (!qu_strz_chrnc (s, n, *z++))
					return QuFalse;
			}
		}
		else
		{
			while (l--)
			{
				if (!qu_strz_chr (s, n, *z++))
					return QuFalse;
			}
		}
		return QuTrue;
	}
	return QuFalse;
}



/*) sub level (size = -1) -> Int
	Return the indentation level of a string.

	A tab is always one level.

	If v{size} is a positive number then only v{size} repetition of spaces
	makes one level.

	@exc
	.EArgValue

	@xmp
		println ('foo'.level)
		println ('  foo'.level)
		println ('\t\t    foo'.level)
		println ('\t    \t      foo'.level (4))

	@out
		0
		0
		2
		4
*/
static qu_int qu__string_level (qu_string self, int argc, qu_any *argv)
{
	long  size = -1;
	char *i    = self->str;
	long  l    = self->len;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
		QU_ARG_vlong (0, size);
	if (size <= 1)
	{
		while (l-- && *i == '\t')
			++i;
		l = i - self->str;
		return qu_int_new (l);
	}
	else
	{
		long count = 0;
		while (l)
		{
			if (*i == '\t')
			{
				++i;
				--l;
				++count;
			}
			else
			{
				long n = size;
				while (n && *i == ' ')
				{
					--n;
					++i;
					--l;
				}
				if (n)
					break;
				++count;
			}
		}
		return qu_int_new (count);
	}
}



/*) sub linecount -> Int
	Return the number of lines.

	This method only counts C{'\n'} thus yields the same result for both
	C{'\n'} and C{'\r\n'} linebreaks.

	@noexc

	@xmp
		s = ' \n \n \r\n '
		println (s.linecount)

	@out
		3
*/
static qu_int qu__string_linecount (qu_string self)
{
	char *s = self->str;
	long  l = self->len;
	long  n = 0;

	while (l--)
	{
		if (*s++ == '\n')
			++n;
	}
	assert (n <= QU_INT_MAX);
	return qu_int_new (n);
}



/*) sub lines -> Array
	Return a new Array containing the lines.

	The lines do not contain the line break.

	This is more efficient than C{s.eachline as Array} and C{s.split ('\n')}.

	@noexc
		s = 'abc\nde\nfgh'
		a = s.lines
		println (a)

	@out
		['abc', 'de', 'fgh']
*/
static qu_array qu__string_lines (qu_string self)
{
	qu_array a = qu_array_new (8);
	long     n = self->len;
	char    *i = self->str;
	char    *j = i;

	while (n > 0)
	{
		while (n && *i != '\n')
		{
			--n;
			++i;
		}
		qu_array_push (a, qu_string_new (j, i - j));
		j = ++i;
		--n;
	}
	return a;
}



/*) sub ljust (width: Integer, fill = ' ') -> String
	Return a new left justified string of length v{width} by padding v{fill}.

	@exc
	.EArgValue

	@xmp
		s = 'foo'
		println (s.ljust (5).quote)

	@out
		'foo  '
*/
static qu_string qu__string_ljust (qu_string self, int argc, qu_any *argv)
{
	qu_string fill = qu_string_byte (' ');
	long      width;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong (0, width);
	if (argc > 1)
	{
		fill = argv[1];
		QU_ARG_CHECK (qu_string_is (fill) && fill->len);
	}
	if (self->len >= width)
		return self;

    return qu_string_pad (self, 0, width - self->len, fill->str[0]);
}



/*) sub match (what) -> String
	@def
	sub match (what, nocase= false) -> String
	sub match (pattern: Regex, *arg) -> String
	@p
	Find string in the String.

	Here v{what} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	The returned value is an empty String if no match found, the part of the String
	that matches otherwise.

	@exc
	.EArgValue

	@xmp
		s = 'FooBarBaz'
		r = s.match ({'baz', 'bar'}, true)
		println (r)
		println (s.index (r))

		r = s.match (['baz', 'bar'])
		println (r.quote)
		println (s.index (r))

	@out
		Bar
		3
		''
		-1

	@p
	The second form expects a %{Regex} object.
	It simply calls C{pattern.match (self, *arg)}.

	@xmp
		use Regex
		r = Regex (`[abc]x`)
		s = 'aaa ax xxx'
		println (s.match (r))

	@out
		ax
*/
static qu_string qu__string_match (qu_string self, int argc, qu_any *argv)
{
	qu_any what;

	QU_ARG_VCOUNT (1, 2);
	what = argv[0];

	if (qu_classid (what) == QU_CLASS_ID_Regex)
	{
		static qu_string name = 0;

		if (name == 0)
			name = qu_string_new_staticz ("match");

		if (argc == 1)
			what = qu_object_call1 (what, name, self);
		else
			what = qu_object_call2 (what, name, self, argv[1]);

		QU_ENSURE (what);

		/* this should never happen */
		QU_REQUIRE (QuEImplement, qu_string_is (what));

		return what;
	}
	else
	{
		int     nocase = 0;
		long    l = self->len;
		char   *i = self->str;
		qu_any  v;

		if (argc > 1)
			QU_ARG_vbool (1, nocase);

		while (l)
		{
			qu_string x = qu_strz_startwith (i, l, nocase, what, &v);
			if (x)
				return qu_string_new (i, x->len);
			--l;
			++i;
		}
		return QuStringEMPTY;
	}
}



/*) sub ord -> Int
	Get the unsigned 8-bits of the first character.

	The returned value is an Int in 0..255.

	See #{asc} to get the signed value.

	@noexc

	@xmp
		x = '\xff'
		println (x.ord)

	@out
		255
*/
qu_int qu__string_ord (qu_string self)
{
	int v = (int) ((unsigned char) self->str[0]);
	return qu_int_newi (v);
}



/*) sub parseform (*arg) -> Array
	@def
	sub parseform -> Array
	sub parseform (special: Dict) -> Array
	@p
	This parses a string in the form of:
	@prog
		"... X{xxx} ... X{xxx} ..."

	The items will be a pair of 2 values:
	@ul
	.nil followed by a String is the plain part
	.a single character tag followed by the content of the tag

	@p
	If you supply v{special} then the tag character will be replaced by the value
	corresponding to the key in the Dict.

	@exc
	.EArgValue

	@xmp
		s = 'Hello ${a b c} how are @{d e f}?'
		a = s.parseform
		println (a)

		s = '${HELLO} ${who}'
		a = s.parseform (['HELLO': 123])
		println (a)

	@out
		[nil, 'Hello ', '$', 'a b c', nil, ' how are ', '@', 'd e f']
		[123, 'HELLO', nil, ' ', '$', 'who']
*/
static qu_array qu__string_parseform (qu_string self, int argc, qu_any *argv)
{
	char     *i = self->str;
	long      l = self->len;
	char     *j = i;
	qu_array  a = qu_array_new (4);
	qu_dict   d = 0;

	QU_ARG_VMAX (1);
	if (argc)
		QU_ARG_VTYPE (0, d, qu_dict);

	while (l)
	{
		if (l > 1 && i[1] == '{')
		{
			qu_string k, v;

			if (j < i)
			{
				qu_array_push (a, QuNil);
				qu_array_push (a, qu_string_new (j, i - j));
			}
			k = qu_string_byte (i[0]);
			i += 2;
			j = i;
			l -= 2;
			while (l && *i != '}')
			{
				--l;
				++i;
			}
			v = qu_string_new (j, i - j);
			if (d)
			{
				qu_stringy u = qu_dict_getitem (d, v);
				if (u)
					k = u;
			}
			qu_array_push (a, k);
			qu_array_push (a, v);

			if (l == 0)
				break;
			--l;
			j = ++i;
		}
		else
		{
			--l;
			++i;
		}
	}
	if (i > j)
	{
		qu_array_push (a, QuNil);
		qu_array_push (a, qu_string_new (j, i - j));
	}
	return a;
}



/*) sub rindex (what, offset = self.__len - 1, nocase = false) -> Int
	Return the index where v{what} starts in self, -1 if not found.
	Optionally start looking from v{offset} and ignore case.

	Here v{what} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	If v{offset} is negative then count from the end.

	Unlike #{index}, this method does the search in reverse order towards the
	start of the string.

	@exc
	.EArgValue

	@xmp
		println ('..xy..xy'.rindex ('xy'))
		println ('..XY..ab'.rindex (const ['xy', 'ab']))

	@out
		6
		6

	@see
	#{rindex}
*/
static qu_int qu__string_rindex (qu_string self, int argc, qu_any *argv)
{
	long       l = self->len;
	qu_any     what;
	long       ofs = l - 1;
	int        nocase = 0;
	char      *i;
	qu_string  x;
	qu_any     v;

	QU_ARG_VCOUNT (1, 3);
	what = argv[0];
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
			QU_ARG_vbool (2, nocase);

		if (ofs < 0)
			ofs += l;
		if (ofs < 0 || ofs >= l)
			return QuInt_1;
	}
	i = self->str + ofs;
	l = ofs + 1;
	while (l)
	{
		x = qu_strz_startwith (i, l, nocase, what, &v);
		if (x)
			return qu_int_new (i - self->str);
		--i;
		--l;
	}
	return QuInt_1;
}



/*) sub rjust (width: Integer, fill = ' ') -> String
	Return a new left justified string of length v{width} by padding v{fill}.

	@exc
	.EArgValue

	@xmp
		s = 'foo'
		println (s.rjust (5).quote)

	@out
		'  foo'
*/
static qu_string qu__string_rjust (qu_string self, int argc, qu_any *argv)
{
	qu_string fill = qu_string_byte (' ');
	long      width;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong (0, width);
	if (argc > 1)
	{
		fill = argv[1];
		QU_ARG_CHECK (qu_string_is (fill) && fill->len);
	}
	if (self->len >= width)
		return self;
    return qu_string_pad (self, width - self->len, 0, fill->str[0]);
}



/*) sub split (*arg) -> Array
	@def
	sub split (max = 0)-> Array
	sub split (delimiter, max = 0, options = 't') -> Array
	sub split (regex, max = 0, options = '') -> Array
	@p
	Split the String into an Array of strings.

	The first form splits the String by any space delimiter.

	@xmp
		s = ' a \t  b \n  c \r d \v '
		a = s.split
		println (a)

	@out
		['a', 'b', 'c', 'd']

	@p
	In the second form, v{delimiter} may be a String or an Array, Set or Dict that
	yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	v{max} is the maximum number of items to return.
	A non-positive v{max] means no limit.

	The possible options are:
	@ul
	.C{"n"} case insensitive
	.C{"s"} remove spaces surrounding the delimiter
	.C{"q"} account for quoted strings
	.C{"u"} unquote (remove quotes), only if C{"q"} specified
	.C{"t"} trim after removing quotes, only if C{"u"} specified

	@xmp
		s = '  a, b ; c ,d  '
		println (s.split (const [';' ',']))

		s = ' " fXo " X `bxr` x \'baz\'  '
		println (s.split ('x', 0, "nsqut"))

	@out
		['a', 'b', 'c', 'd']
		['fXo', 'bar', 'baz']

	@p
	Following is an example on how you could split key/value pairs from a
	string with an HTML-like syntax.

	@xmp
		s = '  foo ="1;"" ;  bar =  baz=3  , fox="faz"  '
		d = [:]
		for s in s.split (const [';' ',' '\t'])
			try
				(k, v) = s.split ('=', 2)
				d[k] = v.unquote
			catch
				println ('?' s.unquote)
			;;
		;;
		println (k ' -> ' v) for k, v in d

	@out
		?
		foo -> "1
		bar -> baz=3
		fox -> faz

	@p
	The last form simply calls C{regex.split (self, max, options)}.
	See the %{Regex} module for more information.

	@xmp
		use Regex

		s = '  a, b ; c ,d  '
		r = Regex (`[[:space:]]*[,;][[:space:]]*`)
		a = s.split (r)
		println (a)

	@out
		[' a', 'b', 'c', 'd  ']
*/
static qu_array _string_split_byspaces (qu_string self, long max)
{
	qu_array  r = qu_array_new (4);
	long      l = self->len;
	char     *i = self->str;
	char     *j;

	--max;
	while (max > 0)
	{
		while (l && qu_ascii_isspace (*i))
		{
			++i;
			--l;
		}
		if (l == 0)
			break;
		j = i;
		while (l && !qu_ascii_isspace (*i))
		{
			--l;
			++i;
		}
		QU_ENSURE (qu_array_push (r, qu_string_new (j, i - j)));
		--max;
	}
	if (l)
		QU_ENSURE (qu_array_push (r, qu_string_new_trim (i, l)));
	return r;
}


static qu_array _string_split_byregex (qu_string self, int argc, qu_any *argv)
{
	static qu_string name = 0;
	qu_array r;

	if (name == 0)
		name = qu_string_new_staticz ("split");

	assert (argc > 0);

	if (argc == 1)
		r = qu_object_call1 (argv[0], name, self);

	else if (argc == 2)
		r = qu_object_call2 (argv[0], name, self, argv[1]);

	else if (argc == 3)
		r = qu_object_call3 (argv[0], name, self, argv[1], argv[2]);

	else
		QU_THROW (QuEArgCount);

	QU_ENSURE (r);
	QU_REQUIRE (QuEImplement, qu_array_is (r));
	return r;
}


#define STRING_SPLIT_NOCASE  0x01
#define STRING_SPLIT_SPACE   0x02
#define STRING_SPLIT_QUOTE   0x04
#define STRING_SPLIT_UNQUOTE 0x08
#define STRING_SPLIT_TRIM    0x10

static qu_any _string_split_bydelim_push (qu_array r, char *start, char *end, long op)
{
	if (op & STRING_SPLIT_SPACE)
	{
		while (start < end && qu_ascii_isspace (*start))
			++start;
		while (end > start && qu_ascii_isspace (*end))
			--end;
	}
	if (op & STRING_SPLIT_QUOTE)
	{
		if (op & STRING_SPLIT_UNQUOTE)
		{
			char q = *start;
			if (end > start && *end == q && (q == '"' || q == '\'' || q == '`'))
			{
				++start;
				--end;
				if (op & STRING_SPLIT_TRIM)
				{
					while (start < end && qu_ascii_isspace (*start))
						++start;
					while (end > start && qu_ascii_isspace (*end))
						--end;
				}
			}
		}
	}
	if (end >= start)
		return qu_array_push (r, qu_string_new (start, end - start + 1));
	return qu_array_push (r, QuStringEMPTY);
}


static qu_array _string_split_bydelim (qu_string self, int argc, qu_any *argv)
{
	long       max = LONG_MAX;
	long       op  = STRING_SPLIT_SPACE;
	qu_any     spec;
	qu_array   r;
	long       l;
	char      *i, *j, q;
	qu_string  delim = 0;
	qu_any     v;

	assert (argc > 0);
	spec = argv[0];

	QU_ARG_VMAX (3);
	if (argc > 1)
	{
		QU_ARG_vlong (1, max);
		if (max <= 0)
			max = LONG_MAX;
		if (argc > 2)
		{
			qu_string  s;
			long       l;
			char      *z;

			QU_ARG_VTYPE (2, s, qu_string);
			l = s->len;
			z = s->str;
			op = 0;
			while (l--)
			{
				switch (*z++)
				{
					case 'n': op |= STRING_SPLIT_NOCASE;  break;
					case 's': op |= STRING_SPLIT_SPACE;   break;
					case 'q': op |= STRING_SPLIT_QUOTE;   break;
					case 'u': op |= STRING_SPLIT_UNQUOTE; break;
					case 't': op |= STRING_SPLIT_TRIM;    break;
				}
			}
		}
	}

	l = self->len;
	i = j = self->str;
	r = qu_array_new (4);

	max--;
	while (l && max)
	{
		while (l)
		{
			if (op & STRING_SPLIT_QUOTE)
			{
				q = *i;
				if (q == '"' || q == '\'' || q == '`')
				{
					--l;
					++i;
					while (l)
					{
						if (*i == '\\' && q != '`')
						{
							--l;
							++i;
						}
						else if (*i == q)
						{
							--l;
							++i;
							break;
						}
						else
						{
							--l;
							++i;
						}
					}
				}
				if (l < 1)
					break;
			}

			delim = qu_strz_startwith (i, l, op & STRING_SPLIT_NOCASE, spec, &v);
			if (delim && delim->len)
				goto found;
			--l;
			++i;
		}
		break;

	  found:

		QU_ENSURE (_string_split_bydelim_push (r, j, i - 1, op));
		max--;
		l -= delim->len;
		i += delim->len;
		j = i;
	}
	i = self->str + self->len;
	if (i > j)
		QU_ENSURE (_string_split_bydelim_push (r, j, i - 1, op));
	return r;
}



static qu_array qu__string_split (qu_string self, int argc, qu_any *argv)
{
	if (argc == 0 || (argc == 1 && qu_integer_is (argv[0])))
	{
		long max = LONG_MAX;
		if (argc > 0)
		{
			QU_ARG_vlong (0, max);
			if (max <= 0)
				max = LONG_MAX;
		}
		return _string_split_byspaces (self, max);
	}
	else if (qu_classid (argv[0]) == QU_CLASS_ID_Regex)
	{
		return _string_split_byregex (self, argc, argv);
	}
	else
	{
		return _string_split_bydelim (self, argc, argv);
	}
}



/*) sub splitat (*indexes) -> Array
	@def
	sub splitat (*indexes) -> Array
	sub splitat (indexes: Array) -> Array
	@p
	Split the String at the specified v{indexes}.

	The indexes must be sorted from low to high.
	Use -{Array.sort} if necessary.

	This is useful when you are scanning the String and keep scanned indexes
	in an Array.

	@exc
	.EArgValue

	@xmp
		s = 'abcdefghijkl'
		a = [1 4 7 100]
		r = s.splitat (a)    # same as r = s.splitat (*a)
		println (r)

	@out
		['a', 'bcd', 'efg', 'hijkl']
*/
static qu_array qu__string_splitat (qu_string self, int argc, qu_any *argv)
{
	long     l = self->len;
	char    *s = self->str;
	long     i = 0;
	long     j, m, n;
	qu_array r;
	qu_any   x;

	if (argc && qu_array_is (argv[0]))
	{
		QU_ARG_VEXACT (1);
		argc = ((qu_array) argv[0])->len;
		argv = ((qu_array) argv[0])->objs;
	}

	r = qu_array_new (argc);
	while (argc--)
	{
		x = *argv++;
		QU_ARG_long (x, j);
		if (j >= l || j < i)
			break;
		n = j - i;
		m = l - j;
		if (n > m)
			n = m;
		QU_ENSURE (qu_array_push (r, qu_string_new (s + i, n)));
		i = j;
	}
	if (i < l)
		QU_ENSURE (qu_array_push (r, qu_string_new (s + i, l - i)));
	return r;
}



/*) sub splitkv (vdelim = '=', idelim = '', multi = false, array = false) -> Any
	Split string into key/value pairs.

	v{vdelim} is the value delimiter. v{idelim} is the item delimiter.
	Both v{vdelim} and v{idelim} may be strings with more than one
	character but they may not be empty. All characters are treated as
	alternatives.

	If v{multi} is true then allow duplicate keys and collect the values in an
	Array. Otherwise the last prevails.

	If v{array} is true then return an Array instead of a Dict, thus push the
	pairs consecutively in an Array.

	@exc
	.EArgValue

	@xmp
		s = 'foo="1" bar=2 foo="uhm"'
		println (s.splitkv)

		s = 'foo="1"; bar=2, foo="uhm"'
		println (s.splitkv ('=', ',;'))

	@out
		['foo':'uhm' 'bar':'2']
		['foo':'uhm' 'bar':'2']
*/
static qu_any qu__string_splitkv (qu_string self, int argc, qu_any *argv)
{
	qu_string vdelim = qu_string_byte ('=');
	qu_string idelim = QuStringEMPTY;
	int       multi  = 0;
	int       ary    = 0;

	QU_ARG_VCOUNT (0, 4);
	if (argc > 0)
	{
		vdelim = argv[0];
		QU_ARG_CHECK (qu_string_is (vdelim) && vdelim->len);
		if (argc > 1)
		{
			idelim = argv[1];
			QU_ARG_TYPE (idelim, qu_string);
			if (argc > 2)
			{
				QU_ARG_vbool (2, multi);
				if (argc > 3)
					QU_ARG_vbool (3, ary);
			}
		}
	}

	/*begin*/
	{
		char *sv = vdelim->str;
		char *si = idelim->str;
		long  lv = vdelim->len;
		long  li = idelim->len;

		long      l = self->len;
		char     *s = self->str;
		qu_any    res = ary ? (qu_any) qu_array_new (2) : (qu_any) qu_dict_new (2);

		char     *key, *val;
		long      lkey, lval;
		int       k;
		qu_string ks, vs;

		while (l)
		{
			if (qu_ascii_isspace (*s))
			{
				l--;
				s++;
				continue;
			}
			key = s;
			lkey = 0;
			k = 0;
			while (l && !qu_ascii_isspace (*s))
			{
				if (memchr (sv, *s, lv))
				{
					k = 1;
					l--;
					s++;
					break;
				}
				lkey++;
				l--;
				s++;
			}
			if (k == 0)
			{
				while (l && qu_ascii_isspace (*s))
				{
					l--;
					s++;
				}
				if (memchr (sv, *s, lv))
				{
					l--;
					s++;
					while (l && qu_ascii_isspace (*s))
					{
						l--;
						s++;
					}
				}
			}
			if (l && *s == '"')
			{
				val = ++s;
				l--;
				while (l && *s != '"')
				{
					l--;
					s++;
				}
				lval = s - val;
				if (l)
				{
					--l;
					++s;
				}
			}
			else
			{
				val = s;
				while (l && !qu_ascii_isspace (*s))
				{
					if (li && memchr (si, *s, li))
						break;
					--l;
					++s;
				}
				lval = s - val;
			}

			ks = qu_string_new (key, lkey);
			vs = qu_string_new (val, lval);

			if (multi)
			{
				long   j = 0;
				qu_any x;

				if (ary)
				{
					j = qu_array_index (res, ks, 0);
					x = 0;
					if (j >= 0)
						x = ((qu_array) res)->objs[j];
				}
				else
				{
					x = qu_dict_getitem ((qu_dict) res, ks);
				}

				if (x)
				{
					if (qu_array_is (x))
						qu_array_push ((qu_array) x, ks);
					else
					{
						qu_array a = qu_array_new_ex (2, x, vs);
						if (ary)
							((qu_array) res)->objs[j] = a;
						else
							qu_dict_setitem ((qu_dict) res, ks, a);
					}
				}
				else
				{
					goto push;
				}
			}
			else
			{
			  push:
				if (ary)
				{
					qu_array_push ((qu_array) res, ks);
					qu_array_push ((qu_array) res, vs);
				}
				else
				{
					qu_dict_setitem ((qu_dict) res, ks, vs);
				}
			}
			if (li)
			{
				while (l && qu_ascii_isspace (*s))
				{
					--l;
					++s;
				}
				if (memchr (si, *s, li))
				{
					--l;
					++s;
				}
			}
		}
		return res;
	}
}



/*) sub splitnum (how = 'B') -> Array
	Break the string by separating integer numbers within the string with the others.

	Optionally set v{how} to:
	@ops
	.| C{b} | collect both, don't convert to Integer
	.| C{B} | collect both, convert to Integer
	.| C{n} | collect numbers only, don't convert to Integer
	.| C{N} | collect numbers only, convert to Integer
	.| C{s} | collect strings only

	@exc
	.EArgValue

	@xmp
		s = 'x-a12.100p'
		println (s.splitnum)
		println (s.splitnum ('N'))
		println (s.splitnum ('n'))
		println (s.splitnum ('s'))

	@out
		['x-a', 12, '.', 100, 'p']
		[12, 100]
		['12', '100']
		['x-a', '.', 'p']
*/
static qu_array qu__string_splitnum (qu_string self, int argc, qu_any *argv)
{
	char        op = 'B';
	long        l = self->len;
	char       *i = self->str;
	qu_array    a;
	char       *p, *k;
	qu_integer  v;

	QU_ARG_VCOUNT (0, 1);
	if (argc > 0)
	{
		qu_string x;
		QU_ARG_VTYPE (0, x, qu_string);
		op = x->str[0];
		QU_ARG_CHECK (op == 'B' || op == 'b' || op == 'n' || op =='N' || op == 's' || op == 'S');
	}

	if (l == 0)
		return qu_array_new (0);

	a = qu_array_new (4);
	p = i;
	while (l)
	{
		if (qu_ascii_isdigit (*i))
		{
			k = i;

			if (i > p && (op == 'B' || op == 'b' || op == 'S' || op == 's'))
				qu_array_push (a, qu_string_new (p, i - p));

			while (qu_ascii_isdigit (*i))
			{
				++i;
				--l;
			}

			if (op == 'b' || op == 'n')
			{
				qu_array_push (a, qu_string_new (k, i - k));
			}
			else if (op == 'B' || op == 'N')
			{
				/*HACK -- qu_int_scan errs when there is a dot after the digit */
				char c = *i;
				*i = '\0';
				v = (qu_integer) qu_int_scan (&k, 0);
				*i = c;
				if (v == 0)
				{
					*i = '\0';
					v = (qu_integer) qu_long_scan (&k, 0);
					*i = c;
					if (v == 0)
						/* this should never happen unless the string is extra huge */
						v = (qu_integer) QuInt0;
				}
				qu_array_push (a, v);
			}
			p = i;
		}
		else
		{
			++i;
			--l;
		}
	}
	if (i > p && (op == 'B' || op == 'b' || op == 'S' || op == 's'))
		qu_array_push (a, qu_string_new (p, i - p));
	return a;
}



/*) sub startwith (what, offset = 0, nocase = false, word = false) -> Any
	Whether or not the String starts with v{what}, optionally case insensitive,
	from a specified v{offset}, and whether it should end with a word boundary.

	v{what} may be a String or an Array, Set or Dict that yields String objects.
	Non String items/keys are ignored. In case of a Dict then the keys are used
	and the corresponding value returned.

	Returns nil if none matches, the matched item otherwise.

	@exc
	.EArgValue

	@xmp
		s = 'foo'
		println (s.startwith ('fo'))

	@out
		fo

	@see
	#{endwith}
*/
static qu_any qu__string_startwith (qu_string self, int argc, qu_any *argv)
{
	qu_any  what;
	long    ofs = 0;
	int     nocase = 0;
	int     word = 0;
	qu_any  v;

	QU_ARG_VCOUNT (1, 4);
	what = argv[0];
	if (argc > 1)
	{
		QU_ARG_vlong (1, ofs);
		if (argc > 2)
		{
			QU_ARG_vbool (2, nocase);
			if (argc > 3)
				QU_ARG_vbool (3, word);
		}

		if (ofs < 0)
			ofs += self->len;
		if (ofs < 0 || ofs >= self->len)
			return QuNil;
	}
	what = qu_strz_startwith (self->str + ofs, self->len - ofs, nocase, what, &v);
	if (what)
	{
		if (word && qu_stringy_is (what))
		{
			long i = ofs + ((qu_string) what)->len;
			if (i < self->len && qu_ascii_isalpha (self->str[i]))
				return QuNil;
		}
		return v ? v : what;
	}
	return QuNil;
}



/*) sub startswith (what, offset = 0, nocase = false, word = false) -> Boolean
	This method wraps #{starwith} and returns false if it returns nil,
	true otherwise.

	It may be useful in a switch statement.

	@exc
	.EArgValue

	@xmp
		s = 'bar'
		switch s
			case .startswith ('a')
				println (1)
				break

			case .startswith ('b')
				println (2)
				break
		;;

	@out
		2
*/
static qu_any qu__string_startswith (qu_string self, int argc, qu_any *argv)
{
	qu_any x = qu__string_startwith (self, argc, argv);

	QU_ENSURE (x);
	return (x == (qu_any) QuNil) ? QuFalse : QuTrue;
}



/*) sub tail (sep = '', ofs = 0, nocase = false) -> String
	Return the part of the string after the separator v{sep}.

	v{sep} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	If v{sep} is an empty String then it means any number of spaces.

	Optionally start of v{offset}.
	If this is a negative value then count from the end.

	If the seperator is not found then an empty String is returned.

	@exc
	.EArgValue

	@xmp
		s = '123 \n \t 45678'
		println (s.tail)

	@out
		45678
*/
static qu_string qu__string_tail (qu_string self, int argc, qu_any *argv)
{
	qu_any     sep = 0;
	int        nocase = 0;
	long       ofs = 0;
	char      *i = self->str;
	long       l = self->len;
	qu_string  x;
	qu_any     v;

	QU_ARG_VCOUNT (0, 3);
	if (argc > 0)
	{
		sep = argv[0];
		if (qu_stringy_is (sep) && ((qu_stringy) sep)->len == 0)
			sep = 0;
		if (argc > 1)
		{
			QU_ARG_vlong (1, ofs);
			if (argc > 2)
				QU_ARG_vbool (2, nocase);

			if (ofs < 0)
				ofs += l;
			if (ofs < 0)
				return self;
			if (ofs >= l)
				return QuStringEMPTY;
		}
	}
	i += ofs;
	l -= ofs;

	if (sep == 0)
	{
		while (l)
		{
			if (qu_ascii_isspace (*i))
			{
				while (l && qu_ascii_isspace (*i))
				{
					--l;
					++i;
				}
				return qu_string_new (i, self->str + self->len - i);
			}
			--l;
			++i;
		}
	}
	else
	{
		while (l)
		{
			x = qu_strz_startwith (i, l, nocase, sep, &v);
			if (x)
				return qu_string_new (i + x->len, l - x->len);
			++i;
			--l;
		}
	}
	return QuStringEMPTY;
}



/*----------------------------------------------------------------------------
	specific instance methods - "modifying"
----------------------------------------------------------------------------*/

/*) sub indent (num = 1, tab = '\t') -> String
	Indent the string.

	v{tab} may not be empty.

	v{num} must be a non-negative integer.

	@exc
	.EArgValue
	.EMemory

	@p
	@code
		|    xxxxxxxx           |        xxxxxxxx
		|        xxxxx    ->    |            xxxxx
		|xxxx                   |    xxxx
*/
static qu_string qu__string_indent (qu_string self, int argc, qu_any *argv)
{
	char       tab = '\t';
	long       num = 1;
	char      *i, *p;
	long       l, m, n;
	qu_string  r;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, num);
		QU_ARG_CHECK  (num >= 0);
		if (argc > 1)
		{
			qu_string t = argv[1];
			QU_ARG_CHECK (qu_string_is (t) && t->len);
			tab = t->str[0];
		}
	}

	if (num == 0)
		return self;

	i = self->str;
	l = self->len;
	n = 0;
	m = QU_MEM_MAX - l;
	while (l)
	{
		if (*i == '\n')
		{
			--l;
			++i;
		}
		else
		{
			QU_MEM_CHECK (m > num);
			n += num;
			m -= num;
			while (l && *i != '\n')
			{
				--l;
				++i;
			}
		}
	}

	if (n == 0)
		return self;

	l = self->len;
	i = self->str;
	r = qu_string_newl (n + l);
	p = r->str;
	while (l)
	{
		if (*i == '\n')
		{
			--l;
			*p++ = *i++;
		}
		else
		{
			m = num;
			while (m--)
				*p++ = tab;
			*p++ = *i++;
			--l;
			while (l && *i != '\n')
			{
				--l;
				*p++ = *i++;
			}
		}
	}
	return r;
}



/*) sub interval (delim: String, intv: Integer) -> String
	Format self into intervals.

	v{delim} is the delimiter string to use.

	v{intv} is the interval. It must be a positive -{Integer}.

	@exc
	.EArgValue

	@xmp
		println ('1234567'.interval ('..', 2))

	@out
		1..23..45..67
*/
static qu_string qu__string_interval (qu_string self, qu_string delim, qu_integer intvl)
{
	long interval;

	QU_ARG_TYPE  (delim, qu_string);
	QU_ARG_long  (intvl, interval);
	QU_ARG_CHECK (interval > 0);

	if (interval >= self->len)
		return self;
	else
	{
		long       len = self->len + (self->len - 1) / interval * delim->len;
		qu_string  s = qu_string_newl (len);
		char      *k = s->str;
		char      *i = k + len - 1;
		char      *j = self->str + self->len - 1;
		char      *d;
		long       n;

		for (;;)
		{
			n = interval;
			while (n--)
			{
				if (i < k)
					return s;
				*i-- = *j--;
			}
			n = delim->len;
			d = delim->str + n - 1;
			while (n--)
				*i-- = *d--;
		}
		return s;
	}
}



/*) sub ltrim -> String
	Return a new string by removing leading spaces.

	@noexc

	@xmp
		println ('  123 '.ltrim.quote)

	@out
		'123 '
*/
#define qu__string_ltrim qu_string_ltrim
qu_string qu__string_ltrim (qu_string self)
{
	char *p = self->str;
	char *e = p + self->len;

	while (qu_ascii_isspace (*p))
		p++;
	return qu_string_new (p, e - p);
}



/*) sub pad (left: Integer, right: Integer, fill = ' ') -> String
	Return a new string by padding v{fill} to the v{left} and v{right}.

	@exc
	.EMemory
	.EArgValue

	@xmp
		s = 'abc'
		println (s.pad (2, 2, '.'))

	@out
		..abc..
*/
static qu_string qu__string_pad (qu_string self, int argc, qu_any *argv)
{
	char fill = ' ';
	long left, right;

	QU_ARG_VCOUNT (2, 3);
	QU_ARG_vlong  (0, left);
	QU_ARG_vlong  (1, right);
	if (argc > 2)
	{
		qu_string f = argv[2];
		QU_ARG_CHECK (qu_string_is (f) && f->len);
		fill = f->str[0];
	}
	return qu_string_pad (self, left, right, fill);
}



/*) sub quote (quote = "'") -> String
	Return a new quoted String with characaters escaped as necessary.

	You can optionally set the v{quote} as C{'} or C{"} or C{`} but no other.
	If v{quote} is C{`} and String contains non-printable characters then
	then the v{quote} is forced to C{'}.

	@exc
	.EArgValue
	.EMemory

	@xmp
		s = 'foo\tbar'
		a = [1, s]
		println (a)
		println (s.quote)
		println (s)

	@out
		[1, 'foo\tbar']
		'foo\tbar'
		foo     bar
*/
static qu_string qu__string_quote (qu_string self, int argc, qu_any *argv)
{
	char quote = '\'';

	if (argc)
	{
		qu_string s = argv[0];

		QU_ARG_TYPE (s, qu_string);
		quote = s->str[0];
		QU_ARG_CHECK (quote == '\'' || quote == '"' || quote == '`');
	}
	return qu_strz_quote (self->str, self->len, quote);
}



/*) sub rcount (s: String, nocase = false) -> Int
	Count the number of characters not in this String.

	@exc
	.EArgValue

	@xmp
		s = 'abacdefa'
		println (s.count ('a'))

	@out
		5
*/
static qu_int qu__string_rcount (qu_string self, int argc, qu_any *argv)
{
	qu_string  s;
	int        nocase = 0;
	char      *i = self->str, c;
	long       l = self->len, n = 0;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_VTYPE (0, s, qu_string);
	if (argc > 1)
		QU_ARG_vbool (1, nocase);

	if (s->len == 0)
		return QuInt0;

	if (nocase)
	{
		c = qu_ascii_toupper (s->str[0]);
		while (l--)
		{
			if (c != qu_ascii_toupper (*i))
				++n;
			++i;
		}
	}
	else
	{
		c = s->str[0];
		while (l--)
		{
			if (c != *i++)
				++n;
		}
	}

	assert (n <= QU_INT_MAX);
	return qu_int_new (n);
}



/*) sub replace (*arg) -> String
	@def
	sub replace (what, rep = '', max = -1, nocase = false) -> String
	sub replace (what: Dict, max = -1, nocase = false) -> String
	sub replace (pattern: Regex, *arg) -> String
	@p
	Return a new String by replacing string parts specified by v{what}.

	The last form simply calls C{pattern.replace (self, *arg)}.

	v{what} may be a String or an Array, Set or Dict that yields String objects.
	In case of a Dict then the keys are used and the values ignored.
	Non String items are ignored.

	v{rep} is the replacement string.

	v{max} is the maximum number of replacements to perform.
	A non-positive value means no limit.

	v{nocase} is true to do a case insensitive match.

	@exc
	.EArgValue
	.EMemory

	@xmp
		use Regex

		sub rep (a)
			return 'x'
		;;

		r = Regex (`([bc])(...)r`)
		s = '__b123r__c456r__'
		println (s.replace (r, @rep))

		s = 'foo Foo BAR foo bar'
		println (s.replace (['foo', 'bar'], 'X', 3, true))

	@out
		__x__x__
		X X X foo bar
*/
static qu_string qu__string_replace (qu_string self, int argc, qu_any *argv)
{
	qu_any     what;
	qu_string  with = QuStringEMPTY;
	long       max = -1;
	int        nocase = 0;
	long       l;
	char      *i;
	qu_string  r, x;
	qu_any     v;

	QU_ARG_VCOUNT (1, 4);
	what = argv[0];

	if (qu_classid (what) == QU_CLASS_ID_Regex)
	{
		static qu_string name = 0;

		if (name == 0)
			name = qu_string_new_staticz ("replace");
		if (argc == 1)
			what = qu_object_call1 (what, name, self);
		else if (argc == 2)
			what = qu_object_call2 (what, name, self, argv[1]);
		else if (argc == 3)
			what = qu_object_call3 (what, name, self, argv[1], argv[2]);
		else
			what = qu_object_call4 (what, name, self, argv[1], argv[2], argv[3]);

		QU_ENSURE (what);
		QU_REQUIRE (QuEImplement, qu_string_is (what));
		return what;
	}
	if (argc > 1)
	{
		QU_ARG_VTYPE (1, with, qu_string);
		if (argc > 2)
		{
			QU_ARG_vlong (2, max);
			if (max <= 0)
				max = -1;
			if (argc > 3)
				QU_ARG_vbool (3, nocase);
		}
	}
	l = self->len;
	i = self->str;
	r = qu_string_buffer_new (16);

	while (l && max)
	{
		x = qu_strz_startwith (i, l, nocase, what, &v);
		if (x)
		{
			--max;
			if (v)
				v = qu_sys_string (v);
			QU_ENSURE (qu_string_buffer_puts (r, v ? v : with));
			l -= x->len;
			i += x->len;
		}
		else
		{
			QU_ENSURE (qu_string_buffer_put (r, i, 1));
			--l;
			++i;
		}
	}
	if (l)
		QU_ENSURE (qu_string_buffer_put (r, i, l));
	return qu_string_buffer_return (r);
}



/*) sub reverse -> String
	Return a new string with reversed bytes.

	@noexc

	@xmp
		println ('abcde'.reverse)

	@out
		edcba
*/
static qu_string qu__string_reverse (qu_string self)
{
	long l;

	l = self->len;
	if (l < 2)
		return self;
	else
	{
		char *i = self->str;

		if (l == 2 && i[0] == i[1])
			return self;
		else
		{
			qu_string   s = qu_string_newl (l);
			char *j = s->str;

			i += l - 1;
			while (l--)
				*j++ = *i--;
			return s;
		}
	}
}



/*) sub rtrim -> String
	Return a new String by removing trailing spaces.

	@noexc

	@xmp
		x = ' 123 \r\t\v\f\n '
		y = x.rtrim
		println (y.quote)

	@out
		' 123'
*/
#define qu__string_rtrim qu_string_rtrim
qu_string qu__string_rtrim (qu_string self)
{
	long  l = self->len;
	char *s = self->str;
	char *p = s + l;

	do { --p; } while (p >= s && qu_ascii_isspace (*p));
	if (p < s)
		return QuStringEMPTY;
	return qu_string_new (s, p - s + 1);
}



/*) sub scrap (s = ' ', translate = const [:], nocase = false) -> self
	Return a new String by removing consecutive repeating characters.

	Optionally v{translate} the characters specified by a Dict before the equality
	test, and ignore case.

	@exc
	.EModify
	.EArgValue

	@xmp
		d =
		[
			'\t': ' '
			'\r': ' '
			';' : ','
		]
		s = 'a,,;  \t   b///c'.modify
		r = s.scrap (' ,/', d)
		println (r)

	@out
		a, b/c
*/
static qu_string qu__string_scrap (qu_string self, int argc, qu_any *argv)
{
	qu_string  cset    = qu_string_byte (' ');
	qu_dict    trans   = 0;
	int        nocase  = 0;
	char       current = 0;
	int        matched = 0;
	char      *s, *q, c;
	long       l;
	qu_string  r, x;

	QU_ARG_VCOUNT (0, 3);
	if (argc > 0)
	{
		QU_ARG_VTYPE (0, cset, qu_string);
		if (argc > 1)
		{
			QU_ARG_VTYPE (1, trans, qu_dict);
			if (trans->len == 0)
				trans = 0;
			if (argc > 2)
				QU_ARG_vbool (2, nocase);
		}
	}

	r = qu_string_buffer_new (self->len);
	q = r->str;
	s = self->str;
	l = self->len;
	while (l--)
	{
		c = *s++;
		if (trans)
		{
			x = qu_dict_getitem (trans, qu_string_byte (c));
			if (x && qu_string_is (x) && x->len)
			{
				c = x->str[0];
				if (nocase)
					c = qu_ascii_toupper (c);
			}
		}
		if (matched && (c == current || (nocase && c == qu_ascii_toupper (current))))
		{
			continue;
		}
		else
		{
			if (STRZINCHR (cset->str, cset->len, c, nocase))
			{
				current = c;
				matched = 1;
			}
			else
				matched = 0;
			*q++ = c;
		}
	}
	r->len = q - r->str;
	return qu_string_buffer_return (r);
}




/*) sub start (rtrim = true)
	Return a new String by removing leading blank lines and optionally trailing spaces.

	@exc
	.EArgValue

	@xmp
		x = '  \v\r\t\n  foo  \nbar  \r\t\v\n\n'
		y = x.start
		println (y.quote)

	@out
		'  foo  \nbar'
*/
static qu_string qu__string_start (qu_string self, int argc, qu_any *argv)
{
	int   rtrim = 1;
	char *s = self->str;
	char *b = s;
	char *e = s + self->len - 1;
	long  n;

	QU_ARG_VMAX (1);
	if (argc > 0)
		QU_ARG_vbool (0, rtrim);
	if (rtrim)
	{
		while (e >= s && qu_ascii_isspace (*e))
			--e;
		if (e < s)
			return QuStringEMPTY;
		if (e == s)
			return qu_string_byte (*e);
	}
	while (s <= e && qu_ascii_isspace (*s))
		++s;
	if (s > e)
		return QuStringEMPTY;
	if (s == e)
		return qu_string_byte (*s);
	while (s > b && s[-1] != '\n')
		--s;
	n = e - s + 1;
	if (n == self->len)
		return self;
	return qu_string_new (s, n);
}



/*) sub strip (len: Integer, pad = '...') -> String
	Return a string by cutting self to a specified length and optionally
	append v{pad}.

	If v{len} >= the string length, then return self.

	@exc
	.EArgValue

	@xmp
		s = "12345"
		println (s.strip (3, '...'))

	@out
		123...
*/
static qu_string qu__string_strip (qu_string self, int argc, qu_any *argv)
{
	qu_string  pad = 0;
	long       l;

	QU_ARG_VCOUNT (1, 2);
	QU_ARG_vlong  (0, l);
	QU_ARG_CHECK  (l >= 0);
	if (argc > 1)
		QU_ARG_VTYPE (1, pad, qu_string);

	if (l >= self->len)
	{
		return self;
	}
	else if (pad == 0)
	{
		qu_string s = qu_string_newl (l + 3);

		memcpy (s->str, self->str, 3);
		memcpy (s->str + 3, "...", 3);
		return s;
	}
	else
	{
		long n = pad->len;

		if (n == 0)
			return qu_string_new (self->str, l);
		else
		{
			qu_string s = qu_string_newl (l + n);

			memcpy (s->str, self->str, l);
			memcpy (s->str + l, pad->str, n);
			return s;
		}
	}
}



/*) sub swapcase -> String
	Return a new string with the lettercase swapped.

	@noexc

	@xmp
		x = 'Foo'
		y = x.swapcase
		println (y)

	@out
		fOO
*/
static qu_string qu__string_swapcase (qu_string self)
{
	char *s = self->str;
	long  l = self->len;

	if (l == 0)
	{
		return QuStringEMPTY;
	}
	else if (l == 1)
	{
		return qu_string_byte (qu_ascii_toswap (*s));
	}
	else
	{
		qu_string  r = qu_string_dup (self);
		char      *p = r->str;

		while (l--)
		{
			*p++ = qu_ascii_toswap (*s);
			++s;
		}
		return r;
	}
}



/*) sub title -> String
	Return a new String by capitalizing first letter or words.

	@noexc

	@xmp
		s = 'a boy and a girl'
		println (s.title)

	@out
		A Boy And A Girl
*/
static qu_string qu__string_title (qu_string self)
{
	qu_string  r = qu_string_dup (self);
	char      *s = r->str;
	long       l = r->len;

	while (l--)
	{
		if (qu_ascii_isspace (*s))
		{
			s++;
		}
		else
		{
			*s = qu_ascii_toupper (*s);
			++s;
			while (l && !qu_ascii_isspace (*s))
			{
				--l;
				++s;
			}
		}
	}
	return r;
}



/*) sub translate (x, ucase = false) -> String
	Return a new String by replacing characters specified by v{x}.

	v{x} is either a Dict, Array, String or Function.

	If v{x} is a Dict then the keys are the characters to replace and the values
	are the replacement strings.

	If v{x} is an Array then the items are the replacements. The first item is
	the replacement for character 0x00, the second for 0x01, etc. It does not
	need to have 256 items.

	If v{x} is a String then the characters represent the replacement
	characters. The length does not need to be 256.

	If v{x} is a Function then that Function will be called with the character
	to replace as the single argument. The returned value is the replacement.

	If v{x} is a Dict or Array and the replacement is a Function then that
	Function will be called with the character to replace as the single argument.
	The returned value is the replacement.

	If the replacement is not a String then it is ignored.
	If the Function throws an exception, it is also ignored.

	v{ucase} is true then the replaced character is converted to an uppercase
	letter before the translation.

	@exc
	.EArgValue
	.EMemory

	@xmp
		s = 'foo.bar-baz'
		println (s.translate (['.': 'DOT', '-': 'DASH']))

	@out
		fooDOTbarDASHbaz

	@see
	See also -{printc} that uses the same functionality as this method.
*/
static qu_string qu__string_translate (qu_string self, int argc, qu_any *argv)
{
	qu_any     x;
	int        ucase = 0;
	long       l = self->len;
	char      *s = self->str;
	qu_string  b;

	QU_ARG_VCOUNT (1, 2);
	x = argv[0];
	if (argc > 1)
		QU_ARG_vbool (1, ucase);

	b = qu_string_buffer_new (l);
	while (l--)
	{
		char   c = *s++;
		qu_string p = qu_strz_translate (ucase ? qu_ascii_toupper (c) : c, x);

		if (p)
			QU_ENSURE (qu_string_buffer_puts (b, p));
		else
			QU_ENSURE (qu_string_buffer_putc (b, c));
	}
	return qu_string_buffer_return (b);
}



/*) sub trim -> String
	Return a new String by removing leading and trailing spaces.
	Thus combines #{ltrim} and #{rtrim}.

	@noexc

	@xmp
		println (' 123  '.trim.quote)

	@out
		'123'
*/
#define qu__string_trim qu_string_trim
qu_string qu__string_trim (qu_string self)
{
	char *s = self->str;
	char *q = s + self->len;

	while (qu_ascii_isspace (*s))
		++s;
	while (q > s && qu_ascii_isspace (q[-1]))
		--q;
	return qu_string_new (s, q - s);
}



/*) sub unfold -> String
	Unfold C{'\r\n'} continuations.

	This method is useful for things such as HTTP headers.

	@noexc

	@xmp
		x = 'foo: 1\r\n  \t bar\r\nbaz: 2'
		y = s.unfold
		println (y.quote)

	@out
		'foo bar'
*/
static qu_string qu__string_unfold (qu_string self)
{
	long  l = self->len;
	char *i = self->str;

	while (l > 2)
	{
		if (   i[0] == '\r'
			&& i[1] == '\n'
			&& (    i[2] == ' '
				||  i[2] == '\t'
				||  i[2] == '\v'
				||  i[2] == '\f'
				||  i[2] == '\n'
				|| (i[2] == '\r' && l > 3 && i[3] != '\n')))
		{
			qu_string  s = qu_string_newl (self->len);
			char      *j = s->str;

			l = self->len;
			i = self->str;
			while (l)
			{
				if (l > 2
					&& i[0] == '\r'
					&& i[1] == '\n'
					&& (    i[2] == ' '
						||  i[2] == '\t'
						||  i[2] == '\v'
						||  i[2] == '\f'
						||  i[2] == '\n'
						|| (i[2] == '\r' && l > 3 && i[3] != '\n')))
				{
					*j++ = ' ';
					i += 2;
					l -= 2;
					while (l
						&& (    i[0] == ' '
							||  i[0] == '\t'
							||  i[0] == '\v'
							||  i[0] == '\f'
							||  i[0] == '\n'
							|| (i[0] == '\r' && l > 1 && i[1] != '\n')))
					{
						--l;
						++i;
					}
				}
				else
				{
					--l;
					*j++ = *i++;
				}
			}

			l = j - s->str;

			if (l == 0)
				return QuStringEMPTY;

			else if (l == 1)
				return qu_string_byte (s->str[0]);

			else
			{
				s->len = l;
				return s;
			}
		}
		i++;
		l--;
	}
	return self;
}



/*) sub unquote -> String
	Return a new String by removing quotes at the start and end of the String.

	If the quotes are not placed at the start and end then self is returned.

	@noexc

	@xmp
		s = '"hello"'
		println (s.unquote)

	@out
		hello
*/
static qu_string qu__string_unquote (qu_string self)
{
	char *s = self->str;

	if (self->len > 1
		&& ((s[0] == '"' || s[0] == '\'' || s[0] == '`')
			&& s[0] == s[self->len - 1]))
		return qu_string_new (s + 1, self->len - 2);
	return self;
}



/*) sub unindent (max = -1, char = '\t') -> String
	Unindent string.

	v{char} must not be empty and v{max} must be a non-negative integer.

	@exc
	.EArgValue

	@p
	@code
		|		xxxxxxxx        |	xxxxxxxx
		|			xxxxx   ->  |		xxxxx
		|	xxxx                |xxxx
*/
static qu_string qu__string_unindent (qu_string self, int argc, qu_any *argv)
{
	char tab = '\t';
	long max = LONG_MAX;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, max);
		if (max < 0)
			max = LONG_MAX;
		if (argc > 1)
		{
			qu_string t = argv[1];
			QU_ARG_CHECK (qu_string_is (t) && t->len);
			tab = t->str[0];
		}
	}

	if (max == 0)
		return self;
	else
	{
		char *i = self->str;
		long  l = self->len;
		long  ind = LONG_MAX;
		long  n;

		while (l > 0)
		{
			if (*i == '\n')
			{
				i++;
				l--;
			}
			else
			{
				n = 0;
				while (l && *i == tab && n < max)
				{
					i++;
					n++;
					l--;
				}
				if (n < ind)
					ind = n;
				if (ind == 0)
					return self;
				while (l && *i != '\n')
				{
					l--;
					i++;
				}
			}
		}

		l = self->len;
		i = self->str;

		/*begin*/
		{
			qu_string  b = qu_string_buffer_new (l);
			char      *k;

			while (l > 0)
			{
				n = ind;
				while (l && *i == tab && n)
				{
					i++;
					n--;
					l--;
				}
				k = i;
				while (l)
				{
					if (*i == '\n')
					{
						l--;
						i++;
						break;
					}
					l--;
					i++;
				}
				QU_ENSURE (qu_string_buffer_put (b, k, i - k));
			}

			return qu_string_buffer_return (b);
		}
	}
}


/*) sub words -> Set
	Return a -{Set} containing unique "words" in the string.

	"Words" are separated by spaces (#{isspace}).
	They may be enclosed in single, double and tick quotes.
	Note however that no escape characters within the quotes are recognized.

	@noexc

	@xmp
		s = ` -a 'foo' foo "foo bar" baz  `
		println (s.words)

	@out
		{'foo', '-a', 'foo bar', 'baz'}
*/
static qu_set qu__string_words (qu_string self)
{
	long     l = self->len;
	char    *i = self->str;
	qu_set   t = qu_set_new (8);
	char    *k, c;

	while (l--)
	{
		c = *i++;
		if (qu_ascii_isspace (c))
		{
			continue;
		}
		else if (c == '"' || c == '\'' || c == '`')
		{
			k = i;
			while (l && *i != c)
			{
				++i;
				--l;
			}
			qu_set_addz (t, k, i - k);
			if (l)
			{
				++i;
				--l;
			}
		}
		else
		{
			k = i - 1;
			while (l && !qu_ascii_isspace (*i))
			{
				++i;
				--l;
			}
			qu_set_addz (t, k, i - k);
		}
	}
	return t;
}



/*----------------------------------------------------------------------------
	specific instance methods - conversion
----------------------------------------------------------------------------*/

/*) sub toident -> String
	Return an identifier string by converting punctuations and funny characters
	to alpha or digits suitable for reference.

	@noexc

	@xmp
		x = 'F-o^ o  9'
		y = x.toident
		println (y)

	@out
		F-ov_o__9
*/
static qu_string qu__string_toident (qu_string self)
{
	char *s = self->str;
	long  l = self->len;

	if (l == 0)
	{
		return QuStringEMPTY;
	}
	else if (l == 1)
	{
		return qu_string_byte (qu_ascii_toident (*s));
	}
	else
	{
		qu_string  r = qu_string_dup (self);
		char      *p = r->str;

		while (l--)
		{
			*p++ = qu_ascii_toident (*s);
			++s;
		}
		return r;
	}
}



/*) sub tolower -> String
	Return a new string by converting all characters to lowercase.

	@noexc

	@xmp
		x = 'ABC'
		y = s.tolower
		println (y)

	@out
		abc
*/
static qu_string qu__string_tolower (qu_string self)
{
	char *s = self->str;
	long  l = self->len;

	if (l == 0)
	{
		return QuStringEMPTY;
	}
	else if (l == 1)
	{
		return qu_string_byte (qu_ascii_tolower (*s));
	}
	else
	{
		qu_string  r = qu_string_dup (self);
		char      *p = r->str;

		while (l--)
		{
			*p++ = qu_ascii_tolower (*s);
			++s;
		}
		return r;
	}
}



/*) sub toprint -> String
	Return a new string by converting non-printable characters to spaces.

	Printable characters are C{' ' to '~'} plus C{'\t'} and C{'\n'}.
	Thus is different from f{isprint}.

	@noexc

	@xmp
		x = 'a\x00b\x01c'
		y = s.toprint
		println (y)

	@out
		a b c
*/
static qu_string qu__string_toprint (qu_string self)
{
	char *s = self->str;
	long  l = self->len;

	if (l == 0)
	{
		return QuStringEMPTY;
	}
	else if (l == 1)
	{
		return qu_string_byte (qu_ascii_toprint (*s));
	}
	else
	{
		qu_string  r = qu_string_dup (self);
		char      *p = r->str;

		while (l--)
		{
			*p++ = qu_ascii_toprint (*s);
			++s;
		}
		return r;
	}
}



/*) sub toupper -> String
	Return a new string by converting all characters to uppercase.

	@noexc

	@xmp
		x = 'abc'
		y = s.toupper
		println (y)

	@out
		ABC
*/
static qu_string qu__string_toupper (qu_string self)
{
	char *s = self->str;
	long  l = self->len;

	if (l == 0)
	{
		return QuStringEMPTY;
	}
	else if (l == 1)
	{
		return qu_string_byte (qu_ascii_toupper (*s));
	}
	else
	{
		qu_string  r = qu_string_dup (self);
		char      *p = r->str;

		while (l--)
		{
			*p++ = qu_ascii_toupper (*s);
			++s;
		}
		return r;
	}
}



/*) sub toheader (i: Integer = 0, n: Integer = self.len) -> String
	Return a new string by converting the characters to header format optionally
	start from index v{i} and convert v{n} characters.

	If v{i} is negative then count from the end.
	If v{n} is negative it means all character from v{i}.

	@exc
	.EIndex

	@xmp
		x = 'foo-bar: 123'
		y = x.toheader
		println (y)

	@out
		Foo-Bar: 123
*/
static qu_string qu__string_toheader (qu_string self, int argc, qu_any *argv)
{
	char *s;
	long  l = self->len;
	long  i = 0;
	long  n = l;
	int   k = 0;

	QU_ARG_VCOUNT (0, 2);
	if (argc > 0)
	{
		QU_ARG_vlong (0, i);
		QU_ENSURE_INDEX (i, l);
		if (argc > 1)
		{
			QU_ARG_vlong (1, n);
			if (n < 0 || n > l - i)
				n = l - i;
		}
	}

	if (l == 0)
		return QuStringEMPTY;

	if (l == 1)
	{
		char c = self->str[0];
		c = qu_ascii_toupper (c);
		return qu_string_byte (c);
	}

	self = qu_string_dup (self);
	s = self->str;
	while (n--)
	{
		char c = s[i++];

		if (c == ':')
			break;
		if (qu_ascii_ispunct (c) || qu_ascii_isspace (c))
		{
			if (n < 1)
				break;
			s[i] = qu_ascii_toupper (s[i]);
			--n;
		}
		else if (k == 0)
		{
			s[i - 1] = qu_ascii_toupper (c);
			k = 1;
		}
	}
	return self;
}



/*----------------------------------------------------------------------------
	specific instance methods - tests
----------------------------------------------------------------------------*/

/*) doc tests
	The String class provides a number of character test methods.

	They all take the form:
	@def
		sub isxxx         -> Boolean
		sub isxxx (index) -> Boolean

	@p
	The first form returns true if all characters satisfy the test, return
	false otherwise (including if the String is empty).

	The second return true if the character at the specified index statisfies
	the test. A negative index means to count from the end. An index out of bound
	returns false.

	See also #{constants}.

	@def
		sub isalnum  -> Boolean  # A-Z a-z 0-9
		sub isalpha  -> Boolean  # A-Z a-z
		sub isbdigit -> Boolean  # 0-1
		sub isblank  -> Boolean  # space \t
		sub iscntrl  -> Boolean  # \x01-\x1f \x7f-\xff
		sub isdigit  -> Boolean  # 0-9
		sub isgraph  -> Boolean  # '!' to '~'
		sub ishi     -> Boolean  # \x80-\xff
		sub isident  -> Boolean  # isalpha + '_'
		sub isidnum  -> Boolean  # isalnum + '_'
		sub islo     -> Boolean  # \x00-\x7f
		sub islower  -> Boolean  # a-z
		sub isname   -> Boolean  # alpha + ' ' + '.'
		sub isodigit -> Boolean  # 0-7
		sub ispath   -> Boolean  # valid file path characters
		sub isprint  -> Boolean  # ' ' to '~'
		sub ispunct  -> Boolean  # punctuations within ' ' to '~'
		sub ispwd    -> Boolean  # valid password characters
		sub isspace  -> Boolean  # iswhite + \n
		sub istext   -> Boolean  # isprint + \t + \n
		sub isupper  -> Boolean  # A-Z
		sub iswhite  -> Boolean  # space \t \v \f \r
		sub isword   -> Boolean  # except \x00 and isspace
		sub isxdigit -> Boolean  # 0-9 A-F a-f

	@noexc

	@xmp
		x = 'abc?'
		println (x.isalnum)

		x = 'ABC123'
		println (x.isalnum)

	@out
		false
		true
*/

unsigned char QU_ASCII_ISHI [256] =
{
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
};


unsigned char QU_ASCII_ISLO [256] =
{
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};


static qu_boolean _string_test (qu_string self, int argc, qu_any *argv, const unsigned char *cset)
{
	long l = self->len;

	if (l == 0)
		return QuFalse;
	else
	{
		char *z = self->str;
		if (argc == 1 && qu_integer_is (argv[0]))
		{
			long i = qu_integer_toli (argv[0]);
			if (errno)
				return QuFalse;
			if (i < 0)
				i += l;
			if (i < 0 || i >= l)
				return QuFalse;
			return cset [255 & z[i]] ? QuTrue : QuFalse;
		}
		else
		{
			while (l--)
			{
				if (!cset [255 & *z++])
					return QuFalse;
			}
			return QuTrue;
		}
	}
}

qu_boolean qu__string_isalnum  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISALNUM ); }
qu_boolean qu__string_isalpha  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISALPHA ); }
qu_boolean qu__string_isbdigit (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISBDIGIT); }
qu_boolean qu__string_isblank  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISBLANK ); }
qu_boolean qu__string_iscntrl  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISCNTRL ); }
qu_boolean qu__string_isdigit  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISDIGIT ); }
qu_boolean qu__string_isgraph  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISGRAPH ); }
qu_boolean qu__string_ishi     (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISHI    ); }
qu_boolean qu__string_isident  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISIDENT ); }
qu_boolean qu__string_isidnum  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISIDNUM ); }
qu_boolean qu__string_islo     (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISLO    ); }
qu_boolean qu__string_islower  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISLOWER ); }
qu_boolean qu__string_isname   (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISNAME  ); }
qu_boolean qu__string_isodigit (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISODIGIT); }
qu_boolean qu__string_ispath   (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISPATH  ); }
qu_boolean qu__string_isprint  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISPRINT ); }
qu_boolean qu__string_ispwd    (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISPWD   ); }
qu_boolean qu__string_ispunct  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISPUNCT ); }
qu_boolean qu__string_isspace  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISSPACE ); }
qu_boolean qu__string_istext   (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISTEXT  ); }
qu_boolean qu__string_isupper  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISUPPER ); }
qu_boolean qu__string_isxdigit (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISXDIGIT); }
qu_boolean qu__string_iswhite  (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISWHITE ); }
qu_boolean qu__string_isword   (qu_string self, int argc, qu_any *argv)  { return _string_test (self, argc, argv, QU_ASCII_ISWORD  ); }



/*----------------------------------------------------------------------------
	interface - hash
----------------------------------------------------------------------------*/

long qu_string_hash (qu_string self)
{
	if (self->hash != -1)
		return self->hash;
	self->hash = qu_strz_hash (self->str, self->len);
	return self->hash;
}



/*-----------------------------------------------------------------------------
	inteface - print
-----------------------------------------------------------------------------*/

static int _string_print_tick_quoted (qu_string self)
{
	if (!qu_sys_printz ("`", 1))
		return 0;

	if (!qu_sys_printz (self->str, self->len))
		return 0;

	if (!qu_sys_printz ("`", 1))
		return 0;
	return 1;
}



static int _string_print_single_or_double_quoted (qu_string self, char quot)
{
	char b [128], *q = b, c;
	long i, n;

	if (!qu_sys_printz (&quot, 1))
		return 0;

	for (i = 0; i < self->len; ++i)
	{
		c = self->str[i];

		/* account for the longest escape sequence */
		n = q - b;
		if (n >= sizeof (b) - 4)
		{
			if (!qu_sys_printz (b, n))
				return 0;
			q = b;
		}

		switch (c)
		{
			default:
				if (32 <= c && c <= 126) *q++ = c;
				else { sprintf (q, "\\x%02x", (unsigned char) c); q += 4; }
				continue;

			case '\'':
				if (quot == '\'') { *q++ = '\\'; *q++ = '\''; }
				else *q++ = '\'';
				continue;

			case '"':
				if (quot == '"') { *q++ = '\\'; *q++ = '"'; }
				else *q++ = '"';
				continue;

			case '\\': *q++ = '\\'; *q++ = '\\'; continue;
			case '\t': *q++ = '\\'; *q++ = 't';  continue;
			case '\v': *q++ = '\\'; *q++ = 'v';  continue;
			case '\f': *q++ = '\\'; *q++ = 'f';  continue;
			case '\r': *q++ = '\\'; *q++ = 'r';  continue;
			case '\n': *q++ = '\\'; *q++ = 'n';  continue;
			case '\a': *q++ = '\\'; *q++ = 'a';  continue;
			case '\b': *q++ = '\\'; *q++ = 'b';  continue;
			case '\e': *q++ = '\\'; *q++ = 'e';  continue;
		}
	}
	n = q - b;
	if (n)
	{
		if (!qu_sys_printz (b, n))
			return 0;
	}
	if (qu_sys_printz (&quot, 1) == 0)
		return 0;
	return 1;
}



static int _string_print_quoted (qu_string self, char quot)
{
	if (self->len == 0)
	{
		char buf[2] = {quot, quot};
		return qu_sys_printz (buf, 2) != 0;
	}
	else if (quot == '`')
	{
		long  l = self->len;
		char *s = self->str;

		while (l--)
		{
			if (!qu_ascii_istext (*s))
				return _string_print_single_or_double_quoted (self, '\'');
			++s;
		}
		return _string_print_tick_quoted (self);
	}
	else
	{
		return _string_print_single_or_double_quoted (self, quot);
	}
}



/*
	Print qu_string to stdout in quoted or plain form.
*/
qu_null qu_string_print (qu_string self, char quot)
{
	if (quot == 0)
	{
		QU_ENSURE (qu_sys_printz (self->str, self->len));
	}
	else
	{
		QU_ENSURE (_string_print_quoted (self, quot));
	}
	return QuNil;
}



/*-----------------------------------------------------------------------------
	interface - format
-----------------------------------------------------------------------------*/

static qu_string qu__string_qformat (qu_string self)
{
	long       l = self->len;
	char      *s = self->str;
	long       n = 0;
	qu_string  r;
	char      *q;

	while (l--)
	{
		if (*s == '\'' || *s == '"')
			++n;
		++n;
		++s;
	}

	r = qu_string_newl (n);
	q = r->str;
	l = self->len;
	s = self->str;

	while (l--)
	{
		if (*s == '\'')
		{
			*q++ = '\\';
			*q++ = '\'';
			++s;
		}
		else if (*s == '"')
		{
			*q++ = '\\';
			*q++ = '"';
			++s;
		}
		else
		{
			*q++ = *s++;
		}
	}
	return r;
}



qu_string qu_string_format (qu_string self, qu_format_s *f)
{
	qu_string s = self;

	switch (f->mode)
	{
		case 'c': s = qu_string_byte (s->str[0]); break;
		case 'h': s = qu__string_htescape (s); break;
		case 'q': s = qu__string_qformat  (s); break;
		case 'U': s = qu__string_toupper  (s); break;
		case 'L': s = qu__string_tolower  (s); break;
		case 'p': return qu_format_ptr (f, s);
		case 'y': return qu_format_yesno (f, s);
	}
	f->mode = 's';
	return qu_format_align (f, (qu_string) s);
}



/*-----------------------------------------------------------------------------
	interface - tools
-----------------------------------------------------------------------------*/

qu_string qu_string_pad (qu_string self, long left, long right, char fill)
{
    qu_string r;

	if (left < 0)
		left = 0;

	if (right < 0)
		right = 0;

	if (left == 0 && right == 0)
		return self;

	QU_MEM_CHECK (self->len < QU_MEM_MAX - left - right);
	r = qu_string_newl (left + self->len + right);

	if (left)
		memset (r->str, fill, left);
	memcpy (r->str + left, self->str, self->len);
	if (right)
		memset (r->str + left + self->len, fill, right);
    return r;
}



int qu_string_eq (qu_string self, qu_string x)
{
	if (self->len == x->len)
	{
		if (self->hash != -1 && x->hash != -1 && x->hash != self->hash)
			return 0;

		if (memcmp (self->str, x->str, self->len) == 0)
			return 1;
	}
	return 0;
}



qu_string qu_string_delete (qu_string self, qu_string what)
{
	return qu__string_replace (self, 1, (qu_any *) &what);
}



qu_string qu_string_interval (qu_string self, qu_string delim, long intvl)
{
	return qu__string_interval (self, delim, qu_integer_newl (intvl));
}



void qu_string_cfunc_name (const char *func, char *buf, size_t size)
{
	if (strlen (func) + 2 >= size)
	{
		memcpy (buf, func, size - 1);
		buf [size - 1] = '\0';
	}
	else
	{
		char *i = (char *) func;
		char *p = buf;

		if ((*i == 'q' || *i == 'Q') && i[1] == 'u' && i[2] == '_')
			i += 3;

		while (*i && !qu_ascii_isalpha (*i))
			++i;

		if (*i)
		{
			*p++ = qu_ascii_toupper (*i);
			++i;
		}

		while (qu_ascii_isalpha (*i))
			*p++ = *i++;

		if (*i == '_')
		{
			++i;
			*p++ = '.';
			if (*i == 'd' && i[1] == 'o' && i[2] == '_')
				i += 2;
			if (*i == '_')
			{
				++i;
				*p++ = '_';
				*p++ = '_';
			}
		}

		while (*i)
			*p++ = *i++;
		*p = '\0';
	}
}



/*----------------------------------------------------------------------------
	init
----------------------------------------------------------------------------*/

static char Characters [512];

void qu_string__setup ()
{
	qu_class_setup_final (QuStringClass, qu_string_s, qu__string__destroy, 0);

	StaticBuckets = qu_mem_calloc (STRING_STATIC_INIT_SIZE, sizeof (entry_s));

	QuStringEMPTY         = qu_gc_new_static (QuStringClass);
	QuStringEMPTY->flags |= QU_STRING_FLAG_STATIC;
	QuStringEMPTY->len    = 0;
	QuStringEMPTY->hash   = 0;
	QuStringEMPTY->str    = "";

	{
		qu_string s;
		int    i;

		for (i = 0; i < 256; i++)
		{
			Characters [i * 2] = i;
			Characters [i * 2 + 1] = '\0';

			s         = qu_gc_new_static (QuStringClass);
			s->flags |= QU_STRING_FLAG_STATIC;
			s->len    = 1;
			s->hash   = i;
			s->str    = &Characters [i * 2];

			__QuStringByteList [i] = s;
		}
	}
}



void qu_string__init ()
{
	/* standard class methods */
	qu_class_method_class  (QuStringClass, QU_NAME_new, qu__string_new, 1);

	/* specific class methods */
	qu_class_method_classz (QuStringClass, "chr"  , qu__string_chr ,  1);
	qu_class_method_classz (QuStringClass, "make" , qu__string_make, -1);

	/* standard instance methods */
	qu_class_method_inst  (QuStringClass, QU_NAME__as      , qu__string__as      ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__cmp     , qu__string__cmp     ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__each    , qu__string__each    ,  0);
	qu_class_method_inst  (QuStringClass, QU_NAME__eq      , qu__string__eq      ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__format  , qu__string__format  ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__getitem , qu__string__getitem ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__getslice, qu__string__getslice,  2);
	qu_class_method_inst  (QuStringClass, QU_NAME__hash    , qu__string__hash    ,  0);
	qu_class_method_inst  (QuStringClass, QU_NAME__iand    , qu__string__iand    ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__in      , qu__string__in      ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__ior     , qu__string__ior     ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__ixor    , qu__string__ixor    ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__len     , qu__string__len     ,  0);
	qu_class_alias        (QuStringClass, QU_NAME__len     , QU_NAME_len);
	qu_class_method_inst  (QuStringClass, QU_NAME__like    , qu__string__like    ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__lshift  , qu__string__lshift  ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__minus   , qu__string__minus   ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__mul     , qu__string__mul     ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__plus    , qu__string__plus    ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__print   , qu__string__print   ,  0);
	qu_class_method_inst  (QuStringClass, QU_NAME__rem     , qu__string__rem     ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__rshift  , qu__string__rshift  ,  1);
	qu_class_method_inst  (QuStringClass, QU_NAME__string  , qu__string__string  ,  0);
	qu_class_method_inst  (QuStringClass, QU_NAME__true    , qu__string__true    ,  0);

	/* specific instance methods */
	qu_class_method_instz (QuStringClass, "asc"        , qu__string_asc         ,  0);
	qu_class_method_instz (QuStringClass, "bytes"      , qu__string_bytes       , -1);
	qu_class_method_instz (QuStringClass, "capitalize" , qu__string_capitalize  ,  0);
	qu_class_method_instz (QuStringClass, "center"     , qu__string_center      , -1);
	qu_class_method_instz (QuStringClass, "chop"       , qu__string_chop        , -1);
	qu_class_method_instz (QuStringClass, "chomp"      , qu__string_chomp       , -1);
	qu_class_method_instz (QuStringClass, "contains"   , qu__string_contains    , -1);
	qu_class_method_instz (QuStringClass, "count"      , qu__string_count       , -1);
	qu_class_method_instz (QuStringClass, "eachline"   , qu__string_eachline    ,  0);
	qu_class_method_instz (QuStringClass, "eachsection", qu__string_eachsection ,  1);
	qu_class_method_instz (QuStringClass, "eachsegment", qu__string_eachsegment ,  1);
	qu_class_method_instz (QuStringClass, "endwith"    , qu__string_endwith     , -1);
	qu_class_method_instz (QuStringClass, "endswith"   , qu__string_endswith    , -1);
	qu_class_method_instz (QuStringClass, "extract"    , qu__string_extract     , -1);
	qu_class_method_instz (QuStringClass, "fixline"    , qu__string_fixline     , -1);
	qu_class_method_instz (QuStringClass, "flatten"    , qu__string_flatten     ,  0);
	qu_class_method_instz (QuStringClass, "head"       , qu__string_head        , -1);
	qu_class_method_instz (QuStringClass, "htescape"   , qu__string_htescape    ,  0);
	qu_class_method_instz (QuStringClass, "indent"     , qu__string_indent      , -1);
	qu_class_method_instz (QuStringClass, "index"      , qu__string_index       , -1);
	qu_class_method_instz (QuStringClass, "inset"      , qu__string_inset       , -1);
	qu_class_method_instz (QuStringClass, "interval"   , qu__string_interval    ,  2);
	qu_class_method_instz (QuStringClass, "level"      , qu__string_level       , -1);
	qu_class_method_instz (QuStringClass, "linecount"  , qu__string_linecount   ,  0);
	qu_class_method_instz (QuStringClass, "lines"      , qu__string_lines       ,  0);
	qu_class_method_instz (QuStringClass, "ljust"      , qu__string_ljust       , -1);
	qu_class_method_instz (QuStringClass, "ltrim"      , qu__string_ltrim       ,  0);
	qu_class_method_instz (QuStringClass, "match"      , qu__string_match       , -1);
	qu_class_method_instz (QuStringClass, "ord"        , qu__string_ord         ,  0);
	qu_class_method_instz (QuStringClass, "pad"        , qu__string_pad         , -1);
	qu_class_method_instz (QuStringClass, "parseform"  , qu__string_parseform   , -1);
	qu_class_method_instz (QuStringClass, "quote"      , qu__string_quote       , -1);
	qu_class_method_instz (QuStringClass, "rcount"     , qu__string_rcount      , -1);
	qu_class_method_instz (QuStringClass, "replace"    , qu__string_replace     , -1);
	qu_class_method_instz (QuStringClass, "reverse"    , qu__string_reverse     ,  0);
	qu_class_method_instz (QuStringClass, "rindex"     , qu__string_rindex      , -1);
	qu_class_method_instz (QuStringClass, "rjust"      , qu__string_rjust       , -1);
	qu_class_method_instz (QuStringClass, "rtrim"      , qu__string_rtrim       ,  0);
	qu_class_method_instz (QuStringClass, "scrap"      , qu__string_scrap       , -1);
	qu_class_method_instz (QuStringClass, "split"      , qu__string_split       , -1);
	qu_class_method_instz (QuStringClass, "splitat"    , qu__string_splitat     , -1);
	qu_class_method_instz (QuStringClass, "splitkv"    , qu__string_splitkv     , -1);
	qu_class_method_instz (QuStringClass, "splitnum"   , qu__string_splitnum    , -1);
	qu_class_method_instz (QuStringClass, "startwith"  , qu__string_startwith   , -1);
	qu_class_method_instz (QuStringClass, "startswith" , qu__string_startswith  , -1);
	qu_class_method_instz (QuStringClass, "start"      , qu__string_start       , -1);
	qu_class_method_instz (QuStringClass, "strip"      , qu__string_strip       , -1);
	qu_class_method_instz (QuStringClass, "swapcase"   , qu__string_swapcase    ,  0);
	qu_class_method_instz (QuStringClass, "tail"       , qu__string_tail        , -1);
	qu_class_method_instz (QuStringClass, "title"      , qu__string_title       ,  0);
	qu_class_method_instz (QuStringClass, "translate"  , qu__string_translate   , -1);
	qu_class_method_instz (QuStringClass, "trim"       , qu__string_trim        ,  0);
	qu_class_method_instz (QuStringClass, "unfold"     , qu__string_unfold      ,  0);
	qu_class_method_instz (QuStringClass, "unquote"    , qu__string_unquote     ,  0);
	qu_class_method_instz (QuStringClass, "unindent"   , qu__string_unindent    , -1);
	qu_class_method_instz (QuStringClass, "words"      , qu__string_words       ,  0);

	qu_class_method_instz (QuStringClass, "toident"    , qu__string_toident     ,  0);
	qu_class_method_instz (QuStringClass, "tolower"    , qu__string_tolower     ,  0);
	qu_class_method_instz (QuStringClass, "toprint"    , qu__string_toprint     ,  0);
	qu_class_method_instz (QuStringClass, "toupper"    , qu__string_toupper     ,  0);
	qu_class_method_instz (QuStringClass, "toheader"   , qu__string_toheader    , -1);

	qu_class_method_instz (QuStringClass, "isalnum"    , qu__string_isalnum     , -1);
	qu_class_method_instz (QuStringClass, "isalpha"    , qu__string_isalpha     , -1);
	qu_class_method_instz (QuStringClass, "isbdigit"   , qu__string_isbdigit    , -1);
	qu_class_method_instz (QuStringClass, "isblank"    , qu__string_isblank     , -1);
	qu_class_method_instz (QuStringClass, "iscntrl"    , qu__string_iscntrl     , -1);
	qu_class_method_instz (QuStringClass, "isdigit"    , qu__string_isdigit     , -1);
	qu_class_method_instz (QuStringClass, "isgraph"    , qu__string_isgraph     , -1);
	qu_class_method_instz (QuStringClass, "ishi"       , qu__string_ishi        , -1);
	qu_class_method_instz (QuStringClass, "isident"    , qu__string_isident     , -1);
	qu_class_method_instz (QuStringClass, "isidnum"    , qu__string_isidnum     , -1);
	qu_class_method_instz (QuStringClass, "islo"       , qu__string_islo        , -1);
	qu_class_method_instz (QuStringClass, "islower"    , qu__string_islower     , -1);
	qu_class_method_instz (QuStringClass, "isname"     , qu__string_isname      , -1);
	qu_class_method_instz (QuStringClass, "isodigit"   , qu__string_isodigit    , -1);
	qu_class_method_instz (QuStringClass, "ispath"     , qu__string_ispath      , -1);
	qu_class_method_instz (QuStringClass, "isprint"    , qu__string_isprint     , -1);
	qu_class_method_instz (QuStringClass, "ispwd"      , qu__string_ispwd       , -1);
	qu_class_method_instz (QuStringClass, "ispunct"    , qu__string_ispunct     , -1);
	qu_class_method_instz (QuStringClass, "isspace"    , qu__string_isspace     , -1);
	qu_class_method_instz (QuStringClass, "istext"     , qu__string_istext      , -1);
	qu_class_method_instz (QuStringClass, "isupper"    , qu__string_isupper     , -1);
	qu_class_method_instz (QuStringClass, "isxdigit"   , qu__string_isxdigit    , -1);
	qu_class_method_instz (QuStringClass, "iswhite"    , qu__string_iswhite     , -1);
	qu_class_method_instz (QuStringClass, "isword"     , qu__string_isword      , -1);

	QuStringALL8   = qu_string_new_static  (ALL8, 256);
	QuStringALL7   = qu_string_new_static  (ALL7, 128);
	QuStringALNUM  = qu_string_new_staticz (ALNUM);
	QuStringALPHA  = qu_string_new_staticz (ALPHA);
	QuStringCNTRL  = qu_string_new_staticz (CNTRL);
	QuStringBDIGIT = qu_string_new_staticz (BDIGIT);
	QuStringBLANK  = qu_string_new_staticz (BLANK);
	QuStringDIGIT  = qu_string_new_staticz (DIGIT);
	QuStringGRAPH  = qu_string_new_staticz (GRAPH);
	QuStringIDENT  = qu_string_new_staticz (IDENT);
	QuStringIDNUM  = qu_string_new_staticz (IDNUM);
	QuStringLOWER  = qu_string_new_staticz (LOWER);
	QuStringNAME   = qu_string_new_staticz (NAME);
	QuStringODIGIT = qu_string_new_staticz (ODIGIT);
	QuStringPATH   = qu_string_new_staticz (PATH);
	QuStringPRINT  = qu_string_new_staticz (PRINT);
	QuStringPUNCT  = qu_string_new_staticz (PUNCT);
	QuStringPWD    = qu_string_new_staticz (PWD);
	QuStringSPACE  = qu_string_new_staticz (SPACE);
	QuStringUPPER  = qu_string_new_staticz (UPPER);
	QuStringXDIGIT = qu_string_new_staticz (XDIGIT);
	QuStringWHITE  = qu_string_new_staticz (WHITE);
	QuStringWORD   = qu_string_new_staticz (WORD);

	qu_class_constz (QuStringClass, "ALL8"  , QuStringALL8);
	qu_class_constz (QuStringClass, "ALL7"  , QuStringALL7);
	qu_class_constz (QuStringClass, "ALNUM" , QuStringALNUM);
	qu_class_constz (QuStringClass, "ALPHA" , QuStringALPHA);
	qu_class_constz (QuStringClass, "CNTRL" , QuStringCNTRL);
	qu_class_constz (QuStringClass, "BDIGIT", QuStringBDIGIT);
	qu_class_constz (QuStringClass, "BLANK" , QuStringBLANK);
	qu_class_constz (QuStringClass, "DIGIT" , QuStringDIGIT);
	qu_class_constz (QuStringClass, "GRAPH" , QuStringGRAPH);
	qu_class_constz (QuStringClass, "IDENT" , QuStringIDENT);
	qu_class_constz (QuStringClass, "IDNUM" , QuStringIDNUM);
	qu_class_constz (QuStringClass, "LOWER" , QuStringLOWER);
	qu_class_constz (QuStringClass, "NAME"  , QuStringNAME);
	qu_class_constz (QuStringClass, "ODIGIT", QuStringODIGIT);
	qu_class_constz (QuStringClass, "PATH"  , QuStringPATH);
	qu_class_constz (QuStringClass, "PRINT" , QuStringPRINT);
	qu_class_constz (QuStringClass, "PUNCT" , QuStringPUNCT);
	qu_class_constz (QuStringClass, "PWD"   , QuStringPWD);
	qu_class_constz (QuStringClass, "SPACE" , QuStringSPACE);
	qu_class_constz (QuStringClass, "UPPER" , QuStringUPPER);
	qu_class_constz (QuStringClass, "XDIGIT", QuStringXDIGIT);
	qu_class_constz (QuStringClass, "WHITE" , QuStringWHITE);
	qu_class_constz (QuStringClass, "WORD"  , QuStringWORD);

	/* finalize declaration */
	qu_class_enddecl (QuStringClass);
}
